/* Generated by Cython 0.25.2 */

/* BEGIN: Cython Metadata
{
    "distutils": {},
    "module_name": "pymola.generated.ModelicaParser"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03020000)
    #error Cython requires Python 2.6+ or Python 3.2+.
#else
#define CYTHON_ABI "0_25_2"
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x03030000 || (PY_MAJOR_VERSION == 2 && PY_VERSION_HEX >= 0x02070000)
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_FASTCALL
  #define METH_FASTCALL 0x80
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject **args,
                                              Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : PyInstanceMethod_New(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#endif

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__pymola__generated__ModelicaParser
#define __PYX_HAVE_API__pymola__generated__ModelicaParser
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#ifdef PYREX_WITHOUT_ASSERTIONS
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER) && defined (_M_X64)
    #define __Pyx_sst_abs(value) _abs64(value)
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyObject_AsSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#if PY_MAJOR_VERSION < 3
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u)
{
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#else
#define __Pyx_Py_UNICODE_strlen Py_UNICODE_strlen
#endif
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
#define __Pyx_PyBool_FromLong(b) ((b) ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False))
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c));
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */

static PyObject *__pyx_m;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "pymola/generated/ModelicaParser.pyx",
};

/*--- Type declarations ---*/
struct __pyx_defaults;
typedef struct __pyx_defaults __pyx_defaults;
struct __pyx_defaults {
  PyObject *__pyx_arg_output;
};

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* GetModuleGlobalName.proto */
static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name);

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PyObjectLookupSpecial.proto */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x02070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name) {
    PyObject *res;
    PyTypeObject *tp = Py_TYPE(obj);
#if PY_MAJOR_VERSION < 3
    if (unlikely(PyInstance_Check(obj)))
        return __Pyx_PyObject_GetAttrStr(obj, attr_name);
#endif
    res = _PyType_Lookup(tp, attr_name);
    if (likely(res)) {
        descrgetfunc f = Py_TYPE(res)->tp_descr_get;
        if (!f) {
            Py_INCREF(res);
        } else {
            res = f(res, obj, (PyObject *)tp);
        }
    } else {
        PyErr_SetObject(PyExc_AttributeError, attr_name);
    }
    return res;
}
#else
#define __Pyx_PyObject_LookupSpecial(o,n) __Pyx_PyObject_GetAttrStr(o,n)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = PyThreadState_GET();
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o,n,NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, long intval, int inplace);
#else
#define __Pyx_PyInt_EqObjC(op1, op2, intval, inplace)\
    PyObject_RichCompare(op1, op2, Py_EQ)
    #endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, long intval, int inplace);
#else
#define __Pyx_PyInt_SubtractObjC(op1, op2, intval, inplace)\
    (inplace ? PyNumber_InPlaceSubtract(op1, op2) : PyNumber_Subtract(op1, op2))
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* SetItemInt.proto */
#define __Pyx_SetItemInt(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_Fast(o, (Py_ssize_t)i, v, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_SetItemInt_Generic(o, to_py_func(i), v)))
static CYTHON_INLINE int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v);
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v,
                                               int is_list, int wraparound, int boundscheck);

/* CalculateMetaclass.proto */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* CythonFunction.proto */
#define __Pyx_CyFunction_USED 1
#include <structmember.h>
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_NewEx(ml, flags, qualname, self, module, globals, code)\
    __Pyx_CyFunction_New(__pyx_CyFunctionType, ml, flags, qualname, self, module, globals, code)
static PyObject *__Pyx_CyFunction_New(PyTypeObject *, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

/* Py3ClassCreate.proto */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name, PyObject *qualname,
                                           PyObject *mkw, PyObject *modname, PyObject *doc);
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases, PyObject *dict,
                                      PyObject *mkw, int calculate_metaclass, int allow_py2_metaclass);

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CyFunctionClassCell.proto */
static CYTHON_INLINE int __Pyx_CyFunction_InitClassCell(PyObject *cyfunctions, PyObject *classobj);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CStringEquals.proto */
static CYTHON_INLINE int __Pyx_StrEq(const char *, const char *);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);


/* Module declarations from 'pymola.generated.ModelicaParser' */
#define __Pyx_MODULE_NAME "pymola.generated.ModelicaParser"
int __pyx_module_is_main_pymola__generated__ModelicaParser = 0;

/* Implementation of 'pymola.generated.ModelicaParser' */
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_super;
static const char __pyx_k_[] = "\003\346\202\213\352\234\252\350\204\263\353\247\255\344\205\274\343\257\247\347\236\206\345\245\244\003^";
static const char __pyx_k_0[] = "\003\002\002\002\306\236\306\237\0050\031\002\306\237+\003\002\002\002\306\240\306\242";
static const char __pyx_k_1[] = "\002\002\002\306\260\306\261\003\002\002\002\306\2611\003\002\002\002\306\262\306\260";
static const char __pyx_k_2[] = "\003\002\002\002\312\207\312\205\003\002\002\002\312\210\312\216\0072\002\002\312\211";
static const char __pyx_k_3[] = "\002\002\002\313\212\313\213\007\020\002\002\313\213\313\214\0073\002\002\313\214";
static const char __pyx_k_6[] = "\002\002\002\313\263\313\264\007\020\002\002\313\264\313\265\0076\002\002\313\265";
static const char __pyx_k_7[] = "\307\201\307\200\003\002\002\002\307\201\307\202\003\002\002\002\307\2027\003\002";
static const char __pyx_k_8[] = "\306\277\003\002\002\002\306\277\307\201\003\002\002\002\307\200\307\202\0058\035\002";
static const char __pyx_k_9[] = "\307\210\003\002\002\002\307\214\307\212\003\002\002\002\307\2159\003\002\002\002\307\216";
static const char __pyx_k_A[] = "\003@\003@\007@\316\273\n@\014@\016@\316\276\013@\003@\005@\317\201\n@\003A";
static const char __pyx_k_B[] = "\003\002\002\002>\307\236\003\002\002\002@\307\241\003\002\002\002B\307\252\003\002\002\002";
static const char __pyx_k_C[] = "\316\264\007\025\002\002\316\264}\003\002\002\002\316\265\316\274\005\302\204C\002";
static const char __pyx_k_D[] = "\307\235\307\237\005D#\002\307\236\307\234\003\002\002\002\307\236\307\235\003\002";
static const char __pyx_k_E[] = "\302\210E\002\304\271\304\270\003\002\002\002\304\271\304\272\003\002\002\002\304\272";
static const char __pyx_k_F[] = "F\002\306\234\306\233\003\002\002\002\306\234\306\235\003\002\002\002\306\235\306\236";
static const char __pyx_k_G[] = "\310\205G\003\002\002\002\310\206\310\207\005,\027\002\310\207\310\210\005.\030";
static const char __pyx_k_H[] = "\002\002\002\303\230\303\231\003\002\002\002\303\231\303\232\005\302\216H\002\303\232";
static const char __pyx_k_I[] = "\002\002\002\303\274\303\275\003\002\002\002\303\275\303\276\005\302\220I\002\303\276";
static const char __pyx_k_J[] = "\304\246\005\302\222J\002\304\245\304\244\003\002\002\002\304\245\304\246\003\002";
static const char __pyx_k_L[] = "\307\266\003\002\002\002\307\266\307\274\003\002\002\002\307\267\307\272\005L'\002";
static const char __pyx_k_M[] = "\003\002\002\002\310\244\310\232\003\002\002\002\310\245M\003\002\002\002\310\246\310\250";
static const char __pyx_k_N[] = "\007\032\002\002\304\225\304\231\005\032\016\002\304\226\304\231\005N(\002\304\227";
static const char __pyx_k_O[] = "\002\002\002\310\260O\003\002\002\002\310\261\310\257\003\002\002\002\310\262\310\264";
static const char __pyx_k_P[] = "\304\231\005P)\002\304\230\304\222\003\002\002\002\304\230\304\224\003\002\002\002\304\230";
static const char __pyx_k_Q[] = "\002\002\307\237?\003\002\002\002\307\240\307\242\007Q\002\002\307\241\307\240\003\002";
static const char __pyx_k_R[] = "\007R\002\002\302\256\302\255\003\002\002\002\302\256\302\257\003\002\002\002\302\257";
static const char __pyx_k_S[] = "\007S\002\002\302\242\302\241\003\002\002\002\302\242\302\243\003\002\002\002\302\243";
static const char __pyx_k_T[] = "\310\251\003\002\002\002\310\251\310\257\007,\002\002\310\252\310\253\005T+\002\310\253";
static const char __pyx_k_U[] = "\302\246\005\003\002\002\002\302\247\302\251\007U\002\002\302\250\302\247\003\002\002";
static const char __pyx_k_V[] = "\003\002\002\002\305\212\035\003\002\002\002\305\213\305\215\007V\002\002\305\214\305\213";
static const char __pyx_k_W[] = "\305\220\305\222\003\002\002\002\305\221\305\223\007W\002\002\305\222\305\221\003";
static const char __pyx_k_X[] = "\007X\002\002\305\225\305\224\003\002\002\002\305\225\305\226\003\002\002\002\305\226";
static const char __pyx_k_Y[] = "\007Y\002\002\310\247\310\246\003\002\002\002\310\247\310\250\003\002\002\002\310\250";
static const char __pyx_k_Z[] = "Z\311\255\003\002\002\002\\\312\226\003\002\002\002^\312\277\003\002\002\002`\313\215";
static const char __pyx_k_c[] = "\003\002\002\002\313\241c\003\002\002\002\313\242\313\240\003\002\002\002\313\243\313\246";
static const char __pyx_k_e[] = "\003\002\002\002\313\246\313\247\003\002\002\002\313\247e\003\002\002\002\313\250\313\251";
static const char __pyx_k_i[] = "i";
static const char __pyx_k_o[] = "\315\213\314\271\003\002\002\002\315\213\314\272\003\002\002\002\315\214o\003\002\002";
static const char __pyx_k_p[] = "_p";
static const char __pyx_k_q[] = "\315\216\003\002\002\002\315\224\315\225\003\002\002\002\315\225q\003\002\002\002\315\226";
static const char __pyx_k_r[] = "\002\315\215\315\224\005r:\002\315\216\315\217\007\024\002\002\315\217\315\222";
static const char __pyx_k_s[] = "\315\264s\003\002\002\002\315\265\315\263\003\002\002\002\315\266\316\227\007\\\002";
static const char __pyx_k_t[] = "\315\241\005t;\002\315\240\315\226\003\002\002\002\315\240\315\231\003\002\002\002\315\240";
static const char __pyx_k_u[] = "\002\002\002\316\227u\003\002\002\002\316\230\316\232\007\036\002\002\316\231\316\230";
static const char __pyx_k_v[] = "\002\002\303\253\303\254\005v<\002\303\254\303\255\007\027\002\002\303\255\303\262";
static const char __pyx_k_w[] = "\002\002\002\316\240\316\241\003\002\002\002\316\241w\003\002\002\002\316\242\316\240";
static const char __pyx_k_y[] = "\316\244\003\002\002\002\316\245\316\246\003\002\002\002\316\246y\003\002\002\002\316\247";
static const char __pyx_k_z[] = "\005z>\002\304\262\304\263\007\021\002\002\304\263\304\265\003\002\002\002\304\264";
static const char __pyx_k_LA[] = "LA";
static const char __pyx_k_LT[] = "LT";
static const char __pyx_k_WS[] = "WS";
static const char __pyx_k__3[] = "\320\214\004\002\t\002\004\003\t\003\004\004\t\004\004\005\t\005\004\006\t\006\004\007\t\007";
static const char __pyx_k__5[] = "\004\010\t\010\004\t\t\t\004\n\t\n\004\013\t\013\004\014\t\014\004\r\t\r\004\016";
static const char __pyx_k__7[] = "\t\016\004\017\t\017\004\020\t\020\004\021\t\021\004\022\t\022\004\023\t\023";
static const char __pyx_k__9[] = "\004\024\t\024\004\025\t\025\004\026\t\026\004\027\t\027\004\030\t\030\004\031";
static const char __pyx_k_d3[] = "\313\235\007\027\002\002\313\235\313\237\005d3\002\313\236\313\234\003\002\002";
static const char __pyx_k_if[] = "'if'";
static const char __pyx_k_in[] = "'in'";
static const char __pyx_k_io[] = "io";
static const char __pyx_k_la[] = "_la";
static const char __pyx_k_n8[] = "\005n8\002\307\206\307\204\003\002\002\002\307\206\307\207\003\002\002\002\307\207\307\215";
static const char __pyx_k_op[] = "op";
static const char __pyx_k_or[] = "'or'";
static const char __pyx_k_re[] = "re";
static const char __pyx_k_0_T[] = "0\002\002\311\273\311\274\005T+\002\311\274\311\275\007\003\002\002\311\275\311\277";
static const char __pyx_k_0_X[] = "\0070\002\002\312\231\312\232\005X-\002\312\232\312\233\007\003\002\002\312\233";
static const char __pyx_k_1_2[] = "\311\251\007/\002\002\311\244\311\251\005\\/\002\311\245\311\251\005`1\002\311\246";
static const char __pyx_k_1_3[] = "\002\002\002\312\240\312\236\003\002\002\002\312\241\312\242\0071\002\002\312\242";
static const char __pyx_k_2_2[] = "\0052\032\002\306\254\306\255\007\027\002\002\306\255\306\257\0052\032\002";
static const char __pyx_k_2_X[] = "\312\261\312\267\0072\002\002\312\262\312\263\005X-\002\312\263\312\264\007";
static const char __pyx_k_4_7[] = "4.7";
static const char __pyx_k_4_X[] = "\313\225\0074\002\002\313\220\313\221\005X-\002\313\221\313\222\007\003\002\002";
static const char __pyx_k_6_4[] = "\003\002\002\002\306\263\306\265\0056\034\002\306\264\306\266\0054\033\002\306\265";
static const char __pyx_k_6_8[] = "\002\0026\306\274\003\002\002\0028\307\214\003\002\002\002:\307\216\003\002\002\002<\307\224";
static const char __pyx_k_7_i[] = "\314\221\314\222\007\020\002\002\314\222\314\223\0077\002\002\314\223i\003\002";
static const char __pyx_k_7_k[] = "\314\260\007\020\002\002\314\260\314\261\0077\002\002\314\261k\003\002\002\002\314\262";
static const char __pyx_k_8_2[] = "\003\002\002\002\314\200\313\276\003\002\002\002\314\201\314\202\0078\002\002\314\202";
static const char __pyx_k_8_3[] = "\002\002\002\314\236\314\234\003\002\002\002\314\237\314\240\0078\002\002\314\240\314\241";
static const char __pyx_k_9_z[] = "\314\263\0079\002\002\314\263\314\264\007\023\002\002\314\264\314\265\005z>\002\314\265";
static const char __pyx_k_ATN[] = "ATN";
static const char __pyx_k_A_2[] = "\307\246\003\002\002\002\307\250\307\247\003\002\002\002\307\251A\003\002\002\002\307\252";
static const char __pyx_k_A_3[] = "\002\002\002\316\277\317\201\005\302\200A\002\317\200\316\265\003\002\002\002\317\200";
static const char __pyx_k_A_4[] = "A\002\317\222\317\221\003\002\002\002\317\222\317\223\003\002\002\002\317\223\317\224";
static const char __pyx_k_B_2[] = "\316\277\003\002\002\002\317\201\003\002\002\002\317\202\317\207\005\302\202B";
static const char __pyx_k_B_3[] = "\002\317\203\317\204\007\027\002\002\317\204\317\206\005\302\202B\002\317\205\317\203";
static const char __pyx_k_B_F[] = "\307\250\003\002\002\002\307\246\307\251\005B\"\002\307\247\307\251\005F$\002\307\250";
static const char __pyx_k_C_2[] = "\316\266\316\267\007\027\002\002\316\267\316\273\005\302\204C\002\316\270\316\271";
static const char __pyx_k_C_3[] = "\302\204C\002\317\215\302\203\003\002\002\002\317\216\317\217\007\017\002\002\317\217";
static const char __pyx_k_DFA[] = "DFA";
static const char __pyx_k_D_2[] = "\316\203\007\023\002\002\316\203\316\204\005\302\206D\002\316\204\316\205\007\025";
static const char __pyx_k_EOF[] = "EOF";
static const char __pyx_k_E_2[] = "\003\002\002\002\307\274\307\273\003\002\002\002\307\275E\003\002\002\002\307\276\310\201";
static const char __pyx_k_E_E[] = "\005\302\210E\002\316\210\316\211\007\003\002\002\316\211\316\213\005\302\210E\002";
static const char __pyx_k_F_2[] = "\002\002\307\272\307\275\003\002\002\002\307\273\307\275\005F$\002\307\274\307\271";
static const char __pyx_k_G_2[] = "\317\260\005\302\214G\002\317\257\317\255\003\002\002\002\317\260\317\263\003\002";
static const char __pyx_k_H_2[] = "\002\002\002\303\265\303\266\007\025\002\002\303\266\303\267\005\302\216H\002\303\267";
static const char __pyx_k_H_3[] = "\304\220\005\302\216H\002\304\220\023\003\002\002\002\304\221\304\232\005\032\016";
static const char __pyx_k_H_4[] = "\003\002\002\002\305\254\305\255\005(\025\002\305\255\305\256\005\302\216H\002\305\256";
static const char __pyx_k_H_5[] = "\003\002\002\002\306\201\306\202\003\002\002\002\306\202\306\203\005\302\216H\002\306\203";
static const char __pyx_k_H_6[] = "\307\270\307\272\005H%\002\307\271\307\267\003\002\002\002\307\271\307\270\003\002";
static const char __pyx_k_H_7[] = "\002\002\002\310\227\310\230\003\002\002\002\310\230\310\231\005\302\216H\002\310\231";
static const char __pyx_k_H_8[] = "\310\242\310\243\007\025\002\002\310\243\310\245\005\302\216H\002\310\244\310\220";
static const char __pyx_k_H_K[] = "\002\310\213\310\214\005\302\216H\002\310\214K\003\002\002\002\310\215\310\216\005";
static const char __pyx_k_H_Z[] = "\007\025\002\002\303\246\304\203\005\302\216H\002\303\247\303\250\007Z\002\002\303\250";
static const char __pyx_k_I_J[] = "\002\002\002\317\272\317\274\005\302\220I\002\317\273\317\275\005\302\222J\002\317\274";
static const char __pyx_k_I_r[] = "\315\241\003\002\002\002\315\235\315\236\007I\002\002\315\236\315\241\005r:\006\315\237";
static const char __pyx_k_J_2[] = "\304\253\005\302\222J\002\304\253\304\254\007\003\002\002\304\254\304\256\003\002";
static const char __pyx_k_J_3[] = "\003\002\002\002\306\220\306\222\003\002\002\002\306\221\306\223\005\302\222J\002\306\222";
static const char __pyx_k_L_2[] = "\002\315\267\316\227\007[\002\002\315\270\316\227\007L\002\002\315\271\316\227\007";
static const char __pyx_k_L_H[] = "\007\034\002\002\307\277\310\202\005L'\002\310\200\310\202\005H%\002\310\201\307\277";
static const char __pyx_k_M_z[] = "M\002\002\315\272\315\273\005z>\002\315\273\315\274\005|?\002\315\274\316\227";
static const char __pyx_k_N_2[] = "\002\002\316\205\316\227\003\002\002\002\316\206\316\207\007N\002\002\316\207\316\214";
static const char __pyx_k_N_G[] = "N\002\002\317\254\317\261\005\302\214G\002\317\255\317\256\007\027\002\002\317\256";
static const char __pyx_k_O_2[] = "\003\002\002\002\316\217\316\220\007O\002\002\316\220\316\227\003\002\002\002\316\221";
static const char __pyx_k_O_3[] = "\003\002\002\002\317\263\317\261\003\002\002\002\317\264\317\265\007O\002\002\317\265";
static const char __pyx_k_P_2[] = "\320\207\302\221\003\002\002\002\320\210\320\211\007P\002\002\320\211\320\212\005";
static const char __pyx_k_Q_2[] = "\002\002\002\310\274Q\003\002\002\002\310\275\310\273\003\002\002\002\310\276\310\277";
static const char __pyx_k_S_2[] = "\305\220\007S\002\002\305\217\305\216\003\002\002\002\305\217\305\220\003\002\002\002";
static const char __pyx_k_S_3[] = "\003\002\002\002\305\235\305\237\003\002\002\002\305\236\305\240\007S\002\002\305\237";
static const char __pyx_k_S_4[] = "\007S\002\002\307\244\307\243\003\002\002\002\307\244\307\245\003\002\002\002\307\245";
static const char __pyx_k_S_5[] = "\003\002\002\002\307\264\307\266\007S\002\002\307\265\307\264\003\002\002\002\307\265";
static const char __pyx_k_S_6[] = "\311\205\003\002\002\002\311\211\311\206\003\002\002\002\311\212S\003\002\002\002\311\213";
static const char __pyx_k_T_2[] = "\005T+\002\311\261\311\262\007\003\002\002\311\262\311\264\003\002\002\002\311\263\311\260";
static const char __pyx_k_T_3[] = "\312\212\005T+\002\312\212\312\213\007\003\002\002\312\213\312\215\003\002\002\002\312\214";
static const char __pyx_k_T_4[] = "\313\203\005T+\002\313\203\313\204\007\003\002\002\313\204\313\206\003\002\002\002\313\205";
static const char __pyx_k_T_z[] = "\302\224\302\226\007T\002\002\302\225\302\227\005z>\002\302\226\302\225\003\002";
static const char __pyx_k_V_2[] = "\002\002\305\233\305\235\007V\002\002\305\234\305\233\003\002\002\002\305\234\305\235";
static const char __pyx_k_W_2[] = "\305\241\305\243\007W\002\002\305\242\305\241\003\002\002\002\305\242\305\243\003";
static const char __pyx_k_W_3[] = "\003\002\002\002\311\250\311\247\003\002\002\002\311\251W\003\002\002\002\311\252\311\253";
static const char __pyx_k_X_2[] = "\002\002\002\305\243\305\245\003\002\002\002\305\244\305\246\007X\002\002\305\245\305\244";
static const char __pyx_k_X_3[] = "\310\265\003\002\002\002\310\265\310\273\007-\002\002\310\266\310\267\005X-\002\310\267";
static const char __pyx_k_X_4[] = "\313\254\005X-\002\313\254\313\255\007\003\002\002\313\255\313\257\003\002\002\002\313\256";
static const char __pyx_k_Y_2[] = "\007Y\002\002\310\263\310\262\003\002\002\002\310\263\310\264\003\002\002\002\310\264";
static const char __pyx_k_Y_z[] = "\316\200\007Y\002\002\316\200\316\227\005|?\002\316\201\316\227\005z>\002\316\202";
static const char __pyx_k_Z_0[] = "\003\002\002\002\311\202\311\212\005Z.\002\311\203\311\212\005^0\002\311\204\311\212";
static const char __pyx_k_Z_2[] = "\303\215\007\020\002\002\303\215\303\216\007Z\002\002\303\216\304\203\003\002\002\002";
static const char __pyx_k_Z_3[] = "\303\217\303\220\007Z\002\002\303\220\303\221\007\021\002\002\303\221\303\222\005";
static const char __pyx_k_Z_4[] = "\304\203\003\002\002\002\303\233\303\234\007Z\002\002\303\234\303\235\007\021\002\002";
static const char __pyx_k_Z_5[] = "\304\203\003\002\002\002\303\270\303\271\007\030\002\002\303\271\303\273\007Z\002\002";
static const char __pyx_k_Z_6[] = "\303\277\005\024\013\002\303\277\304\200\007\020\002\002\304\200\304\201\007Z\002";
static const char __pyx_k_Z_7[] = "\021\003\002\002\002\304\215\304\213\003\002\002\002\304\216\304\217\007Z\002\002\304\217";
static const char __pyx_k_Z_8[] = "\304\266\304\267\007Z\002\002\304\267\304\271\007\023\002\002\304\270\304\272\005";
static const char __pyx_k_Z_9[] = "\305\260!\003\002\002\002\305\261\306\200\007\035\002\002\305\262\305\263\007Z\002";
static const char __pyx_k_Z_F[] = "Z\002\002\306\275\306\277\005\302\212F\002\306\276\306\275\003\002\002\002\306\276";
static const char __pyx_k_Z_I[] = "\007Z\002\002\303\212\303\213\005\302\220I\002\303\213\303\214\005\024\013\002\303\214";
static const char __pyx_k_Z_Z[] = "\007Z\002\002\303\256\303\257\007\027\002\002\303\257\303\261\007Z\002\002\303\260";
static const char __pyx_k__11[] = "\t\031\004\032\t\032\004\033\t\033\004\034\t\034\004\035\t\035\004\036\t\036";
static const char __pyx_k__13[] = "\004\037\t\037\004 \t \004!\t!\004\"\t\"\004#\t#\004$\t$\004%\t%\004&\t";
static const char __pyx_k__15[] = "&\004'\t'\004(\t(\004)\t)\004*\t*\004+\t+\004,\t,\004-\t-\004.\t.\004";
static const char __pyx_k__21[] = "\n\002\003\002\005\002\302\232\n\002\003\002\007\002\302\235\n\002\014\002\016\002\302\240";
static const char __pyx_k__23[] = "\013\002\003\003\005\003\302\243\n\003\003\003\003\003\003\003\003\004\005\004\302\251\n\004";
static const char __pyx_k__25[] = "\003\004\003\004\003\004\003\005\005\005\302\257\n\005\003\005\003\005\003\006\003\006\003\006\005\006";
static const char __pyx_k__27[] = "\302\266\n\006\003\006\003\006\003\006\005\006\302\273\n\006\003\006\003\006\003\006\003\006\005";
static const char __pyx_k__29[] = "\006\303\201\n\006\003\006\005\006\303\204\n\006\003\006\003\006\005\006\303\210\n\006\003";
static const char __pyx_k__31[] = "\007\003\007\003\007\003\007\003\007\003\007\003\007\003\007\003\007\003\007\003\007\005\007\303\225\n";
static const char __pyx_k__33[] = "\007\003\007\005\007\303\230\n\007\003\007\003\007\003\007\003\007\003\007\003\007\003\007\005\007\303\241";
static const char __pyx_k__35[] = "\n\007\003\007\005\007\303\244\n\007\003\007\003\007\003\007\003\007\003\007\003\007\003\007\003\007";
static const char __pyx_k__37[] = "\003\007\003\007\003\007\007\007\303\261\n\007\014\007\016\007\303\264\013\007\003\007\003\007";
static const char __pyx_k__39[] = "\003\007\003\007\003\007\003\007\005\007\303\274\n\007\003\007\003\007\003\007\003\007\003\007\005\007";
static const char __pyx_k__41[] = "\304\203\n\007\003\010\003\010\003\t\003\t\003\t\007\t\304\212\n\t\014\t\016\t";
static const char __pyx_k__43[] = "\304\215\013\t\003\n\003\n\003\n\003\013\003\013\003\013\003\013\003\013\003\013";
static const char __pyx_k__45[] = "\003\013\007\013\304\231\n\013\014\013\016\013\304\234\013\013\003\013\003\013";
static const char __pyx_k__47[] = "\005\013\304\240\n\013\003\013\005\013\304\243\n\013\003\013\005\013\304\246";
static const char __pyx_k__49[] = "\n\013\003\013\005\013\304\251\n\013\003\013\003\013\003\013\005\013\304\256\n";
static const char __pyx_k__51[] = "\013\003\014\003\014\003\r\003\r\003\r\005\r\304\265\n\r\003\r\003\r\003\r\005\r";
static const char __pyx_k__53[] = "\304\272\n\r\003\r\003\r\003\016\003\016\003\016\007\016\305\201\n\016\014\016";
static const char __pyx_k__55[] = "\016\016\305\204\013\016\003\017\003\017\003\017\003\017\005\017\305\212\n\017";
static const char __pyx_k__57[] = "\003\020\005\020\305\215\n\020\003\020\005\020\305\220\n\020\003\020\005\020\305\223";
static const char __pyx_k__59[] = "\n\020\003\020\005\020\305\226\n\020\003\020\003\020\005\020\305\232\n\020\003";
static const char __pyx_k__61[] = "\021\005\021\305\235\n\021\003\021\005\021\305\240\n\021\003\021\005\021\305\243";
static const char __pyx_k__63[] = "\n\021\003\021\005\021\305\246\n\021\003\021\003\021\003\021\005\021\305\253\n";
static const char __pyx_k__65[] = "\021\003\021\003\021\003\021\005\021\305\260\n\021\003\022\003\022\003\022\003\022";
static const char __pyx_k__67[] = "\003\022\003\022\003\022\003\022\003\022\003\022\003\022\005\022\305\275\n\022\005";
static const char __pyx_k__69[] = "\022\305\277\n\022\005\022\306\201\n\022\003\022\003\022\003\023\003\023\003\023";
static const char __pyx_k__71[] = "\007\023\306\210\n\023\014\023\016\023\306\213\013\023\003\024\003\024\003\024";
static const char __pyx_k__73[] = "\005\024\306\220\n\024\003\024\005\024\306\223\n\024\003\025\003\025\003\025\005";
static const char __pyx_k__75[] = "\025\306\230\n\025\003\026\003\026\003\026\005\026\306\235\n\026\003\026\003\026";
static const char __pyx_k__77[] = "\003\027\005\027\306\242\n\027\003\027\005\027\306\245\n\027\003\027\005\027\306\250";
static const char __pyx_k__79[] = "\n\027\003\030\003\030\003\031\003\031\003\031\007\031\306\257\n\031\014\031\016";
static const char __pyx_k__81[] = "\031\306\262\013\031\003\032\003\032\005\032\306\266\n\032\003\032\003\032\003";
static const char __pyx_k__83[] = "\033\003\033\003\033\003\034\003\034\005\034\306\277\n\034\003\034\005\034\307\202";
static const char __pyx_k__85[] = "\n\034\003\035\003\035\003\035\005\035\307\207\n\035\003\035\003\035\003\035\003";
static const char __pyx_k__87[] = "\035\005\035\307\215\n\035\003\036\003\036\005\036\307\221\n\036\003\036\003\036";
static const char __pyx_k__89[] = "\003\037\003\037\003\037\007\037\307\230\n\037\014\037\016\037\307\233\013\037";
static const char __pyx_k__91[] = "\003 \003 \005 \307\237\n \003!\005!\307\242\n!\003!\005!\307\245\n!\003!\003";
static const char __pyx_k__93[] = "!\005!\307\251\n!\003\"\003\"\005\"\307\255\n\"\003\"\003\"\003#\003#\005#\307\263";
static const char __pyx_k__95[] = "\n#\003#\005#\307\266\n#\003#\003#\005#\307\272\n#\003#\005#\307\275\n#\003";
static const char __pyx_k__97[] = "$\003$\003$\005$\310\202\n$\003$\005$\310\205\n$\003%\003%\003%\003%\003&\003&\003";
static const char __pyx_k__99[] = "&\003'\003'\003'\003'\003'\003'\005'\310\224\n'\003'\005'\310\227";
static const char __pyx_k_alt[] = "_alt";
static const char __pyx_k_and[] = "'and'";
static const char __pyx_k_atn[] = "atn";
static const char __pyx_k_buf[] = "buf";
static const char __pyx_k_ctx[] = "_ctx";
static const char __pyx_k_der[] = "'der'";
static const char __pyx_k_doc[] = "__doc__";
static const char __pyx_k_end[] = "'end'";
static const char __pyx_k_for[] = "'for'";
static const char __pyx_k_get[] = "get";
static const char __pyx_k_not[] = "'not'";
static const char __pyx_k_r_2[] = "\315\227\010:\001\002\315\227\315\230\007:\002\002\315\230\315\241\005r:\013\315\231";
static const char __pyx_k_r_3[] = "r:\n\315\245\315\246\014\010\002\002\315\246\315\247\t\010\002\002\315\247\315\262";
static const char __pyx_k_r_4[] = "\005r:\t\315\250\315\251\014\007\002\002\315\251\315\252\t\t\002\002\315\252\315\262";
static const char __pyx_k_r_5[] = "\005r:\005\315\261\315\242\003\002\002\002\315\261\315\245\003\002\002\002\315\261\315\250";
static const char __pyx_k_r_J[] = "\005r:\010\315\253\315\254\014\005\002\002\315\254\315\255\007J\002\002\315\255\315\262";
static const char __pyx_k_r_K[] = "\005r:\006\315\256\315\257\014\004\002\002\315\257\315\260\007K\002\002\315\260\315\262";
static const char __pyx_k_r_r[] = "\005r:\002\315\220\315\221\007\024\002\002\315\221\315\223\005r:\002\315\222\315\220";
static const char __pyx_k_sys[] = "sys";
static const char __pyx_k_t_t[] = "\315\232\005t;\002\315\232\315\233\t\006\002\002\315\233\315\234\005t;\002\315\234";
static const char __pyx_k_v_2[] = "\306\225\007\"\002\002\306\225\306\227\005v<\002\306\226\306\230\005:\036\002\306\227";
static const char __pyx_k_v_3[] = "\317\220\005v<\002\317\220\317\222\007\023\002\002\317\221\317\223\005\302\200";
static const char __pyx_k_v_F[] = "\016\010\002\303\222\303\224\005v<\002\303\223\303\225\005\302\212F\002\303\224";
static const char __pyx_k_x_x[] = "\316\254\005x=\002\316\250\316\251\007\036\002\002\316\251\316\253\005x=\002\316\252";
static const char __pyx_k_x_z[] = "\003\002\002\002x\316\243\003\002\002\002z\316\247\003\002\002\002|\316\257\003\002\002\002";
static const char __pyx_k_z_2[] = "\002\305\263\305\264\007\021\002\002\305\264\306\201\005z>\002\305\265\305\276";
static const char __pyx_k_z_3[] = "\005z>\002\305\266\305\274\007\036\002\002\305\267\305\275\007\037\002\002\305\270";
static const char __pyx_k_z_4[] = "\005z>\002\306\216\306\220\005:\036\002\306\217\306\216\003\002\002\002\306\217\306\220";
static const char __pyx_k_z_5[] = "\003\002\002\002\306\251\306\252\005z>\002\306\252/\003\002\002\002\306\253\306\260";
static const char __pyx_k_z_6[] = "\311\251\003\002\002\002\311\224\311\225\007\023\002\002\311\225\311\232\005z>\002";
static const char __pyx_k_z_7[] = "\311\226\311\227\007\027\002\002\311\227\311\231\005z>\002\311\230\311\226\003";
static const char __pyx_k_z_8[] = "\307\254\005z>\002\307\253\307\255\0058\035\002\307\254\307\253\003\002\002\002\307\254";
static const char __pyx_k_z_9[] = "\311\236\007\025\002\002\311\236\311\237\007+\002\002\311\237\311\240\005z>\002\311\240";
static const char __pyx_k_z_F[] = "\005\016\010\002\310\221\310\223\005z>\002\310\222\310\224\005\302\212F\002\310\223";
static const char __pyx_k_1_n8[] = "\311\270\311\271\0071\002\002\311\271\311\272\005n8\002\311\272\312\200\007";
static const char __pyx_k_2_n8[] = "\315\211\0072\002\002\315\211\315\212\005n8\002\315\212\315\214\003\002\002\002";
static const char __pyx_k_3_b2[] = "_\003\002\002\002\313\215\313\216\0073\002\002\313\216\313\217\005b2\002\313\217";
static const char __pyx_k_7_n8[] = "\002\002\314\224\314\225\0077\002\002\314\225\314\226\005n8\002\314\226\314\234";
static const char __pyx_k_B_CH[] = "::@B\003\002CH\002\321\257\002\302\231\003\002\002\002\004\302\242\003\002\002\002\006\302\250";
static const char __pyx_k_EACH[] = "EACH";
static const char __pyx_k_J_rK[] = "J\002\003rK\002\004\006\010\n\014\016\020\022\024\026\030\032\034\036 \"$&(*";
static const char __pyx_k_T__0[] = "T__0";
static const char __pyx_k_T__1[] = "T__1";
static const char __pyx_k_T__2[] = "T__2";
static const char __pyx_k_T__3[] = "T__3";
static const char __pyx_k_T__4[] = "T__4";
static const char __pyx_k_T__5[] = "T__5";
static const char __pyx_k_T__6[] = "T__6";
static const char __pyx_k_T__7[] = "T__7";
static const char __pyx_k_T__8[] = "T__8";
static const char __pyx_k_T__9[] = "T__9";
static const char __pyx_k_Z_10[] = "#\003\002\002\002\306\204\306\211\007Z\002\002\306\205\306\206\007\027\002\002\306\206";
static const char __pyx_k_Z_11[] = "\010\005\002\310\216\310\217\007Z\002\002\310\217\310\244\007\021\002\002\310\220\310\221";
static const char __pyx_k_Z_12[] = "\317\212\317\213\007Z\002\002\317\213\317\214\007\021\002\002\317\214\317\215\005";
static const char __pyx_k__101[] = "\n'\003'\003'\003'\003'\003'\005'\310\236\n'\003'\005'\310\241\n";
static const char __pyx_k__103[] = "'\003'\003'\005'\310\245\n'\003(\005(\310\250\n(\003(\003(\003(\003(\007";
static const char __pyx_k__105[] = "(\310\256\n(\014(\016(\310\261\013(\003)\005)\310\264\n)\003)\003)\003)\003";
static const char __pyx_k__107[] = ")\007)\310\272\n)\014)\016)\310\275\013)\003*\003*\003*\003*\003*\003*\003*\003";
static const char __pyx_k__109[] = "*\003*\003*\003*\005*\311\212\n*\003+\003+\003+\003,\003,\003,\003,\005,\311\223\n";
static const char __pyx_k__111[] = ",\003,\003,\003,\003,\007,\311\231\n,\014,\016,\311\234\013,\003,\003,\003,\003";
static const char __pyx_k__113[] = ",\003,\003,\003,\003,\003,\003,\003,\005,\311\251\n,\003-\003-\003-\003.\003.\003.\003";
static const char __pyx_k__115[] = ".\003.\003.\007.\311\264\n.\014.\016.\311\267\013.\003.\003.\003.\003.\003.\003";
static const char __pyx_k__117[] = ".\007.\311\277\n.\014.\016.\312\202\013.\007.\312\204\n.\014.\016.\312\207";
static const char __pyx_k__119[] = "\013.\003.\003.\003.\003.\007.\312\215\n.\014.\016.\312\220\013.\005.\312\222";
static const char __pyx_k__121[] = "\n.\003.\003.\003.\003/\003/\003/\003/\003/\003/\007/\312\235\n/\014/\016/\312\240";
static const char __pyx_k__123[] = "\013/\003/\003/\003/\003/\003/\003/\007/\312\250\n/\014/\016/\312\253\013/\007";
static const char __pyx_k__125[] = "/\312\255\n/\014/\016/\312\260\013/\003/\003/\003/\003/\007/\312\266\n/\014";
static const char __pyx_k__144[] = ":\003:\003:\003:\003:\003:\003:\003:\003:\003:\005:\315\241\n:\003:\003:\003:\003:\003";
static const char __pyx_k__146[] = ":\003:\003:\003:\003:\003:\003:\003:\003:\003:\003:\007:\315\262\n:\014:\016:\315\265";
static const char __pyx_k__148[] = "\013:\003;\003;\003;\003;\003;\003;\003;\003;\003;\003;\003;\003;\003;\003;\003;\003;\003";
static const char __pyx_k__150[] = ";\003;\003;\003;\007;\316\213\n;\014;\016;\316\216\013;\003;\003;\003;\003;\003";
static const char __pyx_k__152[] = ";\003;\003;\005;\316\227\n;\003<\005<\316\232\n<\003<\003<\003<\007<\316\237\n";
static const char __pyx_k__154[] = "<\014<\016<\316\242\013<\003=\003=\005=\316\246\n=\003>\003>\003>\007>\316\253";
static const char __pyx_k__156[] = "\n>\014>\016>\316\256\013>\003?\003?\005?\316\262\n?\003?\003?\003@\003@\003@";
static const char __pyx_k__167[] = "\302\202\302\204\302\206\302\210\302\212\302\214\302\216\302\220\302\222";
static const char __pyx_k__169[] = "\002\n\003\002\r\016\003\002#$\003\002%'\003\002()\003\002;<\004\002\037\037=?\004\002";
static const char __pyx_k__172[] = "\003\002\002\002\010\302\256\003\002\002\002\n\303\207\003\002\002\002\014\304\202\003\002\002";
static const char __pyx_k__174[] = "\002\016\304\204\003\002\002\002\020\304\206\003\002\002\002\022\304\216\003\002\002\002";
static const char __pyx_k__176[] = "\024\304\221\003\002\002\002\026\304\257\003\002\002\002\030\304\264\003\002\002\002\032";
static const char __pyx_k__178[] = "\305\202\003\002\002\002\034\305\211\003\002\002\002\036\305\214\003\002\002\002 \305\234";
static const char __pyx_k__180[] = "\003\002\002\002\"\305\261\003\002\002\002$\306\204\003\002\002\002&\306\214\003\002\002\002";
static const char __pyx_k__182[] = "(\306\224\003\002\002\002*\306\231\003\002\002\002,\306\241\003\002\002\002.\306\251\003";
static const char __pyx_k__195[] = "~\317\200\003\002\002\002\302\200\317\202\003\002\002\002\302\202\317\212\003\002\002";
static const char __pyx_k__197[] = "\002\302\204\317\227\003\002\002\002\302\206\317\232\003\002\002\002\302\210\317\243";
static const char __pyx_k__199[] = "\003\002\002\002\302\212\317\253\003\002\002\002\302\214\317\270\003\002\002\002\302\216";
static const char __pyx_k__201[] = "\317\272\003\002\002\002\302\220\320\206\003\002\002\002\302\222\320\210\003\002\002\002";
static const char __pyx_k__204[] = "\002\002\302\226\302\227\003\002\002\002\302\227\302\230\003\002\002\002\302\230\302\232";
static const char __pyx_k__206[] = "\007\003\002\002\302\231\302\224\003\002\002\002\302\231\302\232\003\002\002\002\302\232";
static const char __pyx_k__208[] = "\302\236\003\002\002\002\302\233\302\235\005\004\003\002\302\234\302\233\003\002\002\002";
static const char __pyx_k__210[] = "\302\235\302\240\003\002\002\002\302\236\302\234\003\002\002\002\302\236\302\237\003";
static const char __pyx_k__212[] = "\002\002\002\302\237\003\003\002\002\002\302\240\302\236\003\002\002\002\302\241\302\243";
static const char __pyx_k__215[] = "\302\244\003\002\002\002\302\244\302\245\005\006\004\002\302\245\302\246\007\003\002\002";
static const char __pyx_k__218[] = "\002\302\250\302\251\003\002\002\002\302\251\302\252\003\002\002\002\302\252\302\253";
static const char __pyx_k__220[] = "\005\010\005\002\302\253\302\254\005\014\007\002\302\254\007\003\002\002\002\302\255\302\257";
static const char __pyx_k__223[] = "\302\260\003\002\002\002\302\260\302\261\005\n\006\002\302\261\t\003\002\002\002\302\262";
static const char __pyx_k__225[] = "\303\210\007\004\002\002\302\263\303\210\007\005\002\002\302\264\302\266\007\006\002\002";
static const char __pyx_k__227[] = "\302\265\302\264\003\002\002\002\302\265\302\266\003\002\002\002\302\266\302\267\003";
static const char __pyx_k__229[] = "\002\002\002\302\267\303\210\007\007\002\002\302\270\303\210\007\010\002\002\302\271\302\273";
static const char __pyx_k__231[] = "\007\t\002\002\302\272\302\271\003\002\002\002\302\272\302\273\003\002\002\002\302\273";
static const char __pyx_k__233[] = "\302\274\003\002\002\002\302\274\303\210\007\n\002\002\302\275\303\210\007\013\002";
static const char __pyx_k__235[] = "\002\302\276\303\210\007\014\002\002\302\277\303\201\t\002\002\002\303\200\302\277";
static const char __pyx_k__237[] = "\003\002\002\002\303\200\303\201\003\002\002\002\303\201\303\203\003\002\002\002\303\202";
static const char __pyx_k__239[] = "\303\204\007\006\002\002\303\203\303\202\003\002\002\002\303\203\303\204\003\002\002\002";
static const char __pyx_k__241[] = "\303\204\303\205\003\002\002\002\303\205\303\210\007\017\002\002\303\206\303\210";
static const char __pyx_k__243[] = "\007\006\002\002\303\207\302\262\003\002\002\002\303\207\302\263\003\002\002\002\303\207";
static const char __pyx_k__245[] = "\302\265\003\002\002\002\303\207\302\270\003\002\002\002\303\207\302\272\003\002\002\002";
static const char __pyx_k__247[] = "\303\207\302\275\003\002\002\002\303\207\302\276\003\002\002\002\303\207\303\200\003";
static const char __pyx_k__249[] = "\002\002\002\303\207\303\206\003\002\002\002\303\210\013\003\002\002\002\303\211\303\212";
static const char __pyx_k__255[] = "\303\223\003\002\002\002\303\224\303\225\003\002\002\002\303\225\303\227\003\002\002\002";
static const char __pyx_k__257[] = "\303\226\303\230\005:\036\002\303\227\303\226\003\002\002\002\303\227\303\230\003";
static const char __pyx_k__261[] = "\303\235\303\236\007\022\002\002\303\236\303\243\007\023\002\002\303\237\303\241";
static const char __pyx_k__263[] = "\005\020\t\002\303\240\303\237\003\002\002\002\303\240\303\241\003\002\002\002\303\241";
static const char __pyx_k__265[] = "\303\244\003\002\002\002\303\242\303\244\007\024\002\002\303\243\303\240\003\002\002";
static const char __pyx_k__267[] = "\002\303\243\303\242\003\002\002\002\303\244\303\245\003\002\002\002\303\245\303\246";
static const char __pyx_k__270[] = "\303\251\007\021\002\002\303\251\303\252\007\026\002\002\303\252\303\253\007\023";
static const char __pyx_k__274[] = "\303\256\003\002\002\002\303\261\303\264\003\002\002\002\303\262\303\260\003\002\002\002";
static const char __pyx_k__276[] = "\303\262\303\263\003\002\002\002\303\263\303\265\003\002\002\002\303\264\303\262\003";
static const char __pyx_k__280[] = "\303\272\303\274\005:\036\002\303\273\303\272\003\002\002\002\303\273\303\274\003";
static const char __pyx_k__284[] = "\002\304\201\304\203\003\002\002\002\304\202\303\211\003\002\002\002\304\202\303\217";
static const char __pyx_k__286[] = "\003\002\002\002\304\202\303\233\003\002\002\002\304\202\303\247\003\002\002\002\304\202";
static const char __pyx_k__288[] = "\303\270\003\002\002\002\304\203\r\003\002\002\002\304\204\304\205\005,\027\002\304\205";
static const char __pyx_k__290[] = "\017\003\002\002\002\304\206\304\213\005\022\n\002\304\207\304\210\007\027\002\002";
static const char __pyx_k__292[] = "\304\210\304\212\005\022\n\002\304\211\304\207\003\002\002\002\304\212\304\215";
static const char __pyx_k__294[] = "\003\002\002\002\304\213\304\211\003\002\002\002\304\213\304\214\003\002\002\002\304\214";
static const char __pyx_k__298[] = "\002\304\222\304\223\007\031\002\002\304\223\304\231\005\032\016\002\304\224\304\225";
static const char __pyx_k__302[] = "\304\226\003\002\002\002\304\230\304\227\003\002\002\002\304\231\304\234\003\002\002\002";
static const char __pyx_k__304[] = "\304\232\304\230\003\002\002\002\304\232\304\233\003\002\002\002\304\233\304\250\003";
static const char __pyx_k__306[] = "\002\002\002\304\234\304\232\003\002\002\002\304\235\304\237\007\033\002\002\304\236";
static const char __pyx_k__308[] = "\304\240\005\026\014\002\304\237\304\236\003\002\002\002\304\237\304\240\003\002\002";
static const char __pyx_k__310[] = "\002\304\240\304\242\003\002\002\002\304\241\304\243\005\030\r\002\304\242\304\241";
static const char __pyx_k__312[] = "\003\002\002\002\304\242\304\243\003\002\002\002\304\243\304\245\003\002\002\002\304\244";
static const char __pyx_k__315[] = "\002\002\304\246\304\247\003\002\002\002\304\247\304\251\007\024\002\002\304\250\304\235";
static const char __pyx_k__317[] = "\003\002\002\002\304\250\304\251\003\002\002\002\304\251\304\255\003\002\002\002\304\252";
static const char __pyx_k__320[] = "\002\002\304\255\304\252\003\002\002\002\304\255\304\256\003\002\002\002\304\256\025";
static const char __pyx_k__322[] = "\003\002\002\002\304\257\304\260\007[\002\002\304\260\027\003\002\002\002\304\261\304\262";
static const char __pyx_k__325[] = "\304\261\003\002\002\002\304\264\304\265\003\002\002\002\304\265\304\266\003\002\002\002";
static const char __pyx_k__329[] = "\304\273\003\002\002\002\304\273\304\274\007\025\002\002\304\274\031\003\002\002\002\304\275";
static const char __pyx_k__331[] = "\304\276\005\034\017\002\304\276\304\277\007\003\002\002\304\277\305\201\003\002\002";
static const char __pyx_k__333[] = "\002\305\200\304\275\003\002\002\002\305\201\305\204\003\002\002\002\305\202\305\200";
static const char __pyx_k__335[] = "\003\002\002\002\305\202\305\203\003\002\002\002\305\203\033\003\002\002\002\305\204\305\202";
static const char __pyx_k__337[] = "\003\002\002\002\305\205\305\212\005\"\022\002\305\206\305\212\005&\024\002\305\207";
static const char __pyx_k__339[] = "\305\212\005\036\020\002\305\210\305\212\005 \021\002\305\211\305\205\003\002\002";
static const char __pyx_k__341[] = "\002\305\211\305\206\003\002\002\002\305\211\305\207\003\002\002\002\305\211\305\210";
static const char __pyx_k__344[] = "\003\002\002\002\305\214\305\215\003\002\002\002\305\215\305\217\003\002\002\002\305\216";
static const char __pyx_k__348[] = "\002\002\002\305\222\305\223\003\002\002\002\305\223\305\225\003\002\002\002\305\224\305\226";
static const char __pyx_k__351[] = "\305\231\003\002\002\002\305\227\305\232\005\006\004\002\305\230\305\232\005*\026\002";
static const char __pyx_k__353[] = "\305\231\305\227\003\002\002\002\305\231\305\230\003\002\002\002\305\232\037\003\002";
static const char __pyx_k__357[] = "\305\236\003\002\002\002\305\237\305\240\003\002\002\002\305\240\305\242\003\002\002\002";
static const char __pyx_k__361[] = "\003\002\002\002\305\245\305\246\003\002\002\002\305\246\305\247\003\002\002\002\305\247";
static const char __pyx_k__363[] = "\305\252\007\034\002\002\305\250\305\253\005\006\004\002\305\251\305\253\005*\026";
static const char __pyx_k__365[] = "\002\305\252\305\250\003\002\002\002\305\252\305\251\003\002\002\002\305\253\305\257";
static const char __pyx_k__368[] = "\305\260\003\002\002\002\305\257\305\254\003\002\002\002\305\257\305\260\003\002\002\002";
static const char __pyx_k__373[] = "\305\271\007 \002\002\305\271\305\272\005$\023\002\305\272\305\273\007!\002\002\305\273";
static const char __pyx_k__375[] = "\305\275\003\002\002\002\305\274\305\267\003\002\002\002\305\274\305\270\003\002\002\002";
static const char __pyx_k__377[] = "\305\275\305\277\003\002\002\002\305\276\305\266\003\002\002\002\305\276\305\277\003";
static const char __pyx_k__379[] = "\002\002\002\305\277\306\201\003\002\002\002\306\200\305\262\003\002\002\002\306\200\305\265";
static const char __pyx_k__383[] = "\306\210\005$\023\002\306\207\306\205\003\002\002\002\306\210\306\213\003\002\002\002";
static const char __pyx_k__385[] = "\306\211\306\207\003\002\002\002\306\211\306\212\003\002\002\002\306\212%\003\002\002";
static const char __pyx_k__387[] = "\002\306\213\306\211\003\002\002\002\306\214\306\215\007\030\002\002\306\215\306\217";
static const char __pyx_k__391[] = "\306\221\003\002\002\002\306\222\306\223\003\002\002\002\306\223'\003\002\002\002\306\224";
static const char __pyx_k__394[] = "\306\226\003\002\002\002\306\227\306\230\003\002\002\002\306\230)\003\002\002\002\306\231";
static const char __pyx_k__396[] = "\306\232\005,\027\002\306\232\306\234\005.\030\002\306\233\306\235\005\302\212";
static const char __pyx_k__400[] = "\t\003\002\002\306\241\306\240\003\002\002\002\306\241\306\242\003\002\002\002\306\242";
static const char __pyx_k__402[] = "\306\244\003\002\002\002\306\243\306\245\t\004\002\002\306\244\306\243\003\002\002\002";
static const char __pyx_k__404[] = "\306\244\306\245\003\002\002\002\306\245\306\247\003\002\002\002\306\246\306\250\t";
static const char __pyx_k__406[] = "\005\002\002\306\247\306\246\003\002\002\002\306\247\306\250\003\002\002\002\306\250-";
static const char __pyx_k__410[] = "\306\256\306\254\003\002\002\002\306\257\306\262\003\002\002\002\306\260\306\256\003";
static const char __pyx_k__414[] = "\306\264\003\002\002\002\306\265\306\266\003\002\002\002\306\266\306\267\003\002\002\002";
static const char __pyx_k__421[] = "\002\002\307\203\307\206\005:\036\002\307\204\307\205\007\021\002\002\307\205\307\207";
static const char __pyx_k__427[] = "\307\220\007\023\002\002\307\217\307\221\005<\037\002\307\220\307\217\003\002\002";
static const char __pyx_k__429[] = "\002\307\220\307\221\003\002\002\002\307\221\307\222\003\002\002\002\307\222\307\223";
static const char __pyx_k__431[] = "\007\025\002\002\307\223;\003\002\002\002\307\224\307\231\005> \002\307\225\307\226";
static const char __pyx_k__433[] = "\007\027\002\002\307\226\307\230\005> \002\307\227\307\225\003\002\002\002\307\230";
static const char __pyx_k__435[] = "\307\233\003\002\002\002\307\231\307\227\003\002\002\002\307\231\307\232\003\002\002\002";
static const char __pyx_k__437[] = "\307\232=\003\002\002\002\307\233\307\231\003\002\002\002\307\234\307\237\005@!\002";
static const char __pyx_k__441[] = "\002\002\307\241\307\242\003\002\002\002\307\242\307\244\003\002\002\002\307\243\307\245";
static const char __pyx_k__447[] = "\307\255\003\002\002\002\307\255\307\256\003\002\002\002\307\256\307\257\005\302\220";
static const char __pyx_k__450[] = "\002\307\262\307\261\003\002\002\002\307\262\307\263\003\002\002\002\307\263\307\265";
static const char __pyx_k__458[] = "\003\002\002\002\310\201\310\200\003\002\002\002\310\202\310\204\003\002\002\002\310\203";
static const char __pyx_k__460[] = "\310\205\005(\025\002\310\204\310\203\003\002\002\002\310\204\310\205\003\002\002\002";
static const char __pyx_k__467[] = "\310\222\003\002\002\002\310\223\310\224\003\002\002\002\310\224\310\226\003\002\002\002";
static const char __pyx_k__469[] = "\310\225\310\227\005:\036\002\310\226\310\225\003\002\002\002\310\226\310\227\003";
static const char __pyx_k__472[] = "\310\245\003\002\002\002\310\232\310\233\007\022\002\002\310\233\310\240\007\023\002";
static const char __pyx_k__474[] = "\002\310\234\310\236\005\020\t\002\310\235\310\234\003\002\002\002\310\235\310\236";
static const char __pyx_k__476[] = "\003\002\002\002\310\236\310\241\003\002\002\002\310\237\310\241\007\024\002\002\310\240";
static const char __pyx_k__478[] = "\310\235\003\002\002\002\310\240\310\237\003\002\002\002\310\241\310\242\003\002\002\002";
static const char __pyx_k__484[] = "\310\254\007\003\002\002\310\254\310\256\003\002\002\002\310\255\310\252\003\002\002\002";
static const char __pyx_k__486[] = "\310\256\310\261\003\002\002\002\310\257\310\255\003\002\002\002\310\257\310\260\003";
static const char __pyx_k__491[] = "\310\270\007\003\002\002\310\270\310\272\003\002\002\002\310\271\310\266\003\002\002\002";
static const char __pyx_k__493[] = "\310\272\310\275\003\002\002\002\310\273\310\271\003\002\002\002\310\273\310\274\003";
static const char __pyx_k__499[] = "\005|?\002\311\210\311\212\003\002\002\002\311\211\310\276\003\002\002\002\311\211\311\202";
static const char __pyx_k__501[] = "\003\002\002\002\311\211\311\203\003\002\002\002\311\211\311\204\003\002\002\002\311\211";
static const char __pyx_k__506[] = "\311\223\005|?\002\311\222\311\217\003\002\002\002\311\222\311\221\003\002\002\002\311\223";
static const char __pyx_k__510[] = "\002\002\002\311\231\311\234\003\002\002\002\311\232\311\230\003\002\002\002\311\232\311\233";
static const char __pyx_k__512[] = "\003\002\002\002\311\233\311\235\003\002\002\002\311\234\311\232\003\002\002\002\311\235";
static const char __pyx_k__515[] = "\311\241\005|?\002\311\241\311\251\003\002\002\002\311\242\311\251\007.\002\002\311\243";
static const char __pyx_k__519[] = "\311\250\311\224\003\002\002\002\311\250\311\242\003\002\002\002\311\250\311\243\003";
static const char __pyx_k__521[] = "\002\002\002\311\250\311\244\003\002\002\002\311\250\311\245\003\002\002\002\311\250\311\246";
static const char __pyx_k__527[] = "\003\002\002\002\311\264\311\267\003\002\002\002\311\265\311\263\003\002\002\002\311\265";
static const char __pyx_k__529[] = "\311\266\003\002\002\002\311\266\312\205\003\002\002\002\311\267\311\265\003\002\002\002";
static const char __pyx_k__533[] = "\003\002\002\002\311\276\311\273\003\002\002\002\311\277\312\202\003\002\002\002\312\200";
static const char __pyx_k__535[] = "\311\276\003\002\002\002\312\200\312\201\003\002\002\002\312\201\312\204\003\002\002\002";
static const char __pyx_k__537[] = "\312\202\312\200\003\002\002\002\312\203\311\270\003\002\002\002\312\204\312\207\003";
static const char __pyx_k__539[] = "\002\002\002\312\205\312\203\003\002\002\002\312\205\312\206\003\002\002\002\312\206\312\221";
static const char __pyx_k__543[] = "\312\211\003\002\002\002\312\215\312\220\003\002\002\002\312\216\312\214\003\002\002\002";
static const char __pyx_k__545[] = "\312\216\312\217\003\002\002\002\312\217\312\222\003\002\002\002\312\220\312\216\003";
static const char __pyx_k__547[] = "\002\002\002\312\221\312\210\003\002\002\002\312\221\312\222\003\002\002\002\312\222\312\223";
static const char __pyx_k__549[] = "\003\002\002\002\312\223\312\224\007\020\002\002\312\224\312\225\007*\002\002\312\225";
static const char __pyx_k__553[] = "\312\235\003\002\002\002\312\234\312\231\003\002\002\002\312\235\312\240\003\002\002\002";
static const char __pyx_k__555[] = "\312\236\312\234\003\002\002\002\312\236\312\237\003\002\002\002\312\237\312\256\003";
static const char __pyx_k__559[] = "\312\246\007\003\002\002\312\246\312\250\003\002\002\002\312\247\312\244\003\002\002\002";
static const char __pyx_k__561[] = "\312\250\312\253\003\002\002\002\312\251\312\247\003\002\002\002\312\251\312\252\003";
static const char __pyx_k__563[] = "\002\002\002\312\252\312\255\003\002\002\002\312\253\312\251\003\002\002\002\312\254\312\241";
static const char __pyx_k__565[] = "\003\002\002\002\312\255\312\260\003\002\002\002\312\256\312\254\003\002\002\002\312\256";
static const char __pyx_k__567[] = "\312\257\003\002\002\002\312\257\312\272\003\002\002\002\312\260\312\256\003\002\002\002";
static const char __pyx_k__570[] = "\003\002\002\312\264\312\266\003\002\002\002\312\265\312\262\003\002\002\002\312\266\312\271";
static const char __pyx_k__572[] = "\003\002\002\002\312\267\312\265\003\002\002\002\312\267\312\270\003\002\002\002\312\270";
static const char __pyx_k__574[] = "\312\273\003\002\002\002\312\271\312\267\003\002\002\002\312\272\312\261\003\002\002\002";
static const char __pyx_k__576[] = "\312\272\312\273\003\002\002\002\312\273\312\274\003\002\002\002\312\274\312\275\007";
static const char __pyx_k__578[] = "\020\002\002\312\275\312\276\007*\002\002\312\276]\003\002\002\002\312\277\313\200";
static const char __pyx_k__582[] = "\313\202\003\002\002\002\313\206\313\211\003\002\002\002\313\207\313\205\003\002\002\002";
static const char __pyx_k__584[] = "\313\207\313\210\003\002\002\002\313\210\313\212\003\002\002\002\313\211\313\207\003";
static const char __pyx_k__589[] = "\313\222\313\224\003\002\002\002\313\223\313\220\003\002\002\002\313\224\313\227\003";
static const char __pyx_k__591[] = "\002\002\002\313\225\313\223\003\002\002\002\313\225\313\226\003\002\002\002\313\226\313\230";
static const char __pyx_k__593[] = "\003\002\002\002\313\227\313\225\003\002\002\002\313\230\313\231\007\020\002\002\313\231";
static const char __pyx_k__597[] = "\002\313\237\313\242\003\002\002\002\313\240\313\236\003\002\002\002\313\240\313\241";
static const char __pyx_k__604[] = "\313\253\003\002\002\002\313\257\313\262\003\002\002\002\313\260\313\256\003\002\002\002";
static const char __pyx_k__606[] = "\313\260\313\261\003\002\002\002\313\261\313\263\003\002\002\002\313\262\313\260\003";
static const char __pyx_k__611[] = "\313\273\313\275\003\002\002\002\313\274\313\271\003\002\002\002\313\275\314\200\003";
static const char __pyx_k__613[] = "\002\002\002\313\276\313\274\003\002\002\002\313\276\313\277\003\002\002\002\313\277\314\216";
static const char __pyx_k__617[] = "\314\206\007\003\002\002\314\206\314\210\003\002\002\002\314\207\314\204\003\002\002\002";
static const char __pyx_k__619[] = "\314\210\314\213\003\002\002\002\314\211\314\207\003\002\002\002\314\211\314\212\003";
static const char __pyx_k__621[] = "\002\002\002\314\212\314\215\003\002\002\002\314\213\314\211\003\002\002\002\314\214\314\201";
static const char __pyx_k__623[] = "\003\002\002\002\314\215\314\220\003\002\002\002\314\216\314\214\003\002\002\002\314\216";
static const char __pyx_k__625[] = "\314\217\003\002\002\002\314\217\314\221\003\002\002\002\314\220\314\216\003\002\002\002";
static const char __pyx_k__630[] = "\314\233\003\002\002\002\314\232\314\227\003\002\002\002\314\233\314\236\003\002\002\002";
static const char __pyx_k__632[] = "\314\234\314\232\003\002\002\002\314\234\314\235\003\002\002\002\314\235\314\254\003";
static const char __pyx_k__636[] = "\007\003\002\002\314\244\314\246\003\002\002\002\314\245\314\242\003\002\002\002\314\246";
static const char __pyx_k__638[] = "\314\251\003\002\002\002\314\247\314\245\003\002\002\002\314\247\314\250\003\002\002\002";
static const char __pyx_k__640[] = "\314\250\314\253\003\002\002\002\314\251\314\247\003\002\002\002\314\252\314\237\003";
static const char __pyx_k__642[] = "\002\002\002\314\253\314\256\003\002\002\002\314\254\314\252\003\002\002\002\314\254\314\255";
static const char __pyx_k__644[] = "\003\002\002\002\314\255\314\257\003\002\002\002\314\256\314\254\003\002\002\002\314\257";
static const char __pyx_k__653[] = "\002\002\315\204\315\207\003\002\002\002\315\205\315\203\003\002\002\002\315\205\315\206";
static const char __pyx_k__655[] = "\003\002\002\002\315\206\315\210\003\002\002\002\315\207\315\205\003\002\002\002\315\210";
static const char __pyx_k__661[] = "\003\002\002\002\315\222\315\223\003\002\002\002\315\223\315\225\003\002\002\002\315\224";
static const char __pyx_k__668[] = "\315\235\003\002\002\002\315\240\315\237\003\002\002\002\315\241\315\263\003\002\002\002";
static const char __pyx_k__670[] = "\315\242\315\243\014\t\002\002\315\243\315\244\t\007\002\002\315\244\315\262\005";
static const char __pyx_k__677[] = "\003\002\002\002\315\261\315\253\003\002\002\002\315\261\315\256\003\002\002\002\315\262";
static const char __pyx_k__679[] = "\315\265\003\002\002\002\315\263\315\261\003\002\002\002\315\263\315\264\003\002\002\002";
static const char __pyx_k__684[] = "\003\002\002\002\315\275\315\276\007\026\002\002\315\276\316\227\005|?\002\315\277";
static const char __pyx_k__690[] = "\316\212\316\210\003\002\002\002\316\213\316\216\003\002\002\002\316\214\316\212\003";
static const char __pyx_k__692[] = "\002\002\002\316\214\316\215\003\002\002\002\316\215\316\217\003\002\002\002\316\216\316\214";
static const char __pyx_k__695[] = "\316\222\007 \002\002\316\222\316\223\005~@\002\316\223\316\224\007!\002\002\316\224";
static const char __pyx_k__697[] = "\316\227\003\002\002\002\316\225\316\227\007\020\002\002\316\226\315\266\003\002\002";
static const char __pyx_k__699[] = "\002\316\226\315\267\003\002\002\002\316\226\315\270\003\002\002\002\316\226\315\271";
static const char __pyx_k__701[] = "\003\002\002\002\316\226\315\272\003\002\002\002\316\226\315\275\003\002\002\002\316\226";
static const char __pyx_k__703[] = "\315\277\003\002\002\002\316\226\316\201\003\002\002\002\316\226\316\202\003\002\002\002";
static const char __pyx_k__705[] = "\316\226\316\206\003\002\002\002\316\226\316\221\003\002\002\002\316\226\316\225\003";
static const char __pyx_k__708[] = "\003\002\002\002\316\231\316\232\003\002\002\002\316\232\316\233\003\002\002\002\316\233";
static const char __pyx_k__711[] = "\316\236\316\234\003\002\002\002\316\237\316\242\003\002\002\002\316\240\316\236\003";
static const char __pyx_k__717[] = "\316\250\003\002\002\002\316\253\316\256\003\002\002\002\316\254\316\252\003\002\002\002";
static const char __pyx_k__719[] = "\316\254\316\255\003\002\002\002\316\255{\003\002\002\002\316\256\316\254\003\002\002";
static const char __pyx_k__721[] = "\002\316\257\316\261\007\023\002\002\316\260\316\262\005~@\002\316\261\316\260";
static const char __pyx_k__723[] = "\003\002\002\002\316\261\316\262\003\002\002\002\316\262\316\263\003\002\002\002\316\263";
static const char __pyx_k__728[] = "\316\270\003\002\002\002\316\273\316\276\003\002\002\002\316\274\316\272\003\002\002\002";
static const char __pyx_k__730[] = "\316\274\316\275\003\002\002\002\316\275\317\201\003\002\002\002\316\276\316\274\003";
static const char __pyx_k__735[] = "\003\002\002\002\317\206\317\211\003\002\002\002\317\207\317\205\003\002\002\002\317\207";
static const char __pyx_k__737[] = "\317\210\003\002\002\002\317\210\302\201\003\002\002\002\317\211\317\207\003\002\002\002";
static const char __pyx_k__743[] = "\003\002\002\002\317\224\317\225\007\025\002\002\317\225\317\230\003\002\002\002\317\226";
static const char __pyx_k__747[] = "\317\233\003\002\002\002\317\233\317\240\003\002\002\002\317\234\317\235\007\027\002";
static const char __pyx_k__750[] = "\002\002\002\317\240\317\236\003\002\002\002\317\240\317\241\003\002\002\002\317\241\302\207";
static const char __pyx_k__754[] = "\317\252\003\002\002\002\317\250\317\246\003\002\002\002\317\250\317\251\003\002\002\002";
static const char __pyx_k__756[] = "\317\251\302\211\003\002\002\002\317\252\317\250\003\002\002\002\317\253\317\254\007";
static const char __pyx_k__760[] = "\002\002\317\261\317\257\003\002\002\002\317\261\317\262\003\002\002\002\317\262\317\264";
static const char __pyx_k__764[] = "\317\270\317\266\003\002\002\002\317\270\317\267\003\002\002\002\317\271\302\215\003";
static const char __pyx_k__767[] = "\317\273\003\002\002\002\317\274\317\275\003\002\002\002\317\275\302\217\003\002\002\002";
static const char __pyx_k__769[] = "\317\276\320\203\007[\002\002\317\277\320\200\007@\002\002\320\200\320\202\007[";
static const char __pyx_k__771[] = "\002\002\320\201\317\277\003\002\002\002\320\202\320\205\003\002\002\002\320\203\320\201";
static const char __pyx_k__773[] = "\003\002\002\002\320\203\320\204\003\002\002\002\320\204\320\207\003\002\002\002\320\205";
static const char __pyx_k__775[] = "\320\203\003\002\002\002\320\206\317\276\003\002\002\002\320\206\320\207\003\002\002\002";
static const char __pyx_k__778[] = ":\036\002\320\212\302\223\003\002\002\002\302\207\302\226\302\231\302\236\302\242";
static const char __pyx_k__780[] = "\302\250\302\256\302\265\302\272\303\200\303\203\303\207\303\224\303\227";
static const char __pyx_k__782[] = "\303\240\303\243\303\262\303\273\304\202\304\213\304\230\304\232\304\237";
static const char __pyx_k__784[] = "\304\242\304\245\304\250\304\255\304\264\304\271\305\202\305\211\305\214";
static const char __pyx_k__786[] = "\305\217\305\222\305\225\305\231\305\234\305\237\305\242\305\245\305\252";
static const char __pyx_k__788[] = "\305\257\305\274\305\276\306\200\306\211\306\217\306\222\306\227\306\234";
static const char __pyx_k__790[] = "\306\241\306\244\306\247\306\260\306\265\306\276\307\201\307\206\307\214";
static const char __pyx_k__792[] = "\307\220\307\231\307\236\307\241\307\244\307\250\307\254\307\262\307\265";
static const char __pyx_k__794[] = "\307\271\307\274\310\201\310\204\310\223\310\226\310\235\310\240\310\244";
static const char __pyx_k__796[] = "\310\247\310\257\310\263\310\273\311\211\311\222\311\232\311\250\311\265";
static const char __pyx_k__798[] = "\312\200\312\205\312\216\312\221\312\236\312\251\312\256\312\267\312\272";
static const char __pyx_k__800[] = "\313\207\313\225\313\240\313\246\313\260\313\276\314\211\314\216\314\234";
static const char __pyx_k__802[] = "\314\247\314\254\315\205\315\213\315\222\315\224\315\240\315\261\315\263";
static const char __pyx_k__804[] = "\316\214\316\226\316\231\316\240\316\245\316\254\316\261\316\272\316\274";
static const char __pyx_k__806[] = "\317\200\317\207\317\222\317\227\317\232\317\240\317\250\317\261\317\270";
static const char __pyx_k__808[] = "\317\274\320\203\320\206";
static const char __pyx_k__975[] = "*";
static const char __pyx_k__978[] = "';'";
static const char __pyx_k__979[] = "'='";
static const char __pyx_k__980[] = "'('";
static const char __pyx_k__981[] = "':'";
static const char __pyx_k__982[] = "')'";
static const char __pyx_k__983[] = "','";
static const char __pyx_k__984[] = "'.'";
static const char __pyx_k__985[] = "'*'";
static const char __pyx_k__986[] = "'{'";
static const char __pyx_k__987[] = "'}'";
static const char __pyx_k__988[] = "':='";
static const char __pyx_k__989[] = "'-'";
static const char __pyx_k__990[] = "'^'";
static const char __pyx_k__991[] = "'.^'";
static const char __pyx_k__992[] = "'/'";
static const char __pyx_k__993[] = "'.*'";
static const char __pyx_k__994[] = "'./'";
static const char __pyx_k__995[] = "'+'";
static const char __pyx_k__996[] = "'.+'";
static const char __pyx_k__997[] = "'.-'";
static const char __pyx_k__998[] = "'<'";
static const char __pyx_k__999[] = "'<='";
static const char __pyx_k_each[] = "'each'";
static const char __pyx_k_else[] = "'else'";
static const char __pyx_k_epro[] = "epro";
static const char __pyx_k_epub[] = "epub";
static const char __pyx_k_exit[] = "__exit__";
static const char __pyx_k_expr[] = "expr";
static const char __pyx_k_flow[] = "'flow'";
static const char __pyx_k_init[] = "__init__";
static const char __pyx_k_la_2[] = "la_";
static const char __pyx_k_loop[] = "'loop'";
static const char __pyx_k_m_p9[] = "\314\270m\003\002\002\002\314\271\315\214\005p9\002\314\272\314\273\007*\002\002\314\273";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_n8_0[] = "\007*\002\002\311\256\311\257\005n8\002\311\257\311\265\0070\002\002\311\260\311\261";
static const char __pyx_k_n8_2[] = "\003\002\002\002\307\210\307\211\007\021\002\002\307\211\307\215\005n8\002\307\212";
static const char __pyx_k_n8_3[] = "\307\213\007+\002\002\307\213\307\215\005n8\002\307\214\307\203\003\002\002\002\307\214";
static const char __pyx_k_n8_4[] = "[\003\002\002\002\312\226\312\227\007*\002\002\312\227\312\230\005n8\002\312\230\312\236";
static const char __pyx_k_n8_5[] = "*\002\002\306\272\306\273\005n8\002\306\2735\003\002\002\002\306\274\306\276\007";
static const char __pyx_k_n8_6[] = "\315\201\315\202\005n8\002\315\202\315\204\003\002\002\002\315\203\314\276\003\002";
static const char __pyx_k_n8_7[] = "\317\230\005n8\002\317\227\317\216\003\002\002\002\317\227\317\226\003\002\002\002\317\230";
static const char __pyx_k_n8_8[] = "\302\205\003\002\002\002\317\231\317\233\005n8\002\317\232\317\231\003\002\002\002\317\232";
static const char __pyx_k_n8_9[] = "\002\317\235\317\237\005n8\002\317\236\317\234\003\002\002\002\317\237\317\242\003";
static const char __pyx_k_name[] = "name";
static const char __pyx_k_pred[] = "pred";
static const char __pyx_k_pure[] = "'pure'";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_stop[] = "stop";
static const char __pyx_k_sync[] = "sync";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_then[] = "'then'";
static const char __pyx_k_true[] = "'true'";
static const char __pyx_k_type[] = "'type'";
static const char __pyx_k_when[] = "'when'";
static const char __pyx_k_z_10[] = "\314\266\007\027\002\002\314\266\314\267\005z>\002\314\267\314\270\007\025\002\002";
static const char __pyx_k_z_n8[] = "\311\222\005z>\002\311\217\311\220\007+\002\002\311\220\311\223\005n8\002\311\221";
static const char __pyx_k_0_0_0[] = "/\016/\312\271\013/\005/\312\273\n/\003/\003/\003/\0030\0030\0030\003";
static const char __pyx_k_0_2_4[] = "\002\002\0020\306\253\003\002\002\0022\306\263\003\002\002\0024\306\271\003\002";
static const char __pyx_k_0_T_2[] = "\313\276\0070\002\002\313\271\313\272\005T+\002\313\272\313\273\007\003\002\002";
static const char __pyx_k_0_X_2[] = "\0070\002\002\314\227\314\230\005X-\002\314\230\314\231\007\003\002\002\314\231";
static const char __pyx_k_FINAL[] = "FINAL";
static const char __pyx_k_H_3_2[] = "\306\267\306\270\005\302\216H\002\306\2703\003\002\002\002\306\271\306\272\007";
static const char __pyx_k_IDENT[] = "IDENT";
static const char __pyx_k_INNER[] = "INNER";
static const char __pyx_k_J_I_6[] = "\002\310\210\310\211\005J&\002\310\211I\003\002\002\002\310\212\310\213\0056\034";
static const char __pyx_k_OUTER[] = "OUTER";
static const char __pyx_k_R_H_U[] = "\311\214\005R*\002\311\214\311\215\005\302\216H\002\311\215U\003\002\002\002\311\216";
static const char __pyx_k_T_V_X[] = "\003\002\002\002T\311\213\003\002\002\002V\311\250\003\002\002\002X\311\252\003\002\002\002";
static const char __pyx_k_T__10[] = "T__10";
static const char __pyx_k_T__11[] = "T__11";
static const char __pyx_k_T__12[] = "T__12";
static const char __pyx_k_T__13[] = "T__13";
static const char __pyx_k_T__14[] = "T__14";
static const char __pyx_k_T__15[] = "T__15";
static const char __pyx_k_T__16[] = "T__16";
static const char __pyx_k_T__17[] = "T__17";
static const char __pyx_k_T__18[] = "T__18";
static const char __pyx_k_T__19[] = "T__19";
static const char __pyx_k_T__20[] = "T__20";
static const char __pyx_k_T__21[] = "T__21";
static const char __pyx_k_T__22[] = "T__22";
static const char __pyx_k_T__23[] = "T__23";
static const char __pyx_k_T__24[] = "T__24";
static const char __pyx_k_T__25[] = "T__25";
static const char __pyx_k_T__26[] = "T__26";
static const char __pyx_k_T__27[] = "T__27";
static const char __pyx_k_T__28[] = "T__28";
static const char __pyx_k_T__29[] = "T__29";
static const char __pyx_k_T__30[] = "T__30";
static const char __pyx_k_T__31[] = "T__31";
static const char __pyx_k_T__32[] = "T__32";
static const char __pyx_k_T__33[] = "T__33";
static const char __pyx_k_T__34[] = "T__34";
static const char __pyx_k_T__35[] = "T__35";
static const char __pyx_k_T__36[] = "T__36";
static const char __pyx_k_T__37[] = "T__37";
static const char __pyx_k_T__38[] = "T__38";
static const char __pyx_k_T__39[] = "T__39";
static const char __pyx_k_T__40[] = "T__40";
static const char __pyx_k_T__41[] = "T__41";
static const char __pyx_k_T__42[] = "T__42";
static const char __pyx_k_T__43[] = "T__43";
static const char __pyx_k_T__44[] = "T__44";
static const char __pyx_k_T__45[] = "T__45";
static const char __pyx_k_T__46[] = "T__46";
static const char __pyx_k_T__47[] = "T__47";
static const char __pyx_k_T__48[] = "T__48";
static const char __pyx_k_T__49[] = "T__49";
static const char __pyx_k_T__50[] = "T__50";
static const char __pyx_k_T__51[] = "T__51";
static const char __pyx_k_T__52[] = "T__52";
static const char __pyx_k_T__53[] = "T__53";
static const char __pyx_k_T__54[] = "T__54";
static const char __pyx_k_T__55[] = "T__55";
static const char __pyx_k_T__56[] = "T__56";
static const char __pyx_k_T__57[] = "T__57";
static const char __pyx_k_T__58[] = "T__58";
static const char __pyx_k_T__59[] = "T__59";
static const char __pyx_k_T__60[] = "T__60";
static const char __pyx_k_T__61[] = "T__61";
static const char __pyx_k_T__62[] = "T__62";
static const char __pyx_k_T__63[] = "T__63";
static const char __pyx_k_T__64[] = "T__64";
static const char __pyx_k_T__65[] = "T__65";
static const char __pyx_k_T__66[] = "T__66";
static const char __pyx_k_T__67[] = "T__67";
static const char __pyx_k_T__68[] = "T__68";
static const char __pyx_k_T__69[] = "T__69";
static const char __pyx_k_T__70[] = "T__70";
static const char __pyx_k_T__71[] = "T__71";
static const char __pyx_k_T__72[] = "T__72";
static const char __pyx_k_T__73[] = "T__73";
static const char __pyx_k_T__74[] = "T__74";
static const char __pyx_k_T__75[] = "T__75";
static const char __pyx_k_T__76[] = "T__76";
static const char __pyx_k_T__77[] = "T__77";
static const char __pyx_k_Token[] = "Token";
static const char __pyx_k_V_H_Y[] = "\005V,\002\311\253\311\254\005\302\216H\002\311\254Y\003\002\002\002\311\255\311\256";
static const char __pyx_k_Z_F_2[] = "\003\002\002\002\316\243\316\245\007Z\002\002\316\244\316\246\005\302\212F\002\316\245";
static const char __pyx_k_Z_Z_2[] = "\316\240\007Z\002\002\316\234\316\235\007\036\002\002\316\235\316\237\007Z\002\002";
static const char __pyx_k__1000[] = "'>'";
static const char __pyx_k__1001[] = "'>='";
static const char __pyx_k__1002[] = "'=='";
static const char __pyx_k__1003[] = "'<>'";
static const char __pyx_k__1004[] = "'['";
static const char __pyx_k__1005[] = "']'";
static const char __pyx_k_b_d_f[] = "\003\002\002\002b\313\233\003\002\002\002d\313\243\003\002\002\002f\313\250\003\002\002\002";
static const char __pyx_k_block[] = "'block'";
static const char __pyx_k_break[] = "'break'";
static const char __pyx_k_class[] = "'class'";
static const char __pyx_k_ctx_2[] = "ctx";
static const char __pyx_k_enter[] = "__enter__";
static const char __pyx_k_epriv[] = "epriv";
static const char __pyx_k_f4_j6[] = "\311\251\005f4\002\311\247\311\251\005j6\002\311\250\311\216\003\002\002\002";
static const char __pyx_k_false[] = "'false'";
static const char __pyx_k_final[] = "'final'";
static const char __pyx_k_inner[] = "'inner'";
static const char __pyx_k_input[] = "input";
static const char __pyx_k_match[] = "match";
static const char __pyx_k_model[] = "'model'";
static const char __pyx_k_n8_10[] = "\003\002\002\002\317\242\317\240\003\002\002\002\317\243\317\250\005n8\002\317\244\317\245";
static const char __pyx_k_n8_11[] = "\007\027\002\002\317\245\317\247\005n8\002\317\246\317\244\003\002\002\002\317\247";
static const char __pyx_k_n8_12[] = "\302\213\003\002\002\002\317\266\317\271\007\024\002\002\317\267\317\271\005n8\002";
static const char __pyx_k_outer[] = "'outer'";
static const char __pyx_k_p9_n8[] = "\005p9\002\310\277\311\200\007\021\002\002\311\200\311\201\005n8\002\311\201\311\212";
static const char __pyx_k_state[] = "state";
static const char __pyx_k_super[] = "super";
static const char __pyx_k_token[] = "token";
static const char __pyx_k_while[] = "'while'";
static const char __pyx_k_write[] = "write";
static const char __pyx_k_1_n8_0[] = "\314\277\0071\002\002\314\277\315\200\005n8\002\315\200\315\201\0070\002\002";
static const char __pyx_k_3_a_d3[] = "\313\232\0073\002\002\313\232a\003\002\002\002\313\233\313\240\005d3\002\313\234";
static const char __pyx_k_3_b2_2[] = "\0073\002\002\316\271\316\273\005b2\002\316\272\316\266\003\002\002\002\316\272";
static const char __pyx_k_3_b2_4[] = "\0073\002\002\313\200\313\201\005b2\002\313\201\313\207\0074\002\002\313\202";
static const char __pyx_k_6_n8_4[] = "\0076\002\002\313\251\313\252\005n8\002\313\252\313\260\0074\002\002\313\253";
static const char __pyx_k_Parser[] = "Parser";
static const char __pyx_k_STRING[] = "STRING";
static const char __pyx_k_TextIO[] = "TextIO";
static const char __pyx_k_WITHIN[] = "WITHIN";
static const char __pyx_k_Z_5_n8[] = "\007Z\002\002\313\244\313\245\0075\002\002\313\245\313\247\005n8\002\313\246\313\244";
static const char __pyx_k_accept[] = "accept";
static const char __pyx_k_antlr4[] = "antlr4";
static const char __pyx_k_elseif[] = "'elseif'";
static const char __pyx_k_g_7_n8[] = "g\003\002\002\002\313\266\313\267\0077\002\002\313\267\313\270\005n8\002\313\270";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_impure[] = "'impure'";
static const char __pyx_k_interp[] = "_interp";
static const char __pyx_k_module[] = "__module__";
static const char __pyx_k_n8_0_T[] = "\314\203\005n8\002\314\203\314\211\0070\002\002\314\204\314\205\005T+\002\314\205";
static const char __pyx_k_n8_0_X[] = "\312\243\005n8\002\312\243\312\251\0070\002\002\312\244\312\245\005X-\002\312\245";
static const char __pyx_k_output[] = "output";
static const char __pyx_k_parent[] = "parent";
static const char __pyx_k_parser[] = "parser";
static const char __pyx_k_public[] = "'public'";
static const char __pyx_k_record[] = "'record'";
static const char __pyx_k_return[] = "'return'";
static const char __pyx_k_stdout[] = "stdout";
static const char __pyx_k_stream[] = "'stream'";
static const char __pyx_k_within[] = "'within'";
static const char __pyx_k_COMMENT[] = "COMMENT";
static const char __pyx_k_D_F_H_J[] = "D\307\260\003\002\002\002F\307\276\003\002\002\002H\310\206\003\002\002\002J\310\212\003";
static const char __pyx_k_INITIAL[] = "INITIAL";
static const char __pyx_k_INVALID[] = "<INVALID>";
static const char __pyx_k_I_C_V_Q[] = "I\002\307\257C\003\002\002\002\307\260\307\262\007V\002\002\307\261\307\263\007Q\002";
static const char __pyx_k_L_N_P_R[] = "\002\002\002L\310\215\003\002\002\002N\310\247\003\002\002\002P\310\263\003\002\002\002R\311\211";
static const char __pyx_k_PARTIAL[] = "PARTIAL";
static const char __pyx_k_comment[] = "comment";
static const char __pyx_k_connect[] = "'connect'";
static const char __pyx_k_consume[] = "consume";
static const char __pyx_k_element[] = "element";
static const char __pyx_k_extends[] = "'extends'";
static const char __pyx_k_h_j_l_n[] = "h\313\266\003\002\002\002j\314\224\003\002\002\002l\314\262\003\002\002\002n\315\213\003";
static const char __pyx_k_initial[] = "'initial'";
static const char __pyx_k_input_2[] = "_input";
static const char __pyx_k_input_3[] = "'input'";
static const char __pyx_k_l7_h5_v[] = "\005l7\002\311\205\311\212\005h5\002\311\206\311\207\005v<\002\311\207\311\210";
static const char __pyx_k_n8_0_n8[] = "\314\274\005n8\002\314\274\314\275\0070\002\002\314\275\315\205\005n8\002\314\276";
static const char __pyx_k_p_r_t_v[] = "\002\002\002p\315\215\003\002\002\002r\315\240\003\002\002\002t\316\226\003\002\002\002v\316\231";
static const char __pyx_k_package[] = "'package'";
static const char __pyx_k_partial[] = "'partial'";
static const char __pyx_k_prepare[] = "__prepare__";
static const char __pyx_k_prevctx[] = "_prevctx";
static const char __pyx_k_primary[] = "primary";
static const char __pyx_k_recover[] = "recover";
static const char __pyx_k_sempred[] = "sempred";
static const char __pyx_k_visitor[] = "visitor";
static const char __pyx_k_StringIO[] = "StringIO";
static const char __pyx_k_argument[] = "argument";
static const char __pyx_k_constant[] = "'constant'";
static const char __pyx_k_copyFrom[] = "copyFrom";
static const char __pyx_k_discrete[] = "'discrete'";
static const char __pyx_k_elsewhen[] = "'elsewhen'";
static const char __pyx_k_equation[] = "equation";
static const char __pyx_k_exitName[] = "exitName";
static const char __pyx_k_exitRule[] = "exitRule";
static const char __pyx_k_external[] = "'external'";
static const char __pyx_k_function[] = "'function'";
static const char __pyx_k_getToken[] = "getToken";
static const char __pyx_k_getvalue[] = "getvalue";
static const char __pyx_k_import_2[] = "'import'";
static const char __pyx_k_listener[] = "listener";
static const char __pyx_k_localctx[] = "localctx";
static const char __pyx_k_n8_0_X_2[] = "\005n8\002\314\241\314\247\0070\002\002\314\242\314\243\005X-\002\314\243\314\244";
static const char __pyx_k_operator[] = "'operator'";
static const char __pyx_k_output_2[] = "'output'";
static const char __pyx_k_precpred[] = "precpred";
static const char __pyx_k_qualname[] = "__qualname__";
static const char __pyx_k_REDECLARE[] = "REDECLARE";
static const char __pyx_k_RULE_expr[] = "RULE_expr";
static const char __pyx_k_RULE_name[] = "RULE_name";
static const char __pyx_k_algorithm[] = "'algorithm'";
static const char __pyx_k_comp_elem[] = "comp_elem";
static const char __pyx_k_connector[] = "'connector'";
static const char __pyx_k_enterName[] = "enterName";
static const char __pyx_k_enterRule[] = "enterRule";
static const char __pyx_k_enum_list[] = "enum_list";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_exception[] = "exception";
static const char __pyx_k_for_index[] = "for_index";
static const char __pyx_k_getTokens[] = "getTokens";
static const char __pyx_k_metaclass[] = "__metaclass__";
static const char __pyx_k_parameter[] = "'parameter'";
static const char __pyx_k_parentctx[] = "_parentctx";
static const char __pyx_k_predIndex[] = "predIndex";
static const char __pyx_k_protected[] = "'protected'";
static const char __pyx_k_redeclare[] = "'redeclare'";
static const char __pyx_k_ruleIndex[] = "ruleIndex";
static const char __pyx_k_ruleNames[] = "ruleNames";
static const char __pyx_k_statement[] = "statement";
static const char __pyx_k_subscript[] = "subscript";
static const char __pyx_k_typing_io[] = "typing.io";
static const char __pyx_k_visitName[] = "visitName";
static const char __pyx_k_annotation[] = "annotation";
static const char __pyx_k_class_elem[] = "class_elem";
static const char __pyx_k_class_type[] = "class_type";
static const char __pyx_k_equation_2[] = "'equation'";
static const char __pyx_k_errHandler[] = "_errHandler";
static const char __pyx_k_expandable[] = "'expandable'";
static const char __pyx_k_expression[] = "expression";
static const char __pyx_k_predicates[] = "_predicates";
static const char __pyx_k_startState[] = "_startState";
static const char __pyx_k_ExprContext[] = "ExprContext";
static const char __pyx_k_Modelica_g4[] = "Modelica.g4";
static const char __pyx_k_NameContext[] = "NameContext";
static const char __pyx_k_RuleContext[] = "RuleContext";
static const char __pyx_k_TokenStream[] = "TokenStream";
static const char __pyx_k_base_prefix[] = "base_prefix";
static const char __pyx_k_composition[] = "composition";
static const char __pyx_k_declaration[] = "declaration";
static const char __pyx_k_deserialize[] = "deserialize";
static const char __pyx_k_enumeration[] = "'enumeration'";
static const char __pyx_k_exitComment[] = "exitComment";
static const char __pyx_k_exitElement[] = "exitElement";
static const char __pyx_k_exitExpr_or[] = "exitExpr_or";
static const char __pyx_k_for_indices[] = "for_indices";
static const char __pyx_k_if_equation[] = "if_equation";
static const char __pyx_k_import_list[] = "import_list";
static const char __pyx_k_parentState[] = "_parentState";
static const char __pyx_k_replaceable[] = "'replaceable'";
static const char __pyx_k_reportError[] = "reportError";
static const char __pyx_k_reportMatch[] = "reportMatch";
static const char __pyx_k_type_prefix[] = "type_prefix";
static const char __pyx_k_ENCAPSULATED[] = "ENCAPSULATED";
static const char __pyx_k_RULE_comment[] = "RULE_comment";
static const char __pyx_k_RULE_element[] = "RULE_element";
static const char __pyx_k_RULE_primary[] = "RULE_primary";
static const char __pyx_k_annotation_2[] = "'annotation'";
static const char __pyx_k_checkVersion[] = "checkVersion";
static const char __pyx_k_element_list[] = "element_list";
static const char __pyx_k_encapsulated[] = "'encapsulated'";
static const char __pyx_k_enterComment[] = "enterComment";
static const char __pyx_k_enterElement[] = "enterElement";
static const char __pyx_k_enterExpr_or[] = "enterExpr_or";
static const char __pyx_k_exitArgument[] = "exitArgument";
static const char __pyx_k_exitEquation[] = "exitEquation";
static const char __pyx_k_exitExpr_add[] = "exitExpr_add";
static const char __pyx_k_exitExpr_and[] = "exitExpr_and";
static const char __pyx_k_exitExpr_exp[] = "exitExpr_exp";
static const char __pyx_k_exitExpr_mul[] = "exitExpr_mul";
static const char __pyx_k_exitExpr_neg[] = "exitExpr_neg";
static const char __pyx_k_exitExpr_not[] = "exitExpr_not";
static const char __pyx_k_exitExpr_rel[] = "exitExpr_rel";
static const char __pyx_k_expr_sempred[] = "expr_sempred";
static const char __pyx_k_for_equation[] = "for_equation";
static const char __pyx_k_getRuleIndex[] = "getRuleIndex";
static const char __pyx_k_if_statement[] = "if_statement";
static const char __pyx_k_literalNames[] = "literalNames";
static const char __pyx_k_modification[] = "modification";
static const char __pyx_k_visitComment[] = "visitComment";
static const char __pyx_k_visitElement[] = "visitElement";
static const char __pyx_k_visitExpr_or[] = "visitExpr_or";
static const char __pyx_k_A_A_B_B_C_C_D[] = ";\004<\t<\004=\t=\004>\t>\004?\t?\004@\t@\004A\tA\004B\tB\004C\tC\004D\t";
static const char __pyx_k_RULE_argument[] = "RULE_argument";
static const char __pyx_k_RULE_equation[] = "RULE_equation";
static const char __pyx_k_argument_list[] = "argument_list";
static const char __pyx_k_constrainedby[] = "'constrainedby'";
static const char __pyx_k_enterArgument[] = "enterArgument";
static const char __pyx_k_enterEquation[] = "enterEquation";
static const char __pyx_k_enterExpr_add[] = "enterExpr_add";
static const char __pyx_k_enterExpr_and[] = "enterExpr_and";
static const char __pyx_k_enterExpr_exp[] = "enterExpr_exp";
static const char __pyx_k_enterExpr_mul[] = "enterExpr_mul";
static const char __pyx_k_enterExpr_neg[] = "enterExpr_neg";
static const char __pyx_k_enterExpr_not[] = "enterExpr_not";
static const char __pyx_k_enterExpr_rel[] = "enterExpr_rel";
static const char __pyx_k_enterOuterAlt[] = "enterOuterAlt";
static const char __pyx_k_exitEnum_list[] = "exitEnum_list";
static const char __pyx_k_exitFor_index[] = "exitFor_index";
static const char __pyx_k_exitStatement[] = "exitStatement";
static const char __pyx_k_exitSubscript[] = "exitSubscript";
static const char __pyx_k_for_statement[] = "for_statement";
static const char __pyx_k_import_clause[] = "import_clause";
static const char __pyx_k_invokingState[] = "invokingState";
static const char __pyx_k_recoverInline[] = "recoverInline";
static const char __pyx_k_serializedATN[] = "serializedATN";
static const char __pyx_k_symbolicNames[] = "symbolicNames";
static const char __pyx_k_visitArgument[] = "visitArgument";
static const char __pyx_k_visitChildren[] = "visitChildren";
static const char __pyx_k_visitEquation[] = "visitEquation";
static const char __pyx_k_visitExpr_add[] = "visitExpr_add";
static const char __pyx_k_visitExpr_and[] = "visitExpr_and";
static const char __pyx_k_visitExpr_exp[] = "visitExpr_exp";
static const char __pyx_k_visitExpr_mul[] = "visitExpr_mul";
static const char __pyx_k_visitExpr_neg[] = "visitExpr_neg";
static const char __pyx_k_visitExpr_not[] = "visitExpr_not";
static const char __pyx_k_visitExpr_rel[] = "visitExpr_rel";
static const char __pyx_k_when_equation[] = "when_equation";
static const char __pyx_k_CommentContext[] = "CommentContext";
static const char __pyx_k_ElementContext[] = "ElementContext";
static const char __pyx_k_Expr_orContext[] = "Expr_orContext";
static const char __pyx_k_ModelicaParser[] = "ModelicaParser";
static const char __pyx_k_PrimaryContext[] = "PrimaryContext";
static const char __pyx_k_RULE_enum_list[] = "RULE_enum_list";
static const char __pyx_k_RULE_for_index[] = "RULE_for_index";
static const char __pyx_k_RULE_statement[] = "RULE_statement";
static const char __pyx_k_RULE_subscript[] = "RULE_subscript";
static const char __pyx_k_class_prefixes[] = "class_prefixes";
static const char __pyx_k_component_list[] = "component_list";
static const char __pyx_k_connect_clause[] = "connect_clause";
static const char __pyx_k_decisionsToDFA[] = "decisionsToDFA";
static const char __pyx_k_enterEnum_list[] = "enterEnum_list";
static const char __pyx_k_enterFor_index[] = "enterFor_index";
static const char __pyx_k_enterStatement[] = "enterStatement";
static const char __pyx_k_enterSubscript[] = "enterSubscript";
static const char __pyx_k_exitAnnotation[] = "exitAnnotation";
static const char __pyx_k_exitClass_type[] = "exitClass_type";
static const char __pyx_k_ext_annotation[] = "ext_annotation";
static const char __pyx_k_extends_clause[] = "extends_clause";
static const char __pyx_k_named_argument[] = "named_argument";
static const char __pyx_k_parseListeners[] = "_parseListeners";
static const char __pyx_k_string_comment[] = "string_comment";
static const char __pyx_k_type_specifier[] = "type_specifier";
static const char __pyx_k_visitEnum_list[] = "visitEnum_list";
static const char __pyx_k_visitFor_index[] = "visitFor_index";
static const char __pyx_k_visitStatement[] = "visitStatement";
static const char __pyx_k_visitSubscript[] = "visitSubscript";
static const char __pyx_k_when_statement[] = "when_statement";
static const char __pyx_k_0_0_0_0_0_0_0_0[] = "0\0030\0030\0070\313\206\n0\0140\0160\313\211\0130\003";
static const char __pyx_k_2_2_2_2_2_2_3_3[] = "\0032\0072\313\237\n2\0142\0162\313\242\0132\0033\0033";
static const char __pyx_k_4_4_4_4_4_4_4_4[] = "4\313\257\n4\0144\0164\313\262\0134\0034\0034\0034\003";
static const char __pyx_k_5_5_5_5_5_5_5_5[] = "\314\200\0135\0035\0035\0035\0035\0035\0035\0075\314\210";
static const char __pyx_k_6_6_6_6_6_6_6_6[] = "6\0076\314\233\n6\0146\0166\314\236\0136\0036\0036\003";
static const char __pyx_k_ATNDeserializer[] = "ATNDeserializer";
static const char __pyx_k_ArgumentContext[] = "ArgumentContext";
static const char __pyx_k_EquationContext[] = "EquationContext";
static const char __pyx_k_Expr_addContext[] = "Expr_addContext";
static const char __pyx_k_Expr_andContext[] = "Expr_andContext";
static const char __pyx_k_Expr_expContext[] = "Expr_expContext";
static const char __pyx_k_Expr_mulContext[] = "Expr_mulContext";
static const char __pyx_k_Expr_negContext[] = "Expr_negContext";
static const char __pyx_k_Expr_notContext[] = "Expr_notContext";
static const char __pyx_k_Expr_relContext[] = "Expr_relContext";
static const char __pyx_k_RULE_annotation[] = "RULE_annotation";
static const char __pyx_k_RULE_class_type[] = "RULE_class_type";
static const char __pyx_k_RULE_expression[] = "RULE_expression";
static const char __pyx_k_UNSIGNED_NUMBER[] = "UNSIGNED_NUMBER";
static const char __pyx_k_adaptivePredict[] = "adaptivePredict";
static const char __pyx_k_class_specifier[] = "class_specifier";
static const char __pyx_k_comp_annotation[] = "comp_annotation";
static const char __pyx_k_decisionToState[] = "decisionToState";
static const char __pyx_k_enterAnnotation[] = "enterAnnotation";
static const char __pyx_k_enterClass_type[] = "enterClass_type";
static const char __pyx_k_exitBase_prefix[] = "exitBase_prefix";
static const char __pyx_k_exitComposition[] = "exitComposition";
static const char __pyx_k_exitDeclaration[] = "exitDeclaration";
static const char __pyx_k_exitEquation_if[] = "exitEquation_if";
static const char __pyx_k_exitFor_indices[] = "exitFor_indices";
static const char __pyx_k_exitIf_equation[] = "exitIf_equation";
static const char __pyx_k_exitImport_list[] = "exitImport_list";
static const char __pyx_k_exitPrimary_end[] = "exitPrimary_end";
static const char __pyx_k_exitType_prefix[] = "exitType_prefix";
static const char __pyx_k_expression_list[] = "expression_list";
static const char __pyx_k_grammarFileName[] = "grammarFileName";
static const char __pyx_k_named_arguments[] = "named_arguments";
static const char __pyx_k_regular_element[] = "regular_element";
static const char __pyx_k_visitAnnotation[] = "visitAnnotation";
static const char __pyx_k_visitClass_type[] = "visitClass_type";
static const char __pyx_k_while_statement[] = "while_statement";
static const char __pyx_k_Enum_listContext[] = "Enum_listContext";
static const char __pyx_k_For_indexContext[] = "For_indexContext";
static const char __pyx_k_ParseTreeVisitor[] = "ParseTreeVisitor";
static const char __pyx_k_RULE_base_prefix[] = "RULE_base_prefix";
static const char __pyx_k_RULE_composition[] = "RULE_composition";
static const char __pyx_k_RULE_declaration[] = "RULE_declaration";
static const char __pyx_k_RULE_for_indices[] = "RULE_for_indices";
static const char __pyx_k_RULE_if_equation[] = "RULE_if_equation";
static const char __pyx_k_RULE_import_list[] = "RULE_import_list";
static const char __pyx_k_RULE_type_prefix[] = "RULE_type_prefix";
static const char __pyx_k_StatementContext[] = "StatementContext";
static const char __pyx_k_SubscriptContext[] = "SubscriptContext";
static const char __pyx_k_array_subscripts[] = "array_subscripts";
static const char __pyx_k_class_definition[] = "class_definition";
static const char __pyx_k_component_clause[] = "component_clause";
static const char __pyx_k_enterBase_prefix[] = "enterBase_prefix";
static const char __pyx_k_enterComposition[] = "enterComposition";
static const char __pyx_k_enterDeclaration[] = "enterDeclaration";
static const char __pyx_k_enterEquation_if[] = "enterEquation_if";
static const char __pyx_k_enterFor_indices[] = "enterFor_indices";
static const char __pyx_k_enterIf_equation[] = "enterIf_equation";
static const char __pyx_k_enterImport_list[] = "enterImport_list";
static const char __pyx_k_enterPrimary_end[] = "enterPrimary_end";
static const char __pyx_k_enterType_prefix[] = "enterType_prefix";
static const char __pyx_k_equation_options[] = "equation_options";
static const char __pyx_k_equation_section[] = "equation_section";
static const char __pyx_k_exitElement_list[] = "exitElement_list";
static const char __pyx_k_exitEquation_for[] = "exitEquation_for";
static const char __pyx_k_exitExpr_primary[] = "exitExpr_primary";
static const char __pyx_k_exitFor_equation[] = "exitFor_equation";
static const char __pyx_k_exitIf_statement[] = "exitIf_statement";
static const char __pyx_k_exitPrimary_true[] = "exitPrimary_true";
static const char __pyx_k_exitStatement_if[] = "exitStatement_if";
static const char __pyx_k_visitBase_prefix[] = "visitBase_prefix";
static const char __pyx_k_visitComposition[] = "visitComposition";
static const char __pyx_k_visitDeclaration[] = "visitDeclaration";
static const char __pyx_k_visitEquation_if[] = "visitEquation_if";
static const char __pyx_k_visitFor_indices[] = "visitFor_indices";
static const char __pyx_k_visitIf_equation[] = "visitIf_equation";
static const char __pyx_k_visitImport_list[] = "visitImport_list";
static const char __pyx_k_visitPrimary_end[] = "visitPrimary_end";
static const char __pyx_k_visitType_prefix[] = "visitType_prefix";
static const char __pyx_k_1_1_1_1_1_1_1_2_2[] = "\n1\0141\0161\313\227\0131\0031\0031\0031\0032\0032";
static const char __pyx_k_3_3_3_4_4_4_4_4_4[] = "\0033\0053\313\247\n3\0034\0034\0034\0034\0034\0034\007";
static const char __pyx_k_5_5_5_5_5_5_5_5_2[] = "\n5\0145\0165\314\213\0135\0075\314\215\n5\0145\0165";
static const char __pyx_k_5_5_5_5_6_6_6_6_6[] = "\314\220\0135\0035\0035\0035\0036\0036\0036\0036\0036\003";
static const char __pyx_k_6_6_6_6_6_6_6_6_2[] = "6\0036\0036\0036\0076\314\246\n6\0146\0166\314\251\013";
static const char __pyx_k_6_6_6_6_6_6_6_6_3[] = "6\0076\314\253\n6\0146\0166\314\256\0136\0036\0036\003";
static const char __pyx_k_AnnotationContext[] = "AnnotationContext";
static const char __pyx_k_Class_typeContext[] = "Class_typeContext";
static const char __pyx_k_ExpressionContext[] = "ExpressionContext";
static const char __pyx_k_ParseTreeListener[] = "ParseTreeListener";
static const char __pyx_k_ParserRuleContext[] = "ParserRuleContext";
static const char __pyx_k_RULE_element_list[] = "RULE_element_list";
static const char __pyx_k_RULE_for_equation[] = "RULE_for_equation";
static const char __pyx_k_RULE_if_statement[] = "RULE_if_statement";
static const char __pyx_k_RULE_modification[] = "RULE_modification";
static const char __pyx_k_algorithm_section[] = "algorithm_section";
static const char __pyx_k_component_clause1[] = "component_clause1";
static const char __pyx_k_enterElement_list[] = "enterElement_list";
static const char __pyx_k_enterEquation_for[] = "enterEquation_for";
static const char __pyx_k_enterExpr_primary[] = "enterExpr_primary";
static const char __pyx_k_enterFor_equation[] = "enterFor_equation";
static const char __pyx_k_enterIf_statement[] = "enterIf_statement";
static const char __pyx_k_enterPrimary_true[] = "enterPrimary_true";
static const char __pyx_k_enterStatement_if[] = "enterStatement_if";
static const char __pyx_k_exitArgument_list[] = "exitArgument_list";
static const char __pyx_k_exitEquation_when[] = "exitEquation_when";
static const char __pyx_k_exitExpression_if[] = "exitExpression_if";
static const char __pyx_k_exitFor_statement[] = "exitFor_statement";
static const char __pyx_k_exitImport_clause[] = "exitImport_clause";
static const char __pyx_k_exitPrimary_false[] = "exitPrimary_false";
static const char __pyx_k_exitStatement_for[] = "exitStatement_for";
static const char __pyx_k_exitWhen_equation[] = "exitWhen_equation";
static const char __pyx_k_function_argument[] = "function_argument";
static const char __pyx_k_simple_expression[] = "simple_expression";
static const char __pyx_k_statement_options[] = "statement_options";
static const char __pyx_k_stored_definition[] = "stored_definition";
static const char __pyx_k_visitElement_list[] = "visitElement_list";
static const char __pyx_k_visitEquation_for[] = "visitEquation_for";
static const char __pyx_k_visitExpr_primary[] = "visitExpr_primary";
static const char __pyx_k_visitFor_equation[] = "visitFor_equation";
static const char __pyx_k_visitIf_statement[] = "visitIf_statement";
static const char __pyx_k_visitPrimary_true[] = "visitPrimary_true";
static const char __pyx_k_visitStatement_if[] = "visitStatement_if";
static const char __pyx_k_Base_prefixContext[] = "Base_prefixContext";
static const char __pyx_k_CompositionContext[] = "CompositionContext";
static const char __pyx_k_DeclarationContext[] = "DeclarationContext";
static const char __pyx_k_Equation_ifContext[] = "Equation_ifContext";
static const char __pyx_k_For_indicesContext[] = "For_indicesContext";
static const char __pyx_k_INVALID_ALT_NUMBER[] = "INVALID_ALT_NUMBER";
static const char __pyx_k_If_equationContext[] = "If_equationContext";
static const char __pyx_k_Import_listContext[] = "Import_listContext";
static const char __pyx_k_ParserATNSimulator[] = "ParserATNSimulator";
static const char __pyx_k_Primary_endContext[] = "Primary_endContext";
static const char __pyx_k_RULE_argument_list[] = "RULE_argument_list";
static const char __pyx_k_RULE_for_statement[] = "RULE_for_statement";
static const char __pyx_k_RULE_import_clause[] = "RULE_import_clause";
static const char __pyx_k_RULE_when_equation[] = "RULE_when_equation";
static const char __pyx_k_Type_prefixContext[] = "Type_prefixContext";
static const char __pyx_k_class_modification[] = "class_modification";
static const char __pyx_k_enterArgument_list[] = "enterArgument_list";
static const char __pyx_k_enterEquation_when[] = "enterEquation_when";
static const char __pyx_k_enterExpression_if[] = "enterExpression_if";
static const char __pyx_k_enterFor_statement[] = "enterFor_statement";
static const char __pyx_k_enterImport_clause[] = "enterImport_clause";
static const char __pyx_k_enterPrimary_false[] = "enterPrimary_false";
static const char __pyx_k_enterRecursionRule[] = "enterRecursionRule";
static const char __pyx_k_enterStatement_for[] = "enterStatement_for";
static const char __pyx_k_enterWhen_equation[] = "enterWhen_equation";
static const char __pyx_k_exitClass_prefixes[] = "exitClass_prefixes";
static const char __pyx_k_exitClass_spec_der[] = "exitClass_spec_der";
static const char __pyx_k_exitComponent_list[] = "exitComponent_list";
static const char __pyx_k_exitConnect_clause[] = "exitConnect_clause";
static const char __pyx_k_exitExtends_clause[] = "exitExtends_clause";
static const char __pyx_k_exitNamed_argument[] = "exitNamed_argument";
static const char __pyx_k_exitPrimary_string[] = "exitPrimary_string";
static const char __pyx_k_exitStatement_when[] = "exitStatement_when";
static const char __pyx_k_exitString_comment[] = "exitString_comment";
static const char __pyx_k_exitType_specifier[] = "exitType_specifier";
static const char __pyx_k_exitWhen_statement[] = "exitWhen_statement";
static const char __pyx_k_function_arguments[] = "function_arguments";
static const char __pyx_k_function_call_args[] = "function_call_args";
static const char __pyx_k_sharedContextCache[] = "sharedContextCache";
static const char __pyx_k_visitArgument_list[] = "visitArgument_list";
static const char __pyx_k_visitEquation_when[] = "visitEquation_when";
static const char __pyx_k_visitExpression_if[] = "visitExpression_if";
static const char __pyx_k_visitFor_statement[] = "visitFor_statement";
static const char __pyx_k_visitImport_clause[] = "visitImport_clause";
static const char __pyx_k_visitPrimary_false[] = "visitPrimary_false";
static const char __pyx_k_visitStatement_for[] = "visitStatement_for";
static const char __pyx_k_visitWhen_equation[] = "visitWhen_equation";
static const char __pyx_k_0_0_0_1_1_1_1_1_1_1[] = "0\0030\0030\0031\0031\0031\0031\0031\0031\0071\313\224";
static const char __pyx_k_0_0_1_1_2_2_3_3_4_4[] = "/\t/\0040\t0\0041\t1\0042\t2\0043\t3\0044\t4";
static const char __pyx_k_5_5_5_5_5_5_5_5_5_5[] = "5\0035\0035\0035\0035\0035\0075\313\275\n5\0145\0165";
static const char __pyx_k_5_5_6_6_7_7_8_8_9_9[] = "\0045\t5\0046\t6\0047\t7\0048\t8\0049\t9\004:\t:\004;\t";
static const char __pyx_k_Element_listContext[] = "Element_listContext";
static const char __pyx_k_Equation_forContext[] = "Equation_forContext";
static const char __pyx_k_Expr_primaryContext[] = "Expr_primaryContext";
static const char __pyx_k_For_equationContext[] = "For_equationContext";
static const char __pyx_k_If_statementContext[] = "If_statementContext";
static const char __pyx_k_ModelicaParser_expr[] = "ModelicaParser.expr";
static const char __pyx_k_ModelicaParser_name[] = "ModelicaParser.name";
static const char __pyx_k_ModificationContext[] = "ModificationContext";
static const char __pyx_k_Primary_trueContext[] = "Primary_trueContext";
static const char __pyx_k_RULE_class_prefixes[] = "RULE_class_prefixes";
static const char __pyx_k_RULE_component_list[] = "RULE_component_list";
static const char __pyx_k_RULE_connect_clause[] = "RULE_connect_clause";
static const char __pyx_k_RULE_extends_clause[] = "RULE_extends_clause";
static const char __pyx_k_RULE_named_argument[] = "RULE_named_argument";
static const char __pyx_k_RULE_string_comment[] = "RULE_string_comment";
static const char __pyx_k_RULE_type_specifier[] = "RULE_type_specifier";
static const char __pyx_k_RULE_when_statement[] = "RULE_when_statement";
static const char __pyx_k_Statement_ifContext[] = "Statement_ifContext";
static const char __pyx_k_antlr4_error_Errors[] = "antlr4.error.Errors";
static const char __pyx_k_component_reference[] = "component_reference";
static const char __pyx_k_condition_attribute[] = "condition_attribute";
static const char __pyx_k_constraining_clause[] = "constraining_clause";
static const char __pyx_k_element_replaceable[] = "element_replaceable";
static const char __pyx_k_enterClass_prefixes[] = "enterClass_prefixes";
static const char __pyx_k_enterClass_spec_der[] = "enterClass_spec_der";
static const char __pyx_k_enterComponent_list[] = "enterComponent_list";
static const char __pyx_k_enterConnect_clause[] = "enterConnect_clause";
static const char __pyx_k_enterExtends_clause[] = "enterExtends_clause";
static const char __pyx_k_enterNamed_argument[] = "enterNamed_argument";
static const char __pyx_k_enterPrimary_string[] = "enterPrimary_string";
static const char __pyx_k_enterStatement_when[] = "enterStatement_when";
static const char __pyx_k_enterString_comment[] = "enterString_comment";
static const char __pyx_k_enterType_specifier[] = "enterType_specifier";
static const char __pyx_k_enterWhen_statement[] = "enterWhen_statement";
static const char __pyx_k_enumeration_literal[] = "enumeration_literal";
static const char __pyx_k_exitClass_spec_base[] = "exitClass_spec_base";
static const char __pyx_k_exitClass_spec_comp[] = "exitClass_spec_comp";
static const char __pyx_k_exitClass_spec_enum[] = "exitClass_spec_enum";
static const char __pyx_k_exitEquation_simple[] = "exitEquation_simple";
static const char __pyx_k_exitExpression_list[] = "exitExpression_list";
static const char __pyx_k_exitNamed_arguments[] = "exitNamed_arguments";
static const char __pyx_k_exitPrimary_initial[] = "exitPrimary_initial";
static const char __pyx_k_exitRegular_element[] = "exitRegular_element";
static const char __pyx_k_exitStatement_break[] = "exitStatement_break";
static const char __pyx_k_exitStatement_while[] = "exitStatement_while";
static const char __pyx_k_exitWhile_statement[] = "exitWhile_statement";
static const char __pyx_k_getTypedRuleContext[] = "getTypedRuleContext";
static const char __pyx_k_replaceable_element[] = "replaceable_element";
static const char __pyx_k_visitClass_prefixes[] = "visitClass_prefixes";
static const char __pyx_k_visitClass_spec_der[] = "visitClass_spec_der";
static const char __pyx_k_visitComponent_list[] = "visitComponent_list";
static const char __pyx_k_visitConnect_clause[] = "visitConnect_clause";
static const char __pyx_k_visitExtends_clause[] = "visitExtends_clause";
static const char __pyx_k_visitNamed_argument[] = "visitNamed_argument";
static const char __pyx_k_visitPrimary_string[] = "visitPrimary_string";
static const char __pyx_k_visitStatement_when[] = "visitStatement_when";
static const char __pyx_k_visitString_comment[] = "visitString_comment";
static const char __pyx_k_visitType_specifier[] = "visitType_specifier";
static const char __pyx_k_visitWhen_statement[] = "visitWhen_statement";
static const char __pyx_k_Argument_listContext[] = "Argument_listContext";
static const char __pyx_k_Equation_whenContext[] = "Equation_whenContext";
static const char __pyx_k_Expression_ifContext[] = "Expression_ifContext";
static const char __pyx_k_For_statementContext[] = "For_statementContext";
static const char __pyx_k_Import_clauseContext[] = "Import_clauseContext";
static const char __pyx_k_NoViableAltException[] = "NoViableAltException";
static const char __pyx_k_Primary_falseContext[] = "Primary_falseContext";
static const char __pyx_k_RULE_class_specifier[] = "RULE_class_specifier";
static const char __pyx_k_RULE_expression_list[] = "RULE_expression_list";
static const char __pyx_k_RULE_named_arguments[] = "RULE_named_arguments";
static const char __pyx_k_RULE_regular_element[] = "RULE_regular_element";
static const char __pyx_k_RULE_while_statement[] = "RULE_while_statement";
static const char __pyx_k_RecognitionException[] = "RecognitionException";
static const char __pyx_k_Statement_forContext[] = "Statement_forContext";
static const char __pyx_k_When_equationContext[] = "When_equationContext";
static const char __pyx_k_element_modification[] = "element_modification";
static const char __pyx_k_enterClass_spec_base[] = "enterClass_spec_base";
static const char __pyx_k_enterClass_spec_comp[] = "enterClass_spec_comp";
static const char __pyx_k_enterClass_spec_enum[] = "enterClass_spec_enum";
static const char __pyx_k_enterEquation_simple[] = "enterEquation_simple";
static const char __pyx_k_enterExpression_list[] = "enterExpression_list";
static const char __pyx_k_enterNamed_arguments[] = "enterNamed_arguments";
static const char __pyx_k_enterPrimary_initial[] = "enterPrimary_initial";
static const char __pyx_k_enterRegular_element[] = "enterRegular_element";
static const char __pyx_k_enterStatement_break[] = "enterStatement_break";
static const char __pyx_k_enterStatement_while[] = "enterStatement_while";
static const char __pyx_k_enterWhile_statement[] = "enterWhile_statement";
static const char __pyx_k_exitArray_subscripts[] = "exitArray_subscripts";
static const char __pyx_k_exitClass_definition[] = "exitClass_definition";
static const char __pyx_k_exitComponent_clause[] = "exitComponent_clause";
static const char __pyx_k_exitEquation_section[] = "exitEquation_section";
static const char __pyx_k_exitPrimary_function[] = "exitPrimary_function";
static const char __pyx_k_exitStatement_return[] = "exitStatement_return";
static const char __pyx_k_getTypedRuleContexts[] = "getTypedRuleContexts";
static const char __pyx_k_triggerExitRuleEvent[] = "triggerExitRuleEvent";
static const char __pyx_k_visitClass_spec_base[] = "visitClass_spec_base";
static const char __pyx_k_visitClass_spec_comp[] = "visitClass_spec_comp";
static const char __pyx_k_visitClass_spec_enum[] = "visitClass_spec_enum";
static const char __pyx_k_visitEquation_simple[] = "visitEquation_simple";
static const char __pyx_k_visitExpression_list[] = "visitExpression_list";
static const char __pyx_k_visitNamed_arguments[] = "visitNamed_arguments";
static const char __pyx_k_visitPrimary_initial[] = "visitPrimary_initial";
static const char __pyx_k_visitRegular_element[] = "visitRegular_element";
static const char __pyx_k_visitStatement_break[] = "visitStatement_break";
static const char __pyx_k_visitStatement_while[] = "visitStatement_while";
static const char __pyx_k_visitWhile_statement[] = "visitWhile_statement";
static const char __pyx_k_8_8_9_9_9_9_9_9_9_9_9[] = "\0058\315\214\n8\0039\0039\0039\0039\0039\0059\315\223\n9\0059\315\225\n9\003";
static const char __pyx_k_C_C_C_C_C_C_C_C_C_D_D[] = "\003C\003C\005C\317\223\nC\003C\003C\003C\005C\317\230\nC\003D\005D\317\233\n";
static const char __pyx_k_Class_prefixesContext[] = "Class_prefixesContext";
static const char __pyx_k_Class_spec_derContext[] = "Class_spec_derContext";
static const char __pyx_k_Component_listContext[] = "Component_listContext";
static const char __pyx_k_Connect_clauseContext[] = "Connect_clauseContext";
static const char __pyx_k_Extends_clauseContext[] = "Extends_clauseContext";
static const char __pyx_k_I_I_I_I_I_I_I_I_J_J_J[] = "I\007I\320\202\nI\014I\016I\320\205\013I\005I\320\207\nI\003J\003J\003J\003";
static const char __pyx_k_ModelicaParser___init[] = "ModelicaParser.__init__";
static const char __pyx_k_Named_argumentContext[] = "Named_argumentContext";
static const char __pyx_k_Primary_stringContext[] = "Primary_stringContext";
static const char __pyx_k_RULE_array_subscripts[] = "RULE_array_subscripts";
static const char __pyx_k_RULE_class_definition[] = "RULE_class_definition";
static const char __pyx_k_RULE_component_clause[] = "RULE_component_clause";
static const char __pyx_k_RULE_equation_options[] = "RULE_equation_options";
static const char __pyx_k_RULE_equation_section[] = "RULE_equation_section";
static const char __pyx_k_Statement_whenContext[] = "Statement_whenContext";
static const char __pyx_k_String_commentContext[] = "String_commentContext";
static const char __pyx_k_Type_specifierContext[] = "Type_specifierContext";
static const char __pyx_k_When_statementContext[] = "When_statementContext";
static const char __pyx_k_component_declaration[] = "component_declaration";
static const char __pyx_k_element_redeclaration[] = "element_redeclaration";
static const char __pyx_k_enterArray_subscripts[] = "enterArray_subscripts";
static const char __pyx_k_enterClass_definition[] = "enterClass_definition";
static const char __pyx_k_enterComponent_clause[] = "enterComponent_clause";
static const char __pyx_k_enterEquation_section[] = "enterEquation_section";
static const char __pyx_k_enterPrimary_function[] = "enterPrimary_function";
static const char __pyx_k_enterStatement_return[] = "enterStatement_return";
static const char __pyx_k_exitAlgorithm_section[] = "exitAlgorithm_section";
static const char __pyx_k_exitArgument_function[] = "exitArgument_function";
static const char __pyx_k_exitComponent_clause1[] = "exitComponent_clause1";
static const char __pyx_k_exitEquation_function[] = "exitEquation_function";
static const char __pyx_k_exitExpression_simple[] = "exitExpression_simple";
static const char __pyx_k_exitSimple_expression[] = "exitSimple_expression";
static const char __pyx_k_exitStored_definition[] = "exitStored_definition";
static const char __pyx_k_visitArray_subscripts[] = "visitArray_subscripts";
static const char __pyx_k_visitClass_definition[] = "visitClass_definition";
static const char __pyx_k_visitComponent_clause[] = "visitComponent_clause";
static const char __pyx_k_visitEquation_section[] = "visitEquation_section";
static const char __pyx_k_visitPrimary_function[] = "visitPrimary_function";
static const char __pyx_k_visitStatement_return[] = "visitStatement_return";
static const char __pyx_k_Class_spec_baseContext[] = "Class_spec_baseContext";
static const char __pyx_k_Class_spec_compContext[] = "Class_spec_compContext";
static const char __pyx_k_Class_spec_enumContext[] = "Class_spec_enumContext";
static const char __pyx_k_Class_specifierContext[] = "Class_specifierContext";
static const char __pyx_k_Equation_simpleContext[] = "Equation_simpleContext";
static const char __pyx_k_Expression_listContext[] = "Expression_listContext";
static const char __pyx_k_ModelicaParser_comment[] = "ModelicaParser.comment";
static const char __pyx_k_ModelicaParser_element[] = "ModelicaParser.element";
static const char __pyx_k_ModelicaParser_primary[] = "ModelicaParser.primary";
static const char __pyx_k_ModelicaParser_sempred[] = "ModelicaParser.sempred";
static const char __pyx_k_Named_argumentsContext[] = "Named_argumentsContext";
static const char __pyx_k_PredictionContextCache[] = "PredictionContextCache";
static const char __pyx_k_Primary_initialContext[] = "Primary_initialContext";
static const char __pyx_k_RULE_algorithm_section[] = "RULE_algorithm_section";
static const char __pyx_k_RULE_component_clause1[] = "RULE_component_clause1";
static const char __pyx_k_RULE_function_argument[] = "RULE_function_argument";
static const char __pyx_k_RULE_simple_expression[] = "RULE_simple_expression";
static const char __pyx_k_RULE_statement_options[] = "RULE_statement_options";
static const char __pyx_k_RULE_stored_definition[] = "RULE_stored_definition";
static const char __pyx_k_Regular_elementContext[] = "Regular_elementContext";
static const char __pyx_k_Statement_breakContext[] = "Statement_breakContext";
static const char __pyx_k_Statement_whileContext[] = "Statement_whileContext";
static const char __pyx_k_While_statementContext[] = "While_statementContext";
static const char __pyx_k_component_declaration1[] = "component_declaration1";
static const char __pyx_k_enterAlgorithm_section[] = "enterAlgorithm_section";
static const char __pyx_k_enterArgument_function[] = "enterArgument_function";
static const char __pyx_k_enterComponent_clause1[] = "enterComponent_clause1";
static const char __pyx_k_enterEquation_function[] = "enterEquation_function";
static const char __pyx_k_enterExpression_simple[] = "enterExpression_simple";
static const char __pyx_k_enterSimple_expression[] = "enterSimple_expression";
static const char __pyx_k_enterStored_definition[] = "enterStored_definition";
static const char __pyx_k_exitClass_modification[] = "exitClass_modification";
static const char __pyx_k_exitClass_spec_extends[] = "exitClass_spec_extends";
static const char __pyx_k_exitFunction_arguments[] = "exitFunction_arguments";
static const char __pyx_k_exitFunction_call_args[] = "exitFunction_call_args";
static const char __pyx_k_exitModification_class[] = "exitModification_class";
static const char __pyx_k_exitPrimary_derivative[] = "exitPrimary_derivative";
static const char __pyx_k_external_function_call[] = "external_function_call";
static const char __pyx_k_language_specification[] = "language_specification";
static const char __pyx_k_output_expression_list[] = "output_expression_list";
static const char __pyx_k_short_class_definition[] = "short_class_definition";
static const char __pyx_k_visitAlgorithm_section[] = "visitAlgorithm_section";
static const char __pyx_k_visitArgument_function[] = "visitArgument_function";
static const char __pyx_k_visitComponent_clause1[] = "visitComponent_clause1";
static const char __pyx_k_visitEquation_function[] = "visitEquation_function";
static const char __pyx_k_visitExpression_simple[] = "visitExpression_simple";
static const char __pyx_k_visitSimple_expression[] = "visitSimple_expression";
static const char __pyx_k_visitStored_definition[] = "visitStored_definition";
static const char __pyx_k_Array_subscriptsContext[] = "Array_subscriptsContext";
static const char __pyx_k_Class_definitionContext[] = "Class_definitionContext";
static const char __pyx_k_Component_clauseContext[] = "Component_clauseContext";
static const char __pyx_k_D_D_D_D_D_D_D_D_E_E_E_E[] = "D\003D\003D\007D\317\237\nD\014D\016D\317\242\013D\003E\003E\003E\007E\317\247";
static const char __pyx_k_E_E_E_E_F_F_F_F_F_F_F_F[] = "\nE\014E\016E\317\252\013E\003F\003F\003F\003F\007F\317\260\nF\014F\016F\317\263";
static const char __pyx_k_Equation_optionsContext[] = "Equation_optionsContext";
static const char __pyx_k_Equation_sectionContext[] = "Equation_sectionContext";
static const char __pyx_k_ModelicaParser_argument[] = "ModelicaParser.argument";
static const char __pyx_k_ModelicaParser_equation[] = "ModelicaParser.equation";
static const char __pyx_k_No_predicate_with_index[] = "No predicate with index:";
static const char __pyx_k_Primary_functionContext[] = "Primary_functionContext";
static const char __pyx_k_RULE_class_modification[] = "RULE_class_modification";
static const char __pyx_k_RULE_function_arguments[] = "RULE_function_arguments";
static const char __pyx_k_RULE_function_call_args[] = "RULE_function_call_args";
static const char __pyx_k_Statement_returnContext[] = "Statement_returnContext";
static const char __pyx_k_enterClass_modification[] = "enterClass_modification";
static const char __pyx_k_enterClass_spec_extends[] = "enterClass_spec_extends";
static const char __pyx_k_enterFunction_arguments[] = "enterFunction_arguments";
static const char __pyx_k_enterFunction_call_args[] = "enterFunction_call_args";
static const char __pyx_k_enterModification_class[] = "enterModification_class";
static const char __pyx_k_enterPrimary_derivative[] = "enterPrimary_derivative";
static const char __pyx_k_exitArgument_expression[] = "exitArgument_expression";
static const char __pyx_k_exitComponent_reference[] = "exitComponent_reference";
static const char __pyx_k_exitCondition_attribute[] = "exitCondition_attribute";
static const char __pyx_k_exitConstraining_clause[] = "exitConstraining_clause";
static const char __pyx_k_exitElement_replaceable[] = "exitElement_replaceable";
static const char __pyx_k_exitEnumeration_literal[] = "exitEnumeration_literal";
static const char __pyx_k_exitReplaceable_element[] = "exitReplaceable_element";
static const char __pyx_k_pushNewRecursionContext[] = "pushNewRecursionContext";
static const char __pyx_k_stored_definition_class[] = "stored_definition_class";
static const char __pyx_k_unrollRecursionContexts[] = "unrollRecursionContexts";
static const char __pyx_k_visitClass_modification[] = "visitClass_modification";
static const char __pyx_k_visitClass_spec_extends[] = "visitClass_spec_extends";
static const char __pyx_k_visitFunction_arguments[] = "visitFunction_arguments";
static const char __pyx_k_visitFunction_call_args[] = "visitFunction_call_args";
static const char __pyx_k_visitModification_class[] = "visitModification_class";
static const char __pyx_k_visitPrimary_derivative[] = "visitPrimary_derivative";
static const char __pyx_k_Algorithm_sectionContext[] = "Algorithm_sectionContext";
static const char __pyx_k_Argument_functionContext[] = "Argument_functionContext";
static const char __pyx_k_Component_clause1Context[] = "Component_clause1Context";
static const char __pyx_k_Equation_functionContext[] = "Equation_functionContext";
static const char __pyx_k_Expression_simpleContext[] = "Expression_simpleContext";
static const char __pyx_k_FailedPredicateException[] = "FailedPredicateException";
static const char __pyx_k_Function_argumentContext[] = "Function_argumentContext";
static const char __pyx_k_ModelicaParser_enum_list[] = "ModelicaParser.enum_list";
static const char __pyx_k_ModelicaParser_for_index[] = "ModelicaParser.for_index";
static const char __pyx_k_ModelicaParser_statement[] = "ModelicaParser.statement";
static const char __pyx_k_ModelicaParser_subscript[] = "ModelicaParser.subscript";
static const char __pyx_k_RULE_component_reference[] = "RULE_component_reference";
static const char __pyx_k_RULE_condition_attribute[] = "RULE_condition_attribute";
static const char __pyx_k_RULE_constraining_clause[] = "RULE_constraining_clause";
static const char __pyx_k_RULE_element_replaceable[] = "RULE_element_replaceable";
static const char __pyx_k_RULE_enumeration_literal[] = "RULE_enumeration_literal";
static const char __pyx_k_RULE_replaceable_element[] = "RULE_replaceable_element";
static const char __pyx_k_Simple_expressionContext[] = "Simple_expressionContext";
static const char __pyx_k_Statement_optionsContext[] = "Statement_optionsContext";
static const char __pyx_k_Stored_definitionContext[] = "Stored_definitionContext";
static const char __pyx_k_enterArgument_expression[] = "enterArgument_expression";
static const char __pyx_k_enterComponent_reference[] = "enterComponent_reference";
static const char __pyx_k_enterCondition_attribute[] = "enterCondition_attribute";
static const char __pyx_k_enterConstraining_clause[] = "enterConstraining_clause";
static const char __pyx_k_enterElement_replaceable[] = "enterElement_replaceable";
static const char __pyx_k_enterEnumeration_literal[] = "enterEnumeration_literal";
static const char __pyx_k_enterReplaceable_element[] = "enterReplaceable_element";
static const char __pyx_k_exitElement_modification[] = "exitElement_modification";
static const char __pyx_k_visitArgument_expression[] = "visitArgument_expression";
static const char __pyx_k_visitComponent_reference[] = "visitComponent_reference";
static const char __pyx_k_visitCondition_attribute[] = "visitCondition_attribute";
static const char __pyx_k_visitConstraining_clause[] = "visitConstraining_clause";
static const char __pyx_k_visitElement_replaceable[] = "visitElement_replaceable";
static const char __pyx_k_visitEnumeration_literal[] = "visitEnumeration_literal";
static const char __pyx_k_visitReplaceable_element[] = "visitReplaceable_element";
static const char __pyx_k_6_7_7_7_7_7_7_7_8_8_8_8_8[] = "6\0037\0037\0037\0037\0037\0037\0037\0038\0038\0038\0038\0038";
static const char __pyx_k_8_8_8_8_8_8_8_8_8_8_8_8_8[] = "\0038\0038\0038\0038\0038\0078\315\204\n8\0148\0168\315\207\0138\0038\0038\0038";
static const char __pyx_k_A_A_A_A_A_A_A_B_B_B_B_C_C[] = "\003A\003A\007A\317\206\nA\014A\016A\317\211\013A\003B\003B\003B\003B\003C\003C";
static const char __pyx_k_Class_modificationContext[] = "Class_modificationContext";
static const char __pyx_k_Class_spec_extendsContext[] = "Class_spec_extendsContext";
static const char __pyx_k_D_E_E_F_F_G_G_H_H_I_I_J_J[] = "D\004E\tE\004F\tF\004G\tG\004H\tH\004I\tI\004J\tJ\003\002\003\002\005\002\302\227";
static const char __pyx_k_F_F_F_G_G_G_G_H_H_H_H_I_I[] = "\013F\003F\003F\003G\003G\005G\317\271\nG\003H\003H\005H\317\275\nH\003I\003I\003";
static const char __pyx_k_Function_argumentsContext[] = "Function_argumentsContext";
static const char __pyx_k_Function_call_argsContext[] = "Function_call_argsContext";
static const char __pyx_k_ModelicaParser_annotation[] = "ModelicaParser.annotation";
static const char __pyx_k_ModelicaParser_class_type[] = "ModelicaParser.class_type";
static const char __pyx_k_ModelicaParser_expression[] = "ModelicaParser.expression";
static const char __pyx_k_Modification_classContext[] = "Modification_classContext";
static const char __pyx_k_Primary_derivativeContext[] = "Primary_derivativeContext";
static const char __pyx_k_RULE_element_modification[] = "RULE_element_modification";
static const char __pyx_k_enterElement_modification[] = "enterElement_modification";
static const char __pyx_k_exitComponent_declaration[] = "exitComponent_declaration";
static const char __pyx_k_exitElement_redeclaration[] = "exitElement_redeclaration";
static const char __pyx_k_self_precpred_self__ctx_2[] = "self.precpred(self._ctx, 2)";
static const char __pyx_k_self_precpred_self__ctx_3[] = "self.precpred(self._ctx, 3)";
static const char __pyx_k_self_precpred_self__ctx_5[] = "self.precpred(self._ctx, 5)";
static const char __pyx_k_self_precpred_self__ctx_6[] = "self.precpred(self._ctx, 6)";
static const char __pyx_k_self_precpred_self__ctx_7[] = "self.precpred(self._ctx, 7)";
static const char __pyx_k_visitElement_modification[] = "visitElement_modification";
static const char __pyx_k_Argument_expressionContext[] = "Argument_expressionContext";
static const char __pyx_k_Component_referenceContext[] = "Component_referenceContext";
static const char __pyx_k_Condition_attributeContext[] = "Condition_attributeContext";
static const char __pyx_k_Constraining_clauseContext[] = "Constraining_clauseContext";
static const char __pyx_k_Element_replaceableContext[] = "Element_replaceableContext";
static const char __pyx_k_Enumeration_literalContext[] = "Enumeration_literalContext";
static const char __pyx_k_ModelicaParser_ExprContext[] = "ModelicaParser.ExprContext";
static const char __pyx_k_ModelicaParser_NameContext[] = "ModelicaParser.NameContext";
static const char __pyx_k_ModelicaParser_base_prefix[] = "ModelicaParser.base_prefix";
static const char __pyx_k_ModelicaParser_composition[] = "ModelicaParser.composition";
static const char __pyx_k_ModelicaParser_declaration[] = "ModelicaParser.declaration";
static const char __pyx_k_ModelicaParser_for_indices[] = "ModelicaParser.for_indices";
static const char __pyx_k_ModelicaParser_if_equation[] = "ModelicaParser.if_equation";
static const char __pyx_k_ModelicaParser_import_list[] = "ModelicaParser.import_list";
static const char __pyx_k_ModelicaParser_type_prefix[] = "ModelicaParser.type_prefix";
static const char __pyx_k_RULE_component_declaration[] = "RULE_component_declaration";
static const char __pyx_k_RULE_element_redeclaration[] = "RULE_element_redeclaration";
static const char __pyx_k_Replaceable_elementContext[] = "Replaceable_elementContext";
static const char __pyx_k_enterComponent_declaration[] = "enterComponent_declaration";
static const char __pyx_k_enterElement_redeclaration[] = "enterElement_redeclaration";
static const char __pyx_k_exitComponent_declaration1[] = "exitComponent_declaration1";
static const char __pyx_k_exitExternal_function_call[] = "exitExternal_function_call";
static const char __pyx_k_exitLanguage_specification[] = "exitLanguage_specification";
static const char __pyx_k_exitOutput_expression_list[] = "exitOutput_expression_list";
static const char __pyx_k_exitShort_class_definition[] = "exitShort_class_definition";
static const char __pyx_k_visitComponent_declaration[] = "visitComponent_declaration";
static const char __pyx_k_visitElement_redeclaration[] = "visitElement_redeclaration";
static const char __pyx_k_Element_modificationContext[] = "Element_modificationContext";
static const char __pyx_k_ModelicaParser_element_list[] = "ModelicaParser.element_list";
static const char __pyx_k_ModelicaParser_expr_sempred[] = "ModelicaParser.expr_sempred";
static const char __pyx_k_ModelicaParser_for_equation[] = "ModelicaParser.for_equation";
static const char __pyx_k_ModelicaParser_if_statement[] = "ModelicaParser.if_statement";
static const char __pyx_k_ModelicaParser_modification[] = "ModelicaParser.modification";
static const char __pyx_k_RULE_component_declaration1[] = "RULE_component_declaration1";
static const char __pyx_k_RULE_external_function_call[] = "RULE_external_function_call";
static const char __pyx_k_RULE_language_specification[] = "RULE_language_specification";
static const char __pyx_k_RULE_output_expression_list[] = "RULE_output_expression_list";
static const char __pyx_k_RULE_short_class_definition[] = "RULE_short_class_definition";
static const char __pyx_k_component_reference_element[] = "component_reference_element";
static const char __pyx_k_enterComponent_declaration1[] = "enterComponent_declaration1";
static const char __pyx_k_enterExternal_function_call[] = "enterExternal_function_call";
static const char __pyx_k_enterLanguage_specification[] = "enterLanguage_specification";
static const char __pyx_k_enterOutput_expression_list[] = "enterOutput_expression_list";
static const char __pyx_k_enterShort_class_definition[] = "enterShort_class_definition";
static const char __pyx_k_exitEquation_connect_clause[] = "exitEquation_connect_clause";
static const char __pyx_k_exitModification_assignment[] = "exitModification_assignment";
static const char __pyx_k_exitPrimary_expression_list[] = "exitPrimary_expression_list";
static const char __pyx_k_exitPrimary_unsigned_number[] = "exitPrimary_unsigned_number";
static const char __pyx_k_exitStored_definition_class[] = "exitStored_definition_class";
static const char __pyx_k_visitComponent_declaration1[] = "visitComponent_declaration1";
static const char __pyx_k_visitExternal_function_call[] = "visitExternal_function_call";
static const char __pyx_k_visitLanguage_specification[] = "visitLanguage_specification";
static const char __pyx_k_visitOutput_expression_list[] = "visitOutput_expression_list";
static const char __pyx_k_visitShort_class_definition[] = "visitShort_class_definition";
static const char __pyx_k_Component_declarationContext[] = "Component_declarationContext";
static const char __pyx_k_Element_redeclarationContext[] = "Element_redeclarationContext";
static const char __pyx_k_ModelicaParser_argument_list[] = "ModelicaParser.argument_list";
static const char __pyx_k_ModelicaParser_for_statement[] = "ModelicaParser.for_statement";
static const char __pyx_k_ModelicaParser_import_clause[] = "ModelicaParser.import_clause";
static const char __pyx_k_ModelicaParser_when_equation[] = "ModelicaParser.when_equation";
static const char __pyx_k_RULE_stored_definition_class[] = "RULE_stored_definition_class";
static const char __pyx_k_enterEquation_connect_clause[] = "enterEquation_connect_clause";
static const char __pyx_k_enterModification_assignment[] = "enterModification_assignment";
static const char __pyx_k_enterPrimary_expression_list[] = "enterPrimary_expression_list";
static const char __pyx_k_enterPrimary_unsigned_number[] = "enterPrimary_unsigned_number";
static const char __pyx_k_enterStored_definition_class[] = "enterStored_definition_class";
static const char __pyx_k_exitModification_assignment2[] = "exitModification_assignment2";
static const char __pyx_k_visitEquation_connect_clause[] = "visitEquation_connect_clause";
static const char __pyx_k_visitModification_assignment[] = "visitModification_assignment";
static const char __pyx_k_visitPrimary_expression_list[] = "visitPrimary_expression_list";
static const char __pyx_k_visitPrimary_unsigned_number[] = "visitPrimary_unsigned_number";
static const char __pyx_k_visitStored_definition_class[] = "visitStored_definition_class";
static const char __pyx_k_Component_declaration1Context[] = "Component_declaration1Context";
static const char __pyx_k_External_function_callContext[] = "External_function_callContext";
static const char __pyx_k_Language_specificationContext[] = "Language_specificationContext";
static const char __pyx_k_ModelicaParser_CommentContext[] = "ModelicaParser.CommentContext";
static const char __pyx_k_ModelicaParser_ElementContext[] = "ModelicaParser.ElementContext";
static const char __pyx_k_ModelicaParser_Expr_orContext[] = "ModelicaParser.Expr_orContext";
static const char __pyx_k_ModelicaParser_PrimaryContext[] = "ModelicaParser.PrimaryContext";
static const char __pyx_k_ModelicaParser_class_prefixes[] = "ModelicaParser.class_prefixes";
static const char __pyx_k_ModelicaParser_component_list[] = "ModelicaParser.component_list";
static const char __pyx_k_ModelicaParser_connect_clause[] = "ModelicaParser.connect_clause";
static const char __pyx_k_ModelicaParser_extends_clause[] = "ModelicaParser.extends_clause";
static const char __pyx_k_ModelicaParser_named_argument[] = "ModelicaParser.named_argument";
static const char __pyx_k_ModelicaParser_string_comment[] = "ModelicaParser.string_comment";
static const char __pyx_k_ModelicaParser_type_specifier[] = "ModelicaParser.type_specifier";
static const char __pyx_k_ModelicaParser_when_statement[] = "ModelicaParser.when_statement";
static const char __pyx_k_Output_expression_listContext[] = "Output_expression_listContext";
static const char __pyx_k_Short_class_definitionContext[] = "Short_class_definitionContext";
static const char __pyx_k_enterModification_assignment2[] = "enterModification_assignment2";
static const char __pyx_k_visitModification_assignment2[] = "visitModification_assignment2";
static const char __pyx_k_Equation_connect_clauseContext[] = "Equation_connect_clauseContext";
static const char __pyx_k_ModelicaParser_ArgumentContext[] = "ModelicaParser.ArgumentContext";
static const char __pyx_k_ModelicaParser_EquationContext[] = "ModelicaParser.EquationContext";
static const char __pyx_k_ModelicaParser_Expr_addContext[] = "ModelicaParser.Expr_addContext";
static const char __pyx_k_ModelicaParser_Expr_andContext[] = "ModelicaParser.Expr_andContext";
static const char __pyx_k_ModelicaParser_Expr_expContext[] = "ModelicaParser.Expr_expContext";
static const char __pyx_k_ModelicaParser_Expr_mulContext[] = "ModelicaParser.Expr_mulContext";
static const char __pyx_k_ModelicaParser_Expr_negContext[] = "ModelicaParser.Expr_negContext";
static const char __pyx_k_ModelicaParser_Expr_notContext[] = "ModelicaParser.Expr_notContext";
static const char __pyx_k_ModelicaParser_Expr_relContext[] = "ModelicaParser.Expr_relContext";
static const char __pyx_k_ModelicaParser_class_specifier[] = "ModelicaParser.class_specifier";
static const char __pyx_k_ModelicaParser_expression_list[] = "ModelicaParser.expression_list";
static const char __pyx_k_ModelicaParser_named_arguments[] = "ModelicaParser.named_arguments";
static const char __pyx_k_ModelicaParser_regular_element[] = "ModelicaParser.regular_element";
static const char __pyx_k_ModelicaParser_while_statement[] = "ModelicaParser.while_statement";
static const char __pyx_k_Modification_assignmentContext[] = "Modification_assignmentContext";
static const char __pyx_k_Primary_expression_listContext[] = "Primary_expression_listContext";
static const char __pyx_k_Primary_unsigned_numberContext[] = "Primary_unsigned_numberContext";
static const char __pyx_k_Stored_definition_classContext[] = "Stored_definition_classContext";
static const char __pyx_k_exitPrimary_function_arguments[] = "exitPrimary_function_arguments";
static const char __pyx_k_02468_BDFHJLNPRTVXZ_bdfhjlnprtv[] = ",.02468:<>@BDFHJLNPRTVXZ\\^`bdfhjlnprtvxz|~\302\200";
static const char __pyx_k_ModelicaParser_CommentContext_2[] = "ModelicaParser.CommentContext.__init__";
static const char __pyx_k_ModelicaParser_ElementContext_2[] = "ModelicaParser.ElementContext.__init__";
static const char __pyx_k_ModelicaParser_Enum_listContext[] = "ModelicaParser.Enum_listContext";
static const char __pyx_k_ModelicaParser_Equation_connect[] = "ModelicaParser.Equation_connect_clauseContext";
static const char __pyx_k_ModelicaParser_Expr_orContext_2[] = "ModelicaParser.Expr_orContext.__init__";
static const char __pyx_k_ModelicaParser_For_indexContext[] = "ModelicaParser.For_indexContext";
static const char __pyx_k_ModelicaParser_PrimaryContext_2[] = "ModelicaParser.PrimaryContext.__init__";
static const char __pyx_k_ModelicaParser_Primary_function[] = "ModelicaParser.Primary_function_argumentsContext";
static const char __pyx_k_ModelicaParser_Primary_unsigned[] = "ModelicaParser.Primary_unsigned_numberContext";
static const char __pyx_k_ModelicaParser_StatementContext[] = "ModelicaParser.StatementContext";
static const char __pyx_k_ModelicaParser_SubscriptContext[] = "ModelicaParser.SubscriptContext";
static const char __pyx_k_ModelicaParser_array_subscripts[] = "ModelicaParser.array_subscripts";
static const char __pyx_k_ModelicaParser_class_definition[] = "ModelicaParser.class_definition";
static const char __pyx_k_ModelicaParser_component_clause[] = "ModelicaParser.component_clause";
static const char __pyx_k_ModelicaParser_equation_options[] = "ModelicaParser.equation_options";
static const char __pyx_k_ModelicaParser_equation_section[] = "ModelicaParser.equation_section";
static const char __pyx_k_Modification_assignment2Context[] = "Modification_assignment2Context";
static const char __pyx_k_enterPrimary_function_arguments[] = "enterPrimary_function_arguments";
static const char __pyx_k_exitComponent_reference_element[] = "exitComponent_reference_element";
static const char __pyx_k_exitPrimary_component_reference[] = "exitPrimary_component_reference";
static const char __pyx_k_home_jgoppert_git_pymola_pymola[] = "/home/jgoppert/git/pymola/pymola/generated/ModelicaParser.pyx";
static const char __pyx_k_pymola_generated_ModelicaParser[] = "pymola.generated.ModelicaParser";
static const char __pyx_k_visitPrimary_function_arguments[] = "visitPrimary_function_arguments";
static const char __pyx_k_Component_reference_elementConte[] = "Component_reference_elementContext";
static const char __pyx_k_Element_modification_or_replacea[] = "Element_modification_or_replaceableContext";
static const char __pyx_k_ModelicaParser_Algorithm_section[] = "ModelicaParser.Algorithm_sectionContext";
static const char __pyx_k_ModelicaParser_AnnotationContext[] = "ModelicaParser.AnnotationContext";
static const char __pyx_k_ModelicaParser_ArgumentContext_2[] = "ModelicaParser.ArgumentContext.__init__";
static const char __pyx_k_ModelicaParser_ArgumentContext_a[] = "ModelicaParser.ArgumentContext.accept";
static const char __pyx_k_ModelicaParser_ArgumentContext_e[] = "ModelicaParser.ArgumentContext.element_modification_or_replaceable";
static const char __pyx_k_ModelicaParser_ArgumentContext_g[] = "ModelicaParser.ArgumentContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Argument_expressi[] = "ModelicaParser.Argument_expressionContext";
static const char __pyx_k_ModelicaParser_Argument_function[] = "ModelicaParser.Argument_functionContext";
static const char __pyx_k_ModelicaParser_Argument_listCont[] = "ModelicaParser.Argument_listContext";
static const char __pyx_k_ModelicaParser_Array_subscriptsC[] = "ModelicaParser.Array_subscriptsContext";
static const char __pyx_k_ModelicaParser_Base_prefixContex[] = "ModelicaParser.Base_prefixContext";
static const char __pyx_k_ModelicaParser_Class_definitionC[] = "ModelicaParser.Class_definitionContext";
static const char __pyx_k_ModelicaParser_Class_modificatio[] = "ModelicaParser.Class_modificationContext";
static const char __pyx_k_ModelicaParser_Class_prefixesCon[] = "ModelicaParser.Class_prefixesContext";
static const char __pyx_k_ModelicaParser_Class_spec_baseCo[] = "ModelicaParser.Class_spec_baseContext";
static const char __pyx_k_ModelicaParser_Class_spec_compCo[] = "ModelicaParser.Class_spec_compContext";
static const char __pyx_k_ModelicaParser_Class_spec_derCon[] = "ModelicaParser.Class_spec_derContext";
static const char __pyx_k_ModelicaParser_Class_spec_enumCo[] = "ModelicaParser.Class_spec_enumContext";
static const char __pyx_k_ModelicaParser_Class_spec_extend[] = "ModelicaParser.Class_spec_extendsContext";
static const char __pyx_k_ModelicaParser_Class_specifierCo[] = "ModelicaParser.Class_specifierContext";
static const char __pyx_k_ModelicaParser_Class_typeContext[] = "ModelicaParser.Class_typeContext";
static const char __pyx_k_ModelicaParser_CommentContext_ac[] = "ModelicaParser.CommentContext.accept";
static const char __pyx_k_ModelicaParser_CommentContext_an[] = "ModelicaParser.CommentContext.annotation";
static const char __pyx_k_ModelicaParser_CommentContext_en[] = "ModelicaParser.CommentContext.enterRule";
static const char __pyx_k_ModelicaParser_CommentContext_ex[] = "ModelicaParser.CommentContext.exitRule";
static const char __pyx_k_ModelicaParser_CommentContext_ge[] = "ModelicaParser.CommentContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_CommentContext_st[] = "ModelicaParser.CommentContext.string_comment";
static const char __pyx_k_ModelicaParser_Component_clause1[] = "ModelicaParser.Component_clause1Context";
static const char __pyx_k_ModelicaParser_Component_clauseC[] = "ModelicaParser.Component_clauseContext";
static const char __pyx_k_ModelicaParser_Component_declara[] = "ModelicaParser.Component_declarationContext";
static const char __pyx_k_ModelicaParser_Component_listCon[] = "ModelicaParser.Component_listContext";
static const char __pyx_k_ModelicaParser_Component_referen[] = "ModelicaParser.Component_reference_elementContext";
static const char __pyx_k_ModelicaParser_CompositionContex[] = "ModelicaParser.CompositionContext";
static const char __pyx_k_ModelicaParser_Condition_attribu[] = "ModelicaParser.Condition_attributeContext";
static const char __pyx_k_ModelicaParser_Connect_clauseCon[] = "ModelicaParser.Connect_clauseContext";
static const char __pyx_k_ModelicaParser_Constraining_clau[] = "ModelicaParser.Constraining_clauseContext";
static const char __pyx_k_ModelicaParser_DeclarationContex[] = "ModelicaParser.DeclarationContext";
static const char __pyx_k_ModelicaParser_ElementContext_ac[] = "ModelicaParser.ElementContext.accept";
static const char __pyx_k_ModelicaParser_ElementContext_en[] = "ModelicaParser.ElementContext.enterRule";
static const char __pyx_k_ModelicaParser_ElementContext_ex[] = "ModelicaParser.ElementContext.extends_clause";
static const char __pyx_k_ModelicaParser_ElementContext_ge[] = "ModelicaParser.ElementContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_ElementContext_im[] = "ModelicaParser.ElementContext.import_clause";
static const char __pyx_k_ModelicaParser_ElementContext_re[] = "ModelicaParser.ElementContext.regular_element";
static const char __pyx_k_ModelicaParser_Element_listConte[] = "ModelicaParser.Element_listContext";
static const char __pyx_k_ModelicaParser_Element_modificat[] = "ModelicaParser.Element_modification_or_replaceableContext";
static const char __pyx_k_ModelicaParser_Element_redeclara[] = "ModelicaParser.Element_redeclarationContext";
static const char __pyx_k_ModelicaParser_Element_replaceab[] = "ModelicaParser.Element_replaceableContext";
static const char __pyx_k_ModelicaParser_Enumeration_liter[] = "ModelicaParser.Enumeration_literalContext";
static const char __pyx_k_ModelicaParser_EquationContext_2[] = "ModelicaParser.EquationContext.__init__";
static const char __pyx_k_ModelicaParser_EquationContext_a[] = "ModelicaParser.EquationContext.accept";
static const char __pyx_k_ModelicaParser_EquationContext_c[] = "ModelicaParser.EquationContext.comment";
static const char __pyx_k_ModelicaParser_EquationContext_e[] = "ModelicaParser.EquationContext.equation_options";
static const char __pyx_k_ModelicaParser_EquationContext_g[] = "ModelicaParser.EquationContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Equation_forConte[] = "ModelicaParser.Equation_forContext";
static const char __pyx_k_ModelicaParser_Equation_function[] = "ModelicaParser.Equation_functionContext";
static const char __pyx_k_ModelicaParser_Equation_ifContex[] = "ModelicaParser.Equation_ifContext";
static const char __pyx_k_ModelicaParser_Equation_optionsC[] = "ModelicaParser.Equation_optionsContext";
static const char __pyx_k_ModelicaParser_Equation_sectionC[] = "ModelicaParser.Equation_sectionContext";
static const char __pyx_k_ModelicaParser_Equation_simpleCo[] = "ModelicaParser.Equation_simpleContext";
static const char __pyx_k_ModelicaParser_Equation_whenCont[] = "ModelicaParser.Equation_whenContext";
static const char __pyx_k_ModelicaParser_ExprContext___ini[] = "ModelicaParser.ExprContext.__init__";
static const char __pyx_k_ModelicaParser_ExprContext_copyF[] = "ModelicaParser.ExprContext.copyFrom";
static const char __pyx_k_ModelicaParser_ExprContext_getRu[] = "ModelicaParser.ExprContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Expr_addContext_2[] = "ModelicaParser.Expr_addContext.__init__";
static const char __pyx_k_ModelicaParser_Expr_addContext_a[] = "ModelicaParser.Expr_addContext.accept";
static const char __pyx_k_ModelicaParser_Expr_addContext_e[] = "ModelicaParser.Expr_addContext.expr";
static const char __pyx_k_ModelicaParser_Expr_andContext_2[] = "ModelicaParser.Expr_andContext.__init__";
static const char __pyx_k_ModelicaParser_Expr_andContext_a[] = "ModelicaParser.Expr_andContext.accept";
static const char __pyx_k_ModelicaParser_Expr_andContext_e[] = "ModelicaParser.Expr_andContext.expr";
static const char __pyx_k_ModelicaParser_Expr_expContext_2[] = "ModelicaParser.Expr_expContext.__init__";
static const char __pyx_k_ModelicaParser_Expr_expContext_a[] = "ModelicaParser.Expr_expContext.accept";
static const char __pyx_k_ModelicaParser_Expr_expContext_e[] = "ModelicaParser.Expr_expContext.enterRule";
static const char __pyx_k_ModelicaParser_Expr_expContext_p[] = "ModelicaParser.Expr_expContext.primary";
static const char __pyx_k_ModelicaParser_Expr_mulContext_2[] = "ModelicaParser.Expr_mulContext.__init__";
static const char __pyx_k_ModelicaParser_Expr_mulContext_a[] = "ModelicaParser.Expr_mulContext.accept";
static const char __pyx_k_ModelicaParser_Expr_mulContext_e[] = "ModelicaParser.Expr_mulContext.expr";
static const char __pyx_k_ModelicaParser_Expr_negContext_2[] = "ModelicaParser.Expr_negContext.__init__";
static const char __pyx_k_ModelicaParser_Expr_negContext_a[] = "ModelicaParser.Expr_negContext.accept";
static const char __pyx_k_ModelicaParser_Expr_negContext_e[] = "ModelicaParser.Expr_negContext.expr";
static const char __pyx_k_ModelicaParser_Expr_notContext_2[] = "ModelicaParser.Expr_notContext.__init__";
static const char __pyx_k_ModelicaParser_Expr_notContext_a[] = "ModelicaParser.Expr_notContext.accept";
static const char __pyx_k_ModelicaParser_Expr_notContext_e[] = "ModelicaParser.Expr_notContext.expr";
static const char __pyx_k_ModelicaParser_Expr_orContext_ac[] = "ModelicaParser.Expr_orContext.accept";
static const char __pyx_k_ModelicaParser_Expr_orContext_en[] = "ModelicaParser.Expr_orContext.enterRule";
static const char __pyx_k_ModelicaParser_Expr_orContext_ex[] = "ModelicaParser.Expr_orContext.expr";
static const char __pyx_k_ModelicaParser_Expr_primaryConte[] = "ModelicaParser.Expr_primaryContext";
static const char __pyx_k_ModelicaParser_Expr_relContext_2[] = "ModelicaParser.Expr_relContext.__init__";
static const char __pyx_k_ModelicaParser_Expr_relContext_a[] = "ModelicaParser.Expr_relContext.accept";
static const char __pyx_k_ModelicaParser_Expr_relContext_e[] = "ModelicaParser.Expr_relContext.expr";
static const char __pyx_k_ModelicaParser_ExpressionContext[] = "ModelicaParser.ExpressionContext";
static const char __pyx_k_ModelicaParser_Expression_ifCont[] = "ModelicaParser.Expression_ifContext";
static const char __pyx_k_ModelicaParser_Expression_listCo[] = "ModelicaParser.Expression_listContext";
static const char __pyx_k_ModelicaParser_Expression_simple[] = "ModelicaParser.Expression_simpleContext";
static const char __pyx_k_ModelicaParser_Extends_clauseCon[] = "ModelicaParser.Extends_clauseContext";
static const char __pyx_k_ModelicaParser_External_function[] = "ModelicaParser.External_function_callContext";
static const char __pyx_k_ModelicaParser_For_equationConte[] = "ModelicaParser.For_equationContext";
static const char __pyx_k_ModelicaParser_For_indicesContex[] = "ModelicaParser.For_indicesContext";
static const char __pyx_k_ModelicaParser_For_statementCont[] = "ModelicaParser.For_statementContext";
static const char __pyx_k_ModelicaParser_Function_argument[] = "ModelicaParser.Function_argumentsContext";
static const char __pyx_k_ModelicaParser_Function_call_arg[] = "ModelicaParser.Function_call_argsContext";
static const char __pyx_k_ModelicaParser_If_equationContex[] = "ModelicaParser.If_equationContext";
static const char __pyx_k_ModelicaParser_If_statementConte[] = "ModelicaParser.If_statementContext";
static const char __pyx_k_ModelicaParser_Import_clauseCont[] = "ModelicaParser.Import_clauseContext";
static const char __pyx_k_ModelicaParser_Import_listContex[] = "ModelicaParser.Import_listContext";
static const char __pyx_k_ModelicaParser_Language_specific[] = "ModelicaParser.Language_specificationContext";
static const char __pyx_k_ModelicaParser_ModificationConte[] = "ModelicaParser.ModificationContext";
static const char __pyx_k_ModelicaParser_Modification_assi[] = "ModelicaParser.Modification_assignment2Context";
static const char __pyx_k_ModelicaParser_Modification_clas[] = "ModelicaParser.Modification_classContext";
static const char __pyx_k_ModelicaParser_NameContext_IDENT[] = "ModelicaParser.NameContext.IDENT";
static const char __pyx_k_ModelicaParser_NameContext___ini[] = "ModelicaParser.NameContext.__init__";
static const char __pyx_k_ModelicaParser_NameContext_accep[] = "ModelicaParser.NameContext.accept";
static const char __pyx_k_ModelicaParser_NameContext_enter[] = "ModelicaParser.NameContext.enterRule";
static const char __pyx_k_ModelicaParser_NameContext_exitR[] = "ModelicaParser.NameContext.exitRule";
static const char __pyx_k_ModelicaParser_NameContext_getRu[] = "ModelicaParser.NameContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Named_argumentCon[] = "ModelicaParser.Named_argumentContext";
static const char __pyx_k_ModelicaParser_Named_argumentsCo[] = "ModelicaParser.Named_argumentsContext";
static const char __pyx_k_ModelicaParser_Output_expression[] = "ModelicaParser.Output_expression_listContext";
static const char __pyx_k_ModelicaParser_PrimaryContext_co[] = "ModelicaParser.PrimaryContext.copyFrom";
static const char __pyx_k_ModelicaParser_PrimaryContext_ge[] = "ModelicaParser.PrimaryContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Primary_component[] = "ModelicaParser.Primary_component_referenceContext";
static const char __pyx_k_ModelicaParser_Primary_derivativ[] = "ModelicaParser.Primary_derivativeContext";
static const char __pyx_k_ModelicaParser_Primary_endContex[] = "ModelicaParser.Primary_endContext";
static const char __pyx_k_ModelicaParser_Primary_expressio[] = "ModelicaParser.Primary_expression_listContext";
static const char __pyx_k_ModelicaParser_Primary_falseCont[] = "ModelicaParser.Primary_falseContext";
static const char __pyx_k_ModelicaParser_Primary_functionC[] = "ModelicaParser.Primary_functionContext";
static const char __pyx_k_ModelicaParser_Primary_initialCo[] = "ModelicaParser.Primary_initialContext";
static const char __pyx_k_ModelicaParser_Primary_output_ex[] = "ModelicaParser.Primary_output_expression_listContext";
static const char __pyx_k_ModelicaParser_Primary_stringCon[] = "ModelicaParser.Primary_stringContext";
static const char __pyx_k_ModelicaParser_Primary_trueConte[] = "ModelicaParser.Primary_trueContext";
static const char __pyx_k_ModelicaParser_Regular_elementCo[] = "ModelicaParser.Regular_elementContext";
static const char __pyx_k_ModelicaParser_Replaceable_eleme[] = "ModelicaParser.Replaceable_elementContext";
static const char __pyx_k_ModelicaParser_Short_class_defin[] = "ModelicaParser.Short_class_definitionContext";
static const char __pyx_k_ModelicaParser_Simple_expression[] = "ModelicaParser.Simple_expressionContext";
static const char __pyx_k_ModelicaParser_Statement_breakCo[] = "ModelicaParser.Statement_breakContext";
static const char __pyx_k_ModelicaParser_Statement_compone[] = "ModelicaParser.Statement_component_functionContext";
static const char __pyx_k_ModelicaParser_Statement_forCont[] = "ModelicaParser.Statement_forContext";
static const char __pyx_k_ModelicaParser_Statement_ifConte[] = "ModelicaParser.Statement_ifContext";
static const char __pyx_k_ModelicaParser_Statement_options[] = "ModelicaParser.Statement_optionsContext";
static const char __pyx_k_ModelicaParser_Statement_returnC[] = "ModelicaParser.Statement_returnContext";
static const char __pyx_k_ModelicaParser_Statement_whenCon[] = "ModelicaParser.Statement_whenContext";
static const char __pyx_k_ModelicaParser_Statement_whileCo[] = "ModelicaParser.Statement_whileContext";
static const char __pyx_k_ModelicaParser_Stored_definition[] = "ModelicaParser.Stored_definitionContext";
static const char __pyx_k_ModelicaParser_String_commentCon[] = "ModelicaParser.String_commentContext";
static const char __pyx_k_ModelicaParser_Type_prefixContex[] = "ModelicaParser.Type_prefixContext";
static const char __pyx_k_ModelicaParser_Type_specifierCon[] = "ModelicaParser.Type_specifierContext";
static const char __pyx_k_ModelicaParser_When_equationCont[] = "ModelicaParser.When_equationContext";
static const char __pyx_k_ModelicaParser_When_statementCon[] = "ModelicaParser.When_statementContext";
static const char __pyx_k_ModelicaParser_While_statementCo[] = "ModelicaParser.While_statementContext";
static const char __pyx_k_ModelicaParser_algorithm_section[] = "ModelicaParser.algorithm_section";
static const char __pyx_k_ModelicaParser_class_modificatio[] = "ModelicaParser.class_modification";
static const char __pyx_k_ModelicaParser_component_clause1[] = "ModelicaParser.component_clause1";
static const char __pyx_k_ModelicaParser_component_declara[] = "ModelicaParser.component_declaration";
static const char __pyx_k_ModelicaParser_component_referen[] = "ModelicaParser.component_reference_element";
static const char __pyx_k_ModelicaParser_condition_attribu[] = "ModelicaParser.condition_attribute";
static const char __pyx_k_ModelicaParser_constraining_clau[] = "ModelicaParser.constraining_clause";
static const char __pyx_k_ModelicaParser_element_modificat[] = "ModelicaParser.element_modification_or_replaceable";
static const char __pyx_k_ModelicaParser_element_redeclara[] = "ModelicaParser.element_redeclaration";
static const char __pyx_k_ModelicaParser_element_replaceab[] = "ModelicaParser.element_replaceable";
static const char __pyx_k_ModelicaParser_enumeration_liter[] = "ModelicaParser.enumeration_literal";
static const char __pyx_k_ModelicaParser_external_function[] = "ModelicaParser.external_function_call";
static const char __pyx_k_ModelicaParser_function_argument[] = "ModelicaParser.function_arguments";
static const char __pyx_k_ModelicaParser_function_call_arg[] = "ModelicaParser.function_call_args";
static const char __pyx_k_ModelicaParser_language_specific[] = "ModelicaParser.language_specification";
static const char __pyx_k_ModelicaParser_output_expression[] = "ModelicaParser.output_expression_list";
static const char __pyx_k_ModelicaParser_replaceable_eleme[] = "ModelicaParser.replaceable_element";
static const char __pyx_k_ModelicaParser_short_class_defin[] = "ModelicaParser.short_class_definition";
static const char __pyx_k_ModelicaParser_simple_expression[] = "ModelicaParser.simple_expression";
static const char __pyx_k_ModelicaParser_statement_options[] = "ModelicaParser.statement_options";
static const char __pyx_k_ModelicaParser_stored_definition[] = "ModelicaParser.stored_definition";
static const char __pyx_k_Primary_component_referenceConte[] = "Primary_component_referenceContext";
static const char __pyx_k_Primary_function_argumentsContex[] = "Primary_function_argumentsContext";
static const char __pyx_k_Primary_output_expression_listCo[] = "Primary_output_expression_listContext";
static const char __pyx_k_RULE_component_reference_element[] = "RULE_component_reference_element";
static const char __pyx_k_RULE_element_modification_or_rep[] = "RULE_element_modification_or_replaceable";
static const char __pyx_k_Statement_component_functionCont[] = "Statement_component_functionContext";
static const char __pyx_k_Statement_component_referenceCon[] = "Statement_component_referenceContext";
static const char __pyx_k_element_modification_or_replacea[] = "element_modification_or_replaceable";
static const char __pyx_k_enterComponent_reference_element[] = "enterComponent_reference_element";
static const char __pyx_k_enterElement_modification_or_rep[] = "enterElement_modification_or_replaceable";
static const char __pyx_k_enterPrimary_component_reference[] = "enterPrimary_component_reference";
static const char __pyx_k_enterPrimary_output_expression_l[] = "enterPrimary_output_expression_list";
static const char __pyx_k_enterStatement_component_functio[] = "enterStatement_component_function";
static const char __pyx_k_enterStatement_component_referen[] = "enterStatement_component_reference";
static const char __pyx_k_exitElement_modification_or_repl[] = "exitElement_modification_or_replaceable";
static const char __pyx_k_exitPrimary_output_expression_li[] = "exitPrimary_output_expression_list";
static const char __pyx_k_exitStatement_component_function[] = "exitStatement_component_function";
static const char __pyx_k_exitStatement_component_referenc[] = "exitStatement_component_reference";
static const char __pyx_k_visitComponent_reference_element[] = "visitComponent_reference_element";
static const char __pyx_k_visitElement_modification_or_rep[] = "visitElement_modification_or_replaceable";
static const char __pyx_k_visitPrimary_component_reference[] = "visitPrimary_component_reference";
static const char __pyx_k_visitPrimary_output_expression_l[] = "visitPrimary_output_expression_list";
static const char __pyx_k_visitStatement_component_functio[] = "visitStatement_component_function";
static const char __pyx_k_visitStatement_component_referen[] = "visitStatement_component_reference";
static const char __pyx_k_ModelicaParser_Enum_listContext_2[] = "ModelicaParser.Enum_listContext.__init__";
static const char __pyx_k_ModelicaParser_Enum_listContext_3[] = "ModelicaParser.Enum_listContext.enumeration_literal";
static const char __pyx_k_ModelicaParser_Enum_listContext_4[] = "ModelicaParser.Enum_listContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Enum_listContext_5[] = "ModelicaParser.Enum_listContext.enterRule";
static const char __pyx_k_ModelicaParser_Enum_listContext_6[] = "ModelicaParser.Enum_listContext.exitRule";
static const char __pyx_k_ModelicaParser_Enum_listContext_7[] = "ModelicaParser.Enum_listContext.accept";
static const char __pyx_k_ModelicaParser_Equation_connect_2[] = "ModelicaParser.Equation_connect_clauseContext.__init__";
static const char __pyx_k_ModelicaParser_Equation_connect_3[] = "ModelicaParser.Equation_connect_clauseContext.connect_clause";
static const char __pyx_k_ModelicaParser_Equation_connect_4[] = "ModelicaParser.Equation_connect_clauseContext.enterRule";
static const char __pyx_k_ModelicaParser_Equation_connect_5[] = "ModelicaParser.Equation_connect_clauseContext.exitRule";
static const char __pyx_k_ModelicaParser_Equation_connect_6[] = "ModelicaParser.Equation_connect_clauseContext.accept";
static const char __pyx_k_ModelicaParser_For_indexContext_2[] = "ModelicaParser.For_indexContext.__init__";
static const char __pyx_k_ModelicaParser_For_indexContext_3[] = "ModelicaParser.For_indexContext.IDENT";
static const char __pyx_k_ModelicaParser_For_indexContext_4[] = "ModelicaParser.For_indexContext.expression";
static const char __pyx_k_ModelicaParser_For_indexContext_5[] = "ModelicaParser.For_indexContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_For_indexContext_6[] = "ModelicaParser.For_indexContext.enterRule";
static const char __pyx_k_ModelicaParser_For_indexContext_7[] = "ModelicaParser.For_indexContext.exitRule";
static const char __pyx_k_ModelicaParser_For_indexContext_8[] = "ModelicaParser.For_indexContext.accept";
static const char __pyx_k_ModelicaParser_Primary_function_2[] = "ModelicaParser.Primary_function_argumentsContext.__init__";
static const char __pyx_k_ModelicaParser_Primary_function_3[] = "ModelicaParser.Primary_function_argumentsContext.function_arguments";
static const char __pyx_k_ModelicaParser_Primary_function_4[] = "ModelicaParser.Primary_function_argumentsContext.enterRule";
static const char __pyx_k_ModelicaParser_Primary_function_5[] = "ModelicaParser.Primary_function_argumentsContext.exitRule";
static const char __pyx_k_ModelicaParser_Primary_function_6[] = "ModelicaParser.Primary_function_argumentsContext.accept";
static const char __pyx_k_ModelicaParser_Primary_unsigned_2[] = "ModelicaParser.Primary_unsigned_numberContext.__init__";
static const char __pyx_k_ModelicaParser_Primary_unsigned_3[] = "ModelicaParser.Primary_unsigned_numberContext.UNSIGNED_NUMBER";
static const char __pyx_k_ModelicaParser_Primary_unsigned_4[] = "ModelicaParser.Primary_unsigned_numberContext.enterRule";
static const char __pyx_k_ModelicaParser_Primary_unsigned_5[] = "ModelicaParser.Primary_unsigned_numberContext.exitRule";
static const char __pyx_k_ModelicaParser_Primary_unsigned_6[] = "ModelicaParser.Primary_unsigned_numberContext.accept";
static const char __pyx_k_ModelicaParser_StatementContext_2[] = "ModelicaParser.StatementContext.__init__";
static const char __pyx_k_ModelicaParser_StatementContext_3[] = "ModelicaParser.StatementContext.statement_options";
static const char __pyx_k_ModelicaParser_StatementContext_4[] = "ModelicaParser.StatementContext.comment";
static const char __pyx_k_ModelicaParser_StatementContext_5[] = "ModelicaParser.StatementContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_StatementContext_6[] = "ModelicaParser.StatementContext.enterRule";
static const char __pyx_k_ModelicaParser_StatementContext_7[] = "ModelicaParser.StatementContext.exitRule";
static const char __pyx_k_ModelicaParser_StatementContext_8[] = "ModelicaParser.StatementContext.accept";
static const char __pyx_k_ModelicaParser_SubscriptContext_2[] = "ModelicaParser.SubscriptContext.__init__";
static const char __pyx_k_ModelicaParser_SubscriptContext_3[] = "ModelicaParser.SubscriptContext.expression";
static const char __pyx_k_ModelicaParser_SubscriptContext_4[] = "ModelicaParser.SubscriptContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_SubscriptContext_5[] = "ModelicaParser.SubscriptContext.enterRule";
static const char __pyx_k_ModelicaParser_SubscriptContext_6[] = "ModelicaParser.SubscriptContext.exitRule";
static const char __pyx_k_ModelicaParser_SubscriptContext_7[] = "ModelicaParser.SubscriptContext.accept";
static const char __pyx_k_ModelicaParser_Algorithm_section_2[] = "ModelicaParser.Algorithm_sectionContext.__init__";
static const char __pyx_k_ModelicaParser_Algorithm_section_3[] = "ModelicaParser.Algorithm_sectionContext.INITIAL";
static const char __pyx_k_ModelicaParser_Algorithm_section_4[] = "ModelicaParser.Algorithm_sectionContext.statement";
static const char __pyx_k_ModelicaParser_Algorithm_section_5[] = "ModelicaParser.Algorithm_sectionContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Algorithm_section_6[] = "ModelicaParser.Algorithm_sectionContext.enterRule";
static const char __pyx_k_ModelicaParser_Algorithm_section_7[] = "ModelicaParser.Algorithm_sectionContext.exitRule";
static const char __pyx_k_ModelicaParser_Algorithm_section_8[] = "ModelicaParser.Algorithm_sectionContext.accept";
static const char __pyx_k_ModelicaParser_AnnotationContext_2[] = "ModelicaParser.AnnotationContext.__init__";
static const char __pyx_k_ModelicaParser_AnnotationContext_3[] = "ModelicaParser.AnnotationContext.class_modification";
static const char __pyx_k_ModelicaParser_AnnotationContext_4[] = "ModelicaParser.AnnotationContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_AnnotationContext_5[] = "ModelicaParser.AnnotationContext.enterRule";
static const char __pyx_k_ModelicaParser_AnnotationContext_6[] = "ModelicaParser.AnnotationContext.exitRule";
static const char __pyx_k_ModelicaParser_AnnotationContext_7[] = "ModelicaParser.AnnotationContext.accept";
static const char __pyx_k_ModelicaParser_ArgumentContext_e_2[] = "ModelicaParser.ArgumentContext.element_redeclaration";
static const char __pyx_k_ModelicaParser_ArgumentContext_e_3[] = "ModelicaParser.ArgumentContext.enterRule";
static const char __pyx_k_ModelicaParser_ArgumentContext_e_4[] = "ModelicaParser.ArgumentContext.exitRule";
static const char __pyx_k_ModelicaParser_Argument_expressi_2[] = "ModelicaParser.Argument_expressionContext.__init__";
static const char __pyx_k_ModelicaParser_Argument_expressi_3[] = "ModelicaParser.Argument_expressionContext.expression";
static const char __pyx_k_ModelicaParser_Argument_expressi_4[] = "ModelicaParser.Argument_expressionContext.enterRule";
static const char __pyx_k_ModelicaParser_Argument_expressi_5[] = "ModelicaParser.Argument_expressionContext.exitRule";
static const char __pyx_k_ModelicaParser_Argument_expressi_6[] = "ModelicaParser.Argument_expressionContext.accept";
static const char __pyx_k_ModelicaParser_Argument_function_2[] = "ModelicaParser.Argument_functionContext.__init__";
static const char __pyx_k_ModelicaParser_Argument_function_3[] = "ModelicaParser.Argument_functionContext.name";
static const char __pyx_k_ModelicaParser_Argument_function_4[] = "ModelicaParser.Argument_functionContext.named_arguments";
static const char __pyx_k_ModelicaParser_Argument_function_5[] = "ModelicaParser.Argument_functionContext.enterRule";
static const char __pyx_k_ModelicaParser_Argument_function_6[] = "ModelicaParser.Argument_functionContext.exitRule";
static const char __pyx_k_ModelicaParser_Argument_function_7[] = "ModelicaParser.Argument_functionContext.accept";
static const char __pyx_k_ModelicaParser_Argument_listCont_2[] = "ModelicaParser.Argument_listContext.__init__";
static const char __pyx_k_ModelicaParser_Argument_listCont_3[] = "ModelicaParser.Argument_listContext.argument";
static const char __pyx_k_ModelicaParser_Argument_listCont_4[] = "ModelicaParser.Argument_listContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Argument_listCont_5[] = "ModelicaParser.Argument_listContext.enterRule";
static const char __pyx_k_ModelicaParser_Argument_listCont_6[] = "ModelicaParser.Argument_listContext.exitRule";
static const char __pyx_k_ModelicaParser_Argument_listCont_7[] = "ModelicaParser.Argument_listContext.accept";
static const char __pyx_k_ModelicaParser_Array_subscriptsC_2[] = "ModelicaParser.Array_subscriptsContext.__init__";
static const char __pyx_k_ModelicaParser_Array_subscriptsC_3[] = "ModelicaParser.Array_subscriptsContext.subscript";
static const char __pyx_k_ModelicaParser_Array_subscriptsC_4[] = "ModelicaParser.Array_subscriptsContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Array_subscriptsC_5[] = "ModelicaParser.Array_subscriptsContext.enterRule";
static const char __pyx_k_ModelicaParser_Array_subscriptsC_6[] = "ModelicaParser.Array_subscriptsContext.exitRule";
static const char __pyx_k_ModelicaParser_Array_subscriptsC_7[] = "ModelicaParser.Array_subscriptsContext.accept";
static const char __pyx_k_ModelicaParser_Base_prefixContex_2[] = "ModelicaParser.Base_prefixContext.__init__";
static const char __pyx_k_ModelicaParser_Base_prefixContex_3[] = "ModelicaParser.Base_prefixContext.type_prefix";
static const char __pyx_k_ModelicaParser_Base_prefixContex_4[] = "ModelicaParser.Base_prefixContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Base_prefixContex_5[] = "ModelicaParser.Base_prefixContext.enterRule";
static const char __pyx_k_ModelicaParser_Base_prefixContex_6[] = "ModelicaParser.Base_prefixContext.exitRule";
static const char __pyx_k_ModelicaParser_Base_prefixContex_7[] = "ModelicaParser.Base_prefixContext.accept";
static const char __pyx_k_ModelicaParser_Class_definitionC_2[] = "ModelicaParser.Class_definitionContext.__init__";
static const char __pyx_k_ModelicaParser_Class_definitionC_3[] = "ModelicaParser.Class_definitionContext.class_prefixes";
static const char __pyx_k_ModelicaParser_Class_definitionC_4[] = "ModelicaParser.Class_definitionContext.class_specifier";
static const char __pyx_k_ModelicaParser_Class_definitionC_5[] = "ModelicaParser.Class_definitionContext.ENCAPSULATED";
static const char __pyx_k_ModelicaParser_Class_definitionC_6[] = "ModelicaParser.Class_definitionContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Class_definitionC_7[] = "ModelicaParser.Class_definitionContext.enterRule";
static const char __pyx_k_ModelicaParser_Class_definitionC_8[] = "ModelicaParser.Class_definitionContext.exitRule";
static const char __pyx_k_ModelicaParser_Class_definitionC_9[] = "ModelicaParser.Class_definitionContext.accept";
static const char __pyx_k_ModelicaParser_Class_modificatio_2[] = "ModelicaParser.Class_modificationContext.__init__";
static const char __pyx_k_ModelicaParser_Class_modificatio_3[] = "ModelicaParser.Class_modificationContext.argument_list";
static const char __pyx_k_ModelicaParser_Class_modificatio_4[] = "ModelicaParser.Class_modificationContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Class_modificatio_5[] = "ModelicaParser.Class_modificationContext.enterRule";
static const char __pyx_k_ModelicaParser_Class_modificatio_6[] = "ModelicaParser.Class_modificationContext.exitRule";
static const char __pyx_k_ModelicaParser_Class_modificatio_7[] = "ModelicaParser.Class_modificationContext.accept";
static const char __pyx_k_ModelicaParser_Class_prefixesCon_2[] = "ModelicaParser.Class_prefixesContext.__init__";
static const char __pyx_k_ModelicaParser_Class_prefixesCon_3[] = "ModelicaParser.Class_prefixesContext.class_type";
static const char __pyx_k_ModelicaParser_Class_prefixesCon_4[] = "ModelicaParser.Class_prefixesContext.PARTIAL";
static const char __pyx_k_ModelicaParser_Class_prefixesCon_5[] = "ModelicaParser.Class_prefixesContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Class_prefixesCon_6[] = "ModelicaParser.Class_prefixesContext.enterRule";
static const char __pyx_k_ModelicaParser_Class_prefixesCon_7[] = "ModelicaParser.Class_prefixesContext.exitRule";
static const char __pyx_k_ModelicaParser_Class_prefixesCon_8[] = "ModelicaParser.Class_prefixesContext.accept";
static const char __pyx_k_ModelicaParser_Class_spec_baseCo_2[] = "ModelicaParser.Class_spec_baseContext.__init__";
static const char __pyx_k_ModelicaParser_Class_spec_baseCo_3[] = "ModelicaParser.Class_spec_baseContext.IDENT";
static const char __pyx_k_ModelicaParser_Class_spec_baseCo_4[] = "ModelicaParser.Class_spec_baseContext.base_prefix";
static const char __pyx_k_ModelicaParser_Class_spec_baseCo_5[] = "ModelicaParser.Class_spec_baseContext.name";
static const char __pyx_k_ModelicaParser_Class_spec_baseCo_6[] = "ModelicaParser.Class_spec_baseContext.comment";
static const char __pyx_k_ModelicaParser_Class_spec_baseCo_7[] = "ModelicaParser.Class_spec_baseContext.array_subscripts";
static const char __pyx_k_ModelicaParser_Class_spec_baseCo_8[] = "ModelicaParser.Class_spec_baseContext.class_modification";
static const char __pyx_k_ModelicaParser_Class_spec_baseCo_9[] = "ModelicaParser.Class_spec_baseContext.enterRule";
static const char __pyx_k_ModelicaParser_Class_spec_compCo_2[] = "ModelicaParser.Class_spec_compContext.__init__";
static const char __pyx_k_ModelicaParser_Class_spec_compCo_3[] = "ModelicaParser.Class_spec_compContext.IDENT";
static const char __pyx_k_ModelicaParser_Class_spec_compCo_4[] = "ModelicaParser.Class_spec_compContext.string_comment";
static const char __pyx_k_ModelicaParser_Class_spec_compCo_5[] = "ModelicaParser.Class_spec_compContext.composition";
static const char __pyx_k_ModelicaParser_Class_spec_compCo_6[] = "ModelicaParser.Class_spec_compContext.enterRule";
static const char __pyx_k_ModelicaParser_Class_spec_compCo_7[] = "ModelicaParser.Class_spec_compContext.exitRule";
static const char __pyx_k_ModelicaParser_Class_spec_compCo_8[] = "ModelicaParser.Class_spec_compContext.accept";
static const char __pyx_k_ModelicaParser_Class_spec_derCon_2[] = "ModelicaParser.Class_spec_derContext.__init__";
static const char __pyx_k_ModelicaParser_Class_spec_derCon_3[] = "ModelicaParser.Class_spec_derContext.IDENT";
static const char __pyx_k_ModelicaParser_Class_spec_derCon_4[] = "ModelicaParser.Class_spec_derContext.name";
static const char __pyx_k_ModelicaParser_Class_spec_derCon_5[] = "ModelicaParser.Class_spec_derContext.comment";
static const char __pyx_k_ModelicaParser_Class_spec_derCon_6[] = "ModelicaParser.Class_spec_derContext.enterRule";
static const char __pyx_k_ModelicaParser_Class_spec_derCon_7[] = "ModelicaParser.Class_spec_derContext.exitRule";
static const char __pyx_k_ModelicaParser_Class_spec_derCon_8[] = "ModelicaParser.Class_spec_derContext.accept";
static const char __pyx_k_ModelicaParser_Class_spec_enumCo_2[] = "ModelicaParser.Class_spec_enumContext.__init__";
static const char __pyx_k_ModelicaParser_Class_spec_enumCo_3[] = "ModelicaParser.Class_spec_enumContext.IDENT";
static const char __pyx_k_ModelicaParser_Class_spec_enumCo_4[] = "ModelicaParser.Class_spec_enumContext.comment";
static const char __pyx_k_ModelicaParser_Class_spec_enumCo_5[] = "ModelicaParser.Class_spec_enumContext.enum_list";
static const char __pyx_k_ModelicaParser_Class_spec_enumCo_6[] = "ModelicaParser.Class_spec_enumContext.enterRule";
static const char __pyx_k_ModelicaParser_Class_spec_enumCo_7[] = "ModelicaParser.Class_spec_enumContext.exitRule";
static const char __pyx_k_ModelicaParser_Class_spec_enumCo_8[] = "ModelicaParser.Class_spec_enumContext.accept";
static const char __pyx_k_ModelicaParser_Class_spec_extend_2[] = "ModelicaParser.Class_spec_extendsContext.__init__";
static const char __pyx_k_ModelicaParser_Class_spec_extend_3[] = "ModelicaParser.Class_spec_extendsContext.IDENT";
static const char __pyx_k_ModelicaParser_Class_spec_extend_4[] = "ModelicaParser.Class_spec_extendsContext.string_comment";
static const char __pyx_k_ModelicaParser_Class_spec_extend_5[] = "ModelicaParser.Class_spec_extendsContext.composition";
static const char __pyx_k_ModelicaParser_Class_spec_extend_6[] = "ModelicaParser.Class_spec_extendsContext.class_modification";
static const char __pyx_k_ModelicaParser_Class_spec_extend_7[] = "ModelicaParser.Class_spec_extendsContext.enterRule";
static const char __pyx_k_ModelicaParser_Class_spec_extend_8[] = "ModelicaParser.Class_spec_extendsContext.exitRule";
static const char __pyx_k_ModelicaParser_Class_spec_extend_9[] = "ModelicaParser.Class_spec_extendsContext.accept";
static const char __pyx_k_ModelicaParser_Class_specifierCo_2[] = "ModelicaParser.Class_specifierContext.__init__";
static const char __pyx_k_ModelicaParser_Class_specifierCo_3[] = "ModelicaParser.Class_specifierContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Class_specifierCo_4[] = "ModelicaParser.Class_specifierContext.copyFrom";
static const char __pyx_k_ModelicaParser_Class_typeContext_2[] = "ModelicaParser.Class_typeContext.__init__";
static const char __pyx_k_ModelicaParser_Class_typeContext_3[] = "ModelicaParser.Class_typeContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Class_typeContext_4[] = "ModelicaParser.Class_typeContext.enterRule";
static const char __pyx_k_ModelicaParser_Class_typeContext_5[] = "ModelicaParser.Class_typeContext.exitRule";
static const char __pyx_k_ModelicaParser_Class_typeContext_6[] = "ModelicaParser.Class_typeContext.accept";
static const char __pyx_k_ModelicaParser_Component_clause1_2[] = "ModelicaParser.Component_clause1Context.__init__";
static const char __pyx_k_ModelicaParser_Component_clause1_3[] = "ModelicaParser.Component_clause1Context.type_prefix";
static const char __pyx_k_ModelicaParser_Component_clause1_4[] = "ModelicaParser.Component_clause1Context.type_specifier";
static const char __pyx_k_ModelicaParser_Component_clause1_5[] = "ModelicaParser.Component_clause1Context.component_declaration1";
static const char __pyx_k_ModelicaParser_Component_clause1_6[] = "ModelicaParser.Component_clause1Context.getRuleIndex";
static const char __pyx_k_ModelicaParser_Component_clause1_7[] = "ModelicaParser.Component_clause1Context.enterRule";
static const char __pyx_k_ModelicaParser_Component_clause1_8[] = "ModelicaParser.Component_clause1Context.exitRule";
static const char __pyx_k_ModelicaParser_Component_clause1_9[] = "ModelicaParser.Component_clause1Context.accept";
static const char __pyx_k_ModelicaParser_Component_clauseC_2[] = "ModelicaParser.Component_clauseContext.__init__";
static const char __pyx_k_ModelicaParser_Component_clauseC_3[] = "ModelicaParser.Component_clauseContext.type_prefix";
static const char __pyx_k_ModelicaParser_Component_clauseC_4[] = "ModelicaParser.Component_clauseContext.type_specifier";
static const char __pyx_k_ModelicaParser_Component_clauseC_5[] = "ModelicaParser.Component_clauseContext.component_list";
static const char __pyx_k_ModelicaParser_Component_clauseC_6[] = "ModelicaParser.Component_clauseContext.array_subscripts";
static const char __pyx_k_ModelicaParser_Component_clauseC_7[] = "ModelicaParser.Component_clauseContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Component_clauseC_8[] = "ModelicaParser.Component_clauseContext.enterRule";
static const char __pyx_k_ModelicaParser_Component_clauseC_9[] = "ModelicaParser.Component_clauseContext.exitRule";
static const char __pyx_k_ModelicaParser_Component_declara_2[] = "ModelicaParser.Component_declarationContext.__init__";
static const char __pyx_k_ModelicaParser_Component_declara_3[] = "ModelicaParser.Component_declarationContext.declaration";
static const char __pyx_k_ModelicaParser_Component_declara_4[] = "ModelicaParser.Component_declarationContext.comment";
static const char __pyx_k_ModelicaParser_Component_declara_5[] = "ModelicaParser.Component_declarationContext.condition_attribute";
static const char __pyx_k_ModelicaParser_Component_declara_6[] = "ModelicaParser.Component_declarationContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Component_declara_7[] = "ModelicaParser.Component_declarationContext.enterRule";
static const char __pyx_k_ModelicaParser_Component_declara_8[] = "ModelicaParser.Component_declarationContext.exitRule";
static const char __pyx_k_ModelicaParser_Component_declara_9[] = "ModelicaParser.Component_declarationContext.accept";
static const char __pyx_k_ModelicaParser_Component_listCon_2[] = "ModelicaParser.Component_listContext.__init__";
static const char __pyx_k_ModelicaParser_Component_listCon_3[] = "ModelicaParser.Component_listContext.component_declaration";
static const char __pyx_k_ModelicaParser_Component_listCon_4[] = "ModelicaParser.Component_listContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Component_listCon_5[] = "ModelicaParser.Component_listContext.enterRule";
static const char __pyx_k_ModelicaParser_Component_listCon_6[] = "ModelicaParser.Component_listContext.exitRule";
static const char __pyx_k_ModelicaParser_Component_listCon_7[] = "ModelicaParser.Component_listContext.accept";
static const char __pyx_k_ModelicaParser_Component_referen_2[] = "ModelicaParser.Component_reference_elementContext.__init__";
static const char __pyx_k_ModelicaParser_Component_referen_3[] = "ModelicaParser.Component_reference_elementContext.IDENT";
static const char __pyx_k_ModelicaParser_Component_referen_4[] = "ModelicaParser.Component_reference_elementContext.array_subscripts";
static const char __pyx_k_ModelicaParser_Component_referen_5[] = "ModelicaParser.Component_reference_elementContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Component_referen_6[] = "ModelicaParser.Component_reference_elementContext.enterRule";
static const char __pyx_k_ModelicaParser_Component_referen_7[] = "ModelicaParser.Component_reference_elementContext.exitRule";
static const char __pyx_k_ModelicaParser_Component_referen_8[] = "ModelicaParser.Component_reference_elementContext.accept";
static const char __pyx_k_ModelicaParser_Component_referen_9[] = "ModelicaParser.Component_referenceContext";
static const char __pyx_k_ModelicaParser_CompositionContex_2[] = "ModelicaParser.CompositionContext.__init__";
static const char __pyx_k_ModelicaParser_CompositionContex_3[] = "ModelicaParser.CompositionContext.element_list";
static const char __pyx_k_ModelicaParser_CompositionContex_4[] = "ModelicaParser.CompositionContext.equation_section";
static const char __pyx_k_ModelicaParser_CompositionContex_5[] = "ModelicaParser.CompositionContext.algorithm_section";
static const char __pyx_k_ModelicaParser_CompositionContex_6[] = "ModelicaParser.CompositionContext.annotation";
static const char __pyx_k_ModelicaParser_CompositionContex_7[] = "ModelicaParser.CompositionContext.language_specification";
static const char __pyx_k_ModelicaParser_CompositionContex_8[] = "ModelicaParser.CompositionContext.external_function_call";
static const char __pyx_k_ModelicaParser_CompositionContex_9[] = "ModelicaParser.CompositionContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Condition_attribu_2[] = "ModelicaParser.Condition_attributeContext.__init__";
static const char __pyx_k_ModelicaParser_Condition_attribu_3[] = "ModelicaParser.Condition_attributeContext.expression";
static const char __pyx_k_ModelicaParser_Condition_attribu_4[] = "ModelicaParser.Condition_attributeContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Condition_attribu_5[] = "ModelicaParser.Condition_attributeContext.enterRule";
static const char __pyx_k_ModelicaParser_Condition_attribu_6[] = "ModelicaParser.Condition_attributeContext.exitRule";
static const char __pyx_k_ModelicaParser_Condition_attribu_7[] = "ModelicaParser.Condition_attributeContext.accept";
static const char __pyx_k_ModelicaParser_Connect_clauseCon_2[] = "ModelicaParser.Connect_clauseContext.__init__";
static const char __pyx_k_ModelicaParser_Connect_clauseCon_3[] = "ModelicaParser.Connect_clauseContext.component_reference";
static const char __pyx_k_ModelicaParser_Connect_clauseCon_4[] = "ModelicaParser.Connect_clauseContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Connect_clauseCon_5[] = "ModelicaParser.Connect_clauseContext.enterRule";
static const char __pyx_k_ModelicaParser_Connect_clauseCon_6[] = "ModelicaParser.Connect_clauseContext.exitRule";
static const char __pyx_k_ModelicaParser_Connect_clauseCon_7[] = "ModelicaParser.Connect_clauseContext.accept";
static const char __pyx_k_ModelicaParser_Constraining_clau_2[] = "ModelicaParser.Constraining_clauseContext.__init__";
static const char __pyx_k_ModelicaParser_Constraining_clau_3[] = "ModelicaParser.Constraining_clauseContext.name";
static const char __pyx_k_ModelicaParser_Constraining_clau_4[] = "ModelicaParser.Constraining_clauseContext.class_modification";
static const char __pyx_k_ModelicaParser_Constraining_clau_5[] = "ModelicaParser.Constraining_clauseContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Constraining_clau_6[] = "ModelicaParser.Constraining_clauseContext.enterRule";
static const char __pyx_k_ModelicaParser_Constraining_clau_7[] = "ModelicaParser.Constraining_clauseContext.exitRule";
static const char __pyx_k_ModelicaParser_Constraining_clau_8[] = "ModelicaParser.Constraining_clauseContext.accept";
static const char __pyx_k_ModelicaParser_DeclarationContex_2[] = "ModelicaParser.DeclarationContext.__init__";
static const char __pyx_k_ModelicaParser_DeclarationContex_3[] = "ModelicaParser.DeclarationContext.IDENT";
static const char __pyx_k_ModelicaParser_DeclarationContex_4[] = "ModelicaParser.DeclarationContext.array_subscripts";
static const char __pyx_k_ModelicaParser_DeclarationContex_5[] = "ModelicaParser.DeclarationContext.modification";
static const char __pyx_k_ModelicaParser_DeclarationContex_6[] = "ModelicaParser.DeclarationContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_DeclarationContex_7[] = "ModelicaParser.DeclarationContext.enterRule";
static const char __pyx_k_ModelicaParser_DeclarationContex_8[] = "ModelicaParser.DeclarationContext.exitRule";
static const char __pyx_k_ModelicaParser_DeclarationContex_9[] = "ModelicaParser.DeclarationContext.accept";
static const char __pyx_k_ModelicaParser_ElementContext_ex_2[] = "ModelicaParser.ElementContext.exitRule";
static const char __pyx_k_ModelicaParser_ElementContext_re_2[] = "ModelicaParser.ElementContext.replaceable_element";
static const char __pyx_k_ModelicaParser_Element_listConte_2[] = "ModelicaParser.Element_listContext.__init__";
static const char __pyx_k_ModelicaParser_Element_listConte_3[] = "ModelicaParser.Element_listContext.element";
static const char __pyx_k_ModelicaParser_Element_listConte_4[] = "ModelicaParser.Element_listContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Element_listConte_5[] = "ModelicaParser.Element_listContext.enterRule";
static const char __pyx_k_ModelicaParser_Element_listConte_6[] = "ModelicaParser.Element_listContext.exitRule";
static const char __pyx_k_ModelicaParser_Element_listConte_7[] = "ModelicaParser.Element_listContext.accept";
static const char __pyx_k_ModelicaParser_Element_modificat_2[] = "ModelicaParser.Element_modification_or_replaceableContext.__init__";
static const char __pyx_k_ModelicaParser_Element_modificat_3[] = "ModelicaParser.Element_modification_or_replaceableContext.element_modification";
static const char __pyx_k_ModelicaParser_Element_modificat_4[] = "ModelicaParser.Element_modification_or_replaceableContext.element_replaceable";
static const char __pyx_k_ModelicaParser_Element_modificat_5[] = "ModelicaParser.Element_modification_or_replaceableContext.EACH";
static const char __pyx_k_ModelicaParser_Element_modificat_6[] = "ModelicaParser.Element_modification_or_replaceableContext.FINAL";
static const char __pyx_k_ModelicaParser_Element_modificat_7[] = "ModelicaParser.Element_modification_or_replaceableContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Element_modificat_8[] = "ModelicaParser.Element_modification_or_replaceableContext.enterRule";
static const char __pyx_k_ModelicaParser_Element_modificat_9[] = "ModelicaParser.Element_modification_or_replaceableContext.exitRule";
static const char __pyx_k_ModelicaParser_Element_redeclara_2[] = "ModelicaParser.Element_redeclarationContext.__init__";
static const char __pyx_k_ModelicaParser_Element_redeclara_3[] = "ModelicaParser.Element_redeclarationContext.REDECLARE";
static const char __pyx_k_ModelicaParser_Element_redeclara_4[] = "ModelicaParser.Element_redeclarationContext.element_replaceable";
static const char __pyx_k_ModelicaParser_Element_redeclara_5[] = "ModelicaParser.Element_redeclarationContext.EACH";
static const char __pyx_k_ModelicaParser_Element_redeclara_6[] = "ModelicaParser.Element_redeclarationContext.FINAL";
static const char __pyx_k_ModelicaParser_Element_redeclara_7[] = "ModelicaParser.Element_redeclarationContext.short_class_definition";
static const char __pyx_k_ModelicaParser_Element_redeclara_8[] = "ModelicaParser.Element_redeclarationContext.component_clause1";
static const char __pyx_k_ModelicaParser_Element_redeclara_9[] = "ModelicaParser.Element_redeclarationContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Element_replaceab_2[] = "ModelicaParser.Element_replaceableContext.__init__";
static const char __pyx_k_ModelicaParser_Element_replaceab_3[] = "ModelicaParser.Element_replaceableContext.short_class_definition";
static const char __pyx_k_ModelicaParser_Element_replaceab_4[] = "ModelicaParser.Element_replaceableContext.component_clause1";
static const char __pyx_k_ModelicaParser_Element_replaceab_5[] = "ModelicaParser.Element_replaceableContext.constraining_clause";
static const char __pyx_k_ModelicaParser_Element_replaceab_6[] = "ModelicaParser.Element_replaceableContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Element_replaceab_7[] = "ModelicaParser.Element_replaceableContext.enterRule";
static const char __pyx_k_ModelicaParser_Element_replaceab_8[] = "ModelicaParser.Element_replaceableContext.exitRule";
static const char __pyx_k_ModelicaParser_Element_replaceab_9[] = "ModelicaParser.Element_replaceableContext.accept";
static const char __pyx_k_ModelicaParser_Enumeration_liter_2[] = "ModelicaParser.Enumeration_literalContext.__init__";
static const char __pyx_k_ModelicaParser_Enumeration_liter_3[] = "ModelicaParser.Enumeration_literalContext.IDENT";
static const char __pyx_k_ModelicaParser_Enumeration_liter_4[] = "ModelicaParser.Enumeration_literalContext.comment";
static const char __pyx_k_ModelicaParser_Enumeration_liter_5[] = "ModelicaParser.Enumeration_literalContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Enumeration_liter_6[] = "ModelicaParser.Enumeration_literalContext.enterRule";
static const char __pyx_k_ModelicaParser_Enumeration_liter_7[] = "ModelicaParser.Enumeration_literalContext.exitRule";
static const char __pyx_k_ModelicaParser_Enumeration_liter_8[] = "ModelicaParser.Enumeration_literalContext.accept";
static const char __pyx_k_ModelicaParser_EquationContext_e_2[] = "ModelicaParser.EquationContext.enterRule";
static const char __pyx_k_ModelicaParser_EquationContext_e_3[] = "ModelicaParser.EquationContext.exitRule";
static const char __pyx_k_ModelicaParser_Equation_forConte_2[] = "ModelicaParser.Equation_forContext.__init__";
static const char __pyx_k_ModelicaParser_Equation_forConte_3[] = "ModelicaParser.Equation_forContext.for_equation";
static const char __pyx_k_ModelicaParser_Equation_forConte_4[] = "ModelicaParser.Equation_forContext.enterRule";
static const char __pyx_k_ModelicaParser_Equation_forConte_5[] = "ModelicaParser.Equation_forContext.exitRule";
static const char __pyx_k_ModelicaParser_Equation_forConte_6[] = "ModelicaParser.Equation_forContext.accept";
static const char __pyx_k_ModelicaParser_Equation_function_2[] = "ModelicaParser.Equation_functionContext.__init__";
static const char __pyx_k_ModelicaParser_Equation_function_3[] = "ModelicaParser.Equation_functionContext.name";
static const char __pyx_k_ModelicaParser_Equation_function_4[] = "ModelicaParser.Equation_functionContext.function_call_args";
static const char __pyx_k_ModelicaParser_Equation_function_5[] = "ModelicaParser.Equation_functionContext.enterRule";
static const char __pyx_k_ModelicaParser_Equation_function_6[] = "ModelicaParser.Equation_functionContext.exitRule";
static const char __pyx_k_ModelicaParser_Equation_function_7[] = "ModelicaParser.Equation_functionContext.accept";
static const char __pyx_k_ModelicaParser_Equation_ifContex_2[] = "ModelicaParser.Equation_ifContext.__init__";
static const char __pyx_k_ModelicaParser_Equation_ifContex_3[] = "ModelicaParser.Equation_ifContext.if_equation";
static const char __pyx_k_ModelicaParser_Equation_ifContex_4[] = "ModelicaParser.Equation_ifContext.enterRule";
static const char __pyx_k_ModelicaParser_Equation_ifContex_5[] = "ModelicaParser.Equation_ifContext.exitRule";
static const char __pyx_k_ModelicaParser_Equation_ifContex_6[] = "ModelicaParser.Equation_ifContext.accept";
static const char __pyx_k_ModelicaParser_Equation_optionsC_2[] = "ModelicaParser.Equation_optionsContext.__init__";
static const char __pyx_k_ModelicaParser_Equation_optionsC_3[] = "ModelicaParser.Equation_optionsContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Equation_optionsC_4[] = "ModelicaParser.Equation_optionsContext.copyFrom";
static const char __pyx_k_ModelicaParser_Equation_sectionC_2[] = "ModelicaParser.Equation_sectionContext.__init__";
static const char __pyx_k_ModelicaParser_Equation_sectionC_3[] = "ModelicaParser.Equation_sectionContext.INITIAL";
static const char __pyx_k_ModelicaParser_Equation_sectionC_4[] = "ModelicaParser.Equation_sectionContext.equation";
static const char __pyx_k_ModelicaParser_Equation_sectionC_5[] = "ModelicaParser.Equation_sectionContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Equation_sectionC_6[] = "ModelicaParser.Equation_sectionContext.enterRule";
static const char __pyx_k_ModelicaParser_Equation_sectionC_7[] = "ModelicaParser.Equation_sectionContext.exitRule";
static const char __pyx_k_ModelicaParser_Equation_sectionC_8[] = "ModelicaParser.Equation_sectionContext.accept";
static const char __pyx_k_ModelicaParser_Equation_simpleCo_2[] = "ModelicaParser.Equation_simpleContext.__init__";
static const char __pyx_k_ModelicaParser_Equation_simpleCo_3[] = "ModelicaParser.Equation_simpleContext.simple_expression";
static const char __pyx_k_ModelicaParser_Equation_simpleCo_4[] = "ModelicaParser.Equation_simpleContext.expression";
static const char __pyx_k_ModelicaParser_Equation_simpleCo_5[] = "ModelicaParser.Equation_simpleContext.enterRule";
static const char __pyx_k_ModelicaParser_Equation_simpleCo_6[] = "ModelicaParser.Equation_simpleContext.exitRule";
static const char __pyx_k_ModelicaParser_Equation_simpleCo_7[] = "ModelicaParser.Equation_simpleContext.accept";
static const char __pyx_k_ModelicaParser_Equation_whenCont_2[] = "ModelicaParser.Equation_whenContext.__init__";
static const char __pyx_k_ModelicaParser_Equation_whenCont_3[] = "ModelicaParser.Equation_whenContext.when_equation";
static const char __pyx_k_ModelicaParser_Equation_whenCont_4[] = "ModelicaParser.Equation_whenContext.enterRule";
static const char __pyx_k_ModelicaParser_Equation_whenCont_5[] = "ModelicaParser.Equation_whenContext.exitRule";
static const char __pyx_k_ModelicaParser_Equation_whenCont_6[] = "ModelicaParser.Equation_whenContext.accept";
static const char __pyx_k_ModelicaParser_Expr_addContext_e_2[] = "ModelicaParser.Expr_addContext.enterRule";
static const char __pyx_k_ModelicaParser_Expr_addContext_e_3[] = "ModelicaParser.Expr_addContext.exitRule";
static const char __pyx_k_ModelicaParser_Expr_andContext_e_2[] = "ModelicaParser.Expr_andContext.enterRule";
static const char __pyx_k_ModelicaParser_Expr_andContext_e_3[] = "ModelicaParser.Expr_andContext.exitRule";
static const char __pyx_k_ModelicaParser_Expr_expContext_e_2[] = "ModelicaParser.Expr_expContext.exitRule";
static const char __pyx_k_ModelicaParser_Expr_mulContext_e_2[] = "ModelicaParser.Expr_mulContext.enterRule";
static const char __pyx_k_ModelicaParser_Expr_mulContext_e_3[] = "ModelicaParser.Expr_mulContext.exitRule";
static const char __pyx_k_ModelicaParser_Expr_negContext_e_2[] = "ModelicaParser.Expr_negContext.enterRule";
static const char __pyx_k_ModelicaParser_Expr_negContext_e_3[] = "ModelicaParser.Expr_negContext.exitRule";
static const char __pyx_k_ModelicaParser_Expr_notContext_e_2[] = "ModelicaParser.Expr_notContext.enterRule";
static const char __pyx_k_ModelicaParser_Expr_notContext_e_3[] = "ModelicaParser.Expr_notContext.exitRule";
static const char __pyx_k_ModelicaParser_Expr_orContext_ex_2[] = "ModelicaParser.Expr_orContext.exitRule";
static const char __pyx_k_ModelicaParser_Expr_primaryConte_2[] = "ModelicaParser.Expr_primaryContext.__init__";
static const char __pyx_k_ModelicaParser_Expr_primaryConte_3[] = "ModelicaParser.Expr_primaryContext.primary";
static const char __pyx_k_ModelicaParser_Expr_primaryConte_4[] = "ModelicaParser.Expr_primaryContext.enterRule";
static const char __pyx_k_ModelicaParser_Expr_primaryConte_5[] = "ModelicaParser.Expr_primaryContext.exitRule";
static const char __pyx_k_ModelicaParser_Expr_primaryConte_6[] = "ModelicaParser.Expr_primaryContext.accept";
static const char __pyx_k_ModelicaParser_Expr_relContext_e_2[] = "ModelicaParser.Expr_relContext.enterRule";
static const char __pyx_k_ModelicaParser_Expr_relContext_e_3[] = "ModelicaParser.Expr_relContext.exitRule";
static const char __pyx_k_ModelicaParser_ExpressionContext_2[] = "ModelicaParser.ExpressionContext.__init__";
static const char __pyx_k_ModelicaParser_ExpressionContext_3[] = "ModelicaParser.ExpressionContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_ExpressionContext_4[] = "ModelicaParser.ExpressionContext.copyFrom";
static const char __pyx_k_ModelicaParser_Expression_ifCont_2[] = "ModelicaParser.Expression_ifContext.__init__";
static const char __pyx_k_ModelicaParser_Expression_ifCont_3[] = "ModelicaParser.Expression_ifContext.expression";
static const char __pyx_k_ModelicaParser_Expression_ifCont_4[] = "ModelicaParser.Expression_ifContext.enterRule";
static const char __pyx_k_ModelicaParser_Expression_ifCont_5[] = "ModelicaParser.Expression_ifContext.exitRule";
static const char __pyx_k_ModelicaParser_Expression_ifCont_6[] = "ModelicaParser.Expression_ifContext.accept";
static const char __pyx_k_ModelicaParser_Expression_listCo_2[] = "ModelicaParser.Expression_listContext.__init__";
static const char __pyx_k_ModelicaParser_Expression_listCo_3[] = "ModelicaParser.Expression_listContext.expression";
static const char __pyx_k_ModelicaParser_Expression_listCo_4[] = "ModelicaParser.Expression_listContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Expression_listCo_5[] = "ModelicaParser.Expression_listContext.enterRule";
static const char __pyx_k_ModelicaParser_Expression_listCo_6[] = "ModelicaParser.Expression_listContext.exitRule";
static const char __pyx_k_ModelicaParser_Expression_listCo_7[] = "ModelicaParser.Expression_listContext.accept";
static const char __pyx_k_ModelicaParser_Expression_simple_2[] = "ModelicaParser.Expression_simpleContext.__init__";
static const char __pyx_k_ModelicaParser_Expression_simple_3[] = "ModelicaParser.Expression_simpleContext.simple_expression";
static const char __pyx_k_ModelicaParser_Expression_simple_4[] = "ModelicaParser.Expression_simpleContext.enterRule";
static const char __pyx_k_ModelicaParser_Expression_simple_5[] = "ModelicaParser.Expression_simpleContext.exitRule";
static const char __pyx_k_ModelicaParser_Expression_simple_6[] = "ModelicaParser.Expression_simpleContext.accept";
static const char __pyx_k_ModelicaParser_Extends_clauseCon_2[] = "ModelicaParser.Extends_clauseContext.__init__";
static const char __pyx_k_ModelicaParser_Extends_clauseCon_3[] = "ModelicaParser.Extends_clauseContext.component_reference";
static const char __pyx_k_ModelicaParser_Extends_clauseCon_4[] = "ModelicaParser.Extends_clauseContext.class_modification";
static const char __pyx_k_ModelicaParser_Extends_clauseCon_5[] = "ModelicaParser.Extends_clauseContext.annotation";
static const char __pyx_k_ModelicaParser_Extends_clauseCon_6[] = "ModelicaParser.Extends_clauseContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Extends_clauseCon_7[] = "ModelicaParser.Extends_clauseContext.enterRule";
static const char __pyx_k_ModelicaParser_Extends_clauseCon_8[] = "ModelicaParser.Extends_clauseContext.exitRule";
static const char __pyx_k_ModelicaParser_Extends_clauseCon_9[] = "ModelicaParser.Extends_clauseContext.accept";
static const char __pyx_k_ModelicaParser_External_function_2[] = "ModelicaParser.External_function_callContext.__init__";
static const char __pyx_k_ModelicaParser_External_function_3[] = "ModelicaParser.External_function_callContext.IDENT";
static const char __pyx_k_ModelicaParser_External_function_4[] = "ModelicaParser.External_function_callContext.component_reference";
static const char __pyx_k_ModelicaParser_External_function_5[] = "ModelicaParser.External_function_callContext.expression_list";
static const char __pyx_k_ModelicaParser_External_function_6[] = "ModelicaParser.External_function_callContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_External_function_7[] = "ModelicaParser.External_function_callContext.enterRule";
static const char __pyx_k_ModelicaParser_External_function_8[] = "ModelicaParser.External_function_callContext.exitRule";
static const char __pyx_k_ModelicaParser_External_function_9[] = "ModelicaParser.External_function_callContext.accept";
static const char __pyx_k_ModelicaParser_For_equationConte_2[] = "ModelicaParser.For_equationContext.__init__";
static const char __pyx_k_ModelicaParser_For_equationConte_3[] = "ModelicaParser.For_equationContext.for_indices";
static const char __pyx_k_ModelicaParser_For_equationConte_4[] = "ModelicaParser.For_equationContext.equation";
static const char __pyx_k_ModelicaParser_For_equationConte_5[] = "ModelicaParser.For_equationContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_For_equationConte_6[] = "ModelicaParser.For_equationContext.enterRule";
static const char __pyx_k_ModelicaParser_For_equationConte_7[] = "ModelicaParser.For_equationContext.exitRule";
static const char __pyx_k_ModelicaParser_For_equationConte_8[] = "ModelicaParser.For_equationContext.accept";
static const char __pyx_k_ModelicaParser_For_indicesContex_2[] = "ModelicaParser.For_indicesContext.__init__";
static const char __pyx_k_ModelicaParser_For_indicesContex_3[] = "ModelicaParser.For_indicesContext.for_index";
static const char __pyx_k_ModelicaParser_For_indicesContex_4[] = "ModelicaParser.For_indicesContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_For_indicesContex_5[] = "ModelicaParser.For_indicesContext.enterRule";
static const char __pyx_k_ModelicaParser_For_indicesContex_6[] = "ModelicaParser.For_indicesContext.exitRule";
static const char __pyx_k_ModelicaParser_For_indicesContex_7[] = "ModelicaParser.For_indicesContext.accept";
static const char __pyx_k_ModelicaParser_For_statementCont_2[] = "ModelicaParser.For_statementContext.__init__";
static const char __pyx_k_ModelicaParser_For_statementCont_3[] = "ModelicaParser.For_statementContext.for_indices";
static const char __pyx_k_ModelicaParser_For_statementCont_4[] = "ModelicaParser.For_statementContext.statement";
static const char __pyx_k_ModelicaParser_For_statementCont_5[] = "ModelicaParser.For_statementContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_For_statementCont_6[] = "ModelicaParser.For_statementContext.enterRule";
static const char __pyx_k_ModelicaParser_For_statementCont_7[] = "ModelicaParser.For_statementContext.exitRule";
static const char __pyx_k_ModelicaParser_For_statementCont_8[] = "ModelicaParser.For_statementContext.accept";
static const char __pyx_k_ModelicaParser_Function_argument_2[] = "ModelicaParser.Function_argumentsContext.__init__";
static const char __pyx_k_ModelicaParser_Function_argument_3[] = "ModelicaParser.Function_argumentsContext.function_argument";
static const char __pyx_k_ModelicaParser_Function_argument_4[] = "ModelicaParser.Function_argumentsContext.for_indices";
static const char __pyx_k_ModelicaParser_Function_argument_5[] = "ModelicaParser.Function_argumentsContext.named_arguments";
static const char __pyx_k_ModelicaParser_Function_argument_6[] = "ModelicaParser.Function_argumentsContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Function_argument_7[] = "ModelicaParser.Function_argumentsContext.enterRule";
static const char __pyx_k_ModelicaParser_Function_argument_8[] = "ModelicaParser.Function_argumentsContext.exitRule";
static const char __pyx_k_ModelicaParser_Function_argument_9[] = "ModelicaParser.Function_argumentsContext.accept";
static const char __pyx_k_ModelicaParser_Function_call_arg_2[] = "ModelicaParser.Function_call_argsContext.__init__";
static const char __pyx_k_ModelicaParser_Function_call_arg_3[] = "ModelicaParser.Function_call_argsContext.function_arguments";
static const char __pyx_k_ModelicaParser_Function_call_arg_4[] = "ModelicaParser.Function_call_argsContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Function_call_arg_5[] = "ModelicaParser.Function_call_argsContext.enterRule";
static const char __pyx_k_ModelicaParser_Function_call_arg_6[] = "ModelicaParser.Function_call_argsContext.exitRule";
static const char __pyx_k_ModelicaParser_Function_call_arg_7[] = "ModelicaParser.Function_call_argsContext.accept";
static const char __pyx_k_ModelicaParser_If_equationContex_2[] = "ModelicaParser.If_equationContext.__init__";
static const char __pyx_k_ModelicaParser_If_equationContex_3[] = "ModelicaParser.If_equationContext.expression";
static const char __pyx_k_ModelicaParser_If_equationContex_4[] = "ModelicaParser.If_equationContext.equation";
static const char __pyx_k_ModelicaParser_If_equationContex_5[] = "ModelicaParser.If_equationContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_If_equationContex_6[] = "ModelicaParser.If_equationContext.enterRule";
static const char __pyx_k_ModelicaParser_If_equationContex_7[] = "ModelicaParser.If_equationContext.exitRule";
static const char __pyx_k_ModelicaParser_If_equationContex_8[] = "ModelicaParser.If_equationContext.accept";
static const char __pyx_k_ModelicaParser_If_statementConte_2[] = "ModelicaParser.If_statementContext.__init__";
static const char __pyx_k_ModelicaParser_If_statementConte_3[] = "ModelicaParser.If_statementContext.expression";
static const char __pyx_k_ModelicaParser_If_statementConte_4[] = "ModelicaParser.If_statementContext.statement";
static const char __pyx_k_ModelicaParser_If_statementConte_5[] = "ModelicaParser.If_statementContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_If_statementConte_6[] = "ModelicaParser.If_statementContext.enterRule";
static const char __pyx_k_ModelicaParser_If_statementConte_7[] = "ModelicaParser.If_statementContext.exitRule";
static const char __pyx_k_ModelicaParser_If_statementConte_8[] = "ModelicaParser.If_statementContext.accept";
static const char __pyx_k_ModelicaParser_Import_clauseCont_2[] = "ModelicaParser.Import_clauseContext.__init__";
static const char __pyx_k_ModelicaParser_Import_clauseCont_3[] = "ModelicaParser.Import_clauseContext.comment";
static const char __pyx_k_ModelicaParser_Import_clauseCont_4[] = "ModelicaParser.Import_clauseContext.IDENT";
static const char __pyx_k_ModelicaParser_Import_clauseCont_5[] = "ModelicaParser.Import_clauseContext.component_reference";
static const char __pyx_k_ModelicaParser_Import_clauseCont_6[] = "ModelicaParser.Import_clauseContext.import_list";
static const char __pyx_k_ModelicaParser_Import_clauseCont_7[] = "ModelicaParser.Import_clauseContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Import_clauseCont_8[] = "ModelicaParser.Import_clauseContext.enterRule";
static const char __pyx_k_ModelicaParser_Import_clauseCont_9[] = "ModelicaParser.Import_clauseContext.exitRule";
static const char __pyx_k_ModelicaParser_Import_listContex_2[] = "ModelicaParser.Import_listContext.__init__";
static const char __pyx_k_ModelicaParser_Import_listContex_3[] = "ModelicaParser.Import_listContext.IDENT";
static const char __pyx_k_ModelicaParser_Import_listContex_4[] = "ModelicaParser.Import_listContext.import_list";
static const char __pyx_k_ModelicaParser_Import_listContex_5[] = "ModelicaParser.Import_listContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Import_listContex_6[] = "ModelicaParser.Import_listContext.enterRule";
static const char __pyx_k_ModelicaParser_Import_listContex_7[] = "ModelicaParser.Import_listContext.exitRule";
static const char __pyx_k_ModelicaParser_Import_listContex_8[] = "ModelicaParser.Import_listContext.accept";
static const char __pyx_k_ModelicaParser_Language_specific_2[] = "ModelicaParser.Language_specificationContext.__init__";
static const char __pyx_k_ModelicaParser_Language_specific_3[] = "ModelicaParser.Language_specificationContext.STRING";
static const char __pyx_k_ModelicaParser_Language_specific_4[] = "ModelicaParser.Language_specificationContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Language_specific_5[] = "ModelicaParser.Language_specificationContext.enterRule";
static const char __pyx_k_ModelicaParser_Language_specific_6[] = "ModelicaParser.Language_specificationContext.exitRule";
static const char __pyx_k_ModelicaParser_Language_specific_7[] = "ModelicaParser.Language_specificationContext.accept";
static const char __pyx_k_ModelicaParser_ModificationConte_2[] = "ModelicaParser.ModificationContext.__init__";
static const char __pyx_k_ModelicaParser_ModificationConte_3[] = "ModelicaParser.ModificationContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_ModificationConte_4[] = "ModelicaParser.ModificationContext.copyFrom";
static const char __pyx_k_ModelicaParser_Modification_assi_2[] = "ModelicaParser.Modification_assignment2Context.__init__";
static const char __pyx_k_ModelicaParser_Modification_assi_3[] = "ModelicaParser.Modification_assignment2Context.expression";
static const char __pyx_k_ModelicaParser_Modification_assi_4[] = "ModelicaParser.Modification_assignment2Context.enterRule";
static const char __pyx_k_ModelicaParser_Modification_assi_5[] = "ModelicaParser.Modification_assignment2Context.exitRule";
static const char __pyx_k_ModelicaParser_Modification_assi_6[] = "ModelicaParser.Modification_assignment2Context.accept";
static const char __pyx_k_ModelicaParser_Modification_assi_7[] = "ModelicaParser.Modification_assignmentContext";
static const char __pyx_k_ModelicaParser_Modification_assi_8[] = "ModelicaParser.Modification_assignmentContext.__init__";
static const char __pyx_k_ModelicaParser_Modification_assi_9[] = "ModelicaParser.Modification_assignmentContext.expression";
static const char __pyx_k_ModelicaParser_Modification_clas_2[] = "ModelicaParser.Modification_classContext.__init__";
static const char __pyx_k_ModelicaParser_Modification_clas_3[] = "ModelicaParser.Modification_classContext.class_modification";
static const char __pyx_k_ModelicaParser_Modification_clas_4[] = "ModelicaParser.Modification_classContext.expression";
static const char __pyx_k_ModelicaParser_Modification_clas_5[] = "ModelicaParser.Modification_classContext.enterRule";
static const char __pyx_k_ModelicaParser_Modification_clas_6[] = "ModelicaParser.Modification_classContext.exitRule";
static const char __pyx_k_ModelicaParser_Modification_clas_7[] = "ModelicaParser.Modification_classContext.accept";
static const char __pyx_k_ModelicaParser_Named_argumentCon_2[] = "ModelicaParser.Named_argumentContext.__init__";
static const char __pyx_k_ModelicaParser_Named_argumentCon_3[] = "ModelicaParser.Named_argumentContext.IDENT";
static const char __pyx_k_ModelicaParser_Named_argumentCon_4[] = "ModelicaParser.Named_argumentContext.function_argument";
static const char __pyx_k_ModelicaParser_Named_argumentCon_5[] = "ModelicaParser.Named_argumentContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Named_argumentCon_6[] = "ModelicaParser.Named_argumentContext.enterRule";
static const char __pyx_k_ModelicaParser_Named_argumentCon_7[] = "ModelicaParser.Named_argumentContext.exitRule";
static const char __pyx_k_ModelicaParser_Named_argumentCon_8[] = "ModelicaParser.Named_argumentContext.accept";
static const char __pyx_k_ModelicaParser_Named_argumentsCo_2[] = "ModelicaParser.Named_argumentsContext.__init__";
static const char __pyx_k_ModelicaParser_Named_argumentsCo_3[] = "ModelicaParser.Named_argumentsContext.named_argument";
static const char __pyx_k_ModelicaParser_Named_argumentsCo_4[] = "ModelicaParser.Named_argumentsContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Named_argumentsCo_5[] = "ModelicaParser.Named_argumentsContext.enterRule";
static const char __pyx_k_ModelicaParser_Named_argumentsCo_6[] = "ModelicaParser.Named_argumentsContext.exitRule";
static const char __pyx_k_ModelicaParser_Named_argumentsCo_7[] = "ModelicaParser.Named_argumentsContext.accept";
static const char __pyx_k_ModelicaParser_Output_expression_2[] = "ModelicaParser.Output_expression_listContext.__init__";
static const char __pyx_k_ModelicaParser_Output_expression_3[] = "ModelicaParser.Output_expression_listContext.expression";
static const char __pyx_k_ModelicaParser_Output_expression_4[] = "ModelicaParser.Output_expression_listContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Output_expression_5[] = "ModelicaParser.Output_expression_listContext.enterRule";
static const char __pyx_k_ModelicaParser_Output_expression_6[] = "ModelicaParser.Output_expression_listContext.exitRule";
static const char __pyx_k_ModelicaParser_Output_expression_7[] = "ModelicaParser.Output_expression_listContext.accept";
static const char __pyx_k_ModelicaParser_Primary_component_2[] = "ModelicaParser.Primary_component_referenceContext.__init__";
static const char __pyx_k_ModelicaParser_Primary_component_3[] = "ModelicaParser.Primary_component_referenceContext.component_reference";
static const char __pyx_k_ModelicaParser_Primary_component_4[] = "ModelicaParser.Primary_component_referenceContext.enterRule";
static const char __pyx_k_ModelicaParser_Primary_component_5[] = "ModelicaParser.Primary_component_referenceContext.exitRule";
static const char __pyx_k_ModelicaParser_Primary_component_6[] = "ModelicaParser.Primary_component_referenceContext.accept";
static const char __pyx_k_ModelicaParser_Primary_derivativ_2[] = "ModelicaParser.Primary_derivativeContext.__init__";
static const char __pyx_k_ModelicaParser_Primary_derivativ_3[] = "ModelicaParser.Primary_derivativeContext.function_call_args";
static const char __pyx_k_ModelicaParser_Primary_derivativ_4[] = "ModelicaParser.Primary_derivativeContext.enterRule";
static const char __pyx_k_ModelicaParser_Primary_derivativ_5[] = "ModelicaParser.Primary_derivativeContext.exitRule";
static const char __pyx_k_ModelicaParser_Primary_derivativ_6[] = "ModelicaParser.Primary_derivativeContext.accept";
static const char __pyx_k_ModelicaParser_Primary_endContex_2[] = "ModelicaParser.Primary_endContext.__init__";
static const char __pyx_k_ModelicaParser_Primary_endContex_3[] = "ModelicaParser.Primary_endContext.enterRule";
static const char __pyx_k_ModelicaParser_Primary_endContex_4[] = "ModelicaParser.Primary_endContext.exitRule";
static const char __pyx_k_ModelicaParser_Primary_endContex_5[] = "ModelicaParser.Primary_endContext.accept";
static const char __pyx_k_ModelicaParser_Primary_expressio_2[] = "ModelicaParser.Primary_expression_listContext.__init__";
static const char __pyx_k_ModelicaParser_Primary_expressio_3[] = "ModelicaParser.Primary_expression_listContext.expression_list";
static const char __pyx_k_ModelicaParser_Primary_expressio_4[] = "ModelicaParser.Primary_expression_listContext.enterRule";
static const char __pyx_k_ModelicaParser_Primary_expressio_5[] = "ModelicaParser.Primary_expression_listContext.exitRule";
static const char __pyx_k_ModelicaParser_Primary_expressio_6[] = "ModelicaParser.Primary_expression_listContext.accept";
static const char __pyx_k_ModelicaParser_Primary_falseCont_2[] = "ModelicaParser.Primary_falseContext.__init__";
static const char __pyx_k_ModelicaParser_Primary_falseCont_3[] = "ModelicaParser.Primary_falseContext.enterRule";
static const char __pyx_k_ModelicaParser_Primary_falseCont_4[] = "ModelicaParser.Primary_falseContext.exitRule";
static const char __pyx_k_ModelicaParser_Primary_falseCont_5[] = "ModelicaParser.Primary_falseContext.accept";
static const char __pyx_k_ModelicaParser_Primary_functionC_2[] = "ModelicaParser.Primary_functionContext.__init__";
static const char __pyx_k_ModelicaParser_Primary_functionC_3[] = "ModelicaParser.Primary_functionContext.component_reference";
static const char __pyx_k_ModelicaParser_Primary_functionC_4[] = "ModelicaParser.Primary_functionContext.function_call_args";
static const char __pyx_k_ModelicaParser_Primary_functionC_5[] = "ModelicaParser.Primary_functionContext.enterRule";
static const char __pyx_k_ModelicaParser_Primary_functionC_6[] = "ModelicaParser.Primary_functionContext.exitRule";
static const char __pyx_k_ModelicaParser_Primary_functionC_7[] = "ModelicaParser.Primary_functionContext.accept";
static const char __pyx_k_ModelicaParser_Primary_initialCo_2[] = "ModelicaParser.Primary_initialContext.__init__";
static const char __pyx_k_ModelicaParser_Primary_initialCo_3[] = "ModelicaParser.Primary_initialContext.function_call_args";
static const char __pyx_k_ModelicaParser_Primary_initialCo_4[] = "ModelicaParser.Primary_initialContext.enterRule";
static const char __pyx_k_ModelicaParser_Primary_initialCo_5[] = "ModelicaParser.Primary_initialContext.exitRule";
static const char __pyx_k_ModelicaParser_Primary_initialCo_6[] = "ModelicaParser.Primary_initialContext.accept";
static const char __pyx_k_ModelicaParser_Primary_output_ex_2[] = "ModelicaParser.Primary_output_expression_listContext.__init__";
static const char __pyx_k_ModelicaParser_Primary_output_ex_3[] = "ModelicaParser.Primary_output_expression_listContext.output_expression_list";
static const char __pyx_k_ModelicaParser_Primary_output_ex_4[] = "ModelicaParser.Primary_output_expression_listContext.enterRule";
static const char __pyx_k_ModelicaParser_Primary_output_ex_5[] = "ModelicaParser.Primary_output_expression_listContext.exitRule";
static const char __pyx_k_ModelicaParser_Primary_output_ex_6[] = "ModelicaParser.Primary_output_expression_listContext.accept";
static const char __pyx_k_ModelicaParser_Primary_stringCon_2[] = "ModelicaParser.Primary_stringContext.__init__";
static const char __pyx_k_ModelicaParser_Primary_stringCon_3[] = "ModelicaParser.Primary_stringContext.STRING";
static const char __pyx_k_ModelicaParser_Primary_stringCon_4[] = "ModelicaParser.Primary_stringContext.enterRule";
static const char __pyx_k_ModelicaParser_Primary_stringCon_5[] = "ModelicaParser.Primary_stringContext.exitRule";
static const char __pyx_k_ModelicaParser_Primary_stringCon_6[] = "ModelicaParser.Primary_stringContext.accept";
static const char __pyx_k_ModelicaParser_Primary_trueConte_2[] = "ModelicaParser.Primary_trueContext.__init__";
static const char __pyx_k_ModelicaParser_Primary_trueConte_3[] = "ModelicaParser.Primary_trueContext.enterRule";
static const char __pyx_k_ModelicaParser_Primary_trueConte_4[] = "ModelicaParser.Primary_trueContext.exitRule";
static const char __pyx_k_ModelicaParser_Primary_trueConte_5[] = "ModelicaParser.Primary_trueContext.accept";
static const char __pyx_k_ModelicaParser_Regular_elementCo_2[] = "ModelicaParser.Regular_elementContext.__init__";
static const char __pyx_k_ModelicaParser_Regular_elementCo_3[] = "ModelicaParser.Regular_elementContext.REDECLARE";
static const char __pyx_k_ModelicaParser_Regular_elementCo_4[] = "ModelicaParser.Regular_elementContext.FINAL";
static const char __pyx_k_ModelicaParser_Regular_elementCo_5[] = "ModelicaParser.Regular_elementContext.INNER";
static const char __pyx_k_ModelicaParser_Regular_elementCo_6[] = "ModelicaParser.Regular_elementContext.OUTER";
static const char __pyx_k_ModelicaParser_Regular_elementCo_7[] = "ModelicaParser.Regular_elementContext.class_definition";
static const char __pyx_k_ModelicaParser_Regular_elementCo_8[] = "ModelicaParser.Regular_elementContext.component_clause";
static const char __pyx_k_ModelicaParser_Regular_elementCo_9[] = "ModelicaParser.Regular_elementContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Replaceable_eleme_2[] = "ModelicaParser.Replaceable_elementContext.__init__";
static const char __pyx_k_ModelicaParser_Replaceable_eleme_3[] = "ModelicaParser.Replaceable_elementContext.REDECLARE";
static const char __pyx_k_ModelicaParser_Replaceable_eleme_4[] = "ModelicaParser.Replaceable_elementContext.FINAL";
static const char __pyx_k_ModelicaParser_Replaceable_eleme_5[] = "ModelicaParser.Replaceable_elementContext.INNER";
static const char __pyx_k_ModelicaParser_Replaceable_eleme_6[] = "ModelicaParser.Replaceable_elementContext.OUTER";
static const char __pyx_k_ModelicaParser_Replaceable_eleme_7[] = "ModelicaParser.Replaceable_elementContext.class_definition";
static const char __pyx_k_ModelicaParser_Replaceable_eleme_8[] = "ModelicaParser.Replaceable_elementContext.component_clause";
static const char __pyx_k_ModelicaParser_Replaceable_eleme_9[] = "ModelicaParser.Replaceable_elementContext.constraining_clause";
static const char __pyx_k_ModelicaParser_Short_class_defin_2[] = "ModelicaParser.Short_class_definitionContext.__init__";
static const char __pyx_k_ModelicaParser_Short_class_defin_3[] = "ModelicaParser.Short_class_definitionContext.class_prefixes";
static const char __pyx_k_ModelicaParser_Short_class_defin_4[] = "ModelicaParser.Short_class_definitionContext.IDENT";
static const char __pyx_k_ModelicaParser_Short_class_defin_5[] = "ModelicaParser.Short_class_definitionContext.base_prefix";
static const char __pyx_k_ModelicaParser_Short_class_defin_6[] = "ModelicaParser.Short_class_definitionContext.component_reference";
static const char __pyx_k_ModelicaParser_Short_class_defin_7[] = "ModelicaParser.Short_class_definitionContext.comment";
static const char __pyx_k_ModelicaParser_Short_class_defin_8[] = "ModelicaParser.Short_class_definitionContext.array_subscripts";
static const char __pyx_k_ModelicaParser_Short_class_defin_9[] = "ModelicaParser.Short_class_definitionContext.class_modification";
static const char __pyx_k_ModelicaParser_Simple_expression_2[] = "ModelicaParser.Simple_expressionContext.__init__";
static const char __pyx_k_ModelicaParser_Simple_expression_3[] = "ModelicaParser.Simple_expressionContext.expr";
static const char __pyx_k_ModelicaParser_Simple_expression_4[] = "ModelicaParser.Simple_expressionContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Simple_expression_5[] = "ModelicaParser.Simple_expressionContext.enterRule";
static const char __pyx_k_ModelicaParser_Simple_expression_6[] = "ModelicaParser.Simple_expressionContext.exitRule";
static const char __pyx_k_ModelicaParser_Simple_expression_7[] = "ModelicaParser.Simple_expressionContext.accept";
static const char __pyx_k_ModelicaParser_Statement_breakCo_2[] = "ModelicaParser.Statement_breakContext.__init__";
static const char __pyx_k_ModelicaParser_Statement_breakCo_3[] = "ModelicaParser.Statement_breakContext.enterRule";
static const char __pyx_k_ModelicaParser_Statement_breakCo_4[] = "ModelicaParser.Statement_breakContext.exitRule";
static const char __pyx_k_ModelicaParser_Statement_breakCo_5[] = "ModelicaParser.Statement_breakContext.accept";
static const char __pyx_k_ModelicaParser_Statement_compone_2[] = "ModelicaParser.Statement_component_functionContext.__init__";
static const char __pyx_k_ModelicaParser_Statement_compone_3[] = "ModelicaParser.Statement_component_functionContext.component_reference";
static const char __pyx_k_ModelicaParser_Statement_compone_4[] = "ModelicaParser.Statement_component_functionContext.function_call_args";
static const char __pyx_k_ModelicaParser_Statement_compone_5[] = "ModelicaParser.Statement_component_functionContext.enterRule";
static const char __pyx_k_ModelicaParser_Statement_compone_6[] = "ModelicaParser.Statement_component_functionContext.exitRule";
static const char __pyx_k_ModelicaParser_Statement_compone_7[] = "ModelicaParser.Statement_component_functionContext.accept";
static const char __pyx_k_ModelicaParser_Statement_compone_8[] = "ModelicaParser.Statement_component_referenceContext";
static const char __pyx_k_ModelicaParser_Statement_compone_9[] = "ModelicaParser.Statement_component_referenceContext.__init__";
static const char __pyx_k_ModelicaParser_Statement_forCont_2[] = "ModelicaParser.Statement_forContext.__init__";
static const char __pyx_k_ModelicaParser_Statement_forCont_3[] = "ModelicaParser.Statement_forContext.for_statement";
static const char __pyx_k_ModelicaParser_Statement_forCont_4[] = "ModelicaParser.Statement_forContext.enterRule";
static const char __pyx_k_ModelicaParser_Statement_forCont_5[] = "ModelicaParser.Statement_forContext.exitRule";
static const char __pyx_k_ModelicaParser_Statement_forCont_6[] = "ModelicaParser.Statement_forContext.accept";
static const char __pyx_k_ModelicaParser_Statement_ifConte_2[] = "ModelicaParser.Statement_ifContext.__init__";
static const char __pyx_k_ModelicaParser_Statement_ifConte_3[] = "ModelicaParser.Statement_ifContext.if_statement";
static const char __pyx_k_ModelicaParser_Statement_ifConte_4[] = "ModelicaParser.Statement_ifContext.enterRule";
static const char __pyx_k_ModelicaParser_Statement_ifConte_5[] = "ModelicaParser.Statement_ifContext.exitRule";
static const char __pyx_k_ModelicaParser_Statement_ifConte_6[] = "ModelicaParser.Statement_ifContext.accept";
static const char __pyx_k_ModelicaParser_Statement_options_2[] = "ModelicaParser.Statement_optionsContext.__init__";
static const char __pyx_k_ModelicaParser_Statement_options_3[] = "ModelicaParser.Statement_optionsContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Statement_options_4[] = "ModelicaParser.Statement_optionsContext.copyFrom";
static const char __pyx_k_ModelicaParser_Statement_returnC_2[] = "ModelicaParser.Statement_returnContext.__init__";
static const char __pyx_k_ModelicaParser_Statement_returnC_3[] = "ModelicaParser.Statement_returnContext.enterRule";
static const char __pyx_k_ModelicaParser_Statement_returnC_4[] = "ModelicaParser.Statement_returnContext.exitRule";
static const char __pyx_k_ModelicaParser_Statement_returnC_5[] = "ModelicaParser.Statement_returnContext.accept";
static const char __pyx_k_ModelicaParser_Statement_whenCon_2[] = "ModelicaParser.Statement_whenContext.__init__";
static const char __pyx_k_ModelicaParser_Statement_whenCon_3[] = "ModelicaParser.Statement_whenContext.when_statement";
static const char __pyx_k_ModelicaParser_Statement_whenCon_4[] = "ModelicaParser.Statement_whenContext.enterRule";
static const char __pyx_k_ModelicaParser_Statement_whenCon_5[] = "ModelicaParser.Statement_whenContext.exitRule";
static const char __pyx_k_ModelicaParser_Statement_whenCon_6[] = "ModelicaParser.Statement_whenContext.accept";
static const char __pyx_k_ModelicaParser_Statement_whileCo_2[] = "ModelicaParser.Statement_whileContext.__init__";
static const char __pyx_k_ModelicaParser_Statement_whileCo_3[] = "ModelicaParser.Statement_whileContext.while_statement";
static const char __pyx_k_ModelicaParser_Statement_whileCo_4[] = "ModelicaParser.Statement_whileContext.enterRule";
static const char __pyx_k_ModelicaParser_Statement_whileCo_5[] = "ModelicaParser.Statement_whileContext.exitRule";
static const char __pyx_k_ModelicaParser_Statement_whileCo_6[] = "ModelicaParser.Statement_whileContext.accept";
static const char __pyx_k_ModelicaParser_Stored_definition_2[] = "ModelicaParser.Stored_definitionContext.__init__";
static const char __pyx_k_ModelicaParser_Stored_definition_3[] = "ModelicaParser.Stored_definitionContext.WITHIN";
static const char __pyx_k_ModelicaParser_Stored_definition_4[] = "ModelicaParser.Stored_definitionContext.stored_definition_class";
static const char __pyx_k_ModelicaParser_Stored_definition_5[] = "ModelicaParser.Stored_definitionContext.component_reference";
static const char __pyx_k_ModelicaParser_Stored_definition_6[] = "ModelicaParser.Stored_definitionContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Stored_definition_7[] = "ModelicaParser.Stored_definitionContext.enterRule";
static const char __pyx_k_ModelicaParser_Stored_definition_8[] = "ModelicaParser.Stored_definitionContext.exitRule";
static const char __pyx_k_ModelicaParser_Stored_definition_9[] = "ModelicaParser.Stored_definitionContext.accept";
static const char __pyx_k_ModelicaParser_String_commentCon_2[] = "ModelicaParser.String_commentContext.__init__";
static const char __pyx_k_ModelicaParser_String_commentCon_3[] = "ModelicaParser.String_commentContext.STRING";
static const char __pyx_k_ModelicaParser_String_commentCon_4[] = "ModelicaParser.String_commentContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_String_commentCon_5[] = "ModelicaParser.String_commentContext.enterRule";
static const char __pyx_k_ModelicaParser_String_commentCon_6[] = "ModelicaParser.String_commentContext.exitRule";
static const char __pyx_k_ModelicaParser_String_commentCon_7[] = "ModelicaParser.String_commentContext.accept";
static const char __pyx_k_ModelicaParser_Type_prefixContex_2[] = "ModelicaParser.Type_prefixContext.__init__";
static const char __pyx_k_ModelicaParser_Type_prefixContex_3[] = "ModelicaParser.Type_prefixContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Type_prefixContex_4[] = "ModelicaParser.Type_prefixContext.enterRule";
static const char __pyx_k_ModelicaParser_Type_prefixContex_5[] = "ModelicaParser.Type_prefixContext.exitRule";
static const char __pyx_k_ModelicaParser_Type_prefixContex_6[] = "ModelicaParser.Type_prefixContext.accept";
static const char __pyx_k_ModelicaParser_Type_specifierCon_2[] = "ModelicaParser.Type_specifierContext.__init__";
static const char __pyx_k_ModelicaParser_Type_specifierCon_3[] = "ModelicaParser.Type_specifierContext.component_reference";
static const char __pyx_k_ModelicaParser_Type_specifierCon_4[] = "ModelicaParser.Type_specifierContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Type_specifierCon_5[] = "ModelicaParser.Type_specifierContext.enterRule";
static const char __pyx_k_ModelicaParser_Type_specifierCon_6[] = "ModelicaParser.Type_specifierContext.exitRule";
static const char __pyx_k_ModelicaParser_Type_specifierCon_7[] = "ModelicaParser.Type_specifierContext.accept";
static const char __pyx_k_ModelicaParser_When_equationCont_2[] = "ModelicaParser.When_equationContext.__init__";
static const char __pyx_k_ModelicaParser_When_equationCont_3[] = "ModelicaParser.When_equationContext.expression";
static const char __pyx_k_ModelicaParser_When_equationCont_4[] = "ModelicaParser.When_equationContext.equation";
static const char __pyx_k_ModelicaParser_When_equationCont_5[] = "ModelicaParser.When_equationContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_When_equationCont_6[] = "ModelicaParser.When_equationContext.enterRule";
static const char __pyx_k_ModelicaParser_When_equationCont_7[] = "ModelicaParser.When_equationContext.exitRule";
static const char __pyx_k_ModelicaParser_When_equationCont_8[] = "ModelicaParser.When_equationContext.accept";
static const char __pyx_k_ModelicaParser_When_statementCon_2[] = "ModelicaParser.When_statementContext.__init__";
static const char __pyx_k_ModelicaParser_When_statementCon_3[] = "ModelicaParser.When_statementContext.expression";
static const char __pyx_k_ModelicaParser_When_statementCon_4[] = "ModelicaParser.When_statementContext.statement";
static const char __pyx_k_ModelicaParser_When_statementCon_5[] = "ModelicaParser.When_statementContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_When_statementCon_6[] = "ModelicaParser.When_statementContext.enterRule";
static const char __pyx_k_ModelicaParser_When_statementCon_7[] = "ModelicaParser.When_statementContext.exitRule";
static const char __pyx_k_ModelicaParser_When_statementCon_8[] = "ModelicaParser.When_statementContext.accept";
static const char __pyx_k_ModelicaParser_While_statementCo_2[] = "ModelicaParser.While_statementContext.__init__";
static const char __pyx_k_ModelicaParser_While_statementCo_3[] = "ModelicaParser.While_statementContext.expression";
static const char __pyx_k_ModelicaParser_While_statementCo_4[] = "ModelicaParser.While_statementContext.statement";
static const char __pyx_k_ModelicaParser_While_statementCo_5[] = "ModelicaParser.While_statementContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_While_statementCo_6[] = "ModelicaParser.While_statementContext.enterRule";
static const char __pyx_k_ModelicaParser_While_statementCo_7[] = "ModelicaParser.While_statementContext.exitRule";
static const char __pyx_k_ModelicaParser_While_statementCo_8[] = "ModelicaParser.While_statementContext.accept";
static const char __pyx_k_ModelicaParser_component_declara_2[] = "ModelicaParser.component_declaration1";
static const char __pyx_k_ModelicaParser_component_referen_2[] = "ModelicaParser.component_reference";
static const char __pyx_k_ModelicaParser_element_modificat_2[] = "ModelicaParser.element_modification";
static const char __pyx_k_ModelicaParser_function_argument_2[] = "ModelicaParser.function_argument";
static const char __pyx_k_ModelicaParser_stored_definition_2[] = "ModelicaParser.stored_definition_class";
static const char __pyx_k_ModelicaParser_Class_spec_baseCo_10[] = "ModelicaParser.Class_spec_baseContext.exitRule";
static const char __pyx_k_ModelicaParser_Class_spec_baseCo_11[] = "ModelicaParser.Class_spec_baseContext.accept";
static const char __pyx_k_ModelicaParser_Component_clauseC_10[] = "ModelicaParser.Component_clauseContext.accept";
static const char __pyx_k_ModelicaParser_Component_declara_10[] = "ModelicaParser.Component_declaration1Context";
static const char __pyx_k_ModelicaParser_Component_declara_11[] = "ModelicaParser.Component_declaration1Context.__init__";
static const char __pyx_k_ModelicaParser_Component_declara_12[] = "ModelicaParser.Component_declaration1Context.declaration";
static const char __pyx_k_ModelicaParser_Component_declara_13[] = "ModelicaParser.Component_declaration1Context.comment";
static const char __pyx_k_ModelicaParser_Component_declara_14[] = "ModelicaParser.Component_declaration1Context.getRuleIndex";
static const char __pyx_k_ModelicaParser_Component_declara_15[] = "ModelicaParser.Component_declaration1Context.enterRule";
static const char __pyx_k_ModelicaParser_Component_declara_16[] = "ModelicaParser.Component_declaration1Context.exitRule";
static const char __pyx_k_ModelicaParser_Component_declara_17[] = "ModelicaParser.Component_declaration1Context.accept";
static const char __pyx_k_ModelicaParser_Component_referen_10[] = "ModelicaParser.Component_referenceContext.__init__";
static const char __pyx_k_ModelicaParser_Component_referen_11[] = "ModelicaParser.Component_referenceContext.component_reference_element";
static const char __pyx_k_ModelicaParser_Component_referen_12[] = "ModelicaParser.Component_referenceContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Component_referen_13[] = "ModelicaParser.Component_referenceContext.enterRule";
static const char __pyx_k_ModelicaParser_Component_referen_14[] = "ModelicaParser.Component_referenceContext.exitRule";
static const char __pyx_k_ModelicaParser_Component_referen_15[] = "ModelicaParser.Component_referenceContext.accept";
static const char __pyx_k_ModelicaParser_CompositionContex_10[] = "ModelicaParser.CompositionContext.enterRule";
static const char __pyx_k_ModelicaParser_CompositionContex_11[] = "ModelicaParser.CompositionContext.exitRule";
static const char __pyx_k_ModelicaParser_CompositionContex_12[] = "ModelicaParser.CompositionContext.accept";
static const char __pyx_k_ModelicaParser_Element_modificat_10[] = "ModelicaParser.Element_modification_or_replaceableContext.accept";
static const char __pyx_k_ModelicaParser_Element_modificat_11[] = "ModelicaParser.Element_modificationContext";
static const char __pyx_k_ModelicaParser_Element_modificat_12[] = "ModelicaParser.Element_modificationContext.__init__";
static const char __pyx_k_ModelicaParser_Element_modificat_13[] = "ModelicaParser.Element_modificationContext.component_reference";
static const char __pyx_k_ModelicaParser_Element_modificat_14[] = "ModelicaParser.Element_modificationContext.string_comment";
static const char __pyx_k_ModelicaParser_Element_modificat_15[] = "ModelicaParser.Element_modificationContext.modification";
static const char __pyx_k_ModelicaParser_Element_modificat_16[] = "ModelicaParser.Element_modificationContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Element_modificat_17[] = "ModelicaParser.Element_modificationContext.enterRule";
static const char __pyx_k_ModelicaParser_Element_modificat_18[] = "ModelicaParser.Element_modificationContext.exitRule";
static const char __pyx_k_ModelicaParser_Element_modificat_19[] = "ModelicaParser.Element_modificationContext.accept";
static const char __pyx_k_ModelicaParser_Element_redeclara_10[] = "ModelicaParser.Element_redeclarationContext.enterRule";
static const char __pyx_k_ModelicaParser_Element_redeclara_11[] = "ModelicaParser.Element_redeclarationContext.exitRule";
static const char __pyx_k_ModelicaParser_Element_redeclara_12[] = "ModelicaParser.Element_redeclarationContext.accept";
static const char __pyx_k_ModelicaParser_Function_argument_10[] = "ModelicaParser.Function_argumentContext";
static const char __pyx_k_ModelicaParser_Function_argument_11[] = "ModelicaParser.Function_argumentContext.__init__";
static const char __pyx_k_ModelicaParser_Function_argument_12[] = "ModelicaParser.Function_argumentContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Function_argument_13[] = "ModelicaParser.Function_argumentContext.copyFrom";
static const char __pyx_k_ModelicaParser_Import_clauseCont_10[] = "ModelicaParser.Import_clauseContext.accept";
static const char __pyx_k_ModelicaParser_Modification_assi_10[] = "ModelicaParser.Modification_assignmentContext.enterRule";
static const char __pyx_k_ModelicaParser_Modification_assi_11[] = "ModelicaParser.Modification_assignmentContext.exitRule";
static const char __pyx_k_ModelicaParser_Modification_assi_12[] = "ModelicaParser.Modification_assignmentContext.accept";
static const char __pyx_k_ModelicaParser_Regular_elementCo_10[] = "ModelicaParser.Regular_elementContext.enterRule";
static const char __pyx_k_ModelicaParser_Regular_elementCo_11[] = "ModelicaParser.Regular_elementContext.exitRule";
static const char __pyx_k_ModelicaParser_Regular_elementCo_12[] = "ModelicaParser.Regular_elementContext.accept";
static const char __pyx_k_ModelicaParser_Replaceable_eleme_10[] = "ModelicaParser.Replaceable_elementContext.comment";
static const char __pyx_k_ModelicaParser_Replaceable_eleme_11[] = "ModelicaParser.Replaceable_elementContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Replaceable_eleme_12[] = "ModelicaParser.Replaceable_elementContext.enterRule";
static const char __pyx_k_ModelicaParser_Replaceable_eleme_13[] = "ModelicaParser.Replaceable_elementContext.exitRule";
static const char __pyx_k_ModelicaParser_Replaceable_eleme_14[] = "ModelicaParser.Replaceable_elementContext.accept";
static const char __pyx_k_ModelicaParser_Short_class_defin_10[] = "ModelicaParser.Short_class_definitionContext.enum_list";
static const char __pyx_k_ModelicaParser_Short_class_defin_11[] = "ModelicaParser.Short_class_definitionContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Short_class_defin_12[] = "ModelicaParser.Short_class_definitionContext.enterRule";
static const char __pyx_k_ModelicaParser_Short_class_defin_13[] = "ModelicaParser.Short_class_definitionContext.exitRule";
static const char __pyx_k_ModelicaParser_Short_class_defin_14[] = "ModelicaParser.Short_class_definitionContext.accept";
static const char __pyx_k_ModelicaParser_Statement_compone_10[] = "ModelicaParser.Statement_component_referenceContext.component_reference";
static const char __pyx_k_ModelicaParser_Statement_compone_11[] = "ModelicaParser.Statement_component_referenceContext.expression";
static const char __pyx_k_ModelicaParser_Statement_compone_12[] = "ModelicaParser.Statement_component_referenceContext.function_call_args";
static const char __pyx_k_ModelicaParser_Statement_compone_13[] = "ModelicaParser.Statement_component_referenceContext.enterRule";
static const char __pyx_k_ModelicaParser_Statement_compone_14[] = "ModelicaParser.Statement_component_referenceContext.exitRule";
static const char __pyx_k_ModelicaParser_Statement_compone_15[] = "ModelicaParser.Statement_component_referenceContext.accept";
static const char __pyx_k_ModelicaParser_Stored_definition_10[] = "ModelicaParser.Stored_definition_classContext";
static const char __pyx_k_ModelicaParser_Stored_definition_11[] = "ModelicaParser.Stored_definition_classContext.__init__";
static const char __pyx_k_ModelicaParser_Stored_definition_12[] = "ModelicaParser.Stored_definition_classContext.class_definition";
static const char __pyx_k_ModelicaParser_Stored_definition_13[] = "ModelicaParser.Stored_definition_classContext.FINAL";
static const char __pyx_k_ModelicaParser_Stored_definition_14[] = "ModelicaParser.Stored_definition_classContext.getRuleIndex";
static const char __pyx_k_ModelicaParser_Stored_definition_15[] = "ModelicaParser.Stored_definition_classContext.enterRule";
static const char __pyx_k_ModelicaParser_Stored_definition_16[] = "ModelicaParser.Stored_definition_classContext.exitRule";
static const char __pyx_k_ModelicaParser_Stored_definition_17[] = "ModelicaParser.Stored_definition_classContext.accept";
static PyObject *__pyx_kp_u_;
static PyObject *__pyx_kp_u_0;
static PyObject *__pyx_kp_u_02468_BDFHJLNPRTVXZ_bdfhjlnprtv;
static PyObject *__pyx_kp_u_0_0_0;
static PyObject *__pyx_kp_u_0_0_0_0_0_0_0_0;
static PyObject *__pyx_kp_u_0_0_0_1_1_1_1_1_1_1;
static PyObject *__pyx_kp_u_0_0_1_1_2_2_3_3_4_4;
static PyObject *__pyx_kp_u_0_2_4;
static PyObject *__pyx_kp_u_0_T;
static PyObject *__pyx_kp_u_0_T_2;
static PyObject *__pyx_kp_u_0_X;
static PyObject *__pyx_kp_u_0_X_2;
static PyObject *__pyx_kp_u_1;
static PyObject *__pyx_kp_u_1_1_1_1_1_1_1_2_2;
static PyObject *__pyx_kp_u_1_2;
static PyObject *__pyx_kp_u_1_3;
static PyObject *__pyx_kp_u_1_n8;
static PyObject *__pyx_kp_u_1_n8_0;
static PyObject *__pyx_kp_u_2;
static PyObject *__pyx_kp_u_2_2;
static PyObject *__pyx_kp_u_2_2_2_2_2_2_3_3;
static PyObject *__pyx_kp_u_2_X;
static PyObject *__pyx_kp_u_2_n8;
static PyObject *__pyx_kp_u_3;
static PyObject *__pyx_kp_u_3_3_3_4_4_4_4_4_4;
static PyObject *__pyx_kp_u_3_a_d3;
static PyObject *__pyx_kp_u_3_b2;
static PyObject *__pyx_kp_u_3_b2_2;
static PyObject *__pyx_kp_u_3_b2_4;
static PyObject *__pyx_kp_u_4_4_4_4_4_4_4_4;
static PyObject *__pyx_kp_u_4_7;
static PyObject *__pyx_kp_u_4_X;
static PyObject *__pyx_kp_u_5_5_5_5_5_5_5_5;
static PyObject *__pyx_kp_u_5_5_5_5_5_5_5_5_2;
static PyObject *__pyx_kp_u_5_5_5_5_5_5_5_5_5_5;
static PyObject *__pyx_kp_u_5_5_5_5_6_6_6_6_6;
static PyObject *__pyx_kp_u_5_5_6_6_7_7_8_8_9_9;
static PyObject *__pyx_kp_u_6;
static PyObject *__pyx_kp_u_6_4;
static PyObject *__pyx_kp_u_6_6_6_6_6_6_6_6;
static PyObject *__pyx_kp_u_6_6_6_6_6_6_6_6_2;
static PyObject *__pyx_kp_u_6_6_6_6_6_6_6_6_3;
static PyObject *__pyx_kp_u_6_7_7_7_7_7_7_7_8_8_8_8_8;
static PyObject *__pyx_kp_u_6_8;
static PyObject *__pyx_kp_u_6_n8_4;
static PyObject *__pyx_kp_u_7;
static PyObject *__pyx_kp_u_7_i;
static PyObject *__pyx_kp_u_7_k;
static PyObject *__pyx_kp_u_7_n8;
static PyObject *__pyx_kp_u_8;
static PyObject *__pyx_kp_u_8_2;
static PyObject *__pyx_kp_u_8_3;
static PyObject *__pyx_kp_u_8_8_8_8_8_8_8_8_8_8_8_8_8;
static PyObject *__pyx_kp_u_8_8_9_9_9_9_9_9_9_9_9;
static PyObject *__pyx_kp_u_9;
static PyObject *__pyx_kp_u_9_z;
static PyObject *__pyx_kp_u_A;
static PyObject *__pyx_n_s_ATN;
static PyObject *__pyx_n_s_ATNDeserializer;
static PyObject *__pyx_kp_u_A_2;
static PyObject *__pyx_kp_u_A_3;
static PyObject *__pyx_kp_u_A_4;
static PyObject *__pyx_kp_u_A_A_A_A_A_A_A_B_B_B_B_C_C;
static PyObject *__pyx_kp_u_A_A_B_B_C_C_D;
static PyObject *__pyx_n_s_Algorithm_sectionContext;
static PyObject *__pyx_n_s_AnnotationContext;
static PyObject *__pyx_n_s_ArgumentContext;
static PyObject *__pyx_n_s_Argument_expressionContext;
static PyObject *__pyx_n_s_Argument_functionContext;
static PyObject *__pyx_n_s_Argument_listContext;
static PyObject *__pyx_n_s_Array_subscriptsContext;
static PyObject *__pyx_kp_u_B;
static PyObject *__pyx_kp_u_B_2;
static PyObject *__pyx_kp_u_B_3;
static PyObject *__pyx_kp_u_B_CH;
static PyObject *__pyx_kp_u_B_F;
static PyObject *__pyx_n_s_Base_prefixContext;
static PyObject *__pyx_kp_u_C;
static PyObject *__pyx_n_s_COMMENT;
static PyObject *__pyx_n_u_COMMENT;
static PyObject *__pyx_kp_u_C_2;
static PyObject *__pyx_kp_u_C_3;
static PyObject *__pyx_kp_u_C_C_C_C_C_C_C_C_C_D_D;
static PyObject *__pyx_n_s_Class_definitionContext;
static PyObject *__pyx_n_s_Class_modificationContext;
static PyObject *__pyx_n_s_Class_prefixesContext;
static PyObject *__pyx_n_s_Class_spec_baseContext;
static PyObject *__pyx_n_s_Class_spec_compContext;
static PyObject *__pyx_n_s_Class_spec_derContext;
static PyObject *__pyx_n_s_Class_spec_enumContext;
static PyObject *__pyx_n_s_Class_spec_extendsContext;
static PyObject *__pyx_n_s_Class_specifierContext;
static PyObject *__pyx_n_s_Class_typeContext;
static PyObject *__pyx_n_s_CommentContext;
static PyObject *__pyx_n_s_Component_clause1Context;
static PyObject *__pyx_n_s_Component_clauseContext;
static PyObject *__pyx_n_s_Component_declaration1Context;
static PyObject *__pyx_n_s_Component_declarationContext;
static PyObject *__pyx_n_s_Component_listContext;
static PyObject *__pyx_n_s_Component_referenceContext;
static PyObject *__pyx_n_s_Component_reference_elementConte;
static PyObject *__pyx_n_s_CompositionContext;
static PyObject *__pyx_n_s_Condition_attributeContext;
static PyObject *__pyx_n_s_Connect_clauseContext;
static PyObject *__pyx_n_s_Constraining_clauseContext;
static PyObject *__pyx_kp_u_D;
static PyObject *__pyx_n_s_DFA;
static PyObject *__pyx_kp_u_D_2;
static PyObject *__pyx_kp_u_D_D_D_D_D_D_D_D_E_E_E_E;
static PyObject *__pyx_kp_u_D_E_E_F_F_G_G_H_H_I_I_J_J;
static PyObject *__pyx_kp_u_D_F_H_J;
static PyObject *__pyx_n_s_DeclarationContext;
static PyObject *__pyx_kp_u_E;
static PyObject *__pyx_n_s_EACH;
static PyObject *__pyx_n_u_EACH;
static PyObject *__pyx_n_s_ENCAPSULATED;
static PyObject *__pyx_n_u_ENCAPSULATED;
static PyObject *__pyx_n_s_EOF;
static PyObject *__pyx_kp_u_E_2;
static PyObject *__pyx_kp_u_E_E;
static PyObject *__pyx_kp_u_E_E_E_E_F_F_F_F_F_F_F_F;
static PyObject *__pyx_n_s_ElementContext;
static PyObject *__pyx_n_s_Element_listContext;
static PyObject *__pyx_n_s_Element_modificationContext;
static PyObject *__pyx_n_s_Element_modification_or_replacea;
static PyObject *__pyx_n_s_Element_redeclarationContext;
static PyObject *__pyx_n_s_Element_replaceableContext;
static PyObject *__pyx_n_s_Enum_listContext;
static PyObject *__pyx_n_s_Enumeration_literalContext;
static PyObject *__pyx_n_s_EquationContext;
static PyObject *__pyx_n_s_Equation_connect_clauseContext;
static PyObject *__pyx_n_s_Equation_forContext;
static PyObject *__pyx_n_s_Equation_functionContext;
static PyObject *__pyx_n_s_Equation_ifContext;
static PyObject *__pyx_n_s_Equation_optionsContext;
static PyObject *__pyx_n_s_Equation_sectionContext;
static PyObject *__pyx_n_s_Equation_simpleContext;
static PyObject *__pyx_n_s_Equation_whenContext;
static PyObject *__pyx_n_s_ExprContext;
static PyObject *__pyx_n_s_Expr_addContext;
static PyObject *__pyx_n_s_Expr_andContext;
static PyObject *__pyx_n_s_Expr_expContext;
static PyObject *__pyx_n_s_Expr_mulContext;
static PyObject *__pyx_n_s_Expr_negContext;
static PyObject *__pyx_n_s_Expr_notContext;
static PyObject *__pyx_n_s_Expr_orContext;
static PyObject *__pyx_n_s_Expr_primaryContext;
static PyObject *__pyx_n_s_Expr_relContext;
static PyObject *__pyx_n_s_ExpressionContext;
static PyObject *__pyx_n_s_Expression_ifContext;
static PyObject *__pyx_n_s_Expression_listContext;
static PyObject *__pyx_n_s_Expression_simpleContext;
static PyObject *__pyx_n_s_Extends_clauseContext;
static PyObject *__pyx_n_s_External_function_callContext;
static PyObject *__pyx_kp_u_F;
static PyObject *__pyx_n_s_FINAL;
static PyObject *__pyx_n_u_FINAL;
static PyObject *__pyx_kp_u_F_2;
static PyObject *__pyx_kp_u_F_F_F_G_G_G_G_H_H_H_H_I_I;
static PyObject *__pyx_n_s_FailedPredicateException;
static PyObject *__pyx_n_s_For_equationContext;
static PyObject *__pyx_n_s_For_indexContext;
static PyObject *__pyx_n_s_For_indicesContext;
static PyObject *__pyx_n_s_For_statementContext;
static PyObject *__pyx_n_s_Function_argumentContext;
static PyObject *__pyx_n_s_Function_argumentsContext;
static PyObject *__pyx_n_s_Function_call_argsContext;
static PyObject *__pyx_kp_u_G;
static PyObject *__pyx_kp_u_G_2;
static PyObject *__pyx_kp_u_H;
static PyObject *__pyx_kp_u_H_2;
static PyObject *__pyx_kp_u_H_3;
static PyObject *__pyx_kp_u_H_3_2;
static PyObject *__pyx_kp_u_H_4;
static PyObject *__pyx_kp_u_H_5;
static PyObject *__pyx_kp_u_H_6;
static PyObject *__pyx_kp_u_H_7;
static PyObject *__pyx_kp_u_H_8;
static PyObject *__pyx_kp_u_H_K;
static PyObject *__pyx_kp_u_H_Z;
static PyObject *__pyx_kp_u_I;
static PyObject *__pyx_n_s_IDENT;
static PyObject *__pyx_n_u_IDENT;
static PyObject *__pyx_n_s_INITIAL;
static PyObject *__pyx_n_u_INITIAL;
static PyObject *__pyx_n_s_INNER;
static PyObject *__pyx_n_u_INNER;
static PyObject *__pyx_kp_u_INVALID;
static PyObject *__pyx_n_s_INVALID_ALT_NUMBER;
static PyObject *__pyx_kp_u_I_C_V_Q;
static PyObject *__pyx_kp_u_I_I_I_I_I_I_I_I_J_J_J;
static PyObject *__pyx_kp_u_I_J;
static PyObject *__pyx_kp_u_I_r;
static PyObject *__pyx_n_s_If_equationContext;
static PyObject *__pyx_n_s_If_statementContext;
static PyObject *__pyx_n_s_Import_clauseContext;
static PyObject *__pyx_n_s_Import_listContext;
static PyObject *__pyx_kp_u_J;
static PyObject *__pyx_kp_u_J_2;
static PyObject *__pyx_kp_u_J_3;
static PyObject *__pyx_kp_u_J_I_6;
static PyObject *__pyx_kp_u_J_rK;
static PyObject *__pyx_kp_u_L;
static PyObject *__pyx_n_s_LA;
static PyObject *__pyx_n_s_LT;
static PyObject *__pyx_kp_u_L_2;
static PyObject *__pyx_kp_u_L_H;
static PyObject *__pyx_kp_u_L_N_P_R;
static PyObject *__pyx_n_s_Language_specificationContext;
static PyObject *__pyx_kp_u_M;
static PyObject *__pyx_kp_u_M_z;
static PyObject *__pyx_n_s_ModelicaParser;
static PyObject *__pyx_n_s_ModelicaParser_Algorithm_section;
static PyObject *__pyx_n_s_ModelicaParser_Algorithm_section_2;
static PyObject *__pyx_n_s_ModelicaParser_Algorithm_section_3;
static PyObject *__pyx_n_s_ModelicaParser_Algorithm_section_4;
static PyObject *__pyx_n_s_ModelicaParser_Algorithm_section_5;
static PyObject *__pyx_n_s_ModelicaParser_Algorithm_section_6;
static PyObject *__pyx_n_s_ModelicaParser_Algorithm_section_7;
static PyObject *__pyx_n_s_ModelicaParser_Algorithm_section_8;
static PyObject *__pyx_n_s_ModelicaParser_AnnotationContext;
static PyObject *__pyx_n_s_ModelicaParser_AnnotationContext_2;
static PyObject *__pyx_n_s_ModelicaParser_AnnotationContext_3;
static PyObject *__pyx_n_s_ModelicaParser_AnnotationContext_4;
static PyObject *__pyx_n_s_ModelicaParser_AnnotationContext_5;
static PyObject *__pyx_n_s_ModelicaParser_AnnotationContext_6;
static PyObject *__pyx_n_s_ModelicaParser_AnnotationContext_7;
static PyObject *__pyx_n_s_ModelicaParser_ArgumentContext;
static PyObject *__pyx_n_s_ModelicaParser_ArgumentContext_2;
static PyObject *__pyx_n_s_ModelicaParser_ArgumentContext_a;
static PyObject *__pyx_n_s_ModelicaParser_ArgumentContext_e;
static PyObject *__pyx_n_s_ModelicaParser_ArgumentContext_e_2;
static PyObject *__pyx_n_s_ModelicaParser_ArgumentContext_e_3;
static PyObject *__pyx_n_s_ModelicaParser_ArgumentContext_e_4;
static PyObject *__pyx_n_s_ModelicaParser_ArgumentContext_g;
static PyObject *__pyx_n_s_ModelicaParser_Argument_expressi;
static PyObject *__pyx_n_s_ModelicaParser_Argument_expressi_2;
static PyObject *__pyx_n_s_ModelicaParser_Argument_expressi_3;
static PyObject *__pyx_n_s_ModelicaParser_Argument_expressi_4;
static PyObject *__pyx_n_s_ModelicaParser_Argument_expressi_5;
static PyObject *__pyx_n_s_ModelicaParser_Argument_expressi_6;
static PyObject *__pyx_n_s_ModelicaParser_Argument_function;
static PyObject *__pyx_n_s_ModelicaParser_Argument_function_2;
static PyObject *__pyx_n_s_ModelicaParser_Argument_function_3;
static PyObject *__pyx_n_s_ModelicaParser_Argument_function_4;
static PyObject *__pyx_n_s_ModelicaParser_Argument_function_5;
static PyObject *__pyx_n_s_ModelicaParser_Argument_function_6;
static PyObject *__pyx_n_s_ModelicaParser_Argument_function_7;
static PyObject *__pyx_n_s_ModelicaParser_Argument_listCont;
static PyObject *__pyx_n_s_ModelicaParser_Argument_listCont_2;
static PyObject *__pyx_n_s_ModelicaParser_Argument_listCont_3;
static PyObject *__pyx_n_s_ModelicaParser_Argument_listCont_4;
static PyObject *__pyx_n_s_ModelicaParser_Argument_listCont_5;
static PyObject *__pyx_n_s_ModelicaParser_Argument_listCont_6;
static PyObject *__pyx_n_s_ModelicaParser_Argument_listCont_7;
static PyObject *__pyx_n_s_ModelicaParser_Array_subscriptsC;
static PyObject *__pyx_n_s_ModelicaParser_Array_subscriptsC_2;
static PyObject *__pyx_n_s_ModelicaParser_Array_subscriptsC_3;
static PyObject *__pyx_n_s_ModelicaParser_Array_subscriptsC_4;
static PyObject *__pyx_n_s_ModelicaParser_Array_subscriptsC_5;
static PyObject *__pyx_n_s_ModelicaParser_Array_subscriptsC_6;
static PyObject *__pyx_n_s_ModelicaParser_Array_subscriptsC_7;
static PyObject *__pyx_n_s_ModelicaParser_Base_prefixContex;
static PyObject *__pyx_n_s_ModelicaParser_Base_prefixContex_2;
static PyObject *__pyx_n_s_ModelicaParser_Base_prefixContex_3;
static PyObject *__pyx_n_s_ModelicaParser_Base_prefixContex_4;
static PyObject *__pyx_n_s_ModelicaParser_Base_prefixContex_5;
static PyObject *__pyx_n_s_ModelicaParser_Base_prefixContex_6;
static PyObject *__pyx_n_s_ModelicaParser_Base_prefixContex_7;
static PyObject *__pyx_n_s_ModelicaParser_Class_definitionC;
static PyObject *__pyx_n_s_ModelicaParser_Class_definitionC_2;
static PyObject *__pyx_n_s_ModelicaParser_Class_definitionC_3;
static PyObject *__pyx_n_s_ModelicaParser_Class_definitionC_4;
static PyObject *__pyx_n_s_ModelicaParser_Class_definitionC_5;
static PyObject *__pyx_n_s_ModelicaParser_Class_definitionC_6;
static PyObject *__pyx_n_s_ModelicaParser_Class_definitionC_7;
static PyObject *__pyx_n_s_ModelicaParser_Class_definitionC_8;
static PyObject *__pyx_n_s_ModelicaParser_Class_definitionC_9;
static PyObject *__pyx_n_s_ModelicaParser_Class_modificatio;
static PyObject *__pyx_n_s_ModelicaParser_Class_modificatio_2;
static PyObject *__pyx_n_s_ModelicaParser_Class_modificatio_3;
static PyObject *__pyx_n_s_ModelicaParser_Class_modificatio_4;
static PyObject *__pyx_n_s_ModelicaParser_Class_modificatio_5;
static PyObject *__pyx_n_s_ModelicaParser_Class_modificatio_6;
static PyObject *__pyx_n_s_ModelicaParser_Class_modificatio_7;
static PyObject *__pyx_n_s_ModelicaParser_Class_prefixesCon;
static PyObject *__pyx_n_s_ModelicaParser_Class_prefixesCon_2;
static PyObject *__pyx_n_s_ModelicaParser_Class_prefixesCon_3;
static PyObject *__pyx_n_s_ModelicaParser_Class_prefixesCon_4;
static PyObject *__pyx_n_s_ModelicaParser_Class_prefixesCon_5;
static PyObject *__pyx_n_s_ModelicaParser_Class_prefixesCon_6;
static PyObject *__pyx_n_s_ModelicaParser_Class_prefixesCon_7;
static PyObject *__pyx_n_s_ModelicaParser_Class_prefixesCon_8;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_baseCo;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_baseCo_10;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_baseCo_11;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_baseCo_2;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_baseCo_3;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_baseCo_4;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_baseCo_5;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_baseCo_6;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_baseCo_7;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_baseCo_8;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_baseCo_9;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_compCo;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_compCo_2;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_compCo_3;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_compCo_4;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_compCo_5;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_compCo_6;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_compCo_7;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_compCo_8;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_derCon;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_derCon_2;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_derCon_3;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_derCon_4;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_derCon_5;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_derCon_6;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_derCon_7;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_derCon_8;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_enumCo;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_enumCo_2;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_enumCo_3;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_enumCo_4;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_enumCo_5;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_enumCo_6;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_enumCo_7;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_enumCo_8;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_extend;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_extend_2;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_extend_3;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_extend_4;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_extend_5;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_extend_6;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_extend_7;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_extend_8;
static PyObject *__pyx_n_s_ModelicaParser_Class_spec_extend_9;
static PyObject *__pyx_n_s_ModelicaParser_Class_specifierCo;
static PyObject *__pyx_n_s_ModelicaParser_Class_specifierCo_2;
static PyObject *__pyx_n_s_ModelicaParser_Class_specifierCo_3;
static PyObject *__pyx_n_s_ModelicaParser_Class_specifierCo_4;
static PyObject *__pyx_n_s_ModelicaParser_Class_typeContext;
static PyObject *__pyx_n_s_ModelicaParser_Class_typeContext_2;
static PyObject *__pyx_n_s_ModelicaParser_Class_typeContext_3;
static PyObject *__pyx_n_s_ModelicaParser_Class_typeContext_4;
static PyObject *__pyx_n_s_ModelicaParser_Class_typeContext_5;
static PyObject *__pyx_n_s_ModelicaParser_Class_typeContext_6;
static PyObject *__pyx_n_s_ModelicaParser_CommentContext;
static PyObject *__pyx_n_s_ModelicaParser_CommentContext_2;
static PyObject *__pyx_n_s_ModelicaParser_CommentContext_ac;
static PyObject *__pyx_n_s_ModelicaParser_CommentContext_an;
static PyObject *__pyx_n_s_ModelicaParser_CommentContext_en;
static PyObject *__pyx_n_s_ModelicaParser_CommentContext_ex;
static PyObject *__pyx_n_s_ModelicaParser_CommentContext_ge;
static PyObject *__pyx_n_s_ModelicaParser_CommentContext_st;
static PyObject *__pyx_n_s_ModelicaParser_Component_clause1;
static PyObject *__pyx_n_s_ModelicaParser_Component_clause1_2;
static PyObject *__pyx_n_s_ModelicaParser_Component_clause1_3;
static PyObject *__pyx_n_s_ModelicaParser_Component_clause1_4;
static PyObject *__pyx_n_s_ModelicaParser_Component_clause1_5;
static PyObject *__pyx_n_s_ModelicaParser_Component_clause1_6;
static PyObject *__pyx_n_s_ModelicaParser_Component_clause1_7;
static PyObject *__pyx_n_s_ModelicaParser_Component_clause1_8;
static PyObject *__pyx_n_s_ModelicaParser_Component_clause1_9;
static PyObject *__pyx_n_s_ModelicaParser_Component_clauseC;
static PyObject *__pyx_n_s_ModelicaParser_Component_clauseC_10;
static PyObject *__pyx_n_s_ModelicaParser_Component_clauseC_2;
static PyObject *__pyx_n_s_ModelicaParser_Component_clauseC_3;
static PyObject *__pyx_n_s_ModelicaParser_Component_clauseC_4;
static PyObject *__pyx_n_s_ModelicaParser_Component_clauseC_5;
static PyObject *__pyx_n_s_ModelicaParser_Component_clauseC_6;
static PyObject *__pyx_n_s_ModelicaParser_Component_clauseC_7;
static PyObject *__pyx_n_s_ModelicaParser_Component_clauseC_8;
static PyObject *__pyx_n_s_ModelicaParser_Component_clauseC_9;
static PyObject *__pyx_n_s_ModelicaParser_Component_declara;
static PyObject *__pyx_n_s_ModelicaParser_Component_declara_10;
static PyObject *__pyx_n_s_ModelicaParser_Component_declara_11;
static PyObject *__pyx_n_s_ModelicaParser_Component_declara_12;
static PyObject *__pyx_n_s_ModelicaParser_Component_declara_13;
static PyObject *__pyx_n_s_ModelicaParser_Component_declara_14;
static PyObject *__pyx_n_s_ModelicaParser_Component_declara_15;
static PyObject *__pyx_n_s_ModelicaParser_Component_declara_16;
static PyObject *__pyx_n_s_ModelicaParser_Component_declara_17;
static PyObject *__pyx_n_s_ModelicaParser_Component_declara_2;
static PyObject *__pyx_n_s_ModelicaParser_Component_declara_3;
static PyObject *__pyx_n_s_ModelicaParser_Component_declara_4;
static PyObject *__pyx_n_s_ModelicaParser_Component_declara_5;
static PyObject *__pyx_n_s_ModelicaParser_Component_declara_6;
static PyObject *__pyx_n_s_ModelicaParser_Component_declara_7;
static PyObject *__pyx_n_s_ModelicaParser_Component_declara_8;
static PyObject *__pyx_n_s_ModelicaParser_Component_declara_9;
static PyObject *__pyx_n_s_ModelicaParser_Component_listCon;
static PyObject *__pyx_n_s_ModelicaParser_Component_listCon_2;
static PyObject *__pyx_n_s_ModelicaParser_Component_listCon_3;
static PyObject *__pyx_n_s_ModelicaParser_Component_listCon_4;
static PyObject *__pyx_n_s_ModelicaParser_Component_listCon_5;
static PyObject *__pyx_n_s_ModelicaParser_Component_listCon_6;
static PyObject *__pyx_n_s_ModelicaParser_Component_listCon_7;
static PyObject *__pyx_n_s_ModelicaParser_Component_referen;
static PyObject *__pyx_n_s_ModelicaParser_Component_referen_10;
static PyObject *__pyx_n_s_ModelicaParser_Component_referen_11;
static PyObject *__pyx_n_s_ModelicaParser_Component_referen_12;
static PyObject *__pyx_n_s_ModelicaParser_Component_referen_13;
static PyObject *__pyx_n_s_ModelicaParser_Component_referen_14;
static PyObject *__pyx_n_s_ModelicaParser_Component_referen_15;
static PyObject *__pyx_n_s_ModelicaParser_Component_referen_2;
static PyObject *__pyx_n_s_ModelicaParser_Component_referen_3;
static PyObject *__pyx_n_s_ModelicaParser_Component_referen_4;
static PyObject *__pyx_n_s_ModelicaParser_Component_referen_5;
static PyObject *__pyx_n_s_ModelicaParser_Component_referen_6;
static PyObject *__pyx_n_s_ModelicaParser_Component_referen_7;
static PyObject *__pyx_n_s_ModelicaParser_Component_referen_8;
static PyObject *__pyx_n_s_ModelicaParser_Component_referen_9;
static PyObject *__pyx_n_s_ModelicaParser_CompositionContex;
static PyObject *__pyx_n_s_ModelicaParser_CompositionContex_10;
static PyObject *__pyx_n_s_ModelicaParser_CompositionContex_11;
static PyObject *__pyx_n_s_ModelicaParser_CompositionContex_12;
static PyObject *__pyx_n_s_ModelicaParser_CompositionContex_2;
static PyObject *__pyx_n_s_ModelicaParser_CompositionContex_3;
static PyObject *__pyx_n_s_ModelicaParser_CompositionContex_4;
static PyObject *__pyx_n_s_ModelicaParser_CompositionContex_5;
static PyObject *__pyx_n_s_ModelicaParser_CompositionContex_6;
static PyObject *__pyx_n_s_ModelicaParser_CompositionContex_7;
static PyObject *__pyx_n_s_ModelicaParser_CompositionContex_8;
static PyObject *__pyx_n_s_ModelicaParser_CompositionContex_9;
static PyObject *__pyx_n_s_ModelicaParser_Condition_attribu;
static PyObject *__pyx_n_s_ModelicaParser_Condition_attribu_2;
static PyObject *__pyx_n_s_ModelicaParser_Condition_attribu_3;
static PyObject *__pyx_n_s_ModelicaParser_Condition_attribu_4;
static PyObject *__pyx_n_s_ModelicaParser_Condition_attribu_5;
static PyObject *__pyx_n_s_ModelicaParser_Condition_attribu_6;
static PyObject *__pyx_n_s_ModelicaParser_Condition_attribu_7;
static PyObject *__pyx_n_s_ModelicaParser_Connect_clauseCon;
static PyObject *__pyx_n_s_ModelicaParser_Connect_clauseCon_2;
static PyObject *__pyx_n_s_ModelicaParser_Connect_clauseCon_3;
static PyObject *__pyx_n_s_ModelicaParser_Connect_clauseCon_4;
static PyObject *__pyx_n_s_ModelicaParser_Connect_clauseCon_5;
static PyObject *__pyx_n_s_ModelicaParser_Connect_clauseCon_6;
static PyObject *__pyx_n_s_ModelicaParser_Connect_clauseCon_7;
static PyObject *__pyx_n_s_ModelicaParser_Constraining_clau;
static PyObject *__pyx_n_s_ModelicaParser_Constraining_clau_2;
static PyObject *__pyx_n_s_ModelicaParser_Constraining_clau_3;
static PyObject *__pyx_n_s_ModelicaParser_Constraining_clau_4;
static PyObject *__pyx_n_s_ModelicaParser_Constraining_clau_5;
static PyObject *__pyx_n_s_ModelicaParser_Constraining_clau_6;
static PyObject *__pyx_n_s_ModelicaParser_Constraining_clau_7;
static PyObject *__pyx_n_s_ModelicaParser_Constraining_clau_8;
static PyObject *__pyx_n_s_ModelicaParser_DeclarationContex;
static PyObject *__pyx_n_s_ModelicaParser_DeclarationContex_2;
static PyObject *__pyx_n_s_ModelicaParser_DeclarationContex_3;
static PyObject *__pyx_n_s_ModelicaParser_DeclarationContex_4;
static PyObject *__pyx_n_s_ModelicaParser_DeclarationContex_5;
static PyObject *__pyx_n_s_ModelicaParser_DeclarationContex_6;
static PyObject *__pyx_n_s_ModelicaParser_DeclarationContex_7;
static PyObject *__pyx_n_s_ModelicaParser_DeclarationContex_8;
static PyObject *__pyx_n_s_ModelicaParser_DeclarationContex_9;
static PyObject *__pyx_n_s_ModelicaParser_ElementContext;
static PyObject *__pyx_n_s_ModelicaParser_ElementContext_2;
static PyObject *__pyx_n_s_ModelicaParser_ElementContext_ac;
static PyObject *__pyx_n_s_ModelicaParser_ElementContext_en;
static PyObject *__pyx_n_s_ModelicaParser_ElementContext_ex;
static PyObject *__pyx_n_s_ModelicaParser_ElementContext_ex_2;
static PyObject *__pyx_n_s_ModelicaParser_ElementContext_ge;
static PyObject *__pyx_n_s_ModelicaParser_ElementContext_im;
static PyObject *__pyx_n_s_ModelicaParser_ElementContext_re;
static PyObject *__pyx_n_s_ModelicaParser_ElementContext_re_2;
static PyObject *__pyx_n_s_ModelicaParser_Element_listConte;
static PyObject *__pyx_n_s_ModelicaParser_Element_listConte_2;
static PyObject *__pyx_n_s_ModelicaParser_Element_listConte_3;
static PyObject *__pyx_n_s_ModelicaParser_Element_listConte_4;
static PyObject *__pyx_n_s_ModelicaParser_Element_listConte_5;
static PyObject *__pyx_n_s_ModelicaParser_Element_listConte_6;
static PyObject *__pyx_n_s_ModelicaParser_Element_listConte_7;
static PyObject *__pyx_n_s_ModelicaParser_Element_modificat;
static PyObject *__pyx_n_s_ModelicaParser_Element_modificat_10;
static PyObject *__pyx_n_s_ModelicaParser_Element_modificat_11;
static PyObject *__pyx_n_s_ModelicaParser_Element_modificat_12;
static PyObject *__pyx_n_s_ModelicaParser_Element_modificat_13;
static PyObject *__pyx_n_s_ModelicaParser_Element_modificat_14;
static PyObject *__pyx_n_s_ModelicaParser_Element_modificat_15;
static PyObject *__pyx_n_s_ModelicaParser_Element_modificat_16;
static PyObject *__pyx_n_s_ModelicaParser_Element_modificat_17;
static PyObject *__pyx_n_s_ModelicaParser_Element_modificat_18;
static PyObject *__pyx_n_s_ModelicaParser_Element_modificat_19;
static PyObject *__pyx_n_s_ModelicaParser_Element_modificat_2;
static PyObject *__pyx_n_s_ModelicaParser_Element_modificat_3;
static PyObject *__pyx_n_s_ModelicaParser_Element_modificat_4;
static PyObject *__pyx_n_s_ModelicaParser_Element_modificat_5;
static PyObject *__pyx_n_s_ModelicaParser_Element_modificat_6;
static PyObject *__pyx_n_s_ModelicaParser_Element_modificat_7;
static PyObject *__pyx_n_s_ModelicaParser_Element_modificat_8;
static PyObject *__pyx_n_s_ModelicaParser_Element_modificat_9;
static PyObject *__pyx_n_s_ModelicaParser_Element_redeclara;
static PyObject *__pyx_n_s_ModelicaParser_Element_redeclara_10;
static PyObject *__pyx_n_s_ModelicaParser_Element_redeclara_11;
static PyObject *__pyx_n_s_ModelicaParser_Element_redeclara_12;
static PyObject *__pyx_n_s_ModelicaParser_Element_redeclara_2;
static PyObject *__pyx_n_s_ModelicaParser_Element_redeclara_3;
static PyObject *__pyx_n_s_ModelicaParser_Element_redeclara_4;
static PyObject *__pyx_n_s_ModelicaParser_Element_redeclara_5;
static PyObject *__pyx_n_s_ModelicaParser_Element_redeclara_6;
static PyObject *__pyx_n_s_ModelicaParser_Element_redeclara_7;
static PyObject *__pyx_n_s_ModelicaParser_Element_redeclara_8;
static PyObject *__pyx_n_s_ModelicaParser_Element_redeclara_9;
static PyObject *__pyx_n_s_ModelicaParser_Element_replaceab;
static PyObject *__pyx_n_s_ModelicaParser_Element_replaceab_2;
static PyObject *__pyx_n_s_ModelicaParser_Element_replaceab_3;
static PyObject *__pyx_n_s_ModelicaParser_Element_replaceab_4;
static PyObject *__pyx_n_s_ModelicaParser_Element_replaceab_5;
static PyObject *__pyx_n_s_ModelicaParser_Element_replaceab_6;
static PyObject *__pyx_n_s_ModelicaParser_Element_replaceab_7;
static PyObject *__pyx_n_s_ModelicaParser_Element_replaceab_8;
static PyObject *__pyx_n_s_ModelicaParser_Element_replaceab_9;
static PyObject *__pyx_n_s_ModelicaParser_Enum_listContext;
static PyObject *__pyx_n_s_ModelicaParser_Enum_listContext_2;
static PyObject *__pyx_n_s_ModelicaParser_Enum_listContext_3;
static PyObject *__pyx_n_s_ModelicaParser_Enum_listContext_4;
static PyObject *__pyx_n_s_ModelicaParser_Enum_listContext_5;
static PyObject *__pyx_n_s_ModelicaParser_Enum_listContext_6;
static PyObject *__pyx_n_s_ModelicaParser_Enum_listContext_7;
static PyObject *__pyx_n_s_ModelicaParser_Enumeration_liter;
static PyObject *__pyx_n_s_ModelicaParser_Enumeration_liter_2;
static PyObject *__pyx_n_s_ModelicaParser_Enumeration_liter_3;
static PyObject *__pyx_n_s_ModelicaParser_Enumeration_liter_4;
static PyObject *__pyx_n_s_ModelicaParser_Enumeration_liter_5;
static PyObject *__pyx_n_s_ModelicaParser_Enumeration_liter_6;
static PyObject *__pyx_n_s_ModelicaParser_Enumeration_liter_7;
static PyObject *__pyx_n_s_ModelicaParser_Enumeration_liter_8;
static PyObject *__pyx_n_s_ModelicaParser_EquationContext;
static PyObject *__pyx_n_s_ModelicaParser_EquationContext_2;
static PyObject *__pyx_n_s_ModelicaParser_EquationContext_a;
static PyObject *__pyx_n_s_ModelicaParser_EquationContext_c;
static PyObject *__pyx_n_s_ModelicaParser_EquationContext_e;
static PyObject *__pyx_n_s_ModelicaParser_EquationContext_e_2;
static PyObject *__pyx_n_s_ModelicaParser_EquationContext_e_3;
static PyObject *__pyx_n_s_ModelicaParser_EquationContext_g;
static PyObject *__pyx_n_s_ModelicaParser_Equation_connect;
static PyObject *__pyx_n_s_ModelicaParser_Equation_connect_2;
static PyObject *__pyx_n_s_ModelicaParser_Equation_connect_3;
static PyObject *__pyx_n_s_ModelicaParser_Equation_connect_4;
static PyObject *__pyx_n_s_ModelicaParser_Equation_connect_5;
static PyObject *__pyx_n_s_ModelicaParser_Equation_connect_6;
static PyObject *__pyx_n_s_ModelicaParser_Equation_forConte;
static PyObject *__pyx_n_s_ModelicaParser_Equation_forConte_2;
static PyObject *__pyx_n_s_ModelicaParser_Equation_forConte_3;
static PyObject *__pyx_n_s_ModelicaParser_Equation_forConte_4;
static PyObject *__pyx_n_s_ModelicaParser_Equation_forConte_5;
static PyObject *__pyx_n_s_ModelicaParser_Equation_forConte_6;
static PyObject *__pyx_n_s_ModelicaParser_Equation_function;
static PyObject *__pyx_n_s_ModelicaParser_Equation_function_2;
static PyObject *__pyx_n_s_ModelicaParser_Equation_function_3;
static PyObject *__pyx_n_s_ModelicaParser_Equation_function_4;
static PyObject *__pyx_n_s_ModelicaParser_Equation_function_5;
static PyObject *__pyx_n_s_ModelicaParser_Equation_function_6;
static PyObject *__pyx_n_s_ModelicaParser_Equation_function_7;
static PyObject *__pyx_n_s_ModelicaParser_Equation_ifContex;
static PyObject *__pyx_n_s_ModelicaParser_Equation_ifContex_2;
static PyObject *__pyx_n_s_ModelicaParser_Equation_ifContex_3;
static PyObject *__pyx_n_s_ModelicaParser_Equation_ifContex_4;
static PyObject *__pyx_n_s_ModelicaParser_Equation_ifContex_5;
static PyObject *__pyx_n_s_ModelicaParser_Equation_ifContex_6;
static PyObject *__pyx_n_s_ModelicaParser_Equation_optionsC;
static PyObject *__pyx_n_s_ModelicaParser_Equation_optionsC_2;
static PyObject *__pyx_n_s_ModelicaParser_Equation_optionsC_3;
static PyObject *__pyx_n_s_ModelicaParser_Equation_optionsC_4;
static PyObject *__pyx_n_s_ModelicaParser_Equation_sectionC;
static PyObject *__pyx_n_s_ModelicaParser_Equation_sectionC_2;
static PyObject *__pyx_n_s_ModelicaParser_Equation_sectionC_3;
static PyObject *__pyx_n_s_ModelicaParser_Equation_sectionC_4;
static PyObject *__pyx_n_s_ModelicaParser_Equation_sectionC_5;
static PyObject *__pyx_n_s_ModelicaParser_Equation_sectionC_6;
static PyObject *__pyx_n_s_ModelicaParser_Equation_sectionC_7;
static PyObject *__pyx_n_s_ModelicaParser_Equation_sectionC_8;
static PyObject *__pyx_n_s_ModelicaParser_Equation_simpleCo;
static PyObject *__pyx_n_s_ModelicaParser_Equation_simpleCo_2;
static PyObject *__pyx_n_s_ModelicaParser_Equation_simpleCo_3;
static PyObject *__pyx_n_s_ModelicaParser_Equation_simpleCo_4;
static PyObject *__pyx_n_s_ModelicaParser_Equation_simpleCo_5;
static PyObject *__pyx_n_s_ModelicaParser_Equation_simpleCo_6;
static PyObject *__pyx_n_s_ModelicaParser_Equation_simpleCo_7;
static PyObject *__pyx_n_s_ModelicaParser_Equation_whenCont;
static PyObject *__pyx_n_s_ModelicaParser_Equation_whenCont_2;
static PyObject *__pyx_n_s_ModelicaParser_Equation_whenCont_3;
static PyObject *__pyx_n_s_ModelicaParser_Equation_whenCont_4;
static PyObject *__pyx_n_s_ModelicaParser_Equation_whenCont_5;
static PyObject *__pyx_n_s_ModelicaParser_Equation_whenCont_6;
static PyObject *__pyx_n_s_ModelicaParser_ExprContext;
static PyObject *__pyx_n_s_ModelicaParser_ExprContext___ini;
static PyObject *__pyx_n_s_ModelicaParser_ExprContext_copyF;
static PyObject *__pyx_n_s_ModelicaParser_ExprContext_getRu;
static PyObject *__pyx_n_s_ModelicaParser_Expr_addContext;
static PyObject *__pyx_n_s_ModelicaParser_Expr_addContext_2;
static PyObject *__pyx_n_s_ModelicaParser_Expr_addContext_a;
static PyObject *__pyx_n_s_ModelicaParser_Expr_addContext_e;
static PyObject *__pyx_n_s_ModelicaParser_Expr_addContext_e_2;
static PyObject *__pyx_n_s_ModelicaParser_Expr_addContext_e_3;
static PyObject *__pyx_n_s_ModelicaParser_Expr_andContext;
static PyObject *__pyx_n_s_ModelicaParser_Expr_andContext_2;
static PyObject *__pyx_n_s_ModelicaParser_Expr_andContext_a;
static PyObject *__pyx_n_s_ModelicaParser_Expr_andContext_e;
static PyObject *__pyx_n_s_ModelicaParser_Expr_andContext_e_2;
static PyObject *__pyx_n_s_ModelicaParser_Expr_andContext_e_3;
static PyObject *__pyx_n_s_ModelicaParser_Expr_expContext;
static PyObject *__pyx_n_s_ModelicaParser_Expr_expContext_2;
static PyObject *__pyx_n_s_ModelicaParser_Expr_expContext_a;
static PyObject *__pyx_n_s_ModelicaParser_Expr_expContext_e;
static PyObject *__pyx_n_s_ModelicaParser_Expr_expContext_e_2;
static PyObject *__pyx_n_s_ModelicaParser_Expr_expContext_p;
static PyObject *__pyx_n_s_ModelicaParser_Expr_mulContext;
static PyObject *__pyx_n_s_ModelicaParser_Expr_mulContext_2;
static PyObject *__pyx_n_s_ModelicaParser_Expr_mulContext_a;
static PyObject *__pyx_n_s_ModelicaParser_Expr_mulContext_e;
static PyObject *__pyx_n_s_ModelicaParser_Expr_mulContext_e_2;
static PyObject *__pyx_n_s_ModelicaParser_Expr_mulContext_e_3;
static PyObject *__pyx_n_s_ModelicaParser_Expr_negContext;
static PyObject *__pyx_n_s_ModelicaParser_Expr_negContext_2;
static PyObject *__pyx_n_s_ModelicaParser_Expr_negContext_a;
static PyObject *__pyx_n_s_ModelicaParser_Expr_negContext_e;
static PyObject *__pyx_n_s_ModelicaParser_Expr_negContext_e_2;
static PyObject *__pyx_n_s_ModelicaParser_Expr_negContext_e_3;
static PyObject *__pyx_n_s_ModelicaParser_Expr_notContext;
static PyObject *__pyx_n_s_ModelicaParser_Expr_notContext_2;
static PyObject *__pyx_n_s_ModelicaParser_Expr_notContext_a;
static PyObject *__pyx_n_s_ModelicaParser_Expr_notContext_e;
static PyObject *__pyx_n_s_ModelicaParser_Expr_notContext_e_2;
static PyObject *__pyx_n_s_ModelicaParser_Expr_notContext_e_3;
static PyObject *__pyx_n_s_ModelicaParser_Expr_orContext;
static PyObject *__pyx_n_s_ModelicaParser_Expr_orContext_2;
static PyObject *__pyx_n_s_ModelicaParser_Expr_orContext_ac;
static PyObject *__pyx_n_s_ModelicaParser_Expr_orContext_en;
static PyObject *__pyx_n_s_ModelicaParser_Expr_orContext_ex;
static PyObject *__pyx_n_s_ModelicaParser_Expr_orContext_ex_2;
static PyObject *__pyx_n_s_ModelicaParser_Expr_primaryConte;
static PyObject *__pyx_n_s_ModelicaParser_Expr_primaryConte_2;
static PyObject *__pyx_n_s_ModelicaParser_Expr_primaryConte_3;
static PyObject *__pyx_n_s_ModelicaParser_Expr_primaryConte_4;
static PyObject *__pyx_n_s_ModelicaParser_Expr_primaryConte_5;
static PyObject *__pyx_n_s_ModelicaParser_Expr_primaryConte_6;
static PyObject *__pyx_n_s_ModelicaParser_Expr_relContext;
static PyObject *__pyx_n_s_ModelicaParser_Expr_relContext_2;
static PyObject *__pyx_n_s_ModelicaParser_Expr_relContext_a;
static PyObject *__pyx_n_s_ModelicaParser_Expr_relContext_e;
static PyObject *__pyx_n_s_ModelicaParser_Expr_relContext_e_2;
static PyObject *__pyx_n_s_ModelicaParser_Expr_relContext_e_3;
static PyObject *__pyx_n_s_ModelicaParser_ExpressionContext;
static PyObject *__pyx_n_s_ModelicaParser_ExpressionContext_2;
static PyObject *__pyx_n_s_ModelicaParser_ExpressionContext_3;
static PyObject *__pyx_n_s_ModelicaParser_ExpressionContext_4;
static PyObject *__pyx_n_s_ModelicaParser_Expression_ifCont;
static PyObject *__pyx_n_s_ModelicaParser_Expression_ifCont_2;
static PyObject *__pyx_n_s_ModelicaParser_Expression_ifCont_3;
static PyObject *__pyx_n_s_ModelicaParser_Expression_ifCont_4;
static PyObject *__pyx_n_s_ModelicaParser_Expression_ifCont_5;
static PyObject *__pyx_n_s_ModelicaParser_Expression_ifCont_6;
static PyObject *__pyx_n_s_ModelicaParser_Expression_listCo;
static PyObject *__pyx_n_s_ModelicaParser_Expression_listCo_2;
static PyObject *__pyx_n_s_ModelicaParser_Expression_listCo_3;
static PyObject *__pyx_n_s_ModelicaParser_Expression_listCo_4;
static PyObject *__pyx_n_s_ModelicaParser_Expression_listCo_5;
static PyObject *__pyx_n_s_ModelicaParser_Expression_listCo_6;
static PyObject *__pyx_n_s_ModelicaParser_Expression_listCo_7;
static PyObject *__pyx_n_s_ModelicaParser_Expression_simple;
static PyObject *__pyx_n_s_ModelicaParser_Expression_simple_2;
static PyObject *__pyx_n_s_ModelicaParser_Expression_simple_3;
static PyObject *__pyx_n_s_ModelicaParser_Expression_simple_4;
static PyObject *__pyx_n_s_ModelicaParser_Expression_simple_5;
static PyObject *__pyx_n_s_ModelicaParser_Expression_simple_6;
static PyObject *__pyx_n_s_ModelicaParser_Extends_clauseCon;
static PyObject *__pyx_n_s_ModelicaParser_Extends_clauseCon_2;
static PyObject *__pyx_n_s_ModelicaParser_Extends_clauseCon_3;
static PyObject *__pyx_n_s_ModelicaParser_Extends_clauseCon_4;
static PyObject *__pyx_n_s_ModelicaParser_Extends_clauseCon_5;
static PyObject *__pyx_n_s_ModelicaParser_Extends_clauseCon_6;
static PyObject *__pyx_n_s_ModelicaParser_Extends_clauseCon_7;
static PyObject *__pyx_n_s_ModelicaParser_Extends_clauseCon_8;
static PyObject *__pyx_n_s_ModelicaParser_Extends_clauseCon_9;
static PyObject *__pyx_n_s_ModelicaParser_External_function;
static PyObject *__pyx_n_s_ModelicaParser_External_function_2;
static PyObject *__pyx_n_s_ModelicaParser_External_function_3;
static PyObject *__pyx_n_s_ModelicaParser_External_function_4;
static PyObject *__pyx_n_s_ModelicaParser_External_function_5;
static PyObject *__pyx_n_s_ModelicaParser_External_function_6;
static PyObject *__pyx_n_s_ModelicaParser_External_function_7;
static PyObject *__pyx_n_s_ModelicaParser_External_function_8;
static PyObject *__pyx_n_s_ModelicaParser_External_function_9;
static PyObject *__pyx_n_s_ModelicaParser_For_equationConte;
static PyObject *__pyx_n_s_ModelicaParser_For_equationConte_2;
static PyObject *__pyx_n_s_ModelicaParser_For_equationConte_3;
static PyObject *__pyx_n_s_ModelicaParser_For_equationConte_4;
static PyObject *__pyx_n_s_ModelicaParser_For_equationConte_5;
static PyObject *__pyx_n_s_ModelicaParser_For_equationConte_6;
static PyObject *__pyx_n_s_ModelicaParser_For_equationConte_7;
static PyObject *__pyx_n_s_ModelicaParser_For_equationConte_8;
static PyObject *__pyx_n_s_ModelicaParser_For_indexContext;
static PyObject *__pyx_n_s_ModelicaParser_For_indexContext_2;
static PyObject *__pyx_n_s_ModelicaParser_For_indexContext_3;
static PyObject *__pyx_n_s_ModelicaParser_For_indexContext_4;
static PyObject *__pyx_n_s_ModelicaParser_For_indexContext_5;
static PyObject *__pyx_n_s_ModelicaParser_For_indexContext_6;
static PyObject *__pyx_n_s_ModelicaParser_For_indexContext_7;
static PyObject *__pyx_n_s_ModelicaParser_For_indexContext_8;
static PyObject *__pyx_n_s_ModelicaParser_For_indicesContex;
static PyObject *__pyx_n_s_ModelicaParser_For_indicesContex_2;
static PyObject *__pyx_n_s_ModelicaParser_For_indicesContex_3;
static PyObject *__pyx_n_s_ModelicaParser_For_indicesContex_4;
static PyObject *__pyx_n_s_ModelicaParser_For_indicesContex_5;
static PyObject *__pyx_n_s_ModelicaParser_For_indicesContex_6;
static PyObject *__pyx_n_s_ModelicaParser_For_indicesContex_7;
static PyObject *__pyx_n_s_ModelicaParser_For_statementCont;
static PyObject *__pyx_n_s_ModelicaParser_For_statementCont_2;
static PyObject *__pyx_n_s_ModelicaParser_For_statementCont_3;
static PyObject *__pyx_n_s_ModelicaParser_For_statementCont_4;
static PyObject *__pyx_n_s_ModelicaParser_For_statementCont_5;
static PyObject *__pyx_n_s_ModelicaParser_For_statementCont_6;
static PyObject *__pyx_n_s_ModelicaParser_For_statementCont_7;
static PyObject *__pyx_n_s_ModelicaParser_For_statementCont_8;
static PyObject *__pyx_n_s_ModelicaParser_Function_argument;
static PyObject *__pyx_n_s_ModelicaParser_Function_argument_10;
static PyObject *__pyx_n_s_ModelicaParser_Function_argument_11;
static PyObject *__pyx_n_s_ModelicaParser_Function_argument_12;
static PyObject *__pyx_n_s_ModelicaParser_Function_argument_13;
static PyObject *__pyx_n_s_ModelicaParser_Function_argument_2;
static PyObject *__pyx_n_s_ModelicaParser_Function_argument_3;
static PyObject *__pyx_n_s_ModelicaParser_Function_argument_4;
static PyObject *__pyx_n_s_ModelicaParser_Function_argument_5;
static PyObject *__pyx_n_s_ModelicaParser_Function_argument_6;
static PyObject *__pyx_n_s_ModelicaParser_Function_argument_7;
static PyObject *__pyx_n_s_ModelicaParser_Function_argument_8;
static PyObject *__pyx_n_s_ModelicaParser_Function_argument_9;
static PyObject *__pyx_n_s_ModelicaParser_Function_call_arg;
static PyObject *__pyx_n_s_ModelicaParser_Function_call_arg_2;
static PyObject *__pyx_n_s_ModelicaParser_Function_call_arg_3;
static PyObject *__pyx_n_s_ModelicaParser_Function_call_arg_4;
static PyObject *__pyx_n_s_ModelicaParser_Function_call_arg_5;
static PyObject *__pyx_n_s_ModelicaParser_Function_call_arg_6;
static PyObject *__pyx_n_s_ModelicaParser_Function_call_arg_7;
static PyObject *__pyx_n_s_ModelicaParser_If_equationContex;
static PyObject *__pyx_n_s_ModelicaParser_If_equationContex_2;
static PyObject *__pyx_n_s_ModelicaParser_If_equationContex_3;
static PyObject *__pyx_n_s_ModelicaParser_If_equationContex_4;
static PyObject *__pyx_n_s_ModelicaParser_If_equationContex_5;
static PyObject *__pyx_n_s_ModelicaParser_If_equationContex_6;
static PyObject *__pyx_n_s_ModelicaParser_If_equationContex_7;
static PyObject *__pyx_n_s_ModelicaParser_If_equationContex_8;
static PyObject *__pyx_n_s_ModelicaParser_If_statementConte;
static PyObject *__pyx_n_s_ModelicaParser_If_statementConte_2;
static PyObject *__pyx_n_s_ModelicaParser_If_statementConte_3;
static PyObject *__pyx_n_s_ModelicaParser_If_statementConte_4;
static PyObject *__pyx_n_s_ModelicaParser_If_statementConte_5;
static PyObject *__pyx_n_s_ModelicaParser_If_statementConte_6;
static PyObject *__pyx_n_s_ModelicaParser_If_statementConte_7;
static PyObject *__pyx_n_s_ModelicaParser_If_statementConte_8;
static PyObject *__pyx_n_s_ModelicaParser_Import_clauseCont;
static PyObject *__pyx_n_s_ModelicaParser_Import_clauseCont_10;
static PyObject *__pyx_n_s_ModelicaParser_Import_clauseCont_2;
static PyObject *__pyx_n_s_ModelicaParser_Import_clauseCont_3;
static PyObject *__pyx_n_s_ModelicaParser_Import_clauseCont_4;
static PyObject *__pyx_n_s_ModelicaParser_Import_clauseCont_5;
static PyObject *__pyx_n_s_ModelicaParser_Import_clauseCont_6;
static PyObject *__pyx_n_s_ModelicaParser_Import_clauseCont_7;
static PyObject *__pyx_n_s_ModelicaParser_Import_clauseCont_8;
static PyObject *__pyx_n_s_ModelicaParser_Import_clauseCont_9;
static PyObject *__pyx_n_s_ModelicaParser_Import_listContex;
static PyObject *__pyx_n_s_ModelicaParser_Import_listContex_2;
static PyObject *__pyx_n_s_ModelicaParser_Import_listContex_3;
static PyObject *__pyx_n_s_ModelicaParser_Import_listContex_4;
static PyObject *__pyx_n_s_ModelicaParser_Import_listContex_5;
static PyObject *__pyx_n_s_ModelicaParser_Import_listContex_6;
static PyObject *__pyx_n_s_ModelicaParser_Import_listContex_7;
static PyObject *__pyx_n_s_ModelicaParser_Import_listContex_8;
static PyObject *__pyx_n_s_ModelicaParser_Language_specific;
static PyObject *__pyx_n_s_ModelicaParser_Language_specific_2;
static PyObject *__pyx_n_s_ModelicaParser_Language_specific_3;
static PyObject *__pyx_n_s_ModelicaParser_Language_specific_4;
static PyObject *__pyx_n_s_ModelicaParser_Language_specific_5;
static PyObject *__pyx_n_s_ModelicaParser_Language_specific_6;
static PyObject *__pyx_n_s_ModelicaParser_Language_specific_7;
static PyObject *__pyx_n_s_ModelicaParser_ModificationConte;
static PyObject *__pyx_n_s_ModelicaParser_ModificationConte_2;
static PyObject *__pyx_n_s_ModelicaParser_ModificationConte_3;
static PyObject *__pyx_n_s_ModelicaParser_ModificationConte_4;
static PyObject *__pyx_n_s_ModelicaParser_Modification_assi;
static PyObject *__pyx_n_s_ModelicaParser_Modification_assi_10;
static PyObject *__pyx_n_s_ModelicaParser_Modification_assi_11;
static PyObject *__pyx_n_s_ModelicaParser_Modification_assi_12;
static PyObject *__pyx_n_s_ModelicaParser_Modification_assi_2;
static PyObject *__pyx_n_s_ModelicaParser_Modification_assi_3;
static PyObject *__pyx_n_s_ModelicaParser_Modification_assi_4;
static PyObject *__pyx_n_s_ModelicaParser_Modification_assi_5;
static PyObject *__pyx_n_s_ModelicaParser_Modification_assi_6;
static PyObject *__pyx_n_s_ModelicaParser_Modification_assi_7;
static PyObject *__pyx_n_s_ModelicaParser_Modification_assi_8;
static PyObject *__pyx_n_s_ModelicaParser_Modification_assi_9;
static PyObject *__pyx_n_s_ModelicaParser_Modification_clas;
static PyObject *__pyx_n_s_ModelicaParser_Modification_clas_2;
static PyObject *__pyx_n_s_ModelicaParser_Modification_clas_3;
static PyObject *__pyx_n_s_ModelicaParser_Modification_clas_4;
static PyObject *__pyx_n_s_ModelicaParser_Modification_clas_5;
static PyObject *__pyx_n_s_ModelicaParser_Modification_clas_6;
static PyObject *__pyx_n_s_ModelicaParser_Modification_clas_7;
static PyObject *__pyx_n_s_ModelicaParser_NameContext;
static PyObject *__pyx_n_s_ModelicaParser_NameContext_IDENT;
static PyObject *__pyx_n_s_ModelicaParser_NameContext___ini;
static PyObject *__pyx_n_s_ModelicaParser_NameContext_accep;
static PyObject *__pyx_n_s_ModelicaParser_NameContext_enter;
static PyObject *__pyx_n_s_ModelicaParser_NameContext_exitR;
static PyObject *__pyx_n_s_ModelicaParser_NameContext_getRu;
static PyObject *__pyx_n_s_ModelicaParser_Named_argumentCon;
static PyObject *__pyx_n_s_ModelicaParser_Named_argumentCon_2;
static PyObject *__pyx_n_s_ModelicaParser_Named_argumentCon_3;
static PyObject *__pyx_n_s_ModelicaParser_Named_argumentCon_4;
static PyObject *__pyx_n_s_ModelicaParser_Named_argumentCon_5;
static PyObject *__pyx_n_s_ModelicaParser_Named_argumentCon_6;
static PyObject *__pyx_n_s_ModelicaParser_Named_argumentCon_7;
static PyObject *__pyx_n_s_ModelicaParser_Named_argumentCon_8;
static PyObject *__pyx_n_s_ModelicaParser_Named_argumentsCo;
static PyObject *__pyx_n_s_ModelicaParser_Named_argumentsCo_2;
static PyObject *__pyx_n_s_ModelicaParser_Named_argumentsCo_3;
static PyObject *__pyx_n_s_ModelicaParser_Named_argumentsCo_4;
static PyObject *__pyx_n_s_ModelicaParser_Named_argumentsCo_5;
static PyObject *__pyx_n_s_ModelicaParser_Named_argumentsCo_6;
static PyObject *__pyx_n_s_ModelicaParser_Named_argumentsCo_7;
static PyObject *__pyx_n_s_ModelicaParser_Output_expression;
static PyObject *__pyx_n_s_ModelicaParser_Output_expression_2;
static PyObject *__pyx_n_s_ModelicaParser_Output_expression_3;
static PyObject *__pyx_n_s_ModelicaParser_Output_expression_4;
static PyObject *__pyx_n_s_ModelicaParser_Output_expression_5;
static PyObject *__pyx_n_s_ModelicaParser_Output_expression_6;
static PyObject *__pyx_n_s_ModelicaParser_Output_expression_7;
static PyObject *__pyx_n_s_ModelicaParser_PrimaryContext;
static PyObject *__pyx_n_s_ModelicaParser_PrimaryContext_2;
static PyObject *__pyx_n_s_ModelicaParser_PrimaryContext_co;
static PyObject *__pyx_n_s_ModelicaParser_PrimaryContext_ge;
static PyObject *__pyx_n_s_ModelicaParser_Primary_component;
static PyObject *__pyx_n_s_ModelicaParser_Primary_component_2;
static PyObject *__pyx_n_s_ModelicaParser_Primary_component_3;
static PyObject *__pyx_n_s_ModelicaParser_Primary_component_4;
static PyObject *__pyx_n_s_ModelicaParser_Primary_component_5;
static PyObject *__pyx_n_s_ModelicaParser_Primary_component_6;
static PyObject *__pyx_n_s_ModelicaParser_Primary_derivativ;
static PyObject *__pyx_n_s_ModelicaParser_Primary_derivativ_2;
static PyObject *__pyx_n_s_ModelicaParser_Primary_derivativ_3;
static PyObject *__pyx_n_s_ModelicaParser_Primary_derivativ_4;
static PyObject *__pyx_n_s_ModelicaParser_Primary_derivativ_5;
static PyObject *__pyx_n_s_ModelicaParser_Primary_derivativ_6;
static PyObject *__pyx_n_s_ModelicaParser_Primary_endContex;
static PyObject *__pyx_n_s_ModelicaParser_Primary_endContex_2;
static PyObject *__pyx_n_s_ModelicaParser_Primary_endContex_3;
static PyObject *__pyx_n_s_ModelicaParser_Primary_endContex_4;
static PyObject *__pyx_n_s_ModelicaParser_Primary_endContex_5;
static PyObject *__pyx_n_s_ModelicaParser_Primary_expressio;
static PyObject *__pyx_n_s_ModelicaParser_Primary_expressio_2;
static PyObject *__pyx_n_s_ModelicaParser_Primary_expressio_3;
static PyObject *__pyx_n_s_ModelicaParser_Primary_expressio_4;
static PyObject *__pyx_n_s_ModelicaParser_Primary_expressio_5;
static PyObject *__pyx_n_s_ModelicaParser_Primary_expressio_6;
static PyObject *__pyx_n_s_ModelicaParser_Primary_falseCont;
static PyObject *__pyx_n_s_ModelicaParser_Primary_falseCont_2;
static PyObject *__pyx_n_s_ModelicaParser_Primary_falseCont_3;
static PyObject *__pyx_n_s_ModelicaParser_Primary_falseCont_4;
static PyObject *__pyx_n_s_ModelicaParser_Primary_falseCont_5;
static PyObject *__pyx_n_s_ModelicaParser_Primary_function;
static PyObject *__pyx_n_s_ModelicaParser_Primary_functionC;
static PyObject *__pyx_n_s_ModelicaParser_Primary_functionC_2;
static PyObject *__pyx_n_s_ModelicaParser_Primary_functionC_3;
static PyObject *__pyx_n_s_ModelicaParser_Primary_functionC_4;
static PyObject *__pyx_n_s_ModelicaParser_Primary_functionC_5;
static PyObject *__pyx_n_s_ModelicaParser_Primary_functionC_6;
static PyObject *__pyx_n_s_ModelicaParser_Primary_functionC_7;
static PyObject *__pyx_n_s_ModelicaParser_Primary_function_2;
static PyObject *__pyx_n_s_ModelicaParser_Primary_function_3;
static PyObject *__pyx_n_s_ModelicaParser_Primary_function_4;
static PyObject *__pyx_n_s_ModelicaParser_Primary_function_5;
static PyObject *__pyx_n_s_ModelicaParser_Primary_function_6;
static PyObject *__pyx_n_s_ModelicaParser_Primary_initialCo;
static PyObject *__pyx_n_s_ModelicaParser_Primary_initialCo_2;
static PyObject *__pyx_n_s_ModelicaParser_Primary_initialCo_3;
static PyObject *__pyx_n_s_ModelicaParser_Primary_initialCo_4;
static PyObject *__pyx_n_s_ModelicaParser_Primary_initialCo_5;
static PyObject *__pyx_n_s_ModelicaParser_Primary_initialCo_6;
static PyObject *__pyx_n_s_ModelicaParser_Primary_output_ex;
static PyObject *__pyx_n_s_ModelicaParser_Primary_output_ex_2;
static PyObject *__pyx_n_s_ModelicaParser_Primary_output_ex_3;
static PyObject *__pyx_n_s_ModelicaParser_Primary_output_ex_4;
static PyObject *__pyx_n_s_ModelicaParser_Primary_output_ex_5;
static PyObject *__pyx_n_s_ModelicaParser_Primary_output_ex_6;
static PyObject *__pyx_n_s_ModelicaParser_Primary_stringCon;
static PyObject *__pyx_n_s_ModelicaParser_Primary_stringCon_2;
static PyObject *__pyx_n_s_ModelicaParser_Primary_stringCon_3;
static PyObject *__pyx_n_s_ModelicaParser_Primary_stringCon_4;
static PyObject *__pyx_n_s_ModelicaParser_Primary_stringCon_5;
static PyObject *__pyx_n_s_ModelicaParser_Primary_stringCon_6;
static PyObject *__pyx_n_s_ModelicaParser_Primary_trueConte;
static PyObject *__pyx_n_s_ModelicaParser_Primary_trueConte_2;
static PyObject *__pyx_n_s_ModelicaParser_Primary_trueConte_3;
static PyObject *__pyx_n_s_ModelicaParser_Primary_trueConte_4;
static PyObject *__pyx_n_s_ModelicaParser_Primary_trueConte_5;
static PyObject *__pyx_n_s_ModelicaParser_Primary_unsigned;
static PyObject *__pyx_n_s_ModelicaParser_Primary_unsigned_2;
static PyObject *__pyx_n_s_ModelicaParser_Primary_unsigned_3;
static PyObject *__pyx_n_s_ModelicaParser_Primary_unsigned_4;
static PyObject *__pyx_n_s_ModelicaParser_Primary_unsigned_5;
static PyObject *__pyx_n_s_ModelicaParser_Primary_unsigned_6;
static PyObject *__pyx_n_s_ModelicaParser_Regular_elementCo;
static PyObject *__pyx_n_s_ModelicaParser_Regular_elementCo_10;
static PyObject *__pyx_n_s_ModelicaParser_Regular_elementCo_11;
static PyObject *__pyx_n_s_ModelicaParser_Regular_elementCo_12;
static PyObject *__pyx_n_s_ModelicaParser_Regular_elementCo_2;
static PyObject *__pyx_n_s_ModelicaParser_Regular_elementCo_3;
static PyObject *__pyx_n_s_ModelicaParser_Regular_elementCo_4;
static PyObject *__pyx_n_s_ModelicaParser_Regular_elementCo_5;
static PyObject *__pyx_n_s_ModelicaParser_Regular_elementCo_6;
static PyObject *__pyx_n_s_ModelicaParser_Regular_elementCo_7;
static PyObject *__pyx_n_s_ModelicaParser_Regular_elementCo_8;
static PyObject *__pyx_n_s_ModelicaParser_Regular_elementCo_9;
static PyObject *__pyx_n_s_ModelicaParser_Replaceable_eleme;
static PyObject *__pyx_n_s_ModelicaParser_Replaceable_eleme_10;
static PyObject *__pyx_n_s_ModelicaParser_Replaceable_eleme_11;
static PyObject *__pyx_n_s_ModelicaParser_Replaceable_eleme_12;
static PyObject *__pyx_n_s_ModelicaParser_Replaceable_eleme_13;
static PyObject *__pyx_n_s_ModelicaParser_Replaceable_eleme_14;
static PyObject *__pyx_n_s_ModelicaParser_Replaceable_eleme_2;
static PyObject *__pyx_n_s_ModelicaParser_Replaceable_eleme_3;
static PyObject *__pyx_n_s_ModelicaParser_Replaceable_eleme_4;
static PyObject *__pyx_n_s_ModelicaParser_Replaceable_eleme_5;
static PyObject *__pyx_n_s_ModelicaParser_Replaceable_eleme_6;
static PyObject *__pyx_n_s_ModelicaParser_Replaceable_eleme_7;
static PyObject *__pyx_n_s_ModelicaParser_Replaceable_eleme_8;
static PyObject *__pyx_n_s_ModelicaParser_Replaceable_eleme_9;
static PyObject *__pyx_n_s_ModelicaParser_Short_class_defin;
static PyObject *__pyx_n_s_ModelicaParser_Short_class_defin_10;
static PyObject *__pyx_n_s_ModelicaParser_Short_class_defin_11;
static PyObject *__pyx_n_s_ModelicaParser_Short_class_defin_12;
static PyObject *__pyx_n_s_ModelicaParser_Short_class_defin_13;
static PyObject *__pyx_n_s_ModelicaParser_Short_class_defin_14;
static PyObject *__pyx_n_s_ModelicaParser_Short_class_defin_2;
static PyObject *__pyx_n_s_ModelicaParser_Short_class_defin_3;
static PyObject *__pyx_n_s_ModelicaParser_Short_class_defin_4;
static PyObject *__pyx_n_s_ModelicaParser_Short_class_defin_5;
static PyObject *__pyx_n_s_ModelicaParser_Short_class_defin_6;
static PyObject *__pyx_n_s_ModelicaParser_Short_class_defin_7;
static PyObject *__pyx_n_s_ModelicaParser_Short_class_defin_8;
static PyObject *__pyx_n_s_ModelicaParser_Short_class_defin_9;
static PyObject *__pyx_n_s_ModelicaParser_Simple_expression;
static PyObject *__pyx_n_s_ModelicaParser_Simple_expression_2;
static PyObject *__pyx_n_s_ModelicaParser_Simple_expression_3;
static PyObject *__pyx_n_s_ModelicaParser_Simple_expression_4;
static PyObject *__pyx_n_s_ModelicaParser_Simple_expression_5;
static PyObject *__pyx_n_s_ModelicaParser_Simple_expression_6;
static PyObject *__pyx_n_s_ModelicaParser_Simple_expression_7;
static PyObject *__pyx_n_s_ModelicaParser_StatementContext;
static PyObject *__pyx_n_s_ModelicaParser_StatementContext_2;
static PyObject *__pyx_n_s_ModelicaParser_StatementContext_3;
static PyObject *__pyx_n_s_ModelicaParser_StatementContext_4;
static PyObject *__pyx_n_s_ModelicaParser_StatementContext_5;
static PyObject *__pyx_n_s_ModelicaParser_StatementContext_6;
static PyObject *__pyx_n_s_ModelicaParser_StatementContext_7;
static PyObject *__pyx_n_s_ModelicaParser_StatementContext_8;
static PyObject *__pyx_n_s_ModelicaParser_Statement_breakCo;
static PyObject *__pyx_n_s_ModelicaParser_Statement_breakCo_2;
static PyObject *__pyx_n_s_ModelicaParser_Statement_breakCo_3;
static PyObject *__pyx_n_s_ModelicaParser_Statement_breakCo_4;
static PyObject *__pyx_n_s_ModelicaParser_Statement_breakCo_5;
static PyObject *__pyx_n_s_ModelicaParser_Statement_compone;
static PyObject *__pyx_n_s_ModelicaParser_Statement_compone_10;
static PyObject *__pyx_n_s_ModelicaParser_Statement_compone_11;
static PyObject *__pyx_n_s_ModelicaParser_Statement_compone_12;
static PyObject *__pyx_n_s_ModelicaParser_Statement_compone_13;
static PyObject *__pyx_n_s_ModelicaParser_Statement_compone_14;
static PyObject *__pyx_n_s_ModelicaParser_Statement_compone_15;
static PyObject *__pyx_n_s_ModelicaParser_Statement_compone_2;
static PyObject *__pyx_n_s_ModelicaParser_Statement_compone_3;
static PyObject *__pyx_n_s_ModelicaParser_Statement_compone_4;
static PyObject *__pyx_n_s_ModelicaParser_Statement_compone_5;
static PyObject *__pyx_n_s_ModelicaParser_Statement_compone_6;
static PyObject *__pyx_n_s_ModelicaParser_Statement_compone_7;
static PyObject *__pyx_n_s_ModelicaParser_Statement_compone_8;
static PyObject *__pyx_n_s_ModelicaParser_Statement_compone_9;
static PyObject *__pyx_n_s_ModelicaParser_Statement_forCont;
static PyObject *__pyx_n_s_ModelicaParser_Statement_forCont_2;
static PyObject *__pyx_n_s_ModelicaParser_Statement_forCont_3;
static PyObject *__pyx_n_s_ModelicaParser_Statement_forCont_4;
static PyObject *__pyx_n_s_ModelicaParser_Statement_forCont_5;
static PyObject *__pyx_n_s_ModelicaParser_Statement_forCont_6;
static PyObject *__pyx_n_s_ModelicaParser_Statement_ifConte;
static PyObject *__pyx_n_s_ModelicaParser_Statement_ifConte_2;
static PyObject *__pyx_n_s_ModelicaParser_Statement_ifConte_3;
static PyObject *__pyx_n_s_ModelicaParser_Statement_ifConte_4;
static PyObject *__pyx_n_s_ModelicaParser_Statement_ifConte_5;
static PyObject *__pyx_n_s_ModelicaParser_Statement_ifConte_6;
static PyObject *__pyx_n_s_ModelicaParser_Statement_options;
static PyObject *__pyx_n_s_ModelicaParser_Statement_options_2;
static PyObject *__pyx_n_s_ModelicaParser_Statement_options_3;
static PyObject *__pyx_n_s_ModelicaParser_Statement_options_4;
static PyObject *__pyx_n_s_ModelicaParser_Statement_returnC;
static PyObject *__pyx_n_s_ModelicaParser_Statement_returnC_2;
static PyObject *__pyx_n_s_ModelicaParser_Statement_returnC_3;
static PyObject *__pyx_n_s_ModelicaParser_Statement_returnC_4;
static PyObject *__pyx_n_s_ModelicaParser_Statement_returnC_5;
static PyObject *__pyx_n_s_ModelicaParser_Statement_whenCon;
static PyObject *__pyx_n_s_ModelicaParser_Statement_whenCon_2;
static PyObject *__pyx_n_s_ModelicaParser_Statement_whenCon_3;
static PyObject *__pyx_n_s_ModelicaParser_Statement_whenCon_4;
static PyObject *__pyx_n_s_ModelicaParser_Statement_whenCon_5;
static PyObject *__pyx_n_s_ModelicaParser_Statement_whenCon_6;
static PyObject *__pyx_n_s_ModelicaParser_Statement_whileCo;
static PyObject *__pyx_n_s_ModelicaParser_Statement_whileCo_2;
static PyObject *__pyx_n_s_ModelicaParser_Statement_whileCo_3;
static PyObject *__pyx_n_s_ModelicaParser_Statement_whileCo_4;
static PyObject *__pyx_n_s_ModelicaParser_Statement_whileCo_5;
static PyObject *__pyx_n_s_ModelicaParser_Statement_whileCo_6;
static PyObject *__pyx_n_s_ModelicaParser_Stored_definition;
static PyObject *__pyx_n_s_ModelicaParser_Stored_definition_10;
static PyObject *__pyx_n_s_ModelicaParser_Stored_definition_11;
static PyObject *__pyx_n_s_ModelicaParser_Stored_definition_12;
static PyObject *__pyx_n_s_ModelicaParser_Stored_definition_13;
static PyObject *__pyx_n_s_ModelicaParser_Stored_definition_14;
static PyObject *__pyx_n_s_ModelicaParser_Stored_definition_15;
static PyObject *__pyx_n_s_ModelicaParser_Stored_definition_16;
static PyObject *__pyx_n_s_ModelicaParser_Stored_definition_17;
static PyObject *__pyx_n_s_ModelicaParser_Stored_definition_2;
static PyObject *__pyx_n_s_ModelicaParser_Stored_definition_3;
static PyObject *__pyx_n_s_ModelicaParser_Stored_definition_4;
static PyObject *__pyx_n_s_ModelicaParser_Stored_definition_5;
static PyObject *__pyx_n_s_ModelicaParser_Stored_definition_6;
static PyObject *__pyx_n_s_ModelicaParser_Stored_definition_7;
static PyObject *__pyx_n_s_ModelicaParser_Stored_definition_8;
static PyObject *__pyx_n_s_ModelicaParser_Stored_definition_9;
static PyObject *__pyx_n_s_ModelicaParser_String_commentCon;
static PyObject *__pyx_n_s_ModelicaParser_String_commentCon_2;
static PyObject *__pyx_n_s_ModelicaParser_String_commentCon_3;
static PyObject *__pyx_n_s_ModelicaParser_String_commentCon_4;
static PyObject *__pyx_n_s_ModelicaParser_String_commentCon_5;
static PyObject *__pyx_n_s_ModelicaParser_String_commentCon_6;
static PyObject *__pyx_n_s_ModelicaParser_String_commentCon_7;
static PyObject *__pyx_n_s_ModelicaParser_SubscriptContext;
static PyObject *__pyx_n_s_ModelicaParser_SubscriptContext_2;
static PyObject *__pyx_n_s_ModelicaParser_SubscriptContext_3;
static PyObject *__pyx_n_s_ModelicaParser_SubscriptContext_4;
static PyObject *__pyx_n_s_ModelicaParser_SubscriptContext_5;
static PyObject *__pyx_n_s_ModelicaParser_SubscriptContext_6;
static PyObject *__pyx_n_s_ModelicaParser_SubscriptContext_7;
static PyObject *__pyx_n_s_ModelicaParser_Type_prefixContex;
static PyObject *__pyx_n_s_ModelicaParser_Type_prefixContex_2;
static PyObject *__pyx_n_s_ModelicaParser_Type_prefixContex_3;
static PyObject *__pyx_n_s_ModelicaParser_Type_prefixContex_4;
static PyObject *__pyx_n_s_ModelicaParser_Type_prefixContex_5;
static PyObject *__pyx_n_s_ModelicaParser_Type_prefixContex_6;
static PyObject *__pyx_n_s_ModelicaParser_Type_specifierCon;
static PyObject *__pyx_n_s_ModelicaParser_Type_specifierCon_2;
static PyObject *__pyx_n_s_ModelicaParser_Type_specifierCon_3;
static PyObject *__pyx_n_s_ModelicaParser_Type_specifierCon_4;
static PyObject *__pyx_n_s_ModelicaParser_Type_specifierCon_5;
static PyObject *__pyx_n_s_ModelicaParser_Type_specifierCon_6;
static PyObject *__pyx_n_s_ModelicaParser_Type_specifierCon_7;
static PyObject *__pyx_n_s_ModelicaParser_When_equationCont;
static PyObject *__pyx_n_s_ModelicaParser_When_equationCont_2;
static PyObject *__pyx_n_s_ModelicaParser_When_equationCont_3;
static PyObject *__pyx_n_s_ModelicaParser_When_equationCont_4;
static PyObject *__pyx_n_s_ModelicaParser_When_equationCont_5;
static PyObject *__pyx_n_s_ModelicaParser_When_equationCont_6;
static PyObject *__pyx_n_s_ModelicaParser_When_equationCont_7;
static PyObject *__pyx_n_s_ModelicaParser_When_equationCont_8;
static PyObject *__pyx_n_s_ModelicaParser_When_statementCon;
static PyObject *__pyx_n_s_ModelicaParser_When_statementCon_2;
static PyObject *__pyx_n_s_ModelicaParser_When_statementCon_3;
static PyObject *__pyx_n_s_ModelicaParser_When_statementCon_4;
static PyObject *__pyx_n_s_ModelicaParser_When_statementCon_5;
static PyObject *__pyx_n_s_ModelicaParser_When_statementCon_6;
static PyObject *__pyx_n_s_ModelicaParser_When_statementCon_7;
static PyObject *__pyx_n_s_ModelicaParser_When_statementCon_8;
static PyObject *__pyx_n_s_ModelicaParser_While_statementCo;
static PyObject *__pyx_n_s_ModelicaParser_While_statementCo_2;
static PyObject *__pyx_n_s_ModelicaParser_While_statementCo_3;
static PyObject *__pyx_n_s_ModelicaParser_While_statementCo_4;
static PyObject *__pyx_n_s_ModelicaParser_While_statementCo_5;
static PyObject *__pyx_n_s_ModelicaParser_While_statementCo_6;
static PyObject *__pyx_n_s_ModelicaParser_While_statementCo_7;
static PyObject *__pyx_n_s_ModelicaParser_While_statementCo_8;
static PyObject *__pyx_n_s_ModelicaParser___init;
static PyObject *__pyx_n_s_ModelicaParser_algorithm_section;
static PyObject *__pyx_n_s_ModelicaParser_annotation;
static PyObject *__pyx_n_s_ModelicaParser_argument;
static PyObject *__pyx_n_s_ModelicaParser_argument_list;
static PyObject *__pyx_n_s_ModelicaParser_array_subscripts;
static PyObject *__pyx_n_s_ModelicaParser_base_prefix;
static PyObject *__pyx_n_s_ModelicaParser_class_definition;
static PyObject *__pyx_n_s_ModelicaParser_class_modificatio;
static PyObject *__pyx_n_s_ModelicaParser_class_prefixes;
static PyObject *__pyx_n_s_ModelicaParser_class_specifier;
static PyObject *__pyx_n_s_ModelicaParser_class_type;
static PyObject *__pyx_n_s_ModelicaParser_comment;
static PyObject *__pyx_n_s_ModelicaParser_component_clause;
static PyObject *__pyx_n_s_ModelicaParser_component_clause1;
static PyObject *__pyx_n_s_ModelicaParser_component_declara;
static PyObject *__pyx_n_s_ModelicaParser_component_declara_2;
static PyObject *__pyx_n_s_ModelicaParser_component_list;
static PyObject *__pyx_n_s_ModelicaParser_component_referen;
static PyObject *__pyx_n_s_ModelicaParser_component_referen_2;
static PyObject *__pyx_n_s_ModelicaParser_composition;
static PyObject *__pyx_n_s_ModelicaParser_condition_attribu;
static PyObject *__pyx_n_s_ModelicaParser_connect_clause;
static PyObject *__pyx_n_s_ModelicaParser_constraining_clau;
static PyObject *__pyx_n_s_ModelicaParser_declaration;
static PyObject *__pyx_n_s_ModelicaParser_element;
static PyObject *__pyx_n_s_ModelicaParser_element_list;
static PyObject *__pyx_n_s_ModelicaParser_element_modificat;
static PyObject *__pyx_n_s_ModelicaParser_element_modificat_2;
static PyObject *__pyx_n_s_ModelicaParser_element_redeclara;
static PyObject *__pyx_n_s_ModelicaParser_element_replaceab;
static PyObject *__pyx_n_s_ModelicaParser_enum_list;
static PyObject *__pyx_n_s_ModelicaParser_enumeration_liter;
static PyObject *__pyx_n_s_ModelicaParser_equation;
static PyObject *__pyx_n_s_ModelicaParser_equation_options;
static PyObject *__pyx_n_s_ModelicaParser_equation_section;
static PyObject *__pyx_n_s_ModelicaParser_expr;
static PyObject *__pyx_n_s_ModelicaParser_expr_sempred;
static PyObject *__pyx_n_s_ModelicaParser_expression;
static PyObject *__pyx_n_s_ModelicaParser_expression_list;
static PyObject *__pyx_n_s_ModelicaParser_extends_clause;
static PyObject *__pyx_n_s_ModelicaParser_external_function;
static PyObject *__pyx_n_s_ModelicaParser_for_equation;
static PyObject *__pyx_n_s_ModelicaParser_for_index;
static PyObject *__pyx_n_s_ModelicaParser_for_indices;
static PyObject *__pyx_n_s_ModelicaParser_for_statement;
static PyObject *__pyx_n_s_ModelicaParser_function_argument;
static PyObject *__pyx_n_s_ModelicaParser_function_argument_2;
static PyObject *__pyx_n_s_ModelicaParser_function_call_arg;
static PyObject *__pyx_n_s_ModelicaParser_if_equation;
static PyObject *__pyx_n_s_ModelicaParser_if_statement;
static PyObject *__pyx_n_s_ModelicaParser_import_clause;
static PyObject *__pyx_n_s_ModelicaParser_import_list;
static PyObject *__pyx_n_s_ModelicaParser_language_specific;
static PyObject *__pyx_n_s_ModelicaParser_modification;
static PyObject *__pyx_n_s_ModelicaParser_name;
static PyObject *__pyx_n_s_ModelicaParser_named_argument;
static PyObject *__pyx_n_s_ModelicaParser_named_arguments;
static PyObject *__pyx_n_s_ModelicaParser_output_expression;
static PyObject *__pyx_n_s_ModelicaParser_primary;
static PyObject *__pyx_n_s_ModelicaParser_regular_element;
static PyObject *__pyx_n_s_ModelicaParser_replaceable_eleme;
static PyObject *__pyx_n_s_ModelicaParser_sempred;
static PyObject *__pyx_n_s_ModelicaParser_short_class_defin;
static PyObject *__pyx_n_s_ModelicaParser_simple_expression;
static PyObject *__pyx_n_s_ModelicaParser_statement;
static PyObject *__pyx_n_s_ModelicaParser_statement_options;
static PyObject *__pyx_n_s_ModelicaParser_stored_definition;
static PyObject *__pyx_n_s_ModelicaParser_stored_definition_2;
static PyObject *__pyx_n_s_ModelicaParser_string_comment;
static PyObject *__pyx_n_s_ModelicaParser_subscript;
static PyObject *__pyx_n_s_ModelicaParser_type_prefix;
static PyObject *__pyx_n_s_ModelicaParser_type_specifier;
static PyObject *__pyx_n_s_ModelicaParser_when_equation;
static PyObject *__pyx_n_s_ModelicaParser_when_statement;
static PyObject *__pyx_n_s_ModelicaParser_while_statement;
static PyObject *__pyx_kp_u_Modelica_g4;
static PyObject *__pyx_n_s_ModificationContext;
static PyObject *__pyx_n_s_Modification_assignment2Context;
static PyObject *__pyx_n_s_Modification_assignmentContext;
static PyObject *__pyx_n_s_Modification_classContext;
static PyObject *__pyx_kp_u_N;
static PyObject *__pyx_kp_u_N_2;
static PyObject *__pyx_kp_u_N_G;
static PyObject *__pyx_n_s_NameContext;
static PyObject *__pyx_n_s_Named_argumentContext;
static PyObject *__pyx_n_s_Named_argumentsContext;
static PyObject *__pyx_n_s_NoViableAltException;
static PyObject *__pyx_kp_u_No_predicate_with_index;
static PyObject *__pyx_kp_u_O;
static PyObject *__pyx_n_s_OUTER;
static PyObject *__pyx_n_u_OUTER;
static PyObject *__pyx_kp_u_O_2;
static PyObject *__pyx_kp_u_O_3;
static PyObject *__pyx_n_s_Output_expression_listContext;
static PyObject *__pyx_kp_u_P;
static PyObject *__pyx_n_s_PARTIAL;
static PyObject *__pyx_n_u_PARTIAL;
static PyObject *__pyx_kp_u_P_2;
static PyObject *__pyx_n_s_ParseTreeListener;
static PyObject *__pyx_n_s_ParseTreeVisitor;
static PyObject *__pyx_n_s_Parser;
static PyObject *__pyx_n_s_ParserATNSimulator;
static PyObject *__pyx_n_s_ParserRuleContext;
static PyObject *__pyx_n_s_PredictionContextCache;
static PyObject *__pyx_n_s_PrimaryContext;
static PyObject *__pyx_n_s_Primary_component_referenceConte;
static PyObject *__pyx_n_s_Primary_derivativeContext;
static PyObject *__pyx_n_s_Primary_endContext;
static PyObject *__pyx_n_s_Primary_expression_listContext;
static PyObject *__pyx_n_s_Primary_falseContext;
static PyObject *__pyx_n_s_Primary_functionContext;
static PyObject *__pyx_n_s_Primary_function_argumentsContex;
static PyObject *__pyx_n_s_Primary_initialContext;
static PyObject *__pyx_n_s_Primary_output_expression_listCo;
static PyObject *__pyx_n_s_Primary_stringContext;
static PyObject *__pyx_n_s_Primary_trueContext;
static PyObject *__pyx_n_s_Primary_unsigned_numberContext;
static PyObject *__pyx_kp_u_Q;
static PyObject *__pyx_kp_u_Q_2;
static PyObject *__pyx_kp_u_R;
static PyObject *__pyx_n_s_REDECLARE;
static PyObject *__pyx_n_u_REDECLARE;
static PyObject *__pyx_n_s_RULE_algorithm_section;
static PyObject *__pyx_n_s_RULE_annotation;
static PyObject *__pyx_n_s_RULE_argument;
static PyObject *__pyx_n_s_RULE_argument_list;
static PyObject *__pyx_n_s_RULE_array_subscripts;
static PyObject *__pyx_n_s_RULE_base_prefix;
static PyObject *__pyx_n_s_RULE_class_definition;
static PyObject *__pyx_n_s_RULE_class_modification;
static PyObject *__pyx_n_s_RULE_class_prefixes;
static PyObject *__pyx_n_s_RULE_class_specifier;
static PyObject *__pyx_n_s_RULE_class_type;
static PyObject *__pyx_n_s_RULE_comment;
static PyObject *__pyx_n_s_RULE_component_clause;
static PyObject *__pyx_n_s_RULE_component_clause1;
static PyObject *__pyx_n_s_RULE_component_declaration;
static PyObject *__pyx_n_s_RULE_component_declaration1;
static PyObject *__pyx_n_s_RULE_component_list;
static PyObject *__pyx_n_s_RULE_component_reference;
static PyObject *__pyx_n_s_RULE_component_reference_element;
static PyObject *__pyx_n_s_RULE_composition;
static PyObject *__pyx_n_s_RULE_condition_attribute;
static PyObject *__pyx_n_s_RULE_connect_clause;
static PyObject *__pyx_n_s_RULE_constraining_clause;
static PyObject *__pyx_n_s_RULE_declaration;
static PyObject *__pyx_n_s_RULE_element;
static PyObject *__pyx_n_s_RULE_element_list;
static PyObject *__pyx_n_s_RULE_element_modification;
static PyObject *__pyx_n_s_RULE_element_modification_or_rep;
static PyObject *__pyx_n_s_RULE_element_redeclaration;
static PyObject *__pyx_n_s_RULE_element_replaceable;
static PyObject *__pyx_n_s_RULE_enum_list;
static PyObject *__pyx_n_s_RULE_enumeration_literal;
static PyObject *__pyx_n_s_RULE_equation;
static PyObject *__pyx_n_s_RULE_equation_options;
static PyObject *__pyx_n_s_RULE_equation_section;
static PyObject *__pyx_n_s_RULE_expr;
static PyObject *__pyx_n_s_RULE_expression;
static PyObject *__pyx_n_s_RULE_expression_list;
static PyObject *__pyx_n_s_RULE_extends_clause;
static PyObject *__pyx_n_s_RULE_external_function_call;
static PyObject *__pyx_n_s_RULE_for_equation;
static PyObject *__pyx_n_s_RULE_for_index;
static PyObject *__pyx_n_s_RULE_for_indices;
static PyObject *__pyx_n_s_RULE_for_statement;
static PyObject *__pyx_n_s_RULE_function_argument;
static PyObject *__pyx_n_s_RULE_function_arguments;
static PyObject *__pyx_n_s_RULE_function_call_args;
static PyObject *__pyx_n_s_RULE_if_equation;
static PyObject *__pyx_n_s_RULE_if_statement;
static PyObject *__pyx_n_s_RULE_import_clause;
static PyObject *__pyx_n_s_RULE_import_list;
static PyObject *__pyx_n_s_RULE_language_specification;
static PyObject *__pyx_n_s_RULE_modification;
static PyObject *__pyx_n_s_RULE_name;
static PyObject *__pyx_n_s_RULE_named_argument;
static PyObject *__pyx_n_s_RULE_named_arguments;
static PyObject *__pyx_n_s_RULE_output_expression_list;
static PyObject *__pyx_n_s_RULE_primary;
static PyObject *__pyx_n_s_RULE_regular_element;
static PyObject *__pyx_n_s_RULE_replaceable_element;
static PyObject *__pyx_n_s_RULE_short_class_definition;
static PyObject *__pyx_n_s_RULE_simple_expression;
static PyObject *__pyx_n_s_RULE_statement;
static PyObject *__pyx_n_s_RULE_statement_options;
static PyObject *__pyx_n_s_RULE_stored_definition;
static PyObject *__pyx_n_s_RULE_stored_definition_class;
static PyObject *__pyx_n_s_RULE_string_comment;
static PyObject *__pyx_n_s_RULE_subscript;
static PyObject *__pyx_n_s_RULE_type_prefix;
static PyObject *__pyx_n_s_RULE_type_specifier;
static PyObject *__pyx_n_s_RULE_when_equation;
static PyObject *__pyx_n_s_RULE_when_statement;
static PyObject *__pyx_n_s_RULE_while_statement;
static PyObject *__pyx_kp_u_R_H_U;
static PyObject *__pyx_n_s_RecognitionException;
static PyObject *__pyx_n_s_Regular_elementContext;
static PyObject *__pyx_n_s_Replaceable_elementContext;
static PyObject *__pyx_n_s_RuleContext;
static PyObject *__pyx_kp_u_S;
static PyObject *__pyx_n_s_STRING;
static PyObject *__pyx_n_u_STRING;
static PyObject *__pyx_kp_u_S_2;
static PyObject *__pyx_kp_u_S_3;
static PyObject *__pyx_kp_u_S_4;
static PyObject *__pyx_kp_u_S_5;
static PyObject *__pyx_kp_u_S_6;
static PyObject *__pyx_n_s_Short_class_definitionContext;
static PyObject *__pyx_n_s_Simple_expressionContext;
static PyObject *__pyx_n_s_StatementContext;
static PyObject *__pyx_n_s_Statement_breakContext;
static PyObject *__pyx_n_s_Statement_component_functionCont;
static PyObject *__pyx_n_s_Statement_component_referenceCon;
static PyObject *__pyx_n_s_Statement_forContext;
static PyObject *__pyx_n_s_Statement_ifContext;
static PyObject *__pyx_n_s_Statement_optionsContext;
static PyObject *__pyx_n_s_Statement_returnContext;
static PyObject *__pyx_n_s_Statement_whenContext;
static PyObject *__pyx_n_s_Statement_whileContext;
static PyObject *__pyx_n_s_Stored_definitionContext;
static PyObject *__pyx_n_s_Stored_definition_classContext;
static PyObject *__pyx_n_s_StringIO;
static PyObject *__pyx_n_s_String_commentContext;
static PyObject *__pyx_n_s_SubscriptContext;
static PyObject *__pyx_kp_u_T;
static PyObject *__pyx_kp_u_T_2;
static PyObject *__pyx_kp_u_T_3;
static PyObject *__pyx_kp_u_T_4;
static PyObject *__pyx_kp_u_T_V_X;
static PyObject *__pyx_n_s_T__0;
static PyObject *__pyx_n_s_T__1;
static PyObject *__pyx_n_s_T__10;
static PyObject *__pyx_n_s_T__11;
static PyObject *__pyx_n_s_T__12;
static PyObject *__pyx_n_s_T__13;
static PyObject *__pyx_n_s_T__14;
static PyObject *__pyx_n_s_T__15;
static PyObject *__pyx_n_s_T__16;
static PyObject *__pyx_n_s_T__17;
static PyObject *__pyx_n_s_T__18;
static PyObject *__pyx_n_s_T__19;
static PyObject *__pyx_n_s_T__2;
static PyObject *__pyx_n_s_T__20;
static PyObject *__pyx_n_s_T__21;
static PyObject *__pyx_n_s_T__22;
static PyObject *__pyx_n_s_T__23;
static PyObject *__pyx_n_s_T__24;
static PyObject *__pyx_n_s_T__25;
static PyObject *__pyx_n_s_T__26;
static PyObject *__pyx_n_s_T__27;
static PyObject *__pyx_n_s_T__28;
static PyObject *__pyx_n_s_T__29;
static PyObject *__pyx_n_s_T__3;
static PyObject *__pyx_n_s_T__30;
static PyObject *__pyx_n_s_T__31;
static PyObject *__pyx_n_s_T__32;
static PyObject *__pyx_n_s_T__33;
static PyObject *__pyx_n_s_T__34;
static PyObject *__pyx_n_s_T__35;
static PyObject *__pyx_n_s_T__36;
static PyObject *__pyx_n_s_T__37;
static PyObject *__pyx_n_s_T__38;
static PyObject *__pyx_n_s_T__39;
static PyObject *__pyx_n_s_T__4;
static PyObject *__pyx_n_s_T__40;
static PyObject *__pyx_n_s_T__41;
static PyObject *__pyx_n_s_T__42;
static PyObject *__pyx_n_s_T__43;
static PyObject *__pyx_n_s_T__44;
static PyObject *__pyx_n_s_T__45;
static PyObject *__pyx_n_s_T__46;
static PyObject *__pyx_n_s_T__47;
static PyObject *__pyx_n_s_T__48;
static PyObject *__pyx_n_s_T__49;
static PyObject *__pyx_n_s_T__5;
static PyObject *__pyx_n_s_T__50;
static PyObject *__pyx_n_s_T__51;
static PyObject *__pyx_n_s_T__52;
static PyObject *__pyx_n_s_T__53;
static PyObject *__pyx_n_s_T__54;
static PyObject *__pyx_n_s_T__55;
static PyObject *__pyx_n_s_T__56;
static PyObject *__pyx_n_s_T__57;
static PyObject *__pyx_n_s_T__58;
static PyObject *__pyx_n_s_T__59;
static PyObject *__pyx_n_s_T__6;
static PyObject *__pyx_n_s_T__60;
static PyObject *__pyx_n_s_T__61;
static PyObject *__pyx_n_s_T__62;
static PyObject *__pyx_n_s_T__63;
static PyObject *__pyx_n_s_T__64;
static PyObject *__pyx_n_s_T__65;
static PyObject *__pyx_n_s_T__66;
static PyObject *__pyx_n_s_T__67;
static PyObject *__pyx_n_s_T__68;
static PyObject *__pyx_n_s_T__69;
static PyObject *__pyx_n_s_T__7;
static PyObject *__pyx_n_s_T__70;
static PyObject *__pyx_n_s_T__71;
static PyObject *__pyx_n_s_T__72;
static PyObject *__pyx_n_s_T__73;
static PyObject *__pyx_n_s_T__74;
static PyObject *__pyx_n_s_T__75;
static PyObject *__pyx_n_s_T__76;
static PyObject *__pyx_n_s_T__77;
static PyObject *__pyx_n_s_T__8;
static PyObject *__pyx_n_s_T__9;
static PyObject *__pyx_kp_u_T_z;
static PyObject *__pyx_n_s_TextIO;
static PyObject *__pyx_n_s_Token;
static PyObject *__pyx_n_s_TokenStream;
static PyObject *__pyx_n_s_Type_prefixContext;
static PyObject *__pyx_n_s_Type_specifierContext;
static PyObject *__pyx_kp_u_U;
static PyObject *__pyx_n_s_UNSIGNED_NUMBER;
static PyObject *__pyx_n_u_UNSIGNED_NUMBER;
static PyObject *__pyx_kp_u_V;
static PyObject *__pyx_kp_u_V_2;
static PyObject *__pyx_kp_u_V_H_Y;
static PyObject *__pyx_kp_u_W;
static PyObject *__pyx_n_s_WITHIN;
static PyObject *__pyx_n_u_WITHIN;
static PyObject *__pyx_n_s_WS;
static PyObject *__pyx_n_u_WS;
static PyObject *__pyx_kp_u_W_2;
static PyObject *__pyx_kp_u_W_3;
static PyObject *__pyx_n_s_When_equationContext;
static PyObject *__pyx_n_s_When_statementContext;
static PyObject *__pyx_n_s_While_statementContext;
static PyObject *__pyx_kp_u_X;
static PyObject *__pyx_kp_u_X_2;
static PyObject *__pyx_kp_u_X_3;
static PyObject *__pyx_kp_u_X_4;
static PyObject *__pyx_kp_u_Y;
static PyObject *__pyx_kp_u_Y_2;
static PyObject *__pyx_kp_u_Y_z;
static PyObject *__pyx_kp_u_Z;
static PyObject *__pyx_kp_u_Z_0;
static PyObject *__pyx_kp_u_Z_10;
static PyObject *__pyx_kp_u_Z_11;
static PyObject *__pyx_kp_u_Z_12;
static PyObject *__pyx_kp_u_Z_2;
static PyObject *__pyx_kp_u_Z_3;
static PyObject *__pyx_kp_u_Z_4;
static PyObject *__pyx_kp_u_Z_5;
static PyObject *__pyx_kp_u_Z_5_n8;
static PyObject *__pyx_kp_u_Z_6;
static PyObject *__pyx_kp_u_Z_7;
static PyObject *__pyx_kp_u_Z_8;
static PyObject *__pyx_kp_u_Z_9;
static PyObject *__pyx_kp_u_Z_F;
static PyObject *__pyx_kp_u_Z_F_2;
static PyObject *__pyx_kp_u_Z_I;
static PyObject *__pyx_kp_u_Z_Z;
static PyObject *__pyx_kp_u_Z_Z_2;
static PyObject *__pyx_kp_u__1000;
static PyObject *__pyx_kp_u__1001;
static PyObject *__pyx_kp_u__1002;
static PyObject *__pyx_kp_u__1003;
static PyObject *__pyx_kp_u__1004;
static PyObject *__pyx_kp_u__1005;
static PyObject *__pyx_kp_u__101;
static PyObject *__pyx_kp_u__103;
static PyObject *__pyx_kp_u__105;
static PyObject *__pyx_kp_u__107;
static PyObject *__pyx_kp_u__109;
static PyObject *__pyx_kp_u__11;
static PyObject *__pyx_kp_u__111;
static PyObject *__pyx_kp_u__113;
static PyObject *__pyx_kp_u__115;
static PyObject *__pyx_kp_u__117;
static PyObject *__pyx_kp_u__119;
static PyObject *__pyx_kp_u__121;
static PyObject *__pyx_kp_u__123;
static PyObject *__pyx_kp_u__125;
static PyObject *__pyx_kp_u__13;
static PyObject *__pyx_kp_u__144;
static PyObject *__pyx_kp_u__146;
static PyObject *__pyx_kp_u__148;
static PyObject *__pyx_kp_u__15;
static PyObject *__pyx_kp_u__150;
static PyObject *__pyx_kp_u__152;
static PyObject *__pyx_kp_u__154;
static PyObject *__pyx_kp_u__156;
static PyObject *__pyx_kp_u__167;
static PyObject *__pyx_kp_u__169;
static PyObject *__pyx_kp_u__172;
static PyObject *__pyx_kp_u__174;
static PyObject *__pyx_kp_u__176;
static PyObject *__pyx_kp_u__178;
static PyObject *__pyx_kp_u__180;
static PyObject *__pyx_kp_u__182;
static PyObject *__pyx_kp_u__195;
static PyObject *__pyx_kp_u__197;
static PyObject *__pyx_kp_u__199;
static PyObject *__pyx_kp_u__201;
static PyObject *__pyx_kp_u__204;
static PyObject *__pyx_kp_u__206;
static PyObject *__pyx_kp_u__208;
static PyObject *__pyx_kp_u__21;
static PyObject *__pyx_kp_u__210;
static PyObject *__pyx_kp_u__212;
static PyObject *__pyx_kp_u__215;
static PyObject *__pyx_kp_u__218;
static PyObject *__pyx_kp_u__220;
static PyObject *__pyx_kp_u__223;
static PyObject *__pyx_kp_u__225;
static PyObject *__pyx_kp_u__227;
static PyObject *__pyx_kp_u__229;
static PyObject *__pyx_kp_u__23;
static PyObject *__pyx_kp_u__231;
static PyObject *__pyx_kp_u__233;
static PyObject *__pyx_kp_u__235;
static PyObject *__pyx_kp_u__237;
static PyObject *__pyx_kp_u__239;
static PyObject *__pyx_kp_u__241;
static PyObject *__pyx_kp_u__243;
static PyObject *__pyx_kp_u__245;
static PyObject *__pyx_kp_u__247;
static PyObject *__pyx_kp_u__249;
static PyObject *__pyx_kp_u__25;
static PyObject *__pyx_kp_u__255;
static PyObject *__pyx_kp_u__257;
static PyObject *__pyx_kp_u__261;
static PyObject *__pyx_kp_u__263;
static PyObject *__pyx_kp_u__265;
static PyObject *__pyx_kp_u__267;
static PyObject *__pyx_kp_u__27;
static PyObject *__pyx_kp_u__270;
static PyObject *__pyx_kp_u__274;
static PyObject *__pyx_kp_u__276;
static PyObject *__pyx_kp_u__280;
static PyObject *__pyx_kp_u__284;
static PyObject *__pyx_kp_u__286;
static PyObject *__pyx_kp_u__288;
static PyObject *__pyx_kp_u__29;
static PyObject *__pyx_kp_u__290;
static PyObject *__pyx_kp_u__292;
static PyObject *__pyx_kp_u__294;
static PyObject *__pyx_kp_u__298;
static PyObject *__pyx_kp_u__3;
static PyObject *__pyx_kp_u__302;
static PyObject *__pyx_kp_u__304;
static PyObject *__pyx_kp_u__306;
static PyObject *__pyx_kp_u__308;
static PyObject *__pyx_kp_u__31;
static PyObject *__pyx_kp_u__310;
static PyObject *__pyx_kp_u__312;
static PyObject *__pyx_kp_u__315;
static PyObject *__pyx_kp_u__317;
static PyObject *__pyx_kp_u__320;
static PyObject *__pyx_kp_u__322;
static PyObject *__pyx_kp_u__325;
static PyObject *__pyx_kp_u__329;
static PyObject *__pyx_kp_u__33;
static PyObject *__pyx_kp_u__331;
static PyObject *__pyx_kp_u__333;
static PyObject *__pyx_kp_u__335;
static PyObject *__pyx_kp_u__337;
static PyObject *__pyx_kp_u__339;
static PyObject *__pyx_kp_u__341;
static PyObject *__pyx_kp_u__344;
static PyObject *__pyx_kp_u__348;
static PyObject *__pyx_kp_u__35;
static PyObject *__pyx_kp_u__351;
static PyObject *__pyx_kp_u__353;
static PyObject *__pyx_kp_u__357;
static PyObject *__pyx_kp_u__361;
static PyObject *__pyx_kp_u__363;
static PyObject *__pyx_kp_u__365;
static PyObject *__pyx_kp_u__368;
static PyObject *__pyx_kp_u__37;
static PyObject *__pyx_kp_u__373;
static PyObject *__pyx_kp_u__375;
static PyObject *__pyx_kp_u__377;
static PyObject *__pyx_kp_u__379;
static PyObject *__pyx_kp_u__383;
static PyObject *__pyx_kp_u__385;
static PyObject *__pyx_kp_u__387;
static PyObject *__pyx_kp_u__39;
static PyObject *__pyx_kp_u__391;
static PyObject *__pyx_kp_u__394;
static PyObject *__pyx_kp_u__396;
static PyObject *__pyx_kp_u__400;
static PyObject *__pyx_kp_u__402;
static PyObject *__pyx_kp_u__404;
static PyObject *__pyx_kp_u__406;
static PyObject *__pyx_kp_u__41;
static PyObject *__pyx_kp_u__410;
static PyObject *__pyx_kp_u__414;
static PyObject *__pyx_kp_u__421;
static PyObject *__pyx_kp_u__427;
static PyObject *__pyx_kp_u__429;
static PyObject *__pyx_kp_u__43;
static PyObject *__pyx_kp_u__431;
static PyObject *__pyx_kp_u__433;
static PyObject *__pyx_kp_u__435;
static PyObject *__pyx_kp_u__437;
static PyObject *__pyx_kp_u__441;
static PyObject *__pyx_kp_u__447;
static PyObject *__pyx_kp_u__45;
static PyObject *__pyx_kp_u__450;
static PyObject *__pyx_kp_u__458;
static PyObject *__pyx_kp_u__460;
static PyObject *__pyx_kp_u__467;
static PyObject *__pyx_kp_u__469;
static PyObject *__pyx_kp_u__47;
static PyObject *__pyx_kp_u__472;
static PyObject *__pyx_kp_u__474;
static PyObject *__pyx_kp_u__476;
static PyObject *__pyx_kp_u__478;
static PyObject *__pyx_kp_u__484;
static PyObject *__pyx_kp_u__486;
static PyObject *__pyx_kp_u__49;
static PyObject *__pyx_kp_u__491;
static PyObject *__pyx_kp_u__493;
static PyObject *__pyx_kp_u__499;
static PyObject *__pyx_kp_u__5;
static PyObject *__pyx_kp_u__501;
static PyObject *__pyx_kp_u__506;
static PyObject *__pyx_kp_u__51;
static PyObject *__pyx_kp_u__510;
static PyObject *__pyx_kp_u__512;
static PyObject *__pyx_kp_u__515;
static PyObject *__pyx_kp_u__519;
static PyObject *__pyx_kp_u__521;
static PyObject *__pyx_kp_u__527;
static PyObject *__pyx_kp_u__529;
static PyObject *__pyx_kp_u__53;
static PyObject *__pyx_kp_u__533;
static PyObject *__pyx_kp_u__535;
static PyObject *__pyx_kp_u__537;
static PyObject *__pyx_kp_u__539;
static PyObject *__pyx_kp_u__543;
static PyObject *__pyx_kp_u__545;
static PyObject *__pyx_kp_u__547;
static PyObject *__pyx_kp_u__549;
static PyObject *__pyx_kp_u__55;
static PyObject *__pyx_kp_u__553;
static PyObject *__pyx_kp_u__555;
static PyObject *__pyx_kp_u__559;
static PyObject *__pyx_kp_u__561;
static PyObject *__pyx_kp_u__563;
static PyObject *__pyx_kp_u__565;
static PyObject *__pyx_kp_u__567;
static PyObject *__pyx_kp_u__57;
static PyObject *__pyx_kp_u__570;
static PyObject *__pyx_kp_u__572;
static PyObject *__pyx_kp_u__574;
static PyObject *__pyx_kp_u__576;
static PyObject *__pyx_kp_u__578;
static PyObject *__pyx_kp_u__582;
static PyObject *__pyx_kp_u__584;
static PyObject *__pyx_kp_u__589;
static PyObject *__pyx_kp_u__59;
static PyObject *__pyx_kp_u__591;
static PyObject *__pyx_kp_u__593;
static PyObject *__pyx_kp_u__597;
static PyObject *__pyx_kp_u__604;
static PyObject *__pyx_kp_u__606;
static PyObject *__pyx_kp_u__61;
static PyObject *__pyx_kp_u__611;
static PyObject *__pyx_kp_u__613;
static PyObject *__pyx_kp_u__617;
static PyObject *__pyx_kp_u__619;
static PyObject *__pyx_kp_u__621;
static PyObject *__pyx_kp_u__623;
static PyObject *__pyx_kp_u__625;
static PyObject *__pyx_kp_u__63;
static PyObject *__pyx_kp_u__630;
static PyObject *__pyx_kp_u__632;
static PyObject *__pyx_kp_u__636;
static PyObject *__pyx_kp_u__638;
static PyObject *__pyx_kp_u__640;
static PyObject *__pyx_kp_u__642;
static PyObject *__pyx_kp_u__644;
static PyObject *__pyx_kp_u__65;
static PyObject *__pyx_kp_u__653;
static PyObject *__pyx_kp_u__655;
static PyObject *__pyx_kp_u__661;
static PyObject *__pyx_kp_u__668;
static PyObject *__pyx_kp_u__67;
static PyObject *__pyx_kp_u__670;
static PyObject *__pyx_kp_u__677;
static PyObject *__pyx_kp_u__679;
static PyObject *__pyx_kp_u__684;
static PyObject *__pyx_kp_u__69;
static PyObject *__pyx_kp_u__690;
static PyObject *__pyx_kp_u__692;
static PyObject *__pyx_kp_u__695;
static PyObject *__pyx_kp_u__697;
static PyObject *__pyx_kp_u__699;
static PyObject *__pyx_kp_u__7;
static PyObject *__pyx_kp_u__701;
static PyObject *__pyx_kp_u__703;
static PyObject *__pyx_kp_u__705;
static PyObject *__pyx_kp_u__708;
static PyObject *__pyx_kp_u__71;
static PyObject *__pyx_kp_u__711;
static PyObject *__pyx_kp_u__717;
static PyObject *__pyx_kp_u__719;
static PyObject *__pyx_kp_u__721;
static PyObject *__pyx_kp_u__723;
static PyObject *__pyx_kp_u__728;
static PyObject *__pyx_kp_u__73;
static PyObject *__pyx_kp_u__730;
static PyObject *__pyx_kp_u__735;
static PyObject *__pyx_kp_u__737;
static PyObject *__pyx_kp_u__743;
static PyObject *__pyx_kp_u__747;
static PyObject *__pyx_kp_u__75;
static PyObject *__pyx_kp_u__750;
static PyObject *__pyx_kp_u__754;
static PyObject *__pyx_kp_u__756;
static PyObject *__pyx_kp_u__760;
static PyObject *__pyx_kp_u__764;
static PyObject *__pyx_kp_u__767;
static PyObject *__pyx_kp_u__769;
static PyObject *__pyx_kp_u__77;
static PyObject *__pyx_kp_u__771;
static PyObject *__pyx_kp_u__773;
static PyObject *__pyx_kp_u__775;
static PyObject *__pyx_kp_u__778;
static PyObject *__pyx_kp_u__780;
static PyObject *__pyx_n_u__782;
static PyObject *__pyx_n_u__784;
static PyObject *__pyx_n_u__786;
static PyObject *__pyx_n_u__788;
static PyObject *__pyx_kp_u__79;
static PyObject *__pyx_n_u__790;
static PyObject *__pyx_n_u__792;
static PyObject *__pyx_n_u__794;
static PyObject *__pyx_n_u__796;
static PyObject *__pyx_n_u__798;
static PyObject *__pyx_kp_u__800;
static PyObject *__pyx_kp_u__802;
static PyObject *__pyx_n_u__804;
static PyObject *__pyx_n_u__806;
static PyObject *__pyx_n_u__808;
static PyObject *__pyx_kp_u__81;
static PyObject *__pyx_kp_u__83;
static PyObject *__pyx_kp_u__85;
static PyObject *__pyx_kp_u__87;
static PyObject *__pyx_kp_u__89;
static PyObject *__pyx_kp_u__9;
static PyObject *__pyx_kp_u__91;
static PyObject *__pyx_kp_u__93;
static PyObject *__pyx_kp_u__95;
static PyObject *__pyx_kp_u__97;
static PyObject *__pyx_n_s__975;
static PyObject *__pyx_kp_u__978;
static PyObject *__pyx_kp_u__979;
static PyObject *__pyx_kp_u__980;
static PyObject *__pyx_kp_u__981;
static PyObject *__pyx_kp_u__982;
static PyObject *__pyx_kp_u__983;
static PyObject *__pyx_kp_u__984;
static PyObject *__pyx_kp_u__985;
static PyObject *__pyx_kp_u__986;
static PyObject *__pyx_kp_u__987;
static PyObject *__pyx_kp_u__988;
static PyObject *__pyx_kp_u__989;
static PyObject *__pyx_kp_u__99;
static PyObject *__pyx_kp_u__990;
static PyObject *__pyx_kp_u__991;
static PyObject *__pyx_kp_u__992;
static PyObject *__pyx_kp_u__993;
static PyObject *__pyx_kp_u__994;
static PyObject *__pyx_kp_u__995;
static PyObject *__pyx_kp_u__996;
static PyObject *__pyx_kp_u__997;
static PyObject *__pyx_kp_u__998;
static PyObject *__pyx_kp_u__999;
static PyObject *__pyx_n_s_accept;
static PyObject *__pyx_n_s_adaptivePredict;
static PyObject *__pyx_kp_u_algorithm;
static PyObject *__pyx_n_s_algorithm_section;
static PyObject *__pyx_n_u_algorithm_section;
static PyObject *__pyx_n_s_alt;
static PyObject *__pyx_kp_u_and;
static PyObject *__pyx_n_s_annotation;
static PyObject *__pyx_n_u_annotation;
static PyObject *__pyx_kp_u_annotation_2;
static PyObject *__pyx_n_s_antlr4;
static PyObject *__pyx_n_s_antlr4_error_Errors;
static PyObject *__pyx_n_s_argument;
static PyObject *__pyx_n_u_argument;
static PyObject *__pyx_n_s_argument_list;
static PyObject *__pyx_n_u_argument_list;
static PyObject *__pyx_n_s_array_subscripts;
static PyObject *__pyx_n_u_array_subscripts;
static PyObject *__pyx_n_s_atn;
static PyObject *__pyx_kp_u_b_d_f;
static PyObject *__pyx_n_s_base_prefix;
static PyObject *__pyx_n_u_base_prefix;
static PyObject *__pyx_kp_u_block;
static PyObject *__pyx_kp_u_break;
static PyObject *__pyx_n_s_buf;
static PyObject *__pyx_kp_u_c;
static PyObject *__pyx_n_s_checkVersion;
static PyObject *__pyx_kp_u_class;
static PyObject *__pyx_n_s_class_definition;
static PyObject *__pyx_n_u_class_definition;
static PyObject *__pyx_n_s_class_elem;
static PyObject *__pyx_n_s_class_modification;
static PyObject *__pyx_n_u_class_modification;
static PyObject *__pyx_n_s_class_prefixes;
static PyObject *__pyx_n_u_class_prefixes;
static PyObject *__pyx_n_s_class_specifier;
static PyObject *__pyx_n_u_class_specifier;
static PyObject *__pyx_n_s_class_type;
static PyObject *__pyx_n_u_class_type;
static PyObject *__pyx_n_s_comment;
static PyObject *__pyx_n_u_comment;
static PyObject *__pyx_n_s_comp_annotation;
static PyObject *__pyx_n_s_comp_elem;
static PyObject *__pyx_n_s_component_clause;
static PyObject *__pyx_n_u_component_clause;
static PyObject *__pyx_n_s_component_clause1;
static PyObject *__pyx_n_u_component_clause1;
static PyObject *__pyx_n_s_component_declaration;
static PyObject *__pyx_n_u_component_declaration;
static PyObject *__pyx_n_s_component_declaration1;
static PyObject *__pyx_n_u_component_declaration1;
static PyObject *__pyx_n_s_component_list;
static PyObject *__pyx_n_u_component_list;
static PyObject *__pyx_n_s_component_reference;
static PyObject *__pyx_n_u_component_reference;
static PyObject *__pyx_n_s_component_reference_element;
static PyObject *__pyx_n_u_component_reference_element;
static PyObject *__pyx_n_s_composition;
static PyObject *__pyx_n_u_composition;
static PyObject *__pyx_n_s_condition_attribute;
static PyObject *__pyx_n_u_condition_attribute;
static PyObject *__pyx_kp_u_connect;
static PyObject *__pyx_n_s_connect_clause;
static PyObject *__pyx_n_u_connect_clause;
static PyObject *__pyx_kp_u_connector;
static PyObject *__pyx_kp_u_constant;
static PyObject *__pyx_kp_u_constrainedby;
static PyObject *__pyx_n_s_constraining_clause;
static PyObject *__pyx_n_u_constraining_clause;
static PyObject *__pyx_n_s_consume;
static PyObject *__pyx_n_s_copyFrom;
static PyObject *__pyx_n_s_ctx;
static PyObject *__pyx_n_s_ctx_2;
static PyObject *__pyx_kp_u_d3;
static PyObject *__pyx_n_s_decisionToState;
static PyObject *__pyx_n_s_decisionsToDFA;
static PyObject *__pyx_n_s_declaration;
static PyObject *__pyx_n_u_declaration;
static PyObject *__pyx_kp_u_der;
static PyObject *__pyx_n_s_deserialize;
static PyObject *__pyx_kp_u_discrete;
static PyObject *__pyx_n_s_doc;
static PyObject *__pyx_kp_u_e;
static PyObject *__pyx_kp_u_each;
static PyObject *__pyx_n_s_element;
static PyObject *__pyx_n_u_element;
static PyObject *__pyx_n_s_element_list;
static PyObject *__pyx_n_u_element_list;
static PyObject *__pyx_n_s_element_modification;
static PyObject *__pyx_n_u_element_modification;
static PyObject *__pyx_n_s_element_modification_or_replacea;
static PyObject *__pyx_n_u_element_modification_or_replacea;
static PyObject *__pyx_n_s_element_redeclaration;
static PyObject *__pyx_n_u_element_redeclaration;
static PyObject *__pyx_n_s_element_replaceable;
static PyObject *__pyx_n_u_element_replaceable;
static PyObject *__pyx_kp_u_else;
static PyObject *__pyx_kp_u_elseif;
static PyObject *__pyx_kp_u_elsewhen;
static PyObject *__pyx_kp_u_encapsulated;
static PyObject *__pyx_kp_u_end;
static PyObject *__pyx_n_s_enter;
static PyObject *__pyx_n_s_enterAlgorithm_section;
static PyObject *__pyx_n_u_enterAlgorithm_section;
static PyObject *__pyx_n_s_enterAnnotation;
static PyObject *__pyx_n_u_enterAnnotation;
static PyObject *__pyx_n_s_enterArgument;
static PyObject *__pyx_n_u_enterArgument;
static PyObject *__pyx_n_s_enterArgument_expression;
static PyObject *__pyx_n_u_enterArgument_expression;
static PyObject *__pyx_n_s_enterArgument_function;
static PyObject *__pyx_n_u_enterArgument_function;
static PyObject *__pyx_n_s_enterArgument_list;
static PyObject *__pyx_n_u_enterArgument_list;
static PyObject *__pyx_n_s_enterArray_subscripts;
static PyObject *__pyx_n_u_enterArray_subscripts;
static PyObject *__pyx_n_s_enterBase_prefix;
static PyObject *__pyx_n_u_enterBase_prefix;
static PyObject *__pyx_n_s_enterClass_definition;
static PyObject *__pyx_n_u_enterClass_definition;
static PyObject *__pyx_n_s_enterClass_modification;
static PyObject *__pyx_n_u_enterClass_modification;
static PyObject *__pyx_n_s_enterClass_prefixes;
static PyObject *__pyx_n_u_enterClass_prefixes;
static PyObject *__pyx_n_s_enterClass_spec_base;
static PyObject *__pyx_n_u_enterClass_spec_base;
static PyObject *__pyx_n_s_enterClass_spec_comp;
static PyObject *__pyx_n_u_enterClass_spec_comp;
static PyObject *__pyx_n_s_enterClass_spec_der;
static PyObject *__pyx_n_u_enterClass_spec_der;
static PyObject *__pyx_n_s_enterClass_spec_enum;
static PyObject *__pyx_n_u_enterClass_spec_enum;
static PyObject *__pyx_n_s_enterClass_spec_extends;
static PyObject *__pyx_n_u_enterClass_spec_extends;
static PyObject *__pyx_n_s_enterClass_type;
static PyObject *__pyx_n_u_enterClass_type;
static PyObject *__pyx_n_s_enterComment;
static PyObject *__pyx_n_u_enterComment;
static PyObject *__pyx_n_s_enterComponent_clause;
static PyObject *__pyx_n_u_enterComponent_clause;
static PyObject *__pyx_n_s_enterComponent_clause1;
static PyObject *__pyx_n_u_enterComponent_clause1;
static PyObject *__pyx_n_s_enterComponent_declaration;
static PyObject *__pyx_n_u_enterComponent_declaration;
static PyObject *__pyx_n_s_enterComponent_declaration1;
static PyObject *__pyx_n_u_enterComponent_declaration1;
static PyObject *__pyx_n_s_enterComponent_list;
static PyObject *__pyx_n_u_enterComponent_list;
static PyObject *__pyx_n_s_enterComponent_reference;
static PyObject *__pyx_n_u_enterComponent_reference;
static PyObject *__pyx_n_s_enterComponent_reference_element;
static PyObject *__pyx_n_u_enterComponent_reference_element;
static PyObject *__pyx_n_s_enterComposition;
static PyObject *__pyx_n_u_enterComposition;
static PyObject *__pyx_n_s_enterCondition_attribute;
static PyObject *__pyx_n_u_enterCondition_attribute;
static PyObject *__pyx_n_s_enterConnect_clause;
static PyObject *__pyx_n_u_enterConnect_clause;
static PyObject *__pyx_n_s_enterConstraining_clause;
static PyObject *__pyx_n_u_enterConstraining_clause;
static PyObject *__pyx_n_s_enterDeclaration;
static PyObject *__pyx_n_u_enterDeclaration;
static PyObject *__pyx_n_s_enterElement;
static PyObject *__pyx_n_u_enterElement;
static PyObject *__pyx_n_s_enterElement_list;
static PyObject *__pyx_n_u_enterElement_list;
static PyObject *__pyx_n_s_enterElement_modification;
static PyObject *__pyx_n_u_enterElement_modification;
static PyObject *__pyx_n_s_enterElement_modification_or_rep;
static PyObject *__pyx_n_u_enterElement_modification_or_rep;
static PyObject *__pyx_n_s_enterElement_redeclaration;
static PyObject *__pyx_n_u_enterElement_redeclaration;
static PyObject *__pyx_n_s_enterElement_replaceable;
static PyObject *__pyx_n_u_enterElement_replaceable;
static PyObject *__pyx_n_s_enterEnum_list;
static PyObject *__pyx_n_u_enterEnum_list;
static PyObject *__pyx_n_s_enterEnumeration_literal;
static PyObject *__pyx_n_u_enterEnumeration_literal;
static PyObject *__pyx_n_s_enterEquation;
static PyObject *__pyx_n_u_enterEquation;
static PyObject *__pyx_n_s_enterEquation_connect_clause;
static PyObject *__pyx_n_u_enterEquation_connect_clause;
static PyObject *__pyx_n_s_enterEquation_for;
static PyObject *__pyx_n_u_enterEquation_for;
static PyObject *__pyx_n_s_enterEquation_function;
static PyObject *__pyx_n_u_enterEquation_function;
static PyObject *__pyx_n_s_enterEquation_if;
static PyObject *__pyx_n_u_enterEquation_if;
static PyObject *__pyx_n_s_enterEquation_section;
static PyObject *__pyx_n_u_enterEquation_section;
static PyObject *__pyx_n_s_enterEquation_simple;
static PyObject *__pyx_n_u_enterEquation_simple;
static PyObject *__pyx_n_s_enterEquation_when;
static PyObject *__pyx_n_u_enterEquation_when;
static PyObject *__pyx_n_s_enterExpr_add;
static PyObject *__pyx_n_u_enterExpr_add;
static PyObject *__pyx_n_s_enterExpr_and;
static PyObject *__pyx_n_u_enterExpr_and;
static PyObject *__pyx_n_s_enterExpr_exp;
static PyObject *__pyx_n_u_enterExpr_exp;
static PyObject *__pyx_n_s_enterExpr_mul;
static PyObject *__pyx_n_u_enterExpr_mul;
static PyObject *__pyx_n_s_enterExpr_neg;
static PyObject *__pyx_n_u_enterExpr_neg;
static PyObject *__pyx_n_s_enterExpr_not;
static PyObject *__pyx_n_u_enterExpr_not;
static PyObject *__pyx_n_s_enterExpr_or;
static PyObject *__pyx_n_u_enterExpr_or;
static PyObject *__pyx_n_s_enterExpr_primary;
static PyObject *__pyx_n_u_enterExpr_primary;
static PyObject *__pyx_n_s_enterExpr_rel;
static PyObject *__pyx_n_u_enterExpr_rel;
static PyObject *__pyx_n_s_enterExpression_if;
static PyObject *__pyx_n_u_enterExpression_if;
static PyObject *__pyx_n_s_enterExpression_list;
static PyObject *__pyx_n_u_enterExpression_list;
static PyObject *__pyx_n_s_enterExpression_simple;
static PyObject *__pyx_n_u_enterExpression_simple;
static PyObject *__pyx_n_s_enterExtends_clause;
static PyObject *__pyx_n_u_enterExtends_clause;
static PyObject *__pyx_n_s_enterExternal_function_call;
static PyObject *__pyx_n_u_enterExternal_function_call;
static PyObject *__pyx_n_s_enterFor_equation;
static PyObject *__pyx_n_u_enterFor_equation;
static PyObject *__pyx_n_s_enterFor_index;
static PyObject *__pyx_n_u_enterFor_index;
static PyObject *__pyx_n_s_enterFor_indices;
static PyObject *__pyx_n_u_enterFor_indices;
static PyObject *__pyx_n_s_enterFor_statement;
static PyObject *__pyx_n_u_enterFor_statement;
static PyObject *__pyx_n_s_enterFunction_arguments;
static PyObject *__pyx_n_u_enterFunction_arguments;
static PyObject *__pyx_n_s_enterFunction_call_args;
static PyObject *__pyx_n_u_enterFunction_call_args;
static PyObject *__pyx_n_s_enterIf_equation;
static PyObject *__pyx_n_u_enterIf_equation;
static PyObject *__pyx_n_s_enterIf_statement;
static PyObject *__pyx_n_u_enterIf_statement;
static PyObject *__pyx_n_s_enterImport_clause;
static PyObject *__pyx_n_u_enterImport_clause;
static PyObject *__pyx_n_s_enterImport_list;
static PyObject *__pyx_n_u_enterImport_list;
static PyObject *__pyx_n_s_enterLanguage_specification;
static PyObject *__pyx_n_u_enterLanguage_specification;
static PyObject *__pyx_n_s_enterModification_assignment;
static PyObject *__pyx_n_u_enterModification_assignment;
static PyObject *__pyx_n_s_enterModification_assignment2;
static PyObject *__pyx_n_u_enterModification_assignment2;
static PyObject *__pyx_n_s_enterModification_class;
static PyObject *__pyx_n_u_enterModification_class;
static PyObject *__pyx_n_s_enterName;
static PyObject *__pyx_n_u_enterName;
static PyObject *__pyx_n_s_enterNamed_argument;
static PyObject *__pyx_n_u_enterNamed_argument;
static PyObject *__pyx_n_s_enterNamed_arguments;
static PyObject *__pyx_n_u_enterNamed_arguments;
static PyObject *__pyx_n_s_enterOuterAlt;
static PyObject *__pyx_n_s_enterOutput_expression_list;
static PyObject *__pyx_n_u_enterOutput_expression_list;
static PyObject *__pyx_n_s_enterPrimary_component_reference;
static PyObject *__pyx_n_u_enterPrimary_component_reference;
static PyObject *__pyx_n_s_enterPrimary_derivative;
static PyObject *__pyx_n_u_enterPrimary_derivative;
static PyObject *__pyx_n_s_enterPrimary_end;
static PyObject *__pyx_n_u_enterPrimary_end;
static PyObject *__pyx_n_s_enterPrimary_expression_list;
static PyObject *__pyx_n_u_enterPrimary_expression_list;
static PyObject *__pyx_n_s_enterPrimary_false;
static PyObject *__pyx_n_u_enterPrimary_false;
static PyObject *__pyx_n_s_enterPrimary_function;
static PyObject *__pyx_n_u_enterPrimary_function;
static PyObject *__pyx_n_s_enterPrimary_function_arguments;
static PyObject *__pyx_n_u_enterPrimary_function_arguments;
static PyObject *__pyx_n_s_enterPrimary_initial;
static PyObject *__pyx_n_u_enterPrimary_initial;
static PyObject *__pyx_n_s_enterPrimary_output_expression_l;
static PyObject *__pyx_n_u_enterPrimary_output_expression_l;
static PyObject *__pyx_n_s_enterPrimary_string;
static PyObject *__pyx_n_u_enterPrimary_string;
static PyObject *__pyx_n_s_enterPrimary_true;
static PyObject *__pyx_n_u_enterPrimary_true;
static PyObject *__pyx_n_s_enterPrimary_unsigned_number;
static PyObject *__pyx_n_u_enterPrimary_unsigned_number;
static PyObject *__pyx_n_s_enterRecursionRule;
static PyObject *__pyx_n_s_enterRegular_element;
static PyObject *__pyx_n_u_enterRegular_element;
static PyObject *__pyx_n_s_enterReplaceable_element;
static PyObject *__pyx_n_u_enterReplaceable_element;
static PyObject *__pyx_n_s_enterRule;
static PyObject *__pyx_n_s_enterShort_class_definition;
static PyObject *__pyx_n_u_enterShort_class_definition;
static PyObject *__pyx_n_s_enterSimple_expression;
static PyObject *__pyx_n_u_enterSimple_expression;
static PyObject *__pyx_n_s_enterStatement;
static PyObject *__pyx_n_u_enterStatement;
static PyObject *__pyx_n_s_enterStatement_break;
static PyObject *__pyx_n_u_enterStatement_break;
static PyObject *__pyx_n_s_enterStatement_component_functio;
static PyObject *__pyx_n_u_enterStatement_component_functio;
static PyObject *__pyx_n_s_enterStatement_component_referen;
static PyObject *__pyx_n_u_enterStatement_component_referen;
static PyObject *__pyx_n_s_enterStatement_for;
static PyObject *__pyx_n_u_enterStatement_for;
static PyObject *__pyx_n_s_enterStatement_if;
static PyObject *__pyx_n_u_enterStatement_if;
static PyObject *__pyx_n_s_enterStatement_return;
static PyObject *__pyx_n_u_enterStatement_return;
static PyObject *__pyx_n_s_enterStatement_when;
static PyObject *__pyx_n_u_enterStatement_when;
static PyObject *__pyx_n_s_enterStatement_while;
static PyObject *__pyx_n_u_enterStatement_while;
static PyObject *__pyx_n_s_enterStored_definition;
static PyObject *__pyx_n_u_enterStored_definition;
static PyObject *__pyx_n_s_enterStored_definition_class;
static PyObject *__pyx_n_u_enterStored_definition_class;
static PyObject *__pyx_n_s_enterString_comment;
static PyObject *__pyx_n_u_enterString_comment;
static PyObject *__pyx_n_s_enterSubscript;
static PyObject *__pyx_n_u_enterSubscript;
static PyObject *__pyx_n_s_enterType_prefix;
static PyObject *__pyx_n_u_enterType_prefix;
static PyObject *__pyx_n_s_enterType_specifier;
static PyObject *__pyx_n_u_enterType_specifier;
static PyObject *__pyx_n_s_enterWhen_equation;
static PyObject *__pyx_n_u_enterWhen_equation;
static PyObject *__pyx_n_s_enterWhen_statement;
static PyObject *__pyx_n_u_enterWhen_statement;
static PyObject *__pyx_n_s_enterWhile_statement;
static PyObject *__pyx_n_u_enterWhile_statement;
static PyObject *__pyx_n_s_enum_list;
static PyObject *__pyx_n_u_enum_list;
static PyObject *__pyx_n_s_enumerate;
static PyObject *__pyx_kp_u_enumeration;
static PyObject *__pyx_n_s_enumeration_literal;
static PyObject *__pyx_n_u_enumeration_literal;
static PyObject *__pyx_n_s_epriv;
static PyObject *__pyx_n_s_epro;
static PyObject *__pyx_n_s_epub;
static PyObject *__pyx_n_s_equation;
static PyObject *__pyx_n_u_equation;
static PyObject *__pyx_kp_u_equation_2;
static PyObject *__pyx_n_s_equation_options;
static PyObject *__pyx_n_u_equation_options;
static PyObject *__pyx_n_s_equation_section;
static PyObject *__pyx_n_u_equation_section;
static PyObject *__pyx_n_s_errHandler;
static PyObject *__pyx_n_s_exception;
static PyObject *__pyx_n_s_exit;
static PyObject *__pyx_n_s_exitAlgorithm_section;
static PyObject *__pyx_n_u_exitAlgorithm_section;
static PyObject *__pyx_n_s_exitAnnotation;
static PyObject *__pyx_n_u_exitAnnotation;
static PyObject *__pyx_n_s_exitArgument;
static PyObject *__pyx_n_u_exitArgument;
static PyObject *__pyx_n_s_exitArgument_expression;
static PyObject *__pyx_n_u_exitArgument_expression;
static PyObject *__pyx_n_s_exitArgument_function;
static PyObject *__pyx_n_u_exitArgument_function;
static PyObject *__pyx_n_s_exitArgument_list;
static PyObject *__pyx_n_u_exitArgument_list;
static PyObject *__pyx_n_s_exitArray_subscripts;
static PyObject *__pyx_n_u_exitArray_subscripts;
static PyObject *__pyx_n_s_exitBase_prefix;
static PyObject *__pyx_n_u_exitBase_prefix;
static PyObject *__pyx_n_s_exitClass_definition;
static PyObject *__pyx_n_u_exitClass_definition;
static PyObject *__pyx_n_s_exitClass_modification;
static PyObject *__pyx_n_u_exitClass_modification;
static PyObject *__pyx_n_s_exitClass_prefixes;
static PyObject *__pyx_n_u_exitClass_prefixes;
static PyObject *__pyx_n_s_exitClass_spec_base;
static PyObject *__pyx_n_u_exitClass_spec_base;
static PyObject *__pyx_n_s_exitClass_spec_comp;
static PyObject *__pyx_n_u_exitClass_spec_comp;
static PyObject *__pyx_n_s_exitClass_spec_der;
static PyObject *__pyx_n_u_exitClass_spec_der;
static PyObject *__pyx_n_s_exitClass_spec_enum;
static PyObject *__pyx_n_u_exitClass_spec_enum;
static PyObject *__pyx_n_s_exitClass_spec_extends;
static PyObject *__pyx_n_u_exitClass_spec_extends;
static PyObject *__pyx_n_s_exitClass_type;
static PyObject *__pyx_n_u_exitClass_type;
static PyObject *__pyx_n_s_exitComment;
static PyObject *__pyx_n_u_exitComment;
static PyObject *__pyx_n_s_exitComponent_clause;
static PyObject *__pyx_n_u_exitComponent_clause;
static PyObject *__pyx_n_s_exitComponent_clause1;
static PyObject *__pyx_n_u_exitComponent_clause1;
static PyObject *__pyx_n_s_exitComponent_declaration;
static PyObject *__pyx_n_u_exitComponent_declaration;
static PyObject *__pyx_n_s_exitComponent_declaration1;
static PyObject *__pyx_n_u_exitComponent_declaration1;
static PyObject *__pyx_n_s_exitComponent_list;
static PyObject *__pyx_n_u_exitComponent_list;
static PyObject *__pyx_n_s_exitComponent_reference;
static PyObject *__pyx_n_u_exitComponent_reference;
static PyObject *__pyx_n_s_exitComponent_reference_element;
static PyObject *__pyx_n_u_exitComponent_reference_element;
static PyObject *__pyx_n_s_exitComposition;
static PyObject *__pyx_n_u_exitComposition;
static PyObject *__pyx_n_s_exitCondition_attribute;
static PyObject *__pyx_n_u_exitCondition_attribute;
static PyObject *__pyx_n_s_exitConnect_clause;
static PyObject *__pyx_n_u_exitConnect_clause;
static PyObject *__pyx_n_s_exitConstraining_clause;
static PyObject *__pyx_n_u_exitConstraining_clause;
static PyObject *__pyx_n_s_exitDeclaration;
static PyObject *__pyx_n_u_exitDeclaration;
static PyObject *__pyx_n_s_exitElement;
static PyObject *__pyx_n_u_exitElement;
static PyObject *__pyx_n_s_exitElement_list;
static PyObject *__pyx_n_u_exitElement_list;
static PyObject *__pyx_n_s_exitElement_modification;
static PyObject *__pyx_n_u_exitElement_modification;
static PyObject *__pyx_n_s_exitElement_modification_or_repl;
static PyObject *__pyx_n_u_exitElement_modification_or_repl;
static PyObject *__pyx_n_s_exitElement_redeclaration;
static PyObject *__pyx_n_u_exitElement_redeclaration;
static PyObject *__pyx_n_s_exitElement_replaceable;
static PyObject *__pyx_n_u_exitElement_replaceable;
static PyObject *__pyx_n_s_exitEnum_list;
static PyObject *__pyx_n_u_exitEnum_list;
static PyObject *__pyx_n_s_exitEnumeration_literal;
static PyObject *__pyx_n_u_exitEnumeration_literal;
static PyObject *__pyx_n_s_exitEquation;
static PyObject *__pyx_n_u_exitEquation;
static PyObject *__pyx_n_s_exitEquation_connect_clause;
static PyObject *__pyx_n_u_exitEquation_connect_clause;
static PyObject *__pyx_n_s_exitEquation_for;
static PyObject *__pyx_n_u_exitEquation_for;
static PyObject *__pyx_n_s_exitEquation_function;
static PyObject *__pyx_n_u_exitEquation_function;
static PyObject *__pyx_n_s_exitEquation_if;
static PyObject *__pyx_n_u_exitEquation_if;
static PyObject *__pyx_n_s_exitEquation_section;
static PyObject *__pyx_n_u_exitEquation_section;
static PyObject *__pyx_n_s_exitEquation_simple;
static PyObject *__pyx_n_u_exitEquation_simple;
static PyObject *__pyx_n_s_exitEquation_when;
static PyObject *__pyx_n_u_exitEquation_when;
static PyObject *__pyx_n_s_exitExpr_add;
static PyObject *__pyx_n_u_exitExpr_add;
static PyObject *__pyx_n_s_exitExpr_and;
static PyObject *__pyx_n_u_exitExpr_and;
static PyObject *__pyx_n_s_exitExpr_exp;
static PyObject *__pyx_n_u_exitExpr_exp;
static PyObject *__pyx_n_s_exitExpr_mul;
static PyObject *__pyx_n_u_exitExpr_mul;
static PyObject *__pyx_n_s_exitExpr_neg;
static PyObject *__pyx_n_u_exitExpr_neg;
static PyObject *__pyx_n_s_exitExpr_not;
static PyObject *__pyx_n_u_exitExpr_not;
static PyObject *__pyx_n_s_exitExpr_or;
static PyObject *__pyx_n_u_exitExpr_or;
static PyObject *__pyx_n_s_exitExpr_primary;
static PyObject *__pyx_n_u_exitExpr_primary;
static PyObject *__pyx_n_s_exitExpr_rel;
static PyObject *__pyx_n_u_exitExpr_rel;
static PyObject *__pyx_n_s_exitExpression_if;
static PyObject *__pyx_n_u_exitExpression_if;
static PyObject *__pyx_n_s_exitExpression_list;
static PyObject *__pyx_n_u_exitExpression_list;
static PyObject *__pyx_n_s_exitExpression_simple;
static PyObject *__pyx_n_u_exitExpression_simple;
static PyObject *__pyx_n_s_exitExtends_clause;
static PyObject *__pyx_n_u_exitExtends_clause;
static PyObject *__pyx_n_s_exitExternal_function_call;
static PyObject *__pyx_n_u_exitExternal_function_call;
static PyObject *__pyx_n_s_exitFor_equation;
static PyObject *__pyx_n_u_exitFor_equation;
static PyObject *__pyx_n_s_exitFor_index;
static PyObject *__pyx_n_u_exitFor_index;
static PyObject *__pyx_n_s_exitFor_indices;
static PyObject *__pyx_n_u_exitFor_indices;
static PyObject *__pyx_n_s_exitFor_statement;
static PyObject *__pyx_n_u_exitFor_statement;
static PyObject *__pyx_n_s_exitFunction_arguments;
static PyObject *__pyx_n_u_exitFunction_arguments;
static PyObject *__pyx_n_s_exitFunction_call_args;
static PyObject *__pyx_n_u_exitFunction_call_args;
static PyObject *__pyx_n_s_exitIf_equation;
static PyObject *__pyx_n_u_exitIf_equation;
static PyObject *__pyx_n_s_exitIf_statement;
static PyObject *__pyx_n_u_exitIf_statement;
static PyObject *__pyx_n_s_exitImport_clause;
static PyObject *__pyx_n_u_exitImport_clause;
static PyObject *__pyx_n_s_exitImport_list;
static PyObject *__pyx_n_u_exitImport_list;
static PyObject *__pyx_n_s_exitLanguage_specification;
static PyObject *__pyx_n_u_exitLanguage_specification;
static PyObject *__pyx_n_s_exitModification_assignment;
static PyObject *__pyx_n_u_exitModification_assignment;
static PyObject *__pyx_n_s_exitModification_assignment2;
static PyObject *__pyx_n_u_exitModification_assignment2;
static PyObject *__pyx_n_s_exitModification_class;
static PyObject *__pyx_n_u_exitModification_class;
static PyObject *__pyx_n_s_exitName;
static PyObject *__pyx_n_u_exitName;
static PyObject *__pyx_n_s_exitNamed_argument;
static PyObject *__pyx_n_u_exitNamed_argument;
static PyObject *__pyx_n_s_exitNamed_arguments;
static PyObject *__pyx_n_u_exitNamed_arguments;
static PyObject *__pyx_n_s_exitOutput_expression_list;
static PyObject *__pyx_n_u_exitOutput_expression_list;
static PyObject *__pyx_n_s_exitPrimary_component_reference;
static PyObject *__pyx_n_u_exitPrimary_component_reference;
static PyObject *__pyx_n_s_exitPrimary_derivative;
static PyObject *__pyx_n_u_exitPrimary_derivative;
static PyObject *__pyx_n_s_exitPrimary_end;
static PyObject *__pyx_n_u_exitPrimary_end;
static PyObject *__pyx_n_s_exitPrimary_expression_list;
static PyObject *__pyx_n_u_exitPrimary_expression_list;
static PyObject *__pyx_n_s_exitPrimary_false;
static PyObject *__pyx_n_u_exitPrimary_false;
static PyObject *__pyx_n_s_exitPrimary_function;
static PyObject *__pyx_n_u_exitPrimary_function;
static PyObject *__pyx_n_s_exitPrimary_function_arguments;
static PyObject *__pyx_n_u_exitPrimary_function_arguments;
static PyObject *__pyx_n_s_exitPrimary_initial;
static PyObject *__pyx_n_u_exitPrimary_initial;
static PyObject *__pyx_n_s_exitPrimary_output_expression_li;
static PyObject *__pyx_n_u_exitPrimary_output_expression_li;
static PyObject *__pyx_n_s_exitPrimary_string;
static PyObject *__pyx_n_u_exitPrimary_string;
static PyObject *__pyx_n_s_exitPrimary_true;
static PyObject *__pyx_n_u_exitPrimary_true;
static PyObject *__pyx_n_s_exitPrimary_unsigned_number;
static PyObject *__pyx_n_u_exitPrimary_unsigned_number;
static PyObject *__pyx_n_s_exitRegular_element;
static PyObject *__pyx_n_u_exitRegular_element;
static PyObject *__pyx_n_s_exitReplaceable_element;
static PyObject *__pyx_n_u_exitReplaceable_element;
static PyObject *__pyx_n_s_exitRule;
static PyObject *__pyx_n_s_exitShort_class_definition;
static PyObject *__pyx_n_u_exitShort_class_definition;
static PyObject *__pyx_n_s_exitSimple_expression;
static PyObject *__pyx_n_u_exitSimple_expression;
static PyObject *__pyx_n_s_exitStatement;
static PyObject *__pyx_n_u_exitStatement;
static PyObject *__pyx_n_s_exitStatement_break;
static PyObject *__pyx_n_u_exitStatement_break;
static PyObject *__pyx_n_s_exitStatement_component_function;
static PyObject *__pyx_n_u_exitStatement_component_function;
static PyObject *__pyx_n_s_exitStatement_component_referenc;
static PyObject *__pyx_n_u_exitStatement_component_referenc;
static PyObject *__pyx_n_s_exitStatement_for;
static PyObject *__pyx_n_u_exitStatement_for;
static PyObject *__pyx_n_s_exitStatement_if;
static PyObject *__pyx_n_u_exitStatement_if;
static PyObject *__pyx_n_s_exitStatement_return;
static PyObject *__pyx_n_u_exitStatement_return;
static PyObject *__pyx_n_s_exitStatement_when;
static PyObject *__pyx_n_u_exitStatement_when;
static PyObject *__pyx_n_s_exitStatement_while;
static PyObject *__pyx_n_u_exitStatement_while;
static PyObject *__pyx_n_s_exitStored_definition;
static PyObject *__pyx_n_u_exitStored_definition;
static PyObject *__pyx_n_s_exitStored_definition_class;
static PyObject *__pyx_n_u_exitStored_definition_class;
static PyObject *__pyx_n_s_exitString_comment;
static PyObject *__pyx_n_u_exitString_comment;
static PyObject *__pyx_n_s_exitSubscript;
static PyObject *__pyx_n_u_exitSubscript;
static PyObject *__pyx_n_s_exitType_prefix;
static PyObject *__pyx_n_u_exitType_prefix;
static PyObject *__pyx_n_s_exitType_specifier;
static PyObject *__pyx_n_u_exitType_specifier;
static PyObject *__pyx_n_s_exitWhen_equation;
static PyObject *__pyx_n_u_exitWhen_equation;
static PyObject *__pyx_n_s_exitWhen_statement;
static PyObject *__pyx_n_u_exitWhen_statement;
static PyObject *__pyx_n_s_exitWhile_statement;
static PyObject *__pyx_n_u_exitWhile_statement;
static PyObject *__pyx_kp_u_expandable;
static PyObject *__pyx_n_s_expr;
static PyObject *__pyx_n_u_expr;
static PyObject *__pyx_n_s_expr_sempred;
static PyObject *__pyx_n_s_expression;
static PyObject *__pyx_n_u_expression;
static PyObject *__pyx_n_s_expression_list;
static PyObject *__pyx_n_u_expression_list;
static PyObject *__pyx_n_s_ext_annotation;
static PyObject *__pyx_kp_u_extends;
static PyObject *__pyx_n_s_extends_clause;
static PyObject *__pyx_n_u_extends_clause;
static PyObject *__pyx_kp_u_external;
static PyObject *__pyx_n_s_external_function_call;
static PyObject *__pyx_n_u_external_function_call;
static PyObject *__pyx_kp_u_f4_j6;
static PyObject *__pyx_kp_u_false;
static PyObject *__pyx_kp_u_final;
static PyObject *__pyx_kp_u_flow;
static PyObject *__pyx_kp_u_for;
static PyObject *__pyx_n_s_for_equation;
static PyObject *__pyx_n_u_for_equation;
static PyObject *__pyx_n_s_for_index;
static PyObject *__pyx_n_u_for_index;
static PyObject *__pyx_n_s_for_indices;
static PyObject *__pyx_n_u_for_indices;
static PyObject *__pyx_n_s_for_statement;
static PyObject *__pyx_n_u_for_statement;
static PyObject *__pyx_kp_u_function;
static PyObject *__pyx_n_s_function_argument;
static PyObject *__pyx_n_u_function_argument;
static PyObject *__pyx_n_s_function_arguments;
static PyObject *__pyx_n_u_function_arguments;
static PyObject *__pyx_n_s_function_call_args;
static PyObject *__pyx_n_u_function_call_args;
static PyObject *__pyx_kp_u_g_7_n8;
static PyObject *__pyx_n_s_get;
static PyObject *__pyx_n_s_getRuleIndex;
static PyObject *__pyx_n_s_getToken;
static PyObject *__pyx_n_s_getTokens;
static PyObject *__pyx_n_s_getTypedRuleContext;
static PyObject *__pyx_n_s_getTypedRuleContexts;
static PyObject *__pyx_n_s_getvalue;
static PyObject *__pyx_n_s_grammarFileName;
static PyObject *__pyx_kp_u_h_j_l_n;
static PyObject *__pyx_kp_s_home_jgoppert_git_pymola_pymola;
static PyObject *__pyx_n_s_i;
static PyObject *__pyx_kp_u_if;
static PyObject *__pyx_n_s_if_equation;
static PyObject *__pyx_n_u_if_equation;
static PyObject *__pyx_n_s_if_statement;
static PyObject *__pyx_n_u_if_statement;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_kp_u_import_2;
static PyObject *__pyx_n_s_import_clause;
static PyObject *__pyx_n_u_import_clause;
static PyObject *__pyx_n_s_import_list;
static PyObject *__pyx_n_u_import_list;
static PyObject *__pyx_kp_u_impure;
static PyObject *__pyx_kp_u_in;
static PyObject *__pyx_n_s_init;
static PyObject *__pyx_kp_u_initial;
static PyObject *__pyx_kp_u_inner;
static PyObject *__pyx_n_s_input;
static PyObject *__pyx_n_s_input_2;
static PyObject *__pyx_kp_u_input_3;
static PyObject *__pyx_n_s_interp;
static PyObject *__pyx_n_s_invokingState;
static PyObject *__pyx_n_s_io;
static PyObject *__pyx_kp_u_l7_h5_v;
static PyObject *__pyx_n_s_la;
static PyObject *__pyx_n_s_la_2;
static PyObject *__pyx_n_s_language_specification;
static PyObject *__pyx_n_u_language_specification;
static PyObject *__pyx_n_s_listener;
static PyObject *__pyx_n_s_literalNames;
static PyObject *__pyx_n_s_localctx;
static PyObject *__pyx_kp_u_loop;
static PyObject *__pyx_kp_u_m_p9;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_match;
static PyObject *__pyx_n_s_metaclass;
static PyObject *__pyx_kp_u_model;
static PyObject *__pyx_n_s_modification;
static PyObject *__pyx_n_u_modification;
static PyObject *__pyx_n_s_module;
static PyObject *__pyx_kp_u_n8;
static PyObject *__pyx_kp_u_n8_0;
static PyObject *__pyx_kp_u_n8_0_T;
static PyObject *__pyx_kp_u_n8_0_X;
static PyObject *__pyx_kp_u_n8_0_X_2;
static PyObject *__pyx_kp_u_n8_0_n8;
static PyObject *__pyx_kp_u_n8_10;
static PyObject *__pyx_kp_u_n8_11;
static PyObject *__pyx_kp_u_n8_12;
static PyObject *__pyx_kp_u_n8_2;
static PyObject *__pyx_kp_u_n8_3;
static PyObject *__pyx_kp_u_n8_4;
static PyObject *__pyx_kp_u_n8_5;
static PyObject *__pyx_kp_u_n8_6;
static PyObject *__pyx_kp_u_n8_7;
static PyObject *__pyx_kp_u_n8_8;
static PyObject *__pyx_kp_u_n8_9;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_u_name;
static PyObject *__pyx_n_s_named_argument;
static PyObject *__pyx_n_u_named_argument;
static PyObject *__pyx_n_s_named_arguments;
static PyObject *__pyx_n_u_named_arguments;
static PyObject *__pyx_kp_u_not;
static PyObject *__pyx_kp_u_o;
static PyObject *__pyx_n_s_op;
static PyObject *__pyx_kp_u_operator;
static PyObject *__pyx_kp_u_or;
static PyObject *__pyx_kp_u_outer;
static PyObject *__pyx_n_s_output;
static PyObject *__pyx_kp_u_output_2;
static PyObject *__pyx_n_s_output_expression_list;
static PyObject *__pyx_n_u_output_expression_list;
static PyObject *__pyx_n_s_p;
static PyObject *__pyx_kp_u_p9_n8;
static PyObject *__pyx_kp_u_p_r_t_v;
static PyObject *__pyx_kp_u_package;
static PyObject *__pyx_kp_u_parameter;
static PyObject *__pyx_n_s_parent;
static PyObject *__pyx_n_s_parentState;
static PyObject *__pyx_n_s_parentctx;
static PyObject *__pyx_n_s_parseListeners;
static PyObject *__pyx_n_s_parser;
static PyObject *__pyx_kp_u_partial;
static PyObject *__pyx_n_s_precpred;
static PyObject *__pyx_n_s_pred;
static PyObject *__pyx_n_s_predIndex;
static PyObject *__pyx_n_s_predicates;
static PyObject *__pyx_n_s_prepare;
static PyObject *__pyx_n_s_prevctx;
static PyObject *__pyx_n_s_primary;
static PyObject *__pyx_n_u_primary;
static PyObject *__pyx_kp_u_protected;
static PyObject *__pyx_kp_u_public;
static PyObject *__pyx_kp_u_pure;
static PyObject *__pyx_n_s_pushNewRecursionContext;
static PyObject *__pyx_n_s_pymola_generated_ModelicaParser;
static PyObject *__pyx_kp_u_q;
static PyObject *__pyx_n_s_qualname;
static PyObject *__pyx_kp_u_r;
static PyObject *__pyx_kp_u_r_2;
static PyObject *__pyx_kp_u_r_3;
static PyObject *__pyx_kp_u_r_4;
static PyObject *__pyx_kp_u_r_5;
static PyObject *__pyx_kp_u_r_J;
static PyObject *__pyx_kp_u_r_K;
static PyObject *__pyx_kp_u_r_r;
static PyObject *__pyx_n_s_re;
static PyObject *__pyx_kp_u_record;
static PyObject *__pyx_n_s_recover;
static PyObject *__pyx_n_s_recoverInline;
static PyObject *__pyx_kp_u_redeclare;
static PyObject *__pyx_n_s_regular_element;
static PyObject *__pyx_n_u_regular_element;
static PyObject *__pyx_kp_u_replaceable;
static PyObject *__pyx_n_s_replaceable_element;
static PyObject *__pyx_n_u_replaceable_element;
static PyObject *__pyx_n_s_reportError;
static PyObject *__pyx_n_s_reportMatch;
static PyObject *__pyx_kp_u_return;
static PyObject *__pyx_n_s_ruleIndex;
static PyObject *__pyx_n_s_ruleNames;
static PyObject *__pyx_kp_u_s;
static PyObject *__pyx_n_s_self;
static PyObject *__pyx_kp_u_self_precpred_self__ctx_2;
static PyObject *__pyx_kp_u_self_precpred_self__ctx_3;
static PyObject *__pyx_kp_u_self_precpred_self__ctx_5;
static PyObject *__pyx_kp_u_self_precpred_self__ctx_6;
static PyObject *__pyx_kp_u_self_precpred_self__ctx_7;
static PyObject *__pyx_n_s_sempred;
static PyObject *__pyx_n_s_serializedATN;
static PyObject *__pyx_n_s_sharedContextCache;
static PyObject *__pyx_n_s_short_class_definition;
static PyObject *__pyx_n_u_short_class_definition;
static PyObject *__pyx_n_s_simple_expression;
static PyObject *__pyx_n_u_simple_expression;
static PyObject *__pyx_n_s_startState;
static PyObject *__pyx_n_s_state;
static PyObject *__pyx_n_s_statement;
static PyObject *__pyx_n_u_statement;
static PyObject *__pyx_n_s_statement_options;
static PyObject *__pyx_n_u_statement_options;
static PyObject *__pyx_n_s_stdout;
static PyObject *__pyx_n_s_stop;
static PyObject *__pyx_n_s_stored_definition;
static PyObject *__pyx_n_u_stored_definition;
static PyObject *__pyx_n_s_stored_definition_class;
static PyObject *__pyx_n_u_stored_definition_class;
static PyObject *__pyx_kp_u_stream;
static PyObject *__pyx_n_s_string_comment;
static PyObject *__pyx_n_u_string_comment;
static PyObject *__pyx_n_s_subscript;
static PyObject *__pyx_n_u_subscript;
static PyObject *__pyx_n_s_super;
static PyObject *__pyx_n_s_symbolicNames;
static PyObject *__pyx_n_s_sync;
static PyObject *__pyx_n_s_sys;
static PyObject *__pyx_kp_u_t;
static PyObject *__pyx_kp_u_t_t;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_kp_u_then;
static PyObject *__pyx_n_s_token;
static PyObject *__pyx_n_s_triggerExitRuleEvent;
static PyObject *__pyx_kp_u_true;
static PyObject *__pyx_kp_u_type;
static PyObject *__pyx_n_s_type_prefix;
static PyObject *__pyx_n_u_type_prefix;
static PyObject *__pyx_n_s_type_specifier;
static PyObject *__pyx_n_u_type_specifier;
static PyObject *__pyx_n_s_typing_io;
static PyObject *__pyx_kp_u_u;
static PyObject *__pyx_n_s_unrollRecursionContexts;
static PyObject *__pyx_kp_u_v;
static PyObject *__pyx_kp_u_v_2;
static PyObject *__pyx_kp_u_v_3;
static PyObject *__pyx_kp_u_v_F;
static PyObject *__pyx_n_s_visitAlgorithm_section;
static PyObject *__pyx_n_u_visitAlgorithm_section;
static PyObject *__pyx_n_s_visitAnnotation;
static PyObject *__pyx_n_u_visitAnnotation;
static PyObject *__pyx_n_s_visitArgument;
static PyObject *__pyx_n_u_visitArgument;
static PyObject *__pyx_n_s_visitArgument_expression;
static PyObject *__pyx_n_u_visitArgument_expression;
static PyObject *__pyx_n_s_visitArgument_function;
static PyObject *__pyx_n_u_visitArgument_function;
static PyObject *__pyx_n_s_visitArgument_list;
static PyObject *__pyx_n_u_visitArgument_list;
static PyObject *__pyx_n_s_visitArray_subscripts;
static PyObject *__pyx_n_u_visitArray_subscripts;
static PyObject *__pyx_n_s_visitBase_prefix;
static PyObject *__pyx_n_u_visitBase_prefix;
static PyObject *__pyx_n_s_visitChildren;
static PyObject *__pyx_n_s_visitClass_definition;
static PyObject *__pyx_n_u_visitClass_definition;
static PyObject *__pyx_n_s_visitClass_modification;
static PyObject *__pyx_n_u_visitClass_modification;
static PyObject *__pyx_n_s_visitClass_prefixes;
static PyObject *__pyx_n_u_visitClass_prefixes;
static PyObject *__pyx_n_s_visitClass_spec_base;
static PyObject *__pyx_n_u_visitClass_spec_base;
static PyObject *__pyx_n_s_visitClass_spec_comp;
static PyObject *__pyx_n_u_visitClass_spec_comp;
static PyObject *__pyx_n_s_visitClass_spec_der;
static PyObject *__pyx_n_u_visitClass_spec_der;
static PyObject *__pyx_n_s_visitClass_spec_enum;
static PyObject *__pyx_n_u_visitClass_spec_enum;
static PyObject *__pyx_n_s_visitClass_spec_extends;
static PyObject *__pyx_n_u_visitClass_spec_extends;
static PyObject *__pyx_n_s_visitClass_type;
static PyObject *__pyx_n_u_visitClass_type;
static PyObject *__pyx_n_s_visitComment;
static PyObject *__pyx_n_u_visitComment;
static PyObject *__pyx_n_s_visitComponent_clause;
static PyObject *__pyx_n_u_visitComponent_clause;
static PyObject *__pyx_n_s_visitComponent_clause1;
static PyObject *__pyx_n_u_visitComponent_clause1;
static PyObject *__pyx_n_s_visitComponent_declaration;
static PyObject *__pyx_n_u_visitComponent_declaration;
static PyObject *__pyx_n_s_visitComponent_declaration1;
static PyObject *__pyx_n_u_visitComponent_declaration1;
static PyObject *__pyx_n_s_visitComponent_list;
static PyObject *__pyx_n_u_visitComponent_list;
static PyObject *__pyx_n_s_visitComponent_reference;
static PyObject *__pyx_n_u_visitComponent_reference;
static PyObject *__pyx_n_s_visitComponent_reference_element;
static PyObject *__pyx_n_u_visitComponent_reference_element;
static PyObject *__pyx_n_s_visitComposition;
static PyObject *__pyx_n_u_visitComposition;
static PyObject *__pyx_n_s_visitCondition_attribute;
static PyObject *__pyx_n_u_visitCondition_attribute;
static PyObject *__pyx_n_s_visitConnect_clause;
static PyObject *__pyx_n_u_visitConnect_clause;
static PyObject *__pyx_n_s_visitConstraining_clause;
static PyObject *__pyx_n_u_visitConstraining_clause;
static PyObject *__pyx_n_s_visitDeclaration;
static PyObject *__pyx_n_u_visitDeclaration;
static PyObject *__pyx_n_s_visitElement;
static PyObject *__pyx_n_u_visitElement;
static PyObject *__pyx_n_s_visitElement_list;
static PyObject *__pyx_n_u_visitElement_list;
static PyObject *__pyx_n_s_visitElement_modification;
static PyObject *__pyx_n_u_visitElement_modification;
static PyObject *__pyx_n_s_visitElement_modification_or_rep;
static PyObject *__pyx_n_u_visitElement_modification_or_rep;
static PyObject *__pyx_n_s_visitElement_redeclaration;
static PyObject *__pyx_n_u_visitElement_redeclaration;
static PyObject *__pyx_n_s_visitElement_replaceable;
static PyObject *__pyx_n_u_visitElement_replaceable;
static PyObject *__pyx_n_s_visitEnum_list;
static PyObject *__pyx_n_u_visitEnum_list;
static PyObject *__pyx_n_s_visitEnumeration_literal;
static PyObject *__pyx_n_u_visitEnumeration_literal;
static PyObject *__pyx_n_s_visitEquation;
static PyObject *__pyx_n_u_visitEquation;
static PyObject *__pyx_n_s_visitEquation_connect_clause;
static PyObject *__pyx_n_u_visitEquation_connect_clause;
static PyObject *__pyx_n_s_visitEquation_for;
static PyObject *__pyx_n_u_visitEquation_for;
static PyObject *__pyx_n_s_visitEquation_function;
static PyObject *__pyx_n_u_visitEquation_function;
static PyObject *__pyx_n_s_visitEquation_if;
static PyObject *__pyx_n_u_visitEquation_if;
static PyObject *__pyx_n_s_visitEquation_section;
static PyObject *__pyx_n_u_visitEquation_section;
static PyObject *__pyx_n_s_visitEquation_simple;
static PyObject *__pyx_n_u_visitEquation_simple;
static PyObject *__pyx_n_s_visitEquation_when;
static PyObject *__pyx_n_u_visitEquation_when;
static PyObject *__pyx_n_s_visitExpr_add;
static PyObject *__pyx_n_u_visitExpr_add;
static PyObject *__pyx_n_s_visitExpr_and;
static PyObject *__pyx_n_u_visitExpr_and;
static PyObject *__pyx_n_s_visitExpr_exp;
static PyObject *__pyx_n_u_visitExpr_exp;
static PyObject *__pyx_n_s_visitExpr_mul;
static PyObject *__pyx_n_u_visitExpr_mul;
static PyObject *__pyx_n_s_visitExpr_neg;
static PyObject *__pyx_n_u_visitExpr_neg;
static PyObject *__pyx_n_s_visitExpr_not;
static PyObject *__pyx_n_u_visitExpr_not;
static PyObject *__pyx_n_s_visitExpr_or;
static PyObject *__pyx_n_u_visitExpr_or;
static PyObject *__pyx_n_s_visitExpr_primary;
static PyObject *__pyx_n_u_visitExpr_primary;
static PyObject *__pyx_n_s_visitExpr_rel;
static PyObject *__pyx_n_u_visitExpr_rel;
static PyObject *__pyx_n_s_visitExpression_if;
static PyObject *__pyx_n_u_visitExpression_if;
static PyObject *__pyx_n_s_visitExpression_list;
static PyObject *__pyx_n_u_visitExpression_list;
static PyObject *__pyx_n_s_visitExpression_simple;
static PyObject *__pyx_n_u_visitExpression_simple;
static PyObject *__pyx_n_s_visitExtends_clause;
static PyObject *__pyx_n_u_visitExtends_clause;
static PyObject *__pyx_n_s_visitExternal_function_call;
static PyObject *__pyx_n_u_visitExternal_function_call;
static PyObject *__pyx_n_s_visitFor_equation;
static PyObject *__pyx_n_u_visitFor_equation;
static PyObject *__pyx_n_s_visitFor_index;
static PyObject *__pyx_n_u_visitFor_index;
static PyObject *__pyx_n_s_visitFor_indices;
static PyObject *__pyx_n_u_visitFor_indices;
static PyObject *__pyx_n_s_visitFor_statement;
static PyObject *__pyx_n_u_visitFor_statement;
static PyObject *__pyx_n_s_visitFunction_arguments;
static PyObject *__pyx_n_u_visitFunction_arguments;
static PyObject *__pyx_n_s_visitFunction_call_args;
static PyObject *__pyx_n_u_visitFunction_call_args;
static PyObject *__pyx_n_s_visitIf_equation;
static PyObject *__pyx_n_u_visitIf_equation;
static PyObject *__pyx_n_s_visitIf_statement;
static PyObject *__pyx_n_u_visitIf_statement;
static PyObject *__pyx_n_s_visitImport_clause;
static PyObject *__pyx_n_u_visitImport_clause;
static PyObject *__pyx_n_s_visitImport_list;
static PyObject *__pyx_n_u_visitImport_list;
static PyObject *__pyx_n_s_visitLanguage_specification;
static PyObject *__pyx_n_u_visitLanguage_specification;
static PyObject *__pyx_n_s_visitModification_assignment;
static PyObject *__pyx_n_u_visitModification_assignment;
static PyObject *__pyx_n_s_visitModification_assignment2;
static PyObject *__pyx_n_u_visitModification_assignment2;
static PyObject *__pyx_n_s_visitModification_class;
static PyObject *__pyx_n_u_visitModification_class;
static PyObject *__pyx_n_s_visitName;
static PyObject *__pyx_n_u_visitName;
static PyObject *__pyx_n_s_visitNamed_argument;
static PyObject *__pyx_n_u_visitNamed_argument;
static PyObject *__pyx_n_s_visitNamed_arguments;
static PyObject *__pyx_n_u_visitNamed_arguments;
static PyObject *__pyx_n_s_visitOutput_expression_list;
static PyObject *__pyx_n_u_visitOutput_expression_list;
static PyObject *__pyx_n_s_visitPrimary_component_reference;
static PyObject *__pyx_n_u_visitPrimary_component_reference;
static PyObject *__pyx_n_s_visitPrimary_derivative;
static PyObject *__pyx_n_u_visitPrimary_derivative;
static PyObject *__pyx_n_s_visitPrimary_end;
static PyObject *__pyx_n_u_visitPrimary_end;
static PyObject *__pyx_n_s_visitPrimary_expression_list;
static PyObject *__pyx_n_u_visitPrimary_expression_list;
static PyObject *__pyx_n_s_visitPrimary_false;
static PyObject *__pyx_n_u_visitPrimary_false;
static PyObject *__pyx_n_s_visitPrimary_function;
static PyObject *__pyx_n_u_visitPrimary_function;
static PyObject *__pyx_n_s_visitPrimary_function_arguments;
static PyObject *__pyx_n_u_visitPrimary_function_arguments;
static PyObject *__pyx_n_s_visitPrimary_initial;
static PyObject *__pyx_n_u_visitPrimary_initial;
static PyObject *__pyx_n_s_visitPrimary_output_expression_l;
static PyObject *__pyx_n_u_visitPrimary_output_expression_l;
static PyObject *__pyx_n_s_visitPrimary_string;
static PyObject *__pyx_n_u_visitPrimary_string;
static PyObject *__pyx_n_s_visitPrimary_true;
static PyObject *__pyx_n_u_visitPrimary_true;
static PyObject *__pyx_n_s_visitPrimary_unsigned_number;
static PyObject *__pyx_n_u_visitPrimary_unsigned_number;
static PyObject *__pyx_n_s_visitRegular_element;
static PyObject *__pyx_n_u_visitRegular_element;
static PyObject *__pyx_n_s_visitReplaceable_element;
static PyObject *__pyx_n_u_visitReplaceable_element;
static PyObject *__pyx_n_s_visitShort_class_definition;
static PyObject *__pyx_n_u_visitShort_class_definition;
static PyObject *__pyx_n_s_visitSimple_expression;
static PyObject *__pyx_n_u_visitSimple_expression;
static PyObject *__pyx_n_s_visitStatement;
static PyObject *__pyx_n_u_visitStatement;
static PyObject *__pyx_n_s_visitStatement_break;
static PyObject *__pyx_n_u_visitStatement_break;
static PyObject *__pyx_n_s_visitStatement_component_functio;
static PyObject *__pyx_n_u_visitStatement_component_functio;
static PyObject *__pyx_n_s_visitStatement_component_referen;
static PyObject *__pyx_n_u_visitStatement_component_referen;
static PyObject *__pyx_n_s_visitStatement_for;
static PyObject *__pyx_n_u_visitStatement_for;
static PyObject *__pyx_n_s_visitStatement_if;
static PyObject *__pyx_n_u_visitStatement_if;
static PyObject *__pyx_n_s_visitStatement_return;
static PyObject *__pyx_n_u_visitStatement_return;
static PyObject *__pyx_n_s_visitStatement_when;
static PyObject *__pyx_n_u_visitStatement_when;
static PyObject *__pyx_n_s_visitStatement_while;
static PyObject *__pyx_n_u_visitStatement_while;
static PyObject *__pyx_n_s_visitStored_definition;
static PyObject *__pyx_n_u_visitStored_definition;
static PyObject *__pyx_n_s_visitStored_definition_class;
static PyObject *__pyx_n_u_visitStored_definition_class;
static PyObject *__pyx_n_s_visitString_comment;
static PyObject *__pyx_n_u_visitString_comment;
static PyObject *__pyx_n_s_visitSubscript;
static PyObject *__pyx_n_u_visitSubscript;
static PyObject *__pyx_n_s_visitType_prefix;
static PyObject *__pyx_n_u_visitType_prefix;
static PyObject *__pyx_n_s_visitType_specifier;
static PyObject *__pyx_n_u_visitType_specifier;
static PyObject *__pyx_n_s_visitWhen_equation;
static PyObject *__pyx_n_u_visitWhen_equation;
static PyObject *__pyx_n_s_visitWhen_statement;
static PyObject *__pyx_n_u_visitWhen_statement;
static PyObject *__pyx_n_s_visitWhile_statement;
static PyObject *__pyx_n_u_visitWhile_statement;
static PyObject *__pyx_n_s_visitor;
static PyObject *__pyx_kp_u_w;
static PyObject *__pyx_kp_u_when;
static PyObject *__pyx_n_s_when_equation;
static PyObject *__pyx_n_u_when_equation;
static PyObject *__pyx_n_s_when_statement;
static PyObject *__pyx_n_u_when_statement;
static PyObject *__pyx_kp_u_while;
static PyObject *__pyx_n_s_while_statement;
static PyObject *__pyx_n_u_while_statement;
static PyObject *__pyx_kp_u_within;
static PyObject *__pyx_n_s_write;
static PyObject *__pyx_kp_u_x_x;
static PyObject *__pyx_kp_u_x_z;
static PyObject *__pyx_kp_u_y;
static PyObject *__pyx_kp_u_z;
static PyObject *__pyx_kp_u_z_10;
static PyObject *__pyx_kp_u_z_2;
static PyObject *__pyx_kp_u_z_3;
static PyObject *__pyx_kp_u_z_4;
static PyObject *__pyx_kp_u_z_5;
static PyObject *__pyx_kp_u_z_6;
static PyObject *__pyx_kp_u_z_7;
static PyObject *__pyx_kp_u_z_8;
static PyObject *__pyx_kp_u_z_9;
static PyObject *__pyx_kp_u_z_F;
static PyObject *__pyx_kp_u_z_n8;
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_serializedATN(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_2__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_input, PyObject *__pyx_v_output); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Stored_definitionContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_2WITHIN(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_4stored_definition_class(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_6component_reference(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_8getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_10enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_12exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_14accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_2stored_definition(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_2class_definition(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_4FINAL(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_4stored_definition_class(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Class_definitionContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Class_definitionContext_2class_prefixes(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Class_definitionContext_4class_specifier(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Class_definitionContext_6ENCAPSULATED(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Class_definitionContext_8getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Class_definitionContext_10enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Class_definitionContext_12exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Class_definitionContext_14accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_6class_definition(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Class_prefixesContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_2class_type(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_4PARTIAL(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_8class_prefixes(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_17Class_typeContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_17Class_typeContext_2getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_17Class_typeContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_17Class_typeContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_17Class_typeContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_10class_type(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_specifierContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_specifierContext_2getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_specifierContext_4copyFrom(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Class_spec_derContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_2IDENT(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_4name(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_6comment(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_2IDENT(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_4comment(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_6enum_list(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_2IDENT(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_4base_prefix(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_6name(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_8comment(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_10array_subscripts(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_12class_modification(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_14enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_16exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_18accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_compContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_2IDENT(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_4string_comment(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_6composition(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_2IDENT(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_4string_comment(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_6composition(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_8class_modification(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_10enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_12exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_14accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_12class_specifier(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Base_prefixContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Base_prefixContext_2type_prefix(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Base_prefixContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Base_prefixContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Base_prefixContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Base_prefixContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_14base_prefix(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16Enum_listContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16Enum_listContext_2enumeration_literal(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16Enum_listContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16Enum_listContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16Enum_listContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16Enum_listContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_16enum_list(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_2IDENT(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_4comment(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_18enumeration_literal(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18CompositionContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18CompositionContext_2element_list(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18CompositionContext_4equation_section(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18CompositionContext_6algorithm_section(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18CompositionContext_8annotation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18CompositionContext_10language_specification(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18CompositionContext_12external_function_call(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18CompositionContext_14getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18CompositionContext_16enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18CompositionContext_18exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18CompositionContext_20accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_20composition(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Language_specificationContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Language_specificationContext_2STRING(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Language_specificationContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Language_specificationContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Language_specificationContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Language_specificationContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_22language_specification(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29External_function_callContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29External_function_callContext_2IDENT(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29External_function_callContext_4component_reference(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29External_function_callContext_6expression_list(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29External_function_callContext_8getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29External_function_callContext_10enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29External_function_callContext_12exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29External_function_callContext_14accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_24external_function_call(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Element_listContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Element_listContext_2element(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Element_listContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Element_listContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Element_listContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Element_listContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_26element_list(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ElementContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ElementContext_2import_clause(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ElementContext_4extends_clause(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ElementContext_6regular_element(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ElementContext_8replaceable_element(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ElementContext_10getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ElementContext_12enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ElementContext_14exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ElementContext_16accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_28element(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Regular_elementContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Regular_elementContext_2REDECLARE(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Regular_elementContext_4FINAL(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Regular_elementContext_6INNER(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Regular_elementContext_8OUTER(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Regular_elementContext_10class_definition(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Regular_elementContext_12component_clause(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Regular_elementContext_14getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Regular_elementContext_16enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Regular_elementContext_18exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Regular_elementContext_20accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_30regular_element(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_2REDECLARE(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_4FINAL(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_6INNER(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_8OUTER(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_10class_definition(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_12component_clause(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_14constraining_clause(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_16comment(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_18getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_20enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_22exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_24accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_32replaceable_element(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Import_clauseContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Import_clauseContext_2comment(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Import_clauseContext_4IDENT(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Import_clauseContext_6component_reference(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Import_clauseContext_8import_list(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Import_clauseContext_10getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Import_clauseContext_12enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Import_clauseContext_14exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Import_clauseContext_16accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_34import_clause(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Import_listContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Import_listContext_2IDENT(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Import_listContext_4import_list(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Import_listContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Import_listContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Import_listContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Import_listContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_36import_list(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Extends_clauseContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_2component_reference(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_4class_modification(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_6annotation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_8getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_10enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_12exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_14accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_38extends_clause(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_2name(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_4class_modification(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_40constraining_clause(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Component_clauseContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Component_clauseContext_2type_prefix(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Component_clauseContext_4type_specifier(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Component_clauseContext_6component_list(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Component_clauseContext_8array_subscripts(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Component_clauseContext_10getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Component_clauseContext_12enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Component_clauseContext_14exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Component_clauseContext_16accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_42component_clause(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Type_prefixContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Type_prefixContext_2getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Type_prefixContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Type_prefixContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Type_prefixContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_44type_prefix(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Type_specifierContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Type_specifierContext_2component_reference(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Type_specifierContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Type_specifierContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Type_specifierContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Type_specifierContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_46type_specifier(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Component_listContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Component_listContext_2component_declaration(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Component_listContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Component_listContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Component_listContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Component_listContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_48component_list(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_28Component_declarationContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_28Component_declarationContext_2declaration(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_28Component_declarationContext_4comment(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_28Component_declarationContext_6condition_attribute(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_28Component_declarationContext_8getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_28Component_declarationContext_10enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_28Component_declarationContext_12exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_28Component_declarationContext_14accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_50component_declaration(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Condition_attributeContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_2expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_52condition_attribute(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18DeclarationContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18DeclarationContext_2IDENT(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18DeclarationContext_4array_subscripts(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18DeclarationContext_6modification(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18DeclarationContext_8getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18DeclarationContext_10enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18DeclarationContext_12exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18DeclarationContext_14accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_54declaration(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19ModificationContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19ModificationContext_2getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19ModificationContext_4copyFrom(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Modification_classContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Modification_classContext_2class_modification(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Modification_classContext_4expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Modification_classContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Modification_classContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Modification_classContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_31Modification_assignment2Context___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_31Modification_assignment2Context_2expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_31Modification_assignment2Context_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_31Modification_assignment2Context_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_31Modification_assignment2Context_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Modification_assignmentContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Modification_assignmentContext_2expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Modification_assignmentContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Modification_assignmentContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Modification_assignmentContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_56modification(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Class_modificationContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Class_modificationContext_2argument_list(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Class_modificationContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Class_modificationContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Class_modificationContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Class_modificationContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_58class_modification(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Argument_listContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Argument_listContext_2argument(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Argument_listContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Argument_listContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Argument_listContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Argument_listContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_60argument_list(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15ArgumentContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15ArgumentContext_2element_modification_or_replaceable(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15ArgumentContext_4element_redeclaration(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15ArgumentContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15ArgumentContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15ArgumentContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15ArgumentContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_62argument(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_2element_modification(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_4element_replaceable(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_6EACH(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_8FINAL(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_10getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_12enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_14exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_16accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_64element_modification_or_replaceable(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_27Element_modificationContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_27Element_modificationContext_2component_reference(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_27Element_modificationContext_4string_comment(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_27Element_modificationContext_6modification(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_27Element_modificationContext_8getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_27Element_modificationContext_10enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_27Element_modificationContext_12exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_27Element_modificationContext_14accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_66element_modification(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_2REDECLARE(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_4element_replaceable(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_6EACH(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_8FINAL(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_10short_class_definition(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_12component_clause1(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_14getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_16enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_18exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_20accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_68element_redeclaration(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Element_replaceableContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_2short_class_definition(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_4component_clause1(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_6constraining_clause(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_8getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_10enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_12exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_14accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_70element_replaceable(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Component_clause1Context___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Component_clause1Context_2type_prefix(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Component_clause1Context_4type_specifier(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Component_clause1Context_6component_declaration1(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Component_clause1Context_8getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Component_clause1Context_10enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Component_clause1Context_12exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Component_clause1Context_14accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_72component_clause1(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Component_declaration1Context___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_2declaration(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_4comment(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_74component_declaration1(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_2class_prefixes(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_4IDENT(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_6base_prefix(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_8component_reference(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_10comment(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_12array_subscripts(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_14class_modification(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_16enum_list(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_18getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_20enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_22exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_24accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_76short_class_definition(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Equation_sectionContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_2INITIAL(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_4equation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_78equation_section(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_2INITIAL(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_4statement(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_80algorithm_section(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Equation_optionsContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Equation_optionsContext_2getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Equation_optionsContext_4copyFrom(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Equation_whenContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Equation_whenContext_2when_equation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Equation_whenContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Equation_whenContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Equation_whenContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Equation_connect_clauseContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Equation_connect_clauseContext_2connect_clause(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Equation_connect_clauseContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Equation_connect_clauseContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Equation_connect_clauseContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Equation_forContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Equation_forContext_2for_equation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Equation_forContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Equation_forContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Equation_forContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Equation_functionContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Equation_functionContext_2name(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Equation_functionContext_4function_call_args(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Equation_functionContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Equation_functionContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Equation_functionContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Equation_simpleContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_2simple_expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_4expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Equation_ifContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Equation_ifContext_2if_equation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Equation_ifContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Equation_ifContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Equation_ifContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_82equation_options(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15EquationContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15EquationContext_2equation_options(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15EquationContext_4comment(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15EquationContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15EquationContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15EquationContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15EquationContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_84equation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Statement_optionsContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Statement_optionsContext_2getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Statement_optionsContext_4copyFrom(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Statement_breakContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Statement_breakContext_2enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Statement_breakContext_4exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Statement_breakContext_6accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Statement_whileContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Statement_whileContext_2while_statement(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Statement_whileContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Statement_whileContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Statement_whileContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_2component_reference(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_4function_call_args(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Statement_returnContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Statement_returnContext_2enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Statement_returnContext_4exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Statement_returnContext_6accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Statement_forContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Statement_forContext_2for_statement(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Statement_forContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Statement_forContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Statement_forContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Statement_whenContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Statement_whenContext_2when_statement(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Statement_whenContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Statement_whenContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Statement_whenContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Statement_ifContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Statement_ifContext_2if_statement(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Statement_ifContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Statement_ifContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Statement_ifContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_2component_reference(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_4expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_6function_call_args(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_86statement_options(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16StatementContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16StatementContext_2statement_options(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16StatementContext_4comment(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16StatementContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16StatementContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16StatementContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16StatementContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_88statement(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18If_equationContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18If_equationContext_2expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18If_equationContext_4equation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18If_equationContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18If_equationContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18If_equationContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18If_equationContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_90if_equation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19If_statementContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19If_statementContext_2expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19If_statementContext_4statement(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19If_statementContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19If_statementContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19If_statementContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19If_statementContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_92if_statement(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19For_equationContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19For_equationContext_2for_indices(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19For_equationContext_4equation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19For_equationContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19For_equationContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19For_equationContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19For_equationContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_94for_equation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20For_statementContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20For_statementContext_2for_indices(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20For_statementContext_4statement(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20For_statementContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20For_statementContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20For_statementContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20For_statementContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_96for_statement(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18For_indicesContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18For_indicesContext_2for_index(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18For_indicesContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18For_indicesContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18For_indicesContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18For_indicesContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_98for_indices(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16For_indexContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16For_indexContext_2IDENT(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16For_indexContext_4expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16For_indexContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16For_indexContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16For_indexContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16For_indexContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_100for_index(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22While_statementContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22While_statementContext_2expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22While_statementContext_4statement(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22While_statementContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22While_statementContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22While_statementContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22While_statementContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_102while_statement(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20When_equationContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20When_equationContext_2expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20When_equationContext_4equation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20When_equationContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20When_equationContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20When_equationContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20When_equationContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_104when_equation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21When_statementContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21When_statementContext_2expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21When_statementContext_4statement(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21When_statementContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21When_statementContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21When_statementContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21When_statementContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_106when_statement(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Connect_clauseContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_2component_reference(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_108connect_clause(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_17ExpressionContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_17ExpressionContext_2getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_17ExpressionContext_4copyFrom(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Expression_ifContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Expression_ifContext_2expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Expression_ifContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Expression_ifContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Expression_ifContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Expression_simpleContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Expression_simpleContext_2simple_expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Expression_simpleContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Expression_simpleContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Expression_simpleContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_110expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Simple_expressionContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_2expr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_112simple_expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_11ExprContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_11ExprContext_2getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_11ExprContext_4copyFrom(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_negContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_negContext_2expr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_negContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_negContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_negContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_addContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_addContext_2expr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_addContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_addContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_addContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_expContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_expContext_2primary(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_expContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_expContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_expContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14Expr_orContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14Expr_orContext_2expr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14Expr_orContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14Expr_orContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14Expr_orContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Expr_primaryContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Expr_primaryContext_2primary(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Expr_primaryContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Expr_primaryContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Expr_primaryContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_andContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_andContext_2expr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_andContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_andContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_andContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_relContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_relContext_2expr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_relContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_relContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_relContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_notContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_notContext_2expr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_notContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_notContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_notContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_mulContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_mulContext_2expr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_mulContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_mulContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_mulContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_114expr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v__p); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14PrimaryContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14PrimaryContext_2getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14PrimaryContext_4copyFrom(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Primary_stringContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Primary_stringContext_2STRING(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Primary_stringContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Primary_stringContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Primary_stringContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Primary_endContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Primary_endContext_2enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Primary_endContext_4exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Primary_endContext_6accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_37Primary_output_expression_listContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_37Primary_output_expression_listContext_2output_expression_list(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_37Primary_output_expression_listContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_37Primary_output_expression_listContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_37Primary_output_expression_listContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Primary_unsigned_numberContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Primary_unsigned_numberContext_2UNSIGNED_NUMBER(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Primary_unsigned_numberContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Primary_unsigned_numberContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Primary_unsigned_numberContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_33Primary_function_argumentsContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_33Primary_function_argumentsContext_2function_arguments(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_33Primary_function_argumentsContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_33Primary_function_argumentsContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_33Primary_function_argumentsContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Primary_falseContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Primary_falseContext_2enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Primary_falseContext_4exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Primary_falseContext_6accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Primary_derivativeContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Primary_derivativeContext_2function_call_args(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Primary_derivativeContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Primary_derivativeContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Primary_derivativeContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_34Primary_component_referenceContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_34Primary_component_referenceContext_2component_reference(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_34Primary_component_referenceContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_34Primary_component_referenceContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_34Primary_component_referenceContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Primary_expression_listContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Primary_expression_listContext_2expression_list(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Primary_expression_listContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Primary_expression_listContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Primary_expression_listContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Primary_trueContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Primary_trueContext_2enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Primary_trueContext_4exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Primary_trueContext_6accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Primary_functionContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Primary_functionContext_2component_reference(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Primary_functionContext_4function_call_args(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Primary_functionContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Primary_functionContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Primary_functionContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Primary_initialContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Primary_initialContext_2function_call_args(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Primary_initialContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Primary_initialContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Primary_initialContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_116primary(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_11NameContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_11NameContext_2IDENT(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_11NameContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_11NameContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_11NameContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_11NameContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_118name(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_2IDENT(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_4array_subscripts(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_120component_reference_element(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Component_referenceContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Component_referenceContext_2component_reference_element(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Component_referenceContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Component_referenceContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Component_referenceContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Component_referenceContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_122component_reference(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Function_call_argsContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_2function_arguments(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_124function_call_args(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Function_argumentsContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_2function_argument(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_4for_indices(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_6named_arguments(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_8getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_10enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_12exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_14accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_126function_arguments(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Named_argumentsContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_2named_argument(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_128named_arguments(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Named_argumentContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Named_argumentContext_2IDENT(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Named_argumentContext_4function_argument(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Named_argumentContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Named_argumentContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Named_argumentContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Named_argumentContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_130named_argument(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Function_argumentContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Function_argumentContext_2getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Function_argumentContext_4copyFrom(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Argument_expressionContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Argument_expressionContext_2expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Argument_expressionContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Argument_expressionContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Argument_expressionContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Argument_functionContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Argument_functionContext_2name(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Argument_functionContext_4named_arguments(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Argument_functionContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Argument_functionContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Argument_functionContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_132function_argument(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Output_expression_listContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_2expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_134output_expression_list(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Expression_listContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Expression_listContext_2expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Expression_listContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Expression_listContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Expression_listContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Expression_listContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_136expression_list(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_2subscript(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_138array_subscripts(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16SubscriptContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16SubscriptContext_2expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16SubscriptContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16SubscriptContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16SubscriptContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16SubscriptContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_140subscript(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14CommentContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14CommentContext_2string_comment(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14CommentContext_4annotation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14CommentContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14CommentContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14CommentContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14CommentContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_142comment(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21String_commentContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21String_commentContext_2STRING(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21String_commentContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21String_commentContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21String_commentContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21String_commentContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_144string_comment(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_17AnnotationContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_17AnnotationContext_2class_modification(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_17AnnotationContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_17AnnotationContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_17AnnotationContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_17AnnotationContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_146annotation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_148sempred(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_localctx, PyObject *__pyx_v_ruleIndex, PyObject *__pyx_v_predIndex); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_150expr_sempred(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_localctx, PyObject *__pyx_v_predIndex); /* proto */
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_3;
static PyObject *__pyx_int_4;
static PyObject *__pyx_int_5;
static PyObject *__pyx_int_6;
static PyObject *__pyx_int_7;
static PyObject *__pyx_int_8;
static PyObject *__pyx_int_9;
static PyObject *__pyx_int_10;
static PyObject *__pyx_int_11;
static PyObject *__pyx_int_12;
static PyObject *__pyx_int_13;
static PyObject *__pyx_int_14;
static PyObject *__pyx_int_15;
static PyObject *__pyx_int_16;
static PyObject *__pyx_int_17;
static PyObject *__pyx_int_18;
static PyObject *__pyx_int_19;
static PyObject *__pyx_int_20;
static PyObject *__pyx_int_21;
static PyObject *__pyx_int_22;
static PyObject *__pyx_int_23;
static PyObject *__pyx_int_24;
static PyObject *__pyx_int_25;
static PyObject *__pyx_int_26;
static PyObject *__pyx_int_27;
static PyObject *__pyx_int_28;
static PyObject *__pyx_int_29;
static PyObject *__pyx_int_30;
static PyObject *__pyx_int_31;
static PyObject *__pyx_int_32;
static PyObject *__pyx_int_33;
static PyObject *__pyx_int_34;
static PyObject *__pyx_int_35;
static PyObject *__pyx_int_36;
static PyObject *__pyx_int_37;
static PyObject *__pyx_int_38;
static PyObject *__pyx_int_39;
static PyObject *__pyx_int_40;
static PyObject *__pyx_int_41;
static PyObject *__pyx_int_42;
static PyObject *__pyx_int_43;
static PyObject *__pyx_int_44;
static PyObject *__pyx_int_45;
static PyObject *__pyx_int_46;
static PyObject *__pyx_int_47;
static PyObject *__pyx_int_48;
static PyObject *__pyx_int_49;
static PyObject *__pyx_int_50;
static PyObject *__pyx_int_51;
static PyObject *__pyx_int_52;
static PyObject *__pyx_int_53;
static PyObject *__pyx_int_54;
static PyObject *__pyx_int_55;
static PyObject *__pyx_int_56;
static PyObject *__pyx_int_57;
static PyObject *__pyx_int_58;
static PyObject *__pyx_int_59;
static PyObject *__pyx_int_60;
static PyObject *__pyx_int_61;
static PyObject *__pyx_int_62;
static PyObject *__pyx_int_63;
static PyObject *__pyx_int_64;
static PyObject *__pyx_int_65;
static PyObject *__pyx_int_66;
static PyObject *__pyx_int_67;
static PyObject *__pyx_int_68;
static PyObject *__pyx_int_69;
static PyObject *__pyx_int_70;
static PyObject *__pyx_int_71;
static PyObject *__pyx_int_72;
static PyObject *__pyx_int_73;
static PyObject *__pyx_int_74;
static PyObject *__pyx_int_75;
static PyObject *__pyx_int_76;
static PyObject *__pyx_int_77;
static PyObject *__pyx_int_78;
static PyObject *__pyx_int_79;
static PyObject *__pyx_int_80;
static PyObject *__pyx_int_81;
static PyObject *__pyx_int_82;
static PyObject *__pyx_int_83;
static PyObject *__pyx_int_84;
static PyObject *__pyx_int_85;
static PyObject *__pyx_int_86;
static PyObject *__pyx_int_87;
static PyObject *__pyx_int_88;
static PyObject *__pyx_int_89;
static PyObject *__pyx_int_90;
static PyObject *__pyx_int_91;
static PyObject *__pyx_int_92;
static PyObject *__pyx_int_94;
static PyObject *__pyx_int_96;
static PyObject *__pyx_int_98;
static PyObject *__pyx_int_99;
static PyObject *__pyx_int_100;
static PyObject *__pyx_int_102;
static PyObject *__pyx_int_104;
static PyObject *__pyx_int_106;
static PyObject *__pyx_int_108;
static PyObject *__pyx_int_109;
static PyObject *__pyx_int_110;
static PyObject *__pyx_int_111;
static PyObject *__pyx_int_112;
static PyObject *__pyx_int_113;
static PyObject *__pyx_int_114;
static PyObject *__pyx_int_116;
static PyObject *__pyx_int_117;
static PyObject *__pyx_int_118;
static PyObject *__pyx_int_120;
static PyObject *__pyx_int_121;
static PyObject *__pyx_int_122;
static PyObject *__pyx_int_124;
static PyObject *__pyx_int_126;
static PyObject *__pyx_int_128;
static PyObject *__pyx_int_130;
static PyObject *__pyx_int_132;
static PyObject *__pyx_int_134;
static PyObject *__pyx_int_136;
static PyObject *__pyx_int_138;
static PyObject *__pyx_int_140;
static PyObject *__pyx_int_142;
static PyObject *__pyx_int_144;
static PyObject *__pyx_int_146;
static PyObject *__pyx_int_147;
static PyObject *__pyx_int_148;
static PyObject *__pyx_int_150;
static PyObject *__pyx_int_151;
static PyObject *__pyx_int_153;
static PyObject *__pyx_int_156;
static PyObject *__pyx_int_158;
static PyObject *__pyx_int_159;
static PyObject *__pyx_int_160;
static PyObject *__pyx_int_162;
static PyObject *__pyx_int_163;
static PyObject *__pyx_int_165;
static PyObject *__pyx_int_166;
static PyObject *__pyx_int_168;
static PyObject *__pyx_int_169;
static PyObject *__pyx_int_171;
static PyObject *__pyx_int_172;
static PyObject *__pyx_int_174;
static PyObject *__pyx_int_176;
static PyObject *__pyx_int_177;
static PyObject *__pyx_int_178;
static PyObject *__pyx_int_179;
static PyObject *__pyx_int_181;
static PyObject *__pyx_int_182;
static PyObject *__pyx_int_183;
static PyObject *__pyx_int_184;
static PyObject *__pyx_int_186;
static PyObject *__pyx_int_187;
static PyObject *__pyx_int_188;
static PyObject *__pyx_int_189;
static PyObject *__pyx_int_190;
static PyObject *__pyx_int_192;
static PyObject *__pyx_int_193;
static PyObject *__pyx_int_195;
static PyObject *__pyx_int_196;
static PyObject *__pyx_int_197;
static PyObject *__pyx_int_199;
static PyObject *__pyx_int_200;
static PyObject *__pyx_int_201;
static PyObject *__pyx_int_202;
static PyObject *__pyx_int_203;
static PyObject *__pyx_int_205;
static PyObject *__pyx_int_206;
static PyObject *__pyx_int_207;
static PyObject *__pyx_int_208;
static PyObject *__pyx_int_209;
static PyObject *__pyx_int_210;
static PyObject *__pyx_int_212;
static PyObject *__pyx_int_213;
static PyObject *__pyx_int_215;
static PyObject *__pyx_int_217;
static PyObject *__pyx_int_218;
static PyObject *__pyx_int_219;
static PyObject *__pyx_int_220;
static PyObject *__pyx_int_221;
static PyObject *__pyx_int_222;
static PyObject *__pyx_int_224;
static PyObject *__pyx_int_225;
static PyObject *__pyx_int_227;
static PyObject *__pyx_int_228;
static PyObject *__pyx_int_229;
static PyObject *__pyx_int_230;
static PyObject *__pyx_int_231;
static PyObject *__pyx_int_232;
static PyObject *__pyx_int_233;
static PyObject *__pyx_int_234;
static PyObject *__pyx_int_235;
static PyObject *__pyx_int_236;
static PyObject *__pyx_int_237;
static PyObject *__pyx_int_240;
static PyObject *__pyx_int_242;
static PyObject *__pyx_int_243;
static PyObject *__pyx_int_244;
static PyObject *__pyx_int_246;
static PyObject *__pyx_int_247;
static PyObject *__pyx_int_248;
static PyObject *__pyx_int_249;
static PyObject *__pyx_int_251;
static PyObject *__pyx_int_252;
static PyObject *__pyx_int_253;
static PyObject *__pyx_int_254;
static PyObject *__pyx_int_256;
static PyObject *__pyx_int_258;
static PyObject *__pyx_int_260;
static PyObject *__pyx_int_261;
static PyObject *__pyx_int_262;
static PyObject *__pyx_int_265;
static PyObject *__pyx_int_267;
static PyObject *__pyx_int_268;
static PyObject *__pyx_int_269;
static PyObject *__pyx_int_271;
static PyObject *__pyx_int_272;
static PyObject *__pyx_int_273;
static PyObject *__pyx_int_274;
static PyObject *__pyx_int_275;
static PyObject *__pyx_int_276;
static PyObject *__pyx_int_277;
static PyObject *__pyx_int_278;
static PyObject *__pyx_int_280;
static PyObject *__pyx_int_282;
static PyObject *__pyx_int_283;
static PyObject *__pyx_int_284;
static PyObject *__pyx_int_285;
static PyObject *__pyx_int_287;
static PyObject *__pyx_int_288;
static PyObject *__pyx_int_290;
static PyObject *__pyx_int_291;
static PyObject *__pyx_int_293;
static PyObject *__pyx_int_294;
static PyObject *__pyx_int_296;
static PyObject *__pyx_int_297;
static PyObject *__pyx_int_299;
static PyObject *__pyx_int_301;
static PyObject *__pyx_int_303;
static PyObject *__pyx_int_304;
static PyObject *__pyx_int_306;
static PyObject *__pyx_int_308;
static PyObject *__pyx_int_309;
static PyObject *__pyx_int_310;
static PyObject *__pyx_int_311;
static PyObject *__pyx_int_313;
static PyObject *__pyx_int_315;
static PyObject *__pyx_int_316;
static PyObject *__pyx_int_320;
static PyObject *__pyx_int_322;
static PyObject *__pyx_int_323;
static PyObject *__pyx_int_324;
static PyObject *__pyx_int_325;
static PyObject *__pyx_int_326;
static PyObject *__pyx_int_327;
static PyObject *__pyx_int_329;
static PyObject *__pyx_int_330;
static PyObject *__pyx_int_332;
static PyObject *__pyx_int_333;
static PyObject *__pyx_int_335;
static PyObject *__pyx_int_336;
static PyObject *__pyx_int_338;
static PyObject *__pyx_int_339;
static PyObject *__pyx_int_341;
static PyObject *__pyx_int_342;
static PyObject *__pyx_int_343;
static PyObject *__pyx_int_345;
static PyObject *__pyx_int_346;
static PyObject *__pyx_int_348;
static PyObject *__pyx_int_349;
static PyObject *__pyx_int_351;
static PyObject *__pyx_int_352;
static PyObject *__pyx_int_354;
static PyObject *__pyx_int_355;
static PyObject *__pyx_int_357;
static PyObject *__pyx_int_358;
static PyObject *__pyx_int_359;
static PyObject *__pyx_int_360;
static PyObject *__pyx_int_362;
static PyObject *__pyx_int_363;
static PyObject *__pyx_int_365;
static PyObject *__pyx_int_367;
static PyObject *__pyx_int_368;
static PyObject *__pyx_int_369;
static PyObject *__pyx_int_370;
static PyObject *__pyx_int_371;
static PyObject *__pyx_int_372;
static PyObject *__pyx_int_373;
static PyObject *__pyx_int_374;
static PyObject *__pyx_int_375;
static PyObject *__pyx_int_376;
static PyObject *__pyx_int_378;
static PyObject *__pyx_int_380;
static PyObject *__pyx_int_382;
static PyObject *__pyx_int_384;
static PyObject *__pyx_int_386;
static PyObject *__pyx_int_387;
static PyObject *__pyx_int_388;
static PyObject *__pyx_int_391;
static PyObject *__pyx_int_393;
static PyObject *__pyx_int_394;
static PyObject *__pyx_int_395;
static PyObject *__pyx_int_396;
static PyObject *__pyx_int_397;
static PyObject *__pyx_int_399;
static PyObject *__pyx_int_400;
static PyObject *__pyx_int_402;
static PyObject *__pyx_int_403;
static PyObject *__pyx_int_404;
static PyObject *__pyx_int_405;
static PyObject *__pyx_int_407;
static PyObject *__pyx_int_408;
static PyObject *__pyx_int_409;
static PyObject *__pyx_int_410;
static PyObject *__pyx_int_412;
static PyObject *__pyx_int_414;
static PyObject *__pyx_int_415;
static PyObject *__pyx_int_417;
static PyObject *__pyx_int_418;
static PyObject *__pyx_int_420;
static PyObject *__pyx_int_421;
static PyObject *__pyx_int_423;
static PyObject *__pyx_int_425;
static PyObject *__pyx_int_426;
static PyObject *__pyx_int_427;
static PyObject *__pyx_int_430;
static PyObject *__pyx_int_432;
static PyObject *__pyx_int_433;
static PyObject *__pyx_int_434;
static PyObject *__pyx_int_435;
static PyObject *__pyx_int_437;
static PyObject *__pyx_int_439;
static PyObject *__pyx_int_440;
static PyObject *__pyx_int_442;
static PyObject *__pyx_int_443;
static PyObject *__pyx_int_444;
static PyObject *__pyx_int_446;
static PyObject *__pyx_int_447;
static PyObject *__pyx_int_449;
static PyObject *__pyx_int_450;
static PyObject *__pyx_int_451;
static PyObject *__pyx_int_452;
static PyObject *__pyx_int_454;
static PyObject *__pyx_int_455;
static PyObject *__pyx_int_456;
static PyObject *__pyx_int_457;
static PyObject *__pyx_int_458;
static PyObject *__pyx_int_460;
static PyObject *__pyx_int_461;
static PyObject *__pyx_int_462;
static PyObject *__pyx_int_464;
static PyObject *__pyx_int_466;
static PyObject *__pyx_int_467;
static PyObject *__pyx_int_468;
static PyObject *__pyx_int_471;
static PyObject *__pyx_int_473;
static PyObject *__pyx_int_474;
static PyObject *__pyx_int_475;
static PyObject *__pyx_int_476;
static PyObject *__pyx_int_478;
static PyObject *__pyx_int_479;
static PyObject *__pyx_int_481;
static PyObject *__pyx_int_482;
static PyObject *__pyx_int_484;
static PyObject *__pyx_int_485;
static PyObject *__pyx_int_486;
static PyObject *__pyx_int_488;
static PyObject *__pyx_int_489;
static PyObject *__pyx_int_490;
static PyObject *__pyx_int_492;
static PyObject *__pyx_int_494;
static PyObject *__pyx_int_495;
static PyObject *__pyx_int_496;
static PyObject *__pyx_int_498;
static PyObject *__pyx_int_499;
static PyObject *__pyx_int_501;
static PyObject *__pyx_int_502;
static PyObject *__pyx_int_503;
static PyObject *__pyx_int_505;
static PyObject *__pyx_int_506;
static PyObject *__pyx_int_508;
static PyObject *__pyx_int_509;
static PyObject *__pyx_int_510;
static PyObject *__pyx_int_511;
static PyObject *__pyx_int_513;
static PyObject *__pyx_int_514;
static PyObject *__pyx_int_516;
static PyObject *__pyx_int_517;
static PyObject *__pyx_int_518;
static PyObject *__pyx_int_520;
static PyObject *__pyx_int_521;
static PyObject *__pyx_int_523;
static PyObject *__pyx_int_524;
static PyObject *__pyx_int_525;
static PyObject *__pyx_int_526;
static PyObject *__pyx_int_527;
static PyObject *__pyx_int_528;
static PyObject *__pyx_int_529;
static PyObject *__pyx_int_531;
static PyObject *__pyx_int_532;
static PyObject *__pyx_int_534;
static PyObject *__pyx_int_536;
static PyObject *__pyx_int_537;
static PyObject *__pyx_int_538;
static PyObject *__pyx_int_539;
static PyObject *__pyx_int_541;
static PyObject *__pyx_int_542;
static PyObject *__pyx_int_544;
static PyObject *__pyx_int_545;
static PyObject *__pyx_int_546;
static PyObject *__pyx_int_548;
static PyObject *__pyx_int_549;
static PyObject *__pyx_int_551;
static PyObject *__pyx_int_552;
static PyObject *__pyx_int_553;
static PyObject *__pyx_int_557;
static PyObject *__pyx_int_559;
static PyObject *__pyx_int_560;
static PyObject *__pyx_int_561;
static PyObject *__pyx_int_563;
static PyObject *__pyx_int_564;
static PyObject *__pyx_int_565;
static PyObject *__pyx_int_569;
static PyObject *__pyx_int_571;
static PyObject *__pyx_int_572;
static PyObject *__pyx_int_573;
static PyObject *__pyx_int_574;
static PyObject *__pyx_int_576;
static PyObject *__pyx_int_577;
static PyObject *__pyx_int_578;
static PyObject *__pyx_int_579;
static PyObject *__pyx_int_580;
static PyObject *__pyx_int_581;
static PyObject *__pyx_int_583;
static PyObject *__pyx_int_585;
static PyObject *__pyx_int_586;
static PyObject *__pyx_int_588;
static PyObject *__pyx_int_589;
static PyObject *__pyx_int_590;
static PyObject *__pyx_int_591;
static PyObject *__pyx_int_592;
static PyObject *__pyx_int_594;
static PyObject *__pyx_int_595;
static PyObject *__pyx_int_596;
static PyObject *__pyx_int_597;
static PyObject *__pyx_int_600;
static PyObject *__pyx_int_602;
static PyObject *__pyx_int_603;
static PyObject *__pyx_int_604;
static PyObject *__pyx_int_605;
static PyObject *__pyx_int_606;
static PyObject *__pyx_int_608;
static PyObject *__pyx_int_609;
static PyObject *__pyx_int_610;
static PyObject *__pyx_int_611;
static PyObject *__pyx_int_612;
static PyObject *__pyx_int_613;
static PyObject *__pyx_int_614;
static PyObject *__pyx_int_616;
static PyObject *__pyx_int_617;
static PyObject *__pyx_int_619;
static PyObject *__pyx_int_620;
static PyObject *__pyx_int_621;
static PyObject *__pyx_int_622;
static PyObject *__pyx_int_623;
static PyObject *__pyx_int_627;
static PyObject *__pyx_int_629;
static PyObject *__pyx_int_630;
static PyObject *__pyx_int_631;
static PyObject *__pyx_int_632;
static PyObject *__pyx_int_633;
static PyObject *__pyx_int_634;
static PyObject *__pyx_int_638;
static PyObject *__pyx_int_640;
static PyObject *__pyx_int_643;
static PyObject *__pyx_int_645;
static PyObject *__pyx_int_646;
static PyObject *__pyx_int_647;
static PyObject *__pyx_int_648;
static PyObject *__pyx_int_652;
static PyObject *__pyx_int_654;
static PyObject *__pyx_int_655;
static PyObject *__pyx_int_657;
static PyObject *__pyx_int_658;
static PyObject *__pyx_int_660;
static PyObject *__pyx_int_661;
static PyObject *__pyx_int_662;
static PyObject *__pyx_int_663;
static PyObject *__pyx_int_664;
static PyObject *__pyx_int_668;
static PyObject *__pyx_int_670;
static PyObject *__pyx_int_671;
static PyObject *__pyx_int_672;
static PyObject *__pyx_int_673;
static PyObject *__pyx_int_674;
static PyObject *__pyx_int_675;
static PyObject *__pyx_int_679;
static PyObject *__pyx_int_681;
static PyObject *__pyx_int_684;
static PyObject *__pyx_int_686;
static PyObject *__pyx_int_687;
static PyObject *__pyx_int_688;
static PyObject *__pyx_int_689;
static PyObject *__pyx_int_693;
static PyObject *__pyx_int_695;
static PyObject *__pyx_int_696;
static PyObject *__pyx_int_698;
static PyObject *__pyx_int_699;
static PyObject *__pyx_int_701;
static PyObject *__pyx_int_702;
static PyObject *__pyx_int_703;
static PyObject *__pyx_int_704;
static PyObject *__pyx_int_705;
static PyObject *__pyx_int_709;
static PyObject *__pyx_int_711;
static PyObject *__pyx_int_712;
static PyObject *__pyx_int_713;
static PyObject *__pyx_int_715;
static PyObject *__pyx_int_716;
static PyObject *__pyx_int_717;
static PyObject *__pyx_int_718;
static PyObject *__pyx_int_719;
static PyObject *__pyx_int_723;
static PyObject *__pyx_int_725;
static PyObject *__pyx_int_726;
static PyObject *__pyx_int_727;
static PyObject *__pyx_int_729;
static PyObject *__pyx_int_730;
static PyObject *__pyx_int_731;
static PyObject *__pyx_int_734;
static PyObject *__pyx_int_736;
static PyObject *__pyx_int_737;
static PyObject *__pyx_int_738;
static PyObject *__pyx_int_739;
static PyObject *__pyx_int_740;
static PyObject *__pyx_int_742;
static PyObject *__pyx_int_743;
static PyObject *__pyx_int_744;
static PyObject *__pyx_int_745;
static PyObject *__pyx_int_746;
static PyObject *__pyx_int_750;
static PyObject *__pyx_int_752;
static PyObject *__pyx_int_753;
static PyObject *__pyx_int_754;
static PyObject *__pyx_int_756;
static PyObject *__pyx_int_757;
static PyObject *__pyx_int_758;
static PyObject *__pyx_int_759;
static PyObject *__pyx_int_760;
static PyObject *__pyx_int_764;
static PyObject *__pyx_int_766;
static PyObject *__pyx_int_767;
static PyObject *__pyx_int_768;
static PyObject *__pyx_int_769;
static PyObject *__pyx_int_770;
static PyObject *__pyx_int_771;
static PyObject *__pyx_int_775;
static PyObject *__pyx_int_777;
static PyObject *__pyx_int_780;
static PyObject *__pyx_int_782;
static PyObject *__pyx_int_783;
static PyObject *__pyx_int_784;
static PyObject *__pyx_int_786;
static PyObject *__pyx_int_787;
static PyObject *__pyx_int_788;
static PyObject *__pyx_int_789;
static PyObject *__pyx_int_790;
static PyObject *__pyx_int_794;
static PyObject *__pyx_int_796;
static PyObject *__pyx_int_797;
static PyObject *__pyx_int_798;
static PyObject *__pyx_int_799;
static PyObject *__pyx_int_800;
static PyObject *__pyx_int_801;
static PyObject *__pyx_int_805;
static PyObject *__pyx_int_807;
static PyObject *__pyx_int_810;
static PyObject *__pyx_int_812;
static PyObject *__pyx_int_813;
static PyObject *__pyx_int_814;
static PyObject *__pyx_int_816;
static PyObject *__pyx_int_817;
static PyObject *__pyx_int_818;
static PyObject *__pyx_int_819;
static PyObject *__pyx_int_820;
static PyObject *__pyx_int_821;
static PyObject *__pyx_int_823;
static PyObject *__pyx_int_824;
static PyObject *__pyx_int_825;
static PyObject *__pyx_int_826;
static PyObject *__pyx_int_827;
static PyObject *__pyx_int_828;
static PyObject *__pyx_int_829;
static PyObject *__pyx_int_830;
static PyObject *__pyx_int_831;
static PyObject *__pyx_int_835;
static PyObject *__pyx_int_837;
static PyObject *__pyx_int_838;
static PyObject *__pyx_int_839;
static PyObject *__pyx_int_841;
static PyObject *__pyx_int_843;
static PyObject *__pyx_int_844;
static PyObject *__pyx_int_845;
static PyObject *__pyx_int_846;
static PyObject *__pyx_int_847;
static PyObject *__pyx_int_848;
static PyObject *__pyx_int_850;
static PyObject *__pyx_int_853;
static PyObject *__pyx_int_854;
static PyObject *__pyx_int_855;
static PyObject *__pyx_int_856;
static PyObject *__pyx_int_857;
static PyObject *__pyx_int_859;
static PyObject *__pyx_int_860;
static PyObject *__pyx_int_861;
static PyObject *__pyx_int_862;
static PyObject *__pyx_int_864;
static PyObject *__pyx_int_865;
static PyObject *__pyx_int_866;
static PyObject *__pyx_int_867;
static PyObject *__pyx_int_868;
static PyObject *__pyx_int_869;
static PyObject *__pyx_int_870;
static PyObject *__pyx_int_871;
static PyObject *__pyx_int_872;
static PyObject *__pyx_int_873;
static PyObject *__pyx_int_874;
static PyObject *__pyx_int_875;
static PyObject *__pyx_int_876;
static PyObject *__pyx_int_877;
static PyObject *__pyx_int_878;
static PyObject *__pyx_int_879;
static PyObject *__pyx_int_881;
static PyObject *__pyx_int_883;
static PyObject *__pyx_int_884;
static PyObject *__pyx_int_885;
static PyObject *__pyx_int_886;
static PyObject *__pyx_int_887;
static PyObject *__pyx_int_888;
static PyObject *__pyx_int_889;
static PyObject *__pyx_int_891;
static PyObject *__pyx_int_892;
static PyObject *__pyx_int_893;
static PyObject *__pyx_int_894;
static PyObject *__pyx_int_895;
static PyObject *__pyx_int_896;
static PyObject *__pyx_int_897;
static PyObject *__pyx_int_898;
static PyObject *__pyx_int_900;
static PyObject *__pyx_int_901;
static PyObject *__pyx_int_902;
static PyObject *__pyx_int_903;
static PyObject *__pyx_int_906;
static PyObject *__pyx_int_908;
static PyObject *__pyx_int_909;
static PyObject *__pyx_int_911;
static PyObject *__pyx_int_912;
static PyObject *__pyx_int_913;
static PyObject *__pyx_int_915;
static PyObject *__pyx_int_916;
static PyObject *__pyx_int_918;
static PyObject *__pyx_int_919;
static PyObject *__pyx_int_921;
static PyObject *__pyx_int_922;
static PyObject *__pyx_int_923;
static PyObject *__pyx_int_926;
static PyObject *__pyx_int_928;
static PyObject *__pyx_int_929;
static PyObject *__pyx_int_930;
static PyObject *__pyx_int_931;
static PyObject *__pyx_int_933;
static PyObject *__pyx_int_934;
static PyObject *__pyx_int_935;
static PyObject *__pyx_int_938;
static PyObject *__pyx_int_940;
static PyObject *__pyx_int_941;
static PyObject *__pyx_int_942;
static PyObject *__pyx_int_943;
static PyObject *__pyx_int_945;
static PyObject *__pyx_int_947;
static PyObject *__pyx_int_948;
static PyObject *__pyx_int_949;
static PyObject *__pyx_int_950;
static PyObject *__pyx_int_951;
static PyObject *__pyx_int_952;
static PyObject *__pyx_int_954;
static PyObject *__pyx_int_956;
static PyObject *__pyx_int_957;
static PyObject *__pyx_int_958;
static PyObject *__pyx_int_960;
static PyObject *__pyx_int_961;
static PyObject *__pyx_int_962;
static PyObject *__pyx_int_965;
static PyObject *__pyx_int_967;
static PyObject *__pyx_int_968;
static PyObject *__pyx_int_969;
static PyObject *__pyx_int_970;
static PyObject *__pyx_int_972;
static PyObject *__pyx_int_973;
static PyObject *__pyx_int_974;
static PyObject *__pyx_int_975;
static PyObject *__pyx_int_976;
static PyObject *__pyx_int_978;
static PyObject *__pyx_int_980;
static PyObject *__pyx_int_981;
static PyObject *__pyx_int_983;
static PyObject *__pyx_int_984;
static PyObject *__pyx_int_986;
static PyObject *__pyx_int_987;
static PyObject *__pyx_int_990;
static PyObject *__pyx_int_992;
static PyObject *__pyx_int_993;
static PyObject *__pyx_int_994;
static PyObject *__pyx_int_995;
static PyObject *__pyx_int_998;
static PyObject *__pyx_int_1000;
static PyObject *__pyx_int_1001;
static PyObject *__pyx_int_1002;
static PyObject *__pyx_int_1003;
static PyObject *__pyx_int_1004;
static PyObject *__pyx_int_1007;
static PyObject *__pyx_int_1009;
static PyObject *__pyx_int_1010;
static PyObject *__pyx_int_1012;
static PyObject *__pyx_int_1013;
static PyObject *__pyx_int_1014;
static PyObject *__pyx_int_1016;
static PyObject *__pyx_int_1017;
static PyObject *__pyx_int_1018;
static PyObject *__pyx_int_1020;
static PyObject *__pyx_int_1021;
static PyObject *__pyx_int_1022;
static PyObject *__pyx_int_1025;
static PyObject *__pyx_int_1027;
static PyObject *__pyx_int_1028;
static PyObject *__pyx_int_1030;
static PyObject *__pyx_int_1031;
static PyObject *__pyx_int_neg_1;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__28;
static PyObject *__pyx_tuple__30;
static PyObject *__pyx_tuple__32;
static PyObject *__pyx_tuple__34;
static PyObject *__pyx_tuple__36;
static PyObject *__pyx_tuple__38;
static PyObject *__pyx_tuple__40;
static PyObject *__pyx_tuple__42;
static PyObject *__pyx_tuple__44;
static PyObject *__pyx_tuple__46;
static PyObject *__pyx_tuple__48;
static PyObject *__pyx_tuple__50;
static PyObject *__pyx_tuple__52;
static PyObject *__pyx_tuple__54;
static PyObject *__pyx_tuple__56;
static PyObject *__pyx_tuple__58;
static PyObject *__pyx_tuple__60;
static PyObject *__pyx_tuple__62;
static PyObject *__pyx_tuple__64;
static PyObject *__pyx_tuple__66;
static PyObject *__pyx_tuple__68;
static PyObject *__pyx_tuple__70;
static PyObject *__pyx_tuple__72;
static PyObject *__pyx_tuple__74;
static PyObject *__pyx_tuple__76;
static PyObject *__pyx_tuple__78;
static PyObject *__pyx_tuple__80;
static PyObject *__pyx_tuple__82;
static PyObject *__pyx_tuple__84;
static PyObject *__pyx_tuple__86;
static PyObject *__pyx_tuple__88;
static PyObject *__pyx_tuple__90;
static PyObject *__pyx_tuple__92;
static PyObject *__pyx_tuple__94;
static PyObject *__pyx_tuple__96;
static PyObject *__pyx_tuple__98;
static PyObject *__pyx_tuple__100;
static PyObject *__pyx_tuple__102;
static PyObject *__pyx_tuple__104;
static PyObject *__pyx_tuple__106;
static PyObject *__pyx_tuple__108;
static PyObject *__pyx_tuple__110;
static PyObject *__pyx_tuple__112;
static PyObject *__pyx_tuple__114;
static PyObject *__pyx_tuple__116;
static PyObject *__pyx_tuple__118;
static PyObject *__pyx_tuple__120;
static PyObject *__pyx_tuple__122;
static PyObject *__pyx_tuple__124;
static PyObject *__pyx_tuple__126;
static PyObject *__pyx_tuple__127;
static PyObject *__pyx_tuple__128;
static PyObject *__pyx_tuple__129;
static PyObject *__pyx_tuple__130;
static PyObject *__pyx_tuple__131;
static PyObject *__pyx_tuple__132;
static PyObject *__pyx_tuple__133;
static PyObject *__pyx_tuple__134;
static PyObject *__pyx_tuple__135;
static PyObject *__pyx_tuple__136;
static PyObject *__pyx_tuple__137;
static PyObject *__pyx_tuple__138;
static PyObject *__pyx_tuple__139;
static PyObject *__pyx_tuple__140;
static PyObject *__pyx_tuple__141;
static PyObject *__pyx_tuple__142;
static PyObject *__pyx_tuple__143;
static PyObject *__pyx_tuple__145;
static PyObject *__pyx_tuple__147;
static PyObject *__pyx_tuple__149;
static PyObject *__pyx_tuple__151;
static PyObject *__pyx_tuple__153;
static PyObject *__pyx_tuple__155;
static PyObject *__pyx_tuple__157;
static PyObject *__pyx_tuple__158;
static PyObject *__pyx_tuple__159;
static PyObject *__pyx_tuple__160;
static PyObject *__pyx_tuple__161;
static PyObject *__pyx_tuple__162;
static PyObject *__pyx_tuple__163;
static PyObject *__pyx_tuple__164;
static PyObject *__pyx_tuple__165;
static PyObject *__pyx_tuple__166;
static PyObject *__pyx_tuple__168;
static PyObject *__pyx_tuple__170;
static PyObject *__pyx_tuple__171;
static PyObject *__pyx_tuple__173;
static PyObject *__pyx_tuple__175;
static PyObject *__pyx_tuple__177;
static PyObject *__pyx_tuple__179;
static PyObject *__pyx_tuple__181;
static PyObject *__pyx_tuple__183;
static PyObject *__pyx_tuple__184;
static PyObject *__pyx_tuple__185;
static PyObject *__pyx_tuple__186;
static PyObject *__pyx_tuple__187;
static PyObject *__pyx_tuple__188;
static PyObject *__pyx_tuple__189;
static PyObject *__pyx_tuple__190;
static PyObject *__pyx_tuple__191;
static PyObject *__pyx_tuple__192;
static PyObject *__pyx_tuple__193;
static PyObject *__pyx_tuple__194;
static PyObject *__pyx_tuple__196;
static PyObject *__pyx_tuple__198;
static PyObject *__pyx_tuple__200;
static PyObject *__pyx_tuple__202;
static PyObject *__pyx_tuple__203;
static PyObject *__pyx_tuple__205;
static PyObject *__pyx_tuple__207;
static PyObject *__pyx_tuple__209;
static PyObject *__pyx_tuple__211;
static PyObject *__pyx_tuple__213;
static PyObject *__pyx_tuple__214;
static PyObject *__pyx_tuple__216;
static PyObject *__pyx_tuple__217;
static PyObject *__pyx_tuple__219;
static PyObject *__pyx_tuple__221;
static PyObject *__pyx_tuple__222;
static PyObject *__pyx_tuple__224;
static PyObject *__pyx_tuple__226;
static PyObject *__pyx_tuple__228;
static PyObject *__pyx_tuple__230;
static PyObject *__pyx_tuple__232;
static PyObject *__pyx_tuple__234;
static PyObject *__pyx_tuple__236;
static PyObject *__pyx_tuple__238;
static PyObject *__pyx_tuple__240;
static PyObject *__pyx_tuple__242;
static PyObject *__pyx_tuple__244;
static PyObject *__pyx_tuple__246;
static PyObject *__pyx_tuple__248;
static PyObject *__pyx_tuple__250;
static PyObject *__pyx_tuple__251;
static PyObject *__pyx_tuple__252;
static PyObject *__pyx_tuple__253;
static PyObject *__pyx_tuple__254;
static PyObject *__pyx_tuple__256;
static PyObject *__pyx_tuple__258;
static PyObject *__pyx_tuple__259;
static PyObject *__pyx_tuple__260;
static PyObject *__pyx_tuple__262;
static PyObject *__pyx_tuple__264;
static PyObject *__pyx_tuple__266;
static PyObject *__pyx_tuple__268;
static PyObject *__pyx_tuple__269;
static PyObject *__pyx_tuple__271;
static PyObject *__pyx_tuple__272;
static PyObject *__pyx_tuple__273;
static PyObject *__pyx_tuple__275;
static PyObject *__pyx_tuple__277;
static PyObject *__pyx_tuple__278;
static PyObject *__pyx_tuple__279;
static PyObject *__pyx_tuple__281;
static PyObject *__pyx_tuple__282;
static PyObject *__pyx_tuple__283;
static PyObject *__pyx_tuple__285;
static PyObject *__pyx_tuple__287;
static PyObject *__pyx_tuple__289;
static PyObject *__pyx_tuple__291;
static PyObject *__pyx_tuple__293;
static PyObject *__pyx_tuple__295;
static PyObject *__pyx_tuple__296;
static PyObject *__pyx_tuple__297;
static PyObject *__pyx_tuple__299;
static PyObject *__pyx_tuple__300;
static PyObject *__pyx_tuple__301;
static PyObject *__pyx_tuple__303;
static PyObject *__pyx_tuple__305;
static PyObject *__pyx_tuple__307;
static PyObject *__pyx_tuple__309;
static PyObject *__pyx_tuple__311;
static PyObject *__pyx_tuple__313;
static PyObject *__pyx_tuple__314;
static PyObject *__pyx_tuple__316;
static PyObject *__pyx_tuple__318;
static PyObject *__pyx_tuple__319;
static PyObject *__pyx_tuple__321;
static PyObject *__pyx_tuple__323;
static PyObject *__pyx_tuple__324;
static PyObject *__pyx_tuple__326;
static PyObject *__pyx_tuple__327;
static PyObject *__pyx_tuple__328;
static PyObject *__pyx_tuple__330;
static PyObject *__pyx_tuple__332;
static PyObject *__pyx_tuple__334;
static PyObject *__pyx_tuple__336;
static PyObject *__pyx_tuple__338;
static PyObject *__pyx_tuple__340;
static PyObject *__pyx_tuple__342;
static PyObject *__pyx_tuple__343;
static PyObject *__pyx_tuple__345;
static PyObject *__pyx_tuple__346;
static PyObject *__pyx_tuple__347;
static PyObject *__pyx_tuple__349;
static PyObject *__pyx_tuple__350;
static PyObject *__pyx_tuple__352;
static PyObject *__pyx_tuple__354;
static PyObject *__pyx_tuple__355;
static PyObject *__pyx_tuple__356;
static PyObject *__pyx_tuple__358;
static PyObject *__pyx_tuple__359;
static PyObject *__pyx_tuple__360;
static PyObject *__pyx_tuple__362;
static PyObject *__pyx_tuple__364;
static PyObject *__pyx_tuple__366;
static PyObject *__pyx_tuple__367;
static PyObject *__pyx_tuple__369;
static PyObject *__pyx_tuple__370;
static PyObject *__pyx_tuple__371;
static PyObject *__pyx_tuple__372;
static PyObject *__pyx_tuple__374;
static PyObject *__pyx_tuple__376;
static PyObject *__pyx_tuple__378;
static PyObject *__pyx_tuple__380;
static PyObject *__pyx_tuple__381;
static PyObject *__pyx_tuple__382;
static PyObject *__pyx_tuple__384;
static PyObject *__pyx_tuple__386;
static PyObject *__pyx_tuple__388;
static PyObject *__pyx_tuple__389;
static PyObject *__pyx_tuple__390;
static PyObject *__pyx_tuple__392;
static PyObject *__pyx_tuple__393;
static PyObject *__pyx_tuple__395;
static PyObject *__pyx_tuple__397;
static PyObject *__pyx_tuple__398;
static PyObject *__pyx_tuple__399;
static PyObject *__pyx_tuple__401;
static PyObject *__pyx_tuple__403;
static PyObject *__pyx_tuple__405;
static PyObject *__pyx_tuple__407;
static PyObject *__pyx_tuple__408;
static PyObject *__pyx_tuple__409;
static PyObject *__pyx_tuple__411;
static PyObject *__pyx_tuple__412;
static PyObject *__pyx_tuple__413;
static PyObject *__pyx_tuple__415;
static PyObject *__pyx_tuple__416;
static PyObject *__pyx_tuple__417;
static PyObject *__pyx_tuple__418;
static PyObject *__pyx_tuple__419;
static PyObject *__pyx_tuple__420;
static PyObject *__pyx_tuple__422;
static PyObject *__pyx_tuple__423;
static PyObject *__pyx_tuple__424;
static PyObject *__pyx_tuple__425;
static PyObject *__pyx_tuple__426;
static PyObject *__pyx_tuple__428;
static PyObject *__pyx_tuple__430;
static PyObject *__pyx_tuple__432;
static PyObject *__pyx_tuple__434;
static PyObject *__pyx_tuple__436;
static PyObject *__pyx_tuple__438;
static PyObject *__pyx_tuple__439;
static PyObject *__pyx_tuple__440;
static PyObject *__pyx_tuple__442;
static PyObject *__pyx_tuple__443;
static PyObject *__pyx_tuple__444;
static PyObject *__pyx_tuple__445;
static PyObject *__pyx_tuple__446;
static PyObject *__pyx_tuple__448;
static PyObject *__pyx_tuple__449;
static PyObject *__pyx_tuple__451;
static PyObject *__pyx_tuple__452;
static PyObject *__pyx_tuple__453;
static PyObject *__pyx_tuple__454;
static PyObject *__pyx_tuple__455;
static PyObject *__pyx_tuple__456;
static PyObject *__pyx_tuple__457;
static PyObject *__pyx_tuple__459;
static PyObject *__pyx_tuple__461;
static PyObject *__pyx_tuple__462;
static PyObject *__pyx_tuple__463;
static PyObject *__pyx_tuple__464;
static PyObject *__pyx_tuple__465;
static PyObject *__pyx_tuple__466;
static PyObject *__pyx_tuple__468;
static PyObject *__pyx_tuple__470;
static PyObject *__pyx_tuple__471;
static PyObject *__pyx_tuple__473;
static PyObject *__pyx_tuple__475;
static PyObject *__pyx_tuple__477;
static PyObject *__pyx_tuple__479;
static PyObject *__pyx_tuple__480;
static PyObject *__pyx_tuple__481;
static PyObject *__pyx_tuple__482;
static PyObject *__pyx_tuple__483;
static PyObject *__pyx_tuple__485;
static PyObject *__pyx_tuple__487;
static PyObject *__pyx_tuple__488;
static PyObject *__pyx_tuple__489;
static PyObject *__pyx_tuple__490;
static PyObject *__pyx_tuple__492;
static PyObject *__pyx_tuple__494;
static PyObject *__pyx_tuple__495;
static PyObject *__pyx_tuple__496;
static PyObject *__pyx_tuple__497;
static PyObject *__pyx_tuple__498;
static PyObject *__pyx_tuple__500;
static PyObject *__pyx_tuple__502;
static PyObject *__pyx_tuple__503;
static PyObject *__pyx_tuple__504;
static PyObject *__pyx_tuple__505;
static PyObject *__pyx_tuple__507;
static PyObject *__pyx_tuple__508;
static PyObject *__pyx_tuple__509;
static PyObject *__pyx_tuple__511;
static PyObject *__pyx_tuple__513;
static PyObject *__pyx_tuple__514;
static PyObject *__pyx_tuple__516;
static PyObject *__pyx_tuple__517;
static PyObject *__pyx_tuple__518;
static PyObject *__pyx_tuple__520;
static PyObject *__pyx_tuple__522;
static PyObject *__pyx_tuple__523;
static PyObject *__pyx_tuple__524;
static PyObject *__pyx_tuple__525;
static PyObject *__pyx_tuple__526;
static PyObject *__pyx_tuple__528;
static PyObject *__pyx_tuple__530;
static PyObject *__pyx_tuple__531;
static PyObject *__pyx_tuple__532;
static PyObject *__pyx_tuple__534;
static PyObject *__pyx_tuple__536;
static PyObject *__pyx_tuple__538;
static PyObject *__pyx_tuple__540;
static PyObject *__pyx_tuple__541;
static PyObject *__pyx_tuple__542;
static PyObject *__pyx_tuple__544;
static PyObject *__pyx_tuple__546;
static PyObject *__pyx_tuple__548;
static PyObject *__pyx_tuple__550;
static PyObject *__pyx_tuple__551;
static PyObject *__pyx_tuple__552;
static PyObject *__pyx_tuple__554;
static PyObject *__pyx_tuple__556;
static PyObject *__pyx_tuple__557;
static PyObject *__pyx_tuple__558;
static PyObject *__pyx_tuple__560;
static PyObject *__pyx_tuple__562;
static PyObject *__pyx_tuple__564;
static PyObject *__pyx_tuple__566;
static PyObject *__pyx_tuple__568;
static PyObject *__pyx_tuple__569;
static PyObject *__pyx_tuple__571;
static PyObject *__pyx_tuple__573;
static PyObject *__pyx_tuple__575;
static PyObject *__pyx_tuple__577;
static PyObject *__pyx_tuple__579;
static PyObject *__pyx_tuple__580;
static PyObject *__pyx_tuple__581;
static PyObject *__pyx_tuple__583;
static PyObject *__pyx_tuple__585;
static PyObject *__pyx_tuple__586;
static PyObject *__pyx_tuple__587;
static PyObject *__pyx_tuple__588;
static PyObject *__pyx_tuple__590;
static PyObject *__pyx_tuple__592;
static PyObject *__pyx_tuple__594;
static PyObject *__pyx_tuple__595;
static PyObject *__pyx_tuple__596;
static PyObject *__pyx_tuple__598;
static PyObject *__pyx_tuple__599;
static PyObject *__pyx_tuple__600;
static PyObject *__pyx_tuple__601;
static PyObject *__pyx_tuple__602;
static PyObject *__pyx_tuple__603;
static PyObject *__pyx_tuple__605;
static PyObject *__pyx_tuple__607;
static PyObject *__pyx_tuple__608;
static PyObject *__pyx_tuple__609;
static PyObject *__pyx_tuple__610;
static PyObject *__pyx_tuple__612;
static PyObject *__pyx_tuple__614;
static PyObject *__pyx_tuple__615;
static PyObject *__pyx_tuple__616;
static PyObject *__pyx_tuple__618;
static PyObject *__pyx_tuple__620;
static PyObject *__pyx_tuple__622;
static PyObject *__pyx_tuple__624;
static PyObject *__pyx_tuple__626;
static PyObject *__pyx_tuple__627;
static PyObject *__pyx_tuple__628;
static PyObject *__pyx_tuple__629;
static PyObject *__pyx_tuple__631;
static PyObject *__pyx_tuple__633;
static PyObject *__pyx_tuple__634;
static PyObject *__pyx_tuple__635;
static PyObject *__pyx_tuple__637;
static PyObject *__pyx_tuple__639;
static PyObject *__pyx_tuple__641;
static PyObject *__pyx_tuple__643;
static PyObject *__pyx_tuple__645;
static PyObject *__pyx_tuple__646;
static PyObject *__pyx_tuple__647;
static PyObject *__pyx_tuple__648;
static PyObject *__pyx_tuple__649;
static PyObject *__pyx_tuple__650;
static PyObject *__pyx_tuple__651;
static PyObject *__pyx_tuple__652;
static PyObject *__pyx_tuple__654;
static PyObject *__pyx_tuple__656;
static PyObject *__pyx_tuple__657;
static PyObject *__pyx_tuple__658;
static PyObject *__pyx_tuple__659;
static PyObject *__pyx_tuple__660;
static PyObject *__pyx_tuple__662;
static PyObject *__pyx_tuple__663;
static PyObject *__pyx_tuple__664;
static PyObject *__pyx_tuple__665;
static PyObject *__pyx_tuple__666;
static PyObject *__pyx_tuple__667;
static PyObject *__pyx_tuple__669;
static PyObject *__pyx_tuple__671;
static PyObject *__pyx_tuple__672;
static PyObject *__pyx_tuple__673;
static PyObject *__pyx_tuple__674;
static PyObject *__pyx_tuple__675;
static PyObject *__pyx_tuple__676;
static PyObject *__pyx_tuple__678;
static PyObject *__pyx_tuple__680;
static PyObject *__pyx_tuple__681;
static PyObject *__pyx_tuple__682;
static PyObject *__pyx_tuple__683;
static PyObject *__pyx_tuple__685;
static PyObject *__pyx_tuple__686;
static PyObject *__pyx_tuple__687;
static PyObject *__pyx_tuple__688;
static PyObject *__pyx_tuple__689;
static PyObject *__pyx_tuple__691;
static PyObject *__pyx_tuple__693;
static PyObject *__pyx_tuple__694;
static PyObject *__pyx_tuple__696;
static PyObject *__pyx_tuple__698;
static PyObject *__pyx_tuple__700;
static PyObject *__pyx_tuple__702;
static PyObject *__pyx_tuple__704;
static PyObject *__pyx_tuple__706;
static PyObject *__pyx_tuple__707;
static PyObject *__pyx_tuple__709;
static PyObject *__pyx_tuple__710;
static PyObject *__pyx_tuple__712;
static PyObject *__pyx_tuple__713;
static PyObject *__pyx_tuple__714;
static PyObject *__pyx_tuple__715;
static PyObject *__pyx_tuple__716;
static PyObject *__pyx_tuple__718;
static PyObject *__pyx_tuple__720;
static PyObject *__pyx_tuple__722;
static PyObject *__pyx_tuple__724;
static PyObject *__pyx_tuple__725;
static PyObject *__pyx_tuple__726;
static PyObject *__pyx_tuple__727;
static PyObject *__pyx_tuple__729;
static PyObject *__pyx_tuple__731;
static PyObject *__pyx_tuple__732;
static PyObject *__pyx_tuple__733;
static PyObject *__pyx_tuple__734;
static PyObject *__pyx_tuple__736;
static PyObject *__pyx_tuple__738;
static PyObject *__pyx_tuple__739;
static PyObject *__pyx_tuple__740;
static PyObject *__pyx_tuple__741;
static PyObject *__pyx_tuple__742;
static PyObject *__pyx_tuple__744;
static PyObject *__pyx_tuple__745;
static PyObject *__pyx_tuple__746;
static PyObject *__pyx_tuple__748;
static PyObject *__pyx_tuple__749;
static PyObject *__pyx_tuple__751;
static PyObject *__pyx_tuple__752;
static PyObject *__pyx_tuple__753;
static PyObject *__pyx_tuple__755;
static PyObject *__pyx_tuple__757;
static PyObject *__pyx_tuple__758;
static PyObject *__pyx_tuple__759;
static PyObject *__pyx_tuple__761;
static PyObject *__pyx_tuple__762;
static PyObject *__pyx_tuple__763;
static PyObject *__pyx_tuple__765;
static PyObject *__pyx_tuple__766;
static PyObject *__pyx_tuple__768;
static PyObject *__pyx_tuple__770;
static PyObject *__pyx_tuple__772;
static PyObject *__pyx_tuple__774;
static PyObject *__pyx_tuple__776;
static PyObject *__pyx_tuple__777;
static PyObject *__pyx_tuple__779;
static PyObject *__pyx_tuple__781;
static PyObject *__pyx_tuple__783;
static PyObject *__pyx_tuple__785;
static PyObject *__pyx_tuple__787;
static PyObject *__pyx_tuple__789;
static PyObject *__pyx_tuple__791;
static PyObject *__pyx_tuple__793;
static PyObject *__pyx_tuple__795;
static PyObject *__pyx_tuple__797;
static PyObject *__pyx_tuple__799;
static PyObject *__pyx_tuple__801;
static PyObject *__pyx_tuple__803;
static PyObject *__pyx_tuple__805;
static PyObject *__pyx_tuple__807;
static PyObject *__pyx_tuple__809;
static PyObject *__pyx_tuple__810;
static PyObject *__pyx_tuple__811;
static PyObject *__pyx_tuple__812;
static PyObject *__pyx_tuple__813;
static PyObject *__pyx_tuple__814;
static PyObject *__pyx_tuple__815;
static PyObject *__pyx_tuple__816;
static PyObject *__pyx_tuple__817;
static PyObject *__pyx_tuple__818;
static PyObject *__pyx_tuple__819;
static PyObject *__pyx_tuple__820;
static PyObject *__pyx_tuple__821;
static PyObject *__pyx_tuple__822;
static PyObject *__pyx_tuple__823;
static PyObject *__pyx_tuple__824;
static PyObject *__pyx_tuple__825;
static PyObject *__pyx_tuple__826;
static PyObject *__pyx_tuple__827;
static PyObject *__pyx_tuple__828;
static PyObject *__pyx_tuple__829;
static PyObject *__pyx_tuple__830;
static PyObject *__pyx_tuple__831;
static PyObject *__pyx_tuple__832;
static PyObject *__pyx_tuple__833;
static PyObject *__pyx_tuple__834;
static PyObject *__pyx_tuple__835;
static PyObject *__pyx_tuple__836;
static PyObject *__pyx_tuple__837;
static PyObject *__pyx_tuple__838;
static PyObject *__pyx_tuple__839;
static PyObject *__pyx_tuple__840;
static PyObject *__pyx_tuple__841;
static PyObject *__pyx_tuple__842;
static PyObject *__pyx_tuple__843;
static PyObject *__pyx_tuple__844;
static PyObject *__pyx_tuple__845;
static PyObject *__pyx_tuple__846;
static PyObject *__pyx_tuple__847;
static PyObject *__pyx_tuple__848;
static PyObject *__pyx_tuple__849;
static PyObject *__pyx_tuple__850;
static PyObject *__pyx_tuple__851;
static PyObject *__pyx_tuple__852;
static PyObject *__pyx_tuple__853;
static PyObject *__pyx_tuple__854;
static PyObject *__pyx_tuple__855;
static PyObject *__pyx_tuple__856;
static PyObject *__pyx_tuple__857;
static PyObject *__pyx_tuple__858;
static PyObject *__pyx_tuple__859;
static PyObject *__pyx_tuple__860;
static PyObject *__pyx_tuple__861;
static PyObject *__pyx_tuple__862;
static PyObject *__pyx_tuple__863;
static PyObject *__pyx_tuple__864;
static PyObject *__pyx_tuple__865;
static PyObject *__pyx_tuple__866;
static PyObject *__pyx_tuple__867;
static PyObject *__pyx_tuple__868;
static PyObject *__pyx_tuple__869;
static PyObject *__pyx_tuple__870;
static PyObject *__pyx_tuple__871;
static PyObject *__pyx_tuple__872;
static PyObject *__pyx_tuple__873;
static PyObject *__pyx_tuple__874;
static PyObject *__pyx_tuple__875;
static PyObject *__pyx_tuple__876;
static PyObject *__pyx_tuple__877;
static PyObject *__pyx_tuple__878;
static PyObject *__pyx_tuple__879;
static PyObject *__pyx_tuple__880;
static PyObject *__pyx_tuple__881;
static PyObject *__pyx_tuple__882;
static PyObject *__pyx_tuple__883;
static PyObject *__pyx_tuple__884;
static PyObject *__pyx_tuple__885;
static PyObject *__pyx_tuple__886;
static PyObject *__pyx_tuple__887;
static PyObject *__pyx_tuple__888;
static PyObject *__pyx_tuple__889;
static PyObject *__pyx_tuple__890;
static PyObject *__pyx_tuple__891;
static PyObject *__pyx_tuple__892;
static PyObject *__pyx_tuple__893;
static PyObject *__pyx_tuple__894;
static PyObject *__pyx_tuple__895;
static PyObject *__pyx_tuple__896;
static PyObject *__pyx_tuple__897;
static PyObject *__pyx_tuple__898;
static PyObject *__pyx_tuple__899;
static PyObject *__pyx_tuple__900;
static PyObject *__pyx_tuple__901;
static PyObject *__pyx_tuple__902;
static PyObject *__pyx_tuple__903;
static PyObject *__pyx_tuple__904;
static PyObject *__pyx_tuple__905;
static PyObject *__pyx_tuple__906;
static PyObject *__pyx_tuple__907;
static PyObject *__pyx_tuple__908;
static PyObject *__pyx_tuple__909;
static PyObject *__pyx_tuple__910;
static PyObject *__pyx_tuple__911;
static PyObject *__pyx_tuple__912;
static PyObject *__pyx_tuple__913;
static PyObject *__pyx_tuple__914;
static PyObject *__pyx_tuple__915;
static PyObject *__pyx_tuple__916;
static PyObject *__pyx_tuple__917;
static PyObject *__pyx_tuple__918;
static PyObject *__pyx_tuple__919;
static PyObject *__pyx_tuple__920;
static PyObject *__pyx_tuple__921;
static PyObject *__pyx_tuple__922;
static PyObject *__pyx_tuple__923;
static PyObject *__pyx_tuple__924;
static PyObject *__pyx_tuple__925;
static PyObject *__pyx_tuple__926;
static PyObject *__pyx_tuple__927;
static PyObject *__pyx_tuple__928;
static PyObject *__pyx_tuple__929;
static PyObject *__pyx_tuple__930;
static PyObject *__pyx_tuple__931;
static PyObject *__pyx_tuple__932;
static PyObject *__pyx_tuple__933;
static PyObject *__pyx_tuple__934;
static PyObject *__pyx_tuple__935;
static PyObject *__pyx_tuple__936;
static PyObject *__pyx_tuple__937;
static PyObject *__pyx_tuple__938;
static PyObject *__pyx_tuple__939;
static PyObject *__pyx_tuple__940;
static PyObject *__pyx_tuple__941;
static PyObject *__pyx_tuple__942;
static PyObject *__pyx_tuple__943;
static PyObject *__pyx_tuple__944;
static PyObject *__pyx_tuple__945;
static PyObject *__pyx_tuple__946;
static PyObject *__pyx_tuple__947;
static PyObject *__pyx_tuple__948;
static PyObject *__pyx_tuple__949;
static PyObject *__pyx_tuple__950;
static PyObject *__pyx_tuple__951;
static PyObject *__pyx_tuple__952;
static PyObject *__pyx_tuple__953;
static PyObject *__pyx_tuple__954;
static PyObject *__pyx_tuple__955;
static PyObject *__pyx_tuple__956;
static PyObject *__pyx_tuple__957;
static PyObject *__pyx_tuple__958;
static PyObject *__pyx_tuple__959;
static PyObject *__pyx_tuple__960;
static PyObject *__pyx_tuple__961;
static PyObject *__pyx_tuple__962;
static PyObject *__pyx_tuple__963;
static PyObject *__pyx_tuple__964;
static PyObject *__pyx_tuple__965;
static PyObject *__pyx_tuple__966;
static PyObject *__pyx_tuple__967;
static PyObject *__pyx_tuple__968;
static PyObject *__pyx_tuple__969;
static PyObject *__pyx_tuple__970;
static PyObject *__pyx_tuple__971;
static PyObject *__pyx_tuple__972;
static PyObject *__pyx_tuple__973;
static PyObject *__pyx_tuple__974;
static PyObject *__pyx_tuple__976;
static PyObject *__pyx_tuple__1006;
static PyObject *__pyx_tuple__1008;
static PyObject *__pyx_tuple__1010;
static PyObject *__pyx_tuple__1011;
static PyObject *__pyx_tuple__1013;
static PyObject *__pyx_tuple__1015;
static PyObject *__pyx_tuple__1016;
static PyObject *__pyx_tuple__1018;
static PyObject *__pyx_tuple__1020;
static PyObject *__pyx_tuple__1022;
static PyObject *__pyx_tuple__1024;
static PyObject *__pyx_tuple__1026;
static PyObject *__pyx_tuple__1028;
static PyObject *__pyx_tuple__1030;
static PyObject *__pyx_tuple__1031;
static PyObject *__pyx_tuple__1033;
static PyObject *__pyx_tuple__1035;
static PyObject *__pyx_tuple__1037;
static PyObject *__pyx_tuple__1039;
static PyObject *__pyx_tuple__1041;
static PyObject *__pyx_tuple__1043;
static PyObject *__pyx_tuple__1045;
static PyObject *__pyx_tuple__1047;
static PyObject *__pyx_tuple__1048;
static PyObject *__pyx_tuple__1050;
static PyObject *__pyx_tuple__1052;
static PyObject *__pyx_tuple__1054;
static PyObject *__pyx_tuple__1056;
static PyObject *__pyx_tuple__1058;
static PyObject *__pyx_tuple__1060;
static PyObject *__pyx_tuple__1062;
static PyObject *__pyx_tuple__1064;
static PyObject *__pyx_tuple__1066;
static PyObject *__pyx_tuple__1067;
static PyObject *__pyx_tuple__1069;
static PyObject *__pyx_tuple__1071;
static PyObject *__pyx_tuple__1073;
static PyObject *__pyx_tuple__1075;
static PyObject *__pyx_tuple__1077;
static PyObject *__pyx_tuple__1079;
static PyObject *__pyx_tuple__1081;
static PyObject *__pyx_tuple__1083;
static PyObject *__pyx_tuple__1084;
static PyObject *__pyx_tuple__1086;
static PyObject *__pyx_tuple__1088;
static PyObject *__pyx_tuple__1090;
static PyObject *__pyx_tuple__1092;
static PyObject *__pyx_tuple__1094;
static PyObject *__pyx_tuple__1096;
static PyObject *__pyx_tuple__1097;
static PyObject *__pyx_tuple__1099;
static PyObject *__pyx_tuple__1101;
static PyObject *__pyx_tuple__1103;
static PyObject *__pyx_tuple__1105;
static PyObject *__pyx_tuple__1106;
static PyObject *__pyx_tuple__1108;
static PyObject *__pyx_tuple__1110;
static PyObject *__pyx_tuple__1112;
static PyObject *__pyx_tuple__1114;
static PyObject *__pyx_tuple__1116;
static PyObject *__pyx_tuple__1118;
static PyObject *__pyx_tuple__1120;
static PyObject *__pyx_tuple__1122;
static PyObject *__pyx_tuple__1124;
static PyObject *__pyx_tuple__1126;
static PyObject *__pyx_tuple__1128;
static PyObject *__pyx_tuple__1130;
static PyObject *__pyx_tuple__1132;
static PyObject *__pyx_tuple__1134;
static PyObject *__pyx_tuple__1136;
static PyObject *__pyx_tuple__1138;
static PyObject *__pyx_tuple__1140;
static PyObject *__pyx_tuple__1142;
static PyObject *__pyx_tuple__1144;
static PyObject *__pyx_tuple__1146;
static PyObject *__pyx_tuple__1148;
static PyObject *__pyx_tuple__1150;
static PyObject *__pyx_tuple__1152;
static PyObject *__pyx_tuple__1154;
static PyObject *__pyx_tuple__1155;
static PyObject *__pyx_tuple__1157;
static PyObject *__pyx_tuple__1159;
static PyObject *__pyx_tuple__1161;
static PyObject *__pyx_tuple__1163;
static PyObject *__pyx_tuple__1165;
static PyObject *__pyx_tuple__1167;
static PyObject *__pyx_tuple__1169;
static PyObject *__pyx_tuple__1170;
static PyObject *__pyx_tuple__1172;
static PyObject *__pyx_tuple__1174;
static PyObject *__pyx_tuple__1176;
static PyObject *__pyx_tuple__1178;
static PyObject *__pyx_tuple__1180;
static PyObject *__pyx_tuple__1182;
static PyObject *__pyx_tuple__1184;
static PyObject *__pyx_tuple__1186;
static PyObject *__pyx_tuple__1187;
static PyObject *__pyx_tuple__1189;
static PyObject *__pyx_tuple__1191;
static PyObject *__pyx_tuple__1193;
static PyObject *__pyx_tuple__1195;
static PyObject *__pyx_tuple__1197;
static PyObject *__pyx_tuple__1199;
static PyObject *__pyx_tuple__1201;
static PyObject *__pyx_tuple__1202;
static PyObject *__pyx_tuple__1204;
static PyObject *__pyx_tuple__1205;
static PyObject *__pyx_tuple__1207;
static PyObject *__pyx_tuple__1209;
static PyObject *__pyx_tuple__1211;
static PyObject *__pyx_tuple__1213;
static PyObject *__pyx_tuple__1215;
static PyObject *__pyx_tuple__1217;
static PyObject *__pyx_tuple__1218;
static PyObject *__pyx_tuple__1220;
static PyObject *__pyx_tuple__1222;
static PyObject *__pyx_tuple__1224;
static PyObject *__pyx_tuple__1226;
static PyObject *__pyx_tuple__1228;
static PyObject *__pyx_tuple__1230;
static PyObject *__pyx_tuple__1232;
static PyObject *__pyx_tuple__1234;
static PyObject *__pyx_tuple__1235;
static PyObject *__pyx_tuple__1237;
static PyObject *__pyx_tuple__1238;
static PyObject *__pyx_tuple__1240;
static PyObject *__pyx_tuple__1241;
static PyObject *__pyx_tuple__1243;
static PyObject *__pyx_tuple__1244;
static PyObject *__pyx_tuple__1246;
static PyObject *__pyx_tuple__1247;
static PyObject *__pyx_tuple__1249;
static PyObject *__pyx_tuple__1251;
static PyObject *__pyx_tuple__1253;
static PyObject *__pyx_tuple__1255;
static PyObject *__pyx_tuple__1257;
static PyObject *__pyx_tuple__1259;
static PyObject *__pyx_tuple__1261;
static PyObject *__pyx_tuple__1263;
static PyObject *__pyx_tuple__1264;
static PyObject *__pyx_tuple__1266;
static PyObject *__pyx_tuple__1268;
static PyObject *__pyx_tuple__1270;
static PyObject *__pyx_tuple__1272;
static PyObject *__pyx_tuple__1274;
static PyObject *__pyx_tuple__1276;
static PyObject *__pyx_tuple__1278;
static PyObject *__pyx_tuple__1279;
static PyObject *__pyx_tuple__1281;
static PyObject *__pyx_tuple__1283;
static PyObject *__pyx_tuple__1285;
static PyObject *__pyx_tuple__1287;
static PyObject *__pyx_tuple__1289;
static PyObject *__pyx_tuple__1291;
static PyObject *__pyx_tuple__1293;
static PyObject *__pyx_tuple__1295;
static PyObject *__pyx_tuple__1297;
static PyObject *__pyx_tuple__1298;
static PyObject *__pyx_tuple__1300;
static PyObject *__pyx_tuple__1301;
static PyObject *__pyx_tuple__1303;
static PyObject *__pyx_tuple__1305;
static PyObject *__pyx_tuple__1307;
static PyObject *__pyx_tuple__1309;
static PyObject *__pyx_tuple__1311;
static PyObject *__pyx_tuple__1313;
static PyObject *__pyx_tuple__1314;
static PyObject *__pyx_tuple__1316;
static PyObject *__pyx_tuple__1318;
static PyObject *__pyx_tuple__1320;
static PyObject *__pyx_tuple__1322;
static PyObject *__pyx_tuple__1324;
static PyObject *__pyx_tuple__1326;
static PyObject *__pyx_tuple__1328;
static PyObject *__pyx_tuple__1330;
static PyObject *__pyx_tuple__1332;
static PyObject *__pyx_tuple__1334;
static PyObject *__pyx_tuple__1335;
static PyObject *__pyx_tuple__1337;
static PyObject *__pyx_tuple__1339;
static PyObject *__pyx_tuple__1341;
static PyObject *__pyx_tuple__1343;
static PyObject *__pyx_tuple__1345;
static PyObject *__pyx_tuple__1347;
static PyObject *__pyx_tuple__1349;
static PyObject *__pyx_tuple__1351;
static PyObject *__pyx_tuple__1353;
static PyObject *__pyx_tuple__1355;
static PyObject *__pyx_tuple__1357;
static PyObject *__pyx_tuple__1359;
static PyObject *__pyx_tuple__1360;
static PyObject *__pyx_tuple__1362;
static PyObject *__pyx_tuple__1364;
static PyObject *__pyx_tuple__1366;
static PyObject *__pyx_tuple__1368;
static PyObject *__pyx_tuple__1370;
static PyObject *__pyx_tuple__1372;
static PyObject *__pyx_tuple__1374;
static PyObject *__pyx_tuple__1376;
static PyObject *__pyx_tuple__1378;
static PyObject *__pyx_tuple__1380;
static PyObject *__pyx_tuple__1382;
static PyObject *__pyx_tuple__1384;
static PyObject *__pyx_tuple__1386;
static PyObject *__pyx_tuple__1388;
static PyObject *__pyx_tuple__1389;
static PyObject *__pyx_tuple__1391;
static PyObject *__pyx_tuple__1393;
static PyObject *__pyx_tuple__1395;
static PyObject *__pyx_tuple__1397;
static PyObject *__pyx_tuple__1399;
static PyObject *__pyx_tuple__1401;
static PyObject *__pyx_tuple__1403;
static PyObject *__pyx_tuple__1405;
static PyObject *__pyx_tuple__1407;
static PyObject *__pyx_tuple__1409;
static PyObject *__pyx_tuple__1410;
static PyObject *__pyx_tuple__1412;
static PyObject *__pyx_tuple__1414;
static PyObject *__pyx_tuple__1415;
static PyObject *__pyx_tuple__1417;
static PyObject *__pyx_tuple__1419;
static PyObject *__pyx_tuple__1421;
static PyObject *__pyx_tuple__1423;
static PyObject *__pyx_tuple__1425;
static PyObject *__pyx_tuple__1427;
static PyObject *__pyx_tuple__1428;
static PyObject *__pyx_tuple__1430;
static PyObject *__pyx_tuple__1432;
static PyObject *__pyx_tuple__1434;
static PyObject *__pyx_tuple__1436;
static PyObject *__pyx_tuple__1438;
static PyObject *__pyx_tuple__1440;
static PyObject *__pyx_tuple__1442;
static PyObject *__pyx_tuple__1444;
static PyObject *__pyx_tuple__1446;
static PyObject *__pyx_tuple__1447;
static PyObject *__pyx_tuple__1449;
static PyObject *__pyx_tuple__1451;
static PyObject *__pyx_tuple__1453;
static PyObject *__pyx_tuple__1455;
static PyObject *__pyx_tuple__1457;
static PyObject *__pyx_tuple__1459;
static PyObject *__pyx_tuple__1461;
static PyObject *__pyx_tuple__1463;
static PyObject *__pyx_tuple__1464;
static PyObject *__pyx_tuple__1466;
static PyObject *__pyx_tuple__1468;
static PyObject *__pyx_tuple__1470;
static PyObject *__pyx_tuple__1472;
static PyObject *__pyx_tuple__1474;
static PyObject *__pyx_tuple__1476;
static PyObject *__pyx_tuple__1478;
static PyObject *__pyx_tuple__1480;
static PyObject *__pyx_tuple__1482;
static PyObject *__pyx_tuple__1484;
static PyObject *__pyx_tuple__1485;
static PyObject *__pyx_tuple__1487;
static PyObject *__pyx_tuple__1489;
static PyObject *__pyx_tuple__1491;
static PyObject *__pyx_tuple__1493;
static PyObject *__pyx_tuple__1495;
static PyObject *__pyx_tuple__1497;
static PyObject *__pyx_tuple__1498;
static PyObject *__pyx_tuple__1500;
static PyObject *__pyx_tuple__1502;
static PyObject *__pyx_tuple__1504;
static PyObject *__pyx_tuple__1506;
static PyObject *__pyx_tuple__1508;
static PyObject *__pyx_tuple__1510;
static PyObject *__pyx_tuple__1512;
static PyObject *__pyx_tuple__1513;
static PyObject *__pyx_tuple__1515;
static PyObject *__pyx_tuple__1516;
static PyObject *__pyx_tuple__1518;
static PyObject *__pyx_tuple__1520;
static PyObject *__pyx_tuple__1522;
static PyObject *__pyx_tuple__1524;
static PyObject *__pyx_tuple__1526;
static PyObject *__pyx_tuple__1528;
static PyObject *__pyx_tuple__1529;
static PyObject *__pyx_tuple__1531;
static PyObject *__pyx_tuple__1533;
static PyObject *__pyx_tuple__1535;
static PyObject *__pyx_tuple__1537;
static PyObject *__pyx_tuple__1539;
static PyObject *__pyx_tuple__1541;
static PyObject *__pyx_tuple__1543;
static PyObject *__pyx_tuple__1545;
static PyObject *__pyx_tuple__1547;
static PyObject *__pyx_tuple__1548;
static PyObject *__pyx_tuple__1550;
static PyObject *__pyx_tuple__1552;
static PyObject *__pyx_tuple__1554;
static PyObject *__pyx_tuple__1556;
static PyObject *__pyx_tuple__1558;
static PyObject *__pyx_tuple__1560;
static PyObject *__pyx_tuple__1562;
static PyObject *__pyx_tuple__1563;
static PyObject *__pyx_tuple__1565;
static PyObject *__pyx_tuple__1567;
static PyObject *__pyx_tuple__1569;
static PyObject *__pyx_tuple__1571;
static PyObject *__pyx_tuple__1573;
static PyObject *__pyx_tuple__1575;
static PyObject *__pyx_tuple__1577;
static PyObject *__pyx_tuple__1579;
static PyObject *__pyx_tuple__1581;
static PyObject *__pyx_tuple__1582;
static PyObject *__pyx_tuple__1584;
static PyObject *__pyx_tuple__1586;
static PyObject *__pyx_tuple__1588;
static PyObject *__pyx_tuple__1590;
static PyObject *__pyx_tuple__1592;
static PyObject *__pyx_tuple__1594;
static PyObject *__pyx_tuple__1596;
static PyObject *__pyx_tuple__1598;
static PyObject *__pyx_tuple__1600;
static PyObject *__pyx_tuple__1602;
static PyObject *__pyx_tuple__1604;
static PyObject *__pyx_tuple__1606;
static PyObject *__pyx_tuple__1608;
static PyObject *__pyx_tuple__1610;
static PyObject *__pyx_tuple__1612;
static PyObject *__pyx_tuple__1614;
static PyObject *__pyx_tuple__1616;
static PyObject *__pyx_tuple__1618;
static PyObject *__pyx_tuple__1620;
static PyObject *__pyx_tuple__1622;
static PyObject *__pyx_tuple__1623;
static PyObject *__pyx_tuple__1625;
static PyObject *__pyx_tuple__1627;
static PyObject *__pyx_tuple__1629;
static PyObject *__pyx_tuple__1631;
static PyObject *__pyx_tuple__1633;
static PyObject *__pyx_tuple__1635;
static PyObject *__pyx_tuple__1637;
static PyObject *__pyx_tuple__1638;
static PyObject *__pyx_tuple__1640;
static PyObject *__pyx_tuple__1641;
static PyObject *__pyx_tuple__1643;
static PyObject *__pyx_tuple__1645;
static PyObject *__pyx_tuple__1647;
static PyObject *__pyx_tuple__1649;
static PyObject *__pyx_tuple__1651;
static PyObject *__pyx_tuple__1653;
static PyObject *__pyx_tuple__1654;
static PyObject *__pyx_tuple__1656;
static PyObject *__pyx_tuple__1658;
static PyObject *__pyx_tuple__1660;
static PyObject *__pyx_tuple__1662;
static PyObject *__pyx_tuple__1664;
static PyObject *__pyx_tuple__1666;
static PyObject *__pyx_tuple__1668;
static PyObject *__pyx_tuple__1670;
static PyObject *__pyx_tuple__1671;
static PyObject *__pyx_tuple__1673;
static PyObject *__pyx_tuple__1675;
static PyObject *__pyx_tuple__1677;
static PyObject *__pyx_tuple__1679;
static PyObject *__pyx_tuple__1681;
static PyObject *__pyx_tuple__1683;
static PyObject *__pyx_tuple__1685;
static PyObject *__pyx_tuple__1687;
static PyObject *__pyx_tuple__1689;
static PyObject *__pyx_tuple__1691;
static PyObject *__pyx_tuple__1692;
static PyObject *__pyx_tuple__1694;
static PyObject *__pyx_tuple__1696;
static PyObject *__pyx_tuple__1698;
static PyObject *__pyx_tuple__1700;
static PyObject *__pyx_tuple__1702;
static PyObject *__pyx_tuple__1704;
static PyObject *__pyx_tuple__1706;
static PyObject *__pyx_tuple__1708;
static PyObject *__pyx_tuple__1710;
static PyObject *__pyx_tuple__1711;
static PyObject *__pyx_tuple__1713;
static PyObject *__pyx_tuple__1715;
static PyObject *__pyx_tuple__1717;
static PyObject *__pyx_tuple__1719;
static PyObject *__pyx_tuple__1721;
static PyObject *__pyx_tuple__1723;
static PyObject *__pyx_tuple__1725;
static PyObject *__pyx_tuple__1727;
static PyObject *__pyx_tuple__1729;
static PyObject *__pyx_tuple__1731;
static PyObject *__pyx_tuple__1733;
static PyObject *__pyx_tuple__1735;
static PyObject *__pyx_tuple__1736;
static PyObject *__pyx_tuple__1738;
static PyObject *__pyx_tuple__1740;
static PyObject *__pyx_tuple__1742;
static PyObject *__pyx_tuple__1744;
static PyObject *__pyx_tuple__1746;
static PyObject *__pyx_tuple__1748;
static PyObject *__pyx_tuple__1750;
static PyObject *__pyx_tuple__1752;
static PyObject *__pyx_tuple__1754;
static PyObject *__pyx_tuple__1755;
static PyObject *__pyx_tuple__1757;
static PyObject *__pyx_tuple__1759;
static PyObject *__pyx_tuple__1761;
static PyObject *__pyx_tuple__1763;
static PyObject *__pyx_tuple__1765;
static PyObject *__pyx_tuple__1767;
static PyObject *__pyx_tuple__1769;
static PyObject *__pyx_tuple__1771;
static PyObject *__pyx_tuple__1773;
static PyObject *__pyx_tuple__1774;
static PyObject *__pyx_tuple__1776;
static PyObject *__pyx_tuple__1778;
static PyObject *__pyx_tuple__1780;
static PyObject *__pyx_tuple__1782;
static PyObject *__pyx_tuple__1784;
static PyObject *__pyx_tuple__1786;
static PyObject *__pyx_tuple__1788;
static PyObject *__pyx_tuple__1790;
static PyObject *__pyx_tuple__1791;
static PyObject *__pyx_tuple__1793;
static PyObject *__pyx_tuple__1795;
static PyObject *__pyx_tuple__1797;
static PyObject *__pyx_tuple__1799;
static PyObject *__pyx_tuple__1801;
static PyObject *__pyx_tuple__1803;
static PyObject *__pyx_tuple__1805;
static PyObject *__pyx_tuple__1807;
static PyObject *__pyx_tuple__1809;
static PyObject *__pyx_tuple__1811;
static PyObject *__pyx_tuple__1813;
static PyObject *__pyx_tuple__1815;
static PyObject *__pyx_tuple__1817;
static PyObject *__pyx_tuple__1819;
static PyObject *__pyx_tuple__1820;
static PyObject *__pyx_tuple__1822;
static PyObject *__pyx_tuple__1824;
static PyObject *__pyx_tuple__1825;
static PyObject *__pyx_tuple__1827;
static PyObject *__pyx_tuple__1829;
static PyObject *__pyx_tuple__1831;
static PyObject *__pyx_tuple__1833;
static PyObject *__pyx_tuple__1835;
static PyObject *__pyx_tuple__1837;
static PyObject *__pyx_tuple__1838;
static PyObject *__pyx_tuple__1840;
static PyObject *__pyx_tuple__1842;
static PyObject *__pyx_tuple__1843;
static PyObject *__pyx_tuple__1845;
static PyObject *__pyx_tuple__1847;
static PyObject *__pyx_tuple__1849;
static PyObject *__pyx_tuple__1851;
static PyObject *__pyx_tuple__1853;
static PyObject *__pyx_tuple__1855;
static PyObject *__pyx_tuple__1856;
static PyObject *__pyx_tuple__1858;
static PyObject *__pyx_tuple__1860;
static PyObject *__pyx_tuple__1862;
static PyObject *__pyx_tuple__1864;
static PyObject *__pyx_tuple__1866;
static PyObject *__pyx_tuple__1868;
static PyObject *__pyx_tuple__1870;
static PyObject *__pyx_tuple__1872;
static PyObject *__pyx_tuple__1874;
static PyObject *__pyx_tuple__1876;
static PyObject *__pyx_tuple__1878;
static PyObject *__pyx_tuple__1880;
static PyObject *__pyx_tuple__1882;
static PyObject *__pyx_tuple__1884;
static PyObject *__pyx_tuple__1886;
static PyObject *__pyx_tuple__1888;
static PyObject *__pyx_tuple__1890;
static PyObject *__pyx_tuple__1892;
static PyObject *__pyx_tuple__1894;
static PyObject *__pyx_tuple__1896;
static PyObject *__pyx_tuple__1898;
static PyObject *__pyx_tuple__1900;
static PyObject *__pyx_tuple__1902;
static PyObject *__pyx_tuple__1904;
static PyObject *__pyx_tuple__1906;
static PyObject *__pyx_tuple__1908;
static PyObject *__pyx_tuple__1910;
static PyObject *__pyx_tuple__1912;
static PyObject *__pyx_tuple__1914;
static PyObject *__pyx_tuple__1916;
static PyObject *__pyx_tuple__1918;
static PyObject *__pyx_tuple__1920;
static PyObject *__pyx_tuple__1922;
static PyObject *__pyx_tuple__1924;
static PyObject *__pyx_tuple__1926;
static PyObject *__pyx_tuple__1928;
static PyObject *__pyx_tuple__1929;
static PyObject *__pyx_tuple__1931;
static PyObject *__pyx_tuple__1933;
static PyObject *__pyx_tuple__1935;
static PyObject *__pyx_tuple__1937;
static PyObject *__pyx_tuple__1939;
static PyObject *__pyx_tuple__1941;
static PyObject *__pyx_tuple__1943;
static PyObject *__pyx_tuple__1945;
static PyObject *__pyx_tuple__1946;
static PyObject *__pyx_tuple__1948;
static PyObject *__pyx_tuple__1950;
static PyObject *__pyx_tuple__1952;
static PyObject *__pyx_tuple__1954;
static PyObject *__pyx_tuple__1956;
static PyObject *__pyx_tuple__1958;
static PyObject *__pyx_tuple__1960;
static PyObject *__pyx_tuple__1962;
static PyObject *__pyx_tuple__1964;
static PyObject *__pyx_tuple__1966;
static PyObject *__pyx_tuple__1968;
static PyObject *__pyx_tuple__1970;
static PyObject *__pyx_tuple__1972;
static PyObject *__pyx_tuple__1973;
static PyObject *__pyx_tuple__1975;
static PyObject *__pyx_tuple__1977;
static PyObject *__pyx_tuple__1979;
static PyObject *__pyx_tuple__1981;
static PyObject *__pyx_tuple__1983;
static PyObject *__pyx_tuple__1985;
static PyObject *__pyx_tuple__1987;
static PyObject *__pyx_tuple__1989;
static PyObject *__pyx_tuple__1991;
static PyObject *__pyx_tuple__1993;
static PyObject *__pyx_tuple__1995;
static PyObject *__pyx_tuple__1997;
static PyObject *__pyx_tuple__1999;
static PyObject *__pyx_tuple__2001;
static PyObject *__pyx_tuple__2003;
static PyObject *__pyx_tuple__2005;
static PyObject *__pyx_tuple__2007;
static PyObject *__pyx_tuple__2009;
static PyObject *__pyx_tuple__2011;
static PyObject *__pyx_tuple__2013;
static PyObject *__pyx_tuple__2015;
static PyObject *__pyx_tuple__2017;
static PyObject *__pyx_tuple__2019;
static PyObject *__pyx_tuple__2021;
static PyObject *__pyx_tuple__2023;
static PyObject *__pyx_tuple__2025;
static PyObject *__pyx_tuple__2027;
static PyObject *__pyx_tuple__2029;
static PyObject *__pyx_tuple__2031;
static PyObject *__pyx_tuple__2033;
static PyObject *__pyx_tuple__2035;
static PyObject *__pyx_tuple__2037;
static PyObject *__pyx_tuple__2038;
static PyObject *__pyx_tuple__2040;
static PyObject *__pyx_tuple__2042;
static PyObject *__pyx_tuple__2044;
static PyObject *__pyx_tuple__2046;
static PyObject *__pyx_tuple__2048;
static PyObject *__pyx_tuple__2050;
static PyObject *__pyx_tuple__2052;
static PyObject *__pyx_tuple__2054;
static PyObject *__pyx_tuple__2055;
static PyObject *__pyx_tuple__2057;
static PyObject *__pyx_tuple__2058;
static PyObject *__pyx_tuple__2060;
static PyObject *__pyx_tuple__2061;
static PyObject *__pyx_tuple__2063;
static PyObject *__pyx_tuple__2065;
static PyObject *__pyx_tuple__2067;
static PyObject *__pyx_tuple__2069;
static PyObject *__pyx_tuple__2071;
static PyObject *__pyx_tuple__2073;
static PyObject *__pyx_tuple__2074;
static PyObject *__pyx_tuple__2076;
static PyObject *__pyx_tuple__2077;
static PyObject *__pyx_tuple__2079;
static PyObject *__pyx_tuple__2080;
static PyObject *__pyx_tuple__2082;
static PyObject *__pyx_tuple__2084;
static PyObject *__pyx_tuple__2086;
static PyObject *__pyx_tuple__2088;
static PyObject *__pyx_tuple__2090;
static PyObject *__pyx_tuple__2092;
static PyObject *__pyx_tuple__2093;
static PyObject *__pyx_tuple__2095;
static PyObject *__pyx_tuple__2097;
static PyObject *__pyx_tuple__2098;
static PyObject *__pyx_tuple__2100;
static PyObject *__pyx_tuple__2102;
static PyObject *__pyx_tuple__2104;
static PyObject *__pyx_tuple__2106;
static PyObject *__pyx_tuple__2108;
static PyObject *__pyx_tuple__2110;
static PyObject *__pyx_tuple__2111;
static PyObject *__pyx_tuple__2113;
static PyObject *__pyx_tuple__2115;
static PyObject *__pyx_tuple__2116;
static PyObject *__pyx_tuple__2118;
static PyObject *__pyx_tuple__2120;
static PyObject *__pyx_tuple__2122;
static PyObject *__pyx_tuple__2124;
static PyObject *__pyx_tuple__2126;
static PyObject *__pyx_tuple__2128;
static PyObject *__pyx_tuple__2129;
static PyObject *__pyx_tuple__2131;
static PyObject *__pyx_tuple__2132;
static PyObject *__pyx_tuple__2134;
static PyObject *__pyx_tuple__2136;
static PyObject *__pyx_tuple__2138;
static PyObject *__pyx_tuple__2140;
static PyObject *__pyx_tuple__2142;
static PyObject *__pyx_tuple__2144;
static PyObject *__pyx_tuple__2145;
static PyObject *__pyx_tuple__2147;
static PyObject *__pyx_tuple__2149;
static PyObject *__pyx_tuple__2151;
static PyObject *__pyx_tuple__2153;
static PyObject *__pyx_tuple__2155;
static PyObject *__pyx_tuple__2157;
static PyObject *__pyx_tuple__2159;
static PyObject *__pyx_tuple__2161;
static PyObject *__pyx_tuple__2162;
static PyObject *__pyx_tuple__2164;
static PyObject *__pyx_tuple__2166;
static PyObject *__pyx_tuple__2167;
static PyObject *__pyx_tuple__2169;
static PyObject *__pyx_tuple__2171;
static PyObject *__pyx_tuple__2173;
static PyObject *__pyx_tuple__2175;
static PyObject *__pyx_tuple__2177;
static PyObject *__pyx_tuple__2179;
static PyObject *__pyx_tuple__2180;
static PyObject *__pyx_tuple__2182;
static PyObject *__pyx_tuple__2183;
static PyObject *__pyx_tuple__2185;
static PyObject *__pyx_tuple__2186;
static PyObject *__pyx_tuple__2188;
static PyObject *__pyx_tuple__2190;
static PyObject *__pyx_tuple__2192;
static PyObject *__pyx_tuple__2194;
static PyObject *__pyx_tuple__2196;
static PyObject *__pyx_tuple__2198;
static PyObject *__pyx_tuple__2199;
static PyObject *__pyx_tuple__2201;
static PyObject *__pyx_tuple__2202;
static PyObject *__pyx_tuple__2204;
static PyObject *__pyx_tuple__2205;
static PyObject *__pyx_tuple__2207;
static PyObject *__pyx_tuple__2209;
static PyObject *__pyx_tuple__2211;
static PyObject *__pyx_tuple__2213;
static PyObject *__pyx_tuple__2215;
static PyObject *__pyx_tuple__2217;
static PyObject *__pyx_tuple__2218;
static PyObject *__pyx_tuple__2220;
static PyObject *__pyx_tuple__2221;
static PyObject *__pyx_tuple__2223;
static PyObject *__pyx_tuple__2225;
static PyObject *__pyx_tuple__2227;
static PyObject *__pyx_tuple__2229;
static PyObject *__pyx_tuple__2231;
static PyObject *__pyx_tuple__2233;
static PyObject *__pyx_tuple__2234;
static PyObject *__pyx_tuple__2236;
static PyObject *__pyx_tuple__2238;
static PyObject *__pyx_tuple__2240;
static PyObject *__pyx_tuple__2242;
static PyObject *__pyx_tuple__2243;
static PyObject *__pyx_tuple__2245;
static PyObject *__pyx_tuple__2247;
static PyObject *__pyx_tuple__2249;
static PyObject *__pyx_tuple__2251;
static PyObject *__pyx_tuple__2253;
static PyObject *__pyx_tuple__2255;
static PyObject *__pyx_tuple__2257;
static PyObject *__pyx_tuple__2259;
static PyObject *__pyx_tuple__2261;
static PyObject *__pyx_tuple__2263;
static PyObject *__pyx_tuple__2264;
static PyObject *__pyx_tuple__2266;
static PyObject *__pyx_tuple__2267;
static PyObject *__pyx_tuple__2269;
static PyObject *__pyx_tuple__2271;
static PyObject *__pyx_tuple__2273;
static PyObject *__pyx_tuple__2275;
static PyObject *__pyx_tuple__2277;
static PyObject *__pyx_tuple__2279;
static PyObject *__pyx_tuple__2280;
static PyObject *__pyx_tuple__2282;
static PyObject *__pyx_tuple__2284;
static PyObject *__pyx_tuple__2286;
static PyObject *__pyx_tuple__2288;
static PyObject *__pyx_tuple__2290;
static PyObject *__pyx_tuple__2292;
static PyObject *__pyx_tuple__2294;
static PyObject *__pyx_tuple__2296;
static PyObject *__pyx_tuple__2298;
static PyObject *__pyx_tuple__2299;
static PyObject *__pyx_tuple__2301;
static PyObject *__pyx_tuple__2303;
static PyObject *__pyx_tuple__2305;
static PyObject *__pyx_tuple__2307;
static PyObject *__pyx_tuple__2309;
static PyObject *__pyx_tuple__2310;
static PyObject *__pyx_tuple__2312;
static PyObject *__pyx_tuple__2314;
static PyObject *__pyx_tuple__2316;
static PyObject *__pyx_tuple__2318;
static PyObject *__pyx_tuple__2320;
static PyObject *__pyx_tuple__2321;
static PyObject *__pyx_tuple__2323;
static PyObject *__pyx_tuple__2325;
static PyObject *__pyx_tuple__2327;
static PyObject *__pyx_tuple__2329;
static PyObject *__pyx_tuple__2331;
static PyObject *__pyx_tuple__2333;
static PyObject *__pyx_tuple__2335;
static PyObject *__pyx_tuple__2337;
static PyObject *__pyx_tuple__2339;
static PyObject *__pyx_tuple__2341;
static PyObject *__pyx_tuple__2342;
static PyObject *__pyx_tuple__2344;
static PyObject *__pyx_tuple__2346;
static PyObject *__pyx_tuple__2348;
static PyObject *__pyx_tuple__2350;
static PyObject *__pyx_tuple__2352;
static PyObject *__pyx_tuple__2353;
static PyObject *__pyx_tuple__2355;
static PyObject *__pyx_tuple__2357;
static PyObject *__pyx_tuple__2359;
static PyObject *__pyx_tuple__2361;
static PyObject *__pyx_tuple__2363;
static PyObject *__pyx_tuple__2365;
static PyObject *__pyx_tuple__2367;
static PyObject *__pyx_tuple__2369;
static PyObject *__pyx_tuple__2371;
static PyObject *__pyx_tuple__2373;
static PyObject *__pyx_tuple__2374;
static PyObject *__pyx_tuple__2376;
static PyObject *__pyx_tuple__2378;
static PyObject *__pyx_tuple__2380;
static PyObject *__pyx_tuple__2382;
static PyObject *__pyx_tuple__2383;
static PyObject *__pyx_tuple__2385;
static PyObject *__pyx_tuple__2386;
static PyObject *__pyx_tuple__2388;
static PyObject *__pyx_tuple__2390;
static PyObject *__pyx_tuple__2392;
static PyObject *__pyx_tuple__2394;
static PyObject *__pyx_tuple__2396;
static PyObject *__pyx_tuple__2398;
static PyObject *__pyx_tuple__2400;
static PyObject *__pyx_tuple__2402;
static PyObject *__pyx_tuple__2404;
static PyObject *__pyx_tuple__2406;
static PyObject *__pyx_tuple__2408;
static PyObject *__pyx_tuple__2410;
static PyObject *__pyx_tuple__2412;
static PyObject *__pyx_tuple__2414;
static PyObject *__pyx_tuple__2416;
static PyObject *__pyx_tuple__2418;
static PyObject *__pyx_tuple__2420;
static PyObject *__pyx_tuple__2422;
static PyObject *__pyx_tuple__2424;
static PyObject *__pyx_tuple__2426;
static PyObject *__pyx_tuple__2428;
static PyObject *__pyx_tuple__2430;
static PyObject *__pyx_tuple__2432;
static PyObject *__pyx_tuple__2434;
static PyObject *__pyx_tuple__2436;
static PyObject *__pyx_tuple__2438;
static PyObject *__pyx_tuple__2440;
static PyObject *__pyx_tuple__2442;
static PyObject *__pyx_tuple__2444;
static PyObject *__pyx_tuple__2446;
static PyObject *__pyx_tuple__2448;
static PyObject *__pyx_tuple__2450;
static PyObject *__pyx_tuple__2452;
static PyObject *__pyx_tuple__2454;
static PyObject *__pyx_tuple__2456;
static PyObject *__pyx_tuple__2458;
static PyObject *__pyx_tuple__2460;
static PyObject *__pyx_tuple__2462;
static PyObject *__pyx_tuple__2464;
static PyObject *__pyx_tuple__2466;
static PyObject *__pyx_tuple__2468;
static PyObject *__pyx_tuple__2470;
static PyObject *__pyx_tuple__2471;
static PyObject *__pyx_tuple__2473;
static PyObject *__pyx_tuple__2475;
static PyObject *__pyx_tuple__2477;
static PyObject *__pyx_tuple__2479;
static PyObject *__pyx_tuple__2481;
static PyObject *__pyx_tuple__2483;
static PyObject *__pyx_tuple__2485;
static PyObject *__pyx_tuple__2487;
static PyObject *__pyx_tuple__2489;
static PyObject *__pyx_tuple__2491;
static PyObject *__pyx_tuple__2493;
static PyObject *__pyx_tuple__2495;
static PyObject *__pyx_tuple__2497;
static PyObject *__pyx_tuple__2499;
static PyObject *__pyx_tuple__2501;
static PyObject *__pyx_tuple__2503;
static PyObject *__pyx_tuple__2505;
static PyObject *__pyx_tuple__2507;
static PyObject *__pyx_tuple__2509;
static PyObject *__pyx_tuple__2511;
static PyObject *__pyx_tuple__2512;
static PyObject *__pyx_tuple__2514;
static PyObject *__pyx_tuple__2515;
static PyObject *__pyx_tuple__2517;
static PyObject *__pyx_tuple__2519;
static PyObject *__pyx_tuple__2521;
static PyObject *__pyx_tuple__2523;
static PyObject *__pyx_tuple__2525;
static PyObject *__pyx_tuple__2527;
static PyObject *__pyx_tuple__2528;
static PyObject *__pyx_tuple__2530;
static PyObject *__pyx_tuple__2532;
static PyObject *__pyx_tuple__2534;
static PyObject *__pyx_tuple__2536;
static PyObject *__pyx_tuple__2538;
static PyObject *__pyx_tuple__2540;
static PyObject *__pyx_tuple__2542;
static PyObject *__pyx_tuple__2544;
static PyObject *__pyx_tuple__2545;
static PyObject *__pyx_tuple__2547;
static PyObject *__pyx_tuple__2548;
static PyObject *__pyx_tuple__2550;
static PyObject *__pyx_tuple__2552;
static PyObject *__pyx_tuple__2554;
static PyObject *__pyx_tuple__2556;
static PyObject *__pyx_tuple__2558;
static PyObject *__pyx_tuple__2560;
static PyObject *__pyx_tuple__2561;
static PyObject *__pyx_tuple__2563;
static PyObject *__pyx_tuple__2565;
static PyObject *__pyx_tuple__2567;
static PyObject *__pyx_tuple__2569;
static PyObject *__pyx_tuple__2571;
static PyObject *__pyx_tuple__2573;
static PyObject *__pyx_tuple__2575;
static PyObject *__pyx_tuple__2576;
static PyObject *__pyx_tuple__2578;
static PyObject *__pyx_tuple__2579;
static PyObject *__pyx_tuple__2581;
static PyObject *__pyx_tuple__2582;
static PyObject *__pyx_tuple__2584;
static PyObject *__pyx_tuple__2586;
static PyObject *__pyx_tuple__2588;
static PyObject *__pyx_tuple__2590;
static PyObject *__pyx_tuple__2592;
static PyObject *__pyx_tuple__2594;
static PyObject *__pyx_tuple__2596;
static PyObject *__pyx_tuple__2597;
static PyObject *__pyx_tuple__2599;
static PyObject *__pyx_tuple__2600;
static PyObject *__pyx_tuple__2602;
static PyObject *__pyx_tuple__2604;
static PyObject *__pyx_tuple__2606;
static PyObject *__pyx_tuple__2608;
static PyObject *__pyx_tuple__2610;
static PyObject *__pyx_tuple__2612;
static PyObject *__pyx_tuple__2613;
static PyObject *__pyx_tuple__2615;
static PyObject *__pyx_tuple__2617;
static PyObject *__pyx_tuple__2619;
static PyObject *__pyx_tuple__2621;
static PyObject *__pyx_tuple__2623;
static PyObject *__pyx_tuple__2625;
static PyObject *__pyx_tuple__2627;
static PyObject *__pyx_tuple__2629;
static PyObject *__pyx_tuple__2630;
static PyObject *__pyx_tuple__2632;
static PyObject *__pyx_tuple__2634;
static PyObject *__pyx_tuple__2636;
static PyObject *__pyx_tuple__2638;
static PyObject *__pyx_tuple__2640;
static PyObject *__pyx_tuple__2642;
static PyObject *__pyx_tuple__2644;
static PyObject *__pyx_tuple__2646;
static PyObject *__pyx_tuple__2648;
static PyObject *__pyx_tuple__2650;
static PyObject *__pyx_tuple__2652;
static PyObject *__pyx_tuple__2654;
static PyObject *__pyx_tuple__2656;
static PyObject *__pyx_tuple__2658;
static PyObject *__pyx_tuple__2660;
static PyObject *__pyx_tuple__2661;
static PyObject *__pyx_tuple__2663;
static PyObject *__pyx_tuple__2664;
static PyObject *__pyx_tuple__2666;
static PyObject *__pyx_tuple__2668;
static PyObject *__pyx_tuple__2670;
static PyObject *__pyx_tuple__2672;
static PyObject *__pyx_tuple__2674;
static PyObject *__pyx_tuple__2676;
static PyObject *__pyx_tuple__2677;
static PyObject *__pyx_tuple__2679;
static PyObject *__pyx_tuple__2680;
static PyObject *__pyx_tuple__2682;
static PyObject *__pyx_tuple__2684;
static PyObject *__pyx_tuple__2686;
static PyObject *__pyx_tuple__2688;
static PyObject *__pyx_tuple__2690;
static PyObject *__pyx_tuple__2692;
static PyObject *__pyx_tuple__2693;
static PyObject *__pyx_tuple__2695;
static PyObject *__pyx_tuple__2696;
static PyObject *__pyx_tuple__2698;
static PyObject *__pyx_tuple__2700;
static PyObject *__pyx_tuple__2702;
static PyObject *__pyx_tuple__2704;
static PyObject *__pyx_tuple__2706;
static PyObject *__pyx_tuple__2708;
static PyObject *__pyx_tuple__2709;
static PyObject *__pyx_tuple__2711;
static PyObject *__pyx_tuple__2713;
static PyObject *__pyx_tuple__2715;
static PyObject *__pyx_tuple__2717;
static PyObject *__pyx_tuple__2719;
static PyObject *__pyx_tuple__2721;
static PyObject *__pyx_tuple__2723;
static PyObject *__pyx_tuple__2724;
static PyObject *__pyx_tuple__2726;
static PyObject *__pyx_tuple__2728;
static PyObject *__pyx_tuple__2730;
static PyObject *__pyx_tuple__2732;
static PyObject *__pyx_tuple__2734;
static PyObject *__pyx_tuple__2736;
static PyObject *__pyx_tuple__2738;
static PyObject *__pyx_tuple__2740;
static PyObject *__pyx_tuple__2741;
static PyObject *__pyx_tuple__2743;
static PyObject *__pyx_tuple__2744;
static PyObject *__pyx_tuple__2746;
static PyObject *__pyx_tuple__2748;
static PyObject *__pyx_tuple__2750;
static PyObject *__pyx_tuple__2752;
static PyObject *__pyx_tuple__2754;
static PyObject *__pyx_tuple__2756;
static PyObject *__pyx_tuple__2757;
static PyObject *__pyx_tuple__2759;
static PyObject *__pyx_tuple__2761;
static PyObject *__pyx_tuple__2763;
static PyObject *__pyx_tuple__2765;
static PyObject *__pyx_tuple__2767;
static PyObject *__pyx_tuple__2769;
static PyObject *__pyx_tuple__2771;
static PyObject *__pyx_codeobj__977;
static PyObject *__pyx_codeobj__1007;
static PyObject *__pyx_codeobj__1009;
static PyObject *__pyx_codeobj__1012;
static PyObject *__pyx_codeobj__1014;
static PyObject *__pyx_codeobj__1017;
static PyObject *__pyx_codeobj__1019;
static PyObject *__pyx_codeobj__1021;
static PyObject *__pyx_codeobj__1023;
static PyObject *__pyx_codeobj__1025;
static PyObject *__pyx_codeobj__1027;
static PyObject *__pyx_codeobj__1029;
static PyObject *__pyx_codeobj__1032;
static PyObject *__pyx_codeobj__1034;
static PyObject *__pyx_codeobj__1036;
static PyObject *__pyx_codeobj__1038;
static PyObject *__pyx_codeobj__1040;
static PyObject *__pyx_codeobj__1042;
static PyObject *__pyx_codeobj__1044;
static PyObject *__pyx_codeobj__1046;
static PyObject *__pyx_codeobj__1049;
static PyObject *__pyx_codeobj__1051;
static PyObject *__pyx_codeobj__1053;
static PyObject *__pyx_codeobj__1055;
static PyObject *__pyx_codeobj__1057;
static PyObject *__pyx_codeobj__1059;
static PyObject *__pyx_codeobj__1061;
static PyObject *__pyx_codeobj__1063;
static PyObject *__pyx_codeobj__1065;
static PyObject *__pyx_codeobj__1068;
static PyObject *__pyx_codeobj__1070;
static PyObject *__pyx_codeobj__1072;
static PyObject *__pyx_codeobj__1074;
static PyObject *__pyx_codeobj__1076;
static PyObject *__pyx_codeobj__1078;
static PyObject *__pyx_codeobj__1080;
static PyObject *__pyx_codeobj__1082;
static PyObject *__pyx_codeobj__1085;
static PyObject *__pyx_codeobj__1087;
static PyObject *__pyx_codeobj__1089;
static PyObject *__pyx_codeobj__1091;
static PyObject *__pyx_codeobj__1093;
static PyObject *__pyx_codeobj__1095;
static PyObject *__pyx_codeobj__1098;
static PyObject *__pyx_codeobj__1100;
static PyObject *__pyx_codeobj__1102;
static PyObject *__pyx_codeobj__1104;
static PyObject *__pyx_codeobj__1107;
static PyObject *__pyx_codeobj__1109;
static PyObject *__pyx_codeobj__1111;
static PyObject *__pyx_codeobj__1113;
static PyObject *__pyx_codeobj__1115;
static PyObject *__pyx_codeobj__1117;
static PyObject *__pyx_codeobj__1119;
static PyObject *__pyx_codeobj__1121;
static PyObject *__pyx_codeobj__1123;
static PyObject *__pyx_codeobj__1125;
static PyObject *__pyx_codeobj__1127;
static PyObject *__pyx_codeobj__1129;
static PyObject *__pyx_codeobj__1131;
static PyObject *__pyx_codeobj__1133;
static PyObject *__pyx_codeobj__1135;
static PyObject *__pyx_codeobj__1137;
static PyObject *__pyx_codeobj__1139;
static PyObject *__pyx_codeobj__1141;
static PyObject *__pyx_codeobj__1143;
static PyObject *__pyx_codeobj__1145;
static PyObject *__pyx_codeobj__1147;
static PyObject *__pyx_codeobj__1149;
static PyObject *__pyx_codeobj__1151;
static PyObject *__pyx_codeobj__1153;
static PyObject *__pyx_codeobj__1156;
static PyObject *__pyx_codeobj__1158;
static PyObject *__pyx_codeobj__1160;
static PyObject *__pyx_codeobj__1162;
static PyObject *__pyx_codeobj__1164;
static PyObject *__pyx_codeobj__1166;
static PyObject *__pyx_codeobj__1168;
static PyObject *__pyx_codeobj__1171;
static PyObject *__pyx_codeobj__1173;
static PyObject *__pyx_codeobj__1175;
static PyObject *__pyx_codeobj__1177;
static PyObject *__pyx_codeobj__1179;
static PyObject *__pyx_codeobj__1181;
static PyObject *__pyx_codeobj__1183;
static PyObject *__pyx_codeobj__1185;
static PyObject *__pyx_codeobj__1188;
static PyObject *__pyx_codeobj__1190;
static PyObject *__pyx_codeobj__1192;
static PyObject *__pyx_codeobj__1194;
static PyObject *__pyx_codeobj__1196;
static PyObject *__pyx_codeobj__1198;
static PyObject *__pyx_codeobj__1200;
static PyObject *__pyx_codeobj__1203;
static PyObject *__pyx_codeobj__1206;
static PyObject *__pyx_codeobj__1208;
static PyObject *__pyx_codeobj__1210;
static PyObject *__pyx_codeobj__1212;
static PyObject *__pyx_codeobj__1214;
static PyObject *__pyx_codeobj__1216;
static PyObject *__pyx_codeobj__1219;
static PyObject *__pyx_codeobj__1221;
static PyObject *__pyx_codeobj__1223;
static PyObject *__pyx_codeobj__1225;
static PyObject *__pyx_codeobj__1227;
static PyObject *__pyx_codeobj__1229;
static PyObject *__pyx_codeobj__1231;
static PyObject *__pyx_codeobj__1233;
static PyObject *__pyx_codeobj__1236;
static PyObject *__pyx_codeobj__1239;
static PyObject *__pyx_codeobj__1242;
static PyObject *__pyx_codeobj__1245;
static PyObject *__pyx_codeobj__1248;
static PyObject *__pyx_codeobj__1250;
static PyObject *__pyx_codeobj__1252;
static PyObject *__pyx_codeobj__1254;
static PyObject *__pyx_codeobj__1256;
static PyObject *__pyx_codeobj__1258;
static PyObject *__pyx_codeobj__1260;
static PyObject *__pyx_codeobj__1262;
static PyObject *__pyx_codeobj__1265;
static PyObject *__pyx_codeobj__1267;
static PyObject *__pyx_codeobj__1269;
static PyObject *__pyx_codeobj__1271;
static PyObject *__pyx_codeobj__1273;
static PyObject *__pyx_codeobj__1275;
static PyObject *__pyx_codeobj__1277;
static PyObject *__pyx_codeobj__1280;
static PyObject *__pyx_codeobj__1282;
static PyObject *__pyx_codeobj__1284;
static PyObject *__pyx_codeobj__1286;
static PyObject *__pyx_codeobj__1288;
static PyObject *__pyx_codeobj__1290;
static PyObject *__pyx_codeobj__1292;
static PyObject *__pyx_codeobj__1294;
static PyObject *__pyx_codeobj__1296;
static PyObject *__pyx_codeobj__1299;
static PyObject *__pyx_codeobj__1302;
static PyObject *__pyx_codeobj__1304;
static PyObject *__pyx_codeobj__1306;
static PyObject *__pyx_codeobj__1308;
static PyObject *__pyx_codeobj__1310;
static PyObject *__pyx_codeobj__1312;
static PyObject *__pyx_codeobj__1315;
static PyObject *__pyx_codeobj__1317;
static PyObject *__pyx_codeobj__1319;
static PyObject *__pyx_codeobj__1321;
static PyObject *__pyx_codeobj__1323;
static PyObject *__pyx_codeobj__1325;
static PyObject *__pyx_codeobj__1327;
static PyObject *__pyx_codeobj__1329;
static PyObject *__pyx_codeobj__1331;
static PyObject *__pyx_codeobj__1333;
static PyObject *__pyx_codeobj__1336;
static PyObject *__pyx_codeobj__1338;
static PyObject *__pyx_codeobj__1340;
static PyObject *__pyx_codeobj__1342;
static PyObject *__pyx_codeobj__1344;
static PyObject *__pyx_codeobj__1346;
static PyObject *__pyx_codeobj__1348;
static PyObject *__pyx_codeobj__1350;
static PyObject *__pyx_codeobj__1352;
static PyObject *__pyx_codeobj__1354;
static PyObject *__pyx_codeobj__1356;
static PyObject *__pyx_codeobj__1358;
static PyObject *__pyx_codeobj__1361;
static PyObject *__pyx_codeobj__1363;
static PyObject *__pyx_codeobj__1365;
static PyObject *__pyx_codeobj__1367;
static PyObject *__pyx_codeobj__1369;
static PyObject *__pyx_codeobj__1371;
static PyObject *__pyx_codeobj__1373;
static PyObject *__pyx_codeobj__1375;
static PyObject *__pyx_codeobj__1377;
static PyObject *__pyx_codeobj__1379;
static PyObject *__pyx_codeobj__1381;
static PyObject *__pyx_codeobj__1383;
static PyObject *__pyx_codeobj__1385;
static PyObject *__pyx_codeobj__1387;
static PyObject *__pyx_codeobj__1390;
static PyObject *__pyx_codeobj__1392;
static PyObject *__pyx_codeobj__1394;
static PyObject *__pyx_codeobj__1396;
static PyObject *__pyx_codeobj__1398;
static PyObject *__pyx_codeobj__1400;
static PyObject *__pyx_codeobj__1402;
static PyObject *__pyx_codeobj__1404;
static PyObject *__pyx_codeobj__1406;
static PyObject *__pyx_codeobj__1408;
static PyObject *__pyx_codeobj__1411;
static PyObject *__pyx_codeobj__1413;
static PyObject *__pyx_codeobj__1416;
static PyObject *__pyx_codeobj__1418;
static PyObject *__pyx_codeobj__1420;
static PyObject *__pyx_codeobj__1422;
static PyObject *__pyx_codeobj__1424;
static PyObject *__pyx_codeobj__1426;
static PyObject *__pyx_codeobj__1429;
static PyObject *__pyx_codeobj__1431;
static PyObject *__pyx_codeobj__1433;
static PyObject *__pyx_codeobj__1435;
static PyObject *__pyx_codeobj__1437;
static PyObject *__pyx_codeobj__1439;
static PyObject *__pyx_codeobj__1441;
static PyObject *__pyx_codeobj__1443;
static PyObject *__pyx_codeobj__1445;
static PyObject *__pyx_codeobj__1448;
static PyObject *__pyx_codeobj__1450;
static PyObject *__pyx_codeobj__1452;
static PyObject *__pyx_codeobj__1454;
static PyObject *__pyx_codeobj__1456;
static PyObject *__pyx_codeobj__1458;
static PyObject *__pyx_codeobj__1460;
static PyObject *__pyx_codeobj__1462;
static PyObject *__pyx_codeobj__1465;
static PyObject *__pyx_codeobj__1467;
static PyObject *__pyx_codeobj__1469;
static PyObject *__pyx_codeobj__1471;
static PyObject *__pyx_codeobj__1473;
static PyObject *__pyx_codeobj__1475;
static PyObject *__pyx_codeobj__1477;
static PyObject *__pyx_codeobj__1479;
static PyObject *__pyx_codeobj__1481;
static PyObject *__pyx_codeobj__1483;
static PyObject *__pyx_codeobj__1486;
static PyObject *__pyx_codeobj__1488;
static PyObject *__pyx_codeobj__1490;
static PyObject *__pyx_codeobj__1492;
static PyObject *__pyx_codeobj__1494;
static PyObject *__pyx_codeobj__1496;
static PyObject *__pyx_codeobj__1499;
static PyObject *__pyx_codeobj__1501;
static PyObject *__pyx_codeobj__1503;
static PyObject *__pyx_codeobj__1505;
static PyObject *__pyx_codeobj__1507;
static PyObject *__pyx_codeobj__1509;
static PyObject *__pyx_codeobj__1511;
static PyObject *__pyx_codeobj__1514;
static PyObject *__pyx_codeobj__1517;
static PyObject *__pyx_codeobj__1519;
static PyObject *__pyx_codeobj__1521;
static PyObject *__pyx_codeobj__1523;
static PyObject *__pyx_codeobj__1525;
static PyObject *__pyx_codeobj__1527;
static PyObject *__pyx_codeobj__1530;
static PyObject *__pyx_codeobj__1532;
static PyObject *__pyx_codeobj__1534;
static PyObject *__pyx_codeobj__1536;
static PyObject *__pyx_codeobj__1538;
static PyObject *__pyx_codeobj__1540;
static PyObject *__pyx_codeobj__1542;
static PyObject *__pyx_codeobj__1544;
static PyObject *__pyx_codeobj__1546;
static PyObject *__pyx_codeobj__1549;
static PyObject *__pyx_codeobj__1551;
static PyObject *__pyx_codeobj__1553;
static PyObject *__pyx_codeobj__1555;
static PyObject *__pyx_codeobj__1557;
static PyObject *__pyx_codeobj__1559;
static PyObject *__pyx_codeobj__1561;
static PyObject *__pyx_codeobj__1564;
static PyObject *__pyx_codeobj__1566;
static PyObject *__pyx_codeobj__1568;
static PyObject *__pyx_codeobj__1570;
static PyObject *__pyx_codeobj__1572;
static PyObject *__pyx_codeobj__1574;
static PyObject *__pyx_codeobj__1576;
static PyObject *__pyx_codeobj__1578;
static PyObject *__pyx_codeobj__1580;
static PyObject *__pyx_codeobj__1583;
static PyObject *__pyx_codeobj__1585;
static PyObject *__pyx_codeobj__1587;
static PyObject *__pyx_codeobj__1589;
static PyObject *__pyx_codeobj__1591;
static PyObject *__pyx_codeobj__1593;
static PyObject *__pyx_codeobj__1595;
static PyObject *__pyx_codeobj__1597;
static PyObject *__pyx_codeobj__1599;
static PyObject *__pyx_codeobj__1601;
static PyObject *__pyx_codeobj__1603;
static PyObject *__pyx_codeobj__1605;
static PyObject *__pyx_codeobj__1607;
static PyObject *__pyx_codeobj__1609;
static PyObject *__pyx_codeobj__1611;
static PyObject *__pyx_codeobj__1613;
static PyObject *__pyx_codeobj__1615;
static PyObject *__pyx_codeobj__1617;
static PyObject *__pyx_codeobj__1619;
static PyObject *__pyx_codeobj__1621;
static PyObject *__pyx_codeobj__1624;
static PyObject *__pyx_codeobj__1626;
static PyObject *__pyx_codeobj__1628;
static PyObject *__pyx_codeobj__1630;
static PyObject *__pyx_codeobj__1632;
static PyObject *__pyx_codeobj__1634;
static PyObject *__pyx_codeobj__1636;
static PyObject *__pyx_codeobj__1639;
static PyObject *__pyx_codeobj__1642;
static PyObject *__pyx_codeobj__1644;
static PyObject *__pyx_codeobj__1646;
static PyObject *__pyx_codeobj__1648;
static PyObject *__pyx_codeobj__1650;
static PyObject *__pyx_codeobj__1652;
static PyObject *__pyx_codeobj__1655;
static PyObject *__pyx_codeobj__1657;
static PyObject *__pyx_codeobj__1659;
static PyObject *__pyx_codeobj__1661;
static PyObject *__pyx_codeobj__1663;
static PyObject *__pyx_codeobj__1665;
static PyObject *__pyx_codeobj__1667;
static PyObject *__pyx_codeobj__1669;
static PyObject *__pyx_codeobj__1672;
static PyObject *__pyx_codeobj__1674;
static PyObject *__pyx_codeobj__1676;
static PyObject *__pyx_codeobj__1678;
static PyObject *__pyx_codeobj__1680;
static PyObject *__pyx_codeobj__1682;
static PyObject *__pyx_codeobj__1684;
static PyObject *__pyx_codeobj__1686;
static PyObject *__pyx_codeobj__1688;
static PyObject *__pyx_codeobj__1690;
static PyObject *__pyx_codeobj__1693;
static PyObject *__pyx_codeobj__1695;
static PyObject *__pyx_codeobj__1697;
static PyObject *__pyx_codeobj__1699;
static PyObject *__pyx_codeobj__1701;
static PyObject *__pyx_codeobj__1703;
static PyObject *__pyx_codeobj__1705;
static PyObject *__pyx_codeobj__1707;
static PyObject *__pyx_codeobj__1709;
static PyObject *__pyx_codeobj__1712;
static PyObject *__pyx_codeobj__1714;
static PyObject *__pyx_codeobj__1716;
static PyObject *__pyx_codeobj__1718;
static PyObject *__pyx_codeobj__1720;
static PyObject *__pyx_codeobj__1722;
static PyObject *__pyx_codeobj__1724;
static PyObject *__pyx_codeobj__1726;
static PyObject *__pyx_codeobj__1728;
static PyObject *__pyx_codeobj__1730;
static PyObject *__pyx_codeobj__1732;
static PyObject *__pyx_codeobj__1734;
static PyObject *__pyx_codeobj__1737;
static PyObject *__pyx_codeobj__1739;
static PyObject *__pyx_codeobj__1741;
static PyObject *__pyx_codeobj__1743;
static PyObject *__pyx_codeobj__1745;
static PyObject *__pyx_codeobj__1747;
static PyObject *__pyx_codeobj__1749;
static PyObject *__pyx_codeobj__1751;
static PyObject *__pyx_codeobj__1753;
static PyObject *__pyx_codeobj__1756;
static PyObject *__pyx_codeobj__1758;
static PyObject *__pyx_codeobj__1760;
static PyObject *__pyx_codeobj__1762;
static PyObject *__pyx_codeobj__1764;
static PyObject *__pyx_codeobj__1766;
static PyObject *__pyx_codeobj__1768;
static PyObject *__pyx_codeobj__1770;
static PyObject *__pyx_codeobj__1772;
static PyObject *__pyx_codeobj__1775;
static PyObject *__pyx_codeobj__1777;
static PyObject *__pyx_codeobj__1779;
static PyObject *__pyx_codeobj__1781;
static PyObject *__pyx_codeobj__1783;
static PyObject *__pyx_codeobj__1785;
static PyObject *__pyx_codeobj__1787;
static PyObject *__pyx_codeobj__1789;
static PyObject *__pyx_codeobj__1792;
static PyObject *__pyx_codeobj__1794;
static PyObject *__pyx_codeobj__1796;
static PyObject *__pyx_codeobj__1798;
static PyObject *__pyx_codeobj__1800;
static PyObject *__pyx_codeobj__1802;
static PyObject *__pyx_codeobj__1804;
static PyObject *__pyx_codeobj__1806;
static PyObject *__pyx_codeobj__1808;
static PyObject *__pyx_codeobj__1810;
static PyObject *__pyx_codeobj__1812;
static PyObject *__pyx_codeobj__1814;
static PyObject *__pyx_codeobj__1816;
static PyObject *__pyx_codeobj__1818;
static PyObject *__pyx_codeobj__1821;
static PyObject *__pyx_codeobj__1823;
static PyObject *__pyx_codeobj__1826;
static PyObject *__pyx_codeobj__1828;
static PyObject *__pyx_codeobj__1830;
static PyObject *__pyx_codeobj__1832;
static PyObject *__pyx_codeobj__1834;
static PyObject *__pyx_codeobj__1836;
static PyObject *__pyx_codeobj__1839;
static PyObject *__pyx_codeobj__1841;
static PyObject *__pyx_codeobj__1844;
static PyObject *__pyx_codeobj__1846;
static PyObject *__pyx_codeobj__1848;
static PyObject *__pyx_codeobj__1850;
static PyObject *__pyx_codeobj__1852;
static PyObject *__pyx_codeobj__1854;
static PyObject *__pyx_codeobj__1857;
static PyObject *__pyx_codeobj__1859;
static PyObject *__pyx_codeobj__1861;
static PyObject *__pyx_codeobj__1863;
static PyObject *__pyx_codeobj__1865;
static PyObject *__pyx_codeobj__1867;
static PyObject *__pyx_codeobj__1869;
static PyObject *__pyx_codeobj__1871;
static PyObject *__pyx_codeobj__1873;
static PyObject *__pyx_codeobj__1875;
static PyObject *__pyx_codeobj__1877;
static PyObject *__pyx_codeobj__1879;
static PyObject *__pyx_codeobj__1881;
static PyObject *__pyx_codeobj__1883;
static PyObject *__pyx_codeobj__1885;
static PyObject *__pyx_codeobj__1887;
static PyObject *__pyx_codeobj__1889;
static PyObject *__pyx_codeobj__1891;
static PyObject *__pyx_codeobj__1893;
static PyObject *__pyx_codeobj__1895;
static PyObject *__pyx_codeobj__1897;
static PyObject *__pyx_codeobj__1899;
static PyObject *__pyx_codeobj__1901;
static PyObject *__pyx_codeobj__1903;
static PyObject *__pyx_codeobj__1905;
static PyObject *__pyx_codeobj__1907;
static PyObject *__pyx_codeobj__1909;
static PyObject *__pyx_codeobj__1911;
static PyObject *__pyx_codeobj__1913;
static PyObject *__pyx_codeobj__1915;
static PyObject *__pyx_codeobj__1917;
static PyObject *__pyx_codeobj__1919;
static PyObject *__pyx_codeobj__1921;
static PyObject *__pyx_codeobj__1923;
static PyObject *__pyx_codeobj__1925;
static PyObject *__pyx_codeobj__1927;
static PyObject *__pyx_codeobj__1930;
static PyObject *__pyx_codeobj__1932;
static PyObject *__pyx_codeobj__1934;
static PyObject *__pyx_codeobj__1936;
static PyObject *__pyx_codeobj__1938;
static PyObject *__pyx_codeobj__1940;
static PyObject *__pyx_codeobj__1942;
static PyObject *__pyx_codeobj__1944;
static PyObject *__pyx_codeobj__1947;
static PyObject *__pyx_codeobj__1949;
static PyObject *__pyx_codeobj__1951;
static PyObject *__pyx_codeobj__1953;
static PyObject *__pyx_codeobj__1955;
static PyObject *__pyx_codeobj__1957;
static PyObject *__pyx_codeobj__1959;
static PyObject *__pyx_codeobj__1961;
static PyObject *__pyx_codeobj__1963;
static PyObject *__pyx_codeobj__1965;
static PyObject *__pyx_codeobj__1967;
static PyObject *__pyx_codeobj__1969;
static PyObject *__pyx_codeobj__1971;
static PyObject *__pyx_codeobj__1974;
static PyObject *__pyx_codeobj__1976;
static PyObject *__pyx_codeobj__1978;
static PyObject *__pyx_codeobj__1980;
static PyObject *__pyx_codeobj__1982;
static PyObject *__pyx_codeobj__1984;
static PyObject *__pyx_codeobj__1986;
static PyObject *__pyx_codeobj__1988;
static PyObject *__pyx_codeobj__1990;
static PyObject *__pyx_codeobj__1992;
static PyObject *__pyx_codeobj__1994;
static PyObject *__pyx_codeobj__1996;
static PyObject *__pyx_codeobj__1998;
static PyObject *__pyx_codeobj__2000;
static PyObject *__pyx_codeobj__2002;
static PyObject *__pyx_codeobj__2004;
static PyObject *__pyx_codeobj__2006;
static PyObject *__pyx_codeobj__2008;
static PyObject *__pyx_codeobj__2010;
static PyObject *__pyx_codeobj__2012;
static PyObject *__pyx_codeobj__2014;
static PyObject *__pyx_codeobj__2016;
static PyObject *__pyx_codeobj__2018;
static PyObject *__pyx_codeobj__2020;
static PyObject *__pyx_codeobj__2022;
static PyObject *__pyx_codeobj__2024;
static PyObject *__pyx_codeobj__2026;
static PyObject *__pyx_codeobj__2028;
static PyObject *__pyx_codeobj__2030;
static PyObject *__pyx_codeobj__2032;
static PyObject *__pyx_codeobj__2034;
static PyObject *__pyx_codeobj__2036;
static PyObject *__pyx_codeobj__2039;
static PyObject *__pyx_codeobj__2041;
static PyObject *__pyx_codeobj__2043;
static PyObject *__pyx_codeobj__2045;
static PyObject *__pyx_codeobj__2047;
static PyObject *__pyx_codeobj__2049;
static PyObject *__pyx_codeobj__2051;
static PyObject *__pyx_codeobj__2053;
static PyObject *__pyx_codeobj__2056;
static PyObject *__pyx_codeobj__2059;
static PyObject *__pyx_codeobj__2062;
static PyObject *__pyx_codeobj__2064;
static PyObject *__pyx_codeobj__2066;
static PyObject *__pyx_codeobj__2068;
static PyObject *__pyx_codeobj__2070;
static PyObject *__pyx_codeobj__2072;
static PyObject *__pyx_codeobj__2075;
static PyObject *__pyx_codeobj__2078;
static PyObject *__pyx_codeobj__2081;
static PyObject *__pyx_codeobj__2083;
static PyObject *__pyx_codeobj__2085;
static PyObject *__pyx_codeobj__2087;
static PyObject *__pyx_codeobj__2089;
static PyObject *__pyx_codeobj__2091;
static PyObject *__pyx_codeobj__2094;
static PyObject *__pyx_codeobj__2096;
static PyObject *__pyx_codeobj__2099;
static PyObject *__pyx_codeobj__2101;
static PyObject *__pyx_codeobj__2103;
static PyObject *__pyx_codeobj__2105;
static PyObject *__pyx_codeobj__2107;
static PyObject *__pyx_codeobj__2109;
static PyObject *__pyx_codeobj__2112;
static PyObject *__pyx_codeobj__2114;
static PyObject *__pyx_codeobj__2117;
static PyObject *__pyx_codeobj__2119;
static PyObject *__pyx_codeobj__2121;
static PyObject *__pyx_codeobj__2123;
static PyObject *__pyx_codeobj__2125;
static PyObject *__pyx_codeobj__2127;
static PyObject *__pyx_codeobj__2130;
static PyObject *__pyx_codeobj__2133;
static PyObject *__pyx_codeobj__2135;
static PyObject *__pyx_codeobj__2137;
static PyObject *__pyx_codeobj__2139;
static PyObject *__pyx_codeobj__2141;
static PyObject *__pyx_codeobj__2143;
static PyObject *__pyx_codeobj__2146;
static PyObject *__pyx_codeobj__2148;
static PyObject *__pyx_codeobj__2150;
static PyObject *__pyx_codeobj__2152;
static PyObject *__pyx_codeobj__2154;
static PyObject *__pyx_codeobj__2156;
static PyObject *__pyx_codeobj__2158;
static PyObject *__pyx_codeobj__2160;
static PyObject *__pyx_codeobj__2163;
static PyObject *__pyx_codeobj__2165;
static PyObject *__pyx_codeobj__2168;
static PyObject *__pyx_codeobj__2170;
static PyObject *__pyx_codeobj__2172;
static PyObject *__pyx_codeobj__2174;
static PyObject *__pyx_codeobj__2176;
static PyObject *__pyx_codeobj__2178;
static PyObject *__pyx_codeobj__2181;
static PyObject *__pyx_codeobj__2184;
static PyObject *__pyx_codeobj__2187;
static PyObject *__pyx_codeobj__2189;
static PyObject *__pyx_codeobj__2191;
static PyObject *__pyx_codeobj__2193;
static PyObject *__pyx_codeobj__2195;
static PyObject *__pyx_codeobj__2197;
static PyObject *__pyx_codeobj__2200;
static PyObject *__pyx_codeobj__2203;
static PyObject *__pyx_codeobj__2206;
static PyObject *__pyx_codeobj__2208;
static PyObject *__pyx_codeobj__2210;
static PyObject *__pyx_codeobj__2212;
static PyObject *__pyx_codeobj__2214;
static PyObject *__pyx_codeobj__2216;
static PyObject *__pyx_codeobj__2219;
static PyObject *__pyx_codeobj__2222;
static PyObject *__pyx_codeobj__2224;
static PyObject *__pyx_codeobj__2226;
static PyObject *__pyx_codeobj__2228;
static PyObject *__pyx_codeobj__2230;
static PyObject *__pyx_codeobj__2232;
static PyObject *__pyx_codeobj__2235;
static PyObject *__pyx_codeobj__2237;
static PyObject *__pyx_codeobj__2239;
static PyObject *__pyx_codeobj__2241;
static PyObject *__pyx_codeobj__2244;
static PyObject *__pyx_codeobj__2246;
static PyObject *__pyx_codeobj__2248;
static PyObject *__pyx_codeobj__2250;
static PyObject *__pyx_codeobj__2252;
static PyObject *__pyx_codeobj__2254;
static PyObject *__pyx_codeobj__2256;
static PyObject *__pyx_codeobj__2258;
static PyObject *__pyx_codeobj__2260;
static PyObject *__pyx_codeobj__2262;
static PyObject *__pyx_codeobj__2265;
static PyObject *__pyx_codeobj__2268;
static PyObject *__pyx_codeobj__2270;
static PyObject *__pyx_codeobj__2272;
static PyObject *__pyx_codeobj__2274;
static PyObject *__pyx_codeobj__2276;
static PyObject *__pyx_codeobj__2278;
static PyObject *__pyx_codeobj__2281;
static PyObject *__pyx_codeobj__2283;
static PyObject *__pyx_codeobj__2285;
static PyObject *__pyx_codeobj__2287;
static PyObject *__pyx_codeobj__2289;
static PyObject *__pyx_codeobj__2291;
static PyObject *__pyx_codeobj__2293;
static PyObject *__pyx_codeobj__2295;
static PyObject *__pyx_codeobj__2297;
static PyObject *__pyx_codeobj__2300;
static PyObject *__pyx_codeobj__2302;
static PyObject *__pyx_codeobj__2304;
static PyObject *__pyx_codeobj__2306;
static PyObject *__pyx_codeobj__2308;
static PyObject *__pyx_codeobj__2311;
static PyObject *__pyx_codeobj__2313;
static PyObject *__pyx_codeobj__2315;
static PyObject *__pyx_codeobj__2317;
static PyObject *__pyx_codeobj__2319;
static PyObject *__pyx_codeobj__2322;
static PyObject *__pyx_codeobj__2324;
static PyObject *__pyx_codeobj__2326;
static PyObject *__pyx_codeobj__2328;
static PyObject *__pyx_codeobj__2330;
static PyObject *__pyx_codeobj__2332;
static PyObject *__pyx_codeobj__2334;
static PyObject *__pyx_codeobj__2336;
static PyObject *__pyx_codeobj__2338;
static PyObject *__pyx_codeobj__2340;
static PyObject *__pyx_codeobj__2343;
static PyObject *__pyx_codeobj__2345;
static PyObject *__pyx_codeobj__2347;
static PyObject *__pyx_codeobj__2349;
static PyObject *__pyx_codeobj__2351;
static PyObject *__pyx_codeobj__2354;
static PyObject *__pyx_codeobj__2356;
static PyObject *__pyx_codeobj__2358;
static PyObject *__pyx_codeobj__2360;
static PyObject *__pyx_codeobj__2362;
static PyObject *__pyx_codeobj__2364;
static PyObject *__pyx_codeobj__2366;
static PyObject *__pyx_codeobj__2368;
static PyObject *__pyx_codeobj__2370;
static PyObject *__pyx_codeobj__2372;
static PyObject *__pyx_codeobj__2375;
static PyObject *__pyx_codeobj__2377;
static PyObject *__pyx_codeobj__2379;
static PyObject *__pyx_codeobj__2381;
static PyObject *__pyx_codeobj__2384;
static PyObject *__pyx_codeobj__2387;
static PyObject *__pyx_codeobj__2389;
static PyObject *__pyx_codeobj__2391;
static PyObject *__pyx_codeobj__2393;
static PyObject *__pyx_codeobj__2395;
static PyObject *__pyx_codeobj__2397;
static PyObject *__pyx_codeobj__2399;
static PyObject *__pyx_codeobj__2401;
static PyObject *__pyx_codeobj__2403;
static PyObject *__pyx_codeobj__2405;
static PyObject *__pyx_codeobj__2407;
static PyObject *__pyx_codeobj__2409;
static PyObject *__pyx_codeobj__2411;
static PyObject *__pyx_codeobj__2413;
static PyObject *__pyx_codeobj__2415;
static PyObject *__pyx_codeobj__2417;
static PyObject *__pyx_codeobj__2419;
static PyObject *__pyx_codeobj__2421;
static PyObject *__pyx_codeobj__2423;
static PyObject *__pyx_codeobj__2425;
static PyObject *__pyx_codeobj__2427;
static PyObject *__pyx_codeobj__2429;
static PyObject *__pyx_codeobj__2431;
static PyObject *__pyx_codeobj__2433;
static PyObject *__pyx_codeobj__2435;
static PyObject *__pyx_codeobj__2437;
static PyObject *__pyx_codeobj__2439;
static PyObject *__pyx_codeobj__2441;
static PyObject *__pyx_codeobj__2443;
static PyObject *__pyx_codeobj__2445;
static PyObject *__pyx_codeobj__2447;
static PyObject *__pyx_codeobj__2449;
static PyObject *__pyx_codeobj__2451;
static PyObject *__pyx_codeobj__2453;
static PyObject *__pyx_codeobj__2455;
static PyObject *__pyx_codeobj__2457;
static PyObject *__pyx_codeobj__2459;
static PyObject *__pyx_codeobj__2461;
static PyObject *__pyx_codeobj__2463;
static PyObject *__pyx_codeobj__2465;
static PyObject *__pyx_codeobj__2467;
static PyObject *__pyx_codeobj__2469;
static PyObject *__pyx_codeobj__2472;
static PyObject *__pyx_codeobj__2474;
static PyObject *__pyx_codeobj__2476;
static PyObject *__pyx_codeobj__2478;
static PyObject *__pyx_codeobj__2480;
static PyObject *__pyx_codeobj__2482;
static PyObject *__pyx_codeobj__2484;
static PyObject *__pyx_codeobj__2486;
static PyObject *__pyx_codeobj__2488;
static PyObject *__pyx_codeobj__2490;
static PyObject *__pyx_codeobj__2492;
static PyObject *__pyx_codeobj__2494;
static PyObject *__pyx_codeobj__2496;
static PyObject *__pyx_codeobj__2498;
static PyObject *__pyx_codeobj__2500;
static PyObject *__pyx_codeobj__2502;
static PyObject *__pyx_codeobj__2504;
static PyObject *__pyx_codeobj__2506;
static PyObject *__pyx_codeobj__2508;
static PyObject *__pyx_codeobj__2510;
static PyObject *__pyx_codeobj__2513;
static PyObject *__pyx_codeobj__2516;
static PyObject *__pyx_codeobj__2518;
static PyObject *__pyx_codeobj__2520;
static PyObject *__pyx_codeobj__2522;
static PyObject *__pyx_codeobj__2524;
static PyObject *__pyx_codeobj__2526;
static PyObject *__pyx_codeobj__2529;
static PyObject *__pyx_codeobj__2531;
static PyObject *__pyx_codeobj__2533;
static PyObject *__pyx_codeobj__2535;
static PyObject *__pyx_codeobj__2537;
static PyObject *__pyx_codeobj__2539;
static PyObject *__pyx_codeobj__2541;
static PyObject *__pyx_codeobj__2543;
static PyObject *__pyx_codeobj__2546;
static PyObject *__pyx_codeobj__2549;
static PyObject *__pyx_codeobj__2551;
static PyObject *__pyx_codeobj__2553;
static PyObject *__pyx_codeobj__2555;
static PyObject *__pyx_codeobj__2557;
static PyObject *__pyx_codeobj__2559;
static PyObject *__pyx_codeobj__2562;
static PyObject *__pyx_codeobj__2564;
static PyObject *__pyx_codeobj__2566;
static PyObject *__pyx_codeobj__2568;
static PyObject *__pyx_codeobj__2570;
static PyObject *__pyx_codeobj__2572;
static PyObject *__pyx_codeobj__2574;
static PyObject *__pyx_codeobj__2577;
static PyObject *__pyx_codeobj__2580;
static PyObject *__pyx_codeobj__2583;
static PyObject *__pyx_codeobj__2585;
static PyObject *__pyx_codeobj__2587;
static PyObject *__pyx_codeobj__2589;
static PyObject *__pyx_codeobj__2591;
static PyObject *__pyx_codeobj__2593;
static PyObject *__pyx_codeobj__2595;
static PyObject *__pyx_codeobj__2598;
static PyObject *__pyx_codeobj__2601;
static PyObject *__pyx_codeobj__2603;
static PyObject *__pyx_codeobj__2605;
static PyObject *__pyx_codeobj__2607;
static PyObject *__pyx_codeobj__2609;
static PyObject *__pyx_codeobj__2611;
static PyObject *__pyx_codeobj__2614;
static PyObject *__pyx_codeobj__2616;
static PyObject *__pyx_codeobj__2618;
static PyObject *__pyx_codeobj__2620;
static PyObject *__pyx_codeobj__2622;
static PyObject *__pyx_codeobj__2624;
static PyObject *__pyx_codeobj__2626;
static PyObject *__pyx_codeobj__2628;
static PyObject *__pyx_codeobj__2631;
static PyObject *__pyx_codeobj__2633;
static PyObject *__pyx_codeobj__2635;
static PyObject *__pyx_codeobj__2637;
static PyObject *__pyx_codeobj__2639;
static PyObject *__pyx_codeobj__2641;
static PyObject *__pyx_codeobj__2643;
static PyObject *__pyx_codeobj__2645;
static PyObject *__pyx_codeobj__2647;
static PyObject *__pyx_codeobj__2649;
static PyObject *__pyx_codeobj__2651;
static PyObject *__pyx_codeobj__2653;
static PyObject *__pyx_codeobj__2655;
static PyObject *__pyx_codeobj__2657;
static PyObject *__pyx_codeobj__2659;
static PyObject *__pyx_codeobj__2662;
static PyObject *__pyx_codeobj__2665;
static PyObject *__pyx_codeobj__2667;
static PyObject *__pyx_codeobj__2669;
static PyObject *__pyx_codeobj__2671;
static PyObject *__pyx_codeobj__2673;
static PyObject *__pyx_codeobj__2675;
static PyObject *__pyx_codeobj__2678;
static PyObject *__pyx_codeobj__2681;
static PyObject *__pyx_codeobj__2683;
static PyObject *__pyx_codeobj__2685;
static PyObject *__pyx_codeobj__2687;
static PyObject *__pyx_codeobj__2689;
static PyObject *__pyx_codeobj__2691;
static PyObject *__pyx_codeobj__2694;
static PyObject *__pyx_codeobj__2697;
static PyObject *__pyx_codeobj__2699;
static PyObject *__pyx_codeobj__2701;
static PyObject *__pyx_codeobj__2703;
static PyObject *__pyx_codeobj__2705;
static PyObject *__pyx_codeobj__2707;
static PyObject *__pyx_codeobj__2710;
static PyObject *__pyx_codeobj__2712;
static PyObject *__pyx_codeobj__2714;
static PyObject *__pyx_codeobj__2716;
static PyObject *__pyx_codeobj__2718;
static PyObject *__pyx_codeobj__2720;
static PyObject *__pyx_codeobj__2722;
static PyObject *__pyx_codeobj__2725;
static PyObject *__pyx_codeobj__2727;
static PyObject *__pyx_codeobj__2729;
static PyObject *__pyx_codeobj__2731;
static PyObject *__pyx_codeobj__2733;
static PyObject *__pyx_codeobj__2735;
static PyObject *__pyx_codeobj__2737;
static PyObject *__pyx_codeobj__2739;
static PyObject *__pyx_codeobj__2742;
static PyObject *__pyx_codeobj__2745;
static PyObject *__pyx_codeobj__2747;
static PyObject *__pyx_codeobj__2749;
static PyObject *__pyx_codeobj__2751;
static PyObject *__pyx_codeobj__2753;
static PyObject *__pyx_codeobj__2755;
static PyObject *__pyx_codeobj__2758;
static PyObject *__pyx_codeobj__2760;
static PyObject *__pyx_codeobj__2762;
static PyObject *__pyx_codeobj__2764;
static PyObject *__pyx_codeobj__2766;
static PyObject *__pyx_codeobj__2768;
static PyObject *__pyx_codeobj__2770;
static PyObject *__pyx_codeobj__2772;

/* "pymola/generated/ModelicaParser.pyx":8
 * import sys
 * 
 * def serializedATN():             # <<<<<<<<<<<<<<
 *     with StringIO() as buf:
 *         buf.write("\3\u608b\ua72a\u8133\ub9ed\u417c\u3be7\u7786\u5964\3^")
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_1serializedATN(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_1serializedATN = {"serializedATN", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_1serializedATN, METH_NOARGS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_1serializedATN(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("serializedATN (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_serializedATN(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_serializedATN(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_v_buf = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_t_11;
  __Pyx_RefNannySetupContext("serializedATN", 0);

  /* "pymola/generated/ModelicaParser.pyx":9
 * 
 * def serializedATN():
 *     with StringIO() as buf:             # <<<<<<<<<<<<<<
 *         buf.write("\3\u608b\ua72a\u8133\ub9ed\u417c\u3be7\u7786\u5964\3^")
 *         buf.write("\u040c\4\2\t\2\4\3\t\3\4\4\t\4\4\5\t\5\4\6\t\6\4\7\t\7")
 */
  /*with:*/ {
    __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_StringIO); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_3) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 9, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 9, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __Pyx_PyObject_LookupSpecial(__pyx_t_1, __pyx_n_s_exit); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PyObject_LookupSpecial(__pyx_t_1, __pyx_n_s_enter); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 9, __pyx_L3_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    if (__pyx_t_5) {
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 9, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 9, __pyx_L3_error)
    }
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __pyx_t_2;
    __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    /*try:*/ {
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);
        /*try:*/ {
          __pyx_v_buf = __pyx_t_3;
          __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":10
 * def serializedATN():
 *     with StringIO() as buf:
 *         buf.write("\3\u608b\ua72a\u8133\ub9ed\u417c\u3be7\u7786\u5964\3^")             # <<<<<<<<<<<<<<
 *         buf.write("\u040c\4\2\t\2\4\3\t\3\4\4\t\4\4\5\t\5\4\6\t\6\4\7\t\7")
 *         buf.write("\4\b\t\b\4\t\t\t\4\n\t\n\4\13\t\13\4\f\t\f\4\r\t\r\4\16")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 10, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 10, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":11
 *     with StringIO() as buf:
 *         buf.write("\3\u608b\ua72a\u8133\ub9ed\u417c\u3be7\u7786\u5964\3^")
 *         buf.write("\u040c\4\2\t\2\4\3\t\3\4\4\t\4\4\5\t\5\4\6\t\6\4\7\t\7")             # <<<<<<<<<<<<<<
 *         buf.write("\4\b\t\b\4\t\t\t\4\n\t\n\4\13\t\13\4\f\t\f\4\r\t\r\4\16")
 *         buf.write("\t\16\4\17\t\17\4\20\t\20\4\21\t\21\4\22\t\22\4\23\t\23")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 11, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 11, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":12
 *         buf.write("\3\u608b\ua72a\u8133\ub9ed\u417c\u3be7\u7786\u5964\3^")
 *         buf.write("\u040c\4\2\t\2\4\3\t\3\4\4\t\4\4\5\t\5\4\6\t\6\4\7\t\7")
 *         buf.write("\4\b\t\b\4\t\t\t\4\n\t\n\4\13\t\13\4\f\t\f\4\r\t\r\4\16")             # <<<<<<<<<<<<<<
 *         buf.write("\t\16\4\17\t\17\4\20\t\20\4\21\t\21\4\22\t\22\4\23\t\23")
 *         buf.write("\4\24\t\24\4\25\t\25\4\26\t\26\4\27\t\27\4\30\t\30\4\31")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 12, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":13
 *         buf.write("\u040c\4\2\t\2\4\3\t\3\4\4\t\4\4\5\t\5\4\6\t\6\4\7\t\7")
 *         buf.write("\4\b\t\b\4\t\t\t\4\n\t\n\4\13\t\13\4\f\t\f\4\r\t\r\4\16")
 *         buf.write("\t\16\4\17\t\17\4\20\t\20\4\21\t\21\4\22\t\22\4\23\t\23")             # <<<<<<<<<<<<<<
 *         buf.write("\4\24\t\24\4\25\t\25\4\26\t\26\4\27\t\27\4\30\t\30\4\31")
 *         buf.write("\t\31\4\32\t\32\4\33\t\33\4\34\t\34\4\35\t\35\4\36\t\36")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 13, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 13, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":14
 *         buf.write("\4\b\t\b\4\t\t\t\4\n\t\n\4\13\t\13\4\f\t\f\4\r\t\r\4\16")
 *         buf.write("\t\16\4\17\t\17\4\20\t\20\4\21\t\21\4\22\t\22\4\23\t\23")
 *         buf.write("\4\24\t\24\4\25\t\25\4\26\t\26\4\27\t\27\4\30\t\30\4\31")             # <<<<<<<<<<<<<<
 *         buf.write("\t\31\4\32\t\32\4\33\t\33\4\34\t\34\4\35\t\35\4\36\t\36")
 *         buf.write("\4\37\t\37\4 \t \4!\t!\4\"\t\"\4#\t#\4$\t$\4%\t%\4&\t")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 14, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 14, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":15
 *         buf.write("\t\16\4\17\t\17\4\20\t\20\4\21\t\21\4\22\t\22\4\23\t\23")
 *         buf.write("\4\24\t\24\4\25\t\25\4\26\t\26\4\27\t\27\4\30\t\30\4\31")
 *         buf.write("\t\31\4\32\t\32\4\33\t\33\4\34\t\34\4\35\t\35\4\36\t\36")             # <<<<<<<<<<<<<<
 *         buf.write("\4\37\t\37\4 \t \4!\t!\4\"\t\"\4#\t#\4$\t$\4%\t%\4&\t")
 *         buf.write("&\4\'\t\'\4(\t(\4)\t)\4*\t*\4+\t+\4,\t,\4-\t-\4.\t.\4")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 15, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 15, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":16
 *         buf.write("\4\24\t\24\4\25\t\25\4\26\t\26\4\27\t\27\4\30\t\30\4\31")
 *         buf.write("\t\31\4\32\t\32\4\33\t\33\4\34\t\34\4\35\t\35\4\36\t\36")
 *         buf.write("\4\37\t\37\4 \t \4!\t!\4\"\t\"\4#\t#\4$\t$\4%\t%\4&\t")             # <<<<<<<<<<<<<<
 *         buf.write("&\4\'\t\'\4(\t(\4)\t)\4*\t*\4+\t+\4,\t,\4-\t-\4.\t.\4")
 *         buf.write("/\t/\4\60\t\60\4\61\t\61\4\62\t\62\4\63\t\63\4\64\t\64")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 16, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 16, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":17
 *         buf.write("\t\31\4\32\t\32\4\33\t\33\4\34\t\34\4\35\t\35\4\36\t\36")
 *         buf.write("\4\37\t\37\4 \t \4!\t!\4\"\t\"\4#\t#\4$\t$\4%\t%\4&\t")
 *         buf.write("&\4\'\t\'\4(\t(\4)\t)\4*\t*\4+\t+\4,\t,\4-\t-\4.\t.\4")             # <<<<<<<<<<<<<<
 *         buf.write("/\t/\4\60\t\60\4\61\t\61\4\62\t\62\4\63\t\63\4\64\t\64")
 *         buf.write("\4\65\t\65\4\66\t\66\4\67\t\67\48\t8\49\t9\4:\t:\4;\t")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 17, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":18
 *         buf.write("\4\37\t\37\4 \t \4!\t!\4\"\t\"\4#\t#\4$\t$\4%\t%\4&\t")
 *         buf.write("&\4\'\t\'\4(\t(\4)\t)\4*\t*\4+\t+\4,\t,\4-\t-\4.\t.\4")
 *         buf.write("/\t/\4\60\t\60\4\61\t\61\4\62\t\62\4\63\t\63\4\64\t\64")             # <<<<<<<<<<<<<<
 *         buf.write("\4\65\t\65\4\66\t\66\4\67\t\67\48\t8\49\t9\4:\t:\4;\t")
 *         buf.write(";\4<\t<\4=\t=\4>\t>\4?\t?\4@\t@\4A\tA\4B\tB\4C\tC\4D\t")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 18, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__17, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 18, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":19
 *         buf.write("&\4\'\t\'\4(\t(\4)\t)\4*\t*\4+\t+\4,\t,\4-\t-\4.\t.\4")
 *         buf.write("/\t/\4\60\t\60\4\61\t\61\4\62\t\62\4\63\t\63\4\64\t\64")
 *         buf.write("\4\65\t\65\4\66\t\66\4\67\t\67\48\t8\49\t9\4:\t:\4;\t")             # <<<<<<<<<<<<<<
 *         buf.write(";\4<\t<\4=\t=\4>\t>\4?\t?\4@\t@\4A\tA\4B\tB\4C\tC\4D\t")
 *         buf.write("D\4E\tE\4F\tF\4G\tG\4H\tH\4I\tI\4J\tJ\3\2\3\2\5\2\u0097")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 19, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__18, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 19, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":20
 *         buf.write("/\t/\4\60\t\60\4\61\t\61\4\62\t\62\4\63\t\63\4\64\t\64")
 *         buf.write("\4\65\t\65\4\66\t\66\4\67\t\67\48\t8\49\t9\4:\t:\4;\t")
 *         buf.write(";\4<\t<\4=\t=\4>\t>\4?\t?\4@\t@\4A\tA\4B\tB\4C\tC\4D\t")             # <<<<<<<<<<<<<<
 *         buf.write("D\4E\tE\4F\tF\4G\tG\4H\tH\4I\tI\4J\tJ\3\2\3\2\5\2\u0097")
 *         buf.write("\n\2\3\2\5\2\u009a\n\2\3\2\7\2\u009d\n\2\f\2\16\2\u00a0")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 20, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__19, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 20, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":21
 *         buf.write("\4\65\t\65\4\66\t\66\4\67\t\67\48\t8\49\t9\4:\t:\4;\t")
 *         buf.write(";\4<\t<\4=\t=\4>\t>\4?\t?\4@\t@\4A\tA\4B\tB\4C\tC\4D\t")
 *         buf.write("D\4E\tE\4F\tF\4G\tG\4H\tH\4I\tI\4J\tJ\3\2\3\2\5\2\u0097")             # <<<<<<<<<<<<<<
 *         buf.write("\n\2\3\2\5\2\u009a\n\2\3\2\7\2\u009d\n\2\f\2\16\2\u00a0")
 *         buf.write("\13\2\3\3\5\3\u00a3\n\3\3\3\3\3\3\3\3\4\5\4\u00a9\n\4")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 21, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__20, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 21, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":22
 *         buf.write(";\4<\t<\4=\t=\4>\t>\4?\t?\4@\t@\4A\tA\4B\tB\4C\tC\4D\t")
 *         buf.write("D\4E\tE\4F\tF\4G\tG\4H\tH\4I\tI\4J\tJ\3\2\3\2\5\2\u0097")
 *         buf.write("\n\2\3\2\5\2\u009a\n\2\3\2\7\2\u009d\n\2\f\2\16\2\u00a0")             # <<<<<<<<<<<<<<
 *         buf.write("\13\2\3\3\5\3\u00a3\n\3\3\3\3\3\3\3\3\4\5\4\u00a9\n\4")
 *         buf.write("\3\4\3\4\3\4\3\5\5\5\u00af\n\5\3\5\3\5\3\6\3\6\3\6\5\6")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 22, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__22, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 22, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":23
 *         buf.write("D\4E\tE\4F\tF\4G\tG\4H\tH\4I\tI\4J\tJ\3\2\3\2\5\2\u0097")
 *         buf.write("\n\2\3\2\5\2\u009a\n\2\3\2\7\2\u009d\n\2\f\2\16\2\u00a0")
 *         buf.write("\13\2\3\3\5\3\u00a3\n\3\3\3\3\3\3\3\3\4\5\4\u00a9\n\4")             # <<<<<<<<<<<<<<
 *         buf.write("\3\4\3\4\3\4\3\5\5\5\u00af\n\5\3\5\3\5\3\6\3\6\3\6\5\6")
 *         buf.write("\u00b6\n\6\3\6\3\6\3\6\5\6\u00bb\n\6\3\6\3\6\3\6\3\6\5")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 23, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__24, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 23, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":24
 *         buf.write("\n\2\3\2\5\2\u009a\n\2\3\2\7\2\u009d\n\2\f\2\16\2\u00a0")
 *         buf.write("\13\2\3\3\5\3\u00a3\n\3\3\3\3\3\3\3\3\4\5\4\u00a9\n\4")
 *         buf.write("\3\4\3\4\3\4\3\5\5\5\u00af\n\5\3\5\3\5\3\6\3\6\3\6\5\6")             # <<<<<<<<<<<<<<
 *         buf.write("\u00b6\n\6\3\6\3\6\3\6\5\6\u00bb\n\6\3\6\3\6\3\6\3\6\5")
 *         buf.write("\6\u00c1\n\6\3\6\5\6\u00c4\n\6\3\6\3\6\5\6\u00c8\n\6\3")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 24, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__26, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 24, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":25
 *         buf.write("\13\2\3\3\5\3\u00a3\n\3\3\3\3\3\3\3\3\4\5\4\u00a9\n\4")
 *         buf.write("\3\4\3\4\3\4\3\5\5\5\u00af\n\5\3\5\3\5\3\6\3\6\3\6\5\6")
 *         buf.write("\u00b6\n\6\3\6\3\6\3\6\5\6\u00bb\n\6\3\6\3\6\3\6\3\6\5")             # <<<<<<<<<<<<<<
 *         buf.write("\6\u00c1\n\6\3\6\5\6\u00c4\n\6\3\6\3\6\5\6\u00c8\n\6\3")
 *         buf.write("\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\5\7\u00d5\n")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 25, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__28, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 25, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":26
 *         buf.write("\3\4\3\4\3\4\3\5\5\5\u00af\n\5\3\5\3\5\3\6\3\6\3\6\5\6")
 *         buf.write("\u00b6\n\6\3\6\3\6\3\6\5\6\u00bb\n\6\3\6\3\6\3\6\3\6\5")
 *         buf.write("\6\u00c1\n\6\3\6\5\6\u00c4\n\6\3\6\3\6\5\6\u00c8\n\6\3")             # <<<<<<<<<<<<<<
 *         buf.write("\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\5\7\u00d5\n")
 *         buf.write("\7\3\7\5\7\u00d8\n\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\5\7\u00e1")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 26, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__30, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 26, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":27
 *         buf.write("\u00b6\n\6\3\6\3\6\3\6\5\6\u00bb\n\6\3\6\3\6\3\6\3\6\5")
 *         buf.write("\6\u00c1\n\6\3\6\5\6\u00c4\n\6\3\6\3\6\5\6\u00c8\n\6\3")
 *         buf.write("\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\5\7\u00d5\n")             # <<<<<<<<<<<<<<
 *         buf.write("\7\3\7\5\7\u00d8\n\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\5\7\u00e1")
 *         buf.write("\n\7\3\7\5\7\u00e4\n\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 27, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__32, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 27, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":28
 *         buf.write("\6\u00c1\n\6\3\6\5\6\u00c4\n\6\3\6\3\6\5\6\u00c8\n\6\3")
 *         buf.write("\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\5\7\u00d5\n")
 *         buf.write("\7\3\7\5\7\u00d8\n\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\5\7\u00e1")             # <<<<<<<<<<<<<<
 *         buf.write("\n\7\3\7\5\7\u00e4\n\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7")
 *         buf.write("\3\7\3\7\3\7\7\7\u00f1\n\7\f\7\16\7\u00f4\13\7\3\7\3\7")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 28, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__34, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 28, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":29
 *         buf.write("\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\5\7\u00d5\n")
 *         buf.write("\7\3\7\5\7\u00d8\n\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\5\7\u00e1")
 *         buf.write("\n\7\3\7\5\7\u00e4\n\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7")             # <<<<<<<<<<<<<<
 *         buf.write("\3\7\3\7\3\7\7\7\u00f1\n\7\f\7\16\7\u00f4\13\7\3\7\3\7")
 *         buf.write("\3\7\3\7\3\7\3\7\5\7\u00fc\n\7\3\7\3\7\3\7\3\7\3\7\5\7")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 29, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__36, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 29, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":30
 *         buf.write("\7\3\7\5\7\u00d8\n\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\5\7\u00e1")
 *         buf.write("\n\7\3\7\5\7\u00e4\n\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7")
 *         buf.write("\3\7\3\7\3\7\7\7\u00f1\n\7\f\7\16\7\u00f4\13\7\3\7\3\7")             # <<<<<<<<<<<<<<
 *         buf.write("\3\7\3\7\3\7\3\7\5\7\u00fc\n\7\3\7\3\7\3\7\3\7\3\7\5\7")
 *         buf.write("\u0103\n\7\3\b\3\b\3\t\3\t\3\t\7\t\u010a\n\t\f\t\16\t")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 30, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__38, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 30, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":31
 *         buf.write("\n\7\3\7\5\7\u00e4\n\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7")
 *         buf.write("\3\7\3\7\3\7\7\7\u00f1\n\7\f\7\16\7\u00f4\13\7\3\7\3\7")
 *         buf.write("\3\7\3\7\3\7\3\7\5\7\u00fc\n\7\3\7\3\7\3\7\3\7\3\7\5\7")             # <<<<<<<<<<<<<<
 *         buf.write("\u0103\n\7\3\b\3\b\3\t\3\t\3\t\7\t\u010a\n\t\f\t\16\t")
 *         buf.write("\u010d\13\t\3\n\3\n\3\n\3\13\3\13\3\13\3\13\3\13\3\13")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 31, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__40, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 31, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":32
 *         buf.write("\3\7\3\7\3\7\7\7\u00f1\n\7\f\7\16\7\u00f4\13\7\3\7\3\7")
 *         buf.write("\3\7\3\7\3\7\3\7\5\7\u00fc\n\7\3\7\3\7\3\7\3\7\3\7\5\7")
 *         buf.write("\u0103\n\7\3\b\3\b\3\t\3\t\3\t\7\t\u010a\n\t\f\t\16\t")             # <<<<<<<<<<<<<<
 *         buf.write("\u010d\13\t\3\n\3\n\3\n\3\13\3\13\3\13\3\13\3\13\3\13")
 *         buf.write("\3\13\7\13\u0119\n\13\f\13\16\13\u011c\13\13\3\13\3\13")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 32, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__42, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 32, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":33
 *         buf.write("\3\7\3\7\3\7\3\7\5\7\u00fc\n\7\3\7\3\7\3\7\3\7\3\7\5\7")
 *         buf.write("\u0103\n\7\3\b\3\b\3\t\3\t\3\t\7\t\u010a\n\t\f\t\16\t")
 *         buf.write("\u010d\13\t\3\n\3\n\3\n\3\13\3\13\3\13\3\13\3\13\3\13")             # <<<<<<<<<<<<<<
 *         buf.write("\3\13\7\13\u0119\n\13\f\13\16\13\u011c\13\13\3\13\3\13")
 *         buf.write("\5\13\u0120\n\13\3\13\5\13\u0123\n\13\3\13\5\13\u0126")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 33, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__44, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 33, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":34
 *         buf.write("\u0103\n\7\3\b\3\b\3\t\3\t\3\t\7\t\u010a\n\t\f\t\16\t")
 *         buf.write("\u010d\13\t\3\n\3\n\3\n\3\13\3\13\3\13\3\13\3\13\3\13")
 *         buf.write("\3\13\7\13\u0119\n\13\f\13\16\13\u011c\13\13\3\13\3\13")             # <<<<<<<<<<<<<<
 *         buf.write("\5\13\u0120\n\13\3\13\5\13\u0123\n\13\3\13\5\13\u0126")
 *         buf.write("\n\13\3\13\5\13\u0129\n\13\3\13\3\13\3\13\5\13\u012e\n")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 34, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__46, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 34, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":35
 *         buf.write("\u010d\13\t\3\n\3\n\3\n\3\13\3\13\3\13\3\13\3\13\3\13")
 *         buf.write("\3\13\7\13\u0119\n\13\f\13\16\13\u011c\13\13\3\13\3\13")
 *         buf.write("\5\13\u0120\n\13\3\13\5\13\u0123\n\13\3\13\5\13\u0126")             # <<<<<<<<<<<<<<
 *         buf.write("\n\13\3\13\5\13\u0129\n\13\3\13\3\13\3\13\5\13\u012e\n")
 *         buf.write("\13\3\f\3\f\3\r\3\r\3\r\5\r\u0135\n\r\3\r\3\r\3\r\5\r")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 35, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__48, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 35, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":36
 *         buf.write("\3\13\7\13\u0119\n\13\f\13\16\13\u011c\13\13\3\13\3\13")
 *         buf.write("\5\13\u0120\n\13\3\13\5\13\u0123\n\13\3\13\5\13\u0126")
 *         buf.write("\n\13\3\13\5\13\u0129\n\13\3\13\3\13\3\13\5\13\u012e\n")             # <<<<<<<<<<<<<<
 *         buf.write("\13\3\f\3\f\3\r\3\r\3\r\5\r\u0135\n\r\3\r\3\r\3\r\5\r")
 *         buf.write("\u013a\n\r\3\r\3\r\3\16\3\16\3\16\7\16\u0141\n\16\f\16")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 36, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__50, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 36, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":37
 *         buf.write("\5\13\u0120\n\13\3\13\5\13\u0123\n\13\3\13\5\13\u0126")
 *         buf.write("\n\13\3\13\5\13\u0129\n\13\3\13\3\13\3\13\5\13\u012e\n")
 *         buf.write("\13\3\f\3\f\3\r\3\r\3\r\5\r\u0135\n\r\3\r\3\r\3\r\5\r")             # <<<<<<<<<<<<<<
 *         buf.write("\u013a\n\r\3\r\3\r\3\16\3\16\3\16\7\16\u0141\n\16\f\16")
 *         buf.write("\16\16\u0144\13\16\3\17\3\17\3\17\3\17\5\17\u014a\n\17")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 37, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__52, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 37, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":38
 *         buf.write("\n\13\3\13\5\13\u0129\n\13\3\13\3\13\3\13\5\13\u012e\n")
 *         buf.write("\13\3\f\3\f\3\r\3\r\3\r\5\r\u0135\n\r\3\r\3\r\3\r\5\r")
 *         buf.write("\u013a\n\r\3\r\3\r\3\16\3\16\3\16\7\16\u0141\n\16\f\16")             # <<<<<<<<<<<<<<
 *         buf.write("\16\16\u0144\13\16\3\17\3\17\3\17\3\17\5\17\u014a\n\17")
 *         buf.write("\3\20\5\20\u014d\n\20\3\20\5\20\u0150\n\20\3\20\5\20\u0153")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 38, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__54, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 38, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":39
 *         buf.write("\13\3\f\3\f\3\r\3\r\3\r\5\r\u0135\n\r\3\r\3\r\3\r\5\r")
 *         buf.write("\u013a\n\r\3\r\3\r\3\16\3\16\3\16\7\16\u0141\n\16\f\16")
 *         buf.write("\16\16\u0144\13\16\3\17\3\17\3\17\3\17\5\17\u014a\n\17")             # <<<<<<<<<<<<<<
 *         buf.write("\3\20\5\20\u014d\n\20\3\20\5\20\u0150\n\20\3\20\5\20\u0153")
 *         buf.write("\n\20\3\20\5\20\u0156\n\20\3\20\3\20\5\20\u015a\n\20\3")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 39, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__56, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 39, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":40
 *         buf.write("\u013a\n\r\3\r\3\r\3\16\3\16\3\16\7\16\u0141\n\16\f\16")
 *         buf.write("\16\16\u0144\13\16\3\17\3\17\3\17\3\17\5\17\u014a\n\17")
 *         buf.write("\3\20\5\20\u014d\n\20\3\20\5\20\u0150\n\20\3\20\5\20\u0153")             # <<<<<<<<<<<<<<
 *         buf.write("\n\20\3\20\5\20\u0156\n\20\3\20\3\20\5\20\u015a\n\20\3")
 *         buf.write("\21\5\21\u015d\n\21\3\21\5\21\u0160\n\21\3\21\5\21\u0163")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 40, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__58, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 40, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":41
 *         buf.write("\16\16\u0144\13\16\3\17\3\17\3\17\3\17\5\17\u014a\n\17")
 *         buf.write("\3\20\5\20\u014d\n\20\3\20\5\20\u0150\n\20\3\20\5\20\u0153")
 *         buf.write("\n\20\3\20\5\20\u0156\n\20\3\20\3\20\5\20\u015a\n\20\3")             # <<<<<<<<<<<<<<
 *         buf.write("\21\5\21\u015d\n\21\3\21\5\21\u0160\n\21\3\21\5\21\u0163")
 *         buf.write("\n\21\3\21\5\21\u0166\n\21\3\21\3\21\3\21\5\21\u016b\n")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 41, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__60, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 41, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":42
 *         buf.write("\3\20\5\20\u014d\n\20\3\20\5\20\u0150\n\20\3\20\5\20\u0153")
 *         buf.write("\n\20\3\20\5\20\u0156\n\20\3\20\3\20\5\20\u015a\n\20\3")
 *         buf.write("\21\5\21\u015d\n\21\3\21\5\21\u0160\n\21\3\21\5\21\u0163")             # <<<<<<<<<<<<<<
 *         buf.write("\n\21\3\21\5\21\u0166\n\21\3\21\3\21\3\21\5\21\u016b\n")
 *         buf.write("\21\3\21\3\21\3\21\5\21\u0170\n\21\3\22\3\22\3\22\3\22")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 42, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__62, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 42, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":43
 *         buf.write("\n\20\3\20\5\20\u0156\n\20\3\20\3\20\5\20\u015a\n\20\3")
 *         buf.write("\21\5\21\u015d\n\21\3\21\5\21\u0160\n\21\3\21\5\21\u0163")
 *         buf.write("\n\21\3\21\5\21\u0166\n\21\3\21\3\21\3\21\5\21\u016b\n")             # <<<<<<<<<<<<<<
 *         buf.write("\21\3\21\3\21\3\21\5\21\u0170\n\21\3\22\3\22\3\22\3\22")
 *         buf.write("\3\22\3\22\3\22\3\22\3\22\3\22\3\22\5\22\u017d\n\22\5")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 43, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__64, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 43, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":44
 *         buf.write("\21\5\21\u015d\n\21\3\21\5\21\u0160\n\21\3\21\5\21\u0163")
 *         buf.write("\n\21\3\21\5\21\u0166\n\21\3\21\3\21\3\21\5\21\u016b\n")
 *         buf.write("\21\3\21\3\21\3\21\5\21\u0170\n\21\3\22\3\22\3\22\3\22")             # <<<<<<<<<<<<<<
 *         buf.write("\3\22\3\22\3\22\3\22\3\22\3\22\3\22\5\22\u017d\n\22\5")
 *         buf.write("\22\u017f\n\22\5\22\u0181\n\22\3\22\3\22\3\23\3\23\3\23")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 44, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__66, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 44, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":45
 *         buf.write("\n\21\3\21\5\21\u0166\n\21\3\21\3\21\3\21\5\21\u016b\n")
 *         buf.write("\21\3\21\3\21\3\21\5\21\u0170\n\21\3\22\3\22\3\22\3\22")
 *         buf.write("\3\22\3\22\3\22\3\22\3\22\3\22\3\22\5\22\u017d\n\22\5")             # <<<<<<<<<<<<<<
 *         buf.write("\22\u017f\n\22\5\22\u0181\n\22\3\22\3\22\3\23\3\23\3\23")
 *         buf.write("\7\23\u0188\n\23\f\23\16\23\u018b\13\23\3\24\3\24\3\24")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 45, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__68, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 45, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":46
 *         buf.write("\21\3\21\3\21\3\21\5\21\u0170\n\21\3\22\3\22\3\22\3\22")
 *         buf.write("\3\22\3\22\3\22\3\22\3\22\3\22\3\22\5\22\u017d\n\22\5")
 *         buf.write("\22\u017f\n\22\5\22\u0181\n\22\3\22\3\22\3\23\3\23\3\23")             # <<<<<<<<<<<<<<
 *         buf.write("\7\23\u0188\n\23\f\23\16\23\u018b\13\23\3\24\3\24\3\24")
 *         buf.write("\5\24\u0190\n\24\3\24\5\24\u0193\n\24\3\25\3\25\3\25\5")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 46, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__70, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 46, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":47
 *         buf.write("\3\22\3\22\3\22\3\22\3\22\3\22\3\22\5\22\u017d\n\22\5")
 *         buf.write("\22\u017f\n\22\5\22\u0181\n\22\3\22\3\22\3\23\3\23\3\23")
 *         buf.write("\7\23\u0188\n\23\f\23\16\23\u018b\13\23\3\24\3\24\3\24")             # <<<<<<<<<<<<<<
 *         buf.write("\5\24\u0190\n\24\3\24\5\24\u0193\n\24\3\25\3\25\3\25\5")
 *         buf.write("\25\u0198\n\25\3\26\3\26\3\26\5\26\u019d\n\26\3\26\3\26")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 47, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__72, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 47, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":48
 *         buf.write("\22\u017f\n\22\5\22\u0181\n\22\3\22\3\22\3\23\3\23\3\23")
 *         buf.write("\7\23\u0188\n\23\f\23\16\23\u018b\13\23\3\24\3\24\3\24")
 *         buf.write("\5\24\u0190\n\24\3\24\5\24\u0193\n\24\3\25\3\25\3\25\5")             # <<<<<<<<<<<<<<
 *         buf.write("\25\u0198\n\25\3\26\3\26\3\26\5\26\u019d\n\26\3\26\3\26")
 *         buf.write("\3\27\5\27\u01a2\n\27\3\27\5\27\u01a5\n\27\3\27\5\27\u01a8")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 48, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__74, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 48, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":49
 *         buf.write("\7\23\u0188\n\23\f\23\16\23\u018b\13\23\3\24\3\24\3\24")
 *         buf.write("\5\24\u0190\n\24\3\24\5\24\u0193\n\24\3\25\3\25\3\25\5")
 *         buf.write("\25\u0198\n\25\3\26\3\26\3\26\5\26\u019d\n\26\3\26\3\26")             # <<<<<<<<<<<<<<
 *         buf.write("\3\27\5\27\u01a2\n\27\3\27\5\27\u01a5\n\27\3\27\5\27\u01a8")
 *         buf.write("\n\27\3\30\3\30\3\31\3\31\3\31\7\31\u01af\n\31\f\31\16")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 49, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__76, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 49, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":50
 *         buf.write("\5\24\u0190\n\24\3\24\5\24\u0193\n\24\3\25\3\25\3\25\5")
 *         buf.write("\25\u0198\n\25\3\26\3\26\3\26\5\26\u019d\n\26\3\26\3\26")
 *         buf.write("\3\27\5\27\u01a2\n\27\3\27\5\27\u01a5\n\27\3\27\5\27\u01a8")             # <<<<<<<<<<<<<<
 *         buf.write("\n\27\3\30\3\30\3\31\3\31\3\31\7\31\u01af\n\31\f\31\16")
 *         buf.write("\31\u01b2\13\31\3\32\3\32\5\32\u01b6\n\32\3\32\3\32\3")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 50, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__78, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 50, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":51
 *         buf.write("\25\u0198\n\25\3\26\3\26\3\26\5\26\u019d\n\26\3\26\3\26")
 *         buf.write("\3\27\5\27\u01a2\n\27\3\27\5\27\u01a5\n\27\3\27\5\27\u01a8")
 *         buf.write("\n\27\3\30\3\30\3\31\3\31\3\31\7\31\u01af\n\31\f\31\16")             # <<<<<<<<<<<<<<
 *         buf.write("\31\u01b2\13\31\3\32\3\32\5\32\u01b6\n\32\3\32\3\32\3")
 *         buf.write("\33\3\33\3\33\3\34\3\34\5\34\u01bf\n\34\3\34\5\34\u01c2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 51, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__80, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 51, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":52
 *         buf.write("\3\27\5\27\u01a2\n\27\3\27\5\27\u01a5\n\27\3\27\5\27\u01a8")
 *         buf.write("\n\27\3\30\3\30\3\31\3\31\3\31\7\31\u01af\n\31\f\31\16")
 *         buf.write("\31\u01b2\13\31\3\32\3\32\5\32\u01b6\n\32\3\32\3\32\3")             # <<<<<<<<<<<<<<
 *         buf.write("\33\3\33\3\33\3\34\3\34\5\34\u01bf\n\34\3\34\5\34\u01c2")
 *         buf.write("\n\34\3\35\3\35\3\35\5\35\u01c7\n\35\3\35\3\35\3\35\3")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 52, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__82, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 52, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":53
 *         buf.write("\n\27\3\30\3\30\3\31\3\31\3\31\7\31\u01af\n\31\f\31\16")
 *         buf.write("\31\u01b2\13\31\3\32\3\32\5\32\u01b6\n\32\3\32\3\32\3")
 *         buf.write("\33\3\33\3\33\3\34\3\34\5\34\u01bf\n\34\3\34\5\34\u01c2")             # <<<<<<<<<<<<<<
 *         buf.write("\n\34\3\35\3\35\3\35\5\35\u01c7\n\35\3\35\3\35\3\35\3")
 *         buf.write("\35\5\35\u01cd\n\35\3\36\3\36\5\36\u01d1\n\36\3\36\3\36")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 53, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__84, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 53, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":54
 *         buf.write("\31\u01b2\13\31\3\32\3\32\5\32\u01b6\n\32\3\32\3\32\3")
 *         buf.write("\33\3\33\3\33\3\34\3\34\5\34\u01bf\n\34\3\34\5\34\u01c2")
 *         buf.write("\n\34\3\35\3\35\3\35\5\35\u01c7\n\35\3\35\3\35\3\35\3")             # <<<<<<<<<<<<<<
 *         buf.write("\35\5\35\u01cd\n\35\3\36\3\36\5\36\u01d1\n\36\3\36\3\36")
 *         buf.write("\3\37\3\37\3\37\7\37\u01d8\n\37\f\37\16\37\u01db\13\37")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 54, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__86, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 54, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":55
 *         buf.write("\33\3\33\3\33\3\34\3\34\5\34\u01bf\n\34\3\34\5\34\u01c2")
 *         buf.write("\n\34\3\35\3\35\3\35\5\35\u01c7\n\35\3\35\3\35\3\35\3")
 *         buf.write("\35\5\35\u01cd\n\35\3\36\3\36\5\36\u01d1\n\36\3\36\3\36")             # <<<<<<<<<<<<<<
 *         buf.write("\3\37\3\37\3\37\7\37\u01d8\n\37\f\37\16\37\u01db\13\37")
 *         buf.write("\3 \3 \5 \u01df\n \3!\5!\u01e2\n!\3!\5!\u01e5\n!\3!\3")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 55, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__88, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 55, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":56
 *         buf.write("\n\34\3\35\3\35\3\35\5\35\u01c7\n\35\3\35\3\35\3\35\3")
 *         buf.write("\35\5\35\u01cd\n\35\3\36\3\36\5\36\u01d1\n\36\3\36\3\36")
 *         buf.write("\3\37\3\37\3\37\7\37\u01d8\n\37\f\37\16\37\u01db\13\37")             # <<<<<<<<<<<<<<
 *         buf.write("\3 \3 \5 \u01df\n \3!\5!\u01e2\n!\3!\5!\u01e5\n!\3!\3")
 *         buf.write("!\5!\u01e9\n!\3\"\3\"\5\"\u01ed\n\"\3\"\3\"\3#\3#\5#\u01f3")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 56, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__90, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 56, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":57
 *         buf.write("\35\5\35\u01cd\n\35\3\36\3\36\5\36\u01d1\n\36\3\36\3\36")
 *         buf.write("\3\37\3\37\3\37\7\37\u01d8\n\37\f\37\16\37\u01db\13\37")
 *         buf.write("\3 \3 \5 \u01df\n \3!\5!\u01e2\n!\3!\5!\u01e5\n!\3!\3")             # <<<<<<<<<<<<<<
 *         buf.write("!\5!\u01e9\n!\3\"\3\"\5\"\u01ed\n\"\3\"\3\"\3#\3#\5#\u01f3")
 *         buf.write("\n#\3#\5#\u01f6\n#\3#\3#\5#\u01fa\n#\3#\5#\u01fd\n#\3")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 57, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__92, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 57, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":58
 *         buf.write("\3\37\3\37\3\37\7\37\u01d8\n\37\f\37\16\37\u01db\13\37")
 *         buf.write("\3 \3 \5 \u01df\n \3!\5!\u01e2\n!\3!\5!\u01e5\n!\3!\3")
 *         buf.write("!\5!\u01e9\n!\3\"\3\"\5\"\u01ed\n\"\3\"\3\"\3#\3#\5#\u01f3")             # <<<<<<<<<<<<<<
 *         buf.write("\n#\3#\5#\u01f6\n#\3#\3#\5#\u01fa\n#\3#\5#\u01fd\n#\3")
 *         buf.write("$\3$\3$\5$\u0202\n$\3$\5$\u0205\n$\3%\3%\3%\3%\3&\3&\3")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 58, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__94, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 58, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":59
 *         buf.write("\3 \3 \5 \u01df\n \3!\5!\u01e2\n!\3!\5!\u01e5\n!\3!\3")
 *         buf.write("!\5!\u01e9\n!\3\"\3\"\5\"\u01ed\n\"\3\"\3\"\3#\3#\5#\u01f3")
 *         buf.write("\n#\3#\5#\u01f6\n#\3#\3#\5#\u01fa\n#\3#\5#\u01fd\n#\3")             # <<<<<<<<<<<<<<
 *         buf.write("$\3$\3$\5$\u0202\n$\3$\5$\u0205\n$\3%\3%\3%\3%\3&\3&\3")
 *         buf.write("&\3\'\3\'\3\'\3\'\3\'\3\'\5\'\u0214\n\'\3\'\5\'\u0217")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 59, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__96, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 59, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":60
 *         buf.write("!\5!\u01e9\n!\3\"\3\"\5\"\u01ed\n\"\3\"\3\"\3#\3#\5#\u01f3")
 *         buf.write("\n#\3#\5#\u01f6\n#\3#\3#\5#\u01fa\n#\3#\5#\u01fd\n#\3")
 *         buf.write("$\3$\3$\5$\u0202\n$\3$\5$\u0205\n$\3%\3%\3%\3%\3&\3&\3")             # <<<<<<<<<<<<<<
 *         buf.write("&\3\'\3\'\3\'\3\'\3\'\3\'\5\'\u0214\n\'\3\'\5\'\u0217")
 *         buf.write("\n\'\3\'\3\'\3\'\3\'\3\'\5\'\u021e\n\'\3\'\5\'\u0221\n")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 60, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__98, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 60, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":61
 *         buf.write("\n#\3#\5#\u01f6\n#\3#\3#\5#\u01fa\n#\3#\5#\u01fd\n#\3")
 *         buf.write("$\3$\3$\5$\u0202\n$\3$\5$\u0205\n$\3%\3%\3%\3%\3&\3&\3")
 *         buf.write("&\3\'\3\'\3\'\3\'\3\'\3\'\5\'\u0214\n\'\3\'\5\'\u0217")             # <<<<<<<<<<<<<<
 *         buf.write("\n\'\3\'\3\'\3\'\3\'\3\'\5\'\u021e\n\'\3\'\5\'\u0221\n")
 *         buf.write("\'\3\'\3\'\5\'\u0225\n\'\3(\5(\u0228\n(\3(\3(\3(\3(\7")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 61, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__100, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 61, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":62
 *         buf.write("$\3$\3$\5$\u0202\n$\3$\5$\u0205\n$\3%\3%\3%\3%\3&\3&\3")
 *         buf.write("&\3\'\3\'\3\'\3\'\3\'\3\'\5\'\u0214\n\'\3\'\5\'\u0217")
 *         buf.write("\n\'\3\'\3\'\3\'\3\'\3\'\5\'\u021e\n\'\3\'\5\'\u0221\n")             # <<<<<<<<<<<<<<
 *         buf.write("\'\3\'\3\'\5\'\u0225\n\'\3(\5(\u0228\n(\3(\3(\3(\3(\7")
 *         buf.write("(\u022e\n(\f(\16(\u0231\13(\3)\5)\u0234\n)\3)\3)\3)\3")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 62, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__102, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 62, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":63
 *         buf.write("&\3\'\3\'\3\'\3\'\3\'\3\'\5\'\u0214\n\'\3\'\5\'\u0217")
 *         buf.write("\n\'\3\'\3\'\3\'\3\'\3\'\5\'\u021e\n\'\3\'\5\'\u0221\n")
 *         buf.write("\'\3\'\3\'\5\'\u0225\n\'\3(\5(\u0228\n(\3(\3(\3(\3(\7")             # <<<<<<<<<<<<<<
 *         buf.write("(\u022e\n(\f(\16(\u0231\13(\3)\5)\u0234\n)\3)\3)\3)\3")
 *         buf.write(")\7)\u023a\n)\f)\16)\u023d\13)\3*\3*\3*\3*\3*\3*\3*\3")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 63, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__104, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 63, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":64
 *         buf.write("\n\'\3\'\3\'\3\'\3\'\3\'\5\'\u021e\n\'\3\'\5\'\u0221\n")
 *         buf.write("\'\3\'\3\'\5\'\u0225\n\'\3(\5(\u0228\n(\3(\3(\3(\3(\7")
 *         buf.write("(\u022e\n(\f(\16(\u0231\13(\3)\5)\u0234\n)\3)\3)\3)\3")             # <<<<<<<<<<<<<<
 *         buf.write(")\7)\u023a\n)\f)\16)\u023d\13)\3*\3*\3*\3*\3*\3*\3*\3")
 *         buf.write("*\3*\3*\3*\5*\u024a\n*\3+\3+\3+\3,\3,\3,\3,\5,\u0253\n")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 64, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__106, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 64, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":65
 *         buf.write("\'\3\'\3\'\5\'\u0225\n\'\3(\5(\u0228\n(\3(\3(\3(\3(\7")
 *         buf.write("(\u022e\n(\f(\16(\u0231\13(\3)\5)\u0234\n)\3)\3)\3)\3")
 *         buf.write(")\7)\u023a\n)\f)\16)\u023d\13)\3*\3*\3*\3*\3*\3*\3*\3")             # <<<<<<<<<<<<<<
 *         buf.write("*\3*\3*\3*\5*\u024a\n*\3+\3+\3+\3,\3,\3,\3,\5,\u0253\n")
 *         buf.write(",\3,\3,\3,\3,\7,\u0259\n,\f,\16,\u025c\13,\3,\3,\3,\3")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 65, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__108, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 65, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":66
 *         buf.write("(\u022e\n(\f(\16(\u0231\13(\3)\5)\u0234\n)\3)\3)\3)\3")
 *         buf.write(")\7)\u023a\n)\f)\16)\u023d\13)\3*\3*\3*\3*\3*\3*\3*\3")
 *         buf.write("*\3*\3*\3*\5*\u024a\n*\3+\3+\3+\3,\3,\3,\3,\5,\u0253\n")             # <<<<<<<<<<<<<<
 *         buf.write(",\3,\3,\3,\3,\7,\u0259\n,\f,\16,\u025c\13,\3,\3,\3,\3")
 *         buf.write(",\3,\3,\3,\3,\3,\3,\3,\5,\u0269\n,\3-\3-\3-\3.\3.\3.\3")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 66, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__110, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 66, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":67
 *         buf.write(")\7)\u023a\n)\f)\16)\u023d\13)\3*\3*\3*\3*\3*\3*\3*\3")
 *         buf.write("*\3*\3*\3*\5*\u024a\n*\3+\3+\3+\3,\3,\3,\3,\5,\u0253\n")
 *         buf.write(",\3,\3,\3,\3,\7,\u0259\n,\f,\16,\u025c\13,\3,\3,\3,\3")             # <<<<<<<<<<<<<<
 *         buf.write(",\3,\3,\3,\3,\3,\3,\3,\5,\u0269\n,\3-\3-\3-\3.\3.\3.\3")
 *         buf.write(".\3.\3.\7.\u0274\n.\f.\16.\u0277\13.\3.\3.\3.\3.\3.\3")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 67, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__112, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 67, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":68
 *         buf.write("*\3*\3*\3*\5*\u024a\n*\3+\3+\3+\3,\3,\3,\3,\5,\u0253\n")
 *         buf.write(",\3,\3,\3,\3,\7,\u0259\n,\f,\16,\u025c\13,\3,\3,\3,\3")
 *         buf.write(",\3,\3,\3,\3,\3,\3,\3,\5,\u0269\n,\3-\3-\3-\3.\3.\3.\3")             # <<<<<<<<<<<<<<
 *         buf.write(".\3.\3.\7.\u0274\n.\f.\16.\u0277\13.\3.\3.\3.\3.\3.\3")
 *         buf.write(".\7.\u027f\n.\f.\16.\u0282\13.\7.\u0284\n.\f.\16.\u0287")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 68, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__114, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 68, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":69
 *         buf.write(",\3,\3,\3,\3,\7,\u0259\n,\f,\16,\u025c\13,\3,\3,\3,\3")
 *         buf.write(",\3,\3,\3,\3,\3,\3,\3,\5,\u0269\n,\3-\3-\3-\3.\3.\3.\3")
 *         buf.write(".\3.\3.\7.\u0274\n.\f.\16.\u0277\13.\3.\3.\3.\3.\3.\3")             # <<<<<<<<<<<<<<
 *         buf.write(".\7.\u027f\n.\f.\16.\u0282\13.\7.\u0284\n.\f.\16.\u0287")
 *         buf.write("\13.\3.\3.\3.\3.\7.\u028d\n.\f.\16.\u0290\13.\5.\u0292")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 69, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__116, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 69, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":70
 *         buf.write(",\3,\3,\3,\3,\3,\3,\3,\5,\u0269\n,\3-\3-\3-\3.\3.\3.\3")
 *         buf.write(".\3.\3.\7.\u0274\n.\f.\16.\u0277\13.\3.\3.\3.\3.\3.\3")
 *         buf.write(".\7.\u027f\n.\f.\16.\u0282\13.\7.\u0284\n.\f.\16.\u0287")             # <<<<<<<<<<<<<<
 *         buf.write("\13.\3.\3.\3.\3.\7.\u028d\n.\f.\16.\u0290\13.\5.\u0292")
 *         buf.write("\n.\3.\3.\3.\3/\3/\3/\3/\3/\3/\7/\u029d\n/\f/\16/\u02a0")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 70, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__118, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 70, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":71
 *         buf.write(".\3.\3.\7.\u0274\n.\f.\16.\u0277\13.\3.\3.\3.\3.\3.\3")
 *         buf.write(".\7.\u027f\n.\f.\16.\u0282\13.\7.\u0284\n.\f.\16.\u0287")
 *         buf.write("\13.\3.\3.\3.\3.\7.\u028d\n.\f.\16.\u0290\13.\5.\u0292")             # <<<<<<<<<<<<<<
 *         buf.write("\n.\3.\3.\3.\3/\3/\3/\3/\3/\3/\7/\u029d\n/\f/\16/\u02a0")
 *         buf.write("\13/\3/\3/\3/\3/\3/\3/\7/\u02a8\n/\f/\16/\u02ab\13/\7")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 71, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__120, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 71, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":72
 *         buf.write(".\7.\u027f\n.\f.\16.\u0282\13.\7.\u0284\n.\f.\16.\u0287")
 *         buf.write("\13.\3.\3.\3.\3.\7.\u028d\n.\f.\16.\u0290\13.\5.\u0292")
 *         buf.write("\n.\3.\3.\3.\3/\3/\3/\3/\3/\3/\7/\u029d\n/\f/\16/\u02a0")             # <<<<<<<<<<<<<<
 *         buf.write("\13/\3/\3/\3/\3/\3/\3/\7/\u02a8\n/\f/\16/\u02ab\13/\7")
 *         buf.write("/\u02ad\n/\f/\16/\u02b0\13/\3/\3/\3/\3/\7/\u02b6\n/\f")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 72, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__122, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 72, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":73
 *         buf.write("\13.\3.\3.\3.\3.\7.\u028d\n.\f.\16.\u0290\13.\5.\u0292")
 *         buf.write("\n.\3.\3.\3.\3/\3/\3/\3/\3/\3/\7/\u029d\n/\f/\16/\u02a0")
 *         buf.write("\13/\3/\3/\3/\3/\3/\3/\7/\u02a8\n/\f/\16/\u02ab\13/\7")             # <<<<<<<<<<<<<<
 *         buf.write("/\u02ad\n/\f/\16/\u02b0\13/\3/\3/\3/\3/\7/\u02b6\n/\f")
 *         buf.write("/\16/\u02b9\13/\5/\u02bb\n/\3/\3/\3/\3\60\3\60\3\60\3")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 73, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__124, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 73, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":74
 *         buf.write("\n.\3.\3.\3.\3/\3/\3/\3/\3/\3/\7/\u029d\n/\f/\16/\u02a0")
 *         buf.write("\13/\3/\3/\3/\3/\3/\3/\7/\u02a8\n/\f/\16/\u02ab\13/\7")
 *         buf.write("/\u02ad\n/\f/\16/\u02b0\13/\3/\3/\3/\3/\7/\u02b6\n/\f")             # <<<<<<<<<<<<<<
 *         buf.write("/\16/\u02b9\13/\5/\u02bb\n/\3/\3/\3/\3\60\3\60\3\60\3")
 *         buf.write("\60\3\60\3\60\7\60\u02c6\n\60\f\60\16\60\u02c9\13\60\3")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 74, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__126, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 74, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":75
 *         buf.write("\13/\3/\3/\3/\3/\3/\3/\7/\u02a8\n/\f/\16/\u02ab\13/\7")
 *         buf.write("/\u02ad\n/\f/\16/\u02b0\13/\3/\3/\3/\3/\7/\u02b6\n/\f")
 *         buf.write("/\16/\u02b9\13/\5/\u02bb\n/\3/\3/\3/\3\60\3\60\3\60\3")             # <<<<<<<<<<<<<<
 *         buf.write("\60\3\60\3\60\7\60\u02c6\n\60\f\60\16\60\u02c9\13\60\3")
 *         buf.write("\60\3\60\3\60\3\61\3\61\3\61\3\61\3\61\3\61\7\61\u02d4")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 75, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__127, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 75, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":76
 *         buf.write("/\u02ad\n/\f/\16/\u02b0\13/\3/\3/\3/\3/\7/\u02b6\n/\f")
 *         buf.write("/\16/\u02b9\13/\5/\u02bb\n/\3/\3/\3/\3\60\3\60\3\60\3")
 *         buf.write("\60\3\60\3\60\7\60\u02c6\n\60\f\60\16\60\u02c9\13\60\3")             # <<<<<<<<<<<<<<
 *         buf.write("\60\3\60\3\60\3\61\3\61\3\61\3\61\3\61\3\61\7\61\u02d4")
 *         buf.write("\n\61\f\61\16\61\u02d7\13\61\3\61\3\61\3\61\3\62\3\62")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 76, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__128, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 76, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":77
 *         buf.write("/\16/\u02b9\13/\5/\u02bb\n/\3/\3/\3/\3\60\3\60\3\60\3")
 *         buf.write("\60\3\60\3\60\7\60\u02c6\n\60\f\60\16\60\u02c9\13\60\3")
 *         buf.write("\60\3\60\3\60\3\61\3\61\3\61\3\61\3\61\3\61\7\61\u02d4")             # <<<<<<<<<<<<<<
 *         buf.write("\n\61\f\61\16\61\u02d7\13\61\3\61\3\61\3\61\3\62\3\62")
 *         buf.write("\3\62\7\62\u02df\n\62\f\62\16\62\u02e2\13\62\3\63\3\63")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 77, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__129, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 77, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":78
 *         buf.write("\60\3\60\3\60\7\60\u02c6\n\60\f\60\16\60\u02c9\13\60\3")
 *         buf.write("\60\3\60\3\60\3\61\3\61\3\61\3\61\3\61\3\61\7\61\u02d4")
 *         buf.write("\n\61\f\61\16\61\u02d7\13\61\3\61\3\61\3\61\3\62\3\62")             # <<<<<<<<<<<<<<
 *         buf.write("\3\62\7\62\u02df\n\62\f\62\16\62\u02e2\13\62\3\63\3\63")
 *         buf.write("\3\63\5\63\u02e7\n\63\3\64\3\64\3\64\3\64\3\64\3\64\7")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 78, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__130, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 78, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":79
 *         buf.write("\60\3\60\3\60\3\61\3\61\3\61\3\61\3\61\3\61\7\61\u02d4")
 *         buf.write("\n\61\f\61\16\61\u02d7\13\61\3\61\3\61\3\61\3\62\3\62")
 *         buf.write("\3\62\7\62\u02df\n\62\f\62\16\62\u02e2\13\62\3\63\3\63")             # <<<<<<<<<<<<<<
 *         buf.write("\3\63\5\63\u02e7\n\63\3\64\3\64\3\64\3\64\3\64\3\64\7")
 *         buf.write("\64\u02ef\n\64\f\64\16\64\u02f2\13\64\3\64\3\64\3\64\3")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 79, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__131, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 79, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":80
 *         buf.write("\n\61\f\61\16\61\u02d7\13\61\3\61\3\61\3\61\3\62\3\62")
 *         buf.write("\3\62\7\62\u02df\n\62\f\62\16\62\u02e2\13\62\3\63\3\63")
 *         buf.write("\3\63\5\63\u02e7\n\63\3\64\3\64\3\64\3\64\3\64\3\64\7")             # <<<<<<<<<<<<<<
 *         buf.write("\64\u02ef\n\64\f\64\16\64\u02f2\13\64\3\64\3\64\3\64\3")
 *         buf.write("\65\3\65\3\65\3\65\3\65\3\65\7\65\u02fd\n\65\f\65\16\65")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 80, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__132, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 80, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":81
 *         buf.write("\3\62\7\62\u02df\n\62\f\62\16\62\u02e2\13\62\3\63\3\63")
 *         buf.write("\3\63\5\63\u02e7\n\63\3\64\3\64\3\64\3\64\3\64\3\64\7")
 *         buf.write("\64\u02ef\n\64\f\64\16\64\u02f2\13\64\3\64\3\64\3\64\3")             # <<<<<<<<<<<<<<
 *         buf.write("\65\3\65\3\65\3\65\3\65\3\65\7\65\u02fd\n\65\f\65\16\65")
 *         buf.write("\u0300\13\65\3\65\3\65\3\65\3\65\3\65\3\65\7\65\u0308")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 81, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__133, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 81, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":82
 *         buf.write("\3\63\5\63\u02e7\n\63\3\64\3\64\3\64\3\64\3\64\3\64\7")
 *         buf.write("\64\u02ef\n\64\f\64\16\64\u02f2\13\64\3\64\3\64\3\64\3")
 *         buf.write("\65\3\65\3\65\3\65\3\65\3\65\7\65\u02fd\n\65\f\65\16\65")             # <<<<<<<<<<<<<<
 *         buf.write("\u0300\13\65\3\65\3\65\3\65\3\65\3\65\3\65\7\65\u0308")
 *         buf.write("\n\65\f\65\16\65\u030b\13\65\7\65\u030d\n\65\f\65\16\65")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 82, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__134, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 82, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":83
 *         buf.write("\64\u02ef\n\64\f\64\16\64\u02f2\13\64\3\64\3\64\3\64\3")
 *         buf.write("\65\3\65\3\65\3\65\3\65\3\65\7\65\u02fd\n\65\f\65\16\65")
 *         buf.write("\u0300\13\65\3\65\3\65\3\65\3\65\3\65\3\65\7\65\u0308")             # <<<<<<<<<<<<<<
 *         buf.write("\n\65\f\65\16\65\u030b\13\65\7\65\u030d\n\65\f\65\16\65")
 *         buf.write("\u0310\13\65\3\65\3\65\3\65\3\66\3\66\3\66\3\66\3\66\3")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 83, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__135, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 83, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":84
 *         buf.write("\65\3\65\3\65\3\65\3\65\3\65\7\65\u02fd\n\65\f\65\16\65")
 *         buf.write("\u0300\13\65\3\65\3\65\3\65\3\65\3\65\3\65\7\65\u0308")
 *         buf.write("\n\65\f\65\16\65\u030b\13\65\7\65\u030d\n\65\f\65\16\65")             # <<<<<<<<<<<<<<
 *         buf.write("\u0310\13\65\3\65\3\65\3\65\3\66\3\66\3\66\3\66\3\66\3")
 *         buf.write("\66\7\66\u031b\n\66\f\66\16\66\u031e\13\66\3\66\3\66\3")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 84, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__136, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 84, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":85
 *         buf.write("\u0300\13\65\3\65\3\65\3\65\3\65\3\65\3\65\7\65\u0308")
 *         buf.write("\n\65\f\65\16\65\u030b\13\65\7\65\u030d\n\65\f\65\16\65")
 *         buf.write("\u0310\13\65\3\65\3\65\3\65\3\66\3\66\3\66\3\66\3\66\3")             # <<<<<<<<<<<<<<
 *         buf.write("\66\7\66\u031b\n\66\f\66\16\66\u031e\13\66\3\66\3\66\3")
 *         buf.write("\66\3\66\3\66\3\66\7\66\u0326\n\66\f\66\16\66\u0329\13")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 85, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__137, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 85, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":86
 *         buf.write("\n\65\f\65\16\65\u030b\13\65\7\65\u030d\n\65\f\65\16\65")
 *         buf.write("\u0310\13\65\3\65\3\65\3\65\3\66\3\66\3\66\3\66\3\66\3")
 *         buf.write("\66\7\66\u031b\n\66\f\66\16\66\u031e\13\66\3\66\3\66\3")             # <<<<<<<<<<<<<<
 *         buf.write("\66\3\66\3\66\3\66\7\66\u0326\n\66\f\66\16\66\u0329\13")
 *         buf.write("\66\7\66\u032b\n\66\f\66\16\66\u032e\13\66\3\66\3\66\3")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 86, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__138, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 86, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":87
 *         buf.write("\u0310\13\65\3\65\3\65\3\65\3\66\3\66\3\66\3\66\3\66\3")
 *         buf.write("\66\7\66\u031b\n\66\f\66\16\66\u031e\13\66\3\66\3\66\3")
 *         buf.write("\66\3\66\3\66\3\66\7\66\u0326\n\66\f\66\16\66\u0329\13")             # <<<<<<<<<<<<<<
 *         buf.write("\66\7\66\u032b\n\66\f\66\16\66\u032e\13\66\3\66\3\66\3")
 *         buf.write("\66\3\67\3\67\3\67\3\67\3\67\3\67\3\67\38\38\38\38\38")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 87, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__139, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 87, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":88
 *         buf.write("\66\7\66\u031b\n\66\f\66\16\66\u031e\13\66\3\66\3\66\3")
 *         buf.write("\66\3\66\3\66\3\66\7\66\u0326\n\66\f\66\16\66\u0329\13")
 *         buf.write("\66\7\66\u032b\n\66\f\66\16\66\u032e\13\66\3\66\3\66\3")             # <<<<<<<<<<<<<<
 *         buf.write("\66\3\67\3\67\3\67\3\67\3\67\3\67\3\67\38\38\38\38\38")
 *         buf.write("\38\38\38\38\38\78\u0344\n8\f8\168\u0347\138\38\38\38")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 88, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__140, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 88, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":89
 *         buf.write("\66\3\66\3\66\3\66\7\66\u0326\n\66\f\66\16\66\u0329\13")
 *         buf.write("\66\7\66\u032b\n\66\f\66\16\66\u032e\13\66\3\66\3\66\3")
 *         buf.write("\66\3\67\3\67\3\67\3\67\3\67\3\67\3\67\38\38\38\38\38")             # <<<<<<<<<<<<<<
 *         buf.write("\38\38\38\38\38\78\u0344\n8\f8\168\u0347\138\38\38\38")
 *         buf.write("\58\u034c\n8\39\39\39\39\39\59\u0353\n9\59\u0355\n9\3")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 89, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__141, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 89, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":90
 *         buf.write("\66\7\66\u032b\n\66\f\66\16\66\u032e\13\66\3\66\3\66\3")
 *         buf.write("\66\3\67\3\67\3\67\3\67\3\67\3\67\3\67\38\38\38\38\38")
 *         buf.write("\38\38\38\38\38\78\u0344\n8\f8\168\u0347\138\38\38\38")             # <<<<<<<<<<<<<<
 *         buf.write("\58\u034c\n8\39\39\39\39\39\59\u0353\n9\59\u0355\n9\3")
 *         buf.write(":\3:\3:\3:\3:\3:\3:\3:\3:\3:\5:\u0361\n:\3:\3:\3:\3:\3")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 90, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__142, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 90, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":91
 *         buf.write("\66\3\67\3\67\3\67\3\67\3\67\3\67\3\67\38\38\38\38\38")
 *         buf.write("\38\38\38\38\38\78\u0344\n8\f8\168\u0347\138\38\38\38")
 *         buf.write("\58\u034c\n8\39\39\39\39\39\59\u0353\n9\59\u0355\n9\3")             # <<<<<<<<<<<<<<
 *         buf.write(":\3:\3:\3:\3:\3:\3:\3:\3:\3:\5:\u0361\n:\3:\3:\3:\3:\3")
 *         buf.write(":\3:\3:\3:\3:\3:\3:\3:\3:\3:\3:\7:\u0372\n:\f:\16:\u0375")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 91, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__143, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 91, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":92
 *         buf.write("\38\38\38\38\38\78\u0344\n8\f8\168\u0347\138\38\38\38")
 *         buf.write("\58\u034c\n8\39\39\39\39\39\59\u0353\n9\59\u0355\n9\3")
 *         buf.write(":\3:\3:\3:\3:\3:\3:\3:\3:\3:\5:\u0361\n:\3:\3:\3:\3:\3")             # <<<<<<<<<<<<<<
 *         buf.write(":\3:\3:\3:\3:\3:\3:\3:\3:\3:\3:\7:\u0372\n:\f:\16:\u0375")
 *         buf.write("\13:\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 92, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__145, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 92, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":93
 *         buf.write("\58\u034c\n8\39\39\39\39\39\59\u0353\n9\59\u0355\n9\3")
 *         buf.write(":\3:\3:\3:\3:\3:\3:\3:\3:\3:\5:\u0361\n:\3:\3:\3:\3:\3")
 *         buf.write(":\3:\3:\3:\3:\3:\3:\3:\3:\3:\3:\7:\u0372\n:\f:\16:\u0375")             # <<<<<<<<<<<<<<
 *         buf.write("\13:\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3")
 *         buf.write(";\3;\3;\3;\7;\u038b\n;\f;\16;\u038e\13;\3;\3;\3;\3;\3")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 93, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__147, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 93, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":94
 *         buf.write(":\3:\3:\3:\3:\3:\3:\3:\3:\3:\5:\u0361\n:\3:\3:\3:\3:\3")
 *         buf.write(":\3:\3:\3:\3:\3:\3:\3:\3:\3:\3:\7:\u0372\n:\f:\16:\u0375")
 *         buf.write("\13:\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3")             # <<<<<<<<<<<<<<
 *         buf.write(";\3;\3;\3;\7;\u038b\n;\f;\16;\u038e\13;\3;\3;\3;\3;\3")
 *         buf.write(";\3;\3;\5;\u0397\n;\3<\5<\u039a\n<\3<\3<\3<\7<\u039f\n")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 94, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__149, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 94, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":95
 *         buf.write(":\3:\3:\3:\3:\3:\3:\3:\3:\3:\3:\7:\u0372\n:\f:\16:\u0375")
 *         buf.write("\13:\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3")
 *         buf.write(";\3;\3;\3;\7;\u038b\n;\f;\16;\u038e\13;\3;\3;\3;\3;\3")             # <<<<<<<<<<<<<<
 *         buf.write(";\3;\3;\5;\u0397\n;\3<\5<\u039a\n<\3<\3<\3<\7<\u039f\n")
 *         buf.write("<\f<\16<\u03a2\13<\3=\3=\5=\u03a6\n=\3>\3>\3>\7>\u03ab")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 95, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__151, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 95, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":96
 *         buf.write("\13:\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3")
 *         buf.write(";\3;\3;\3;\7;\u038b\n;\f;\16;\u038e\13;\3;\3;\3;\3;\3")
 *         buf.write(";\3;\3;\5;\u0397\n;\3<\5<\u039a\n<\3<\3<\3<\7<\u039f\n")             # <<<<<<<<<<<<<<
 *         buf.write("<\f<\16<\u03a2\13<\3=\3=\5=\u03a6\n=\3>\3>\3>\7>\u03ab")
 *         buf.write("\n>\f>\16>\u03ae\13>\3?\3?\5?\u03b2\n?\3?\3?\3@\3@\3@")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 96, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__153, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 96, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":97
 *         buf.write(";\3;\3;\3;\7;\u038b\n;\f;\16;\u038e\13;\3;\3;\3;\3;\3")
 *         buf.write(";\3;\3;\5;\u0397\n;\3<\5<\u039a\n<\3<\3<\3<\7<\u039f\n")
 *         buf.write("<\f<\16<\u03a2\13<\3=\3=\5=\u03a6\n=\3>\3>\3>\7>\u03ab")             # <<<<<<<<<<<<<<
 *         buf.write("\n>\f>\16>\u03ae\13>\3?\3?\5?\u03b2\n?\3?\3?\3@\3@\3@")
 *         buf.write("\3@\3@\7@\u03bb\n@\f@\16@\u03be\13@\3@\5@\u03c1\n@\3A")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 97, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__155, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 97, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":98
 *         buf.write(";\3;\3;\5;\u0397\n;\3<\5<\u039a\n<\3<\3<\3<\7<\u039f\n")
 *         buf.write("<\f<\16<\u03a2\13<\3=\3=\5=\u03a6\n=\3>\3>\3>\7>\u03ab")
 *         buf.write("\n>\f>\16>\u03ae\13>\3?\3?\5?\u03b2\n?\3?\3?\3@\3@\3@")             # <<<<<<<<<<<<<<
 *         buf.write("\3@\3@\7@\u03bb\n@\f@\16@\u03be\13@\3@\5@\u03c1\n@\3A")
 *         buf.write("\3A\3A\7A\u03c6\nA\fA\16A\u03c9\13A\3B\3B\3B\3B\3C\3C")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 98, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__157, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 98, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":99
 *         buf.write("<\f<\16<\u03a2\13<\3=\3=\5=\u03a6\n=\3>\3>\3>\7>\u03ab")
 *         buf.write("\n>\f>\16>\u03ae\13>\3?\3?\5?\u03b2\n?\3?\3?\3@\3@\3@")
 *         buf.write("\3@\3@\7@\u03bb\n@\f@\16@\u03be\13@\3@\5@\u03c1\n@\3A")             # <<<<<<<<<<<<<<
 *         buf.write("\3A\3A\7A\u03c6\nA\fA\16A\u03c9\13A\3B\3B\3B\3B\3C\3C")
 *         buf.write("\3C\3C\5C\u03d3\nC\3C\3C\3C\5C\u03d8\nC\3D\5D\u03db\n")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 99, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__158, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 99, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":100
 *         buf.write("\n>\f>\16>\u03ae\13>\3?\3?\5?\u03b2\n?\3?\3?\3@\3@\3@")
 *         buf.write("\3@\3@\7@\u03bb\n@\f@\16@\u03be\13@\3@\5@\u03c1\n@\3A")
 *         buf.write("\3A\3A\7A\u03c6\nA\fA\16A\u03c9\13A\3B\3B\3B\3B\3C\3C")             # <<<<<<<<<<<<<<
 *         buf.write("\3C\3C\5C\u03d3\nC\3C\3C\3C\5C\u03d8\nC\3D\5D\u03db\n")
 *         buf.write("D\3D\3D\7D\u03df\nD\fD\16D\u03e2\13D\3E\3E\3E\7E\u03e7")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 100, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__159, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 100, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":101
 *         buf.write("\3@\3@\7@\u03bb\n@\f@\16@\u03be\13@\3@\5@\u03c1\n@\3A")
 *         buf.write("\3A\3A\7A\u03c6\nA\fA\16A\u03c9\13A\3B\3B\3B\3B\3C\3C")
 *         buf.write("\3C\3C\5C\u03d3\nC\3C\3C\3C\5C\u03d8\nC\3D\5D\u03db\n")             # <<<<<<<<<<<<<<
 *         buf.write("D\3D\3D\7D\u03df\nD\fD\16D\u03e2\13D\3E\3E\3E\7E\u03e7")
 *         buf.write("\nE\fE\16E\u03ea\13E\3F\3F\3F\3F\7F\u03f0\nF\fF\16F\u03f3")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 101, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__160, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 101, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":102
 *         buf.write("\3A\3A\7A\u03c6\nA\fA\16A\u03c9\13A\3B\3B\3B\3B\3C\3C")
 *         buf.write("\3C\3C\5C\u03d3\nC\3C\3C\3C\5C\u03d8\nC\3D\5D\u03db\n")
 *         buf.write("D\3D\3D\7D\u03df\nD\fD\16D\u03e2\13D\3E\3E\3E\7E\u03e7")             # <<<<<<<<<<<<<<
 *         buf.write("\nE\fE\16E\u03ea\13E\3F\3F\3F\3F\7F\u03f0\nF\fF\16F\u03f3")
 *         buf.write("\13F\3F\3F\3G\3G\5G\u03f9\nG\3H\3H\5H\u03fd\nH\3I\3I\3")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 102, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__161, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 102, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":103
 *         buf.write("\3C\3C\5C\u03d3\nC\3C\3C\3C\5C\u03d8\nC\3D\5D\u03db\n")
 *         buf.write("D\3D\3D\7D\u03df\nD\fD\16D\u03e2\13D\3E\3E\3E\7E\u03e7")
 *         buf.write("\nE\fE\16E\u03ea\13E\3F\3F\3F\3F\7F\u03f0\nF\fF\16F\u03f3")             # <<<<<<<<<<<<<<
 *         buf.write("\13F\3F\3F\3G\3G\5G\u03f9\nG\3H\3H\5H\u03fd\nH\3I\3I\3")
 *         buf.write("I\7I\u0402\nI\fI\16I\u0405\13I\5I\u0407\nI\3J\3J\3J\3")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 103, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__162, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 103, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":104
 *         buf.write("D\3D\3D\7D\u03df\nD\fD\16D\u03e2\13D\3E\3E\3E\7E\u03e7")
 *         buf.write("\nE\fE\16E\u03ea\13E\3F\3F\3F\3F\7F\u03f0\nF\fF\16F\u03f3")
 *         buf.write("\13F\3F\3F\3G\3G\5G\u03f9\nG\3H\3H\5H\u03fd\nH\3I\3I\3")             # <<<<<<<<<<<<<<
 *         buf.write("I\7I\u0402\nI\fI\16I\u0405\13I\5I\u0407\nI\3J\3J\3J\3")
 *         buf.write("J\2\3rK\2\4\6\b\n\f\16\20\22\24\26\30\32\34\36 \"$&(*")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 104, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__163, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 104, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":105
 *         buf.write("\nE\fE\16E\u03ea\13E\3F\3F\3F\3F\7F\u03f0\nF\fF\16F\u03f3")
 *         buf.write("\13F\3F\3F\3G\3G\5G\u03f9\nG\3H\3H\5H\u03fd\nH\3I\3I\3")
 *         buf.write("I\7I\u0402\nI\fI\16I\u0405\13I\5I\u0407\nI\3J\3J\3J\3")             # <<<<<<<<<<<<<<
 *         buf.write("J\2\3rK\2\4\6\b\n\f\16\20\22\24\26\30\32\34\36 \"$&(*")
 *         buf.write(",.\60\62\64\668:<>@BDFHJLNPRTVXZ\\^`bdfhjlnprtvxz|~\u0080")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 105, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__164, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 105, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":106
 *         buf.write("\13F\3F\3F\3G\3G\5G\u03f9\nG\3H\3H\5H\u03fd\nH\3I\3I\3")
 *         buf.write("I\7I\u0402\nI\fI\16I\u0405\13I\5I\u0407\nI\3J\3J\3J\3")
 *         buf.write("J\2\3rK\2\4\6\b\n\f\16\20\22\24\26\30\32\34\36 \"$&(*")             # <<<<<<<<<<<<<<
 *         buf.write(",.\60\62\64\668:<>@BDFHJLNPRTVXZ\\^`bdfhjlnprtvxz|~\u0080")
 *         buf.write("\u0082\u0084\u0086\u0088\u008a\u008c\u008e\u0090\u0092")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 106, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__165, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 106, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":107
 *         buf.write("I\7I\u0402\nI\fI\16I\u0405\13I\5I\u0407\nI\3J\3J\3J\3")
 *         buf.write("J\2\3rK\2\4\6\b\n\f\16\20\22\24\26\30\32\34\36 \"$&(*")
 *         buf.write(",.\60\62\64\668:<>@BDFHJLNPRTVXZ\\^`bdfhjlnprtvxz|~\u0080")             # <<<<<<<<<<<<<<
 *         buf.write("\u0082\u0084\u0086\u0088\u008a\u008c\u008e\u0090\u0092")
 *         buf.write("\2\n\3\2\r\16\3\2#$\3\2%\'\3\2()\3\2;<\4\2\37\37=?\4\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 107, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__166, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 107, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":108
 *         buf.write("J\2\3rK\2\4\6\b\n\f\16\20\22\24\26\30\32\34\36 \"$&(*")
 *         buf.write(",.\60\62\64\668:<>@BDFHJLNPRTVXZ\\^`bdfhjlnprtvxz|~\u0080")
 *         buf.write("\u0082\u0084\u0086\u0088\u008a\u008c\u008e\u0090\u0092")             # <<<<<<<<<<<<<<
 *         buf.write("\2\n\3\2\r\16\3\2#$\3\2%\'\3\2()\3\2;<\4\2\37\37=?\4\2")
 *         buf.write("::@B\3\2CH\2\u046f\2\u0099\3\2\2\2\4\u00a2\3\2\2\2\6\u00a8")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 108, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__168, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 108, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":109
 *         buf.write(",.\60\62\64\668:<>@BDFHJLNPRTVXZ\\^`bdfhjlnprtvxz|~\u0080")
 *         buf.write("\u0082\u0084\u0086\u0088\u008a\u008c\u008e\u0090\u0092")
 *         buf.write("\2\n\3\2\r\16\3\2#$\3\2%\'\3\2()\3\2;<\4\2\37\37=?\4\2")             # <<<<<<<<<<<<<<
 *         buf.write("::@B\3\2CH\2\u046f\2\u0099\3\2\2\2\4\u00a2\3\2\2\2\6\u00a8")
 *         buf.write("\3\2\2\2\b\u00ae\3\2\2\2\n\u00c7\3\2\2\2\f\u0102\3\2\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 109, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__170, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 109, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":110
 *         buf.write("\u0082\u0084\u0086\u0088\u008a\u008c\u008e\u0090\u0092")
 *         buf.write("\2\n\3\2\r\16\3\2#$\3\2%\'\3\2()\3\2;<\4\2\37\37=?\4\2")
 *         buf.write("::@B\3\2CH\2\u046f\2\u0099\3\2\2\2\4\u00a2\3\2\2\2\6\u00a8")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\b\u00ae\3\2\2\2\n\u00c7\3\2\2\2\f\u0102\3\2\2")
 *         buf.write("\2\16\u0104\3\2\2\2\20\u0106\3\2\2\2\22\u010e\3\2\2\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 110, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__171, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 110, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":111
 *         buf.write("\2\n\3\2\r\16\3\2#$\3\2%\'\3\2()\3\2;<\4\2\37\37=?\4\2")
 *         buf.write("::@B\3\2CH\2\u046f\2\u0099\3\2\2\2\4\u00a2\3\2\2\2\6\u00a8")
 *         buf.write("\3\2\2\2\b\u00ae\3\2\2\2\n\u00c7\3\2\2\2\f\u0102\3\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\16\u0104\3\2\2\2\20\u0106\3\2\2\2\22\u010e\3\2\2\2")
 *         buf.write("\24\u0111\3\2\2\2\26\u012f\3\2\2\2\30\u0134\3\2\2\2\32")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 111, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__173, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 111, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":112
 *         buf.write("::@B\3\2CH\2\u046f\2\u0099\3\2\2\2\4\u00a2\3\2\2\2\6\u00a8")
 *         buf.write("\3\2\2\2\b\u00ae\3\2\2\2\n\u00c7\3\2\2\2\f\u0102\3\2\2")
 *         buf.write("\2\16\u0104\3\2\2\2\20\u0106\3\2\2\2\22\u010e\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\24\u0111\3\2\2\2\26\u012f\3\2\2\2\30\u0134\3\2\2\2\32")
 *         buf.write("\u0142\3\2\2\2\34\u0149\3\2\2\2\36\u014c\3\2\2\2 \u015c")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 112, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__175, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 112, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":113
 *         buf.write("\3\2\2\2\b\u00ae\3\2\2\2\n\u00c7\3\2\2\2\f\u0102\3\2\2")
 *         buf.write("\2\16\u0104\3\2\2\2\20\u0106\3\2\2\2\22\u010e\3\2\2\2")
 *         buf.write("\24\u0111\3\2\2\2\26\u012f\3\2\2\2\30\u0134\3\2\2\2\32")             # <<<<<<<<<<<<<<
 *         buf.write("\u0142\3\2\2\2\34\u0149\3\2\2\2\36\u014c\3\2\2\2 \u015c")
 *         buf.write("\3\2\2\2\"\u0171\3\2\2\2$\u0184\3\2\2\2&\u018c\3\2\2\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 113, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__177, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 113, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":114
 *         buf.write("\2\16\u0104\3\2\2\2\20\u0106\3\2\2\2\22\u010e\3\2\2\2")
 *         buf.write("\24\u0111\3\2\2\2\26\u012f\3\2\2\2\30\u0134\3\2\2\2\32")
 *         buf.write("\u0142\3\2\2\2\34\u0149\3\2\2\2\36\u014c\3\2\2\2 \u015c")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\"\u0171\3\2\2\2$\u0184\3\2\2\2&\u018c\3\2\2\2")
 *         buf.write("(\u0194\3\2\2\2*\u0199\3\2\2\2,\u01a1\3\2\2\2.\u01a9\3")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 114, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__179, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 114, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":115
 *         buf.write("\24\u0111\3\2\2\2\26\u012f\3\2\2\2\30\u0134\3\2\2\2\32")
 *         buf.write("\u0142\3\2\2\2\34\u0149\3\2\2\2\36\u014c\3\2\2\2 \u015c")
 *         buf.write("\3\2\2\2\"\u0171\3\2\2\2$\u0184\3\2\2\2&\u018c\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("(\u0194\3\2\2\2*\u0199\3\2\2\2,\u01a1\3\2\2\2.\u01a9\3")
 *         buf.write("\2\2\2\60\u01ab\3\2\2\2\62\u01b3\3\2\2\2\64\u01b9\3\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 115, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__181, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 115, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":116
 *         buf.write("\u0142\3\2\2\2\34\u0149\3\2\2\2\36\u014c\3\2\2\2 \u015c")
 *         buf.write("\3\2\2\2\"\u0171\3\2\2\2$\u0184\3\2\2\2&\u018c\3\2\2\2")
 *         buf.write("(\u0194\3\2\2\2*\u0199\3\2\2\2,\u01a1\3\2\2\2.\u01a9\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\60\u01ab\3\2\2\2\62\u01b3\3\2\2\2\64\u01b9\3\2")
 *         buf.write("\2\2\66\u01bc\3\2\2\28\u01cc\3\2\2\2:\u01ce\3\2\2\2<\u01d4")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 116, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__183, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 116, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":117
 *         buf.write("\3\2\2\2\"\u0171\3\2\2\2$\u0184\3\2\2\2&\u018c\3\2\2\2")
 *         buf.write("(\u0194\3\2\2\2*\u0199\3\2\2\2,\u01a1\3\2\2\2.\u01a9\3")
 *         buf.write("\2\2\2\60\u01ab\3\2\2\2\62\u01b3\3\2\2\2\64\u01b9\3\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\66\u01bc\3\2\2\28\u01cc\3\2\2\2:\u01ce\3\2\2\2<\u01d4")
 *         buf.write("\3\2\2\2>\u01de\3\2\2\2@\u01e1\3\2\2\2B\u01ea\3\2\2\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 117, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__184, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 117, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":118
 *         buf.write("(\u0194\3\2\2\2*\u0199\3\2\2\2,\u01a1\3\2\2\2.\u01a9\3")
 *         buf.write("\2\2\2\60\u01ab\3\2\2\2\62\u01b3\3\2\2\2\64\u01b9\3\2")
 *         buf.write("\2\2\66\u01bc\3\2\2\28\u01cc\3\2\2\2:\u01ce\3\2\2\2<\u01d4")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2>\u01de\3\2\2\2@\u01e1\3\2\2\2B\u01ea\3\2\2\2")
 *         buf.write("D\u01f0\3\2\2\2F\u01fe\3\2\2\2H\u0206\3\2\2\2J\u020a\3")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 118, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__185, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 118, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":119
 *         buf.write("\2\2\2\60\u01ab\3\2\2\2\62\u01b3\3\2\2\2\64\u01b9\3\2")
 *         buf.write("\2\2\66\u01bc\3\2\2\28\u01cc\3\2\2\2:\u01ce\3\2\2\2<\u01d4")
 *         buf.write("\3\2\2\2>\u01de\3\2\2\2@\u01e1\3\2\2\2B\u01ea\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("D\u01f0\3\2\2\2F\u01fe\3\2\2\2H\u0206\3\2\2\2J\u020a\3")
 *         buf.write("\2\2\2L\u020d\3\2\2\2N\u0227\3\2\2\2P\u0233\3\2\2\2R\u0249")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 119, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__186, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 119, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":120
 *         buf.write("\2\2\66\u01bc\3\2\2\28\u01cc\3\2\2\2:\u01ce\3\2\2\2<\u01d4")
 *         buf.write("\3\2\2\2>\u01de\3\2\2\2@\u01e1\3\2\2\2B\u01ea\3\2\2\2")
 *         buf.write("D\u01f0\3\2\2\2F\u01fe\3\2\2\2H\u0206\3\2\2\2J\u020a\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2L\u020d\3\2\2\2N\u0227\3\2\2\2P\u0233\3\2\2\2R\u0249")
 *         buf.write("\3\2\2\2T\u024b\3\2\2\2V\u0268\3\2\2\2X\u026a\3\2\2\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 120, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__187, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 120, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":121
 *         buf.write("\3\2\2\2>\u01de\3\2\2\2@\u01e1\3\2\2\2B\u01ea\3\2\2\2")
 *         buf.write("D\u01f0\3\2\2\2F\u01fe\3\2\2\2H\u0206\3\2\2\2J\u020a\3")
 *         buf.write("\2\2\2L\u020d\3\2\2\2N\u0227\3\2\2\2P\u0233\3\2\2\2R\u0249")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2T\u024b\3\2\2\2V\u0268\3\2\2\2X\u026a\3\2\2\2")
 *         buf.write("Z\u026d\3\2\2\2\\\u0296\3\2\2\2^\u02bf\3\2\2\2`\u02cd")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 121, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__188, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 121, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":122
 *         buf.write("D\u01f0\3\2\2\2F\u01fe\3\2\2\2H\u0206\3\2\2\2J\u020a\3")
 *         buf.write("\2\2\2L\u020d\3\2\2\2N\u0227\3\2\2\2P\u0233\3\2\2\2R\u0249")
 *         buf.write("\3\2\2\2T\u024b\3\2\2\2V\u0268\3\2\2\2X\u026a\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("Z\u026d\3\2\2\2\\\u0296\3\2\2\2^\u02bf\3\2\2\2`\u02cd")
 *         buf.write("\3\2\2\2b\u02db\3\2\2\2d\u02e3\3\2\2\2f\u02e8\3\2\2\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 122, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__189, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 122, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":123
 *         buf.write("\2\2\2L\u020d\3\2\2\2N\u0227\3\2\2\2P\u0233\3\2\2\2R\u0249")
 *         buf.write("\3\2\2\2T\u024b\3\2\2\2V\u0268\3\2\2\2X\u026a\3\2\2\2")
 *         buf.write("Z\u026d\3\2\2\2\\\u0296\3\2\2\2^\u02bf\3\2\2\2`\u02cd")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2b\u02db\3\2\2\2d\u02e3\3\2\2\2f\u02e8\3\2\2\2")
 *         buf.write("h\u02f6\3\2\2\2j\u0314\3\2\2\2l\u0332\3\2\2\2n\u034b\3")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 123, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__190, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 123, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":124
 *         buf.write("\3\2\2\2T\u024b\3\2\2\2V\u0268\3\2\2\2X\u026a\3\2\2\2")
 *         buf.write("Z\u026d\3\2\2\2\\\u0296\3\2\2\2^\u02bf\3\2\2\2`\u02cd")
 *         buf.write("\3\2\2\2b\u02db\3\2\2\2d\u02e3\3\2\2\2f\u02e8\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("h\u02f6\3\2\2\2j\u0314\3\2\2\2l\u0332\3\2\2\2n\u034b\3")
 *         buf.write("\2\2\2p\u034d\3\2\2\2r\u0360\3\2\2\2t\u0396\3\2\2\2v\u0399")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 124, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__191, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 124, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":125
 *         buf.write("Z\u026d\3\2\2\2\\\u0296\3\2\2\2^\u02bf\3\2\2\2`\u02cd")
 *         buf.write("\3\2\2\2b\u02db\3\2\2\2d\u02e3\3\2\2\2f\u02e8\3\2\2\2")
 *         buf.write("h\u02f6\3\2\2\2j\u0314\3\2\2\2l\u0332\3\2\2\2n\u034b\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2p\u034d\3\2\2\2r\u0360\3\2\2\2t\u0396\3\2\2\2v\u0399")
 *         buf.write("\3\2\2\2x\u03a3\3\2\2\2z\u03a7\3\2\2\2|\u03af\3\2\2\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 125, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__192, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 125, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":126
 *         buf.write("\3\2\2\2b\u02db\3\2\2\2d\u02e3\3\2\2\2f\u02e8\3\2\2\2")
 *         buf.write("h\u02f6\3\2\2\2j\u0314\3\2\2\2l\u0332\3\2\2\2n\u034b\3")
 *         buf.write("\2\2\2p\u034d\3\2\2\2r\u0360\3\2\2\2t\u0396\3\2\2\2v\u0399")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2x\u03a3\3\2\2\2z\u03a7\3\2\2\2|\u03af\3\2\2\2")
 *         buf.write("~\u03c0\3\2\2\2\u0080\u03c2\3\2\2\2\u0082\u03ca\3\2\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 126, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__193, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 126, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":127
 *         buf.write("h\u02f6\3\2\2\2j\u0314\3\2\2\2l\u0332\3\2\2\2n\u034b\3")
 *         buf.write("\2\2\2p\u034d\3\2\2\2r\u0360\3\2\2\2t\u0396\3\2\2\2v\u0399")
 *         buf.write("\3\2\2\2x\u03a3\3\2\2\2z\u03a7\3\2\2\2|\u03af\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("~\u03c0\3\2\2\2\u0080\u03c2\3\2\2\2\u0082\u03ca\3\2\2")
 *         buf.write("\2\u0084\u03d7\3\2\2\2\u0086\u03da\3\2\2\2\u0088\u03e3")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 127, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__194, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 127, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":128
 *         buf.write("\2\2\2p\u034d\3\2\2\2r\u0360\3\2\2\2t\u0396\3\2\2\2v\u0399")
 *         buf.write("\3\2\2\2x\u03a3\3\2\2\2z\u03a7\3\2\2\2|\u03af\3\2\2\2")
 *         buf.write("~\u03c0\3\2\2\2\u0080\u03c2\3\2\2\2\u0082\u03ca\3\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u0084\u03d7\3\2\2\2\u0086\u03da\3\2\2\2\u0088\u03e3")
 *         buf.write("\3\2\2\2\u008a\u03eb\3\2\2\2\u008c\u03f8\3\2\2\2\u008e")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 128, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__196, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 128, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":129
 *         buf.write("\3\2\2\2x\u03a3\3\2\2\2z\u03a7\3\2\2\2|\u03af\3\2\2\2")
 *         buf.write("~\u03c0\3\2\2\2\u0080\u03c2\3\2\2\2\u0082\u03ca\3\2\2")
 *         buf.write("\2\u0084\u03d7\3\2\2\2\u0086\u03da\3\2\2\2\u0088\u03e3")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u008a\u03eb\3\2\2\2\u008c\u03f8\3\2\2\2\u008e")
 *         buf.write("\u03fa\3\2\2\2\u0090\u0406\3\2\2\2\u0092\u0408\3\2\2\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 129, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__198, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 129, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":130
 *         buf.write("~\u03c0\3\2\2\2\u0080\u03c2\3\2\2\2\u0082\u03ca\3\2\2")
 *         buf.write("\2\u0084\u03d7\3\2\2\2\u0086\u03da\3\2\2\2\u0088\u03e3")
 *         buf.write("\3\2\2\2\u008a\u03eb\3\2\2\2\u008c\u03f8\3\2\2\2\u008e")             # <<<<<<<<<<<<<<
 *         buf.write("\u03fa\3\2\2\2\u0090\u0406\3\2\2\2\u0092\u0408\3\2\2\2")
 *         buf.write("\u0094\u0096\7T\2\2\u0095\u0097\5z>\2\u0096\u0095\3\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 130, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__200, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 130, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":131
 *         buf.write("\2\u0084\u03d7\3\2\2\2\u0086\u03da\3\2\2\2\u0088\u03e3")
 *         buf.write("\3\2\2\2\u008a\u03eb\3\2\2\2\u008c\u03f8\3\2\2\2\u008e")
 *         buf.write("\u03fa\3\2\2\2\u0090\u0406\3\2\2\2\u0092\u0408\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0094\u0096\7T\2\2\u0095\u0097\5z>\2\u0096\u0095\3\2")
 *         buf.write("\2\2\u0096\u0097\3\2\2\2\u0097\u0098\3\2\2\2\u0098\u009a")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 131, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__202, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 131, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":132
 *         buf.write("\3\2\2\2\u008a\u03eb\3\2\2\2\u008c\u03f8\3\2\2\2\u008e")
 *         buf.write("\u03fa\3\2\2\2\u0090\u0406\3\2\2\2\u0092\u0408\3\2\2\2")
 *         buf.write("\u0094\u0096\7T\2\2\u0095\u0097\5z>\2\u0096\u0095\3\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\u0096\u0097\3\2\2\2\u0097\u0098\3\2\2\2\u0098\u009a")
 *         buf.write("\7\3\2\2\u0099\u0094\3\2\2\2\u0099\u009a\3\2\2\2\u009a")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 132, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__203, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 132, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":133
 *         buf.write("\u03fa\3\2\2\2\u0090\u0406\3\2\2\2\u0092\u0408\3\2\2\2")
 *         buf.write("\u0094\u0096\7T\2\2\u0095\u0097\5z>\2\u0096\u0095\3\2")
 *         buf.write("\2\2\u0096\u0097\3\2\2\2\u0097\u0098\3\2\2\2\u0098\u009a")             # <<<<<<<<<<<<<<
 *         buf.write("\7\3\2\2\u0099\u0094\3\2\2\2\u0099\u009a\3\2\2\2\u009a")
 *         buf.write("\u009e\3\2\2\2\u009b\u009d\5\4\3\2\u009c\u009b\3\2\2\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 133, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__205, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 133, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":134
 *         buf.write("\u0094\u0096\7T\2\2\u0095\u0097\5z>\2\u0096\u0095\3\2")
 *         buf.write("\2\2\u0096\u0097\3\2\2\2\u0097\u0098\3\2\2\2\u0098\u009a")
 *         buf.write("\7\3\2\2\u0099\u0094\3\2\2\2\u0099\u009a\3\2\2\2\u009a")             # <<<<<<<<<<<<<<
 *         buf.write("\u009e\3\2\2\2\u009b\u009d\5\4\3\2\u009c\u009b\3\2\2\2")
 *         buf.write("\u009d\u00a0\3\2\2\2\u009e\u009c\3\2\2\2\u009e\u009f\3")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 134, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__207, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 134, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":135
 *         buf.write("\2\2\u0096\u0097\3\2\2\2\u0097\u0098\3\2\2\2\u0098\u009a")
 *         buf.write("\7\3\2\2\u0099\u0094\3\2\2\2\u0099\u009a\3\2\2\2\u009a")
 *         buf.write("\u009e\3\2\2\2\u009b\u009d\5\4\3\2\u009c\u009b\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u009d\u00a0\3\2\2\2\u009e\u009c\3\2\2\2\u009e\u009f\3")
 *         buf.write("\2\2\2\u009f\3\3\2\2\2\u00a0\u009e\3\2\2\2\u00a1\u00a3")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 135, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__209, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 135, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":136
 *         buf.write("\7\3\2\2\u0099\u0094\3\2\2\2\u0099\u009a\3\2\2\2\u009a")
 *         buf.write("\u009e\3\2\2\2\u009b\u009d\5\4\3\2\u009c\u009b\3\2\2\2")
 *         buf.write("\u009d\u00a0\3\2\2\2\u009e\u009c\3\2\2\2\u009e\u009f\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u009f\3\3\2\2\2\u00a0\u009e\3\2\2\2\u00a1\u00a3")
 *         buf.write("\7S\2\2\u00a2\u00a1\3\2\2\2\u00a2\u00a3\3\2\2\2\u00a3")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 136, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__211, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 136, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":137
 *         buf.write("\u009e\3\2\2\2\u009b\u009d\5\4\3\2\u009c\u009b\3\2\2\2")
 *         buf.write("\u009d\u00a0\3\2\2\2\u009e\u009c\3\2\2\2\u009e\u009f\3")
 *         buf.write("\2\2\2\u009f\3\3\2\2\2\u00a0\u009e\3\2\2\2\u00a1\u00a3")             # <<<<<<<<<<<<<<
 *         buf.write("\7S\2\2\u00a2\u00a1\3\2\2\2\u00a2\u00a3\3\2\2\2\u00a3")
 *         buf.write("\u00a4\3\2\2\2\u00a4\u00a5\5\6\4\2\u00a5\u00a6\7\3\2\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 137, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__213, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 137, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":138
 *         buf.write("\u009d\u00a0\3\2\2\2\u009e\u009c\3\2\2\2\u009e\u009f\3")
 *         buf.write("\2\2\2\u009f\3\3\2\2\2\u00a0\u009e\3\2\2\2\u00a1\u00a3")
 *         buf.write("\7S\2\2\u00a2\u00a1\3\2\2\2\u00a2\u00a3\3\2\2\2\u00a3")             # <<<<<<<<<<<<<<
 *         buf.write("\u00a4\3\2\2\2\u00a4\u00a5\5\6\4\2\u00a5\u00a6\7\3\2\2")
 *         buf.write("\u00a6\5\3\2\2\2\u00a7\u00a9\7U\2\2\u00a8\u00a7\3\2\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 138, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__214, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 138, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":139
 *         buf.write("\2\2\2\u009f\3\3\2\2\2\u00a0\u009e\3\2\2\2\u00a1\u00a3")
 *         buf.write("\7S\2\2\u00a2\u00a1\3\2\2\2\u00a2\u00a3\3\2\2\2\u00a3")
 *         buf.write("\u00a4\3\2\2\2\u00a4\u00a5\5\6\4\2\u00a5\u00a6\7\3\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u00a6\5\3\2\2\2\u00a7\u00a9\7U\2\2\u00a8\u00a7\3\2\2")
 *         buf.write("\2\u00a8\u00a9\3\2\2\2\u00a9\u00aa\3\2\2\2\u00aa\u00ab")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 139, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__216, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 139, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":140
 *         buf.write("\7S\2\2\u00a2\u00a1\3\2\2\2\u00a2\u00a3\3\2\2\2\u00a3")
 *         buf.write("\u00a4\3\2\2\2\u00a4\u00a5\5\6\4\2\u00a5\u00a6\7\3\2\2")
 *         buf.write("\u00a6\5\3\2\2\2\u00a7\u00a9\7U\2\2\u00a8\u00a7\3\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u00a8\u00a9\3\2\2\2\u00a9\u00aa\3\2\2\2\u00aa\u00ab")
 *         buf.write("\5\b\5\2\u00ab\u00ac\5\f\7\2\u00ac\7\3\2\2\2\u00ad\u00af")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 140, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__217, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 140, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":141
 *         buf.write("\u00a4\3\2\2\2\u00a4\u00a5\5\6\4\2\u00a5\u00a6\7\3\2\2")
 *         buf.write("\u00a6\5\3\2\2\2\u00a7\u00a9\7U\2\2\u00a8\u00a7\3\2\2")
 *         buf.write("\2\u00a8\u00a9\3\2\2\2\u00a9\u00aa\3\2\2\2\u00aa\u00ab")             # <<<<<<<<<<<<<<
 *         buf.write("\5\b\5\2\u00ab\u00ac\5\f\7\2\u00ac\7\3\2\2\2\u00ad\u00af")
 *         buf.write("\7R\2\2\u00ae\u00ad\3\2\2\2\u00ae\u00af\3\2\2\2\u00af")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 141, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__219, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 141, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":142
 *         buf.write("\u00a6\5\3\2\2\2\u00a7\u00a9\7U\2\2\u00a8\u00a7\3\2\2")
 *         buf.write("\2\u00a8\u00a9\3\2\2\2\u00a9\u00aa\3\2\2\2\u00aa\u00ab")
 *         buf.write("\5\b\5\2\u00ab\u00ac\5\f\7\2\u00ac\7\3\2\2\2\u00ad\u00af")             # <<<<<<<<<<<<<<
 *         buf.write("\7R\2\2\u00ae\u00ad\3\2\2\2\u00ae\u00af\3\2\2\2\u00af")
 *         buf.write("\u00b0\3\2\2\2\u00b0\u00b1\5\n\6\2\u00b1\t\3\2\2\2\u00b2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 142, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__221, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 142, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":143
 *         buf.write("\2\u00a8\u00a9\3\2\2\2\u00a9\u00aa\3\2\2\2\u00aa\u00ab")
 *         buf.write("\5\b\5\2\u00ab\u00ac\5\f\7\2\u00ac\7\3\2\2\2\u00ad\u00af")
 *         buf.write("\7R\2\2\u00ae\u00ad\3\2\2\2\u00ae\u00af\3\2\2\2\u00af")             # <<<<<<<<<<<<<<
 *         buf.write("\u00b0\3\2\2\2\u00b0\u00b1\5\n\6\2\u00b1\t\3\2\2\2\u00b2")
 *         buf.write("\u00c8\7\4\2\2\u00b3\u00c8\7\5\2\2\u00b4\u00b6\7\6\2\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 143, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__222, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 143, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":144
 *         buf.write("\5\b\5\2\u00ab\u00ac\5\f\7\2\u00ac\7\3\2\2\2\u00ad\u00af")
 *         buf.write("\7R\2\2\u00ae\u00ad\3\2\2\2\u00ae\u00af\3\2\2\2\u00af")
 *         buf.write("\u00b0\3\2\2\2\u00b0\u00b1\5\n\6\2\u00b1\t\3\2\2\2\u00b2")             # <<<<<<<<<<<<<<
 *         buf.write("\u00c8\7\4\2\2\u00b3\u00c8\7\5\2\2\u00b4\u00b6\7\6\2\2")
 *         buf.write("\u00b5\u00b4\3\2\2\2\u00b5\u00b6\3\2\2\2\u00b6\u00b7\3")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 144, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__224, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 144, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":145
 *         buf.write("\7R\2\2\u00ae\u00ad\3\2\2\2\u00ae\u00af\3\2\2\2\u00af")
 *         buf.write("\u00b0\3\2\2\2\u00b0\u00b1\5\n\6\2\u00b1\t\3\2\2\2\u00b2")
 *         buf.write("\u00c8\7\4\2\2\u00b3\u00c8\7\5\2\2\u00b4\u00b6\7\6\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u00b5\u00b4\3\2\2\2\u00b5\u00b6\3\2\2\2\u00b6\u00b7\3")
 *         buf.write("\2\2\2\u00b7\u00c8\7\7\2\2\u00b8\u00c8\7\b\2\2\u00b9\u00bb")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 145, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__226, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 145, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":146
 *         buf.write("\u00b0\3\2\2\2\u00b0\u00b1\5\n\6\2\u00b1\t\3\2\2\2\u00b2")
 *         buf.write("\u00c8\7\4\2\2\u00b3\u00c8\7\5\2\2\u00b4\u00b6\7\6\2\2")
 *         buf.write("\u00b5\u00b4\3\2\2\2\u00b5\u00b6\3\2\2\2\u00b6\u00b7\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u00b7\u00c8\7\7\2\2\u00b8\u00c8\7\b\2\2\u00b9\u00bb")
 *         buf.write("\7\t\2\2\u00ba\u00b9\3\2\2\2\u00ba\u00bb\3\2\2\2\u00bb")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 146, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__228, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 146, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":147
 *         buf.write("\u00c8\7\4\2\2\u00b3\u00c8\7\5\2\2\u00b4\u00b6\7\6\2\2")
 *         buf.write("\u00b5\u00b4\3\2\2\2\u00b5\u00b6\3\2\2\2\u00b6\u00b7\3")
 *         buf.write("\2\2\2\u00b7\u00c8\7\7\2\2\u00b8\u00c8\7\b\2\2\u00b9\u00bb")             # <<<<<<<<<<<<<<
 *         buf.write("\7\t\2\2\u00ba\u00b9\3\2\2\2\u00ba\u00bb\3\2\2\2\u00bb")
 *         buf.write("\u00bc\3\2\2\2\u00bc\u00c8\7\n\2\2\u00bd\u00c8\7\13\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 147, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__230, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 147, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":148
 *         buf.write("\u00b5\u00b4\3\2\2\2\u00b5\u00b6\3\2\2\2\u00b6\u00b7\3")
 *         buf.write("\2\2\2\u00b7\u00c8\7\7\2\2\u00b8\u00c8\7\b\2\2\u00b9\u00bb")
 *         buf.write("\7\t\2\2\u00ba\u00b9\3\2\2\2\u00ba\u00bb\3\2\2\2\u00bb")             # <<<<<<<<<<<<<<
 *         buf.write("\u00bc\3\2\2\2\u00bc\u00c8\7\n\2\2\u00bd\u00c8\7\13\2")
 *         buf.write("\2\u00be\u00c8\7\f\2\2\u00bf\u00c1\t\2\2\2\u00c0\u00bf")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 148, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__232, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 148, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":149
 *         buf.write("\2\2\2\u00b7\u00c8\7\7\2\2\u00b8\u00c8\7\b\2\2\u00b9\u00bb")
 *         buf.write("\7\t\2\2\u00ba\u00b9\3\2\2\2\u00ba\u00bb\3\2\2\2\u00bb")
 *         buf.write("\u00bc\3\2\2\2\u00bc\u00c8\7\n\2\2\u00bd\u00c8\7\13\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u00be\u00c8\7\f\2\2\u00bf\u00c1\t\2\2\2\u00c0\u00bf")
 *         buf.write("\3\2\2\2\u00c0\u00c1\3\2\2\2\u00c1\u00c3\3\2\2\2\u00c2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 149, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__234, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 149, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":150
 *         buf.write("\7\t\2\2\u00ba\u00b9\3\2\2\2\u00ba\u00bb\3\2\2\2\u00bb")
 *         buf.write("\u00bc\3\2\2\2\u00bc\u00c8\7\n\2\2\u00bd\u00c8\7\13\2")
 *         buf.write("\2\u00be\u00c8\7\f\2\2\u00bf\u00c1\t\2\2\2\u00c0\u00bf")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u00c0\u00c1\3\2\2\2\u00c1\u00c3\3\2\2\2\u00c2")
 *         buf.write("\u00c4\7\6\2\2\u00c3\u00c2\3\2\2\2\u00c3\u00c4\3\2\2\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 150, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__236, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 150, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":151
 *         buf.write("\u00bc\3\2\2\2\u00bc\u00c8\7\n\2\2\u00bd\u00c8\7\13\2")
 *         buf.write("\2\u00be\u00c8\7\f\2\2\u00bf\u00c1\t\2\2\2\u00c0\u00bf")
 *         buf.write("\3\2\2\2\u00c0\u00c1\3\2\2\2\u00c1\u00c3\3\2\2\2\u00c2")             # <<<<<<<<<<<<<<
 *         buf.write("\u00c4\7\6\2\2\u00c3\u00c2\3\2\2\2\u00c3\u00c4\3\2\2\2")
 *         buf.write("\u00c4\u00c5\3\2\2\2\u00c5\u00c8\7\17\2\2\u00c6\u00c8")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 151, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__238, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 151, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":152
 *         buf.write("\2\u00be\u00c8\7\f\2\2\u00bf\u00c1\t\2\2\2\u00c0\u00bf")
 *         buf.write("\3\2\2\2\u00c0\u00c1\3\2\2\2\u00c1\u00c3\3\2\2\2\u00c2")
 *         buf.write("\u00c4\7\6\2\2\u00c3\u00c2\3\2\2\2\u00c3\u00c4\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u00c4\u00c5\3\2\2\2\u00c5\u00c8\7\17\2\2\u00c6\u00c8")
 *         buf.write("\7\6\2\2\u00c7\u00b2\3\2\2\2\u00c7\u00b3\3\2\2\2\u00c7")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 152, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__240, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 152, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":153
 *         buf.write("\3\2\2\2\u00c0\u00c1\3\2\2\2\u00c1\u00c3\3\2\2\2\u00c2")
 *         buf.write("\u00c4\7\6\2\2\u00c3\u00c2\3\2\2\2\u00c3\u00c4\3\2\2\2")
 *         buf.write("\u00c4\u00c5\3\2\2\2\u00c5\u00c8\7\17\2\2\u00c6\u00c8")             # <<<<<<<<<<<<<<
 *         buf.write("\7\6\2\2\u00c7\u00b2\3\2\2\2\u00c7\u00b3\3\2\2\2\u00c7")
 *         buf.write("\u00b5\3\2\2\2\u00c7\u00b8\3\2\2\2\u00c7\u00ba\3\2\2\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 153, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__242, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 153, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":154
 *         buf.write("\u00c4\7\6\2\2\u00c3\u00c2\3\2\2\2\u00c3\u00c4\3\2\2\2")
 *         buf.write("\u00c4\u00c5\3\2\2\2\u00c5\u00c8\7\17\2\2\u00c6\u00c8")
 *         buf.write("\7\6\2\2\u00c7\u00b2\3\2\2\2\u00c7\u00b3\3\2\2\2\u00c7")             # <<<<<<<<<<<<<<
 *         buf.write("\u00b5\3\2\2\2\u00c7\u00b8\3\2\2\2\u00c7\u00ba\3\2\2\2")
 *         buf.write("\u00c7\u00bd\3\2\2\2\u00c7\u00be\3\2\2\2\u00c7\u00c0\3")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 154, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__244, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 154, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":155
 *         buf.write("\u00c4\u00c5\3\2\2\2\u00c5\u00c8\7\17\2\2\u00c6\u00c8")
 *         buf.write("\7\6\2\2\u00c7\u00b2\3\2\2\2\u00c7\u00b3\3\2\2\2\u00c7")
 *         buf.write("\u00b5\3\2\2\2\u00c7\u00b8\3\2\2\2\u00c7\u00ba\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u00c7\u00bd\3\2\2\2\u00c7\u00be\3\2\2\2\u00c7\u00c0\3")
 *         buf.write("\2\2\2\u00c7\u00c6\3\2\2\2\u00c8\13\3\2\2\2\u00c9\u00ca")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 155, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__246, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 155, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":156
 *         buf.write("\7\6\2\2\u00c7\u00b2\3\2\2\2\u00c7\u00b3\3\2\2\2\u00c7")
 *         buf.write("\u00b5\3\2\2\2\u00c7\u00b8\3\2\2\2\u00c7\u00ba\3\2\2\2")
 *         buf.write("\u00c7\u00bd\3\2\2\2\u00c7\u00be\3\2\2\2\u00c7\u00c0\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u00c7\u00c6\3\2\2\2\u00c8\13\3\2\2\2\u00c9\u00ca")
 *         buf.write("\7Z\2\2\u00ca\u00cb\5\u0090I\2\u00cb\u00cc\5\24\13\2\u00cc")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 156, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__248, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 156, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":157
 *         buf.write("\u00b5\3\2\2\2\u00c7\u00b8\3\2\2\2\u00c7\u00ba\3\2\2\2")
 *         buf.write("\u00c7\u00bd\3\2\2\2\u00c7\u00be\3\2\2\2\u00c7\u00c0\3")
 *         buf.write("\2\2\2\u00c7\u00c6\3\2\2\2\u00c8\13\3\2\2\2\u00c9\u00ca")             # <<<<<<<<<<<<<<
 *         buf.write("\7Z\2\2\u00ca\u00cb\5\u0090I\2\u00cb\u00cc\5\24\13\2\u00cc")
 *         buf.write("\u00cd\7\20\2\2\u00cd\u00ce\7Z\2\2\u00ce\u0103\3\2\2\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 157, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__250, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 157, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":158
 *         buf.write("\u00c7\u00bd\3\2\2\2\u00c7\u00be\3\2\2\2\u00c7\u00c0\3")
 *         buf.write("\2\2\2\u00c7\u00c6\3\2\2\2\u00c8\13\3\2\2\2\u00c9\u00ca")
 *         buf.write("\7Z\2\2\u00ca\u00cb\5\u0090I\2\u00cb\u00cc\5\24\13\2\u00cc")             # <<<<<<<<<<<<<<
 *         buf.write("\u00cd\7\20\2\2\u00cd\u00ce\7Z\2\2\u00ce\u0103\3\2\2\2")
 *         buf.write("\u00cf\u00d0\7Z\2\2\u00d0\u00d1\7\21\2\2\u00d1\u00d2\5")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 158, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__251, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 158, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":159
 *         buf.write("\2\2\2\u00c7\u00c6\3\2\2\2\u00c8\13\3\2\2\2\u00c9\u00ca")
 *         buf.write("\7Z\2\2\u00ca\u00cb\5\u0090I\2\u00cb\u00cc\5\24\13\2\u00cc")
 *         buf.write("\u00cd\7\20\2\2\u00cd\u00ce\7Z\2\2\u00ce\u0103\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u00cf\u00d0\7Z\2\2\u00d0\u00d1\7\21\2\2\u00d1\u00d2\5")
 *         buf.write("\16\b\2\u00d2\u00d4\5v<\2\u00d3\u00d5\5\u008aF\2\u00d4")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 159, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__252, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 159, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":160
 *         buf.write("\7Z\2\2\u00ca\u00cb\5\u0090I\2\u00cb\u00cc\5\24\13\2\u00cc")
 *         buf.write("\u00cd\7\20\2\2\u00cd\u00ce\7Z\2\2\u00ce\u0103\3\2\2\2")
 *         buf.write("\u00cf\u00d0\7Z\2\2\u00d0\u00d1\7\21\2\2\u00d1\u00d2\5")             # <<<<<<<<<<<<<<
 *         buf.write("\16\b\2\u00d2\u00d4\5v<\2\u00d3\u00d5\5\u008aF\2\u00d4")
 *         buf.write("\u00d3\3\2\2\2\u00d4\u00d5\3\2\2\2\u00d5\u00d7\3\2\2\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 160, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__253, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 160, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":161
 *         buf.write("\u00cd\7\20\2\2\u00cd\u00ce\7Z\2\2\u00ce\u0103\3\2\2\2")
 *         buf.write("\u00cf\u00d0\7Z\2\2\u00d0\u00d1\7\21\2\2\u00d1\u00d2\5")
 *         buf.write("\16\b\2\u00d2\u00d4\5v<\2\u00d3\u00d5\5\u008aF\2\u00d4")             # <<<<<<<<<<<<<<
 *         buf.write("\u00d3\3\2\2\2\u00d4\u00d5\3\2\2\2\u00d5\u00d7\3\2\2\2")
 *         buf.write("\u00d6\u00d8\5:\36\2\u00d7\u00d6\3\2\2\2\u00d7\u00d8\3")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 161, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__254, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 161, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":162
 *         buf.write("\u00cf\u00d0\7Z\2\2\u00d0\u00d1\7\21\2\2\u00d1\u00d2\5")
 *         buf.write("\16\b\2\u00d2\u00d4\5v<\2\u00d3\u00d5\5\u008aF\2\u00d4")
 *         buf.write("\u00d3\3\2\2\2\u00d4\u00d5\3\2\2\2\u00d5\u00d7\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u00d6\u00d8\5:\36\2\u00d7\u00d6\3\2\2\2\u00d7\u00d8\3")
 *         buf.write("\2\2\2\u00d8\u00d9\3\2\2\2\u00d9\u00da\5\u008eH\2\u00da")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 162, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__256, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 162, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":163
 *         buf.write("\16\b\2\u00d2\u00d4\5v<\2\u00d3\u00d5\5\u008aF\2\u00d4")
 *         buf.write("\u00d3\3\2\2\2\u00d4\u00d5\3\2\2\2\u00d5\u00d7\3\2\2\2")
 *         buf.write("\u00d6\u00d8\5:\36\2\u00d7\u00d6\3\2\2\2\u00d7\u00d8\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u00d8\u00d9\3\2\2\2\u00d9\u00da\5\u008eH\2\u00da")
 *         buf.write("\u0103\3\2\2\2\u00db\u00dc\7Z\2\2\u00dc\u00dd\7\21\2\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 163, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__258, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 163, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":164
 *         buf.write("\u00d3\3\2\2\2\u00d4\u00d5\3\2\2\2\u00d5\u00d7\3\2\2\2")
 *         buf.write("\u00d6\u00d8\5:\36\2\u00d7\u00d6\3\2\2\2\u00d7\u00d8\3")
 *         buf.write("\2\2\2\u00d8\u00d9\3\2\2\2\u00d9\u00da\5\u008eH\2\u00da")             # <<<<<<<<<<<<<<
 *         buf.write("\u0103\3\2\2\2\u00db\u00dc\7Z\2\2\u00dc\u00dd\7\21\2\2")
 *         buf.write("\u00dd\u00de\7\22\2\2\u00de\u00e3\7\23\2\2\u00df\u00e1")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 164, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__259, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 164, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":165
 *         buf.write("\u00d6\u00d8\5:\36\2\u00d7\u00d6\3\2\2\2\u00d7\u00d8\3")
 *         buf.write("\2\2\2\u00d8\u00d9\3\2\2\2\u00d9\u00da\5\u008eH\2\u00da")
 *         buf.write("\u0103\3\2\2\2\u00db\u00dc\7Z\2\2\u00dc\u00dd\7\21\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u00dd\u00de\7\22\2\2\u00de\u00e3\7\23\2\2\u00df\u00e1")
 *         buf.write("\5\20\t\2\u00e0\u00df\3\2\2\2\u00e0\u00e1\3\2\2\2\u00e1")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 165, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__260, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 165, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":166
 *         buf.write("\2\2\2\u00d8\u00d9\3\2\2\2\u00d9\u00da\5\u008eH\2\u00da")
 *         buf.write("\u0103\3\2\2\2\u00db\u00dc\7Z\2\2\u00dc\u00dd\7\21\2\2")
 *         buf.write("\u00dd\u00de\7\22\2\2\u00de\u00e3\7\23\2\2\u00df\u00e1")             # <<<<<<<<<<<<<<
 *         buf.write("\5\20\t\2\u00e0\u00df\3\2\2\2\u00e0\u00e1\3\2\2\2\u00e1")
 *         buf.write("\u00e4\3\2\2\2\u00e2\u00e4\7\24\2\2\u00e3\u00e0\3\2\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 166, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__262, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 166, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":167
 *         buf.write("\u0103\3\2\2\2\u00db\u00dc\7Z\2\2\u00dc\u00dd\7\21\2\2")
 *         buf.write("\u00dd\u00de\7\22\2\2\u00de\u00e3\7\23\2\2\u00df\u00e1")
 *         buf.write("\5\20\t\2\u00e0\u00df\3\2\2\2\u00e0\u00e1\3\2\2\2\u00e1")             # <<<<<<<<<<<<<<
 *         buf.write("\u00e4\3\2\2\2\u00e2\u00e4\7\24\2\2\u00e3\u00e0\3\2\2")
 *         buf.write("\2\u00e3\u00e2\3\2\2\2\u00e4\u00e5\3\2\2\2\u00e5\u00e6")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 167, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__264, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 167, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":168
 *         buf.write("\u00dd\u00de\7\22\2\2\u00de\u00e3\7\23\2\2\u00df\u00e1")
 *         buf.write("\5\20\t\2\u00e0\u00df\3\2\2\2\u00e0\u00e1\3\2\2\2\u00e1")
 *         buf.write("\u00e4\3\2\2\2\u00e2\u00e4\7\24\2\2\u00e3\u00e0\3\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u00e3\u00e2\3\2\2\2\u00e4\u00e5\3\2\2\2\u00e5\u00e6")
 *         buf.write("\7\25\2\2\u00e6\u0103\5\u008eH\2\u00e7\u00e8\7Z\2\2\u00e8")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 168, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__266, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 168, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":169
 *         buf.write("\5\20\t\2\u00e0\u00df\3\2\2\2\u00e0\u00e1\3\2\2\2\u00e1")
 *         buf.write("\u00e4\3\2\2\2\u00e2\u00e4\7\24\2\2\u00e3\u00e0\3\2\2")
 *         buf.write("\2\u00e3\u00e2\3\2\2\2\u00e4\u00e5\3\2\2\2\u00e5\u00e6")             # <<<<<<<<<<<<<<
 *         buf.write("\7\25\2\2\u00e6\u0103\5\u008eH\2\u00e7\u00e8\7Z\2\2\u00e8")
 *         buf.write("\u00e9\7\21\2\2\u00e9\u00ea\7\26\2\2\u00ea\u00eb\7\23")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 169, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__268, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 169, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":170
 *         buf.write("\u00e4\3\2\2\2\u00e2\u00e4\7\24\2\2\u00e3\u00e0\3\2\2")
 *         buf.write("\2\u00e3\u00e2\3\2\2\2\u00e4\u00e5\3\2\2\2\u00e5\u00e6")
 *         buf.write("\7\25\2\2\u00e6\u0103\5\u008eH\2\u00e7\u00e8\7Z\2\2\u00e8")             # <<<<<<<<<<<<<<
 *         buf.write("\u00e9\7\21\2\2\u00e9\u00ea\7\26\2\2\u00ea\u00eb\7\23")
 *         buf.write("\2\2\u00eb\u00ec\5v<\2\u00ec\u00ed\7\27\2\2\u00ed\u00f2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 170, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__269, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 170, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":171
 *         buf.write("\2\u00e3\u00e2\3\2\2\2\u00e4\u00e5\3\2\2\2\u00e5\u00e6")
 *         buf.write("\7\25\2\2\u00e6\u0103\5\u008eH\2\u00e7\u00e8\7Z\2\2\u00e8")
 *         buf.write("\u00e9\7\21\2\2\u00e9\u00ea\7\26\2\2\u00ea\u00eb\7\23")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\u00eb\u00ec\5v<\2\u00ec\u00ed\7\27\2\2\u00ed\u00f2")
 *         buf.write("\7Z\2\2\u00ee\u00ef\7\27\2\2\u00ef\u00f1\7Z\2\2\u00f0")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 171, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__271, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 171, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":172
 *         buf.write("\7\25\2\2\u00e6\u0103\5\u008eH\2\u00e7\u00e8\7Z\2\2\u00e8")
 *         buf.write("\u00e9\7\21\2\2\u00e9\u00ea\7\26\2\2\u00ea\u00eb\7\23")
 *         buf.write("\2\2\u00eb\u00ec\5v<\2\u00ec\u00ed\7\27\2\2\u00ed\u00f2")             # <<<<<<<<<<<<<<
 *         buf.write("\7Z\2\2\u00ee\u00ef\7\27\2\2\u00ef\u00f1\7Z\2\2\u00f0")
 *         buf.write("\u00ee\3\2\2\2\u00f1\u00f4\3\2\2\2\u00f2\u00f0\3\2\2\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 172, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__272, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 172, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":173
 *         buf.write("\u00e9\7\21\2\2\u00e9\u00ea\7\26\2\2\u00ea\u00eb\7\23")
 *         buf.write("\2\2\u00eb\u00ec\5v<\2\u00ec\u00ed\7\27\2\2\u00ed\u00f2")
 *         buf.write("\7Z\2\2\u00ee\u00ef\7\27\2\2\u00ef\u00f1\7Z\2\2\u00f0")             # <<<<<<<<<<<<<<
 *         buf.write("\u00ee\3\2\2\2\u00f1\u00f4\3\2\2\2\u00f2\u00f0\3\2\2\2")
 *         buf.write("\u00f2\u00f3\3\2\2\2\u00f3\u00f5\3\2\2\2\u00f4\u00f2\3")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 173, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__273, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 173, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":174
 *         buf.write("\2\2\u00eb\u00ec\5v<\2\u00ec\u00ed\7\27\2\2\u00ed\u00f2")
 *         buf.write("\7Z\2\2\u00ee\u00ef\7\27\2\2\u00ef\u00f1\7Z\2\2\u00f0")
 *         buf.write("\u00ee\3\2\2\2\u00f1\u00f4\3\2\2\2\u00f2\u00f0\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u00f2\u00f3\3\2\2\2\u00f3\u00f5\3\2\2\2\u00f4\u00f2\3")
 *         buf.write("\2\2\2\u00f5\u00f6\7\25\2\2\u00f6\u00f7\5\u008eH\2\u00f7")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 174, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__275, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 174, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":175
 *         buf.write("\7Z\2\2\u00ee\u00ef\7\27\2\2\u00ef\u00f1\7Z\2\2\u00f0")
 *         buf.write("\u00ee\3\2\2\2\u00f1\u00f4\3\2\2\2\u00f2\u00f0\3\2\2\2")
 *         buf.write("\u00f2\u00f3\3\2\2\2\u00f3\u00f5\3\2\2\2\u00f4\u00f2\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u00f5\u00f6\7\25\2\2\u00f6\u00f7\5\u008eH\2\u00f7")
 *         buf.write("\u0103\3\2\2\2\u00f8\u00f9\7\30\2\2\u00f9\u00fb\7Z\2\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 175, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__277, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 175, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":176
 *         buf.write("\u00ee\3\2\2\2\u00f1\u00f4\3\2\2\2\u00f2\u00f0\3\2\2\2")
 *         buf.write("\u00f2\u00f3\3\2\2\2\u00f3\u00f5\3\2\2\2\u00f4\u00f2\3")
 *         buf.write("\2\2\2\u00f5\u00f6\7\25\2\2\u00f6\u00f7\5\u008eH\2\u00f7")             # <<<<<<<<<<<<<<
 *         buf.write("\u0103\3\2\2\2\u00f8\u00f9\7\30\2\2\u00f9\u00fb\7Z\2\2")
 *         buf.write("\u00fa\u00fc\5:\36\2\u00fb\u00fa\3\2\2\2\u00fb\u00fc\3")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 176, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__278, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 176, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":177
 *         buf.write("\u00f2\u00f3\3\2\2\2\u00f3\u00f5\3\2\2\2\u00f4\u00f2\3")
 *         buf.write("\2\2\2\u00f5\u00f6\7\25\2\2\u00f6\u00f7\5\u008eH\2\u00f7")
 *         buf.write("\u0103\3\2\2\2\u00f8\u00f9\7\30\2\2\u00f9\u00fb\7Z\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u00fa\u00fc\5:\36\2\u00fb\u00fa\3\2\2\2\u00fb\u00fc\3")
 *         buf.write("\2\2\2\u00fc\u00fd\3\2\2\2\u00fd\u00fe\5\u0090I\2\u00fe")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 177, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__279, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 177, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":178
 *         buf.write("\2\2\2\u00f5\u00f6\7\25\2\2\u00f6\u00f7\5\u008eH\2\u00f7")
 *         buf.write("\u0103\3\2\2\2\u00f8\u00f9\7\30\2\2\u00f9\u00fb\7Z\2\2")
 *         buf.write("\u00fa\u00fc\5:\36\2\u00fb\u00fa\3\2\2\2\u00fb\u00fc\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u00fc\u00fd\3\2\2\2\u00fd\u00fe\5\u0090I\2\u00fe")
 *         buf.write("\u00ff\5\24\13\2\u00ff\u0100\7\20\2\2\u0100\u0101\7Z\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 178, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__281, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 178, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":179
 *         buf.write("\u0103\3\2\2\2\u00f8\u00f9\7\30\2\2\u00f9\u00fb\7Z\2\2")
 *         buf.write("\u00fa\u00fc\5:\36\2\u00fb\u00fa\3\2\2\2\u00fb\u00fc\3")
 *         buf.write("\2\2\2\u00fc\u00fd\3\2\2\2\u00fd\u00fe\5\u0090I\2\u00fe")             # <<<<<<<<<<<<<<
 *         buf.write("\u00ff\5\24\13\2\u00ff\u0100\7\20\2\2\u0100\u0101\7Z\2")
 *         buf.write("\2\u0101\u0103\3\2\2\2\u0102\u00c9\3\2\2\2\u0102\u00cf")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 179, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__282, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 179, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":180
 *         buf.write("\u00fa\u00fc\5:\36\2\u00fb\u00fa\3\2\2\2\u00fb\u00fc\3")
 *         buf.write("\2\2\2\u00fc\u00fd\3\2\2\2\u00fd\u00fe\5\u0090I\2\u00fe")
 *         buf.write("\u00ff\5\24\13\2\u00ff\u0100\7\20\2\2\u0100\u0101\7Z\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u0101\u0103\3\2\2\2\u0102\u00c9\3\2\2\2\u0102\u00cf")
 *         buf.write("\3\2\2\2\u0102\u00db\3\2\2\2\u0102\u00e7\3\2\2\2\u0102")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 180, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__283, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 180, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":181
 *         buf.write("\2\2\2\u00fc\u00fd\3\2\2\2\u00fd\u00fe\5\u0090I\2\u00fe")
 *         buf.write("\u00ff\5\24\13\2\u00ff\u0100\7\20\2\2\u0100\u0101\7Z\2")
 *         buf.write("\2\u0101\u0103\3\2\2\2\u0102\u00c9\3\2\2\2\u0102\u00cf")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0102\u00db\3\2\2\2\u0102\u00e7\3\2\2\2\u0102")
 *         buf.write("\u00f8\3\2\2\2\u0103\r\3\2\2\2\u0104\u0105\5,\27\2\u0105")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 181, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__285, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 181, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":182
 *         buf.write("\u00ff\5\24\13\2\u00ff\u0100\7\20\2\2\u0100\u0101\7Z\2")
 *         buf.write("\2\u0101\u0103\3\2\2\2\u0102\u00c9\3\2\2\2\u0102\u00cf")
 *         buf.write("\3\2\2\2\u0102\u00db\3\2\2\2\u0102\u00e7\3\2\2\2\u0102")             # <<<<<<<<<<<<<<
 *         buf.write("\u00f8\3\2\2\2\u0103\r\3\2\2\2\u0104\u0105\5,\27\2\u0105")
 *         buf.write("\17\3\2\2\2\u0106\u010b\5\22\n\2\u0107\u0108\7\27\2\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 182, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__287, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 182, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":183
 *         buf.write("\2\u0101\u0103\3\2\2\2\u0102\u00c9\3\2\2\2\u0102\u00cf")
 *         buf.write("\3\2\2\2\u0102\u00db\3\2\2\2\u0102\u00e7\3\2\2\2\u0102")
 *         buf.write("\u00f8\3\2\2\2\u0103\r\3\2\2\2\u0104\u0105\5,\27\2\u0105")             # <<<<<<<<<<<<<<
 *         buf.write("\17\3\2\2\2\u0106\u010b\5\22\n\2\u0107\u0108\7\27\2\2")
 *         buf.write("\u0108\u010a\5\22\n\2\u0109\u0107\3\2\2\2\u010a\u010d")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 183, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__289, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 183, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":184
 *         buf.write("\3\2\2\2\u0102\u00db\3\2\2\2\u0102\u00e7\3\2\2\2\u0102")
 *         buf.write("\u00f8\3\2\2\2\u0103\r\3\2\2\2\u0104\u0105\5,\27\2\u0105")
 *         buf.write("\17\3\2\2\2\u0106\u010b\5\22\n\2\u0107\u0108\7\27\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0108\u010a\5\22\n\2\u0109\u0107\3\2\2\2\u010a\u010d")
 *         buf.write("\3\2\2\2\u010b\u0109\3\2\2\2\u010b\u010c\3\2\2\2\u010c")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 184, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__291, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 184, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":185
 *         buf.write("\u00f8\3\2\2\2\u0103\r\3\2\2\2\u0104\u0105\5,\27\2\u0105")
 *         buf.write("\17\3\2\2\2\u0106\u010b\5\22\n\2\u0107\u0108\7\27\2\2")
 *         buf.write("\u0108\u010a\5\22\n\2\u0109\u0107\3\2\2\2\u010a\u010d")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u010b\u0109\3\2\2\2\u010b\u010c\3\2\2\2\u010c")
 *         buf.write("\21\3\2\2\2\u010d\u010b\3\2\2\2\u010e\u010f\7Z\2\2\u010f")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 185, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__293, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 185, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":186
 *         buf.write("\17\3\2\2\2\u0106\u010b\5\22\n\2\u0107\u0108\7\27\2\2")
 *         buf.write("\u0108\u010a\5\22\n\2\u0109\u0107\3\2\2\2\u010a\u010d")
 *         buf.write("\3\2\2\2\u010b\u0109\3\2\2\2\u010b\u010c\3\2\2\2\u010c")             # <<<<<<<<<<<<<<
 *         buf.write("\21\3\2\2\2\u010d\u010b\3\2\2\2\u010e\u010f\7Z\2\2\u010f")
 *         buf.write("\u0110\5\u008eH\2\u0110\23\3\2\2\2\u0111\u011a\5\32\16")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 186, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__295, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 186, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":187
 *         buf.write("\u0108\u010a\5\22\n\2\u0109\u0107\3\2\2\2\u010a\u010d")
 *         buf.write("\3\2\2\2\u010b\u0109\3\2\2\2\u010b\u010c\3\2\2\2\u010c")
 *         buf.write("\21\3\2\2\2\u010d\u010b\3\2\2\2\u010e\u010f\7Z\2\2\u010f")             # <<<<<<<<<<<<<<
 *         buf.write("\u0110\5\u008eH\2\u0110\23\3\2\2\2\u0111\u011a\5\32\16")
 *         buf.write("\2\u0112\u0113\7\31\2\2\u0113\u0119\5\32\16\2\u0114\u0115")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 187, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__296, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 187, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":188
 *         buf.write("\3\2\2\2\u010b\u0109\3\2\2\2\u010b\u010c\3\2\2\2\u010c")
 *         buf.write("\21\3\2\2\2\u010d\u010b\3\2\2\2\u010e\u010f\7Z\2\2\u010f")
 *         buf.write("\u0110\5\u008eH\2\u0110\23\3\2\2\2\u0111\u011a\5\32\16")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u0112\u0113\7\31\2\2\u0113\u0119\5\32\16\2\u0114\u0115")
 *         buf.write("\7\32\2\2\u0115\u0119\5\32\16\2\u0116\u0119\5N(\2\u0117")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 188, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__297, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 188, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":189
 *         buf.write("\21\3\2\2\2\u010d\u010b\3\2\2\2\u010e\u010f\7Z\2\2\u010f")
 *         buf.write("\u0110\5\u008eH\2\u0110\23\3\2\2\2\u0111\u011a\5\32\16")
 *         buf.write("\2\u0112\u0113\7\31\2\2\u0113\u0119\5\32\16\2\u0114\u0115")             # <<<<<<<<<<<<<<
 *         buf.write("\7\32\2\2\u0115\u0119\5\32\16\2\u0116\u0119\5N(\2\u0117")
 *         buf.write("\u0119\5P)\2\u0118\u0112\3\2\2\2\u0118\u0114\3\2\2\2\u0118")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 189, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__299, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 189, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":190
 *         buf.write("\u0110\5\u008eH\2\u0110\23\3\2\2\2\u0111\u011a\5\32\16")
 *         buf.write("\2\u0112\u0113\7\31\2\2\u0113\u0119\5\32\16\2\u0114\u0115")
 *         buf.write("\7\32\2\2\u0115\u0119\5\32\16\2\u0116\u0119\5N(\2\u0117")             # <<<<<<<<<<<<<<
 *         buf.write("\u0119\5P)\2\u0118\u0112\3\2\2\2\u0118\u0114\3\2\2\2\u0118")
 *         buf.write("\u0116\3\2\2\2\u0118\u0117\3\2\2\2\u0119\u011c\3\2\2\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 190, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__300, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 190, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":191
 *         buf.write("\2\u0112\u0113\7\31\2\2\u0113\u0119\5\32\16\2\u0114\u0115")
 *         buf.write("\7\32\2\2\u0115\u0119\5\32\16\2\u0116\u0119\5N(\2\u0117")
 *         buf.write("\u0119\5P)\2\u0118\u0112\3\2\2\2\u0118\u0114\3\2\2\2\u0118")             # <<<<<<<<<<<<<<
 *         buf.write("\u0116\3\2\2\2\u0118\u0117\3\2\2\2\u0119\u011c\3\2\2\2")
 *         buf.write("\u011a\u0118\3\2\2\2\u011a\u011b\3\2\2\2\u011b\u0128\3")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 191, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__301, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 191, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":192
 *         buf.write("\7\32\2\2\u0115\u0119\5\32\16\2\u0116\u0119\5N(\2\u0117")
 *         buf.write("\u0119\5P)\2\u0118\u0112\3\2\2\2\u0118\u0114\3\2\2\2\u0118")
 *         buf.write("\u0116\3\2\2\2\u0118\u0117\3\2\2\2\u0119\u011c\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u011a\u0118\3\2\2\2\u011a\u011b\3\2\2\2\u011b\u0128\3")
 *         buf.write("\2\2\2\u011c\u011a\3\2\2\2\u011d\u011f\7\33\2\2\u011e")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 192, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__303, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 192, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":193
 *         buf.write("\u0119\5P)\2\u0118\u0112\3\2\2\2\u0118\u0114\3\2\2\2\u0118")
 *         buf.write("\u0116\3\2\2\2\u0118\u0117\3\2\2\2\u0119\u011c\3\2\2\2")
 *         buf.write("\u011a\u0118\3\2\2\2\u011a\u011b\3\2\2\2\u011b\u0128\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u011c\u011a\3\2\2\2\u011d\u011f\7\33\2\2\u011e")
 *         buf.write("\u0120\5\26\f\2\u011f\u011e\3\2\2\2\u011f\u0120\3\2\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 193, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__305, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 193, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":194
 *         buf.write("\u0116\3\2\2\2\u0118\u0117\3\2\2\2\u0119\u011c\3\2\2\2")
 *         buf.write("\u011a\u0118\3\2\2\2\u011a\u011b\3\2\2\2\u011b\u0128\3")
 *         buf.write("\2\2\2\u011c\u011a\3\2\2\2\u011d\u011f\7\33\2\2\u011e")             # <<<<<<<<<<<<<<
 *         buf.write("\u0120\5\26\f\2\u011f\u011e\3\2\2\2\u011f\u0120\3\2\2")
 *         buf.write("\2\u0120\u0122\3\2\2\2\u0121\u0123\5\30\r\2\u0122\u0121")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 194, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__307, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 194, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":195
 *         buf.write("\u011a\u0118\3\2\2\2\u011a\u011b\3\2\2\2\u011b\u0128\3")
 *         buf.write("\2\2\2\u011c\u011a\3\2\2\2\u011d\u011f\7\33\2\2\u011e")
 *         buf.write("\u0120\5\26\f\2\u011f\u011e\3\2\2\2\u011f\u0120\3\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u0120\u0122\3\2\2\2\u0121\u0123\5\30\r\2\u0122\u0121")
 *         buf.write("\3\2\2\2\u0122\u0123\3\2\2\2\u0123\u0125\3\2\2\2\u0124")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 195, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__309, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 195, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":196
 *         buf.write("\2\2\2\u011c\u011a\3\2\2\2\u011d\u011f\7\33\2\2\u011e")
 *         buf.write("\u0120\5\26\f\2\u011f\u011e\3\2\2\2\u011f\u0120\3\2\2")
 *         buf.write("\2\u0120\u0122\3\2\2\2\u0121\u0123\5\30\r\2\u0122\u0121")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0122\u0123\3\2\2\2\u0123\u0125\3\2\2\2\u0124")
 *         buf.write("\u0126\5\u0092J\2\u0125\u0124\3\2\2\2\u0125\u0126\3\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 196, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__311, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 196, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":197
 *         buf.write("\u0120\5\26\f\2\u011f\u011e\3\2\2\2\u011f\u0120\3\2\2")
 *         buf.write("\2\u0120\u0122\3\2\2\2\u0121\u0123\5\30\r\2\u0122\u0121")
 *         buf.write("\3\2\2\2\u0122\u0123\3\2\2\2\u0123\u0125\3\2\2\2\u0124")             # <<<<<<<<<<<<<<
 *         buf.write("\u0126\5\u0092J\2\u0125\u0124\3\2\2\2\u0125\u0126\3\2")
 *         buf.write("\2\2\u0126\u0127\3\2\2\2\u0127\u0129\7\24\2\2\u0128\u011d")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 197, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__313, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 197, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":198
 *         buf.write("\2\u0120\u0122\3\2\2\2\u0121\u0123\5\30\r\2\u0122\u0121")
 *         buf.write("\3\2\2\2\u0122\u0123\3\2\2\2\u0123\u0125\3\2\2\2\u0124")
 *         buf.write("\u0126\5\u0092J\2\u0125\u0124\3\2\2\2\u0125\u0126\3\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\u0126\u0127\3\2\2\2\u0127\u0129\7\24\2\2\u0128\u011d")
 *         buf.write("\3\2\2\2\u0128\u0129\3\2\2\2\u0129\u012d\3\2\2\2\u012a")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 198, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__314, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 198, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":199
 *         buf.write("\3\2\2\2\u0122\u0123\3\2\2\2\u0123\u0125\3\2\2\2\u0124")
 *         buf.write("\u0126\5\u0092J\2\u0125\u0124\3\2\2\2\u0125\u0126\3\2")
 *         buf.write("\2\2\u0126\u0127\3\2\2\2\u0127\u0129\7\24\2\2\u0128\u011d")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0128\u0129\3\2\2\2\u0129\u012d\3\2\2\2\u012a")
 *         buf.write("\u012b\5\u0092J\2\u012b\u012c\7\3\2\2\u012c\u012e\3\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 199, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__316, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 199, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":200
 *         buf.write("\u0126\5\u0092J\2\u0125\u0124\3\2\2\2\u0125\u0126\3\2")
 *         buf.write("\2\2\u0126\u0127\3\2\2\2\u0127\u0129\7\24\2\2\u0128\u011d")
 *         buf.write("\3\2\2\2\u0128\u0129\3\2\2\2\u0129\u012d\3\2\2\2\u012a")             # <<<<<<<<<<<<<<
 *         buf.write("\u012b\5\u0092J\2\u012b\u012c\7\3\2\2\u012c\u012e\3\2")
 *         buf.write("\2\2\u012d\u012a\3\2\2\2\u012d\u012e\3\2\2\2\u012e\25")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 200, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__318, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 200, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":201
 *         buf.write("\2\2\u0126\u0127\3\2\2\2\u0127\u0129\7\24\2\2\u0128\u011d")
 *         buf.write("\3\2\2\2\u0128\u0129\3\2\2\2\u0129\u012d\3\2\2\2\u012a")
 *         buf.write("\u012b\5\u0092J\2\u012b\u012c\7\3\2\2\u012c\u012e\3\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\u012d\u012a\3\2\2\2\u012d\u012e\3\2\2\2\u012e\25")
 *         buf.write("\3\2\2\2\u012f\u0130\7[\2\2\u0130\27\3\2\2\2\u0131\u0132")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 201, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__319, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 201, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":202
 *         buf.write("\3\2\2\2\u0128\u0129\3\2\2\2\u0129\u012d\3\2\2\2\u012a")
 *         buf.write("\u012b\5\u0092J\2\u012b\u012c\7\3\2\2\u012c\u012e\3\2")
 *         buf.write("\2\2\u012d\u012a\3\2\2\2\u012d\u012e\3\2\2\2\u012e\25")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u012f\u0130\7[\2\2\u0130\27\3\2\2\2\u0131\u0132")
 *         buf.write("\5z>\2\u0132\u0133\7\21\2\2\u0133\u0135\3\2\2\2\u0134")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 202, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__321, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 202, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":203
 *         buf.write("\u012b\5\u0092J\2\u012b\u012c\7\3\2\2\u012c\u012e\3\2")
 *         buf.write("\2\2\u012d\u012a\3\2\2\2\u012d\u012e\3\2\2\2\u012e\25")
 *         buf.write("\3\2\2\2\u012f\u0130\7[\2\2\u0130\27\3\2\2\2\u0131\u0132")             # <<<<<<<<<<<<<<
 *         buf.write("\5z>\2\u0132\u0133\7\21\2\2\u0133\u0135\3\2\2\2\u0134")
 *         buf.write("\u0131\3\2\2\2\u0134\u0135\3\2\2\2\u0135\u0136\3\2\2\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 203, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__323, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 203, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":204
 *         buf.write("\2\2\u012d\u012a\3\2\2\2\u012d\u012e\3\2\2\2\u012e\25")
 *         buf.write("\3\2\2\2\u012f\u0130\7[\2\2\u0130\27\3\2\2\2\u0131\u0132")
 *         buf.write("\5z>\2\u0132\u0133\7\21\2\2\u0133\u0135\3\2\2\2\u0134")             # <<<<<<<<<<<<<<
 *         buf.write("\u0131\3\2\2\2\u0134\u0135\3\2\2\2\u0135\u0136\3\2\2\2")
 *         buf.write("\u0136\u0137\7Z\2\2\u0137\u0139\7\23\2\2\u0138\u013a\5")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 204, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__324, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 204, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":205
 *         buf.write("\3\2\2\2\u012f\u0130\7[\2\2\u0130\27\3\2\2\2\u0131\u0132")
 *         buf.write("\5z>\2\u0132\u0133\7\21\2\2\u0133\u0135\3\2\2\2\u0134")
 *         buf.write("\u0131\3\2\2\2\u0134\u0135\3\2\2\2\u0135\u0136\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0136\u0137\7Z\2\2\u0137\u0139\7\23\2\2\u0138\u013a\5")
 *         buf.write("\u0088E\2\u0139\u0138\3\2\2\2\u0139\u013a\3\2\2\2\u013a")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 205, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__326, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 205, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":206
 *         buf.write("\5z>\2\u0132\u0133\7\21\2\2\u0133\u0135\3\2\2\2\u0134")
 *         buf.write("\u0131\3\2\2\2\u0134\u0135\3\2\2\2\u0135\u0136\3\2\2\2")
 *         buf.write("\u0136\u0137\7Z\2\2\u0137\u0139\7\23\2\2\u0138\u013a\5")             # <<<<<<<<<<<<<<
 *         buf.write("\u0088E\2\u0139\u0138\3\2\2\2\u0139\u013a\3\2\2\2\u013a")
 *         buf.write("\u013b\3\2\2\2\u013b\u013c\7\25\2\2\u013c\31\3\2\2\2\u013d")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 206, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__327, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 206, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":207
 *         buf.write("\u0131\3\2\2\2\u0134\u0135\3\2\2\2\u0135\u0136\3\2\2\2")
 *         buf.write("\u0136\u0137\7Z\2\2\u0137\u0139\7\23\2\2\u0138\u013a\5")
 *         buf.write("\u0088E\2\u0139\u0138\3\2\2\2\u0139\u013a\3\2\2\2\u013a")             # <<<<<<<<<<<<<<
 *         buf.write("\u013b\3\2\2\2\u013b\u013c\7\25\2\2\u013c\31\3\2\2\2\u013d")
 *         buf.write("\u013e\5\34\17\2\u013e\u013f\7\3\2\2\u013f\u0141\3\2\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 207, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__328, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 207, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":208
 *         buf.write("\u0136\u0137\7Z\2\2\u0137\u0139\7\23\2\2\u0138\u013a\5")
 *         buf.write("\u0088E\2\u0139\u0138\3\2\2\2\u0139\u013a\3\2\2\2\u013a")
 *         buf.write("\u013b\3\2\2\2\u013b\u013c\7\25\2\2\u013c\31\3\2\2\2\u013d")             # <<<<<<<<<<<<<<
 *         buf.write("\u013e\5\34\17\2\u013e\u013f\7\3\2\2\u013f\u0141\3\2\2")
 *         buf.write("\2\u0140\u013d\3\2\2\2\u0141\u0144\3\2\2\2\u0142\u0140")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 208, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__330, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 208, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":209
 *         buf.write("\u0088E\2\u0139\u0138\3\2\2\2\u0139\u013a\3\2\2\2\u013a")
 *         buf.write("\u013b\3\2\2\2\u013b\u013c\7\25\2\2\u013c\31\3\2\2\2\u013d")
 *         buf.write("\u013e\5\34\17\2\u013e\u013f\7\3\2\2\u013f\u0141\3\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u0140\u013d\3\2\2\2\u0141\u0144\3\2\2\2\u0142\u0140")
 *         buf.write("\3\2\2\2\u0142\u0143\3\2\2\2\u0143\33\3\2\2\2\u0144\u0142")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 209, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__332, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 209, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":210
 *         buf.write("\u013b\3\2\2\2\u013b\u013c\7\25\2\2\u013c\31\3\2\2\2\u013d")
 *         buf.write("\u013e\5\34\17\2\u013e\u013f\7\3\2\2\u013f\u0141\3\2\2")
 *         buf.write("\2\u0140\u013d\3\2\2\2\u0141\u0144\3\2\2\2\u0142\u0140")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0142\u0143\3\2\2\2\u0143\33\3\2\2\2\u0144\u0142")
 *         buf.write("\3\2\2\2\u0145\u014a\5\"\22\2\u0146\u014a\5&\24\2\u0147")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 210, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__334, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 210, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":211
 *         buf.write("\u013e\5\34\17\2\u013e\u013f\7\3\2\2\u013f\u0141\3\2\2")
 *         buf.write("\2\u0140\u013d\3\2\2\2\u0141\u0144\3\2\2\2\u0142\u0140")
 *         buf.write("\3\2\2\2\u0142\u0143\3\2\2\2\u0143\33\3\2\2\2\u0144\u0142")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0145\u014a\5\"\22\2\u0146\u014a\5&\24\2\u0147")
 *         buf.write("\u014a\5\36\20\2\u0148\u014a\5 \21\2\u0149\u0145\3\2\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 211, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__336, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 211, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":212
 *         buf.write("\2\u0140\u013d\3\2\2\2\u0141\u0144\3\2\2\2\u0142\u0140")
 *         buf.write("\3\2\2\2\u0142\u0143\3\2\2\2\u0143\33\3\2\2\2\u0144\u0142")
 *         buf.write("\3\2\2\2\u0145\u014a\5\"\22\2\u0146\u014a\5&\24\2\u0147")             # <<<<<<<<<<<<<<
 *         buf.write("\u014a\5\36\20\2\u0148\u014a\5 \21\2\u0149\u0145\3\2\2")
 *         buf.write("\2\u0149\u0146\3\2\2\2\u0149\u0147\3\2\2\2\u0149\u0148")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 212, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__338, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 212, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":213
 *         buf.write("\3\2\2\2\u0142\u0143\3\2\2\2\u0143\33\3\2\2\2\u0144\u0142")
 *         buf.write("\3\2\2\2\u0145\u014a\5\"\22\2\u0146\u014a\5&\24\2\u0147")
 *         buf.write("\u014a\5\36\20\2\u0148\u014a\5 \21\2\u0149\u0145\3\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u0149\u0146\3\2\2\2\u0149\u0147\3\2\2\2\u0149\u0148")
 *         buf.write("\3\2\2\2\u014a\35\3\2\2\2\u014b\u014d\7V\2\2\u014c\u014b")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 213, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__340, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 213, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":214
 *         buf.write("\3\2\2\2\u0145\u014a\5\"\22\2\u0146\u014a\5&\24\2\u0147")
 *         buf.write("\u014a\5\36\20\2\u0148\u014a\5 \21\2\u0149\u0145\3\2\2")
 *         buf.write("\2\u0149\u0146\3\2\2\2\u0149\u0147\3\2\2\2\u0149\u0148")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u014a\35\3\2\2\2\u014b\u014d\7V\2\2\u014c\u014b")
 *         buf.write("\3\2\2\2\u014c\u014d\3\2\2\2\u014d\u014f\3\2\2\2\u014e")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 214, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__342, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 214, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":215
 *         buf.write("\u014a\5\36\20\2\u0148\u014a\5 \21\2\u0149\u0145\3\2\2")
 *         buf.write("\2\u0149\u0146\3\2\2\2\u0149\u0147\3\2\2\2\u0149\u0148")
 *         buf.write("\3\2\2\2\u014a\35\3\2\2\2\u014b\u014d\7V\2\2\u014c\u014b")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u014c\u014d\3\2\2\2\u014d\u014f\3\2\2\2\u014e")
 *         buf.write("\u0150\7S\2\2\u014f\u014e\3\2\2\2\u014f\u0150\3\2\2\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 215, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__343, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 215, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":216
 *         buf.write("\2\u0149\u0146\3\2\2\2\u0149\u0147\3\2\2\2\u0149\u0148")
 *         buf.write("\3\2\2\2\u014a\35\3\2\2\2\u014b\u014d\7V\2\2\u014c\u014b")
 *         buf.write("\3\2\2\2\u014c\u014d\3\2\2\2\u014d\u014f\3\2\2\2\u014e")             # <<<<<<<<<<<<<<
 *         buf.write("\u0150\7S\2\2\u014f\u014e\3\2\2\2\u014f\u0150\3\2\2\2")
 *         buf.write("\u0150\u0152\3\2\2\2\u0151\u0153\7W\2\2\u0152\u0151\3")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 216, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__345, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 216, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":217
 *         buf.write("\3\2\2\2\u014a\35\3\2\2\2\u014b\u014d\7V\2\2\u014c\u014b")
 *         buf.write("\3\2\2\2\u014c\u014d\3\2\2\2\u014d\u014f\3\2\2\2\u014e")
 *         buf.write("\u0150\7S\2\2\u014f\u014e\3\2\2\2\u014f\u0150\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0150\u0152\3\2\2\2\u0151\u0153\7W\2\2\u0152\u0151\3")
 *         buf.write("\2\2\2\u0152\u0153\3\2\2\2\u0153\u0155\3\2\2\2\u0154\u0156")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 217, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__346, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 217, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":218
 *         buf.write("\3\2\2\2\u014c\u014d\3\2\2\2\u014d\u014f\3\2\2\2\u014e")
 *         buf.write("\u0150\7S\2\2\u014f\u014e\3\2\2\2\u014f\u0150\3\2\2\2")
 *         buf.write("\u0150\u0152\3\2\2\2\u0151\u0153\7W\2\2\u0152\u0151\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u0152\u0153\3\2\2\2\u0153\u0155\3\2\2\2\u0154\u0156")
 *         buf.write("\7X\2\2\u0155\u0154\3\2\2\2\u0155\u0156\3\2\2\2\u0156")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 218, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__347, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 218, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":219
 *         buf.write("\u0150\7S\2\2\u014f\u014e\3\2\2\2\u014f\u0150\3\2\2\2")
 *         buf.write("\u0150\u0152\3\2\2\2\u0151\u0153\7W\2\2\u0152\u0151\3")
 *         buf.write("\2\2\2\u0152\u0153\3\2\2\2\u0153\u0155\3\2\2\2\u0154\u0156")             # <<<<<<<<<<<<<<
 *         buf.write("\7X\2\2\u0155\u0154\3\2\2\2\u0155\u0156\3\2\2\2\u0156")
 *         buf.write("\u0159\3\2\2\2\u0157\u015a\5\6\4\2\u0158\u015a\5*\26\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 219, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__349, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 219, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":220
 *         buf.write("\u0150\u0152\3\2\2\2\u0151\u0153\7W\2\2\u0152\u0151\3")
 *         buf.write("\2\2\2\u0152\u0153\3\2\2\2\u0153\u0155\3\2\2\2\u0154\u0156")
 *         buf.write("\7X\2\2\u0155\u0154\3\2\2\2\u0155\u0156\3\2\2\2\u0156")             # <<<<<<<<<<<<<<
 *         buf.write("\u0159\3\2\2\2\u0157\u015a\5\6\4\2\u0158\u015a\5*\26\2")
 *         buf.write("\u0159\u0157\3\2\2\2\u0159\u0158\3\2\2\2\u015a\37\3\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 220, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__350, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 220, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":221
 *         buf.write("\2\2\2\u0152\u0153\3\2\2\2\u0153\u0155\3\2\2\2\u0154\u0156")
 *         buf.write("\7X\2\2\u0155\u0154\3\2\2\2\u0155\u0156\3\2\2\2\u0156")
 *         buf.write("\u0159\3\2\2\2\u0157\u015a\5\6\4\2\u0158\u015a\5*\26\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0159\u0157\3\2\2\2\u0159\u0158\3\2\2\2\u015a\37\3\2")
 *         buf.write("\2\2\u015b\u015d\7V\2\2\u015c\u015b\3\2\2\2\u015c\u015d")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 221, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__352, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 221, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":222
 *         buf.write("\7X\2\2\u0155\u0154\3\2\2\2\u0155\u0156\3\2\2\2\u0156")
 *         buf.write("\u0159\3\2\2\2\u0157\u015a\5\6\4\2\u0158\u015a\5*\26\2")
 *         buf.write("\u0159\u0157\3\2\2\2\u0159\u0158\3\2\2\2\u015a\37\3\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\u015b\u015d\7V\2\2\u015c\u015b\3\2\2\2\u015c\u015d")
 *         buf.write("\3\2\2\2\u015d\u015f\3\2\2\2\u015e\u0160\7S\2\2\u015f")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 222, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__354, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 222, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":223
 *         buf.write("\u0159\3\2\2\2\u0157\u015a\5\6\4\2\u0158\u015a\5*\26\2")
 *         buf.write("\u0159\u0157\3\2\2\2\u0159\u0158\3\2\2\2\u015a\37\3\2")
 *         buf.write("\2\2\u015b\u015d\7V\2\2\u015c\u015b\3\2\2\2\u015c\u015d")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u015d\u015f\3\2\2\2\u015e\u0160\7S\2\2\u015f")
 *         buf.write("\u015e\3\2\2\2\u015f\u0160\3\2\2\2\u0160\u0162\3\2\2\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 223, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__355, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 223, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":224
 *         buf.write("\u0159\u0157\3\2\2\2\u0159\u0158\3\2\2\2\u015a\37\3\2")
 *         buf.write("\2\2\u015b\u015d\7V\2\2\u015c\u015b\3\2\2\2\u015c\u015d")
 *         buf.write("\3\2\2\2\u015d\u015f\3\2\2\2\u015e\u0160\7S\2\2\u015f")             # <<<<<<<<<<<<<<
 *         buf.write("\u015e\3\2\2\2\u015f\u0160\3\2\2\2\u0160\u0162\3\2\2\2")
 *         buf.write("\u0161\u0163\7W\2\2\u0162\u0161\3\2\2\2\u0162\u0163\3")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 224, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__356, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 224, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":225
 *         buf.write("\2\2\u015b\u015d\7V\2\2\u015c\u015b\3\2\2\2\u015c\u015d")
 *         buf.write("\3\2\2\2\u015d\u015f\3\2\2\2\u015e\u0160\7S\2\2\u015f")
 *         buf.write("\u015e\3\2\2\2\u015f\u0160\3\2\2\2\u0160\u0162\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0161\u0163\7W\2\2\u0162\u0161\3\2\2\2\u0162\u0163\3")
 *         buf.write("\2\2\2\u0163\u0165\3\2\2\2\u0164\u0166\7X\2\2\u0165\u0164")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 225, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__358, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 225, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":226
 *         buf.write("\3\2\2\2\u015d\u015f\3\2\2\2\u015e\u0160\7S\2\2\u015f")
 *         buf.write("\u015e\3\2\2\2\u015f\u0160\3\2\2\2\u0160\u0162\3\2\2\2")
 *         buf.write("\u0161\u0163\7W\2\2\u0162\u0161\3\2\2\2\u0162\u0163\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u0163\u0165\3\2\2\2\u0164\u0166\7X\2\2\u0165\u0164")
 *         buf.write("\3\2\2\2\u0165\u0166\3\2\2\2\u0166\u0167\3\2\2\2\u0167")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 226, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__359, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 226, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":227
 *         buf.write("\u015e\3\2\2\2\u015f\u0160\3\2\2\2\u0160\u0162\3\2\2\2")
 *         buf.write("\u0161\u0163\7W\2\2\u0162\u0161\3\2\2\2\u0162\u0163\3")
 *         buf.write("\2\2\2\u0163\u0165\3\2\2\2\u0164\u0166\7X\2\2\u0165\u0164")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0165\u0166\3\2\2\2\u0166\u0167\3\2\2\2\u0167")
 *         buf.write("\u016a\7\34\2\2\u0168\u016b\5\6\4\2\u0169\u016b\5*\26")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 227, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__360, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 227, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":228
 *         buf.write("\u0161\u0163\7W\2\2\u0162\u0161\3\2\2\2\u0162\u0163\3")
 *         buf.write("\2\2\2\u0163\u0165\3\2\2\2\u0164\u0166\7X\2\2\u0165\u0164")
 *         buf.write("\3\2\2\2\u0165\u0166\3\2\2\2\u0166\u0167\3\2\2\2\u0167")             # <<<<<<<<<<<<<<
 *         buf.write("\u016a\7\34\2\2\u0168\u016b\5\6\4\2\u0169\u016b\5*\26")
 *         buf.write("\2\u016a\u0168\3\2\2\2\u016a\u0169\3\2\2\2\u016b\u016f")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 228, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__362, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 228, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":229
 *         buf.write("\2\2\2\u0163\u0165\3\2\2\2\u0164\u0166\7X\2\2\u0165\u0164")
 *         buf.write("\3\2\2\2\u0165\u0166\3\2\2\2\u0166\u0167\3\2\2\2\u0167")
 *         buf.write("\u016a\7\34\2\2\u0168\u016b\5\6\4\2\u0169\u016b\5*\26")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u016a\u0168\3\2\2\2\u016a\u0169\3\2\2\2\u016b\u016f")
 *         buf.write("\3\2\2\2\u016c\u016d\5(\25\2\u016d\u016e\5\u008eH\2\u016e")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 229, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__364, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 229, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":230
 *         buf.write("\3\2\2\2\u0165\u0166\3\2\2\2\u0166\u0167\3\2\2\2\u0167")
 *         buf.write("\u016a\7\34\2\2\u0168\u016b\5\6\4\2\u0169\u016b\5*\26")
 *         buf.write("\2\u016a\u0168\3\2\2\2\u016a\u0169\3\2\2\2\u016b\u016f")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u016c\u016d\5(\25\2\u016d\u016e\5\u008eH\2\u016e")
 *         buf.write("\u0170\3\2\2\2\u016f\u016c\3\2\2\2\u016f\u0170\3\2\2\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 230, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__366, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 230, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":231
 *         buf.write("\u016a\7\34\2\2\u0168\u016b\5\6\4\2\u0169\u016b\5*\26")
 *         buf.write("\2\u016a\u0168\3\2\2\2\u016a\u0169\3\2\2\2\u016b\u016f")
 *         buf.write("\3\2\2\2\u016c\u016d\5(\25\2\u016d\u016e\5\u008eH\2\u016e")             # <<<<<<<<<<<<<<
 *         buf.write("\u0170\3\2\2\2\u016f\u016c\3\2\2\2\u016f\u0170\3\2\2\2")
 *         buf.write("\u0170!\3\2\2\2\u0171\u0180\7\35\2\2\u0172\u0173\7Z\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 231, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__367, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 231, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":232
 *         buf.write("\2\u016a\u0168\3\2\2\2\u016a\u0169\3\2\2\2\u016b\u016f")
 *         buf.write("\3\2\2\2\u016c\u016d\5(\25\2\u016d\u016e\5\u008eH\2\u016e")
 *         buf.write("\u0170\3\2\2\2\u016f\u016c\3\2\2\2\u016f\u0170\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0170!\3\2\2\2\u0171\u0180\7\35\2\2\u0172\u0173\7Z\2")
 *         buf.write("\2\u0173\u0174\7\21\2\2\u0174\u0181\5z>\2\u0175\u017e")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 232, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__369, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 232, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":233
 *         buf.write("\3\2\2\2\u016c\u016d\5(\25\2\u016d\u016e\5\u008eH\2\u016e")
 *         buf.write("\u0170\3\2\2\2\u016f\u016c\3\2\2\2\u016f\u0170\3\2\2\2")
 *         buf.write("\u0170!\3\2\2\2\u0171\u0180\7\35\2\2\u0172\u0173\7Z\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u0173\u0174\7\21\2\2\u0174\u0181\5z>\2\u0175\u017e")
 *         buf.write("\5z>\2\u0176\u017c\7\36\2\2\u0177\u017d\7\37\2\2\u0178")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 233, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__370, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 233, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":234
 *         buf.write("\u0170\3\2\2\2\u016f\u016c\3\2\2\2\u016f\u0170\3\2\2\2")
 *         buf.write("\u0170!\3\2\2\2\u0171\u0180\7\35\2\2\u0172\u0173\7Z\2")
 *         buf.write("\2\u0173\u0174\7\21\2\2\u0174\u0181\5z>\2\u0175\u017e")             # <<<<<<<<<<<<<<
 *         buf.write("\5z>\2\u0176\u017c\7\36\2\2\u0177\u017d\7\37\2\2\u0178")
 *         buf.write("\u0179\7 \2\2\u0179\u017a\5$\23\2\u017a\u017b\7!\2\2\u017b")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 234, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__371, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 234, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":235
 *         buf.write("\u0170!\3\2\2\2\u0171\u0180\7\35\2\2\u0172\u0173\7Z\2")
 *         buf.write("\2\u0173\u0174\7\21\2\2\u0174\u0181\5z>\2\u0175\u017e")
 *         buf.write("\5z>\2\u0176\u017c\7\36\2\2\u0177\u017d\7\37\2\2\u0178")             # <<<<<<<<<<<<<<
 *         buf.write("\u0179\7 \2\2\u0179\u017a\5$\23\2\u017a\u017b\7!\2\2\u017b")
 *         buf.write("\u017d\3\2\2\2\u017c\u0177\3\2\2\2\u017c\u0178\3\2\2\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 235, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__372, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 235, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":236
 *         buf.write("\2\u0173\u0174\7\21\2\2\u0174\u0181\5z>\2\u0175\u017e")
 *         buf.write("\5z>\2\u0176\u017c\7\36\2\2\u0177\u017d\7\37\2\2\u0178")
 *         buf.write("\u0179\7 \2\2\u0179\u017a\5$\23\2\u017a\u017b\7!\2\2\u017b")             # <<<<<<<<<<<<<<
 *         buf.write("\u017d\3\2\2\2\u017c\u0177\3\2\2\2\u017c\u0178\3\2\2\2")
 *         buf.write("\u017d\u017f\3\2\2\2\u017e\u0176\3\2\2\2\u017e\u017f\3")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 236, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__374, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 236, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":237
 *         buf.write("\5z>\2\u0176\u017c\7\36\2\2\u0177\u017d\7\37\2\2\u0178")
 *         buf.write("\u0179\7 \2\2\u0179\u017a\5$\23\2\u017a\u017b\7!\2\2\u017b")
 *         buf.write("\u017d\3\2\2\2\u017c\u0177\3\2\2\2\u017c\u0178\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u017d\u017f\3\2\2\2\u017e\u0176\3\2\2\2\u017e\u017f\3")
 *         buf.write("\2\2\2\u017f\u0181\3\2\2\2\u0180\u0172\3\2\2\2\u0180\u0175")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 237, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__376, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 237, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":238
 *         buf.write("\u0179\7 \2\2\u0179\u017a\5$\23\2\u017a\u017b\7!\2\2\u017b")
 *         buf.write("\u017d\3\2\2\2\u017c\u0177\3\2\2\2\u017c\u0178\3\2\2\2")
 *         buf.write("\u017d\u017f\3\2\2\2\u017e\u0176\3\2\2\2\u017e\u017f\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u017f\u0181\3\2\2\2\u0180\u0172\3\2\2\2\u0180\u0175")
 *         buf.write("\3\2\2\2\u0181\u0182\3\2\2\2\u0182\u0183\5\u008eH\2\u0183")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 238, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__378, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 238, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":239
 *         buf.write("\u017d\3\2\2\2\u017c\u0177\3\2\2\2\u017c\u0178\3\2\2\2")
 *         buf.write("\u017d\u017f\3\2\2\2\u017e\u0176\3\2\2\2\u017e\u017f\3")
 *         buf.write("\2\2\2\u017f\u0181\3\2\2\2\u0180\u0172\3\2\2\2\u0180\u0175")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0181\u0182\3\2\2\2\u0182\u0183\5\u008eH\2\u0183")
 *         buf.write("#\3\2\2\2\u0184\u0189\7Z\2\2\u0185\u0186\7\27\2\2\u0186")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 239, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__380, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 239, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":240
 *         buf.write("\u017d\u017f\3\2\2\2\u017e\u0176\3\2\2\2\u017e\u017f\3")
 *         buf.write("\2\2\2\u017f\u0181\3\2\2\2\u0180\u0172\3\2\2\2\u0180\u0175")
 *         buf.write("\3\2\2\2\u0181\u0182\3\2\2\2\u0182\u0183\5\u008eH\2\u0183")             # <<<<<<<<<<<<<<
 *         buf.write("#\3\2\2\2\u0184\u0189\7Z\2\2\u0185\u0186\7\27\2\2\u0186")
 *         buf.write("\u0188\5$\23\2\u0187\u0185\3\2\2\2\u0188\u018b\3\2\2\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 240, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__381, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 240, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":241
 *         buf.write("\2\2\2\u017f\u0181\3\2\2\2\u0180\u0172\3\2\2\2\u0180\u0175")
 *         buf.write("\3\2\2\2\u0181\u0182\3\2\2\2\u0182\u0183\5\u008eH\2\u0183")
 *         buf.write("#\3\2\2\2\u0184\u0189\7Z\2\2\u0185\u0186\7\27\2\2\u0186")             # <<<<<<<<<<<<<<
 *         buf.write("\u0188\5$\23\2\u0187\u0185\3\2\2\2\u0188\u018b\3\2\2\2")
 *         buf.write("\u0189\u0187\3\2\2\2\u0189\u018a\3\2\2\2\u018a%\3\2\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 241, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__382, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 241, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":242
 *         buf.write("\3\2\2\2\u0181\u0182\3\2\2\2\u0182\u0183\5\u008eH\2\u0183")
 *         buf.write("#\3\2\2\2\u0184\u0189\7Z\2\2\u0185\u0186\7\27\2\2\u0186")
 *         buf.write("\u0188\5$\23\2\u0187\u0185\3\2\2\2\u0188\u018b\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0189\u0187\3\2\2\2\u0189\u018a\3\2\2\2\u018a%\3\2\2")
 *         buf.write("\2\u018b\u0189\3\2\2\2\u018c\u018d\7\30\2\2\u018d\u018f")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 242, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__384, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 242, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":243
 *         buf.write("#\3\2\2\2\u0184\u0189\7Z\2\2\u0185\u0186\7\27\2\2\u0186")
 *         buf.write("\u0188\5$\23\2\u0187\u0185\3\2\2\2\u0188\u018b\3\2\2\2")
 *         buf.write("\u0189\u0187\3\2\2\2\u0189\u018a\3\2\2\2\u018a%\3\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u018b\u0189\3\2\2\2\u018c\u018d\7\30\2\2\u018d\u018f")
 *         buf.write("\5z>\2\u018e\u0190\5:\36\2\u018f\u018e\3\2\2\2\u018f\u0190")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 243, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__386, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 243, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":244
 *         buf.write("\u0188\5$\23\2\u0187\u0185\3\2\2\2\u0188\u018b\3\2\2\2")
 *         buf.write("\u0189\u0187\3\2\2\2\u0189\u018a\3\2\2\2\u018a%\3\2\2")
 *         buf.write("\2\u018b\u0189\3\2\2\2\u018c\u018d\7\30\2\2\u018d\u018f")             # <<<<<<<<<<<<<<
 *         buf.write("\5z>\2\u018e\u0190\5:\36\2\u018f\u018e\3\2\2\2\u018f\u0190")
 *         buf.write("\3\2\2\2\u0190\u0192\3\2\2\2\u0191\u0193\5\u0092J\2\u0192")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 244, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__388, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":245
 *         buf.write("\u0189\u0187\3\2\2\2\u0189\u018a\3\2\2\2\u018a%\3\2\2")
 *         buf.write("\2\u018b\u0189\3\2\2\2\u018c\u018d\7\30\2\2\u018d\u018f")
 *         buf.write("\5z>\2\u018e\u0190\5:\36\2\u018f\u018e\3\2\2\2\u018f\u0190")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0190\u0192\3\2\2\2\u0191\u0193\5\u0092J\2\u0192")
 *         buf.write("\u0191\3\2\2\2\u0192\u0193\3\2\2\2\u0193\'\3\2\2\2\u0194")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 245, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__389, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 245, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":246
 *         buf.write("\2\u018b\u0189\3\2\2\2\u018c\u018d\7\30\2\2\u018d\u018f")
 *         buf.write("\5z>\2\u018e\u0190\5:\36\2\u018f\u018e\3\2\2\2\u018f\u0190")
 *         buf.write("\3\2\2\2\u0190\u0192\3\2\2\2\u0191\u0193\5\u0092J\2\u0192")             # <<<<<<<<<<<<<<
 *         buf.write("\u0191\3\2\2\2\u0192\u0193\3\2\2\2\u0193\'\3\2\2\2\u0194")
 *         buf.write("\u0195\7\"\2\2\u0195\u0197\5v<\2\u0196\u0198\5:\36\2\u0197")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 246, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__390, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 246, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":247
 *         buf.write("\5z>\2\u018e\u0190\5:\36\2\u018f\u018e\3\2\2\2\u018f\u0190")
 *         buf.write("\3\2\2\2\u0190\u0192\3\2\2\2\u0191\u0193\5\u0092J\2\u0192")
 *         buf.write("\u0191\3\2\2\2\u0192\u0193\3\2\2\2\u0193\'\3\2\2\2\u0194")             # <<<<<<<<<<<<<<
 *         buf.write("\u0195\7\"\2\2\u0195\u0197\5v<\2\u0196\u0198\5:\36\2\u0197")
 *         buf.write("\u0196\3\2\2\2\u0197\u0198\3\2\2\2\u0198)\3\2\2\2\u0199")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 247, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__392, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 247, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":248
 *         buf.write("\3\2\2\2\u0190\u0192\3\2\2\2\u0191\u0193\5\u0092J\2\u0192")
 *         buf.write("\u0191\3\2\2\2\u0192\u0193\3\2\2\2\u0193\'\3\2\2\2\u0194")
 *         buf.write("\u0195\7\"\2\2\u0195\u0197\5v<\2\u0196\u0198\5:\36\2\u0197")             # <<<<<<<<<<<<<<
 *         buf.write("\u0196\3\2\2\2\u0197\u0198\3\2\2\2\u0198)\3\2\2\2\u0199")
 *         buf.write("\u019a\5,\27\2\u019a\u019c\5.\30\2\u019b\u019d\5\u008a")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 248, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__393, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 248, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":249
 *         buf.write("\u0191\3\2\2\2\u0192\u0193\3\2\2\2\u0193\'\3\2\2\2\u0194")
 *         buf.write("\u0195\7\"\2\2\u0195\u0197\5v<\2\u0196\u0198\5:\36\2\u0197")
 *         buf.write("\u0196\3\2\2\2\u0197\u0198\3\2\2\2\u0198)\3\2\2\2\u0199")             # <<<<<<<<<<<<<<
 *         buf.write("\u019a\5,\27\2\u019a\u019c\5.\30\2\u019b\u019d\5\u008a")
 *         buf.write("F\2\u019c\u019b\3\2\2\2\u019c\u019d\3\2\2\2\u019d\u019e")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 249, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__395, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 249, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":250
 *         buf.write("\u0195\7\"\2\2\u0195\u0197\5v<\2\u0196\u0198\5:\36\2\u0197")
 *         buf.write("\u0196\3\2\2\2\u0197\u0198\3\2\2\2\u0198)\3\2\2\2\u0199")
 *         buf.write("\u019a\5,\27\2\u019a\u019c\5.\30\2\u019b\u019d\5\u008a")             # <<<<<<<<<<<<<<
 *         buf.write("F\2\u019c\u019b\3\2\2\2\u019c\u019d\3\2\2\2\u019d\u019e")
 *         buf.write("\3\2\2\2\u019e\u019f\5\60\31\2\u019f+\3\2\2\2\u01a0\u01a2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 250, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__397, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 250, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":251
 *         buf.write("\u0196\3\2\2\2\u0197\u0198\3\2\2\2\u0198)\3\2\2\2\u0199")
 *         buf.write("\u019a\5,\27\2\u019a\u019c\5.\30\2\u019b\u019d\5\u008a")
 *         buf.write("F\2\u019c\u019b\3\2\2\2\u019c\u019d\3\2\2\2\u019d\u019e")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u019e\u019f\5\60\31\2\u019f+\3\2\2\2\u01a0\u01a2")
 *         buf.write("\t\3\2\2\u01a1\u01a0\3\2\2\2\u01a1\u01a2\3\2\2\2\u01a2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 251, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__398, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 251, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":252
 *         buf.write("\u019a\5,\27\2\u019a\u019c\5.\30\2\u019b\u019d\5\u008a")
 *         buf.write("F\2\u019c\u019b\3\2\2\2\u019c\u019d\3\2\2\2\u019d\u019e")
 *         buf.write("\3\2\2\2\u019e\u019f\5\60\31\2\u019f+\3\2\2\2\u01a0\u01a2")             # <<<<<<<<<<<<<<
 *         buf.write("\t\3\2\2\u01a1\u01a0\3\2\2\2\u01a1\u01a2\3\2\2\2\u01a2")
 *         buf.write("\u01a4\3\2\2\2\u01a3\u01a5\t\4\2\2\u01a4\u01a3\3\2\2\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 252, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__399, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 252, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":253
 *         buf.write("F\2\u019c\u019b\3\2\2\2\u019c\u019d\3\2\2\2\u019d\u019e")
 *         buf.write("\3\2\2\2\u019e\u019f\5\60\31\2\u019f+\3\2\2\2\u01a0\u01a2")
 *         buf.write("\t\3\2\2\u01a1\u01a0\3\2\2\2\u01a1\u01a2\3\2\2\2\u01a2")             # <<<<<<<<<<<<<<
 *         buf.write("\u01a4\3\2\2\2\u01a3\u01a5\t\4\2\2\u01a4\u01a3\3\2\2\2")
 *         buf.write("\u01a4\u01a5\3\2\2\2\u01a5\u01a7\3\2\2\2\u01a6\u01a8\t")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 253, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__401, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 253, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":254
 *         buf.write("\3\2\2\2\u019e\u019f\5\60\31\2\u019f+\3\2\2\2\u01a0\u01a2")
 *         buf.write("\t\3\2\2\u01a1\u01a0\3\2\2\2\u01a1\u01a2\3\2\2\2\u01a2")
 *         buf.write("\u01a4\3\2\2\2\u01a3\u01a5\t\4\2\2\u01a4\u01a3\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u01a4\u01a5\3\2\2\2\u01a5\u01a7\3\2\2\2\u01a6\u01a8\t")
 *         buf.write("\5\2\2\u01a7\u01a6\3\2\2\2\u01a7\u01a8\3\2\2\2\u01a8-")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 254, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__403, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 254, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":255
 *         buf.write("\t\3\2\2\u01a1\u01a0\3\2\2\2\u01a1\u01a2\3\2\2\2\u01a2")
 *         buf.write("\u01a4\3\2\2\2\u01a3\u01a5\t\4\2\2\u01a4\u01a3\3\2\2\2")
 *         buf.write("\u01a4\u01a5\3\2\2\2\u01a5\u01a7\3\2\2\2\u01a6\u01a8\t")             # <<<<<<<<<<<<<<
 *         buf.write("\5\2\2\u01a7\u01a6\3\2\2\2\u01a7\u01a8\3\2\2\2\u01a8-")
 *         buf.write("\3\2\2\2\u01a9\u01aa\5z>\2\u01aa/\3\2\2\2\u01ab\u01b0")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 255, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__405, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 255, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":256
 *         buf.write("\u01a4\3\2\2\2\u01a3\u01a5\t\4\2\2\u01a4\u01a3\3\2\2\2")
 *         buf.write("\u01a4\u01a5\3\2\2\2\u01a5\u01a7\3\2\2\2\u01a6\u01a8\t")
 *         buf.write("\5\2\2\u01a7\u01a6\3\2\2\2\u01a7\u01a8\3\2\2\2\u01a8-")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u01a9\u01aa\5z>\2\u01aa/\3\2\2\2\u01ab\u01b0")
 *         buf.write("\5\62\32\2\u01ac\u01ad\7\27\2\2\u01ad\u01af\5\62\32\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 256, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__407, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 256, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":257
 *         buf.write("\u01a4\u01a5\3\2\2\2\u01a5\u01a7\3\2\2\2\u01a6\u01a8\t")
 *         buf.write("\5\2\2\u01a7\u01a6\3\2\2\2\u01a7\u01a8\3\2\2\2\u01a8-")
 *         buf.write("\3\2\2\2\u01a9\u01aa\5z>\2\u01aa/\3\2\2\2\u01ab\u01b0")             # <<<<<<<<<<<<<<
 *         buf.write("\5\62\32\2\u01ac\u01ad\7\27\2\2\u01ad\u01af\5\62\32\2")
 *         buf.write("\u01ae\u01ac\3\2\2\2\u01af\u01b2\3\2\2\2\u01b0\u01ae\3")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 257, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__408, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 257, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":258
 *         buf.write("\5\2\2\u01a7\u01a6\3\2\2\2\u01a7\u01a8\3\2\2\2\u01a8-")
 *         buf.write("\3\2\2\2\u01a9\u01aa\5z>\2\u01aa/\3\2\2\2\u01ab\u01b0")
 *         buf.write("\5\62\32\2\u01ac\u01ad\7\27\2\2\u01ad\u01af\5\62\32\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u01ae\u01ac\3\2\2\2\u01af\u01b2\3\2\2\2\u01b0\u01ae\3")
 *         buf.write("\2\2\2\u01b0\u01b1\3\2\2\2\u01b1\61\3\2\2\2\u01b2\u01b0")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 258, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__409, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":259
 *         buf.write("\3\2\2\2\u01a9\u01aa\5z>\2\u01aa/\3\2\2\2\u01ab\u01b0")
 *         buf.write("\5\62\32\2\u01ac\u01ad\7\27\2\2\u01ad\u01af\5\62\32\2")
 *         buf.write("\u01ae\u01ac\3\2\2\2\u01af\u01b2\3\2\2\2\u01b0\u01ae\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u01b0\u01b1\3\2\2\2\u01b1\61\3\2\2\2\u01b2\u01b0")
 *         buf.write("\3\2\2\2\u01b3\u01b5\5\66\34\2\u01b4\u01b6\5\64\33\2\u01b5")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 259, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__411, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 259, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":260
 *         buf.write("\5\62\32\2\u01ac\u01ad\7\27\2\2\u01ad\u01af\5\62\32\2")
 *         buf.write("\u01ae\u01ac\3\2\2\2\u01af\u01b2\3\2\2\2\u01b0\u01ae\3")
 *         buf.write("\2\2\2\u01b0\u01b1\3\2\2\2\u01b1\61\3\2\2\2\u01b2\u01b0")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u01b3\u01b5\5\66\34\2\u01b4\u01b6\5\64\33\2\u01b5")
 *         buf.write("\u01b4\3\2\2\2\u01b5\u01b6\3\2\2\2\u01b6\u01b7\3\2\2\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 260, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__412, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 260, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":261
 *         buf.write("\u01ae\u01ac\3\2\2\2\u01af\u01b2\3\2\2\2\u01b0\u01ae\3")
 *         buf.write("\2\2\2\u01b0\u01b1\3\2\2\2\u01b1\61\3\2\2\2\u01b2\u01b0")
 *         buf.write("\3\2\2\2\u01b3\u01b5\5\66\34\2\u01b4\u01b6\5\64\33\2\u01b5")             # <<<<<<<<<<<<<<
 *         buf.write("\u01b4\3\2\2\2\u01b5\u01b6\3\2\2\2\u01b6\u01b7\3\2\2\2")
 *         buf.write("\u01b7\u01b8\5\u008eH\2\u01b8\63\3\2\2\2\u01b9\u01ba\7")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 261, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__413, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 261, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":262
 *         buf.write("\2\2\2\u01b0\u01b1\3\2\2\2\u01b1\61\3\2\2\2\u01b2\u01b0")
 *         buf.write("\3\2\2\2\u01b3\u01b5\5\66\34\2\u01b4\u01b6\5\64\33\2\u01b5")
 *         buf.write("\u01b4\3\2\2\2\u01b5\u01b6\3\2\2\2\u01b6\u01b7\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u01b7\u01b8\5\u008eH\2\u01b8\63\3\2\2\2\u01b9\u01ba\7")
 *         buf.write("*\2\2\u01ba\u01bb\5n8\2\u01bb\65\3\2\2\2\u01bc\u01be\7")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 262, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__415, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 262, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":263
 *         buf.write("\3\2\2\2\u01b3\u01b5\5\66\34\2\u01b4\u01b6\5\64\33\2\u01b5")
 *         buf.write("\u01b4\3\2\2\2\u01b5\u01b6\3\2\2\2\u01b6\u01b7\3\2\2\2")
 *         buf.write("\u01b7\u01b8\5\u008eH\2\u01b8\63\3\2\2\2\u01b9\u01ba\7")             # <<<<<<<<<<<<<<
 *         buf.write("*\2\2\u01ba\u01bb\5n8\2\u01bb\65\3\2\2\2\u01bc\u01be\7")
 *         buf.write("Z\2\2\u01bd\u01bf\5\u008aF\2\u01be\u01bd\3\2\2\2\u01be")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 263, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__416, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 263, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":264
 *         buf.write("\u01b4\3\2\2\2\u01b5\u01b6\3\2\2\2\u01b6\u01b7\3\2\2\2")
 *         buf.write("\u01b7\u01b8\5\u008eH\2\u01b8\63\3\2\2\2\u01b9\u01ba\7")
 *         buf.write("*\2\2\u01ba\u01bb\5n8\2\u01bb\65\3\2\2\2\u01bc\u01be\7")             # <<<<<<<<<<<<<<
 *         buf.write("Z\2\2\u01bd\u01bf\5\u008aF\2\u01be\u01bd\3\2\2\2\u01be")
 *         buf.write("\u01bf\3\2\2\2\u01bf\u01c1\3\2\2\2\u01c0\u01c2\58\35\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 264, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__417, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 264, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":265
 *         buf.write("\u01b7\u01b8\5\u008eH\2\u01b8\63\3\2\2\2\u01b9\u01ba\7")
 *         buf.write("*\2\2\u01ba\u01bb\5n8\2\u01bb\65\3\2\2\2\u01bc\u01be\7")
 *         buf.write("Z\2\2\u01bd\u01bf\5\u008aF\2\u01be\u01bd\3\2\2\2\u01be")             # <<<<<<<<<<<<<<
 *         buf.write("\u01bf\3\2\2\2\u01bf\u01c1\3\2\2\2\u01c0\u01c2\58\35\2")
 *         buf.write("\u01c1\u01c0\3\2\2\2\u01c1\u01c2\3\2\2\2\u01c2\67\3\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 265, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__418, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 265, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":266
 *         buf.write("*\2\2\u01ba\u01bb\5n8\2\u01bb\65\3\2\2\2\u01bc\u01be\7")
 *         buf.write("Z\2\2\u01bd\u01bf\5\u008aF\2\u01be\u01bd\3\2\2\2\u01be")
 *         buf.write("\u01bf\3\2\2\2\u01bf\u01c1\3\2\2\2\u01c0\u01c2\58\35\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u01c1\u01c0\3\2\2\2\u01c1\u01c2\3\2\2\2\u01c2\67\3\2")
 *         buf.write("\2\2\u01c3\u01c6\5:\36\2\u01c4\u01c5\7\21\2\2\u01c5\u01c7")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 266, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__419, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 266, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":267
 *         buf.write("Z\2\2\u01bd\u01bf\5\u008aF\2\u01be\u01bd\3\2\2\2\u01be")
 *         buf.write("\u01bf\3\2\2\2\u01bf\u01c1\3\2\2\2\u01c0\u01c2\58\35\2")
 *         buf.write("\u01c1\u01c0\3\2\2\2\u01c1\u01c2\3\2\2\2\u01c2\67\3\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\u01c3\u01c6\5:\36\2\u01c4\u01c5\7\21\2\2\u01c5\u01c7")
 *         buf.write("\5n8\2\u01c6\u01c4\3\2\2\2\u01c6\u01c7\3\2\2\2\u01c7\u01cd")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 267, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__420, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 267, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":268
 *         buf.write("\u01bf\3\2\2\2\u01bf\u01c1\3\2\2\2\u01c0\u01c2\58\35\2")
 *         buf.write("\u01c1\u01c0\3\2\2\2\u01c1\u01c2\3\2\2\2\u01c2\67\3\2")
 *         buf.write("\2\2\u01c3\u01c6\5:\36\2\u01c4\u01c5\7\21\2\2\u01c5\u01c7")             # <<<<<<<<<<<<<<
 *         buf.write("\5n8\2\u01c6\u01c4\3\2\2\2\u01c6\u01c7\3\2\2\2\u01c7\u01cd")
 *         buf.write("\3\2\2\2\u01c8\u01c9\7\21\2\2\u01c9\u01cd\5n8\2\u01ca")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 268, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__422, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 268, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":269
 *         buf.write("\u01c1\u01c0\3\2\2\2\u01c1\u01c2\3\2\2\2\u01c2\67\3\2")
 *         buf.write("\2\2\u01c3\u01c6\5:\36\2\u01c4\u01c5\7\21\2\2\u01c5\u01c7")
 *         buf.write("\5n8\2\u01c6\u01c4\3\2\2\2\u01c6\u01c7\3\2\2\2\u01c7\u01cd")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u01c8\u01c9\7\21\2\2\u01c9\u01cd\5n8\2\u01ca")
 *         buf.write("\u01cb\7+\2\2\u01cb\u01cd\5n8\2\u01cc\u01c3\3\2\2\2\u01cc")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 269, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__423, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 269, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":270
 *         buf.write("\2\2\u01c3\u01c6\5:\36\2\u01c4\u01c5\7\21\2\2\u01c5\u01c7")
 *         buf.write("\5n8\2\u01c6\u01c4\3\2\2\2\u01c6\u01c7\3\2\2\2\u01c7\u01cd")
 *         buf.write("\3\2\2\2\u01c8\u01c9\7\21\2\2\u01c9\u01cd\5n8\2\u01ca")             # <<<<<<<<<<<<<<
 *         buf.write("\u01cb\7+\2\2\u01cb\u01cd\5n8\2\u01cc\u01c3\3\2\2\2\u01cc")
 *         buf.write("\u01c8\3\2\2\2\u01cc\u01ca\3\2\2\2\u01cd9\3\2\2\2\u01ce")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 270, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__424, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 270, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":271
 *         buf.write("\5n8\2\u01c6\u01c4\3\2\2\2\u01c6\u01c7\3\2\2\2\u01c7\u01cd")
 *         buf.write("\3\2\2\2\u01c8\u01c9\7\21\2\2\u01c9\u01cd\5n8\2\u01ca")
 *         buf.write("\u01cb\7+\2\2\u01cb\u01cd\5n8\2\u01cc\u01c3\3\2\2\2\u01cc")             # <<<<<<<<<<<<<<
 *         buf.write("\u01c8\3\2\2\2\u01cc\u01ca\3\2\2\2\u01cd9\3\2\2\2\u01ce")
 *         buf.write("\u01d0\7\23\2\2\u01cf\u01d1\5<\37\2\u01d0\u01cf\3\2\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 271, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__425, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 271, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":272
 *         buf.write("\3\2\2\2\u01c8\u01c9\7\21\2\2\u01c9\u01cd\5n8\2\u01ca")
 *         buf.write("\u01cb\7+\2\2\u01cb\u01cd\5n8\2\u01cc\u01c3\3\2\2\2\u01cc")
 *         buf.write("\u01c8\3\2\2\2\u01cc\u01ca\3\2\2\2\u01cd9\3\2\2\2\u01ce")             # <<<<<<<<<<<<<<
 *         buf.write("\u01d0\7\23\2\2\u01cf\u01d1\5<\37\2\u01d0\u01cf\3\2\2")
 *         buf.write("\2\u01d0\u01d1\3\2\2\2\u01d1\u01d2\3\2\2\2\u01d2\u01d3")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 272, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__426, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 272, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":273
 *         buf.write("\u01cb\7+\2\2\u01cb\u01cd\5n8\2\u01cc\u01c3\3\2\2\2\u01cc")
 *         buf.write("\u01c8\3\2\2\2\u01cc\u01ca\3\2\2\2\u01cd9\3\2\2\2\u01ce")
 *         buf.write("\u01d0\7\23\2\2\u01cf\u01d1\5<\37\2\u01d0\u01cf\3\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u01d0\u01d1\3\2\2\2\u01d1\u01d2\3\2\2\2\u01d2\u01d3")
 *         buf.write("\7\25\2\2\u01d3;\3\2\2\2\u01d4\u01d9\5> \2\u01d5\u01d6")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 273, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__428, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 273, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":274
 *         buf.write("\u01c8\3\2\2\2\u01cc\u01ca\3\2\2\2\u01cd9\3\2\2\2\u01ce")
 *         buf.write("\u01d0\7\23\2\2\u01cf\u01d1\5<\37\2\u01d0\u01cf\3\2\2")
 *         buf.write("\2\u01d0\u01d1\3\2\2\2\u01d1\u01d2\3\2\2\2\u01d2\u01d3")             # <<<<<<<<<<<<<<
 *         buf.write("\7\25\2\2\u01d3;\3\2\2\2\u01d4\u01d9\5> \2\u01d5\u01d6")
 *         buf.write("\7\27\2\2\u01d6\u01d8\5> \2\u01d7\u01d5\3\2\2\2\u01d8")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 274, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__430, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 274, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":275
 *         buf.write("\u01d0\7\23\2\2\u01cf\u01d1\5<\37\2\u01d0\u01cf\3\2\2")
 *         buf.write("\2\u01d0\u01d1\3\2\2\2\u01d1\u01d2\3\2\2\2\u01d2\u01d3")
 *         buf.write("\7\25\2\2\u01d3;\3\2\2\2\u01d4\u01d9\5> \2\u01d5\u01d6")             # <<<<<<<<<<<<<<
 *         buf.write("\7\27\2\2\u01d6\u01d8\5> \2\u01d7\u01d5\3\2\2\2\u01d8")
 *         buf.write("\u01db\3\2\2\2\u01d9\u01d7\3\2\2\2\u01d9\u01da\3\2\2\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 275, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__432, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 275, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":276
 *         buf.write("\2\u01d0\u01d1\3\2\2\2\u01d1\u01d2\3\2\2\2\u01d2\u01d3")
 *         buf.write("\7\25\2\2\u01d3;\3\2\2\2\u01d4\u01d9\5> \2\u01d5\u01d6")
 *         buf.write("\7\27\2\2\u01d6\u01d8\5> \2\u01d7\u01d5\3\2\2\2\u01d8")             # <<<<<<<<<<<<<<
 *         buf.write("\u01db\3\2\2\2\u01d9\u01d7\3\2\2\2\u01d9\u01da\3\2\2\2")
 *         buf.write("\u01da=\3\2\2\2\u01db\u01d9\3\2\2\2\u01dc\u01df\5@!\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 276, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__434, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 276, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":277
 *         buf.write("\7\25\2\2\u01d3;\3\2\2\2\u01d4\u01d9\5> \2\u01d5\u01d6")
 *         buf.write("\7\27\2\2\u01d6\u01d8\5> \2\u01d7\u01d5\3\2\2\2\u01d8")
 *         buf.write("\u01db\3\2\2\2\u01d9\u01d7\3\2\2\2\u01d9\u01da\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u01da=\3\2\2\2\u01db\u01d9\3\2\2\2\u01dc\u01df\5@!\2")
 *         buf.write("\u01dd\u01df\5D#\2\u01de\u01dc\3\2\2\2\u01de\u01dd\3\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 277, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__436, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 277, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":278
 *         buf.write("\7\27\2\2\u01d6\u01d8\5> \2\u01d7\u01d5\3\2\2\2\u01d8")
 *         buf.write("\u01db\3\2\2\2\u01d9\u01d7\3\2\2\2\u01d9\u01da\3\2\2\2")
 *         buf.write("\u01da=\3\2\2\2\u01db\u01d9\3\2\2\2\u01dc\u01df\5@!\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u01dd\u01df\5D#\2\u01de\u01dc\3\2\2\2\u01de\u01dd\3\2")
 *         buf.write("\2\2\u01df?\3\2\2\2\u01e0\u01e2\7Q\2\2\u01e1\u01e0\3\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 278, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__438, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 278, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":279
 *         buf.write("\u01db\3\2\2\2\u01d9\u01d7\3\2\2\2\u01d9\u01da\3\2\2\2")
 *         buf.write("\u01da=\3\2\2\2\u01db\u01d9\3\2\2\2\u01dc\u01df\5@!\2")
 *         buf.write("\u01dd\u01df\5D#\2\u01de\u01dc\3\2\2\2\u01de\u01dd\3\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\u01df?\3\2\2\2\u01e0\u01e2\7Q\2\2\u01e1\u01e0\3\2")
 *         buf.write("\2\2\u01e1\u01e2\3\2\2\2\u01e2\u01e4\3\2\2\2\u01e3\u01e5")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 279, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__439, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 279, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":280
 *         buf.write("\u01da=\3\2\2\2\u01db\u01d9\3\2\2\2\u01dc\u01df\5@!\2")
 *         buf.write("\u01dd\u01df\5D#\2\u01de\u01dc\3\2\2\2\u01de\u01dd\3\2")
 *         buf.write("\2\2\u01df?\3\2\2\2\u01e0\u01e2\7Q\2\2\u01e1\u01e0\3\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\u01e1\u01e2\3\2\2\2\u01e2\u01e4\3\2\2\2\u01e3\u01e5")
 *         buf.write("\7S\2\2\u01e4\u01e3\3\2\2\2\u01e4\u01e5\3\2\2\2\u01e5")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 280, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__440, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 280, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":281
 *         buf.write("\u01dd\u01df\5D#\2\u01de\u01dc\3\2\2\2\u01de\u01dd\3\2")
 *         buf.write("\2\2\u01df?\3\2\2\2\u01e0\u01e2\7Q\2\2\u01e1\u01e0\3\2")
 *         buf.write("\2\2\u01e1\u01e2\3\2\2\2\u01e2\u01e4\3\2\2\2\u01e3\u01e5")             # <<<<<<<<<<<<<<
 *         buf.write("\7S\2\2\u01e4\u01e3\3\2\2\2\u01e4\u01e5\3\2\2\2\u01e5")
 *         buf.write("\u01e8\3\2\2\2\u01e6\u01e9\5B\"\2\u01e7\u01e9\5F$\2\u01e8")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 281, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__442, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 281, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":282
 *         buf.write("\2\2\u01df?\3\2\2\2\u01e0\u01e2\7Q\2\2\u01e1\u01e0\3\2")
 *         buf.write("\2\2\u01e1\u01e2\3\2\2\2\u01e2\u01e4\3\2\2\2\u01e3\u01e5")
 *         buf.write("\7S\2\2\u01e4\u01e3\3\2\2\2\u01e4\u01e5\3\2\2\2\u01e5")             # <<<<<<<<<<<<<<
 *         buf.write("\u01e8\3\2\2\2\u01e6\u01e9\5B\"\2\u01e7\u01e9\5F$\2\u01e8")
 *         buf.write("\u01e6\3\2\2\2\u01e8\u01e7\3\2\2\2\u01e9A\3\2\2\2\u01ea")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 282, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__443, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 282, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":283
 *         buf.write("\2\2\u01e1\u01e2\3\2\2\2\u01e2\u01e4\3\2\2\2\u01e3\u01e5")
 *         buf.write("\7S\2\2\u01e4\u01e3\3\2\2\2\u01e4\u01e5\3\2\2\2\u01e5")
 *         buf.write("\u01e8\3\2\2\2\u01e6\u01e9\5B\"\2\u01e7\u01e9\5F$\2\u01e8")             # <<<<<<<<<<<<<<
 *         buf.write("\u01e6\3\2\2\2\u01e8\u01e7\3\2\2\2\u01e9A\3\2\2\2\u01ea")
 *         buf.write("\u01ec\5z>\2\u01eb\u01ed\58\35\2\u01ec\u01eb\3\2\2\2\u01ec")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 283, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__444, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 283, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":284
 *         buf.write("\7S\2\2\u01e4\u01e3\3\2\2\2\u01e4\u01e5\3\2\2\2\u01e5")
 *         buf.write("\u01e8\3\2\2\2\u01e6\u01e9\5B\"\2\u01e7\u01e9\5F$\2\u01e8")
 *         buf.write("\u01e6\3\2\2\2\u01e8\u01e7\3\2\2\2\u01e9A\3\2\2\2\u01ea")             # <<<<<<<<<<<<<<
 *         buf.write("\u01ec\5z>\2\u01eb\u01ed\58\35\2\u01ec\u01eb\3\2\2\2\u01ec")
 *         buf.write("\u01ed\3\2\2\2\u01ed\u01ee\3\2\2\2\u01ee\u01ef\5\u0090")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 284, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__445, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 284, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":285
 *         buf.write("\u01e8\3\2\2\2\u01e6\u01e9\5B\"\2\u01e7\u01e9\5F$\2\u01e8")
 *         buf.write("\u01e6\3\2\2\2\u01e8\u01e7\3\2\2\2\u01e9A\3\2\2\2\u01ea")
 *         buf.write("\u01ec\5z>\2\u01eb\u01ed\58\35\2\u01ec\u01eb\3\2\2\2\u01ec")             # <<<<<<<<<<<<<<
 *         buf.write("\u01ed\3\2\2\2\u01ed\u01ee\3\2\2\2\u01ee\u01ef\5\u0090")
 *         buf.write("I\2\u01efC\3\2\2\2\u01f0\u01f2\7V\2\2\u01f1\u01f3\7Q\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 285, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__446, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 285, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":286
 *         buf.write("\u01e6\3\2\2\2\u01e8\u01e7\3\2\2\2\u01e9A\3\2\2\2\u01ea")
 *         buf.write("\u01ec\5z>\2\u01eb\u01ed\58\35\2\u01ec\u01eb\3\2\2\2\u01ec")
 *         buf.write("\u01ed\3\2\2\2\u01ed\u01ee\3\2\2\2\u01ee\u01ef\5\u0090")             # <<<<<<<<<<<<<<
 *         buf.write("I\2\u01efC\3\2\2\2\u01f0\u01f2\7V\2\2\u01f1\u01f3\7Q\2")
 *         buf.write("\2\u01f2\u01f1\3\2\2\2\u01f2\u01f3\3\2\2\2\u01f3\u01f5")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 286, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__448, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 286, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":287
 *         buf.write("\u01ec\5z>\2\u01eb\u01ed\58\35\2\u01ec\u01eb\3\2\2\2\u01ec")
 *         buf.write("\u01ed\3\2\2\2\u01ed\u01ee\3\2\2\2\u01ee\u01ef\5\u0090")
 *         buf.write("I\2\u01efC\3\2\2\2\u01f0\u01f2\7V\2\2\u01f1\u01f3\7Q\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u01f2\u01f1\3\2\2\2\u01f2\u01f3\3\2\2\2\u01f3\u01f5")
 *         buf.write("\3\2\2\2\u01f4\u01f6\7S\2\2\u01f5\u01f4\3\2\2\2\u01f5")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 287, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__449, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 287, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":288
 *         buf.write("\u01ed\3\2\2\2\u01ed\u01ee\3\2\2\2\u01ee\u01ef\5\u0090")
 *         buf.write("I\2\u01efC\3\2\2\2\u01f0\u01f2\7V\2\2\u01f1\u01f3\7Q\2")
 *         buf.write("\2\u01f2\u01f1\3\2\2\2\u01f2\u01f3\3\2\2\2\u01f3\u01f5")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u01f4\u01f6\7S\2\2\u01f5\u01f4\3\2\2\2\u01f5")
 *         buf.write("\u01f6\3\2\2\2\u01f6\u01fc\3\2\2\2\u01f7\u01fa\5L\'\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 288, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__451, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 288, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":289
 *         buf.write("I\2\u01efC\3\2\2\2\u01f0\u01f2\7V\2\2\u01f1\u01f3\7Q\2")
 *         buf.write("\2\u01f2\u01f1\3\2\2\2\u01f2\u01f3\3\2\2\2\u01f3\u01f5")
 *         buf.write("\3\2\2\2\u01f4\u01f6\7S\2\2\u01f5\u01f4\3\2\2\2\u01f5")             # <<<<<<<<<<<<<<
 *         buf.write("\u01f6\3\2\2\2\u01f6\u01fc\3\2\2\2\u01f7\u01fa\5L\'\2")
 *         buf.write("\u01f8\u01fa\5H%\2\u01f9\u01f7\3\2\2\2\u01f9\u01f8\3\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 289, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__452, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 289, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":290
 *         buf.write("\2\u01f2\u01f1\3\2\2\2\u01f2\u01f3\3\2\2\2\u01f3\u01f5")
 *         buf.write("\3\2\2\2\u01f4\u01f6\7S\2\2\u01f5\u01f4\3\2\2\2\u01f5")
 *         buf.write("\u01f6\3\2\2\2\u01f6\u01fc\3\2\2\2\u01f7\u01fa\5L\'\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u01f8\u01fa\5H%\2\u01f9\u01f7\3\2\2\2\u01f9\u01f8\3\2")
 *         buf.write("\2\2\u01fa\u01fd\3\2\2\2\u01fb\u01fd\5F$\2\u01fc\u01f9")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 290, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__453, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 290, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":291
 *         buf.write("\3\2\2\2\u01f4\u01f6\7S\2\2\u01f5\u01f4\3\2\2\2\u01f5")
 *         buf.write("\u01f6\3\2\2\2\u01f6\u01fc\3\2\2\2\u01f7\u01fa\5L\'\2")
 *         buf.write("\u01f8\u01fa\5H%\2\u01f9\u01f7\3\2\2\2\u01f9\u01f8\3\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\u01fa\u01fd\3\2\2\2\u01fb\u01fd\5F$\2\u01fc\u01f9")
 *         buf.write("\3\2\2\2\u01fc\u01fb\3\2\2\2\u01fdE\3\2\2\2\u01fe\u0201")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 291, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__454, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 291, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":292
 *         buf.write("\u01f6\3\2\2\2\u01f6\u01fc\3\2\2\2\u01f7\u01fa\5L\'\2")
 *         buf.write("\u01f8\u01fa\5H%\2\u01f9\u01f7\3\2\2\2\u01f9\u01f8\3\2")
 *         buf.write("\2\2\u01fa\u01fd\3\2\2\2\u01fb\u01fd\5F$\2\u01fc\u01f9")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u01fc\u01fb\3\2\2\2\u01fdE\3\2\2\2\u01fe\u0201")
 *         buf.write("\7\34\2\2\u01ff\u0202\5L\'\2\u0200\u0202\5H%\2\u0201\u01ff")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 292, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__455, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 292, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":293
 *         buf.write("\u01f8\u01fa\5H%\2\u01f9\u01f7\3\2\2\2\u01f9\u01f8\3\2")
 *         buf.write("\2\2\u01fa\u01fd\3\2\2\2\u01fb\u01fd\5F$\2\u01fc\u01f9")
 *         buf.write("\3\2\2\2\u01fc\u01fb\3\2\2\2\u01fdE\3\2\2\2\u01fe\u0201")             # <<<<<<<<<<<<<<
 *         buf.write("\7\34\2\2\u01ff\u0202\5L\'\2\u0200\u0202\5H%\2\u0201\u01ff")
 *         buf.write("\3\2\2\2\u0201\u0200\3\2\2\2\u0202\u0204\3\2\2\2\u0203")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 293, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__456, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 293, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":294
 *         buf.write("\2\2\u01fa\u01fd\3\2\2\2\u01fb\u01fd\5F$\2\u01fc\u01f9")
 *         buf.write("\3\2\2\2\u01fc\u01fb\3\2\2\2\u01fdE\3\2\2\2\u01fe\u0201")
 *         buf.write("\7\34\2\2\u01ff\u0202\5L\'\2\u0200\u0202\5H%\2\u0201\u01ff")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0201\u0200\3\2\2\2\u0202\u0204\3\2\2\2\u0203")
 *         buf.write("\u0205\5(\25\2\u0204\u0203\3\2\2\2\u0204\u0205\3\2\2\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 294, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__457, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 294, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":295
 *         buf.write("\3\2\2\2\u01fc\u01fb\3\2\2\2\u01fdE\3\2\2\2\u01fe\u0201")
 *         buf.write("\7\34\2\2\u01ff\u0202\5L\'\2\u0200\u0202\5H%\2\u0201\u01ff")
 *         buf.write("\3\2\2\2\u0201\u0200\3\2\2\2\u0202\u0204\3\2\2\2\u0203")             # <<<<<<<<<<<<<<
 *         buf.write("\u0205\5(\25\2\u0204\u0203\3\2\2\2\u0204\u0205\3\2\2\2")
 *         buf.write("\u0205G\3\2\2\2\u0206\u0207\5,\27\2\u0207\u0208\5.\30")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 295, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__459, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 295, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":296
 *         buf.write("\7\34\2\2\u01ff\u0202\5L\'\2\u0200\u0202\5H%\2\u0201\u01ff")
 *         buf.write("\3\2\2\2\u0201\u0200\3\2\2\2\u0202\u0204\3\2\2\2\u0203")
 *         buf.write("\u0205\5(\25\2\u0204\u0203\3\2\2\2\u0204\u0205\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0205G\3\2\2\2\u0206\u0207\5,\27\2\u0207\u0208\5.\30")
 *         buf.write("\2\u0208\u0209\5J&\2\u0209I\3\2\2\2\u020a\u020b\5\66\34")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 296, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__461, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 296, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":297
 *         buf.write("\3\2\2\2\u0201\u0200\3\2\2\2\u0202\u0204\3\2\2\2\u0203")
 *         buf.write("\u0205\5(\25\2\u0204\u0203\3\2\2\2\u0204\u0205\3\2\2\2")
 *         buf.write("\u0205G\3\2\2\2\u0206\u0207\5,\27\2\u0207\u0208\5.\30")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u0208\u0209\5J&\2\u0209I\3\2\2\2\u020a\u020b\5\66\34")
 *         buf.write("\2\u020b\u020c\5\u008eH\2\u020cK\3\2\2\2\u020d\u020e\5")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 297, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__462, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 297, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":298
 *         buf.write("\u0205\5(\25\2\u0204\u0203\3\2\2\2\u0204\u0205\3\2\2\2")
 *         buf.write("\u0205G\3\2\2\2\u0206\u0207\5,\27\2\u0207\u0208\5.\30")
 *         buf.write("\2\u0208\u0209\5J&\2\u0209I\3\2\2\2\u020a\u020b\5\66\34")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u020b\u020c\5\u008eH\2\u020cK\3\2\2\2\u020d\u020e\5")
 *         buf.write("\b\5\2\u020e\u020f\7Z\2\2\u020f\u0224\7\21\2\2\u0210\u0211")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 298, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__463, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 298, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":299
 *         buf.write("\u0205G\3\2\2\2\u0206\u0207\5,\27\2\u0207\u0208\5.\30")
 *         buf.write("\2\u0208\u0209\5J&\2\u0209I\3\2\2\2\u020a\u020b\5\66\34")
 *         buf.write("\2\u020b\u020c\5\u008eH\2\u020cK\3\2\2\2\u020d\u020e\5")             # <<<<<<<<<<<<<<
 *         buf.write("\b\5\2\u020e\u020f\7Z\2\2\u020f\u0224\7\21\2\2\u0210\u0211")
 *         buf.write("\5\16\b\2\u0211\u0213\5z>\2\u0212\u0214\5\u008aF\2\u0213")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 299, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__464, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 299, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":300
 *         buf.write("\2\u0208\u0209\5J&\2\u0209I\3\2\2\2\u020a\u020b\5\66\34")
 *         buf.write("\2\u020b\u020c\5\u008eH\2\u020cK\3\2\2\2\u020d\u020e\5")
 *         buf.write("\b\5\2\u020e\u020f\7Z\2\2\u020f\u0224\7\21\2\2\u0210\u0211")             # <<<<<<<<<<<<<<
 *         buf.write("\5\16\b\2\u0211\u0213\5z>\2\u0212\u0214\5\u008aF\2\u0213")
 *         buf.write("\u0212\3\2\2\2\u0213\u0214\3\2\2\2\u0214\u0216\3\2\2\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 300, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__465, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 300, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":301
 *         buf.write("\2\u020b\u020c\5\u008eH\2\u020cK\3\2\2\2\u020d\u020e\5")
 *         buf.write("\b\5\2\u020e\u020f\7Z\2\2\u020f\u0224\7\21\2\2\u0210\u0211")
 *         buf.write("\5\16\b\2\u0211\u0213\5z>\2\u0212\u0214\5\u008aF\2\u0213")             # <<<<<<<<<<<<<<
 *         buf.write("\u0212\3\2\2\2\u0213\u0214\3\2\2\2\u0214\u0216\3\2\2\2")
 *         buf.write("\u0215\u0217\5:\36\2\u0216\u0215\3\2\2\2\u0216\u0217\3")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 301, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__466, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 301, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":302
 *         buf.write("\b\5\2\u020e\u020f\7Z\2\2\u020f\u0224\7\21\2\2\u0210\u0211")
 *         buf.write("\5\16\b\2\u0211\u0213\5z>\2\u0212\u0214\5\u008aF\2\u0213")
 *         buf.write("\u0212\3\2\2\2\u0213\u0214\3\2\2\2\u0214\u0216\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0215\u0217\5:\36\2\u0216\u0215\3\2\2\2\u0216\u0217\3")
 *         buf.write("\2\2\2\u0217\u0218\3\2\2\2\u0218\u0219\5\u008eH\2\u0219")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 302, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__468, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 302, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":303
 *         buf.write("\5\16\b\2\u0211\u0213\5z>\2\u0212\u0214\5\u008aF\2\u0213")
 *         buf.write("\u0212\3\2\2\2\u0213\u0214\3\2\2\2\u0214\u0216\3\2\2\2")
 *         buf.write("\u0215\u0217\5:\36\2\u0216\u0215\3\2\2\2\u0216\u0217\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u0217\u0218\3\2\2\2\u0218\u0219\5\u008eH\2\u0219")
 *         buf.write("\u0225\3\2\2\2\u021a\u021b\7\22\2\2\u021b\u0220\7\23\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 303, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__470, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 303, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":304
 *         buf.write("\u0212\3\2\2\2\u0213\u0214\3\2\2\2\u0214\u0216\3\2\2\2")
 *         buf.write("\u0215\u0217\5:\36\2\u0216\u0215\3\2\2\2\u0216\u0217\3")
 *         buf.write("\2\2\2\u0217\u0218\3\2\2\2\u0218\u0219\5\u008eH\2\u0219")             # <<<<<<<<<<<<<<
 *         buf.write("\u0225\3\2\2\2\u021a\u021b\7\22\2\2\u021b\u0220\7\23\2")
 *         buf.write("\2\u021c\u021e\5\20\t\2\u021d\u021c\3\2\2\2\u021d\u021e")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 304, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__471, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 304, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":305
 *         buf.write("\u0215\u0217\5:\36\2\u0216\u0215\3\2\2\2\u0216\u0217\3")
 *         buf.write("\2\2\2\u0217\u0218\3\2\2\2\u0218\u0219\5\u008eH\2\u0219")
 *         buf.write("\u0225\3\2\2\2\u021a\u021b\7\22\2\2\u021b\u0220\7\23\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u021c\u021e\5\20\t\2\u021d\u021c\3\2\2\2\u021d\u021e")
 *         buf.write("\3\2\2\2\u021e\u0221\3\2\2\2\u021f\u0221\7\24\2\2\u0220")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 305, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__473, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 305, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":306
 *         buf.write("\2\2\2\u0217\u0218\3\2\2\2\u0218\u0219\5\u008eH\2\u0219")
 *         buf.write("\u0225\3\2\2\2\u021a\u021b\7\22\2\2\u021b\u0220\7\23\2")
 *         buf.write("\2\u021c\u021e\5\20\t\2\u021d\u021c\3\2\2\2\u021d\u021e")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u021e\u0221\3\2\2\2\u021f\u0221\7\24\2\2\u0220")
 *         buf.write("\u021d\3\2\2\2\u0220\u021f\3\2\2\2\u0221\u0222\3\2\2\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 306, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__475, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 306, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":307
 *         buf.write("\u0225\3\2\2\2\u021a\u021b\7\22\2\2\u021b\u0220\7\23\2")
 *         buf.write("\2\u021c\u021e\5\20\t\2\u021d\u021c\3\2\2\2\u021d\u021e")
 *         buf.write("\3\2\2\2\u021e\u0221\3\2\2\2\u021f\u0221\7\24\2\2\u0220")             # <<<<<<<<<<<<<<
 *         buf.write("\u021d\3\2\2\2\u0220\u021f\3\2\2\2\u0221\u0222\3\2\2\2")
 *         buf.write("\u0222\u0223\7\25\2\2\u0223\u0225\5\u008eH\2\u0224\u0210")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 307, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__477, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 307, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":308
 *         buf.write("\2\u021c\u021e\5\20\t\2\u021d\u021c\3\2\2\2\u021d\u021e")
 *         buf.write("\3\2\2\2\u021e\u0221\3\2\2\2\u021f\u0221\7\24\2\2\u0220")
 *         buf.write("\u021d\3\2\2\2\u0220\u021f\3\2\2\2\u0221\u0222\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0222\u0223\7\25\2\2\u0223\u0225\5\u008eH\2\u0224\u0210")
 *         buf.write("\3\2\2\2\u0224\u021a\3\2\2\2\u0225M\3\2\2\2\u0226\u0228")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 308, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__479, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 308, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":309
 *         buf.write("\3\2\2\2\u021e\u0221\3\2\2\2\u021f\u0221\7\24\2\2\u0220")
 *         buf.write("\u021d\3\2\2\2\u0220\u021f\3\2\2\2\u0221\u0222\3\2\2\2")
 *         buf.write("\u0222\u0223\7\25\2\2\u0223\u0225\5\u008eH\2\u0224\u0210")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0224\u021a\3\2\2\2\u0225M\3\2\2\2\u0226\u0228")
 *         buf.write("\7Y\2\2\u0227\u0226\3\2\2\2\u0227\u0228\3\2\2\2\u0228")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 309, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__480, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 309, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":310
 *         buf.write("\u021d\3\2\2\2\u0220\u021f\3\2\2\2\u0221\u0222\3\2\2\2")
 *         buf.write("\u0222\u0223\7\25\2\2\u0223\u0225\5\u008eH\2\u0224\u0210")
 *         buf.write("\3\2\2\2\u0224\u021a\3\2\2\2\u0225M\3\2\2\2\u0226\u0228")             # <<<<<<<<<<<<<<
 *         buf.write("\7Y\2\2\u0227\u0226\3\2\2\2\u0227\u0228\3\2\2\2\u0228")
 *         buf.write("\u0229\3\2\2\2\u0229\u022f\7,\2\2\u022a\u022b\5T+\2\u022b")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 310, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__481, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 310, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":311
 *         buf.write("\u0222\u0223\7\25\2\2\u0223\u0225\5\u008eH\2\u0224\u0210")
 *         buf.write("\3\2\2\2\u0224\u021a\3\2\2\2\u0225M\3\2\2\2\u0226\u0228")
 *         buf.write("\7Y\2\2\u0227\u0226\3\2\2\2\u0227\u0228\3\2\2\2\u0228")             # <<<<<<<<<<<<<<
 *         buf.write("\u0229\3\2\2\2\u0229\u022f\7,\2\2\u022a\u022b\5T+\2\u022b")
 *         buf.write("\u022c\7\3\2\2\u022c\u022e\3\2\2\2\u022d\u022a\3\2\2\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 311, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__482, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 311, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":312
 *         buf.write("\3\2\2\2\u0224\u021a\3\2\2\2\u0225M\3\2\2\2\u0226\u0228")
 *         buf.write("\7Y\2\2\u0227\u0226\3\2\2\2\u0227\u0228\3\2\2\2\u0228")
 *         buf.write("\u0229\3\2\2\2\u0229\u022f\7,\2\2\u022a\u022b\5T+\2\u022b")             # <<<<<<<<<<<<<<
 *         buf.write("\u022c\7\3\2\2\u022c\u022e\3\2\2\2\u022d\u022a\3\2\2\2")
 *         buf.write("\u022e\u0231\3\2\2\2\u022f\u022d\3\2\2\2\u022f\u0230\3")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 312, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__483, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 312, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":313
 *         buf.write("\7Y\2\2\u0227\u0226\3\2\2\2\u0227\u0228\3\2\2\2\u0228")
 *         buf.write("\u0229\3\2\2\2\u0229\u022f\7,\2\2\u022a\u022b\5T+\2\u022b")
 *         buf.write("\u022c\7\3\2\2\u022c\u022e\3\2\2\2\u022d\u022a\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u022e\u0231\3\2\2\2\u022f\u022d\3\2\2\2\u022f\u0230\3")
 *         buf.write("\2\2\2\u0230O\3\2\2\2\u0231\u022f\3\2\2\2\u0232\u0234")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 313, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__485, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 313, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":314
 *         buf.write("\u0229\3\2\2\2\u0229\u022f\7,\2\2\u022a\u022b\5T+\2\u022b")
 *         buf.write("\u022c\7\3\2\2\u022c\u022e\3\2\2\2\u022d\u022a\3\2\2\2")
 *         buf.write("\u022e\u0231\3\2\2\2\u022f\u022d\3\2\2\2\u022f\u0230\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u0230O\3\2\2\2\u0231\u022f\3\2\2\2\u0232\u0234")
 *         buf.write("\7Y\2\2\u0233\u0232\3\2\2\2\u0233\u0234\3\2\2\2\u0234")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 314, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__487, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 314, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":315
 *         buf.write("\u022c\7\3\2\2\u022c\u022e\3\2\2\2\u022d\u022a\3\2\2\2")
 *         buf.write("\u022e\u0231\3\2\2\2\u022f\u022d\3\2\2\2\u022f\u0230\3")
 *         buf.write("\2\2\2\u0230O\3\2\2\2\u0231\u022f\3\2\2\2\u0232\u0234")             # <<<<<<<<<<<<<<
 *         buf.write("\7Y\2\2\u0233\u0232\3\2\2\2\u0233\u0234\3\2\2\2\u0234")
 *         buf.write("\u0235\3\2\2\2\u0235\u023b\7-\2\2\u0236\u0237\5X-\2\u0237")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 315, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__488, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 315, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":316
 *         buf.write("\u022e\u0231\3\2\2\2\u022f\u022d\3\2\2\2\u022f\u0230\3")
 *         buf.write("\2\2\2\u0230O\3\2\2\2\u0231\u022f\3\2\2\2\u0232\u0234")
 *         buf.write("\7Y\2\2\u0233\u0232\3\2\2\2\u0233\u0234\3\2\2\2\u0234")             # <<<<<<<<<<<<<<
 *         buf.write("\u0235\3\2\2\2\u0235\u023b\7-\2\2\u0236\u0237\5X-\2\u0237")
 *         buf.write("\u0238\7\3\2\2\u0238\u023a\3\2\2\2\u0239\u0236\3\2\2\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 316, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__489, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 316, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":317
 *         buf.write("\2\2\2\u0230O\3\2\2\2\u0231\u022f\3\2\2\2\u0232\u0234")
 *         buf.write("\7Y\2\2\u0233\u0232\3\2\2\2\u0233\u0234\3\2\2\2\u0234")
 *         buf.write("\u0235\3\2\2\2\u0235\u023b\7-\2\2\u0236\u0237\5X-\2\u0237")             # <<<<<<<<<<<<<<
 *         buf.write("\u0238\7\3\2\2\u0238\u023a\3\2\2\2\u0239\u0236\3\2\2\2")
 *         buf.write("\u023a\u023d\3\2\2\2\u023b\u0239\3\2\2\2\u023b\u023c\3")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 317, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__490, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 317, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":318
 *         buf.write("\7Y\2\2\u0233\u0232\3\2\2\2\u0233\u0234\3\2\2\2\u0234")
 *         buf.write("\u0235\3\2\2\2\u0235\u023b\7-\2\2\u0236\u0237\5X-\2\u0237")
 *         buf.write("\u0238\7\3\2\2\u0238\u023a\3\2\2\2\u0239\u0236\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u023a\u023d\3\2\2\2\u023b\u0239\3\2\2\2\u023b\u023c\3")
 *         buf.write("\2\2\2\u023cQ\3\2\2\2\u023d\u023b\3\2\2\2\u023e\u023f")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 318, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__492, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 318, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":319
 *         buf.write("\u0235\3\2\2\2\u0235\u023b\7-\2\2\u0236\u0237\5X-\2\u0237")
 *         buf.write("\u0238\7\3\2\2\u0238\u023a\3\2\2\2\u0239\u0236\3\2\2\2")
 *         buf.write("\u023a\u023d\3\2\2\2\u023b\u0239\3\2\2\2\u023b\u023c\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u023cQ\3\2\2\2\u023d\u023b\3\2\2\2\u023e\u023f")
 *         buf.write("\5p9\2\u023f\u0240\7\21\2\2\u0240\u0241\5n8\2\u0241\u024a")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 319, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__494, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 319, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":320
 *         buf.write("\u0238\7\3\2\2\u0238\u023a\3\2\2\2\u0239\u0236\3\2\2\2")
 *         buf.write("\u023a\u023d\3\2\2\2\u023b\u0239\3\2\2\2\u023b\u023c\3")
 *         buf.write("\2\2\2\u023cQ\3\2\2\2\u023d\u023b\3\2\2\2\u023e\u023f")             # <<<<<<<<<<<<<<
 *         buf.write("\5p9\2\u023f\u0240\7\21\2\2\u0240\u0241\5n8\2\u0241\u024a")
 *         buf.write("\3\2\2\2\u0242\u024a\5Z.\2\u0243\u024a\5^\60\2\u0244\u024a")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 320, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__495, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 320, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":321
 *         buf.write("\u023a\u023d\3\2\2\2\u023b\u0239\3\2\2\2\u023b\u023c\3")
 *         buf.write("\2\2\2\u023cQ\3\2\2\2\u023d\u023b\3\2\2\2\u023e\u023f")
 *         buf.write("\5p9\2\u023f\u0240\7\21\2\2\u0240\u0241\5n8\2\u0241\u024a")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0242\u024a\5Z.\2\u0243\u024a\5^\60\2\u0244\u024a")
 *         buf.write("\5l\67\2\u0245\u024a\5h\65\2\u0246\u0247\5v<\2\u0247\u0248")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 321, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__496, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 321, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":322
 *         buf.write("\2\2\2\u023cQ\3\2\2\2\u023d\u023b\3\2\2\2\u023e\u023f")
 *         buf.write("\5p9\2\u023f\u0240\7\21\2\2\u0240\u0241\5n8\2\u0241\u024a")
 *         buf.write("\3\2\2\2\u0242\u024a\5Z.\2\u0243\u024a\5^\60\2\u0244\u024a")             # <<<<<<<<<<<<<<
 *         buf.write("\5l\67\2\u0245\u024a\5h\65\2\u0246\u0247\5v<\2\u0247\u0248")
 *         buf.write("\5|?\2\u0248\u024a\3\2\2\2\u0249\u023e\3\2\2\2\u0249\u0242")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 322, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__497, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 322, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":323
 *         buf.write("\5p9\2\u023f\u0240\7\21\2\2\u0240\u0241\5n8\2\u0241\u024a")
 *         buf.write("\3\2\2\2\u0242\u024a\5Z.\2\u0243\u024a\5^\60\2\u0244\u024a")
 *         buf.write("\5l\67\2\u0245\u024a\5h\65\2\u0246\u0247\5v<\2\u0247\u0248")             # <<<<<<<<<<<<<<
 *         buf.write("\5|?\2\u0248\u024a\3\2\2\2\u0249\u023e\3\2\2\2\u0249\u0242")
 *         buf.write("\3\2\2\2\u0249\u0243\3\2\2\2\u0249\u0244\3\2\2\2\u0249")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 323, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__498, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 323, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":324
 *         buf.write("\3\2\2\2\u0242\u024a\5Z.\2\u0243\u024a\5^\60\2\u0244\u024a")
 *         buf.write("\5l\67\2\u0245\u024a\5h\65\2\u0246\u0247\5v<\2\u0247\u0248")
 *         buf.write("\5|?\2\u0248\u024a\3\2\2\2\u0249\u023e\3\2\2\2\u0249\u0242")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0249\u0243\3\2\2\2\u0249\u0244\3\2\2\2\u0249")
 *         buf.write("\u0245\3\2\2\2\u0249\u0246\3\2\2\2\u024aS\3\2\2\2\u024b")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 324, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__500, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 324, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":325
 *         buf.write("\5l\67\2\u0245\u024a\5h\65\2\u0246\u0247\5v<\2\u0247\u0248")
 *         buf.write("\5|?\2\u0248\u024a\3\2\2\2\u0249\u023e\3\2\2\2\u0249\u0242")
 *         buf.write("\3\2\2\2\u0249\u0243\3\2\2\2\u0249\u0244\3\2\2\2\u0249")             # <<<<<<<<<<<<<<
 *         buf.write("\u0245\3\2\2\2\u0249\u0246\3\2\2\2\u024aS\3\2\2\2\u024b")
 *         buf.write("\u024c\5R*\2\u024c\u024d\5\u008eH\2\u024dU\3\2\2\2\u024e")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 325, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__502, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 325, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":326
 *         buf.write("\5|?\2\u0248\u024a\3\2\2\2\u0249\u023e\3\2\2\2\u0249\u0242")
 *         buf.write("\3\2\2\2\u0249\u0243\3\2\2\2\u0249\u0244\3\2\2\2\u0249")
 *         buf.write("\u0245\3\2\2\2\u0249\u0246\3\2\2\2\u024aS\3\2\2\2\u024b")             # <<<<<<<<<<<<<<
 *         buf.write("\u024c\5R*\2\u024c\u024d\5\u008eH\2\u024dU\3\2\2\2\u024e")
 *         buf.write("\u0252\5z>\2\u024f\u0250\7+\2\2\u0250\u0253\5n8\2\u0251")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 326, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__503, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 326, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":327
 *         buf.write("\3\2\2\2\u0249\u0243\3\2\2\2\u0249\u0244\3\2\2\2\u0249")
 *         buf.write("\u0245\3\2\2\2\u0249\u0246\3\2\2\2\u024aS\3\2\2\2\u024b")
 *         buf.write("\u024c\5R*\2\u024c\u024d\5\u008eH\2\u024dU\3\2\2\2\u024e")             # <<<<<<<<<<<<<<
 *         buf.write("\u0252\5z>\2\u024f\u0250\7+\2\2\u0250\u0253\5n8\2\u0251")
 *         buf.write("\u0253\5|?\2\u0252\u024f\3\2\2\2\u0252\u0251\3\2\2\2\u0253")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 327, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__504, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 327, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":328
 *         buf.write("\u0245\3\2\2\2\u0249\u0246\3\2\2\2\u024aS\3\2\2\2\u024b")
 *         buf.write("\u024c\5R*\2\u024c\u024d\5\u008eH\2\u024dU\3\2\2\2\u024e")
 *         buf.write("\u0252\5z>\2\u024f\u0250\7+\2\2\u0250\u0253\5n8\2\u0251")             # <<<<<<<<<<<<<<
 *         buf.write("\u0253\5|?\2\u0252\u024f\3\2\2\2\u0252\u0251\3\2\2\2\u0253")
 *         buf.write("\u0269\3\2\2\2\u0254\u0255\7\23\2\2\u0255\u025a\5z>\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 328, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__505, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 328, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":329
 *         buf.write("\u024c\5R*\2\u024c\u024d\5\u008eH\2\u024dU\3\2\2\2\u024e")
 *         buf.write("\u0252\5z>\2\u024f\u0250\7+\2\2\u0250\u0253\5n8\2\u0251")
 *         buf.write("\u0253\5|?\2\u0252\u024f\3\2\2\2\u0252\u0251\3\2\2\2\u0253")             # <<<<<<<<<<<<<<
 *         buf.write("\u0269\3\2\2\2\u0254\u0255\7\23\2\2\u0255\u025a\5z>\2")
 *         buf.write("\u0256\u0257\7\27\2\2\u0257\u0259\5z>\2\u0258\u0256\3")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 329, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__507, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 329, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":330
 *         buf.write("\u0252\5z>\2\u024f\u0250\7+\2\2\u0250\u0253\5n8\2\u0251")
 *         buf.write("\u0253\5|?\2\u0252\u024f\3\2\2\2\u0252\u0251\3\2\2\2\u0253")
 *         buf.write("\u0269\3\2\2\2\u0254\u0255\7\23\2\2\u0255\u025a\5z>\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0256\u0257\7\27\2\2\u0257\u0259\5z>\2\u0258\u0256\3")
 *         buf.write("\2\2\2\u0259\u025c\3\2\2\2\u025a\u0258\3\2\2\2\u025a\u025b")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 330, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__508, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 330, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":331
 *         buf.write("\u0253\5|?\2\u0252\u024f\3\2\2\2\u0252\u0251\3\2\2\2\u0253")
 *         buf.write("\u0269\3\2\2\2\u0254\u0255\7\23\2\2\u0255\u025a\5z>\2")
 *         buf.write("\u0256\u0257\7\27\2\2\u0257\u0259\5z>\2\u0258\u0256\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u0259\u025c\3\2\2\2\u025a\u0258\3\2\2\2\u025a\u025b")
 *         buf.write("\3\2\2\2\u025b\u025d\3\2\2\2\u025c\u025a\3\2\2\2\u025d")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 331, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__509, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 331, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":332
 *         buf.write("\u0269\3\2\2\2\u0254\u0255\7\23\2\2\u0255\u025a\5z>\2")
 *         buf.write("\u0256\u0257\7\27\2\2\u0257\u0259\5z>\2\u0258\u0256\3")
 *         buf.write("\2\2\2\u0259\u025c\3\2\2\2\u025a\u0258\3\2\2\2\u025a\u025b")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u025b\u025d\3\2\2\2\u025c\u025a\3\2\2\2\u025d")
 *         buf.write("\u025e\7\25\2\2\u025e\u025f\7+\2\2\u025f\u0260\5z>\2\u0260")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 332, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__511, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 332, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":333
 *         buf.write("\u0256\u0257\7\27\2\2\u0257\u0259\5z>\2\u0258\u0256\3")
 *         buf.write("\2\2\2\u0259\u025c\3\2\2\2\u025a\u0258\3\2\2\2\u025a\u025b")
 *         buf.write("\3\2\2\2\u025b\u025d\3\2\2\2\u025c\u025a\3\2\2\2\u025d")             # <<<<<<<<<<<<<<
 *         buf.write("\u025e\7\25\2\2\u025e\u025f\7+\2\2\u025f\u0260\5z>\2\u0260")
 *         buf.write("\u0261\5|?\2\u0261\u0269\3\2\2\2\u0262\u0269\7.\2\2\u0263")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 333, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__513, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 333, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":334
 *         buf.write("\2\2\2\u0259\u025c\3\2\2\2\u025a\u0258\3\2\2\2\u025a\u025b")
 *         buf.write("\3\2\2\2\u025b\u025d\3\2\2\2\u025c\u025a\3\2\2\2\u025d")
 *         buf.write("\u025e\7\25\2\2\u025e\u025f\7+\2\2\u025f\u0260\5z>\2\u0260")             # <<<<<<<<<<<<<<
 *         buf.write("\u0261\5|?\2\u0261\u0269\3\2\2\2\u0262\u0269\7.\2\2\u0263")
 *         buf.write("\u0269\7/\2\2\u0264\u0269\5\\/\2\u0265\u0269\5`\61\2\u0266")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 334, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__514, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 334, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":335
 *         buf.write("\3\2\2\2\u025b\u025d\3\2\2\2\u025c\u025a\3\2\2\2\u025d")
 *         buf.write("\u025e\7\25\2\2\u025e\u025f\7+\2\2\u025f\u0260\5z>\2\u0260")
 *         buf.write("\u0261\5|?\2\u0261\u0269\3\2\2\2\u0262\u0269\7.\2\2\u0263")             # <<<<<<<<<<<<<<
 *         buf.write("\u0269\7/\2\2\u0264\u0269\5\\/\2\u0265\u0269\5`\61\2\u0266")
 *         buf.write("\u0269\5f\64\2\u0267\u0269\5j\66\2\u0268\u024e\3\2\2\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 335, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__516, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 335, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":336
 *         buf.write("\u025e\7\25\2\2\u025e\u025f\7+\2\2\u025f\u0260\5z>\2\u0260")
 *         buf.write("\u0261\5|?\2\u0261\u0269\3\2\2\2\u0262\u0269\7.\2\2\u0263")
 *         buf.write("\u0269\7/\2\2\u0264\u0269\5\\/\2\u0265\u0269\5`\61\2\u0266")             # <<<<<<<<<<<<<<
 *         buf.write("\u0269\5f\64\2\u0267\u0269\5j\66\2\u0268\u024e\3\2\2\2")
 *         buf.write("\u0268\u0254\3\2\2\2\u0268\u0262\3\2\2\2\u0268\u0263\3")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 336, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__517, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 336, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":337
 *         buf.write("\u0261\5|?\2\u0261\u0269\3\2\2\2\u0262\u0269\7.\2\2\u0263")
 *         buf.write("\u0269\7/\2\2\u0264\u0269\5\\/\2\u0265\u0269\5`\61\2\u0266")
 *         buf.write("\u0269\5f\64\2\u0267\u0269\5j\66\2\u0268\u024e\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0268\u0254\3\2\2\2\u0268\u0262\3\2\2\2\u0268\u0263\3")
 *         buf.write("\2\2\2\u0268\u0264\3\2\2\2\u0268\u0265\3\2\2\2\u0268\u0266")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 337, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__518, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 337, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":338
 *         buf.write("\u0269\7/\2\2\u0264\u0269\5\\/\2\u0265\u0269\5`\61\2\u0266")
 *         buf.write("\u0269\5f\64\2\u0267\u0269\5j\66\2\u0268\u024e\3\2\2\2")
 *         buf.write("\u0268\u0254\3\2\2\2\u0268\u0262\3\2\2\2\u0268\u0263\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u0268\u0264\3\2\2\2\u0268\u0265\3\2\2\2\u0268\u0266")
 *         buf.write("\3\2\2\2\u0268\u0267\3\2\2\2\u0269W\3\2\2\2\u026a\u026b")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 338, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__520, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 338, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":339
 *         buf.write("\u0269\5f\64\2\u0267\u0269\5j\66\2\u0268\u024e\3\2\2\2")
 *         buf.write("\u0268\u0254\3\2\2\2\u0268\u0262\3\2\2\2\u0268\u0263\3")
 *         buf.write("\2\2\2\u0268\u0264\3\2\2\2\u0268\u0265\3\2\2\2\u0268\u0266")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0268\u0267\3\2\2\2\u0269W\3\2\2\2\u026a\u026b")
 *         buf.write("\5V,\2\u026b\u026c\5\u008eH\2\u026cY\3\2\2\2\u026d\u026e")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 339, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__522, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 339, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":340
 *         buf.write("\u0268\u0254\3\2\2\2\u0268\u0262\3\2\2\2\u0268\u0263\3")
 *         buf.write("\2\2\2\u0268\u0264\3\2\2\2\u0268\u0265\3\2\2\2\u0268\u0266")
 *         buf.write("\3\2\2\2\u0268\u0267\3\2\2\2\u0269W\3\2\2\2\u026a\u026b")             # <<<<<<<<<<<<<<
 *         buf.write("\5V,\2\u026b\u026c\5\u008eH\2\u026cY\3\2\2\2\u026d\u026e")
 *         buf.write("\7*\2\2\u026e\u026f\5n8\2\u026f\u0275\7\60\2\2\u0270\u0271")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 340, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__523, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 340, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":341
 *         buf.write("\2\2\2\u0268\u0264\3\2\2\2\u0268\u0265\3\2\2\2\u0268\u0266")
 *         buf.write("\3\2\2\2\u0268\u0267\3\2\2\2\u0269W\3\2\2\2\u026a\u026b")
 *         buf.write("\5V,\2\u026b\u026c\5\u008eH\2\u026cY\3\2\2\2\u026d\u026e")             # <<<<<<<<<<<<<<
 *         buf.write("\7*\2\2\u026e\u026f\5n8\2\u026f\u0275\7\60\2\2\u0270\u0271")
 *         buf.write("\5T+\2\u0271\u0272\7\3\2\2\u0272\u0274\3\2\2\2\u0273\u0270")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 341, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__524, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 341, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":342
 *         buf.write("\3\2\2\2\u0268\u0267\3\2\2\2\u0269W\3\2\2\2\u026a\u026b")
 *         buf.write("\5V,\2\u026b\u026c\5\u008eH\2\u026cY\3\2\2\2\u026d\u026e")
 *         buf.write("\7*\2\2\u026e\u026f\5n8\2\u026f\u0275\7\60\2\2\u0270\u0271")             # <<<<<<<<<<<<<<
 *         buf.write("\5T+\2\u0271\u0272\7\3\2\2\u0272\u0274\3\2\2\2\u0273\u0270")
 *         buf.write("\3\2\2\2\u0274\u0277\3\2\2\2\u0275\u0273\3\2\2\2\u0275")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 342, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__525, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 342, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":343
 *         buf.write("\5V,\2\u026b\u026c\5\u008eH\2\u026cY\3\2\2\2\u026d\u026e")
 *         buf.write("\7*\2\2\u026e\u026f\5n8\2\u026f\u0275\7\60\2\2\u0270\u0271")
 *         buf.write("\5T+\2\u0271\u0272\7\3\2\2\u0272\u0274\3\2\2\2\u0273\u0270")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0274\u0277\3\2\2\2\u0275\u0273\3\2\2\2\u0275")
 *         buf.write("\u0276\3\2\2\2\u0276\u0285\3\2\2\2\u0277\u0275\3\2\2\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 343, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__526, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 343, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":344
 *         buf.write("\7*\2\2\u026e\u026f\5n8\2\u026f\u0275\7\60\2\2\u0270\u0271")
 *         buf.write("\5T+\2\u0271\u0272\7\3\2\2\u0272\u0274\3\2\2\2\u0273\u0270")
 *         buf.write("\3\2\2\2\u0274\u0277\3\2\2\2\u0275\u0273\3\2\2\2\u0275")             # <<<<<<<<<<<<<<
 *         buf.write("\u0276\3\2\2\2\u0276\u0285\3\2\2\2\u0277\u0275\3\2\2\2")
 *         buf.write("\u0278\u0279\7\61\2\2\u0279\u027a\5n8\2\u027a\u0280\7")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 344, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__528, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 344, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":345
 *         buf.write("\5T+\2\u0271\u0272\7\3\2\2\u0272\u0274\3\2\2\2\u0273\u0270")
 *         buf.write("\3\2\2\2\u0274\u0277\3\2\2\2\u0275\u0273\3\2\2\2\u0275")
 *         buf.write("\u0276\3\2\2\2\u0276\u0285\3\2\2\2\u0277\u0275\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0278\u0279\7\61\2\2\u0279\u027a\5n8\2\u027a\u0280\7")
 *         buf.write("\60\2\2\u027b\u027c\5T+\2\u027c\u027d\7\3\2\2\u027d\u027f")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 345, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__530, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 345, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":346
 *         buf.write("\3\2\2\2\u0274\u0277\3\2\2\2\u0275\u0273\3\2\2\2\u0275")
 *         buf.write("\u0276\3\2\2\2\u0276\u0285\3\2\2\2\u0277\u0275\3\2\2\2")
 *         buf.write("\u0278\u0279\7\61\2\2\u0279\u027a\5n8\2\u027a\u0280\7")             # <<<<<<<<<<<<<<
 *         buf.write("\60\2\2\u027b\u027c\5T+\2\u027c\u027d\7\3\2\2\u027d\u027f")
 *         buf.write("\3\2\2\2\u027e\u027b\3\2\2\2\u027f\u0282\3\2\2\2\u0280")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 346, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__531, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 346, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":347
 *         buf.write("\u0276\3\2\2\2\u0276\u0285\3\2\2\2\u0277\u0275\3\2\2\2")
 *         buf.write("\u0278\u0279\7\61\2\2\u0279\u027a\5n8\2\u027a\u0280\7")
 *         buf.write("\60\2\2\u027b\u027c\5T+\2\u027c\u027d\7\3\2\2\u027d\u027f")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u027e\u027b\3\2\2\2\u027f\u0282\3\2\2\2\u0280")
 *         buf.write("\u027e\3\2\2\2\u0280\u0281\3\2\2\2\u0281\u0284\3\2\2\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 347, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__532, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 347, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":348
 *         buf.write("\u0278\u0279\7\61\2\2\u0279\u027a\5n8\2\u027a\u0280\7")
 *         buf.write("\60\2\2\u027b\u027c\5T+\2\u027c\u027d\7\3\2\2\u027d\u027f")
 *         buf.write("\3\2\2\2\u027e\u027b\3\2\2\2\u027f\u0282\3\2\2\2\u0280")             # <<<<<<<<<<<<<<
 *         buf.write("\u027e\3\2\2\2\u0280\u0281\3\2\2\2\u0281\u0284\3\2\2\2")
 *         buf.write("\u0282\u0280\3\2\2\2\u0283\u0278\3\2\2\2\u0284\u0287\3")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 348, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__534, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 348, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":349
 *         buf.write("\60\2\2\u027b\u027c\5T+\2\u027c\u027d\7\3\2\2\u027d\u027f")
 *         buf.write("\3\2\2\2\u027e\u027b\3\2\2\2\u027f\u0282\3\2\2\2\u0280")
 *         buf.write("\u027e\3\2\2\2\u0280\u0281\3\2\2\2\u0281\u0284\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0282\u0280\3\2\2\2\u0283\u0278\3\2\2\2\u0284\u0287\3")
 *         buf.write("\2\2\2\u0285\u0283\3\2\2\2\u0285\u0286\3\2\2\2\u0286\u0291")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 349, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__536, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 349, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":350
 *         buf.write("\3\2\2\2\u027e\u027b\3\2\2\2\u027f\u0282\3\2\2\2\u0280")
 *         buf.write("\u027e\3\2\2\2\u0280\u0281\3\2\2\2\u0281\u0284\3\2\2\2")
 *         buf.write("\u0282\u0280\3\2\2\2\u0283\u0278\3\2\2\2\u0284\u0287\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u0285\u0283\3\2\2\2\u0285\u0286\3\2\2\2\u0286\u0291")
 *         buf.write("\3\2\2\2\u0287\u0285\3\2\2\2\u0288\u028e\7\62\2\2\u0289")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 350, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__538, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 350, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":351
 *         buf.write("\u027e\3\2\2\2\u0280\u0281\3\2\2\2\u0281\u0284\3\2\2\2")
 *         buf.write("\u0282\u0280\3\2\2\2\u0283\u0278\3\2\2\2\u0284\u0287\3")
 *         buf.write("\2\2\2\u0285\u0283\3\2\2\2\u0285\u0286\3\2\2\2\u0286\u0291")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0287\u0285\3\2\2\2\u0288\u028e\7\62\2\2\u0289")
 *         buf.write("\u028a\5T+\2\u028a\u028b\7\3\2\2\u028b\u028d\3\2\2\2\u028c")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 351, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__540, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 351, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":352
 *         buf.write("\u0282\u0280\3\2\2\2\u0283\u0278\3\2\2\2\u0284\u0287\3")
 *         buf.write("\2\2\2\u0285\u0283\3\2\2\2\u0285\u0286\3\2\2\2\u0286\u0291")
 *         buf.write("\3\2\2\2\u0287\u0285\3\2\2\2\u0288\u028e\7\62\2\2\u0289")             # <<<<<<<<<<<<<<
 *         buf.write("\u028a\5T+\2\u028a\u028b\7\3\2\2\u028b\u028d\3\2\2\2\u028c")
 *         buf.write("\u0289\3\2\2\2\u028d\u0290\3\2\2\2\u028e\u028c\3\2\2\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 352, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__541, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 352, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":353
 *         buf.write("\2\2\2\u0285\u0283\3\2\2\2\u0285\u0286\3\2\2\2\u0286\u0291")
 *         buf.write("\3\2\2\2\u0287\u0285\3\2\2\2\u0288\u028e\7\62\2\2\u0289")
 *         buf.write("\u028a\5T+\2\u028a\u028b\7\3\2\2\u028b\u028d\3\2\2\2\u028c")             # <<<<<<<<<<<<<<
 *         buf.write("\u0289\3\2\2\2\u028d\u0290\3\2\2\2\u028e\u028c\3\2\2\2")
 *         buf.write("\u028e\u028f\3\2\2\2\u028f\u0292\3\2\2\2\u0290\u028e\3")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 353, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__542, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 353, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":354
 *         buf.write("\3\2\2\2\u0287\u0285\3\2\2\2\u0288\u028e\7\62\2\2\u0289")
 *         buf.write("\u028a\5T+\2\u028a\u028b\7\3\2\2\u028b\u028d\3\2\2\2\u028c")
 *         buf.write("\u0289\3\2\2\2\u028d\u0290\3\2\2\2\u028e\u028c\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u028e\u028f\3\2\2\2\u028f\u0292\3\2\2\2\u0290\u028e\3")
 *         buf.write("\2\2\2\u0291\u0288\3\2\2\2\u0291\u0292\3\2\2\2\u0292\u0293")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 354, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__544, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 354, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":355
 *         buf.write("\u028a\5T+\2\u028a\u028b\7\3\2\2\u028b\u028d\3\2\2\2\u028c")
 *         buf.write("\u0289\3\2\2\2\u028d\u0290\3\2\2\2\u028e\u028c\3\2\2\2")
 *         buf.write("\u028e\u028f\3\2\2\2\u028f\u0292\3\2\2\2\u0290\u028e\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u0291\u0288\3\2\2\2\u0291\u0292\3\2\2\2\u0292\u0293")
 *         buf.write("\3\2\2\2\u0293\u0294\7\20\2\2\u0294\u0295\7*\2\2\u0295")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 355, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__546, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 355, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":356
 *         buf.write("\u0289\3\2\2\2\u028d\u0290\3\2\2\2\u028e\u028c\3\2\2\2")
 *         buf.write("\u028e\u028f\3\2\2\2\u028f\u0292\3\2\2\2\u0290\u028e\3")
 *         buf.write("\2\2\2\u0291\u0288\3\2\2\2\u0291\u0292\3\2\2\2\u0292\u0293")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0293\u0294\7\20\2\2\u0294\u0295\7*\2\2\u0295")
 *         buf.write("[\3\2\2\2\u0296\u0297\7*\2\2\u0297\u0298\5n8\2\u0298\u029e")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 356, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__548, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 356, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":357
 *         buf.write("\u028e\u028f\3\2\2\2\u028f\u0292\3\2\2\2\u0290\u028e\3")
 *         buf.write("\2\2\2\u0291\u0288\3\2\2\2\u0291\u0292\3\2\2\2\u0292\u0293")
 *         buf.write("\3\2\2\2\u0293\u0294\7\20\2\2\u0294\u0295\7*\2\2\u0295")             # <<<<<<<<<<<<<<
 *         buf.write("[\3\2\2\2\u0296\u0297\7*\2\2\u0297\u0298\5n8\2\u0298\u029e")
 *         buf.write("\7\60\2\2\u0299\u029a\5X-\2\u029a\u029b\7\3\2\2\u029b")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 357, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__550, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 357, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":358
 *         buf.write("\2\2\2\u0291\u0288\3\2\2\2\u0291\u0292\3\2\2\2\u0292\u0293")
 *         buf.write("\3\2\2\2\u0293\u0294\7\20\2\2\u0294\u0295\7*\2\2\u0295")
 *         buf.write("[\3\2\2\2\u0296\u0297\7*\2\2\u0297\u0298\5n8\2\u0298\u029e")             # <<<<<<<<<<<<<<
 *         buf.write("\7\60\2\2\u0299\u029a\5X-\2\u029a\u029b\7\3\2\2\u029b")
 *         buf.write("\u029d\3\2\2\2\u029c\u0299\3\2\2\2\u029d\u02a0\3\2\2\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 358, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__551, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 358, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":359
 *         buf.write("\3\2\2\2\u0293\u0294\7\20\2\2\u0294\u0295\7*\2\2\u0295")
 *         buf.write("[\3\2\2\2\u0296\u0297\7*\2\2\u0297\u0298\5n8\2\u0298\u029e")
 *         buf.write("\7\60\2\2\u0299\u029a\5X-\2\u029a\u029b\7\3\2\2\u029b")             # <<<<<<<<<<<<<<
 *         buf.write("\u029d\3\2\2\2\u029c\u0299\3\2\2\2\u029d\u02a0\3\2\2\2")
 *         buf.write("\u029e\u029c\3\2\2\2\u029e\u029f\3\2\2\2\u029f\u02ae\3")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 359, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__552, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 359, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":360
 *         buf.write("[\3\2\2\2\u0296\u0297\7*\2\2\u0297\u0298\5n8\2\u0298\u029e")
 *         buf.write("\7\60\2\2\u0299\u029a\5X-\2\u029a\u029b\7\3\2\2\u029b")
 *         buf.write("\u029d\3\2\2\2\u029c\u0299\3\2\2\2\u029d\u02a0\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u029e\u029c\3\2\2\2\u029e\u029f\3\2\2\2\u029f\u02ae\3")
 *         buf.write("\2\2\2\u02a0\u029e\3\2\2\2\u02a1\u02a2\7\61\2\2\u02a2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 360, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__554, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 360, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":361
 *         buf.write("\7\60\2\2\u0299\u029a\5X-\2\u029a\u029b\7\3\2\2\u029b")
 *         buf.write("\u029d\3\2\2\2\u029c\u0299\3\2\2\2\u029d\u02a0\3\2\2\2")
 *         buf.write("\u029e\u029c\3\2\2\2\u029e\u029f\3\2\2\2\u029f\u02ae\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u02a0\u029e\3\2\2\2\u02a1\u02a2\7\61\2\2\u02a2")
 *         buf.write("\u02a3\5n8\2\u02a3\u02a9\7\60\2\2\u02a4\u02a5\5X-\2\u02a5")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 361, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__556, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 361, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":362
 *         buf.write("\u029d\3\2\2\2\u029c\u0299\3\2\2\2\u029d\u02a0\3\2\2\2")
 *         buf.write("\u029e\u029c\3\2\2\2\u029e\u029f\3\2\2\2\u029f\u02ae\3")
 *         buf.write("\2\2\2\u02a0\u029e\3\2\2\2\u02a1\u02a2\7\61\2\2\u02a2")             # <<<<<<<<<<<<<<
 *         buf.write("\u02a3\5n8\2\u02a3\u02a9\7\60\2\2\u02a4\u02a5\5X-\2\u02a5")
 *         buf.write("\u02a6\7\3\2\2\u02a6\u02a8\3\2\2\2\u02a7\u02a4\3\2\2\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 362, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__557, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 362, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":363
 *         buf.write("\u029e\u029c\3\2\2\2\u029e\u029f\3\2\2\2\u029f\u02ae\3")
 *         buf.write("\2\2\2\u02a0\u029e\3\2\2\2\u02a1\u02a2\7\61\2\2\u02a2")
 *         buf.write("\u02a3\5n8\2\u02a3\u02a9\7\60\2\2\u02a4\u02a5\5X-\2\u02a5")             # <<<<<<<<<<<<<<
 *         buf.write("\u02a6\7\3\2\2\u02a6\u02a8\3\2\2\2\u02a7\u02a4\3\2\2\2")
 *         buf.write("\u02a8\u02ab\3\2\2\2\u02a9\u02a7\3\2\2\2\u02a9\u02aa\3")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 363, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__558, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 363, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":364
 *         buf.write("\2\2\2\u02a0\u029e\3\2\2\2\u02a1\u02a2\7\61\2\2\u02a2")
 *         buf.write("\u02a3\5n8\2\u02a3\u02a9\7\60\2\2\u02a4\u02a5\5X-\2\u02a5")
 *         buf.write("\u02a6\7\3\2\2\u02a6\u02a8\3\2\2\2\u02a7\u02a4\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u02a8\u02ab\3\2\2\2\u02a9\u02a7\3\2\2\2\u02a9\u02aa\3")
 *         buf.write("\2\2\2\u02aa\u02ad\3\2\2\2\u02ab\u02a9\3\2\2\2\u02ac\u02a1")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 364, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__560, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 364, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":365
 *         buf.write("\u02a3\5n8\2\u02a3\u02a9\7\60\2\2\u02a4\u02a5\5X-\2\u02a5")
 *         buf.write("\u02a6\7\3\2\2\u02a6\u02a8\3\2\2\2\u02a7\u02a4\3\2\2\2")
 *         buf.write("\u02a8\u02ab\3\2\2\2\u02a9\u02a7\3\2\2\2\u02a9\u02aa\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u02aa\u02ad\3\2\2\2\u02ab\u02a9\3\2\2\2\u02ac\u02a1")
 *         buf.write("\3\2\2\2\u02ad\u02b0\3\2\2\2\u02ae\u02ac\3\2\2\2\u02ae")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 365, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__562, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 365, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":366
 *         buf.write("\u02a6\7\3\2\2\u02a6\u02a8\3\2\2\2\u02a7\u02a4\3\2\2\2")
 *         buf.write("\u02a8\u02ab\3\2\2\2\u02a9\u02a7\3\2\2\2\u02a9\u02aa\3")
 *         buf.write("\2\2\2\u02aa\u02ad\3\2\2\2\u02ab\u02a9\3\2\2\2\u02ac\u02a1")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u02ad\u02b0\3\2\2\2\u02ae\u02ac\3\2\2\2\u02ae")
 *         buf.write("\u02af\3\2\2\2\u02af\u02ba\3\2\2\2\u02b0\u02ae\3\2\2\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 366, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__564, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 366, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":367
 *         buf.write("\u02a8\u02ab\3\2\2\2\u02a9\u02a7\3\2\2\2\u02a9\u02aa\3")
 *         buf.write("\2\2\2\u02aa\u02ad\3\2\2\2\u02ab\u02a9\3\2\2\2\u02ac\u02a1")
 *         buf.write("\3\2\2\2\u02ad\u02b0\3\2\2\2\u02ae\u02ac\3\2\2\2\u02ae")             # <<<<<<<<<<<<<<
 *         buf.write("\u02af\3\2\2\2\u02af\u02ba\3\2\2\2\u02b0\u02ae\3\2\2\2")
 *         buf.write("\u02b1\u02b7\7\62\2\2\u02b2\u02b3\5X-\2\u02b3\u02b4\7")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 367, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__566, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 367, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":368
 *         buf.write("\2\2\2\u02aa\u02ad\3\2\2\2\u02ab\u02a9\3\2\2\2\u02ac\u02a1")
 *         buf.write("\3\2\2\2\u02ad\u02b0\3\2\2\2\u02ae\u02ac\3\2\2\2\u02ae")
 *         buf.write("\u02af\3\2\2\2\u02af\u02ba\3\2\2\2\u02b0\u02ae\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u02b1\u02b7\7\62\2\2\u02b2\u02b3\5X-\2\u02b3\u02b4\7")
 *         buf.write("\3\2\2\u02b4\u02b6\3\2\2\2\u02b5\u02b2\3\2\2\2\u02b6\u02b9")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 368, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__568, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 368, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":369
 *         buf.write("\3\2\2\2\u02ad\u02b0\3\2\2\2\u02ae\u02ac\3\2\2\2\u02ae")
 *         buf.write("\u02af\3\2\2\2\u02af\u02ba\3\2\2\2\u02b0\u02ae\3\2\2\2")
 *         buf.write("\u02b1\u02b7\7\62\2\2\u02b2\u02b3\5X-\2\u02b3\u02b4\7")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\u02b4\u02b6\3\2\2\2\u02b5\u02b2\3\2\2\2\u02b6\u02b9")
 *         buf.write("\3\2\2\2\u02b7\u02b5\3\2\2\2\u02b7\u02b8\3\2\2\2\u02b8")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 369, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__569, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 369, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":370
 *         buf.write("\u02af\3\2\2\2\u02af\u02ba\3\2\2\2\u02b0\u02ae\3\2\2\2")
 *         buf.write("\u02b1\u02b7\7\62\2\2\u02b2\u02b3\5X-\2\u02b3\u02b4\7")
 *         buf.write("\3\2\2\u02b4\u02b6\3\2\2\2\u02b5\u02b2\3\2\2\2\u02b6\u02b9")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u02b7\u02b5\3\2\2\2\u02b7\u02b8\3\2\2\2\u02b8")
 *         buf.write("\u02bb\3\2\2\2\u02b9\u02b7\3\2\2\2\u02ba\u02b1\3\2\2\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 370, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__571, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 370, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":371
 *         buf.write("\u02b1\u02b7\7\62\2\2\u02b2\u02b3\5X-\2\u02b3\u02b4\7")
 *         buf.write("\3\2\2\u02b4\u02b6\3\2\2\2\u02b5\u02b2\3\2\2\2\u02b6\u02b9")
 *         buf.write("\3\2\2\2\u02b7\u02b5\3\2\2\2\u02b7\u02b8\3\2\2\2\u02b8")             # <<<<<<<<<<<<<<
 *         buf.write("\u02bb\3\2\2\2\u02b9\u02b7\3\2\2\2\u02ba\u02b1\3\2\2\2")
 *         buf.write("\u02ba\u02bb\3\2\2\2\u02bb\u02bc\3\2\2\2\u02bc\u02bd\7")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 371, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__573, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 371, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":372
 *         buf.write("\3\2\2\u02b4\u02b6\3\2\2\2\u02b5\u02b2\3\2\2\2\u02b6\u02b9")
 *         buf.write("\3\2\2\2\u02b7\u02b5\3\2\2\2\u02b7\u02b8\3\2\2\2\u02b8")
 *         buf.write("\u02bb\3\2\2\2\u02b9\u02b7\3\2\2\2\u02ba\u02b1\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u02ba\u02bb\3\2\2\2\u02bb\u02bc\3\2\2\2\u02bc\u02bd\7")
 *         buf.write("\20\2\2\u02bd\u02be\7*\2\2\u02be]\3\2\2\2\u02bf\u02c0")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 372, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__575, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 372, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":373
 *         buf.write("\3\2\2\2\u02b7\u02b5\3\2\2\2\u02b7\u02b8\3\2\2\2\u02b8")
 *         buf.write("\u02bb\3\2\2\2\u02b9\u02b7\3\2\2\2\u02ba\u02b1\3\2\2\2")
 *         buf.write("\u02ba\u02bb\3\2\2\2\u02bb\u02bc\3\2\2\2\u02bc\u02bd\7")             # <<<<<<<<<<<<<<
 *         buf.write("\20\2\2\u02bd\u02be\7*\2\2\u02be]\3\2\2\2\u02bf\u02c0")
 *         buf.write("\7\63\2\2\u02c0\u02c1\5b\62\2\u02c1\u02c7\7\64\2\2\u02c2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 373, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__577, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 373, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":374
 *         buf.write("\u02bb\3\2\2\2\u02b9\u02b7\3\2\2\2\u02ba\u02b1\3\2\2\2")
 *         buf.write("\u02ba\u02bb\3\2\2\2\u02bb\u02bc\3\2\2\2\u02bc\u02bd\7")
 *         buf.write("\20\2\2\u02bd\u02be\7*\2\2\u02be]\3\2\2\2\u02bf\u02c0")             # <<<<<<<<<<<<<<
 *         buf.write("\7\63\2\2\u02c0\u02c1\5b\62\2\u02c1\u02c7\7\64\2\2\u02c2")
 *         buf.write("\u02c3\5T+\2\u02c3\u02c4\7\3\2\2\u02c4\u02c6\3\2\2\2\u02c5")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 374, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__579, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 374, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":375
 *         buf.write("\u02ba\u02bb\3\2\2\2\u02bb\u02bc\3\2\2\2\u02bc\u02bd\7")
 *         buf.write("\20\2\2\u02bd\u02be\7*\2\2\u02be]\3\2\2\2\u02bf\u02c0")
 *         buf.write("\7\63\2\2\u02c0\u02c1\5b\62\2\u02c1\u02c7\7\64\2\2\u02c2")             # <<<<<<<<<<<<<<
 *         buf.write("\u02c3\5T+\2\u02c3\u02c4\7\3\2\2\u02c4\u02c6\3\2\2\2\u02c5")
 *         buf.write("\u02c2\3\2\2\2\u02c6\u02c9\3\2\2\2\u02c7\u02c5\3\2\2\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 375, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__580, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 375, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":376
 *         buf.write("\20\2\2\u02bd\u02be\7*\2\2\u02be]\3\2\2\2\u02bf\u02c0")
 *         buf.write("\7\63\2\2\u02c0\u02c1\5b\62\2\u02c1\u02c7\7\64\2\2\u02c2")
 *         buf.write("\u02c3\5T+\2\u02c3\u02c4\7\3\2\2\u02c4\u02c6\3\2\2\2\u02c5")             # <<<<<<<<<<<<<<
 *         buf.write("\u02c2\3\2\2\2\u02c6\u02c9\3\2\2\2\u02c7\u02c5\3\2\2\2")
 *         buf.write("\u02c7\u02c8\3\2\2\2\u02c8\u02ca\3\2\2\2\u02c9\u02c7\3")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 376, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__581, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 376, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":377
 *         buf.write("\7\63\2\2\u02c0\u02c1\5b\62\2\u02c1\u02c7\7\64\2\2\u02c2")
 *         buf.write("\u02c3\5T+\2\u02c3\u02c4\7\3\2\2\u02c4\u02c6\3\2\2\2\u02c5")
 *         buf.write("\u02c2\3\2\2\2\u02c6\u02c9\3\2\2\2\u02c7\u02c5\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u02c7\u02c8\3\2\2\2\u02c8\u02ca\3\2\2\2\u02c9\u02c7\3")
 *         buf.write("\2\2\2\u02ca\u02cb\7\20\2\2\u02cb\u02cc\7\63\2\2\u02cc")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 377, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__583, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 377, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":378
 *         buf.write("\u02c3\5T+\2\u02c3\u02c4\7\3\2\2\u02c4\u02c6\3\2\2\2\u02c5")
 *         buf.write("\u02c2\3\2\2\2\u02c6\u02c9\3\2\2\2\u02c7\u02c5\3\2\2\2")
 *         buf.write("\u02c7\u02c8\3\2\2\2\u02c8\u02ca\3\2\2\2\u02c9\u02c7\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u02ca\u02cb\7\20\2\2\u02cb\u02cc\7\63\2\2\u02cc")
 *         buf.write("_\3\2\2\2\u02cd\u02ce\7\63\2\2\u02ce\u02cf\5b\62\2\u02cf")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 378, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__585, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 378, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":379
 *         buf.write("\u02c2\3\2\2\2\u02c6\u02c9\3\2\2\2\u02c7\u02c5\3\2\2\2")
 *         buf.write("\u02c7\u02c8\3\2\2\2\u02c8\u02ca\3\2\2\2\u02c9\u02c7\3")
 *         buf.write("\2\2\2\u02ca\u02cb\7\20\2\2\u02cb\u02cc\7\63\2\2\u02cc")             # <<<<<<<<<<<<<<
 *         buf.write("_\3\2\2\2\u02cd\u02ce\7\63\2\2\u02ce\u02cf\5b\62\2\u02cf")
 *         buf.write("\u02d5\7\64\2\2\u02d0\u02d1\5X-\2\u02d1\u02d2\7\3\2\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 379, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__586, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 379, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":380
 *         buf.write("\u02c7\u02c8\3\2\2\2\u02c8\u02ca\3\2\2\2\u02c9\u02c7\3")
 *         buf.write("\2\2\2\u02ca\u02cb\7\20\2\2\u02cb\u02cc\7\63\2\2\u02cc")
 *         buf.write("_\3\2\2\2\u02cd\u02ce\7\63\2\2\u02ce\u02cf\5b\62\2\u02cf")             # <<<<<<<<<<<<<<
 *         buf.write("\u02d5\7\64\2\2\u02d0\u02d1\5X-\2\u02d1\u02d2\7\3\2\2")
 *         buf.write("\u02d2\u02d4\3\2\2\2\u02d3\u02d0\3\2\2\2\u02d4\u02d7\3")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 380, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__587, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 380, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":381
 *         buf.write("\2\2\2\u02ca\u02cb\7\20\2\2\u02cb\u02cc\7\63\2\2\u02cc")
 *         buf.write("_\3\2\2\2\u02cd\u02ce\7\63\2\2\u02ce\u02cf\5b\62\2\u02cf")
 *         buf.write("\u02d5\7\64\2\2\u02d0\u02d1\5X-\2\u02d1\u02d2\7\3\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u02d2\u02d4\3\2\2\2\u02d3\u02d0\3\2\2\2\u02d4\u02d7\3")
 *         buf.write("\2\2\2\u02d5\u02d3\3\2\2\2\u02d5\u02d6\3\2\2\2\u02d6\u02d8")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 381, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__588, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 381, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":382
 *         buf.write("_\3\2\2\2\u02cd\u02ce\7\63\2\2\u02ce\u02cf\5b\62\2\u02cf")
 *         buf.write("\u02d5\7\64\2\2\u02d0\u02d1\5X-\2\u02d1\u02d2\7\3\2\2")
 *         buf.write("\u02d2\u02d4\3\2\2\2\u02d3\u02d0\3\2\2\2\u02d4\u02d7\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u02d5\u02d3\3\2\2\2\u02d5\u02d6\3\2\2\2\u02d6\u02d8")
 *         buf.write("\3\2\2\2\u02d7\u02d5\3\2\2\2\u02d8\u02d9\7\20\2\2\u02d9")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 382, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__590, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 382, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":383
 *         buf.write("\u02d5\7\64\2\2\u02d0\u02d1\5X-\2\u02d1\u02d2\7\3\2\2")
 *         buf.write("\u02d2\u02d4\3\2\2\2\u02d3\u02d0\3\2\2\2\u02d4\u02d7\3")
 *         buf.write("\2\2\2\u02d5\u02d3\3\2\2\2\u02d5\u02d6\3\2\2\2\u02d6\u02d8")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u02d7\u02d5\3\2\2\2\u02d8\u02d9\7\20\2\2\u02d9")
 *         buf.write("\u02da\7\63\2\2\u02daa\3\2\2\2\u02db\u02e0\5d\63\2\u02dc")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 383, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__592, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 383, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":384
 *         buf.write("\u02d2\u02d4\3\2\2\2\u02d3\u02d0\3\2\2\2\u02d4\u02d7\3")
 *         buf.write("\2\2\2\u02d5\u02d3\3\2\2\2\u02d5\u02d6\3\2\2\2\u02d6\u02d8")
 *         buf.write("\3\2\2\2\u02d7\u02d5\3\2\2\2\u02d8\u02d9\7\20\2\2\u02d9")             # <<<<<<<<<<<<<<
 *         buf.write("\u02da\7\63\2\2\u02daa\3\2\2\2\u02db\u02e0\5d\63\2\u02dc")
 *         buf.write("\u02dd\7\27\2\2\u02dd\u02df\5d\63\2\u02de\u02dc\3\2\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 384, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__594, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 384, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":385
 *         buf.write("\2\2\2\u02d5\u02d3\3\2\2\2\u02d5\u02d6\3\2\2\2\u02d6\u02d8")
 *         buf.write("\3\2\2\2\u02d7\u02d5\3\2\2\2\u02d8\u02d9\7\20\2\2\u02d9")
 *         buf.write("\u02da\7\63\2\2\u02daa\3\2\2\2\u02db\u02e0\5d\63\2\u02dc")             # <<<<<<<<<<<<<<
 *         buf.write("\u02dd\7\27\2\2\u02dd\u02df\5d\63\2\u02de\u02dc\3\2\2")
 *         buf.write("\2\u02df\u02e2\3\2\2\2\u02e0\u02de\3\2\2\2\u02e0\u02e1")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 385, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__595, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 385, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":386
 *         buf.write("\3\2\2\2\u02d7\u02d5\3\2\2\2\u02d8\u02d9\7\20\2\2\u02d9")
 *         buf.write("\u02da\7\63\2\2\u02daa\3\2\2\2\u02db\u02e0\5d\63\2\u02dc")
 *         buf.write("\u02dd\7\27\2\2\u02dd\u02df\5d\63\2\u02de\u02dc\3\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u02df\u02e2\3\2\2\2\u02e0\u02de\3\2\2\2\u02e0\u02e1")
 *         buf.write("\3\2\2\2\u02e1c\3\2\2\2\u02e2\u02e0\3\2\2\2\u02e3\u02e6")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 386, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__596, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 386, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":387
 *         buf.write("\u02da\7\63\2\2\u02daa\3\2\2\2\u02db\u02e0\5d\63\2\u02dc")
 *         buf.write("\u02dd\7\27\2\2\u02dd\u02df\5d\63\2\u02de\u02dc\3\2\2")
 *         buf.write("\2\u02df\u02e2\3\2\2\2\u02e0\u02de\3\2\2\2\u02e0\u02e1")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u02e1c\3\2\2\2\u02e2\u02e0\3\2\2\2\u02e3\u02e6")
 *         buf.write("\7Z\2\2\u02e4\u02e5\7\65\2\2\u02e5\u02e7\5n8\2\u02e6\u02e4")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 387, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__598, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 387, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":388
 *         buf.write("\u02dd\7\27\2\2\u02dd\u02df\5d\63\2\u02de\u02dc\3\2\2")
 *         buf.write("\2\u02df\u02e2\3\2\2\2\u02e0\u02de\3\2\2\2\u02e0\u02e1")
 *         buf.write("\3\2\2\2\u02e1c\3\2\2\2\u02e2\u02e0\3\2\2\2\u02e3\u02e6")             # <<<<<<<<<<<<<<
 *         buf.write("\7Z\2\2\u02e4\u02e5\7\65\2\2\u02e5\u02e7\5n8\2\u02e6\u02e4")
 *         buf.write("\3\2\2\2\u02e6\u02e7\3\2\2\2\u02e7e\3\2\2\2\u02e8\u02e9")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 388, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__599, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 388, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":389
 *         buf.write("\2\u02df\u02e2\3\2\2\2\u02e0\u02de\3\2\2\2\u02e0\u02e1")
 *         buf.write("\3\2\2\2\u02e1c\3\2\2\2\u02e2\u02e0\3\2\2\2\u02e3\u02e6")
 *         buf.write("\7Z\2\2\u02e4\u02e5\7\65\2\2\u02e5\u02e7\5n8\2\u02e6\u02e4")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u02e6\u02e7\3\2\2\2\u02e7e\3\2\2\2\u02e8\u02e9")
 *         buf.write("\7\66\2\2\u02e9\u02ea\5n8\2\u02ea\u02f0\7\64\2\2\u02eb")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 389, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__600, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 389, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":390
 *         buf.write("\3\2\2\2\u02e1c\3\2\2\2\u02e2\u02e0\3\2\2\2\u02e3\u02e6")
 *         buf.write("\7Z\2\2\u02e4\u02e5\7\65\2\2\u02e5\u02e7\5n8\2\u02e6\u02e4")
 *         buf.write("\3\2\2\2\u02e6\u02e7\3\2\2\2\u02e7e\3\2\2\2\u02e8\u02e9")             # <<<<<<<<<<<<<<
 *         buf.write("\7\66\2\2\u02e9\u02ea\5n8\2\u02ea\u02f0\7\64\2\2\u02eb")
 *         buf.write("\u02ec\5X-\2\u02ec\u02ed\7\3\2\2\u02ed\u02ef\3\2\2\2\u02ee")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 390, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__601, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 390, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":391
 *         buf.write("\7Z\2\2\u02e4\u02e5\7\65\2\2\u02e5\u02e7\5n8\2\u02e6\u02e4")
 *         buf.write("\3\2\2\2\u02e6\u02e7\3\2\2\2\u02e7e\3\2\2\2\u02e8\u02e9")
 *         buf.write("\7\66\2\2\u02e9\u02ea\5n8\2\u02ea\u02f0\7\64\2\2\u02eb")             # <<<<<<<<<<<<<<
 *         buf.write("\u02ec\5X-\2\u02ec\u02ed\7\3\2\2\u02ed\u02ef\3\2\2\2\u02ee")
 *         buf.write("\u02eb\3\2\2\2\u02ef\u02f2\3\2\2\2\u02f0\u02ee\3\2\2\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 391, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__602, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 391, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":392
 *         buf.write("\3\2\2\2\u02e6\u02e7\3\2\2\2\u02e7e\3\2\2\2\u02e8\u02e9")
 *         buf.write("\7\66\2\2\u02e9\u02ea\5n8\2\u02ea\u02f0\7\64\2\2\u02eb")
 *         buf.write("\u02ec\5X-\2\u02ec\u02ed\7\3\2\2\u02ed\u02ef\3\2\2\2\u02ee")             # <<<<<<<<<<<<<<
 *         buf.write("\u02eb\3\2\2\2\u02ef\u02f2\3\2\2\2\u02f0\u02ee\3\2\2\2")
 *         buf.write("\u02f0\u02f1\3\2\2\2\u02f1\u02f3\3\2\2\2\u02f2\u02f0\3")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 392, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__603, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 392, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":393
 *         buf.write("\7\66\2\2\u02e9\u02ea\5n8\2\u02ea\u02f0\7\64\2\2\u02eb")
 *         buf.write("\u02ec\5X-\2\u02ec\u02ed\7\3\2\2\u02ed\u02ef\3\2\2\2\u02ee")
 *         buf.write("\u02eb\3\2\2\2\u02ef\u02f2\3\2\2\2\u02f0\u02ee\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u02f0\u02f1\3\2\2\2\u02f1\u02f3\3\2\2\2\u02f2\u02f0\3")
 *         buf.write("\2\2\2\u02f3\u02f4\7\20\2\2\u02f4\u02f5\7\66\2\2\u02f5")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 393, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__605, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 393, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":394
 *         buf.write("\u02ec\5X-\2\u02ec\u02ed\7\3\2\2\u02ed\u02ef\3\2\2\2\u02ee")
 *         buf.write("\u02eb\3\2\2\2\u02ef\u02f2\3\2\2\2\u02f0\u02ee\3\2\2\2")
 *         buf.write("\u02f0\u02f1\3\2\2\2\u02f1\u02f3\3\2\2\2\u02f2\u02f0\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u02f3\u02f4\7\20\2\2\u02f4\u02f5\7\66\2\2\u02f5")
 *         buf.write("g\3\2\2\2\u02f6\u02f7\7\67\2\2\u02f7\u02f8\5n8\2\u02f8")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 394, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__607, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 394, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":395
 *         buf.write("\u02eb\3\2\2\2\u02ef\u02f2\3\2\2\2\u02f0\u02ee\3\2\2\2")
 *         buf.write("\u02f0\u02f1\3\2\2\2\u02f1\u02f3\3\2\2\2\u02f2\u02f0\3")
 *         buf.write("\2\2\2\u02f3\u02f4\7\20\2\2\u02f4\u02f5\7\66\2\2\u02f5")             # <<<<<<<<<<<<<<
 *         buf.write("g\3\2\2\2\u02f6\u02f7\7\67\2\2\u02f7\u02f8\5n8\2\u02f8")
 *         buf.write("\u02fe\7\60\2\2\u02f9\u02fa\5T+\2\u02fa\u02fb\7\3\2\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 395, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__608, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 395, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":396
 *         buf.write("\u02f0\u02f1\3\2\2\2\u02f1\u02f3\3\2\2\2\u02f2\u02f0\3")
 *         buf.write("\2\2\2\u02f3\u02f4\7\20\2\2\u02f4\u02f5\7\66\2\2\u02f5")
 *         buf.write("g\3\2\2\2\u02f6\u02f7\7\67\2\2\u02f7\u02f8\5n8\2\u02f8")             # <<<<<<<<<<<<<<
 *         buf.write("\u02fe\7\60\2\2\u02f9\u02fa\5T+\2\u02fa\u02fb\7\3\2\2")
 *         buf.write("\u02fb\u02fd\3\2\2\2\u02fc\u02f9\3\2\2\2\u02fd\u0300\3")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 396, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__609, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 396, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":397
 *         buf.write("\2\2\2\u02f3\u02f4\7\20\2\2\u02f4\u02f5\7\66\2\2\u02f5")
 *         buf.write("g\3\2\2\2\u02f6\u02f7\7\67\2\2\u02f7\u02f8\5n8\2\u02f8")
 *         buf.write("\u02fe\7\60\2\2\u02f9\u02fa\5T+\2\u02fa\u02fb\7\3\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u02fb\u02fd\3\2\2\2\u02fc\u02f9\3\2\2\2\u02fd\u0300\3")
 *         buf.write("\2\2\2\u02fe\u02fc\3\2\2\2\u02fe\u02ff\3\2\2\2\u02ff\u030e")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 397, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__610, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 397, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":398
 *         buf.write("g\3\2\2\2\u02f6\u02f7\7\67\2\2\u02f7\u02f8\5n8\2\u02f8")
 *         buf.write("\u02fe\7\60\2\2\u02f9\u02fa\5T+\2\u02fa\u02fb\7\3\2\2")
 *         buf.write("\u02fb\u02fd\3\2\2\2\u02fc\u02f9\3\2\2\2\u02fd\u0300\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u02fe\u02fc\3\2\2\2\u02fe\u02ff\3\2\2\2\u02ff\u030e")
 *         buf.write("\3\2\2\2\u0300\u02fe\3\2\2\2\u0301\u0302\78\2\2\u0302")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 398, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__612, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 398, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":399
 *         buf.write("\u02fe\7\60\2\2\u02f9\u02fa\5T+\2\u02fa\u02fb\7\3\2\2")
 *         buf.write("\u02fb\u02fd\3\2\2\2\u02fc\u02f9\3\2\2\2\u02fd\u0300\3")
 *         buf.write("\2\2\2\u02fe\u02fc\3\2\2\2\u02fe\u02ff\3\2\2\2\u02ff\u030e")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0300\u02fe\3\2\2\2\u0301\u0302\78\2\2\u0302")
 *         buf.write("\u0303\5n8\2\u0303\u0309\7\60\2\2\u0304\u0305\5T+\2\u0305")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 399, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__614, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 399, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":400
 *         buf.write("\u02fb\u02fd\3\2\2\2\u02fc\u02f9\3\2\2\2\u02fd\u0300\3")
 *         buf.write("\2\2\2\u02fe\u02fc\3\2\2\2\u02fe\u02ff\3\2\2\2\u02ff\u030e")
 *         buf.write("\3\2\2\2\u0300\u02fe\3\2\2\2\u0301\u0302\78\2\2\u0302")             # <<<<<<<<<<<<<<
 *         buf.write("\u0303\5n8\2\u0303\u0309\7\60\2\2\u0304\u0305\5T+\2\u0305")
 *         buf.write("\u0306\7\3\2\2\u0306\u0308\3\2\2\2\u0307\u0304\3\2\2\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 400, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__615, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 400, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":401
 *         buf.write("\2\2\2\u02fe\u02fc\3\2\2\2\u02fe\u02ff\3\2\2\2\u02ff\u030e")
 *         buf.write("\3\2\2\2\u0300\u02fe\3\2\2\2\u0301\u0302\78\2\2\u0302")
 *         buf.write("\u0303\5n8\2\u0303\u0309\7\60\2\2\u0304\u0305\5T+\2\u0305")             # <<<<<<<<<<<<<<
 *         buf.write("\u0306\7\3\2\2\u0306\u0308\3\2\2\2\u0307\u0304\3\2\2\2")
 *         buf.write("\u0308\u030b\3\2\2\2\u0309\u0307\3\2\2\2\u0309\u030a\3")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 401, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__616, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 401, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":402
 *         buf.write("\3\2\2\2\u0300\u02fe\3\2\2\2\u0301\u0302\78\2\2\u0302")
 *         buf.write("\u0303\5n8\2\u0303\u0309\7\60\2\2\u0304\u0305\5T+\2\u0305")
 *         buf.write("\u0306\7\3\2\2\u0306\u0308\3\2\2\2\u0307\u0304\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0308\u030b\3\2\2\2\u0309\u0307\3\2\2\2\u0309\u030a\3")
 *         buf.write("\2\2\2\u030a\u030d\3\2\2\2\u030b\u0309\3\2\2\2\u030c\u0301")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 402, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__618, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 402, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":403
 *         buf.write("\u0303\5n8\2\u0303\u0309\7\60\2\2\u0304\u0305\5T+\2\u0305")
 *         buf.write("\u0306\7\3\2\2\u0306\u0308\3\2\2\2\u0307\u0304\3\2\2\2")
 *         buf.write("\u0308\u030b\3\2\2\2\u0309\u0307\3\2\2\2\u0309\u030a\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u030a\u030d\3\2\2\2\u030b\u0309\3\2\2\2\u030c\u0301")
 *         buf.write("\3\2\2\2\u030d\u0310\3\2\2\2\u030e\u030c\3\2\2\2\u030e")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 403, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__620, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 403, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":404
 *         buf.write("\u0306\7\3\2\2\u0306\u0308\3\2\2\2\u0307\u0304\3\2\2\2")
 *         buf.write("\u0308\u030b\3\2\2\2\u0309\u0307\3\2\2\2\u0309\u030a\3")
 *         buf.write("\2\2\2\u030a\u030d\3\2\2\2\u030b\u0309\3\2\2\2\u030c\u0301")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u030d\u0310\3\2\2\2\u030e\u030c\3\2\2\2\u030e")
 *         buf.write("\u030f\3\2\2\2\u030f\u0311\3\2\2\2\u0310\u030e\3\2\2\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 404, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__622, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 404, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":405
 *         buf.write("\u0308\u030b\3\2\2\2\u0309\u0307\3\2\2\2\u0309\u030a\3")
 *         buf.write("\2\2\2\u030a\u030d\3\2\2\2\u030b\u0309\3\2\2\2\u030c\u0301")
 *         buf.write("\3\2\2\2\u030d\u0310\3\2\2\2\u030e\u030c\3\2\2\2\u030e")             # <<<<<<<<<<<<<<
 *         buf.write("\u030f\3\2\2\2\u030f\u0311\3\2\2\2\u0310\u030e\3\2\2\2")
 *         buf.write("\u0311\u0312\7\20\2\2\u0312\u0313\7\67\2\2\u0313i\3\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 405, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__624, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 405, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":406
 *         buf.write("\2\2\2\u030a\u030d\3\2\2\2\u030b\u0309\3\2\2\2\u030c\u0301")
 *         buf.write("\3\2\2\2\u030d\u0310\3\2\2\2\u030e\u030c\3\2\2\2\u030e")
 *         buf.write("\u030f\3\2\2\2\u030f\u0311\3\2\2\2\u0310\u030e\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0311\u0312\7\20\2\2\u0312\u0313\7\67\2\2\u0313i\3\2")
 *         buf.write("\2\2\u0314\u0315\7\67\2\2\u0315\u0316\5n8\2\u0316\u031c")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 406, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__626, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 406, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":407
 *         buf.write("\3\2\2\2\u030d\u0310\3\2\2\2\u030e\u030c\3\2\2\2\u030e")
 *         buf.write("\u030f\3\2\2\2\u030f\u0311\3\2\2\2\u0310\u030e\3\2\2\2")
 *         buf.write("\u0311\u0312\7\20\2\2\u0312\u0313\7\67\2\2\u0313i\3\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\u0314\u0315\7\67\2\2\u0315\u0316\5n8\2\u0316\u031c")
 *         buf.write("\7\60\2\2\u0317\u0318\5X-\2\u0318\u0319\7\3\2\2\u0319")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 407, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__627, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 407, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":408
 *         buf.write("\u030f\3\2\2\2\u030f\u0311\3\2\2\2\u0310\u030e\3\2\2\2")
 *         buf.write("\u0311\u0312\7\20\2\2\u0312\u0313\7\67\2\2\u0313i\3\2")
 *         buf.write("\2\2\u0314\u0315\7\67\2\2\u0315\u0316\5n8\2\u0316\u031c")             # <<<<<<<<<<<<<<
 *         buf.write("\7\60\2\2\u0317\u0318\5X-\2\u0318\u0319\7\3\2\2\u0319")
 *         buf.write("\u031b\3\2\2\2\u031a\u0317\3\2\2\2\u031b\u031e\3\2\2\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 408, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__628, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 408, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":409
 *         buf.write("\u0311\u0312\7\20\2\2\u0312\u0313\7\67\2\2\u0313i\3\2")
 *         buf.write("\2\2\u0314\u0315\7\67\2\2\u0315\u0316\5n8\2\u0316\u031c")
 *         buf.write("\7\60\2\2\u0317\u0318\5X-\2\u0318\u0319\7\3\2\2\u0319")             # <<<<<<<<<<<<<<
 *         buf.write("\u031b\3\2\2\2\u031a\u0317\3\2\2\2\u031b\u031e\3\2\2\2")
 *         buf.write("\u031c\u031a\3\2\2\2\u031c\u031d\3\2\2\2\u031d\u032c\3")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 409, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__629, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 409, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":410
 *         buf.write("\2\2\u0314\u0315\7\67\2\2\u0315\u0316\5n8\2\u0316\u031c")
 *         buf.write("\7\60\2\2\u0317\u0318\5X-\2\u0318\u0319\7\3\2\2\u0319")
 *         buf.write("\u031b\3\2\2\2\u031a\u0317\3\2\2\2\u031b\u031e\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u031c\u031a\3\2\2\2\u031c\u031d\3\2\2\2\u031d\u032c\3")
 *         buf.write("\2\2\2\u031e\u031c\3\2\2\2\u031f\u0320\78\2\2\u0320\u0321")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 410, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__631, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 410, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":411
 *         buf.write("\7\60\2\2\u0317\u0318\5X-\2\u0318\u0319\7\3\2\2\u0319")
 *         buf.write("\u031b\3\2\2\2\u031a\u0317\3\2\2\2\u031b\u031e\3\2\2\2")
 *         buf.write("\u031c\u031a\3\2\2\2\u031c\u031d\3\2\2\2\u031d\u032c\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u031e\u031c\3\2\2\2\u031f\u0320\78\2\2\u0320\u0321")
 *         buf.write("\5n8\2\u0321\u0327\7\60\2\2\u0322\u0323\5X-\2\u0323\u0324")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 411, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__633, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 411, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":412
 *         buf.write("\u031b\3\2\2\2\u031a\u0317\3\2\2\2\u031b\u031e\3\2\2\2")
 *         buf.write("\u031c\u031a\3\2\2\2\u031c\u031d\3\2\2\2\u031d\u032c\3")
 *         buf.write("\2\2\2\u031e\u031c\3\2\2\2\u031f\u0320\78\2\2\u0320\u0321")             # <<<<<<<<<<<<<<
 *         buf.write("\5n8\2\u0321\u0327\7\60\2\2\u0322\u0323\5X-\2\u0323\u0324")
 *         buf.write("\7\3\2\2\u0324\u0326\3\2\2\2\u0325\u0322\3\2\2\2\u0326")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 412, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__634, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 412, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":413
 *         buf.write("\u031c\u031a\3\2\2\2\u031c\u031d\3\2\2\2\u031d\u032c\3")
 *         buf.write("\2\2\2\u031e\u031c\3\2\2\2\u031f\u0320\78\2\2\u0320\u0321")
 *         buf.write("\5n8\2\u0321\u0327\7\60\2\2\u0322\u0323\5X-\2\u0323\u0324")             # <<<<<<<<<<<<<<
 *         buf.write("\7\3\2\2\u0324\u0326\3\2\2\2\u0325\u0322\3\2\2\2\u0326")
 *         buf.write("\u0329\3\2\2\2\u0327\u0325\3\2\2\2\u0327\u0328\3\2\2\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 413, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__635, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 413, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":414
 *         buf.write("\2\2\2\u031e\u031c\3\2\2\2\u031f\u0320\78\2\2\u0320\u0321")
 *         buf.write("\5n8\2\u0321\u0327\7\60\2\2\u0322\u0323\5X-\2\u0323\u0324")
 *         buf.write("\7\3\2\2\u0324\u0326\3\2\2\2\u0325\u0322\3\2\2\2\u0326")             # <<<<<<<<<<<<<<
 *         buf.write("\u0329\3\2\2\2\u0327\u0325\3\2\2\2\u0327\u0328\3\2\2\2")
 *         buf.write("\u0328\u032b\3\2\2\2\u0329\u0327\3\2\2\2\u032a\u031f\3")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 414, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__637, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 414, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":415
 *         buf.write("\5n8\2\u0321\u0327\7\60\2\2\u0322\u0323\5X-\2\u0323\u0324")
 *         buf.write("\7\3\2\2\u0324\u0326\3\2\2\2\u0325\u0322\3\2\2\2\u0326")
 *         buf.write("\u0329\3\2\2\2\u0327\u0325\3\2\2\2\u0327\u0328\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0328\u032b\3\2\2\2\u0329\u0327\3\2\2\2\u032a\u031f\3")
 *         buf.write("\2\2\2\u032b\u032e\3\2\2\2\u032c\u032a\3\2\2\2\u032c\u032d")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 415, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__639, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 415, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":416
 *         buf.write("\7\3\2\2\u0324\u0326\3\2\2\2\u0325\u0322\3\2\2\2\u0326")
 *         buf.write("\u0329\3\2\2\2\u0327\u0325\3\2\2\2\u0327\u0328\3\2\2\2")
 *         buf.write("\u0328\u032b\3\2\2\2\u0329\u0327\3\2\2\2\u032a\u031f\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u032b\u032e\3\2\2\2\u032c\u032a\3\2\2\2\u032c\u032d")
 *         buf.write("\3\2\2\2\u032d\u032f\3\2\2\2\u032e\u032c\3\2\2\2\u032f")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 416, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__641, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 416, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":417
 *         buf.write("\u0329\3\2\2\2\u0327\u0325\3\2\2\2\u0327\u0328\3\2\2\2")
 *         buf.write("\u0328\u032b\3\2\2\2\u0329\u0327\3\2\2\2\u032a\u031f\3")
 *         buf.write("\2\2\2\u032b\u032e\3\2\2\2\u032c\u032a\3\2\2\2\u032c\u032d")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u032d\u032f\3\2\2\2\u032e\u032c\3\2\2\2\u032f")
 *         buf.write("\u0330\7\20\2\2\u0330\u0331\7\67\2\2\u0331k\3\2\2\2\u0332")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 417, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__643, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 417, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":418
 *         buf.write("\u0328\u032b\3\2\2\2\u0329\u0327\3\2\2\2\u032a\u031f\3")
 *         buf.write("\2\2\2\u032b\u032e\3\2\2\2\u032c\u032a\3\2\2\2\u032c\u032d")
 *         buf.write("\3\2\2\2\u032d\u032f\3\2\2\2\u032e\u032c\3\2\2\2\u032f")             # <<<<<<<<<<<<<<
 *         buf.write("\u0330\7\20\2\2\u0330\u0331\7\67\2\2\u0331k\3\2\2\2\u0332")
 *         buf.write("\u0333\79\2\2\u0333\u0334\7\23\2\2\u0334\u0335\5z>\2\u0335")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 418, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__645, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 418, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":419
 *         buf.write("\2\2\2\u032b\u032e\3\2\2\2\u032c\u032a\3\2\2\2\u032c\u032d")
 *         buf.write("\3\2\2\2\u032d\u032f\3\2\2\2\u032e\u032c\3\2\2\2\u032f")
 *         buf.write("\u0330\7\20\2\2\u0330\u0331\7\67\2\2\u0331k\3\2\2\2\u0332")             # <<<<<<<<<<<<<<
 *         buf.write("\u0333\79\2\2\u0333\u0334\7\23\2\2\u0334\u0335\5z>\2\u0335")
 *         buf.write("\u0336\7\27\2\2\u0336\u0337\5z>\2\u0337\u0338\7\25\2\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 419, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__646, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 419, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":420
 *         buf.write("\3\2\2\2\u032d\u032f\3\2\2\2\u032e\u032c\3\2\2\2\u032f")
 *         buf.write("\u0330\7\20\2\2\u0330\u0331\7\67\2\2\u0331k\3\2\2\2\u0332")
 *         buf.write("\u0333\79\2\2\u0333\u0334\7\23\2\2\u0334\u0335\5z>\2\u0335")             # <<<<<<<<<<<<<<
 *         buf.write("\u0336\7\27\2\2\u0336\u0337\5z>\2\u0337\u0338\7\25\2\2")
 *         buf.write("\u0338m\3\2\2\2\u0339\u034c\5p9\2\u033a\u033b\7*\2\2\u033b")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 420, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__647, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 420, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":421
 *         buf.write("\u0330\7\20\2\2\u0330\u0331\7\67\2\2\u0331k\3\2\2\2\u0332")
 *         buf.write("\u0333\79\2\2\u0333\u0334\7\23\2\2\u0334\u0335\5z>\2\u0335")
 *         buf.write("\u0336\7\27\2\2\u0336\u0337\5z>\2\u0337\u0338\7\25\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0338m\3\2\2\2\u0339\u034c\5p9\2\u033a\u033b\7*\2\2\u033b")
 *         buf.write("\u033c\5n8\2\u033c\u033d\7\60\2\2\u033d\u0345\5n8\2\u033e")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 421, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__648, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 421, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":422
 *         buf.write("\u0333\79\2\2\u0333\u0334\7\23\2\2\u0334\u0335\5z>\2\u0335")
 *         buf.write("\u0336\7\27\2\2\u0336\u0337\5z>\2\u0337\u0338\7\25\2\2")
 *         buf.write("\u0338m\3\2\2\2\u0339\u034c\5p9\2\u033a\u033b\7*\2\2\u033b")             # <<<<<<<<<<<<<<
 *         buf.write("\u033c\5n8\2\u033c\u033d\7\60\2\2\u033d\u0345\5n8\2\u033e")
 *         buf.write("\u033f\7\61\2\2\u033f\u0340\5n8\2\u0340\u0341\7\60\2\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 422, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__649, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 422, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":423
 *         buf.write("\u0336\7\27\2\2\u0336\u0337\5z>\2\u0337\u0338\7\25\2\2")
 *         buf.write("\u0338m\3\2\2\2\u0339\u034c\5p9\2\u033a\u033b\7*\2\2\u033b")
 *         buf.write("\u033c\5n8\2\u033c\u033d\7\60\2\2\u033d\u0345\5n8\2\u033e")             # <<<<<<<<<<<<<<
 *         buf.write("\u033f\7\61\2\2\u033f\u0340\5n8\2\u0340\u0341\7\60\2\2")
 *         buf.write("\u0341\u0342\5n8\2\u0342\u0344\3\2\2\2\u0343\u033e\3\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 423, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__650, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 423, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":424
 *         buf.write("\u0338m\3\2\2\2\u0339\u034c\5p9\2\u033a\u033b\7*\2\2\u033b")
 *         buf.write("\u033c\5n8\2\u033c\u033d\7\60\2\2\u033d\u0345\5n8\2\u033e")
 *         buf.write("\u033f\7\61\2\2\u033f\u0340\5n8\2\u0340\u0341\7\60\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0341\u0342\5n8\2\u0342\u0344\3\2\2\2\u0343\u033e\3\2")
 *         buf.write("\2\2\u0344\u0347\3\2\2\2\u0345\u0343\3\2\2\2\u0345\u0346")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 424, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__651, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 424, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":425
 *         buf.write("\u033c\5n8\2\u033c\u033d\7\60\2\2\u033d\u0345\5n8\2\u033e")
 *         buf.write("\u033f\7\61\2\2\u033f\u0340\5n8\2\u0340\u0341\7\60\2\2")
 *         buf.write("\u0341\u0342\5n8\2\u0342\u0344\3\2\2\2\u0343\u033e\3\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\u0344\u0347\3\2\2\2\u0345\u0343\3\2\2\2\u0345\u0346")
 *         buf.write("\3\2\2\2\u0346\u0348\3\2\2\2\u0347\u0345\3\2\2\2\u0348")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 425, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__652, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 425, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":426
 *         buf.write("\u033f\7\61\2\2\u033f\u0340\5n8\2\u0340\u0341\7\60\2\2")
 *         buf.write("\u0341\u0342\5n8\2\u0342\u0344\3\2\2\2\u0343\u033e\3\2")
 *         buf.write("\2\2\u0344\u0347\3\2\2\2\u0345\u0343\3\2\2\2\u0345\u0346")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0346\u0348\3\2\2\2\u0347\u0345\3\2\2\2\u0348")
 *         buf.write("\u0349\7\62\2\2\u0349\u034a\5n8\2\u034a\u034c\3\2\2\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 426, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__654, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 426, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":427
 *         buf.write("\u0341\u0342\5n8\2\u0342\u0344\3\2\2\2\u0343\u033e\3\2")
 *         buf.write("\2\2\u0344\u0347\3\2\2\2\u0345\u0343\3\2\2\2\u0345\u0346")
 *         buf.write("\3\2\2\2\u0346\u0348\3\2\2\2\u0347\u0345\3\2\2\2\u0348")             # <<<<<<<<<<<<<<
 *         buf.write("\u0349\7\62\2\2\u0349\u034a\5n8\2\u034a\u034c\3\2\2\2")
 *         buf.write("\u034b\u0339\3\2\2\2\u034b\u033a\3\2\2\2\u034co\3\2\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 427, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__656, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 427, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":428
 *         buf.write("\2\2\u0344\u0347\3\2\2\2\u0345\u0343\3\2\2\2\u0345\u0346")
 *         buf.write("\3\2\2\2\u0346\u0348\3\2\2\2\u0347\u0345\3\2\2\2\u0348")
 *         buf.write("\u0349\7\62\2\2\u0349\u034a\5n8\2\u034a\u034c\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u034b\u0339\3\2\2\2\u034b\u033a\3\2\2\2\u034co\3\2\2")
 *         buf.write("\2\u034d\u0354\5r:\2\u034e\u034f\7\24\2\2\u034f\u0352")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 428, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__657, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 428, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":429
 *         buf.write("\3\2\2\2\u0346\u0348\3\2\2\2\u0347\u0345\3\2\2\2\u0348")
 *         buf.write("\u0349\7\62\2\2\u0349\u034a\5n8\2\u034a\u034c\3\2\2\2")
 *         buf.write("\u034b\u0339\3\2\2\2\u034b\u033a\3\2\2\2\u034co\3\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u034d\u0354\5r:\2\u034e\u034f\7\24\2\2\u034f\u0352")
 *         buf.write("\5r:\2\u0350\u0351\7\24\2\2\u0351\u0353\5r:\2\u0352\u0350")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 429, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__658, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 429, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":430
 *         buf.write("\u0349\7\62\2\2\u0349\u034a\5n8\2\u034a\u034c\3\2\2\2")
 *         buf.write("\u034b\u0339\3\2\2\2\u034b\u033a\3\2\2\2\u034co\3\2\2")
 *         buf.write("\2\u034d\u0354\5r:\2\u034e\u034f\7\24\2\2\u034f\u0352")             # <<<<<<<<<<<<<<
 *         buf.write("\5r:\2\u0350\u0351\7\24\2\2\u0351\u0353\5r:\2\u0352\u0350")
 *         buf.write("\3\2\2\2\u0352\u0353\3\2\2\2\u0353\u0355\3\2\2\2\u0354")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 430, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__659, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 430, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":431
 *         buf.write("\u034b\u0339\3\2\2\2\u034b\u033a\3\2\2\2\u034co\3\2\2")
 *         buf.write("\2\u034d\u0354\5r:\2\u034e\u034f\7\24\2\2\u034f\u0352")
 *         buf.write("\5r:\2\u0350\u0351\7\24\2\2\u0351\u0353\5r:\2\u0352\u0350")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0352\u0353\3\2\2\2\u0353\u0355\3\2\2\2\u0354")
 *         buf.write("\u034e\3\2\2\2\u0354\u0355\3\2\2\2\u0355q\3\2\2\2\u0356")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 431, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__660, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 431, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":432
 *         buf.write("\2\u034d\u0354\5r:\2\u034e\u034f\7\24\2\2\u034f\u0352")
 *         buf.write("\5r:\2\u0350\u0351\7\24\2\2\u0351\u0353\5r:\2\u0352\u0350")
 *         buf.write("\3\2\2\2\u0352\u0353\3\2\2\2\u0353\u0355\3\2\2\2\u0354")             # <<<<<<<<<<<<<<
 *         buf.write("\u034e\3\2\2\2\u0354\u0355\3\2\2\2\u0355q\3\2\2\2\u0356")
 *         buf.write("\u0357\b:\1\2\u0357\u0358\7:\2\2\u0358\u0361\5r:\13\u0359")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 432, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__662, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 432, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":433
 *         buf.write("\5r:\2\u0350\u0351\7\24\2\2\u0351\u0353\5r:\2\u0352\u0350")
 *         buf.write("\3\2\2\2\u0352\u0353\3\2\2\2\u0353\u0355\3\2\2\2\u0354")
 *         buf.write("\u034e\3\2\2\2\u0354\u0355\3\2\2\2\u0355q\3\2\2\2\u0356")             # <<<<<<<<<<<<<<
 *         buf.write("\u0357\b:\1\2\u0357\u0358\7:\2\2\u0358\u0361\5r:\13\u0359")
 *         buf.write("\u035a\5t;\2\u035a\u035b\t\6\2\2\u035b\u035c\5t;\2\u035c")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 433, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__663, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 433, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":434
 *         buf.write("\3\2\2\2\u0352\u0353\3\2\2\2\u0353\u0355\3\2\2\2\u0354")
 *         buf.write("\u034e\3\2\2\2\u0354\u0355\3\2\2\2\u0355q\3\2\2\2\u0356")
 *         buf.write("\u0357\b:\1\2\u0357\u0358\7:\2\2\u0358\u0361\5r:\13\u0359")             # <<<<<<<<<<<<<<
 *         buf.write("\u035a\5t;\2\u035a\u035b\t\6\2\2\u035b\u035c\5t;\2\u035c")
 *         buf.write("\u0361\3\2\2\2\u035d\u035e\7I\2\2\u035e\u0361\5r:\6\u035f")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 434, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__664, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 434, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":435
 *         buf.write("\u034e\3\2\2\2\u0354\u0355\3\2\2\2\u0355q\3\2\2\2\u0356")
 *         buf.write("\u0357\b:\1\2\u0357\u0358\7:\2\2\u0358\u0361\5r:\13\u0359")
 *         buf.write("\u035a\5t;\2\u035a\u035b\t\6\2\2\u035b\u035c\5t;\2\u035c")             # <<<<<<<<<<<<<<
 *         buf.write("\u0361\3\2\2\2\u035d\u035e\7I\2\2\u035e\u0361\5r:\6\u035f")
 *         buf.write("\u0361\5t;\2\u0360\u0356\3\2\2\2\u0360\u0359\3\2\2\2\u0360")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 435, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__665, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 435, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":436
 *         buf.write("\u0357\b:\1\2\u0357\u0358\7:\2\2\u0358\u0361\5r:\13\u0359")
 *         buf.write("\u035a\5t;\2\u035a\u035b\t\6\2\2\u035b\u035c\5t;\2\u035c")
 *         buf.write("\u0361\3\2\2\2\u035d\u035e\7I\2\2\u035e\u0361\5r:\6\u035f")             # <<<<<<<<<<<<<<
 *         buf.write("\u0361\5t;\2\u0360\u0356\3\2\2\2\u0360\u0359\3\2\2\2\u0360")
 *         buf.write("\u035d\3\2\2\2\u0360\u035f\3\2\2\2\u0361\u0373\3\2\2\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 436, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__666, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 436, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":437
 *         buf.write("\u035a\5t;\2\u035a\u035b\t\6\2\2\u035b\u035c\5t;\2\u035c")
 *         buf.write("\u0361\3\2\2\2\u035d\u035e\7I\2\2\u035e\u0361\5r:\6\u035f")
 *         buf.write("\u0361\5t;\2\u0360\u0356\3\2\2\2\u0360\u0359\3\2\2\2\u0360")             # <<<<<<<<<<<<<<
 *         buf.write("\u035d\3\2\2\2\u0360\u035f\3\2\2\2\u0361\u0373\3\2\2\2")
 *         buf.write("\u0362\u0363\f\t\2\2\u0363\u0364\t\7\2\2\u0364\u0372\5")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 437, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__667, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 437, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":438
 *         buf.write("\u0361\3\2\2\2\u035d\u035e\7I\2\2\u035e\u0361\5r:\6\u035f")
 *         buf.write("\u0361\5t;\2\u0360\u0356\3\2\2\2\u0360\u0359\3\2\2\2\u0360")
 *         buf.write("\u035d\3\2\2\2\u0360\u035f\3\2\2\2\u0361\u0373\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0362\u0363\f\t\2\2\u0363\u0364\t\7\2\2\u0364\u0372\5")
 *         buf.write("r:\n\u0365\u0366\f\b\2\2\u0366\u0367\t\b\2\2\u0367\u0372")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 438, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__669, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 438, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":439
 *         buf.write("\u0361\5t;\2\u0360\u0356\3\2\2\2\u0360\u0359\3\2\2\2\u0360")
 *         buf.write("\u035d\3\2\2\2\u0360\u035f\3\2\2\2\u0361\u0373\3\2\2\2")
 *         buf.write("\u0362\u0363\f\t\2\2\u0363\u0364\t\7\2\2\u0364\u0372\5")             # <<<<<<<<<<<<<<
 *         buf.write("r:\n\u0365\u0366\f\b\2\2\u0366\u0367\t\b\2\2\u0367\u0372")
 *         buf.write("\5r:\t\u0368\u0369\f\7\2\2\u0369\u036a\t\t\2\2\u036a\u0372")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 439, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__671, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 439, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":440
 *         buf.write("\u035d\3\2\2\2\u0360\u035f\3\2\2\2\u0361\u0373\3\2\2\2")
 *         buf.write("\u0362\u0363\f\t\2\2\u0363\u0364\t\7\2\2\u0364\u0372\5")
 *         buf.write("r:\n\u0365\u0366\f\b\2\2\u0366\u0367\t\b\2\2\u0367\u0372")             # <<<<<<<<<<<<<<
 *         buf.write("\5r:\t\u0368\u0369\f\7\2\2\u0369\u036a\t\t\2\2\u036a\u0372")
 *         buf.write("\5r:\b\u036b\u036c\f\5\2\2\u036c\u036d\7J\2\2\u036d\u0372")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 440, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__672, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 440, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":441
 *         buf.write("\u0362\u0363\f\t\2\2\u0363\u0364\t\7\2\2\u0364\u0372\5")
 *         buf.write("r:\n\u0365\u0366\f\b\2\2\u0366\u0367\t\b\2\2\u0367\u0372")
 *         buf.write("\5r:\t\u0368\u0369\f\7\2\2\u0369\u036a\t\t\2\2\u036a\u0372")             # <<<<<<<<<<<<<<
 *         buf.write("\5r:\b\u036b\u036c\f\5\2\2\u036c\u036d\7J\2\2\u036d\u0372")
 *         buf.write("\5r:\6\u036e\u036f\f\4\2\2\u036f\u0370\7K\2\2\u0370\u0372")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 441, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__673, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 441, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":442
 *         buf.write("r:\n\u0365\u0366\f\b\2\2\u0366\u0367\t\b\2\2\u0367\u0372")
 *         buf.write("\5r:\t\u0368\u0369\f\7\2\2\u0369\u036a\t\t\2\2\u036a\u0372")
 *         buf.write("\5r:\b\u036b\u036c\f\5\2\2\u036c\u036d\7J\2\2\u036d\u0372")             # <<<<<<<<<<<<<<
 *         buf.write("\5r:\6\u036e\u036f\f\4\2\2\u036f\u0370\7K\2\2\u0370\u0372")
 *         buf.write("\5r:\5\u0371\u0362\3\2\2\2\u0371\u0365\3\2\2\2\u0371\u0368")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 442, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__674, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 442, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":443
 *         buf.write("\5r:\t\u0368\u0369\f\7\2\2\u0369\u036a\t\t\2\2\u036a\u0372")
 *         buf.write("\5r:\b\u036b\u036c\f\5\2\2\u036c\u036d\7J\2\2\u036d\u0372")
 *         buf.write("\5r:\6\u036e\u036f\f\4\2\2\u036f\u0370\7K\2\2\u0370\u0372")             # <<<<<<<<<<<<<<
 *         buf.write("\5r:\5\u0371\u0362\3\2\2\2\u0371\u0365\3\2\2\2\u0371\u0368")
 *         buf.write("\3\2\2\2\u0371\u036b\3\2\2\2\u0371\u036e\3\2\2\2\u0372")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 443, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__675, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 443, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":444
 *         buf.write("\5r:\b\u036b\u036c\f\5\2\2\u036c\u036d\7J\2\2\u036d\u0372")
 *         buf.write("\5r:\6\u036e\u036f\f\4\2\2\u036f\u0370\7K\2\2\u0370\u0372")
 *         buf.write("\5r:\5\u0371\u0362\3\2\2\2\u0371\u0365\3\2\2\2\u0371\u0368")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0371\u036b\3\2\2\2\u0371\u036e\3\2\2\2\u0372")
 *         buf.write("\u0375\3\2\2\2\u0373\u0371\3\2\2\2\u0373\u0374\3\2\2\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 444, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__676, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 444, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":445
 *         buf.write("\5r:\6\u036e\u036f\f\4\2\2\u036f\u0370\7K\2\2\u0370\u0372")
 *         buf.write("\5r:\5\u0371\u0362\3\2\2\2\u0371\u0365\3\2\2\2\u0371\u0368")
 *         buf.write("\3\2\2\2\u0371\u036b\3\2\2\2\u0371\u036e\3\2\2\2\u0372")             # <<<<<<<<<<<<<<
 *         buf.write("\u0375\3\2\2\2\u0373\u0371\3\2\2\2\u0373\u0374\3\2\2\2")
 *         buf.write("\u0374s\3\2\2\2\u0375\u0373\3\2\2\2\u0376\u0397\7\\\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 445, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__678, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 445, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":446
 *         buf.write("\5r:\5\u0371\u0362\3\2\2\2\u0371\u0365\3\2\2\2\u0371\u0368")
 *         buf.write("\3\2\2\2\u0371\u036b\3\2\2\2\u0371\u036e\3\2\2\2\u0372")
 *         buf.write("\u0375\3\2\2\2\u0373\u0371\3\2\2\2\u0373\u0374\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0374s\3\2\2\2\u0375\u0373\3\2\2\2\u0376\u0397\7\\\2")
 *         buf.write("\2\u0377\u0397\7[\2\2\u0378\u0397\7L\2\2\u0379\u0397\7")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 446, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__680, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 446, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":447
 *         buf.write("\3\2\2\2\u0371\u036b\3\2\2\2\u0371\u036e\3\2\2\2\u0372")
 *         buf.write("\u0375\3\2\2\2\u0373\u0371\3\2\2\2\u0373\u0374\3\2\2\2")
 *         buf.write("\u0374s\3\2\2\2\u0375\u0373\3\2\2\2\u0376\u0397\7\\\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u0377\u0397\7[\2\2\u0378\u0397\7L\2\2\u0379\u0397\7")
 *         buf.write("M\2\2\u037a\u037b\5z>\2\u037b\u037c\5|?\2\u037c\u0397")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 447, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__681, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 447, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":448
 *         buf.write("\u0375\3\2\2\2\u0373\u0371\3\2\2\2\u0373\u0374\3\2\2\2")
 *         buf.write("\u0374s\3\2\2\2\u0375\u0373\3\2\2\2\u0376\u0397\7\\\2")
 *         buf.write("\2\u0377\u0397\7[\2\2\u0378\u0397\7L\2\2\u0379\u0397\7")             # <<<<<<<<<<<<<<
 *         buf.write("M\2\2\u037a\u037b\5z>\2\u037b\u037c\5|?\2\u037c\u0397")
 *         buf.write("\3\2\2\2\u037d\u037e\7\26\2\2\u037e\u0397\5|?\2\u037f")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 448, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__682, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 448, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":449
 *         buf.write("\u0374s\3\2\2\2\u0375\u0373\3\2\2\2\u0376\u0397\7\\\2")
 *         buf.write("\2\u0377\u0397\7[\2\2\u0378\u0397\7L\2\2\u0379\u0397\7")
 *         buf.write("M\2\2\u037a\u037b\5z>\2\u037b\u037c\5|?\2\u037c\u0397")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u037d\u037e\7\26\2\2\u037e\u0397\5|?\2\u037f")
 *         buf.write("\u0380\7Y\2\2\u0380\u0397\5|?\2\u0381\u0397\5z>\2\u0382")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 449, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__683, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 449, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":450
 *         buf.write("\2\u0377\u0397\7[\2\2\u0378\u0397\7L\2\2\u0379\u0397\7")
 *         buf.write("M\2\2\u037a\u037b\5z>\2\u037b\u037c\5|?\2\u037c\u0397")
 *         buf.write("\3\2\2\2\u037d\u037e\7\26\2\2\u037e\u0397\5|?\2\u037f")             # <<<<<<<<<<<<<<
 *         buf.write("\u0380\7Y\2\2\u0380\u0397\5|?\2\u0381\u0397\5z>\2\u0382")
 *         buf.write("\u0383\7\23\2\2\u0383\u0384\5\u0086D\2\u0384\u0385\7\25")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 450, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__685, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 450, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":451
 *         buf.write("M\2\2\u037a\u037b\5z>\2\u037b\u037c\5|?\2\u037c\u0397")
 *         buf.write("\3\2\2\2\u037d\u037e\7\26\2\2\u037e\u0397\5|?\2\u037f")
 *         buf.write("\u0380\7Y\2\2\u0380\u0397\5|?\2\u0381\u0397\5z>\2\u0382")             # <<<<<<<<<<<<<<
 *         buf.write("\u0383\7\23\2\2\u0383\u0384\5\u0086D\2\u0384\u0385\7\25")
 *         buf.write("\2\2\u0385\u0397\3\2\2\2\u0386\u0387\7N\2\2\u0387\u038c")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 451, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__686, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 451, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":452
 *         buf.write("\3\2\2\2\u037d\u037e\7\26\2\2\u037e\u0397\5|?\2\u037f")
 *         buf.write("\u0380\7Y\2\2\u0380\u0397\5|?\2\u0381\u0397\5z>\2\u0382")
 *         buf.write("\u0383\7\23\2\2\u0383\u0384\5\u0086D\2\u0384\u0385\7\25")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\u0385\u0397\3\2\2\2\u0386\u0387\7N\2\2\u0387\u038c")
 *         buf.write("\5\u0088E\2\u0388\u0389\7\3\2\2\u0389\u038b\5\u0088E\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 452, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__687, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 452, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":453
 *         buf.write("\u0380\7Y\2\2\u0380\u0397\5|?\2\u0381\u0397\5z>\2\u0382")
 *         buf.write("\u0383\7\23\2\2\u0383\u0384\5\u0086D\2\u0384\u0385\7\25")
 *         buf.write("\2\2\u0385\u0397\3\2\2\2\u0386\u0387\7N\2\2\u0387\u038c")             # <<<<<<<<<<<<<<
 *         buf.write("\5\u0088E\2\u0388\u0389\7\3\2\2\u0389\u038b\5\u0088E\2")
 *         buf.write("\u038a\u0388\3\2\2\2\u038b\u038e\3\2\2\2\u038c\u038a\3")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 453, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__688, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 453, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":454
 *         buf.write("\u0383\7\23\2\2\u0383\u0384\5\u0086D\2\u0384\u0385\7\25")
 *         buf.write("\2\2\u0385\u0397\3\2\2\2\u0386\u0387\7N\2\2\u0387\u038c")
 *         buf.write("\5\u0088E\2\u0388\u0389\7\3\2\2\u0389\u038b\5\u0088E\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u038a\u0388\3\2\2\2\u038b\u038e\3\2\2\2\u038c\u038a\3")
 *         buf.write("\2\2\2\u038c\u038d\3\2\2\2\u038d\u038f\3\2\2\2\u038e\u038c")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 454, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__689, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 454, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":455
 *         buf.write("\2\2\u0385\u0397\3\2\2\2\u0386\u0387\7N\2\2\u0387\u038c")
 *         buf.write("\5\u0088E\2\u0388\u0389\7\3\2\2\u0389\u038b\5\u0088E\2")
 *         buf.write("\u038a\u0388\3\2\2\2\u038b\u038e\3\2\2\2\u038c\u038a\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u038c\u038d\3\2\2\2\u038d\u038f\3\2\2\2\u038e\u038c")
 *         buf.write("\3\2\2\2\u038f\u0390\7O\2\2\u0390\u0397\3\2\2\2\u0391")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 455, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__691, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 455, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":456
 *         buf.write("\5\u0088E\2\u0388\u0389\7\3\2\2\u0389\u038b\5\u0088E\2")
 *         buf.write("\u038a\u0388\3\2\2\2\u038b\u038e\3\2\2\2\u038c\u038a\3")
 *         buf.write("\2\2\2\u038c\u038d\3\2\2\2\u038d\u038f\3\2\2\2\u038e\u038c")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u038f\u0390\7O\2\2\u0390\u0397\3\2\2\2\u0391")
 *         buf.write("\u0392\7 \2\2\u0392\u0393\5~@\2\u0393\u0394\7!\2\2\u0394")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 456, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__693, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 456, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":457
 *         buf.write("\u038a\u0388\3\2\2\2\u038b\u038e\3\2\2\2\u038c\u038a\3")
 *         buf.write("\2\2\2\u038c\u038d\3\2\2\2\u038d\u038f\3\2\2\2\u038e\u038c")
 *         buf.write("\3\2\2\2\u038f\u0390\7O\2\2\u0390\u0397\3\2\2\2\u0391")             # <<<<<<<<<<<<<<
 *         buf.write("\u0392\7 \2\2\u0392\u0393\5~@\2\u0393\u0394\7!\2\2\u0394")
 *         buf.write("\u0397\3\2\2\2\u0395\u0397\7\20\2\2\u0396\u0376\3\2\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 457, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__694, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 457, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":458
 *         buf.write("\2\2\2\u038c\u038d\3\2\2\2\u038d\u038f\3\2\2\2\u038e\u038c")
 *         buf.write("\3\2\2\2\u038f\u0390\7O\2\2\u0390\u0397\3\2\2\2\u0391")
 *         buf.write("\u0392\7 \2\2\u0392\u0393\5~@\2\u0393\u0394\7!\2\2\u0394")             # <<<<<<<<<<<<<<
 *         buf.write("\u0397\3\2\2\2\u0395\u0397\7\20\2\2\u0396\u0376\3\2\2")
 *         buf.write("\2\u0396\u0377\3\2\2\2\u0396\u0378\3\2\2\2\u0396\u0379")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 458, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__696, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 458, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":459
 *         buf.write("\3\2\2\2\u038f\u0390\7O\2\2\u0390\u0397\3\2\2\2\u0391")
 *         buf.write("\u0392\7 \2\2\u0392\u0393\5~@\2\u0393\u0394\7!\2\2\u0394")
 *         buf.write("\u0397\3\2\2\2\u0395\u0397\7\20\2\2\u0396\u0376\3\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u0396\u0377\3\2\2\2\u0396\u0378\3\2\2\2\u0396\u0379")
 *         buf.write("\3\2\2\2\u0396\u037a\3\2\2\2\u0396\u037d\3\2\2\2\u0396")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 459, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__698, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 459, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":460
 *         buf.write("\u0392\7 \2\2\u0392\u0393\5~@\2\u0393\u0394\7!\2\2\u0394")
 *         buf.write("\u0397\3\2\2\2\u0395\u0397\7\20\2\2\u0396\u0376\3\2\2")
 *         buf.write("\2\u0396\u0377\3\2\2\2\u0396\u0378\3\2\2\2\u0396\u0379")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0396\u037a\3\2\2\2\u0396\u037d\3\2\2\2\u0396")
 *         buf.write("\u037f\3\2\2\2\u0396\u0381\3\2\2\2\u0396\u0382\3\2\2\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 460, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__700, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 460, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":461
 *         buf.write("\u0397\3\2\2\2\u0395\u0397\7\20\2\2\u0396\u0376\3\2\2")
 *         buf.write("\2\u0396\u0377\3\2\2\2\u0396\u0378\3\2\2\2\u0396\u0379")
 *         buf.write("\3\2\2\2\u0396\u037a\3\2\2\2\u0396\u037d\3\2\2\2\u0396")             # <<<<<<<<<<<<<<
 *         buf.write("\u037f\3\2\2\2\u0396\u0381\3\2\2\2\u0396\u0382\3\2\2\2")
 *         buf.write("\u0396\u0386\3\2\2\2\u0396\u0391\3\2\2\2\u0396\u0395\3")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 461, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__702, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 461, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":462
 *         buf.write("\2\u0396\u0377\3\2\2\2\u0396\u0378\3\2\2\2\u0396\u0379")
 *         buf.write("\3\2\2\2\u0396\u037a\3\2\2\2\u0396\u037d\3\2\2\2\u0396")
 *         buf.write("\u037f\3\2\2\2\u0396\u0381\3\2\2\2\u0396\u0382\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0396\u0386\3\2\2\2\u0396\u0391\3\2\2\2\u0396\u0395\3")
 *         buf.write("\2\2\2\u0397u\3\2\2\2\u0398\u039a\7\36\2\2\u0399\u0398")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 462, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__704, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 462, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":463
 *         buf.write("\3\2\2\2\u0396\u037a\3\2\2\2\u0396\u037d\3\2\2\2\u0396")
 *         buf.write("\u037f\3\2\2\2\u0396\u0381\3\2\2\2\u0396\u0382\3\2\2\2")
 *         buf.write("\u0396\u0386\3\2\2\2\u0396\u0391\3\2\2\2\u0396\u0395\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u0397u\3\2\2\2\u0398\u039a\7\36\2\2\u0399\u0398")
 *         buf.write("\3\2\2\2\u0399\u039a\3\2\2\2\u039a\u039b\3\2\2\2\u039b")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 463, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__706, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 463, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":464
 *         buf.write("\u037f\3\2\2\2\u0396\u0381\3\2\2\2\u0396\u0382\3\2\2\2")
 *         buf.write("\u0396\u0386\3\2\2\2\u0396\u0391\3\2\2\2\u0396\u0395\3")
 *         buf.write("\2\2\2\u0397u\3\2\2\2\u0398\u039a\7\36\2\2\u0399\u0398")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0399\u039a\3\2\2\2\u039a\u039b\3\2\2\2\u039b")
 *         buf.write("\u03a0\7Z\2\2\u039c\u039d\7\36\2\2\u039d\u039f\7Z\2\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 464, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__707, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 464, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":465
 *         buf.write("\u0396\u0386\3\2\2\2\u0396\u0391\3\2\2\2\u0396\u0395\3")
 *         buf.write("\2\2\2\u0397u\3\2\2\2\u0398\u039a\7\36\2\2\u0399\u0398")
 *         buf.write("\3\2\2\2\u0399\u039a\3\2\2\2\u039a\u039b\3\2\2\2\u039b")             # <<<<<<<<<<<<<<
 *         buf.write("\u03a0\7Z\2\2\u039c\u039d\7\36\2\2\u039d\u039f\7Z\2\2")
 *         buf.write("\u039e\u039c\3\2\2\2\u039f\u03a2\3\2\2\2\u03a0\u039e\3")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 465, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__709, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 465, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":466
 *         buf.write("\2\2\2\u0397u\3\2\2\2\u0398\u039a\7\36\2\2\u0399\u0398")
 *         buf.write("\3\2\2\2\u0399\u039a\3\2\2\2\u039a\u039b\3\2\2\2\u039b")
 *         buf.write("\u03a0\7Z\2\2\u039c\u039d\7\36\2\2\u039d\u039f\7Z\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u039e\u039c\3\2\2\2\u039f\u03a2\3\2\2\2\u03a0\u039e\3")
 *         buf.write("\2\2\2\u03a0\u03a1\3\2\2\2\u03a1w\3\2\2\2\u03a2\u03a0")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 466, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__710, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 466, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":467
 *         buf.write("\3\2\2\2\u0399\u039a\3\2\2\2\u039a\u039b\3\2\2\2\u039b")
 *         buf.write("\u03a0\7Z\2\2\u039c\u039d\7\36\2\2\u039d\u039f\7Z\2\2")
 *         buf.write("\u039e\u039c\3\2\2\2\u039f\u03a2\3\2\2\2\u03a0\u039e\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u03a0\u03a1\3\2\2\2\u03a1w\3\2\2\2\u03a2\u03a0")
 *         buf.write("\3\2\2\2\u03a3\u03a5\7Z\2\2\u03a4\u03a6\5\u008aF\2\u03a5")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 467, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__712, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 467, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":468
 *         buf.write("\u03a0\7Z\2\2\u039c\u039d\7\36\2\2\u039d\u039f\7Z\2\2")
 *         buf.write("\u039e\u039c\3\2\2\2\u039f\u03a2\3\2\2\2\u03a0\u039e\3")
 *         buf.write("\2\2\2\u03a0\u03a1\3\2\2\2\u03a1w\3\2\2\2\u03a2\u03a0")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u03a3\u03a5\7Z\2\2\u03a4\u03a6\5\u008aF\2\u03a5")
 *         buf.write("\u03a4\3\2\2\2\u03a5\u03a6\3\2\2\2\u03a6y\3\2\2\2\u03a7")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 468, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__713, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 468, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":469
 *         buf.write("\u039e\u039c\3\2\2\2\u039f\u03a2\3\2\2\2\u03a0\u039e\3")
 *         buf.write("\2\2\2\u03a0\u03a1\3\2\2\2\u03a1w\3\2\2\2\u03a2\u03a0")
 *         buf.write("\3\2\2\2\u03a3\u03a5\7Z\2\2\u03a4\u03a6\5\u008aF\2\u03a5")             # <<<<<<<<<<<<<<
 *         buf.write("\u03a4\3\2\2\2\u03a5\u03a6\3\2\2\2\u03a6y\3\2\2\2\u03a7")
 *         buf.write("\u03ac\5x=\2\u03a8\u03a9\7\36\2\2\u03a9\u03ab\5x=\2\u03aa")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 469, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__714, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 469, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":470
 *         buf.write("\2\2\2\u03a0\u03a1\3\2\2\2\u03a1w\3\2\2\2\u03a2\u03a0")
 *         buf.write("\3\2\2\2\u03a3\u03a5\7Z\2\2\u03a4\u03a6\5\u008aF\2\u03a5")
 *         buf.write("\u03a4\3\2\2\2\u03a5\u03a6\3\2\2\2\u03a6y\3\2\2\2\u03a7")             # <<<<<<<<<<<<<<
 *         buf.write("\u03ac\5x=\2\u03a8\u03a9\7\36\2\2\u03a9\u03ab\5x=\2\u03aa")
 *         buf.write("\u03a8\3\2\2\2\u03ab\u03ae\3\2\2\2\u03ac\u03aa\3\2\2\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 470, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__715, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 470, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":471
 *         buf.write("\3\2\2\2\u03a3\u03a5\7Z\2\2\u03a4\u03a6\5\u008aF\2\u03a5")
 *         buf.write("\u03a4\3\2\2\2\u03a5\u03a6\3\2\2\2\u03a6y\3\2\2\2\u03a7")
 *         buf.write("\u03ac\5x=\2\u03a8\u03a9\7\36\2\2\u03a9\u03ab\5x=\2\u03aa")             # <<<<<<<<<<<<<<
 *         buf.write("\u03a8\3\2\2\2\u03ab\u03ae\3\2\2\2\u03ac\u03aa\3\2\2\2")
 *         buf.write("\u03ac\u03ad\3\2\2\2\u03ad{\3\2\2\2\u03ae\u03ac\3\2\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 471, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__716, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 471, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":472
 *         buf.write("\u03a4\3\2\2\2\u03a5\u03a6\3\2\2\2\u03a6y\3\2\2\2\u03a7")
 *         buf.write("\u03ac\5x=\2\u03a8\u03a9\7\36\2\2\u03a9\u03ab\5x=\2\u03aa")
 *         buf.write("\u03a8\3\2\2\2\u03ab\u03ae\3\2\2\2\u03ac\u03aa\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u03ac\u03ad\3\2\2\2\u03ad{\3\2\2\2\u03ae\u03ac\3\2\2")
 *         buf.write("\2\u03af\u03b1\7\23\2\2\u03b0\u03b2\5~@\2\u03b1\u03b0")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 472, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__718, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 472, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":473
 *         buf.write("\u03ac\5x=\2\u03a8\u03a9\7\36\2\2\u03a9\u03ab\5x=\2\u03aa")
 *         buf.write("\u03a8\3\2\2\2\u03ab\u03ae\3\2\2\2\u03ac\u03aa\3\2\2\2")
 *         buf.write("\u03ac\u03ad\3\2\2\2\u03ad{\3\2\2\2\u03ae\u03ac\3\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u03af\u03b1\7\23\2\2\u03b0\u03b2\5~@\2\u03b1\u03b0")
 *         buf.write("\3\2\2\2\u03b1\u03b2\3\2\2\2\u03b2\u03b3\3\2\2\2\u03b3")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 473, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__720, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 473, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":474
 *         buf.write("\u03a8\3\2\2\2\u03ab\u03ae\3\2\2\2\u03ac\u03aa\3\2\2\2")
 *         buf.write("\u03ac\u03ad\3\2\2\2\u03ad{\3\2\2\2\u03ae\u03ac\3\2\2")
 *         buf.write("\2\u03af\u03b1\7\23\2\2\u03b0\u03b2\5~@\2\u03b1\u03b0")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u03b1\u03b2\3\2\2\2\u03b2\u03b3\3\2\2\2\u03b3")
 *         buf.write("\u03b4\7\25\2\2\u03b4}\3\2\2\2\u03b5\u03bc\5\u0084C\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 474, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__722, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 474, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":475
 *         buf.write("\u03ac\u03ad\3\2\2\2\u03ad{\3\2\2\2\u03ae\u03ac\3\2\2")
 *         buf.write("\2\u03af\u03b1\7\23\2\2\u03b0\u03b2\5~@\2\u03b1\u03b0")
 *         buf.write("\3\2\2\2\u03b1\u03b2\3\2\2\2\u03b2\u03b3\3\2\2\2\u03b3")             # <<<<<<<<<<<<<<
 *         buf.write("\u03b4\7\25\2\2\u03b4}\3\2\2\2\u03b5\u03bc\5\u0084C\2")
 *         buf.write("\u03b6\u03b7\7\27\2\2\u03b7\u03bb\5\u0084C\2\u03b8\u03b9")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 475, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__724, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 475, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":476
 *         buf.write("\2\u03af\u03b1\7\23\2\2\u03b0\u03b2\5~@\2\u03b1\u03b0")
 *         buf.write("\3\2\2\2\u03b1\u03b2\3\2\2\2\u03b2\u03b3\3\2\2\2\u03b3")
 *         buf.write("\u03b4\7\25\2\2\u03b4}\3\2\2\2\u03b5\u03bc\5\u0084C\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u03b6\u03b7\7\27\2\2\u03b7\u03bb\5\u0084C\2\u03b8\u03b9")
 *         buf.write("\7\63\2\2\u03b9\u03bb\5b\62\2\u03ba\u03b6\3\2\2\2\u03ba")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 476, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__725, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 476, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":477
 *         buf.write("\3\2\2\2\u03b1\u03b2\3\2\2\2\u03b2\u03b3\3\2\2\2\u03b3")
 *         buf.write("\u03b4\7\25\2\2\u03b4}\3\2\2\2\u03b5\u03bc\5\u0084C\2")
 *         buf.write("\u03b6\u03b7\7\27\2\2\u03b7\u03bb\5\u0084C\2\u03b8\u03b9")             # <<<<<<<<<<<<<<
 *         buf.write("\7\63\2\2\u03b9\u03bb\5b\62\2\u03ba\u03b6\3\2\2\2\u03ba")
 *         buf.write("\u03b8\3\2\2\2\u03bb\u03be\3\2\2\2\u03bc\u03ba\3\2\2\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 477, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__726, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 477, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":478
 *         buf.write("\u03b4\7\25\2\2\u03b4}\3\2\2\2\u03b5\u03bc\5\u0084C\2")
 *         buf.write("\u03b6\u03b7\7\27\2\2\u03b7\u03bb\5\u0084C\2\u03b8\u03b9")
 *         buf.write("\7\63\2\2\u03b9\u03bb\5b\62\2\u03ba\u03b6\3\2\2\2\u03ba")             # <<<<<<<<<<<<<<
 *         buf.write("\u03b8\3\2\2\2\u03bb\u03be\3\2\2\2\u03bc\u03ba\3\2\2\2")
 *         buf.write("\u03bc\u03bd\3\2\2\2\u03bd\u03c1\3\2\2\2\u03be\u03bc\3")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 478, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__727, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 478, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":479
 *         buf.write("\u03b6\u03b7\7\27\2\2\u03b7\u03bb\5\u0084C\2\u03b8\u03b9")
 *         buf.write("\7\63\2\2\u03b9\u03bb\5b\62\2\u03ba\u03b6\3\2\2\2\u03ba")
 *         buf.write("\u03b8\3\2\2\2\u03bb\u03be\3\2\2\2\u03bc\u03ba\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u03bc\u03bd\3\2\2\2\u03bd\u03c1\3\2\2\2\u03be\u03bc\3")
 *         buf.write("\2\2\2\u03bf\u03c1\5\u0080A\2\u03c0\u03b5\3\2\2\2\u03c0")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 479, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__729, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 479, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":480
 *         buf.write("\7\63\2\2\u03b9\u03bb\5b\62\2\u03ba\u03b6\3\2\2\2\u03ba")
 *         buf.write("\u03b8\3\2\2\2\u03bb\u03be\3\2\2\2\u03bc\u03ba\3\2\2\2")
 *         buf.write("\u03bc\u03bd\3\2\2\2\u03bd\u03c1\3\2\2\2\u03be\u03bc\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u03bf\u03c1\5\u0080A\2\u03c0\u03b5\3\2\2\2\u03c0")
 *         buf.write("\u03bf\3\2\2\2\u03c1\177\3\2\2\2\u03c2\u03c7\5\u0082B")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 480, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__731, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 480, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":481
 *         buf.write("\u03b8\3\2\2\2\u03bb\u03be\3\2\2\2\u03bc\u03ba\3\2\2\2")
 *         buf.write("\u03bc\u03bd\3\2\2\2\u03bd\u03c1\3\2\2\2\u03be\u03bc\3")
 *         buf.write("\2\2\2\u03bf\u03c1\5\u0080A\2\u03c0\u03b5\3\2\2\2\u03c0")             # <<<<<<<<<<<<<<
 *         buf.write("\u03bf\3\2\2\2\u03c1\177\3\2\2\2\u03c2\u03c7\5\u0082B")
 *         buf.write("\2\u03c3\u03c4\7\27\2\2\u03c4\u03c6\5\u0082B\2\u03c5\u03c3")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 481, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__732, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 481, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":482
 *         buf.write("\u03bc\u03bd\3\2\2\2\u03bd\u03c1\3\2\2\2\u03be\u03bc\3")
 *         buf.write("\2\2\2\u03bf\u03c1\5\u0080A\2\u03c0\u03b5\3\2\2\2\u03c0")
 *         buf.write("\u03bf\3\2\2\2\u03c1\177\3\2\2\2\u03c2\u03c7\5\u0082B")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u03c3\u03c4\7\27\2\2\u03c4\u03c6\5\u0082B\2\u03c5\u03c3")
 *         buf.write("\3\2\2\2\u03c6\u03c9\3\2\2\2\u03c7\u03c5\3\2\2\2\u03c7")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 482, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__733, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 482, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":483
 *         buf.write("\2\2\2\u03bf\u03c1\5\u0080A\2\u03c0\u03b5\3\2\2\2\u03c0")
 *         buf.write("\u03bf\3\2\2\2\u03c1\177\3\2\2\2\u03c2\u03c7\5\u0082B")
 *         buf.write("\2\u03c3\u03c4\7\27\2\2\u03c4\u03c6\5\u0082B\2\u03c5\u03c3")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u03c6\u03c9\3\2\2\2\u03c7\u03c5\3\2\2\2\u03c7")
 *         buf.write("\u03c8\3\2\2\2\u03c8\u0081\3\2\2\2\u03c9\u03c7\3\2\2\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 483, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__734, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 483, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":484
 *         buf.write("\u03bf\3\2\2\2\u03c1\177\3\2\2\2\u03c2\u03c7\5\u0082B")
 *         buf.write("\2\u03c3\u03c4\7\27\2\2\u03c4\u03c6\5\u0082B\2\u03c5\u03c3")
 *         buf.write("\3\2\2\2\u03c6\u03c9\3\2\2\2\u03c7\u03c5\3\2\2\2\u03c7")             # <<<<<<<<<<<<<<
 *         buf.write("\u03c8\3\2\2\2\u03c8\u0081\3\2\2\2\u03c9\u03c7\3\2\2\2")
 *         buf.write("\u03ca\u03cb\7Z\2\2\u03cb\u03cc\7\21\2\2\u03cc\u03cd\5")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 484, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__736, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 484, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":485
 *         buf.write("\2\u03c3\u03c4\7\27\2\2\u03c4\u03c6\5\u0082B\2\u03c5\u03c3")
 *         buf.write("\3\2\2\2\u03c6\u03c9\3\2\2\2\u03c7\u03c5\3\2\2\2\u03c7")
 *         buf.write("\u03c8\3\2\2\2\u03c8\u0081\3\2\2\2\u03c9\u03c7\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u03ca\u03cb\7Z\2\2\u03cb\u03cc\7\21\2\2\u03cc\u03cd\5")
 *         buf.write("\u0084C\2\u03cd\u0083\3\2\2\2\u03ce\u03cf\7\17\2\2\u03cf")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 485, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__738, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 485, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":486
 *         buf.write("\3\2\2\2\u03c6\u03c9\3\2\2\2\u03c7\u03c5\3\2\2\2\u03c7")
 *         buf.write("\u03c8\3\2\2\2\u03c8\u0081\3\2\2\2\u03c9\u03c7\3\2\2\2")
 *         buf.write("\u03ca\u03cb\7Z\2\2\u03cb\u03cc\7\21\2\2\u03cc\u03cd\5")             # <<<<<<<<<<<<<<
 *         buf.write("\u0084C\2\u03cd\u0083\3\2\2\2\u03ce\u03cf\7\17\2\2\u03cf")
 *         buf.write("\u03d0\5v<\2\u03d0\u03d2\7\23\2\2\u03d1\u03d3\5\u0080")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 486, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__739, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 486, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":487
 *         buf.write("\u03c8\3\2\2\2\u03c8\u0081\3\2\2\2\u03c9\u03c7\3\2\2\2")
 *         buf.write("\u03ca\u03cb\7Z\2\2\u03cb\u03cc\7\21\2\2\u03cc\u03cd\5")
 *         buf.write("\u0084C\2\u03cd\u0083\3\2\2\2\u03ce\u03cf\7\17\2\2\u03cf")             # <<<<<<<<<<<<<<
 *         buf.write("\u03d0\5v<\2\u03d0\u03d2\7\23\2\2\u03d1\u03d3\5\u0080")
 *         buf.write("A\2\u03d2\u03d1\3\2\2\2\u03d2\u03d3\3\2\2\2\u03d3\u03d4")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 487, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__740, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 487, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":488
 *         buf.write("\u03ca\u03cb\7Z\2\2\u03cb\u03cc\7\21\2\2\u03cc\u03cd\5")
 *         buf.write("\u0084C\2\u03cd\u0083\3\2\2\2\u03ce\u03cf\7\17\2\2\u03cf")
 *         buf.write("\u03d0\5v<\2\u03d0\u03d2\7\23\2\2\u03d1\u03d3\5\u0080")             # <<<<<<<<<<<<<<
 *         buf.write("A\2\u03d2\u03d1\3\2\2\2\u03d2\u03d3\3\2\2\2\u03d3\u03d4")
 *         buf.write("\3\2\2\2\u03d4\u03d5\7\25\2\2\u03d5\u03d8\3\2\2\2\u03d6")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 488, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__741, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 488, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":489
 *         buf.write("\u0084C\2\u03cd\u0083\3\2\2\2\u03ce\u03cf\7\17\2\2\u03cf")
 *         buf.write("\u03d0\5v<\2\u03d0\u03d2\7\23\2\2\u03d1\u03d3\5\u0080")
 *         buf.write("A\2\u03d2\u03d1\3\2\2\2\u03d2\u03d3\3\2\2\2\u03d3\u03d4")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u03d4\u03d5\7\25\2\2\u03d5\u03d8\3\2\2\2\u03d6")
 *         buf.write("\u03d8\5n8\2\u03d7\u03ce\3\2\2\2\u03d7\u03d6\3\2\2\2\u03d8")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 489, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__742, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 489, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":490
 *         buf.write("\u03d0\5v<\2\u03d0\u03d2\7\23\2\2\u03d1\u03d3\5\u0080")
 *         buf.write("A\2\u03d2\u03d1\3\2\2\2\u03d2\u03d3\3\2\2\2\u03d3\u03d4")
 *         buf.write("\3\2\2\2\u03d4\u03d5\7\25\2\2\u03d5\u03d8\3\2\2\2\u03d6")             # <<<<<<<<<<<<<<
 *         buf.write("\u03d8\5n8\2\u03d7\u03ce\3\2\2\2\u03d7\u03d6\3\2\2\2\u03d8")
 *         buf.write("\u0085\3\2\2\2\u03d9\u03db\5n8\2\u03da\u03d9\3\2\2\2\u03da")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 490, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__744, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 490, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":491
 *         buf.write("A\2\u03d2\u03d1\3\2\2\2\u03d2\u03d3\3\2\2\2\u03d3\u03d4")
 *         buf.write("\3\2\2\2\u03d4\u03d5\7\25\2\2\u03d5\u03d8\3\2\2\2\u03d6")
 *         buf.write("\u03d8\5n8\2\u03d7\u03ce\3\2\2\2\u03d7\u03d6\3\2\2\2\u03d8")             # <<<<<<<<<<<<<<
 *         buf.write("\u0085\3\2\2\2\u03d9\u03db\5n8\2\u03da\u03d9\3\2\2\2\u03da")
 *         buf.write("\u03db\3\2\2\2\u03db\u03e0\3\2\2\2\u03dc\u03dd\7\27\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 491, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__745, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 491, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":492
 *         buf.write("\3\2\2\2\u03d4\u03d5\7\25\2\2\u03d5\u03d8\3\2\2\2\u03d6")
 *         buf.write("\u03d8\5n8\2\u03d7\u03ce\3\2\2\2\u03d7\u03d6\3\2\2\2\u03d8")
 *         buf.write("\u0085\3\2\2\2\u03d9\u03db\5n8\2\u03da\u03d9\3\2\2\2\u03da")             # <<<<<<<<<<<<<<
 *         buf.write("\u03db\3\2\2\2\u03db\u03e0\3\2\2\2\u03dc\u03dd\7\27\2")
 *         buf.write("\2\u03dd\u03df\5n8\2\u03de\u03dc\3\2\2\2\u03df\u03e2\3")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 492, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__746, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 492, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":493
 *         buf.write("\u03d8\5n8\2\u03d7\u03ce\3\2\2\2\u03d7\u03d6\3\2\2\2\u03d8")
 *         buf.write("\u0085\3\2\2\2\u03d9\u03db\5n8\2\u03da\u03d9\3\2\2\2\u03da")
 *         buf.write("\u03db\3\2\2\2\u03db\u03e0\3\2\2\2\u03dc\u03dd\7\27\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u03dd\u03df\5n8\2\u03de\u03dc\3\2\2\2\u03df\u03e2\3")
 *         buf.write("\2\2\2\u03e0\u03de\3\2\2\2\u03e0\u03e1\3\2\2\2\u03e1\u0087")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 493, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__748, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 493, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":494
 *         buf.write("\u0085\3\2\2\2\u03d9\u03db\5n8\2\u03da\u03d9\3\2\2\2\u03da")
 *         buf.write("\u03db\3\2\2\2\u03db\u03e0\3\2\2\2\u03dc\u03dd\7\27\2")
 *         buf.write("\2\u03dd\u03df\5n8\2\u03de\u03dc\3\2\2\2\u03df\u03e2\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u03e0\u03de\3\2\2\2\u03e0\u03e1\3\2\2\2\u03e1\u0087")
 *         buf.write("\3\2\2\2\u03e2\u03e0\3\2\2\2\u03e3\u03e8\5n8\2\u03e4\u03e5")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 494, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__749, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 494, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":495
 *         buf.write("\u03db\3\2\2\2\u03db\u03e0\3\2\2\2\u03dc\u03dd\7\27\2")
 *         buf.write("\2\u03dd\u03df\5n8\2\u03de\u03dc\3\2\2\2\u03df\u03e2\3")
 *         buf.write("\2\2\2\u03e0\u03de\3\2\2\2\u03e0\u03e1\3\2\2\2\u03e1\u0087")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u03e2\u03e0\3\2\2\2\u03e3\u03e8\5n8\2\u03e4\u03e5")
 *         buf.write("\7\27\2\2\u03e5\u03e7\5n8\2\u03e6\u03e4\3\2\2\2\u03e7")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 495, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__751, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 495, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":496
 *         buf.write("\2\u03dd\u03df\5n8\2\u03de\u03dc\3\2\2\2\u03df\u03e2\3")
 *         buf.write("\2\2\2\u03e0\u03de\3\2\2\2\u03e0\u03e1\3\2\2\2\u03e1\u0087")
 *         buf.write("\3\2\2\2\u03e2\u03e0\3\2\2\2\u03e3\u03e8\5n8\2\u03e4\u03e5")             # <<<<<<<<<<<<<<
 *         buf.write("\7\27\2\2\u03e5\u03e7\5n8\2\u03e6\u03e4\3\2\2\2\u03e7")
 *         buf.write("\u03ea\3\2\2\2\u03e8\u03e6\3\2\2\2\u03e8\u03e9\3\2\2\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 496, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__752, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 496, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":497
 *         buf.write("\2\2\2\u03e0\u03de\3\2\2\2\u03e0\u03e1\3\2\2\2\u03e1\u0087")
 *         buf.write("\3\2\2\2\u03e2\u03e0\3\2\2\2\u03e3\u03e8\5n8\2\u03e4\u03e5")
 *         buf.write("\7\27\2\2\u03e5\u03e7\5n8\2\u03e6\u03e4\3\2\2\2\u03e7")             # <<<<<<<<<<<<<<
 *         buf.write("\u03ea\3\2\2\2\u03e8\u03e6\3\2\2\2\u03e8\u03e9\3\2\2\2")
 *         buf.write("\u03e9\u0089\3\2\2\2\u03ea\u03e8\3\2\2\2\u03eb\u03ec\7")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 497, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__753, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 497, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":498
 *         buf.write("\3\2\2\2\u03e2\u03e0\3\2\2\2\u03e3\u03e8\5n8\2\u03e4\u03e5")
 *         buf.write("\7\27\2\2\u03e5\u03e7\5n8\2\u03e6\u03e4\3\2\2\2\u03e7")
 *         buf.write("\u03ea\3\2\2\2\u03e8\u03e6\3\2\2\2\u03e8\u03e9\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u03e9\u0089\3\2\2\2\u03ea\u03e8\3\2\2\2\u03eb\u03ec\7")
 *         buf.write("N\2\2\u03ec\u03f1\5\u008cG\2\u03ed\u03ee\7\27\2\2\u03ee")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 498, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__755, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 498, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":499
 *         buf.write("\7\27\2\2\u03e5\u03e7\5n8\2\u03e6\u03e4\3\2\2\2\u03e7")
 *         buf.write("\u03ea\3\2\2\2\u03e8\u03e6\3\2\2\2\u03e8\u03e9\3\2\2\2")
 *         buf.write("\u03e9\u0089\3\2\2\2\u03ea\u03e8\3\2\2\2\u03eb\u03ec\7")             # <<<<<<<<<<<<<<
 *         buf.write("N\2\2\u03ec\u03f1\5\u008cG\2\u03ed\u03ee\7\27\2\2\u03ee")
 *         buf.write("\u03f0\5\u008cG\2\u03ef\u03ed\3\2\2\2\u03f0\u03f3\3\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 499, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__757, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 499, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":500
 *         buf.write("\u03ea\3\2\2\2\u03e8\u03e6\3\2\2\2\u03e8\u03e9\3\2\2\2")
 *         buf.write("\u03e9\u0089\3\2\2\2\u03ea\u03e8\3\2\2\2\u03eb\u03ec\7")
 *         buf.write("N\2\2\u03ec\u03f1\5\u008cG\2\u03ed\u03ee\7\27\2\2\u03ee")             # <<<<<<<<<<<<<<
 *         buf.write("\u03f0\5\u008cG\2\u03ef\u03ed\3\2\2\2\u03f0\u03f3\3\2")
 *         buf.write("\2\2\u03f1\u03ef\3\2\2\2\u03f1\u03f2\3\2\2\2\u03f2\u03f4")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 500, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__758, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 500, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":501
 *         buf.write("\u03e9\u0089\3\2\2\2\u03ea\u03e8\3\2\2\2\u03eb\u03ec\7")
 *         buf.write("N\2\2\u03ec\u03f1\5\u008cG\2\u03ed\u03ee\7\27\2\2\u03ee")
 *         buf.write("\u03f0\5\u008cG\2\u03ef\u03ed\3\2\2\2\u03f0\u03f3\3\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\u03f1\u03ef\3\2\2\2\u03f1\u03f2\3\2\2\2\u03f2\u03f4")
 *         buf.write("\3\2\2\2\u03f3\u03f1\3\2\2\2\u03f4\u03f5\7O\2\2\u03f5")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 501, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__759, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 501, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":502
 *         buf.write("N\2\2\u03ec\u03f1\5\u008cG\2\u03ed\u03ee\7\27\2\2\u03ee")
 *         buf.write("\u03f0\5\u008cG\2\u03ef\u03ed\3\2\2\2\u03f0\u03f3\3\2")
 *         buf.write("\2\2\u03f1\u03ef\3\2\2\2\u03f1\u03f2\3\2\2\2\u03f2\u03f4")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u03f3\u03f1\3\2\2\2\u03f4\u03f5\7O\2\2\u03f5")
 *         buf.write("\u008b\3\2\2\2\u03f6\u03f9\7\24\2\2\u03f7\u03f9\5n8\2")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 502, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__761, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 502, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":503
 *         buf.write("\u03f0\5\u008cG\2\u03ef\u03ed\3\2\2\2\u03f0\u03f3\3\2")
 *         buf.write("\2\2\u03f1\u03ef\3\2\2\2\u03f1\u03f2\3\2\2\2\u03f2\u03f4")
 *         buf.write("\3\2\2\2\u03f3\u03f1\3\2\2\2\u03f4\u03f5\7O\2\2\u03f5")             # <<<<<<<<<<<<<<
 *         buf.write("\u008b\3\2\2\2\u03f6\u03f9\7\24\2\2\u03f7\u03f9\5n8\2")
 *         buf.write("\u03f8\u03f6\3\2\2\2\u03f8\u03f7\3\2\2\2\u03f9\u008d\3")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 503, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__762, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 503, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":504
 *         buf.write("\2\2\u03f1\u03ef\3\2\2\2\u03f1\u03f2\3\2\2\2\u03f2\u03f4")
 *         buf.write("\3\2\2\2\u03f3\u03f1\3\2\2\2\u03f4\u03f5\7O\2\2\u03f5")
 *         buf.write("\u008b\3\2\2\2\u03f6\u03f9\7\24\2\2\u03f7\u03f9\5n8\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u03f8\u03f6\3\2\2\2\u03f8\u03f7\3\2\2\2\u03f9\u008d\3")
 *         buf.write("\2\2\2\u03fa\u03fc\5\u0090I\2\u03fb\u03fd\5\u0092J\2\u03fc")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 504, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__763, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 504, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":505
 *         buf.write("\3\2\2\2\u03f3\u03f1\3\2\2\2\u03f4\u03f5\7O\2\2\u03f5")
 *         buf.write("\u008b\3\2\2\2\u03f6\u03f9\7\24\2\2\u03f7\u03f9\5n8\2")
 *         buf.write("\u03f8\u03f6\3\2\2\2\u03f8\u03f7\3\2\2\2\u03f9\u008d\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u03fa\u03fc\5\u0090I\2\u03fb\u03fd\5\u0092J\2\u03fc")
 *         buf.write("\u03fb\3\2\2\2\u03fc\u03fd\3\2\2\2\u03fd\u008f\3\2\2\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 505, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__765, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 505, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":506
 *         buf.write("\u008b\3\2\2\2\u03f6\u03f9\7\24\2\2\u03f7\u03f9\5n8\2")
 *         buf.write("\u03f8\u03f6\3\2\2\2\u03f8\u03f7\3\2\2\2\u03f9\u008d\3")
 *         buf.write("\2\2\2\u03fa\u03fc\5\u0090I\2\u03fb\u03fd\5\u0092J\2\u03fc")             # <<<<<<<<<<<<<<
 *         buf.write("\u03fb\3\2\2\2\u03fc\u03fd\3\2\2\2\u03fd\u008f\3\2\2\2")
 *         buf.write("\u03fe\u0403\7[\2\2\u03ff\u0400\7@\2\2\u0400\u0402\7[")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 506, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__766, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 506, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":507
 *         buf.write("\u03f8\u03f6\3\2\2\2\u03f8\u03f7\3\2\2\2\u03f9\u008d\3")
 *         buf.write("\2\2\2\u03fa\u03fc\5\u0090I\2\u03fb\u03fd\5\u0092J\2\u03fc")
 *         buf.write("\u03fb\3\2\2\2\u03fc\u03fd\3\2\2\2\u03fd\u008f\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u03fe\u0403\7[\2\2\u03ff\u0400\7@\2\2\u0400\u0402\7[")
 *         buf.write("\2\2\u0401\u03ff\3\2\2\2\u0402\u0405\3\2\2\2\u0403\u0401")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 507, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__768, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 507, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":508
 *         buf.write("\2\2\2\u03fa\u03fc\5\u0090I\2\u03fb\u03fd\5\u0092J\2\u03fc")
 *         buf.write("\u03fb\3\2\2\2\u03fc\u03fd\3\2\2\2\u03fd\u008f\3\2\2\2")
 *         buf.write("\u03fe\u0403\7[\2\2\u03ff\u0400\7@\2\2\u0400\u0402\7[")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\u0401\u03ff\3\2\2\2\u0402\u0405\3\2\2\2\u0403\u0401")
 *         buf.write("\3\2\2\2\u0403\u0404\3\2\2\2\u0404\u0407\3\2\2\2\u0405")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 508, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__770, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 508, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":509
 *         buf.write("\u03fb\3\2\2\2\u03fc\u03fd\3\2\2\2\u03fd\u008f\3\2\2\2")
 *         buf.write("\u03fe\u0403\7[\2\2\u03ff\u0400\7@\2\2\u0400\u0402\7[")
 *         buf.write("\2\2\u0401\u03ff\3\2\2\2\u0402\u0405\3\2\2\2\u0403\u0401")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0403\u0404\3\2\2\2\u0404\u0407\3\2\2\2\u0405")
 *         buf.write("\u0403\3\2\2\2\u0406\u03fe\3\2\2\2\u0406\u0407\3\2\2\2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 509, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__772, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 509, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":510
 *         buf.write("\u03fe\u0403\7[\2\2\u03ff\u0400\7@\2\2\u0400\u0402\7[")
 *         buf.write("\2\2\u0401\u03ff\3\2\2\2\u0402\u0405\3\2\2\2\u0403\u0401")
 *         buf.write("\3\2\2\2\u0403\u0404\3\2\2\2\u0404\u0407\3\2\2\2\u0405")             # <<<<<<<<<<<<<<
 *         buf.write("\u0403\3\2\2\2\u0406\u03fe\3\2\2\2\u0406\u0407\3\2\2\2")
 *         buf.write("\u0407\u0091\3\2\2\2\u0408\u0409\7P\2\2\u0409\u040a\5")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 510, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__774, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 510, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":511
 *         buf.write("\2\2\u0401\u03ff\3\2\2\2\u0402\u0405\3\2\2\2\u0403\u0401")
 *         buf.write("\3\2\2\2\u0403\u0404\3\2\2\2\u0404\u0407\3\2\2\2\u0405")
 *         buf.write("\u0403\3\2\2\2\u0406\u03fe\3\2\2\2\u0406\u0407\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0407\u0091\3\2\2\2\u0408\u0409\7P\2\2\u0409\u040a\5")
 *         buf.write(":\36\2\u040a\u0093\3\2\2\2\u0087\u0096\u0099\u009e\u00a2")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 511, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__776, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 511, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":512
 *         buf.write("\3\2\2\2\u0403\u0404\3\2\2\2\u0404\u0407\3\2\2\2\u0405")
 *         buf.write("\u0403\3\2\2\2\u0406\u03fe\3\2\2\2\u0406\u0407\3\2\2\2")
 *         buf.write("\u0407\u0091\3\2\2\2\u0408\u0409\7P\2\2\u0409\u040a\5")             # <<<<<<<<<<<<<<
 *         buf.write(":\36\2\u040a\u0093\3\2\2\2\u0087\u0096\u0099\u009e\u00a2")
 *         buf.write("\u00a8\u00ae\u00b5\u00ba\u00c0\u00c3\u00c7\u00d4\u00d7")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 512, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__777, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 512, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":513
 *         buf.write("\u0403\3\2\2\2\u0406\u03fe\3\2\2\2\u0406\u0407\3\2\2\2")
 *         buf.write("\u0407\u0091\3\2\2\2\u0408\u0409\7P\2\2\u0409\u040a\5")
 *         buf.write(":\36\2\u040a\u0093\3\2\2\2\u0087\u0096\u0099\u009e\u00a2")             # <<<<<<<<<<<<<<
 *         buf.write("\u00a8\u00ae\u00b5\u00ba\u00c0\u00c3\u00c7\u00d4\u00d7")
 *         buf.write("\u00e0\u00e3\u00f2\u00fb\u0102\u010b\u0118\u011a\u011f")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 513, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__779, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 513, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":514
 *         buf.write("\u0407\u0091\3\2\2\2\u0408\u0409\7P\2\2\u0409\u040a\5")
 *         buf.write(":\36\2\u040a\u0093\3\2\2\2\u0087\u0096\u0099\u009e\u00a2")
 *         buf.write("\u00a8\u00ae\u00b5\u00ba\u00c0\u00c3\u00c7\u00d4\u00d7")             # <<<<<<<<<<<<<<
 *         buf.write("\u00e0\u00e3\u00f2\u00fb\u0102\u010b\u0118\u011a\u011f")
 *         buf.write("\u0122\u0125\u0128\u012d\u0134\u0139\u0142\u0149\u014c")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 514, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__781, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 514, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":515
 *         buf.write(":\36\2\u040a\u0093\3\2\2\2\u0087\u0096\u0099\u009e\u00a2")
 *         buf.write("\u00a8\u00ae\u00b5\u00ba\u00c0\u00c3\u00c7\u00d4\u00d7")
 *         buf.write("\u00e0\u00e3\u00f2\u00fb\u0102\u010b\u0118\u011a\u011f")             # <<<<<<<<<<<<<<
 *         buf.write("\u0122\u0125\u0128\u012d\u0134\u0139\u0142\u0149\u014c")
 *         buf.write("\u014f\u0152\u0155\u0159\u015c\u015f\u0162\u0165\u016a")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 515, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__783, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 515, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":516
 *         buf.write("\u00a8\u00ae\u00b5\u00ba\u00c0\u00c3\u00c7\u00d4\u00d7")
 *         buf.write("\u00e0\u00e3\u00f2\u00fb\u0102\u010b\u0118\u011a\u011f")
 *         buf.write("\u0122\u0125\u0128\u012d\u0134\u0139\u0142\u0149\u014c")             # <<<<<<<<<<<<<<
 *         buf.write("\u014f\u0152\u0155\u0159\u015c\u015f\u0162\u0165\u016a")
 *         buf.write("\u016f\u017c\u017e\u0180\u0189\u018f\u0192\u0197\u019c")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 516, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__785, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 516, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":517
 *         buf.write("\u00e0\u00e3\u00f2\u00fb\u0102\u010b\u0118\u011a\u011f")
 *         buf.write("\u0122\u0125\u0128\u012d\u0134\u0139\u0142\u0149\u014c")
 *         buf.write("\u014f\u0152\u0155\u0159\u015c\u015f\u0162\u0165\u016a")             # <<<<<<<<<<<<<<
 *         buf.write("\u016f\u017c\u017e\u0180\u0189\u018f\u0192\u0197\u019c")
 *         buf.write("\u01a1\u01a4\u01a7\u01b0\u01b5\u01be\u01c1\u01c6\u01cc")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 517, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__787, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 517, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":518
 *         buf.write("\u0122\u0125\u0128\u012d\u0134\u0139\u0142\u0149\u014c")
 *         buf.write("\u014f\u0152\u0155\u0159\u015c\u015f\u0162\u0165\u016a")
 *         buf.write("\u016f\u017c\u017e\u0180\u0189\u018f\u0192\u0197\u019c")             # <<<<<<<<<<<<<<
 *         buf.write("\u01a1\u01a4\u01a7\u01b0\u01b5\u01be\u01c1\u01c6\u01cc")
 *         buf.write("\u01d0\u01d9\u01de\u01e1\u01e4\u01e8\u01ec\u01f2\u01f5")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 518, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__789, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 518, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":519
 *         buf.write("\u014f\u0152\u0155\u0159\u015c\u015f\u0162\u0165\u016a")
 *         buf.write("\u016f\u017c\u017e\u0180\u0189\u018f\u0192\u0197\u019c")
 *         buf.write("\u01a1\u01a4\u01a7\u01b0\u01b5\u01be\u01c1\u01c6\u01cc")             # <<<<<<<<<<<<<<
 *         buf.write("\u01d0\u01d9\u01de\u01e1\u01e4\u01e8\u01ec\u01f2\u01f5")
 *         buf.write("\u01f9\u01fc\u0201\u0204\u0213\u0216\u021d\u0220\u0224")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 519, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__791, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 519, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":520
 *         buf.write("\u016f\u017c\u017e\u0180\u0189\u018f\u0192\u0197\u019c")
 *         buf.write("\u01a1\u01a4\u01a7\u01b0\u01b5\u01be\u01c1\u01c6\u01cc")
 *         buf.write("\u01d0\u01d9\u01de\u01e1\u01e4\u01e8\u01ec\u01f2\u01f5")             # <<<<<<<<<<<<<<
 *         buf.write("\u01f9\u01fc\u0201\u0204\u0213\u0216\u021d\u0220\u0224")
 *         buf.write("\u0227\u022f\u0233\u023b\u0249\u0252\u025a\u0268\u0275")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 520, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__793, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 520, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":521
 *         buf.write("\u01a1\u01a4\u01a7\u01b0\u01b5\u01be\u01c1\u01c6\u01cc")
 *         buf.write("\u01d0\u01d9\u01de\u01e1\u01e4\u01e8\u01ec\u01f2\u01f5")
 *         buf.write("\u01f9\u01fc\u0201\u0204\u0213\u0216\u021d\u0220\u0224")             # <<<<<<<<<<<<<<
 *         buf.write("\u0227\u022f\u0233\u023b\u0249\u0252\u025a\u0268\u0275")
 *         buf.write("\u0280\u0285\u028e\u0291\u029e\u02a9\u02ae\u02b7\u02ba")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 521, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__795, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 521, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":522
 *         buf.write("\u01d0\u01d9\u01de\u01e1\u01e4\u01e8\u01ec\u01f2\u01f5")
 *         buf.write("\u01f9\u01fc\u0201\u0204\u0213\u0216\u021d\u0220\u0224")
 *         buf.write("\u0227\u022f\u0233\u023b\u0249\u0252\u025a\u0268\u0275")             # <<<<<<<<<<<<<<
 *         buf.write("\u0280\u0285\u028e\u0291\u029e\u02a9\u02ae\u02b7\u02ba")
 *         buf.write("\u02c7\u02d5\u02e0\u02e6\u02f0\u02fe\u0309\u030e\u031c")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 522, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__797, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 522, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":523
 *         buf.write("\u01f9\u01fc\u0201\u0204\u0213\u0216\u021d\u0220\u0224")
 *         buf.write("\u0227\u022f\u0233\u023b\u0249\u0252\u025a\u0268\u0275")
 *         buf.write("\u0280\u0285\u028e\u0291\u029e\u02a9\u02ae\u02b7\u02ba")             # <<<<<<<<<<<<<<
 *         buf.write("\u02c7\u02d5\u02e0\u02e6\u02f0\u02fe\u0309\u030e\u031c")
 *         buf.write("\u0327\u032c\u0345\u034b\u0352\u0354\u0360\u0371\u0373")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 523, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__799, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 523, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":524
 *         buf.write("\u0227\u022f\u0233\u023b\u0249\u0252\u025a\u0268\u0275")
 *         buf.write("\u0280\u0285\u028e\u0291\u029e\u02a9\u02ae\u02b7\u02ba")
 *         buf.write("\u02c7\u02d5\u02e0\u02e6\u02f0\u02fe\u0309\u030e\u031c")             # <<<<<<<<<<<<<<
 *         buf.write("\u0327\u032c\u0345\u034b\u0352\u0354\u0360\u0371\u0373")
 *         buf.write("\u038c\u0396\u0399\u03a0\u03a5\u03ac\u03b1\u03ba\u03bc")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 524, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__801, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 524, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":525
 *         buf.write("\u0280\u0285\u028e\u0291\u029e\u02a9\u02ae\u02b7\u02ba")
 *         buf.write("\u02c7\u02d5\u02e0\u02e6\u02f0\u02fe\u0309\u030e\u031c")
 *         buf.write("\u0327\u032c\u0345\u034b\u0352\u0354\u0360\u0371\u0373")             # <<<<<<<<<<<<<<
 *         buf.write("\u038c\u0396\u0399\u03a0\u03a5\u03ac\u03b1\u03ba\u03bc")
 *         buf.write("\u03c0\u03c7\u03d2\u03d7\u03da\u03e0\u03e8\u03f1\u03f8")
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 525, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__803, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 525, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":526
 *         buf.write("\u02c7\u02d5\u02e0\u02e6\u02f0\u02fe\u0309\u030e\u031c")
 *         buf.write("\u0327\u032c\u0345\u034b\u0352\u0354\u0360\u0371\u0373")
 *         buf.write("\u038c\u0396\u0399\u03a0\u03a5\u03ac\u03b1\u03ba\u03bc")             # <<<<<<<<<<<<<<
 *         buf.write("\u03c0\u03c7\u03d2\u03d7\u03da\u03e0\u03e8\u03f1\u03f8")
 *         buf.write("\u03fc\u0403\u0406")
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 526, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__805, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 526, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":527
 *         buf.write("\u0327\u032c\u0345\u034b\u0352\u0354\u0360\u0371\u0373")
 *         buf.write("\u038c\u0396\u0399\u03a0\u03a5\u03ac\u03b1\u03ba\u03bc")
 *         buf.write("\u03c0\u03c7\u03d2\u03d7\u03da\u03e0\u03e8\u03f1\u03f8")             # <<<<<<<<<<<<<<
 *         buf.write("\u03fc\u0403\u0406")
 *         return buf.getvalue()
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 527, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__807, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 527, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":528
 *         buf.write("\u038c\u0396\u0399\u03a0\u03a5\u03ac\u03b1\u03ba\u03bc")
 *         buf.write("\u03c0\u03c7\u03d2\u03d7\u03da\u03e0\u03e8\u03f1\u03f8")
 *         buf.write("\u03fc\u0403\u0406")             # <<<<<<<<<<<<<<
 *         return buf.getvalue()
 * 
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 528, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__809, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 528, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":529
 *         buf.write("\u03c0\u03c7\u03d2\u03d7\u03da\u03e0\u03e8\u03f1\u03f8")
 *         buf.write("\u03fc\u0403\u0406")
 *         return buf.getvalue()             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __Pyx_XDECREF(__pyx_r);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf, __pyx_n_s_getvalue); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 529, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 529, __pyx_L7_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 529, __pyx_L7_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_r = __pyx_t_1;
          __pyx_t_1 = 0;
          goto __pyx_L11_try_return;

          /* "pymola/generated/ModelicaParser.pyx":9
 * 
 * def serializedATN():
 *     with StringIO() as buf:             # <<<<<<<<<<<<<<
 *         buf.write("\3\u608b\ua72a\u8133\ub9ed\u417c\u3be7\u7786\u5964\3^")
 *         buf.write("\u040c\4\2\t\2\4\3\t\3\4\4\t\4\4\5\t\5\4\6\t\6\4\7\t\7")
 */
        }
        __pyx_L7_error:;
        __Pyx_PyThreadState_assign
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        /*except:*/ {
          __Pyx_AddTraceback("pymola.generated.ModelicaParser.serializedATN", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_3, &__pyx_t_2) < 0) __PYX_ERR(0, 9, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = PyTuple_Pack(3, __pyx_t_1, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 9, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 9, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_9);
          __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (__pyx_t_10 < 0) __PYX_ERR(0, 9, __pyx_L9_except_error)
          __pyx_t_11 = ((!(__pyx_t_10 != 0)) != 0);
          if (__pyx_t_11) {
            __Pyx_GIVEREF(__pyx_t_1);
            __Pyx_GIVEREF(__pyx_t_3);
            __Pyx_XGIVEREF(__pyx_t_2);
            __Pyx_ErrRestoreWithState(__pyx_t_1, __pyx_t_3, __pyx_t_2);
            __pyx_t_1 = 0; __pyx_t_3 = 0; __pyx_t_2 = 0; 
            __PYX_ERR(0, 9, __pyx_L9_except_error)
          }
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          goto __pyx_L8_exception_handled;
        }
        __pyx_L9_except_error:;
        __Pyx_PyThreadState_assign
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
        goto __pyx_L1_error;
        __pyx_L11_try_return:;
        __Pyx_PyThreadState_assign
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
        goto __pyx_L4_return;
        __pyx_L8_exception_handled:;
        __Pyx_PyThreadState_assign
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
      }
    }
    /*finally:*/ {
      /*normal exit:*/{
        if (__pyx_t_4) {
          __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__810, NULL);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 9, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        goto __pyx_L6;
      }
      __pyx_L4_return: {
        __pyx_t_8 = __pyx_r;
        __pyx_r = 0;
        if (__pyx_t_4) {
          __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__811, NULL);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 9, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __pyx_r = __pyx_t_8;
        __pyx_t_8 = 0;
        goto __pyx_L0;
      }
      __pyx_L6:;
    }
    goto __pyx_L18;
    __pyx_L3_error:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L1_error;
    __pyx_L18:;
  }

  /* "pymola/generated/ModelicaParser.pyx":8
 * import sys
 * 
 * def serializedATN():             # <<<<<<<<<<<<<<
 *     with StringIO() as buf:
 *         buf.write("\3\u608b\ua72a\u8133\ub9ed\u417c\u3be7\u7786\u5964\3^")
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.serializedATN", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_buf);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":775
 *     WS=92
 * 
 *     def __init__(self, input:TokenStream, output:TextIO = sys.stdout):             # <<<<<<<<<<<<<<
 *         super().__init__(input, output)
 *         self.checkVersion("4.7")
 */

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_2__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__defaults__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 775, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_output);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_output);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_output);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 775, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_2, 1, Py_None);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_input = 0;
  PyObject *__pyx_v_output = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_input,&__pyx_n_s_output,0};
    PyObject* values[3] = {0,0,0};
    __pyx_defaults *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self);
    values[2] = __pyx_dynamic_args->__pyx_arg_output;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_input)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, 1); __PYX_ERR(0, 775, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_output);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 775, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_input = values[1];
    __pyx_v_output = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 775, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser___init__(__pyx_self, __pyx_v_self, __pyx_v_input, __pyx_v_output);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_input, PyObject *__pyx_v_output) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":776
 * 
 *     def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
 *         super().__init__(input, output)             # <<<<<<<<<<<<<<
 *         self.checkVersion("4.7")
 *         self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 776, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 776, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 776, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 776, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_input, __pyx_v_output};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 776, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_input, __pyx_v_output};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 776, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 776, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_input);
    __Pyx_GIVEREF(__pyx_v_input);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_input);
    __Pyx_INCREF(__pyx_v_output);
    __Pyx_GIVEREF(__pyx_v_output);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_output);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 776, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":777
 *     def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
 *         super().__init__(input, output)
 *         self.checkVersion("4.7")             # <<<<<<<<<<<<<<
 *         self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
 *         self._predicates = None
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_checkVersion); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 777, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__812, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 777, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pymola/generated/ModelicaParser.pyx":778
 *         super().__init__(input, output)
 *         self.checkVersion("4.7")
 *         self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)             # <<<<<<<<<<<<<<
 *         self._predicates = None
 * 
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserATNSimulator); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 778, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_atn); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 778, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_decisionsToDFA); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 778, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_sharedContextCache); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 778, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[5] = {__pyx_t_7, __pyx_v_self, __pyx_t_5, __pyx_t_2, __pyx_t_6};
    __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_4, 4+__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 778, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[5] = {__pyx_t_7, __pyx_v_self, __pyx_t_5, __pyx_t_2, __pyx_t_6};
    __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_4, 4+__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 778, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else
  #endif
  {
    __pyx_t_8 = PyTuple_New(4+__pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 778, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (__pyx_t_7) {
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_7); __pyx_t_7 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_4, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_4, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_4, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_8, 3+__pyx_t_4, __pyx_t_6);
    __pyx_t_5 = 0;
    __pyx_t_2 = 0;
    __pyx_t_6 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_8, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 778, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_interp, __pyx_t_3) < 0) __PYX_ERR(0, 778, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pymola/generated/ModelicaParser.pyx":779
 *         self.checkVersion("4.7")
 *         self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
 *         self._predicates = None             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_predicates, Py_None) < 0) __PYX_ERR(0, 779, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":775
 *     WS=92
 * 
 *     def __init__(self, input:TokenStream, output:TextIO = sys.stdout):             # <<<<<<<<<<<<<<
 *         super().__init__(input, output)
 *         self.checkVersion("4.7")
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":785
 *     class Stored_definitionContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 785, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 785, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 785, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Stored_definitionContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Stored_definitionContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Stored_definitionContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":786
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 786, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 786, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 786, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 786, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 786, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":787
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def WITHIN(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 787, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":785
 *     class Stored_definitionContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Stored_definitionContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":789
 *             self.parser = parser
 * 
 *         def WITHIN(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.WITHIN, 0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_3WITHIN(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_3WITHIN = {"WITHIN", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_3WITHIN, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_3WITHIN(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("WITHIN (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_2WITHIN(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_2WITHIN(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("WITHIN", 0);

  /* "pymola/generated/ModelicaParser.pyx":790
 * 
 *         def WITHIN(self):
 *             return self.getToken(ModelicaParser.WITHIN, 0)             # <<<<<<<<<<<<<<
 * 
 *         def stored_definition_class(self, i:int=None):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getToken); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_WITHIN); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 790, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 790, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 790, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 790, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":789
 *             self.parser = parser
 * 
 *         def WITHIN(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.WITHIN, 0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Stored_definitionContext.WITHIN", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":792
 *             return self.getToken(ModelicaParser.WITHIN, 0)
 * 
 *         def stored_definition_class(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Stored_definition_classContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_5stored_definition_class(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_5stored_definition_class = {"stored_definition_class", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_5stored_definition_class, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_5stored_definition_class(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("stored_definition_class (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "stored_definition_class") < 0)) __PYX_ERR(0, 792, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("stored_definition_class", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 792, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Stored_definitionContext.stored_definition_class", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_4stored_definition_class(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_4stored_definition_class(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("stored_definition_class", 0);

  /* "pymola/generated/ModelicaParser.pyx":793
 * 
 *         def stored_definition_class(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.Stored_definition_classContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":794
 *         def stored_definition_class(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Stored_definition_classContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.Stored_definition_classContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 794, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 794, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_Stored_definition_classContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 794, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 794, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 794, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 794, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 794, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 794, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":793
 * 
 *         def stored_definition_class(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.Stored_definition_classContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":796
 *                 return self.getTypedRuleContexts(ModelicaParser.Stored_definition_classContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.Stored_definition_classContext,i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 796, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 796, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Stored_definition_classContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 796, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 796, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 796, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 796, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 796, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":792
 *             return self.getToken(ModelicaParser.WITHIN, 0)
 * 
 *         def stored_definition_class(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Stored_definition_classContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Stored_definitionContext.stored_definition_class", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":799
 * 
 * 
 *         def component_reference(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_7component_reference(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_7component_reference = {"component_reference", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_7component_reference, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_7component_reference(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("component_reference (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_6component_reference(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_6component_reference(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("component_reference", 0);

  /* "pymola/generated/ModelicaParser.pyx":800
 * 
 *         def component_reference(self):
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 800, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 800, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Component_referenceContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 800, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 800, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 800, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 800, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 800, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":799
 * 
 * 
 *         def component_reference(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Stored_definitionContext.component_reference", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":803
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_stored_definition
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_9getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_9getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_9getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_9getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_8getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_8getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":804
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_stored_definition             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 804, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_stored_definition); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 804, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":803
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_stored_definition
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Stored_definitionContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":806
 *             return ModelicaParser.RULE_stored_definition
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStored_definition" ):
 *                 listener.enterStored_definition(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_11enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_11enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_11enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_11enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 806, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 806, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 806, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Stored_definitionContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_10enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_10enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":807
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterStored_definition" ):             # <<<<<<<<<<<<<<
 *                 listener.enterStored_definition(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterStored_definition); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 807, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":808
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterStored_definition" ):
 *                 listener.enterStored_definition(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterStored_definition); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 808, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 808, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 808, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 808, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 808, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 808, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":807
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterStored_definition" ):             # <<<<<<<<<<<<<<
 *                 listener.enterStored_definition(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":806
 *             return ModelicaParser.RULE_stored_definition
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStored_definition" ):
 *                 listener.enterStored_definition(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Stored_definitionContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":810
 *                 listener.enterStored_definition(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStored_definition" ):
 *                 listener.exitStored_definition(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_13exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_13exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_13exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_13exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 810, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 810, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 810, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Stored_definitionContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_12exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_12exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":811
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitStored_definition" ):             # <<<<<<<<<<<<<<
 *                 listener.exitStored_definition(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitStored_definition); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 811, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":812
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitStored_definition" ):
 *                 listener.exitStored_definition(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitStored_definition); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 812, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 812, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 812, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 812, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 812, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 812, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":811
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitStored_definition" ):             # <<<<<<<<<<<<<<
 *                 listener.exitStored_definition(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":810
 *                 listener.enterStored_definition(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStored_definition" ):
 *                 listener.exitStored_definition(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Stored_definitionContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":814
 *                 listener.exitStored_definition(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStored_definition" ):
 *                 return visitor.visitStored_definition(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_15accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_15accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_15accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_15accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 814, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 814, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 814, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Stored_definitionContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_14accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_14accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":815
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitStored_definition" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitStored_definition(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitStored_definition); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 815, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":816
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitStored_definition" ):
 *                 return visitor.visitStored_definition(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitStored_definition); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 816, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 816, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 816, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 816, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 816, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 816, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":815
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitStored_definition" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitStored_definition(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":818
 *                 return visitor.visitStored_definition(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 818, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 818, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 818, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 818, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 818, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 818, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":814
 *                 listener.exitStored_definition(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStored_definition" ):
 *                 return visitor.visitStored_definition(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Stored_definitionContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":823
 * 
 * 
 *     def stored_definition(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Stored_definitionContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_3stored_definition(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_3stored_definition = {"stored_definition", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_3stored_definition, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_3stored_definition(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("stored_definition (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_2stored_definition(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_2stored_definition(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  char const *__pyx_t_22;
  __Pyx_RefNannySetupContext("stored_definition", 0);

  /* "pymola/generated/ModelicaParser.pyx":825
 *     def stored_definition(self):
 * 
 *         localctx = ModelicaParser.Stored_definitionContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 0, self.RULE_stored_definition)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 825, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Stored_definitionContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 825, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 825, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 825, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 825, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 825, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 825, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 825, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":826
 * 
 *         localctx = ModelicaParser.Stored_definitionContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 0, self.RULE_stored_definition)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_stored_definition); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_0, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 826, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_0, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 826, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 826, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_0);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 826, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":827
 *         localctx = ModelicaParser.Stored_definitionContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 0, self.RULE_stored_definition)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 827, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":828
 *         self.enterRule(localctx, 0, self.RULE_stored_definition)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 151
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":829
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 151
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 829, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 829, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 829, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 829, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 829, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":830
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 151             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_151) < 0) __PYX_ERR(0, 830, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":831
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 151
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.WITHIN:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 831, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 831, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 831, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 831, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 831, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 831, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 831, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":832
 *             self.state = 151
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.WITHIN:
 *                 self.state = 146
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 832, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 832, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__813, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 832, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v__la = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":833
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.WITHIN:             # <<<<<<<<<<<<<<
 *                 self.state = 146
 *                 self.match(ModelicaParser.WITHIN)
 */
        __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 833, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_WITHIN); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 833, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 833, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 833, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":834
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.WITHIN:
 *                 self.state = 146             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.WITHIN)
 *                 self.state = 148
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_146) < 0) __PYX_ERR(0, 834, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":835
 *             if _la==ModelicaParser.WITHIN:
 *                 self.state = 146
 *                 self.match(ModelicaParser.WITHIN)             # <<<<<<<<<<<<<<
 *                 self.state = 148
 *                 self._errHandler.sync(self)
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 835, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 835, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_WITHIN); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 835, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 835, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 835, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 835, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 835, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 835, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":836
 *                 self.state = 146
 *                 self.match(ModelicaParser.WITHIN)
 *                 self.state = 148             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_148) < 0) __PYX_ERR(0, 836, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":837
 *                 self.match(ModelicaParser.WITHIN)
 *                 self.state = 148
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.IDENT:
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 837, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_sync); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 837, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 837, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 837, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 837, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 837, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 837, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":838
 *                 self.state = 148
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if _la==ModelicaParser.IDENT:
 *                     self.state = 147
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 838, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 838, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__814, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 838, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":839
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.IDENT:             # <<<<<<<<<<<<<<
 *                     self.state = 147
 *                     self.component_reference()
 */
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 839, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 839, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 839, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 839, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (__pyx_t_11) {

            /* "pymola/generated/ModelicaParser.pyx":840
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.IDENT:
 *                     self.state = 147             # <<<<<<<<<<<<<<
 *                     self.component_reference()
 * 
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_147) < 0) __PYX_ERR(0, 840, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":841
 *                 if _la==ModelicaParser.IDENT:
 *                     self.state = 147
 *                     self.component_reference()             # <<<<<<<<<<<<<<
 * 
 * 
 */
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_component_reference); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 841, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_3 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
              __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
              if (likely(__pyx_t_3)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                __Pyx_INCREF(__pyx_t_3);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_4, function);
              }
            }
            if (__pyx_t_3) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 841, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else {
              __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 841, __pyx_L6_error)
            }
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":839
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.IDENT:             # <<<<<<<<<<<<<<
 *                     self.state = 147
 *                     self.component_reference()
 */
          }

          /* "pymola/generated/ModelicaParser.pyx":844
 * 
 * 
 *                 self.state = 150             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__0)
 * 
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_150) < 0) __PYX_ERR(0, 844, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":845
 * 
 *                 self.state = 150
 *                 self.match(ModelicaParser.T__0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 845, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 845, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 845, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (!__pyx_t_3) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 845, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 845, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 845, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            {
              __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 845, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
              __Pyx_GIVEREF(__pyx_t_7);
              PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_7);
              __pyx_t_7 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 845, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":833
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.WITHIN:             # <<<<<<<<<<<<<<
 *                 self.state = 146
 *                 self.match(ModelicaParser.WITHIN)
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":848
 * 
 * 
 *             self.state = 156             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_156) < 0) __PYX_ERR(0, 848, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":849
 * 
 *             self.state = 156
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__1) | (1 << ModelicaParser.T__2) | (1 << ModelicaParser.T__3) | (1 << ModelicaParser.T__4) | (1 << ModelicaParser.T__5) | (1 << ModelicaParser.T__6) | (1 << ModelicaParser.T__7) | (1 << ModelicaParser.T__8) | (1 << ModelicaParser.T__9) | (1 << ModelicaParser.T__10) | (1 << ModelicaParser.T__11) | (1 << ModelicaParser.T__12))) != 0) or ((((_la - 80)) & ~0x3f) == 0 and ((1 << (_la - 80)) & ((1 << (ModelicaParser.PARTIAL - 80)) | (1 << (ModelicaParser.FINAL - 80)) | (1 << (ModelicaParser.ENCAPSULATED - 80)))) != 0):
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 849, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 849, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        if (!__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 849, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 849, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 849, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 849, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4); __pyx_t_4 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 849, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":850
 *             self.state = 156
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__1) | (1 << ModelicaParser.T__2) | (1 << ModelicaParser.T__3) | (1 << ModelicaParser.T__4) | (1 << ModelicaParser.T__5) | (1 << ModelicaParser.T__6) | (1 << ModelicaParser.T__7) | (1 << ModelicaParser.T__8) | (1 << ModelicaParser.T__9) | (1 << ModelicaParser.T__10) | (1 << ModelicaParser.T__11) | (1 << ModelicaParser.T__12))) != 0) or ((((_la - 80)) & ~0x3f) == 0 and ((1 << (_la - 80)) & ((1 << (ModelicaParser.PARTIAL - 80)) | (1 << (ModelicaParser.FINAL - 80)) | (1 << (ModelicaParser.ENCAPSULATED - 80)))) != 0):
 *                 self.state = 153
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 850, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 850, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__815, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 850, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":851
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__1) | (1 << ModelicaParser.T__2) | (1 << ModelicaParser.T__3) | (1 << ModelicaParser.T__4) | (1 << ModelicaParser.T__5) | (1 << ModelicaParser.T__6) | (1 << ModelicaParser.T__7) | (1 << ModelicaParser.T__8) | (1 << ModelicaParser.T__9) | (1 << ModelicaParser.T__10) | (1 << ModelicaParser.T__11) | (1 << ModelicaParser.T__12))) != 0) or ((((_la - 80)) & ~0x3f) == 0 and ((1 << (_la - 80)) & ((1 << (ModelicaParser.PARTIAL - 80)) | (1 << (ModelicaParser.FINAL - 80)) | (1 << (ModelicaParser.ENCAPSULATED - 80)))) != 0):             # <<<<<<<<<<<<<<
 *                 self.state = 153
 *                 self.stored_definition_class()
 */
        while (1) {
          __pyx_t_1 = __Pyx_PyInt_From_long((~0x3f)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_2 = PyNumber_And(__pyx_v__la, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_t_2, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (!__pyx_t_12) {
            goto __pyx_L19_next_or;
          } else {
          }
          __pyx_t_1 = PyNumber_Lshift(__pyx_int_1, __pyx_v__la); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Lshift(__pyx_int_1, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = PyNumber_Or(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Or(__pyx_t_4, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Lshift(__pyx_int_1, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = PyNumber_Or(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Or(__pyx_t_4, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Lshift(__pyx_int_1, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = PyNumber_Or(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Or(__pyx_t_4, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Lshift(__pyx_int_1, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = PyNumber_Or(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Or(__pyx_t_4, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__10); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Lshift(__pyx_int_1, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = PyNumber_Or(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__11); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Or(__pyx_t_4, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__12); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Lshift(__pyx_int_1, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = PyNumber_Or(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_And(__pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = PyObject_RichCompare(__pyx_t_7, __pyx_int_0, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (!__pyx_t_12) {
          } else {
            __pyx_t_11 = __pyx_t_12;
            goto __pyx_L18_bool_binop_done;
          }
          __pyx_L19_next_or:;
          __pyx_t_4 = __Pyx_PyInt_SubtractObjC(__pyx_v__la, __pyx_int_80, 80, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = __Pyx_PyInt_From_long((~0x3f)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_1 = PyNumber_And(__pyx_t_4, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_PyInt_EqObjC(__pyx_t_1, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (__pyx_t_12) {
          } else {
            __pyx_t_11 = __pyx_t_12;
            goto __pyx_L18_bool_binop_done;
          }
          __pyx_t_7 = __Pyx_PyInt_SubtractObjC(__pyx_v__la, __pyx_int_80, 80, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_1 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_PARTIAL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_PyInt_SubtractObjC(__pyx_t_4, __pyx_int_80, 80, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_FINAL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_PyInt_SubtractObjC(__pyx_t_2, __pyx_int_80, 80, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Or(__pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_ENCAPSULATED); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_t_4, __pyx_int_80, 80, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Or(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = PyNumber_And(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyObject_RichCompare(__pyx_t_4, __pyx_int_0, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 851, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_11 = __pyx_t_12;
          __pyx_L18_bool_binop_done:;
          if (!__pyx_t_11) break;

          /* "pymola/generated/ModelicaParser.pyx":852
 *             _la = self._input.LA(1)
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__1) | (1 << ModelicaParser.T__2) | (1 << ModelicaParser.T__3) | (1 << ModelicaParser.T__4) | (1 << ModelicaParser.T__5) | (1 << ModelicaParser.T__6) | (1 << ModelicaParser.T__7) | (1 << ModelicaParser.T__8) | (1 << ModelicaParser.T__9) | (1 << ModelicaParser.T__10) | (1 << ModelicaParser.T__11) | (1 << ModelicaParser.T__12))) != 0) or ((((_la - 80)) & ~0x3f) == 0 and ((1 << (_la - 80)) & ((1 << (ModelicaParser.PARTIAL - 80)) | (1 << (ModelicaParser.FINAL - 80)) | (1 << (ModelicaParser.ENCAPSULATED - 80)))) != 0):
 *                 self.state = 153             # <<<<<<<<<<<<<<
 *                 self.stored_definition_class()
 *                 self.state = 158
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_153) < 0) __PYX_ERR(0, 852, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":853
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__1) | (1 << ModelicaParser.T__2) | (1 << ModelicaParser.T__3) | (1 << ModelicaParser.T__4) | (1 << ModelicaParser.T__5) | (1 << ModelicaParser.T__6) | (1 << ModelicaParser.T__7) | (1 << ModelicaParser.T__8) | (1 << ModelicaParser.T__9) | (1 << ModelicaParser.T__10) | (1 << ModelicaParser.T__11) | (1 << ModelicaParser.T__12))) != 0) or ((((_la - 80)) & ~0x3f) == 0 and ((1 << (_la - 80)) & ((1 << (ModelicaParser.PARTIAL - 80)) | (1 << (ModelicaParser.FINAL - 80)) | (1 << (ModelicaParser.ENCAPSULATED - 80)))) != 0):
 *                 self.state = 153
 *                 self.stored_definition_class()             # <<<<<<<<<<<<<<
 *                 self.state = 158
 *                 self._errHandler.sync(self)
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_stored_definition_class); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 853, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_1 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (__pyx_t_1) {
            __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 853, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          } else {
            __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 853, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":854
 *                 self.state = 153
 *                 self.stored_definition_class()
 *                 self.state = 158             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_158) < 0) __PYX_ERR(0, 854, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":855
 *                 self.stored_definition_class()
 *                 self.state = 158
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 * 
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 855, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_sync); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 855, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_1, function);
            }
          }
          if (!__pyx_t_4) {
            __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 855, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_1)) {
              PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
              __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 855, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_2);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
              PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
              __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 855, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_2);
            } else
            #endif
            {
              __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 855, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4); __pyx_t_4 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self);
              __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 855, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":856
 *                 self.state = 158
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *         except RecognitionException as re:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 856, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_LA); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 856, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__816, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 856, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_2);
          __pyx_t_2 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":828
 *         self.enterRule(localctx, 0, self.RULE_stored_definition)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 151
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "pymola/generated/ModelicaParser.pyx":858
 *                 _la = self._input.LA(1)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 858, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.stored_definition", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_1, &__pyx_t_7) < 0) __PYX_ERR(0, 858, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_v_re = __pyx_t_1;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":859
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 859, __pyx_L27_error)

          /* "pymola/generated/ModelicaParser.pyx":860
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 860, __pyx_L27_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 860, __pyx_L27_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_self, __pyx_v_re};
            __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 860, __pyx_L27_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_self, __pyx_v_re};
            __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 860, __pyx_L27_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 860, __pyx_L27_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_13, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 860, __pyx_L27_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "pymola/generated/ModelicaParser.pyx":861
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 861, __pyx_L27_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 861, __pyx_L27_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 861, __pyx_L27_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 861, __pyx_L27_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 861, __pyx_L27_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 861, __pyx_L27_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":858
 *                 _la = self._input.LA(1)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L28;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L27_error:;
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_6 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L8_except_error;
          }
          __pyx_L28:;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":828
 *         self.enterRule(localctx, 0, self.RULE_stored_definition)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 151
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":863
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 863, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      if (__pyx_t_2) {
        __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 863, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      } else {
        __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 863, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_14 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 863, __pyx_L34_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_1, function);
          }
        }
        if (__pyx_t_2) {
          __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 863, __pyx_L34_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        } else {
          __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 863, __pyx_L34_error)
        }
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
      __pyx_L34_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":864
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Stored_definition_classContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":823
 * 
 * 
 *     def stored_definition(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Stored_definitionContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.stored_definition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":868
 *     class Stored_definition_classContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 868, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 868, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 868, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Stored_definition_classContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":869
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 869, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 869, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 869, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 869, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 869, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 869, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 869, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 869, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":870
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def class_definition(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 870, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":868
 *     class Stored_definition_classContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Stored_definition_classContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":872
 *             self.parser = parser
 * 
 *         def class_definition(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_definitionContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_3class_definition(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_3class_definition = {"class_definition", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_3class_definition, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_3class_definition(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("class_definition (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_2class_definition(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_2class_definition(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("class_definition", 0);

  /* "pymola/generated/ModelicaParser.pyx":873
 * 
 *         def class_definition(self):
 *             return self.getTypedRuleContext(ModelicaParser.Class_definitionContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 873, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 873, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Class_definitionContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 873, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 873, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 873, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 873, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 873, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":872
 *             self.parser = parser
 * 
 *         def class_definition(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_definitionContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Stored_definition_classContext.class_definition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":876
 * 
 * 
 *         def FINAL(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.FINAL, 0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_5FINAL(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_5FINAL = {"FINAL", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_5FINAL, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_5FINAL(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("FINAL (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_4FINAL(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_4FINAL(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("FINAL", 0);

  /* "pymola/generated/ModelicaParser.pyx":877
 * 
 *         def FINAL(self):
 *             return self.getToken(ModelicaParser.FINAL, 0)             # <<<<<<<<<<<<<<
 * 
 *         def getRuleIndex(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getToken); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 877, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 877, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_FINAL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 877, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 877, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 877, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 877, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 877, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":876
 * 
 * 
 *         def FINAL(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.FINAL, 0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Stored_definition_classContext.FINAL", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":879
 *             return self.getToken(ModelicaParser.FINAL, 0)
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_stored_definition_class
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_7getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_7getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_6getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":880
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_stored_definition_class             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_stored_definition_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":879
 *             return self.getToken(ModelicaParser.FINAL, 0)
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_stored_definition_class
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Stored_definition_classContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":882
 *             return ModelicaParser.RULE_stored_definition_class
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStored_definition_class" ):
 *                 listener.enterStored_definition_class(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_9enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_9enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 882, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 882, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 882, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Stored_definition_classContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_8enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":883
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterStored_definition_class" ):             # <<<<<<<<<<<<<<
 *                 listener.enterStored_definition_class(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterStored_definition_class); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 883, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":884
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterStored_definition_class" ):
 *                 listener.enterStored_definition_class(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterStored_definition_class); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 884, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 884, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 884, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 884, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 884, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 884, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":883
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterStored_definition_class" ):             # <<<<<<<<<<<<<<
 *                 listener.enterStored_definition_class(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":882
 *             return ModelicaParser.RULE_stored_definition_class
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStored_definition_class" ):
 *                 listener.enterStored_definition_class(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Stored_definition_classContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":886
 *                 listener.enterStored_definition_class(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStored_definition_class" ):
 *                 listener.exitStored_definition_class(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_11exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_11exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 886, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 886, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 886, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Stored_definition_classContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_10exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":887
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitStored_definition_class" ):             # <<<<<<<<<<<<<<
 *                 listener.exitStored_definition_class(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitStored_definition_class); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 887, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":888
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitStored_definition_class" ):
 *                 listener.exitStored_definition_class(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitStored_definition_class); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 888, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 888, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 888, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 888, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 888, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 888, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":887
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitStored_definition_class" ):             # <<<<<<<<<<<<<<
 *                 listener.exitStored_definition_class(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":886
 *                 listener.enterStored_definition_class(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStored_definition_class" ):
 *                 listener.exitStored_definition_class(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Stored_definition_classContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":890
 *                 listener.exitStored_definition_class(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStored_definition_class" ):
 *                 return visitor.visitStored_definition_class(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_13accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_13accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 890, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 890, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 890, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Stored_definition_classContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_12accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":891
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitStored_definition_class" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitStored_definition_class(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitStored_definition_class); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 891, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":892
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitStored_definition_class" ):
 *                 return visitor.visitStored_definition_class(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitStored_definition_class); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 892, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 892, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 892, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 892, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 892, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 892, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":891
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitStored_definition_class" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitStored_definition_class(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":894
 *                 return visitor.visitStored_definition_class(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 894, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 894, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 894, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 894, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 894, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 894, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":890
 *                 listener.exitStored_definition_class(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStored_definition_class" ):
 *                 return visitor.visitStored_definition_class(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Stored_definition_classContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":899
 * 
 * 
 *     def stored_definition_class(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Stored_definition_classContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_5stored_definition_class(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_5stored_definition_class = {"stored_definition_class", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_5stored_definition_class, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_5stored_definition_class(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("stored_definition_class (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_4stored_definition_class(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_4stored_definition_class(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  char const *__pyx_t_21;
  __Pyx_RefNannySetupContext("stored_definition_class", 0);

  /* "pymola/generated/ModelicaParser.pyx":901
 *     def stored_definition_class(self):
 * 
 *         localctx = ModelicaParser.Stored_definition_classContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 2, self.RULE_stored_definition_class)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 901, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Stored_definition_classContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 901, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 901, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 901, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 901, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 901, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 901, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 901, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":902
 * 
 *         localctx = ModelicaParser.Stored_definition_classContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 2, self.RULE_stored_definition_class)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 902, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_stored_definition_class); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 902, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_2, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 902, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_2, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 902, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 902, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_2);
    __Pyx_GIVEREF(__pyx_int_2);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_2);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 902, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":903
 *         localctx = ModelicaParser.Stored_definition_classContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 2, self.RULE_stored_definition_class)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 903, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":904
 *         self.enterRule(localctx, 2, self.RULE_stored_definition_class)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 160
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":905
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 160
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 905, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 905, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 905, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 905, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 905, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":906
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 160             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_160) < 0) __PYX_ERR(0, 906, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":907
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 160
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.FINAL:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 907, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 907, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 907, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 907, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 907, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 907, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 907, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":908
 *             self.state = 160
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.FINAL:
 *                 self.state = 159
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 908, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 908, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__817, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 908, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v__la = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":909
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.FINAL:             # <<<<<<<<<<<<<<
 *                 self.state = 159
 *                 self.match(ModelicaParser.FINAL)
 */
        __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 909, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_FINAL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 909, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 909, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 909, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":910
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.FINAL:
 *                 self.state = 159             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.FINAL)
 * 
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_159) < 0) __PYX_ERR(0, 910, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":911
 *             if _la==ModelicaParser.FINAL:
 *                 self.state = 159
 *                 self.match(ModelicaParser.FINAL)             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 911, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 911, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_FINAL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 911, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 911, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 911, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 911, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 911, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 911, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":909
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.FINAL:             # <<<<<<<<<<<<<<
 *                 self.state = 159
 *                 self.match(ModelicaParser.FINAL)
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":914
 * 
 * 
 *             self.state = 162             # <<<<<<<<<<<<<<
 *             self.class_definition()
 *             self.state = 163
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_162) < 0) __PYX_ERR(0, 914, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":915
 * 
 *             self.state = 162
 *             self.class_definition()             # <<<<<<<<<<<<<<
 *             self.state = 163
 *             self.match(ModelicaParser.T__0)
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_class_definition); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 915, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 915, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 915, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":916
 *             self.state = 162
 *             self.class_definition()
 *             self.state = 163             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__0)
 *         except RecognitionException as re:
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_163) < 0) __PYX_ERR(0, 916, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":917
 *             self.class_definition()
 *             self.state = 163
 *             self.match(ModelicaParser.T__0)             # <<<<<<<<<<<<<<
 *         except RecognitionException as re:
 *             localctx.exception = re
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 917, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 917, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 917, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 917, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_3};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 917, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_3};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 917, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 917, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
            __Pyx_GIVEREF(__pyx_t_3);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_3);
            __pyx_t_3 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 917, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":904
 *         self.enterRule(localctx, 2, self.RULE_stored_definition_class)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 160
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":918
 *             self.state = 163
 *             self.match(ModelicaParser.T__0)
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 918, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.stored_definition_class", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_7, &__pyx_t_2) < 0) __PYX_ERR(0, 918, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_7);
        __pyx_v_re = __pyx_t_7;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":919
 *             self.match(ModelicaParser.T__0)
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 919, __pyx_L20_error)

          /* "pymola/generated/ModelicaParser.pyx":920
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 920, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 920, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 920, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 920, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_12 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 920, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_12);
            if (__pyx_t_4) {
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_4); __pyx_t_4 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_12, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 920, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":921
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 921, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 921, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_12);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_12, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 921, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 921, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 921, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 921, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":918
 *             self.state = 163
 *             self.match(ModelicaParser.T__0)
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L21;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L20_error:;
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __pyx_t_6 = __pyx_lineno; __pyx_t_13 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
            }
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_ErrRestore(__pyx_t_15, __pyx_t_16, __pyx_t_17);
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_13; __pyx_filename = __pyx_t_14;
            goto __pyx_L8_except_error;
          }
          __pyx_L21:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":904
 *         self.enterRule(localctx, 2, self.RULE_stored_definition_class)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 160
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":923
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 923, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 923, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 923, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_19, &__pyx_t_18);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_18);
      __pyx_t_13 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_21 = __pyx_filename;
      {
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 923, __pyx_L27_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 923, __pyx_L27_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 923, __pyx_L27_error)
        }
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __pyx_lineno = __pyx_t_13; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_21;
      goto __pyx_L1_error;
      __pyx_L27_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":924
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Class_definitionContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":899
 * 
 * 
 *     def stored_definition_class(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Stored_definition_classContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.stored_definition_class", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":928
 *     class Class_definitionContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Class_definitionContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Class_definitionContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Class_definitionContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Class_definitionContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 928, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 928, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 928, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_definitionContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Class_definitionContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Class_definitionContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":929
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 929, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 929, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 929, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 929, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 929, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 929, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 929, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 929, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":930
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def class_prefixes(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 930, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":928
 *     class Class_definitionContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_definitionContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":932
 *             self.parser = parser
 * 
 *         def class_prefixes(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_prefixesContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Class_definitionContext_3class_prefixes(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Class_definitionContext_3class_prefixes = {"class_prefixes", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Class_definitionContext_3class_prefixes, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Class_definitionContext_3class_prefixes(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("class_prefixes (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Class_definitionContext_2class_prefixes(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Class_definitionContext_2class_prefixes(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("class_prefixes", 0);

  /* "pymola/generated/ModelicaParser.pyx":933
 * 
 *         def class_prefixes(self):
 *             return self.getTypedRuleContext(ModelicaParser.Class_prefixesContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 933, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 933, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Class_prefixesContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 933, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 933, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 933, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 933, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 933, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":932
 *             self.parser = parser
 * 
 *         def class_prefixes(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_prefixesContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_definitionContext.class_prefixes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":936
 * 
 * 
 *         def class_specifier(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_specifierContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Class_definitionContext_5class_specifier(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Class_definitionContext_5class_specifier = {"class_specifier", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Class_definitionContext_5class_specifier, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Class_definitionContext_5class_specifier(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("class_specifier (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Class_definitionContext_4class_specifier(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Class_definitionContext_4class_specifier(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("class_specifier", 0);

  /* "pymola/generated/ModelicaParser.pyx":937
 * 
 *         def class_specifier(self):
 *             return self.getTypedRuleContext(ModelicaParser.Class_specifierContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 937, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 937, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Class_specifierContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 937, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 937, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 937, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 937, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 937, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":936
 * 
 * 
 *         def class_specifier(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_specifierContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_definitionContext.class_specifier", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":940
 * 
 * 
 *         def ENCAPSULATED(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.ENCAPSULATED, 0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Class_definitionContext_7ENCAPSULATED(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Class_definitionContext_7ENCAPSULATED = {"ENCAPSULATED", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Class_definitionContext_7ENCAPSULATED, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Class_definitionContext_7ENCAPSULATED(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ENCAPSULATED (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Class_definitionContext_6ENCAPSULATED(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Class_definitionContext_6ENCAPSULATED(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("ENCAPSULATED", 0);

  /* "pymola/generated/ModelicaParser.pyx":941
 * 
 *         def ENCAPSULATED(self):
 *             return self.getToken(ModelicaParser.ENCAPSULATED, 0)             # <<<<<<<<<<<<<<
 * 
 *         def getRuleIndex(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getToken); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 941, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 941, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_ENCAPSULATED); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 941, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 941, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 941, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 941, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 941, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":940
 * 
 * 
 *         def ENCAPSULATED(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.ENCAPSULATED, 0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_definitionContext.ENCAPSULATED", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":943
 *             return self.getToken(ModelicaParser.ENCAPSULATED, 0)
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_class_definition
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Class_definitionContext_9getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Class_definitionContext_9getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Class_definitionContext_9getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Class_definitionContext_9getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Class_definitionContext_8getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Class_definitionContext_8getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":944
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_class_definition             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 944, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_class_definition); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 944, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":943
 *             return self.getToken(ModelicaParser.ENCAPSULATED, 0)
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_class_definition
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_definitionContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":946
 *             return ModelicaParser.RULE_class_definition
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterClass_definition" ):
 *                 listener.enterClass_definition(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Class_definitionContext_11enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Class_definitionContext_11enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Class_definitionContext_11enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Class_definitionContext_11enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 946, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 946, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 946, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_definitionContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Class_definitionContext_10enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Class_definitionContext_10enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":947
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterClass_definition" ):             # <<<<<<<<<<<<<<
 *                 listener.enterClass_definition(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterClass_definition); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 947, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":948
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterClass_definition" ):
 *                 listener.enterClass_definition(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterClass_definition); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 948, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 948, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 948, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 948, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 948, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 948, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":947
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterClass_definition" ):             # <<<<<<<<<<<<<<
 *                 listener.enterClass_definition(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":946
 *             return ModelicaParser.RULE_class_definition
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterClass_definition" ):
 *                 listener.enterClass_definition(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_definitionContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":950
 *                 listener.enterClass_definition(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitClass_definition" ):
 *                 listener.exitClass_definition(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Class_definitionContext_13exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Class_definitionContext_13exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Class_definitionContext_13exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Class_definitionContext_13exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 950, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 950, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 950, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_definitionContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Class_definitionContext_12exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Class_definitionContext_12exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":951
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitClass_definition" ):             # <<<<<<<<<<<<<<
 *                 listener.exitClass_definition(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitClass_definition); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 951, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":952
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitClass_definition" ):
 *                 listener.exitClass_definition(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitClass_definition); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 952, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 952, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 952, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 952, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 952, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 952, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":951
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitClass_definition" ):             # <<<<<<<<<<<<<<
 *                 listener.exitClass_definition(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":950
 *                 listener.enterClass_definition(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitClass_definition" ):
 *                 listener.exitClass_definition(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_definitionContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":954
 *                 listener.exitClass_definition(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitClass_definition" ):
 *                 return visitor.visitClass_definition(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Class_definitionContext_15accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Class_definitionContext_15accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Class_definitionContext_15accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Class_definitionContext_15accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 954, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 954, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 954, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_definitionContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Class_definitionContext_14accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Class_definitionContext_14accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":955
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitClass_definition" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitClass_definition(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitClass_definition); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 955, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":956
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitClass_definition" ):
 *                 return visitor.visitClass_definition(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitClass_definition); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 956, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 956, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 956, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 956, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 956, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 956, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":955
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitClass_definition" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitClass_definition(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":958
 *                 return visitor.visitClass_definition(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 958, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 958, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 958, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 958, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 958, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 958, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":954
 *                 listener.exitClass_definition(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitClass_definition" ):
 *                 return visitor.visitClass_definition(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_definitionContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":963
 * 
 * 
 *     def class_definition(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Class_definitionContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_7class_definition(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_7class_definition = {"class_definition", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_7class_definition, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_7class_definition(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("class_definition (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_6class_definition(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_6class_definition(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  char const *__pyx_t_21;
  __Pyx_RefNannySetupContext("class_definition", 0);

  /* "pymola/generated/ModelicaParser.pyx":965
 *     def class_definition(self):
 * 
 *         localctx = ModelicaParser.Class_definitionContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 4, self.RULE_class_definition)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 965, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Class_definitionContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 965, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 965, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 965, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 965, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 965, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 965, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 965, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":966
 * 
 *         localctx = ModelicaParser.Class_definitionContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 4, self.RULE_class_definition)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 966, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_class_definition); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 966, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_4, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 966, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_4, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 966, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 966, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_4);
    __Pyx_GIVEREF(__pyx_int_4);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_4);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 966, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":967
 *         localctx = ModelicaParser.Class_definitionContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 4, self.RULE_class_definition)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 967, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":968
 *         self.enterRule(localctx, 4, self.RULE_class_definition)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 166
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":969
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 166
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 969, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 969, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 969, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 969, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 969, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":970
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 166             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_166) < 0) __PYX_ERR(0, 970, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":971
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 166
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.ENCAPSULATED:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 971, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 971, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 971, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 971, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 971, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 971, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 971, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":972
 *             self.state = 166
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.ENCAPSULATED:
 *                 self.state = 165
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 972, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 972, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__818, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 972, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v__la = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":973
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.ENCAPSULATED:             # <<<<<<<<<<<<<<
 *                 self.state = 165
 *                 self.match(ModelicaParser.ENCAPSULATED)
 */
        __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 973, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ENCAPSULATED); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 973, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 973, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 973, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":974
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.ENCAPSULATED:
 *                 self.state = 165             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.ENCAPSULATED)
 * 
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_165) < 0) __PYX_ERR(0, 974, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":975
 *             if _la==ModelicaParser.ENCAPSULATED:
 *                 self.state = 165
 *                 self.match(ModelicaParser.ENCAPSULATED)             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 975, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 975, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_ENCAPSULATED); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 975, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 975, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 975, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 975, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 975, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 975, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":973
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.ENCAPSULATED:             # <<<<<<<<<<<<<<
 *                 self.state = 165
 *                 self.match(ModelicaParser.ENCAPSULATED)
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":978
 * 
 * 
 *             self.state = 168             # <<<<<<<<<<<<<<
 *             self.class_prefixes()
 *             self.state = 169
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_168) < 0) __PYX_ERR(0, 978, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":979
 * 
 *             self.state = 168
 *             self.class_prefixes()             # <<<<<<<<<<<<<<
 *             self.state = 169
 *             self.class_specifier()
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_class_prefixes); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 979, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 979, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 979, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":980
 *             self.state = 168
 *             self.class_prefixes()
 *             self.state = 169             # <<<<<<<<<<<<<<
 *             self.class_specifier()
 *         except RecognitionException as re:
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_169) < 0) __PYX_ERR(0, 980, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":981
 *             self.class_prefixes()
 *             self.state = 169
 *             self.class_specifier()             # <<<<<<<<<<<<<<
 *         except RecognitionException as re:
 *             localctx.exception = re
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_class_specifier); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 981, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 981, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 981, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":968
 *         self.enterRule(localctx, 4, self.RULE_class_definition)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 166
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":982
 *             self.state = 169
 *             self.class_specifier()
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 982, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.class_definition", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_7, &__pyx_t_4) < 0) __PYX_ERR(0, 982, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __pyx_v_re = __pyx_t_7;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":983
 *             self.class_specifier()
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 983, __pyx_L20_error)

          /* "pymola/generated/ModelicaParser.pyx":984
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 984, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 984, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 984, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 984, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_12 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 984, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_12);
            if (__pyx_t_2) {
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_2); __pyx_t_2 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_12, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 984, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":985
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 985, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 985, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_12);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_12, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 985, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 985, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 985, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 985, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":982
 *             self.state = 169
 *             self.class_specifier()
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L21;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L20_error:;
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __pyx_t_6 = __pyx_lineno; __pyx_t_13 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
            }
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_ErrRestore(__pyx_t_15, __pyx_t_16, __pyx_t_17);
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_13; __pyx_filename = __pyx_t_14;
            goto __pyx_L8_except_error;
          }
          __pyx_L21:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":968
 *         self.enterRule(localctx, 4, self.RULE_class_definition)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 166
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":987
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 987, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 987, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 987, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_19, &__pyx_t_18);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_18);
      __pyx_t_13 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_21 = __pyx_filename;
      {
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 987, __pyx_L27_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 987, __pyx_L27_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 987, __pyx_L27_error)
        }
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __pyx_lineno = __pyx_t_13; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_21;
      goto __pyx_L1_error;
      __pyx_L27_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":988
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Class_prefixesContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":963
 * 
 * 
 *     def class_definition(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Class_definitionContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.class_definition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":992
 *     class Class_prefixesContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 992, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 992, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 992, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_prefixesContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Class_prefixesContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Class_prefixesContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":993
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 993, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 993, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 993, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 993, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 993, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 993, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 993, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 993, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":994
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def class_type(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 994, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":992
 *     class Class_prefixesContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_prefixesContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":996
 *             self.parser = parser
 * 
 *         def class_type(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_typeContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_3class_type(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_3class_type = {"class_type", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_3class_type, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_3class_type(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("class_type (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_2class_type(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_2class_type(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("class_type", 0);

  /* "pymola/generated/ModelicaParser.pyx":997
 * 
 *         def class_type(self):
 *             return self.getTypedRuleContext(ModelicaParser.Class_typeContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 997, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 997, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Class_typeContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 997, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 997, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 997, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 997, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 997, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":996
 *             self.parser = parser
 * 
 *         def class_type(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_typeContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_prefixesContext.class_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1000
 * 
 * 
 *         def PARTIAL(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.PARTIAL, 0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_5PARTIAL(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_5PARTIAL = {"PARTIAL", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_5PARTIAL, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_5PARTIAL(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("PARTIAL (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_4PARTIAL(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_4PARTIAL(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("PARTIAL", 0);

  /* "pymola/generated/ModelicaParser.pyx":1001
 * 
 *         def PARTIAL(self):
 *             return self.getToken(ModelicaParser.PARTIAL, 0)             # <<<<<<<<<<<<<<
 * 
 *         def getRuleIndex(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getToken); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1001, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1001, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_PARTIAL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1001, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1001, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1001, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1001, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1001, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1000
 * 
 * 
 *         def PARTIAL(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.PARTIAL, 0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_prefixesContext.PARTIAL", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1003
 *             return self.getToken(ModelicaParser.PARTIAL, 0)
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_class_prefixes
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_7getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_7getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_6getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":1004
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_class_prefixes             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1004, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_class_prefixes); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1004, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1003
 *             return self.getToken(ModelicaParser.PARTIAL, 0)
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_class_prefixes
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_prefixesContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1006
 *             return ModelicaParser.RULE_class_prefixes
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterClass_prefixes" ):
 *                 listener.enterClass_prefixes(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_9enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_9enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 1006, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 1006, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1006, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_prefixesContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_8enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":1007
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterClass_prefixes" ):             # <<<<<<<<<<<<<<
 *                 listener.enterClass_prefixes(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterClass_prefixes); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 1007, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1008
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterClass_prefixes" ):
 *                 listener.enterClass_prefixes(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterClass_prefixes); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1008, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1008, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1008, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1008, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1008, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1008, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":1007
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterClass_prefixes" ):             # <<<<<<<<<<<<<<
 *                 listener.enterClass_prefixes(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1006
 *             return ModelicaParser.RULE_class_prefixes
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterClass_prefixes" ):
 *                 listener.enterClass_prefixes(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_prefixesContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1010
 *                 listener.enterClass_prefixes(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitClass_prefixes" ):
 *                 listener.exitClass_prefixes(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_11exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_11exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 1010, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 1010, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1010, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_prefixesContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_10exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":1011
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitClass_prefixes" ):             # <<<<<<<<<<<<<<
 *                 listener.exitClass_prefixes(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitClass_prefixes); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 1011, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1012
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitClass_prefixes" ):
 *                 listener.exitClass_prefixes(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitClass_prefixes); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1012, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1012, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1012, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1012, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1012, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1012, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":1011
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitClass_prefixes" ):             # <<<<<<<<<<<<<<
 *                 listener.exitClass_prefixes(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1010
 *                 listener.enterClass_prefixes(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitClass_prefixes" ):
 *                 listener.exitClass_prefixes(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_prefixesContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1014
 *                 listener.exitClass_prefixes(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitClass_prefixes" ):
 *                 return visitor.visitClass_prefixes(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_13accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_13accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 1014, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 1014, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1014, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_prefixesContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_12accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":1015
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitClass_prefixes" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitClass_prefixes(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitClass_prefixes); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 1015, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1016
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitClass_prefixes" ):
 *                 return visitor.visitClass_prefixes(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitClass_prefixes); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1016, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1016, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1016, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1016, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1016, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1016, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":1015
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitClass_prefixes" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitClass_prefixes(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1018
 *                 return visitor.visitClass_prefixes(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1018, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1018, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1018, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1018, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1018, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1018, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":1014
 *                 listener.exitClass_prefixes(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitClass_prefixes" ):
 *                 return visitor.visitClass_prefixes(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_prefixesContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1023
 * 
 * 
 *     def class_prefixes(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Class_prefixesContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_9class_prefixes(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_9class_prefixes = {"class_prefixes", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_9class_prefixes, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_9class_prefixes(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("class_prefixes (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_8class_prefixes(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_8class_prefixes(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  char const *__pyx_t_21;
  __Pyx_RefNannySetupContext("class_prefixes", 0);

  /* "pymola/generated/ModelicaParser.pyx":1025
 *     def class_prefixes(self):
 * 
 *         localctx = ModelicaParser.Class_prefixesContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 6, self.RULE_class_prefixes)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1025, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Class_prefixesContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1025, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1025, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1025, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1025, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1025, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1025, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1025, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1026
 * 
 *         localctx = ModelicaParser.Class_prefixesContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 6, self.RULE_class_prefixes)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1026, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_class_prefixes); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1026, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_6, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1026, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_6, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1026, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1026, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_6);
    __Pyx_GIVEREF(__pyx_int_6);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_6);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1026, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1027
 *         localctx = ModelicaParser.Class_prefixesContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 6, self.RULE_class_prefixes)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 1027, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1028
 *         self.enterRule(localctx, 6, self.RULE_class_prefixes)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 172
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":1029
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 172
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1029, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1029, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1029, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1029, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1029, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":1030
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 172             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_172) < 0) __PYX_ERR(0, 1030, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":1031
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 172
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.PARTIAL:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1031, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1031, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1031, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1031, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1031, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1031, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1031, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":1032
 *             self.state = 172
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.PARTIAL:
 *                 self.state = 171
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1032, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1032, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__819, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1032, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v__la = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":1033
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.PARTIAL:             # <<<<<<<<<<<<<<
 *                 self.state = 171
 *                 self.match(ModelicaParser.PARTIAL)
 */
        __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1033, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_PARTIAL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1033, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1033, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1033, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":1034
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.PARTIAL:
 *                 self.state = 171             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.PARTIAL)
 * 
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_171) < 0) __PYX_ERR(0, 1034, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1035
 *             if _la==ModelicaParser.PARTIAL:
 *                 self.state = 171
 *                 self.match(ModelicaParser.PARTIAL)             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1035, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1035, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_PARTIAL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1035, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1035, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1035, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1035, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1035, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1035, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1033
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.PARTIAL:             # <<<<<<<<<<<<<<
 *                 self.state = 171
 *                 self.match(ModelicaParser.PARTIAL)
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":1038
 * 
 * 
 *             self.state = 174             # <<<<<<<<<<<<<<
 *             self.class_type()
 *         except RecognitionException as re:
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_174) < 0) __PYX_ERR(0, 1038, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":1039
 * 
 *             self.state = 174
 *             self.class_type()             # <<<<<<<<<<<<<<
 *         except RecognitionException as re:
 *             localctx.exception = re
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_class_type); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1039, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1039, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1039, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":1028
 *         self.enterRule(localctx, 6, self.RULE_class_prefixes)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 172
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":1040
 *             self.state = 174
 *             self.class_type()
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1040, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.class_prefixes", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_7, &__pyx_t_4) < 0) __PYX_ERR(0, 1040, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __pyx_v_re = __pyx_t_7;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":1041
 *             self.class_type()
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 1041, __pyx_L20_error)

          /* "pymola/generated/ModelicaParser.pyx":1042
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1042, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1042, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1042, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1042, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_12 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1042, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_12);
            if (__pyx_t_2) {
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_2); __pyx_t_2 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_12, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1042, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1043
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1043, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1043, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_12);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_12, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1043, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1043, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1043, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1043, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":1040
 *             self.state = 174
 *             self.class_type()
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L21;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L20_error:;
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __pyx_t_6 = __pyx_lineno; __pyx_t_13 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
            }
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_ErrRestore(__pyx_t_15, __pyx_t_16, __pyx_t_17);
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_13; __pyx_filename = __pyx_t_14;
            goto __pyx_L8_except_error;
          }
          __pyx_L21:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":1028
 *         self.enterRule(localctx, 6, self.RULE_class_prefixes)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 172
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":1045
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1045, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1045, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1045, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_19, &__pyx_t_18);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_18);
      __pyx_t_13 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_21 = __pyx_filename;
      {
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1045, __pyx_L27_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1045, __pyx_L27_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1045, __pyx_L27_error)
        }
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __pyx_lineno = __pyx_t_13; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_21;
      goto __pyx_L1_error;
      __pyx_L27_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":1046
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Class_typeContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1023
 * 
 * 
 *     def class_prefixes(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Class_prefixesContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.class_prefixes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1050
 *     class Class_typeContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_17Class_typeContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_17Class_typeContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_17Class_typeContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_17Class_typeContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 1050, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 1050, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1050, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_typeContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_17Class_typeContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_17Class_typeContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":1051
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 1051, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1051, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1051, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1051, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1051, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1051, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1051, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1051, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1052
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 1052, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1050
 *     class Class_typeContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_typeContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1055
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_class_type
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_17Class_typeContext_3getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_17Class_typeContext_3getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_17Class_typeContext_3getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_17Class_typeContext_3getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_17Class_typeContext_2getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_17Class_typeContext_2getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":1056
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_class_type             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1056, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_class_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1056, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1055
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_class_type
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_typeContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1058
 *             return ModelicaParser.RULE_class_type
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterClass_type" ):
 *                 listener.enterClass_type(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_17Class_typeContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_17Class_typeContext_5enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_17Class_typeContext_5enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_17Class_typeContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 1058, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 1058, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1058, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_typeContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_17Class_typeContext_4enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_17Class_typeContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":1059
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterClass_type" ):             # <<<<<<<<<<<<<<
 *                 listener.enterClass_type(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterClass_type); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 1059, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1060
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterClass_type" ):
 *                 listener.enterClass_type(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterClass_type); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1060, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1060, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1060, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1060, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1060, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1060, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":1059
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterClass_type" ):             # <<<<<<<<<<<<<<
 *                 listener.enterClass_type(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1058
 *             return ModelicaParser.RULE_class_type
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterClass_type" ):
 *                 listener.enterClass_type(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_typeContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1062
 *                 listener.enterClass_type(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitClass_type" ):
 *                 listener.exitClass_type(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_17Class_typeContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_17Class_typeContext_7exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_17Class_typeContext_7exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_17Class_typeContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 1062, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 1062, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1062, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_typeContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_17Class_typeContext_6exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_17Class_typeContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":1063
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitClass_type" ):             # <<<<<<<<<<<<<<
 *                 listener.exitClass_type(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitClass_type); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 1063, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1064
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitClass_type" ):
 *                 listener.exitClass_type(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitClass_type); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1064, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1064, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1064, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1064, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1064, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1064, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":1063
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitClass_type" ):             # <<<<<<<<<<<<<<
 *                 listener.exitClass_type(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1062
 *                 listener.enterClass_type(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitClass_type" ):
 *                 listener.exitClass_type(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_typeContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1066
 *                 listener.exitClass_type(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitClass_type" ):
 *                 return visitor.visitClass_type(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_17Class_typeContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_17Class_typeContext_9accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_17Class_typeContext_9accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_17Class_typeContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 1066, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 1066, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1066, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_typeContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_17Class_typeContext_8accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_17Class_typeContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":1067
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitClass_type" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitClass_type(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitClass_type); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 1067, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1068
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitClass_type" ):
 *                 return visitor.visitClass_type(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitClass_type); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1068, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1068, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1068, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1068, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1068, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1068, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":1067
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitClass_type" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitClass_type(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1070
 *                 return visitor.visitClass_type(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1070, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1070, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1070, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1070, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1070, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1070, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":1066
 *                 listener.exitClass_type(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitClass_type" ):
 *                 return visitor.visitClass_type(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_typeContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1075
 * 
 * 
 *     def class_type(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Class_typeContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_11class_type(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_11class_type = {"class_type", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_11class_type, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_11class_type(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("class_type (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_10class_type(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_10class_type(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v_la_ = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  char const *__pyx_t_22;
  __Pyx_RefNannySetupContext("class_type", 0);

  /* "pymola/generated/ModelicaParser.pyx":1077
 *     def class_type(self):
 * 
 *         localctx = ModelicaParser.Class_typeContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 8, self.RULE_class_type)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1077, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Class_typeContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1077, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1077, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1077, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1077, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1077, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1077, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1077, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1078
 * 
 *         localctx = ModelicaParser.Class_typeContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 8, self.RULE_class_type)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_class_type); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_8, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1078, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_8, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1078, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1078, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_8);
    __Pyx_GIVEREF(__pyx_int_8);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_8);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1078, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1079
 *         localctx = ModelicaParser.Class_typeContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 8, self.RULE_class_type)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.state = 197
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 1079, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1080
 *         self.enterRule(localctx, 8, self.RULE_class_type)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.state = 197
 *             self._errHandler.sync(self)
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":1081
 *         self._la = 0 # Token type
 *         try:
 *             self.state = 197             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             la_ = self._interp.adaptivePredict(self._input,10,self._ctx)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_197) < 0) __PYX_ERR(0, 1081, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":1082
 *         try:
 *             self.state = 197
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             la_ = self._interp.adaptivePredict(self._input,10,self._ctx)
 *             if la_ == 1:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1082, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1082, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1082, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1082, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1082, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1082, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1082, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":1083
 *             self.state = 197
 *             self._errHandler.sync(self)
 *             la_ = self._interp.adaptivePredict(self._input,10,self._ctx)             # <<<<<<<<<<<<<<
 *             if la_ == 1:
 *                 self.enterOuterAlt(localctx, 1)
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_interp); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1083, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_adaptivePredict); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1083, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1083, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1083, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_t_2, __pyx_int_10, __pyx_t_3};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1083, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_t_2, __pyx_int_10, __pyx_t_3};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1083, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1083, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_2);
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_t_2);
          __Pyx_INCREF(__pyx_int_10);
          __Pyx_GIVEREF(__pyx_int_10);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_int_10);
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_6, __pyx_t_3);
          __pyx_t_2 = 0;
          __pyx_t_3 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1083, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v_la_ = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":1084
 *             self._errHandler.sync(self)
 *             la_ = self._interp.adaptivePredict(self._input,10,self._ctx)
 *             if la_ == 1:             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 1)
 *                 self.state = 176
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1084, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1084, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":1085
 *             la_ = self._interp.adaptivePredict(self._input,10,self._ctx)
 *             if la_ == 1:
 *                 self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *                 self.state = 176
 *                 self.match(ModelicaParser.T__1)
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1085, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_localctx, __pyx_int_1};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1085, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_localctx, __pyx_int_1};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1085, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1085, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_1);
            __Pyx_GIVEREF(__pyx_int_1);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_int_1);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1085, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1086
 *             if la_ == 1:
 *                 self.enterOuterAlt(localctx, 1)
 *                 self.state = 176             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__1)
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_176) < 0) __PYX_ERR(0, 1086, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1087
 *                 self.enterOuterAlt(localctx, 1)
 *                 self.state = 176
 *                 self.match(ModelicaParser.T__1)             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1087, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1087, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1087, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_3) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1087, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_5};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1087, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_5};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1087, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            {
              __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1087, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
              __Pyx_GIVEREF(__pyx_t_5);
              PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_5);
              __pyx_t_5 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1087, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1084
 *             self._errHandler.sync(self)
 *             la_ = self._interp.adaptivePredict(self._input,10,self._ctx)
 *             if la_ == 1:             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 1)
 *                 self.state = 176
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":1090
 *                 pass
 * 
 *             elif la_ == 2:             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 2)
 *                 self.state = 177
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1090, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1090, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":1091
 * 
 *             elif la_ == 2:
 *                 self.enterOuterAlt(localctx, 2)             # <<<<<<<<<<<<<<
 *                 self.state = 177
 *                 self.match(ModelicaParser.T__2)
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1091, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1091, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1091, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1091, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            if (__pyx_t_2) {
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_2);
            __Pyx_GIVEREF(__pyx_int_2);
            PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_int_2);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1091, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1092
 *             elif la_ == 2:
 *                 self.enterOuterAlt(localctx, 2)
 *                 self.state = 177             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__2)
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_177) < 0) __PYX_ERR(0, 1092, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1093
 *                 self.enterOuterAlt(localctx, 2)
 *                 self.state = 177
 *                 self.match(ModelicaParser.T__2)             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1093, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1093, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_T__2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1093, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_5) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1093, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_2};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1093, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_2};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1093, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1093, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
              __Pyx_GIVEREF(__pyx_t_2);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_t_2);
              __pyx_t_2 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1093, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1090
 *                 pass
 * 
 *             elif la_ == 2:             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 2)
 *                 self.state = 177
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":1096
 *                 pass
 * 
 *             elif la_ == 3:             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 3)
 *                 self.state = 179
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_3, 3, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1096, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1096, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":1097
 * 
 *             elif la_ == 3:
 *                 self.enterOuterAlt(localctx, 3)             # <<<<<<<<<<<<<<
 *                 self.state = 179
 *                 self._errHandler.sync(self)
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1097, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_3};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1097, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_3};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1097, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1097, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_3);
            __Pyx_GIVEREF(__pyx_int_3);
            PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_3);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1097, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1098
 *             elif la_ == 3:
 *                 self.enterOuterAlt(localctx, 3)
 *                 self.state = 179             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_179) < 0) __PYX_ERR(0, 1098, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1099
 *                 self.enterOuterAlt(localctx, 3)
 *                 self.state = 179
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__3:
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1099, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1099, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1099, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1099, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1099, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1099, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1099, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1100
 *                 self.state = 179
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if _la==ModelicaParser.T__3:
 *                     self.state = 178
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1100, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1100, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__820, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1100, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_v__la = __pyx_t_1;
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1101
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__3:             # <<<<<<<<<<<<<<
 *                     self.state = 178
 *                     self.match(ModelicaParser.T__3)
 */
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1101, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1101, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1101, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1101, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (__pyx_t_11) {

            /* "pymola/generated/ModelicaParser.pyx":1102
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__3:
 *                     self.state = 178             # <<<<<<<<<<<<<<
 *                     self.match(ModelicaParser.T__3)
 * 
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_178) < 0) __PYX_ERR(0, 1102, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":1103
 *                 if _la==ModelicaParser.T__3:
 *                     self.state = 178
 *                     self.match(ModelicaParser.T__3)             # <<<<<<<<<<<<<<
 * 
 * 
 */
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1103, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1103, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1103, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_3)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_3);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
              }
            }
            if (!__pyx_t_3) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1103, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_2)) {
                PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_7};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1103, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
                PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_7};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1103, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              } else
              #endif
              {
                __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1103, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_5);
                __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
                __Pyx_GIVEREF(__pyx_t_7);
                PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_t_7);
                __pyx_t_7 = 0;
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1103, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":1101
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__3:             # <<<<<<<<<<<<<<
 *                     self.state = 178
 *                     self.match(ModelicaParser.T__3)
 */
          }

          /* "pymola/generated/ModelicaParser.pyx":1106
 * 
 * 
 *                 self.state = 181             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__4)
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_181) < 0) __PYX_ERR(0, 1106, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1107
 * 
 *                 self.state = 181
 *                 self.match(ModelicaParser.T__4)             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1107, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1107, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_T__4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1107, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_5) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1107, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1107, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1107, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1107, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
              __Pyx_GIVEREF(__pyx_t_7);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_t_7);
              __pyx_t_7 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1107, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1096
 *                 pass
 * 
 *             elif la_ == 3:             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 3)
 *                 self.state = 179
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":1110
 *                 pass
 * 
 *             elif la_ == 4:             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 4)
 *                 self.state = 182
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_4, 4, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1110, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1110, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":1111
 * 
 *             elif la_ == 4:
 *                 self.enterOuterAlt(localctx, 4)             # <<<<<<<<<<<<<<
 *                 self.state = 182
 *                 self.match(ModelicaParser.T__5)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1111, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_4};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1111, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_4};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1111, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1111, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_4);
            __Pyx_GIVEREF(__pyx_int_4);
            PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_4);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1111, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1112
 *             elif la_ == 4:
 *                 self.enterOuterAlt(localctx, 4)
 *                 self.state = 182             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__5)
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_182) < 0) __PYX_ERR(0, 1112, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1113
 *                 self.enterOuterAlt(localctx, 4)
 *                 self.state = 182
 *                 self.match(ModelicaParser.T__5)             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1113, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1113, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1113, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1113, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1113, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1113, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1113, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1113, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1110
 *                 pass
 * 
 *             elif la_ == 4:             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 4)
 *                 self.state = 182
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":1116
 *                 pass
 * 
 *             elif la_ == 5:             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 5)
 *                 self.state = 184
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_5, 5, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1116, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1116, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":1117
 * 
 *             elif la_ == 5:
 *                 self.enterOuterAlt(localctx, 5)             # <<<<<<<<<<<<<<
 *                 self.state = 184
 *                 self._errHandler.sync(self)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1117, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_localctx, __pyx_int_5};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1117, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_localctx, __pyx_int_5};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1117, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1117, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_5);
            __Pyx_GIVEREF(__pyx_int_5);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_int_5);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1117, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1118
 *             elif la_ == 5:
 *                 self.enterOuterAlt(localctx, 5)
 *                 self.state = 184             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_184) < 0) __PYX_ERR(0, 1118, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1119
 *                 self.enterOuterAlt(localctx, 5)
 *                 self.state = 184
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__6:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1119, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_sync); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1119, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1119, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1119, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1119, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1119, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1119, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1120
 *                 self.state = 184
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if _la==ModelicaParser.T__6:
 *                     self.state = 183
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1120, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1120, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__821, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1120, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_v__la = __pyx_t_1;
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1121
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__6:             # <<<<<<<<<<<<<<
 *                     self.state = 183
 *                     self.match(ModelicaParser.T__6)
 */
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1121, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1121, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1121, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1121, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (__pyx_t_11) {

            /* "pymola/generated/ModelicaParser.pyx":1122
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__6:
 *                     self.state = 183             # <<<<<<<<<<<<<<
 *                     self.match(ModelicaParser.T__6)
 * 
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_183) < 0) __PYX_ERR(0, 1122, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":1123
 *                 if _la==ModelicaParser.T__6:
 *                     self.state = 183
 *                     self.match(ModelicaParser.T__6)             # <<<<<<<<<<<<<<
 * 
 * 
 */
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1123, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1123, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_T__6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1123, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_5 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
              if (likely(__pyx_t_5)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                __Pyx_INCREF(__pyx_t_5);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_3, function);
              }
            }
            if (!__pyx_t_5) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1123, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_3)) {
                PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_2};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1123, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
                PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_2};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1123, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              } else
              #endif
              {
                __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1123, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_7);
                __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
                __Pyx_GIVEREF(__pyx_t_2);
                PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_2);
                __pyx_t_2 = 0;
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1123, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":1121
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__6:             # <<<<<<<<<<<<<<
 *                     self.state = 183
 *                     self.match(ModelicaParser.T__6)
 */
          }

          /* "pymola/generated/ModelicaParser.pyx":1126
 * 
 * 
 *                 self.state = 186             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__7)
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_186) < 0) __PYX_ERR(0, 1126, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1127
 * 
 *                 self.state = 186
 *                 self.match(ModelicaParser.T__7)             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1127, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1127, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1127, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1127, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1127, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1127, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            } else
            #endif
            {
              __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1127, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
              __Pyx_GIVEREF(__pyx_t_2);
              PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_t_2);
              __pyx_t_2 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1127, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1116
 *                 pass
 * 
 *             elif la_ == 5:             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 5)
 *                 self.state = 184
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":1130
 *                 pass
 * 
 *             elif la_ == 6:             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 6)
 *                 self.state = 187
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_6, 6, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1130, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1130, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":1131
 * 
 *             elif la_ == 6:
 *                 self.enterOuterAlt(localctx, 6)             # <<<<<<<<<<<<<<
 *                 self.state = 187
 *                 self.match(ModelicaParser.T__8)
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1131, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_localctx, __pyx_int_6};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1131, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_localctx, __pyx_int_6};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1131, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1131, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_6);
            __Pyx_GIVEREF(__pyx_int_6);
            PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_6);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1131, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1132
 *             elif la_ == 6:
 *                 self.enterOuterAlt(localctx, 6)
 *                 self.state = 187             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__8)
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_187) < 0) __PYX_ERR(0, 1132, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1133
 *                 self.enterOuterAlt(localctx, 6)
 *                 self.state = 187
 *                 self.match(ModelicaParser.T__8)             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1133, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1133, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1133, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1133, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_5};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1133, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_5};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1133, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            {
              __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1133, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_GIVEREF(__pyx_t_5);
              PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_5);
              __pyx_t_5 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1133, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1130
 *                 pass
 * 
 *             elif la_ == 6:             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 6)
 *                 self.state = 187
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":1136
 *                 pass
 * 
 *             elif la_ == 7:             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 7)
 *                 self.state = 188
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_7, 7, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1136, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1136, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":1137
 * 
 *             elif la_ == 7:
 *                 self.enterOuterAlt(localctx, 7)             # <<<<<<<<<<<<<<
 *                 self.state = 188
 *                 self.match(ModelicaParser.T__9)
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1137, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_7 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_localctx, __pyx_int_7};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1137, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_localctx, __pyx_int_7};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1137, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1137, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            if (__pyx_t_7) {
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_7);
            __Pyx_GIVEREF(__pyx_int_7);
            PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_int_7);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1137, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1138
 *             elif la_ == 7:
 *                 self.enterOuterAlt(localctx, 7)
 *                 self.state = 188             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__9)
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_188) < 0) __PYX_ERR(0, 1138, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1139
 *                 self.enterOuterAlt(localctx, 7)
 *                 self.state = 188
 *                 self.match(ModelicaParser.T__9)             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1139, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1139, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_T__9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1139, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (!__pyx_t_5) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1139, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1139, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1139, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            {
              __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1139, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5); __pyx_t_5 = NULL;
              __Pyx_GIVEREF(__pyx_t_7);
              PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_7);
              __pyx_t_7 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1139, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1136
 *                 pass
 * 
 *             elif la_ == 7:             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 7)
 *                 self.state = 188
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":1142
 *                 pass
 * 
 *             elif la_ == 8:             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 8)
 *                 self.state = 190
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_8, 8, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1142, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1142, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":1143
 * 
 *             elif la_ == 8:
 *                 self.enterOuterAlt(localctx, 8)             # <<<<<<<<<<<<<<
 *                 self.state = 190
 *                 self._errHandler.sync(self)
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1143, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_8};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1143, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_8};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1143, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1143, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            if (__pyx_t_2) {
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_8);
            __Pyx_GIVEREF(__pyx_int_8);
            PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_8);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1143, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1144
 *             elif la_ == 8:
 *                 self.enterOuterAlt(localctx, 8)
 *                 self.state = 190             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_190) < 0) __PYX_ERR(0, 1144, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1145
 *                 self.enterOuterAlt(localctx, 8)
 *                 self.state = 190
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__10 or _la==ModelicaParser.T__11:
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1145, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1145, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_3) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1145, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1145, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1145, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1145, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1145, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1146
 *                 self.state = 190
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if _la==ModelicaParser.T__10 or _la==ModelicaParser.T__11:
 *                     self.state = 189
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1146, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1146, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__822, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1146, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_v__la = __pyx_t_1;
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1147
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__10 or _la==ModelicaParser.T__11:             # <<<<<<<<<<<<<<
 *                     self.state = 189
 *                     _la = self._input.LA(1)
 */
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1147, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__10); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1147, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1147, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 1147, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (!__pyx_t_12) {
          } else {
            __pyx_t_11 = __pyx_t_12;
            goto __pyx_L18_bool_binop_done;
          }
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1147, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__11); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1147, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1147, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 1147, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_11 = __pyx_t_12;
          __pyx_L18_bool_binop_done:;
          if (__pyx_t_11) {

            /* "pymola/generated/ModelicaParser.pyx":1148
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__10 or _la==ModelicaParser.T__11:
 *                     self.state = 189             # <<<<<<<<<<<<<<
 *                     _la = self._input.LA(1)
 *                     if not(_la==ModelicaParser.T__10 or _la==ModelicaParser.T__11):
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_189) < 0) __PYX_ERR(0, 1148, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":1149
 *                 if _la==ModelicaParser.T__10 or _la==ModelicaParser.T__11:
 *                     self.state = 189
 *                     _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                     if not(_la==ModelicaParser.T__10 or _la==ModelicaParser.T__11):
 *                         self._errHandler.recoverInline(self)
 */
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1149, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1149, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__823, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1149, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_1);
            __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":1150
 *                     self.state = 189
 *                     _la = self._input.LA(1)
 *                     if not(_la==ModelicaParser.T__10 or _la==ModelicaParser.T__11):             # <<<<<<<<<<<<<<
 *                         self._errHandler.recoverInline(self)
 *                     else:
 */
            __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1150, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__10); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1150, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1150, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 1150, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (!__pyx_t_12) {
            } else {
              __pyx_t_11 = __pyx_t_12;
              goto __pyx_L21_bool_binop_done;
            }
            __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1150, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__11); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1150, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1150, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 1150, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_11 = __pyx_t_12;
            __pyx_L21_bool_binop_done:;
            __pyx_t_12 = ((!__pyx_t_11) != 0);
            if (__pyx_t_12) {

              /* "pymola/generated/ModelicaParser.pyx":1151
 *                     _la = self._input.LA(1)
 *                     if not(_la==ModelicaParser.T__10 or _la==ModelicaParser.T__11):
 *                         self._errHandler.recoverInline(self)             # <<<<<<<<<<<<<<
 *                     else:
 *                         self._errHandler.reportMatch(self)
 */
              __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1151, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_recoverInline); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1151, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __pyx_t_7 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
                __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
                if (likely(__pyx_t_7)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                  __Pyx_INCREF(__pyx_t_7);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_2, function);
                }
              }
              if (!__pyx_t_7) {
                __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1151, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
              } else {
                #if CYTHON_FAST_PYCALL
                if (PyFunction_Check(__pyx_t_2)) {
                  PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
                  __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1151, __pyx_L6_error)
                  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
                  __Pyx_GOTREF(__pyx_t_1);
                } else
                #endif
                #if CYTHON_FAST_PYCCALL
                if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
                  PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
                  __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1151, __pyx_L6_error)
                  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
                  __Pyx_GOTREF(__pyx_t_1);
                } else
                #endif
                {
                  __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1151, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_3);
                  __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
                  __Pyx_INCREF(__pyx_v_self);
                  __Pyx_GIVEREF(__pyx_v_self);
                  PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
                  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1151, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_1);
                  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                }
              }
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "pymola/generated/ModelicaParser.pyx":1150
 *                     self.state = 189
 *                     _la = self._input.LA(1)
 *                     if not(_la==ModelicaParser.T__10 or _la==ModelicaParser.T__11):             # <<<<<<<<<<<<<<
 *                         self._errHandler.recoverInline(self)
 *                     else:
 */
              goto __pyx_L20;
            }

            /* "pymola/generated/ModelicaParser.pyx":1153
 *                         self._errHandler.recoverInline(self)
 *                     else:
 *                         self._errHandler.reportMatch(self)             # <<<<<<<<<<<<<<
 *                         self.consume()
 * 
 */
            /*else*/ {
              __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1153, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_reportMatch); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1153, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __pyx_t_2 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
                __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
                if (likely(__pyx_t_2)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                  __Pyx_INCREF(__pyx_t_2);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_3, function);
                }
              }
              if (!__pyx_t_2) {
                __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1153, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
              } else {
                #if CYTHON_FAST_PYCALL
                if (PyFunction_Check(__pyx_t_3)) {
                  PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
                  __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1153, __pyx_L6_error)
                  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
                  __Pyx_GOTREF(__pyx_t_1);
                } else
                #endif
                #if CYTHON_FAST_PYCCALL
                if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
                  PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
                  __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1153, __pyx_L6_error)
                  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
                  __Pyx_GOTREF(__pyx_t_1);
                } else
                #endif
                {
                  __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1153, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_7);
                  __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
                  __Pyx_INCREF(__pyx_v_self);
                  __Pyx_GIVEREF(__pyx_v_self);
                  PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self);
                  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1153, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_1);
                  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
                }
              }
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "pymola/generated/ModelicaParser.pyx":1154
 *                     else:
 *                         self._errHandler.reportMatch(self)
 *                         self.consume()             # <<<<<<<<<<<<<<
 * 
 * 
 */
              __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_consume); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1154, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __pyx_t_7 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
                __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
                if (likely(__pyx_t_7)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                  __Pyx_INCREF(__pyx_t_7);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_3, function);
                }
              }
              if (__pyx_t_7) {
                __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1154, __pyx_L6_error)
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              } else {
                __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1154, __pyx_L6_error)
              }
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            }
            __pyx_L20:;

            /* "pymola/generated/ModelicaParser.pyx":1147
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__10 or _la==ModelicaParser.T__11:             # <<<<<<<<<<<<<<
 *                     self.state = 189
 *                     _la = self._input.LA(1)
 */
          }

          /* "pymola/generated/ModelicaParser.pyx":1157
 * 
 * 
 *                 self.state = 193             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_193) < 0) __PYX_ERR(0, 1157, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1158
 * 
 *                 self.state = 193
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__3:
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1158, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1158, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_3) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1158, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1158, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1158, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1158, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1158, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1159
 *                 self.state = 193
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if _la==ModelicaParser.T__3:
 *                     self.state = 192
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1159, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1159, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__824, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1159, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1160
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__3:             # <<<<<<<<<<<<<<
 *                     self.state = 192
 *                     self.match(ModelicaParser.T__3)
 */
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1160, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1160, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1160, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 1160, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (__pyx_t_12) {

            /* "pymola/generated/ModelicaParser.pyx":1161
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__3:
 *                     self.state = 192             # <<<<<<<<<<<<<<
 *                     self.match(ModelicaParser.T__3)
 * 
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_192) < 0) __PYX_ERR(0, 1161, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":1162
 *                 if _la==ModelicaParser.T__3:
 *                     self.state = 192
 *                     self.match(ModelicaParser.T__3)             # <<<<<<<<<<<<<<
 * 
 * 
 */
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1162, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1162, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1162, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_2 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
              __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
              if (likely(__pyx_t_2)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
                __Pyx_INCREF(__pyx_t_2);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_7, function);
              }
            }
            if (!__pyx_t_2) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1162, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_7)) {
                PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1162, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
                PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1162, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              } else
              #endif
              {
                __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1162, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_5);
                __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
                __Pyx_GIVEREF(__pyx_t_3);
                PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_t_3);
                __pyx_t_3 = 0;
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1162, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":1160
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__3:             # <<<<<<<<<<<<<<
 *                     self.state = 192
 *                     self.match(ModelicaParser.T__3)
 */
          }

          /* "pymola/generated/ModelicaParser.pyx":1165
 * 
 * 
 *                 self.state = 195             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__12)
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_195) < 0) __PYX_ERR(0, 1165, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1166
 * 
 *                 self.state = 195
 *                 self.match(ModelicaParser.T__12)             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1166, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1166, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_T__12); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1166, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_5) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1166, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1166, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1166, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1166, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5); __pyx_t_5 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1166, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1142
 *                 pass
 * 
 *             elif la_ == 8:             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 8)
 *                 self.state = 190
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":1169
 *                 pass
 * 
 *             elif la_ == 9:             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 9)
 *                 self.state = 196
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_9, 9, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1169, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 1169, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_12) {

          /* "pymola/generated/ModelicaParser.pyx":1170
 * 
 *             elif la_ == 9:
 *                 self.enterOuterAlt(localctx, 9)             # <<<<<<<<<<<<<<
 *                 self.state = 196
 *                 self.match(ModelicaParser.T__3)
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1170, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_9};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1170, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_9};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1170, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1170, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_2) {
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_9);
            __Pyx_GIVEREF(__pyx_int_9);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_int_9);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1170, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1171
 *             elif la_ == 9:
 *                 self.enterOuterAlt(localctx, 9)
 *                 self.state = 196             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__3)
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_196) < 0) __PYX_ERR(0, 1171, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1172
 *                 self.enterOuterAlt(localctx, 9)
 *                 self.state = 196
 *                 self.match(ModelicaParser.T__3)             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1172, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1172, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1172, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_3) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1172, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_2};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1172, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_2};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1172, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            } else
            #endif
            {
              __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1172, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
              __Pyx_GIVEREF(__pyx_t_2);
              PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_t_2);
              __pyx_t_2 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1172, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1169
 *                 pass
 * 
 *             elif la_ == 9:             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 9)
 *                 self.state = 196
 */
        }
        __pyx_L14:;

        /* "pymola/generated/ModelicaParser.pyx":1080
 *         self.enterRule(localctx, 8, self.RULE_class_type)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.state = 197
 *             self._errHandler.sync(self)
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":1176
 * 
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1176, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.class_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_7, &__pyx_t_5) < 0) __PYX_ERR(0, 1176, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_7);
        __pyx_v_re = __pyx_t_7;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":1177
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 1177, __pyx_L29_error)

          /* "pymola/generated/ModelicaParser.pyx":1178
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1178, __pyx_L29_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_reportError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1178, __pyx_L29_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1178, __pyx_L29_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1178, __pyx_L29_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1178, __pyx_L29_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_13, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1178, __pyx_L29_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1179
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1179, __pyx_L29_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_recover); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1179, __pyx_L29_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1179, __pyx_L29_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1179, __pyx_L29_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1179, __pyx_L29_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_4) {
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __pyx_t_4 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1179, __pyx_L29_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":1176
 * 
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L30;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L29_error:;
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_6 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L8_except_error;
          }
          __pyx_L30:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":1080
 *         self.enterRule(localctx, 8, self.RULE_class_type)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.state = 197
 *             self._errHandler.sync(self)
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":1181
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1181, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1181, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_5 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1181, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_14 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1181, __pyx_L36_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1181, __pyx_L36_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_5 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1181, __pyx_L36_error)
        }
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
      __pyx_L36_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":1182
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Class_specifierContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1075
 * 
 * 
 *     def class_type(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Class_typeContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.class_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v_la_);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1186
 *     class Class_specifierContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_specifierContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_specifierContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_specifierContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_specifierContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 1186, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 1186, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1186, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_specifierContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Class_specifierContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_specifierContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":1187
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 1187, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1187, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1187, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1187, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1187, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1188
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 1188, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1186
 *     class Class_specifierContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_specifierContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1191
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_class_specifier
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_specifierContext_3getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_specifierContext_3getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_specifierContext_3getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_specifierContext_3getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Class_specifierContext_2getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_specifierContext_2getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":1192
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_class_specifier             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_class_specifier); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1191
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_class_specifier
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_specifierContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1195
 * 
 * 
 *         def copyFrom(self, ctx:ParserRuleContext):             # <<<<<<<<<<<<<<
 *             super().copyFrom(ctx)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_specifierContext_5copyFrom(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_specifierContext_5copyFrom = {"copyFrom", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_specifierContext_5copyFrom, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_specifierContext_5copyFrom(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copyFrom (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx_2,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("copyFrom", 1, 2, 2, 1); __PYX_ERR(0, 1195, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "copyFrom") < 0)) __PYX_ERR(0, 1195, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("copyFrom", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1195, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_specifierContext.copyFrom", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Class_specifierContext_4copyFrom(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_specifierContext_4copyFrom(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("copyFrom", 0);

  /* "pymola/generated/ModelicaParser.pyx":1196
 * 
 *         def copyFrom(self, ctx:ParserRuleContext):
 *             super().copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 1196, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1196, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1196, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1196, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1196, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1196, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1195
 * 
 * 
 *         def copyFrom(self, ctx:ParserRuleContext):             # <<<<<<<<<<<<<<
 *             super().copyFrom(ctx)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_specifierContext.copyFrom", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1202
 *     class Class_spec_derContext(Class_specifierContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Class_specifierContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 1202, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 1202, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 1202, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1202, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_derContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Class_spec_derContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Class_spec_derContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":1203
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Class_specifierContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 1203, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1203, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1203, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1203, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1203, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1203, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1204
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Class_specifierContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def IDENT(self, i:int=None):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1204, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1204, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1204, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1204, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1202
 *     class Class_spec_derContext(Class_specifierContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Class_specifierContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_derContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1206
 *             self.copyFrom(ctx)
 * 
 *         def IDENT(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTokens(ModelicaParser.IDENT)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_3IDENT(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_3IDENT = {"IDENT", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_3IDENT, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_3IDENT(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("IDENT (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "IDENT") < 0)) __PYX_ERR(0, 1206, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("IDENT", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1206, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_derContext.IDENT", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_2IDENT(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_2IDENT(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("IDENT", 0);

  /* "pymola/generated/ModelicaParser.pyx":1207
 * 
 *         def IDENT(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTokens(ModelicaParser.IDENT)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1208
 *         def IDENT(self, i:int=None):
 *             if i is None:
 *                 return self.getTokens(ModelicaParser.IDENT)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getToken(ModelicaParser.IDENT, i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTokens); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1208, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1208, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1208, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1208, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1208, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":1207
 * 
 *         def IDENT(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTokens(ModelicaParser.IDENT)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1210
 *                 return self.getTokens(ModelicaParser.IDENT)
 *             else:
 *                 return self.getToken(ModelicaParser.IDENT, i)             # <<<<<<<<<<<<<<
 *         def name(self):
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getToken); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1210, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1210, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1210, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1210, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1210, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1210, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1210, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":1206
 *             self.copyFrom(ctx)
 * 
 *         def IDENT(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTokens(ModelicaParser.IDENT)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_derContext.IDENT", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1211
 *             else:
 *                 return self.getToken(ModelicaParser.IDENT, i)
 *         def name(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_5name(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_5name = {"name", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_5name, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_5name(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("name (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_4name(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_4name(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("name", 0);

  /* "pymola/generated/ModelicaParser.pyx":1212
 *                 return self.getToken(ModelicaParser.IDENT, i)
 *         def name(self):
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)             # <<<<<<<<<<<<<<
 * 
 *         def comment(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_NameContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1212, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1212, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1212, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1212, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1211
 *             else:
 *                 return self.getToken(ModelicaParser.IDENT, i)
 *         def name(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_derContext.name", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1214
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)
 * 
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_7comment(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_7comment = {"comment", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_7comment, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_7comment(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("comment (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_6comment(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_6comment(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("comment", 0);

  /* "pymola/generated/ModelicaParser.pyx":1215
 * 
 *         def comment(self):
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_CommentContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1215, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1215, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1215, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1215, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1214
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)
 * 
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_derContext.comment", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1218
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterClass_spec_der" ):
 *                 listener.enterClass_spec_der(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_9enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_9enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 1218, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 1218, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1218, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_derContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_8enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":1219
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterClass_spec_der" ):             # <<<<<<<<<<<<<<
 *                 listener.enterClass_spec_der(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterClass_spec_der); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 1219, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1220
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterClass_spec_der" ):
 *                 listener.enterClass_spec_der(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterClass_spec_der); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1220, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1220, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1220, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1220, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1220, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1220, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":1219
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterClass_spec_der" ):             # <<<<<<<<<<<<<<
 *                 listener.enterClass_spec_der(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1218
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterClass_spec_der" ):
 *                 listener.enterClass_spec_der(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_derContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1222
 *                 listener.enterClass_spec_der(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitClass_spec_der" ):
 *                 listener.exitClass_spec_der(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_11exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_11exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 1222, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 1222, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1222, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_derContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_10exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":1223
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitClass_spec_der" ):             # <<<<<<<<<<<<<<
 *                 listener.exitClass_spec_der(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitClass_spec_der); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 1223, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1224
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitClass_spec_der" ):
 *                 listener.exitClass_spec_der(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitClass_spec_der); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1224, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1224, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1224, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1224, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1224, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1224, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":1223
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitClass_spec_der" ):             # <<<<<<<<<<<<<<
 *                 listener.exitClass_spec_der(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1222
 *                 listener.enterClass_spec_der(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitClass_spec_der" ):
 *                 listener.exitClass_spec_der(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_derContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1226
 *                 listener.exitClass_spec_der(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitClass_spec_der" ):
 *                 return visitor.visitClass_spec_der(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_13accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_13accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 1226, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 1226, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1226, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_derContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_12accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":1227
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitClass_spec_der" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitClass_spec_der(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitClass_spec_der); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 1227, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1228
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitClass_spec_der" ):
 *                 return visitor.visitClass_spec_der(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitClass_spec_der); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1228, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1228, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1228, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1228, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1228, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1228, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":1227
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitClass_spec_der" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitClass_spec_der(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1230
 *                 return visitor.visitClass_spec_der(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1230, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1230, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1230, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1230, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1230, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1230, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":1226
 *                 listener.exitClass_spec_der(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitClass_spec_der" ):
 *                 return visitor.visitClass_spec_der(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_derContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1235
 *     class Class_spec_enumContext(Class_specifierContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Class_specifierContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 1235, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 1235, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 1235, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1235, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_enumContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":1236
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Class_specifierContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 1236, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1236, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1236, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1236, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1236, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1236, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1236, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1236, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1236, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1237
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Class_specifierContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def IDENT(self):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1237, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1237, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1237, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1237, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1237, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1235
 *     class Class_spec_enumContext(Class_specifierContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Class_specifierContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_enumContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1239
 *             self.copyFrom(ctx)
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 *         def comment(self):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_3IDENT(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_3IDENT = {"IDENT", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_3IDENT, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_3IDENT(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("IDENT (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_2IDENT(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_2IDENT(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("IDENT", 0);

  /* "pymola/generated/ModelicaParser.pyx":1240
 * 
 *         def IDENT(self):
 *             return self.getToken(ModelicaParser.IDENT, 0)             # <<<<<<<<<<<<<<
 *         def comment(self):
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getToken); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1240, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1240, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1240, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1240, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1239
 *             self.copyFrom(ctx)
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 *         def comment(self):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_enumContext.IDENT", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1241
 *         def IDENT(self):
 *             return self.getToken(ModelicaParser.IDENT, 0)
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_5comment(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_5comment = {"comment", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_5comment, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_5comment(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("comment (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_4comment(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_4comment(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("comment", 0);

  /* "pymola/generated/ModelicaParser.pyx":1242
 *             return self.getToken(ModelicaParser.IDENT, 0)
 *         def comment(self):
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)             # <<<<<<<<<<<<<<
 * 
 *         def enum_list(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_CommentContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1242, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1242, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1242, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1242, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1241
 *         def IDENT(self):
 *             return self.getToken(ModelicaParser.IDENT, 0)
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_enumContext.comment", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1244
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 *         def enum_list(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Enum_listContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_7enum_list(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_7enum_list = {"enum_list", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_7enum_list, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_7enum_list(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enum_list (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_6enum_list(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_6enum_list(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enum_list", 0);

  /* "pymola/generated/ModelicaParser.pyx":1245
 * 
 *         def enum_list(self):
 *             return self.getTypedRuleContext(ModelicaParser.Enum_listContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Enum_listContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1245, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1245, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1245, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1245, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1244
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 *         def enum_list(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Enum_listContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_enumContext.enum_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1248
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterClass_spec_enum" ):
 *                 listener.enterClass_spec_enum(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_9enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_9enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 1248, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 1248, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1248, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_enumContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_8enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":1249
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterClass_spec_enum" ):             # <<<<<<<<<<<<<<
 *                 listener.enterClass_spec_enum(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterClass_spec_enum); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 1249, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1250
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterClass_spec_enum" ):
 *                 listener.enterClass_spec_enum(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterClass_spec_enum); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1250, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1250, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1250, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1250, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1250, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":1249
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterClass_spec_enum" ):             # <<<<<<<<<<<<<<
 *                 listener.enterClass_spec_enum(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1248
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterClass_spec_enum" ):
 *                 listener.enterClass_spec_enum(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_enumContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1252
 *                 listener.enterClass_spec_enum(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitClass_spec_enum" ):
 *                 listener.exitClass_spec_enum(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_11exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_11exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 1252, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 1252, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1252, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_enumContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_10exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":1253
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitClass_spec_enum" ):             # <<<<<<<<<<<<<<
 *                 listener.exitClass_spec_enum(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitClass_spec_enum); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 1253, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1254
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitClass_spec_enum" ):
 *                 listener.exitClass_spec_enum(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitClass_spec_enum); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1254, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1254, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1254, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1254, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1254, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1254, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":1253
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitClass_spec_enum" ):             # <<<<<<<<<<<<<<
 *                 listener.exitClass_spec_enum(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1252
 *                 listener.enterClass_spec_enum(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitClass_spec_enum" ):
 *                 listener.exitClass_spec_enum(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_enumContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1256
 *                 listener.exitClass_spec_enum(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitClass_spec_enum" ):
 *                 return visitor.visitClass_spec_enum(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_13accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_13accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 1256, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 1256, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1256, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_enumContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_12accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":1257
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitClass_spec_enum" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitClass_spec_enum(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitClass_spec_enum); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 1257, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1258
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitClass_spec_enum" ):
 *                 return visitor.visitClass_spec_enum(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitClass_spec_enum); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1258, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1258, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1258, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1258, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1258, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1258, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":1257
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitClass_spec_enum" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitClass_spec_enum(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1260
 *                 return visitor.visitClass_spec_enum(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1260, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1260, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1260, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1260, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1260, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1260, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":1256
 *                 listener.exitClass_spec_enum(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitClass_spec_enum" ):
 *                 return visitor.visitClass_spec_enum(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_enumContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1265
 *     class Class_spec_baseContext(Class_specifierContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Class_specifierContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 1265, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 1265, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 1265, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1265, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_baseContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":1266
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Class_specifierContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 1266, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1266, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1266, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1266, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1266, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1266, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1267
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Class_specifierContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def IDENT(self):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1267, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1267, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1267, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1267, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1267, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1265
 *     class Class_spec_baseContext(Class_specifierContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Class_specifierContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_baseContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1269
 *             self.copyFrom(ctx)
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 *         def base_prefix(self):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_3IDENT(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_3IDENT = {"IDENT", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_3IDENT, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_3IDENT(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("IDENT (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_2IDENT(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_2IDENT(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("IDENT", 0);

  /* "pymola/generated/ModelicaParser.pyx":1270
 * 
 *         def IDENT(self):
 *             return self.getToken(ModelicaParser.IDENT, 0)             # <<<<<<<<<<<<<<
 *         def base_prefix(self):
 *             return self.getTypedRuleContext(ModelicaParser.Base_prefixContext,0)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getToken); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1270, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1270, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1270, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1270, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1269
 *             self.copyFrom(ctx)
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 *         def base_prefix(self):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_baseContext.IDENT", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1271
 *         def IDENT(self):
 *             return self.getToken(ModelicaParser.IDENT, 0)
 *         def base_prefix(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Base_prefixContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_5base_prefix(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_5base_prefix = {"base_prefix", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_5base_prefix, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_5base_prefix(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("base_prefix (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_4base_prefix(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_4base_prefix(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("base_prefix", 0);

  /* "pymola/generated/ModelicaParser.pyx":1272
 *             return self.getToken(ModelicaParser.IDENT, 0)
 *         def base_prefix(self):
 *             return self.getTypedRuleContext(ModelicaParser.Base_prefixContext,0)             # <<<<<<<<<<<<<<
 * 
 *         def name(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Base_prefixContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1272, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1272, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1272, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1272, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1271
 *         def IDENT(self):
 *             return self.getToken(ModelicaParser.IDENT, 0)
 *         def base_prefix(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Base_prefixContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_baseContext.base_prefix", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1274
 *             return self.getTypedRuleContext(ModelicaParser.Base_prefixContext,0)
 * 
 *         def name(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_7name(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_7name = {"name", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_7name, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_7name(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("name (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_6name(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_6name(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("name", 0);

  /* "pymola/generated/ModelicaParser.pyx":1275
 * 
 *         def name(self):
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)             # <<<<<<<<<<<<<<
 * 
 *         def comment(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_NameContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1275, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1275, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1274
 *             return self.getTypedRuleContext(ModelicaParser.Base_prefixContext,0)
 * 
 *         def name(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_baseContext.name", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1277
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)
 * 
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_9comment(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_9comment = {"comment", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_9comment, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_9comment(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("comment (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_8comment(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_8comment(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("comment", 0);

  /* "pymola/generated/ModelicaParser.pyx":1278
 * 
 *         def comment(self):
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)             # <<<<<<<<<<<<<<
 * 
 *         def array_subscripts(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1278, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1278, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_CommentContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1278, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1278, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1278, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1278, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1278, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1277
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)
 * 
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_baseContext.comment", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1280
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 *         def array_subscripts(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Array_subscriptsContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_11array_subscripts(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_11array_subscripts = {"array_subscripts", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_11array_subscripts, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_11array_subscripts(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("array_subscripts (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_10array_subscripts(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_10array_subscripts(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("array_subscripts", 0);

  /* "pymola/generated/ModelicaParser.pyx":1281
 * 
 *         def array_subscripts(self):
 *             return self.getTypedRuleContext(ModelicaParser.Array_subscriptsContext,0)             # <<<<<<<<<<<<<<
 * 
 *         def class_modification(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Array_subscriptsContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1281, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1281, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1281, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1281, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1280
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 *         def array_subscripts(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Array_subscriptsContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_baseContext.array_subscripts", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1283
 *             return self.getTypedRuleContext(ModelicaParser.Array_subscriptsContext,0)
 * 
 *         def class_modification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_modificationContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_13class_modification(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_13class_modification = {"class_modification", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_13class_modification, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_13class_modification(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("class_modification (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_12class_modification(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_12class_modification(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("class_modification", 0);

  /* "pymola/generated/ModelicaParser.pyx":1284
 * 
 *         def class_modification(self):
 *             return self.getTypedRuleContext(ModelicaParser.Class_modificationContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Class_modificationContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1284, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1284, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1284, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1284, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1283
 *             return self.getTypedRuleContext(ModelicaParser.Array_subscriptsContext,0)
 * 
 *         def class_modification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_modificationContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_baseContext.class_modification", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1287
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterClass_spec_base" ):
 *                 listener.enterClass_spec_base(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_15enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_15enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_15enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_15enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 1287, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 1287, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1287, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_baseContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_14enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_14enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":1288
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterClass_spec_base" ):             # <<<<<<<<<<<<<<
 *                 listener.enterClass_spec_base(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterClass_spec_base); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 1288, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1289
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterClass_spec_base" ):
 *                 listener.enterClass_spec_base(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterClass_spec_base); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1289, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1289, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1289, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1289, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1289, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1289, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":1288
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterClass_spec_base" ):             # <<<<<<<<<<<<<<
 *                 listener.enterClass_spec_base(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1287
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterClass_spec_base" ):
 *                 listener.enterClass_spec_base(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_baseContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1291
 *                 listener.enterClass_spec_base(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitClass_spec_base" ):
 *                 listener.exitClass_spec_base(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_17exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_17exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_17exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_17exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 1291, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 1291, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1291, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_baseContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_16exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_16exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":1292
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitClass_spec_base" ):             # <<<<<<<<<<<<<<
 *                 listener.exitClass_spec_base(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitClass_spec_base); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 1292, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1293
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitClass_spec_base" ):
 *                 listener.exitClass_spec_base(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitClass_spec_base); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1293, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1293, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1293, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1293, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1293, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1293, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":1292
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitClass_spec_base" ):             # <<<<<<<<<<<<<<
 *                 listener.exitClass_spec_base(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1291
 *                 listener.enterClass_spec_base(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitClass_spec_base" ):
 *                 listener.exitClass_spec_base(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_baseContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1295
 *                 listener.exitClass_spec_base(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitClass_spec_base" ):
 *                 return visitor.visitClass_spec_base(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_19accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_19accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_19accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_19accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 1295, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 1295, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1295, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_baseContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_18accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_18accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":1296
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitClass_spec_base" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitClass_spec_base(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitClass_spec_base); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 1296, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1297
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitClass_spec_base" ):
 *                 return visitor.visitClass_spec_base(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitClass_spec_base); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1297, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1297, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1297, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1297, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1297, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1297, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":1296
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitClass_spec_base" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitClass_spec_base(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1299
 *                 return visitor.visitClass_spec_base(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1299, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1299, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1299, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1299, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1299, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1299, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":1295
 *                 listener.exitClass_spec_base(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitClass_spec_base" ):
 *                 return visitor.visitClass_spec_base(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_baseContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1304
 *     class Class_spec_compContext(Class_specifierContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Class_specifierContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 1304, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 1304, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 1304, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1304, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_compContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_compContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_compContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":1305
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Class_specifierContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 1305, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1305, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1305, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1305, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1305, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1305, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1306
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Class_specifierContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def IDENT(self, i:int=None):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1306, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1306, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1306, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1306, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1306, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1304
 *     class Class_spec_compContext(Class_specifierContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Class_specifierContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_compContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1308
 *             self.copyFrom(ctx)
 * 
 *         def IDENT(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTokens(ModelicaParser.IDENT)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_3IDENT(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_3IDENT = {"IDENT", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_3IDENT, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_3IDENT(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("IDENT (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "IDENT") < 0)) __PYX_ERR(0, 1308, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("IDENT", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1308, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_compContext.IDENT", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_2IDENT(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_2IDENT(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("IDENT", 0);

  /* "pymola/generated/ModelicaParser.pyx":1309
 * 
 *         def IDENT(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTokens(ModelicaParser.IDENT)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1310
 *         def IDENT(self, i:int=None):
 *             if i is None:
 *                 return self.getTokens(ModelicaParser.IDENT)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getToken(ModelicaParser.IDENT, i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTokens); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1310, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1310, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1310, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1310, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1310, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1310, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1310, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1310, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":1309
 * 
 *         def IDENT(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTokens(ModelicaParser.IDENT)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1312
 *                 return self.getTokens(ModelicaParser.IDENT)
 *             else:
 *                 return self.getToken(ModelicaParser.IDENT, i)             # <<<<<<<<<<<<<<
 *         def string_comment(self):
 *             return self.getTypedRuleContext(ModelicaParser.String_commentContext,0)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getToken); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1312, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1312, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1312, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1312, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1312, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1312, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1312, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":1308
 *             self.copyFrom(ctx)
 * 
 *         def IDENT(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTokens(ModelicaParser.IDENT)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_compContext.IDENT", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1313
 *             else:
 *                 return self.getToken(ModelicaParser.IDENT, i)
 *         def string_comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.String_commentContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_5string_comment(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_5string_comment = {"string_comment", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_5string_comment, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_5string_comment(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("string_comment (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_4string_comment(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_4string_comment(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("string_comment", 0);

  /* "pymola/generated/ModelicaParser.pyx":1314
 *                 return self.getToken(ModelicaParser.IDENT, i)
 *         def string_comment(self):
 *             return self.getTypedRuleContext(ModelicaParser.String_commentContext,0)             # <<<<<<<<<<<<<<
 * 
 *         def composition(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_String_commentContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1314, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1314, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1314, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1314, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1313
 *             else:
 *                 return self.getToken(ModelicaParser.IDENT, i)
 *         def string_comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.String_commentContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_compContext.string_comment", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1316
 *             return self.getTypedRuleContext(ModelicaParser.String_commentContext,0)
 * 
 *         def composition(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CompositionContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_7composition(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_7composition = {"composition", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_7composition, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_7composition(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("composition (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_6composition(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_6composition(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("composition", 0);

  /* "pymola/generated/ModelicaParser.pyx":1317
 * 
 *         def composition(self):
 *             return self.getTypedRuleContext(ModelicaParser.CompositionContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1317, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1317, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_CompositionContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1317, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1317, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1317, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1317, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1317, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1316
 *             return self.getTypedRuleContext(ModelicaParser.String_commentContext,0)
 * 
 *         def composition(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CompositionContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_compContext.composition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1320
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterClass_spec_comp" ):
 *                 listener.enterClass_spec_comp(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_9enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_9enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 1320, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 1320, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1320, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_compContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_8enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":1321
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterClass_spec_comp" ):             # <<<<<<<<<<<<<<
 *                 listener.enterClass_spec_comp(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterClass_spec_comp); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 1321, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1322
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterClass_spec_comp" ):
 *                 listener.enterClass_spec_comp(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterClass_spec_comp); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1322, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1322, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1322, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1322, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1322, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1322, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":1321
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterClass_spec_comp" ):             # <<<<<<<<<<<<<<
 *                 listener.enterClass_spec_comp(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1320
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterClass_spec_comp" ):
 *                 listener.enterClass_spec_comp(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_compContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1324
 *                 listener.enterClass_spec_comp(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitClass_spec_comp" ):
 *                 listener.exitClass_spec_comp(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_11exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_11exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 1324, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 1324, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1324, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_compContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_10exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":1325
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitClass_spec_comp" ):             # <<<<<<<<<<<<<<
 *                 listener.exitClass_spec_comp(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitClass_spec_comp); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 1325, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1326
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitClass_spec_comp" ):
 *                 listener.exitClass_spec_comp(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitClass_spec_comp); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1326, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1326, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1326, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1326, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1326, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1326, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":1325
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitClass_spec_comp" ):             # <<<<<<<<<<<<<<
 *                 listener.exitClass_spec_comp(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1324
 *                 listener.enterClass_spec_comp(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitClass_spec_comp" ):
 *                 listener.exitClass_spec_comp(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_compContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1328
 *                 listener.exitClass_spec_comp(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitClass_spec_comp" ):
 *                 return visitor.visitClass_spec_comp(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_13accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_13accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 1328, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 1328, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1328, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_compContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_12accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":1329
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitClass_spec_comp" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitClass_spec_comp(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitClass_spec_comp); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 1329, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1330
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitClass_spec_comp" ):
 *                 return visitor.visitClass_spec_comp(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitClass_spec_comp); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1330, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1330, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1330, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1330, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1330, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1330, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":1329
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitClass_spec_comp" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitClass_spec_comp(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1332
 *                 return visitor.visitClass_spec_comp(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1332, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1332, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1332, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1332, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1332, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1332, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":1328
 *                 listener.exitClass_spec_comp(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitClass_spec_comp" ):
 *                 return visitor.visitClass_spec_comp(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_compContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1337
 *     class Class_spec_extendsContext(Class_specifierContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Class_specifierContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 1337, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 1337, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 1337, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1337, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_extendsContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":1338
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Class_specifierContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 1338, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1338, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1338, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1338, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1338, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1338, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1338, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1338, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1338, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1339
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Class_specifierContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def IDENT(self, i:int=None):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1339, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1339, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1339, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1339, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1339, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1337
 *     class Class_spec_extendsContext(Class_specifierContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Class_specifierContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_extendsContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1341
 *             self.copyFrom(ctx)
 * 
 *         def IDENT(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTokens(ModelicaParser.IDENT)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_3IDENT(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_3IDENT = {"IDENT", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_3IDENT, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_3IDENT(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("IDENT (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "IDENT") < 0)) __PYX_ERR(0, 1341, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("IDENT", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1341, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_extendsContext.IDENT", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_2IDENT(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_2IDENT(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("IDENT", 0);

  /* "pymola/generated/ModelicaParser.pyx":1342
 * 
 *         def IDENT(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTokens(ModelicaParser.IDENT)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1343
 *         def IDENT(self, i:int=None):
 *             if i is None:
 *                 return self.getTokens(ModelicaParser.IDENT)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getToken(ModelicaParser.IDENT, i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTokens); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1343, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1343, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1343, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1343, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1343, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1343, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1343, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1343, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":1342
 * 
 *         def IDENT(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTokens(ModelicaParser.IDENT)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1345
 *                 return self.getTokens(ModelicaParser.IDENT)
 *             else:
 *                 return self.getToken(ModelicaParser.IDENT, i)             # <<<<<<<<<<<<<<
 *         def string_comment(self):
 *             return self.getTypedRuleContext(ModelicaParser.String_commentContext,0)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getToken); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1345, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1345, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1345, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1345, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1345, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1345, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1345, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":1341
 *             self.copyFrom(ctx)
 * 
 *         def IDENT(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTokens(ModelicaParser.IDENT)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_extendsContext.IDENT", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1346
 *             else:
 *                 return self.getToken(ModelicaParser.IDENT, i)
 *         def string_comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.String_commentContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_5string_comment(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_5string_comment = {"string_comment", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_5string_comment, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_5string_comment(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("string_comment (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_4string_comment(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_4string_comment(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("string_comment", 0);

  /* "pymola/generated/ModelicaParser.pyx":1347
 *                 return self.getToken(ModelicaParser.IDENT, i)
 *         def string_comment(self):
 *             return self.getTypedRuleContext(ModelicaParser.String_commentContext,0)             # <<<<<<<<<<<<<<
 * 
 *         def composition(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_String_commentContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1347, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1347, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1347, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1347, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1346
 *             else:
 *                 return self.getToken(ModelicaParser.IDENT, i)
 *         def string_comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.String_commentContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_extendsContext.string_comment", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1349
 *             return self.getTypedRuleContext(ModelicaParser.String_commentContext,0)
 * 
 *         def composition(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CompositionContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_7composition(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_7composition = {"composition", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_7composition, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_7composition(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("composition (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_6composition(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_6composition(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("composition", 0);

  /* "pymola/generated/ModelicaParser.pyx":1350
 * 
 *         def composition(self):
 *             return self.getTypedRuleContext(ModelicaParser.CompositionContext,0)             # <<<<<<<<<<<<<<
 * 
 *         def class_modification(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_CompositionContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1350, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1350, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1350, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1350, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1349
 *             return self.getTypedRuleContext(ModelicaParser.String_commentContext,0)
 * 
 *         def composition(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CompositionContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_extendsContext.composition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1352
 *             return self.getTypedRuleContext(ModelicaParser.CompositionContext,0)
 * 
 *         def class_modification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_modificationContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_9class_modification(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_9class_modification = {"class_modification", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_9class_modification, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_9class_modification(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("class_modification (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_8class_modification(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_8class_modification(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("class_modification", 0);

  /* "pymola/generated/ModelicaParser.pyx":1353
 * 
 *         def class_modification(self):
 *             return self.getTypedRuleContext(ModelicaParser.Class_modificationContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1353, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1353, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Class_modificationContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1353, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1353, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1353, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1353, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1353, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1352
 *             return self.getTypedRuleContext(ModelicaParser.CompositionContext,0)
 * 
 *         def class_modification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_modificationContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_extendsContext.class_modification", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1356
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterClass_spec_extends" ):
 *                 listener.enterClass_spec_extends(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_11enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_11enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_11enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_11enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 1356, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 1356, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1356, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_extendsContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_10enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_10enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":1357
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterClass_spec_extends" ):             # <<<<<<<<<<<<<<
 *                 listener.enterClass_spec_extends(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterClass_spec_extends); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 1357, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1358
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterClass_spec_extends" ):
 *                 listener.enterClass_spec_extends(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterClass_spec_extends); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1358, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1358, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1358, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1358, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1358, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1358, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":1357
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterClass_spec_extends" ):             # <<<<<<<<<<<<<<
 *                 listener.enterClass_spec_extends(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1356
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterClass_spec_extends" ):
 *                 listener.enterClass_spec_extends(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_extendsContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1360
 *                 listener.enterClass_spec_extends(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitClass_spec_extends" ):
 *                 listener.exitClass_spec_extends(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_13exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_13exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_13exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_13exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 1360, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 1360, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1360, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_extendsContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_12exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_12exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":1361
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitClass_spec_extends" ):             # <<<<<<<<<<<<<<
 *                 listener.exitClass_spec_extends(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitClass_spec_extends); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 1361, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1362
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitClass_spec_extends" ):
 *                 listener.exitClass_spec_extends(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitClass_spec_extends); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1362, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1362, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1362, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1362, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1362, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1362, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":1361
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitClass_spec_extends" ):             # <<<<<<<<<<<<<<
 *                 listener.exitClass_spec_extends(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1360
 *                 listener.enterClass_spec_extends(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitClass_spec_extends" ):
 *                 listener.exitClass_spec_extends(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_extendsContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1364
 *                 listener.exitClass_spec_extends(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitClass_spec_extends" ):
 *                 return visitor.visitClass_spec_extends(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_15accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_15accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_15accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_15accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 1364, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 1364, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1364, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_extendsContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_14accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_14accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":1365
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitClass_spec_extends" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitClass_spec_extends(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitClass_spec_extends); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 1365, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1366
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitClass_spec_extends" ):
 *                 return visitor.visitClass_spec_extends(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitClass_spec_extends); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1366, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1366, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1366, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1366, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1366, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1366, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":1365
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitClass_spec_extends" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitClass_spec_extends(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1368
 *                 return visitor.visitClass_spec_extends(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1368, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1368, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1368, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1368, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1368, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1368, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":1364
 *                 listener.exitClass_spec_extends(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitClass_spec_extends" ):
 *                 return visitor.visitClass_spec_extends(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_spec_extendsContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1372
 * 
 * 
 *     def class_specifier(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Class_specifierContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_13class_specifier(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_13class_specifier = {"class_specifier", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_13class_specifier, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_13class_specifier(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("class_specifier (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_12class_specifier(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_12class_specifier(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v_la_ = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_token = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  char const *__pyx_t_22;
  __Pyx_RefNannySetupContext("class_specifier", 0);

  /* "pymola/generated/ModelicaParser.pyx":1374
 *     def class_specifier(self):
 * 
 *         localctx = ModelicaParser.Class_specifierContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 10, self.RULE_class_specifier)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Class_specifierContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1374, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1374, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1374, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1374, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1375
 * 
 *         localctx = ModelicaParser.Class_specifierContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 10, self.RULE_class_specifier)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_class_specifier); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_10, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1375, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_10, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1375, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1375, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_10);
    __Pyx_GIVEREF(__pyx_int_10);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_10);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1375, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1376
 *         localctx = ModelicaParser.Class_specifierContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 10, self.RULE_class_specifier)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.state = 256
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 1376, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1377
 *         self.enterRule(localctx, 10, self.RULE_class_specifier)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.state = 256
 *             self._errHandler.sync(self)
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":1378
 *         self._la = 0 # Token type
 *         try:
 *             self.state = 256             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             la_ = self._interp.adaptivePredict(self._input,17,self._ctx)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_256) < 0) __PYX_ERR(0, 1378, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":1379
 *         try:
 *             self.state = 256
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             la_ = self._interp.adaptivePredict(self._input,17,self._ctx)
 *             if la_ == 1:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1379, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1379, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1379, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1379, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1379, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1379, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1379, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":1380
 *             self.state = 256
 *             self._errHandler.sync(self)
 *             la_ = self._interp.adaptivePredict(self._input,17,self._ctx)             # <<<<<<<<<<<<<<
 *             if la_ == 1:
 *                 localctx = ModelicaParser.Class_spec_compContext(self, localctx)
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_interp); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1380, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_adaptivePredict); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1380, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1380, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1380, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_t_2, __pyx_int_17, __pyx_t_3};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1380, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_t_2, __pyx_int_17, __pyx_t_3};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1380, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1380, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_2);
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_t_2);
          __Pyx_INCREF(__pyx_int_17);
          __Pyx_GIVEREF(__pyx_int_17);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_int_17);
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_6, __pyx_t_3);
          __pyx_t_2 = 0;
          __pyx_t_3 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1380, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v_la_ = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":1381
 *             self._errHandler.sync(self)
 *             la_ = self._interp.adaptivePredict(self._input,17,self._ctx)
 *             if la_ == 1:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Class_spec_compContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 1)
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1381, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1381, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":1382
 *             la_ = self._interp.adaptivePredict(self._input,17,self._ctx)
 *             if la_ == 1:
 *                 localctx = ModelicaParser.Class_spec_compContext(self, localctx)             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 1)
 *                 self.state = 199
 */
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1382, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Class_spec_compContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1382, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1382, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1382, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1382, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_7) {
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_v_localctx);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1382, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1383
 *             if la_ == 1:
 *                 localctx = ModelicaParser.Class_spec_compContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *                 self.state = 199
 *                 self.match(ModelicaParser.IDENT)
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1383, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_1};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1383, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_1};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1383, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1383, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_1);
            __Pyx_GIVEREF(__pyx_int_1);
            PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1383, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1384
 *                 localctx = ModelicaParser.Class_spec_compContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 1)
 *                 self.state = 199             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.IDENT)
 *                 self.state = 200
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_199) < 0) __PYX_ERR(0, 1384, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1385
 *                 self.enterOuterAlt(localctx, 1)
 *                 self.state = 199
 *                 self.match(ModelicaParser.IDENT)             # <<<<<<<<<<<<<<
 *                 self.state = 200
 *                 self.string_comment()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1385, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1385, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1385, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1385, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_5)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1385, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1385, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1385, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_7); __pyx_t_7 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1385, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1386
 *                 self.state = 199
 *                 self.match(ModelicaParser.IDENT)
 *                 self.state = 200             # <<<<<<<<<<<<<<
 *                 self.string_comment()
 *                 self.state = 201
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_200) < 0) __PYX_ERR(0, 1386, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1387
 *                 self.match(ModelicaParser.IDENT)
 *                 self.state = 200
 *                 self.string_comment()             # <<<<<<<<<<<<<<
 *                 self.state = 201
 *                 self.composition()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_string_comment); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1387, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
            }
          }
          if (__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1387, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1387, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1388
 *                 self.state = 200
 *                 self.string_comment()
 *                 self.state = 201             # <<<<<<<<<<<<<<
 *                 self.composition()
 *                 self.state = 202
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_201) < 0) __PYX_ERR(0, 1388, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1389
 *                 self.string_comment()
 *                 self.state = 201
 *                 self.composition()             # <<<<<<<<<<<<<<
 *                 self.state = 202
 *                 self.match(ModelicaParser.T__13)
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_composition); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1389, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
            }
          }
          if (__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1389, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1389, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1390
 *                 self.state = 201
 *                 self.composition()
 *                 self.state = 202             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__13)
 *                 self.state = 203
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_202) < 0) __PYX_ERR(0, 1390, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1391
 *                 self.composition()
 *                 self.state = 202
 *                 self.match(ModelicaParser.T__13)             # <<<<<<<<<<<<<<
 *                 self.state = 203
 *                 self.match(ModelicaParser.IDENT)
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1391, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1391, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__13); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1391, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1391, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_5)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1391, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1391, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1391, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1391, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1392
 *                 self.state = 202
 *                 self.match(ModelicaParser.T__13)
 *                 self.state = 203             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.IDENT)
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_203) < 0) __PYX_ERR(0, 1392, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1393
 *                 self.match(ModelicaParser.T__13)
 *                 self.state = 203
 *                 self.match(ModelicaParser.IDENT)             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1393, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1393, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1393, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1393, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_5)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1393, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1393, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1393, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_7); __pyx_t_7 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1393, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1381
 *             self._errHandler.sync(self)
 *             la_ = self._interp.adaptivePredict(self._input,17,self._ctx)
 *             if la_ == 1:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Class_spec_compContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 1)
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":1396
 *                 pass
 * 
 *             elif la_ == 2:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Class_spec_baseContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 2)
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1396, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1396, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":1397
 * 
 *             elif la_ == 2:
 *                 localctx = ModelicaParser.Class_spec_baseContext(self, localctx)             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 2)
 *                 self.state = 205
 */
          __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1397, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_Class_spec_baseContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1397, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1397, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1397, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1397, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_v_localctx);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1397, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1398
 *             elif la_ == 2:
 *                 localctx = ModelicaParser.Class_spec_baseContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 2)             # <<<<<<<<<<<<<<
 *                 self.state = 205
 *                 self.match(ModelicaParser.IDENT)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1398, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1398, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1398, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1398, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_2);
            __Pyx_GIVEREF(__pyx_int_2);
            PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_int_2);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1398, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1399
 *                 localctx = ModelicaParser.Class_spec_baseContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 2)
 *                 self.state = 205             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.IDENT)
 *                 self.state = 206
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_205) < 0) __PYX_ERR(0, 1399, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1400
 *                 self.enterOuterAlt(localctx, 2)
 *                 self.state = 205
 *                 self.match(ModelicaParser.IDENT)             # <<<<<<<<<<<<<<
 *                 self.state = 206
 *                 self.match(ModelicaParser.T__14)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1400, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1400, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1400, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_5) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1400, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1400, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1400, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1400, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1400, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1401
 *                 self.state = 205
 *                 self.match(ModelicaParser.IDENT)
 *                 self.state = 206             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__14)
 *                 self.state = 207
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_206) < 0) __PYX_ERR(0, 1401, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1402
 *                 self.match(ModelicaParser.IDENT)
 *                 self.state = 206
 *                 self.match(ModelicaParser.T__14)             # <<<<<<<<<<<<<<
 *                 self.state = 207
 *                 self.base_prefix()
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1402, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1402, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__14); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1402, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1402, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1402, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1402, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1402, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1402, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1403
 *                 self.state = 206
 *                 self.match(ModelicaParser.T__14)
 *                 self.state = 207             # <<<<<<<<<<<<<<
 *                 self.base_prefix()
 *                 self.state = 208
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_207) < 0) __PYX_ERR(0, 1403, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1404
 *                 self.match(ModelicaParser.T__14)
 *                 self.state = 207
 *                 self.base_prefix()             # <<<<<<<<<<<<<<
 *                 self.state = 208
 *                 self.name()
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_base_prefix); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1404, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (__pyx_t_5) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1404, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1404, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1405
 *                 self.state = 207
 *                 self.base_prefix()
 *                 self.state = 208             # <<<<<<<<<<<<<<
 *                 self.name()
 *                 self.state = 210
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_208) < 0) __PYX_ERR(0, 1405, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1406
 *                 self.base_prefix()
 *                 self.state = 208
 *                 self.name()             # <<<<<<<<<<<<<<
 *                 self.state = 210
 *                 self._errHandler.sync(self)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1406, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (__pyx_t_5) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1406, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1406, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1407
 *                 self.state = 208
 *                 self.name()
 *                 self.state = 210             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_210) < 0) __PYX_ERR(0, 1407, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1408
 *                 self.name()
 *                 self.state = 210
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__75:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1408, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_sync); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1408, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1408, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_5)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1408, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1408, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1408, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1408, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1409
 *                 self.state = 210
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if _la==ModelicaParser.T__75:
 *                     self.state = 209
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1409, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1409, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_tuple__825, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1409, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_v__la = __pyx_t_1;
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1410
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__75:             # <<<<<<<<<<<<<<
 *                     self.state = 209
 *                     self.array_subscripts()
 */
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1410, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__75); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1410, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1410, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1410, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (__pyx_t_11) {

            /* "pymola/generated/ModelicaParser.pyx":1411
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__75:
 *                     self.state = 209             # <<<<<<<<<<<<<<
 *                     self.array_subscripts()
 * 
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_209) < 0) __PYX_ERR(0, 1411, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":1412
 *                 if _la==ModelicaParser.T__75:
 *                     self.state = 209
 *                     self.array_subscripts()             # <<<<<<<<<<<<<<
 * 
 * 
 */
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_array_subscripts); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1412, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_3 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
              __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
              if (likely(__pyx_t_3)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                __Pyx_INCREF(__pyx_t_3);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_5, function);
              }
            }
            if (__pyx_t_3) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1412, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else {
              __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1412, __pyx_L6_error)
            }
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":1410
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__75:             # <<<<<<<<<<<<<<
 *                     self.state = 209
 *                     self.array_subscripts()
 */
          }

          /* "pymola/generated/ModelicaParser.pyx":1415
 * 
 * 
 *                 self.state = 213             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_213) < 0) __PYX_ERR(0, 1415, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1416
 * 
 *                 self.state = 213
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__16:
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1416, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_sync); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1416, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (!__pyx_t_5) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1416, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1416, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1416, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1416, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5); __pyx_t_5 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1416, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1417
 *                 self.state = 213
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if _la==ModelicaParser.T__16:
 *                     self.state = 212
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1417, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1417, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__826, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1417, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1418
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__16:             # <<<<<<<<<<<<<<
 *                     self.state = 212
 *                     self.class_modification()
 */
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1418, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__16); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1418, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1418, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1418, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (__pyx_t_11) {

            /* "pymola/generated/ModelicaParser.pyx":1419
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__16:
 *                     self.state = 212             # <<<<<<<<<<<<<<
 *                     self.class_modification()
 * 
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_212) < 0) __PYX_ERR(0, 1419, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":1420
 *                 if _la==ModelicaParser.T__16:
 *                     self.state = 212
 *                     self.class_modification()             # <<<<<<<<<<<<<<
 * 
 * 
 */
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_class_modification); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1420, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_2 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
              __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
              if (likely(__pyx_t_2)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                __Pyx_INCREF(__pyx_t_2);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_3, function);
              }
            }
            if (__pyx_t_2) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1420, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            } else {
              __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1420, __pyx_L6_error)
            }
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":1418
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__16:             # <<<<<<<<<<<<<<
 *                     self.state = 212
 *                     self.class_modification()
 */
          }

          /* "pymola/generated/ModelicaParser.pyx":1423
 * 
 * 
 *                 self.state = 215             # <<<<<<<<<<<<<<
 *                 self.comment()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_215) < 0) __PYX_ERR(0, 1423, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1424
 * 
 *                 self.state = 215
 *                 self.comment()             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_comment); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1424, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1424, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1424, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1396
 *                 pass
 * 
 *             elif la_ == 2:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Class_spec_baseContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 2)
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":1427
 *                 pass
 * 
 *             elif la_ == 3:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Class_spec_enumContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 3)
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_3, 3, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1427, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1427, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":1428
 * 
 *             elif la_ == 3:
 *                 localctx = ModelicaParser.Class_spec_enumContext(self, localctx)             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 3)
 *                 self.state = 217
 */
          __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1428, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Class_spec_enumContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1428, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1428, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1428, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1428, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_v_localctx);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1428, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1429
 *             elif la_ == 3:
 *                 localctx = ModelicaParser.Class_spec_enumContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 3)             # <<<<<<<<<<<<<<
 *                 self.state = 217
 *                 self.match(ModelicaParser.IDENT)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1429, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_localctx, __pyx_int_3};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1429, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_localctx, __pyx_int_3};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1429, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1429, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_3);
            __Pyx_GIVEREF(__pyx_int_3);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_int_3);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1429, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1430
 *                 localctx = ModelicaParser.Class_spec_enumContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 3)
 *                 self.state = 217             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.IDENT)
 *                 self.state = 218
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_217) < 0) __PYX_ERR(0, 1430, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1431
 *                 self.enterOuterAlt(localctx, 3)
 *                 self.state = 217
 *                 self.match(ModelicaParser.IDENT)             # <<<<<<<<<<<<<<
 *                 self.state = 218
 *                 self.match(ModelicaParser.T__14)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1431, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1431, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1431, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_3) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1431, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_5};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1431, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_5};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1431, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            {
              __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1431, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
              __Pyx_GIVEREF(__pyx_t_5);
              PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_5);
              __pyx_t_5 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1431, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1432
 *                 self.state = 217
 *                 self.match(ModelicaParser.IDENT)
 *                 self.state = 218             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__14)
 *                 self.state = 219
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_218) < 0) __PYX_ERR(0, 1432, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1433
 *                 self.match(ModelicaParser.IDENT)
 *                 self.state = 218
 *                 self.match(ModelicaParser.T__14)             # <<<<<<<<<<<<<<
 *                 self.state = 219
 *                 self.match(ModelicaParser.T__15)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1433, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1433, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__14); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1433, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1433, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_5};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1433, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_5};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1433, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1433, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
              __Pyx_GIVEREF(__pyx_t_5);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_t_5);
              __pyx_t_5 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1433, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1434
 *                 self.state = 218
 *                 self.match(ModelicaParser.T__14)
 *                 self.state = 219             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__15)
 *                 self.state = 220
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_219) < 0) __PYX_ERR(0, 1434, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1435
 *                 self.match(ModelicaParser.T__14)
 *                 self.state = 219
 *                 self.match(ModelicaParser.T__15)             # <<<<<<<<<<<<<<
 *                 self.state = 220
 *                 self.match(ModelicaParser.T__16)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1435, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1435, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__15); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1435, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_3) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1435, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_5};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1435, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_5};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1435, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            {
              __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1435, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
              __Pyx_GIVEREF(__pyx_t_5);
              PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_5);
              __pyx_t_5 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1435, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1436
 *                 self.state = 219
 *                 self.match(ModelicaParser.T__15)
 *                 self.state = 220             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__16)
 *                 self.state = 225
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_220) < 0) __PYX_ERR(0, 1436, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1437
 *                 self.match(ModelicaParser.T__15)
 *                 self.state = 220
 *                 self.match(ModelicaParser.T__16)             # <<<<<<<<<<<<<<
 *                 self.state = 225
 *                 self._errHandler.sync(self)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1437, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1437, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__16); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1437, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1437, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_5};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1437, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_5};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1437, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1437, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
              __Pyx_GIVEREF(__pyx_t_5);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_t_5);
              __pyx_t_5 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1437, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1438
 *                 self.state = 220
 *                 self.match(ModelicaParser.T__16)
 *                 self.state = 225             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 token = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_225) < 0) __PYX_ERR(0, 1438, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1439
 *                 self.match(ModelicaParser.T__16)
 *                 self.state = 225
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 token = self._input.LA(1)
 *                 if token in [ModelicaParser.T__18, ModelicaParser.IDENT]:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1439, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_sync); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1439, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1439, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1439, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1439, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1439, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1439, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1440
 *                 self.state = 225
 *                 self._errHandler.sync(self)
 *                 token = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if token in [ModelicaParser.T__18, ModelicaParser.IDENT]:
 *                     self.state = 222
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1440, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1440, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__827, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1440, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_v_token = __pyx_t_1;
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1441
 *                 self._errHandler.sync(self)
 *                 token = self._input.LA(1)
 *                 if token in [ModelicaParser.T__18, ModelicaParser.IDENT]:             # <<<<<<<<<<<<<<
 *                     self.state = 222
 *                     self._errHandler.sync(self)
 */
          __Pyx_INCREF(__pyx_v_token);
          __pyx_t_1 = __pyx_v_token;
          __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1441, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__18); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1441, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1441, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 1441, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (!__pyx_t_12) {
          } else {
            __pyx_t_11 = __pyx_t_12;
            goto __pyx_L18_bool_binop_done;
          }
          __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1441, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1441, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1441, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 1441, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_11 = __pyx_t_12;
          __pyx_L18_bool_binop_done:;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_12 = (__pyx_t_11 != 0);
          if (__pyx_t_12) {

            /* "pymola/generated/ModelicaParser.pyx":1442
 *                 token = self._input.LA(1)
 *                 if token in [ModelicaParser.T__18, ModelicaParser.IDENT]:
 *                     self.state = 222             # <<<<<<<<<<<<<<
 *                     self._errHandler.sync(self)
 *                     _la = self._input.LA(1)
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_222) < 0) __PYX_ERR(0, 1442, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":1443
 *                 if token in [ModelicaParser.T__18, ModelicaParser.IDENT]:
 *                     self.state = 222
 *                     self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                     _la = self._input.LA(1)
 *                     if _la==ModelicaParser.IDENT:
 */
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1443, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1443, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
              __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
              if (likely(__pyx_t_3)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                __Pyx_INCREF(__pyx_t_3);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_5, function);
              }
            }
            if (!__pyx_t_3) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1443, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_5)) {
                PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1443, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
                PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1443, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else
              #endif
              {
                __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1443, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
                __Pyx_INCREF(__pyx_v_self);
                __Pyx_GIVEREF(__pyx_v_self);
                PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1443, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":1444
 *                     self.state = 222
 *                     self._errHandler.sync(self)
 *                     _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                     if _la==ModelicaParser.IDENT:
 *                         self.state = 221
 */
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1444, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1444, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_tuple__828, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1444, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_v__la = __pyx_t_1;
            __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":1445
 *                     self._errHandler.sync(self)
 *                     _la = self._input.LA(1)
 *                     if _la==ModelicaParser.IDENT:             # <<<<<<<<<<<<<<
 *                         self.state = 221
 *                         self.enum_list()
 */
            __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1445, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1445, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1445, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 1445, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (__pyx_t_12) {

              /* "pymola/generated/ModelicaParser.pyx":1446
 *                     _la = self._input.LA(1)
 *                     if _la==ModelicaParser.IDENT:
 *                         self.state = 221             # <<<<<<<<<<<<<<
 *                         self.enum_list()
 * 
 */
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_221) < 0) __PYX_ERR(0, 1446, __pyx_L6_error)

              /* "pymola/generated/ModelicaParser.pyx":1447
 *                     if _la==ModelicaParser.IDENT:
 *                         self.state = 221
 *                         self.enum_list()             # <<<<<<<<<<<<<<
 * 
 * 
 */
              __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enum_list); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1447, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_2 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
                __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
                if (likely(__pyx_t_2)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                  __Pyx_INCREF(__pyx_t_2);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_5, function);
                }
              }
              if (__pyx_t_2) {
                __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1447, __pyx_L6_error)
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              } else {
                __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1447, __pyx_L6_error)
              }
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "pymola/generated/ModelicaParser.pyx":1445
 *                     self._errHandler.sync(self)
 *                     _la = self._input.LA(1)
 *                     if _la==ModelicaParser.IDENT:             # <<<<<<<<<<<<<<
 *                         self.state = 221
 *                         self.enum_list()
 */
            }

            /* "pymola/generated/ModelicaParser.pyx":1441
 *                 self._errHandler.sync(self)
 *                 token = self._input.LA(1)
 *                 if token in [ModelicaParser.T__18, ModelicaParser.IDENT]:             # <<<<<<<<<<<<<<
 *                     self.state = 222
 *                     self._errHandler.sync(self)
 */
            goto __pyx_L17;
          }

          /* "pymola/generated/ModelicaParser.pyx":1451
 * 
 *                     pass
 *                 elif token in [ModelicaParser.T__17]:             # <<<<<<<<<<<<<<
 *                     self.state = 224
 *                     self.match(ModelicaParser.T__17)
 */
          __Pyx_INCREF(__pyx_v_token);
          __pyx_t_1 = __pyx_v_token;
          __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1451, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_T__17); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1451, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1451, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 1451, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_11 = (__pyx_t_12 != 0);
          if (__pyx_t_11) {

            /* "pymola/generated/ModelicaParser.pyx":1452
 *                     pass
 *                 elif token in [ModelicaParser.T__17]:
 *                     self.state = 224             # <<<<<<<<<<<<<<
 *                     self.match(ModelicaParser.T__17)
 *                     pass
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_224) < 0) __PYX_ERR(0, 1452, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":1453
 *                 elif token in [ModelicaParser.T__17]:
 *                     self.state = 224
 *                     self.match(ModelicaParser.T__17)             # <<<<<<<<<<<<<<
 *                     pass
 *                 else:
 */
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1453, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1453, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__17); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1453, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_2 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
              __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
              if (likely(__pyx_t_2)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                __Pyx_INCREF(__pyx_t_2);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_5, function);
              }
            }
            if (!__pyx_t_2) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1453, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_5)) {
                PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1453, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
                PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1453, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              } else
              #endif
              {
                __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1453, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_7);
                __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
                __Pyx_GIVEREF(__pyx_t_3);
                PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_3);
                __pyx_t_3 = 0;
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1453, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":1451
 * 
 *                     pass
 *                 elif token in [ModelicaParser.T__17]:             # <<<<<<<<<<<<<<
 *                     self.state = 224
 *                     self.match(ModelicaParser.T__17)
 */
            goto __pyx_L17;
          }

          /* "pymola/generated/ModelicaParser.pyx":1456
 *                     pass
 *                 else:
 *                     raise NoViableAltException(self)             # <<<<<<<<<<<<<<
 * 
 *                 self.state = 227
 */
          /*else*/ {
            __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_NoViableAltException); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1456, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_7 = NULL;
            if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
              __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
              if (likely(__pyx_t_7)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                __Pyx_INCREF(__pyx_t_7);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_5, function);
              }
            }
            if (!__pyx_t_7) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1456, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_5)) {
                PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1456, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
                PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1456, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else
              #endif
              {
                __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1456, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
                __Pyx_INCREF(__pyx_v_self);
                __Pyx_GIVEREF(__pyx_v_self);
                PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1456, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_Raise(__pyx_t_1, 0, 0, 0);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __PYX_ERR(0, 1456, __pyx_L6_error)
          }
          __pyx_L17:;

          /* "pymola/generated/ModelicaParser.pyx":1458
 *                     raise NoViableAltException(self)
 * 
 *                 self.state = 227             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__18)
 *                 self.state = 228
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_227) < 0) __PYX_ERR(0, 1458, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1459
 * 
 *                 self.state = 227
 *                 self.match(ModelicaParser.T__18)             # <<<<<<<<<<<<<<
 *                 self.state = 228
 *                 self.comment()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1459, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1459, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__18); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1459, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
            }
          }
          if (!__pyx_t_3) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1459, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_5)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1459, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1459, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            {
              __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1459, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
              __Pyx_GIVEREF(__pyx_t_7);
              PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_7);
              __pyx_t_7 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1459, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1460
 *                 self.state = 227
 *                 self.match(ModelicaParser.T__18)
 *                 self.state = 228             # <<<<<<<<<<<<<<
 *                 self.comment()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_228) < 0) __PYX_ERR(0, 1460, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1461
 *                 self.match(ModelicaParser.T__18)
 *                 self.state = 228
 *                 self.comment()             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_comment); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1461, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
            }
          }
          if (__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1461, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1461, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1427
 *                 pass
 * 
 *             elif la_ == 3:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Class_spec_enumContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 3)
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":1464
 *                 pass
 * 
 *             elif la_ == 4:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Class_spec_derContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 4)
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_4, 4, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1464, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1464, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":1465
 * 
 *             elif la_ == 4:
 *                 localctx = ModelicaParser.Class_spec_derContext(self, localctx)             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 4)
 *                 self.state = 229
 */
          __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1465, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_Class_spec_derContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1465, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1465, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1465, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1465, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_localctx);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1465, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1466
 *             elif la_ == 4:
 *                 localctx = ModelicaParser.Class_spec_derContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 4)             # <<<<<<<<<<<<<<
 *                 self.state = 229
 *                 self.match(ModelicaParser.IDENT)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1466, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_localctx, __pyx_int_4};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1466, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_localctx, __pyx_int_4};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1466, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1466, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            if (__pyx_t_7) {
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_4);
            __Pyx_GIVEREF(__pyx_int_4);
            PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_int_4);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1466, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1467
 *                 localctx = ModelicaParser.Class_spec_derContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 4)
 *                 self.state = 229             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.IDENT)
 *                 self.state = 230
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_229) < 0) __PYX_ERR(0, 1467, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1468
 *                 self.enterOuterAlt(localctx, 4)
 *                 self.state = 229
 *                 self.match(ModelicaParser.IDENT)             # <<<<<<<<<<<<<<
 *                 self.state = 230
 *                 self.match(ModelicaParser.T__14)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1468, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1468, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1468, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_5) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1468, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1468, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1468, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1468, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
              __Pyx_GIVEREF(__pyx_t_7);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_t_7);
              __pyx_t_7 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1468, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1469
 *                 self.state = 229
 *                 self.match(ModelicaParser.IDENT)
 *                 self.state = 230             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__14)
 *                 self.state = 231
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_230) < 0) __PYX_ERR(0, 1469, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1470
 *                 self.match(ModelicaParser.IDENT)
 *                 self.state = 230
 *                 self.match(ModelicaParser.T__14)             # <<<<<<<<<<<<<<
 *                 self.state = 231
 *                 self.match(ModelicaParser.T__19)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1470, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1470, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__14); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1470, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_3) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1470, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1470, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1470, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            {
              __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1470, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
              __Pyx_GIVEREF(__pyx_t_7);
              PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_t_7);
              __pyx_t_7 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1470, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1471
 *                 self.state = 230
 *                 self.match(ModelicaParser.T__14)
 *                 self.state = 231             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__19)
 *                 self.state = 232
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_231) < 0) __PYX_ERR(0, 1471, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1472
 *                 self.match(ModelicaParser.T__14)
 *                 self.state = 231
 *                 self.match(ModelicaParser.T__19)             # <<<<<<<<<<<<<<
 *                 self.state = 232
 *                 self.match(ModelicaParser.T__16)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1472, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1472, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_T__19); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1472, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_5) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1472, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1472, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1472, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1472, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
              __Pyx_GIVEREF(__pyx_t_7);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_t_7);
              __pyx_t_7 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1472, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1473
 *                 self.state = 231
 *                 self.match(ModelicaParser.T__19)
 *                 self.state = 232             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__16)
 *                 self.state = 233
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_232) < 0) __PYX_ERR(0, 1473, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1474
 *                 self.match(ModelicaParser.T__19)
 *                 self.state = 232
 *                 self.match(ModelicaParser.T__16)             # <<<<<<<<<<<<<<
 *                 self.state = 233
 *                 self.name()
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1474, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1474, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__16); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1474, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_3) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1474, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1474, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1474, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            {
              __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1474, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
              __Pyx_GIVEREF(__pyx_t_7);
              PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_t_7);
              __pyx_t_7 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1474, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1475
 *                 self.state = 232
 *                 self.match(ModelicaParser.T__16)
 *                 self.state = 233             # <<<<<<<<<<<<<<
 *                 self.name()
 *                 self.state = 234
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_233) < 0) __PYX_ERR(0, 1475, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1476
 *                 self.match(ModelicaParser.T__16)
 *                 self.state = 233
 *                 self.name()             # <<<<<<<<<<<<<<
 *                 self.state = 234
 *                 self.match(ModelicaParser.T__20)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1476, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (__pyx_t_5) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1476, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1476, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1477
 *                 self.state = 233
 *                 self.name()
 *                 self.state = 234             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__20)
 *                 self.state = 235
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_234) < 0) __PYX_ERR(0, 1477, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1478
 *                 self.name()
 *                 self.state = 234
 *                 self.match(ModelicaParser.T__20)             # <<<<<<<<<<<<<<
 *                 self.state = 235
 *                 self.match(ModelicaParser.IDENT)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1478, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1478, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_T__20); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1478, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_5) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1478, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1478, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1478, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1478, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
              __Pyx_GIVEREF(__pyx_t_7);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_t_7);
              __pyx_t_7 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1478, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1479
 *                 self.state = 234
 *                 self.match(ModelicaParser.T__20)
 *                 self.state = 235             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.IDENT)
 *                 self.state = 240
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_235) < 0) __PYX_ERR(0, 1479, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1480
 *                 self.match(ModelicaParser.T__20)
 *                 self.state = 235
 *                 self.match(ModelicaParser.IDENT)             # <<<<<<<<<<<<<<
 *                 self.state = 240
 *                 self._errHandler.sync(self)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1480, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1480, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1480, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_3) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1480, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1480, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1480, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            {
              __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1480, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
              __Pyx_GIVEREF(__pyx_t_7);
              PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_t_7);
              __pyx_t_7 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1480, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1481
 *                 self.state = 235
 *                 self.match(ModelicaParser.IDENT)
 *                 self.state = 240             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_240) < 0) __PYX_ERR(0, 1481, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1482
 *                 self.match(ModelicaParser.IDENT)
 *                 self.state = 240
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 *                 while _la==ModelicaParser.T__20:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1482, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_sync); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1482, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1482, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_5)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1482, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1482, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1482, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1482, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1483
 *                 self.state = 240
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 while _la==ModelicaParser.T__20:
 *                     self.state = 236
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1483, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1483, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_tuple__829, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1483, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_v__la = __pyx_t_1;
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1484
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 while _la==ModelicaParser.T__20:             # <<<<<<<<<<<<<<
 *                     self.state = 236
 *                     self.match(ModelicaParser.T__20)
 */
          while (1) {
            __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1484, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__20); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1484, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1484, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1484, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (!__pyx_t_11) break;

            /* "pymola/generated/ModelicaParser.pyx":1485
 *                 _la = self._input.LA(1)
 *                 while _la==ModelicaParser.T__20:
 *                     self.state = 236             # <<<<<<<<<<<<<<
 *                     self.match(ModelicaParser.T__20)
 *                     self.state = 237
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_236) < 0) __PYX_ERR(0, 1485, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":1486
 *                 while _la==ModelicaParser.T__20:
 *                     self.state = 236
 *                     self.match(ModelicaParser.T__20)             # <<<<<<<<<<<<<<
 *                     self.state = 237
 *                     self.match(ModelicaParser.IDENT)
 */
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1486, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1486, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__20); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1486, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __pyx_t_7 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
              __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
              if (likely(__pyx_t_7)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                __Pyx_INCREF(__pyx_t_7);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_5, function);
              }
            }
            if (!__pyx_t_7) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1486, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_5)) {
                PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1486, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
                PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1486, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              } else
              #endif
              {
                __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1486, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
                __Pyx_GIVEREF(__pyx_t_2);
                PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_t_2);
                __pyx_t_2 = 0;
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1486, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":1487
 *                     self.state = 236
 *                     self.match(ModelicaParser.T__20)
 *                     self.state = 237             # <<<<<<<<<<<<<<
 *                     self.match(ModelicaParser.IDENT)
 *                     self.state = 242
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_237) < 0) __PYX_ERR(0, 1487, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":1488
 *                     self.match(ModelicaParser.T__20)
 *                     self.state = 237
 *                     self.match(ModelicaParser.IDENT)             # <<<<<<<<<<<<<<
 *                     self.state = 242
 *                     self._errHandler.sync(self)
 */
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1488, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1488, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1488, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
              __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
              if (likely(__pyx_t_3)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                __Pyx_INCREF(__pyx_t_3);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_5, function);
              }
            }
            if (!__pyx_t_3) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1488, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_5)) {
                PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_2};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1488, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
                PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_2};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1488, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              } else
              #endif
              {
                __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1488, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_7);
                __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
                __Pyx_GIVEREF(__pyx_t_2);
                PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_2);
                __pyx_t_2 = 0;
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1488, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":1489
 *                     self.state = 237
 *                     self.match(ModelicaParser.IDENT)
 *                     self.state = 242             # <<<<<<<<<<<<<<
 *                     self._errHandler.sync(self)
 *                     _la = self._input.LA(1)
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_242) < 0) __PYX_ERR(0, 1489, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":1490
 *                     self.match(ModelicaParser.IDENT)
 *                     self.state = 242
 *                     self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                     _la = self._input.LA(1)
 * 
 */
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1490, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1490, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_5 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
              if (likely(__pyx_t_5)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
                __Pyx_INCREF(__pyx_t_5);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_7, function);
              }
            }
            if (!__pyx_t_5) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1490, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_7)) {
                PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1490, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
                PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1490, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else
              #endif
              {
                __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1490, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5); __pyx_t_5 = NULL;
                __Pyx_INCREF(__pyx_v_self);
                __Pyx_GIVEREF(__pyx_v_self);
                PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1490, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":1491
 *                     self.state = 242
 *                     self._errHandler.sync(self)
 *                     _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *                 self.state = 243
 */
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1491, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1491, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__830, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1491, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_1);
            __pyx_t_1 = 0;
          }

          /* "pymola/generated/ModelicaParser.pyx":1493
 *                     _la = self._input.LA(1)
 * 
 *                 self.state = 243             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__18)
 *                 self.state = 244
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_243) < 0) __PYX_ERR(0, 1493, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1494
 * 
 *                 self.state = 243
 *                 self.match(ModelicaParser.T__18)             # <<<<<<<<<<<<<<
 *                 self.state = 244
 *                 self.comment()
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1494, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1494, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__18); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1494, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1494, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_5};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1494, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_5};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1494, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1494, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_GIVEREF(__pyx_t_5);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_t_5);
              __pyx_t_5 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1494, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1495
 *                 self.state = 243
 *                 self.match(ModelicaParser.T__18)
 *                 self.state = 244             # <<<<<<<<<<<<<<
 *                 self.comment()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_244) < 0) __PYX_ERR(0, 1495, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1496
 *                 self.match(ModelicaParser.T__18)
 *                 self.state = 244
 *                 self.comment()             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_comment); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1496, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (__pyx_t_3) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1496, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1496, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1464
 *                 pass
 * 
 *             elif la_ == 4:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Class_spec_derContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 4)
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":1499
 *                 pass
 * 
 *             elif la_ == 5:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Class_spec_extendsContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 5)
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_5, 5, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1499, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1499, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":1500
 * 
 *             elif la_ == 5:
 *                 localctx = ModelicaParser.Class_spec_extendsContext(self, localctx)             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 5)
 *                 self.state = 246
 */
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1500, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Class_spec_extendsContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1500, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1500, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1500, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1500, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            if (__pyx_t_7) {
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_v_localctx);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1500, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1501
 *             elif la_ == 5:
 *                 localctx = ModelicaParser.Class_spec_extendsContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 5)             # <<<<<<<<<<<<<<
 *                 self.state = 246
 *                 self.match(ModelicaParser.T__21)
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1501, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_localctx, __pyx_int_5};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1501, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_localctx, __pyx_int_5};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1501, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1501, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_5);
            __Pyx_GIVEREF(__pyx_int_5);
            PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_5);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1501, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1502
 *                 localctx = ModelicaParser.Class_spec_extendsContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 5)
 *                 self.state = 246             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__21)
 *                 self.state = 247
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_246) < 0) __PYX_ERR(0, 1502, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1503
 *                 self.enterOuterAlt(localctx, 5)
 *                 self.state = 246
 *                 self.match(ModelicaParser.T__21)             # <<<<<<<<<<<<<<
 *                 self.state = 247
 *                 self.match(ModelicaParser.IDENT)
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1503, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1503, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__21); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1503, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1503, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_5};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1503, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_5};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1503, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            {
              __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1503, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_7); __pyx_t_7 = NULL;
              __Pyx_GIVEREF(__pyx_t_5);
              PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_5);
              __pyx_t_5 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1503, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1504
 *                 self.state = 246
 *                 self.match(ModelicaParser.T__21)
 *                 self.state = 247             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.IDENT)
 *                 self.state = 249
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_247) < 0) __PYX_ERR(0, 1504, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1505
 *                 self.match(ModelicaParser.T__21)
 *                 self.state = 247
 *                 self.match(ModelicaParser.IDENT)             # <<<<<<<<<<<<<<
 *                 self.state = 249
 *                 self._errHandler.sync(self)
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1505, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1505, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1505, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1505, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_5};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1505, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_5};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1505, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            {
              __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1505, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_GIVEREF(__pyx_t_5);
              PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_5);
              __pyx_t_5 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1505, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1506
 *                 self.state = 247
 *                 self.match(ModelicaParser.IDENT)
 *                 self.state = 249             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_249) < 0) __PYX_ERR(0, 1506, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1507
 *                 self.match(ModelicaParser.IDENT)
 *                 self.state = 249
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__16:
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1507, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1507, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_3) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1507, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1507, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1507, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1507, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1507, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1508
 *                 self.state = 249
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if _la==ModelicaParser.T__16:
 *                     self.state = 248
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1508, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1508, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__831, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1508, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_v__la = __pyx_t_1;
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1509
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__16:             # <<<<<<<<<<<<<<
 *                     self.state = 248
 *                     self.class_modification()
 */
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1509, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__16); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1509, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1509, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1509, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (__pyx_t_11) {

            /* "pymola/generated/ModelicaParser.pyx":1510
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__16:
 *                     self.state = 248             # <<<<<<<<<<<<<<
 *                     self.class_modification()
 * 
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_248) < 0) __PYX_ERR(0, 1510, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":1511
 *                 if _la==ModelicaParser.T__16:
 *                     self.state = 248
 *                     self.class_modification()             # <<<<<<<<<<<<<<
 * 
 * 
 */
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_class_modification); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1511, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_5 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
              if (likely(__pyx_t_5)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
                __Pyx_INCREF(__pyx_t_5);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_7, function);
              }
            }
            if (__pyx_t_5) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1511, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else {
              __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1511, __pyx_L6_error)
            }
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":1509
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__16:             # <<<<<<<<<<<<<<
 *                     self.state = 248
 *                     self.class_modification()
 */
          }

          /* "pymola/generated/ModelicaParser.pyx":1514
 * 
 * 
 *                 self.state = 251             # <<<<<<<<<<<<<<
 *                 self.string_comment()
 *                 self.state = 252
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_251) < 0) __PYX_ERR(0, 1514, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1515
 * 
 *                 self.state = 251
 *                 self.string_comment()             # <<<<<<<<<<<<<<
 *                 self.state = 252
 *                 self.composition()
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_string_comment); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1515, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (__pyx_t_5) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1515, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1515, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1516
 *                 self.state = 251
 *                 self.string_comment()
 *                 self.state = 252             # <<<<<<<<<<<<<<
 *                 self.composition()
 *                 self.state = 253
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_252) < 0) __PYX_ERR(0, 1516, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1517
 *                 self.string_comment()
 *                 self.state = 252
 *                 self.composition()             # <<<<<<<<<<<<<<
 *                 self.state = 253
 *                 self.match(ModelicaParser.T__13)
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_composition); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1517, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (__pyx_t_5) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1517, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1517, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1518
 *                 self.state = 252
 *                 self.composition()
 *                 self.state = 253             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__13)
 *                 self.state = 254
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_253) < 0) __PYX_ERR(0, 1518, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1519
 *                 self.composition()
 *                 self.state = 253
 *                 self.match(ModelicaParser.T__13)             # <<<<<<<<<<<<<<
 *                 self.state = 254
 *                 self.match(ModelicaParser.IDENT)
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1519, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1519, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_T__13); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1519, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_5) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1519, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1519, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1519, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1519, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5); __pyx_t_5 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1519, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1520
 *                 self.state = 253
 *                 self.match(ModelicaParser.T__13)
 *                 self.state = 254             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.IDENT)
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_254) < 0) __PYX_ERR(0, 1520, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1521
 *                 self.match(ModelicaParser.T__13)
 *                 self.state = 254
 *                 self.match(ModelicaParser.IDENT)             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1521, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1521, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1521, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1521, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1521, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1521, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1521, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1521, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1499
 *                 pass
 * 
 *             elif la_ == 5:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Class_spec_extendsContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 5)
 */
        }
        __pyx_L14:;

        /* "pymola/generated/ModelicaParser.pyx":1377
 *         self.enterRule(localctx, 10, self.RULE_class_specifier)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.state = 256
 *             self._errHandler.sync(self)
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":1525
 * 
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1525, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.class_specifier", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_7, &__pyx_t_5) < 0) __PYX_ERR(0, 1525, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_7);
        __pyx_v_re = __pyx_t_7;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":1526
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 1526, __pyx_L29_error)

          /* "pymola/generated/ModelicaParser.pyx":1527
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1527, __pyx_L29_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_reportError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1527, __pyx_L29_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1527, __pyx_L29_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1527, __pyx_L29_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1527, __pyx_L29_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_2) {
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_2); __pyx_t_2 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_13, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1527, __pyx_L29_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1528
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1528, __pyx_L29_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_recover); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1528, __pyx_L29_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1528, __pyx_L29_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1528, __pyx_L29_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1528, __pyx_L29_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (__pyx_t_4) {
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1528, __pyx_L29_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":1525
 * 
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L30;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L29_error:;
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_6 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L8_except_error;
          }
          __pyx_L30:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":1377
 *         self.enterRule(localctx, 10, self.RULE_class_specifier)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.state = 256
 *             self._errHandler.sync(self)
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":1530
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1530, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1530, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_5 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1530, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_14 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1530, __pyx_L36_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1530, __pyx_L36_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_5 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1530, __pyx_L36_error)
        }
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
      __pyx_L36_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":1531
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Base_prefixContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1372
 * 
 * 
 *     def class_specifier(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Class_specifierContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.class_specifier", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v_la_);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_token);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1535
 *     class Base_prefixContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Base_prefixContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18Base_prefixContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18Base_prefixContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Base_prefixContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 1535, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 1535, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1535, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Base_prefixContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18Base_prefixContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Base_prefixContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":1536
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 1536, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1536, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1536, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1536, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1536, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1537
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def type_prefix(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 1537, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1535
 *     class Base_prefixContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Base_prefixContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1539
 *             self.parser = parser
 * 
 *         def type_prefix(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Type_prefixContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Base_prefixContext_3type_prefix(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18Base_prefixContext_3type_prefix = {"type_prefix", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18Base_prefixContext_3type_prefix, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Base_prefixContext_3type_prefix(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("type_prefix (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18Base_prefixContext_2type_prefix(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Base_prefixContext_2type_prefix(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("type_prefix", 0);

  /* "pymola/generated/ModelicaParser.pyx":1540
 * 
 *         def type_prefix(self):
 *             return self.getTypedRuleContext(ModelicaParser.Type_prefixContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Type_prefixContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1540, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1540, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1539
 *             self.parser = parser
 * 
 *         def type_prefix(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Type_prefixContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Base_prefixContext.type_prefix", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1543
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_base_prefix
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Base_prefixContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18Base_prefixContext_5getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18Base_prefixContext_5getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Base_prefixContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18Base_prefixContext_4getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Base_prefixContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":1544
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_base_prefix             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_base_prefix); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1543
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_base_prefix
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Base_prefixContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1546
 *             return ModelicaParser.RULE_base_prefix
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterBase_prefix" ):
 *                 listener.enterBase_prefix(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Base_prefixContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18Base_prefixContext_7enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18Base_prefixContext_7enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Base_prefixContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 1546, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 1546, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1546, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Base_prefixContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18Base_prefixContext_6enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Base_prefixContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":1547
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterBase_prefix" ):             # <<<<<<<<<<<<<<
 *                 listener.enterBase_prefix(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterBase_prefix); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 1547, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1548
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterBase_prefix" ):
 *                 listener.enterBase_prefix(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterBase_prefix); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1548, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1548, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1548, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1548, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1548, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1548, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":1547
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterBase_prefix" ):             # <<<<<<<<<<<<<<
 *                 listener.enterBase_prefix(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1546
 *             return ModelicaParser.RULE_base_prefix
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterBase_prefix" ):
 *                 listener.enterBase_prefix(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Base_prefixContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1550
 *                 listener.enterBase_prefix(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitBase_prefix" ):
 *                 listener.exitBase_prefix(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Base_prefixContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18Base_prefixContext_9exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18Base_prefixContext_9exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Base_prefixContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 1550, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 1550, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1550, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Base_prefixContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18Base_prefixContext_8exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Base_prefixContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":1551
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitBase_prefix" ):             # <<<<<<<<<<<<<<
 *                 listener.exitBase_prefix(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitBase_prefix); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 1551, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1552
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitBase_prefix" ):
 *                 listener.exitBase_prefix(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitBase_prefix); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1552, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1552, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1552, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1552, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1552, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1552, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":1551
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitBase_prefix" ):             # <<<<<<<<<<<<<<
 *                 listener.exitBase_prefix(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1550
 *                 listener.enterBase_prefix(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitBase_prefix" ):
 *                 listener.exitBase_prefix(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Base_prefixContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1554
 *                 listener.exitBase_prefix(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitBase_prefix" ):
 *                 return visitor.visitBase_prefix(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Base_prefixContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18Base_prefixContext_11accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18Base_prefixContext_11accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Base_prefixContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 1554, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 1554, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1554, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Base_prefixContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18Base_prefixContext_10accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Base_prefixContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":1555
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitBase_prefix" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitBase_prefix(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitBase_prefix); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 1555, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1556
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitBase_prefix" ):
 *                 return visitor.visitBase_prefix(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitBase_prefix); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1556, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1556, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1556, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1556, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1556, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1556, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":1555
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitBase_prefix" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitBase_prefix(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1558
 *                 return visitor.visitBase_prefix(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1558, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1558, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1558, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1558, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1558, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1558, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":1554
 *                 listener.exitBase_prefix(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitBase_prefix" ):
 *                 return visitor.visitBase_prefix(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Base_prefixContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1563
 * 
 * 
 *     def base_prefix(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Base_prefixContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_15base_prefix(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_15base_prefix = {"base_prefix", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_15base_prefix, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_15base_prefix(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("base_prefix (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_14base_prefix(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_14base_prefix(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  char const *__pyx_t_20;
  __Pyx_RefNannySetupContext("base_prefix", 0);

  /* "pymola/generated/ModelicaParser.pyx":1565
 *     def base_prefix(self):
 * 
 *         localctx = ModelicaParser.Base_prefixContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 12, self.RULE_base_prefix)
 *         try:
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Base_prefixContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1565, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1565, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1565, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1565, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1566
 * 
 *         localctx = ModelicaParser.Base_prefixContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 12, self.RULE_base_prefix)             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_base_prefix); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_12, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1566, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_12, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1566, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1566, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_12);
    __Pyx_GIVEREF(__pyx_int_12);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_12);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1566, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1567
 *         localctx = ModelicaParser.Base_prefixContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 12, self.RULE_base_prefix)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 258
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":1568
 *         self.enterRule(localctx, 12, self.RULE_base_prefix)
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 258
 *             self.type_prefix()
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1568, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1568, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1568, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1568, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1568, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":1569
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 258             # <<<<<<<<<<<<<<
 *             self.type_prefix()
 *         except RecognitionException as re:
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_258) < 0) __PYX_ERR(0, 1569, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":1570
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 258
 *             self.type_prefix()             # <<<<<<<<<<<<<<
 *         except RecognitionException as re:
 *             localctx.exception = re
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_type_prefix); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1570, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1570, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1570, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":1567
 *         localctx = ModelicaParser.Base_prefixContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 12, self.RULE_base_prefix)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 258
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":1571
 *             self.state = 258
 *             self.type_prefix()
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1571, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.base_prefix", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_3, &__pyx_t_7) < 0) __PYX_ERR(0, 1571, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_3);
        __pyx_v_re = __pyx_t_3;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":1572
 *             self.type_prefix()
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 1572, __pyx_L19_error)

          /* "pymola/generated/ModelicaParser.pyx":1573
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1573, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1573, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1573, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1573, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_11 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1573, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_11);
            if (__pyx_t_4) {
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_4); __pyx_t_4 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_11, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1573, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1574
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1574, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1574, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_11)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1574, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_11)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1574, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1574, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1574, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":1571
 *             self.state = 258
 *             self.type_prefix()
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L20;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L19_error:;
            __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_14);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __pyx_t_6 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_17);
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
            }
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
            __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
            goto __pyx_L8_except_error;
          }
          __pyx_L20:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":1567
 *         localctx = ModelicaParser.Base_prefixContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 12, self.RULE_base_prefix)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 258
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":1576
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1576, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1576, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1576, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_18, &__pyx_t_17);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_18);
      __Pyx_XGOTREF(__pyx_t_17);
      __pyx_t_12 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_20 = __pyx_filename;
      {
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1576, __pyx_L26_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1576, __pyx_L26_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1576, __pyx_L26_error)
        }
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_18, __pyx_t_17);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0;
      __pyx_lineno = __pyx_t_12; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_20;
      goto __pyx_L1_error;
      __pyx_L26_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_18, __pyx_t_17);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":1577
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Enum_listContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1563
 * 
 * 
 *     def base_prefix(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Base_prefixContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.base_prefix", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1581
 *     class Enum_listContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16Enum_listContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_16Enum_listContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_16Enum_listContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16Enum_listContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 1581, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 1581, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1581, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Enum_listContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_16Enum_listContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16Enum_listContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":1582
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 1582, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1582, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1582, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1582, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1582, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1583
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def enumeration_literal(self, i:int=None):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 1583, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1581
 *     class Enum_listContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Enum_listContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1585
 *             self.parser = parser
 * 
 *         def enumeration_literal(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Enumeration_literalContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16Enum_listContext_3enumeration_literal(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_16Enum_listContext_3enumeration_literal = {"enumeration_literal", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_16Enum_listContext_3enumeration_literal, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16Enum_listContext_3enumeration_literal(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enumeration_literal (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enumeration_literal") < 0)) __PYX_ERR(0, 1585, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enumeration_literal", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1585, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Enum_listContext.enumeration_literal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_16Enum_listContext_2enumeration_literal(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16Enum_listContext_2enumeration_literal(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("enumeration_literal", 0);

  /* "pymola/generated/ModelicaParser.pyx":1586
 * 
 *         def enumeration_literal(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.Enumeration_literalContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1587
 *         def enumeration_literal(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Enumeration_literalContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.Enumeration_literalContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1587, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1587, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_Enumeration_literalContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1587, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1587, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1587, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1587, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1587, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1587, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":1586
 * 
 *         def enumeration_literal(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.Enumeration_literalContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1589
 *                 return self.getTypedRuleContexts(ModelicaParser.Enumeration_literalContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.Enumeration_literalContext,i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1589, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1589, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Enumeration_literalContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1589, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1589, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1589, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1589, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1589, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":1585
 *             self.parser = parser
 * 
 *         def enumeration_literal(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Enumeration_literalContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Enum_listContext.enumeration_literal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1592
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_enum_list
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16Enum_listContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_16Enum_listContext_5getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_16Enum_listContext_5getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16Enum_listContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_16Enum_listContext_4getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16Enum_listContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":1593
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_enum_list             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_enum_list); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1592
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_enum_list
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Enum_listContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1595
 *             return ModelicaParser.RULE_enum_list
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterEnum_list" ):
 *                 listener.enterEnum_list(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16Enum_listContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_16Enum_listContext_7enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_16Enum_listContext_7enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16Enum_listContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 1595, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 1595, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1595, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Enum_listContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_16Enum_listContext_6enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16Enum_listContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":1596
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterEnum_list" ):             # <<<<<<<<<<<<<<
 *                 listener.enterEnum_list(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterEnum_list); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 1596, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1597
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterEnum_list" ):
 *                 listener.enterEnum_list(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterEnum_list); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1597, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1597, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1597, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1597, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1597, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1597, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":1596
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterEnum_list" ):             # <<<<<<<<<<<<<<
 *                 listener.enterEnum_list(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1595
 *             return ModelicaParser.RULE_enum_list
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterEnum_list" ):
 *                 listener.enterEnum_list(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Enum_listContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1599
 *                 listener.enterEnum_list(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitEnum_list" ):
 *                 listener.exitEnum_list(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16Enum_listContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_16Enum_listContext_9exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_16Enum_listContext_9exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16Enum_listContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 1599, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 1599, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1599, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Enum_listContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_16Enum_listContext_8exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16Enum_listContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":1600
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitEnum_list" ):             # <<<<<<<<<<<<<<
 *                 listener.exitEnum_list(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitEnum_list); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 1600, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1601
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitEnum_list" ):
 *                 listener.exitEnum_list(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitEnum_list); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1601, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1601, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1601, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1601, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1601, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1601, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":1600
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitEnum_list" ):             # <<<<<<<<<<<<<<
 *                 listener.exitEnum_list(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1599
 *                 listener.enterEnum_list(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitEnum_list" ):
 *                 listener.exitEnum_list(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Enum_listContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1603
 *                 listener.exitEnum_list(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitEnum_list" ):
 *                 return visitor.visitEnum_list(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16Enum_listContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_16Enum_listContext_11accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_16Enum_listContext_11accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16Enum_listContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 1603, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 1603, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1603, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Enum_listContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_16Enum_listContext_10accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16Enum_listContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":1604
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitEnum_list" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitEnum_list(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitEnum_list); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 1604, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1605
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitEnum_list" ):
 *                 return visitor.visitEnum_list(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitEnum_list); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1605, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1605, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1605, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1605, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1605, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1605, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":1604
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitEnum_list" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitEnum_list(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1607
 *                 return visitor.visitEnum_list(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1607, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1607, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1607, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1607, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1607, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1607, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":1603
 *                 listener.exitEnum_list(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitEnum_list" ):
 *                 return visitor.visitEnum_list(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Enum_listContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1612
 * 
 * 
 *     def enum_list(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Enum_listContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_17enum_list(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_17enum_list = {"enum_list", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_17enum_list, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_17enum_list(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enum_list (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_16enum_list(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_16enum_list(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  char const *__pyx_t_21;
  __Pyx_RefNannySetupContext("enum_list", 0);

  /* "pymola/generated/ModelicaParser.pyx":1614
 *     def enum_list(self):
 * 
 *         localctx = ModelicaParser.Enum_listContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 14, self.RULE_enum_list)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1614, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Enum_listContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1614, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1614, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1614, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1614, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1614, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1614, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1614, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1615
 * 
 *         localctx = ModelicaParser.Enum_listContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 14, self.RULE_enum_list)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_enum_list); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_14, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1615, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_14, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1615, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1615, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_14);
    __Pyx_GIVEREF(__pyx_int_14);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_14);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1615, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1616
 *         localctx = ModelicaParser.Enum_listContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 14, self.RULE_enum_list)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 1616, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1617
 *         self.enterRule(localctx, 14, self.RULE_enum_list)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 260
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":1618
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 260
 *             self.enumeration_literal()
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1618, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1618, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1618, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1618, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1618, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":1619
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 260             # <<<<<<<<<<<<<<
 *             self.enumeration_literal()
 *             self.state = 265
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_260) < 0) __PYX_ERR(0, 1619, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":1620
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 260
 *             self.enumeration_literal()             # <<<<<<<<<<<<<<
 *             self.state = 265
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enumeration_literal); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1620, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1620, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1620, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":1621
 *             self.state = 260
 *             self.enumeration_literal()
 *             self.state = 265             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_265) < 0) __PYX_ERR(0, 1621, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":1622
 *             self.enumeration_literal()
 *             self.state = 265
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             while _la==ModelicaParser.T__20:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1622, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1622, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1622, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1622, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1622, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1622, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1622, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":1623
 *             self.state = 265
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             while _la==ModelicaParser.T__20:
 *                 self.state = 261
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1623, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1623, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__832, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1623, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v__la = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":1624
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             while _la==ModelicaParser.T__20:             # <<<<<<<<<<<<<<
 *                 self.state = 261
 *                 self.match(ModelicaParser.T__20)
 */
        while (1) {
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1624, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__20); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1624, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1624, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1624, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (!__pyx_t_11) break;

          /* "pymola/generated/ModelicaParser.pyx":1625
 *             _la = self._input.LA(1)
 *             while _la==ModelicaParser.T__20:
 *                 self.state = 261             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__20)
 *                 self.state = 262
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_261) < 0) __PYX_ERR(0, 1625, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1626
 *             while _la==ModelicaParser.T__20:
 *                 self.state = 261
 *                 self.match(ModelicaParser.T__20)             # <<<<<<<<<<<<<<
 *                 self.state = 262
 *                 self.enumeration_literal()
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1626, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1626, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__20); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1626, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1626, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1626, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1626, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1626, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1626, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1627
 *                 self.state = 261
 *                 self.match(ModelicaParser.T__20)
 *                 self.state = 262             # <<<<<<<<<<<<<<
 *                 self.enumeration_literal()
 *                 self.state = 267
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_262) < 0) __PYX_ERR(0, 1627, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1628
 *                 self.match(ModelicaParser.T__20)
 *                 self.state = 262
 *                 self.enumeration_literal()             # <<<<<<<<<<<<<<
 *                 self.state = 267
 *                 self._errHandler.sync(self)
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enumeration_literal); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1628, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_4 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (__pyx_t_4) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1628, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1628, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1629
 *                 self.state = 262
 *                 self.enumeration_literal()
 *                 self.state = 267             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_267) < 0) __PYX_ERR(0, 1629, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1630
 *                 self.enumeration_literal()
 *                 self.state = 267
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 * 
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1630, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_sync); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1630, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1630, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1630, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1630, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1630, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1630, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1631
 *                 self.state = 267
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *         except RecognitionException as re:
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1631, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1631, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__833, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1631, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_1);
          __pyx_t_1 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":1617
 *         self.enterRule(localctx, 14, self.RULE_enum_list)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 260
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":1633
 *                 _la = self._input.LA(1)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1633, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.enum_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_4, &__pyx_t_3) < 0) __PYX_ERR(0, 1633, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __pyx_v_re = __pyx_t_4;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":1634
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 1634, __pyx_L21_error)

          /* "pymola/generated/ModelicaParser.pyx":1635
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1635, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1635, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1635, __pyx_L21_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1635, __pyx_L21_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          {
            __pyx_t_12 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1635, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_12);
            if (__pyx_t_2) {
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_2); __pyx_t_2 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_12, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1635, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1636
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1636, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1636, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_12);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_12, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1636, __pyx_L21_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1636, __pyx_L21_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1636, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_2, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1636, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":1633
 *                 _la = self._input.LA(1)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L22;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L21_error:;
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __pyx_t_6 = __pyx_lineno; __pyx_t_13 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
            }
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_ErrRestore(__pyx_t_15, __pyx_t_16, __pyx_t_17);
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_13; __pyx_filename = __pyx_t_14;
            goto __pyx_L8_except_error;
          }
          __pyx_L22:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":1617
 *         self.enterRule(localctx, 14, self.RULE_enum_list)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 260
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":1638
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1638, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1638, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1638, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_19, &__pyx_t_18);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_18);
      __pyx_t_13 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_21 = __pyx_filename;
      {
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1638, __pyx_L28_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1638, __pyx_L28_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1638, __pyx_L28_error)
        }
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __pyx_lineno = __pyx_t_13; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_21;
      goto __pyx_L1_error;
      __pyx_L28_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":1639
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Enumeration_literalContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1612
 * 
 * 
 *     def enum_list(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Enum_listContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.enum_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1643
 *     class Enumeration_literalContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 1643, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 1643, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1643, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Enumeration_literalContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":1644
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 1644, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1644, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1644, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1644, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1644, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1644, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1644, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1644, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1645
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def IDENT(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 1645, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1643
 *     class Enumeration_literalContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Enumeration_literalContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1647
 *             self.parser = parser
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_3IDENT(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_3IDENT = {"IDENT", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_3IDENT, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_3IDENT(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("IDENT (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_2IDENT(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_2IDENT(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("IDENT", 0);

  /* "pymola/generated/ModelicaParser.pyx":1648
 * 
 *         def IDENT(self):
 *             return self.getToken(ModelicaParser.IDENT, 0)             # <<<<<<<<<<<<<<
 * 
 *         def comment(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getToken); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1648, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1648, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1648, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1648, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1648, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1648, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1648, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1647
 *             self.parser = parser
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Enumeration_literalContext.IDENT", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1650
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_5comment(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_5comment = {"comment", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_5comment, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_5comment(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("comment (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_4comment(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_4comment(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("comment", 0);

  /* "pymola/generated/ModelicaParser.pyx":1651
 * 
 *         def comment(self):
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_CommentContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1651, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1651, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1651, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1651, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1650
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Enumeration_literalContext.comment", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1654
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_enumeration_literal
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_7getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_7getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_6getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":1655
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_enumeration_literal             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_enumeration_literal); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1654
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_enumeration_literal
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Enumeration_literalContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1657
 *             return ModelicaParser.RULE_enumeration_literal
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterEnumeration_literal" ):
 *                 listener.enterEnumeration_literal(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_9enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_9enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 1657, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 1657, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1657, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Enumeration_literalContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_8enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":1658
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterEnumeration_literal" ):             # <<<<<<<<<<<<<<
 *                 listener.enterEnumeration_literal(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterEnumeration_literal); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 1658, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1659
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterEnumeration_literal" ):
 *                 listener.enterEnumeration_literal(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterEnumeration_literal); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1659, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1659, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1659, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1659, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1659, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1659, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":1658
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterEnumeration_literal" ):             # <<<<<<<<<<<<<<
 *                 listener.enterEnumeration_literal(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1657
 *             return ModelicaParser.RULE_enumeration_literal
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterEnumeration_literal" ):
 *                 listener.enterEnumeration_literal(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Enumeration_literalContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1661
 *                 listener.enterEnumeration_literal(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitEnumeration_literal" ):
 *                 listener.exitEnumeration_literal(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_11exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_11exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 1661, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 1661, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1661, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Enumeration_literalContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_10exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":1662
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitEnumeration_literal" ):             # <<<<<<<<<<<<<<
 *                 listener.exitEnumeration_literal(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitEnumeration_literal); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 1662, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1663
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitEnumeration_literal" ):
 *                 listener.exitEnumeration_literal(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitEnumeration_literal); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1663, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1663, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1663, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1663, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1663, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1663, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":1662
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitEnumeration_literal" ):             # <<<<<<<<<<<<<<
 *                 listener.exitEnumeration_literal(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1661
 *                 listener.enterEnumeration_literal(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitEnumeration_literal" ):
 *                 listener.exitEnumeration_literal(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Enumeration_literalContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1665
 *                 listener.exitEnumeration_literal(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitEnumeration_literal" ):
 *                 return visitor.visitEnumeration_literal(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_13accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_13accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 1665, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 1665, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1665, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Enumeration_literalContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_12accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":1666
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitEnumeration_literal" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitEnumeration_literal(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitEnumeration_literal); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 1666, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1667
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitEnumeration_literal" ):
 *                 return visitor.visitEnumeration_literal(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitEnumeration_literal); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1667, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1667, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1667, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1667, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1667, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1667, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":1666
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitEnumeration_literal" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitEnumeration_literal(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1669
 *                 return visitor.visitEnumeration_literal(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1669, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1669, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1669, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1669, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1669, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1669, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":1665
 *                 listener.exitEnumeration_literal(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitEnumeration_literal" ):
 *                 return visitor.visitEnumeration_literal(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Enumeration_literalContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1674
 * 
 * 
 *     def enumeration_literal(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Enumeration_literalContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_19enumeration_literal(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_19enumeration_literal = {"enumeration_literal", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_19enumeration_literal, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_19enumeration_literal(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enumeration_literal (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_18enumeration_literal(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_18enumeration_literal(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  char const *__pyx_t_20;
  __Pyx_RefNannySetupContext("enumeration_literal", 0);

  /* "pymola/generated/ModelicaParser.pyx":1676
 *     def enumeration_literal(self):
 * 
 *         localctx = ModelicaParser.Enumeration_literalContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 16, self.RULE_enumeration_literal)
 *         try:
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1676, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Enumeration_literalContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1676, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1676, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1676, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1676, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1676, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1676, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1676, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1677
 * 
 *         localctx = ModelicaParser.Enumeration_literalContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 16, self.RULE_enumeration_literal)             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1677, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_enumeration_literal); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1677, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_16, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1677, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_16, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1677, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1677, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_16);
    __Pyx_GIVEREF(__pyx_int_16);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_16);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1677, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1678
 *         localctx = ModelicaParser.Enumeration_literalContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 16, self.RULE_enumeration_literal)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 268
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":1679
 *         self.enterRule(localctx, 16, self.RULE_enumeration_literal)
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 268
 *             self.match(ModelicaParser.IDENT)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1679, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1679, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1679, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1679, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1679, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":1680
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 268             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.IDENT)
 *             self.state = 269
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_268) < 0) __PYX_ERR(0, 1680, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":1681
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 268
 *             self.match(ModelicaParser.IDENT)             # <<<<<<<<<<<<<<
 *             self.state = 269
 *             self.comment()
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1681, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1681, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1681, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1681, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1681, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1681, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1681, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1681, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":1682
 *             self.state = 268
 *             self.match(ModelicaParser.IDENT)
 *             self.state = 269             # <<<<<<<<<<<<<<
 *             self.comment()
 *         except RecognitionException as re:
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_269) < 0) __PYX_ERR(0, 1682, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":1683
 *             self.match(ModelicaParser.IDENT)
 *             self.state = 269
 *             self.comment()             # <<<<<<<<<<<<<<
 *         except RecognitionException as re:
 *             localctx.exception = re
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_comment); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1683, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1683, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1683, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":1678
 *         localctx = ModelicaParser.Enumeration_literalContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 16, self.RULE_enumeration_literal)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 268
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":1684
 *             self.state = 269
 *             self.comment()
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1684, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.enumeration_literal", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_3, &__pyx_t_4) < 0) __PYX_ERR(0, 1684, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __pyx_v_re = __pyx_t_3;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":1685
 *             self.comment()
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 1685, __pyx_L19_error)

          /* "pymola/generated/ModelicaParser.pyx":1686
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1686, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1686, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1686, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1686, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_11 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1686, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_11);
            if (__pyx_t_7) {
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_7); __pyx_t_7 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_11, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1686, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1687
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1687, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1687, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_11)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1687, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_11)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1687, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1687, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_7);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1687, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":1684
 *             self.state = 269
 *             self.comment()
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L20;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L19_error:;
            __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_14);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __pyx_t_6 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_17);
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
            }
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
            __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
            goto __pyx_L8_except_error;
          }
          __pyx_L20:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":1678
 *         localctx = ModelicaParser.Enumeration_literalContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 16, self.RULE_enumeration_literal)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 268
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":1689
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1689, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1689, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1689, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_18, &__pyx_t_17);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_18);
      __Pyx_XGOTREF(__pyx_t_17);
      __pyx_t_12 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_20 = __pyx_filename;
      {
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1689, __pyx_L26_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1689, __pyx_L26_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1689, __pyx_L26_error)
        }
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_18, __pyx_t_17);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0;
      __pyx_lineno = __pyx_t_12; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_20;
      goto __pyx_L1_error;
      __pyx_L26_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_18, __pyx_t_17);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":1690
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class CompositionContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1674
 * 
 * 
 *     def enumeration_literal(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Enumeration_literalContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.enumeration_literal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1694
 *     class CompositionContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18CompositionContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18CompositionContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18CompositionContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18CompositionContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 1694, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 1694, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1694, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.CompositionContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18CompositionContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18CompositionContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":1695
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 *             self.epriv = None # Element_listContext
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 1695, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1695, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1695, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1695, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1695, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1695, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1695, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1695, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1696
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 *             self.epriv = None # Element_listContext
 *             self.epub = None # Element_listContext
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 1696, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1697
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 *             self.epriv = None # Element_listContext             # <<<<<<<<<<<<<<
 *             self.epub = None # Element_listContext
 *             self.epro = None # Element_listContext
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_epriv, Py_None) < 0) __PYX_ERR(0, 1697, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1698
 *             self.parser = parser
 *             self.epriv = None # Element_listContext
 *             self.epub = None # Element_listContext             # <<<<<<<<<<<<<<
 *             self.epro = None # Element_listContext
 *             self.ext_annotation = None # AnnotationContext
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_epub, Py_None) < 0) __PYX_ERR(0, 1698, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1699
 *             self.epriv = None # Element_listContext
 *             self.epub = None # Element_listContext
 *             self.epro = None # Element_listContext             # <<<<<<<<<<<<<<
 *             self.ext_annotation = None # AnnotationContext
 *             self.comp_annotation = None # AnnotationContext
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_epro, Py_None) < 0) __PYX_ERR(0, 1699, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1700
 *             self.epub = None # Element_listContext
 *             self.epro = None # Element_listContext
 *             self.ext_annotation = None # AnnotationContext             # <<<<<<<<<<<<<<
 *             self.comp_annotation = None # AnnotationContext
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_ext_annotation, Py_None) < 0) __PYX_ERR(0, 1700, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1701
 *             self.epro = None # Element_listContext
 *             self.ext_annotation = None # AnnotationContext
 *             self.comp_annotation = None # AnnotationContext             # <<<<<<<<<<<<<<
 * 
 *         def element_list(self, i:int=None):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_comp_annotation, Py_None) < 0) __PYX_ERR(0, 1701, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1694
 *     class CompositionContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.CompositionContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1703
 *             self.comp_annotation = None # AnnotationContext
 * 
 *         def element_list(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Element_listContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18CompositionContext_3element_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18CompositionContext_3element_list = {"element_list", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18CompositionContext_3element_list, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18CompositionContext_3element_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("element_list (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "element_list") < 0)) __PYX_ERR(0, 1703, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("element_list", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1703, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.CompositionContext.element_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18CompositionContext_2element_list(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18CompositionContext_2element_list(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("element_list", 0);

  /* "pymola/generated/ModelicaParser.pyx":1704
 * 
 *         def element_list(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.Element_listContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1705
 *         def element_list(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Element_listContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.Element_listContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1705, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1705, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_Element_listContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1705, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1705, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1705, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1705, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1705, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1705, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":1704
 * 
 *         def element_list(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.Element_listContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1707
 *                 return self.getTypedRuleContexts(ModelicaParser.Element_listContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.Element_listContext,i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1707, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1707, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Element_listContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1707, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1707, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1707, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1707, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1707, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":1703
 *             self.comp_annotation = None # AnnotationContext
 * 
 *         def element_list(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Element_listContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.CompositionContext.element_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1710
 * 
 * 
 *         def equation_section(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Equation_sectionContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18CompositionContext_5equation_section(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18CompositionContext_5equation_section = {"equation_section", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18CompositionContext_5equation_section, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18CompositionContext_5equation_section(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("equation_section (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "equation_section") < 0)) __PYX_ERR(0, 1710, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("equation_section", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1710, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.CompositionContext.equation_section", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18CompositionContext_4equation_section(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18CompositionContext_4equation_section(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("equation_section", 0);

  /* "pymola/generated/ModelicaParser.pyx":1711
 * 
 *         def equation_section(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.Equation_sectionContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1712
 *         def equation_section(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Equation_sectionContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.Equation_sectionContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1712, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1712, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_Equation_sectionContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1712, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1712, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1712, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1712, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1712, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1712, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":1711
 * 
 *         def equation_section(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.Equation_sectionContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1714
 *                 return self.getTypedRuleContexts(ModelicaParser.Equation_sectionContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.Equation_sectionContext,i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1714, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1714, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Equation_sectionContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1714, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1714, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1714, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1714, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1714, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":1710
 * 
 * 
 *         def equation_section(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Equation_sectionContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.CompositionContext.equation_section", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1717
 * 
 * 
 *         def algorithm_section(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Algorithm_sectionContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18CompositionContext_7algorithm_section(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18CompositionContext_7algorithm_section = {"algorithm_section", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18CompositionContext_7algorithm_section, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18CompositionContext_7algorithm_section(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("algorithm_section (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "algorithm_section") < 0)) __PYX_ERR(0, 1717, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("algorithm_section", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1717, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.CompositionContext.algorithm_section", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18CompositionContext_6algorithm_section(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18CompositionContext_6algorithm_section(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("algorithm_section", 0);

  /* "pymola/generated/ModelicaParser.pyx":1718
 * 
 *         def algorithm_section(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.Algorithm_sectionContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1719
 *         def algorithm_section(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Algorithm_sectionContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.Algorithm_sectionContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1719, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1719, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_Algorithm_sectionContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1719, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1719, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1719, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1719, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1719, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1719, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":1718
 * 
 *         def algorithm_section(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.Algorithm_sectionContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1721
 *                 return self.getTypedRuleContexts(ModelicaParser.Algorithm_sectionContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.Algorithm_sectionContext,i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1721, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1721, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Algorithm_sectionContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1721, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1721, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1721, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1721, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1721, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":1717
 * 
 * 
 *         def algorithm_section(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Algorithm_sectionContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.CompositionContext.algorithm_section", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1724
 * 
 * 
 *         def annotation(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.AnnotationContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18CompositionContext_9annotation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18CompositionContext_9annotation = {"annotation", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18CompositionContext_9annotation, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18CompositionContext_9annotation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("annotation (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "annotation") < 0)) __PYX_ERR(0, 1724, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("annotation", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1724, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.CompositionContext.annotation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18CompositionContext_8annotation(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18CompositionContext_8annotation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("annotation", 0);

  /* "pymola/generated/ModelicaParser.pyx":1725
 * 
 *         def annotation(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.AnnotationContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1726
 *         def annotation(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.AnnotationContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.AnnotationContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1726, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1726, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_AnnotationContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1726, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1726, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1726, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1726, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1726, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1726, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":1725
 * 
 *         def annotation(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.AnnotationContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1728
 *                 return self.getTypedRuleContexts(ModelicaParser.AnnotationContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.AnnotationContext,i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1728, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1728, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_AnnotationContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1728, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1728, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1728, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1728, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1728, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":1724
 * 
 * 
 *         def annotation(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.AnnotationContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.CompositionContext.annotation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1731
 * 
 * 
 *         def language_specification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Language_specificationContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18CompositionContext_11language_specification(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18CompositionContext_11language_specification = {"language_specification", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18CompositionContext_11language_specification, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18CompositionContext_11language_specification(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("language_specification (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18CompositionContext_10language_specification(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18CompositionContext_10language_specification(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("language_specification", 0);

  /* "pymola/generated/ModelicaParser.pyx":1732
 * 
 *         def language_specification(self):
 *             return self.getTypedRuleContext(ModelicaParser.Language_specificationContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1732, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1732, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Language_specificationContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1732, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1732, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1732, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1732, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1732, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1731
 * 
 * 
 *         def language_specification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Language_specificationContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.CompositionContext.language_specification", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1735
 * 
 * 
 *         def external_function_call(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.External_function_callContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18CompositionContext_13external_function_call(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18CompositionContext_13external_function_call = {"external_function_call", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18CompositionContext_13external_function_call, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18CompositionContext_13external_function_call(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("external_function_call (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18CompositionContext_12external_function_call(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18CompositionContext_12external_function_call(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("external_function_call", 0);

  /* "pymola/generated/ModelicaParser.pyx":1736
 * 
 *         def external_function_call(self):
 *             return self.getTypedRuleContext(ModelicaParser.External_function_callContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_External_function_callContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1736, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1736, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1736, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1736, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1735
 * 
 * 
 *         def external_function_call(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.External_function_callContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.CompositionContext.external_function_call", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1739
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_composition
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18CompositionContext_15getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18CompositionContext_15getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18CompositionContext_15getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18CompositionContext_15getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18CompositionContext_14getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18CompositionContext_14getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":1740
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_composition             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1740, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_composition); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1740, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1739
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_composition
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.CompositionContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1742
 *             return ModelicaParser.RULE_composition
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterComposition" ):
 *                 listener.enterComposition(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18CompositionContext_17enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18CompositionContext_17enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18CompositionContext_17enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18CompositionContext_17enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 1742, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 1742, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1742, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.CompositionContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18CompositionContext_16enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18CompositionContext_16enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":1743
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterComposition" ):             # <<<<<<<<<<<<<<
 *                 listener.enterComposition(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterComposition); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 1743, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1744
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterComposition" ):
 *                 listener.enterComposition(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterComposition); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1744, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1744, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1744, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1744, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1744, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1744, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":1743
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterComposition" ):             # <<<<<<<<<<<<<<
 *                 listener.enterComposition(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1742
 *             return ModelicaParser.RULE_composition
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterComposition" ):
 *                 listener.enterComposition(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.CompositionContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1746
 *                 listener.enterComposition(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitComposition" ):
 *                 listener.exitComposition(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18CompositionContext_19exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18CompositionContext_19exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18CompositionContext_19exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18CompositionContext_19exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 1746, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 1746, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1746, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.CompositionContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18CompositionContext_18exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18CompositionContext_18exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":1747
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitComposition" ):             # <<<<<<<<<<<<<<
 *                 listener.exitComposition(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitComposition); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 1747, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1748
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitComposition" ):
 *                 listener.exitComposition(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitComposition); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1748, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1748, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1748, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1748, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1748, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1748, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":1747
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitComposition" ):             # <<<<<<<<<<<<<<
 *                 listener.exitComposition(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1746
 *                 listener.enterComposition(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitComposition" ):
 *                 listener.exitComposition(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.CompositionContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1750
 *                 listener.exitComposition(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitComposition" ):
 *                 return visitor.visitComposition(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18CompositionContext_21accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18CompositionContext_21accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18CompositionContext_21accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18CompositionContext_21accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 1750, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 1750, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1750, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.CompositionContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18CompositionContext_20accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18CompositionContext_20accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":1751
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitComposition" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitComposition(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitComposition); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 1751, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1752
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitComposition" ):
 *                 return visitor.visitComposition(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitComposition); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1752, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1752, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1752, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1752, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1752, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1752, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":1751
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitComposition" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitComposition(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1754
 *                 return visitor.visitComposition(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1754, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1754, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1754, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1754, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1754, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1754, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":1750
 *                 listener.exitComposition(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitComposition" ):
 *                 return visitor.visitComposition(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.CompositionContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1759
 * 
 * 
 *     def composition(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.CompositionContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_21composition(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_21composition = {"composition", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_21composition, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_21composition(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("composition (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_20composition(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_20composition(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_la_ = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  char const *__pyx_t_22;
  __Pyx_RefNannySetupContext("composition", 0);

  /* "pymola/generated/ModelicaParser.pyx":1761
 *     def composition(self):
 * 
 *         localctx = ModelicaParser.CompositionContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 18, self.RULE_composition)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1761, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_CompositionContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1761, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1761, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1761, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1761, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1761, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1761, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1761, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1762
 * 
 *         localctx = ModelicaParser.CompositionContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 18, self.RULE_composition)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1762, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_composition); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1762, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_18, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1762, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_18, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1762, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1762, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_18);
    __Pyx_GIVEREF(__pyx_int_18);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_18);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1762, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1763
 *         localctx = ModelicaParser.CompositionContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 18, self.RULE_composition)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 1763, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1764
 *         self.enterRule(localctx, 18, self.RULE_composition)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 271
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":1765
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 271
 *             localctx.epriv = self.element_list()
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1765, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1765, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1765, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1765, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1765, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":1766
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 271             # <<<<<<<<<<<<<<
 *             localctx.epriv = self.element_list()
 *             self.state = 280
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_271) < 0) __PYX_ERR(0, 1766, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":1767
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 271
 *             localctx.epriv = self.element_list()             # <<<<<<<<<<<<<<
 *             self.state = 280
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_element_list); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1767, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1767, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1767, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_epriv, __pyx_t_1) < 0) __PYX_ERR(0, 1767, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":1768
 *             self.state = 271
 *             localctx.epriv = self.element_list()
 *             self.state = 280             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_280) < 0) __PYX_ERR(0, 1768, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":1769
 *             localctx.epriv = self.element_list()
 *             self.state = 280
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__22) | (1 << ModelicaParser.T__23) | (1 << ModelicaParser.T__41) | (1 << ModelicaParser.T__42))) != 0) or _la==ModelicaParser.INITIAL:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1769, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1769, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1769, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1769, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1769, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1769, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1769, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":1770
 *             self.state = 280
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__22) | (1 << ModelicaParser.T__23) | (1 << ModelicaParser.T__41) | (1 << ModelicaParser.T__42))) != 0) or _la==ModelicaParser.INITIAL:
 *                 self.state = 278
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1770, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1770, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__834, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1770, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v__la = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":1771
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__22) | (1 << ModelicaParser.T__23) | (1 << ModelicaParser.T__41) | (1 << ModelicaParser.T__42))) != 0) or _la==ModelicaParser.INITIAL:             # <<<<<<<<<<<<<<
 *                 self.state = 278
 *                 self._errHandler.sync(self)
 */
        while (1) {
          __pyx_t_1 = __Pyx_PyInt_From_long((~0x3f)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1771, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_7 = PyNumber_And(__pyx_v__la, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1771, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_t_7, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1771, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 1771, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (!__pyx_t_12) {
            goto __pyx_L17_next_or;
          } else {
          }
          __pyx_t_1 = PyNumber_Lshift(__pyx_int_1, __pyx_v__la); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1771, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1771, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__22); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1771, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1771, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1771, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__23); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1771, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1771, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyNumber_Or(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1771, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1771, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__41); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1771, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1771, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Or(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1771, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1771, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__42); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1771, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1771, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyNumber_Or(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1771, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_And(__pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1771, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_int_0, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1771, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 1771, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (!__pyx_t_12) {
          } else {
            __pyx_t_11 = __pyx_t_12;
            goto __pyx_L16_bool_binop_done;
          }
          __pyx_L17_next_or:;
          __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1771, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_INITIAL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1771, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyObject_RichCompare(__pyx_v__la, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1771, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 1771, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_11 = __pyx_t_12;
          __pyx_L16_bool_binop_done:;
          if (!__pyx_t_11) break;

          /* "pymola/generated/ModelicaParser.pyx":1772
 *             _la = self._input.LA(1)
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__22) | (1 << ModelicaParser.T__23) | (1 << ModelicaParser.T__41) | (1 << ModelicaParser.T__42))) != 0) or _la==ModelicaParser.INITIAL:
 *                 self.state = 278             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 la_ = self._interp.adaptivePredict(self._input,19,self._ctx)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_278) < 0) __PYX_ERR(0, 1772, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1773
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__22) | (1 << ModelicaParser.T__23) | (1 << ModelicaParser.T__41) | (1 << ModelicaParser.T__42))) != 0) or _la==ModelicaParser.INITIAL:
 *                 self.state = 278
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 la_ = self._interp.adaptivePredict(self._input,19,self._ctx)
 *                 if la_ == 1:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1773, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_sync); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1773, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_1, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1773, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_1)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1773, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_3);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1773, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_3);
            } else
            #endif
            {
              __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1773, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self);
              __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1773, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1774
 *                 self.state = 278
 *                 self._errHandler.sync(self)
 *                 la_ = self._interp.adaptivePredict(self._input,19,self._ctx)             # <<<<<<<<<<<<<<
 *                 if la_ == 1:
 *                     self.state = 272
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_interp); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1774, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_adaptivePredict); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1774, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1774, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1774, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_4 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_t_1, __pyx_int_19, __pyx_t_2};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1774, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_t_1, __pyx_int_19, __pyx_t_2};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1774, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_5 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1774, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            if (__pyx_t_4) {
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
            }
            __Pyx_GIVEREF(__pyx_t_1);
            PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_t_1);
            __Pyx_INCREF(__pyx_int_19);
            __Pyx_GIVEREF(__pyx_int_19);
            PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_int_19);
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_6, __pyx_t_2);
            __pyx_t_1 = 0;
            __pyx_t_2 = 0;
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1774, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_XDECREF_SET(__pyx_v_la_, __pyx_t_3);
          __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1775
 *                 self._errHandler.sync(self)
 *                 la_ = self._interp.adaptivePredict(self._input,19,self._ctx)
 *                 if la_ == 1:             # <<<<<<<<<<<<<<
 *                     self.state = 272
 *                     self.match(ModelicaParser.T__22)
 */
          __pyx_t_3 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1775, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1775, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (__pyx_t_11) {

            /* "pymola/generated/ModelicaParser.pyx":1776
 *                 la_ = self._interp.adaptivePredict(self._input,19,self._ctx)
 *                 if la_ == 1:
 *                     self.state = 272             # <<<<<<<<<<<<<<
 *                     self.match(ModelicaParser.T__22)
 *                     self.state = 273
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_272) < 0) __PYX_ERR(0, 1776, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":1777
 *                 if la_ == 1:
 *                     self.state = 272
 *                     self.match(ModelicaParser.T__22)             # <<<<<<<<<<<<<<
 *                     self.state = 273
 *                     localctx.epub = self.element_list()
 */
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1777, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1777, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_T__22); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1777, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_5 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
              if (likely(__pyx_t_5)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
                __Pyx_INCREF(__pyx_t_5);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_7, function);
              }
            }
            if (!__pyx_t_5) {
              __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1777, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_3);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_7)) {
                PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_2};
                __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1777, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
                PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_2};
                __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1777, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              } else
              #endif
              {
                __pyx_t_1 = PyTuple_New(1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1777, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5); __pyx_t_5 = NULL;
                __Pyx_GIVEREF(__pyx_t_2);
                PyTuple_SET_ITEM(__pyx_t_1, 0+1, __pyx_t_2);
                __pyx_t_2 = 0;
                __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_1, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1777, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

            /* "pymola/generated/ModelicaParser.pyx":1778
 *                     self.state = 272
 *                     self.match(ModelicaParser.T__22)
 *                     self.state = 273             # <<<<<<<<<<<<<<
 *                     localctx.epub = self.element_list()
 *                     pass
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_273) < 0) __PYX_ERR(0, 1778, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":1779
 *                     self.match(ModelicaParser.T__22)
 *                     self.state = 273
 *                     localctx.epub = self.element_list()             # <<<<<<<<<<<<<<
 *                     pass
 * 
 */
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_element_list); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1779, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_1 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
              __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
              if (likely(__pyx_t_1)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
                __Pyx_INCREF(__pyx_t_1);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_7, function);
              }
            }
            if (__pyx_t_1) {
              __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1779, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            } else {
              __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1779, __pyx_L6_error)
            }
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_epub, __pyx_t_3) < 0) __PYX_ERR(0, 1779, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

            /* "pymola/generated/ModelicaParser.pyx":1775
 *                 self._errHandler.sync(self)
 *                 la_ = self._interp.adaptivePredict(self._input,19,self._ctx)
 *                 if la_ == 1:             # <<<<<<<<<<<<<<
 *                     self.state = 272
 *                     self.match(ModelicaParser.T__22)
 */
            goto __pyx_L19;
          }

          /* "pymola/generated/ModelicaParser.pyx":1782
 *                     pass
 * 
 *                 elif la_ == 2:             # <<<<<<<<<<<<<<
 *                     self.state = 274
 *                     self.match(ModelicaParser.T__23)
 */
          __pyx_t_3 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1782, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1782, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (__pyx_t_11) {

            /* "pymola/generated/ModelicaParser.pyx":1783
 * 
 *                 elif la_ == 2:
 *                     self.state = 274             # <<<<<<<<<<<<<<
 *                     self.match(ModelicaParser.T__23)
 *                     self.state = 275
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_274) < 0) __PYX_ERR(0, 1783, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":1784
 *                 elif la_ == 2:
 *                     self.state = 274
 *                     self.match(ModelicaParser.T__23)             # <<<<<<<<<<<<<<
 *                     self.state = 275
 *                     localctx.epro = self.element_list()
 */
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1784, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1784, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__23); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1784, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
              __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
              if (likely(__pyx_t_1)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
                __Pyx_INCREF(__pyx_t_1);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_7, function);
              }
            }
            if (!__pyx_t_1) {
              __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1784, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_3);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_7)) {
                PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_t_2};
                __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1784, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
                PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_t_2};
                __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1784, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              } else
              #endif
              {
                __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1784, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_5);
                __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1); __pyx_t_1 = NULL;
                __Pyx_GIVEREF(__pyx_t_2);
                PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_t_2);
                __pyx_t_2 = 0;
                __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1784, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

            /* "pymola/generated/ModelicaParser.pyx":1785
 *                     self.state = 274
 *                     self.match(ModelicaParser.T__23)
 *                     self.state = 275             # <<<<<<<<<<<<<<
 *                     localctx.epro = self.element_list()
 *                     pass
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_275) < 0) __PYX_ERR(0, 1785, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":1786
 *                     self.match(ModelicaParser.T__23)
 *                     self.state = 275
 *                     localctx.epro = self.element_list()             # <<<<<<<<<<<<<<
 *                     pass
 * 
 */
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_element_list); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1786, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_5 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
              if (likely(__pyx_t_5)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
                __Pyx_INCREF(__pyx_t_5);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_7, function);
              }
            }
            if (__pyx_t_5) {
              __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1786, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else {
              __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1786, __pyx_L6_error)
            }
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_epro, __pyx_t_3) < 0) __PYX_ERR(0, 1786, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

            /* "pymola/generated/ModelicaParser.pyx":1782
 *                     pass
 * 
 *                 elif la_ == 2:             # <<<<<<<<<<<<<<
 *                     self.state = 274
 *                     self.match(ModelicaParser.T__23)
 */
            goto __pyx_L19;
          }

          /* "pymola/generated/ModelicaParser.pyx":1789
 *                     pass
 * 
 *                 elif la_ == 3:             # <<<<<<<<<<<<<<
 *                     self.state = 276
 *                     self.equation_section()
 */
          __pyx_t_3 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_3, 3, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1789, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1789, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (__pyx_t_11) {

            /* "pymola/generated/ModelicaParser.pyx":1790
 * 
 *                 elif la_ == 3:
 *                     self.state = 276             # <<<<<<<<<<<<<<
 *                     self.equation_section()
 *                     pass
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_276) < 0) __PYX_ERR(0, 1790, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":1791
 *                 elif la_ == 3:
 *                     self.state = 276
 *                     self.equation_section()             # <<<<<<<<<<<<<<
 *                     pass
 * 
 */
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_equation_section); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1791, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_5 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
              if (likely(__pyx_t_5)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
                __Pyx_INCREF(__pyx_t_5);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_7, function);
              }
            }
            if (__pyx_t_5) {
              __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1791, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else {
              __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1791, __pyx_L6_error)
            }
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

            /* "pymola/generated/ModelicaParser.pyx":1789
 *                     pass
 * 
 *                 elif la_ == 3:             # <<<<<<<<<<<<<<
 *                     self.state = 276
 *                     self.equation_section()
 */
            goto __pyx_L19;
          }

          /* "pymola/generated/ModelicaParser.pyx":1794
 *                     pass
 * 
 *                 elif la_ == 4:             # <<<<<<<<<<<<<<
 *                     self.state = 277
 *                     self.algorithm_section()
 */
          __pyx_t_3 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_4, 4, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1794, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1794, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (__pyx_t_11) {

            /* "pymola/generated/ModelicaParser.pyx":1795
 * 
 *                 elif la_ == 4:
 *                     self.state = 277             # <<<<<<<<<<<<<<
 *                     self.algorithm_section()
 *                     pass
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_277) < 0) __PYX_ERR(0, 1795, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":1796
 *                 elif la_ == 4:
 *                     self.state = 277
 *                     self.algorithm_section()             # <<<<<<<<<<<<<<
 *                     pass
 * 
 */
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_algorithm_section); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1796, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_5 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
              if (likely(__pyx_t_5)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
                __Pyx_INCREF(__pyx_t_5);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_7, function);
              }
            }
            if (__pyx_t_5) {
              __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1796, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else {
              __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1796, __pyx_L6_error)
            }
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

            /* "pymola/generated/ModelicaParser.pyx":1794
 *                     pass
 * 
 *                 elif la_ == 4:             # <<<<<<<<<<<<<<
 *                     self.state = 277
 *                     self.algorithm_section()
 */
          }
          __pyx_L19:;

          /* "pymola/generated/ModelicaParser.pyx":1800
 * 
 * 
 *                 self.state = 282             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_282) < 0) __PYX_ERR(0, 1800, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1801
 * 
 *                 self.state = 282
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 * 
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1801, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_sync); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1801, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1801, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_5)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
              __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1801, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_3);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
              __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1801, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_3);
            } else
            #endif
            {
              __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1801, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_7); __pyx_t_7 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
              __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1801, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1802
 *                 self.state = 282
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *             self.state = 294
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1802, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_LA); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1802, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_tuple__835, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1802, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_3);
          __pyx_t_3 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":1804
 *                 _la = self._input.LA(1)
 * 
 *             self.state = 294             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_294) < 0) __PYX_ERR(0, 1804, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":1805
 * 
 *             self.state = 294
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__24:
 */
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1805, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1805, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_5 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        if (!__pyx_t_5) {
          __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1805, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1805, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1805, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1805, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1805, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "pymola/generated/ModelicaParser.pyx":1806
 *             self.state = 294
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.T__24:
 *                 self.state = 283
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1806, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_LA); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1806, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__836, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1806, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_3);
        __pyx_t_3 = 0;

        /* "pymola/generated/ModelicaParser.pyx":1807
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__24:             # <<<<<<<<<<<<<<
 *                 self.state = 283
 *                 self.match(ModelicaParser.T__24)
 */
        __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1807, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__24); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1807, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyObject_RichCompare(__pyx_v__la, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1807, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1807, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":1808
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__24:
 *                 self.state = 283             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__24)
 *                 self.state = 285
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_283) < 0) __PYX_ERR(0, 1808, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1809
 *             if _la==ModelicaParser.T__24:
 *                 self.state = 283
 *                 self.match(ModelicaParser.T__24)             # <<<<<<<<<<<<<<
 *                 self.state = 285
 *                 self._errHandler.sync(self)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1809, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1809, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__24); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1809, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1809, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_5};
              __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1809, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_5};
              __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1809, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            {
              __pyx_t_1 = PyTuple_New(1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1809, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_7); __pyx_t_7 = NULL;
              __Pyx_GIVEREF(__pyx_t_5);
              PyTuple_SET_ITEM(__pyx_t_1, 0+1, __pyx_t_5);
              __pyx_t_5 = 0;
              __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1809, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1810
 *                 self.state = 283
 *                 self.match(ModelicaParser.T__24)
 *                 self.state = 285             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_285) < 0) __PYX_ERR(0, 1810, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1811
 *                 self.match(ModelicaParser.T__24)
 *                 self.state = 285
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.STRING:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1811, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_sync); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1811, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_1, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1811, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_1)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1811, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_3);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1811, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_3);
            } else
            #endif
            {
              __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1811, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
              __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1811, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1812
 *                 self.state = 285
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if _la==ModelicaParser.STRING:
 *                     self.state = 284
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1812, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_LA); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1812, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__837, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1812, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_3);
          __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1813
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.STRING:             # <<<<<<<<<<<<<<
 *                     self.state = 284
 *                     self.language_specification()
 */
          __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1813, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_STRING); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1813, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyObject_RichCompare(__pyx_v__la, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1813, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1813, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (__pyx_t_11) {

            /* "pymola/generated/ModelicaParser.pyx":1814
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.STRING:
 *                     self.state = 284             # <<<<<<<<<<<<<<
 *                     self.language_specification()
 * 
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_284) < 0) __PYX_ERR(0, 1814, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":1815
 *                 if _la==ModelicaParser.STRING:
 *                     self.state = 284
 *                     self.language_specification()             # <<<<<<<<<<<<<<
 * 
 * 
 */
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_language_specification); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1815, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_5 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
              if (likely(__pyx_t_5)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
                __Pyx_INCREF(__pyx_t_5);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_1, function);
              }
            }
            if (__pyx_t_5) {
              __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1815, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else {
              __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1815, __pyx_L6_error)
            }
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

            /* "pymola/generated/ModelicaParser.pyx":1813
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.STRING:             # <<<<<<<<<<<<<<
 *                     self.state = 284
 *                     self.language_specification()
 */
          }

          /* "pymola/generated/ModelicaParser.pyx":1818
 * 
 * 
 *                 self.state = 288             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_288) < 0) __PYX_ERR(0, 1818, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1819
 * 
 *                 self.state = 288
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.IDENT:
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1819, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_sync); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1819, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
            }
          }
          if (!__pyx_t_1) {
            __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1819, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_5)) {
              PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_v_self};
              __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1819, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_GOTREF(__pyx_t_3);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
              PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_v_self};
              __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1819, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_GOTREF(__pyx_t_3);
            } else
            #endif
            {
              __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1819, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1); __pyx_t_1 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
              __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1819, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1820
 *                 self.state = 288
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if _la==ModelicaParser.IDENT:
 *                     self.state = 287
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1820, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_LA); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1820, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_tuple__838, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1820, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_3);
          __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1821
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.IDENT:             # <<<<<<<<<<<<<<
 *                     self.state = 287
 *                     self.external_function_call()
 */
          __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1821, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1821, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyObject_RichCompare(__pyx_v__la, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1821, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1821, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (__pyx_t_11) {

            /* "pymola/generated/ModelicaParser.pyx":1822
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.IDENT:
 *                     self.state = 287             # <<<<<<<<<<<<<<
 *                     self.external_function_call()
 * 
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_287) < 0) __PYX_ERR(0, 1822, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":1823
 *                 if _la==ModelicaParser.IDENT:
 *                     self.state = 287
 *                     self.external_function_call()             # <<<<<<<<<<<<<<
 * 
 * 
 */
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_external_function_call); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1823, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_2 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
              __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
              if (likely(__pyx_t_2)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                __Pyx_INCREF(__pyx_t_2);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_5, function);
              }
            }
            if (__pyx_t_2) {
              __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1823, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            } else {
              __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1823, __pyx_L6_error)
            }
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

            /* "pymola/generated/ModelicaParser.pyx":1821
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.IDENT:             # <<<<<<<<<<<<<<
 *                     self.state = 287
 *                     self.external_function_call()
 */
          }

          /* "pymola/generated/ModelicaParser.pyx":1826
 * 
 * 
 *                 self.state = 291             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_291) < 0) __PYX_ERR(0, 1826, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1827
 * 
 *                 self.state = 291
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__77:
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1827, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1827, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_5) {
            __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1827, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
              __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1827, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_3);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
              __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1827, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_3);
            } else
            #endif
            {
              __pyx_t_1 = PyTuple_New(1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1827, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5); __pyx_t_5 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_1, 0+1, __pyx_v_self);
              __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1827, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1828
 *                 self.state = 291
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if _la==ModelicaParser.T__77:
 *                     self.state = 290
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1828, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_LA); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1828, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__839, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1828, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_3);
          __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1829
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__77:             # <<<<<<<<<<<<<<
 *                     self.state = 290
 *                     localctx.ext_annotation = self.annotation()
 */
          __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1829, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__77); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1829, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyObject_RichCompare(__pyx_v__la, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1829, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1829, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (__pyx_t_11) {

            /* "pymola/generated/ModelicaParser.pyx":1830
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__77:
 *                     self.state = 290             # <<<<<<<<<<<<<<
 *                     localctx.ext_annotation = self.annotation()
 * 
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_290) < 0) __PYX_ERR(0, 1830, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":1831
 *                 if _la==ModelicaParser.T__77:
 *                     self.state = 290
 *                     localctx.ext_annotation = self.annotation()             # <<<<<<<<<<<<<<
 * 
 * 
 */
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_annotation); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1831, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_1 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_1)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_1);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
              }
            }
            if (__pyx_t_1) {
              __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1831, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            } else {
              __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1831, __pyx_L6_error)
            }
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_ext_annotation, __pyx_t_3) < 0) __PYX_ERR(0, 1831, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

            /* "pymola/generated/ModelicaParser.pyx":1829
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__77:             # <<<<<<<<<<<<<<
 *                     self.state = 290
 *                     localctx.ext_annotation = self.annotation()
 */
          }

          /* "pymola/generated/ModelicaParser.pyx":1834
 * 
 * 
 *                 self.state = 293             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__17)
 * 
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_293) < 0) __PYX_ERR(0, 1834, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1835
 * 
 *                 self.state = 293
 *                 self.match(ModelicaParser.T__17)             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1835, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1835, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__17); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1835, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_1) {
            __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1835, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_t_5};
              __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1835, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_t_5};
              __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1835, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            {
              __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1835, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_1); __pyx_t_1 = NULL;
              __Pyx_GIVEREF(__pyx_t_5);
              PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_5);
              __pyx_t_5 = 0;
              __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1835, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1807
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__24:             # <<<<<<<<<<<<<<
 *                 self.state = 283
 *                 self.match(ModelicaParser.T__24)
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":1838
 * 
 * 
 *             self.state = 299             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_299) < 0) __PYX_ERR(0, 1838, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":1839
 * 
 *             self.state = 299
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__77:
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1839, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1839, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_2) {
          __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1839, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1839, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1839, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1839, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1839, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "pymola/generated/ModelicaParser.pyx":1840
 *             self.state = 299
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.T__77:
 *                 self.state = 296
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1840, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1840, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__840, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1840, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_3);
        __pyx_t_3 = 0;

        /* "pymola/generated/ModelicaParser.pyx":1841
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__77:             # <<<<<<<<<<<<<<
 *                 self.state = 296
 *                 localctx.comp_annotation = self.annotation()
 */
        __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1841, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__77); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1841, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyObject_RichCompare(__pyx_v__la, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1841, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1841, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":1842
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__77:
 *                 self.state = 296             # <<<<<<<<<<<<<<
 *                 localctx.comp_annotation = self.annotation()
 *                 self.state = 297
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_296) < 0) __PYX_ERR(0, 1842, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1843
 *             if _la==ModelicaParser.T__77:
 *                 self.state = 296
 *                 localctx.comp_annotation = self.annotation()             # <<<<<<<<<<<<<<
 *                 self.state = 297
 *                 self.match(ModelicaParser.T__0)
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_annotation); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1843, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (__pyx_t_5) {
            __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1843, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else {
            __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1843, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_comp_annotation, __pyx_t_3) < 0) __PYX_ERR(0, 1843, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1844
 *                 self.state = 296
 *                 localctx.comp_annotation = self.annotation()
 *                 self.state = 297             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__0)
 * 
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_297) < 0) __PYX_ERR(0, 1844, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1845
 *                 localctx.comp_annotation = self.annotation()
 *                 self.state = 297
 *                 self.match(ModelicaParser.T__0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1845, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1845, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_T__0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1845, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_5) {
            __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1845, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_2};
              __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1845, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_2};
              __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1845, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            } else
            #endif
            {
              __pyx_t_1 = PyTuple_New(1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1845, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5); __pyx_t_5 = NULL;
              __Pyx_GIVEREF(__pyx_t_2);
              PyTuple_SET_ITEM(__pyx_t_1, 0+1, __pyx_t_2);
              __pyx_t_2 = 0;
              __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_1, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1845, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1841
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__77:             # <<<<<<<<<<<<<<
 *                 self.state = 296
 *                 localctx.comp_annotation = self.annotation()
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":1764
 *         self.enterRule(localctx, 18, self.RULE_composition)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 271
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "pymola/generated/ModelicaParser.pyx":1848
 * 
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1848, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.composition", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_7, &__pyx_t_1) < 0) __PYX_ERR(0, 1848, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_7);
        __pyx_v_re = __pyx_t_7;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":1849
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 1849, __pyx_L30_error)

          /* "pymola/generated/ModelicaParser.pyx":1850
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1850, __pyx_L30_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_reportError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1850, __pyx_L30_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1850, __pyx_L30_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1850, __pyx_L30_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1850, __pyx_L30_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_13, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1850, __pyx_L30_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1851
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1851, __pyx_L30_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_recover); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1851, __pyx_L30_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1851, __pyx_L30_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1851, __pyx_L30_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1851, __pyx_L30_error)
            __Pyx_GOTREF(__pyx_t_5);
            if (__pyx_t_4) {
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1851, __pyx_L30_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":1848
 * 
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L31;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L30_error:;
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_6 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L8_except_error;
          }
          __pyx_L31:;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":1764
 *         self.enterRule(localctx, 18, self.RULE_composition)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 271
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":1853
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1853, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_3 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      if (__pyx_t_3) {
        __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1853, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      } else {
        __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1853, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_14 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1853, __pyx_L37_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1853, __pyx_L37_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1853, __pyx_L37_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
      __pyx_L37_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":1854
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Language_specificationContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1759
 * 
 * 
 *     def composition(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.CompositionContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.composition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_la_);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1858
 *     class Language_specificationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Language_specificationContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_29Language_specificationContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_29Language_specificationContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Language_specificationContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 1858, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 1858, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1858, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Language_specificationContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_29Language_specificationContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Language_specificationContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":1859
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 1859, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1859, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1859, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1859, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1859, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1860
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def STRING(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 1860, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1858
 *     class Language_specificationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Language_specificationContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1862
 *             self.parser = parser
 * 
 *         def STRING(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.STRING, 0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Language_specificationContext_3STRING(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_29Language_specificationContext_3STRING = {"STRING", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_29Language_specificationContext_3STRING, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Language_specificationContext_3STRING(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("STRING (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_29Language_specificationContext_2STRING(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Language_specificationContext_2STRING(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("STRING", 0);

  /* "pymola/generated/ModelicaParser.pyx":1863
 * 
 *         def STRING(self):
 *             return self.getToken(ModelicaParser.STRING, 0)             # <<<<<<<<<<<<<<
 * 
 *         def getRuleIndex(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getToken); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1863, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1863, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_STRING); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1863, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1863, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1863, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1863, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1863, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1862
 *             self.parser = parser
 * 
 *         def STRING(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.STRING, 0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Language_specificationContext.STRING", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1865
 *             return self.getToken(ModelicaParser.STRING, 0)
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_language_specification
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Language_specificationContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_29Language_specificationContext_5getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_29Language_specificationContext_5getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Language_specificationContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_29Language_specificationContext_4getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Language_specificationContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":1866
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_language_specification             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_language_specification); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1865
 *             return self.getToken(ModelicaParser.STRING, 0)
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_language_specification
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Language_specificationContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1868
 *             return ModelicaParser.RULE_language_specification
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterLanguage_specification" ):
 *                 listener.enterLanguage_specification(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Language_specificationContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_29Language_specificationContext_7enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_29Language_specificationContext_7enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Language_specificationContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 1868, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 1868, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1868, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Language_specificationContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_29Language_specificationContext_6enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Language_specificationContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":1869
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterLanguage_specification" ):             # <<<<<<<<<<<<<<
 *                 listener.enterLanguage_specification(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterLanguage_specification); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 1869, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1870
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterLanguage_specification" ):
 *                 listener.enterLanguage_specification(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterLanguage_specification); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1870, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1870, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1870, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1870, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1870, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1870, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":1869
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterLanguage_specification" ):             # <<<<<<<<<<<<<<
 *                 listener.enterLanguage_specification(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1868
 *             return ModelicaParser.RULE_language_specification
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterLanguage_specification" ):
 *                 listener.enterLanguage_specification(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Language_specificationContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1872
 *                 listener.enterLanguage_specification(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitLanguage_specification" ):
 *                 listener.exitLanguage_specification(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Language_specificationContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_29Language_specificationContext_9exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_29Language_specificationContext_9exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Language_specificationContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 1872, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 1872, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1872, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Language_specificationContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_29Language_specificationContext_8exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Language_specificationContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":1873
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitLanguage_specification" ):             # <<<<<<<<<<<<<<
 *                 listener.exitLanguage_specification(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitLanguage_specification); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 1873, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1874
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitLanguage_specification" ):
 *                 listener.exitLanguage_specification(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitLanguage_specification); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1874, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1874, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1874, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1874, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1874, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1874, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":1873
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitLanguage_specification" ):             # <<<<<<<<<<<<<<
 *                 listener.exitLanguage_specification(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1872
 *                 listener.enterLanguage_specification(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitLanguage_specification" ):
 *                 listener.exitLanguage_specification(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Language_specificationContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1876
 *                 listener.exitLanguage_specification(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitLanguage_specification" ):
 *                 return visitor.visitLanguage_specification(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Language_specificationContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_29Language_specificationContext_11accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_29Language_specificationContext_11accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Language_specificationContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 1876, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 1876, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1876, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Language_specificationContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_29Language_specificationContext_10accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Language_specificationContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":1877
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitLanguage_specification" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitLanguage_specification(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitLanguage_specification); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 1877, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1878
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitLanguage_specification" ):
 *                 return visitor.visitLanguage_specification(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitLanguage_specification); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1878, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1878, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1878, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1878, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1878, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1878, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":1877
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitLanguage_specification" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitLanguage_specification(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1880
 *                 return visitor.visitLanguage_specification(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1880, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1880, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1880, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1880, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1880, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1880, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":1876
 *                 listener.exitLanguage_specification(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitLanguage_specification" ):
 *                 return visitor.visitLanguage_specification(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Language_specificationContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1885
 * 
 * 
 *     def language_specification(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Language_specificationContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_23language_specification(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_23language_specification = {"language_specification", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_23language_specification, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_23language_specification(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("language_specification (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_22language_specification(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_22language_specification(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  char const *__pyx_t_20;
  __Pyx_RefNannySetupContext("language_specification", 0);

  /* "pymola/generated/ModelicaParser.pyx":1887
 *     def language_specification(self):
 * 
 *         localctx = ModelicaParser.Language_specificationContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 20, self.RULE_language_specification)
 *         try:
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1887, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Language_specificationContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1887, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1887, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1887, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1887, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1887, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1887, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1887, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1888
 * 
 *         localctx = ModelicaParser.Language_specificationContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 20, self.RULE_language_specification)             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1888, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_language_specification); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1888, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_20, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1888, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_20, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1888, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1888, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_20);
    __Pyx_GIVEREF(__pyx_int_20);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_20);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1888, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1889
 *         localctx = ModelicaParser.Language_specificationContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 20, self.RULE_language_specification)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 301
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":1890
 *         self.enterRule(localctx, 20, self.RULE_language_specification)
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 301
 *             self.match(ModelicaParser.STRING)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1890, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1890, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1890, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1890, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1890, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":1891
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 301             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.STRING)
 *         except RecognitionException as re:
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_301) < 0) __PYX_ERR(0, 1891, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":1892
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 301
 *             self.match(ModelicaParser.STRING)             # <<<<<<<<<<<<<<
 *         except RecognitionException as re:
 *             localctx.exception = re
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1892, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1892, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_STRING); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1892, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1892, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1892, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1892, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1892, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1892, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":1889
 *         localctx = ModelicaParser.Language_specificationContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 20, self.RULE_language_specification)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 301
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":1893
 *             self.state = 301
 *             self.match(ModelicaParser.STRING)
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1893, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.language_specification", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_3, &__pyx_t_4) < 0) __PYX_ERR(0, 1893, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __pyx_v_re = __pyx_t_3;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":1894
 *             self.match(ModelicaParser.STRING)
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 1894, __pyx_L19_error)

          /* "pymola/generated/ModelicaParser.pyx":1895
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1895, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1895, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1895, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1895, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_11 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1895, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_11);
            if (__pyx_t_7) {
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_7); __pyx_t_7 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_11, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1895, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1896
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1896, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1896, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_11)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1896, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_11)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1896, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1896, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_7);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1896, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":1893
 *             self.state = 301
 *             self.match(ModelicaParser.STRING)
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L20;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L19_error:;
            __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_14);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __pyx_t_6 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_17);
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
            }
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
            __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
            goto __pyx_L8_except_error;
          }
          __pyx_L20:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":1889
 *         localctx = ModelicaParser.Language_specificationContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 20, self.RULE_language_specification)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 301
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":1898
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1898, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1898, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1898, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_18, &__pyx_t_17);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_18);
      __Pyx_XGOTREF(__pyx_t_17);
      __pyx_t_12 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_20 = __pyx_filename;
      {
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1898, __pyx_L26_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1898, __pyx_L26_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1898, __pyx_L26_error)
        }
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_18, __pyx_t_17);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0;
      __pyx_lineno = __pyx_t_12; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_20;
      goto __pyx_L1_error;
      __pyx_L26_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_18, __pyx_t_17);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":1899
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class External_function_callContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1885
 * 
 * 
 *     def language_specification(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Language_specificationContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.language_specification", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1903
 *     class External_function_callContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29External_function_callContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_29External_function_callContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_29External_function_callContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29External_function_callContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 1903, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 1903, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1903, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.External_function_callContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_29External_function_callContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29External_function_callContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":1904
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 1904, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1904, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1904, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1904, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1904, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1905
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def IDENT(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 1905, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1903
 *     class External_function_callContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.External_function_callContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1907
 *             self.parser = parser
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29External_function_callContext_3IDENT(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_29External_function_callContext_3IDENT = {"IDENT", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_29External_function_callContext_3IDENT, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29External_function_callContext_3IDENT(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("IDENT (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_29External_function_callContext_2IDENT(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29External_function_callContext_2IDENT(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("IDENT", 0);

  /* "pymola/generated/ModelicaParser.pyx":1908
 * 
 *         def IDENT(self):
 *             return self.getToken(ModelicaParser.IDENT, 0)             # <<<<<<<<<<<<<<
 * 
 *         def component_reference(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getToken); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1908, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1908, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1908, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1908, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1908, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1908, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1908, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1907
 *             self.parser = parser
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.External_function_callContext.IDENT", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1910
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 *         def component_reference(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29External_function_callContext_5component_reference(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_29External_function_callContext_5component_reference = {"component_reference", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_29External_function_callContext_5component_reference, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29External_function_callContext_5component_reference(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("component_reference (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_29External_function_callContext_4component_reference(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29External_function_callContext_4component_reference(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("component_reference", 0);

  /* "pymola/generated/ModelicaParser.pyx":1911
 * 
 *         def component_reference(self):
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1911, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1911, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Component_referenceContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1911, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1911, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1911, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1911, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1911, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1910
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 *         def component_reference(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.External_function_callContext.component_reference", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1914
 * 
 * 
 *         def expression_list(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Expression_listContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29External_function_callContext_7expression_list(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_29External_function_callContext_7expression_list = {"expression_list", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_29External_function_callContext_7expression_list, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29External_function_callContext_7expression_list(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("expression_list (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_29External_function_callContext_6expression_list(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29External_function_callContext_6expression_list(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("expression_list", 0);

  /* "pymola/generated/ModelicaParser.pyx":1915
 * 
 *         def expression_list(self):
 *             return self.getTypedRuleContext(ModelicaParser.Expression_listContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1915, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1915, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Expression_listContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1915, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1915, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1915, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1915, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1915, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1914
 * 
 * 
 *         def expression_list(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Expression_listContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.External_function_callContext.expression_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1918
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_external_function_call
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29External_function_callContext_9getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_29External_function_callContext_9getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_29External_function_callContext_9getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29External_function_callContext_9getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_29External_function_callContext_8getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29External_function_callContext_8getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":1919
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_external_function_call             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1919, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_external_function_call); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1919, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1918
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_external_function_call
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.External_function_callContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1921
 *             return ModelicaParser.RULE_external_function_call
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExternal_function_call" ):
 *                 listener.enterExternal_function_call(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29External_function_callContext_11enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_29External_function_callContext_11enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_29External_function_callContext_11enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29External_function_callContext_11enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 1921, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 1921, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1921, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.External_function_callContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_29External_function_callContext_10enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29External_function_callContext_10enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":1922
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExternal_function_call" ):             # <<<<<<<<<<<<<<
 *                 listener.enterExternal_function_call(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterExternal_function_call); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 1922, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1923
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExternal_function_call" ):
 *                 listener.enterExternal_function_call(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterExternal_function_call); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1923, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1923, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1923, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1923, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1923, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1923, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":1922
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExternal_function_call" ):             # <<<<<<<<<<<<<<
 *                 listener.enterExternal_function_call(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1921
 *             return ModelicaParser.RULE_external_function_call
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExternal_function_call" ):
 *                 listener.enterExternal_function_call(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.External_function_callContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1925
 *                 listener.enterExternal_function_call(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExternal_function_call" ):
 *                 listener.exitExternal_function_call(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29External_function_callContext_13exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_29External_function_callContext_13exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_29External_function_callContext_13exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29External_function_callContext_13exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 1925, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 1925, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1925, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.External_function_callContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_29External_function_callContext_12exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29External_function_callContext_12exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":1926
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExternal_function_call" ):             # <<<<<<<<<<<<<<
 *                 listener.exitExternal_function_call(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitExternal_function_call); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 1926, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1927
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExternal_function_call" ):
 *                 listener.exitExternal_function_call(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitExternal_function_call); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1927, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1927, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1927, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1927, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1927, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1927, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":1926
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExternal_function_call" ):             # <<<<<<<<<<<<<<
 *                 listener.exitExternal_function_call(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1925
 *                 listener.enterExternal_function_call(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExternal_function_call" ):
 *                 listener.exitExternal_function_call(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.External_function_callContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1929
 *                 listener.exitExternal_function_call(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExternal_function_call" ):
 *                 return visitor.visitExternal_function_call(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29External_function_callContext_15accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_29External_function_callContext_15accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_29External_function_callContext_15accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29External_function_callContext_15accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 1929, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 1929, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1929, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.External_function_callContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_29External_function_callContext_14accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29External_function_callContext_14accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":1930
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExternal_function_call" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitExternal_function_call(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitExternal_function_call); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 1930, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1931
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExternal_function_call" ):
 *                 return visitor.visitExternal_function_call(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitExternal_function_call); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1931, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1931, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1931, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1931, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1931, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1931, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":1930
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExternal_function_call" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitExternal_function_call(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1933
 *                 return visitor.visitExternal_function_call(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1933, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1933, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1933, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1933, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1933, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1933, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":1929
 *                 listener.exitExternal_function_call(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExternal_function_call" ):
 *                 return visitor.visitExternal_function_call(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.External_function_callContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1938
 * 
 * 
 *     def external_function_call(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.External_function_callContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_25external_function_call(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_25external_function_call = {"external_function_call", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_25external_function_call, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_25external_function_call(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("external_function_call (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_24external_function_call(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_24external_function_call(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v_la_ = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  char const *__pyx_t_22;
  __Pyx_RefNannySetupContext("external_function_call", 0);

  /* "pymola/generated/ModelicaParser.pyx":1940
 *     def external_function_call(self):
 * 
 *         localctx = ModelicaParser.External_function_callContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 22, self.RULE_external_function_call)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1940, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_External_function_callContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1940, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1940, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1940, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1940, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1940, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1940, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1940, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1941
 * 
 *         localctx = ModelicaParser.External_function_callContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 22, self.RULE_external_function_call)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1941, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_external_function_call); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1941, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_22, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1941, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_22, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1941, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1941, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_22);
    __Pyx_GIVEREF(__pyx_int_22);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_22);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1941, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1942
 *         localctx = ModelicaParser.External_function_callContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 22, self.RULE_external_function_call)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 1942, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1943
 *         self.enterRule(localctx, 22, self.RULE_external_function_call)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 306
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":1944
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 306
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1944, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1944, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1944, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1944, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1944, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":1945
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 306             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             la_ = self._interp.adaptivePredict(self._input,26,self._ctx)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_306) < 0) __PYX_ERR(0, 1945, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":1946
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 306
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             la_ = self._interp.adaptivePredict(self._input,26,self._ctx)
 *             if la_ == 1:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1946, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1946, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1946, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1946, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1946, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1946, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1946, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":1947
 *             self.state = 306
 *             self._errHandler.sync(self)
 *             la_ = self._interp.adaptivePredict(self._input,26,self._ctx)             # <<<<<<<<<<<<<<
 *             if la_ == 1:
 *                 self.state = 303
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_interp); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1947, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_adaptivePredict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1947, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1947, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1947, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_2)) {
          PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_t_7, __pyx_int_26, __pyx_t_3};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1947, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
          PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_t_7, __pyx_int_26, __pyx_t_3};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1947, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1947, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_7);
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_t_7);
          __Pyx_INCREF(__pyx_int_26);
          __Pyx_GIVEREF(__pyx_int_26);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_int_26);
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_6, __pyx_t_3);
          __pyx_t_7 = 0;
          __pyx_t_3 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1947, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_v_la_ = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":1948
 *             self._errHandler.sync(self)
 *             la_ = self._interp.adaptivePredict(self._input,26,self._ctx)
 *             if la_ == 1:             # <<<<<<<<<<<<<<
 *                 self.state = 303
 *                 self.component_reference()
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1948, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1948, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":1949
 *             la_ = self._interp.adaptivePredict(self._input,26,self._ctx)
 *             if la_ == 1:
 *                 self.state = 303             # <<<<<<<<<<<<<<
 *                 self.component_reference()
 *                 self.state = 304
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_303) < 0) __PYX_ERR(0, 1949, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1950
 *             if la_ == 1:
 *                 self.state = 303
 *                 self.component_reference()             # <<<<<<<<<<<<<<
 *                 self.state = 304
 *                 self.match(ModelicaParser.T__14)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_component_reference); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1950, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (__pyx_t_5) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1950, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1950, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1951
 *                 self.state = 303
 *                 self.component_reference()
 *                 self.state = 304             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__14)
 * 
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_304) < 0) __PYX_ERR(0, 1951, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1952
 *                 self.component_reference()
 *                 self.state = 304
 *                 self.match(ModelicaParser.T__14)             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1952, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1952, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_T__14); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1952, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_5) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1952, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1952, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1952, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1952, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1952, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1948
 *             self._errHandler.sync(self)
 *             la_ = self._interp.adaptivePredict(self._input,26,self._ctx)
 *             if la_ == 1:             # <<<<<<<<<<<<<<
 *                 self.state = 303
 *                 self.component_reference()
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":1955
 * 
 * 
 *             self.state = 308             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.IDENT)
 *             self.state = 309
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_308) < 0) __PYX_ERR(0, 1955, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":1956
 * 
 *             self.state = 308
 *             self.match(ModelicaParser.IDENT)             # <<<<<<<<<<<<<<
 *             self.state = 309
 *             self.match(ModelicaParser.T__16)
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1956, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1956, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1956, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1956, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_3};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1956, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_3};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1956, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          } else
          #endif
          {
            __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1956, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_GIVEREF(__pyx_t_3);
            PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_t_3);
            __pyx_t_3 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1956, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":1957
 *             self.state = 308
 *             self.match(ModelicaParser.IDENT)
 *             self.state = 309             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__16)
 *             self.state = 311
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_309) < 0) __PYX_ERR(0, 1957, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":1958
 *             self.match(ModelicaParser.IDENT)
 *             self.state = 309
 *             self.match(ModelicaParser.T__16)             # <<<<<<<<<<<<<<
 *             self.state = 311
 *             self._errHandler.sync(self)
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1958, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1958, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_T__16); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1958, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_5 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        if (!__pyx_t_5) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1958, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1958, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1958, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1958, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
            __Pyx_GIVEREF(__pyx_t_3);
            PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_3);
            __pyx_t_3 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1958, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":1959
 *             self.state = 309
 *             self.match(ModelicaParser.T__16)
 *             self.state = 311             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_311) < 0) __PYX_ERR(0, 1959, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":1960
 *             self.match(ModelicaParser.T__16)
 *             self.state = 311
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__13) | (1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__19) | (1 << ModelicaParser.T__29) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__55))) != 0) or ((((_la - 71)) & ~0x3f) == 0 and ((1 << (_la - 71)) & ((1 << (ModelicaParser.T__70 - 71)) | (1 << (ModelicaParser.T__73 - 71)) | (1 << (ModelicaParser.T__74 - 71)) | (1 << (ModelicaParser.T__75 - 71)) | (1 << (ModelicaParser.INITIAL - 71)) | (1 << (ModelicaParser.IDENT - 71)) | (1 << (ModelicaParser.STRING - 71)) | (1 << (ModelicaParser.UNSIGNED_NUMBER - 71)))) != 0):
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1960, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1960, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_2) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1960, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1960, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1960, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1960, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1960, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":1961
 *             self.state = 311
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__13) | (1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__19) | (1 << ModelicaParser.T__29) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__55))) != 0) or ((((_la - 71)) & ~0x3f) == 0 and ((1 << (_la - 71)) & ((1 << (ModelicaParser.T__70 - 71)) | (1 << (ModelicaParser.T__73 - 71)) | (1 << (ModelicaParser.T__74 - 71)) | (1 << (ModelicaParser.T__75 - 71)) | (1 << (ModelicaParser.INITIAL - 71)) | (1 << (ModelicaParser.IDENT - 71)) | (1 << (ModelicaParser.STRING - 71)) | (1 << (ModelicaParser.UNSIGNED_NUMBER - 71)))) != 0):
 *                 self.state = 310
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1961, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1961, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__841, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1961, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v__la = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":1962
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__13) | (1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__19) | (1 << ModelicaParser.T__29) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__55))) != 0) or ((((_la - 71)) & ~0x3f) == 0 and ((1 << (_la - 71)) & ((1 << (ModelicaParser.T__70 - 71)) | (1 << (ModelicaParser.T__73 - 71)) | (1 << (ModelicaParser.T__74 - 71)) | (1 << (ModelicaParser.T__75 - 71)) | (1 << (ModelicaParser.INITIAL - 71)) | (1 << (ModelicaParser.IDENT - 71)) | (1 << (ModelicaParser.STRING - 71)) | (1 << (ModelicaParser.UNSIGNED_NUMBER - 71)))) != 0):             # <<<<<<<<<<<<<<
 *                 self.state = 310
 *                 self.expression_list()
 */
        __pyx_t_1 = __Pyx_PyInt_From_long((~0x3f)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = PyNumber_And(__pyx_v__la, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_t_7, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (!__pyx_t_12) {
          goto __pyx_L17_next_or;
        } else {
        }
        __pyx_t_1 = PyNumber_Lshift(__pyx_int_1, __pyx_v__la); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__13); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__16); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_Or(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__19); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyNumber_Or(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__29); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_Or(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__39); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyNumber_Or(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__55); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_Or(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_And(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_3, __pyx_int_0, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L16_bool_binop_done;
        }
        __pyx_L17_next_or:;
        __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_v__la, __pyx_int_71, 71, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_PyInt_From_long((~0x3f)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = PyNumber_And(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyInt_EqObjC(__pyx_t_1, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L16_bool_binop_done;
        }
        __pyx_t_3 = __Pyx_PyInt_SubtractObjC(__pyx_v__la, __pyx_int_71, 71, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__70); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyInt_SubtractObjC(__pyx_t_2, __pyx_int_71, 71, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__73); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyInt_SubtractObjC(__pyx_t_7, __pyx_int_71, 71, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_Or(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__74); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyInt_SubtractObjC(__pyx_t_2, __pyx_int_71, 71, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyNumber_Or(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__75); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_t_3, __pyx_int_71, 71, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_Or(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_INITIAL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyInt_SubtractObjC(__pyx_t_7, __pyx_int_71, 71, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_Or(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyInt_SubtractObjC(__pyx_t_2, __pyx_int_71, 71, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyNumber_Or(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_STRING); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_t_3, __pyx_int_71, 71, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_Or(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_UNSIGNED_NUMBER); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyInt_SubtractObjC(__pyx_t_7, __pyx_int_71, 71, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_Or(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyNumber_And(__pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyObject_RichCompare(__pyx_t_7, __pyx_int_0, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 1962, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_11 = __pyx_t_12;
        __pyx_L16_bool_binop_done:;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":1963
 *             _la = self._input.LA(1)
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__13) | (1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__19) | (1 << ModelicaParser.T__29) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__55))) != 0) or ((((_la - 71)) & ~0x3f) == 0 and ((1 << (_la - 71)) & ((1 << (ModelicaParser.T__70 - 71)) | (1 << (ModelicaParser.T__73 - 71)) | (1 << (ModelicaParser.T__74 - 71)) | (1 << (ModelicaParser.T__75 - 71)) | (1 << (ModelicaParser.INITIAL - 71)) | (1 << (ModelicaParser.IDENT - 71)) | (1 << (ModelicaParser.STRING - 71)) | (1 << (ModelicaParser.UNSIGNED_NUMBER - 71)))) != 0):
 *                 self.state = 310             # <<<<<<<<<<<<<<
 *                 self.expression_list()
 * 
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_310) < 0) __PYX_ERR(0, 1963, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":1964
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__13) | (1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__19) | (1 << ModelicaParser.T__29) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__55))) != 0) or ((((_la - 71)) & ~0x3f) == 0 and ((1 << (_la - 71)) & ((1 << (ModelicaParser.T__70 - 71)) | (1 << (ModelicaParser.T__73 - 71)) | (1 << (ModelicaParser.T__74 - 71)) | (1 << (ModelicaParser.T__75 - 71)) | (1 << (ModelicaParser.INITIAL - 71)) | (1 << (ModelicaParser.IDENT - 71)) | (1 << (ModelicaParser.STRING - 71)) | (1 << (ModelicaParser.UNSIGNED_NUMBER - 71)))) != 0):
 *                 self.state = 310
 *                 self.expression_list()             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expression_list); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1964, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_1 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (__pyx_t_1) {
            __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1964, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          } else {
            __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1964, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1962
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__13) | (1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__19) | (1 << ModelicaParser.T__29) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__55))) != 0) or ((((_la - 71)) & ~0x3f) == 0 and ((1 << (_la - 71)) & ((1 << (ModelicaParser.T__70 - 71)) | (1 << (ModelicaParser.T__73 - 71)) | (1 << (ModelicaParser.T__74 - 71)) | (1 << (ModelicaParser.T__75 - 71)) | (1 << (ModelicaParser.INITIAL - 71)) | (1 << (ModelicaParser.IDENT - 71)) | (1 << (ModelicaParser.STRING - 71)) | (1 << (ModelicaParser.UNSIGNED_NUMBER - 71)))) != 0):             # <<<<<<<<<<<<<<
 *                 self.state = 310
 *                 self.expression_list()
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":1967
 * 
 * 
 *             self.state = 313             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__18)
 *         except RecognitionException as re:
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_313) < 0) __PYX_ERR(0, 1967, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":1968
 * 
 *             self.state = 313
 *             self.match(ModelicaParser.T__18)             # <<<<<<<<<<<<<<
 *         except RecognitionException as re:
 *             localctx.exception = re
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1968, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1968, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__18); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1968, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_1) {
          __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1968, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_3);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_t_2};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1968, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_t_2};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1968, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1968, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1); __pyx_t_1 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1968, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "pymola/generated/ModelicaParser.pyx":1943
 *         self.enterRule(localctx, 22, self.RULE_external_function_call)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 306
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "pymola/generated/ModelicaParser.pyx":1969
 *             self.state = 313
 *             self.match(ModelicaParser.T__18)
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1969, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.external_function_call", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_7, &__pyx_t_5) < 0) __PYX_ERR(0, 1969, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_7);
        __pyx_v_re = __pyx_t_7;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":1970
 *             self.match(ModelicaParser.T__18)
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 1970, __pyx_L25_error)

          /* "pymola/generated/ModelicaParser.pyx":1971
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1971, __pyx_L25_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_reportError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1971, __pyx_L25_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1971, __pyx_L25_error)
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1971, __pyx_L25_error)
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1971, __pyx_L25_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_1) {
              __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_1); __pyx_t_1 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_13, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1971, __pyx_L25_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":1972
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1972, __pyx_L25_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_recover); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1972, __pyx_L25_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1972, __pyx_L25_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1972, __pyx_L25_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_1 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1972, __pyx_L25_error)
            __Pyx_GOTREF(__pyx_t_1);
            if (__pyx_t_4) {
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4); __pyx_t_4 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1972, __pyx_L25_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":1969
 *             self.state = 313
 *             self.match(ModelicaParser.T__18)
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L26;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L25_error:;
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_6 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L8_except_error;
          }
          __pyx_L26:;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":1943
 *         self.enterRule(localctx, 22, self.RULE_external_function_call)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 306
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":1974
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1974, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_3 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      if (__pyx_t_3) {
        __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1974, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      } else {
        __pyx_t_5 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1974, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_14 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1974, __pyx_L32_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (__pyx_t_3) {
          __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1974, __pyx_L32_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else {
          __pyx_t_5 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1974, __pyx_L32_error)
        }
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
      __pyx_L32_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":1975
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Element_listContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1938
 * 
 * 
 *     def external_function_call(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.External_function_callContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.external_function_call", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v_la_);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1979
 *     class Element_listContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Element_listContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19Element_listContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19Element_listContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Element_listContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 1979, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 1979, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1979, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_listContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19Element_listContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Element_listContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":1980
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 1980, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1980, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1980, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1980, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1980, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1980, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1980, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1980, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1981
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def element(self, i:int=None):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 1981, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1979
 *     class Element_listContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_listContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1983
 *             self.parser = parser
 * 
 *         def element(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ElementContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Element_listContext_3element(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19Element_listContext_3element = {"element", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19Element_listContext_3element, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Element_listContext_3element(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("element (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "element") < 0)) __PYX_ERR(0, 1983, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("element", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1983, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_listContext.element", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19Element_listContext_2element(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Element_listContext_2element(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("element", 0);

  /* "pymola/generated/ModelicaParser.pyx":1984
 * 
 *         def element(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.ElementContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1985
 *         def element(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ElementContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.ElementContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1985, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1985, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_ElementContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1985, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1985, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1985, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1985, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1985, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1985, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":1984
 * 
 *         def element(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.ElementContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1987
 *                 return self.getTypedRuleContexts(ModelicaParser.ElementContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.ElementContext,i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1987, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1987, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_ElementContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1987, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1987, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1987, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1987, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1987, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":1983
 *             self.parser = parser
 * 
 *         def element(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ElementContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_listContext.element", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1990
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_element_list
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Element_listContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19Element_listContext_5getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19Element_listContext_5getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Element_listContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19Element_listContext_4getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Element_listContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":1991
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_element_list             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1991, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_element_list); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1991, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":1990
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_element_list
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_listContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1993
 *             return ModelicaParser.RULE_element_list
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterElement_list" ):
 *                 listener.enterElement_list(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Element_listContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19Element_listContext_7enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19Element_listContext_7enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Element_listContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 1993, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 1993, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1993, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_listContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19Element_listContext_6enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Element_listContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":1994
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterElement_list" ):             # <<<<<<<<<<<<<<
 *                 listener.enterElement_list(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterElement_list); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 1994, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1995
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterElement_list" ):
 *                 listener.enterElement_list(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterElement_list); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1995, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1995, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1995, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1995, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1995, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1995, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":1994
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterElement_list" ):             # <<<<<<<<<<<<<<
 *                 listener.enterElement_list(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1993
 *             return ModelicaParser.RULE_element_list
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterElement_list" ):
 *                 listener.enterElement_list(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_listContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":1997
 *                 listener.enterElement_list(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitElement_list" ):
 *                 listener.exitElement_list(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Element_listContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19Element_listContext_9exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19Element_listContext_9exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Element_listContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 1997, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 1997, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1997, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_listContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19Element_listContext_8exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Element_listContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":1998
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitElement_list" ):             # <<<<<<<<<<<<<<
 *                 listener.exitElement_list(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitElement_list); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 1998, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":1999
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitElement_list" ):
 *                 listener.exitElement_list(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitElement_list); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1999, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1999, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1999, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1999, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1999, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1999, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":1998
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitElement_list" ):             # <<<<<<<<<<<<<<
 *                 listener.exitElement_list(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":1997
 *                 listener.enterElement_list(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitElement_list" ):
 *                 listener.exitElement_list(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_listContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2001
 *                 listener.exitElement_list(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitElement_list" ):
 *                 return visitor.visitElement_list(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Element_listContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19Element_listContext_11accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19Element_listContext_11accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Element_listContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 2001, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 2001, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2001, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_listContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19Element_listContext_10accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Element_listContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":2002
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitElement_list" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitElement_list(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitElement_list); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 2002, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":2003
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitElement_list" ):
 *                 return visitor.visitElement_list(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitElement_list); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2003, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2003, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2003, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2003, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2003, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2003, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":2002
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitElement_list" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitElement_list(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":2005
 *                 return visitor.visitElement_list(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2005, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2005, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2005, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2005, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2005, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2005, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":2001
 *                 listener.exitElement_list(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitElement_list" ):
 *                 return visitor.visitElement_list(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_listContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2010
 * 
 * 
 *     def element_list(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Element_listContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_27element_list(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_27element_list = {"element_list", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_27element_list, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_27element_list(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("element_list (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_26element_list(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_26element_list(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  char const *__pyx_t_22;
  __Pyx_RefNannySetupContext("element_list", 0);

  /* "pymola/generated/ModelicaParser.pyx":2012
 *     def element_list(self):
 * 
 *         localctx = ModelicaParser.Element_listContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 24, self.RULE_element_list)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2012, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Element_listContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2012, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2012, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2012, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2012, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2012, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2012, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2012, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2013
 * 
 *         localctx = ModelicaParser.Element_listContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 24, self.RULE_element_list)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_element_list); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_24, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2013, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_24, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2013, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2013, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_24);
    __Pyx_GIVEREF(__pyx_int_24);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_24);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2013, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2014
 *         localctx = ModelicaParser.Element_listContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 24, self.RULE_element_list)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 2014, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2015
 *         self.enterRule(localctx, 24, self.RULE_element_list)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 320
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":2016
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 320
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2016, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2016, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2016, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2016, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2016, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2017
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 320             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_320) < 0) __PYX_ERR(0, 2017, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":2018
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 320
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__1) | (1 << ModelicaParser.T__2) | (1 << ModelicaParser.T__3) | (1 << ModelicaParser.T__4) | (1 << ModelicaParser.T__5) | (1 << ModelicaParser.T__6) | (1 << ModelicaParser.T__7) | (1 << ModelicaParser.T__8) | (1 << ModelicaParser.T__9) | (1 << ModelicaParser.T__10) | (1 << ModelicaParser.T__11) | (1 << ModelicaParser.T__12) | (1 << ModelicaParser.T__21) | (1 << ModelicaParser.T__25) | (1 << ModelicaParser.T__26) | (1 << ModelicaParser.T__32) | (1 << ModelicaParser.T__33) | (1 << ModelicaParser.T__34) | (1 << ModelicaParser.T__35) | (1 << ModelicaParser.T__36) | (1 << ModelicaParser.T__37) | (1 << ModelicaParser.T__38))) != 0) or ((((_la - 80)) & ~0x3f) == 0 and ((1 << (_la - 80)) & ((1 << (ModelicaParser.PARTIAL - 80)) | (1 << (ModelicaParser.FINAL - 80)) | (1 << (ModelicaParser.ENCAPSULATED - 80)) | (1 << (ModelicaParser.REDECLARE - 80)) | (1 << (ModelicaParser.INNER - 80)) | (1 << (ModelicaParser.OUTER - 80)) | (1 << (ModelicaParser.IDENT - 80)))) != 0):
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2018, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2018, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2018, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2018, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2018, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2018, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2018, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2019
 *             self.state = 320
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__1) | (1 << ModelicaParser.T__2) | (1 << ModelicaParser.T__3) | (1 << ModelicaParser.T__4) | (1 << ModelicaParser.T__5) | (1 << ModelicaParser.T__6) | (1 << ModelicaParser.T__7) | (1 << ModelicaParser.T__8) | (1 << ModelicaParser.T__9) | (1 << ModelicaParser.T__10) | (1 << ModelicaParser.T__11) | (1 << ModelicaParser.T__12) | (1 << ModelicaParser.T__21) | (1 << ModelicaParser.T__25) | (1 << ModelicaParser.T__26) | (1 << ModelicaParser.T__32) | (1 << ModelicaParser.T__33) | (1 << ModelicaParser.T__34) | (1 << ModelicaParser.T__35) | (1 << ModelicaParser.T__36) | (1 << ModelicaParser.T__37) | (1 << ModelicaParser.T__38))) != 0) or ((((_la - 80)) & ~0x3f) == 0 and ((1 << (_la - 80)) & ((1 << (ModelicaParser.PARTIAL - 80)) | (1 << (ModelicaParser.FINAL - 80)) | (1 << (ModelicaParser.ENCAPSULATED - 80)) | (1 << (ModelicaParser.REDECLARE - 80)) | (1 << (ModelicaParser.INNER - 80)) | (1 << (ModelicaParser.OUTER - 80)) | (1 << (ModelicaParser.IDENT - 80)))) != 0):
 *                 self.state = 315
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2019, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2019, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__842, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2019, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v__la = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2020
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__1) | (1 << ModelicaParser.T__2) | (1 << ModelicaParser.T__3) | (1 << ModelicaParser.T__4) | (1 << ModelicaParser.T__5) | (1 << ModelicaParser.T__6) | (1 << ModelicaParser.T__7) | (1 << ModelicaParser.T__8) | (1 << ModelicaParser.T__9) | (1 << ModelicaParser.T__10) | (1 << ModelicaParser.T__11) | (1 << ModelicaParser.T__12) | (1 << ModelicaParser.T__21) | (1 << ModelicaParser.T__25) | (1 << ModelicaParser.T__26) | (1 << ModelicaParser.T__32) | (1 << ModelicaParser.T__33) | (1 << ModelicaParser.T__34) | (1 << ModelicaParser.T__35) | (1 << ModelicaParser.T__36) | (1 << ModelicaParser.T__37) | (1 << ModelicaParser.T__38))) != 0) or ((((_la - 80)) & ~0x3f) == 0 and ((1 << (_la - 80)) & ((1 << (ModelicaParser.PARTIAL - 80)) | (1 << (ModelicaParser.FINAL - 80)) | (1 << (ModelicaParser.ENCAPSULATED - 80)) | (1 << (ModelicaParser.REDECLARE - 80)) | (1 << (ModelicaParser.INNER - 80)) | (1 << (ModelicaParser.OUTER - 80)) | (1 << (ModelicaParser.IDENT - 80)))) != 0):             # <<<<<<<<<<<<<<
 *                 self.state = 315
 *                 self.element()
 */
        while (1) {
          __pyx_t_1 = __Pyx_PyInt_From_long((~0x3f)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_7 = PyNumber_And(__pyx_v__la, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_t_7, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (!__pyx_t_12) {
            goto __pyx_L17_next_or;
          } else {
          }
          __pyx_t_1 = PyNumber_Lshift(__pyx_int_1, __pyx_v__la); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyNumber_Or(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Or(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyNumber_Or(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Or(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyNumber_Or(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Or(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyNumber_Or(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Or(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyNumber_Or(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__11); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Or(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__12); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyNumber_Or(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__21); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Or(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__25); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyNumber_Or(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__26); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Or(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__32); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyNumber_Or(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__33); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Or(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__34); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyNumber_Or(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__35); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Or(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__36); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyNumber_Or(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__37); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Or(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__38); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyNumber_Or(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_And(__pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_int_0, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (!__pyx_t_12) {
          } else {
            __pyx_t_11 = __pyx_t_12;
            goto __pyx_L16_bool_binop_done;
          }
          __pyx_L17_next_or:;
          __pyx_t_3 = __Pyx_PyInt_SubtractObjC(__pyx_v__la, __pyx_int_80, 80, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = __Pyx_PyInt_From_long((~0x3f)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_1 = PyNumber_And(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_t_1, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          if (__pyx_t_12) {
          } else {
            __pyx_t_11 = __pyx_t_12;
            goto __pyx_L16_bool_binop_done;
          }
          __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_v__la, __pyx_int_80, 80, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_1 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_PARTIAL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_t_3, __pyx_int_80, 80, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_FINAL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_t_7, __pyx_int_80, 80, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Or(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_ENCAPSULATED); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_PyInt_SubtractObjC(__pyx_t_3, __pyx_int_80, 80, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Or(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_REDECLARE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = __Pyx_PyInt_SubtractObjC(__pyx_t_2, __pyx_int_80, 80, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyNumber_Or(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_INNER); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_t_7, __pyx_int_80, 80, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Or(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_OUTER); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_PyInt_SubtractObjC(__pyx_t_3, __pyx_int_80, 80, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Or(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = __Pyx_PyInt_SubtractObjC(__pyx_t_2, __pyx_int_80, 80, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyNumber_Or(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_And(__pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_int_0, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2020, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_11 = __pyx_t_12;
          __pyx_L16_bool_binop_done:;
          if (!__pyx_t_11) break;

          /* "pymola/generated/ModelicaParser.pyx":2021
 *             _la = self._input.LA(1)
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__1) | (1 << ModelicaParser.T__2) | (1 << ModelicaParser.T__3) | (1 << ModelicaParser.T__4) | (1 << ModelicaParser.T__5) | (1 << ModelicaParser.T__6) | (1 << ModelicaParser.T__7) | (1 << ModelicaParser.T__8) | (1 << ModelicaParser.T__9) | (1 << ModelicaParser.T__10) | (1 << ModelicaParser.T__11) | (1 << ModelicaParser.T__12) | (1 << ModelicaParser.T__21) | (1 << ModelicaParser.T__25) | (1 << ModelicaParser.T__26) | (1 << ModelicaParser.T__32) | (1 << ModelicaParser.T__33) | (1 << ModelicaParser.T__34) | (1 << ModelicaParser.T__35) | (1 << ModelicaParser.T__36) | (1 << ModelicaParser.T__37) | (1 << ModelicaParser.T__38))) != 0) or ((((_la - 80)) & ~0x3f) == 0 and ((1 << (_la - 80)) & ((1 << (ModelicaParser.PARTIAL - 80)) | (1 << (ModelicaParser.FINAL - 80)) | (1 << (ModelicaParser.ENCAPSULATED - 80)) | (1 << (ModelicaParser.REDECLARE - 80)) | (1 << (ModelicaParser.INNER - 80)) | (1 << (ModelicaParser.OUTER - 80)) | (1 << (ModelicaParser.IDENT - 80)))) != 0):
 *                 self.state = 315             # <<<<<<<<<<<<<<
 *                 self.element()
 *                 self.state = 316
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_315) < 0) __PYX_ERR(0, 2021, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":2022
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__1) | (1 << ModelicaParser.T__2) | (1 << ModelicaParser.T__3) | (1 << ModelicaParser.T__4) | (1 << ModelicaParser.T__5) | (1 << ModelicaParser.T__6) | (1 << ModelicaParser.T__7) | (1 << ModelicaParser.T__8) | (1 << ModelicaParser.T__9) | (1 << ModelicaParser.T__10) | (1 << ModelicaParser.T__11) | (1 << ModelicaParser.T__12) | (1 << ModelicaParser.T__21) | (1 << ModelicaParser.T__25) | (1 << ModelicaParser.T__26) | (1 << ModelicaParser.T__32) | (1 << ModelicaParser.T__33) | (1 << ModelicaParser.T__34) | (1 << ModelicaParser.T__35) | (1 << ModelicaParser.T__36) | (1 << ModelicaParser.T__37) | (1 << ModelicaParser.T__38))) != 0) or ((((_la - 80)) & ~0x3f) == 0 and ((1 << (_la - 80)) & ((1 << (ModelicaParser.PARTIAL - 80)) | (1 << (ModelicaParser.FINAL - 80)) | (1 << (ModelicaParser.ENCAPSULATED - 80)) | (1 << (ModelicaParser.REDECLARE - 80)) | (1 << (ModelicaParser.INNER - 80)) | (1 << (ModelicaParser.OUTER - 80)) | (1 << (ModelicaParser.IDENT - 80)))) != 0):
 *                 self.state = 315
 *                 self.element()             # <<<<<<<<<<<<<<
 *                 self.state = 316
 *                 self.match(ModelicaParser.T__0)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_element); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2022, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_1 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (__pyx_t_1) {
            __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2022, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          } else {
            __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2022, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2023
 *                 self.state = 315
 *                 self.element()
 *                 self.state = 316             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__0)
 *                 self.state = 322
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_316) < 0) __PYX_ERR(0, 2023, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":2024
 *                 self.element()
 *                 self.state = 316
 *                 self.match(ModelicaParser.T__0)             # <<<<<<<<<<<<<<
 *                 self.state = 322
 *                 self._errHandler.sync(self)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2024, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2024, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2024, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_1) {
            __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2024, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_t_7};
              __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2024, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_t_7};
              __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2024, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2024, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1); __pyx_t_1 = NULL;
              __Pyx_GIVEREF(__pyx_t_7);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_7);
              __pyx_t_7 = 0;
              __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2024, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2025
 *                 self.state = 316
 *                 self.match(ModelicaParser.T__0)
 *                 self.state = 322             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_322) < 0) __PYX_ERR(0, 2025, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":2026
 *                 self.match(ModelicaParser.T__0)
 *                 self.state = 322
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 * 
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2026, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_sync); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2026, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2026, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2026, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_3);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2026, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_3);
            } else
            #endif
            {
              __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2026, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self);
              __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2026, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2027
 *                 self.state = 322
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *         except RecognitionException as re:
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2027, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_LA); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2027, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__843, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2027, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_3);
          __pyx_t_3 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":2015
 *         self.enterRule(localctx, 24, self.RULE_element_list)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 320
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "pymola/generated/ModelicaParser.pyx":2029
 *                 _la = self._input.LA(1)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2029, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.element_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_4, &__pyx_t_7) < 0) __PYX_ERR(0, 2029, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_4);
        __pyx_v_re = __pyx_t_4;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":2030
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 2030, __pyx_L25_error)

          /* "pymola/generated/ModelicaParser.pyx":2031
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2031, __pyx_L25_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2031, __pyx_L25_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2031, __pyx_L25_error)
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2031, __pyx_L25_error)
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2031, __pyx_L25_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_1) {
              __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_1); __pyx_t_1 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_13, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2031, __pyx_L25_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2032
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2032, __pyx_L25_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2032, __pyx_L25_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2032, __pyx_L25_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2032, __pyx_L25_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_1 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2032, __pyx_L25_error)
            __Pyx_GOTREF(__pyx_t_1);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2032, __pyx_L25_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":2029
 *                 _la = self._input.LA(1)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L26;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L25_error:;
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_6 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L8_except_error;
          }
          __pyx_L26:;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":2015
 *         self.enterRule(localctx, 24, self.RULE_element_list)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 320
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":2034
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2034, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (__pyx_t_3) {
        __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2034, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      } else {
        __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2034, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_14 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2034, __pyx_L32_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (__pyx_t_3) {
          __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2034, __pyx_L32_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else {
          __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2034, __pyx_L32_error)
        }
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
      __pyx_L32_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":2035
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class ElementContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2010
 * 
 * 
 *     def element_list(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Element_listContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.element_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2039
 *     class ElementContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ElementContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ElementContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ElementContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ElementContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 2039, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 2039, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2039, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ElementContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ElementContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ElementContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":2040
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 2040, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2040, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2040, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2040, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2040, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2040, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2040, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2040, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2041
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def import_clause(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 2041, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2039
 *     class ElementContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ElementContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2043
 *             self.parser = parser
 * 
 *         def import_clause(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Import_clauseContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ElementContext_3import_clause(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ElementContext_3import_clause = {"import_clause", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ElementContext_3import_clause, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ElementContext_3import_clause(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("import_clause (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ElementContext_2import_clause(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ElementContext_2import_clause(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("import_clause", 0);

  /* "pymola/generated/ModelicaParser.pyx":2044
 * 
 *         def import_clause(self):
 *             return self.getTypedRuleContext(ModelicaParser.Import_clauseContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Import_clauseContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2044, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2044, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2044, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2044, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2043
 *             self.parser = parser
 * 
 *         def import_clause(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Import_clauseContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ElementContext.import_clause", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2047
 * 
 * 
 *         def extends_clause(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Extends_clauseContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ElementContext_5extends_clause(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ElementContext_5extends_clause = {"extends_clause", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ElementContext_5extends_clause, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ElementContext_5extends_clause(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("extends_clause (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ElementContext_4extends_clause(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ElementContext_4extends_clause(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("extends_clause", 0);

  /* "pymola/generated/ModelicaParser.pyx":2048
 * 
 *         def extends_clause(self):
 *             return self.getTypedRuleContext(ModelicaParser.Extends_clauseContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2048, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2048, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Extends_clauseContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2048, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2048, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2048, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2048, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2048, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2047
 * 
 * 
 *         def extends_clause(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Extends_clauseContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ElementContext.extends_clause", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2051
 * 
 * 
 *         def regular_element(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Regular_elementContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ElementContext_7regular_element(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ElementContext_7regular_element = {"regular_element", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ElementContext_7regular_element, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ElementContext_7regular_element(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("regular_element (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ElementContext_6regular_element(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ElementContext_6regular_element(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("regular_element", 0);

  /* "pymola/generated/ModelicaParser.pyx":2052
 * 
 *         def regular_element(self):
 *             return self.getTypedRuleContext(ModelicaParser.Regular_elementContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2052, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2052, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Regular_elementContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2052, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2052, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2052, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2052, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2052, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2051
 * 
 * 
 *         def regular_element(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Regular_elementContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ElementContext.regular_element", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2055
 * 
 * 
 *         def replaceable_element(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Replaceable_elementContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ElementContext_9replaceable_element(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ElementContext_9replaceable_element = {"replaceable_element", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ElementContext_9replaceable_element, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ElementContext_9replaceable_element(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("replaceable_element (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ElementContext_8replaceable_element(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ElementContext_8replaceable_element(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("replaceable_element", 0);

  /* "pymola/generated/ModelicaParser.pyx":2056
 * 
 *         def replaceable_element(self):
 *             return self.getTypedRuleContext(ModelicaParser.Replaceable_elementContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2056, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2056, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Replaceable_elementContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2056, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2056, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2056, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2056, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2056, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2055
 * 
 * 
 *         def replaceable_element(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Replaceable_elementContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ElementContext.replaceable_element", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2059
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_element
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ElementContext_11getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ElementContext_11getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ElementContext_11getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ElementContext_11getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ElementContext_10getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ElementContext_10getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":2060
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_element             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2060, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_element); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2060, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2059
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_element
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ElementContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2062
 *             return ModelicaParser.RULE_element
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterElement" ):
 *                 listener.enterElement(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ElementContext_13enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ElementContext_13enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ElementContext_13enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ElementContext_13enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 2062, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 2062, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2062, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ElementContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ElementContext_12enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ElementContext_12enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":2063
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterElement" ):             # <<<<<<<<<<<<<<
 *                 listener.enterElement(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterElement); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 2063, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":2064
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterElement" ):
 *                 listener.enterElement(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterElement); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2064, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2064, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2064, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2064, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2064, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2064, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":2063
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterElement" ):             # <<<<<<<<<<<<<<
 *                 listener.enterElement(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":2062
 *             return ModelicaParser.RULE_element
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterElement" ):
 *                 listener.enterElement(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ElementContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2066
 *                 listener.enterElement(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitElement" ):
 *                 listener.exitElement(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ElementContext_15exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ElementContext_15exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ElementContext_15exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ElementContext_15exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 2066, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 2066, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2066, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ElementContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ElementContext_14exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ElementContext_14exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":2067
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitElement" ):             # <<<<<<<<<<<<<<
 *                 listener.exitElement(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitElement); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 2067, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":2068
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitElement" ):
 *                 listener.exitElement(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitElement); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2068, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2068, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2068, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2068, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2068, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2068, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":2067
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitElement" ):             # <<<<<<<<<<<<<<
 *                 listener.exitElement(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":2066
 *                 listener.enterElement(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitElement" ):
 *                 listener.exitElement(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ElementContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2070
 *                 listener.exitElement(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitElement" ):
 *                 return visitor.visitElement(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ElementContext_17accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ElementContext_17accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ElementContext_17accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ElementContext_17accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 2070, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 2070, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2070, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ElementContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ElementContext_16accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ElementContext_16accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":2071
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitElement" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitElement(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitElement); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 2071, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":2072
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitElement" ):
 *                 return visitor.visitElement(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitElement); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2072, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2072, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2072, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2072, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2072, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2072, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":2071
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitElement" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitElement(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":2074
 *                 return visitor.visitElement(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2074, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2074, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2074, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2074, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2074, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2074, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":2070
 *                 listener.exitElement(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitElement" ):
 *                 return visitor.visitElement(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ElementContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2079
 * 
 * 
 *     def element(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.ElementContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_29element(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_29element = {"element", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_29element, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_29element(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("element (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_28element(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_28element(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v_la_ = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  char const *__pyx_t_21;
  __Pyx_RefNannySetupContext("element", 0);

  /* "pymola/generated/ModelicaParser.pyx":2081
 *     def element(self):
 * 
 *         localctx = ModelicaParser.ElementContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 26, self.RULE_element)
 *         try:
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2081, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_ElementContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2081, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2081, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2081, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2081, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2081, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2081, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2081, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2082
 * 
 *         localctx = ModelicaParser.ElementContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 26, self.RULE_element)             # <<<<<<<<<<<<<<
 *         try:
 *             self.state = 327
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2082, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_element); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2082, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_26, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2082, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_26, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2082, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2082, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_26);
    __Pyx_GIVEREF(__pyx_int_26);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_26);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2082, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2083
 *         localctx = ModelicaParser.ElementContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 26, self.RULE_element)
 *         try:             # <<<<<<<<<<<<<<
 *             self.state = 327
 *             self._errHandler.sync(self)
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":2084
 *         self.enterRule(localctx, 26, self.RULE_element)
 *         try:
 *             self.state = 327             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             la_ = self._interp.adaptivePredict(self._input,29,self._ctx)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_327) < 0) __PYX_ERR(0, 2084, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":2085
 *         try:
 *             self.state = 327
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             la_ = self._interp.adaptivePredict(self._input,29,self._ctx)
 *             if la_ == 1:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2085, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2085, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2085, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2085, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2085, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2085, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2085, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2086
 *             self.state = 327
 *             self._errHandler.sync(self)
 *             la_ = self._interp.adaptivePredict(self._input,29,self._ctx)             # <<<<<<<<<<<<<<
 *             if la_ == 1:
 *                 self.enterOuterAlt(localctx, 1)
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_interp); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2086, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_adaptivePredict); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2086, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2086, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2086, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_t_2, __pyx_int_29, __pyx_t_3};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2086, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_t_2, __pyx_int_29, __pyx_t_3};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2086, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2086, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_2);
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_t_2);
          __Pyx_INCREF(__pyx_int_29);
          __Pyx_GIVEREF(__pyx_int_29);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_int_29);
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_6, __pyx_t_3);
          __pyx_t_2 = 0;
          __pyx_t_3 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2086, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v_la_ = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2087
 *             self._errHandler.sync(self)
 *             la_ = self._interp.adaptivePredict(self._input,29,self._ctx)
 *             if la_ == 1:             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 1)
 *                 self.state = 323
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2087, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2087, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":2088
 *             la_ = self._interp.adaptivePredict(self._input,29,self._ctx)
 *             if la_ == 1:
 *                 self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *                 self.state = 323
 *                 self.import_clause()
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2088, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_localctx, __pyx_int_1};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2088, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_localctx, __pyx_int_1};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2088, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2088, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_1);
            __Pyx_GIVEREF(__pyx_int_1);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_int_1);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2088, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2089
 *             if la_ == 1:
 *                 self.enterOuterAlt(localctx, 1)
 *                 self.state = 323             # <<<<<<<<<<<<<<
 *                 self.import_clause()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_323) < 0) __PYX_ERR(0, 2089, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":2090
 *                 self.enterOuterAlt(localctx, 1)
 *                 self.state = 323
 *                 self.import_clause()             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_import_clause); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2090, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (__pyx_t_3) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2090, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2090, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2087
 *             self._errHandler.sync(self)
 *             la_ = self._interp.adaptivePredict(self._input,29,self._ctx)
 *             if la_ == 1:             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 1)
 *                 self.state = 323
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":2093
 *                 pass
 * 
 *             elif la_ == 2:             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 2)
 *                 self.state = 324
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2093, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2093, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":2094
 * 
 *             elif la_ == 2:
 *                 self.enterOuterAlt(localctx, 2)             # <<<<<<<<<<<<<<
 *                 self.state = 324
 *                 self.extends_clause()
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2094, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2094, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2094, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2094, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_2);
            __Pyx_GIVEREF(__pyx_int_2);
            PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_int_2);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2094, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2095
 *             elif la_ == 2:
 *                 self.enterOuterAlt(localctx, 2)
 *                 self.state = 324             # <<<<<<<<<<<<<<
 *                 self.extends_clause()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_324) < 0) __PYX_ERR(0, 2095, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":2096
 *                 self.enterOuterAlt(localctx, 2)
 *                 self.state = 324
 *                 self.extends_clause()             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_extends_clause); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2096, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (__pyx_t_5) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2096, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2096, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2093
 *                 pass
 * 
 *             elif la_ == 2:             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 2)
 *                 self.state = 324
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":2099
 *                 pass
 * 
 *             elif la_ == 3:             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 3)
 *                 self.state = 325
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_3, 3, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2099, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2099, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":2100
 * 
 *             elif la_ == 3:
 *                 self.enterOuterAlt(localctx, 3)             # <<<<<<<<<<<<<<
 *                 self.state = 325
 *                 self.regular_element()
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2100, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_localctx, __pyx_int_3};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2100, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_localctx, __pyx_int_3};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2100, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2100, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_3);
            __Pyx_GIVEREF(__pyx_int_3);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_int_3);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2100, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2101
 *             elif la_ == 3:
 *                 self.enterOuterAlt(localctx, 3)
 *                 self.state = 325             # <<<<<<<<<<<<<<
 *                 self.regular_element()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_325) < 0) __PYX_ERR(0, 2101, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":2102
 *                 self.enterOuterAlt(localctx, 3)
 *                 self.state = 325
 *                 self.regular_element()             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_regular_element); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2102, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (__pyx_t_3) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2102, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2102, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2099
 *                 pass
 * 
 *             elif la_ == 3:             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 3)
 *                 self.state = 325
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":2105
 *                 pass
 * 
 *             elif la_ == 4:             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 4)
 *                 self.state = 326
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_4, 4, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2105, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2105, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":2106
 * 
 *             elif la_ == 4:
 *                 self.enterOuterAlt(localctx, 4)             # <<<<<<<<<<<<<<
 *                 self.state = 326
 *                 self.replaceable_element()
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2106, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_4};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2106, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_4};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2106, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2106, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_4);
            __Pyx_GIVEREF(__pyx_int_4);
            PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_int_4);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2106, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2107
 *             elif la_ == 4:
 *                 self.enterOuterAlt(localctx, 4)
 *                 self.state = 326             # <<<<<<<<<<<<<<
 *                 self.replaceable_element()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_326) < 0) __PYX_ERR(0, 2107, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":2108
 *                 self.enterOuterAlt(localctx, 4)
 *                 self.state = 326
 *                 self.replaceable_element()             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_replaceable_element); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2108, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (__pyx_t_5) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2108, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2108, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2105
 *                 pass
 * 
 *             elif la_ == 4:             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 4)
 *                 self.state = 326
 */
        }
        __pyx_L14:;

        /* "pymola/generated/ModelicaParser.pyx":2083
 *         localctx = ModelicaParser.ElementContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 26, self.RULE_element)
 *         try:             # <<<<<<<<<<<<<<
 *             self.state = 327
 *             self._errHandler.sync(self)
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":2112
 * 
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2112, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.element", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_7, &__pyx_t_5) < 0) __PYX_ERR(0, 2112, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_7);
        __pyx_v_re = __pyx_t_7;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":2113
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 2113, __pyx_L20_error)

          /* "pymola/generated/ModelicaParser.pyx":2114
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2114, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_reportError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2114, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2114, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2114, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_12 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 2114, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_12);
            if (__pyx_t_2) {
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_2); __pyx_t_2 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_12, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2114, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2115
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2115, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_recover); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 2115, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_12);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_12, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2115, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2115, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2115, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (__pyx_t_4) {
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2115, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":2112
 * 
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L21;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L20_error:;
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __pyx_t_6 = __pyx_lineno; __pyx_t_13 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
            }
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_ErrRestore(__pyx_t_15, __pyx_t_16, __pyx_t_17);
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_13; __pyx_filename = __pyx_t_14;
            goto __pyx_L8_except_error;
          }
          __pyx_L21:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":2083
 *         localctx = ModelicaParser.ElementContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 26, self.RULE_element)
 *         try:             # <<<<<<<<<<<<<<
 *             self.state = 327
 *             self._errHandler.sync(self)
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":2117
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2117, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2117, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_5 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2117, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_19, &__pyx_t_18);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_18);
      __pyx_t_13 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_21 = __pyx_filename;
      {
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2117, __pyx_L27_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2117, __pyx_L27_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_5 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2117, __pyx_L27_error)
        }
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __pyx_lineno = __pyx_t_13; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_21;
      goto __pyx_L1_error;
      __pyx_L27_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":2118
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Regular_elementContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2079
 * 
 * 
 *     def element(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.ElementContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.element", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v_la_);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2122
 *     class Regular_elementContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Regular_elementContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Regular_elementContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Regular_elementContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Regular_elementContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 2122, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 2122, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2122, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Regular_elementContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Regular_elementContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Regular_elementContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":2123
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 *             self.class_elem = None # Class_definitionContext
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 2123, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2123, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2123, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2124
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 *             self.class_elem = None # Class_definitionContext
 *             self.comp_elem = None # Component_clauseContext
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 2124, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2125
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 *             self.class_elem = None # Class_definitionContext             # <<<<<<<<<<<<<<
 *             self.comp_elem = None # Component_clauseContext
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_class_elem, Py_None) < 0) __PYX_ERR(0, 2125, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2126
 *             self.parser = parser
 *             self.class_elem = None # Class_definitionContext
 *             self.comp_elem = None # Component_clauseContext             # <<<<<<<<<<<<<<
 * 
 *         def REDECLARE(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_comp_elem, Py_None) < 0) __PYX_ERR(0, 2126, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2122
 *     class Regular_elementContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Regular_elementContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2128
 *             self.comp_elem = None # Component_clauseContext
 * 
 *         def REDECLARE(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.REDECLARE, 0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Regular_elementContext_3REDECLARE(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Regular_elementContext_3REDECLARE = {"REDECLARE", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Regular_elementContext_3REDECLARE, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Regular_elementContext_3REDECLARE(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("REDECLARE (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Regular_elementContext_2REDECLARE(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Regular_elementContext_2REDECLARE(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("REDECLARE", 0);

  /* "pymola/generated/ModelicaParser.pyx":2129
 * 
 *         def REDECLARE(self):
 *             return self.getToken(ModelicaParser.REDECLARE, 0)             # <<<<<<<<<<<<<<
 * 
 *         def FINAL(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getToken); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_REDECLARE); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2129, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2129, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2129, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2129, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2128
 *             self.comp_elem = None # Component_clauseContext
 * 
 *         def REDECLARE(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.REDECLARE, 0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Regular_elementContext.REDECLARE", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2131
 *             return self.getToken(ModelicaParser.REDECLARE, 0)
 * 
 *         def FINAL(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.FINAL, 0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Regular_elementContext_5FINAL(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Regular_elementContext_5FINAL = {"FINAL", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Regular_elementContext_5FINAL, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Regular_elementContext_5FINAL(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("FINAL (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Regular_elementContext_4FINAL(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Regular_elementContext_4FINAL(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("FINAL", 0);

  /* "pymola/generated/ModelicaParser.pyx":2132
 * 
 *         def FINAL(self):
 *             return self.getToken(ModelicaParser.FINAL, 0)             # <<<<<<<<<<<<<<
 * 
 *         def INNER(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getToken); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_FINAL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2132, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2132, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2132, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2132, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2131
 *             return self.getToken(ModelicaParser.REDECLARE, 0)
 * 
 *         def FINAL(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.FINAL, 0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Regular_elementContext.FINAL", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2134
 *             return self.getToken(ModelicaParser.FINAL, 0)
 * 
 *         def INNER(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.INNER, 0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Regular_elementContext_7INNER(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Regular_elementContext_7INNER = {"INNER", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Regular_elementContext_7INNER, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Regular_elementContext_7INNER(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("INNER (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Regular_elementContext_6INNER(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Regular_elementContext_6INNER(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("INNER", 0);

  /* "pymola/generated/ModelicaParser.pyx":2135
 * 
 *         def INNER(self):
 *             return self.getToken(ModelicaParser.INNER, 0)             # <<<<<<<<<<<<<<
 * 
 *         def OUTER(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getToken); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_INNER); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2135, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2135, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2134
 *             return self.getToken(ModelicaParser.FINAL, 0)
 * 
 *         def INNER(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.INNER, 0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Regular_elementContext.INNER", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2137
 *             return self.getToken(ModelicaParser.INNER, 0)
 * 
 *         def OUTER(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.OUTER, 0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Regular_elementContext_9OUTER(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Regular_elementContext_9OUTER = {"OUTER", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Regular_elementContext_9OUTER, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Regular_elementContext_9OUTER(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("OUTER (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Regular_elementContext_8OUTER(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Regular_elementContext_8OUTER(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("OUTER", 0);

  /* "pymola/generated/ModelicaParser.pyx":2138
 * 
 *         def OUTER(self):
 *             return self.getToken(ModelicaParser.OUTER, 0)             # <<<<<<<<<<<<<<
 * 
 *         def class_definition(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getToken); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_OUTER); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2138, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2138, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2138, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2138, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2137
 *             return self.getToken(ModelicaParser.INNER, 0)
 * 
 *         def OUTER(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.OUTER, 0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Regular_elementContext.OUTER", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2140
 *             return self.getToken(ModelicaParser.OUTER, 0)
 * 
 *         def class_definition(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_definitionContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Regular_elementContext_11class_definition(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Regular_elementContext_11class_definition = {"class_definition", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Regular_elementContext_11class_definition, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Regular_elementContext_11class_definition(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("class_definition (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Regular_elementContext_10class_definition(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Regular_elementContext_10class_definition(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("class_definition", 0);

  /* "pymola/generated/ModelicaParser.pyx":2141
 * 
 *         def class_definition(self):
 *             return self.getTypedRuleContext(ModelicaParser.Class_definitionContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Class_definitionContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2141, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2141, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2141, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2141, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2140
 *             return self.getToken(ModelicaParser.OUTER, 0)
 * 
 *         def class_definition(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_definitionContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Regular_elementContext.class_definition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2144
 * 
 * 
 *         def component_clause(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_clauseContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Regular_elementContext_13component_clause(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Regular_elementContext_13component_clause = {"component_clause", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Regular_elementContext_13component_clause, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Regular_elementContext_13component_clause(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("component_clause (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Regular_elementContext_12component_clause(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Regular_elementContext_12component_clause(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("component_clause", 0);

  /* "pymola/generated/ModelicaParser.pyx":2145
 * 
 *         def component_clause(self):
 *             return self.getTypedRuleContext(ModelicaParser.Component_clauseContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Component_clauseContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2145, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2145, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2145, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2145, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2144
 * 
 * 
 *         def component_clause(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_clauseContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Regular_elementContext.component_clause", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2148
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_regular_element
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Regular_elementContext_15getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Regular_elementContext_15getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Regular_elementContext_15getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Regular_elementContext_15getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Regular_elementContext_14getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Regular_elementContext_14getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":2149
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_regular_element             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_regular_element); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2148
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_regular_element
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Regular_elementContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2151
 *             return ModelicaParser.RULE_regular_element
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterRegular_element" ):
 *                 listener.enterRegular_element(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Regular_elementContext_17enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Regular_elementContext_17enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Regular_elementContext_17enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Regular_elementContext_17enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 2151, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 2151, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2151, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Regular_elementContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Regular_elementContext_16enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Regular_elementContext_16enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":2152
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterRegular_element" ):             # <<<<<<<<<<<<<<
 *                 listener.enterRegular_element(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterRegular_element); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 2152, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":2153
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterRegular_element" ):
 *                 listener.enterRegular_element(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterRegular_element); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2153, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2153, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2153, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2153, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2153, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":2152
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterRegular_element" ):             # <<<<<<<<<<<<<<
 *                 listener.enterRegular_element(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":2151
 *             return ModelicaParser.RULE_regular_element
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterRegular_element" ):
 *                 listener.enterRegular_element(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Regular_elementContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2155
 *                 listener.enterRegular_element(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitRegular_element" ):
 *                 listener.exitRegular_element(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Regular_elementContext_19exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Regular_elementContext_19exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Regular_elementContext_19exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Regular_elementContext_19exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 2155, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 2155, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2155, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Regular_elementContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Regular_elementContext_18exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Regular_elementContext_18exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":2156
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitRegular_element" ):             # <<<<<<<<<<<<<<
 *                 listener.exitRegular_element(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitRegular_element); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 2156, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":2157
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitRegular_element" ):
 *                 listener.exitRegular_element(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitRegular_element); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2157, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2157, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2157, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2157, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2157, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2157, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":2156
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitRegular_element" ):             # <<<<<<<<<<<<<<
 *                 listener.exitRegular_element(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":2155
 *                 listener.enterRegular_element(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitRegular_element" ):
 *                 listener.exitRegular_element(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Regular_elementContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2159
 *                 listener.exitRegular_element(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitRegular_element" ):
 *                 return visitor.visitRegular_element(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Regular_elementContext_21accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Regular_elementContext_21accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Regular_elementContext_21accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Regular_elementContext_21accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 2159, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 2159, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2159, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Regular_elementContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Regular_elementContext_20accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Regular_elementContext_20accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":2160
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitRegular_element" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitRegular_element(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitRegular_element); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 2160, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":2161
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitRegular_element" ):
 *                 return visitor.visitRegular_element(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitRegular_element); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2161, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2161, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2161, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2161, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2161, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2161, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":2160
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitRegular_element" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitRegular_element(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":2163
 *                 return visitor.visitRegular_element(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2163, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2163, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2163, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2163, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2163, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2163, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":2159
 *                 listener.exitRegular_element(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitRegular_element" ):
 *                 return visitor.visitRegular_element(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Regular_elementContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2168
 * 
 * 
 *     def regular_element(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Regular_elementContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_31regular_element(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_31regular_element = {"regular_element", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_31regular_element, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_31regular_element(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("regular_element (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_30regular_element(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_30regular_element(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_token = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  char const *__pyx_t_22;
  __Pyx_RefNannySetupContext("regular_element", 0);

  /* "pymola/generated/ModelicaParser.pyx":2170
 *     def regular_element(self):
 * 
 *         localctx = ModelicaParser.Regular_elementContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 28, self.RULE_regular_element)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Regular_elementContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2170, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2170, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2170, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2170, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2171
 * 
 *         localctx = ModelicaParser.Regular_elementContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 28, self.RULE_regular_element)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_regular_element); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_28, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2171, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_28, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2171, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2171, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_28);
    __Pyx_GIVEREF(__pyx_int_28);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_28);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2171, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2172
 *         localctx = ModelicaParser.Regular_elementContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 28, self.RULE_regular_element)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 2172, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2173
 *         self.enterRule(localctx, 28, self.RULE_regular_element)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 330
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":2174
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 330
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2174, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2174, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2174, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2174, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2174, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2175
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 330             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_330) < 0) __PYX_ERR(0, 2175, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":2176
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 330
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.REDECLARE:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2176, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2176, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2176, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2176, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2176, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2176, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2176, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2177
 *             self.state = 330
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.REDECLARE:
 *                 self.state = 329
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2177, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2177, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__844, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2177, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v__la = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2178
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.REDECLARE:             # <<<<<<<<<<<<<<
 *                 self.state = 329
 *                 self.match(ModelicaParser.REDECLARE)
 */
        __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2178, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_REDECLARE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2178, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2178, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2178, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":2179
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.REDECLARE:
 *                 self.state = 329             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.REDECLARE)
 * 
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_329) < 0) __PYX_ERR(0, 2179, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":2180
 *             if _la==ModelicaParser.REDECLARE:
 *                 self.state = 329
 *                 self.match(ModelicaParser.REDECLARE)             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2180, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2180, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_REDECLARE); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2180, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2180, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2180, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2180, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2180, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2180, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2178
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.REDECLARE:             # <<<<<<<<<<<<<<
 *                 self.state = 329
 *                 self.match(ModelicaParser.REDECLARE)
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":2183
 * 
 * 
 *             self.state = 333             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_333) < 0) __PYX_ERR(0, 2183, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":2184
 * 
 *             self.state = 333
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.FINAL:
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2184, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_sync); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2184, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2184, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2184, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2184, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2184, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2184, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2185
 *             self.state = 333
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.FINAL:
 *                 self.state = 332
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2185, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2185, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__845, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2185, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2186
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.FINAL:             # <<<<<<<<<<<<<<
 *                 self.state = 332
 *                 self.match(ModelicaParser.FINAL)
 */
        __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2186, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_FINAL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2186, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2186, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2186, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":2187
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.FINAL:
 *                 self.state = 332             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.FINAL)
 * 
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_332) < 0) __PYX_ERR(0, 2187, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":2188
 *             if _la==ModelicaParser.FINAL:
 *                 self.state = 332
 *                 self.match(ModelicaParser.FINAL)             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2188, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2188, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_FINAL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2188, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (!__pyx_t_3) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2188, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2188, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2188, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            {
              __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2188, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
              __Pyx_GIVEREF(__pyx_t_7);
              PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_7);
              __pyx_t_7 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2188, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2186
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.FINAL:             # <<<<<<<<<<<<<<
 *                 self.state = 332
 *                 self.match(ModelicaParser.FINAL)
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":2191
 * 
 * 
 *             self.state = 336             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_336) < 0) __PYX_ERR(0, 2191, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":2192
 * 
 *             self.state = 336
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.INNER:
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2192, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2192, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        if (!__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2192, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2192, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2192, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2192, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4); __pyx_t_4 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2192, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2193
 *             self.state = 336
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.INNER:
 *                 self.state = 335
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2193, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2193, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__846, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2193, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2194
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.INNER:             # <<<<<<<<<<<<<<
 *                 self.state = 335
 *                 self.match(ModelicaParser.INNER)
 */
        __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2194, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_INNER); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2194, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2194, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2194, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":2195
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.INNER:
 *                 self.state = 335             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.INNER)
 * 
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_335) < 0) __PYX_ERR(0, 2195, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":2196
 *             if _la==ModelicaParser.INNER:
 *                 self.state = 335
 *                 self.match(ModelicaParser.INNER)             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2196, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2196, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_INNER); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2196, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2196, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_4};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2196, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_4};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2196, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2196, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
              __Pyx_GIVEREF(__pyx_t_4);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_t_4);
              __pyx_t_4 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2196, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2194
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.INNER:             # <<<<<<<<<<<<<<
 *                 self.state = 335
 *                 self.match(ModelicaParser.INNER)
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":2199
 * 
 * 
 *             self.state = 339             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_339) < 0) __PYX_ERR(0, 2199, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":2200
 * 
 *             self.state = 339
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.OUTER:
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2200, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_sync); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2200, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_2) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2200, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2200, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2200, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2200, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2200, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2201
 *             self.state = 339
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.OUTER:
 *                 self.state = 338
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2201, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2201, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__847, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2201, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2202
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.OUTER:             # <<<<<<<<<<<<<<
 *                 self.state = 338
 *                 self.match(ModelicaParser.OUTER)
 */
        __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2202, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_OUTER); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2202, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2202, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2202, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":2203
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.OUTER:
 *                 self.state = 338             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.OUTER)
 * 
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_338) < 0) __PYX_ERR(0, 2203, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":2204
 *             if _la==ModelicaParser.OUTER:
 *                 self.state = 338
 *                 self.match(ModelicaParser.OUTER)             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2204, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2204, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_OUTER); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2204, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (!__pyx_t_4) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2204, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_2};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2204, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_2};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2204, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            } else
            #endif
            {
              __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2204, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4); __pyx_t_4 = NULL;
              __Pyx_GIVEREF(__pyx_t_2);
              PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_2);
              __pyx_t_2 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2204, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2202
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.OUTER:             # <<<<<<<<<<<<<<
 *                 self.state = 338
 *                 self.match(ModelicaParser.OUTER)
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":2207
 * 
 * 
 *             self.state = 343             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_343) < 0) __PYX_ERR(0, 2207, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":2208
 * 
 *             self.state = 343
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.T__1, ModelicaParser.T__2, ModelicaParser.T__3, ModelicaParser.T__4, ModelicaParser.T__5, ModelicaParser.T__6, ModelicaParser.T__7, ModelicaParser.T__8, ModelicaParser.T__9, ModelicaParser.T__10, ModelicaParser.T__11, ModelicaParser.T__12, ModelicaParser.PARTIAL, ModelicaParser.ENCAPSULATED]:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2208, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2208, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2208, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2208, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2208, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2208, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2208, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2209
 *             self.state = 343
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if token in [ModelicaParser.T__1, ModelicaParser.T__2, ModelicaParser.T__3, ModelicaParser.T__4, ModelicaParser.T__5, ModelicaParser.T__6, ModelicaParser.T__7, ModelicaParser.T__8, ModelicaParser.T__9, ModelicaParser.T__10, ModelicaParser.T__11, ModelicaParser.T__12, ModelicaParser.PARTIAL, ModelicaParser.ENCAPSULATED]:
 *                 self.state = 341
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2209, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2209, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__848, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2209, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v_token = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2210
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.T__1, ModelicaParser.T__2, ModelicaParser.T__3, ModelicaParser.T__4, ModelicaParser.T__5, ModelicaParser.T__6, ModelicaParser.T__7, ModelicaParser.T__8, ModelicaParser.T__9, ModelicaParser.T__10, ModelicaParser.T__11, ModelicaParser.T__12, ModelicaParser.PARTIAL, ModelicaParser.ENCAPSULATED]:             # <<<<<<<<<<<<<<
 *                 self.state = 341
 *                 localctx.class_elem = self.class_definition()
 */
        __Pyx_INCREF(__pyx_v_token);
        __pyx_t_1 = __pyx_v_token;
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L19_bool_binop_done;
        }
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L19_bool_binop_done;
        }
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L19_bool_binop_done;
        }
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L19_bool_binop_done;
        }
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L19_bool_binop_done;
        }
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L19_bool_binop_done;
        }
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L19_bool_binop_done;
        }
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L19_bool_binop_done;
        }
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L19_bool_binop_done;
        }
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L19_bool_binop_done;
        }
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__11); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L19_bool_binop_done;
        }
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__12); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L19_bool_binop_done;
        }
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_PARTIAL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L19_bool_binop_done;
        }
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_ENCAPSULATED); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2210, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_11 = __pyx_t_12;
        __pyx_L19_bool_binop_done:;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_12 = (__pyx_t_11 != 0);
        if (__pyx_t_12) {

          /* "pymola/generated/ModelicaParser.pyx":2211
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.T__1, ModelicaParser.T__2, ModelicaParser.T__3, ModelicaParser.T__4, ModelicaParser.T__5, ModelicaParser.T__6, ModelicaParser.T__7, ModelicaParser.T__8, ModelicaParser.T__9, ModelicaParser.T__10, ModelicaParser.T__11, ModelicaParser.T__12, ModelicaParser.PARTIAL, ModelicaParser.ENCAPSULATED]:
 *                 self.state = 341             # <<<<<<<<<<<<<<
 *                 localctx.class_elem = self.class_definition()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_341) < 0) __PYX_ERR(0, 2211, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":2212
 *             if token in [ModelicaParser.T__1, ModelicaParser.T__2, ModelicaParser.T__3, ModelicaParser.T__4, ModelicaParser.T__5, ModelicaParser.T__6, ModelicaParser.T__7, ModelicaParser.T__8, ModelicaParser.T__9, ModelicaParser.T__10, ModelicaParser.T__11, ModelicaParser.T__12, ModelicaParser.PARTIAL, ModelicaParser.ENCAPSULATED]:
 *                 self.state = 341
 *                 localctx.class_elem = self.class_definition()             # <<<<<<<<<<<<<<
 *                 pass
 *             elif token in [ModelicaParser.T__32, ModelicaParser.T__33, ModelicaParser.T__34, ModelicaParser.T__35, ModelicaParser.T__36, ModelicaParser.T__37, ModelicaParser.T__38, ModelicaParser.IDENT]:
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_class_definition); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2212, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2212, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2212, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_class_elem, __pyx_t_1) < 0) __PYX_ERR(0, 2212, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2210
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.T__1, ModelicaParser.T__2, ModelicaParser.T__3, ModelicaParser.T__4, ModelicaParser.T__5, ModelicaParser.T__6, ModelicaParser.T__7, ModelicaParser.T__8, ModelicaParser.T__9, ModelicaParser.T__10, ModelicaParser.T__11, ModelicaParser.T__12, ModelicaParser.PARTIAL, ModelicaParser.ENCAPSULATED]:             # <<<<<<<<<<<<<<
 *                 self.state = 341
 *                 localctx.class_elem = self.class_definition()
 */
          goto __pyx_L18;
        }

        /* "pymola/generated/ModelicaParser.pyx":2214
 *                 localctx.class_elem = self.class_definition()
 *                 pass
 *             elif token in [ModelicaParser.T__32, ModelicaParser.T__33, ModelicaParser.T__34, ModelicaParser.T__35, ModelicaParser.T__36, ModelicaParser.T__37, ModelicaParser.T__38, ModelicaParser.IDENT]:             # <<<<<<<<<<<<<<
 *                 self.state = 342
 *                 localctx.comp_elem = self.component_clause()
 */
        __Pyx_INCREF(__pyx_v_token);
        __pyx_t_1 = __pyx_v_token;
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2214, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__32); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2214, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2214, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2214, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L33_bool_binop_done;
        }
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2214, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__33); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2214, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2214, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2214, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L33_bool_binop_done;
        }
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2214, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__34); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2214, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2214, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2214, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L33_bool_binop_done;
        }
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2214, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__35); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2214, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2214, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2214, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L33_bool_binop_done;
        }
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2214, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__36); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2214, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2214, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2214, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L33_bool_binop_done;
        }
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2214, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__37); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2214, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2214, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2214, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L33_bool_binop_done;
        }
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2214, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__38); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2214, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2214, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2214, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L33_bool_binop_done;
        }
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2214, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2214, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2214, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2214, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_12 = __pyx_t_11;
        __pyx_L33_bool_binop_done:;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_11 = (__pyx_t_12 != 0);
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":2215
 *                 pass
 *             elif token in [ModelicaParser.T__32, ModelicaParser.T__33, ModelicaParser.T__34, ModelicaParser.T__35, ModelicaParser.T__36, ModelicaParser.T__37, ModelicaParser.T__38, ModelicaParser.IDENT]:
 *                 self.state = 342             # <<<<<<<<<<<<<<
 *                 localctx.comp_elem = self.component_clause()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_342) < 0) __PYX_ERR(0, 2215, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":2216
 *             elif token in [ModelicaParser.T__32, ModelicaParser.T__33, ModelicaParser.T__34, ModelicaParser.T__35, ModelicaParser.T__36, ModelicaParser.T__37, ModelicaParser.T__38, ModelicaParser.IDENT]:
 *                 self.state = 342
 *                 localctx.comp_elem = self.component_clause()             # <<<<<<<<<<<<<<
 *                 pass
 *             else:
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_component_clause); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2216, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2216, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2216, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_comp_elem, __pyx_t_1) < 0) __PYX_ERR(0, 2216, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2214
 *                 localctx.class_elem = self.class_definition()
 *                 pass
 *             elif token in [ModelicaParser.T__32, ModelicaParser.T__33, ModelicaParser.T__34, ModelicaParser.T__35, ModelicaParser.T__36, ModelicaParser.T__37, ModelicaParser.T__38, ModelicaParser.IDENT]:             # <<<<<<<<<<<<<<
 *                 self.state = 342
 *                 localctx.comp_elem = self.component_clause()
 */
          goto __pyx_L18;
        }

        /* "pymola/generated/ModelicaParser.pyx":2219
 *                 pass
 *             else:
 *                 raise NoViableAltException(self)             # <<<<<<<<<<<<<<
 * 
 *         except RecognitionException as re:
 */
        /*else*/ {
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_NoViableAltException); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2219, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2219, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2219, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2219, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2219, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2219, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_Raise(__pyx_t_1, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __PYX_ERR(0, 2219, __pyx_L6_error)
        }
        __pyx_L18:;

        /* "pymola/generated/ModelicaParser.pyx":2173
 *         self.enterRule(localctx, 28, self.RULE_regular_element)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 330
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":2221
 *                 raise NoViableAltException(self)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2221, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.regular_element", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_7, &__pyx_t_3) < 0) __PYX_ERR(0, 2221, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_7);
        __pyx_v_re = __pyx_t_7;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":2222
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 2222, __pyx_L46_error)

          /* "pymola/generated/ModelicaParser.pyx":2223
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2223, __pyx_L46_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2223, __pyx_L46_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2223, __pyx_L46_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2223, __pyx_L46_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2223, __pyx_L46_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_4) {
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_4); __pyx_t_4 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_13, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2223, __pyx_L46_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2224
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2224, __pyx_L46_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2224, __pyx_L46_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2224, __pyx_L46_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2224, __pyx_L46_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2224, __pyx_L46_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2224, __pyx_L46_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":2221
 *                 raise NoViableAltException(self)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L47;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L46_error:;
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_6 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L8_except_error;
          }
          __pyx_L47:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":2173
 *         self.enterRule(localctx, 28, self.RULE_regular_element)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 330
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":2226
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2226, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2226, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2226, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_14 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2226, __pyx_L53_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2226, __pyx_L53_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2226, __pyx_L53_error)
        }
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
      __pyx_L53_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":2227
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Replaceable_elementContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2168
 * 
 * 
 *     def regular_element(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Regular_elementContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.regular_element", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_token);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2231
 *     class Replaceable_elementContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 2231, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 2231, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2231, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Replaceable_elementContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":2232
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 *             self.class_elem = None # Class_definitionContext
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 2232, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2232, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2232, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2232, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2232, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2233
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 *             self.class_elem = None # Class_definitionContext
 *             self.comp_elem = None # Component_clauseContext
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 2233, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2234
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 *             self.class_elem = None # Class_definitionContext             # <<<<<<<<<<<<<<
 *             self.comp_elem = None # Component_clauseContext
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_class_elem, Py_None) < 0) __PYX_ERR(0, 2234, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2235
 *             self.parser = parser
 *             self.class_elem = None # Class_definitionContext
 *             self.comp_elem = None # Component_clauseContext             # <<<<<<<<<<<<<<
 * 
 *         def REDECLARE(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_comp_elem, Py_None) < 0) __PYX_ERR(0, 2235, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2231
 *     class Replaceable_elementContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Replaceable_elementContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2237
 *             self.comp_elem = None # Component_clauseContext
 * 
 *         def REDECLARE(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.REDECLARE, 0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_3REDECLARE(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_3REDECLARE = {"REDECLARE", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_3REDECLARE, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_3REDECLARE(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("REDECLARE (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_2REDECLARE(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_2REDECLARE(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("REDECLARE", 0);

  /* "pymola/generated/ModelicaParser.pyx":2238
 * 
 *         def REDECLARE(self):
 *             return self.getToken(ModelicaParser.REDECLARE, 0)             # <<<<<<<<<<<<<<
 * 
 *         def FINAL(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getToken); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_REDECLARE); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2238, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2238, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2238, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2238, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2237
 *             self.comp_elem = None # Component_clauseContext
 * 
 *         def REDECLARE(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.REDECLARE, 0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Replaceable_elementContext.REDECLARE", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2240
 *             return self.getToken(ModelicaParser.REDECLARE, 0)
 * 
 *         def FINAL(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.FINAL, 0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_5FINAL(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_5FINAL = {"FINAL", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_5FINAL, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_5FINAL(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("FINAL (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_4FINAL(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_4FINAL(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("FINAL", 0);

  /* "pymola/generated/ModelicaParser.pyx":2241
 * 
 *         def FINAL(self):
 *             return self.getToken(ModelicaParser.FINAL, 0)             # <<<<<<<<<<<<<<
 * 
 *         def INNER(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getToken); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_FINAL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2241, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2241, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2241, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2241, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2240
 *             return self.getToken(ModelicaParser.REDECLARE, 0)
 * 
 *         def FINAL(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.FINAL, 0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Replaceable_elementContext.FINAL", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2243
 *             return self.getToken(ModelicaParser.FINAL, 0)
 * 
 *         def INNER(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.INNER, 0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_7INNER(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_7INNER = {"INNER", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_7INNER, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_7INNER(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("INNER (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_6INNER(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_6INNER(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("INNER", 0);

  /* "pymola/generated/ModelicaParser.pyx":2244
 * 
 *         def INNER(self):
 *             return self.getToken(ModelicaParser.INNER, 0)             # <<<<<<<<<<<<<<
 * 
 *         def OUTER(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getToken); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_INNER); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2244, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2244, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2244, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2244, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2243
 *             return self.getToken(ModelicaParser.FINAL, 0)
 * 
 *         def INNER(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.INNER, 0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Replaceable_elementContext.INNER", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2246
 *             return self.getToken(ModelicaParser.INNER, 0)
 * 
 *         def OUTER(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.OUTER, 0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_9OUTER(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_9OUTER = {"OUTER", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_9OUTER, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_9OUTER(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("OUTER (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_8OUTER(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_8OUTER(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("OUTER", 0);

  /* "pymola/generated/ModelicaParser.pyx":2247
 * 
 *         def OUTER(self):
 *             return self.getToken(ModelicaParser.OUTER, 0)             # <<<<<<<<<<<<<<
 * 
 *         def class_definition(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getToken); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_OUTER); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2247, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2247, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2247, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2247, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2246
 *             return self.getToken(ModelicaParser.INNER, 0)
 * 
 *         def OUTER(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.OUTER, 0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Replaceable_elementContext.OUTER", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2249
 *             return self.getToken(ModelicaParser.OUTER, 0)
 * 
 *         def class_definition(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_definitionContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_11class_definition(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_11class_definition = {"class_definition", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_11class_definition, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_11class_definition(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("class_definition (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_10class_definition(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_10class_definition(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("class_definition", 0);

  /* "pymola/generated/ModelicaParser.pyx":2250
 * 
 *         def class_definition(self):
 *             return self.getTypedRuleContext(ModelicaParser.Class_definitionContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Class_definitionContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2250, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2250, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2249
 *             return self.getToken(ModelicaParser.OUTER, 0)
 * 
 *         def class_definition(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_definitionContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Replaceable_elementContext.class_definition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2253
 * 
 * 
 *         def component_clause(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_clauseContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_13component_clause(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_13component_clause = {"component_clause", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_13component_clause, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_13component_clause(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("component_clause (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_12component_clause(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_12component_clause(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("component_clause", 0);

  /* "pymola/generated/ModelicaParser.pyx":2254
 * 
 *         def component_clause(self):
 *             return self.getTypedRuleContext(ModelicaParser.Component_clauseContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Component_clauseContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2254, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2254, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2254, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2254, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2253
 * 
 * 
 *         def component_clause(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_clauseContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Replaceable_elementContext.component_clause", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2257
 * 
 * 
 *         def constraining_clause(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Constraining_clauseContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_15constraining_clause(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_15constraining_clause = {"constraining_clause", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_15constraining_clause, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_15constraining_clause(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("constraining_clause (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_14constraining_clause(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_14constraining_clause(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("constraining_clause", 0);

  /* "pymola/generated/ModelicaParser.pyx":2258
 * 
 *         def constraining_clause(self):
 *             return self.getTypedRuleContext(ModelicaParser.Constraining_clauseContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Constraining_clauseContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2258, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2258, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2258, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2258, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2257
 * 
 * 
 *         def constraining_clause(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Constraining_clauseContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Replaceable_elementContext.constraining_clause", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2261
 * 
 * 
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_17comment(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_17comment = {"comment", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_17comment, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_17comment(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("comment (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_16comment(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_16comment(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("comment", 0);

  /* "pymola/generated/ModelicaParser.pyx":2262
 * 
 *         def comment(self):
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_CommentContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2262, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2262, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2262, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2262, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2261
 * 
 * 
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Replaceable_elementContext.comment", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2265
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_replaceable_element
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_19getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_19getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_19getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_19getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_18getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_18getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":2266
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_replaceable_element             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_replaceable_element); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2265
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_replaceable_element
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Replaceable_elementContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2268
 *             return ModelicaParser.RULE_replaceable_element
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterReplaceable_element" ):
 *                 listener.enterReplaceable_element(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_21enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_21enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_21enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_21enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 2268, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 2268, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2268, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Replaceable_elementContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_20enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_20enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":2269
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterReplaceable_element" ):             # <<<<<<<<<<<<<<
 *                 listener.enterReplaceable_element(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterReplaceable_element); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 2269, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":2270
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterReplaceable_element" ):
 *                 listener.enterReplaceable_element(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterReplaceable_element); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2270, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2270, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2270, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2270, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2270, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2270, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":2269
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterReplaceable_element" ):             # <<<<<<<<<<<<<<
 *                 listener.enterReplaceable_element(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":2268
 *             return ModelicaParser.RULE_replaceable_element
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterReplaceable_element" ):
 *                 listener.enterReplaceable_element(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Replaceable_elementContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2272
 *                 listener.enterReplaceable_element(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitReplaceable_element" ):
 *                 listener.exitReplaceable_element(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_23exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_23exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_23exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_23exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 2272, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 2272, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2272, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Replaceable_elementContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_22exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_22exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":2273
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitReplaceable_element" ):             # <<<<<<<<<<<<<<
 *                 listener.exitReplaceable_element(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitReplaceable_element); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 2273, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":2274
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitReplaceable_element" ):
 *                 listener.exitReplaceable_element(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitReplaceable_element); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2274, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2274, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2274, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2274, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2274, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":2273
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitReplaceable_element" ):             # <<<<<<<<<<<<<<
 *                 listener.exitReplaceable_element(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":2272
 *                 listener.enterReplaceable_element(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitReplaceable_element" ):
 *                 listener.exitReplaceable_element(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Replaceable_elementContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2276
 *                 listener.exitReplaceable_element(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitReplaceable_element" ):
 *                 return visitor.visitReplaceable_element(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_25accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_25accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_25accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_25accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 2276, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 2276, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2276, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Replaceable_elementContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_24accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_24accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":2277
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitReplaceable_element" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitReplaceable_element(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitReplaceable_element); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 2277, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":2278
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitReplaceable_element" ):
 *                 return visitor.visitReplaceable_element(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitReplaceable_element); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2278, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2278, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2278, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2278, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2278, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2278, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":2277
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitReplaceable_element" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitReplaceable_element(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":2280
 *                 return visitor.visitReplaceable_element(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2280, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2280, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2280, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2280, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2280, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2280, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":2276
 *                 listener.exitReplaceable_element(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitReplaceable_element" ):
 *                 return visitor.visitReplaceable_element(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Replaceable_elementContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2285
 * 
 * 
 *     def replaceable_element(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Replaceable_elementContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_33replaceable_element(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_33replaceable_element = {"replaceable_element", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_33replaceable_element, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_33replaceable_element(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("replaceable_element (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_32replaceable_element(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_32replaceable_element(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_token = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  char const *__pyx_t_22;
  __Pyx_RefNannySetupContext("replaceable_element", 0);

  /* "pymola/generated/ModelicaParser.pyx":2287
 *     def replaceable_element(self):
 * 
 *         localctx = ModelicaParser.Replaceable_elementContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 30, self.RULE_replaceable_element)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Replaceable_elementContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2287, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2287, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2287, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2287, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2288
 * 
 *         localctx = ModelicaParser.Replaceable_elementContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 30, self.RULE_replaceable_element)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_replaceable_element); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_30, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2288, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_30, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2288, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2288, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_30);
    __Pyx_GIVEREF(__pyx_int_30);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_30);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2288, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2289
 *         localctx = ModelicaParser.Replaceable_elementContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 30, self.RULE_replaceable_element)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 2289, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2290
 *         self.enterRule(localctx, 30, self.RULE_replaceable_element)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 346
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":2291
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 346
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2291, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2291, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2291, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2291, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2291, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2292
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 346             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_346) < 0) __PYX_ERR(0, 2292, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":2293
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 346
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.REDECLARE:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2293, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2293, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2293, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2293, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2293, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2293, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2293, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2294
 *             self.state = 346
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.REDECLARE:
 *                 self.state = 345
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2294, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2294, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__849, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2294, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v__la = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2295
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.REDECLARE:             # <<<<<<<<<<<<<<
 *                 self.state = 345
 *                 self.match(ModelicaParser.REDECLARE)
 */
        __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2295, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_REDECLARE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2295, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2295, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2295, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":2296
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.REDECLARE:
 *                 self.state = 345             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.REDECLARE)
 * 
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_345) < 0) __PYX_ERR(0, 2296, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":2297
 *             if _la==ModelicaParser.REDECLARE:
 *                 self.state = 345
 *                 self.match(ModelicaParser.REDECLARE)             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2297, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2297, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_REDECLARE); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2297, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2297, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2297, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2297, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2297, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2297, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2295
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.REDECLARE:             # <<<<<<<<<<<<<<
 *                 self.state = 345
 *                 self.match(ModelicaParser.REDECLARE)
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":2300
 * 
 * 
 *             self.state = 349             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_349) < 0) __PYX_ERR(0, 2300, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":2301
 * 
 *             self.state = 349
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.FINAL:
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2301, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_sync); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2301, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2301, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2301, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2301, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2301, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2301, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2302
 *             self.state = 349
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.FINAL:
 *                 self.state = 348
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2302, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2302, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__850, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2302, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2303
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.FINAL:             # <<<<<<<<<<<<<<
 *                 self.state = 348
 *                 self.match(ModelicaParser.FINAL)
 */
        __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2303, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_FINAL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2303, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2303, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2303, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":2304
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.FINAL:
 *                 self.state = 348             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.FINAL)
 * 
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_348) < 0) __PYX_ERR(0, 2304, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":2305
 *             if _la==ModelicaParser.FINAL:
 *                 self.state = 348
 *                 self.match(ModelicaParser.FINAL)             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2305, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2305, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_FINAL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2305, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (!__pyx_t_3) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2305, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2305, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2305, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            {
              __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2305, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
              __Pyx_GIVEREF(__pyx_t_7);
              PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_7);
              __pyx_t_7 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2305, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2303
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.FINAL:             # <<<<<<<<<<<<<<
 *                 self.state = 348
 *                 self.match(ModelicaParser.FINAL)
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":2308
 * 
 * 
 *             self.state = 352             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_352) < 0) __PYX_ERR(0, 2308, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":2309
 * 
 *             self.state = 352
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.INNER:
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2309, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2309, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        if (!__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2309, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2309, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2309, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2309, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4); __pyx_t_4 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2309, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2310
 *             self.state = 352
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.INNER:
 *                 self.state = 351
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2310, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2310, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__851, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2310, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2311
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.INNER:             # <<<<<<<<<<<<<<
 *                 self.state = 351
 *                 self.match(ModelicaParser.INNER)
 */
        __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2311, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_INNER); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2311, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2311, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2311, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":2312
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.INNER:
 *                 self.state = 351             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.INNER)
 * 
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_351) < 0) __PYX_ERR(0, 2312, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":2313
 *             if _la==ModelicaParser.INNER:
 *                 self.state = 351
 *                 self.match(ModelicaParser.INNER)             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2313, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2313, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_INNER); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2313, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2313, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_4};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2313, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_4};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2313, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2313, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
              __Pyx_GIVEREF(__pyx_t_4);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_t_4);
              __pyx_t_4 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2313, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2311
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.INNER:             # <<<<<<<<<<<<<<
 *                 self.state = 351
 *                 self.match(ModelicaParser.INNER)
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":2316
 * 
 * 
 *             self.state = 355             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_355) < 0) __PYX_ERR(0, 2316, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":2317
 * 
 *             self.state = 355
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.OUTER:
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2317, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_sync); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2317, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_2) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2317, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2317, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2317, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2317, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2317, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2318
 *             self.state = 355
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.OUTER:
 *                 self.state = 354
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2318, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2318, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__852, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2318, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2319
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.OUTER:             # <<<<<<<<<<<<<<
 *                 self.state = 354
 *                 self.match(ModelicaParser.OUTER)
 */
        __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2319, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_OUTER); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2319, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2319, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2319, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":2320
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.OUTER:
 *                 self.state = 354             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.OUTER)
 * 
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_354) < 0) __PYX_ERR(0, 2320, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":2321
 *             if _la==ModelicaParser.OUTER:
 *                 self.state = 354
 *                 self.match(ModelicaParser.OUTER)             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2321, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2321, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_OUTER); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2321, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (!__pyx_t_4) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2321, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_2};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2321, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_2};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2321, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            } else
            #endif
            {
              __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2321, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4); __pyx_t_4 = NULL;
              __Pyx_GIVEREF(__pyx_t_2);
              PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_2);
              __pyx_t_2 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2321, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2319
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.OUTER:             # <<<<<<<<<<<<<<
 *                 self.state = 354
 *                 self.match(ModelicaParser.OUTER)
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":2324
 * 
 * 
 *             self.state = 357             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__25)
 *             self.state = 360
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_357) < 0) __PYX_ERR(0, 2324, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":2325
 * 
 *             self.state = 357
 *             self.match(ModelicaParser.T__25)             # <<<<<<<<<<<<<<
 *             self.state = 360
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2325, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2325, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__25); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2325, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2325, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2325, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2325, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2325, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2325, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2326
 *             self.state = 357
 *             self.match(ModelicaParser.T__25)
 *             self.state = 360             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_360) < 0) __PYX_ERR(0, 2326, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":2327
 *             self.match(ModelicaParser.T__25)
 *             self.state = 360
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.T__1, ModelicaParser.T__2, ModelicaParser.T__3, ModelicaParser.T__4, ModelicaParser.T__5, ModelicaParser.T__6, ModelicaParser.T__7, ModelicaParser.T__8, ModelicaParser.T__9, ModelicaParser.T__10, ModelicaParser.T__11, ModelicaParser.T__12, ModelicaParser.PARTIAL, ModelicaParser.ENCAPSULATED]:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2327, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2327, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2327, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2327, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2327, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2327, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2327, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2328
 *             self.state = 360
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if token in [ModelicaParser.T__1, ModelicaParser.T__2, ModelicaParser.T__3, ModelicaParser.T__4, ModelicaParser.T__5, ModelicaParser.T__6, ModelicaParser.T__7, ModelicaParser.T__8, ModelicaParser.T__9, ModelicaParser.T__10, ModelicaParser.T__11, ModelicaParser.T__12, ModelicaParser.PARTIAL, ModelicaParser.ENCAPSULATED]:
 *                 self.state = 358
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2328, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2328, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__853, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2328, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_v_token = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2329
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.T__1, ModelicaParser.T__2, ModelicaParser.T__3, ModelicaParser.T__4, ModelicaParser.T__5, ModelicaParser.T__6, ModelicaParser.T__7, ModelicaParser.T__8, ModelicaParser.T__9, ModelicaParser.T__10, ModelicaParser.T__11, ModelicaParser.T__12, ModelicaParser.PARTIAL, ModelicaParser.ENCAPSULATED]:             # <<<<<<<<<<<<<<
 *                 self.state = 358
 *                 localctx.class_elem = self.class_definition()
 */
        __Pyx_INCREF(__pyx_v_token);
        __pyx_t_1 = __pyx_v_token;
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L19_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L19_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L19_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L19_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L19_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L19_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L19_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L19_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L19_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L19_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__11); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L19_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__12); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L19_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_PARTIAL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L19_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_ENCAPSULATED); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2329, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __pyx_t_12;
        __pyx_L19_bool_binop_done:;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_12 = (__pyx_t_11 != 0);
        if (__pyx_t_12) {

          /* "pymola/generated/ModelicaParser.pyx":2330
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.T__1, ModelicaParser.T__2, ModelicaParser.T__3, ModelicaParser.T__4, ModelicaParser.T__5, ModelicaParser.T__6, ModelicaParser.T__7, ModelicaParser.T__8, ModelicaParser.T__9, ModelicaParser.T__10, ModelicaParser.T__11, ModelicaParser.T__12, ModelicaParser.PARTIAL, ModelicaParser.ENCAPSULATED]:
 *                 self.state = 358             # <<<<<<<<<<<<<<
 *                 localctx.class_elem = self.class_definition()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_358) < 0) __PYX_ERR(0, 2330, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":2331
 *             if token in [ModelicaParser.T__1, ModelicaParser.T__2, ModelicaParser.T__3, ModelicaParser.T__4, ModelicaParser.T__5, ModelicaParser.T__6, ModelicaParser.T__7, ModelicaParser.T__8, ModelicaParser.T__9, ModelicaParser.T__10, ModelicaParser.T__11, ModelicaParser.T__12, ModelicaParser.PARTIAL, ModelicaParser.ENCAPSULATED]:
 *                 self.state = 358
 *                 localctx.class_elem = self.class_definition()             # <<<<<<<<<<<<<<
 *                 pass
 *             elif token in [ModelicaParser.T__32, ModelicaParser.T__33, ModelicaParser.T__34, ModelicaParser.T__35, ModelicaParser.T__36, ModelicaParser.T__37, ModelicaParser.T__38, ModelicaParser.IDENT]:
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_class_definition); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2331, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2331, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2331, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_class_elem, __pyx_t_1) < 0) __PYX_ERR(0, 2331, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2329
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.T__1, ModelicaParser.T__2, ModelicaParser.T__3, ModelicaParser.T__4, ModelicaParser.T__5, ModelicaParser.T__6, ModelicaParser.T__7, ModelicaParser.T__8, ModelicaParser.T__9, ModelicaParser.T__10, ModelicaParser.T__11, ModelicaParser.T__12, ModelicaParser.PARTIAL, ModelicaParser.ENCAPSULATED]:             # <<<<<<<<<<<<<<
 *                 self.state = 358
 *                 localctx.class_elem = self.class_definition()
 */
          goto __pyx_L18;
        }

        /* "pymola/generated/ModelicaParser.pyx":2333
 *                 localctx.class_elem = self.class_definition()
 *                 pass
 *             elif token in [ModelicaParser.T__32, ModelicaParser.T__33, ModelicaParser.T__34, ModelicaParser.T__35, ModelicaParser.T__36, ModelicaParser.T__37, ModelicaParser.T__38, ModelicaParser.IDENT]:             # <<<<<<<<<<<<<<
 *                 self.state = 359
 *                 localctx.comp_elem = self.component_clause()
 */
        __Pyx_INCREF(__pyx_v_token);
        __pyx_t_1 = __pyx_v_token;
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2333, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__32); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2333, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2333, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2333, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L33_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2333, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__33); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2333, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2333, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2333, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L33_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2333, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__34); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2333, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2333, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2333, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L33_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2333, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__35); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2333, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2333, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2333, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L33_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2333, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__36); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2333, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2333, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2333, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L33_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2333, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__37); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2333, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2333, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2333, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L33_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2333, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__38); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2333, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2333, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2333, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L33_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2333, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2333, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2333, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2333, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_12 = __pyx_t_11;
        __pyx_L33_bool_binop_done:;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_11 = (__pyx_t_12 != 0);
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":2334
 *                 pass
 *             elif token in [ModelicaParser.T__32, ModelicaParser.T__33, ModelicaParser.T__34, ModelicaParser.T__35, ModelicaParser.T__36, ModelicaParser.T__37, ModelicaParser.T__38, ModelicaParser.IDENT]:
 *                 self.state = 359             # <<<<<<<<<<<<<<
 *                 localctx.comp_elem = self.component_clause()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_359) < 0) __PYX_ERR(0, 2334, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":2335
 *             elif token in [ModelicaParser.T__32, ModelicaParser.T__33, ModelicaParser.T__34, ModelicaParser.T__35, ModelicaParser.T__36, ModelicaParser.T__37, ModelicaParser.T__38, ModelicaParser.IDENT]:
 *                 self.state = 359
 *                 localctx.comp_elem = self.component_clause()             # <<<<<<<<<<<<<<
 *                 pass
 *             else:
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_component_clause); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2335, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2335, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2335, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_comp_elem, __pyx_t_1) < 0) __PYX_ERR(0, 2335, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2333
 *                 localctx.class_elem = self.class_definition()
 *                 pass
 *             elif token in [ModelicaParser.T__32, ModelicaParser.T__33, ModelicaParser.T__34, ModelicaParser.T__35, ModelicaParser.T__36, ModelicaParser.T__37, ModelicaParser.T__38, ModelicaParser.IDENT]:             # <<<<<<<<<<<<<<
 *                 self.state = 359
 *                 localctx.comp_elem = self.component_clause()
 */
          goto __pyx_L18;
        }

        /* "pymola/generated/ModelicaParser.pyx":2338
 *                 pass
 *             else:
 *                 raise NoViableAltException(self)             # <<<<<<<<<<<<<<
 * 
 *             self.state = 365
 */
        /*else*/ {
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_NoViableAltException); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2338, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2338, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2338, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2338, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2338, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2338, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_Raise(__pyx_t_1, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __PYX_ERR(0, 2338, __pyx_L6_error)
        }
        __pyx_L18:;

        /* "pymola/generated/ModelicaParser.pyx":2340
 *                 raise NoViableAltException(self)
 * 
 *             self.state = 365             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_365) < 0) __PYX_ERR(0, 2340, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":2341
 * 
 *             self.state = 365
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__31:
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2341, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_sync); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2341, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2341, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2341, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2341, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2341, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2341, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2342
 *             self.state = 365
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.T__31:
 *                 self.state = 362
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2342, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2342, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__854, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2342, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2343
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__31:             # <<<<<<<<<<<<<<
 *                 self.state = 362
 *                 self.constraining_clause()
 */
        __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2343, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__31); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2343, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2343, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2343, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":2344
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__31:
 *                 self.state = 362             # <<<<<<<<<<<<<<
 *                 self.constraining_clause()
 *                 self.state = 363
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_362) < 0) __PYX_ERR(0, 2344, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":2345
 *             if _la==ModelicaParser.T__31:
 *                 self.state = 362
 *                 self.constraining_clause()             # <<<<<<<<<<<<<<
 *                 self.state = 363
 *                 self.comment()
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_constraining_clause); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2345, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2345, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2345, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2346
 *                 self.state = 362
 *                 self.constraining_clause()
 *                 self.state = 363             # <<<<<<<<<<<<<<
 *                 self.comment()
 * 
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_363) < 0) __PYX_ERR(0, 2346, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":2347
 *                 self.constraining_clause()
 *                 self.state = 363
 *                 self.comment()             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_comment); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2347, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2347, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2347, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2343
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__31:             # <<<<<<<<<<<<<<
 *                 self.state = 362
 *                 self.constraining_clause()
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":2290
 *         self.enterRule(localctx, 30, self.RULE_replaceable_element)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 346
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":2350
 * 
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2350, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.replaceable_element", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_3, &__pyx_t_2) < 0) __PYX_ERR(0, 2350, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __pyx_v_re = __pyx_t_3;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":2351
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 2351, __pyx_L47_error)

          /* "pymola/generated/ModelicaParser.pyx":2352
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2352, __pyx_L47_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2352, __pyx_L47_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_re};
            __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2352, __pyx_L47_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_re};
            __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2352, __pyx_L47_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2352, __pyx_L47_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_7) {
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_7); __pyx_t_7 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_13, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2352, __pyx_L47_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2353
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2353, __pyx_L47_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2353, __pyx_L47_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2353, __pyx_L47_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2353, __pyx_L47_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2353, __pyx_L47_error)
            __Pyx_GOTREF(__pyx_t_7);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_7, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2353, __pyx_L47_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":2350
 * 
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L48;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L47_error:;
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_6 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L8_except_error;
          }
          __pyx_L48:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":2290
 *         self.enterRule(localctx, 30, self.RULE_replaceable_element)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 346
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":2355
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2355, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2355, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2355, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_14 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2355, __pyx_L54_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2355, __pyx_L54_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2355, __pyx_L54_error)
        }
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
      __pyx_L54_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":2356
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Import_clauseContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2285
 * 
 * 
 *     def replaceable_element(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Replaceable_elementContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.replaceable_element", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_token);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2360
 *     class Import_clauseContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Import_clauseContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20Import_clauseContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20Import_clauseContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Import_clauseContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 2360, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 2360, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2360, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Import_clauseContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20Import_clauseContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Import_clauseContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":2361
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 2361, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2361, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2361, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2361, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2361, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2361, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2361, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2361, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2362
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def comment(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 2362, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2360
 *     class Import_clauseContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Import_clauseContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2364
 *             self.parser = parser
 * 
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Import_clauseContext_3comment(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20Import_clauseContext_3comment = {"comment", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20Import_clauseContext_3comment, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Import_clauseContext_3comment(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("comment (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20Import_clauseContext_2comment(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Import_clauseContext_2comment(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("comment", 0);

  /* "pymola/generated/ModelicaParser.pyx":2365
 * 
 *         def comment(self):
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2365, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2365, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_CommentContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2365, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2365, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2365, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2365, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2365, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2364
 *             self.parser = parser
 * 
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Import_clauseContext.comment", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2368
 * 
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Import_clauseContext_5IDENT(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20Import_clauseContext_5IDENT = {"IDENT", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20Import_clauseContext_5IDENT, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Import_clauseContext_5IDENT(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("IDENT (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20Import_clauseContext_4IDENT(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Import_clauseContext_4IDENT(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("IDENT", 0);

  /* "pymola/generated/ModelicaParser.pyx":2369
 * 
 *         def IDENT(self):
 *             return self.getToken(ModelicaParser.IDENT, 0)             # <<<<<<<<<<<<<<
 * 
 *         def component_reference(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getToken); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2369, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2369, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2369, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2369, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2368
 * 
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Import_clauseContext.IDENT", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2371
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 *         def component_reference(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Import_clauseContext_7component_reference(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20Import_clauseContext_7component_reference = {"component_reference", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20Import_clauseContext_7component_reference, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Import_clauseContext_7component_reference(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("component_reference (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20Import_clauseContext_6component_reference(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Import_clauseContext_6component_reference(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("component_reference", 0);

  /* "pymola/generated/ModelicaParser.pyx":2372
 * 
 *         def component_reference(self):
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Component_referenceContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2372, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2372, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2372, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2372, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2371
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 *         def component_reference(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Import_clauseContext.component_reference", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2375
 * 
 * 
 *         def import_list(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Import_listContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Import_clauseContext_9import_list(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20Import_clauseContext_9import_list = {"import_list", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20Import_clauseContext_9import_list, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Import_clauseContext_9import_list(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("import_list (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20Import_clauseContext_8import_list(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Import_clauseContext_8import_list(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("import_list", 0);

  /* "pymola/generated/ModelicaParser.pyx":2376
 * 
 *         def import_list(self):
 *             return self.getTypedRuleContext(ModelicaParser.Import_listContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Import_listContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2376, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2376, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2376, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2376, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2375
 * 
 * 
 *         def import_list(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Import_listContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Import_clauseContext.import_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2379
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_import_clause
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Import_clauseContext_11getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20Import_clauseContext_11getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20Import_clauseContext_11getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Import_clauseContext_11getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20Import_clauseContext_10getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Import_clauseContext_10getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":2380
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_import_clause             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_import_clause); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2379
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_import_clause
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Import_clauseContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2382
 *             return ModelicaParser.RULE_import_clause
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterImport_clause" ):
 *                 listener.enterImport_clause(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Import_clauseContext_13enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20Import_clauseContext_13enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20Import_clauseContext_13enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Import_clauseContext_13enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 2382, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 2382, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2382, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Import_clauseContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20Import_clauseContext_12enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Import_clauseContext_12enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":2383
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterImport_clause" ):             # <<<<<<<<<<<<<<
 *                 listener.enterImport_clause(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterImport_clause); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 2383, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":2384
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterImport_clause" ):
 *                 listener.enterImport_clause(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterImport_clause); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2384, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2384, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2384, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2384, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2384, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2384, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":2383
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterImport_clause" ):             # <<<<<<<<<<<<<<
 *                 listener.enterImport_clause(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":2382
 *             return ModelicaParser.RULE_import_clause
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterImport_clause" ):
 *                 listener.enterImport_clause(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Import_clauseContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2386
 *                 listener.enterImport_clause(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitImport_clause" ):
 *                 listener.exitImport_clause(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Import_clauseContext_15exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20Import_clauseContext_15exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20Import_clauseContext_15exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Import_clauseContext_15exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 2386, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 2386, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2386, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Import_clauseContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20Import_clauseContext_14exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Import_clauseContext_14exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":2387
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitImport_clause" ):             # <<<<<<<<<<<<<<
 *                 listener.exitImport_clause(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitImport_clause); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 2387, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":2388
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitImport_clause" ):
 *                 listener.exitImport_clause(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitImport_clause); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2388, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2388, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2388, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2388, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2388, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2388, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":2387
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitImport_clause" ):             # <<<<<<<<<<<<<<
 *                 listener.exitImport_clause(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":2386
 *                 listener.enterImport_clause(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitImport_clause" ):
 *                 listener.exitImport_clause(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Import_clauseContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2390
 *                 listener.exitImport_clause(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitImport_clause" ):
 *                 return visitor.visitImport_clause(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Import_clauseContext_17accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20Import_clauseContext_17accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20Import_clauseContext_17accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Import_clauseContext_17accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 2390, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 2390, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2390, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Import_clauseContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20Import_clauseContext_16accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Import_clauseContext_16accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":2391
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitImport_clause" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitImport_clause(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitImport_clause); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 2391, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":2392
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitImport_clause" ):
 *                 return visitor.visitImport_clause(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitImport_clause); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2392, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2392, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2392, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2392, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2392, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2392, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":2391
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitImport_clause" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitImport_clause(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":2394
 *                 return visitor.visitImport_clause(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2394, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2394, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2394, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2394, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2394, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":2390
 *                 listener.exitImport_clause(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitImport_clause" ):
 *                 return visitor.visitImport_clause(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Import_clauseContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2399
 * 
 * 
 *     def import_clause(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Import_clauseContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_35import_clause(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_35import_clause = {"import_clause", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_35import_clause, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_35import_clause(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("import_clause (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_34import_clause(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_34import_clause(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v_la_ = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_token = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  char const *__pyx_t_22;
  __Pyx_RefNannySetupContext("import_clause", 0);

  /* "pymola/generated/ModelicaParser.pyx":2401
 *     def import_clause(self):
 * 
 *         localctx = ModelicaParser.Import_clauseContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 32, self.RULE_import_clause)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Import_clauseContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2401, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2401, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2402
 * 
 *         localctx = ModelicaParser.Import_clauseContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 32, self.RULE_import_clause)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2402, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_import_clause); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2402, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_32, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2402, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_32, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2402, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2402, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_32);
    __Pyx_GIVEREF(__pyx_int_32);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_32);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2402, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2403
 *         localctx = ModelicaParser.Import_clauseContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 32, self.RULE_import_clause)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 2403, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2404
 *         self.enterRule(localctx, 32, self.RULE_import_clause)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 367
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":2405
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 367
 *             self.match(ModelicaParser.T__26)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2405, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2405, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2405, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2405, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2405, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2406
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 367             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__26)
 *             self.state = 382
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_367) < 0) __PYX_ERR(0, 2406, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":2407
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 367
 *             self.match(ModelicaParser.T__26)             # <<<<<<<<<<<<<<
 *             self.state = 382
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2407, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2407, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__26); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2407, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2407, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2407, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2407, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2407, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2407, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2408
 *             self.state = 367
 *             self.match(ModelicaParser.T__26)
 *             self.state = 382             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             la_ = self._interp.adaptivePredict(self._input,43,self._ctx)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_382) < 0) __PYX_ERR(0, 2408, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":2409
 *             self.match(ModelicaParser.T__26)
 *             self.state = 382
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             la_ = self._interp.adaptivePredict(self._input,43,self._ctx)
 *             if la_ == 1:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2409, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2409, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2409, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2409, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2409, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2409, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2409, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2410
 *             self.state = 382
 *             self._errHandler.sync(self)
 *             la_ = self._interp.adaptivePredict(self._input,43,self._ctx)             # <<<<<<<<<<<<<<
 *             if la_ == 1:
 *                 self.state = 368
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_interp); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2410, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_adaptivePredict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2410, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2410, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2410, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_2)) {
          PyObject *__pyx_temp[4] = {__pyx_t_7, __pyx_t_4, __pyx_int_43, __pyx_t_3};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2410, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
          PyObject *__pyx_temp[4] = {__pyx_t_7, __pyx_t_4, __pyx_int_43, __pyx_t_3};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2410, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2410, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_7) {
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_t_4);
          __Pyx_INCREF(__pyx_int_43);
          __Pyx_GIVEREF(__pyx_int_43);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_int_43);
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_6, __pyx_t_3);
          __pyx_t_4 = 0;
          __pyx_t_3 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2410, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_v_la_ = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2411
 *             self._errHandler.sync(self)
 *             la_ = self._interp.adaptivePredict(self._input,43,self._ctx)
 *             if la_ == 1:             # <<<<<<<<<<<<<<
 *                 self.state = 368
 *                 self.match(ModelicaParser.IDENT)
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2411, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2411, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":2412
 *             la_ = self._interp.adaptivePredict(self._input,43,self._ctx)
 *             if la_ == 1:
 *                 self.state = 368             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.IDENT)
 *                 self.state = 369
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_368) < 0) __PYX_ERR(0, 2412, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":2413
 *             if la_ == 1:
 *                 self.state = 368
 *                 self.match(ModelicaParser.IDENT)             # <<<<<<<<<<<<<<
 *                 self.state = 369
 *                 self.match(ModelicaParser.T__14)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2413, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2413, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2413, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_5) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2413, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2413, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2413, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2413, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2413, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2414
 *                 self.state = 368
 *                 self.match(ModelicaParser.IDENT)
 *                 self.state = 369             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__14)
 *                 self.state = 370
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_369) < 0) __PYX_ERR(0, 2414, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":2415
 *                 self.match(ModelicaParser.IDENT)
 *                 self.state = 369
 *                 self.match(ModelicaParser.T__14)             # <<<<<<<<<<<<<<
 *                 self.state = 370
 *                 self.component_reference()
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2415, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2415, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__14); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2415, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_4) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2415, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2415, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2415, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2415, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2415, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2416
 *                 self.state = 369
 *                 self.match(ModelicaParser.T__14)
 *                 self.state = 370             # <<<<<<<<<<<<<<
 *                 self.component_reference()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_370) < 0) __PYX_ERR(0, 2416, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":2417
 *                 self.match(ModelicaParser.T__14)
 *                 self.state = 370
 *                 self.component_reference()             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_component_reference); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2417, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (__pyx_t_5) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2417, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2417, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2411
 *             self._errHandler.sync(self)
 *             la_ = self._interp.adaptivePredict(self._input,43,self._ctx)
 *             if la_ == 1:             # <<<<<<<<<<<<<<
 *                 self.state = 368
 *                 self.match(ModelicaParser.IDENT)
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":2420
 *                 pass
 * 
 *             elif la_ == 2:             # <<<<<<<<<<<<<<
 *                 self.state = 371
 *                 self.component_reference()
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2420, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2420, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":2421
 * 
 *             elif la_ == 2:
 *                 self.state = 371             # <<<<<<<<<<<<<<
 *                 self.component_reference()
 *                 self.state = 380
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_371) < 0) __PYX_ERR(0, 2421, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":2422
 *             elif la_ == 2:
 *                 self.state = 371
 *                 self.component_reference()             # <<<<<<<<<<<<<<
 *                 self.state = 380
 *                 self._errHandler.sync(self)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_component_reference); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2422, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (__pyx_t_5) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2422, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2422, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2423
 *                 self.state = 371
 *                 self.component_reference()
 *                 self.state = 380             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_380) < 0) __PYX_ERR(0, 2423, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":2424
 *                 self.component_reference()
 *                 self.state = 380
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__27:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2424, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_sync); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2424, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2424, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_5)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2424, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2424, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2424, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2424, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2425
 *                 self.state = 380
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if _la==ModelicaParser.T__27:
 *                     self.state = 372
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2425, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2425, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_tuple__855, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2425, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_v__la = __pyx_t_1;
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2426
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__27:             # <<<<<<<<<<<<<<
 *                     self.state = 372
 *                     self.match(ModelicaParser.T__27)
 */
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2426, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__27); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2426, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2426, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2426, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (__pyx_t_11) {

            /* "pymola/generated/ModelicaParser.pyx":2427
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__27:
 *                     self.state = 372             # <<<<<<<<<<<<<<
 *                     self.match(ModelicaParser.T__27)
 *                     self.state = 378
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_372) < 0) __PYX_ERR(0, 2427, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":2428
 *                 if _la==ModelicaParser.T__27:
 *                     self.state = 372
 *                     self.match(ModelicaParser.T__27)             # <<<<<<<<<<<<<<
 *                     self.state = 378
 *                     self._errHandler.sync(self)
 */
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2428, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2428, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__27); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2428, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
              __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
              if (likely(__pyx_t_3)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                __Pyx_INCREF(__pyx_t_3);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_5, function);
              }
            }
            if (!__pyx_t_3) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2428, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_5)) {
                PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_2};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2428, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
                PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_2};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2428, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              } else
              #endif
              {
                __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2428, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __pyx_t_3 = NULL;
                __Pyx_GIVEREF(__pyx_t_2);
                PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_2);
                __pyx_t_2 = 0;
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2428, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":2429
 *                     self.state = 372
 *                     self.match(ModelicaParser.T__27)
 *                     self.state = 378             # <<<<<<<<<<<<<<
 *                     self._errHandler.sync(self)
 *                     token = self._input.LA(1)
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_378) < 0) __PYX_ERR(0, 2429, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":2430
 *                     self.match(ModelicaParser.T__27)
 *                     self.state = 378
 *                     self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                     token = self._input.LA(1)
 *                     if token in [ModelicaParser.T__28]:
 */
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2430, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_sync); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2430, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_5 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
              if (likely(__pyx_t_5)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                __Pyx_INCREF(__pyx_t_5);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_4, function);
              }
            }
            if (!__pyx_t_5) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2430, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_4)) {
                PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2430, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
                PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2430, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else
              #endif
              {
                __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2430, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5); __pyx_t_5 = NULL;
                __Pyx_INCREF(__pyx_v_self);
                __Pyx_GIVEREF(__pyx_v_self);
                PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2430, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":2431
 *                     self.state = 378
 *                     self._errHandler.sync(self)
 *                     token = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                     if token in [ModelicaParser.T__28]:
 *                         self.state = 373
 */
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2431, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2431, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__856, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2431, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_v_token = __pyx_t_1;
            __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":2432
 *                     self._errHandler.sync(self)
 *                     token = self._input.LA(1)
 *                     if token in [ModelicaParser.T__28]:             # <<<<<<<<<<<<<<
 *                         self.state = 373
 *                         self.match(ModelicaParser.T__28)
 */
            __Pyx_INCREF(__pyx_v_token);
            __pyx_t_1 = __pyx_v_token;
            __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2432, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__28); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2432, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2432, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2432, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_12 = (__pyx_t_11 != 0);
            if (__pyx_t_12) {

              /* "pymola/generated/ModelicaParser.pyx":2433
 *                     token = self._input.LA(1)
 *                     if token in [ModelicaParser.T__28]:
 *                         self.state = 373             # <<<<<<<<<<<<<<
 *                         self.match(ModelicaParser.T__28)
 *                         pass
 */
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_373) < 0) __PYX_ERR(0, 2433, __pyx_L6_error)

              /* "pymola/generated/ModelicaParser.pyx":2434
 *                     if token in [ModelicaParser.T__28]:
 *                         self.state = 373
 *                         self.match(ModelicaParser.T__28)             # <<<<<<<<<<<<<<
 *                         pass
 *                     elif token in [ModelicaParser.T__29]:
 */
              __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2434, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2434, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__28); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2434, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __pyx_t_2 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
                __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
                if (likely(__pyx_t_2)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                  __Pyx_INCREF(__pyx_t_2);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_4, function);
                }
              }
              if (!__pyx_t_2) {
                __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2434, __pyx_L6_error)
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else {
                #if CYTHON_FAST_PYCALL
                if (PyFunction_Check(__pyx_t_4)) {
                  PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_5};
                  __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2434, __pyx_L6_error)
                  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
                  __Pyx_GOTREF(__pyx_t_1);
                  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                } else
                #endif
                #if CYTHON_FAST_PYCCALL
                if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
                  PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_5};
                  __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2434, __pyx_L6_error)
                  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
                  __Pyx_GOTREF(__pyx_t_1);
                  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                } else
                #endif
                {
                  __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2434, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_3);
                  __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
                  __Pyx_GIVEREF(__pyx_t_5);
                  PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_t_5);
                  __pyx_t_5 = 0;
                  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2434, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_1);
                  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                }
              }
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "pymola/generated/ModelicaParser.pyx":2432
 *                     self._errHandler.sync(self)
 *                     token = self._input.LA(1)
 *                     if token in [ModelicaParser.T__28]:             # <<<<<<<<<<<<<<
 *                         self.state = 373
 *                         self.match(ModelicaParser.T__28)
 */
              goto __pyx_L16;
            }

            /* "pymola/generated/ModelicaParser.pyx":2436
 *                         self.match(ModelicaParser.T__28)
 *                         pass
 *                     elif token in [ModelicaParser.T__29]:             # <<<<<<<<<<<<<<
 *                         self.state = 374
 *                         self.match(ModelicaParser.T__29)
 */
            __Pyx_INCREF(__pyx_v_token);
            __pyx_t_1 = __pyx_v_token;
            __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2436, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__29); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2436, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2436, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2436, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_11 = (__pyx_t_12 != 0);
            if (__pyx_t_11) {

              /* "pymola/generated/ModelicaParser.pyx":2437
 *                         pass
 *                     elif token in [ModelicaParser.T__29]:
 *                         self.state = 374             # <<<<<<<<<<<<<<
 *                         self.match(ModelicaParser.T__29)
 *                         self.state = 375
 */
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_374) < 0) __PYX_ERR(0, 2437, __pyx_L6_error)

              /* "pymola/generated/ModelicaParser.pyx":2438
 *                     elif token in [ModelicaParser.T__29]:
 *                         self.state = 374
 *                         self.match(ModelicaParser.T__29)             # <<<<<<<<<<<<<<
 *                         self.state = 375
 *                         self.import_list()
 */
              __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2438, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2438, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__29); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2438, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __pyx_t_3 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
                __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
                if (likely(__pyx_t_3)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                  __Pyx_INCREF(__pyx_t_3);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_4, function);
                }
              }
              if (!__pyx_t_3) {
                __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2438, __pyx_L6_error)
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else {
                #if CYTHON_FAST_PYCALL
                if (PyFunction_Check(__pyx_t_4)) {
                  PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_5};
                  __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2438, __pyx_L6_error)
                  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                  __Pyx_GOTREF(__pyx_t_1);
                  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                } else
                #endif
                #if CYTHON_FAST_PYCCALL
                if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
                  PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_5};
                  __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2438, __pyx_L6_error)
                  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                  __Pyx_GOTREF(__pyx_t_1);
                  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                } else
                #endif
                {
                  __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2438, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_2);
                  __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
                  __Pyx_GIVEREF(__pyx_t_5);
                  PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_5);
                  __pyx_t_5 = 0;
                  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2438, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_1);
                  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                }
              }
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "pymola/generated/ModelicaParser.pyx":2439
 *                         self.state = 374
 *                         self.match(ModelicaParser.T__29)
 *                         self.state = 375             # <<<<<<<<<<<<<<
 *                         self.import_list()
 *                         self.state = 376
 */
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_375) < 0) __PYX_ERR(0, 2439, __pyx_L6_error)

              /* "pymola/generated/ModelicaParser.pyx":2440
 *                         self.match(ModelicaParser.T__29)
 *                         self.state = 375
 *                         self.import_list()             # <<<<<<<<<<<<<<
 *                         self.state = 376
 *                         self.match(ModelicaParser.T__30)
 */
              __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_import_list); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2440, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_2 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
                __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
                if (likely(__pyx_t_2)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                  __Pyx_INCREF(__pyx_t_2);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_4, function);
                }
              }
              if (__pyx_t_2) {
                __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2440, __pyx_L6_error)
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              } else {
                __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2440, __pyx_L6_error)
              }
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "pymola/generated/ModelicaParser.pyx":2441
 *                         self.state = 375
 *                         self.import_list()
 *                         self.state = 376             # <<<<<<<<<<<<<<
 *                         self.match(ModelicaParser.T__30)
 *                         pass
 */
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_376) < 0) __PYX_ERR(0, 2441, __pyx_L6_error)

              /* "pymola/generated/ModelicaParser.pyx":2442
 *                         self.import_list()
 *                         self.state = 376
 *                         self.match(ModelicaParser.T__30)             # <<<<<<<<<<<<<<
 *                         pass
 *                     else:
 */
              __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2442, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2442, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__30); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2442, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __pyx_t_2 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
                __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
                if (likely(__pyx_t_2)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                  __Pyx_INCREF(__pyx_t_2);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_4, function);
                }
              }
              if (!__pyx_t_2) {
                __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2442, __pyx_L6_error)
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else {
                #if CYTHON_FAST_PYCALL
                if (PyFunction_Check(__pyx_t_4)) {
                  PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_5};
                  __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2442, __pyx_L6_error)
                  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
                  __Pyx_GOTREF(__pyx_t_1);
                  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                } else
                #endif
                #if CYTHON_FAST_PYCCALL
                if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
                  PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_5};
                  __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2442, __pyx_L6_error)
                  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
                  __Pyx_GOTREF(__pyx_t_1);
                  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                } else
                #endif
                {
                  __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2442, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_3);
                  __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
                  __Pyx_GIVEREF(__pyx_t_5);
                  PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_t_5);
                  __pyx_t_5 = 0;
                  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2442, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_1);
                  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                }
              }
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "pymola/generated/ModelicaParser.pyx":2436
 *                         self.match(ModelicaParser.T__28)
 *                         pass
 *                     elif token in [ModelicaParser.T__29]:             # <<<<<<<<<<<<<<
 *                         self.state = 374
 *                         self.match(ModelicaParser.T__29)
 */
              goto __pyx_L16;
            }

            /* "pymola/generated/ModelicaParser.pyx":2445
 *                         pass
 *                     else:
 *                         raise NoViableAltException(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
            /*else*/ {
              __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_NoViableAltException); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2445, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_3 = NULL;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
                __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
                if (likely(__pyx_t_3)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                  __Pyx_INCREF(__pyx_t_3);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_4, function);
                }
              }
              if (!__pyx_t_3) {
                __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2445, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
              } else {
                #if CYTHON_FAST_PYCALL
                if (PyFunction_Check(__pyx_t_4)) {
                  PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
                  __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2445, __pyx_L6_error)
                  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                  __Pyx_GOTREF(__pyx_t_1);
                } else
                #endif
                #if CYTHON_FAST_PYCCALL
                if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
                  PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
                  __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2445, __pyx_L6_error)
                  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                  __Pyx_GOTREF(__pyx_t_1);
                } else
                #endif
                {
                  __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2445, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_5);
                  __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
                  __Pyx_INCREF(__pyx_v_self);
                  __Pyx_GIVEREF(__pyx_v_self);
                  PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
                  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2445, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_1);
                  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                }
              }
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_Raise(__pyx_t_1, 0, 0, 0);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __PYX_ERR(0, 2445, __pyx_L6_error)
            }
            __pyx_L16:;

            /* "pymola/generated/ModelicaParser.pyx":2426
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__27:             # <<<<<<<<<<<<<<
 *                     self.state = 372
 *                     self.match(ModelicaParser.T__27)
 */
          }

          /* "pymola/generated/ModelicaParser.pyx":2420
 *                 pass
 * 
 *             elif la_ == 2:             # <<<<<<<<<<<<<<
 *                 self.state = 371
 *                 self.component_reference()
 */
        }
        __pyx_L14:;

        /* "pymola/generated/ModelicaParser.pyx":2452
 * 
 * 
 *             self.state = 384             # <<<<<<<<<<<<<<
 *             self.comment()
 *         except RecognitionException as re:
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_384) < 0) __PYX_ERR(0, 2452, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":2453
 * 
 *             self.state = 384
 *             self.comment()             # <<<<<<<<<<<<<<
 *         except RecognitionException as re:
 *             localctx.exception = re
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_comment); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2453, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (__pyx_t_5) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2453, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2453, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2404
 *         self.enterRule(localctx, 32, self.RULE_import_clause)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 367
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":2454
 *             self.state = 384
 *             self.comment()
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2454, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.import_clause", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_4, &__pyx_t_5) < 0) __PYX_ERR(0, 2454, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __pyx_v_re = __pyx_t_4;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":2455
 *             self.comment()
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 2455, __pyx_L22_error)

          /* "pymola/generated/ModelicaParser.pyx":2456
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2456, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_reportError); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2456, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2456, __pyx_L22_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2456, __pyx_L22_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2456, __pyx_L22_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_2) {
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_2); __pyx_t_2 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_13, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2456, __pyx_L22_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2457
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2457, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_recover); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2457, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2457, __pyx_L22_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2457, __pyx_L22_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2457, __pyx_L22_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (__pyx_t_7) {
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_7); __pyx_t_7 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2457, __pyx_L22_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":2454
 *             self.state = 384
 *             self.comment()
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L23;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L22_error:;
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_6 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L8_except_error;
          }
          __pyx_L23:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":2404
 *         self.enterRule(localctx, 32, self.RULE_import_clause)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 367
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":2459
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2459, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2459, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_5 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2459, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_14 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2459, __pyx_L29_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2459, __pyx_L29_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_5 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2459, __pyx_L29_error)
        }
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
      __pyx_L29_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":2460
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Import_listContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2399
 * 
 * 
 *     def import_clause(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Import_clauseContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.import_clause", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v_la_);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_token);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2464
 *     class Import_listContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Import_listContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18Import_listContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18Import_listContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Import_listContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 2464, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 2464, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2464, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Import_listContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18Import_listContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Import_listContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":2465
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 2465, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2465, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2465, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2465, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2465, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2465, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2465, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2465, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2466
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def IDENT(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 2466, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2464
 *     class Import_listContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Import_listContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2468
 *             self.parser = parser
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Import_listContext_3IDENT(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18Import_listContext_3IDENT = {"IDENT", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18Import_listContext_3IDENT, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Import_listContext_3IDENT(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("IDENT (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18Import_listContext_2IDENT(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Import_listContext_2IDENT(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("IDENT", 0);

  /* "pymola/generated/ModelicaParser.pyx":2469
 * 
 *         def IDENT(self):
 *             return self.getToken(ModelicaParser.IDENT, 0)             # <<<<<<<<<<<<<<
 * 
 *         def import_list(self, i:int=None):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getToken); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2469, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2469, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2469, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2469, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2468
 *             self.parser = parser
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Import_listContext.IDENT", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2471
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 *         def import_list(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Import_listContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Import_listContext_5import_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18Import_listContext_5import_list = {"import_list", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18Import_listContext_5import_list, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Import_listContext_5import_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("import_list (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "import_list") < 0)) __PYX_ERR(0, 2471, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("import_list", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2471, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Import_listContext.import_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18Import_listContext_4import_list(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Import_listContext_4import_list(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("import_list", 0);

  /* "pymola/generated/ModelicaParser.pyx":2472
 * 
 *         def import_list(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.Import_listContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":2473
 *         def import_list(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Import_listContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.Import_listContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2473, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2473, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_Import_listContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2473, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2473, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2473, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2473, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2473, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2473, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":2472
 * 
 *         def import_list(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.Import_listContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":2475
 *                 return self.getTypedRuleContexts(ModelicaParser.Import_listContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.Import_listContext,i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2475, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2475, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Import_listContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2475, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2475, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2475, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2475, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2475, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":2471
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 *         def import_list(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Import_listContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Import_listContext.import_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2478
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_import_list
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Import_listContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18Import_listContext_7getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18Import_listContext_7getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Import_listContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18Import_listContext_6getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Import_listContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":2479
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_import_list             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2479, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_import_list); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2479, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2478
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_import_list
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Import_listContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2481
 *             return ModelicaParser.RULE_import_list
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterImport_list" ):
 *                 listener.enterImport_list(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Import_listContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18Import_listContext_9enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18Import_listContext_9enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Import_listContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 2481, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 2481, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2481, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Import_listContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18Import_listContext_8enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Import_listContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":2482
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterImport_list" ):             # <<<<<<<<<<<<<<
 *                 listener.enterImport_list(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterImport_list); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 2482, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":2483
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterImport_list" ):
 *                 listener.enterImport_list(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterImport_list); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2483, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2483, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2483, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2483, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2483, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2483, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":2482
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterImport_list" ):             # <<<<<<<<<<<<<<
 *                 listener.enterImport_list(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":2481
 *             return ModelicaParser.RULE_import_list
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterImport_list" ):
 *                 listener.enterImport_list(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Import_listContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2485
 *                 listener.enterImport_list(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitImport_list" ):
 *                 listener.exitImport_list(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Import_listContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18Import_listContext_11exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18Import_listContext_11exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Import_listContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 2485, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 2485, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2485, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Import_listContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18Import_listContext_10exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Import_listContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":2486
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitImport_list" ):             # <<<<<<<<<<<<<<
 *                 listener.exitImport_list(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitImport_list); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 2486, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":2487
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitImport_list" ):
 *                 listener.exitImport_list(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitImport_list); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2487, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2487, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2487, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2487, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2487, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2487, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":2486
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitImport_list" ):             # <<<<<<<<<<<<<<
 *                 listener.exitImport_list(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":2485
 *                 listener.enterImport_list(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitImport_list" ):
 *                 listener.exitImport_list(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Import_listContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2489
 *                 listener.exitImport_list(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitImport_list" ):
 *                 return visitor.visitImport_list(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Import_listContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18Import_listContext_13accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18Import_listContext_13accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Import_listContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 2489, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 2489, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2489, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Import_listContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18Import_listContext_12accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Import_listContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":2490
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitImport_list" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitImport_list(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitImport_list); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 2490, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":2491
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitImport_list" ):
 *                 return visitor.visitImport_list(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitImport_list); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2491, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2491, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2491, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2491, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2491, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2491, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":2490
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitImport_list" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitImport_list(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":2493
 *                 return visitor.visitImport_list(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2493, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2493, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2493, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2493, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2493, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2493, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":2489
 *                 listener.exitImport_list(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitImport_list" ):
 *                 return visitor.visitImport_list(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Import_listContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2498
 * 
 * 
 *     def import_list(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Import_listContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_37import_list(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_37import_list = {"import_list", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_37import_list, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_37import_list(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("import_list (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_36import_list(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_36import_list(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__alt = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  char const *__pyx_t_22;
  __Pyx_RefNannySetupContext("import_list", 0);

  /* "pymola/generated/ModelicaParser.pyx":2500
 *     def import_list(self):
 * 
 *         localctx = ModelicaParser.Import_listContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 34, self.RULE_import_list)
 *         try:
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Import_listContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2500, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2500, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2500, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2500, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2501
 * 
 *         localctx = ModelicaParser.Import_listContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 34, self.RULE_import_list)             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_import_list); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_34, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2501, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_34, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2501, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2501, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_34);
    __Pyx_GIVEREF(__pyx_int_34);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_34);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2501, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2502
 *         localctx = ModelicaParser.Import_listContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 34, self.RULE_import_list)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 386
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":2503
 *         self.enterRule(localctx, 34, self.RULE_import_list)
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 386
 *             self.match(ModelicaParser.IDENT)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2503, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2503, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2503, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2503, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2503, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2504
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 386             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.IDENT)
 *             self.state = 391
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_386) < 0) __PYX_ERR(0, 2504, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":2505
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 386
 *             self.match(ModelicaParser.IDENT)             # <<<<<<<<<<<<<<
 *             self.state = 391
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2505, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2505, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2505, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2505, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2505, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2505, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2505, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2505, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2506
 *             self.state = 386
 *             self.match(ModelicaParser.IDENT)
 *             self.state = 391             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _alt = self._interp.adaptivePredict(self._input,44,self._ctx)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_391) < 0) __PYX_ERR(0, 2506, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":2507
 *             self.match(ModelicaParser.IDENT)
 *             self.state = 391
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _alt = self._interp.adaptivePredict(self._input,44,self._ctx)
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2507, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2507, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2507, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2507, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2507, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2507, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2507, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2508
 *             self.state = 391
 *             self._errHandler.sync(self)
 *             _alt = self._interp.adaptivePredict(self._input,44,self._ctx)             # <<<<<<<<<<<<<<
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                 if _alt==1:
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_interp); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2508, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_adaptivePredict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2508, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2508, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2508, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_2)) {
          PyObject *__pyx_temp[4] = {__pyx_t_7, __pyx_t_4, __pyx_int_44, __pyx_t_3};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2508, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
          PyObject *__pyx_temp[4] = {__pyx_t_7, __pyx_t_4, __pyx_int_44, __pyx_t_3};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2508, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2508, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_7) {
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_t_4);
          __Pyx_INCREF(__pyx_int_44);
          __Pyx_GIVEREF(__pyx_int_44);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_int_44);
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_6, __pyx_t_3);
          __pyx_t_4 = 0;
          __pyx_t_3 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2508, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_v__alt = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2509
 *             self._errHandler.sync(self)
 *             _alt = self._interp.adaptivePredict(self._input,44,self._ctx)
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:             # <<<<<<<<<<<<<<
 *                 if _alt==1:
 *                     self.state = 387
 */
        while (1) {
          __pyx_t_1 = PyObject_RichCompare(__pyx_v__alt, __pyx_int_2, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2509, __pyx_L6_error)
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2509, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (__pyx_t_12) {
          } else {
            __pyx_t_11 = __pyx_t_12;
            goto __pyx_L16_bool_binop_done;
          }
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ATN); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2509, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_INVALID_ALT_NUMBER); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2509, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyObject_RichCompare(__pyx_v__alt, __pyx_t_2, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2509, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2509, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_11 = __pyx_t_12;
          __pyx_L16_bool_binop_done:;
          if (!__pyx_t_11) break;

          /* "pymola/generated/ModelicaParser.pyx":2510
 *             _alt = self._interp.adaptivePredict(self._input,44,self._ctx)
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                 if _alt==1:             # <<<<<<<<<<<<<<
 *                     self.state = 387
 *                     self.match(ModelicaParser.T__20)
 */
          __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v__alt, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2510, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2510, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (__pyx_t_11) {

            /* "pymola/generated/ModelicaParser.pyx":2511
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                 if _alt==1:
 *                     self.state = 387             # <<<<<<<<<<<<<<
 *                     self.match(ModelicaParser.T__20)
 *                     self.state = 388
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_387) < 0) __PYX_ERR(0, 2511, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":2512
 *                 if _alt==1:
 *                     self.state = 387
 *                     self.match(ModelicaParser.T__20)             # <<<<<<<<<<<<<<
 *                     self.state = 388
 *                     self.import_list()
 */
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2512, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2512, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_T__20); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2512, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_5 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_5)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_5);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
              }
            }
            if (!__pyx_t_5) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2512, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_2)) {
                PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2512, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
                PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2512, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              } else
              #endif
              {
                __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2512, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
                __Pyx_GIVEREF(__pyx_t_3);
                PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_3);
                __pyx_t_3 = 0;
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2512, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":2513
 *                     self.state = 387
 *                     self.match(ModelicaParser.T__20)
 *                     self.state = 388             # <<<<<<<<<<<<<<
 *                     self.import_list()
 *                 self.state = 393
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_388) < 0) __PYX_ERR(0, 2513, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":2514
 *                     self.match(ModelicaParser.T__20)
 *                     self.state = 388
 *                     self.import_list()             # <<<<<<<<<<<<<<
 *                 self.state = 393
 *                 self._errHandler.sync(self)
 */
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_import_list); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2514, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_4 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_4)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_4);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
              }
            }
            if (__pyx_t_4) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2514, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            } else {
              __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2514, __pyx_L6_error)
            }
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":2510
 *             _alt = self._interp.adaptivePredict(self._input,44,self._ctx)
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                 if _alt==1:             # <<<<<<<<<<<<<<
 *                     self.state = 387
 *                     self.match(ModelicaParser.T__20)
 */
          }

          /* "pymola/generated/ModelicaParser.pyx":2515
 *                     self.state = 388
 *                     self.import_list()
 *                 self.state = 393             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _alt = self._interp.adaptivePredict(self._input,44,self._ctx)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_393) < 0) __PYX_ERR(0, 2515, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":2516
 *                     self.import_list()
 *                 self.state = 393
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _alt = self._interp.adaptivePredict(self._input,44,self._ctx)
 * 
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2516, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_sync); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2516, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2516, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2516, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2516, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2516, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2516, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2517
 *                 self.state = 393
 *                 self._errHandler.sync(self)
 *                 _alt = self._interp.adaptivePredict(self._input,44,self._ctx)             # <<<<<<<<<<<<<<
 * 
 *         except RecognitionException as re:
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_interp); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2517, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_adaptivePredict); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2517, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2517, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2517, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_t_4, __pyx_int_44, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2517, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_t_4, __pyx_int_44, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2517, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2517, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_GIVEREF(__pyx_t_4);
            PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_t_4);
            __Pyx_INCREF(__pyx_int_44);
            __Pyx_GIVEREF(__pyx_int_44);
            PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_44);
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_2);
            __pyx_t_4 = 0;
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2517, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF_SET(__pyx_v__alt, __pyx_t_1);
          __pyx_t_1 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":2502
 *         localctx = ModelicaParser.Import_listContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 34, self.RULE_import_list)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 386
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":2519
 *                 _alt = self._interp.adaptivePredict(self._input,44,self._ctx)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2519, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.import_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_3, &__pyx_t_7) < 0) __PYX_ERR(0, 2519, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_3);
        __pyx_v_re = __pyx_t_3;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":2520
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 2520, __pyx_L24_error)

          /* "pymola/generated/ModelicaParser.pyx":2521
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2521, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2521, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2521, __pyx_L24_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2521, __pyx_L24_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2521, __pyx_L24_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_4) {
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_4); __pyx_t_4 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_13, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2521, __pyx_L24_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2522
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2522, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2522, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2522, __pyx_L24_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2522, __pyx_L24_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2522, __pyx_L24_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2522, __pyx_L24_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":2519
 *                 _alt = self._interp.adaptivePredict(self._input,44,self._ctx)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L25;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L24_error:;
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_6 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L8_except_error;
          }
          __pyx_L25:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":2502
 *         localctx = ModelicaParser.Import_listContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 34, self.RULE_import_list)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 386
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":2524
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2524, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2524, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2524, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_14 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2524, __pyx_L31_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2524, __pyx_L31_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2524, __pyx_L31_error)
        }
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
      __pyx_L31_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":2525
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Extends_clauseContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2498
 * 
 * 
 *     def import_list(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Import_listContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.import_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__alt);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2529
 *     class Extends_clauseContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 2529, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 2529, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2529, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Extends_clauseContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Extends_clauseContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Extends_clauseContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":2530
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 2530, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2530, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2530, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2530, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2530, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2531
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def component_reference(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 2531, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2529
 *     class Extends_clauseContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Extends_clauseContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2533
 *             self.parser = parser
 * 
 *         def component_reference(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_3component_reference(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_3component_reference = {"component_reference", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_3component_reference, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_3component_reference(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("component_reference (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_2component_reference(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_2component_reference(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("component_reference", 0);

  /* "pymola/generated/ModelicaParser.pyx":2534
 * 
 *         def component_reference(self):
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Component_referenceContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2534, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2534, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2534, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2534, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2533
 *             self.parser = parser
 * 
 *         def component_reference(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Extends_clauseContext.component_reference", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2537
 * 
 * 
 *         def class_modification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_modificationContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_5class_modification(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_5class_modification = {"class_modification", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_5class_modification, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_5class_modification(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("class_modification (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_4class_modification(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_4class_modification(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("class_modification", 0);

  /* "pymola/generated/ModelicaParser.pyx":2538
 * 
 *         def class_modification(self):
 *             return self.getTypedRuleContext(ModelicaParser.Class_modificationContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Class_modificationContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2538, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2538, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2538, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2538, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2537
 * 
 * 
 *         def class_modification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_modificationContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Extends_clauseContext.class_modification", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2541
 * 
 * 
 *         def annotation(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.AnnotationContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_7annotation(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_7annotation = {"annotation", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_7annotation, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_7annotation(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("annotation (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_6annotation(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_6annotation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("annotation", 0);

  /* "pymola/generated/ModelicaParser.pyx":2542
 * 
 *         def annotation(self):
 *             return self.getTypedRuleContext(ModelicaParser.AnnotationContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_AnnotationContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2542, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2542, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2542, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2542, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2541
 * 
 * 
 *         def annotation(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.AnnotationContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Extends_clauseContext.annotation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2545
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_extends_clause
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_9getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_9getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_9getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_9getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_8getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_8getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":2546
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_extends_clause             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_extends_clause); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2545
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_extends_clause
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Extends_clauseContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2548
 *             return ModelicaParser.RULE_extends_clause
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExtends_clause" ):
 *                 listener.enterExtends_clause(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_11enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_11enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_11enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_11enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 2548, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 2548, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2548, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Extends_clauseContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_10enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_10enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":2549
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExtends_clause" ):             # <<<<<<<<<<<<<<
 *                 listener.enterExtends_clause(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterExtends_clause); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 2549, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":2550
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExtends_clause" ):
 *                 listener.enterExtends_clause(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterExtends_clause); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2550, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2550, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2550, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2550, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2550, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2550, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":2549
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExtends_clause" ):             # <<<<<<<<<<<<<<
 *                 listener.enterExtends_clause(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":2548
 *             return ModelicaParser.RULE_extends_clause
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExtends_clause" ):
 *                 listener.enterExtends_clause(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Extends_clauseContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2552
 *                 listener.enterExtends_clause(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExtends_clause" ):
 *                 listener.exitExtends_clause(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_13exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_13exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_13exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_13exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 2552, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 2552, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2552, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Extends_clauseContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_12exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_12exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":2553
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExtends_clause" ):             # <<<<<<<<<<<<<<
 *                 listener.exitExtends_clause(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitExtends_clause); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 2553, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":2554
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExtends_clause" ):
 *                 listener.exitExtends_clause(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitExtends_clause); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2554, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2554, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2554, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2554, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2554, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2554, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":2553
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExtends_clause" ):             # <<<<<<<<<<<<<<
 *                 listener.exitExtends_clause(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":2552
 *                 listener.enterExtends_clause(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExtends_clause" ):
 *                 listener.exitExtends_clause(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Extends_clauseContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2556
 *                 listener.exitExtends_clause(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExtends_clause" ):
 *                 return visitor.visitExtends_clause(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_15accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_15accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_15accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_15accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 2556, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 2556, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2556, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Extends_clauseContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_14accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_14accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":2557
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExtends_clause" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitExtends_clause(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitExtends_clause); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 2557, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":2558
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExtends_clause" ):
 *                 return visitor.visitExtends_clause(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitExtends_clause); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2558, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2558, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2558, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2558, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2558, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2558, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":2557
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExtends_clause" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitExtends_clause(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":2560
 *                 return visitor.visitExtends_clause(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2560, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2560, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2560, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2560, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2560, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2560, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":2556
 *                 listener.exitExtends_clause(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExtends_clause" ):
 *                 return visitor.visitExtends_clause(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Extends_clauseContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2565
 * 
 * 
 *     def extends_clause(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Extends_clauseContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_39extends_clause(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_39extends_clause = {"extends_clause", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_39extends_clause, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_39extends_clause(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("extends_clause (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_38extends_clause(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_38extends_clause(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  char const *__pyx_t_21;
  __Pyx_RefNannySetupContext("extends_clause", 0);

  /* "pymola/generated/ModelicaParser.pyx":2567
 *     def extends_clause(self):
 * 
 *         localctx = ModelicaParser.Extends_clauseContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 36, self.RULE_extends_clause)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Extends_clauseContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2567, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2567, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2567, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2567, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2568
 * 
 *         localctx = ModelicaParser.Extends_clauseContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 36, self.RULE_extends_clause)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2568, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_extends_clause); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2568, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_36, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2568, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_36, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2568, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2568, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_36);
    __Pyx_GIVEREF(__pyx_int_36);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_36);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2568, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2569
 *         localctx = ModelicaParser.Extends_clauseContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 36, self.RULE_extends_clause)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 2569, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2570
 *         self.enterRule(localctx, 36, self.RULE_extends_clause)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 394
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":2571
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 394
 *             self.match(ModelicaParser.T__21)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2571, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2571, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2571, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2571, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2571, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2572
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 394             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__21)
 *             self.state = 395
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_394) < 0) __PYX_ERR(0, 2572, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":2573
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 394
 *             self.match(ModelicaParser.T__21)             # <<<<<<<<<<<<<<
 *             self.state = 395
 *             self.component_reference()
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2573, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2573, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__21); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2573, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2573, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2573, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2573, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2573, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2573, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2574
 *             self.state = 394
 *             self.match(ModelicaParser.T__21)
 *             self.state = 395             # <<<<<<<<<<<<<<
 *             self.component_reference()
 *             self.state = 397
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_395) < 0) __PYX_ERR(0, 2574, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":2575
 *             self.match(ModelicaParser.T__21)
 *             self.state = 395
 *             self.component_reference()             # <<<<<<<<<<<<<<
 *             self.state = 397
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_component_reference); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2575, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2575, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2575, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2576
 *             self.state = 395
 *             self.component_reference()
 *             self.state = 397             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_397) < 0) __PYX_ERR(0, 2576, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":2577
 *             self.component_reference()
 *             self.state = 397
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__16:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2577, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2577, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2577, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2577, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2577, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2577, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2577, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2578
 *             self.state = 397
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.T__16:
 *                 self.state = 396
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2578, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2578, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__857, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2578, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_v__la = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2579
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__16:             # <<<<<<<<<<<<<<
 *                 self.state = 396
 *                 self.class_modification()
 */
        __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2579, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__16); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2579, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2579, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2579, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":2580
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__16:
 *                 self.state = 396             # <<<<<<<<<<<<<<
 *                 self.class_modification()
 * 
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_396) < 0) __PYX_ERR(0, 2580, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":2581
 *             if _la==ModelicaParser.T__16:
 *                 self.state = 396
 *                 self.class_modification()             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_class_modification); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2581, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2581, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2581, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2579
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__16:             # <<<<<<<<<<<<<<
 *                 self.state = 396
 *                 self.class_modification()
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":2584
 * 
 * 
 *             self.state = 400             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_400) < 0) __PYX_ERR(0, 2584, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":2585
 * 
 *             self.state = 400
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__77:
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2585, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2585, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        if (!__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2585, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2585, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2585, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2585, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __pyx_t_4 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2585, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2586
 *             self.state = 400
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.T__77:
 *                 self.state = 399
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2586, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2586, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__858, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2586, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2587
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__77:             # <<<<<<<<<<<<<<
 *                 self.state = 399
 *                 self.annotation()
 */
        __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2587, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__77); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2587, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2587, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2587, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":2588
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__77:
 *                 self.state = 399             # <<<<<<<<<<<<<<
 *                 self.annotation()
 * 
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_399) < 0) __PYX_ERR(0, 2588, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":2589
 *             if _la==ModelicaParser.T__77:
 *                 self.state = 399
 *                 self.annotation()             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_annotation); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2589, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (__pyx_t_3) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2589, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2589, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2587
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__77:             # <<<<<<<<<<<<<<
 *                 self.state = 399
 *                 self.annotation()
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":2570
 *         self.enterRule(localctx, 36, self.RULE_extends_clause)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 394
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":2592
 * 
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2592, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.extends_clause", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3) < 0) __PYX_ERR(0, 2592, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __pyx_v_re = __pyx_t_2;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":2593
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 2593, __pyx_L21_error)

          /* "pymola/generated/ModelicaParser.pyx":2594
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2594, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2594, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_re};
            __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2594, __pyx_L21_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_re};
            __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2594, __pyx_L21_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          {
            __pyx_t_12 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 2594, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_12);
            if (__pyx_t_7) {
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_7); __pyx_t_7 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_12, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2594, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2595
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2595, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 2595, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_12);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_12, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2595, __pyx_L21_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2595, __pyx_L21_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2595, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_7);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_7, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2595, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":2592
 * 
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L22;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L21_error:;
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __pyx_t_6 = __pyx_lineno; __pyx_t_13 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
            }
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_ErrRestore(__pyx_t_15, __pyx_t_16, __pyx_t_17);
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_13; __pyx_filename = __pyx_t_14;
            goto __pyx_L8_except_error;
          }
          __pyx_L22:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":2570
 *         self.enterRule(localctx, 36, self.RULE_extends_clause)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 394
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":2597
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2597, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2597, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2597, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_19, &__pyx_t_18);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_18);
      __pyx_t_13 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_21 = __pyx_filename;
      {
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2597, __pyx_L28_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2597, __pyx_L28_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2597, __pyx_L28_error)
        }
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __pyx_lineno = __pyx_t_13; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_21;
      goto __pyx_L1_error;
      __pyx_L28_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":2598
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Constraining_clauseContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2565
 * 
 * 
 *     def extends_clause(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Extends_clauseContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.extends_clause", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2602
 *     class Constraining_clauseContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 2602, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 2602, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2602, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Constraining_clauseContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":2603
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 2603, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2603, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2603, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2603, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2603, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2603, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2603, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2603, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2604
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def name(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 2604, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2602
 *     class Constraining_clauseContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Constraining_clauseContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2606
 *             self.parser = parser
 * 
 *         def name(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_3name(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_3name = {"name", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_3name, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_3name(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("name (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_2name(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_2name(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("name", 0);

  /* "pymola/generated/ModelicaParser.pyx":2607
 * 
 *         def name(self):
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2607, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2607, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_NameContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2607, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2607, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2607, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2607, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2607, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2606
 *             self.parser = parser
 * 
 *         def name(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Constraining_clauseContext.name", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2610
 * 
 * 
 *         def class_modification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_modificationContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_5class_modification(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_5class_modification = {"class_modification", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_5class_modification, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_5class_modification(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("class_modification (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_4class_modification(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_4class_modification(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("class_modification", 0);

  /* "pymola/generated/ModelicaParser.pyx":2611
 * 
 *         def class_modification(self):
 *             return self.getTypedRuleContext(ModelicaParser.Class_modificationContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2611, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2611, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Class_modificationContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2611, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2611, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2611, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2611, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2611, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2610
 * 
 * 
 *         def class_modification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_modificationContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Constraining_clauseContext.class_modification", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2614
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_constraining_clause
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_7getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_7getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_6getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":2615
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_constraining_clause             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_constraining_clause); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2614
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_constraining_clause
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Constraining_clauseContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2617
 *             return ModelicaParser.RULE_constraining_clause
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterConstraining_clause" ):
 *                 listener.enterConstraining_clause(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_9enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_9enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 2617, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 2617, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2617, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Constraining_clauseContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_8enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":2618
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterConstraining_clause" ):             # <<<<<<<<<<<<<<
 *                 listener.enterConstraining_clause(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterConstraining_clause); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 2618, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":2619
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterConstraining_clause" ):
 *                 listener.enterConstraining_clause(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterConstraining_clause); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2619, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2619, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2619, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2619, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2619, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2619, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":2618
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterConstraining_clause" ):             # <<<<<<<<<<<<<<
 *                 listener.enterConstraining_clause(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":2617
 *             return ModelicaParser.RULE_constraining_clause
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterConstraining_clause" ):
 *                 listener.enterConstraining_clause(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Constraining_clauseContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2621
 *                 listener.enterConstraining_clause(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitConstraining_clause" ):
 *                 listener.exitConstraining_clause(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_11exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_11exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 2621, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 2621, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2621, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Constraining_clauseContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_10exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":2622
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitConstraining_clause" ):             # <<<<<<<<<<<<<<
 *                 listener.exitConstraining_clause(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitConstraining_clause); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 2622, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":2623
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitConstraining_clause" ):
 *                 listener.exitConstraining_clause(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitConstraining_clause); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2623, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2623, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2623, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2623, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2623, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2623, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":2622
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitConstraining_clause" ):             # <<<<<<<<<<<<<<
 *                 listener.exitConstraining_clause(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":2621
 *                 listener.enterConstraining_clause(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitConstraining_clause" ):
 *                 listener.exitConstraining_clause(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Constraining_clauseContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2625
 *                 listener.exitConstraining_clause(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitConstraining_clause" ):
 *                 return visitor.visitConstraining_clause(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_13accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_13accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 2625, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 2625, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2625, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Constraining_clauseContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_12accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":2626
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitConstraining_clause" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitConstraining_clause(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitConstraining_clause); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 2626, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":2627
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitConstraining_clause" ):
 *                 return visitor.visitConstraining_clause(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitConstraining_clause); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2627, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2627, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2627, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2627, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2627, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2627, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":2626
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitConstraining_clause" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitConstraining_clause(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":2629
 *                 return visitor.visitConstraining_clause(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2629, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2629, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2629, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2629, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2629, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2629, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":2625
 *                 listener.exitConstraining_clause(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitConstraining_clause" ):
 *                 return visitor.visitConstraining_clause(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Constraining_clauseContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2634
 * 
 * 
 *     def constraining_clause(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Constraining_clauseContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_41constraining_clause(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_41constraining_clause = {"constraining_clause", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_41constraining_clause, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_41constraining_clause(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("constraining_clause (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_40constraining_clause(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_40constraining_clause(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  char const *__pyx_t_21;
  __Pyx_RefNannySetupContext("constraining_clause", 0);

  /* "pymola/generated/ModelicaParser.pyx":2636
 *     def constraining_clause(self):
 * 
 *         localctx = ModelicaParser.Constraining_clauseContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 38, self.RULE_constraining_clause)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2636, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Constraining_clauseContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2636, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2636, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2636, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2636, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2636, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2636, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2636, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2637
 * 
 *         localctx = ModelicaParser.Constraining_clauseContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 38, self.RULE_constraining_clause)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_constraining_clause); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_38, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2637, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_38, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2637, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2637, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_38);
    __Pyx_GIVEREF(__pyx_int_38);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_38);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2637, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2638
 *         localctx = ModelicaParser.Constraining_clauseContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 38, self.RULE_constraining_clause)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 2638, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2639
 *         self.enterRule(localctx, 38, self.RULE_constraining_clause)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 402
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":2640
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 402
 *             self.match(ModelicaParser.T__31)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2640, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2640, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2640, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2640, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2640, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2641
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 402             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__31)
 *             self.state = 403
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_402) < 0) __PYX_ERR(0, 2641, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":2642
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 402
 *             self.match(ModelicaParser.T__31)             # <<<<<<<<<<<<<<
 *             self.state = 403
 *             self.name()
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2642, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2642, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__31); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2642, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2642, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2642, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2642, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2642, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2642, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2643
 *             self.state = 402
 *             self.match(ModelicaParser.T__31)
 *             self.state = 403             # <<<<<<<<<<<<<<
 *             self.name()
 *             self.state = 405
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_403) < 0) __PYX_ERR(0, 2643, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":2644
 *             self.match(ModelicaParser.T__31)
 *             self.state = 403
 *             self.name()             # <<<<<<<<<<<<<<
 *             self.state = 405
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2644, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2644, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2644, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2645
 *             self.state = 403
 *             self.name()
 *             self.state = 405             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_405) < 0) __PYX_ERR(0, 2645, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":2646
 *             self.name()
 *             self.state = 405
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__16:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2646, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2646, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2646, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2646, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2646, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2646, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2646, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2647
 *             self.state = 405
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.T__16:
 *                 self.state = 404
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2647, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2647, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__859, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2647, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_v__la = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2648
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__16:             # <<<<<<<<<<<<<<
 *                 self.state = 404
 *                 self.class_modification()
 */
        __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2648, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__16); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2648, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2648, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2648, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":2649
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__16:
 *                 self.state = 404             # <<<<<<<<<<<<<<
 *                 self.class_modification()
 * 
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_404) < 0) __PYX_ERR(0, 2649, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":2650
 *             if _la==ModelicaParser.T__16:
 *                 self.state = 404
 *                 self.class_modification()             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_class_modification); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2650, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2650, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2650, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2648
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__16:             # <<<<<<<<<<<<<<
 *                 self.state = 404
 *                 self.class_modification()
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":2639
 *         self.enterRule(localctx, 38, self.RULE_constraining_clause)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 402
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":2653
 * 
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2653, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.constraining_clause", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_4, &__pyx_t_2) < 0) __PYX_ERR(0, 2653, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __pyx_v_re = __pyx_t_4;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":2654
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 2654, __pyx_L20_error)

          /* "pymola/generated/ModelicaParser.pyx":2655
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2655, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2655, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2655, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2655, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_12 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 2655, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_12);
            if (__pyx_t_7) {
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_7); __pyx_t_7 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_12, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2655, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2656
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2656, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 2656, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_12);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_12, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2656, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2656, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2656, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_7);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2656, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":2653
 * 
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L21;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L20_error:;
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __pyx_t_6 = __pyx_lineno; __pyx_t_13 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
            }
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_ErrRestore(__pyx_t_15, __pyx_t_16, __pyx_t_17);
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_13; __pyx_filename = __pyx_t_14;
            goto __pyx_L8_except_error;
          }
          __pyx_L21:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":2639
 *         self.enterRule(localctx, 38, self.RULE_constraining_clause)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 402
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":2658
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2658, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2658, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2658, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_19, &__pyx_t_18);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_18);
      __pyx_t_13 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_21 = __pyx_filename;
      {
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2658, __pyx_L27_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2658, __pyx_L27_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2658, __pyx_L27_error)
        }
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __pyx_lineno = __pyx_t_13; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_21;
      goto __pyx_L1_error;
      __pyx_L27_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":2659
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Component_clauseContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2634
 * 
 * 
 *     def constraining_clause(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Constraining_clauseContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.constraining_clause", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2663
 *     class Component_clauseContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Component_clauseContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Component_clauseContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Component_clauseContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Component_clauseContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 2663, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 2663, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2663, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_clauseContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Component_clauseContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Component_clauseContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":2664
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 2664, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2664, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2664, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2664, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2664, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2664, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2664, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2664, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2665
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def type_prefix(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 2665, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2663
 *     class Component_clauseContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_clauseContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2667
 *             self.parser = parser
 * 
 *         def type_prefix(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Type_prefixContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Component_clauseContext_3type_prefix(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Component_clauseContext_3type_prefix = {"type_prefix", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Component_clauseContext_3type_prefix, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Component_clauseContext_3type_prefix(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("type_prefix (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Component_clauseContext_2type_prefix(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Component_clauseContext_2type_prefix(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("type_prefix", 0);

  /* "pymola/generated/ModelicaParser.pyx":2668
 * 
 *         def type_prefix(self):
 *             return self.getTypedRuleContext(ModelicaParser.Type_prefixContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2668, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2668, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Type_prefixContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2668, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2668, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2668, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2668, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2668, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2667
 *             self.parser = parser
 * 
 *         def type_prefix(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Type_prefixContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_clauseContext.type_prefix", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2671
 * 
 * 
 *         def type_specifier(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Type_specifierContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Component_clauseContext_5type_specifier(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Component_clauseContext_5type_specifier = {"type_specifier", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Component_clauseContext_5type_specifier, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Component_clauseContext_5type_specifier(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("type_specifier (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Component_clauseContext_4type_specifier(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Component_clauseContext_4type_specifier(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("type_specifier", 0);

  /* "pymola/generated/ModelicaParser.pyx":2672
 * 
 *         def type_specifier(self):
 *             return self.getTypedRuleContext(ModelicaParser.Type_specifierContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Type_specifierContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2672, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2672, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2672, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2672, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2671
 * 
 * 
 *         def type_specifier(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Type_specifierContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_clauseContext.type_specifier", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2675
 * 
 * 
 *         def component_list(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_listContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Component_clauseContext_7component_list(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Component_clauseContext_7component_list = {"component_list", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Component_clauseContext_7component_list, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Component_clauseContext_7component_list(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("component_list (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Component_clauseContext_6component_list(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Component_clauseContext_6component_list(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("component_list", 0);

  /* "pymola/generated/ModelicaParser.pyx":2676
 * 
 *         def component_list(self):
 *             return self.getTypedRuleContext(ModelicaParser.Component_listContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2676, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2676, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Component_listContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2676, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2676, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2676, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2676, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2676, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2675
 * 
 * 
 *         def component_list(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_listContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_clauseContext.component_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2679
 * 
 * 
 *         def array_subscripts(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Array_subscriptsContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Component_clauseContext_9array_subscripts(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Component_clauseContext_9array_subscripts = {"array_subscripts", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Component_clauseContext_9array_subscripts, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Component_clauseContext_9array_subscripts(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("array_subscripts (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Component_clauseContext_8array_subscripts(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Component_clauseContext_8array_subscripts(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("array_subscripts", 0);

  /* "pymola/generated/ModelicaParser.pyx":2680
 * 
 *         def array_subscripts(self):
 *             return self.getTypedRuleContext(ModelicaParser.Array_subscriptsContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2680, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2680, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Array_subscriptsContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2680, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2680, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2680, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2680, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2680, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2679
 * 
 * 
 *         def array_subscripts(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Array_subscriptsContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_clauseContext.array_subscripts", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2683
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_component_clause
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Component_clauseContext_11getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Component_clauseContext_11getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Component_clauseContext_11getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Component_clauseContext_11getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Component_clauseContext_10getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Component_clauseContext_10getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":2684
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_component_clause             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2684, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_component_clause); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2684, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2683
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_component_clause
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_clauseContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2686
 *             return ModelicaParser.RULE_component_clause
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterComponent_clause" ):
 *                 listener.enterComponent_clause(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Component_clauseContext_13enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Component_clauseContext_13enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Component_clauseContext_13enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Component_clauseContext_13enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 2686, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 2686, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2686, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_clauseContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Component_clauseContext_12enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Component_clauseContext_12enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":2687
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterComponent_clause" ):             # <<<<<<<<<<<<<<
 *                 listener.enterComponent_clause(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterComponent_clause); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 2687, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":2688
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterComponent_clause" ):
 *                 listener.enterComponent_clause(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterComponent_clause); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2688, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2688, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2688, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2688, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2688, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2688, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":2687
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterComponent_clause" ):             # <<<<<<<<<<<<<<
 *                 listener.enterComponent_clause(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":2686
 *             return ModelicaParser.RULE_component_clause
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterComponent_clause" ):
 *                 listener.enterComponent_clause(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_clauseContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2690
 *                 listener.enterComponent_clause(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitComponent_clause" ):
 *                 listener.exitComponent_clause(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Component_clauseContext_15exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Component_clauseContext_15exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Component_clauseContext_15exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Component_clauseContext_15exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 2690, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 2690, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2690, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_clauseContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Component_clauseContext_14exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Component_clauseContext_14exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":2691
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitComponent_clause" ):             # <<<<<<<<<<<<<<
 *                 listener.exitComponent_clause(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitComponent_clause); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 2691, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":2692
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitComponent_clause" ):
 *                 listener.exitComponent_clause(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitComponent_clause); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2692, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2692, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2692, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2692, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2692, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2692, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":2691
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitComponent_clause" ):             # <<<<<<<<<<<<<<
 *                 listener.exitComponent_clause(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":2690
 *                 listener.enterComponent_clause(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitComponent_clause" ):
 *                 listener.exitComponent_clause(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_clauseContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2694
 *                 listener.exitComponent_clause(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitComponent_clause" ):
 *                 return visitor.visitComponent_clause(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Component_clauseContext_17accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Component_clauseContext_17accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Component_clauseContext_17accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Component_clauseContext_17accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 2694, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 2694, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2694, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_clauseContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Component_clauseContext_16accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Component_clauseContext_16accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":2695
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitComponent_clause" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitComponent_clause(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitComponent_clause); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 2695, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":2696
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitComponent_clause" ):
 *                 return visitor.visitComponent_clause(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitComponent_clause); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2696, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2696, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2696, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2696, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2696, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2696, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":2695
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitComponent_clause" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitComponent_clause(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":2698
 *                 return visitor.visitComponent_clause(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2698, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2698, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2698, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2698, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2698, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2698, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":2694
 *                 listener.exitComponent_clause(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitComponent_clause" ):
 *                 return visitor.visitComponent_clause(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_clauseContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2703
 * 
 * 
 *     def component_clause(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Component_clauseContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_43component_clause(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_43component_clause = {"component_clause", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_43component_clause, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_43component_clause(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("component_clause (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_42component_clause(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_42component_clause(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  char const *__pyx_t_21;
  __Pyx_RefNannySetupContext("component_clause", 0);

  /* "pymola/generated/ModelicaParser.pyx":2705
 *     def component_clause(self):
 * 
 *         localctx = ModelicaParser.Component_clauseContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 40, self.RULE_component_clause)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2705, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Component_clauseContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2705, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2705, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2705, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2705, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2705, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2705, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2705, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2706
 * 
 *         localctx = ModelicaParser.Component_clauseContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 40, self.RULE_component_clause)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_component_clause); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_40, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2706, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_40, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2706, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2706, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_40);
    __Pyx_GIVEREF(__pyx_int_40);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_40);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2706, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2707
 *         localctx = ModelicaParser.Component_clauseContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 40, self.RULE_component_clause)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 2707, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2708
 *         self.enterRule(localctx, 40, self.RULE_component_clause)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 407
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":2709
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 407
 *             self.type_prefix()
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2709, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2709, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2709, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2709, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2709, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2710
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 407             # <<<<<<<<<<<<<<
 *             self.type_prefix()
 *             self.state = 408
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_407) < 0) __PYX_ERR(0, 2710, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":2711
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 407
 *             self.type_prefix()             # <<<<<<<<<<<<<<
 *             self.state = 408
 *             self.type_specifier()
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_type_prefix); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2711, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2711, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2711, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2712
 *             self.state = 407
 *             self.type_prefix()
 *             self.state = 408             # <<<<<<<<<<<<<<
 *             self.type_specifier()
 *             self.state = 410
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_408) < 0) __PYX_ERR(0, 2712, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":2713
 *             self.type_prefix()
 *             self.state = 408
 *             self.type_specifier()             # <<<<<<<<<<<<<<
 *             self.state = 410
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_type_specifier); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2713, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2713, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2713, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2714
 *             self.state = 408
 *             self.type_specifier()
 *             self.state = 410             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_410) < 0) __PYX_ERR(0, 2714, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":2715
 *             self.type_specifier()
 *             self.state = 410
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__75:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2715, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2715, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2715, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2715, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2715, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2715, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2715, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2716
 *             self.state = 410
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.T__75:
 *                 self.state = 409
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2716, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2716, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__860, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2716, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v__la = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2717
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__75:             # <<<<<<<<<<<<<<
 *                 self.state = 409
 *                 self.array_subscripts()
 */
        __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2717, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__75); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2717, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2717, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2717, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":2718
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__75:
 *                 self.state = 409             # <<<<<<<<<<<<<<
 *                 self.array_subscripts()
 * 
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_409) < 0) __PYX_ERR(0, 2718, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":2719
 *             if _la==ModelicaParser.T__75:
 *                 self.state = 409
 *                 self.array_subscripts()             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_array_subscripts); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2719, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2719, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2719, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2717
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__75:             # <<<<<<<<<<<<<<
 *                 self.state = 409
 *                 self.array_subscripts()
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":2722
 * 
 * 
 *             self.state = 412             # <<<<<<<<<<<<<<
 *             self.component_list()
 *         except RecognitionException as re:
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_412) < 0) __PYX_ERR(0, 2722, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":2723
 * 
 *             self.state = 412
 *             self.component_list()             # <<<<<<<<<<<<<<
 *         except RecognitionException as re:
 *             localctx.exception = re
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_component_list); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2723, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (__pyx_t_2) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2723, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2723, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2708
 *         self.enterRule(localctx, 40, self.RULE_component_clause)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 407
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":2724
 *             self.state = 412
 *             self.component_list()
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2724, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.component_clause", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_7, &__pyx_t_2) < 0) __PYX_ERR(0, 2724, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_7);
        __pyx_v_re = __pyx_t_7;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":2725
 *             self.component_list()
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 2725, __pyx_L20_error)

          /* "pymola/generated/ModelicaParser.pyx":2726
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2726, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2726, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2726, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2726, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_12 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 2726, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_12);
            if (__pyx_t_4) {
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_4); __pyx_t_4 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_12, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2726, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2727
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2727, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 2727, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_12);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_12, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2727, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2727, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2727, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2727, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":2724
 *             self.state = 412
 *             self.component_list()
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L21;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L20_error:;
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __pyx_t_6 = __pyx_lineno; __pyx_t_13 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
            }
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_ErrRestore(__pyx_t_15, __pyx_t_16, __pyx_t_17);
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_13; __pyx_filename = __pyx_t_14;
            goto __pyx_L8_except_error;
          }
          __pyx_L21:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":2708
 *         self.enterRule(localctx, 40, self.RULE_component_clause)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 407
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":2729
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2729, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2729, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2729, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_19, &__pyx_t_18);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_18);
      __pyx_t_13 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_21 = __pyx_filename;
      {
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2729, __pyx_L27_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2729, __pyx_L27_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2729, __pyx_L27_error)
        }
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __pyx_lineno = __pyx_t_13; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_21;
      goto __pyx_L1_error;
      __pyx_L27_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":2730
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Type_prefixContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2703
 * 
 * 
 *     def component_clause(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Component_clauseContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.component_clause", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2734
 *     class Type_prefixContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Type_prefixContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18Type_prefixContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18Type_prefixContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Type_prefixContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 2734, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 2734, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2734, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Type_prefixContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18Type_prefixContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Type_prefixContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":2735
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 2735, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2735, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2735, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2735, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2735, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2735, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2735, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2735, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2736
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 2736, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2734
 *     class Type_prefixContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Type_prefixContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2739
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_type_prefix
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Type_prefixContext_3getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18Type_prefixContext_3getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18Type_prefixContext_3getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Type_prefixContext_3getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18Type_prefixContext_2getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Type_prefixContext_2getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":2740
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_type_prefix             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2740, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_type_prefix); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2740, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2739
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_type_prefix
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Type_prefixContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2742
 *             return ModelicaParser.RULE_type_prefix
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterType_prefix" ):
 *                 listener.enterType_prefix(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Type_prefixContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18Type_prefixContext_5enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18Type_prefixContext_5enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Type_prefixContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 2742, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 2742, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2742, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Type_prefixContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18Type_prefixContext_4enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Type_prefixContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":2743
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterType_prefix" ):             # <<<<<<<<<<<<<<
 *                 listener.enterType_prefix(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterType_prefix); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 2743, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":2744
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterType_prefix" ):
 *                 listener.enterType_prefix(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterType_prefix); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2744, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2744, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2744, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2744, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2744, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2744, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":2743
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterType_prefix" ):             # <<<<<<<<<<<<<<
 *                 listener.enterType_prefix(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":2742
 *             return ModelicaParser.RULE_type_prefix
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterType_prefix" ):
 *                 listener.enterType_prefix(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Type_prefixContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2746
 *                 listener.enterType_prefix(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitType_prefix" ):
 *                 listener.exitType_prefix(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Type_prefixContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18Type_prefixContext_7exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18Type_prefixContext_7exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Type_prefixContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 2746, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 2746, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2746, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Type_prefixContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18Type_prefixContext_6exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Type_prefixContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":2747
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitType_prefix" ):             # <<<<<<<<<<<<<<
 *                 listener.exitType_prefix(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitType_prefix); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 2747, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":2748
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitType_prefix" ):
 *                 listener.exitType_prefix(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitType_prefix); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2748, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2748, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2748, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2748, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2748, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2748, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":2747
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitType_prefix" ):             # <<<<<<<<<<<<<<
 *                 listener.exitType_prefix(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":2746
 *                 listener.enterType_prefix(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitType_prefix" ):
 *                 listener.exitType_prefix(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Type_prefixContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2750
 *                 listener.exitType_prefix(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitType_prefix" ):
 *                 return visitor.visitType_prefix(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Type_prefixContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18Type_prefixContext_9accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18Type_prefixContext_9accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Type_prefixContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 2750, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 2750, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2750, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Type_prefixContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18Type_prefixContext_8accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Type_prefixContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":2751
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitType_prefix" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitType_prefix(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitType_prefix); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 2751, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":2752
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitType_prefix" ):
 *                 return visitor.visitType_prefix(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitType_prefix); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2752, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2752, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2752, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2752, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2752, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2752, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":2751
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitType_prefix" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitType_prefix(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":2754
 *                 return visitor.visitType_prefix(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2754, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2754, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2754, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2754, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2754, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2754, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":2750
 *                 listener.exitType_prefix(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitType_prefix" ):
 *                 return visitor.visitType_prefix(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Type_prefixContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2759
 * 
 * 
 *     def type_prefix(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Type_prefixContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_45type_prefix(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_45type_prefix = {"type_prefix", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_45type_prefix, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_45type_prefix(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("type_prefix (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_44type_prefix(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_44type_prefix(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  char const *__pyx_t_22;
  __Pyx_RefNannySetupContext("type_prefix", 0);

  /* "pymola/generated/ModelicaParser.pyx":2761
 *     def type_prefix(self):
 * 
 *         localctx = ModelicaParser.Type_prefixContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 42, self.RULE_type_prefix)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2761, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Type_prefixContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2761, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2761, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2761, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2761, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2761, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2761, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2761, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2762
 * 
 *         localctx = ModelicaParser.Type_prefixContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 42, self.RULE_type_prefix)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2762, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_type_prefix); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2762, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_42, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2762, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_42, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2762, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2762, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_42);
    __Pyx_GIVEREF(__pyx_int_42);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_42);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2762, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2763
 *         localctx = ModelicaParser.Type_prefixContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 42, self.RULE_type_prefix)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 2763, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2764
 *         self.enterRule(localctx, 42, self.RULE_type_prefix)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 415
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":2765
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 415
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2765, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2765, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2765, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2765, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2765, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2766
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 415             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_415) < 0) __PYX_ERR(0, 2766, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":2767
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 415
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__32 or _la==ModelicaParser.T__33:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2767, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2767, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2767, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2767, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2767, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2767, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2767, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2768
 *             self.state = 415
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.T__32 or _la==ModelicaParser.T__33:
 *                 self.state = 414
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2768, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2768, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__861, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2768, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v__la = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2769
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__32 or _la==ModelicaParser.T__33:             # <<<<<<<<<<<<<<
 *                 self.state = 414
 *                 _la = self._input.LA(1)
 */
        __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2769, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__32); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2769, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2769, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2769, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2769, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__33); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2769, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2769, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2769, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_11 = __pyx_t_12;
        __pyx_L15_bool_binop_done:;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":2770
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__32 or _la==ModelicaParser.T__33:
 *                 self.state = 414             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 *                 if not(_la==ModelicaParser.T__32 or _la==ModelicaParser.T__33):
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_414) < 0) __PYX_ERR(0, 2770, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":2771
 *             if _la==ModelicaParser.T__32 or _la==ModelicaParser.T__33:
 *                 self.state = 414
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if not(_la==ModelicaParser.T__32 or _la==ModelicaParser.T__33):
 *                     self._errHandler.recoverInline(self)
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2771, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2771, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__862, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2771, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2772
 *                 self.state = 414
 *                 _la = self._input.LA(1)
 *                 if not(_la==ModelicaParser.T__32 or _la==ModelicaParser.T__33):             # <<<<<<<<<<<<<<
 *                     self._errHandler.recoverInline(self)
 *                 else:
 */
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2772, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__32); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2772, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2772, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2772, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (!__pyx_t_12) {
          } else {
            __pyx_t_11 = __pyx_t_12;
            goto __pyx_L18_bool_binop_done;
          }
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2772, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__33); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2772, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2772, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2772, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_11 = __pyx_t_12;
          __pyx_L18_bool_binop_done:;
          __pyx_t_12 = ((!__pyx_t_11) != 0);
          if (__pyx_t_12) {

            /* "pymola/generated/ModelicaParser.pyx":2773
 *                 _la = self._input.LA(1)
 *                 if not(_la==ModelicaParser.T__32 or _la==ModelicaParser.T__33):
 *                     self._errHandler.recoverInline(self)             # <<<<<<<<<<<<<<
 *                 else:
 *                     self._errHandler.reportMatch(self)
 */
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2773, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_recoverInline); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2773, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __pyx_t_7 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_7)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_7);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
              }
            }
            if (!__pyx_t_7) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2773, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_2)) {
                PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2773, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
                PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2773, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else
              #endif
              {
                __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2773, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
                __Pyx_INCREF(__pyx_v_self);
                __Pyx_GIVEREF(__pyx_v_self);
                PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2773, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":2772
 *                 self.state = 414
 *                 _la = self._input.LA(1)
 *                 if not(_la==ModelicaParser.T__32 or _la==ModelicaParser.T__33):             # <<<<<<<<<<<<<<
 *                     self._errHandler.recoverInline(self)
 *                 else:
 */
            goto __pyx_L17;
          }

          /* "pymola/generated/ModelicaParser.pyx":2775
 *                     self._errHandler.recoverInline(self)
 *                 else:
 *                     self._errHandler.reportMatch(self)             # <<<<<<<<<<<<<<
 *                     self.consume()
 * 
 */
          /*else*/ {
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2775, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_reportMatch); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2775, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_2 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
              __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
              if (likely(__pyx_t_2)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                __Pyx_INCREF(__pyx_t_2);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_3, function);
              }
            }
            if (!__pyx_t_2) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2775, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_3)) {
                PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2775, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
                PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2775, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else
              #endif
              {
                __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2775, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_7);
                __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
                __Pyx_INCREF(__pyx_v_self);
                __Pyx_GIVEREF(__pyx_v_self);
                PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self);
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2775, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":2776
 *                 else:
 *                     self._errHandler.reportMatch(self)
 *                     self.consume()             # <<<<<<<<<<<<<<
 * 
 * 
 */
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_consume); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2776, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_7 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
              __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
              if (likely(__pyx_t_7)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                __Pyx_INCREF(__pyx_t_7);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_3, function);
              }
            }
            if (__pyx_t_7) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2776, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else {
              __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2776, __pyx_L6_error)
            }
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
          __pyx_L17:;

          /* "pymola/generated/ModelicaParser.pyx":2769
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__32 or _la==ModelicaParser.T__33:             # <<<<<<<<<<<<<<
 *                 self.state = 414
 *                 _la = self._input.LA(1)
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":2779
 * 
 * 
 *             self.state = 418             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_418) < 0) __PYX_ERR(0, 2779, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":2780
 * 
 *             self.state = 418
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__34) | (1 << ModelicaParser.T__35) | (1 << ModelicaParser.T__36))) != 0):
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2780, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2780, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2780, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2780, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2780, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2780, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2780, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2781
 *             self.state = 418
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__34) | (1 << ModelicaParser.T__35) | (1 << ModelicaParser.T__36))) != 0):
 *                 self.state = 417
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2781, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2781, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__863, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2781, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2782
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__34) | (1 << ModelicaParser.T__35) | (1 << ModelicaParser.T__36))) != 0):             # <<<<<<<<<<<<<<
 *                 self.state = 417
 *                 _la = self._input.LA(1)
 */
        __pyx_t_1 = __Pyx_PyInt_From_long((~0x3f)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2782, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = PyNumber_And(__pyx_v__la, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2782, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_t_7, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2782, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2782, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L21_bool_binop_done;
        }
        __pyx_t_1 = PyNumber_Lshift(__pyx_int_1, __pyx_v__la); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2782, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2782, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__34); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2782, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2782, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2782, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__35); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2782, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2782, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_Or(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2782, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2782, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__36); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2782, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2782, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyNumber_Or(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2782, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_And(__pyx_t_1, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2782, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyObject_RichCompare(__pyx_t_2, __pyx_int_0, Py_NE); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2782, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2782, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_12 = __pyx_t_11;
        __pyx_L21_bool_binop_done:;
        if (__pyx_t_12) {

          /* "pymola/generated/ModelicaParser.pyx":2783
 *             _la = self._input.LA(1)
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__34) | (1 << ModelicaParser.T__35) | (1 << ModelicaParser.T__36))) != 0):
 *                 self.state = 417             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 *                 if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__34) | (1 << ModelicaParser.T__35) | (1 << ModelicaParser.T__36))) != 0)):
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_417) < 0) __PYX_ERR(0, 2783, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":2784
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__34) | (1 << ModelicaParser.T__35) | (1 << ModelicaParser.T__36))) != 0):
 *                 self.state = 417
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__34) | (1 << ModelicaParser.T__35) | (1 << ModelicaParser.T__36))) != 0)):
 *                     self._errHandler.recoverInline(self)
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2784, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_LA); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2784, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__864, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2784, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_7);
          __pyx_t_7 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2785
 *                 self.state = 417
 *                 _la = self._input.LA(1)
 *                 if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__34) | (1 << ModelicaParser.T__35) | (1 << ModelicaParser.T__36))) != 0)):             # <<<<<<<<<<<<<<
 *                     self._errHandler.recoverInline(self)
 *                 else:
 */
          __pyx_t_7 = __Pyx_PyInt_From_long((~0x3f)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2785, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = PyNumber_And(__pyx_v__la, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2785, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_PyInt_EqObjC(__pyx_t_2, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2785, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2785, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (__pyx_t_11) {
          } else {
            __pyx_t_12 = __pyx_t_11;
            goto __pyx_L24_bool_binop_done;
          }
          __pyx_t_7 = PyNumber_Lshift(__pyx_int_1, __pyx_v__la); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2785, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2785, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__34); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2785, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2785, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2785, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__35); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2785, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2785, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyNumber_Or(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2785, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2785, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__36); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2785, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2785, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Or(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2785, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyNumber_And(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2785, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_int_0, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2785, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2785, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_12 = __pyx_t_11;
          __pyx_L24_bool_binop_done:;
          __pyx_t_11 = ((!__pyx_t_12) != 0);
          if (__pyx_t_11) {

            /* "pymola/generated/ModelicaParser.pyx":2786
 *                 _la = self._input.LA(1)
 *                 if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__34) | (1 << ModelicaParser.T__35) | (1 << ModelicaParser.T__36))) != 0)):
 *                     self._errHandler.recoverInline(self)             # <<<<<<<<<<<<<<
 *                 else:
 *                     self._errHandler.reportMatch(self)
 */
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2786, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_recoverInline); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2786, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
              __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
              if (likely(__pyx_t_1)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
                __Pyx_INCREF(__pyx_t_1);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_7, function);
              }
            }
            if (!__pyx_t_1) {
              __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2786, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_7)) {
                PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_v_self};
                __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2786, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                __Pyx_GOTREF(__pyx_t_2);
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
                PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_v_self};
                __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2786, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                __Pyx_GOTREF(__pyx_t_2);
              } else
              #endif
              {
                __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2786, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1); __pyx_t_1 = NULL;
                __Pyx_INCREF(__pyx_v_self);
                __Pyx_GIVEREF(__pyx_v_self);
                PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
                __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2786, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

            /* "pymola/generated/ModelicaParser.pyx":2785
 *                 self.state = 417
 *                 _la = self._input.LA(1)
 *                 if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__34) | (1 << ModelicaParser.T__35) | (1 << ModelicaParser.T__36))) != 0)):             # <<<<<<<<<<<<<<
 *                     self._errHandler.recoverInline(self)
 *                 else:
 */
            goto __pyx_L23;
          }

          /* "pymola/generated/ModelicaParser.pyx":2788
 *                     self._errHandler.recoverInline(self)
 *                 else:
 *                     self._errHandler.reportMatch(self)             # <<<<<<<<<<<<<<
 *                     self.consume()
 * 
 */
          /*else*/ {
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2788, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_reportMatch); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2788, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __pyx_t_7 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
              __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
              if (likely(__pyx_t_7)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                __Pyx_INCREF(__pyx_t_7);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_3, function);
              }
            }
            if (!__pyx_t_7) {
              __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2788, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_3)) {
                PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
                __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2788, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
                __Pyx_GOTREF(__pyx_t_2);
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
                PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
                __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2788, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
                __Pyx_GOTREF(__pyx_t_2);
              } else
              #endif
              {
                __pyx_t_1 = PyTuple_New(1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2788, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_7); __pyx_t_7 = NULL;
                __Pyx_INCREF(__pyx_v_self);
                __Pyx_GIVEREF(__pyx_v_self);
                PyTuple_SET_ITEM(__pyx_t_1, 0+1, __pyx_v_self);
                __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2788, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

            /* "pymola/generated/ModelicaParser.pyx":2789
 *                 else:
 *                     self._errHandler.reportMatch(self)
 *                     self.consume()             # <<<<<<<<<<<<<<
 * 
 * 
 */
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_consume); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2789, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_1 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
              __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
              if (likely(__pyx_t_1)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                __Pyx_INCREF(__pyx_t_1);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_3, function);
              }
            }
            if (__pyx_t_1) {
              __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2789, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            } else {
              __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2789, __pyx_L6_error)
            }
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __pyx_L23:;

          /* "pymola/generated/ModelicaParser.pyx":2782
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__34) | (1 << ModelicaParser.T__35) | (1 << ModelicaParser.T__36))) != 0):             # <<<<<<<<<<<<<<
 *                 self.state = 417
 *                 _la = self._input.LA(1)
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":2792
 * 
 * 
 *             self.state = 421             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_421) < 0) __PYX_ERR(0, 2792, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":2793
 * 
 *             self.state = 421
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__37 or _la==ModelicaParser.T__38:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2793, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2793, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_1, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2793, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_1)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2793, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2793, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2793, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2793, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2794
 *             self.state = 421
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.T__37 or _la==ModelicaParser.T__38:
 *                 self.state = 420
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2794, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_LA); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2794, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__865, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2794, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2795
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__37 or _la==ModelicaParser.T__38:             # <<<<<<<<<<<<<<
 *                 self.state = 420
 *                 _la = self._input.LA(1)
 */
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2795, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__37); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2795, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_v__la, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2795, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2795, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L27_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2795, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__38); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2795, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_v__la, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2795, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2795, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_11 = __pyx_t_12;
        __pyx_L27_bool_binop_done:;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":2796
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__37 or _la==ModelicaParser.T__38:
 *                 self.state = 420             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 *                 if not(_la==ModelicaParser.T__37 or _la==ModelicaParser.T__38):
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_420) < 0) __PYX_ERR(0, 2796, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":2797
 *             if _la==ModelicaParser.T__37 or _la==ModelicaParser.T__38:
 *                 self.state = 420
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if not(_la==ModelicaParser.T__37 or _la==ModelicaParser.T__38):
 *                     self._errHandler.recoverInline(self)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2797, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_LA); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2797, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__866, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2797, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_2);
          __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2798
 *                 self.state = 420
 *                 _la = self._input.LA(1)
 *                 if not(_la==ModelicaParser.T__37 or _la==ModelicaParser.T__38):             # <<<<<<<<<<<<<<
 *                     self._errHandler.recoverInline(self)
 *                 else:
 */
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2798, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__37); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2798, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyObject_RichCompare(__pyx_v__la, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2798, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2798, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          if (!__pyx_t_12) {
          } else {
            __pyx_t_11 = __pyx_t_12;
            goto __pyx_L30_bool_binop_done;
          }
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2798, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__38); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2798, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyObject_RichCompare(__pyx_v__la, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2798, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2798, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_11 = __pyx_t_12;
          __pyx_L30_bool_binop_done:;
          __pyx_t_12 = ((!__pyx_t_11) != 0);
          if (__pyx_t_12) {

            /* "pymola/generated/ModelicaParser.pyx":2799
 *                 _la = self._input.LA(1)
 *                 if not(_la==ModelicaParser.T__37 or _la==ModelicaParser.T__38):
 *                     self._errHandler.recoverInline(self)             # <<<<<<<<<<<<<<
 *                 else:
 *                     self._errHandler.reportMatch(self)
 */
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2799, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_recoverInline); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2799, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
              __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
              if (likely(__pyx_t_1)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
                __Pyx_INCREF(__pyx_t_1);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_7, function);
              }
            }
            if (!__pyx_t_1) {
              __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2799, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_7)) {
                PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_v_self};
                __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2799, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                __Pyx_GOTREF(__pyx_t_2);
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
                PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_v_self};
                __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2799, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                __Pyx_GOTREF(__pyx_t_2);
              } else
              #endif
              {
                __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2799, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1); __pyx_t_1 = NULL;
                __Pyx_INCREF(__pyx_v_self);
                __Pyx_GIVEREF(__pyx_v_self);
                PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
                __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2799, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

            /* "pymola/generated/ModelicaParser.pyx":2798
 *                 self.state = 420
 *                 _la = self._input.LA(1)
 *                 if not(_la==ModelicaParser.T__37 or _la==ModelicaParser.T__38):             # <<<<<<<<<<<<<<
 *                     self._errHandler.recoverInline(self)
 *                 else:
 */
            goto __pyx_L29;
          }

          /* "pymola/generated/ModelicaParser.pyx":2801
 *                     self._errHandler.recoverInline(self)
 *                 else:
 *                     self._errHandler.reportMatch(self)             # <<<<<<<<<<<<<<
 *                     self.consume()
 * 
 */
          /*else*/ {
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2801, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_reportMatch); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2801, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __pyx_t_7 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
              __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
              if (likely(__pyx_t_7)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                __Pyx_INCREF(__pyx_t_7);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_3, function);
              }
            }
            if (!__pyx_t_7) {
              __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2801, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_3)) {
                PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
                __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2801, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
                __Pyx_GOTREF(__pyx_t_2);
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
                PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
                __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2801, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
                __Pyx_GOTREF(__pyx_t_2);
              } else
              #endif
              {
                __pyx_t_1 = PyTuple_New(1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2801, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_7); __pyx_t_7 = NULL;
                __Pyx_INCREF(__pyx_v_self);
                __Pyx_GIVEREF(__pyx_v_self);
                PyTuple_SET_ITEM(__pyx_t_1, 0+1, __pyx_v_self);
                __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2801, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

            /* "pymola/generated/ModelicaParser.pyx":2802
 *                 else:
 *                     self._errHandler.reportMatch(self)
 *                     self.consume()             # <<<<<<<<<<<<<<
 * 
 * 
 */
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_consume); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2802, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_1 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
              __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
              if (likely(__pyx_t_1)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                __Pyx_INCREF(__pyx_t_1);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_3, function);
              }
            }
            if (__pyx_t_1) {
              __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2802, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            } else {
              __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2802, __pyx_L6_error)
            }
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __pyx_L29:;

          /* "pymola/generated/ModelicaParser.pyx":2795
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__37 or _la==ModelicaParser.T__38:             # <<<<<<<<<<<<<<
 *                 self.state = 420
 *                 _la = self._input.LA(1)
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":2764
 *         self.enterRule(localctx, 42, self.RULE_type_prefix)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 415
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "pymola/generated/ModelicaParser.pyx":2805
 * 
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2805, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.type_prefix", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_3, &__pyx_t_1) < 0) __PYX_ERR(0, 2805, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __pyx_v_re = __pyx_t_3;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":2806
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 2806, __pyx_L37_error)

          /* "pymola/generated/ModelicaParser.pyx":2807
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2807, __pyx_L37_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2807, __pyx_L37_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2807, __pyx_L37_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2807, __pyx_L37_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2807, __pyx_L37_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_4) {
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_4); __pyx_t_4 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_13, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2807, __pyx_L37_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2808
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2808, __pyx_L37_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2808, __pyx_L37_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2808, __pyx_L37_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2808, __pyx_L37_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2808, __pyx_L37_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_4, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2808, __pyx_L37_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":2805
 * 
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L38;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L37_error:;
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_6 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L8_except_error;
          }
          __pyx_L38:;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":2764
 *         self.enterRule(localctx, 42, self.RULE_type_prefix)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 415
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":2810
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2810, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_2) {
        __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2810, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      } else {
        __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2810, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_14 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2810, __pyx_L44_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_2) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2810, __pyx_L44_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2810, __pyx_L44_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
      __pyx_L44_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":2811
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Type_specifierContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2759
 * 
 * 
 *     def type_prefix(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Type_prefixContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.type_prefix", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2815
 *     class Type_specifierContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Type_specifierContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Type_specifierContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Type_specifierContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Type_specifierContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 2815, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 2815, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2815, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Type_specifierContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Type_specifierContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Type_specifierContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":2816
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 2816, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2816, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2816, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2816, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2816, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2816, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2816, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2816, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2817
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def component_reference(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 2817, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2815
 *     class Type_specifierContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Type_specifierContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2819
 *             self.parser = parser
 * 
 *         def component_reference(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Type_specifierContext_3component_reference(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Type_specifierContext_3component_reference = {"component_reference", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Type_specifierContext_3component_reference, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Type_specifierContext_3component_reference(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("component_reference (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Type_specifierContext_2component_reference(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Type_specifierContext_2component_reference(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("component_reference", 0);

  /* "pymola/generated/ModelicaParser.pyx":2820
 * 
 *         def component_reference(self):
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Component_referenceContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2820, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2820, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2820, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2820, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2819
 *             self.parser = parser
 * 
 *         def component_reference(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Type_specifierContext.component_reference", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2823
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_type_specifier
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Type_specifierContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Type_specifierContext_5getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Type_specifierContext_5getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Type_specifierContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Type_specifierContext_4getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Type_specifierContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":2824
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_type_specifier             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2824, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_type_specifier); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2824, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2823
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_type_specifier
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Type_specifierContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2826
 *             return ModelicaParser.RULE_type_specifier
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterType_specifier" ):
 *                 listener.enterType_specifier(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Type_specifierContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Type_specifierContext_7enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Type_specifierContext_7enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Type_specifierContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 2826, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 2826, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2826, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Type_specifierContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Type_specifierContext_6enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Type_specifierContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":2827
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterType_specifier" ):             # <<<<<<<<<<<<<<
 *                 listener.enterType_specifier(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterType_specifier); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 2827, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":2828
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterType_specifier" ):
 *                 listener.enterType_specifier(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterType_specifier); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2828, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2828, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2828, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2828, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2828, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2828, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":2827
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterType_specifier" ):             # <<<<<<<<<<<<<<
 *                 listener.enterType_specifier(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":2826
 *             return ModelicaParser.RULE_type_specifier
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterType_specifier" ):
 *                 listener.enterType_specifier(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Type_specifierContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2830
 *                 listener.enterType_specifier(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitType_specifier" ):
 *                 listener.exitType_specifier(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Type_specifierContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Type_specifierContext_9exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Type_specifierContext_9exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Type_specifierContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 2830, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 2830, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2830, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Type_specifierContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Type_specifierContext_8exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Type_specifierContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":2831
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitType_specifier" ):             # <<<<<<<<<<<<<<
 *                 listener.exitType_specifier(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitType_specifier); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 2831, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":2832
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitType_specifier" ):
 *                 listener.exitType_specifier(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitType_specifier); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2832, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2832, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2832, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2832, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2832, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2832, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":2831
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitType_specifier" ):             # <<<<<<<<<<<<<<
 *                 listener.exitType_specifier(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":2830
 *                 listener.enterType_specifier(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitType_specifier" ):
 *                 listener.exitType_specifier(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Type_specifierContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2834
 *                 listener.exitType_specifier(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitType_specifier" ):
 *                 return visitor.visitType_specifier(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Type_specifierContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Type_specifierContext_11accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Type_specifierContext_11accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Type_specifierContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 2834, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 2834, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2834, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Type_specifierContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Type_specifierContext_10accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Type_specifierContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":2835
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitType_specifier" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitType_specifier(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitType_specifier); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 2835, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":2836
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitType_specifier" ):
 *                 return visitor.visitType_specifier(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitType_specifier); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2836, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2836, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2836, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2836, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2836, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2836, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":2835
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitType_specifier" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitType_specifier(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":2838
 *                 return visitor.visitType_specifier(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2838, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2838, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2838, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2838, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2838, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2838, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":2834
 *                 listener.exitType_specifier(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitType_specifier" ):
 *                 return visitor.visitType_specifier(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Type_specifierContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2843
 * 
 * 
 *     def type_specifier(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Type_specifierContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_47type_specifier(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_47type_specifier = {"type_specifier", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_47type_specifier, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_47type_specifier(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("type_specifier (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_46type_specifier(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_46type_specifier(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  char const *__pyx_t_20;
  __Pyx_RefNannySetupContext("type_specifier", 0);

  /* "pymola/generated/ModelicaParser.pyx":2845
 *     def type_specifier(self):
 * 
 *         localctx = ModelicaParser.Type_specifierContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 44, self.RULE_type_specifier)
 *         try:
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2845, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Type_specifierContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2845, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2845, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2845, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2845, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2845, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2845, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2845, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2846
 * 
 *         localctx = ModelicaParser.Type_specifierContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 44, self.RULE_type_specifier)             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2846, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_type_specifier); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2846, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_44, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2846, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_44, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2846, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2846, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_44);
    __Pyx_GIVEREF(__pyx_int_44);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_44);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2846, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2847
 *         localctx = ModelicaParser.Type_specifierContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 44, self.RULE_type_specifier)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 423
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":2848
 *         self.enterRule(localctx, 44, self.RULE_type_specifier)
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 423
 *             self.component_reference()
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2848, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2848, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2848, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2848, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2848, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2849
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 423             # <<<<<<<<<<<<<<
 *             self.component_reference()
 *         except RecognitionException as re:
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_423) < 0) __PYX_ERR(0, 2849, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":2850
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 423
 *             self.component_reference()             # <<<<<<<<<<<<<<
 *         except RecognitionException as re:
 *             localctx.exception = re
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_component_reference); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2850, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2850, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2850, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2847
 *         localctx = ModelicaParser.Type_specifierContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 44, self.RULE_type_specifier)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 423
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":2851
 *             self.state = 423
 *             self.component_reference()
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2851, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.type_specifier", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_3, &__pyx_t_7) < 0) __PYX_ERR(0, 2851, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_3);
        __pyx_v_re = __pyx_t_3;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":2852
 *             self.component_reference()
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 2852, __pyx_L19_error)

          /* "pymola/generated/ModelicaParser.pyx":2853
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2853, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2853, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2853, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2853, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_11 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2853, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_11);
            if (__pyx_t_4) {
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_4); __pyx_t_4 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_11, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2853, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2854
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2854, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2854, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_11)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2854, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_11)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2854, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2854, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2854, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":2851
 *             self.state = 423
 *             self.component_reference()
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L20;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L19_error:;
            __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_14);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __pyx_t_6 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_17);
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
            }
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
            __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
            goto __pyx_L8_except_error;
          }
          __pyx_L20:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":2847
 *         localctx = ModelicaParser.Type_specifierContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 44, self.RULE_type_specifier)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 423
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":2856
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2856, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2856, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2856, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_18, &__pyx_t_17);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_18);
      __Pyx_XGOTREF(__pyx_t_17);
      __pyx_t_12 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_20 = __pyx_filename;
      {
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2856, __pyx_L26_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2856, __pyx_L26_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2856, __pyx_L26_error)
        }
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_18, __pyx_t_17);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0;
      __pyx_lineno = __pyx_t_12; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_20;
      goto __pyx_L1_error;
      __pyx_L26_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_18, __pyx_t_17);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":2857
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Component_listContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2843
 * 
 * 
 *     def type_specifier(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Type_specifierContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.type_specifier", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2861
 *     class Component_listContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Component_listContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Component_listContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Component_listContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Component_listContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 2861, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 2861, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2861, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_listContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Component_listContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Component_listContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":2862
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 2862, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2862, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2862, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2862, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2862, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2862, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2862, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2862, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2863
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def component_declaration(self, i:int=None):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 2863, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2861
 *     class Component_listContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_listContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2865
 *             self.parser = parser
 * 
 *         def component_declaration(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Component_declarationContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Component_listContext_3component_declaration(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Component_listContext_3component_declaration = {"component_declaration", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Component_listContext_3component_declaration, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Component_listContext_3component_declaration(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("component_declaration (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "component_declaration") < 0)) __PYX_ERR(0, 2865, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("component_declaration", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2865, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_listContext.component_declaration", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Component_listContext_2component_declaration(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Component_listContext_2component_declaration(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("component_declaration", 0);

  /* "pymola/generated/ModelicaParser.pyx":2866
 * 
 *         def component_declaration(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.Component_declarationContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":2867
 *         def component_declaration(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Component_declarationContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.Component_declarationContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2867, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2867, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_Component_declarationContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2867, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2867, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2867, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2867, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2867, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2867, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":2866
 * 
 *         def component_declaration(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.Component_declarationContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":2869
 *                 return self.getTypedRuleContexts(ModelicaParser.Component_declarationContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.Component_declarationContext,i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2869, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2869, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Component_declarationContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2869, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2869, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2869, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2869, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2869, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":2865
 *             self.parser = parser
 * 
 *         def component_declaration(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Component_declarationContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_listContext.component_declaration", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2872
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_component_list
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Component_listContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Component_listContext_5getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Component_listContext_5getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Component_listContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Component_listContext_4getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Component_listContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":2873
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_component_list             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2873, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_component_list); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2873, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2872
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_component_list
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_listContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2875
 *             return ModelicaParser.RULE_component_list
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterComponent_list" ):
 *                 listener.enterComponent_list(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Component_listContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Component_listContext_7enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Component_listContext_7enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Component_listContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 2875, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 2875, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2875, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_listContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Component_listContext_6enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Component_listContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":2876
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterComponent_list" ):             # <<<<<<<<<<<<<<
 *                 listener.enterComponent_list(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterComponent_list); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 2876, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":2877
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterComponent_list" ):
 *                 listener.enterComponent_list(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterComponent_list); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2877, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2877, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2877, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2877, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2877, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2877, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":2876
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterComponent_list" ):             # <<<<<<<<<<<<<<
 *                 listener.enterComponent_list(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":2875
 *             return ModelicaParser.RULE_component_list
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterComponent_list" ):
 *                 listener.enterComponent_list(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_listContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2879
 *                 listener.enterComponent_list(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitComponent_list" ):
 *                 listener.exitComponent_list(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Component_listContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Component_listContext_9exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Component_listContext_9exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Component_listContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 2879, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 2879, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2879, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_listContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Component_listContext_8exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Component_listContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":2880
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitComponent_list" ):             # <<<<<<<<<<<<<<
 *                 listener.exitComponent_list(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitComponent_list); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 2880, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":2881
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitComponent_list" ):
 *                 listener.exitComponent_list(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitComponent_list); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2881, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2881, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2881, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2881, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2881, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2881, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":2880
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitComponent_list" ):             # <<<<<<<<<<<<<<
 *                 listener.exitComponent_list(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":2879
 *                 listener.enterComponent_list(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitComponent_list" ):
 *                 listener.exitComponent_list(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_listContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2883
 *                 listener.exitComponent_list(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitComponent_list" ):
 *                 return visitor.visitComponent_list(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Component_listContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Component_listContext_11accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Component_listContext_11accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Component_listContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 2883, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 2883, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2883, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_listContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Component_listContext_10accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Component_listContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":2884
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitComponent_list" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitComponent_list(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitComponent_list); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 2884, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":2885
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitComponent_list" ):
 *                 return visitor.visitComponent_list(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitComponent_list); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2885, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2885, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2885, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2885, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2885, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2885, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":2884
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitComponent_list" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitComponent_list(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":2887
 *                 return visitor.visitComponent_list(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2887, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2887, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2887, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2887, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2887, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2887, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":2883
 *                 listener.exitComponent_list(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitComponent_list" ):
 *                 return visitor.visitComponent_list(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_listContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2892
 * 
 * 
 *     def component_list(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Component_listContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_49component_list(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_49component_list = {"component_list", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_49component_list, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_49component_list(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("component_list (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_48component_list(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_48component_list(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  char const *__pyx_t_21;
  __Pyx_RefNannySetupContext("component_list", 0);

  /* "pymola/generated/ModelicaParser.pyx":2894
 *     def component_list(self):
 * 
 *         localctx = ModelicaParser.Component_listContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 46, self.RULE_component_list)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2894, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Component_listContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2894, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2894, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2894, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2894, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2894, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2894, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2894, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2895
 * 
 *         localctx = ModelicaParser.Component_listContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 46, self.RULE_component_list)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2895, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_component_list); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2895, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_46, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2895, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_46, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2895, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2895, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_46);
    __Pyx_GIVEREF(__pyx_int_46);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_46);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2895, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2896
 *         localctx = ModelicaParser.Component_listContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 46, self.RULE_component_list)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 2896, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2897
 *         self.enterRule(localctx, 46, self.RULE_component_list)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 425
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":2898
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 425
 *             self.component_declaration()
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2898, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2898, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2898, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2898, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2898, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2899
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 425             # <<<<<<<<<<<<<<
 *             self.component_declaration()
 *             self.state = 430
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_425) < 0) __PYX_ERR(0, 2899, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":2900
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 425
 *             self.component_declaration()             # <<<<<<<<<<<<<<
 *             self.state = 430
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_component_declaration); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2900, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2900, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2900, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2901
 *             self.state = 425
 *             self.component_declaration()
 *             self.state = 430             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_430) < 0) __PYX_ERR(0, 2901, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":2902
 *             self.component_declaration()
 *             self.state = 430
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             while _la==ModelicaParser.T__20:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2902, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2902, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2902, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2902, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2902, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2902, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2902, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2903
 *             self.state = 430
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             while _la==ModelicaParser.T__20:
 *                 self.state = 426
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2903, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2903, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__867, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2903, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v__la = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2904
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             while _la==ModelicaParser.T__20:             # <<<<<<<<<<<<<<
 *                 self.state = 426
 *                 self.match(ModelicaParser.T__20)
 */
        while (1) {
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2904, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__20); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2904, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2904, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2904, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (!__pyx_t_11) break;

          /* "pymola/generated/ModelicaParser.pyx":2905
 *             _la = self._input.LA(1)
 *             while _la==ModelicaParser.T__20:
 *                 self.state = 426             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__20)
 *                 self.state = 427
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_426) < 0) __PYX_ERR(0, 2905, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":2906
 *             while _la==ModelicaParser.T__20:
 *                 self.state = 426
 *                 self.match(ModelicaParser.T__20)             # <<<<<<<<<<<<<<
 *                 self.state = 427
 *                 self.component_declaration()
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2906, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2906, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__20); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2906, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2906, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2906, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2906, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2906, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2906, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2907
 *                 self.state = 426
 *                 self.match(ModelicaParser.T__20)
 *                 self.state = 427             # <<<<<<<<<<<<<<
 *                 self.component_declaration()
 *                 self.state = 432
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_427) < 0) __PYX_ERR(0, 2907, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":2908
 *                 self.match(ModelicaParser.T__20)
 *                 self.state = 427
 *                 self.component_declaration()             # <<<<<<<<<<<<<<
 *                 self.state = 432
 *                 self._errHandler.sync(self)
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_component_declaration); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2908, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_4 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (__pyx_t_4) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2908, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2908, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2909
 *                 self.state = 427
 *                 self.component_declaration()
 *                 self.state = 432             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_432) < 0) __PYX_ERR(0, 2909, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":2910
 *                 self.component_declaration()
 *                 self.state = 432
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 * 
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2910, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_sync); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2910, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2910, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2910, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2910, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2910, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2910, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2911
 *                 self.state = 432
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *         except RecognitionException as re:
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2911, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2911, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__868, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2911, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_1);
          __pyx_t_1 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":2897
 *         self.enterRule(localctx, 46, self.RULE_component_list)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 425
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":2913
 *                 _la = self._input.LA(1)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2913, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.component_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_4, &__pyx_t_3) < 0) __PYX_ERR(0, 2913, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __pyx_v_re = __pyx_t_4;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":2914
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 2914, __pyx_L21_error)

          /* "pymola/generated/ModelicaParser.pyx":2915
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2915, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2915, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2915, __pyx_L21_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2915, __pyx_L21_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          {
            __pyx_t_12 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 2915, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_12);
            if (__pyx_t_2) {
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_2); __pyx_t_2 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_12, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2915, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2916
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2916, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 2916, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_12);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_12, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2916, __pyx_L21_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2916, __pyx_L21_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2916, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_2, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2916, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":2913
 *                 _la = self._input.LA(1)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L22;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L21_error:;
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __pyx_t_6 = __pyx_lineno; __pyx_t_13 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
            }
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_ErrRestore(__pyx_t_15, __pyx_t_16, __pyx_t_17);
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_13; __pyx_filename = __pyx_t_14;
            goto __pyx_L8_except_error;
          }
          __pyx_L22:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":2897
 *         self.enterRule(localctx, 46, self.RULE_component_list)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 425
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":2918
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2918, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2918, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2918, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_19, &__pyx_t_18);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_18);
      __pyx_t_13 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_21 = __pyx_filename;
      {
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2918, __pyx_L28_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2918, __pyx_L28_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2918, __pyx_L28_error)
        }
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __pyx_lineno = __pyx_t_13; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_21;
      goto __pyx_L1_error;
      __pyx_L28_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":2919
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Component_declarationContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2892
 * 
 * 
 *     def component_list(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Component_listContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.component_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2923
 *     class Component_declarationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_28Component_declarationContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_28Component_declarationContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_28Component_declarationContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_28Component_declarationContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 2923, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 2923, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2923, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_declarationContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_28Component_declarationContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_28Component_declarationContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":2924
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 2924, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2924, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2924, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2924, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2924, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2924, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2924, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2924, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2925
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def declaration(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 2925, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2923
 *     class Component_declarationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_declarationContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2927
 *             self.parser = parser
 * 
 *         def declaration(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.DeclarationContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_28Component_declarationContext_3declaration(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_28Component_declarationContext_3declaration = {"declaration", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_28Component_declarationContext_3declaration, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_28Component_declarationContext_3declaration(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("declaration (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_28Component_declarationContext_2declaration(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_28Component_declarationContext_2declaration(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("declaration", 0);

  /* "pymola/generated/ModelicaParser.pyx":2928
 * 
 *         def declaration(self):
 *             return self.getTypedRuleContext(ModelicaParser.DeclarationContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2928, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2928, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_DeclarationContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2928, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2928, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2928, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2928, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2928, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2927
 *             self.parser = parser
 * 
 *         def declaration(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.DeclarationContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_declarationContext.declaration", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2931
 * 
 * 
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_28Component_declarationContext_5comment(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_28Component_declarationContext_5comment = {"comment", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_28Component_declarationContext_5comment, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_28Component_declarationContext_5comment(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("comment (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_28Component_declarationContext_4comment(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_28Component_declarationContext_4comment(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("comment", 0);

  /* "pymola/generated/ModelicaParser.pyx":2932
 * 
 *         def comment(self):
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2932, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2932, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_CommentContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2932, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2932, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2932, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2932, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2932, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2931
 * 
 * 
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_declarationContext.comment", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2935
 * 
 * 
 *         def condition_attribute(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Condition_attributeContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_28Component_declarationContext_7condition_attribute(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_28Component_declarationContext_7condition_attribute = {"condition_attribute", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_28Component_declarationContext_7condition_attribute, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_28Component_declarationContext_7condition_attribute(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("condition_attribute (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_28Component_declarationContext_6condition_attribute(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_28Component_declarationContext_6condition_attribute(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("condition_attribute", 0);

  /* "pymola/generated/ModelicaParser.pyx":2936
 * 
 *         def condition_attribute(self):
 *             return self.getTypedRuleContext(ModelicaParser.Condition_attributeContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2936, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2936, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Condition_attributeContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2936, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2936, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2936, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2936, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2936, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2935
 * 
 * 
 *         def condition_attribute(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Condition_attributeContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_declarationContext.condition_attribute", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2939
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_component_declaration
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_28Component_declarationContext_9getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_28Component_declarationContext_9getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_28Component_declarationContext_9getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_28Component_declarationContext_9getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_28Component_declarationContext_8getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_28Component_declarationContext_8getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":2940
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_component_declaration             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2940, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_component_declaration); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2940, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2939
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_component_declaration
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_declarationContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2942
 *             return ModelicaParser.RULE_component_declaration
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterComponent_declaration" ):
 *                 listener.enterComponent_declaration(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_28Component_declarationContext_11enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_28Component_declarationContext_11enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_28Component_declarationContext_11enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_28Component_declarationContext_11enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 2942, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 2942, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2942, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_declarationContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_28Component_declarationContext_10enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_28Component_declarationContext_10enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":2943
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterComponent_declaration" ):             # <<<<<<<<<<<<<<
 *                 listener.enterComponent_declaration(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterComponent_declaration); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 2943, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":2944
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterComponent_declaration" ):
 *                 listener.enterComponent_declaration(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterComponent_declaration); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2944, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2944, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2944, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2944, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2944, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2944, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":2943
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterComponent_declaration" ):             # <<<<<<<<<<<<<<
 *                 listener.enterComponent_declaration(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":2942
 *             return ModelicaParser.RULE_component_declaration
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterComponent_declaration" ):
 *                 listener.enterComponent_declaration(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_declarationContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2946
 *                 listener.enterComponent_declaration(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitComponent_declaration" ):
 *                 listener.exitComponent_declaration(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_28Component_declarationContext_13exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_28Component_declarationContext_13exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_28Component_declarationContext_13exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_28Component_declarationContext_13exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 2946, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 2946, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2946, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_declarationContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_28Component_declarationContext_12exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_28Component_declarationContext_12exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":2947
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitComponent_declaration" ):             # <<<<<<<<<<<<<<
 *                 listener.exitComponent_declaration(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitComponent_declaration); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 2947, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":2948
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitComponent_declaration" ):
 *                 listener.exitComponent_declaration(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitComponent_declaration); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2948, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2948, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2948, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2948, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2948, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2948, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":2947
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitComponent_declaration" ):             # <<<<<<<<<<<<<<
 *                 listener.exitComponent_declaration(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":2946
 *                 listener.enterComponent_declaration(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitComponent_declaration" ):
 *                 listener.exitComponent_declaration(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_declarationContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2950
 *                 listener.exitComponent_declaration(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitComponent_declaration" ):
 *                 return visitor.visitComponent_declaration(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_28Component_declarationContext_15accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_28Component_declarationContext_15accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_28Component_declarationContext_15accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_28Component_declarationContext_15accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 2950, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 2950, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2950, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_declarationContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_28Component_declarationContext_14accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_28Component_declarationContext_14accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":2951
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitComponent_declaration" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitComponent_declaration(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitComponent_declaration); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 2951, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":2952
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitComponent_declaration" ):
 *                 return visitor.visitComponent_declaration(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitComponent_declaration); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2952, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2952, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2952, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2952, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2952, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2952, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":2951
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitComponent_declaration" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitComponent_declaration(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":2954
 *                 return visitor.visitComponent_declaration(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2954, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2954, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2954, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2954, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2954, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2954, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":2950
 *                 listener.exitComponent_declaration(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitComponent_declaration" ):
 *                 return visitor.visitComponent_declaration(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_declarationContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2959
 * 
 * 
 *     def component_declaration(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Component_declarationContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_51component_declaration(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_51component_declaration = {"component_declaration", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_51component_declaration, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_51component_declaration(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("component_declaration (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_50component_declaration(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_50component_declaration(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  char const *__pyx_t_21;
  __Pyx_RefNannySetupContext("component_declaration", 0);

  /* "pymola/generated/ModelicaParser.pyx":2961
 *     def component_declaration(self):
 * 
 *         localctx = ModelicaParser.Component_declarationContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 48, self.RULE_component_declaration)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Component_declarationContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2961, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2961, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2961, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2961, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2962
 * 
 *         localctx = ModelicaParser.Component_declarationContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 48, self.RULE_component_declaration)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2962, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_component_declaration); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2962, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_48, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2962, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_48, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2962, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2962, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_48);
    __Pyx_GIVEREF(__pyx_int_48);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_48);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2962, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2963
 *         localctx = ModelicaParser.Component_declarationContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 48, self.RULE_component_declaration)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 2963, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2964
 *         self.enterRule(localctx, 48, self.RULE_component_declaration)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 433
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":2965
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 433
 *             self.declaration()
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2965, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2965, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2965, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2965, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2965, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2966
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 433             # <<<<<<<<<<<<<<
 *             self.declaration()
 *             self.state = 435
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_433) < 0) __PYX_ERR(0, 2966, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":2967
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 433
 *             self.declaration()             # <<<<<<<<<<<<<<
 *             self.state = 435
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_declaration); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2967, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2967, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2967, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2968
 *             self.state = 433
 *             self.declaration()
 *             self.state = 435             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_435) < 0) __PYX_ERR(0, 2968, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":2969
 *             self.declaration()
 *             self.state = 435
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__39:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2969, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2969, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2969, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2969, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2969, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2969, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2969, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2970
 *             self.state = 435
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.T__39:
 *                 self.state = 434
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2970, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2970, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__869, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2970, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v__la = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2971
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__39:             # <<<<<<<<<<<<<<
 *                 self.state = 434
 *                 self.condition_attribute()
 */
        __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2971, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__39); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2971, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2971, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2971, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":2972
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__39:
 *                 self.state = 434             # <<<<<<<<<<<<<<
 *                 self.condition_attribute()
 * 
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_434) < 0) __PYX_ERR(0, 2972, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":2973
 *             if _la==ModelicaParser.T__39:
 *                 self.state = 434
 *                 self.condition_attribute()             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_condition_attribute); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2973, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2973, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2973, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2971
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__39:             # <<<<<<<<<<<<<<
 *                 self.state = 434
 *                 self.condition_attribute()
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":2976
 * 
 * 
 *             self.state = 437             # <<<<<<<<<<<<<<
 *             self.comment()
 *         except RecognitionException as re:
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_437) < 0) __PYX_ERR(0, 2976, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":2977
 * 
 *             self.state = 437
 *             self.comment()             # <<<<<<<<<<<<<<
 *         except RecognitionException as re:
 *             localctx.exception = re
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_comment); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2977, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (__pyx_t_2) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2977, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2977, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":2964
 *         self.enterRule(localctx, 48, self.RULE_component_declaration)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 433
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":2978
 *             self.state = 437
 *             self.comment()
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2978, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.component_declaration", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_7, &__pyx_t_2) < 0) __PYX_ERR(0, 2978, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_7);
        __pyx_v_re = __pyx_t_7;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":2979
 *             self.comment()
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 2979, __pyx_L20_error)

          /* "pymola/generated/ModelicaParser.pyx":2980
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2980, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2980, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2980, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2980, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_12 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 2980, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_12);
            if (__pyx_t_4) {
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_4); __pyx_t_4 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_12, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2980, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":2981
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2981, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 2981, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_12);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_12, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2981, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2981, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2981, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2981, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":2978
 *             self.state = 437
 *             self.comment()
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L21;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L20_error:;
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __pyx_t_6 = __pyx_lineno; __pyx_t_13 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
            }
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_ErrRestore(__pyx_t_15, __pyx_t_16, __pyx_t_17);
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_13; __pyx_filename = __pyx_t_14;
            goto __pyx_L8_except_error;
          }
          __pyx_L21:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":2964
 *         self.enterRule(localctx, 48, self.RULE_component_declaration)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 433
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":2983
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2983, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2983, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2983, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_19, &__pyx_t_18);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_18);
      __pyx_t_13 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_21 = __pyx_filename;
      {
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2983, __pyx_L27_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2983, __pyx_L27_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2983, __pyx_L27_error)
        }
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __pyx_lineno = __pyx_t_13; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_21;
      goto __pyx_L1_error;
      __pyx_L27_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":2984
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Condition_attributeContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2959
 * 
 * 
 *     def component_declaration(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Component_declarationContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.component_declaration", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2988
 *     class Condition_attributeContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 2988, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 2988, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2988, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Condition_attributeContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Condition_attributeContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Condition_attributeContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":2989
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 2989, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2989, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2989, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2989, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2989, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2989, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2989, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2989, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2990
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def expression(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 2990, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2988
 *     class Condition_attributeContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Condition_attributeContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2992
 *             self.parser = parser
 * 
 *         def expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_3expression(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_3expression = {"expression", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_3expression, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_3expression(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("expression (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_2expression(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_2expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("expression", 0);

  /* "pymola/generated/ModelicaParser.pyx":2993
 * 
 *         def expression(self):
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2993, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2993, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_ExpressionContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2993, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2993, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2993, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2993, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2993, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2992
 *             self.parser = parser
 * 
 *         def expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Condition_attributeContext.expression", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2996
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_condition_attribute
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_5getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_5getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_4getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":2997
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_condition_attribute             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2997, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_condition_attribute); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2997, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":2996
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_condition_attribute
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Condition_attributeContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":2999
 *             return ModelicaParser.RULE_condition_attribute
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterCondition_attribute" ):
 *                 listener.enterCondition_attribute(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_7enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_7enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 2999, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 2999, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2999, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Condition_attributeContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_6enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":3000
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterCondition_attribute" ):             # <<<<<<<<<<<<<<
 *                 listener.enterCondition_attribute(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterCondition_attribute); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3000, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3001
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterCondition_attribute" ):
 *                 listener.enterCondition_attribute(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterCondition_attribute); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3001, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3001, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3001, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3001, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3001, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3001, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":3000
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterCondition_attribute" ):             # <<<<<<<<<<<<<<
 *                 listener.enterCondition_attribute(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":2999
 *             return ModelicaParser.RULE_condition_attribute
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterCondition_attribute" ):
 *                 listener.enterCondition_attribute(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Condition_attributeContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3003
 *                 listener.enterCondition_attribute(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitCondition_attribute" ):
 *                 listener.exitCondition_attribute(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_9exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_9exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 3003, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 3003, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3003, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Condition_attributeContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_8exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":3004
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitCondition_attribute" ):             # <<<<<<<<<<<<<<
 *                 listener.exitCondition_attribute(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitCondition_attribute); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3004, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3005
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitCondition_attribute" ):
 *                 listener.exitCondition_attribute(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitCondition_attribute); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3005, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3005, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3005, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3005, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3005, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3005, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":3004
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitCondition_attribute" ):             # <<<<<<<<<<<<<<
 *                 listener.exitCondition_attribute(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3003
 *                 listener.enterCondition_attribute(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitCondition_attribute" ):
 *                 listener.exitCondition_attribute(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Condition_attributeContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3007
 *                 listener.exitCondition_attribute(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitCondition_attribute" ):
 *                 return visitor.visitCondition_attribute(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_11accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_11accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 3007, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 3007, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3007, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Condition_attributeContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_10accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":3008
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitCondition_attribute" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitCondition_attribute(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitCondition_attribute); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3008, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3009
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitCondition_attribute" ):
 *                 return visitor.visitCondition_attribute(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitCondition_attribute); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3009, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3009, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3009, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3009, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3009, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3009, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":3008
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitCondition_attribute" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitCondition_attribute(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3011
 *                 return visitor.visitCondition_attribute(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3011, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3011, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3011, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3011, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3011, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3011, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":3007
 *                 listener.exitCondition_attribute(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitCondition_attribute" ):
 *                 return visitor.visitCondition_attribute(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Condition_attributeContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3016
 * 
 * 
 *     def condition_attribute(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Condition_attributeContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_53condition_attribute(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_53condition_attribute = {"condition_attribute", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_53condition_attribute, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_53condition_attribute(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("condition_attribute (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_52condition_attribute(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_52condition_attribute(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  char const *__pyx_t_20;
  __Pyx_RefNannySetupContext("condition_attribute", 0);

  /* "pymola/generated/ModelicaParser.pyx":3018
 *     def condition_attribute(self):
 * 
 *         localctx = ModelicaParser.Condition_attributeContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 50, self.RULE_condition_attribute)
 *         try:
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Condition_attributeContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3018, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3018, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3018, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3018, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3019
 * 
 *         localctx = ModelicaParser.Condition_attributeContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 50, self.RULE_condition_attribute)             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3019, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_condition_attribute); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3019, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_50, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3019, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_50, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3019, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3019, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_50);
    __Pyx_GIVEREF(__pyx_int_50);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_50);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3019, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3020
 *         localctx = ModelicaParser.Condition_attributeContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 50, self.RULE_condition_attribute)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 439
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":3021
 *         self.enterRule(localctx, 50, self.RULE_condition_attribute)
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 439
 *             self.match(ModelicaParser.T__39)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3021, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3021, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3021, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3021, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3021, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3022
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 439             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__39)
 *             self.state = 440
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_439) < 0) __PYX_ERR(0, 3022, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":3023
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 439
 *             self.match(ModelicaParser.T__39)             # <<<<<<<<<<<<<<
 *             self.state = 440
 *             self.expression()
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3023, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3023, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__39); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3023, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3023, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3023, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3023, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3023, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3023, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3024
 *             self.state = 439
 *             self.match(ModelicaParser.T__39)
 *             self.state = 440             # <<<<<<<<<<<<<<
 *             self.expression()
 *         except RecognitionException as re:
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_440) < 0) __PYX_ERR(0, 3024, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":3025
 *             self.match(ModelicaParser.T__39)
 *             self.state = 440
 *             self.expression()             # <<<<<<<<<<<<<<
 *         except RecognitionException as re:
 *             localctx.exception = re
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expression); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3025, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3025, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3025, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3020
 *         localctx = ModelicaParser.Condition_attributeContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 50, self.RULE_condition_attribute)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 439
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":3026
 *             self.state = 440
 *             self.expression()
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3026, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.condition_attribute", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_3, &__pyx_t_4) < 0) __PYX_ERR(0, 3026, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __pyx_v_re = __pyx_t_3;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":3027
 *             self.expression()
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 3027, __pyx_L19_error)

          /* "pymola/generated/ModelicaParser.pyx":3028
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3028, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3028, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3028, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3028, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_11 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 3028, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_11);
            if (__pyx_t_7) {
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_7); __pyx_t_7 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_11, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3028, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3029
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3029, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 3029, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_11)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3029, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_11)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3029, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3029, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_7);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3029, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":3026
 *             self.state = 440
 *             self.expression()
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L20;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L19_error:;
            __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_14);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __pyx_t_6 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_17);
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
            }
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
            __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
            goto __pyx_L8_except_error;
          }
          __pyx_L20:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":3020
 *         localctx = ModelicaParser.Condition_attributeContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 50, self.RULE_condition_attribute)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 439
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":3031
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3031, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3031, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3031, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_18, &__pyx_t_17);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_18);
      __Pyx_XGOTREF(__pyx_t_17);
      __pyx_t_12 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_20 = __pyx_filename;
      {
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3031, __pyx_L26_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3031, __pyx_L26_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3031, __pyx_L26_error)
        }
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_18, __pyx_t_17);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0;
      __pyx_lineno = __pyx_t_12; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_20;
      goto __pyx_L1_error;
      __pyx_L26_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_18, __pyx_t_17);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":3032
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class DeclarationContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3016
 * 
 * 
 *     def condition_attribute(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Condition_attributeContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.condition_attribute", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3036
 *     class DeclarationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18DeclarationContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18DeclarationContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18DeclarationContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18DeclarationContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 3036, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 3036, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3036, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.DeclarationContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18DeclarationContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18DeclarationContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":3037
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 3037, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3037, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3037, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3037, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3037, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3037, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3037, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3037, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3038
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def IDENT(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 3038, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3036
 *     class DeclarationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.DeclarationContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3040
 *             self.parser = parser
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18DeclarationContext_3IDENT(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18DeclarationContext_3IDENT = {"IDENT", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18DeclarationContext_3IDENT, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18DeclarationContext_3IDENT(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("IDENT (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18DeclarationContext_2IDENT(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18DeclarationContext_2IDENT(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("IDENT", 0);

  /* "pymola/generated/ModelicaParser.pyx":3041
 * 
 *         def IDENT(self):
 *             return self.getToken(ModelicaParser.IDENT, 0)             # <<<<<<<<<<<<<<
 * 
 *         def array_subscripts(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getToken); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3041, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3041, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3041, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3041, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3041, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3041, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3041, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3040
 *             self.parser = parser
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.DeclarationContext.IDENT", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3043
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 *         def array_subscripts(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Array_subscriptsContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18DeclarationContext_5array_subscripts(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18DeclarationContext_5array_subscripts = {"array_subscripts", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18DeclarationContext_5array_subscripts, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18DeclarationContext_5array_subscripts(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("array_subscripts (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18DeclarationContext_4array_subscripts(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18DeclarationContext_4array_subscripts(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("array_subscripts", 0);

  /* "pymola/generated/ModelicaParser.pyx":3044
 * 
 *         def array_subscripts(self):
 *             return self.getTypedRuleContext(ModelicaParser.Array_subscriptsContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Array_subscriptsContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3044, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3044, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3044, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3044, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3043
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 *         def array_subscripts(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Array_subscriptsContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.DeclarationContext.array_subscripts", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3047
 * 
 * 
 *         def modification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ModificationContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18DeclarationContext_7modification(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18DeclarationContext_7modification = {"modification", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18DeclarationContext_7modification, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18DeclarationContext_7modification(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("modification (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18DeclarationContext_6modification(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18DeclarationContext_6modification(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("modification", 0);

  /* "pymola/generated/ModelicaParser.pyx":3048
 * 
 *         def modification(self):
 *             return self.getTypedRuleContext(ModelicaParser.ModificationContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3048, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3048, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_ModificationContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3048, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3048, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3048, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3048, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3048, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3047
 * 
 * 
 *         def modification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ModificationContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.DeclarationContext.modification", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3051
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_declaration
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18DeclarationContext_9getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18DeclarationContext_9getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18DeclarationContext_9getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18DeclarationContext_9getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18DeclarationContext_8getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18DeclarationContext_8getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":3052
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_declaration             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3052, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_declaration); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3052, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3051
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_declaration
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.DeclarationContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3054
 *             return ModelicaParser.RULE_declaration
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterDeclaration" ):
 *                 listener.enterDeclaration(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18DeclarationContext_11enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18DeclarationContext_11enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18DeclarationContext_11enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18DeclarationContext_11enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 3054, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 3054, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3054, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.DeclarationContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18DeclarationContext_10enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18DeclarationContext_10enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":3055
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterDeclaration" ):             # <<<<<<<<<<<<<<
 *                 listener.enterDeclaration(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterDeclaration); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3055, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3056
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterDeclaration" ):
 *                 listener.enterDeclaration(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterDeclaration); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3056, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3056, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3056, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3056, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3056, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3056, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":3055
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterDeclaration" ):             # <<<<<<<<<<<<<<
 *                 listener.enterDeclaration(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3054
 *             return ModelicaParser.RULE_declaration
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterDeclaration" ):
 *                 listener.enterDeclaration(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.DeclarationContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3058
 *                 listener.enterDeclaration(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitDeclaration" ):
 *                 listener.exitDeclaration(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18DeclarationContext_13exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18DeclarationContext_13exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18DeclarationContext_13exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18DeclarationContext_13exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 3058, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 3058, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3058, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.DeclarationContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18DeclarationContext_12exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18DeclarationContext_12exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":3059
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitDeclaration" ):             # <<<<<<<<<<<<<<
 *                 listener.exitDeclaration(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitDeclaration); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3059, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3060
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitDeclaration" ):
 *                 listener.exitDeclaration(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitDeclaration); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3060, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3060, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3060, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3060, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3060, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3060, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":3059
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitDeclaration" ):             # <<<<<<<<<<<<<<
 *                 listener.exitDeclaration(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3058
 *                 listener.enterDeclaration(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitDeclaration" ):
 *                 listener.exitDeclaration(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.DeclarationContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3062
 *                 listener.exitDeclaration(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitDeclaration" ):
 *                 return visitor.visitDeclaration(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18DeclarationContext_15accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18DeclarationContext_15accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18DeclarationContext_15accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18DeclarationContext_15accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 3062, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 3062, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3062, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.DeclarationContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18DeclarationContext_14accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18DeclarationContext_14accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":3063
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitDeclaration" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitDeclaration(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitDeclaration); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3063, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3064
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitDeclaration" ):
 *                 return visitor.visitDeclaration(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitDeclaration); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3064, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3064, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3064, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3064, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3064, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3064, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":3063
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitDeclaration" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitDeclaration(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3066
 *                 return visitor.visitDeclaration(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3066, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3066, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3066, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3066, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3066, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3066, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":3062
 *                 listener.exitDeclaration(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitDeclaration" ):
 *                 return visitor.visitDeclaration(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.DeclarationContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3071
 * 
 * 
 *     def declaration(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.DeclarationContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_55declaration(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_55declaration = {"declaration", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_55declaration, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_55declaration(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("declaration (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_54declaration(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_54declaration(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  char const *__pyx_t_22;
  __Pyx_RefNannySetupContext("declaration", 0);

  /* "pymola/generated/ModelicaParser.pyx":3073
 *     def declaration(self):
 * 
 *         localctx = ModelicaParser.DeclarationContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 52, self.RULE_declaration)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3073, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_DeclarationContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3073, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3073, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3073, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3073, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3073, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3073, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3073, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3074
 * 
 *         localctx = ModelicaParser.DeclarationContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 52, self.RULE_declaration)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3074, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_declaration); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3074, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_52, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3074, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_52, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3074, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3074, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_52);
    __Pyx_GIVEREF(__pyx_int_52);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_52);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3074, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3075
 *         localctx = ModelicaParser.DeclarationContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 52, self.RULE_declaration)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 3075, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3076
 *         self.enterRule(localctx, 52, self.RULE_declaration)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 442
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":3077
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 442
 *             self.match(ModelicaParser.IDENT)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3077, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3077, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3077, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3077, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3077, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3078
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 442             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.IDENT)
 *             self.state = 444
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_442) < 0) __PYX_ERR(0, 3078, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":3079
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 442
 *             self.match(ModelicaParser.IDENT)             # <<<<<<<<<<<<<<
 *             self.state = 444
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3079, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3079, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3079, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3079, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3079, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3079, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3079, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3079, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3080
 *             self.state = 442
 *             self.match(ModelicaParser.IDENT)
 *             self.state = 444             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_444) < 0) __PYX_ERR(0, 3080, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":3081
 *             self.match(ModelicaParser.IDENT)
 *             self.state = 444
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__75:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3081, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3081, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3081, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3081, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3081, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3081, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3081, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3082
 *             self.state = 444
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.T__75:
 *                 self.state = 443
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3082, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3082, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__870, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3082, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_v__la = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3083
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__75:             # <<<<<<<<<<<<<<
 *                 self.state = 443
 *                 self.array_subscripts()
 */
        __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3083, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__75); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3083, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3083, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 3083, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":3084
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__75:
 *                 self.state = 443             # <<<<<<<<<<<<<<
 *                 self.array_subscripts()
 * 
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_443) < 0) __PYX_ERR(0, 3084, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":3085
 *             if _la==ModelicaParser.T__75:
 *                 self.state = 443
 *                 self.array_subscripts()             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_array_subscripts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3085, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3085, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3085, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3083
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__75:             # <<<<<<<<<<<<<<
 *                 self.state = 443
 *                 self.array_subscripts()
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":3088
 * 
 * 
 *             self.state = 447             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_447) < 0) __PYX_ERR(0, 3088, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":3089
 * 
 *             self.state = 447
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__14) | (1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__40))) != 0):
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3089, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3089, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        if (!__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3089, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3089, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3089, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3089, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __pyx_t_4 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3089, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3090
 *             self.state = 447
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__14) | (1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__40))) != 0):
 *                 self.state = 446
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3090, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3090, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__871, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3090, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3091
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__14) | (1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__40))) != 0):             # <<<<<<<<<<<<<<
 *                 self.state = 446
 *                 self.modification()
 */
        __pyx_t_1 = __Pyx_PyInt_From_long((~0x3f)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3091, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = PyNumber_And(__pyx_v__la, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3091, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_t_2, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3091, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3091, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L16_bool_binop_done;
        }
        __pyx_t_1 = PyNumber_Lshift(__pyx_int_1, __pyx_v__la); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3091, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3091, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__14); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3091, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3091, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3091, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__16); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3091, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_Lshift(__pyx_int_1, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3091, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyNumber_Or(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3091, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3091, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__40); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3091, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3091, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_Or(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3091, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_And(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3091, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_3, __pyx_int_0, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3091, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3091, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_11 = __pyx_t_12;
        __pyx_L16_bool_binop_done:;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":3092
 *             _la = self._input.LA(1)
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__14) | (1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__40))) != 0):
 *                 self.state = 446             # <<<<<<<<<<<<<<
 *                 self.modification()
 * 
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_446) < 0) __PYX_ERR(0, 3092, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":3093
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__14) | (1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__40))) != 0):
 *                 self.state = 446
 *                 self.modification()             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_modification); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3093, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (__pyx_t_1) {
            __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3093, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          } else {
            __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3093, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3091
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__14) | (1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__40))) != 0):             # <<<<<<<<<<<<<<
 *                 self.state = 446
 *                 self.modification()
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":3076
 *         self.enterRule(localctx, 52, self.RULE_declaration)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 442
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "pymola/generated/ModelicaParser.pyx":3096
 * 
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3096, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.declaration", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_3, &__pyx_t_1) < 0) __PYX_ERR(0, 3096, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __pyx_v_re = __pyx_t_3;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":3097
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 3097, __pyx_L23_error)

          /* "pymola/generated/ModelicaParser.pyx":3098
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3098, __pyx_L23_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3098, __pyx_L23_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_re};
            __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3098, __pyx_L23_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_re};
            __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3098, __pyx_L23_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3098, __pyx_L23_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_7) {
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_7); __pyx_t_7 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_13, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3098, __pyx_L23_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3099
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3099, __pyx_L23_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3099, __pyx_L23_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3099, __pyx_L23_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3099, __pyx_L23_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3099, __pyx_L23_error)
            __Pyx_GOTREF(__pyx_t_7);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_7, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3099, __pyx_L23_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":3096
 * 
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L24;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L23_error:;
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_6 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L8_except_error;
          }
          __pyx_L24:;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":3076
 *         self.enterRule(localctx, 52, self.RULE_declaration)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 442
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":3101
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3101, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_2) {
        __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3101, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      } else {
        __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3101, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_14 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3101, __pyx_L30_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_2) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3101, __pyx_L30_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3101, __pyx_L30_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
      __pyx_L30_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":3102
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class ModificationContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3071
 * 
 * 
 *     def declaration(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.DeclarationContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.declaration", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3106
 *     class ModificationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19ModificationContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19ModificationContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19ModificationContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19ModificationContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 3106, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 3106, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3106, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModificationContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19ModificationContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19ModificationContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":3107
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 3107, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3107, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3107, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3107, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3107, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3108
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 3108, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3106
 *     class ModificationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModificationContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3111
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_modification
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19ModificationContext_3getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19ModificationContext_3getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19ModificationContext_3getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19ModificationContext_3getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19ModificationContext_2getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19ModificationContext_2getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":3112
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_modification             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_modification); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3111
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_modification
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModificationContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3115
 * 
 * 
 *         def copyFrom(self, ctx:ParserRuleContext):             # <<<<<<<<<<<<<<
 *             super().copyFrom(ctx)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19ModificationContext_5copyFrom(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19ModificationContext_5copyFrom = {"copyFrom", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19ModificationContext_5copyFrom, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19ModificationContext_5copyFrom(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copyFrom (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx_2,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("copyFrom", 1, 2, 2, 1); __PYX_ERR(0, 3115, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "copyFrom") < 0)) __PYX_ERR(0, 3115, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("copyFrom", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3115, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModificationContext.copyFrom", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19ModificationContext_4copyFrom(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19ModificationContext_4copyFrom(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("copyFrom", 0);

  /* "pymola/generated/ModelicaParser.pyx":3116
 * 
 *         def copyFrom(self, ctx:ParserRuleContext):
 *             super().copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 3116, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3116, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3116, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3116, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3116, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3116, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3115
 * 
 * 
 *         def copyFrom(self, ctx:ParserRuleContext):             # <<<<<<<<<<<<<<
 *             super().copyFrom(ctx)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModificationContext.copyFrom", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3122
 *     class Modification_classContext(ModificationContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ModificationContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Modification_classContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_25Modification_classContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_25Modification_classContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Modification_classContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 3122, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 3122, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 3122, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3122, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Modification_classContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_25Modification_classContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Modification_classContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":3123
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ModificationContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 3123, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3123, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3123, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3123, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3123, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3124
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ModificationContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def class_modification(self):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3124, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3124, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3124, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3124, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3122
 *     class Modification_classContext(ModificationContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ModificationContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Modification_classContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3126
 *             self.copyFrom(ctx)
 * 
 *         def class_modification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_modificationContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Modification_classContext_3class_modification(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_25Modification_classContext_3class_modification = {"class_modification", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_25Modification_classContext_3class_modification, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Modification_classContext_3class_modification(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("class_modification (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_25Modification_classContext_2class_modification(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Modification_classContext_2class_modification(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("class_modification", 0);

  /* "pymola/generated/ModelicaParser.pyx":3127
 * 
 *         def class_modification(self):
 *             return self.getTypedRuleContext(ModelicaParser.Class_modificationContext,0)             # <<<<<<<<<<<<<<
 * 
 *         def expression(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Class_modificationContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3127, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3127, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3126
 *             self.copyFrom(ctx)
 * 
 *         def class_modification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_modificationContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Modification_classContext.class_modification", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3129
 *             return self.getTypedRuleContext(ModelicaParser.Class_modificationContext,0)
 * 
 *         def expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Modification_classContext_5expression(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_25Modification_classContext_5expression = {"expression", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_25Modification_classContext_5expression, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Modification_classContext_5expression(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("expression (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_25Modification_classContext_4expression(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Modification_classContext_4expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("expression", 0);

  /* "pymola/generated/ModelicaParser.pyx":3130
 * 
 *         def expression(self):
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_ExpressionContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3130, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3130, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3130, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3130, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3129
 *             return self.getTypedRuleContext(ModelicaParser.Class_modificationContext,0)
 * 
 *         def expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Modification_classContext.expression", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3133
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterModification_class" ):
 *                 listener.enterModification_class(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Modification_classContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_25Modification_classContext_7enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_25Modification_classContext_7enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Modification_classContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 3133, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 3133, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3133, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Modification_classContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_25Modification_classContext_6enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Modification_classContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":3134
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterModification_class" ):             # <<<<<<<<<<<<<<
 *                 listener.enterModification_class(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterModification_class); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3134, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3135
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterModification_class" ):
 *                 listener.enterModification_class(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterModification_class); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3135, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3135, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3135, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3135, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3135, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":3134
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterModification_class" ):             # <<<<<<<<<<<<<<
 *                 listener.enterModification_class(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3133
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterModification_class" ):
 *                 listener.enterModification_class(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Modification_classContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3137
 *                 listener.enterModification_class(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitModification_class" ):
 *                 listener.exitModification_class(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Modification_classContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_25Modification_classContext_9exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_25Modification_classContext_9exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Modification_classContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 3137, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 3137, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3137, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Modification_classContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_25Modification_classContext_8exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Modification_classContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":3138
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitModification_class" ):             # <<<<<<<<<<<<<<
 *                 listener.exitModification_class(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitModification_class); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3138, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3139
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitModification_class" ):
 *                 listener.exitModification_class(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitModification_class); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3139, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3139, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3139, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3139, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3139, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":3138
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitModification_class" ):             # <<<<<<<<<<<<<<
 *                 listener.exitModification_class(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3137
 *                 listener.enterModification_class(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitModification_class" ):
 *                 listener.exitModification_class(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Modification_classContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3141
 *                 listener.exitModification_class(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitModification_class" ):
 *                 return visitor.visitModification_class(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Modification_classContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_25Modification_classContext_11accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_25Modification_classContext_11accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Modification_classContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 3141, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 3141, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3141, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Modification_classContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_25Modification_classContext_10accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Modification_classContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":3142
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitModification_class" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitModification_class(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitModification_class); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3142, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3143
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitModification_class" ):
 *                 return visitor.visitModification_class(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitModification_class); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3143, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3143, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3143, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3143, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3143, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3143, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":3142
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitModification_class" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitModification_class(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3145
 *                 return visitor.visitModification_class(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3145, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3145, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3145, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3145, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3145, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3145, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":3141
 *                 listener.exitModification_class(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitModification_class" ):
 *                 return visitor.visitModification_class(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Modification_classContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3150
 *     class Modification_assignment2Context(ModificationContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ModificationContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_31Modification_assignment2Context_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_31Modification_assignment2Context_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_31Modification_assignment2Context_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_31Modification_assignment2Context_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 3150, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 3150, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 3150, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3150, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Modification_assignment2Context.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_31Modification_assignment2Context___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_31Modification_assignment2Context___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":3151
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ModificationContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 3151, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3151, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3151, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3151, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3151, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3152
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ModificationContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def expression(self):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3152, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3152, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3152, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3152, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3150
 *     class Modification_assignment2Context(ModificationContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ModificationContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Modification_assignment2Context.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3154
 *             self.copyFrom(ctx)
 * 
 *         def expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_31Modification_assignment2Context_3expression(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_31Modification_assignment2Context_3expression = {"expression", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_31Modification_assignment2Context_3expression, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_31Modification_assignment2Context_3expression(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("expression (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_31Modification_assignment2Context_2expression(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_31Modification_assignment2Context_2expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("expression", 0);

  /* "pymola/generated/ModelicaParser.pyx":3155
 * 
 *         def expression(self):
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_ExpressionContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3155, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3155, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3155, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3155, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3154
 *             self.copyFrom(ctx)
 * 
 *         def expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Modification_assignment2Context.expression", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3158
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterModification_assignment2" ):
 *                 listener.enterModification_assignment2(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_31Modification_assignment2Context_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_31Modification_assignment2Context_5enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_31Modification_assignment2Context_5enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_31Modification_assignment2Context_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 3158, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 3158, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3158, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Modification_assignment2Context.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_31Modification_assignment2Context_4enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_31Modification_assignment2Context_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":3159
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterModification_assignment2" ):             # <<<<<<<<<<<<<<
 *                 listener.enterModification_assignment2(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterModification_assignment2); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3159, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3160
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterModification_assignment2" ):
 *                 listener.enterModification_assignment2(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterModification_assignment2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3160, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3160, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3160, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3160, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3160, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3160, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":3159
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterModification_assignment2" ):             # <<<<<<<<<<<<<<
 *                 listener.enterModification_assignment2(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3158
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterModification_assignment2" ):
 *                 listener.enterModification_assignment2(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Modification_assignment2Context.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3162
 *                 listener.enterModification_assignment2(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitModification_assignment2" ):
 *                 listener.exitModification_assignment2(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_31Modification_assignment2Context_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_31Modification_assignment2Context_7exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_31Modification_assignment2Context_7exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_31Modification_assignment2Context_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 3162, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 3162, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3162, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Modification_assignment2Context.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_31Modification_assignment2Context_6exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_31Modification_assignment2Context_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":3163
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitModification_assignment2" ):             # <<<<<<<<<<<<<<
 *                 listener.exitModification_assignment2(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitModification_assignment2); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3163, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3164
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitModification_assignment2" ):
 *                 listener.exitModification_assignment2(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitModification_assignment2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3164, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3164, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3164, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3164, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3164, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":3163
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitModification_assignment2" ):             # <<<<<<<<<<<<<<
 *                 listener.exitModification_assignment2(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3162
 *                 listener.enterModification_assignment2(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitModification_assignment2" ):
 *                 listener.exitModification_assignment2(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Modification_assignment2Context.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3166
 *                 listener.exitModification_assignment2(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitModification_assignment2" ):
 *                 return visitor.visitModification_assignment2(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_31Modification_assignment2Context_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_31Modification_assignment2Context_9accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_31Modification_assignment2Context_9accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_31Modification_assignment2Context_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 3166, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 3166, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3166, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Modification_assignment2Context.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_31Modification_assignment2Context_8accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_31Modification_assignment2Context_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":3167
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitModification_assignment2" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitModification_assignment2(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitModification_assignment2); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3167, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3168
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitModification_assignment2" ):
 *                 return visitor.visitModification_assignment2(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitModification_assignment2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3168, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3168, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3168, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3168, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3168, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3168, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":3167
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitModification_assignment2" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitModification_assignment2(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3170
 *                 return visitor.visitModification_assignment2(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3170, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3170, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3170, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3170, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3170, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3170, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":3166
 *                 listener.exitModification_assignment2(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitModification_assignment2" ):
 *                 return visitor.visitModification_assignment2(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Modification_assignment2Context.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3175
 *     class Modification_assignmentContext(ModificationContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ModificationContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Modification_assignmentContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_30Modification_assignmentContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_30Modification_assignmentContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Modification_assignmentContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 3175, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 3175, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 3175, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3175, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Modification_assignmentContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_30Modification_assignmentContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Modification_assignmentContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":3176
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ModificationContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 3176, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3176, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3176, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3176, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3176, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3176, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3177
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ModificationContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def expression(self):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3177, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3177, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3177, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3177, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3175
 *     class Modification_assignmentContext(ModificationContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ModificationContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Modification_assignmentContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3179
 *             self.copyFrom(ctx)
 * 
 *         def expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Modification_assignmentContext_3expression(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_30Modification_assignmentContext_3expression = {"expression", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_30Modification_assignmentContext_3expression, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Modification_assignmentContext_3expression(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("expression (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_30Modification_assignmentContext_2expression(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Modification_assignmentContext_2expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("expression", 0);

  /* "pymola/generated/ModelicaParser.pyx":3180
 * 
 *         def expression(self):
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_ExpressionContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3180, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3180, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3180, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3180, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3179
 *             self.copyFrom(ctx)
 * 
 *         def expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Modification_assignmentContext.expression", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3183
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterModification_assignment" ):
 *                 listener.enterModification_assignment(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Modification_assignmentContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_30Modification_assignmentContext_5enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_30Modification_assignmentContext_5enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Modification_assignmentContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 3183, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 3183, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3183, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Modification_assignmentContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_30Modification_assignmentContext_4enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Modification_assignmentContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":3184
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterModification_assignment" ):             # <<<<<<<<<<<<<<
 *                 listener.enterModification_assignment(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterModification_assignment); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3184, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3185
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterModification_assignment" ):
 *                 listener.enterModification_assignment(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterModification_assignment); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3185, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3185, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3185, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3185, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3185, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3185, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":3184
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterModification_assignment" ):             # <<<<<<<<<<<<<<
 *                 listener.enterModification_assignment(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3183
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterModification_assignment" ):
 *                 listener.enterModification_assignment(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Modification_assignmentContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3187
 *                 listener.enterModification_assignment(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitModification_assignment" ):
 *                 listener.exitModification_assignment(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Modification_assignmentContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_30Modification_assignmentContext_7exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_30Modification_assignmentContext_7exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Modification_assignmentContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 3187, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 3187, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3187, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Modification_assignmentContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_30Modification_assignmentContext_6exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Modification_assignmentContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":3188
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitModification_assignment" ):             # <<<<<<<<<<<<<<
 *                 listener.exitModification_assignment(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitModification_assignment); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3188, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3189
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitModification_assignment" ):
 *                 listener.exitModification_assignment(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitModification_assignment); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3189, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3189, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3189, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3189, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3189, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3189, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":3188
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitModification_assignment" ):             # <<<<<<<<<<<<<<
 *                 listener.exitModification_assignment(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3187
 *                 listener.enterModification_assignment(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitModification_assignment" ):
 *                 listener.exitModification_assignment(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Modification_assignmentContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3191
 *                 listener.exitModification_assignment(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitModification_assignment" ):
 *                 return visitor.visitModification_assignment(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Modification_assignmentContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_30Modification_assignmentContext_9accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_30Modification_assignmentContext_9accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Modification_assignmentContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 3191, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 3191, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3191, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Modification_assignmentContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_30Modification_assignmentContext_8accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Modification_assignmentContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":3192
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitModification_assignment" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitModification_assignment(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitModification_assignment); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3192, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3193
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitModification_assignment" ):
 *                 return visitor.visitModification_assignment(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitModification_assignment); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3193, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3193, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3193, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3193, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3193, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3193, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":3192
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitModification_assignment" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitModification_assignment(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3195
 *                 return visitor.visitModification_assignment(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3195, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3195, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3195, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3195, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3195, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3195, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":3191
 *                 listener.exitModification_assignment(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitModification_assignment" ):
 *                 return visitor.visitModification_assignment(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Modification_assignmentContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3199
 * 
 * 
 *     def modification(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.ModificationContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_57modification(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_57modification = {"modification", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_57modification, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_57modification(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("modification (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_56modification(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_56modification(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v_token = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  char const *__pyx_t_22;
  __Pyx_RefNannySetupContext("modification", 0);

  /* "pymola/generated/ModelicaParser.pyx":3201
 *     def modification(self):
 * 
 *         localctx = ModelicaParser.ModificationContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 54, self.RULE_modification)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_ModificationContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3201, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3201, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3201, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3201, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3202
 * 
 *         localctx = ModelicaParser.ModificationContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 54, self.RULE_modification)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_modification); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_54, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3202, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_54, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3202, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3202, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_54);
    __Pyx_GIVEREF(__pyx_int_54);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_54);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3202, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3203
 *         localctx = ModelicaParser.ModificationContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 54, self.RULE_modification)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.state = 458
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 3203, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3204
 *         self.enterRule(localctx, 54, self.RULE_modification)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.state = 458
 *             self._errHandler.sync(self)
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":3205
 *         self._la = 0 # Token type
 *         try:
 *             self.state = 458             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_458) < 0) __PYX_ERR(0, 3205, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":3206
 *         try:
 *             self.state = 458
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.T__16]:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3206, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3206, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3206, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3206, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3206, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3206, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3206, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3207
 *             self.state = 458
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if token in [ModelicaParser.T__16]:
 *                 localctx = ModelicaParser.Modification_classContext(self, localctx)
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3207, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3207, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__872, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3207, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_v_token = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3208
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.T__16]:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Modification_classContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 1)
 */
        __Pyx_INCREF(__pyx_v_token);
        __pyx_t_1 = __pyx_v_token;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3208, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__16); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3208, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3208, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 3208, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_12 = (__pyx_t_11 != 0);
        if (__pyx_t_12) {

          /* "pymola/generated/ModelicaParser.pyx":3209
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.T__16]:
 *                 localctx = ModelicaParser.Modification_classContext(self, localctx)             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 1)
 *                 self.state = 449
 */
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3209, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Modification_classContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3209, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3209, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3209, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3209, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_2) {
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_v_localctx);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3209, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3210
 *             if token in [ModelicaParser.T__16]:
 *                 localctx = ModelicaParser.Modification_classContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *                 self.state = 449
 *                 self.class_modification()
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3210, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_1};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3210, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_1};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3210, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3210, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_1);
            __Pyx_GIVEREF(__pyx_int_1);
            PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_1);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3210, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3211
 *                 localctx = ModelicaParser.Modification_classContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 1)
 *                 self.state = 449             # <<<<<<<<<<<<<<
 *                 self.class_modification()
 *                 self.state = 452
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_449) < 0) __PYX_ERR(0, 3211, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":3212
 *                 self.enterOuterAlt(localctx, 1)
 *                 self.state = 449
 *                 self.class_modification()             # <<<<<<<<<<<<<<
 *                 self.state = 452
 *                 self._errHandler.sync(self)
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_class_modification); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3212, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3212, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3212, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3213
 *                 self.state = 449
 *                 self.class_modification()
 *                 self.state = 452             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_452) < 0) __PYX_ERR(0, 3213, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":3214
 *                 self.class_modification()
 *                 self.state = 452
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__14:
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3214, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3214, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3214, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3214, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3214, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3214, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3214, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3215
 *                 self.state = 452
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if _la==ModelicaParser.T__14:
 *                     self.state = 450
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3215, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3215, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__873, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3215, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_v__la = __pyx_t_1;
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3216
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__14:             # <<<<<<<<<<<<<<
 *                     self.state = 450
 *                     self.match(ModelicaParser.T__14)
 */
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3216, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__14); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3216, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3216, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3216, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (__pyx_t_12) {

            /* "pymola/generated/ModelicaParser.pyx":3217
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__14:
 *                     self.state = 450             # <<<<<<<<<<<<<<
 *                     self.match(ModelicaParser.T__14)
 *                     self.state = 451
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_450) < 0) __PYX_ERR(0, 3217, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":3218
 *                 if _la==ModelicaParser.T__14:
 *                     self.state = 450
 *                     self.match(ModelicaParser.T__14)             # <<<<<<<<<<<<<<
 *                     self.state = 451
 *                     self.expression()
 */
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3218, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3218, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__14); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3218, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_3)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_3);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
              }
            }
            if (!__pyx_t_3) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3218, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_2)) {
                PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_7};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3218, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
                PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_7};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3218, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              } else
              #endif
              {
                __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3218, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __pyx_t_3 = NULL;
                __Pyx_GIVEREF(__pyx_t_7);
                PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_7);
                __pyx_t_7 = 0;
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3218, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":3219
 *                     self.state = 450
 *                     self.match(ModelicaParser.T__14)
 *                     self.state = 451             # <<<<<<<<<<<<<<
 *                     self.expression()
 * 
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_451) < 0) __PYX_ERR(0, 3219, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":3220
 *                     self.match(ModelicaParser.T__14)
 *                     self.state = 451
 *                     self.expression()             # <<<<<<<<<<<<<<
 * 
 * 
 */
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expression); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3220, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_4 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_4)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_4);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
              }
            }
            if (__pyx_t_4) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3220, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            } else {
              __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3220, __pyx_L6_error)
            }
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":3216
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__14:             # <<<<<<<<<<<<<<
 *                     self.state = 450
 *                     self.match(ModelicaParser.T__14)
 */
          }

          /* "pymola/generated/ModelicaParser.pyx":3208
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.T__16]:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Modification_classContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 1)
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":3224
 * 
 *                 pass
 *             elif token in [ModelicaParser.T__14]:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Modification_assignmentContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 2)
 */
        __Pyx_INCREF(__pyx_v_token);
        __pyx_t_1 = __pyx_v_token;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3224, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__14); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3224, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3224, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3224, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_11 = (__pyx_t_12 != 0);
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":3225
 *                 pass
 *             elif token in [ModelicaParser.T__14]:
 *                 localctx = ModelicaParser.Modification_assignmentContext(self, localctx)             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 2)
 *                 self.state = 454
 */
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3225, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Modification_assignmentContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3225, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3225, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3225, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3225, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            if (__pyx_t_2) {
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_localctx);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3225, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3226
 *             elif token in [ModelicaParser.T__14]:
 *                 localctx = ModelicaParser.Modification_assignmentContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 2)             # <<<<<<<<<<<<<<
 *                 self.state = 454
 *                 self.match(ModelicaParser.T__14)
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3226, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_localctx, __pyx_int_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3226, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_localctx, __pyx_int_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3226, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3226, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (__pyx_t_7) {
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_7); __pyx_t_7 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_2);
            __Pyx_GIVEREF(__pyx_int_2);
            PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_2);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3226, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3227
 *                 localctx = ModelicaParser.Modification_assignmentContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 2)
 *                 self.state = 454             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__14)
 *                 self.state = 455
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_454) < 0) __PYX_ERR(0, 3227, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":3228
 *                 self.enterOuterAlt(localctx, 2)
 *                 self.state = 454
 *                 self.match(ModelicaParser.T__14)             # <<<<<<<<<<<<<<
 *                 self.state = 455
 *                 self.expression()
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3228, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3228, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__14); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3228, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3228, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3228, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3228, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3228, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_GIVEREF(__pyx_t_7);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_t_7);
              __pyx_t_7 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3228, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3229
 *                 self.state = 454
 *                 self.match(ModelicaParser.T__14)
 *                 self.state = 455             # <<<<<<<<<<<<<<
 *                 self.expression()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_455) < 0) __PYX_ERR(0, 3229, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":3230
 *                 self.match(ModelicaParser.T__14)
 *                 self.state = 455
 *                 self.expression()             # <<<<<<<<<<<<<<
 *                 pass
 *             elif token in [ModelicaParser.T__40]:
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expression); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3230, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (__pyx_t_3) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3230, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3230, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3224
 * 
 *                 pass
 *             elif token in [ModelicaParser.T__14]:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Modification_assignmentContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 2)
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":3232
 *                 self.expression()
 *                 pass
 *             elif token in [ModelicaParser.T__40]:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Modification_assignment2Context(self, localctx)
 *                 self.enterOuterAlt(localctx, 3)
 */
        __Pyx_INCREF(__pyx_v_token);
        __pyx_t_1 = __pyx_v_token;
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3232, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__40); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3232, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3232, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 3232, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_12 = (__pyx_t_11 != 0);
        if (__pyx_t_12) {

          /* "pymola/generated/ModelicaParser.pyx":3233
 *                 pass
 *             elif token in [ModelicaParser.T__40]:
 *                 localctx = ModelicaParser.Modification_assignment2Context(self, localctx)             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 3)
 *                 self.state = 456
 */
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3233, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_Modification_assignment2Context); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3233, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3233, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3233, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3233, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            if (__pyx_t_4) {
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4); __pyx_t_4 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_localctx);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3233, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3234
 *             elif token in [ModelicaParser.T__40]:
 *                 localctx = ModelicaParser.Modification_assignment2Context(self, localctx)
 *                 self.enterOuterAlt(localctx, 3)             # <<<<<<<<<<<<<<
 *                 self.state = 456
 *                 self.match(ModelicaParser.T__40)
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3234, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_7 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_localctx, __pyx_int_3};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3234, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_localctx, __pyx_int_3};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3234, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3234, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (__pyx_t_7) {
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_3);
            __Pyx_GIVEREF(__pyx_int_3);
            PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_int_3);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3234, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3235
 *                 localctx = ModelicaParser.Modification_assignment2Context(self, localctx)
 *                 self.enterOuterAlt(localctx, 3)
 *                 self.state = 456             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__40)
 *                 self.state = 457
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_456) < 0) __PYX_ERR(0, 3235, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":3236
 *                 self.enterOuterAlt(localctx, 3)
 *                 self.state = 456
 *                 self.match(ModelicaParser.T__40)             # <<<<<<<<<<<<<<
 *                 self.state = 457
 *                 self.expression()
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3236, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3236, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__40); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3236, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (!__pyx_t_4) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3236, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3236, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3236, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            {
              __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3236, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
              __Pyx_GIVEREF(__pyx_t_7);
              PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_7);
              __pyx_t_7 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3236, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3237
 *                 self.state = 456
 *                 self.match(ModelicaParser.T__40)
 *                 self.state = 457             # <<<<<<<<<<<<<<
 *                 self.expression()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_457) < 0) __PYX_ERR(0, 3237, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":3238
 *                 self.match(ModelicaParser.T__40)
 *                 self.state = 457
 *                 self.expression()             # <<<<<<<<<<<<<<
 *                 pass
 *             else:
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expression); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3238, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3238, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3238, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3232
 *                 self.expression()
 *                 pass
 *             elif token in [ModelicaParser.T__40]:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Modification_assignment2Context(self, localctx)
 *                 self.enterOuterAlt(localctx, 3)
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":3241
 *                 pass
 *             else:
 *                 raise NoViableAltException(self)             # <<<<<<<<<<<<<<
 * 
 *         except RecognitionException as re:
 */
        /*else*/ {
          __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_NoViableAltException); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3241, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3241, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3241, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3241, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3241, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3241, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_Raise(__pyx_t_1, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __PYX_ERR(0, 3241, __pyx_L6_error)
        }
        __pyx_L14:;

        /* "pymola/generated/ModelicaParser.pyx":3204
 *         self.enterRule(localctx, 54, self.RULE_modification)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.state = 458
 *             self._errHandler.sync(self)
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":3243
 *                 raise NoViableAltException(self)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3243, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.modification", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_3, &__pyx_t_7) < 0) __PYX_ERR(0, 3243, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_3);
        __pyx_v_re = __pyx_t_3;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":3244
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 3244, __pyx_L21_error)

          /* "pymola/generated/ModelicaParser.pyx":3245
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3245, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3245, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3245, __pyx_L21_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3245, __pyx_L21_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3245, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_4) {
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_4); __pyx_t_4 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_13, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3245, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3246
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3246, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3246, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3246, __pyx_L21_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3246, __pyx_L21_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3246, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3246, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":3243
 *                 raise NoViableAltException(self)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L22;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L21_error:;
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_6 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L8_except_error;
          }
          __pyx_L22:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":3204
 *         self.enterRule(localctx, 54, self.RULE_modification)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.state = 458
 *             self._errHandler.sync(self)
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":3248
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3248, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3248, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3248, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_14 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3248, __pyx_L28_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3248, __pyx_L28_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3248, __pyx_L28_error)
        }
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
      __pyx_L28_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":3249
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Class_modificationContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3199
 * 
 * 
 *     def modification(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.ModificationContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.modification", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v_token);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3253
 *     class Class_modificationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_modificationContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_25Class_modificationContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_modificationContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_modificationContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 3253, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 3253, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3253, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_modificationContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_25Class_modificationContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Class_modificationContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":3254
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 3254, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3254, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3254, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3254, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3254, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3255
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def argument_list(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 3255, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3253
 *     class Class_modificationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_modificationContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3257
 *             self.parser = parser
 * 
 *         def argument_list(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Argument_listContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_modificationContext_3argument_list(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_25Class_modificationContext_3argument_list = {"argument_list", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_modificationContext_3argument_list, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_modificationContext_3argument_list(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("argument_list (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_25Class_modificationContext_2argument_list(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Class_modificationContext_2argument_list(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("argument_list", 0);

  /* "pymola/generated/ModelicaParser.pyx":3258
 * 
 *         def argument_list(self):
 *             return self.getTypedRuleContext(ModelicaParser.Argument_listContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Argument_listContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3258, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3258, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3258, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3258, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3257
 *             self.parser = parser
 * 
 *         def argument_list(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Argument_listContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_modificationContext.argument_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3261
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_class_modification
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_modificationContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_25Class_modificationContext_5getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_modificationContext_5getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_modificationContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_25Class_modificationContext_4getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Class_modificationContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":3262
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_class_modification             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_class_modification); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3261
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_class_modification
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_modificationContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3264
 *             return ModelicaParser.RULE_class_modification
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterClass_modification" ):
 *                 listener.enterClass_modification(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_modificationContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_25Class_modificationContext_7enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_modificationContext_7enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_modificationContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 3264, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 3264, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3264, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_modificationContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_25Class_modificationContext_6enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Class_modificationContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":3265
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterClass_modification" ):             # <<<<<<<<<<<<<<
 *                 listener.enterClass_modification(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterClass_modification); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3265, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3266
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterClass_modification" ):
 *                 listener.enterClass_modification(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterClass_modification); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3266, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3266, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3266, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3266, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3266, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3266, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":3265
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterClass_modification" ):             # <<<<<<<<<<<<<<
 *                 listener.enterClass_modification(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3264
 *             return ModelicaParser.RULE_class_modification
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterClass_modification" ):
 *                 listener.enterClass_modification(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_modificationContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3268
 *                 listener.enterClass_modification(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitClass_modification" ):
 *                 listener.exitClass_modification(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_modificationContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_25Class_modificationContext_9exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_modificationContext_9exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_modificationContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 3268, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 3268, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3268, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_modificationContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_25Class_modificationContext_8exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Class_modificationContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":3269
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitClass_modification" ):             # <<<<<<<<<<<<<<
 *                 listener.exitClass_modification(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitClass_modification); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3269, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3270
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitClass_modification" ):
 *                 listener.exitClass_modification(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitClass_modification); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3270, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3270, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3270, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3270, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3270, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3270, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":3269
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitClass_modification" ):             # <<<<<<<<<<<<<<
 *                 listener.exitClass_modification(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3268
 *                 listener.enterClass_modification(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitClass_modification" ):
 *                 listener.exitClass_modification(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_modificationContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3272
 *                 listener.exitClass_modification(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitClass_modification" ):
 *                 return visitor.visitClass_modification(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_modificationContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_25Class_modificationContext_11accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_modificationContext_11accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Class_modificationContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 3272, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 3272, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3272, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_modificationContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_25Class_modificationContext_10accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Class_modificationContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":3273
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitClass_modification" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitClass_modification(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitClass_modification); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3273, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3274
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitClass_modification" ):
 *                 return visitor.visitClass_modification(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitClass_modification); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3274, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3274, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3274, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3274, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3274, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":3273
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitClass_modification" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitClass_modification(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3276
 *                 return visitor.visitClass_modification(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3276, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3276, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3276, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3276, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3276, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":3272
 *                 listener.exitClass_modification(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitClass_modification" ):
 *                 return visitor.visitClass_modification(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Class_modificationContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3281
 * 
 * 
 *     def class_modification(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Class_modificationContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_59class_modification(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_59class_modification = {"class_modification", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_59class_modification, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_59class_modification(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("class_modification (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_58class_modification(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_58class_modification(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  char const *__pyx_t_22;
  __Pyx_RefNannySetupContext("class_modification", 0);

  /* "pymola/generated/ModelicaParser.pyx":3283
 *     def class_modification(self):
 * 
 *         localctx = ModelicaParser.Class_modificationContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 56, self.RULE_class_modification)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Class_modificationContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3283, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3283, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3283, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3283, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3284
 * 
 *         localctx = ModelicaParser.Class_modificationContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 56, self.RULE_class_modification)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_class_modification); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_56, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3284, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_56, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3284, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3284, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_56);
    __Pyx_GIVEREF(__pyx_int_56);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_56);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3284, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3285
 *         localctx = ModelicaParser.Class_modificationContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 56, self.RULE_class_modification)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 3285, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3286
 *         self.enterRule(localctx, 56, self.RULE_class_modification)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 460
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":3287
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 460
 *             self.match(ModelicaParser.T__16)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3287, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3287, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3287, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3287, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3287, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3288
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 460             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__16)
 *             self.state = 462
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_460) < 0) __PYX_ERR(0, 3288, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":3289
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 460
 *             self.match(ModelicaParser.T__16)             # <<<<<<<<<<<<<<
 *             self.state = 462
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3289, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3289, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__16); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3289, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3289, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3289, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3289, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3289, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3289, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3290
 *             self.state = 460
 *             self.match(ModelicaParser.T__16)
 *             self.state = 462             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_462) < 0) __PYX_ERR(0, 3290, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":3291
 *             self.match(ModelicaParser.T__16)
 *             self.state = 462
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if ((((_la - 26)) & ~0x3f) == 0 and ((1 << (_la - 26)) & ((1 << (ModelicaParser.T__25 - 26)) | (1 << (ModelicaParser.EACH - 26)) | (1 << (ModelicaParser.FINAL - 26)) | (1 << (ModelicaParser.REDECLARE - 26)) | (1 << (ModelicaParser.IDENT - 26)))) != 0):
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3291, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3291, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3291, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3291, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3291, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3291, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3291, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3292
 *             self.state = 462
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if ((((_la - 26)) & ~0x3f) == 0 and ((1 << (_la - 26)) & ((1 << (ModelicaParser.T__25 - 26)) | (1 << (ModelicaParser.EACH - 26)) | (1 << (ModelicaParser.FINAL - 26)) | (1 << (ModelicaParser.REDECLARE - 26)) | (1 << (ModelicaParser.IDENT - 26)))) != 0):
 *                 self.state = 461
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3292, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3292, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__874, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3292, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_v__la = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3293
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if ((((_la - 26)) & ~0x3f) == 0 and ((1 << (_la - 26)) & ((1 << (ModelicaParser.T__25 - 26)) | (1 << (ModelicaParser.EACH - 26)) | (1 << (ModelicaParser.FINAL - 26)) | (1 << (ModelicaParser.REDECLARE - 26)) | (1 << (ModelicaParser.IDENT - 26)))) != 0):             # <<<<<<<<<<<<<<
 *                 self.state = 461
 *                 self.argument_list()
 */
        __pyx_t_1 = __Pyx_PyInt_SubtractObjC(__pyx_v__la, __pyx_int_26, 26, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3293, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_PyInt_From_long((~0x3f)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3293, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = PyNumber_And(__pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3293, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyInt_EqObjC(__pyx_t_2, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3293, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3293, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_PyInt_SubtractObjC(__pyx_v__la, __pyx_int_26, 26, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3293, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3293, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3293, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__25); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3293, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyInt_SubtractObjC(__pyx_t_1, __pyx_int_26, 26, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3293, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyNumber_Lshift(__pyx_int_1, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3293, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3293, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_EACH); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3293, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyInt_SubtractObjC(__pyx_t_3, __pyx_int_26, 26, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3293, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_Lshift(__pyx_int_1, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3293, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyNumber_Or(__pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3293, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3293, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_FINAL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3293, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyInt_SubtractObjC(__pyx_t_1, __pyx_int_26, 26, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3293, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3293, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_Or(__pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3293, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3293, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_REDECLARE); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3293, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyInt_SubtractObjC(__pyx_t_4, __pyx_int_26, 26, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3293, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyNumber_Lshift(__pyx_int_1, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3293, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyNumber_Or(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3293, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3293, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3293, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyInt_SubtractObjC(__pyx_t_3, __pyx_int_26, 26, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3293, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_Lshift(__pyx_int_1, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3293, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyNumber_Or(__pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3293, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_And(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3293, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_3, __pyx_int_0, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3293, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3293, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __pyx_t_12;
        __pyx_L15_bool_binop_done:;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":3294
 *             _la = self._input.LA(1)
 *             if ((((_la - 26)) & ~0x3f) == 0 and ((1 << (_la - 26)) & ((1 << (ModelicaParser.T__25 - 26)) | (1 << (ModelicaParser.EACH - 26)) | (1 << (ModelicaParser.FINAL - 26)) | (1 << (ModelicaParser.REDECLARE - 26)) | (1 << (ModelicaParser.IDENT - 26)))) != 0):
 *                 self.state = 461             # <<<<<<<<<<<<<<
 *                 self.argument_list()
 * 
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_461) < 0) __PYX_ERR(0, 3294, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":3295
 *             if ((((_la - 26)) & ~0x3f) == 0 and ((1 << (_la - 26)) & ((1 << (ModelicaParser.T__25 - 26)) | (1 << (ModelicaParser.EACH - 26)) | (1 << (ModelicaParser.FINAL - 26)) | (1 << (ModelicaParser.REDECLARE - 26)) | (1 << (ModelicaParser.IDENT - 26)))) != 0):
 *                 self.state = 461
 *                 self.argument_list()             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_argument_list); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3295, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (__pyx_t_2) {
            __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3295, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else {
            __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3295, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3293
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if ((((_la - 26)) & ~0x3f) == 0 and ((1 << (_la - 26)) & ((1 << (ModelicaParser.T__25 - 26)) | (1 << (ModelicaParser.EACH - 26)) | (1 << (ModelicaParser.FINAL - 26)) | (1 << (ModelicaParser.REDECLARE - 26)) | (1 << (ModelicaParser.IDENT - 26)))) != 0):             # <<<<<<<<<<<<<<
 *                 self.state = 461
 *                 self.argument_list()
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":3298
 * 
 * 
 *             self.state = 464             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__18)
 *         except RecognitionException as re:
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_464) < 0) __PYX_ERR(0, 3298, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":3299
 * 
 *             self.state = 464
 *             self.match(ModelicaParser.T__18)             # <<<<<<<<<<<<<<
 *         except RecognitionException as re:
 *             localctx.exception = re
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3299, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3299, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__18); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3299, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_2) {
          __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3299, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_GOTREF(__pyx_t_4);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_1};
            __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3299, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_1};
            __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3299, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3299, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
            __Pyx_GIVEREF(__pyx_t_1);
            PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_1);
            __pyx_t_1 = 0;
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3299, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3286
 *         self.enterRule(localctx, 56, self.RULE_class_modification)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 460
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "pymola/generated/ModelicaParser.pyx":3300
 *             self.state = 464
 *             self.match(ModelicaParser.T__18)
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3300, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.class_modification", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_3, &__pyx_t_7) < 0) __PYX_ERR(0, 3300, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_3);
        __pyx_v_re = __pyx_t_3;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":3301
 *             self.match(ModelicaParser.T__18)
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 3301, __pyx_L22_error)

          /* "pymola/generated/ModelicaParser.pyx":3302
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3302, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3302, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_re};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3302, __pyx_L22_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_re};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3302, __pyx_L22_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3302, __pyx_L22_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_2) {
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_2); __pyx_t_2 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_13, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3302, __pyx_L22_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3303
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3303, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3303, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3303, __pyx_L22_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3303, __pyx_L22_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3303, __pyx_L22_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3303, __pyx_L22_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":3300
 *             self.state = 464
 *             self.match(ModelicaParser.T__18)
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L23;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L22_error:;
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_6 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L8_except_error;
          }
          __pyx_L23:;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":3286
 *         self.enterRule(localctx, 56, self.RULE_class_modification)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 460
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":3305
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3305, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3305, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3305, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_14 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3305, __pyx_L29_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_4) {
          __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3305, __pyx_L29_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else {
          __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3305, __pyx_L29_error)
        }
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
      __pyx_L29_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":3306
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Argument_listContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3281
 * 
 * 
 *     def class_modification(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Class_modificationContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.class_modification", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3310
 *     class Argument_listContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Argument_listContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20Argument_listContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20Argument_listContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Argument_listContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 3310, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 3310, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3310, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Argument_listContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20Argument_listContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Argument_listContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":3311
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 3311, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3311, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3311, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3311, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3311, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3312
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def argument(self, i:int=None):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 3312, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3310
 *     class Argument_listContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Argument_listContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3314
 *             self.parser = parser
 * 
 *         def argument(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ArgumentContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Argument_listContext_3argument(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20Argument_listContext_3argument = {"argument", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20Argument_listContext_3argument, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Argument_listContext_3argument(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("argument (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "argument") < 0)) __PYX_ERR(0, 3314, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("argument", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3314, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Argument_listContext.argument", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20Argument_listContext_2argument(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Argument_listContext_2argument(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("argument", 0);

  /* "pymola/generated/ModelicaParser.pyx":3315
 * 
 *         def argument(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.ArgumentContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3316
 *         def argument(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ArgumentContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.ArgumentContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3316, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3316, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_ArgumentContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3316, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3316, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3316, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3316, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3316, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3316, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":3315
 * 
 *         def argument(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.ArgumentContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3318
 *                 return self.getTypedRuleContexts(ModelicaParser.ArgumentContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.ArgumentContext,i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3318, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3318, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_ArgumentContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3318, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3318, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3318, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3318, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3318, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":3314
 *             self.parser = parser
 * 
 *         def argument(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ArgumentContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Argument_listContext.argument", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3321
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_argument_list
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Argument_listContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20Argument_listContext_5getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20Argument_listContext_5getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Argument_listContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20Argument_listContext_4getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Argument_listContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":3322
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_argument_list             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3322, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_argument_list); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3322, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3321
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_argument_list
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Argument_listContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3324
 *             return ModelicaParser.RULE_argument_list
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterArgument_list" ):
 *                 listener.enterArgument_list(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Argument_listContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20Argument_listContext_7enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20Argument_listContext_7enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Argument_listContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 3324, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 3324, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3324, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Argument_listContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20Argument_listContext_6enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Argument_listContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":3325
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterArgument_list" ):             # <<<<<<<<<<<<<<
 *                 listener.enterArgument_list(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterArgument_list); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3325, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3326
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterArgument_list" ):
 *                 listener.enterArgument_list(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterArgument_list); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3326, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3326, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3326, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3326, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3326, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3326, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":3325
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterArgument_list" ):             # <<<<<<<<<<<<<<
 *                 listener.enterArgument_list(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3324
 *             return ModelicaParser.RULE_argument_list
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterArgument_list" ):
 *                 listener.enterArgument_list(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Argument_listContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3328
 *                 listener.enterArgument_list(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitArgument_list" ):
 *                 listener.exitArgument_list(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Argument_listContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20Argument_listContext_9exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20Argument_listContext_9exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Argument_listContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 3328, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 3328, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3328, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Argument_listContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20Argument_listContext_8exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Argument_listContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":3329
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitArgument_list" ):             # <<<<<<<<<<<<<<
 *                 listener.exitArgument_list(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitArgument_list); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3329, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3330
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitArgument_list" ):
 *                 listener.exitArgument_list(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitArgument_list); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3330, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3330, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3330, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3330, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3330, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3330, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":3329
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitArgument_list" ):             # <<<<<<<<<<<<<<
 *                 listener.exitArgument_list(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3328
 *                 listener.enterArgument_list(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitArgument_list" ):
 *                 listener.exitArgument_list(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Argument_listContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3332
 *                 listener.exitArgument_list(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitArgument_list" ):
 *                 return visitor.visitArgument_list(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Argument_listContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20Argument_listContext_11accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20Argument_listContext_11accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Argument_listContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 3332, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 3332, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3332, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Argument_listContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20Argument_listContext_10accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Argument_listContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":3333
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitArgument_list" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitArgument_list(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitArgument_list); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3333, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3334
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitArgument_list" ):
 *                 return visitor.visitArgument_list(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitArgument_list); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3334, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3334, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3334, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3334, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3334, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3334, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":3333
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitArgument_list" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitArgument_list(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3336
 *                 return visitor.visitArgument_list(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3336, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3336, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3336, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3336, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3336, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3336, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":3332
 *                 listener.exitArgument_list(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitArgument_list" ):
 *                 return visitor.visitArgument_list(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Argument_listContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3341
 * 
 * 
 *     def argument_list(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Argument_listContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_61argument_list(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_61argument_list = {"argument_list", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_61argument_list, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_61argument_list(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("argument_list (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_60argument_list(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_60argument_list(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  char const *__pyx_t_21;
  __Pyx_RefNannySetupContext("argument_list", 0);

  /* "pymola/generated/ModelicaParser.pyx":3343
 *     def argument_list(self):
 * 
 *         localctx = ModelicaParser.Argument_listContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 58, self.RULE_argument_list)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Argument_listContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3343, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3343, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3343, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3343, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3344
 * 
 *         localctx = ModelicaParser.Argument_listContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 58, self.RULE_argument_list)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3344, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_argument_list); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3344, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_58, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3344, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_58, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3344, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3344, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_58);
    __Pyx_GIVEREF(__pyx_int_58);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_58);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3344, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3345
 *         localctx = ModelicaParser.Argument_listContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 58, self.RULE_argument_list)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 3345, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3346
 *         self.enterRule(localctx, 58, self.RULE_argument_list)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 466
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":3347
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 466
 *             self.argument()
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3347, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3347, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3347, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3347, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3347, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3348
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 466             # <<<<<<<<<<<<<<
 *             self.argument()
 *             self.state = 471
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_466) < 0) __PYX_ERR(0, 3348, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":3349
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 466
 *             self.argument()             # <<<<<<<<<<<<<<
 *             self.state = 471
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_argument); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3349, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3349, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3349, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3350
 *             self.state = 466
 *             self.argument()
 *             self.state = 471             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_471) < 0) __PYX_ERR(0, 3350, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":3351
 *             self.argument()
 *             self.state = 471
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             while _la==ModelicaParser.T__20:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3351, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3351, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3351, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3351, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3351, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3351, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3351, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3352
 *             self.state = 471
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             while _la==ModelicaParser.T__20:
 *                 self.state = 467
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3352, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3352, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__875, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3352, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v__la = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3353
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             while _la==ModelicaParser.T__20:             # <<<<<<<<<<<<<<
 *                 self.state = 467
 *                 self.match(ModelicaParser.T__20)
 */
        while (1) {
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3353, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__20); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3353, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3353, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 3353, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (!__pyx_t_11) break;

          /* "pymola/generated/ModelicaParser.pyx":3354
 *             _la = self._input.LA(1)
 *             while _la==ModelicaParser.T__20:
 *                 self.state = 467             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__20)
 *                 self.state = 468
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_467) < 0) __PYX_ERR(0, 3354, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":3355
 *             while _la==ModelicaParser.T__20:
 *                 self.state = 467
 *                 self.match(ModelicaParser.T__20)             # <<<<<<<<<<<<<<
 *                 self.state = 468
 *                 self.argument()
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3355, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3355, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__20); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3355, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3355, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3355, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3355, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3355, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3355, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3356
 *                 self.state = 467
 *                 self.match(ModelicaParser.T__20)
 *                 self.state = 468             # <<<<<<<<<<<<<<
 *                 self.argument()
 *                 self.state = 473
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_468) < 0) __PYX_ERR(0, 3356, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":3357
 *                 self.match(ModelicaParser.T__20)
 *                 self.state = 468
 *                 self.argument()             # <<<<<<<<<<<<<<
 *                 self.state = 473
 *                 self._errHandler.sync(self)
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_argument); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3357, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_4 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (__pyx_t_4) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3357, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3357, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3358
 *                 self.state = 468
 *                 self.argument()
 *                 self.state = 473             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_473) < 0) __PYX_ERR(0, 3358, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":3359
 *                 self.argument()
 *                 self.state = 473
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 * 
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3359, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_sync); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3359, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3359, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3359, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3359, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3359, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3359, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3360
 *                 self.state = 473
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *         except RecognitionException as re:
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3360, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3360, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__876, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3360, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_1);
          __pyx_t_1 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":3346
 *         self.enterRule(localctx, 58, self.RULE_argument_list)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 466
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":3362
 *                 _la = self._input.LA(1)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3362, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.argument_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_4, &__pyx_t_3) < 0) __PYX_ERR(0, 3362, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __pyx_v_re = __pyx_t_4;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":3363
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 3363, __pyx_L21_error)

          /* "pymola/generated/ModelicaParser.pyx":3364
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3364, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3364, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3364, __pyx_L21_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3364, __pyx_L21_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          {
            __pyx_t_12 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 3364, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_12);
            if (__pyx_t_2) {
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_2); __pyx_t_2 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_12, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3364, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3365
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3365, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 3365, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_12);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_12, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3365, __pyx_L21_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3365, __pyx_L21_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3365, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_2, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3365, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":3362
 *                 _la = self._input.LA(1)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L22;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L21_error:;
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __pyx_t_6 = __pyx_lineno; __pyx_t_13 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
            }
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_ErrRestore(__pyx_t_15, __pyx_t_16, __pyx_t_17);
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_13; __pyx_filename = __pyx_t_14;
            goto __pyx_L8_except_error;
          }
          __pyx_L22:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":3346
 *         self.enterRule(localctx, 58, self.RULE_argument_list)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 466
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":3367
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3367, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3367, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3367, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_19, &__pyx_t_18);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_18);
      __pyx_t_13 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_21 = __pyx_filename;
      {
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3367, __pyx_L28_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3367, __pyx_L28_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3367, __pyx_L28_error)
        }
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __pyx_lineno = __pyx_t_13; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_21;
      goto __pyx_L1_error;
      __pyx_L28_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":3368
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class ArgumentContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3341
 * 
 * 
 *     def argument_list(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Argument_listContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.argument_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3372
 *     class ArgumentContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15ArgumentContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15ArgumentContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15ArgumentContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15ArgumentContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 3372, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 3372, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3372, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ArgumentContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15ArgumentContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15ArgumentContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":3373
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 3373, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3373, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3373, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3373, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3373, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3374
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def element_modification_or_replaceable(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 3374, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3372
 *     class ArgumentContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ArgumentContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3376
 *             self.parser = parser
 * 
 *         def element_modification_or_replaceable(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Element_modification_or_replaceableContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15ArgumentContext_3element_modification_or_replaceable(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15ArgumentContext_3element_modification_or_replaceable = {"element_modification_or_replaceable", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15ArgumentContext_3element_modification_or_replaceable, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15ArgumentContext_3element_modification_or_replaceable(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("element_modification_or_replaceable (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15ArgumentContext_2element_modification_or_replaceable(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15ArgumentContext_2element_modification_or_replaceable(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("element_modification_or_replaceable", 0);

  /* "pymola/generated/ModelicaParser.pyx":3377
 * 
 *         def element_modification_or_replaceable(self):
 *             return self.getTypedRuleContext(ModelicaParser.Element_modification_or_replaceableContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Element_modification_or_replacea); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3377, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3377, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3377, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3377, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3376
 *             self.parser = parser
 * 
 *         def element_modification_or_replaceable(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Element_modification_or_replaceableContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ArgumentContext.element_modification_or_replaceable", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3380
 * 
 * 
 *         def element_redeclaration(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Element_redeclarationContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15ArgumentContext_5element_redeclaration(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15ArgumentContext_5element_redeclaration = {"element_redeclaration", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15ArgumentContext_5element_redeclaration, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15ArgumentContext_5element_redeclaration(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("element_redeclaration (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15ArgumentContext_4element_redeclaration(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15ArgumentContext_4element_redeclaration(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("element_redeclaration", 0);

  /* "pymola/generated/ModelicaParser.pyx":3381
 * 
 *         def element_redeclaration(self):
 *             return self.getTypedRuleContext(ModelicaParser.Element_redeclarationContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Element_redeclarationContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3381, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3381, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3381, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3381, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3380
 * 
 * 
 *         def element_redeclaration(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Element_redeclarationContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ArgumentContext.element_redeclaration", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3384
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_argument
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15ArgumentContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15ArgumentContext_7getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15ArgumentContext_7getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15ArgumentContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15ArgumentContext_6getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15ArgumentContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":3385
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_argument             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_argument); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3384
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_argument
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ArgumentContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3387
 *             return ModelicaParser.RULE_argument
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterArgument" ):
 *                 listener.enterArgument(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15ArgumentContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15ArgumentContext_9enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15ArgumentContext_9enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15ArgumentContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 3387, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 3387, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3387, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ArgumentContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15ArgumentContext_8enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15ArgumentContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":3388
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterArgument" ):             # <<<<<<<<<<<<<<
 *                 listener.enterArgument(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterArgument); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3388, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3389
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterArgument" ):
 *                 listener.enterArgument(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterArgument); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3389, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3389, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3389, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3389, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3389, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3389, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":3388
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterArgument" ):             # <<<<<<<<<<<<<<
 *                 listener.enterArgument(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3387
 *             return ModelicaParser.RULE_argument
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterArgument" ):
 *                 listener.enterArgument(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ArgumentContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3391
 *                 listener.enterArgument(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitArgument" ):
 *                 listener.exitArgument(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15ArgumentContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15ArgumentContext_11exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15ArgumentContext_11exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15ArgumentContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 3391, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 3391, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3391, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ArgumentContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15ArgumentContext_10exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15ArgumentContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":3392
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitArgument" ):             # <<<<<<<<<<<<<<
 *                 listener.exitArgument(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitArgument); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3392, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3393
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitArgument" ):
 *                 listener.exitArgument(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitArgument); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3393, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3393, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3393, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3393, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3393, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3393, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":3392
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitArgument" ):             # <<<<<<<<<<<<<<
 *                 listener.exitArgument(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3391
 *                 listener.enterArgument(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitArgument" ):
 *                 listener.exitArgument(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ArgumentContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3395
 *                 listener.exitArgument(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitArgument" ):
 *                 return visitor.visitArgument(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15ArgumentContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15ArgumentContext_13accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15ArgumentContext_13accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15ArgumentContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 3395, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 3395, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3395, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ArgumentContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15ArgumentContext_12accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15ArgumentContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":3396
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitArgument" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitArgument(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitArgument); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3396, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3397
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitArgument" ):
 *                 return visitor.visitArgument(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitArgument); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3397, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3397, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3397, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3397, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3397, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":3396
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitArgument" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitArgument(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3399
 *                 return visitor.visitArgument(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3399, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3399, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3399, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3399, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3399, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3399, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":3395
 *                 listener.exitArgument(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitArgument" ):
 *                 return visitor.visitArgument(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ArgumentContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3404
 * 
 * 
 *     def argument(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.ArgumentContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_63argument(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_63argument = {"argument", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_63argument, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_63argument(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("argument (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_62argument(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_62argument(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v_token = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  char const *__pyx_t_22;
  __Pyx_RefNannySetupContext("argument", 0);

  /* "pymola/generated/ModelicaParser.pyx":3406
 *     def argument(self):
 * 
 *         localctx = ModelicaParser.ArgumentContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 60, self.RULE_argument)
 *         try:
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_ArgumentContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3406, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3406, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3406, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3406, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3407
 * 
 *         localctx = ModelicaParser.ArgumentContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 60, self.RULE_argument)             # <<<<<<<<<<<<<<
 *         try:
 *             self.state = 476
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_argument); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_60, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3407, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_60, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3407, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3407, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_60);
    __Pyx_GIVEREF(__pyx_int_60);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_60);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3407, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3408
 *         localctx = ModelicaParser.ArgumentContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 60, self.RULE_argument)
 *         try:             # <<<<<<<<<<<<<<
 *             self.state = 476
 *             self._errHandler.sync(self)
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":3409
 *         self.enterRule(localctx, 60, self.RULE_argument)
 *         try:
 *             self.state = 476             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_476) < 0) __PYX_ERR(0, 3409, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":3410
 *         try:
 *             self.state = 476
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.T__25, ModelicaParser.EACH, ModelicaParser.FINAL, ModelicaParser.IDENT]:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3410, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3410, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3410, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3410, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3410, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3410, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3410, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3411
 *             self.state = 476
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if token in [ModelicaParser.T__25, ModelicaParser.EACH, ModelicaParser.FINAL, ModelicaParser.IDENT]:
 *                 self.enterOuterAlt(localctx, 1)
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3411, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3411, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__877, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3411, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_v_token = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3412
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.T__25, ModelicaParser.EACH, ModelicaParser.FINAL, ModelicaParser.IDENT]:             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 1)
 *                 self.state = 474
 */
        __Pyx_INCREF(__pyx_v_token);
        __pyx_t_1 = __pyx_v_token;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3412, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__25); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3412, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3412, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3412, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3412, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_EACH); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3412, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3412, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3412, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3412, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_FINAL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3412, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3412, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3412, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3412, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3412, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3412, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3412, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_11 = __pyx_t_12;
        __pyx_L15_bool_binop_done:;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_12 = (__pyx_t_11 != 0);
        if (__pyx_t_12) {

          /* "pymola/generated/ModelicaParser.pyx":3413
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.T__25, ModelicaParser.EACH, ModelicaParser.FINAL, ModelicaParser.IDENT]:
 *                 self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *                 self.state = 474
 *                 self.element_modification_or_replaceable()
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3413, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_localctx, __pyx_int_1};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3413, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_localctx, __pyx_int_1};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3413, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3413, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_7) {
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_1);
            __Pyx_GIVEREF(__pyx_int_1);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_int_1);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3413, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3414
 *             if token in [ModelicaParser.T__25, ModelicaParser.EACH, ModelicaParser.FINAL, ModelicaParser.IDENT]:
 *                 self.enterOuterAlt(localctx, 1)
 *                 self.state = 474             # <<<<<<<<<<<<<<
 *                 self.element_modification_or_replaceable()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_474) < 0) __PYX_ERR(0, 3414, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":3415
 *                 self.enterOuterAlt(localctx, 1)
 *                 self.state = 474
 *                 self.element_modification_or_replaceable()             # <<<<<<<<<<<<<<
 *                 pass
 *             elif token in [ModelicaParser.REDECLARE]:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_element_modification_or_replacea); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3415, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (__pyx_t_3) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3415, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3415, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3412
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.T__25, ModelicaParser.EACH, ModelicaParser.FINAL, ModelicaParser.IDENT]:             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 1)
 *                 self.state = 474
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":3417
 *                 self.element_modification_or_replaceable()
 *                 pass
 *             elif token in [ModelicaParser.REDECLARE]:             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 2)
 *                 self.state = 475
 */
        __Pyx_INCREF(__pyx_v_token);
        __pyx_t_1 = __pyx_v_token;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3417, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_REDECLARE); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3417, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3417, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3417, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_11 = (__pyx_t_12 != 0);
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":3418
 *                 pass
 *             elif token in [ModelicaParser.REDECLARE]:
 *                 self.enterOuterAlt(localctx, 2)             # <<<<<<<<<<<<<<
 *                 self.state = 475
 *                 self.element_redeclaration()
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3418, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3418, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3418, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3418, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_2);
            __Pyx_GIVEREF(__pyx_int_2);
            PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_2);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3418, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3419
 *             elif token in [ModelicaParser.REDECLARE]:
 *                 self.enterOuterAlt(localctx, 2)
 *                 self.state = 475             # <<<<<<<<<<<<<<
 *                 self.element_redeclaration()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_475) < 0) __PYX_ERR(0, 3419, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":3420
 *                 self.enterOuterAlt(localctx, 2)
 *                 self.state = 475
 *                 self.element_redeclaration()             # <<<<<<<<<<<<<<
 *                 pass
 *             else:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_element_redeclaration); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3420, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3420, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3420, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3417
 *                 self.element_modification_or_replaceable()
 *                 pass
 *             elif token in [ModelicaParser.REDECLARE]:             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 2)
 *                 self.state = 475
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":3423
 *                 pass
 *             else:
 *                 raise NoViableAltException(self)             # <<<<<<<<<<<<<<
 * 
 *         except RecognitionException as re:
 */
        /*else*/ {
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_NoViableAltException); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3423, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3423, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3423, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3423, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3423, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3423, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_Raise(__pyx_t_1, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __PYX_ERR(0, 3423, __pyx_L6_error)
        }
        __pyx_L14:;

        /* "pymola/generated/ModelicaParser.pyx":3408
 *         localctx = ModelicaParser.ArgumentContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 60, self.RULE_argument)
 *         try:             # <<<<<<<<<<<<<<
 *             self.state = 476
 *             self._errHandler.sync(self)
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":3425
 *                 raise NoViableAltException(self)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3425, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.argument", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3) < 0) __PYX_ERR(0, 3425, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __pyx_v_re = __pyx_t_2;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":3426
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 3426, __pyx_L24_error)

          /* "pymola/generated/ModelicaParser.pyx":3427
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3427, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3427, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3427, __pyx_L24_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3427, __pyx_L24_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3427, __pyx_L24_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_4) {
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_4); __pyx_t_4 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_13, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3427, __pyx_L24_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3428
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3428, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3428, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3428, __pyx_L24_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3428, __pyx_L24_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3428, __pyx_L24_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_4, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3428, __pyx_L24_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":3425
 *                 raise NoViableAltException(self)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L25;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L24_error:;
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_6 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L8_except_error;
          }
          __pyx_L25:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":3408
 *         localctx = ModelicaParser.ArgumentContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 60, self.RULE_argument)
 *         try:             # <<<<<<<<<<<<<<
 *             self.state = 476
 *             self._errHandler.sync(self)
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":3430
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3430, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3430, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3430, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_14 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3430, __pyx_L31_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3430, __pyx_L31_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3430, __pyx_L31_error)
        }
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
      __pyx_L31_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":3431
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Element_modification_or_replaceableContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3404
 * 
 * 
 *     def argument(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.ArgumentContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.argument", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v_token);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3435
 *     class Element_modification_or_replaceableContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 3435, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 3435, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3435, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_modification_or_replaceableContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":3436
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 3436, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3436, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3436, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3436, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3436, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3437
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def element_modification(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 3437, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3435
 *     class Element_modification_or_replaceableContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_modification_or_replaceableContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3439
 *             self.parser = parser
 * 
 *         def element_modification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Element_modificationContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_3element_modification(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_3element_modification = {"element_modification", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_3element_modification, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_3element_modification(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("element_modification (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_2element_modification(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_2element_modification(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("element_modification", 0);

  /* "pymola/generated/ModelicaParser.pyx":3440
 * 
 *         def element_modification(self):
 *             return self.getTypedRuleContext(ModelicaParser.Element_modificationContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Element_modificationContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3440, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3440, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3440, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3440, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3439
 *             self.parser = parser
 * 
 *         def element_modification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Element_modificationContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_modification_or_replaceableContext.element_modification", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3443
 * 
 * 
 *         def element_replaceable(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Element_replaceableContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_5element_replaceable(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_5element_replaceable = {"element_replaceable", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_5element_replaceable, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_5element_replaceable(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("element_replaceable (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_4element_replaceable(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_4element_replaceable(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("element_replaceable", 0);

  /* "pymola/generated/ModelicaParser.pyx":3444
 * 
 *         def element_replaceable(self):
 *             return self.getTypedRuleContext(ModelicaParser.Element_replaceableContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3444, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3444, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Element_replaceableContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3444, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3444, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3444, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3444, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3444, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3443
 * 
 * 
 *         def element_replaceable(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Element_replaceableContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_modification_or_replaceableContext.element_replaceable", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3447
 * 
 * 
 *         def EACH(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.EACH, 0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_7EACH(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_7EACH = {"EACH", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_7EACH, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_7EACH(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("EACH (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_6EACH(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_6EACH(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("EACH", 0);

  /* "pymola/generated/ModelicaParser.pyx":3448
 * 
 *         def EACH(self):
 *             return self.getToken(ModelicaParser.EACH, 0)             # <<<<<<<<<<<<<<
 * 
 *         def FINAL(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getToken); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_EACH); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3448, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3448, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3448, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3448, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3447
 * 
 * 
 *         def EACH(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.EACH, 0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_modification_or_replaceableContext.EACH", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3450
 *             return self.getToken(ModelicaParser.EACH, 0)
 * 
 *         def FINAL(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.FINAL, 0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_9FINAL(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_9FINAL = {"FINAL", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_9FINAL, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_9FINAL(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("FINAL (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_8FINAL(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_8FINAL(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("FINAL", 0);

  /* "pymola/generated/ModelicaParser.pyx":3451
 * 
 *         def FINAL(self):
 *             return self.getToken(ModelicaParser.FINAL, 0)             # <<<<<<<<<<<<<<
 * 
 *         def getRuleIndex(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getToken); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_FINAL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3451, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3451, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3451, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3451, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3450
 *             return self.getToken(ModelicaParser.EACH, 0)
 * 
 *         def FINAL(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.FINAL, 0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_modification_or_replaceableContext.FINAL", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3453
 *             return self.getToken(ModelicaParser.FINAL, 0)
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_element_modification_or_replaceable
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_11getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_11getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_11getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_11getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_10getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_10getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":3454
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_element_modification_or_replaceable             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3454, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_element_modification_or_rep); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3454, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3453
 *             return self.getToken(ModelicaParser.FINAL, 0)
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_element_modification_or_replaceable
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_modification_or_replaceableContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3456
 *             return ModelicaParser.RULE_element_modification_or_replaceable
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterElement_modification_or_replaceable" ):
 *                 listener.enterElement_modification_or_replaceable(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_13enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_13enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_13enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_13enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 3456, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 3456, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3456, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_modification_or_replaceableContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_12enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_12enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":3457
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterElement_modification_or_replaceable" ):             # <<<<<<<<<<<<<<
 *                 listener.enterElement_modification_or_replaceable(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterElement_modification_or_rep); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3457, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3458
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterElement_modification_or_replaceable" ):
 *                 listener.enterElement_modification_or_replaceable(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterElement_modification_or_rep); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3458, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3458, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3458, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3458, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3458, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3458, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":3457
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterElement_modification_or_replaceable" ):             # <<<<<<<<<<<<<<
 *                 listener.enterElement_modification_or_replaceable(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3456
 *             return ModelicaParser.RULE_element_modification_or_replaceable
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterElement_modification_or_replaceable" ):
 *                 listener.enterElement_modification_or_replaceable(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_modification_or_replaceableContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3460
 *                 listener.enterElement_modification_or_replaceable(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitElement_modification_or_replaceable" ):
 *                 listener.exitElement_modification_or_replaceable(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_15exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_15exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_15exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_15exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 3460, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 3460, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3460, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_modification_or_replaceableContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_14exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_14exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":3461
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitElement_modification_or_replaceable" ):             # <<<<<<<<<<<<<<
 *                 listener.exitElement_modification_or_replaceable(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitElement_modification_or_repl); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3461, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3462
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitElement_modification_or_replaceable" ):
 *                 listener.exitElement_modification_or_replaceable(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitElement_modification_or_repl); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3462, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3462, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3462, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3462, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3462, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3462, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":3461
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitElement_modification_or_replaceable" ):             # <<<<<<<<<<<<<<
 *                 listener.exitElement_modification_or_replaceable(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3460
 *                 listener.enterElement_modification_or_replaceable(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitElement_modification_or_replaceable" ):
 *                 listener.exitElement_modification_or_replaceable(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_modification_or_replaceableContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3464
 *                 listener.exitElement_modification_or_replaceable(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitElement_modification_or_replaceable" ):
 *                 return visitor.visitElement_modification_or_replaceable(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_17accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_17accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_17accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_17accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 3464, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 3464, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3464, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_modification_or_replaceableContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_16accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_16accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":3465
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitElement_modification_or_replaceable" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitElement_modification_or_replaceable(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitElement_modification_or_rep); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3465, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3466
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitElement_modification_or_replaceable" ):
 *                 return visitor.visitElement_modification_or_replaceable(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitElement_modification_or_rep); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3466, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3466, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3466, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3466, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3466, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3466, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":3465
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitElement_modification_or_replaceable" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitElement_modification_or_replaceable(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3468
 *                 return visitor.visitElement_modification_or_replaceable(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3468, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3468, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3468, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3468, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3468, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3468, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":3464
 *                 listener.exitElement_modification_or_replaceable(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitElement_modification_or_replaceable" ):
 *                 return visitor.visitElement_modification_or_replaceable(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_modification_or_replaceableContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3473
 * 
 * 
 *     def element_modification_or_replaceable(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Element_modification_or_replaceableContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_65element_modification_or_replaceable(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_65element_modification_or_replaceable = {"element_modification_or_replaceable", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_65element_modification_or_replaceable, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_65element_modification_or_replaceable(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("element_modification_or_replaceable (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_64element_modification_or_replaceable(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_64element_modification_or_replaceable(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_token = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  char const *__pyx_t_22;
  __Pyx_RefNannySetupContext("element_modification_or_replaceable", 0);

  /* "pymola/generated/ModelicaParser.pyx":3475
 *     def element_modification_or_replaceable(self):
 * 
 *         localctx = ModelicaParser.Element_modification_or_replaceableContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 62, self.RULE_element_modification_or_replaceable)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Element_modification_or_replacea); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3475, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3475, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3475, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3475, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3476
 * 
 *         localctx = ModelicaParser.Element_modification_or_replaceableContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 62, self.RULE_element_modification_or_replaceable)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_element_modification_or_rep); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_62, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3476, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_62, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3476, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3476, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_62);
    __Pyx_GIVEREF(__pyx_int_62);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_62);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3476, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3477
 *         localctx = ModelicaParser.Element_modification_or_replaceableContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 62, self.RULE_element_modification_or_replaceable)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 3477, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3478
 *         self.enterRule(localctx, 62, self.RULE_element_modification_or_replaceable)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 479
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":3479
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 479
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3479, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3479, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3479, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3479, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3479, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3480
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 479             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_479) < 0) __PYX_ERR(0, 3480, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":3481
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 479
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.EACH:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3481, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3481, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3481, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3481, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3481, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3481, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3481, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3482
 *             self.state = 479
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.EACH:
 *                 self.state = 478
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3482, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3482, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__878, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3482, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v__la = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3483
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.EACH:             # <<<<<<<<<<<<<<
 *                 self.state = 478
 *                 self.match(ModelicaParser.EACH)
 */
        __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3483, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_EACH); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3483, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3483, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 3483, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":3484
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.EACH:
 *                 self.state = 478             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.EACH)
 * 
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_478) < 0) __PYX_ERR(0, 3484, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":3485
 *             if _la==ModelicaParser.EACH:
 *                 self.state = 478
 *                 self.match(ModelicaParser.EACH)             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3485, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3485, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_EACH); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3485, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3485, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3485, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3485, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3485, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3485, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3483
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.EACH:             # <<<<<<<<<<<<<<
 *                 self.state = 478
 *                 self.match(ModelicaParser.EACH)
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":3488
 * 
 * 
 *             self.state = 482             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_482) < 0) __PYX_ERR(0, 3488, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":3489
 * 
 *             self.state = 482
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.FINAL:
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3489, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_sync); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3489, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3489, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3489, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3489, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3489, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3489, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3490
 *             self.state = 482
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.FINAL:
 *                 self.state = 481
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3490, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3490, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__879, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3490, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3491
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.FINAL:             # <<<<<<<<<<<<<<
 *                 self.state = 481
 *                 self.match(ModelicaParser.FINAL)
 */
        __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3491, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_FINAL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3491, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3491, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 3491, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":3492
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.FINAL:
 *                 self.state = 481             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.FINAL)
 * 
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_481) < 0) __PYX_ERR(0, 3492, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":3493
 *             if _la==ModelicaParser.FINAL:
 *                 self.state = 481
 *                 self.match(ModelicaParser.FINAL)             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3493, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3493, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_FINAL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3493, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (!__pyx_t_3) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3493, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3493, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3493, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            {
              __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3493, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
              __Pyx_GIVEREF(__pyx_t_7);
              PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_7);
              __pyx_t_7 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3493, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3491
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.FINAL:             # <<<<<<<<<<<<<<
 *                 self.state = 481
 *                 self.match(ModelicaParser.FINAL)
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":3496
 * 
 * 
 *             self.state = 486             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_486) < 0) __PYX_ERR(0, 3496, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":3497
 * 
 *             self.state = 486
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.IDENT]:
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3497, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3497, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        if (!__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3497, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3497, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3497, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3497, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4); __pyx_t_4 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3497, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3498
 *             self.state = 486
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if token in [ModelicaParser.IDENT]:
 *                 self.state = 484
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3498, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3498, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__880, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3498, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_v_token = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3499
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.IDENT]:             # <<<<<<<<<<<<<<
 *                 self.state = 484
 *                 self.element_modification()
 */
        __Pyx_INCREF(__pyx_v_token);
        __pyx_t_1 = __pyx_v_token;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3499, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3499, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3499, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 3499, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_12 = (__pyx_t_11 != 0);
        if (__pyx_t_12) {

          /* "pymola/generated/ModelicaParser.pyx":3500
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.IDENT]:
 *                 self.state = 484             # <<<<<<<<<<<<<<
 *                 self.element_modification()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_484) < 0) __PYX_ERR(0, 3500, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":3501
 *             if token in [ModelicaParser.IDENT]:
 *                 self.state = 484
 *                 self.element_modification()             # <<<<<<<<<<<<<<
 *                 pass
 *             elif token in [ModelicaParser.T__25]:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_element_modification); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3501, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3501, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3501, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3499
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.IDENT]:             # <<<<<<<<<<<<<<
 *                 self.state = 484
 *                 self.element_modification()
 */
          goto __pyx_L16;
        }

        /* "pymola/generated/ModelicaParser.pyx":3503
 *                 self.element_modification()
 *                 pass
 *             elif token in [ModelicaParser.T__25]:             # <<<<<<<<<<<<<<
 *                 self.state = 485
 *                 self.element_replaceable()
 */
        __Pyx_INCREF(__pyx_v_token);
        __pyx_t_1 = __pyx_v_token;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3503, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__25); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3503, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3503, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3503, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_11 = (__pyx_t_12 != 0);
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":3504
 *                 pass
 *             elif token in [ModelicaParser.T__25]:
 *                 self.state = 485             # <<<<<<<<<<<<<<
 *                 self.element_replaceable()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_485) < 0) __PYX_ERR(0, 3504, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":3505
 *             elif token in [ModelicaParser.T__25]:
 *                 self.state = 485
 *                 self.element_replaceable()             # <<<<<<<<<<<<<<
 *                 pass
 *             else:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_element_replaceable); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3505, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3505, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3505, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3503
 *                 self.element_modification()
 *                 pass
 *             elif token in [ModelicaParser.T__25]:             # <<<<<<<<<<<<<<
 *                 self.state = 485
 *                 self.element_replaceable()
 */
          goto __pyx_L16;
        }

        /* "pymola/generated/ModelicaParser.pyx":3508
 *                 pass
 *             else:
 *                 raise NoViableAltException(self)             # <<<<<<<<<<<<<<
 * 
 *         except RecognitionException as re:
 */
        /*else*/ {
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_NoViableAltException); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3508, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3508, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3508, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3508, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3508, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3508, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_Raise(__pyx_t_1, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __PYX_ERR(0, 3508, __pyx_L6_error)
        }
        __pyx_L16:;

        /* "pymola/generated/ModelicaParser.pyx":3478
 *         self.enterRule(localctx, 62, self.RULE_element_modification_or_replaceable)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 479
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":3510
 *                 raise NoViableAltException(self)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3510, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.element_modification_or_replaceable", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_2, &__pyx_t_4) < 0) __PYX_ERR(0, 3510, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_2);
        __pyx_v_re = __pyx_t_2;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":3511
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 3511, __pyx_L22_error)

          /* "pymola/generated/ModelicaParser.pyx":3512
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3512, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3512, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3512, __pyx_L22_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3512, __pyx_L22_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3512, __pyx_L22_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_13, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3512, __pyx_L22_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3513
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3513, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3513, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3513, __pyx_L22_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3513, __pyx_L22_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3513, __pyx_L22_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_3, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3513, __pyx_L22_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":3510
 *                 raise NoViableAltException(self)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L23;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L22_error:;
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_6 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L8_except_error;
          }
          __pyx_L23:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":3478
 *         self.enterRule(localctx, 62, self.RULE_element_modification_or_replaceable)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 479
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":3515
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3515, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3515, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3515, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_14 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3515, __pyx_L29_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3515, __pyx_L29_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3515, __pyx_L29_error)
        }
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
      __pyx_L29_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":3516
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Element_modificationContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3473
 * 
 * 
 *     def element_modification_or_replaceable(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Element_modification_or_replaceableContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.element_modification_or_replaceable", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_token);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3520
 *     class Element_modificationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_27Element_modificationContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_27Element_modificationContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_27Element_modificationContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_27Element_modificationContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 3520, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 3520, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3520, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_modificationContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_27Element_modificationContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_27Element_modificationContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":3521
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 3521, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3521, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3521, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3521, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3521, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3522
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def component_reference(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 3522, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3520
 *     class Element_modificationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_modificationContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3524
 *             self.parser = parser
 * 
 *         def component_reference(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_27Element_modificationContext_3component_reference(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_27Element_modificationContext_3component_reference = {"component_reference", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_27Element_modificationContext_3component_reference, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_27Element_modificationContext_3component_reference(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("component_reference (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_27Element_modificationContext_2component_reference(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_27Element_modificationContext_2component_reference(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("component_reference", 0);

  /* "pymola/generated/ModelicaParser.pyx":3525
 * 
 *         def component_reference(self):
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Component_referenceContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3525, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3525, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3525, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3525, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3524
 *             self.parser = parser
 * 
 *         def component_reference(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_modificationContext.component_reference", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3528
 * 
 * 
 *         def string_comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.String_commentContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_27Element_modificationContext_5string_comment(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_27Element_modificationContext_5string_comment = {"string_comment", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_27Element_modificationContext_5string_comment, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_27Element_modificationContext_5string_comment(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("string_comment (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_27Element_modificationContext_4string_comment(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_27Element_modificationContext_4string_comment(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("string_comment", 0);

  /* "pymola/generated/ModelicaParser.pyx":3529
 * 
 *         def string_comment(self):
 *             return self.getTypedRuleContext(ModelicaParser.String_commentContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_String_commentContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3529, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3529, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3529, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3529, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3528
 * 
 * 
 *         def string_comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.String_commentContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_modificationContext.string_comment", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3532
 * 
 * 
 *         def modification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ModificationContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_27Element_modificationContext_7modification(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_27Element_modificationContext_7modification = {"modification", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_27Element_modificationContext_7modification, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_27Element_modificationContext_7modification(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("modification (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_27Element_modificationContext_6modification(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_27Element_modificationContext_6modification(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("modification", 0);

  /* "pymola/generated/ModelicaParser.pyx":3533
 * 
 *         def modification(self):
 *             return self.getTypedRuleContext(ModelicaParser.ModificationContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_ModificationContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3533, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3533, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3533, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3533, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3532
 * 
 * 
 *         def modification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ModificationContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_modificationContext.modification", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3536
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_element_modification
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_27Element_modificationContext_9getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_27Element_modificationContext_9getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_27Element_modificationContext_9getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_27Element_modificationContext_9getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_27Element_modificationContext_8getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_27Element_modificationContext_8getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":3537
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_element_modification             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_element_modification); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3536
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_element_modification
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_modificationContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3539
 *             return ModelicaParser.RULE_element_modification
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterElement_modification" ):
 *                 listener.enterElement_modification(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_27Element_modificationContext_11enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_27Element_modificationContext_11enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_27Element_modificationContext_11enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_27Element_modificationContext_11enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 3539, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 3539, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3539, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_modificationContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_27Element_modificationContext_10enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_27Element_modificationContext_10enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":3540
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterElement_modification" ):             # <<<<<<<<<<<<<<
 *                 listener.enterElement_modification(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterElement_modification); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3540, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3541
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterElement_modification" ):
 *                 listener.enterElement_modification(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterElement_modification); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3541, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3541, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3541, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3541, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3541, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":3540
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterElement_modification" ):             # <<<<<<<<<<<<<<
 *                 listener.enterElement_modification(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3539
 *             return ModelicaParser.RULE_element_modification
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterElement_modification" ):
 *                 listener.enterElement_modification(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_modificationContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3543
 *                 listener.enterElement_modification(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitElement_modification" ):
 *                 listener.exitElement_modification(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_27Element_modificationContext_13exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_27Element_modificationContext_13exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_27Element_modificationContext_13exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_27Element_modificationContext_13exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 3543, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 3543, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3543, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_modificationContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_27Element_modificationContext_12exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_27Element_modificationContext_12exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":3544
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitElement_modification" ):             # <<<<<<<<<<<<<<
 *                 listener.exitElement_modification(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitElement_modification); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3544, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3545
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitElement_modification" ):
 *                 listener.exitElement_modification(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitElement_modification); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3545, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3545, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3545, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3545, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3545, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3545, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":3544
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitElement_modification" ):             # <<<<<<<<<<<<<<
 *                 listener.exitElement_modification(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3543
 *                 listener.enterElement_modification(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitElement_modification" ):
 *                 listener.exitElement_modification(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_modificationContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3547
 *                 listener.exitElement_modification(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitElement_modification" ):
 *                 return visitor.visitElement_modification(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_27Element_modificationContext_15accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_27Element_modificationContext_15accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_27Element_modificationContext_15accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_27Element_modificationContext_15accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 3547, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 3547, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3547, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_modificationContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_27Element_modificationContext_14accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_27Element_modificationContext_14accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":3548
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitElement_modification" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitElement_modification(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitElement_modification); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3548, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3549
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitElement_modification" ):
 *                 return visitor.visitElement_modification(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitElement_modification); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3549, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3549, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3549, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3549, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3549, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3549, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":3548
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitElement_modification" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitElement_modification(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3551
 *                 return visitor.visitElement_modification(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3551, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3551, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3551, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3551, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3551, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3551, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":3547
 *                 listener.exitElement_modification(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitElement_modification" ):
 *                 return visitor.visitElement_modification(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_modificationContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3556
 * 
 * 
 *     def element_modification(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Element_modificationContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_67element_modification(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_67element_modification = {"element_modification", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_67element_modification, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_67element_modification(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("element_modification (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_66element_modification(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_66element_modification(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  char const *__pyx_t_22;
  __Pyx_RefNannySetupContext("element_modification", 0);

  /* "pymola/generated/ModelicaParser.pyx":3558
 *     def element_modification(self):
 * 
 *         localctx = ModelicaParser.Element_modificationContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 64, self.RULE_element_modification)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3558, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Element_modificationContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3558, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3558, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3558, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3558, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3558, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3558, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3558, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3559
 * 
 *         localctx = ModelicaParser.Element_modificationContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 64, self.RULE_element_modification)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3559, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_element_modification); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3559, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_64, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3559, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_64, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3559, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3559, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_64);
    __Pyx_GIVEREF(__pyx_int_64);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_64);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3559, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3560
 *         localctx = ModelicaParser.Element_modificationContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 64, self.RULE_element_modification)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 3560, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3561
 *         self.enterRule(localctx, 64, self.RULE_element_modification)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 488
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":3562
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 488
 *             self.component_reference()
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3562, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3562, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3562, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3562, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3562, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3563
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 488             # <<<<<<<<<<<<<<
 *             self.component_reference()
 *             self.state = 490
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_488) < 0) __PYX_ERR(0, 3563, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":3564
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 488
 *             self.component_reference()             # <<<<<<<<<<<<<<
 *             self.state = 490
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_component_reference); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3564, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3564, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3564, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3565
 *             self.state = 488
 *             self.component_reference()
 *             self.state = 490             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_490) < 0) __PYX_ERR(0, 3565, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":3566
 *             self.component_reference()
 *             self.state = 490
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__14) | (1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__40))) != 0):
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3566, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3566, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3566, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3566, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3566, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3566, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3566, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3567
 *             self.state = 490
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__14) | (1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__40))) != 0):
 *                 self.state = 489
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3567, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3567, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__881, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3567, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v__la = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3568
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__14) | (1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__40))) != 0):             # <<<<<<<<<<<<<<
 *                 self.state = 489
 *                 self.modification()
 */
        __pyx_t_1 = __Pyx_PyInt_From_long((~0x3f)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3568, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = PyNumber_And(__pyx_v__la, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3568, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_t_7, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3568, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3568, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_1 = PyNumber_Lshift(__pyx_int_1, __pyx_v__la); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3568, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3568, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__14); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3568, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3568, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3568, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__16); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3568, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3568, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_Or(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3568, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3568, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__40); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3568, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3568, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyNumber_Or(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3568, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_And(__pyx_t_1, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3568, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyObject_RichCompare(__pyx_t_2, __pyx_int_0, Py_NE); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3568, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3568, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_11 = __pyx_t_12;
        __pyx_L15_bool_binop_done:;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":3569
 *             _la = self._input.LA(1)
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__14) | (1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__40))) != 0):
 *                 self.state = 489             # <<<<<<<<<<<<<<
 *                 self.modification()
 * 
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_489) < 0) __PYX_ERR(0, 3569, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":3570
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__14) | (1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__40))) != 0):
 *                 self.state = 489
 *                 self.modification()             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_modification); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3570, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_1 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (__pyx_t_1) {
            __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3570, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          } else {
            __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3570, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3568
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__14) | (1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__40))) != 0):             # <<<<<<<<<<<<<<
 *                 self.state = 489
 *                 self.modification()
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":3573
 * 
 * 
 *             self.state = 492             # <<<<<<<<<<<<<<
 *             self.string_comment()
 *         except RecognitionException as re:
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_492) < 0) __PYX_ERR(0, 3573, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":3574
 * 
 *             self.state = 492
 *             self.string_comment()             # <<<<<<<<<<<<<<
 *         except RecognitionException as re:
 *             localctx.exception = re
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_string_comment); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3574, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3574, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3574, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3561
 *         self.enterRule(localctx, 64, self.RULE_element_modification)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 488
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "pymola/generated/ModelicaParser.pyx":3575
 *             self.state = 492
 *             self.string_comment()
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3575, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.element_modification", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_2, &__pyx_t_1) < 0) __PYX_ERR(0, 3575, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_2);
        __pyx_v_re = __pyx_t_2;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":3576
 *             self.string_comment()
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 3576, __pyx_L22_error)

          /* "pymola/generated/ModelicaParser.pyx":3577
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3577, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3577, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3577, __pyx_L22_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3577, __pyx_L22_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3577, __pyx_L22_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_4) {
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_4); __pyx_t_4 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_13, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3577, __pyx_L22_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3578
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3578, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3578, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3578, __pyx_L22_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3578, __pyx_L22_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3578, __pyx_L22_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3578, __pyx_L22_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":3575
 *             self.state = 492
 *             self.string_comment()
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L23;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L22_error:;
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_6 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L8_except_error;
          }
          __pyx_L23:;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":3561
 *         self.enterRule(localctx, 64, self.RULE_element_modification)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 488
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":3580
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3580, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      if (__pyx_t_7) {
        __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3580, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      } else {
        __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3580, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_14 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3580, __pyx_L29_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        if (__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3580, __pyx_L29_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3580, __pyx_L29_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
      __pyx_L29_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":3581
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Element_redeclarationContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3556
 * 
 * 
 *     def element_modification(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Element_modificationContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.element_modification", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3585
 *     class Element_redeclarationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 3585, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 3585, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3585, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_redeclarationContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":3586
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 3586, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3586, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3586, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3586, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3586, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3587
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def REDECLARE(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 3587, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3585
 *     class Element_redeclarationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_redeclarationContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3589
 *             self.parser = parser
 * 
 *         def REDECLARE(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.REDECLARE, 0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_3REDECLARE(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_3REDECLARE = {"REDECLARE", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_3REDECLARE, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_3REDECLARE(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("REDECLARE (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_2REDECLARE(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_2REDECLARE(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("REDECLARE", 0);

  /* "pymola/generated/ModelicaParser.pyx":3590
 * 
 *         def REDECLARE(self):
 *             return self.getToken(ModelicaParser.REDECLARE, 0)             # <<<<<<<<<<<<<<
 * 
 *         def element_replaceable(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getToken); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_REDECLARE); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3590, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3590, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3590, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3590, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3589
 *             self.parser = parser
 * 
 *         def REDECLARE(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.REDECLARE, 0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_redeclarationContext.REDECLARE", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3592
 *             return self.getToken(ModelicaParser.REDECLARE, 0)
 * 
 *         def element_replaceable(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Element_replaceableContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_5element_replaceable(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_5element_replaceable = {"element_replaceable", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_5element_replaceable, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_5element_replaceable(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("element_replaceable (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_4element_replaceable(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_4element_replaceable(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("element_replaceable", 0);

  /* "pymola/generated/ModelicaParser.pyx":3593
 * 
 *         def element_replaceable(self):
 *             return self.getTypedRuleContext(ModelicaParser.Element_replaceableContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Element_replaceableContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3593, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3593, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3593, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3593, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3592
 *             return self.getToken(ModelicaParser.REDECLARE, 0)
 * 
 *         def element_replaceable(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Element_replaceableContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_redeclarationContext.element_replaceable", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3596
 * 
 * 
 *         def EACH(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.EACH, 0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_7EACH(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_7EACH = {"EACH", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_7EACH, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_7EACH(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("EACH (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_6EACH(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_6EACH(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("EACH", 0);

  /* "pymola/generated/ModelicaParser.pyx":3597
 * 
 *         def EACH(self):
 *             return self.getToken(ModelicaParser.EACH, 0)             # <<<<<<<<<<<<<<
 * 
 *         def FINAL(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getToken); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3597, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3597, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_EACH); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3597, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3597, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3597, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3597, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3597, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3596
 * 
 * 
 *         def EACH(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.EACH, 0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_redeclarationContext.EACH", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3599
 *             return self.getToken(ModelicaParser.EACH, 0)
 * 
 *         def FINAL(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.FINAL, 0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_9FINAL(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_9FINAL = {"FINAL", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_9FINAL, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_9FINAL(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("FINAL (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_8FINAL(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_8FINAL(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("FINAL", 0);

  /* "pymola/generated/ModelicaParser.pyx":3600
 * 
 *         def FINAL(self):
 *             return self.getToken(ModelicaParser.FINAL, 0)             # <<<<<<<<<<<<<<
 * 
 *         def short_class_definition(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getToken); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_FINAL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3600, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3600, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3600, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3600, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3599
 *             return self.getToken(ModelicaParser.EACH, 0)
 * 
 *         def FINAL(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.FINAL, 0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_redeclarationContext.FINAL", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3602
 *             return self.getToken(ModelicaParser.FINAL, 0)
 * 
 *         def short_class_definition(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Short_class_definitionContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_11short_class_definition(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_11short_class_definition = {"short_class_definition", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_11short_class_definition, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_11short_class_definition(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("short_class_definition (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_10short_class_definition(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_10short_class_definition(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("short_class_definition", 0);

  /* "pymola/generated/ModelicaParser.pyx":3603
 * 
 *         def short_class_definition(self):
 *             return self.getTypedRuleContext(ModelicaParser.Short_class_definitionContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3603, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3603, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Short_class_definitionContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3603, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3603, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3603, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3603, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3603, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3602
 *             return self.getToken(ModelicaParser.FINAL, 0)
 * 
 *         def short_class_definition(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Short_class_definitionContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_redeclarationContext.short_class_definition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3606
 * 
 * 
 *         def component_clause1(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_clause1Context,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_13component_clause1(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_13component_clause1 = {"component_clause1", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_13component_clause1, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_13component_clause1(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("component_clause1 (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_12component_clause1(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_12component_clause1(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("component_clause1", 0);

  /* "pymola/generated/ModelicaParser.pyx":3607
 * 
 *         def component_clause1(self):
 *             return self.getTypedRuleContext(ModelicaParser.Component_clause1Context,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3607, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3607, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Component_clause1Context); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3607, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3607, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3607, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3607, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3607, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3606
 * 
 * 
 *         def component_clause1(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_clause1Context,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_redeclarationContext.component_clause1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3610
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_element_redeclaration
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_15getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_15getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_15getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_15getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_14getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_14getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":3611
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_element_redeclaration             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3611, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_element_redeclaration); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3611, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3610
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_element_redeclaration
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_redeclarationContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3613
 *             return ModelicaParser.RULE_element_redeclaration
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterElement_redeclaration" ):
 *                 listener.enterElement_redeclaration(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_17enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_17enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_17enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_17enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 3613, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 3613, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3613, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_redeclarationContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_16enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_16enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":3614
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterElement_redeclaration" ):             # <<<<<<<<<<<<<<
 *                 listener.enterElement_redeclaration(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterElement_redeclaration); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3614, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3615
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterElement_redeclaration" ):
 *                 listener.enterElement_redeclaration(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterElement_redeclaration); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3615, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3615, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3615, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3615, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3615, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3615, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":3614
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterElement_redeclaration" ):             # <<<<<<<<<<<<<<
 *                 listener.enterElement_redeclaration(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3613
 *             return ModelicaParser.RULE_element_redeclaration
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterElement_redeclaration" ):
 *                 listener.enterElement_redeclaration(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_redeclarationContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3617
 *                 listener.enterElement_redeclaration(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitElement_redeclaration" ):
 *                 listener.exitElement_redeclaration(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_19exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_19exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_19exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_19exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 3617, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 3617, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3617, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_redeclarationContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_18exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_18exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":3618
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitElement_redeclaration" ):             # <<<<<<<<<<<<<<
 *                 listener.exitElement_redeclaration(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitElement_redeclaration); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3618, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3619
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitElement_redeclaration" ):
 *                 listener.exitElement_redeclaration(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitElement_redeclaration); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3619, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3619, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3619, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3619, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3619, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3619, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":3618
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitElement_redeclaration" ):             # <<<<<<<<<<<<<<
 *                 listener.exitElement_redeclaration(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3617
 *                 listener.enterElement_redeclaration(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitElement_redeclaration" ):
 *                 listener.exitElement_redeclaration(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_redeclarationContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3621
 *                 listener.exitElement_redeclaration(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitElement_redeclaration" ):
 *                 return visitor.visitElement_redeclaration(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_21accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_21accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_21accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_21accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 3621, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 3621, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3621, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_redeclarationContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_20accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_20accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":3622
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitElement_redeclaration" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitElement_redeclaration(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitElement_redeclaration); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3622, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3623
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitElement_redeclaration" ):
 *                 return visitor.visitElement_redeclaration(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitElement_redeclaration); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3623, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3623, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3623, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3623, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3623, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3623, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":3622
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitElement_redeclaration" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitElement_redeclaration(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3625
 *                 return visitor.visitElement_redeclaration(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3625, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3625, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3625, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3625, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3625, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3625, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":3621
 *                 listener.exitElement_redeclaration(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitElement_redeclaration" ):
 *                 return visitor.visitElement_redeclaration(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_redeclarationContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3630
 * 
 * 
 *     def element_redeclaration(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Element_redeclarationContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_69element_redeclaration(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_69element_redeclaration = {"element_redeclaration", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_69element_redeclaration, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_69element_redeclaration(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("element_redeclaration (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_68element_redeclaration(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_68element_redeclaration(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_token = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  char const *__pyx_t_22;
  __Pyx_RefNannySetupContext("element_redeclaration", 0);

  /* "pymola/generated/ModelicaParser.pyx":3632
 *     def element_redeclaration(self):
 * 
 *         localctx = ModelicaParser.Element_redeclarationContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 66, self.RULE_element_redeclaration)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Element_redeclarationContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3632, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3632, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3632, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3632, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3633
 * 
 *         localctx = ModelicaParser.Element_redeclarationContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 66, self.RULE_element_redeclaration)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_element_redeclaration); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_66, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3633, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_66, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3633, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3633, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_66);
    __Pyx_GIVEREF(__pyx_int_66);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_66);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3633, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3634
 *         localctx = ModelicaParser.Element_redeclarationContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 66, self.RULE_element_redeclaration)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 3634, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3635
 *         self.enterRule(localctx, 66, self.RULE_element_redeclaration)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 494
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":3636
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 494
 *             self.match(ModelicaParser.REDECLARE)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3636, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3636, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3636, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3636, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3636, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3637
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 494             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.REDECLARE)
 *             self.state = 496
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_494) < 0) __PYX_ERR(0, 3637, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":3638
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 494
 *             self.match(ModelicaParser.REDECLARE)             # <<<<<<<<<<<<<<
 *             self.state = 496
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3638, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3638, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_REDECLARE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3638, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3638, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3638, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3638, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3638, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3638, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3639
 *             self.state = 494
 *             self.match(ModelicaParser.REDECLARE)
 *             self.state = 496             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_496) < 0) __PYX_ERR(0, 3639, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":3640
 *             self.match(ModelicaParser.REDECLARE)
 *             self.state = 496
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.EACH:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3640, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3640, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3640, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3640, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3640, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3640, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3640, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3641
 *             self.state = 496
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.EACH:
 *                 self.state = 495
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3641, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3641, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__882, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3641, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_v__la = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3642
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.EACH:             # <<<<<<<<<<<<<<
 *                 self.state = 495
 *                 self.match(ModelicaParser.EACH)
 */
        __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3642, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_EACH); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3642, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3642, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 3642, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":3643
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.EACH:
 *                 self.state = 495             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.EACH)
 * 
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_495) < 0) __PYX_ERR(0, 3643, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":3644
 *             if _la==ModelicaParser.EACH:
 *                 self.state = 495
 *                 self.match(ModelicaParser.EACH)             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3644, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3644, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_EACH); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3644, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3644, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3644, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3644, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3644, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3644, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3642
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.EACH:             # <<<<<<<<<<<<<<
 *                 self.state = 495
 *                 self.match(ModelicaParser.EACH)
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":3647
 * 
 * 
 *             self.state = 499             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_499) < 0) __PYX_ERR(0, 3647, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":3648
 * 
 *             self.state = 499
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.FINAL:
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3648, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3648, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3648, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3648, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3648, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3648, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __pyx_t_4 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3648, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3649
 *             self.state = 499
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.FINAL:
 *                 self.state = 498
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3649, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3649, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__883, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3649, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3650
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.FINAL:             # <<<<<<<<<<<<<<
 *                 self.state = 498
 *                 self.match(ModelicaParser.FINAL)
 */
        __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3650, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_FINAL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3650, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3650, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 3650, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":3651
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.FINAL:
 *                 self.state = 498             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.FINAL)
 * 
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_498) < 0) __PYX_ERR(0, 3651, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":3652
 *             if _la==ModelicaParser.FINAL:
 *                 self.state = 498
 *                 self.match(ModelicaParser.FINAL)             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3652, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3652, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_FINAL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3652, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_3) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3652, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_4};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3652, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_4};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3652, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            } else
            #endif
            {
              __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3652, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
              __Pyx_GIVEREF(__pyx_t_4);
              PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_4);
              __pyx_t_4 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3652, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3650
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.FINAL:             # <<<<<<<<<<<<<<
 *                 self.state = 498
 *                 self.match(ModelicaParser.FINAL)
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":3655
 * 
 * 
 *             self.state = 506             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_506) < 0) __PYX_ERR(0, 3655, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":3656
 * 
 *             self.state = 506
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.T__1, ModelicaParser.T__2, ModelicaParser.T__3, ModelicaParser.T__4, ModelicaParser.T__5, ModelicaParser.T__6, ModelicaParser.T__7, ModelicaParser.T__8, ModelicaParser.T__9, ModelicaParser.T__10, ModelicaParser.T__11, ModelicaParser.T__12, ModelicaParser.T__32, ModelicaParser.T__33, ModelicaParser.T__34, ModelicaParser.T__35, ModelicaParser.T__36, ModelicaParser.T__37, ModelicaParser.T__38, ModelicaParser.PARTIAL, ModelicaParser.IDENT]:
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3656, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3656, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3656, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3656, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3656, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3656, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3656, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3657
 *             self.state = 506
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if token in [ModelicaParser.T__1, ModelicaParser.T__2, ModelicaParser.T__3, ModelicaParser.T__4, ModelicaParser.T__5, ModelicaParser.T__6, ModelicaParser.T__7, ModelicaParser.T__8, ModelicaParser.T__9, ModelicaParser.T__10, ModelicaParser.T__11, ModelicaParser.T__12, ModelicaParser.T__32, ModelicaParser.T__33, ModelicaParser.T__34, ModelicaParser.T__35, ModelicaParser.T__36, ModelicaParser.T__37, ModelicaParser.T__38, ModelicaParser.PARTIAL, ModelicaParser.IDENT]:
 *                 self.state = 503
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3657, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3657, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__884, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3657, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_v_token = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3658
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.T__1, ModelicaParser.T__2, ModelicaParser.T__3, ModelicaParser.T__4, ModelicaParser.T__5, ModelicaParser.T__6, ModelicaParser.T__7, ModelicaParser.T__8, ModelicaParser.T__9, ModelicaParser.T__10, ModelicaParser.T__11, ModelicaParser.T__12, ModelicaParser.T__32, ModelicaParser.T__33, ModelicaParser.T__34, ModelicaParser.T__35, ModelicaParser.T__36, ModelicaParser.T__37, ModelicaParser.T__38, ModelicaParser.PARTIAL, ModelicaParser.IDENT]:             # <<<<<<<<<<<<<<
 *                 self.state = 503
 *                 self._errHandler.sync(self)
 */
        __Pyx_INCREF(__pyx_v_token);
        __pyx_t_1 = __pyx_v_token;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L17_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L17_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L17_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L17_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L17_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L17_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L17_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L17_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L17_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__10); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L17_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__11); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L17_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__12); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L17_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__32); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L17_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__33); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L17_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__34); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L17_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__35); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L17_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__36); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L17_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__37); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L17_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__38); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L17_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_PARTIAL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L17_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_11 = __pyx_t_12;
        __pyx_L17_bool_binop_done:;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_12 = (__pyx_t_11 != 0);
        if (__pyx_t_12) {

          /* "pymola/generated/ModelicaParser.pyx":3659
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.T__1, ModelicaParser.T__2, ModelicaParser.T__3, ModelicaParser.T__4, ModelicaParser.T__5, ModelicaParser.T__6, ModelicaParser.T__7, ModelicaParser.T__8, ModelicaParser.T__9, ModelicaParser.T__10, ModelicaParser.T__11, ModelicaParser.T__12, ModelicaParser.T__32, ModelicaParser.T__33, ModelicaParser.T__34, ModelicaParser.T__35, ModelicaParser.T__36, ModelicaParser.T__37, ModelicaParser.T__38, ModelicaParser.PARTIAL, ModelicaParser.IDENT]:
 *                 self.state = 503             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 token = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_503) < 0) __PYX_ERR(0, 3659, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":3660
 *             if token in [ModelicaParser.T__1, ModelicaParser.T__2, ModelicaParser.T__3, ModelicaParser.T__4, ModelicaParser.T__5, ModelicaParser.T__6, ModelicaParser.T__7, ModelicaParser.T__8, ModelicaParser.T__9, ModelicaParser.T__10, ModelicaParser.T__11, ModelicaParser.T__12, ModelicaParser.T__32, ModelicaParser.T__33, ModelicaParser.T__34, ModelicaParser.T__35, ModelicaParser.T__36, ModelicaParser.T__37, ModelicaParser.T__38, ModelicaParser.PARTIAL, ModelicaParser.IDENT]:
 *                 self.state = 503
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 token = self._input.LA(1)
 *                 if token in [ModelicaParser.T__1, ModelicaParser.T__2, ModelicaParser.T__3, ModelicaParser.T__4, ModelicaParser.T__5, ModelicaParser.T__6, ModelicaParser.T__7, ModelicaParser.T__8, ModelicaParser.T__9, ModelicaParser.T__10, ModelicaParser.T__11, ModelicaParser.T__12, ModelicaParser.PARTIAL]:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3660, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_sync); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3660, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3660, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3660, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3660, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3660, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3660, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3661
 *                 self.state = 503
 *                 self._errHandler.sync(self)
 *                 token = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if token in [ModelicaParser.T__1, ModelicaParser.T__2, ModelicaParser.T__3, ModelicaParser.T__4, ModelicaParser.T__5, ModelicaParser.T__6, ModelicaParser.T__7, ModelicaParser.T__8, ModelicaParser.T__9, ModelicaParser.T__10, ModelicaParser.T__11, ModelicaParser.T__12, ModelicaParser.PARTIAL]:
 *                     self.state = 501
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3661, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3661, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__885, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3661, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF_SET(__pyx_v_token, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3662
 *                 self._errHandler.sync(self)
 *                 token = self._input.LA(1)
 *                 if token in [ModelicaParser.T__1, ModelicaParser.T__2, ModelicaParser.T__3, ModelicaParser.T__4, ModelicaParser.T__5, ModelicaParser.T__6, ModelicaParser.T__7, ModelicaParser.T__8, ModelicaParser.T__9, ModelicaParser.T__10, ModelicaParser.T__11, ModelicaParser.T__12, ModelicaParser.PARTIAL]:             # <<<<<<<<<<<<<<
 *                     self.state = 501
 *                     self.short_class_definition()
 */
          __Pyx_INCREF(__pyx_v_token);
          __pyx_t_1 = __pyx_v_token;
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (!__pyx_t_11) {
          } else {
            __pyx_t_12 = __pyx_t_11;
            goto __pyx_L39_bool_binop_done;
          }
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (!__pyx_t_11) {
          } else {
            __pyx_t_12 = __pyx_t_11;
            goto __pyx_L39_bool_binop_done;
          }
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (!__pyx_t_11) {
          } else {
            __pyx_t_12 = __pyx_t_11;
            goto __pyx_L39_bool_binop_done;
          }
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (!__pyx_t_11) {
          } else {
            __pyx_t_12 = __pyx_t_11;
            goto __pyx_L39_bool_binop_done;
          }
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (!__pyx_t_11) {
          } else {
            __pyx_t_12 = __pyx_t_11;
            goto __pyx_L39_bool_binop_done;
          }
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (!__pyx_t_11) {
          } else {
            __pyx_t_12 = __pyx_t_11;
            goto __pyx_L39_bool_binop_done;
          }
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (!__pyx_t_11) {
          } else {
            __pyx_t_12 = __pyx_t_11;
            goto __pyx_L39_bool_binop_done;
          }
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (!__pyx_t_11) {
          } else {
            __pyx_t_12 = __pyx_t_11;
            goto __pyx_L39_bool_binop_done;
          }
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (!__pyx_t_11) {
          } else {
            __pyx_t_12 = __pyx_t_11;
            goto __pyx_L39_bool_binop_done;
          }
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__10); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (!__pyx_t_11) {
          } else {
            __pyx_t_12 = __pyx_t_11;
            goto __pyx_L39_bool_binop_done;
          }
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__11); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (!__pyx_t_11) {
          } else {
            __pyx_t_12 = __pyx_t_11;
            goto __pyx_L39_bool_binop_done;
          }
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__12); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (!__pyx_t_11) {
          } else {
            __pyx_t_12 = __pyx_t_11;
            goto __pyx_L39_bool_binop_done;
          }
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_PARTIAL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 3662, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_12 = __pyx_t_11;
          __pyx_L39_bool_binop_done:;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_11 = (__pyx_t_12 != 0);
          if (__pyx_t_11) {

            /* "pymola/generated/ModelicaParser.pyx":3663
 *                 token = self._input.LA(1)
 *                 if token in [ModelicaParser.T__1, ModelicaParser.T__2, ModelicaParser.T__3, ModelicaParser.T__4, ModelicaParser.T__5, ModelicaParser.T__6, ModelicaParser.T__7, ModelicaParser.T__8, ModelicaParser.T__9, ModelicaParser.T__10, ModelicaParser.T__11, ModelicaParser.T__12, ModelicaParser.PARTIAL]:
 *                     self.state = 501             # <<<<<<<<<<<<<<
 *                     self.short_class_definition()
 *                     pass
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_501) < 0) __PYX_ERR(0, 3663, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":3664
 *                 if token in [ModelicaParser.T__1, ModelicaParser.T__2, ModelicaParser.T__3, ModelicaParser.T__4, ModelicaParser.T__5, ModelicaParser.T__6, ModelicaParser.T__7, ModelicaParser.T__8, ModelicaParser.T__9, ModelicaParser.T__10, ModelicaParser.T__11, ModelicaParser.T__12, ModelicaParser.PARTIAL]:
 *                     self.state = 501
 *                     self.short_class_definition()             # <<<<<<<<<<<<<<
 *                     pass
 *                 elif token in [ModelicaParser.T__32, ModelicaParser.T__33, ModelicaParser.T__34, ModelicaParser.T__35, ModelicaParser.T__36, ModelicaParser.T__37, ModelicaParser.T__38, ModelicaParser.IDENT]:
 */
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_short_class_definition); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3664, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_7 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
              __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
              if (likely(__pyx_t_7)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                __Pyx_INCREF(__pyx_t_7);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_4, function);
              }
            }
            if (__pyx_t_7) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3664, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else {
              __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3664, __pyx_L6_error)
            }
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":3662
 *                 self._errHandler.sync(self)
 *                 token = self._input.LA(1)
 *                 if token in [ModelicaParser.T__1, ModelicaParser.T__2, ModelicaParser.T__3, ModelicaParser.T__4, ModelicaParser.T__5, ModelicaParser.T__6, ModelicaParser.T__7, ModelicaParser.T__8, ModelicaParser.T__9, ModelicaParser.T__10, ModelicaParser.T__11, ModelicaParser.T__12, ModelicaParser.PARTIAL]:             # <<<<<<<<<<<<<<
 *                     self.state = 501
 *                     self.short_class_definition()
 */
            goto __pyx_L38;
          }

          /* "pymola/generated/ModelicaParser.pyx":3666
 *                     self.short_class_definition()
 *                     pass
 *                 elif token in [ModelicaParser.T__32, ModelicaParser.T__33, ModelicaParser.T__34, ModelicaParser.T__35, ModelicaParser.T__36, ModelicaParser.T__37, ModelicaParser.T__38, ModelicaParser.IDENT]:             # <<<<<<<<<<<<<<
 *                     self.state = 502
 *                     self.component_clause1()
 */
          __Pyx_INCREF(__pyx_v_token);
          __pyx_t_1 = __pyx_v_token;
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3666, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__32); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3666, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3666, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3666, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (!__pyx_t_12) {
          } else {
            __pyx_t_11 = __pyx_t_12;
            goto __pyx_L52_bool_binop_done;
          }
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3666, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__33); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3666, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3666, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3666, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (!__pyx_t_12) {
          } else {
            __pyx_t_11 = __pyx_t_12;
            goto __pyx_L52_bool_binop_done;
          }
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3666, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__34); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3666, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3666, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3666, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (!__pyx_t_12) {
          } else {
            __pyx_t_11 = __pyx_t_12;
            goto __pyx_L52_bool_binop_done;
          }
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3666, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__35); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3666, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3666, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3666, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (!__pyx_t_12) {
          } else {
            __pyx_t_11 = __pyx_t_12;
            goto __pyx_L52_bool_binop_done;
          }
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3666, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__36); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3666, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3666, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3666, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (!__pyx_t_12) {
          } else {
            __pyx_t_11 = __pyx_t_12;
            goto __pyx_L52_bool_binop_done;
          }
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3666, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__37); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3666, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3666, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3666, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (!__pyx_t_12) {
          } else {
            __pyx_t_11 = __pyx_t_12;
            goto __pyx_L52_bool_binop_done;
          }
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3666, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__38); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3666, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3666, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3666, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (!__pyx_t_12) {
          } else {
            __pyx_t_11 = __pyx_t_12;
            goto __pyx_L52_bool_binop_done;
          }
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3666, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3666, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3666, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3666, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_11 = __pyx_t_12;
          __pyx_L52_bool_binop_done:;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_12 = (__pyx_t_11 != 0);
          if (__pyx_t_12) {

            /* "pymola/generated/ModelicaParser.pyx":3667
 *                     pass
 *                 elif token in [ModelicaParser.T__32, ModelicaParser.T__33, ModelicaParser.T__34, ModelicaParser.T__35, ModelicaParser.T__36, ModelicaParser.T__37, ModelicaParser.T__38, ModelicaParser.IDENT]:
 *                     self.state = 502             # <<<<<<<<<<<<<<
 *                     self.component_clause1()
 *                     pass
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_502) < 0) __PYX_ERR(0, 3667, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":3668
 *                 elif token in [ModelicaParser.T__32, ModelicaParser.T__33, ModelicaParser.T__34, ModelicaParser.T__35, ModelicaParser.T__36, ModelicaParser.T__37, ModelicaParser.T__38, ModelicaParser.IDENT]:
 *                     self.state = 502
 *                     self.component_clause1()             # <<<<<<<<<<<<<<
 *                     pass
 *                 else:
 */
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_component_clause1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3668, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_7 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
              __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
              if (likely(__pyx_t_7)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                __Pyx_INCREF(__pyx_t_7);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_4, function);
              }
            }
            if (__pyx_t_7) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3668, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else {
              __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3668, __pyx_L6_error)
            }
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":3666
 *                     self.short_class_definition()
 *                     pass
 *                 elif token in [ModelicaParser.T__32, ModelicaParser.T__33, ModelicaParser.T__34, ModelicaParser.T__35, ModelicaParser.T__36, ModelicaParser.T__37, ModelicaParser.T__38, ModelicaParser.IDENT]:             # <<<<<<<<<<<<<<
 *                     self.state = 502
 *                     self.component_clause1()
 */
            goto __pyx_L38;
          }

          /* "pymola/generated/ModelicaParser.pyx":3671
 *                     pass
 *                 else:
 *                     raise NoViableAltException(self)             # <<<<<<<<<<<<<<
 * 
 *                 pass
 */
          /*else*/ {
            __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_NoViableAltException); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3671, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_7 = NULL;
            if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
              __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
              if (likely(__pyx_t_7)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                __Pyx_INCREF(__pyx_t_7);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_4, function);
              }
            }
            if (!__pyx_t_7) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3671, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_4)) {
                PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3671, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
                PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3671, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else
              #endif
              {
                __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3671, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_7); __pyx_t_7 = NULL;
                __Pyx_INCREF(__pyx_v_self);
                __Pyx_GIVEREF(__pyx_v_self);
                PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3671, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_Raise(__pyx_t_1, 0, 0, 0);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __PYX_ERR(0, 3671, __pyx_L6_error)
          }
          __pyx_L38:;

          /* "pymola/generated/ModelicaParser.pyx":3658
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.T__1, ModelicaParser.T__2, ModelicaParser.T__3, ModelicaParser.T__4, ModelicaParser.T__5, ModelicaParser.T__6, ModelicaParser.T__7, ModelicaParser.T__8, ModelicaParser.T__9, ModelicaParser.T__10, ModelicaParser.T__11, ModelicaParser.T__12, ModelicaParser.T__32, ModelicaParser.T__33, ModelicaParser.T__34, ModelicaParser.T__35, ModelicaParser.T__36, ModelicaParser.T__37, ModelicaParser.T__38, ModelicaParser.PARTIAL, ModelicaParser.IDENT]:             # <<<<<<<<<<<<<<
 *                 self.state = 503
 *                 self._errHandler.sync(self)
 */
          goto __pyx_L16;
        }

        /* "pymola/generated/ModelicaParser.pyx":3674
 * 
 *                 pass
 *             elif token in [ModelicaParser.T__25]:             # <<<<<<<<<<<<<<
 *                 self.state = 505
 *                 self.element_replaceable()
 */
        __Pyx_INCREF(__pyx_v_token);
        __pyx_t_1 = __pyx_v_token;
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3674, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__25); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3674, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3674, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3674, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_11 = (__pyx_t_12 != 0);
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":3675
 *                 pass
 *             elif token in [ModelicaParser.T__25]:
 *                 self.state = 505             # <<<<<<<<<<<<<<
 *                 self.element_replaceable()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_505) < 0) __PYX_ERR(0, 3675, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":3676
 *             elif token in [ModelicaParser.T__25]:
 *                 self.state = 505
 *                 self.element_replaceable()             # <<<<<<<<<<<<<<
 *                 pass
 *             else:
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_element_replaceable); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3676, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3676, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3676, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3674
 * 
 *                 pass
 *             elif token in [ModelicaParser.T__25]:             # <<<<<<<<<<<<<<
 *                 self.state = 505
 *                 self.element_replaceable()
 */
          goto __pyx_L16;
        }

        /* "pymola/generated/ModelicaParser.pyx":3679
 *                 pass
 *             else:
 *                 raise NoViableAltException(self)             # <<<<<<<<<<<<<<
 * 
 *         except RecognitionException as re:
 */
        /*else*/ {
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_NoViableAltException); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3679, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3679, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3679, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3679, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3679, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3679, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_Raise(__pyx_t_1, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __PYX_ERR(0, 3679, __pyx_L6_error)
        }
        __pyx_L16:;

        /* "pymola/generated/ModelicaParser.pyx":3635
 *         self.enterRule(localctx, 66, self.RULE_element_redeclaration)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 494
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":3681
 *                 raise NoViableAltException(self)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3681, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.element_redeclaration", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_4, &__pyx_t_7) < 0) __PYX_ERR(0, 3681, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_4);
        __pyx_v_re = __pyx_t_4;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":3682
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 3682, __pyx_L65_error)

          /* "pymola/generated/ModelicaParser.pyx":3683
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3683, __pyx_L65_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3683, __pyx_L65_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3683, __pyx_L65_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3683, __pyx_L65_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3683, __pyx_L65_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_13, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3683, __pyx_L65_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3684
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3684, __pyx_L65_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3684, __pyx_L65_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3684, __pyx_L65_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3684, __pyx_L65_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3684, __pyx_L65_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3684, __pyx_L65_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":3681
 *                 raise NoViableAltException(self)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L66;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L65_error:;
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_6 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L8_except_error;
          }
          __pyx_L66:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":3635
 *         self.enterRule(localctx, 66, self.RULE_element_redeclaration)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 494
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":3686
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3686, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3686, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3686, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_14 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3686, __pyx_L72_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3686, __pyx_L72_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3686, __pyx_L72_error)
        }
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
      __pyx_L72_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":3687
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Element_replaceableContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3630
 * 
 * 
 *     def element_redeclaration(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Element_redeclarationContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.element_redeclaration", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_token);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3691
 *     class Element_replaceableContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 3691, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 3691, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3691, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_replaceableContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Element_replaceableContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Element_replaceableContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":3692
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 3692, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3692, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3692, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3692, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3692, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3693
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def short_class_definition(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 3693, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3691
 *     class Element_replaceableContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_replaceableContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3695
 *             self.parser = parser
 * 
 *         def short_class_definition(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Short_class_definitionContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_3short_class_definition(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_3short_class_definition = {"short_class_definition", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_3short_class_definition, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_3short_class_definition(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("short_class_definition (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_2short_class_definition(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_2short_class_definition(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("short_class_definition", 0);

  /* "pymola/generated/ModelicaParser.pyx":3696
 * 
 *         def short_class_definition(self):
 *             return self.getTypedRuleContext(ModelicaParser.Short_class_definitionContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Short_class_definitionContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3696, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3696, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3696, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3696, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3695
 *             self.parser = parser
 * 
 *         def short_class_definition(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Short_class_definitionContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_replaceableContext.short_class_definition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3699
 * 
 * 
 *         def component_clause1(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_clause1Context,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_5component_clause1(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_5component_clause1 = {"component_clause1", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_5component_clause1, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_5component_clause1(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("component_clause1 (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_4component_clause1(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_4component_clause1(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("component_clause1", 0);

  /* "pymola/generated/ModelicaParser.pyx":3700
 * 
 *         def component_clause1(self):
 *             return self.getTypedRuleContext(ModelicaParser.Component_clause1Context,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Component_clause1Context); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3700, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3700, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3700, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3700, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3699
 * 
 * 
 *         def component_clause1(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_clause1Context,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_replaceableContext.component_clause1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3703
 * 
 * 
 *         def constraining_clause(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Constraining_clauseContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_7constraining_clause(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_7constraining_clause = {"constraining_clause", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_7constraining_clause, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_7constraining_clause(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("constraining_clause (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_6constraining_clause(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_6constraining_clause(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("constraining_clause", 0);

  /* "pymola/generated/ModelicaParser.pyx":3704
 * 
 *         def constraining_clause(self):
 *             return self.getTypedRuleContext(ModelicaParser.Constraining_clauseContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3704, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3704, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Constraining_clauseContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3704, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3704, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3704, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3704, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3704, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3703
 * 
 * 
 *         def constraining_clause(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Constraining_clauseContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_replaceableContext.constraining_clause", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3707
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_element_replaceable
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_9getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_9getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_9getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_9getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_8getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_8getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":3708
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_element_replaceable             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_element_replaceable); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3707
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_element_replaceable
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_replaceableContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3710
 *             return ModelicaParser.RULE_element_replaceable
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterElement_replaceable" ):
 *                 listener.enterElement_replaceable(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_11enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_11enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_11enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_11enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 3710, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 3710, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3710, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_replaceableContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_10enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_10enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":3711
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterElement_replaceable" ):             # <<<<<<<<<<<<<<
 *                 listener.enterElement_replaceable(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterElement_replaceable); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3711, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3712
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterElement_replaceable" ):
 *                 listener.enterElement_replaceable(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterElement_replaceable); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3712, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3712, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3712, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3712, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3712, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3712, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":3711
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterElement_replaceable" ):             # <<<<<<<<<<<<<<
 *                 listener.enterElement_replaceable(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3710
 *             return ModelicaParser.RULE_element_replaceable
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterElement_replaceable" ):
 *                 listener.enterElement_replaceable(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_replaceableContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3714
 *                 listener.enterElement_replaceable(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitElement_replaceable" ):
 *                 listener.exitElement_replaceable(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_13exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_13exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_13exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_13exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 3714, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 3714, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3714, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_replaceableContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_12exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_12exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":3715
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitElement_replaceable" ):             # <<<<<<<<<<<<<<
 *                 listener.exitElement_replaceable(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitElement_replaceable); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3715, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3716
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitElement_replaceable" ):
 *                 listener.exitElement_replaceable(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitElement_replaceable); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3716, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3716, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3716, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3716, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3716, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3716, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":3715
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitElement_replaceable" ):             # <<<<<<<<<<<<<<
 *                 listener.exitElement_replaceable(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3714
 *                 listener.enterElement_replaceable(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitElement_replaceable" ):
 *                 listener.exitElement_replaceable(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_replaceableContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3718
 *                 listener.exitElement_replaceable(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitElement_replaceable" ):
 *                 return visitor.visitElement_replaceable(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_15accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_15accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_15accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_15accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 3718, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 3718, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3718, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_replaceableContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_14accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_14accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":3719
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitElement_replaceable" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitElement_replaceable(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitElement_replaceable); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3719, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3720
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitElement_replaceable" ):
 *                 return visitor.visitElement_replaceable(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitElement_replaceable); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3720, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3720, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3720, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3720, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3720, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3720, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":3719
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitElement_replaceable" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitElement_replaceable(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3722
 *                 return visitor.visitElement_replaceable(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3722, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3722, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3722, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3722, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3722, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3722, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":3718
 *                 listener.exitElement_replaceable(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitElement_replaceable" ):
 *                 return visitor.visitElement_replaceable(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Element_replaceableContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3727
 * 
 * 
 *     def element_replaceable(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Element_replaceableContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_71element_replaceable(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_71element_replaceable = {"element_replaceable", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_71element_replaceable, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_71element_replaceable(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("element_replaceable (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_70element_replaceable(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_70element_replaceable(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v_token = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  char const *__pyx_t_22;
  __Pyx_RefNannySetupContext("element_replaceable", 0);

  /* "pymola/generated/ModelicaParser.pyx":3729
 *     def element_replaceable(self):
 * 
 *         localctx = ModelicaParser.Element_replaceableContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 68, self.RULE_element_replaceable)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3729, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Element_replaceableContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3729, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3729, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3729, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3729, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3729, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3729, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3729, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3730
 * 
 *         localctx = ModelicaParser.Element_replaceableContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 68, self.RULE_element_replaceable)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_element_replaceable); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_68, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3730, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_68, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3730, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_68);
    __Pyx_GIVEREF(__pyx_int_68);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_68);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3731
 *         localctx = ModelicaParser.Element_replaceableContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 68, self.RULE_element_replaceable)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 3731, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3732
 *         self.enterRule(localctx, 68, self.RULE_element_replaceable)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 508
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":3733
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 508
 *             self.match(ModelicaParser.T__25)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3733, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3733, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3733, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3733, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3733, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3734
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 508             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__25)
 *             self.state = 511
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_508) < 0) __PYX_ERR(0, 3734, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":3735
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 508
 *             self.match(ModelicaParser.T__25)             # <<<<<<<<<<<<<<
 *             self.state = 511
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3735, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3735, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__25); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3735, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3735, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3735, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3735, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3735, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3735, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3736
 *             self.state = 508
 *             self.match(ModelicaParser.T__25)
 *             self.state = 511             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_511) < 0) __PYX_ERR(0, 3736, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":3737
 *             self.match(ModelicaParser.T__25)
 *             self.state = 511
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.T__1, ModelicaParser.T__2, ModelicaParser.T__3, ModelicaParser.T__4, ModelicaParser.T__5, ModelicaParser.T__6, ModelicaParser.T__7, ModelicaParser.T__8, ModelicaParser.T__9, ModelicaParser.T__10, ModelicaParser.T__11, ModelicaParser.T__12, ModelicaParser.PARTIAL]:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3737, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3737, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3737, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3737, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3737, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3737, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3737, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3738
 *             self.state = 511
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if token in [ModelicaParser.T__1, ModelicaParser.T__2, ModelicaParser.T__3, ModelicaParser.T__4, ModelicaParser.T__5, ModelicaParser.T__6, ModelicaParser.T__7, ModelicaParser.T__8, ModelicaParser.T__9, ModelicaParser.T__10, ModelicaParser.T__11, ModelicaParser.T__12, ModelicaParser.PARTIAL]:
 *                 self.state = 509
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3738, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3738, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__886, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3738, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_v_token = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3739
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.T__1, ModelicaParser.T__2, ModelicaParser.T__3, ModelicaParser.T__4, ModelicaParser.T__5, ModelicaParser.T__6, ModelicaParser.T__7, ModelicaParser.T__8, ModelicaParser.T__9, ModelicaParser.T__10, ModelicaParser.T__11, ModelicaParser.T__12, ModelicaParser.PARTIAL]:             # <<<<<<<<<<<<<<
 *                 self.state = 509
 *                 self.short_class_definition()
 */
        __Pyx_INCREF(__pyx_v_token);
        __pyx_t_1 = __pyx_v_token;
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__11); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__12); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_PARTIAL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3739, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __pyx_t_12;
        __pyx_L15_bool_binop_done:;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_12 = (__pyx_t_11 != 0);
        if (__pyx_t_12) {

          /* "pymola/generated/ModelicaParser.pyx":3740
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.T__1, ModelicaParser.T__2, ModelicaParser.T__3, ModelicaParser.T__4, ModelicaParser.T__5, ModelicaParser.T__6, ModelicaParser.T__7, ModelicaParser.T__8, ModelicaParser.T__9, ModelicaParser.T__10, ModelicaParser.T__11, ModelicaParser.T__12, ModelicaParser.PARTIAL]:
 *                 self.state = 509             # <<<<<<<<<<<<<<
 *                 self.short_class_definition()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_509) < 0) __PYX_ERR(0, 3740, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":3741
 *             if token in [ModelicaParser.T__1, ModelicaParser.T__2, ModelicaParser.T__3, ModelicaParser.T__4, ModelicaParser.T__5, ModelicaParser.T__6, ModelicaParser.T__7, ModelicaParser.T__8, ModelicaParser.T__9, ModelicaParser.T__10, ModelicaParser.T__11, ModelicaParser.T__12, ModelicaParser.PARTIAL]:
 *                 self.state = 509
 *                 self.short_class_definition()             # <<<<<<<<<<<<<<
 *                 pass
 *             elif token in [ModelicaParser.T__32, ModelicaParser.T__33, ModelicaParser.T__34, ModelicaParser.T__35, ModelicaParser.T__36, ModelicaParser.T__37, ModelicaParser.T__38, ModelicaParser.IDENT]:
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_short_class_definition); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3741, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3741, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3741, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3739
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.T__1, ModelicaParser.T__2, ModelicaParser.T__3, ModelicaParser.T__4, ModelicaParser.T__5, ModelicaParser.T__6, ModelicaParser.T__7, ModelicaParser.T__8, ModelicaParser.T__9, ModelicaParser.T__10, ModelicaParser.T__11, ModelicaParser.T__12, ModelicaParser.PARTIAL]:             # <<<<<<<<<<<<<<
 *                 self.state = 509
 *                 self.short_class_definition()
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":3743
 *                 self.short_class_definition()
 *                 pass
 *             elif token in [ModelicaParser.T__32, ModelicaParser.T__33, ModelicaParser.T__34, ModelicaParser.T__35, ModelicaParser.T__36, ModelicaParser.T__37, ModelicaParser.T__38, ModelicaParser.IDENT]:             # <<<<<<<<<<<<<<
 *                 self.state = 510
 *                 self.component_clause1()
 */
        __Pyx_INCREF(__pyx_v_token);
        __pyx_t_1 = __pyx_v_token;
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3743, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__32); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3743, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3743, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 3743, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L28_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3743, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__33); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3743, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3743, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 3743, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L28_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3743, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__34); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3743, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3743, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 3743, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L28_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3743, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__35); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3743, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3743, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 3743, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L28_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3743, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__36); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3743, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3743, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 3743, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L28_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3743, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__37); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3743, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3743, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 3743, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L28_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3743, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__38); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3743, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3743, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 3743, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L28_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3743, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3743, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3743, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 3743, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_12 = __pyx_t_11;
        __pyx_L28_bool_binop_done:;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_11 = (__pyx_t_12 != 0);
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":3744
 *                 pass
 *             elif token in [ModelicaParser.T__32, ModelicaParser.T__33, ModelicaParser.T__34, ModelicaParser.T__35, ModelicaParser.T__36, ModelicaParser.T__37, ModelicaParser.T__38, ModelicaParser.IDENT]:
 *                 self.state = 510             # <<<<<<<<<<<<<<
 *                 self.component_clause1()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_510) < 0) __PYX_ERR(0, 3744, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":3745
 *             elif token in [ModelicaParser.T__32, ModelicaParser.T__33, ModelicaParser.T__34, ModelicaParser.T__35, ModelicaParser.T__36, ModelicaParser.T__37, ModelicaParser.T__38, ModelicaParser.IDENT]:
 *                 self.state = 510
 *                 self.component_clause1()             # <<<<<<<<<<<<<<
 *                 pass
 *             else:
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_component_clause1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3745, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3745, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3745, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3743
 *                 self.short_class_definition()
 *                 pass
 *             elif token in [ModelicaParser.T__32, ModelicaParser.T__33, ModelicaParser.T__34, ModelicaParser.T__35, ModelicaParser.T__36, ModelicaParser.T__37, ModelicaParser.T__38, ModelicaParser.IDENT]:             # <<<<<<<<<<<<<<
 *                 self.state = 510
 *                 self.component_clause1()
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":3748
 *                 pass
 *             else:
 *                 raise NoViableAltException(self)             # <<<<<<<<<<<<<<
 * 
 *             self.state = 514
 */
        /*else*/ {
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_NoViableAltException); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3748, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3748, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3748, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3748, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3748, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3748, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_Raise(__pyx_t_1, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __PYX_ERR(0, 3748, __pyx_L6_error)
        }
        __pyx_L14:;

        /* "pymola/generated/ModelicaParser.pyx":3750
 *                 raise NoViableAltException(self)
 * 
 *             self.state = 514             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_514) < 0) __PYX_ERR(0, 3750, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":3751
 * 
 *             self.state = 514
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__31:
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3751, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_sync); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3751, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3751, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3751, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3751, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3751, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3751, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3752
 *             self.state = 514
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.T__31:
 *                 self.state = 513
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3752, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3752, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__887, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3752, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_v__la = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3753
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__31:             # <<<<<<<<<<<<<<
 *                 self.state = 513
 *                 self.constraining_clause()
 */
        __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3753, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__31); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3753, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3753, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 3753, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":3754
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__31:
 *                 self.state = 513             # <<<<<<<<<<<<<<
 *                 self.constraining_clause()
 * 
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_513) < 0) __PYX_ERR(0, 3754, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":3755
 *             if _la==ModelicaParser.T__31:
 *                 self.state = 513
 *                 self.constraining_clause()             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_constraining_clause); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3755, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3755, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3755, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3753
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__31:             # <<<<<<<<<<<<<<
 *                 self.state = 513
 *                 self.constraining_clause()
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":3732
 *         self.enterRule(localctx, 68, self.RULE_element_replaceable)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 508
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":3758
 * 
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3758, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.element_replaceable", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_3, &__pyx_t_2) < 0) __PYX_ERR(0, 3758, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __pyx_v_re = __pyx_t_3;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":3759
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 3759, __pyx_L42_error)

          /* "pymola/generated/ModelicaParser.pyx":3760
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3760, __pyx_L42_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3760, __pyx_L42_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_re};
            __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3760, __pyx_L42_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_re};
            __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3760, __pyx_L42_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3760, __pyx_L42_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_7) {
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_7); __pyx_t_7 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_13, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3760, __pyx_L42_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3761
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3761, __pyx_L42_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3761, __pyx_L42_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3761, __pyx_L42_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3761, __pyx_L42_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3761, __pyx_L42_error)
            __Pyx_GOTREF(__pyx_t_7);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_7, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3761, __pyx_L42_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":3758
 * 
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L43;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L42_error:;
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_6 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L8_except_error;
          }
          __pyx_L43:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":3732
 *         self.enterRule(localctx, 68, self.RULE_element_replaceable)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 508
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":3763
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3763, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3763, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3763, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_14 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3763, __pyx_L49_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3763, __pyx_L49_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3763, __pyx_L49_error)
        }
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
      __pyx_L49_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":3764
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Component_clause1Context(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3727
 * 
 * 
 *     def element_replaceable(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Element_replaceableContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.element_replaceable", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v_token);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3768
 *     class Component_clause1Context(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Component_clause1Context_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Component_clause1Context_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Component_clause1Context_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Component_clause1Context_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 3768, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 3768, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3768, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_clause1Context.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Component_clause1Context___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Component_clause1Context___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":3769
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 3769, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3769, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3769, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3769, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3769, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3769, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3769, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3769, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3770
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def type_prefix(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 3770, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3768
 *     class Component_clause1Context(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_clause1Context.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3772
 *             self.parser = parser
 * 
 *         def type_prefix(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Type_prefixContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Component_clause1Context_3type_prefix(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Component_clause1Context_3type_prefix = {"type_prefix", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Component_clause1Context_3type_prefix, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Component_clause1Context_3type_prefix(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("type_prefix (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Component_clause1Context_2type_prefix(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Component_clause1Context_2type_prefix(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("type_prefix", 0);

  /* "pymola/generated/ModelicaParser.pyx":3773
 * 
 *         def type_prefix(self):
 *             return self.getTypedRuleContext(ModelicaParser.Type_prefixContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3773, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3773, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Type_prefixContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3773, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3773, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3773, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3773, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3773, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3772
 *             self.parser = parser
 * 
 *         def type_prefix(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Type_prefixContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_clause1Context.type_prefix", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3776
 * 
 * 
 *         def type_specifier(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Type_specifierContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Component_clause1Context_5type_specifier(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Component_clause1Context_5type_specifier = {"type_specifier", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Component_clause1Context_5type_specifier, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Component_clause1Context_5type_specifier(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("type_specifier (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Component_clause1Context_4type_specifier(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Component_clause1Context_4type_specifier(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("type_specifier", 0);

  /* "pymola/generated/ModelicaParser.pyx":3777
 * 
 *         def type_specifier(self):
 *             return self.getTypedRuleContext(ModelicaParser.Type_specifierContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3777, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3777, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Type_specifierContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3777, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3777, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3777, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3777, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3777, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3776
 * 
 * 
 *         def type_specifier(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Type_specifierContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_clause1Context.type_specifier", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3780
 * 
 * 
 *         def component_declaration1(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_declaration1Context,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Component_clause1Context_7component_declaration1(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Component_clause1Context_7component_declaration1 = {"component_declaration1", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Component_clause1Context_7component_declaration1, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Component_clause1Context_7component_declaration1(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("component_declaration1 (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Component_clause1Context_6component_declaration1(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Component_clause1Context_6component_declaration1(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("component_declaration1", 0);

  /* "pymola/generated/ModelicaParser.pyx":3781
 * 
 *         def component_declaration1(self):
 *             return self.getTypedRuleContext(ModelicaParser.Component_declaration1Context,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3781, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3781, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Component_declaration1Context); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3781, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3781, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3781, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3781, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3781, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3780
 * 
 * 
 *         def component_declaration1(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_declaration1Context,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_clause1Context.component_declaration1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3784
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_component_clause1
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Component_clause1Context_9getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Component_clause1Context_9getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Component_clause1Context_9getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Component_clause1Context_9getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Component_clause1Context_8getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Component_clause1Context_8getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":3785
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_component_clause1             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3785, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_component_clause1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3785, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3784
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_component_clause1
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_clause1Context.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3787
 *             return ModelicaParser.RULE_component_clause1
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterComponent_clause1" ):
 *                 listener.enterComponent_clause1(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Component_clause1Context_11enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Component_clause1Context_11enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Component_clause1Context_11enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Component_clause1Context_11enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 3787, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 3787, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3787, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_clause1Context.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Component_clause1Context_10enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Component_clause1Context_10enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":3788
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterComponent_clause1" ):             # <<<<<<<<<<<<<<
 *                 listener.enterComponent_clause1(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterComponent_clause1); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3788, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3789
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterComponent_clause1" ):
 *                 listener.enterComponent_clause1(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterComponent_clause1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3789, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3789, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3789, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3789, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3789, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3789, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":3788
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterComponent_clause1" ):             # <<<<<<<<<<<<<<
 *                 listener.enterComponent_clause1(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3787
 *             return ModelicaParser.RULE_component_clause1
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterComponent_clause1" ):
 *                 listener.enterComponent_clause1(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_clause1Context.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3791
 *                 listener.enterComponent_clause1(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitComponent_clause1" ):
 *                 listener.exitComponent_clause1(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Component_clause1Context_13exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Component_clause1Context_13exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Component_clause1Context_13exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Component_clause1Context_13exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 3791, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 3791, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3791, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_clause1Context.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Component_clause1Context_12exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Component_clause1Context_12exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":3792
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitComponent_clause1" ):             # <<<<<<<<<<<<<<
 *                 listener.exitComponent_clause1(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitComponent_clause1); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3792, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3793
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitComponent_clause1" ):
 *                 listener.exitComponent_clause1(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitComponent_clause1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3793, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3793, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3793, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3793, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3793, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3793, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":3792
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitComponent_clause1" ):             # <<<<<<<<<<<<<<
 *                 listener.exitComponent_clause1(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3791
 *                 listener.enterComponent_clause1(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitComponent_clause1" ):
 *                 listener.exitComponent_clause1(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_clause1Context.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3795
 *                 listener.exitComponent_clause1(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitComponent_clause1" ):
 *                 return visitor.visitComponent_clause1(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Component_clause1Context_15accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Component_clause1Context_15accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Component_clause1Context_15accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Component_clause1Context_15accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 3795, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 3795, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3795, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_clause1Context.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Component_clause1Context_14accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Component_clause1Context_14accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":3796
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitComponent_clause1" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitComponent_clause1(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitComponent_clause1); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3796, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3797
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitComponent_clause1" ):
 *                 return visitor.visitComponent_clause1(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitComponent_clause1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3797, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3797, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3797, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3797, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3797, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3797, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":3796
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitComponent_clause1" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitComponent_clause1(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3799
 *                 return visitor.visitComponent_clause1(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3799, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3799, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3799, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3799, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3799, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3799, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":3795
 *                 listener.exitComponent_clause1(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitComponent_clause1" ):
 *                 return visitor.visitComponent_clause1(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_clause1Context.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3804
 * 
 * 
 *     def component_clause1(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Component_clause1Context(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_73component_clause1(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_73component_clause1 = {"component_clause1", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_73component_clause1, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_73component_clause1(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("component_clause1 (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_72component_clause1(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_72component_clause1(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  char const *__pyx_t_20;
  __Pyx_RefNannySetupContext("component_clause1", 0);

  /* "pymola/generated/ModelicaParser.pyx":3806
 *     def component_clause1(self):
 * 
 *         localctx = ModelicaParser.Component_clause1Context(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 70, self.RULE_component_clause1)
 *         try:
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3806, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Component_clause1Context); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3806, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3806, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3806, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3806, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3806, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3806, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3806, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3807
 * 
 *         localctx = ModelicaParser.Component_clause1Context(self, self._ctx, self.state)
 *         self.enterRule(localctx, 70, self.RULE_component_clause1)             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3807, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_component_clause1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3807, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_70, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3807, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_70, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3807, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3807, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_70);
    __Pyx_GIVEREF(__pyx_int_70);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_70);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3807, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3808
 *         localctx = ModelicaParser.Component_clause1Context(self, self._ctx, self.state)
 *         self.enterRule(localctx, 70, self.RULE_component_clause1)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 516
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":3809
 *         self.enterRule(localctx, 70, self.RULE_component_clause1)
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 516
 *             self.type_prefix()
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3809, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3809, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3809, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3809, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3809, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3810
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 516             # <<<<<<<<<<<<<<
 *             self.type_prefix()
 *             self.state = 517
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_516) < 0) __PYX_ERR(0, 3810, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":3811
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 516
 *             self.type_prefix()             # <<<<<<<<<<<<<<
 *             self.state = 517
 *             self.type_specifier()
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_type_prefix); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3811, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3811, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3811, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3812
 *             self.state = 516
 *             self.type_prefix()
 *             self.state = 517             # <<<<<<<<<<<<<<
 *             self.type_specifier()
 *             self.state = 518
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_517) < 0) __PYX_ERR(0, 3812, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":3813
 *             self.type_prefix()
 *             self.state = 517
 *             self.type_specifier()             # <<<<<<<<<<<<<<
 *             self.state = 518
 *             self.component_declaration1()
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_type_specifier); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3813, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3813, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3813, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3814
 *             self.state = 517
 *             self.type_specifier()
 *             self.state = 518             # <<<<<<<<<<<<<<
 *             self.component_declaration1()
 *         except RecognitionException as re:
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_518) < 0) __PYX_ERR(0, 3814, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":3815
 *             self.type_specifier()
 *             self.state = 518
 *             self.component_declaration1()             # <<<<<<<<<<<<<<
 *         except RecognitionException as re:
 *             localctx.exception = re
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_component_declaration1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3815, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3815, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3815, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3808
 *         localctx = ModelicaParser.Component_clause1Context(self, self._ctx, self.state)
 *         self.enterRule(localctx, 70, self.RULE_component_clause1)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 516
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":3816
 *             self.state = 518
 *             self.component_declaration1()
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3816, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.component_clause1", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_3, &__pyx_t_7) < 0) __PYX_ERR(0, 3816, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_3);
        __pyx_v_re = __pyx_t_3;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":3817
 *             self.component_declaration1()
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 3817, __pyx_L19_error)

          /* "pymola/generated/ModelicaParser.pyx":3818
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3818, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3818, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3818, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3818, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_11 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 3818, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_11);
            if (__pyx_t_4) {
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_4); __pyx_t_4 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_11, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3818, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3819
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3819, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 3819, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_11)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3819, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_11)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3819, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3819, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3819, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":3816
 *             self.state = 518
 *             self.component_declaration1()
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L20;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L19_error:;
            __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_14);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __pyx_t_6 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_17);
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
            }
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
            __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
            goto __pyx_L8_except_error;
          }
          __pyx_L20:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":3808
 *         localctx = ModelicaParser.Component_clause1Context(self, self._ctx, self.state)
 *         self.enterRule(localctx, 70, self.RULE_component_clause1)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 516
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":3821
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3821, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3821, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3821, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_18, &__pyx_t_17);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_18);
      __Pyx_XGOTREF(__pyx_t_17);
      __pyx_t_12 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_20 = __pyx_filename;
      {
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3821, __pyx_L26_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3821, __pyx_L26_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3821, __pyx_L26_error)
        }
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_18, __pyx_t_17);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0;
      __pyx_lineno = __pyx_t_12; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_20;
      goto __pyx_L1_error;
      __pyx_L26_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_18, __pyx_t_17);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":3822
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Component_declaration1Context(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3804
 * 
 * 
 *     def component_clause1(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Component_clause1Context(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.component_clause1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3826
 *     class Component_declaration1Context(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 3826, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 3826, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3826, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_declaration1Context.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_29Component_declaration1Context___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Component_declaration1Context___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":3827
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 3827, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3827, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3827, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3827, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3827, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3827, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3828
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def declaration(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 3828, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3826
 *     class Component_declaration1Context(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_declaration1Context.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3830
 *             self.parser = parser
 * 
 *         def declaration(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.DeclarationContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_3declaration(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_3declaration = {"declaration", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_3declaration, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_3declaration(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("declaration (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_2declaration(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_2declaration(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("declaration", 0);

  /* "pymola/generated/ModelicaParser.pyx":3831
 * 
 *         def declaration(self):
 *             return self.getTypedRuleContext(ModelicaParser.DeclarationContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_DeclarationContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3831, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3831, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3831, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3831, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3830
 *             self.parser = parser
 * 
 *         def declaration(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.DeclarationContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_declaration1Context.declaration", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3834
 * 
 * 
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_5comment(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_5comment = {"comment", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_5comment, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_5comment(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("comment (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_4comment(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_4comment(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("comment", 0);

  /* "pymola/generated/ModelicaParser.pyx":3835
 * 
 *         def comment(self):
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3835, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3835, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_CommentContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3835, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3835, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3835, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3835, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3835, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3834
 * 
 * 
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_declaration1Context.comment", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3838
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_component_declaration1
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_7getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_7getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_6getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":3839
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_component_declaration1             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3839, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_component_declaration1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3839, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3838
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_component_declaration1
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_declaration1Context.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3841
 *             return ModelicaParser.RULE_component_declaration1
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterComponent_declaration1" ):
 *                 listener.enterComponent_declaration1(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_9enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_9enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 3841, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 3841, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3841, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_declaration1Context.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_8enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":3842
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterComponent_declaration1" ):             # <<<<<<<<<<<<<<
 *                 listener.enterComponent_declaration1(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterComponent_declaration1); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3842, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3843
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterComponent_declaration1" ):
 *                 listener.enterComponent_declaration1(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterComponent_declaration1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3843, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3843, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3843, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3843, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3843, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3843, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":3842
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterComponent_declaration1" ):             # <<<<<<<<<<<<<<
 *                 listener.enterComponent_declaration1(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3841
 *             return ModelicaParser.RULE_component_declaration1
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterComponent_declaration1" ):
 *                 listener.enterComponent_declaration1(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_declaration1Context.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3845
 *                 listener.enterComponent_declaration1(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitComponent_declaration1" ):
 *                 listener.exitComponent_declaration1(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_11exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_11exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 3845, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 3845, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3845, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_declaration1Context.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_10exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":3846
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitComponent_declaration1" ):             # <<<<<<<<<<<<<<
 *                 listener.exitComponent_declaration1(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitComponent_declaration1); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3846, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3847
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitComponent_declaration1" ):
 *                 listener.exitComponent_declaration1(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitComponent_declaration1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3847, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3847, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3847, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3847, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3847, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3847, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":3846
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitComponent_declaration1" ):             # <<<<<<<<<<<<<<
 *                 listener.exitComponent_declaration1(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3845
 *                 listener.enterComponent_declaration1(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitComponent_declaration1" ):
 *                 listener.exitComponent_declaration1(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_declaration1Context.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3849
 *                 listener.exitComponent_declaration1(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitComponent_declaration1" ):
 *                 return visitor.visitComponent_declaration1(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_13accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_13accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 3849, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 3849, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3849, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_declaration1Context.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_12accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":3850
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitComponent_declaration1" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitComponent_declaration1(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitComponent_declaration1); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3850, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3851
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitComponent_declaration1" ):
 *                 return visitor.visitComponent_declaration1(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitComponent_declaration1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3851, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3851, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3851, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3851, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3851, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3851, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":3850
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitComponent_declaration1" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitComponent_declaration1(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3853
 *                 return visitor.visitComponent_declaration1(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3853, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3853, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3853, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3853, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3853, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3853, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":3849
 *                 listener.exitComponent_declaration1(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitComponent_declaration1" ):
 *                 return visitor.visitComponent_declaration1(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_declaration1Context.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3858
 * 
 * 
 *     def component_declaration1(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Component_declaration1Context(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_75component_declaration1(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_75component_declaration1 = {"component_declaration1", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_75component_declaration1, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_75component_declaration1(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("component_declaration1 (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_74component_declaration1(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_74component_declaration1(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  char const *__pyx_t_20;
  __Pyx_RefNannySetupContext("component_declaration1", 0);

  /* "pymola/generated/ModelicaParser.pyx":3860
 *     def component_declaration1(self):
 * 
 *         localctx = ModelicaParser.Component_declaration1Context(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 72, self.RULE_component_declaration1)
 *         try:
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3860, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Component_declaration1Context); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3860, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3860, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3860, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3860, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3860, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3860, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3860, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3861
 * 
 *         localctx = ModelicaParser.Component_declaration1Context(self, self._ctx, self.state)
 *         self.enterRule(localctx, 72, self.RULE_component_declaration1)             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_component_declaration1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_72, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3861, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_72, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3861, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3861, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_72);
    __Pyx_GIVEREF(__pyx_int_72);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_72);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3861, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3862
 *         localctx = ModelicaParser.Component_declaration1Context(self, self._ctx, self.state)
 *         self.enterRule(localctx, 72, self.RULE_component_declaration1)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 520
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":3863
 *         self.enterRule(localctx, 72, self.RULE_component_declaration1)
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 520
 *             self.declaration()
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3863, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3863, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3863, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3863, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3863, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3864
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 520             # <<<<<<<<<<<<<<
 *             self.declaration()
 *             self.state = 521
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_520) < 0) __PYX_ERR(0, 3864, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":3865
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 520
 *             self.declaration()             # <<<<<<<<<<<<<<
 *             self.state = 521
 *             self.comment()
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_declaration); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3865, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3865, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3865, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3866
 *             self.state = 520
 *             self.declaration()
 *             self.state = 521             # <<<<<<<<<<<<<<
 *             self.comment()
 *         except RecognitionException as re:
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_521) < 0) __PYX_ERR(0, 3866, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":3867
 *             self.declaration()
 *             self.state = 521
 *             self.comment()             # <<<<<<<<<<<<<<
 *         except RecognitionException as re:
 *             localctx.exception = re
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_comment); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3867, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3867, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3867, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3862
 *         localctx = ModelicaParser.Component_declaration1Context(self, self._ctx, self.state)
 *         self.enterRule(localctx, 72, self.RULE_component_declaration1)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 520
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":3868
 *             self.state = 521
 *             self.comment()
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3868, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.component_declaration1", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_3, &__pyx_t_7) < 0) __PYX_ERR(0, 3868, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_3);
        __pyx_v_re = __pyx_t_3;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":3869
 *             self.comment()
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 3869, __pyx_L19_error)

          /* "pymola/generated/ModelicaParser.pyx":3870
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3870, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3870, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3870, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3870, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_11 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 3870, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_11);
            if (__pyx_t_4) {
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_4); __pyx_t_4 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_11, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3870, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3871
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3871, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 3871, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_11)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3871, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_11)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3871, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3871, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3871, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":3868
 *             self.state = 521
 *             self.comment()
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L20;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L19_error:;
            __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_14);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __pyx_t_6 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_17);
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
            }
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
            __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
            goto __pyx_L8_except_error;
          }
          __pyx_L20:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":3862
 *         localctx = ModelicaParser.Component_declaration1Context(self, self._ctx, self.state)
 *         self.enterRule(localctx, 72, self.RULE_component_declaration1)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 520
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":3873
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3873, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3873, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3873, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_18, &__pyx_t_17);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_18);
      __Pyx_XGOTREF(__pyx_t_17);
      __pyx_t_12 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_20 = __pyx_filename;
      {
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3873, __pyx_L26_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3873, __pyx_L26_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3873, __pyx_L26_error)
        }
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_18, __pyx_t_17);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0;
      __pyx_lineno = __pyx_t_12; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_20;
      goto __pyx_L1_error;
      __pyx_L26_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_18, __pyx_t_17);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":3874
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Short_class_definitionContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3858
 * 
 * 
 *     def component_declaration1(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Component_declaration1Context(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.component_declaration1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3878
 *     class Short_class_definitionContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 3878, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 3878, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3878, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Short_class_definitionContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":3879
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 3879, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3879, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3879, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3879, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3879, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3879, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3879, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3879, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3880
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def class_prefixes(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 3880, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3878
 *     class Short_class_definitionContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Short_class_definitionContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3882
 *             self.parser = parser
 * 
 *         def class_prefixes(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_prefixesContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_3class_prefixes(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_3class_prefixes = {"class_prefixes", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_3class_prefixes, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_3class_prefixes(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("class_prefixes (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_2class_prefixes(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_2class_prefixes(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("class_prefixes", 0);

  /* "pymola/generated/ModelicaParser.pyx":3883
 * 
 *         def class_prefixes(self):
 *             return self.getTypedRuleContext(ModelicaParser.Class_prefixesContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3883, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3883, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Class_prefixesContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3883, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3883, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3883, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3883, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3883, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3882
 *             self.parser = parser
 * 
 *         def class_prefixes(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_prefixesContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Short_class_definitionContext.class_prefixes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3886
 * 
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_5IDENT(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_5IDENT = {"IDENT", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_5IDENT, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_5IDENT(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("IDENT (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_4IDENT(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_4IDENT(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("IDENT", 0);

  /* "pymola/generated/ModelicaParser.pyx":3887
 * 
 *         def IDENT(self):
 *             return self.getToken(ModelicaParser.IDENT, 0)             # <<<<<<<<<<<<<<
 * 
 *         def base_prefix(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getToken); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3887, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3887, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3887, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3887, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3887, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3887, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3887, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3886
 * 
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Short_class_definitionContext.IDENT", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3889
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 *         def base_prefix(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Base_prefixContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_7base_prefix(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_7base_prefix = {"base_prefix", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_7base_prefix, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_7base_prefix(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("base_prefix (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_6base_prefix(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_6base_prefix(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("base_prefix", 0);

  /* "pymola/generated/ModelicaParser.pyx":3890
 * 
 *         def base_prefix(self):
 *             return self.getTypedRuleContext(ModelicaParser.Base_prefixContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3890, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3890, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Base_prefixContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3890, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3890, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3890, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3890, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3890, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3889
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 *         def base_prefix(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Base_prefixContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Short_class_definitionContext.base_prefix", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3893
 * 
 * 
 *         def component_reference(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_9component_reference(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_9component_reference = {"component_reference", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_9component_reference, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_9component_reference(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("component_reference (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_8component_reference(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_8component_reference(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("component_reference", 0);

  /* "pymola/generated/ModelicaParser.pyx":3894
 * 
 *         def component_reference(self):
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3894, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3894, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Component_referenceContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3894, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3894, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3894, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3894, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3894, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3893
 * 
 * 
 *         def component_reference(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Short_class_definitionContext.component_reference", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3897
 * 
 * 
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_11comment(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_11comment = {"comment", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_11comment, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_11comment(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("comment (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_10comment(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_10comment(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("comment", 0);

  /* "pymola/generated/ModelicaParser.pyx":3898
 * 
 *         def comment(self):
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3898, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3898, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_CommentContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3898, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3898, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3898, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3898, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3898, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3897
 * 
 * 
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Short_class_definitionContext.comment", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3901
 * 
 * 
 *         def array_subscripts(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Array_subscriptsContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_13array_subscripts(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_13array_subscripts = {"array_subscripts", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_13array_subscripts, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_13array_subscripts(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("array_subscripts (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_12array_subscripts(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_12array_subscripts(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("array_subscripts", 0);

  /* "pymola/generated/ModelicaParser.pyx":3902
 * 
 *         def array_subscripts(self):
 *             return self.getTypedRuleContext(ModelicaParser.Array_subscriptsContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3902, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3902, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Array_subscriptsContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3902, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3902, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3902, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3902, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3902, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3901
 * 
 * 
 *         def array_subscripts(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Array_subscriptsContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Short_class_definitionContext.array_subscripts", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3905
 * 
 * 
 *         def class_modification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_modificationContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_15class_modification(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_15class_modification = {"class_modification", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_15class_modification, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_15class_modification(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("class_modification (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_14class_modification(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_14class_modification(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("class_modification", 0);

  /* "pymola/generated/ModelicaParser.pyx":3906
 * 
 *         def class_modification(self):
 *             return self.getTypedRuleContext(ModelicaParser.Class_modificationContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3906, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3906, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Class_modificationContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3906, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3906, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3906, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3906, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3906, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3905
 * 
 * 
 *         def class_modification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_modificationContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Short_class_definitionContext.class_modification", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3909
 * 
 * 
 *         def enum_list(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Enum_listContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_17enum_list(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_17enum_list = {"enum_list", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_17enum_list, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_17enum_list(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enum_list (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_16enum_list(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_16enum_list(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enum_list", 0);

  /* "pymola/generated/ModelicaParser.pyx":3910
 * 
 *         def enum_list(self):
 *             return self.getTypedRuleContext(ModelicaParser.Enum_listContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3910, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3910, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Enum_listContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3910, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3910, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3910, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3910, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3910, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3909
 * 
 * 
 *         def enum_list(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Enum_listContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Short_class_definitionContext.enum_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3913
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_short_class_definition
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_19getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_19getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_19getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_19getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_18getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_18getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":3914
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_short_class_definition             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3914, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_short_class_definition); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3914, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3913
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_short_class_definition
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Short_class_definitionContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3916
 *             return ModelicaParser.RULE_short_class_definition
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterShort_class_definition" ):
 *                 listener.enterShort_class_definition(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_21enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_21enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_21enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_21enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 3916, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 3916, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3916, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Short_class_definitionContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_20enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_20enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":3917
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterShort_class_definition" ):             # <<<<<<<<<<<<<<
 *                 listener.enterShort_class_definition(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterShort_class_definition); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3917, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3918
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterShort_class_definition" ):
 *                 listener.enterShort_class_definition(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterShort_class_definition); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3918, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3918, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3918, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3918, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3918, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3918, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":3917
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterShort_class_definition" ):             # <<<<<<<<<<<<<<
 *                 listener.enterShort_class_definition(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3916
 *             return ModelicaParser.RULE_short_class_definition
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterShort_class_definition" ):
 *                 listener.enterShort_class_definition(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Short_class_definitionContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3920
 *                 listener.enterShort_class_definition(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitShort_class_definition" ):
 *                 listener.exitShort_class_definition(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_23exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_23exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_23exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_23exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 3920, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 3920, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3920, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Short_class_definitionContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_22exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_22exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":3921
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitShort_class_definition" ):             # <<<<<<<<<<<<<<
 *                 listener.exitShort_class_definition(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitShort_class_definition); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3921, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3922
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitShort_class_definition" ):
 *                 listener.exitShort_class_definition(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitShort_class_definition); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3922, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3922, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3922, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3922, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3922, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3922, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":3921
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitShort_class_definition" ):             # <<<<<<<<<<<<<<
 *                 listener.exitShort_class_definition(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3920
 *                 listener.enterShort_class_definition(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitShort_class_definition" ):
 *                 listener.exitShort_class_definition(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Short_class_definitionContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3924
 *                 listener.exitShort_class_definition(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitShort_class_definition" ):
 *                 return visitor.visitShort_class_definition(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_25accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_25accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_25accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_25accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 3924, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 3924, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3924, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Short_class_definitionContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_24accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_24accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":3925
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitShort_class_definition" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitShort_class_definition(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitShort_class_definition); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 3925, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":3926
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitShort_class_definition" ):
 *                 return visitor.visitShort_class_definition(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitShort_class_definition); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3926, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3926, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3926, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3926, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3926, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3926, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":3925
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitShort_class_definition" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitShort_class_definition(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":3928
 *                 return visitor.visitShort_class_definition(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3928, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3928, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3928, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3928, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3928, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3928, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":3924
 *                 listener.exitShort_class_definition(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitShort_class_definition" ):
 *                 return visitor.visitShort_class_definition(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Short_class_definitionContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":3933
 * 
 * 
 *     def short_class_definition(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Short_class_definitionContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_77short_class_definition(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_77short_class_definition = {"short_class_definition", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_77short_class_definition, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_77short_class_definition(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("short_class_definition (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_76short_class_definition(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_76short_class_definition(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v_token = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  char const *__pyx_t_22;
  __Pyx_RefNannySetupContext("short_class_definition", 0);

  /* "pymola/generated/ModelicaParser.pyx":3935
 *     def short_class_definition(self):
 * 
 *         localctx = ModelicaParser.Short_class_definitionContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 74, self.RULE_short_class_definition)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3935, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Short_class_definitionContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3935, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3935, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3935, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3935, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3935, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3935, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3935, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3936
 * 
 *         localctx = ModelicaParser.Short_class_definitionContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 74, self.RULE_short_class_definition)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3936, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_short_class_definition); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3936, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_74, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3936, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_74, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3936, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3936, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_74);
    __Pyx_GIVEREF(__pyx_int_74);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_74);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3936, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3937
 *         localctx = ModelicaParser.Short_class_definitionContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 74, self.RULE_short_class_definition)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 3937, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3938
 *         self.enterRule(localctx, 74, self.RULE_short_class_definition)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 523
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":3939
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 523
 *             self.class_prefixes()
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3939, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3939, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3939, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3939, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3939, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3940
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 523             # <<<<<<<<<<<<<<
 *             self.class_prefixes()
 *             self.state = 524
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_523) < 0) __PYX_ERR(0, 3940, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":3941
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 523
 *             self.class_prefixes()             # <<<<<<<<<<<<<<
 *             self.state = 524
 *             self.match(ModelicaParser.IDENT)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_class_prefixes); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3941, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3941, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3941, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3942
 *             self.state = 523
 *             self.class_prefixes()
 *             self.state = 524             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.IDENT)
 *             self.state = 525
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_524) < 0) __PYX_ERR(0, 3942, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":3943
 *             self.class_prefixes()
 *             self.state = 524
 *             self.match(ModelicaParser.IDENT)             # <<<<<<<<<<<<<<
 *             self.state = 525
 *             self.match(ModelicaParser.T__14)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3943, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3943, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3943, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3943, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3943, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3943, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3943, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3943, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3944
 *             self.state = 524
 *             self.match(ModelicaParser.IDENT)
 *             self.state = 525             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__14)
 *             self.state = 546
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_525) < 0) __PYX_ERR(0, 3944, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":3945
 *             self.match(ModelicaParser.IDENT)
 *             self.state = 525
 *             self.match(ModelicaParser.T__14)             # <<<<<<<<<<<<<<
 *             self.state = 546
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3945, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3945, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__14); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3945, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3945, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3945, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3945, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3945, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4); __pyx_t_4 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3945, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3946
 *             self.state = 525
 *             self.match(ModelicaParser.T__14)
 *             self.state = 546             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_546) < 0) __PYX_ERR(0, 3946, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":3947
 *             self.match(ModelicaParser.T__14)
 *             self.state = 546
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.T__32, ModelicaParser.T__33, ModelicaParser.T__34, ModelicaParser.T__35, ModelicaParser.T__36, ModelicaParser.T__37, ModelicaParser.T__38, ModelicaParser.IDENT]:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3947, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3947, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3947, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3947, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3947, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3947, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3947, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3948
 *             self.state = 546
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if token in [ModelicaParser.T__32, ModelicaParser.T__33, ModelicaParser.T__34, ModelicaParser.T__35, ModelicaParser.T__36, ModelicaParser.T__37, ModelicaParser.T__38, ModelicaParser.IDENT]:
 *                 self.state = 526
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3948, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3948, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__888, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3948, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v_token = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":3949
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.T__32, ModelicaParser.T__33, ModelicaParser.T__34, ModelicaParser.T__35, ModelicaParser.T__36, ModelicaParser.T__37, ModelicaParser.T__38, ModelicaParser.IDENT]:             # <<<<<<<<<<<<<<
 *                 self.state = 526
 *                 self.base_prefix()
 */
        __Pyx_INCREF(__pyx_v_token);
        __pyx_t_1 = __pyx_v_token;
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3949, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__32); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3949, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3949, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3949, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3949, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__33); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3949, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3949, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3949, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3949, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__34); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3949, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3949, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3949, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3949, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__35); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3949, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3949, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3949, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3949, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__36); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3949, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3949, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3949, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3949, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__37); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3949, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3949, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3949, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3949, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__38); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3949, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3949, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3949, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3949, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3949, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3949, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3949, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_11 = __pyx_t_12;
        __pyx_L15_bool_binop_done:;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_12 = (__pyx_t_11 != 0);
        if (__pyx_t_12) {

          /* "pymola/generated/ModelicaParser.pyx":3950
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.T__32, ModelicaParser.T__33, ModelicaParser.T__34, ModelicaParser.T__35, ModelicaParser.T__36, ModelicaParser.T__37, ModelicaParser.T__38, ModelicaParser.IDENT]:
 *                 self.state = 526             # <<<<<<<<<<<<<<
 *                 self.base_prefix()
 *                 self.state = 527
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_526) < 0) __PYX_ERR(0, 3950, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":3951
 *             if token in [ModelicaParser.T__32, ModelicaParser.T__33, ModelicaParser.T__34, ModelicaParser.T__35, ModelicaParser.T__36, ModelicaParser.T__37, ModelicaParser.T__38, ModelicaParser.IDENT]:
 *                 self.state = 526
 *                 self.base_prefix()             # <<<<<<<<<<<<<<
 *                 self.state = 527
 *                 self.component_reference()
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_base_prefix); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3951, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3951, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3951, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3952
 *                 self.state = 526
 *                 self.base_prefix()
 *                 self.state = 527             # <<<<<<<<<<<<<<
 *                 self.component_reference()
 *                 self.state = 529
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_527) < 0) __PYX_ERR(0, 3952, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":3953
 *                 self.base_prefix()
 *                 self.state = 527
 *                 self.component_reference()             # <<<<<<<<<<<<<<
 *                 self.state = 529
 *                 self._errHandler.sync(self)
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_component_reference); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3953, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3953, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3953, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3954
 *                 self.state = 527
 *                 self.component_reference()
 *                 self.state = 529             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_529) < 0) __PYX_ERR(0, 3954, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":3955
 *                 self.component_reference()
 *                 self.state = 529
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__75:
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3955, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3955, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3955, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3955, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3955, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3955, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3955, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3956
 *                 self.state = 529
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if _la==ModelicaParser.T__75:
 *                     self.state = 528
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3956, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3956, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__889, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3956, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_v__la = __pyx_t_1;
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3957
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__75:             # <<<<<<<<<<<<<<
 *                     self.state = 528
 *                     self.array_subscripts()
 */
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3957, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__75); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3957, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3957, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3957, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (__pyx_t_12) {

            /* "pymola/generated/ModelicaParser.pyx":3958
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__75:
 *                     self.state = 528             # <<<<<<<<<<<<<<
 *                     self.array_subscripts()
 * 
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_528) < 0) __PYX_ERR(0, 3958, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":3959
 *                 if _la==ModelicaParser.T__75:
 *                     self.state = 528
 *                     self.array_subscripts()             # <<<<<<<<<<<<<<
 * 
 * 
 */
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_array_subscripts); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3959, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_3 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_3)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_3);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
              }
            }
            if (__pyx_t_3) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3959, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else {
              __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3959, __pyx_L6_error)
            }
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":3957
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__75:             # <<<<<<<<<<<<<<
 *                     self.state = 528
 *                     self.array_subscripts()
 */
          }

          /* "pymola/generated/ModelicaParser.pyx":3962
 * 
 * 
 *                 self.state = 532             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_532) < 0) __PYX_ERR(0, 3962, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":3963
 * 
 *                 self.state = 532
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__16:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3963, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_sync); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3963, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3963, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3963, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3963, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3963, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3963, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3964
 *                 self.state = 532
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if _la==ModelicaParser.T__16:
 *                     self.state = 531
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3964, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3964, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__890, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3964, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3965
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__16:             # <<<<<<<<<<<<<<
 *                     self.state = 531
 *                     self.class_modification()
 */
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3965, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__16); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3965, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3965, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3965, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (__pyx_t_12) {

            /* "pymola/generated/ModelicaParser.pyx":3966
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__16:
 *                     self.state = 531             # <<<<<<<<<<<<<<
 *                     self.class_modification()
 * 
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_531) < 0) __PYX_ERR(0, 3966, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":3967
 *                 if _la==ModelicaParser.T__16:
 *                     self.state = 531
 *                     self.class_modification()             # <<<<<<<<<<<<<<
 * 
 * 
 */
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_class_modification); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3967, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_7 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
              __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
              if (likely(__pyx_t_7)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                __Pyx_INCREF(__pyx_t_7);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_3, function);
              }
            }
            if (__pyx_t_7) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3967, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else {
              __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3967, __pyx_L6_error)
            }
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":3965
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__16:             # <<<<<<<<<<<<<<
 *                     self.state = 531
 *                     self.class_modification()
 */
          }

          /* "pymola/generated/ModelicaParser.pyx":3970
 * 
 * 
 *                 self.state = 534             # <<<<<<<<<<<<<<
 *                 self.comment()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_534) < 0) __PYX_ERR(0, 3970, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":3971
 * 
 *                 self.state = 534
 *                 self.comment()             # <<<<<<<<<<<<<<
 *                 pass
 *             elif token in [ModelicaParser.T__15]:
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_comment); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3971, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3971, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3971, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3949
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.T__32, ModelicaParser.T__33, ModelicaParser.T__34, ModelicaParser.T__35, ModelicaParser.T__36, ModelicaParser.T__37, ModelicaParser.T__38, ModelicaParser.IDENT]:             # <<<<<<<<<<<<<<
 *                 self.state = 526
 *                 self.base_prefix()
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":3973
 *                 self.comment()
 *                 pass
 *             elif token in [ModelicaParser.T__15]:             # <<<<<<<<<<<<<<
 *                 self.state = 536
 *                 self.match(ModelicaParser.T__15)
 */
        __Pyx_INCREF(__pyx_v_token);
        __pyx_t_1 = __pyx_v_token;
        __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3973, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__15); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3973, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3973, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3973, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_11 = (__pyx_t_12 != 0);
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":3974
 *                 pass
 *             elif token in [ModelicaParser.T__15]:
 *                 self.state = 536             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__15)
 *                 self.state = 537
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_536) < 0) __PYX_ERR(0, 3974, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":3975
 *             elif token in [ModelicaParser.T__15]:
 *                 self.state = 536
 *                 self.match(ModelicaParser.T__15)             # <<<<<<<<<<<<<<
 *                 self.state = 537
 *                 self.match(ModelicaParser.T__16)
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3975, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3975, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__15); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3975, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3975, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3975, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3975, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3975, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
              __Pyx_GIVEREF(__pyx_t_2);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_2);
              __pyx_t_2 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3975, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3976
 *                 self.state = 536
 *                 self.match(ModelicaParser.T__15)
 *                 self.state = 537             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__16)
 *                 self.state = 542
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_537) < 0) __PYX_ERR(0, 3976, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":3977
 *                 self.match(ModelicaParser.T__15)
 *                 self.state = 537
 *                 self.match(ModelicaParser.T__16)             # <<<<<<<<<<<<<<
 *                 self.state = 542
 *                 self._errHandler.sync(self)
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3977, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3977, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__16); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3977, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (!__pyx_t_4) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3977, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_2};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3977, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_2};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3977, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            } else
            #endif
            {
              __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3977, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4); __pyx_t_4 = NULL;
              __Pyx_GIVEREF(__pyx_t_2);
              PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_2);
              __pyx_t_2 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3977, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3978
 *                 self.state = 537
 *                 self.match(ModelicaParser.T__16)
 *                 self.state = 542             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 token = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_542) < 0) __PYX_ERR(0, 3978, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":3979
 *                 self.match(ModelicaParser.T__16)
 *                 self.state = 542
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 token = self._input.LA(1)
 *                 if token in [ModelicaParser.T__18, ModelicaParser.IDENT]:
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3979, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3979, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_3) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3979, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3979, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3979, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3979, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3979, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3980
 *                 self.state = 542
 *                 self._errHandler.sync(self)
 *                 token = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if token in [ModelicaParser.T__18, ModelicaParser.IDENT]:
 *                     self.state = 539
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3980, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3980, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__891, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3980, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF_SET(__pyx_v_token, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3981
 *                 self._errHandler.sync(self)
 *                 token = self._input.LA(1)
 *                 if token in [ModelicaParser.T__18, ModelicaParser.IDENT]:             # <<<<<<<<<<<<<<
 *                     self.state = 539
 *                     self._errHandler.sync(self)
 */
          __Pyx_INCREF(__pyx_v_token);
          __pyx_t_1 = __pyx_v_token;
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3981, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__18); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3981, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3981, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3981, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (!__pyx_t_12) {
          } else {
            __pyx_t_11 = __pyx_t_12;
            goto __pyx_L26_bool_binop_done;
          }
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3981, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3981, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3981, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3981, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_11 = __pyx_t_12;
          __pyx_L26_bool_binop_done:;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_12 = (__pyx_t_11 != 0);
          if (__pyx_t_12) {

            /* "pymola/generated/ModelicaParser.pyx":3982
 *                 token = self._input.LA(1)
 *                 if token in [ModelicaParser.T__18, ModelicaParser.IDENT]:
 *                     self.state = 539             # <<<<<<<<<<<<<<
 *                     self._errHandler.sync(self)
 *                     _la = self._input.LA(1)
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_539) < 0) __PYX_ERR(0, 3982, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":3983
 *                 if token in [ModelicaParser.T__18, ModelicaParser.IDENT]:
 *                     self.state = 539
 *                     self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                     _la = self._input.LA(1)
 *                     if _la==ModelicaParser.IDENT:
 */
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3983, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3983, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __pyx_t_7 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_7)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_7);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
              }
            }
            if (!__pyx_t_7) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3983, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_2)) {
                PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3983, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
                PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3983, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else
              #endif
              {
                __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3983, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
                __Pyx_INCREF(__pyx_v_self);
                __Pyx_GIVEREF(__pyx_v_self);
                PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3983, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":3984
 *                     self.state = 539
 *                     self._errHandler.sync(self)
 *                     _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                     if _la==ModelicaParser.IDENT:
 *                         self.state = 538
 */
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3984, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3984, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__892, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3984, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_v__la = __pyx_t_1;
            __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":3985
 *                     self._errHandler.sync(self)
 *                     _la = self._input.LA(1)
 *                     if _la==ModelicaParser.IDENT:             # <<<<<<<<<<<<<<
 *                         self.state = 538
 *                         self.enum_list()
 */
            __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3985, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3985, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3985, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3985, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (__pyx_t_12) {

              /* "pymola/generated/ModelicaParser.pyx":3986
 *                     _la = self._input.LA(1)
 *                     if _la==ModelicaParser.IDENT:
 *                         self.state = 538             # <<<<<<<<<<<<<<
 *                         self.enum_list()
 * 
 */
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_538) < 0) __PYX_ERR(0, 3986, __pyx_L6_error)

              /* "pymola/generated/ModelicaParser.pyx":3987
 *                     if _la==ModelicaParser.IDENT:
 *                         self.state = 538
 *                         self.enum_list()             # <<<<<<<<<<<<<<
 * 
 * 
 */
              __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enum_list); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3987, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __pyx_t_3 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
                __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
                if (likely(__pyx_t_3)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                  __Pyx_INCREF(__pyx_t_3);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_2, function);
                }
              }
              if (__pyx_t_3) {
                __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3987, __pyx_L6_error)
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              } else {
                __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3987, __pyx_L6_error)
              }
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "pymola/generated/ModelicaParser.pyx":3985
 *                     self._errHandler.sync(self)
 *                     _la = self._input.LA(1)
 *                     if _la==ModelicaParser.IDENT:             # <<<<<<<<<<<<<<
 *                         self.state = 538
 *                         self.enum_list()
 */
            }

            /* "pymola/generated/ModelicaParser.pyx":3981
 *                 self._errHandler.sync(self)
 *                 token = self._input.LA(1)
 *                 if token in [ModelicaParser.T__18, ModelicaParser.IDENT]:             # <<<<<<<<<<<<<<
 *                     self.state = 539
 *                     self._errHandler.sync(self)
 */
            goto __pyx_L25;
          }

          /* "pymola/generated/ModelicaParser.pyx":3991
 * 
 *                     pass
 *                 elif token in [ModelicaParser.T__17]:             # <<<<<<<<<<<<<<
 *                     self.state = 541
 *                     self.match(ModelicaParser.T__17)
 */
          __Pyx_INCREF(__pyx_v_token);
          __pyx_t_1 = __pyx_v_token;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3991, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__17); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3991, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3991, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 3991, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_11 = (__pyx_t_12 != 0);
          if (__pyx_t_11) {

            /* "pymola/generated/ModelicaParser.pyx":3992
 *                     pass
 *                 elif token in [ModelicaParser.T__17]:
 *                     self.state = 541             # <<<<<<<<<<<<<<
 *                     self.match(ModelicaParser.T__17)
 *                     pass
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_541) < 0) __PYX_ERR(0, 3992, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":3993
 *                 elif token in [ModelicaParser.T__17]:
 *                     self.state = 541
 *                     self.match(ModelicaParser.T__17)             # <<<<<<<<<<<<<<
 *                     pass
 *                 else:
 */
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3993, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3993, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__17); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3993, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_3)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_3);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
              }
            }
            if (!__pyx_t_3) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3993, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_2)) {
                PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_7};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3993, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
                PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_7};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3993, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              } else
              #endif
              {
                __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3993, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __pyx_t_3 = NULL;
                __Pyx_GIVEREF(__pyx_t_7);
                PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_7);
                __pyx_t_7 = 0;
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3993, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":3991
 * 
 *                     pass
 *                 elif token in [ModelicaParser.T__17]:             # <<<<<<<<<<<<<<
 *                     self.state = 541
 *                     self.match(ModelicaParser.T__17)
 */
            goto __pyx_L25;
          }

          /* "pymola/generated/ModelicaParser.pyx":3996
 *                     pass
 *                 else:
 *                     raise NoViableAltException(self)             # <<<<<<<<<<<<<<
 * 
 *                 self.state = 544
 */
          /*else*/ {
            __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_NoViableAltException); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3996, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_4 = NULL;
            if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_4)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_4);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
              }
            }
            if (!__pyx_t_4) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3996, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_2)) {
                PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3996, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
                PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3996, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else
              #endif
              {
                __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3996, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_7);
                __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4); __pyx_t_4 = NULL;
                __Pyx_INCREF(__pyx_v_self);
                __Pyx_GIVEREF(__pyx_v_self);
                PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self);
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3996, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_Raise(__pyx_t_1, 0, 0, 0);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __PYX_ERR(0, 3996, __pyx_L6_error)
          }
          __pyx_L25:;

          /* "pymola/generated/ModelicaParser.pyx":3998
 *                     raise NoViableAltException(self)
 * 
 *                 self.state = 544             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__18)
 *                 self.state = 545
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_544) < 0) __PYX_ERR(0, 3998, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":3999
 * 
 *                 self.state = 544
 *                 self.match(ModelicaParser.T__18)             # <<<<<<<<<<<<<<
 *                 self.state = 545
 *                 self.comment()
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3999, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3999, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__18); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3999, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3999, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_4};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3999, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_4};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3999, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3999, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
              __Pyx_GIVEREF(__pyx_t_4);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_t_4);
              __pyx_t_4 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3999, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4000
 *                 self.state = 544
 *                 self.match(ModelicaParser.T__18)
 *                 self.state = 545             # <<<<<<<<<<<<<<
 *                 self.comment()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_545) < 0) __PYX_ERR(0, 4000, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":4001
 *                 self.match(ModelicaParser.T__18)
 *                 self.state = 545
 *                 self.comment()             # <<<<<<<<<<<<<<
 *                 pass
 *             else:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_comment); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4001, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (__pyx_t_3) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4001, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4001, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":3973
 *                 self.comment()
 *                 pass
 *             elif token in [ModelicaParser.T__15]:             # <<<<<<<<<<<<<<
 *                 self.state = 536
 *                 self.match(ModelicaParser.T__15)
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":4004
 *                 pass
 *             else:
 *                 raise NoViableAltException(self)             # <<<<<<<<<<<<<<
 * 
 *         except RecognitionException as re:
 */
        /*else*/ {
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_NoViableAltException); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4004, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_3) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4004, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4004, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4004, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4004, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __pyx_t_3 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4004, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_Raise(__pyx_t_1, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __PYX_ERR(0, 4004, __pyx_L6_error)
        }
        __pyx_L14:;

        /* "pymola/generated/ModelicaParser.pyx":3938
 *         self.enterRule(localctx, 74, self.RULE_short_class_definition)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 523
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":4006
 *                 raise NoViableAltException(self)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4006, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.short_class_definition", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_2, &__pyx_t_4) < 0) __PYX_ERR(0, 4006, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_2);
        __pyx_v_re = __pyx_t_2;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":4007
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 4007, __pyx_L34_error)

          /* "pymola/generated/ModelicaParser.pyx":4008
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4008, __pyx_L34_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4008, __pyx_L34_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4008, __pyx_L34_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4008, __pyx_L34_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4008, __pyx_L34_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_7) {
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_7); __pyx_t_7 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_13, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4008, __pyx_L34_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4009
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4009, __pyx_L34_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4009, __pyx_L34_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4009, __pyx_L34_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4009, __pyx_L34_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4009, __pyx_L34_error)
            __Pyx_GOTREF(__pyx_t_7);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4009, __pyx_L34_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":4006
 *                 raise NoViableAltException(self)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L35;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L34_error:;
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_6 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L8_except_error;
          }
          __pyx_L35:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":3938
 *         self.enterRule(localctx, 74, self.RULE_short_class_definition)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 523
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":4011
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4011, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4011, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4011, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_14 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4011, __pyx_L41_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4011, __pyx_L41_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4011, __pyx_L41_error)
        }
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
      __pyx_L41_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":4012
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Equation_sectionContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":3933
 * 
 * 
 *     def short_class_definition(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Short_class_definitionContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.short_class_definition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v_token);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4016
 *     class Equation_sectionContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 4016, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 4016, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4016, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_sectionContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Equation_sectionContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Equation_sectionContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":4017
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 4017, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4017, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4017, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4017, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4017, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4017, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4017, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4017, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4018
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def INITIAL(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 4018, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4016
 *     class Equation_sectionContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_sectionContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4020
 *             self.parser = parser
 * 
 *         def INITIAL(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.INITIAL, 0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_3INITIAL(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_3INITIAL = {"INITIAL", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_3INITIAL, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_3INITIAL(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("INITIAL (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_2INITIAL(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_2INITIAL(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("INITIAL", 0);

  /* "pymola/generated/ModelicaParser.pyx":4021
 * 
 *         def INITIAL(self):
 *             return self.getToken(ModelicaParser.INITIAL, 0)             # <<<<<<<<<<<<<<
 * 
 *         def equation(self, i:int=None):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getToken); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4021, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4021, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_INITIAL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4021, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4021, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4021, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4021, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4021, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":4020
 *             self.parser = parser
 * 
 *         def INITIAL(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.INITIAL, 0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_sectionContext.INITIAL", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4023
 *             return self.getToken(ModelicaParser.INITIAL, 0)
 * 
 *         def equation(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.EquationContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_5equation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_5equation = {"equation", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_5equation, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_5equation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("equation (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "equation") < 0)) __PYX_ERR(0, 4023, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("equation", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4023, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_sectionContext.equation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_4equation(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_4equation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("equation", 0);

  /* "pymola/generated/ModelicaParser.pyx":4024
 * 
 *         def equation(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.EquationContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4025
 *         def equation(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.EquationContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.EquationContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4025, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4025, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_EquationContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4025, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4025, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4025, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4025, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4025, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4025, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":4024
 * 
 *         def equation(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.EquationContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4027
 *                 return self.getTypedRuleContexts(ModelicaParser.EquationContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.EquationContext,i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4027, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4027, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_EquationContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4027, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4027, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4027, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4027, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4027, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":4023
 *             return self.getToken(ModelicaParser.INITIAL, 0)
 * 
 *         def equation(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.EquationContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_sectionContext.equation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4030
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_equation_section
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_7getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_7getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_6getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":4031
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_equation_section             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4031, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_equation_section); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4031, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":4030
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_equation_section
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_sectionContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4033
 *             return ModelicaParser.RULE_equation_section
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterEquation_section" ):
 *                 listener.enterEquation_section(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_9enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_9enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 4033, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 4033, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4033, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_sectionContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_8enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":4034
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterEquation_section" ):             # <<<<<<<<<<<<<<
 *                 listener.enterEquation_section(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterEquation_section); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4034, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4035
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterEquation_section" ):
 *                 listener.enterEquation_section(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterEquation_section); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4035, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4035, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4035, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4035, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4035, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4035, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":4034
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterEquation_section" ):             # <<<<<<<<<<<<<<
 *                 listener.enterEquation_section(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4033
 *             return ModelicaParser.RULE_equation_section
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterEquation_section" ):
 *                 listener.enterEquation_section(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_sectionContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4037
 *                 listener.enterEquation_section(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitEquation_section" ):
 *                 listener.exitEquation_section(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_11exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_11exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 4037, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 4037, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4037, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_sectionContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_10exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":4038
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitEquation_section" ):             # <<<<<<<<<<<<<<
 *                 listener.exitEquation_section(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitEquation_section); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4038, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4039
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitEquation_section" ):
 *                 listener.exitEquation_section(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitEquation_section); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4039, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4039, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4039, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4039, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4039, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4039, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":4038
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitEquation_section" ):             # <<<<<<<<<<<<<<
 *                 listener.exitEquation_section(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4037
 *                 listener.enterEquation_section(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitEquation_section" ):
 *                 listener.exitEquation_section(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_sectionContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4041
 *                 listener.exitEquation_section(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitEquation_section" ):
 *                 return visitor.visitEquation_section(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_13accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_13accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 4041, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 4041, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4041, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_sectionContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_12accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":4042
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitEquation_section" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitEquation_section(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitEquation_section); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4042, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4043
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitEquation_section" ):
 *                 return visitor.visitEquation_section(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitEquation_section); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4043, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4043, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4043, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4043, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4043, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4043, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":4042
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitEquation_section" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitEquation_section(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4045
 *                 return visitor.visitEquation_section(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4045, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4045, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4045, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4045, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4045, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4045, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":4041
 *                 listener.exitEquation_section(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitEquation_section" ):
 *                 return visitor.visitEquation_section(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_sectionContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4050
 * 
 * 
 *     def equation_section(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Equation_sectionContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_79equation_section(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_79equation_section = {"equation_section", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_79equation_section, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_79equation_section(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("equation_section (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_78equation_section(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_78equation_section(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v__alt = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  char const *__pyx_t_22;
  __Pyx_RefNannySetupContext("equation_section", 0);

  /* "pymola/generated/ModelicaParser.pyx":4052
 *     def equation_section(self):
 * 
 *         localctx = ModelicaParser.Equation_sectionContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 76, self.RULE_equation_section)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4052, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Equation_sectionContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4052, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4052, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4052, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4052, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4052, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4052, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4052, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4053
 * 
 *         localctx = ModelicaParser.Equation_sectionContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 76, self.RULE_equation_section)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4053, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_equation_section); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4053, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_76, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4053, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_76, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4053, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4053, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_76);
    __Pyx_GIVEREF(__pyx_int_76);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_76);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4053, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4054
 *         localctx = ModelicaParser.Equation_sectionContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 76, self.RULE_equation_section)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 4054, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4055
 *         self.enterRule(localctx, 76, self.RULE_equation_section)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 549
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":4056
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 549
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4056, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4056, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4056, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4056, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4056, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":4057
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 549             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_549) < 0) __PYX_ERR(0, 4057, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":4058
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 549
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.INITIAL:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4058, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4058, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4058, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4058, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4058, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4058, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4058, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":4059
 *             self.state = 549
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.INITIAL:
 *                 self.state = 548
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4059, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4059, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__893, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4059, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v__la = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":4060
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.INITIAL:             # <<<<<<<<<<<<<<
 *                 self.state = 548
 *                 self.match(ModelicaParser.INITIAL)
 */
        __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4060, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_INITIAL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4060, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4060, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 4060, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":4061
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.INITIAL:
 *                 self.state = 548             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.INITIAL)
 * 
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_548) < 0) __PYX_ERR(0, 4061, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":4062
 *             if _la==ModelicaParser.INITIAL:
 *                 self.state = 548
 *                 self.match(ModelicaParser.INITIAL)             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4062, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4062, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_INITIAL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4062, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4062, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4062, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4062, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4062, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4062, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4060
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.INITIAL:             # <<<<<<<<<<<<<<
 *                 self.state = 548
 *                 self.match(ModelicaParser.INITIAL)
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":4065
 * 
 * 
 *             self.state = 551             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__41)
 *             self.state = 557
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_551) < 0) __PYX_ERR(0, 4065, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":4066
 * 
 *             self.state = 551
 *             self.match(ModelicaParser.T__41)             # <<<<<<<<<<<<<<
 *             self.state = 557
 *             self._errHandler.sync(self)
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4066, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4066, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__41); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4066, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4066, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_3};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4066, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_3};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4066, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4066, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
            __Pyx_GIVEREF(__pyx_t_3);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_3);
            __pyx_t_3 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4066, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":4067
 *             self.state = 551
 *             self.match(ModelicaParser.T__41)
 *             self.state = 557             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _alt = self._interp.adaptivePredict(self._input,77,self._ctx)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_557) < 0) __PYX_ERR(0, 4067, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":4068
 *             self.match(ModelicaParser.T__41)
 *             self.state = 557
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _alt = self._interp.adaptivePredict(self._input,77,self._ctx)
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4068, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4068, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4068, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4068, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4068, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4068, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4068, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":4069
 *             self.state = 557
 *             self._errHandler.sync(self)
 *             _alt = self._interp.adaptivePredict(self._input,77,self._ctx)             # <<<<<<<<<<<<<<
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                 if _alt==1:
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_interp); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4069, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_adaptivePredict); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4069, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4069, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4069, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_4 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_t_2, __pyx_int_77, __pyx_t_7};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4069, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_t_2, __pyx_int_77, __pyx_t_7};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4069, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4069, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_2);
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_t_2);
          __Pyx_INCREF(__pyx_int_77);
          __Pyx_GIVEREF(__pyx_int_77);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_int_77);
          __Pyx_GIVEREF(__pyx_t_7);
          PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_6, __pyx_t_7);
          __pyx_t_2 = 0;
          __pyx_t_7 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4069, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_v__alt = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":4070
 *             self._errHandler.sync(self)
 *             _alt = self._interp.adaptivePredict(self._input,77,self._ctx)
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:             # <<<<<<<<<<<<<<
 *                 if _alt==1:
 *                     self.state = 552
 */
        while (1) {
          __pyx_t_1 = PyObject_RichCompare(__pyx_v__alt, __pyx_int_2, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4070, __pyx_L6_error)
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 4070, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (__pyx_t_12) {
          } else {
            __pyx_t_11 = __pyx_t_12;
            goto __pyx_L17_bool_binop_done;
          }
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ATN); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4070, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_INVALID_ALT_NUMBER); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4070, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyObject_RichCompare(__pyx_v__alt, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4070, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 4070, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_11 = __pyx_t_12;
          __pyx_L17_bool_binop_done:;
          if (!__pyx_t_11) break;

          /* "pymola/generated/ModelicaParser.pyx":4071
 *             _alt = self._interp.adaptivePredict(self._input,77,self._ctx)
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                 if _alt==1:             # <<<<<<<<<<<<<<
 *                     self.state = 552
 *                     self.equation()
 */
          __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v__alt, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4071, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 4071, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (__pyx_t_11) {

            /* "pymola/generated/ModelicaParser.pyx":4072
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                 if _alt==1:
 *                     self.state = 552             # <<<<<<<<<<<<<<
 *                     self.equation()
 *                     self.state = 553
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_552) < 0) __PYX_ERR(0, 4072, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":4073
 *                 if _alt==1:
 *                     self.state = 552
 *                     self.equation()             # <<<<<<<<<<<<<<
 *                     self.state = 553
 *                     self.match(ModelicaParser.T__0)
 */
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_equation); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4073, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_5 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
              if (likely(__pyx_t_5)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                __Pyx_INCREF(__pyx_t_5);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_3, function);
              }
            }
            if (__pyx_t_5) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4073, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else {
              __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4073, __pyx_L6_error)
            }
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":4074
 *                     self.state = 552
 *                     self.equation()
 *                     self.state = 553             # <<<<<<<<<<<<<<
 *                     self.match(ModelicaParser.T__0)
 *                 self.state = 559
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_553) < 0) __PYX_ERR(0, 4074, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":4075
 *                     self.equation()
 *                     self.state = 553
 *                     self.match(ModelicaParser.T__0)             # <<<<<<<<<<<<<<
 *                 self.state = 559
 *                 self._errHandler.sync(self)
 */
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4075, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4075, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_T__0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4075, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_5 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
              if (likely(__pyx_t_5)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                __Pyx_INCREF(__pyx_t_5);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_3, function);
              }
            }
            if (!__pyx_t_5) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4075, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_3)) {
                PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_7};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4075, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
                PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_7};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4075, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              } else
              #endif
              {
                __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4075, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5); __pyx_t_5 = NULL;
                __Pyx_GIVEREF(__pyx_t_7);
                PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_7);
                __pyx_t_7 = 0;
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4075, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":4071
 *             _alt = self._interp.adaptivePredict(self._input,77,self._ctx)
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                 if _alt==1:             # <<<<<<<<<<<<<<
 *                     self.state = 552
 *                     self.equation()
 */
          }

          /* "pymola/generated/ModelicaParser.pyx":4076
 *                     self.state = 553
 *                     self.match(ModelicaParser.T__0)
 *                 self.state = 559             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _alt = self._interp.adaptivePredict(self._input,77,self._ctx)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_559) < 0) __PYX_ERR(0, 4076, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":4077
 *                     self.match(ModelicaParser.T__0)
 *                 self.state = 559
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _alt = self._interp.adaptivePredict(self._input,77,self._ctx)
 * 
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4077, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4077, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_3) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4077, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4077, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4077, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4077, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4077, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4078
 *                 self.state = 559
 *                 self._errHandler.sync(self)
 *                 _alt = self._interp.adaptivePredict(self._input,77,self._ctx)             # <<<<<<<<<<<<<<
 * 
 *         except RecognitionException as re:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_interp); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4078, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_adaptivePredict); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4078, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4078, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4078, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_t_2, __pyx_int_77, __pyx_t_3};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4078, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_t_2, __pyx_int_77, __pyx_t_3};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4078, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4078, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_t_2);
            __Pyx_INCREF(__pyx_int_77);
            __Pyx_GIVEREF(__pyx_int_77);
            PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_int_77);
            __Pyx_GIVEREF(__pyx_t_3);
            PyTuple_SET_ITEM(__pyx_t_4, 2+__pyx_t_6, __pyx_t_3);
            __pyx_t_2 = 0;
            __pyx_t_3 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4078, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF_SET(__pyx_v__alt, __pyx_t_1);
          __pyx_t_1 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":4055
 *         self.enterRule(localctx, 76, self.RULE_equation_section)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 549
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":4080
 *                 _alt = self._interp.adaptivePredict(self._input,77,self._ctx)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4080, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.equation_section", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_7, &__pyx_t_4) < 0) __PYX_ERR(0, 4080, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __pyx_v_re = __pyx_t_7;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":4081
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 4081, __pyx_L25_error)

          /* "pymola/generated/ModelicaParser.pyx":4082
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4082, __pyx_L25_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4082, __pyx_L25_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4082, __pyx_L25_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4082, __pyx_L25_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4082, __pyx_L25_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_2) {
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_2); __pyx_t_2 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_13, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4082, __pyx_L25_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4083
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4083, __pyx_L25_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4083, __pyx_L25_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4083, __pyx_L25_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4083, __pyx_L25_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4083, __pyx_L25_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4083, __pyx_L25_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":4080
 *                 _alt = self._interp.adaptivePredict(self._input,77,self._ctx)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L26;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L25_error:;
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_6 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L8_except_error;
          }
          __pyx_L26:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":4055
 *         self.enterRule(localctx, 76, self.RULE_equation_section)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 549
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":4085
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4085, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4085, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4085, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_14 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4085, __pyx_L32_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4085, __pyx_L32_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4085, __pyx_L32_error)
        }
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
      __pyx_L32_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":4086
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Algorithm_sectionContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":4050
 * 
 * 
 *     def equation_section(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Equation_sectionContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.equation_section", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v__alt);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4090
 *     class Algorithm_sectionContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 4090, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 4090, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4090, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Algorithm_sectionContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":4091
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 4091, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4091, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4091, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4091, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4091, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4091, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4091, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4091, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4092
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def INITIAL(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 4092, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4090
 *     class Algorithm_sectionContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Algorithm_sectionContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4094
 *             self.parser = parser
 * 
 *         def INITIAL(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.INITIAL, 0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_3INITIAL(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_3INITIAL = {"INITIAL", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_3INITIAL, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_3INITIAL(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("INITIAL (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_2INITIAL(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_2INITIAL(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("INITIAL", 0);

  /* "pymola/generated/ModelicaParser.pyx":4095
 * 
 *         def INITIAL(self):
 *             return self.getToken(ModelicaParser.INITIAL, 0)             # <<<<<<<<<<<<<<
 * 
 *         def statement(self, i:int=None):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getToken); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4095, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4095, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_INITIAL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4095, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4095, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4095, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4095, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4095, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":4094
 *             self.parser = parser
 * 
 *         def INITIAL(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.INITIAL, 0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Algorithm_sectionContext.INITIAL", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4097
 *             return self.getToken(ModelicaParser.INITIAL, 0)
 * 
 *         def statement(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.StatementContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_5statement(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_5statement = {"statement", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_5statement, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_5statement(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("statement (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "statement") < 0)) __PYX_ERR(0, 4097, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("statement", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4097, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Algorithm_sectionContext.statement", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_4statement(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_4statement(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("statement", 0);

  /* "pymola/generated/ModelicaParser.pyx":4098
 * 
 *         def statement(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.StatementContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4099
 *         def statement(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.StatementContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.StatementContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4099, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4099, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_StatementContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4099, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4099, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4099, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4099, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4099, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4099, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":4098
 * 
 *         def statement(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.StatementContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4101
 *                 return self.getTypedRuleContexts(ModelicaParser.StatementContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.StatementContext,i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_StatementContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4101, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4101, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4101, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4101, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":4097
 *             return self.getToken(ModelicaParser.INITIAL, 0)
 * 
 *         def statement(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.StatementContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Algorithm_sectionContext.statement", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4104
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_algorithm_section
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_7getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_7getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_6getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":4105
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_algorithm_section             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_algorithm_section); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":4104
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_algorithm_section
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Algorithm_sectionContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4107
 *             return ModelicaParser.RULE_algorithm_section
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterAlgorithm_section" ):
 *                 listener.enterAlgorithm_section(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_9enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_9enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 4107, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 4107, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4107, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Algorithm_sectionContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_8enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":4108
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterAlgorithm_section" ):             # <<<<<<<<<<<<<<
 *                 listener.enterAlgorithm_section(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterAlgorithm_section); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4108, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4109
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterAlgorithm_section" ):
 *                 listener.enterAlgorithm_section(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterAlgorithm_section); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4109, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4109, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4109, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4109, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4109, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4109, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":4108
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterAlgorithm_section" ):             # <<<<<<<<<<<<<<
 *                 listener.enterAlgorithm_section(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4107
 *             return ModelicaParser.RULE_algorithm_section
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterAlgorithm_section" ):
 *                 listener.enterAlgorithm_section(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Algorithm_sectionContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4111
 *                 listener.enterAlgorithm_section(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitAlgorithm_section" ):
 *                 listener.exitAlgorithm_section(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_11exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_11exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 4111, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 4111, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4111, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Algorithm_sectionContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_10exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":4112
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitAlgorithm_section" ):             # <<<<<<<<<<<<<<
 *                 listener.exitAlgorithm_section(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitAlgorithm_section); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4112, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4113
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitAlgorithm_section" ):
 *                 listener.exitAlgorithm_section(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitAlgorithm_section); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4113, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4113, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4113, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4113, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4113, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4113, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":4112
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitAlgorithm_section" ):             # <<<<<<<<<<<<<<
 *                 listener.exitAlgorithm_section(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4111
 *                 listener.enterAlgorithm_section(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitAlgorithm_section" ):
 *                 listener.exitAlgorithm_section(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Algorithm_sectionContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4115
 *                 listener.exitAlgorithm_section(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitAlgorithm_section" ):
 *                 return visitor.visitAlgorithm_section(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_13accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_13accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 4115, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 4115, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4115, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Algorithm_sectionContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_12accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":4116
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitAlgorithm_section" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitAlgorithm_section(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitAlgorithm_section); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4116, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4117
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitAlgorithm_section" ):
 *                 return visitor.visitAlgorithm_section(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitAlgorithm_section); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4117, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4117, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4117, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4117, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4117, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":4116
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitAlgorithm_section" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitAlgorithm_section(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4119
 *                 return visitor.visitAlgorithm_section(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4119, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4119, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4119, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4119, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4119, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":4115
 *                 listener.exitAlgorithm_section(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitAlgorithm_section" ):
 *                 return visitor.visitAlgorithm_section(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Algorithm_sectionContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4124
 * 
 * 
 *     def algorithm_section(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Algorithm_sectionContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_81algorithm_section(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_81algorithm_section = {"algorithm_section", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_81algorithm_section, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_81algorithm_section(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("algorithm_section (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_80algorithm_section(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_80algorithm_section(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  char const *__pyx_t_22;
  __Pyx_RefNannySetupContext("algorithm_section", 0);

  /* "pymola/generated/ModelicaParser.pyx":4126
 *     def algorithm_section(self):
 * 
 *         localctx = ModelicaParser.Algorithm_sectionContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 78, self.RULE_algorithm_section)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Algorithm_sectionContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4126, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4126, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4126, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4126, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4127
 * 
 *         localctx = ModelicaParser.Algorithm_sectionContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 78, self.RULE_algorithm_section)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_algorithm_section); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_78, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4127, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_78, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4127, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_78);
    __Pyx_GIVEREF(__pyx_int_78);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_78);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4128
 *         localctx = ModelicaParser.Algorithm_sectionContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 78, self.RULE_algorithm_section)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 4128, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4129
 *         self.enterRule(localctx, 78, self.RULE_algorithm_section)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 561
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":4130
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 561
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4130, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4130, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4130, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4130, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4130, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":4131
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 561             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_561) < 0) __PYX_ERR(0, 4131, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":4132
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 561
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.INITIAL:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4132, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4132, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4132, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4132, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4132, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4132, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4132, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":4133
 *             self.state = 561
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.INITIAL:
 *                 self.state = 560
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4133, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4133, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__894, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4133, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v__la = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":4134
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.INITIAL:             # <<<<<<<<<<<<<<
 *                 self.state = 560
 *                 self.match(ModelicaParser.INITIAL)
 */
        __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4134, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_INITIAL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4134, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4134, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 4134, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":4135
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.INITIAL:
 *                 self.state = 560             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.INITIAL)
 * 
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_560) < 0) __PYX_ERR(0, 4135, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":4136
 *             if _la==ModelicaParser.INITIAL:
 *                 self.state = 560
 *                 self.match(ModelicaParser.INITIAL)             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4136, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4136, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_INITIAL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4136, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4136, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4136, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4136, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4136, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4136, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4134
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.INITIAL:             # <<<<<<<<<<<<<<
 *                 self.state = 560
 *                 self.match(ModelicaParser.INITIAL)
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":4139
 * 
 * 
 *             self.state = 563             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__42)
 *             self.state = 569
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_563) < 0) __PYX_ERR(0, 4139, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":4140
 * 
 *             self.state = 563
 *             self.match(ModelicaParser.T__42)             # <<<<<<<<<<<<<<
 *             self.state = 569
 *             self._errHandler.sync(self)
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4140, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4140, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__42); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4140, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4140, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_3};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4140, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_3};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4140, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4140, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
            __Pyx_GIVEREF(__pyx_t_3);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_3);
            __pyx_t_3 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4140, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":4141
 *             self.state = 563
 *             self.match(ModelicaParser.T__42)
 *             self.state = 569             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_569) < 0) __PYX_ERR(0, 4141, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":4142
 *             self.match(ModelicaParser.T__42)
 *             self.state = 569
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4142, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4142, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4142, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4142, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4142, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4142, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4142, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":4143
 *             self.state = 569
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:
 *                 self.state = 564
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4143, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4143, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__895, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4143, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":4144
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:             # <<<<<<<<<<<<<<
 *                 self.state = 564
 *                 self.statement()
 */
        while (1) {
          __pyx_t_1 = __Pyx_PyInt_From_long((~0x3f)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4144, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_2 = PyNumber_And(__pyx_v__la, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4144, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_t_2, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4144, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 4144, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (!__pyx_t_12) {
            goto __pyx_L18_next_or;
          } else {
          }
          __pyx_t_1 = PyNumber_Lshift(__pyx_int_1, __pyx_v__la); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4144, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4144, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__16); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4144, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4144, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4144, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__39); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4144, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4144, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Or(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4144, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4144, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__43); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4144, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4144, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Or(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4144, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4144, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__44); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4144, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4144, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Or(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4144, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4144, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__48); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4144, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4144, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Or(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4144, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4144, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__51); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4144, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4144, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Or(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4144, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4144, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__52); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4144, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4144, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Or(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4144, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyNumber_And(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4144, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyObject_RichCompare(__pyx_t_3, __pyx_int_0, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4144, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 4144, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          if (!__pyx_t_12) {
          } else {
            __pyx_t_11 = __pyx_t_12;
            goto __pyx_L17_bool_binop_done;
          }
          __pyx_L18_next_or:;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4144, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4144, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyObject_RichCompare(__pyx_v__la, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4144, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 4144, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_11 = __pyx_t_12;
          __pyx_L17_bool_binop_done:;
          if (!__pyx_t_11) break;

          /* "pymola/generated/ModelicaParser.pyx":4145
 *             _la = self._input.LA(1)
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:
 *                 self.state = 564             # <<<<<<<<<<<<<<
 *                 self.statement()
 *                 self.state = 565
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_564) < 0) __PYX_ERR(0, 4145, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":4146
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:
 *                 self.state = 564
 *                 self.statement()             # <<<<<<<<<<<<<<
 *                 self.state = 565
 *                 self.match(ModelicaParser.T__0)
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_statement); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4146, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (__pyx_t_1) {
            __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4146, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          } else {
            __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4146, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4147
 *                 self.state = 564
 *                 self.statement()
 *                 self.state = 565             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__0)
 *                 self.state = 571
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_565) < 0) __PYX_ERR(0, 4147, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":4148
 *                 self.statement()
 *                 self.state = 565
 *                 self.match(ModelicaParser.T__0)             # <<<<<<<<<<<<<<
 *                 self.state = 571
 *                 self._errHandler.sync(self)
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4148, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4148, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4148, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (!__pyx_t_1) {
            __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4148, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_t_7};
              __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4148, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_t_7};
              __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4148, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4148, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1); __pyx_t_1 = NULL;
              __Pyx_GIVEREF(__pyx_t_7);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_7);
              __pyx_t_7 = 0;
              __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4148, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4149
 *                 self.state = 565
 *                 self.match(ModelicaParser.T__0)
 *                 self.state = 571             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_571) < 0) __PYX_ERR(0, 4149, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":4150
 *                 self.match(ModelicaParser.T__0)
 *                 self.state = 571
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 * 
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4150, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4150, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (!__pyx_t_3) {
            __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4150, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
              __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4150, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_2);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
              __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4150, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_2);
            } else
            #endif
            {
              __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4150, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self);
              __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4150, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4151
 *                 self.state = 571
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *         except RecognitionException as re:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4151, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_LA); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4151, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__896, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4151, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_2);
          __pyx_t_2 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":4129
 *         self.enterRule(localctx, 78, self.RULE_algorithm_section)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 561
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "pymola/generated/ModelicaParser.pyx":4153
 *                 _la = self._input.LA(1)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4153, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.algorithm_section", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_4, &__pyx_t_7) < 0) __PYX_ERR(0, 4153, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_4);
        __pyx_v_re = __pyx_t_4;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":4154
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 4154, __pyx_L25_error)

          /* "pymola/generated/ModelicaParser.pyx":4155
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4155, __pyx_L25_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4155, __pyx_L25_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4155, __pyx_L25_error)
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4155, __pyx_L25_error)
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4155, __pyx_L25_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_1) {
              __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_1); __pyx_t_1 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_13, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4155, __pyx_L25_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4156
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4156, __pyx_L25_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4156, __pyx_L25_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4156, __pyx_L25_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4156, __pyx_L25_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_1 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4156, __pyx_L25_error)
            __Pyx_GOTREF(__pyx_t_1);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_1, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4156, __pyx_L25_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":4153
 *                 _la = self._input.LA(1)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L26;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L25_error:;
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_6 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L8_except_error;
          }
          __pyx_L26:;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":4129
 *         self.enterRule(localctx, 78, self.RULE_algorithm_section)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 561
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":4158
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4158, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (__pyx_t_2) {
        __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4158, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      } else {
        __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4158, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_14 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4158, __pyx_L32_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (__pyx_t_2) {
          __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4158, __pyx_L32_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        } else {
          __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4158, __pyx_L32_error)
        }
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
      __pyx_L32_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":4159
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Equation_optionsContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":4124
 * 
 * 
 *     def algorithm_section(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Algorithm_sectionContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.algorithm_section", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4163
 *     class Equation_optionsContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Equation_optionsContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Equation_optionsContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Equation_optionsContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Equation_optionsContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 4163, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 4163, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4163, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_optionsContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Equation_optionsContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Equation_optionsContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":4164
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 4164, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4164, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4164, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4165
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 4165, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4163
 *     class Equation_optionsContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_optionsContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4168
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_equation_options
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Equation_optionsContext_3getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Equation_optionsContext_3getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Equation_optionsContext_3getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Equation_optionsContext_3getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Equation_optionsContext_2getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Equation_optionsContext_2getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":4169
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_equation_options             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_equation_options); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":4168
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_equation_options
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_optionsContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4172
 * 
 * 
 *         def copyFrom(self, ctx:ParserRuleContext):             # <<<<<<<<<<<<<<
 *             super().copyFrom(ctx)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Equation_optionsContext_5copyFrom(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Equation_optionsContext_5copyFrom = {"copyFrom", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Equation_optionsContext_5copyFrom, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Equation_optionsContext_5copyFrom(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copyFrom (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx_2,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("copyFrom", 1, 2, 2, 1); __PYX_ERR(0, 4172, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "copyFrom") < 0)) __PYX_ERR(0, 4172, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("copyFrom", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4172, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_optionsContext.copyFrom", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Equation_optionsContext_4copyFrom(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Equation_optionsContext_4copyFrom(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("copyFrom", 0);

  /* "pymola/generated/ModelicaParser.pyx":4173
 * 
 *         def copyFrom(self, ctx:ParserRuleContext):
 *             super().copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 4173, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4173, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4173, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4173, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4173, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4173, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4172
 * 
 * 
 *         def copyFrom(self, ctx:ParserRuleContext):             # <<<<<<<<<<<<<<
 *             super().copyFrom(ctx)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_optionsContext.copyFrom", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4179
 *     class Equation_whenContext(Equation_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Equation_whenContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20Equation_whenContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20Equation_whenContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Equation_whenContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 4179, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 4179, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 4179, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4179, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_whenContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20Equation_whenContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Equation_whenContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":4180
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 4180, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4180, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4180, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4180, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4180, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4180, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4181
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def when_equation(self):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4181, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4181, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4181, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4181, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4181, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4179
 *     class Equation_whenContext(Equation_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_whenContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4183
 *             self.copyFrom(ctx)
 * 
 *         def when_equation(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.When_equationContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Equation_whenContext_3when_equation(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20Equation_whenContext_3when_equation = {"when_equation", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20Equation_whenContext_3when_equation, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Equation_whenContext_3when_equation(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("when_equation (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20Equation_whenContext_2when_equation(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Equation_whenContext_2when_equation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("when_equation", 0);

  /* "pymola/generated/ModelicaParser.pyx":4184
 * 
 *         def when_equation(self):
 *             return self.getTypedRuleContext(ModelicaParser.When_equationContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_When_equationContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4184, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4184, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4184, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4184, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":4183
 *             self.copyFrom(ctx)
 * 
 *         def when_equation(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.When_equationContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_whenContext.when_equation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4187
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterEquation_when" ):
 *                 listener.enterEquation_when(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Equation_whenContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20Equation_whenContext_5enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20Equation_whenContext_5enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Equation_whenContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 4187, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 4187, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4187, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_whenContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20Equation_whenContext_4enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Equation_whenContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":4188
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterEquation_when" ):             # <<<<<<<<<<<<<<
 *                 listener.enterEquation_when(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterEquation_when); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4188, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4189
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterEquation_when" ):
 *                 listener.enterEquation_when(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterEquation_when); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4189, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4189, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4189, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4189, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4189, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4189, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":4188
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterEquation_when" ):             # <<<<<<<<<<<<<<
 *                 listener.enterEquation_when(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4187
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterEquation_when" ):
 *                 listener.enterEquation_when(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_whenContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4191
 *                 listener.enterEquation_when(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitEquation_when" ):
 *                 listener.exitEquation_when(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Equation_whenContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20Equation_whenContext_7exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20Equation_whenContext_7exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Equation_whenContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 4191, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 4191, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4191, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_whenContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20Equation_whenContext_6exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Equation_whenContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":4192
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitEquation_when" ):             # <<<<<<<<<<<<<<
 *                 listener.exitEquation_when(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitEquation_when); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4192, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4193
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitEquation_when" ):
 *                 listener.exitEquation_when(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitEquation_when); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4193, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4193, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4193, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4193, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4193, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4193, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":4192
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitEquation_when" ):             # <<<<<<<<<<<<<<
 *                 listener.exitEquation_when(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4191
 *                 listener.enterEquation_when(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitEquation_when" ):
 *                 listener.exitEquation_when(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_whenContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4195
 *                 listener.exitEquation_when(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitEquation_when" ):
 *                 return visitor.visitEquation_when(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Equation_whenContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20Equation_whenContext_9accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20Equation_whenContext_9accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Equation_whenContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 4195, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 4195, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4195, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_whenContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20Equation_whenContext_8accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Equation_whenContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":4196
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitEquation_when" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitEquation_when(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitEquation_when); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4196, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4197
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitEquation_when" ):
 *                 return visitor.visitEquation_when(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitEquation_when); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4197, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4197, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4197, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4197, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4197, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4197, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":4196
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitEquation_when" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitEquation_when(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4199
 *                 return visitor.visitEquation_when(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4199, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4199, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4199, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4199, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4199, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4199, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":4195
 *                 listener.exitEquation_when(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitEquation_when" ):
 *                 return visitor.visitEquation_when(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_whenContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4204
 *     class Equation_connect_clauseContext(Equation_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Equation_connect_clauseContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_30Equation_connect_clauseContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_30Equation_connect_clauseContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Equation_connect_clauseContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 4204, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 4204, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 4204, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4204, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_connect_clauseContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_30Equation_connect_clauseContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Equation_connect_clauseContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":4205
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 4205, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4205, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4205, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4205, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4205, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4205, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4206
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def connect_clause(self):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4206, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4206, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4206, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4206, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4206, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4204
 *     class Equation_connect_clauseContext(Equation_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_connect_clauseContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4208
 *             self.copyFrom(ctx)
 * 
 *         def connect_clause(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Connect_clauseContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Equation_connect_clauseContext_3connect_clause(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_30Equation_connect_clauseContext_3connect_clause = {"connect_clause", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_30Equation_connect_clauseContext_3connect_clause, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Equation_connect_clauseContext_3connect_clause(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("connect_clause (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_30Equation_connect_clauseContext_2connect_clause(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Equation_connect_clauseContext_2connect_clause(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("connect_clause", 0);

  /* "pymola/generated/ModelicaParser.pyx":4209
 * 
 *         def connect_clause(self):
 *             return self.getTypedRuleContext(ModelicaParser.Connect_clauseContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Connect_clauseContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4209, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4209, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4209, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4209, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":4208
 *             self.copyFrom(ctx)
 * 
 *         def connect_clause(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Connect_clauseContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_connect_clauseContext.connect_clause", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4212
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterEquation_connect_clause" ):
 *                 listener.enterEquation_connect_clause(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Equation_connect_clauseContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_30Equation_connect_clauseContext_5enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_30Equation_connect_clauseContext_5enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Equation_connect_clauseContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 4212, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 4212, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4212, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_connect_clauseContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_30Equation_connect_clauseContext_4enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Equation_connect_clauseContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":4213
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterEquation_connect_clause" ):             # <<<<<<<<<<<<<<
 *                 listener.enterEquation_connect_clause(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterEquation_connect_clause); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4213, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4214
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterEquation_connect_clause" ):
 *                 listener.enterEquation_connect_clause(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterEquation_connect_clause); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4214, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4214, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4214, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4214, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4214, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":4213
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterEquation_connect_clause" ):             # <<<<<<<<<<<<<<
 *                 listener.enterEquation_connect_clause(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4212
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterEquation_connect_clause" ):
 *                 listener.enterEquation_connect_clause(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_connect_clauseContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4216
 *                 listener.enterEquation_connect_clause(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitEquation_connect_clause" ):
 *                 listener.exitEquation_connect_clause(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Equation_connect_clauseContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_30Equation_connect_clauseContext_7exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_30Equation_connect_clauseContext_7exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Equation_connect_clauseContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 4216, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 4216, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4216, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_connect_clauseContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_30Equation_connect_clauseContext_6exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Equation_connect_clauseContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":4217
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitEquation_connect_clause" ):             # <<<<<<<<<<<<<<
 *                 listener.exitEquation_connect_clause(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitEquation_connect_clause); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4217, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4218
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitEquation_connect_clause" ):
 *                 listener.exitEquation_connect_clause(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitEquation_connect_clause); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4218, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4218, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4218, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4218, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4218, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4218, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":4217
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitEquation_connect_clause" ):             # <<<<<<<<<<<<<<
 *                 listener.exitEquation_connect_clause(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4216
 *                 listener.enterEquation_connect_clause(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitEquation_connect_clause" ):
 *                 listener.exitEquation_connect_clause(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_connect_clauseContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4220
 *                 listener.exitEquation_connect_clause(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitEquation_connect_clause" ):
 *                 return visitor.visitEquation_connect_clause(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Equation_connect_clauseContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_30Equation_connect_clauseContext_9accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_30Equation_connect_clauseContext_9accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Equation_connect_clauseContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 4220, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 4220, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4220, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_connect_clauseContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_30Equation_connect_clauseContext_8accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Equation_connect_clauseContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":4221
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitEquation_connect_clause" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitEquation_connect_clause(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitEquation_connect_clause); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4221, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4222
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitEquation_connect_clause" ):
 *                 return visitor.visitEquation_connect_clause(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitEquation_connect_clause); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4222, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4222, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4222, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4222, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4222, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4222, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":4221
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitEquation_connect_clause" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitEquation_connect_clause(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4224
 *                 return visitor.visitEquation_connect_clause(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4224, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4224, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4224, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4224, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4224, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4224, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":4220
 *                 listener.exitEquation_connect_clause(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitEquation_connect_clause" ):
 *                 return visitor.visitEquation_connect_clause(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_connect_clauseContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4229
 *     class Equation_forContext(Equation_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Equation_forContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19Equation_forContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19Equation_forContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Equation_forContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 4229, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 4229, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 4229, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4229, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_forContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19Equation_forContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Equation_forContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":4230
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 4230, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4230, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4230, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4230, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4230, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4230, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4231
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def for_equation(self):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4231, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4231, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4231, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4231, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4231, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4229
 *     class Equation_forContext(Equation_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_forContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4233
 *             self.copyFrom(ctx)
 * 
 *         def for_equation(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.For_equationContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Equation_forContext_3for_equation(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19Equation_forContext_3for_equation = {"for_equation", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19Equation_forContext_3for_equation, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Equation_forContext_3for_equation(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("for_equation (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19Equation_forContext_2for_equation(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Equation_forContext_2for_equation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("for_equation", 0);

  /* "pymola/generated/ModelicaParser.pyx":4234
 * 
 *         def for_equation(self):
 *             return self.getTypedRuleContext(ModelicaParser.For_equationContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_For_equationContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4234, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4234, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4234, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4234, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":4233
 *             self.copyFrom(ctx)
 * 
 *         def for_equation(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.For_equationContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_forContext.for_equation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4237
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterEquation_for" ):
 *                 listener.enterEquation_for(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Equation_forContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19Equation_forContext_5enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19Equation_forContext_5enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Equation_forContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 4237, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 4237, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4237, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_forContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19Equation_forContext_4enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Equation_forContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":4238
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterEquation_for" ):             # <<<<<<<<<<<<<<
 *                 listener.enterEquation_for(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterEquation_for); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4238, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4239
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterEquation_for" ):
 *                 listener.enterEquation_for(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterEquation_for); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4239, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4239, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4239, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4239, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4239, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4239, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":4238
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterEquation_for" ):             # <<<<<<<<<<<<<<
 *                 listener.enterEquation_for(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4237
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterEquation_for" ):
 *                 listener.enterEquation_for(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_forContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4241
 *                 listener.enterEquation_for(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitEquation_for" ):
 *                 listener.exitEquation_for(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Equation_forContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19Equation_forContext_7exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19Equation_forContext_7exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Equation_forContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 4241, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 4241, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4241, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_forContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19Equation_forContext_6exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Equation_forContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":4242
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitEquation_for" ):             # <<<<<<<<<<<<<<
 *                 listener.exitEquation_for(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitEquation_for); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4242, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4243
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitEquation_for" ):
 *                 listener.exitEquation_for(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitEquation_for); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4243, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4243, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4243, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4243, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4243, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4243, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":4242
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitEquation_for" ):             # <<<<<<<<<<<<<<
 *                 listener.exitEquation_for(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4241
 *                 listener.enterEquation_for(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitEquation_for" ):
 *                 listener.exitEquation_for(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_forContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4245
 *                 listener.exitEquation_for(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitEquation_for" ):
 *                 return visitor.visitEquation_for(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Equation_forContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19Equation_forContext_9accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19Equation_forContext_9accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Equation_forContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 4245, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 4245, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4245, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_forContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19Equation_forContext_8accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Equation_forContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":4246
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitEquation_for" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitEquation_for(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitEquation_for); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4246, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4247
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitEquation_for" ):
 *                 return visitor.visitEquation_for(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitEquation_for); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4247, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4247, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4247, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4247, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4247, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4247, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":4246
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitEquation_for" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitEquation_for(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4249
 *                 return visitor.visitEquation_for(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4249, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4249, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4249, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4249, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4249, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":4245
 *                 listener.exitEquation_for(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitEquation_for" ):
 *                 return visitor.visitEquation_for(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_forContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4254
 *     class Equation_functionContext(Equation_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Equation_functionContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Equation_functionContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Equation_functionContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Equation_functionContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 4254, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 4254, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 4254, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4254, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_functionContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Equation_functionContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Equation_functionContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":4255
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 4255, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4255, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4255, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4255, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4255, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4255, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4256
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def name(self):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4256, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4256, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4256, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4256, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4256, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4254
 *     class Equation_functionContext(Equation_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_functionContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4258
 *             self.copyFrom(ctx)
 * 
 *         def name(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Equation_functionContext_3name(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Equation_functionContext_3name = {"name", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Equation_functionContext_3name, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Equation_functionContext_3name(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("name (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Equation_functionContext_2name(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Equation_functionContext_2name(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("name", 0);

  /* "pymola/generated/ModelicaParser.pyx":4259
 * 
 *         def name(self):
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)             # <<<<<<<<<<<<<<
 * 
 *         def function_call_args(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_NameContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4259, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4259, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4259, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4259, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":4258
 *             self.copyFrom(ctx)
 * 
 *         def name(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_functionContext.name", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4261
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)
 * 
 *         def function_call_args(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Function_call_argsContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Equation_functionContext_5function_call_args(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Equation_functionContext_5function_call_args = {"function_call_args", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Equation_functionContext_5function_call_args, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Equation_functionContext_5function_call_args(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("function_call_args (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Equation_functionContext_4function_call_args(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Equation_functionContext_4function_call_args(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("function_call_args", 0);

  /* "pymola/generated/ModelicaParser.pyx":4262
 * 
 *         def function_call_args(self):
 *             return self.getTypedRuleContext(ModelicaParser.Function_call_argsContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Function_call_argsContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4262, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4262, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4262, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4262, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":4261
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)
 * 
 *         def function_call_args(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Function_call_argsContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_functionContext.function_call_args", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4265
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterEquation_function" ):
 *                 listener.enterEquation_function(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Equation_functionContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Equation_functionContext_7enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Equation_functionContext_7enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Equation_functionContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 4265, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 4265, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4265, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_functionContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Equation_functionContext_6enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Equation_functionContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":4266
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterEquation_function" ):             # <<<<<<<<<<<<<<
 *                 listener.enterEquation_function(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterEquation_function); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4266, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4267
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterEquation_function" ):
 *                 listener.enterEquation_function(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterEquation_function); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4267, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4267, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4267, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4267, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4267, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4267, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":4266
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterEquation_function" ):             # <<<<<<<<<<<<<<
 *                 listener.enterEquation_function(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4265
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterEquation_function" ):
 *                 listener.enterEquation_function(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_functionContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4269
 *                 listener.enterEquation_function(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitEquation_function" ):
 *                 listener.exitEquation_function(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Equation_functionContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Equation_functionContext_9exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Equation_functionContext_9exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Equation_functionContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 4269, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 4269, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4269, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_functionContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Equation_functionContext_8exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Equation_functionContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":4270
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitEquation_function" ):             # <<<<<<<<<<<<<<
 *                 listener.exitEquation_function(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitEquation_function); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4270, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4271
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitEquation_function" ):
 *                 listener.exitEquation_function(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitEquation_function); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4271, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4271, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4271, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4271, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4271, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4271, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":4270
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitEquation_function" ):             # <<<<<<<<<<<<<<
 *                 listener.exitEquation_function(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4269
 *                 listener.enterEquation_function(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitEquation_function" ):
 *                 listener.exitEquation_function(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_functionContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4273
 *                 listener.exitEquation_function(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitEquation_function" ):
 *                 return visitor.visitEquation_function(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Equation_functionContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Equation_functionContext_11accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Equation_functionContext_11accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Equation_functionContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 4273, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 4273, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4273, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_functionContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Equation_functionContext_10accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Equation_functionContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":4274
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitEquation_function" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitEquation_function(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitEquation_function); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4274, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4275
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitEquation_function" ):
 *                 return visitor.visitEquation_function(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitEquation_function); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4275, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4275, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4275, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4275, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4275, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":4274
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitEquation_function" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitEquation_function(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4277
 *                 return visitor.visitEquation_function(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4277, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4277, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4277, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4277, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4277, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4277, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":4273
 *                 listener.exitEquation_function(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitEquation_function" ):
 *                 return visitor.visitEquation_function(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_functionContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4282
 *     class Equation_simpleContext(Equation_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 4282, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 4282, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 4282, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4282, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_simpleContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Equation_simpleContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Equation_simpleContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":4283
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 4283, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4283, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4283, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4283, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4283, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4283, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4284
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def simple_expression(self):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4284, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4284, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4284, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4284, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4284, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4282
 *     class Equation_simpleContext(Equation_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_simpleContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4286
 *             self.copyFrom(ctx)
 * 
 *         def simple_expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Simple_expressionContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_3simple_expression(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_3simple_expression = {"simple_expression", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_3simple_expression, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_3simple_expression(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("simple_expression (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_2simple_expression(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_2simple_expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("simple_expression", 0);

  /* "pymola/generated/ModelicaParser.pyx":4287
 * 
 *         def simple_expression(self):
 *             return self.getTypedRuleContext(ModelicaParser.Simple_expressionContext,0)             # <<<<<<<<<<<<<<
 * 
 *         def expression(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Simple_expressionContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4287, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4287, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4287, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4287, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":4286
 *             self.copyFrom(ctx)
 * 
 *         def simple_expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Simple_expressionContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_simpleContext.simple_expression", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4289
 *             return self.getTypedRuleContext(ModelicaParser.Simple_expressionContext,0)
 * 
 *         def expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_5expression(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_5expression = {"expression", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_5expression, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_5expression(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("expression (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_4expression(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_4expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("expression", 0);

  /* "pymola/generated/ModelicaParser.pyx":4290
 * 
 *         def expression(self):
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_ExpressionContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4290, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4290, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4290, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4290, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":4289
 *             return self.getTypedRuleContext(ModelicaParser.Simple_expressionContext,0)
 * 
 *         def expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_simpleContext.expression", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4293
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterEquation_simple" ):
 *                 listener.enterEquation_simple(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_7enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_7enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 4293, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 4293, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4293, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_simpleContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_6enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":4294
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterEquation_simple" ):             # <<<<<<<<<<<<<<
 *                 listener.enterEquation_simple(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterEquation_simple); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4294, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4295
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterEquation_simple" ):
 *                 listener.enterEquation_simple(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterEquation_simple); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4295, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4295, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4295, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4295, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4295, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4295, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":4294
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterEquation_simple" ):             # <<<<<<<<<<<<<<
 *                 listener.enterEquation_simple(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4293
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterEquation_simple" ):
 *                 listener.enterEquation_simple(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_simpleContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4297
 *                 listener.enterEquation_simple(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitEquation_simple" ):
 *                 listener.exitEquation_simple(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_9exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_9exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 4297, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 4297, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4297, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_simpleContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_8exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":4298
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitEquation_simple" ):             # <<<<<<<<<<<<<<
 *                 listener.exitEquation_simple(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitEquation_simple); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4298, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4299
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitEquation_simple" ):
 *                 listener.exitEquation_simple(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitEquation_simple); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4299, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4299, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4299, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4299, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4299, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4299, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":4298
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitEquation_simple" ):             # <<<<<<<<<<<<<<
 *                 listener.exitEquation_simple(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4297
 *                 listener.enterEquation_simple(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitEquation_simple" ):
 *                 listener.exitEquation_simple(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_simpleContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4301
 *                 listener.exitEquation_simple(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitEquation_simple" ):
 *                 return visitor.visitEquation_simple(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_11accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_11accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 4301, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 4301, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4301, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_simpleContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_10accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":4302
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitEquation_simple" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitEquation_simple(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitEquation_simple); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4302, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4303
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitEquation_simple" ):
 *                 return visitor.visitEquation_simple(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitEquation_simple); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4303, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4303, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4303, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4303, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4303, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4303, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":4302
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitEquation_simple" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitEquation_simple(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4305
 *                 return visitor.visitEquation_simple(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4305, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4305, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4305, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4305, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4305, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4305, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":4301
 *                 listener.exitEquation_simple(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitEquation_simple" ):
 *                 return visitor.visitEquation_simple(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_simpleContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4310
 *     class Equation_ifContext(Equation_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Equation_ifContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18Equation_ifContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18Equation_ifContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Equation_ifContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 4310, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 4310, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 4310, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4310, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_ifContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18Equation_ifContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Equation_ifContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":4311
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 4311, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4311, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4311, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4311, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4311, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4311, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4312
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def if_equation(self):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4312, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4312, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4312, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4312, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4312, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4312, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4310
 *     class Equation_ifContext(Equation_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_ifContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4314
 *             self.copyFrom(ctx)
 * 
 *         def if_equation(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.If_equationContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Equation_ifContext_3if_equation(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18Equation_ifContext_3if_equation = {"if_equation", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18Equation_ifContext_3if_equation, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Equation_ifContext_3if_equation(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("if_equation (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18Equation_ifContext_2if_equation(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Equation_ifContext_2if_equation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("if_equation", 0);

  /* "pymola/generated/ModelicaParser.pyx":4315
 * 
 *         def if_equation(self):
 *             return self.getTypedRuleContext(ModelicaParser.If_equationContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_If_equationContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4315, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4315, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4315, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4315, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":4314
 *             self.copyFrom(ctx)
 * 
 *         def if_equation(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.If_equationContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_ifContext.if_equation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4318
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterEquation_if" ):
 *                 listener.enterEquation_if(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Equation_ifContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18Equation_ifContext_5enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18Equation_ifContext_5enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Equation_ifContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 4318, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 4318, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4318, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_ifContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18Equation_ifContext_4enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Equation_ifContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":4319
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterEquation_if" ):             # <<<<<<<<<<<<<<
 *                 listener.enterEquation_if(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterEquation_if); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4319, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4320
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterEquation_if" ):
 *                 listener.enterEquation_if(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterEquation_if); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4320, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4320, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4320, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4320, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4320, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4320, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":4319
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterEquation_if" ):             # <<<<<<<<<<<<<<
 *                 listener.enterEquation_if(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4318
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterEquation_if" ):
 *                 listener.enterEquation_if(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_ifContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4322
 *                 listener.enterEquation_if(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitEquation_if" ):
 *                 listener.exitEquation_if(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Equation_ifContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18Equation_ifContext_7exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18Equation_ifContext_7exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Equation_ifContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 4322, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 4322, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4322, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_ifContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18Equation_ifContext_6exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Equation_ifContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":4323
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitEquation_if" ):             # <<<<<<<<<<<<<<
 *                 listener.exitEquation_if(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitEquation_if); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4323, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4324
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitEquation_if" ):
 *                 listener.exitEquation_if(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitEquation_if); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4324, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4324, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4324, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4324, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4324, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4324, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":4323
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitEquation_if" ):             # <<<<<<<<<<<<<<
 *                 listener.exitEquation_if(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4322
 *                 listener.enterEquation_if(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitEquation_if" ):
 *                 listener.exitEquation_if(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_ifContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4326
 *                 listener.exitEquation_if(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitEquation_if" ):
 *                 return visitor.visitEquation_if(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Equation_ifContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18Equation_ifContext_9accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18Equation_ifContext_9accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Equation_ifContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 4326, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 4326, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4326, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_ifContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18Equation_ifContext_8accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Equation_ifContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":4327
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitEquation_if" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitEquation_if(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitEquation_if); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4327, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4328
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitEquation_if" ):
 *                 return visitor.visitEquation_if(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitEquation_if); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4328, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4328, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4328, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4328, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4328, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4328, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":4327
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitEquation_if" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitEquation_if(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4330
 *                 return visitor.visitEquation_if(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4330, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4330, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4330, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4330, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4330, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4330, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":4326
 *                 listener.exitEquation_if(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitEquation_if" ):
 *                 return visitor.visitEquation_if(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Equation_ifContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4334
 * 
 * 
 *     def equation_options(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Equation_optionsContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_83equation_options(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_83equation_options = {"equation_options", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_83equation_options, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_83equation_options(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("equation_options (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_82equation_options(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_82equation_options(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v_la_ = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  char const *__pyx_t_21;
  __Pyx_RefNannySetupContext("equation_options", 0);

  /* "pymola/generated/ModelicaParser.pyx":4336
 *     def equation_options(self):
 * 
 *         localctx = ModelicaParser.Equation_optionsContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 80, self.RULE_equation_options)
 *         try:
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Equation_optionsContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4336, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4336, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4336, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4336, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4337
 * 
 *         localctx = ModelicaParser.Equation_optionsContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 80, self.RULE_equation_options)             # <<<<<<<<<<<<<<
 *         try:
 *             self.state = 583
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_equation_options); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_80, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4337, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_80, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4337, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4337, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_80);
    __Pyx_GIVEREF(__pyx_int_80);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_80);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4337, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4338
 *         localctx = ModelicaParser.Equation_optionsContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 80, self.RULE_equation_options)
 *         try:             # <<<<<<<<<<<<<<
 *             self.state = 583
 *             self._errHandler.sync(self)
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":4339
 *         self.enterRule(localctx, 80, self.RULE_equation_options)
 *         try:
 *             self.state = 583             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             la_ = self._interp.adaptivePredict(self._input,80,self._ctx)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_583) < 0) __PYX_ERR(0, 4339, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":4340
 *         try:
 *             self.state = 583
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             la_ = self._interp.adaptivePredict(self._input,80,self._ctx)
 *             if la_ == 1:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4340, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4340, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4340, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4340, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4340, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4340, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4340, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":4341
 *             self.state = 583
 *             self._errHandler.sync(self)
 *             la_ = self._interp.adaptivePredict(self._input,80,self._ctx)             # <<<<<<<<<<<<<<
 *             if la_ == 1:
 *                 localctx = ModelicaParser.Equation_simpleContext(self, localctx)
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_interp); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4341, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_adaptivePredict); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4341, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4341, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4341, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_t_2, __pyx_int_80, __pyx_t_3};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4341, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_t_2, __pyx_int_80, __pyx_t_3};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4341, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4341, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_2);
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_t_2);
          __Pyx_INCREF(__pyx_int_80);
          __Pyx_GIVEREF(__pyx_int_80);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_int_80);
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_6, __pyx_t_3);
          __pyx_t_2 = 0;
          __pyx_t_3 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4341, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v_la_ = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":4342
 *             self._errHandler.sync(self)
 *             la_ = self._interp.adaptivePredict(self._input,80,self._ctx)
 *             if la_ == 1:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Equation_simpleContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 1)
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4342, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 4342, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":4343
 *             la_ = self._interp.adaptivePredict(self._input,80,self._ctx)
 *             if la_ == 1:
 *                 localctx = ModelicaParser.Equation_simpleContext(self, localctx)             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 1)
 *                 self.state = 572
 */
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4343, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Equation_simpleContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4343, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4343, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4343, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4343, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_7) {
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_v_localctx);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4343, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4344
 *             if la_ == 1:
 *                 localctx = ModelicaParser.Equation_simpleContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *                 self.state = 572
 *                 self.simple_expression()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4344, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_1};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4344, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_1};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4344, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4344, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_1);
            __Pyx_GIVEREF(__pyx_int_1);
            PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4344, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4345
 *                 localctx = ModelicaParser.Equation_simpleContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 1)
 *                 self.state = 572             # <<<<<<<<<<<<<<
 *                 self.simple_expression()
 *                 self.state = 573
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_572) < 0) __PYX_ERR(0, 4345, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":4346
 *                 self.enterOuterAlt(localctx, 1)
 *                 self.state = 572
 *                 self.simple_expression()             # <<<<<<<<<<<<<<
 *                 self.state = 573
 *                 self.match(ModelicaParser.T__14)
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_simple_expression); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4346, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
            }
          }
          if (__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4346, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4346, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4347
 *                 self.state = 572
 *                 self.simple_expression()
 *                 self.state = 573             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__14)
 *                 self.state = 574
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_573) < 0) __PYX_ERR(0, 4347, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":4348
 *                 self.simple_expression()
 *                 self.state = 573
 *                 self.match(ModelicaParser.T__14)             # <<<<<<<<<<<<<<
 *                 self.state = 574
 *                 self.expression()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4348, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4348, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__14); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4348, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4348, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_5)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4348, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4348, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4348, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_7); __pyx_t_7 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4348, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4349
 *                 self.state = 573
 *                 self.match(ModelicaParser.T__14)
 *                 self.state = 574             # <<<<<<<<<<<<<<
 *                 self.expression()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_574) < 0) __PYX_ERR(0, 4349, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":4350
 *                 self.match(ModelicaParser.T__14)
 *                 self.state = 574
 *                 self.expression()             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expression); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4350, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
            }
          }
          if (__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4350, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4350, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4342
 *             self._errHandler.sync(self)
 *             la_ = self._interp.adaptivePredict(self._input,80,self._ctx)
 *             if la_ == 1:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Equation_simpleContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 1)
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":4353
 *                 pass
 * 
 *             elif la_ == 2:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Equation_ifContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 2)
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4353, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 4353, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":4354
 * 
 *             elif la_ == 2:
 *                 localctx = ModelicaParser.Equation_ifContext(self, localctx)             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 2)
 *                 self.state = 576
 */
          __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4354, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_Equation_ifContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4354, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4354, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4354, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4354, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_v_localctx);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4354, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4355
 *             elif la_ == 2:
 *                 localctx = ModelicaParser.Equation_ifContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 2)             # <<<<<<<<<<<<<<
 *                 self.state = 576
 *                 self.if_equation()
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4355, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4355, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4355, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4355, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_2);
            __Pyx_GIVEREF(__pyx_int_2);
            PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_int_2);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4355, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4356
 *                 localctx = ModelicaParser.Equation_ifContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 2)
 *                 self.state = 576             # <<<<<<<<<<<<<<
 *                 self.if_equation()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_576) < 0) __PYX_ERR(0, 4356, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":4357
 *                 self.enterOuterAlt(localctx, 2)
 *                 self.state = 576
 *                 self.if_equation()             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_if_equation); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4357, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (__pyx_t_5) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4357, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4357, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4353
 *                 pass
 * 
 *             elif la_ == 2:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Equation_ifContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 2)
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":4360
 *                 pass
 * 
 *             elif la_ == 3:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Equation_forContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 3)
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_3, 3, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4360, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 4360, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":4361
 * 
 *             elif la_ == 3:
 *                 localctx = ModelicaParser.Equation_forContext(self, localctx)             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 3)
 *                 self.state = 577
 */
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4361, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Equation_forContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4361, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4361, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4361, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4361, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_2) {
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_v_localctx);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4361, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4362
 *             elif la_ == 3:
 *                 localctx = ModelicaParser.Equation_forContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 3)             # <<<<<<<<<<<<<<
 *                 self.state = 577
 *                 self.for_equation()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4362, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_3};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4362, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_3};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4362, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4362, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_3);
            __Pyx_GIVEREF(__pyx_int_3);
            PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_3);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4362, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4363
 *                 localctx = ModelicaParser.Equation_forContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 3)
 *                 self.state = 577             # <<<<<<<<<<<<<<
 *                 self.for_equation()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_577) < 0) __PYX_ERR(0, 4363, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":4364
 *                 self.enterOuterAlt(localctx, 3)
 *                 self.state = 577
 *                 self.for_equation()             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_for_equation); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4364, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
            }
          }
          if (__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4364, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4364, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4360
 *                 pass
 * 
 *             elif la_ == 3:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Equation_forContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 3)
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":4367
 *                 pass
 * 
 *             elif la_ == 4:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Equation_connect_clauseContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 4)
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_4, 4, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4367, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 4367, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":4368
 * 
 *             elif la_ == 4:
 *                 localctx = ModelicaParser.Equation_connect_clauseContext(self, localctx)             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 4)
 *                 self.state = 578
 */
          __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4368, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_Equation_connect_clauseContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4368, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4368, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4368, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4368, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_v_localctx);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4368, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4369
 *             elif la_ == 4:
 *                 localctx = ModelicaParser.Equation_connect_clauseContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 4)             # <<<<<<<<<<<<<<
 *                 self.state = 578
 *                 self.connect_clause()
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4369, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_4};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4369, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_4};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4369, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4369, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_4);
            __Pyx_GIVEREF(__pyx_int_4);
            PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_int_4);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4369, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4370
 *                 localctx = ModelicaParser.Equation_connect_clauseContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 4)
 *                 self.state = 578             # <<<<<<<<<<<<<<
 *                 self.connect_clause()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_578) < 0) __PYX_ERR(0, 4370, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":4371
 *                 self.enterOuterAlt(localctx, 4)
 *                 self.state = 578
 *                 self.connect_clause()             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_connect_clause); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4371, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (__pyx_t_5) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4371, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4371, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4367
 *                 pass
 * 
 *             elif la_ == 4:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Equation_connect_clauseContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 4)
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":4374
 *                 pass
 * 
 *             elif la_ == 5:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Equation_whenContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 5)
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_5, 5, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4374, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 4374, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":4375
 * 
 *             elif la_ == 5:
 *                 localctx = ModelicaParser.Equation_whenContext(self, localctx)             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 5)
 *                 self.state = 579
 */
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4375, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Equation_whenContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4375, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4375, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4375, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4375, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_2) {
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_v_localctx);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4375, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4376
 *             elif la_ == 5:
 *                 localctx = ModelicaParser.Equation_whenContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 5)             # <<<<<<<<<<<<<<
 *                 self.state = 579
 *                 self.when_equation()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4376, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_5};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4376, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_5};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4376, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4376, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_5);
            __Pyx_GIVEREF(__pyx_int_5);
            PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_5);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4376, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4377
 *                 localctx = ModelicaParser.Equation_whenContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 5)
 *                 self.state = 579             # <<<<<<<<<<<<<<
 *                 self.when_equation()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_579) < 0) __PYX_ERR(0, 4377, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":4378
 *                 self.enterOuterAlt(localctx, 5)
 *                 self.state = 579
 *                 self.when_equation()             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_when_equation); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4378, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
            }
          }
          if (__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4378, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4378, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4374
 *                 pass
 * 
 *             elif la_ == 5:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Equation_whenContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 5)
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":4381
 *                 pass
 * 
 *             elif la_ == 6:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Equation_functionContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 6)
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_6, 6, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4381, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 4381, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":4382
 * 
 *             elif la_ == 6:
 *                 localctx = ModelicaParser.Equation_functionContext(self, localctx)             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 6)
 *                 self.state = 580
 */
          __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4382, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_Equation_functionContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4382, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4382, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4382, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4382, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_v_localctx);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4382, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4383
 *             elif la_ == 6:
 *                 localctx = ModelicaParser.Equation_functionContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 6)             # <<<<<<<<<<<<<<
 *                 self.state = 580
 *                 self.name()
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4383, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_6};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4383, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_6};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4383, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4383, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_6);
            __Pyx_GIVEREF(__pyx_int_6);
            PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_int_6);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4383, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4384
 *                 localctx = ModelicaParser.Equation_functionContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 6)
 *                 self.state = 580             # <<<<<<<<<<<<<<
 *                 self.name()
 *                 self.state = 581
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_580) < 0) __PYX_ERR(0, 4384, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":4385
 *                 self.enterOuterAlt(localctx, 6)
 *                 self.state = 580
 *                 self.name()             # <<<<<<<<<<<<<<
 *                 self.state = 581
 *                 self.function_call_args()
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4385, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (__pyx_t_5) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4385, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4385, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4386
 *                 self.state = 580
 *                 self.name()
 *                 self.state = 581             # <<<<<<<<<<<<<<
 *                 self.function_call_args()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_581) < 0) __PYX_ERR(0, 4386, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":4387
 *                 self.name()
 *                 self.state = 581
 *                 self.function_call_args()             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_function_call_args); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4387, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (__pyx_t_5) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4387, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4387, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4381
 *                 pass
 * 
 *             elif la_ == 6:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Equation_functionContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 6)
 */
        }
        __pyx_L14:;

        /* "pymola/generated/ModelicaParser.pyx":4338
 *         localctx = ModelicaParser.Equation_optionsContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 80, self.RULE_equation_options)
 *         try:             # <<<<<<<<<<<<<<
 *             self.state = 583
 *             self._errHandler.sync(self)
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":4391
 * 
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4391, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.equation_options", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_2, &__pyx_t_5) < 0) __PYX_ERR(0, 4391, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_2);
        __pyx_v_re = __pyx_t_2;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":4392
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 4392, __pyx_L20_error)

          /* "pymola/generated/ModelicaParser.pyx":4393
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4393, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_reportError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4393, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4393, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4393, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_12 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 4393, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_12);
            if (__pyx_t_7) {
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_7); __pyx_t_7 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_12, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4393, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4394
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4394, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_recover); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 4394, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_12);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_12, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4394, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4394, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4394, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_7);
            if (__pyx_t_4) {
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4); __pyx_t_4 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4394, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":4391
 * 
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L21;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L20_error:;
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __pyx_t_6 = __pyx_lineno; __pyx_t_13 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
            }
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_ErrRestore(__pyx_t_15, __pyx_t_16, __pyx_t_17);
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_13; __pyx_filename = __pyx_t_14;
            goto __pyx_L8_except_error;
          }
          __pyx_L21:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":4338
 *         localctx = ModelicaParser.Equation_optionsContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 80, self.RULE_equation_options)
 *         try:             # <<<<<<<<<<<<<<
 *             self.state = 583
 *             self._errHandler.sync(self)
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":4396
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4396, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4396, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_5 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4396, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_19, &__pyx_t_18);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_18);
      __pyx_t_13 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_21 = __pyx_filename;
      {
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4396, __pyx_L27_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4396, __pyx_L27_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_5 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4396, __pyx_L27_error)
        }
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __pyx_lineno = __pyx_t_13; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_21;
      goto __pyx_L1_error;
      __pyx_L27_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":4397
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class EquationContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":4334
 * 
 * 
 *     def equation_options(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Equation_optionsContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.equation_options", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v_la_);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4401
 *     class EquationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15EquationContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15EquationContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15EquationContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15EquationContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 4401, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 4401, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4401, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.EquationContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15EquationContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15EquationContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":4402
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 4402, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4402, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4402, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4402, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4402, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4402, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4402, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4402, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4403
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def equation_options(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 4403, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4401
 *     class EquationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.EquationContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4405
 *             self.parser = parser
 * 
 *         def equation_options(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Equation_optionsContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15EquationContext_3equation_options(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15EquationContext_3equation_options = {"equation_options", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15EquationContext_3equation_options, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15EquationContext_3equation_options(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("equation_options (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15EquationContext_2equation_options(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15EquationContext_2equation_options(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("equation_options", 0);

  /* "pymola/generated/ModelicaParser.pyx":4406
 * 
 *         def equation_options(self):
 *             return self.getTypedRuleContext(ModelicaParser.Equation_optionsContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Equation_optionsContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4406, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4406, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4406, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4406, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":4405
 *             self.parser = parser
 * 
 *         def equation_options(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Equation_optionsContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.EquationContext.equation_options", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4409
 * 
 * 
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15EquationContext_5comment(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15EquationContext_5comment = {"comment", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15EquationContext_5comment, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15EquationContext_5comment(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("comment (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15EquationContext_4comment(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15EquationContext_4comment(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("comment", 0);

  /* "pymola/generated/ModelicaParser.pyx":4410
 * 
 *         def comment(self):
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_CommentContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4410, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4410, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":4409
 * 
 * 
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.EquationContext.comment", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4413
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_equation
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15EquationContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15EquationContext_7getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15EquationContext_7getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15EquationContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15EquationContext_6getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15EquationContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":4414
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_equation             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4414, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_equation); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4414, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":4413
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_equation
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.EquationContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4416
 *             return ModelicaParser.RULE_equation
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterEquation" ):
 *                 listener.enterEquation(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15EquationContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15EquationContext_9enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15EquationContext_9enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15EquationContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 4416, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 4416, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4416, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.EquationContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15EquationContext_8enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15EquationContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":4417
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterEquation" ):             # <<<<<<<<<<<<<<
 *                 listener.enterEquation(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterEquation); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4417, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4418
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterEquation" ):
 *                 listener.enterEquation(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterEquation); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4418, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4418, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4418, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4418, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4418, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":4417
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterEquation" ):             # <<<<<<<<<<<<<<
 *                 listener.enterEquation(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4416
 *             return ModelicaParser.RULE_equation
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterEquation" ):
 *                 listener.enterEquation(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.EquationContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4420
 *                 listener.enterEquation(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitEquation" ):
 *                 listener.exitEquation(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15EquationContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15EquationContext_11exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15EquationContext_11exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15EquationContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 4420, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 4420, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4420, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.EquationContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15EquationContext_10exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15EquationContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":4421
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitEquation" ):             # <<<<<<<<<<<<<<
 *                 listener.exitEquation(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitEquation); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4421, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4422
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitEquation" ):
 *                 listener.exitEquation(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitEquation); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4422, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4422, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4422, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4422, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4422, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4422, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":4421
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitEquation" ):             # <<<<<<<<<<<<<<
 *                 listener.exitEquation(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4420
 *                 listener.enterEquation(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitEquation" ):
 *                 listener.exitEquation(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.EquationContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4424
 *                 listener.exitEquation(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitEquation" ):
 *                 return visitor.visitEquation(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15EquationContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15EquationContext_13accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15EquationContext_13accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15EquationContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 4424, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 4424, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4424, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.EquationContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15EquationContext_12accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15EquationContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":4425
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitEquation" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitEquation(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitEquation); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4425, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4426
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitEquation" ):
 *                 return visitor.visitEquation(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitEquation); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4426, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4426, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4426, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4426, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4426, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4426, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":4425
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitEquation" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitEquation(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4428
 *                 return visitor.visitEquation(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4428, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4428, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4428, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4428, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4428, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4428, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":4424
 *                 listener.exitEquation(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitEquation" ):
 *                 return visitor.visitEquation(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.EquationContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4433
 * 
 * 
 *     def equation(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.EquationContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_85equation(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_85equation = {"equation", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_85equation, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_85equation(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("equation (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_84equation(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_84equation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  char const *__pyx_t_20;
  __Pyx_RefNannySetupContext("equation", 0);

  /* "pymola/generated/ModelicaParser.pyx":4435
 *     def equation(self):
 * 
 *         localctx = ModelicaParser.EquationContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 82, self.RULE_equation)
 *         try:
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_EquationContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4435, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4435, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4435, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4435, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4436
 * 
 *         localctx = ModelicaParser.EquationContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 82, self.RULE_equation)             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_equation); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_82, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4436, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_82, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4436, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4436, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_82);
    __Pyx_GIVEREF(__pyx_int_82);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_82);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4436, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4437
 *         localctx = ModelicaParser.EquationContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 82, self.RULE_equation)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 585
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":4438
 *         self.enterRule(localctx, 82, self.RULE_equation)
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 585
 *             self.equation_options()
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4438, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4438, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4438, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4438, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4438, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":4439
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 585             # <<<<<<<<<<<<<<
 *             self.equation_options()
 *             self.state = 586
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_585) < 0) __PYX_ERR(0, 4439, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":4440
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 585
 *             self.equation_options()             # <<<<<<<<<<<<<<
 *             self.state = 586
 *             self.comment()
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_equation_options); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4440, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4440, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4440, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":4441
 *             self.state = 585
 *             self.equation_options()
 *             self.state = 586             # <<<<<<<<<<<<<<
 *             self.comment()
 *         except RecognitionException as re:
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_586) < 0) __PYX_ERR(0, 4441, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":4442
 *             self.equation_options()
 *             self.state = 586
 *             self.comment()             # <<<<<<<<<<<<<<
 *         except RecognitionException as re:
 *             localctx.exception = re
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_comment); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4442, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4442, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4442, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":4437
 *         localctx = ModelicaParser.EquationContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 82, self.RULE_equation)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 585
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":4443
 *             self.state = 586
 *             self.comment()
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4443, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.equation", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_3, &__pyx_t_7) < 0) __PYX_ERR(0, 4443, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_3);
        __pyx_v_re = __pyx_t_3;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":4444
 *             self.comment()
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 4444, __pyx_L19_error)

          /* "pymola/generated/ModelicaParser.pyx":4445
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4445, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4445, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4445, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4445, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_11 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 4445, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_11);
            if (__pyx_t_4) {
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_4); __pyx_t_4 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_11, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4445, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4446
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4446, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 4446, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_11)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4446, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_11)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4446, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4446, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4446, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":4443
 *             self.state = 586
 *             self.comment()
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L20;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L19_error:;
            __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_14);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __pyx_t_6 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_17);
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
            }
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
            __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
            goto __pyx_L8_except_error;
          }
          __pyx_L20:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":4437
 *         localctx = ModelicaParser.EquationContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 82, self.RULE_equation)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 585
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":4448
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4448, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4448, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4448, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_18, &__pyx_t_17);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_18);
      __Pyx_XGOTREF(__pyx_t_17);
      __pyx_t_12 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_20 = __pyx_filename;
      {
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4448, __pyx_L26_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4448, __pyx_L26_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4448, __pyx_L26_error)
        }
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_18, __pyx_t_17);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0;
      __pyx_lineno = __pyx_t_12; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_20;
      goto __pyx_L1_error;
      __pyx_L26_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_18, __pyx_t_17);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":4449
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Statement_optionsContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":4433
 * 
 * 
 *     def equation(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.EquationContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.equation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4453
 *     class Statement_optionsContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Statement_optionsContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Statement_optionsContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Statement_optionsContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Statement_optionsContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 4453, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 4453, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4453, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_optionsContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Statement_optionsContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Statement_optionsContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":4454
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 4454, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4454, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4454, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4454, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4454, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4454, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4454, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4454, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4455
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 4455, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4453
 *     class Statement_optionsContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_optionsContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4458
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_statement_options
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Statement_optionsContext_3getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Statement_optionsContext_3getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Statement_optionsContext_3getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Statement_optionsContext_3getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Statement_optionsContext_2getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Statement_optionsContext_2getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":4459
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_statement_options             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4459, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_statement_options); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4459, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":4458
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_statement_options
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_optionsContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4462
 * 
 * 
 *         def copyFrom(self, ctx:ParserRuleContext):             # <<<<<<<<<<<<<<
 *             super().copyFrom(ctx)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Statement_optionsContext_5copyFrom(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Statement_optionsContext_5copyFrom = {"copyFrom", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Statement_optionsContext_5copyFrom, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Statement_optionsContext_5copyFrom(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copyFrom (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx_2,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("copyFrom", 1, 2, 2, 1); __PYX_ERR(0, 4462, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "copyFrom") < 0)) __PYX_ERR(0, 4462, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("copyFrom", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4462, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_optionsContext.copyFrom", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Statement_optionsContext_4copyFrom(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Statement_optionsContext_4copyFrom(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("copyFrom", 0);

  /* "pymola/generated/ModelicaParser.pyx":4463
 * 
 *         def copyFrom(self, ctx:ParserRuleContext):
 *             super().copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 4463, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4463, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4463, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4463, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4463, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4463, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4463, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4463, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4463, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4462
 * 
 * 
 *         def copyFrom(self, ctx:ParserRuleContext):             # <<<<<<<<<<<<<<
 *             super().copyFrom(ctx)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_optionsContext.copyFrom", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4469
 *     class Statement_breakContext(Statement_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Statement_breakContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Statement_breakContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Statement_breakContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Statement_breakContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 4469, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 4469, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 4469, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4469, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_breakContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Statement_breakContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Statement_breakContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":4470
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 4470, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4470, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4470, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4470, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4470, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4470, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4470, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4470, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4470, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4471
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4471, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4471, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4471, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4471, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4471, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4471, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4469
 *     class Statement_breakContext(Statement_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_breakContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4474
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStatement_break" ):
 *                 listener.enterStatement_break(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Statement_breakContext_3enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Statement_breakContext_3enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Statement_breakContext_3enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Statement_breakContext_3enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 4474, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 4474, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4474, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_breakContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Statement_breakContext_2enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Statement_breakContext_2enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":4475
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterStatement_break" ):             # <<<<<<<<<<<<<<
 *                 listener.enterStatement_break(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterStatement_break); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4475, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4476
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterStatement_break" ):
 *                 listener.enterStatement_break(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterStatement_break); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4476, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4476, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4476, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4476, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4476, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4476, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":4475
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterStatement_break" ):             # <<<<<<<<<<<<<<
 *                 listener.enterStatement_break(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4474
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStatement_break" ):
 *                 listener.enterStatement_break(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_breakContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4478
 *                 listener.enterStatement_break(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStatement_break" ):
 *                 listener.exitStatement_break(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Statement_breakContext_5exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Statement_breakContext_5exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Statement_breakContext_5exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Statement_breakContext_5exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 4478, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 4478, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4478, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_breakContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Statement_breakContext_4exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Statement_breakContext_4exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":4479
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitStatement_break" ):             # <<<<<<<<<<<<<<
 *                 listener.exitStatement_break(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitStatement_break); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4479, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4480
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitStatement_break" ):
 *                 listener.exitStatement_break(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitStatement_break); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4480, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4480, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4480, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4480, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4480, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4480, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":4479
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitStatement_break" ):             # <<<<<<<<<<<<<<
 *                 listener.exitStatement_break(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4478
 *                 listener.enterStatement_break(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStatement_break" ):
 *                 listener.exitStatement_break(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_breakContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4482
 *                 listener.exitStatement_break(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStatement_break" ):
 *                 return visitor.visitStatement_break(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Statement_breakContext_7accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Statement_breakContext_7accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Statement_breakContext_7accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Statement_breakContext_7accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 4482, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 4482, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4482, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_breakContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Statement_breakContext_6accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Statement_breakContext_6accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":4483
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitStatement_break" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitStatement_break(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitStatement_break); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4483, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4484
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitStatement_break" ):
 *                 return visitor.visitStatement_break(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitStatement_break); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4484, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4484, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4484, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4484, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4484, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4484, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":4483
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitStatement_break" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitStatement_break(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4486
 *                 return visitor.visitStatement_break(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4486, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4486, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4486, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4486, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4486, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4486, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":4482
 *                 listener.exitStatement_break(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStatement_break" ):
 *                 return visitor.visitStatement_break(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_breakContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4491
 *     class Statement_whileContext(Statement_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Statement_whileContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Statement_whileContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Statement_whileContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Statement_whileContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 4491, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 4491, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 4491, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4491, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_whileContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Statement_whileContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Statement_whileContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":4492
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 4492, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4492, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4492, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4492, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4492, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4492, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4492, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4492, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4492, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4493
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def while_statement(self):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4493, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4493, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4493, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4493, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4493, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4491
 *     class Statement_whileContext(Statement_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_whileContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4495
 *             self.copyFrom(ctx)
 * 
 *         def while_statement(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.While_statementContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Statement_whileContext_3while_statement(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Statement_whileContext_3while_statement = {"while_statement", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Statement_whileContext_3while_statement, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Statement_whileContext_3while_statement(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("while_statement (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Statement_whileContext_2while_statement(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Statement_whileContext_2while_statement(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("while_statement", 0);

  /* "pymola/generated/ModelicaParser.pyx":4496
 * 
 *         def while_statement(self):
 *             return self.getTypedRuleContext(ModelicaParser.While_statementContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_While_statementContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4496, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4496, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4496, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4496, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":4495
 *             self.copyFrom(ctx)
 * 
 *         def while_statement(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.While_statementContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_whileContext.while_statement", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4499
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStatement_while" ):
 *                 listener.enterStatement_while(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Statement_whileContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Statement_whileContext_5enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Statement_whileContext_5enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Statement_whileContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 4499, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 4499, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4499, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_whileContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Statement_whileContext_4enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Statement_whileContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":4500
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterStatement_while" ):             # <<<<<<<<<<<<<<
 *                 listener.enterStatement_while(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterStatement_while); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4500, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4501
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterStatement_while" ):
 *                 listener.enterStatement_while(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterStatement_while); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4501, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4501, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4501, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4501, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4501, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4501, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":4500
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterStatement_while" ):             # <<<<<<<<<<<<<<
 *                 listener.enterStatement_while(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4499
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStatement_while" ):
 *                 listener.enterStatement_while(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_whileContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4503
 *                 listener.enterStatement_while(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStatement_while" ):
 *                 listener.exitStatement_while(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Statement_whileContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Statement_whileContext_7exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Statement_whileContext_7exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Statement_whileContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 4503, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 4503, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4503, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_whileContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Statement_whileContext_6exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Statement_whileContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":4504
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitStatement_while" ):             # <<<<<<<<<<<<<<
 *                 listener.exitStatement_while(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitStatement_while); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4504, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4505
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitStatement_while" ):
 *                 listener.exitStatement_while(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitStatement_while); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4505, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4505, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4505, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4505, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4505, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4505, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":4504
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitStatement_while" ):             # <<<<<<<<<<<<<<
 *                 listener.exitStatement_while(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4503
 *                 listener.enterStatement_while(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStatement_while" ):
 *                 listener.exitStatement_while(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_whileContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4507
 *                 listener.exitStatement_while(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStatement_while" ):
 *                 return visitor.visitStatement_while(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Statement_whileContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Statement_whileContext_9accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Statement_whileContext_9accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Statement_whileContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 4507, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 4507, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4507, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_whileContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Statement_whileContext_8accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Statement_whileContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":4508
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitStatement_while" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitStatement_while(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitStatement_while); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4508, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4509
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitStatement_while" ):
 *                 return visitor.visitStatement_while(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitStatement_while); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4509, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4509, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4509, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4509, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4509, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4509, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":4508
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitStatement_while" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitStatement_while(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4511
 *                 return visitor.visitStatement_while(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4511, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4511, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4511, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4511, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4511, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4511, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":4507
 *                 listener.exitStatement_while(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStatement_while" ):
 *                 return visitor.visitStatement_while(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_whileContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4516
 *     class Statement_component_functionContext(Statement_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 4516, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 4516, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 4516, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4516, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_component_functionContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":4517
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 4517, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4517, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4517, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4517, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4517, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4517, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4518
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def component_reference(self, i:int=None):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4518, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4518, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4518, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4518, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4518, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4516
 *     class Statement_component_functionContext(Statement_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_component_functionContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4520
 *             self.copyFrom(ctx)
 * 
 *         def component_reference(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Component_referenceContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_3component_reference(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_3component_reference = {"component_reference", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_3component_reference, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_3component_reference(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("component_reference (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "component_reference") < 0)) __PYX_ERR(0, 4520, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("component_reference", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4520, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_component_functionContext.component_reference", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_2component_reference(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_2component_reference(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("component_reference", 0);

  /* "pymola/generated/ModelicaParser.pyx":4521
 * 
 *         def component_reference(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.Component_referenceContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4522
 *         def component_reference(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Component_referenceContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4522, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4522, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_Component_referenceContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4522, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4522, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4522, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4522, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4522, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4522, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":4521
 * 
 *         def component_reference(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.Component_referenceContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4524
 *                 return self.getTypedRuleContexts(ModelicaParser.Component_referenceContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,i)             # <<<<<<<<<<<<<<
 * 
 *         def function_call_args(self):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4524, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4524, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Component_referenceContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4524, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4524, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4524, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4524, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4524, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":4520
 *             self.copyFrom(ctx)
 * 
 *         def component_reference(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Component_referenceContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_component_functionContext.component_reference", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4526
 *                 return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,i)
 * 
 *         def function_call_args(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Function_call_argsContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_5function_call_args(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_5function_call_args = {"function_call_args", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_5function_call_args, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_5function_call_args(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("function_call_args (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_4function_call_args(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_4function_call_args(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("function_call_args", 0);

  /* "pymola/generated/ModelicaParser.pyx":4527
 * 
 *         def function_call_args(self):
 *             return self.getTypedRuleContext(ModelicaParser.Function_call_argsContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4527, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4527, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Function_call_argsContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4527, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4527, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4527, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4527, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4527, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":4526
 *                 return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,i)
 * 
 *         def function_call_args(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Function_call_argsContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_component_functionContext.function_call_args", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4530
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStatement_component_function" ):
 *                 listener.enterStatement_component_function(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_7enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_7enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 4530, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 4530, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4530, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_component_functionContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_6enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":4531
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterStatement_component_function" ):             # <<<<<<<<<<<<<<
 *                 listener.enterStatement_component_function(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterStatement_component_functio); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4531, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4532
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterStatement_component_function" ):
 *                 listener.enterStatement_component_function(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterStatement_component_functio); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4532, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4532, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4532, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4532, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4532, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4532, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":4531
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterStatement_component_function" ):             # <<<<<<<<<<<<<<
 *                 listener.enterStatement_component_function(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4530
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStatement_component_function" ):
 *                 listener.enterStatement_component_function(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_component_functionContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4534
 *                 listener.enterStatement_component_function(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStatement_component_function" ):
 *                 listener.exitStatement_component_function(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_9exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_9exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 4534, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 4534, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4534, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_component_functionContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_8exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":4535
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitStatement_component_function" ):             # <<<<<<<<<<<<<<
 *                 listener.exitStatement_component_function(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitStatement_component_function); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4535, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4536
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitStatement_component_function" ):
 *                 listener.exitStatement_component_function(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitStatement_component_function); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4536, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4536, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4536, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4536, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4536, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4536, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":4535
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitStatement_component_function" ):             # <<<<<<<<<<<<<<
 *                 listener.exitStatement_component_function(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4534
 *                 listener.enterStatement_component_function(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStatement_component_function" ):
 *                 listener.exitStatement_component_function(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_component_functionContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4538
 *                 listener.exitStatement_component_function(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStatement_component_function" ):
 *                 return visitor.visitStatement_component_function(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_11accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_11accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 4538, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 4538, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4538, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_component_functionContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_10accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":4539
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitStatement_component_function" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitStatement_component_function(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitStatement_component_functio); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4539, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4540
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitStatement_component_function" ):
 *                 return visitor.visitStatement_component_function(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitStatement_component_functio); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4540, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4540, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4540, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4540, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4540, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":4539
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitStatement_component_function" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitStatement_component_function(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4542
 *                 return visitor.visitStatement_component_function(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4542, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4542, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4542, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4542, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4542, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4542, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":4538
 *                 listener.exitStatement_component_function(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStatement_component_function" ):
 *                 return visitor.visitStatement_component_function(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_component_functionContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4547
 *     class Statement_returnContext(Statement_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Statement_returnContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Statement_returnContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Statement_returnContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Statement_returnContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 4547, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 4547, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 4547, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4547, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_returnContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Statement_returnContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Statement_returnContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":4548
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 4548, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4548, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4548, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4548, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4548, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4548, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4548, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4548, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4548, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4549
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4549, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4549, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4549, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4549, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4549, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4547
 *     class Statement_returnContext(Statement_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_returnContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4552
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStatement_return" ):
 *                 listener.enterStatement_return(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Statement_returnContext_3enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Statement_returnContext_3enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Statement_returnContext_3enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Statement_returnContext_3enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 4552, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 4552, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4552, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_returnContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Statement_returnContext_2enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Statement_returnContext_2enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":4553
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterStatement_return" ):             # <<<<<<<<<<<<<<
 *                 listener.enterStatement_return(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterStatement_return); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4553, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4554
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterStatement_return" ):
 *                 listener.enterStatement_return(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterStatement_return); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4554, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4554, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4554, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4554, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4554, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4554, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":4553
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterStatement_return" ):             # <<<<<<<<<<<<<<
 *                 listener.enterStatement_return(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4552
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStatement_return" ):
 *                 listener.enterStatement_return(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_returnContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4556
 *                 listener.enterStatement_return(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStatement_return" ):
 *                 listener.exitStatement_return(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Statement_returnContext_5exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Statement_returnContext_5exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Statement_returnContext_5exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Statement_returnContext_5exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 4556, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 4556, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4556, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_returnContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Statement_returnContext_4exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Statement_returnContext_4exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":4557
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitStatement_return" ):             # <<<<<<<<<<<<<<
 *                 listener.exitStatement_return(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitStatement_return); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4557, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4558
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitStatement_return" ):
 *                 listener.exitStatement_return(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitStatement_return); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4558, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4558, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4558, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4558, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4558, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4558, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":4557
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitStatement_return" ):             # <<<<<<<<<<<<<<
 *                 listener.exitStatement_return(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4556
 *                 listener.enterStatement_return(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStatement_return" ):
 *                 listener.exitStatement_return(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_returnContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4560
 *                 listener.exitStatement_return(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStatement_return" ):
 *                 return visitor.visitStatement_return(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Statement_returnContext_7accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Statement_returnContext_7accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Statement_returnContext_7accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Statement_returnContext_7accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 4560, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 4560, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4560, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_returnContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Statement_returnContext_6accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Statement_returnContext_6accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":4561
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitStatement_return" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitStatement_return(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitStatement_return); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4561, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4562
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitStatement_return" ):
 *                 return visitor.visitStatement_return(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitStatement_return); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4562, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4562, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4562, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4562, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4562, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4562, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":4561
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitStatement_return" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitStatement_return(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4564
 *                 return visitor.visitStatement_return(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4564, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4564, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4564, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4564, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4564, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":4560
 *                 listener.exitStatement_return(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStatement_return" ):
 *                 return visitor.visitStatement_return(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_returnContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4569
 *     class Statement_forContext(Statement_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Statement_forContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20Statement_forContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20Statement_forContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Statement_forContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 4569, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 4569, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 4569, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4569, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_forContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20Statement_forContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Statement_forContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":4570
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 4570, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4570, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4570, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4570, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4570, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4570, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4571
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def for_statement(self):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4571, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4571, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4571, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4571, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4571, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4571, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4569
 *     class Statement_forContext(Statement_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_forContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4573
 *             self.copyFrom(ctx)
 * 
 *         def for_statement(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.For_statementContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Statement_forContext_3for_statement(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20Statement_forContext_3for_statement = {"for_statement", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20Statement_forContext_3for_statement, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Statement_forContext_3for_statement(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("for_statement (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20Statement_forContext_2for_statement(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Statement_forContext_2for_statement(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("for_statement", 0);

  /* "pymola/generated/ModelicaParser.pyx":4574
 * 
 *         def for_statement(self):
 *             return self.getTypedRuleContext(ModelicaParser.For_statementContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_For_statementContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4574, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4574, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4574, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4574, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":4573
 *             self.copyFrom(ctx)
 * 
 *         def for_statement(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.For_statementContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_forContext.for_statement", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4577
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStatement_for" ):
 *                 listener.enterStatement_for(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Statement_forContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20Statement_forContext_5enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20Statement_forContext_5enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Statement_forContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 4577, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 4577, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4577, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_forContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20Statement_forContext_4enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Statement_forContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":4578
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterStatement_for" ):             # <<<<<<<<<<<<<<
 *                 listener.enterStatement_for(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterStatement_for); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4578, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4579
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterStatement_for" ):
 *                 listener.enterStatement_for(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterStatement_for); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4579, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4579, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4579, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4579, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4579, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4579, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":4578
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterStatement_for" ):             # <<<<<<<<<<<<<<
 *                 listener.enterStatement_for(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4577
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStatement_for" ):
 *                 listener.enterStatement_for(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_forContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4581
 *                 listener.enterStatement_for(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStatement_for" ):
 *                 listener.exitStatement_for(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Statement_forContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20Statement_forContext_7exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20Statement_forContext_7exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Statement_forContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 4581, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 4581, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4581, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_forContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20Statement_forContext_6exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Statement_forContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":4582
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitStatement_for" ):             # <<<<<<<<<<<<<<
 *                 listener.exitStatement_for(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitStatement_for); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4582, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4583
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitStatement_for" ):
 *                 listener.exitStatement_for(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitStatement_for); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4583, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4583, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4583, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4583, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4583, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4583, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":4582
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitStatement_for" ):             # <<<<<<<<<<<<<<
 *                 listener.exitStatement_for(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4581
 *                 listener.enterStatement_for(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStatement_for" ):
 *                 listener.exitStatement_for(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_forContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4585
 *                 listener.exitStatement_for(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStatement_for" ):
 *                 return visitor.visitStatement_for(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Statement_forContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20Statement_forContext_9accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20Statement_forContext_9accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Statement_forContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 4585, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 4585, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4585, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_forContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20Statement_forContext_8accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Statement_forContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":4586
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitStatement_for" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitStatement_for(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitStatement_for); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4586, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4587
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitStatement_for" ):
 *                 return visitor.visitStatement_for(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitStatement_for); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4587, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4587, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4587, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4587, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4587, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4587, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":4586
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitStatement_for" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitStatement_for(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4589
 *                 return visitor.visitStatement_for(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4589, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4589, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4589, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4589, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4589, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4589, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":4585
 *                 listener.exitStatement_for(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStatement_for" ):
 *                 return visitor.visitStatement_for(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_forContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4594
 *     class Statement_whenContext(Statement_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Statement_whenContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Statement_whenContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Statement_whenContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Statement_whenContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 4594, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 4594, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 4594, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4594, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_whenContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Statement_whenContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Statement_whenContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":4595
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 4595, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4595, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4595, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4595, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4595, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4595, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4596
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def when_statement(self):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4596, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4596, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4596, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4596, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4596, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4596, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4594
 *     class Statement_whenContext(Statement_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_whenContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4598
 *             self.copyFrom(ctx)
 * 
 *         def when_statement(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.When_statementContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Statement_whenContext_3when_statement(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Statement_whenContext_3when_statement = {"when_statement", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Statement_whenContext_3when_statement, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Statement_whenContext_3when_statement(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("when_statement (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Statement_whenContext_2when_statement(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Statement_whenContext_2when_statement(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("when_statement", 0);

  /* "pymola/generated/ModelicaParser.pyx":4599
 * 
 *         def when_statement(self):
 *             return self.getTypedRuleContext(ModelicaParser.When_statementContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_When_statementContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4599, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4599, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4599, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4599, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":4598
 *             self.copyFrom(ctx)
 * 
 *         def when_statement(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.When_statementContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_whenContext.when_statement", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4602
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStatement_when" ):
 *                 listener.enterStatement_when(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Statement_whenContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Statement_whenContext_5enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Statement_whenContext_5enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Statement_whenContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 4602, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 4602, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4602, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_whenContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Statement_whenContext_4enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Statement_whenContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":4603
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterStatement_when" ):             # <<<<<<<<<<<<<<
 *                 listener.enterStatement_when(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterStatement_when); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4603, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4604
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterStatement_when" ):
 *                 listener.enterStatement_when(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterStatement_when); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4604, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4604, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4604, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4604, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4604, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4604, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":4603
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterStatement_when" ):             # <<<<<<<<<<<<<<
 *                 listener.enterStatement_when(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4602
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStatement_when" ):
 *                 listener.enterStatement_when(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_whenContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4606
 *                 listener.enterStatement_when(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStatement_when" ):
 *                 listener.exitStatement_when(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Statement_whenContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Statement_whenContext_7exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Statement_whenContext_7exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Statement_whenContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 4606, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 4606, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4606, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_whenContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Statement_whenContext_6exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Statement_whenContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":4607
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitStatement_when" ):             # <<<<<<<<<<<<<<
 *                 listener.exitStatement_when(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitStatement_when); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4607, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4608
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitStatement_when" ):
 *                 listener.exitStatement_when(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitStatement_when); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4608, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4608, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4608, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4608, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4608, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4608, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":4607
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitStatement_when" ):             # <<<<<<<<<<<<<<
 *                 listener.exitStatement_when(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4606
 *                 listener.enterStatement_when(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStatement_when" ):
 *                 listener.exitStatement_when(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_whenContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4610
 *                 listener.exitStatement_when(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStatement_when" ):
 *                 return visitor.visitStatement_when(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Statement_whenContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Statement_whenContext_9accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Statement_whenContext_9accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Statement_whenContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 4610, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 4610, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4610, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_whenContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Statement_whenContext_8accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Statement_whenContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":4611
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitStatement_when" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitStatement_when(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitStatement_when); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4611, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4612
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitStatement_when" ):
 *                 return visitor.visitStatement_when(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitStatement_when); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4612, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4612, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4612, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4612, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4612, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4612, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":4611
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitStatement_when" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitStatement_when(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4614
 *                 return visitor.visitStatement_when(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4614, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4614, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4614, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4614, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4614, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4614, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":4610
 *                 listener.exitStatement_when(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStatement_when" ):
 *                 return visitor.visitStatement_when(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_whenContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4619
 *     class Statement_ifContext(Statement_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Statement_ifContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19Statement_ifContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19Statement_ifContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Statement_ifContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 4619, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 4619, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 4619, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4619, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_ifContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19Statement_ifContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Statement_ifContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":4620
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 4620, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4620, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4620, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4620, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4620, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4620, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4620, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4620, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4620, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4621
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def if_statement(self):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4621, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4621, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4621, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4621, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4621, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4621, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4619
 *     class Statement_ifContext(Statement_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_ifContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4623
 *             self.copyFrom(ctx)
 * 
 *         def if_statement(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.If_statementContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Statement_ifContext_3if_statement(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19Statement_ifContext_3if_statement = {"if_statement", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19Statement_ifContext_3if_statement, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Statement_ifContext_3if_statement(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("if_statement (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19Statement_ifContext_2if_statement(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Statement_ifContext_2if_statement(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("if_statement", 0);

  /* "pymola/generated/ModelicaParser.pyx":4624
 * 
 *         def if_statement(self):
 *             return self.getTypedRuleContext(ModelicaParser.If_statementContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_If_statementContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4624, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4624, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4624, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4624, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":4623
 *             self.copyFrom(ctx)
 * 
 *         def if_statement(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.If_statementContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_ifContext.if_statement", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4627
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStatement_if" ):
 *                 listener.enterStatement_if(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Statement_ifContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19Statement_ifContext_5enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19Statement_ifContext_5enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Statement_ifContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 4627, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 4627, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4627, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_ifContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19Statement_ifContext_4enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Statement_ifContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":4628
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterStatement_if" ):             # <<<<<<<<<<<<<<
 *                 listener.enterStatement_if(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterStatement_if); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4628, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4629
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterStatement_if" ):
 *                 listener.enterStatement_if(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterStatement_if); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4629, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4629, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4629, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4629, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4629, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4629, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":4628
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterStatement_if" ):             # <<<<<<<<<<<<<<
 *                 listener.enterStatement_if(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4627
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStatement_if" ):
 *                 listener.enterStatement_if(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_ifContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4631
 *                 listener.enterStatement_if(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStatement_if" ):
 *                 listener.exitStatement_if(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Statement_ifContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19Statement_ifContext_7exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19Statement_ifContext_7exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Statement_ifContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 4631, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 4631, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4631, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_ifContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19Statement_ifContext_6exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Statement_ifContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":4632
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitStatement_if" ):             # <<<<<<<<<<<<<<
 *                 listener.exitStatement_if(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitStatement_if); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4632, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4633
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitStatement_if" ):
 *                 listener.exitStatement_if(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitStatement_if); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4633, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4633, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4633, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4633, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4633, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4633, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":4632
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitStatement_if" ):             # <<<<<<<<<<<<<<
 *                 listener.exitStatement_if(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4631
 *                 listener.enterStatement_if(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStatement_if" ):
 *                 listener.exitStatement_if(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_ifContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4635
 *                 listener.exitStatement_if(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStatement_if" ):
 *                 return visitor.visitStatement_if(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Statement_ifContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19Statement_ifContext_9accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19Statement_ifContext_9accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Statement_ifContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 4635, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 4635, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4635, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_ifContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19Statement_ifContext_8accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Statement_ifContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":4636
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitStatement_if" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitStatement_if(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitStatement_if); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4636, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4637
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitStatement_if" ):
 *                 return visitor.visitStatement_if(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitStatement_if); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4637, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4637, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4637, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4637, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4637, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4637, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":4636
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitStatement_if" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitStatement_if(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4639
 *                 return visitor.visitStatement_if(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4639, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4639, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4639, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4639, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4639, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4639, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":4635
 *                 listener.exitStatement_if(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStatement_if" ):
 *                 return visitor.visitStatement_if(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_ifContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4644
 *     class Statement_component_referenceContext(Statement_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 4644, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 4644, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 4644, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4644, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_component_referenceContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":4645
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 4645, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4645, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4645, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4645, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4645, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4645, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4646
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def component_reference(self):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4646, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4646, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4646, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4646, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4646, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4644
 *     class Statement_component_referenceContext(Statement_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_component_referenceContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4648
 *             self.copyFrom(ctx)
 * 
 *         def component_reference(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_3component_reference(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_3component_reference = {"component_reference", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_3component_reference, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_3component_reference(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("component_reference (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_2component_reference(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_2component_reference(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("component_reference", 0);

  /* "pymola/generated/ModelicaParser.pyx":4649
 * 
 *         def component_reference(self):
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)             # <<<<<<<<<<<<<<
 * 
 *         def expression(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Component_referenceContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4649, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4649, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4649, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4649, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":4648
 *             self.copyFrom(ctx)
 * 
 *         def component_reference(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_component_referenceContext.component_reference", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4651
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 *         def expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_5expression(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_5expression = {"expression", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_5expression, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_5expression(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("expression (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_4expression(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_4expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("expression", 0);

  /* "pymola/generated/ModelicaParser.pyx":4652
 * 
 *         def expression(self):
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)             # <<<<<<<<<<<<<<
 * 
 *         def function_call_args(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_ExpressionContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4652, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4652, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4652, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4652, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":4651
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 *         def expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_component_referenceContext.expression", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4654
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 *         def function_call_args(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Function_call_argsContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_7function_call_args(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_7function_call_args = {"function_call_args", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_7function_call_args, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_7function_call_args(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("function_call_args (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_6function_call_args(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_6function_call_args(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("function_call_args", 0);

  /* "pymola/generated/ModelicaParser.pyx":4655
 * 
 *         def function_call_args(self):
 *             return self.getTypedRuleContext(ModelicaParser.Function_call_argsContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Function_call_argsContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4655, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4655, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4655, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4655, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":4654
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 *         def function_call_args(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Function_call_argsContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_component_referenceContext.function_call_args", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4658
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStatement_component_reference" ):
 *                 listener.enterStatement_component_reference(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_9enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_9enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 4658, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 4658, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4658, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_component_referenceContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_8enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":4659
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterStatement_component_reference" ):             # <<<<<<<<<<<<<<
 *                 listener.enterStatement_component_reference(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterStatement_component_referen); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4659, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4660
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterStatement_component_reference" ):
 *                 listener.enterStatement_component_reference(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterStatement_component_referen); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4660, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4660, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4660, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4660, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4660, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4660, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":4659
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterStatement_component_reference" ):             # <<<<<<<<<<<<<<
 *                 listener.enterStatement_component_reference(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4658
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStatement_component_reference" ):
 *                 listener.enterStatement_component_reference(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_component_referenceContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4662
 *                 listener.enterStatement_component_reference(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStatement_component_reference" ):
 *                 listener.exitStatement_component_reference(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_11exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_11exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 4662, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 4662, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4662, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_component_referenceContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_10exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":4663
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitStatement_component_reference" ):             # <<<<<<<<<<<<<<
 *                 listener.exitStatement_component_reference(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitStatement_component_referenc); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4663, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4664
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitStatement_component_reference" ):
 *                 listener.exitStatement_component_reference(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitStatement_component_referenc); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4664, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4664, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4664, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4664, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4664, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4664, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":4663
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitStatement_component_reference" ):             # <<<<<<<<<<<<<<
 *                 listener.exitStatement_component_reference(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4662
 *                 listener.enterStatement_component_reference(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStatement_component_reference" ):
 *                 listener.exitStatement_component_reference(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_component_referenceContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4666
 *                 listener.exitStatement_component_reference(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStatement_component_reference" ):
 *                 return visitor.visitStatement_component_reference(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_13accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_13accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 4666, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 4666, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4666, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_component_referenceContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_12accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":4667
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitStatement_component_reference" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitStatement_component_reference(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitStatement_component_referen); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4667, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4668
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitStatement_component_reference" ):
 *                 return visitor.visitStatement_component_reference(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitStatement_component_referen); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4668, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4668, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4668, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4668, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4668, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4668, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":4667
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitStatement_component_reference" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitStatement_component_reference(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4670
 *                 return visitor.visitStatement_component_reference(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4670, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4670, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4670, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4670, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4670, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4670, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":4666
 *                 listener.exitStatement_component_reference(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStatement_component_reference" ):
 *                 return visitor.visitStatement_component_reference(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Statement_component_referenceContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4674
 * 
 * 
 *     def statement_options(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Statement_optionsContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_87statement_options(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_87statement_options = {"statement_options", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_87statement_options, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_87statement_options(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("statement_options (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_86statement_options(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_86statement_options(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v_token = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  char const *__pyx_t_22;
  __Pyx_RefNannySetupContext("statement_options", 0);

  /* "pymola/generated/ModelicaParser.pyx":4676
 *     def statement_options(self):
 * 
 *         localctx = ModelicaParser.Statement_optionsContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 84, self.RULE_statement_options)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4676, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Statement_optionsContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4676, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4676, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4676, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4676, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4676, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4676, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4676, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4677
 * 
 *         localctx = ModelicaParser.Statement_optionsContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 84, self.RULE_statement_options)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4677, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_statement_options); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4677, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_84, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4677, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_84, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4677, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4677, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_84);
    __Pyx_GIVEREF(__pyx_int_84);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_84);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4677, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4678
 *         localctx = ModelicaParser.Statement_optionsContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 84, self.RULE_statement_options)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.state = 614
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 4678, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4679
 *         self.enterRule(localctx, 84, self.RULE_statement_options)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.state = 614
 *             self._errHandler.sync(self)
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":4680
 *         self._la = 0 # Token type
 *         try:
 *             self.state = 614             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_614) < 0) __PYX_ERR(0, 4680, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":4681
 *         try:
 *             self.state = 614
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.IDENT]:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4681, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4681, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4681, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4681, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4681, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4681, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4681, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":4682
 *             self.state = 614
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if token in [ModelicaParser.IDENT]:
 *                 localctx = ModelicaParser.Statement_component_referenceContext(self, localctx)
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4682, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4682, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__897, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4682, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_v_token = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":4683
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.IDENT]:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Statement_component_referenceContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 1)
 */
        __Pyx_INCREF(__pyx_v_token);
        __pyx_t_1 = __pyx_v_token;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4683, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4683, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4683, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 4683, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_12 = (__pyx_t_11 != 0);
        if (__pyx_t_12) {

          /* "pymola/generated/ModelicaParser.pyx":4684
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.IDENT]:
 *                 localctx = ModelicaParser.Statement_component_referenceContext(self, localctx)             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 1)
 *                 self.state = 588
 */
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4684, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Statement_component_referenceCon); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4684, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4684, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4684, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4684, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_2) {
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_v_localctx);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4684, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4685
 *             if token in [ModelicaParser.IDENT]:
 *                 localctx = ModelicaParser.Statement_component_referenceContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *                 self.state = 588
 *                 self.component_reference()
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4685, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_1};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4685, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_1};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4685, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4685, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_1);
            __Pyx_GIVEREF(__pyx_int_1);
            PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_1);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4685, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4686
 *                 localctx = ModelicaParser.Statement_component_referenceContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 1)
 *                 self.state = 588             # <<<<<<<<<<<<<<
 *                 self.component_reference()
 *                 self.state = 592
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_588) < 0) __PYX_ERR(0, 4686, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":4687
 *                 self.enterOuterAlt(localctx, 1)
 *                 self.state = 588
 *                 self.component_reference()             # <<<<<<<<<<<<<<
 *                 self.state = 592
 *                 self._errHandler.sync(self)
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_component_reference); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4687, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4687, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4687, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4688
 *                 self.state = 588
 *                 self.component_reference()
 *                 self.state = 592             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 token = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_592) < 0) __PYX_ERR(0, 4688, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":4689
 *                 self.component_reference()
 *                 self.state = 592
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 token = self._input.LA(1)
 *                 if token in [ModelicaParser.T__40]:
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4689, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4689, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4689, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4689, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4689, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4689, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4689, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4690
 *                 self.state = 592
 *                 self._errHandler.sync(self)
 *                 token = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if token in [ModelicaParser.T__40]:
 *                     self.state = 589
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4690, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4690, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__898, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4690, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF_SET(__pyx_v_token, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4691
 *                 self._errHandler.sync(self)
 *                 token = self._input.LA(1)
 *                 if token in [ModelicaParser.T__40]:             # <<<<<<<<<<<<<<
 *                     self.state = 589
 *                     self.match(ModelicaParser.T__40)
 */
          __Pyx_INCREF(__pyx_v_token);
          __pyx_t_1 = __pyx_v_token;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4691, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__40); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4691, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4691, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 4691, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_11 = (__pyx_t_12 != 0);
          if (__pyx_t_11) {

            /* "pymola/generated/ModelicaParser.pyx":4692
 *                 token = self._input.LA(1)
 *                 if token in [ModelicaParser.T__40]:
 *                     self.state = 589             # <<<<<<<<<<<<<<
 *                     self.match(ModelicaParser.T__40)
 *                     self.state = 590
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_589) < 0) __PYX_ERR(0, 4692, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":4693
 *                 if token in [ModelicaParser.T__40]:
 *                     self.state = 589
 *                     self.match(ModelicaParser.T__40)             # <<<<<<<<<<<<<<
 *                     self.state = 590
 *                     self.expression()
 */
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4693, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4693, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__40); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4693, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_3)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_3);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
              }
            }
            if (!__pyx_t_3) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4693, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_2)) {
                PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_7};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4693, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
                PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_7};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4693, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              } else
              #endif
              {
                __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4693, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __pyx_t_3 = NULL;
                __Pyx_GIVEREF(__pyx_t_7);
                PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_7);
                __pyx_t_7 = 0;
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4693, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":4694
 *                     self.state = 589
 *                     self.match(ModelicaParser.T__40)
 *                     self.state = 590             # <<<<<<<<<<<<<<
 *                     self.expression()
 *                     pass
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_590) < 0) __PYX_ERR(0, 4694, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":4695
 *                     self.match(ModelicaParser.T__40)
 *                     self.state = 590
 *                     self.expression()             # <<<<<<<<<<<<<<
 *                     pass
 *                 elif token in [ModelicaParser.T__16]:
 */
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expression); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4695, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_4 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_4)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_4);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
              }
            }
            if (__pyx_t_4) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4695, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            } else {
              __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4695, __pyx_L6_error)
            }
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":4691
 *                 self._errHandler.sync(self)
 *                 token = self._input.LA(1)
 *                 if token in [ModelicaParser.T__40]:             # <<<<<<<<<<<<<<
 *                     self.state = 589
 *                     self.match(ModelicaParser.T__40)
 */
            goto __pyx_L15;
          }

          /* "pymola/generated/ModelicaParser.pyx":4697
 *                     self.expression()
 *                     pass
 *                 elif token in [ModelicaParser.T__16]:             # <<<<<<<<<<<<<<
 *                     self.state = 591
 *                     self.function_call_args()
 */
          __Pyx_INCREF(__pyx_v_token);
          __pyx_t_1 = __pyx_v_token;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4697, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__16); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4697, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4697, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 4697, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_12 = (__pyx_t_11 != 0);
          if (__pyx_t_12) {

            /* "pymola/generated/ModelicaParser.pyx":4698
 *                     pass
 *                 elif token in [ModelicaParser.T__16]:
 *                     self.state = 591             # <<<<<<<<<<<<<<
 *                     self.function_call_args()
 *                     pass
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_591) < 0) __PYX_ERR(0, 4698, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":4699
 *                 elif token in [ModelicaParser.T__16]:
 *                     self.state = 591
 *                     self.function_call_args()             # <<<<<<<<<<<<<<
 *                     pass
 *                 else:
 */
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_function_call_args); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4699, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_4 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_4)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_4);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
              }
            }
            if (__pyx_t_4) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4699, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            } else {
              __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4699, __pyx_L6_error)
            }
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":4697
 *                     self.expression()
 *                     pass
 *                 elif token in [ModelicaParser.T__16]:             # <<<<<<<<<<<<<<
 *                     self.state = 591
 *                     self.function_call_args()
 */
            goto __pyx_L15;
          }

          /* "pymola/generated/ModelicaParser.pyx":4702
 *                     pass
 *                 else:
 *                     raise NoViableAltException(self)             # <<<<<<<<<<<<<<
 * 
 *                 pass
 */
          /*else*/ {
            __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_NoViableAltException); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4702, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_4 = NULL;
            if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_4)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_4);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
              }
            }
            if (!__pyx_t_4) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4702, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_2)) {
                PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4702, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
                PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4702, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else
              #endif
              {
                __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4702, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_7);
                __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4); __pyx_t_4 = NULL;
                __Pyx_INCREF(__pyx_v_self);
                __Pyx_GIVEREF(__pyx_v_self);
                PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self);
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4702, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_Raise(__pyx_t_1, 0, 0, 0);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __PYX_ERR(0, 4702, __pyx_L6_error)
          }
          __pyx_L15:;

          /* "pymola/generated/ModelicaParser.pyx":4683
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.IDENT]:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Statement_component_referenceContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 1)
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":4705
 * 
 *                 pass
 *             elif token in [ModelicaParser.T__16]:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Statement_component_functionContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 2)
 */
        __Pyx_INCREF(__pyx_v_token);
        __pyx_t_1 = __pyx_v_token;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4705, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__16); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4705, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4705, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 4705, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_11 = (__pyx_t_12 != 0);
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":4706
 *                 pass
 *             elif token in [ModelicaParser.T__16]:
 *                 localctx = ModelicaParser.Statement_component_functionContext(self, localctx)             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 2)
 *                 self.state = 594
 */
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4706, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Statement_component_functionCont); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4706, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4706, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4706, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4706, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (__pyx_t_2) {
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_v_localctx);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4706, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4707
 *             elif token in [ModelicaParser.T__16]:
 *                 localctx = ModelicaParser.Statement_component_functionContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 2)             # <<<<<<<<<<<<<<
 *                 self.state = 594
 *                 self.match(ModelicaParser.T__16)
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4707, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_4 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4707, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4707, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4707, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (__pyx_t_4) {
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_2);
            __Pyx_GIVEREF(__pyx_int_2);
            PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_2);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4707, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4708
 *                 localctx = ModelicaParser.Statement_component_functionContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 2)
 *                 self.state = 594             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__16)
 *                 self.state = 595
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_594) < 0) __PYX_ERR(0, 4708, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":4709
 *                 self.enterOuterAlt(localctx, 2)
 *                 self.state = 594
 *                 self.match(ModelicaParser.T__16)             # <<<<<<<<<<<<<<
 *                 self.state = 595
 *                 self.component_reference()
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4709, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4709, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__16); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4709, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4709, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_4};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4709, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_4};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4709, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4709, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_GIVEREF(__pyx_t_4);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_t_4);
              __pyx_t_4 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4709, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4710
 *                 self.state = 594
 *                 self.match(ModelicaParser.T__16)
 *                 self.state = 595             # <<<<<<<<<<<<<<
 *                 self.component_reference()
 *                 self.state = 600
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_595) < 0) __PYX_ERR(0, 4710, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":4711
 *                 self.match(ModelicaParser.T__16)
 *                 self.state = 595
 *                 self.component_reference()             # <<<<<<<<<<<<<<
 *                 self.state = 600
 *                 self._errHandler.sync(self)
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_component_reference); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4711, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (__pyx_t_3) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4711, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4711, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4712
 *                 self.state = 595
 *                 self.component_reference()
 *                 self.state = 600             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_600) < 0) __PYX_ERR(0, 4712, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":4713
 *                 self.component_reference()
 *                 self.state = 600
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 *                 while _la==ModelicaParser.T__20:
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4713, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_sync); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4713, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4713, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4713, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4713, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4713, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4713, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4714
 *                 self.state = 600
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 while _la==ModelicaParser.T__20:
 *                     self.state = 596
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4714, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4714, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__899, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4714, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_v__la = __pyx_t_1;
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4715
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 while _la==ModelicaParser.T__20:             # <<<<<<<<<<<<<<
 *                     self.state = 596
 *                     self.match(ModelicaParser.T__20)
 */
          while (1) {
            __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4715, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__20); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4715, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4715, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 4715, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (!__pyx_t_11) break;

            /* "pymola/generated/ModelicaParser.pyx":4716
 *                 _la = self._input.LA(1)
 *                 while _la==ModelicaParser.T__20:
 *                     self.state = 596             # <<<<<<<<<<<<<<
 *                     self.match(ModelicaParser.T__20)
 *                     self.state = 597
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_596) < 0) __PYX_ERR(0, 4716, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":4717
 *                 while _la==ModelicaParser.T__20:
 *                     self.state = 596
 *                     self.match(ModelicaParser.T__20)             # <<<<<<<<<<<<<<
 *                     self.state = 597
 *                     self.component_reference()
 */
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4717, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4717, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__20); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4717, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
              __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
              if (likely(__pyx_t_4)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                __Pyx_INCREF(__pyx_t_4);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_3, function);
              }
            }
            if (!__pyx_t_4) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4717, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_3)) {
                PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_7};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4717, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
                PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_7};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4717, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              } else
              #endif
              {
                __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4717, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
                __Pyx_GIVEREF(__pyx_t_7);
                PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_7);
                __pyx_t_7 = 0;
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4717, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":4718
 *                     self.state = 596
 *                     self.match(ModelicaParser.T__20)
 *                     self.state = 597             # <<<<<<<<<<<<<<
 *                     self.component_reference()
 *                     self.state = 602
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_597) < 0) __PYX_ERR(0, 4718, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":4719
 *                     self.match(ModelicaParser.T__20)
 *                     self.state = 597
 *                     self.component_reference()             # <<<<<<<<<<<<<<
 *                     self.state = 602
 *                     self._errHandler.sync(self)
 */
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_component_reference); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4719, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_2 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
              __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
              if (likely(__pyx_t_2)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                __Pyx_INCREF(__pyx_t_2);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_3, function);
              }
            }
            if (__pyx_t_2) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4719, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            } else {
              __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4719, __pyx_L6_error)
            }
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":4720
 *                     self.state = 597
 *                     self.component_reference()
 *                     self.state = 602             # <<<<<<<<<<<<<<
 *                     self._errHandler.sync(self)
 *                     _la = self._input.LA(1)
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_602) < 0) __PYX_ERR(0, 4720, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":4721
 *                     self.component_reference()
 *                     self.state = 602
 *                     self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                     _la = self._input.LA(1)
 * 
 */
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4721, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4721, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_3)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_3);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
              }
            }
            if (!__pyx_t_3) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4721, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_2)) {
                PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4721, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
                PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4721, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else
              #endif
              {
                __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4721, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_7);
                __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
                __Pyx_INCREF(__pyx_v_self);
                __Pyx_GIVEREF(__pyx_v_self);
                PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self);
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4721, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":4722
 *                     self.state = 602
 *                     self._errHandler.sync(self)
 *                     _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *                 self.state = 603
 */
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4722, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4722, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__900, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4722, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_1);
            __pyx_t_1 = 0;
          }

          /* "pymola/generated/ModelicaParser.pyx":4724
 *                     _la = self._input.LA(1)
 * 
 *                 self.state = 603             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__18)
 *                 self.state = 604
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_603) < 0) __PYX_ERR(0, 4724, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":4725
 * 
 *                 self.state = 603
 *                 self.match(ModelicaParser.T__18)             # <<<<<<<<<<<<<<
 *                 self.state = 604
 *                 self.match(ModelicaParser.T__40)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4725, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4725, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__18); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4725, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4725, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4725, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4725, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4725, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4725, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4726
 *                 self.state = 603
 *                 self.match(ModelicaParser.T__18)
 *                 self.state = 604             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__40)
 *                 self.state = 605
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_604) < 0) __PYX_ERR(0, 4726, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":4727
 *                 self.match(ModelicaParser.T__18)
 *                 self.state = 604
 *                 self.match(ModelicaParser.T__40)             # <<<<<<<<<<<<<<
 *                 self.state = 605
 *                 self.component_reference()
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4727, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4727, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__40); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4727, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_4) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4727, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4727, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4727, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4727, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4); __pyx_t_4 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4727, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4728
 *                 self.state = 604
 *                 self.match(ModelicaParser.T__40)
 *                 self.state = 605             # <<<<<<<<<<<<<<
 *                 self.component_reference()
 *                 self.state = 606
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_605) < 0) __PYX_ERR(0, 4728, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":4729
 *                 self.match(ModelicaParser.T__40)
 *                 self.state = 605
 *                 self.component_reference()             # <<<<<<<<<<<<<<
 *                 self.state = 606
 *                 self.function_call_args()
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_component_reference); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4729, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4729, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4729, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4730
 *                 self.state = 605
 *                 self.component_reference()
 *                 self.state = 606             # <<<<<<<<<<<<<<
 *                 self.function_call_args()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_606) < 0) __PYX_ERR(0, 4730, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":4731
 *                 self.component_reference()
 *                 self.state = 606
 *                 self.function_call_args()             # <<<<<<<<<<<<<<
 *                 pass
 *             elif token in [ModelicaParser.T__43]:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_function_call_args); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4731, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4731, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4731, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4705
 * 
 *                 pass
 *             elif token in [ModelicaParser.T__16]:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Statement_component_functionContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 2)
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":4733
 *                 self.function_call_args()
 *                 pass
 *             elif token in [ModelicaParser.T__43]:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Statement_breakContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 3)
 */
        __Pyx_INCREF(__pyx_v_token);
        __pyx_t_1 = __pyx_v_token;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4733, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__43); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4733, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4733, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 4733, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_12 = (__pyx_t_11 != 0);
        if (__pyx_t_12) {

          /* "pymola/generated/ModelicaParser.pyx":4734
 *                 pass
 *             elif token in [ModelicaParser.T__43]:
 *                 localctx = ModelicaParser.Statement_breakContext(self, localctx)             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 3)
 *                 self.state = 608
 */
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4734, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Statement_breakContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4734, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4734, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4734, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4734, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_2) {
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_v_localctx);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4734, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4735
 *             elif token in [ModelicaParser.T__43]:
 *                 localctx = ModelicaParser.Statement_breakContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 3)             # <<<<<<<<<<<<<<
 *                 self.state = 608
 *                 self.match(ModelicaParser.T__43)
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4735, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_3};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4735, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_3};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4735, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4735, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_3);
            __Pyx_GIVEREF(__pyx_int_3);
            PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_3);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4735, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4736
 *                 localctx = ModelicaParser.Statement_breakContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 3)
 *                 self.state = 608             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__43)
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_608) < 0) __PYX_ERR(0, 4736, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":4737
 *                 self.enterOuterAlt(localctx, 3)
 *                 self.state = 608
 *                 self.match(ModelicaParser.T__43)             # <<<<<<<<<<<<<<
 *                 pass
 *             elif token in [ModelicaParser.T__44]:
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4737, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4737, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__43); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4737, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4737, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4737, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4737, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4737, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4737, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4733
 *                 self.function_call_args()
 *                 pass
 *             elif token in [ModelicaParser.T__43]:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Statement_breakContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 3)
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":4739
 *                 self.match(ModelicaParser.T__43)
 *                 pass
 *             elif token in [ModelicaParser.T__44]:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Statement_returnContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 4)
 */
        __Pyx_INCREF(__pyx_v_token);
        __pyx_t_1 = __pyx_v_token;
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4739, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__44); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4739, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4739, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 4739, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_11 = (__pyx_t_12 != 0);
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":4740
 *                 pass
 *             elif token in [ModelicaParser.T__44]:
 *                 localctx = ModelicaParser.Statement_returnContext(self, localctx)             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 4)
 *                 self.state = 609
 */
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4740, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Statement_returnContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4740, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4740, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4740, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4740, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_7) {
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_v_localctx);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4740, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4741
 *             elif token in [ModelicaParser.T__44]:
 *                 localctx = ModelicaParser.Statement_returnContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 4)             # <<<<<<<<<<<<<<
 *                 self.state = 609
 *                 self.match(ModelicaParser.T__44)
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4741, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_4};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4741, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_4};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4741, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4741, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_4);
            __Pyx_GIVEREF(__pyx_int_4);
            PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_4);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4741, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4742
 *                 localctx = ModelicaParser.Statement_returnContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 4)
 *                 self.state = 609             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__44)
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_609) < 0) __PYX_ERR(0, 4742, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":4743
 *                 self.enterOuterAlt(localctx, 4)
 *                 self.state = 609
 *                 self.match(ModelicaParser.T__44)             # <<<<<<<<<<<<<<
 *                 pass
 *             elif token in [ModelicaParser.T__39]:
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4743, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4743, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__44); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4743, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4743, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4743, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4743, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4743, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_7); __pyx_t_7 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4743, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4739
 *                 self.match(ModelicaParser.T__43)
 *                 pass
 *             elif token in [ModelicaParser.T__44]:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Statement_returnContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 4)
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":4745
 *                 self.match(ModelicaParser.T__44)
 *                 pass
 *             elif token in [ModelicaParser.T__39]:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Statement_ifContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 5)
 */
        __Pyx_INCREF(__pyx_v_token);
        __pyx_t_1 = __pyx_v_token;
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4745, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__39); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4745, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4745, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 4745, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_12 = (__pyx_t_11 != 0);
        if (__pyx_t_12) {

          /* "pymola/generated/ModelicaParser.pyx":4746
 *                 pass
 *             elif token in [ModelicaParser.T__39]:
 *                 localctx = ModelicaParser.Statement_ifContext(self, localctx)             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 5)
 *                 self.state = 610
 */
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4746, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_Statement_ifContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4746, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4746, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4746, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4746, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_4) {
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __pyx_t_4 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_v_localctx);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4746, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4747
 *             elif token in [ModelicaParser.T__39]:
 *                 localctx = ModelicaParser.Statement_ifContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 5)             # <<<<<<<<<<<<<<
 *                 self.state = 610
 *                 self.if_statement()
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4747, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_5};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4747, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_5};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4747, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4747, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_5);
            __Pyx_GIVEREF(__pyx_int_5);
            PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_int_5);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4747, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4748
 *                 localctx = ModelicaParser.Statement_ifContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 5)
 *                 self.state = 610             # <<<<<<<<<<<<<<
 *                 self.if_statement()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_610) < 0) __PYX_ERR(0, 4748, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":4749
 *                 self.enterOuterAlt(localctx, 5)
 *                 self.state = 610
 *                 self.if_statement()             # <<<<<<<<<<<<<<
 *                 pass
 *             elif token in [ModelicaParser.T__48]:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_if_statement); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4749, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_4 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (__pyx_t_4) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4749, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4749, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4745
 *                 self.match(ModelicaParser.T__44)
 *                 pass
 *             elif token in [ModelicaParser.T__39]:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Statement_ifContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 5)
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":4751
 *                 self.if_statement()
 *                 pass
 *             elif token in [ModelicaParser.T__48]:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Statement_forContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 6)
 */
        __Pyx_INCREF(__pyx_v_token);
        __pyx_t_1 = __pyx_v_token;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4751, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__48); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4751, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4751, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 4751, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_11 = (__pyx_t_12 != 0);
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":4752
 *                 pass
 *             elif token in [ModelicaParser.T__48]:
 *                 localctx = ModelicaParser.Statement_forContext(self, localctx)             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 6)
 *                 self.state = 611
 */
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4752, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Statement_forContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4752, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4752, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4752, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4752, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_2) {
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_v_localctx);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4752, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4753
 *             elif token in [ModelicaParser.T__48]:
 *                 localctx = ModelicaParser.Statement_forContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 6)             # <<<<<<<<<<<<<<
 *                 self.state = 611
 *                 self.for_statement()
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4753, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_6};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4753, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_6};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4753, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4753, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_6);
            __Pyx_GIVEREF(__pyx_int_6);
            PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_6);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4753, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4754
 *                 localctx = ModelicaParser.Statement_forContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 6)
 *                 self.state = 611             # <<<<<<<<<<<<<<
 *                 self.for_statement()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_611) < 0) __PYX_ERR(0, 4754, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":4755
 *                 self.enterOuterAlt(localctx, 6)
 *                 self.state = 611
 *                 self.for_statement()             # <<<<<<<<<<<<<<
 *                 pass
 *             elif token in [ModelicaParser.T__51]:
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_for_statement); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4755, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4755, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4755, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4751
 *                 self.if_statement()
 *                 pass
 *             elif token in [ModelicaParser.T__48]:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Statement_forContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 6)
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":4757
 *                 self.for_statement()
 *                 pass
 *             elif token in [ModelicaParser.T__51]:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Statement_whileContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 7)
 */
        __Pyx_INCREF(__pyx_v_token);
        __pyx_t_1 = __pyx_v_token;
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4757, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__51); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4757, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4757, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 4757, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_12 = (__pyx_t_11 != 0);
        if (__pyx_t_12) {

          /* "pymola/generated/ModelicaParser.pyx":4758
 *                 pass
 *             elif token in [ModelicaParser.T__51]:
 *                 localctx = ModelicaParser.Statement_whileContext(self, localctx)             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 7)
 *                 self.state = 612
 */
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4758, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_Statement_whileContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4758, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4758, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4758, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4758, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_4) {
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __pyx_t_4 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_v_localctx);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4758, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4759
 *             elif token in [ModelicaParser.T__51]:
 *                 localctx = ModelicaParser.Statement_whileContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 7)             # <<<<<<<<<<<<<<
 *                 self.state = 612
 *                 self.while_statement()
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4759, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_7};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4759, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_7};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4759, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4759, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_7);
            __Pyx_GIVEREF(__pyx_int_7);
            PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_int_7);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4759, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4760
 *                 localctx = ModelicaParser.Statement_whileContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 7)
 *                 self.state = 612             # <<<<<<<<<<<<<<
 *                 self.while_statement()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_612) < 0) __PYX_ERR(0, 4760, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":4761
 *                 self.enterOuterAlt(localctx, 7)
 *                 self.state = 612
 *                 self.while_statement()             # <<<<<<<<<<<<<<
 *                 pass
 *             elif token in [ModelicaParser.T__52]:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_while_statement); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4761, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_4 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (__pyx_t_4) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4761, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4761, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4757
 *                 self.for_statement()
 *                 pass
 *             elif token in [ModelicaParser.T__51]:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Statement_whileContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 7)
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":4763
 *                 self.while_statement()
 *                 pass
 *             elif token in [ModelicaParser.T__52]:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Statement_whenContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 8)
 */
        __Pyx_INCREF(__pyx_v_token);
        __pyx_t_1 = __pyx_v_token;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4763, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__52); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4763, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4763, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 4763, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_11 = (__pyx_t_12 != 0);
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":4764
 *                 pass
 *             elif token in [ModelicaParser.T__52]:
 *                 localctx = ModelicaParser.Statement_whenContext(self, localctx)             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 8)
 *                 self.state = 613
 */
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4764, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Statement_whenContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4764, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4764, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4764, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4764, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_2) {
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_v_localctx);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4764, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4765
 *             elif token in [ModelicaParser.T__52]:
 *                 localctx = ModelicaParser.Statement_whenContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 8)             # <<<<<<<<<<<<<<
 *                 self.state = 613
 *                 self.when_statement()
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4765, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_8};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4765, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_8};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4765, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4765, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_8);
            __Pyx_GIVEREF(__pyx_int_8);
            PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_8);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4765, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4766
 *                 localctx = ModelicaParser.Statement_whenContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 8)
 *                 self.state = 613             # <<<<<<<<<<<<<<
 *                 self.when_statement()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_613) < 0) __PYX_ERR(0, 4766, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":4767
 *                 self.enterOuterAlt(localctx, 8)
 *                 self.state = 613
 *                 self.when_statement()             # <<<<<<<<<<<<<<
 *                 pass
 *             else:
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_when_statement); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4767, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4767, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4767, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4763
 *                 self.while_statement()
 *                 pass
 *             elif token in [ModelicaParser.T__52]:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Statement_whenContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 8)
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":4770
 *                 pass
 *             else:
 *                 raise NoViableAltException(self)             # <<<<<<<<<<<<<<
 * 
 *         except RecognitionException as re:
 */
        /*else*/ {
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_NoViableAltException); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4770, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4770, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4770, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4770, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4770, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4770, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_Raise(__pyx_t_1, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __PYX_ERR(0, 4770, __pyx_L6_error)
        }
        __pyx_L14:;

        /* "pymola/generated/ModelicaParser.pyx":4679
 *         self.enterRule(localctx, 84, self.RULE_statement_options)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.state = 614
 *             self._errHandler.sync(self)
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":4772
 *                 raise NoViableAltException(self)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4772, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.statement_options", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_4, &__pyx_t_3) < 0) __PYX_ERR(0, 4772, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __pyx_v_re = __pyx_t_4;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":4773
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 4773, __pyx_L23_error)

          /* "pymola/generated/ModelicaParser.pyx":4774
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4774, __pyx_L23_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4774, __pyx_L23_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4774, __pyx_L23_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4774, __pyx_L23_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4774, __pyx_L23_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_7) {
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_7); __pyx_t_7 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_13, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4774, __pyx_L23_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4775
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4775, __pyx_L23_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4775, __pyx_L23_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4775, __pyx_L23_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4775, __pyx_L23_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4775, __pyx_L23_error)
            __Pyx_GOTREF(__pyx_t_7);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4775, __pyx_L23_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":4772
 *                 raise NoViableAltException(self)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L24;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L23_error:;
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_6 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L8_except_error;
          }
          __pyx_L24:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":4679
 *         self.enterRule(localctx, 84, self.RULE_statement_options)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.state = 614
 *             self._errHandler.sync(self)
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":4777
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4777, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4777, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4777, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_14 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4777, __pyx_L30_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4777, __pyx_L30_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4777, __pyx_L30_error)
        }
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
      __pyx_L30_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":4778
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class StatementContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":4674
 * 
 * 
 *     def statement_options(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Statement_optionsContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.statement_options", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v_token);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4782
 *     class StatementContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16StatementContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_16StatementContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_16StatementContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16StatementContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 4782, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 4782, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4782, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.StatementContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_16StatementContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16StatementContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":4783
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 4783, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4783, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4783, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4783, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4783, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4784
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def statement_options(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 4784, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4782
 *     class StatementContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.StatementContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4786
 *             self.parser = parser
 * 
 *         def statement_options(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Statement_optionsContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16StatementContext_3statement_options(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_16StatementContext_3statement_options = {"statement_options", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_16StatementContext_3statement_options, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16StatementContext_3statement_options(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("statement_options (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_16StatementContext_2statement_options(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16StatementContext_2statement_options(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("statement_options", 0);

  /* "pymola/generated/ModelicaParser.pyx":4787
 * 
 *         def statement_options(self):
 *             return self.getTypedRuleContext(ModelicaParser.Statement_optionsContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4787, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4787, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Statement_optionsContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4787, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4787, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4787, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4787, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4787, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":4786
 *             self.parser = parser
 * 
 *         def statement_options(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Statement_optionsContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.StatementContext.statement_options", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4790
 * 
 * 
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16StatementContext_5comment(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_16StatementContext_5comment = {"comment", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_16StatementContext_5comment, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16StatementContext_5comment(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("comment (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_16StatementContext_4comment(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16StatementContext_4comment(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("comment", 0);

  /* "pymola/generated/ModelicaParser.pyx":4791
 * 
 *         def comment(self):
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4791, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4791, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_CommentContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4791, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4791, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4791, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4791, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4791, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":4790
 * 
 * 
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.StatementContext.comment", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4794
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_statement
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16StatementContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_16StatementContext_7getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_16StatementContext_7getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16StatementContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_16StatementContext_6getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16StatementContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":4795
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_statement             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_statement); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":4794
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_statement
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.StatementContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4797
 *             return ModelicaParser.RULE_statement
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStatement" ):
 *                 listener.enterStatement(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16StatementContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_16StatementContext_9enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_16StatementContext_9enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16StatementContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 4797, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 4797, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4797, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.StatementContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_16StatementContext_8enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16StatementContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":4798
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterStatement" ):             # <<<<<<<<<<<<<<
 *                 listener.enterStatement(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterStatement); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4798, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4799
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterStatement" ):
 *                 listener.enterStatement(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterStatement); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4799, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4799, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4799, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4799, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4799, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4799, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":4798
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterStatement" ):             # <<<<<<<<<<<<<<
 *                 listener.enterStatement(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4797
 *             return ModelicaParser.RULE_statement
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStatement" ):
 *                 listener.enterStatement(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.StatementContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4801
 *                 listener.enterStatement(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStatement" ):
 *                 listener.exitStatement(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16StatementContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_16StatementContext_11exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_16StatementContext_11exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16StatementContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 4801, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 4801, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4801, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.StatementContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_16StatementContext_10exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16StatementContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":4802
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitStatement" ):             # <<<<<<<<<<<<<<
 *                 listener.exitStatement(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitStatement); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4802, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4803
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitStatement" ):
 *                 listener.exitStatement(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitStatement); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4803, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4803, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4803, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4803, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4803, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4803, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":4802
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitStatement" ):             # <<<<<<<<<<<<<<
 *                 listener.exitStatement(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4801
 *                 listener.enterStatement(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStatement" ):
 *                 listener.exitStatement(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.StatementContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4805
 *                 listener.exitStatement(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStatement" ):
 *                 return visitor.visitStatement(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16StatementContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_16StatementContext_13accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_16StatementContext_13accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16StatementContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 4805, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 4805, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4805, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.StatementContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_16StatementContext_12accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16StatementContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":4806
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitStatement" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitStatement(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitStatement); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4806, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4807
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitStatement" ):
 *                 return visitor.visitStatement(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitStatement); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4807, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4807, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4807, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4807, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4807, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4807, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":4806
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitStatement" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitStatement(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4809
 *                 return visitor.visitStatement(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4809, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4809, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4809, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4809, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4809, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4809, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":4805
 *                 listener.exitStatement(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStatement" ):
 *                 return visitor.visitStatement(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.StatementContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4814
 * 
 * 
 *     def statement(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.StatementContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_89statement(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_89statement = {"statement", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_89statement, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_89statement(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("statement (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_88statement(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_88statement(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  char const *__pyx_t_20;
  __Pyx_RefNannySetupContext("statement", 0);

  /* "pymola/generated/ModelicaParser.pyx":4816
 *     def statement(self):
 * 
 *         localctx = ModelicaParser.StatementContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 86, self.RULE_statement)
 *         try:
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4816, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_StatementContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4816, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4816, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4816, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4816, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4816, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4816, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4816, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4817
 * 
 *         localctx = ModelicaParser.StatementContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 86, self.RULE_statement)             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4817, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_statement); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4817, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_86, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4817, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_86, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4817, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4817, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_86);
    __Pyx_GIVEREF(__pyx_int_86);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_86);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4817, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4818
 *         localctx = ModelicaParser.StatementContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 86, self.RULE_statement)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 616
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":4819
 *         self.enterRule(localctx, 86, self.RULE_statement)
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 616
 *             self.statement_options()
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4819, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4819, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4819, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4819, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4819, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":4820
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 616             # <<<<<<<<<<<<<<
 *             self.statement_options()
 *             self.state = 617
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_616) < 0) __PYX_ERR(0, 4820, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":4821
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 616
 *             self.statement_options()             # <<<<<<<<<<<<<<
 *             self.state = 617
 *             self.comment()
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_statement_options); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4821, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4821, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4821, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":4822
 *             self.state = 616
 *             self.statement_options()
 *             self.state = 617             # <<<<<<<<<<<<<<
 *             self.comment()
 *         except RecognitionException as re:
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_617) < 0) __PYX_ERR(0, 4822, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":4823
 *             self.statement_options()
 *             self.state = 617
 *             self.comment()             # <<<<<<<<<<<<<<
 *         except RecognitionException as re:
 *             localctx.exception = re
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_comment); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4823, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4823, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4823, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":4818
 *         localctx = ModelicaParser.StatementContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 86, self.RULE_statement)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 616
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":4824
 *             self.state = 617
 *             self.comment()
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4824, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.statement", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_3, &__pyx_t_7) < 0) __PYX_ERR(0, 4824, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_3);
        __pyx_v_re = __pyx_t_3;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":4825
 *             self.comment()
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 4825, __pyx_L19_error)

          /* "pymola/generated/ModelicaParser.pyx":4826
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4826, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4826, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4826, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4826, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_11 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 4826, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_11);
            if (__pyx_t_4) {
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_4); __pyx_t_4 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_11, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4826, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4827
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4827, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 4827, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_11)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4827, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_11)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4827, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4827, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4827, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":4824
 *             self.state = 617
 *             self.comment()
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L20;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L19_error:;
            __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_14);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __pyx_t_6 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_17);
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
            }
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
            __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
            goto __pyx_L8_except_error;
          }
          __pyx_L20:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":4818
 *         localctx = ModelicaParser.StatementContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 86, self.RULE_statement)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 616
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":4829
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4829, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4829, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4829, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_18, &__pyx_t_17);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_18);
      __Pyx_XGOTREF(__pyx_t_17);
      __pyx_t_12 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_20 = __pyx_filename;
      {
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4829, __pyx_L26_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4829, __pyx_L26_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4829, __pyx_L26_error)
        }
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_18, __pyx_t_17);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0;
      __pyx_lineno = __pyx_t_12; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_20;
      goto __pyx_L1_error;
      __pyx_L26_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_18, __pyx_t_17);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":4830
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class If_equationContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":4814
 * 
 * 
 *     def statement(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.StatementContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.statement", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4834
 *     class If_equationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18If_equationContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18If_equationContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18If_equationContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18If_equationContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 4834, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 4834, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4834, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.If_equationContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18If_equationContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18If_equationContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":4835
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 4835, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4835, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4835, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4835, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4835, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4835, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4835, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4835, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4836
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def expression(self, i:int=None):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 4836, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4834
 *     class If_equationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.If_equationContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4838
 *             self.parser = parser
 * 
 *         def expression(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18If_equationContext_3expression(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18If_equationContext_3expression = {"expression", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18If_equationContext_3expression, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18If_equationContext_3expression(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("expression (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "expression") < 0)) __PYX_ERR(0, 4838, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("expression", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4838, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.If_equationContext.expression", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18If_equationContext_2expression(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18If_equationContext_2expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("expression", 0);

  /* "pymola/generated/ModelicaParser.pyx":4839
 * 
 *         def expression(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4840
 *         def expression(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.ExpressionContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4840, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4840, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_ExpressionContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4840, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4840, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4840, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4840, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4840, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4840, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":4839
 * 
 *         def expression(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4842
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.ExpressionContext,i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4842, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4842, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_ExpressionContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4842, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4842, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4842, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4842, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4842, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":4838
 *             self.parser = parser
 * 
 *         def expression(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.If_equationContext.expression", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4845
 * 
 * 
 *         def equation(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.EquationContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18If_equationContext_5equation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18If_equationContext_5equation = {"equation", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18If_equationContext_5equation, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18If_equationContext_5equation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("equation (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "equation") < 0)) __PYX_ERR(0, 4845, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("equation", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4845, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.If_equationContext.equation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18If_equationContext_4equation(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18If_equationContext_4equation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("equation", 0);

  /* "pymola/generated/ModelicaParser.pyx":4846
 * 
 *         def equation(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.EquationContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4847
 *         def equation(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.EquationContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.EquationContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4847, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4847, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_EquationContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4847, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4847, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4847, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4847, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4847, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4847, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":4846
 * 
 *         def equation(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.EquationContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4849
 *                 return self.getTypedRuleContexts(ModelicaParser.EquationContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.EquationContext,i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4849, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4849, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_EquationContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4849, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4849, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4849, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4849, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4849, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":4845
 * 
 * 
 *         def equation(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.EquationContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.If_equationContext.equation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4852
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_if_equation
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18If_equationContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18If_equationContext_7getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18If_equationContext_7getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18If_equationContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18If_equationContext_6getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18If_equationContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":4853
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_if_equation             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4853, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_if_equation); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4853, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":4852
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_if_equation
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.If_equationContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4855
 *             return ModelicaParser.RULE_if_equation
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterIf_equation" ):
 *                 listener.enterIf_equation(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18If_equationContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18If_equationContext_9enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18If_equationContext_9enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18If_equationContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 4855, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 4855, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4855, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.If_equationContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18If_equationContext_8enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18If_equationContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":4856
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterIf_equation" ):             # <<<<<<<<<<<<<<
 *                 listener.enterIf_equation(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterIf_equation); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4856, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4857
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterIf_equation" ):
 *                 listener.enterIf_equation(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterIf_equation); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4857, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4857, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4857, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4857, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4857, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4857, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":4856
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterIf_equation" ):             # <<<<<<<<<<<<<<
 *                 listener.enterIf_equation(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4855
 *             return ModelicaParser.RULE_if_equation
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterIf_equation" ):
 *                 listener.enterIf_equation(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.If_equationContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4859
 *                 listener.enterIf_equation(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitIf_equation" ):
 *                 listener.exitIf_equation(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18If_equationContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18If_equationContext_11exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18If_equationContext_11exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18If_equationContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 4859, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 4859, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4859, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.If_equationContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18If_equationContext_10exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18If_equationContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":4860
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitIf_equation" ):             # <<<<<<<<<<<<<<
 *                 listener.exitIf_equation(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitIf_equation); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4860, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4861
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitIf_equation" ):
 *                 listener.exitIf_equation(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitIf_equation); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4861, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4861, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4861, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4861, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4861, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4861, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":4860
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitIf_equation" ):             # <<<<<<<<<<<<<<
 *                 listener.exitIf_equation(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4859
 *                 listener.enterIf_equation(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitIf_equation" ):
 *                 listener.exitIf_equation(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.If_equationContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4863
 *                 listener.exitIf_equation(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitIf_equation" ):
 *                 return visitor.visitIf_equation(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18If_equationContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18If_equationContext_13accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18If_equationContext_13accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18If_equationContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 4863, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 4863, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4863, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.If_equationContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18If_equationContext_12accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18If_equationContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":4864
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitIf_equation" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitIf_equation(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitIf_equation); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4864, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4865
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitIf_equation" ):
 *                 return visitor.visitIf_equation(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitIf_equation); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4865, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4865, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4865, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4865, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4865, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4865, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":4864
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitIf_equation" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitIf_equation(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4867
 *                 return visitor.visitIf_equation(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4867, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4867, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4867, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4867, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4867, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4867, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":4863
 *                 listener.exitIf_equation(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitIf_equation" ):
 *                 return visitor.visitIf_equation(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.If_equationContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4872
 * 
 * 
 *     def if_equation(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.If_equationContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_91if_equation(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_91if_equation = {"if_equation", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_91if_equation, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_91if_equation(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("if_equation (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_90if_equation(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_90if_equation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__alt = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  char const *__pyx_t_22;
  __Pyx_RefNannySetupContext("if_equation", 0);

  /* "pymola/generated/ModelicaParser.pyx":4874
 *     def if_equation(self):
 * 
 *         localctx = ModelicaParser.If_equationContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 88, self.RULE_if_equation)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4874, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_If_equationContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4874, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4874, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4874, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4874, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4874, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4874, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4874, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4875
 * 
 *         localctx = ModelicaParser.If_equationContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 88, self.RULE_if_equation)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4875, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_if_equation); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4875, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_88, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4875, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_88, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4875, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4875, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_88);
    __Pyx_GIVEREF(__pyx_int_88);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_88);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4875, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4876
 *         localctx = ModelicaParser.If_equationContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 88, self.RULE_if_equation)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 4876, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4877
 *         self.enterRule(localctx, 88, self.RULE_if_equation)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 619
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":4878
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 619
 *             self.match(ModelicaParser.T__39)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4878, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4878, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4878, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4878, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4878, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":4879
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 619             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__39)
 *             self.state = 620
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_619) < 0) __PYX_ERR(0, 4879, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":4880
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 619
 *             self.match(ModelicaParser.T__39)             # <<<<<<<<<<<<<<
 *             self.state = 620
 *             self.expression()
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4880, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4880, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__39); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4880, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4880, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4880, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4880, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4880, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4880, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":4881
 *             self.state = 619
 *             self.match(ModelicaParser.T__39)
 *             self.state = 620             # <<<<<<<<<<<<<<
 *             self.expression()
 *             self.state = 621
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_620) < 0) __PYX_ERR(0, 4881, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":4882
 *             self.match(ModelicaParser.T__39)
 *             self.state = 620
 *             self.expression()             # <<<<<<<<<<<<<<
 *             self.state = 621
 *             self.match(ModelicaParser.T__45)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expression); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4882, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4882, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4882, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":4883
 *             self.state = 620
 *             self.expression()
 *             self.state = 621             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__45)
 *             self.state = 627
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_621) < 0) __PYX_ERR(0, 4883, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":4884
 *             self.expression()
 *             self.state = 621
 *             self.match(ModelicaParser.T__45)             # <<<<<<<<<<<<<<
 *             self.state = 627
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4884, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4884, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__45); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4884, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4884, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4884, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4884, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4884, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4); __pyx_t_4 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4884, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":4885
 *             self.state = 621
 *             self.match(ModelicaParser.T__45)
 *             self.state = 627             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _alt = self._interp.adaptivePredict(self._input,84,self._ctx)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_627) < 0) __PYX_ERR(0, 4885, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":4886
 *             self.match(ModelicaParser.T__45)
 *             self.state = 627
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _alt = self._interp.adaptivePredict(self._input,84,self._ctx)
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4886, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4886, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4886, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4886, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4886, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4886, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4886, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":4887
 *             self.state = 627
 *             self._errHandler.sync(self)
 *             _alt = self._interp.adaptivePredict(self._input,84,self._ctx)             # <<<<<<<<<<<<<<
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                 if _alt==1:
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_interp); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4887, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_adaptivePredict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4887, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4887, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4887, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_2)) {
          PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_t_7, __pyx_int_84, __pyx_t_3};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4887, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
          PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_t_7, __pyx_int_84, __pyx_t_3};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4887, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4887, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_7);
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_t_7);
          __Pyx_INCREF(__pyx_int_84);
          __Pyx_GIVEREF(__pyx_int_84);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_int_84);
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_6, __pyx_t_3);
          __pyx_t_7 = 0;
          __pyx_t_3 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4887, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_v__alt = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":4888
 *             self._errHandler.sync(self)
 *             _alt = self._interp.adaptivePredict(self._input,84,self._ctx)
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:             # <<<<<<<<<<<<<<
 *                 if _alt==1:
 *                     self.state = 622
 */
        while (1) {
          __pyx_t_1 = PyObject_RichCompare(__pyx_v__alt, __pyx_int_2, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4888, __pyx_L6_error)
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 4888, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (__pyx_t_12) {
          } else {
            __pyx_t_11 = __pyx_t_12;
            goto __pyx_L16_bool_binop_done;
          }
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ATN); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4888, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_INVALID_ALT_NUMBER); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4888, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyObject_RichCompare(__pyx_v__alt, __pyx_t_2, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4888, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 4888, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_11 = __pyx_t_12;
          __pyx_L16_bool_binop_done:;
          if (!__pyx_t_11) break;

          /* "pymola/generated/ModelicaParser.pyx":4889
 *             _alt = self._interp.adaptivePredict(self._input,84,self._ctx)
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                 if _alt==1:             # <<<<<<<<<<<<<<
 *                     self.state = 622
 *                     self.equation()
 */
          __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v__alt, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4889, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 4889, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (__pyx_t_11) {

            /* "pymola/generated/ModelicaParser.pyx":4890
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                 if _alt==1:
 *                     self.state = 622             # <<<<<<<<<<<<<<
 *                     self.equation()
 *                     self.state = 623
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_622) < 0) __PYX_ERR(0, 4890, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":4891
 *                 if _alt==1:
 *                     self.state = 622
 *                     self.equation()             # <<<<<<<<<<<<<<
 *                     self.state = 623
 *                     self.match(ModelicaParser.T__0)
 */
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_equation); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4891, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_5 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_5)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_5);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
              }
            }
            if (__pyx_t_5) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4891, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else {
              __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4891, __pyx_L6_error)
            }
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":4892
 *                     self.state = 622
 *                     self.equation()
 *                     self.state = 623             # <<<<<<<<<<<<<<
 *                     self.match(ModelicaParser.T__0)
 *                 self.state = 629
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_623) < 0) __PYX_ERR(0, 4892, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":4893
 *                     self.equation()
 *                     self.state = 623
 *                     self.match(ModelicaParser.T__0)             # <<<<<<<<<<<<<<
 *                 self.state = 629
 *                 self._errHandler.sync(self)
 */
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4893, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4893, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_T__0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4893, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_5 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_5)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_5);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
              }
            }
            if (!__pyx_t_5) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4893, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_2)) {
                PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4893, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
                PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4893, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              } else
              #endif
              {
                __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4893, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_7);
                __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
                __Pyx_GIVEREF(__pyx_t_3);
                PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_3);
                __pyx_t_3 = 0;
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4893, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":4889
 *             _alt = self._interp.adaptivePredict(self._input,84,self._ctx)
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                 if _alt==1:             # <<<<<<<<<<<<<<
 *                     self.state = 622
 *                     self.equation()
 */
          }

          /* "pymola/generated/ModelicaParser.pyx":4894
 *                     self.state = 623
 *                     self.match(ModelicaParser.T__0)
 *                 self.state = 629             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _alt = self._interp.adaptivePredict(self._input,84,self._ctx)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_629) < 0) __PYX_ERR(0, 4894, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":4895
 *                     self.match(ModelicaParser.T__0)
 *                 self.state = 629
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _alt = self._interp.adaptivePredict(self._input,84,self._ctx)
 * 
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4895, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4895, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4895, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4895, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4895, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4895, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4895, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4896
 *                 self.state = 629
 *                 self._errHandler.sync(self)
 *                 _alt = self._interp.adaptivePredict(self._input,84,self._ctx)             # <<<<<<<<<<<<<<
 * 
 *             self.state = 643
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_interp); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4896, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_adaptivePredict); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4896, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4896, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4896, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_t_7, __pyx_int_84, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4896, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_t_7, __pyx_int_84, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4896, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4896, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_GIVEREF(__pyx_t_7);
            PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_t_7);
            __Pyx_INCREF(__pyx_int_84);
            __Pyx_GIVEREF(__pyx_int_84);
            PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_int_84);
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_4, 2+__pyx_t_6, __pyx_t_2);
            __pyx_t_7 = 0;
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4896, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF_SET(__pyx_v__alt, __pyx_t_1);
          __pyx_t_1 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":4898
 *                 _alt = self._interp.adaptivePredict(self._input,84,self._ctx)
 * 
 *             self.state = 643             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_643) < 0) __PYX_ERR(0, 4898, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":4899
 * 
 *             self.state = 643
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             while _la==ModelicaParser.T__46:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4899, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4899, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4899, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4899, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4899, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4899, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4899, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":4900
 *             self.state = 643
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             while _la==ModelicaParser.T__46:
 *                 self.state = 630
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4900, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4900, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__901, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4900, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_v__la = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":4901
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             while _la==ModelicaParser.T__46:             # <<<<<<<<<<<<<<
 *                 self.state = 630
 *                 self.match(ModelicaParser.T__46)
 */
        while (1) {
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4901, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__46); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4901, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4901, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 4901, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (!__pyx_t_11) break;

          /* "pymola/generated/ModelicaParser.pyx":4902
 *             _la = self._input.LA(1)
 *             while _la==ModelicaParser.T__46:
 *                 self.state = 630             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__46)
 *                 self.state = 631
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_630) < 0) __PYX_ERR(0, 4902, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":4903
 *             while _la==ModelicaParser.T__46:
 *                 self.state = 630
 *                 self.match(ModelicaParser.T__46)             # <<<<<<<<<<<<<<
 *                 self.state = 631
 *                 self.expression()
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4903, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4903, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__46); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4903, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4903, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4903, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4903, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4903, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4903, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4904
 *                 self.state = 630
 *                 self.match(ModelicaParser.T__46)
 *                 self.state = 631             # <<<<<<<<<<<<<<
 *                 self.expression()
 *                 self.state = 632
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_631) < 0) __PYX_ERR(0, 4904, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":4905
 *                 self.match(ModelicaParser.T__46)
 *                 self.state = 631
 *                 self.expression()             # <<<<<<<<<<<<<<
 *                 self.state = 632
 *                 self.match(ModelicaParser.T__45)
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expression); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4905, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4905, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4905, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4906
 *                 self.state = 631
 *                 self.expression()
 *                 self.state = 632             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__45)
 *                 self.state = 638
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_632) < 0) __PYX_ERR(0, 4906, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":4907
 *                 self.expression()
 *                 self.state = 632
 *                 self.match(ModelicaParser.T__45)             # <<<<<<<<<<<<<<
 *                 self.state = 638
 *                 self._errHandler.sync(self)
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4907, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4907, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__45); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4907, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4907, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4907, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4907, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4907, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_7); __pyx_t_7 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4907, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4908
 *                 self.state = 632
 *                 self.match(ModelicaParser.T__45)
 *                 self.state = 638             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _alt = self._interp.adaptivePredict(self._input,85,self._ctx)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_638) < 0) __PYX_ERR(0, 4908, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":4909
 *                 self.match(ModelicaParser.T__45)
 *                 self.state = 638
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _alt = self._interp.adaptivePredict(self._input,85,self._ctx)
 *                 while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4909, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4909, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_4) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4909, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4909, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4909, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4909, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __pyx_t_4 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4909, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4910
 *                 self.state = 638
 *                 self._errHandler.sync(self)
 *                 _alt = self._interp.adaptivePredict(self._input,85,self._ctx)             # <<<<<<<<<<<<<<
 *                 while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                     if _alt==1:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_interp); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4910, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_adaptivePredict); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4910, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4910, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4910, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[4] = {__pyx_t_7, __pyx_t_2, __pyx_int_85, __pyx_t_4};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4910, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[4] = {__pyx_t_7, __pyx_t_2, __pyx_int_85, __pyx_t_4};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4910, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          } else
          #endif
          {
            __pyx_t_5 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4910, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            if (__pyx_t_7) {
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
            }
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_t_2);
            __Pyx_INCREF(__pyx_int_85);
            __Pyx_GIVEREF(__pyx_int_85);
            PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_int_85);
            __Pyx_GIVEREF(__pyx_t_4);
            PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_6, __pyx_t_4);
            __pyx_t_2 = 0;
            __pyx_t_4 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4910, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF_SET(__pyx_v__alt, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4911
 *                 self._errHandler.sync(self)
 *                 _alt = self._interp.adaptivePredict(self._input,85,self._ctx)
 *                 while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:             # <<<<<<<<<<<<<<
 *                     if _alt==1:
 *                         self.state = 633
 */
          while (1) {
            __pyx_t_1 = PyObject_RichCompare(__pyx_v__alt, __pyx_int_2, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4911, __pyx_L6_error)
            __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 4911, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (__pyx_t_12) {
            } else {
              __pyx_t_11 = __pyx_t_12;
              goto __pyx_L23_bool_binop_done;
            }
            __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ATN); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4911, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_INVALID_ALT_NUMBER); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4911, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = PyObject_RichCompare(__pyx_v__alt, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4911, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 4911, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_11 = __pyx_t_12;
            __pyx_L23_bool_binop_done:;
            if (!__pyx_t_11) break;

            /* "pymola/generated/ModelicaParser.pyx":4912
 *                 _alt = self._interp.adaptivePredict(self._input,85,self._ctx)
 *                 while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                     if _alt==1:             # <<<<<<<<<<<<<<
 *                         self.state = 633
 *                         self.equation()
 */
            __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v__alt, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4912, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 4912, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (__pyx_t_11) {

              /* "pymola/generated/ModelicaParser.pyx":4913
 *                 while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                     if _alt==1:
 *                         self.state = 633             # <<<<<<<<<<<<<<
 *                         self.equation()
 *                         self.state = 634
 */
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_633) < 0) __PYX_ERR(0, 4913, __pyx_L6_error)

              /* "pymola/generated/ModelicaParser.pyx":4914
 *                     if _alt==1:
 *                         self.state = 633
 *                         self.equation()             # <<<<<<<<<<<<<<
 *                         self.state = 634
 *                         self.match(ModelicaParser.T__0)
 */
              __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_equation); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4914, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __pyx_t_5 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
                __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
                if (likely(__pyx_t_5)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                  __Pyx_INCREF(__pyx_t_5);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_3, function);
                }
              }
              if (__pyx_t_5) {
                __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4914, __pyx_L6_error)
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              } else {
                __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4914, __pyx_L6_error)
              }
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "pymola/generated/ModelicaParser.pyx":4915
 *                         self.state = 633
 *                         self.equation()
 *                         self.state = 634             # <<<<<<<<<<<<<<
 *                         self.match(ModelicaParser.T__0)
 *                     self.state = 640
 */
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_634) < 0) __PYX_ERR(0, 4915, __pyx_L6_error)

              /* "pymola/generated/ModelicaParser.pyx":4916
 *                         self.equation()
 *                         self.state = 634
 *                         self.match(ModelicaParser.T__0)             # <<<<<<<<<<<<<<
 *                     self.state = 640
 *                     self._errHandler.sync(self)
 */
              __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4916, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4916, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_T__0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4916, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __pyx_t_5 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
                __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
                if (likely(__pyx_t_5)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                  __Pyx_INCREF(__pyx_t_5);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_3, function);
                }
              }
              if (!__pyx_t_5) {
                __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4916, __pyx_L6_error)
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else {
                #if CYTHON_FAST_PYCALL
                if (PyFunction_Check(__pyx_t_3)) {
                  PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_4};
                  __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4916, __pyx_L6_error)
                  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                  __Pyx_GOTREF(__pyx_t_1);
                  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                } else
                #endif
                #if CYTHON_FAST_PYCCALL
                if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
                  PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_4};
                  __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4916, __pyx_L6_error)
                  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                  __Pyx_GOTREF(__pyx_t_1);
                  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                } else
                #endif
                {
                  __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4916, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_2);
                  __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5); __pyx_t_5 = NULL;
                  __Pyx_GIVEREF(__pyx_t_4);
                  PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_4);
                  __pyx_t_4 = 0;
                  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4916, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_1);
                  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                }
              }
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "pymola/generated/ModelicaParser.pyx":4912
 *                 _alt = self._interp.adaptivePredict(self._input,85,self._ctx)
 *                 while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                     if _alt==1:             # <<<<<<<<<<<<<<
 *                         self.state = 633
 *                         self.equation()
 */
            }

            /* "pymola/generated/ModelicaParser.pyx":4917
 *                         self.state = 634
 *                         self.match(ModelicaParser.T__0)
 *                     self.state = 640             # <<<<<<<<<<<<<<
 *                     self._errHandler.sync(self)
 *                     _alt = self._interp.adaptivePredict(self._input,85,self._ctx)
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_640) < 0) __PYX_ERR(0, 4917, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":4918
 *                         self.match(ModelicaParser.T__0)
 *                     self.state = 640
 *                     self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                     _alt = self._interp.adaptivePredict(self._input,85,self._ctx)
 * 
 */
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4918, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4918, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_3)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_3);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
              }
            }
            if (!__pyx_t_3) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4918, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_2)) {
                PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4918, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
                PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4918, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else
              #endif
              {
                __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4918, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __pyx_t_3 = NULL;
                __Pyx_INCREF(__pyx_v_self);
                __Pyx_GIVEREF(__pyx_v_self);
                PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_self);
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4918, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":4919
 *                     self.state = 640
 *                     self._errHandler.sync(self)
 *                     _alt = self._interp.adaptivePredict(self._input,85,self._ctx)             # <<<<<<<<<<<<<<
 * 
 *                 self.state = 645
 */
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_interp); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4919, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_adaptivePredict); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4919, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4919, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4919, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_5 = NULL;
            __pyx_t_6 = 0;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
              if (likely(__pyx_t_5)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                __Pyx_INCREF(__pyx_t_5);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_4, function);
                __pyx_t_6 = 1;
              }
            }
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_t_2, __pyx_int_85, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4919, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_t_2, __pyx_int_85, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4919, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4919, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              if (__pyx_t_5) {
                __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
              }
              __Pyx_GIVEREF(__pyx_t_2);
              PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_t_2);
              __Pyx_INCREF(__pyx_int_85);
              __Pyx_GIVEREF(__pyx_int_85);
              PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_85);
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_3);
              __pyx_t_2 = 0;
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4919, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            }
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF_SET(__pyx_v__alt, __pyx_t_1);
            __pyx_t_1 = 0;
          }

          /* "pymola/generated/ModelicaParser.pyx":4921
 *                     _alt = self._interp.adaptivePredict(self._input,85,self._ctx)
 * 
 *                 self.state = 645             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_645) < 0) __PYX_ERR(0, 4921, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":4922
 * 
 *                 self.state = 645
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 * 
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4922, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4922, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_4) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4922, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4922, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4922, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4922, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __pyx_t_4 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4922, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4923
 *                 self.state = 645
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *             self.state = 655
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4923, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4923, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__902, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4923, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_1);
          __pyx_t_1 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":4925
 *                 _la = self._input.LA(1)
 * 
 *             self.state = 655             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_655) < 0) __PYX_ERR(0, 4925, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":4926
 * 
 *             self.state = 655
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__47:
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4926, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_sync); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4926, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4926, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4926, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4926, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4926, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4926, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":4927
 *             self.state = 655
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.T__47:
 *                 self.state = 646
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4927, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4927, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__903, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4927, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":4928
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__47:             # <<<<<<<<<<<<<<
 *                 self.state = 646
 *                 self.match(ModelicaParser.T__47)
 */
        __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4928, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__47); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4928, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4928, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 4928, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":4929
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__47:
 *                 self.state = 646             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__47)
 *                 self.state = 652
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_646) < 0) __PYX_ERR(0, 4929, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":4930
 *             if _la==ModelicaParser.T__47:
 *                 self.state = 646
 *                 self.match(ModelicaParser.T__47)             # <<<<<<<<<<<<<<
 *                 self.state = 652
 *                 self._errHandler.sync(self)
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4930, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4930, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__47); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4930, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (!__pyx_t_4) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4930, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4930, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4930, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            {
              __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4930, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
              __Pyx_GIVEREF(__pyx_t_7);
              PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_7);
              __pyx_t_7 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4930, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4931
 *                 self.state = 646
 *                 self.match(ModelicaParser.T__47)
 *                 self.state = 652             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _alt = self._interp.adaptivePredict(self._input,87,self._ctx)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_652) < 0) __PYX_ERR(0, 4931, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":4932
 *                 self.match(ModelicaParser.T__47)
 *                 self.state = 652
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _alt = self._interp.adaptivePredict(self._input,87,self._ctx)
 *                 while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4932, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4932, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_3) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4932, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4932, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4932, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4932, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4932, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4933
 *                 self.state = 652
 *                 self._errHandler.sync(self)
 *                 _alt = self._interp.adaptivePredict(self._input,87,self._ctx)             # <<<<<<<<<<<<<<
 *                 while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                     if _alt==1:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_interp); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4933, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_adaptivePredict); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4933, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4933, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4933, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_4 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_t_2, __pyx_int_87, __pyx_t_3};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4933, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_t_2, __pyx_int_87, __pyx_t_3};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4933, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          } else
          #endif
          {
            __pyx_t_5 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4933, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            if (__pyx_t_4) {
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
            }
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_t_2);
            __Pyx_INCREF(__pyx_int_87);
            __Pyx_GIVEREF(__pyx_int_87);
            PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_int_87);
            __Pyx_GIVEREF(__pyx_t_3);
            PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_6, __pyx_t_3);
            __pyx_t_2 = 0;
            __pyx_t_3 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4933, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF_SET(__pyx_v__alt, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4934
 *                 self._errHandler.sync(self)
 *                 _alt = self._interp.adaptivePredict(self._input,87,self._ctx)
 *                 while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:             # <<<<<<<<<<<<<<
 *                     if _alt==1:
 *                         self.state = 647
 */
          while (1) {
            __pyx_t_1 = PyObject_RichCompare(__pyx_v__alt, __pyx_int_2, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4934, __pyx_L6_error)
            __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 4934, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (__pyx_t_12) {
            } else {
              __pyx_t_11 = __pyx_t_12;
              goto __pyx_L29_bool_binop_done;
            }
            __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ATN); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4934, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_INVALID_ALT_NUMBER); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4934, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = PyObject_RichCompare(__pyx_v__alt, __pyx_t_7, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4934, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 4934, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_11 = __pyx_t_12;
            __pyx_L29_bool_binop_done:;
            if (!__pyx_t_11) break;

            /* "pymola/generated/ModelicaParser.pyx":4935
 *                 _alt = self._interp.adaptivePredict(self._input,87,self._ctx)
 *                 while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                     if _alt==1:             # <<<<<<<<<<<<<<
 *                         self.state = 647
 *                         self.equation()
 */
            __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v__alt, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4935, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 4935, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (__pyx_t_11) {

              /* "pymola/generated/ModelicaParser.pyx":4936
 *                 while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                     if _alt==1:
 *                         self.state = 647             # <<<<<<<<<<<<<<
 *                         self.equation()
 *                         self.state = 648
 */
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_647) < 0) __PYX_ERR(0, 4936, __pyx_L6_error)

              /* "pymola/generated/ModelicaParser.pyx":4937
 *                     if _alt==1:
 *                         self.state = 647
 *                         self.equation()             # <<<<<<<<<<<<<<
 *                         self.state = 648
 *                         self.match(ModelicaParser.T__0)
 */
              __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_equation); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4937, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __pyx_t_5 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
                __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
                if (likely(__pyx_t_5)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
                  __Pyx_INCREF(__pyx_t_5);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_7, function);
                }
              }
              if (__pyx_t_5) {
                __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4937, __pyx_L6_error)
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              } else {
                __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4937, __pyx_L6_error)
              }
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "pymola/generated/ModelicaParser.pyx":4938
 *                         self.state = 647
 *                         self.equation()
 *                         self.state = 648             # <<<<<<<<<<<<<<
 *                         self.match(ModelicaParser.T__0)
 *                     self.state = 654
 */
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_648) < 0) __PYX_ERR(0, 4938, __pyx_L6_error)

              /* "pymola/generated/ModelicaParser.pyx":4939
 *                         self.equation()
 *                         self.state = 648
 *                         self.match(ModelicaParser.T__0)             # <<<<<<<<<<<<<<
 *                     self.state = 654
 *                     self._errHandler.sync(self)
 */
              __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4939, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4939, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_T__0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4939, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __pyx_t_5 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
                __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
                if (likely(__pyx_t_5)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
                  __Pyx_INCREF(__pyx_t_5);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_7, function);
                }
              }
              if (!__pyx_t_5) {
                __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4939, __pyx_L6_error)
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else {
                #if CYTHON_FAST_PYCALL
                if (PyFunction_Check(__pyx_t_7)) {
                  PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
                  __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4939, __pyx_L6_error)
                  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                  __Pyx_GOTREF(__pyx_t_1);
                  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                } else
                #endif
                #if CYTHON_FAST_PYCCALL
                if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
                  PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
                  __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4939, __pyx_L6_error)
                  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                  __Pyx_GOTREF(__pyx_t_1);
                  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                } else
                #endif
                {
                  __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4939, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_2);
                  __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5); __pyx_t_5 = NULL;
                  __Pyx_GIVEREF(__pyx_t_3);
                  PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_3);
                  __pyx_t_3 = 0;
                  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4939, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_1);
                  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                }
              }
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "pymola/generated/ModelicaParser.pyx":4935
 *                 _alt = self._interp.adaptivePredict(self._input,87,self._ctx)
 *                 while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                     if _alt==1:             # <<<<<<<<<<<<<<
 *                         self.state = 647
 *                         self.equation()
 */
            }

            /* "pymola/generated/ModelicaParser.pyx":4940
 *                         self.state = 648
 *                         self.match(ModelicaParser.T__0)
 *                     self.state = 654             # <<<<<<<<<<<<<<
 *                     self._errHandler.sync(self)
 *                     _alt = self._interp.adaptivePredict(self._input,87,self._ctx)
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_654) < 0) __PYX_ERR(0, 4940, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":4941
 *                         self.match(ModelicaParser.T__0)
 *                     self.state = 654
 *                     self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                     _alt = self._interp.adaptivePredict(self._input,87,self._ctx)
 * 
 */
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4941, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4941, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __pyx_t_7 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_7)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_7);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
              }
            }
            if (!__pyx_t_7) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4941, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_2)) {
                PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4941, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
                PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4941, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else
              #endif
              {
                __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4941, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
                __Pyx_INCREF(__pyx_v_self);
                __Pyx_GIVEREF(__pyx_v_self);
                PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4941, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":4942
 *                     self.state = 654
 *                     self._errHandler.sync(self)
 *                     _alt = self._interp.adaptivePredict(self._input,87,self._ctx)             # <<<<<<<<<<<<<<
 * 
 * 
 */
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_interp); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4942, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_adaptivePredict); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4942, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4942, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4942, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_5 = NULL;
            __pyx_t_6 = 0;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
              if (likely(__pyx_t_5)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                __Pyx_INCREF(__pyx_t_5);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_3, function);
                __pyx_t_6 = 1;
              }
            }
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_t_2, __pyx_int_87, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4942, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_t_2, __pyx_int_87, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4942, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4942, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              if (__pyx_t_5) {
                __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
              }
              __Pyx_GIVEREF(__pyx_t_2);
              PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_t_2);
              __Pyx_INCREF(__pyx_int_87);
              __Pyx_GIVEREF(__pyx_int_87);
              PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_int_87);
              __Pyx_GIVEREF(__pyx_t_7);
              PyTuple_SET_ITEM(__pyx_t_4, 2+__pyx_t_6, __pyx_t_7);
              __pyx_t_2 = 0;
              __pyx_t_7 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4942, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF_SET(__pyx_v__alt, __pyx_t_1);
            __pyx_t_1 = 0;
          }

          /* "pymola/generated/ModelicaParser.pyx":4928
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__47:             # <<<<<<<<<<<<<<
 *                 self.state = 646
 *                 self.match(ModelicaParser.T__47)
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":4946
 * 
 * 
 *             self.state = 657             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__13)
 *             self.state = 658
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_657) < 0) __PYX_ERR(0, 4946, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":4947
 * 
 *             self.state = 657
 *             self.match(ModelicaParser.T__13)             # <<<<<<<<<<<<<<
 *             self.state = 658
 *             self.match(ModelicaParser.T__39)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4947, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4947, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__13); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4947, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4947, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_7};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4947, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_7};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4947, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4947, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
            __Pyx_GIVEREF(__pyx_t_7);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_7);
            __pyx_t_7 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4947, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":4948
 *             self.state = 657
 *             self.match(ModelicaParser.T__13)
 *             self.state = 658             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__39)
 *         except RecognitionException as re:
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_658) < 0) __PYX_ERR(0, 4948, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":4949
 *             self.match(ModelicaParser.T__13)
 *             self.state = 658
 *             self.match(ModelicaParser.T__39)             # <<<<<<<<<<<<<<
 *         except RecognitionException as re:
 *             localctx.exception = re
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4949, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4949, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__39); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4949, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_2) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4949, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_7};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4949, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_7};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4949, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4949, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
            __Pyx_GIVEREF(__pyx_t_7);
            PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_7);
            __pyx_t_7 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4949, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":4877
 *         self.enterRule(localctx, 88, self.RULE_if_equation)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 619
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":4950
 *             self.state = 658
 *             self.match(ModelicaParser.T__39)
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4950, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.if_equation", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_3, &__pyx_t_4) < 0) __PYX_ERR(0, 4950, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __pyx_v_re = __pyx_t_3;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":4951
 *             self.match(ModelicaParser.T__39)
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 4951, __pyx_L37_error)

          /* "pymola/generated/ModelicaParser.pyx":4952
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4952, __pyx_L37_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4952, __pyx_L37_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4952, __pyx_L37_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4952, __pyx_L37_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4952, __pyx_L37_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_2) {
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_2); __pyx_t_2 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_13, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4952, __pyx_L37_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "pymola/generated/ModelicaParser.pyx":4953
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4953, __pyx_L37_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4953, __pyx_L37_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4953, __pyx_L37_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4953, __pyx_L37_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4953, __pyx_L37_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_2, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4953, __pyx_L37_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":4950
 *             self.state = 658
 *             self.match(ModelicaParser.T__39)
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L38;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L37_error:;
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_6 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L8_except_error;
          }
          __pyx_L38:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":4877
 *         self.enterRule(localctx, 88, self.RULE_if_equation)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 619
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":4955
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4955, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4955, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4955, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_14 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4955, __pyx_L44_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4955, __pyx_L44_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4955, __pyx_L44_error)
        }
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
      __pyx_L44_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":4956
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class If_statementContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":4872
 * 
 * 
 *     def if_equation(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.If_equationContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.if_equation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__alt);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4960
 *     class If_statementContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19If_statementContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19If_statementContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19If_statementContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19If_statementContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 4960, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 4960, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4960, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.If_statementContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19If_statementContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19If_statementContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":4961
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 4961, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4961, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4961, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4961, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4961, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4962
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def expression(self, i:int=None):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 4962, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4960
 *     class If_statementContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.If_statementContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4964
 *             self.parser = parser
 * 
 *         def expression(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19If_statementContext_3expression(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19If_statementContext_3expression = {"expression", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19If_statementContext_3expression, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19If_statementContext_3expression(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("expression (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "expression") < 0)) __PYX_ERR(0, 4964, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("expression", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4964, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.If_statementContext.expression", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19If_statementContext_2expression(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19If_statementContext_2expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("expression", 0);

  /* "pymola/generated/ModelicaParser.pyx":4965
 * 
 *         def expression(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4966
 *         def expression(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.ExpressionContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4966, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4966, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_ExpressionContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4966, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4966, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4966, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4966, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4966, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4966, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":4965
 * 
 *         def expression(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4968
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.ExpressionContext,i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4968, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4968, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_ExpressionContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4968, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4968, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4968, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4968, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4968, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":4964
 *             self.parser = parser
 * 
 *         def expression(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.If_statementContext.expression", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4971
 * 
 * 
 *         def statement(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.StatementContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19If_statementContext_5statement(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19If_statementContext_5statement = {"statement", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19If_statementContext_5statement, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19If_statementContext_5statement(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("statement (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "statement") < 0)) __PYX_ERR(0, 4971, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("statement", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4971, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.If_statementContext.statement", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19If_statementContext_4statement(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19If_statementContext_4statement(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("statement", 0);

  /* "pymola/generated/ModelicaParser.pyx":4972
 * 
 *         def statement(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.StatementContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4973
 *         def statement(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.StatementContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.StatementContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4973, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4973, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_StatementContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4973, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4973, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4973, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4973, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4973, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4973, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":4972
 * 
 *         def statement(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.StatementContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4975
 *                 return self.getTypedRuleContexts(ModelicaParser.StatementContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.StatementContext,i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4975, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4975, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_StatementContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4975, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4975, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4975, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4975, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4975, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":4971
 * 
 * 
 *         def statement(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.StatementContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.If_statementContext.statement", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4978
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_if_statement
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19If_statementContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19If_statementContext_7getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19If_statementContext_7getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19If_statementContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19If_statementContext_6getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19If_statementContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":4979
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_if_statement             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4979, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_if_statement); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4979, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":4978
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_if_statement
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.If_statementContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4981
 *             return ModelicaParser.RULE_if_statement
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterIf_statement" ):
 *                 listener.enterIf_statement(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19If_statementContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19If_statementContext_9enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19If_statementContext_9enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19If_statementContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 4981, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 4981, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4981, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.If_statementContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19If_statementContext_8enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19If_statementContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":4982
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterIf_statement" ):             # <<<<<<<<<<<<<<
 *                 listener.enterIf_statement(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterIf_statement); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4982, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4983
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterIf_statement" ):
 *                 listener.enterIf_statement(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterIf_statement); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4983, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4983, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4983, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4983, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4983, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4983, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":4982
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterIf_statement" ):             # <<<<<<<<<<<<<<
 *                 listener.enterIf_statement(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4981
 *             return ModelicaParser.RULE_if_statement
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterIf_statement" ):
 *                 listener.enterIf_statement(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.If_statementContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4985
 *                 listener.enterIf_statement(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitIf_statement" ):
 *                 listener.exitIf_statement(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19If_statementContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19If_statementContext_11exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19If_statementContext_11exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19If_statementContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 4985, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 4985, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4985, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.If_statementContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19If_statementContext_10exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19If_statementContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":4986
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitIf_statement" ):             # <<<<<<<<<<<<<<
 *                 listener.exitIf_statement(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitIf_statement); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4986, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4987
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitIf_statement" ):
 *                 listener.exitIf_statement(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitIf_statement); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4987, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4987, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4987, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4987, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4987, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4987, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":4986
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitIf_statement" ):             # <<<<<<<<<<<<<<
 *                 listener.exitIf_statement(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4985
 *                 listener.enterIf_statement(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitIf_statement" ):
 *                 listener.exitIf_statement(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.If_statementContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4989
 *                 listener.exitIf_statement(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitIf_statement" ):
 *                 return visitor.visitIf_statement(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19If_statementContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19If_statementContext_13accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19If_statementContext_13accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19If_statementContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 4989, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 4989, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4989, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.If_statementContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19If_statementContext_12accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19If_statementContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":4990
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitIf_statement" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitIf_statement(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitIf_statement); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 4990, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":4991
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitIf_statement" ):
 *                 return visitor.visitIf_statement(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitIf_statement); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4991, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4991, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4991, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4991, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4991, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4991, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":4990
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitIf_statement" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitIf_statement(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":4993
 *                 return visitor.visitIf_statement(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4993, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4993, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4993, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4993, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4993, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4993, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":4989
 *                 listener.exitIf_statement(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitIf_statement" ):
 *                 return visitor.visitIf_statement(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.If_statementContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":4998
 * 
 * 
 *     def if_statement(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.If_statementContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_93if_statement(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_93if_statement = {"if_statement", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_93if_statement, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_93if_statement(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("if_statement (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_92if_statement(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_92if_statement(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  char const *__pyx_t_22;
  __Pyx_RefNannySetupContext("if_statement", 0);

  /* "pymola/generated/ModelicaParser.pyx":5000
 *     def if_statement(self):
 * 
 *         localctx = ModelicaParser.If_statementContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 90, self.RULE_if_statement)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5000, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_If_statementContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5000, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5000, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5000, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5000, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5000, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5000, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5000, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5001
 * 
 *         localctx = ModelicaParser.If_statementContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 90, self.RULE_if_statement)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5001, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_if_statement); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5001, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_90, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5001, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_90, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5001, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5001, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_90);
    __Pyx_GIVEREF(__pyx_int_90);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_90);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5001, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5002
 *         localctx = ModelicaParser.If_statementContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 90, self.RULE_if_statement)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 5002, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5003
 *         self.enterRule(localctx, 90, self.RULE_if_statement)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 660
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":5004
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 660
 *             self.match(ModelicaParser.T__39)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5004, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5004, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5004, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5004, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5004, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5005
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 660             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__39)
 *             self.state = 661
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_660) < 0) __PYX_ERR(0, 5005, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5006
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 660
 *             self.match(ModelicaParser.T__39)             # <<<<<<<<<<<<<<
 *             self.state = 661
 *             self.expression()
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5006, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5006, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__39); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5006, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5006, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5006, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5006, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5006, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5006, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5007
 *             self.state = 660
 *             self.match(ModelicaParser.T__39)
 *             self.state = 661             # <<<<<<<<<<<<<<
 *             self.expression()
 *             self.state = 662
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_661) < 0) __PYX_ERR(0, 5007, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5008
 *             self.match(ModelicaParser.T__39)
 *             self.state = 661
 *             self.expression()             # <<<<<<<<<<<<<<
 *             self.state = 662
 *             self.match(ModelicaParser.T__45)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expression); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5008, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5008, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5008, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5009
 *             self.state = 661
 *             self.expression()
 *             self.state = 662             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__45)
 *             self.state = 668
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_662) < 0) __PYX_ERR(0, 5009, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5010
 *             self.expression()
 *             self.state = 662
 *             self.match(ModelicaParser.T__45)             # <<<<<<<<<<<<<<
 *             self.state = 668
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5010, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5010, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__45); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5010, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5010, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5010, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5010, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5010, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4); __pyx_t_4 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5010, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5011
 *             self.state = 662
 *             self.match(ModelicaParser.T__45)
 *             self.state = 668             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_668) < 0) __PYX_ERR(0, 5011, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5012
 *             self.match(ModelicaParser.T__45)
 *             self.state = 668
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5012, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5012, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5012, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5012, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5012, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5012, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5012, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5013
 *             self.state = 668
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:
 *                 self.state = 663
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5013, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5013, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__904, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5013, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v__la = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5014
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:             # <<<<<<<<<<<<<<
 *                 self.state = 663
 *                 self.statement()
 */
        while (1) {
          __pyx_t_1 = __Pyx_PyInt_From_long((~0x3f)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5014, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_7 = PyNumber_And(__pyx_v__la, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5014, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_t_7, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5014, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5014, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (!__pyx_t_12) {
            goto __pyx_L17_next_or;
          } else {
          }
          __pyx_t_1 = PyNumber_Lshift(__pyx_int_1, __pyx_v__la); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5014, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5014, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__16); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5014, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5014, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5014, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__39); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5014, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5014, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyNumber_Or(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5014, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5014, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__43); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5014, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5014, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Or(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5014, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5014, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__44); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5014, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5014, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyNumber_Or(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5014, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5014, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__48); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5014, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5014, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Or(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5014, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5014, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__51); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5014, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5014, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyNumber_Or(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5014, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5014, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__52); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5014, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5014, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Or(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5014, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_And(__pyx_t_1, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5014, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyObject_RichCompare(__pyx_t_2, __pyx_int_0, Py_NE); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5014, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5014, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (!__pyx_t_12) {
          } else {
            __pyx_t_11 = __pyx_t_12;
            goto __pyx_L16_bool_binop_done;
          }
          __pyx_L17_next_or:;
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5014, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5014, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyObject_RichCompare(__pyx_v__la, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5014, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5014, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_11 = __pyx_t_12;
          __pyx_L16_bool_binop_done:;
          if (!__pyx_t_11) break;

          /* "pymola/generated/ModelicaParser.pyx":5015
 *             _la = self._input.LA(1)
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:
 *                 self.state = 663             # <<<<<<<<<<<<<<
 *                 self.statement()
 *                 self.state = 664
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_663) < 0) __PYX_ERR(0, 5015, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5016
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:
 *                 self.state = 663
 *                 self.statement()             # <<<<<<<<<<<<<<
 *                 self.state = 664
 *                 self.match(ModelicaParser.T__0)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_statement); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5016, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_1 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (__pyx_t_1) {
            __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5016, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          } else {
            __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5016, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5017
 *                 self.state = 663
 *                 self.statement()
 *                 self.state = 664             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__0)
 *                 self.state = 670
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_664) < 0) __PYX_ERR(0, 5017, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5018
 *                 self.statement()
 *                 self.state = 664
 *                 self.match(ModelicaParser.T__0)             # <<<<<<<<<<<<<<
 *                 self.state = 670
 *                 self._errHandler.sync(self)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5018, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5018, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5018, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_1) {
            __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5018, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_t_3};
              __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5018, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_t_3};
              __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5018, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5018, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1); __pyx_t_1 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5018, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5019
 *                 self.state = 664
 *                 self.match(ModelicaParser.T__0)
 *                 self.state = 670             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_670) < 0) __PYX_ERR(0, 5019, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5020
 *                 self.match(ModelicaParser.T__0)
 *                 self.state = 670
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 * 
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_sync); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5020, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5020, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5020, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_7);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5020, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_7);
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5020, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
              __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5020, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5021
 *                 self.state = 670
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *             self.state = 684
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5021, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_LA); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5021, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__905, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5021, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_7);
          __pyx_t_7 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":5023
 *                 _la = self._input.LA(1)
 * 
 *             self.state = 684             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_684) < 0) __PYX_ERR(0, 5023, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5024
 * 
 *             self.state = 684
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             while _la==ModelicaParser.T__46:
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5024, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_sync); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5024, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_4) {
          __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_self); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5024, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
            __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5024, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
            __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5024, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5024, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5024, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5025
 *             self.state = 684
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             while _la==ModelicaParser.T__46:
 *                 self.state = 671
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5025, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_LA); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5025, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__906, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5025, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5026
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             while _la==ModelicaParser.T__46:             # <<<<<<<<<<<<<<
 *                 self.state = 671
 *                 self.match(ModelicaParser.T__46)
 */
        while (1) {
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5026, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__46); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5026, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyObject_RichCompare(__pyx_v__la, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5026, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 5026, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (!__pyx_t_11) break;

          /* "pymola/generated/ModelicaParser.pyx":5027
 *             _la = self._input.LA(1)
 *             while _la==ModelicaParser.T__46:
 *                 self.state = 671             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__46)
 *                 self.state = 672
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_671) < 0) __PYX_ERR(0, 5027, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5028
 *             while _la==ModelicaParser.T__46:
 *                 self.state = 671
 *                 self.match(ModelicaParser.T__46)             # <<<<<<<<<<<<<<
 *                 self.state = 672
 *                 self.expression()
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5028, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5028, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__46); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5028, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5028, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_4};
              __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5028, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_4};
              __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5028, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            } else
            #endif
            {
              __pyx_t_1 = PyTuple_New(1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5028, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_GIVEREF(__pyx_t_4);
              PyTuple_SET_ITEM(__pyx_t_1, 0+1, __pyx_t_4);
              __pyx_t_4 = 0;
              __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_1, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5028, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5029
 *                 self.state = 671
 *                 self.match(ModelicaParser.T__46)
 *                 self.state = 672             # <<<<<<<<<<<<<<
 *                 self.expression()
 *                 self.state = 673
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_672) < 0) __PYX_ERR(0, 5029, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5030
 *                 self.match(ModelicaParser.T__46)
 *                 self.state = 672
 *                 self.expression()             # <<<<<<<<<<<<<<
 *                 self.state = 673
 *                 self.match(ModelicaParser.T__45)
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expression); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5030, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (__pyx_t_1) {
            __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5030, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          } else {
            __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5030, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5031
 *                 self.state = 672
 *                 self.expression()
 *                 self.state = 673             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__45)
 *                 self.state = 679
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_673) < 0) __PYX_ERR(0, 5031, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5032
 *                 self.expression()
 *                 self.state = 673
 *                 self.match(ModelicaParser.T__45)             # <<<<<<<<<<<<<<
 *                 self.state = 679
 *                 self._errHandler.sync(self)
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5032, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5032, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__45); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5032, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (!__pyx_t_1) {
            __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5032, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_t_4};
              __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5032, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_t_4};
              __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5032, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            } else
            #endif
            {
              __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5032, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1); __pyx_t_1 = NULL;
              __Pyx_GIVEREF(__pyx_t_4);
              PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_4);
              __pyx_t_4 = 0;
              __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5032, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5033
 *                 self.state = 673
 *                 self.match(ModelicaParser.T__45)
 *                 self.state = 679             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_679) < 0) __PYX_ERR(0, 5033, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5034
 *                 self.match(ModelicaParser.T__45)
 *                 self.state = 679
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 *                 while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5034, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5034, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_3) {
            __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5034, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
              __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5034, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_7);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
              __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5034, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_7);
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5034, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __pyx_t_3 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_self);
              __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5034, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5035
 *                 self.state = 679
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:
 *                     self.state = 674
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5035, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_LA); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5035, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__907, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5035, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_7);
          __pyx_t_7 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5036
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:             # <<<<<<<<<<<<<<
 *                     self.state = 674
 *                     self.statement()
 */
          while (1) {
            __pyx_t_7 = __Pyx_PyInt_From_long((~0x3f)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5036, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_2 = PyNumber_And(__pyx_v__la, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5036, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __pyx_t_7 = __Pyx_PyInt_EqObjC(__pyx_t_2, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5036, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5036, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            if (!__pyx_t_12) {
              goto __pyx_L24_next_or;
            } else {
            }
            __pyx_t_7 = PyNumber_Lshift(__pyx_int_1, __pyx_v__la); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5036, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5036, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__16); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5036, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5036, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5036, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__39); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5036, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5036, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = PyNumber_Or(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5036, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5036, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__43); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5036, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5036, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_2 = PyNumber_Or(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5036, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5036, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__44); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5036, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5036, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = PyNumber_Or(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5036, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5036, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__48); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5036, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5036, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_2 = PyNumber_Or(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5036, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5036, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__51); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5036, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5036, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = PyNumber_Or(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5036, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5036, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__52); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5036, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5036, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_2 = PyNumber_Or(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5036, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = PyNumber_And(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5036, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_2 = PyObject_RichCompare(__pyx_t_4, __pyx_int_0, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5036, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5036, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (!__pyx_t_12) {
            } else {
              __pyx_t_11 = __pyx_t_12;
              goto __pyx_L23_bool_binop_done;
            }
            __pyx_L24_next_or:;
            __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5036, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5036, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_2 = PyObject_RichCompare(__pyx_v__la, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5036, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5036, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_11 = __pyx_t_12;
            __pyx_L23_bool_binop_done:;
            if (!__pyx_t_11) break;

            /* "pymola/generated/ModelicaParser.pyx":5037
 *                 _la = self._input.LA(1)
 *                 while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:
 *                     self.state = 674             # <<<<<<<<<<<<<<
 *                     self.statement()
 *                     self.state = 675
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_674) < 0) __PYX_ERR(0, 5037, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":5038
 *                 while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:
 *                     self.state = 674
 *                     self.statement()             # <<<<<<<<<<<<<<
 *                     self.state = 675
 *                     self.match(ModelicaParser.T__0)
 */
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_statement); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5038, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_7 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
              __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
              if (likely(__pyx_t_7)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                __Pyx_INCREF(__pyx_t_7);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_4, function);
              }
            }
            if (__pyx_t_7) {
              __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5038, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else {
              __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5038, __pyx_L6_error)
            }
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

            /* "pymola/generated/ModelicaParser.pyx":5039
 *                     self.state = 674
 *                     self.statement()
 *                     self.state = 675             # <<<<<<<<<<<<<<
 *                     self.match(ModelicaParser.T__0)
 *                     self.state = 681
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_675) < 0) __PYX_ERR(0, 5039, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":5040
 *                     self.statement()
 *                     self.state = 675
 *                     self.match(ModelicaParser.T__0)             # <<<<<<<<<<<<<<
 *                     self.state = 681
 *                     self._errHandler.sync(self)
 */
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5040, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5040, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5040, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __pyx_t_7 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
              __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
              if (likely(__pyx_t_7)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                __Pyx_INCREF(__pyx_t_7);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_4, function);
              }
            }
            if (!__pyx_t_7) {
              __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5040, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_2);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_4)) {
                PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_3};
                __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5040, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
                PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_3};
                __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5040, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              } else
              #endif
              {
                __pyx_t_1 = PyTuple_New(1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5040, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_7); __pyx_t_7 = NULL;
                __Pyx_GIVEREF(__pyx_t_3);
                PyTuple_SET_ITEM(__pyx_t_1, 0+1, __pyx_t_3);
                __pyx_t_3 = 0;
                __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5040, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

            /* "pymola/generated/ModelicaParser.pyx":5041
 *                     self.state = 675
 *                     self.match(ModelicaParser.T__0)
 *                     self.state = 681             # <<<<<<<<<<<<<<
 *                     self._errHandler.sync(self)
 *                     _la = self._input.LA(1)
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_681) < 0) __PYX_ERR(0, 5041, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":5042
 *                     self.match(ModelicaParser.T__0)
 *                     self.state = 681
 *                     self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                     _la = self._input.LA(1)
 * 
 */
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5042, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_sync); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5042, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
              __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
              if (likely(__pyx_t_4)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
                __Pyx_INCREF(__pyx_t_4);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_1, function);
              }
            }
            if (!__pyx_t_4) {
              __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5042, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_1)) {
                PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
                __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5042, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_GOTREF(__pyx_t_2);
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
                PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
                __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5042, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_GOTREF(__pyx_t_2);
              } else
              #endif
              {
                __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5042, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __pyx_t_4 = NULL;
                __Pyx_INCREF(__pyx_v_self);
                __Pyx_GIVEREF(__pyx_v_self);
                PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
                __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5042, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

            /* "pymola/generated/ModelicaParser.pyx":5043
 *                     self.state = 681
 *                     self._errHandler.sync(self)
 *                     _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *                 self.state = 686
 */
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5043, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_LA); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5043, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__908, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5043, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_2);
            __pyx_t_2 = 0;
          }

          /* "pymola/generated/ModelicaParser.pyx":5045
 *                     _la = self._input.LA(1)
 * 
 *                 self.state = 686             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_686) < 0) __PYX_ERR(0, 5045, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5046
 * 
 *                 self.state = 686
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 * 
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5046, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_sync); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5046, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (!__pyx_t_1) {
            __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5046, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_v_self};
              __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5046, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_GOTREF(__pyx_t_2);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_v_self};
              __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5046, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_GOTREF(__pyx_t_2);
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5046, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1); __pyx_t_1 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_self);
              __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5046, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5047
 *                 self.state = 686
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *             self.state = 696
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5047, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_LA); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5047, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__909, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5047, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_2);
          __pyx_t_2 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":5049
 *                 _la = self._input.LA(1)
 * 
 *             self.state = 696             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_696) < 0) __PYX_ERR(0, 5049, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5050
 * 
 *             self.state = 696
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__47:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5050, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5050, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5050, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5050, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5050, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_1 = PyTuple_New(1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5050, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_1, 0+1, __pyx_v_self);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5050, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5051
 *             self.state = 696
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.T__47:
 *                 self.state = 687
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5051, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_LA); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5051, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__910, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5051, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5052
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__47:             # <<<<<<<<<<<<<<
 *                 self.state = 687
 *                 self.match(ModelicaParser.T__47)
 */
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5052, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__47); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5052, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_v__la, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5052, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 5052, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":5053
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__47:
 *                 self.state = 687             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__47)
 *                 self.state = 693
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_687) < 0) __PYX_ERR(0, 5053, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5054
 *             if _la==ModelicaParser.T__47:
 *                 self.state = 687
 *                 self.match(ModelicaParser.T__47)             # <<<<<<<<<<<<<<
 *                 self.state = 693
 *                 self._errHandler.sync(self)
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5054, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5054, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__47); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5054, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (!__pyx_t_1) {
            __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5054, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_t_3};
              __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5054, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_t_3};
              __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5054, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5054, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_1); __pyx_t_1 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5054, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5055
 *                 self.state = 687
 *                 self.match(ModelicaParser.T__47)
 *                 self.state = 693             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_693) < 0) __PYX_ERR(0, 5055, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5056
 *                 self.match(ModelicaParser.T__47)
 *                 self.state = 693
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 *                 while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5056, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5056, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_4) {
            __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5056, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
              __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5056, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_2);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
              __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5056, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_2);
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5056, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __pyx_t_4 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
              __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5056, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5057
 *                 self.state = 693
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:
 *                     self.state = 688
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5057, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5057, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__911, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5057, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_2);
          __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5058
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:             # <<<<<<<<<<<<<<
 *                     self.state = 688
 *                     self.statement()
 */
          while (1) {
            __pyx_t_2 = __Pyx_PyInt_From_long((~0x3f)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5058, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_7 = PyNumber_And(__pyx_v__la, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5058, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_t_7, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5058, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5058, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (!__pyx_t_12) {
              goto __pyx_L30_next_or;
            } else {
            }
            __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_v__la); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5058, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5058, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__16); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5058, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __pyx_t_7 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5058, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5058, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__39); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5058, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = PyNumber_Lshift(__pyx_int_1, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5058, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = PyNumber_Or(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5058, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5058, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__43); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5058, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5058, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __pyx_t_7 = PyNumber_Or(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5058, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5058, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__44); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5058, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = PyNumber_Lshift(__pyx_int_1, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5058, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = PyNumber_Or(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5058, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5058, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__48); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5058, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5058, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __pyx_t_7 = PyNumber_Or(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5058, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5058, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__51); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5058, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = PyNumber_Lshift(__pyx_int_1, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5058, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = PyNumber_Or(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5058, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5058, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__52); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5058, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5058, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __pyx_t_7 = PyNumber_Or(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5058, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = PyNumber_And(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5058, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __pyx_t_7 = PyObject_RichCompare(__pyx_t_3, __pyx_int_0, Py_NE); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5058, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5058, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            if (!__pyx_t_12) {
            } else {
              __pyx_t_11 = __pyx_t_12;
              goto __pyx_L29_bool_binop_done;
            }
            __pyx_L30_next_or:;
            __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5058, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5058, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __pyx_t_7 = PyObject_RichCompare(__pyx_v__la, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5058, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5058, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __pyx_t_11 = __pyx_t_12;
            __pyx_L29_bool_binop_done:;
            if (!__pyx_t_11) break;

            /* "pymola/generated/ModelicaParser.pyx":5059
 *                 _la = self._input.LA(1)
 *                 while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:
 *                     self.state = 688             # <<<<<<<<<<<<<<
 *                     self.statement()
 *                     self.state = 689
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_688) < 0) __PYX_ERR(0, 5059, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":5060
 *                 while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:
 *                     self.state = 688
 *                     self.statement()             # <<<<<<<<<<<<<<
 *                     self.state = 689
 *                     self.match(ModelicaParser.T__0)
 */
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_statement); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5060, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_2 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
              __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
              if (likely(__pyx_t_2)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                __Pyx_INCREF(__pyx_t_2);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_3, function);
              }
            }
            if (__pyx_t_2) {
              __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5060, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            } else {
              __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5060, __pyx_L6_error)
            }
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

            /* "pymola/generated/ModelicaParser.pyx":5061
 *                     self.state = 688
 *                     self.statement()
 *                     self.state = 689             # <<<<<<<<<<<<<<
 *                     self.match(ModelicaParser.T__0)
 *                     self.state = 695
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_689) < 0) __PYX_ERR(0, 5061, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":5062
 *                     self.statement()
 *                     self.state = 689
 *                     self.match(ModelicaParser.T__0)             # <<<<<<<<<<<<<<
 *                     self.state = 695
 *                     self._errHandler.sync(self)
 */
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5062, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5062, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5062, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_2 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
              __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
              if (likely(__pyx_t_2)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                __Pyx_INCREF(__pyx_t_2);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_3, function);
              }
            }
            if (!__pyx_t_2) {
              __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5062, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_7);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_3)) {
                PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_4};
                __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5062, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
                __Pyx_GOTREF(__pyx_t_7);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
                PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_4};
                __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5062, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
                __Pyx_GOTREF(__pyx_t_7);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              } else
              #endif
              {
                __pyx_t_1 = PyTuple_New(1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5062, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2); __pyx_t_2 = NULL;
                __Pyx_GIVEREF(__pyx_t_4);
                PyTuple_SET_ITEM(__pyx_t_1, 0+1, __pyx_t_4);
                __pyx_t_4 = 0;
                __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_1, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5062, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_7);
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

            /* "pymola/generated/ModelicaParser.pyx":5063
 *                     self.state = 689
 *                     self.match(ModelicaParser.T__0)
 *                     self.state = 695             # <<<<<<<<<<<<<<
 *                     self._errHandler.sync(self)
 *                     _la = self._input.LA(1)
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_695) < 0) __PYX_ERR(0, 5063, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":5064
 *                     self.match(ModelicaParser.T__0)
 *                     self.state = 695
 *                     self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                     _la = self._input.LA(1)
 * 
 */
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5064, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5064, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
              __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
              if (likely(__pyx_t_3)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
                __Pyx_INCREF(__pyx_t_3);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_1, function);
              }
            }
            if (!__pyx_t_3) {
              __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_self); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5064, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_1)) {
                PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
                __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5064, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_GOTREF(__pyx_t_7);
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
                PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
                __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5064, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_GOTREF(__pyx_t_7);
              } else
              #endif
              {
                __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5064, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __pyx_t_3 = NULL;
                __Pyx_INCREF(__pyx_v_self);
                __Pyx_GIVEREF(__pyx_v_self);
                PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_self);
                __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_4, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5064, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_7);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

            /* "pymola/generated/ModelicaParser.pyx":5065
 *                     self.state = 695
 *                     self._errHandler.sync(self)
 *                     _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 * 
 */
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5065, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_LA); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5065, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__912, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5065, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_7);
            __pyx_t_7 = 0;
          }

          /* "pymola/generated/ModelicaParser.pyx":5052
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__47:             # <<<<<<<<<<<<<<
 *                 self.state = 687
 *                 self.match(ModelicaParser.T__47)
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":5069
 * 
 * 
 *             self.state = 698             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__13)
 *             self.state = 699
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_698) < 0) __PYX_ERR(0, 5069, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5070
 * 
 *             self.state = 698
 *             self.match(ModelicaParser.T__13)             # <<<<<<<<<<<<<<
 *             self.state = 699
 *             self.match(ModelicaParser.T__39)
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5070, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5070, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__13); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5070, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_1, function);
          }
        }
        if (!__pyx_t_4) {
          __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5070, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_7);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_1)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_3};
            __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5070, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_3};
            __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5070, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5070, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
            __Pyx_GIVEREF(__pyx_t_3);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_3);
            __pyx_t_3 = 0;
            __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5070, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5071
 *             self.state = 698
 *             self.match(ModelicaParser.T__13)
 *             self.state = 699             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__39)
 *         except RecognitionException as re:
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_699) < 0) __PYX_ERR(0, 5071, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5072
 *             self.match(ModelicaParser.T__13)
 *             self.state = 699
 *             self.match(ModelicaParser.T__39)             # <<<<<<<<<<<<<<
 *         except RecognitionException as re:
 *             localctx.exception = re
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5072, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5072, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__39); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5072, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_1, function);
          }
        }
        if (!__pyx_t_2) {
          __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5072, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_7);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_1)) {
            PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
            __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5072, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
            PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
            __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5072, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5072, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
            __Pyx_GIVEREF(__pyx_t_3);
            PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_3);
            __pyx_t_3 = 0;
            __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_4, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5072, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5003
 *         self.enterRule(localctx, 90, self.RULE_if_statement)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 660
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "pymola/generated/ModelicaParser.pyx":5073
 *             self.state = 699
 *             self.match(ModelicaParser.T__39)
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5073, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.if_statement", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_1, &__pyx_t_4) < 0) __PYX_ERR(0, 5073, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_v_re = __pyx_t_1;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":5074
 *             self.match(ModelicaParser.T__39)
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 5074, __pyx_L37_error)

          /* "pymola/generated/ModelicaParser.pyx":5075
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5075, __pyx_L37_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5075, __pyx_L37_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5075, __pyx_L37_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5075, __pyx_L37_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 5075, __pyx_L37_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_2) {
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_2); __pyx_t_2 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_13, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5075, __pyx_L37_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5076
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5076, __pyx_L37_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 5076, __pyx_L37_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5076, __pyx_L37_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5076, __pyx_L37_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5076, __pyx_L37_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5076, __pyx_L37_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":5073
 *             self.state = 699
 *             self.match(ModelicaParser.T__39)
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L38;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L37_error:;
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_6 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L8_except_error;
          }
          __pyx_L38:;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":5003
 *         self.enterRule(localctx, 90, self.RULE_if_statement)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 660
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":5078
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5078, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      if (__pyx_t_7) {
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5078, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      } else {
        __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5078, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_14 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5078, __pyx_L44_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_1, function);
          }
        }
        if (__pyx_t_7) {
          __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5078, __pyx_L44_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else {
          __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5078, __pyx_L44_error)
        }
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
      __pyx_L44_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":5079
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class For_equationContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":4998
 * 
 * 
 *     def if_statement(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.If_statementContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.if_statement", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5083
 *     class For_equationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19For_equationContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19For_equationContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19For_equationContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19For_equationContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 5083, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 5083, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5083, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_equationContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19For_equationContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19For_equationContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":5084
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 5084, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5084, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5084, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5084, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5084, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5084, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5084, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5084, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5085
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def for_indices(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 5085, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5083
 *     class For_equationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_equationContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5087
 *             self.parser = parser
 * 
 *         def for_indices(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.For_indicesContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19For_equationContext_3for_indices(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19For_equationContext_3for_indices = {"for_indices", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19For_equationContext_3for_indices, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19For_equationContext_3for_indices(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("for_indices (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19For_equationContext_2for_indices(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19For_equationContext_2for_indices(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("for_indices", 0);

  /* "pymola/generated/ModelicaParser.pyx":5088
 * 
 *         def for_indices(self):
 *             return self.getTypedRuleContext(ModelicaParser.For_indicesContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5088, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5088, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_For_indicesContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5088, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5088, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5088, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5088, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5088, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":5087
 *             self.parser = parser
 * 
 *         def for_indices(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.For_indicesContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_equationContext.for_indices", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5091
 * 
 * 
 *         def equation(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.EquationContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19For_equationContext_5equation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19For_equationContext_5equation = {"equation", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19For_equationContext_5equation, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19For_equationContext_5equation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("equation (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "equation") < 0)) __PYX_ERR(0, 5091, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("equation", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5091, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_equationContext.equation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19For_equationContext_4equation(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19For_equationContext_4equation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("equation", 0);

  /* "pymola/generated/ModelicaParser.pyx":5092
 * 
 *         def equation(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.EquationContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5093
 *         def equation(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.EquationContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.EquationContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5093, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5093, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_EquationContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5093, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5093, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5093, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5093, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5093, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5093, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":5092
 * 
 *         def equation(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.EquationContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5095
 *                 return self.getTypedRuleContexts(ModelicaParser.EquationContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.EquationContext,i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5095, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5095, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_EquationContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5095, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5095, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5095, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5095, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5095, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":5091
 * 
 * 
 *         def equation(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.EquationContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_equationContext.equation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5098
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_for_equation
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19For_equationContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19For_equationContext_7getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19For_equationContext_7getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19For_equationContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19For_equationContext_6getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19For_equationContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":5099
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_for_equation             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5099, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_for_equation); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5099, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":5098
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_for_equation
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_equationContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5101
 *             return ModelicaParser.RULE_for_equation
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterFor_equation" ):
 *                 listener.enterFor_equation(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19For_equationContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19For_equationContext_9enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19For_equationContext_9enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19For_equationContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 5101, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 5101, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5101, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_equationContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19For_equationContext_8enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19For_equationContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":5102
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterFor_equation" ):             # <<<<<<<<<<<<<<
 *                 listener.enterFor_equation(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterFor_equation); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5102, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5103
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterFor_equation" ):
 *                 listener.enterFor_equation(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterFor_equation); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5103, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5103, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5103, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5103, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5103, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5103, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":5102
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterFor_equation" ):             # <<<<<<<<<<<<<<
 *                 listener.enterFor_equation(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5101
 *             return ModelicaParser.RULE_for_equation
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterFor_equation" ):
 *                 listener.enterFor_equation(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_equationContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5105
 *                 listener.enterFor_equation(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitFor_equation" ):
 *                 listener.exitFor_equation(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19For_equationContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19For_equationContext_11exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19For_equationContext_11exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19For_equationContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 5105, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 5105, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5105, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_equationContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19For_equationContext_10exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19For_equationContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":5106
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitFor_equation" ):             # <<<<<<<<<<<<<<
 *                 listener.exitFor_equation(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitFor_equation); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5106, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5107
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitFor_equation" ):
 *                 listener.exitFor_equation(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitFor_equation); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5107, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5107, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5107, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5107, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5107, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5107, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":5106
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitFor_equation" ):             # <<<<<<<<<<<<<<
 *                 listener.exitFor_equation(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5105
 *                 listener.enterFor_equation(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitFor_equation" ):
 *                 listener.exitFor_equation(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_equationContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5109
 *                 listener.exitFor_equation(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitFor_equation" ):
 *                 return visitor.visitFor_equation(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19For_equationContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19For_equationContext_13accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19For_equationContext_13accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19For_equationContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 5109, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 5109, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5109, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_equationContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19For_equationContext_12accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19For_equationContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":5110
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitFor_equation" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitFor_equation(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitFor_equation); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5110, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5111
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitFor_equation" ):
 *                 return visitor.visitFor_equation(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitFor_equation); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5111, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5111, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5111, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5111, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5111, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5111, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":5110
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitFor_equation" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitFor_equation(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5113
 *                 return visitor.visitFor_equation(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5113, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5113, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5113, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5113, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5113, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5113, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":5109
 *                 listener.exitFor_equation(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitFor_equation" ):
 *                 return visitor.visitFor_equation(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_equationContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5118
 * 
 * 
 *     def for_equation(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.For_equationContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_95for_equation(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_95for_equation = {"for_equation", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_95for_equation, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_95for_equation(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("for_equation (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_94for_equation(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_94for_equation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__alt = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  char const *__pyx_t_22;
  __Pyx_RefNannySetupContext("for_equation", 0);

  /* "pymola/generated/ModelicaParser.pyx":5120
 *     def for_equation(self):
 * 
 *         localctx = ModelicaParser.For_equationContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 92, self.RULE_for_equation)
 *         try:
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_For_equationContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5120, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5120, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5120, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5120, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5121
 * 
 *         localctx = ModelicaParser.For_equationContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 92, self.RULE_for_equation)             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_for_equation); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_92, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5121, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_92, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5121, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_92);
    __Pyx_GIVEREF(__pyx_int_92);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_92);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5122
 *         localctx = ModelicaParser.For_equationContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 92, self.RULE_for_equation)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 701
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":5123
 *         self.enterRule(localctx, 92, self.RULE_for_equation)
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 701
 *             self.match(ModelicaParser.T__48)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5123, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5123, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5123, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5123, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5123, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5124
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 701             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__48)
 *             self.state = 702
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_701) < 0) __PYX_ERR(0, 5124, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5125
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 701
 *             self.match(ModelicaParser.T__48)             # <<<<<<<<<<<<<<
 *             self.state = 702
 *             self.for_indices()
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5125, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5125, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__48); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5125, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5125, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5125, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5125, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5125, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5125, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5126
 *             self.state = 701
 *             self.match(ModelicaParser.T__48)
 *             self.state = 702             # <<<<<<<<<<<<<<
 *             self.for_indices()
 *             self.state = 703
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_702) < 0) __PYX_ERR(0, 5126, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5127
 *             self.match(ModelicaParser.T__48)
 *             self.state = 702
 *             self.for_indices()             # <<<<<<<<<<<<<<
 *             self.state = 703
 *             self.match(ModelicaParser.T__49)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_for_indices); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5127, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5127, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5127, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5128
 *             self.state = 702
 *             self.for_indices()
 *             self.state = 703             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__49)
 *             self.state = 709
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_703) < 0) __PYX_ERR(0, 5128, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5129
 *             self.for_indices()
 *             self.state = 703
 *             self.match(ModelicaParser.T__49)             # <<<<<<<<<<<<<<
 *             self.state = 709
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5129, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5129, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__49); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5129, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5129, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5129, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5129, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5129, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4); __pyx_t_4 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5129, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5130
 *             self.state = 703
 *             self.match(ModelicaParser.T__49)
 *             self.state = 709             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _alt = self._interp.adaptivePredict(self._input,94,self._ctx)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_709) < 0) __PYX_ERR(0, 5130, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5131
 *             self.match(ModelicaParser.T__49)
 *             self.state = 709
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _alt = self._interp.adaptivePredict(self._input,94,self._ctx)
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5131, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5131, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5131, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5131, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5131, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5131, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5131, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5132
 *             self.state = 709
 *             self._errHandler.sync(self)
 *             _alt = self._interp.adaptivePredict(self._input,94,self._ctx)             # <<<<<<<<<<<<<<
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                 if _alt==1:
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_interp); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5132, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_adaptivePredict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5132, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5132, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5132, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_2)) {
          PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_t_7, __pyx_int_94, __pyx_t_3};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5132, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
          PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_t_7, __pyx_int_94, __pyx_t_3};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5132, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5132, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_7);
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_t_7);
          __Pyx_INCREF(__pyx_int_94);
          __Pyx_GIVEREF(__pyx_int_94);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_int_94);
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_6, __pyx_t_3);
          __pyx_t_7 = 0;
          __pyx_t_3 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5132, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_v__alt = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5133
 *             self._errHandler.sync(self)
 *             _alt = self._interp.adaptivePredict(self._input,94,self._ctx)
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:             # <<<<<<<<<<<<<<
 *                 if _alt==1:
 *                     self.state = 704
 */
        while (1) {
          __pyx_t_1 = PyObject_RichCompare(__pyx_v__alt, __pyx_int_2, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5133, __pyx_L6_error)
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5133, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (__pyx_t_12) {
          } else {
            __pyx_t_11 = __pyx_t_12;
            goto __pyx_L16_bool_binop_done;
          }
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ATN); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5133, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_INVALID_ALT_NUMBER); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5133, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyObject_RichCompare(__pyx_v__alt, __pyx_t_2, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5133, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5133, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_11 = __pyx_t_12;
          __pyx_L16_bool_binop_done:;
          if (!__pyx_t_11) break;

          /* "pymola/generated/ModelicaParser.pyx":5134
 *             _alt = self._interp.adaptivePredict(self._input,94,self._ctx)
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                 if _alt==1:             # <<<<<<<<<<<<<<
 *                     self.state = 704
 *                     self.equation()
 */
          __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v__alt, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5134, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 5134, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (__pyx_t_11) {

            /* "pymola/generated/ModelicaParser.pyx":5135
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                 if _alt==1:
 *                     self.state = 704             # <<<<<<<<<<<<<<
 *                     self.equation()
 *                     self.state = 705
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_704) < 0) __PYX_ERR(0, 5135, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":5136
 *                 if _alt==1:
 *                     self.state = 704
 *                     self.equation()             # <<<<<<<<<<<<<<
 *                     self.state = 705
 *                     self.match(ModelicaParser.T__0)
 */
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_equation); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5136, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_5 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_5)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_5);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
              }
            }
            if (__pyx_t_5) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5136, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else {
              __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5136, __pyx_L6_error)
            }
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":5137
 *                     self.state = 704
 *                     self.equation()
 *                     self.state = 705             # <<<<<<<<<<<<<<
 *                     self.match(ModelicaParser.T__0)
 *                 self.state = 711
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_705) < 0) __PYX_ERR(0, 5137, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":5138
 *                     self.equation()
 *                     self.state = 705
 *                     self.match(ModelicaParser.T__0)             # <<<<<<<<<<<<<<
 *                 self.state = 711
 *                 self._errHandler.sync(self)
 */
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5138, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5138, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_T__0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5138, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_5 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_5)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_5);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
              }
            }
            if (!__pyx_t_5) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5138, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_2)) {
                PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5138, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
                PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5138, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              } else
              #endif
              {
                __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5138, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_7);
                __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
                __Pyx_GIVEREF(__pyx_t_3);
                PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_3);
                __pyx_t_3 = 0;
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5138, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":5134
 *             _alt = self._interp.adaptivePredict(self._input,94,self._ctx)
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                 if _alt==1:             # <<<<<<<<<<<<<<
 *                     self.state = 704
 *                     self.equation()
 */
          }

          /* "pymola/generated/ModelicaParser.pyx":5139
 *                     self.state = 705
 *                     self.match(ModelicaParser.T__0)
 *                 self.state = 711             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _alt = self._interp.adaptivePredict(self._input,94,self._ctx)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_711) < 0) __PYX_ERR(0, 5139, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5140
 *                     self.match(ModelicaParser.T__0)
 *                 self.state = 711
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _alt = self._interp.adaptivePredict(self._input,94,self._ctx)
 * 
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5140, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5140, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5140, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5140, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5140, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5140, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5140, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5141
 *                 self.state = 711
 *                 self._errHandler.sync(self)
 *                 _alt = self._interp.adaptivePredict(self._input,94,self._ctx)             # <<<<<<<<<<<<<<
 * 
 *             self.state = 712
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_interp); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5141, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_adaptivePredict); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5141, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5141, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5141, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_t_7, __pyx_int_94, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5141, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_t_7, __pyx_int_94, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5141, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5141, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_GIVEREF(__pyx_t_7);
            PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_t_7);
            __Pyx_INCREF(__pyx_int_94);
            __Pyx_GIVEREF(__pyx_int_94);
            PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_int_94);
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_4, 2+__pyx_t_6, __pyx_t_2);
            __pyx_t_7 = 0;
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5141, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF_SET(__pyx_v__alt, __pyx_t_1);
          __pyx_t_1 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":5143
 *                 _alt = self._interp.adaptivePredict(self._input,94,self._ctx)
 * 
 *             self.state = 712             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__13)
 *             self.state = 713
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_712) < 0) __PYX_ERR(0, 5143, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5144
 * 
 *             self.state = 712
 *             self.match(ModelicaParser.T__13)             # <<<<<<<<<<<<<<
 *             self.state = 713
 *             self.match(ModelicaParser.T__48)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5144, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5144, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__13); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5144, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5144, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5144, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5144, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5144, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4); __pyx_t_4 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5144, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5145
 *             self.state = 712
 *             self.match(ModelicaParser.T__13)
 *             self.state = 713             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__48)
 *         except RecognitionException as re:
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_713) < 0) __PYX_ERR(0, 5145, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5146
 *             self.match(ModelicaParser.T__13)
 *             self.state = 713
 *             self.match(ModelicaParser.T__48)             # <<<<<<<<<<<<<<
 *         except RecognitionException as re:
 *             localctx.exception = re
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5146, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5146, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__48); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5146, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5146, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5146, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5146, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5146, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5146, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5122
 *         localctx = ModelicaParser.For_equationContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 92, self.RULE_for_equation)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 701
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":5147
 *             self.state = 713
 *             self.match(ModelicaParser.T__48)
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5147, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.for_equation", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_3, &__pyx_t_4) < 0) __PYX_ERR(0, 5147, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __pyx_v_re = __pyx_t_3;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":5148
 *             self.match(ModelicaParser.T__48)
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 5148, __pyx_L24_error)

          /* "pymola/generated/ModelicaParser.pyx":5149
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5149, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5149, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5149, __pyx_L24_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5149, __pyx_L24_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 5149, __pyx_L24_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_7) {
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_7); __pyx_t_7 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_13, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5149, __pyx_L24_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5150
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5150, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 5150, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5150, __pyx_L24_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5150, __pyx_L24_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5150, __pyx_L24_error)
            __Pyx_GOTREF(__pyx_t_7);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5150, __pyx_L24_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":5147
 *             self.state = 713
 *             self.match(ModelicaParser.T__48)
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L25;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L24_error:;
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_6 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L8_except_error;
          }
          __pyx_L25:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":5122
 *         localctx = ModelicaParser.For_equationContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 92, self.RULE_for_equation)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 701
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":5152
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5152, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5152, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5152, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_14 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5152, __pyx_L31_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5152, __pyx_L31_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5152, __pyx_L31_error)
        }
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
      __pyx_L31_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":5153
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class For_statementContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":5118
 * 
 * 
 *     def for_equation(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.For_equationContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.for_equation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__alt);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5157
 *     class For_statementContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20For_statementContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20For_statementContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20For_statementContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20For_statementContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 5157, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 5157, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5157, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_statementContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20For_statementContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20For_statementContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":5158
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 5158, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5158, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5158, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5158, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5158, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5159
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def for_indices(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 5159, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5157
 *     class For_statementContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_statementContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5161
 *             self.parser = parser
 * 
 *         def for_indices(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.For_indicesContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20For_statementContext_3for_indices(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20For_statementContext_3for_indices = {"for_indices", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20For_statementContext_3for_indices, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20For_statementContext_3for_indices(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("for_indices (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20For_statementContext_2for_indices(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20For_statementContext_2for_indices(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("for_indices", 0);

  /* "pymola/generated/ModelicaParser.pyx":5162
 * 
 *         def for_indices(self):
 *             return self.getTypedRuleContext(ModelicaParser.For_indicesContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_For_indicesContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5162, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5162, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5162, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5162, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":5161
 *             self.parser = parser
 * 
 *         def for_indices(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.For_indicesContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_statementContext.for_indices", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5165
 * 
 * 
 *         def statement(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.StatementContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20For_statementContext_5statement(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20For_statementContext_5statement = {"statement", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20For_statementContext_5statement, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20For_statementContext_5statement(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("statement (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "statement") < 0)) __PYX_ERR(0, 5165, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("statement", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5165, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_statementContext.statement", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20For_statementContext_4statement(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20For_statementContext_4statement(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("statement", 0);

  /* "pymola/generated/ModelicaParser.pyx":5166
 * 
 *         def statement(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.StatementContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5167
 *         def statement(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.StatementContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.StatementContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5167, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5167, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_StatementContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5167, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5167, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5167, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5167, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5167, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5167, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":5166
 * 
 *         def statement(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.StatementContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5169
 *                 return self.getTypedRuleContexts(ModelicaParser.StatementContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.StatementContext,i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5169, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5169, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_StatementContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5169, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5169, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5169, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5169, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5169, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":5165
 * 
 * 
 *         def statement(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.StatementContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_statementContext.statement", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5172
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_for_statement
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20For_statementContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20For_statementContext_7getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20For_statementContext_7getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20For_statementContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20For_statementContext_6getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20For_statementContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":5173
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_for_statement             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_for_statement); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":5172
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_for_statement
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_statementContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5175
 *             return ModelicaParser.RULE_for_statement
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterFor_statement" ):
 *                 listener.enterFor_statement(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20For_statementContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20For_statementContext_9enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20For_statementContext_9enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20For_statementContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 5175, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 5175, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5175, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_statementContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20For_statementContext_8enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20For_statementContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":5176
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterFor_statement" ):             # <<<<<<<<<<<<<<
 *                 listener.enterFor_statement(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterFor_statement); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5176, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5177
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterFor_statement" ):
 *                 listener.enterFor_statement(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterFor_statement); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5177, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5177, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5177, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5177, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5177, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":5176
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterFor_statement" ):             # <<<<<<<<<<<<<<
 *                 listener.enterFor_statement(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5175
 *             return ModelicaParser.RULE_for_statement
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterFor_statement" ):
 *                 listener.enterFor_statement(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_statementContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5179
 *                 listener.enterFor_statement(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitFor_statement" ):
 *                 listener.exitFor_statement(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20For_statementContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20For_statementContext_11exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20For_statementContext_11exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20For_statementContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 5179, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 5179, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5179, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_statementContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20For_statementContext_10exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20For_statementContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":5180
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitFor_statement" ):             # <<<<<<<<<<<<<<
 *                 listener.exitFor_statement(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitFor_statement); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5180, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5181
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitFor_statement" ):
 *                 listener.exitFor_statement(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitFor_statement); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5181, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5181, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5181, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5181, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5181, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5181, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":5180
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitFor_statement" ):             # <<<<<<<<<<<<<<
 *                 listener.exitFor_statement(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5179
 *                 listener.enterFor_statement(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitFor_statement" ):
 *                 listener.exitFor_statement(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_statementContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5183
 *                 listener.exitFor_statement(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitFor_statement" ):
 *                 return visitor.visitFor_statement(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20For_statementContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20For_statementContext_13accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20For_statementContext_13accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20For_statementContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 5183, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 5183, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5183, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_statementContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20For_statementContext_12accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20For_statementContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":5184
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitFor_statement" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitFor_statement(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitFor_statement); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5184, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5185
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitFor_statement" ):
 *                 return visitor.visitFor_statement(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitFor_statement); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5185, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5185, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5185, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5185, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5185, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5185, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":5184
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitFor_statement" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitFor_statement(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5187
 *                 return visitor.visitFor_statement(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5187, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5187, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5187, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5187, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5187, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5187, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":5183
 *                 listener.exitFor_statement(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitFor_statement" ):
 *                 return visitor.visitFor_statement(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_statementContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5192
 * 
 * 
 *     def for_statement(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.For_statementContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_97for_statement(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_97for_statement = {"for_statement", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_97for_statement, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_97for_statement(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("for_statement (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_96for_statement(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_96for_statement(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  char const *__pyx_t_22;
  __Pyx_RefNannySetupContext("for_statement", 0);

  /* "pymola/generated/ModelicaParser.pyx":5194
 *     def for_statement(self):
 * 
 *         localctx = ModelicaParser.For_statementContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 94, self.RULE_for_statement)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_For_statementContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5194, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5194, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5194, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5194, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5195
 * 
 *         localctx = ModelicaParser.For_statementContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 94, self.RULE_for_statement)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_for_statement); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_94, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5195, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_94, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5195, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5195, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_94);
    __Pyx_GIVEREF(__pyx_int_94);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_94);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5195, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5196
 *         localctx = ModelicaParser.For_statementContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 94, self.RULE_for_statement)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 5196, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5197
 *         self.enterRule(localctx, 94, self.RULE_for_statement)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 715
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":5198
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 715
 *             self.match(ModelicaParser.T__48)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5198, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5198, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5198, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5198, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5198, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5199
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 715             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__48)
 *             self.state = 716
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_715) < 0) __PYX_ERR(0, 5199, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5200
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 715
 *             self.match(ModelicaParser.T__48)             # <<<<<<<<<<<<<<
 *             self.state = 716
 *             self.for_indices()
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5200, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5200, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__48); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5200, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5200, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5200, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5200, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5200, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5200, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5201
 *             self.state = 715
 *             self.match(ModelicaParser.T__48)
 *             self.state = 716             # <<<<<<<<<<<<<<
 *             self.for_indices()
 *             self.state = 717
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_716) < 0) __PYX_ERR(0, 5201, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5202
 *             self.match(ModelicaParser.T__48)
 *             self.state = 716
 *             self.for_indices()             # <<<<<<<<<<<<<<
 *             self.state = 717
 *             self.match(ModelicaParser.T__49)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_for_indices); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5202, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5202, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5202, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5203
 *             self.state = 716
 *             self.for_indices()
 *             self.state = 717             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__49)
 *             self.state = 723
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_717) < 0) __PYX_ERR(0, 5203, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5204
 *             self.for_indices()
 *             self.state = 717
 *             self.match(ModelicaParser.T__49)             # <<<<<<<<<<<<<<
 *             self.state = 723
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5204, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5204, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__49); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5204, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5204, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5204, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5204, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5204, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4); __pyx_t_4 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5204, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5205
 *             self.state = 717
 *             self.match(ModelicaParser.T__49)
 *             self.state = 723             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_723) < 0) __PYX_ERR(0, 5205, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5206
 *             self.match(ModelicaParser.T__49)
 *             self.state = 723
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5206, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5206, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5206, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5206, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5206, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5206, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5206, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5207
 *             self.state = 723
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:
 *                 self.state = 718
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5207, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5207, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__913, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5207, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v__la = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5208
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:             # <<<<<<<<<<<<<<
 *                 self.state = 718
 *                 self.statement()
 */
        while (1) {
          __pyx_t_1 = __Pyx_PyInt_From_long((~0x3f)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5208, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_7 = PyNumber_And(__pyx_v__la, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5208, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_t_7, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5208, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5208, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (!__pyx_t_12) {
            goto __pyx_L17_next_or;
          } else {
          }
          __pyx_t_1 = PyNumber_Lshift(__pyx_int_1, __pyx_v__la); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5208, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5208, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__16); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5208, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5208, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5208, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__39); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5208, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5208, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyNumber_Or(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5208, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5208, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__43); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5208, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5208, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Or(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5208, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5208, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__44); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5208, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5208, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyNumber_Or(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5208, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5208, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__48); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5208, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5208, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Or(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5208, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5208, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__51); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5208, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5208, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyNumber_Or(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5208, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5208, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__52); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5208, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5208, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Or(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5208, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_And(__pyx_t_1, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5208, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyObject_RichCompare(__pyx_t_2, __pyx_int_0, Py_NE); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5208, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5208, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (!__pyx_t_12) {
          } else {
            __pyx_t_11 = __pyx_t_12;
            goto __pyx_L16_bool_binop_done;
          }
          __pyx_L17_next_or:;
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5208, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5208, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyObject_RichCompare(__pyx_v__la, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5208, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5208, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_11 = __pyx_t_12;
          __pyx_L16_bool_binop_done:;
          if (!__pyx_t_11) break;

          /* "pymola/generated/ModelicaParser.pyx":5209
 *             _la = self._input.LA(1)
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:
 *                 self.state = 718             # <<<<<<<<<<<<<<
 *                 self.statement()
 *                 self.state = 719
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_718) < 0) __PYX_ERR(0, 5209, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5210
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:
 *                 self.state = 718
 *                 self.statement()             # <<<<<<<<<<<<<<
 *                 self.state = 719
 *                 self.match(ModelicaParser.T__0)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_statement); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5210, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_1 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (__pyx_t_1) {
            __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5210, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          } else {
            __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5210, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5211
 *                 self.state = 718
 *                 self.statement()
 *                 self.state = 719             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__0)
 *                 self.state = 725
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_719) < 0) __PYX_ERR(0, 5211, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5212
 *                 self.statement()
 *                 self.state = 719
 *                 self.match(ModelicaParser.T__0)             # <<<<<<<<<<<<<<
 *                 self.state = 725
 *                 self._errHandler.sync(self)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5212, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5212, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5212, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_1) {
            __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5212, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_t_3};
              __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5212, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_t_3};
              __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5212, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5212, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1); __pyx_t_1 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5212, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5213
 *                 self.state = 719
 *                 self.match(ModelicaParser.T__0)
 *                 self.state = 725             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_725) < 0) __PYX_ERR(0, 5213, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5214
 *                 self.match(ModelicaParser.T__0)
 *                 self.state = 725
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 * 
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5214, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_sync); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5214, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5214, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5214, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_7);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5214, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_7);
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5214, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
              __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5214, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5215
 *                 self.state = 725
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *             self.state = 726
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5215, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_LA); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5215, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__914, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5215, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_7);
          __pyx_t_7 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":5217
 *                 _la = self._input.LA(1)
 * 
 *             self.state = 726             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__13)
 *             self.state = 727
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_726) < 0) __PYX_ERR(0, 5217, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5218
 * 
 *             self.state = 726
 *             self.match(ModelicaParser.T__13)             # <<<<<<<<<<<<<<
 *             self.state = 727
 *             self.match(ModelicaParser.T__48)
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5218, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5218, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__13); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5218, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5218, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_7);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_2};
            __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5218, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_2};
            __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5218, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_1 = PyTuple_New(1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5218, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_1, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_1, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5218, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5219
 *             self.state = 726
 *             self.match(ModelicaParser.T__13)
 *             self.state = 727             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__48)
 *         except RecognitionException as re:
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_727) < 0) __PYX_ERR(0, 5219, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5220
 *             self.match(ModelicaParser.T__13)
 *             self.state = 727
 *             self.match(ModelicaParser.T__48)             # <<<<<<<<<<<<<<
 *         except RecognitionException as re:
 *             localctx.exception = re
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5220, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5220, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__48); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5220, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (!__pyx_t_1) {
          __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5220, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_7);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_t_2};
            __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5220, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_t_2};
            __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5220, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5220, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1); __pyx_t_1 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5220, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5197
 *         self.enterRule(localctx, 94, self.RULE_for_statement)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 715
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "pymola/generated/ModelicaParser.pyx":5221
 *             self.state = 727
 *             self.match(ModelicaParser.T__48)
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5221, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.for_statement", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_4, &__pyx_t_3) < 0) __PYX_ERR(0, 5221, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __pyx_v_re = __pyx_t_4;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":5222
 *             self.match(ModelicaParser.T__48)
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 5222, __pyx_L24_error)

          /* "pymola/generated/ModelicaParser.pyx":5223
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5223, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5223, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5223, __pyx_L24_error)
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5223, __pyx_L24_error)
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 5223, __pyx_L24_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_1) {
              __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_1); __pyx_t_1 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_13, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5223, __pyx_L24_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5224
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5224, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 5224, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5224, __pyx_L24_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5224, __pyx_L24_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_1 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5224, __pyx_L24_error)
            __Pyx_GOTREF(__pyx_t_1);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5224, __pyx_L24_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":5221
 *             self.state = 727
 *             self.match(ModelicaParser.T__48)
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L25;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L24_error:;
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_6 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L8_except_error;
          }
          __pyx_L25:;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":5197
 *         self.enterRule(localctx, 94, self.RULE_for_statement)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 715
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":5226
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5226, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (__pyx_t_7) {
        __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5226, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      } else {
        __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5226, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_14 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5226, __pyx_L31_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (__pyx_t_7) {
          __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5226, __pyx_L31_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else {
          __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5226, __pyx_L31_error)
        }
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
      __pyx_L31_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":5227
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class For_indicesContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":5192
 * 
 * 
 *     def for_statement(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.For_statementContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.for_statement", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5231
 *     class For_indicesContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18For_indicesContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18For_indicesContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18For_indicesContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18For_indicesContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 5231, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 5231, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5231, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_indicesContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18For_indicesContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18For_indicesContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":5232
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 5232, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5232, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5232, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5232, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5232, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5233
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def for_index(self, i:int=None):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 5233, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5231
 *     class For_indicesContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_indicesContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5235
 *             self.parser = parser
 * 
 *         def for_index(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.For_indexContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18For_indicesContext_3for_index(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18For_indicesContext_3for_index = {"for_index", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18For_indicesContext_3for_index, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18For_indicesContext_3for_index(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("for_index (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "for_index") < 0)) __PYX_ERR(0, 5235, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("for_index", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5235, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_indicesContext.for_index", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18For_indicesContext_2for_index(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18For_indicesContext_2for_index(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("for_index", 0);

  /* "pymola/generated/ModelicaParser.pyx":5236
 * 
 *         def for_index(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.For_indexContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5237
 *         def for_index(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.For_indexContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.For_indexContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5237, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5237, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_For_indexContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5237, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5237, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5237, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5237, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5237, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5237, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":5236
 * 
 *         def for_index(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.For_indexContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5239
 *                 return self.getTypedRuleContexts(ModelicaParser.For_indexContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.For_indexContext,i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5239, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5239, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_For_indexContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5239, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5239, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5239, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5239, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5239, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":5235
 *             self.parser = parser
 * 
 *         def for_index(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.For_indexContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_indicesContext.for_index", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5242
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_for_indices
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18For_indicesContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18For_indicesContext_5getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18For_indicesContext_5getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18For_indicesContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18For_indicesContext_4getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18For_indicesContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":5243
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_for_indices             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_for_indices); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":5242
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_for_indices
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_indicesContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5245
 *             return ModelicaParser.RULE_for_indices
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterFor_indices" ):
 *                 listener.enterFor_indices(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18For_indicesContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18For_indicesContext_7enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18For_indicesContext_7enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18For_indicesContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 5245, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 5245, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5245, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_indicesContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18For_indicesContext_6enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18For_indicesContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":5246
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterFor_indices" ):             # <<<<<<<<<<<<<<
 *                 listener.enterFor_indices(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterFor_indices); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5246, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5247
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterFor_indices" ):
 *                 listener.enterFor_indices(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterFor_indices); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5247, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5247, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5247, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5247, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5247, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5247, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":5246
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterFor_indices" ):             # <<<<<<<<<<<<<<
 *                 listener.enterFor_indices(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5245
 *             return ModelicaParser.RULE_for_indices
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterFor_indices" ):
 *                 listener.enterFor_indices(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_indicesContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5249
 *                 listener.enterFor_indices(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitFor_indices" ):
 *                 listener.exitFor_indices(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18For_indicesContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18For_indicesContext_9exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18For_indicesContext_9exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18For_indicesContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 5249, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 5249, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5249, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_indicesContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18For_indicesContext_8exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18For_indicesContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":5250
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitFor_indices" ):             # <<<<<<<<<<<<<<
 *                 listener.exitFor_indices(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitFor_indices); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5250, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5251
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitFor_indices" ):
 *                 listener.exitFor_indices(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitFor_indices); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5251, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5251, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5251, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5251, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5251, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":5250
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitFor_indices" ):             # <<<<<<<<<<<<<<
 *                 listener.exitFor_indices(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5249
 *                 listener.enterFor_indices(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitFor_indices" ):
 *                 listener.exitFor_indices(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_indicesContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5253
 *                 listener.exitFor_indices(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitFor_indices" ):
 *                 return visitor.visitFor_indices(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18For_indicesContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18For_indicesContext_11accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18For_indicesContext_11accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18For_indicesContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 5253, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 5253, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5253, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_indicesContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18For_indicesContext_10accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18For_indicesContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":5254
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitFor_indices" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitFor_indices(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitFor_indices); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5254, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5255
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitFor_indices" ):
 *                 return visitor.visitFor_indices(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitFor_indices); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5255, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5255, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5255, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5255, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5255, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5255, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":5254
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitFor_indices" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitFor_indices(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5257
 *                 return visitor.visitFor_indices(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5257, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5257, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5257, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5257, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5257, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5257, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":5253
 *                 listener.exitFor_indices(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitFor_indices" ):
 *                 return visitor.visitFor_indices(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_indicesContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5262
 * 
 * 
 *     def for_indices(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.For_indicesContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_99for_indices(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_99for_indices = {"for_indices", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_99for_indices, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_99for_indices(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("for_indices (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_98for_indices(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_98for_indices(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__alt = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  char const *__pyx_t_22;
  __Pyx_RefNannySetupContext("for_indices", 0);

  /* "pymola/generated/ModelicaParser.pyx":5264
 *     def for_indices(self):
 * 
 *         localctx = ModelicaParser.For_indicesContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 96, self.RULE_for_indices)
 *         try:
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_For_indicesContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5264, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5264, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5264, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5264, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5265
 * 
 *         localctx = ModelicaParser.For_indicesContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 96, self.RULE_for_indices)             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_for_indices); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_96, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5265, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_96, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5265, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5265, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_96);
    __Pyx_GIVEREF(__pyx_int_96);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_96);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5265, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5266
 *         localctx = ModelicaParser.For_indicesContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 96, self.RULE_for_indices)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 729
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":5267
 *         self.enterRule(localctx, 96, self.RULE_for_indices)
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 729
 *             self.for_index()
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5267, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5267, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5267, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5267, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5267, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5268
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 729             # <<<<<<<<<<<<<<
 *             self.for_index()
 *             self.state = 734
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_729) < 0) __PYX_ERR(0, 5268, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5269
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 729
 *             self.for_index()             # <<<<<<<<<<<<<<
 *             self.state = 734
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_for_index); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5269, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5269, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5269, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5270
 *             self.state = 729
 *             self.for_index()
 *             self.state = 734             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _alt = self._interp.adaptivePredict(self._input,96,self._ctx)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_734) < 0) __PYX_ERR(0, 5270, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5271
 *             self.for_index()
 *             self.state = 734
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _alt = self._interp.adaptivePredict(self._input,96,self._ctx)
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5271, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5271, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5271, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5271, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5271, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5271, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5271, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5272
 *             self.state = 734
 *             self._errHandler.sync(self)
 *             _alt = self._interp.adaptivePredict(self._input,96,self._ctx)             # <<<<<<<<<<<<<<
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                 if _alt==1:
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_interp); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5272, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_adaptivePredict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5272, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5272, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5272, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_2)) {
          PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_t_7, __pyx_int_96, __pyx_t_3};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5272, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
          PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_t_7, __pyx_int_96, __pyx_t_3};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5272, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5272, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_7);
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_t_7);
          __Pyx_INCREF(__pyx_int_96);
          __Pyx_GIVEREF(__pyx_int_96);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_int_96);
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_6, __pyx_t_3);
          __pyx_t_7 = 0;
          __pyx_t_3 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5272, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_v__alt = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5273
 *             self._errHandler.sync(self)
 *             _alt = self._interp.adaptivePredict(self._input,96,self._ctx)
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:             # <<<<<<<<<<<<<<
 *                 if _alt==1:
 *                     self.state = 730
 */
        while (1) {
          __pyx_t_1 = PyObject_RichCompare(__pyx_v__alt, __pyx_int_2, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5273, __pyx_L6_error)
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5273, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (__pyx_t_12) {
          } else {
            __pyx_t_11 = __pyx_t_12;
            goto __pyx_L16_bool_binop_done;
          }
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ATN); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5273, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_INVALID_ALT_NUMBER); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5273, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyObject_RichCompare(__pyx_v__alt, __pyx_t_2, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5273, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5273, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_11 = __pyx_t_12;
          __pyx_L16_bool_binop_done:;
          if (!__pyx_t_11) break;

          /* "pymola/generated/ModelicaParser.pyx":5274
 *             _alt = self._interp.adaptivePredict(self._input,96,self._ctx)
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                 if _alt==1:             # <<<<<<<<<<<<<<
 *                     self.state = 730
 *                     self.match(ModelicaParser.T__20)
 */
          __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v__alt, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5274, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 5274, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (__pyx_t_11) {

            /* "pymola/generated/ModelicaParser.pyx":5275
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                 if _alt==1:
 *                     self.state = 730             # <<<<<<<<<<<<<<
 *                     self.match(ModelicaParser.T__20)
 *                     self.state = 731
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_730) < 0) __PYX_ERR(0, 5275, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":5276
 *                 if _alt==1:
 *                     self.state = 730
 *                     self.match(ModelicaParser.T__20)             # <<<<<<<<<<<<<<
 *                     self.state = 731
 *                     self.for_index()
 */
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5276, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5276, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_T__20); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5276, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_5 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_5)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_5);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
              }
            }
            if (!__pyx_t_5) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5276, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_2)) {
                PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5276, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
                PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5276, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              } else
              #endif
              {
                __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5276, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_7);
                __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
                __Pyx_GIVEREF(__pyx_t_3);
                PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_3);
                __pyx_t_3 = 0;
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5276, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":5277
 *                     self.state = 730
 *                     self.match(ModelicaParser.T__20)
 *                     self.state = 731             # <<<<<<<<<<<<<<
 *                     self.for_index()
 *                 self.state = 736
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_731) < 0) __PYX_ERR(0, 5277, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":5278
 *                     self.match(ModelicaParser.T__20)
 *                     self.state = 731
 *                     self.for_index()             # <<<<<<<<<<<<<<
 *                 self.state = 736
 *                 self._errHandler.sync(self)
 */
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_for_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5278, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_7 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_7)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_7);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
              }
            }
            if (__pyx_t_7) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5278, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else {
              __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5278, __pyx_L6_error)
            }
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":5274
 *             _alt = self._interp.adaptivePredict(self._input,96,self._ctx)
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                 if _alt==1:             # <<<<<<<<<<<<<<
 *                     self.state = 730
 *                     self.match(ModelicaParser.T__20)
 */
          }

          /* "pymola/generated/ModelicaParser.pyx":5279
 *                     self.state = 731
 *                     self.for_index()
 *                 self.state = 736             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _alt = self._interp.adaptivePredict(self._input,96,self._ctx)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_736) < 0) __PYX_ERR(0, 5279, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5280
 *                     self.for_index()
 *                 self.state = 736
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _alt = self._interp.adaptivePredict(self._input,96,self._ctx)
 * 
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5280, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5280, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5280, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5280, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5280, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5280, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5280, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5281
 *                 self.state = 736
 *                 self._errHandler.sync(self)
 *                 _alt = self._interp.adaptivePredict(self._input,96,self._ctx)             # <<<<<<<<<<<<<<
 * 
 *         except RecognitionException as re:
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_interp); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5281, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_adaptivePredict); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5281, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5281, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5281, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_t_7, __pyx_int_96, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5281, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_t_7, __pyx_int_96, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5281, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5281, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_GIVEREF(__pyx_t_7);
            PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_t_7);
            __Pyx_INCREF(__pyx_int_96);
            __Pyx_GIVEREF(__pyx_int_96);
            PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_int_96);
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_4, 2+__pyx_t_6, __pyx_t_2);
            __pyx_t_7 = 0;
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5281, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF_SET(__pyx_v__alt, __pyx_t_1);
          __pyx_t_1 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":5266
 *         localctx = ModelicaParser.For_indicesContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 96, self.RULE_for_indices)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 729
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":5283
 *                 _alt = self._interp.adaptivePredict(self._input,96,self._ctx)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5283, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.for_indices", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_3, &__pyx_t_4) < 0) __PYX_ERR(0, 5283, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __pyx_v_re = __pyx_t_3;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":5284
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 5284, __pyx_L24_error)

          /* "pymola/generated/ModelicaParser.pyx":5285
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5285, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5285, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5285, __pyx_L24_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5285, __pyx_L24_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 5285, __pyx_L24_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_7) {
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_7); __pyx_t_7 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_13, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5285, __pyx_L24_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5286
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5286, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 5286, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5286, __pyx_L24_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5286, __pyx_L24_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5286, __pyx_L24_error)
            __Pyx_GOTREF(__pyx_t_7);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5286, __pyx_L24_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":5283
 *                 _alt = self._interp.adaptivePredict(self._input,96,self._ctx)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L25;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L24_error:;
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_6 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L8_except_error;
          }
          __pyx_L25:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":5266
 *         localctx = ModelicaParser.For_indicesContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 96, self.RULE_for_indices)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 729
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":5288
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5288, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5288, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5288, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_14 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5288, __pyx_L31_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5288, __pyx_L31_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5288, __pyx_L31_error)
        }
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
      __pyx_L31_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":5289
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class For_indexContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":5262
 * 
 * 
 *     def for_indices(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.For_indicesContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.for_indices", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__alt);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5293
 *     class For_indexContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16For_indexContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_16For_indexContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_16For_indexContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16For_indexContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 5293, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 5293, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5293, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_indexContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_16For_indexContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16For_indexContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":5294
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 5294, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5294, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5294, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5294, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5294, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5295
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def IDENT(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 5295, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5293
 *     class For_indexContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_indexContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5297
 *             self.parser = parser
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16For_indexContext_3IDENT(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_16For_indexContext_3IDENT = {"IDENT", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_16For_indexContext_3IDENT, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16For_indexContext_3IDENT(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("IDENT (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_16For_indexContext_2IDENT(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16For_indexContext_2IDENT(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("IDENT", 0);

  /* "pymola/generated/ModelicaParser.pyx":5298
 * 
 *         def IDENT(self):
 *             return self.getToken(ModelicaParser.IDENT, 0)             # <<<<<<<<<<<<<<
 * 
 *         def expression(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getToken); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5298, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5298, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5298, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5298, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":5297
 *             self.parser = parser
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_indexContext.IDENT", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5300
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 *         def expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16For_indexContext_5expression(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_16For_indexContext_5expression = {"expression", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_16For_indexContext_5expression, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16For_indexContext_5expression(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("expression (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_16For_indexContext_4expression(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16For_indexContext_4expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("expression", 0);

  /* "pymola/generated/ModelicaParser.pyx":5301
 * 
 *         def expression(self):
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5301, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5301, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_ExpressionContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5301, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5301, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5301, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5301, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5301, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":5300
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 *         def expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_indexContext.expression", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5304
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_for_index
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16For_indexContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_16For_indexContext_7getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_16For_indexContext_7getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16For_indexContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_16For_indexContext_6getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16For_indexContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":5305
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_for_index             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_for_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":5304
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_for_index
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_indexContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5307
 *             return ModelicaParser.RULE_for_index
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterFor_index" ):
 *                 listener.enterFor_index(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16For_indexContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_16For_indexContext_9enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_16For_indexContext_9enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16For_indexContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 5307, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 5307, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5307, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_indexContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_16For_indexContext_8enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16For_indexContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":5308
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterFor_index" ):             # <<<<<<<<<<<<<<
 *                 listener.enterFor_index(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterFor_index); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5308, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5309
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterFor_index" ):
 *                 listener.enterFor_index(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterFor_index); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5309, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5309, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5309, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5309, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5309, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5309, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":5308
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterFor_index" ):             # <<<<<<<<<<<<<<
 *                 listener.enterFor_index(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5307
 *             return ModelicaParser.RULE_for_index
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterFor_index" ):
 *                 listener.enterFor_index(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_indexContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5311
 *                 listener.enterFor_index(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitFor_index" ):
 *                 listener.exitFor_index(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16For_indexContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_16For_indexContext_11exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_16For_indexContext_11exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16For_indexContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 5311, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 5311, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5311, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_indexContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_16For_indexContext_10exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16For_indexContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":5312
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitFor_index" ):             # <<<<<<<<<<<<<<
 *                 listener.exitFor_index(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitFor_index); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5312, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5313
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitFor_index" ):
 *                 listener.exitFor_index(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitFor_index); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5313, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5313, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5313, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5313, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5313, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5313, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":5312
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitFor_index" ):             # <<<<<<<<<<<<<<
 *                 listener.exitFor_index(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5311
 *                 listener.enterFor_index(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitFor_index" ):
 *                 listener.exitFor_index(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_indexContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5315
 *                 listener.exitFor_index(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitFor_index" ):
 *                 return visitor.visitFor_index(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16For_indexContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_16For_indexContext_13accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_16For_indexContext_13accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16For_indexContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 5315, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 5315, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5315, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_indexContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_16For_indexContext_12accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16For_indexContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":5316
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitFor_index" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitFor_index(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitFor_index); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5316, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5317
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitFor_index" ):
 *                 return visitor.visitFor_index(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitFor_index); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5317, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5317, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5317, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5317, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5317, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5317, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":5316
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitFor_index" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitFor_index(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5319
 *                 return visitor.visitFor_index(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5319, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5319, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5319, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5319, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5319, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5319, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":5315
 *                 listener.exitFor_index(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitFor_index" ):
 *                 return visitor.visitFor_index(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.For_indexContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5324
 * 
 * 
 *     def for_index(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.For_indexContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_101for_index(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_101for_index = {"for_index", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_101for_index, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_101for_index(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("for_index (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_100for_index(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_100for_index(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  char const *__pyx_t_21;
  __Pyx_RefNannySetupContext("for_index", 0);

  /* "pymola/generated/ModelicaParser.pyx":5326
 *     def for_index(self):
 * 
 *         localctx = ModelicaParser.For_indexContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 98, self.RULE_for_index)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_For_indexContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5326, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5326, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5326, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5326, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5327
 * 
 *         localctx = ModelicaParser.For_indexContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 98, self.RULE_for_index)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_for_index); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_98, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5327, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_98, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5327, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5327, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_98);
    __Pyx_GIVEREF(__pyx_int_98);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_98);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5327, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5328
 *         localctx = ModelicaParser.For_indexContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 98, self.RULE_for_index)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 5328, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5329
 *         self.enterRule(localctx, 98, self.RULE_for_index)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 737
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":5330
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 737
 *             self.match(ModelicaParser.IDENT)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5330, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5330, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5330, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5330, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5330, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5331
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 737             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.IDENT)
 *             self.state = 740
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_737) < 0) __PYX_ERR(0, 5331, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5332
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 737
 *             self.match(ModelicaParser.IDENT)             # <<<<<<<<<<<<<<
 *             self.state = 740
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5332, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5332, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5332, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5332, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5332, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5332, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5332, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5332, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5333
 *             self.state = 737
 *             self.match(ModelicaParser.IDENT)
 *             self.state = 740             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_740) < 0) __PYX_ERR(0, 5333, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5334
 *             self.match(ModelicaParser.IDENT)
 *             self.state = 740
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__50:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5334, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5334, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5334, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5334, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5334, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5334, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5334, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5335
 *             self.state = 740
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.T__50:
 *                 self.state = 738
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5335, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5335, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__915, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5335, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_v__la = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5336
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__50:             # <<<<<<<<<<<<<<
 *                 self.state = 738
 *                 self.match(ModelicaParser.T__50)
 */
        __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5336, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__50); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5336, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5336, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 5336, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":5337
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__50:
 *                 self.state = 738             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__50)
 *                 self.state = 739
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_738) < 0) __PYX_ERR(0, 5337, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5338
 *             if _la==ModelicaParser.T__50:
 *                 self.state = 738
 *                 self.match(ModelicaParser.T__50)             # <<<<<<<<<<<<<<
 *                 self.state = 739
 *                 self.expression()
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5338, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5338, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__50); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5338, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5338, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5338, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5338, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5338, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5338, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5339
 *                 self.state = 738
 *                 self.match(ModelicaParser.T__50)
 *                 self.state = 739             # <<<<<<<<<<<<<<
 *                 self.expression()
 * 
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_739) < 0) __PYX_ERR(0, 5339, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5340
 *                 self.match(ModelicaParser.T__50)
 *                 self.state = 739
 *                 self.expression()             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expression); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5340, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5340, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5340, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5336
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__50:             # <<<<<<<<<<<<<<
 *                 self.state = 738
 *                 self.match(ModelicaParser.T__50)
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":5329
 *         self.enterRule(localctx, 98, self.RULE_for_index)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 737
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":5343
 * 
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5343, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.for_index", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_4, &__pyx_t_7) < 0) __PYX_ERR(0, 5343, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_4);
        __pyx_v_re = __pyx_t_4;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":5344
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 5344, __pyx_L20_error)

          /* "pymola/generated/ModelicaParser.pyx":5345
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5345, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5345, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5345, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5345, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_12 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 5345, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_12);
            if (__pyx_t_2) {
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_2); __pyx_t_2 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_12, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5345, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5346
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5346, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 5346, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_12);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_12, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5346, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5346, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5346, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5346, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":5343
 * 
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L21;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L20_error:;
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __pyx_t_6 = __pyx_lineno; __pyx_t_13 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
            }
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_ErrRestore(__pyx_t_15, __pyx_t_16, __pyx_t_17);
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_13; __pyx_filename = __pyx_t_14;
            goto __pyx_L8_except_error;
          }
          __pyx_L21:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":5329
 *         self.enterRule(localctx, 98, self.RULE_for_index)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 737
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":5348
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5348, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5348, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5348, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_19, &__pyx_t_18);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_18);
      __pyx_t_13 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_21 = __pyx_filename;
      {
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5348, __pyx_L27_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5348, __pyx_L27_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5348, __pyx_L27_error)
        }
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __pyx_lineno = __pyx_t_13; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_21;
      goto __pyx_L1_error;
      __pyx_L27_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":5349
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class While_statementContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":5324
 * 
 * 
 *     def for_index(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.For_indexContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.for_index", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5353
 *     class While_statementContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22While_statementContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22While_statementContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22While_statementContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22While_statementContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 5353, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 5353, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5353, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.While_statementContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22While_statementContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22While_statementContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":5354
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 5354, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5354, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5354, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5354, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5354, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5355
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def expression(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 5355, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5353
 *     class While_statementContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.While_statementContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5357
 *             self.parser = parser
 * 
 *         def expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22While_statementContext_3expression(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22While_statementContext_3expression = {"expression", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22While_statementContext_3expression, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22While_statementContext_3expression(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("expression (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22While_statementContext_2expression(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22While_statementContext_2expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("expression", 0);

  /* "pymola/generated/ModelicaParser.pyx":5358
 * 
 *         def expression(self):
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_ExpressionContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5358, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5358, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5358, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5358, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":5357
 *             self.parser = parser
 * 
 *         def expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.While_statementContext.expression", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5361
 * 
 * 
 *         def statement(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.StatementContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22While_statementContext_5statement(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22While_statementContext_5statement = {"statement", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22While_statementContext_5statement, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22While_statementContext_5statement(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("statement (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "statement") < 0)) __PYX_ERR(0, 5361, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("statement", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5361, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.While_statementContext.statement", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22While_statementContext_4statement(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22While_statementContext_4statement(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("statement", 0);

  /* "pymola/generated/ModelicaParser.pyx":5362
 * 
 *         def statement(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.StatementContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5363
 *         def statement(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.StatementContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.StatementContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5363, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5363, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_StatementContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5363, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5363, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5363, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5363, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5363, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5363, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":5362
 * 
 *         def statement(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.StatementContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5365
 *                 return self.getTypedRuleContexts(ModelicaParser.StatementContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.StatementContext,i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5365, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5365, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_StatementContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5365, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5365, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5365, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5365, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5365, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":5361
 * 
 * 
 *         def statement(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.StatementContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.While_statementContext.statement", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5368
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_while_statement
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22While_statementContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22While_statementContext_7getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22While_statementContext_7getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22While_statementContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22While_statementContext_6getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22While_statementContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":5369
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_while_statement             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_while_statement); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":5368
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_while_statement
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.While_statementContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5371
 *             return ModelicaParser.RULE_while_statement
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterWhile_statement" ):
 *                 listener.enterWhile_statement(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22While_statementContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22While_statementContext_9enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22While_statementContext_9enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22While_statementContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 5371, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 5371, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5371, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.While_statementContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22While_statementContext_8enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22While_statementContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":5372
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterWhile_statement" ):             # <<<<<<<<<<<<<<
 *                 listener.enterWhile_statement(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterWhile_statement); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5372, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5373
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterWhile_statement" ):
 *                 listener.enterWhile_statement(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterWhile_statement); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5373, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5373, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5373, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5373, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5373, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5373, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":5372
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterWhile_statement" ):             # <<<<<<<<<<<<<<
 *                 listener.enterWhile_statement(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5371
 *             return ModelicaParser.RULE_while_statement
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterWhile_statement" ):
 *                 listener.enterWhile_statement(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.While_statementContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5375
 *                 listener.enterWhile_statement(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitWhile_statement" ):
 *                 listener.exitWhile_statement(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22While_statementContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22While_statementContext_11exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22While_statementContext_11exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22While_statementContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 5375, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 5375, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5375, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.While_statementContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22While_statementContext_10exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22While_statementContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":5376
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitWhile_statement" ):             # <<<<<<<<<<<<<<
 *                 listener.exitWhile_statement(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitWhile_statement); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5376, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5377
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitWhile_statement" ):
 *                 listener.exitWhile_statement(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitWhile_statement); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5377, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5377, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5377, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5377, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5377, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5377, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":5376
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitWhile_statement" ):             # <<<<<<<<<<<<<<
 *                 listener.exitWhile_statement(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5375
 *                 listener.enterWhile_statement(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitWhile_statement" ):
 *                 listener.exitWhile_statement(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.While_statementContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5379
 *                 listener.exitWhile_statement(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitWhile_statement" ):
 *                 return visitor.visitWhile_statement(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22While_statementContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22While_statementContext_13accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22While_statementContext_13accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22While_statementContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 5379, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 5379, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5379, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.While_statementContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22While_statementContext_12accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22While_statementContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":5380
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitWhile_statement" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitWhile_statement(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitWhile_statement); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5380, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5381
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitWhile_statement" ):
 *                 return visitor.visitWhile_statement(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitWhile_statement); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5381, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5381, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5381, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5381, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5381, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5381, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":5380
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitWhile_statement" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitWhile_statement(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5383
 *                 return visitor.visitWhile_statement(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5383, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5383, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5383, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5383, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5383, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5383, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":5379
 *                 listener.exitWhile_statement(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitWhile_statement" ):
 *                 return visitor.visitWhile_statement(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.While_statementContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5388
 * 
 * 
 *     def while_statement(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.While_statementContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_103while_statement(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_103while_statement = {"while_statement", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_103while_statement, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_103while_statement(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("while_statement (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_102while_statement(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_102while_statement(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  char const *__pyx_t_22;
  __Pyx_RefNannySetupContext("while_statement", 0);

  /* "pymola/generated/ModelicaParser.pyx":5390
 *     def while_statement(self):
 * 
 *         localctx = ModelicaParser.While_statementContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 100, self.RULE_while_statement)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5390, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_While_statementContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5390, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5390, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5390, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5390, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5390, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5390, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5390, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5391
 * 
 *         localctx = ModelicaParser.While_statementContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 100, self.RULE_while_statement)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_while_statement); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_100, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5391, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_100, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5391, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5391, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_100);
    __Pyx_GIVEREF(__pyx_int_100);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_100);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5391, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5392
 *         localctx = ModelicaParser.While_statementContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 100, self.RULE_while_statement)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 5392, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5393
 *         self.enterRule(localctx, 100, self.RULE_while_statement)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 742
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":5394
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 742
 *             self.match(ModelicaParser.T__51)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5394, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5394, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5394, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5394, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5394, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5395
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 742             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__51)
 *             self.state = 743
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_742) < 0) __PYX_ERR(0, 5395, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5396
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 742
 *             self.match(ModelicaParser.T__51)             # <<<<<<<<<<<<<<
 *             self.state = 743
 *             self.expression()
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5396, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5396, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__51); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5396, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5396, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5396, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5396, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5396, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5396, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5397
 *             self.state = 742
 *             self.match(ModelicaParser.T__51)
 *             self.state = 743             # <<<<<<<<<<<<<<
 *             self.expression()
 *             self.state = 744
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_743) < 0) __PYX_ERR(0, 5397, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5398
 *             self.match(ModelicaParser.T__51)
 *             self.state = 743
 *             self.expression()             # <<<<<<<<<<<<<<
 *             self.state = 744
 *             self.match(ModelicaParser.T__49)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expression); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5398, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5398, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5398, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5399
 *             self.state = 743
 *             self.expression()
 *             self.state = 744             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__49)
 *             self.state = 750
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_744) < 0) __PYX_ERR(0, 5399, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5400
 *             self.expression()
 *             self.state = 744
 *             self.match(ModelicaParser.T__49)             # <<<<<<<<<<<<<<
 *             self.state = 750
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5400, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5400, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__49); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5400, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5400, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5400, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5400, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5400, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4); __pyx_t_4 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5400, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5401
 *             self.state = 744
 *             self.match(ModelicaParser.T__49)
 *             self.state = 750             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_750) < 0) __PYX_ERR(0, 5401, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5402
 *             self.match(ModelicaParser.T__49)
 *             self.state = 750
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5402, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5402, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5402, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5402, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5402, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5402, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5402, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5403
 *             self.state = 750
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:
 *                 self.state = 745
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5403, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5403, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__916, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5403, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v__la = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5404
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:             # <<<<<<<<<<<<<<
 *                 self.state = 745
 *                 self.statement()
 */
        while (1) {
          __pyx_t_1 = __Pyx_PyInt_From_long((~0x3f)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5404, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_7 = PyNumber_And(__pyx_v__la, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5404, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_t_7, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5404, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5404, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (!__pyx_t_12) {
            goto __pyx_L17_next_or;
          } else {
          }
          __pyx_t_1 = PyNumber_Lshift(__pyx_int_1, __pyx_v__la); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5404, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5404, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__16); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5404, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5404, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5404, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__39); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5404, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5404, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyNumber_Or(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5404, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5404, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__43); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5404, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5404, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Or(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5404, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5404, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__44); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5404, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5404, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyNumber_Or(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5404, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5404, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__48); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5404, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5404, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Or(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5404, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5404, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__51); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5404, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5404, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyNumber_Or(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5404, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5404, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__52); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5404, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5404, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Or(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5404, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_And(__pyx_t_1, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5404, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyObject_RichCompare(__pyx_t_2, __pyx_int_0, Py_NE); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5404, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5404, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (!__pyx_t_12) {
          } else {
            __pyx_t_11 = __pyx_t_12;
            goto __pyx_L16_bool_binop_done;
          }
          __pyx_L17_next_or:;
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5404, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5404, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyObject_RichCompare(__pyx_v__la, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5404, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5404, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_11 = __pyx_t_12;
          __pyx_L16_bool_binop_done:;
          if (!__pyx_t_11) break;

          /* "pymola/generated/ModelicaParser.pyx":5405
 *             _la = self._input.LA(1)
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:
 *                 self.state = 745             # <<<<<<<<<<<<<<
 *                 self.statement()
 *                 self.state = 746
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_745) < 0) __PYX_ERR(0, 5405, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5406
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:
 *                 self.state = 745
 *                 self.statement()             # <<<<<<<<<<<<<<
 *                 self.state = 746
 *                 self.match(ModelicaParser.T__0)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_statement); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5406, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_1 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (__pyx_t_1) {
            __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5406, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          } else {
            __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5406, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5407
 *                 self.state = 745
 *                 self.statement()
 *                 self.state = 746             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__0)
 *                 self.state = 752
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_746) < 0) __PYX_ERR(0, 5407, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5408
 *                 self.statement()
 *                 self.state = 746
 *                 self.match(ModelicaParser.T__0)             # <<<<<<<<<<<<<<
 *                 self.state = 752
 *                 self._errHandler.sync(self)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5408, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5408, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5408, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_1) {
            __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5408, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_t_3};
              __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5408, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_t_3};
              __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5408, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5408, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1); __pyx_t_1 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5408, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5409
 *                 self.state = 746
 *                 self.match(ModelicaParser.T__0)
 *                 self.state = 752             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_752) < 0) __PYX_ERR(0, 5409, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5410
 *                 self.match(ModelicaParser.T__0)
 *                 self.state = 752
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 * 
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5410, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_sync); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5410, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5410, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5410, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_7);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5410, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_7);
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5410, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
              __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5410, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5411
 *                 self.state = 752
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *             self.state = 753
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5411, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_LA); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5411, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__917, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5411, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_7);
          __pyx_t_7 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":5413
 *                 _la = self._input.LA(1)
 * 
 *             self.state = 753             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__13)
 *             self.state = 754
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_753) < 0) __PYX_ERR(0, 5413, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5414
 * 
 *             self.state = 753
 *             self.match(ModelicaParser.T__13)             # <<<<<<<<<<<<<<
 *             self.state = 754
 *             self.match(ModelicaParser.T__51)
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5414, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5414, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__13); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5414, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5414, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_7);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_2};
            __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5414, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_2};
            __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5414, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_1 = PyTuple_New(1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5414, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_1, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_1, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5414, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5415
 *             self.state = 753
 *             self.match(ModelicaParser.T__13)
 *             self.state = 754             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__51)
 *         except RecognitionException as re:
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_754) < 0) __PYX_ERR(0, 5415, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5416
 *             self.match(ModelicaParser.T__13)
 *             self.state = 754
 *             self.match(ModelicaParser.T__51)             # <<<<<<<<<<<<<<
 *         except RecognitionException as re:
 *             localctx.exception = re
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5416, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5416, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__51); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5416, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (!__pyx_t_1) {
          __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5416, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_7);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_t_2};
            __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5416, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_t_2};
            __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5416, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5416, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1); __pyx_t_1 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5416, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5393
 *         self.enterRule(localctx, 100, self.RULE_while_statement)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 742
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "pymola/generated/ModelicaParser.pyx":5417
 *             self.state = 754
 *             self.match(ModelicaParser.T__51)
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5417, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.while_statement", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_4, &__pyx_t_3) < 0) __PYX_ERR(0, 5417, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __pyx_v_re = __pyx_t_4;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":5418
 *             self.match(ModelicaParser.T__51)
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 5418, __pyx_L24_error)

          /* "pymola/generated/ModelicaParser.pyx":5419
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5419, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5419, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5419, __pyx_L24_error)
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5419, __pyx_L24_error)
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 5419, __pyx_L24_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_1) {
              __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_1); __pyx_t_1 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_13, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5419, __pyx_L24_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5420
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5420, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 5420, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5420, __pyx_L24_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5420, __pyx_L24_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_1 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5420, __pyx_L24_error)
            __Pyx_GOTREF(__pyx_t_1);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5420, __pyx_L24_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":5417
 *             self.state = 754
 *             self.match(ModelicaParser.T__51)
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L25;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L24_error:;
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_6 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L8_except_error;
          }
          __pyx_L25:;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":5393
 *         self.enterRule(localctx, 100, self.RULE_while_statement)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 742
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":5422
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5422, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (__pyx_t_7) {
        __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5422, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      } else {
        __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5422, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_14 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5422, __pyx_L31_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (__pyx_t_7) {
          __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5422, __pyx_L31_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else {
          __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5422, __pyx_L31_error)
        }
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
      __pyx_L31_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":5423
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class When_equationContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":5388
 * 
 * 
 *     def while_statement(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.While_statementContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.while_statement", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5427
 *     class When_equationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20When_equationContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20When_equationContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20When_equationContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20When_equationContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 5427, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 5427, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5427, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.When_equationContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20When_equationContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20When_equationContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":5428
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 5428, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5428, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5428, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5428, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5428, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5428, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5428, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5428, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5429
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def expression(self, i:int=None):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 5429, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5427
 *     class When_equationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.When_equationContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5431
 *             self.parser = parser
 * 
 *         def expression(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20When_equationContext_3expression(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20When_equationContext_3expression = {"expression", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20When_equationContext_3expression, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20When_equationContext_3expression(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("expression (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "expression") < 0)) __PYX_ERR(0, 5431, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("expression", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5431, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.When_equationContext.expression", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20When_equationContext_2expression(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20When_equationContext_2expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("expression", 0);

  /* "pymola/generated/ModelicaParser.pyx":5432
 * 
 *         def expression(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5433
 *         def expression(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.ExpressionContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5433, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5433, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_ExpressionContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5433, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5433, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5433, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5433, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5433, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5433, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":5432
 * 
 *         def expression(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5435
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.ExpressionContext,i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5435, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5435, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_ExpressionContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5435, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5435, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5435, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5435, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5435, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":5431
 *             self.parser = parser
 * 
 *         def expression(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.When_equationContext.expression", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5438
 * 
 * 
 *         def equation(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.EquationContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20When_equationContext_5equation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20When_equationContext_5equation = {"equation", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20When_equationContext_5equation, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20When_equationContext_5equation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("equation (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "equation") < 0)) __PYX_ERR(0, 5438, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("equation", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5438, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.When_equationContext.equation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20When_equationContext_4equation(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20When_equationContext_4equation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("equation", 0);

  /* "pymola/generated/ModelicaParser.pyx":5439
 * 
 *         def equation(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.EquationContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5440
 *         def equation(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.EquationContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.EquationContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5440, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5440, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_EquationContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5440, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5440, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5440, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5440, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5440, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5440, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":5439
 * 
 *         def equation(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.EquationContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5442
 *                 return self.getTypedRuleContexts(ModelicaParser.EquationContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.EquationContext,i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5442, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5442, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_EquationContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5442, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5442, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5442, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5442, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5442, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":5438
 * 
 * 
 *         def equation(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.EquationContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.When_equationContext.equation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5445
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_when_equation
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20When_equationContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20When_equationContext_7getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20When_equationContext_7getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20When_equationContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20When_equationContext_6getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20When_equationContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":5446
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_when_equation             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5446, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_when_equation); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5446, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":5445
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_when_equation
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.When_equationContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5448
 *             return ModelicaParser.RULE_when_equation
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterWhen_equation" ):
 *                 listener.enterWhen_equation(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20When_equationContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20When_equationContext_9enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20When_equationContext_9enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20When_equationContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 5448, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 5448, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5448, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.When_equationContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20When_equationContext_8enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20When_equationContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":5449
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterWhen_equation" ):             # <<<<<<<<<<<<<<
 *                 listener.enterWhen_equation(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterWhen_equation); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5449, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5450
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterWhen_equation" ):
 *                 listener.enterWhen_equation(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterWhen_equation); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5450, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5450, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5450, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5450, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5450, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5450, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":5449
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterWhen_equation" ):             # <<<<<<<<<<<<<<
 *                 listener.enterWhen_equation(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5448
 *             return ModelicaParser.RULE_when_equation
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterWhen_equation" ):
 *                 listener.enterWhen_equation(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.When_equationContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5452
 *                 listener.enterWhen_equation(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitWhen_equation" ):
 *                 listener.exitWhen_equation(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20When_equationContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20When_equationContext_11exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20When_equationContext_11exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20When_equationContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 5452, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 5452, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5452, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.When_equationContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20When_equationContext_10exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20When_equationContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":5453
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitWhen_equation" ):             # <<<<<<<<<<<<<<
 *                 listener.exitWhen_equation(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitWhen_equation); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5453, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5454
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitWhen_equation" ):
 *                 listener.exitWhen_equation(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitWhen_equation); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5454, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5454, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5454, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5454, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5454, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5454, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":5453
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitWhen_equation" ):             # <<<<<<<<<<<<<<
 *                 listener.exitWhen_equation(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5452
 *                 listener.enterWhen_equation(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitWhen_equation" ):
 *                 listener.exitWhen_equation(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.When_equationContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5456
 *                 listener.exitWhen_equation(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitWhen_equation" ):
 *                 return visitor.visitWhen_equation(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20When_equationContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20When_equationContext_13accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20When_equationContext_13accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20When_equationContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 5456, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 5456, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5456, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.When_equationContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20When_equationContext_12accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20When_equationContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":5457
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitWhen_equation" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitWhen_equation(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitWhen_equation); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5457, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5458
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitWhen_equation" ):
 *                 return visitor.visitWhen_equation(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitWhen_equation); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5458, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5458, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5458, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5458, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5458, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5458, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":5457
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitWhen_equation" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitWhen_equation(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5460
 *                 return visitor.visitWhen_equation(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5460, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5460, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5460, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5460, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5460, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5460, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":5456
 *                 listener.exitWhen_equation(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitWhen_equation" ):
 *                 return visitor.visitWhen_equation(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.When_equationContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5465
 * 
 * 
 *     def when_equation(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.When_equationContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_105when_equation(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_105when_equation = {"when_equation", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_105when_equation, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_105when_equation(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("when_equation (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_104when_equation(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_104when_equation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__alt = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  char const *__pyx_t_22;
  __Pyx_RefNannySetupContext("when_equation", 0);

  /* "pymola/generated/ModelicaParser.pyx":5467
 *     def when_equation(self):
 * 
 *         localctx = ModelicaParser.When_equationContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 102, self.RULE_when_equation)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5467, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_When_equationContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5467, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5467, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5467, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5467, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5467, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5467, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5467, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5468
 * 
 *         localctx = ModelicaParser.When_equationContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 102, self.RULE_when_equation)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5468, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_when_equation); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5468, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_102, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5468, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_102, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5468, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5468, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_102);
    __Pyx_GIVEREF(__pyx_int_102);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_102);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5468, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5469
 *         localctx = ModelicaParser.When_equationContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 102, self.RULE_when_equation)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 5469, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5470
 *         self.enterRule(localctx, 102, self.RULE_when_equation)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 756
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":5471
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 756
 *             self.match(ModelicaParser.T__52)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5471, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5471, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5471, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5471, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5471, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5472
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 756             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__52)
 *             self.state = 757
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_756) < 0) __PYX_ERR(0, 5472, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5473
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 756
 *             self.match(ModelicaParser.T__52)             # <<<<<<<<<<<<<<
 *             self.state = 757
 *             self.expression()
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5473, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5473, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__52); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5473, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5473, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5473, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5473, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5473, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5473, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5474
 *             self.state = 756
 *             self.match(ModelicaParser.T__52)
 *             self.state = 757             # <<<<<<<<<<<<<<
 *             self.expression()
 *             self.state = 758
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_757) < 0) __PYX_ERR(0, 5474, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5475
 *             self.match(ModelicaParser.T__52)
 *             self.state = 757
 *             self.expression()             # <<<<<<<<<<<<<<
 *             self.state = 758
 *             self.match(ModelicaParser.T__45)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expression); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5475, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5475, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5475, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5476
 *             self.state = 757
 *             self.expression()
 *             self.state = 758             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__45)
 *             self.state = 764
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_758) < 0) __PYX_ERR(0, 5476, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5477
 *             self.expression()
 *             self.state = 758
 *             self.match(ModelicaParser.T__45)             # <<<<<<<<<<<<<<
 *             self.state = 764
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5477, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5477, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__45); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5477, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5477, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5477, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5477, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5477, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4); __pyx_t_4 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5477, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5478
 *             self.state = 758
 *             self.match(ModelicaParser.T__45)
 *             self.state = 764             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _alt = self._interp.adaptivePredict(self._input,99,self._ctx)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_764) < 0) __PYX_ERR(0, 5478, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5479
 *             self.match(ModelicaParser.T__45)
 *             self.state = 764
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _alt = self._interp.adaptivePredict(self._input,99,self._ctx)
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5479, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5479, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5479, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5479, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5479, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5479, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5479, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5480
 *             self.state = 764
 *             self._errHandler.sync(self)
 *             _alt = self._interp.adaptivePredict(self._input,99,self._ctx)             # <<<<<<<<<<<<<<
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                 if _alt==1:
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_interp); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5480, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_adaptivePredict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5480, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5480, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5480, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_2)) {
          PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_t_7, __pyx_int_99, __pyx_t_3};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5480, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
          PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_t_7, __pyx_int_99, __pyx_t_3};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5480, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5480, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_7);
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_t_7);
          __Pyx_INCREF(__pyx_int_99);
          __Pyx_GIVEREF(__pyx_int_99);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_int_99);
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_6, __pyx_t_3);
          __pyx_t_7 = 0;
          __pyx_t_3 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5480, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_v__alt = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5481
 *             self._errHandler.sync(self)
 *             _alt = self._interp.adaptivePredict(self._input,99,self._ctx)
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:             # <<<<<<<<<<<<<<
 *                 if _alt==1:
 *                     self.state = 759
 */
        while (1) {
          __pyx_t_1 = PyObject_RichCompare(__pyx_v__alt, __pyx_int_2, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5481, __pyx_L6_error)
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5481, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (__pyx_t_12) {
          } else {
            __pyx_t_11 = __pyx_t_12;
            goto __pyx_L16_bool_binop_done;
          }
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ATN); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5481, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_INVALID_ALT_NUMBER); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5481, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyObject_RichCompare(__pyx_v__alt, __pyx_t_2, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5481, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5481, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_11 = __pyx_t_12;
          __pyx_L16_bool_binop_done:;
          if (!__pyx_t_11) break;

          /* "pymola/generated/ModelicaParser.pyx":5482
 *             _alt = self._interp.adaptivePredict(self._input,99,self._ctx)
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                 if _alt==1:             # <<<<<<<<<<<<<<
 *                     self.state = 759
 *                     self.equation()
 */
          __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v__alt, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5482, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 5482, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (__pyx_t_11) {

            /* "pymola/generated/ModelicaParser.pyx":5483
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                 if _alt==1:
 *                     self.state = 759             # <<<<<<<<<<<<<<
 *                     self.equation()
 *                     self.state = 760
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_759) < 0) __PYX_ERR(0, 5483, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":5484
 *                 if _alt==1:
 *                     self.state = 759
 *                     self.equation()             # <<<<<<<<<<<<<<
 *                     self.state = 760
 *                     self.match(ModelicaParser.T__0)
 */
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_equation); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5484, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_5 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_5)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_5);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
              }
            }
            if (__pyx_t_5) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5484, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else {
              __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5484, __pyx_L6_error)
            }
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":5485
 *                     self.state = 759
 *                     self.equation()
 *                     self.state = 760             # <<<<<<<<<<<<<<
 *                     self.match(ModelicaParser.T__0)
 *                 self.state = 766
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_760) < 0) __PYX_ERR(0, 5485, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":5486
 *                     self.equation()
 *                     self.state = 760
 *                     self.match(ModelicaParser.T__0)             # <<<<<<<<<<<<<<
 *                 self.state = 766
 *                 self._errHandler.sync(self)
 */
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5486, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5486, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_T__0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5486, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_5 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_5)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_5);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
              }
            }
            if (!__pyx_t_5) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5486, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_2)) {
                PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5486, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
                PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5486, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              } else
              #endif
              {
                __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5486, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_7);
                __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
                __Pyx_GIVEREF(__pyx_t_3);
                PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_3);
                __pyx_t_3 = 0;
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5486, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":5482
 *             _alt = self._interp.adaptivePredict(self._input,99,self._ctx)
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                 if _alt==1:             # <<<<<<<<<<<<<<
 *                     self.state = 759
 *                     self.equation()
 */
          }

          /* "pymola/generated/ModelicaParser.pyx":5487
 *                     self.state = 760
 *                     self.match(ModelicaParser.T__0)
 *                 self.state = 766             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _alt = self._interp.adaptivePredict(self._input,99,self._ctx)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_766) < 0) __PYX_ERR(0, 5487, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5488
 *                     self.match(ModelicaParser.T__0)
 *                 self.state = 766
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _alt = self._interp.adaptivePredict(self._input,99,self._ctx)
 * 
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5488, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5488, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5488, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5488, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5488, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5488, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5488, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5489
 *                 self.state = 766
 *                 self._errHandler.sync(self)
 *                 _alt = self._interp.adaptivePredict(self._input,99,self._ctx)             # <<<<<<<<<<<<<<
 * 
 *             self.state = 780
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_interp); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5489, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_adaptivePredict); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5489, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5489, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5489, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_t_7, __pyx_int_99, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5489, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_t_7, __pyx_int_99, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5489, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5489, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_GIVEREF(__pyx_t_7);
            PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_t_7);
            __Pyx_INCREF(__pyx_int_99);
            __Pyx_GIVEREF(__pyx_int_99);
            PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_int_99);
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_4, 2+__pyx_t_6, __pyx_t_2);
            __pyx_t_7 = 0;
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5489, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF_SET(__pyx_v__alt, __pyx_t_1);
          __pyx_t_1 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":5491
 *                 _alt = self._interp.adaptivePredict(self._input,99,self._ctx)
 * 
 *             self.state = 780             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_780) < 0) __PYX_ERR(0, 5491, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5492
 * 
 *             self.state = 780
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             while _la==ModelicaParser.T__53:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5492, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5492, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5492, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5492, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5492, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5492, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5492, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5493
 *             self.state = 780
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             while _la==ModelicaParser.T__53:
 *                 self.state = 767
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5493, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5493, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__918, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5493, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_v__la = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5494
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             while _la==ModelicaParser.T__53:             # <<<<<<<<<<<<<<
 *                 self.state = 767
 *                 self.match(ModelicaParser.T__53)
 */
        while (1) {
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5494, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__53); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5494, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5494, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 5494, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (!__pyx_t_11) break;

          /* "pymola/generated/ModelicaParser.pyx":5495
 *             _la = self._input.LA(1)
 *             while _la==ModelicaParser.T__53:
 *                 self.state = 767             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__53)
 *                 self.state = 768
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_767) < 0) __PYX_ERR(0, 5495, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5496
 *             while _la==ModelicaParser.T__53:
 *                 self.state = 767
 *                 self.match(ModelicaParser.T__53)             # <<<<<<<<<<<<<<
 *                 self.state = 768
 *                 self.expression()
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5496, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5496, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__53); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5496, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5496, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5496, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5496, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5496, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5496, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5497
 *                 self.state = 767
 *                 self.match(ModelicaParser.T__53)
 *                 self.state = 768             # <<<<<<<<<<<<<<
 *                 self.expression()
 *                 self.state = 769
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_768) < 0) __PYX_ERR(0, 5497, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5498
 *                 self.match(ModelicaParser.T__53)
 *                 self.state = 768
 *                 self.expression()             # <<<<<<<<<<<<<<
 *                 self.state = 769
 *                 self.match(ModelicaParser.T__45)
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expression); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5498, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5498, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5498, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5499
 *                 self.state = 768
 *                 self.expression()
 *                 self.state = 769             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__45)
 *                 self.state = 775
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_769) < 0) __PYX_ERR(0, 5499, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5500
 *                 self.expression()
 *                 self.state = 769
 *                 self.match(ModelicaParser.T__45)             # <<<<<<<<<<<<<<
 *                 self.state = 775
 *                 self._errHandler.sync(self)
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5500, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5500, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__45); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5500, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5500, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5500, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5500, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5500, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_7); __pyx_t_7 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5500, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5501
 *                 self.state = 769
 *                 self.match(ModelicaParser.T__45)
 *                 self.state = 775             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _alt = self._interp.adaptivePredict(self._input,100,self._ctx)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_775) < 0) __PYX_ERR(0, 5501, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5502
 *                 self.match(ModelicaParser.T__45)
 *                 self.state = 775
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _alt = self._interp.adaptivePredict(self._input,100,self._ctx)
 *                 while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5502, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5502, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_4) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5502, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5502, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5502, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5502, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __pyx_t_4 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5502, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5503
 *                 self.state = 775
 *                 self._errHandler.sync(self)
 *                 _alt = self._interp.adaptivePredict(self._input,100,self._ctx)             # <<<<<<<<<<<<<<
 *                 while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                     if _alt==1:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_interp); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5503, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_adaptivePredict); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5503, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5503, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5503, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[4] = {__pyx_t_7, __pyx_t_2, __pyx_int_100, __pyx_t_4};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5503, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[4] = {__pyx_t_7, __pyx_t_2, __pyx_int_100, __pyx_t_4};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5503, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          } else
          #endif
          {
            __pyx_t_5 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5503, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            if (__pyx_t_7) {
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
            }
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_t_2);
            __Pyx_INCREF(__pyx_int_100);
            __Pyx_GIVEREF(__pyx_int_100);
            PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_int_100);
            __Pyx_GIVEREF(__pyx_t_4);
            PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_6, __pyx_t_4);
            __pyx_t_2 = 0;
            __pyx_t_4 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5503, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF_SET(__pyx_v__alt, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5504
 *                 self._errHandler.sync(self)
 *                 _alt = self._interp.adaptivePredict(self._input,100,self._ctx)
 *                 while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:             # <<<<<<<<<<<<<<
 *                     if _alt==1:
 *                         self.state = 770
 */
          while (1) {
            __pyx_t_1 = PyObject_RichCompare(__pyx_v__alt, __pyx_int_2, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5504, __pyx_L6_error)
            __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5504, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (__pyx_t_12) {
            } else {
              __pyx_t_11 = __pyx_t_12;
              goto __pyx_L23_bool_binop_done;
            }
            __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ATN); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5504, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_INVALID_ALT_NUMBER); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5504, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = PyObject_RichCompare(__pyx_v__alt, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5504, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5504, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_11 = __pyx_t_12;
            __pyx_L23_bool_binop_done:;
            if (!__pyx_t_11) break;

            /* "pymola/generated/ModelicaParser.pyx":5505
 *                 _alt = self._interp.adaptivePredict(self._input,100,self._ctx)
 *                 while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                     if _alt==1:             # <<<<<<<<<<<<<<
 *                         self.state = 770
 *                         self.equation()
 */
            __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v__alt, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5505, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 5505, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (__pyx_t_11) {

              /* "pymola/generated/ModelicaParser.pyx":5506
 *                 while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                     if _alt==1:
 *                         self.state = 770             # <<<<<<<<<<<<<<
 *                         self.equation()
 *                         self.state = 771
 */
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_770) < 0) __PYX_ERR(0, 5506, __pyx_L6_error)

              /* "pymola/generated/ModelicaParser.pyx":5507
 *                     if _alt==1:
 *                         self.state = 770
 *                         self.equation()             # <<<<<<<<<<<<<<
 *                         self.state = 771
 *                         self.match(ModelicaParser.T__0)
 */
              __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_equation); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5507, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __pyx_t_5 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
                __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
                if (likely(__pyx_t_5)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                  __Pyx_INCREF(__pyx_t_5);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_3, function);
                }
              }
              if (__pyx_t_5) {
                __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5507, __pyx_L6_error)
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              } else {
                __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5507, __pyx_L6_error)
              }
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "pymola/generated/ModelicaParser.pyx":5508
 *                         self.state = 770
 *                         self.equation()
 *                         self.state = 771             # <<<<<<<<<<<<<<
 *                         self.match(ModelicaParser.T__0)
 *                     self.state = 777
 */
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_771) < 0) __PYX_ERR(0, 5508, __pyx_L6_error)

              /* "pymola/generated/ModelicaParser.pyx":5509
 *                         self.equation()
 *                         self.state = 771
 *                         self.match(ModelicaParser.T__0)             # <<<<<<<<<<<<<<
 *                     self.state = 777
 *                     self._errHandler.sync(self)
 */
              __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5509, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5509, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_T__0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5509, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __pyx_t_5 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
                __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
                if (likely(__pyx_t_5)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                  __Pyx_INCREF(__pyx_t_5);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_3, function);
                }
              }
              if (!__pyx_t_5) {
                __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5509, __pyx_L6_error)
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else {
                #if CYTHON_FAST_PYCALL
                if (PyFunction_Check(__pyx_t_3)) {
                  PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_4};
                  __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5509, __pyx_L6_error)
                  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                  __Pyx_GOTREF(__pyx_t_1);
                  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                } else
                #endif
                #if CYTHON_FAST_PYCCALL
                if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
                  PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_4};
                  __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5509, __pyx_L6_error)
                  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                  __Pyx_GOTREF(__pyx_t_1);
                  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                } else
                #endif
                {
                  __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5509, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_2);
                  __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5); __pyx_t_5 = NULL;
                  __Pyx_GIVEREF(__pyx_t_4);
                  PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_4);
                  __pyx_t_4 = 0;
                  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5509, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_1);
                  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                }
              }
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "pymola/generated/ModelicaParser.pyx":5505
 *                 _alt = self._interp.adaptivePredict(self._input,100,self._ctx)
 *                 while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                     if _alt==1:             # <<<<<<<<<<<<<<
 *                         self.state = 770
 *                         self.equation()
 */
            }

            /* "pymola/generated/ModelicaParser.pyx":5510
 *                         self.state = 771
 *                         self.match(ModelicaParser.T__0)
 *                     self.state = 777             # <<<<<<<<<<<<<<
 *                     self._errHandler.sync(self)
 *                     _alt = self._interp.adaptivePredict(self._input,100,self._ctx)
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_777) < 0) __PYX_ERR(0, 5510, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":5511
 *                         self.match(ModelicaParser.T__0)
 *                     self.state = 777
 *                     self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                     _alt = self._interp.adaptivePredict(self._input,100,self._ctx)
 * 
 */
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5511, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5511, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_3)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_3);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
              }
            }
            if (!__pyx_t_3) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5511, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_2)) {
                PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5511, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
                PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5511, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else
              #endif
              {
                __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5511, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __pyx_t_3 = NULL;
                __Pyx_INCREF(__pyx_v_self);
                __Pyx_GIVEREF(__pyx_v_self);
                PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_self);
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5511, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":5512
 *                     self.state = 777
 *                     self._errHandler.sync(self)
 *                     _alt = self._interp.adaptivePredict(self._input,100,self._ctx)             # <<<<<<<<<<<<<<
 * 
 *                 self.state = 782
 */
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_interp); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5512, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_adaptivePredict); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5512, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5512, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5512, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_5 = NULL;
            __pyx_t_6 = 0;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
              if (likely(__pyx_t_5)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                __Pyx_INCREF(__pyx_t_5);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_4, function);
                __pyx_t_6 = 1;
              }
            }
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_t_2, __pyx_int_100, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5512, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_t_2, __pyx_int_100, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5512, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5512, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              if (__pyx_t_5) {
                __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
              }
              __Pyx_GIVEREF(__pyx_t_2);
              PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_t_2);
              __Pyx_INCREF(__pyx_int_100);
              __Pyx_GIVEREF(__pyx_int_100);
              PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_100);
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_3);
              __pyx_t_2 = 0;
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5512, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            }
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF_SET(__pyx_v__alt, __pyx_t_1);
            __pyx_t_1 = 0;
          }

          /* "pymola/generated/ModelicaParser.pyx":5514
 *                     _alt = self._interp.adaptivePredict(self._input,100,self._ctx)
 * 
 *                 self.state = 782             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_782) < 0) __PYX_ERR(0, 5514, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5515
 * 
 *                 self.state = 782
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 * 
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5515, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5515, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_4) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5515, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5515, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5515, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5515, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __pyx_t_4 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5515, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5516
 *                 self.state = 782
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *             self.state = 783
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5516, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5516, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__919, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5516, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_1);
          __pyx_t_1 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":5518
 *                 _la = self._input.LA(1)
 * 
 *             self.state = 783             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__13)
 *             self.state = 784
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_783) < 0) __PYX_ERR(0, 5518, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5519
 * 
 *             self.state = 783
 *             self.match(ModelicaParser.T__13)             # <<<<<<<<<<<<<<
 *             self.state = 784
 *             self.match(ModelicaParser.T__52)
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5519, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5519, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__13); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5519, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5519, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_4};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5519, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_4};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5519, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5519, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_GIVEREF(__pyx_t_4);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_4);
            __pyx_t_4 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5519, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5520
 *             self.state = 783
 *             self.match(ModelicaParser.T__13)
 *             self.state = 784             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__52)
 *         except RecognitionException as re:
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_784) < 0) __PYX_ERR(0, 5520, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5521
 *             self.match(ModelicaParser.T__13)
 *             self.state = 784
 *             self.match(ModelicaParser.T__52)             # <<<<<<<<<<<<<<
 *         except RecognitionException as re:
 *             localctx.exception = re
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5521, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5521, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__52); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5521, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_2) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5521, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_4};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5521, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_4};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5521, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5521, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
            __Pyx_GIVEREF(__pyx_t_4);
            PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_t_4);
            __pyx_t_4 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5521, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5470
 *         self.enterRule(localctx, 102, self.RULE_when_equation)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 756
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":5522
 *             self.state = 784
 *             self.match(ModelicaParser.T__52)
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5522, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.when_equation", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_7, &__pyx_t_3) < 0) __PYX_ERR(0, 5522, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_7);
        __pyx_v_re = __pyx_t_7;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":5523
 *             self.match(ModelicaParser.T__52)
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 5523, __pyx_L31_error)

          /* "pymola/generated/ModelicaParser.pyx":5524
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5524, __pyx_L31_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5524, __pyx_L31_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_re};
            __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5524, __pyx_L31_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_re};
            __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5524, __pyx_L31_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 5524, __pyx_L31_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_2) {
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_2); __pyx_t_2 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_13, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5524, __pyx_L31_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5525
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5525, __pyx_L31_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 5525, __pyx_L31_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5525, __pyx_L31_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5525, __pyx_L31_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5525, __pyx_L31_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_2, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5525, __pyx_L31_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":5522
 *             self.state = 784
 *             self.match(ModelicaParser.T__52)
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L32;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L31_error:;
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_6 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L8_except_error;
          }
          __pyx_L32:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":5470
 *         self.enterRule(localctx, 102, self.RULE_when_equation)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 756
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":5527
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5527, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5527, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5527, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_14 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5527, __pyx_L38_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5527, __pyx_L38_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5527, __pyx_L38_error)
        }
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
      __pyx_L38_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":5528
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class When_statementContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":5465
 * 
 * 
 *     def when_equation(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.When_equationContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.when_equation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__alt);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5532
 *     class When_statementContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21When_statementContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21When_statementContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21When_statementContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21When_statementContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 5532, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 5532, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5532, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.When_statementContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21When_statementContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21When_statementContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":5533
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 5533, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5533, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5533, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5533, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5533, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5534
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def expression(self, i:int=None):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 5534, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5532
 *     class When_statementContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.When_statementContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5536
 *             self.parser = parser
 * 
 *         def expression(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21When_statementContext_3expression(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21When_statementContext_3expression = {"expression", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21When_statementContext_3expression, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21When_statementContext_3expression(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("expression (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "expression") < 0)) __PYX_ERR(0, 5536, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("expression", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5536, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.When_statementContext.expression", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21When_statementContext_2expression(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21When_statementContext_2expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("expression", 0);

  /* "pymola/generated/ModelicaParser.pyx":5537
 * 
 *         def expression(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5538
 *         def expression(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.ExpressionContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5538, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5538, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_ExpressionContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5538, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5538, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5538, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5538, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5538, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5538, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":5537
 * 
 *         def expression(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5540
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.ExpressionContext,i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_ExpressionContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5540, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5540, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5540, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5540, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":5536
 *             self.parser = parser
 * 
 *         def expression(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.When_statementContext.expression", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5543
 * 
 * 
 *         def statement(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.StatementContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21When_statementContext_5statement(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21When_statementContext_5statement = {"statement", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21When_statementContext_5statement, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21When_statementContext_5statement(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("statement (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "statement") < 0)) __PYX_ERR(0, 5543, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("statement", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5543, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.When_statementContext.statement", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21When_statementContext_4statement(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21When_statementContext_4statement(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("statement", 0);

  /* "pymola/generated/ModelicaParser.pyx":5544
 * 
 *         def statement(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.StatementContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5545
 *         def statement(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.StatementContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.StatementContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5545, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5545, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_StatementContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5545, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5545, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5545, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5545, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5545, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5545, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":5544
 * 
 *         def statement(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.StatementContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5547
 *                 return self.getTypedRuleContexts(ModelicaParser.StatementContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.StatementContext,i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5547, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5547, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_StatementContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5547, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5547, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5547, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5547, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5547, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":5543
 * 
 * 
 *         def statement(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.StatementContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.When_statementContext.statement", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5550
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_when_statement
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21When_statementContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21When_statementContext_7getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21When_statementContext_7getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21When_statementContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21When_statementContext_6getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21When_statementContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":5551
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_when_statement             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5551, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_when_statement); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5551, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":5550
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_when_statement
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.When_statementContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5553
 *             return ModelicaParser.RULE_when_statement
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterWhen_statement" ):
 *                 listener.enterWhen_statement(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21When_statementContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21When_statementContext_9enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21When_statementContext_9enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21When_statementContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 5553, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 5553, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5553, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.When_statementContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21When_statementContext_8enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21When_statementContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":5554
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterWhen_statement" ):             # <<<<<<<<<<<<<<
 *                 listener.enterWhen_statement(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterWhen_statement); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5554, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5555
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterWhen_statement" ):
 *                 listener.enterWhen_statement(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterWhen_statement); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5555, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5555, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5555, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5555, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5555, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5555, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":5554
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterWhen_statement" ):             # <<<<<<<<<<<<<<
 *                 listener.enterWhen_statement(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5553
 *             return ModelicaParser.RULE_when_statement
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterWhen_statement" ):
 *                 listener.enterWhen_statement(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.When_statementContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5557
 *                 listener.enterWhen_statement(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitWhen_statement" ):
 *                 listener.exitWhen_statement(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21When_statementContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21When_statementContext_11exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21When_statementContext_11exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21When_statementContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 5557, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 5557, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5557, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.When_statementContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21When_statementContext_10exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21When_statementContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":5558
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitWhen_statement" ):             # <<<<<<<<<<<<<<
 *                 listener.exitWhen_statement(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitWhen_statement); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5558, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5559
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitWhen_statement" ):
 *                 listener.exitWhen_statement(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitWhen_statement); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5559, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5559, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5559, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5559, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5559, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5559, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":5558
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitWhen_statement" ):             # <<<<<<<<<<<<<<
 *                 listener.exitWhen_statement(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5557
 *                 listener.enterWhen_statement(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitWhen_statement" ):
 *                 listener.exitWhen_statement(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.When_statementContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5561
 *                 listener.exitWhen_statement(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitWhen_statement" ):
 *                 return visitor.visitWhen_statement(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21When_statementContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21When_statementContext_13accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21When_statementContext_13accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21When_statementContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 5561, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 5561, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5561, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.When_statementContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21When_statementContext_12accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21When_statementContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":5562
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitWhen_statement" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitWhen_statement(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitWhen_statement); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5562, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5563
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitWhen_statement" ):
 *                 return visitor.visitWhen_statement(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitWhen_statement); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5563, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5563, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5563, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5563, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5563, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5563, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":5562
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitWhen_statement" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitWhen_statement(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5565
 *                 return visitor.visitWhen_statement(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5565, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5565, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5565, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5565, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5565, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5565, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":5561
 *                 listener.exitWhen_statement(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitWhen_statement" ):
 *                 return visitor.visitWhen_statement(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.When_statementContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5570
 * 
 * 
 *     def when_statement(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.When_statementContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_107when_statement(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_107when_statement = {"when_statement", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_107when_statement, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_107when_statement(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("when_statement (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_106when_statement(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_106when_statement(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  char const *__pyx_t_22;
  __Pyx_RefNannySetupContext("when_statement", 0);

  /* "pymola/generated/ModelicaParser.pyx":5572
 *     def when_statement(self):
 * 
 *         localctx = ModelicaParser.When_statementContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 104, self.RULE_when_statement)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_When_statementContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5572, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5572, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5572, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5572, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5573
 * 
 *         localctx = ModelicaParser.When_statementContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 104, self.RULE_when_statement)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_when_statement); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_104, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5573, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_104, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5573, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5573, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_104);
    __Pyx_GIVEREF(__pyx_int_104);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_104);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5573, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5574
 *         localctx = ModelicaParser.When_statementContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 104, self.RULE_when_statement)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 5574, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5575
 *         self.enterRule(localctx, 104, self.RULE_when_statement)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 786
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":5576
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 786
 *             self.match(ModelicaParser.T__52)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5576, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5576, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5576, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5576, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5576, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5577
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 786             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__52)
 *             self.state = 787
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_786) < 0) __PYX_ERR(0, 5577, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5578
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 786
 *             self.match(ModelicaParser.T__52)             # <<<<<<<<<<<<<<
 *             self.state = 787
 *             self.expression()
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5578, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5578, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__52); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5578, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5578, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5578, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5578, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5578, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5578, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5579
 *             self.state = 786
 *             self.match(ModelicaParser.T__52)
 *             self.state = 787             # <<<<<<<<<<<<<<
 *             self.expression()
 *             self.state = 788
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_787) < 0) __PYX_ERR(0, 5579, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5580
 *             self.match(ModelicaParser.T__52)
 *             self.state = 787
 *             self.expression()             # <<<<<<<<<<<<<<
 *             self.state = 788
 *             self.match(ModelicaParser.T__45)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expression); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5580, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5580, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5580, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5581
 *             self.state = 787
 *             self.expression()
 *             self.state = 788             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__45)
 *             self.state = 794
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_788) < 0) __PYX_ERR(0, 5581, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5582
 *             self.expression()
 *             self.state = 788
 *             self.match(ModelicaParser.T__45)             # <<<<<<<<<<<<<<
 *             self.state = 794
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5582, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5582, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__45); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5582, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5582, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5582, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5582, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5582, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4); __pyx_t_4 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5582, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5583
 *             self.state = 788
 *             self.match(ModelicaParser.T__45)
 *             self.state = 794             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_794) < 0) __PYX_ERR(0, 5583, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5584
 *             self.match(ModelicaParser.T__45)
 *             self.state = 794
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5584, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5584, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5584, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5584, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5584, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5584, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5584, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5585
 *             self.state = 794
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:
 *                 self.state = 789
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5585, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5585, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__920, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5585, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v__la = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5586
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:             # <<<<<<<<<<<<<<
 *                 self.state = 789
 *                 self.statement()
 */
        while (1) {
          __pyx_t_1 = __Pyx_PyInt_From_long((~0x3f)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5586, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_7 = PyNumber_And(__pyx_v__la, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5586, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_t_7, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5586, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5586, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (!__pyx_t_12) {
            goto __pyx_L17_next_or;
          } else {
          }
          __pyx_t_1 = PyNumber_Lshift(__pyx_int_1, __pyx_v__la); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5586, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5586, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__16); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5586, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5586, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5586, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__39); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5586, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5586, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyNumber_Or(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5586, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5586, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__43); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5586, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5586, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Or(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5586, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5586, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__44); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5586, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5586, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyNumber_Or(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5586, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5586, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__48); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5586, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5586, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Or(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5586, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5586, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__51); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5586, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5586, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyNumber_Or(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5586, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5586, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__52); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5586, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5586, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Or(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5586, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_And(__pyx_t_1, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5586, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyObject_RichCompare(__pyx_t_2, __pyx_int_0, Py_NE); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5586, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5586, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (!__pyx_t_12) {
          } else {
            __pyx_t_11 = __pyx_t_12;
            goto __pyx_L16_bool_binop_done;
          }
          __pyx_L17_next_or:;
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5586, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5586, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyObject_RichCompare(__pyx_v__la, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5586, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5586, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_11 = __pyx_t_12;
          __pyx_L16_bool_binop_done:;
          if (!__pyx_t_11) break;

          /* "pymola/generated/ModelicaParser.pyx":5587
 *             _la = self._input.LA(1)
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:
 *                 self.state = 789             # <<<<<<<<<<<<<<
 *                 self.statement()
 *                 self.state = 790
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_789) < 0) __PYX_ERR(0, 5587, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5588
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:
 *                 self.state = 789
 *                 self.statement()             # <<<<<<<<<<<<<<
 *                 self.state = 790
 *                 self.match(ModelicaParser.T__0)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_statement); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5588, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_1 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (__pyx_t_1) {
            __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5588, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          } else {
            __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5588, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5589
 *                 self.state = 789
 *                 self.statement()
 *                 self.state = 790             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__0)
 *                 self.state = 796
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_790) < 0) __PYX_ERR(0, 5589, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5590
 *                 self.statement()
 *                 self.state = 790
 *                 self.match(ModelicaParser.T__0)             # <<<<<<<<<<<<<<
 *                 self.state = 796
 *                 self._errHandler.sync(self)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5590, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5590, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5590, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_1) {
            __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5590, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_t_3};
              __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5590, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_t_3};
              __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5590, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5590, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1); __pyx_t_1 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5590, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5591
 *                 self.state = 790
 *                 self.match(ModelicaParser.T__0)
 *                 self.state = 796             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_796) < 0) __PYX_ERR(0, 5591, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5592
 *                 self.match(ModelicaParser.T__0)
 *                 self.state = 796
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 * 
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5592, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_sync); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5592, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5592, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5592, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_7);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5592, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_7);
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5592, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
              __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5592, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5593
 *                 self.state = 796
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *             self.state = 810
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5593, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_LA); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5593, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__921, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5593, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_7);
          __pyx_t_7 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":5595
 *                 _la = self._input.LA(1)
 * 
 *             self.state = 810             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_810) < 0) __PYX_ERR(0, 5595, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5596
 * 
 *             self.state = 810
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             while _la==ModelicaParser.T__53:
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5596, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_sync); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5596, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_4) {
          __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_self); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5596, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
            __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5596, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
            __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5596, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5596, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5596, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5597
 *             self.state = 810
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             while _la==ModelicaParser.T__53:
 *                 self.state = 797
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5597, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_LA); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5597, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__922, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5597, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5598
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             while _la==ModelicaParser.T__53:             # <<<<<<<<<<<<<<
 *                 self.state = 797
 *                 self.match(ModelicaParser.T__53)
 */
        while (1) {
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5598, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__53); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5598, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyObject_RichCompare(__pyx_v__la, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5598, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 5598, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (!__pyx_t_11) break;

          /* "pymola/generated/ModelicaParser.pyx":5599
 *             _la = self._input.LA(1)
 *             while _la==ModelicaParser.T__53:
 *                 self.state = 797             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__53)
 *                 self.state = 798
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_797) < 0) __PYX_ERR(0, 5599, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5600
 *             while _la==ModelicaParser.T__53:
 *                 self.state = 797
 *                 self.match(ModelicaParser.T__53)             # <<<<<<<<<<<<<<
 *                 self.state = 798
 *                 self.expression()
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5600, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5600, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__53); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5600, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5600, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_4};
              __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5600, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_4};
              __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5600, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            } else
            #endif
            {
              __pyx_t_1 = PyTuple_New(1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5600, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_GIVEREF(__pyx_t_4);
              PyTuple_SET_ITEM(__pyx_t_1, 0+1, __pyx_t_4);
              __pyx_t_4 = 0;
              __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_1, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5600, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5601
 *                 self.state = 797
 *                 self.match(ModelicaParser.T__53)
 *                 self.state = 798             # <<<<<<<<<<<<<<
 *                 self.expression()
 *                 self.state = 799
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_798) < 0) __PYX_ERR(0, 5601, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5602
 *                 self.match(ModelicaParser.T__53)
 *                 self.state = 798
 *                 self.expression()             # <<<<<<<<<<<<<<
 *                 self.state = 799
 *                 self.match(ModelicaParser.T__45)
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expression); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5602, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (__pyx_t_1) {
            __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5602, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          } else {
            __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5602, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5603
 *                 self.state = 798
 *                 self.expression()
 *                 self.state = 799             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__45)
 *                 self.state = 805
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_799) < 0) __PYX_ERR(0, 5603, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5604
 *                 self.expression()
 *                 self.state = 799
 *                 self.match(ModelicaParser.T__45)             # <<<<<<<<<<<<<<
 *                 self.state = 805
 *                 self._errHandler.sync(self)
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5604, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5604, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__45); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5604, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (!__pyx_t_1) {
            __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5604, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_t_4};
              __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5604, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_t_4};
              __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5604, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            } else
            #endif
            {
              __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5604, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1); __pyx_t_1 = NULL;
              __Pyx_GIVEREF(__pyx_t_4);
              PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_4);
              __pyx_t_4 = 0;
              __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5604, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5605
 *                 self.state = 799
 *                 self.match(ModelicaParser.T__45)
 *                 self.state = 805             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_805) < 0) __PYX_ERR(0, 5605, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5606
 *                 self.match(ModelicaParser.T__45)
 *                 self.state = 805
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 *                 while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5606, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5606, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_3) {
            __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5606, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
              __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5606, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_7);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
              __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5606, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_7);
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5606, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __pyx_t_3 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_self);
              __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5606, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5607
 *                 self.state = 805
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:
 *                     self.state = 800
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5607, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_LA); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5607, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__923, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5607, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_7);
          __pyx_t_7 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5608
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:             # <<<<<<<<<<<<<<
 *                     self.state = 800
 *                     self.statement()
 */
          while (1) {
            __pyx_t_7 = __Pyx_PyInt_From_long((~0x3f)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5608, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_2 = PyNumber_And(__pyx_v__la, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5608, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __pyx_t_7 = __Pyx_PyInt_EqObjC(__pyx_t_2, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5608, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5608, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            if (!__pyx_t_12) {
              goto __pyx_L24_next_or;
            } else {
            }
            __pyx_t_7 = PyNumber_Lshift(__pyx_int_1, __pyx_v__la); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5608, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5608, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__16); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5608, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5608, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5608, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__39); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5608, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5608, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = PyNumber_Or(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5608, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5608, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__43); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5608, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5608, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_2 = PyNumber_Or(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5608, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5608, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__44); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5608, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5608, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = PyNumber_Or(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5608, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5608, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__48); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5608, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5608, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_2 = PyNumber_Or(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5608, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5608, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__51); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5608, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5608, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = PyNumber_Or(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5608, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5608, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__52); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5608, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5608, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_2 = PyNumber_Or(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5608, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = PyNumber_And(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5608, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_2 = PyObject_RichCompare(__pyx_t_4, __pyx_int_0, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5608, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5608, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (!__pyx_t_12) {
            } else {
              __pyx_t_11 = __pyx_t_12;
              goto __pyx_L23_bool_binop_done;
            }
            __pyx_L24_next_or:;
            __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5608, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5608, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_2 = PyObject_RichCompare(__pyx_v__la, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5608, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5608, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_11 = __pyx_t_12;
            __pyx_L23_bool_binop_done:;
            if (!__pyx_t_11) break;

            /* "pymola/generated/ModelicaParser.pyx":5609
 *                 _la = self._input.LA(1)
 *                 while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:
 *                     self.state = 800             # <<<<<<<<<<<<<<
 *                     self.statement()
 *                     self.state = 801
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_800) < 0) __PYX_ERR(0, 5609, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":5610
 *                 while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:
 *                     self.state = 800
 *                     self.statement()             # <<<<<<<<<<<<<<
 *                     self.state = 801
 *                     self.match(ModelicaParser.T__0)
 */
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_statement); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5610, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_7 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
              __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
              if (likely(__pyx_t_7)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                __Pyx_INCREF(__pyx_t_7);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_4, function);
              }
            }
            if (__pyx_t_7) {
              __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5610, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else {
              __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5610, __pyx_L6_error)
            }
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

            /* "pymola/generated/ModelicaParser.pyx":5611
 *                     self.state = 800
 *                     self.statement()
 *                     self.state = 801             # <<<<<<<<<<<<<<
 *                     self.match(ModelicaParser.T__0)
 *                     self.state = 807
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_801) < 0) __PYX_ERR(0, 5611, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":5612
 *                     self.statement()
 *                     self.state = 801
 *                     self.match(ModelicaParser.T__0)             # <<<<<<<<<<<<<<
 *                     self.state = 807
 *                     self._errHandler.sync(self)
 */
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5612, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5612, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5612, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __pyx_t_7 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
              __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
              if (likely(__pyx_t_7)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                __Pyx_INCREF(__pyx_t_7);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_4, function);
              }
            }
            if (!__pyx_t_7) {
              __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5612, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_2);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_4)) {
                PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_3};
                __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5612, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
                PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_3};
                __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5612, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              } else
              #endif
              {
                __pyx_t_1 = PyTuple_New(1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5612, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_7); __pyx_t_7 = NULL;
                __Pyx_GIVEREF(__pyx_t_3);
                PyTuple_SET_ITEM(__pyx_t_1, 0+1, __pyx_t_3);
                __pyx_t_3 = 0;
                __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5612, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

            /* "pymola/generated/ModelicaParser.pyx":5613
 *                     self.state = 801
 *                     self.match(ModelicaParser.T__0)
 *                     self.state = 807             # <<<<<<<<<<<<<<
 *                     self._errHandler.sync(self)
 *                     _la = self._input.LA(1)
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_807) < 0) __PYX_ERR(0, 5613, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":5614
 *                     self.match(ModelicaParser.T__0)
 *                     self.state = 807
 *                     self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                     _la = self._input.LA(1)
 * 
 */
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5614, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_sync); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5614, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
              __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
              if (likely(__pyx_t_4)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
                __Pyx_INCREF(__pyx_t_4);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_1, function);
              }
            }
            if (!__pyx_t_4) {
              __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5614, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_1)) {
                PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
                __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5614, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_GOTREF(__pyx_t_2);
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
                PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
                __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5614, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_GOTREF(__pyx_t_2);
              } else
              #endif
              {
                __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5614, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __pyx_t_4 = NULL;
                __Pyx_INCREF(__pyx_v_self);
                __Pyx_GIVEREF(__pyx_v_self);
                PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
                __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5614, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

            /* "pymola/generated/ModelicaParser.pyx":5615
 *                     self.state = 807
 *                     self._errHandler.sync(self)
 *                     _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *                 self.state = 812
 */
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5615, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_LA); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5615, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__924, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5615, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_2);
            __pyx_t_2 = 0;
          }

          /* "pymola/generated/ModelicaParser.pyx":5617
 *                     _la = self._input.LA(1)
 * 
 *                 self.state = 812             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_812) < 0) __PYX_ERR(0, 5617, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5618
 * 
 *                 self.state = 812
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 * 
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5618, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_sync); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5618, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (!__pyx_t_1) {
            __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5618, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_v_self};
              __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5618, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_GOTREF(__pyx_t_2);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_v_self};
              __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5618, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_GOTREF(__pyx_t_2);
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5618, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1); __pyx_t_1 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_self);
              __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5618, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5619
 *                 self.state = 812
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *             self.state = 813
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5619, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_LA); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5619, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__925, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5619, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_2);
          __pyx_t_2 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":5621
 *                 _la = self._input.LA(1)
 * 
 *             self.state = 813             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__13)
 *             self.state = 814
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_813) < 0) __PYX_ERR(0, 5621, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5622
 * 
 *             self.state = 813
 *             self.match(ModelicaParser.T__13)             # <<<<<<<<<<<<<<
 *             self.state = 814
 *             self.match(ModelicaParser.T__52)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5622, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5622, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__13); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5622, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_4) {
          __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5622, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_1};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5622, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_1};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5622, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5622, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4); __pyx_t_4 = NULL;
            __Pyx_GIVEREF(__pyx_t_1);
            PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_1);
            __pyx_t_1 = 0;
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5622, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5623
 *             self.state = 813
 *             self.match(ModelicaParser.T__13)
 *             self.state = 814             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__52)
 *         except RecognitionException as re:
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_814) < 0) __PYX_ERR(0, 5623, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5624
 *             self.match(ModelicaParser.T__13)
 *             self.state = 814
 *             self.match(ModelicaParser.T__52)             # <<<<<<<<<<<<<<
 *         except RecognitionException as re:
 *             localctx.exception = re
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5624, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5624, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__52); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5624, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5624, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_1};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5624, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_1};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5624, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5624, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_GIVEREF(__pyx_t_1);
            PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_1);
            __pyx_t_1 = 0;
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5624, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5575
 *         self.enterRule(localctx, 104, self.RULE_when_statement)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 786
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "pymola/generated/ModelicaParser.pyx":5625
 *             self.state = 814
 *             self.match(ModelicaParser.T__52)
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5625, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.when_statement", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4) < 0) __PYX_ERR(0, 5625, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __pyx_v_re = __pyx_t_3;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":5626
 *             self.match(ModelicaParser.T__52)
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 5626, __pyx_L31_error)

          /* "pymola/generated/ModelicaParser.pyx":5627
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5627, __pyx_L31_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5627, __pyx_L31_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_re};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5627, __pyx_L31_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_re};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5627, __pyx_L31_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 5627, __pyx_L31_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_7) {
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_7); __pyx_t_7 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_13, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5627, __pyx_L31_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5628
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5628, __pyx_L31_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 5628, __pyx_L31_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5628, __pyx_L31_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5628, __pyx_L31_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5628, __pyx_L31_error)
            __Pyx_GOTREF(__pyx_t_7);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5628, __pyx_L31_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":5625
 *             self.state = 814
 *             self.match(ModelicaParser.T__52)
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L32;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L31_error:;
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_6 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L8_except_error;
          }
          __pyx_L32:;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":5575
 *         self.enterRule(localctx, 104, self.RULE_when_statement)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 786
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":5630
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5630, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_2) {
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5630, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      } else {
        __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5630, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_14 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5630, __pyx_L38_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_2) {
          __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5630, __pyx_L38_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        } else {
          __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5630, __pyx_L38_error)
        }
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
      __pyx_L38_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":5631
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Connect_clauseContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":5570
 * 
 * 
 *     def when_statement(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.When_statementContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.when_statement", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5635
 *     class Connect_clauseContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 5635, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 5635, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5635, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Connect_clauseContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Connect_clauseContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Connect_clauseContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":5636
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 5636, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5636, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5636, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5636, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5636, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5636, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5636, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5636, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5637
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def component_reference(self, i:int=None):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 5637, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5635
 *     class Connect_clauseContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Connect_clauseContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5639
 *             self.parser = parser
 * 
 *         def component_reference(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Component_referenceContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_3component_reference(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_3component_reference = {"component_reference", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_3component_reference, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_3component_reference(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("component_reference (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "component_reference") < 0)) __PYX_ERR(0, 5639, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("component_reference", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5639, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Connect_clauseContext.component_reference", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_2component_reference(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_2component_reference(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("component_reference", 0);

  /* "pymola/generated/ModelicaParser.pyx":5640
 * 
 *         def component_reference(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.Component_referenceContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5641
 *         def component_reference(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Component_referenceContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5641, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5641, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_Component_referenceContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5641, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5641, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5641, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5641, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5641, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5641, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":5640
 * 
 *         def component_reference(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.Component_referenceContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5643
 *                 return self.getTypedRuleContexts(ModelicaParser.Component_referenceContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5643, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5643, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Component_referenceContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5643, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5643, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5643, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5643, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5643, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":5639
 *             self.parser = parser
 * 
 *         def component_reference(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Component_referenceContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Connect_clauseContext.component_reference", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5646
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_connect_clause
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_5getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_5getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_4getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":5647
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_connect_clause             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5647, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_connect_clause); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5647, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":5646
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_connect_clause
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Connect_clauseContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5649
 *             return ModelicaParser.RULE_connect_clause
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterConnect_clause" ):
 *                 listener.enterConnect_clause(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_7enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_7enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 5649, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 5649, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5649, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Connect_clauseContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_6enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":5650
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterConnect_clause" ):             # <<<<<<<<<<<<<<
 *                 listener.enterConnect_clause(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterConnect_clause); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5650, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5651
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterConnect_clause" ):
 *                 listener.enterConnect_clause(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterConnect_clause); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5651, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5651, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5651, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5651, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5651, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5651, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":5650
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterConnect_clause" ):             # <<<<<<<<<<<<<<
 *                 listener.enterConnect_clause(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5649
 *             return ModelicaParser.RULE_connect_clause
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterConnect_clause" ):
 *                 listener.enterConnect_clause(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Connect_clauseContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5653
 *                 listener.enterConnect_clause(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitConnect_clause" ):
 *                 listener.exitConnect_clause(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_9exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_9exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 5653, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 5653, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5653, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Connect_clauseContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_8exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":5654
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitConnect_clause" ):             # <<<<<<<<<<<<<<
 *                 listener.exitConnect_clause(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitConnect_clause); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5654, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5655
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitConnect_clause" ):
 *                 listener.exitConnect_clause(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitConnect_clause); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5655, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5655, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5655, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5655, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5655, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5655, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":5654
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitConnect_clause" ):             # <<<<<<<<<<<<<<
 *                 listener.exitConnect_clause(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5653
 *                 listener.enterConnect_clause(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitConnect_clause" ):
 *                 listener.exitConnect_clause(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Connect_clauseContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5657
 *                 listener.exitConnect_clause(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitConnect_clause" ):
 *                 return visitor.visitConnect_clause(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_11accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_11accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 5657, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 5657, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5657, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Connect_clauseContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_10accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":5658
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitConnect_clause" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitConnect_clause(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitConnect_clause); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5658, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5659
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitConnect_clause" ):
 *                 return visitor.visitConnect_clause(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitConnect_clause); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5659, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5659, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5659, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5659, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5659, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5659, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":5658
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitConnect_clause" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitConnect_clause(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5661
 *                 return visitor.visitConnect_clause(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5661, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5661, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5661, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5661, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5661, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5661, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":5657
 *                 listener.exitConnect_clause(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitConnect_clause" ):
 *                 return visitor.visitConnect_clause(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Connect_clauseContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5666
 * 
 * 
 *     def connect_clause(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Connect_clauseContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_109connect_clause(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_109connect_clause = {"connect_clause", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_109connect_clause, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_109connect_clause(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("connect_clause (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_108connect_clause(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_108connect_clause(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  char const *__pyx_t_20;
  __Pyx_RefNannySetupContext("connect_clause", 0);

  /* "pymola/generated/ModelicaParser.pyx":5668
 *     def connect_clause(self):
 * 
 *         localctx = ModelicaParser.Connect_clauseContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 106, self.RULE_connect_clause)
 *         try:
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5668, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Connect_clauseContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5668, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5668, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5668, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5668, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5668, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5668, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5668, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5669
 * 
 *         localctx = ModelicaParser.Connect_clauseContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 106, self.RULE_connect_clause)             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5669, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_connect_clause); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5669, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_106, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5669, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_106, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5669, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5669, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_106);
    __Pyx_GIVEREF(__pyx_int_106);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_106);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5669, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5670
 *         localctx = ModelicaParser.Connect_clauseContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 106, self.RULE_connect_clause)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 816
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":5671
 *         self.enterRule(localctx, 106, self.RULE_connect_clause)
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 816
 *             self.match(ModelicaParser.T__54)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5671, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5671, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5671, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5671, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5671, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5672
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 816             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__54)
 *             self.state = 817
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_816) < 0) __PYX_ERR(0, 5672, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5673
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 816
 *             self.match(ModelicaParser.T__54)             # <<<<<<<<<<<<<<
 *             self.state = 817
 *             self.match(ModelicaParser.T__16)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5673, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5673, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__54); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5673, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5673, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5673, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5673, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5673, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5673, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5674
 *             self.state = 816
 *             self.match(ModelicaParser.T__54)
 *             self.state = 817             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__16)
 *             self.state = 818
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_817) < 0) __PYX_ERR(0, 5674, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5675
 *             self.match(ModelicaParser.T__54)
 *             self.state = 817
 *             self.match(ModelicaParser.T__16)             # <<<<<<<<<<<<<<
 *             self.state = 818
 *             self.component_reference()
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5675, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5675, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__16); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5675, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5675, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5675, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5675, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5675, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4); __pyx_t_4 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5675, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5676
 *             self.state = 817
 *             self.match(ModelicaParser.T__16)
 *             self.state = 818             # <<<<<<<<<<<<<<
 *             self.component_reference()
 *             self.state = 819
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_818) < 0) __PYX_ERR(0, 5676, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5677
 *             self.match(ModelicaParser.T__16)
 *             self.state = 818
 *             self.component_reference()             # <<<<<<<<<<<<<<
 *             self.state = 819
 *             self.match(ModelicaParser.T__20)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_component_reference); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5677, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5677, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5677, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5678
 *             self.state = 818
 *             self.component_reference()
 *             self.state = 819             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__20)
 *             self.state = 820
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_819) < 0) __PYX_ERR(0, 5678, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5679
 *             self.component_reference()
 *             self.state = 819
 *             self.match(ModelicaParser.T__20)             # <<<<<<<<<<<<<<
 *             self.state = 820
 *             self.component_reference()
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5679, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5679, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__20); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5679, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5679, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5679, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5679, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5679, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5679, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5680
 *             self.state = 819
 *             self.match(ModelicaParser.T__20)
 *             self.state = 820             # <<<<<<<<<<<<<<
 *             self.component_reference()
 *             self.state = 821
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_820) < 0) __PYX_ERR(0, 5680, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5681
 *             self.match(ModelicaParser.T__20)
 *             self.state = 820
 *             self.component_reference()             # <<<<<<<<<<<<<<
 *             self.state = 821
 *             self.match(ModelicaParser.T__18)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_component_reference); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5681, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5681, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5681, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5682
 *             self.state = 820
 *             self.component_reference()
 *             self.state = 821             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__18)
 *         except RecognitionException as re:
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_821) < 0) __PYX_ERR(0, 5682, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5683
 *             self.component_reference()
 *             self.state = 821
 *             self.match(ModelicaParser.T__18)             # <<<<<<<<<<<<<<
 *         except RecognitionException as re:
 *             localctx.exception = re
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5683, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5683, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__18); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5683, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5683, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5683, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5683, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5683, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4); __pyx_t_4 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5683, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5670
 *         localctx = ModelicaParser.Connect_clauseContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 106, self.RULE_connect_clause)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 816
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":5684
 *             self.state = 821
 *             self.match(ModelicaParser.T__18)
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5684, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.connect_clause", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_3, &__pyx_t_7) < 0) __PYX_ERR(0, 5684, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_3);
        __pyx_v_re = __pyx_t_3;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":5685
 *             self.match(ModelicaParser.T__18)
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 5685, __pyx_L19_error)

          /* "pymola/generated/ModelicaParser.pyx":5686
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5686, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5686, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5686, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5686, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_11 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 5686, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_11);
            if (__pyx_t_4) {
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_4); __pyx_t_4 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_11, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5686, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5687
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5687, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 5687, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_11)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5687, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_11)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5687, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5687, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5687, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":5684
 *             self.state = 821
 *             self.match(ModelicaParser.T__18)
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L20;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L19_error:;
            __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_14);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __pyx_t_6 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_17);
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
            }
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
            __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
            goto __pyx_L8_except_error;
          }
          __pyx_L20:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":5670
 *         localctx = ModelicaParser.Connect_clauseContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 106, self.RULE_connect_clause)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 816
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":5689
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5689, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5689, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5689, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_18, &__pyx_t_17);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_18);
      __Pyx_XGOTREF(__pyx_t_17);
      __pyx_t_12 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_20 = __pyx_filename;
      {
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5689, __pyx_L26_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5689, __pyx_L26_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5689, __pyx_L26_error)
        }
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_18, __pyx_t_17);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0;
      __pyx_lineno = __pyx_t_12; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_20;
      goto __pyx_L1_error;
      __pyx_L26_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_18, __pyx_t_17);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":5690
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class ExpressionContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":5666
 * 
 * 
 *     def connect_clause(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Connect_clauseContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.connect_clause", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5694
 *     class ExpressionContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_17ExpressionContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_17ExpressionContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_17ExpressionContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_17ExpressionContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 5694, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 5694, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5694, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ExpressionContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_17ExpressionContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_17ExpressionContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":5695
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 5695, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5695, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5695, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5695, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5695, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5695, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5695, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5695, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5696
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 5696, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5694
 *     class ExpressionContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ExpressionContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5699
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_expression
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_17ExpressionContext_3getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_17ExpressionContext_3getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_17ExpressionContext_3getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_17ExpressionContext_3getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_17ExpressionContext_2getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_17ExpressionContext_2getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":5700
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_expression             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_expression); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":5699
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_expression
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ExpressionContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5703
 * 
 * 
 *         def copyFrom(self, ctx:ParserRuleContext):             # <<<<<<<<<<<<<<
 *             super().copyFrom(ctx)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_17ExpressionContext_5copyFrom(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_17ExpressionContext_5copyFrom = {"copyFrom", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_17ExpressionContext_5copyFrom, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_17ExpressionContext_5copyFrom(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copyFrom (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx_2,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("copyFrom", 1, 2, 2, 1); __PYX_ERR(0, 5703, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "copyFrom") < 0)) __PYX_ERR(0, 5703, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("copyFrom", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5703, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ExpressionContext.copyFrom", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_17ExpressionContext_4copyFrom(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_17ExpressionContext_4copyFrom(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("copyFrom", 0);

  /* "pymola/generated/ModelicaParser.pyx":5704
 * 
 *         def copyFrom(self, ctx:ParserRuleContext):
 *             super().copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 5704, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5704, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5704, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5704, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5704, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5704, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5704, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5704, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5704, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5703
 * 
 * 
 *         def copyFrom(self, ctx:ParserRuleContext):             # <<<<<<<<<<<<<<
 *             super().copyFrom(ctx)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ExpressionContext.copyFrom", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5710
 *     class Expression_ifContext(ExpressionContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExpressionContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Expression_ifContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20Expression_ifContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20Expression_ifContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Expression_ifContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 5710, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 5710, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 5710, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5710, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expression_ifContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20Expression_ifContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Expression_ifContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":5711
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExpressionContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 5711, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5711, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5711, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5711, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5711, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5711, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5711, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5711, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5711, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5712
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExpressionContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def expression(self, i:int=None):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5712, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5712, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5712, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5712, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5712, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5712, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5710
 *     class Expression_ifContext(ExpressionContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExpressionContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expression_ifContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5714
 *             self.copyFrom(ctx)
 * 
 *         def expression(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Expression_ifContext_3expression(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20Expression_ifContext_3expression = {"expression", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20Expression_ifContext_3expression, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Expression_ifContext_3expression(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("expression (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "expression") < 0)) __PYX_ERR(0, 5714, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("expression", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5714, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expression_ifContext.expression", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20Expression_ifContext_2expression(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Expression_ifContext_2expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("expression", 0);

  /* "pymola/generated/ModelicaParser.pyx":5715
 * 
 *         def expression(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5716
 *         def expression(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.ExpressionContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5716, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5716, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_ExpressionContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5716, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5716, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5716, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5716, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5716, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5716, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":5715
 * 
 *         def expression(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5718
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.ExpressionContext,i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5718, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5718, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_ExpressionContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5718, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5718, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5718, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5718, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5718, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":5714
 *             self.copyFrom(ctx)
 * 
 *         def expression(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expression_ifContext.expression", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5721
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpression_if" ):
 *                 listener.enterExpression_if(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Expression_ifContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20Expression_ifContext_5enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20Expression_ifContext_5enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Expression_ifContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 5721, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 5721, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5721, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expression_ifContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20Expression_ifContext_4enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Expression_ifContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":5722
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExpression_if" ):             # <<<<<<<<<<<<<<
 *                 listener.enterExpression_if(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterExpression_if); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5722, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5723
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExpression_if" ):
 *                 listener.enterExpression_if(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterExpression_if); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5723, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5723, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5723, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5723, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5723, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5723, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":5722
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExpression_if" ):             # <<<<<<<<<<<<<<
 *                 listener.enterExpression_if(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5721
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpression_if" ):
 *                 listener.enterExpression_if(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expression_ifContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5725
 *                 listener.enterExpression_if(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpression_if" ):
 *                 listener.exitExpression_if(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Expression_ifContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20Expression_ifContext_7exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20Expression_ifContext_7exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Expression_ifContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 5725, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 5725, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5725, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expression_ifContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20Expression_ifContext_6exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Expression_ifContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":5726
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExpression_if" ):             # <<<<<<<<<<<<<<
 *                 listener.exitExpression_if(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitExpression_if); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5726, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5727
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExpression_if" ):
 *                 listener.exitExpression_if(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitExpression_if); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5727, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5727, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5727, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5727, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5727, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5727, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":5726
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExpression_if" ):             # <<<<<<<<<<<<<<
 *                 listener.exitExpression_if(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5725
 *                 listener.enterExpression_if(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpression_if" ):
 *                 listener.exitExpression_if(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expression_ifContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5729
 *                 listener.exitExpression_if(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpression_if" ):
 *                 return visitor.visitExpression_if(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Expression_ifContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20Expression_ifContext_9accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20Expression_ifContext_9accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Expression_ifContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 5729, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 5729, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5729, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expression_ifContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20Expression_ifContext_8accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Expression_ifContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":5730
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExpression_if" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitExpression_if(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitExpression_if); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5730, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5731
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExpression_if" ):
 *                 return visitor.visitExpression_if(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitExpression_if); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5731, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5731, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5731, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5731, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5731, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5731, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":5730
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExpression_if" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitExpression_if(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5733
 *                 return visitor.visitExpression_if(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5733, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5733, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5733, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5733, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5733, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5733, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":5729
 *                 listener.exitExpression_if(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpression_if" ):
 *                 return visitor.visitExpression_if(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expression_ifContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5738
 *     class Expression_simpleContext(ExpressionContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExpressionContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Expression_simpleContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Expression_simpleContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Expression_simpleContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Expression_simpleContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 5738, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 5738, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 5738, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5738, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expression_simpleContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Expression_simpleContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Expression_simpleContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":5739
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExpressionContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 5739, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5739, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5739, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5739, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5739, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5739, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5739, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5739, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5739, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5740
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExpressionContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def simple_expression(self):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5740, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5740, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5740, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5740, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5740, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5740, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5738
 *     class Expression_simpleContext(ExpressionContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExpressionContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expression_simpleContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5742
 *             self.copyFrom(ctx)
 * 
 *         def simple_expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Simple_expressionContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Expression_simpleContext_3simple_expression(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Expression_simpleContext_3simple_expression = {"simple_expression", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Expression_simpleContext_3simple_expression, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Expression_simpleContext_3simple_expression(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("simple_expression (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Expression_simpleContext_2simple_expression(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Expression_simpleContext_2simple_expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("simple_expression", 0);

  /* "pymola/generated/ModelicaParser.pyx":5743
 * 
 *         def simple_expression(self):
 *             return self.getTypedRuleContext(ModelicaParser.Simple_expressionContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5743, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5743, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Simple_expressionContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5743, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5743, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5743, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5743, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5743, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":5742
 *             self.copyFrom(ctx)
 * 
 *         def simple_expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Simple_expressionContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expression_simpleContext.simple_expression", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5746
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpression_simple" ):
 *                 listener.enterExpression_simple(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Expression_simpleContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Expression_simpleContext_5enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Expression_simpleContext_5enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Expression_simpleContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 5746, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 5746, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5746, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expression_simpleContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Expression_simpleContext_4enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Expression_simpleContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":5747
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExpression_simple" ):             # <<<<<<<<<<<<<<
 *                 listener.enterExpression_simple(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterExpression_simple); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5747, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5748
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExpression_simple" ):
 *                 listener.enterExpression_simple(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterExpression_simple); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5748, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5748, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5748, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5748, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5748, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5748, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":5747
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExpression_simple" ):             # <<<<<<<<<<<<<<
 *                 listener.enterExpression_simple(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5746
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpression_simple" ):
 *                 listener.enterExpression_simple(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expression_simpleContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5750
 *                 listener.enterExpression_simple(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpression_simple" ):
 *                 listener.exitExpression_simple(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Expression_simpleContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Expression_simpleContext_7exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Expression_simpleContext_7exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Expression_simpleContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 5750, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 5750, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5750, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expression_simpleContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Expression_simpleContext_6exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Expression_simpleContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":5751
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExpression_simple" ):             # <<<<<<<<<<<<<<
 *                 listener.exitExpression_simple(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitExpression_simple); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5751, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5752
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExpression_simple" ):
 *                 listener.exitExpression_simple(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitExpression_simple); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5752, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5752, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5752, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5752, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5752, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5752, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":5751
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExpression_simple" ):             # <<<<<<<<<<<<<<
 *                 listener.exitExpression_simple(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5750
 *                 listener.enterExpression_simple(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpression_simple" ):
 *                 listener.exitExpression_simple(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expression_simpleContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5754
 *                 listener.exitExpression_simple(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpression_simple" ):
 *                 return visitor.visitExpression_simple(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Expression_simpleContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Expression_simpleContext_9accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Expression_simpleContext_9accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Expression_simpleContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 5754, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 5754, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5754, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expression_simpleContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Expression_simpleContext_8accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Expression_simpleContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":5755
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExpression_simple" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitExpression_simple(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitExpression_simple); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5755, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5756
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExpression_simple" ):
 *                 return visitor.visitExpression_simple(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitExpression_simple); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5756, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5756, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5756, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5756, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5756, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5756, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":5755
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExpression_simple" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitExpression_simple(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5758
 *                 return visitor.visitExpression_simple(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5758, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5758, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5758, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5758, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5758, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5758, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":5754
 *                 listener.exitExpression_simple(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpression_simple" ):
 *                 return visitor.visitExpression_simple(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expression_simpleContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5762
 * 
 * 
 *     def expression(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.ExpressionContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_111expression(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_111expression = {"expression", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_111expression, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_111expression(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("expression (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_110expression(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_110expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v_token = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  char const *__pyx_t_22;
  __Pyx_RefNannySetupContext("expression", 0);

  /* "pymola/generated/ModelicaParser.pyx":5764
 *     def expression(self):
 * 
 *         localctx = ModelicaParser.ExpressionContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 108, self.RULE_expression)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5764, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_ExpressionContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5764, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5764, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5764, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5764, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5764, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5764, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5764, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5765
 * 
 *         localctx = ModelicaParser.ExpressionContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 108, self.RULE_expression)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5765, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_expression); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5765, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_108, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5765, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_108, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5765, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5765, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_108);
    __Pyx_GIVEREF(__pyx_int_108);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_108);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5765, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5766
 *         localctx = ModelicaParser.ExpressionContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 108, self.RULE_expression)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.state = 841
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 5766, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5767
 *         self.enterRule(localctx, 108, self.RULE_expression)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.state = 841
 *             self._errHandler.sync(self)
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":5768
 *         self._la = 0 # Token type
 *         try:
 *             self.state = 841             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_841) < 0) __PYX_ERR(0, 5768, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5769
 *         try:
 *             self.state = 841
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.T__13, ModelicaParser.T__16, ModelicaParser.T__19, ModelicaParser.T__29, ModelicaParser.T__55, ModelicaParser.T__70, ModelicaParser.T__73, ModelicaParser.T__74, ModelicaParser.T__75, ModelicaParser.INITIAL, ModelicaParser.IDENT, ModelicaParser.STRING, ModelicaParser.UNSIGNED_NUMBER]:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5769, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5769, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5769, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5769, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5769, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5769, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5769, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5770
 *             self.state = 841
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if token in [ModelicaParser.T__13, ModelicaParser.T__16, ModelicaParser.T__19, ModelicaParser.T__29, ModelicaParser.T__55, ModelicaParser.T__70, ModelicaParser.T__73, ModelicaParser.T__74, ModelicaParser.T__75, ModelicaParser.INITIAL, ModelicaParser.IDENT, ModelicaParser.STRING, ModelicaParser.UNSIGNED_NUMBER]:
 *                 localctx = ModelicaParser.Expression_simpleContext(self, localctx)
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5770, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5770, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__926, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5770, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_v_token = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5771
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.T__13, ModelicaParser.T__16, ModelicaParser.T__19, ModelicaParser.T__29, ModelicaParser.T__55, ModelicaParser.T__70, ModelicaParser.T__73, ModelicaParser.T__74, ModelicaParser.T__75, ModelicaParser.INITIAL, ModelicaParser.IDENT, ModelicaParser.STRING, ModelicaParser.UNSIGNED_NUMBER]:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Expression_simpleContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 1)
 */
        __Pyx_INCREF(__pyx_v_token);
        __pyx_t_1 = __pyx_v_token;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__13); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__16); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__19); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__29); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__55); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__70); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__73); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__74); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__75); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_INITIAL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_STRING); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_UNSIGNED_NUMBER); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5771, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_11 = __pyx_t_12;
        __pyx_L15_bool_binop_done:;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_12 = (__pyx_t_11 != 0);
        if (__pyx_t_12) {

          /* "pymola/generated/ModelicaParser.pyx":5772
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.T__13, ModelicaParser.T__16, ModelicaParser.T__19, ModelicaParser.T__29, ModelicaParser.T__55, ModelicaParser.T__70, ModelicaParser.T__73, ModelicaParser.T__74, ModelicaParser.T__75, ModelicaParser.INITIAL, ModelicaParser.IDENT, ModelicaParser.STRING, ModelicaParser.UNSIGNED_NUMBER]:
 *                 localctx = ModelicaParser.Expression_simpleContext(self, localctx)             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 1)
 *                 self.state = 823
 */
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5772, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Expression_simpleContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5772, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5772, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5772, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5772, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_2) {
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_v_localctx);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5772, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5773
 *             if token in [ModelicaParser.T__13, ModelicaParser.T__16, ModelicaParser.T__19, ModelicaParser.T__29, ModelicaParser.T__55, ModelicaParser.T__70, ModelicaParser.T__73, ModelicaParser.T__74, ModelicaParser.T__75, ModelicaParser.INITIAL, ModelicaParser.IDENT, ModelicaParser.STRING, ModelicaParser.UNSIGNED_NUMBER]:
 *                 localctx = ModelicaParser.Expression_simpleContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *                 self.state = 823
 *                 self.simple_expression()
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5773, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_1};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5773, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_1};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5773, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5773, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_1);
            __Pyx_GIVEREF(__pyx_int_1);
            PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_1);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5773, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5774
 *                 localctx = ModelicaParser.Expression_simpleContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 1)
 *                 self.state = 823             # <<<<<<<<<<<<<<
 *                 self.simple_expression()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_823) < 0) __PYX_ERR(0, 5774, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5775
 *                 self.enterOuterAlt(localctx, 1)
 *                 self.state = 823
 *                 self.simple_expression()             # <<<<<<<<<<<<<<
 *                 pass
 *             elif token in [ModelicaParser.T__39]:
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_simple_expression); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5775, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5775, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5775, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5771
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.T__13, ModelicaParser.T__16, ModelicaParser.T__19, ModelicaParser.T__29, ModelicaParser.T__55, ModelicaParser.T__70, ModelicaParser.T__73, ModelicaParser.T__74, ModelicaParser.T__75, ModelicaParser.INITIAL, ModelicaParser.IDENT, ModelicaParser.STRING, ModelicaParser.UNSIGNED_NUMBER]:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Expression_simpleContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 1)
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":5777
 *                 self.simple_expression()
 *                 pass
 *             elif token in [ModelicaParser.T__39]:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Expression_ifContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 2)
 */
        __Pyx_INCREF(__pyx_v_token);
        __pyx_t_1 = __pyx_v_token;
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5777, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__39); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5777, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5777, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5777, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_11 = (__pyx_t_12 != 0);
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":5778
 *                 pass
 *             elif token in [ModelicaParser.T__39]:
 *                 localctx = ModelicaParser.Expression_ifContext(self, localctx)             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 2)
 *                 self.state = 824
 */
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5778, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Expression_ifContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5778, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5778, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5778, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5778, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_7) {
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_v_localctx);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5778, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5779
 *             elif token in [ModelicaParser.T__39]:
 *                 localctx = ModelicaParser.Expression_ifContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 2)             # <<<<<<<<<<<<<<
 *                 self.state = 824
 *                 self.match(ModelicaParser.T__39)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5779, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5779, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5779, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5779, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_2);
            __Pyx_GIVEREF(__pyx_int_2);
            PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_2);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5779, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5780
 *                 localctx = ModelicaParser.Expression_ifContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 2)
 *                 self.state = 824             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__39)
 *                 self.state = 825
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_824) < 0) __PYX_ERR(0, 5780, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5781
 *                 self.enterOuterAlt(localctx, 2)
 *                 self.state = 824
 *                 self.match(ModelicaParser.T__39)             # <<<<<<<<<<<<<<
 *                 self.state = 825
 *                 self.expression()
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5781, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5781, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__39); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5781, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5781, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5781, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5781, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5781, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5781, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5782
 *                 self.state = 824
 *                 self.match(ModelicaParser.T__39)
 *                 self.state = 825             # <<<<<<<<<<<<<<
 *                 self.expression()
 *                 self.state = 826
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_825) < 0) __PYX_ERR(0, 5782, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5783
 *                 self.match(ModelicaParser.T__39)
 *                 self.state = 825
 *                 self.expression()             # <<<<<<<<<<<<<<
 *                 self.state = 826
 *                 self.match(ModelicaParser.T__45)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expression); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5783, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_4 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (__pyx_t_4) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5783, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5783, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5784
 *                 self.state = 825
 *                 self.expression()
 *                 self.state = 826             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__45)
 *                 self.state = 827
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_826) < 0) __PYX_ERR(0, 5784, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5785
 *                 self.expression()
 *                 self.state = 826
 *                 self.match(ModelicaParser.T__45)             # <<<<<<<<<<<<<<
 *                 self.state = 827
 *                 self.expression()
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5785, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5785, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__45); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5785, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_4) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5785, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5785, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5785, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5785, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4); __pyx_t_4 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5785, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5786
 *                 self.state = 826
 *                 self.match(ModelicaParser.T__45)
 *                 self.state = 827             # <<<<<<<<<<<<<<
 *                 self.expression()
 *                 self.state = 835
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_827) < 0) __PYX_ERR(0, 5786, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5787
 *                 self.match(ModelicaParser.T__45)
 *                 self.state = 827
 *                 self.expression()             # <<<<<<<<<<<<<<
 *                 self.state = 835
 *                 self._errHandler.sync(self)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expression); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5787, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5787, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5787, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5788
 *                 self.state = 827
 *                 self.expression()
 *                 self.state = 835             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_835) < 0) __PYX_ERR(0, 5788, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5789
 *                 self.expression()
 *                 self.state = 835
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 *                 while _la==ModelicaParser.T__46:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5789, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5789, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5789, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5789, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5789, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5789, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5789, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5790
 *                 self.state = 835
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 while _la==ModelicaParser.T__46:
 *                     self.state = 828
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5790, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5790, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__927, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5790, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_v__la = __pyx_t_1;
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5791
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 while _la==ModelicaParser.T__46:             # <<<<<<<<<<<<<<
 *                     self.state = 828
 *                     self.match(ModelicaParser.T__46)
 */
          while (1) {
            __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5791, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__46); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5791, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5791, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 5791, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (!__pyx_t_11) break;

            /* "pymola/generated/ModelicaParser.pyx":5792
 *                 _la = self._input.LA(1)
 *                 while _la==ModelicaParser.T__46:
 *                     self.state = 828             # <<<<<<<<<<<<<<
 *                     self.match(ModelicaParser.T__46)
 *                     self.state = 829
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_828) < 0) __PYX_ERR(0, 5792, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":5793
 *                 while _la==ModelicaParser.T__46:
 *                     self.state = 828
 *                     self.match(ModelicaParser.T__46)             # <<<<<<<<<<<<<<
 *                     self.state = 829
 *                     self.expression()
 */
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5793, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5793, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__46); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5793, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
              __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
              if (likely(__pyx_t_3)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
                __Pyx_INCREF(__pyx_t_3);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_7, function);
              }
            }
            if (!__pyx_t_3) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5793, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_7)) {
                PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_2};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5793, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
                PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_2};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5793, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              } else
              #endif
              {
                __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5793, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __pyx_t_3 = NULL;
                __Pyx_GIVEREF(__pyx_t_2);
                PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_2);
                __pyx_t_2 = 0;
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5793, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":5794
 *                     self.state = 828
 *                     self.match(ModelicaParser.T__46)
 *                     self.state = 829             # <<<<<<<<<<<<<<
 *                     self.expression()
 *                     self.state = 830
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_829) < 0) __PYX_ERR(0, 5794, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":5795
 *                     self.match(ModelicaParser.T__46)
 *                     self.state = 829
 *                     self.expression()             # <<<<<<<<<<<<<<
 *                     self.state = 830
 *                     self.match(ModelicaParser.T__45)
 */
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expression); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5795, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_4 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
              __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
              if (likely(__pyx_t_4)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
                __Pyx_INCREF(__pyx_t_4);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_7, function);
              }
            }
            if (__pyx_t_4) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5795, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            } else {
              __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5795, __pyx_L6_error)
            }
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":5796
 *                     self.state = 829
 *                     self.expression()
 *                     self.state = 830             # <<<<<<<<<<<<<<
 *                     self.match(ModelicaParser.T__45)
 *                     self.state = 831
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_830) < 0) __PYX_ERR(0, 5796, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":5797
 *                     self.expression()
 *                     self.state = 830
 *                     self.match(ModelicaParser.T__45)             # <<<<<<<<<<<<<<
 *                     self.state = 831
 *                     self.expression()
 */
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5797, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5797, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__45); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5797, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
              __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
              if (likely(__pyx_t_4)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
                __Pyx_INCREF(__pyx_t_4);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_7, function);
              }
            }
            if (!__pyx_t_4) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5797, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_7)) {
                PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_2};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5797, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
                PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_2};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5797, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              } else
              #endif
              {
                __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5797, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __pyx_t_4 = NULL;
                __Pyx_GIVEREF(__pyx_t_2);
                PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_t_2);
                __pyx_t_2 = 0;
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5797, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":5798
 *                     self.state = 830
 *                     self.match(ModelicaParser.T__45)
 *                     self.state = 831             # <<<<<<<<<<<<<<
 *                     self.expression()
 *                     self.state = 837
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_831) < 0) __PYX_ERR(0, 5798, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":5799
 *                     self.match(ModelicaParser.T__45)
 *                     self.state = 831
 *                     self.expression()             # <<<<<<<<<<<<<<
 *                     self.state = 837
 *                     self._errHandler.sync(self)
 */
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expression); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5799, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_3 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
              __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
              if (likely(__pyx_t_3)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
                __Pyx_INCREF(__pyx_t_3);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_7, function);
              }
            }
            if (__pyx_t_3) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5799, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else {
              __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5799, __pyx_L6_error)
            }
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":5800
 *                     self.state = 831
 *                     self.expression()
 *                     self.state = 837             # <<<<<<<<<<<<<<
 *                     self._errHandler.sync(self)
 *                     _la = self._input.LA(1)
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_837) < 0) __PYX_ERR(0, 5800, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":5801
 *                     self.expression()
 *                     self.state = 837
 *                     self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                     _la = self._input.LA(1)
 * 
 */
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5801, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_sync); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5801, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __pyx_t_7 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
              __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
              if (likely(__pyx_t_7)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                __Pyx_INCREF(__pyx_t_7);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_3, function);
              }
            }
            if (!__pyx_t_7) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5801, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_3)) {
                PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5801, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
                PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5801, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else
              #endif
              {
                __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5801, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_7); __pyx_t_7 = NULL;
                __Pyx_INCREF(__pyx_v_self);
                __Pyx_GIVEREF(__pyx_v_self);
                PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5801, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":5802
 *                     self.state = 837
 *                     self._errHandler.sync(self)
 *                     _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *                 self.state = 838
 */
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5802, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5802, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__928, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5802, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_1);
            __pyx_t_1 = 0;
          }

          /* "pymola/generated/ModelicaParser.pyx":5804
 *                     _la = self._input.LA(1)
 * 
 *                 self.state = 838             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__47)
 *                 self.state = 839
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_838) < 0) __PYX_ERR(0, 5804, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5805
 * 
 *                 self.state = 838
 *                 self.match(ModelicaParser.T__47)             # <<<<<<<<<<<<<<
 *                 self.state = 839
 *                 self.expression()
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5805, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5805, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__47); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5805, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5805, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5805, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5805, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5805, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_GIVEREF(__pyx_t_7);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_7);
              __pyx_t_7 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5805, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5806
 *                 self.state = 838
 *                 self.match(ModelicaParser.T__47)
 *                 self.state = 839             # <<<<<<<<<<<<<<
 *                 self.expression()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_839) < 0) __PYX_ERR(0, 5806, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5807
 *                 self.match(ModelicaParser.T__47)
 *                 self.state = 839
 *                 self.expression()             # <<<<<<<<<<<<<<
 *                 pass
 *             else:
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expression); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5807, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_4 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (__pyx_t_4) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5807, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5807, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5777
 *                 self.simple_expression()
 *                 pass
 *             elif token in [ModelicaParser.T__39]:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Expression_ifContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 2)
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":5810
 *                 pass
 *             else:
 *                 raise NoViableAltException(self)             # <<<<<<<<<<<<<<
 * 
 *         except RecognitionException as re:
 */
        /*else*/ {
          __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_NoViableAltException); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5810, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_4 = NULL;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (!__pyx_t_4) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5810, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5810, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5810, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5810, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4); __pyx_t_4 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5810, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_Raise(__pyx_t_1, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __PYX_ERR(0, 5810, __pyx_L6_error)
        }
        __pyx_L14:;

        /* "pymola/generated/ModelicaParser.pyx":5767
 *         self.enterRule(localctx, 108, self.RULE_expression)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.state = 841
 *             self._errHandler.sync(self)
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":5812
 *                 raise NoViableAltException(self)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5812, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.expression", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_3, &__pyx_t_7) < 0) __PYX_ERR(0, 5812, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_3);
        __pyx_v_re = __pyx_t_3;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":5813
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 5813, __pyx_L35_error)

          /* "pymola/generated/ModelicaParser.pyx":5814
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5814, __pyx_L35_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5814, __pyx_L35_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_re};
            __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5814, __pyx_L35_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_re};
            __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5814, __pyx_L35_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 5814, __pyx_L35_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_2) {
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_2); __pyx_t_2 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_13, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5814, __pyx_L35_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5815
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5815, __pyx_L35_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 5815, __pyx_L35_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5815, __pyx_L35_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5815, __pyx_L35_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5815, __pyx_L35_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_2, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5815, __pyx_L35_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":5812
 *                 raise NoViableAltException(self)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L36;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L35_error:;
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_6 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L8_except_error;
          }
          __pyx_L36:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":5767
 *         self.enterRule(localctx, 108, self.RULE_expression)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.state = 841
 *             self._errHandler.sync(self)
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":5817
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5817, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5817, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5817, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_14 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5817, __pyx_L42_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5817, __pyx_L42_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5817, __pyx_L42_error)
        }
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
      __pyx_L42_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":5818
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Simple_expressionContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":5762
 * 
 * 
 *     def expression(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.ExpressionContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.expression", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v_token);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5822
 *     class Simple_expressionContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 5822, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 5822, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5822, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Simple_expressionContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Simple_expressionContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Simple_expressionContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":5823
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 5823, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5823, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5823, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5823, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5823, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5823, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5823, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5823, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5824
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def expr(self, i:int=None):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 5824, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5822
 *     class Simple_expressionContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Simple_expressionContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5826
 *             self.parser = parser
 * 
 *         def expr(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_3expr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_3expr = {"expr", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_3expr, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_3expr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("expr (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "expr") < 0)) __PYX_ERR(0, 5826, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("expr", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5826, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Simple_expressionContext.expr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_2expr(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_2expr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("expr", 0);

  /* "pymola/generated/ModelicaParser.pyx":5827
 * 
 *         def expr(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5828
 *         def expr(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.ExprContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5828, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5828, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_ExprContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5828, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5828, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5828, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5828, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5828, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5828, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":5827
 * 
 *         def expr(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5830
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.ExprContext,i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5830, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5830, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_ExprContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5830, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5830, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5830, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5830, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5830, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":5826
 *             self.parser = parser
 * 
 *         def expr(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Simple_expressionContext.expr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5833
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_simple_expression
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_5getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_5getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_4getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":5834
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_simple_expression             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_simple_expression); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":5833
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_simple_expression
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Simple_expressionContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5836
 *             return ModelicaParser.RULE_simple_expression
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterSimple_expression" ):
 *                 listener.enterSimple_expression(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_7enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_7enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 5836, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 5836, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5836, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Simple_expressionContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_6enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":5837
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterSimple_expression" ):             # <<<<<<<<<<<<<<
 *                 listener.enterSimple_expression(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterSimple_expression); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5837, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5838
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterSimple_expression" ):
 *                 listener.enterSimple_expression(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterSimple_expression); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5838, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5838, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5838, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5838, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5838, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5838, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":5837
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterSimple_expression" ):             # <<<<<<<<<<<<<<
 *                 listener.enterSimple_expression(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5836
 *             return ModelicaParser.RULE_simple_expression
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterSimple_expression" ):
 *                 listener.enterSimple_expression(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Simple_expressionContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5840
 *                 listener.enterSimple_expression(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitSimple_expression" ):
 *                 listener.exitSimple_expression(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_9exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_9exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 5840, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 5840, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5840, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Simple_expressionContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_8exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":5841
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitSimple_expression" ):             # <<<<<<<<<<<<<<
 *                 listener.exitSimple_expression(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitSimple_expression); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5841, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5842
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitSimple_expression" ):
 *                 listener.exitSimple_expression(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitSimple_expression); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5842, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5842, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5842, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5842, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5842, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5842, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":5841
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitSimple_expression" ):             # <<<<<<<<<<<<<<
 *                 listener.exitSimple_expression(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5840
 *                 listener.enterSimple_expression(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitSimple_expression" ):
 *                 listener.exitSimple_expression(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Simple_expressionContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5844
 *                 listener.exitSimple_expression(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitSimple_expression" ):
 *                 return visitor.visitSimple_expression(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_11accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_11accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 5844, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 5844, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5844, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Simple_expressionContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_10accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":5845
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitSimple_expression" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitSimple_expression(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitSimple_expression); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5845, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5846
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitSimple_expression" ):
 *                 return visitor.visitSimple_expression(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitSimple_expression); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5846, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5846, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5846, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5846, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5846, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5846, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":5845
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitSimple_expression" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitSimple_expression(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5848
 *                 return visitor.visitSimple_expression(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5848, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5848, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5848, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5848, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5848, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5848, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":5844
 *                 listener.exitSimple_expression(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitSimple_expression" ):
 *                 return visitor.visitSimple_expression(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Simple_expressionContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5853
 * 
 * 
 *     def simple_expression(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Simple_expressionContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_113simple_expression(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_113simple_expression = {"simple_expression", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_113simple_expression, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_113simple_expression(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("simple_expression (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_112simple_expression(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_112simple_expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  char const *__pyx_t_21;
  __Pyx_RefNannySetupContext("simple_expression", 0);

  /* "pymola/generated/ModelicaParser.pyx":5855
 *     def simple_expression(self):
 * 
 *         localctx = ModelicaParser.Simple_expressionContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 110, self.RULE_simple_expression)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5855, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Simple_expressionContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5855, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5855, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5855, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5855, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5855, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5855, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5855, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5856
 * 
 *         localctx = ModelicaParser.Simple_expressionContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 110, self.RULE_simple_expression)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_simple_expression); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_110, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5856, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_110, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5856, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5856, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_110);
    __Pyx_GIVEREF(__pyx_int_110);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_110);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5856, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5857
 *         localctx = ModelicaParser.Simple_expressionContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 110, self.RULE_simple_expression)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 5857, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5858
 *         self.enterRule(localctx, 110, self.RULE_simple_expression)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 843
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":5859
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 843
 *             self.expr(0)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5859, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5859, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5859, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5859, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5859, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5860
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 843             # <<<<<<<<<<<<<<
 *             self.expr(0)
 *             self.state = 850
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_843) < 0) __PYX_ERR(0, 5860, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5861
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 843
 *             self.expr(0)             # <<<<<<<<<<<<<<
 *             self.state = 850
 *             self._errHandler.sync(self)
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5861, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__929, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5861, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5862
 *             self.state = 843
 *             self.expr(0)
 *             self.state = 850             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_850) < 0) __PYX_ERR(0, 5862, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":5863
 *             self.expr(0)
 *             self.state = 850
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__17:
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5863, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5863, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_1) {
          __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5863, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_v_self};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5863, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_v_self};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5863, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5863, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1); __pyx_t_1 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5863, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5864
 *             self.state = 850
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.T__17:
 *                 self.state = 844
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5864, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5864, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__930, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5864, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v__la = __pyx_t_3;
        __pyx_t_3 = 0;

        /* "pymola/generated/ModelicaParser.pyx":5865
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__17:             # <<<<<<<<<<<<<<
 *                 self.state = 844
 *                 self.match(ModelicaParser.T__17)
 */
        __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5865, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__17); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5865, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyObject_RichCompare(__pyx_v__la, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5865, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 5865, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":5866
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__17:
 *                 self.state = 844             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__17)
 *                 self.state = 845
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_844) < 0) __PYX_ERR(0, 5866, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5867
 *             if _la==ModelicaParser.T__17:
 *                 self.state = 844
 *                 self.match(ModelicaParser.T__17)             # <<<<<<<<<<<<<<
 *                 self.state = 845
 *                 self.expr(0)
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5867, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5867, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__17); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5867, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5867, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_1};
              __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5867, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_1};
              __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5867, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5867, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_GIVEREF(__pyx_t_1);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_1);
              __pyx_t_1 = 0;
              __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5867, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5868
 *                 self.state = 844
 *                 self.match(ModelicaParser.T__17)
 *                 self.state = 845             # <<<<<<<<<<<<<<
 *                 self.expr(0)
 *                 self.state = 848
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_845) < 0) __PYX_ERR(0, 5868, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5869
 *                 self.match(ModelicaParser.T__17)
 *                 self.state = 845
 *                 self.expr(0)             # <<<<<<<<<<<<<<
 *                 self.state = 848
 *                 self._errHandler.sync(self)
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5869, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__931, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5869, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5870
 *                 self.state = 845
 *                 self.expr(0)
 *                 self.state = 848             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_848) < 0) __PYX_ERR(0, 5870, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":5871
 *                 self.expr(0)
 *                 self.state = 848
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__17:
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5871, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5871, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (!__pyx_t_3) {
            __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5871, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
              __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5871, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_7);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
              __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5871, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_7);
            } else
            #endif
            {
              __pyx_t_1 = PyTuple_New(1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5871, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3); __pyx_t_3 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_1, 0+1, __pyx_v_self);
              __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_1, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5871, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5872
 *                 self.state = 848
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if _la==ModelicaParser.T__17:
 *                     self.state = 846
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5872, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_LA); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5872, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__932, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5872, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_7);
          __pyx_t_7 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5873
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__17:             # <<<<<<<<<<<<<<
 *                     self.state = 846
 *                     self.match(ModelicaParser.T__17)
 */
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5873, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__17); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5873, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyObject_RichCompare(__pyx_v__la, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5873, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 5873, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (__pyx_t_11) {

            /* "pymola/generated/ModelicaParser.pyx":5874
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__17:
 *                     self.state = 846             # <<<<<<<<<<<<<<
 *                     self.match(ModelicaParser.T__17)
 *                     self.state = 847
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_846) < 0) __PYX_ERR(0, 5874, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":5875
 *                 if _la==ModelicaParser.T__17:
 *                     self.state = 846
 *                     self.match(ModelicaParser.T__17)             # <<<<<<<<<<<<<<
 *                     self.state = 847
 *                     self.expr(0)
 */
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5875, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5875, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__17); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5875, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
              __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
              if (likely(__pyx_t_1)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                __Pyx_INCREF(__pyx_t_1);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_4, function);
              }
            }
            if (!__pyx_t_1) {
              __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5875, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_7);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_4)) {
                PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_t_3};
                __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5875, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                __Pyx_GOTREF(__pyx_t_7);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
                PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_t_3};
                __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5875, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                __Pyx_GOTREF(__pyx_t_7);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              } else
              #endif
              {
                __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5875, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1); __pyx_t_1 = NULL;
                __Pyx_GIVEREF(__pyx_t_3);
                PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_3);
                __pyx_t_3 = 0;
                __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5875, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_7);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

            /* "pymola/generated/ModelicaParser.pyx":5876
 *                     self.state = 846
 *                     self.match(ModelicaParser.T__17)
 *                     self.state = 847             # <<<<<<<<<<<<<<
 *                     self.expr(0)
 * 
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_847) < 0) __PYX_ERR(0, 5876, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":5877
 *                     self.match(ModelicaParser.T__17)
 *                     self.state = 847
 *                     self.expr(0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5877, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__933, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5877, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

            /* "pymola/generated/ModelicaParser.pyx":5873
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.T__17:             # <<<<<<<<<<<<<<
 *                     self.state = 846
 *                     self.match(ModelicaParser.T__17)
 */
          }

          /* "pymola/generated/ModelicaParser.pyx":5865
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__17:             # <<<<<<<<<<<<<<
 *                 self.state = 844
 *                 self.match(ModelicaParser.T__17)
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":5858
 *         self.enterRule(localctx, 110, self.RULE_simple_expression)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 843
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "pymola/generated/ModelicaParser.pyx":5882
 * 
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5882, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.simple_expression", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_7, &__pyx_t_2) < 0) __PYX_ERR(0, 5882, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_7);
        __pyx_v_re = __pyx_t_7;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":5883
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 5883, __pyx_L21_error)

          /* "pymola/generated/ModelicaParser.pyx":5884
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5884, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5884, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5884, __pyx_L21_error)
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5884, __pyx_L21_error)
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_12 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 5884, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_12);
            if (__pyx_t_1) {
              __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_1); __pyx_t_1 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_12, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5884, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":5885
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5885, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 5885, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_12);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_12, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5885, __pyx_L21_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5885, __pyx_L21_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_1 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5885, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_1);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_1, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5885, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":5882
 * 
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L22;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L21_error:;
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __pyx_t_6 = __pyx_lineno; __pyx_t_13 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
            }
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_ErrRestore(__pyx_t_15, __pyx_t_16, __pyx_t_17);
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_13; __pyx_filename = __pyx_t_14;
            goto __pyx_L8_except_error;
          }
          __pyx_L22:;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":5858
 *         self.enterRule(localctx, 110, self.RULE_simple_expression)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 843
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":5887
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5887, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5887, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5887, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_19, &__pyx_t_18);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_18);
      __pyx_t_13 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_21 = __pyx_filename;
      {
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5887, __pyx_L28_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (__pyx_t_4) {
          __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5887, __pyx_L28_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else {
          __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5887, __pyx_L28_error)
        }
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __pyx_lineno = __pyx_t_13; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_21;
      goto __pyx_L1_error;
      __pyx_L28_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":5888
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class ExprContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":5853
 * 
 * 
 *     def simple_expression(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Simple_expressionContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.simple_expression", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5892
 *     class ExprContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_11ExprContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_11ExprContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_11ExprContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_11ExprContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 5892, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 5892, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5892, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ExprContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_11ExprContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_11ExprContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":5893
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 5893, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5893, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5893, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5893, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5893, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5893, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5893, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5893, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5894
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 5894, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5892
 *     class ExprContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ExprContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5897
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_expr
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_11ExprContext_3getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_11ExprContext_3getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_11ExprContext_3getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_11ExprContext_3getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_11ExprContext_2getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_11ExprContext_2getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":5898
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_expr             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5898, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_expr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5898, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":5897
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_expr
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ExprContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5901
 * 
 * 
 *         def copyFrom(self, ctx:ParserRuleContext):             # <<<<<<<<<<<<<<
 *             super().copyFrom(ctx)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_11ExprContext_5copyFrom(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_11ExprContext_5copyFrom = {"copyFrom", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_11ExprContext_5copyFrom, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_11ExprContext_5copyFrom(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copyFrom (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx_2,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("copyFrom", 1, 2, 2, 1); __PYX_ERR(0, 5901, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "copyFrom") < 0)) __PYX_ERR(0, 5901, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("copyFrom", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5901, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ExprContext.copyFrom", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_11ExprContext_4copyFrom(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_11ExprContext_4copyFrom(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("copyFrom", 0);

  /* "pymola/generated/ModelicaParser.pyx":5902
 * 
 *         def copyFrom(self, ctx:ParserRuleContext):
 *             super().copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 5902, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5902, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5902, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5902, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5902, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5902, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5902, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5902, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5902, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5901
 * 
 * 
 *         def copyFrom(self, ctx:ParserRuleContext):             # <<<<<<<<<<<<<<
 *             super().copyFrom(ctx)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ExprContext.copyFrom", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5907
 *     class Expr_negContext(ExprContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.op = None # Token
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_negContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_negContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_negContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_negContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 5907, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 5907, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 5907, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5907, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_negContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_negContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_negContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":5908
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.op = None # Token
 *             self.copyFrom(ctx)
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 5908, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5908, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5908, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5908, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5908, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5908, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5908, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5908, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5908, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5909
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext
 *             super().__init__(parser)
 *             self.op = None # Token             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_op, Py_None) < 0) __PYX_ERR(0, 5909, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5910
 *             super().__init__(parser)
 *             self.op = None # Token
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def expr(self):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5910, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5910, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5910, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5910, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5910, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5910, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5907
 *     class Expr_negContext(ExprContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.op = None # Token
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_negContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5912
 *             self.copyFrom(ctx)
 * 
 *         def expr(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExprContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_negContext_3expr(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_negContext_3expr = {"expr", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_negContext_3expr, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_negContext_3expr(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("expr (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_negContext_2expr(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_negContext_2expr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("expr", 0);

  /* "pymola/generated/ModelicaParser.pyx":5913
 * 
 *         def expr(self):
 *             return self.getTypedRuleContext(ModelicaParser.ExprContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5913, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5913, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_ExprContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5913, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5913, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5913, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5913, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5913, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":5912
 *             self.copyFrom(ctx)
 * 
 *         def expr(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExprContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_negContext.expr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5916
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpr_neg" ):
 *                 listener.enterExpr_neg(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_negContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_negContext_5enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_negContext_5enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_negContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 5916, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 5916, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5916, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_negContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_negContext_4enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_negContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":5917
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExpr_neg" ):             # <<<<<<<<<<<<<<
 *                 listener.enterExpr_neg(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterExpr_neg); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5917, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5918
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExpr_neg" ):
 *                 listener.enterExpr_neg(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterExpr_neg); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5918, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5918, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5918, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5918, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5918, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5918, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":5917
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExpr_neg" ):             # <<<<<<<<<<<<<<
 *                 listener.enterExpr_neg(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5916
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpr_neg" ):
 *                 listener.enterExpr_neg(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_negContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5920
 *                 listener.enterExpr_neg(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpr_neg" ):
 *                 listener.exitExpr_neg(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_negContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_negContext_7exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_negContext_7exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_negContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 5920, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 5920, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5920, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_negContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_negContext_6exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_negContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":5921
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExpr_neg" ):             # <<<<<<<<<<<<<<
 *                 listener.exitExpr_neg(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitExpr_neg); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5921, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5922
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExpr_neg" ):
 *                 listener.exitExpr_neg(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitExpr_neg); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5922, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5922, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5922, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5922, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5922, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5922, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":5921
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExpr_neg" ):             # <<<<<<<<<<<<<<
 *                 listener.exitExpr_neg(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5920
 *                 listener.enterExpr_neg(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpr_neg" ):
 *                 listener.exitExpr_neg(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_negContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5924
 *                 listener.exitExpr_neg(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpr_neg" ):
 *                 return visitor.visitExpr_neg(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_negContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_negContext_9accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_negContext_9accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_negContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 5924, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 5924, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5924, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_negContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_negContext_8accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_negContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":5925
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExpr_neg" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitExpr_neg(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitExpr_neg); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5925, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5926
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExpr_neg" ):
 *                 return visitor.visitExpr_neg(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitExpr_neg); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5926, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5926, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5926, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5926, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5926, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5926, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":5925
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExpr_neg" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitExpr_neg(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5928
 *                 return visitor.visitExpr_neg(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5928, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5928, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5928, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5928, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5928, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5928, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":5924
 *                 listener.exitExpr_neg(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpr_neg" ):
 *                 return visitor.visitExpr_neg(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_negContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5933
 *     class Expr_addContext(ExprContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.op = None # Token
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_addContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_addContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_addContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_addContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 5933, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 5933, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 5933, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5933, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_addContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_addContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_addContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":5934
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.op = None # Token
 *             self.copyFrom(ctx)
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 5934, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5934, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5934, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5934, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5934, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5934, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5934, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5934, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5934, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5935
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext
 *             super().__init__(parser)
 *             self.op = None # Token             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_op, Py_None) < 0) __PYX_ERR(0, 5935, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5936
 *             super().__init__(parser)
 *             self.op = None # Token
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def expr(self, i:int=None):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5936, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5936, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5936, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5936, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5936, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5936, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5933
 *     class Expr_addContext(ExprContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.op = None # Token
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_addContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5938
 *             self.copyFrom(ctx)
 * 
 *         def expr(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_addContext_3expr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_addContext_3expr = {"expr", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_addContext_3expr, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_addContext_3expr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("expr (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "expr") < 0)) __PYX_ERR(0, 5938, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("expr", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5938, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_addContext.expr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_addContext_2expr(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_addContext_2expr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("expr", 0);

  /* "pymola/generated/ModelicaParser.pyx":5939
 * 
 *         def expr(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5940
 *         def expr(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.ExprContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5940, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5940, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_ExprContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5940, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5940, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5940, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5940, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5940, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5940, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":5939
 * 
 *         def expr(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5942
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.ExprContext,i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5942, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5942, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_ExprContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5942, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5942, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5942, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5942, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5942, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":5938
 *             self.copyFrom(ctx)
 * 
 *         def expr(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_addContext.expr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5945
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpr_add" ):
 *                 listener.enterExpr_add(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_addContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_addContext_5enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_addContext_5enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_addContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 5945, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 5945, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5945, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_addContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_addContext_4enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_addContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":5946
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExpr_add" ):             # <<<<<<<<<<<<<<
 *                 listener.enterExpr_add(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterExpr_add); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5946, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5947
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExpr_add" ):
 *                 listener.enterExpr_add(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterExpr_add); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5947, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5947, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5947, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5947, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5947, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5947, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":5946
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExpr_add" ):             # <<<<<<<<<<<<<<
 *                 listener.enterExpr_add(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5945
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpr_add" ):
 *                 listener.enterExpr_add(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_addContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5949
 *                 listener.enterExpr_add(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpr_add" ):
 *                 listener.exitExpr_add(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_addContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_addContext_7exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_addContext_7exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_addContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 5949, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 5949, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5949, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_addContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_addContext_6exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_addContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":5950
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExpr_add" ):             # <<<<<<<<<<<<<<
 *                 listener.exitExpr_add(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitExpr_add); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5950, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5951
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExpr_add" ):
 *                 listener.exitExpr_add(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitExpr_add); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5951, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5951, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5951, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5951, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5951, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5951, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":5950
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExpr_add" ):             # <<<<<<<<<<<<<<
 *                 listener.exitExpr_add(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5949
 *                 listener.enterExpr_add(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpr_add" ):
 *                 listener.exitExpr_add(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_addContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5953
 *                 listener.exitExpr_add(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpr_add" ):
 *                 return visitor.visitExpr_add(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_addContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_addContext_9accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_addContext_9accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_addContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 5953, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 5953, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5953, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_addContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_addContext_8accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_addContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":5954
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExpr_add" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitExpr_add(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitExpr_add); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5954, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5955
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExpr_add" ):
 *                 return visitor.visitExpr_add(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitExpr_add); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5955, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5955, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5955, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5955, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5955, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5955, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":5954
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExpr_add" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitExpr_add(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5957
 *                 return visitor.visitExpr_add(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5957, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5957, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5957, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5957, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5957, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5957, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":5953
 *                 listener.exitExpr_add(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpr_add" ):
 *                 return visitor.visitExpr_add(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_addContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5962
 *     class Expr_expContext(ExprContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.op = None # Token
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_expContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_expContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_expContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_expContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 5962, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 5962, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 5962, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5962, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_expContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_expContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_expContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":5963
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.op = None # Token
 *             self.copyFrom(ctx)
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 5963, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5963, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5963, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5963, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5963, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5963, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5963, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5963, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5963, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5964
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext
 *             super().__init__(parser)
 *             self.op = None # Token             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_op, Py_None) < 0) __PYX_ERR(0, 5964, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5965
 *             super().__init__(parser)
 *             self.op = None # Token
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def primary(self, i:int=None):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5965, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5965, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5965, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5965, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5965, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5965, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5962
 *     class Expr_expContext(ExprContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.op = None # Token
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_expContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5967
 *             self.copyFrom(ctx)
 * 
 *         def primary(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.PrimaryContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_expContext_3primary(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_expContext_3primary = {"primary", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_expContext_3primary, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_expContext_3primary(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("primary (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "primary") < 0)) __PYX_ERR(0, 5967, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("primary", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5967, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_expContext.primary", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_expContext_2primary(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_expContext_2primary(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("primary", 0);

  /* "pymola/generated/ModelicaParser.pyx":5968
 * 
 *         def primary(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.PrimaryContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5969
 *         def primary(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.PrimaryContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.PrimaryContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5969, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5969, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_PrimaryContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5969, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5969, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5969, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5969, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5969, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5969, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":5968
 * 
 *         def primary(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.PrimaryContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5971
 *                 return self.getTypedRuleContexts(ModelicaParser.PrimaryContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.PrimaryContext,i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5971, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5971, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_PrimaryContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5971, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5971, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5971, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5971, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5971, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":5967
 *             self.copyFrom(ctx)
 * 
 *         def primary(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.PrimaryContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_expContext.primary", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5974
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpr_exp" ):
 *                 listener.enterExpr_exp(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_expContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_expContext_5enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_expContext_5enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_expContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 5974, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 5974, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5974, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_expContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_expContext_4enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_expContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":5975
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExpr_exp" ):             # <<<<<<<<<<<<<<
 *                 listener.enterExpr_exp(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterExpr_exp); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5975, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5976
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExpr_exp" ):
 *                 listener.enterExpr_exp(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterExpr_exp); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5976, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5976, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5976, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5976, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5976, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5976, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":5975
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExpr_exp" ):             # <<<<<<<<<<<<<<
 *                 listener.enterExpr_exp(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5974
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpr_exp" ):
 *                 listener.enterExpr_exp(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_expContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5978
 *                 listener.enterExpr_exp(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpr_exp" ):
 *                 listener.exitExpr_exp(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_expContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_expContext_7exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_expContext_7exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_expContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 5978, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 5978, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5978, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_expContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_expContext_6exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_expContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":5979
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExpr_exp" ):             # <<<<<<<<<<<<<<
 *                 listener.exitExpr_exp(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitExpr_exp); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5979, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5980
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExpr_exp" ):
 *                 listener.exitExpr_exp(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitExpr_exp); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5980, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5980, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5980, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5980, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5980, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5980, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":5979
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExpr_exp" ):             # <<<<<<<<<<<<<<
 *                 listener.exitExpr_exp(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5978
 *                 listener.enterExpr_exp(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpr_exp" ):
 *                 listener.exitExpr_exp(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_expContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5982
 *                 listener.exitExpr_exp(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpr_exp" ):
 *                 return visitor.visitExpr_exp(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_expContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_expContext_9accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_expContext_9accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_expContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 5982, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 5982, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5982, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_expContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_expContext_8accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_expContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":5983
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExpr_exp" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitExpr_exp(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitExpr_exp); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 5983, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5984
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExpr_exp" ):
 *                 return visitor.visitExpr_exp(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitExpr_exp); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5984, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5984, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5984, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5984, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5984, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5984, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":5983
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExpr_exp" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitExpr_exp(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5986
 *                 return visitor.visitExpr_exp(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5986, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5986, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5986, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5986, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5986, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5986, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":5982
 *                 listener.exitExpr_exp(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpr_exp" ):
 *                 return visitor.visitExpr_exp(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_expContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5991
 *     class Expr_orContext(ExprContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14Expr_orContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14Expr_orContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14Expr_orContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14Expr_orContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 5991, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 5991, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 5991, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5991, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_orContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14Expr_orContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14Expr_orContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":5992
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 5992, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5992, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5992, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5992, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5992, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5992, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5992, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5992, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5992, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5993
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def expr(self, i:int=None):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5993, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5993, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5993, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5993, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5993, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5993, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5991
 *     class Expr_orContext(ExprContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_orContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":5995
 *             self.copyFrom(ctx)
 * 
 *         def expr(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14Expr_orContext_3expr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14Expr_orContext_3expr = {"expr", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14Expr_orContext_3expr, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14Expr_orContext_3expr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("expr (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "expr") < 0)) __PYX_ERR(0, 5995, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("expr", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5995, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_orContext.expr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14Expr_orContext_2expr(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14Expr_orContext_2expr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("expr", 0);

  /* "pymola/generated/ModelicaParser.pyx":5996
 * 
 *         def expr(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":5997
 *         def expr(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.ExprContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5997, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5997, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_ExprContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5997, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5997, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5997, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5997, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5997, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5997, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":5996
 * 
 *         def expr(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":5999
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.ExprContext,i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5999, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5999, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_ExprContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5999, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5999, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5999, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5999, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5999, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":5995
 *             self.copyFrom(ctx)
 * 
 *         def expr(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_orContext.expr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6002
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpr_or" ):
 *                 listener.enterExpr_or(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14Expr_orContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14Expr_orContext_5enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14Expr_orContext_5enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14Expr_orContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 6002, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 6002, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6002, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_orContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14Expr_orContext_4enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14Expr_orContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6003
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExpr_or" ):             # <<<<<<<<<<<<<<
 *                 listener.enterExpr_or(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterExpr_or); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6003, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6004
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExpr_or" ):
 *                 listener.enterExpr_or(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterExpr_or); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6004, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6004, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6004, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6004, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6004, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6004, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6003
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExpr_or" ):             # <<<<<<<<<<<<<<
 *                 listener.enterExpr_or(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6002
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpr_or" ):
 *                 listener.enterExpr_or(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_orContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6006
 *                 listener.enterExpr_or(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpr_or" ):
 *                 listener.exitExpr_or(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14Expr_orContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14Expr_orContext_7exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14Expr_orContext_7exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14Expr_orContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 6006, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 6006, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6006, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_orContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14Expr_orContext_6exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14Expr_orContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6007
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExpr_or" ):             # <<<<<<<<<<<<<<
 *                 listener.exitExpr_or(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitExpr_or); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6007, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6008
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExpr_or" ):
 *                 listener.exitExpr_or(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitExpr_or); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6008, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6008, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6008, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6008, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6008, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6008, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6007
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExpr_or" ):             # <<<<<<<<<<<<<<
 *                 listener.exitExpr_or(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6006
 *                 listener.enterExpr_or(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpr_or" ):
 *                 listener.exitExpr_or(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_orContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6010
 *                 listener.exitExpr_or(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpr_or" ):
 *                 return visitor.visitExpr_or(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14Expr_orContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14Expr_orContext_9accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14Expr_orContext_9accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14Expr_orContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 6010, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 6010, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6010, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_orContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14Expr_orContext_8accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14Expr_orContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":6011
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExpr_or" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitExpr_or(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitExpr_or); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6011, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6012
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExpr_or" ):
 *                 return visitor.visitExpr_or(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitExpr_or); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6012, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6012, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6012, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6012, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6012, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6012, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":6011
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExpr_or" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitExpr_or(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6014
 *                 return visitor.visitExpr_or(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6014, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6014, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6014, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6014, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6014, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6014, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":6010
 *                 listener.exitExpr_or(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpr_or" ):
 *                 return visitor.visitExpr_or(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_orContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6019
 *     class Expr_primaryContext(ExprContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Expr_primaryContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19Expr_primaryContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19Expr_primaryContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Expr_primaryContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 6019, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 6019, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 6019, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6019, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_primaryContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19Expr_primaryContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Expr_primaryContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":6020
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 6020, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6020, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6020, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6020, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6020, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6020, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6020, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6020, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6020, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6021
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def primary(self):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6021, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6021, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6021, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6021, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6021, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6021, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6019
 *     class Expr_primaryContext(ExprContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_primaryContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6023
 *             self.copyFrom(ctx)
 * 
 *         def primary(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.PrimaryContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Expr_primaryContext_3primary(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19Expr_primaryContext_3primary = {"primary", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19Expr_primaryContext_3primary, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Expr_primaryContext_3primary(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("primary (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19Expr_primaryContext_2primary(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Expr_primaryContext_2primary(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("primary", 0);

  /* "pymola/generated/ModelicaParser.pyx":6024
 * 
 *         def primary(self):
 *             return self.getTypedRuleContext(ModelicaParser.PrimaryContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6024, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6024, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_PrimaryContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6024, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6024, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6024, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6024, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6024, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":6023
 *             self.copyFrom(ctx)
 * 
 *         def primary(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.PrimaryContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_primaryContext.primary", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6027
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpr_primary" ):
 *                 listener.enterExpr_primary(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Expr_primaryContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19Expr_primaryContext_5enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19Expr_primaryContext_5enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Expr_primaryContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 6027, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 6027, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6027, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_primaryContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19Expr_primaryContext_4enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Expr_primaryContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6028
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExpr_primary" ):             # <<<<<<<<<<<<<<
 *                 listener.enterExpr_primary(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterExpr_primary); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6028, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6029
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExpr_primary" ):
 *                 listener.enterExpr_primary(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterExpr_primary); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6029, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6029, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6029, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6029, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6029, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6029, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6028
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExpr_primary" ):             # <<<<<<<<<<<<<<
 *                 listener.enterExpr_primary(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6027
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpr_primary" ):
 *                 listener.enterExpr_primary(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_primaryContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6031
 *                 listener.enterExpr_primary(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpr_primary" ):
 *                 listener.exitExpr_primary(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Expr_primaryContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19Expr_primaryContext_7exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19Expr_primaryContext_7exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Expr_primaryContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 6031, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 6031, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6031, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_primaryContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19Expr_primaryContext_6exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Expr_primaryContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6032
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExpr_primary" ):             # <<<<<<<<<<<<<<
 *                 listener.exitExpr_primary(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitExpr_primary); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6032, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6033
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExpr_primary" ):
 *                 listener.exitExpr_primary(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitExpr_primary); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6033, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6033, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6033, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6033, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6033, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6033, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6032
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExpr_primary" ):             # <<<<<<<<<<<<<<
 *                 listener.exitExpr_primary(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6031
 *                 listener.enterExpr_primary(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpr_primary" ):
 *                 listener.exitExpr_primary(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_primaryContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6035
 *                 listener.exitExpr_primary(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpr_primary" ):
 *                 return visitor.visitExpr_primary(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Expr_primaryContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19Expr_primaryContext_9accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19Expr_primaryContext_9accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Expr_primaryContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 6035, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 6035, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6035, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_primaryContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19Expr_primaryContext_8accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Expr_primaryContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":6036
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExpr_primary" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitExpr_primary(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitExpr_primary); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6036, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6037
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExpr_primary" ):
 *                 return visitor.visitExpr_primary(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitExpr_primary); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6037, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6037, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6037, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6037, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6037, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6037, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":6036
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExpr_primary" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitExpr_primary(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6039
 *                 return visitor.visitExpr_primary(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6039, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6039, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6039, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6039, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6039, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6039, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":6035
 *                 listener.exitExpr_primary(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpr_primary" ):
 *                 return visitor.visitExpr_primary(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_primaryContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6044
 *     class Expr_andContext(ExprContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_andContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_andContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_andContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_andContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 6044, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 6044, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 6044, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6044, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_andContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_andContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_andContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":6045
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 6045, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6045, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6045, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6045, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6045, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6045, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6045, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6045, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6045, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6046
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def expr(self, i:int=None):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6046, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6046, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6046, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6046, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6046, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6046, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6044
 *     class Expr_andContext(ExprContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_andContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6048
 *             self.copyFrom(ctx)
 * 
 *         def expr(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_andContext_3expr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_andContext_3expr = {"expr", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_andContext_3expr, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_andContext_3expr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("expr (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "expr") < 0)) __PYX_ERR(0, 6048, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("expr", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6048, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_andContext.expr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_andContext_2expr(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_andContext_2expr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("expr", 0);

  /* "pymola/generated/ModelicaParser.pyx":6049
 * 
 *         def expr(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6050
 *         def expr(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.ExprContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6050, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6050, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_ExprContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6050, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6050, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6050, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6050, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6050, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6050, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":6049
 * 
 *         def expr(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6052
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.ExprContext,i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6052, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6052, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_ExprContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6052, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6052, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6052, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6052, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6052, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":6048
 *             self.copyFrom(ctx)
 * 
 *         def expr(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_andContext.expr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6055
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpr_and" ):
 *                 listener.enterExpr_and(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_andContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_andContext_5enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_andContext_5enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_andContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 6055, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 6055, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6055, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_andContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_andContext_4enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_andContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6056
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExpr_and" ):             # <<<<<<<<<<<<<<
 *                 listener.enterExpr_and(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterExpr_and); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6056, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6057
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExpr_and" ):
 *                 listener.enterExpr_and(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterExpr_and); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6057, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6057, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6057, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6057, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6057, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6057, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6056
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExpr_and" ):             # <<<<<<<<<<<<<<
 *                 listener.enterExpr_and(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6055
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpr_and" ):
 *                 listener.enterExpr_and(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_andContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6059
 *                 listener.enterExpr_and(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpr_and" ):
 *                 listener.exitExpr_and(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_andContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_andContext_7exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_andContext_7exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_andContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 6059, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 6059, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6059, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_andContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_andContext_6exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_andContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6060
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExpr_and" ):             # <<<<<<<<<<<<<<
 *                 listener.exitExpr_and(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitExpr_and); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6060, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6061
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExpr_and" ):
 *                 listener.exitExpr_and(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitExpr_and); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6061, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6061, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6061, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6061, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6061, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6061, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6060
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExpr_and" ):             # <<<<<<<<<<<<<<
 *                 listener.exitExpr_and(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6059
 *                 listener.enterExpr_and(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpr_and" ):
 *                 listener.exitExpr_and(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_andContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6063
 *                 listener.exitExpr_and(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpr_and" ):
 *                 return visitor.visitExpr_and(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_andContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_andContext_9accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_andContext_9accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_andContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 6063, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 6063, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6063, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_andContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_andContext_8accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_andContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":6064
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExpr_and" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitExpr_and(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitExpr_and); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6064, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6065
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExpr_and" ):
 *                 return visitor.visitExpr_and(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitExpr_and); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6065, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6065, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6065, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6065, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6065, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6065, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":6064
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExpr_and" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitExpr_and(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6067
 *                 return visitor.visitExpr_and(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6067, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6067, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6067, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6067, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6067, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6067, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":6063
 *                 listener.exitExpr_and(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpr_and" ):
 *                 return visitor.visitExpr_and(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_andContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6072
 *     class Expr_relContext(ExprContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.op = None # Token
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_relContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_relContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_relContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_relContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 6072, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 6072, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 6072, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6072, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_relContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_relContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_relContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":6073
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.op = None # Token
 *             self.copyFrom(ctx)
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 6073, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6073, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6073, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6073, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6073, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6073, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6073, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6073, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6073, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6074
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext
 *             super().__init__(parser)
 *             self.op = None # Token             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_op, Py_None) < 0) __PYX_ERR(0, 6074, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6075
 *             super().__init__(parser)
 *             self.op = None # Token
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def expr(self, i:int=None):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6075, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6075, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6075, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6075, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6075, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6075, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6072
 *     class Expr_relContext(ExprContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.op = None # Token
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_relContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6077
 *             self.copyFrom(ctx)
 * 
 *         def expr(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_relContext_3expr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_relContext_3expr = {"expr", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_relContext_3expr, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_relContext_3expr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("expr (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "expr") < 0)) __PYX_ERR(0, 6077, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("expr", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6077, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_relContext.expr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_relContext_2expr(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_relContext_2expr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("expr", 0);

  /* "pymola/generated/ModelicaParser.pyx":6078
 * 
 *         def expr(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6079
 *         def expr(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.ExprContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6079, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6079, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_ExprContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6079, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6079, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6079, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6079, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6079, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6079, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":6078
 * 
 *         def expr(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6081
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.ExprContext,i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6081, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6081, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_ExprContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6081, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6081, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6081, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6081, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6081, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":6077
 *             self.copyFrom(ctx)
 * 
 *         def expr(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_relContext.expr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6084
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpr_rel" ):
 *                 listener.enterExpr_rel(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_relContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_relContext_5enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_relContext_5enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_relContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 6084, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 6084, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6084, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_relContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_relContext_4enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_relContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6085
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExpr_rel" ):             # <<<<<<<<<<<<<<
 *                 listener.enterExpr_rel(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterExpr_rel); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6085, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6086
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExpr_rel" ):
 *                 listener.enterExpr_rel(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterExpr_rel); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6086, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6086, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6086, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6086, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6086, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6086, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6085
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExpr_rel" ):             # <<<<<<<<<<<<<<
 *                 listener.enterExpr_rel(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6084
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpr_rel" ):
 *                 listener.enterExpr_rel(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_relContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6088
 *                 listener.enterExpr_rel(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpr_rel" ):
 *                 listener.exitExpr_rel(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_relContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_relContext_7exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_relContext_7exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_relContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 6088, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 6088, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6088, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_relContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_relContext_6exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_relContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6089
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExpr_rel" ):             # <<<<<<<<<<<<<<
 *                 listener.exitExpr_rel(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitExpr_rel); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6089, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6090
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExpr_rel" ):
 *                 listener.exitExpr_rel(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitExpr_rel); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6090, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6090, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6090, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6090, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6090, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6090, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6089
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExpr_rel" ):             # <<<<<<<<<<<<<<
 *                 listener.exitExpr_rel(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6088
 *                 listener.enterExpr_rel(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpr_rel" ):
 *                 listener.exitExpr_rel(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_relContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6092
 *                 listener.exitExpr_rel(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpr_rel" ):
 *                 return visitor.visitExpr_rel(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_relContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_relContext_9accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_relContext_9accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_relContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 6092, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 6092, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6092, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_relContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_relContext_8accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_relContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":6093
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExpr_rel" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitExpr_rel(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitExpr_rel); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6093, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6094
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExpr_rel" ):
 *                 return visitor.visitExpr_rel(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitExpr_rel); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6094, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6094, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6094, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6094, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6094, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6094, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":6093
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExpr_rel" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitExpr_rel(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6096
 *                 return visitor.visitExpr_rel(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6096, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6096, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6096, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6096, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6096, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6096, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":6092
 *                 listener.exitExpr_rel(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpr_rel" ):
 *                 return visitor.visitExpr_rel(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_relContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6101
 *     class Expr_notContext(ExprContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_notContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_notContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_notContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_notContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 6101, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 6101, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 6101, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6101, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_notContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_notContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_notContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":6102
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 6102, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6102, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6102, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6102, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6102, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6103
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def expr(self):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6103, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6103, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6103, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6103, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6103, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6101
 *     class Expr_notContext(ExprContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_notContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6105
 *             self.copyFrom(ctx)
 * 
 *         def expr(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExprContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_notContext_3expr(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_notContext_3expr = {"expr", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_notContext_3expr, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_notContext_3expr(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("expr (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_notContext_2expr(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_notContext_2expr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("expr", 0);

  /* "pymola/generated/ModelicaParser.pyx":6106
 * 
 *         def expr(self):
 *             return self.getTypedRuleContext(ModelicaParser.ExprContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_ExprContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6106, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6106, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":6105
 *             self.copyFrom(ctx)
 * 
 *         def expr(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExprContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_notContext.expr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6109
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpr_not" ):
 *                 listener.enterExpr_not(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_notContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_notContext_5enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_notContext_5enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_notContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 6109, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 6109, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6109, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_notContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_notContext_4enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_notContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6110
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExpr_not" ):             # <<<<<<<<<<<<<<
 *                 listener.enterExpr_not(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterExpr_not); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6110, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6111
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExpr_not" ):
 *                 listener.enterExpr_not(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterExpr_not); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6111, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6111, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6111, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6111, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6111, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6111, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6110
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExpr_not" ):             # <<<<<<<<<<<<<<
 *                 listener.enterExpr_not(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6109
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpr_not" ):
 *                 listener.enterExpr_not(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_notContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6113
 *                 listener.enterExpr_not(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpr_not" ):
 *                 listener.exitExpr_not(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_notContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_notContext_7exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_notContext_7exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_notContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 6113, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 6113, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6113, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_notContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_notContext_6exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_notContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6114
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExpr_not" ):             # <<<<<<<<<<<<<<
 *                 listener.exitExpr_not(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitExpr_not); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6114, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6115
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExpr_not" ):
 *                 listener.exitExpr_not(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitExpr_not); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6115, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6115, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6115, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6115, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6115, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6114
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExpr_not" ):             # <<<<<<<<<<<<<<
 *                 listener.exitExpr_not(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6113
 *                 listener.enterExpr_not(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpr_not" ):
 *                 listener.exitExpr_not(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_notContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6117
 *                 listener.exitExpr_not(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpr_not" ):
 *                 return visitor.visitExpr_not(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_notContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_notContext_9accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_notContext_9accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_notContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 6117, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 6117, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6117, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_notContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_notContext_8accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_notContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":6118
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExpr_not" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitExpr_not(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitExpr_not); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6118, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6119
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExpr_not" ):
 *                 return visitor.visitExpr_not(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitExpr_not); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6119, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6119, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6119, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6119, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6119, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":6118
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExpr_not" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitExpr_not(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6121
 *                 return visitor.visitExpr_not(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6121, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6121, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6121, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6121, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6121, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":6117
 *                 listener.exitExpr_not(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpr_not" ):
 *                 return visitor.visitExpr_not(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_notContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6126
 *     class Expr_mulContext(ExprContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.op = None # Token
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_mulContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_mulContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_mulContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_mulContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 6126, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 6126, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 6126, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6126, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_mulContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_mulContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_mulContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":6127
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.op = None # Token
 *             self.copyFrom(ctx)
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 6127, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6127, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6127, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6127, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6127, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6128
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext
 *             super().__init__(parser)
 *             self.op = None # Token             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_op, Py_None) < 0) __PYX_ERR(0, 6128, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6129
 *             super().__init__(parser)
 *             self.op = None # Token
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def expr(self, i:int=None):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6129, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6129, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6129, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6129, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6129, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6126
 *     class Expr_mulContext(ExprContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.op = None # Token
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_mulContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6131
 *             self.copyFrom(ctx)
 * 
 *         def expr(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_mulContext_3expr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_mulContext_3expr = {"expr", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_mulContext_3expr, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_mulContext_3expr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("expr (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "expr") < 0)) __PYX_ERR(0, 6131, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("expr", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6131, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_mulContext.expr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_mulContext_2expr(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_mulContext_2expr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("expr", 0);

  /* "pymola/generated/ModelicaParser.pyx":6132
 * 
 *         def expr(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6133
 *         def expr(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.ExprContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6133, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6133, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_ExprContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6133, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6133, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6133, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6133, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6133, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6133, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":6132
 * 
 *         def expr(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6135
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.ExprContext,i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_ExprContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6135, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6135, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6135, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6135, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":6131
 *             self.copyFrom(ctx)
 * 
 *         def expr(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_mulContext.expr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6138
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpr_mul" ):
 *                 listener.enterExpr_mul(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_mulContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_mulContext_5enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_mulContext_5enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_mulContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 6138, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 6138, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6138, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_mulContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_mulContext_4enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_mulContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6139
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExpr_mul" ):             # <<<<<<<<<<<<<<
 *                 listener.enterExpr_mul(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterExpr_mul); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6139, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6140
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExpr_mul" ):
 *                 listener.enterExpr_mul(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterExpr_mul); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6140, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6140, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6140, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6140, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6140, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6139
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExpr_mul" ):             # <<<<<<<<<<<<<<
 *                 listener.enterExpr_mul(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6138
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpr_mul" ):
 *                 listener.enterExpr_mul(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_mulContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6142
 *                 listener.enterExpr_mul(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpr_mul" ):
 *                 listener.exitExpr_mul(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_mulContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_mulContext_7exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_mulContext_7exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_mulContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 6142, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 6142, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6142, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_mulContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_mulContext_6exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_mulContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6143
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExpr_mul" ):             # <<<<<<<<<<<<<<
 *                 listener.exitExpr_mul(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitExpr_mul); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6143, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6144
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExpr_mul" ):
 *                 listener.exitExpr_mul(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitExpr_mul); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6144, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6144, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6144, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6144, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6144, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6143
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExpr_mul" ):             # <<<<<<<<<<<<<<
 *                 listener.exitExpr_mul(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6142
 *                 listener.enterExpr_mul(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpr_mul" ):
 *                 listener.exitExpr_mul(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_mulContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6146
 *                 listener.exitExpr_mul(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpr_mul" ):
 *                 return visitor.visitExpr_mul(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_mulContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_mulContext_9accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_mulContext_9accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_15Expr_mulContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 6146, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 6146, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6146, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_mulContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_mulContext_8accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_15Expr_mulContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":6147
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExpr_mul" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitExpr_mul(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitExpr_mul); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6147, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6148
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExpr_mul" ):
 *                 return visitor.visitExpr_mul(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitExpr_mul); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6148, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6148, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6148, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6148, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6148, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":6147
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExpr_mul" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitExpr_mul(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6150
 *                 return visitor.visitExpr_mul(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6150, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6150, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6150, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6150, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6150, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":6146
 *                 listener.exitExpr_mul(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpr_mul" ):
 *                 return visitor.visitExpr_mul(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expr_mulContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6154
 * 
 * 
 *     def expr(self, _p:int=0):             # <<<<<<<<<<<<<<
 *         _parentctx = self._ctx
 *         _parentState = self.state
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_115expr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_115expr = {"expr", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_115expr, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_115expr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v__p = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("expr (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_p,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)__pyx_int_0));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_p);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "expr") < 0)) __PYX_ERR(0, 6154, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v__p = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("expr", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6154, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.expr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_114expr(__pyx_self, __pyx_v_self, __pyx_v__p);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_114expr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v__p) {
  PyObject *__pyx_v__parentctx = NULL;
  PyObject *__pyx_v__parentState = NULL;
  PyObject *__pyx_v_localctx = NULL;
  CYTHON_UNUSED PyObject *__pyx_v__prevctx = NULL;
  long __pyx_v__startState;
  PyObject *__pyx_v_la_ = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v__alt = NULL;
  PyObject *__pyx_v_FailedPredicateException = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  char const *__pyx_t_22;
  __Pyx_RefNannySetupContext("expr", 0);

  /* "pymola/generated/ModelicaParser.pyx":6155
 * 
 *     def expr(self, _p:int=0):
 *         _parentctx = self._ctx             # <<<<<<<<<<<<<<
 *         _parentState = self.state
 *         localctx = ModelicaParser.ExprContext(self, self._ctx, _parentState)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__parentctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6156
 *     def expr(self, _p:int=0):
 *         _parentctx = self._ctx
 *         _parentState = self.state             # <<<<<<<<<<<<<<
 *         localctx = ModelicaParser.ExprContext(self, self._ctx, _parentState)
 *         _prevctx = localctx
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__parentState = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6157
 *         _parentctx = self._ctx
 *         _parentState = self.state
 *         localctx = ModelicaParser.ExprContext(self, self._ctx, _parentState)             # <<<<<<<<<<<<<<
 *         _prevctx = localctx
 *         _startState = 112
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_ExprContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_self, __pyx_t_2, __pyx_v__parentState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6157, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_self, __pyx_t_2, __pyx_v__parentState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6157, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6157, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_t_2);
    __Pyx_INCREF(__pyx_v__parentState);
    __Pyx_GIVEREF(__pyx_v__parentState);
    PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_5, __pyx_v__parentState);
    __pyx_t_2 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6157, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6158
 *         _parentState = self.state
 *         localctx = ModelicaParser.ExprContext(self, self._ctx, _parentState)
 *         _prevctx = localctx             # <<<<<<<<<<<<<<
 *         _startState = 112
 *         self.enterRecursionRule(localctx, 112, self.RULE_expr, _p)
 */
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_v__prevctx = __pyx_v_localctx;

  /* "pymola/generated/ModelicaParser.pyx":6159
 *         localctx = ModelicaParser.ExprContext(self, self._ctx, _parentState)
 *         _prevctx = localctx
 *         _startState = 112             # <<<<<<<<<<<<<<
 *         self.enterRecursionRule(localctx, 112, self.RULE_expr, _p)
 *         self._la = 0 # Token type
 */
  __pyx_v__startState = 0x70;

  /* "pymola/generated/ModelicaParser.pyx":6160
 *         _prevctx = localctx
 *         _startState = 112
 *         self.enterRecursionRule(localctx, 112, self.RULE_expr, _p)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRecursionRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_expr); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_2 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[5] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_112, __pyx_t_6, __pyx_v__p};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 4+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6160, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[5] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_112, __pyx_t_6, __pyx_v__p};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 4+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6160, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else
  #endif
  {
    __pyx_t_4 = PyTuple_New(4+__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6160, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_5, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_112);
    __Pyx_GIVEREF(__pyx_int_112);
    PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_5, __pyx_int_112);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_4, 2+__pyx_t_5, __pyx_t_6);
    __Pyx_INCREF(__pyx_v__p);
    __Pyx_GIVEREF(__pyx_v__p);
    PyTuple_SET_ITEM(__pyx_t_4, 3+__pyx_t_5, __pyx_v__p);
    __pyx_t_6 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6160, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6161
 *         _startState = 112
 *         self.enterRecursionRule(localctx, 112, self.RULE_expr, _p)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 6161, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6162
 *         self.enterRecursionRule(localctx, 112, self.RULE_expr, _p)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 862
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":6163
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 862
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6163, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6163, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6163, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6163, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6163, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":6164
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 862             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             la_ = self._interp.adaptivePredict(self._input,109,self._ctx)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_862) < 0) __PYX_ERR(0, 6164, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":6165
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 862
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             la_ = self._interp.adaptivePredict(self._input,109,self._ctx)
 *             if la_ == 1:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6165, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6165, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6165, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_6)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6165, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6165, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6165, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6165, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":6166
 *             self.state = 862
 *             self._errHandler.sync(self)
 *             la_ = self._interp.adaptivePredict(self._input,109,self._ctx)             # <<<<<<<<<<<<<<
 *             if la_ == 1:
 *                 localctx = ModelicaParser.Expr_negContext(self, localctx)
 */
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_interp); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6166, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_adaptivePredict); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6166, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6166, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6166, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_5 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_5 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[4] = {__pyx_t_2, __pyx_t_6, __pyx_int_109, __pyx_t_3};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6166, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[4] = {__pyx_t_2, __pyx_t_6, __pyx_int_109, __pyx_t_3};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6166, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_10 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6166, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_10);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_6);
          PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_5, __pyx_t_6);
          __Pyx_INCREF(__pyx_int_109);
          __Pyx_GIVEREF(__pyx_int_109);
          PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_5, __pyx_int_109);
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_10, 2+__pyx_t_5, __pyx_t_3);
          __pyx_t_6 = 0;
          __pyx_t_3 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6166, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_v_la_ = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":6167
 *             self._errHandler.sync(self)
 *             la_ = self._interp.adaptivePredict(self._input,109,self._ctx)
 *             if la_ == 1:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Expr_negContext(self, localctx)
 *                 self._ctx = localctx
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6167, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 6167, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":6168
 *             la_ = self._interp.adaptivePredict(self._input,109,self._ctx)
 *             if la_ == 1:
 *                 localctx = ModelicaParser.Expr_negContext(self, localctx)             # <<<<<<<<<<<<<<
 *                 self._ctx = localctx
 *                 _prevctx = localctx
 */
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6168, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_Expr_negContext); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6168, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          __pyx_t_5 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_10);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_10, function);
              __pyx_t_5 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_10)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6168, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6168, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6168, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_4) {
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __pyx_t_4 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_5, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_5, __pyx_v_localctx);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6168, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6169
 *             if la_ == 1:
 *                 localctx = ModelicaParser.Expr_negContext(self, localctx)
 *                 self._ctx = localctx             # <<<<<<<<<<<<<<
 *                 _prevctx = localctx
 * 
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_ctx, __pyx_v_localctx) < 0) __PYX_ERR(0, 6169, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6170
 *                 localctx = ModelicaParser.Expr_negContext(self, localctx)
 *                 self._ctx = localctx
 *                 _prevctx = localctx             # <<<<<<<<<<<<<<
 * 
 *                 self.state = 853
 */
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_DECREF_SET(__pyx_v__prevctx, __pyx_v_localctx);

          /* "pymola/generated/ModelicaParser.pyx":6172
 *                 _prevctx = localctx
 * 
 *                 self.state = 853             # <<<<<<<<<<<<<<
 *                 localctx.op = self.match(ModelicaParser.T__55)
 *                 self.state = 854
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_853) < 0) __PYX_ERR(0, 6172, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6173
 * 
 *                 self.state = 853
 *                 localctx.op = self.match(ModelicaParser.T__55)             # <<<<<<<<<<<<<<
 *                 self.state = 854
 *                 self.expr(9)
 */
          __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6173, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6173, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__55); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6173, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_10);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_10, function);
            }
          }
          if (!__pyx_t_3) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6173, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_10)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_4};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6173, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_4};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6173, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            } else
            #endif
            {
              __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6173, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
              __Pyx_GIVEREF(__pyx_t_4);
              PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_4);
              __pyx_t_4 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6173, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_op, __pyx_t_1) < 0) __PYX_ERR(0, 6173, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6174
 *                 self.state = 853
 *                 localctx.op = self.match(ModelicaParser.T__55)
 *                 self.state = 854             # <<<<<<<<<<<<<<
 *                 self.expr(9)
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_854) < 0) __PYX_ERR(0, 6174, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6175
 *                 localctx.op = self.match(ModelicaParser.T__55)
 *                 self.state = 854
 *                 self.expr(9)             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6175, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__934, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6175, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6167
 *             self._errHandler.sync(self)
 *             la_ = self._interp.adaptivePredict(self._input,109,self._ctx)
 *             if la_ == 1:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Expr_negContext(self, localctx)
 *                 self._ctx = localctx
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":6178
 *                 pass
 * 
 *             elif la_ == 2:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Expr_expContext(self, localctx)
 *                 self._ctx = localctx
 */
        __pyx_t_10 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6178, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 6178, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":6179
 * 
 *             elif la_ == 2:
 *                 localctx = ModelicaParser.Expr_expContext(self, localctx)             # <<<<<<<<<<<<<<
 *                 self._ctx = localctx
 *                 _prevctx = localctx
 */
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6179, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_Expr_expContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6179, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = NULL;
          __pyx_t_5 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_6, function);
              __pyx_t_5 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_6)) {
            PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_10 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6179, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_10);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
            PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_10 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6179, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_10);
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6179, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (__pyx_t_1) {
              __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1); __pyx_t_1 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_5, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_5, __pyx_v_localctx);
            __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_4, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6179, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_10);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_10);
          __pyx_t_10 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6180
 *             elif la_ == 2:
 *                 localctx = ModelicaParser.Expr_expContext(self, localctx)
 *                 self._ctx = localctx             # <<<<<<<<<<<<<<
 *                 _prevctx = localctx
 *                 self.state = 855
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_ctx, __pyx_v_localctx) < 0) __PYX_ERR(0, 6180, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6181
 *                 localctx = ModelicaParser.Expr_expContext(self, localctx)
 *                 self._ctx = localctx
 *                 _prevctx = localctx             # <<<<<<<<<<<<<<
 *                 self.state = 855
 *                 self.primary()
 */
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_DECREF_SET(__pyx_v__prevctx, __pyx_v_localctx);

          /* "pymola/generated/ModelicaParser.pyx":6182
 *                 self._ctx = localctx
 *                 _prevctx = localctx
 *                 self.state = 855             # <<<<<<<<<<<<<<
 *                 self.primary()
 *                 self.state = 856
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_855) < 0) __PYX_ERR(0, 6182, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6183
 *                 _prevctx = localctx
 *                 self.state = 855
 *                 self.primary()             # <<<<<<<<<<<<<<
 *                 self.state = 856
 *                 localctx.op = self._input.LT(1)
 */
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_primary); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6183, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_4 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_6, function);
            }
          }
          if (__pyx_t_4) {
            __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_4); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6183, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          } else {
            __pyx_t_10 = __Pyx_PyObject_CallNoArg(__pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6183, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6184
 *                 self.state = 855
 *                 self.primary()
 *                 self.state = 856             # <<<<<<<<<<<<<<
 *                 localctx.op = self._input.LT(1)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_856) < 0) __PYX_ERR(0, 6184, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6185
 *                 self.primary()
 *                 self.state = 856
 *                 localctx.op = self._input.LT(1)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 *                 if not(_la==ModelicaParser.T__56 or _la==ModelicaParser.T__57):
 */
          __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6185, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_LT); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6185, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_tuple__935, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6185, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_op, __pyx_t_10) < 0) __PYX_ERR(0, 6185, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6186
 *                 self.state = 856
 *                 localctx.op = self._input.LT(1)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if not(_la==ModelicaParser.T__56 or _la==ModelicaParser.T__57):
 *                     localctx.op = self._errHandler.recoverInline(self)
 */
          __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6186, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_LA); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6186, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_tuple__936, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6186, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_v__la = __pyx_t_10;
          __pyx_t_10 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6187
 *                 localctx.op = self._input.LT(1)
 *                 _la = self._input.LA(1)
 *                 if not(_la==ModelicaParser.T__56 or _la==ModelicaParser.T__57):             # <<<<<<<<<<<<<<
 *                     localctx.op = self._errHandler.recoverInline(self)
 *                 else:
 */
          __pyx_t_10 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6187, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_T__56); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6187, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_t_10 = PyObject_RichCompare(__pyx_v__la, __pyx_t_6, Py_EQ); __Pyx_XGOTREF(__pyx_t_10); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6187, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 6187, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          if (!__pyx_t_12) {
          } else {
            __pyx_t_11 = __pyx_t_12;
            goto __pyx_L16_bool_binop_done;
          }
          __pyx_t_10 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6187, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_T__57); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6187, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_t_10 = PyObject_RichCompare(__pyx_v__la, __pyx_t_6, Py_EQ); __Pyx_XGOTREF(__pyx_t_10); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6187, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 6187, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_t_11 = __pyx_t_12;
          __pyx_L16_bool_binop_done:;
          __pyx_t_12 = ((!__pyx_t_11) != 0);
          if (__pyx_t_12) {

            /* "pymola/generated/ModelicaParser.pyx":6188
 *                 _la = self._input.LA(1)
 *                 if not(_la==ModelicaParser.T__56 or _la==ModelicaParser.T__57):
 *                     localctx.op = self._errHandler.recoverInline(self)             # <<<<<<<<<<<<<<
 *                 else:
 *                     self._errHandler.reportMatch(self)
 */
            __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6188, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_6);
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_recoverInline); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6188, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __pyx_t_6 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
              __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
              if (likely(__pyx_t_6)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                __Pyx_INCREF(__pyx_t_6);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_4, function);
              }
            }
            if (!__pyx_t_6) {
              __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6188, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_10);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_4)) {
                PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
                __pyx_t_10 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6188, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
                __Pyx_GOTREF(__pyx_t_10);
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
                PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
                __pyx_t_10 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6188, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
                __Pyx_GOTREF(__pyx_t_10);
              } else
              #endif
              {
                __pyx_t_1 = PyTuple_New(1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6188, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_6); __pyx_t_6 = NULL;
                __Pyx_INCREF(__pyx_v_self);
                __Pyx_GIVEREF(__pyx_v_self);
                PyTuple_SET_ITEM(__pyx_t_1, 0+1, __pyx_v_self);
                __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_1, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6188, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_10);
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_op, __pyx_t_10) < 0) __PYX_ERR(0, 6188, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

            /* "pymola/generated/ModelicaParser.pyx":6187
 *                 localctx.op = self._input.LT(1)
 *                 _la = self._input.LA(1)
 *                 if not(_la==ModelicaParser.T__56 or _la==ModelicaParser.T__57):             # <<<<<<<<<<<<<<
 *                     localctx.op = self._errHandler.recoverInline(self)
 *                 else:
 */
            goto __pyx_L15;
          }

          /* "pymola/generated/ModelicaParser.pyx":6190
 *                     localctx.op = self._errHandler.recoverInline(self)
 *                 else:
 *                     self._errHandler.reportMatch(self)             # <<<<<<<<<<<<<<
 *                     self.consume()
 *                 self.state = 857
 */
          /*else*/ {
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6190, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_reportMatch); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6190, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
              __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
              if (likely(__pyx_t_4)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
                __Pyx_INCREF(__pyx_t_4);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_1, function);
              }
            }
            if (!__pyx_t_4) {
              __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_self); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6190, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_10);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_1)) {
                PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
                __pyx_t_10 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6190, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_GOTREF(__pyx_t_10);
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
                PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
                __pyx_t_10 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6190, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_GOTREF(__pyx_t_10);
              } else
              #endif
              {
                __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6190, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
                __Pyx_INCREF(__pyx_v_self);
                __Pyx_GIVEREF(__pyx_v_self);
                PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
                __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_6, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6190, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_10);
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

            /* "pymola/generated/ModelicaParser.pyx":6191
 *                 else:
 *                     self._errHandler.reportMatch(self)
 *                     self.consume()             # <<<<<<<<<<<<<<
 *                 self.state = 857
 *                 self.primary()
 */
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_consume); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6191, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_6 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
              __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
              if (likely(__pyx_t_6)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
                __Pyx_INCREF(__pyx_t_6);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_1, function);
              }
            }
            if (__pyx_t_6) {
              __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6191, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            } else {
              __pyx_t_10 = __Pyx_PyObject_CallNoArg(__pyx_t_1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6191, __pyx_L6_error)
            }
            __Pyx_GOTREF(__pyx_t_10);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          }
          __pyx_L15:;

          /* "pymola/generated/ModelicaParser.pyx":6192
 *                     self._errHandler.reportMatch(self)
 *                     self.consume()
 *                 self.state = 857             # <<<<<<<<<<<<<<
 *                 self.primary()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_857) < 0) __PYX_ERR(0, 6192, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6193
 *                     self.consume()
 *                 self.state = 857
 *                 self.primary()             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_primary); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6193, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_6 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
            __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
            if (likely(__pyx_t_6)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
              __Pyx_INCREF(__pyx_t_6);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_1, function);
            }
          }
          if (__pyx_t_6) {
            __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6193, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          } else {
            __pyx_t_10 = __Pyx_PyObject_CallNoArg(__pyx_t_1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6193, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6178
 *                 pass
 * 
 *             elif la_ == 2:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Expr_expContext(self, localctx)
 *                 self._ctx = localctx
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":6196
 *                 pass
 * 
 *             elif la_ == 3:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Expr_notContext(self, localctx)
 *                 self._ctx = localctx
 */
        __pyx_t_10 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_3, 3, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6196, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 6196, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (__pyx_t_12) {

          /* "pymola/generated/ModelicaParser.pyx":6197
 * 
 *             elif la_ == 3:
 *                 localctx = ModelicaParser.Expr_notContext(self, localctx)             # <<<<<<<<<<<<<<
 *                 self._ctx = localctx
 *                 _prevctx = localctx
 */
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6197, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_Expr_notContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6197, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = NULL;
          __pyx_t_5 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_6, function);
              __pyx_t_5 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_6)) {
            PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_10 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6197, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_10);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
            PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_10 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6197, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_10);
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6197, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (__pyx_t_1) {
              __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1); __pyx_t_1 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_5, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_5, __pyx_v_localctx);
            __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_4, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6197, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_10);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_10);
          __pyx_t_10 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6198
 *             elif la_ == 3:
 *                 localctx = ModelicaParser.Expr_notContext(self, localctx)
 *                 self._ctx = localctx             # <<<<<<<<<<<<<<
 *                 _prevctx = localctx
 *                 self.state = 859
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_ctx, __pyx_v_localctx) < 0) __PYX_ERR(0, 6198, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6199
 *                 localctx = ModelicaParser.Expr_notContext(self, localctx)
 *                 self._ctx = localctx
 *                 _prevctx = localctx             # <<<<<<<<<<<<<<
 *                 self.state = 859
 *                 self.match(ModelicaParser.T__70)
 */
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_DECREF_SET(__pyx_v__prevctx, __pyx_v_localctx);

          /* "pymola/generated/ModelicaParser.pyx":6200
 *                 self._ctx = localctx
 *                 _prevctx = localctx
 *                 self.state = 859             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__70)
 *                 self.state = 860
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_859) < 0) __PYX_ERR(0, 6200, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6201
 *                 _prevctx = localctx
 *                 self.state = 859
 *                 self.match(ModelicaParser.T__70)             # <<<<<<<<<<<<<<
 *                 self.state = 860
 *                 self.expr(4)
 */
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6201, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6201, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__70); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6201, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_6, function);
            }
          }
          if (!__pyx_t_4) {
            __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6201, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_10);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_6)) {
              PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_1};
              __pyx_t_10 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6201, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_10);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
              PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_1};
              __pyx_t_10 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6201, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_10);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6201, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __pyx_t_4 = NULL;
              __Pyx_GIVEREF(__pyx_t_1);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_t_1);
              __pyx_t_1 = 0;
              __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_3, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6201, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_10);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6202
 *                 self.state = 859
 *                 self.match(ModelicaParser.T__70)
 *                 self.state = 860             # <<<<<<<<<<<<<<
 *                 self.expr(4)
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_860) < 0) __PYX_ERR(0, 6202, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6203
 *                 self.match(ModelicaParser.T__70)
 *                 self.state = 860
 *                 self.expr(4)             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6203, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_tuple__937, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6203, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6196
 *                 pass
 * 
 *             elif la_ == 3:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Expr_notContext(self, localctx)
 *                 self._ctx = localctx
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":6206
 *                 pass
 * 
 *             elif la_ == 4:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Expr_primaryContext(self, localctx)
 *                 self._ctx = localctx
 */
        __pyx_t_6 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_4, 4, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6206, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 6206, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (__pyx_t_12) {

          /* "pymola/generated/ModelicaParser.pyx":6207
 * 
 *             elif la_ == 4:
 *                 localctx = ModelicaParser.Expr_primaryContext(self, localctx)             # <<<<<<<<<<<<<<
 *                 self._ctx = localctx
 *                 _prevctx = localctx
 */
          __pyx_t_10 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6207, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_Expr_primaryContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6207, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_t_10 = NULL;
          __pyx_t_5 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_10)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_10);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
              __pyx_t_5 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6207, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_GOTREF(__pyx_t_6);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6207, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_GOTREF(__pyx_t_6);
          } else
          #endif
          {
            __pyx_t_1 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6207, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            if (__pyx_t_10) {
              __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_10); __pyx_t_10 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_5, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_5, __pyx_v_localctx);
            __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_1, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6207, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_6);
          __pyx_t_6 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6208
 *             elif la_ == 4:
 *                 localctx = ModelicaParser.Expr_primaryContext(self, localctx)
 *                 self._ctx = localctx             # <<<<<<<<<<<<<<
 *                 _prevctx = localctx
 *                 self.state = 861
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_ctx, __pyx_v_localctx) < 0) __PYX_ERR(0, 6208, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6209
 *                 localctx = ModelicaParser.Expr_primaryContext(self, localctx)
 *                 self._ctx = localctx
 *                 _prevctx = localctx             # <<<<<<<<<<<<<<
 *                 self.state = 861
 *                 self.primary()
 */
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_DECREF_SET(__pyx_v__prevctx, __pyx_v_localctx);

          /* "pymola/generated/ModelicaParser.pyx":6210
 *                 self._ctx = localctx
 *                 _prevctx = localctx
 *                 self.state = 861             # <<<<<<<<<<<<<<
 *                 self.primary()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_861) < 0) __PYX_ERR(0, 6210, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6211
 *                 _prevctx = localctx
 *                 self.state = 861
 *                 self.primary()             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_primary); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6211, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (__pyx_t_1) {
            __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6211, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          } else {
            __pyx_t_6 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6211, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6206
 *                 pass
 * 
 *             elif la_ == 4:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Expr_primaryContext(self, localctx)
 *                 self._ctx = localctx
 */
        }
        __pyx_L14:;

        /* "pymola/generated/ModelicaParser.pyx":6215
 * 
 * 
 *             self._ctx.stop = self._input.LT(-1)             # <<<<<<<<<<<<<<
 *             self.state = 881
 *             self._errHandler.sync(self)
 */
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6215, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_LT); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6215, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__938, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6215, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6215, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (__Pyx_PyObject_SetAttrStr(__pyx_t_3, __pyx_n_s_stop, __pyx_t_6) < 0) __PYX_ERR(0, 6215, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "pymola/generated/ModelicaParser.pyx":6216
 * 
 *             self._ctx.stop = self._input.LT(-1)
 *             self.state = 881             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _alt = self._interp.adaptivePredict(self._input,111,self._ctx)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_881) < 0) __PYX_ERR(0, 6216, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":6217
 *             self._ctx.stop = self._input.LT(-1)
 *             self.state = 881
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _alt = self._interp.adaptivePredict(self._input,111,self._ctx)
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 */
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6217, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_sync); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6217, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_1, function);
          }
        }
        if (!__pyx_t_6) {
          __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6217, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_1)) {
            PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6217, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
            PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6217, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_10 = PyTuple_New(1+1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6217, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_10);
            __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_6); __pyx_t_6 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_10, 0+1, __pyx_v_self);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_10, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6217, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "pymola/generated/ModelicaParser.pyx":6218
 *             self.state = 881
 *             self._errHandler.sync(self)
 *             _alt = self._interp.adaptivePredict(self._input,111,self._ctx)             # <<<<<<<<<<<<<<
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                 if _alt==1:
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_interp); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6218, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_adaptivePredict); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6218, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6218, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6218, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_10);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_10, function);
            __pyx_t_5 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_10)) {
          PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_t_1, __pyx_int_111, __pyx_t_6};
          __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6218, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
          PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_t_1, __pyx_int_111, __pyx_t_6};
          __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6218, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        } else
        #endif
        {
          __pyx_t_2 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6218, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_1);
          PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_5, __pyx_t_1);
          __Pyx_INCREF(__pyx_int_111);
          __Pyx_GIVEREF(__pyx_int_111);
          PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_5, __pyx_int_111);
          __Pyx_GIVEREF(__pyx_t_6);
          PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_5, __pyx_t_6);
          __pyx_t_1 = 0;
          __pyx_t_6 = 0;
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6218, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_v__alt = __pyx_t_3;
        __pyx_t_3 = 0;

        /* "pymola/generated/ModelicaParser.pyx":6219
 *             self._errHandler.sync(self)
 *             _alt = self._interp.adaptivePredict(self._input,111,self._ctx)
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:             # <<<<<<<<<<<<<<
 *                 if _alt==1:
 *                     if self._parseListeners is not None:
 */
        while (1) {
          __pyx_t_3 = PyObject_RichCompare(__pyx_v__alt, __pyx_int_2, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6219, __pyx_L6_error)
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 6219, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (__pyx_t_11) {
          } else {
            __pyx_t_12 = __pyx_t_11;
            goto __pyx_L20_bool_binop_done;
          }
          __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ATN); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6219, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_INVALID_ALT_NUMBER); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6219, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyObject_RichCompare(__pyx_v__alt, __pyx_t_10, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6219, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 6219, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_12 = __pyx_t_11;
          __pyx_L20_bool_binop_done:;
          if (!__pyx_t_12) break;

          /* "pymola/generated/ModelicaParser.pyx":6220
 *             _alt = self._interp.adaptivePredict(self._input,111,self._ctx)
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                 if _alt==1:             # <<<<<<<<<<<<<<
 *                     if self._parseListeners is not None:
 *                         self.triggerExitRuleEvent()
 */
          __pyx_t_3 = __Pyx_PyInt_EqObjC(__pyx_v__alt, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6220, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 6220, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (__pyx_t_12) {

            /* "pymola/generated/ModelicaParser.pyx":6221
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                 if _alt==1:
 *                     if self._parseListeners is not None:             # <<<<<<<<<<<<<<
 *                         self.triggerExitRuleEvent()
 *                     _prevctx = localctx
 */
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_parseListeners); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6221, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_12 = (__pyx_t_3 != Py_None);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_11 = (__pyx_t_12 != 0);
            if (__pyx_t_11) {

              /* "pymola/generated/ModelicaParser.pyx":6222
 *                 if _alt==1:
 *                     if self._parseListeners is not None:
 *                         self.triggerExitRuleEvent()             # <<<<<<<<<<<<<<
 *                     _prevctx = localctx
 *                     self.state = 879
 */
              __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_triggerExitRuleEvent); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6222, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_10);
              __pyx_t_2 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
                __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_10);
                if (likely(__pyx_t_2)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
                  __Pyx_INCREF(__pyx_t_2);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_10, function);
                }
              }
              if (__pyx_t_2) {
                __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6222, __pyx_L6_error)
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              } else {
                __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6222, __pyx_L6_error)
              }
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

              /* "pymola/generated/ModelicaParser.pyx":6221
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                 if _alt==1:
 *                     if self._parseListeners is not None:             # <<<<<<<<<<<<<<
 *                         self.triggerExitRuleEvent()
 *                     _prevctx = localctx
 */
            }

            /* "pymola/generated/ModelicaParser.pyx":6223
 *                     if self._parseListeners is not None:
 *                         self.triggerExitRuleEvent()
 *                     _prevctx = localctx             # <<<<<<<<<<<<<<
 *                     self.state = 879
 *                     self._errHandler.sync(self)
 */
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_DECREF_SET(__pyx_v__prevctx, __pyx_v_localctx);

            /* "pymola/generated/ModelicaParser.pyx":6224
 *                         self.triggerExitRuleEvent()
 *                     _prevctx = localctx
 *                     self.state = 879             # <<<<<<<<<<<<<<
 *                     self._errHandler.sync(self)
 *                     la_ = self._interp.adaptivePredict(self._input,110,self._ctx)
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_879) < 0) __PYX_ERR(0, 6224, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":6225
 *                     _prevctx = localctx
 *                     self.state = 879
 *                     self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                     la_ = self._interp.adaptivePredict(self._input,110,self._ctx)
 *                     if la_ == 1:
 */
            __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6225, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_10);
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6225, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            __pyx_t_10 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_10)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_10);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
              }
            }
            if (!__pyx_t_10) {
              __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6225, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_2)) {
                PyObject *__pyx_temp[2] = {__pyx_t_10, __pyx_v_self};
                __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6225, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
                __Pyx_GOTREF(__pyx_t_3);
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
                PyObject *__pyx_temp[2] = {__pyx_t_10, __pyx_v_self};
                __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6225, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
                __Pyx_GOTREF(__pyx_t_3);
              } else
              #endif
              {
                __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6225, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_10); __pyx_t_10 = NULL;
                __Pyx_INCREF(__pyx_v_self);
                __Pyx_GIVEREF(__pyx_v_self);
                PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
                __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6225, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

            /* "pymola/generated/ModelicaParser.pyx":6226
 *                     self.state = 879
 *                     self._errHandler.sync(self)
 *                     la_ = self._interp.adaptivePredict(self._input,110,self._ctx)             # <<<<<<<<<<<<<<
 *                     if la_ == 1:
 *                         localctx = ModelicaParser.Expr_mulContext(self, ModelicaParser.ExprContext(self, _parentctx, _parentState))
 */
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_interp); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6226, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_adaptivePredict); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6226, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6226, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6226, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_10);
            __pyx_t_1 = NULL;
            __pyx_t_5 = 0;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
              __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
              if (likely(__pyx_t_1)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                __Pyx_INCREF(__pyx_t_1);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_6, function);
                __pyx_t_5 = 1;
              }
            }
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_6)) {
              PyObject *__pyx_temp[4] = {__pyx_t_1, __pyx_t_2, __pyx_int_110, __pyx_t_10};
              __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6226, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
              PyObject *__pyx_temp[4] = {__pyx_t_1, __pyx_t_2, __pyx_int_110, __pyx_t_10};
              __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6226, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6226, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              if (__pyx_t_1) {
                __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1); __pyx_t_1 = NULL;
              }
              __Pyx_GIVEREF(__pyx_t_2);
              PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_5, __pyx_t_2);
              __Pyx_INCREF(__pyx_int_110);
              __Pyx_GIVEREF(__pyx_int_110);
              PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_5, __pyx_int_110);
              __Pyx_GIVEREF(__pyx_t_10);
              PyTuple_SET_ITEM(__pyx_t_4, 2+__pyx_t_5, __pyx_t_10);
              __pyx_t_2 = 0;
              __pyx_t_10 = 0;
              __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6226, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_DECREF_SET(__pyx_v_la_, __pyx_t_3);
            __pyx_t_3 = 0;

            /* "pymola/generated/ModelicaParser.pyx":6227
 *                     self._errHandler.sync(self)
 *                     la_ = self._interp.adaptivePredict(self._input,110,self._ctx)
 *                     if la_ == 1:             # <<<<<<<<<<<<<<
 *                         localctx = ModelicaParser.Expr_mulContext(self, ModelicaParser.ExprContext(self, _parentctx, _parentState))
 *                         self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
 */
            __pyx_t_3 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6227, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 6227, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (__pyx_t_11) {

              /* "pymola/generated/ModelicaParser.pyx":6228
 *                     la_ = self._interp.adaptivePredict(self._input,110,self._ctx)
 *                     if la_ == 1:
 *                         localctx = ModelicaParser.Expr_mulContext(self, ModelicaParser.ExprContext(self, _parentctx, _parentState))             # <<<<<<<<<<<<<<
 *                         self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
 *                         self.state = 864
 */
              __pyx_t_6 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6228, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_6);
              __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_Expr_mulContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6228, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __pyx_t_10 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6228, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_10);
              __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_ExprContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6228, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __pyx_t_10 = NULL;
              __pyx_t_5 = 0;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
                __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_2);
                if (likely(__pyx_t_10)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                  __Pyx_INCREF(__pyx_t_10);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_2, function);
                  __pyx_t_5 = 1;
                }
              }
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_2)) {
                PyObject *__pyx_temp[4] = {__pyx_t_10, __pyx_v_self, __pyx_v__parentctx, __pyx_v__parentState};
                __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6228, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
                __Pyx_GOTREF(__pyx_t_6);
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
                PyObject *__pyx_temp[4] = {__pyx_t_10, __pyx_v_self, __pyx_v__parentctx, __pyx_v__parentState};
                __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6228, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
                __Pyx_GOTREF(__pyx_t_6);
              } else
              #endif
              {
                __pyx_t_1 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6228, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                if (__pyx_t_10) {
                  __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_10); __pyx_t_10 = NULL;
                }
                __Pyx_INCREF(__pyx_v_self);
                __Pyx_GIVEREF(__pyx_v_self);
                PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_5, __pyx_v_self);
                __Pyx_INCREF(__pyx_v__parentctx);
                __Pyx_GIVEREF(__pyx_v__parentctx);
                PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_5, __pyx_v__parentctx);
                __Pyx_INCREF(__pyx_v__parentState);
                __Pyx_GIVEREF(__pyx_v__parentState);
                PyTuple_SET_ITEM(__pyx_t_1, 2+__pyx_t_5, __pyx_v__parentState);
                __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6228, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              }
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __pyx_t_2 = NULL;
              __pyx_t_5 = 0;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
                __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
                if (likely(__pyx_t_2)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                  __Pyx_INCREF(__pyx_t_2);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_4, function);
                  __pyx_t_5 = 1;
                }
              }
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_4)) {
                PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_t_6};
                __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6228, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
                PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_t_6};
                __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6228, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              } else
              #endif
              {
                __pyx_t_1 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6228, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                if (__pyx_t_2) {
                  __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2); __pyx_t_2 = NULL;
                }
                __Pyx_INCREF(__pyx_v_self);
                __Pyx_GIVEREF(__pyx_v_self);
                PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_5, __pyx_v_self);
                __Pyx_GIVEREF(__pyx_t_6);
                PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_5, __pyx_t_6);
                __pyx_t_6 = 0;
                __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_1, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6228, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              }
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_3);
              __pyx_t_3 = 0;

              /* "pymola/generated/ModelicaParser.pyx":6229
 *                     if la_ == 1:
 *                         localctx = ModelicaParser.Expr_mulContext(self, ModelicaParser.ExprContext(self, _parentctx, _parentState))
 *                         self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)             # <<<<<<<<<<<<<<
 *                         self.state = 864
 *                         if not self.precpred(self._ctx, 7):
 */
              __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_pushNewRecursionContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6229, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v__startState); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6229, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_expr); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6229, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_6);
              __pyx_t_2 = NULL;
              __pyx_t_5 = 0;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
                __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
                if (likely(__pyx_t_2)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                  __Pyx_INCREF(__pyx_t_2);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_4, function);
                  __pyx_t_5 = 1;
                }
              }
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_4)) {
                PyObject *__pyx_temp[4] = {__pyx_t_2, __pyx_v_localctx, __pyx_t_1, __pyx_t_6};
                __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6229, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
                PyObject *__pyx_temp[4] = {__pyx_t_2, __pyx_v_localctx, __pyx_t_1, __pyx_t_6};
                __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6229, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              } else
              #endif
              {
                __pyx_t_10 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6229, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_10);
                if (__pyx_t_2) {
                  __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_2); __pyx_t_2 = NULL;
                }
                __Pyx_INCREF(__pyx_v_localctx);
                __Pyx_GIVEREF(__pyx_v_localctx);
                PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_5, __pyx_v_localctx);
                __Pyx_GIVEREF(__pyx_t_1);
                PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_5, __pyx_t_1);
                __Pyx_GIVEREF(__pyx_t_6);
                PyTuple_SET_ITEM(__pyx_t_10, 2+__pyx_t_5, __pyx_t_6);
                __pyx_t_1 = 0;
                __pyx_t_6 = 0;
                __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_10, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6229, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              }
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

              /* "pymola/generated/ModelicaParser.pyx":6230
 *                         localctx = ModelicaParser.Expr_mulContext(self, ModelicaParser.ExprContext(self, _parentctx, _parentState))
 *                         self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
 *                         self.state = 864             # <<<<<<<<<<<<<<
 *                         if not self.precpred(self._ctx, 7):
 *                             from antlr4.error.Errors import FailedPredicateException
 */
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_864) < 0) __PYX_ERR(0, 6230, __pyx_L6_error)

              /* "pymola/generated/ModelicaParser.pyx":6231
 *                         self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
 *                         self.state = 864
 *                         if not self.precpred(self._ctx, 7):             # <<<<<<<<<<<<<<
 *                             from antlr4.error.Errors import FailedPredicateException
 *                             raise FailedPredicateException(self, "self.precpred(self._ctx, 7)")
 */
              __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_precpred); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6231, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6231, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_10);
              __pyx_t_6 = NULL;
              __pyx_t_5 = 0;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
                __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
                if (likely(__pyx_t_6)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                  __Pyx_INCREF(__pyx_t_6);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_4, function);
                  __pyx_t_5 = 1;
                }
              }
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_4)) {
                PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_int_7};
                __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6231, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
                PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_int_7};
                __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6231, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              } else
              #endif
              {
                __pyx_t_1 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6231, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                if (__pyx_t_6) {
                  __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_6); __pyx_t_6 = NULL;
                }
                __Pyx_GIVEREF(__pyx_t_10);
                PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_5, __pyx_t_10);
                __Pyx_INCREF(__pyx_int_7);
                __Pyx_GIVEREF(__pyx_int_7);
                PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_5, __pyx_int_7);
                __pyx_t_10 = 0;
                __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_1, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6231, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              }
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 6231, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __pyx_t_12 = ((!__pyx_t_11) != 0);
              if (__pyx_t_12) {

                /* "pymola/generated/ModelicaParser.pyx":6232
 *                         self.state = 864
 *                         if not self.precpred(self._ctx, 7):
 *                             from antlr4.error.Errors import FailedPredicateException             # <<<<<<<<<<<<<<
 *                             raise FailedPredicateException(self, "self.precpred(self._ctx, 7)")
 *                         self.state = 865
 */
                __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6232, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_INCREF(__pyx_n_s_FailedPredicateException);
                __Pyx_GIVEREF(__pyx_n_s_FailedPredicateException);
                PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_FailedPredicateException);
                __pyx_t_4 = __Pyx_Import(__pyx_n_s_antlr4_error_Errors, __pyx_t_3, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6232, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_FailedPredicateException); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6232, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_INCREF(__pyx_t_3);
                __pyx_v_FailedPredicateException = __pyx_t_3;
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

                /* "pymola/generated/ModelicaParser.pyx":6233
 *                         if not self.precpred(self._ctx, 7):
 *                             from antlr4.error.Errors import FailedPredicateException
 *                             raise FailedPredicateException(self, "self.precpred(self._ctx, 7)")             # <<<<<<<<<<<<<<
 *                         self.state = 865
 *                         localctx.op = self._input.LT(1)
 */
                __Pyx_INCREF(__pyx_v_FailedPredicateException);
                __pyx_t_3 = __pyx_v_FailedPredicateException; __pyx_t_1 = NULL;
                __pyx_t_5 = 0;
                if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
                  __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
                  if (likely(__pyx_t_1)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                    __Pyx_INCREF(__pyx_t_1);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_3, function);
                    __pyx_t_5 = 1;
                  }
                }
                #if CYTHON_FAST_PYCALL
                if (PyFunction_Check(__pyx_t_3)) {
                  PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_self, __pyx_kp_u_self_precpred_self__ctx_7};
                  __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6233, __pyx_L6_error)
                  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                  __Pyx_GOTREF(__pyx_t_4);
                } else
                #endif
                #if CYTHON_FAST_PYCCALL
                if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
                  PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_self, __pyx_kp_u_self_precpred_self__ctx_7};
                  __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6233, __pyx_L6_error)
                  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                  __Pyx_GOTREF(__pyx_t_4);
                } else
                #endif
                {
                  __pyx_t_10 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6233, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_10);
                  if (__pyx_t_1) {
                    __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_1); __pyx_t_1 = NULL;
                  }
                  __Pyx_INCREF(__pyx_v_self);
                  __Pyx_GIVEREF(__pyx_v_self);
                  PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_5, __pyx_v_self);
                  __Pyx_INCREF(__pyx_kp_u_self_precpred_self__ctx_7);
                  __Pyx_GIVEREF(__pyx_kp_u_self_precpred_self__ctx_7);
                  PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_5, __pyx_kp_u_self_precpred_self__ctx_7);
                  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6233, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_4);
                  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
                }
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_Raise(__pyx_t_4, 0, 0, 0);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                __PYX_ERR(0, 6233, __pyx_L6_error)

                /* "pymola/generated/ModelicaParser.pyx":6231
 *                         self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
 *                         self.state = 864
 *                         if not self.precpred(self._ctx, 7):             # <<<<<<<<<<<<<<
 *                             from antlr4.error.Errors import FailedPredicateException
 *                             raise FailedPredicateException(self, "self.precpred(self._ctx, 7)")
 */
              }

              /* "pymola/generated/ModelicaParser.pyx":6234
 *                             from antlr4.error.Errors import FailedPredicateException
 *                             raise FailedPredicateException(self, "self.precpred(self._ctx, 7)")
 *                         self.state = 865             # <<<<<<<<<<<<<<
 *                         localctx.op = self._input.LT(1)
 *                         _la = self._input.LA(1)
 */
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_865) < 0) __PYX_ERR(0, 6234, __pyx_L6_error)

              /* "pymola/generated/ModelicaParser.pyx":6235
 *                             raise FailedPredicateException(self, "self.precpred(self._ctx, 7)")
 *                         self.state = 865
 *                         localctx.op = self._input.LT(1)             # <<<<<<<<<<<<<<
 *                         _la = self._input.LA(1)
 *                         if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__28) | (1 << ModelicaParser.T__58) | (1 << ModelicaParser.T__59) | (1 << ModelicaParser.T__60))) != 0)):
 */
              __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6235, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_LT); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6235, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__939, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6235, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_op, __pyx_t_4) < 0) __PYX_ERR(0, 6235, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

              /* "pymola/generated/ModelicaParser.pyx":6236
 *                         self.state = 865
 *                         localctx.op = self._input.LT(1)
 *                         _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                         if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__28) | (1 << ModelicaParser.T__58) | (1 << ModelicaParser.T__59) | (1 << ModelicaParser.T__60))) != 0)):
 *                             localctx.op = self._errHandler.recoverInline(self)
 */
              __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6236, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_LA); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6236, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__940, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6236, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_XDECREF_SET(__pyx_v__la, __pyx_t_4);
              __pyx_t_4 = 0;

              /* "pymola/generated/ModelicaParser.pyx":6237
 *                         localctx.op = self._input.LT(1)
 *                         _la = self._input.LA(1)
 *                         if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__28) | (1 << ModelicaParser.T__58) | (1 << ModelicaParser.T__59) | (1 << ModelicaParser.T__60))) != 0)):             # <<<<<<<<<<<<<<
 *                             localctx.op = self._errHandler.recoverInline(self)
 *                         else:
 */
              __pyx_t_4 = __Pyx_PyInt_From_long((~0x3f)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6237, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_3 = PyNumber_And(__pyx_v__la, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6237, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __pyx_t_4 = __Pyx_PyInt_EqObjC(__pyx_t_3, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6237, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 6237, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              if (__pyx_t_11) {
              } else {
                __pyx_t_12 = __pyx_t_11;
                goto __pyx_L27_bool_binop_done;
              }
              __pyx_t_4 = PyNumber_Lshift(__pyx_int_1, __pyx_v__la); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6237, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6237, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__28); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6237, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_10);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __pyx_t_3 = PyNumber_Lshift(__pyx_int_1, __pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6237, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __pyx_t_10 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6237, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_10);
              __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_T__58); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6237, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __pyx_t_10 = PyNumber_Lshift(__pyx_int_1, __pyx_t_1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6237, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_10);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_t_1 = PyNumber_Or(__pyx_t_3, __pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6237, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __pyx_t_10 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6237, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_10);
              __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_T__59); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6237, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __pyx_t_10 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6237, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_10);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __pyx_t_3 = PyNumber_Or(__pyx_t_1, __pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6237, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __pyx_t_10 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6237, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_10);
              __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_T__60); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6237, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __pyx_t_10 = PyNumber_Lshift(__pyx_int_1, __pyx_t_1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6237, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_10);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_t_1 = PyNumber_Or(__pyx_t_3, __pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6237, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __pyx_t_10 = PyNumber_And(__pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6237, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_10);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_t_1 = PyObject_RichCompare(__pyx_t_10, __pyx_int_0, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6237, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 6237, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_t_12 = __pyx_t_11;
              __pyx_L27_bool_binop_done:;
              __pyx_t_11 = ((!__pyx_t_12) != 0);
              if (__pyx_t_11) {

                /* "pymola/generated/ModelicaParser.pyx":6238
 *                         _la = self._input.LA(1)
 *                         if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__28) | (1 << ModelicaParser.T__58) | (1 << ModelicaParser.T__59) | (1 << ModelicaParser.T__60))) != 0)):
 *                             localctx.op = self._errHandler.recoverInline(self)             # <<<<<<<<<<<<<<
 *                         else:
 *                             self._errHandler.reportMatch(self)
 */
                __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6238, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_10);
                __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_recoverInline); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6238, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
                __pyx_t_10 = NULL;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
                  __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_4);
                  if (likely(__pyx_t_10)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                    __Pyx_INCREF(__pyx_t_10);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_4, function);
                  }
                }
                if (!__pyx_t_10) {
                  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6238, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_1);
                } else {
                  #if CYTHON_FAST_PYCALL
                  if (PyFunction_Check(__pyx_t_4)) {
                    PyObject *__pyx_temp[2] = {__pyx_t_10, __pyx_v_self};
                    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6238, __pyx_L6_error)
                    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
                    __Pyx_GOTREF(__pyx_t_1);
                  } else
                  #endif
                  #if CYTHON_FAST_PYCCALL
                  if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
                    PyObject *__pyx_temp[2] = {__pyx_t_10, __pyx_v_self};
                    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6238, __pyx_L6_error)
                    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
                    __Pyx_GOTREF(__pyx_t_1);
                  } else
                  #endif
                  {
                    __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6238, __pyx_L6_error)
                    __Pyx_GOTREF(__pyx_t_3);
                    __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_10); __pyx_t_10 = NULL;
                    __Pyx_INCREF(__pyx_v_self);
                    __Pyx_GIVEREF(__pyx_v_self);
                    PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
                    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6238, __pyx_L6_error)
                    __Pyx_GOTREF(__pyx_t_1);
                    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                  }
                }
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_op, __pyx_t_1) < 0) __PYX_ERR(0, 6238, __pyx_L6_error)
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

                /* "pymola/generated/ModelicaParser.pyx":6237
 *                         localctx.op = self._input.LT(1)
 *                         _la = self._input.LA(1)
 *                         if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__28) | (1 << ModelicaParser.T__58) | (1 << ModelicaParser.T__59) | (1 << ModelicaParser.T__60))) != 0)):             # <<<<<<<<<<<<<<
 *                             localctx.op = self._errHandler.recoverInline(self)
 *                         else:
 */
                goto __pyx_L26;
              }

              /* "pymola/generated/ModelicaParser.pyx":6240
 *                             localctx.op = self._errHandler.recoverInline(self)
 *                         else:
 *                             self._errHandler.reportMatch(self)             # <<<<<<<<<<<<<<
 *                             self.consume()
 *                         self.state = 866
 */
              /*else*/ {
                __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6240, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_4);
                __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_reportMatch); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6240, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                __pyx_t_4 = NULL;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
                  __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
                  if (likely(__pyx_t_4)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                    __Pyx_INCREF(__pyx_t_4);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_3, function);
                  }
                }
                if (!__pyx_t_4) {
                  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6240, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_1);
                } else {
                  #if CYTHON_FAST_PYCALL
                  if (PyFunction_Check(__pyx_t_3)) {
                    PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
                    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6240, __pyx_L6_error)
                    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                    __Pyx_GOTREF(__pyx_t_1);
                  } else
                  #endif
                  #if CYTHON_FAST_PYCCALL
                  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
                    PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
                    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6240, __pyx_L6_error)
                    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                    __Pyx_GOTREF(__pyx_t_1);
                  } else
                  #endif
                  {
                    __pyx_t_10 = PyTuple_New(1+1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6240, __pyx_L6_error)
                    __Pyx_GOTREF(__pyx_t_10);
                    __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_4); __pyx_t_4 = NULL;
                    __Pyx_INCREF(__pyx_v_self);
                    __Pyx_GIVEREF(__pyx_v_self);
                    PyTuple_SET_ITEM(__pyx_t_10, 0+1, __pyx_v_self);
                    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6240, __pyx_L6_error)
                    __Pyx_GOTREF(__pyx_t_1);
                    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
                  }
                }
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

                /* "pymola/generated/ModelicaParser.pyx":6241
 *                         else:
 *                             self._errHandler.reportMatch(self)
 *                             self.consume()             # <<<<<<<<<<<<<<
 *                         self.state = 866
 *                         self.expr(8)
 */
                __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_consume); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6241, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_3);
                __pyx_t_10 = NULL;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
                  __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
                  if (likely(__pyx_t_10)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                    __Pyx_INCREF(__pyx_t_10);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_3, function);
                  }
                }
                if (__pyx_t_10) {
                  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6241, __pyx_L6_error)
                  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
                } else {
                  __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6241, __pyx_L6_error)
                }
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              }
              __pyx_L26:;

              /* "pymola/generated/ModelicaParser.pyx":6242
 *                             self._errHandler.reportMatch(self)
 *                             self.consume()
 *                         self.state = 866             # <<<<<<<<<<<<<<
 *                         self.expr(8)
 *                         pass
 */
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_866) < 0) __PYX_ERR(0, 6242, __pyx_L6_error)

              /* "pymola/generated/ModelicaParser.pyx":6243
 *                             self.consume()
 *                         self.state = 866
 *                         self.expr(8)             # <<<<<<<<<<<<<<
 *                         pass
 * 
 */
              __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6243, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__941, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6243, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

              /* "pymola/generated/ModelicaParser.pyx":6227
 *                     self._errHandler.sync(self)
 *                     la_ = self._interp.adaptivePredict(self._input,110,self._ctx)
 *                     if la_ == 1:             # <<<<<<<<<<<<<<
 *                         localctx = ModelicaParser.Expr_mulContext(self, ModelicaParser.ExprContext(self, _parentctx, _parentState))
 *                         self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
 */
              goto __pyx_L24;
            }

            /* "pymola/generated/ModelicaParser.pyx":6246
 *                         pass
 * 
 *                     elif la_ == 2:             # <<<<<<<<<<<<<<
 *                         localctx = ModelicaParser.Expr_addContext(self, ModelicaParser.ExprContext(self, _parentctx, _parentState))
 *                         self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
 */
            __pyx_t_3 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6246, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 6246, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (__pyx_t_11) {

              /* "pymola/generated/ModelicaParser.pyx":6247
 * 
 *                     elif la_ == 2:
 *                         localctx = ModelicaParser.Expr_addContext(self, ModelicaParser.ExprContext(self, _parentctx, _parentState))             # <<<<<<<<<<<<<<
 *                         self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
 *                         self.state = 867
 */
              __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6247, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_Expr_addContext); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6247, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_10);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6247, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_ExprContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6247, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __pyx_t_4 = NULL;
              __pyx_t_5 = 0;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
                __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
                if (likely(__pyx_t_4)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                  __Pyx_INCREF(__pyx_t_4);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_6, function);
                  __pyx_t_5 = 1;
                }
              }
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_6)) {
                PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_self, __pyx_v__parentctx, __pyx_v__parentState};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6247, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
                PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_self, __pyx_v__parentctx, __pyx_v__parentState};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6247, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else
              #endif
              {
                __pyx_t_2 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6247, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_2);
                if (__pyx_t_4) {
                  __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
                }
                __Pyx_INCREF(__pyx_v_self);
                __Pyx_GIVEREF(__pyx_v_self);
                PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_5, __pyx_v_self);
                __Pyx_INCREF(__pyx_v__parentctx);
                __Pyx_GIVEREF(__pyx_v__parentctx);
                PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_5, __pyx_v__parentctx);
                __Pyx_INCREF(__pyx_v__parentState);
                __Pyx_GIVEREF(__pyx_v__parentState);
                PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_5, __pyx_v__parentState);
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6247, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              }
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __pyx_t_6 = NULL;
              __pyx_t_5 = 0;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
                __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_10);
                if (likely(__pyx_t_6)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
                  __Pyx_INCREF(__pyx_t_6);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_10, function);
                  __pyx_t_5 = 1;
                }
              }
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_10)) {
                PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_self, __pyx_t_1};
                __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6247, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
                PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_self, __pyx_t_1};
                __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6247, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              } else
              #endif
              {
                __pyx_t_2 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6247, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_2);
                if (__pyx_t_6) {
                  __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_6); __pyx_t_6 = NULL;
                }
                __Pyx_INCREF(__pyx_v_self);
                __Pyx_GIVEREF(__pyx_v_self);
                PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_5, __pyx_v_self);
                __Pyx_GIVEREF(__pyx_t_1);
                PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_5, __pyx_t_1);
                __pyx_t_1 = 0;
                __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6247, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              }
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_3);
              __pyx_t_3 = 0;

              /* "pymola/generated/ModelicaParser.pyx":6248
 *                     elif la_ == 2:
 *                         localctx = ModelicaParser.Expr_addContext(self, ModelicaParser.ExprContext(self, _parentctx, _parentState))
 *                         self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)             # <<<<<<<<<<<<<<
 *                         self.state = 867
 *                         if not self.precpred(self._ctx, 6):
 */
              __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_pushNewRecursionContext); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6248, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_10);
              __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v__startState); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6248, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_expr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6248, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_6 = NULL;
              __pyx_t_5 = 0;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
                __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_10);
                if (likely(__pyx_t_6)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
                  __Pyx_INCREF(__pyx_t_6);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_10, function);
                  __pyx_t_5 = 1;
                }
              }
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_10)) {
                PyObject *__pyx_temp[4] = {__pyx_t_6, __pyx_v_localctx, __pyx_t_2, __pyx_t_1};
                __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6248, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
                PyObject *__pyx_temp[4] = {__pyx_t_6, __pyx_v_localctx, __pyx_t_2, __pyx_t_1};
                __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6248, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              } else
              #endif
              {
                __pyx_t_4 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6248, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_4);
                if (__pyx_t_6) {
                  __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_6); __pyx_t_6 = NULL;
                }
                __Pyx_INCREF(__pyx_v_localctx);
                __Pyx_GIVEREF(__pyx_v_localctx);
                PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_5, __pyx_v_localctx);
                __Pyx_GIVEREF(__pyx_t_2);
                PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_5, __pyx_t_2);
                __Pyx_GIVEREF(__pyx_t_1);
                PyTuple_SET_ITEM(__pyx_t_4, 2+__pyx_t_5, __pyx_t_1);
                __pyx_t_2 = 0;
                __pyx_t_1 = 0;
                __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6248, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              }
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

              /* "pymola/generated/ModelicaParser.pyx":6249
 *                         localctx = ModelicaParser.Expr_addContext(self, ModelicaParser.ExprContext(self, _parentctx, _parentState))
 *                         self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
 *                         self.state = 867             # <<<<<<<<<<<<<<
 *                         if not self.precpred(self._ctx, 6):
 *                             from antlr4.error.Errors import FailedPredicateException
 */
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_867) < 0) __PYX_ERR(0, 6249, __pyx_L6_error)

              /* "pymola/generated/ModelicaParser.pyx":6250
 *                         self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
 *                         self.state = 867
 *                         if not self.precpred(self._ctx, 6):             # <<<<<<<<<<<<<<
 *                             from antlr4.error.Errors import FailedPredicateException
 *                             raise FailedPredicateException(self, "self.precpred(self._ctx, 6)")
 */
              __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_precpred); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6250, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_10);
              __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6250, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_1 = NULL;
              __pyx_t_5 = 0;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
                __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_10);
                if (likely(__pyx_t_1)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
                  __Pyx_INCREF(__pyx_t_1);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_10, function);
                  __pyx_t_5 = 1;
                }
              }
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_10)) {
                PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_t_4, __pyx_int_6};
                __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6250, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
                PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_t_4, __pyx_int_6};
                __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6250, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              } else
              #endif
              {
                __pyx_t_2 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6250, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_2);
                if (__pyx_t_1) {
                  __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1); __pyx_t_1 = NULL;
                }
                __Pyx_GIVEREF(__pyx_t_4);
                PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_5, __pyx_t_4);
                __Pyx_INCREF(__pyx_int_6);
                __Pyx_GIVEREF(__pyx_int_6);
                PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_5, __pyx_int_6);
                __pyx_t_4 = 0;
                __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6250, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              }
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 6250, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __pyx_t_12 = ((!__pyx_t_11) != 0);
              if (__pyx_t_12) {

                /* "pymola/generated/ModelicaParser.pyx":6251
 *                         self.state = 867
 *                         if not self.precpred(self._ctx, 6):
 *                             from antlr4.error.Errors import FailedPredicateException             # <<<<<<<<<<<<<<
 *                             raise FailedPredicateException(self, "self.precpred(self._ctx, 6)")
 *                         self.state = 868
 */
                __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6251, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_INCREF(__pyx_n_s_FailedPredicateException);
                __Pyx_GIVEREF(__pyx_n_s_FailedPredicateException);
                PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_FailedPredicateException);
                __pyx_t_10 = __Pyx_Import(__pyx_n_s_antlr4_error_Errors, __pyx_t_3, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6251, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_10);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_10, __pyx_n_s_FailedPredicateException); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6251, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_INCREF(__pyx_t_3);
                __pyx_v_FailedPredicateException = __pyx_t_3;
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

                /* "pymola/generated/ModelicaParser.pyx":6252
 *                         if not self.precpred(self._ctx, 6):
 *                             from antlr4.error.Errors import FailedPredicateException
 *                             raise FailedPredicateException(self, "self.precpred(self._ctx, 6)")             # <<<<<<<<<<<<<<
 *                         self.state = 868
 *                         localctx.op = self._input.LT(1)
 */
                __Pyx_INCREF(__pyx_v_FailedPredicateException);
                __pyx_t_3 = __pyx_v_FailedPredicateException; __pyx_t_2 = NULL;
                __pyx_t_5 = 0;
                if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
                  __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
                  if (likely(__pyx_t_2)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                    __Pyx_INCREF(__pyx_t_2);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_3, function);
                    __pyx_t_5 = 1;
                  }
                }
                #if CYTHON_FAST_PYCALL
                if (PyFunction_Check(__pyx_t_3)) {
                  PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_kp_u_self_precpred_self__ctx_6};
                  __pyx_t_10 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6252, __pyx_L6_error)
                  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
                  __Pyx_GOTREF(__pyx_t_10);
                } else
                #endif
                #if CYTHON_FAST_PYCCALL
                if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
                  PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_kp_u_self_precpred_self__ctx_6};
                  __pyx_t_10 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6252, __pyx_L6_error)
                  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
                  __Pyx_GOTREF(__pyx_t_10);
                } else
                #endif
                {
                  __pyx_t_4 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6252, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_4);
                  if (__pyx_t_2) {
                    __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
                  }
                  __Pyx_INCREF(__pyx_v_self);
                  __Pyx_GIVEREF(__pyx_v_self);
                  PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_5, __pyx_v_self);
                  __Pyx_INCREF(__pyx_kp_u_self_precpred_self__ctx_6);
                  __Pyx_GIVEREF(__pyx_kp_u_self_precpred_self__ctx_6);
                  PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_5, __pyx_kp_u_self_precpred_self__ctx_6);
                  __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6252, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_10);
                  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                }
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_Raise(__pyx_t_10, 0, 0, 0);
                __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
                __PYX_ERR(0, 6252, __pyx_L6_error)

                /* "pymola/generated/ModelicaParser.pyx":6250
 *                         self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
 *                         self.state = 867
 *                         if not self.precpred(self._ctx, 6):             # <<<<<<<<<<<<<<
 *                             from antlr4.error.Errors import FailedPredicateException
 *                             raise FailedPredicateException(self, "self.precpred(self._ctx, 6)")
 */
              }

              /* "pymola/generated/ModelicaParser.pyx":6253
 *                             from antlr4.error.Errors import FailedPredicateException
 *                             raise FailedPredicateException(self, "self.precpred(self._ctx, 6)")
 *                         self.state = 868             # <<<<<<<<<<<<<<
 *                         localctx.op = self._input.LT(1)
 *                         _la = self._input.LA(1)
 */
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_868) < 0) __PYX_ERR(0, 6253, __pyx_L6_error)

              /* "pymola/generated/ModelicaParser.pyx":6254
 *                             raise FailedPredicateException(self, "self.precpred(self._ctx, 6)")
 *                         self.state = 868
 *                         localctx.op = self._input.LT(1)             # <<<<<<<<<<<<<<
 *                         _la = self._input.LA(1)
 *                         if not(((((_la - 56)) & ~0x3f) == 0 and ((1 << (_la - 56)) & ((1 << (ModelicaParser.T__55 - 56)) | (1 << (ModelicaParser.T__61 - 56)) | (1 << (ModelicaParser.T__62 - 56)) | (1 << (ModelicaParser.T__63 - 56)))) != 0)):
 */
              __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6254, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_10);
              __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_LT); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6254, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__942, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6254, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_10);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_op, __pyx_t_10) < 0) __PYX_ERR(0, 6254, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

              /* "pymola/generated/ModelicaParser.pyx":6255
 *                         self.state = 868
 *                         localctx.op = self._input.LT(1)
 *                         _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                         if not(((((_la - 56)) & ~0x3f) == 0 and ((1 << (_la - 56)) & ((1 << (ModelicaParser.T__55 - 56)) | (1 << (ModelicaParser.T__61 - 56)) | (1 << (ModelicaParser.T__62 - 56)) | (1 << (ModelicaParser.T__63 - 56)))) != 0)):
 *                             localctx.op = self._errHandler.recoverInline(self)
 */
              __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6255, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_10);
              __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_LA); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6255, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__943, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6255, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_10);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_XDECREF_SET(__pyx_v__la, __pyx_t_10);
              __pyx_t_10 = 0;

              /* "pymola/generated/ModelicaParser.pyx":6256
 *                         localctx.op = self._input.LT(1)
 *                         _la = self._input.LA(1)
 *                         if not(((((_la - 56)) & ~0x3f) == 0 and ((1 << (_la - 56)) & ((1 << (ModelicaParser.T__55 - 56)) | (1 << (ModelicaParser.T__61 - 56)) | (1 << (ModelicaParser.T__62 - 56)) | (1 << (ModelicaParser.T__63 - 56)))) != 0)):             # <<<<<<<<<<<<<<
 *                             localctx.op = self._errHandler.recoverInline(self)
 *                         else:
 */
              __pyx_t_10 = __Pyx_PyInt_SubtractObjC(__pyx_v__la, __pyx_int_56, 56, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6256, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_10);
              __pyx_t_3 = __Pyx_PyInt_From_long((~0x3f)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6256, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __pyx_t_4 = PyNumber_And(__pyx_t_10, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6256, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __pyx_t_3 = __Pyx_PyInt_EqObjC(__pyx_t_4, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6256, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 6256, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              if (__pyx_t_11) {
              } else {
                __pyx_t_12 = __pyx_t_11;
                goto __pyx_L31_bool_binop_done;
              }
              __pyx_t_3 = __Pyx_PyInt_SubtractObjC(__pyx_v__la, __pyx_int_56, 56, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6256, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __pyx_t_4 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6256, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6256, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__55); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6256, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_10);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __pyx_t_3 = __Pyx_PyInt_SubtractObjC(__pyx_t_10, __pyx_int_56, 56, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6256, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __pyx_t_10 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6256, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_10);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6256, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__61); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6256, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __pyx_t_3 = __Pyx_PyInt_SubtractObjC(__pyx_t_2, __pyx_int_56, 56, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6256, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6256, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __pyx_t_3 = PyNumber_Or(__pyx_t_10, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6256, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6256, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__62); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6256, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_10);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_t_10, __pyx_int_56, 56, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6256, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __pyx_t_10 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6256, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_10);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __pyx_t_2 = PyNumber_Or(__pyx_t_3, __pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6256, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __pyx_t_10 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6256, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_10);
              __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_T__63); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6256, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __pyx_t_10 = __Pyx_PyInt_SubtractObjC(__pyx_t_3, __pyx_int_56, 56, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6256, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_10);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __pyx_t_3 = PyNumber_Lshift(__pyx_int_1, __pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6256, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __pyx_t_10 = PyNumber_Or(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6256, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_10);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __pyx_t_3 = PyNumber_And(__pyx_t_4, __pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6256, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __pyx_t_10 = PyObject_RichCompare(__pyx_t_3, __pyx_int_0, Py_NE); __Pyx_XGOTREF(__pyx_t_10); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6256, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 6256, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __pyx_t_12 = __pyx_t_11;
              __pyx_L31_bool_binop_done:;
              __pyx_t_11 = ((!__pyx_t_12) != 0);
              if (__pyx_t_11) {

                /* "pymola/generated/ModelicaParser.pyx":6257
 *                         _la = self._input.LA(1)
 *                         if not(((((_la - 56)) & ~0x3f) == 0 and ((1 << (_la - 56)) & ((1 << (ModelicaParser.T__55 - 56)) | (1 << (ModelicaParser.T__61 - 56)) | (1 << (ModelicaParser.T__62 - 56)) | (1 << (ModelicaParser.T__63 - 56)))) != 0)):
 *                             localctx.op = self._errHandler.recoverInline(self)             # <<<<<<<<<<<<<<
 *                         else:
 *                             self._errHandler.reportMatch(self)
 */
                __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6257, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_3);
                __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_recoverInline); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6257, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                __pyx_t_3 = NULL;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
                  __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
                  if (likely(__pyx_t_3)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                    __Pyx_INCREF(__pyx_t_3);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_4, function);
                  }
                }
                if (!__pyx_t_3) {
                  __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6257, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_10);
                } else {
                  #if CYTHON_FAST_PYCALL
                  if (PyFunction_Check(__pyx_t_4)) {
                    PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
                    __pyx_t_10 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6257, __pyx_L6_error)
                    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                    __Pyx_GOTREF(__pyx_t_10);
                  } else
                  #endif
                  #if CYTHON_FAST_PYCCALL
                  if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
                    PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
                    __pyx_t_10 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6257, __pyx_L6_error)
                    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                    __Pyx_GOTREF(__pyx_t_10);
                  } else
                  #endif
                  {
                    __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6257, __pyx_L6_error)
                    __Pyx_GOTREF(__pyx_t_2);
                    __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
                    __Pyx_INCREF(__pyx_v_self);
                    __Pyx_GIVEREF(__pyx_v_self);
                    PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
                    __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6257, __pyx_L6_error)
                    __Pyx_GOTREF(__pyx_t_10);
                    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                  }
                }
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_op, __pyx_t_10) < 0) __PYX_ERR(0, 6257, __pyx_L6_error)
                __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

                /* "pymola/generated/ModelicaParser.pyx":6256
 *                         localctx.op = self._input.LT(1)
 *                         _la = self._input.LA(1)
 *                         if not(((((_la - 56)) & ~0x3f) == 0 and ((1 << (_la - 56)) & ((1 << (ModelicaParser.T__55 - 56)) | (1 << (ModelicaParser.T__61 - 56)) | (1 << (ModelicaParser.T__62 - 56)) | (1 << (ModelicaParser.T__63 - 56)))) != 0)):             # <<<<<<<<<<<<<<
 *                             localctx.op = self._errHandler.recoverInline(self)
 *                         else:
 */
                goto __pyx_L30;
              }

              /* "pymola/generated/ModelicaParser.pyx":6259
 *                             localctx.op = self._errHandler.recoverInline(self)
 *                         else:
 *                             self._errHandler.reportMatch(self)             # <<<<<<<<<<<<<<
 *                             self.consume()
 *                         self.state = 869
 */
              /*else*/ {
                __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6259, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_4);
                __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_reportMatch); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6259, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                __pyx_t_4 = NULL;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
                  __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
                  if (likely(__pyx_t_4)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                    __Pyx_INCREF(__pyx_t_4);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_2, function);
                  }
                }
                if (!__pyx_t_4) {
                  __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6259, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_10);
                } else {
                  #if CYTHON_FAST_PYCALL
                  if (PyFunction_Check(__pyx_t_2)) {
                    PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
                    __pyx_t_10 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6259, __pyx_L6_error)
                    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                    __Pyx_GOTREF(__pyx_t_10);
                  } else
                  #endif
                  #if CYTHON_FAST_PYCCALL
                  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
                    PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
                    __pyx_t_10 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6259, __pyx_L6_error)
                    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                    __Pyx_GOTREF(__pyx_t_10);
                  } else
                  #endif
                  {
                    __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6259, __pyx_L6_error)
                    __Pyx_GOTREF(__pyx_t_3);
                    __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __pyx_t_4 = NULL;
                    __Pyx_INCREF(__pyx_v_self);
                    __Pyx_GIVEREF(__pyx_v_self);
                    PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
                    __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6259, __pyx_L6_error)
                    __Pyx_GOTREF(__pyx_t_10);
                    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                  }
                }
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

                /* "pymola/generated/ModelicaParser.pyx":6260
 *                         else:
 *                             self._errHandler.reportMatch(self)
 *                             self.consume()             # <<<<<<<<<<<<<<
 *                         self.state = 869
 *                         self.expr(7)
 */
                __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_consume); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6260, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_2);
                __pyx_t_3 = NULL;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
                  __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
                  if (likely(__pyx_t_3)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                    __Pyx_INCREF(__pyx_t_3);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_2, function);
                  }
                }
                if (__pyx_t_3) {
                  __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6260, __pyx_L6_error)
                  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                } else {
                  __pyx_t_10 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6260, __pyx_L6_error)
                }
                __Pyx_GOTREF(__pyx_t_10);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              }
              __pyx_L30:;

              /* "pymola/generated/ModelicaParser.pyx":6261
 *                             self._errHandler.reportMatch(self)
 *                             self.consume()
 *                         self.state = 869             # <<<<<<<<<<<<<<
 *                         self.expr(7)
 *                         pass
 */
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_869) < 0) __PYX_ERR(0, 6261, __pyx_L6_error)

              /* "pymola/generated/ModelicaParser.pyx":6262
 *                             self.consume()
 *                         self.state = 869
 *                         self.expr(7)             # <<<<<<<<<<<<<<
 *                         pass
 * 
 */
              __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6262, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_10);
              __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_tuple__944, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6262, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

              /* "pymola/generated/ModelicaParser.pyx":6246
 *                         pass
 * 
 *                     elif la_ == 2:             # <<<<<<<<<<<<<<
 *                         localctx = ModelicaParser.Expr_addContext(self, ModelicaParser.ExprContext(self, _parentctx, _parentState))
 *                         self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
 */
              goto __pyx_L24;
            }

            /* "pymola/generated/ModelicaParser.pyx":6265
 *                         pass
 * 
 *                     elif la_ == 3:             # <<<<<<<<<<<<<<
 *                         localctx = ModelicaParser.Expr_relContext(self, ModelicaParser.ExprContext(self, _parentctx, _parentState))
 *                         self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
 */
            __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_3, 3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6265, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 6265, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (__pyx_t_11) {

              /* "pymola/generated/ModelicaParser.pyx":6266
 * 
 *                     elif la_ == 3:
 *                         localctx = ModelicaParser.Expr_relContext(self, ModelicaParser.ExprContext(self, _parentctx, _parentState))             # <<<<<<<<<<<<<<
 *                         self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
 *                         self.state = 870
 */
              __pyx_t_10 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6266, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_10);
              __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_Expr_relContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6266, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6266, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_ExprContext); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6266, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __pyx_t_4 = NULL;
              __pyx_t_5 = 0;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
                __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
                if (likely(__pyx_t_4)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
                  __Pyx_INCREF(__pyx_t_4);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_1, function);
                  __pyx_t_5 = 1;
                }
              }
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_1)) {
                PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_self, __pyx_v__parentctx, __pyx_v__parentState};
                __pyx_t_10 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6266, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_GOTREF(__pyx_t_10);
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
                PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_self, __pyx_v__parentctx, __pyx_v__parentState};
                __pyx_t_10 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6266, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_GOTREF(__pyx_t_10);
              } else
              #endif
              {
                __pyx_t_6 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6266, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_6);
                if (__pyx_t_4) {
                  __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
                }
                __Pyx_INCREF(__pyx_v_self);
                __Pyx_GIVEREF(__pyx_v_self);
                PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_self);
                __Pyx_INCREF(__pyx_v__parentctx);
                __Pyx_GIVEREF(__pyx_v__parentctx);
                PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v__parentctx);
                __Pyx_INCREF(__pyx_v__parentState);
                __Pyx_GIVEREF(__pyx_v__parentState);
                PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_5, __pyx_v__parentState);
                __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_6, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6266, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_10);
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              }
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_t_1 = NULL;
              __pyx_t_5 = 0;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
                __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
                if (likely(__pyx_t_1)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                  __Pyx_INCREF(__pyx_t_1);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_3, function);
                  __pyx_t_5 = 1;
                }
              }
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_3)) {
                PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_self, __pyx_t_10};
                __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6266, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
                PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_self, __pyx_t_10};
                __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6266, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              } else
              #endif
              {
                __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6266, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_6);
                if (__pyx_t_1) {
                  __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1); __pyx_t_1 = NULL;
                }
                __Pyx_INCREF(__pyx_v_self);
                __Pyx_GIVEREF(__pyx_v_self);
                PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_self);
                __Pyx_GIVEREF(__pyx_t_10);
                PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_t_10);
                __pyx_t_10 = 0;
                __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6266, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              }
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_2);
              __pyx_t_2 = 0;

              /* "pymola/generated/ModelicaParser.pyx":6267
 *                     elif la_ == 3:
 *                         localctx = ModelicaParser.Expr_relContext(self, ModelicaParser.ExprContext(self, _parentctx, _parentState))
 *                         self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)             # <<<<<<<<<<<<<<
 *                         self.state = 870
 *                         if not self.precpred(self._ctx, 5):
 */
              __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_pushNewRecursionContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6267, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __pyx_t_6 = __Pyx_PyInt_From_long(__pyx_v__startState); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6267, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_6);
              __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_expr); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6267, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_10);
              __pyx_t_1 = NULL;
              __pyx_t_5 = 0;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
                __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
                if (likely(__pyx_t_1)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                  __Pyx_INCREF(__pyx_t_1);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_3, function);
                  __pyx_t_5 = 1;
                }
              }
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_3)) {
                PyObject *__pyx_temp[4] = {__pyx_t_1, __pyx_v_localctx, __pyx_t_6, __pyx_t_10};
                __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6267, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
                PyObject *__pyx_temp[4] = {__pyx_t_1, __pyx_v_localctx, __pyx_t_6, __pyx_t_10};
                __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6267, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              } else
              #endif
              {
                __pyx_t_4 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6267, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_4);
                if (__pyx_t_1) {
                  __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1); __pyx_t_1 = NULL;
                }
                __Pyx_INCREF(__pyx_v_localctx);
                __Pyx_GIVEREF(__pyx_v_localctx);
                PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_5, __pyx_v_localctx);
                __Pyx_GIVEREF(__pyx_t_6);
                PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_5, __pyx_t_6);
                __Pyx_GIVEREF(__pyx_t_10);
                PyTuple_SET_ITEM(__pyx_t_4, 2+__pyx_t_5, __pyx_t_10);
                __pyx_t_6 = 0;
                __pyx_t_10 = 0;
                __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6267, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              }
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

              /* "pymola/generated/ModelicaParser.pyx":6268
 *                         localctx = ModelicaParser.Expr_relContext(self, ModelicaParser.ExprContext(self, _parentctx, _parentState))
 *                         self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
 *                         self.state = 870             # <<<<<<<<<<<<<<
 *                         if not self.precpred(self._ctx, 5):
 *                             from antlr4.error.Errors import FailedPredicateException
 */
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_870) < 0) __PYX_ERR(0, 6268, __pyx_L6_error)

              /* "pymola/generated/ModelicaParser.pyx":6269
 *                         self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
 *                         self.state = 870
 *                         if not self.precpred(self._ctx, 5):             # <<<<<<<<<<<<<<
 *                             from antlr4.error.Errors import FailedPredicateException
 *                             raise FailedPredicateException(self, "self.precpred(self._ctx, 5)")
 */
              __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_precpred); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6269, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6269, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_10 = NULL;
              __pyx_t_5 = 0;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
                __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
                if (likely(__pyx_t_10)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                  __Pyx_INCREF(__pyx_t_10);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_3, function);
                  __pyx_t_5 = 1;
                }
              }
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_3)) {
                PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_4, __pyx_int_5};
                __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6269, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
                PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_4, __pyx_int_5};
                __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6269, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              } else
              #endif
              {
                __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6269, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_6);
                if (__pyx_t_10) {
                  __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_10); __pyx_t_10 = NULL;
                }
                __Pyx_GIVEREF(__pyx_t_4);
                PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
                __Pyx_INCREF(__pyx_int_5);
                __Pyx_GIVEREF(__pyx_int_5);
                PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_5);
                __pyx_t_4 = 0;
                __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6269, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              }
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 6269, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __pyx_t_12 = ((!__pyx_t_11) != 0);
              if (__pyx_t_12) {

                /* "pymola/generated/ModelicaParser.pyx":6270
 *                         self.state = 870
 *                         if not self.precpred(self._ctx, 5):
 *                             from antlr4.error.Errors import FailedPredicateException             # <<<<<<<<<<<<<<
 *                             raise FailedPredicateException(self, "self.precpred(self._ctx, 5)")
 *                         self.state = 871
 */
                __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6270, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_INCREF(__pyx_n_s_FailedPredicateException);
                __Pyx_GIVEREF(__pyx_n_s_FailedPredicateException);
                PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_FailedPredicateException);
                __pyx_t_3 = __Pyx_Import(__pyx_n_s_antlr4_error_Errors, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6270, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_FailedPredicateException); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6270, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_2);
                __pyx_v_FailedPredicateException = __pyx_t_2;
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

                /* "pymola/generated/ModelicaParser.pyx":6271
 *                         if not self.precpred(self._ctx, 5):
 *                             from antlr4.error.Errors import FailedPredicateException
 *                             raise FailedPredicateException(self, "self.precpred(self._ctx, 5)")             # <<<<<<<<<<<<<<
 *                         self.state = 871
 *                         localctx.op = self._input.LT(1)
 */
                __Pyx_INCREF(__pyx_v_FailedPredicateException);
                __pyx_t_2 = __pyx_v_FailedPredicateException; __pyx_t_6 = NULL;
                __pyx_t_5 = 0;
                if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
                  __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
                  if (likely(__pyx_t_6)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                    __Pyx_INCREF(__pyx_t_6);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_2, function);
                    __pyx_t_5 = 1;
                  }
                }
                #if CYTHON_FAST_PYCALL
                if (PyFunction_Check(__pyx_t_2)) {
                  PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_self, __pyx_kp_u_self_precpred_self__ctx_5};
                  __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6271, __pyx_L6_error)
                  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
                  __Pyx_GOTREF(__pyx_t_3);
                } else
                #endif
                #if CYTHON_FAST_PYCCALL
                if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
                  PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_self, __pyx_kp_u_self_precpred_self__ctx_5};
                  __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6271, __pyx_L6_error)
                  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
                  __Pyx_GOTREF(__pyx_t_3);
                } else
                #endif
                {
                  __pyx_t_4 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6271, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_4);
                  if (__pyx_t_6) {
                    __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_6); __pyx_t_6 = NULL;
                  }
                  __Pyx_INCREF(__pyx_v_self);
                  __Pyx_GIVEREF(__pyx_v_self);
                  PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_5, __pyx_v_self);
                  __Pyx_INCREF(__pyx_kp_u_self_precpred_self__ctx_5);
                  __Pyx_GIVEREF(__pyx_kp_u_self_precpred_self__ctx_5);
                  PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_5, __pyx_kp_u_self_precpred_self__ctx_5);
                  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6271, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_3);
                  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                }
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                __Pyx_Raise(__pyx_t_3, 0, 0, 0);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                __PYX_ERR(0, 6271, __pyx_L6_error)

                /* "pymola/generated/ModelicaParser.pyx":6269
 *                         self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
 *                         self.state = 870
 *                         if not self.precpred(self._ctx, 5):             # <<<<<<<<<<<<<<
 *                             from antlr4.error.Errors import FailedPredicateException
 *                             raise FailedPredicateException(self, "self.precpred(self._ctx, 5)")
 */
              }

              /* "pymola/generated/ModelicaParser.pyx":6272
 *                             from antlr4.error.Errors import FailedPredicateException
 *                             raise FailedPredicateException(self, "self.precpred(self._ctx, 5)")
 *                         self.state = 871             # <<<<<<<<<<<<<<
 *                         localctx.op = self._input.LT(1)
 *                         _la = self._input.LA(1)
 */
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_871) < 0) __PYX_ERR(0, 6272, __pyx_L6_error)

              /* "pymola/generated/ModelicaParser.pyx":6273
 *                             raise FailedPredicateException(self, "self.precpred(self._ctx, 5)")
 *                         self.state = 871
 *                         localctx.op = self._input.LT(1)             # <<<<<<<<<<<<<<
 *                         _la = self._input.LA(1)
 *                         if not(((((_la - 65)) & ~0x3f) == 0 and ((1 << (_la - 65)) & ((1 << (ModelicaParser.T__64 - 65)) | (1 << (ModelicaParser.T__65 - 65)) | (1 << (ModelicaParser.T__66 - 65)) | (1 << (ModelicaParser.T__67 - 65)) | (1 << (ModelicaParser.T__68 - 65)) | (1 << (ModelicaParser.T__69 - 65)))) != 0)):
 */
              __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6273, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_LT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6273, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__945, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6273, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_op, __pyx_t_3) < 0) __PYX_ERR(0, 6273, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

              /* "pymola/generated/ModelicaParser.pyx":6274
 *                         self.state = 871
 *                         localctx.op = self._input.LT(1)
 *                         _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                         if not(((((_la - 65)) & ~0x3f) == 0 and ((1 << (_la - 65)) & ((1 << (ModelicaParser.T__64 - 65)) | (1 << (ModelicaParser.T__65 - 65)) | (1 << (ModelicaParser.T__66 - 65)) | (1 << (ModelicaParser.T__67 - 65)) | (1 << (ModelicaParser.T__68 - 65)) | (1 << (ModelicaParser.T__69 - 65)))) != 0)):
 *                             localctx.op = self._errHandler.recoverInline(self)
 */
              __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6274, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_LA); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6274, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__946, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6274, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_XDECREF_SET(__pyx_v__la, __pyx_t_3);
              __pyx_t_3 = 0;

              /* "pymola/generated/ModelicaParser.pyx":6275
 *                         localctx.op = self._input.LT(1)
 *                         _la = self._input.LA(1)
 *                         if not(((((_la - 65)) & ~0x3f) == 0 and ((1 << (_la - 65)) & ((1 << (ModelicaParser.T__64 - 65)) | (1 << (ModelicaParser.T__65 - 65)) | (1 << (ModelicaParser.T__66 - 65)) | (1 << (ModelicaParser.T__67 - 65)) | (1 << (ModelicaParser.T__68 - 65)) | (1 << (ModelicaParser.T__69 - 65)))) != 0)):             # <<<<<<<<<<<<<<
 *                             localctx.op = self._errHandler.recoverInline(self)
 *                         else:
 */
              __pyx_t_3 = __Pyx_PyInt_SubtractObjC(__pyx_v__la, __pyx_int_65, 65, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6275, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __pyx_t_2 = __Pyx_PyInt_From_long((~0x3f)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6275, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __pyx_t_4 = PyNumber_And(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6275, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_t_4, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6275, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 6275, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              if (__pyx_t_11) {
              } else {
                __pyx_t_12 = __pyx_t_11;
                goto __pyx_L35_bool_binop_done;
              }
              __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_v__la, __pyx_int_65, 65, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6275, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __pyx_t_4 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6275, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6275, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__64); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6275, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_t_3, __pyx_int_65, 65, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6275, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __pyx_t_3 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6275, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6275, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__65); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6275, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_t_6, __pyx_int_65, 65, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6275, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __pyx_t_6 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6275, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __pyx_t_2 = PyNumber_Or(__pyx_t_3, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6275, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __pyx_t_6 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6275, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_6);
              __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_T__66); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6275, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __pyx_t_6 = __Pyx_PyInt_SubtractObjC(__pyx_t_3, __pyx_int_65, 65, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6275, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __pyx_t_3 = PyNumber_Lshift(__pyx_int_1, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6275, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __pyx_t_6 = PyNumber_Or(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6275, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6275, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__67); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6275, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __pyx_t_3 = __Pyx_PyInt_SubtractObjC(__pyx_t_2, __pyx_int_65, 65, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6275, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6275, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __pyx_t_3 = PyNumber_Or(__pyx_t_6, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6275, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6275, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__68); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6275, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_t_6, __pyx_int_65, 65, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6275, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __pyx_t_6 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6275, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __pyx_t_2 = PyNumber_Or(__pyx_t_3, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6275, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __pyx_t_6 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6275, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_6);
              __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_T__69); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6275, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __pyx_t_6 = __Pyx_PyInt_SubtractObjC(__pyx_t_3, __pyx_int_65, 65, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6275, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __pyx_t_3 = PyNumber_Lshift(__pyx_int_1, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6275, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __pyx_t_6 = PyNumber_Or(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6275, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __pyx_t_3 = PyNumber_And(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6275, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __pyx_t_6 = PyObject_RichCompare(__pyx_t_3, __pyx_int_0, Py_NE); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6275, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 6275, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __pyx_t_12 = __pyx_t_11;
              __pyx_L35_bool_binop_done:;
              __pyx_t_11 = ((!__pyx_t_12) != 0);
              if (__pyx_t_11) {

                /* "pymola/generated/ModelicaParser.pyx":6276
 *                         _la = self._input.LA(1)
 *                         if not(((((_la - 65)) & ~0x3f) == 0 and ((1 << (_la - 65)) & ((1 << (ModelicaParser.T__64 - 65)) | (1 << (ModelicaParser.T__65 - 65)) | (1 << (ModelicaParser.T__66 - 65)) | (1 << (ModelicaParser.T__67 - 65)) | (1 << (ModelicaParser.T__68 - 65)) | (1 << (ModelicaParser.T__69 - 65)))) != 0)):
 *                             localctx.op = self._errHandler.recoverInline(self)             # <<<<<<<<<<<<<<
 *                         else:
 *                             self._errHandler.reportMatch(self)
 */
                __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6276, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_3);
                __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_recoverInline); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6276, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                __pyx_t_3 = NULL;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
                  __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
                  if (likely(__pyx_t_3)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                    __Pyx_INCREF(__pyx_t_3);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_4, function);
                  }
                }
                if (!__pyx_t_3) {
                  __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6276, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_6);
                } else {
                  #if CYTHON_FAST_PYCALL
                  if (PyFunction_Check(__pyx_t_4)) {
                    PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
                    __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6276, __pyx_L6_error)
                    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                    __Pyx_GOTREF(__pyx_t_6);
                  } else
                  #endif
                  #if CYTHON_FAST_PYCCALL
                  if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
                    PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
                    __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6276, __pyx_L6_error)
                    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                    __Pyx_GOTREF(__pyx_t_6);
                  } else
                  #endif
                  {
                    __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6276, __pyx_L6_error)
                    __Pyx_GOTREF(__pyx_t_2);
                    __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
                    __Pyx_INCREF(__pyx_v_self);
                    __Pyx_GIVEREF(__pyx_v_self);
                    PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
                    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6276, __pyx_L6_error)
                    __Pyx_GOTREF(__pyx_t_6);
                    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                  }
                }
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_op, __pyx_t_6) < 0) __PYX_ERR(0, 6276, __pyx_L6_error)
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

                /* "pymola/generated/ModelicaParser.pyx":6275
 *                         localctx.op = self._input.LT(1)
 *                         _la = self._input.LA(1)
 *                         if not(((((_la - 65)) & ~0x3f) == 0 and ((1 << (_la - 65)) & ((1 << (ModelicaParser.T__64 - 65)) | (1 << (ModelicaParser.T__65 - 65)) | (1 << (ModelicaParser.T__66 - 65)) | (1 << (ModelicaParser.T__67 - 65)) | (1 << (ModelicaParser.T__68 - 65)) | (1 << (ModelicaParser.T__69 - 65)))) != 0)):             # <<<<<<<<<<<<<<
 *                             localctx.op = self._errHandler.recoverInline(self)
 *                         else:
 */
                goto __pyx_L34;
              }

              /* "pymola/generated/ModelicaParser.pyx":6278
 *                             localctx.op = self._errHandler.recoverInline(self)
 *                         else:
 *                             self._errHandler.reportMatch(self)             # <<<<<<<<<<<<<<
 *                             self.consume()
 *                         self.state = 872
 */
              /*else*/ {
                __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6278, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_4);
                __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_reportMatch); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6278, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                __pyx_t_4 = NULL;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
                  __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
                  if (likely(__pyx_t_4)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                    __Pyx_INCREF(__pyx_t_4);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_2, function);
                  }
                }
                if (!__pyx_t_4) {
                  __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6278, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_6);
                } else {
                  #if CYTHON_FAST_PYCALL
                  if (PyFunction_Check(__pyx_t_2)) {
                    PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
                    __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6278, __pyx_L6_error)
                    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                    __Pyx_GOTREF(__pyx_t_6);
                  } else
                  #endif
                  #if CYTHON_FAST_PYCCALL
                  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
                    PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
                    __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6278, __pyx_L6_error)
                    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                    __Pyx_GOTREF(__pyx_t_6);
                  } else
                  #endif
                  {
                    __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6278, __pyx_L6_error)
                    __Pyx_GOTREF(__pyx_t_3);
                    __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __pyx_t_4 = NULL;
                    __Pyx_INCREF(__pyx_v_self);
                    __Pyx_GIVEREF(__pyx_v_self);
                    PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
                    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6278, __pyx_L6_error)
                    __Pyx_GOTREF(__pyx_t_6);
                    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                  }
                }
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

                /* "pymola/generated/ModelicaParser.pyx":6279
 *                         else:
 *                             self._errHandler.reportMatch(self)
 *                             self.consume()             # <<<<<<<<<<<<<<
 *                         self.state = 872
 *                         self.expr(6)
 */
                __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_consume); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6279, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_2);
                __pyx_t_3 = NULL;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
                  __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
                  if (likely(__pyx_t_3)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                    __Pyx_INCREF(__pyx_t_3);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_2, function);
                  }
                }
                if (__pyx_t_3) {
                  __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6279, __pyx_L6_error)
                  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                } else {
                  __pyx_t_6 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6279, __pyx_L6_error)
                }
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              }
              __pyx_L34:;

              /* "pymola/generated/ModelicaParser.pyx":6280
 *                             self._errHandler.reportMatch(self)
 *                             self.consume()
 *                         self.state = 872             # <<<<<<<<<<<<<<
 *                         self.expr(6)
 *                         pass
 */
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_872) < 0) __PYX_ERR(0, 6280, __pyx_L6_error)

              /* "pymola/generated/ModelicaParser.pyx":6281
 *                             self.consume()
 *                         self.state = 872
 *                         self.expr(6)             # <<<<<<<<<<<<<<
 *                         pass
 * 
 */
              __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6281, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_6);
              __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_tuple__947, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6281, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

              /* "pymola/generated/ModelicaParser.pyx":6265
 *                         pass
 * 
 *                     elif la_ == 3:             # <<<<<<<<<<<<<<
 *                         localctx = ModelicaParser.Expr_relContext(self, ModelicaParser.ExprContext(self, _parentctx, _parentState))
 *                         self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
 */
              goto __pyx_L24;
            }

            /* "pymola/generated/ModelicaParser.pyx":6284
 *                         pass
 * 
 *                     elif la_ == 4:             # <<<<<<<<<<<<<<
 *                         localctx = ModelicaParser.Expr_andContext(self, ModelicaParser.ExprContext(self, _parentctx, _parentState))
 *                         self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
 */
            __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_4, 4, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6284, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 6284, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (__pyx_t_11) {

              /* "pymola/generated/ModelicaParser.pyx":6285
 * 
 *                     elif la_ == 4:
 *                         localctx = ModelicaParser.Expr_andContext(self, ModelicaParser.ExprContext(self, _parentctx, _parentState))             # <<<<<<<<<<<<<<
 *                         self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
 *                         self.state = 873
 */
              __pyx_t_6 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6285, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_6);
              __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_Expr_andContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6285, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6285, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_ExprContext); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6285, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_10);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __pyx_t_4 = NULL;
              __pyx_t_5 = 0;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
                __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_10);
                if (likely(__pyx_t_4)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
                  __Pyx_INCREF(__pyx_t_4);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_10, function);
                  __pyx_t_5 = 1;
                }
              }
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_10)) {
                PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_self, __pyx_v__parentctx, __pyx_v__parentState};
                __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6285, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_GOTREF(__pyx_t_6);
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
                PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_self, __pyx_v__parentctx, __pyx_v__parentState};
                __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6285, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_GOTREF(__pyx_t_6);
              } else
              #endif
              {
                __pyx_t_1 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6285, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                if (__pyx_t_4) {
                  __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4); __pyx_t_4 = NULL;
                }
                __Pyx_INCREF(__pyx_v_self);
                __Pyx_GIVEREF(__pyx_v_self);
                PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_5, __pyx_v_self);
                __Pyx_INCREF(__pyx_v__parentctx);
                __Pyx_GIVEREF(__pyx_v__parentctx);
                PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_5, __pyx_v__parentctx);
                __Pyx_INCREF(__pyx_v__parentState);
                __Pyx_GIVEREF(__pyx_v__parentState);
                PyTuple_SET_ITEM(__pyx_t_1, 2+__pyx_t_5, __pyx_v__parentState);
                __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_1, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6285, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              }
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __pyx_t_10 = NULL;
              __pyx_t_5 = 0;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
                __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
                if (likely(__pyx_t_10)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                  __Pyx_INCREF(__pyx_t_10);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_3, function);
                  __pyx_t_5 = 1;
                }
              }
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_3)) {
                PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_v_self, __pyx_t_6};
                __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6285, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
                PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_v_self, __pyx_t_6};
                __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6285, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              } else
              #endif
              {
                __pyx_t_1 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6285, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                if (__pyx_t_10) {
                  __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_10); __pyx_t_10 = NULL;
                }
                __Pyx_INCREF(__pyx_v_self);
                __Pyx_GIVEREF(__pyx_v_self);
                PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_5, __pyx_v_self);
                __Pyx_GIVEREF(__pyx_t_6);
                PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_5, __pyx_t_6);
                __pyx_t_6 = 0;
                __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6285, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              }
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_2);
              __pyx_t_2 = 0;

              /* "pymola/generated/ModelicaParser.pyx":6286
 *                     elif la_ == 4:
 *                         localctx = ModelicaParser.Expr_andContext(self, ModelicaParser.ExprContext(self, _parentctx, _parentState))
 *                         self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)             # <<<<<<<<<<<<<<
 *                         self.state = 873
 *                         if not self.precpred(self._ctx, 3):
 */
              __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_pushNewRecursionContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6286, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v__startState); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6286, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_expr); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6286, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_6);
              __pyx_t_10 = NULL;
              __pyx_t_5 = 0;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
                __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
                if (likely(__pyx_t_10)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                  __Pyx_INCREF(__pyx_t_10);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_3, function);
                  __pyx_t_5 = 1;
                }
              }
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_3)) {
                PyObject *__pyx_temp[4] = {__pyx_t_10, __pyx_v_localctx, __pyx_t_1, __pyx_t_6};
                __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6286, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
                PyObject *__pyx_temp[4] = {__pyx_t_10, __pyx_v_localctx, __pyx_t_1, __pyx_t_6};
                __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6286, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              } else
              #endif
              {
                __pyx_t_4 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6286, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_4);
                if (__pyx_t_10) {
                  __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_10); __pyx_t_10 = NULL;
                }
                __Pyx_INCREF(__pyx_v_localctx);
                __Pyx_GIVEREF(__pyx_v_localctx);
                PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_5, __pyx_v_localctx);
                __Pyx_GIVEREF(__pyx_t_1);
                PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_5, __pyx_t_1);
                __Pyx_GIVEREF(__pyx_t_6);
                PyTuple_SET_ITEM(__pyx_t_4, 2+__pyx_t_5, __pyx_t_6);
                __pyx_t_1 = 0;
                __pyx_t_6 = 0;
                __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6286, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              }
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

              /* "pymola/generated/ModelicaParser.pyx":6287
 *                         localctx = ModelicaParser.Expr_andContext(self, ModelicaParser.ExprContext(self, _parentctx, _parentState))
 *                         self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
 *                         self.state = 873             # <<<<<<<<<<<<<<
 *                         if not self.precpred(self._ctx, 3):
 *                             from antlr4.error.Errors import FailedPredicateException
 */
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_873) < 0) __PYX_ERR(0, 6287, __pyx_L6_error)

              /* "pymola/generated/ModelicaParser.pyx":6288
 *                         self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
 *                         self.state = 873
 *                         if not self.precpred(self._ctx, 3):             # <<<<<<<<<<<<<<
 *                             from antlr4.error.Errors import FailedPredicateException
 *                             raise FailedPredicateException(self, "self.precpred(self._ctx, 3)")
 */
              __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_precpred); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6288, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6288, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_6 = NULL;
              __pyx_t_5 = 0;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
                __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
                if (likely(__pyx_t_6)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                  __Pyx_INCREF(__pyx_t_6);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_3, function);
                  __pyx_t_5 = 1;
                }
              }
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_3)) {
                PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_4, __pyx_int_3};
                __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6288, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
                PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_4, __pyx_int_3};
                __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6288, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              } else
              #endif
              {
                __pyx_t_1 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6288, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                if (__pyx_t_6) {
                  __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_6); __pyx_t_6 = NULL;
                }
                __Pyx_GIVEREF(__pyx_t_4);
                PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_5, __pyx_t_4);
                __Pyx_INCREF(__pyx_int_3);
                __Pyx_GIVEREF(__pyx_int_3);
                PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_5, __pyx_int_3);
                __pyx_t_4 = 0;
                __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6288, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              }
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 6288, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __pyx_t_12 = ((!__pyx_t_11) != 0);
              if (__pyx_t_12) {

                /* "pymola/generated/ModelicaParser.pyx":6289
 *                         self.state = 873
 *                         if not self.precpred(self._ctx, 3):
 *                             from antlr4.error.Errors import FailedPredicateException             # <<<<<<<<<<<<<<
 *                             raise FailedPredicateException(self, "self.precpred(self._ctx, 3)")
 *                         self.state = 874
 */
                __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6289, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_INCREF(__pyx_n_s_FailedPredicateException);
                __Pyx_GIVEREF(__pyx_n_s_FailedPredicateException);
                PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_FailedPredicateException);
                __pyx_t_3 = __Pyx_Import(__pyx_n_s_antlr4_error_Errors, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6289, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_FailedPredicateException); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6289, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_2);
                __pyx_v_FailedPredicateException = __pyx_t_2;
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

                /* "pymola/generated/ModelicaParser.pyx":6290
 *                         if not self.precpred(self._ctx, 3):
 *                             from antlr4.error.Errors import FailedPredicateException
 *                             raise FailedPredicateException(self, "self.precpred(self._ctx, 3)")             # <<<<<<<<<<<<<<
 *                         self.state = 874
 *                         self.match(ModelicaParser.T__71)
 */
                __Pyx_INCREF(__pyx_v_FailedPredicateException);
                __pyx_t_2 = __pyx_v_FailedPredicateException; __pyx_t_1 = NULL;
                __pyx_t_5 = 0;
                if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
                  __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
                  if (likely(__pyx_t_1)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                    __Pyx_INCREF(__pyx_t_1);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_2, function);
                    __pyx_t_5 = 1;
                  }
                }
                #if CYTHON_FAST_PYCALL
                if (PyFunction_Check(__pyx_t_2)) {
                  PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_self, __pyx_kp_u_self_precpred_self__ctx_3};
                  __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6290, __pyx_L6_error)
                  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                  __Pyx_GOTREF(__pyx_t_3);
                } else
                #endif
                #if CYTHON_FAST_PYCCALL
                if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
                  PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_self, __pyx_kp_u_self_precpred_self__ctx_3};
                  __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6290, __pyx_L6_error)
                  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                  __Pyx_GOTREF(__pyx_t_3);
                } else
                #endif
                {
                  __pyx_t_4 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6290, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_4);
                  if (__pyx_t_1) {
                    __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1); __pyx_t_1 = NULL;
                  }
                  __Pyx_INCREF(__pyx_v_self);
                  __Pyx_GIVEREF(__pyx_v_self);
                  PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_5, __pyx_v_self);
                  __Pyx_INCREF(__pyx_kp_u_self_precpred_self__ctx_3);
                  __Pyx_GIVEREF(__pyx_kp_u_self_precpred_self__ctx_3);
                  PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_5, __pyx_kp_u_self_precpred_self__ctx_3);
                  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6290, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_3);
                  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                }
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                __Pyx_Raise(__pyx_t_3, 0, 0, 0);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                __PYX_ERR(0, 6290, __pyx_L6_error)

                /* "pymola/generated/ModelicaParser.pyx":6288
 *                         self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
 *                         self.state = 873
 *                         if not self.precpred(self._ctx, 3):             # <<<<<<<<<<<<<<
 *                             from antlr4.error.Errors import FailedPredicateException
 *                             raise FailedPredicateException(self, "self.precpred(self._ctx, 3)")
 */
              }

              /* "pymola/generated/ModelicaParser.pyx":6291
 *                             from antlr4.error.Errors import FailedPredicateException
 *                             raise FailedPredicateException(self, "self.precpred(self._ctx, 3)")
 *                         self.state = 874             # <<<<<<<<<<<<<<
 *                         self.match(ModelicaParser.T__71)
 *                         self.state = 875
 */
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_874) < 0) __PYX_ERR(0, 6291, __pyx_L6_error)

              /* "pymola/generated/ModelicaParser.pyx":6292
 *                             raise FailedPredicateException(self, "self.precpred(self._ctx, 3)")
 *                         self.state = 874
 *                         self.match(ModelicaParser.T__71)             # <<<<<<<<<<<<<<
 *                         self.state = 875
 *                         self.expr(4)
 */
              __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6292, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6292, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__71); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6292, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __pyx_t_4 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
                __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
                if (likely(__pyx_t_4)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                  __Pyx_INCREF(__pyx_t_4);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_2, function);
                }
              }
              if (!__pyx_t_4) {
                __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6292, __pyx_L6_error)
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                __Pyx_GOTREF(__pyx_t_3);
              } else {
                #if CYTHON_FAST_PYCALL
                if (PyFunction_Check(__pyx_t_2)) {
                  PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_1};
                  __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6292, __pyx_L6_error)
                  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                  __Pyx_GOTREF(__pyx_t_3);
                  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                } else
                #endif
                #if CYTHON_FAST_PYCCALL
                if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
                  PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_1};
                  __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6292, __pyx_L6_error)
                  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                  __Pyx_GOTREF(__pyx_t_3);
                  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                } else
                #endif
                {
                  __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6292, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_6);
                  __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
                  __Pyx_GIVEREF(__pyx_t_1);
                  PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_1);
                  __pyx_t_1 = 0;
                  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6292, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_3);
                  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                }
              }
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

              /* "pymola/generated/ModelicaParser.pyx":6293
 *                         self.state = 874
 *                         self.match(ModelicaParser.T__71)
 *                         self.state = 875             # <<<<<<<<<<<<<<
 *                         self.expr(4)
 *                         pass
 */
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_875) < 0) __PYX_ERR(0, 6293, __pyx_L6_error)

              /* "pymola/generated/ModelicaParser.pyx":6294
 *                         self.match(ModelicaParser.T__71)
 *                         self.state = 875
 *                         self.expr(4)             # <<<<<<<<<<<<<<
 *                         pass
 * 
 */
              __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6294, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__948, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6294, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

              /* "pymola/generated/ModelicaParser.pyx":6284
 *                         pass
 * 
 *                     elif la_ == 4:             # <<<<<<<<<<<<<<
 *                         localctx = ModelicaParser.Expr_andContext(self, ModelicaParser.ExprContext(self, _parentctx, _parentState))
 *                         self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
 */
              goto __pyx_L24;
            }

            /* "pymola/generated/ModelicaParser.pyx":6297
 *                         pass
 * 
 *                     elif la_ == 5:             # <<<<<<<<<<<<<<
 *                         localctx = ModelicaParser.Expr_orContext(self, ModelicaParser.ExprContext(self, _parentctx, _parentState))
 *                         self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
 */
            __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_5, 5, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6297, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 6297, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (__pyx_t_12) {

              /* "pymola/generated/ModelicaParser.pyx":6298
 * 
 *                     elif la_ == 5:
 *                         localctx = ModelicaParser.Expr_orContext(self, ModelicaParser.ExprContext(self, _parentctx, _parentState))             # <<<<<<<<<<<<<<
 *                         self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
 *                         self.state = 876
 */
              __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6298, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Expr_orContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6298, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6298, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ExprContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6298, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_t_1 = NULL;
              __pyx_t_5 = 0;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
                __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
                if (likely(__pyx_t_1)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                  __Pyx_INCREF(__pyx_t_1);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_4, function);
                  __pyx_t_5 = 1;
                }
              }
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_4)) {
                PyObject *__pyx_temp[4] = {__pyx_t_1, __pyx_v_self, __pyx_v__parentctx, __pyx_v__parentState};
                __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6298, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                __Pyx_GOTREF(__pyx_t_3);
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
                PyObject *__pyx_temp[4] = {__pyx_t_1, __pyx_v_self, __pyx_v__parentctx, __pyx_v__parentState};
                __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6298, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                __Pyx_GOTREF(__pyx_t_3);
              } else
              #endif
              {
                __pyx_t_10 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6298, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_10);
                if (__pyx_t_1) {
                  __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_1); __pyx_t_1 = NULL;
                }
                __Pyx_INCREF(__pyx_v_self);
                __Pyx_GIVEREF(__pyx_v_self);
                PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_5, __pyx_v_self);
                __Pyx_INCREF(__pyx_v__parentctx);
                __Pyx_GIVEREF(__pyx_v__parentctx);
                PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_5, __pyx_v__parentctx);
                __Pyx_INCREF(__pyx_v__parentState);
                __Pyx_GIVEREF(__pyx_v__parentState);
                PyTuple_SET_ITEM(__pyx_t_10, 2+__pyx_t_5, __pyx_v__parentState);
                __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_10, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6298, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              }
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __pyx_t_4 = NULL;
              __pyx_t_5 = 0;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
                __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
                if (likely(__pyx_t_4)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                  __Pyx_INCREF(__pyx_t_4);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_6, function);
                  __pyx_t_5 = 1;
                }
              }
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_6)) {
                PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_t_3};
                __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6298, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
                PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_t_3};
                __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6298, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              } else
              #endif
              {
                __pyx_t_10 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6298, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_10);
                if (__pyx_t_4) {
                  __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_4); __pyx_t_4 = NULL;
                }
                __Pyx_INCREF(__pyx_v_self);
                __Pyx_GIVEREF(__pyx_v_self);
                PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_5, __pyx_v_self);
                __Pyx_GIVEREF(__pyx_t_3);
                PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_5, __pyx_t_3);
                __pyx_t_3 = 0;
                __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_10, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6298, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              }
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_2);
              __pyx_t_2 = 0;

              /* "pymola/generated/ModelicaParser.pyx":6299
 *                     elif la_ == 5:
 *                         localctx = ModelicaParser.Expr_orContext(self, ModelicaParser.ExprContext(self, _parentctx, _parentState))
 *                         self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)             # <<<<<<<<<<<<<<
 *                         self.state = 876
 *                         if not self.precpred(self._ctx, 2):
 */
              __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_pushNewRecursionContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6299, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_6);
              __pyx_t_10 = __Pyx_PyInt_From_long(__pyx_v__startState); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6299, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_10);
              __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_expr); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6299, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __pyx_t_4 = NULL;
              __pyx_t_5 = 0;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
                __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
                if (likely(__pyx_t_4)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                  __Pyx_INCREF(__pyx_t_4);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_6, function);
                  __pyx_t_5 = 1;
                }
              }
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_6)) {
                PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_t_10, __pyx_t_3};
                __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6299, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
                PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_t_10, __pyx_t_3};
                __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6299, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              } else
              #endif
              {
                __pyx_t_1 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6299, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                if (__pyx_t_4) {
                  __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4); __pyx_t_4 = NULL;
                }
                __Pyx_INCREF(__pyx_v_localctx);
                __Pyx_GIVEREF(__pyx_v_localctx);
                PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_5, __pyx_v_localctx);
                __Pyx_GIVEREF(__pyx_t_10);
                PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_5, __pyx_t_10);
                __Pyx_GIVEREF(__pyx_t_3);
                PyTuple_SET_ITEM(__pyx_t_1, 2+__pyx_t_5, __pyx_t_3);
                __pyx_t_10 = 0;
                __pyx_t_3 = 0;
                __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6299, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              }
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

              /* "pymola/generated/ModelicaParser.pyx":6300
 *                         localctx = ModelicaParser.Expr_orContext(self, ModelicaParser.ExprContext(self, _parentctx, _parentState))
 *                         self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
 *                         self.state = 876             # <<<<<<<<<<<<<<
 *                         if not self.precpred(self._ctx, 2):
 *                             from antlr4.error.Errors import FailedPredicateException
 */
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_876) < 0) __PYX_ERR(0, 6300, __pyx_L6_error)

              /* "pymola/generated/ModelicaParser.pyx":6301
 *                         self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
 *                         self.state = 876
 *                         if not self.precpred(self._ctx, 2):             # <<<<<<<<<<<<<<
 *                             from antlr4.error.Errors import FailedPredicateException
 *                             raise FailedPredicateException(self, "self.precpred(self._ctx, 2)")
 */
              __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_precpred); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6301, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_6);
              __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6301, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_3 = NULL;
              __pyx_t_5 = 0;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
                __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
                if (likely(__pyx_t_3)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                  __Pyx_INCREF(__pyx_t_3);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_6, function);
                  __pyx_t_5 = 1;
                }
              }
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_6)) {
                PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_1, __pyx_int_2};
                __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6301, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
                PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_1, __pyx_int_2};
                __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6301, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              } else
              #endif
              {
                __pyx_t_10 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6301, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_10);
                if (__pyx_t_3) {
                  __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_3); __pyx_t_3 = NULL;
                }
                __Pyx_GIVEREF(__pyx_t_1);
                PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_5, __pyx_t_1);
                __Pyx_INCREF(__pyx_int_2);
                __Pyx_GIVEREF(__pyx_int_2);
                PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_5, __pyx_int_2);
                __pyx_t_1 = 0;
                __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_10, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6301, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              }
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 6301, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __pyx_t_11 = ((!__pyx_t_12) != 0);
              if (__pyx_t_11) {

                /* "pymola/generated/ModelicaParser.pyx":6302
 *                         self.state = 876
 *                         if not self.precpred(self._ctx, 2):
 *                             from antlr4.error.Errors import FailedPredicateException             # <<<<<<<<<<<<<<
 *                             raise FailedPredicateException(self, "self.precpred(self._ctx, 2)")
 *                         self.state = 877
 */
                __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6302, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_INCREF(__pyx_n_s_FailedPredicateException);
                __Pyx_GIVEREF(__pyx_n_s_FailedPredicateException);
                PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_FailedPredicateException);
                __pyx_t_6 = __Pyx_Import(__pyx_n_s_antlr4_error_Errors, __pyx_t_2, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6302, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_6, __pyx_n_s_FailedPredicateException); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6302, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_2);
                __pyx_v_FailedPredicateException = __pyx_t_2;
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

                /* "pymola/generated/ModelicaParser.pyx":6303
 *                         if not self.precpred(self._ctx, 2):
 *                             from antlr4.error.Errors import FailedPredicateException
 *                             raise FailedPredicateException(self, "self.precpred(self._ctx, 2)")             # <<<<<<<<<<<<<<
 *                         self.state = 877
 *                         self.match(ModelicaParser.T__72)
 */
                __Pyx_INCREF(__pyx_v_FailedPredicateException);
                __pyx_t_2 = __pyx_v_FailedPredicateException; __pyx_t_10 = NULL;
                __pyx_t_5 = 0;
                if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
                  __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_2);
                  if (likely(__pyx_t_10)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                    __Pyx_INCREF(__pyx_t_10);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_2, function);
                    __pyx_t_5 = 1;
                  }
                }
                #if CYTHON_FAST_PYCALL
                if (PyFunction_Check(__pyx_t_2)) {
                  PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_v_self, __pyx_kp_u_self_precpred_self__ctx_2};
                  __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6303, __pyx_L6_error)
                  __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
                  __Pyx_GOTREF(__pyx_t_6);
                } else
                #endif
                #if CYTHON_FAST_PYCCALL
                if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
                  PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_v_self, __pyx_kp_u_self_precpred_self__ctx_2};
                  __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6303, __pyx_L6_error)
                  __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
                  __Pyx_GOTREF(__pyx_t_6);
                } else
                #endif
                {
                  __pyx_t_1 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6303, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_1);
                  if (__pyx_t_10) {
                    __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_10); __pyx_t_10 = NULL;
                  }
                  __Pyx_INCREF(__pyx_v_self);
                  __Pyx_GIVEREF(__pyx_v_self);
                  PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_5, __pyx_v_self);
                  __Pyx_INCREF(__pyx_kp_u_self_precpred_self__ctx_2);
                  __Pyx_GIVEREF(__pyx_kp_u_self_precpred_self__ctx_2);
                  PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_5, __pyx_kp_u_self_precpred_self__ctx_2);
                  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6303, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_6);
                  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                }
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                __Pyx_Raise(__pyx_t_6, 0, 0, 0);
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                __PYX_ERR(0, 6303, __pyx_L6_error)

                /* "pymola/generated/ModelicaParser.pyx":6301
 *                         self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
 *                         self.state = 876
 *                         if not self.precpred(self._ctx, 2):             # <<<<<<<<<<<<<<
 *                             from antlr4.error.Errors import FailedPredicateException
 *                             raise FailedPredicateException(self, "self.precpred(self._ctx, 2)")
 */
              }

              /* "pymola/generated/ModelicaParser.pyx":6304
 *                             from antlr4.error.Errors import FailedPredicateException
 *                             raise FailedPredicateException(self, "self.precpred(self._ctx, 2)")
 *                         self.state = 877             # <<<<<<<<<<<<<<
 *                         self.match(ModelicaParser.T__72)
 *                         self.state = 878
 */
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_877) < 0) __PYX_ERR(0, 6304, __pyx_L6_error)

              /* "pymola/generated/ModelicaParser.pyx":6305
 *                             raise FailedPredicateException(self, "self.precpred(self._ctx, 2)")
 *                         self.state = 877
 *                         self.match(ModelicaParser.T__72)             # <<<<<<<<<<<<<<
 *                         self.state = 878
 *                         self.expr(3)
 */
              __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6305, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6305, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__72); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6305, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_10);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_t_1 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
                __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
                if (likely(__pyx_t_1)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                  __Pyx_INCREF(__pyx_t_1);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_2, function);
                }
              }
              if (!__pyx_t_1) {
                __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6305, __pyx_L6_error)
                __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
                __Pyx_GOTREF(__pyx_t_6);
              } else {
                #if CYTHON_FAST_PYCALL
                if (PyFunction_Check(__pyx_t_2)) {
                  PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_t_10};
                  __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6305, __pyx_L6_error)
                  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                  __Pyx_GOTREF(__pyx_t_6);
                  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
                } else
                #endif
                #if CYTHON_FAST_PYCCALL
                if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
                  PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_t_10};
                  __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6305, __pyx_L6_error)
                  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                  __Pyx_GOTREF(__pyx_t_6);
                  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
                } else
                #endif
                {
                  __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6305, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_3);
                  __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1); __pyx_t_1 = NULL;
                  __Pyx_GIVEREF(__pyx_t_10);
                  PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_t_10);
                  __pyx_t_10 = 0;
                  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6305, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_6);
                  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                }
              }
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

              /* "pymola/generated/ModelicaParser.pyx":6306
 *                         self.state = 877
 *                         self.match(ModelicaParser.T__72)
 *                         self.state = 878             # <<<<<<<<<<<<<<
 *                         self.expr(3)
 *                         pass
 */
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_878) < 0) __PYX_ERR(0, 6306, __pyx_L6_error)

              /* "pymola/generated/ModelicaParser.pyx":6307
 *                         self.match(ModelicaParser.T__72)
 *                         self.state = 878
 *                         self.expr(3)             # <<<<<<<<<<<<<<
 *                         pass
 * 
 */
              __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6307, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_6);
              __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_tuple__949, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6307, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

              /* "pymola/generated/ModelicaParser.pyx":6297
 *                         pass
 * 
 *                     elif la_ == 5:             # <<<<<<<<<<<<<<
 *                         localctx = ModelicaParser.Expr_orContext(self, ModelicaParser.ExprContext(self, _parentctx, _parentState))
 *                         self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
 */
            }
            __pyx_L24:;

            /* "pymola/generated/ModelicaParser.pyx":6220
 *             _alt = self._interp.adaptivePredict(self._input,111,self._ctx)
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                 if _alt==1:             # <<<<<<<<<<<<<<
 *                     if self._parseListeners is not None:
 *                         self.triggerExitRuleEvent()
 */
          }

          /* "pymola/generated/ModelicaParser.pyx":6311
 * 
 * 
 *                 self.state = 883             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _alt = self._interp.adaptivePredict(self._input,111,self._ctx)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_883) < 0) __PYX_ERR(0, 6311, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6312
 * 
 *                 self.state = 883
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _alt = self._interp.adaptivePredict(self._input,111,self._ctx)
 * 
 */
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6312, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_sync); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6312, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_6 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_6)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_6);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (!__pyx_t_6) {
            __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6312, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
              __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6312, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_GOTREF(__pyx_t_2);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
              __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6312, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_GOTREF(__pyx_t_2);
            } else
            #endif
            {
              __pyx_t_10 = PyTuple_New(1+1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6312, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_10);
              __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_6); __pyx_t_6 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_10, 0+1, __pyx_v_self);
              __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_10, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6312, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6313
 *                 self.state = 883
 *                 self._errHandler.sync(self)
 *                 _alt = self._interp.adaptivePredict(self._input,111,self._ctx)             # <<<<<<<<<<<<<<
 * 
 *         except RecognitionException as re:
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_interp); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6313, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_adaptivePredict); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6313, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6313, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6313, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_1 = NULL;
          __pyx_t_5 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_10);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_10, function);
              __pyx_t_5 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_10)) {
            PyObject *__pyx_temp[4] = {__pyx_t_1, __pyx_t_3, __pyx_int_111, __pyx_t_6};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6313, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
            PyObject *__pyx_temp[4] = {__pyx_t_1, __pyx_t_3, __pyx_int_111, __pyx_t_6};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6313, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6313, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (__pyx_t_1) {
              __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1); __pyx_t_1 = NULL;
            }
            __Pyx_GIVEREF(__pyx_t_3);
            PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_5, __pyx_t_3);
            __Pyx_INCREF(__pyx_int_111);
            __Pyx_GIVEREF(__pyx_int_111);
            PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_5, __pyx_int_111);
            __Pyx_GIVEREF(__pyx_t_6);
            PyTuple_SET_ITEM(__pyx_t_4, 2+__pyx_t_5, __pyx_t_6);
            __pyx_t_3 = 0;
            __pyx_t_6 = 0;
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6313, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF_SET(__pyx_v__alt, __pyx_t_2);
          __pyx_t_2 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":6162
 *         self.enterRecursionRule(localctx, 112, self.RULE_expr, _p)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 862
 */
      }
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "pymola/generated/ModelicaParser.pyx":6315
 *                 _alt = self._interp.adaptivePredict(self._input,111,self._ctx)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6315, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_5 = __Pyx_PyErr_ExceptionMatches(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (__pyx_t_5) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.expr", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_10, &__pyx_t_4) < 0) __PYX_ERR(0, 6315, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_10);
        __pyx_v_re = __pyx_t_10;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":6316
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 6316, __pyx_L44_error)

          /* "pymola/generated/ModelicaParser.pyx":6317
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6317, __pyx_L44_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_reportError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6317, __pyx_L44_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          __pyx_t_5 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_1, function);
              __pyx_t_5 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_1)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_self, __pyx_v_re};
            __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6317, __pyx_L44_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_6);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_self, __pyx_v_re};
            __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6317, __pyx_L44_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_6);
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 6317, __pyx_L44_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_5, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_5, __pyx_v_re);
            __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_13, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6317, __pyx_L44_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6318
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.unrollRecursionContexts(_parentctx)
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6318, __pyx_L44_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_recover); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 6318, __pyx_L44_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = NULL;
          __pyx_t_5 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_5 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_self, __pyx_v_re};
            __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6318, __pyx_L44_error)
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_6);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_self, __pyx_v_re};
            __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6318, __pyx_L44_error)
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_6);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6318, __pyx_L44_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_1) {
              __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1); __pyx_t_1 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_5, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_5, __pyx_v_re);
            __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_3, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6318, __pyx_L44_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":6315
 *                 _alt = self._interp.adaptivePredict(self._input,111,self._ctx)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L45;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L44_error:;
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_5 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_5; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L8_except_error;
          }
          __pyx_L45:;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":6162
 *         self.enterRecursionRule(localctx, 112, self.RULE_expr, _p)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 862
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":6320
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.unrollRecursionContexts(_parentctx)             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_unrollRecursionContexts); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6320, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_10);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_10, function);
        }
      }
      if (!__pyx_t_2) {
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_v__parentctx); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6320, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_10)) {
          PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v__parentctx};
          __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6320, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_4);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
          PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v__parentctx};
          __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6320, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_4);
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6320, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2); __pyx_t_2 = NULL;
          __Pyx_INCREF(__pyx_v__parentctx);
          __Pyx_GIVEREF(__pyx_v__parentctx);
          PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v__parentctx);
          __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6320, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_7 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_9, &__pyx_t_8, &__pyx_t_7) < 0)) __Pyx_ErrFetch(&__pyx_t_9, &__pyx_t_8, &__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_14 = __pyx_lineno; __pyx_t_5 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {
        __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_unrollRecursionContexts); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6320, __pyx_L51_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_6 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_10);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_10, function);
          }
        }
        if (!__pyx_t_6) {
          __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_v__parentctx); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6320, __pyx_L51_error)
          __Pyx_GOTREF(__pyx_t_4);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_10)) {
            PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v__parentctx};
            __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6320, __pyx_L51_error)
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
            PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v__parentctx};
            __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6320, __pyx_L51_error)
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6320, __pyx_L51_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_6); __pyx_t_6 = NULL;
            __Pyx_INCREF(__pyx_v__parentctx);
            __Pyx_GIVEREF(__pyx_v__parentctx);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v__parentctx);
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_2, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6320, __pyx_L51_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ErrRestore(__pyx_t_9, __pyx_t_8, __pyx_t_7);
      __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_7 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_5; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
      __pyx_L51_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":6321
 *         finally:
 *             self.unrollRecursionContexts(_parentctx)
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class PrimaryContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":6154
 * 
 * 
 *     def expr(self, _p:int=0):             # <<<<<<<<<<<<<<
 *         _parentctx = self._ctx
 *         _parentState = self.state
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.expr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v__parentctx);
  __Pyx_XDECREF(__pyx_v__parentState);
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__prevctx);
  __Pyx_XDECREF(__pyx_v_la_);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v__alt);
  __Pyx_XDECREF(__pyx_v_FailedPredicateException);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6325
 *     class PrimaryContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14PrimaryContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14PrimaryContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14PrimaryContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14PrimaryContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 6325, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 6325, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6325, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.PrimaryContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14PrimaryContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14PrimaryContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":6326
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 6326, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6326, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6326, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6326, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6326, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6327
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 6327, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6325
 *     class PrimaryContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.PrimaryContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6330
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_primary
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14PrimaryContext_3getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14PrimaryContext_3getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14PrimaryContext_3getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14PrimaryContext_3getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14PrimaryContext_2getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14PrimaryContext_2getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":6331
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_primary             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6331, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_primary); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6331, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":6330
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_primary
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.PrimaryContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6334
 * 
 * 
 *         def copyFrom(self, ctx:ParserRuleContext):             # <<<<<<<<<<<<<<
 *             super().copyFrom(ctx)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14PrimaryContext_5copyFrom(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14PrimaryContext_5copyFrom = {"copyFrom", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14PrimaryContext_5copyFrom, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14PrimaryContext_5copyFrom(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copyFrom (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx_2,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("copyFrom", 1, 2, 2, 1); __PYX_ERR(0, 6334, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "copyFrom") < 0)) __PYX_ERR(0, 6334, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("copyFrom", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6334, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.PrimaryContext.copyFrom", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14PrimaryContext_4copyFrom(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14PrimaryContext_4copyFrom(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("copyFrom", 0);

  /* "pymola/generated/ModelicaParser.pyx":6335
 * 
 *         def copyFrom(self, ctx:ParserRuleContext):
 *             super().copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 6335, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6335, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6335, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6335, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6335, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6335, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6334
 * 
 * 
 *         def copyFrom(self, ctx:ParserRuleContext):             # <<<<<<<<<<<<<<
 *             super().copyFrom(ctx)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.PrimaryContext.copyFrom", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6341
 *     class Primary_stringContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Primary_stringContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Primary_stringContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Primary_stringContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Primary_stringContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 6341, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 6341, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 6341, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6341, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_stringContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Primary_stringContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Primary_stringContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":6342
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 6342, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6342, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6342, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6342, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6342, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6342, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6343
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def STRING(self):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6343, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6343, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6343, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6343, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6343, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6341
 *     class Primary_stringContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_stringContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6345
 *             self.copyFrom(ctx)
 * 
 *         def STRING(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.STRING, 0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Primary_stringContext_3STRING(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Primary_stringContext_3STRING = {"STRING", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Primary_stringContext_3STRING, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Primary_stringContext_3STRING(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("STRING (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Primary_stringContext_2STRING(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Primary_stringContext_2STRING(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("STRING", 0);

  /* "pymola/generated/ModelicaParser.pyx":6346
 * 
 *         def STRING(self):
 *             return self.getToken(ModelicaParser.STRING, 0)             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getToken); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6346, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6346, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_STRING); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6346, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6346, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6346, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6346, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6346, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":6345
 *             self.copyFrom(ctx)
 * 
 *         def STRING(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.STRING, 0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_stringContext.STRING", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6348
 *             return self.getToken(ModelicaParser.STRING, 0)
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_string" ):
 *                 listener.enterPrimary_string(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Primary_stringContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Primary_stringContext_5enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Primary_stringContext_5enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Primary_stringContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 6348, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 6348, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6348, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_stringContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Primary_stringContext_4enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Primary_stringContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6349
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterPrimary_string" ):             # <<<<<<<<<<<<<<
 *                 listener.enterPrimary_string(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterPrimary_string); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6349, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6350
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterPrimary_string" ):
 *                 listener.enterPrimary_string(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterPrimary_string); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6350, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6350, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6350, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6350, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6350, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6350, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6349
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterPrimary_string" ):             # <<<<<<<<<<<<<<
 *                 listener.enterPrimary_string(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6348
 *             return self.getToken(ModelicaParser.STRING, 0)
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_string" ):
 *                 listener.enterPrimary_string(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_stringContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6352
 *                 listener.enterPrimary_string(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_string" ):
 *                 listener.exitPrimary_string(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Primary_stringContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Primary_stringContext_7exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Primary_stringContext_7exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Primary_stringContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 6352, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 6352, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6352, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_stringContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Primary_stringContext_6exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Primary_stringContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6353
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitPrimary_string" ):             # <<<<<<<<<<<<<<
 *                 listener.exitPrimary_string(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitPrimary_string); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6353, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6354
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitPrimary_string" ):
 *                 listener.exitPrimary_string(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitPrimary_string); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6354, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6354, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6354, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6354, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6354, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6354, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6353
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitPrimary_string" ):             # <<<<<<<<<<<<<<
 *                 listener.exitPrimary_string(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6352
 *                 listener.enterPrimary_string(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_string" ):
 *                 listener.exitPrimary_string(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_stringContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6356
 *                 listener.exitPrimary_string(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_string" ):
 *                 return visitor.visitPrimary_string(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Primary_stringContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Primary_stringContext_9accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Primary_stringContext_9accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Primary_stringContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 6356, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 6356, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6356, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_stringContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Primary_stringContext_8accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Primary_stringContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":6357
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitPrimary_string" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitPrimary_string(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitPrimary_string); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6357, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6358
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitPrimary_string" ):
 *                 return visitor.visitPrimary_string(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitPrimary_string); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6358, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6358, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6358, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6358, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6358, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6358, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":6357
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitPrimary_string" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitPrimary_string(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6360
 *                 return visitor.visitPrimary_string(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6360, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6360, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6360, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6360, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6360, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6360, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":6356
 *                 listener.exitPrimary_string(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_string" ):
 *                 return visitor.visitPrimary_string(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_stringContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6365
 *     class Primary_endContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Primary_endContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18Primary_endContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18Primary_endContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Primary_endContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 6365, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 6365, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 6365, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6365, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_endContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18Primary_endContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Primary_endContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":6366
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 6366, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6366, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6366, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6366, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6366, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6366, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6367
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6367, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6367, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6367, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6367, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6367, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6367, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6365
 *     class Primary_endContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_endContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6370
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_end" ):
 *                 listener.enterPrimary_end(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Primary_endContext_3enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18Primary_endContext_3enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18Primary_endContext_3enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Primary_endContext_3enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 6370, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 6370, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6370, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_endContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18Primary_endContext_2enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Primary_endContext_2enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6371
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterPrimary_end" ):             # <<<<<<<<<<<<<<
 *                 listener.enterPrimary_end(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterPrimary_end); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6371, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6372
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterPrimary_end" ):
 *                 listener.enterPrimary_end(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterPrimary_end); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6372, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6372, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6372, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6372, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6372, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6372, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6371
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterPrimary_end" ):             # <<<<<<<<<<<<<<
 *                 listener.enterPrimary_end(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6370
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_end" ):
 *                 listener.enterPrimary_end(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_endContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6374
 *                 listener.enterPrimary_end(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_end" ):
 *                 listener.exitPrimary_end(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Primary_endContext_5exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18Primary_endContext_5exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18Primary_endContext_5exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Primary_endContext_5exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 6374, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 6374, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6374, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_endContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18Primary_endContext_4exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Primary_endContext_4exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6375
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitPrimary_end" ):             # <<<<<<<<<<<<<<
 *                 listener.exitPrimary_end(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitPrimary_end); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6375, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6376
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitPrimary_end" ):
 *                 listener.exitPrimary_end(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitPrimary_end); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6376, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6376, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6376, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6376, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6376, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6376, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6375
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitPrimary_end" ):             # <<<<<<<<<<<<<<
 *                 listener.exitPrimary_end(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6374
 *                 listener.enterPrimary_end(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_end" ):
 *                 listener.exitPrimary_end(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_endContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6378
 *                 listener.exitPrimary_end(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_end" ):
 *                 return visitor.visitPrimary_end(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Primary_endContext_7accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_18Primary_endContext_7accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_18Primary_endContext_7accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_18Primary_endContext_7accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 6378, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 6378, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6378, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_endContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_18Primary_endContext_6accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_18Primary_endContext_6accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":6379
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitPrimary_end" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitPrimary_end(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitPrimary_end); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6379, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6380
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitPrimary_end" ):
 *                 return visitor.visitPrimary_end(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitPrimary_end); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6380, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6380, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6380, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6380, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6380, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6380, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":6379
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitPrimary_end" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitPrimary_end(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6382
 *                 return visitor.visitPrimary_end(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6382, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6382, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6382, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6382, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6382, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":6378
 *                 listener.exitPrimary_end(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_end" ):
 *                 return visitor.visitPrimary_end(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_endContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6387
 *     class Primary_output_expression_listContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_37Primary_output_expression_listContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_37Primary_output_expression_listContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_37Primary_output_expression_listContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_37Primary_output_expression_listContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 6387, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 6387, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 6387, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6387, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_output_expression_listContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_37Primary_output_expression_listContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_37Primary_output_expression_listContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":6388
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 6388, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6388, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6388, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6388, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6388, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6388, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6389
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def output_expression_list(self):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6389, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6389, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6389, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6389, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6389, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6389, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6387
 *     class Primary_output_expression_listContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_output_expression_listContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6391
 *             self.copyFrom(ctx)
 * 
 *         def output_expression_list(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Output_expression_listContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_37Primary_output_expression_listContext_3output_expression_list(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_37Primary_output_expression_listContext_3output_expression_list = {"output_expression_list", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_37Primary_output_expression_listContext_3output_expression_list, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_37Primary_output_expression_listContext_3output_expression_list(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("output_expression_list (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_37Primary_output_expression_listContext_2output_expression_list(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_37Primary_output_expression_listContext_2output_expression_list(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("output_expression_list", 0);

  /* "pymola/generated/ModelicaParser.pyx":6392
 * 
 *         def output_expression_list(self):
 *             return self.getTypedRuleContext(ModelicaParser.Output_expression_listContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6392, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6392, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Output_expression_listContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6392, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6392, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6392, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6392, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6392, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":6391
 *             self.copyFrom(ctx)
 * 
 *         def output_expression_list(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Output_expression_listContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_output_expression_listContext.output_expression_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6395
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_output_expression_list" ):
 *                 listener.enterPrimary_output_expression_list(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_37Primary_output_expression_listContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_37Primary_output_expression_listContext_5enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_37Primary_output_expression_listContext_5enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_37Primary_output_expression_listContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 6395, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 6395, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6395, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_output_expression_listContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_37Primary_output_expression_listContext_4enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_37Primary_output_expression_listContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6396
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterPrimary_output_expression_list" ):             # <<<<<<<<<<<<<<
 *                 listener.enterPrimary_output_expression_list(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterPrimary_output_expression_l); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6396, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6397
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterPrimary_output_expression_list" ):
 *                 listener.enterPrimary_output_expression_list(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterPrimary_output_expression_l); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6397, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6397, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6397, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6397, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6397, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6396
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterPrimary_output_expression_list" ):             # <<<<<<<<<<<<<<
 *                 listener.enterPrimary_output_expression_list(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6395
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_output_expression_list" ):
 *                 listener.enterPrimary_output_expression_list(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_output_expression_listContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6399
 *                 listener.enterPrimary_output_expression_list(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_output_expression_list" ):
 *                 listener.exitPrimary_output_expression_list(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_37Primary_output_expression_listContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_37Primary_output_expression_listContext_7exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_37Primary_output_expression_listContext_7exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_37Primary_output_expression_listContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 6399, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 6399, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6399, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_output_expression_listContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_37Primary_output_expression_listContext_6exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_37Primary_output_expression_listContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6400
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitPrimary_output_expression_list" ):             # <<<<<<<<<<<<<<
 *                 listener.exitPrimary_output_expression_list(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitPrimary_output_expression_li); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6400, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6401
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitPrimary_output_expression_list" ):
 *                 listener.exitPrimary_output_expression_list(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitPrimary_output_expression_li); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6401, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6401, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6401, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6401, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6401, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6400
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitPrimary_output_expression_list" ):             # <<<<<<<<<<<<<<
 *                 listener.exitPrimary_output_expression_list(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6399
 *                 listener.enterPrimary_output_expression_list(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_output_expression_list" ):
 *                 listener.exitPrimary_output_expression_list(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_output_expression_listContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6403
 *                 listener.exitPrimary_output_expression_list(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_output_expression_list" ):
 *                 return visitor.visitPrimary_output_expression_list(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_37Primary_output_expression_listContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_37Primary_output_expression_listContext_9accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_37Primary_output_expression_listContext_9accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_37Primary_output_expression_listContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 6403, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 6403, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6403, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_output_expression_listContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_37Primary_output_expression_listContext_8accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_37Primary_output_expression_listContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":6404
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitPrimary_output_expression_list" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitPrimary_output_expression_list(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitPrimary_output_expression_l); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6404, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6405
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitPrimary_output_expression_list" ):
 *                 return visitor.visitPrimary_output_expression_list(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitPrimary_output_expression_l); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6405, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6405, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6405, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6405, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6405, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":6404
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitPrimary_output_expression_list" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitPrimary_output_expression_list(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6407
 *                 return visitor.visitPrimary_output_expression_list(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6407, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6407, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6407, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6407, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6407, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6407, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":6403
 *                 listener.exitPrimary_output_expression_list(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_output_expression_list" ):
 *                 return visitor.visitPrimary_output_expression_list(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_output_expression_listContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6412
 *     class Primary_unsigned_numberContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Primary_unsigned_numberContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_30Primary_unsigned_numberContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_30Primary_unsigned_numberContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Primary_unsigned_numberContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 6412, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 6412, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 6412, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6412, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_unsigned_numberContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_30Primary_unsigned_numberContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Primary_unsigned_numberContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":6413
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 6413, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6413, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6413, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6413, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6413, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6414
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def UNSIGNED_NUMBER(self):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6414, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6414, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6414, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6414, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6414, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6414, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6412
 *     class Primary_unsigned_numberContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_unsigned_numberContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6416
 *             self.copyFrom(ctx)
 * 
 *         def UNSIGNED_NUMBER(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.UNSIGNED_NUMBER, 0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Primary_unsigned_numberContext_3UNSIGNED_NUMBER(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_30Primary_unsigned_numberContext_3UNSIGNED_NUMBER = {"UNSIGNED_NUMBER", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_30Primary_unsigned_numberContext_3UNSIGNED_NUMBER, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Primary_unsigned_numberContext_3UNSIGNED_NUMBER(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("UNSIGNED_NUMBER (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_30Primary_unsigned_numberContext_2UNSIGNED_NUMBER(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Primary_unsigned_numberContext_2UNSIGNED_NUMBER(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("UNSIGNED_NUMBER", 0);

  /* "pymola/generated/ModelicaParser.pyx":6417
 * 
 *         def UNSIGNED_NUMBER(self):
 *             return self.getToken(ModelicaParser.UNSIGNED_NUMBER, 0)             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getToken); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_UNSIGNED_NUMBER); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6417, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6417, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6417, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6417, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":6416
 *             self.copyFrom(ctx)
 * 
 *         def UNSIGNED_NUMBER(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.UNSIGNED_NUMBER, 0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_unsigned_numberContext.UNSIGNED_NUMBER", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6419
 *             return self.getToken(ModelicaParser.UNSIGNED_NUMBER, 0)
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_unsigned_number" ):
 *                 listener.enterPrimary_unsigned_number(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Primary_unsigned_numberContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_30Primary_unsigned_numberContext_5enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_30Primary_unsigned_numberContext_5enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Primary_unsigned_numberContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 6419, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 6419, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6419, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_unsigned_numberContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_30Primary_unsigned_numberContext_4enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Primary_unsigned_numberContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6420
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterPrimary_unsigned_number" ):             # <<<<<<<<<<<<<<
 *                 listener.enterPrimary_unsigned_number(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterPrimary_unsigned_number); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6420, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6421
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterPrimary_unsigned_number" ):
 *                 listener.enterPrimary_unsigned_number(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterPrimary_unsigned_number); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6421, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6421, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6421, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6421, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6421, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6421, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6420
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterPrimary_unsigned_number" ):             # <<<<<<<<<<<<<<
 *                 listener.enterPrimary_unsigned_number(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6419
 *             return self.getToken(ModelicaParser.UNSIGNED_NUMBER, 0)
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_unsigned_number" ):
 *                 listener.enterPrimary_unsigned_number(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_unsigned_numberContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6423
 *                 listener.enterPrimary_unsigned_number(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_unsigned_number" ):
 *                 listener.exitPrimary_unsigned_number(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Primary_unsigned_numberContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_30Primary_unsigned_numberContext_7exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_30Primary_unsigned_numberContext_7exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Primary_unsigned_numberContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 6423, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 6423, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6423, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_unsigned_numberContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_30Primary_unsigned_numberContext_6exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Primary_unsigned_numberContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6424
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitPrimary_unsigned_number" ):             # <<<<<<<<<<<<<<
 *                 listener.exitPrimary_unsigned_number(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitPrimary_unsigned_number); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6424, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6425
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitPrimary_unsigned_number" ):
 *                 listener.exitPrimary_unsigned_number(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitPrimary_unsigned_number); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6425, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6425, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6425, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6425, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6425, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6425, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6424
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitPrimary_unsigned_number" ):             # <<<<<<<<<<<<<<
 *                 listener.exitPrimary_unsigned_number(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6423
 *                 listener.enterPrimary_unsigned_number(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_unsigned_number" ):
 *                 listener.exitPrimary_unsigned_number(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_unsigned_numberContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6427
 *                 listener.exitPrimary_unsigned_number(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_unsigned_number" ):
 *                 return visitor.visitPrimary_unsigned_number(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Primary_unsigned_numberContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_30Primary_unsigned_numberContext_9accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_30Primary_unsigned_numberContext_9accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Primary_unsigned_numberContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 6427, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 6427, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6427, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_unsigned_numberContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_30Primary_unsigned_numberContext_8accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Primary_unsigned_numberContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":6428
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitPrimary_unsigned_number" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitPrimary_unsigned_number(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitPrimary_unsigned_number); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6428, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6429
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitPrimary_unsigned_number" ):
 *                 return visitor.visitPrimary_unsigned_number(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitPrimary_unsigned_number); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6429, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6429, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6429, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6429, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6429, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":6428
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitPrimary_unsigned_number" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitPrimary_unsigned_number(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6431
 *                 return visitor.visitPrimary_unsigned_number(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6431, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6431, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6431, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6431, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6431, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6431, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":6427
 *                 listener.exitPrimary_unsigned_number(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_unsigned_number" ):
 *                 return visitor.visitPrimary_unsigned_number(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_unsigned_numberContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6436
 *     class Primary_function_argumentsContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_33Primary_function_argumentsContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_33Primary_function_argumentsContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_33Primary_function_argumentsContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_33Primary_function_argumentsContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 6436, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 6436, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 6436, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6436, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_function_argumentsContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_33Primary_function_argumentsContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_33Primary_function_argumentsContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":6437
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 6437, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6437, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6437, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6437, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6437, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6437, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6437, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6437, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6437, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6438
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def function_arguments(self):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6438, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6438, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6438, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6438, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6438, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6438, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6436
 *     class Primary_function_argumentsContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_function_argumentsContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6440
 *             self.copyFrom(ctx)
 * 
 *         def function_arguments(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Function_argumentsContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_33Primary_function_argumentsContext_3function_arguments(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_33Primary_function_argumentsContext_3function_arguments = {"function_arguments", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_33Primary_function_argumentsContext_3function_arguments, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_33Primary_function_argumentsContext_3function_arguments(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("function_arguments (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_33Primary_function_argumentsContext_2function_arguments(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_33Primary_function_argumentsContext_2function_arguments(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("function_arguments", 0);

  /* "pymola/generated/ModelicaParser.pyx":6441
 * 
 *         def function_arguments(self):
 *             return self.getTypedRuleContext(ModelicaParser.Function_argumentsContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6441, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6441, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Function_argumentsContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6441, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6441, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6441, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6441, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6441, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":6440
 *             self.copyFrom(ctx)
 * 
 *         def function_arguments(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Function_argumentsContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_function_argumentsContext.function_arguments", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6444
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_function_arguments" ):
 *                 listener.enterPrimary_function_arguments(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_33Primary_function_argumentsContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_33Primary_function_argumentsContext_5enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_33Primary_function_argumentsContext_5enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_33Primary_function_argumentsContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 6444, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 6444, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6444, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_function_argumentsContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_33Primary_function_argumentsContext_4enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_33Primary_function_argumentsContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6445
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterPrimary_function_arguments" ):             # <<<<<<<<<<<<<<
 *                 listener.enterPrimary_function_arguments(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterPrimary_function_arguments); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6445, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6446
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterPrimary_function_arguments" ):
 *                 listener.enterPrimary_function_arguments(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterPrimary_function_arguments); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6446, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6446, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6446, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6446, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6446, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6446, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6445
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterPrimary_function_arguments" ):             # <<<<<<<<<<<<<<
 *                 listener.enterPrimary_function_arguments(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6444
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_function_arguments" ):
 *                 listener.enterPrimary_function_arguments(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_function_argumentsContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6448
 *                 listener.enterPrimary_function_arguments(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_function_arguments" ):
 *                 listener.exitPrimary_function_arguments(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_33Primary_function_argumentsContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_33Primary_function_argumentsContext_7exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_33Primary_function_argumentsContext_7exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_33Primary_function_argumentsContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 6448, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 6448, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6448, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_function_argumentsContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_33Primary_function_argumentsContext_6exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_33Primary_function_argumentsContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6449
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitPrimary_function_arguments" ):             # <<<<<<<<<<<<<<
 *                 listener.exitPrimary_function_arguments(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitPrimary_function_arguments); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6449, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6450
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitPrimary_function_arguments" ):
 *                 listener.exitPrimary_function_arguments(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitPrimary_function_arguments); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6450, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6450, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6450, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6450, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6450, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6450, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6449
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitPrimary_function_arguments" ):             # <<<<<<<<<<<<<<
 *                 listener.exitPrimary_function_arguments(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6448
 *                 listener.enterPrimary_function_arguments(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_function_arguments" ):
 *                 listener.exitPrimary_function_arguments(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_function_argumentsContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6452
 *                 listener.exitPrimary_function_arguments(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_function_arguments" ):
 *                 return visitor.visitPrimary_function_arguments(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_33Primary_function_argumentsContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_33Primary_function_argumentsContext_9accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_33Primary_function_argumentsContext_9accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_33Primary_function_argumentsContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 6452, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 6452, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6452, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_function_argumentsContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_33Primary_function_argumentsContext_8accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_33Primary_function_argumentsContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":6453
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitPrimary_function_arguments" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitPrimary_function_arguments(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitPrimary_function_arguments); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6453, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6454
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitPrimary_function_arguments" ):
 *                 return visitor.visitPrimary_function_arguments(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitPrimary_function_arguments); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6454, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6454, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6454, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6454, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6454, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6454, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":6453
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitPrimary_function_arguments" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitPrimary_function_arguments(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6456
 *                 return visitor.visitPrimary_function_arguments(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6456, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6456, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6456, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6456, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6456, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6456, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":6452
 *                 listener.exitPrimary_function_arguments(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_function_arguments" ):
 *                 return visitor.visitPrimary_function_arguments(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_function_argumentsContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6461
 *     class Primary_falseContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Primary_falseContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20Primary_falseContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20Primary_falseContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Primary_falseContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 6461, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 6461, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 6461, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6461, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_falseContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20Primary_falseContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Primary_falseContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":6462
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 6462, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6462, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6462, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6462, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6462, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6462, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6463
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6463, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6463, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6463, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6463, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6463, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6463, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6461
 *     class Primary_falseContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_falseContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6466
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_false" ):
 *                 listener.enterPrimary_false(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Primary_falseContext_3enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20Primary_falseContext_3enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20Primary_falseContext_3enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Primary_falseContext_3enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 6466, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 6466, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6466, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_falseContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20Primary_falseContext_2enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Primary_falseContext_2enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6467
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterPrimary_false" ):             # <<<<<<<<<<<<<<
 *                 listener.enterPrimary_false(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterPrimary_false); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6467, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6468
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterPrimary_false" ):
 *                 listener.enterPrimary_false(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterPrimary_false); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6468, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6468, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6468, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6468, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6468, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6468, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6467
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterPrimary_false" ):             # <<<<<<<<<<<<<<
 *                 listener.enterPrimary_false(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6466
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_false" ):
 *                 listener.enterPrimary_false(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_falseContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6470
 *                 listener.enterPrimary_false(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_false" ):
 *                 listener.exitPrimary_false(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Primary_falseContext_5exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20Primary_falseContext_5exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20Primary_falseContext_5exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Primary_falseContext_5exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 6470, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 6470, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6470, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_falseContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20Primary_falseContext_4exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Primary_falseContext_4exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6471
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitPrimary_false" ):             # <<<<<<<<<<<<<<
 *                 listener.exitPrimary_false(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitPrimary_false); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6471, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6472
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitPrimary_false" ):
 *                 listener.exitPrimary_false(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitPrimary_false); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6472, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6472, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6472, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6472, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6472, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6472, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6471
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitPrimary_false" ):             # <<<<<<<<<<<<<<
 *                 listener.exitPrimary_false(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6470
 *                 listener.enterPrimary_false(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_false" ):
 *                 listener.exitPrimary_false(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_falseContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6474
 *                 listener.exitPrimary_false(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_false" ):
 *                 return visitor.visitPrimary_false(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Primary_falseContext_7accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_20Primary_falseContext_7accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_20Primary_falseContext_7accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_20Primary_falseContext_7accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 6474, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 6474, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6474, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_falseContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_20Primary_falseContext_6accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_20Primary_falseContext_6accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":6475
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitPrimary_false" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitPrimary_false(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitPrimary_false); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6475, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6476
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitPrimary_false" ):
 *                 return visitor.visitPrimary_false(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitPrimary_false); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6476, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6476, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6476, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6476, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6476, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6476, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":6475
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitPrimary_false" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitPrimary_false(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6478
 *                 return visitor.visitPrimary_false(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6478, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6478, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6478, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6478, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6478, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6478, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":6474
 *                 listener.exitPrimary_false(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_false" ):
 *                 return visitor.visitPrimary_false(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_falseContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6483
 *     class Primary_derivativeContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Primary_derivativeContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_25Primary_derivativeContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_25Primary_derivativeContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Primary_derivativeContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 6483, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 6483, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 6483, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6483, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_derivativeContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_25Primary_derivativeContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Primary_derivativeContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":6484
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 6484, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6484, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6484, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6484, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6484, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6484, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6485
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def function_call_args(self):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6485, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6485, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6485, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6485, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6485, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6483
 *     class Primary_derivativeContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_derivativeContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6487
 *             self.copyFrom(ctx)
 * 
 *         def function_call_args(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Function_call_argsContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Primary_derivativeContext_3function_call_args(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_25Primary_derivativeContext_3function_call_args = {"function_call_args", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_25Primary_derivativeContext_3function_call_args, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Primary_derivativeContext_3function_call_args(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("function_call_args (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_25Primary_derivativeContext_2function_call_args(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Primary_derivativeContext_2function_call_args(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("function_call_args", 0);

  /* "pymola/generated/ModelicaParser.pyx":6488
 * 
 *         def function_call_args(self):
 *             return self.getTypedRuleContext(ModelicaParser.Function_call_argsContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Function_call_argsContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6488, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6488, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6488, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6488, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":6487
 *             self.copyFrom(ctx)
 * 
 *         def function_call_args(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Function_call_argsContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_derivativeContext.function_call_args", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6491
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_derivative" ):
 *                 listener.enterPrimary_derivative(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Primary_derivativeContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_25Primary_derivativeContext_5enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_25Primary_derivativeContext_5enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Primary_derivativeContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 6491, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 6491, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6491, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_derivativeContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_25Primary_derivativeContext_4enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Primary_derivativeContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6492
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterPrimary_derivative" ):             # <<<<<<<<<<<<<<
 *                 listener.enterPrimary_derivative(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterPrimary_derivative); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6492, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6493
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterPrimary_derivative" ):
 *                 listener.enterPrimary_derivative(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterPrimary_derivative); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6493, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6493, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6493, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6493, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6493, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6493, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6492
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterPrimary_derivative" ):             # <<<<<<<<<<<<<<
 *                 listener.enterPrimary_derivative(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6491
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_derivative" ):
 *                 listener.enterPrimary_derivative(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_derivativeContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6495
 *                 listener.enterPrimary_derivative(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_derivative" ):
 *                 listener.exitPrimary_derivative(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Primary_derivativeContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_25Primary_derivativeContext_7exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_25Primary_derivativeContext_7exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Primary_derivativeContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 6495, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 6495, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6495, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_derivativeContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_25Primary_derivativeContext_6exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Primary_derivativeContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6496
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitPrimary_derivative" ):             # <<<<<<<<<<<<<<
 *                 listener.exitPrimary_derivative(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitPrimary_derivative); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6496, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6497
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitPrimary_derivative" ):
 *                 listener.exitPrimary_derivative(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitPrimary_derivative); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6497, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6497, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6497, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6497, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6497, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6497, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6496
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitPrimary_derivative" ):             # <<<<<<<<<<<<<<
 *                 listener.exitPrimary_derivative(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6495
 *                 listener.enterPrimary_derivative(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_derivative" ):
 *                 listener.exitPrimary_derivative(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_derivativeContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6499
 *                 listener.exitPrimary_derivative(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_derivative" ):
 *                 return visitor.visitPrimary_derivative(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Primary_derivativeContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_25Primary_derivativeContext_9accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_25Primary_derivativeContext_9accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Primary_derivativeContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 6499, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 6499, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6499, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_derivativeContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_25Primary_derivativeContext_8accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Primary_derivativeContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":6500
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitPrimary_derivative" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitPrimary_derivative(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitPrimary_derivative); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6500, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6501
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitPrimary_derivative" ):
 *                 return visitor.visitPrimary_derivative(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitPrimary_derivative); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6501, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6501, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6501, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6501, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6501, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6501, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":6500
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitPrimary_derivative" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitPrimary_derivative(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6503
 *                 return visitor.visitPrimary_derivative(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6503, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6503, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6503, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6503, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6503, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6503, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":6499
 *                 listener.exitPrimary_derivative(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_derivative" ):
 *                 return visitor.visitPrimary_derivative(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_derivativeContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6508
 *     class Primary_component_referenceContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_34Primary_component_referenceContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_34Primary_component_referenceContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_34Primary_component_referenceContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_34Primary_component_referenceContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 6508, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 6508, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 6508, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6508, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_component_referenceContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_34Primary_component_referenceContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_34Primary_component_referenceContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":6509
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 6509, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6509, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6509, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6509, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6509, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6509, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6510
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def component_reference(self):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6510, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6510, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6510, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6510, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6508
 *     class Primary_component_referenceContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_component_referenceContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6512
 *             self.copyFrom(ctx)
 * 
 *         def component_reference(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_34Primary_component_referenceContext_3component_reference(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_34Primary_component_referenceContext_3component_reference = {"component_reference", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_34Primary_component_referenceContext_3component_reference, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_34Primary_component_referenceContext_3component_reference(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("component_reference (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_34Primary_component_referenceContext_2component_reference(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_34Primary_component_referenceContext_2component_reference(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("component_reference", 0);

  /* "pymola/generated/ModelicaParser.pyx":6513
 * 
 *         def component_reference(self):
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Component_referenceContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6513, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6513, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6513, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6513, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":6512
 *             self.copyFrom(ctx)
 * 
 *         def component_reference(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_component_referenceContext.component_reference", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6516
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_component_reference" ):
 *                 listener.enterPrimary_component_reference(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_34Primary_component_referenceContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_34Primary_component_referenceContext_5enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_34Primary_component_referenceContext_5enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_34Primary_component_referenceContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 6516, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 6516, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6516, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_component_referenceContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_34Primary_component_referenceContext_4enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_34Primary_component_referenceContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6517
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterPrimary_component_reference" ):             # <<<<<<<<<<<<<<
 *                 listener.enterPrimary_component_reference(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterPrimary_component_reference); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6517, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6518
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterPrimary_component_reference" ):
 *                 listener.enterPrimary_component_reference(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterPrimary_component_reference); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6518, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6518, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6518, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6518, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6518, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6518, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6517
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterPrimary_component_reference" ):             # <<<<<<<<<<<<<<
 *                 listener.enterPrimary_component_reference(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6516
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_component_reference" ):
 *                 listener.enterPrimary_component_reference(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_component_referenceContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6520
 *                 listener.enterPrimary_component_reference(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_component_reference" ):
 *                 listener.exitPrimary_component_reference(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_34Primary_component_referenceContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_34Primary_component_referenceContext_7exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_34Primary_component_referenceContext_7exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_34Primary_component_referenceContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 6520, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 6520, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6520, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_component_referenceContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_34Primary_component_referenceContext_6exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_34Primary_component_referenceContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6521
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitPrimary_component_reference" ):             # <<<<<<<<<<<<<<
 *                 listener.exitPrimary_component_reference(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitPrimary_component_reference); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6521, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6522
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitPrimary_component_reference" ):
 *                 listener.exitPrimary_component_reference(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitPrimary_component_reference); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6522, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6522, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6522, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6522, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6522, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6522, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6521
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitPrimary_component_reference" ):             # <<<<<<<<<<<<<<
 *                 listener.exitPrimary_component_reference(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6520
 *                 listener.enterPrimary_component_reference(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_component_reference" ):
 *                 listener.exitPrimary_component_reference(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_component_referenceContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6524
 *                 listener.exitPrimary_component_reference(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_component_reference" ):
 *                 return visitor.visitPrimary_component_reference(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_34Primary_component_referenceContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_34Primary_component_referenceContext_9accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_34Primary_component_referenceContext_9accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_34Primary_component_referenceContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 6524, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 6524, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6524, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_component_referenceContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_34Primary_component_referenceContext_8accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_34Primary_component_referenceContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":6525
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitPrimary_component_reference" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitPrimary_component_reference(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitPrimary_component_reference); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6525, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6526
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitPrimary_component_reference" ):
 *                 return visitor.visitPrimary_component_reference(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitPrimary_component_reference); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6526, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6526, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6526, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6526, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6526, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6526, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":6525
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitPrimary_component_reference" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitPrimary_component_reference(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6528
 *                 return visitor.visitPrimary_component_reference(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6528, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6528, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6528, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6528, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6528, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6528, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":6524
 *                 listener.exitPrimary_component_reference(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_component_reference" ):
 *                 return visitor.visitPrimary_component_reference(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_component_referenceContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6533
 *     class Primary_expression_listContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Primary_expression_listContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_30Primary_expression_listContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_30Primary_expression_listContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Primary_expression_listContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 6533, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 6533, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 6533, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6533, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_expression_listContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_30Primary_expression_listContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Primary_expression_listContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":6534
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 6534, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6534, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6534, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6534, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6534, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6534, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6535
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def expression_list(self, i:int=None):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6535, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6535, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6535, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6535, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6535, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6533
 *     class Primary_expression_listContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_expression_listContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6537
 *             self.copyFrom(ctx)
 * 
 *         def expression_list(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Expression_listContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Primary_expression_listContext_3expression_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_30Primary_expression_listContext_3expression_list = {"expression_list", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_30Primary_expression_listContext_3expression_list, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Primary_expression_listContext_3expression_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("expression_list (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "expression_list") < 0)) __PYX_ERR(0, 6537, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("expression_list", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6537, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_expression_listContext.expression_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_30Primary_expression_listContext_2expression_list(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Primary_expression_listContext_2expression_list(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("expression_list", 0);

  /* "pymola/generated/ModelicaParser.pyx":6538
 * 
 *         def expression_list(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.Expression_listContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6539
 *         def expression_list(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Expression_listContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.Expression_listContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6539, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6539, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_Expression_listContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6539, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6539, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6539, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6539, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6539, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6539, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":6538
 * 
 *         def expression_list(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.Expression_listContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6541
 *                 return self.getTypedRuleContexts(ModelicaParser.Expression_listContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.Expression_listContext,i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Expression_listContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6541, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6541, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6541, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6541, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":6537
 *             self.copyFrom(ctx)
 * 
 *         def expression_list(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Expression_listContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_expression_listContext.expression_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6544
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_expression_list" ):
 *                 listener.enterPrimary_expression_list(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Primary_expression_listContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_30Primary_expression_listContext_5enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_30Primary_expression_listContext_5enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Primary_expression_listContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 6544, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 6544, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6544, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_expression_listContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_30Primary_expression_listContext_4enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Primary_expression_listContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6545
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterPrimary_expression_list" ):             # <<<<<<<<<<<<<<
 *                 listener.enterPrimary_expression_list(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterPrimary_expression_list); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6545, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6546
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterPrimary_expression_list" ):
 *                 listener.enterPrimary_expression_list(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterPrimary_expression_list); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6546, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6546, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6546, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6546, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6546, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6546, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6545
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterPrimary_expression_list" ):             # <<<<<<<<<<<<<<
 *                 listener.enterPrimary_expression_list(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6544
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_expression_list" ):
 *                 listener.enterPrimary_expression_list(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_expression_listContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6548
 *                 listener.enterPrimary_expression_list(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_expression_list" ):
 *                 listener.exitPrimary_expression_list(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Primary_expression_listContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_30Primary_expression_listContext_7exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_30Primary_expression_listContext_7exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Primary_expression_listContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 6548, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 6548, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6548, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_expression_listContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_30Primary_expression_listContext_6exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Primary_expression_listContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6549
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitPrimary_expression_list" ):             # <<<<<<<<<<<<<<
 *                 listener.exitPrimary_expression_list(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitPrimary_expression_list); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6549, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6550
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitPrimary_expression_list" ):
 *                 listener.exitPrimary_expression_list(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitPrimary_expression_list); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6550, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6550, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6550, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6550, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6550, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6550, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6549
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitPrimary_expression_list" ):             # <<<<<<<<<<<<<<
 *                 listener.exitPrimary_expression_list(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6548
 *                 listener.enterPrimary_expression_list(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_expression_list" ):
 *                 listener.exitPrimary_expression_list(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_expression_listContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6552
 *                 listener.exitPrimary_expression_list(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_expression_list" ):
 *                 return visitor.visitPrimary_expression_list(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Primary_expression_listContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_30Primary_expression_listContext_9accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_30Primary_expression_listContext_9accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_30Primary_expression_listContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 6552, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 6552, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6552, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_expression_listContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_30Primary_expression_listContext_8accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_30Primary_expression_listContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":6553
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitPrimary_expression_list" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitPrimary_expression_list(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitPrimary_expression_list); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6553, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6554
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitPrimary_expression_list" ):
 *                 return visitor.visitPrimary_expression_list(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitPrimary_expression_list); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6554, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6554, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6554, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6554, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6554, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6554, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":6553
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitPrimary_expression_list" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitPrimary_expression_list(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6556
 *                 return visitor.visitPrimary_expression_list(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6556, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6556, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6556, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6556, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6556, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6556, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":6552
 *                 listener.exitPrimary_expression_list(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_expression_list" ):
 *                 return visitor.visitPrimary_expression_list(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_expression_listContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6561
 *     class Primary_trueContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Primary_trueContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19Primary_trueContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19Primary_trueContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Primary_trueContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 6561, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 6561, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 6561, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6561, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_trueContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19Primary_trueContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Primary_trueContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":6562
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 6562, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6562, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6562, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6562, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6562, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6562, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6562, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6562, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6562, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6563
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6563, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6563, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6563, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6563, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6563, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6561
 *     class Primary_trueContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_trueContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6566
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_true" ):
 *                 listener.enterPrimary_true(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Primary_trueContext_3enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19Primary_trueContext_3enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19Primary_trueContext_3enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Primary_trueContext_3enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 6566, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 6566, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6566, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_trueContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19Primary_trueContext_2enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Primary_trueContext_2enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6567
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterPrimary_true" ):             # <<<<<<<<<<<<<<
 *                 listener.enterPrimary_true(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterPrimary_true); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6567, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6568
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterPrimary_true" ):
 *                 listener.enterPrimary_true(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterPrimary_true); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6568, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6568, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6568, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6568, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6568, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6568, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6567
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterPrimary_true" ):             # <<<<<<<<<<<<<<
 *                 listener.enterPrimary_true(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6566
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_true" ):
 *                 listener.enterPrimary_true(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_trueContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6570
 *                 listener.enterPrimary_true(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_true" ):
 *                 listener.exitPrimary_true(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Primary_trueContext_5exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19Primary_trueContext_5exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19Primary_trueContext_5exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Primary_trueContext_5exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 6570, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 6570, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6570, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_trueContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19Primary_trueContext_4exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Primary_trueContext_4exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6571
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitPrimary_true" ):             # <<<<<<<<<<<<<<
 *                 listener.exitPrimary_true(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitPrimary_true); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6571, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6572
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitPrimary_true" ):
 *                 listener.exitPrimary_true(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitPrimary_true); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6572, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6572, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6572, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6572, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6572, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6572, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6571
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitPrimary_true" ):             # <<<<<<<<<<<<<<
 *                 listener.exitPrimary_true(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6570
 *                 listener.enterPrimary_true(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_true" ):
 *                 listener.exitPrimary_true(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_trueContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6574
 *                 listener.exitPrimary_true(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_true" ):
 *                 return visitor.visitPrimary_true(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Primary_trueContext_7accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_19Primary_trueContext_7accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_19Primary_trueContext_7accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_19Primary_trueContext_7accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 6574, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 6574, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6574, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_trueContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_19Primary_trueContext_6accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_19Primary_trueContext_6accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":6575
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitPrimary_true" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitPrimary_true(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitPrimary_true); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6575, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6576
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitPrimary_true" ):
 *                 return visitor.visitPrimary_true(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitPrimary_true); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6576, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6576, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6576, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6576, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6576, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6576, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":6575
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitPrimary_true" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitPrimary_true(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6578
 *                 return visitor.visitPrimary_true(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6578, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6578, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6578, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6578, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6578, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6578, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":6574
 *                 listener.exitPrimary_true(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_true" ):
 *                 return visitor.visitPrimary_true(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_trueContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6583
 *     class Primary_functionContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Primary_functionContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Primary_functionContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Primary_functionContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Primary_functionContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 6583, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 6583, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 6583, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6583, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_functionContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Primary_functionContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Primary_functionContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":6584
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 6584, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6584, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6584, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6584, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6584, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6584, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6585
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def component_reference(self):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6585, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6585, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6585, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6585, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6585, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6583
 *     class Primary_functionContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_functionContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6587
 *             self.copyFrom(ctx)
 * 
 *         def component_reference(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Primary_functionContext_3component_reference(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Primary_functionContext_3component_reference = {"component_reference", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Primary_functionContext_3component_reference, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Primary_functionContext_3component_reference(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("component_reference (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Primary_functionContext_2component_reference(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Primary_functionContext_2component_reference(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("component_reference", 0);

  /* "pymola/generated/ModelicaParser.pyx":6588
 * 
 *         def component_reference(self):
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)             # <<<<<<<<<<<<<<
 * 
 *         def function_call_args(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Component_referenceContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6588, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6588, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6588, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6588, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":6587
 *             self.copyFrom(ctx)
 * 
 *         def component_reference(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_functionContext.component_reference", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6590
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 *         def function_call_args(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Function_call_argsContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Primary_functionContext_5function_call_args(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Primary_functionContext_5function_call_args = {"function_call_args", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Primary_functionContext_5function_call_args, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Primary_functionContext_5function_call_args(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("function_call_args (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Primary_functionContext_4function_call_args(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Primary_functionContext_4function_call_args(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("function_call_args", 0);

  /* "pymola/generated/ModelicaParser.pyx":6591
 * 
 *         def function_call_args(self):
 *             return self.getTypedRuleContext(ModelicaParser.Function_call_argsContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Function_call_argsContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6591, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6591, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6591, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6591, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":6590
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 *         def function_call_args(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Function_call_argsContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_functionContext.function_call_args", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6594
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_function" ):
 *                 listener.enterPrimary_function(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Primary_functionContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Primary_functionContext_7enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Primary_functionContext_7enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Primary_functionContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 6594, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 6594, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6594, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_functionContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Primary_functionContext_6enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Primary_functionContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6595
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterPrimary_function" ):             # <<<<<<<<<<<<<<
 *                 listener.enterPrimary_function(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterPrimary_function); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6595, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6596
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterPrimary_function" ):
 *                 listener.enterPrimary_function(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterPrimary_function); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6596, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6596, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6596, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6596, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6596, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6596, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6595
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterPrimary_function" ):             # <<<<<<<<<<<<<<
 *                 listener.enterPrimary_function(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6594
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_function" ):
 *                 listener.enterPrimary_function(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_functionContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6598
 *                 listener.enterPrimary_function(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_function" ):
 *                 listener.exitPrimary_function(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Primary_functionContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Primary_functionContext_9exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Primary_functionContext_9exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Primary_functionContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 6598, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 6598, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6598, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_functionContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Primary_functionContext_8exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Primary_functionContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6599
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitPrimary_function" ):             # <<<<<<<<<<<<<<
 *                 listener.exitPrimary_function(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitPrimary_function); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6599, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6600
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitPrimary_function" ):
 *                 listener.exitPrimary_function(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitPrimary_function); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6600, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6600, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6600, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6600, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6600, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6600, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6599
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitPrimary_function" ):             # <<<<<<<<<<<<<<
 *                 listener.exitPrimary_function(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6598
 *                 listener.enterPrimary_function(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_function" ):
 *                 listener.exitPrimary_function(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_functionContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6602
 *                 listener.exitPrimary_function(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_function" ):
 *                 return visitor.visitPrimary_function(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Primary_functionContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Primary_functionContext_11accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Primary_functionContext_11accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Primary_functionContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 6602, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 6602, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6602, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_functionContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Primary_functionContext_10accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Primary_functionContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":6603
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitPrimary_function" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitPrimary_function(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitPrimary_function); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6603, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6604
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitPrimary_function" ):
 *                 return visitor.visitPrimary_function(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitPrimary_function); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6604, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6604, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6604, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6604, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6604, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6604, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":6603
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitPrimary_function" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitPrimary_function(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6606
 *                 return visitor.visitPrimary_function(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6606, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6606, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6606, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6606, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6606, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6606, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":6602
 *                 listener.exitPrimary_function(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_function" ):
 *                 return visitor.visitPrimary_function(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_functionContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6611
 *     class Primary_initialContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Primary_initialContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Primary_initialContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Primary_initialContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Primary_initialContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 6611, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 6611, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 6611, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6611, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_initialContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Primary_initialContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Primary_initialContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":6612
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 6612, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6612, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6612, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6612, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6612, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6612, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6613
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def function_call_args(self):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6613, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6613, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6613, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6613, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6613, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6611
 *     class Primary_initialContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_initialContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6615
 *             self.copyFrom(ctx)
 * 
 *         def function_call_args(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Function_call_argsContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Primary_initialContext_3function_call_args(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Primary_initialContext_3function_call_args = {"function_call_args", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Primary_initialContext_3function_call_args, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Primary_initialContext_3function_call_args(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("function_call_args (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Primary_initialContext_2function_call_args(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Primary_initialContext_2function_call_args(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("function_call_args", 0);

  /* "pymola/generated/ModelicaParser.pyx":6616
 * 
 *         def function_call_args(self):
 *             return self.getTypedRuleContext(ModelicaParser.Function_call_argsContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Function_call_argsContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6616, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6616, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6616, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6616, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":6615
 *             self.copyFrom(ctx)
 * 
 *         def function_call_args(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Function_call_argsContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_initialContext.function_call_args", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6619
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_initial" ):
 *                 listener.enterPrimary_initial(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Primary_initialContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Primary_initialContext_5enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Primary_initialContext_5enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Primary_initialContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 6619, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 6619, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6619, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_initialContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Primary_initialContext_4enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Primary_initialContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6620
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterPrimary_initial" ):             # <<<<<<<<<<<<<<
 *                 listener.enterPrimary_initial(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterPrimary_initial); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6620, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6621
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterPrimary_initial" ):
 *                 listener.enterPrimary_initial(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterPrimary_initial); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6621, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6621, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6621, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6621, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6621, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6621, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6620
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterPrimary_initial" ):             # <<<<<<<<<<<<<<
 *                 listener.enterPrimary_initial(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6619
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_initial" ):
 *                 listener.enterPrimary_initial(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_initialContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6623
 *                 listener.enterPrimary_initial(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_initial" ):
 *                 listener.exitPrimary_initial(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Primary_initialContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Primary_initialContext_7exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Primary_initialContext_7exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Primary_initialContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 6623, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 6623, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6623, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_initialContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Primary_initialContext_6exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Primary_initialContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6624
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitPrimary_initial" ):             # <<<<<<<<<<<<<<
 *                 listener.exitPrimary_initial(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitPrimary_initial); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6624, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6625
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitPrimary_initial" ):
 *                 listener.exitPrimary_initial(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitPrimary_initial); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6625, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6625, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6625, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6625, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6625, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6625, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6624
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitPrimary_initial" ):             # <<<<<<<<<<<<<<
 *                 listener.exitPrimary_initial(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6623
 *                 listener.enterPrimary_initial(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_initial" ):
 *                 listener.exitPrimary_initial(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_initialContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6627
 *                 listener.exitPrimary_initial(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_initial" ):
 *                 return visitor.visitPrimary_initial(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Primary_initialContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Primary_initialContext_9accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Primary_initialContext_9accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Primary_initialContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 6627, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 6627, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6627, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_initialContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Primary_initialContext_8accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Primary_initialContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":6628
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitPrimary_initial" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitPrimary_initial(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitPrimary_initial); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6628, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6629
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitPrimary_initial" ):
 *                 return visitor.visitPrimary_initial(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitPrimary_initial); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6629, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6629, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6629, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6629, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6629, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6629, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":6628
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitPrimary_initial" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitPrimary_initial(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6631
 *                 return visitor.visitPrimary_initial(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6631, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6631, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6631, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6631, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6631, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6631, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":6627
 *                 listener.exitPrimary_initial(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_initial" ):
 *                 return visitor.visitPrimary_initial(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Primary_initialContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6635
 * 
 * 
 *     def primary(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.PrimaryContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_117primary(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_117primary = {"primary", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_117primary, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_117primary(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("primary (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_116primary(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_116primary(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v_la_ = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  char const *__pyx_t_21;
  __Pyx_RefNannySetupContext("primary", 0);

  /* "pymola/generated/ModelicaParser.pyx":6637
 *     def primary(self):
 * 
 *         localctx = ModelicaParser.PrimaryContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 114, self.RULE_primary)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_PrimaryContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6637, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6637, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6637, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6637, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6638
 * 
 *         localctx = ModelicaParser.PrimaryContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 114, self.RULE_primary)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6638, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_primary); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6638, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_114, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6638, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_114, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6638, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6638, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_114);
    __Pyx_GIVEREF(__pyx_int_114);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_114);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6638, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6639
 *         localctx = ModelicaParser.PrimaryContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 114, self.RULE_primary)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.state = 916
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 6639, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6640
 *         self.enterRule(localctx, 114, self.RULE_primary)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.state = 916
 *             self._errHandler.sync(self)
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":6641
 *         self._la = 0 # Token type
 *         try:
 *             self.state = 916             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             la_ = self._interp.adaptivePredict(self._input,113,self._ctx)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_916) < 0) __PYX_ERR(0, 6641, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":6642
 *         try:
 *             self.state = 916
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             la_ = self._interp.adaptivePredict(self._input,113,self._ctx)
 *             if la_ == 1:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6642, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6642, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6642, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6642, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6642, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6642, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6642, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":6643
 *             self.state = 916
 *             self._errHandler.sync(self)
 *             la_ = self._interp.adaptivePredict(self._input,113,self._ctx)             # <<<<<<<<<<<<<<
 *             if la_ == 1:
 *                 localctx = ModelicaParser.Primary_unsigned_numberContext(self, localctx)
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_interp); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6643, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_adaptivePredict); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6643, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6643, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6643, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_t_2, __pyx_int_113, __pyx_t_3};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6643, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_t_2, __pyx_int_113, __pyx_t_3};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6643, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6643, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_2);
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_t_2);
          __Pyx_INCREF(__pyx_int_113);
          __Pyx_GIVEREF(__pyx_int_113);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_int_113);
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_6, __pyx_t_3);
          __pyx_t_2 = 0;
          __pyx_t_3 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6643, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v_la_ = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":6644
 *             self._errHandler.sync(self)
 *             la_ = self._interp.adaptivePredict(self._input,113,self._ctx)
 *             if la_ == 1:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Primary_unsigned_numberContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 1)
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6644, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 6644, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":6645
 *             la_ = self._interp.adaptivePredict(self._input,113,self._ctx)
 *             if la_ == 1:
 *                 localctx = ModelicaParser.Primary_unsigned_numberContext(self, localctx)             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 1)
 *                 self.state = 884
 */
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6645, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Primary_unsigned_numberContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6645, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6645, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6645, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6645, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_7) {
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_v_localctx);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6645, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6646
 *             if la_ == 1:
 *                 localctx = ModelicaParser.Primary_unsigned_numberContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *                 self.state = 884
 *                 self.match(ModelicaParser.UNSIGNED_NUMBER)
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6646, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_1};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6646, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_1};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6646, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6646, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_1);
            __Pyx_GIVEREF(__pyx_int_1);
            PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6646, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6647
 *                 localctx = ModelicaParser.Primary_unsigned_numberContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 1)
 *                 self.state = 884             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.UNSIGNED_NUMBER)
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_884) < 0) __PYX_ERR(0, 6647, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6648
 *                 self.enterOuterAlt(localctx, 1)
 *                 self.state = 884
 *                 self.match(ModelicaParser.UNSIGNED_NUMBER)             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6648, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6648, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_UNSIGNED_NUMBER); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6648, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6648, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_5)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6648, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6648, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6648, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_7); __pyx_t_7 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6648, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6644
 *             self._errHandler.sync(self)
 *             la_ = self._interp.adaptivePredict(self._input,113,self._ctx)
 *             if la_ == 1:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Primary_unsigned_numberContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 1)
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":6651
 *                 pass
 * 
 *             elif la_ == 2:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Primary_stringContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 2)
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6651, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 6651, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":6652
 * 
 *             elif la_ == 2:
 *                 localctx = ModelicaParser.Primary_stringContext(self, localctx)             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 2)
 *                 self.state = 885
 */
          __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6652, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_Primary_stringContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6652, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6652, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6652, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6652, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_v_localctx);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6652, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6653
 *             elif la_ == 2:
 *                 localctx = ModelicaParser.Primary_stringContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 2)             # <<<<<<<<<<<<<<
 *                 self.state = 885
 *                 self.match(ModelicaParser.STRING)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6653, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6653, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6653, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6653, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_2);
            __Pyx_GIVEREF(__pyx_int_2);
            PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_int_2);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6653, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6654
 *                 localctx = ModelicaParser.Primary_stringContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 2)
 *                 self.state = 885             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.STRING)
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_885) < 0) __PYX_ERR(0, 6654, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6655
 *                 self.enterOuterAlt(localctx, 2)
 *                 self.state = 885
 *                 self.match(ModelicaParser.STRING)             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6655, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6655, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_STRING); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6655, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_5) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6655, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6655, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6655, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6655, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6655, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6651
 *                 pass
 * 
 *             elif la_ == 2:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Primary_stringContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 2)
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":6658
 *                 pass
 * 
 *             elif la_ == 3:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Primary_falseContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 3)
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_3, 3, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6658, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 6658, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":6659
 * 
 *             elif la_ == 3:
 *                 localctx = ModelicaParser.Primary_falseContext(self, localctx)             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 3)
 *                 self.state = 886
 */
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6659, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Primary_falseContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6659, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6659, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6659, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6659, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_2) {
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_v_localctx);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6659, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6660
 *             elif la_ == 3:
 *                 localctx = ModelicaParser.Primary_falseContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 3)             # <<<<<<<<<<<<<<
 *                 self.state = 886
 *                 self.match(ModelicaParser.T__73)
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6660, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_3};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6660, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_3};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6660, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6660, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_3);
            __Pyx_GIVEREF(__pyx_int_3);
            PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_3);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6660, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6661
 *                 localctx = ModelicaParser.Primary_falseContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 3)
 *                 self.state = 886             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__73)
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_886) < 0) __PYX_ERR(0, 6661, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6662
 *                 self.enterOuterAlt(localctx, 3)
 *                 self.state = 886
 *                 self.match(ModelicaParser.T__73)             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6662, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6662, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__73); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6662, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6662, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6662, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6662, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6662, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6662, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6658
 *                 pass
 * 
 *             elif la_ == 3:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Primary_falseContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 3)
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":6665
 *                 pass
 * 
 *             elif la_ == 4:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Primary_trueContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 4)
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_4, 4, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6665, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 6665, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":6666
 * 
 *             elif la_ == 4:
 *                 localctx = ModelicaParser.Primary_trueContext(self, localctx)             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 4)
 *                 self.state = 887
 */
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6666, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Primary_trueContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6666, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6666, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6666, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6666, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_7) {
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_v_localctx);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6666, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6667
 *             elif la_ == 4:
 *                 localctx = ModelicaParser.Primary_trueContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 4)             # <<<<<<<<<<<<<<
 *                 self.state = 887
 *                 self.match(ModelicaParser.T__74)
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6667, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_4};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6667, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_4};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6667, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6667, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_4);
            __Pyx_GIVEREF(__pyx_int_4);
            PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_4);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6667, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6668
 *                 localctx = ModelicaParser.Primary_trueContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 4)
 *                 self.state = 887             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__74)
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_887) < 0) __PYX_ERR(0, 6668, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6669
 *                 self.enterOuterAlt(localctx, 4)
 *                 self.state = 887
 *                 self.match(ModelicaParser.T__74)             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6669, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6669, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__74); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6669, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6669, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_5)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6669, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6669, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6669, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_7); __pyx_t_7 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6669, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6665
 *                 pass
 * 
 *             elif la_ == 4:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Primary_trueContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 4)
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":6672
 *                 pass
 * 
 *             elif la_ == 5:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Primary_functionContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 5)
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_5, 5, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6672, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 6672, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":6673
 * 
 *             elif la_ == 5:
 *                 localctx = ModelicaParser.Primary_functionContext(self, localctx)             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 5)
 *                 self.state = 888
 */
          __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6673, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_Primary_functionContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6673, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6673, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6673, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6673, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_v_localctx);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6673, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6674
 *             elif la_ == 5:
 *                 localctx = ModelicaParser.Primary_functionContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 5)             # <<<<<<<<<<<<<<
 *                 self.state = 888
 *                 self.component_reference()
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6674, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_5};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6674, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_5};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6674, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6674, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_5);
            __Pyx_GIVEREF(__pyx_int_5);
            PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_int_5);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6674, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6675
 *                 localctx = ModelicaParser.Primary_functionContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 5)
 *                 self.state = 888             # <<<<<<<<<<<<<<
 *                 self.component_reference()
 *                 self.state = 889
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_888) < 0) __PYX_ERR(0, 6675, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6676
 *                 self.enterOuterAlt(localctx, 5)
 *                 self.state = 888
 *                 self.component_reference()             # <<<<<<<<<<<<<<
 *                 self.state = 889
 *                 self.function_call_args()
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_component_reference); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6676, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (__pyx_t_5) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6676, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6676, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6677
 *                 self.state = 888
 *                 self.component_reference()
 *                 self.state = 889             # <<<<<<<<<<<<<<
 *                 self.function_call_args()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_889) < 0) __PYX_ERR(0, 6677, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6678
 *                 self.component_reference()
 *                 self.state = 889
 *                 self.function_call_args()             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_function_call_args); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6678, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (__pyx_t_5) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6678, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6678, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6672
 *                 pass
 * 
 *             elif la_ == 5:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Primary_functionContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 5)
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":6681
 *                 pass
 * 
 *             elif la_ == 6:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Primary_derivativeContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 6)
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_6, 6, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6681, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 6681, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":6682
 * 
 *             elif la_ == 6:
 *                 localctx = ModelicaParser.Primary_derivativeContext(self, localctx)             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 6)
 *                 self.state = 891
 */
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6682, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Primary_derivativeContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6682, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6682, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6682, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6682, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_2) {
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_v_localctx);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6682, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6683
 *             elif la_ == 6:
 *                 localctx = ModelicaParser.Primary_derivativeContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 6)             # <<<<<<<<<<<<<<
 *                 self.state = 891
 *                 self.match(ModelicaParser.T__19)
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6683, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_6};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6683, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_6};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6683, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6683, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_6);
            __Pyx_GIVEREF(__pyx_int_6);
            PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_6);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6683, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6684
 *                 localctx = ModelicaParser.Primary_derivativeContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 6)
 *                 self.state = 891             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__19)
 *                 self.state = 892
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_891) < 0) __PYX_ERR(0, 6684, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6685
 *                 self.enterOuterAlt(localctx, 6)
 *                 self.state = 891
 *                 self.match(ModelicaParser.T__19)             # <<<<<<<<<<<<<<
 *                 self.state = 892
 *                 self.function_call_args()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6685, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6685, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__19); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6685, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6685, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_5)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6685, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6685, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6685, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6685, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6686
 *                 self.state = 891
 *                 self.match(ModelicaParser.T__19)
 *                 self.state = 892             # <<<<<<<<<<<<<<
 *                 self.function_call_args()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_892) < 0) __PYX_ERR(0, 6686, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6687
 *                 self.match(ModelicaParser.T__19)
 *                 self.state = 892
 *                 self.function_call_args()             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_function_call_args); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6687, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
            }
          }
          if (__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6687, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6687, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6681
 *                 pass
 * 
 *             elif la_ == 6:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Primary_derivativeContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 6)
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":6690
 *                 pass
 * 
 *             elif la_ == 7:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Primary_initialContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 7)
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_7, 7, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6690, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 6690, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":6691
 * 
 *             elif la_ == 7:
 *                 localctx = ModelicaParser.Primary_initialContext(self, localctx)             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 7)
 *                 self.state = 893
 */
          __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6691, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_Primary_initialContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6691, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6691, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6691, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6691, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_v_localctx);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6691, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6692
 *             elif la_ == 7:
 *                 localctx = ModelicaParser.Primary_initialContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 7)             # <<<<<<<<<<<<<<
 *                 self.state = 893
 *                 self.match(ModelicaParser.INITIAL)
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6692, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_7};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6692, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_7};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6692, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6692, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_7);
            __Pyx_GIVEREF(__pyx_int_7);
            PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_int_7);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6692, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6693
 *                 localctx = ModelicaParser.Primary_initialContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 7)
 *                 self.state = 893             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.INITIAL)
 *                 self.state = 894
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_893) < 0) __PYX_ERR(0, 6693, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6694
 *                 self.enterOuterAlt(localctx, 7)
 *                 self.state = 893
 *                 self.match(ModelicaParser.INITIAL)             # <<<<<<<<<<<<<<
 *                 self.state = 894
 *                 self.function_call_args()
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6694, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6694, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_INITIAL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6694, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_5) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6694, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6694, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6694, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6694, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5); __pyx_t_5 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6694, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6695
 *                 self.state = 893
 *                 self.match(ModelicaParser.INITIAL)
 *                 self.state = 894             # <<<<<<<<<<<<<<
 *                 self.function_call_args()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_894) < 0) __PYX_ERR(0, 6695, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6696
 *                 self.match(ModelicaParser.INITIAL)
 *                 self.state = 894
 *                 self.function_call_args()             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_function_call_args); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6696, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6696, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6696, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6690
 *                 pass
 * 
 *             elif la_ == 7:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Primary_initialContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 7)
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":6699
 *                 pass
 * 
 *             elif la_ == 8:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Primary_component_referenceContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 8)
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_8, 8, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6699, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 6699, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":6700
 * 
 *             elif la_ == 8:
 *                 localctx = ModelicaParser.Primary_component_referenceContext(self, localctx)             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 8)
 *                 self.state = 895
 */
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6700, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Primary_component_referenceConte); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6700, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6700, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6700, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6700, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_7) {
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_v_localctx);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6700, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6701
 *             elif la_ == 8:
 *                 localctx = ModelicaParser.Primary_component_referenceContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 8)             # <<<<<<<<<<<<<<
 *                 self.state = 895
 *                 self.component_reference()
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6701, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_8};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6701, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_8};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6701, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6701, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_8);
            __Pyx_GIVEREF(__pyx_int_8);
            PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_8);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6701, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6702
 *                 localctx = ModelicaParser.Primary_component_referenceContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 8)
 *                 self.state = 895             # <<<<<<<<<<<<<<
 *                 self.component_reference()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_895) < 0) __PYX_ERR(0, 6702, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6703
 *                 self.enterOuterAlt(localctx, 8)
 *                 self.state = 895
 *                 self.component_reference()             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_component_reference); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6703, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6703, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6703, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6699
 *                 pass
 * 
 *             elif la_ == 8:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Primary_component_referenceContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 8)
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":6706
 *                 pass
 * 
 *             elif la_ == 9:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Primary_output_expression_listContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 9)
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_9, 9, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6706, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 6706, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":6707
 * 
 *             elif la_ == 9:
 *                 localctx = ModelicaParser.Primary_output_expression_listContext(self, localctx)             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 9)
 *                 self.state = 896
 */
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6707, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Primary_output_expression_listCo); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6707, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6707, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6707, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6707, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_2) {
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_v_localctx);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6707, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6708
 *             elif la_ == 9:
 *                 localctx = ModelicaParser.Primary_output_expression_listContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 9)             # <<<<<<<<<<<<<<
 *                 self.state = 896
 *                 self.match(ModelicaParser.T__16)
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6708, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_9};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6708, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_9};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6708, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6708, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_9);
            __Pyx_GIVEREF(__pyx_int_9);
            PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_9);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6708, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6709
 *                 localctx = ModelicaParser.Primary_output_expression_listContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 9)
 *                 self.state = 896             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__16)
 *                 self.state = 897
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_896) < 0) __PYX_ERR(0, 6709, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6710
 *                 self.enterOuterAlt(localctx, 9)
 *                 self.state = 896
 *                 self.match(ModelicaParser.T__16)             # <<<<<<<<<<<<<<
 *                 self.state = 897
 *                 self.output_expression_list()
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6710, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6710, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__16); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6710, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6710, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6710, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6710, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6710, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6710, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6711
 *                 self.state = 896
 *                 self.match(ModelicaParser.T__16)
 *                 self.state = 897             # <<<<<<<<<<<<<<
 *                 self.output_expression_list()
 *                 self.state = 898
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_897) < 0) __PYX_ERR(0, 6711, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6712
 *                 self.match(ModelicaParser.T__16)
 *                 self.state = 897
 *                 self.output_expression_list()             # <<<<<<<<<<<<<<
 *                 self.state = 898
 *                 self.match(ModelicaParser.T__18)
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_output_expression_list); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6712, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (__pyx_t_5) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6712, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6712, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6713
 *                 self.state = 897
 *                 self.output_expression_list()
 *                 self.state = 898             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__18)
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_898) < 0) __PYX_ERR(0, 6713, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6714
 *                 self.output_expression_list()
 *                 self.state = 898
 *                 self.match(ModelicaParser.T__18)             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6714, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6714, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_T__18); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6714, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_5) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6714, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6714, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6714, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6714, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5); __pyx_t_5 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6714, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6706
 *                 pass
 * 
 *             elif la_ == 9:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Primary_output_expression_listContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 9)
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":6717
 *                 pass
 * 
 *             elif la_ == 10:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Primary_expression_listContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 10)
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_10, 10, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6717, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 6717, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":6718
 * 
 *             elif la_ == 10:
 *                 localctx = ModelicaParser.Primary_expression_listContext(self, localctx)             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 10)
 *                 self.state = 900
 */
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6718, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Primary_expression_listContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6718, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6718, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6718, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6718, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_7) {
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_v_localctx);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6718, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6719
 *             elif la_ == 10:
 *                 localctx = ModelicaParser.Primary_expression_listContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 10)             # <<<<<<<<<<<<<<
 *                 self.state = 900
 *                 self.match(ModelicaParser.T__75)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6719, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_10};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6719, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_10};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6719, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6719, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_10);
            __Pyx_GIVEREF(__pyx_int_10);
            PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_10);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6719, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6720
 *                 localctx = ModelicaParser.Primary_expression_listContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 10)
 *                 self.state = 900             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__75)
 *                 self.state = 901
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_900) < 0) __PYX_ERR(0, 6720, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6721
 *                 self.enterOuterAlt(localctx, 10)
 *                 self.state = 900
 *                 self.match(ModelicaParser.T__75)             # <<<<<<<<<<<<<<
 *                 self.state = 901
 *                 self.expression_list()
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6721, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6721, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__75); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6721, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6721, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6721, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6721, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6721, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6721, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6722
 *                 self.state = 900
 *                 self.match(ModelicaParser.T__75)
 *                 self.state = 901             # <<<<<<<<<<<<<<
 *                 self.expression_list()
 *                 self.state = 906
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_901) < 0) __PYX_ERR(0, 6722, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6723
 *                 self.match(ModelicaParser.T__75)
 *                 self.state = 901
 *                 self.expression_list()             # <<<<<<<<<<<<<<
 *                 self.state = 906
 *                 self._errHandler.sync(self)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expression_list); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6723, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (__pyx_t_5) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6723, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6723, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6724
 *                 self.state = 901
 *                 self.expression_list()
 *                 self.state = 906             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_906) < 0) __PYX_ERR(0, 6724, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6725
 *                 self.expression_list()
 *                 self.state = 906
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 *                 while _la==ModelicaParser.T__0:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6725, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_sync); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6725, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6725, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_5)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6725, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6725, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6725, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6725, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6726
 *                 self.state = 906
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 while _la==ModelicaParser.T__0:
 *                     self.state = 902
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6726, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6726, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_tuple__950, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6726, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_v__la = __pyx_t_1;
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6727
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 while _la==ModelicaParser.T__0:             # <<<<<<<<<<<<<<
 *                     self.state = 902
 *                     self.match(ModelicaParser.T__0)
 */
          while (1) {
            __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6727, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6727, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6727, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 6727, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (!__pyx_t_11) break;

            /* "pymola/generated/ModelicaParser.pyx":6728
 *                 _la = self._input.LA(1)
 *                 while _la==ModelicaParser.T__0:
 *                     self.state = 902             # <<<<<<<<<<<<<<
 *                     self.match(ModelicaParser.T__0)
 *                     self.state = 903
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_902) < 0) __PYX_ERR(0, 6728, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":6729
 *                 while _la==ModelicaParser.T__0:
 *                     self.state = 902
 *                     self.match(ModelicaParser.T__0)             # <<<<<<<<<<<<<<
 *                     self.state = 903
 *                     self.expression_list()
 */
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6729, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6729, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6729, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
              __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
              if (likely(__pyx_t_3)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                __Pyx_INCREF(__pyx_t_3);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_5, function);
              }
            }
            if (!__pyx_t_3) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6729, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_5)) {
                PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_2};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6729, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
                PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_2};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6729, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              } else
              #endif
              {
                __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6729, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_7);
                __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
                __Pyx_GIVEREF(__pyx_t_2);
                PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_2);
                __pyx_t_2 = 0;
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6729, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":6730
 *                     self.state = 902
 *                     self.match(ModelicaParser.T__0)
 *                     self.state = 903             # <<<<<<<<<<<<<<
 *                     self.expression_list()
 *                     self.state = 908
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_903) < 0) __PYX_ERR(0, 6730, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":6731
 *                     self.match(ModelicaParser.T__0)
 *                     self.state = 903
 *                     self.expression_list()             # <<<<<<<<<<<<<<
 *                     self.state = 908
 *                     self._errHandler.sync(self)
 */
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expression_list); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6731, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_7 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
              __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
              if (likely(__pyx_t_7)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                __Pyx_INCREF(__pyx_t_7);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_5, function);
              }
            }
            if (__pyx_t_7) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6731, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else {
              __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6731, __pyx_L6_error)
            }
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":6732
 *                     self.state = 903
 *                     self.expression_list()
 *                     self.state = 908             # <<<<<<<<<<<<<<
 *                     self._errHandler.sync(self)
 *                     _la = self._input.LA(1)
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_908) < 0) __PYX_ERR(0, 6732, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":6733
 *                     self.expression_list()
 *                     self.state = 908
 *                     self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                     _la = self._input.LA(1)
 * 
 */
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6733, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6733, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_5 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
              if (likely(__pyx_t_5)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
                __Pyx_INCREF(__pyx_t_5);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_7, function);
              }
            }
            if (!__pyx_t_5) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6733, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_7)) {
                PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6733, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
                PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6733, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else
              #endif
              {
                __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6733, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5); __pyx_t_5 = NULL;
                __Pyx_INCREF(__pyx_v_self);
                __Pyx_GIVEREF(__pyx_v_self);
                PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6733, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":6734
 *                     self.state = 908
 *                     self._errHandler.sync(self)
 *                     _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *                 self.state = 909
 */
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6734, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6734, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__951, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6734, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_1);
            __pyx_t_1 = 0;
          }

          /* "pymola/generated/ModelicaParser.pyx":6736
 *                     _la = self._input.LA(1)
 * 
 *                 self.state = 909             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__76)
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_909) < 0) __PYX_ERR(0, 6736, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6737
 * 
 *                 self.state = 909
 *                 self.match(ModelicaParser.T__76)             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6737, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6737, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__76); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6737, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6737, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_5};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6737, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_5};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6737, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6737, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_GIVEREF(__pyx_t_5);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_t_5);
              __pyx_t_5 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6737, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6717
 *                 pass
 * 
 *             elif la_ == 10:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Primary_expression_listContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 10)
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":6740
 *                 pass
 * 
 *             elif la_ == 11:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Primary_function_argumentsContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 11)
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_11, 11, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6740, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 6740, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":6741
 * 
 *             elif la_ == 11:
 *                 localctx = ModelicaParser.Primary_function_argumentsContext(self, localctx)             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 11)
 *                 self.state = 911
 */
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6741, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Primary_function_argumentsContex); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6741, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6741, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6741, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6741, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            if (__pyx_t_7) {
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_v_localctx);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6741, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6742
 *             elif la_ == 11:
 *                 localctx = ModelicaParser.Primary_function_argumentsContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 11)             # <<<<<<<<<<<<<<
 *                 self.state = 911
 *                 self.match(ModelicaParser.T__29)
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6742, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_localctx, __pyx_int_11};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6742, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_localctx, __pyx_int_11};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6742, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6742, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_11);
            __Pyx_GIVEREF(__pyx_int_11);
            PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_11);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6742, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6743
 *                 localctx = ModelicaParser.Primary_function_argumentsContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 11)
 *                 self.state = 911             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__29)
 *                 self.state = 912
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_911) < 0) __PYX_ERR(0, 6743, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6744
 *                 self.enterOuterAlt(localctx, 11)
 *                 self.state = 911
 *                 self.match(ModelicaParser.T__29)             # <<<<<<<<<<<<<<
 *                 self.state = 912
 *                 self.function_arguments()
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6744, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6744, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__29); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6744, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6744, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_5};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6744, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_5};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6744, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            {
              __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6744, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_7); __pyx_t_7 = NULL;
              __Pyx_GIVEREF(__pyx_t_5);
              PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_5);
              __pyx_t_5 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6744, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6745
 *                 self.state = 911
 *                 self.match(ModelicaParser.T__29)
 *                 self.state = 912             # <<<<<<<<<<<<<<
 *                 self.function_arguments()
 *                 self.state = 913
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_912) < 0) __PYX_ERR(0, 6745, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6746
 *                 self.match(ModelicaParser.T__29)
 *                 self.state = 912
 *                 self.function_arguments()             # <<<<<<<<<<<<<<
 *                 self.state = 913
 *                 self.match(ModelicaParser.T__30)
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_function_arguments); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6746, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6746, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6746, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6747
 *                 self.state = 912
 *                 self.function_arguments()
 *                 self.state = 913             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__30)
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_913) < 0) __PYX_ERR(0, 6747, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6748
 *                 self.function_arguments()
 *                 self.state = 913
 *                 self.match(ModelicaParser.T__30)             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6748, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6748, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__30); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6748, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6748, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_5};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6748, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_5};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6748, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            {
              __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6748, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_GIVEREF(__pyx_t_5);
              PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_5);
              __pyx_t_5 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6748, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6740
 *                 pass
 * 
 *             elif la_ == 11:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Primary_function_argumentsContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 11)
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":6751
 *                 pass
 * 
 *             elif la_ == 12:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Primary_endContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 12)
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_12, 12, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6751, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 6751, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":6752
 * 
 *             elif la_ == 12:
 *                 localctx = ModelicaParser.Primary_endContext(self, localctx)             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 12)
 *                 self.state = 915
 */
          __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6752, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Primary_endContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6752, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6752, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6752, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6752, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_v_localctx);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6752, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6753
 *             elif la_ == 12:
 *                 localctx = ModelicaParser.Primary_endContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 12)             # <<<<<<<<<<<<<<
 *                 self.state = 915
 *                 self.match(ModelicaParser.T__13)
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6753, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_localctx, __pyx_int_12};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6753, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_localctx, __pyx_int_12};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6753, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6753, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_12);
            __Pyx_GIVEREF(__pyx_int_12);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_int_12);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6753, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6754
 *                 localctx = ModelicaParser.Primary_endContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 12)
 *                 self.state = 915             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__13)
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_915) < 0) __PYX_ERR(0, 6754, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6755
 *                 self.enterOuterAlt(localctx, 12)
 *                 self.state = 915
 *                 self.match(ModelicaParser.T__13)             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6755, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6755, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__13); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6755, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_3) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6755, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_5};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6755, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_5};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6755, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            {
              __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6755, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
              __Pyx_GIVEREF(__pyx_t_5);
              PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_5);
              __pyx_t_5 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6755, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6751
 *                 pass
 * 
 *             elif la_ == 12:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Primary_endContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 12)
 */
        }
        __pyx_L14:;

        /* "pymola/generated/ModelicaParser.pyx":6640
 *         self.enterRule(localctx, 114, self.RULE_primary)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.state = 916
 *             self._errHandler.sync(self)
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":6759
 * 
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6759, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.primary", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_7, &__pyx_t_2) < 0) __PYX_ERR(0, 6759, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_7);
        __pyx_v_re = __pyx_t_7;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":6760
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 6760, __pyx_L22_error)

          /* "pymola/generated/ModelicaParser.pyx":6761
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6761, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_reportError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6761, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_self, __pyx_v_re};
            __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6761, __pyx_L22_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_5);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_self, __pyx_v_re};
            __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6761, __pyx_L22_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_5);
          } else
          #endif
          {
            __pyx_t_12 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 6761, __pyx_L22_error)
            __Pyx_GOTREF(__pyx_t_12);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_12, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6761, __pyx_L22_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6762
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6762, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_recover); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 6762, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_12);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_12, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6762, __pyx_L22_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_5);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6762, __pyx_L22_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_5);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6762, __pyx_L22_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_4) {
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __pyx_t_4 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_3, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6762, __pyx_L22_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":6759
 * 
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L23;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L22_error:;
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __pyx_t_6 = __pyx_lineno; __pyx_t_13 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
            }
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_ErrRestore(__pyx_t_15, __pyx_t_16, __pyx_t_17);
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_13; __pyx_filename = __pyx_t_14;
            goto __pyx_L8_except_error;
          }
          __pyx_L23:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":6640
 *         self.enterRule(localctx, 114, self.RULE_primary)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.state = 916
 *             self._errHandler.sync(self)
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":6764
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6764, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6764, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6764, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_19, &__pyx_t_18);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_18);
      __pyx_t_13 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_21 = __pyx_filename;
      {
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6764, __pyx_L29_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6764, __pyx_L29_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6764, __pyx_L29_error)
        }
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __pyx_lineno = __pyx_t_13; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_21;
      goto __pyx_L1_error;
      __pyx_L29_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":6765
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class NameContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":6635
 * 
 * 
 *     def primary(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.PrimaryContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.primary", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v_la_);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6769
 *     class NameContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_11NameContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_11NameContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_11NameContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_11NameContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 6769, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 6769, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6769, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.NameContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_11NameContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_11NameContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":6770
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 6770, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6770, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6770, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6770, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6770, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6770, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6770, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6770, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6771
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def IDENT(self, i:int=None):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 6771, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6769
 *     class NameContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.NameContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6773
 *             self.parser = parser
 * 
 *         def IDENT(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTokens(ModelicaParser.IDENT)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_11NameContext_3IDENT(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_11NameContext_3IDENT = {"IDENT", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_11NameContext_3IDENT, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_11NameContext_3IDENT(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("IDENT (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "IDENT") < 0)) __PYX_ERR(0, 6773, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("IDENT", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6773, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.NameContext.IDENT", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_11NameContext_2IDENT(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_11NameContext_2IDENT(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("IDENT", 0);

  /* "pymola/generated/ModelicaParser.pyx":6774
 * 
 *         def IDENT(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTokens(ModelicaParser.IDENT)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6775
 *         def IDENT(self, i:int=None):
 *             if i is None:
 *                 return self.getTokens(ModelicaParser.IDENT)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getToken(ModelicaParser.IDENT, i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTokens); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6775, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6775, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6775, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6775, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6775, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6775, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6775, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6775, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":6774
 * 
 *         def IDENT(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTokens(ModelicaParser.IDENT)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6777
 *                 return self.getTokens(ModelicaParser.IDENT)
 *             else:
 *                 return self.getToken(ModelicaParser.IDENT, i)             # <<<<<<<<<<<<<<
 * 
 *         def getRuleIndex(self):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getToken); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6777, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6777, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6777, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6777, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6777, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6777, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6777, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":6773
 *             self.parser = parser
 * 
 *         def IDENT(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTokens(ModelicaParser.IDENT)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.NameContext.IDENT", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6779
 *                 return self.getToken(ModelicaParser.IDENT, i)
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_name
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_11NameContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_11NameContext_5getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_11NameContext_5getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_11NameContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_11NameContext_4getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_11NameContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":6780
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_name             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":6779
 *                 return self.getToken(ModelicaParser.IDENT, i)
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_name
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.NameContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6782
 *             return ModelicaParser.RULE_name
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterName" ):
 *                 listener.enterName(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_11NameContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_11NameContext_7enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_11NameContext_7enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_11NameContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 6782, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 6782, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6782, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.NameContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_11NameContext_6enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_11NameContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6783
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterName" ):             # <<<<<<<<<<<<<<
 *                 listener.enterName(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterName); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6783, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6784
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterName" ):
 *                 listener.enterName(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterName); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6784, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6784, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6784, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6784, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6784, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6784, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6783
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterName" ):             # <<<<<<<<<<<<<<
 *                 listener.enterName(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6782
 *             return ModelicaParser.RULE_name
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterName" ):
 *                 listener.enterName(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.NameContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6786
 *                 listener.enterName(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitName" ):
 *                 listener.exitName(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_11NameContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_11NameContext_9exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_11NameContext_9exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_11NameContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 6786, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 6786, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6786, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.NameContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_11NameContext_8exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_11NameContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6787
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitName" ):             # <<<<<<<<<<<<<<
 *                 listener.exitName(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitName); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6787, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6788
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitName" ):
 *                 listener.exitName(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitName); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6788, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6788, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6788, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6788, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6788, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6788, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6787
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitName" ):             # <<<<<<<<<<<<<<
 *                 listener.exitName(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6786
 *                 listener.enterName(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitName" ):
 *                 listener.exitName(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.NameContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6790
 *                 listener.exitName(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitName" ):
 *                 return visitor.visitName(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_11NameContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_11NameContext_11accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_11NameContext_11accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_11NameContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 6790, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 6790, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6790, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.NameContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_11NameContext_10accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_11NameContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":6791
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitName" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitName(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitName); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6791, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6792
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitName" ):
 *                 return visitor.visitName(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitName); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6792, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6792, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6792, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6792, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6792, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6792, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":6791
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitName" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitName(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6794
 *                 return visitor.visitName(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6794, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6794, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6794, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6794, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6794, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6794, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":6790
 *                 listener.exitName(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitName" ):
 *                 return visitor.visitName(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.NameContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6799
 * 
 * 
 *     def name(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.NameContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_119name(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_119name = {"name", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_119name, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_119name(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("name (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_118name(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_118name(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  char const *__pyx_t_21;
  __Pyx_RefNannySetupContext("name", 0);

  /* "pymola/generated/ModelicaParser.pyx":6801
 *     def name(self):
 * 
 *         localctx = ModelicaParser.NameContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 116, self.RULE_name)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_NameContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6801, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6801, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6801, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6801, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6802
 * 
 *         localctx = ModelicaParser.NameContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 116, self.RULE_name)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6802, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_name); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6802, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_116, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6802, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_116, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6802, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6802, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_116);
    __Pyx_GIVEREF(__pyx_int_116);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_116);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6802, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6803
 *         localctx = ModelicaParser.NameContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 116, self.RULE_name)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 6803, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6804
 *         self.enterRule(localctx, 116, self.RULE_name)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 919
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":6805
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 919
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6805, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6805, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6805, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6805, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6805, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":6806
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 919             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_919) < 0) __PYX_ERR(0, 6806, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":6807
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 919
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__27:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6807, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6807, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6807, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6807, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6807, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6807, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6807, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":6808
 *             self.state = 919
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.T__27:
 *                 self.state = 918
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6808, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6808, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__952, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6808, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v__la = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":6809
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__27:             # <<<<<<<<<<<<<<
 *                 self.state = 918
 *                 self.match(ModelicaParser.T__27)
 */
        __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6809, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__27); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6809, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6809, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 6809, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":6810
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__27:
 *                 self.state = 918             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__27)
 * 
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_918) < 0) __PYX_ERR(0, 6810, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6811
 *             if _la==ModelicaParser.T__27:
 *                 self.state = 918
 *                 self.match(ModelicaParser.T__27)             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6811, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6811, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__27); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6811, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6811, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6811, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6811, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6811, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6811, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6809
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__27:             # <<<<<<<<<<<<<<
 *                 self.state = 918
 *                 self.match(ModelicaParser.T__27)
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":6814
 * 
 * 
 *             self.state = 921             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.IDENT)
 *             self.state = 926
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_921) < 0) __PYX_ERR(0, 6814, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":6815
 * 
 *             self.state = 921
 *             self.match(ModelicaParser.IDENT)             # <<<<<<<<<<<<<<
 *             self.state = 926
 *             self._errHandler.sync(self)
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6815, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6815, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6815, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6815, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_3};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6815, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_3};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6815, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6815, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
            __Pyx_GIVEREF(__pyx_t_3);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_3);
            __pyx_t_3 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6815, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":6816
 *             self.state = 921
 *             self.match(ModelicaParser.IDENT)
 *             self.state = 926             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_926) < 0) __PYX_ERR(0, 6816, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":6817
 *             self.match(ModelicaParser.IDENT)
 *             self.state = 926
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             while _la==ModelicaParser.T__27:
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6817, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6817, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6817, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6817, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6817, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6817, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6817, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":6818
 *             self.state = 926
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             while _la==ModelicaParser.T__27:
 *                 self.state = 922
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6818, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6818, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__953, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6818, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":6819
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             while _la==ModelicaParser.T__27:             # <<<<<<<<<<<<<<
 *                 self.state = 922
 *                 self.match(ModelicaParser.T__27)
 */
        while (1) {
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6819, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__27); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6819, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6819, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 6819, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (!__pyx_t_11) break;

          /* "pymola/generated/ModelicaParser.pyx":6820
 *             _la = self._input.LA(1)
 *             while _la==ModelicaParser.T__27:
 *                 self.state = 922             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__27)
 *                 self.state = 923
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_922) < 0) __PYX_ERR(0, 6820, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6821
 *             while _la==ModelicaParser.T__27:
 *                 self.state = 922
 *                 self.match(ModelicaParser.T__27)             # <<<<<<<<<<<<<<
 *                 self.state = 923
 *                 self.match(ModelicaParser.IDENT)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6821, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6821, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__27); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6821, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_3) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6821, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6821, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6821, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6821, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __pyx_t_3 = NULL;
              __Pyx_GIVEREF(__pyx_t_7);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_7);
              __pyx_t_7 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6821, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6822
 *                 self.state = 922
 *                 self.match(ModelicaParser.T__27)
 *                 self.state = 923             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.IDENT)
 *                 self.state = 928
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_923) < 0) __PYX_ERR(0, 6822, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6823
 *                 self.match(ModelicaParser.T__27)
 *                 self.state = 923
 *                 self.match(ModelicaParser.IDENT)             # <<<<<<<<<<<<<<
 *                 self.state = 928
 *                 self._errHandler.sync(self)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6823, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6823, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6823, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_4) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6823, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6823, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6823, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6823, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __pyx_t_4 = NULL;
              __Pyx_GIVEREF(__pyx_t_7);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_t_7);
              __pyx_t_7 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6823, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6824
 *                 self.state = 923
 *                 self.match(ModelicaParser.IDENT)
 *                 self.state = 928             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_928) < 0) __PYX_ERR(0, 6824, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6825
 *                 self.match(ModelicaParser.IDENT)
 *                 self.state = 928
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 * 
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6825, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_sync); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6825, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6825, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6825, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6825, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6825, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6825, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6826
 *                 self.state = 928
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *         except RecognitionException as re:
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6826, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6826, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__954, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6826, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_1);
          __pyx_t_1 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":6804
 *         self.enterRule(localctx, 116, self.RULE_name)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 919
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":6828
 *                 _la = self._input.LA(1)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6828, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.name", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_3, &__pyx_t_7) < 0) __PYX_ERR(0, 6828, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_3);
        __pyx_v_re = __pyx_t_3;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":6829
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 6829, __pyx_L22_error)

          /* "pymola/generated/ModelicaParser.pyx":6830
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6830, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6830, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6830, __pyx_L22_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6830, __pyx_L22_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_12 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 6830, __pyx_L22_error)
            __Pyx_GOTREF(__pyx_t_12);
            if (__pyx_t_4) {
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_4); __pyx_t_4 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_12, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6830, __pyx_L22_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6831
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6831, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 6831, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_12);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_12, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6831, __pyx_L22_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6831, __pyx_L22_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6831, __pyx_L22_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6831, __pyx_L22_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":6828
 *                 _la = self._input.LA(1)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L23;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L22_error:;
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __pyx_t_6 = __pyx_lineno; __pyx_t_13 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
            }
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_ErrRestore(__pyx_t_15, __pyx_t_16, __pyx_t_17);
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_13; __pyx_filename = __pyx_t_14;
            goto __pyx_L8_except_error;
          }
          __pyx_L23:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":6804
 *         self.enterRule(localctx, 116, self.RULE_name)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 919
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":6833
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6833, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6833, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6833, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_19, &__pyx_t_18);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_18);
      __pyx_t_13 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_21 = __pyx_filename;
      {
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6833, __pyx_L29_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6833, __pyx_L29_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6833, __pyx_L29_error)
        }
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __pyx_lineno = __pyx_t_13; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_21;
      goto __pyx_L1_error;
      __pyx_L29_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":6834
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Component_reference_elementContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":6799
 * 
 * 
 *     def name(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.NameContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.name", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6838
 *     class Component_reference_elementContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 6838, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 6838, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6838, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_reference_elementContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":6839
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 6839, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6839, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6839, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6839, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6839, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6839, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6839, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6839, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6840
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def IDENT(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 6840, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6838
 *     class Component_reference_elementContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_reference_elementContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6842
 *             self.parser = parser
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_3IDENT(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_3IDENT = {"IDENT", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_3IDENT, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_3IDENT(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("IDENT (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_2IDENT(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_2IDENT(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("IDENT", 0);

  /* "pymola/generated/ModelicaParser.pyx":6843
 * 
 *         def IDENT(self):
 *             return self.getToken(ModelicaParser.IDENT, 0)             # <<<<<<<<<<<<<<
 * 
 *         def array_subscripts(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getToken); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6843, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6843, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6843, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6843, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6843, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6843, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6843, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":6842
 *             self.parser = parser
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_reference_elementContext.IDENT", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6845
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 *         def array_subscripts(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Array_subscriptsContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_5array_subscripts(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_5array_subscripts = {"array_subscripts", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_5array_subscripts, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_5array_subscripts(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("array_subscripts (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_4array_subscripts(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_4array_subscripts(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("array_subscripts", 0);

  /* "pymola/generated/ModelicaParser.pyx":6846
 * 
 *         def array_subscripts(self):
 *             return self.getTypedRuleContext(ModelicaParser.Array_subscriptsContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6846, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6846, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Array_subscriptsContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6846, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6846, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6846, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6846, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6846, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":6845
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 *         def array_subscripts(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Array_subscriptsContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_reference_elementContext.array_subscripts", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6849
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_component_reference_element
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_7getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_7getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_6getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":6850
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_component_reference_element             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6850, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_component_reference_element); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6850, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":6849
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_component_reference_element
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_reference_elementContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6852
 *             return ModelicaParser.RULE_component_reference_element
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterComponent_reference_element" ):
 *                 listener.enterComponent_reference_element(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_9enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_9enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 6852, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 6852, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6852, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_reference_elementContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_8enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6853
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterComponent_reference_element" ):             # <<<<<<<<<<<<<<
 *                 listener.enterComponent_reference_element(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterComponent_reference_element); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6853, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6854
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterComponent_reference_element" ):
 *                 listener.enterComponent_reference_element(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterComponent_reference_element); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6854, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6854, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6854, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6854, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6854, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6854, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6853
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterComponent_reference_element" ):             # <<<<<<<<<<<<<<
 *                 listener.enterComponent_reference_element(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6852
 *             return ModelicaParser.RULE_component_reference_element
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterComponent_reference_element" ):
 *                 listener.enterComponent_reference_element(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_reference_elementContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6856
 *                 listener.enterComponent_reference_element(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitComponent_reference_element" ):
 *                 listener.exitComponent_reference_element(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_11exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_11exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 6856, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 6856, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6856, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_reference_elementContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_10exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6857
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitComponent_reference_element" ):             # <<<<<<<<<<<<<<
 *                 listener.exitComponent_reference_element(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitComponent_reference_element); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6857, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6858
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitComponent_reference_element" ):
 *                 listener.exitComponent_reference_element(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitComponent_reference_element); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6858, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6858, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6858, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6858, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6858, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6858, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6857
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitComponent_reference_element" ):             # <<<<<<<<<<<<<<
 *                 listener.exitComponent_reference_element(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6856
 *                 listener.enterComponent_reference_element(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitComponent_reference_element" ):
 *                 listener.exitComponent_reference_element(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_reference_elementContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6860
 *                 listener.exitComponent_reference_element(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitComponent_reference_element" ):
 *                 return visitor.visitComponent_reference_element(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_13accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_13accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 6860, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 6860, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6860, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_reference_elementContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_12accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":6861
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitComponent_reference_element" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitComponent_reference_element(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitComponent_reference_element); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6861, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6862
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitComponent_reference_element" ):
 *                 return visitor.visitComponent_reference_element(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitComponent_reference_element); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6862, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6862, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6862, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6862, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6862, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6862, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":6861
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitComponent_reference_element" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitComponent_reference_element(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6864
 *                 return visitor.visitComponent_reference_element(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6864, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6864, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6864, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6864, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6864, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6864, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":6860
 *                 listener.exitComponent_reference_element(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitComponent_reference_element" ):
 *                 return visitor.visitComponent_reference_element(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_reference_elementContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6869
 * 
 * 
 *     def component_reference_element(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Component_reference_elementContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_121component_reference_element(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_121component_reference_element = {"component_reference_element", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_121component_reference_element, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_121component_reference_element(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("component_reference_element (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_120component_reference_element(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_120component_reference_element(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v_la_ = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  char const *__pyx_t_21;
  __Pyx_RefNannySetupContext("component_reference_element", 0);

  /* "pymola/generated/ModelicaParser.pyx":6871
 *     def component_reference_element(self):
 * 
 *         localctx = ModelicaParser.Component_reference_elementContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 118, self.RULE_component_reference_element)
 *         try:
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6871, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Component_reference_elementConte); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6871, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6871, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6871, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6871, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6871, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6871, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6871, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6872
 * 
 *         localctx = ModelicaParser.Component_reference_elementContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 118, self.RULE_component_reference_element)             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6872, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_component_reference_element); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6872, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_118, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6872, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_118, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6872, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6872, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_118);
    __Pyx_GIVEREF(__pyx_int_118);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_118);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6872, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6873
 *         localctx = ModelicaParser.Component_reference_elementContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 118, self.RULE_component_reference_element)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 929
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":6874
 *         self.enterRule(localctx, 118, self.RULE_component_reference_element)
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 929
 *             self.match(ModelicaParser.IDENT)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6874, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6874, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6874, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6874, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6874, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":6875
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 929             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.IDENT)
 *             self.state = 931
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_929) < 0) __PYX_ERR(0, 6875, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":6876
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 929
 *             self.match(ModelicaParser.IDENT)             # <<<<<<<<<<<<<<
 *             self.state = 931
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6876, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6876, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6876, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6876, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6876, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6876, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6876, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6876, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":6877
 *             self.state = 929
 *             self.match(ModelicaParser.IDENT)
 *             self.state = 931             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             la_ = self._interp.adaptivePredict(self._input,116,self._ctx)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_931) < 0) __PYX_ERR(0, 6877, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":6878
 *             self.match(ModelicaParser.IDENT)
 *             self.state = 931
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             la_ = self._interp.adaptivePredict(self._input,116,self._ctx)
 *             if la_ == 1:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6878, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6878, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6878, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6878, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6878, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6878, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6878, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":6879
 *             self.state = 931
 *             self._errHandler.sync(self)
 *             la_ = self._interp.adaptivePredict(self._input,116,self._ctx)             # <<<<<<<<<<<<<<
 *             if la_ == 1:
 *                 self.state = 930
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_interp); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6879, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_adaptivePredict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6879, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6879, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6879, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_2)) {
          PyObject *__pyx_temp[4] = {__pyx_t_7, __pyx_t_4, __pyx_int_116, __pyx_t_3};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6879, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
          PyObject *__pyx_temp[4] = {__pyx_t_7, __pyx_t_4, __pyx_int_116, __pyx_t_3};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6879, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6879, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_7) {
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_t_4);
          __Pyx_INCREF(__pyx_int_116);
          __Pyx_GIVEREF(__pyx_int_116);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_int_116);
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_6, __pyx_t_3);
          __pyx_t_4 = 0;
          __pyx_t_3 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6879, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_v_la_ = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":6880
 *             self._errHandler.sync(self)
 *             la_ = self._interp.adaptivePredict(self._input,116,self._ctx)
 *             if la_ == 1:             # <<<<<<<<<<<<<<
 *                 self.state = 930
 *                 self.array_subscripts()
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6880, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 6880, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":6881
 *             la_ = self._interp.adaptivePredict(self._input,116,self._ctx)
 *             if la_ == 1:
 *                 self.state = 930             # <<<<<<<<<<<<<<
 *                 self.array_subscripts()
 * 
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_930) < 0) __PYX_ERR(0, 6881, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6882
 *             if la_ == 1:
 *                 self.state = 930
 *                 self.array_subscripts()             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_array_subscripts); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6882, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (__pyx_t_5) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6882, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6882, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6880
 *             self._errHandler.sync(self)
 *             la_ = self._interp.adaptivePredict(self._input,116,self._ctx)
 *             if la_ == 1:             # <<<<<<<<<<<<<<
 *                 self.state = 930
 *                 self.array_subscripts()
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":6873
 *         localctx = ModelicaParser.Component_reference_elementContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 118, self.RULE_component_reference_element)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 929
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":6885
 * 
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6885, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.component_reference_element", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_2, &__pyx_t_5) < 0) __PYX_ERR(0, 6885, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_2);
        __pyx_v_re = __pyx_t_2;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":6886
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 6886, __pyx_L20_error)

          /* "pymola/generated/ModelicaParser.pyx":6887
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6887, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_reportError); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6887, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6887, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6887, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_12 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 6887, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_12);
            if (__pyx_t_4) {
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_4); __pyx_t_4 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_12, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6887, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6888
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6888, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_recover); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 6888, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_12);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_12, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6888, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6888, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6888, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (__pyx_t_7) {
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6888, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":6885
 * 
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L21;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L20_error:;
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __pyx_t_6 = __pyx_lineno; __pyx_t_13 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
            }
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_ErrRestore(__pyx_t_15, __pyx_t_16, __pyx_t_17);
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_13; __pyx_filename = __pyx_t_14;
            goto __pyx_L8_except_error;
          }
          __pyx_L21:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":6873
 *         localctx = ModelicaParser.Component_reference_elementContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 118, self.RULE_component_reference_element)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 929
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":6890
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6890, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6890, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_5 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6890, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_19, &__pyx_t_18);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_18);
      __pyx_t_13 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_21 = __pyx_filename;
      {
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6890, __pyx_L27_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6890, __pyx_L27_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_5 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6890, __pyx_L27_error)
        }
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __pyx_lineno = __pyx_t_13; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_21;
      goto __pyx_L1_error;
      __pyx_L27_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":6891
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Component_referenceContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":6869
 * 
 * 
 *     def component_reference_element(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Component_reference_elementContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.component_reference_element", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v_la_);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6895
 *     class Component_referenceContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Component_referenceContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Component_referenceContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Component_referenceContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Component_referenceContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 6895, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 6895, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6895, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_referenceContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Component_referenceContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Component_referenceContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":6896
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 6896, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6896, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6896, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6896, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6896, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6896, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6896, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6896, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6897
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def component_reference_element(self, i:int=None):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 6897, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6895
 *     class Component_referenceContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_referenceContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6899
 *             self.parser = parser
 * 
 *         def component_reference_element(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Component_reference_elementContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Component_referenceContext_3component_reference_element(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Component_referenceContext_3component_reference_element = {"component_reference_element", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Component_referenceContext_3component_reference_element, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Component_referenceContext_3component_reference_element(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("component_reference_element (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "component_reference_element") < 0)) __PYX_ERR(0, 6899, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("component_reference_element", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6899, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_referenceContext.component_reference_element", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Component_referenceContext_2component_reference_element(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Component_referenceContext_2component_reference_element(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("component_reference_element", 0);

  /* "pymola/generated/ModelicaParser.pyx":6900
 * 
 *         def component_reference_element(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.Component_reference_elementContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6901
 *         def component_reference_element(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Component_reference_elementContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.Component_reference_elementContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6901, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6901, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_Component_reference_elementConte); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6901, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6901, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6901, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6901, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6901, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6901, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":6900
 * 
 *         def component_reference_element(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.Component_reference_elementContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6903
 *                 return self.getTypedRuleContexts(ModelicaParser.Component_reference_elementContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.Component_reference_elementContext,i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6903, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6903, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Component_reference_elementConte); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6903, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6903, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6903, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6903, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6903, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":6899
 *             self.parser = parser
 * 
 *         def component_reference_element(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Component_reference_elementContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_referenceContext.component_reference_element", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6906
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_component_reference
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Component_referenceContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Component_referenceContext_5getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Component_referenceContext_5getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Component_referenceContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Component_referenceContext_4getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Component_referenceContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":6907
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_component_reference             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6907, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_component_reference); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6907, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":6906
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_component_reference
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_referenceContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6909
 *             return ModelicaParser.RULE_component_reference
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterComponent_reference" ):
 *                 listener.enterComponent_reference(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Component_referenceContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Component_referenceContext_7enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Component_referenceContext_7enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Component_referenceContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 6909, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 6909, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6909, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_referenceContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Component_referenceContext_6enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Component_referenceContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6910
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterComponent_reference" ):             # <<<<<<<<<<<<<<
 *                 listener.enterComponent_reference(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterComponent_reference); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6910, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6911
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterComponent_reference" ):
 *                 listener.enterComponent_reference(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterComponent_reference); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6911, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6911, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6911, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6911, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6911, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6911, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6910
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterComponent_reference" ):             # <<<<<<<<<<<<<<
 *                 listener.enterComponent_reference(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6909
 *             return ModelicaParser.RULE_component_reference
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterComponent_reference" ):
 *                 listener.enterComponent_reference(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_referenceContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6913
 *                 listener.enterComponent_reference(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitComponent_reference" ):
 *                 listener.exitComponent_reference(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Component_referenceContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Component_referenceContext_9exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Component_referenceContext_9exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Component_referenceContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 6913, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 6913, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6913, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_referenceContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Component_referenceContext_8exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Component_referenceContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6914
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitComponent_reference" ):             # <<<<<<<<<<<<<<
 *                 listener.exitComponent_reference(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitComponent_reference); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6914, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6915
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitComponent_reference" ):
 *                 listener.exitComponent_reference(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitComponent_reference); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6915, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6915, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6915, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6915, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6915, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6915, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6914
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitComponent_reference" ):             # <<<<<<<<<<<<<<
 *                 listener.exitComponent_reference(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6913
 *                 listener.enterComponent_reference(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitComponent_reference" ):
 *                 listener.exitComponent_reference(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_referenceContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6917
 *                 listener.exitComponent_reference(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitComponent_reference" ):
 *                 return visitor.visitComponent_reference(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Component_referenceContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Component_referenceContext_11accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Component_referenceContext_11accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Component_referenceContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 6917, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 6917, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6917, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_referenceContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Component_referenceContext_10accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Component_referenceContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":6918
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitComponent_reference" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitComponent_reference(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitComponent_reference); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6918, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6919
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitComponent_reference" ):
 *                 return visitor.visitComponent_reference(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitComponent_reference); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6919, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6919, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6919, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6919, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6919, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6919, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":6918
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitComponent_reference" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitComponent_reference(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6921
 *                 return visitor.visitComponent_reference(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6921, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6921, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6921, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6921, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6921, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6921, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":6917
 *                 listener.exitComponent_reference(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitComponent_reference" ):
 *                 return visitor.visitComponent_reference(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Component_referenceContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6926
 * 
 * 
 *     def component_reference(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Component_referenceContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_123component_reference(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_123component_reference = {"component_reference", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_123component_reference, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_123component_reference(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("component_reference (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_122component_reference(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_122component_reference(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__alt = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  char const *__pyx_t_22;
  __Pyx_RefNannySetupContext("component_reference", 0);

  /* "pymola/generated/ModelicaParser.pyx":6928
 *     def component_reference(self):
 * 
 *         localctx = ModelicaParser.Component_referenceContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 120, self.RULE_component_reference)
 *         try:
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6928, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Component_referenceContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6928, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6928, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6928, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6928, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6928, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6928, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6928, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6929
 * 
 *         localctx = ModelicaParser.Component_referenceContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 120, self.RULE_component_reference)             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6929, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_component_reference); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6929, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_120, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6929, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_120, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6929, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6929, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_120);
    __Pyx_GIVEREF(__pyx_int_120);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_120);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6929, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6930
 *         localctx = ModelicaParser.Component_referenceContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 120, self.RULE_component_reference)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 933
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":6931
 *         self.enterRule(localctx, 120, self.RULE_component_reference)
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 933
 *             self.component_reference_element()
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6931, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6931, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6931, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6931, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6931, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":6932
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 933             # <<<<<<<<<<<<<<
 *             self.component_reference_element()
 *             self.state = 938
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_933) < 0) __PYX_ERR(0, 6932, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":6933
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 933
 *             self.component_reference_element()             # <<<<<<<<<<<<<<
 *             self.state = 938
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_component_reference_element); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6933, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6933, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6933, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":6934
 *             self.state = 933
 *             self.component_reference_element()
 *             self.state = 938             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _alt = self._interp.adaptivePredict(self._input,117,self._ctx)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_938) < 0) __PYX_ERR(0, 6934, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":6935
 *             self.component_reference_element()
 *             self.state = 938
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _alt = self._interp.adaptivePredict(self._input,117,self._ctx)
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6935, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6935, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6935, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6935, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6935, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6935, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6935, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":6936
 *             self.state = 938
 *             self._errHandler.sync(self)
 *             _alt = self._interp.adaptivePredict(self._input,117,self._ctx)             # <<<<<<<<<<<<<<
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                 if _alt==1:
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_interp); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6936, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_adaptivePredict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6936, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6936, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6936, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_2)) {
          PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_t_7, __pyx_int_117, __pyx_t_3};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6936, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
          PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_t_7, __pyx_int_117, __pyx_t_3};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6936, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6936, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_7);
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_t_7);
          __Pyx_INCREF(__pyx_int_117);
          __Pyx_GIVEREF(__pyx_int_117);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_int_117);
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_6, __pyx_t_3);
          __pyx_t_7 = 0;
          __pyx_t_3 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6936, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_v__alt = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":6937
 *             self._errHandler.sync(self)
 *             _alt = self._interp.adaptivePredict(self._input,117,self._ctx)
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:             # <<<<<<<<<<<<<<
 *                 if _alt==1:
 *                     self.state = 934
 */
        while (1) {
          __pyx_t_1 = PyObject_RichCompare(__pyx_v__alt, __pyx_int_2, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6937, __pyx_L6_error)
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 6937, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (__pyx_t_12) {
          } else {
            __pyx_t_11 = __pyx_t_12;
            goto __pyx_L16_bool_binop_done;
          }
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ATN); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6937, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_INVALID_ALT_NUMBER); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6937, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyObject_RichCompare(__pyx_v__alt, __pyx_t_2, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6937, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 6937, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_11 = __pyx_t_12;
          __pyx_L16_bool_binop_done:;
          if (!__pyx_t_11) break;

          /* "pymola/generated/ModelicaParser.pyx":6938
 *             _alt = self._interp.adaptivePredict(self._input,117,self._ctx)
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                 if _alt==1:             # <<<<<<<<<<<<<<
 *                     self.state = 934
 *                     self.match(ModelicaParser.T__27)
 */
          __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v__alt, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6938, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 6938, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (__pyx_t_11) {

            /* "pymola/generated/ModelicaParser.pyx":6939
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                 if _alt==1:
 *                     self.state = 934             # <<<<<<<<<<<<<<
 *                     self.match(ModelicaParser.T__27)
 *                     self.state = 935
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_934) < 0) __PYX_ERR(0, 6939, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":6940
 *                 if _alt==1:
 *                     self.state = 934
 *                     self.match(ModelicaParser.T__27)             # <<<<<<<<<<<<<<
 *                     self.state = 935
 *                     self.component_reference_element()
 */
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6940, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6940, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_T__27); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6940, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_5 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_5)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_5);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
              }
            }
            if (!__pyx_t_5) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6940, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_2)) {
                PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6940, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
                PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6940, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              } else
              #endif
              {
                __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6940, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_7);
                __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
                __Pyx_GIVEREF(__pyx_t_3);
                PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_3);
                __pyx_t_3 = 0;
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6940, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":6941
 *                     self.state = 934
 *                     self.match(ModelicaParser.T__27)
 *                     self.state = 935             # <<<<<<<<<<<<<<
 *                     self.component_reference_element()
 *                 self.state = 940
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_935) < 0) __PYX_ERR(0, 6941, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":6942
 *                     self.match(ModelicaParser.T__27)
 *                     self.state = 935
 *                     self.component_reference_element()             # <<<<<<<<<<<<<<
 *                 self.state = 940
 *                 self._errHandler.sync(self)
 */
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_component_reference_element); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6942, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_7 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_7)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_7);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
              }
            }
            if (__pyx_t_7) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6942, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else {
              __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6942, __pyx_L6_error)
            }
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":6938
 *             _alt = self._interp.adaptivePredict(self._input,117,self._ctx)
 *             while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
 *                 if _alt==1:             # <<<<<<<<<<<<<<
 *                     self.state = 934
 *                     self.match(ModelicaParser.T__27)
 */
          }

          /* "pymola/generated/ModelicaParser.pyx":6943
 *                     self.state = 935
 *                     self.component_reference_element()
 *                 self.state = 940             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _alt = self._interp.adaptivePredict(self._input,117,self._ctx)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_940) < 0) __PYX_ERR(0, 6943, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6944
 *                     self.component_reference_element()
 *                 self.state = 940
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _alt = self._interp.adaptivePredict(self._input,117,self._ctx)
 * 
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6944, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6944, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6944, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6944, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6944, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6944, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6944, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6945
 *                 self.state = 940
 *                 self._errHandler.sync(self)
 *                 _alt = self._interp.adaptivePredict(self._input,117,self._ctx)             # <<<<<<<<<<<<<<
 * 
 *         except RecognitionException as re:
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_interp); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6945, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_adaptivePredict); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6945, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6945, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6945, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_t_7, __pyx_int_117, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6945, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_t_7, __pyx_int_117, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6945, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6945, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_GIVEREF(__pyx_t_7);
            PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_t_7);
            __Pyx_INCREF(__pyx_int_117);
            __Pyx_GIVEREF(__pyx_int_117);
            PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_int_117);
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_4, 2+__pyx_t_6, __pyx_t_2);
            __pyx_t_7 = 0;
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6945, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF_SET(__pyx_v__alt, __pyx_t_1);
          __pyx_t_1 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":6930
 *         localctx = ModelicaParser.Component_referenceContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 120, self.RULE_component_reference)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 933
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":6947
 *                 _alt = self._interp.adaptivePredict(self._input,117,self._ctx)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6947, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.component_reference", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_3, &__pyx_t_4) < 0) __PYX_ERR(0, 6947, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __pyx_v_re = __pyx_t_3;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":6948
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 6948, __pyx_L24_error)

          /* "pymola/generated/ModelicaParser.pyx":6949
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6949, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6949, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6949, __pyx_L24_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6949, __pyx_L24_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 6949, __pyx_L24_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_7) {
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_7); __pyx_t_7 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_13, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6949, __pyx_L24_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6950
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6950, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 6950, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6950, __pyx_L24_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6950, __pyx_L24_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6950, __pyx_L24_error)
            __Pyx_GOTREF(__pyx_t_7);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6950, __pyx_L24_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":6947
 *                 _alt = self._interp.adaptivePredict(self._input,117,self._ctx)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L25;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L24_error:;
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_6 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L8_except_error;
          }
          __pyx_L25:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":6930
 *         localctx = ModelicaParser.Component_referenceContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 120, self.RULE_component_reference)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 933
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":6952
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6952, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6952, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6952, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_14 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6952, __pyx_L31_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6952, __pyx_L31_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6952, __pyx_L31_error)
        }
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
      __pyx_L31_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":6953
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Function_call_argsContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":6926
 * 
 * 
 *     def component_reference(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Component_referenceContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.component_reference", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__alt);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6957
 *     class Function_call_argsContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 6957, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 6957, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6957, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Function_call_argsContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_25Function_call_argsContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Function_call_argsContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":6958
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 6958, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6958, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6958, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6958, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6958, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6958, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6958, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6958, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6959
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def function_arguments(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 6959, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6957
 *     class Function_call_argsContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Function_call_argsContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6961
 *             self.parser = parser
 * 
 *         def function_arguments(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Function_argumentsContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_3function_arguments(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_3function_arguments = {"function_arguments", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_3function_arguments, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_3function_arguments(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("function_arguments (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_2function_arguments(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_2function_arguments(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("function_arguments", 0);

  /* "pymola/generated/ModelicaParser.pyx":6962
 * 
 *         def function_arguments(self):
 *             return self.getTypedRuleContext(ModelicaParser.Function_argumentsContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6962, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6962, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Function_argumentsContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6962, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6962, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6962, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6962, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6962, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":6961
 *             self.parser = parser
 * 
 *         def function_arguments(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Function_argumentsContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Function_call_argsContext.function_arguments", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6965
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_function_call_args
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_5getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_5getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_4getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":6966
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_function_call_args             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6966, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_function_call_args); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6966, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":6965
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_function_call_args
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Function_call_argsContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6968
 *             return ModelicaParser.RULE_function_call_args
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterFunction_call_args" ):
 *                 listener.enterFunction_call_args(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_7enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_7enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 6968, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 6968, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6968, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Function_call_argsContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_6enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6969
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterFunction_call_args" ):             # <<<<<<<<<<<<<<
 *                 listener.enterFunction_call_args(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterFunction_call_args); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6969, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6970
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterFunction_call_args" ):
 *                 listener.enterFunction_call_args(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterFunction_call_args); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6970, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6970, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6970, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6970, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6970, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6970, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6969
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterFunction_call_args" ):             # <<<<<<<<<<<<<<
 *                 listener.enterFunction_call_args(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6968
 *             return ModelicaParser.RULE_function_call_args
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterFunction_call_args" ):
 *                 listener.enterFunction_call_args(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Function_call_argsContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6972
 *                 listener.enterFunction_call_args(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitFunction_call_args" ):
 *                 listener.exitFunction_call_args(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_9exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_9exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 6972, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 6972, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6972, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Function_call_argsContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_8exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":6973
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitFunction_call_args" ):             # <<<<<<<<<<<<<<
 *                 listener.exitFunction_call_args(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitFunction_call_args); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6973, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6974
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitFunction_call_args" ):
 *                 listener.exitFunction_call_args(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitFunction_call_args); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6974, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6974, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6974, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6974, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6974, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6974, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":6973
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitFunction_call_args" ):             # <<<<<<<<<<<<<<
 *                 listener.exitFunction_call_args(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6972
 *                 listener.enterFunction_call_args(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitFunction_call_args" ):
 *                 listener.exitFunction_call_args(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Function_call_argsContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6976
 *                 listener.exitFunction_call_args(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitFunction_call_args" ):
 *                 return visitor.visitFunction_call_args(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_11accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_11accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 6976, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 6976, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 6976, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Function_call_argsContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_10accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":6977
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitFunction_call_args" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitFunction_call_args(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitFunction_call_args); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 6977, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":6978
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitFunction_call_args" ):
 *                 return visitor.visitFunction_call_args(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitFunction_call_args); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6978, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6978, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6978, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6978, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6978, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6978, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":6977
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitFunction_call_args" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitFunction_call_args(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":6980
 *                 return visitor.visitFunction_call_args(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6980, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6980, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6980, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6980, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6980, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6980, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":6976
 *                 listener.exitFunction_call_args(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitFunction_call_args" ):
 *                 return visitor.visitFunction_call_args(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Function_call_argsContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":6985
 * 
 * 
 *     def function_call_args(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Function_call_argsContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_125function_call_args(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_125function_call_args = {"function_call_args", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_125function_call_args, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_125function_call_args(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("function_call_args (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_124function_call_args(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_124function_call_args(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  char const *__pyx_t_22;
  __Pyx_RefNannySetupContext("function_call_args", 0);

  /* "pymola/generated/ModelicaParser.pyx":6987
 *     def function_call_args(self):
 * 
 *         localctx = ModelicaParser.Function_call_argsContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 122, self.RULE_function_call_args)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6987, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Function_call_argsContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6987, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6987, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6987, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6987, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6987, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6987, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6987, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6988
 * 
 *         localctx = ModelicaParser.Function_call_argsContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 122, self.RULE_function_call_args)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6988, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_function_call_args); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6988, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_122, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6988, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_122, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6988, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6988, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_122);
    __Pyx_GIVEREF(__pyx_int_122);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_122);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6988, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6989
 *         localctx = ModelicaParser.Function_call_argsContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 122, self.RULE_function_call_args)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 6989, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6990
 *         self.enterRule(localctx, 122, self.RULE_function_call_args)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 941
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":6991
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 941
 *             self.match(ModelicaParser.T__16)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6991, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6991, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6991, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6991, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6991, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":6992
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 941             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__16)
 *             self.state = 943
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_941) < 0) __PYX_ERR(0, 6992, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":6993
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 941
 *             self.match(ModelicaParser.T__16)             # <<<<<<<<<<<<<<
 *             self.state = 943
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6993, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6993, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__16); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6993, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6993, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6993, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6993, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6993, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6993, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":6994
 *             self.state = 941
 *             self.match(ModelicaParser.T__16)
 *             self.state = 943             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_943) < 0) __PYX_ERR(0, 6994, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":6995
 *             self.match(ModelicaParser.T__16)
 *             self.state = 943
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__12) | (1 << ModelicaParser.T__13) | (1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__19) | (1 << ModelicaParser.T__29) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__55))) != 0) or ((((_la - 71)) & ~0x3f) == 0 and ((1 << (_la - 71)) & ((1 << (ModelicaParser.T__70 - 71)) | (1 << (ModelicaParser.T__73 - 71)) | (1 << (ModelicaParser.T__74 - 71)) | (1 << (ModelicaParser.T__75 - 71)) | (1 << (ModelicaParser.INITIAL - 71)) | (1 << (ModelicaParser.IDENT - 71)) | (1 << (ModelicaParser.STRING - 71)) | (1 << (ModelicaParser.UNSIGNED_NUMBER - 71)))) != 0):
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6995, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6995, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6995, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6995, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6995, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6995, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6995, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":6996
 *             self.state = 943
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__12) | (1 << ModelicaParser.T__13) | (1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__19) | (1 << ModelicaParser.T__29) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__55))) != 0) or ((((_la - 71)) & ~0x3f) == 0 and ((1 << (_la - 71)) & ((1 << (ModelicaParser.T__70 - 71)) | (1 << (ModelicaParser.T__73 - 71)) | (1 << (ModelicaParser.T__74 - 71)) | (1 << (ModelicaParser.T__75 - 71)) | (1 << (ModelicaParser.INITIAL - 71)) | (1 << (ModelicaParser.IDENT - 71)) | (1 << (ModelicaParser.STRING - 71)) | (1 << (ModelicaParser.UNSIGNED_NUMBER - 71)))) != 0):
 *                 self.state = 942
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6996, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6996, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__955, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6996, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_v__la = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":6997
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__12) | (1 << ModelicaParser.T__13) | (1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__19) | (1 << ModelicaParser.T__29) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__55))) != 0) or ((((_la - 71)) & ~0x3f) == 0 and ((1 << (_la - 71)) & ((1 << (ModelicaParser.T__70 - 71)) | (1 << (ModelicaParser.T__73 - 71)) | (1 << (ModelicaParser.T__74 - 71)) | (1 << (ModelicaParser.T__75 - 71)) | (1 << (ModelicaParser.INITIAL - 71)) | (1 << (ModelicaParser.IDENT - 71)) | (1 << (ModelicaParser.STRING - 71)) | (1 << (ModelicaParser.UNSIGNED_NUMBER - 71)))) != 0):             # <<<<<<<<<<<<<<
 *                 self.state = 942
 *                 self.function_arguments()
 */
        __pyx_t_1 = __Pyx_PyInt_From_long((~0x3f)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = PyNumber_And(__pyx_v__la, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_t_4, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (!__pyx_t_12) {
          goto __pyx_L16_next_or;
        } else {
        }
        __pyx_t_1 = PyNumber_Lshift(__pyx_int_1, __pyx_v__la); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__12); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__13); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_Or(__pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__16); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyNumber_Or(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__19); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_Or(__pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__29); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyNumber_Or(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__39); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_Or(__pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__55); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyNumber_Or(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_And(__pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_2, __pyx_int_0, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_L16_next_or:;
        __pyx_t_4 = __Pyx_PyInt_SubtractObjC(__pyx_v__la, __pyx_int_71, 71, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyInt_From_long((~0x3f)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = PyNumber_And(__pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_t_1, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_v__la, __pyx_int_71, 71, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__70); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_t_4, __pyx_int_71, 71, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__73); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_t_3, __pyx_int_71, 71, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_Or(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__74); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyInt_SubtractObjC(__pyx_t_4, __pyx_int_71, 71, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_Or(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__75); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyInt_SubtractObjC(__pyx_t_2, __pyx_int_71, 71, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyNumber_Or(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_INITIAL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_t_3, __pyx_int_71, 71, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_Or(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyInt_SubtractObjC(__pyx_t_4, __pyx_int_71, 71, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_Or(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_STRING); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyInt_SubtractObjC(__pyx_t_2, __pyx_int_71, 71, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyNumber_Or(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_UNSIGNED_NUMBER); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_t_3, __pyx_int_71, 71, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_Or(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_And(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_3, __pyx_int_0, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 6997, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_11 = __pyx_t_12;
        __pyx_L15_bool_binop_done:;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":6998
 *             _la = self._input.LA(1)
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__12) | (1 << ModelicaParser.T__13) | (1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__19) | (1 << ModelicaParser.T__29) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__55))) != 0) or ((((_la - 71)) & ~0x3f) == 0 and ((1 << (_la - 71)) & ((1 << (ModelicaParser.T__70 - 71)) | (1 << (ModelicaParser.T__73 - 71)) | (1 << (ModelicaParser.T__74 - 71)) | (1 << (ModelicaParser.T__75 - 71)) | (1 << (ModelicaParser.INITIAL - 71)) | (1 << (ModelicaParser.IDENT - 71)) | (1 << (ModelicaParser.STRING - 71)) | (1 << (ModelicaParser.UNSIGNED_NUMBER - 71)))) != 0):
 *                 self.state = 942             # <<<<<<<<<<<<<<
 *                 self.function_arguments()
 * 
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_942) < 0) __PYX_ERR(0, 6998, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":6999
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__12) | (1 << ModelicaParser.T__13) | (1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__19) | (1 << ModelicaParser.T__29) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__55))) != 0) or ((((_la - 71)) & ~0x3f) == 0 and ((1 << (_la - 71)) & ((1 << (ModelicaParser.T__70 - 71)) | (1 << (ModelicaParser.T__73 - 71)) | (1 << (ModelicaParser.T__74 - 71)) | (1 << (ModelicaParser.T__75 - 71)) | (1 << (ModelicaParser.INITIAL - 71)) | (1 << (ModelicaParser.IDENT - 71)) | (1 << (ModelicaParser.STRING - 71)) | (1 << (ModelicaParser.UNSIGNED_NUMBER - 71)))) != 0):
 *                 self.state = 942
 *                 self.function_arguments()             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_function_arguments); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6999, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (__pyx_t_1) {
            __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6999, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          } else {
            __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6999, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":6997
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__12) | (1 << ModelicaParser.T__13) | (1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__19) | (1 << ModelicaParser.T__29) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__55))) != 0) or ((((_la - 71)) & ~0x3f) == 0 and ((1 << (_la - 71)) & ((1 << (ModelicaParser.T__70 - 71)) | (1 << (ModelicaParser.T__73 - 71)) | (1 << (ModelicaParser.T__74 - 71)) | (1 << (ModelicaParser.T__75 - 71)) | (1 << (ModelicaParser.INITIAL - 71)) | (1 << (ModelicaParser.IDENT - 71)) | (1 << (ModelicaParser.STRING - 71)) | (1 << (ModelicaParser.UNSIGNED_NUMBER - 71)))) != 0):             # <<<<<<<<<<<<<<
 *                 self.state = 942
 *                 self.function_arguments()
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":7002
 * 
 * 
 *             self.state = 945             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__18)
 *         except RecognitionException as re:
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_945) < 0) __PYX_ERR(0, 7002, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":7003
 * 
 *             self.state = 945
 *             self.match(ModelicaParser.T__18)             # <<<<<<<<<<<<<<
 *         except RecognitionException as re:
 *             localctx.exception = re
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7003, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7003, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__18); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7003, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_1) {
          __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7003, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_t_4};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7003, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_t_4};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7003, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7003, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_1); __pyx_t_1 = NULL;
            __Pyx_GIVEREF(__pyx_t_4);
            PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_4);
            __pyx_t_4 = 0;
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7003, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "pymola/generated/ModelicaParser.pyx":6990
 *         self.enterRule(localctx, 122, self.RULE_function_call_args)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 941
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "pymola/generated/ModelicaParser.pyx":7004
 *             self.state = 945
 *             self.match(ModelicaParser.T__18)
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7004, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.function_call_args", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_3, &__pyx_t_7) < 0) __PYX_ERR(0, 7004, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_3);
        __pyx_v_re = __pyx_t_3;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":7005
 *             self.match(ModelicaParser.T__18)
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 7005, __pyx_L24_error)

          /* "pymola/generated/ModelicaParser.pyx":7006
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7006, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7006, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_self, __pyx_v_re};
            __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7006, __pyx_L24_error)
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_self, __pyx_v_re};
            __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7006, __pyx_L24_error)
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 7006, __pyx_L24_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_1) {
              __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_1); __pyx_t_1 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_13, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7006, __pyx_L24_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7007
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7007, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 7007, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7007, __pyx_L24_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7007, __pyx_L24_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          {
            __pyx_t_1 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7007, __pyx_L24_error)
            __Pyx_GOTREF(__pyx_t_1);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_1, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7007, __pyx_L24_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":7004
 *             self.state = 945
 *             self.match(ModelicaParser.T__18)
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L25;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L24_error:;
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_6 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L8_except_error;
          }
          __pyx_L25:;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":6990
 *         self.enterRule(localctx, 122, self.RULE_function_call_args)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 941
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":7009
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7009, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_2) {
        __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7009, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      } else {
        __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7009, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_14 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7009, __pyx_L31_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_2) {
          __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7009, __pyx_L31_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        } else {
          __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7009, __pyx_L31_error)
        }
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
      __pyx_L31_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":7010
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Function_argumentsContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":6985
 * 
 * 
 *     def function_call_args(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Function_call_argsContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.function_call_args", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7014
 *     class Function_argumentsContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 7014, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 7014, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7014, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Function_argumentsContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_25Function_argumentsContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Function_argumentsContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":7015
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 7015, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7015, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7015, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7015, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7015, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7016
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def function_argument(self, i:int=None):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 7016, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7014
 *     class Function_argumentsContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Function_argumentsContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7018
 *             self.parser = parser
 * 
 *         def function_argument(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Function_argumentContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_3function_argument(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_3function_argument = {"function_argument", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_3function_argument, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_3function_argument(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("function_argument (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "function_argument") < 0)) __PYX_ERR(0, 7018, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("function_argument", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7018, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Function_argumentsContext.function_argument", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_2function_argument(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_2function_argument(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("function_argument", 0);

  /* "pymola/generated/ModelicaParser.pyx":7019
 * 
 *         def function_argument(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.Function_argumentContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7020
 *         def function_argument(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Function_argumentContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.Function_argumentContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7020, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7020, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_Function_argumentContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7020, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7020, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7020, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7020, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7020, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7020, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":7019
 * 
 *         def function_argument(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.Function_argumentContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7022
 *                 return self.getTypedRuleContexts(ModelicaParser.Function_argumentContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.Function_argumentContext,i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7022, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7022, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Function_argumentContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7022, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7022, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7022, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7022, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7022, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":7018
 *             self.parser = parser
 * 
 *         def function_argument(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Function_argumentContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Function_argumentsContext.function_argument", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7025
 * 
 * 
 *         def for_indices(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.For_indicesContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_5for_indices(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_5for_indices = {"for_indices", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_5for_indices, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_5for_indices(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("for_indices (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "for_indices") < 0)) __PYX_ERR(0, 7025, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("for_indices", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7025, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Function_argumentsContext.for_indices", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_4for_indices(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_4for_indices(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("for_indices", 0);

  /* "pymola/generated/ModelicaParser.pyx":7026
 * 
 *         def for_indices(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.For_indicesContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7027
 *         def for_indices(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.For_indicesContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.For_indicesContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7027, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7027, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_For_indicesContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7027, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7027, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7027, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7027, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7027, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7027, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":7026
 * 
 *         def for_indices(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.For_indicesContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7029
 *                 return self.getTypedRuleContexts(ModelicaParser.For_indicesContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.For_indicesContext,i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7029, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7029, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_For_indicesContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7029, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7029, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7029, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7029, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7029, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":7025
 * 
 * 
 *         def for_indices(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.For_indicesContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Function_argumentsContext.for_indices", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7032
 * 
 * 
 *         def named_arguments(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Named_argumentsContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_7named_arguments(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_7named_arguments = {"named_arguments", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_7named_arguments, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_7named_arguments(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("named_arguments (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_6named_arguments(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_6named_arguments(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("named_arguments", 0);

  /* "pymola/generated/ModelicaParser.pyx":7033
 * 
 *         def named_arguments(self):
 *             return self.getTypedRuleContext(ModelicaParser.Named_argumentsContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7033, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7033, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Named_argumentsContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7033, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7033, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7033, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7033, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7033, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":7032
 * 
 * 
 *         def named_arguments(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Named_argumentsContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Function_argumentsContext.named_arguments", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7036
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_function_arguments
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_9getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_9getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_9getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_9getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_8getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_8getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":7037
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_function_arguments             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7037, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_function_arguments); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7037, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":7036
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_function_arguments
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Function_argumentsContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7039
 *             return ModelicaParser.RULE_function_arguments
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterFunction_arguments" ):
 *                 listener.enterFunction_arguments(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_11enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_11enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_11enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_11enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 7039, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 7039, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7039, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Function_argumentsContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_10enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_10enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":7040
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterFunction_arguments" ):             # <<<<<<<<<<<<<<
 *                 listener.enterFunction_arguments(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterFunction_arguments); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 7040, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7041
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterFunction_arguments" ):
 *                 listener.enterFunction_arguments(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterFunction_arguments); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7041, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7041, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7041, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7041, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7041, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7041, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":7040
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterFunction_arguments" ):             # <<<<<<<<<<<<<<
 *                 listener.enterFunction_arguments(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7039
 *             return ModelicaParser.RULE_function_arguments
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterFunction_arguments" ):
 *                 listener.enterFunction_arguments(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Function_argumentsContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7043
 *                 listener.enterFunction_arguments(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitFunction_arguments" ):
 *                 listener.exitFunction_arguments(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_13exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_13exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_13exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_13exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 7043, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 7043, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7043, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Function_argumentsContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_12exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_12exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":7044
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitFunction_arguments" ):             # <<<<<<<<<<<<<<
 *                 listener.exitFunction_arguments(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitFunction_arguments); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 7044, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7045
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitFunction_arguments" ):
 *                 listener.exitFunction_arguments(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitFunction_arguments); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7045, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7045, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7045, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7045, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7045, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7045, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":7044
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitFunction_arguments" ):             # <<<<<<<<<<<<<<
 *                 listener.exitFunction_arguments(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7043
 *                 listener.enterFunction_arguments(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitFunction_arguments" ):
 *                 listener.exitFunction_arguments(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Function_argumentsContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7047
 *                 listener.exitFunction_arguments(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitFunction_arguments" ):
 *                 return visitor.visitFunction_arguments(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_15accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_15accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_15accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_15accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 7047, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 7047, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7047, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Function_argumentsContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_14accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_14accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":7048
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitFunction_arguments" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitFunction_arguments(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitFunction_arguments); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 7048, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7049
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitFunction_arguments" ):
 *                 return visitor.visitFunction_arguments(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitFunction_arguments); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7049, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7049, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7049, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7049, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7049, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7049, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":7048
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitFunction_arguments" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitFunction_arguments(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7051
 *                 return visitor.visitFunction_arguments(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7051, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7051, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7051, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7051, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7051, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7051, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":7047
 *                 listener.exitFunction_arguments(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitFunction_arguments" ):
 *                 return visitor.visitFunction_arguments(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Function_argumentsContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7056
 * 
 * 
 *     def function_arguments(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Function_argumentsContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_127function_arguments(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_127function_arguments = {"function_arguments", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_127function_arguments, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_127function_arguments(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("function_arguments (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_126function_arguments(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_126function_arguments(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v_la_ = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_token = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  char const *__pyx_t_22;
  __Pyx_RefNannySetupContext("function_arguments", 0);

  /* "pymola/generated/ModelicaParser.pyx":7058
 *     def function_arguments(self):
 * 
 *         localctx = ModelicaParser.Function_argumentsContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 124, self.RULE_function_arguments)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7058, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Function_argumentsContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7058, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7058, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7058, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7058, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7058, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7058, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7058, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7059
 * 
 *         localctx = ModelicaParser.Function_argumentsContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 124, self.RULE_function_arguments)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7059, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_function_arguments); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7059, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_124, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7059, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_124, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7059, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7059, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_124);
    __Pyx_GIVEREF(__pyx_int_124);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_124);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7059, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7060
 *         localctx = ModelicaParser.Function_argumentsContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 124, self.RULE_function_arguments)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.state = 958
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 7060, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7061
 *         self.enterRule(localctx, 124, self.RULE_function_arguments)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.state = 958
 *             self._errHandler.sync(self)
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":7062
 *         self._la = 0 # Token type
 *         try:
 *             self.state = 958             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             la_ = self._interp.adaptivePredict(self._input,121,self._ctx)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_958) < 0) __PYX_ERR(0, 7062, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":7063
 *         try:
 *             self.state = 958
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             la_ = self._interp.adaptivePredict(self._input,121,self._ctx)
 *             if la_ == 1:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7063, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7063, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7063, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7063, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7063, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7063, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7063, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":7064
 *             self.state = 958
 *             self._errHandler.sync(self)
 *             la_ = self._interp.adaptivePredict(self._input,121,self._ctx)             # <<<<<<<<<<<<<<
 *             if la_ == 1:
 *                 self.enterOuterAlt(localctx, 1)
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_interp); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7064, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_adaptivePredict); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7064, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7064, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7064, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_t_2, __pyx_int_121, __pyx_t_3};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7064, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_t_2, __pyx_int_121, __pyx_t_3};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7064, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7064, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_2);
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_t_2);
          __Pyx_INCREF(__pyx_int_121);
          __Pyx_GIVEREF(__pyx_int_121);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_int_121);
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_6, __pyx_t_3);
          __pyx_t_2 = 0;
          __pyx_t_3 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7064, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v_la_ = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":7065
 *             self._errHandler.sync(self)
 *             la_ = self._interp.adaptivePredict(self._input,121,self._ctx)
 *             if la_ == 1:             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 1)
 *                 self.state = 947
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7065, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 7065, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":7066
 *             la_ = self._interp.adaptivePredict(self._input,121,self._ctx)
 *             if la_ == 1:
 *                 self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *                 self.state = 947
 *                 self.function_argument()
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7066, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_localctx, __pyx_int_1};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7066, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_localctx, __pyx_int_1};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7066, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7066, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_1);
            __Pyx_GIVEREF(__pyx_int_1);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_int_1);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7066, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7067
 *             if la_ == 1:
 *                 self.enterOuterAlt(localctx, 1)
 *                 self.state = 947             # <<<<<<<<<<<<<<
 *                 self.function_argument()
 *                 self.state = 954
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_947) < 0) __PYX_ERR(0, 7067, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":7068
 *                 self.enterOuterAlt(localctx, 1)
 *                 self.state = 947
 *                 self.function_argument()             # <<<<<<<<<<<<<<
 *                 self.state = 954
 *                 self._errHandler.sync(self)
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_function_argument); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7068, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (__pyx_t_3) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7068, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7068, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7069
 *                 self.state = 947
 *                 self.function_argument()
 *                 self.state = 954             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_954) < 0) __PYX_ERR(0, 7069, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":7070
 *                 self.function_argument()
 *                 self.state = 954
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 *                 while _la==ModelicaParser.T__20 or _la==ModelicaParser.T__48:
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7070, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_sync); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7070, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7070, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7070, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7070, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7070, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7070, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7071
 *                 self.state = 954
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 while _la==ModelicaParser.T__20 or _la==ModelicaParser.T__48:
 *                     self.state = 952
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7071, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7071, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__956, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7071, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_v__la = __pyx_t_1;
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7072
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 while _la==ModelicaParser.T__20 or _la==ModelicaParser.T__48:             # <<<<<<<<<<<<<<
 *                     self.state = 952
 *                     self._errHandler.sync(self)
 */
          while (1) {
            __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7072, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__20); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7072, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7072, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 7072, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (!__pyx_t_12) {
            } else {
              __pyx_t_11 = __pyx_t_12;
              goto __pyx_L17_bool_binop_done;
            }
            __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7072, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__48); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7072, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7072, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 7072, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_11 = __pyx_t_12;
            __pyx_L17_bool_binop_done:;
            if (!__pyx_t_11) break;

            /* "pymola/generated/ModelicaParser.pyx":7073
 *                 _la = self._input.LA(1)
 *                 while _la==ModelicaParser.T__20 or _la==ModelicaParser.T__48:
 *                     self.state = 952             # <<<<<<<<<<<<<<
 *                     self._errHandler.sync(self)
 *                     token = self._input.LA(1)
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_952) < 0) __PYX_ERR(0, 7073, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":7074
 *                 while _la==ModelicaParser.T__20 or _la==ModelicaParser.T__48:
 *                     self.state = 952
 *                     self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                     token = self._input.LA(1)
 *                     if token in [ModelicaParser.T__20]:
 */
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7074, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7074, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
              __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
              if (likely(__pyx_t_3)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                __Pyx_INCREF(__pyx_t_3);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_5, function);
              }
            }
            if (!__pyx_t_3) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7074, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_5)) {
                PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7074, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
                PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7074, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else
              #endif
              {
                __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7074, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_7);
                __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
                __Pyx_INCREF(__pyx_v_self);
                __Pyx_GIVEREF(__pyx_v_self);
                PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self);
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7074, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":7075
 *                     self.state = 952
 *                     self._errHandler.sync(self)
 *                     token = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                     if token in [ModelicaParser.T__20]:
 *                         self.state = 948
 */
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7075, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7075, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_tuple__957, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7075, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF_SET(__pyx_v_token, __pyx_t_1);
            __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":7076
 *                     self._errHandler.sync(self)
 *                     token = self._input.LA(1)
 *                     if token in [ModelicaParser.T__20]:             # <<<<<<<<<<<<<<
 *                         self.state = 948
 *                         self.match(ModelicaParser.T__20)
 */
            __Pyx_INCREF(__pyx_v_token);
            __pyx_t_1 = __pyx_v_token;
            __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7076, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_T__20); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7076, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_5 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7076, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 7076, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_12 = (__pyx_t_11 != 0);
            if (__pyx_t_12) {

              /* "pymola/generated/ModelicaParser.pyx":7077
 *                     token = self._input.LA(1)
 *                     if token in [ModelicaParser.T__20]:
 *                         self.state = 948             # <<<<<<<<<<<<<<
 *                         self.match(ModelicaParser.T__20)
 *                         self.state = 949
 */
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_948) < 0) __PYX_ERR(0, 7077, __pyx_L6_error)

              /* "pymola/generated/ModelicaParser.pyx":7078
 *                     if token in [ModelicaParser.T__20]:
 *                         self.state = 948
 *                         self.match(ModelicaParser.T__20)             # <<<<<<<<<<<<<<
 *                         self.state = 949
 *                         self.function_argument()
 */
              __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7078, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7078, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__20); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7078, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __pyx_t_7 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
                __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
                if (likely(__pyx_t_7)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                  __Pyx_INCREF(__pyx_t_7);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_5, function);
                }
              }
              if (!__pyx_t_7) {
                __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7078, __pyx_L6_error)
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else {
                #if CYTHON_FAST_PYCALL
                if (PyFunction_Check(__pyx_t_5)) {
                  PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_3};
                  __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7078, __pyx_L6_error)
                  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
                  __Pyx_GOTREF(__pyx_t_1);
                  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                } else
                #endif
                #if CYTHON_FAST_PYCCALL
                if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
                  PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_3};
                  __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7078, __pyx_L6_error)
                  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
                  __Pyx_GOTREF(__pyx_t_1);
                  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                } else
                #endif
                {
                  __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7078, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_2);
                  __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_7); __pyx_t_7 = NULL;
                  __Pyx_GIVEREF(__pyx_t_3);
                  PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_3);
                  __pyx_t_3 = 0;
                  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7078, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_1);
                  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                }
              }
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "pymola/generated/ModelicaParser.pyx":7079
 *                         self.state = 948
 *                         self.match(ModelicaParser.T__20)
 *                         self.state = 949             # <<<<<<<<<<<<<<
 *                         self.function_argument()
 *                         pass
 */
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_949) < 0) __PYX_ERR(0, 7079, __pyx_L6_error)

              /* "pymola/generated/ModelicaParser.pyx":7080
 *                         self.match(ModelicaParser.T__20)
 *                         self.state = 949
 *                         self.function_argument()             # <<<<<<<<<<<<<<
 *                         pass
 *                     elif token in [ModelicaParser.T__48]:
 */
              __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_function_argument); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7080, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_2 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
                __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
                if (likely(__pyx_t_2)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                  __Pyx_INCREF(__pyx_t_2);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_5, function);
                }
              }
              if (__pyx_t_2) {
                __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7080, __pyx_L6_error)
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              } else {
                __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7080, __pyx_L6_error)
              }
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "pymola/generated/ModelicaParser.pyx":7076
 *                     self._errHandler.sync(self)
 *                     token = self._input.LA(1)
 *                     if token in [ModelicaParser.T__20]:             # <<<<<<<<<<<<<<
 *                         self.state = 948
 *                         self.match(ModelicaParser.T__20)
 */
              goto __pyx_L19;
            }

            /* "pymola/generated/ModelicaParser.pyx":7082
 *                         self.function_argument()
 *                         pass
 *                     elif token in [ModelicaParser.T__48]:             # <<<<<<<<<<<<<<
 *                         self.state = 950
 *                         self.match(ModelicaParser.T__48)
 */
            __Pyx_INCREF(__pyx_v_token);
            __pyx_t_1 = __pyx_v_token;
            __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7082, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_T__48); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7082, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_5 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7082, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 7082, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_11 = (__pyx_t_12 != 0);
            if (__pyx_t_11) {

              /* "pymola/generated/ModelicaParser.pyx":7083
 *                         pass
 *                     elif token in [ModelicaParser.T__48]:
 *                         self.state = 950             # <<<<<<<<<<<<<<
 *                         self.match(ModelicaParser.T__48)
 *                         self.state = 951
 */
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_950) < 0) __PYX_ERR(0, 7083, __pyx_L6_error)

              /* "pymola/generated/ModelicaParser.pyx":7084
 *                     elif token in [ModelicaParser.T__48]:
 *                         self.state = 950
 *                         self.match(ModelicaParser.T__48)             # <<<<<<<<<<<<<<
 *                         self.state = 951
 *                         self.for_indices()
 */
              __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7084, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7084, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__48); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7084, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __pyx_t_2 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
                __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
                if (likely(__pyx_t_2)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                  __Pyx_INCREF(__pyx_t_2);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_5, function);
                }
              }
              if (!__pyx_t_2) {
                __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7084, __pyx_L6_error)
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else {
                #if CYTHON_FAST_PYCALL
                if (PyFunction_Check(__pyx_t_5)) {
                  PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
                  __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7084, __pyx_L6_error)
                  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
                  __Pyx_GOTREF(__pyx_t_1);
                  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                } else
                #endif
                #if CYTHON_FAST_PYCCALL
                if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
                  PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
                  __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7084, __pyx_L6_error)
                  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
                  __Pyx_GOTREF(__pyx_t_1);
                  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                } else
                #endif
                {
                  __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7084, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_7);
                  __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
                  __Pyx_GIVEREF(__pyx_t_3);
                  PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_3);
                  __pyx_t_3 = 0;
                  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7084, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_1);
                  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
                }
              }
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "pymola/generated/ModelicaParser.pyx":7085
 *                         self.state = 950
 *                         self.match(ModelicaParser.T__48)
 *                         self.state = 951             # <<<<<<<<<<<<<<
 *                         self.for_indices()
 *                         pass
 */
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_951) < 0) __PYX_ERR(0, 7085, __pyx_L6_error)

              /* "pymola/generated/ModelicaParser.pyx":7086
 *                         self.match(ModelicaParser.T__48)
 *                         self.state = 951
 *                         self.for_indices()             # <<<<<<<<<<<<<<
 *                         pass
 *                     else:
 */
              __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_for_indices); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7086, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_7 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
                __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
                if (likely(__pyx_t_7)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                  __Pyx_INCREF(__pyx_t_7);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_5, function);
                }
              }
              if (__pyx_t_7) {
                __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7086, __pyx_L6_error)
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              } else {
                __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7086, __pyx_L6_error)
              }
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "pymola/generated/ModelicaParser.pyx":7082
 *                         self.function_argument()
 *                         pass
 *                     elif token in [ModelicaParser.T__48]:             # <<<<<<<<<<<<<<
 *                         self.state = 950
 *                         self.match(ModelicaParser.T__48)
 */
              goto __pyx_L19;
            }

            /* "pymola/generated/ModelicaParser.pyx":7089
 *                         pass
 *                     else:
 *                         raise NoViableAltException(self)             # <<<<<<<<<<<<<<
 * 
 *                     self.state = 956
 */
            /*else*/ {
              __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_NoViableAltException); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7089, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_7 = NULL;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
                __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
                if (likely(__pyx_t_7)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                  __Pyx_INCREF(__pyx_t_7);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_5, function);
                }
              }
              if (!__pyx_t_7) {
                __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7089, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
              } else {
                #if CYTHON_FAST_PYCALL
                if (PyFunction_Check(__pyx_t_5)) {
                  PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
                  __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7089, __pyx_L6_error)
                  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
                  __Pyx_GOTREF(__pyx_t_1);
                } else
                #endif
                #if CYTHON_FAST_PYCCALL
                if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
                  PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
                  __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7089, __pyx_L6_error)
                  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
                  __Pyx_GOTREF(__pyx_t_1);
                } else
                #endif
                {
                  __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7089, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_3);
                  __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
                  __Pyx_INCREF(__pyx_v_self);
                  __Pyx_GIVEREF(__pyx_v_self);
                  PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
                  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7089, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_1);
                  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                }
              }
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_Raise(__pyx_t_1, 0, 0, 0);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __PYX_ERR(0, 7089, __pyx_L6_error)
            }
            __pyx_L19:;

            /* "pymola/generated/ModelicaParser.pyx":7091
 *                         raise NoViableAltException(self)
 * 
 *                     self.state = 956             # <<<<<<<<<<<<<<
 *                     self._errHandler.sync(self)
 *                     _la = self._input.LA(1)
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_956) < 0) __PYX_ERR(0, 7091, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":7092
 * 
 *                     self.state = 956
 *                     self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                     _la = self._input.LA(1)
 * 
 */
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7092, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_sync); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7092, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_5 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
              if (likely(__pyx_t_5)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                __Pyx_INCREF(__pyx_t_5);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_3, function);
              }
            }
            if (!__pyx_t_5) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7092, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_3)) {
                PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7092, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
                PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7092, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else
              #endif
              {
                __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7092, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_7);
                __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
                __Pyx_INCREF(__pyx_v_self);
                __Pyx_GIVEREF(__pyx_v_self);
                PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self);
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7092, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":7093
 *                     self.state = 956
 *                     self._errHandler.sync(self)
 *                     _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *                 pass
 */
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7093, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7093, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__958, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7093, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_1);
            __pyx_t_1 = 0;
          }

          /* "pymola/generated/ModelicaParser.pyx":7065
 *             self._errHandler.sync(self)
 *             la_ = self._interp.adaptivePredict(self._input,121,self._ctx)
 *             if la_ == 1:             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 1)
 *                 self.state = 947
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":7097
 *                 pass
 * 
 *             elif la_ == 2:             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 2)
 *                 self.state = 957
 */
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_la_, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7097, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 7097, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":7098
 * 
 *             elif la_ == 2:
 *                 self.enterOuterAlt(localctx, 2)             # <<<<<<<<<<<<<<
 *                 self.state = 957
 *                 self.named_arguments()
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7098, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_7 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_localctx, __pyx_int_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7098, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_localctx, __pyx_int_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7098, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7098, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            if (__pyx_t_7) {
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_2);
            __Pyx_GIVEREF(__pyx_int_2);
            PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_int_2);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7098, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7099
 *             elif la_ == 2:
 *                 self.enterOuterAlt(localctx, 2)
 *                 self.state = 957             # <<<<<<<<<<<<<<
 *                 self.named_arguments()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_957) < 0) __PYX_ERR(0, 7099, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":7100
 *                 self.enterOuterAlt(localctx, 2)
 *                 self.state = 957
 *                 self.named_arguments()             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_named_arguments); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7100, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          if (__pyx_t_5) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7100, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7100, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7097
 *                 pass
 * 
 *             elif la_ == 2:             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 2)
 *                 self.state = 957
 */
        }
        __pyx_L14:;

        /* "pymola/generated/ModelicaParser.pyx":7061
 *         self.enterRule(localctx, 124, self.RULE_function_arguments)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.state = 958
 *             self._errHandler.sync(self)
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":7104
 * 
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7104, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.function_arguments", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_3, &__pyx_t_5) < 0) __PYX_ERR(0, 7104, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_3);
        __pyx_v_re = __pyx_t_3;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":7105
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 7105, __pyx_L25_error)

          /* "pymola/generated/ModelicaParser.pyx":7106
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7106, __pyx_L25_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_reportError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7106, __pyx_L25_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7106, __pyx_L25_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7106, __pyx_L25_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 7106, __pyx_L25_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_2) {
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_2); __pyx_t_2 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_13, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7106, __pyx_L25_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7107
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7107, __pyx_L25_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_recover); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 7107, __pyx_L25_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7107, __pyx_L25_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7107, __pyx_L25_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7107, __pyx_L25_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (__pyx_t_4) {
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_2, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7107, __pyx_L25_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":7104
 * 
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L26;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L25_error:;
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_6 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L8_except_error;
          }
          __pyx_L26:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":7061
 *         self.enterRule(localctx, 124, self.RULE_function_arguments)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.state = 958
 *             self._errHandler.sync(self)
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":7109
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7109, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7109, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_5 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7109, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_14 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7109, __pyx_L32_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7109, __pyx_L32_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_5 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7109, __pyx_L32_error)
        }
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
      __pyx_L32_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":7110
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Named_argumentsContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":7056
 * 
 * 
 *     def function_arguments(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Function_argumentsContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.function_arguments", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v_la_);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_token);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7114
 *     class Named_argumentsContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 7114, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 7114, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7114, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Named_argumentsContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Named_argumentsContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Named_argumentsContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":7115
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 7115, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7115, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7115, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7116
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def named_argument(self, i:int=None):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 7116, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7114
 *     class Named_argumentsContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Named_argumentsContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7118
 *             self.parser = parser
 * 
 *         def named_argument(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Named_argumentContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_3named_argument(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_3named_argument = {"named_argument", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_3named_argument, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_3named_argument(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("named_argument (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "named_argument") < 0)) __PYX_ERR(0, 7118, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("named_argument", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7118, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Named_argumentsContext.named_argument", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_2named_argument(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_2named_argument(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("named_argument", 0);

  /* "pymola/generated/ModelicaParser.pyx":7119
 * 
 *         def named_argument(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.Named_argumentContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7120
 *         def named_argument(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Named_argumentContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.Named_argumentContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7120, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7120, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_Named_argumentContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7120, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7120, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7120, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7120, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7120, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7120, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":7119
 * 
 *         def named_argument(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.Named_argumentContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7122
 *                 return self.getTypedRuleContexts(ModelicaParser.Named_argumentContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.Named_argumentContext,i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Named_argumentContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7122, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7122, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7122, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7122, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":7118
 *             self.parser = parser
 * 
 *         def named_argument(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Named_argumentContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Named_argumentsContext.named_argument", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7125
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_named_arguments
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_5getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_5getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_4getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":7126
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_named_arguments             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_named_arguments); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":7125
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_named_arguments
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Named_argumentsContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7128
 *             return ModelicaParser.RULE_named_arguments
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterNamed_arguments" ):
 *                 listener.enterNamed_arguments(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_7enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_7enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 7128, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 7128, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7128, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Named_argumentsContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_6enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":7129
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterNamed_arguments" ):             # <<<<<<<<<<<<<<
 *                 listener.enterNamed_arguments(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterNamed_arguments); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 7129, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7130
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterNamed_arguments" ):
 *                 listener.enterNamed_arguments(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterNamed_arguments); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7130, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7130, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7130, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7130, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7130, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7130, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":7129
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterNamed_arguments" ):             # <<<<<<<<<<<<<<
 *                 listener.enterNamed_arguments(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7128
 *             return ModelicaParser.RULE_named_arguments
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterNamed_arguments" ):
 *                 listener.enterNamed_arguments(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Named_argumentsContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7132
 *                 listener.enterNamed_arguments(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitNamed_arguments" ):
 *                 listener.exitNamed_arguments(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_9exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_9exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 7132, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 7132, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7132, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Named_argumentsContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_8exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":7133
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitNamed_arguments" ):             # <<<<<<<<<<<<<<
 *                 listener.exitNamed_arguments(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitNamed_arguments); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 7133, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7134
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitNamed_arguments" ):
 *                 listener.exitNamed_arguments(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitNamed_arguments); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7134, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7134, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7134, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7134, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7134, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":7133
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitNamed_arguments" ):             # <<<<<<<<<<<<<<
 *                 listener.exitNamed_arguments(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7132
 *                 listener.enterNamed_arguments(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitNamed_arguments" ):
 *                 listener.exitNamed_arguments(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Named_argumentsContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7136
 *                 listener.exitNamed_arguments(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitNamed_arguments" ):
 *                 return visitor.visitNamed_arguments(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_11accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_11accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 7136, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 7136, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7136, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Named_argumentsContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_10accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":7137
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitNamed_arguments" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitNamed_arguments(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitNamed_arguments); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 7137, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7138
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitNamed_arguments" ):
 *                 return visitor.visitNamed_arguments(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitNamed_arguments); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7138, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7138, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7138, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7138, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7138, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7138, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":7137
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitNamed_arguments" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitNamed_arguments(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7140
 *                 return visitor.visitNamed_arguments(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7140, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7140, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7140, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7140, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7140, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":7136
 *                 listener.exitNamed_arguments(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitNamed_arguments" ):
 *                 return visitor.visitNamed_arguments(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Named_argumentsContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7145
 * 
 * 
 *     def named_arguments(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Named_argumentsContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_129named_arguments(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_129named_arguments = {"named_arguments", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_129named_arguments, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_129named_arguments(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("named_arguments (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_128named_arguments(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_128named_arguments(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  char const *__pyx_t_21;
  __Pyx_RefNannySetupContext("named_arguments", 0);

  /* "pymola/generated/ModelicaParser.pyx":7147
 *     def named_arguments(self):
 * 
 *         localctx = ModelicaParser.Named_argumentsContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 126, self.RULE_named_arguments)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Named_argumentsContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7147, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7147, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7148
 * 
 *         localctx = ModelicaParser.Named_argumentsContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 126, self.RULE_named_arguments)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_named_arguments); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_126, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7148, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_126, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7148, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_126);
    __Pyx_GIVEREF(__pyx_int_126);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_126);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7149
 *         localctx = ModelicaParser.Named_argumentsContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 126, self.RULE_named_arguments)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 7149, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7150
 *         self.enterRule(localctx, 126, self.RULE_named_arguments)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 960
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":7151
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 960
 *             self.named_argument()
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7151, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7151, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7151, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7151, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7151, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":7152
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 960             # <<<<<<<<<<<<<<
 *             self.named_argument()
 *             self.state = 965
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_960) < 0) __PYX_ERR(0, 7152, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":7153
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 960
 *             self.named_argument()             # <<<<<<<<<<<<<<
 *             self.state = 965
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_named_argument); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7153, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7153, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7153, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":7154
 *             self.state = 960
 *             self.named_argument()
 *             self.state = 965             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_965) < 0) __PYX_ERR(0, 7154, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":7155
 *             self.named_argument()
 *             self.state = 965
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             while _la==ModelicaParser.T__20:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7155, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7155, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7155, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7155, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7155, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7155, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7155, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":7156
 *             self.state = 965
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             while _la==ModelicaParser.T__20:
 *                 self.state = 961
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7156, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7156, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__959, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7156, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v__la = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":7157
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             while _la==ModelicaParser.T__20:             # <<<<<<<<<<<<<<
 *                 self.state = 961
 *                 self.match(ModelicaParser.T__20)
 */
        while (1) {
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7157, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__20); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7157, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7157, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 7157, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (!__pyx_t_11) break;

          /* "pymola/generated/ModelicaParser.pyx":7158
 *             _la = self._input.LA(1)
 *             while _la==ModelicaParser.T__20:
 *                 self.state = 961             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__20)
 *                 self.state = 962
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_961) < 0) __PYX_ERR(0, 7158, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":7159
 *             while _la==ModelicaParser.T__20:
 *                 self.state = 961
 *                 self.match(ModelicaParser.T__20)             # <<<<<<<<<<<<<<
 *                 self.state = 962
 *                 self.named_argument()
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7159, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7159, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__20); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7159, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7159, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7159, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7159, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7159, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7159, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7160
 *                 self.state = 961
 *                 self.match(ModelicaParser.T__20)
 *                 self.state = 962             # <<<<<<<<<<<<<<
 *                 self.named_argument()
 *                 self.state = 967
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_962) < 0) __PYX_ERR(0, 7160, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":7161
 *                 self.match(ModelicaParser.T__20)
 *                 self.state = 962
 *                 self.named_argument()             # <<<<<<<<<<<<<<
 *                 self.state = 967
 *                 self._errHandler.sync(self)
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_named_argument); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7161, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_4 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (__pyx_t_4) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7161, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7161, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7162
 *                 self.state = 962
 *                 self.named_argument()
 *                 self.state = 967             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_967) < 0) __PYX_ERR(0, 7162, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":7163
 *                 self.named_argument()
 *                 self.state = 967
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 * 
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7163, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_sync); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7163, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7163, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7163, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7163, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7163, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7163, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7164
 *                 self.state = 967
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *         except RecognitionException as re:
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7164, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7164, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__960, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7164, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_1);
          __pyx_t_1 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":7150
 *         self.enterRule(localctx, 126, self.RULE_named_arguments)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 960
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":7166
 *                 _la = self._input.LA(1)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7166, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.named_arguments", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_4, &__pyx_t_3) < 0) __PYX_ERR(0, 7166, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __pyx_v_re = __pyx_t_4;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":7167
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 7167, __pyx_L21_error)

          /* "pymola/generated/ModelicaParser.pyx":7168
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7168, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7168, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7168, __pyx_L21_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7168, __pyx_L21_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          {
            __pyx_t_12 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 7168, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_12);
            if (__pyx_t_2) {
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_2); __pyx_t_2 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_12, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7168, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7169
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7169, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 7169, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_12);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_12, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7169, __pyx_L21_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7169, __pyx_L21_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7169, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_2, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7169, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":7166
 *                 _la = self._input.LA(1)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L22;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L21_error:;
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __pyx_t_6 = __pyx_lineno; __pyx_t_13 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
            }
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_ErrRestore(__pyx_t_15, __pyx_t_16, __pyx_t_17);
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_13; __pyx_filename = __pyx_t_14;
            goto __pyx_L8_except_error;
          }
          __pyx_L22:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":7150
 *         self.enterRule(localctx, 126, self.RULE_named_arguments)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 960
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":7171
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7171, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7171, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7171, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_19, &__pyx_t_18);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_18);
      __pyx_t_13 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_21 = __pyx_filename;
      {
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7171, __pyx_L28_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7171, __pyx_L28_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7171, __pyx_L28_error)
        }
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __pyx_lineno = __pyx_t_13; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_21;
      goto __pyx_L1_error;
      __pyx_L28_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":7172
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Named_argumentContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":7145
 * 
 * 
 *     def named_arguments(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Named_argumentsContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.named_arguments", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7176
 *     class Named_argumentContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Named_argumentContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Named_argumentContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Named_argumentContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Named_argumentContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 7176, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 7176, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7176, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Named_argumentContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Named_argumentContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Named_argumentContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":7177
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 7177, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7177, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7177, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7178
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def IDENT(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 7178, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7176
 *     class Named_argumentContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Named_argumentContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7180
 *             self.parser = parser
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Named_argumentContext_3IDENT(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Named_argumentContext_3IDENT = {"IDENT", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Named_argumentContext_3IDENT, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Named_argumentContext_3IDENT(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("IDENT (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Named_argumentContext_2IDENT(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Named_argumentContext_2IDENT(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("IDENT", 0);

  /* "pymola/generated/ModelicaParser.pyx":7181
 * 
 *         def IDENT(self):
 *             return self.getToken(ModelicaParser.IDENT, 0)             # <<<<<<<<<<<<<<
 * 
 *         def function_argument(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getToken); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7181, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7181, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7181, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7181, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":7180
 *             self.parser = parser
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Named_argumentContext.IDENT", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7183
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 *         def function_argument(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Function_argumentContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Named_argumentContext_5function_argument(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Named_argumentContext_5function_argument = {"function_argument", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Named_argumentContext_5function_argument, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Named_argumentContext_5function_argument(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("function_argument (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Named_argumentContext_4function_argument(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Named_argumentContext_4function_argument(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("function_argument", 0);

  /* "pymola/generated/ModelicaParser.pyx":7184
 * 
 *         def function_argument(self):
 *             return self.getTypedRuleContext(ModelicaParser.Function_argumentContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Function_argumentContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7184, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7184, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7184, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7184, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":7183
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 *         def function_argument(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Function_argumentContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Named_argumentContext.function_argument", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7187
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_named_argument
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Named_argumentContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Named_argumentContext_7getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Named_argumentContext_7getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Named_argumentContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Named_argumentContext_6getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Named_argumentContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":7188
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_named_argument             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_named_argument); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":7187
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_named_argument
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Named_argumentContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7190
 *             return ModelicaParser.RULE_named_argument
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterNamed_argument" ):
 *                 listener.enterNamed_argument(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Named_argumentContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Named_argumentContext_9enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Named_argumentContext_9enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Named_argumentContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 7190, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 7190, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7190, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Named_argumentContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Named_argumentContext_8enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Named_argumentContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":7191
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterNamed_argument" ):             # <<<<<<<<<<<<<<
 *                 listener.enterNamed_argument(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterNamed_argument); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 7191, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7192
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterNamed_argument" ):
 *                 listener.enterNamed_argument(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterNamed_argument); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7192, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7192, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7192, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7192, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7192, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7192, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":7191
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterNamed_argument" ):             # <<<<<<<<<<<<<<
 *                 listener.enterNamed_argument(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7190
 *             return ModelicaParser.RULE_named_argument
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterNamed_argument" ):
 *                 listener.enterNamed_argument(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Named_argumentContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7194
 *                 listener.enterNamed_argument(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitNamed_argument" ):
 *                 listener.exitNamed_argument(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Named_argumentContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Named_argumentContext_11exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Named_argumentContext_11exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Named_argumentContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 7194, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 7194, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7194, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Named_argumentContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Named_argumentContext_10exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Named_argumentContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":7195
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitNamed_argument" ):             # <<<<<<<<<<<<<<
 *                 listener.exitNamed_argument(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitNamed_argument); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 7195, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7196
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitNamed_argument" ):
 *                 listener.exitNamed_argument(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitNamed_argument); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7196, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7196, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7196, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7196, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7196, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7196, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":7195
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitNamed_argument" ):             # <<<<<<<<<<<<<<
 *                 listener.exitNamed_argument(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7194
 *                 listener.enterNamed_argument(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitNamed_argument" ):
 *                 listener.exitNamed_argument(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Named_argumentContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7198
 *                 listener.exitNamed_argument(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitNamed_argument" ):
 *                 return visitor.visitNamed_argument(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Named_argumentContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21Named_argumentContext_13accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21Named_argumentContext_13accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21Named_argumentContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 7198, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 7198, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7198, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Named_argumentContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21Named_argumentContext_12accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21Named_argumentContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":7199
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitNamed_argument" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitNamed_argument(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitNamed_argument); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 7199, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7200
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitNamed_argument" ):
 *                 return visitor.visitNamed_argument(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitNamed_argument); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7200, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7200, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7200, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7200, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7200, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7200, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":7199
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitNamed_argument" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitNamed_argument(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7202
 *                 return visitor.visitNamed_argument(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7202, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7202, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7202, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7202, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7202, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7202, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":7198
 *                 listener.exitNamed_argument(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitNamed_argument" ):
 *                 return visitor.visitNamed_argument(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Named_argumentContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7207
 * 
 * 
 *     def named_argument(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Named_argumentContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_131named_argument(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_131named_argument = {"named_argument", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_131named_argument, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_131named_argument(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("named_argument (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_130named_argument(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_130named_argument(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  char const *__pyx_t_20;
  __Pyx_RefNannySetupContext("named_argument", 0);

  /* "pymola/generated/ModelicaParser.pyx":7209
 *     def named_argument(self):
 * 
 *         localctx = ModelicaParser.Named_argumentContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 128, self.RULE_named_argument)
 *         try:
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Named_argumentContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7209, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7209, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7209, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7209, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7210
 * 
 *         localctx = ModelicaParser.Named_argumentContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 128, self.RULE_named_argument)             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_named_argument); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_128, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7210, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_128, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7210, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7210, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_128);
    __Pyx_GIVEREF(__pyx_int_128);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_128);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7210, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7211
 *         localctx = ModelicaParser.Named_argumentContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 128, self.RULE_named_argument)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 968
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":7212
 *         self.enterRule(localctx, 128, self.RULE_named_argument)
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 968
 *             self.match(ModelicaParser.IDENT)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7212, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7212, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7212, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7212, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7212, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":7213
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 968             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.IDENT)
 *             self.state = 969
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_968) < 0) __PYX_ERR(0, 7213, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":7214
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 968
 *             self.match(ModelicaParser.IDENT)             # <<<<<<<<<<<<<<
 *             self.state = 969
 *             self.match(ModelicaParser.T__14)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7214, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7214, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7214, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7214, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7214, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7214, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7214, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7214, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":7215
 *             self.state = 968
 *             self.match(ModelicaParser.IDENT)
 *             self.state = 969             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__14)
 *             self.state = 970
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_969) < 0) __PYX_ERR(0, 7215, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":7216
 *             self.match(ModelicaParser.IDENT)
 *             self.state = 969
 *             self.match(ModelicaParser.T__14)             # <<<<<<<<<<<<<<
 *             self.state = 970
 *             self.function_argument()
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7216, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7216, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__14); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7216, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7216, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7216, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7216, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7216, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4); __pyx_t_4 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7216, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":7217
 *             self.state = 969
 *             self.match(ModelicaParser.T__14)
 *             self.state = 970             # <<<<<<<<<<<<<<
 *             self.function_argument()
 *         except RecognitionException as re:
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_970) < 0) __PYX_ERR(0, 7217, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":7218
 *             self.match(ModelicaParser.T__14)
 *             self.state = 970
 *             self.function_argument()             # <<<<<<<<<<<<<<
 *         except RecognitionException as re:
 *             localctx.exception = re
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_function_argument); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7218, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7218, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7218, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":7211
 *         localctx = ModelicaParser.Named_argumentContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 128, self.RULE_named_argument)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 968
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":7219
 *             self.state = 970
 *             self.function_argument()
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7219, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.named_argument", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_3, &__pyx_t_7) < 0) __PYX_ERR(0, 7219, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_3);
        __pyx_v_re = __pyx_t_3;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":7220
 *             self.function_argument()
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 7220, __pyx_L19_error)

          /* "pymola/generated/ModelicaParser.pyx":7221
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7221, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7221, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7221, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7221, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_11 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 7221, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_11);
            if (__pyx_t_4) {
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_4); __pyx_t_4 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_11, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7221, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7222
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7222, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 7222, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_11)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7222, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_11)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7222, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7222, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7222, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":7219
 *             self.state = 970
 *             self.function_argument()
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L20;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L19_error:;
            __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_14);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __pyx_t_6 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_17);
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
            }
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
            __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
            goto __pyx_L8_except_error;
          }
          __pyx_L20:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":7211
 *         localctx = ModelicaParser.Named_argumentContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 128, self.RULE_named_argument)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 968
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":7224
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7224, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7224, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7224, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_18, &__pyx_t_17);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_18);
      __Pyx_XGOTREF(__pyx_t_17);
      __pyx_t_12 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_20 = __pyx_filename;
      {
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7224, __pyx_L26_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7224, __pyx_L26_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7224, __pyx_L26_error)
        }
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_18, __pyx_t_17);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0;
      __pyx_lineno = __pyx_t_12; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_20;
      goto __pyx_L1_error;
      __pyx_L26_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_18, __pyx_t_17);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":7225
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Function_argumentContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":7207
 * 
 * 
 *     def named_argument(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Named_argumentContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.named_argument", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7229
 *     class Function_argumentContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Function_argumentContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Function_argumentContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Function_argumentContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Function_argumentContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 7229, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 7229, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7229, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Function_argumentContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Function_argumentContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Function_argumentContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":7230
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 7230, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7230, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7230, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7230, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7230, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7231
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 7231, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7229
 *     class Function_argumentContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Function_argumentContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7234
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_function_argument
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Function_argumentContext_3getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Function_argumentContext_3getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Function_argumentContext_3getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Function_argumentContext_3getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Function_argumentContext_2getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Function_argumentContext_2getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":7235
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_function_argument             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_function_argument); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":7234
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_function_argument
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Function_argumentContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7238
 * 
 * 
 *         def copyFrom(self, ctx:ParserRuleContext):             # <<<<<<<<<<<<<<
 *             super().copyFrom(ctx)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Function_argumentContext_5copyFrom(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Function_argumentContext_5copyFrom = {"copyFrom", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Function_argumentContext_5copyFrom, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Function_argumentContext_5copyFrom(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copyFrom (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx_2,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("copyFrom", 1, 2, 2, 1); __PYX_ERR(0, 7238, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "copyFrom") < 0)) __PYX_ERR(0, 7238, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("copyFrom", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7238, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Function_argumentContext.copyFrom", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Function_argumentContext_4copyFrom(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Function_argumentContext_4copyFrom(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("copyFrom", 0);

  /* "pymola/generated/ModelicaParser.pyx":7239
 * 
 *         def copyFrom(self, ctx:ParserRuleContext):
 *             super().copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 7239, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7239, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7239, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7239, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7239, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7239, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7238
 * 
 * 
 *         def copyFrom(self, ctx:ParserRuleContext):             # <<<<<<<<<<<<<<
 *             super().copyFrom(ctx)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Function_argumentContext.copyFrom", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7245
 *     class Argument_expressionContext(Function_argumentContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Function_argumentContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Argument_expressionContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Argument_expressionContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Argument_expressionContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Argument_expressionContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 7245, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 7245, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 7245, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7245, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Argument_expressionContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Argument_expressionContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Argument_expressionContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":7246
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Function_argumentContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 7246, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7246, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7246, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7246, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7246, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7246, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7247
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Function_argumentContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def expression(self):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7247, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7247, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7247, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7247, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7247, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7245
 *     class Argument_expressionContext(Function_argumentContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Function_argumentContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Argument_expressionContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7249
 *             self.copyFrom(ctx)
 * 
 *         def expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Argument_expressionContext_3expression(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Argument_expressionContext_3expression = {"expression", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Argument_expressionContext_3expression, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Argument_expressionContext_3expression(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("expression (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Argument_expressionContext_2expression(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Argument_expressionContext_2expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("expression", 0);

  /* "pymola/generated/ModelicaParser.pyx":7250
 * 
 *         def expression(self):
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_ExpressionContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7250, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7250, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":7249
 *             self.copyFrom(ctx)
 * 
 *         def expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Argument_expressionContext.expression", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7253
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterArgument_expression" ):
 *                 listener.enterArgument_expression(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Argument_expressionContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Argument_expressionContext_5enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Argument_expressionContext_5enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Argument_expressionContext_5enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 7253, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 7253, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7253, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Argument_expressionContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Argument_expressionContext_4enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Argument_expressionContext_4enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":7254
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterArgument_expression" ):             # <<<<<<<<<<<<<<
 *                 listener.enterArgument_expression(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterArgument_expression); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 7254, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7255
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterArgument_expression" ):
 *                 listener.enterArgument_expression(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterArgument_expression); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7255, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7255, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7255, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7255, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7255, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7255, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":7254
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterArgument_expression" ):             # <<<<<<<<<<<<<<
 *                 listener.enterArgument_expression(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7253
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterArgument_expression" ):
 *                 listener.enterArgument_expression(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Argument_expressionContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7257
 *                 listener.enterArgument_expression(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitArgument_expression" ):
 *                 listener.exitArgument_expression(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Argument_expressionContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Argument_expressionContext_7exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Argument_expressionContext_7exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Argument_expressionContext_7exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 7257, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 7257, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7257, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Argument_expressionContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Argument_expressionContext_6exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Argument_expressionContext_6exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":7258
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitArgument_expression" ):             # <<<<<<<<<<<<<<
 *                 listener.exitArgument_expression(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitArgument_expression); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 7258, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7259
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitArgument_expression" ):
 *                 listener.exitArgument_expression(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitArgument_expression); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7259, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7259, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7259, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7259, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7259, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7259, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":7258
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitArgument_expression" ):             # <<<<<<<<<<<<<<
 *                 listener.exitArgument_expression(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7257
 *                 listener.enterArgument_expression(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitArgument_expression" ):
 *                 listener.exitArgument_expression(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Argument_expressionContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7261
 *                 listener.exitArgument_expression(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitArgument_expression" ):
 *                 return visitor.visitArgument_expression(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Argument_expressionContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_26Argument_expressionContext_9accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_26Argument_expressionContext_9accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_26Argument_expressionContext_9accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 7261, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 7261, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7261, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Argument_expressionContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_26Argument_expressionContext_8accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_26Argument_expressionContext_8accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":7262
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitArgument_expression" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitArgument_expression(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitArgument_expression); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 7262, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7263
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitArgument_expression" ):
 *                 return visitor.visitArgument_expression(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitArgument_expression); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7263, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7263, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7263, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7263, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7263, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7263, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":7262
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitArgument_expression" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitArgument_expression(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7265
 *                 return visitor.visitArgument_expression(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7265, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7265, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7265, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7265, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7265, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7265, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":7261
 *                 listener.exitArgument_expression(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitArgument_expression" ):
 *                 return visitor.visitArgument_expression(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Argument_expressionContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7270
 *     class Argument_functionContext(Function_argumentContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Function_argumentContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Argument_functionContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Argument_functionContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Argument_functionContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Argument_functionContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_ctx_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 7270, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 7270, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 7270, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_ctx = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7270, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Argument_functionContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Argument_functionContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Argument_functionContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":7271
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Function_argumentContext
 *             super().__init__(parser)             # <<<<<<<<<<<<<<
 *             self.copyFrom(ctx)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 7271, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7271, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7271, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_parser};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7271, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7271, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_parser);
      __Pyx_GIVEREF(__pyx_v_parser);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_parser);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7271, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7272
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Function_argumentContext
 *             super().__init__(parser)
 *             self.copyFrom(ctx)             # <<<<<<<<<<<<<<
 * 
 *         def name(self):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyFrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7272, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7272, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_ctx};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7272, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7272, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_ctx);
      __Pyx_GIVEREF(__pyx_v_ctx);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_ctx);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7272, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7270
 *     class Argument_functionContext(Function_argumentContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Function_argumentContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Argument_functionContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7274
 *             self.copyFrom(ctx)
 * 
 *         def name(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Argument_functionContext_3name(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Argument_functionContext_3name = {"name", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Argument_functionContext_3name, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Argument_functionContext_3name(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("name (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Argument_functionContext_2name(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Argument_functionContext_2name(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("name", 0);

  /* "pymola/generated/ModelicaParser.pyx":7275
 * 
 *         def name(self):
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)             # <<<<<<<<<<<<<<
 * 
 *         def named_arguments(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_NameContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7275, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7275, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":7274
 *             self.copyFrom(ctx)
 * 
 *         def name(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Argument_functionContext.name", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7277
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)
 * 
 *         def named_arguments(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Named_argumentsContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Argument_functionContext_5named_arguments(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Argument_functionContext_5named_arguments = {"named_arguments", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Argument_functionContext_5named_arguments, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Argument_functionContext_5named_arguments(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("named_arguments (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Argument_functionContext_4named_arguments(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Argument_functionContext_4named_arguments(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("named_arguments", 0);

  /* "pymola/generated/ModelicaParser.pyx":7278
 * 
 *         def named_arguments(self):
 *             return self.getTypedRuleContext(ModelicaParser.Named_argumentsContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7278, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7278, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Named_argumentsContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7278, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7278, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7278, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7278, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7278, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":7277
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)
 * 
 *         def named_arguments(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Named_argumentsContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Argument_functionContext.named_arguments", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7281
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterArgument_function" ):
 *                 listener.enterArgument_function(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Argument_functionContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Argument_functionContext_7enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Argument_functionContext_7enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Argument_functionContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 7281, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 7281, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7281, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Argument_functionContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Argument_functionContext_6enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Argument_functionContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":7282
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterArgument_function" ):             # <<<<<<<<<<<<<<
 *                 listener.enterArgument_function(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterArgument_function); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 7282, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7283
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterArgument_function" ):
 *                 listener.enterArgument_function(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterArgument_function); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7283, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7283, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7283, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7283, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7283, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7283, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":7282
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterArgument_function" ):             # <<<<<<<<<<<<<<
 *                 listener.enterArgument_function(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7281
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterArgument_function" ):
 *                 listener.enterArgument_function(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Argument_functionContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7285
 *                 listener.enterArgument_function(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitArgument_function" ):
 *                 listener.exitArgument_function(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Argument_functionContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Argument_functionContext_9exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Argument_functionContext_9exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Argument_functionContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 7285, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 7285, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7285, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Argument_functionContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Argument_functionContext_8exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Argument_functionContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":7286
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitArgument_function" ):             # <<<<<<<<<<<<<<
 *                 listener.exitArgument_function(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitArgument_function); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 7286, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7287
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitArgument_function" ):
 *                 listener.exitArgument_function(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitArgument_function); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7287, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7287, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7287, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7287, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7287, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7287, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":7286
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitArgument_function" ):             # <<<<<<<<<<<<<<
 *                 listener.exitArgument_function(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7285
 *                 listener.enterArgument_function(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitArgument_function" ):
 *                 listener.exitArgument_function(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Argument_functionContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7289
 *                 listener.exitArgument_function(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitArgument_function" ):
 *                 return visitor.visitArgument_function(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Argument_functionContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_24Argument_functionContext_11accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_24Argument_functionContext_11accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_24Argument_functionContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 7289, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 7289, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7289, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Argument_functionContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_24Argument_functionContext_10accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_24Argument_functionContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":7290
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitArgument_function" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitArgument_function(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitArgument_function); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 7290, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7291
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitArgument_function" ):
 *                 return visitor.visitArgument_function(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitArgument_function); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7291, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7291, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7291, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7291, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7291, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7291, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":7290
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitArgument_function" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitArgument_function(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7293
 *                 return visitor.visitArgument_function(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7293, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7293, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7293, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7293, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7293, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7293, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":7289
 *                 listener.exitArgument_function(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitArgument_function" ):
 *                 return visitor.visitArgument_function(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Argument_functionContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7297
 * 
 * 
 *     def function_argument(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Function_argumentContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_133function_argument(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_133function_argument = {"function_argument", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_133function_argument, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_133function_argument(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("function_argument (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_132function_argument(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_132function_argument(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v_token = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  char const *__pyx_t_22;
  __Pyx_RefNannySetupContext("function_argument", 0);

  /* "pymola/generated/ModelicaParser.pyx":7299
 *     def function_argument(self):
 * 
 *         localctx = ModelicaParser.Function_argumentContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 130, self.RULE_function_argument)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Function_argumentContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7299, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7299, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7299, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7299, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7300
 * 
 *         localctx = ModelicaParser.Function_argumentContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 130, self.RULE_function_argument)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_function_argument); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_130, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7300, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_130, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7300, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7300, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_130);
    __Pyx_GIVEREF(__pyx_int_130);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_130);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7300, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7301
 *         localctx = ModelicaParser.Function_argumentContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 130, self.RULE_function_argument)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.state = 981
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 7301, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7302
 *         self.enterRule(localctx, 130, self.RULE_function_argument)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.state = 981
 *             self._errHandler.sync(self)
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":7303
 *         self._la = 0 # Token type
 *         try:
 *             self.state = 981             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_981) < 0) __PYX_ERR(0, 7303, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":7304
 *         try:
 *             self.state = 981
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.T__12]:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7304, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7304, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7304, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7304, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7304, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7304, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7304, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":7305
 *             self.state = 981
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if token in [ModelicaParser.T__12]:
 *                 localctx = ModelicaParser.Argument_functionContext(self, localctx)
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7305, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7305, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__961, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7305, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_v_token = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":7306
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.T__12]:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Argument_functionContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 1)
 */
        __Pyx_INCREF(__pyx_v_token);
        __pyx_t_1 = __pyx_v_token;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7306, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__12); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7306, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7306, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 7306, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_12 = (__pyx_t_11 != 0);
        if (__pyx_t_12) {

          /* "pymola/generated/ModelicaParser.pyx":7307
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.T__12]:
 *                 localctx = ModelicaParser.Argument_functionContext(self, localctx)             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 1)
 *                 self.state = 972
 */
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7307, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Argument_functionContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7307, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7307, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7307, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7307, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_2) {
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_v_localctx);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7307, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7308
 *             if token in [ModelicaParser.T__12]:
 *                 localctx = ModelicaParser.Argument_functionContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *                 self.state = 972
 *                 self.match(ModelicaParser.T__12)
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7308, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_1};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7308, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_localctx, __pyx_int_1};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7308, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7308, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_1);
            __Pyx_GIVEREF(__pyx_int_1);
            PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_1);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7308, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7309
 *                 localctx = ModelicaParser.Argument_functionContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 1)
 *                 self.state = 972             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__12)
 *                 self.state = 973
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_972) < 0) __PYX_ERR(0, 7309, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":7310
 *                 self.enterOuterAlt(localctx, 1)
 *                 self.state = 972
 *                 self.match(ModelicaParser.T__12)             # <<<<<<<<<<<<<<
 *                 self.state = 973
 *                 self.name()
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7310, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7310, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__12); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7310, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7310, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7310, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7310, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7310, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7310, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7311
 *                 self.state = 972
 *                 self.match(ModelicaParser.T__12)
 *                 self.state = 973             # <<<<<<<<<<<<<<
 *                 self.name()
 *                 self.state = 974
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_973) < 0) __PYX_ERR(0, 7311, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":7312
 *                 self.match(ModelicaParser.T__12)
 *                 self.state = 973
 *                 self.name()             # <<<<<<<<<<<<<<
 *                 self.state = 974
 *                 self.match(ModelicaParser.T__16)
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_name); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7312, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_4 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (__pyx_t_4) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7312, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7312, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7313
 *                 self.state = 973
 *                 self.name()
 *                 self.state = 974             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__16)
 *                 self.state = 976
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_974) < 0) __PYX_ERR(0, 7313, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":7314
 *                 self.name()
 *                 self.state = 974
 *                 self.match(ModelicaParser.T__16)             # <<<<<<<<<<<<<<
 *                 self.state = 976
 *                 self._errHandler.sync(self)
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7314, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7314, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_T__16); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7314, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_4) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7314, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7314, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7314, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7314, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7314, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7315
 *                 self.state = 974
 *                 self.match(ModelicaParser.T__16)
 *                 self.state = 976             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_976) < 0) __PYX_ERR(0, 7315, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":7316
 *                 self.match(ModelicaParser.T__16)
 *                 self.state = 976
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.IDENT:
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7316, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7316, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7316, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7316, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7316, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7316, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7316, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7317
 *                 self.state = 976
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if _la==ModelicaParser.IDENT:
 *                     self.state = 975
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7317, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7317, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__962, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7317, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_v__la = __pyx_t_1;
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7318
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.IDENT:             # <<<<<<<<<<<<<<
 *                     self.state = 975
 *                     self.named_arguments()
 */
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7318, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7318, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7318, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 7318, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (__pyx_t_12) {

            /* "pymola/generated/ModelicaParser.pyx":7319
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.IDENT:
 *                     self.state = 975             # <<<<<<<<<<<<<<
 *                     self.named_arguments()
 * 
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_975) < 0) __PYX_ERR(0, 7319, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":7320
 *                 if _la==ModelicaParser.IDENT:
 *                     self.state = 975
 *                     self.named_arguments()             # <<<<<<<<<<<<<<
 * 
 * 
 */
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_named_arguments); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7320, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_3 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_3)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_3);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
              }
            }
            if (__pyx_t_3) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7320, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else {
              __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7320, __pyx_L6_error)
            }
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":7318
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 if _la==ModelicaParser.IDENT:             # <<<<<<<<<<<<<<
 *                     self.state = 975
 *                     self.named_arguments()
 */
          }

          /* "pymola/generated/ModelicaParser.pyx":7323
 * 
 * 
 *                 self.state = 978             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__18)
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_978) < 0) __PYX_ERR(0, 7323, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":7324
 * 
 *                 self.state = 978
 *                 self.match(ModelicaParser.T__18)             # <<<<<<<<<<<<<<
 *                 pass
 *             elif token in [ModelicaParser.T__13, ModelicaParser.T__16, ModelicaParser.T__19, ModelicaParser.T__29, ModelicaParser.T__39, ModelicaParser.T__55, ModelicaParser.T__70, ModelicaParser.T__73, ModelicaParser.T__74, ModelicaParser.T__75, ModelicaParser.INITIAL, ModelicaParser.IDENT, ModelicaParser.STRING, ModelicaParser.UNSIGNED_NUMBER]:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7324, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7324, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__18); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7324, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_3) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7324, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7324, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7324, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7324, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __pyx_t_3 = NULL;
              __Pyx_GIVEREF(__pyx_t_7);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_7);
              __pyx_t_7 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7324, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7306
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.T__12]:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Argument_functionContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 1)
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":7326
 *                 self.match(ModelicaParser.T__18)
 *                 pass
 *             elif token in [ModelicaParser.T__13, ModelicaParser.T__16, ModelicaParser.T__19, ModelicaParser.T__29, ModelicaParser.T__39, ModelicaParser.T__55, ModelicaParser.T__70, ModelicaParser.T__73, ModelicaParser.T__74, ModelicaParser.T__75, ModelicaParser.INITIAL, ModelicaParser.IDENT, ModelicaParser.STRING, ModelicaParser.UNSIGNED_NUMBER]:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Argument_expressionContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 2)
 */
        __Pyx_INCREF(__pyx_v_token);
        __pyx_t_1 = __pyx_v_token;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__13); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L16_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__16); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L16_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__19); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L16_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__29); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L16_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__39); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L16_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__55); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L16_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__70); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L16_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__73); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L16_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__74); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L16_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__75); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L16_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_INITIAL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L16_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L16_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_STRING); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L16_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_UNSIGNED_NUMBER); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 7326, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __pyx_t_11;
        __pyx_L16_bool_binop_done:;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_11 = (__pyx_t_12 != 0);
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":7327
 *                 pass
 *             elif token in [ModelicaParser.T__13, ModelicaParser.T__16, ModelicaParser.T__19, ModelicaParser.T__29, ModelicaParser.T__39, ModelicaParser.T__55, ModelicaParser.T__70, ModelicaParser.T__73, ModelicaParser.T__74, ModelicaParser.T__75, ModelicaParser.INITIAL, ModelicaParser.IDENT, ModelicaParser.STRING, ModelicaParser.UNSIGNED_NUMBER]:
 *                 localctx = ModelicaParser.Argument_expressionContext(self, localctx)             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 2)
 *                 self.state = 980
 */
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7327, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Argument_expressionContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7327, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7327, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_localctx};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7327, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7327, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            if (__pyx_t_2) {
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_localctx);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7327, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF_SET(__pyx_v_localctx, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7328
 *             elif token in [ModelicaParser.T__13, ModelicaParser.T__16, ModelicaParser.T__19, ModelicaParser.T__29, ModelicaParser.T__39, ModelicaParser.T__55, ModelicaParser.T__70, ModelicaParser.T__73, ModelicaParser.T__74, ModelicaParser.T__75, ModelicaParser.INITIAL, ModelicaParser.IDENT, ModelicaParser.STRING, ModelicaParser.UNSIGNED_NUMBER]:
 *                 localctx = ModelicaParser.Argument_expressionContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 2)             # <<<<<<<<<<<<<<
 *                 self.state = 980
 *                 self.expression()
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7328, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_localctx, __pyx_int_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7328, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_localctx, __pyx_int_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7328, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7328, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (__pyx_t_7) {
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_7); __pyx_t_7 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_2);
            __Pyx_GIVEREF(__pyx_int_2);
            PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_2);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7328, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7329
 *                 localctx = ModelicaParser.Argument_expressionContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 2)
 *                 self.state = 980             # <<<<<<<<<<<<<<
 *                 self.expression()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_980) < 0) __PYX_ERR(0, 7329, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":7330
 *                 self.enterOuterAlt(localctx, 2)
 *                 self.state = 980
 *                 self.expression()             # <<<<<<<<<<<<<<
 *                 pass
 *             else:
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expression); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7330, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7330, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7330, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7326
 *                 self.match(ModelicaParser.T__18)
 *                 pass
 *             elif token in [ModelicaParser.T__13, ModelicaParser.T__16, ModelicaParser.T__19, ModelicaParser.T__29, ModelicaParser.T__39, ModelicaParser.T__55, ModelicaParser.T__70, ModelicaParser.T__73, ModelicaParser.T__74, ModelicaParser.T__75, ModelicaParser.INITIAL, ModelicaParser.IDENT, ModelicaParser.STRING, ModelicaParser.UNSIGNED_NUMBER]:             # <<<<<<<<<<<<<<
 *                 localctx = ModelicaParser.Argument_expressionContext(self, localctx)
 *                 self.enterOuterAlt(localctx, 2)
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":7333
 *                 pass
 *             else:
 *                 raise NoViableAltException(self)             # <<<<<<<<<<<<<<
 * 
 *         except RecognitionException as re:
 */
        /*else*/ {
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_NoViableAltException); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7333, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7333, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7333, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7333, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7333, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7333, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_Raise(__pyx_t_1, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __PYX_ERR(0, 7333, __pyx_L6_error)
        }
        __pyx_L14:;

        /* "pymola/generated/ModelicaParser.pyx":7302
 *         self.enterRule(localctx, 130, self.RULE_function_argument)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.state = 981
 *             self._errHandler.sync(self)
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":7335
 *                 raise NoViableAltException(self)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7335, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.function_argument", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_4, &__pyx_t_7) < 0) __PYX_ERR(0, 7335, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_4);
        __pyx_v_re = __pyx_t_4;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":7336
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 7336, __pyx_L35_error)

          /* "pymola/generated/ModelicaParser.pyx":7337
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7337, __pyx_L35_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7337, __pyx_L35_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7337, __pyx_L35_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7337, __pyx_L35_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 7337, __pyx_L35_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_13, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7337, __pyx_L35_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7338
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7338, __pyx_L35_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 7338, __pyx_L35_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7338, __pyx_L35_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7338, __pyx_L35_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7338, __pyx_L35_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7338, __pyx_L35_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":7335
 *                 raise NoViableAltException(self)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L36;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L35_error:;
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_6 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L8_except_error;
          }
          __pyx_L36:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":7302
 *         self.enterRule(localctx, 130, self.RULE_function_argument)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.state = 981
 *             self._errHandler.sync(self)
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":7340
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7340, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7340, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7340, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_14 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7340, __pyx_L42_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7340, __pyx_L42_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7340, __pyx_L42_error)
        }
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
      __pyx_L42_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":7341
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Output_expression_listContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":7297
 * 
 * 
 *     def function_argument(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Function_argumentContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.function_argument", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v_token);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7345
 *     class Output_expression_listContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 7345, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 7345, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7345, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Output_expression_listContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_29Output_expression_listContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Output_expression_listContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":7346
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 7346, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7346, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7346, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7346, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7346, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7346, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7346, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7346, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7347
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def expression(self, i:int=None):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 7347, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7345
 *     class Output_expression_listContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Output_expression_listContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7349
 *             self.parser = parser
 * 
 *         def expression(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_3expression(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_3expression = {"expression", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_3expression, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_3expression(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("expression (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "expression") < 0)) __PYX_ERR(0, 7349, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("expression", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7349, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Output_expression_listContext.expression", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_2expression(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_2expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("expression", 0);

  /* "pymola/generated/ModelicaParser.pyx":7350
 * 
 *         def expression(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7351
 *         def expression(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.ExpressionContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7351, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7351, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_ExpressionContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7351, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7351, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7351, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7351, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7351, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7351, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":7350
 * 
 *         def expression(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7353
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.ExpressionContext,i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7353, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7353, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_ExpressionContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7353, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7353, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7353, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7353, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7353, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":7349
 *             self.parser = parser
 * 
 *         def expression(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Output_expression_listContext.expression", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7356
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_output_expression_list
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_5getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_5getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_4getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":7357
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_output_expression_list             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_output_expression_list); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":7356
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_output_expression_list
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Output_expression_listContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7359
 *             return ModelicaParser.RULE_output_expression_list
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterOutput_expression_list" ):
 *                 listener.enterOutput_expression_list(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_7enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_7enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 7359, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 7359, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7359, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Output_expression_listContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_6enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":7360
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterOutput_expression_list" ):             # <<<<<<<<<<<<<<
 *                 listener.enterOutput_expression_list(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterOutput_expression_list); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 7360, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7361
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterOutput_expression_list" ):
 *                 listener.enterOutput_expression_list(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterOutput_expression_list); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7361, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7361, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7361, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7361, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7361, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7361, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":7360
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterOutput_expression_list" ):             # <<<<<<<<<<<<<<
 *                 listener.enterOutput_expression_list(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7359
 *             return ModelicaParser.RULE_output_expression_list
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterOutput_expression_list" ):
 *                 listener.enterOutput_expression_list(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Output_expression_listContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7363
 *                 listener.enterOutput_expression_list(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitOutput_expression_list" ):
 *                 listener.exitOutput_expression_list(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_9exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_9exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 7363, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 7363, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7363, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Output_expression_listContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_8exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":7364
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitOutput_expression_list" ):             # <<<<<<<<<<<<<<
 *                 listener.exitOutput_expression_list(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitOutput_expression_list); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 7364, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7365
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitOutput_expression_list" ):
 *                 listener.exitOutput_expression_list(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitOutput_expression_list); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7365, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7365, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7365, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7365, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7365, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7365, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":7364
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitOutput_expression_list" ):             # <<<<<<<<<<<<<<
 *                 listener.exitOutput_expression_list(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7363
 *                 listener.enterOutput_expression_list(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitOutput_expression_list" ):
 *                 listener.exitOutput_expression_list(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Output_expression_listContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7367
 *                 listener.exitOutput_expression_list(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitOutput_expression_list" ):
 *                 return visitor.visitOutput_expression_list(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_11accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_11accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 7367, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 7367, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7367, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Output_expression_listContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_10accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":7368
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitOutput_expression_list" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitOutput_expression_list(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitOutput_expression_list); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 7368, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7369
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitOutput_expression_list" ):
 *                 return visitor.visitOutput_expression_list(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitOutput_expression_list); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7369, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7369, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7369, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7369, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7369, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7369, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":7368
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitOutput_expression_list" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitOutput_expression_list(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7371
 *                 return visitor.visitOutput_expression_list(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7371, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7371, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7371, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7371, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7371, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7371, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":7367
 *                 listener.exitOutput_expression_list(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitOutput_expression_list" ):
 *                 return visitor.visitOutput_expression_list(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Output_expression_listContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7376
 * 
 * 
 *     def output_expression_list(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Output_expression_listContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_135output_expression_list(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_135output_expression_list = {"output_expression_list", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_135output_expression_list, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_135output_expression_list(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("output_expression_list (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_134output_expression_list(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_134output_expression_list(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  char const *__pyx_t_22;
  __Pyx_RefNannySetupContext("output_expression_list", 0);

  /* "pymola/generated/ModelicaParser.pyx":7378
 *     def output_expression_list(self):
 * 
 *         localctx = ModelicaParser.Output_expression_listContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 132, self.RULE_output_expression_list)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7378, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Output_expression_listContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7378, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7378, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7378, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7378, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7378, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7378, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7378, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7379
 * 
 *         localctx = ModelicaParser.Output_expression_listContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 132, self.RULE_output_expression_list)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_output_expression_list); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_132, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7379, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_132, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7379, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7379, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_132);
    __Pyx_GIVEREF(__pyx_int_132);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_132);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7379, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7380
 *         localctx = ModelicaParser.Output_expression_listContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 132, self.RULE_output_expression_list)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 7380, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7381
 *         self.enterRule(localctx, 132, self.RULE_output_expression_list)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 984
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":7382
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 984
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7382, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7382, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7382, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7382, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7382, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":7383
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 984             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_984) < 0) __PYX_ERR(0, 7383, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":7384
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 984
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__13) | (1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__19) | (1 << ModelicaParser.T__29) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__55))) != 0) or ((((_la - 71)) & ~0x3f) == 0 and ((1 << (_la - 71)) & ((1 << (ModelicaParser.T__70 - 71)) | (1 << (ModelicaParser.T__73 - 71)) | (1 << (ModelicaParser.T__74 - 71)) | (1 << (ModelicaParser.T__75 - 71)) | (1 << (ModelicaParser.INITIAL - 71)) | (1 << (ModelicaParser.IDENT - 71)) | (1 << (ModelicaParser.STRING - 71)) | (1 << (ModelicaParser.UNSIGNED_NUMBER - 71)))) != 0):
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7384, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7384, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7384, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7384, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7384, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7384, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7384, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":7385
 *             self.state = 984
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__13) | (1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__19) | (1 << ModelicaParser.T__29) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__55))) != 0) or ((((_la - 71)) & ~0x3f) == 0 and ((1 << (_la - 71)) & ((1 << (ModelicaParser.T__70 - 71)) | (1 << (ModelicaParser.T__73 - 71)) | (1 << (ModelicaParser.T__74 - 71)) | (1 << (ModelicaParser.T__75 - 71)) | (1 << (ModelicaParser.INITIAL - 71)) | (1 << (ModelicaParser.IDENT - 71)) | (1 << (ModelicaParser.STRING - 71)) | (1 << (ModelicaParser.UNSIGNED_NUMBER - 71)))) != 0):
 *                 self.state = 983
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7385, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7385, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__963, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7385, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v__la = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":7386
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__13) | (1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__19) | (1 << ModelicaParser.T__29) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__55))) != 0) or ((((_la - 71)) & ~0x3f) == 0 and ((1 << (_la - 71)) & ((1 << (ModelicaParser.T__70 - 71)) | (1 << (ModelicaParser.T__73 - 71)) | (1 << (ModelicaParser.T__74 - 71)) | (1 << (ModelicaParser.T__75 - 71)) | (1 << (ModelicaParser.INITIAL - 71)) | (1 << (ModelicaParser.IDENT - 71)) | (1 << (ModelicaParser.STRING - 71)) | (1 << (ModelicaParser.UNSIGNED_NUMBER - 71)))) != 0):             # <<<<<<<<<<<<<<
 *                 self.state = 983
 *                 self.expression()
 */
        __pyx_t_1 = __Pyx_PyInt_From_long((~0x3f)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = PyNumber_And(__pyx_v__la, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_t_7, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (!__pyx_t_12) {
          goto __pyx_L16_next_or;
        } else {
        }
        __pyx_t_1 = PyNumber_Lshift(__pyx_int_1, __pyx_v__la); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__13); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__16); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_Or(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__19); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyNumber_Or(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__29); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_Or(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__39); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyNumber_Or(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__55); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_Or(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_And(__pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_int_0, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_L16_next_or:;
        __pyx_t_3 = __Pyx_PyInt_SubtractObjC(__pyx_v__la, __pyx_int_71, 71, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_PyInt_From_long((~0x3f)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = PyNumber_And(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_t_1, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (__pyx_t_12) {
        } else {
          __pyx_t_11 = __pyx_t_12;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_v__la, __pyx_int_71, 71, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__70); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_t_3, __pyx_int_71, 71, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__73); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_t_7, __pyx_int_71, 71, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_Or(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__74); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyInt_SubtractObjC(__pyx_t_3, __pyx_int_71, 71, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyNumber_Or(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__75); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyInt_SubtractObjC(__pyx_t_2, __pyx_int_71, 71, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_Or(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_INITIAL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_t_7, __pyx_int_71, 71, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_Or(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyInt_SubtractObjC(__pyx_t_3, __pyx_int_71, 71, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_Lshift(__pyx_int_1, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyNumber_Or(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_STRING); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyInt_SubtractObjC(__pyx_t_2, __pyx_int_71, 71, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_Lshift(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_Or(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_UNSIGNED_NUMBER); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_t_7, __pyx_int_71, 71, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyNumber_Lshift(__pyx_int_1, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_Or(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyNumber_And(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_7, __pyx_int_0, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 7386, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_11 = __pyx_t_12;
        __pyx_L15_bool_binop_done:;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":7387
 *             _la = self._input.LA(1)
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__13) | (1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__19) | (1 << ModelicaParser.T__29) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__55))) != 0) or ((((_la - 71)) & ~0x3f) == 0 and ((1 << (_la - 71)) & ((1 << (ModelicaParser.T__70 - 71)) | (1 << (ModelicaParser.T__73 - 71)) | (1 << (ModelicaParser.T__74 - 71)) | (1 << (ModelicaParser.T__75 - 71)) | (1 << (ModelicaParser.INITIAL - 71)) | (1 << (ModelicaParser.IDENT - 71)) | (1 << (ModelicaParser.STRING - 71)) | (1 << (ModelicaParser.UNSIGNED_NUMBER - 71)))) != 0):
 *                 self.state = 983             # <<<<<<<<<<<<<<
 *                 self.expression()
 * 
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_983) < 0) __PYX_ERR(0, 7387, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":7388
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__13) | (1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__19) | (1 << ModelicaParser.T__29) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__55))) != 0) or ((((_la - 71)) & ~0x3f) == 0 and ((1 << (_la - 71)) & ((1 << (ModelicaParser.T__70 - 71)) | (1 << (ModelicaParser.T__73 - 71)) | (1 << (ModelicaParser.T__74 - 71)) | (1 << (ModelicaParser.T__75 - 71)) | (1 << (ModelicaParser.INITIAL - 71)) | (1 << (ModelicaParser.IDENT - 71)) | (1 << (ModelicaParser.STRING - 71)) | (1 << (ModelicaParser.UNSIGNED_NUMBER - 71)))) != 0):
 *                 self.state = 983
 *                 self.expression()             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expression); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7388, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_1 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (__pyx_t_1) {
            __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7388, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          } else {
            __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7388, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7386
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__13) | (1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__19) | (1 << ModelicaParser.T__29) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__55))) != 0) or ((((_la - 71)) & ~0x3f) == 0 and ((1 << (_la - 71)) & ((1 << (ModelicaParser.T__70 - 71)) | (1 << (ModelicaParser.T__73 - 71)) | (1 << (ModelicaParser.T__74 - 71)) | (1 << (ModelicaParser.T__75 - 71)) | (1 << (ModelicaParser.INITIAL - 71)) | (1 << (ModelicaParser.IDENT - 71)) | (1 << (ModelicaParser.STRING - 71)) | (1 << (ModelicaParser.UNSIGNED_NUMBER - 71)))) != 0):             # <<<<<<<<<<<<<<
 *                 self.state = 983
 *                 self.expression()
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":7391
 * 
 * 
 *             self.state = 990             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_990) < 0) __PYX_ERR(0, 7391, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":7392
 * 
 *             self.state = 990
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             while _la==ModelicaParser.T__20:
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7392, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_sync); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7392, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_1, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7392, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_1)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7392, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7392, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7392, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7392, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "pymola/generated/ModelicaParser.pyx":7393
 *             self.state = 990
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             while _la==ModelicaParser.T__20:
 *                 self.state = 986
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7393, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_LA); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7393, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__964, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7393, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "pymola/generated/ModelicaParser.pyx":7394
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             while _la==ModelicaParser.T__20:             # <<<<<<<<<<<<<<
 *                 self.state = 986
 *                 self.match(ModelicaParser.T__20)
 */
        while (1) {
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7394, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__20); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7394, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyObject_RichCompare(__pyx_v__la, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7394, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 7394, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          if (!__pyx_t_11) break;

          /* "pymola/generated/ModelicaParser.pyx":7395
 *             _la = self._input.LA(1)
 *             while _la==ModelicaParser.T__20:
 *                 self.state = 986             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__20)
 *                 self.state = 987
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_986) < 0) __PYX_ERR(0, 7395, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":7396
 *             while _la==ModelicaParser.T__20:
 *                 self.state = 986
 *                 self.match(ModelicaParser.T__20)             # <<<<<<<<<<<<<<
 *                 self.state = 987
 *                 self.expression()
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7396, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7396, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__20); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7396, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_1, function);
            }
          }
          if (!__pyx_t_3) {
            __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7396, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_1)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_7};
              __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7396, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_7};
              __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7396, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7396, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __pyx_t_3 = NULL;
              __Pyx_GIVEREF(__pyx_t_7);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_7);
              __pyx_t_7 = 0;
              __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7396, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7397
 *                 self.state = 986
 *                 self.match(ModelicaParser.T__20)
 *                 self.state = 987             # <<<<<<<<<<<<<<
 *                 self.expression()
 *                 self.state = 992
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_987) < 0) __PYX_ERR(0, 7397, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":7398
 *                 self.match(ModelicaParser.T__20)
 *                 self.state = 987
 *                 self.expression()             # <<<<<<<<<<<<<<
 *                 self.state = 992
 *                 self._errHandler.sync(self)
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expression); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7398, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_4 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_1, function);
            }
          }
          if (__pyx_t_4) {
            __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7398, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          } else {
            __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7398, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7399
 *                 self.state = 987
 *                 self.expression()
 *                 self.state = 992             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_992) < 0) __PYX_ERR(0, 7399, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":7400
 *                 self.expression()
 *                 self.state = 992
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 * 
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7400, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_sync); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7400, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (!__pyx_t_1) {
            __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7400, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_v_self};
              __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7400, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_GOTREF(__pyx_t_2);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_v_self};
              __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7400, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_GOTREF(__pyx_t_2);
            } else
            #endif
            {
              __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7400, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_1); __pyx_t_1 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self);
              __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7400, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7401
 *                 self.state = 992
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *         except RecognitionException as re:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7401, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_LA); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7401, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__965, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7401, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_2);
          __pyx_t_2 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":7381
 *         self.enterRule(localctx, 132, self.RULE_output_expression_list)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 984
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "pymola/generated/ModelicaParser.pyx":7403
 *                 _la = self._input.LA(1)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7403, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.output_expression_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_4, &__pyx_t_7) < 0) __PYX_ERR(0, 7403, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_4);
        __pyx_v_re = __pyx_t_4;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":7404
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 7404, __pyx_L26_error)

          /* "pymola/generated/ModelicaParser.pyx":7405
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7405, __pyx_L26_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7405, __pyx_L26_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_self, __pyx_v_re};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7405, __pyx_L26_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_self, __pyx_v_re};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7405, __pyx_L26_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 7405, __pyx_L26_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_13, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7405, __pyx_L26_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7406
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7406, __pyx_L26_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 7406, __pyx_L26_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7406, __pyx_L26_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7406, __pyx_L26_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7406, __pyx_L26_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7406, __pyx_L26_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":7403
 *                 _la = self._input.LA(1)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L27;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L26_error:;
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_6 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L8_except_error;
          }
          __pyx_L27:;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":7381
 *         self.enterRule(localctx, 132, self.RULE_output_expression_list)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 984
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":7408
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7408, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (__pyx_t_2) {
        __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7408, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      } else {
        __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7408, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_14 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7408, __pyx_L33_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (__pyx_t_2) {
          __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7408, __pyx_L33_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        } else {
          __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7408, __pyx_L33_error)
        }
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
      __pyx_L33_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":7409
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Expression_listContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":7376
 * 
 * 
 *     def output_expression_list(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Output_expression_listContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.output_expression_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7413
 *     class Expression_listContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Expression_listContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Expression_listContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Expression_listContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Expression_listContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 7413, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 7413, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7413, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expression_listContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Expression_listContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Expression_listContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":7414
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 7414, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7414, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7414, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7414, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7414, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7414, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7414, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7414, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7415
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def expression(self, i:int=None):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 7415, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7413
 *     class Expression_listContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expression_listContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7417
 *             self.parser = parser
 * 
 *         def expression(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Expression_listContext_3expression(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Expression_listContext_3expression = {"expression", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Expression_listContext_3expression, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Expression_listContext_3expression(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("expression (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "expression") < 0)) __PYX_ERR(0, 7417, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("expression", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7417, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expression_listContext.expression", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Expression_listContext_2expression(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Expression_listContext_2expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("expression", 0);

  /* "pymola/generated/ModelicaParser.pyx":7418
 * 
 *         def expression(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7419
 *         def expression(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.ExpressionContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7419, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7419, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_ExpressionContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7419, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7419, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7419, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7419, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7419, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7419, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":7418
 * 
 *         def expression(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7421
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.ExpressionContext,i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7421, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7421, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_ExpressionContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7421, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7421, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7421, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7421, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7421, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":7417
 *             self.parser = parser
 * 
 *         def expression(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expression_listContext.expression", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7424
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_expression_list
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Expression_listContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Expression_listContext_5getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Expression_listContext_5getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Expression_listContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Expression_listContext_4getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Expression_listContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":7425
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_expression_list             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_expression_list); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":7424
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_expression_list
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expression_listContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7427
 *             return ModelicaParser.RULE_expression_list
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpression_list" ):
 *                 listener.enterExpression_list(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Expression_listContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Expression_listContext_7enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Expression_listContext_7enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Expression_listContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 7427, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 7427, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7427, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expression_listContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Expression_listContext_6enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Expression_listContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":7428
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExpression_list" ):             # <<<<<<<<<<<<<<
 *                 listener.enterExpression_list(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterExpression_list); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 7428, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7429
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExpression_list" ):
 *                 listener.enterExpression_list(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterExpression_list); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7429, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7429, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7429, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7429, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7429, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":7428
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterExpression_list" ):             # <<<<<<<<<<<<<<
 *                 listener.enterExpression_list(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7427
 *             return ModelicaParser.RULE_expression_list
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpression_list" ):
 *                 listener.enterExpression_list(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expression_listContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7431
 *                 listener.enterExpression_list(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpression_list" ):
 *                 listener.exitExpression_list(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Expression_listContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Expression_listContext_9exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Expression_listContext_9exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Expression_listContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 7431, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 7431, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7431, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expression_listContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Expression_listContext_8exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Expression_listContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":7432
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExpression_list" ):             # <<<<<<<<<<<<<<
 *                 listener.exitExpression_list(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitExpression_list); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 7432, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7433
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExpression_list" ):
 *                 listener.exitExpression_list(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitExpression_list); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7433, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7433, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7433, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7433, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7433, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7433, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":7432
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitExpression_list" ):             # <<<<<<<<<<<<<<
 *                 listener.exitExpression_list(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7431
 *                 listener.enterExpression_list(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpression_list" ):
 *                 listener.exitExpression_list(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expression_listContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7435
 *                 listener.exitExpression_list(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpression_list" ):
 *                 return visitor.visitExpression_list(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Expression_listContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_22Expression_listContext_11accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_22Expression_listContext_11accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_22Expression_listContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 7435, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 7435, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7435, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expression_listContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_22Expression_listContext_10accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_22Expression_listContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":7436
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExpression_list" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitExpression_list(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitExpression_list); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 7436, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7437
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExpression_list" ):
 *                 return visitor.visitExpression_list(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitExpression_list); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7437, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7437, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7437, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7437, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7437, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7437, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":7436
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitExpression_list" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitExpression_list(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7439
 *                 return visitor.visitExpression_list(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7439, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7439, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7439, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7439, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7439, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7439, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":7435
 *                 listener.exitExpression_list(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpression_list" ):
 *                 return visitor.visitExpression_list(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Expression_listContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7444
 * 
 * 
 *     def expression_list(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Expression_listContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_137expression_list(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_137expression_list = {"expression_list", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_137expression_list, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_137expression_list(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("expression_list (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_136expression_list(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_136expression_list(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  char const *__pyx_t_21;
  __Pyx_RefNannySetupContext("expression_list", 0);

  /* "pymola/generated/ModelicaParser.pyx":7446
 *     def expression_list(self):
 * 
 *         localctx = ModelicaParser.Expression_listContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 134, self.RULE_expression_list)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7446, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Expression_listContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7446, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7446, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7446, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7446, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7446, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7446, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7446, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7447
 * 
 *         localctx = ModelicaParser.Expression_listContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 134, self.RULE_expression_list)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_expression_list); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_134, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7447, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_134, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7447, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7447, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_134);
    __Pyx_GIVEREF(__pyx_int_134);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_134);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7447, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7448
 *         localctx = ModelicaParser.Expression_listContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 134, self.RULE_expression_list)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 7448, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7449
 *         self.enterRule(localctx, 134, self.RULE_expression_list)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 993
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":7450
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 993
 *             self.expression()
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7450, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7450, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7450, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7450, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7450, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":7451
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 993             # <<<<<<<<<<<<<<
 *             self.expression()
 *             self.state = 998
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_993) < 0) __PYX_ERR(0, 7451, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":7452
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 993
 *             self.expression()             # <<<<<<<<<<<<<<
 *             self.state = 998
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expression); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7452, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7452, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7452, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":7453
 *             self.state = 993
 *             self.expression()
 *             self.state = 998             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_998) < 0) __PYX_ERR(0, 7453, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":7454
 *             self.expression()
 *             self.state = 998
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             while _la==ModelicaParser.T__20:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7454, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7454, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7454, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7454, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7454, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7454, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7454, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":7455
 *             self.state = 998
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             while _la==ModelicaParser.T__20:
 *                 self.state = 994
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7455, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7455, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__966, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7455, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v__la = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":7456
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             while _la==ModelicaParser.T__20:             # <<<<<<<<<<<<<<
 *                 self.state = 994
 *                 self.match(ModelicaParser.T__20)
 */
        while (1) {
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7456, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__20); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7456, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7456, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 7456, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (!__pyx_t_11) break;

          /* "pymola/generated/ModelicaParser.pyx":7457
 *             _la = self._input.LA(1)
 *             while _la==ModelicaParser.T__20:
 *                 self.state = 994             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__20)
 *                 self.state = 995
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_994) < 0) __PYX_ERR(0, 7457, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":7458
 *             while _la==ModelicaParser.T__20:
 *                 self.state = 994
 *                 self.match(ModelicaParser.T__20)             # <<<<<<<<<<<<<<
 *                 self.state = 995
 *                 self.expression()
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7458, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7458, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__20); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7458, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7458, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7458, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7458, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7458, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7458, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7459
 *                 self.state = 994
 *                 self.match(ModelicaParser.T__20)
 *                 self.state = 995             # <<<<<<<<<<<<<<
 *                 self.expression()
 *                 self.state = 1000
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_995) < 0) __PYX_ERR(0, 7459, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":7460
 *                 self.match(ModelicaParser.T__20)
 *                 self.state = 995
 *                 self.expression()             # <<<<<<<<<<<<<<
 *                 self.state = 1000
 *                 self._errHandler.sync(self)
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expression); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7460, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_4 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (__pyx_t_4) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7460, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7460, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7461
 *                 self.state = 995
 *                 self.expression()
 *                 self.state = 1000             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_1000) < 0) __PYX_ERR(0, 7461, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":7462
 *                 self.expression()
 *                 self.state = 1000
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 * 
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7462, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_sync); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7462, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7462, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7462, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7462, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7462, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7462, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7463
 *                 self.state = 1000
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *         except RecognitionException as re:
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7463, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7463, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__967, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7463, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_1);
          __pyx_t_1 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":7449
 *         self.enterRule(localctx, 134, self.RULE_expression_list)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 993
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":7465
 *                 _la = self._input.LA(1)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7465, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.expression_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_4, &__pyx_t_3) < 0) __PYX_ERR(0, 7465, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __pyx_v_re = __pyx_t_4;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":7466
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 7466, __pyx_L21_error)

          /* "pymola/generated/ModelicaParser.pyx":7467
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7467, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7467, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7467, __pyx_L21_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7467, __pyx_L21_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          {
            __pyx_t_12 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 7467, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_12);
            if (__pyx_t_2) {
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_2); __pyx_t_2 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_12, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7467, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7468
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7468, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 7468, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_12);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_12, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7468, __pyx_L21_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7468, __pyx_L21_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7468, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_2, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7468, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":7465
 *                 _la = self._input.LA(1)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L22;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L21_error:;
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __pyx_t_6 = __pyx_lineno; __pyx_t_13 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
            }
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_ErrRestore(__pyx_t_15, __pyx_t_16, __pyx_t_17);
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_13; __pyx_filename = __pyx_t_14;
            goto __pyx_L8_except_error;
          }
          __pyx_L22:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":7449
 *         self.enterRule(localctx, 134, self.RULE_expression_list)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 993
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":7470
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7470, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7470, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7470, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_19, &__pyx_t_18);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_18);
      __pyx_t_13 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_21 = __pyx_filename;
      {
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7470, __pyx_L28_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7470, __pyx_L28_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7470, __pyx_L28_error)
        }
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __pyx_lineno = __pyx_t_13; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_21;
      goto __pyx_L1_error;
      __pyx_L28_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":7471
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class Array_subscriptsContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":7444
 * 
 * 
 *     def expression_list(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Expression_listContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.expression_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7475
 *     class Array_subscriptsContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 7475, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 7475, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7475, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Array_subscriptsContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":7476
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 7476, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7476, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7476, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7476, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7476, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7477
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def subscript(self, i:int=None):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 7477, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7475
 *     class Array_subscriptsContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Array_subscriptsContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7479
 *             self.parser = parser
 * 
 *         def subscript(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.SubscriptContext)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_3subscript(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_3subscript = {"subscript", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_3subscript, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_3subscript(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("subscript (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "subscript") < 0)) __PYX_ERR(0, 7479, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("subscript", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7479, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Array_subscriptsContext.subscript", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_2subscript(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_2subscript(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("subscript", 0);

  /* "pymola/generated/ModelicaParser.pyx":7480
 * 
 *         def subscript(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.SubscriptContext)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7481
 *         def subscript(self, i:int=None):
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.SubscriptContext)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.SubscriptContext,i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContexts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7481, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7481, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_SubscriptContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7481, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7481, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7481, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7481, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7481, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7481, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":7480
 * 
 *         def subscript(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTypedRuleContexts(ModelicaParser.SubscriptContext)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7483
 *                 return self.getTypedRuleContexts(ModelicaParser.SubscriptContext)
 *             else:
 *                 return self.getTypedRuleContext(ModelicaParser.SubscriptContext,i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7483, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7483, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_SubscriptContext); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7483, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7483, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7483, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7483, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7483, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":7479
 *             self.parser = parser
 * 
 *         def subscript(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.SubscriptContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Array_subscriptsContext.subscript", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7486
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_array_subscripts
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_5getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_5getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_4getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":7487
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_array_subscripts             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7487, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_array_subscripts); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7487, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":7486
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_array_subscripts
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Array_subscriptsContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7489
 *             return ModelicaParser.RULE_array_subscripts
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterArray_subscripts" ):
 *                 listener.enterArray_subscripts(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_7enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_7enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 7489, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 7489, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7489, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Array_subscriptsContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_6enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":7490
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterArray_subscripts" ):             # <<<<<<<<<<<<<<
 *                 listener.enterArray_subscripts(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterArray_subscripts); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 7490, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7491
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterArray_subscripts" ):
 *                 listener.enterArray_subscripts(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterArray_subscripts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7491, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7491, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7491, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7491, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7491, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7491, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":7490
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterArray_subscripts" ):             # <<<<<<<<<<<<<<
 *                 listener.enterArray_subscripts(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7489
 *             return ModelicaParser.RULE_array_subscripts
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterArray_subscripts" ):
 *                 listener.enterArray_subscripts(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Array_subscriptsContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7493
 *                 listener.enterArray_subscripts(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitArray_subscripts" ):
 *                 listener.exitArray_subscripts(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_9exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_9exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 7493, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 7493, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7493, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Array_subscriptsContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_8exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":7494
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitArray_subscripts" ):             # <<<<<<<<<<<<<<
 *                 listener.exitArray_subscripts(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitArray_subscripts); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 7494, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7495
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitArray_subscripts" ):
 *                 listener.exitArray_subscripts(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitArray_subscripts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7495, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7495, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7495, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7495, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7495, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7495, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":7494
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitArray_subscripts" ):             # <<<<<<<<<<<<<<
 *                 listener.exitArray_subscripts(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7493
 *                 listener.enterArray_subscripts(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitArray_subscripts" ):
 *                 listener.exitArray_subscripts(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Array_subscriptsContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7497
 *                 listener.exitArray_subscripts(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitArray_subscripts" ):
 *                 return visitor.visitArray_subscripts(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_11accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_11accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 7497, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 7497, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7497, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Array_subscriptsContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_10accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":7498
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitArray_subscripts" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitArray_subscripts(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitArray_subscripts); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 7498, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7499
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitArray_subscripts" ):
 *                 return visitor.visitArray_subscripts(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitArray_subscripts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7499, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7499, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7499, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7499, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7499, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7499, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":7498
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitArray_subscripts" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitArray_subscripts(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7501
 *                 return visitor.visitArray_subscripts(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7501, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7501, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7501, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7501, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7501, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7501, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":7497
 *                 listener.exitArray_subscripts(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitArray_subscripts" ):
 *                 return visitor.visitArray_subscripts(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.Array_subscriptsContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7506
 * 
 * 
 *     def array_subscripts(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Array_subscriptsContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_139array_subscripts(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_139array_subscripts = {"array_subscripts", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_139array_subscripts, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_139array_subscripts(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("array_subscripts (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_138array_subscripts(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_138array_subscripts(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  char const *__pyx_t_21;
  __Pyx_RefNannySetupContext("array_subscripts", 0);

  /* "pymola/generated/ModelicaParser.pyx":7508
 *     def array_subscripts(self):
 * 
 *         localctx = ModelicaParser.Array_subscriptsContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 136, self.RULE_array_subscripts)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7508, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Array_subscriptsContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7508, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7508, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7508, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7508, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7508, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7508, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7508, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7509
 * 
 *         localctx = ModelicaParser.Array_subscriptsContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 136, self.RULE_array_subscripts)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_array_subscripts); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_136, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7509, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_136, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7509, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7509, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_136);
    __Pyx_GIVEREF(__pyx_int_136);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_136);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7509, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7510
 *         localctx = ModelicaParser.Array_subscriptsContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 136, self.RULE_array_subscripts)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 7510, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7511
 *         self.enterRule(localctx, 136, self.RULE_array_subscripts)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 1001
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":7512
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 1001
 *             self.match(ModelicaParser.T__75)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7512, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7512, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7512, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7512, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7512, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":7513
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 1001             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__75)
 *             self.state = 1002
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_1001) < 0) __PYX_ERR(0, 7513, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":7514
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 1001
 *             self.match(ModelicaParser.T__75)             # <<<<<<<<<<<<<<
 *             self.state = 1002
 *             self.subscript()
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7514, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7514, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__75); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7514, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7514, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7514, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7514, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7514, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7514, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":7515
 *             self.state = 1001
 *             self.match(ModelicaParser.T__75)
 *             self.state = 1002             # <<<<<<<<<<<<<<
 *             self.subscript()
 *             self.state = 1007
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_1002) < 0) __PYX_ERR(0, 7515, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":7516
 *             self.match(ModelicaParser.T__75)
 *             self.state = 1002
 *             self.subscript()             # <<<<<<<<<<<<<<
 *             self.state = 1007
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_subscript); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7516, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7516, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7516, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":7517
 *             self.state = 1002
 *             self.subscript()
 *             self.state = 1007             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_1007) < 0) __PYX_ERR(0, 7517, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":7518
 *             self.subscript()
 *             self.state = 1007
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             while _la==ModelicaParser.T__20:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7518, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7518, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7518, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7518, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7518, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7518, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7518, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":7519
 *             self.state = 1007
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             while _la==ModelicaParser.T__20:
 *                 self.state = 1003
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7519, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7519, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__968, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7519, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_v__la = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":7520
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             while _la==ModelicaParser.T__20:             # <<<<<<<<<<<<<<
 *                 self.state = 1003
 *                 self.match(ModelicaParser.T__20)
 */
        while (1) {
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7520, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__20); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7520, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7520, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 7520, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (!__pyx_t_11) break;

          /* "pymola/generated/ModelicaParser.pyx":7521
 *             _la = self._input.LA(1)
 *             while _la==ModelicaParser.T__20:
 *                 self.state = 1003             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__20)
 *                 self.state = 1004
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_1003) < 0) __PYX_ERR(0, 7521, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":7522
 *             while _la==ModelicaParser.T__20:
 *                 self.state = 1003
 *                 self.match(ModelicaParser.T__20)             # <<<<<<<<<<<<<<
 *                 self.state = 1004
 *                 self.subscript()
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7522, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7522, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__20); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7522, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7522, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7522, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7522, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7522, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7522, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7523
 *                 self.state = 1003
 *                 self.match(ModelicaParser.T__20)
 *                 self.state = 1004             # <<<<<<<<<<<<<<
 *                 self.subscript()
 *                 self.state = 1009
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_1004) < 0) __PYX_ERR(0, 7523, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":7524
 *                 self.match(ModelicaParser.T__20)
 *                 self.state = 1004
 *                 self.subscript()             # <<<<<<<<<<<<<<
 *                 self.state = 1009
 *                 self._errHandler.sync(self)
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_subscript); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7524, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7524, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7524, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7525
 *                 self.state = 1004
 *                 self.subscript()
 *                 self.state = 1009             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_1009) < 0) __PYX_ERR(0, 7525, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":7526
 *                 self.subscript()
 *                 self.state = 1009
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 * 
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7526, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7526, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_4) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7526, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7526, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7526, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7526, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __pyx_t_4 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7526, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7527
 *                 self.state = 1009
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *             self.state = 1010
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7527, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7527, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__969, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7527, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_1);
          __pyx_t_1 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":7529
 *                 _la = self._input.LA(1)
 * 
 *             self.state = 1010             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__76)
 *         except RecognitionException as re:
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_1010) < 0) __PYX_ERR(0, 7529, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":7530
 * 
 *             self.state = 1010
 *             self.match(ModelicaParser.T__76)             # <<<<<<<<<<<<<<
 *         except RecognitionException as re:
 *             localctx.exception = re
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7530, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7530, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__76); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7530, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7530, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_4};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7530, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_4};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7530, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7530, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_GIVEREF(__pyx_t_4);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_4);
            __pyx_t_4 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7530, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":7511
 *         self.enterRule(localctx, 136, self.RULE_array_subscripts)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 1001
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":7531
 *             self.state = 1010
 *             self.match(ModelicaParser.T__76)
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7531, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.array_subscripts", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_7, &__pyx_t_2) < 0) __PYX_ERR(0, 7531, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_7);
        __pyx_v_re = __pyx_t_7;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":7532
 *             self.match(ModelicaParser.T__76)
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 7532, __pyx_L21_error)

          /* "pymola/generated/ModelicaParser.pyx":7533
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7533, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7533, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_self, __pyx_v_re};
            __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7533, __pyx_L21_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_self, __pyx_v_re};
            __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7533, __pyx_L21_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          {
            __pyx_t_12 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 7533, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_12);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_12, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7533, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7534
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7534, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 7534, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_12);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_12, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7534, __pyx_L21_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7534, __pyx_L21_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7534, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7534, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":7531
 *             self.state = 1010
 *             self.match(ModelicaParser.T__76)
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L22;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L21_error:;
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __pyx_t_6 = __pyx_lineno; __pyx_t_13 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
            }
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_ErrRestore(__pyx_t_15, __pyx_t_16, __pyx_t_17);
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_13; __pyx_filename = __pyx_t_14;
            goto __pyx_L8_except_error;
          }
          __pyx_L22:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":7511
 *         self.enterRule(localctx, 136, self.RULE_array_subscripts)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 1001
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":7536
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7536, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7536, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7536, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_19, &__pyx_t_18);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_18);
      __pyx_t_13 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_21 = __pyx_filename;
      {
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7536, __pyx_L28_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7536, __pyx_L28_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7536, __pyx_L28_error)
        }
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __pyx_lineno = __pyx_t_13; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_21;
      goto __pyx_L1_error;
      __pyx_L28_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":7537
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class SubscriptContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":7506
 * 
 * 
 *     def array_subscripts(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Array_subscriptsContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.array_subscripts", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7541
 *     class SubscriptContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16SubscriptContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_16SubscriptContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_16SubscriptContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16SubscriptContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 7541, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 7541, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7541, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.SubscriptContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_16SubscriptContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16SubscriptContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":7542
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 7542, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7542, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7542, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7542, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7542, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7543
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def expression(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 7543, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7541
 *     class SubscriptContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.SubscriptContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7545
 *             self.parser = parser
 * 
 *         def expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16SubscriptContext_3expression(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_16SubscriptContext_3expression = {"expression", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_16SubscriptContext_3expression, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16SubscriptContext_3expression(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("expression (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_16SubscriptContext_2expression(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16SubscriptContext_2expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("expression", 0);

  /* "pymola/generated/ModelicaParser.pyx":7546
 * 
 *         def expression(self):
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_ExpressionContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7546, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7546, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7546, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7546, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":7545
 *             self.parser = parser
 * 
 *         def expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.SubscriptContext.expression", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7549
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_subscript
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16SubscriptContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_16SubscriptContext_5getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_16SubscriptContext_5getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16SubscriptContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_16SubscriptContext_4getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16SubscriptContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":7550
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_subscript             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_subscript); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":7549
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_subscript
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.SubscriptContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7552
 *             return ModelicaParser.RULE_subscript
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterSubscript" ):
 *                 listener.enterSubscript(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16SubscriptContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_16SubscriptContext_7enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_16SubscriptContext_7enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16SubscriptContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 7552, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 7552, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7552, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.SubscriptContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_16SubscriptContext_6enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16SubscriptContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":7553
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterSubscript" ):             # <<<<<<<<<<<<<<
 *                 listener.enterSubscript(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterSubscript); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 7553, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7554
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterSubscript" ):
 *                 listener.enterSubscript(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterSubscript); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7554, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7554, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7554, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7554, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7554, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7554, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":7553
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterSubscript" ):             # <<<<<<<<<<<<<<
 *                 listener.enterSubscript(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7552
 *             return ModelicaParser.RULE_subscript
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterSubscript" ):
 *                 listener.enterSubscript(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.SubscriptContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7556
 *                 listener.enterSubscript(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitSubscript" ):
 *                 listener.exitSubscript(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16SubscriptContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_16SubscriptContext_9exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_16SubscriptContext_9exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16SubscriptContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 7556, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 7556, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7556, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.SubscriptContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_16SubscriptContext_8exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16SubscriptContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":7557
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitSubscript" ):             # <<<<<<<<<<<<<<
 *                 listener.exitSubscript(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitSubscript); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 7557, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7558
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitSubscript" ):
 *                 listener.exitSubscript(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitSubscript); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7558, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7558, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7558, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7558, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7558, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7558, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":7557
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitSubscript" ):             # <<<<<<<<<<<<<<
 *                 listener.exitSubscript(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7556
 *                 listener.enterSubscript(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitSubscript" ):
 *                 listener.exitSubscript(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.SubscriptContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7560
 *                 listener.exitSubscript(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitSubscript" ):
 *                 return visitor.visitSubscript(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16SubscriptContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_16SubscriptContext_11accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_16SubscriptContext_11accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_16SubscriptContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 7560, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 7560, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7560, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.SubscriptContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_16SubscriptContext_10accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_16SubscriptContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":7561
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitSubscript" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitSubscript(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitSubscript); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 7561, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7562
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitSubscript" ):
 *                 return visitor.visitSubscript(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitSubscript); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7562, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7562, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7562, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7562, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7562, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7562, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":7561
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitSubscript" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitSubscript(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7564
 *                 return visitor.visitSubscript(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7564, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7564, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7564, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7564, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7564, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":7560
 *                 listener.exitSubscript(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitSubscript" ):
 *                 return visitor.visitSubscript(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.SubscriptContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7569
 * 
 * 
 *     def subscript(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.SubscriptContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_141subscript(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_141subscript = {"subscript", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_141subscript, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_141subscript(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("subscript (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_140subscript(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_140subscript(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v_token = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  char const *__pyx_t_22;
  __Pyx_RefNannySetupContext("subscript", 0);

  /* "pymola/generated/ModelicaParser.pyx":7571
 *     def subscript(self):
 * 
 *         localctx = ModelicaParser.SubscriptContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 138, self.RULE_subscript)
 *         try:
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7571, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_SubscriptContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7571, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7571, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7571, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7571, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7571, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7571, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7571, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7572
 * 
 *         localctx = ModelicaParser.SubscriptContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 138, self.RULE_subscript)             # <<<<<<<<<<<<<<
 *         try:
 *             self.state = 1014
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_subscript); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_138, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7572, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_138, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7572, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7572, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_138);
    __Pyx_GIVEREF(__pyx_int_138);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_138);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7572, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7573
 *         localctx = ModelicaParser.SubscriptContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 138, self.RULE_subscript)
 *         try:             # <<<<<<<<<<<<<<
 *             self.state = 1014
 *             self._errHandler.sync(self)
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":7574
 *         self.enterRule(localctx, 138, self.RULE_subscript)
 *         try:
 *             self.state = 1014             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_1014) < 0) __PYX_ERR(0, 7574, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":7575
 *         try:
 *             self.state = 1014
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.T__17]:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7575, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7575, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7575, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7575, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7575, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7575, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7575, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":7576
 *             self.state = 1014
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if token in [ModelicaParser.T__17]:
 *                 self.enterOuterAlt(localctx, 1)
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7576, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7576, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__970, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7576, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_v_token = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":7577
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.T__17]:             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 1)
 *                 self.state = 1012
 */
        __Pyx_INCREF(__pyx_v_token);
        __pyx_t_1 = __pyx_v_token;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7577, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__17); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7577, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7577, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 7577, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_12 = (__pyx_t_11 != 0);
        if (__pyx_t_12) {

          /* "pymola/generated/ModelicaParser.pyx":7578
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.T__17]:
 *                 self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *                 self.state = 1012
 *                 self.match(ModelicaParser.T__17)
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7578, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_localctx, __pyx_int_1};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7578, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_localctx, __pyx_int_1};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7578, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7578, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_7) {
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_1);
            __Pyx_GIVEREF(__pyx_int_1);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_int_1);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7578, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7579
 *             if token in [ModelicaParser.T__17]:
 *                 self.enterOuterAlt(localctx, 1)
 *                 self.state = 1012             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.T__17)
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_1012) < 0) __PYX_ERR(0, 7579, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":7580
 *                 self.enterOuterAlt(localctx, 1)
 *                 self.state = 1012
 *                 self.match(ModelicaParser.T__17)             # <<<<<<<<<<<<<<
 *                 pass
 *             elif token in [ModelicaParser.T__13, ModelicaParser.T__16, ModelicaParser.T__19, ModelicaParser.T__29, ModelicaParser.T__39, ModelicaParser.T__55, ModelicaParser.T__70, ModelicaParser.T__73, ModelicaParser.T__74, ModelicaParser.T__75, ModelicaParser.INITIAL, ModelicaParser.IDENT, ModelicaParser.STRING, ModelicaParser.UNSIGNED_NUMBER]:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7580, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7580, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__17); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7580, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_3) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7580, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7580, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_7};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7580, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7580, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __pyx_t_3 = NULL;
              __Pyx_GIVEREF(__pyx_t_7);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_7);
              __pyx_t_7 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7580, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7577
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)
 *             if token in [ModelicaParser.T__17]:             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 1)
 *                 self.state = 1012
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":7582
 *                 self.match(ModelicaParser.T__17)
 *                 pass
 *             elif token in [ModelicaParser.T__13, ModelicaParser.T__16, ModelicaParser.T__19, ModelicaParser.T__29, ModelicaParser.T__39, ModelicaParser.T__55, ModelicaParser.T__70, ModelicaParser.T__73, ModelicaParser.T__74, ModelicaParser.T__75, ModelicaParser.INITIAL, ModelicaParser.IDENT, ModelicaParser.STRING, ModelicaParser.UNSIGNED_NUMBER]:             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 2)
 *                 self.state = 1013
 */
        __Pyx_INCREF(__pyx_v_token);
        __pyx_t_1 = __pyx_v_token;
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__13); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__16); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__19); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__29); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__39); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__55); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__70); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__73); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__74); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_T__75); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_INITIAL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_IDENT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_STRING); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!__pyx_t_11) {
        } else {
          __pyx_t_12 = __pyx_t_11;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_UNSIGNED_NUMBER); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 7582, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __pyx_t_11;
        __pyx_L15_bool_binop_done:;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_11 = (__pyx_t_12 != 0);
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":7583
 *                 pass
 *             elif token in [ModelicaParser.T__13, ModelicaParser.T__16, ModelicaParser.T__19, ModelicaParser.T__29, ModelicaParser.T__39, ModelicaParser.T__55, ModelicaParser.T__70, ModelicaParser.T__73, ModelicaParser.T__74, ModelicaParser.T__75, ModelicaParser.INITIAL, ModelicaParser.IDENT, ModelicaParser.STRING, ModelicaParser.UNSIGNED_NUMBER]:
 *                 self.enterOuterAlt(localctx, 2)             # <<<<<<<<<<<<<<
 *                 self.state = 1013
 *                 self.expression()
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7583, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_4 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7583, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7583, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7583, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            if (__pyx_t_4) {
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4); __pyx_t_4 = NULL;
            }
            __Pyx_INCREF(__pyx_v_localctx);
            __Pyx_GIVEREF(__pyx_v_localctx);
            PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
            __Pyx_INCREF(__pyx_int_2);
            __Pyx_GIVEREF(__pyx_int_2);
            PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_2);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7583, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7584
 *             elif token in [ModelicaParser.T__13, ModelicaParser.T__16, ModelicaParser.T__19, ModelicaParser.T__29, ModelicaParser.T__39, ModelicaParser.T__55, ModelicaParser.T__70, ModelicaParser.T__73, ModelicaParser.T__74, ModelicaParser.T__75, ModelicaParser.INITIAL, ModelicaParser.IDENT, ModelicaParser.STRING, ModelicaParser.UNSIGNED_NUMBER]:
 *                 self.enterOuterAlt(localctx, 2)
 *                 self.state = 1013             # <<<<<<<<<<<<<<
 *                 self.expression()
 *                 pass
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_1013) < 0) __PYX_ERR(0, 7584, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":7585
 *                 self.enterOuterAlt(localctx, 2)
 *                 self.state = 1013
 *                 self.expression()             # <<<<<<<<<<<<<<
 *                 pass
 *             else:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expression); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7585, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7585, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7585, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7582
 *                 self.match(ModelicaParser.T__17)
 *                 pass
 *             elif token in [ModelicaParser.T__13, ModelicaParser.T__16, ModelicaParser.T__19, ModelicaParser.T__29, ModelicaParser.T__39, ModelicaParser.T__55, ModelicaParser.T__70, ModelicaParser.T__73, ModelicaParser.T__74, ModelicaParser.T__75, ModelicaParser.INITIAL, ModelicaParser.IDENT, ModelicaParser.STRING, ModelicaParser.UNSIGNED_NUMBER]:             # <<<<<<<<<<<<<<
 *                 self.enterOuterAlt(localctx, 2)
 *                 self.state = 1013
 */
          goto __pyx_L14;
        }

        /* "pymola/generated/ModelicaParser.pyx":7588
 *                 pass
 *             else:
 *                 raise NoViableAltException(self)             # <<<<<<<<<<<<<<
 * 
 *         except RecognitionException as re:
 */
        /*else*/ {
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_NoViableAltException); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7588, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7588, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7588, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7588, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7588, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7588, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_Raise(__pyx_t_1, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __PYX_ERR(0, 7588, __pyx_L6_error)
        }
        __pyx_L14:;

        /* "pymola/generated/ModelicaParser.pyx":7573
 *         localctx = ModelicaParser.SubscriptContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 138, self.RULE_subscript)
 *         try:             # <<<<<<<<<<<<<<
 *             self.state = 1014
 *             self._errHandler.sync(self)
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":7590
 *                 raise NoViableAltException(self)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7590, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.subscript", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_2, &__pyx_t_4) < 0) __PYX_ERR(0, 7590, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_2);
        __pyx_v_re = __pyx_t_2;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":7591
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 7591, __pyx_L34_error)

          /* "pymola/generated/ModelicaParser.pyx":7592
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7592, __pyx_L34_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7592, __pyx_L34_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7592, __pyx_L34_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7592, __pyx_L34_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 7592, __pyx_L34_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_13, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7592, __pyx_L34_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7593
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7593, __pyx_L34_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 7593, __pyx_L34_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7593, __pyx_L34_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7593, __pyx_L34_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7593, __pyx_L34_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_3, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7593, __pyx_L34_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":7590
 *                 raise NoViableAltException(self)
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L35;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L34_error:;
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_6 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L8_except_error;
          }
          __pyx_L35:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":7573
 *         localctx = ModelicaParser.SubscriptContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 138, self.RULE_subscript)
 *         try:             # <<<<<<<<<<<<<<
 *             self.state = 1014
 *             self._errHandler.sync(self)
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":7595
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7595, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7595, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7595, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_14 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7595, __pyx_L41_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7595, __pyx_L41_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7595, __pyx_L41_error)
        }
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
      __pyx_L41_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":7596
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class CommentContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":7569
 * 
 * 
 *     def subscript(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.SubscriptContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.subscript", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v_token);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7600
 *     class CommentContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14CommentContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14CommentContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14CommentContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14CommentContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 7600, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 7600, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7600, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.CommentContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14CommentContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14CommentContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":7601
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 7601, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7601, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7601, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7601, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7601, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7601, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7601, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7601, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7602
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def string_comment(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 7602, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7600
 *     class CommentContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.CommentContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7604
 *             self.parser = parser
 * 
 *         def string_comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.String_commentContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14CommentContext_3string_comment(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14CommentContext_3string_comment = {"string_comment", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14CommentContext_3string_comment, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14CommentContext_3string_comment(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("string_comment (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14CommentContext_2string_comment(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14CommentContext_2string_comment(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("string_comment", 0);

  /* "pymola/generated/ModelicaParser.pyx":7605
 * 
 *         def string_comment(self):
 *             return self.getTypedRuleContext(ModelicaParser.String_commentContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7605, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7605, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_String_commentContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7605, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7605, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7605, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7605, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7605, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":7604
 *             self.parser = parser
 * 
 *         def string_comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.String_commentContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.CommentContext.string_comment", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7608
 * 
 * 
 *         def annotation(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.AnnotationContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14CommentContext_5annotation(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14CommentContext_5annotation = {"annotation", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14CommentContext_5annotation, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14CommentContext_5annotation(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("annotation (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14CommentContext_4annotation(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14CommentContext_4annotation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("annotation", 0);

  /* "pymola/generated/ModelicaParser.pyx":7609
 * 
 *         def annotation(self):
 *             return self.getTypedRuleContext(ModelicaParser.AnnotationContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7609, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7609, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_AnnotationContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7609, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7609, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7609, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7609, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7609, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":7608
 * 
 * 
 *         def annotation(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.AnnotationContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.CommentContext.annotation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7612
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_comment
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14CommentContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14CommentContext_7getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14CommentContext_7getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14CommentContext_7getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14CommentContext_6getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14CommentContext_6getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":7613
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_comment             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_comment); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":7612
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_comment
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.CommentContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7615
 *             return ModelicaParser.RULE_comment
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterComment" ):
 *                 listener.enterComment(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14CommentContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14CommentContext_9enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14CommentContext_9enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14CommentContext_9enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 7615, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 7615, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7615, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.CommentContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14CommentContext_8enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14CommentContext_8enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":7616
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterComment" ):             # <<<<<<<<<<<<<<
 *                 listener.enterComment(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterComment); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 7616, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7617
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterComment" ):
 *                 listener.enterComment(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterComment); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7617, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7617, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7617, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7617, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7617, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7617, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":7616
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterComment" ):             # <<<<<<<<<<<<<<
 *                 listener.enterComment(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7615
 *             return ModelicaParser.RULE_comment
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterComment" ):
 *                 listener.enterComment(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.CommentContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7619
 *                 listener.enterComment(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitComment" ):
 *                 listener.exitComment(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14CommentContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14CommentContext_11exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14CommentContext_11exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14CommentContext_11exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 7619, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 7619, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7619, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.CommentContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14CommentContext_10exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14CommentContext_10exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":7620
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitComment" ):             # <<<<<<<<<<<<<<
 *                 listener.exitComment(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitComment); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 7620, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7621
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitComment" ):
 *                 listener.exitComment(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitComment); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7621, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7621, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7621, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7621, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7621, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7621, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":7620
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitComment" ):             # <<<<<<<<<<<<<<
 *                 listener.exitComment(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7619
 *                 listener.enterComment(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitComment" ):
 *                 listener.exitComment(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.CommentContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7623
 *                 listener.exitComment(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitComment" ):
 *                 return visitor.visitComment(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14CommentContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14CommentContext_13accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14CommentContext_13accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14CommentContext_13accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 7623, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 7623, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7623, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.CommentContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14CommentContext_12accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14CommentContext_12accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":7624
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitComment" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitComment(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitComment); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 7624, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7625
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitComment" ):
 *                 return visitor.visitComment(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitComment); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7625, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7625, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7625, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7625, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7625, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7625, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":7624
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitComment" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitComment(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7627
 *                 return visitor.visitComment(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7627, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7627, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7627, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7627, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7627, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7627, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":7623
 *                 listener.exitComment(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitComment" ):
 *                 return visitor.visitComment(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.CommentContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7632
 * 
 * 
 *     def comment(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.CommentContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_143comment(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_143comment = {"comment", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_143comment, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_143comment(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("comment (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_142comment(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_142comment(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  char const *__pyx_t_21;
  __Pyx_RefNannySetupContext("comment", 0);

  /* "pymola/generated/ModelicaParser.pyx":7634
 *     def comment(self):
 * 
 *         localctx = ModelicaParser.CommentContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 140, self.RULE_comment)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_CommentContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7634, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7634, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7634, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7634, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7635
 * 
 *         localctx = ModelicaParser.CommentContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 140, self.RULE_comment)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_comment); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_140, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7635, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_140, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7635, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7635, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_140);
    __Pyx_GIVEREF(__pyx_int_140);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_140);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7635, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7636
 *         localctx = ModelicaParser.CommentContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 140, self.RULE_comment)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 7636, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7637
 *         self.enterRule(localctx, 140, self.RULE_comment)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 1016
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":7638
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 1016
 *             self.string_comment()
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7638, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7638, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7638, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7638, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7638, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":7639
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 1016             # <<<<<<<<<<<<<<
 *             self.string_comment()
 *             self.state = 1018
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_1016) < 0) __PYX_ERR(0, 7639, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":7640
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 1016
 *             self.string_comment()             # <<<<<<<<<<<<<<
 *             self.state = 1018
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_string_comment); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7640, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7640, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7640, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":7641
 *             self.state = 1016
 *             self.string_comment()
 *             self.state = 1018             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_1018) < 0) __PYX_ERR(0, 7641, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":7642
 *             self.string_comment()
 *             self.state = 1018
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__77:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7642, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7642, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7642, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7642, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7642, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7642, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7642, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":7643
 *             self.state = 1018
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.T__77:
 *                 self.state = 1017
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7643, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7643, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__971, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7643, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v__la = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":7644
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__77:             # <<<<<<<<<<<<<<
 *                 self.state = 1017
 *                 self.annotation()
 */
        __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7644, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__77); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7644, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7644, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 7644, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":7645
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__77:
 *                 self.state = 1017             # <<<<<<<<<<<<<<
 *                 self.annotation()
 * 
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_1017) < 0) __PYX_ERR(0, 7645, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":7646
 *             if _la==ModelicaParser.T__77:
 *                 self.state = 1017
 *                 self.annotation()             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_annotation); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7646, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7646, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else {
            __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7646, __pyx_L6_error)
          }
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7644
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.T__77:             # <<<<<<<<<<<<<<
 *                 self.state = 1017
 *                 self.annotation()
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":7637
 *         self.enterRule(localctx, 140, self.RULE_comment)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 1016
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":7649
 * 
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7649, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.comment", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_7, &__pyx_t_2) < 0) __PYX_ERR(0, 7649, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_7);
        __pyx_v_re = __pyx_t_7;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":7650
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 7650, __pyx_L20_error)

          /* "pymola/generated/ModelicaParser.pyx":7651
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7651, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7651, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7651, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7651, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_12 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 7651, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_12);
            if (__pyx_t_4) {
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_4); __pyx_t_4 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_12, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7651, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7652
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7652, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 7652, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_12);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_12, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7652, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7652, __pyx_L20_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7652, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7652, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":7649
 * 
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L21;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L20_error:;
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __pyx_t_6 = __pyx_lineno; __pyx_t_13 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
            }
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_ErrRestore(__pyx_t_15, __pyx_t_16, __pyx_t_17);
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_13; __pyx_filename = __pyx_t_14;
            goto __pyx_L8_except_error;
          }
          __pyx_L21:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":7637
 *         self.enterRule(localctx, 140, self.RULE_comment)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 1016
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":7654
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7654, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7654, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7654, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_19, &__pyx_t_18);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_18);
      __pyx_t_13 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_21 = __pyx_filename;
      {
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7654, __pyx_L27_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7654, __pyx_L27_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7654, __pyx_L27_error)
        }
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __pyx_lineno = __pyx_t_13; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_21;
      goto __pyx_L1_error;
      __pyx_L27_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":7655
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class String_commentContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":7632
 * 
 * 
 *     def comment(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.CommentContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.comment", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7659
 *     class String_commentContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21String_commentContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21String_commentContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21String_commentContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21String_commentContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 7659, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 7659, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7659, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.String_commentContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21String_commentContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21String_commentContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":7660
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 7660, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7660, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7660, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7660, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7660, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7660, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7660, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7660, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7661
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def STRING(self, i:int=None):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 7661, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7659
 *     class String_commentContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.String_commentContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7663
 *             self.parser = parser
 * 
 *         def STRING(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTokens(ModelicaParser.STRING)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21String_commentContext_3STRING(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21String_commentContext_3STRING = {"STRING", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21String_commentContext_3STRING, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21String_commentContext_3STRING(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("STRING (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_i);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "STRING") < 0)) __PYX_ERR(0, 7663, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("STRING", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7663, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.String_commentContext.STRING", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21String_commentContext_2STRING(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21String_commentContext_2STRING(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("STRING", 0);

  /* "pymola/generated/ModelicaParser.pyx":7664
 * 
 *         def STRING(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTokens(ModelicaParser.STRING)
 *             else:
 */
  __pyx_t_1 = (__pyx_v_i == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7665
 *         def STRING(self, i:int=None):
 *             if i is None:
 *                 return self.getTokens(ModelicaParser.STRING)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.getToken(ModelicaParser.STRING, i)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTokens); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7665, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7665, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_STRING); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7665, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7665, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7665, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7665, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7665, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7665, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":7664
 * 
 *         def STRING(self, i:int=None):
 *             if i is None:             # <<<<<<<<<<<<<<
 *                 return self.getTokens(ModelicaParser.STRING)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7667
 *                 return self.getTokens(ModelicaParser.STRING)
 *             else:
 *                 return self.getToken(ModelicaParser.STRING, i)             # <<<<<<<<<<<<<<
 * 
 *         def getRuleIndex(self):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getToken); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7667, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7667, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_STRING); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7667, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7667, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_i};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7667, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7667, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_v_i);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7667, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":7663
 *             self.parser = parser
 * 
 *         def STRING(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTokens(ModelicaParser.STRING)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.String_commentContext.STRING", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7669
 *                 return self.getToken(ModelicaParser.STRING, i)
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_string_comment
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21String_commentContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21String_commentContext_5getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21String_commentContext_5getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21String_commentContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21String_commentContext_4getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21String_commentContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":7670
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_string_comment             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7670, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_string_comment); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7670, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":7669
 *                 return self.getToken(ModelicaParser.STRING, i)
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_string_comment
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.String_commentContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7672
 *             return ModelicaParser.RULE_string_comment
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterString_comment" ):
 *                 listener.enterString_comment(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21String_commentContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21String_commentContext_7enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21String_commentContext_7enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21String_commentContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 7672, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 7672, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7672, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.String_commentContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21String_commentContext_6enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21String_commentContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":7673
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterString_comment" ):             # <<<<<<<<<<<<<<
 *                 listener.enterString_comment(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterString_comment); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 7673, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7674
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterString_comment" ):
 *                 listener.enterString_comment(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterString_comment); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7674, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7674, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7674, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7674, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7674, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7674, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":7673
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterString_comment" ):             # <<<<<<<<<<<<<<
 *                 listener.enterString_comment(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7672
 *             return ModelicaParser.RULE_string_comment
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterString_comment" ):
 *                 listener.enterString_comment(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.String_commentContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7676
 *                 listener.enterString_comment(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitString_comment" ):
 *                 listener.exitString_comment(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21String_commentContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21String_commentContext_9exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21String_commentContext_9exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21String_commentContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 7676, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 7676, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7676, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.String_commentContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21String_commentContext_8exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21String_commentContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":7677
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitString_comment" ):             # <<<<<<<<<<<<<<
 *                 listener.exitString_comment(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitString_comment); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 7677, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7678
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitString_comment" ):
 *                 listener.exitString_comment(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitString_comment); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7678, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7678, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7678, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7678, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7678, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7678, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":7677
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitString_comment" ):             # <<<<<<<<<<<<<<
 *                 listener.exitString_comment(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7676
 *                 listener.enterString_comment(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitString_comment" ):
 *                 listener.exitString_comment(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.String_commentContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7680
 *                 listener.exitString_comment(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitString_comment" ):
 *                 return visitor.visitString_comment(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21String_commentContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_21String_commentContext_11accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_21String_commentContext_11accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_21String_commentContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 7680, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 7680, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7680, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.String_commentContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_21String_commentContext_10accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_21String_commentContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":7681
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitString_comment" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitString_comment(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitString_comment); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 7681, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7682
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitString_comment" ):
 *                 return visitor.visitString_comment(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitString_comment); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7682, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7682, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7682, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7682, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7682, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7682, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":7681
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitString_comment" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitString_comment(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7684
 *                 return visitor.visitString_comment(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7684, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7684, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7684, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7684, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7684, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7684, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":7680
 *                 listener.exitString_comment(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitString_comment" ):
 *                 return visitor.visitString_comment(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.String_commentContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7689
 * 
 * 
 *     def string_comment(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.String_commentContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_145string_comment(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_145string_comment = {"string_comment", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_145string_comment, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_145string_comment(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("string_comment (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_144string_comment(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_144string_comment(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v__la = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  char const *__pyx_t_21;
  __Pyx_RefNannySetupContext("string_comment", 0);

  /* "pymola/generated/ModelicaParser.pyx":7691
 *     def string_comment(self):
 * 
 *         localctx = ModelicaParser.String_commentContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 142, self.RULE_string_comment)
 *         self._la = 0 # Token type
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7691, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_String_commentContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7691, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7691, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7691, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7691, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7691, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7691, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7691, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7692
 * 
 *         localctx = ModelicaParser.String_commentContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 142, self.RULE_string_comment)             # <<<<<<<<<<<<<<
 *         self._la = 0 # Token type
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_string_comment); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_142, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7692, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_142, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7692, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7692, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_142);
    __Pyx_GIVEREF(__pyx_int_142);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_142);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7692, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7693
 *         localctx = ModelicaParser.String_commentContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 142, self.RULE_string_comment)
 *         self._la = 0 # Token type             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_la, __pyx_int_0) < 0) __PYX_ERR(0, 7693, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7694
 *         self.enterRule(localctx, 142, self.RULE_string_comment)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 1028
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":7695
 *         self._la = 0 # Token type
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 1028
 *             self._errHandler.sync(self)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7695, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7695, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7695, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7695, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7695, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":7696
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 1028             # <<<<<<<<<<<<<<
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_1028) < 0) __PYX_ERR(0, 7696, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":7697
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 1028
 *             self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.STRING:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7697, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sync); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7697, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7697, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7697, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7697, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7697, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_self);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7697, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":7698
 *             self.state = 1028
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.STRING:
 *                 self.state = 1020
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7698, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7698, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__972, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7698, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v__la = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":7699
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.STRING:             # <<<<<<<<<<<<<<
 *                 self.state = 1020
 *                 self.match(ModelicaParser.STRING)
 */
        __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7699, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_STRING); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7699, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7699, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 7699, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_11) {

          /* "pymola/generated/ModelicaParser.pyx":7700
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.STRING:
 *                 self.state = 1020             # <<<<<<<<<<<<<<
 *                 self.match(ModelicaParser.STRING)
 *                 self.state = 1025
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_1020) < 0) __PYX_ERR(0, 7700, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":7701
 *             if _la==ModelicaParser.STRING:
 *                 self.state = 1020
 *                 self.match(ModelicaParser.STRING)             # <<<<<<<<<<<<<<
 *                 self.state = 1025
 *                 self._errHandler.sync(self)
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7701, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7701, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_STRING); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7701, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7701, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7701, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
              PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_3};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7701, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7701, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7701, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7702
 *                 self.state = 1020
 *                 self.match(ModelicaParser.STRING)
 *                 self.state = 1025             # <<<<<<<<<<<<<<
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_1025) < 0) __PYX_ERR(0, 7702, __pyx_L6_error)

          /* "pymola/generated/ModelicaParser.pyx":7703
 *                 self.match(ModelicaParser.STRING)
 *                 self.state = 1025
 *                 self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 *                 while _la==ModelicaParser.T__61:
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7703, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_sync); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7703, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7703, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7703, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_self};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7703, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7703, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_self);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7703, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7704
 *                 self.state = 1025
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 while _la==ModelicaParser.T__61:
 *                     self.state = 1021
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7704, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7704, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__973, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7704, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7705
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)
 *                 while _la==ModelicaParser.T__61:             # <<<<<<<<<<<<<<
 *                     self.state = 1021
 *                     self.match(ModelicaParser.T__61)
 */
          while (1) {
            __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7705, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T__61); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7705, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = PyObject_RichCompare(__pyx_v__la, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7705, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 7705, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (!__pyx_t_11) break;

            /* "pymola/generated/ModelicaParser.pyx":7706
 *                 _la = self._input.LA(1)
 *                 while _la==ModelicaParser.T__61:
 *                     self.state = 1021             # <<<<<<<<<<<<<<
 *                     self.match(ModelicaParser.T__61)
 *                     self.state = 1022
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_1021) < 0) __PYX_ERR(0, 7706, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":7707
 *                 while _la==ModelicaParser.T__61:
 *                     self.state = 1021
 *                     self.match(ModelicaParser.T__61)             # <<<<<<<<<<<<<<
 *                     self.state = 1022
 *                     self.match(ModelicaParser.STRING)
 */
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7707, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7707, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_T__61); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7707, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
              __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
              if (likely(__pyx_t_3)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                __Pyx_INCREF(__pyx_t_3);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_4, function);
              }
            }
            if (!__pyx_t_3) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7707, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_4)) {
                PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_7};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7707, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
                PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_7};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7707, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              } else
              #endif
              {
                __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7707, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
                __Pyx_GIVEREF(__pyx_t_7);
                PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_7);
                __pyx_t_7 = 0;
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7707, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":7708
 *                     self.state = 1021
 *                     self.match(ModelicaParser.T__61)
 *                     self.state = 1022             # <<<<<<<<<<<<<<
 *                     self.match(ModelicaParser.STRING)
 *                     self.state = 1027
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_1022) < 0) __PYX_ERR(0, 7708, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":7709
 *                     self.match(ModelicaParser.T__61)
 *                     self.state = 1022
 *                     self.match(ModelicaParser.STRING)             # <<<<<<<<<<<<<<
 *                     self.state = 1027
 *                     self._errHandler.sync(self)
 */
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7709, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7709, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_STRING); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7709, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_2 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
              __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
              if (likely(__pyx_t_2)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                __Pyx_INCREF(__pyx_t_2);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_4, function);
              }
            }
            if (!__pyx_t_2) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7709, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_4)) {
                PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_7};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7709, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
                PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_7};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7709, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              } else
              #endif
              {
                __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7709, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
                __Pyx_GIVEREF(__pyx_t_7);
                PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_t_7);
                __pyx_t_7 = 0;
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7709, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":7710
 *                     self.state = 1022
 *                     self.match(ModelicaParser.STRING)
 *                     self.state = 1027             # <<<<<<<<<<<<<<
 *                     self._errHandler.sync(self)
 *                     _la = self._input.LA(1)
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_1027) < 0) __PYX_ERR(0, 7710, __pyx_L6_error)

            /* "pymola/generated/ModelicaParser.pyx":7711
 *                     self.match(ModelicaParser.STRING)
 *                     self.state = 1027
 *                     self._errHandler.sync(self)             # <<<<<<<<<<<<<<
 *                     _la = self._input.LA(1)
 * 
 */
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7711, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_sync); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7711, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
              __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
              if (likely(__pyx_t_4)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                __Pyx_INCREF(__pyx_t_4);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_3, function);
              }
            }
            if (!__pyx_t_4) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7711, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_3)) {
                PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7711, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
                PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_self};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7711, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_GOTREF(__pyx_t_1);
              } else
              #endif
              {
                __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7711, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_7);
                __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4); __pyx_t_4 = NULL;
                __Pyx_INCREF(__pyx_v_self);
                __Pyx_GIVEREF(__pyx_v_self);
                PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self);
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7711, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              }
            }
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "pymola/generated/ModelicaParser.pyx":7712
 *                     self.state = 1027
 *                     self._errHandler.sync(self)
 *                     _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 * 
 */
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_input_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7712, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LA); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7712, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__974, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7712, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF_SET(__pyx_v__la, __pyx_t_1);
            __pyx_t_1 = 0;
          }

          /* "pymola/generated/ModelicaParser.pyx":7699
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)
 *             if _la==ModelicaParser.STRING:             # <<<<<<<<<<<<<<
 *                 self.state = 1020
 *                 self.match(ModelicaParser.STRING)
 */
        }

        /* "pymola/generated/ModelicaParser.pyx":7694
 *         self.enterRule(localctx, 142, self.RULE_string_comment)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 1028
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":7716
 * 
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7716, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.string_comment", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_3, &__pyx_t_7) < 0) __PYX_ERR(0, 7716, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_3);
        __pyx_v_re = __pyx_t_3;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":7717
 * 
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 7717, __pyx_L22_error)

          /* "pymola/generated/ModelicaParser.pyx":7718
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7718, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7718, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_re};
            __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7718, __pyx_L22_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self, __pyx_v_re};
            __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7718, __pyx_L22_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          {
            __pyx_t_12 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 7718, __pyx_L22_error)
            __Pyx_GOTREF(__pyx_t_12);
            if (__pyx_t_2) {
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_2); __pyx_t_2 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_12, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7718, __pyx_L22_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7719
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7719, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 7719, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_12);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_12, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7719, __pyx_L22_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7719, __pyx_L22_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7719, __pyx_L22_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_2, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7719, __pyx_L22_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":7716
 * 
 * 
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L23;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L22_error:;
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __pyx_t_6 = __pyx_lineno; __pyx_t_13 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
            }
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_ErrRestore(__pyx_t_15, __pyx_t_16, __pyx_t_17);
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_13; __pyx_filename = __pyx_t_14;
            goto __pyx_L8_except_error;
          }
          __pyx_L23:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":7694
 *         self.enterRule(localctx, 142, self.RULE_string_comment)
 *         self._la = 0 # Token type
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 1028
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":7721
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7721, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7721, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7721, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_19, &__pyx_t_18);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_18);
      __pyx_t_13 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_21 = __pyx_filename;
      {
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7721, __pyx_L29_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7721, __pyx_L29_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7721, __pyx_L29_error)
        }
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __pyx_lineno = __pyx_t_13; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_21;
      goto __pyx_L1_error;
      __pyx_L29_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":7722
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 *     class AnnotationContext(ParserRuleContext):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":7689
 * 
 * 
 *     def string_comment(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.String_commentContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.string_comment", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v__la);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7726
 *     class AnnotationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_17AnnotationContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_17AnnotationContext_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_17AnnotationContext_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_17AnnotationContext_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parser = 0;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_invokingState = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parser,&__pyx_n_s_parent,&__pyx_n_s_invokingState,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_neg_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parser)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 7726, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_parent);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_invokingState);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 7726, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_parser = values[1];
    __pyx_v_parent = values[2];
    __pyx_v_invokingState = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7726, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.AnnotationContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_17AnnotationContext___init__(__pyx_self, __pyx_v_self, __pyx_v_parser, __pyx_v_parent, __pyx_v_invokingState);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_17AnnotationContext___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parser, PyObject *__pyx_v_parent, PyObject *__pyx_v_invokingState) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pymola/generated/ModelicaParser.pyx":7727
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)             # <<<<<<<<<<<<<<
 *             self.parser = parser
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 7727, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7727, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7727, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7727, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7727, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_parent, __pyx_v_invokingState};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7727, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7727, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_parent);
    __Pyx_INCREF(__pyx_v_invokingState);
    __Pyx_GIVEREF(__pyx_v_invokingState);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_invokingState);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7727, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7728
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 *             super().__init__(parent, invokingState)
 *             self.parser = parser             # <<<<<<<<<<<<<<
 * 
 *         def class_modification(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parser, __pyx_v_parser) < 0) __PYX_ERR(0, 7728, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7726
 *     class AnnotationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.AnnotationContext.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7730
 *             self.parser = parser
 * 
 *         def class_modification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_modificationContext,0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_17AnnotationContext_3class_modification(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_17AnnotationContext_3class_modification = {"class_modification", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_17AnnotationContext_3class_modification, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_17AnnotationContext_3class_modification(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("class_modification (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_17AnnotationContext_2class_modification(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_17AnnotationContext_2class_modification(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("class_modification", 0);

  /* "pymola/generated/ModelicaParser.pyx":7731
 * 
 *         def class_modification(self):
 *             return self.getTypedRuleContext(ModelicaParser.Class_modificationContext,0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTypedRuleContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7731, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7731, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Class_modificationContext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7731, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7731, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7731, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7731, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7731, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":7730
 *             self.parser = parser
 * 
 *         def class_modification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_modificationContext,0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.AnnotationContext.class_modification", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7734
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_annotation
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_17AnnotationContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_17AnnotationContext_5getRuleIndex = {"getRuleIndex", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_17AnnotationContext_5getRuleIndex, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_17AnnotationContext_5getRuleIndex(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRuleIndex (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_17AnnotationContext_4getRuleIndex(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_17AnnotationContext_4getRuleIndex(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getRuleIndex", 0);

  /* "pymola/generated/ModelicaParser.pyx":7735
 * 
 *         def getRuleIndex(self):
 *             return ModelicaParser.RULE_annotation             # <<<<<<<<<<<<<<
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7735, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_RULE_annotation); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7735, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":7734
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_annotation
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.AnnotationContext.getRuleIndex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7737
 *             return ModelicaParser.RULE_annotation
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterAnnotation" ):
 *                 listener.enterAnnotation(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_17AnnotationContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_17AnnotationContext_7enterRule = {"enterRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_17AnnotationContext_7enterRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_17AnnotationContext_7enterRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, 1); __PYX_ERR(0, 7737, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRule") < 0)) __PYX_ERR(0, 7737, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7737, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.AnnotationContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_17AnnotationContext_6enterRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_17AnnotationContext_6enterRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("enterRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":7738
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterAnnotation" ):             # <<<<<<<<<<<<<<
 *                 listener.enterAnnotation(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_enterAnnotation); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 7738, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7739
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterAnnotation" ):
 *                 listener.enterAnnotation(self)             # <<<<<<<<<<<<<<
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_enterAnnotation); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7739, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7739, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7739, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7739, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7739, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7739, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":7738
 * 
 *         def enterRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "enterAnnotation" ):             # <<<<<<<<<<<<<<
 *                 listener.enterAnnotation(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7737
 *             return ModelicaParser.RULE_annotation
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterAnnotation" ):
 *                 listener.enterAnnotation(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.AnnotationContext.enterRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7741
 *                 listener.enterAnnotation(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitAnnotation" ):
 *                 listener.exitAnnotation(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_17AnnotationContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_17AnnotationContext_9exitRule = {"exitRule", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_17AnnotationContext_9exitRule, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_17AnnotationContext_9exitRule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_listener = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_listener)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, 1); __PYX_ERR(0, 7741, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRule") < 0)) __PYX_ERR(0, 7741, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRule", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7741, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.AnnotationContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_17AnnotationContext_8exitRule(__pyx_self, __pyx_v_self, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_17AnnotationContext_8exitRule(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_listener) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("exitRule", 0);

  /* "pymola/generated/ModelicaParser.pyx":7742
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitAnnotation" ):             # <<<<<<<<<<<<<<
 *                 listener.exitAnnotation(self)
 * 
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_listener, __pyx_n_u_exitAnnotation); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 7742, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7743
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitAnnotation" ):
 *                 listener.exitAnnotation(self)             # <<<<<<<<<<<<<<
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_exitAnnotation); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7743, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7743, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7743, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7743, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7743, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7743, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pymola/generated/ModelicaParser.pyx":7742
 * 
 *         def exitRule(self, listener:ParseTreeListener):
 *             if hasattr( listener, "exitAnnotation" ):             # <<<<<<<<<<<<<<
 *                 listener.exitAnnotation(self)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7741
 *                 listener.enterAnnotation(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitAnnotation" ):
 *                 listener.exitAnnotation(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.AnnotationContext.exitRule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7745
 *                 listener.exitAnnotation(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitAnnotation" ):
 *                 return visitor.visitAnnotation(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_17AnnotationContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_17AnnotationContext_11accept = {"accept", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_17AnnotationContext_11accept, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_17AnnotationContext_11accept(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_visitor = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_visitor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_visitor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, 1); __PYX_ERR(0, 7745, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "accept") < 0)) __PYX_ERR(0, 7745, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_visitor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7745, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.AnnotationContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_17AnnotationContext_10accept(__pyx_self, __pyx_v_self, __pyx_v_visitor);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_17AnnotationContext_10accept(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_visitor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pymola/generated/ModelicaParser.pyx":7746
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitAnnotation" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitAnnotation(self)
 *             else:
 */
  __pyx_t_1 = PyObject_HasAttr(__pyx_v_visitor, __pyx_n_u_visitAnnotation); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 7746, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7747
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitAnnotation" ):
 *                 return visitor.visitAnnotation(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return visitor.visitChildren(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitAnnotation); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7747, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7747, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7747, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7747, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7747, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7747, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":7746
 * 
 *         def accept(self, visitor:ParseTreeVisitor):
 *             if hasattr( visitor, "visitAnnotation" ):             # <<<<<<<<<<<<<<
 *                 return visitor.visitAnnotation(self)
 *             else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7749
 *                 return visitor.visitAnnotation(self)
 *             else:
 *                 return visitor.visitChildren(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_visitor, __pyx_n_s_visitChildren); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7749, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7749, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7749, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7749, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7749, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_self);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7749, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":7745
 *                 listener.exitAnnotation(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitAnnotation" ):
 *                 return visitor.visitAnnotation(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.AnnotationContext.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7754
 * 
 * 
 *     def annotation(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.AnnotationContext(self, self._ctx, self.state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_147annotation(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_147annotation = {"annotation", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_147annotation, METH_O, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_147annotation(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("annotation (wrapper)", 0);
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_146annotation(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_146annotation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_localctx = NULL;
  PyObject *__pyx_v_re = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  char const *__pyx_t_20;
  __Pyx_RefNannySetupContext("annotation", 0);

  /* "pymola/generated/ModelicaParser.pyx":7756
 *     def annotation(self):
 * 
 *         localctx = ModelicaParser.AnnotationContext(self, self._ctx, self.state)             # <<<<<<<<<<<<<<
 *         self.enterRule(localctx, 144, self.RULE_annotation)
 *         try:
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7756, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_AnnotationContext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7756, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7756, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7756, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7756, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_self, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7756, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7756, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7756, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_localctx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7757
 * 
 *         localctx = ModelicaParser.AnnotationContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 144, self.RULE_annotation)             # <<<<<<<<<<<<<<
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7757, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_RULE_annotation); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7757, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_144, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7757, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_localctx, __pyx_int_144, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7757, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7757, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_localctx);
    __Pyx_GIVEREF(__pyx_v_localctx);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_localctx);
    __Pyx_INCREF(__pyx_int_144);
    __Pyx_GIVEREF(__pyx_int_144);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_int_144);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7757, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7758
 *         localctx = ModelicaParser.AnnotationContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 144, self.RULE_annotation)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 1030
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "pymola/generated/ModelicaParser.pyx":7759
 *         self.enterRule(localctx, 144, self.RULE_annotation)
 *         try:
 *             self.enterOuterAlt(localctx, 1)             # <<<<<<<<<<<<<<
 *             self.state = 1030
 *             self.match(ModelicaParser.T__77)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enterOuterAlt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7759, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7759, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_localctx, __pyx_int_1};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7759, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7759, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_localctx);
          __Pyx_GIVEREF(__pyx_v_localctx);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_localctx);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_1);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7759, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":7760
 *         try:
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 1030             # <<<<<<<<<<<<<<
 *             self.match(ModelicaParser.T__77)
 *             self.state = 1031
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_1030) < 0) __PYX_ERR(0, 7760, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":7761
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 1030
 *             self.match(ModelicaParser.T__77)             # <<<<<<<<<<<<<<
 *             self.state = 1031
 *             self.class_modification()
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7761, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7761, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_T__77); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7761, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7761, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7761, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_2};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7761, __pyx_L6_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7761, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7761, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":7762
 *             self.state = 1030
 *             self.match(ModelicaParser.T__77)
 *             self.state = 1031             # <<<<<<<<<<<<<<
 *             self.class_modification()
 *         except RecognitionException as re:
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_state, __pyx_int_1031) < 0) __PYX_ERR(0, 7762, __pyx_L6_error)

        /* "pymola/generated/ModelicaParser.pyx":7763
 *             self.match(ModelicaParser.T__77)
 *             self.state = 1031
 *             self.class_modification()             # <<<<<<<<<<<<<<
 *         except RecognitionException as re:
 *             localctx.exception = re
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_class_modification); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7763, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7763, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7763, __pyx_L6_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pymola/generated/ModelicaParser.pyx":7758
 *         localctx = ModelicaParser.AnnotationContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 144, self.RULE_annotation)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 1030
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L13_try_end;
      __pyx_L6_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pymola/generated/ModelicaParser.pyx":7764
 *             self.state = 1031
 *             self.class_modification()
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_RecognitionException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7764, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {
        __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.annotation", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_3, &__pyx_t_4) < 0) __PYX_ERR(0, 7764, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __pyx_v_re = __pyx_t_3;
        /*try:*/ {

          /* "pymola/generated/ModelicaParser.pyx":7765
 *             self.class_modification()
 *         except RecognitionException as re:
 *             localctx.exception = re             # <<<<<<<<<<<<<<
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_localctx, __pyx_n_s_exception, __pyx_v_re) < 0) __PYX_ERR(0, 7765, __pyx_L19_error)

          /* "pymola/generated/ModelicaParser.pyx":7766
 *         except RecognitionException as re:
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)             # <<<<<<<<<<<<<<
 *             self._errHandler.recover(self, re)
 *         finally:
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7766, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_reportError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7766, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7766, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7766, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_11 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 7766, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_11);
            if (__pyx_t_7) {
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_7); __pyx_t_7 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_11, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7766, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pymola/generated/ModelicaParser.pyx":7767
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 *             self._errHandler.recover(self, re)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.exitRule()
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7767, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_recover); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 7767, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_11)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7767, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_11)) {
            PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self, __pyx_v_re};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7767, __pyx_L19_error)
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7767, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_7);
            if (__pyx_t_5) {
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
            }
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_re);
            __Pyx_GIVEREF(__pyx_v_re);
            PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_re);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7767, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }

        /* "pymola/generated/ModelicaParser.pyx":7764
 *             self.state = 1031
 *             self.class_modification()
 *         except RecognitionException as re:             # <<<<<<<<<<<<<<
 *             localctx.exception = re
 *             self._errHandler.reportError(self, re)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_re);
            __pyx_v_re = NULL;
            goto __pyx_L20;
          }
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __pyx_L19_error:;
            __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_14);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __pyx_t_6 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_re);
              __pyx_v_re = NULL;
            }
            __Pyx_PyThreadState_assign
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_17);
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
            }
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
            __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
            goto __pyx_L8_except_error;
          }
          __pyx_L20:;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "pymola/generated/ModelicaParser.pyx":7758
 *         localctx = ModelicaParser.AnnotationContext(self, self._ctx, self.state)
 *         self.enterRule(localctx, 144, self.RULE_annotation)
 *         try:             # <<<<<<<<<<<<<<
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 1030
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L13_try_end:;
    }
  }

  /* "pymola/generated/ModelicaParser.pyx":7769
 *             self._errHandler.recover(self, re)
 *         finally:
 *             self.exitRule()             # <<<<<<<<<<<<<<
 *         return localctx
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7769, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_1) {
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7769, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7769, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_18, &__pyx_t_17);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_18);
      __Pyx_XGOTREF(__pyx_t_17);
      __pyx_t_12 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_20 = __pyx_filename;
      {
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exitRule); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7769, __pyx_L26_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_1) {
          __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7769, __pyx_L26_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7769, __pyx_L26_error)
        }
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_18, __pyx_t_17);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0;
      __pyx_lineno = __pyx_t_12; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_20;
      goto __pyx_L1_error;
      __pyx_L26_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_18, __pyx_t_17);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "pymola/generated/ModelicaParser.pyx":7770
 *         finally:
 *             self.exitRule()
 *         return localctx             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_localctx);
  __pyx_r = __pyx_v_localctx;
  goto __pyx_L0;

  /* "pymola/generated/ModelicaParser.pyx":7754
 * 
 * 
 *     def annotation(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.AnnotationContext(self, self._ctx, self.state)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.annotation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_localctx);
  __Pyx_XDECREF(__pyx_v_re);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7774
 * 
 * 
 *     def sempred(self, localctx:RuleContext, ruleIndex:int, predIndex:int):             # <<<<<<<<<<<<<<
 *         if self._predicates == None:
 *             self._predicates = dict()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_149sempred(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_149sempred = {"sempred", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_149sempred, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_149sempred(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_localctx = 0;
  PyObject *__pyx_v_ruleIndex = 0;
  PyObject *__pyx_v_predIndex = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sempred (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_localctx,&__pyx_n_s_ruleIndex,&__pyx_n_s_predIndex,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_localctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sempred", 1, 4, 4, 1); __PYX_ERR(0, 7774, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ruleIndex)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sempred", 1, 4, 4, 2); __PYX_ERR(0, 7774, __pyx_L3_error)
        }
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_predIndex)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sempred", 1, 4, 4, 3); __PYX_ERR(0, 7774, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "sempred") < 0)) __PYX_ERR(0, 7774, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_self = values[0];
    __pyx_v_localctx = values[1];
    __pyx_v_ruleIndex = values[2];
    __pyx_v_predIndex = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sempred", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7774, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.sempred", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_148sempred(__pyx_self, __pyx_v_self, __pyx_v_localctx, __pyx_v_ruleIndex, __pyx_v_predIndex);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_148sempred(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_localctx, PyObject *__pyx_v_ruleIndex, PyObject *__pyx_v_predIndex) {
  PyObject *__pyx_v_pred = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("sempred", 0);

  /* "pymola/generated/ModelicaParser.pyx":7775
 * 
 *     def sempred(self, localctx:RuleContext, ruleIndex:int, predIndex:int):
 *         if self._predicates == None:             # <<<<<<<<<<<<<<
 *             self._predicates = dict()
 *         self._predicates[56] = self.expr_sempred
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_predicates); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7775, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, Py_None, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7775, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 7775, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "pymola/generated/ModelicaParser.pyx":7776
 *     def sempred(self, localctx:RuleContext, ruleIndex:int, predIndex:int):
 *         if self._predicates == None:
 *             self._predicates = dict()             # <<<<<<<<<<<<<<
 *         self._predicates[56] = self.expr_sempred
 *         pred = self._predicates.get(ruleIndex, None)
 */
    __pyx_t_2 = PyDict_New(); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7776, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_predicates, __pyx_t_2) < 0) __PYX_ERR(0, 7776, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "pymola/generated/ModelicaParser.pyx":7775
 * 
 *     def sempred(self, localctx:RuleContext, ruleIndex:int, predIndex:int):
 *         if self._predicates == None:             # <<<<<<<<<<<<<<
 *             self._predicates = dict()
 *         self._predicates[56] = self.expr_sempred
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7777
 *         if self._predicates == None:
 *             self._predicates = dict()
 *         self._predicates[56] = self.expr_sempred             # <<<<<<<<<<<<<<
 *         pred = self._predicates.get(ruleIndex, None)
 *         if pred is None:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr_sempred); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7777, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_predicates); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7777, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(__Pyx_SetItemInt(__pyx_t_1, 56, __pyx_t_2, long, 1, __Pyx_PyInt_From_long, 0, 0, 0) < 0)) __PYX_ERR(0, 7777, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7778
 *             self._predicates = dict()
 *         self._predicates[56] = self.expr_sempred
 *         pred = self._predicates.get(ruleIndex, None)             # <<<<<<<<<<<<<<
 *         if pred is None:
 *             raise Exception("No predicate with index:" + str(ruleIndex))
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_predicates); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7778, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_get); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7778, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_4)) {
    PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_ruleIndex, Py_None};
    __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7778, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GOTREF(__pyx_t_2);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
    PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_ruleIndex, Py_None};
    __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7778, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GOTREF(__pyx_t_2);
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7778, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_1) {
      __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1); __pyx_t_1 = NULL;
    }
    __Pyx_INCREF(__pyx_v_ruleIndex);
    __Pyx_GIVEREF(__pyx_v_ruleIndex);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_ruleIndex);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, Py_None);
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7778, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_pred = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7779
 *         self._predicates[56] = self.expr_sempred
 *         pred = self._predicates.get(ruleIndex, None)
 *         if pred is None:             # <<<<<<<<<<<<<<
 *             raise Exception("No predicate with index:" + str(ruleIndex))
 *         else:
 */
  __pyx_t_3 = (__pyx_v_pred == Py_None);
  __pyx_t_7 = (__pyx_t_3 != 0);
  if (__pyx_t_7) {

    /* "pymola/generated/ModelicaParser.pyx":7780
 *         pred = self._predicates.get(ruleIndex, None)
 *         if pred is None:
 *             raise Exception("No predicate with index:" + str(ruleIndex))             # <<<<<<<<<<<<<<
 *         else:
 *             return pred(localctx, predIndex)
 */
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7780, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_ruleIndex);
    __Pyx_GIVEREF(__pyx_v_ruleIndex);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_ruleIndex);
    __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)(&PyUnicode_Type)), __pyx_t_2, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7780, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyUnicode_Concat(__pyx_kp_u_No_predicate_with_index, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7780, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7780, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])), __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7780, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 7780, __pyx_L1_error)

    /* "pymola/generated/ModelicaParser.pyx":7779
 *         self._predicates[56] = self.expr_sempred
 *         pred = self._predicates.get(ruleIndex, None)
 *         if pred is None:             # <<<<<<<<<<<<<<
 *             raise Exception("No predicate with index:" + str(ruleIndex))
 *         else:
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7782
 *             raise Exception("No predicate with index:" + str(ruleIndex))
 *         else:
 *             return pred(localctx, predIndex)             # <<<<<<<<<<<<<<
 * 
 *     def expr_sempred(self, localctx:ExprContext, predIndex:int):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_pred);
    __pyx_t_4 = __pyx_v_pred; __pyx_t_6 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_5 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_localctx, __pyx_v_predIndex};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7782, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_localctx, __pyx_v_predIndex};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7782, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else
    #endif
    {
      __pyx_t_1 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7782, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_INCREF(__pyx_v_localctx);
      __Pyx_GIVEREF(__pyx_v_localctx);
      PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_5, __pyx_v_localctx);
      __Pyx_INCREF(__pyx_v_predIndex);
      __Pyx_GIVEREF(__pyx_v_predIndex);
      PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_5, __pyx_v_predIndex);
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7782, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "pymola/generated/ModelicaParser.pyx":7774
 * 
 * 
 *     def sempred(self, localctx:RuleContext, ruleIndex:int, predIndex:int):             # <<<<<<<<<<<<<<
 *         if self._predicates == None:
 *             self._predicates = dict()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.sempred", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pred);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaParser.pyx":7784
 *             return pred(localctx, predIndex)
 * 
 *     def expr_sempred(self, localctx:ExprContext, predIndex:int):             # <<<<<<<<<<<<<<
 *             if predIndex == 0:
 *                 return self.precpred(self._ctx, 7)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_151expr_sempred(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_151expr_sempred = {"expr_sempred", (PyCFunction)__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_151expr_sempred, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_14ModelicaParser_14ModelicaParser_151expr_sempred(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_localctx = 0;
  PyObject *__pyx_v_predIndex = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("expr_sempred (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_localctx,&__pyx_n_s_predIndex,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_localctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("expr_sempred", 1, 3, 3, 1); __PYX_ERR(0, 7784, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_predIndex)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("expr_sempred", 1, 3, 3, 2); __PYX_ERR(0, 7784, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "expr_sempred") < 0)) __PYX_ERR(0, 7784, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_localctx = values[1];
    __pyx_v_predIndex = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("expr_sempred", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 7784, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.expr_sempred", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_150expr_sempred(__pyx_self, __pyx_v_self, __pyx_v_localctx, __pyx_v_predIndex);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_14ModelicaParser_14ModelicaParser_150expr_sempred(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_localctx, PyObject *__pyx_v_predIndex) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("expr_sempred", 0);

  /* "pymola/generated/ModelicaParser.pyx":7785
 * 
 *     def expr_sempred(self, localctx:ExprContext, predIndex:int):
 *             if predIndex == 0:             # <<<<<<<<<<<<<<
 *                 return self.precpred(self._ctx, 7)
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_predIndex, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7785, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 7785, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7786
 *     def expr_sempred(self, localctx:ExprContext, predIndex:int):
 *             if predIndex == 0:
 *                 return self.precpred(self._ctx, 7)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_precpred); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7786, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7786, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_4, __pyx_int_7};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7786, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_4, __pyx_int_7};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7786, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    {
      __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7786, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_t_4);
      __Pyx_INCREF(__pyx_int_7);
      __Pyx_GIVEREF(__pyx_int_7);
      PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_7);
      __pyx_t_4 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7786, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":7785
 * 
 *     def expr_sempred(self, localctx:ExprContext, predIndex:int):
 *             if predIndex == 0:             # <<<<<<<<<<<<<<
 *                 return self.precpred(self._ctx, 7)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7789
 * 
 * 
 *             if predIndex == 1:             # <<<<<<<<<<<<<<
 *                 return self.precpred(self._ctx, 6)
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_predIndex, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7789, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 7789, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7790
 * 
 *             if predIndex == 1:
 *                 return self.precpred(self._ctx, 6)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_precpred); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7790, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7790, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_4 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_7, __pyx_int_6};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7790, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_7, __pyx_int_6};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7790, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7790, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_4) {
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_t_7);
      __Pyx_INCREF(__pyx_int_6);
      __Pyx_GIVEREF(__pyx_int_6);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_int_6);
      __pyx_t_7 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7790, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":7789
 * 
 * 
 *             if predIndex == 1:             # <<<<<<<<<<<<<<
 *                 return self.precpred(self._ctx, 6)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7793
 * 
 * 
 *             if predIndex == 2:             # <<<<<<<<<<<<<<
 *                 return self.precpred(self._ctx, 5)
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_predIndex, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7793, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 7793, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7794
 * 
 *             if predIndex == 2:
 *                 return self.precpred(self._ctx, 5)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_precpred); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7794, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7794, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_5, __pyx_int_5};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7794, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_5, __pyx_int_5};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7794, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7794, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_t_5);
      __Pyx_INCREF(__pyx_int_5);
      __Pyx_GIVEREF(__pyx_int_5);
      PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_int_5);
      __pyx_t_5 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7794, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":7793
 * 
 * 
 *             if predIndex == 2:             # <<<<<<<<<<<<<<
 *                 return self.precpred(self._ctx, 5)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7797
 * 
 * 
 *             if predIndex == 3:             # <<<<<<<<<<<<<<
 *                 return self.precpred(self._ctx, 3)
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_predIndex, __pyx_int_3, 3, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7797, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 7797, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7798
 * 
 *             if predIndex == 3:
 *                 return self.precpred(self._ctx, 3)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_precpred); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7798, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7798, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_4, __pyx_int_3};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7798, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_4, __pyx_int_3};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7798, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    {
      __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7798, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_t_4);
      __Pyx_INCREF(__pyx_int_3);
      __Pyx_GIVEREF(__pyx_int_3);
      PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_3);
      __pyx_t_4 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7798, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":7797
 * 
 * 
 *             if predIndex == 3:             # <<<<<<<<<<<<<<
 *                 return self.precpred(self._ctx, 3)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7801
 * 
 * 
 *             if predIndex == 4:             # <<<<<<<<<<<<<<
 *                 return self.precpred(self._ctx, 2)
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_predIndex, __pyx_int_4, 4, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 7801, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "pymola/generated/ModelicaParser.pyx":7802
 * 
 *             if predIndex == 4:
 *                 return self.precpred(self._ctx, 2)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_precpred); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7802, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ctx); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7802, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_4 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_7, __pyx_int_2};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7802, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_7, __pyx_int_2};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7802, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7802, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_4) {
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_t_7);
      __Pyx_INCREF(__pyx_int_2);
      __Pyx_GIVEREF(__pyx_int_2);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_int_2);
      __pyx_t_7 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7802, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "pymola/generated/ModelicaParser.pyx":7801
 * 
 * 
 *             if predIndex == 4:             # <<<<<<<<<<<<<<
 *                 return self.precpred(self._ctx, 2)
 * 
 */
  }

  /* "pymola/generated/ModelicaParser.pyx":7784
 *             return pred(localctx, predIndex)
 * 
 *     def expr_sempred(self, localctx:ExprContext, predIndex:int):             # <<<<<<<<<<<<<<
 *             if predIndex == 0:
 *                 return self.precpred(self._ctx, 7)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pymola.generated.ModelicaParser.ModelicaParser.expr_sempred", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

static int __pyx_import_star_set(PyObject *o, PyObject* py_name, char *name) {
  static const char* internal_type_names[] = {
    "__pyx_ctuple_long",
    "__pyx_ctuple_long_struct",
    0
  };
  const char** type_name = internal_type_names;
  while (*type_name) {
    if (__Pyx_StrEq(name, *type_name)) {
      PyErr_Format(PyExc_TypeError, "Cannot overwrite C type %s", name);
      goto bad;
    }
    type_name++;
  }
  if (0);
  else {
    if (PyObject_SetAttr(__pyx_m, py_name, o) < 0) goto bad;
  }
  return 0;
  bad:
  return -1;
}

























































































































static int
__Pyx_import_all_from(PyObject *locals, PyObject *v)
{
    PyObject *all = PyObject_GetAttrString(v, "__all__");
    PyObject *dict, *name, *value;
    int skip_leading_underscores = 0;
    int pos, err;

    if (all == NULL) {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError))
            return -1;
        PyErr_Clear();
        dict = PyObject_GetAttrString(v, "__dict__");
        if (dict == NULL) {
            if (!PyErr_ExceptionMatches(PyExc_AttributeError))
                return -1;
            PyErr_SetString(PyExc_ImportError,
            "from-import-* object has no __dict__ and no __all__");
            return -1;
        }
#if PY_MAJOR_VERSION < 3
        all = PyObject_CallMethod(dict, (char *)"keys", NULL);
#else
        all = PyMapping_Keys(dict);
#endif
        Py_DECREF(dict);
        if (all == NULL)
            return -1;
        skip_leading_underscores = 1;
    }

    for (pos = 0, err = 0; ; pos++) {
        name = PySequence_GetItem(all, pos);
        if (name == NULL) {
            if (!PyErr_ExceptionMatches(PyExc_IndexError))
                err = -1;
            else
                PyErr_Clear();
            break;
        }
        if (skip_leading_underscores &&
#if PY_MAJOR_VERSION < 3
            PyString_Check(name) &&
            PyString_AS_STRING(name)[0] == '_')
#else
            PyUnicode_Check(name) &&
            PyUnicode_AS_UNICODE(name)[0] == '_')
#endif
        {
            Py_DECREF(name);
            continue;
        }
        value = PyObject_GetAttr(v, name);
        if (value == NULL)
            err = -1;
        else if (PyDict_CheckExact(locals))
            err = PyDict_SetItem(locals, name, value);
        else
            err = PyObject_SetItem(locals, name, value);
        Py_DECREF(name);
        Py_XDECREF(value);
        if (err != 0)
            break;
    }
    Py_DECREF(all);
    return err;
}


static int __pyx_import_star(PyObject* m) {

    int i;
    int ret = -1;
    char* s;
    PyObject *locals = 0;
    PyObject *list = 0;
#if PY_MAJOR_VERSION >= 3
    PyObject *utf8_name = 0;
#endif
    PyObject *name;
    PyObject *item;

    locals = PyDict_New();              if (!locals) goto bad;
    if (__Pyx_import_all_from(locals, m) < 0) goto bad;
    list = PyDict_Items(locals);        if (!list) goto bad;

    for(i=0; i<PyList_GET_SIZE(list); i++) {
        name = PyTuple_GET_ITEM(PyList_GET_ITEM(list, i), 0);
        item = PyTuple_GET_ITEM(PyList_GET_ITEM(list, i), 1);
#if PY_MAJOR_VERSION >= 3
        utf8_name = PyUnicode_AsUTF8String(name);
        if (!utf8_name) goto bad;
        s = PyBytes_AS_STRING(utf8_name);
        if (__pyx_import_star_set(item, name, s) < 0) goto bad;
        Py_DECREF(utf8_name); utf8_name = 0;
#else
        s = PyString_AsString(name);
        if (!s) goto bad;
        if (__pyx_import_star_set(item, name, s) < 0) goto bad;
#endif
    }
    ret = 0;

bad:
    Py_XDECREF(locals);
    Py_XDECREF(list);
#if PY_MAJOR_VERSION >= 3
    Py_XDECREF(utf8_name);
#endif
    return ret;
}



#if PY_MAJOR_VERSION >= 3
static struct PyModuleDef __pyx_moduledef = {
  #if PY_VERSION_HEX < 0x03020000
    { PyObject_HEAD_INIT(NULL) NULL, 0, NULL },
  #else
    PyModuleDef_HEAD_INIT,
  #endif
    "ModelicaParser",
    0, /* m_doc */
    -1, /* m_size */
    __pyx_methods /* m_methods */,
    NULL, /* m_reload */
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_u_, __pyx_k_, sizeof(__pyx_k_), 0, 1, 0, 0},
  {&__pyx_kp_u_0, __pyx_k_0, sizeof(__pyx_k_0), 0, 1, 0, 0},
  {&__pyx_kp_u_02468_BDFHJLNPRTVXZ_bdfhjlnprtv, __pyx_k_02468_BDFHJLNPRTVXZ_bdfhjlnprtv, sizeof(__pyx_k_02468_BDFHJLNPRTVXZ_bdfhjlnprtv), 0, 1, 0, 0},
  {&__pyx_kp_u_0_0_0, __pyx_k_0_0_0, sizeof(__pyx_k_0_0_0), 0, 1, 0, 0},
  {&__pyx_kp_u_0_0_0_0_0_0_0_0, __pyx_k_0_0_0_0_0_0_0_0, sizeof(__pyx_k_0_0_0_0_0_0_0_0), 0, 1, 0, 0},
  {&__pyx_kp_u_0_0_0_1_1_1_1_1_1_1, __pyx_k_0_0_0_1_1_1_1_1_1_1, sizeof(__pyx_k_0_0_0_1_1_1_1_1_1_1), 0, 1, 0, 0},
  {&__pyx_kp_u_0_0_1_1_2_2_3_3_4_4, __pyx_k_0_0_1_1_2_2_3_3_4_4, sizeof(__pyx_k_0_0_1_1_2_2_3_3_4_4), 0, 1, 0, 0},
  {&__pyx_kp_u_0_2_4, __pyx_k_0_2_4, sizeof(__pyx_k_0_2_4), 0, 1, 0, 0},
  {&__pyx_kp_u_0_T, __pyx_k_0_T, sizeof(__pyx_k_0_T), 0, 1, 0, 0},
  {&__pyx_kp_u_0_T_2, __pyx_k_0_T_2, sizeof(__pyx_k_0_T_2), 0, 1, 0, 0},
  {&__pyx_kp_u_0_X, __pyx_k_0_X, sizeof(__pyx_k_0_X), 0, 1, 0, 0},
  {&__pyx_kp_u_0_X_2, __pyx_k_0_X_2, sizeof(__pyx_k_0_X_2), 0, 1, 0, 0},
  {&__pyx_kp_u_1, __pyx_k_1, sizeof(__pyx_k_1), 0, 1, 0, 0},
  {&__pyx_kp_u_1_1_1_1_1_1_1_2_2, __pyx_k_1_1_1_1_1_1_1_2_2, sizeof(__pyx_k_1_1_1_1_1_1_1_2_2), 0, 1, 0, 0},
  {&__pyx_kp_u_1_2, __pyx_k_1_2, sizeof(__pyx_k_1_2), 0, 1, 0, 0},
  {&__pyx_kp_u_1_3, __pyx_k_1_3, sizeof(__pyx_k_1_3), 0, 1, 0, 0},
  {&__pyx_kp_u_1_n8, __pyx_k_1_n8, sizeof(__pyx_k_1_n8), 0, 1, 0, 0},
  {&__pyx_kp_u_1_n8_0, __pyx_k_1_n8_0, sizeof(__pyx_k_1_n8_0), 0, 1, 0, 0},
  {&__pyx_kp_u_2, __pyx_k_2, sizeof(__pyx_k_2), 0, 1, 0, 0},
  {&__pyx_kp_u_2_2, __pyx_k_2_2, sizeof(__pyx_k_2_2), 0, 1, 0, 0},
  {&__pyx_kp_u_2_2_2_2_2_2_3_3, __pyx_k_2_2_2_2_2_2_3_3, sizeof(__pyx_k_2_2_2_2_2_2_3_3), 0, 1, 0, 0},
  {&__pyx_kp_u_2_X, __pyx_k_2_X, sizeof(__pyx_k_2_X), 0, 1, 0, 0},
  {&__pyx_kp_u_2_n8, __pyx_k_2_n8, sizeof(__pyx_k_2_n8), 0, 1, 0, 0},
  {&__pyx_kp_u_3, __pyx_k_3, sizeof(__pyx_k_3), 0, 1, 0, 0},
  {&__pyx_kp_u_3_3_3_4_4_4_4_4_4, __pyx_k_3_3_3_4_4_4_4_4_4, sizeof(__pyx_k_3_3_3_4_4_4_4_4_4), 0, 1, 0, 0},
  {&__pyx_kp_u_3_a_d3, __pyx_k_3_a_d3, sizeof(__pyx_k_3_a_d3), 0, 1, 0, 0},
  {&__pyx_kp_u_3_b2, __pyx_k_3_b2, sizeof(__pyx_k_3_b2), 0, 1, 0, 0},
  {&__pyx_kp_u_3_b2_2, __pyx_k_3_b2_2, sizeof(__pyx_k_3_b2_2), 0, 1, 0, 0},
  {&__pyx_kp_u_3_b2_4, __pyx_k_3_b2_4, sizeof(__pyx_k_3_b2_4), 0, 1, 0, 0},
  {&__pyx_kp_u_4_4_4_4_4_4_4_4, __pyx_k_4_4_4_4_4_4_4_4, sizeof(__pyx_k_4_4_4_4_4_4_4_4), 0, 1, 0, 0},
  {&__pyx_kp_u_4_7, __pyx_k_4_7, sizeof(__pyx_k_4_7), 0, 1, 0, 0},
  {&__pyx_kp_u_4_X, __pyx_k_4_X, sizeof(__pyx_k_4_X), 0, 1, 0, 0},
  {&__pyx_kp_u_5_5_5_5_5_5_5_5, __pyx_k_5_5_5_5_5_5_5_5, sizeof(__pyx_k_5_5_5_5_5_5_5_5), 0, 1, 0, 0},
  {&__pyx_kp_u_5_5_5_5_5_5_5_5_2, __pyx_k_5_5_5_5_5_5_5_5_2, sizeof(__pyx_k_5_5_5_5_5_5_5_5_2), 0, 1, 0, 0},
  {&__pyx_kp_u_5_5_5_5_5_5_5_5_5_5, __pyx_k_5_5_5_5_5_5_5_5_5_5, sizeof(__pyx_k_5_5_5_5_5_5_5_5_5_5), 0, 1, 0, 0},
  {&__pyx_kp_u_5_5_5_5_6_6_6_6_6, __pyx_k_5_5_5_5_6_6_6_6_6, sizeof(__pyx_k_5_5_5_5_6_6_6_6_6), 0, 1, 0, 0},
  {&__pyx_kp_u_5_5_6_6_7_7_8_8_9_9, __pyx_k_5_5_6_6_7_7_8_8_9_9, sizeof(__pyx_k_5_5_6_6_7_7_8_8_9_9), 0, 1, 0, 0},
  {&__pyx_kp_u_6, __pyx_k_6, sizeof(__pyx_k_6), 0, 1, 0, 0},
  {&__pyx_kp_u_6_4, __pyx_k_6_4, sizeof(__pyx_k_6_4), 0, 1, 0, 0},
  {&__pyx_kp_u_6_6_6_6_6_6_6_6, __pyx_k_6_6_6_6_6_6_6_6, sizeof(__pyx_k_6_6_6_6_6_6_6_6), 0, 1, 0, 0},
  {&__pyx_kp_u_6_6_6_6_6_6_6_6_2, __pyx_k_6_6_6_6_6_6_6_6_2, sizeof(__pyx_k_6_6_6_6_6_6_6_6_2), 0, 1, 0, 0},
  {&__pyx_kp_u_6_6_6_6_6_6_6_6_3, __pyx_k_6_6_6_6_6_6_6_6_3, sizeof(__pyx_k_6_6_6_6_6_6_6_6_3), 0, 1, 0, 0},
  {&__pyx_kp_u_6_7_7_7_7_7_7_7_8_8_8_8_8, __pyx_k_6_7_7_7_7_7_7_7_8_8_8_8_8, sizeof(__pyx_k_6_7_7_7_7_7_7_7_8_8_8_8_8), 0, 1, 0, 0},
  {&__pyx_kp_u_6_8, __pyx_k_6_8, sizeof(__pyx_k_6_8), 0, 1, 0, 0},
  {&__pyx_kp_u_6_n8_4, __pyx_k_6_n8_4, sizeof(__pyx_k_6_n8_4), 0, 1, 0, 0},
  {&__pyx_kp_u_7, __pyx_k_7, sizeof(__pyx_k_7), 0, 1, 0, 0},
  {&__pyx_kp_u_7_i, __pyx_k_7_i, sizeof(__pyx_k_7_i), 0, 1, 0, 0},
  {&__pyx_kp_u_7_k, __pyx_k_7_k, sizeof(__pyx_k_7_k), 0, 1, 0, 0},
  {&__pyx_kp_u_7_n8, __pyx_k_7_n8, sizeof(__pyx_k_7_n8), 0, 1, 0, 0},
  {&__pyx_kp_u_8, __pyx_k_8, sizeof(__pyx_k_8), 0, 1, 0, 0},
  {&__pyx_kp_u_8_2, __pyx_k_8_2, sizeof(__pyx_k_8_2), 0, 1, 0, 0},
  {&__pyx_kp_u_8_3, __pyx_k_8_3, sizeof(__pyx_k_8_3), 0, 1, 0, 0},
  {&__pyx_kp_u_8_8_8_8_8_8_8_8_8_8_8_8_8, __pyx_k_8_8_8_8_8_8_8_8_8_8_8_8_8, sizeof(__pyx_k_8_8_8_8_8_8_8_8_8_8_8_8_8), 0, 1, 0, 0},
  {&__pyx_kp_u_8_8_9_9_9_9_9_9_9_9_9, __pyx_k_8_8_9_9_9_9_9_9_9_9_9, sizeof(__pyx_k_8_8_9_9_9_9_9_9_9_9_9), 0, 1, 0, 0},
  {&__pyx_kp_u_9, __pyx_k_9, sizeof(__pyx_k_9), 0, 1, 0, 0},
  {&__pyx_kp_u_9_z, __pyx_k_9_z, sizeof(__pyx_k_9_z), 0, 1, 0, 0},
  {&__pyx_kp_u_A, __pyx_k_A, sizeof(__pyx_k_A), 0, 1, 0, 0},
  {&__pyx_n_s_ATN, __pyx_k_ATN, sizeof(__pyx_k_ATN), 0, 0, 1, 1},
  {&__pyx_n_s_ATNDeserializer, __pyx_k_ATNDeserializer, sizeof(__pyx_k_ATNDeserializer), 0, 0, 1, 1},
  {&__pyx_kp_u_A_2, __pyx_k_A_2, sizeof(__pyx_k_A_2), 0, 1, 0, 0},
  {&__pyx_kp_u_A_3, __pyx_k_A_3, sizeof(__pyx_k_A_3), 0, 1, 0, 0},
  {&__pyx_kp_u_A_4, __pyx_k_A_4, sizeof(__pyx_k_A_4), 0, 1, 0, 0},
  {&__pyx_kp_u_A_A_A_A_A_A_A_B_B_B_B_C_C, __pyx_k_A_A_A_A_A_A_A_B_B_B_B_C_C, sizeof(__pyx_k_A_A_A_A_A_A_A_B_B_B_B_C_C), 0, 1, 0, 0},
  {&__pyx_kp_u_A_A_B_B_C_C_D, __pyx_k_A_A_B_B_C_C_D, sizeof(__pyx_k_A_A_B_B_C_C_D), 0, 1, 0, 0},
  {&__pyx_n_s_Algorithm_sectionContext, __pyx_k_Algorithm_sectionContext, sizeof(__pyx_k_Algorithm_sectionContext), 0, 0, 1, 1},
  {&__pyx_n_s_AnnotationContext, __pyx_k_AnnotationContext, sizeof(__pyx_k_AnnotationContext), 0, 0, 1, 1},
  {&__pyx_n_s_ArgumentContext, __pyx_k_ArgumentContext, sizeof(__pyx_k_ArgumentContext), 0, 0, 1, 1},
  {&__pyx_n_s_Argument_expressionContext, __pyx_k_Argument_expressionContext, sizeof(__pyx_k_Argument_expressionContext), 0, 0, 1, 1},
  {&__pyx_n_s_Argument_functionContext, __pyx_k_Argument_functionContext, sizeof(__pyx_k_Argument_functionContext), 0, 0, 1, 1},
  {&__pyx_n_s_Argument_listContext, __pyx_k_Argument_listContext, sizeof(__pyx_k_Argument_listContext), 0, 0, 1, 1},
  {&__pyx_n_s_Array_subscriptsContext, __pyx_k_Array_subscriptsContext, sizeof(__pyx_k_Array_subscriptsContext), 0, 0, 1, 1},
  {&__pyx_kp_u_B, __pyx_k_B, sizeof(__pyx_k_B), 0, 1, 0, 0},
  {&__pyx_kp_u_B_2, __pyx_k_B_2, sizeof(__pyx_k_B_2), 0, 1, 0, 0},
  {&__pyx_kp_u_B_3, __pyx_k_B_3, sizeof(__pyx_k_B_3), 0, 1, 0, 0},
  {&__pyx_kp_u_B_CH, __pyx_k_B_CH, sizeof(__pyx_k_B_CH), 0, 1, 0, 0},
  {&__pyx_kp_u_B_F, __pyx_k_B_F, sizeof(__pyx_k_B_F), 0, 1, 0, 0},
  {&__pyx_n_s_Base_prefixContext, __pyx_k_Base_prefixContext, sizeof(__pyx_k_Base_prefixContext), 0, 0, 1, 1},
  {&__pyx_kp_u_C, __pyx_k_C, sizeof(__pyx_k_C), 0, 1, 0, 0},
  {&__pyx_n_s_COMMENT, __pyx_k_COMMENT, sizeof(__pyx_k_COMMENT), 0, 0, 1, 1},
  {&__pyx_n_u_COMMENT, __pyx_k_COMMENT, sizeof(__pyx_k_COMMENT), 0, 1, 0, 1},
  {&__pyx_kp_u_C_2, __pyx_k_C_2, sizeof(__pyx_k_C_2), 0, 1, 0, 0},
  {&__pyx_kp_u_C_3, __pyx_k_C_3, sizeof(__pyx_k_C_3), 0, 1, 0, 0},
  {&__pyx_kp_u_C_C_C_C_C_C_C_C_C_D_D, __pyx_k_C_C_C_C_C_C_C_C_C_D_D, sizeof(__pyx_k_C_C_C_C_C_C_C_C_C_D_D), 0, 1, 0, 0},
  {&__pyx_n_s_Class_definitionContext, __pyx_k_Class_definitionContext, sizeof(__pyx_k_Class_definitionContext), 0, 0, 1, 1},
  {&__pyx_n_s_Class_modificationContext, __pyx_k_Class_modificationContext, sizeof(__pyx_k_Class_modificationContext), 0, 0, 1, 1},
  {&__pyx_n_s_Class_prefixesContext, __pyx_k_Class_prefixesContext, sizeof(__pyx_k_Class_prefixesContext), 0, 0, 1, 1},
  {&__pyx_n_s_Class_spec_baseContext, __pyx_k_Class_spec_baseContext, sizeof(__pyx_k_Class_spec_baseContext), 0, 0, 1, 1},
  {&__pyx_n_s_Class_spec_compContext, __pyx_k_Class_spec_compContext, sizeof(__pyx_k_Class_spec_compContext), 0, 0, 1, 1},
  {&__pyx_n_s_Class_spec_derContext, __pyx_k_Class_spec_derContext, sizeof(__pyx_k_Class_spec_derContext), 0, 0, 1, 1},
  {&__pyx_n_s_Class_spec_enumContext, __pyx_k_Class_spec_enumContext, sizeof(__pyx_k_Class_spec_enumContext), 0, 0, 1, 1},
  {&__pyx_n_s_Class_spec_extendsContext, __pyx_k_Class_spec_extendsContext, sizeof(__pyx_k_Class_spec_extendsContext), 0, 0, 1, 1},
  {&__pyx_n_s_Class_specifierContext, __pyx_k_Class_specifierContext, sizeof(__pyx_k_Class_specifierContext), 0, 0, 1, 1},
  {&__pyx_n_s_Class_typeContext, __pyx_k_Class_typeContext, sizeof(__pyx_k_Class_typeContext), 0, 0, 1, 1},
  {&__pyx_n_s_CommentContext, __pyx_k_CommentContext, sizeof(__pyx_k_CommentContext), 0, 0, 1, 1},
  {&__pyx_n_s_Component_clause1Context, __pyx_k_Component_clause1Context, sizeof(__pyx_k_Component_clause1Context), 0, 0, 1, 1},
  {&__pyx_n_s_Component_clauseContext, __pyx_k_Component_clauseContext, sizeof(__pyx_k_Component_clauseContext), 0, 0, 1, 1},
  {&__pyx_n_s_Component_declaration1Context, __pyx_k_Component_declaration1Context, sizeof(__pyx_k_Component_declaration1Context), 0, 0, 1, 1},
  {&__pyx_n_s_Component_declarationContext, __pyx_k_Component_declarationContext, sizeof(__pyx_k_Component_declarationContext), 0, 0, 1, 1},
  {&__pyx_n_s_Component_listContext, __pyx_k_Component_listContext, sizeof(__pyx_k_Component_listContext), 0, 0, 1, 1},
  {&__pyx_n_s_Component_referenceContext, __pyx_k_Component_referenceContext, sizeof(__pyx_k_Component_referenceContext), 0, 0, 1, 1},
  {&__pyx_n_s_Component_reference_elementConte, __pyx_k_Component_reference_elementConte, sizeof(__pyx_k_Component_reference_elementConte), 0, 0, 1, 1},
  {&__pyx_n_s_CompositionContext, __pyx_k_CompositionContext, sizeof(__pyx_k_CompositionContext), 0, 0, 1, 1},
  {&__pyx_n_s_Condition_attributeContext, __pyx_k_Condition_attributeContext, sizeof(__pyx_k_Condition_attributeContext), 0, 0, 1, 1},
  {&__pyx_n_s_Connect_clauseContext, __pyx_k_Connect_clauseContext, sizeof(__pyx_k_Connect_clauseContext), 0, 0, 1, 1},
  {&__pyx_n_s_Constraining_clauseContext, __pyx_k_Constraining_clauseContext, sizeof(__pyx_k_Constraining_clauseContext), 0, 0, 1, 1},
  {&__pyx_kp_u_D, __pyx_k_D, sizeof(__pyx_k_D), 0, 1, 0, 0},
  {&__pyx_n_s_DFA, __pyx_k_DFA, sizeof(__pyx_k_DFA), 0, 0, 1, 1},
  {&__pyx_kp_u_D_2, __pyx_k_D_2, sizeof(__pyx_k_D_2), 0, 1, 0, 0},
  {&__pyx_kp_u_D_D_D_D_D_D_D_D_E_E_E_E, __pyx_k_D_D_D_D_D_D_D_D_E_E_E_E, sizeof(__pyx_k_D_D_D_D_D_D_D_D_E_E_E_E), 0, 1, 0, 0},
  {&__pyx_kp_u_D_E_E_F_F_G_G_H_H_I_I_J_J, __pyx_k_D_E_E_F_F_G_G_H_H_I_I_J_J, sizeof(__pyx_k_D_E_E_F_F_G_G_H_H_I_I_J_J), 0, 1, 0, 0},
  {&__pyx_kp_u_D_F_H_J, __pyx_k_D_F_H_J, sizeof(__pyx_k_D_F_H_J), 0, 1, 0, 0},
  {&__pyx_n_s_DeclarationContext, __pyx_k_DeclarationContext, sizeof(__pyx_k_DeclarationContext), 0, 0, 1, 1},
  {&__pyx_kp_u_E, __pyx_k_E, sizeof(__pyx_k_E), 0, 1, 0, 0},
  {&__pyx_n_s_EACH, __pyx_k_EACH, sizeof(__pyx_k_EACH), 0, 0, 1, 1},
  {&__pyx_n_u_EACH, __pyx_k_EACH, sizeof(__pyx_k_EACH), 0, 1, 0, 1},
  {&__pyx_n_s_ENCAPSULATED, __pyx_k_ENCAPSULATED, sizeof(__pyx_k_ENCAPSULATED), 0, 0, 1, 1},
  {&__pyx_n_u_ENCAPSULATED, __pyx_k_ENCAPSULATED, sizeof(__pyx_k_ENCAPSULATED), 0, 1, 0, 1},
  {&__pyx_n_s_EOF, __pyx_k_EOF, sizeof(__pyx_k_EOF), 0, 0, 1, 1},
  {&__pyx_kp_u_E_2, __pyx_k_E_2, sizeof(__pyx_k_E_2), 0, 1, 0, 0},
  {&__pyx_kp_u_E_E, __pyx_k_E_E, sizeof(__pyx_k_E_E), 0, 1, 0, 0},
  {&__pyx_kp_u_E_E_E_E_F_F_F_F_F_F_F_F, __pyx_k_E_E_E_E_F_F_F_F_F_F_F_F, sizeof(__pyx_k_E_E_E_E_F_F_F_F_F_F_F_F), 0, 1, 0, 0},
  {&__pyx_n_s_ElementContext, __pyx_k_ElementContext, sizeof(__pyx_k_ElementContext), 0, 0, 1, 1},
  {&__pyx_n_s_Element_listContext, __pyx_k_Element_listContext, sizeof(__pyx_k_Element_listContext), 0, 0, 1, 1},
  {&__pyx_n_s_Element_modificationContext, __pyx_k_Element_modificationContext, sizeof(__pyx_k_Element_modificationContext), 0, 0, 1, 1},
  {&__pyx_n_s_Element_modification_or_replacea, __pyx_k_Element_modification_or_replacea, sizeof(__pyx_k_Element_modification_or_replacea), 0, 0, 1, 1},
  {&__pyx_n_s_Element_redeclarationContext, __pyx_k_Element_redeclarationContext, sizeof(__pyx_k_Element_redeclarationContext), 0, 0, 1, 1},
  {&__pyx_n_s_Element_replaceableContext, __pyx_k_Element_replaceableContext, sizeof(__pyx_k_Element_replaceableContext), 0, 0, 1, 1},
  {&__pyx_n_s_Enum_listContext, __pyx_k_Enum_listContext, sizeof(__pyx_k_Enum_listContext), 0, 0, 1, 1},
  {&__pyx_n_s_Enumeration_literalContext, __pyx_k_Enumeration_literalContext, sizeof(__pyx_k_Enumeration_literalContext), 0, 0, 1, 1},
  {&__pyx_n_s_EquationContext, __pyx_k_EquationContext, sizeof(__pyx_k_EquationContext), 0, 0, 1, 1},
  {&__pyx_n_s_Equation_connect_clauseContext, __pyx_k_Equation_connect_clauseContext, sizeof(__pyx_k_Equation_connect_clauseContext), 0, 0, 1, 1},
  {&__pyx_n_s_Equation_forContext, __pyx_k_Equation_forContext, sizeof(__pyx_k_Equation_forContext), 0, 0, 1, 1},
  {&__pyx_n_s_Equation_functionContext, __pyx_k_Equation_functionContext, sizeof(__pyx_k_Equation_functionContext), 0, 0, 1, 1},
  {&__pyx_n_s_Equation_ifContext, __pyx_k_Equation_ifContext, sizeof(__pyx_k_Equation_ifContext), 0, 0, 1, 1},
  {&__pyx_n_s_Equation_optionsContext, __pyx_k_Equation_optionsContext, sizeof(__pyx_k_Equation_optionsContext), 0, 0, 1, 1},
  {&__pyx_n_s_Equation_sectionContext, __pyx_k_Equation_sectionContext, sizeof(__pyx_k_Equation_sectionContext), 0, 0, 1, 1},
  {&__pyx_n_s_Equation_simpleContext, __pyx_k_Equation_simpleContext, sizeof(__pyx_k_Equation_simpleContext), 0, 0, 1, 1},
  {&__pyx_n_s_Equation_whenContext, __pyx_k_Equation_whenContext, sizeof(__pyx_k_Equation_whenContext), 0, 0, 1, 1},
  {&__pyx_n_s_ExprContext, __pyx_k_ExprContext, sizeof(__pyx_k_ExprContext), 0, 0, 1, 1},
  {&__pyx_n_s_Expr_addContext, __pyx_k_Expr_addContext, sizeof(__pyx_k_Expr_addContext), 0, 0, 1, 1},
  {&__pyx_n_s_Expr_andContext, __pyx_k_Expr_andContext, sizeof(__pyx_k_Expr_andContext), 0, 0, 1, 1},
  {&__pyx_n_s_Expr_expContext, __pyx_k_Expr_expContext, sizeof(__pyx_k_Expr_expContext), 0, 0, 1, 1},
  {&__pyx_n_s_Expr_mulContext, __pyx_k_Expr_mulContext, sizeof(__pyx_k_Expr_mulContext), 0, 0, 1, 1},
  {&__pyx_n_s_Expr_negContext, __pyx_k_Expr_negContext, sizeof(__pyx_k_Expr_negContext), 0, 0, 1, 1},
  {&__pyx_n_s_Expr_notContext, __pyx_k_Expr_notContext, sizeof(__pyx_k_Expr_notContext), 0, 0, 1, 1},
  {&__pyx_n_s_Expr_orContext, __pyx_k_Expr_orContext, sizeof(__pyx_k_Expr_orContext), 0, 0, 1, 1},
  {&__pyx_n_s_Expr_primaryContext, __pyx_k_Expr_primaryContext, sizeof(__pyx_k_Expr_primaryContext), 0, 0, 1, 1},
  {&__pyx_n_s_Expr_relContext, __pyx_k_Expr_relContext, sizeof(__pyx_k_Expr_relContext), 0, 0, 1, 1},
  {&__pyx_n_s_ExpressionContext, __pyx_k_ExpressionContext, sizeof(__pyx_k_ExpressionContext), 0, 0, 1, 1},
  {&__pyx_n_s_Expression_ifContext, __pyx_k_Expression_ifContext, sizeof(__pyx_k_Expression_ifContext), 0, 0, 1, 1},
  {&__pyx_n_s_Expression_listContext, __pyx_k_Expression_listContext, sizeof(__pyx_k_Expression_listContext), 0, 0, 1, 1},
  {&__pyx_n_s_Expression_simpleContext, __pyx_k_Expression_simpleContext, sizeof(__pyx_k_Expression_simpleContext), 0, 0, 1, 1},
  {&__pyx_n_s_Extends_clauseContext, __pyx_k_Extends_clauseContext, sizeof(__pyx_k_Extends_clauseContext), 0, 0, 1, 1},
  {&__pyx_n_s_External_function_callContext, __pyx_k_External_function_callContext, sizeof(__pyx_k_External_function_callContext), 0, 0, 1, 1},
  {&__pyx_kp_u_F, __pyx_k_F, sizeof(__pyx_k_F), 0, 1, 0, 0},
  {&__pyx_n_s_FINAL, __pyx_k_FINAL, sizeof(__pyx_k_FINAL), 0, 0, 1, 1},
  {&__pyx_n_u_FINAL, __pyx_k_FINAL, sizeof(__pyx_k_FINAL), 0, 1, 0, 1},
  {&__pyx_kp_u_F_2, __pyx_k_F_2, sizeof(__pyx_k_F_2), 0, 1, 0, 0},
  {&__pyx_kp_u_F_F_F_G_G_G_G_H_H_H_H_I_I, __pyx_k_F_F_F_G_G_G_G_H_H_H_H_I_I, sizeof(__pyx_k_F_F_F_G_G_G_G_H_H_H_H_I_I), 0, 1, 0, 0},
  {&__pyx_n_s_FailedPredicateException, __pyx_k_FailedPredicateException, sizeof(__pyx_k_FailedPredicateException), 0, 0, 1, 1},
  {&__pyx_n_s_For_equationContext, __pyx_k_For_equationContext, sizeof(__pyx_k_For_equationContext), 0, 0, 1, 1},
  {&__pyx_n_s_For_indexContext, __pyx_k_For_indexContext, sizeof(__pyx_k_For_indexContext), 0, 0, 1, 1},
  {&__pyx_n_s_For_indicesContext, __pyx_k_For_indicesContext, sizeof(__pyx_k_For_indicesContext), 0, 0, 1, 1},
  {&__pyx_n_s_For_statementContext, __pyx_k_For_statementContext, sizeof(__pyx_k_For_statementContext), 0, 0, 1, 1},
  {&__pyx_n_s_Function_argumentContext, __pyx_k_Function_argumentContext, sizeof(__pyx_k_Function_argumentContext), 0, 0, 1, 1},
  {&__pyx_n_s_Function_argumentsContext, __pyx_k_Function_argumentsContext, sizeof(__pyx_k_Function_argumentsContext), 0, 0, 1, 1},
  {&__pyx_n_s_Function_call_argsContext, __pyx_k_Function_call_argsContext, sizeof(__pyx_k_Function_call_argsContext), 0, 0, 1, 1},
  {&__pyx_kp_u_G, __pyx_k_G, sizeof(__pyx_k_G), 0, 1, 0, 0},
  {&__pyx_kp_u_G_2, __pyx_k_G_2, sizeof(__pyx_k_G_2), 0, 1, 0, 0},
  {&__pyx_kp_u_H, __pyx_k_H, sizeof(__pyx_k_H), 0, 1, 0, 0},
  {&__pyx_kp_u_H_2, __pyx_k_H_2, sizeof(__pyx_k_H_2), 0, 1, 0, 0},
  {&__pyx_kp_u_H_3, __pyx_k_H_3, sizeof(__pyx_k_H_3), 0, 1, 0, 0},
  {&__pyx_kp_u_H_3_2, __pyx_k_H_3_2, sizeof(__pyx_k_H_3_2), 0, 1, 0, 0},
  {&__pyx_kp_u_H_4, __pyx_k_H_4, sizeof(__pyx_k_H_4), 0, 1, 0, 0},
  {&__pyx_kp_u_H_5, __pyx_k_H_5, sizeof(__pyx_k_H_5), 0, 1, 0, 0},
  {&__pyx_kp_u_H_6, __pyx_k_H_6, sizeof(__pyx_k_H_6), 0, 1, 0, 0},
  {&__pyx_kp_u_H_7, __pyx_k_H_7, sizeof(__pyx_k_H_7), 0, 1, 0, 0},
  {&__pyx_kp_u_H_8, __pyx_k_H_8, sizeof(__pyx_k_H_8), 0, 1, 0, 0},
  {&__pyx_kp_u_H_K, __pyx_k_H_K, sizeof(__pyx_k_H_K), 0, 1, 0, 0},
  {&__pyx_kp_u_H_Z, __pyx_k_H_Z, sizeof(__pyx_k_H_Z), 0, 1, 0, 0},
  {&__pyx_kp_u_I, __pyx_k_I, sizeof(__pyx_k_I), 0, 1, 0, 0},
  {&__pyx_n_s_IDENT, __pyx_k_IDENT, sizeof(__pyx_k_IDENT), 0, 0, 1, 1},
  {&__pyx_n_u_IDENT, __pyx_k_IDENT, sizeof(__pyx_k_IDENT), 0, 1, 0, 1},
  {&__pyx_n_s_INITIAL, __pyx_k_INITIAL, sizeof(__pyx_k_INITIAL), 0, 0, 1, 1},
  {&__pyx_n_u_INITIAL, __pyx_k_INITIAL, sizeof(__pyx_k_INITIAL), 0, 1, 0, 1},
  {&__pyx_n_s_INNER, __pyx_k_INNER, sizeof(__pyx_k_INNER), 0, 0, 1, 1},
  {&__pyx_n_u_INNER, __pyx_k_INNER, sizeof(__pyx_k_INNER), 0, 1, 0, 1},
  {&__pyx_kp_u_INVALID, __pyx_k_INVALID, sizeof(__pyx_k_INVALID), 0, 1, 0, 0},
  {&__pyx_n_s_INVALID_ALT_NUMBER, __pyx_k_INVALID_ALT_NUMBER, sizeof(__pyx_k_INVALID_ALT_NUMBER), 0, 0, 1, 1},
  {&__pyx_kp_u_I_C_V_Q, __pyx_k_I_C_V_Q, sizeof(__pyx_k_I_C_V_Q), 0, 1, 0, 0},
  {&__pyx_kp_u_I_I_I_I_I_I_I_I_J_J_J, __pyx_k_I_I_I_I_I_I_I_I_J_J_J, sizeof(__pyx_k_I_I_I_I_I_I_I_I_J_J_J), 0, 1, 0, 0},
  {&__pyx_kp_u_I_J, __pyx_k_I_J, sizeof(__pyx_k_I_J), 0, 1, 0, 0},
  {&__pyx_kp_u_I_r, __pyx_k_I_r, sizeof(__pyx_k_I_r), 0, 1, 0, 0},
  {&__pyx_n_s_If_equationContext, __pyx_k_If_equationContext, sizeof(__pyx_k_If_equationContext), 0, 0, 1, 1},
  {&__pyx_n_s_If_statementContext, __pyx_k_If_statementContext, sizeof(__pyx_k_If_statementContext), 0, 0, 1, 1},
  {&__pyx_n_s_Import_clauseContext, __pyx_k_Import_clauseContext, sizeof(__pyx_k_Import_clauseContext), 0, 0, 1, 1},
  {&__pyx_n_s_Import_listContext, __pyx_k_Import_listContext, sizeof(__pyx_k_Import_listContext), 0, 0, 1, 1},
  {&__pyx_kp_u_J, __pyx_k_J, sizeof(__pyx_k_J), 0, 1, 0, 0},
  {&__pyx_kp_u_J_2, __pyx_k_J_2, sizeof(__pyx_k_J_2), 0, 1, 0, 0},
  {&__pyx_kp_u_J_3, __pyx_k_J_3, sizeof(__pyx_k_J_3), 0, 1, 0, 0},
  {&__pyx_kp_u_J_I_6, __pyx_k_J_I_6, sizeof(__pyx_k_J_I_6), 0, 1, 0, 0},
  {&__pyx_kp_u_J_rK, __pyx_k_J_rK, sizeof(__pyx_k_J_rK), 0, 1, 0, 0},
  {&__pyx_kp_u_L, __pyx_k_L, sizeof(__pyx_k_L), 0, 1, 0, 0},
  {&__pyx_n_s_LA, __pyx_k_LA, sizeof(__pyx_k_LA), 0, 0, 1, 1},
  {&__pyx_n_s_LT, __pyx_k_LT, sizeof(__pyx_k_LT), 0, 0, 1, 1},
  {&__pyx_kp_u_L_2, __pyx_k_L_2, sizeof(__pyx_k_L_2), 0, 1, 0, 0},
  {&__pyx_kp_u_L_H, __pyx_k_L_H, sizeof(__pyx_k_L_H), 0, 1, 0, 0},
  {&__pyx_kp_u_L_N_P_R, __pyx_k_L_N_P_R, sizeof(__pyx_k_L_N_P_R), 0, 1, 0, 0},
  {&__pyx_n_s_Language_specificationContext, __pyx_k_Language_specificationContext, sizeof(__pyx_k_Language_specificationContext), 0, 0, 1, 1},
  {&__pyx_kp_u_M, __pyx_k_M, sizeof(__pyx_k_M), 0, 1, 0, 0},
  {&__pyx_kp_u_M_z, __pyx_k_M_z, sizeof(__pyx_k_M_z), 0, 1, 0, 0},
  {&__pyx_n_s_ModelicaParser, __pyx_k_ModelicaParser, sizeof(__pyx_k_ModelicaParser), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Algorithm_section, __pyx_k_ModelicaParser_Algorithm_section, sizeof(__pyx_k_ModelicaParser_Algorithm_section), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Algorithm_section_2, __pyx_k_ModelicaParser_Algorithm_section_2, sizeof(__pyx_k_ModelicaParser_Algorithm_section_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Algorithm_section_3, __pyx_k_ModelicaParser_Algorithm_section_3, sizeof(__pyx_k_ModelicaParser_Algorithm_section_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Algorithm_section_4, __pyx_k_ModelicaParser_Algorithm_section_4, sizeof(__pyx_k_ModelicaParser_Algorithm_section_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Algorithm_section_5, __pyx_k_ModelicaParser_Algorithm_section_5, sizeof(__pyx_k_ModelicaParser_Algorithm_section_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Algorithm_section_6, __pyx_k_ModelicaParser_Algorithm_section_6, sizeof(__pyx_k_ModelicaParser_Algorithm_section_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Algorithm_section_7, __pyx_k_ModelicaParser_Algorithm_section_7, sizeof(__pyx_k_ModelicaParser_Algorithm_section_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Algorithm_section_8, __pyx_k_ModelicaParser_Algorithm_section_8, sizeof(__pyx_k_ModelicaParser_Algorithm_section_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_AnnotationContext, __pyx_k_ModelicaParser_AnnotationContext, sizeof(__pyx_k_ModelicaParser_AnnotationContext), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_AnnotationContext_2, __pyx_k_ModelicaParser_AnnotationContext_2, sizeof(__pyx_k_ModelicaParser_AnnotationContext_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_AnnotationContext_3, __pyx_k_ModelicaParser_AnnotationContext_3, sizeof(__pyx_k_ModelicaParser_AnnotationContext_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_AnnotationContext_4, __pyx_k_ModelicaParser_AnnotationContext_4, sizeof(__pyx_k_ModelicaParser_AnnotationContext_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_AnnotationContext_5, __pyx_k_ModelicaParser_AnnotationContext_5, sizeof(__pyx_k_ModelicaParser_AnnotationContext_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_AnnotationContext_6, __pyx_k_ModelicaParser_AnnotationContext_6, sizeof(__pyx_k_ModelicaParser_AnnotationContext_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_AnnotationContext_7, __pyx_k_ModelicaParser_AnnotationContext_7, sizeof(__pyx_k_ModelicaParser_AnnotationContext_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_ArgumentContext, __pyx_k_ModelicaParser_ArgumentContext, sizeof(__pyx_k_ModelicaParser_ArgumentContext), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_ArgumentContext_2, __pyx_k_ModelicaParser_ArgumentContext_2, sizeof(__pyx_k_ModelicaParser_ArgumentContext_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_ArgumentContext_a, __pyx_k_ModelicaParser_ArgumentContext_a, sizeof(__pyx_k_ModelicaParser_ArgumentContext_a), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_ArgumentContext_e, __pyx_k_ModelicaParser_ArgumentContext_e, sizeof(__pyx_k_ModelicaParser_ArgumentContext_e), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_ArgumentContext_e_2, __pyx_k_ModelicaParser_ArgumentContext_e_2, sizeof(__pyx_k_ModelicaParser_ArgumentContext_e_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_ArgumentContext_e_3, __pyx_k_ModelicaParser_ArgumentContext_e_3, sizeof(__pyx_k_ModelicaParser_ArgumentContext_e_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_ArgumentContext_e_4, __pyx_k_ModelicaParser_ArgumentContext_e_4, sizeof(__pyx_k_ModelicaParser_ArgumentContext_e_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_ArgumentContext_g, __pyx_k_ModelicaParser_ArgumentContext_g, sizeof(__pyx_k_ModelicaParser_ArgumentContext_g), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Argument_expressi, __pyx_k_ModelicaParser_Argument_expressi, sizeof(__pyx_k_ModelicaParser_Argument_expressi), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Argument_expressi_2, __pyx_k_ModelicaParser_Argument_expressi_2, sizeof(__pyx_k_ModelicaParser_Argument_expressi_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Argument_expressi_3, __pyx_k_ModelicaParser_Argument_expressi_3, sizeof(__pyx_k_ModelicaParser_Argument_expressi_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Argument_expressi_4, __pyx_k_ModelicaParser_Argument_expressi_4, sizeof(__pyx_k_ModelicaParser_Argument_expressi_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Argument_expressi_5, __pyx_k_ModelicaParser_Argument_expressi_5, sizeof(__pyx_k_ModelicaParser_Argument_expressi_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Argument_expressi_6, __pyx_k_ModelicaParser_Argument_expressi_6, sizeof(__pyx_k_ModelicaParser_Argument_expressi_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Argument_function, __pyx_k_ModelicaParser_Argument_function, sizeof(__pyx_k_ModelicaParser_Argument_function), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Argument_function_2, __pyx_k_ModelicaParser_Argument_function_2, sizeof(__pyx_k_ModelicaParser_Argument_function_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Argument_function_3, __pyx_k_ModelicaParser_Argument_function_3, sizeof(__pyx_k_ModelicaParser_Argument_function_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Argument_function_4, __pyx_k_ModelicaParser_Argument_function_4, sizeof(__pyx_k_ModelicaParser_Argument_function_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Argument_function_5, __pyx_k_ModelicaParser_Argument_function_5, sizeof(__pyx_k_ModelicaParser_Argument_function_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Argument_function_6, __pyx_k_ModelicaParser_Argument_function_6, sizeof(__pyx_k_ModelicaParser_Argument_function_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Argument_function_7, __pyx_k_ModelicaParser_Argument_function_7, sizeof(__pyx_k_ModelicaParser_Argument_function_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Argument_listCont, __pyx_k_ModelicaParser_Argument_listCont, sizeof(__pyx_k_ModelicaParser_Argument_listCont), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Argument_listCont_2, __pyx_k_ModelicaParser_Argument_listCont_2, sizeof(__pyx_k_ModelicaParser_Argument_listCont_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Argument_listCont_3, __pyx_k_ModelicaParser_Argument_listCont_3, sizeof(__pyx_k_ModelicaParser_Argument_listCont_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Argument_listCont_4, __pyx_k_ModelicaParser_Argument_listCont_4, sizeof(__pyx_k_ModelicaParser_Argument_listCont_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Argument_listCont_5, __pyx_k_ModelicaParser_Argument_listCont_5, sizeof(__pyx_k_ModelicaParser_Argument_listCont_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Argument_listCont_6, __pyx_k_ModelicaParser_Argument_listCont_6, sizeof(__pyx_k_ModelicaParser_Argument_listCont_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Argument_listCont_7, __pyx_k_ModelicaParser_Argument_listCont_7, sizeof(__pyx_k_ModelicaParser_Argument_listCont_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Array_subscriptsC, __pyx_k_ModelicaParser_Array_subscriptsC, sizeof(__pyx_k_ModelicaParser_Array_subscriptsC), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Array_subscriptsC_2, __pyx_k_ModelicaParser_Array_subscriptsC_2, sizeof(__pyx_k_ModelicaParser_Array_subscriptsC_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Array_subscriptsC_3, __pyx_k_ModelicaParser_Array_subscriptsC_3, sizeof(__pyx_k_ModelicaParser_Array_subscriptsC_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Array_subscriptsC_4, __pyx_k_ModelicaParser_Array_subscriptsC_4, sizeof(__pyx_k_ModelicaParser_Array_subscriptsC_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Array_subscriptsC_5, __pyx_k_ModelicaParser_Array_subscriptsC_5, sizeof(__pyx_k_ModelicaParser_Array_subscriptsC_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Array_subscriptsC_6, __pyx_k_ModelicaParser_Array_subscriptsC_6, sizeof(__pyx_k_ModelicaParser_Array_subscriptsC_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Array_subscriptsC_7, __pyx_k_ModelicaParser_Array_subscriptsC_7, sizeof(__pyx_k_ModelicaParser_Array_subscriptsC_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Base_prefixContex, __pyx_k_ModelicaParser_Base_prefixContex, sizeof(__pyx_k_ModelicaParser_Base_prefixContex), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Base_prefixContex_2, __pyx_k_ModelicaParser_Base_prefixContex_2, sizeof(__pyx_k_ModelicaParser_Base_prefixContex_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Base_prefixContex_3, __pyx_k_ModelicaParser_Base_prefixContex_3, sizeof(__pyx_k_ModelicaParser_Base_prefixContex_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Base_prefixContex_4, __pyx_k_ModelicaParser_Base_prefixContex_4, sizeof(__pyx_k_ModelicaParser_Base_prefixContex_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Base_prefixContex_5, __pyx_k_ModelicaParser_Base_prefixContex_5, sizeof(__pyx_k_ModelicaParser_Base_prefixContex_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Base_prefixContex_6, __pyx_k_ModelicaParser_Base_prefixContex_6, sizeof(__pyx_k_ModelicaParser_Base_prefixContex_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Base_prefixContex_7, __pyx_k_ModelicaParser_Base_prefixContex_7, sizeof(__pyx_k_ModelicaParser_Base_prefixContex_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_definitionC, __pyx_k_ModelicaParser_Class_definitionC, sizeof(__pyx_k_ModelicaParser_Class_definitionC), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_definitionC_2, __pyx_k_ModelicaParser_Class_definitionC_2, sizeof(__pyx_k_ModelicaParser_Class_definitionC_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_definitionC_3, __pyx_k_ModelicaParser_Class_definitionC_3, sizeof(__pyx_k_ModelicaParser_Class_definitionC_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_definitionC_4, __pyx_k_ModelicaParser_Class_definitionC_4, sizeof(__pyx_k_ModelicaParser_Class_definitionC_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_definitionC_5, __pyx_k_ModelicaParser_Class_definitionC_5, sizeof(__pyx_k_ModelicaParser_Class_definitionC_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_definitionC_6, __pyx_k_ModelicaParser_Class_definitionC_6, sizeof(__pyx_k_ModelicaParser_Class_definitionC_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_definitionC_7, __pyx_k_ModelicaParser_Class_definitionC_7, sizeof(__pyx_k_ModelicaParser_Class_definitionC_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_definitionC_8, __pyx_k_ModelicaParser_Class_definitionC_8, sizeof(__pyx_k_ModelicaParser_Class_definitionC_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_definitionC_9, __pyx_k_ModelicaParser_Class_definitionC_9, sizeof(__pyx_k_ModelicaParser_Class_definitionC_9), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_modificatio, __pyx_k_ModelicaParser_Class_modificatio, sizeof(__pyx_k_ModelicaParser_Class_modificatio), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_modificatio_2, __pyx_k_ModelicaParser_Class_modificatio_2, sizeof(__pyx_k_ModelicaParser_Class_modificatio_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_modificatio_3, __pyx_k_ModelicaParser_Class_modificatio_3, sizeof(__pyx_k_ModelicaParser_Class_modificatio_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_modificatio_4, __pyx_k_ModelicaParser_Class_modificatio_4, sizeof(__pyx_k_ModelicaParser_Class_modificatio_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_modificatio_5, __pyx_k_ModelicaParser_Class_modificatio_5, sizeof(__pyx_k_ModelicaParser_Class_modificatio_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_modificatio_6, __pyx_k_ModelicaParser_Class_modificatio_6, sizeof(__pyx_k_ModelicaParser_Class_modificatio_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_modificatio_7, __pyx_k_ModelicaParser_Class_modificatio_7, sizeof(__pyx_k_ModelicaParser_Class_modificatio_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_prefixesCon, __pyx_k_ModelicaParser_Class_prefixesCon, sizeof(__pyx_k_ModelicaParser_Class_prefixesCon), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_prefixesCon_2, __pyx_k_ModelicaParser_Class_prefixesCon_2, sizeof(__pyx_k_ModelicaParser_Class_prefixesCon_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_prefixesCon_3, __pyx_k_ModelicaParser_Class_prefixesCon_3, sizeof(__pyx_k_ModelicaParser_Class_prefixesCon_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_prefixesCon_4, __pyx_k_ModelicaParser_Class_prefixesCon_4, sizeof(__pyx_k_ModelicaParser_Class_prefixesCon_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_prefixesCon_5, __pyx_k_ModelicaParser_Class_prefixesCon_5, sizeof(__pyx_k_ModelicaParser_Class_prefixesCon_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_prefixesCon_6, __pyx_k_ModelicaParser_Class_prefixesCon_6, sizeof(__pyx_k_ModelicaParser_Class_prefixesCon_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_prefixesCon_7, __pyx_k_ModelicaParser_Class_prefixesCon_7, sizeof(__pyx_k_ModelicaParser_Class_prefixesCon_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_prefixesCon_8, __pyx_k_ModelicaParser_Class_prefixesCon_8, sizeof(__pyx_k_ModelicaParser_Class_prefixesCon_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_baseCo, __pyx_k_ModelicaParser_Class_spec_baseCo, sizeof(__pyx_k_ModelicaParser_Class_spec_baseCo), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_baseCo_10, __pyx_k_ModelicaParser_Class_spec_baseCo_10, sizeof(__pyx_k_ModelicaParser_Class_spec_baseCo_10), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_baseCo_11, __pyx_k_ModelicaParser_Class_spec_baseCo_11, sizeof(__pyx_k_ModelicaParser_Class_spec_baseCo_11), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_baseCo_2, __pyx_k_ModelicaParser_Class_spec_baseCo_2, sizeof(__pyx_k_ModelicaParser_Class_spec_baseCo_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_baseCo_3, __pyx_k_ModelicaParser_Class_spec_baseCo_3, sizeof(__pyx_k_ModelicaParser_Class_spec_baseCo_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_baseCo_4, __pyx_k_ModelicaParser_Class_spec_baseCo_4, sizeof(__pyx_k_ModelicaParser_Class_spec_baseCo_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_baseCo_5, __pyx_k_ModelicaParser_Class_spec_baseCo_5, sizeof(__pyx_k_ModelicaParser_Class_spec_baseCo_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_baseCo_6, __pyx_k_ModelicaParser_Class_spec_baseCo_6, sizeof(__pyx_k_ModelicaParser_Class_spec_baseCo_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_baseCo_7, __pyx_k_ModelicaParser_Class_spec_baseCo_7, sizeof(__pyx_k_ModelicaParser_Class_spec_baseCo_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_baseCo_8, __pyx_k_ModelicaParser_Class_spec_baseCo_8, sizeof(__pyx_k_ModelicaParser_Class_spec_baseCo_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_baseCo_9, __pyx_k_ModelicaParser_Class_spec_baseCo_9, sizeof(__pyx_k_ModelicaParser_Class_spec_baseCo_9), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_compCo, __pyx_k_ModelicaParser_Class_spec_compCo, sizeof(__pyx_k_ModelicaParser_Class_spec_compCo), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_compCo_2, __pyx_k_ModelicaParser_Class_spec_compCo_2, sizeof(__pyx_k_ModelicaParser_Class_spec_compCo_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_compCo_3, __pyx_k_ModelicaParser_Class_spec_compCo_3, sizeof(__pyx_k_ModelicaParser_Class_spec_compCo_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_compCo_4, __pyx_k_ModelicaParser_Class_spec_compCo_4, sizeof(__pyx_k_ModelicaParser_Class_spec_compCo_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_compCo_5, __pyx_k_ModelicaParser_Class_spec_compCo_5, sizeof(__pyx_k_ModelicaParser_Class_spec_compCo_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_compCo_6, __pyx_k_ModelicaParser_Class_spec_compCo_6, sizeof(__pyx_k_ModelicaParser_Class_spec_compCo_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_compCo_7, __pyx_k_ModelicaParser_Class_spec_compCo_7, sizeof(__pyx_k_ModelicaParser_Class_spec_compCo_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_compCo_8, __pyx_k_ModelicaParser_Class_spec_compCo_8, sizeof(__pyx_k_ModelicaParser_Class_spec_compCo_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_derCon, __pyx_k_ModelicaParser_Class_spec_derCon, sizeof(__pyx_k_ModelicaParser_Class_spec_derCon), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_derCon_2, __pyx_k_ModelicaParser_Class_spec_derCon_2, sizeof(__pyx_k_ModelicaParser_Class_spec_derCon_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_derCon_3, __pyx_k_ModelicaParser_Class_spec_derCon_3, sizeof(__pyx_k_ModelicaParser_Class_spec_derCon_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_derCon_4, __pyx_k_ModelicaParser_Class_spec_derCon_4, sizeof(__pyx_k_ModelicaParser_Class_spec_derCon_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_derCon_5, __pyx_k_ModelicaParser_Class_spec_derCon_5, sizeof(__pyx_k_ModelicaParser_Class_spec_derCon_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_derCon_6, __pyx_k_ModelicaParser_Class_spec_derCon_6, sizeof(__pyx_k_ModelicaParser_Class_spec_derCon_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_derCon_7, __pyx_k_ModelicaParser_Class_spec_derCon_7, sizeof(__pyx_k_ModelicaParser_Class_spec_derCon_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_derCon_8, __pyx_k_ModelicaParser_Class_spec_derCon_8, sizeof(__pyx_k_ModelicaParser_Class_spec_derCon_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_enumCo, __pyx_k_ModelicaParser_Class_spec_enumCo, sizeof(__pyx_k_ModelicaParser_Class_spec_enumCo), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_enumCo_2, __pyx_k_ModelicaParser_Class_spec_enumCo_2, sizeof(__pyx_k_ModelicaParser_Class_spec_enumCo_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_enumCo_3, __pyx_k_ModelicaParser_Class_spec_enumCo_3, sizeof(__pyx_k_ModelicaParser_Class_spec_enumCo_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_enumCo_4, __pyx_k_ModelicaParser_Class_spec_enumCo_4, sizeof(__pyx_k_ModelicaParser_Class_spec_enumCo_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_enumCo_5, __pyx_k_ModelicaParser_Class_spec_enumCo_5, sizeof(__pyx_k_ModelicaParser_Class_spec_enumCo_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_enumCo_6, __pyx_k_ModelicaParser_Class_spec_enumCo_6, sizeof(__pyx_k_ModelicaParser_Class_spec_enumCo_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_enumCo_7, __pyx_k_ModelicaParser_Class_spec_enumCo_7, sizeof(__pyx_k_ModelicaParser_Class_spec_enumCo_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_enumCo_8, __pyx_k_ModelicaParser_Class_spec_enumCo_8, sizeof(__pyx_k_ModelicaParser_Class_spec_enumCo_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_extend, __pyx_k_ModelicaParser_Class_spec_extend, sizeof(__pyx_k_ModelicaParser_Class_spec_extend), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_extend_2, __pyx_k_ModelicaParser_Class_spec_extend_2, sizeof(__pyx_k_ModelicaParser_Class_spec_extend_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_extend_3, __pyx_k_ModelicaParser_Class_spec_extend_3, sizeof(__pyx_k_ModelicaParser_Class_spec_extend_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_extend_4, __pyx_k_ModelicaParser_Class_spec_extend_4, sizeof(__pyx_k_ModelicaParser_Class_spec_extend_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_extend_5, __pyx_k_ModelicaParser_Class_spec_extend_5, sizeof(__pyx_k_ModelicaParser_Class_spec_extend_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_extend_6, __pyx_k_ModelicaParser_Class_spec_extend_6, sizeof(__pyx_k_ModelicaParser_Class_spec_extend_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_extend_7, __pyx_k_ModelicaParser_Class_spec_extend_7, sizeof(__pyx_k_ModelicaParser_Class_spec_extend_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_extend_8, __pyx_k_ModelicaParser_Class_spec_extend_8, sizeof(__pyx_k_ModelicaParser_Class_spec_extend_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_spec_extend_9, __pyx_k_ModelicaParser_Class_spec_extend_9, sizeof(__pyx_k_ModelicaParser_Class_spec_extend_9), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_specifierCo, __pyx_k_ModelicaParser_Class_specifierCo, sizeof(__pyx_k_ModelicaParser_Class_specifierCo), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_specifierCo_2, __pyx_k_ModelicaParser_Class_specifierCo_2, sizeof(__pyx_k_ModelicaParser_Class_specifierCo_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_specifierCo_3, __pyx_k_ModelicaParser_Class_specifierCo_3, sizeof(__pyx_k_ModelicaParser_Class_specifierCo_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_specifierCo_4, __pyx_k_ModelicaParser_Class_specifierCo_4, sizeof(__pyx_k_ModelicaParser_Class_specifierCo_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_typeContext, __pyx_k_ModelicaParser_Class_typeContext, sizeof(__pyx_k_ModelicaParser_Class_typeContext), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_typeContext_2, __pyx_k_ModelicaParser_Class_typeContext_2, sizeof(__pyx_k_ModelicaParser_Class_typeContext_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_typeContext_3, __pyx_k_ModelicaParser_Class_typeContext_3, sizeof(__pyx_k_ModelicaParser_Class_typeContext_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_typeContext_4, __pyx_k_ModelicaParser_Class_typeContext_4, sizeof(__pyx_k_ModelicaParser_Class_typeContext_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_typeContext_5, __pyx_k_ModelicaParser_Class_typeContext_5, sizeof(__pyx_k_ModelicaParser_Class_typeContext_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Class_typeContext_6, __pyx_k_ModelicaParser_Class_typeContext_6, sizeof(__pyx_k_ModelicaParser_Class_typeContext_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_CommentContext, __pyx_k_ModelicaParser_CommentContext, sizeof(__pyx_k_ModelicaParser_CommentContext), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_CommentContext_2, __pyx_k_ModelicaParser_CommentContext_2, sizeof(__pyx_k_ModelicaParser_CommentContext_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_CommentContext_ac, __pyx_k_ModelicaParser_CommentContext_ac, sizeof(__pyx_k_ModelicaParser_CommentContext_ac), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_CommentContext_an, __pyx_k_ModelicaParser_CommentContext_an, sizeof(__pyx_k_ModelicaParser_CommentContext_an), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_CommentContext_en, __pyx_k_ModelicaParser_CommentContext_en, sizeof(__pyx_k_ModelicaParser_CommentContext_en), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_CommentContext_ex, __pyx_k_ModelicaParser_CommentContext_ex, sizeof(__pyx_k_ModelicaParser_CommentContext_ex), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_CommentContext_ge, __pyx_k_ModelicaParser_CommentContext_ge, sizeof(__pyx_k_ModelicaParser_CommentContext_ge), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_CommentContext_st, __pyx_k_ModelicaParser_CommentContext_st, sizeof(__pyx_k_ModelicaParser_CommentContext_st), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_clause1, __pyx_k_ModelicaParser_Component_clause1, sizeof(__pyx_k_ModelicaParser_Component_clause1), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_clause1_2, __pyx_k_ModelicaParser_Component_clause1_2, sizeof(__pyx_k_ModelicaParser_Component_clause1_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_clause1_3, __pyx_k_ModelicaParser_Component_clause1_3, sizeof(__pyx_k_ModelicaParser_Component_clause1_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_clause1_4, __pyx_k_ModelicaParser_Component_clause1_4, sizeof(__pyx_k_ModelicaParser_Component_clause1_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_clause1_5, __pyx_k_ModelicaParser_Component_clause1_5, sizeof(__pyx_k_ModelicaParser_Component_clause1_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_clause1_6, __pyx_k_ModelicaParser_Component_clause1_6, sizeof(__pyx_k_ModelicaParser_Component_clause1_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_clause1_7, __pyx_k_ModelicaParser_Component_clause1_7, sizeof(__pyx_k_ModelicaParser_Component_clause1_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_clause1_8, __pyx_k_ModelicaParser_Component_clause1_8, sizeof(__pyx_k_ModelicaParser_Component_clause1_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_clause1_9, __pyx_k_ModelicaParser_Component_clause1_9, sizeof(__pyx_k_ModelicaParser_Component_clause1_9), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_clauseC, __pyx_k_ModelicaParser_Component_clauseC, sizeof(__pyx_k_ModelicaParser_Component_clauseC), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_clauseC_10, __pyx_k_ModelicaParser_Component_clauseC_10, sizeof(__pyx_k_ModelicaParser_Component_clauseC_10), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_clauseC_2, __pyx_k_ModelicaParser_Component_clauseC_2, sizeof(__pyx_k_ModelicaParser_Component_clauseC_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_clauseC_3, __pyx_k_ModelicaParser_Component_clauseC_3, sizeof(__pyx_k_ModelicaParser_Component_clauseC_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_clauseC_4, __pyx_k_ModelicaParser_Component_clauseC_4, sizeof(__pyx_k_ModelicaParser_Component_clauseC_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_clauseC_5, __pyx_k_ModelicaParser_Component_clauseC_5, sizeof(__pyx_k_ModelicaParser_Component_clauseC_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_clauseC_6, __pyx_k_ModelicaParser_Component_clauseC_6, sizeof(__pyx_k_ModelicaParser_Component_clauseC_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_clauseC_7, __pyx_k_ModelicaParser_Component_clauseC_7, sizeof(__pyx_k_ModelicaParser_Component_clauseC_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_clauseC_8, __pyx_k_ModelicaParser_Component_clauseC_8, sizeof(__pyx_k_ModelicaParser_Component_clauseC_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_clauseC_9, __pyx_k_ModelicaParser_Component_clauseC_9, sizeof(__pyx_k_ModelicaParser_Component_clauseC_9), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_declara, __pyx_k_ModelicaParser_Component_declara, sizeof(__pyx_k_ModelicaParser_Component_declara), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_declara_10, __pyx_k_ModelicaParser_Component_declara_10, sizeof(__pyx_k_ModelicaParser_Component_declara_10), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_declara_11, __pyx_k_ModelicaParser_Component_declara_11, sizeof(__pyx_k_ModelicaParser_Component_declara_11), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_declara_12, __pyx_k_ModelicaParser_Component_declara_12, sizeof(__pyx_k_ModelicaParser_Component_declara_12), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_declara_13, __pyx_k_ModelicaParser_Component_declara_13, sizeof(__pyx_k_ModelicaParser_Component_declara_13), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_declara_14, __pyx_k_ModelicaParser_Component_declara_14, sizeof(__pyx_k_ModelicaParser_Component_declara_14), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_declara_15, __pyx_k_ModelicaParser_Component_declara_15, sizeof(__pyx_k_ModelicaParser_Component_declara_15), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_declara_16, __pyx_k_ModelicaParser_Component_declara_16, sizeof(__pyx_k_ModelicaParser_Component_declara_16), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_declara_17, __pyx_k_ModelicaParser_Component_declara_17, sizeof(__pyx_k_ModelicaParser_Component_declara_17), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_declara_2, __pyx_k_ModelicaParser_Component_declara_2, sizeof(__pyx_k_ModelicaParser_Component_declara_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_declara_3, __pyx_k_ModelicaParser_Component_declara_3, sizeof(__pyx_k_ModelicaParser_Component_declara_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_declara_4, __pyx_k_ModelicaParser_Component_declara_4, sizeof(__pyx_k_ModelicaParser_Component_declara_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_declara_5, __pyx_k_ModelicaParser_Component_declara_5, sizeof(__pyx_k_ModelicaParser_Component_declara_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_declara_6, __pyx_k_ModelicaParser_Component_declara_6, sizeof(__pyx_k_ModelicaParser_Component_declara_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_declara_7, __pyx_k_ModelicaParser_Component_declara_7, sizeof(__pyx_k_ModelicaParser_Component_declara_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_declara_8, __pyx_k_ModelicaParser_Component_declara_8, sizeof(__pyx_k_ModelicaParser_Component_declara_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_declara_9, __pyx_k_ModelicaParser_Component_declara_9, sizeof(__pyx_k_ModelicaParser_Component_declara_9), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_listCon, __pyx_k_ModelicaParser_Component_listCon, sizeof(__pyx_k_ModelicaParser_Component_listCon), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_listCon_2, __pyx_k_ModelicaParser_Component_listCon_2, sizeof(__pyx_k_ModelicaParser_Component_listCon_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_listCon_3, __pyx_k_ModelicaParser_Component_listCon_3, sizeof(__pyx_k_ModelicaParser_Component_listCon_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_listCon_4, __pyx_k_ModelicaParser_Component_listCon_4, sizeof(__pyx_k_ModelicaParser_Component_listCon_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_listCon_5, __pyx_k_ModelicaParser_Component_listCon_5, sizeof(__pyx_k_ModelicaParser_Component_listCon_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_listCon_6, __pyx_k_ModelicaParser_Component_listCon_6, sizeof(__pyx_k_ModelicaParser_Component_listCon_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_listCon_7, __pyx_k_ModelicaParser_Component_listCon_7, sizeof(__pyx_k_ModelicaParser_Component_listCon_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_referen, __pyx_k_ModelicaParser_Component_referen, sizeof(__pyx_k_ModelicaParser_Component_referen), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_referen_10, __pyx_k_ModelicaParser_Component_referen_10, sizeof(__pyx_k_ModelicaParser_Component_referen_10), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_referen_11, __pyx_k_ModelicaParser_Component_referen_11, sizeof(__pyx_k_ModelicaParser_Component_referen_11), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_referen_12, __pyx_k_ModelicaParser_Component_referen_12, sizeof(__pyx_k_ModelicaParser_Component_referen_12), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_referen_13, __pyx_k_ModelicaParser_Component_referen_13, sizeof(__pyx_k_ModelicaParser_Component_referen_13), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_referen_14, __pyx_k_ModelicaParser_Component_referen_14, sizeof(__pyx_k_ModelicaParser_Component_referen_14), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_referen_15, __pyx_k_ModelicaParser_Component_referen_15, sizeof(__pyx_k_ModelicaParser_Component_referen_15), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_referen_2, __pyx_k_ModelicaParser_Component_referen_2, sizeof(__pyx_k_ModelicaParser_Component_referen_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_referen_3, __pyx_k_ModelicaParser_Component_referen_3, sizeof(__pyx_k_ModelicaParser_Component_referen_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_referen_4, __pyx_k_ModelicaParser_Component_referen_4, sizeof(__pyx_k_ModelicaParser_Component_referen_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_referen_5, __pyx_k_ModelicaParser_Component_referen_5, sizeof(__pyx_k_ModelicaParser_Component_referen_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_referen_6, __pyx_k_ModelicaParser_Component_referen_6, sizeof(__pyx_k_ModelicaParser_Component_referen_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_referen_7, __pyx_k_ModelicaParser_Component_referen_7, sizeof(__pyx_k_ModelicaParser_Component_referen_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_referen_8, __pyx_k_ModelicaParser_Component_referen_8, sizeof(__pyx_k_ModelicaParser_Component_referen_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Component_referen_9, __pyx_k_ModelicaParser_Component_referen_9, sizeof(__pyx_k_ModelicaParser_Component_referen_9), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_CompositionContex, __pyx_k_ModelicaParser_CompositionContex, sizeof(__pyx_k_ModelicaParser_CompositionContex), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_CompositionContex_10, __pyx_k_ModelicaParser_CompositionContex_10, sizeof(__pyx_k_ModelicaParser_CompositionContex_10), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_CompositionContex_11, __pyx_k_ModelicaParser_CompositionContex_11, sizeof(__pyx_k_ModelicaParser_CompositionContex_11), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_CompositionContex_12, __pyx_k_ModelicaParser_CompositionContex_12, sizeof(__pyx_k_ModelicaParser_CompositionContex_12), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_CompositionContex_2, __pyx_k_ModelicaParser_CompositionContex_2, sizeof(__pyx_k_ModelicaParser_CompositionContex_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_CompositionContex_3, __pyx_k_ModelicaParser_CompositionContex_3, sizeof(__pyx_k_ModelicaParser_CompositionContex_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_CompositionContex_4, __pyx_k_ModelicaParser_CompositionContex_4, sizeof(__pyx_k_ModelicaParser_CompositionContex_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_CompositionContex_5, __pyx_k_ModelicaParser_CompositionContex_5, sizeof(__pyx_k_ModelicaParser_CompositionContex_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_CompositionContex_6, __pyx_k_ModelicaParser_CompositionContex_6, sizeof(__pyx_k_ModelicaParser_CompositionContex_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_CompositionContex_7, __pyx_k_ModelicaParser_CompositionContex_7, sizeof(__pyx_k_ModelicaParser_CompositionContex_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_CompositionContex_8, __pyx_k_ModelicaParser_CompositionContex_8, sizeof(__pyx_k_ModelicaParser_CompositionContex_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_CompositionContex_9, __pyx_k_ModelicaParser_CompositionContex_9, sizeof(__pyx_k_ModelicaParser_CompositionContex_9), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Condition_attribu, __pyx_k_ModelicaParser_Condition_attribu, sizeof(__pyx_k_ModelicaParser_Condition_attribu), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Condition_attribu_2, __pyx_k_ModelicaParser_Condition_attribu_2, sizeof(__pyx_k_ModelicaParser_Condition_attribu_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Condition_attribu_3, __pyx_k_ModelicaParser_Condition_attribu_3, sizeof(__pyx_k_ModelicaParser_Condition_attribu_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Condition_attribu_4, __pyx_k_ModelicaParser_Condition_attribu_4, sizeof(__pyx_k_ModelicaParser_Condition_attribu_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Condition_attribu_5, __pyx_k_ModelicaParser_Condition_attribu_5, sizeof(__pyx_k_ModelicaParser_Condition_attribu_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Condition_attribu_6, __pyx_k_ModelicaParser_Condition_attribu_6, sizeof(__pyx_k_ModelicaParser_Condition_attribu_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Condition_attribu_7, __pyx_k_ModelicaParser_Condition_attribu_7, sizeof(__pyx_k_ModelicaParser_Condition_attribu_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Connect_clauseCon, __pyx_k_ModelicaParser_Connect_clauseCon, sizeof(__pyx_k_ModelicaParser_Connect_clauseCon), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Connect_clauseCon_2, __pyx_k_ModelicaParser_Connect_clauseCon_2, sizeof(__pyx_k_ModelicaParser_Connect_clauseCon_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Connect_clauseCon_3, __pyx_k_ModelicaParser_Connect_clauseCon_3, sizeof(__pyx_k_ModelicaParser_Connect_clauseCon_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Connect_clauseCon_4, __pyx_k_ModelicaParser_Connect_clauseCon_4, sizeof(__pyx_k_ModelicaParser_Connect_clauseCon_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Connect_clauseCon_5, __pyx_k_ModelicaParser_Connect_clauseCon_5, sizeof(__pyx_k_ModelicaParser_Connect_clauseCon_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Connect_clauseCon_6, __pyx_k_ModelicaParser_Connect_clauseCon_6, sizeof(__pyx_k_ModelicaParser_Connect_clauseCon_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Connect_clauseCon_7, __pyx_k_ModelicaParser_Connect_clauseCon_7, sizeof(__pyx_k_ModelicaParser_Connect_clauseCon_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Constraining_clau, __pyx_k_ModelicaParser_Constraining_clau, sizeof(__pyx_k_ModelicaParser_Constraining_clau), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Constraining_clau_2, __pyx_k_ModelicaParser_Constraining_clau_2, sizeof(__pyx_k_ModelicaParser_Constraining_clau_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Constraining_clau_3, __pyx_k_ModelicaParser_Constraining_clau_3, sizeof(__pyx_k_ModelicaParser_Constraining_clau_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Constraining_clau_4, __pyx_k_ModelicaParser_Constraining_clau_4, sizeof(__pyx_k_ModelicaParser_Constraining_clau_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Constraining_clau_5, __pyx_k_ModelicaParser_Constraining_clau_5, sizeof(__pyx_k_ModelicaParser_Constraining_clau_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Constraining_clau_6, __pyx_k_ModelicaParser_Constraining_clau_6, sizeof(__pyx_k_ModelicaParser_Constraining_clau_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Constraining_clau_7, __pyx_k_ModelicaParser_Constraining_clau_7, sizeof(__pyx_k_ModelicaParser_Constraining_clau_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Constraining_clau_8, __pyx_k_ModelicaParser_Constraining_clau_8, sizeof(__pyx_k_ModelicaParser_Constraining_clau_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_DeclarationContex, __pyx_k_ModelicaParser_DeclarationContex, sizeof(__pyx_k_ModelicaParser_DeclarationContex), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_DeclarationContex_2, __pyx_k_ModelicaParser_DeclarationContex_2, sizeof(__pyx_k_ModelicaParser_DeclarationContex_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_DeclarationContex_3, __pyx_k_ModelicaParser_DeclarationContex_3, sizeof(__pyx_k_ModelicaParser_DeclarationContex_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_DeclarationContex_4, __pyx_k_ModelicaParser_DeclarationContex_4, sizeof(__pyx_k_ModelicaParser_DeclarationContex_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_DeclarationContex_5, __pyx_k_ModelicaParser_DeclarationContex_5, sizeof(__pyx_k_ModelicaParser_DeclarationContex_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_DeclarationContex_6, __pyx_k_ModelicaParser_DeclarationContex_6, sizeof(__pyx_k_ModelicaParser_DeclarationContex_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_DeclarationContex_7, __pyx_k_ModelicaParser_DeclarationContex_7, sizeof(__pyx_k_ModelicaParser_DeclarationContex_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_DeclarationContex_8, __pyx_k_ModelicaParser_DeclarationContex_8, sizeof(__pyx_k_ModelicaParser_DeclarationContex_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_DeclarationContex_9, __pyx_k_ModelicaParser_DeclarationContex_9, sizeof(__pyx_k_ModelicaParser_DeclarationContex_9), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_ElementContext, __pyx_k_ModelicaParser_ElementContext, sizeof(__pyx_k_ModelicaParser_ElementContext), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_ElementContext_2, __pyx_k_ModelicaParser_ElementContext_2, sizeof(__pyx_k_ModelicaParser_ElementContext_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_ElementContext_ac, __pyx_k_ModelicaParser_ElementContext_ac, sizeof(__pyx_k_ModelicaParser_ElementContext_ac), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_ElementContext_en, __pyx_k_ModelicaParser_ElementContext_en, sizeof(__pyx_k_ModelicaParser_ElementContext_en), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_ElementContext_ex, __pyx_k_ModelicaParser_ElementContext_ex, sizeof(__pyx_k_ModelicaParser_ElementContext_ex), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_ElementContext_ex_2, __pyx_k_ModelicaParser_ElementContext_ex_2, sizeof(__pyx_k_ModelicaParser_ElementContext_ex_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_ElementContext_ge, __pyx_k_ModelicaParser_ElementContext_ge, sizeof(__pyx_k_ModelicaParser_ElementContext_ge), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_ElementContext_im, __pyx_k_ModelicaParser_ElementContext_im, sizeof(__pyx_k_ModelicaParser_ElementContext_im), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_ElementContext_re, __pyx_k_ModelicaParser_ElementContext_re, sizeof(__pyx_k_ModelicaParser_ElementContext_re), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_ElementContext_re_2, __pyx_k_ModelicaParser_ElementContext_re_2, sizeof(__pyx_k_ModelicaParser_ElementContext_re_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_listConte, __pyx_k_ModelicaParser_Element_listConte, sizeof(__pyx_k_ModelicaParser_Element_listConte), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_listConte_2, __pyx_k_ModelicaParser_Element_listConte_2, sizeof(__pyx_k_ModelicaParser_Element_listConte_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_listConte_3, __pyx_k_ModelicaParser_Element_listConte_3, sizeof(__pyx_k_ModelicaParser_Element_listConte_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_listConte_4, __pyx_k_ModelicaParser_Element_listConte_4, sizeof(__pyx_k_ModelicaParser_Element_listConte_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_listConte_5, __pyx_k_ModelicaParser_Element_listConte_5, sizeof(__pyx_k_ModelicaParser_Element_listConte_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_listConte_6, __pyx_k_ModelicaParser_Element_listConte_6, sizeof(__pyx_k_ModelicaParser_Element_listConte_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_listConte_7, __pyx_k_ModelicaParser_Element_listConte_7, sizeof(__pyx_k_ModelicaParser_Element_listConte_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_modificat, __pyx_k_ModelicaParser_Element_modificat, sizeof(__pyx_k_ModelicaParser_Element_modificat), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_modificat_10, __pyx_k_ModelicaParser_Element_modificat_10, sizeof(__pyx_k_ModelicaParser_Element_modificat_10), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_modificat_11, __pyx_k_ModelicaParser_Element_modificat_11, sizeof(__pyx_k_ModelicaParser_Element_modificat_11), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_modificat_12, __pyx_k_ModelicaParser_Element_modificat_12, sizeof(__pyx_k_ModelicaParser_Element_modificat_12), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_modificat_13, __pyx_k_ModelicaParser_Element_modificat_13, sizeof(__pyx_k_ModelicaParser_Element_modificat_13), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_modificat_14, __pyx_k_ModelicaParser_Element_modificat_14, sizeof(__pyx_k_ModelicaParser_Element_modificat_14), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_modificat_15, __pyx_k_ModelicaParser_Element_modificat_15, sizeof(__pyx_k_ModelicaParser_Element_modificat_15), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_modificat_16, __pyx_k_ModelicaParser_Element_modificat_16, sizeof(__pyx_k_ModelicaParser_Element_modificat_16), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_modificat_17, __pyx_k_ModelicaParser_Element_modificat_17, sizeof(__pyx_k_ModelicaParser_Element_modificat_17), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_modificat_18, __pyx_k_ModelicaParser_Element_modificat_18, sizeof(__pyx_k_ModelicaParser_Element_modificat_18), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_modificat_19, __pyx_k_ModelicaParser_Element_modificat_19, sizeof(__pyx_k_ModelicaParser_Element_modificat_19), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_modificat_2, __pyx_k_ModelicaParser_Element_modificat_2, sizeof(__pyx_k_ModelicaParser_Element_modificat_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_modificat_3, __pyx_k_ModelicaParser_Element_modificat_3, sizeof(__pyx_k_ModelicaParser_Element_modificat_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_modificat_4, __pyx_k_ModelicaParser_Element_modificat_4, sizeof(__pyx_k_ModelicaParser_Element_modificat_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_modificat_5, __pyx_k_ModelicaParser_Element_modificat_5, sizeof(__pyx_k_ModelicaParser_Element_modificat_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_modificat_6, __pyx_k_ModelicaParser_Element_modificat_6, sizeof(__pyx_k_ModelicaParser_Element_modificat_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_modificat_7, __pyx_k_ModelicaParser_Element_modificat_7, sizeof(__pyx_k_ModelicaParser_Element_modificat_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_modificat_8, __pyx_k_ModelicaParser_Element_modificat_8, sizeof(__pyx_k_ModelicaParser_Element_modificat_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_modificat_9, __pyx_k_ModelicaParser_Element_modificat_9, sizeof(__pyx_k_ModelicaParser_Element_modificat_9), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_redeclara, __pyx_k_ModelicaParser_Element_redeclara, sizeof(__pyx_k_ModelicaParser_Element_redeclara), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_redeclara_10, __pyx_k_ModelicaParser_Element_redeclara_10, sizeof(__pyx_k_ModelicaParser_Element_redeclara_10), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_redeclara_11, __pyx_k_ModelicaParser_Element_redeclara_11, sizeof(__pyx_k_ModelicaParser_Element_redeclara_11), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_redeclara_12, __pyx_k_ModelicaParser_Element_redeclara_12, sizeof(__pyx_k_ModelicaParser_Element_redeclara_12), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_redeclara_2, __pyx_k_ModelicaParser_Element_redeclara_2, sizeof(__pyx_k_ModelicaParser_Element_redeclara_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_redeclara_3, __pyx_k_ModelicaParser_Element_redeclara_3, sizeof(__pyx_k_ModelicaParser_Element_redeclara_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_redeclara_4, __pyx_k_ModelicaParser_Element_redeclara_4, sizeof(__pyx_k_ModelicaParser_Element_redeclara_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_redeclara_5, __pyx_k_ModelicaParser_Element_redeclara_5, sizeof(__pyx_k_ModelicaParser_Element_redeclara_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_redeclara_6, __pyx_k_ModelicaParser_Element_redeclara_6, sizeof(__pyx_k_ModelicaParser_Element_redeclara_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_redeclara_7, __pyx_k_ModelicaParser_Element_redeclara_7, sizeof(__pyx_k_ModelicaParser_Element_redeclara_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_redeclara_8, __pyx_k_ModelicaParser_Element_redeclara_8, sizeof(__pyx_k_ModelicaParser_Element_redeclara_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_redeclara_9, __pyx_k_ModelicaParser_Element_redeclara_9, sizeof(__pyx_k_ModelicaParser_Element_redeclara_9), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_replaceab, __pyx_k_ModelicaParser_Element_replaceab, sizeof(__pyx_k_ModelicaParser_Element_replaceab), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_replaceab_2, __pyx_k_ModelicaParser_Element_replaceab_2, sizeof(__pyx_k_ModelicaParser_Element_replaceab_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_replaceab_3, __pyx_k_ModelicaParser_Element_replaceab_3, sizeof(__pyx_k_ModelicaParser_Element_replaceab_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_replaceab_4, __pyx_k_ModelicaParser_Element_replaceab_4, sizeof(__pyx_k_ModelicaParser_Element_replaceab_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_replaceab_5, __pyx_k_ModelicaParser_Element_replaceab_5, sizeof(__pyx_k_ModelicaParser_Element_replaceab_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_replaceab_6, __pyx_k_ModelicaParser_Element_replaceab_6, sizeof(__pyx_k_ModelicaParser_Element_replaceab_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_replaceab_7, __pyx_k_ModelicaParser_Element_replaceab_7, sizeof(__pyx_k_ModelicaParser_Element_replaceab_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_replaceab_8, __pyx_k_ModelicaParser_Element_replaceab_8, sizeof(__pyx_k_ModelicaParser_Element_replaceab_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Element_replaceab_9, __pyx_k_ModelicaParser_Element_replaceab_9, sizeof(__pyx_k_ModelicaParser_Element_replaceab_9), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Enum_listContext, __pyx_k_ModelicaParser_Enum_listContext, sizeof(__pyx_k_ModelicaParser_Enum_listContext), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Enum_listContext_2, __pyx_k_ModelicaParser_Enum_listContext_2, sizeof(__pyx_k_ModelicaParser_Enum_listContext_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Enum_listContext_3, __pyx_k_ModelicaParser_Enum_listContext_3, sizeof(__pyx_k_ModelicaParser_Enum_listContext_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Enum_listContext_4, __pyx_k_ModelicaParser_Enum_listContext_4, sizeof(__pyx_k_ModelicaParser_Enum_listContext_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Enum_listContext_5, __pyx_k_ModelicaParser_Enum_listContext_5, sizeof(__pyx_k_ModelicaParser_Enum_listContext_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Enum_listContext_6, __pyx_k_ModelicaParser_Enum_listContext_6, sizeof(__pyx_k_ModelicaParser_Enum_listContext_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Enum_listContext_7, __pyx_k_ModelicaParser_Enum_listContext_7, sizeof(__pyx_k_ModelicaParser_Enum_listContext_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Enumeration_liter, __pyx_k_ModelicaParser_Enumeration_liter, sizeof(__pyx_k_ModelicaParser_Enumeration_liter), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Enumeration_liter_2, __pyx_k_ModelicaParser_Enumeration_liter_2, sizeof(__pyx_k_ModelicaParser_Enumeration_liter_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Enumeration_liter_3, __pyx_k_ModelicaParser_Enumeration_liter_3, sizeof(__pyx_k_ModelicaParser_Enumeration_liter_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Enumeration_liter_4, __pyx_k_ModelicaParser_Enumeration_liter_4, sizeof(__pyx_k_ModelicaParser_Enumeration_liter_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Enumeration_liter_5, __pyx_k_ModelicaParser_Enumeration_liter_5, sizeof(__pyx_k_ModelicaParser_Enumeration_liter_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Enumeration_liter_6, __pyx_k_ModelicaParser_Enumeration_liter_6, sizeof(__pyx_k_ModelicaParser_Enumeration_liter_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Enumeration_liter_7, __pyx_k_ModelicaParser_Enumeration_liter_7, sizeof(__pyx_k_ModelicaParser_Enumeration_liter_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Enumeration_liter_8, __pyx_k_ModelicaParser_Enumeration_liter_8, sizeof(__pyx_k_ModelicaParser_Enumeration_liter_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_EquationContext, __pyx_k_ModelicaParser_EquationContext, sizeof(__pyx_k_ModelicaParser_EquationContext), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_EquationContext_2, __pyx_k_ModelicaParser_EquationContext_2, sizeof(__pyx_k_ModelicaParser_EquationContext_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_EquationContext_a, __pyx_k_ModelicaParser_EquationContext_a, sizeof(__pyx_k_ModelicaParser_EquationContext_a), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_EquationContext_c, __pyx_k_ModelicaParser_EquationContext_c, sizeof(__pyx_k_ModelicaParser_EquationContext_c), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_EquationContext_e, __pyx_k_ModelicaParser_EquationContext_e, sizeof(__pyx_k_ModelicaParser_EquationContext_e), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_EquationContext_e_2, __pyx_k_ModelicaParser_EquationContext_e_2, sizeof(__pyx_k_ModelicaParser_EquationContext_e_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_EquationContext_e_3, __pyx_k_ModelicaParser_EquationContext_e_3, sizeof(__pyx_k_ModelicaParser_EquationContext_e_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_EquationContext_g, __pyx_k_ModelicaParser_EquationContext_g, sizeof(__pyx_k_ModelicaParser_EquationContext_g), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_connect, __pyx_k_ModelicaParser_Equation_connect, sizeof(__pyx_k_ModelicaParser_Equation_connect), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_connect_2, __pyx_k_ModelicaParser_Equation_connect_2, sizeof(__pyx_k_ModelicaParser_Equation_connect_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_connect_3, __pyx_k_ModelicaParser_Equation_connect_3, sizeof(__pyx_k_ModelicaParser_Equation_connect_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_connect_4, __pyx_k_ModelicaParser_Equation_connect_4, sizeof(__pyx_k_ModelicaParser_Equation_connect_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_connect_5, __pyx_k_ModelicaParser_Equation_connect_5, sizeof(__pyx_k_ModelicaParser_Equation_connect_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_connect_6, __pyx_k_ModelicaParser_Equation_connect_6, sizeof(__pyx_k_ModelicaParser_Equation_connect_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_forConte, __pyx_k_ModelicaParser_Equation_forConte, sizeof(__pyx_k_ModelicaParser_Equation_forConte), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_forConte_2, __pyx_k_ModelicaParser_Equation_forConte_2, sizeof(__pyx_k_ModelicaParser_Equation_forConte_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_forConte_3, __pyx_k_ModelicaParser_Equation_forConte_3, sizeof(__pyx_k_ModelicaParser_Equation_forConte_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_forConte_4, __pyx_k_ModelicaParser_Equation_forConte_4, sizeof(__pyx_k_ModelicaParser_Equation_forConte_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_forConte_5, __pyx_k_ModelicaParser_Equation_forConte_5, sizeof(__pyx_k_ModelicaParser_Equation_forConte_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_forConte_6, __pyx_k_ModelicaParser_Equation_forConte_6, sizeof(__pyx_k_ModelicaParser_Equation_forConte_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_function, __pyx_k_ModelicaParser_Equation_function, sizeof(__pyx_k_ModelicaParser_Equation_function), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_function_2, __pyx_k_ModelicaParser_Equation_function_2, sizeof(__pyx_k_ModelicaParser_Equation_function_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_function_3, __pyx_k_ModelicaParser_Equation_function_3, sizeof(__pyx_k_ModelicaParser_Equation_function_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_function_4, __pyx_k_ModelicaParser_Equation_function_4, sizeof(__pyx_k_ModelicaParser_Equation_function_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_function_5, __pyx_k_ModelicaParser_Equation_function_5, sizeof(__pyx_k_ModelicaParser_Equation_function_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_function_6, __pyx_k_ModelicaParser_Equation_function_6, sizeof(__pyx_k_ModelicaParser_Equation_function_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_function_7, __pyx_k_ModelicaParser_Equation_function_7, sizeof(__pyx_k_ModelicaParser_Equation_function_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_ifContex, __pyx_k_ModelicaParser_Equation_ifContex, sizeof(__pyx_k_ModelicaParser_Equation_ifContex), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_ifContex_2, __pyx_k_ModelicaParser_Equation_ifContex_2, sizeof(__pyx_k_ModelicaParser_Equation_ifContex_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_ifContex_3, __pyx_k_ModelicaParser_Equation_ifContex_3, sizeof(__pyx_k_ModelicaParser_Equation_ifContex_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_ifContex_4, __pyx_k_ModelicaParser_Equation_ifContex_4, sizeof(__pyx_k_ModelicaParser_Equation_ifContex_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_ifContex_5, __pyx_k_ModelicaParser_Equation_ifContex_5, sizeof(__pyx_k_ModelicaParser_Equation_ifContex_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_ifContex_6, __pyx_k_ModelicaParser_Equation_ifContex_6, sizeof(__pyx_k_ModelicaParser_Equation_ifContex_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_optionsC, __pyx_k_ModelicaParser_Equation_optionsC, sizeof(__pyx_k_ModelicaParser_Equation_optionsC), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_optionsC_2, __pyx_k_ModelicaParser_Equation_optionsC_2, sizeof(__pyx_k_ModelicaParser_Equation_optionsC_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_optionsC_3, __pyx_k_ModelicaParser_Equation_optionsC_3, sizeof(__pyx_k_ModelicaParser_Equation_optionsC_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_optionsC_4, __pyx_k_ModelicaParser_Equation_optionsC_4, sizeof(__pyx_k_ModelicaParser_Equation_optionsC_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_sectionC, __pyx_k_ModelicaParser_Equation_sectionC, sizeof(__pyx_k_ModelicaParser_Equation_sectionC), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_sectionC_2, __pyx_k_ModelicaParser_Equation_sectionC_2, sizeof(__pyx_k_ModelicaParser_Equation_sectionC_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_sectionC_3, __pyx_k_ModelicaParser_Equation_sectionC_3, sizeof(__pyx_k_ModelicaParser_Equation_sectionC_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_sectionC_4, __pyx_k_ModelicaParser_Equation_sectionC_4, sizeof(__pyx_k_ModelicaParser_Equation_sectionC_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_sectionC_5, __pyx_k_ModelicaParser_Equation_sectionC_5, sizeof(__pyx_k_ModelicaParser_Equation_sectionC_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_sectionC_6, __pyx_k_ModelicaParser_Equation_sectionC_6, sizeof(__pyx_k_ModelicaParser_Equation_sectionC_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_sectionC_7, __pyx_k_ModelicaParser_Equation_sectionC_7, sizeof(__pyx_k_ModelicaParser_Equation_sectionC_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_sectionC_8, __pyx_k_ModelicaParser_Equation_sectionC_8, sizeof(__pyx_k_ModelicaParser_Equation_sectionC_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_simpleCo, __pyx_k_ModelicaParser_Equation_simpleCo, sizeof(__pyx_k_ModelicaParser_Equation_simpleCo), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_simpleCo_2, __pyx_k_ModelicaParser_Equation_simpleCo_2, sizeof(__pyx_k_ModelicaParser_Equation_simpleCo_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_simpleCo_3, __pyx_k_ModelicaParser_Equation_simpleCo_3, sizeof(__pyx_k_ModelicaParser_Equation_simpleCo_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_simpleCo_4, __pyx_k_ModelicaParser_Equation_simpleCo_4, sizeof(__pyx_k_ModelicaParser_Equation_simpleCo_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_simpleCo_5, __pyx_k_ModelicaParser_Equation_simpleCo_5, sizeof(__pyx_k_ModelicaParser_Equation_simpleCo_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_simpleCo_6, __pyx_k_ModelicaParser_Equation_simpleCo_6, sizeof(__pyx_k_ModelicaParser_Equation_simpleCo_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_simpleCo_7, __pyx_k_ModelicaParser_Equation_simpleCo_7, sizeof(__pyx_k_ModelicaParser_Equation_simpleCo_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_whenCont, __pyx_k_ModelicaParser_Equation_whenCont, sizeof(__pyx_k_ModelicaParser_Equation_whenCont), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_whenCont_2, __pyx_k_ModelicaParser_Equation_whenCont_2, sizeof(__pyx_k_ModelicaParser_Equation_whenCont_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_whenCont_3, __pyx_k_ModelicaParser_Equation_whenCont_3, sizeof(__pyx_k_ModelicaParser_Equation_whenCont_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_whenCont_4, __pyx_k_ModelicaParser_Equation_whenCont_4, sizeof(__pyx_k_ModelicaParser_Equation_whenCont_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_whenCont_5, __pyx_k_ModelicaParser_Equation_whenCont_5, sizeof(__pyx_k_ModelicaParser_Equation_whenCont_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Equation_whenCont_6, __pyx_k_ModelicaParser_Equation_whenCont_6, sizeof(__pyx_k_ModelicaParser_Equation_whenCont_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_ExprContext, __pyx_k_ModelicaParser_ExprContext, sizeof(__pyx_k_ModelicaParser_ExprContext), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_ExprContext___ini, __pyx_k_ModelicaParser_ExprContext___ini, sizeof(__pyx_k_ModelicaParser_ExprContext___ini), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_ExprContext_copyF, __pyx_k_ModelicaParser_ExprContext_copyF, sizeof(__pyx_k_ModelicaParser_ExprContext_copyF), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_ExprContext_getRu, __pyx_k_ModelicaParser_ExprContext_getRu, sizeof(__pyx_k_ModelicaParser_ExprContext_getRu), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_addContext, __pyx_k_ModelicaParser_Expr_addContext, sizeof(__pyx_k_ModelicaParser_Expr_addContext), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_addContext_2, __pyx_k_ModelicaParser_Expr_addContext_2, sizeof(__pyx_k_ModelicaParser_Expr_addContext_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_addContext_a, __pyx_k_ModelicaParser_Expr_addContext_a, sizeof(__pyx_k_ModelicaParser_Expr_addContext_a), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_addContext_e, __pyx_k_ModelicaParser_Expr_addContext_e, sizeof(__pyx_k_ModelicaParser_Expr_addContext_e), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_addContext_e_2, __pyx_k_ModelicaParser_Expr_addContext_e_2, sizeof(__pyx_k_ModelicaParser_Expr_addContext_e_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_addContext_e_3, __pyx_k_ModelicaParser_Expr_addContext_e_3, sizeof(__pyx_k_ModelicaParser_Expr_addContext_e_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_andContext, __pyx_k_ModelicaParser_Expr_andContext, sizeof(__pyx_k_ModelicaParser_Expr_andContext), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_andContext_2, __pyx_k_ModelicaParser_Expr_andContext_2, sizeof(__pyx_k_ModelicaParser_Expr_andContext_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_andContext_a, __pyx_k_ModelicaParser_Expr_andContext_a, sizeof(__pyx_k_ModelicaParser_Expr_andContext_a), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_andContext_e, __pyx_k_ModelicaParser_Expr_andContext_e, sizeof(__pyx_k_ModelicaParser_Expr_andContext_e), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_andContext_e_2, __pyx_k_ModelicaParser_Expr_andContext_e_2, sizeof(__pyx_k_ModelicaParser_Expr_andContext_e_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_andContext_e_3, __pyx_k_ModelicaParser_Expr_andContext_e_3, sizeof(__pyx_k_ModelicaParser_Expr_andContext_e_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_expContext, __pyx_k_ModelicaParser_Expr_expContext, sizeof(__pyx_k_ModelicaParser_Expr_expContext), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_expContext_2, __pyx_k_ModelicaParser_Expr_expContext_2, sizeof(__pyx_k_ModelicaParser_Expr_expContext_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_expContext_a, __pyx_k_ModelicaParser_Expr_expContext_a, sizeof(__pyx_k_ModelicaParser_Expr_expContext_a), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_expContext_e, __pyx_k_ModelicaParser_Expr_expContext_e, sizeof(__pyx_k_ModelicaParser_Expr_expContext_e), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_expContext_e_2, __pyx_k_ModelicaParser_Expr_expContext_e_2, sizeof(__pyx_k_ModelicaParser_Expr_expContext_e_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_expContext_p, __pyx_k_ModelicaParser_Expr_expContext_p, sizeof(__pyx_k_ModelicaParser_Expr_expContext_p), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_mulContext, __pyx_k_ModelicaParser_Expr_mulContext, sizeof(__pyx_k_ModelicaParser_Expr_mulContext), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_mulContext_2, __pyx_k_ModelicaParser_Expr_mulContext_2, sizeof(__pyx_k_ModelicaParser_Expr_mulContext_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_mulContext_a, __pyx_k_ModelicaParser_Expr_mulContext_a, sizeof(__pyx_k_ModelicaParser_Expr_mulContext_a), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_mulContext_e, __pyx_k_ModelicaParser_Expr_mulContext_e, sizeof(__pyx_k_ModelicaParser_Expr_mulContext_e), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_mulContext_e_2, __pyx_k_ModelicaParser_Expr_mulContext_e_2, sizeof(__pyx_k_ModelicaParser_Expr_mulContext_e_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_mulContext_e_3, __pyx_k_ModelicaParser_Expr_mulContext_e_3, sizeof(__pyx_k_ModelicaParser_Expr_mulContext_e_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_negContext, __pyx_k_ModelicaParser_Expr_negContext, sizeof(__pyx_k_ModelicaParser_Expr_negContext), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_negContext_2, __pyx_k_ModelicaParser_Expr_negContext_2, sizeof(__pyx_k_ModelicaParser_Expr_negContext_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_negContext_a, __pyx_k_ModelicaParser_Expr_negContext_a, sizeof(__pyx_k_ModelicaParser_Expr_negContext_a), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_negContext_e, __pyx_k_ModelicaParser_Expr_negContext_e, sizeof(__pyx_k_ModelicaParser_Expr_negContext_e), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_negContext_e_2, __pyx_k_ModelicaParser_Expr_negContext_e_2, sizeof(__pyx_k_ModelicaParser_Expr_negContext_e_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_negContext_e_3, __pyx_k_ModelicaParser_Expr_negContext_e_3, sizeof(__pyx_k_ModelicaParser_Expr_negContext_e_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_notContext, __pyx_k_ModelicaParser_Expr_notContext, sizeof(__pyx_k_ModelicaParser_Expr_notContext), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_notContext_2, __pyx_k_ModelicaParser_Expr_notContext_2, sizeof(__pyx_k_ModelicaParser_Expr_notContext_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_notContext_a, __pyx_k_ModelicaParser_Expr_notContext_a, sizeof(__pyx_k_ModelicaParser_Expr_notContext_a), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_notContext_e, __pyx_k_ModelicaParser_Expr_notContext_e, sizeof(__pyx_k_ModelicaParser_Expr_notContext_e), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_notContext_e_2, __pyx_k_ModelicaParser_Expr_notContext_e_2, sizeof(__pyx_k_ModelicaParser_Expr_notContext_e_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_notContext_e_3, __pyx_k_ModelicaParser_Expr_notContext_e_3, sizeof(__pyx_k_ModelicaParser_Expr_notContext_e_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_orContext, __pyx_k_ModelicaParser_Expr_orContext, sizeof(__pyx_k_ModelicaParser_Expr_orContext), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_orContext_2, __pyx_k_ModelicaParser_Expr_orContext_2, sizeof(__pyx_k_ModelicaParser_Expr_orContext_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_orContext_ac, __pyx_k_ModelicaParser_Expr_orContext_ac, sizeof(__pyx_k_ModelicaParser_Expr_orContext_ac), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_orContext_en, __pyx_k_ModelicaParser_Expr_orContext_en, sizeof(__pyx_k_ModelicaParser_Expr_orContext_en), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_orContext_ex, __pyx_k_ModelicaParser_Expr_orContext_ex, sizeof(__pyx_k_ModelicaParser_Expr_orContext_ex), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_orContext_ex_2, __pyx_k_ModelicaParser_Expr_orContext_ex_2, sizeof(__pyx_k_ModelicaParser_Expr_orContext_ex_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_primaryConte, __pyx_k_ModelicaParser_Expr_primaryConte, sizeof(__pyx_k_ModelicaParser_Expr_primaryConte), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_primaryConte_2, __pyx_k_ModelicaParser_Expr_primaryConte_2, sizeof(__pyx_k_ModelicaParser_Expr_primaryConte_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_primaryConte_3, __pyx_k_ModelicaParser_Expr_primaryConte_3, sizeof(__pyx_k_ModelicaParser_Expr_primaryConte_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_primaryConte_4, __pyx_k_ModelicaParser_Expr_primaryConte_4, sizeof(__pyx_k_ModelicaParser_Expr_primaryConte_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_primaryConte_5, __pyx_k_ModelicaParser_Expr_primaryConte_5, sizeof(__pyx_k_ModelicaParser_Expr_primaryConte_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_primaryConte_6, __pyx_k_ModelicaParser_Expr_primaryConte_6, sizeof(__pyx_k_ModelicaParser_Expr_primaryConte_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_relContext, __pyx_k_ModelicaParser_Expr_relContext, sizeof(__pyx_k_ModelicaParser_Expr_relContext), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_relContext_2, __pyx_k_ModelicaParser_Expr_relContext_2, sizeof(__pyx_k_ModelicaParser_Expr_relContext_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_relContext_a, __pyx_k_ModelicaParser_Expr_relContext_a, sizeof(__pyx_k_ModelicaParser_Expr_relContext_a), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_relContext_e, __pyx_k_ModelicaParser_Expr_relContext_e, sizeof(__pyx_k_ModelicaParser_Expr_relContext_e), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_relContext_e_2, __pyx_k_ModelicaParser_Expr_relContext_e_2, sizeof(__pyx_k_ModelicaParser_Expr_relContext_e_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expr_relContext_e_3, __pyx_k_ModelicaParser_Expr_relContext_e_3, sizeof(__pyx_k_ModelicaParser_Expr_relContext_e_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_ExpressionContext, __pyx_k_ModelicaParser_ExpressionContext, sizeof(__pyx_k_ModelicaParser_ExpressionContext), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_ExpressionContext_2, __pyx_k_ModelicaParser_ExpressionContext_2, sizeof(__pyx_k_ModelicaParser_ExpressionContext_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_ExpressionContext_3, __pyx_k_ModelicaParser_ExpressionContext_3, sizeof(__pyx_k_ModelicaParser_ExpressionContext_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_ExpressionContext_4, __pyx_k_ModelicaParser_ExpressionContext_4, sizeof(__pyx_k_ModelicaParser_ExpressionContext_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expression_ifCont, __pyx_k_ModelicaParser_Expression_ifCont, sizeof(__pyx_k_ModelicaParser_Expression_ifCont), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expression_ifCont_2, __pyx_k_ModelicaParser_Expression_ifCont_2, sizeof(__pyx_k_ModelicaParser_Expression_ifCont_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expression_ifCont_3, __pyx_k_ModelicaParser_Expression_ifCont_3, sizeof(__pyx_k_ModelicaParser_Expression_ifCont_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expression_ifCont_4, __pyx_k_ModelicaParser_Expression_ifCont_4, sizeof(__pyx_k_ModelicaParser_Expression_ifCont_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expression_ifCont_5, __pyx_k_ModelicaParser_Expression_ifCont_5, sizeof(__pyx_k_ModelicaParser_Expression_ifCont_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expression_ifCont_6, __pyx_k_ModelicaParser_Expression_ifCont_6, sizeof(__pyx_k_ModelicaParser_Expression_ifCont_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expression_listCo, __pyx_k_ModelicaParser_Expression_listCo, sizeof(__pyx_k_ModelicaParser_Expression_listCo), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expression_listCo_2, __pyx_k_ModelicaParser_Expression_listCo_2, sizeof(__pyx_k_ModelicaParser_Expression_listCo_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expression_listCo_3, __pyx_k_ModelicaParser_Expression_listCo_3, sizeof(__pyx_k_ModelicaParser_Expression_listCo_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expression_listCo_4, __pyx_k_ModelicaParser_Expression_listCo_4, sizeof(__pyx_k_ModelicaParser_Expression_listCo_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expression_listCo_5, __pyx_k_ModelicaParser_Expression_listCo_5, sizeof(__pyx_k_ModelicaParser_Expression_listCo_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expression_listCo_6, __pyx_k_ModelicaParser_Expression_listCo_6, sizeof(__pyx_k_ModelicaParser_Expression_listCo_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expression_listCo_7, __pyx_k_ModelicaParser_Expression_listCo_7, sizeof(__pyx_k_ModelicaParser_Expression_listCo_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expression_simple, __pyx_k_ModelicaParser_Expression_simple, sizeof(__pyx_k_ModelicaParser_Expression_simple), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expression_simple_2, __pyx_k_ModelicaParser_Expression_simple_2, sizeof(__pyx_k_ModelicaParser_Expression_simple_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expression_simple_3, __pyx_k_ModelicaParser_Expression_simple_3, sizeof(__pyx_k_ModelicaParser_Expression_simple_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expression_simple_4, __pyx_k_ModelicaParser_Expression_simple_4, sizeof(__pyx_k_ModelicaParser_Expression_simple_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expression_simple_5, __pyx_k_ModelicaParser_Expression_simple_5, sizeof(__pyx_k_ModelicaParser_Expression_simple_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Expression_simple_6, __pyx_k_ModelicaParser_Expression_simple_6, sizeof(__pyx_k_ModelicaParser_Expression_simple_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Extends_clauseCon, __pyx_k_ModelicaParser_Extends_clauseCon, sizeof(__pyx_k_ModelicaParser_Extends_clauseCon), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Extends_clauseCon_2, __pyx_k_ModelicaParser_Extends_clauseCon_2, sizeof(__pyx_k_ModelicaParser_Extends_clauseCon_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Extends_clauseCon_3, __pyx_k_ModelicaParser_Extends_clauseCon_3, sizeof(__pyx_k_ModelicaParser_Extends_clauseCon_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Extends_clauseCon_4, __pyx_k_ModelicaParser_Extends_clauseCon_4, sizeof(__pyx_k_ModelicaParser_Extends_clauseCon_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Extends_clauseCon_5, __pyx_k_ModelicaParser_Extends_clauseCon_5, sizeof(__pyx_k_ModelicaParser_Extends_clauseCon_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Extends_clauseCon_6, __pyx_k_ModelicaParser_Extends_clauseCon_6, sizeof(__pyx_k_ModelicaParser_Extends_clauseCon_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Extends_clauseCon_7, __pyx_k_ModelicaParser_Extends_clauseCon_7, sizeof(__pyx_k_ModelicaParser_Extends_clauseCon_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Extends_clauseCon_8, __pyx_k_ModelicaParser_Extends_clauseCon_8, sizeof(__pyx_k_ModelicaParser_Extends_clauseCon_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Extends_clauseCon_9, __pyx_k_ModelicaParser_Extends_clauseCon_9, sizeof(__pyx_k_ModelicaParser_Extends_clauseCon_9), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_External_function, __pyx_k_ModelicaParser_External_function, sizeof(__pyx_k_ModelicaParser_External_function), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_External_function_2, __pyx_k_ModelicaParser_External_function_2, sizeof(__pyx_k_ModelicaParser_External_function_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_External_function_3, __pyx_k_ModelicaParser_External_function_3, sizeof(__pyx_k_ModelicaParser_External_function_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_External_function_4, __pyx_k_ModelicaParser_External_function_4, sizeof(__pyx_k_ModelicaParser_External_function_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_External_function_5, __pyx_k_ModelicaParser_External_function_5, sizeof(__pyx_k_ModelicaParser_External_function_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_External_function_6, __pyx_k_ModelicaParser_External_function_6, sizeof(__pyx_k_ModelicaParser_External_function_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_External_function_7, __pyx_k_ModelicaParser_External_function_7, sizeof(__pyx_k_ModelicaParser_External_function_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_External_function_8, __pyx_k_ModelicaParser_External_function_8, sizeof(__pyx_k_ModelicaParser_External_function_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_External_function_9, __pyx_k_ModelicaParser_External_function_9, sizeof(__pyx_k_ModelicaParser_External_function_9), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_For_equationConte, __pyx_k_ModelicaParser_For_equationConte, sizeof(__pyx_k_ModelicaParser_For_equationConte), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_For_equationConte_2, __pyx_k_ModelicaParser_For_equationConte_2, sizeof(__pyx_k_ModelicaParser_For_equationConte_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_For_equationConte_3, __pyx_k_ModelicaParser_For_equationConte_3, sizeof(__pyx_k_ModelicaParser_For_equationConte_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_For_equationConte_4, __pyx_k_ModelicaParser_For_equationConte_4, sizeof(__pyx_k_ModelicaParser_For_equationConte_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_For_equationConte_5, __pyx_k_ModelicaParser_For_equationConte_5, sizeof(__pyx_k_ModelicaParser_For_equationConte_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_For_equationConte_6, __pyx_k_ModelicaParser_For_equationConte_6, sizeof(__pyx_k_ModelicaParser_For_equationConte_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_For_equationConte_7, __pyx_k_ModelicaParser_For_equationConte_7, sizeof(__pyx_k_ModelicaParser_For_equationConte_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_For_equationConte_8, __pyx_k_ModelicaParser_For_equationConte_8, sizeof(__pyx_k_ModelicaParser_For_equationConte_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_For_indexContext, __pyx_k_ModelicaParser_For_indexContext, sizeof(__pyx_k_ModelicaParser_For_indexContext), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_For_indexContext_2, __pyx_k_ModelicaParser_For_indexContext_2, sizeof(__pyx_k_ModelicaParser_For_indexContext_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_For_indexContext_3, __pyx_k_ModelicaParser_For_indexContext_3, sizeof(__pyx_k_ModelicaParser_For_indexContext_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_For_indexContext_4, __pyx_k_ModelicaParser_For_indexContext_4, sizeof(__pyx_k_ModelicaParser_For_indexContext_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_For_indexContext_5, __pyx_k_ModelicaParser_For_indexContext_5, sizeof(__pyx_k_ModelicaParser_For_indexContext_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_For_indexContext_6, __pyx_k_ModelicaParser_For_indexContext_6, sizeof(__pyx_k_ModelicaParser_For_indexContext_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_For_indexContext_7, __pyx_k_ModelicaParser_For_indexContext_7, sizeof(__pyx_k_ModelicaParser_For_indexContext_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_For_indexContext_8, __pyx_k_ModelicaParser_For_indexContext_8, sizeof(__pyx_k_ModelicaParser_For_indexContext_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_For_indicesContex, __pyx_k_ModelicaParser_For_indicesContex, sizeof(__pyx_k_ModelicaParser_For_indicesContex), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_For_indicesContex_2, __pyx_k_ModelicaParser_For_indicesContex_2, sizeof(__pyx_k_ModelicaParser_For_indicesContex_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_For_indicesContex_3, __pyx_k_ModelicaParser_For_indicesContex_3, sizeof(__pyx_k_ModelicaParser_For_indicesContex_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_For_indicesContex_4, __pyx_k_ModelicaParser_For_indicesContex_4, sizeof(__pyx_k_ModelicaParser_For_indicesContex_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_For_indicesContex_5, __pyx_k_ModelicaParser_For_indicesContex_5, sizeof(__pyx_k_ModelicaParser_For_indicesContex_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_For_indicesContex_6, __pyx_k_ModelicaParser_For_indicesContex_6, sizeof(__pyx_k_ModelicaParser_For_indicesContex_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_For_indicesContex_7, __pyx_k_ModelicaParser_For_indicesContex_7, sizeof(__pyx_k_ModelicaParser_For_indicesContex_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_For_statementCont, __pyx_k_ModelicaParser_For_statementCont, sizeof(__pyx_k_ModelicaParser_For_statementCont), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_For_statementCont_2, __pyx_k_ModelicaParser_For_statementCont_2, sizeof(__pyx_k_ModelicaParser_For_statementCont_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_For_statementCont_3, __pyx_k_ModelicaParser_For_statementCont_3, sizeof(__pyx_k_ModelicaParser_For_statementCont_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_For_statementCont_4, __pyx_k_ModelicaParser_For_statementCont_4, sizeof(__pyx_k_ModelicaParser_For_statementCont_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_For_statementCont_5, __pyx_k_ModelicaParser_For_statementCont_5, sizeof(__pyx_k_ModelicaParser_For_statementCont_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_For_statementCont_6, __pyx_k_ModelicaParser_For_statementCont_6, sizeof(__pyx_k_ModelicaParser_For_statementCont_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_For_statementCont_7, __pyx_k_ModelicaParser_For_statementCont_7, sizeof(__pyx_k_ModelicaParser_For_statementCont_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_For_statementCont_8, __pyx_k_ModelicaParser_For_statementCont_8, sizeof(__pyx_k_ModelicaParser_For_statementCont_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Function_argument, __pyx_k_ModelicaParser_Function_argument, sizeof(__pyx_k_ModelicaParser_Function_argument), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Function_argument_10, __pyx_k_ModelicaParser_Function_argument_10, sizeof(__pyx_k_ModelicaParser_Function_argument_10), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Function_argument_11, __pyx_k_ModelicaParser_Function_argument_11, sizeof(__pyx_k_ModelicaParser_Function_argument_11), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Function_argument_12, __pyx_k_ModelicaParser_Function_argument_12, sizeof(__pyx_k_ModelicaParser_Function_argument_12), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Function_argument_13, __pyx_k_ModelicaParser_Function_argument_13, sizeof(__pyx_k_ModelicaParser_Function_argument_13), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Function_argument_2, __pyx_k_ModelicaParser_Function_argument_2, sizeof(__pyx_k_ModelicaParser_Function_argument_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Function_argument_3, __pyx_k_ModelicaParser_Function_argument_3, sizeof(__pyx_k_ModelicaParser_Function_argument_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Function_argument_4, __pyx_k_ModelicaParser_Function_argument_4, sizeof(__pyx_k_ModelicaParser_Function_argument_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Function_argument_5, __pyx_k_ModelicaParser_Function_argument_5, sizeof(__pyx_k_ModelicaParser_Function_argument_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Function_argument_6, __pyx_k_ModelicaParser_Function_argument_6, sizeof(__pyx_k_ModelicaParser_Function_argument_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Function_argument_7, __pyx_k_ModelicaParser_Function_argument_7, sizeof(__pyx_k_ModelicaParser_Function_argument_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Function_argument_8, __pyx_k_ModelicaParser_Function_argument_8, sizeof(__pyx_k_ModelicaParser_Function_argument_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Function_argument_9, __pyx_k_ModelicaParser_Function_argument_9, sizeof(__pyx_k_ModelicaParser_Function_argument_9), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Function_call_arg, __pyx_k_ModelicaParser_Function_call_arg, sizeof(__pyx_k_ModelicaParser_Function_call_arg), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Function_call_arg_2, __pyx_k_ModelicaParser_Function_call_arg_2, sizeof(__pyx_k_ModelicaParser_Function_call_arg_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Function_call_arg_3, __pyx_k_ModelicaParser_Function_call_arg_3, sizeof(__pyx_k_ModelicaParser_Function_call_arg_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Function_call_arg_4, __pyx_k_ModelicaParser_Function_call_arg_4, sizeof(__pyx_k_ModelicaParser_Function_call_arg_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Function_call_arg_5, __pyx_k_ModelicaParser_Function_call_arg_5, sizeof(__pyx_k_ModelicaParser_Function_call_arg_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Function_call_arg_6, __pyx_k_ModelicaParser_Function_call_arg_6, sizeof(__pyx_k_ModelicaParser_Function_call_arg_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Function_call_arg_7, __pyx_k_ModelicaParser_Function_call_arg_7, sizeof(__pyx_k_ModelicaParser_Function_call_arg_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_If_equationContex, __pyx_k_ModelicaParser_If_equationContex, sizeof(__pyx_k_ModelicaParser_If_equationContex), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_If_equationContex_2, __pyx_k_ModelicaParser_If_equationContex_2, sizeof(__pyx_k_ModelicaParser_If_equationContex_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_If_equationContex_3, __pyx_k_ModelicaParser_If_equationContex_3, sizeof(__pyx_k_ModelicaParser_If_equationContex_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_If_equationContex_4, __pyx_k_ModelicaParser_If_equationContex_4, sizeof(__pyx_k_ModelicaParser_If_equationContex_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_If_equationContex_5, __pyx_k_ModelicaParser_If_equationContex_5, sizeof(__pyx_k_ModelicaParser_If_equationContex_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_If_equationContex_6, __pyx_k_ModelicaParser_If_equationContex_6, sizeof(__pyx_k_ModelicaParser_If_equationContex_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_If_equationContex_7, __pyx_k_ModelicaParser_If_equationContex_7, sizeof(__pyx_k_ModelicaParser_If_equationContex_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_If_equationContex_8, __pyx_k_ModelicaParser_If_equationContex_8, sizeof(__pyx_k_ModelicaParser_If_equationContex_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_If_statementConte, __pyx_k_ModelicaParser_If_statementConte, sizeof(__pyx_k_ModelicaParser_If_statementConte), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_If_statementConte_2, __pyx_k_ModelicaParser_If_statementConte_2, sizeof(__pyx_k_ModelicaParser_If_statementConte_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_If_statementConte_3, __pyx_k_ModelicaParser_If_statementConte_3, sizeof(__pyx_k_ModelicaParser_If_statementConte_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_If_statementConte_4, __pyx_k_ModelicaParser_If_statementConte_4, sizeof(__pyx_k_ModelicaParser_If_statementConte_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_If_statementConte_5, __pyx_k_ModelicaParser_If_statementConte_5, sizeof(__pyx_k_ModelicaParser_If_statementConte_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_If_statementConte_6, __pyx_k_ModelicaParser_If_statementConte_6, sizeof(__pyx_k_ModelicaParser_If_statementConte_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_If_statementConte_7, __pyx_k_ModelicaParser_If_statementConte_7, sizeof(__pyx_k_ModelicaParser_If_statementConte_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_If_statementConte_8, __pyx_k_ModelicaParser_If_statementConte_8, sizeof(__pyx_k_ModelicaParser_If_statementConte_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Import_clauseCont, __pyx_k_ModelicaParser_Import_clauseCont, sizeof(__pyx_k_ModelicaParser_Import_clauseCont), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Import_clauseCont_10, __pyx_k_ModelicaParser_Import_clauseCont_10, sizeof(__pyx_k_ModelicaParser_Import_clauseCont_10), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Import_clauseCont_2, __pyx_k_ModelicaParser_Import_clauseCont_2, sizeof(__pyx_k_ModelicaParser_Import_clauseCont_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Import_clauseCont_3, __pyx_k_ModelicaParser_Import_clauseCont_3, sizeof(__pyx_k_ModelicaParser_Import_clauseCont_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Import_clauseCont_4, __pyx_k_ModelicaParser_Import_clauseCont_4, sizeof(__pyx_k_ModelicaParser_Import_clauseCont_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Import_clauseCont_5, __pyx_k_ModelicaParser_Import_clauseCont_5, sizeof(__pyx_k_ModelicaParser_Import_clauseCont_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Import_clauseCont_6, __pyx_k_ModelicaParser_Import_clauseCont_6, sizeof(__pyx_k_ModelicaParser_Import_clauseCont_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Import_clauseCont_7, __pyx_k_ModelicaParser_Import_clauseCont_7, sizeof(__pyx_k_ModelicaParser_Import_clauseCont_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Import_clauseCont_8, __pyx_k_ModelicaParser_Import_clauseCont_8, sizeof(__pyx_k_ModelicaParser_Import_clauseCont_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Import_clauseCont_9, __pyx_k_ModelicaParser_Import_clauseCont_9, sizeof(__pyx_k_ModelicaParser_Import_clauseCont_9), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Import_listContex, __pyx_k_ModelicaParser_Import_listContex, sizeof(__pyx_k_ModelicaParser_Import_listContex), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Import_listContex_2, __pyx_k_ModelicaParser_Import_listContex_2, sizeof(__pyx_k_ModelicaParser_Import_listContex_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Import_listContex_3, __pyx_k_ModelicaParser_Import_listContex_3, sizeof(__pyx_k_ModelicaParser_Import_listContex_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Import_listContex_4, __pyx_k_ModelicaParser_Import_listContex_4, sizeof(__pyx_k_ModelicaParser_Import_listContex_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Import_listContex_5, __pyx_k_ModelicaParser_Import_listContex_5, sizeof(__pyx_k_ModelicaParser_Import_listContex_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Import_listContex_6, __pyx_k_ModelicaParser_Import_listContex_6, sizeof(__pyx_k_ModelicaParser_Import_listContex_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Import_listContex_7, __pyx_k_ModelicaParser_Import_listContex_7, sizeof(__pyx_k_ModelicaParser_Import_listContex_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Import_listContex_8, __pyx_k_ModelicaParser_Import_listContex_8, sizeof(__pyx_k_ModelicaParser_Import_listContex_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Language_specific, __pyx_k_ModelicaParser_Language_specific, sizeof(__pyx_k_ModelicaParser_Language_specific), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Language_specific_2, __pyx_k_ModelicaParser_Language_specific_2, sizeof(__pyx_k_ModelicaParser_Language_specific_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Language_specific_3, __pyx_k_ModelicaParser_Language_specific_3, sizeof(__pyx_k_ModelicaParser_Language_specific_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Language_specific_4, __pyx_k_ModelicaParser_Language_specific_4, sizeof(__pyx_k_ModelicaParser_Language_specific_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Language_specific_5, __pyx_k_ModelicaParser_Language_specific_5, sizeof(__pyx_k_ModelicaParser_Language_specific_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Language_specific_6, __pyx_k_ModelicaParser_Language_specific_6, sizeof(__pyx_k_ModelicaParser_Language_specific_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Language_specific_7, __pyx_k_ModelicaParser_Language_specific_7, sizeof(__pyx_k_ModelicaParser_Language_specific_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_ModificationConte, __pyx_k_ModelicaParser_ModificationConte, sizeof(__pyx_k_ModelicaParser_ModificationConte), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_ModificationConte_2, __pyx_k_ModelicaParser_ModificationConte_2, sizeof(__pyx_k_ModelicaParser_ModificationConte_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_ModificationConte_3, __pyx_k_ModelicaParser_ModificationConte_3, sizeof(__pyx_k_ModelicaParser_ModificationConte_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_ModificationConte_4, __pyx_k_ModelicaParser_ModificationConte_4, sizeof(__pyx_k_ModelicaParser_ModificationConte_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Modification_assi, __pyx_k_ModelicaParser_Modification_assi, sizeof(__pyx_k_ModelicaParser_Modification_assi), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Modification_assi_10, __pyx_k_ModelicaParser_Modification_assi_10, sizeof(__pyx_k_ModelicaParser_Modification_assi_10), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Modification_assi_11, __pyx_k_ModelicaParser_Modification_assi_11, sizeof(__pyx_k_ModelicaParser_Modification_assi_11), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Modification_assi_12, __pyx_k_ModelicaParser_Modification_assi_12, sizeof(__pyx_k_ModelicaParser_Modification_assi_12), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Modification_assi_2, __pyx_k_ModelicaParser_Modification_assi_2, sizeof(__pyx_k_ModelicaParser_Modification_assi_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Modification_assi_3, __pyx_k_ModelicaParser_Modification_assi_3, sizeof(__pyx_k_ModelicaParser_Modification_assi_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Modification_assi_4, __pyx_k_ModelicaParser_Modification_assi_4, sizeof(__pyx_k_ModelicaParser_Modification_assi_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Modification_assi_5, __pyx_k_ModelicaParser_Modification_assi_5, sizeof(__pyx_k_ModelicaParser_Modification_assi_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Modification_assi_6, __pyx_k_ModelicaParser_Modification_assi_6, sizeof(__pyx_k_ModelicaParser_Modification_assi_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Modification_assi_7, __pyx_k_ModelicaParser_Modification_assi_7, sizeof(__pyx_k_ModelicaParser_Modification_assi_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Modification_assi_8, __pyx_k_ModelicaParser_Modification_assi_8, sizeof(__pyx_k_ModelicaParser_Modification_assi_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Modification_assi_9, __pyx_k_ModelicaParser_Modification_assi_9, sizeof(__pyx_k_ModelicaParser_Modification_assi_9), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Modification_clas, __pyx_k_ModelicaParser_Modification_clas, sizeof(__pyx_k_ModelicaParser_Modification_clas), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Modification_clas_2, __pyx_k_ModelicaParser_Modification_clas_2, sizeof(__pyx_k_ModelicaParser_Modification_clas_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Modification_clas_3, __pyx_k_ModelicaParser_Modification_clas_3, sizeof(__pyx_k_ModelicaParser_Modification_clas_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Modification_clas_4, __pyx_k_ModelicaParser_Modification_clas_4, sizeof(__pyx_k_ModelicaParser_Modification_clas_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Modification_clas_5, __pyx_k_ModelicaParser_Modification_clas_5, sizeof(__pyx_k_ModelicaParser_Modification_clas_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Modification_clas_6, __pyx_k_ModelicaParser_Modification_clas_6, sizeof(__pyx_k_ModelicaParser_Modification_clas_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Modification_clas_7, __pyx_k_ModelicaParser_Modification_clas_7, sizeof(__pyx_k_ModelicaParser_Modification_clas_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_NameContext, __pyx_k_ModelicaParser_NameContext, sizeof(__pyx_k_ModelicaParser_NameContext), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_NameContext_IDENT, __pyx_k_ModelicaParser_NameContext_IDENT, sizeof(__pyx_k_ModelicaParser_NameContext_IDENT), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_NameContext___ini, __pyx_k_ModelicaParser_NameContext___ini, sizeof(__pyx_k_ModelicaParser_NameContext___ini), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_NameContext_accep, __pyx_k_ModelicaParser_NameContext_accep, sizeof(__pyx_k_ModelicaParser_NameContext_accep), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_NameContext_enter, __pyx_k_ModelicaParser_NameContext_enter, sizeof(__pyx_k_ModelicaParser_NameContext_enter), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_NameContext_exitR, __pyx_k_ModelicaParser_NameContext_exitR, sizeof(__pyx_k_ModelicaParser_NameContext_exitR), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_NameContext_getRu, __pyx_k_ModelicaParser_NameContext_getRu, sizeof(__pyx_k_ModelicaParser_NameContext_getRu), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Named_argumentCon, __pyx_k_ModelicaParser_Named_argumentCon, sizeof(__pyx_k_ModelicaParser_Named_argumentCon), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Named_argumentCon_2, __pyx_k_ModelicaParser_Named_argumentCon_2, sizeof(__pyx_k_ModelicaParser_Named_argumentCon_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Named_argumentCon_3, __pyx_k_ModelicaParser_Named_argumentCon_3, sizeof(__pyx_k_ModelicaParser_Named_argumentCon_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Named_argumentCon_4, __pyx_k_ModelicaParser_Named_argumentCon_4, sizeof(__pyx_k_ModelicaParser_Named_argumentCon_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Named_argumentCon_5, __pyx_k_ModelicaParser_Named_argumentCon_5, sizeof(__pyx_k_ModelicaParser_Named_argumentCon_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Named_argumentCon_6, __pyx_k_ModelicaParser_Named_argumentCon_6, sizeof(__pyx_k_ModelicaParser_Named_argumentCon_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Named_argumentCon_7, __pyx_k_ModelicaParser_Named_argumentCon_7, sizeof(__pyx_k_ModelicaParser_Named_argumentCon_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Named_argumentCon_8, __pyx_k_ModelicaParser_Named_argumentCon_8, sizeof(__pyx_k_ModelicaParser_Named_argumentCon_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Named_argumentsCo, __pyx_k_ModelicaParser_Named_argumentsCo, sizeof(__pyx_k_ModelicaParser_Named_argumentsCo), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Named_argumentsCo_2, __pyx_k_ModelicaParser_Named_argumentsCo_2, sizeof(__pyx_k_ModelicaParser_Named_argumentsCo_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Named_argumentsCo_3, __pyx_k_ModelicaParser_Named_argumentsCo_3, sizeof(__pyx_k_ModelicaParser_Named_argumentsCo_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Named_argumentsCo_4, __pyx_k_ModelicaParser_Named_argumentsCo_4, sizeof(__pyx_k_ModelicaParser_Named_argumentsCo_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Named_argumentsCo_5, __pyx_k_ModelicaParser_Named_argumentsCo_5, sizeof(__pyx_k_ModelicaParser_Named_argumentsCo_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Named_argumentsCo_6, __pyx_k_ModelicaParser_Named_argumentsCo_6, sizeof(__pyx_k_ModelicaParser_Named_argumentsCo_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Named_argumentsCo_7, __pyx_k_ModelicaParser_Named_argumentsCo_7, sizeof(__pyx_k_ModelicaParser_Named_argumentsCo_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Output_expression, __pyx_k_ModelicaParser_Output_expression, sizeof(__pyx_k_ModelicaParser_Output_expression), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Output_expression_2, __pyx_k_ModelicaParser_Output_expression_2, sizeof(__pyx_k_ModelicaParser_Output_expression_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Output_expression_3, __pyx_k_ModelicaParser_Output_expression_3, sizeof(__pyx_k_ModelicaParser_Output_expression_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Output_expression_4, __pyx_k_ModelicaParser_Output_expression_4, sizeof(__pyx_k_ModelicaParser_Output_expression_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Output_expression_5, __pyx_k_ModelicaParser_Output_expression_5, sizeof(__pyx_k_ModelicaParser_Output_expression_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Output_expression_6, __pyx_k_ModelicaParser_Output_expression_6, sizeof(__pyx_k_ModelicaParser_Output_expression_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Output_expression_7, __pyx_k_ModelicaParser_Output_expression_7, sizeof(__pyx_k_ModelicaParser_Output_expression_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_PrimaryContext, __pyx_k_ModelicaParser_PrimaryContext, sizeof(__pyx_k_ModelicaParser_PrimaryContext), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_PrimaryContext_2, __pyx_k_ModelicaParser_PrimaryContext_2, sizeof(__pyx_k_ModelicaParser_PrimaryContext_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_PrimaryContext_co, __pyx_k_ModelicaParser_PrimaryContext_co, sizeof(__pyx_k_ModelicaParser_PrimaryContext_co), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_PrimaryContext_ge, __pyx_k_ModelicaParser_PrimaryContext_ge, sizeof(__pyx_k_ModelicaParser_PrimaryContext_ge), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_component, __pyx_k_ModelicaParser_Primary_component, sizeof(__pyx_k_ModelicaParser_Primary_component), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_component_2, __pyx_k_ModelicaParser_Primary_component_2, sizeof(__pyx_k_ModelicaParser_Primary_component_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_component_3, __pyx_k_ModelicaParser_Primary_component_3, sizeof(__pyx_k_ModelicaParser_Primary_component_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_component_4, __pyx_k_ModelicaParser_Primary_component_4, sizeof(__pyx_k_ModelicaParser_Primary_component_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_component_5, __pyx_k_ModelicaParser_Primary_component_5, sizeof(__pyx_k_ModelicaParser_Primary_component_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_component_6, __pyx_k_ModelicaParser_Primary_component_6, sizeof(__pyx_k_ModelicaParser_Primary_component_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_derivativ, __pyx_k_ModelicaParser_Primary_derivativ, sizeof(__pyx_k_ModelicaParser_Primary_derivativ), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_derivativ_2, __pyx_k_ModelicaParser_Primary_derivativ_2, sizeof(__pyx_k_ModelicaParser_Primary_derivativ_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_derivativ_3, __pyx_k_ModelicaParser_Primary_derivativ_3, sizeof(__pyx_k_ModelicaParser_Primary_derivativ_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_derivativ_4, __pyx_k_ModelicaParser_Primary_derivativ_4, sizeof(__pyx_k_ModelicaParser_Primary_derivativ_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_derivativ_5, __pyx_k_ModelicaParser_Primary_derivativ_5, sizeof(__pyx_k_ModelicaParser_Primary_derivativ_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_derivativ_6, __pyx_k_ModelicaParser_Primary_derivativ_6, sizeof(__pyx_k_ModelicaParser_Primary_derivativ_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_endContex, __pyx_k_ModelicaParser_Primary_endContex, sizeof(__pyx_k_ModelicaParser_Primary_endContex), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_endContex_2, __pyx_k_ModelicaParser_Primary_endContex_2, sizeof(__pyx_k_ModelicaParser_Primary_endContex_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_endContex_3, __pyx_k_ModelicaParser_Primary_endContex_3, sizeof(__pyx_k_ModelicaParser_Primary_endContex_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_endContex_4, __pyx_k_ModelicaParser_Primary_endContex_4, sizeof(__pyx_k_ModelicaParser_Primary_endContex_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_endContex_5, __pyx_k_ModelicaParser_Primary_endContex_5, sizeof(__pyx_k_ModelicaParser_Primary_endContex_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_expressio, __pyx_k_ModelicaParser_Primary_expressio, sizeof(__pyx_k_ModelicaParser_Primary_expressio), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_expressio_2, __pyx_k_ModelicaParser_Primary_expressio_2, sizeof(__pyx_k_ModelicaParser_Primary_expressio_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_expressio_3, __pyx_k_ModelicaParser_Primary_expressio_3, sizeof(__pyx_k_ModelicaParser_Primary_expressio_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_expressio_4, __pyx_k_ModelicaParser_Primary_expressio_4, sizeof(__pyx_k_ModelicaParser_Primary_expressio_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_expressio_5, __pyx_k_ModelicaParser_Primary_expressio_5, sizeof(__pyx_k_ModelicaParser_Primary_expressio_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_expressio_6, __pyx_k_ModelicaParser_Primary_expressio_6, sizeof(__pyx_k_ModelicaParser_Primary_expressio_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_falseCont, __pyx_k_ModelicaParser_Primary_falseCont, sizeof(__pyx_k_ModelicaParser_Primary_falseCont), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_falseCont_2, __pyx_k_ModelicaParser_Primary_falseCont_2, sizeof(__pyx_k_ModelicaParser_Primary_falseCont_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_falseCont_3, __pyx_k_ModelicaParser_Primary_falseCont_3, sizeof(__pyx_k_ModelicaParser_Primary_falseCont_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_falseCont_4, __pyx_k_ModelicaParser_Primary_falseCont_4, sizeof(__pyx_k_ModelicaParser_Primary_falseCont_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_falseCont_5, __pyx_k_ModelicaParser_Primary_falseCont_5, sizeof(__pyx_k_ModelicaParser_Primary_falseCont_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_function, __pyx_k_ModelicaParser_Primary_function, sizeof(__pyx_k_ModelicaParser_Primary_function), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_functionC, __pyx_k_ModelicaParser_Primary_functionC, sizeof(__pyx_k_ModelicaParser_Primary_functionC), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_functionC_2, __pyx_k_ModelicaParser_Primary_functionC_2, sizeof(__pyx_k_ModelicaParser_Primary_functionC_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_functionC_3, __pyx_k_ModelicaParser_Primary_functionC_3, sizeof(__pyx_k_ModelicaParser_Primary_functionC_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_functionC_4, __pyx_k_ModelicaParser_Primary_functionC_4, sizeof(__pyx_k_ModelicaParser_Primary_functionC_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_functionC_5, __pyx_k_ModelicaParser_Primary_functionC_5, sizeof(__pyx_k_ModelicaParser_Primary_functionC_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_functionC_6, __pyx_k_ModelicaParser_Primary_functionC_6, sizeof(__pyx_k_ModelicaParser_Primary_functionC_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_functionC_7, __pyx_k_ModelicaParser_Primary_functionC_7, sizeof(__pyx_k_ModelicaParser_Primary_functionC_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_function_2, __pyx_k_ModelicaParser_Primary_function_2, sizeof(__pyx_k_ModelicaParser_Primary_function_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_function_3, __pyx_k_ModelicaParser_Primary_function_3, sizeof(__pyx_k_ModelicaParser_Primary_function_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_function_4, __pyx_k_ModelicaParser_Primary_function_4, sizeof(__pyx_k_ModelicaParser_Primary_function_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_function_5, __pyx_k_ModelicaParser_Primary_function_5, sizeof(__pyx_k_ModelicaParser_Primary_function_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_function_6, __pyx_k_ModelicaParser_Primary_function_6, sizeof(__pyx_k_ModelicaParser_Primary_function_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_initialCo, __pyx_k_ModelicaParser_Primary_initialCo, sizeof(__pyx_k_ModelicaParser_Primary_initialCo), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_initialCo_2, __pyx_k_ModelicaParser_Primary_initialCo_2, sizeof(__pyx_k_ModelicaParser_Primary_initialCo_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_initialCo_3, __pyx_k_ModelicaParser_Primary_initialCo_3, sizeof(__pyx_k_ModelicaParser_Primary_initialCo_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_initialCo_4, __pyx_k_ModelicaParser_Primary_initialCo_4, sizeof(__pyx_k_ModelicaParser_Primary_initialCo_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_initialCo_5, __pyx_k_ModelicaParser_Primary_initialCo_5, sizeof(__pyx_k_ModelicaParser_Primary_initialCo_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_initialCo_6, __pyx_k_ModelicaParser_Primary_initialCo_6, sizeof(__pyx_k_ModelicaParser_Primary_initialCo_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_output_ex, __pyx_k_ModelicaParser_Primary_output_ex, sizeof(__pyx_k_ModelicaParser_Primary_output_ex), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_output_ex_2, __pyx_k_ModelicaParser_Primary_output_ex_2, sizeof(__pyx_k_ModelicaParser_Primary_output_ex_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_output_ex_3, __pyx_k_ModelicaParser_Primary_output_ex_3, sizeof(__pyx_k_ModelicaParser_Primary_output_ex_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_output_ex_4, __pyx_k_ModelicaParser_Primary_output_ex_4, sizeof(__pyx_k_ModelicaParser_Primary_output_ex_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_output_ex_5, __pyx_k_ModelicaParser_Primary_output_ex_5, sizeof(__pyx_k_ModelicaParser_Primary_output_ex_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_output_ex_6, __pyx_k_ModelicaParser_Primary_output_ex_6, sizeof(__pyx_k_ModelicaParser_Primary_output_ex_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_stringCon, __pyx_k_ModelicaParser_Primary_stringCon, sizeof(__pyx_k_ModelicaParser_Primary_stringCon), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_stringCon_2, __pyx_k_ModelicaParser_Primary_stringCon_2, sizeof(__pyx_k_ModelicaParser_Primary_stringCon_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_stringCon_3, __pyx_k_ModelicaParser_Primary_stringCon_3, sizeof(__pyx_k_ModelicaParser_Primary_stringCon_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_stringCon_4, __pyx_k_ModelicaParser_Primary_stringCon_4, sizeof(__pyx_k_ModelicaParser_Primary_stringCon_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_stringCon_5, __pyx_k_ModelicaParser_Primary_stringCon_5, sizeof(__pyx_k_ModelicaParser_Primary_stringCon_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_stringCon_6, __pyx_k_ModelicaParser_Primary_stringCon_6, sizeof(__pyx_k_ModelicaParser_Primary_stringCon_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_trueConte, __pyx_k_ModelicaParser_Primary_trueConte, sizeof(__pyx_k_ModelicaParser_Primary_trueConte), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_trueConte_2, __pyx_k_ModelicaParser_Primary_trueConte_2, sizeof(__pyx_k_ModelicaParser_Primary_trueConte_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_trueConte_3, __pyx_k_ModelicaParser_Primary_trueConte_3, sizeof(__pyx_k_ModelicaParser_Primary_trueConte_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_trueConte_4, __pyx_k_ModelicaParser_Primary_trueConte_4, sizeof(__pyx_k_ModelicaParser_Primary_trueConte_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_trueConte_5, __pyx_k_ModelicaParser_Primary_trueConte_5, sizeof(__pyx_k_ModelicaParser_Primary_trueConte_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_unsigned, __pyx_k_ModelicaParser_Primary_unsigned, sizeof(__pyx_k_ModelicaParser_Primary_unsigned), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_unsigned_2, __pyx_k_ModelicaParser_Primary_unsigned_2, sizeof(__pyx_k_ModelicaParser_Primary_unsigned_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_unsigned_3, __pyx_k_ModelicaParser_Primary_unsigned_3, sizeof(__pyx_k_ModelicaParser_Primary_unsigned_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_unsigned_4, __pyx_k_ModelicaParser_Primary_unsigned_4, sizeof(__pyx_k_ModelicaParser_Primary_unsigned_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_unsigned_5, __pyx_k_ModelicaParser_Primary_unsigned_5, sizeof(__pyx_k_ModelicaParser_Primary_unsigned_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Primary_unsigned_6, __pyx_k_ModelicaParser_Primary_unsigned_6, sizeof(__pyx_k_ModelicaParser_Primary_unsigned_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Regular_elementCo, __pyx_k_ModelicaParser_Regular_elementCo, sizeof(__pyx_k_ModelicaParser_Regular_elementCo), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Regular_elementCo_10, __pyx_k_ModelicaParser_Regular_elementCo_10, sizeof(__pyx_k_ModelicaParser_Regular_elementCo_10), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Regular_elementCo_11, __pyx_k_ModelicaParser_Regular_elementCo_11, sizeof(__pyx_k_ModelicaParser_Regular_elementCo_11), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Regular_elementCo_12, __pyx_k_ModelicaParser_Regular_elementCo_12, sizeof(__pyx_k_ModelicaParser_Regular_elementCo_12), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Regular_elementCo_2, __pyx_k_ModelicaParser_Regular_elementCo_2, sizeof(__pyx_k_ModelicaParser_Regular_elementCo_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Regular_elementCo_3, __pyx_k_ModelicaParser_Regular_elementCo_3, sizeof(__pyx_k_ModelicaParser_Regular_elementCo_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Regular_elementCo_4, __pyx_k_ModelicaParser_Regular_elementCo_4, sizeof(__pyx_k_ModelicaParser_Regular_elementCo_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Regular_elementCo_5, __pyx_k_ModelicaParser_Regular_elementCo_5, sizeof(__pyx_k_ModelicaParser_Regular_elementCo_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Regular_elementCo_6, __pyx_k_ModelicaParser_Regular_elementCo_6, sizeof(__pyx_k_ModelicaParser_Regular_elementCo_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Regular_elementCo_7, __pyx_k_ModelicaParser_Regular_elementCo_7, sizeof(__pyx_k_ModelicaParser_Regular_elementCo_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Regular_elementCo_8, __pyx_k_ModelicaParser_Regular_elementCo_8, sizeof(__pyx_k_ModelicaParser_Regular_elementCo_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Regular_elementCo_9, __pyx_k_ModelicaParser_Regular_elementCo_9, sizeof(__pyx_k_ModelicaParser_Regular_elementCo_9), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Replaceable_eleme, __pyx_k_ModelicaParser_Replaceable_eleme, sizeof(__pyx_k_ModelicaParser_Replaceable_eleme), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Replaceable_eleme_10, __pyx_k_ModelicaParser_Replaceable_eleme_10, sizeof(__pyx_k_ModelicaParser_Replaceable_eleme_10), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Replaceable_eleme_11, __pyx_k_ModelicaParser_Replaceable_eleme_11, sizeof(__pyx_k_ModelicaParser_Replaceable_eleme_11), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Replaceable_eleme_12, __pyx_k_ModelicaParser_Replaceable_eleme_12, sizeof(__pyx_k_ModelicaParser_Replaceable_eleme_12), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Replaceable_eleme_13, __pyx_k_ModelicaParser_Replaceable_eleme_13, sizeof(__pyx_k_ModelicaParser_Replaceable_eleme_13), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Replaceable_eleme_14, __pyx_k_ModelicaParser_Replaceable_eleme_14, sizeof(__pyx_k_ModelicaParser_Replaceable_eleme_14), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Replaceable_eleme_2, __pyx_k_ModelicaParser_Replaceable_eleme_2, sizeof(__pyx_k_ModelicaParser_Replaceable_eleme_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Replaceable_eleme_3, __pyx_k_ModelicaParser_Replaceable_eleme_3, sizeof(__pyx_k_ModelicaParser_Replaceable_eleme_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Replaceable_eleme_4, __pyx_k_ModelicaParser_Replaceable_eleme_4, sizeof(__pyx_k_ModelicaParser_Replaceable_eleme_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Replaceable_eleme_5, __pyx_k_ModelicaParser_Replaceable_eleme_5, sizeof(__pyx_k_ModelicaParser_Replaceable_eleme_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Replaceable_eleme_6, __pyx_k_ModelicaParser_Replaceable_eleme_6, sizeof(__pyx_k_ModelicaParser_Replaceable_eleme_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Replaceable_eleme_7, __pyx_k_ModelicaParser_Replaceable_eleme_7, sizeof(__pyx_k_ModelicaParser_Replaceable_eleme_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Replaceable_eleme_8, __pyx_k_ModelicaParser_Replaceable_eleme_8, sizeof(__pyx_k_ModelicaParser_Replaceable_eleme_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Replaceable_eleme_9, __pyx_k_ModelicaParser_Replaceable_eleme_9, sizeof(__pyx_k_ModelicaParser_Replaceable_eleme_9), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Short_class_defin, __pyx_k_ModelicaParser_Short_class_defin, sizeof(__pyx_k_ModelicaParser_Short_class_defin), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Short_class_defin_10, __pyx_k_ModelicaParser_Short_class_defin_10, sizeof(__pyx_k_ModelicaParser_Short_class_defin_10), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Short_class_defin_11, __pyx_k_ModelicaParser_Short_class_defin_11, sizeof(__pyx_k_ModelicaParser_Short_class_defin_11), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Short_class_defin_12, __pyx_k_ModelicaParser_Short_class_defin_12, sizeof(__pyx_k_ModelicaParser_Short_class_defin_12), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Short_class_defin_13, __pyx_k_ModelicaParser_Short_class_defin_13, sizeof(__pyx_k_ModelicaParser_Short_class_defin_13), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Short_class_defin_14, __pyx_k_ModelicaParser_Short_class_defin_14, sizeof(__pyx_k_ModelicaParser_Short_class_defin_14), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Short_class_defin_2, __pyx_k_ModelicaParser_Short_class_defin_2, sizeof(__pyx_k_ModelicaParser_Short_class_defin_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Short_class_defin_3, __pyx_k_ModelicaParser_Short_class_defin_3, sizeof(__pyx_k_ModelicaParser_Short_class_defin_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Short_class_defin_4, __pyx_k_ModelicaParser_Short_class_defin_4, sizeof(__pyx_k_ModelicaParser_Short_class_defin_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Short_class_defin_5, __pyx_k_ModelicaParser_Short_class_defin_5, sizeof(__pyx_k_ModelicaParser_Short_class_defin_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Short_class_defin_6, __pyx_k_ModelicaParser_Short_class_defin_6, sizeof(__pyx_k_ModelicaParser_Short_class_defin_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Short_class_defin_7, __pyx_k_ModelicaParser_Short_class_defin_7, sizeof(__pyx_k_ModelicaParser_Short_class_defin_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Short_class_defin_8, __pyx_k_ModelicaParser_Short_class_defin_8, sizeof(__pyx_k_ModelicaParser_Short_class_defin_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Short_class_defin_9, __pyx_k_ModelicaParser_Short_class_defin_9, sizeof(__pyx_k_ModelicaParser_Short_class_defin_9), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Simple_expression, __pyx_k_ModelicaParser_Simple_expression, sizeof(__pyx_k_ModelicaParser_Simple_expression), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Simple_expression_2, __pyx_k_ModelicaParser_Simple_expression_2, sizeof(__pyx_k_ModelicaParser_Simple_expression_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Simple_expression_3, __pyx_k_ModelicaParser_Simple_expression_3, sizeof(__pyx_k_ModelicaParser_Simple_expression_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Simple_expression_4, __pyx_k_ModelicaParser_Simple_expression_4, sizeof(__pyx_k_ModelicaParser_Simple_expression_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Simple_expression_5, __pyx_k_ModelicaParser_Simple_expression_5, sizeof(__pyx_k_ModelicaParser_Simple_expression_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Simple_expression_6, __pyx_k_ModelicaParser_Simple_expression_6, sizeof(__pyx_k_ModelicaParser_Simple_expression_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Simple_expression_7, __pyx_k_ModelicaParser_Simple_expression_7, sizeof(__pyx_k_ModelicaParser_Simple_expression_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_StatementContext, __pyx_k_ModelicaParser_StatementContext, sizeof(__pyx_k_ModelicaParser_StatementContext), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_StatementContext_2, __pyx_k_ModelicaParser_StatementContext_2, sizeof(__pyx_k_ModelicaParser_StatementContext_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_StatementContext_3, __pyx_k_ModelicaParser_StatementContext_3, sizeof(__pyx_k_ModelicaParser_StatementContext_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_StatementContext_4, __pyx_k_ModelicaParser_StatementContext_4, sizeof(__pyx_k_ModelicaParser_StatementContext_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_StatementContext_5, __pyx_k_ModelicaParser_StatementContext_5, sizeof(__pyx_k_ModelicaParser_StatementContext_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_StatementContext_6, __pyx_k_ModelicaParser_StatementContext_6, sizeof(__pyx_k_ModelicaParser_StatementContext_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_StatementContext_7, __pyx_k_ModelicaParser_StatementContext_7, sizeof(__pyx_k_ModelicaParser_StatementContext_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_StatementContext_8, __pyx_k_ModelicaParser_StatementContext_8, sizeof(__pyx_k_ModelicaParser_StatementContext_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_breakCo, __pyx_k_ModelicaParser_Statement_breakCo, sizeof(__pyx_k_ModelicaParser_Statement_breakCo), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_breakCo_2, __pyx_k_ModelicaParser_Statement_breakCo_2, sizeof(__pyx_k_ModelicaParser_Statement_breakCo_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_breakCo_3, __pyx_k_ModelicaParser_Statement_breakCo_3, sizeof(__pyx_k_ModelicaParser_Statement_breakCo_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_breakCo_4, __pyx_k_ModelicaParser_Statement_breakCo_4, sizeof(__pyx_k_ModelicaParser_Statement_breakCo_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_breakCo_5, __pyx_k_ModelicaParser_Statement_breakCo_5, sizeof(__pyx_k_ModelicaParser_Statement_breakCo_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_compone, __pyx_k_ModelicaParser_Statement_compone, sizeof(__pyx_k_ModelicaParser_Statement_compone), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_compone_10, __pyx_k_ModelicaParser_Statement_compone_10, sizeof(__pyx_k_ModelicaParser_Statement_compone_10), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_compone_11, __pyx_k_ModelicaParser_Statement_compone_11, sizeof(__pyx_k_ModelicaParser_Statement_compone_11), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_compone_12, __pyx_k_ModelicaParser_Statement_compone_12, sizeof(__pyx_k_ModelicaParser_Statement_compone_12), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_compone_13, __pyx_k_ModelicaParser_Statement_compone_13, sizeof(__pyx_k_ModelicaParser_Statement_compone_13), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_compone_14, __pyx_k_ModelicaParser_Statement_compone_14, sizeof(__pyx_k_ModelicaParser_Statement_compone_14), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_compone_15, __pyx_k_ModelicaParser_Statement_compone_15, sizeof(__pyx_k_ModelicaParser_Statement_compone_15), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_compone_2, __pyx_k_ModelicaParser_Statement_compone_2, sizeof(__pyx_k_ModelicaParser_Statement_compone_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_compone_3, __pyx_k_ModelicaParser_Statement_compone_3, sizeof(__pyx_k_ModelicaParser_Statement_compone_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_compone_4, __pyx_k_ModelicaParser_Statement_compone_4, sizeof(__pyx_k_ModelicaParser_Statement_compone_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_compone_5, __pyx_k_ModelicaParser_Statement_compone_5, sizeof(__pyx_k_ModelicaParser_Statement_compone_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_compone_6, __pyx_k_ModelicaParser_Statement_compone_6, sizeof(__pyx_k_ModelicaParser_Statement_compone_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_compone_7, __pyx_k_ModelicaParser_Statement_compone_7, sizeof(__pyx_k_ModelicaParser_Statement_compone_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_compone_8, __pyx_k_ModelicaParser_Statement_compone_8, sizeof(__pyx_k_ModelicaParser_Statement_compone_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_compone_9, __pyx_k_ModelicaParser_Statement_compone_9, sizeof(__pyx_k_ModelicaParser_Statement_compone_9), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_forCont, __pyx_k_ModelicaParser_Statement_forCont, sizeof(__pyx_k_ModelicaParser_Statement_forCont), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_forCont_2, __pyx_k_ModelicaParser_Statement_forCont_2, sizeof(__pyx_k_ModelicaParser_Statement_forCont_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_forCont_3, __pyx_k_ModelicaParser_Statement_forCont_3, sizeof(__pyx_k_ModelicaParser_Statement_forCont_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_forCont_4, __pyx_k_ModelicaParser_Statement_forCont_4, sizeof(__pyx_k_ModelicaParser_Statement_forCont_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_forCont_5, __pyx_k_ModelicaParser_Statement_forCont_5, sizeof(__pyx_k_ModelicaParser_Statement_forCont_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_forCont_6, __pyx_k_ModelicaParser_Statement_forCont_6, sizeof(__pyx_k_ModelicaParser_Statement_forCont_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_ifConte, __pyx_k_ModelicaParser_Statement_ifConte, sizeof(__pyx_k_ModelicaParser_Statement_ifConte), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_ifConte_2, __pyx_k_ModelicaParser_Statement_ifConte_2, sizeof(__pyx_k_ModelicaParser_Statement_ifConte_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_ifConte_3, __pyx_k_ModelicaParser_Statement_ifConte_3, sizeof(__pyx_k_ModelicaParser_Statement_ifConte_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_ifConte_4, __pyx_k_ModelicaParser_Statement_ifConte_4, sizeof(__pyx_k_ModelicaParser_Statement_ifConte_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_ifConte_5, __pyx_k_ModelicaParser_Statement_ifConte_5, sizeof(__pyx_k_ModelicaParser_Statement_ifConte_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_ifConte_6, __pyx_k_ModelicaParser_Statement_ifConte_6, sizeof(__pyx_k_ModelicaParser_Statement_ifConte_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_options, __pyx_k_ModelicaParser_Statement_options, sizeof(__pyx_k_ModelicaParser_Statement_options), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_options_2, __pyx_k_ModelicaParser_Statement_options_2, sizeof(__pyx_k_ModelicaParser_Statement_options_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_options_3, __pyx_k_ModelicaParser_Statement_options_3, sizeof(__pyx_k_ModelicaParser_Statement_options_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_options_4, __pyx_k_ModelicaParser_Statement_options_4, sizeof(__pyx_k_ModelicaParser_Statement_options_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_returnC, __pyx_k_ModelicaParser_Statement_returnC, sizeof(__pyx_k_ModelicaParser_Statement_returnC), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_returnC_2, __pyx_k_ModelicaParser_Statement_returnC_2, sizeof(__pyx_k_ModelicaParser_Statement_returnC_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_returnC_3, __pyx_k_ModelicaParser_Statement_returnC_3, sizeof(__pyx_k_ModelicaParser_Statement_returnC_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_returnC_4, __pyx_k_ModelicaParser_Statement_returnC_4, sizeof(__pyx_k_ModelicaParser_Statement_returnC_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_returnC_5, __pyx_k_ModelicaParser_Statement_returnC_5, sizeof(__pyx_k_ModelicaParser_Statement_returnC_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_whenCon, __pyx_k_ModelicaParser_Statement_whenCon, sizeof(__pyx_k_ModelicaParser_Statement_whenCon), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_whenCon_2, __pyx_k_ModelicaParser_Statement_whenCon_2, sizeof(__pyx_k_ModelicaParser_Statement_whenCon_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_whenCon_3, __pyx_k_ModelicaParser_Statement_whenCon_3, sizeof(__pyx_k_ModelicaParser_Statement_whenCon_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_whenCon_4, __pyx_k_ModelicaParser_Statement_whenCon_4, sizeof(__pyx_k_ModelicaParser_Statement_whenCon_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_whenCon_5, __pyx_k_ModelicaParser_Statement_whenCon_5, sizeof(__pyx_k_ModelicaParser_Statement_whenCon_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_whenCon_6, __pyx_k_ModelicaParser_Statement_whenCon_6, sizeof(__pyx_k_ModelicaParser_Statement_whenCon_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_whileCo, __pyx_k_ModelicaParser_Statement_whileCo, sizeof(__pyx_k_ModelicaParser_Statement_whileCo), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_whileCo_2, __pyx_k_ModelicaParser_Statement_whileCo_2, sizeof(__pyx_k_ModelicaParser_Statement_whileCo_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_whileCo_3, __pyx_k_ModelicaParser_Statement_whileCo_3, sizeof(__pyx_k_ModelicaParser_Statement_whileCo_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_whileCo_4, __pyx_k_ModelicaParser_Statement_whileCo_4, sizeof(__pyx_k_ModelicaParser_Statement_whileCo_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_whileCo_5, __pyx_k_ModelicaParser_Statement_whileCo_5, sizeof(__pyx_k_ModelicaParser_Statement_whileCo_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Statement_whileCo_6, __pyx_k_ModelicaParser_Statement_whileCo_6, sizeof(__pyx_k_ModelicaParser_Statement_whileCo_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Stored_definition, __pyx_k_ModelicaParser_Stored_definition, sizeof(__pyx_k_ModelicaParser_Stored_definition), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Stored_definition_10, __pyx_k_ModelicaParser_Stored_definition_10, sizeof(__pyx_k_ModelicaParser_Stored_definition_10), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Stored_definition_11, __pyx_k_ModelicaParser_Stored_definition_11, sizeof(__pyx_k_ModelicaParser_Stored_definition_11), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Stored_definition_12, __pyx_k_ModelicaParser_Stored_definition_12, sizeof(__pyx_k_ModelicaParser_Stored_definition_12), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Stored_definition_13, __pyx_k_ModelicaParser_Stored_definition_13, sizeof(__pyx_k_ModelicaParser_Stored_definition_13), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Stored_definition_14, __pyx_k_ModelicaParser_Stored_definition_14, sizeof(__pyx_k_ModelicaParser_Stored_definition_14), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Stored_definition_15, __pyx_k_ModelicaParser_Stored_definition_15, sizeof(__pyx_k_ModelicaParser_Stored_definition_15), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Stored_definition_16, __pyx_k_ModelicaParser_Stored_definition_16, sizeof(__pyx_k_ModelicaParser_Stored_definition_16), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Stored_definition_17, __pyx_k_ModelicaParser_Stored_definition_17, sizeof(__pyx_k_ModelicaParser_Stored_definition_17), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Stored_definition_2, __pyx_k_ModelicaParser_Stored_definition_2, sizeof(__pyx_k_ModelicaParser_Stored_definition_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Stored_definition_3, __pyx_k_ModelicaParser_Stored_definition_3, sizeof(__pyx_k_ModelicaParser_Stored_definition_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Stored_definition_4, __pyx_k_ModelicaParser_Stored_definition_4, sizeof(__pyx_k_ModelicaParser_Stored_definition_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Stored_definition_5, __pyx_k_ModelicaParser_Stored_definition_5, sizeof(__pyx_k_ModelicaParser_Stored_definition_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Stored_definition_6, __pyx_k_ModelicaParser_Stored_definition_6, sizeof(__pyx_k_ModelicaParser_Stored_definition_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Stored_definition_7, __pyx_k_ModelicaParser_Stored_definition_7, sizeof(__pyx_k_ModelicaParser_Stored_definition_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Stored_definition_8, __pyx_k_ModelicaParser_Stored_definition_8, sizeof(__pyx_k_ModelicaParser_Stored_definition_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Stored_definition_9, __pyx_k_ModelicaParser_Stored_definition_9, sizeof(__pyx_k_ModelicaParser_Stored_definition_9), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_String_commentCon, __pyx_k_ModelicaParser_String_commentCon, sizeof(__pyx_k_ModelicaParser_String_commentCon), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_String_commentCon_2, __pyx_k_ModelicaParser_String_commentCon_2, sizeof(__pyx_k_ModelicaParser_String_commentCon_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_String_commentCon_3, __pyx_k_ModelicaParser_String_commentCon_3, sizeof(__pyx_k_ModelicaParser_String_commentCon_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_String_commentCon_4, __pyx_k_ModelicaParser_String_commentCon_4, sizeof(__pyx_k_ModelicaParser_String_commentCon_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_String_commentCon_5, __pyx_k_ModelicaParser_String_commentCon_5, sizeof(__pyx_k_ModelicaParser_String_commentCon_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_String_commentCon_6, __pyx_k_ModelicaParser_String_commentCon_6, sizeof(__pyx_k_ModelicaParser_String_commentCon_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_String_commentCon_7, __pyx_k_ModelicaParser_String_commentCon_7, sizeof(__pyx_k_ModelicaParser_String_commentCon_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_SubscriptContext, __pyx_k_ModelicaParser_SubscriptContext, sizeof(__pyx_k_ModelicaParser_SubscriptContext), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_SubscriptContext_2, __pyx_k_ModelicaParser_SubscriptContext_2, sizeof(__pyx_k_ModelicaParser_SubscriptContext_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_SubscriptContext_3, __pyx_k_ModelicaParser_SubscriptContext_3, sizeof(__pyx_k_ModelicaParser_SubscriptContext_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_SubscriptContext_4, __pyx_k_ModelicaParser_SubscriptContext_4, sizeof(__pyx_k_ModelicaParser_SubscriptContext_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_SubscriptContext_5, __pyx_k_ModelicaParser_SubscriptContext_5, sizeof(__pyx_k_ModelicaParser_SubscriptContext_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_SubscriptContext_6, __pyx_k_ModelicaParser_SubscriptContext_6, sizeof(__pyx_k_ModelicaParser_SubscriptContext_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_SubscriptContext_7, __pyx_k_ModelicaParser_SubscriptContext_7, sizeof(__pyx_k_ModelicaParser_SubscriptContext_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Type_prefixContex, __pyx_k_ModelicaParser_Type_prefixContex, sizeof(__pyx_k_ModelicaParser_Type_prefixContex), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Type_prefixContex_2, __pyx_k_ModelicaParser_Type_prefixContex_2, sizeof(__pyx_k_ModelicaParser_Type_prefixContex_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Type_prefixContex_3, __pyx_k_ModelicaParser_Type_prefixContex_3, sizeof(__pyx_k_ModelicaParser_Type_prefixContex_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Type_prefixContex_4, __pyx_k_ModelicaParser_Type_prefixContex_4, sizeof(__pyx_k_ModelicaParser_Type_prefixContex_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Type_prefixContex_5, __pyx_k_ModelicaParser_Type_prefixContex_5, sizeof(__pyx_k_ModelicaParser_Type_prefixContex_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Type_prefixContex_6, __pyx_k_ModelicaParser_Type_prefixContex_6, sizeof(__pyx_k_ModelicaParser_Type_prefixContex_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Type_specifierCon, __pyx_k_ModelicaParser_Type_specifierCon, sizeof(__pyx_k_ModelicaParser_Type_specifierCon), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Type_specifierCon_2, __pyx_k_ModelicaParser_Type_specifierCon_2, sizeof(__pyx_k_ModelicaParser_Type_specifierCon_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Type_specifierCon_3, __pyx_k_ModelicaParser_Type_specifierCon_3, sizeof(__pyx_k_ModelicaParser_Type_specifierCon_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Type_specifierCon_4, __pyx_k_ModelicaParser_Type_specifierCon_4, sizeof(__pyx_k_ModelicaParser_Type_specifierCon_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Type_specifierCon_5, __pyx_k_ModelicaParser_Type_specifierCon_5, sizeof(__pyx_k_ModelicaParser_Type_specifierCon_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Type_specifierCon_6, __pyx_k_ModelicaParser_Type_specifierCon_6, sizeof(__pyx_k_ModelicaParser_Type_specifierCon_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_Type_specifierCon_7, __pyx_k_ModelicaParser_Type_specifierCon_7, sizeof(__pyx_k_ModelicaParser_Type_specifierCon_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_When_equationCont, __pyx_k_ModelicaParser_When_equationCont, sizeof(__pyx_k_ModelicaParser_When_equationCont), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_When_equationCont_2, __pyx_k_ModelicaParser_When_equationCont_2, sizeof(__pyx_k_ModelicaParser_When_equationCont_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_When_equationCont_3, __pyx_k_ModelicaParser_When_equationCont_3, sizeof(__pyx_k_ModelicaParser_When_equationCont_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_When_equationCont_4, __pyx_k_ModelicaParser_When_equationCont_4, sizeof(__pyx_k_ModelicaParser_When_equationCont_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_When_equationCont_5, __pyx_k_ModelicaParser_When_equationCont_5, sizeof(__pyx_k_ModelicaParser_When_equationCont_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_When_equationCont_6, __pyx_k_ModelicaParser_When_equationCont_6, sizeof(__pyx_k_ModelicaParser_When_equationCont_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_When_equationCont_7, __pyx_k_ModelicaParser_When_equationCont_7, sizeof(__pyx_k_ModelicaParser_When_equationCont_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_When_equationCont_8, __pyx_k_ModelicaParser_When_equationCont_8, sizeof(__pyx_k_ModelicaParser_When_equationCont_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_When_statementCon, __pyx_k_ModelicaParser_When_statementCon, sizeof(__pyx_k_ModelicaParser_When_statementCon), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_When_statementCon_2, __pyx_k_ModelicaParser_When_statementCon_2, sizeof(__pyx_k_ModelicaParser_When_statementCon_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_When_statementCon_3, __pyx_k_ModelicaParser_When_statementCon_3, sizeof(__pyx_k_ModelicaParser_When_statementCon_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_When_statementCon_4, __pyx_k_ModelicaParser_When_statementCon_4, sizeof(__pyx_k_ModelicaParser_When_statementCon_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_When_statementCon_5, __pyx_k_ModelicaParser_When_statementCon_5, sizeof(__pyx_k_ModelicaParser_When_statementCon_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_When_statementCon_6, __pyx_k_ModelicaParser_When_statementCon_6, sizeof(__pyx_k_ModelicaParser_When_statementCon_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_When_statementCon_7, __pyx_k_ModelicaParser_When_statementCon_7, sizeof(__pyx_k_ModelicaParser_When_statementCon_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_When_statementCon_8, __pyx_k_ModelicaParser_When_statementCon_8, sizeof(__pyx_k_ModelicaParser_When_statementCon_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_While_statementCo, __pyx_k_ModelicaParser_While_statementCo, sizeof(__pyx_k_ModelicaParser_While_statementCo), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_While_statementCo_2, __pyx_k_ModelicaParser_While_statementCo_2, sizeof(__pyx_k_ModelicaParser_While_statementCo_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_While_statementCo_3, __pyx_k_ModelicaParser_While_statementCo_3, sizeof(__pyx_k_ModelicaParser_While_statementCo_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_While_statementCo_4, __pyx_k_ModelicaParser_While_statementCo_4, sizeof(__pyx_k_ModelicaParser_While_statementCo_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_While_statementCo_5, __pyx_k_ModelicaParser_While_statementCo_5, sizeof(__pyx_k_ModelicaParser_While_statementCo_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_While_statementCo_6, __pyx_k_ModelicaParser_While_statementCo_6, sizeof(__pyx_k_ModelicaParser_While_statementCo_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_While_statementCo_7, __pyx_k_ModelicaParser_While_statementCo_7, sizeof(__pyx_k_ModelicaParser_While_statementCo_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_While_statementCo_8, __pyx_k_ModelicaParser_While_statementCo_8, sizeof(__pyx_k_ModelicaParser_While_statementCo_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser___init, __pyx_k_ModelicaParser___init, sizeof(__pyx_k_ModelicaParser___init), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_algorithm_section, __pyx_k_ModelicaParser_algorithm_section, sizeof(__pyx_k_ModelicaParser_algorithm_section), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_annotation, __pyx_k_ModelicaParser_annotation, sizeof(__pyx_k_ModelicaParser_annotation), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_argument, __pyx_k_ModelicaParser_argument, sizeof(__pyx_k_ModelicaParser_argument), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_argument_list, __pyx_k_ModelicaParser_argument_list, sizeof(__pyx_k_ModelicaParser_argument_list), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_array_subscripts, __pyx_k_ModelicaParser_array_subscripts, sizeof(__pyx_k_ModelicaParser_array_subscripts), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_base_prefix, __pyx_k_ModelicaParser_base_prefix, sizeof(__pyx_k_ModelicaParser_base_prefix), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_class_definition, __pyx_k_ModelicaParser_class_definition, sizeof(__pyx_k_ModelicaParser_class_definition), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_class_modificatio, __pyx_k_ModelicaParser_class_modificatio, sizeof(__pyx_k_ModelicaParser_class_modificatio), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_class_prefixes, __pyx_k_ModelicaParser_class_prefixes, sizeof(__pyx_k_ModelicaParser_class_prefixes), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_class_specifier, __pyx_k_ModelicaParser_class_specifier, sizeof(__pyx_k_ModelicaParser_class_specifier), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_class_type, __pyx_k_ModelicaParser_class_type, sizeof(__pyx_k_ModelicaParser_class_type), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_comment, __pyx_k_ModelicaParser_comment, sizeof(__pyx_k_ModelicaParser_comment), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_component_clause, __pyx_k_ModelicaParser_component_clause, sizeof(__pyx_k_ModelicaParser_component_clause), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_component_clause1, __pyx_k_ModelicaParser_component_clause1, sizeof(__pyx_k_ModelicaParser_component_clause1), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_component_declara, __pyx_k_ModelicaParser_component_declara, sizeof(__pyx_k_ModelicaParser_component_declara), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_component_declara_2, __pyx_k_ModelicaParser_component_declara_2, sizeof(__pyx_k_ModelicaParser_component_declara_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_component_list, __pyx_k_ModelicaParser_component_list, sizeof(__pyx_k_ModelicaParser_component_list), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_component_referen, __pyx_k_ModelicaParser_component_referen, sizeof(__pyx_k_ModelicaParser_component_referen), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_component_referen_2, __pyx_k_ModelicaParser_component_referen_2, sizeof(__pyx_k_ModelicaParser_component_referen_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_composition, __pyx_k_ModelicaParser_composition, sizeof(__pyx_k_ModelicaParser_composition), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_condition_attribu, __pyx_k_ModelicaParser_condition_attribu, sizeof(__pyx_k_ModelicaParser_condition_attribu), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_connect_clause, __pyx_k_ModelicaParser_connect_clause, sizeof(__pyx_k_ModelicaParser_connect_clause), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_constraining_clau, __pyx_k_ModelicaParser_constraining_clau, sizeof(__pyx_k_ModelicaParser_constraining_clau), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_declaration, __pyx_k_ModelicaParser_declaration, sizeof(__pyx_k_ModelicaParser_declaration), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_element, __pyx_k_ModelicaParser_element, sizeof(__pyx_k_ModelicaParser_element), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_element_list, __pyx_k_ModelicaParser_element_list, sizeof(__pyx_k_ModelicaParser_element_list), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_element_modificat, __pyx_k_ModelicaParser_element_modificat, sizeof(__pyx_k_ModelicaParser_element_modificat), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_element_modificat_2, __pyx_k_ModelicaParser_element_modificat_2, sizeof(__pyx_k_ModelicaParser_element_modificat_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_element_redeclara, __pyx_k_ModelicaParser_element_redeclara, sizeof(__pyx_k_ModelicaParser_element_redeclara), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_element_replaceab, __pyx_k_ModelicaParser_element_replaceab, sizeof(__pyx_k_ModelicaParser_element_replaceab), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_enum_list, __pyx_k_ModelicaParser_enum_list, sizeof(__pyx_k_ModelicaParser_enum_list), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_enumeration_liter, __pyx_k_ModelicaParser_enumeration_liter, sizeof(__pyx_k_ModelicaParser_enumeration_liter), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_equation, __pyx_k_ModelicaParser_equation, sizeof(__pyx_k_ModelicaParser_equation), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_equation_options, __pyx_k_ModelicaParser_equation_options, sizeof(__pyx_k_ModelicaParser_equation_options), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_equation_section, __pyx_k_ModelicaParser_equation_section, sizeof(__pyx_k_ModelicaParser_equation_section), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_expr, __pyx_k_ModelicaParser_expr, sizeof(__pyx_k_ModelicaParser_expr), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_expr_sempred, __pyx_k_ModelicaParser_expr_sempred, sizeof(__pyx_k_ModelicaParser_expr_sempred), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_expression, __pyx_k_ModelicaParser_expression, sizeof(__pyx_k_ModelicaParser_expression), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_expression_list, __pyx_k_ModelicaParser_expression_list, sizeof(__pyx_k_ModelicaParser_expression_list), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_extends_clause, __pyx_k_ModelicaParser_extends_clause, sizeof(__pyx_k_ModelicaParser_extends_clause), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_external_function, __pyx_k_ModelicaParser_external_function, sizeof(__pyx_k_ModelicaParser_external_function), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_for_equation, __pyx_k_ModelicaParser_for_equation, sizeof(__pyx_k_ModelicaParser_for_equation), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_for_index, __pyx_k_ModelicaParser_for_index, sizeof(__pyx_k_ModelicaParser_for_index), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_for_indices, __pyx_k_ModelicaParser_for_indices, sizeof(__pyx_k_ModelicaParser_for_indices), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_for_statement, __pyx_k_ModelicaParser_for_statement, sizeof(__pyx_k_ModelicaParser_for_statement), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_function_argument, __pyx_k_ModelicaParser_function_argument, sizeof(__pyx_k_ModelicaParser_function_argument), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_function_argument_2, __pyx_k_ModelicaParser_function_argument_2, sizeof(__pyx_k_ModelicaParser_function_argument_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_function_call_arg, __pyx_k_ModelicaParser_function_call_arg, sizeof(__pyx_k_ModelicaParser_function_call_arg), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_if_equation, __pyx_k_ModelicaParser_if_equation, sizeof(__pyx_k_ModelicaParser_if_equation), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_if_statement, __pyx_k_ModelicaParser_if_statement, sizeof(__pyx_k_ModelicaParser_if_statement), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_import_clause, __pyx_k_ModelicaParser_import_clause, sizeof(__pyx_k_ModelicaParser_import_clause), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_import_list, __pyx_k_ModelicaParser_import_list, sizeof(__pyx_k_ModelicaParser_import_list), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_language_specific, __pyx_k_ModelicaParser_language_specific, sizeof(__pyx_k_ModelicaParser_language_specific), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_modification, __pyx_k_ModelicaParser_modification, sizeof(__pyx_k_ModelicaParser_modification), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_name, __pyx_k_ModelicaParser_name, sizeof(__pyx_k_ModelicaParser_name), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_named_argument, __pyx_k_ModelicaParser_named_argument, sizeof(__pyx_k_ModelicaParser_named_argument), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_named_arguments, __pyx_k_ModelicaParser_named_arguments, sizeof(__pyx_k_ModelicaParser_named_arguments), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_output_expression, __pyx_k_ModelicaParser_output_expression, sizeof(__pyx_k_ModelicaParser_output_expression), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_primary, __pyx_k_ModelicaParser_primary, sizeof(__pyx_k_ModelicaParser_primary), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_regular_element, __pyx_k_ModelicaParser_regular_element, sizeof(__pyx_k_ModelicaParser_regular_element), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_replaceable_eleme, __pyx_k_ModelicaParser_replaceable_eleme, sizeof(__pyx_k_ModelicaParser_replaceable_eleme), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_sempred, __pyx_k_ModelicaParser_sempred, sizeof(__pyx_k_ModelicaParser_sempred), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_short_class_defin, __pyx_k_ModelicaParser_short_class_defin, sizeof(__pyx_k_ModelicaParser_short_class_defin), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_simple_expression, __pyx_k_ModelicaParser_simple_expression, sizeof(__pyx_k_ModelicaParser_simple_expression), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_statement, __pyx_k_ModelicaParser_statement, sizeof(__pyx_k_ModelicaParser_statement), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_statement_options, __pyx_k_ModelicaParser_statement_options, sizeof(__pyx_k_ModelicaParser_statement_options), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_stored_definition, __pyx_k_ModelicaParser_stored_definition, sizeof(__pyx_k_ModelicaParser_stored_definition), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_stored_definition_2, __pyx_k_ModelicaParser_stored_definition_2, sizeof(__pyx_k_ModelicaParser_stored_definition_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_string_comment, __pyx_k_ModelicaParser_string_comment, sizeof(__pyx_k_ModelicaParser_string_comment), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_subscript, __pyx_k_ModelicaParser_subscript, sizeof(__pyx_k_ModelicaParser_subscript), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_type_prefix, __pyx_k_ModelicaParser_type_prefix, sizeof(__pyx_k_ModelicaParser_type_prefix), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_type_specifier, __pyx_k_ModelicaParser_type_specifier, sizeof(__pyx_k_ModelicaParser_type_specifier), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_when_equation, __pyx_k_ModelicaParser_when_equation, sizeof(__pyx_k_ModelicaParser_when_equation), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_when_statement, __pyx_k_ModelicaParser_when_statement, sizeof(__pyx_k_ModelicaParser_when_statement), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser_while_statement, __pyx_k_ModelicaParser_while_statement, sizeof(__pyx_k_ModelicaParser_while_statement), 0, 0, 1, 1},
  {&__pyx_kp_u_Modelica_g4, __pyx_k_Modelica_g4, sizeof(__pyx_k_Modelica_g4), 0, 1, 0, 0},
  {&__pyx_n_s_ModificationContext, __pyx_k_ModificationContext, sizeof(__pyx_k_ModificationContext), 0, 0, 1, 1},
  {&__pyx_n_s_Modification_assignment2Context, __pyx_k_Modification_assignment2Context, sizeof(__pyx_k_Modification_assignment2Context), 0, 0, 1, 1},
  {&__pyx_n_s_Modification_assignmentContext, __pyx_k_Modification_assignmentContext, sizeof(__pyx_k_Modification_assignmentContext), 0, 0, 1, 1},
  {&__pyx_n_s_Modification_classContext, __pyx_k_Modification_classContext, sizeof(__pyx_k_Modification_classContext), 0, 0, 1, 1},
  {&__pyx_kp_u_N, __pyx_k_N, sizeof(__pyx_k_N), 0, 1, 0, 0},
  {&__pyx_kp_u_N_2, __pyx_k_N_2, sizeof(__pyx_k_N_2), 0, 1, 0, 0},
  {&__pyx_kp_u_N_G, __pyx_k_N_G, sizeof(__pyx_k_N_G), 0, 1, 0, 0},
  {&__pyx_n_s_NameContext, __pyx_k_NameContext, sizeof(__pyx_k_NameContext), 0, 0, 1, 1},
  {&__pyx_n_s_Named_argumentContext, __pyx_k_Named_argumentContext, sizeof(__pyx_k_Named_argumentContext), 0, 0, 1, 1},
  {&__pyx_n_s_Named_argumentsContext, __pyx_k_Named_argumentsContext, sizeof(__pyx_k_Named_argumentsContext), 0, 0, 1, 1},
  {&__pyx_n_s_NoViableAltException, __pyx_k_NoViableAltException, sizeof(__pyx_k_NoViableAltException), 0, 0, 1, 1},
  {&__pyx_kp_u_No_predicate_with_index, __pyx_k_No_predicate_with_index, sizeof(__pyx_k_No_predicate_with_index), 0, 1, 0, 0},
  {&__pyx_kp_u_O, __pyx_k_O, sizeof(__pyx_k_O), 0, 1, 0, 0},
  {&__pyx_n_s_OUTER, __pyx_k_OUTER, sizeof(__pyx_k_OUTER), 0, 0, 1, 1},
  {&__pyx_n_u_OUTER, __pyx_k_OUTER, sizeof(__pyx_k_OUTER), 0, 1, 0, 1},
  {&__pyx_kp_u_O_2, __pyx_k_O_2, sizeof(__pyx_k_O_2), 0, 1, 0, 0},
  {&__pyx_kp_u_O_3, __pyx_k_O_3, sizeof(__pyx_k_O_3), 0, 1, 0, 0},
  {&__pyx_n_s_Output_expression_listContext, __pyx_k_Output_expression_listContext, sizeof(__pyx_k_Output_expression_listContext), 0, 0, 1, 1},
  {&__pyx_kp_u_P, __pyx_k_P, sizeof(__pyx_k_P), 0, 1, 0, 0},
  {&__pyx_n_s_PARTIAL, __pyx_k_PARTIAL, sizeof(__pyx_k_PARTIAL), 0, 0, 1, 1},
  {&__pyx_n_u_PARTIAL, __pyx_k_PARTIAL, sizeof(__pyx_k_PARTIAL), 0, 1, 0, 1},
  {&__pyx_kp_u_P_2, __pyx_k_P_2, sizeof(__pyx_k_P_2), 0, 1, 0, 0},
  {&__pyx_n_s_ParseTreeListener, __pyx_k_ParseTreeListener, sizeof(__pyx_k_ParseTreeListener), 0, 0, 1, 1},
  {&__pyx_n_s_ParseTreeVisitor, __pyx_k_ParseTreeVisitor, sizeof(__pyx_k_ParseTreeVisitor), 0, 0, 1, 1},
  {&__pyx_n_s_Parser, __pyx_k_Parser, sizeof(__pyx_k_Parser), 0, 0, 1, 1},
  {&__pyx_n_s_ParserATNSimulator, __pyx_k_ParserATNSimulator, sizeof(__pyx_k_ParserATNSimulator), 0, 0, 1, 1},
  {&__pyx_n_s_ParserRuleContext, __pyx_k_ParserRuleContext, sizeof(__pyx_k_ParserRuleContext), 0, 0, 1, 1},
  {&__pyx_n_s_PredictionContextCache, __pyx_k_PredictionContextCache, sizeof(__pyx_k_PredictionContextCache), 0, 0, 1, 1},
  {&__pyx_n_s_PrimaryContext, __pyx_k_PrimaryContext, sizeof(__pyx_k_PrimaryContext), 0, 0, 1, 1},
  {&__pyx_n_s_Primary_component_referenceConte, __pyx_k_Primary_component_referenceConte, sizeof(__pyx_k_Primary_component_referenceConte), 0, 0, 1, 1},
  {&__pyx_n_s_Primary_derivativeContext, __pyx_k_Primary_derivativeContext, sizeof(__pyx_k_Primary_derivativeContext), 0, 0, 1, 1},
  {&__pyx_n_s_Primary_endContext, __pyx_k_Primary_endContext, sizeof(__pyx_k_Primary_endContext), 0, 0, 1, 1},
  {&__pyx_n_s_Primary_expression_listContext, __pyx_k_Primary_expression_listContext, sizeof(__pyx_k_Primary_expression_listContext), 0, 0, 1, 1},
  {&__pyx_n_s_Primary_falseContext, __pyx_k_Primary_falseContext, sizeof(__pyx_k_Primary_falseContext), 0, 0, 1, 1},
  {&__pyx_n_s_Primary_functionContext, __pyx_k_Primary_functionContext, sizeof(__pyx_k_Primary_functionContext), 0, 0, 1, 1},
  {&__pyx_n_s_Primary_function_argumentsContex, __pyx_k_Primary_function_argumentsContex, sizeof(__pyx_k_Primary_function_argumentsContex), 0, 0, 1, 1},
  {&__pyx_n_s_Primary_initialContext, __pyx_k_Primary_initialContext, sizeof(__pyx_k_Primary_initialContext), 0, 0, 1, 1},
  {&__pyx_n_s_Primary_output_expression_listCo, __pyx_k_Primary_output_expression_listCo, sizeof(__pyx_k_Primary_output_expression_listCo), 0, 0, 1, 1},
  {&__pyx_n_s_Primary_stringContext, __pyx_k_Primary_stringContext, sizeof(__pyx_k_Primary_stringContext), 0, 0, 1, 1},
  {&__pyx_n_s_Primary_trueContext, __pyx_k_Primary_trueContext, sizeof(__pyx_k_Primary_trueContext), 0, 0, 1, 1},
  {&__pyx_n_s_Primary_unsigned_numberContext, __pyx_k_Primary_unsigned_numberContext, sizeof(__pyx_k_Primary_unsigned_numberContext), 0, 0, 1, 1},
  {&__pyx_kp_u_Q, __pyx_k_Q, sizeof(__pyx_k_Q), 0, 1, 0, 0},
  {&__pyx_kp_u_Q_2, __pyx_k_Q_2, sizeof(__pyx_k_Q_2), 0, 1, 0, 0},
  {&__pyx_kp_u_R, __pyx_k_R, sizeof(__pyx_k_R), 0, 1, 0, 0},
  {&__pyx_n_s_REDECLARE, __pyx_k_REDECLARE, sizeof(__pyx_k_REDECLARE), 0, 0, 1, 1},
  {&__pyx_n_u_REDECLARE, __pyx_k_REDECLARE, sizeof(__pyx_k_REDECLARE), 0, 1, 0, 1},
  {&__pyx_n_s_RULE_algorithm_section, __pyx_k_RULE_algorithm_section, sizeof(__pyx_k_RULE_algorithm_section), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_annotation, __pyx_k_RULE_annotation, sizeof(__pyx_k_RULE_annotation), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_argument, __pyx_k_RULE_argument, sizeof(__pyx_k_RULE_argument), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_argument_list, __pyx_k_RULE_argument_list, sizeof(__pyx_k_RULE_argument_list), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_array_subscripts, __pyx_k_RULE_array_subscripts, sizeof(__pyx_k_RULE_array_subscripts), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_base_prefix, __pyx_k_RULE_base_prefix, sizeof(__pyx_k_RULE_base_prefix), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_class_definition, __pyx_k_RULE_class_definition, sizeof(__pyx_k_RULE_class_definition), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_class_modification, __pyx_k_RULE_class_modification, sizeof(__pyx_k_RULE_class_modification), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_class_prefixes, __pyx_k_RULE_class_prefixes, sizeof(__pyx_k_RULE_class_prefixes), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_class_specifier, __pyx_k_RULE_class_specifier, sizeof(__pyx_k_RULE_class_specifier), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_class_type, __pyx_k_RULE_class_type, sizeof(__pyx_k_RULE_class_type), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_comment, __pyx_k_RULE_comment, sizeof(__pyx_k_RULE_comment), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_component_clause, __pyx_k_RULE_component_clause, sizeof(__pyx_k_RULE_component_clause), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_component_clause1, __pyx_k_RULE_component_clause1, sizeof(__pyx_k_RULE_component_clause1), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_component_declaration, __pyx_k_RULE_component_declaration, sizeof(__pyx_k_RULE_component_declaration), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_component_declaration1, __pyx_k_RULE_component_declaration1, sizeof(__pyx_k_RULE_component_declaration1), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_component_list, __pyx_k_RULE_component_list, sizeof(__pyx_k_RULE_component_list), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_component_reference, __pyx_k_RULE_component_reference, sizeof(__pyx_k_RULE_component_reference), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_component_reference_element, __pyx_k_RULE_component_reference_element, sizeof(__pyx_k_RULE_component_reference_element), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_composition, __pyx_k_RULE_composition, sizeof(__pyx_k_RULE_composition), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_condition_attribute, __pyx_k_RULE_condition_attribute, sizeof(__pyx_k_RULE_condition_attribute), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_connect_clause, __pyx_k_RULE_connect_clause, sizeof(__pyx_k_RULE_connect_clause), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_constraining_clause, __pyx_k_RULE_constraining_clause, sizeof(__pyx_k_RULE_constraining_clause), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_declaration, __pyx_k_RULE_declaration, sizeof(__pyx_k_RULE_declaration), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_element, __pyx_k_RULE_element, sizeof(__pyx_k_RULE_element), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_element_list, __pyx_k_RULE_element_list, sizeof(__pyx_k_RULE_element_list), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_element_modification, __pyx_k_RULE_element_modification, sizeof(__pyx_k_RULE_element_modification), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_element_modification_or_rep, __pyx_k_RULE_element_modification_or_rep, sizeof(__pyx_k_RULE_element_modification_or_rep), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_element_redeclaration, __pyx_k_RULE_element_redeclaration, sizeof(__pyx_k_RULE_element_redeclaration), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_element_replaceable, __pyx_k_RULE_element_replaceable, sizeof(__pyx_k_RULE_element_replaceable), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_enum_list, __pyx_k_RULE_enum_list, sizeof(__pyx_k_RULE_enum_list), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_enumeration_literal, __pyx_k_RULE_enumeration_literal, sizeof(__pyx_k_RULE_enumeration_literal), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_equation, __pyx_k_RULE_equation, sizeof(__pyx_k_RULE_equation), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_equation_options, __pyx_k_RULE_equation_options, sizeof(__pyx_k_RULE_equation_options), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_equation_section, __pyx_k_RULE_equation_section, sizeof(__pyx_k_RULE_equation_section), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_expr, __pyx_k_RULE_expr, sizeof(__pyx_k_RULE_expr), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_expression, __pyx_k_RULE_expression, sizeof(__pyx_k_RULE_expression), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_expression_list, __pyx_k_RULE_expression_list, sizeof(__pyx_k_RULE_expression_list), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_extends_clause, __pyx_k_RULE_extends_clause, sizeof(__pyx_k_RULE_extends_clause), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_external_function_call, __pyx_k_RULE_external_function_call, sizeof(__pyx_k_RULE_external_function_call), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_for_equation, __pyx_k_RULE_for_equation, sizeof(__pyx_k_RULE_for_equation), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_for_index, __pyx_k_RULE_for_index, sizeof(__pyx_k_RULE_for_index), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_for_indices, __pyx_k_RULE_for_indices, sizeof(__pyx_k_RULE_for_indices), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_for_statement, __pyx_k_RULE_for_statement, sizeof(__pyx_k_RULE_for_statement), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_function_argument, __pyx_k_RULE_function_argument, sizeof(__pyx_k_RULE_function_argument), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_function_arguments, __pyx_k_RULE_function_arguments, sizeof(__pyx_k_RULE_function_arguments), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_function_call_args, __pyx_k_RULE_function_call_args, sizeof(__pyx_k_RULE_function_call_args), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_if_equation, __pyx_k_RULE_if_equation, sizeof(__pyx_k_RULE_if_equation), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_if_statement, __pyx_k_RULE_if_statement, sizeof(__pyx_k_RULE_if_statement), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_import_clause, __pyx_k_RULE_import_clause, sizeof(__pyx_k_RULE_import_clause), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_import_list, __pyx_k_RULE_import_list, sizeof(__pyx_k_RULE_import_list), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_language_specification, __pyx_k_RULE_language_specification, sizeof(__pyx_k_RULE_language_specification), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_modification, __pyx_k_RULE_modification, sizeof(__pyx_k_RULE_modification), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_name, __pyx_k_RULE_name, sizeof(__pyx_k_RULE_name), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_named_argument, __pyx_k_RULE_named_argument, sizeof(__pyx_k_RULE_named_argument), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_named_arguments, __pyx_k_RULE_named_arguments, sizeof(__pyx_k_RULE_named_arguments), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_output_expression_list, __pyx_k_RULE_output_expression_list, sizeof(__pyx_k_RULE_output_expression_list), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_primary, __pyx_k_RULE_primary, sizeof(__pyx_k_RULE_primary), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_regular_element, __pyx_k_RULE_regular_element, sizeof(__pyx_k_RULE_regular_element), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_replaceable_element, __pyx_k_RULE_replaceable_element, sizeof(__pyx_k_RULE_replaceable_element), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_short_class_definition, __pyx_k_RULE_short_class_definition, sizeof(__pyx_k_RULE_short_class_definition), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_simple_expression, __pyx_k_RULE_simple_expression, sizeof(__pyx_k_RULE_simple_expression), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_statement, __pyx_k_RULE_statement, sizeof(__pyx_k_RULE_statement), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_statement_options, __pyx_k_RULE_statement_options, sizeof(__pyx_k_RULE_statement_options), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_stored_definition, __pyx_k_RULE_stored_definition, sizeof(__pyx_k_RULE_stored_definition), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_stored_definition_class, __pyx_k_RULE_stored_definition_class, sizeof(__pyx_k_RULE_stored_definition_class), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_string_comment, __pyx_k_RULE_string_comment, sizeof(__pyx_k_RULE_string_comment), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_subscript, __pyx_k_RULE_subscript, sizeof(__pyx_k_RULE_subscript), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_type_prefix, __pyx_k_RULE_type_prefix, sizeof(__pyx_k_RULE_type_prefix), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_type_specifier, __pyx_k_RULE_type_specifier, sizeof(__pyx_k_RULE_type_specifier), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_when_equation, __pyx_k_RULE_when_equation, sizeof(__pyx_k_RULE_when_equation), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_when_statement, __pyx_k_RULE_when_statement, sizeof(__pyx_k_RULE_when_statement), 0, 0, 1, 1},
  {&__pyx_n_s_RULE_while_statement, __pyx_k_RULE_while_statement, sizeof(__pyx_k_RULE_while_statement), 0, 0, 1, 1},
  {&__pyx_kp_u_R_H_U, __pyx_k_R_H_U, sizeof(__pyx_k_R_H_U), 0, 1, 0, 0},
  {&__pyx_n_s_RecognitionException, __pyx_k_RecognitionException, sizeof(__pyx_k_RecognitionException), 0, 0, 1, 1},
  {&__pyx_n_s_Regular_elementContext, __pyx_k_Regular_elementContext, sizeof(__pyx_k_Regular_elementContext), 0, 0, 1, 1},
  {&__pyx_n_s_Replaceable_elementContext, __pyx_k_Replaceable_elementContext, sizeof(__pyx_k_Replaceable_elementContext), 0, 0, 1, 1},
  {&__pyx_n_s_RuleContext, __pyx_k_RuleContext, sizeof(__pyx_k_RuleContext), 0, 0, 1, 1},
  {&__pyx_kp_u_S, __pyx_k_S, sizeof(__pyx_k_S), 0, 1, 0, 0},
  {&__pyx_n_s_STRING, __pyx_k_STRING, sizeof(__pyx_k_STRING), 0, 0, 1, 1},
  {&__pyx_n_u_STRING, __pyx_k_STRING, sizeof(__pyx_k_STRING), 0, 1, 0, 1},
  {&__pyx_kp_u_S_2, __pyx_k_S_2, sizeof(__pyx_k_S_2), 0, 1, 0, 0},
  {&__pyx_kp_u_S_3, __pyx_k_S_3, sizeof(__pyx_k_S_3), 0, 1, 0, 0},
  {&__pyx_kp_u_S_4, __pyx_k_S_4, sizeof(__pyx_k_S_4), 0, 1, 0, 0},
  {&__pyx_kp_u_S_5, __pyx_k_S_5, sizeof(__pyx_k_S_5), 0, 1, 0, 0},
  {&__pyx_kp_u_S_6, __pyx_k_S_6, sizeof(__pyx_k_S_6), 0, 1, 0, 0},
  {&__pyx_n_s_Short_class_definitionContext, __pyx_k_Short_class_definitionContext, sizeof(__pyx_k_Short_class_definitionContext), 0, 0, 1, 1},
  {&__pyx_n_s_Simple_expressionContext, __pyx_k_Simple_expressionContext, sizeof(__pyx_k_Simple_expressionContext), 0, 0, 1, 1},
  {&__pyx_n_s_StatementContext, __pyx_k_StatementContext, sizeof(__pyx_k_StatementContext), 0, 0, 1, 1},
  {&__pyx_n_s_Statement_breakContext, __pyx_k_Statement_breakContext, sizeof(__pyx_k_Statement_breakContext), 0, 0, 1, 1},
  {&__pyx_n_s_Statement_component_functionCont, __pyx_k_Statement_component_functionCont, sizeof(__pyx_k_Statement_component_functionCont), 0, 0, 1, 1},
  {&__pyx_n_s_Statement_component_referenceCon, __pyx_k_Statement_component_referenceCon, sizeof(__pyx_k_Statement_component_referenceCon), 0, 0, 1, 1},
  {&__pyx_n_s_Statement_forContext, __pyx_k_Statement_forContext, sizeof(__pyx_k_Statement_forContext), 0, 0, 1, 1},
  {&__pyx_n_s_Statement_ifContext, __pyx_k_Statement_ifContext, sizeof(__pyx_k_Statement_ifContext), 0, 0, 1, 1},
  {&__pyx_n_s_Statement_optionsContext, __pyx_k_Statement_optionsContext, sizeof(__pyx_k_Statement_optionsContext), 0, 0, 1, 1},
  {&__pyx_n_s_Statement_returnContext, __pyx_k_Statement_returnContext, sizeof(__pyx_k_Statement_returnContext), 0, 0, 1, 1},
  {&__pyx_n_s_Statement_whenContext, __pyx_k_Statement_whenContext, sizeof(__pyx_k_Statement_whenContext), 0, 0, 1, 1},
  {&__pyx_n_s_Statement_whileContext, __pyx_k_Statement_whileContext, sizeof(__pyx_k_Statement_whileContext), 0, 0, 1, 1},
  {&__pyx_n_s_Stored_definitionContext, __pyx_k_Stored_definitionContext, sizeof(__pyx_k_Stored_definitionContext), 0, 0, 1, 1},
  {&__pyx_n_s_Stored_definition_classContext, __pyx_k_Stored_definition_classContext, sizeof(__pyx_k_Stored_definition_classContext), 0, 0, 1, 1},
  {&__pyx_n_s_StringIO, __pyx_k_StringIO, sizeof(__pyx_k_StringIO), 0, 0, 1, 1},
  {&__pyx_n_s_String_commentContext, __pyx_k_String_commentContext, sizeof(__pyx_k_String_commentContext), 0, 0, 1, 1},
  {&__pyx_n_s_SubscriptContext, __pyx_k_SubscriptContext, sizeof(__pyx_k_SubscriptContext), 0, 0, 1, 1},
  {&__pyx_kp_u_T, __pyx_k_T, sizeof(__pyx_k_T), 0, 1, 0, 0},
  {&__pyx_kp_u_T_2, __pyx_k_T_2, sizeof(__pyx_k_T_2), 0, 1, 0, 0},
  {&__pyx_kp_u_T_3, __pyx_k_T_3, sizeof(__pyx_k_T_3), 0, 1, 0, 0},
  {&__pyx_kp_u_T_4, __pyx_k_T_4, sizeof(__pyx_k_T_4), 0, 1, 0, 0},
  {&__pyx_kp_u_T_V_X, __pyx_k_T_V_X, sizeof(__pyx_k_T_V_X), 0, 1, 0, 0},
  {&__pyx_n_s_T__0, __pyx_k_T__0, sizeof(__pyx_k_T__0), 0, 0, 1, 1},
  {&__pyx_n_s_T__1, __pyx_k_T__1, sizeof(__pyx_k_T__1), 0, 0, 1, 1},
  {&__pyx_n_s_T__10, __pyx_k_T__10, sizeof(__pyx_k_T__10), 0, 0, 1, 1},
  {&__pyx_n_s_T__11, __pyx_k_T__11, sizeof(__pyx_k_T__11), 0, 0, 1, 1},
  {&__pyx_n_s_T__12, __pyx_k_T__12, sizeof(__pyx_k_T__12), 0, 0, 1, 1},
  {&__pyx_n_s_T__13, __pyx_k_T__13, sizeof(__pyx_k_T__13), 0, 0, 1, 1},
  {&__pyx_n_s_T__14, __pyx_k_T__14, sizeof(__pyx_k_T__14), 0, 0, 1, 1},
  {&__pyx_n_s_T__15, __pyx_k_T__15, sizeof(__pyx_k_T__15), 0, 0, 1, 1},
  {&__pyx_n_s_T__16, __pyx_k_T__16, sizeof(__pyx_k_T__16), 0, 0, 1, 1},
  {&__pyx_n_s_T__17, __pyx_k_T__17, sizeof(__pyx_k_T__17), 0, 0, 1, 1},
  {&__pyx_n_s_T__18, __pyx_k_T__18, sizeof(__pyx_k_T__18), 0, 0, 1, 1},
  {&__pyx_n_s_T__19, __pyx_k_T__19, sizeof(__pyx_k_T__19), 0, 0, 1, 1},
  {&__pyx_n_s_T__2, __pyx_k_T__2, sizeof(__pyx_k_T__2), 0, 0, 1, 1},
  {&__pyx_n_s_T__20, __pyx_k_T__20, sizeof(__pyx_k_T__20), 0, 0, 1, 1},
  {&__pyx_n_s_T__21, __pyx_k_T__21, sizeof(__pyx_k_T__21), 0, 0, 1, 1},
  {&__pyx_n_s_T__22, __pyx_k_T__22, sizeof(__pyx_k_T__22), 0, 0, 1, 1},
  {&__pyx_n_s_T__23, __pyx_k_T__23, sizeof(__pyx_k_T__23), 0, 0, 1, 1},
  {&__pyx_n_s_T__24, __pyx_k_T__24, sizeof(__pyx_k_T__24), 0, 0, 1, 1},
  {&__pyx_n_s_T__25, __pyx_k_T__25, sizeof(__pyx_k_T__25), 0, 0, 1, 1},
  {&__pyx_n_s_T__26, __pyx_k_T__26, sizeof(__pyx_k_T__26), 0, 0, 1, 1},
  {&__pyx_n_s_T__27, __pyx_k_T__27, sizeof(__pyx_k_T__27), 0, 0, 1, 1},
  {&__pyx_n_s_T__28, __pyx_k_T__28, sizeof(__pyx_k_T__28), 0, 0, 1, 1},
  {&__pyx_n_s_T__29, __pyx_k_T__29, sizeof(__pyx_k_T__29), 0, 0, 1, 1},
  {&__pyx_n_s_T__3, __pyx_k_T__3, sizeof(__pyx_k_T__3), 0, 0, 1, 1},
  {&__pyx_n_s_T__30, __pyx_k_T__30, sizeof(__pyx_k_T__30), 0, 0, 1, 1},
  {&__pyx_n_s_T__31, __pyx_k_T__31, sizeof(__pyx_k_T__31), 0, 0, 1, 1},
  {&__pyx_n_s_T__32, __pyx_k_T__32, sizeof(__pyx_k_T__32), 0, 0, 1, 1},
  {&__pyx_n_s_T__33, __pyx_k_T__33, sizeof(__pyx_k_T__33), 0, 0, 1, 1},
  {&__pyx_n_s_T__34, __pyx_k_T__34, sizeof(__pyx_k_T__34), 0, 0, 1, 1},
  {&__pyx_n_s_T__35, __pyx_k_T__35, sizeof(__pyx_k_T__35), 0, 0, 1, 1},
  {&__pyx_n_s_T__36, __pyx_k_T__36, sizeof(__pyx_k_T__36), 0, 0, 1, 1},
  {&__pyx_n_s_T__37, __pyx_k_T__37, sizeof(__pyx_k_T__37), 0, 0, 1, 1},
  {&__pyx_n_s_T__38, __pyx_k_T__38, sizeof(__pyx_k_T__38), 0, 0, 1, 1},
  {&__pyx_n_s_T__39, __pyx_k_T__39, sizeof(__pyx_k_T__39), 0, 0, 1, 1},
  {&__pyx_n_s_T__4, __pyx_k_T__4, sizeof(__pyx_k_T__4), 0, 0, 1, 1},
  {&__pyx_n_s_T__40, __pyx_k_T__40, sizeof(__pyx_k_T__40), 0, 0, 1, 1},
  {&__pyx_n_s_T__41, __pyx_k_T__41, sizeof(__pyx_k_T__41), 0, 0, 1, 1},
  {&__pyx_n_s_T__42, __pyx_k_T__42, sizeof(__pyx_k_T__42), 0, 0, 1, 1},
  {&__pyx_n_s_T__43, __pyx_k_T__43, sizeof(__pyx_k_T__43), 0, 0, 1, 1},
  {&__pyx_n_s_T__44, __pyx_k_T__44, sizeof(__pyx_k_T__44), 0, 0, 1, 1},
  {&__pyx_n_s_T__45, __pyx_k_T__45, sizeof(__pyx_k_T__45), 0, 0, 1, 1},
  {&__pyx_n_s_T__46, __pyx_k_T__46, sizeof(__pyx_k_T__46), 0, 0, 1, 1},
  {&__pyx_n_s_T__47, __pyx_k_T__47, sizeof(__pyx_k_T__47), 0, 0, 1, 1},
  {&__pyx_n_s_T__48, __pyx_k_T__48, sizeof(__pyx_k_T__48), 0, 0, 1, 1},
  {&__pyx_n_s_T__49, __pyx_k_T__49, sizeof(__pyx_k_T__49), 0, 0, 1, 1},
  {&__pyx_n_s_T__5, __pyx_k_T__5, sizeof(__pyx_k_T__5), 0, 0, 1, 1},
  {&__pyx_n_s_T__50, __pyx_k_T__50, sizeof(__pyx_k_T__50), 0, 0, 1, 1},
  {&__pyx_n_s_T__51, __pyx_k_T__51, sizeof(__pyx_k_T__51), 0, 0, 1, 1},
  {&__pyx_n_s_T__52, __pyx_k_T__52, sizeof(__pyx_k_T__52), 0, 0, 1, 1},
  {&__pyx_n_s_T__53, __pyx_k_T__53, sizeof(__pyx_k_T__53), 0, 0, 1, 1},
  {&__pyx_n_s_T__54, __pyx_k_T__54, sizeof(__pyx_k_T__54), 0, 0, 1, 1},
  {&__pyx_n_s_T__55, __pyx_k_T__55, sizeof(__pyx_k_T__55), 0, 0, 1, 1},
  {&__pyx_n_s_T__56, __pyx_k_T__56, sizeof(__pyx_k_T__56), 0, 0, 1, 1},
  {&__pyx_n_s_T__57, __pyx_k_T__57, sizeof(__pyx_k_T__57), 0, 0, 1, 1},
  {&__pyx_n_s_T__58, __pyx_k_T__58, sizeof(__pyx_k_T__58), 0, 0, 1, 1},
  {&__pyx_n_s_T__59, __pyx_k_T__59, sizeof(__pyx_k_T__59), 0, 0, 1, 1},
  {&__pyx_n_s_T__6, __pyx_k_T__6, sizeof(__pyx_k_T__6), 0, 0, 1, 1},
  {&__pyx_n_s_T__60, __pyx_k_T__60, sizeof(__pyx_k_T__60), 0, 0, 1, 1},
  {&__pyx_n_s_T__61, __pyx_k_T__61, sizeof(__pyx_k_T__61), 0, 0, 1, 1},
  {&__pyx_n_s_T__62, __pyx_k_T__62, sizeof(__pyx_k_T__62), 0, 0, 1, 1},
  {&__pyx_n_s_T__63, __pyx_k_T__63, sizeof(__pyx_k_T__63), 0, 0, 1, 1},
  {&__pyx_n_s_T__64, __pyx_k_T__64, sizeof(__pyx_k_T__64), 0, 0, 1, 1},
  {&__pyx_n_s_T__65, __pyx_k_T__65, sizeof(__pyx_k_T__65), 0, 0, 1, 1},
  {&__pyx_n_s_T__66, __pyx_k_T__66, sizeof(__pyx_k_T__66), 0, 0, 1, 1},
  {&__pyx_n_s_T__67, __pyx_k_T__67, sizeof(__pyx_k_T__67), 0, 0, 1, 1},
  {&__pyx_n_s_T__68, __pyx_k_T__68, sizeof(__pyx_k_T__68), 0, 0, 1, 1},
  {&__pyx_n_s_T__69, __pyx_k_T__69, sizeof(__pyx_k_T__69), 0, 0, 1, 1},
  {&__pyx_n_s_T__7, __pyx_k_T__7, sizeof(__pyx_k_T__7), 0, 0, 1, 1},
  {&__pyx_n_s_T__70, __pyx_k_T__70, sizeof(__pyx_k_T__70), 0, 0, 1, 1},
  {&__pyx_n_s_T__71, __pyx_k_T__71, sizeof(__pyx_k_T__71), 0, 0, 1, 1},
  {&__pyx_n_s_T__72, __pyx_k_T__72, sizeof(__pyx_k_T__72), 0, 0, 1, 1},
  {&__pyx_n_s_T__73, __pyx_k_T__73, sizeof(__pyx_k_T__73), 0, 0, 1, 1},
  {&__pyx_n_s_T__74, __pyx_k_T__74, sizeof(__pyx_k_T__74), 0, 0, 1, 1},
  {&__pyx_n_s_T__75, __pyx_k_T__75, sizeof(__pyx_k_T__75), 0, 0, 1, 1},
  {&__pyx_n_s_T__76, __pyx_k_T__76, sizeof(__pyx_k_T__76), 0, 0, 1, 1},
  {&__pyx_n_s_T__77, __pyx_k_T__77, sizeof(__pyx_k_T__77), 0, 0, 1, 1},
  {&__pyx_n_s_T__8, __pyx_k_T__8, sizeof(__pyx_k_T__8), 0, 0, 1, 1},
  {&__pyx_n_s_T__9, __pyx_k_T__9, sizeof(__pyx_k_T__9), 0, 0, 1, 1},
  {&__pyx_kp_u_T_z, __pyx_k_T_z, sizeof(__pyx_k_T_z), 0, 1, 0, 0},
  {&__pyx_n_s_TextIO, __pyx_k_TextIO, sizeof(__pyx_k_TextIO), 0, 0, 1, 1},
  {&__pyx_n_s_Token, __pyx_k_Token, sizeof(__pyx_k_Token), 0, 0, 1, 1},
  {&__pyx_n_s_TokenStream, __pyx_k_TokenStream, sizeof(__pyx_k_TokenStream), 0, 0, 1, 1},
  {&__pyx_n_s_Type_prefixContext, __pyx_k_Type_prefixContext, sizeof(__pyx_k_Type_prefixContext), 0, 0, 1, 1},
  {&__pyx_n_s_Type_specifierContext, __pyx_k_Type_specifierContext, sizeof(__pyx_k_Type_specifierContext), 0, 0, 1, 1},
  {&__pyx_kp_u_U, __pyx_k_U, sizeof(__pyx_k_U), 0, 1, 0, 0},
  {&__pyx_n_s_UNSIGNED_NUMBER, __pyx_k_UNSIGNED_NUMBER, sizeof(__pyx_k_UNSIGNED_NUMBER), 0, 0, 1, 1},
  {&__pyx_n_u_UNSIGNED_NUMBER, __pyx_k_UNSIGNED_NUMBER, sizeof(__pyx_k_UNSIGNED_NUMBER), 0, 1, 0, 1},
  {&__pyx_kp_u_V, __pyx_k_V, sizeof(__pyx_k_V), 0, 1, 0, 0},
  {&__pyx_kp_u_V_2, __pyx_k_V_2, sizeof(__pyx_k_V_2), 0, 1, 0, 0},
  {&__pyx_kp_u_V_H_Y, __pyx_k_V_H_Y, sizeof(__pyx_k_V_H_Y), 0, 1, 0, 0},
  {&__pyx_kp_u_W, __pyx_k_W, sizeof(__pyx_k_W), 0, 1, 0, 0},
  {&__pyx_n_s_WITHIN, __pyx_k_WITHIN, sizeof(__pyx_k_WITHIN), 0, 0, 1, 1},
  {&__pyx_n_u_WITHIN, __pyx_k_WITHIN, sizeof(__pyx_k_WITHIN), 0, 1, 0, 1},
  {&__pyx_n_s_WS, __pyx_k_WS, sizeof(__pyx_k_WS), 0, 0, 1, 1},
  {&__pyx_n_u_WS, __pyx_k_WS, sizeof(__pyx_k_WS), 0, 1, 0, 1},
  {&__pyx_kp_u_W_2, __pyx_k_W_2, sizeof(__pyx_k_W_2), 0, 1, 0, 0},
  {&__pyx_kp_u_W_3, __pyx_k_W_3, sizeof(__pyx_k_W_3), 0, 1, 0, 0},
  {&__pyx_n_s_When_equationContext, __pyx_k_When_equationContext, sizeof(__pyx_k_When_equationContext), 0, 0, 1, 1},
  {&__pyx_n_s_When_statementContext, __pyx_k_When_statementContext, sizeof(__pyx_k_When_statementContext), 0, 0, 1, 1},
  {&__pyx_n_s_While_statementContext, __pyx_k_While_statementContext, sizeof(__pyx_k_While_statementContext), 0, 0, 1, 1},
  {&__pyx_kp_u_X, __pyx_k_X, sizeof(__pyx_k_X), 0, 1, 0, 0},
  {&__pyx_kp_u_X_2, __pyx_k_X_2, sizeof(__pyx_k_X_2), 0, 1, 0, 0},
  {&__pyx_kp_u_X_3, __pyx_k_X_3, sizeof(__pyx_k_X_3), 0, 1, 0, 0},
  {&__pyx_kp_u_X_4, __pyx_k_X_4, sizeof(__pyx_k_X_4), 0, 1, 0, 0},
  {&__pyx_kp_u_Y, __pyx_k_Y, sizeof(__pyx_k_Y), 0, 1, 0, 0},
  {&__pyx_kp_u_Y_2, __pyx_k_Y_2, sizeof(__pyx_k_Y_2), 0, 1, 0, 0},
  {&__pyx_kp_u_Y_z, __pyx_k_Y_z, sizeof(__pyx_k_Y_z), 0, 1, 0, 0},
  {&__pyx_kp_u_Z, __pyx_k_Z, sizeof(__pyx_k_Z), 0, 1, 0, 0},
  {&__pyx_kp_u_Z_0, __pyx_k_Z_0, sizeof(__pyx_k_Z_0), 0, 1, 0, 0},
  {&__pyx_kp_u_Z_10, __pyx_k_Z_10, sizeof(__pyx_k_Z_10), 0, 1, 0, 0},
  {&__pyx_kp_u_Z_11, __pyx_k_Z_11, sizeof(__pyx_k_Z_11), 0, 1, 0, 0},
  {&__pyx_kp_u_Z_12, __pyx_k_Z_12, sizeof(__pyx_k_Z_12), 0, 1, 0, 0},
  {&__pyx_kp_u_Z_2, __pyx_k_Z_2, sizeof(__pyx_k_Z_2), 0, 1, 0, 0},
  {&__pyx_kp_u_Z_3, __pyx_k_Z_3, sizeof(__pyx_k_Z_3), 0, 1, 0, 0},
  {&__pyx_kp_u_Z_4, __pyx_k_Z_4, sizeof(__pyx_k_Z_4), 0, 1, 0, 0},
  {&__pyx_kp_u_Z_5, __pyx_k_Z_5, sizeof(__pyx_k_Z_5), 0, 1, 0, 0},
  {&__pyx_kp_u_Z_5_n8, __pyx_k_Z_5_n8, sizeof(__pyx_k_Z_5_n8), 0, 1, 0, 0},
  {&__pyx_kp_u_Z_6, __pyx_k_Z_6, sizeof(__pyx_k_Z_6), 0, 1, 0, 0},
  {&__pyx_kp_u_Z_7, __pyx_k_Z_7, sizeof(__pyx_k_Z_7), 0, 1, 0, 0},
  {&__pyx_kp_u_Z_8, __pyx_k_Z_8, sizeof(__pyx_k_Z_8), 0, 1, 0, 0},
  {&__pyx_kp_u_Z_9, __pyx_k_Z_9, sizeof(__pyx_k_Z_9), 0, 1, 0, 0},
  {&__pyx_kp_u_Z_F, __pyx_k_Z_F, sizeof(__pyx_k_Z_F), 0, 1, 0, 0},
  {&__pyx_kp_u_Z_F_2, __pyx_k_Z_F_2, sizeof(__pyx_k_Z_F_2), 0, 1, 0, 0},
  {&__pyx_kp_u_Z_I, __pyx_k_Z_I, sizeof(__pyx_k_Z_I), 0, 1, 0, 0},
  {&__pyx_kp_u_Z_Z, __pyx_k_Z_Z, sizeof(__pyx_k_Z_Z), 0, 1, 0, 0},
  {&__pyx_kp_u_Z_Z_2, __pyx_k_Z_Z_2, sizeof(__pyx_k_Z_Z_2), 0, 1, 0, 0},
  {&__pyx_kp_u__1000, __pyx_k__1000, sizeof(__pyx_k__1000), 0, 1, 0, 0},
  {&__pyx_kp_u__1001, __pyx_k__1001, sizeof(__pyx_k__1001), 0, 1, 0, 0},
  {&__pyx_kp_u__1002, __pyx_k__1002, sizeof(__pyx_k__1002), 0, 1, 0, 0},
  {&__pyx_kp_u__1003, __pyx_k__1003, sizeof(__pyx_k__1003), 0, 1, 0, 0},
  {&__pyx_kp_u__1004, __pyx_k__1004, sizeof(__pyx_k__1004), 0, 1, 0, 0},
  {&__pyx_kp_u__1005, __pyx_k__1005, sizeof(__pyx_k__1005), 0, 1, 0, 0},
  {&__pyx_kp_u__101, __pyx_k__101, sizeof(__pyx_k__101), 0, 1, 0, 0},
  {&__pyx_kp_u__103, __pyx_k__103, sizeof(__pyx_k__103), 0, 1, 0, 0},
  {&__pyx_kp_u__105, __pyx_k__105, sizeof(__pyx_k__105), 0, 1, 0, 0},
  {&__pyx_kp_u__107, __pyx_k__107, sizeof(__pyx_k__107), 0, 1, 0, 0},
  {&__pyx_kp_u__109, __pyx_k__109, sizeof(__pyx_k__109), 0, 1, 0, 0},
  {&__pyx_kp_u__11, __pyx_k__11, sizeof(__pyx_k__11), 0, 1, 0, 0},
  {&__pyx_kp_u__111, __pyx_k__111, sizeof(__pyx_k__111), 0, 1, 0, 0},
  {&__pyx_kp_u__113, __pyx_k__113, sizeof(__pyx_k__113), 0, 1, 0, 0},
  {&__pyx_kp_u__115, __pyx_k__115, sizeof(__pyx_k__115), 0, 1, 0, 0},
  {&__pyx_kp_u__117, __pyx_k__117, sizeof(__pyx_k__117), 0, 1, 0, 0},
  {&__pyx_kp_u__119, __pyx_k__119, sizeof(__pyx_k__119), 0, 1, 0, 0},
  {&__pyx_kp_u__121, __pyx_k__121, sizeof(__pyx_k__121), 0, 1, 0, 0},
  {&__pyx_kp_u__123, __pyx_k__123, sizeof(__pyx_k__123), 0, 1, 0, 0},
  {&__pyx_kp_u__125, __pyx_k__125, sizeof(__pyx_k__125), 0, 1, 0, 0},
  {&__pyx_kp_u__13, __pyx_k__13, sizeof(__pyx_k__13), 0, 1, 0, 0},
  {&__pyx_kp_u__144, __pyx_k__144, sizeof(__pyx_k__144), 0, 1, 0, 0},
  {&__pyx_kp_u__146, __pyx_k__146, sizeof(__pyx_k__146), 0, 1, 0, 0},
  {&__pyx_kp_u__148, __pyx_k__148, sizeof(__pyx_k__148), 0, 1, 0, 0},
  {&__pyx_kp_u__15, __pyx_k__15, sizeof(__pyx_k__15), 0, 1, 0, 0},
  {&__pyx_kp_u__150, __pyx_k__150, sizeof(__pyx_k__150), 0, 1, 0, 0},
  {&__pyx_kp_u__152, __pyx_k__152, sizeof(__pyx_k__152), 0, 1, 0, 0},
  {&__pyx_kp_u__154, __pyx_k__154, sizeof(__pyx_k__154), 0, 1, 0, 0},
  {&__pyx_kp_u__156, __pyx_k__156, sizeof(__pyx_k__156), 0, 1, 0, 0},
  {&__pyx_kp_u__167, __pyx_k__167, sizeof(__pyx_k__167), 0, 1, 0, 0},
  {&__pyx_kp_u__169, __pyx_k__169, sizeof(__pyx_k__169), 0, 1, 0, 0},
  {&__pyx_kp_u__172, __pyx_k__172, sizeof(__pyx_k__172), 0, 1, 0, 0},
  {&__pyx_kp_u__174, __pyx_k__174, sizeof(__pyx_k__174), 0, 1, 0, 0},
  {&__pyx_kp_u__176, __pyx_k__176, sizeof(__pyx_k__176), 0, 1, 0, 0},
  {&__pyx_kp_u__178, __pyx_k__178, sizeof(__pyx_k__178), 0, 1, 0, 0},
  {&__pyx_kp_u__180, __pyx_k__180, sizeof(__pyx_k__180), 0, 1, 0, 0},
  {&__pyx_kp_u__182, __pyx_k__182, sizeof(__pyx_k__182), 0, 1, 0, 0},
  {&__pyx_kp_u__195, __pyx_k__195, sizeof(__pyx_k__195), 0, 1, 0, 0},
  {&__pyx_kp_u__197, __pyx_k__197, sizeof(__pyx_k__197), 0, 1, 0, 0},
  {&__pyx_kp_u__199, __pyx_k__199, sizeof(__pyx_k__199), 0, 1, 0, 0},
  {&__pyx_kp_u__201, __pyx_k__201, sizeof(__pyx_k__201), 0, 1, 0, 0},
  {&__pyx_kp_u__204, __pyx_k__204, sizeof(__pyx_k__204), 0, 1, 0, 0},
  {&__pyx_kp_u__206, __pyx_k__206, sizeof(__pyx_k__206), 0, 1, 0, 0},
  {&__pyx_kp_u__208, __pyx_k__208, sizeof(__pyx_k__208), 0, 1, 0, 0},
  {&__pyx_kp_u__21, __pyx_k__21, sizeof(__pyx_k__21), 0, 1, 0, 0},
  {&__pyx_kp_u__210, __pyx_k__210, sizeof(__pyx_k__210), 0, 1, 0, 0},
  {&__pyx_kp_u__212, __pyx_k__212, sizeof(__pyx_k__212), 0, 1, 0, 0},
  {&__pyx_kp_u__215, __pyx_k__215, sizeof(__pyx_k__215), 0, 1, 0, 0},
  {&__pyx_kp_u__218, __pyx_k__218, sizeof(__pyx_k__218), 0, 1, 0, 0},
  {&__pyx_kp_u__220, __pyx_k__220, sizeof(__pyx_k__220), 0, 1, 0, 0},
  {&__pyx_kp_u__223, __pyx_k__223, sizeof(__pyx_k__223), 0, 1, 0, 0},
  {&__pyx_kp_u__225, __pyx_k__225, sizeof(__pyx_k__225), 0, 1, 0, 0},
  {&__pyx_kp_u__227, __pyx_k__227, sizeof(__pyx_k__227), 0, 1, 0, 0},
  {&__pyx_kp_u__229, __pyx_k__229, sizeof(__pyx_k__229), 0, 1, 0, 0},
  {&__pyx_kp_u__23, __pyx_k__23, sizeof(__pyx_k__23), 0, 1, 0, 0},
  {&__pyx_kp_u__231, __pyx_k__231, sizeof(__pyx_k__231), 0, 1, 0, 0},
  {&__pyx_kp_u__233, __pyx_k__233, sizeof(__pyx_k__233), 0, 1, 0, 0},
  {&__pyx_kp_u__235, __pyx_k__235, sizeof(__pyx_k__235), 0, 1, 0, 0},
  {&__pyx_kp_u__237, __pyx_k__237, sizeof(__pyx_k__237), 0, 1, 0, 0},
  {&__pyx_kp_u__239, __pyx_k__239, sizeof(__pyx_k__239), 0, 1, 0, 0},
  {&__pyx_kp_u__241, __pyx_k__241, sizeof(__pyx_k__241), 0, 1, 0, 0},
  {&__pyx_kp_u__243, __pyx_k__243, sizeof(__pyx_k__243), 0, 1, 0, 0},
  {&__pyx_kp_u__245, __pyx_k__245, sizeof(__pyx_k__245), 0, 1, 0, 0},
  {&__pyx_kp_u__247, __pyx_k__247, sizeof(__pyx_k__247), 0, 1, 0, 0},
  {&__pyx_kp_u__249, __pyx_k__249, sizeof(__pyx_k__249), 0, 1, 0, 0},
  {&__pyx_kp_u__25, __pyx_k__25, sizeof(__pyx_k__25), 0, 1, 0, 0},
  {&__pyx_kp_u__255, __pyx_k__255, sizeof(__pyx_k__255), 0, 1, 0, 0},
  {&__pyx_kp_u__257, __pyx_k__257, sizeof(__pyx_k__257), 0, 1, 0, 0},
  {&__pyx_kp_u__261, __pyx_k__261, sizeof(__pyx_k__261), 0, 1, 0, 0},
  {&__pyx_kp_u__263, __pyx_k__263, sizeof(__pyx_k__263), 0, 1, 0, 0},
  {&__pyx_kp_u__265, __pyx_k__265, sizeof(__pyx_k__265), 0, 1, 0, 0},
  {&__pyx_kp_u__267, __pyx_k__267, sizeof(__pyx_k__267), 0, 1, 0, 0},
  {&__pyx_kp_u__27, __pyx_k__27, sizeof(__pyx_k__27), 0, 1, 0, 0},
  {&__pyx_kp_u__270, __pyx_k__270, sizeof(__pyx_k__270), 0, 1, 0, 0},
  {&__pyx_kp_u__274, __pyx_k__274, sizeof(__pyx_k__274), 0, 1, 0, 0},
  {&__pyx_kp_u__276, __pyx_k__276, sizeof(__pyx_k__276), 0, 1, 0, 0},
  {&__pyx_kp_u__280, __pyx_k__280, sizeof(__pyx_k__280), 0, 1, 0, 0},
  {&__pyx_kp_u__284, __pyx_k__284, sizeof(__pyx_k__284), 0, 1, 0, 0},
  {&__pyx_kp_u__286, __pyx_k__286, sizeof(__pyx_k__286), 0, 1, 0, 0},
  {&__pyx_kp_u__288, __pyx_k__288, sizeof(__pyx_k__288), 0, 1, 0, 0},
  {&__pyx_kp_u__29, __pyx_k__29, sizeof(__pyx_k__29), 0, 1, 0, 0},
  {&__pyx_kp_u__290, __pyx_k__290, sizeof(__pyx_k__290), 0, 1, 0, 0},
  {&__pyx_kp_u__292, __pyx_k__292, sizeof(__pyx_k__292), 0, 1, 0, 0},
  {&__pyx_kp_u__294, __pyx_k__294, sizeof(__pyx_k__294), 0, 1, 0, 0},
  {&__pyx_kp_u__298, __pyx_k__298, sizeof(__pyx_k__298), 0, 1, 0, 0},
  {&__pyx_kp_u__3, __pyx_k__3, sizeof(__pyx_k__3), 0, 1, 0, 0},
  {&__pyx_kp_u__302, __pyx_k__302, sizeof(__pyx_k__302), 0, 1, 0, 0},
  {&__pyx_kp_u__304, __pyx_k__304, sizeof(__pyx_k__304), 0, 1, 0, 0},
  {&__pyx_kp_u__306, __pyx_k__306, sizeof(__pyx_k__306), 0, 1, 0, 0},
  {&__pyx_kp_u__308, __pyx_k__308, sizeof(__pyx_k__308), 0, 1, 0, 0},
  {&__pyx_kp_u__31, __pyx_k__31, sizeof(__pyx_k__31), 0, 1, 0, 0},
  {&__pyx_kp_u__310, __pyx_k__310, sizeof(__pyx_k__310), 0, 1, 0, 0},
  {&__pyx_kp_u__312, __pyx_k__312, sizeof(__pyx_k__312), 0, 1, 0, 0},
  {&__pyx_kp_u__315, __pyx_k__315, sizeof(__pyx_k__315), 0, 1, 0, 0},
  {&__pyx_kp_u__317, __pyx_k__317, sizeof(__pyx_k__317), 0, 1, 0, 0},
  {&__pyx_kp_u__320, __pyx_k__320, sizeof(__pyx_k__320), 0, 1, 0, 0},
  {&__pyx_kp_u__322, __pyx_k__322, sizeof(__pyx_k__322), 0, 1, 0, 0},
  {&__pyx_kp_u__325, __pyx_k__325, sizeof(__pyx_k__325), 0, 1, 0, 0},
  {&__pyx_kp_u__329, __pyx_k__329, sizeof(__pyx_k__329), 0, 1, 0, 0},
  {&__pyx_kp_u__33, __pyx_k__33, sizeof(__pyx_k__33), 0, 1, 0, 0},
  {&__pyx_kp_u__331, __pyx_k__331, sizeof(__pyx_k__331), 0, 1, 0, 0},
  {&__pyx_kp_u__333, __pyx_k__333, sizeof(__pyx_k__333), 0, 1, 0, 0},
  {&__pyx_kp_u__335, __pyx_k__335, sizeof(__pyx_k__335), 0, 1, 0, 0},
  {&__pyx_kp_u__337, __pyx_k__337, sizeof(__pyx_k__337), 0, 1, 0, 0},
  {&__pyx_kp_u__339, __pyx_k__339, sizeof(__pyx_k__339), 0, 1, 0, 0},
  {&__pyx_kp_u__341, __pyx_k__341, sizeof(__pyx_k__341), 0, 1, 0, 0},
  {&__pyx_kp_u__344, __pyx_k__344, sizeof(__pyx_k__344), 0, 1, 0, 0},
  {&__pyx_kp_u__348, __pyx_k__348, sizeof(__pyx_k__348), 0, 1, 0, 0},
  {&__pyx_kp_u__35, __pyx_k__35, sizeof(__pyx_k__35), 0, 1, 0, 0},
  {&__pyx_kp_u__351, __pyx_k__351, sizeof(__pyx_k__351), 0, 1, 0, 0},
  {&__pyx_kp_u__353, __pyx_k__353, sizeof(__pyx_k__353), 0, 1, 0, 0},
  {&__pyx_kp_u__357, __pyx_k__357, sizeof(__pyx_k__357), 0, 1, 0, 0},
  {&__pyx_kp_u__361, __pyx_k__361, sizeof(__pyx_k__361), 0, 1, 0, 0},
  {&__pyx_kp_u__363, __pyx_k__363, sizeof(__pyx_k__363), 0, 1, 0, 0},
  {&__pyx_kp_u__365, __pyx_k__365, sizeof(__pyx_k__365), 0, 1, 0, 0},
  {&__pyx_kp_u__368, __pyx_k__368, sizeof(__pyx_k__368), 0, 1, 0, 0},
  {&__pyx_kp_u__37, __pyx_k__37, sizeof(__pyx_k__37), 0, 1, 0, 0},
  {&__pyx_kp_u__373, __pyx_k__373, sizeof(__pyx_k__373), 0, 1, 0, 0},
  {&__pyx_kp_u__375, __pyx_k__375, sizeof(__pyx_k__375), 0, 1, 0, 0},
  {&__pyx_kp_u__377, __pyx_k__377, sizeof(__pyx_k__377), 0, 1, 0, 0},
  {&__pyx_kp_u__379, __pyx_k__379, sizeof(__pyx_k__379), 0, 1, 0, 0},
  {&__pyx_kp_u__383, __pyx_k__383, sizeof(__pyx_k__383), 0, 1, 0, 0},
  {&__pyx_kp_u__385, __pyx_k__385, sizeof(__pyx_k__385), 0, 1, 0, 0},
  {&__pyx_kp_u__387, __pyx_k__387, sizeof(__pyx_k__387), 0, 1, 0, 0},
  {&__pyx_kp_u__39, __pyx_k__39, sizeof(__pyx_k__39), 0, 1, 0, 0},
  {&__pyx_kp_u__391, __pyx_k__391, sizeof(__pyx_k__391), 0, 1, 0, 0},
  {&__pyx_kp_u__394, __pyx_k__394, sizeof(__pyx_k__394), 0, 1, 0, 0},
  {&__pyx_kp_u__396, __pyx_k__396, sizeof(__pyx_k__396), 0, 1, 0, 0},
  {&__pyx_kp_u__400, __pyx_k__400, sizeof(__pyx_k__400), 0, 1, 0, 0},
  {&__pyx_kp_u__402, __pyx_k__402, sizeof(__pyx_k__402), 0, 1, 0, 0},
  {&__pyx_kp_u__404, __pyx_k__404, sizeof(__pyx_k__404), 0, 1, 0, 0},
  {&__pyx_kp_u__406, __pyx_k__406, sizeof(__pyx_k__406), 0, 1, 0, 0},
  {&__pyx_kp_u__41, __pyx_k__41, sizeof(__pyx_k__41), 0, 1, 0, 0},
  {&__pyx_kp_u__410, __pyx_k__410, sizeof(__pyx_k__410), 0, 1, 0, 0},
  {&__pyx_kp_u__414, __pyx_k__414, sizeof(__pyx_k__414), 0, 1, 0, 0},
  {&__pyx_kp_u__421, __pyx_k__421, sizeof(__pyx_k__421), 0, 1, 0, 0},
  {&__pyx_kp_u__427, __pyx_k__427, sizeof(__pyx_k__427), 0, 1, 0, 0},
  {&__pyx_kp_u__429, __pyx_k__429, sizeof(__pyx_k__429), 0, 1, 0, 0},
  {&__pyx_kp_u__43, __pyx_k__43, sizeof(__pyx_k__43), 0, 1, 0, 0},
  {&__pyx_kp_u__431, __pyx_k__431, sizeof(__pyx_k__431), 0, 1, 0, 0},
  {&__pyx_kp_u__433, __pyx_k__433, sizeof(__pyx_k__433), 0, 1, 0, 0},
  {&__pyx_kp_u__435, __pyx_k__435, sizeof(__pyx_k__435), 0, 1, 0, 0},
  {&__pyx_kp_u__437, __pyx_k__437, sizeof(__pyx_k__437), 0, 1, 0, 0},
  {&__pyx_kp_u__441, __pyx_k__441, sizeof(__pyx_k__441), 0, 1, 0, 0},
  {&__pyx_kp_u__447, __pyx_k__447, sizeof(__pyx_k__447), 0, 1, 0, 0},
  {&__pyx_kp_u__45, __pyx_k__45, sizeof(__pyx_k__45), 0, 1, 0, 0},
  {&__pyx_kp_u__450, __pyx_k__450, sizeof(__pyx_k__450), 0, 1, 0, 0},
  {&__pyx_kp_u__458, __pyx_k__458, sizeof(__pyx_k__458), 0, 1, 0, 0},
  {&__pyx_kp_u__460, __pyx_k__460, sizeof(__pyx_k__460), 0, 1, 0, 0},
  {&__pyx_kp_u__467, __pyx_k__467, sizeof(__pyx_k__467), 0, 1, 0, 0},
  {&__pyx_kp_u__469, __pyx_k__469, sizeof(__pyx_k__469), 0, 1, 0, 0},
  {&__pyx_kp_u__47, __pyx_k__47, sizeof(__pyx_k__47), 0, 1, 0, 0},
  {&__pyx_kp_u__472, __pyx_k__472, sizeof(__pyx_k__472), 0, 1, 0, 0},
  {&__pyx_kp_u__474, __pyx_k__474, sizeof(__pyx_k__474), 0, 1, 0, 0},
  {&__pyx_kp_u__476, __pyx_k__476, sizeof(__pyx_k__476), 0, 1, 0, 0},
  {&__pyx_kp_u__478, __pyx_k__478, sizeof(__pyx_k__478), 0, 1, 0, 0},
  {&__pyx_kp_u__484, __pyx_k__484, sizeof(__pyx_k__484), 0, 1, 0, 0},
  {&__pyx_kp_u__486, __pyx_k__486, sizeof(__pyx_k__486), 0, 1, 0, 0},
  {&__pyx_kp_u__49, __pyx_k__49, sizeof(__pyx_k__49), 0, 1, 0, 0},
  {&__pyx_kp_u__491, __pyx_k__491, sizeof(__pyx_k__491), 0, 1, 0, 0},
  {&__pyx_kp_u__493, __pyx_k__493, sizeof(__pyx_k__493), 0, 1, 0, 0},
  {&__pyx_kp_u__499, __pyx_k__499, sizeof(__pyx_k__499), 0, 1, 0, 0},
  {&__pyx_kp_u__5, __pyx_k__5, sizeof(__pyx_k__5), 0, 1, 0, 0},
  {&__pyx_kp_u__501, __pyx_k__501, sizeof(__pyx_k__501), 0, 1, 0, 0},
  {&__pyx_kp_u__506, __pyx_k__506, sizeof(__pyx_k__506), 0, 1, 0, 0},
  {&__pyx_kp_u__51, __pyx_k__51, sizeof(__pyx_k__51), 0, 1, 0, 0},
  {&__pyx_kp_u__510, __pyx_k__510, sizeof(__pyx_k__510), 0, 1, 0, 0},
  {&__pyx_kp_u__512, __pyx_k__512, sizeof(__pyx_k__512), 0, 1, 0, 0},
  {&__pyx_kp_u__515, __pyx_k__515, sizeof(__pyx_k__515), 0, 1, 0, 0},
  {&__pyx_kp_u__519, __pyx_k__519, sizeof(__pyx_k__519), 0, 1, 0, 0},
  {&__pyx_kp_u__521, __pyx_k__521, sizeof(__pyx_k__521), 0, 1, 0, 0},
  {&__pyx_kp_u__527, __pyx_k__527, sizeof(__pyx_k__527), 0, 1, 0, 0},
  {&__pyx_kp_u__529, __pyx_k__529, sizeof(__pyx_k__529), 0, 1, 0, 0},
  {&__pyx_kp_u__53, __pyx_k__53, sizeof(__pyx_k__53), 0, 1, 0, 0},
  {&__pyx_kp_u__533, __pyx_k__533, sizeof(__pyx_k__533), 0, 1, 0, 0},
  {&__pyx_kp_u__535, __pyx_k__535, sizeof(__pyx_k__535), 0, 1, 0, 0},
  {&__pyx_kp_u__537, __pyx_k__537, sizeof(__pyx_k__537), 0, 1, 0, 0},
  {&__pyx_kp_u__539, __pyx_k__539, sizeof(__pyx_k__539), 0, 1, 0, 0},
  {&__pyx_kp_u__543, __pyx_k__543, sizeof(__pyx_k__543), 0, 1, 0, 0},
  {&__pyx_kp_u__545, __pyx_k__545, sizeof(__pyx_k__545), 0, 1, 0, 0},
  {&__pyx_kp_u__547, __pyx_k__547, sizeof(__pyx_k__547), 0, 1, 0, 0},
  {&__pyx_kp_u__549, __pyx_k__549, sizeof(__pyx_k__549), 0, 1, 0, 0},
  {&__pyx_kp_u__55, __pyx_k__55, sizeof(__pyx_k__55), 0, 1, 0, 0},
  {&__pyx_kp_u__553, __pyx_k__553, sizeof(__pyx_k__553), 0, 1, 0, 0},
  {&__pyx_kp_u__555, __pyx_k__555, sizeof(__pyx_k__555), 0, 1, 0, 0},
  {&__pyx_kp_u__559, __pyx_k__559, sizeof(__pyx_k__559), 0, 1, 0, 0},
  {&__pyx_kp_u__561, __pyx_k__561, sizeof(__pyx_k__561), 0, 1, 0, 0},
  {&__pyx_kp_u__563, __pyx_k__563, sizeof(__pyx_k__563), 0, 1, 0, 0},
  {&__pyx_kp_u__565, __pyx_k__565, sizeof(__pyx_k__565), 0, 1, 0, 0},
  {&__pyx_kp_u__567, __pyx_k__567, sizeof(__pyx_k__567), 0, 1, 0, 0},
  {&__pyx_kp_u__57, __pyx_k__57, sizeof(__pyx_k__57), 0, 1, 0, 0},
  {&__pyx_kp_u__570, __pyx_k__570, sizeof(__pyx_k__570), 0, 1, 0, 0},
  {&__pyx_kp_u__572, __pyx_k__572, sizeof(__pyx_k__572), 0, 1, 0, 0},
  {&__pyx_kp_u__574, __pyx_k__574, sizeof(__pyx_k__574), 0, 1, 0, 0},
  {&__pyx_kp_u__576, __pyx_k__576, sizeof(__pyx_k__576), 0, 1, 0, 0},
  {&__pyx_kp_u__578, __pyx_k__578, sizeof(__pyx_k__578), 0, 1, 0, 0},
  {&__pyx_kp_u__582, __pyx_k__582, sizeof(__pyx_k__582), 0, 1, 0, 0},
  {&__pyx_kp_u__584, __pyx_k__584, sizeof(__pyx_k__584), 0, 1, 0, 0},
  {&__pyx_kp_u__589, __pyx_k__589, sizeof(__pyx_k__589), 0, 1, 0, 0},
  {&__pyx_kp_u__59, __pyx_k__59, sizeof(__pyx_k__59), 0, 1, 0, 0},
  {&__pyx_kp_u__591, __pyx_k__591, sizeof(__pyx_k__591), 0, 1, 0, 0},
  {&__pyx_kp_u__593, __pyx_k__593, sizeof(__pyx_k__593), 0, 1, 0, 0},
  {&__pyx_kp_u__597, __pyx_k__597, sizeof(__pyx_k__597), 0, 1, 0, 0},
  {&__pyx_kp_u__604, __pyx_k__604, sizeof(__pyx_k__604), 0, 1, 0, 0},
  {&__pyx_kp_u__606, __pyx_k__606, sizeof(__pyx_k__606), 0, 1, 0, 0},
  {&__pyx_kp_u__61, __pyx_k__61, sizeof(__pyx_k__61), 0, 1, 0, 0},
  {&__pyx_kp_u__611, __pyx_k__611, sizeof(__pyx_k__611), 0, 1, 0, 0},
  {&__pyx_kp_u__613, __pyx_k__613, sizeof(__pyx_k__613), 0, 1, 0, 0},
  {&__pyx_kp_u__617, __pyx_k__617, sizeof(__pyx_k__617), 0, 1, 0, 0},
  {&__pyx_kp_u__619, __pyx_k__619, sizeof(__pyx_k__619), 0, 1, 0, 0},
  {&__pyx_kp_u__621, __pyx_k__621, sizeof(__pyx_k__621), 0, 1, 0, 0},
  {&__pyx_kp_u__623, __pyx_k__623, sizeof(__pyx_k__623), 0, 1, 0, 0},
  {&__pyx_kp_u__625, __pyx_k__625, sizeof(__pyx_k__625), 0, 1, 0, 0},
  {&__pyx_kp_u__63, __pyx_k__63, sizeof(__pyx_k__63), 0, 1, 0, 0},
  {&__pyx_kp_u__630, __pyx_k__630, sizeof(__pyx_k__630), 0, 1, 0, 0},
  {&__pyx_kp_u__632, __pyx_k__632, sizeof(__pyx_k__632), 0, 1, 0, 0},
  {&__pyx_kp_u__636, __pyx_k__636, sizeof(__pyx_k__636), 0, 1, 0, 0},
  {&__pyx_kp_u__638, __pyx_k__638, sizeof(__pyx_k__638), 0, 1, 0, 0},
  {&__pyx_kp_u__640, __pyx_k__640, sizeof(__pyx_k__640), 0, 1, 0, 0},
  {&__pyx_kp_u__642, __pyx_k__642, sizeof(__pyx_k__642), 0, 1, 0, 0},
  {&__pyx_kp_u__644, __pyx_k__644, sizeof(__pyx_k__644), 0, 1, 0, 0},
  {&__pyx_kp_u__65, __pyx_k__65, sizeof(__pyx_k__65), 0, 1, 0, 0},
  {&__pyx_kp_u__653, __pyx_k__653, sizeof(__pyx_k__653), 0, 1, 0, 0},
  {&__pyx_kp_u__655, __pyx_k__655, sizeof(__pyx_k__655), 0, 1, 0, 0},
  {&__pyx_kp_u__661, __pyx_k__661, sizeof(__pyx_k__661), 0, 1, 0, 0},
  {&__pyx_kp_u__668, __pyx_k__668, sizeof(__pyx_k__668), 0, 1, 0, 0},
  {&__pyx_kp_u__67, __pyx_k__67, sizeof(__pyx_k__67), 0, 1, 0, 0},
  {&__pyx_kp_u__670, __pyx_k__670, sizeof(__pyx_k__670), 0, 1, 0, 0},
  {&__pyx_kp_u__677, __pyx_k__677, sizeof(__pyx_k__677), 0, 1, 0, 0},
  {&__pyx_kp_u__679, __pyx_k__679, sizeof(__pyx_k__679), 0, 1, 0, 0},
  {&__pyx_kp_u__684, __pyx_k__684, sizeof(__pyx_k__684), 0, 1, 0, 0},
  {&__pyx_kp_u__69, __pyx_k__69, sizeof(__pyx_k__69), 0, 1, 0, 0},
  {&__pyx_kp_u__690, __pyx_k__690, sizeof(__pyx_k__690), 0, 1, 0, 0},
  {&__pyx_kp_u__692, __pyx_k__692, sizeof(__pyx_k__692), 0, 1, 0, 0},
  {&__pyx_kp_u__695, __pyx_k__695, sizeof(__pyx_k__695), 0, 1, 0, 0},
  {&__pyx_kp_u__697, __pyx_k__697, sizeof(__pyx_k__697), 0, 1, 0, 0},
  {&__pyx_kp_u__699, __pyx_k__699, sizeof(__pyx_k__699), 0, 1, 0, 0},
  {&__pyx_kp_u__7, __pyx_k__7, sizeof(__pyx_k__7), 0, 1, 0, 0},
  {&__pyx_kp_u__701, __pyx_k__701, sizeof(__pyx_k__701), 0, 1, 0, 0},
  {&__pyx_kp_u__703, __pyx_k__703, sizeof(__pyx_k__703), 0, 1, 0, 0},
  {&__pyx_kp_u__705, __pyx_k__705, sizeof(__pyx_k__705), 0, 1, 0, 0},
  {&__pyx_kp_u__708, __pyx_k__708, sizeof(__pyx_k__708), 0, 1, 0, 0},
  {&__pyx_kp_u__71, __pyx_k__71, sizeof(__pyx_k__71), 0, 1, 0, 0},
  {&__pyx_kp_u__711, __pyx_k__711, sizeof(__pyx_k__711), 0, 1, 0, 0},
  {&__pyx_kp_u__717, __pyx_k__717, sizeof(__pyx_k__717), 0, 1, 0, 0},
  {&__pyx_kp_u__719, __pyx_k__719, sizeof(__pyx_k__719), 0, 1, 0, 0},
  {&__pyx_kp_u__721, __pyx_k__721, sizeof(__pyx_k__721), 0, 1, 0, 0},
  {&__pyx_kp_u__723, __pyx_k__723, sizeof(__pyx_k__723), 0, 1, 0, 0},
  {&__pyx_kp_u__728, __pyx_k__728, sizeof(__pyx_k__728), 0, 1, 0, 0},
  {&__pyx_kp_u__73, __pyx_k__73, sizeof(__pyx_k__73), 0, 1, 0, 0},
  {&__pyx_kp_u__730, __pyx_k__730, sizeof(__pyx_k__730), 0, 1, 0, 0},
  {&__pyx_kp_u__735, __pyx_k__735, sizeof(__pyx_k__735), 0, 1, 0, 0},
  {&__pyx_kp_u__737, __pyx_k__737, sizeof(__pyx_k__737), 0, 1, 0, 0},
  {&__pyx_kp_u__743, __pyx_k__743, sizeof(__pyx_k__743), 0, 1, 0, 0},
  {&__pyx_kp_u__747, __pyx_k__747, sizeof(__pyx_k__747), 0, 1, 0, 0},
  {&__pyx_kp_u__75, __pyx_k__75, sizeof(__pyx_k__75), 0, 1, 0, 0},
  {&__pyx_kp_u__750, __pyx_k__750, sizeof(__pyx_k__750), 0, 1, 0, 0},
  {&__pyx_kp_u__754, __pyx_k__754, sizeof(__pyx_k__754), 0, 1, 0, 0},
  {&__pyx_kp_u__756, __pyx_k__756, sizeof(__pyx_k__756), 0, 1, 0, 0},
  {&__pyx_kp_u__760, __pyx_k__760, sizeof(__pyx_k__760), 0, 1, 0, 0},
  {&__pyx_kp_u__764, __pyx_k__764, sizeof(__pyx_k__764), 0, 1, 0, 0},
  {&__pyx_kp_u__767, __pyx_k__767, sizeof(__pyx_k__767), 0, 1, 0, 0},
  {&__pyx_kp_u__769, __pyx_k__769, sizeof(__pyx_k__769), 0, 1, 0, 0},
  {&__pyx_kp_u__77, __pyx_k__77, sizeof(__pyx_k__77), 0, 1, 0, 0},
  {&__pyx_kp_u__771, __pyx_k__771, sizeof(__pyx_k__771), 0, 1, 0, 0},
  {&__pyx_kp_u__773, __pyx_k__773, sizeof(__pyx_k__773), 0, 1, 0, 0},
  {&__pyx_kp_u__775, __pyx_k__775, sizeof(__pyx_k__775), 0, 1, 0, 0},
  {&__pyx_kp_u__778, __pyx_k__778, sizeof(__pyx_k__778), 0, 1, 0, 0},
  {&__pyx_kp_u__780, __pyx_k__780, sizeof(__pyx_k__780), 0, 1, 0, 0},
  {&__pyx_n_u__782, __pyx_k__782, sizeof(__pyx_k__782), 0, 1, 0, 1},
  {&__pyx_n_u__784, __pyx_k__784, sizeof(__pyx_k__784), 0, 1, 0, 1},
  {&__pyx_n_u__786, __pyx_k__786, sizeof(__pyx_k__786), 0, 1, 0, 1},
  {&__pyx_n_u__788, __pyx_k__788, sizeof(__pyx_k__788), 0, 1, 0, 1},
  {&__pyx_kp_u__79, __pyx_k__79, sizeof(__pyx_k__79), 0, 1, 0, 0},
  {&__pyx_n_u__790, __pyx_k__790, sizeof(__pyx_k__790), 0, 1, 0, 1},
  {&__pyx_n_u__792, __pyx_k__792, sizeof(__pyx_k__792), 0, 1, 0, 1},
  {&__pyx_n_u__794, __pyx_k__794, sizeof(__pyx_k__794), 0, 1, 0, 1},
  {&__pyx_n_u__796, __pyx_k__796, sizeof(__pyx_k__796), 0, 1, 0, 1},
  {&__pyx_n_u__798, __pyx_k__798, sizeof(__pyx_k__798), 0, 1, 0, 1},
  {&__pyx_kp_u__800, __pyx_k__800, sizeof(__pyx_k__800), 0, 1, 0, 0},
  {&__pyx_kp_u__802, __pyx_k__802, sizeof(__pyx_k__802), 0, 1, 0, 0},
  {&__pyx_n_u__804, __pyx_k__804, sizeof(__pyx_k__804), 0, 1, 0, 1},
  {&__pyx_n_u__806, __pyx_k__806, sizeof(__pyx_k__806), 0, 1, 0, 1},
  {&__pyx_n_u__808, __pyx_k__808, sizeof(__pyx_k__808), 0, 1, 0, 1},
  {&__pyx_kp_u__81, __pyx_k__81, sizeof(__pyx_k__81), 0, 1, 0, 0},
  {&__pyx_kp_u__83, __pyx_k__83, sizeof(__pyx_k__83), 0, 1, 0, 0},
  {&__pyx_kp_u__85, __pyx_k__85, sizeof(__pyx_k__85), 0, 1, 0, 0},
  {&__pyx_kp_u__87, __pyx_k__87, sizeof(__pyx_k__87), 0, 1, 0, 0},
  {&__pyx_kp_u__89, __pyx_k__89, sizeof(__pyx_k__89), 0, 1, 0, 0},
  {&__pyx_kp_u__9, __pyx_k__9, sizeof(__pyx_k__9), 0, 1, 0, 0},
  {&__pyx_kp_u__91, __pyx_k__91, sizeof(__pyx_k__91), 0, 1, 0, 0},
  {&__pyx_kp_u__93, __pyx_k__93, sizeof(__pyx_k__93), 0, 1, 0, 0},
  {&__pyx_kp_u__95, __pyx_k__95, sizeof(__pyx_k__95), 0, 1, 0, 0},
  {&__pyx_kp_u__97, __pyx_k__97, sizeof(__pyx_k__97), 0, 1, 0, 0},
  {&__pyx_n_s__975, __pyx_k__975, sizeof(__pyx_k__975), 0, 0, 1, 1},
  {&__pyx_kp_u__978, __pyx_k__978, sizeof(__pyx_k__978), 0, 1, 0, 0},
  {&__pyx_kp_u__979, __pyx_k__979, sizeof(__pyx_k__979), 0, 1, 0, 0},
  {&__pyx_kp_u__980, __pyx_k__980, sizeof(__pyx_k__980), 0, 1, 0, 0},
  {&__pyx_kp_u__981, __pyx_k__981, sizeof(__pyx_k__981), 0, 1, 0, 0},
  {&__pyx_kp_u__982, __pyx_k__982, sizeof(__pyx_k__982), 0, 1, 0, 0},
  {&__pyx_kp_u__983, __pyx_k__983, sizeof(__pyx_k__983), 0, 1, 0, 0},
  {&__pyx_kp_u__984, __pyx_k__984, sizeof(__pyx_k__984), 0, 1, 0, 0},
  {&__pyx_kp_u__985, __pyx_k__985, sizeof(__pyx_k__985), 0, 1, 0, 0},
  {&__pyx_kp_u__986, __pyx_k__986, sizeof(__pyx_k__986), 0, 1, 0, 0},
  {&__pyx_kp_u__987, __pyx_k__987, sizeof(__pyx_k__987), 0, 1, 0, 0},
  {&__pyx_kp_u__988, __pyx_k__988, sizeof(__pyx_k__988), 0, 1, 0, 0},
  {&__pyx_kp_u__989, __pyx_k__989, sizeof(__pyx_k__989), 0, 1, 0, 0},
  {&__pyx_kp_u__99, __pyx_k__99, sizeof(__pyx_k__99), 0, 1, 0, 0},
  {&__pyx_kp_u__990, __pyx_k__990, sizeof(__pyx_k__990), 0, 1, 0, 0},
  {&__pyx_kp_u__991, __pyx_k__991, sizeof(__pyx_k__991), 0, 1, 0, 0},
  {&__pyx_kp_u__992, __pyx_k__992, sizeof(__pyx_k__992), 0, 1, 0, 0},
  {&__pyx_kp_u__993, __pyx_k__993, sizeof(__pyx_k__993), 0, 1, 0, 0},
  {&__pyx_kp_u__994, __pyx_k__994, sizeof(__pyx_k__994), 0, 1, 0, 0},
  {&__pyx_kp_u__995, __pyx_k__995, sizeof(__pyx_k__995), 0, 1, 0, 0},
  {&__pyx_kp_u__996, __pyx_k__996, sizeof(__pyx_k__996), 0, 1, 0, 0},
  {&__pyx_kp_u__997, __pyx_k__997, sizeof(__pyx_k__997), 0, 1, 0, 0},
  {&__pyx_kp_u__998, __pyx_k__998, sizeof(__pyx_k__998), 0, 1, 0, 0},
  {&__pyx_kp_u__999, __pyx_k__999, sizeof(__pyx_k__999), 0, 1, 0, 0},
  {&__pyx_n_s_accept, __pyx_k_accept, sizeof(__pyx_k_accept), 0, 0, 1, 1},
  {&__pyx_n_s_adaptivePredict, __pyx_k_adaptivePredict, sizeof(__pyx_k_adaptivePredict), 0, 0, 1, 1},
  {&__pyx_kp_u_algorithm, __pyx_k_algorithm, sizeof(__pyx_k_algorithm), 0, 1, 0, 0},
  {&__pyx_n_s_algorithm_section, __pyx_k_algorithm_section, sizeof(__pyx_k_algorithm_section), 0, 0, 1, 1},
  {&__pyx_n_u_algorithm_section, __pyx_k_algorithm_section, sizeof(__pyx_k_algorithm_section), 0, 1, 0, 1},
  {&__pyx_n_s_alt, __pyx_k_alt, sizeof(__pyx_k_alt), 0, 0, 1, 1},
  {&__pyx_kp_u_and, __pyx_k_and, sizeof(__pyx_k_and), 0, 1, 0, 0},
  {&__pyx_n_s_annotation, __pyx_k_annotation, sizeof(__pyx_k_annotation), 0, 0, 1, 1},
  {&__pyx_n_u_annotation, __pyx_k_annotation, sizeof(__pyx_k_annotation), 0, 1, 0, 1},
  {&__pyx_kp_u_annotation_2, __pyx_k_annotation_2, sizeof(__pyx_k_annotation_2), 0, 1, 0, 0},
  {&__pyx_n_s_antlr4, __pyx_k_antlr4, sizeof(__pyx_k_antlr4), 0, 0, 1, 1},
  {&__pyx_n_s_antlr4_error_Errors, __pyx_k_antlr4_error_Errors, sizeof(__pyx_k_antlr4_error_Errors), 0, 0, 1, 1},
  {&__pyx_n_s_argument, __pyx_k_argument, sizeof(__pyx_k_argument), 0, 0, 1, 1},
  {&__pyx_n_u_argument, __pyx_k_argument, sizeof(__pyx_k_argument), 0, 1, 0, 1},
  {&__pyx_n_s_argument_list, __pyx_k_argument_list, sizeof(__pyx_k_argument_list), 0, 0, 1, 1},
  {&__pyx_n_u_argument_list, __pyx_k_argument_list, sizeof(__pyx_k_argument_list), 0, 1, 0, 1},
  {&__pyx_n_s_array_subscripts, __pyx_k_array_subscripts, sizeof(__pyx_k_array_subscripts), 0, 0, 1, 1},
  {&__pyx_n_u_array_subscripts, __pyx_k_array_subscripts, sizeof(__pyx_k_array_subscripts), 0, 1, 0, 1},
  {&__pyx_n_s_atn, __pyx_k_atn, sizeof(__pyx_k_atn), 0, 0, 1, 1},
  {&__pyx_kp_u_b_d_f, __pyx_k_b_d_f, sizeof(__pyx_k_b_d_f), 0, 1, 0, 0},
  {&__pyx_n_s_base_prefix, __pyx_k_base_prefix, sizeof(__pyx_k_base_prefix), 0, 0, 1, 1},
  {&__pyx_n_u_base_prefix, __pyx_k_base_prefix, sizeof(__pyx_k_base_prefix), 0, 1, 0, 1},
  {&__pyx_kp_u_block, __pyx_k_block, sizeof(__pyx_k_block), 0, 1, 0, 0},
  {&__pyx_kp_u_break, __pyx_k_break, sizeof(__pyx_k_break), 0, 1, 0, 0},
  {&__pyx_n_s_buf, __pyx_k_buf, sizeof(__pyx_k_buf), 0, 0, 1, 1},
  {&__pyx_kp_u_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 1, 0, 0},
  {&__pyx_n_s_checkVersion, __pyx_k_checkVersion, sizeof(__pyx_k_checkVersion), 0, 0, 1, 1},
  {&__pyx_kp_u_class, __pyx_k_class, sizeof(__pyx_k_class), 0, 1, 0, 0},
  {&__pyx_n_s_class_definition, __pyx_k_class_definition, sizeof(__pyx_k_class_definition), 0, 0, 1, 1},
  {&__pyx_n_u_class_definition, __pyx_k_class_definition, sizeof(__pyx_k_class_definition), 0, 1, 0, 1},
  {&__pyx_n_s_class_elem, __pyx_k_class_elem, sizeof(__pyx_k_class_elem), 0, 0, 1, 1},
  {&__pyx_n_s_class_modification, __pyx_k_class_modification, sizeof(__pyx_k_class_modification), 0, 0, 1, 1},
  {&__pyx_n_u_class_modification, __pyx_k_class_modification, sizeof(__pyx_k_class_modification), 0, 1, 0, 1},
  {&__pyx_n_s_class_prefixes, __pyx_k_class_prefixes, sizeof(__pyx_k_class_prefixes), 0, 0, 1, 1},
  {&__pyx_n_u_class_prefixes, __pyx_k_class_prefixes, sizeof(__pyx_k_class_prefixes), 0, 1, 0, 1},
  {&__pyx_n_s_class_specifier, __pyx_k_class_specifier, sizeof(__pyx_k_class_specifier), 0, 0, 1, 1},
  {&__pyx_n_u_class_specifier, __pyx_k_class_specifier, sizeof(__pyx_k_class_specifier), 0, 1, 0, 1},
  {&__pyx_n_s_class_type, __pyx_k_class_type, sizeof(__pyx_k_class_type), 0, 0, 1, 1},
  {&__pyx_n_u_class_type, __pyx_k_class_type, sizeof(__pyx_k_class_type), 0, 1, 0, 1},
  {&__pyx_n_s_comment, __pyx_k_comment, sizeof(__pyx_k_comment), 0, 0, 1, 1},
  {&__pyx_n_u_comment, __pyx_k_comment, sizeof(__pyx_k_comment), 0, 1, 0, 1},
  {&__pyx_n_s_comp_annotation, __pyx_k_comp_annotation, sizeof(__pyx_k_comp_annotation), 0, 0, 1, 1},
  {&__pyx_n_s_comp_elem, __pyx_k_comp_elem, sizeof(__pyx_k_comp_elem), 0, 0, 1, 1},
  {&__pyx_n_s_component_clause, __pyx_k_component_clause, sizeof(__pyx_k_component_clause), 0, 0, 1, 1},
  {&__pyx_n_u_component_clause, __pyx_k_component_clause, sizeof(__pyx_k_component_clause), 0, 1, 0, 1},
  {&__pyx_n_s_component_clause1, __pyx_k_component_clause1, sizeof(__pyx_k_component_clause1), 0, 0, 1, 1},
  {&__pyx_n_u_component_clause1, __pyx_k_component_clause1, sizeof(__pyx_k_component_clause1), 0, 1, 0, 1},
  {&__pyx_n_s_component_declaration, __pyx_k_component_declaration, sizeof(__pyx_k_component_declaration), 0, 0, 1, 1},
  {&__pyx_n_u_component_declaration, __pyx_k_component_declaration, sizeof(__pyx_k_component_declaration), 0, 1, 0, 1},
  {&__pyx_n_s_component_declaration1, __pyx_k_component_declaration1, sizeof(__pyx_k_component_declaration1), 0, 0, 1, 1},
  {&__pyx_n_u_component_declaration1, __pyx_k_component_declaration1, sizeof(__pyx_k_component_declaration1), 0, 1, 0, 1},
  {&__pyx_n_s_component_list, __pyx_k_component_list, sizeof(__pyx_k_component_list), 0, 0, 1, 1},
  {&__pyx_n_u_component_list, __pyx_k_component_list, sizeof(__pyx_k_component_list), 0, 1, 0, 1},
  {&__pyx_n_s_component_reference, __pyx_k_component_reference, sizeof(__pyx_k_component_reference), 0, 0, 1, 1},
  {&__pyx_n_u_component_reference, __pyx_k_component_reference, sizeof(__pyx_k_component_reference), 0, 1, 0, 1},
  {&__pyx_n_s_component_reference_element, __pyx_k_component_reference_element, sizeof(__pyx_k_component_reference_element), 0, 0, 1, 1},
  {&__pyx_n_u_component_reference_element, __pyx_k_component_reference_element, sizeof(__pyx_k_component_reference_element), 0, 1, 0, 1},
  {&__pyx_n_s_composition, __pyx_k_composition, sizeof(__pyx_k_composition), 0, 0, 1, 1},
  {&__pyx_n_u_composition, __pyx_k_composition, sizeof(__pyx_k_composition), 0, 1, 0, 1},
  {&__pyx_n_s_condition_attribute, __pyx_k_condition_attribute, sizeof(__pyx_k_condition_attribute), 0, 0, 1, 1},
  {&__pyx_n_u_condition_attribute, __pyx_k_condition_attribute, sizeof(__pyx_k_condition_attribute), 0, 1, 0, 1},
  {&__pyx_kp_u_connect, __pyx_k_connect, sizeof(__pyx_k_connect), 0, 1, 0, 0},
  {&__pyx_n_s_connect_clause, __pyx_k_connect_clause, sizeof(__pyx_k_connect_clause), 0, 0, 1, 1},
  {&__pyx_n_u_connect_clause, __pyx_k_connect_clause, sizeof(__pyx_k_connect_clause), 0, 1, 0, 1},
  {&__pyx_kp_u_connector, __pyx_k_connector, sizeof(__pyx_k_connector), 0, 1, 0, 0},
  {&__pyx_kp_u_constant, __pyx_k_constant, sizeof(__pyx_k_constant), 0, 1, 0, 0},
  {&__pyx_kp_u_constrainedby, __pyx_k_constrainedby, sizeof(__pyx_k_constrainedby), 0, 1, 0, 0},
  {&__pyx_n_s_constraining_clause, __pyx_k_constraining_clause, sizeof(__pyx_k_constraining_clause), 0, 0, 1, 1},
  {&__pyx_n_u_constraining_clause, __pyx_k_constraining_clause, sizeof(__pyx_k_constraining_clause), 0, 1, 0, 1},
  {&__pyx_n_s_consume, __pyx_k_consume, sizeof(__pyx_k_consume), 0, 0, 1, 1},
  {&__pyx_n_s_copyFrom, __pyx_k_copyFrom, sizeof(__pyx_k_copyFrom), 0, 0, 1, 1},
  {&__pyx_n_s_ctx, __pyx_k_ctx, sizeof(__pyx_k_ctx), 0, 0, 1, 1},
  {&__pyx_n_s_ctx_2, __pyx_k_ctx_2, sizeof(__pyx_k_ctx_2), 0, 0, 1, 1},
  {&__pyx_kp_u_d3, __pyx_k_d3, sizeof(__pyx_k_d3), 0, 1, 0, 0},
  {&__pyx_n_s_decisionToState, __pyx_k_decisionToState, sizeof(__pyx_k_decisionToState), 0, 0, 1, 1},
  {&__pyx_n_s_decisionsToDFA, __pyx_k_decisionsToDFA, sizeof(__pyx_k_decisionsToDFA), 0, 0, 1, 1},
  {&__pyx_n_s_declaration, __pyx_k_declaration, sizeof(__pyx_k_declaration), 0, 0, 1, 1},
  {&__pyx_n_u_declaration, __pyx_k_declaration, sizeof(__pyx_k_declaration), 0, 1, 0, 1},
  {&__pyx_kp_u_der, __pyx_k_der, sizeof(__pyx_k_der), 0, 1, 0, 0},
  {&__pyx_n_s_deserialize, __pyx_k_deserialize, sizeof(__pyx_k_deserialize), 0, 0, 1, 1},
  {&__pyx_kp_u_discrete, __pyx_k_discrete, sizeof(__pyx_k_discrete), 0, 1, 0, 0},
  {&__pyx_n_s_doc, __pyx_k_doc, sizeof(__pyx_k_doc), 0, 0, 1, 1},
  {&__pyx_kp_u_e, __pyx_k_e, sizeof(__pyx_k_e), 0, 1, 0, 0},
  {&__pyx_kp_u_each, __pyx_k_each, sizeof(__pyx_k_each), 0, 1, 0, 0},
  {&__pyx_n_s_element, __pyx_k_element, sizeof(__pyx_k_element), 0, 0, 1, 1},
  {&__pyx_n_u_element, __pyx_k_element, sizeof(__pyx_k_element), 0, 1, 0, 1},
  {&__pyx_n_s_element_list, __pyx_k_element_list, sizeof(__pyx_k_element_list), 0, 0, 1, 1},
  {&__pyx_n_u_element_list, __pyx_k_element_list, sizeof(__pyx_k_element_list), 0, 1, 0, 1},
  {&__pyx_n_s_element_modification, __pyx_k_element_modification, sizeof(__pyx_k_element_modification), 0, 0, 1, 1},
  {&__pyx_n_u_element_modification, __pyx_k_element_modification, sizeof(__pyx_k_element_modification), 0, 1, 0, 1},
  {&__pyx_n_s_element_modification_or_replacea, __pyx_k_element_modification_or_replacea, sizeof(__pyx_k_element_modification_or_replacea), 0, 0, 1, 1},
  {&__pyx_n_u_element_modification_or_replacea, __pyx_k_element_modification_or_replacea, sizeof(__pyx_k_element_modification_or_replacea), 0, 1, 0, 1},
  {&__pyx_n_s_element_redeclaration, __pyx_k_element_redeclaration, sizeof(__pyx_k_element_redeclaration), 0, 0, 1, 1},
  {&__pyx_n_u_element_redeclaration, __pyx_k_element_redeclaration, sizeof(__pyx_k_element_redeclaration), 0, 1, 0, 1},
  {&__pyx_n_s_element_replaceable, __pyx_k_element_replaceable, sizeof(__pyx_k_element_replaceable), 0, 0, 1, 1},
  {&__pyx_n_u_element_replaceable, __pyx_k_element_replaceable, sizeof(__pyx_k_element_replaceable), 0, 1, 0, 1},
  {&__pyx_kp_u_else, __pyx_k_else, sizeof(__pyx_k_else), 0, 1, 0, 0},
  {&__pyx_kp_u_elseif, __pyx_k_elseif, sizeof(__pyx_k_elseif), 0, 1, 0, 0},
  {&__pyx_kp_u_elsewhen, __pyx_k_elsewhen, sizeof(__pyx_k_elsewhen), 0, 1, 0, 0},
  {&__pyx_kp_u_encapsulated, __pyx_k_encapsulated, sizeof(__pyx_k_encapsulated), 0, 1, 0, 0},
  {&__pyx_kp_u_end, __pyx_k_end, sizeof(__pyx_k_end), 0, 1, 0, 0},
  {&__pyx_n_s_enter, __pyx_k_enter, sizeof(__pyx_k_enter), 0, 0, 1, 1},
  {&__pyx_n_s_enterAlgorithm_section, __pyx_k_enterAlgorithm_section, sizeof(__pyx_k_enterAlgorithm_section), 0, 0, 1, 1},
  {&__pyx_n_u_enterAlgorithm_section, __pyx_k_enterAlgorithm_section, sizeof(__pyx_k_enterAlgorithm_section), 0, 1, 0, 1},
  {&__pyx_n_s_enterAnnotation, __pyx_k_enterAnnotation, sizeof(__pyx_k_enterAnnotation), 0, 0, 1, 1},
  {&__pyx_n_u_enterAnnotation, __pyx_k_enterAnnotation, sizeof(__pyx_k_enterAnnotation), 0, 1, 0, 1},
  {&__pyx_n_s_enterArgument, __pyx_k_enterArgument, sizeof(__pyx_k_enterArgument), 0, 0, 1, 1},
  {&__pyx_n_u_enterArgument, __pyx_k_enterArgument, sizeof(__pyx_k_enterArgument), 0, 1, 0, 1},
  {&__pyx_n_s_enterArgument_expression, __pyx_k_enterArgument_expression, sizeof(__pyx_k_enterArgument_expression), 0, 0, 1, 1},
  {&__pyx_n_u_enterArgument_expression, __pyx_k_enterArgument_expression, sizeof(__pyx_k_enterArgument_expression), 0, 1, 0, 1},
  {&__pyx_n_s_enterArgument_function, __pyx_k_enterArgument_function, sizeof(__pyx_k_enterArgument_function), 0, 0, 1, 1},
  {&__pyx_n_u_enterArgument_function, __pyx_k_enterArgument_function, sizeof(__pyx_k_enterArgument_function), 0, 1, 0, 1},
  {&__pyx_n_s_enterArgument_list, __pyx_k_enterArgument_list, sizeof(__pyx_k_enterArgument_list), 0, 0, 1, 1},
  {&__pyx_n_u_enterArgument_list, __pyx_k_enterArgument_list, sizeof(__pyx_k_enterArgument_list), 0, 1, 0, 1},
  {&__pyx_n_s_enterArray_subscripts, __pyx_k_enterArray_subscripts, sizeof(__pyx_k_enterArray_subscripts), 0, 0, 1, 1},
  {&__pyx_n_u_enterArray_subscripts, __pyx_k_enterArray_subscripts, sizeof(__pyx_k_enterArray_subscripts), 0, 1, 0, 1},
  {&__pyx_n_s_enterBase_prefix, __pyx_k_enterBase_prefix, sizeof(__pyx_k_enterBase_prefix), 0, 0, 1, 1},
  {&__pyx_n_u_enterBase_prefix, __pyx_k_enterBase_prefix, sizeof(__pyx_k_enterBase_prefix), 0, 1, 0, 1},
  {&__pyx_n_s_enterClass_definition, __pyx_k_enterClass_definition, sizeof(__pyx_k_enterClass_definition), 0, 0, 1, 1},
  {&__pyx_n_u_enterClass_definition, __pyx_k_enterClass_definition, sizeof(__pyx_k_enterClass_definition), 0, 1, 0, 1},
  {&__pyx_n_s_enterClass_modification, __pyx_k_enterClass_modification, sizeof(__pyx_k_enterClass_modification), 0, 0, 1, 1},
  {&__pyx_n_u_enterClass_modification, __pyx_k_enterClass_modification, sizeof(__pyx_k_enterClass_modification), 0, 1, 0, 1},
  {&__pyx_n_s_enterClass_prefixes, __pyx_k_enterClass_prefixes, sizeof(__pyx_k_enterClass_prefixes), 0, 0, 1, 1},
  {&__pyx_n_u_enterClass_prefixes, __pyx_k_enterClass_prefixes, sizeof(__pyx_k_enterClass_prefixes), 0, 1, 0, 1},
  {&__pyx_n_s_enterClass_spec_base, __pyx_k_enterClass_spec_base, sizeof(__pyx_k_enterClass_spec_base), 0, 0, 1, 1},
  {&__pyx_n_u_enterClass_spec_base, __pyx_k_enterClass_spec_base, sizeof(__pyx_k_enterClass_spec_base), 0, 1, 0, 1},
  {&__pyx_n_s_enterClass_spec_comp, __pyx_k_enterClass_spec_comp, sizeof(__pyx_k_enterClass_spec_comp), 0, 0, 1, 1},
  {&__pyx_n_u_enterClass_spec_comp, __pyx_k_enterClass_spec_comp, sizeof(__pyx_k_enterClass_spec_comp), 0, 1, 0, 1},
  {&__pyx_n_s_enterClass_spec_der, __pyx_k_enterClass_spec_der, sizeof(__pyx_k_enterClass_spec_der), 0, 0, 1, 1},
  {&__pyx_n_u_enterClass_spec_der, __pyx_k_enterClass_spec_der, sizeof(__pyx_k_enterClass_spec_der), 0, 1, 0, 1},
  {&__pyx_n_s_enterClass_spec_enum, __pyx_k_enterClass_spec_enum, sizeof(__pyx_k_enterClass_spec_enum), 0, 0, 1, 1},
  {&__pyx_n_u_enterClass_spec_enum, __pyx_k_enterClass_spec_enum, sizeof(__pyx_k_enterClass_spec_enum), 0, 1, 0, 1},
  {&__pyx_n_s_enterClass_spec_extends, __pyx_k_enterClass_spec_extends, sizeof(__pyx_k_enterClass_spec_extends), 0, 0, 1, 1},
  {&__pyx_n_u_enterClass_spec_extends, __pyx_k_enterClass_spec_extends, sizeof(__pyx_k_enterClass_spec_extends), 0, 1, 0, 1},
  {&__pyx_n_s_enterClass_type, __pyx_k_enterClass_type, sizeof(__pyx_k_enterClass_type), 0, 0, 1, 1},
  {&__pyx_n_u_enterClass_type, __pyx_k_enterClass_type, sizeof(__pyx_k_enterClass_type), 0, 1, 0, 1},
  {&__pyx_n_s_enterComment, __pyx_k_enterComment, sizeof(__pyx_k_enterComment), 0, 0, 1, 1},
  {&__pyx_n_u_enterComment, __pyx_k_enterComment, sizeof(__pyx_k_enterComment), 0, 1, 0, 1},
  {&__pyx_n_s_enterComponent_clause, __pyx_k_enterComponent_clause, sizeof(__pyx_k_enterComponent_clause), 0, 0, 1, 1},
  {&__pyx_n_u_enterComponent_clause, __pyx_k_enterComponent_clause, sizeof(__pyx_k_enterComponent_clause), 0, 1, 0, 1},
  {&__pyx_n_s_enterComponent_clause1, __pyx_k_enterComponent_clause1, sizeof(__pyx_k_enterComponent_clause1), 0, 0, 1, 1},
  {&__pyx_n_u_enterComponent_clause1, __pyx_k_enterComponent_clause1, sizeof(__pyx_k_enterComponent_clause1), 0, 1, 0, 1},
  {&__pyx_n_s_enterComponent_declaration, __pyx_k_enterComponent_declaration, sizeof(__pyx_k_enterComponent_declaration), 0, 0, 1, 1},
  {&__pyx_n_u_enterComponent_declaration, __pyx_k_enterComponent_declaration, sizeof(__pyx_k_enterComponent_declaration), 0, 1, 0, 1},
  {&__pyx_n_s_enterComponent_declaration1, __pyx_k_enterComponent_declaration1, sizeof(__pyx_k_enterComponent_declaration1), 0, 0, 1, 1},
  {&__pyx_n_u_enterComponent_declaration1, __pyx_k_enterComponent_declaration1, sizeof(__pyx_k_enterComponent_declaration1), 0, 1, 0, 1},
  {&__pyx_n_s_enterComponent_list, __pyx_k_enterComponent_list, sizeof(__pyx_k_enterComponent_list), 0, 0, 1, 1},
  {&__pyx_n_u_enterComponent_list, __pyx_k_enterComponent_list, sizeof(__pyx_k_enterComponent_list), 0, 1, 0, 1},
  {&__pyx_n_s_enterComponent_reference, __pyx_k_enterComponent_reference, sizeof(__pyx_k_enterComponent_reference), 0, 0, 1, 1},
  {&__pyx_n_u_enterComponent_reference, __pyx_k_enterComponent_reference, sizeof(__pyx_k_enterComponent_reference), 0, 1, 0, 1},
  {&__pyx_n_s_enterComponent_reference_element, __pyx_k_enterComponent_reference_element, sizeof(__pyx_k_enterComponent_reference_element), 0, 0, 1, 1},
  {&__pyx_n_u_enterComponent_reference_element, __pyx_k_enterComponent_reference_element, sizeof(__pyx_k_enterComponent_reference_element), 0, 1, 0, 1},
  {&__pyx_n_s_enterComposition, __pyx_k_enterComposition, sizeof(__pyx_k_enterComposition), 0, 0, 1, 1},
  {&__pyx_n_u_enterComposition, __pyx_k_enterComposition, sizeof(__pyx_k_enterComposition), 0, 1, 0, 1},
  {&__pyx_n_s_enterCondition_attribute, __pyx_k_enterCondition_attribute, sizeof(__pyx_k_enterCondition_attribute), 0, 0, 1, 1},
  {&__pyx_n_u_enterCondition_attribute, __pyx_k_enterCondition_attribute, sizeof(__pyx_k_enterCondition_attribute), 0, 1, 0, 1},
  {&__pyx_n_s_enterConnect_clause, __pyx_k_enterConnect_clause, sizeof(__pyx_k_enterConnect_clause), 0, 0, 1, 1},
  {&__pyx_n_u_enterConnect_clause, __pyx_k_enterConnect_clause, sizeof(__pyx_k_enterConnect_clause), 0, 1, 0, 1},
  {&__pyx_n_s_enterConstraining_clause, __pyx_k_enterConstraining_clause, sizeof(__pyx_k_enterConstraining_clause), 0, 0, 1, 1},
  {&__pyx_n_u_enterConstraining_clause, __pyx_k_enterConstraining_clause, sizeof(__pyx_k_enterConstraining_clause), 0, 1, 0, 1},
  {&__pyx_n_s_enterDeclaration, __pyx_k_enterDeclaration, sizeof(__pyx_k_enterDeclaration), 0, 0, 1, 1},
  {&__pyx_n_u_enterDeclaration, __pyx_k_enterDeclaration, sizeof(__pyx_k_enterDeclaration), 0, 1, 0, 1},
  {&__pyx_n_s_enterElement, __pyx_k_enterElement, sizeof(__pyx_k_enterElement), 0, 0, 1, 1},
  {&__pyx_n_u_enterElement, __pyx_k_enterElement, sizeof(__pyx_k_enterElement), 0, 1, 0, 1},
  {&__pyx_n_s_enterElement_list, __pyx_k_enterElement_list, sizeof(__pyx_k_enterElement_list), 0, 0, 1, 1},
  {&__pyx_n_u_enterElement_list, __pyx_k_enterElement_list, sizeof(__pyx_k_enterElement_list), 0, 1, 0, 1},
  {&__pyx_n_s_enterElement_modification, __pyx_k_enterElement_modification, sizeof(__pyx_k_enterElement_modification), 0, 0, 1, 1},
  {&__pyx_n_u_enterElement_modification, __pyx_k_enterElement_modification, sizeof(__pyx_k_enterElement_modification), 0, 1, 0, 1},
  {&__pyx_n_s_enterElement_modification_or_rep, __pyx_k_enterElement_modification_or_rep, sizeof(__pyx_k_enterElement_modification_or_rep), 0, 0, 1, 1},
  {&__pyx_n_u_enterElement_modification_or_rep, __pyx_k_enterElement_modification_or_rep, sizeof(__pyx_k_enterElement_modification_or_rep), 0, 1, 0, 1},
  {&__pyx_n_s_enterElement_redeclaration, __pyx_k_enterElement_redeclaration, sizeof(__pyx_k_enterElement_redeclaration), 0, 0, 1, 1},
  {&__pyx_n_u_enterElement_redeclaration, __pyx_k_enterElement_redeclaration, sizeof(__pyx_k_enterElement_redeclaration), 0, 1, 0, 1},
  {&__pyx_n_s_enterElement_replaceable, __pyx_k_enterElement_replaceable, sizeof(__pyx_k_enterElement_replaceable), 0, 0, 1, 1},
  {&__pyx_n_u_enterElement_replaceable, __pyx_k_enterElement_replaceable, sizeof(__pyx_k_enterElement_replaceable), 0, 1, 0, 1},
  {&__pyx_n_s_enterEnum_list, __pyx_k_enterEnum_list, sizeof(__pyx_k_enterEnum_list), 0, 0, 1, 1},
  {&__pyx_n_u_enterEnum_list, __pyx_k_enterEnum_list, sizeof(__pyx_k_enterEnum_list), 0, 1, 0, 1},
  {&__pyx_n_s_enterEnumeration_literal, __pyx_k_enterEnumeration_literal, sizeof(__pyx_k_enterEnumeration_literal), 0, 0, 1, 1},
  {&__pyx_n_u_enterEnumeration_literal, __pyx_k_enterEnumeration_literal, sizeof(__pyx_k_enterEnumeration_literal), 0, 1, 0, 1},
  {&__pyx_n_s_enterEquation, __pyx_k_enterEquation, sizeof(__pyx_k_enterEquation), 0, 0, 1, 1},
  {&__pyx_n_u_enterEquation, __pyx_k_enterEquation, sizeof(__pyx_k_enterEquation), 0, 1, 0, 1},
  {&__pyx_n_s_enterEquation_connect_clause, __pyx_k_enterEquation_connect_clause, sizeof(__pyx_k_enterEquation_connect_clause), 0, 0, 1, 1},
  {&__pyx_n_u_enterEquation_connect_clause, __pyx_k_enterEquation_connect_clause, sizeof(__pyx_k_enterEquation_connect_clause), 0, 1, 0, 1},
  {&__pyx_n_s_enterEquation_for, __pyx_k_enterEquation_for, sizeof(__pyx_k_enterEquation_for), 0, 0, 1, 1},
  {&__pyx_n_u_enterEquation_for, __pyx_k_enterEquation_for, sizeof(__pyx_k_enterEquation_for), 0, 1, 0, 1},
  {&__pyx_n_s_enterEquation_function, __pyx_k_enterEquation_function, sizeof(__pyx_k_enterEquation_function), 0, 0, 1, 1},
  {&__pyx_n_u_enterEquation_function, __pyx_k_enterEquation_function, sizeof(__pyx_k_enterEquation_function), 0, 1, 0, 1},
  {&__pyx_n_s_enterEquation_if, __pyx_k_enterEquation_if, sizeof(__pyx_k_enterEquation_if), 0, 0, 1, 1},
  {&__pyx_n_u_enterEquation_if, __pyx_k_enterEquation_if, sizeof(__pyx_k_enterEquation_if), 0, 1, 0, 1},
  {&__pyx_n_s_enterEquation_section, __pyx_k_enterEquation_section, sizeof(__pyx_k_enterEquation_section), 0, 0, 1, 1},
  {&__pyx_n_u_enterEquation_section, __pyx_k_enterEquation_section, sizeof(__pyx_k_enterEquation_section), 0, 1, 0, 1},
  {&__pyx_n_s_enterEquation_simple, __pyx_k_enterEquation_simple, sizeof(__pyx_k_enterEquation_simple), 0, 0, 1, 1},
  {&__pyx_n_u_enterEquation_simple, __pyx_k_enterEquation_simple, sizeof(__pyx_k_enterEquation_simple), 0, 1, 0, 1},
  {&__pyx_n_s_enterEquation_when, __pyx_k_enterEquation_when, sizeof(__pyx_k_enterEquation_when), 0, 0, 1, 1},
  {&__pyx_n_u_enterEquation_when, __pyx_k_enterEquation_when, sizeof(__pyx_k_enterEquation_when), 0, 1, 0, 1},
  {&__pyx_n_s_enterExpr_add, __pyx_k_enterExpr_add, sizeof(__pyx_k_enterExpr_add), 0, 0, 1, 1},
  {&__pyx_n_u_enterExpr_add, __pyx_k_enterExpr_add, sizeof(__pyx_k_enterExpr_add), 0, 1, 0, 1},
  {&__pyx_n_s_enterExpr_and, __pyx_k_enterExpr_and, sizeof(__pyx_k_enterExpr_and), 0, 0, 1, 1},
  {&__pyx_n_u_enterExpr_and, __pyx_k_enterExpr_and, sizeof(__pyx_k_enterExpr_and), 0, 1, 0, 1},
  {&__pyx_n_s_enterExpr_exp, __pyx_k_enterExpr_exp, sizeof(__pyx_k_enterExpr_exp), 0, 0, 1, 1},
  {&__pyx_n_u_enterExpr_exp, __pyx_k_enterExpr_exp, sizeof(__pyx_k_enterExpr_exp), 0, 1, 0, 1},
  {&__pyx_n_s_enterExpr_mul, __pyx_k_enterExpr_mul, sizeof(__pyx_k_enterExpr_mul), 0, 0, 1, 1},
  {&__pyx_n_u_enterExpr_mul, __pyx_k_enterExpr_mul, sizeof(__pyx_k_enterExpr_mul), 0, 1, 0, 1},
  {&__pyx_n_s_enterExpr_neg, __pyx_k_enterExpr_neg, sizeof(__pyx_k_enterExpr_neg), 0, 0, 1, 1},
  {&__pyx_n_u_enterExpr_neg, __pyx_k_enterExpr_neg, sizeof(__pyx_k_enterExpr_neg), 0, 1, 0, 1},
  {&__pyx_n_s_enterExpr_not, __pyx_k_enterExpr_not, sizeof(__pyx_k_enterExpr_not), 0, 0, 1, 1},
  {&__pyx_n_u_enterExpr_not, __pyx_k_enterExpr_not, sizeof(__pyx_k_enterExpr_not), 0, 1, 0, 1},
  {&__pyx_n_s_enterExpr_or, __pyx_k_enterExpr_or, sizeof(__pyx_k_enterExpr_or), 0, 0, 1, 1},
  {&__pyx_n_u_enterExpr_or, __pyx_k_enterExpr_or, sizeof(__pyx_k_enterExpr_or), 0, 1, 0, 1},
  {&__pyx_n_s_enterExpr_primary, __pyx_k_enterExpr_primary, sizeof(__pyx_k_enterExpr_primary), 0, 0, 1, 1},
  {&__pyx_n_u_enterExpr_primary, __pyx_k_enterExpr_primary, sizeof(__pyx_k_enterExpr_primary), 0, 1, 0, 1},
  {&__pyx_n_s_enterExpr_rel, __pyx_k_enterExpr_rel, sizeof(__pyx_k_enterExpr_rel), 0, 0, 1, 1},
  {&__pyx_n_u_enterExpr_rel, __pyx_k_enterExpr_rel, sizeof(__pyx_k_enterExpr_rel), 0, 1, 0, 1},
  {&__pyx_n_s_enterExpression_if, __pyx_k_enterExpression_if, sizeof(__pyx_k_enterExpression_if), 0, 0, 1, 1},
  {&__pyx_n_u_enterExpression_if, __pyx_k_enterExpression_if, sizeof(__pyx_k_enterExpression_if), 0, 1, 0, 1},
  {&__pyx_n_s_enterExpression_list, __pyx_k_enterExpression_list, sizeof(__pyx_k_enterExpression_list), 0, 0, 1, 1},
  {&__pyx_n_u_enterExpression_list, __pyx_k_enterExpression_list, sizeof(__pyx_k_enterExpression_list), 0, 1, 0, 1},
  {&__pyx_n_s_enterExpression_simple, __pyx_k_enterExpression_simple, sizeof(__pyx_k_enterExpression_simple), 0, 0, 1, 1},
  {&__pyx_n_u_enterExpression_simple, __pyx_k_enterExpression_simple, sizeof(__pyx_k_enterExpression_simple), 0, 1, 0, 1},
  {&__pyx_n_s_enterExtends_clause, __pyx_k_enterExtends_clause, sizeof(__pyx_k_enterExtends_clause), 0, 0, 1, 1},
  {&__pyx_n_u_enterExtends_clause, __pyx_k_enterExtends_clause, sizeof(__pyx_k_enterExtends_clause), 0, 1, 0, 1},
  {&__pyx_n_s_enterExternal_function_call, __pyx_k_enterExternal_function_call, sizeof(__pyx_k_enterExternal_function_call), 0, 0, 1, 1},
  {&__pyx_n_u_enterExternal_function_call, __pyx_k_enterExternal_function_call, sizeof(__pyx_k_enterExternal_function_call), 0, 1, 0, 1},
  {&__pyx_n_s_enterFor_equation, __pyx_k_enterFor_equation, sizeof(__pyx_k_enterFor_equation), 0, 0, 1, 1},
  {&__pyx_n_u_enterFor_equation, __pyx_k_enterFor_equation, sizeof(__pyx_k_enterFor_equation), 0, 1, 0, 1},
  {&__pyx_n_s_enterFor_index, __pyx_k_enterFor_index, sizeof(__pyx_k_enterFor_index), 0, 0, 1, 1},
  {&__pyx_n_u_enterFor_index, __pyx_k_enterFor_index, sizeof(__pyx_k_enterFor_index), 0, 1, 0, 1},
  {&__pyx_n_s_enterFor_indices, __pyx_k_enterFor_indices, sizeof(__pyx_k_enterFor_indices), 0, 0, 1, 1},
  {&__pyx_n_u_enterFor_indices, __pyx_k_enterFor_indices, sizeof(__pyx_k_enterFor_indices), 0, 1, 0, 1},
  {&__pyx_n_s_enterFor_statement, __pyx_k_enterFor_statement, sizeof(__pyx_k_enterFor_statement), 0, 0, 1, 1},
  {&__pyx_n_u_enterFor_statement, __pyx_k_enterFor_statement, sizeof(__pyx_k_enterFor_statement), 0, 1, 0, 1},
  {&__pyx_n_s_enterFunction_arguments, __pyx_k_enterFunction_arguments, sizeof(__pyx_k_enterFunction_arguments), 0, 0, 1, 1},
  {&__pyx_n_u_enterFunction_arguments, __pyx_k_enterFunction_arguments, sizeof(__pyx_k_enterFunction_arguments), 0, 1, 0, 1},
  {&__pyx_n_s_enterFunction_call_args, __pyx_k_enterFunction_call_args, sizeof(__pyx_k_enterFunction_call_args), 0, 0, 1, 1},
  {&__pyx_n_u_enterFunction_call_args, __pyx_k_enterFunction_call_args, sizeof(__pyx_k_enterFunction_call_args), 0, 1, 0, 1},
  {&__pyx_n_s_enterIf_equation, __pyx_k_enterIf_equation, sizeof(__pyx_k_enterIf_equation), 0, 0, 1, 1},
  {&__pyx_n_u_enterIf_equation, __pyx_k_enterIf_equation, sizeof(__pyx_k_enterIf_equation), 0, 1, 0, 1},
  {&__pyx_n_s_enterIf_statement, __pyx_k_enterIf_statement, sizeof(__pyx_k_enterIf_statement), 0, 0, 1, 1},
  {&__pyx_n_u_enterIf_statement, __pyx_k_enterIf_statement, sizeof(__pyx_k_enterIf_statement), 0, 1, 0, 1},
  {&__pyx_n_s_enterImport_clause, __pyx_k_enterImport_clause, sizeof(__pyx_k_enterImport_clause), 0, 0, 1, 1},
  {&__pyx_n_u_enterImport_clause, __pyx_k_enterImport_clause, sizeof(__pyx_k_enterImport_clause), 0, 1, 0, 1},
  {&__pyx_n_s_enterImport_list, __pyx_k_enterImport_list, sizeof(__pyx_k_enterImport_list), 0, 0, 1, 1},
  {&__pyx_n_u_enterImport_list, __pyx_k_enterImport_list, sizeof(__pyx_k_enterImport_list), 0, 1, 0, 1},
  {&__pyx_n_s_enterLanguage_specification, __pyx_k_enterLanguage_specification, sizeof(__pyx_k_enterLanguage_specification), 0, 0, 1, 1},
  {&__pyx_n_u_enterLanguage_specification, __pyx_k_enterLanguage_specification, sizeof(__pyx_k_enterLanguage_specification), 0, 1, 0, 1},
  {&__pyx_n_s_enterModification_assignment, __pyx_k_enterModification_assignment, sizeof(__pyx_k_enterModification_assignment), 0, 0, 1, 1},
  {&__pyx_n_u_enterModification_assignment, __pyx_k_enterModification_assignment, sizeof(__pyx_k_enterModification_assignment), 0, 1, 0, 1},
  {&__pyx_n_s_enterModification_assignment2, __pyx_k_enterModification_assignment2, sizeof(__pyx_k_enterModification_assignment2), 0, 0, 1, 1},
  {&__pyx_n_u_enterModification_assignment2, __pyx_k_enterModification_assignment2, sizeof(__pyx_k_enterModification_assignment2), 0, 1, 0, 1},
  {&__pyx_n_s_enterModification_class, __pyx_k_enterModification_class, sizeof(__pyx_k_enterModification_class), 0, 0, 1, 1},
  {&__pyx_n_u_enterModification_class, __pyx_k_enterModification_class, sizeof(__pyx_k_enterModification_class), 0, 1, 0, 1},
  {&__pyx_n_s_enterName, __pyx_k_enterName, sizeof(__pyx_k_enterName), 0, 0, 1, 1},
  {&__pyx_n_u_enterName, __pyx_k_enterName, sizeof(__pyx_k_enterName), 0, 1, 0, 1},
  {&__pyx_n_s_enterNamed_argument, __pyx_k_enterNamed_argument, sizeof(__pyx_k_enterNamed_argument), 0, 0, 1, 1},
  {&__pyx_n_u_enterNamed_argument, __pyx_k_enterNamed_argument, sizeof(__pyx_k_enterNamed_argument), 0, 1, 0, 1},
  {&__pyx_n_s_enterNamed_arguments, __pyx_k_enterNamed_arguments, sizeof(__pyx_k_enterNamed_arguments), 0, 0, 1, 1},
  {&__pyx_n_u_enterNamed_arguments, __pyx_k_enterNamed_arguments, sizeof(__pyx_k_enterNamed_arguments), 0, 1, 0, 1},
  {&__pyx_n_s_enterOuterAlt, __pyx_k_enterOuterAlt, sizeof(__pyx_k_enterOuterAlt), 0, 0, 1, 1},
  {&__pyx_n_s_enterOutput_expression_list, __pyx_k_enterOutput_expression_list, sizeof(__pyx_k_enterOutput_expression_list), 0, 0, 1, 1},
  {&__pyx_n_u_enterOutput_expression_list, __pyx_k_enterOutput_expression_list, sizeof(__pyx_k_enterOutput_expression_list), 0, 1, 0, 1},
  {&__pyx_n_s_enterPrimary_component_reference, __pyx_k_enterPrimary_component_reference, sizeof(__pyx_k_enterPrimary_component_reference), 0, 0, 1, 1},
  {&__pyx_n_u_enterPrimary_component_reference, __pyx_k_enterPrimary_component_reference, sizeof(__pyx_k_enterPrimary_component_reference), 0, 1, 0, 1},
  {&__pyx_n_s_enterPrimary_derivative, __pyx_k_enterPrimary_derivative, sizeof(__pyx_k_enterPrimary_derivative), 0, 0, 1, 1},
  {&__pyx_n_u_enterPrimary_derivative, __pyx_k_enterPrimary_derivative, sizeof(__pyx_k_enterPrimary_derivative), 0, 1, 0, 1},
  {&__pyx_n_s_enterPrimary_end, __pyx_k_enterPrimary_end, sizeof(__pyx_k_enterPrimary_end), 0, 0, 1, 1},
  {&__pyx_n_u_enterPrimary_end, __pyx_k_enterPrimary_end, sizeof(__pyx_k_enterPrimary_end), 0, 1, 0, 1},
  {&__pyx_n_s_enterPrimary_expression_list, __pyx_k_enterPrimary_expression_list, sizeof(__pyx_k_enterPrimary_expression_list), 0, 0, 1, 1},
  {&__pyx_n_u_enterPrimary_expression_list, __pyx_k_enterPrimary_expression_list, sizeof(__pyx_k_enterPrimary_expression_list), 0, 1, 0, 1},
  {&__pyx_n_s_enterPrimary_false, __pyx_k_enterPrimary_false, sizeof(__pyx_k_enterPrimary_false), 0, 0, 1, 1},
  {&__pyx_n_u_enterPrimary_false, __pyx_k_enterPrimary_false, sizeof(__pyx_k_enterPrimary_false), 0, 1, 0, 1},
  {&__pyx_n_s_enterPrimary_function, __pyx_k_enterPrimary_function, sizeof(__pyx_k_enterPrimary_function), 0, 0, 1, 1},
  {&__pyx_n_u_enterPrimary_function, __pyx_k_enterPrimary_function, sizeof(__pyx_k_enterPrimary_function), 0, 1, 0, 1},
  {&__pyx_n_s_enterPrimary_function_arguments, __pyx_k_enterPrimary_function_arguments, sizeof(__pyx_k_enterPrimary_function_arguments), 0, 0, 1, 1},
  {&__pyx_n_u_enterPrimary_function_arguments, __pyx_k_enterPrimary_function_arguments, sizeof(__pyx_k_enterPrimary_function_arguments), 0, 1, 0, 1},
  {&__pyx_n_s_enterPrimary_initial, __pyx_k_enterPrimary_initial, sizeof(__pyx_k_enterPrimary_initial), 0, 0, 1, 1},
  {&__pyx_n_u_enterPrimary_initial, __pyx_k_enterPrimary_initial, sizeof(__pyx_k_enterPrimary_initial), 0, 1, 0, 1},
  {&__pyx_n_s_enterPrimary_output_expression_l, __pyx_k_enterPrimary_output_expression_l, sizeof(__pyx_k_enterPrimary_output_expression_l), 0, 0, 1, 1},
  {&__pyx_n_u_enterPrimary_output_expression_l, __pyx_k_enterPrimary_output_expression_l, sizeof(__pyx_k_enterPrimary_output_expression_l), 0, 1, 0, 1},
  {&__pyx_n_s_enterPrimary_string, __pyx_k_enterPrimary_string, sizeof(__pyx_k_enterPrimary_string), 0, 0, 1, 1},
  {&__pyx_n_u_enterPrimary_string, __pyx_k_enterPrimary_string, sizeof(__pyx_k_enterPrimary_string), 0, 1, 0, 1},
  {&__pyx_n_s_enterPrimary_true, __pyx_k_enterPrimary_true, sizeof(__pyx_k_enterPrimary_true), 0, 0, 1, 1},
  {&__pyx_n_u_enterPrimary_true, __pyx_k_enterPrimary_true, sizeof(__pyx_k_enterPrimary_true), 0, 1, 0, 1},
  {&__pyx_n_s_enterPrimary_unsigned_number, __pyx_k_enterPrimary_unsigned_number, sizeof(__pyx_k_enterPrimary_unsigned_number), 0, 0, 1, 1},
  {&__pyx_n_u_enterPrimary_unsigned_number, __pyx_k_enterPrimary_unsigned_number, sizeof(__pyx_k_enterPrimary_unsigned_number), 0, 1, 0, 1},
  {&__pyx_n_s_enterRecursionRule, __pyx_k_enterRecursionRule, sizeof(__pyx_k_enterRecursionRule), 0, 0, 1, 1},
  {&__pyx_n_s_enterRegular_element, __pyx_k_enterRegular_element, sizeof(__pyx_k_enterRegular_element), 0, 0, 1, 1},
  {&__pyx_n_u_enterRegular_element, __pyx_k_enterRegular_element, sizeof(__pyx_k_enterRegular_element), 0, 1, 0, 1},
  {&__pyx_n_s_enterReplaceable_element, __pyx_k_enterReplaceable_element, sizeof(__pyx_k_enterReplaceable_element), 0, 0, 1, 1},
  {&__pyx_n_u_enterReplaceable_element, __pyx_k_enterReplaceable_element, sizeof(__pyx_k_enterReplaceable_element), 0, 1, 0, 1},
  {&__pyx_n_s_enterRule, __pyx_k_enterRule, sizeof(__pyx_k_enterRule), 0, 0, 1, 1},
  {&__pyx_n_s_enterShort_class_definition, __pyx_k_enterShort_class_definition, sizeof(__pyx_k_enterShort_class_definition), 0, 0, 1, 1},
  {&__pyx_n_u_enterShort_class_definition, __pyx_k_enterShort_class_definition, sizeof(__pyx_k_enterShort_class_definition), 0, 1, 0, 1},
  {&__pyx_n_s_enterSimple_expression, __pyx_k_enterSimple_expression, sizeof(__pyx_k_enterSimple_expression), 0, 0, 1, 1},
  {&__pyx_n_u_enterSimple_expression, __pyx_k_enterSimple_expression, sizeof(__pyx_k_enterSimple_expression), 0, 1, 0, 1},
  {&__pyx_n_s_enterStatement, __pyx_k_enterStatement, sizeof(__pyx_k_enterStatement), 0, 0, 1, 1},
  {&__pyx_n_u_enterStatement, __pyx_k_enterStatement, sizeof(__pyx_k_enterStatement), 0, 1, 0, 1},
  {&__pyx_n_s_enterStatement_break, __pyx_k_enterStatement_break, sizeof(__pyx_k_enterStatement_break), 0, 0, 1, 1},
  {&__pyx_n_u_enterStatement_break, __pyx_k_enterStatement_break, sizeof(__pyx_k_enterStatement_break), 0, 1, 0, 1},
  {&__pyx_n_s_enterStatement_component_functio, __pyx_k_enterStatement_component_functio, sizeof(__pyx_k_enterStatement_component_functio), 0, 0, 1, 1},
  {&__pyx_n_u_enterStatement_component_functio, __pyx_k_enterStatement_component_functio, sizeof(__pyx_k_enterStatement_component_functio), 0, 1, 0, 1},
  {&__pyx_n_s_enterStatement_component_referen, __pyx_k_enterStatement_component_referen, sizeof(__pyx_k_enterStatement_component_referen), 0, 0, 1, 1},
  {&__pyx_n_u_enterStatement_component_referen, __pyx_k_enterStatement_component_referen, sizeof(__pyx_k_enterStatement_component_referen), 0, 1, 0, 1},
  {&__pyx_n_s_enterStatement_for, __pyx_k_enterStatement_for, sizeof(__pyx_k_enterStatement_for), 0, 0, 1, 1},
  {&__pyx_n_u_enterStatement_for, __pyx_k_enterStatement_for, sizeof(__pyx_k_enterStatement_for), 0, 1, 0, 1},
  {&__pyx_n_s_enterStatement_if, __pyx_k_enterStatement_if, sizeof(__pyx_k_enterStatement_if), 0, 0, 1, 1},
  {&__pyx_n_u_enterStatement_if, __pyx_k_enterStatement_if, sizeof(__pyx_k_enterStatement_if), 0, 1, 0, 1},
  {&__pyx_n_s_enterStatement_return, __pyx_k_enterStatement_return, sizeof(__pyx_k_enterStatement_return), 0, 0, 1, 1},
  {&__pyx_n_u_enterStatement_return, __pyx_k_enterStatement_return, sizeof(__pyx_k_enterStatement_return), 0, 1, 0, 1},
  {&__pyx_n_s_enterStatement_when, __pyx_k_enterStatement_when, sizeof(__pyx_k_enterStatement_when), 0, 0, 1, 1},
  {&__pyx_n_u_enterStatement_when, __pyx_k_enterStatement_when, sizeof(__pyx_k_enterStatement_when), 0, 1, 0, 1},
  {&__pyx_n_s_enterStatement_while, __pyx_k_enterStatement_while, sizeof(__pyx_k_enterStatement_while), 0, 0, 1, 1},
  {&__pyx_n_u_enterStatement_while, __pyx_k_enterStatement_while, sizeof(__pyx_k_enterStatement_while), 0, 1, 0, 1},
  {&__pyx_n_s_enterStored_definition, __pyx_k_enterStored_definition, sizeof(__pyx_k_enterStored_definition), 0, 0, 1, 1},
  {&__pyx_n_u_enterStored_definition, __pyx_k_enterStored_definition, sizeof(__pyx_k_enterStored_definition), 0, 1, 0, 1},
  {&__pyx_n_s_enterStored_definition_class, __pyx_k_enterStored_definition_class, sizeof(__pyx_k_enterStored_definition_class), 0, 0, 1, 1},
  {&__pyx_n_u_enterStored_definition_class, __pyx_k_enterStored_definition_class, sizeof(__pyx_k_enterStored_definition_class), 0, 1, 0, 1},
  {&__pyx_n_s_enterString_comment, __pyx_k_enterString_comment, sizeof(__pyx_k_enterString_comment), 0, 0, 1, 1},
  {&__pyx_n_u_enterString_comment, __pyx_k_enterString_comment, sizeof(__pyx_k_enterString_comment), 0, 1, 0, 1},
  {&__pyx_n_s_enterSubscript, __pyx_k_enterSubscript, sizeof(__pyx_k_enterSubscript), 0, 0, 1, 1},
  {&__pyx_n_u_enterSubscript, __pyx_k_enterSubscript, sizeof(__pyx_k_enterSubscript), 0, 1, 0, 1},
  {&__pyx_n_s_enterType_prefix, __pyx_k_enterType_prefix, sizeof(__pyx_k_enterType_prefix), 0, 0, 1, 1},
  {&__pyx_n_u_enterType_prefix, __pyx_k_enterType_prefix, sizeof(__pyx_k_enterType_prefix), 0, 1, 0, 1},
  {&__pyx_n_s_enterType_specifier, __pyx_k_enterType_specifier, sizeof(__pyx_k_enterType_specifier), 0, 0, 1, 1},
  {&__pyx_n_u_enterType_specifier, __pyx_k_enterType_specifier, sizeof(__pyx_k_enterType_specifier), 0, 1, 0, 1},
  {&__pyx_n_s_enterWhen_equation, __pyx_k_enterWhen_equation, sizeof(__pyx_k_enterWhen_equation), 0, 0, 1, 1},
  {&__pyx_n_u_enterWhen_equation, __pyx_k_enterWhen_equation, sizeof(__pyx_k_enterWhen_equation), 0, 1, 0, 1},
  {&__pyx_n_s_enterWhen_statement, __pyx_k_enterWhen_statement, sizeof(__pyx_k_enterWhen_statement), 0, 0, 1, 1},
  {&__pyx_n_u_enterWhen_statement, __pyx_k_enterWhen_statement, sizeof(__pyx_k_enterWhen_statement), 0, 1, 0, 1},
  {&__pyx_n_s_enterWhile_statement, __pyx_k_enterWhile_statement, sizeof(__pyx_k_enterWhile_statement), 0, 0, 1, 1},
  {&__pyx_n_u_enterWhile_statement, __pyx_k_enterWhile_statement, sizeof(__pyx_k_enterWhile_statement), 0, 1, 0, 1},
  {&__pyx_n_s_enum_list, __pyx_k_enum_list, sizeof(__pyx_k_enum_list), 0, 0, 1, 1},
  {&__pyx_n_u_enum_list, __pyx_k_enum_list, sizeof(__pyx_k_enum_list), 0, 1, 0, 1},
  {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {&__pyx_kp_u_enumeration, __pyx_k_enumeration, sizeof(__pyx_k_enumeration), 0, 1, 0, 0},
  {&__pyx_n_s_enumeration_literal, __pyx_k_enumeration_literal, sizeof(__pyx_k_enumeration_literal), 0, 0, 1, 1},
  {&__pyx_n_u_enumeration_literal, __pyx_k_enumeration_literal, sizeof(__pyx_k_enumeration_literal), 0, 1, 0, 1},
  {&__pyx_n_s_epriv, __pyx_k_epriv, sizeof(__pyx_k_epriv), 0, 0, 1, 1},
  {&__pyx_n_s_epro, __pyx_k_epro, sizeof(__pyx_k_epro), 0, 0, 1, 1},
  {&__pyx_n_s_epub, __pyx_k_epub, sizeof(__pyx_k_epub), 0, 0, 1, 1},
  {&__pyx_n_s_equation, __pyx_k_equation, sizeof(__pyx_k_equation), 0, 0, 1, 1},
  {&__pyx_n_u_equation, __pyx_k_equation, sizeof(__pyx_k_equation), 0, 1, 0, 1},
  {&__pyx_kp_u_equation_2, __pyx_k_equation_2, sizeof(__pyx_k_equation_2), 0, 1, 0, 0},
  {&__pyx_n_s_equation_options, __pyx_k_equation_options, sizeof(__pyx_k_equation_options), 0, 0, 1, 1},
  {&__pyx_n_u_equation_options, __pyx_k_equation_options, sizeof(__pyx_k_equation_options), 0, 1, 0, 1},
  {&__pyx_n_s_equation_section, __pyx_k_equation_section, sizeof(__pyx_k_equation_section), 0, 0, 1, 1},
  {&__pyx_n_u_equation_section, __pyx_k_equation_section, sizeof(__pyx_k_equation_section), 0, 1, 0, 1},
  {&__pyx_n_s_errHandler, __pyx_k_errHandler, sizeof(__pyx_k_errHandler), 0, 0, 1, 1},
  {&__pyx_n_s_exception, __pyx_k_exception, sizeof(__pyx_k_exception), 0, 0, 1, 1},
  {&__pyx_n_s_exit, __pyx_k_exit, sizeof(__pyx_k_exit), 0, 0, 1, 1},
  {&__pyx_n_s_exitAlgorithm_section, __pyx_k_exitAlgorithm_section, sizeof(__pyx_k_exitAlgorithm_section), 0, 0, 1, 1},
  {&__pyx_n_u_exitAlgorithm_section, __pyx_k_exitAlgorithm_section, sizeof(__pyx_k_exitAlgorithm_section), 0, 1, 0, 1},
  {&__pyx_n_s_exitAnnotation, __pyx_k_exitAnnotation, sizeof(__pyx_k_exitAnnotation), 0, 0, 1, 1},
  {&__pyx_n_u_exitAnnotation, __pyx_k_exitAnnotation, sizeof(__pyx_k_exitAnnotation), 0, 1, 0, 1},
  {&__pyx_n_s_exitArgument, __pyx_k_exitArgument, sizeof(__pyx_k_exitArgument), 0, 0, 1, 1},
  {&__pyx_n_u_exitArgument, __pyx_k_exitArgument, sizeof(__pyx_k_exitArgument), 0, 1, 0, 1},
  {&__pyx_n_s_exitArgument_expression, __pyx_k_exitArgument_expression, sizeof(__pyx_k_exitArgument_expression), 0, 0, 1, 1},
  {&__pyx_n_u_exitArgument_expression, __pyx_k_exitArgument_expression, sizeof(__pyx_k_exitArgument_expression), 0, 1, 0, 1},
  {&__pyx_n_s_exitArgument_function, __pyx_k_exitArgument_function, sizeof(__pyx_k_exitArgument_function), 0, 0, 1, 1},
  {&__pyx_n_u_exitArgument_function, __pyx_k_exitArgument_function, sizeof(__pyx_k_exitArgument_function), 0, 1, 0, 1},
  {&__pyx_n_s_exitArgument_list, __pyx_k_exitArgument_list, sizeof(__pyx_k_exitArgument_list), 0, 0, 1, 1},
  {&__pyx_n_u_exitArgument_list, __pyx_k_exitArgument_list, sizeof(__pyx_k_exitArgument_list), 0, 1, 0, 1},
  {&__pyx_n_s_exitArray_subscripts, __pyx_k_exitArray_subscripts, sizeof(__pyx_k_exitArray_subscripts), 0, 0, 1, 1},
  {&__pyx_n_u_exitArray_subscripts, __pyx_k_exitArray_subscripts, sizeof(__pyx_k_exitArray_subscripts), 0, 1, 0, 1},
  {&__pyx_n_s_exitBase_prefix, __pyx_k_exitBase_prefix, sizeof(__pyx_k_exitBase_prefix), 0, 0, 1, 1},
  {&__pyx_n_u_exitBase_prefix, __pyx_k_exitBase_prefix, sizeof(__pyx_k_exitBase_prefix), 0, 1, 0, 1},
  {&__pyx_n_s_exitClass_definition, __pyx_k_exitClass_definition, sizeof(__pyx_k_exitClass_definition), 0, 0, 1, 1},
  {&__pyx_n_u_exitClass_definition, __pyx_k_exitClass_definition, sizeof(__pyx_k_exitClass_definition), 0, 1, 0, 1},
  {&__pyx_n_s_exitClass_modification, __pyx_k_exitClass_modification, sizeof(__pyx_k_exitClass_modification), 0, 0, 1, 1},
  {&__pyx_n_u_exitClass_modification, __pyx_k_exitClass_modification, sizeof(__pyx_k_exitClass_modification), 0, 1, 0, 1},
  {&__pyx_n_s_exitClass_prefixes, __pyx_k_exitClass_prefixes, sizeof(__pyx_k_exitClass_prefixes), 0, 0, 1, 1},
  {&__pyx_n_u_exitClass_prefixes, __pyx_k_exitClass_prefixes, sizeof(__pyx_k_exitClass_prefixes), 0, 1, 0, 1},
  {&__pyx_n_s_exitClass_spec_base, __pyx_k_exitClass_spec_base, sizeof(__pyx_k_exitClass_spec_base), 0, 0, 1, 1},
  {&__pyx_n_u_exitClass_spec_base, __pyx_k_exitClass_spec_base, sizeof(__pyx_k_exitClass_spec_base), 0, 1, 0, 1},
  {&__pyx_n_s_exitClass_spec_comp, __pyx_k_exitClass_spec_comp, sizeof(__pyx_k_exitClass_spec_comp), 0, 0, 1, 1},
  {&__pyx_n_u_exitClass_spec_comp, __pyx_k_exitClass_spec_comp, sizeof(__pyx_k_exitClass_spec_comp), 0, 1, 0, 1},
  {&__pyx_n_s_exitClass_spec_der, __pyx_k_exitClass_spec_der, sizeof(__pyx_k_exitClass_spec_der), 0, 0, 1, 1},
  {&__pyx_n_u_exitClass_spec_der, __pyx_k_exitClass_spec_der, sizeof(__pyx_k_exitClass_spec_der), 0, 1, 0, 1},
  {&__pyx_n_s_exitClass_spec_enum, __pyx_k_exitClass_spec_enum, sizeof(__pyx_k_exitClass_spec_enum), 0, 0, 1, 1},
  {&__pyx_n_u_exitClass_spec_enum, __pyx_k_exitClass_spec_enum, sizeof(__pyx_k_exitClass_spec_enum), 0, 1, 0, 1},
  {&__pyx_n_s_exitClass_spec_extends, __pyx_k_exitClass_spec_extends, sizeof(__pyx_k_exitClass_spec_extends), 0, 0, 1, 1},
  {&__pyx_n_u_exitClass_spec_extends, __pyx_k_exitClass_spec_extends, sizeof(__pyx_k_exitClass_spec_extends), 0, 1, 0, 1},
  {&__pyx_n_s_exitClass_type, __pyx_k_exitClass_type, sizeof(__pyx_k_exitClass_type), 0, 0, 1, 1},
  {&__pyx_n_u_exitClass_type, __pyx_k_exitClass_type, sizeof(__pyx_k_exitClass_type), 0, 1, 0, 1},
  {&__pyx_n_s_exitComment, __pyx_k_exitComment, sizeof(__pyx_k_exitComment), 0, 0, 1, 1},
  {&__pyx_n_u_exitComment, __pyx_k_exitComment, sizeof(__pyx_k_exitComment), 0, 1, 0, 1},
  {&__pyx_n_s_exitComponent_clause, __pyx_k_exitComponent_clause, sizeof(__pyx_k_exitComponent_clause), 0, 0, 1, 1},
  {&__pyx_n_u_exitComponent_clause, __pyx_k_exitComponent_clause, sizeof(__pyx_k_exitComponent_clause), 0, 1, 0, 1},
  {&__pyx_n_s_exitComponent_clause1, __pyx_k_exitComponent_clause1, sizeof(__pyx_k_exitComponent_clause1), 0, 0, 1, 1},
  {&__pyx_n_u_exitComponent_clause1, __pyx_k_exitComponent_clause1, sizeof(__pyx_k_exitComponent_clause1), 0, 1, 0, 1},
  {&__pyx_n_s_exitComponent_declaration, __pyx_k_exitComponent_declaration, sizeof(__pyx_k_exitComponent_declaration), 0, 0, 1, 1},
  {&__pyx_n_u_exitComponent_declaration, __pyx_k_exitComponent_declaration, sizeof(__pyx_k_exitComponent_declaration), 0, 1, 0, 1},
  {&__pyx_n_s_exitComponent_declaration1, __pyx_k_exitComponent_declaration1, sizeof(__pyx_k_exitComponent_declaration1), 0, 0, 1, 1},
  {&__pyx_n_u_exitComponent_declaration1, __pyx_k_exitComponent_declaration1, sizeof(__pyx_k_exitComponent_declaration1), 0, 1, 0, 1},
  {&__pyx_n_s_exitComponent_list, __pyx_k_exitComponent_list, sizeof(__pyx_k_exitComponent_list), 0, 0, 1, 1},
  {&__pyx_n_u_exitComponent_list, __pyx_k_exitComponent_list, sizeof(__pyx_k_exitComponent_list), 0, 1, 0, 1},
  {&__pyx_n_s_exitComponent_reference, __pyx_k_exitComponent_reference, sizeof(__pyx_k_exitComponent_reference), 0, 0, 1, 1},
  {&__pyx_n_u_exitComponent_reference, __pyx_k_exitComponent_reference, sizeof(__pyx_k_exitComponent_reference), 0, 1, 0, 1},
  {&__pyx_n_s_exitComponent_reference_element, __pyx_k_exitComponent_reference_element, sizeof(__pyx_k_exitComponent_reference_element), 0, 0, 1, 1},
  {&__pyx_n_u_exitComponent_reference_element, __pyx_k_exitComponent_reference_element, sizeof(__pyx_k_exitComponent_reference_element), 0, 1, 0, 1},
  {&__pyx_n_s_exitComposition, __pyx_k_exitComposition, sizeof(__pyx_k_exitComposition), 0, 0, 1, 1},
  {&__pyx_n_u_exitComposition, __pyx_k_exitComposition, sizeof(__pyx_k_exitComposition), 0, 1, 0, 1},
  {&__pyx_n_s_exitCondition_attribute, __pyx_k_exitCondition_attribute, sizeof(__pyx_k_exitCondition_attribute), 0, 0, 1, 1},
  {&__pyx_n_u_exitCondition_attribute, __pyx_k_exitCondition_attribute, sizeof(__pyx_k_exitCondition_attribute), 0, 1, 0, 1},
  {&__pyx_n_s_exitConnect_clause, __pyx_k_exitConnect_clause, sizeof(__pyx_k_exitConnect_clause), 0, 0, 1, 1},
  {&__pyx_n_u_exitConnect_clause, __pyx_k_exitConnect_clause, sizeof(__pyx_k_exitConnect_clause), 0, 1, 0, 1},
  {&__pyx_n_s_exitConstraining_clause, __pyx_k_exitConstraining_clause, sizeof(__pyx_k_exitConstraining_clause), 0, 0, 1, 1},
  {&__pyx_n_u_exitConstraining_clause, __pyx_k_exitConstraining_clause, sizeof(__pyx_k_exitConstraining_clause), 0, 1, 0, 1},
  {&__pyx_n_s_exitDeclaration, __pyx_k_exitDeclaration, sizeof(__pyx_k_exitDeclaration), 0, 0, 1, 1},
  {&__pyx_n_u_exitDeclaration, __pyx_k_exitDeclaration, sizeof(__pyx_k_exitDeclaration), 0, 1, 0, 1},
  {&__pyx_n_s_exitElement, __pyx_k_exitElement, sizeof(__pyx_k_exitElement), 0, 0, 1, 1},
  {&__pyx_n_u_exitElement, __pyx_k_exitElement, sizeof(__pyx_k_exitElement), 0, 1, 0, 1},
  {&__pyx_n_s_exitElement_list, __pyx_k_exitElement_list, sizeof(__pyx_k_exitElement_list), 0, 0, 1, 1},
  {&__pyx_n_u_exitElement_list, __pyx_k_exitElement_list, sizeof(__pyx_k_exitElement_list), 0, 1, 0, 1},
  {&__pyx_n_s_exitElement_modification, __pyx_k_exitElement_modification, sizeof(__pyx_k_exitElement_modification), 0, 0, 1, 1},
  {&__pyx_n_u_exitElement_modification, __pyx_k_exitElement_modification, sizeof(__pyx_k_exitElement_modification), 0, 1, 0, 1},
  {&__pyx_n_s_exitElement_modification_or_repl, __pyx_k_exitElement_modification_or_repl, sizeof(__pyx_k_exitElement_modification_or_repl), 0, 0, 1, 1},
  {&__pyx_n_u_exitElement_modification_or_repl, __pyx_k_exitElement_modification_or_repl, sizeof(__pyx_k_exitElement_modification_or_repl), 0, 1, 0, 1},
  {&__pyx_n_s_exitElement_redeclaration, __pyx_k_exitElement_redeclaration, sizeof(__pyx_k_exitElement_redeclaration), 0, 0, 1, 1},
  {&__pyx_n_u_exitElement_redeclaration, __pyx_k_exitElement_redeclaration, sizeof(__pyx_k_exitElement_redeclaration), 0, 1, 0, 1},
  {&__pyx_n_s_exitElement_replaceable, __pyx_k_exitElement_replaceable, sizeof(__pyx_k_exitElement_replaceable), 0, 0, 1, 1},
  {&__pyx_n_u_exitElement_replaceable, __pyx_k_exitElement_replaceable, sizeof(__pyx_k_exitElement_replaceable), 0, 1, 0, 1},
  {&__pyx_n_s_exitEnum_list, __pyx_k_exitEnum_list, sizeof(__pyx_k_exitEnum_list), 0, 0, 1, 1},
  {&__pyx_n_u_exitEnum_list, __pyx_k_exitEnum_list, sizeof(__pyx_k_exitEnum_list), 0, 1, 0, 1},
  {&__pyx_n_s_exitEnumeration_literal, __pyx_k_exitEnumeration_literal, sizeof(__pyx_k_exitEnumeration_literal), 0, 0, 1, 1},
  {&__pyx_n_u_exitEnumeration_literal, __pyx_k_exitEnumeration_literal, sizeof(__pyx_k_exitEnumeration_literal), 0, 1, 0, 1},
  {&__pyx_n_s_exitEquation, __pyx_k_exitEquation, sizeof(__pyx_k_exitEquation), 0, 0, 1, 1},
  {&__pyx_n_u_exitEquation, __pyx_k_exitEquation, sizeof(__pyx_k_exitEquation), 0, 1, 0, 1},
  {&__pyx_n_s_exitEquation_connect_clause, __pyx_k_exitEquation_connect_clause, sizeof(__pyx_k_exitEquation_connect_clause), 0, 0, 1, 1},
  {&__pyx_n_u_exitEquation_connect_clause, __pyx_k_exitEquation_connect_clause, sizeof(__pyx_k_exitEquation_connect_clause), 0, 1, 0, 1},
  {&__pyx_n_s_exitEquation_for, __pyx_k_exitEquation_for, sizeof(__pyx_k_exitEquation_for), 0, 0, 1, 1},
  {&__pyx_n_u_exitEquation_for, __pyx_k_exitEquation_for, sizeof(__pyx_k_exitEquation_for), 0, 1, 0, 1},
  {&__pyx_n_s_exitEquation_function, __pyx_k_exitEquation_function, sizeof(__pyx_k_exitEquation_function), 0, 0, 1, 1},
  {&__pyx_n_u_exitEquation_function, __pyx_k_exitEquation_function, sizeof(__pyx_k_exitEquation_function), 0, 1, 0, 1},
  {&__pyx_n_s_exitEquation_if, __pyx_k_exitEquation_if, sizeof(__pyx_k_exitEquation_if), 0, 0, 1, 1},
  {&__pyx_n_u_exitEquation_if, __pyx_k_exitEquation_if, sizeof(__pyx_k_exitEquation_if), 0, 1, 0, 1},
  {&__pyx_n_s_exitEquation_section, __pyx_k_exitEquation_section, sizeof(__pyx_k_exitEquation_section), 0, 0, 1, 1},
  {&__pyx_n_u_exitEquation_section, __pyx_k_exitEquation_section, sizeof(__pyx_k_exitEquation_section), 0, 1, 0, 1},
  {&__pyx_n_s_exitEquation_simple, __pyx_k_exitEquation_simple, sizeof(__pyx_k_exitEquation_simple), 0, 0, 1, 1},
  {&__pyx_n_u_exitEquation_simple, __pyx_k_exitEquation_simple, sizeof(__pyx_k_exitEquation_simple), 0, 1, 0, 1},
  {&__pyx_n_s_exitEquation_when, __pyx_k_exitEquation_when, sizeof(__pyx_k_exitEquation_when), 0, 0, 1, 1},
  {&__pyx_n_u_exitEquation_when, __pyx_k_exitEquation_when, sizeof(__pyx_k_exitEquation_when), 0, 1, 0, 1},
  {&__pyx_n_s_exitExpr_add, __pyx_k_exitExpr_add, sizeof(__pyx_k_exitExpr_add), 0, 0, 1, 1},
  {&__pyx_n_u_exitExpr_add, __pyx_k_exitExpr_add, sizeof(__pyx_k_exitExpr_add), 0, 1, 0, 1},
  {&__pyx_n_s_exitExpr_and, __pyx_k_exitExpr_and, sizeof(__pyx_k_exitExpr_and), 0, 0, 1, 1},
  {&__pyx_n_u_exitExpr_and, __pyx_k_exitExpr_and, sizeof(__pyx_k_exitExpr_and), 0, 1, 0, 1},
  {&__pyx_n_s_exitExpr_exp, __pyx_k_exitExpr_exp, sizeof(__pyx_k_exitExpr_exp), 0, 0, 1, 1},
  {&__pyx_n_u_exitExpr_exp, __pyx_k_exitExpr_exp, sizeof(__pyx_k_exitExpr_exp), 0, 1, 0, 1},
  {&__pyx_n_s_exitExpr_mul, __pyx_k_exitExpr_mul, sizeof(__pyx_k_exitExpr_mul), 0, 0, 1, 1},
  {&__pyx_n_u_exitExpr_mul, __pyx_k_exitExpr_mul, sizeof(__pyx_k_exitExpr_mul), 0, 1, 0, 1},
  {&__pyx_n_s_exitExpr_neg, __pyx_k_exitExpr_neg, sizeof(__pyx_k_exitExpr_neg), 0, 0, 1, 1},
  {&__pyx_n_u_exitExpr_neg, __pyx_k_exitExpr_neg, sizeof(__pyx_k_exitExpr_neg), 0, 1, 0, 1},
  {&__pyx_n_s_exitExpr_not, __pyx_k_exitExpr_not, sizeof(__pyx_k_exitExpr_not), 0, 0, 1, 1},
  {&__pyx_n_u_exitExpr_not, __pyx_k_exitExpr_not, sizeof(__pyx_k_exitExpr_not), 0, 1, 0, 1},
  {&__pyx_n_s_exitExpr_or, __pyx_k_exitExpr_or, sizeof(__pyx_k_exitExpr_or), 0, 0, 1, 1},
  {&__pyx_n_u_exitExpr_or, __pyx_k_exitExpr_or, sizeof(__pyx_k_exitExpr_or), 0, 1, 0, 1},
  {&__pyx_n_s_exitExpr_primary, __pyx_k_exitExpr_primary, sizeof(__pyx_k_exitExpr_primary), 0, 0, 1, 1},
  {&__pyx_n_u_exitExpr_primary, __pyx_k_exitExpr_primary, sizeof(__pyx_k_exitExpr_primary), 0, 1, 0, 1},
  {&__pyx_n_s_exitExpr_rel, __pyx_k_exitExpr_rel, sizeof(__pyx_k_exitExpr_rel), 0, 0, 1, 1},
  {&__pyx_n_u_exitExpr_rel, __pyx_k_exitExpr_rel, sizeof(__pyx_k_exitExpr_rel), 0, 1, 0, 1},
  {&__pyx_n_s_exitExpression_if, __pyx_k_exitExpression_if, sizeof(__pyx_k_exitExpression_if), 0, 0, 1, 1},
  {&__pyx_n_u_exitExpression_if, __pyx_k_exitExpression_if, sizeof(__pyx_k_exitExpression_if), 0, 1, 0, 1},
  {&__pyx_n_s_exitExpression_list, __pyx_k_exitExpression_list, sizeof(__pyx_k_exitExpression_list), 0, 0, 1, 1},
  {&__pyx_n_u_exitExpression_list, __pyx_k_exitExpression_list, sizeof(__pyx_k_exitExpression_list), 0, 1, 0, 1},
  {&__pyx_n_s_exitExpression_simple, __pyx_k_exitExpression_simple, sizeof(__pyx_k_exitExpression_simple), 0, 0, 1, 1},
  {&__pyx_n_u_exitExpression_simple, __pyx_k_exitExpression_simple, sizeof(__pyx_k_exitExpression_simple), 0, 1, 0, 1},
  {&__pyx_n_s_exitExtends_clause, __pyx_k_exitExtends_clause, sizeof(__pyx_k_exitExtends_clause), 0, 0, 1, 1},
  {&__pyx_n_u_exitExtends_clause, __pyx_k_exitExtends_clause, sizeof(__pyx_k_exitExtends_clause), 0, 1, 0, 1},
  {&__pyx_n_s_exitExternal_function_call, __pyx_k_exitExternal_function_call, sizeof(__pyx_k_exitExternal_function_call), 0, 0, 1, 1},
  {&__pyx_n_u_exitExternal_function_call, __pyx_k_exitExternal_function_call, sizeof(__pyx_k_exitExternal_function_call), 0, 1, 0, 1},
  {&__pyx_n_s_exitFor_equation, __pyx_k_exitFor_equation, sizeof(__pyx_k_exitFor_equation), 0, 0, 1, 1},
  {&__pyx_n_u_exitFor_equation, __pyx_k_exitFor_equation, sizeof(__pyx_k_exitFor_equation), 0, 1, 0, 1},
  {&__pyx_n_s_exitFor_index, __pyx_k_exitFor_index, sizeof(__pyx_k_exitFor_index), 0, 0, 1, 1},
  {&__pyx_n_u_exitFor_index, __pyx_k_exitFor_index, sizeof(__pyx_k_exitFor_index), 0, 1, 0, 1},
  {&__pyx_n_s_exitFor_indices, __pyx_k_exitFor_indices, sizeof(__pyx_k_exitFor_indices), 0, 0, 1, 1},
  {&__pyx_n_u_exitFor_indices, __pyx_k_exitFor_indices, sizeof(__pyx_k_exitFor_indices), 0, 1, 0, 1},
  {&__pyx_n_s_exitFor_statement, __pyx_k_exitFor_statement, sizeof(__pyx_k_exitFor_statement), 0, 0, 1, 1},
  {&__pyx_n_u_exitFor_statement, __pyx_k_exitFor_statement, sizeof(__pyx_k_exitFor_statement), 0, 1, 0, 1},
  {&__pyx_n_s_exitFunction_arguments, __pyx_k_exitFunction_arguments, sizeof(__pyx_k_exitFunction_arguments), 0, 0, 1, 1},
  {&__pyx_n_u_exitFunction_arguments, __pyx_k_exitFunction_arguments, sizeof(__pyx_k_exitFunction_arguments), 0, 1, 0, 1},
  {&__pyx_n_s_exitFunction_call_args, __pyx_k_exitFunction_call_args, sizeof(__pyx_k_exitFunction_call_args), 0, 0, 1, 1},
  {&__pyx_n_u_exitFunction_call_args, __pyx_k_exitFunction_call_args, sizeof(__pyx_k_exitFunction_call_args), 0, 1, 0, 1},
  {&__pyx_n_s_exitIf_equation, __pyx_k_exitIf_equation, sizeof(__pyx_k_exitIf_equation), 0, 0, 1, 1},
  {&__pyx_n_u_exitIf_equation, __pyx_k_exitIf_equation, sizeof(__pyx_k_exitIf_equation), 0, 1, 0, 1},
  {&__pyx_n_s_exitIf_statement, __pyx_k_exitIf_statement, sizeof(__pyx_k_exitIf_statement), 0, 0, 1, 1},
  {&__pyx_n_u_exitIf_statement, __pyx_k_exitIf_statement, sizeof(__pyx_k_exitIf_statement), 0, 1, 0, 1},
  {&__pyx_n_s_exitImport_clause, __pyx_k_exitImport_clause, sizeof(__pyx_k_exitImport_clause), 0, 0, 1, 1},
  {&__pyx_n_u_exitImport_clause, __pyx_k_exitImport_clause, sizeof(__pyx_k_exitImport_clause), 0, 1, 0, 1},
  {&__pyx_n_s_exitImport_list, __pyx_k_exitImport_list, sizeof(__pyx_k_exitImport_list), 0, 0, 1, 1},
  {&__pyx_n_u_exitImport_list, __pyx_k_exitImport_list, sizeof(__pyx_k_exitImport_list), 0, 1, 0, 1},
  {&__pyx_n_s_exitLanguage_specification, __pyx_k_exitLanguage_specification, sizeof(__pyx_k_exitLanguage_specification), 0, 0, 1, 1},
  {&__pyx_n_u_exitLanguage_specification, __pyx_k_exitLanguage_specification, sizeof(__pyx_k_exitLanguage_specification), 0, 1, 0, 1},
  {&__pyx_n_s_exitModification_assignment, __pyx_k_exitModification_assignment, sizeof(__pyx_k_exitModification_assignment), 0, 0, 1, 1},
  {&__pyx_n_u_exitModification_assignment, __pyx_k_exitModification_assignment, sizeof(__pyx_k_exitModification_assignment), 0, 1, 0, 1},
  {&__pyx_n_s_exitModification_assignment2, __pyx_k_exitModification_assignment2, sizeof(__pyx_k_exitModification_assignment2), 0, 0, 1, 1},
  {&__pyx_n_u_exitModification_assignment2, __pyx_k_exitModification_assignment2, sizeof(__pyx_k_exitModification_assignment2), 0, 1, 0, 1},
  {&__pyx_n_s_exitModification_class, __pyx_k_exitModification_class, sizeof(__pyx_k_exitModification_class), 0, 0, 1, 1},
  {&__pyx_n_u_exitModification_class, __pyx_k_exitModification_class, sizeof(__pyx_k_exitModification_class), 0, 1, 0, 1},
  {&__pyx_n_s_exitName, __pyx_k_exitName, sizeof(__pyx_k_exitName), 0, 0, 1, 1},
  {&__pyx_n_u_exitName, __pyx_k_exitName, sizeof(__pyx_k_exitName), 0, 1, 0, 1},
  {&__pyx_n_s_exitNamed_argument, __pyx_k_exitNamed_argument, sizeof(__pyx_k_exitNamed_argument), 0, 0, 1, 1},
  {&__pyx_n_u_exitNamed_argument, __pyx_k_exitNamed_argument, sizeof(__pyx_k_exitNamed_argument), 0, 1, 0, 1},
  {&__pyx_n_s_exitNamed_arguments, __pyx_k_exitNamed_arguments, sizeof(__pyx_k_exitNamed_arguments), 0, 0, 1, 1},
  {&__pyx_n_u_exitNamed_arguments, __pyx_k_exitNamed_arguments, sizeof(__pyx_k_exitNamed_arguments), 0, 1, 0, 1},
  {&__pyx_n_s_exitOutput_expression_list, __pyx_k_exitOutput_expression_list, sizeof(__pyx_k_exitOutput_expression_list), 0, 0, 1, 1},
  {&__pyx_n_u_exitOutput_expression_list, __pyx_k_exitOutput_expression_list, sizeof(__pyx_k_exitOutput_expression_list), 0, 1, 0, 1},
  {&__pyx_n_s_exitPrimary_component_reference, __pyx_k_exitPrimary_component_reference, sizeof(__pyx_k_exitPrimary_component_reference), 0, 0, 1, 1},
  {&__pyx_n_u_exitPrimary_component_reference, __pyx_k_exitPrimary_component_reference, sizeof(__pyx_k_exitPrimary_component_reference), 0, 1, 0, 1},
  {&__pyx_n_s_exitPrimary_derivative, __pyx_k_exitPrimary_derivative, sizeof(__pyx_k_exitPrimary_derivative), 0, 0, 1, 1},
  {&__pyx_n_u_exitPrimary_derivative, __pyx_k_exitPrimary_derivative, sizeof(__pyx_k_exitPrimary_derivative), 0, 1, 0, 1},
  {&__pyx_n_s_exitPrimary_end, __pyx_k_exitPrimary_end, sizeof(__pyx_k_exitPrimary_end), 0, 0, 1, 1},
  {&__pyx_n_u_exitPrimary_end, __pyx_k_exitPrimary_end, sizeof(__pyx_k_exitPrimary_end), 0, 1, 0, 1},
  {&__pyx_n_s_exitPrimary_expression_list, __pyx_k_exitPrimary_expression_list, sizeof(__pyx_k_exitPrimary_expression_list), 0, 0, 1, 1},
  {&__pyx_n_u_exitPrimary_expression_list, __pyx_k_exitPrimary_expression_list, sizeof(__pyx_k_exitPrimary_expression_list), 0, 1, 0, 1},
  {&__pyx_n_s_exitPrimary_false, __pyx_k_exitPrimary_false, sizeof(__pyx_k_exitPrimary_false), 0, 0, 1, 1},
  {&__pyx_n_u_exitPrimary_false, __pyx_k_exitPrimary_false, sizeof(__pyx_k_exitPrimary_false), 0, 1, 0, 1},
  {&__pyx_n_s_exitPrimary_function, __pyx_k_exitPrimary_function, sizeof(__pyx_k_exitPrimary_function), 0, 0, 1, 1},
  {&__pyx_n_u_exitPrimary_function, __pyx_k_exitPrimary_function, sizeof(__pyx_k_exitPrimary_function), 0, 1, 0, 1},
  {&__pyx_n_s_exitPrimary_function_arguments, __pyx_k_exitPrimary_function_arguments, sizeof(__pyx_k_exitPrimary_function_arguments), 0, 0, 1, 1},
  {&__pyx_n_u_exitPrimary_function_arguments, __pyx_k_exitPrimary_function_arguments, sizeof(__pyx_k_exitPrimary_function_arguments), 0, 1, 0, 1},
  {&__pyx_n_s_exitPrimary_initial, __pyx_k_exitPrimary_initial, sizeof(__pyx_k_exitPrimary_initial), 0, 0, 1, 1},
  {&__pyx_n_u_exitPrimary_initial, __pyx_k_exitPrimary_initial, sizeof(__pyx_k_exitPrimary_initial), 0, 1, 0, 1},
  {&__pyx_n_s_exitPrimary_output_expression_li, __pyx_k_exitPrimary_output_expression_li, sizeof(__pyx_k_exitPrimary_output_expression_li), 0, 0, 1, 1},
  {&__pyx_n_u_exitPrimary_output_expression_li, __pyx_k_exitPrimary_output_expression_li, sizeof(__pyx_k_exitPrimary_output_expression_li), 0, 1, 0, 1},
  {&__pyx_n_s_exitPrimary_string, __pyx_k_exitPrimary_string, sizeof(__pyx_k_exitPrimary_string), 0, 0, 1, 1},
  {&__pyx_n_u_exitPrimary_string, __pyx_k_exitPrimary_string, sizeof(__pyx_k_exitPrimary_string), 0, 1, 0, 1},
  {&__pyx_n_s_exitPrimary_true, __pyx_k_exitPrimary_true, sizeof(__pyx_k_exitPrimary_true), 0, 0, 1, 1},
  {&__pyx_n_u_exitPrimary_true, __pyx_k_exitPrimary_true, sizeof(__pyx_k_exitPrimary_true), 0, 1, 0, 1},
  {&__pyx_n_s_exitPrimary_unsigned_number, __pyx_k_exitPrimary_unsigned_number, sizeof(__pyx_k_exitPrimary_unsigned_number), 0, 0, 1, 1},
  {&__pyx_n_u_exitPrimary_unsigned_number, __pyx_k_exitPrimary_unsigned_number, sizeof(__pyx_k_exitPrimary_unsigned_number), 0, 1, 0, 1},
  {&__pyx_n_s_exitRegular_element, __pyx_k_exitRegular_element, sizeof(__pyx_k_exitRegular_element), 0, 0, 1, 1},
  {&__pyx_n_u_exitRegular_element, __pyx_k_exitRegular_element, sizeof(__pyx_k_exitRegular_element), 0, 1, 0, 1},
  {&__pyx_n_s_exitReplaceable_element, __pyx_k_exitReplaceable_element, sizeof(__pyx_k_exitReplaceable_element), 0, 0, 1, 1},
  {&__pyx_n_u_exitReplaceable_element, __pyx_k_exitReplaceable_element, sizeof(__pyx_k_exitReplaceable_element), 0, 1, 0, 1},
  {&__pyx_n_s_exitRule, __pyx_k_exitRule, sizeof(__pyx_k_exitRule), 0, 0, 1, 1},
  {&__pyx_n_s_exitShort_class_definition, __pyx_k_exitShort_class_definition, sizeof(__pyx_k_exitShort_class_definition), 0, 0, 1, 1},
  {&__pyx_n_u_exitShort_class_definition, __pyx_k_exitShort_class_definition, sizeof(__pyx_k_exitShort_class_definition), 0, 1, 0, 1},
  {&__pyx_n_s_exitSimple_expression, __pyx_k_exitSimple_expression, sizeof(__pyx_k_exitSimple_expression), 0, 0, 1, 1},
  {&__pyx_n_u_exitSimple_expression, __pyx_k_exitSimple_expression, sizeof(__pyx_k_exitSimple_expression), 0, 1, 0, 1},
  {&__pyx_n_s_exitStatement, __pyx_k_exitStatement, sizeof(__pyx_k_exitStatement), 0, 0, 1, 1},
  {&__pyx_n_u_exitStatement, __pyx_k_exitStatement, sizeof(__pyx_k_exitStatement), 0, 1, 0, 1},
  {&__pyx_n_s_exitStatement_break, __pyx_k_exitStatement_break, sizeof(__pyx_k_exitStatement_break), 0, 0, 1, 1},
  {&__pyx_n_u_exitStatement_break, __pyx_k_exitStatement_break, sizeof(__pyx_k_exitStatement_break), 0, 1, 0, 1},
  {&__pyx_n_s_exitStatement_component_function, __pyx_k_exitStatement_component_function, sizeof(__pyx_k_exitStatement_component_function), 0, 0, 1, 1},
  {&__pyx_n_u_exitStatement_component_function, __pyx_k_exitStatement_component_function, sizeof(__pyx_k_exitStatement_component_function), 0, 1, 0, 1},
  {&__pyx_n_s_exitStatement_component_referenc, __pyx_k_exitStatement_component_referenc, sizeof(__pyx_k_exitStatement_component_referenc), 0, 0, 1, 1},
  {&__pyx_n_u_exitStatement_component_referenc, __pyx_k_exitStatement_component_referenc, sizeof(__pyx_k_exitStatement_component_referenc), 0, 1, 0, 1},
  {&__pyx_n_s_exitStatement_for, __pyx_k_exitStatement_for, sizeof(__pyx_k_exitStatement_for), 0, 0, 1, 1},
  {&__pyx_n_u_exitStatement_for, __pyx_k_exitStatement_for, sizeof(__pyx_k_exitStatement_for), 0, 1, 0, 1},
  {&__pyx_n_s_exitStatement_if, __pyx_k_exitStatement_if, sizeof(__pyx_k_exitStatement_if), 0, 0, 1, 1},
  {&__pyx_n_u_exitStatement_if, __pyx_k_exitStatement_if, sizeof(__pyx_k_exitStatement_if), 0, 1, 0, 1},
  {&__pyx_n_s_exitStatement_return, __pyx_k_exitStatement_return, sizeof(__pyx_k_exitStatement_return), 0, 0, 1, 1},
  {&__pyx_n_u_exitStatement_return, __pyx_k_exitStatement_return, sizeof(__pyx_k_exitStatement_return), 0, 1, 0, 1},
  {&__pyx_n_s_exitStatement_when, __pyx_k_exitStatement_when, sizeof(__pyx_k_exitStatement_when), 0, 0, 1, 1},
  {&__pyx_n_u_exitStatement_when, __pyx_k_exitStatement_when, sizeof(__pyx_k_exitStatement_when), 0, 1, 0, 1},
  {&__pyx_n_s_exitStatement_while, __pyx_k_exitStatement_while, sizeof(__pyx_k_exitStatement_while), 0, 0, 1, 1},
  {&__pyx_n_u_exitStatement_while, __pyx_k_exitStatement_while, sizeof(__pyx_k_exitStatement_while), 0, 1, 0, 1},
  {&__pyx_n_s_exitStored_definition, __pyx_k_exitStored_definition, sizeof(__pyx_k_exitStored_definition), 0, 0, 1, 1},
  {&__pyx_n_u_exitStored_definition, __pyx_k_exitStored_definition, sizeof(__pyx_k_exitStored_definition), 0, 1, 0, 1},
  {&__pyx_n_s_exitStored_definition_class, __pyx_k_exitStored_definition_class, sizeof(__pyx_k_exitStored_definition_class), 0, 0, 1, 1},
  {&__pyx_n_u_exitStored_definition_class, __pyx_k_exitStored_definition_class, sizeof(__pyx_k_exitStored_definition_class), 0, 1, 0, 1},
  {&__pyx_n_s_exitString_comment, __pyx_k_exitString_comment, sizeof(__pyx_k_exitString_comment), 0, 0, 1, 1},
  {&__pyx_n_u_exitString_comment, __pyx_k_exitString_comment, sizeof(__pyx_k_exitString_comment), 0, 1, 0, 1},
  {&__pyx_n_s_exitSubscript, __pyx_k_exitSubscript, sizeof(__pyx_k_exitSubscript), 0, 0, 1, 1},
  {&__pyx_n_u_exitSubscript, __pyx_k_exitSubscript, sizeof(__pyx_k_exitSubscript), 0, 1, 0, 1},
  {&__pyx_n_s_exitType_prefix, __pyx_k_exitType_prefix, sizeof(__pyx_k_exitType_prefix), 0, 0, 1, 1},
  {&__pyx_n_u_exitType_prefix, __pyx_k_exitType_prefix, sizeof(__pyx_k_exitType_prefix), 0, 1, 0, 1},
  {&__pyx_n_s_exitType_specifier, __pyx_k_exitType_specifier, sizeof(__pyx_k_exitType_specifier), 0, 0, 1, 1},
  {&__pyx_n_u_exitType_specifier, __pyx_k_exitType_specifier, sizeof(__pyx_k_exitType_specifier), 0, 1, 0, 1},
  {&__pyx_n_s_exitWhen_equation, __pyx_k_exitWhen_equation, sizeof(__pyx_k_exitWhen_equation), 0, 0, 1, 1},
  {&__pyx_n_u_exitWhen_equation, __pyx_k_exitWhen_equation, sizeof(__pyx_k_exitWhen_equation), 0, 1, 0, 1},
  {&__pyx_n_s_exitWhen_statement, __pyx_k_exitWhen_statement, sizeof(__pyx_k_exitWhen_statement), 0, 0, 1, 1},
  {&__pyx_n_u_exitWhen_statement, __pyx_k_exitWhen_statement, sizeof(__pyx_k_exitWhen_statement), 0, 1, 0, 1},
  {&__pyx_n_s_exitWhile_statement, __pyx_k_exitWhile_statement, sizeof(__pyx_k_exitWhile_statement), 0, 0, 1, 1},
  {&__pyx_n_u_exitWhile_statement, __pyx_k_exitWhile_statement, sizeof(__pyx_k_exitWhile_statement), 0, 1, 0, 1},
  {&__pyx_kp_u_expandable, __pyx_k_expandable, sizeof(__pyx_k_expandable), 0, 1, 0, 0},
  {&__pyx_n_s_expr, __pyx_k_expr, sizeof(__pyx_k_expr), 0, 0, 1, 1},
  {&__pyx_n_u_expr, __pyx_k_expr, sizeof(__pyx_k_expr), 0, 1, 0, 1},
  {&__pyx_n_s_expr_sempred, __pyx_k_expr_sempred, sizeof(__pyx_k_expr_sempred), 0, 0, 1, 1},
  {&__pyx_n_s_expression, __pyx_k_expression, sizeof(__pyx_k_expression), 0, 0, 1, 1},
  {&__pyx_n_u_expression, __pyx_k_expression, sizeof(__pyx_k_expression), 0, 1, 0, 1},
  {&__pyx_n_s_expression_list, __pyx_k_expression_list, sizeof(__pyx_k_expression_list), 0, 0, 1, 1},
  {&__pyx_n_u_expression_list, __pyx_k_expression_list, sizeof(__pyx_k_expression_list), 0, 1, 0, 1},
  {&__pyx_n_s_ext_annotation, __pyx_k_ext_annotation, sizeof(__pyx_k_ext_annotation), 0, 0, 1, 1},
  {&__pyx_kp_u_extends, __pyx_k_extends, sizeof(__pyx_k_extends), 0, 1, 0, 0},
  {&__pyx_n_s_extends_clause, __pyx_k_extends_clause, sizeof(__pyx_k_extends_clause), 0, 0, 1, 1},
  {&__pyx_n_u_extends_clause, __pyx_k_extends_clause, sizeof(__pyx_k_extends_clause), 0, 1, 0, 1},
  {&__pyx_kp_u_external, __pyx_k_external, sizeof(__pyx_k_external), 0, 1, 0, 0},
  {&__pyx_n_s_external_function_call, __pyx_k_external_function_call, sizeof(__pyx_k_external_function_call), 0, 0, 1, 1},
  {&__pyx_n_u_external_function_call, __pyx_k_external_function_call, sizeof(__pyx_k_external_function_call), 0, 1, 0, 1},
  {&__pyx_kp_u_f4_j6, __pyx_k_f4_j6, sizeof(__pyx_k_f4_j6), 0, 1, 0, 0},
  {&__pyx_kp_u_false, __pyx_k_false, sizeof(__pyx_k_false), 0, 1, 0, 0},
  {&__pyx_kp_u_final, __pyx_k_final, sizeof(__pyx_k_final), 0, 1, 0, 0},
  {&__pyx_kp_u_flow, __pyx_k_flow, sizeof(__pyx_k_flow), 0, 1, 0, 0},
  {&__pyx_kp_u_for, __pyx_k_for, sizeof(__pyx_k_for), 0, 1, 0, 0},
  {&__pyx_n_s_for_equation, __pyx_k_for_equation, sizeof(__pyx_k_for_equation), 0, 0, 1, 1},
  {&__pyx_n_u_for_equation, __pyx_k_for_equation, sizeof(__pyx_k_for_equation), 0, 1, 0, 1},
  {&__pyx_n_s_for_index, __pyx_k_for_index, sizeof(__pyx_k_for_index), 0, 0, 1, 1},
  {&__pyx_n_u_for_index, __pyx_k_for_index, sizeof(__pyx_k_for_index), 0, 1, 0, 1},
  {&__pyx_n_s_for_indices, __pyx_k_for_indices, sizeof(__pyx_k_for_indices), 0, 0, 1, 1},
  {&__pyx_n_u_for_indices, __pyx_k_for_indices, sizeof(__pyx_k_for_indices), 0, 1, 0, 1},
  {&__pyx_n_s_for_statement, __pyx_k_for_statement, sizeof(__pyx_k_for_statement), 0, 0, 1, 1},
  {&__pyx_n_u_for_statement, __pyx_k_for_statement, sizeof(__pyx_k_for_statement), 0, 1, 0, 1},
  {&__pyx_kp_u_function, __pyx_k_function, sizeof(__pyx_k_function), 0, 1, 0, 0},
  {&__pyx_n_s_function_argument, __pyx_k_function_argument, sizeof(__pyx_k_function_argument), 0, 0, 1, 1},
  {&__pyx_n_u_function_argument, __pyx_k_function_argument, sizeof(__pyx_k_function_argument), 0, 1, 0, 1},
  {&__pyx_n_s_function_arguments, __pyx_k_function_arguments, sizeof(__pyx_k_function_arguments), 0, 0, 1, 1},
  {&__pyx_n_u_function_arguments, __pyx_k_function_arguments, sizeof(__pyx_k_function_arguments), 0, 1, 0, 1},
  {&__pyx_n_s_function_call_args, __pyx_k_function_call_args, sizeof(__pyx_k_function_call_args), 0, 0, 1, 1},
  {&__pyx_n_u_function_call_args, __pyx_k_function_call_args, sizeof(__pyx_k_function_call_args), 0, 1, 0, 1},
  {&__pyx_kp_u_g_7_n8, __pyx_k_g_7_n8, sizeof(__pyx_k_g_7_n8), 0, 1, 0, 0},
  {&__pyx_n_s_get, __pyx_k_get, sizeof(__pyx_k_get), 0, 0, 1, 1},
  {&__pyx_n_s_getRuleIndex, __pyx_k_getRuleIndex, sizeof(__pyx_k_getRuleIndex), 0, 0, 1, 1},
  {&__pyx_n_s_getToken, __pyx_k_getToken, sizeof(__pyx_k_getToken), 0, 0, 1, 1},
  {&__pyx_n_s_getTokens, __pyx_k_getTokens, sizeof(__pyx_k_getTokens), 0, 0, 1, 1},
  {&__pyx_n_s_getTypedRuleContext, __pyx_k_getTypedRuleContext, sizeof(__pyx_k_getTypedRuleContext), 0, 0, 1, 1},
  {&__pyx_n_s_getTypedRuleContexts, __pyx_k_getTypedRuleContexts, sizeof(__pyx_k_getTypedRuleContexts), 0, 0, 1, 1},
  {&__pyx_n_s_getvalue, __pyx_k_getvalue, sizeof(__pyx_k_getvalue), 0, 0, 1, 1},
  {&__pyx_n_s_grammarFileName, __pyx_k_grammarFileName, sizeof(__pyx_k_grammarFileName), 0, 0, 1, 1},
  {&__pyx_kp_u_h_j_l_n, __pyx_k_h_j_l_n, sizeof(__pyx_k_h_j_l_n), 0, 1, 0, 0},
  {&__pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_k_home_jgoppert_git_pymola_pymola, sizeof(__pyx_k_home_jgoppert_git_pymola_pymola), 0, 0, 1, 0},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_kp_u_if, __pyx_k_if, sizeof(__pyx_k_if), 0, 1, 0, 0},
  {&__pyx_n_s_if_equation, __pyx_k_if_equation, sizeof(__pyx_k_if_equation), 0, 0, 1, 1},
  {&__pyx_n_u_if_equation, __pyx_k_if_equation, sizeof(__pyx_k_if_equation), 0, 1, 0, 1},
  {&__pyx_n_s_if_statement, __pyx_k_if_statement, sizeof(__pyx_k_if_statement), 0, 0, 1, 1},
  {&__pyx_n_u_if_statement, __pyx_k_if_statement, sizeof(__pyx_k_if_statement), 0, 1, 0, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_kp_u_import_2, __pyx_k_import_2, sizeof(__pyx_k_import_2), 0, 1, 0, 0},
  {&__pyx_n_s_import_clause, __pyx_k_import_clause, sizeof(__pyx_k_import_clause), 0, 0, 1, 1},
  {&__pyx_n_u_import_clause, __pyx_k_import_clause, sizeof(__pyx_k_import_clause), 0, 1, 0, 1},
  {&__pyx_n_s_import_list, __pyx_k_import_list, sizeof(__pyx_k_import_list), 0, 0, 1, 1},
  {&__pyx_n_u_import_list, __pyx_k_import_list, sizeof(__pyx_k_import_list), 0, 1, 0, 1},
  {&__pyx_kp_u_impure, __pyx_k_impure, sizeof(__pyx_k_impure), 0, 1, 0, 0},
  {&__pyx_kp_u_in, __pyx_k_in, sizeof(__pyx_k_in), 0, 1, 0, 0},
  {&__pyx_n_s_init, __pyx_k_init, sizeof(__pyx_k_init), 0, 0, 1, 1},
  {&__pyx_kp_u_initial, __pyx_k_initial, sizeof(__pyx_k_initial), 0, 1, 0, 0},
  {&__pyx_kp_u_inner, __pyx_k_inner, sizeof(__pyx_k_inner), 0, 1, 0, 0},
  {&__pyx_n_s_input, __pyx_k_input, sizeof(__pyx_k_input), 0, 0, 1, 1},
  {&__pyx_n_s_input_2, __pyx_k_input_2, sizeof(__pyx_k_input_2), 0, 0, 1, 1},
  {&__pyx_kp_u_input_3, __pyx_k_input_3, sizeof(__pyx_k_input_3), 0, 1, 0, 0},
  {&__pyx_n_s_interp, __pyx_k_interp, sizeof(__pyx_k_interp), 0, 0, 1, 1},
  {&__pyx_n_s_invokingState, __pyx_k_invokingState, sizeof(__pyx_k_invokingState), 0, 0, 1, 1},
  {&__pyx_n_s_io, __pyx_k_io, sizeof(__pyx_k_io), 0, 0, 1, 1},
  {&__pyx_kp_u_l7_h5_v, __pyx_k_l7_h5_v, sizeof(__pyx_k_l7_h5_v), 0, 1, 0, 0},
  {&__pyx_n_s_la, __pyx_k_la, sizeof(__pyx_k_la), 0, 0, 1, 1},
  {&__pyx_n_s_la_2, __pyx_k_la_2, sizeof(__pyx_k_la_2), 0, 0, 1, 1},
  {&__pyx_n_s_language_specification, __pyx_k_language_specification, sizeof(__pyx_k_language_specification), 0, 0, 1, 1},
  {&__pyx_n_u_language_specification, __pyx_k_language_specification, sizeof(__pyx_k_language_specification), 0, 1, 0, 1},
  {&__pyx_n_s_listener, __pyx_k_listener, sizeof(__pyx_k_listener), 0, 0, 1, 1},
  {&__pyx_n_s_literalNames, __pyx_k_literalNames, sizeof(__pyx_k_literalNames), 0, 0, 1, 1},
  {&__pyx_n_s_localctx, __pyx_k_localctx, sizeof(__pyx_k_localctx), 0, 0, 1, 1},
  {&__pyx_kp_u_loop, __pyx_k_loop, sizeof(__pyx_k_loop), 0, 1, 0, 0},
  {&__pyx_kp_u_m_p9, __pyx_k_m_p9, sizeof(__pyx_k_m_p9), 0, 1, 0, 0},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_match, __pyx_k_match, sizeof(__pyx_k_match), 0, 0, 1, 1},
  {&__pyx_n_s_metaclass, __pyx_k_metaclass, sizeof(__pyx_k_metaclass), 0, 0, 1, 1},
  {&__pyx_kp_u_model, __pyx_k_model, sizeof(__pyx_k_model), 0, 1, 0, 0},
  {&__pyx_n_s_modification, __pyx_k_modification, sizeof(__pyx_k_modification), 0, 0, 1, 1},
  {&__pyx_n_u_modification, __pyx_k_modification, sizeof(__pyx_k_modification), 0, 1, 0, 1},
  {&__pyx_n_s_module, __pyx_k_module, sizeof(__pyx_k_module), 0, 0, 1, 1},
  {&__pyx_kp_u_n8, __pyx_k_n8, sizeof(__pyx_k_n8), 0, 1, 0, 0},
  {&__pyx_kp_u_n8_0, __pyx_k_n8_0, sizeof(__pyx_k_n8_0), 0, 1, 0, 0},
  {&__pyx_kp_u_n8_0_T, __pyx_k_n8_0_T, sizeof(__pyx_k_n8_0_T), 0, 1, 0, 0},
  {&__pyx_kp_u_n8_0_X, __pyx_k_n8_0_X, sizeof(__pyx_k_n8_0_X), 0, 1, 0, 0},
  {&__pyx_kp_u_n8_0_X_2, __pyx_k_n8_0_X_2, sizeof(__pyx_k_n8_0_X_2), 0, 1, 0, 0},
  {&__pyx_kp_u_n8_0_n8, __pyx_k_n8_0_n8, sizeof(__pyx_k_n8_0_n8), 0, 1, 0, 0},
  {&__pyx_kp_u_n8_10, __pyx_k_n8_10, sizeof(__pyx_k_n8_10), 0, 1, 0, 0},
  {&__pyx_kp_u_n8_11, __pyx_k_n8_11, sizeof(__pyx_k_n8_11), 0, 1, 0, 0},
  {&__pyx_kp_u_n8_12, __pyx_k_n8_12, sizeof(__pyx_k_n8_12), 0, 1, 0, 0},
  {&__pyx_kp_u_n8_2, __pyx_k_n8_2, sizeof(__pyx_k_n8_2), 0, 1, 0, 0},
  {&__pyx_kp_u_n8_3, __pyx_k_n8_3, sizeof(__pyx_k_n8_3), 0, 1, 0, 0},
  {&__pyx_kp_u_n8_4, __pyx_k_n8_4, sizeof(__pyx_k_n8_4), 0, 1, 0, 0},
  {&__pyx_kp_u_n8_5, __pyx_k_n8_5, sizeof(__pyx_k_n8_5), 0, 1, 0, 0},
  {&__pyx_kp_u_n8_6, __pyx_k_n8_6, sizeof(__pyx_k_n8_6), 0, 1, 0, 0},
  {&__pyx_kp_u_n8_7, __pyx_k_n8_7, sizeof(__pyx_k_n8_7), 0, 1, 0, 0},
  {&__pyx_kp_u_n8_8, __pyx_k_n8_8, sizeof(__pyx_k_n8_8), 0, 1, 0, 0},
  {&__pyx_kp_u_n8_9, __pyx_k_n8_9, sizeof(__pyx_k_n8_9), 0, 1, 0, 0},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_u_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 1, 0, 1},
  {&__pyx_n_s_named_argument, __pyx_k_named_argument, sizeof(__pyx_k_named_argument), 0, 0, 1, 1},
  {&__pyx_n_u_named_argument, __pyx_k_named_argument, sizeof(__pyx_k_named_argument), 0, 1, 0, 1},
  {&__pyx_n_s_named_arguments, __pyx_k_named_arguments, sizeof(__pyx_k_named_arguments), 0, 0, 1, 1},
  {&__pyx_n_u_named_arguments, __pyx_k_named_arguments, sizeof(__pyx_k_named_arguments), 0, 1, 0, 1},
  {&__pyx_kp_u_not, __pyx_k_not, sizeof(__pyx_k_not), 0, 1, 0, 0},
  {&__pyx_kp_u_o, __pyx_k_o, sizeof(__pyx_k_o), 0, 1, 0, 0},
  {&__pyx_n_s_op, __pyx_k_op, sizeof(__pyx_k_op), 0, 0, 1, 1},
  {&__pyx_kp_u_operator, __pyx_k_operator, sizeof(__pyx_k_operator), 0, 1, 0, 0},
  {&__pyx_kp_u_or, __pyx_k_or, sizeof(__pyx_k_or), 0, 1, 0, 0},
  {&__pyx_kp_u_outer, __pyx_k_outer, sizeof(__pyx_k_outer), 0, 1, 0, 0},
  {&__pyx_n_s_output, __pyx_k_output, sizeof(__pyx_k_output), 0, 0, 1, 1},
  {&__pyx_kp_u_output_2, __pyx_k_output_2, sizeof(__pyx_k_output_2), 0, 1, 0, 0},
  {&__pyx_n_s_output_expression_list, __pyx_k_output_expression_list, sizeof(__pyx_k_output_expression_list), 0, 0, 1, 1},
  {&__pyx_n_u_output_expression_list, __pyx_k_output_expression_list, sizeof(__pyx_k_output_expression_list), 0, 1, 0, 1},
  {&__pyx_n_s_p, __pyx_k_p, sizeof(__pyx_k_p), 0, 0, 1, 1},
  {&__pyx_kp_u_p9_n8, __pyx_k_p9_n8, sizeof(__pyx_k_p9_n8), 0, 1, 0, 0},
  {&__pyx_kp_u_p_r_t_v, __pyx_k_p_r_t_v, sizeof(__pyx_k_p_r_t_v), 0, 1, 0, 0},
  {&__pyx_kp_u_package, __pyx_k_package, sizeof(__pyx_k_package), 0, 1, 0, 0},
  {&__pyx_kp_u_parameter, __pyx_k_parameter, sizeof(__pyx_k_parameter), 0, 1, 0, 0},
  {&__pyx_n_s_parent, __pyx_k_parent, sizeof(__pyx_k_parent), 0, 0, 1, 1},
  {&__pyx_n_s_parentState, __pyx_k_parentState, sizeof(__pyx_k_parentState), 0, 0, 1, 1},
  {&__pyx_n_s_parentctx, __pyx_k_parentctx, sizeof(__pyx_k_parentctx), 0, 0, 1, 1},
  {&__pyx_n_s_parseListeners, __pyx_k_parseListeners, sizeof(__pyx_k_parseListeners), 0, 0, 1, 1},
  {&__pyx_n_s_parser, __pyx_k_parser, sizeof(__pyx_k_parser), 0, 0, 1, 1},
  {&__pyx_kp_u_partial, __pyx_k_partial, sizeof(__pyx_k_partial), 0, 1, 0, 0},
  {&__pyx_n_s_precpred, __pyx_k_precpred, sizeof(__pyx_k_precpred), 0, 0, 1, 1},
  {&__pyx_n_s_pred, __pyx_k_pred, sizeof(__pyx_k_pred), 0, 0, 1, 1},
  {&__pyx_n_s_predIndex, __pyx_k_predIndex, sizeof(__pyx_k_predIndex), 0, 0, 1, 1},
  {&__pyx_n_s_predicates, __pyx_k_predicates, sizeof(__pyx_k_predicates), 0, 0, 1, 1},
  {&__pyx_n_s_prepare, __pyx_k_prepare, sizeof(__pyx_k_prepare), 0, 0, 1, 1},
  {&__pyx_n_s_prevctx, __pyx_k_prevctx, sizeof(__pyx_k_prevctx), 0, 0, 1, 1},
  {&__pyx_n_s_primary, __pyx_k_primary, sizeof(__pyx_k_primary), 0, 0, 1, 1},
  {&__pyx_n_u_primary, __pyx_k_primary, sizeof(__pyx_k_primary), 0, 1, 0, 1},
  {&__pyx_kp_u_protected, __pyx_k_protected, sizeof(__pyx_k_protected), 0, 1, 0, 0},
  {&__pyx_kp_u_public, __pyx_k_public, sizeof(__pyx_k_public), 0, 1, 0, 0},
  {&__pyx_kp_u_pure, __pyx_k_pure, sizeof(__pyx_k_pure), 0, 1, 0, 0},
  {&__pyx_n_s_pushNewRecursionContext, __pyx_k_pushNewRecursionContext, sizeof(__pyx_k_pushNewRecursionContext), 0, 0, 1, 1},
  {&__pyx_n_s_pymola_generated_ModelicaParser, __pyx_k_pymola_generated_ModelicaParser, sizeof(__pyx_k_pymola_generated_ModelicaParser), 0, 0, 1, 1},
  {&__pyx_kp_u_q, __pyx_k_q, sizeof(__pyx_k_q), 0, 1, 0, 0},
  {&__pyx_n_s_qualname, __pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 0, 1, 1},
  {&__pyx_kp_u_r, __pyx_k_r, sizeof(__pyx_k_r), 0, 1, 0, 0},
  {&__pyx_kp_u_r_2, __pyx_k_r_2, sizeof(__pyx_k_r_2), 0, 1, 0, 0},
  {&__pyx_kp_u_r_3, __pyx_k_r_3, sizeof(__pyx_k_r_3), 0, 1, 0, 0},
  {&__pyx_kp_u_r_4, __pyx_k_r_4, sizeof(__pyx_k_r_4), 0, 1, 0, 0},
  {&__pyx_kp_u_r_5, __pyx_k_r_5, sizeof(__pyx_k_r_5), 0, 1, 0, 0},
  {&__pyx_kp_u_r_J, __pyx_k_r_J, sizeof(__pyx_k_r_J), 0, 1, 0, 0},
  {&__pyx_kp_u_r_K, __pyx_k_r_K, sizeof(__pyx_k_r_K), 0, 1, 0, 0},
  {&__pyx_kp_u_r_r, __pyx_k_r_r, sizeof(__pyx_k_r_r), 0, 1, 0, 0},
  {&__pyx_n_s_re, __pyx_k_re, sizeof(__pyx_k_re), 0, 0, 1, 1},
  {&__pyx_kp_u_record, __pyx_k_record, sizeof(__pyx_k_record), 0, 1, 0, 0},
  {&__pyx_n_s_recover, __pyx_k_recover, sizeof(__pyx_k_recover), 0, 0, 1, 1},
  {&__pyx_n_s_recoverInline, __pyx_k_recoverInline, sizeof(__pyx_k_recoverInline), 0, 0, 1, 1},
  {&__pyx_kp_u_redeclare, __pyx_k_redeclare, sizeof(__pyx_k_redeclare), 0, 1, 0, 0},
  {&__pyx_n_s_regular_element, __pyx_k_regular_element, sizeof(__pyx_k_regular_element), 0, 0, 1, 1},
  {&__pyx_n_u_regular_element, __pyx_k_regular_element, sizeof(__pyx_k_regular_element), 0, 1, 0, 1},
  {&__pyx_kp_u_replaceable, __pyx_k_replaceable, sizeof(__pyx_k_replaceable), 0, 1, 0, 0},
  {&__pyx_n_s_replaceable_element, __pyx_k_replaceable_element, sizeof(__pyx_k_replaceable_element), 0, 0, 1, 1},
  {&__pyx_n_u_replaceable_element, __pyx_k_replaceable_element, sizeof(__pyx_k_replaceable_element), 0, 1, 0, 1},
  {&__pyx_n_s_reportError, __pyx_k_reportError, sizeof(__pyx_k_reportError), 0, 0, 1, 1},
  {&__pyx_n_s_reportMatch, __pyx_k_reportMatch, sizeof(__pyx_k_reportMatch), 0, 0, 1, 1},
  {&__pyx_kp_u_return, __pyx_k_return, sizeof(__pyx_k_return), 0, 1, 0, 0},
  {&__pyx_n_s_ruleIndex, __pyx_k_ruleIndex, sizeof(__pyx_k_ruleIndex), 0, 0, 1, 1},
  {&__pyx_n_s_ruleNames, __pyx_k_ruleNames, sizeof(__pyx_k_ruleNames), 0, 0, 1, 1},
  {&__pyx_kp_u_s, __pyx_k_s, sizeof(__pyx_k_s), 0, 1, 0, 0},
  {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {&__pyx_kp_u_self_precpred_self__ctx_2, __pyx_k_self_precpred_self__ctx_2, sizeof(__pyx_k_self_precpred_self__ctx_2), 0, 1, 0, 0},
  {&__pyx_kp_u_self_precpred_self__ctx_3, __pyx_k_self_precpred_self__ctx_3, sizeof(__pyx_k_self_precpred_self__ctx_3), 0, 1, 0, 0},
  {&__pyx_kp_u_self_precpred_self__ctx_5, __pyx_k_self_precpred_self__ctx_5, sizeof(__pyx_k_self_precpred_self__ctx_5), 0, 1, 0, 0},
  {&__pyx_kp_u_self_precpred_self__ctx_6, __pyx_k_self_precpred_self__ctx_6, sizeof(__pyx_k_self_precpred_self__ctx_6), 0, 1, 0, 0},
  {&__pyx_kp_u_self_precpred_self__ctx_7, __pyx_k_self_precpred_self__ctx_7, sizeof(__pyx_k_self_precpred_self__ctx_7), 0, 1, 0, 0},
  {&__pyx_n_s_sempred, __pyx_k_sempred, sizeof(__pyx_k_sempred), 0, 0, 1, 1},
  {&__pyx_n_s_serializedATN, __pyx_k_serializedATN, sizeof(__pyx_k_serializedATN), 0, 0, 1, 1},
  {&__pyx_n_s_sharedContextCache, __pyx_k_sharedContextCache, sizeof(__pyx_k_sharedContextCache), 0, 0, 1, 1},
  {&__pyx_n_s_short_class_definition, __pyx_k_short_class_definition, sizeof(__pyx_k_short_class_definition), 0, 0, 1, 1},
  {&__pyx_n_u_short_class_definition, __pyx_k_short_class_definition, sizeof(__pyx_k_short_class_definition), 0, 1, 0, 1},
  {&__pyx_n_s_simple_expression, __pyx_k_simple_expression, sizeof(__pyx_k_simple_expression), 0, 0, 1, 1},
  {&__pyx_n_u_simple_expression, __pyx_k_simple_expression, sizeof(__pyx_k_simple_expression), 0, 1, 0, 1},
  {&__pyx_n_s_startState, __pyx_k_startState, sizeof(__pyx_k_startState), 0, 0, 1, 1},
  {&__pyx_n_s_state, __pyx_k_state, sizeof(__pyx_k_state), 0, 0, 1, 1},
  {&__pyx_n_s_statement, __pyx_k_statement, sizeof(__pyx_k_statement), 0, 0, 1, 1},
  {&__pyx_n_u_statement, __pyx_k_statement, sizeof(__pyx_k_statement), 0, 1, 0, 1},
  {&__pyx_n_s_statement_options, __pyx_k_statement_options, sizeof(__pyx_k_statement_options), 0, 0, 1, 1},
  {&__pyx_n_u_statement_options, __pyx_k_statement_options, sizeof(__pyx_k_statement_options), 0, 1, 0, 1},
  {&__pyx_n_s_stdout, __pyx_k_stdout, sizeof(__pyx_k_stdout), 0, 0, 1, 1},
  {&__pyx_n_s_stop, __pyx_k_stop, sizeof(__pyx_k_stop), 0, 0, 1, 1},
  {&__pyx_n_s_stored_definition, __pyx_k_stored_definition, sizeof(__pyx_k_stored_definition), 0, 0, 1, 1},
  {&__pyx_n_u_stored_definition, __pyx_k_stored_definition, sizeof(__pyx_k_stored_definition), 0, 1, 0, 1},
  {&__pyx_n_s_stored_definition_class, __pyx_k_stored_definition_class, sizeof(__pyx_k_stored_definition_class), 0, 0, 1, 1},
  {&__pyx_n_u_stored_definition_class, __pyx_k_stored_definition_class, sizeof(__pyx_k_stored_definition_class), 0, 1, 0, 1},
  {&__pyx_kp_u_stream, __pyx_k_stream, sizeof(__pyx_k_stream), 0, 1, 0, 0},
  {&__pyx_n_s_string_comment, __pyx_k_string_comment, sizeof(__pyx_k_string_comment), 0, 0, 1, 1},
  {&__pyx_n_u_string_comment, __pyx_k_string_comment, sizeof(__pyx_k_string_comment), 0, 1, 0, 1},
  {&__pyx_n_s_subscript, __pyx_k_subscript, sizeof(__pyx_k_subscript), 0, 0, 1, 1},
  {&__pyx_n_u_subscript, __pyx_k_subscript, sizeof(__pyx_k_subscript), 0, 1, 0, 1},
  {&__pyx_n_s_super, __pyx_k_super, sizeof(__pyx_k_super), 0, 0, 1, 1},
  {&__pyx_n_s_symbolicNames, __pyx_k_symbolicNames, sizeof(__pyx_k_symbolicNames), 0, 0, 1, 1},
  {&__pyx_n_s_sync, __pyx_k_sync, sizeof(__pyx_k_sync), 0, 0, 1, 1},
  {&__pyx_n_s_sys, __pyx_k_sys, sizeof(__pyx_k_sys), 0, 0, 1, 1},
  {&__pyx_kp_u_t, __pyx_k_t, sizeof(__pyx_k_t), 0, 1, 0, 0},
  {&__pyx_kp_u_t_t, __pyx_k_t_t, sizeof(__pyx_k_t_t), 0, 1, 0, 0},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_kp_u_then, __pyx_k_then, sizeof(__pyx_k_then), 0, 1, 0, 0},
  {&__pyx_n_s_token, __pyx_k_token, sizeof(__pyx_k_token), 0, 0, 1, 1},
  {&__pyx_n_s_triggerExitRuleEvent, __pyx_k_triggerExitRuleEvent, sizeof(__pyx_k_triggerExitRuleEvent), 0, 0, 1, 1},
  {&__pyx_kp_u_true, __pyx_k_true, sizeof(__pyx_k_true), 0, 1, 0, 0},
  {&__pyx_kp_u_type, __pyx_k_type, sizeof(__pyx_k_type), 0, 1, 0, 0},
  {&__pyx_n_s_type_prefix, __pyx_k_type_prefix, sizeof(__pyx_k_type_prefix), 0, 0, 1, 1},
  {&__pyx_n_u_type_prefix, __pyx_k_type_prefix, sizeof(__pyx_k_type_prefix), 0, 1, 0, 1},
  {&__pyx_n_s_type_specifier, __pyx_k_type_specifier, sizeof(__pyx_k_type_specifier), 0, 0, 1, 1},
  {&__pyx_n_u_type_specifier, __pyx_k_type_specifier, sizeof(__pyx_k_type_specifier), 0, 1, 0, 1},
  {&__pyx_n_s_typing_io, __pyx_k_typing_io, sizeof(__pyx_k_typing_io), 0, 0, 1, 1},
  {&__pyx_kp_u_u, __pyx_k_u, sizeof(__pyx_k_u), 0, 1, 0, 0},
  {&__pyx_n_s_unrollRecursionContexts, __pyx_k_unrollRecursionContexts, sizeof(__pyx_k_unrollRecursionContexts), 0, 0, 1, 1},
  {&__pyx_kp_u_v, __pyx_k_v, sizeof(__pyx_k_v), 0, 1, 0, 0},
  {&__pyx_kp_u_v_2, __pyx_k_v_2, sizeof(__pyx_k_v_2), 0, 1, 0, 0},
  {&__pyx_kp_u_v_3, __pyx_k_v_3, sizeof(__pyx_k_v_3), 0, 1, 0, 0},
  {&__pyx_kp_u_v_F, __pyx_k_v_F, sizeof(__pyx_k_v_F), 0, 1, 0, 0},
  {&__pyx_n_s_visitAlgorithm_section, __pyx_k_visitAlgorithm_section, sizeof(__pyx_k_visitAlgorithm_section), 0, 0, 1, 1},
  {&__pyx_n_u_visitAlgorithm_section, __pyx_k_visitAlgorithm_section, sizeof(__pyx_k_visitAlgorithm_section), 0, 1, 0, 1},
  {&__pyx_n_s_visitAnnotation, __pyx_k_visitAnnotation, sizeof(__pyx_k_visitAnnotation), 0, 0, 1, 1},
  {&__pyx_n_u_visitAnnotation, __pyx_k_visitAnnotation, sizeof(__pyx_k_visitAnnotation), 0, 1, 0, 1},
  {&__pyx_n_s_visitArgument, __pyx_k_visitArgument, sizeof(__pyx_k_visitArgument), 0, 0, 1, 1},
  {&__pyx_n_u_visitArgument, __pyx_k_visitArgument, sizeof(__pyx_k_visitArgument), 0, 1, 0, 1},
  {&__pyx_n_s_visitArgument_expression, __pyx_k_visitArgument_expression, sizeof(__pyx_k_visitArgument_expression), 0, 0, 1, 1},
  {&__pyx_n_u_visitArgument_expression, __pyx_k_visitArgument_expression, sizeof(__pyx_k_visitArgument_expression), 0, 1, 0, 1},
  {&__pyx_n_s_visitArgument_function, __pyx_k_visitArgument_function, sizeof(__pyx_k_visitArgument_function), 0, 0, 1, 1},
  {&__pyx_n_u_visitArgument_function, __pyx_k_visitArgument_function, sizeof(__pyx_k_visitArgument_function), 0, 1, 0, 1},
  {&__pyx_n_s_visitArgument_list, __pyx_k_visitArgument_list, sizeof(__pyx_k_visitArgument_list), 0, 0, 1, 1},
  {&__pyx_n_u_visitArgument_list, __pyx_k_visitArgument_list, sizeof(__pyx_k_visitArgument_list), 0, 1, 0, 1},
  {&__pyx_n_s_visitArray_subscripts, __pyx_k_visitArray_subscripts, sizeof(__pyx_k_visitArray_subscripts), 0, 0, 1, 1},
  {&__pyx_n_u_visitArray_subscripts, __pyx_k_visitArray_subscripts, sizeof(__pyx_k_visitArray_subscripts), 0, 1, 0, 1},
  {&__pyx_n_s_visitBase_prefix, __pyx_k_visitBase_prefix, sizeof(__pyx_k_visitBase_prefix), 0, 0, 1, 1},
  {&__pyx_n_u_visitBase_prefix, __pyx_k_visitBase_prefix, sizeof(__pyx_k_visitBase_prefix), 0, 1, 0, 1},
  {&__pyx_n_s_visitChildren, __pyx_k_visitChildren, sizeof(__pyx_k_visitChildren), 0, 0, 1, 1},
  {&__pyx_n_s_visitClass_definition, __pyx_k_visitClass_definition, sizeof(__pyx_k_visitClass_definition), 0, 0, 1, 1},
  {&__pyx_n_u_visitClass_definition, __pyx_k_visitClass_definition, sizeof(__pyx_k_visitClass_definition), 0, 1, 0, 1},
  {&__pyx_n_s_visitClass_modification, __pyx_k_visitClass_modification, sizeof(__pyx_k_visitClass_modification), 0, 0, 1, 1},
  {&__pyx_n_u_visitClass_modification, __pyx_k_visitClass_modification, sizeof(__pyx_k_visitClass_modification), 0, 1, 0, 1},
  {&__pyx_n_s_visitClass_prefixes, __pyx_k_visitClass_prefixes, sizeof(__pyx_k_visitClass_prefixes), 0, 0, 1, 1},
  {&__pyx_n_u_visitClass_prefixes, __pyx_k_visitClass_prefixes, sizeof(__pyx_k_visitClass_prefixes), 0, 1, 0, 1},
  {&__pyx_n_s_visitClass_spec_base, __pyx_k_visitClass_spec_base, sizeof(__pyx_k_visitClass_spec_base), 0, 0, 1, 1},
  {&__pyx_n_u_visitClass_spec_base, __pyx_k_visitClass_spec_base, sizeof(__pyx_k_visitClass_spec_base), 0, 1, 0, 1},
  {&__pyx_n_s_visitClass_spec_comp, __pyx_k_visitClass_spec_comp, sizeof(__pyx_k_visitClass_spec_comp), 0, 0, 1, 1},
  {&__pyx_n_u_visitClass_spec_comp, __pyx_k_visitClass_spec_comp, sizeof(__pyx_k_visitClass_spec_comp), 0, 1, 0, 1},
  {&__pyx_n_s_visitClass_spec_der, __pyx_k_visitClass_spec_der, sizeof(__pyx_k_visitClass_spec_der), 0, 0, 1, 1},
  {&__pyx_n_u_visitClass_spec_der, __pyx_k_visitClass_spec_der, sizeof(__pyx_k_visitClass_spec_der), 0, 1, 0, 1},
  {&__pyx_n_s_visitClass_spec_enum, __pyx_k_visitClass_spec_enum, sizeof(__pyx_k_visitClass_spec_enum), 0, 0, 1, 1},
  {&__pyx_n_u_visitClass_spec_enum, __pyx_k_visitClass_spec_enum, sizeof(__pyx_k_visitClass_spec_enum), 0, 1, 0, 1},
  {&__pyx_n_s_visitClass_spec_extends, __pyx_k_visitClass_spec_extends, sizeof(__pyx_k_visitClass_spec_extends), 0, 0, 1, 1},
  {&__pyx_n_u_visitClass_spec_extends, __pyx_k_visitClass_spec_extends, sizeof(__pyx_k_visitClass_spec_extends), 0, 1, 0, 1},
  {&__pyx_n_s_visitClass_type, __pyx_k_visitClass_type, sizeof(__pyx_k_visitClass_type), 0, 0, 1, 1},
  {&__pyx_n_u_visitClass_type, __pyx_k_visitClass_type, sizeof(__pyx_k_visitClass_type), 0, 1, 0, 1},
  {&__pyx_n_s_visitComment, __pyx_k_visitComment, sizeof(__pyx_k_visitComment), 0, 0, 1, 1},
  {&__pyx_n_u_visitComment, __pyx_k_visitComment, sizeof(__pyx_k_visitComment), 0, 1, 0, 1},
  {&__pyx_n_s_visitComponent_clause, __pyx_k_visitComponent_clause, sizeof(__pyx_k_visitComponent_clause), 0, 0, 1, 1},
  {&__pyx_n_u_visitComponent_clause, __pyx_k_visitComponent_clause, sizeof(__pyx_k_visitComponent_clause), 0, 1, 0, 1},
  {&__pyx_n_s_visitComponent_clause1, __pyx_k_visitComponent_clause1, sizeof(__pyx_k_visitComponent_clause1), 0, 0, 1, 1},
  {&__pyx_n_u_visitComponent_clause1, __pyx_k_visitComponent_clause1, sizeof(__pyx_k_visitComponent_clause1), 0, 1, 0, 1},
  {&__pyx_n_s_visitComponent_declaration, __pyx_k_visitComponent_declaration, sizeof(__pyx_k_visitComponent_declaration), 0, 0, 1, 1},
  {&__pyx_n_u_visitComponent_declaration, __pyx_k_visitComponent_declaration, sizeof(__pyx_k_visitComponent_declaration), 0, 1, 0, 1},
  {&__pyx_n_s_visitComponent_declaration1, __pyx_k_visitComponent_declaration1, sizeof(__pyx_k_visitComponent_declaration1), 0, 0, 1, 1},
  {&__pyx_n_u_visitComponent_declaration1, __pyx_k_visitComponent_declaration1, sizeof(__pyx_k_visitComponent_declaration1), 0, 1, 0, 1},
  {&__pyx_n_s_visitComponent_list, __pyx_k_visitComponent_list, sizeof(__pyx_k_visitComponent_list), 0, 0, 1, 1},
  {&__pyx_n_u_visitComponent_list, __pyx_k_visitComponent_list, sizeof(__pyx_k_visitComponent_list), 0, 1, 0, 1},
  {&__pyx_n_s_visitComponent_reference, __pyx_k_visitComponent_reference, sizeof(__pyx_k_visitComponent_reference), 0, 0, 1, 1},
  {&__pyx_n_u_visitComponent_reference, __pyx_k_visitComponent_reference, sizeof(__pyx_k_visitComponent_reference), 0, 1, 0, 1},
  {&__pyx_n_s_visitComponent_reference_element, __pyx_k_visitComponent_reference_element, sizeof(__pyx_k_visitComponent_reference_element), 0, 0, 1, 1},
  {&__pyx_n_u_visitComponent_reference_element, __pyx_k_visitComponent_reference_element, sizeof(__pyx_k_visitComponent_reference_element), 0, 1, 0, 1},
  {&__pyx_n_s_visitComposition, __pyx_k_visitComposition, sizeof(__pyx_k_visitComposition), 0, 0, 1, 1},
  {&__pyx_n_u_visitComposition, __pyx_k_visitComposition, sizeof(__pyx_k_visitComposition), 0, 1, 0, 1},
  {&__pyx_n_s_visitCondition_attribute, __pyx_k_visitCondition_attribute, sizeof(__pyx_k_visitCondition_attribute), 0, 0, 1, 1},
  {&__pyx_n_u_visitCondition_attribute, __pyx_k_visitCondition_attribute, sizeof(__pyx_k_visitCondition_attribute), 0, 1, 0, 1},
  {&__pyx_n_s_visitConnect_clause, __pyx_k_visitConnect_clause, sizeof(__pyx_k_visitConnect_clause), 0, 0, 1, 1},
  {&__pyx_n_u_visitConnect_clause, __pyx_k_visitConnect_clause, sizeof(__pyx_k_visitConnect_clause), 0, 1, 0, 1},
  {&__pyx_n_s_visitConstraining_clause, __pyx_k_visitConstraining_clause, sizeof(__pyx_k_visitConstraining_clause), 0, 0, 1, 1},
  {&__pyx_n_u_visitConstraining_clause, __pyx_k_visitConstraining_clause, sizeof(__pyx_k_visitConstraining_clause), 0, 1, 0, 1},
  {&__pyx_n_s_visitDeclaration, __pyx_k_visitDeclaration, sizeof(__pyx_k_visitDeclaration), 0, 0, 1, 1},
  {&__pyx_n_u_visitDeclaration, __pyx_k_visitDeclaration, sizeof(__pyx_k_visitDeclaration), 0, 1, 0, 1},
  {&__pyx_n_s_visitElement, __pyx_k_visitElement, sizeof(__pyx_k_visitElement), 0, 0, 1, 1},
  {&__pyx_n_u_visitElement, __pyx_k_visitElement, sizeof(__pyx_k_visitElement), 0, 1, 0, 1},
  {&__pyx_n_s_visitElement_list, __pyx_k_visitElement_list, sizeof(__pyx_k_visitElement_list), 0, 0, 1, 1},
  {&__pyx_n_u_visitElement_list, __pyx_k_visitElement_list, sizeof(__pyx_k_visitElement_list), 0, 1, 0, 1},
  {&__pyx_n_s_visitElement_modification, __pyx_k_visitElement_modification, sizeof(__pyx_k_visitElement_modification), 0, 0, 1, 1},
  {&__pyx_n_u_visitElement_modification, __pyx_k_visitElement_modification, sizeof(__pyx_k_visitElement_modification), 0, 1, 0, 1},
  {&__pyx_n_s_visitElement_modification_or_rep, __pyx_k_visitElement_modification_or_rep, sizeof(__pyx_k_visitElement_modification_or_rep), 0, 0, 1, 1},
  {&__pyx_n_u_visitElement_modification_or_rep, __pyx_k_visitElement_modification_or_rep, sizeof(__pyx_k_visitElement_modification_or_rep), 0, 1, 0, 1},
  {&__pyx_n_s_visitElement_redeclaration, __pyx_k_visitElement_redeclaration, sizeof(__pyx_k_visitElement_redeclaration), 0, 0, 1, 1},
  {&__pyx_n_u_visitElement_redeclaration, __pyx_k_visitElement_redeclaration, sizeof(__pyx_k_visitElement_redeclaration), 0, 1, 0, 1},
  {&__pyx_n_s_visitElement_replaceable, __pyx_k_visitElement_replaceable, sizeof(__pyx_k_visitElement_replaceable), 0, 0, 1, 1},
  {&__pyx_n_u_visitElement_replaceable, __pyx_k_visitElement_replaceable, sizeof(__pyx_k_visitElement_replaceable), 0, 1, 0, 1},
  {&__pyx_n_s_visitEnum_list, __pyx_k_visitEnum_list, sizeof(__pyx_k_visitEnum_list), 0, 0, 1, 1},
  {&__pyx_n_u_visitEnum_list, __pyx_k_visitEnum_list, sizeof(__pyx_k_visitEnum_list), 0, 1, 0, 1},
  {&__pyx_n_s_visitEnumeration_literal, __pyx_k_visitEnumeration_literal, sizeof(__pyx_k_visitEnumeration_literal), 0, 0, 1, 1},
  {&__pyx_n_u_visitEnumeration_literal, __pyx_k_visitEnumeration_literal, sizeof(__pyx_k_visitEnumeration_literal), 0, 1, 0, 1},
  {&__pyx_n_s_visitEquation, __pyx_k_visitEquation, sizeof(__pyx_k_visitEquation), 0, 0, 1, 1},
  {&__pyx_n_u_visitEquation, __pyx_k_visitEquation, sizeof(__pyx_k_visitEquation), 0, 1, 0, 1},
  {&__pyx_n_s_visitEquation_connect_clause, __pyx_k_visitEquation_connect_clause, sizeof(__pyx_k_visitEquation_connect_clause), 0, 0, 1, 1},
  {&__pyx_n_u_visitEquation_connect_clause, __pyx_k_visitEquation_connect_clause, sizeof(__pyx_k_visitEquation_connect_clause), 0, 1, 0, 1},
  {&__pyx_n_s_visitEquation_for, __pyx_k_visitEquation_for, sizeof(__pyx_k_visitEquation_for), 0, 0, 1, 1},
  {&__pyx_n_u_visitEquation_for, __pyx_k_visitEquation_for, sizeof(__pyx_k_visitEquation_for), 0, 1, 0, 1},
  {&__pyx_n_s_visitEquation_function, __pyx_k_visitEquation_function, sizeof(__pyx_k_visitEquation_function), 0, 0, 1, 1},
  {&__pyx_n_u_visitEquation_function, __pyx_k_visitEquation_function, sizeof(__pyx_k_visitEquation_function), 0, 1, 0, 1},
  {&__pyx_n_s_visitEquation_if, __pyx_k_visitEquation_if, sizeof(__pyx_k_visitEquation_if), 0, 0, 1, 1},
  {&__pyx_n_u_visitEquation_if, __pyx_k_visitEquation_if, sizeof(__pyx_k_visitEquation_if), 0, 1, 0, 1},
  {&__pyx_n_s_visitEquation_section, __pyx_k_visitEquation_section, sizeof(__pyx_k_visitEquation_section), 0, 0, 1, 1},
  {&__pyx_n_u_visitEquation_section, __pyx_k_visitEquation_section, sizeof(__pyx_k_visitEquation_section), 0, 1, 0, 1},
  {&__pyx_n_s_visitEquation_simple, __pyx_k_visitEquation_simple, sizeof(__pyx_k_visitEquation_simple), 0, 0, 1, 1},
  {&__pyx_n_u_visitEquation_simple, __pyx_k_visitEquation_simple, sizeof(__pyx_k_visitEquation_simple), 0, 1, 0, 1},
  {&__pyx_n_s_visitEquation_when, __pyx_k_visitEquation_when, sizeof(__pyx_k_visitEquation_when), 0, 0, 1, 1},
  {&__pyx_n_u_visitEquation_when, __pyx_k_visitEquation_when, sizeof(__pyx_k_visitEquation_when), 0, 1, 0, 1},
  {&__pyx_n_s_visitExpr_add, __pyx_k_visitExpr_add, sizeof(__pyx_k_visitExpr_add), 0, 0, 1, 1},
  {&__pyx_n_u_visitExpr_add, __pyx_k_visitExpr_add, sizeof(__pyx_k_visitExpr_add), 0, 1, 0, 1},
  {&__pyx_n_s_visitExpr_and, __pyx_k_visitExpr_and, sizeof(__pyx_k_visitExpr_and), 0, 0, 1, 1},
  {&__pyx_n_u_visitExpr_and, __pyx_k_visitExpr_and, sizeof(__pyx_k_visitExpr_and), 0, 1, 0, 1},
  {&__pyx_n_s_visitExpr_exp, __pyx_k_visitExpr_exp, sizeof(__pyx_k_visitExpr_exp), 0, 0, 1, 1},
  {&__pyx_n_u_visitExpr_exp, __pyx_k_visitExpr_exp, sizeof(__pyx_k_visitExpr_exp), 0, 1, 0, 1},
  {&__pyx_n_s_visitExpr_mul, __pyx_k_visitExpr_mul, sizeof(__pyx_k_visitExpr_mul), 0, 0, 1, 1},
  {&__pyx_n_u_visitExpr_mul, __pyx_k_visitExpr_mul, sizeof(__pyx_k_visitExpr_mul), 0, 1, 0, 1},
  {&__pyx_n_s_visitExpr_neg, __pyx_k_visitExpr_neg, sizeof(__pyx_k_visitExpr_neg), 0, 0, 1, 1},
  {&__pyx_n_u_visitExpr_neg, __pyx_k_visitExpr_neg, sizeof(__pyx_k_visitExpr_neg), 0, 1, 0, 1},
  {&__pyx_n_s_visitExpr_not, __pyx_k_visitExpr_not, sizeof(__pyx_k_visitExpr_not), 0, 0, 1, 1},
  {&__pyx_n_u_visitExpr_not, __pyx_k_visitExpr_not, sizeof(__pyx_k_visitExpr_not), 0, 1, 0, 1},
  {&__pyx_n_s_visitExpr_or, __pyx_k_visitExpr_or, sizeof(__pyx_k_visitExpr_or), 0, 0, 1, 1},
  {&__pyx_n_u_visitExpr_or, __pyx_k_visitExpr_or, sizeof(__pyx_k_visitExpr_or), 0, 1, 0, 1},
  {&__pyx_n_s_visitExpr_primary, __pyx_k_visitExpr_primary, sizeof(__pyx_k_visitExpr_primary), 0, 0, 1, 1},
  {&__pyx_n_u_visitExpr_primary, __pyx_k_visitExpr_primary, sizeof(__pyx_k_visitExpr_primary), 0, 1, 0, 1},
  {&__pyx_n_s_visitExpr_rel, __pyx_k_visitExpr_rel, sizeof(__pyx_k_visitExpr_rel), 0, 0, 1, 1},
  {&__pyx_n_u_visitExpr_rel, __pyx_k_visitExpr_rel, sizeof(__pyx_k_visitExpr_rel), 0, 1, 0, 1},
  {&__pyx_n_s_visitExpression_if, __pyx_k_visitExpression_if, sizeof(__pyx_k_visitExpression_if), 0, 0, 1, 1},
  {&__pyx_n_u_visitExpression_if, __pyx_k_visitExpression_if, sizeof(__pyx_k_visitExpression_if), 0, 1, 0, 1},
  {&__pyx_n_s_visitExpression_list, __pyx_k_visitExpression_list, sizeof(__pyx_k_visitExpression_list), 0, 0, 1, 1},
  {&__pyx_n_u_visitExpression_list, __pyx_k_visitExpression_list, sizeof(__pyx_k_visitExpression_list), 0, 1, 0, 1},
  {&__pyx_n_s_visitExpression_simple, __pyx_k_visitExpression_simple, sizeof(__pyx_k_visitExpression_simple), 0, 0, 1, 1},
  {&__pyx_n_u_visitExpression_simple, __pyx_k_visitExpression_simple, sizeof(__pyx_k_visitExpression_simple), 0, 1, 0, 1},
  {&__pyx_n_s_visitExtends_clause, __pyx_k_visitExtends_clause, sizeof(__pyx_k_visitExtends_clause), 0, 0, 1, 1},
  {&__pyx_n_u_visitExtends_clause, __pyx_k_visitExtends_clause, sizeof(__pyx_k_visitExtends_clause), 0, 1, 0, 1},
  {&__pyx_n_s_visitExternal_function_call, __pyx_k_visitExternal_function_call, sizeof(__pyx_k_visitExternal_function_call), 0, 0, 1, 1},
  {&__pyx_n_u_visitExternal_function_call, __pyx_k_visitExternal_function_call, sizeof(__pyx_k_visitExternal_function_call), 0, 1, 0, 1},
  {&__pyx_n_s_visitFor_equation, __pyx_k_visitFor_equation, sizeof(__pyx_k_visitFor_equation), 0, 0, 1, 1},
  {&__pyx_n_u_visitFor_equation, __pyx_k_visitFor_equation, sizeof(__pyx_k_visitFor_equation), 0, 1, 0, 1},
  {&__pyx_n_s_visitFor_index, __pyx_k_visitFor_index, sizeof(__pyx_k_visitFor_index), 0, 0, 1, 1},
  {&__pyx_n_u_visitFor_index, __pyx_k_visitFor_index, sizeof(__pyx_k_visitFor_index), 0, 1, 0, 1},
  {&__pyx_n_s_visitFor_indices, __pyx_k_visitFor_indices, sizeof(__pyx_k_visitFor_indices), 0, 0, 1, 1},
  {&__pyx_n_u_visitFor_indices, __pyx_k_visitFor_indices, sizeof(__pyx_k_visitFor_indices), 0, 1, 0, 1},
  {&__pyx_n_s_visitFor_statement, __pyx_k_visitFor_statement, sizeof(__pyx_k_visitFor_statement), 0, 0, 1, 1},
  {&__pyx_n_u_visitFor_statement, __pyx_k_visitFor_statement, sizeof(__pyx_k_visitFor_statement), 0, 1, 0, 1},
  {&__pyx_n_s_visitFunction_arguments, __pyx_k_visitFunction_arguments, sizeof(__pyx_k_visitFunction_arguments), 0, 0, 1, 1},
  {&__pyx_n_u_visitFunction_arguments, __pyx_k_visitFunction_arguments, sizeof(__pyx_k_visitFunction_arguments), 0, 1, 0, 1},
  {&__pyx_n_s_visitFunction_call_args, __pyx_k_visitFunction_call_args, sizeof(__pyx_k_visitFunction_call_args), 0, 0, 1, 1},
  {&__pyx_n_u_visitFunction_call_args, __pyx_k_visitFunction_call_args, sizeof(__pyx_k_visitFunction_call_args), 0, 1, 0, 1},
  {&__pyx_n_s_visitIf_equation, __pyx_k_visitIf_equation, sizeof(__pyx_k_visitIf_equation), 0, 0, 1, 1},
  {&__pyx_n_u_visitIf_equation, __pyx_k_visitIf_equation, sizeof(__pyx_k_visitIf_equation), 0, 1, 0, 1},
  {&__pyx_n_s_visitIf_statement, __pyx_k_visitIf_statement, sizeof(__pyx_k_visitIf_statement), 0, 0, 1, 1},
  {&__pyx_n_u_visitIf_statement, __pyx_k_visitIf_statement, sizeof(__pyx_k_visitIf_statement), 0, 1, 0, 1},
  {&__pyx_n_s_visitImport_clause, __pyx_k_visitImport_clause, sizeof(__pyx_k_visitImport_clause), 0, 0, 1, 1},
  {&__pyx_n_u_visitImport_clause, __pyx_k_visitImport_clause, sizeof(__pyx_k_visitImport_clause), 0, 1, 0, 1},
  {&__pyx_n_s_visitImport_list, __pyx_k_visitImport_list, sizeof(__pyx_k_visitImport_list), 0, 0, 1, 1},
  {&__pyx_n_u_visitImport_list, __pyx_k_visitImport_list, sizeof(__pyx_k_visitImport_list), 0, 1, 0, 1},
  {&__pyx_n_s_visitLanguage_specification, __pyx_k_visitLanguage_specification, sizeof(__pyx_k_visitLanguage_specification), 0, 0, 1, 1},
  {&__pyx_n_u_visitLanguage_specification, __pyx_k_visitLanguage_specification, sizeof(__pyx_k_visitLanguage_specification), 0, 1, 0, 1},
  {&__pyx_n_s_visitModification_assignment, __pyx_k_visitModification_assignment, sizeof(__pyx_k_visitModification_assignment), 0, 0, 1, 1},
  {&__pyx_n_u_visitModification_assignment, __pyx_k_visitModification_assignment, sizeof(__pyx_k_visitModification_assignment), 0, 1, 0, 1},
  {&__pyx_n_s_visitModification_assignment2, __pyx_k_visitModification_assignment2, sizeof(__pyx_k_visitModification_assignment2), 0, 0, 1, 1},
  {&__pyx_n_u_visitModification_assignment2, __pyx_k_visitModification_assignment2, sizeof(__pyx_k_visitModification_assignment2), 0, 1, 0, 1},
  {&__pyx_n_s_visitModification_class, __pyx_k_visitModification_class, sizeof(__pyx_k_visitModification_class), 0, 0, 1, 1},
  {&__pyx_n_u_visitModification_class, __pyx_k_visitModification_class, sizeof(__pyx_k_visitModification_class), 0, 1, 0, 1},
  {&__pyx_n_s_visitName, __pyx_k_visitName, sizeof(__pyx_k_visitName), 0, 0, 1, 1},
  {&__pyx_n_u_visitName, __pyx_k_visitName, sizeof(__pyx_k_visitName), 0, 1, 0, 1},
  {&__pyx_n_s_visitNamed_argument, __pyx_k_visitNamed_argument, sizeof(__pyx_k_visitNamed_argument), 0, 0, 1, 1},
  {&__pyx_n_u_visitNamed_argument, __pyx_k_visitNamed_argument, sizeof(__pyx_k_visitNamed_argument), 0, 1, 0, 1},
  {&__pyx_n_s_visitNamed_arguments, __pyx_k_visitNamed_arguments, sizeof(__pyx_k_visitNamed_arguments), 0, 0, 1, 1},
  {&__pyx_n_u_visitNamed_arguments, __pyx_k_visitNamed_arguments, sizeof(__pyx_k_visitNamed_arguments), 0, 1, 0, 1},
  {&__pyx_n_s_visitOutput_expression_list, __pyx_k_visitOutput_expression_list, sizeof(__pyx_k_visitOutput_expression_list), 0, 0, 1, 1},
  {&__pyx_n_u_visitOutput_expression_list, __pyx_k_visitOutput_expression_list, sizeof(__pyx_k_visitOutput_expression_list), 0, 1, 0, 1},
  {&__pyx_n_s_visitPrimary_component_reference, __pyx_k_visitPrimary_component_reference, sizeof(__pyx_k_visitPrimary_component_reference), 0, 0, 1, 1},
  {&__pyx_n_u_visitPrimary_component_reference, __pyx_k_visitPrimary_component_reference, sizeof(__pyx_k_visitPrimary_component_reference), 0, 1, 0, 1},
  {&__pyx_n_s_visitPrimary_derivative, __pyx_k_visitPrimary_derivative, sizeof(__pyx_k_visitPrimary_derivative), 0, 0, 1, 1},
  {&__pyx_n_u_visitPrimary_derivative, __pyx_k_visitPrimary_derivative, sizeof(__pyx_k_visitPrimary_derivative), 0, 1, 0, 1},
  {&__pyx_n_s_visitPrimary_end, __pyx_k_visitPrimary_end, sizeof(__pyx_k_visitPrimary_end), 0, 0, 1, 1},
  {&__pyx_n_u_visitPrimary_end, __pyx_k_visitPrimary_end, sizeof(__pyx_k_visitPrimary_end), 0, 1, 0, 1},
  {&__pyx_n_s_visitPrimary_expression_list, __pyx_k_visitPrimary_expression_list, sizeof(__pyx_k_visitPrimary_expression_list), 0, 0, 1, 1},
  {&__pyx_n_u_visitPrimary_expression_list, __pyx_k_visitPrimary_expression_list, sizeof(__pyx_k_visitPrimary_expression_list), 0, 1, 0, 1},
  {&__pyx_n_s_visitPrimary_false, __pyx_k_visitPrimary_false, sizeof(__pyx_k_visitPrimary_false), 0, 0, 1, 1},
  {&__pyx_n_u_visitPrimary_false, __pyx_k_visitPrimary_false, sizeof(__pyx_k_visitPrimary_false), 0, 1, 0, 1},
  {&__pyx_n_s_visitPrimary_function, __pyx_k_visitPrimary_function, sizeof(__pyx_k_visitPrimary_function), 0, 0, 1, 1},
  {&__pyx_n_u_visitPrimary_function, __pyx_k_visitPrimary_function, sizeof(__pyx_k_visitPrimary_function), 0, 1, 0, 1},
  {&__pyx_n_s_visitPrimary_function_arguments, __pyx_k_visitPrimary_function_arguments, sizeof(__pyx_k_visitPrimary_function_arguments), 0, 0, 1, 1},
  {&__pyx_n_u_visitPrimary_function_arguments, __pyx_k_visitPrimary_function_arguments, sizeof(__pyx_k_visitPrimary_function_arguments), 0, 1, 0, 1},
  {&__pyx_n_s_visitPrimary_initial, __pyx_k_visitPrimary_initial, sizeof(__pyx_k_visitPrimary_initial), 0, 0, 1, 1},
  {&__pyx_n_u_visitPrimary_initial, __pyx_k_visitPrimary_initial, sizeof(__pyx_k_visitPrimary_initial), 0, 1, 0, 1},
  {&__pyx_n_s_visitPrimary_output_expression_l, __pyx_k_visitPrimary_output_expression_l, sizeof(__pyx_k_visitPrimary_output_expression_l), 0, 0, 1, 1},
  {&__pyx_n_u_visitPrimary_output_expression_l, __pyx_k_visitPrimary_output_expression_l, sizeof(__pyx_k_visitPrimary_output_expression_l), 0, 1, 0, 1},
  {&__pyx_n_s_visitPrimary_string, __pyx_k_visitPrimary_string, sizeof(__pyx_k_visitPrimary_string), 0, 0, 1, 1},
  {&__pyx_n_u_visitPrimary_string, __pyx_k_visitPrimary_string, sizeof(__pyx_k_visitPrimary_string), 0, 1, 0, 1},
  {&__pyx_n_s_visitPrimary_true, __pyx_k_visitPrimary_true, sizeof(__pyx_k_visitPrimary_true), 0, 0, 1, 1},
  {&__pyx_n_u_visitPrimary_true, __pyx_k_visitPrimary_true, sizeof(__pyx_k_visitPrimary_true), 0, 1, 0, 1},
  {&__pyx_n_s_visitPrimary_unsigned_number, __pyx_k_visitPrimary_unsigned_number, sizeof(__pyx_k_visitPrimary_unsigned_number), 0, 0, 1, 1},
  {&__pyx_n_u_visitPrimary_unsigned_number, __pyx_k_visitPrimary_unsigned_number, sizeof(__pyx_k_visitPrimary_unsigned_number), 0, 1, 0, 1},
  {&__pyx_n_s_visitRegular_element, __pyx_k_visitRegular_element, sizeof(__pyx_k_visitRegular_element), 0, 0, 1, 1},
  {&__pyx_n_u_visitRegular_element, __pyx_k_visitRegular_element, sizeof(__pyx_k_visitRegular_element), 0, 1, 0, 1},
  {&__pyx_n_s_visitReplaceable_element, __pyx_k_visitReplaceable_element, sizeof(__pyx_k_visitReplaceable_element), 0, 0, 1, 1},
  {&__pyx_n_u_visitReplaceable_element, __pyx_k_visitReplaceable_element, sizeof(__pyx_k_visitReplaceable_element), 0, 1, 0, 1},
  {&__pyx_n_s_visitShort_class_definition, __pyx_k_visitShort_class_definition, sizeof(__pyx_k_visitShort_class_definition), 0, 0, 1, 1},
  {&__pyx_n_u_visitShort_class_definition, __pyx_k_visitShort_class_definition, sizeof(__pyx_k_visitShort_class_definition), 0, 1, 0, 1},
  {&__pyx_n_s_visitSimple_expression, __pyx_k_visitSimple_expression, sizeof(__pyx_k_visitSimple_expression), 0, 0, 1, 1},
  {&__pyx_n_u_visitSimple_expression, __pyx_k_visitSimple_expression, sizeof(__pyx_k_visitSimple_expression), 0, 1, 0, 1},
  {&__pyx_n_s_visitStatement, __pyx_k_visitStatement, sizeof(__pyx_k_visitStatement), 0, 0, 1, 1},
  {&__pyx_n_u_visitStatement, __pyx_k_visitStatement, sizeof(__pyx_k_visitStatement), 0, 1, 0, 1},
  {&__pyx_n_s_visitStatement_break, __pyx_k_visitStatement_break, sizeof(__pyx_k_visitStatement_break), 0, 0, 1, 1},
  {&__pyx_n_u_visitStatement_break, __pyx_k_visitStatement_break, sizeof(__pyx_k_visitStatement_break), 0, 1, 0, 1},
  {&__pyx_n_s_visitStatement_component_functio, __pyx_k_visitStatement_component_functio, sizeof(__pyx_k_visitStatement_component_functio), 0, 0, 1, 1},
  {&__pyx_n_u_visitStatement_component_functio, __pyx_k_visitStatement_component_functio, sizeof(__pyx_k_visitStatement_component_functio), 0, 1, 0, 1},
  {&__pyx_n_s_visitStatement_component_referen, __pyx_k_visitStatement_component_referen, sizeof(__pyx_k_visitStatement_component_referen), 0, 0, 1, 1},
  {&__pyx_n_u_visitStatement_component_referen, __pyx_k_visitStatement_component_referen, sizeof(__pyx_k_visitStatement_component_referen), 0, 1, 0, 1},
  {&__pyx_n_s_visitStatement_for, __pyx_k_visitStatement_for, sizeof(__pyx_k_visitStatement_for), 0, 0, 1, 1},
  {&__pyx_n_u_visitStatement_for, __pyx_k_visitStatement_for, sizeof(__pyx_k_visitStatement_for), 0, 1, 0, 1},
  {&__pyx_n_s_visitStatement_if, __pyx_k_visitStatement_if, sizeof(__pyx_k_visitStatement_if), 0, 0, 1, 1},
  {&__pyx_n_u_visitStatement_if, __pyx_k_visitStatement_if, sizeof(__pyx_k_visitStatement_if), 0, 1, 0, 1},
  {&__pyx_n_s_visitStatement_return, __pyx_k_visitStatement_return, sizeof(__pyx_k_visitStatement_return), 0, 0, 1, 1},
  {&__pyx_n_u_visitStatement_return, __pyx_k_visitStatement_return, sizeof(__pyx_k_visitStatement_return), 0, 1, 0, 1},
  {&__pyx_n_s_visitStatement_when, __pyx_k_visitStatement_when, sizeof(__pyx_k_visitStatement_when), 0, 0, 1, 1},
  {&__pyx_n_u_visitStatement_when, __pyx_k_visitStatement_when, sizeof(__pyx_k_visitStatement_when), 0, 1, 0, 1},
  {&__pyx_n_s_visitStatement_while, __pyx_k_visitStatement_while, sizeof(__pyx_k_visitStatement_while), 0, 0, 1, 1},
  {&__pyx_n_u_visitStatement_while, __pyx_k_visitStatement_while, sizeof(__pyx_k_visitStatement_while), 0, 1, 0, 1},
  {&__pyx_n_s_visitStored_definition, __pyx_k_visitStored_definition, sizeof(__pyx_k_visitStored_definition), 0, 0, 1, 1},
  {&__pyx_n_u_visitStored_definition, __pyx_k_visitStored_definition, sizeof(__pyx_k_visitStored_definition), 0, 1, 0, 1},
  {&__pyx_n_s_visitStored_definition_class, __pyx_k_visitStored_definition_class, sizeof(__pyx_k_visitStored_definition_class), 0, 0, 1, 1},
  {&__pyx_n_u_visitStored_definition_class, __pyx_k_visitStored_definition_class, sizeof(__pyx_k_visitStored_definition_class), 0, 1, 0, 1},
  {&__pyx_n_s_visitString_comment, __pyx_k_visitString_comment, sizeof(__pyx_k_visitString_comment), 0, 0, 1, 1},
  {&__pyx_n_u_visitString_comment, __pyx_k_visitString_comment, sizeof(__pyx_k_visitString_comment), 0, 1, 0, 1},
  {&__pyx_n_s_visitSubscript, __pyx_k_visitSubscript, sizeof(__pyx_k_visitSubscript), 0, 0, 1, 1},
  {&__pyx_n_u_visitSubscript, __pyx_k_visitSubscript, sizeof(__pyx_k_visitSubscript), 0, 1, 0, 1},
  {&__pyx_n_s_visitType_prefix, __pyx_k_visitType_prefix, sizeof(__pyx_k_visitType_prefix), 0, 0, 1, 1},
  {&__pyx_n_u_visitType_prefix, __pyx_k_visitType_prefix, sizeof(__pyx_k_visitType_prefix), 0, 1, 0, 1},
  {&__pyx_n_s_visitType_specifier, __pyx_k_visitType_specifier, sizeof(__pyx_k_visitType_specifier), 0, 0, 1, 1},
  {&__pyx_n_u_visitType_specifier, __pyx_k_visitType_specifier, sizeof(__pyx_k_visitType_specifier), 0, 1, 0, 1},
  {&__pyx_n_s_visitWhen_equation, __pyx_k_visitWhen_equation, sizeof(__pyx_k_visitWhen_equation), 0, 0, 1, 1},
  {&__pyx_n_u_visitWhen_equation, __pyx_k_visitWhen_equation, sizeof(__pyx_k_visitWhen_equation), 0, 1, 0, 1},
  {&__pyx_n_s_visitWhen_statement, __pyx_k_visitWhen_statement, sizeof(__pyx_k_visitWhen_statement), 0, 0, 1, 1},
  {&__pyx_n_u_visitWhen_statement, __pyx_k_visitWhen_statement, sizeof(__pyx_k_visitWhen_statement), 0, 1, 0, 1},
  {&__pyx_n_s_visitWhile_statement, __pyx_k_visitWhile_statement, sizeof(__pyx_k_visitWhile_statement), 0, 0, 1, 1},
  {&__pyx_n_u_visitWhile_statement, __pyx_k_visitWhile_statement, sizeof(__pyx_k_visitWhile_statement), 0, 1, 0, 1},
  {&__pyx_n_s_visitor, __pyx_k_visitor, sizeof(__pyx_k_visitor), 0, 0, 1, 1},
  {&__pyx_kp_u_w, __pyx_k_w, sizeof(__pyx_k_w), 0, 1, 0, 0},
  {&__pyx_kp_u_when, __pyx_k_when, sizeof(__pyx_k_when), 0, 1, 0, 0},
  {&__pyx_n_s_when_equation, __pyx_k_when_equation, sizeof(__pyx_k_when_equation), 0, 0, 1, 1},
  {&__pyx_n_u_when_equation, __pyx_k_when_equation, sizeof(__pyx_k_when_equation), 0, 1, 0, 1},
  {&__pyx_n_s_when_statement, __pyx_k_when_statement, sizeof(__pyx_k_when_statement), 0, 0, 1, 1},
  {&__pyx_n_u_when_statement, __pyx_k_when_statement, sizeof(__pyx_k_when_statement), 0, 1, 0, 1},
  {&__pyx_kp_u_while, __pyx_k_while, sizeof(__pyx_k_while), 0, 1, 0, 0},
  {&__pyx_n_s_while_statement, __pyx_k_while_statement, sizeof(__pyx_k_while_statement), 0, 0, 1, 1},
  {&__pyx_n_u_while_statement, __pyx_k_while_statement, sizeof(__pyx_k_while_statement), 0, 1, 0, 1},
  {&__pyx_kp_u_within, __pyx_k_within, sizeof(__pyx_k_within), 0, 1, 0, 0},
  {&__pyx_n_s_write, __pyx_k_write, sizeof(__pyx_k_write), 0, 0, 1, 1},
  {&__pyx_kp_u_x_x, __pyx_k_x_x, sizeof(__pyx_k_x_x), 0, 1, 0, 0},
  {&__pyx_kp_u_x_z, __pyx_k_x_z, sizeof(__pyx_k_x_z), 0, 1, 0, 0},
  {&__pyx_kp_u_y, __pyx_k_y, sizeof(__pyx_k_y), 0, 1, 0, 0},
  {&__pyx_kp_u_z, __pyx_k_z, sizeof(__pyx_k_z), 0, 1, 0, 0},
  {&__pyx_kp_u_z_10, __pyx_k_z_10, sizeof(__pyx_k_z_10), 0, 1, 0, 0},
  {&__pyx_kp_u_z_2, __pyx_k_z_2, sizeof(__pyx_k_z_2), 0, 1, 0, 0},
  {&__pyx_kp_u_z_3, __pyx_k_z_3, sizeof(__pyx_k_z_3), 0, 1, 0, 0},
  {&__pyx_kp_u_z_4, __pyx_k_z_4, sizeof(__pyx_k_z_4), 0, 1, 0, 0},
  {&__pyx_kp_u_z_5, __pyx_k_z_5, sizeof(__pyx_k_z_5), 0, 1, 0, 0},
  {&__pyx_kp_u_z_6, __pyx_k_z_6, sizeof(__pyx_k_z_6), 0, 1, 0, 0},
  {&__pyx_kp_u_z_7, __pyx_k_z_7, sizeof(__pyx_k_z_7), 0, 1, 0, 0},
  {&__pyx_kp_u_z_8, __pyx_k_z_8, sizeof(__pyx_k_z_8), 0, 1, 0, 0},
  {&__pyx_kp_u_z_9, __pyx_k_z_9, sizeof(__pyx_k_z_9), 0, 1, 0, 0},
  {&__pyx_kp_u_z_F, __pyx_k_z_F, sizeof(__pyx_k_z_F), 0, 1, 0, 0},
  {&__pyx_kp_u_z_n8, __pyx_k_z_n8, sizeof(__pyx_k_z_n8), 0, 1, 0, 0},
  {0, 0, 0, 0, 0, 0, 0}
};
static int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(0, 538, __pyx_L1_error)
  __pyx_builtin_super = __Pyx_GetBuiltinName(__pyx_n_s_super); if (!__pyx_builtin_super) __PYX_ERR(0, 776, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "pymola/generated/ModelicaParser.pyx":10
 * def serializedATN():
 *     with StringIO() as buf:
 *         buf.write("\3\u608b\ua72a\u8133\ub9ed\u417c\u3be7\u7786\u5964\3^")             # <<<<<<<<<<<<<<
 *         buf.write("\u040c\4\2\t\2\4\3\t\3\4\4\t\4\4\5\t\5\4\6\t\6\4\7\t\7")
 *         buf.write("\4\b\t\b\4\t\t\t\4\n\t\n\4\13\t\13\4\f\t\f\4\r\t\r\4\16")
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_u_); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "pymola/generated/ModelicaParser.pyx":11
 *     with StringIO() as buf:
 *         buf.write("\3\u608b\ua72a\u8133\ub9ed\u417c\u3be7\u7786\u5964\3^")
 *         buf.write("\u040c\4\2\t\2\4\3\t\3\4\4\t\4\4\5\t\5\4\6\t\6\4\7\t\7")             # <<<<<<<<<<<<<<
 *         buf.write("\4\b\t\b\4\t\t\t\4\n\t\n\4\13\t\13\4\f\t\f\4\r\t\r\4\16")
 *         buf.write("\t\16\4\17\t\17\4\20\t\20\4\21\t\21\4\22\t\22\4\23\t\23")
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_u__3); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "pymola/generated/ModelicaParser.pyx":12
 *         buf.write("\3\u608b\ua72a\u8133\ub9ed\u417c\u3be7\u7786\u5964\3^")
 *         buf.write("\u040c\4\2\t\2\4\3\t\3\4\4\t\4\4\5\t\5\4\6\t\6\4\7\t\7")
 *         buf.write("\4\b\t\b\4\t\t\t\4\n\t\n\4\13\t\13\4\f\t\f\4\r\t\r\4\16")             # <<<<<<<<<<<<<<
 *         buf.write("\t\16\4\17\t\17\4\20\t\20\4\21\t\21\4\22\t\22\4\23\t\23")
 *         buf.write("\4\24\t\24\4\25\t\25\4\26\t\26\4\27\t\27\4\30\t\30\4\31")
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_u__5); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "pymola/generated/ModelicaParser.pyx":13
 *         buf.write("\u040c\4\2\t\2\4\3\t\3\4\4\t\4\4\5\t\5\4\6\t\6\4\7\t\7")
 *         buf.write("\4\b\t\b\4\t\t\t\4\n\t\n\4\13\t\13\4\f\t\f\4\r\t\r\4\16")
 *         buf.write("\t\16\4\17\t\17\4\20\t\20\4\21\t\21\4\22\t\22\4\23\t\23")             # <<<<<<<<<<<<<<
 *         buf.write("\4\24\t\24\4\25\t\25\4\26\t\26\4\27\t\27\4\30\t\30\4\31")
 *         buf.write("\t\31\4\32\t\32\4\33\t\33\4\34\t\34\4\35\t\35\4\36\t\36")
 */
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_kp_u__7); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "pymola/generated/ModelicaParser.pyx":14
 *         buf.write("\4\b\t\b\4\t\t\t\4\n\t\n\4\13\t\13\4\f\t\f\4\r\t\r\4\16")
 *         buf.write("\t\16\4\17\t\17\4\20\t\20\4\21\t\21\4\22\t\22\4\23\t\23")
 *         buf.write("\4\24\t\24\4\25\t\25\4\26\t\26\4\27\t\27\4\30\t\30\4\31")             # <<<<<<<<<<<<<<
 *         buf.write("\t\31\4\32\t\32\4\33\t\33\4\34\t\34\4\35\t\35\4\36\t\36")
 *         buf.write("\4\37\t\37\4 \t \4!\t!\4\"\t\"\4#\t#\4$\t$\4%\t%\4&\t")
 */
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_kp_u__9); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "pymola/generated/ModelicaParser.pyx":15
 *         buf.write("\t\16\4\17\t\17\4\20\t\20\4\21\t\21\4\22\t\22\4\23\t\23")
 *         buf.write("\4\24\t\24\4\25\t\25\4\26\t\26\4\27\t\27\4\30\t\30\4\31")
 *         buf.write("\t\31\4\32\t\32\4\33\t\33\4\34\t\34\4\35\t\35\4\36\t\36")             # <<<<<<<<<<<<<<
 *         buf.write("\4\37\t\37\4 \t \4!\t!\4\"\t\"\4#\t#\4$\t$\4%\t%\4&\t")
 *         buf.write("&\4\'\t\'\4(\t(\4)\t)\4*\t*\4+\t+\4,\t,\4-\t-\4.\t.\4")
 */
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_kp_u__11); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "pymola/generated/ModelicaParser.pyx":16
 *         buf.write("\4\24\t\24\4\25\t\25\4\26\t\26\4\27\t\27\4\30\t\30\4\31")
 *         buf.write("\t\31\4\32\t\32\4\33\t\33\4\34\t\34\4\35\t\35\4\36\t\36")
 *         buf.write("\4\37\t\37\4 \t \4!\t!\4\"\t\"\4#\t#\4$\t$\4%\t%\4&\t")             # <<<<<<<<<<<<<<
 *         buf.write("&\4\'\t\'\4(\t(\4)\t)\4*\t*\4+\t+\4,\t,\4-\t-\4.\t.\4")
 *         buf.write("/\t/\4\60\t\60\4\61\t\61\4\62\t\62\4\63\t\63\4\64\t\64")
 */
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_kp_u__13); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "pymola/generated/ModelicaParser.pyx":17
 *         buf.write("\t\31\4\32\t\32\4\33\t\33\4\34\t\34\4\35\t\35\4\36\t\36")
 *         buf.write("\4\37\t\37\4 \t \4!\t!\4\"\t\"\4#\t#\4$\t$\4%\t%\4&\t")
 *         buf.write("&\4\'\t\'\4(\t(\4)\t)\4*\t*\4+\t+\4,\t,\4-\t-\4.\t.\4")             # <<<<<<<<<<<<<<
 *         buf.write("/\t/\4\60\t\60\4\61\t\61\4\62\t\62\4\63\t\63\4\64\t\64")
 *         buf.write("\4\65\t\65\4\66\t\66\4\67\t\67\48\t8\49\t9\4:\t:\4;\t")
 */
  __pyx_tuple__16 = PyTuple_Pack(1, __pyx_kp_u__15); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "pymola/generated/ModelicaParser.pyx":18
 *         buf.write("\4\37\t\37\4 \t \4!\t!\4\"\t\"\4#\t#\4$\t$\4%\t%\4&\t")
 *         buf.write("&\4\'\t\'\4(\t(\4)\t)\4*\t*\4+\t+\4,\t,\4-\t-\4.\t.\4")
 *         buf.write("/\t/\4\60\t\60\4\61\t\61\4\62\t\62\4\63\t\63\4\64\t\64")             # <<<<<<<<<<<<<<
 *         buf.write("\4\65\t\65\4\66\t\66\4\67\t\67\48\t8\49\t9\4:\t:\4;\t")
 *         buf.write(";\4<\t<\4=\t=\4>\t>\4?\t?\4@\t@\4A\tA\4B\tB\4C\tC\4D\t")
 */
  __pyx_tuple__17 = PyTuple_Pack(1, __pyx_kp_u_0_0_1_1_2_2_3_3_4_4); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "pymola/generated/ModelicaParser.pyx":19
 *         buf.write("&\4\'\t\'\4(\t(\4)\t)\4*\t*\4+\t+\4,\t,\4-\t-\4.\t.\4")
 *         buf.write("/\t/\4\60\t\60\4\61\t\61\4\62\t\62\4\63\t\63\4\64\t\64")
 *         buf.write("\4\65\t\65\4\66\t\66\4\67\t\67\48\t8\49\t9\4:\t:\4;\t")             # <<<<<<<<<<<<<<
 *         buf.write(";\4<\t<\4=\t=\4>\t>\4?\t?\4@\t@\4A\tA\4B\tB\4C\tC\4D\t")
 *         buf.write("D\4E\tE\4F\tF\4G\tG\4H\tH\4I\tI\4J\tJ\3\2\3\2\5\2\u0097")
 */
  __pyx_tuple__18 = PyTuple_Pack(1, __pyx_kp_u_5_5_6_6_7_7_8_8_9_9); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);

  /* "pymola/generated/ModelicaParser.pyx":20
 *         buf.write("/\t/\4\60\t\60\4\61\t\61\4\62\t\62\4\63\t\63\4\64\t\64")
 *         buf.write("\4\65\t\65\4\66\t\66\4\67\t\67\48\t8\49\t9\4:\t:\4;\t")
 *         buf.write(";\4<\t<\4=\t=\4>\t>\4?\t?\4@\t@\4A\tA\4B\tB\4C\tC\4D\t")             # <<<<<<<<<<<<<<
 *         buf.write("D\4E\tE\4F\tF\4G\tG\4H\tH\4I\tI\4J\tJ\3\2\3\2\5\2\u0097")
 *         buf.write("\n\2\3\2\5\2\u009a\n\2\3\2\7\2\u009d\n\2\f\2\16\2\u00a0")
 */
  __pyx_tuple__19 = PyTuple_Pack(1, __pyx_kp_u_A_A_B_B_C_C_D); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);

  /* "pymola/generated/ModelicaParser.pyx":21
 *         buf.write("\4\65\t\65\4\66\t\66\4\67\t\67\48\t8\49\t9\4:\t:\4;\t")
 *         buf.write(";\4<\t<\4=\t=\4>\t>\4?\t?\4@\t@\4A\tA\4B\tB\4C\tC\4D\t")
 *         buf.write("D\4E\tE\4F\tF\4G\tG\4H\tH\4I\tI\4J\tJ\3\2\3\2\5\2\u0097")             # <<<<<<<<<<<<<<
 *         buf.write("\n\2\3\2\5\2\u009a\n\2\3\2\7\2\u009d\n\2\f\2\16\2\u00a0")
 *         buf.write("\13\2\3\3\5\3\u00a3\n\3\3\3\3\3\3\3\3\4\5\4\u00a9\n\4")
 */
  __pyx_tuple__20 = PyTuple_Pack(1, __pyx_kp_u_D_E_E_F_F_G_G_H_H_I_I_J_J); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);

  /* "pymola/generated/ModelicaParser.pyx":22
 *         buf.write(";\4<\t<\4=\t=\4>\t>\4?\t?\4@\t@\4A\tA\4B\tB\4C\tC\4D\t")
 *         buf.write("D\4E\tE\4F\tF\4G\tG\4H\tH\4I\tI\4J\tJ\3\2\3\2\5\2\u0097")
 *         buf.write("\n\2\3\2\5\2\u009a\n\2\3\2\7\2\u009d\n\2\f\2\16\2\u00a0")             # <<<<<<<<<<<<<<
 *         buf.write("\13\2\3\3\5\3\u00a3\n\3\3\3\3\3\3\3\3\4\5\4\u00a9\n\4")
 *         buf.write("\3\4\3\4\3\4\3\5\5\5\u00af\n\5\3\5\3\5\3\6\3\6\3\6\5\6")
 */
  __pyx_tuple__22 = PyTuple_Pack(1, __pyx_kp_u__21); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);

  /* "pymola/generated/ModelicaParser.pyx":23
 *         buf.write("D\4E\tE\4F\tF\4G\tG\4H\tH\4I\tI\4J\tJ\3\2\3\2\5\2\u0097")
 *         buf.write("\n\2\3\2\5\2\u009a\n\2\3\2\7\2\u009d\n\2\f\2\16\2\u00a0")
 *         buf.write("\13\2\3\3\5\3\u00a3\n\3\3\3\3\3\3\3\3\4\5\4\u00a9\n\4")             # <<<<<<<<<<<<<<
 *         buf.write("\3\4\3\4\3\4\3\5\5\5\u00af\n\5\3\5\3\5\3\6\3\6\3\6\5\6")
 *         buf.write("\u00b6\n\6\3\6\3\6\3\6\5\6\u00bb\n\6\3\6\3\6\3\6\3\6\5")
 */
  __pyx_tuple__24 = PyTuple_Pack(1, __pyx_kp_u__23); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);

  /* "pymola/generated/ModelicaParser.pyx":24
 *         buf.write("\n\2\3\2\5\2\u009a\n\2\3\2\7\2\u009d\n\2\f\2\16\2\u00a0")
 *         buf.write("\13\2\3\3\5\3\u00a3\n\3\3\3\3\3\3\3\3\4\5\4\u00a9\n\4")
 *         buf.write("\3\4\3\4\3\4\3\5\5\5\u00af\n\5\3\5\3\5\3\6\3\6\3\6\5\6")             # <<<<<<<<<<<<<<
 *         buf.write("\u00b6\n\6\3\6\3\6\3\6\5\6\u00bb\n\6\3\6\3\6\3\6\3\6\5")
 *         buf.write("\6\u00c1\n\6\3\6\5\6\u00c4\n\6\3\6\3\6\5\6\u00c8\n\6\3")
 */
  __pyx_tuple__26 = PyTuple_Pack(1, __pyx_kp_u__25); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);

  /* "pymola/generated/ModelicaParser.pyx":25
 *         buf.write("\13\2\3\3\5\3\u00a3\n\3\3\3\3\3\3\3\3\4\5\4\u00a9\n\4")
 *         buf.write("\3\4\3\4\3\4\3\5\5\5\u00af\n\5\3\5\3\5\3\6\3\6\3\6\5\6")
 *         buf.write("\u00b6\n\6\3\6\3\6\3\6\5\6\u00bb\n\6\3\6\3\6\3\6\3\6\5")             # <<<<<<<<<<<<<<
 *         buf.write("\6\u00c1\n\6\3\6\5\6\u00c4\n\6\3\6\3\6\5\6\u00c8\n\6\3")
 *         buf.write("\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\5\7\u00d5\n")
 */
  __pyx_tuple__28 = PyTuple_Pack(1, __pyx_kp_u__27); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);

  /* "pymola/generated/ModelicaParser.pyx":26
 *         buf.write("\3\4\3\4\3\4\3\5\5\5\u00af\n\5\3\5\3\5\3\6\3\6\3\6\5\6")
 *         buf.write("\u00b6\n\6\3\6\3\6\3\6\5\6\u00bb\n\6\3\6\3\6\3\6\3\6\5")
 *         buf.write("\6\u00c1\n\6\3\6\5\6\u00c4\n\6\3\6\3\6\5\6\u00c8\n\6\3")             # <<<<<<<<<<<<<<
 *         buf.write("\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\5\7\u00d5\n")
 *         buf.write("\7\3\7\5\7\u00d8\n\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\5\7\u00e1")
 */
  __pyx_tuple__30 = PyTuple_Pack(1, __pyx_kp_u__29); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);

  /* "pymola/generated/ModelicaParser.pyx":27
 *         buf.write("\u00b6\n\6\3\6\3\6\3\6\5\6\u00bb\n\6\3\6\3\6\3\6\3\6\5")
 *         buf.write("\6\u00c1\n\6\3\6\5\6\u00c4\n\6\3\6\3\6\5\6\u00c8\n\6\3")
 *         buf.write("\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\5\7\u00d5\n")             # <<<<<<<<<<<<<<
 *         buf.write("\7\3\7\5\7\u00d8\n\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\5\7\u00e1")
 *         buf.write("\n\7\3\7\5\7\u00e4\n\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7")
 */
  __pyx_tuple__32 = PyTuple_Pack(1, __pyx_kp_u__31); if (unlikely(!__pyx_tuple__32)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__32);
  __Pyx_GIVEREF(__pyx_tuple__32);

  /* "pymola/generated/ModelicaParser.pyx":28
 *         buf.write("\6\u00c1\n\6\3\6\5\6\u00c4\n\6\3\6\3\6\5\6\u00c8\n\6\3")
 *         buf.write("\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\5\7\u00d5\n")
 *         buf.write("\7\3\7\5\7\u00d8\n\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\5\7\u00e1")             # <<<<<<<<<<<<<<
 *         buf.write("\n\7\3\7\5\7\u00e4\n\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7")
 *         buf.write("\3\7\3\7\3\7\7\7\u00f1\n\7\f\7\16\7\u00f4\13\7\3\7\3\7")
 */
  __pyx_tuple__34 = PyTuple_Pack(1, __pyx_kp_u__33); if (unlikely(!__pyx_tuple__34)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__34);
  __Pyx_GIVEREF(__pyx_tuple__34);

  /* "pymola/generated/ModelicaParser.pyx":29
 *         buf.write("\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\5\7\u00d5\n")
 *         buf.write("\7\3\7\5\7\u00d8\n\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\5\7\u00e1")
 *         buf.write("\n\7\3\7\5\7\u00e4\n\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7")             # <<<<<<<<<<<<<<
 *         buf.write("\3\7\3\7\3\7\7\7\u00f1\n\7\f\7\16\7\u00f4\13\7\3\7\3\7")
 *         buf.write("\3\7\3\7\3\7\3\7\5\7\u00fc\n\7\3\7\3\7\3\7\3\7\3\7\5\7")
 */
  __pyx_tuple__36 = PyTuple_Pack(1, __pyx_kp_u__35); if (unlikely(!__pyx_tuple__36)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__36);
  __Pyx_GIVEREF(__pyx_tuple__36);

  /* "pymola/generated/ModelicaParser.pyx":30
 *         buf.write("\7\3\7\5\7\u00d8\n\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\5\7\u00e1")
 *         buf.write("\n\7\3\7\5\7\u00e4\n\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7")
 *         buf.write("\3\7\3\7\3\7\7\7\u00f1\n\7\f\7\16\7\u00f4\13\7\3\7\3\7")             # <<<<<<<<<<<<<<
 *         buf.write("\3\7\3\7\3\7\3\7\5\7\u00fc\n\7\3\7\3\7\3\7\3\7\3\7\5\7")
 *         buf.write("\u0103\n\7\3\b\3\b\3\t\3\t\3\t\7\t\u010a\n\t\f\t\16\t")
 */
  __pyx_tuple__38 = PyTuple_Pack(1, __pyx_kp_u__37); if (unlikely(!__pyx_tuple__38)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__38);
  __Pyx_GIVEREF(__pyx_tuple__38);

  /* "pymola/generated/ModelicaParser.pyx":31
 *         buf.write("\n\7\3\7\5\7\u00e4\n\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7")
 *         buf.write("\3\7\3\7\3\7\7\7\u00f1\n\7\f\7\16\7\u00f4\13\7\3\7\3\7")
 *         buf.write("\3\7\3\7\3\7\3\7\5\7\u00fc\n\7\3\7\3\7\3\7\3\7\3\7\5\7")             # <<<<<<<<<<<<<<
 *         buf.write("\u0103\n\7\3\b\3\b\3\t\3\t\3\t\7\t\u010a\n\t\f\t\16\t")
 *         buf.write("\u010d\13\t\3\n\3\n\3\n\3\13\3\13\3\13\3\13\3\13\3\13")
 */
  __pyx_tuple__40 = PyTuple_Pack(1, __pyx_kp_u__39); if (unlikely(!__pyx_tuple__40)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__40);
  __Pyx_GIVEREF(__pyx_tuple__40);

  /* "pymola/generated/ModelicaParser.pyx":32
 *         buf.write("\3\7\3\7\3\7\7\7\u00f1\n\7\f\7\16\7\u00f4\13\7\3\7\3\7")
 *         buf.write("\3\7\3\7\3\7\3\7\5\7\u00fc\n\7\3\7\3\7\3\7\3\7\3\7\5\7")
 *         buf.write("\u0103\n\7\3\b\3\b\3\t\3\t\3\t\7\t\u010a\n\t\f\t\16\t")             # <<<<<<<<<<<<<<
 *         buf.write("\u010d\13\t\3\n\3\n\3\n\3\13\3\13\3\13\3\13\3\13\3\13")
 *         buf.write("\3\13\7\13\u0119\n\13\f\13\16\13\u011c\13\13\3\13\3\13")
 */
  __pyx_tuple__42 = PyTuple_Pack(1, __pyx_kp_u__41); if (unlikely(!__pyx_tuple__42)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__42);
  __Pyx_GIVEREF(__pyx_tuple__42);

  /* "pymola/generated/ModelicaParser.pyx":33
 *         buf.write("\3\7\3\7\3\7\3\7\5\7\u00fc\n\7\3\7\3\7\3\7\3\7\3\7\5\7")
 *         buf.write("\u0103\n\7\3\b\3\b\3\t\3\t\3\t\7\t\u010a\n\t\f\t\16\t")
 *         buf.write("\u010d\13\t\3\n\3\n\3\n\3\13\3\13\3\13\3\13\3\13\3\13")             # <<<<<<<<<<<<<<
 *         buf.write("\3\13\7\13\u0119\n\13\f\13\16\13\u011c\13\13\3\13\3\13")
 *         buf.write("\5\13\u0120\n\13\3\13\5\13\u0123\n\13\3\13\5\13\u0126")
 */
  __pyx_tuple__44 = PyTuple_Pack(1, __pyx_kp_u__43); if (unlikely(!__pyx_tuple__44)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__44);
  __Pyx_GIVEREF(__pyx_tuple__44);

  /* "pymola/generated/ModelicaParser.pyx":34
 *         buf.write("\u0103\n\7\3\b\3\b\3\t\3\t\3\t\7\t\u010a\n\t\f\t\16\t")
 *         buf.write("\u010d\13\t\3\n\3\n\3\n\3\13\3\13\3\13\3\13\3\13\3\13")
 *         buf.write("\3\13\7\13\u0119\n\13\f\13\16\13\u011c\13\13\3\13\3\13")             # <<<<<<<<<<<<<<
 *         buf.write("\5\13\u0120\n\13\3\13\5\13\u0123\n\13\3\13\5\13\u0126")
 *         buf.write("\n\13\3\13\5\13\u0129\n\13\3\13\3\13\3\13\5\13\u012e\n")
 */
  __pyx_tuple__46 = PyTuple_Pack(1, __pyx_kp_u__45); if (unlikely(!__pyx_tuple__46)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__46);
  __Pyx_GIVEREF(__pyx_tuple__46);

  /* "pymola/generated/ModelicaParser.pyx":35
 *         buf.write("\u010d\13\t\3\n\3\n\3\n\3\13\3\13\3\13\3\13\3\13\3\13")
 *         buf.write("\3\13\7\13\u0119\n\13\f\13\16\13\u011c\13\13\3\13\3\13")
 *         buf.write("\5\13\u0120\n\13\3\13\5\13\u0123\n\13\3\13\5\13\u0126")             # <<<<<<<<<<<<<<
 *         buf.write("\n\13\3\13\5\13\u0129\n\13\3\13\3\13\3\13\5\13\u012e\n")
 *         buf.write("\13\3\f\3\f\3\r\3\r\3\r\5\r\u0135\n\r\3\r\3\r\3\r\5\r")
 */
  __pyx_tuple__48 = PyTuple_Pack(1, __pyx_kp_u__47); if (unlikely(!__pyx_tuple__48)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__48);
  __Pyx_GIVEREF(__pyx_tuple__48);

  /* "pymola/generated/ModelicaParser.pyx":36
 *         buf.write("\3\13\7\13\u0119\n\13\f\13\16\13\u011c\13\13\3\13\3\13")
 *         buf.write("\5\13\u0120\n\13\3\13\5\13\u0123\n\13\3\13\5\13\u0126")
 *         buf.write("\n\13\3\13\5\13\u0129\n\13\3\13\3\13\3\13\5\13\u012e\n")             # <<<<<<<<<<<<<<
 *         buf.write("\13\3\f\3\f\3\r\3\r\3\r\5\r\u0135\n\r\3\r\3\r\3\r\5\r")
 *         buf.write("\u013a\n\r\3\r\3\r\3\16\3\16\3\16\7\16\u0141\n\16\f\16")
 */
  __pyx_tuple__50 = PyTuple_Pack(1, __pyx_kp_u__49); if (unlikely(!__pyx_tuple__50)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__50);
  __Pyx_GIVEREF(__pyx_tuple__50);

  /* "pymola/generated/ModelicaParser.pyx":37
 *         buf.write("\5\13\u0120\n\13\3\13\5\13\u0123\n\13\3\13\5\13\u0126")
 *         buf.write("\n\13\3\13\5\13\u0129\n\13\3\13\3\13\3\13\5\13\u012e\n")
 *         buf.write("\13\3\f\3\f\3\r\3\r\3\r\5\r\u0135\n\r\3\r\3\r\3\r\5\r")             # <<<<<<<<<<<<<<
 *         buf.write("\u013a\n\r\3\r\3\r\3\16\3\16\3\16\7\16\u0141\n\16\f\16")
 *         buf.write("\16\16\u0144\13\16\3\17\3\17\3\17\3\17\5\17\u014a\n\17")
 */
  __pyx_tuple__52 = PyTuple_Pack(1, __pyx_kp_u__51); if (unlikely(!__pyx_tuple__52)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__52);
  __Pyx_GIVEREF(__pyx_tuple__52);

  /* "pymola/generated/ModelicaParser.pyx":38
 *         buf.write("\n\13\3\13\5\13\u0129\n\13\3\13\3\13\3\13\5\13\u012e\n")
 *         buf.write("\13\3\f\3\f\3\r\3\r\3\r\5\r\u0135\n\r\3\r\3\r\3\r\5\r")
 *         buf.write("\u013a\n\r\3\r\3\r\3\16\3\16\3\16\7\16\u0141\n\16\f\16")             # <<<<<<<<<<<<<<
 *         buf.write("\16\16\u0144\13\16\3\17\3\17\3\17\3\17\5\17\u014a\n\17")
 *         buf.write("\3\20\5\20\u014d\n\20\3\20\5\20\u0150\n\20\3\20\5\20\u0153")
 */
  __pyx_tuple__54 = PyTuple_Pack(1, __pyx_kp_u__53); if (unlikely(!__pyx_tuple__54)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__54);
  __Pyx_GIVEREF(__pyx_tuple__54);

  /* "pymola/generated/ModelicaParser.pyx":39
 *         buf.write("\13\3\f\3\f\3\r\3\r\3\r\5\r\u0135\n\r\3\r\3\r\3\r\5\r")
 *         buf.write("\u013a\n\r\3\r\3\r\3\16\3\16\3\16\7\16\u0141\n\16\f\16")
 *         buf.write("\16\16\u0144\13\16\3\17\3\17\3\17\3\17\5\17\u014a\n\17")             # <<<<<<<<<<<<<<
 *         buf.write("\3\20\5\20\u014d\n\20\3\20\5\20\u0150\n\20\3\20\5\20\u0153")
 *         buf.write("\n\20\3\20\5\20\u0156\n\20\3\20\3\20\5\20\u015a\n\20\3")
 */
  __pyx_tuple__56 = PyTuple_Pack(1, __pyx_kp_u__55); if (unlikely(!__pyx_tuple__56)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__56);
  __Pyx_GIVEREF(__pyx_tuple__56);

  /* "pymola/generated/ModelicaParser.pyx":40
 *         buf.write("\u013a\n\r\3\r\3\r\3\16\3\16\3\16\7\16\u0141\n\16\f\16")
 *         buf.write("\16\16\u0144\13\16\3\17\3\17\3\17\3\17\5\17\u014a\n\17")
 *         buf.write("\3\20\5\20\u014d\n\20\3\20\5\20\u0150\n\20\3\20\5\20\u0153")             # <<<<<<<<<<<<<<
 *         buf.write("\n\20\3\20\5\20\u0156\n\20\3\20\3\20\5\20\u015a\n\20\3")
 *         buf.write("\21\5\21\u015d\n\21\3\21\5\21\u0160\n\21\3\21\5\21\u0163")
 */
  __pyx_tuple__58 = PyTuple_Pack(1, __pyx_kp_u__57); if (unlikely(!__pyx_tuple__58)) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__58);
  __Pyx_GIVEREF(__pyx_tuple__58);

  /* "pymola/generated/ModelicaParser.pyx":41
 *         buf.write("\16\16\u0144\13\16\3\17\3\17\3\17\3\17\5\17\u014a\n\17")
 *         buf.write("\3\20\5\20\u014d\n\20\3\20\5\20\u0150\n\20\3\20\5\20\u0153")
 *         buf.write("\n\20\3\20\5\20\u0156\n\20\3\20\3\20\5\20\u015a\n\20\3")             # <<<<<<<<<<<<<<
 *         buf.write("\21\5\21\u015d\n\21\3\21\5\21\u0160\n\21\3\21\5\21\u0163")
 *         buf.write("\n\21\3\21\5\21\u0166\n\21\3\21\3\21\3\21\5\21\u016b\n")
 */
  __pyx_tuple__60 = PyTuple_Pack(1, __pyx_kp_u__59); if (unlikely(!__pyx_tuple__60)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__60);
  __Pyx_GIVEREF(__pyx_tuple__60);

  /* "pymola/generated/ModelicaParser.pyx":42
 *         buf.write("\3\20\5\20\u014d\n\20\3\20\5\20\u0150\n\20\3\20\5\20\u0153")
 *         buf.write("\n\20\3\20\5\20\u0156\n\20\3\20\3\20\5\20\u015a\n\20\3")
 *         buf.write("\21\5\21\u015d\n\21\3\21\5\21\u0160\n\21\3\21\5\21\u0163")             # <<<<<<<<<<<<<<
 *         buf.write("\n\21\3\21\5\21\u0166\n\21\3\21\3\21\3\21\5\21\u016b\n")
 *         buf.write("\21\3\21\3\21\3\21\5\21\u0170\n\21\3\22\3\22\3\22\3\22")
 */
  __pyx_tuple__62 = PyTuple_Pack(1, __pyx_kp_u__61); if (unlikely(!__pyx_tuple__62)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__62);
  __Pyx_GIVEREF(__pyx_tuple__62);

  /* "pymola/generated/ModelicaParser.pyx":43
 *         buf.write("\n\20\3\20\5\20\u0156\n\20\3\20\3\20\5\20\u015a\n\20\3")
 *         buf.write("\21\5\21\u015d\n\21\3\21\5\21\u0160\n\21\3\21\5\21\u0163")
 *         buf.write("\n\21\3\21\5\21\u0166\n\21\3\21\3\21\3\21\5\21\u016b\n")             # <<<<<<<<<<<<<<
 *         buf.write("\21\3\21\3\21\3\21\5\21\u0170\n\21\3\22\3\22\3\22\3\22")
 *         buf.write("\3\22\3\22\3\22\3\22\3\22\3\22\3\22\5\22\u017d\n\22\5")
 */
  __pyx_tuple__64 = PyTuple_Pack(1, __pyx_kp_u__63); if (unlikely(!__pyx_tuple__64)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__64);
  __Pyx_GIVEREF(__pyx_tuple__64);

  /* "pymola/generated/ModelicaParser.pyx":44
 *         buf.write("\21\5\21\u015d\n\21\3\21\5\21\u0160\n\21\3\21\5\21\u0163")
 *         buf.write("\n\21\3\21\5\21\u0166\n\21\3\21\3\21\3\21\5\21\u016b\n")
 *         buf.write("\21\3\21\3\21\3\21\5\21\u0170\n\21\3\22\3\22\3\22\3\22")             # <<<<<<<<<<<<<<
 *         buf.write("\3\22\3\22\3\22\3\22\3\22\3\22\3\22\5\22\u017d\n\22\5")
 *         buf.write("\22\u017f\n\22\5\22\u0181\n\22\3\22\3\22\3\23\3\23\3\23")
 */
  __pyx_tuple__66 = PyTuple_Pack(1, __pyx_kp_u__65); if (unlikely(!__pyx_tuple__66)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__66);
  __Pyx_GIVEREF(__pyx_tuple__66);

  /* "pymola/generated/ModelicaParser.pyx":45
 *         buf.write("\n\21\3\21\5\21\u0166\n\21\3\21\3\21\3\21\5\21\u016b\n")
 *         buf.write("\21\3\21\3\21\3\21\5\21\u0170\n\21\3\22\3\22\3\22\3\22")
 *         buf.write("\3\22\3\22\3\22\3\22\3\22\3\22\3\22\5\22\u017d\n\22\5")             # <<<<<<<<<<<<<<
 *         buf.write("\22\u017f\n\22\5\22\u0181\n\22\3\22\3\22\3\23\3\23\3\23")
 *         buf.write("\7\23\u0188\n\23\f\23\16\23\u018b\13\23\3\24\3\24\3\24")
 */
  __pyx_tuple__68 = PyTuple_Pack(1, __pyx_kp_u__67); if (unlikely(!__pyx_tuple__68)) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__68);
  __Pyx_GIVEREF(__pyx_tuple__68);

  /* "pymola/generated/ModelicaParser.pyx":46
 *         buf.write("\21\3\21\3\21\3\21\5\21\u0170\n\21\3\22\3\22\3\22\3\22")
 *         buf.write("\3\22\3\22\3\22\3\22\3\22\3\22\3\22\5\22\u017d\n\22\5")
 *         buf.write("\22\u017f\n\22\5\22\u0181\n\22\3\22\3\22\3\23\3\23\3\23")             # <<<<<<<<<<<<<<
 *         buf.write("\7\23\u0188\n\23\f\23\16\23\u018b\13\23\3\24\3\24\3\24")
 *         buf.write("\5\24\u0190\n\24\3\24\5\24\u0193\n\24\3\25\3\25\3\25\5")
 */
  __pyx_tuple__70 = PyTuple_Pack(1, __pyx_kp_u__69); if (unlikely(!__pyx_tuple__70)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__70);
  __Pyx_GIVEREF(__pyx_tuple__70);

  /* "pymola/generated/ModelicaParser.pyx":47
 *         buf.write("\3\22\3\22\3\22\3\22\3\22\3\22\3\22\5\22\u017d\n\22\5")
 *         buf.write("\22\u017f\n\22\5\22\u0181\n\22\3\22\3\22\3\23\3\23\3\23")
 *         buf.write("\7\23\u0188\n\23\f\23\16\23\u018b\13\23\3\24\3\24\3\24")             # <<<<<<<<<<<<<<
 *         buf.write("\5\24\u0190\n\24\3\24\5\24\u0193\n\24\3\25\3\25\3\25\5")
 *         buf.write("\25\u0198\n\25\3\26\3\26\3\26\5\26\u019d\n\26\3\26\3\26")
 */
  __pyx_tuple__72 = PyTuple_Pack(1, __pyx_kp_u__71); if (unlikely(!__pyx_tuple__72)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__72);
  __Pyx_GIVEREF(__pyx_tuple__72);

  /* "pymola/generated/ModelicaParser.pyx":48
 *         buf.write("\22\u017f\n\22\5\22\u0181\n\22\3\22\3\22\3\23\3\23\3\23")
 *         buf.write("\7\23\u0188\n\23\f\23\16\23\u018b\13\23\3\24\3\24\3\24")
 *         buf.write("\5\24\u0190\n\24\3\24\5\24\u0193\n\24\3\25\3\25\3\25\5")             # <<<<<<<<<<<<<<
 *         buf.write("\25\u0198\n\25\3\26\3\26\3\26\5\26\u019d\n\26\3\26\3\26")
 *         buf.write("\3\27\5\27\u01a2\n\27\3\27\5\27\u01a5\n\27\3\27\5\27\u01a8")
 */
  __pyx_tuple__74 = PyTuple_Pack(1, __pyx_kp_u__73); if (unlikely(!__pyx_tuple__74)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__74);
  __Pyx_GIVEREF(__pyx_tuple__74);

  /* "pymola/generated/ModelicaParser.pyx":49
 *         buf.write("\7\23\u0188\n\23\f\23\16\23\u018b\13\23\3\24\3\24\3\24")
 *         buf.write("\5\24\u0190\n\24\3\24\5\24\u0193\n\24\3\25\3\25\3\25\5")
 *         buf.write("\25\u0198\n\25\3\26\3\26\3\26\5\26\u019d\n\26\3\26\3\26")             # <<<<<<<<<<<<<<
 *         buf.write("\3\27\5\27\u01a2\n\27\3\27\5\27\u01a5\n\27\3\27\5\27\u01a8")
 *         buf.write("\n\27\3\30\3\30\3\31\3\31\3\31\7\31\u01af\n\31\f\31\16")
 */
  __pyx_tuple__76 = PyTuple_Pack(1, __pyx_kp_u__75); if (unlikely(!__pyx_tuple__76)) __PYX_ERR(0, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__76);
  __Pyx_GIVEREF(__pyx_tuple__76);

  /* "pymola/generated/ModelicaParser.pyx":50
 *         buf.write("\5\24\u0190\n\24\3\24\5\24\u0193\n\24\3\25\3\25\3\25\5")
 *         buf.write("\25\u0198\n\25\3\26\3\26\3\26\5\26\u019d\n\26\3\26\3\26")
 *         buf.write("\3\27\5\27\u01a2\n\27\3\27\5\27\u01a5\n\27\3\27\5\27\u01a8")             # <<<<<<<<<<<<<<
 *         buf.write("\n\27\3\30\3\30\3\31\3\31\3\31\7\31\u01af\n\31\f\31\16")
 *         buf.write("\31\u01b2\13\31\3\32\3\32\5\32\u01b6\n\32\3\32\3\32\3")
 */
  __pyx_tuple__78 = PyTuple_Pack(1, __pyx_kp_u__77); if (unlikely(!__pyx_tuple__78)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__78);
  __Pyx_GIVEREF(__pyx_tuple__78);

  /* "pymola/generated/ModelicaParser.pyx":51
 *         buf.write("\25\u0198\n\25\3\26\3\26\3\26\5\26\u019d\n\26\3\26\3\26")
 *         buf.write("\3\27\5\27\u01a2\n\27\3\27\5\27\u01a5\n\27\3\27\5\27\u01a8")
 *         buf.write("\n\27\3\30\3\30\3\31\3\31\3\31\7\31\u01af\n\31\f\31\16")             # <<<<<<<<<<<<<<
 *         buf.write("\31\u01b2\13\31\3\32\3\32\5\32\u01b6\n\32\3\32\3\32\3")
 *         buf.write("\33\3\33\3\33\3\34\3\34\5\34\u01bf\n\34\3\34\5\34\u01c2")
 */
  __pyx_tuple__80 = PyTuple_Pack(1, __pyx_kp_u__79); if (unlikely(!__pyx_tuple__80)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__80);
  __Pyx_GIVEREF(__pyx_tuple__80);

  /* "pymola/generated/ModelicaParser.pyx":52
 *         buf.write("\3\27\5\27\u01a2\n\27\3\27\5\27\u01a5\n\27\3\27\5\27\u01a8")
 *         buf.write("\n\27\3\30\3\30\3\31\3\31\3\31\7\31\u01af\n\31\f\31\16")
 *         buf.write("\31\u01b2\13\31\3\32\3\32\5\32\u01b6\n\32\3\32\3\32\3")             # <<<<<<<<<<<<<<
 *         buf.write("\33\3\33\3\33\3\34\3\34\5\34\u01bf\n\34\3\34\5\34\u01c2")
 *         buf.write("\n\34\3\35\3\35\3\35\5\35\u01c7\n\35\3\35\3\35\3\35\3")
 */
  __pyx_tuple__82 = PyTuple_Pack(1, __pyx_kp_u__81); if (unlikely(!__pyx_tuple__82)) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__82);
  __Pyx_GIVEREF(__pyx_tuple__82);

  /* "pymola/generated/ModelicaParser.pyx":53
 *         buf.write("\n\27\3\30\3\30\3\31\3\31\3\31\7\31\u01af\n\31\f\31\16")
 *         buf.write("\31\u01b2\13\31\3\32\3\32\5\32\u01b6\n\32\3\32\3\32\3")
 *         buf.write("\33\3\33\3\33\3\34\3\34\5\34\u01bf\n\34\3\34\5\34\u01c2")             # <<<<<<<<<<<<<<
 *         buf.write("\n\34\3\35\3\35\3\35\5\35\u01c7\n\35\3\35\3\35\3\35\3")
 *         buf.write("\35\5\35\u01cd\n\35\3\36\3\36\5\36\u01d1\n\36\3\36\3\36")
 */
  __pyx_tuple__84 = PyTuple_Pack(1, __pyx_kp_u__83); if (unlikely(!__pyx_tuple__84)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__84);
  __Pyx_GIVEREF(__pyx_tuple__84);

  /* "pymola/generated/ModelicaParser.pyx":54
 *         buf.write("\31\u01b2\13\31\3\32\3\32\5\32\u01b6\n\32\3\32\3\32\3")
 *         buf.write("\33\3\33\3\33\3\34\3\34\5\34\u01bf\n\34\3\34\5\34\u01c2")
 *         buf.write("\n\34\3\35\3\35\3\35\5\35\u01c7\n\35\3\35\3\35\3\35\3")             # <<<<<<<<<<<<<<
 *         buf.write("\35\5\35\u01cd\n\35\3\36\3\36\5\36\u01d1\n\36\3\36\3\36")
 *         buf.write("\3\37\3\37\3\37\7\37\u01d8\n\37\f\37\16\37\u01db\13\37")
 */
  __pyx_tuple__86 = PyTuple_Pack(1, __pyx_kp_u__85); if (unlikely(!__pyx_tuple__86)) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__86);
  __Pyx_GIVEREF(__pyx_tuple__86);

  /* "pymola/generated/ModelicaParser.pyx":55
 *         buf.write("\33\3\33\3\33\3\34\3\34\5\34\u01bf\n\34\3\34\5\34\u01c2")
 *         buf.write("\n\34\3\35\3\35\3\35\5\35\u01c7\n\35\3\35\3\35\3\35\3")
 *         buf.write("\35\5\35\u01cd\n\35\3\36\3\36\5\36\u01d1\n\36\3\36\3\36")             # <<<<<<<<<<<<<<
 *         buf.write("\3\37\3\37\3\37\7\37\u01d8\n\37\f\37\16\37\u01db\13\37")
 *         buf.write("\3 \3 \5 \u01df\n \3!\5!\u01e2\n!\3!\5!\u01e5\n!\3!\3")
 */
  __pyx_tuple__88 = PyTuple_Pack(1, __pyx_kp_u__87); if (unlikely(!__pyx_tuple__88)) __PYX_ERR(0, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__88);
  __Pyx_GIVEREF(__pyx_tuple__88);

  /* "pymola/generated/ModelicaParser.pyx":56
 *         buf.write("\n\34\3\35\3\35\3\35\5\35\u01c7\n\35\3\35\3\35\3\35\3")
 *         buf.write("\35\5\35\u01cd\n\35\3\36\3\36\5\36\u01d1\n\36\3\36\3\36")
 *         buf.write("\3\37\3\37\3\37\7\37\u01d8\n\37\f\37\16\37\u01db\13\37")             # <<<<<<<<<<<<<<
 *         buf.write("\3 \3 \5 \u01df\n \3!\5!\u01e2\n!\3!\5!\u01e5\n!\3!\3")
 *         buf.write("!\5!\u01e9\n!\3\"\3\"\5\"\u01ed\n\"\3\"\3\"\3#\3#\5#\u01f3")
 */
  __pyx_tuple__90 = PyTuple_Pack(1, __pyx_kp_u__89); if (unlikely(!__pyx_tuple__90)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__90);
  __Pyx_GIVEREF(__pyx_tuple__90);

  /* "pymola/generated/ModelicaParser.pyx":57
 *         buf.write("\35\5\35\u01cd\n\35\3\36\3\36\5\36\u01d1\n\36\3\36\3\36")
 *         buf.write("\3\37\3\37\3\37\7\37\u01d8\n\37\f\37\16\37\u01db\13\37")
 *         buf.write("\3 \3 \5 \u01df\n \3!\5!\u01e2\n!\3!\5!\u01e5\n!\3!\3")             # <<<<<<<<<<<<<<
 *         buf.write("!\5!\u01e9\n!\3\"\3\"\5\"\u01ed\n\"\3\"\3\"\3#\3#\5#\u01f3")
 *         buf.write("\n#\3#\5#\u01f6\n#\3#\3#\5#\u01fa\n#\3#\5#\u01fd\n#\3")
 */
  __pyx_tuple__92 = PyTuple_Pack(1, __pyx_kp_u__91); if (unlikely(!__pyx_tuple__92)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__92);
  __Pyx_GIVEREF(__pyx_tuple__92);

  /* "pymola/generated/ModelicaParser.pyx":58
 *         buf.write("\3\37\3\37\3\37\7\37\u01d8\n\37\f\37\16\37\u01db\13\37")
 *         buf.write("\3 \3 \5 \u01df\n \3!\5!\u01e2\n!\3!\5!\u01e5\n!\3!\3")
 *         buf.write("!\5!\u01e9\n!\3\"\3\"\5\"\u01ed\n\"\3\"\3\"\3#\3#\5#\u01f3")             # <<<<<<<<<<<<<<
 *         buf.write("\n#\3#\5#\u01f6\n#\3#\3#\5#\u01fa\n#\3#\5#\u01fd\n#\3")
 *         buf.write("$\3$\3$\5$\u0202\n$\3$\5$\u0205\n$\3%\3%\3%\3%\3&\3&\3")
 */
  __pyx_tuple__94 = PyTuple_Pack(1, __pyx_kp_u__93); if (unlikely(!__pyx_tuple__94)) __PYX_ERR(0, 58, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__94);
  __Pyx_GIVEREF(__pyx_tuple__94);

  /* "pymola/generated/ModelicaParser.pyx":59
 *         buf.write("\3 \3 \5 \u01df\n \3!\5!\u01e2\n!\3!\5!\u01e5\n!\3!\3")
 *         buf.write("!\5!\u01e9\n!\3\"\3\"\5\"\u01ed\n\"\3\"\3\"\3#\3#\5#\u01f3")
 *         buf.write("\n#\3#\5#\u01f6\n#\3#\3#\5#\u01fa\n#\3#\5#\u01fd\n#\3")             # <<<<<<<<<<<<<<
 *         buf.write("$\3$\3$\5$\u0202\n$\3$\5$\u0205\n$\3%\3%\3%\3%\3&\3&\3")
 *         buf.write("&\3\'\3\'\3\'\3\'\3\'\3\'\5\'\u0214\n\'\3\'\5\'\u0217")
 */
  __pyx_tuple__96 = PyTuple_Pack(1, __pyx_kp_u__95); if (unlikely(!__pyx_tuple__96)) __PYX_ERR(0, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__96);
  __Pyx_GIVEREF(__pyx_tuple__96);

  /* "pymola/generated/ModelicaParser.pyx":60
 *         buf.write("!\5!\u01e9\n!\3\"\3\"\5\"\u01ed\n\"\3\"\3\"\3#\3#\5#\u01f3")
 *         buf.write("\n#\3#\5#\u01f6\n#\3#\3#\5#\u01fa\n#\3#\5#\u01fd\n#\3")
 *         buf.write("$\3$\3$\5$\u0202\n$\3$\5$\u0205\n$\3%\3%\3%\3%\3&\3&\3")             # <<<<<<<<<<<<<<
 *         buf.write("&\3\'\3\'\3\'\3\'\3\'\3\'\5\'\u0214\n\'\3\'\5\'\u0217")
 *         buf.write("\n\'\3\'\3\'\3\'\3\'\3\'\5\'\u021e\n\'\3\'\5\'\u0221\n")
 */
  __pyx_tuple__98 = PyTuple_Pack(1, __pyx_kp_u__97); if (unlikely(!__pyx_tuple__98)) __PYX_ERR(0, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__98);
  __Pyx_GIVEREF(__pyx_tuple__98);

  /* "pymola/generated/ModelicaParser.pyx":61
 *         buf.write("\n#\3#\5#\u01f6\n#\3#\3#\5#\u01fa\n#\3#\5#\u01fd\n#\3")
 *         buf.write("$\3$\3$\5$\u0202\n$\3$\5$\u0205\n$\3%\3%\3%\3%\3&\3&\3")
 *         buf.write("&\3\'\3\'\3\'\3\'\3\'\3\'\5\'\u0214\n\'\3\'\5\'\u0217")             # <<<<<<<<<<<<<<
 *         buf.write("\n\'\3\'\3\'\3\'\3\'\3\'\5\'\u021e\n\'\3\'\5\'\u0221\n")
 *         buf.write("\'\3\'\3\'\5\'\u0225\n\'\3(\5(\u0228\n(\3(\3(\3(\3(\7")
 */
  __pyx_tuple__100 = PyTuple_Pack(1, __pyx_kp_u__99); if (unlikely(!__pyx_tuple__100)) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__100);
  __Pyx_GIVEREF(__pyx_tuple__100);

  /* "pymola/generated/ModelicaParser.pyx":62
 *         buf.write("$\3$\3$\5$\u0202\n$\3$\5$\u0205\n$\3%\3%\3%\3%\3&\3&\3")
 *         buf.write("&\3\'\3\'\3\'\3\'\3\'\3\'\5\'\u0214\n\'\3\'\5\'\u0217")
 *         buf.write("\n\'\3\'\3\'\3\'\3\'\3\'\5\'\u021e\n\'\3\'\5\'\u0221\n")             # <<<<<<<<<<<<<<
 *         buf.write("\'\3\'\3\'\5\'\u0225\n\'\3(\5(\u0228\n(\3(\3(\3(\3(\7")
 *         buf.write("(\u022e\n(\f(\16(\u0231\13(\3)\5)\u0234\n)\3)\3)\3)\3")
 */
  __pyx_tuple__102 = PyTuple_Pack(1, __pyx_kp_u__101); if (unlikely(!__pyx_tuple__102)) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__102);
  __Pyx_GIVEREF(__pyx_tuple__102);

  /* "pymola/generated/ModelicaParser.pyx":63
 *         buf.write("&\3\'\3\'\3\'\3\'\3\'\3\'\5\'\u0214\n\'\3\'\5\'\u0217")
 *         buf.write("\n\'\3\'\3\'\3\'\3\'\3\'\5\'\u021e\n\'\3\'\5\'\u0221\n")
 *         buf.write("\'\3\'\3\'\5\'\u0225\n\'\3(\5(\u0228\n(\3(\3(\3(\3(\7")             # <<<<<<<<<<<<<<
 *         buf.write("(\u022e\n(\f(\16(\u0231\13(\3)\5)\u0234\n)\3)\3)\3)\3")
 *         buf.write(")\7)\u023a\n)\f)\16)\u023d\13)\3*\3*\3*\3*\3*\3*\3*\3")
 */
  __pyx_tuple__104 = PyTuple_Pack(1, __pyx_kp_u__103); if (unlikely(!__pyx_tuple__104)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__104);
  __Pyx_GIVEREF(__pyx_tuple__104);

  /* "pymola/generated/ModelicaParser.pyx":64
 *         buf.write("\n\'\3\'\3\'\3\'\3\'\3\'\5\'\u021e\n\'\3\'\5\'\u0221\n")
 *         buf.write("\'\3\'\3\'\5\'\u0225\n\'\3(\5(\u0228\n(\3(\3(\3(\3(\7")
 *         buf.write("(\u022e\n(\f(\16(\u0231\13(\3)\5)\u0234\n)\3)\3)\3)\3")             # <<<<<<<<<<<<<<
 *         buf.write(")\7)\u023a\n)\f)\16)\u023d\13)\3*\3*\3*\3*\3*\3*\3*\3")
 *         buf.write("*\3*\3*\3*\5*\u024a\n*\3+\3+\3+\3,\3,\3,\3,\5,\u0253\n")
 */
  __pyx_tuple__106 = PyTuple_Pack(1, __pyx_kp_u__105); if (unlikely(!__pyx_tuple__106)) __PYX_ERR(0, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__106);
  __Pyx_GIVEREF(__pyx_tuple__106);

  /* "pymola/generated/ModelicaParser.pyx":65
 *         buf.write("\'\3\'\3\'\5\'\u0225\n\'\3(\5(\u0228\n(\3(\3(\3(\3(\7")
 *         buf.write("(\u022e\n(\f(\16(\u0231\13(\3)\5)\u0234\n)\3)\3)\3)\3")
 *         buf.write(")\7)\u023a\n)\f)\16)\u023d\13)\3*\3*\3*\3*\3*\3*\3*\3")             # <<<<<<<<<<<<<<
 *         buf.write("*\3*\3*\3*\5*\u024a\n*\3+\3+\3+\3,\3,\3,\3,\5,\u0253\n")
 *         buf.write(",\3,\3,\3,\3,\7,\u0259\n,\f,\16,\u025c\13,\3,\3,\3,\3")
 */
  __pyx_tuple__108 = PyTuple_Pack(1, __pyx_kp_u__107); if (unlikely(!__pyx_tuple__108)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__108);
  __Pyx_GIVEREF(__pyx_tuple__108);

  /* "pymola/generated/ModelicaParser.pyx":66
 *         buf.write("(\u022e\n(\f(\16(\u0231\13(\3)\5)\u0234\n)\3)\3)\3)\3")
 *         buf.write(")\7)\u023a\n)\f)\16)\u023d\13)\3*\3*\3*\3*\3*\3*\3*\3")
 *         buf.write("*\3*\3*\3*\5*\u024a\n*\3+\3+\3+\3,\3,\3,\3,\5,\u0253\n")             # <<<<<<<<<<<<<<
 *         buf.write(",\3,\3,\3,\3,\7,\u0259\n,\f,\16,\u025c\13,\3,\3,\3,\3")
 *         buf.write(",\3,\3,\3,\3,\3,\3,\3,\5,\u0269\n,\3-\3-\3-\3.\3.\3.\3")
 */
  __pyx_tuple__110 = PyTuple_Pack(1, __pyx_kp_u__109); if (unlikely(!__pyx_tuple__110)) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__110);
  __Pyx_GIVEREF(__pyx_tuple__110);

  /* "pymola/generated/ModelicaParser.pyx":67
 *         buf.write(")\7)\u023a\n)\f)\16)\u023d\13)\3*\3*\3*\3*\3*\3*\3*\3")
 *         buf.write("*\3*\3*\3*\5*\u024a\n*\3+\3+\3+\3,\3,\3,\3,\5,\u0253\n")
 *         buf.write(",\3,\3,\3,\3,\7,\u0259\n,\f,\16,\u025c\13,\3,\3,\3,\3")             # <<<<<<<<<<<<<<
 *         buf.write(",\3,\3,\3,\3,\3,\3,\3,\5,\u0269\n,\3-\3-\3-\3.\3.\3.\3")
 *         buf.write(".\3.\3.\7.\u0274\n.\f.\16.\u0277\13.\3.\3.\3.\3.\3.\3")
 */
  __pyx_tuple__112 = PyTuple_Pack(1, __pyx_kp_u__111); if (unlikely(!__pyx_tuple__112)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__112);
  __Pyx_GIVEREF(__pyx_tuple__112);

  /* "pymola/generated/ModelicaParser.pyx":68
 *         buf.write("*\3*\3*\3*\5*\u024a\n*\3+\3+\3+\3,\3,\3,\3,\5,\u0253\n")
 *         buf.write(",\3,\3,\3,\3,\7,\u0259\n,\f,\16,\u025c\13,\3,\3,\3,\3")
 *         buf.write(",\3,\3,\3,\3,\3,\3,\3,\5,\u0269\n,\3-\3-\3-\3.\3.\3.\3")             # <<<<<<<<<<<<<<
 *         buf.write(".\3.\3.\7.\u0274\n.\f.\16.\u0277\13.\3.\3.\3.\3.\3.\3")
 *         buf.write(".\7.\u027f\n.\f.\16.\u0282\13.\7.\u0284\n.\f.\16.\u0287")
 */
  __pyx_tuple__114 = PyTuple_Pack(1, __pyx_kp_u__113); if (unlikely(!__pyx_tuple__114)) __PYX_ERR(0, 68, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__114);
  __Pyx_GIVEREF(__pyx_tuple__114);

  /* "pymola/generated/ModelicaParser.pyx":69
 *         buf.write(",\3,\3,\3,\3,\7,\u0259\n,\f,\16,\u025c\13,\3,\3,\3,\3")
 *         buf.write(",\3,\3,\3,\3,\3,\3,\3,\5,\u0269\n,\3-\3-\3-\3.\3.\3.\3")
 *         buf.write(".\3.\3.\7.\u0274\n.\f.\16.\u0277\13.\3.\3.\3.\3.\3.\3")             # <<<<<<<<<<<<<<
 *         buf.write(".\7.\u027f\n.\f.\16.\u0282\13.\7.\u0284\n.\f.\16.\u0287")
 *         buf.write("\13.\3.\3.\3.\3.\7.\u028d\n.\f.\16.\u0290\13.\5.\u0292")
 */
  __pyx_tuple__116 = PyTuple_Pack(1, __pyx_kp_u__115); if (unlikely(!__pyx_tuple__116)) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__116);
  __Pyx_GIVEREF(__pyx_tuple__116);

  /* "pymola/generated/ModelicaParser.pyx":70
 *         buf.write(",\3,\3,\3,\3,\3,\3,\3,\5,\u0269\n,\3-\3-\3-\3.\3.\3.\3")
 *         buf.write(".\3.\3.\7.\u0274\n.\f.\16.\u0277\13.\3.\3.\3.\3.\3.\3")
 *         buf.write(".\7.\u027f\n.\f.\16.\u0282\13.\7.\u0284\n.\f.\16.\u0287")             # <<<<<<<<<<<<<<
 *         buf.write("\13.\3.\3.\3.\3.\7.\u028d\n.\f.\16.\u0290\13.\5.\u0292")
 *         buf.write("\n.\3.\3.\3.\3/\3/\3/\3/\3/\3/\7/\u029d\n/\f/\16/\u02a0")
 */
  __pyx_tuple__118 = PyTuple_Pack(1, __pyx_kp_u__117); if (unlikely(!__pyx_tuple__118)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__118);
  __Pyx_GIVEREF(__pyx_tuple__118);

  /* "pymola/generated/ModelicaParser.pyx":71
 *         buf.write(".\3.\3.\7.\u0274\n.\f.\16.\u0277\13.\3.\3.\3.\3.\3.\3")
 *         buf.write(".\7.\u027f\n.\f.\16.\u0282\13.\7.\u0284\n.\f.\16.\u0287")
 *         buf.write("\13.\3.\3.\3.\3.\7.\u028d\n.\f.\16.\u0290\13.\5.\u0292")             # <<<<<<<<<<<<<<
 *         buf.write("\n.\3.\3.\3.\3/\3/\3/\3/\3/\3/\7/\u029d\n/\f/\16/\u02a0")
 *         buf.write("\13/\3/\3/\3/\3/\3/\3/\7/\u02a8\n/\f/\16/\u02ab\13/\7")
 */
  __pyx_tuple__120 = PyTuple_Pack(1, __pyx_kp_u__119); if (unlikely(!__pyx_tuple__120)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__120);
  __Pyx_GIVEREF(__pyx_tuple__120);

  /* "pymola/generated/ModelicaParser.pyx":72
 *         buf.write(".\7.\u027f\n.\f.\16.\u0282\13.\7.\u0284\n.\f.\16.\u0287")
 *         buf.write("\13.\3.\3.\3.\3.\7.\u028d\n.\f.\16.\u0290\13.\5.\u0292")
 *         buf.write("\n.\3.\3.\3.\3/\3/\3/\3/\3/\3/\7/\u029d\n/\f/\16/\u02a0")             # <<<<<<<<<<<<<<
 *         buf.write("\13/\3/\3/\3/\3/\3/\3/\7/\u02a8\n/\f/\16/\u02ab\13/\7")
 *         buf.write("/\u02ad\n/\f/\16/\u02b0\13/\3/\3/\3/\3/\7/\u02b6\n/\f")
 */
  __pyx_tuple__122 = PyTuple_Pack(1, __pyx_kp_u__121); if (unlikely(!__pyx_tuple__122)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__122);
  __Pyx_GIVEREF(__pyx_tuple__122);

  /* "pymola/generated/ModelicaParser.pyx":73
 *         buf.write("\13.\3.\3.\3.\3.\7.\u028d\n.\f.\16.\u0290\13.\5.\u0292")
 *         buf.write("\n.\3.\3.\3.\3/\3/\3/\3/\3/\3/\7/\u029d\n/\f/\16/\u02a0")
 *         buf.write("\13/\3/\3/\3/\3/\3/\3/\7/\u02a8\n/\f/\16/\u02ab\13/\7")             # <<<<<<<<<<<<<<
 *         buf.write("/\u02ad\n/\f/\16/\u02b0\13/\3/\3/\3/\3/\7/\u02b6\n/\f")
 *         buf.write("/\16/\u02b9\13/\5/\u02bb\n/\3/\3/\3/\3\60\3\60\3\60\3")
 */
  __pyx_tuple__124 = PyTuple_Pack(1, __pyx_kp_u__123); if (unlikely(!__pyx_tuple__124)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__124);
  __Pyx_GIVEREF(__pyx_tuple__124);

  /* "pymola/generated/ModelicaParser.pyx":74
 *         buf.write("\n.\3.\3.\3.\3/\3/\3/\3/\3/\3/\7/\u029d\n/\f/\16/\u02a0")
 *         buf.write("\13/\3/\3/\3/\3/\3/\3/\7/\u02a8\n/\f/\16/\u02ab\13/\7")
 *         buf.write("/\u02ad\n/\f/\16/\u02b0\13/\3/\3/\3/\3/\7/\u02b6\n/\f")             # <<<<<<<<<<<<<<
 *         buf.write("/\16/\u02b9\13/\5/\u02bb\n/\3/\3/\3/\3\60\3\60\3\60\3")
 *         buf.write("\60\3\60\3\60\7\60\u02c6\n\60\f\60\16\60\u02c9\13\60\3")
 */
  __pyx_tuple__126 = PyTuple_Pack(1, __pyx_kp_u__125); if (unlikely(!__pyx_tuple__126)) __PYX_ERR(0, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__126);
  __Pyx_GIVEREF(__pyx_tuple__126);

  /* "pymola/generated/ModelicaParser.pyx":75
 *         buf.write("\13/\3/\3/\3/\3/\3/\3/\7/\u02a8\n/\f/\16/\u02ab\13/\7")
 *         buf.write("/\u02ad\n/\f/\16/\u02b0\13/\3/\3/\3/\3/\7/\u02b6\n/\f")
 *         buf.write("/\16/\u02b9\13/\5/\u02bb\n/\3/\3/\3/\3\60\3\60\3\60\3")             # <<<<<<<<<<<<<<
 *         buf.write("\60\3\60\3\60\7\60\u02c6\n\60\f\60\16\60\u02c9\13\60\3")
 *         buf.write("\60\3\60\3\60\3\61\3\61\3\61\3\61\3\61\3\61\7\61\u02d4")
 */
  __pyx_tuple__127 = PyTuple_Pack(1, __pyx_kp_u_0_0_0); if (unlikely(!__pyx_tuple__127)) __PYX_ERR(0, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__127);
  __Pyx_GIVEREF(__pyx_tuple__127);

  /* "pymola/generated/ModelicaParser.pyx":76
 *         buf.write("/\u02ad\n/\f/\16/\u02b0\13/\3/\3/\3/\3/\7/\u02b6\n/\f")
 *         buf.write("/\16/\u02b9\13/\5/\u02bb\n/\3/\3/\3/\3\60\3\60\3\60\3")
 *         buf.write("\60\3\60\3\60\7\60\u02c6\n\60\f\60\16\60\u02c9\13\60\3")             # <<<<<<<<<<<<<<
 *         buf.write("\60\3\60\3\60\3\61\3\61\3\61\3\61\3\61\3\61\7\61\u02d4")
 *         buf.write("\n\61\f\61\16\61\u02d7\13\61\3\61\3\61\3\61\3\62\3\62")
 */
  __pyx_tuple__128 = PyTuple_Pack(1, __pyx_kp_u_0_0_0_0_0_0_0_0); if (unlikely(!__pyx_tuple__128)) __PYX_ERR(0, 76, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__128);
  __Pyx_GIVEREF(__pyx_tuple__128);

  /* "pymola/generated/ModelicaParser.pyx":77
 *         buf.write("/\16/\u02b9\13/\5/\u02bb\n/\3/\3/\3/\3\60\3\60\3\60\3")
 *         buf.write("\60\3\60\3\60\7\60\u02c6\n\60\f\60\16\60\u02c9\13\60\3")
 *         buf.write("\60\3\60\3\60\3\61\3\61\3\61\3\61\3\61\3\61\7\61\u02d4")             # <<<<<<<<<<<<<<
 *         buf.write("\n\61\f\61\16\61\u02d7\13\61\3\61\3\61\3\61\3\62\3\62")
 *         buf.write("\3\62\7\62\u02df\n\62\f\62\16\62\u02e2\13\62\3\63\3\63")
 */
  __pyx_tuple__129 = PyTuple_Pack(1, __pyx_kp_u_0_0_0_1_1_1_1_1_1_1); if (unlikely(!__pyx_tuple__129)) __PYX_ERR(0, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__129);
  __Pyx_GIVEREF(__pyx_tuple__129);

  /* "pymola/generated/ModelicaParser.pyx":78
 *         buf.write("\60\3\60\3\60\7\60\u02c6\n\60\f\60\16\60\u02c9\13\60\3")
 *         buf.write("\60\3\60\3\60\3\61\3\61\3\61\3\61\3\61\3\61\7\61\u02d4")
 *         buf.write("\n\61\f\61\16\61\u02d7\13\61\3\61\3\61\3\61\3\62\3\62")             # <<<<<<<<<<<<<<
 *         buf.write("\3\62\7\62\u02df\n\62\f\62\16\62\u02e2\13\62\3\63\3\63")
 *         buf.write("\3\63\5\63\u02e7\n\63\3\64\3\64\3\64\3\64\3\64\3\64\7")
 */
  __pyx_tuple__130 = PyTuple_Pack(1, __pyx_kp_u_1_1_1_1_1_1_1_2_2); if (unlikely(!__pyx_tuple__130)) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__130);
  __Pyx_GIVEREF(__pyx_tuple__130);

  /* "pymola/generated/ModelicaParser.pyx":79
 *         buf.write("\60\3\60\3\60\3\61\3\61\3\61\3\61\3\61\3\61\7\61\u02d4")
 *         buf.write("\n\61\f\61\16\61\u02d7\13\61\3\61\3\61\3\61\3\62\3\62")
 *         buf.write("\3\62\7\62\u02df\n\62\f\62\16\62\u02e2\13\62\3\63\3\63")             # <<<<<<<<<<<<<<
 *         buf.write("\3\63\5\63\u02e7\n\63\3\64\3\64\3\64\3\64\3\64\3\64\7")
 *         buf.write("\64\u02ef\n\64\f\64\16\64\u02f2\13\64\3\64\3\64\3\64\3")
 */
  __pyx_tuple__131 = PyTuple_Pack(1, __pyx_kp_u_2_2_2_2_2_2_3_3); if (unlikely(!__pyx_tuple__131)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__131);
  __Pyx_GIVEREF(__pyx_tuple__131);

  /* "pymola/generated/ModelicaParser.pyx":80
 *         buf.write("\n\61\f\61\16\61\u02d7\13\61\3\61\3\61\3\61\3\62\3\62")
 *         buf.write("\3\62\7\62\u02df\n\62\f\62\16\62\u02e2\13\62\3\63\3\63")
 *         buf.write("\3\63\5\63\u02e7\n\63\3\64\3\64\3\64\3\64\3\64\3\64\7")             # <<<<<<<<<<<<<<
 *         buf.write("\64\u02ef\n\64\f\64\16\64\u02f2\13\64\3\64\3\64\3\64\3")
 *         buf.write("\65\3\65\3\65\3\65\3\65\3\65\7\65\u02fd\n\65\f\65\16\65")
 */
  __pyx_tuple__132 = PyTuple_Pack(1, __pyx_kp_u_3_3_3_4_4_4_4_4_4); if (unlikely(!__pyx_tuple__132)) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__132);
  __Pyx_GIVEREF(__pyx_tuple__132);

  /* "pymola/generated/ModelicaParser.pyx":81
 *         buf.write("\3\62\7\62\u02df\n\62\f\62\16\62\u02e2\13\62\3\63\3\63")
 *         buf.write("\3\63\5\63\u02e7\n\63\3\64\3\64\3\64\3\64\3\64\3\64\7")
 *         buf.write("\64\u02ef\n\64\f\64\16\64\u02f2\13\64\3\64\3\64\3\64\3")             # <<<<<<<<<<<<<<
 *         buf.write("\65\3\65\3\65\3\65\3\65\3\65\7\65\u02fd\n\65\f\65\16\65")
 *         buf.write("\u0300\13\65\3\65\3\65\3\65\3\65\3\65\3\65\7\65\u0308")
 */
  __pyx_tuple__133 = PyTuple_Pack(1, __pyx_kp_u_4_4_4_4_4_4_4_4); if (unlikely(!__pyx_tuple__133)) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__133);
  __Pyx_GIVEREF(__pyx_tuple__133);

  /* "pymola/generated/ModelicaParser.pyx":82
 *         buf.write("\3\63\5\63\u02e7\n\63\3\64\3\64\3\64\3\64\3\64\3\64\7")
 *         buf.write("\64\u02ef\n\64\f\64\16\64\u02f2\13\64\3\64\3\64\3\64\3")
 *         buf.write("\65\3\65\3\65\3\65\3\65\3\65\7\65\u02fd\n\65\f\65\16\65")             # <<<<<<<<<<<<<<
 *         buf.write("\u0300\13\65\3\65\3\65\3\65\3\65\3\65\3\65\7\65\u0308")
 *         buf.write("\n\65\f\65\16\65\u030b\13\65\7\65\u030d\n\65\f\65\16\65")
 */
  __pyx_tuple__134 = PyTuple_Pack(1, __pyx_kp_u_5_5_5_5_5_5_5_5_5_5); if (unlikely(!__pyx_tuple__134)) __PYX_ERR(0, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__134);
  __Pyx_GIVEREF(__pyx_tuple__134);

  /* "pymola/generated/ModelicaParser.pyx":83
 *         buf.write("\64\u02ef\n\64\f\64\16\64\u02f2\13\64\3\64\3\64\3\64\3")
 *         buf.write("\65\3\65\3\65\3\65\3\65\3\65\7\65\u02fd\n\65\f\65\16\65")
 *         buf.write("\u0300\13\65\3\65\3\65\3\65\3\65\3\65\3\65\7\65\u0308")             # <<<<<<<<<<<<<<
 *         buf.write("\n\65\f\65\16\65\u030b\13\65\7\65\u030d\n\65\f\65\16\65")
 *         buf.write("\u0310\13\65\3\65\3\65\3\65\3\66\3\66\3\66\3\66\3\66\3")
 */
  __pyx_tuple__135 = PyTuple_Pack(1, __pyx_kp_u_5_5_5_5_5_5_5_5); if (unlikely(!__pyx_tuple__135)) __PYX_ERR(0, 83, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__135);
  __Pyx_GIVEREF(__pyx_tuple__135);

  /* "pymola/generated/ModelicaParser.pyx":84
 *         buf.write("\65\3\65\3\65\3\65\3\65\3\65\7\65\u02fd\n\65\f\65\16\65")
 *         buf.write("\u0300\13\65\3\65\3\65\3\65\3\65\3\65\3\65\7\65\u0308")
 *         buf.write("\n\65\f\65\16\65\u030b\13\65\7\65\u030d\n\65\f\65\16\65")             # <<<<<<<<<<<<<<
 *         buf.write("\u0310\13\65\3\65\3\65\3\65\3\66\3\66\3\66\3\66\3\66\3")
 *         buf.write("\66\7\66\u031b\n\66\f\66\16\66\u031e\13\66\3\66\3\66\3")
 */
  __pyx_tuple__136 = PyTuple_Pack(1, __pyx_kp_u_5_5_5_5_5_5_5_5_2); if (unlikely(!__pyx_tuple__136)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__136);
  __Pyx_GIVEREF(__pyx_tuple__136);

  /* "pymola/generated/ModelicaParser.pyx":85
 *         buf.write("\u0300\13\65\3\65\3\65\3\65\3\65\3\65\3\65\7\65\u0308")
 *         buf.write("\n\65\f\65\16\65\u030b\13\65\7\65\u030d\n\65\f\65\16\65")
 *         buf.write("\u0310\13\65\3\65\3\65\3\65\3\66\3\66\3\66\3\66\3\66\3")             # <<<<<<<<<<<<<<
 *         buf.write("\66\7\66\u031b\n\66\f\66\16\66\u031e\13\66\3\66\3\66\3")
 *         buf.write("\66\3\66\3\66\3\66\7\66\u0326\n\66\f\66\16\66\u0329\13")
 */
  __pyx_tuple__137 = PyTuple_Pack(1, __pyx_kp_u_5_5_5_5_6_6_6_6_6); if (unlikely(!__pyx_tuple__137)) __PYX_ERR(0, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__137);
  __Pyx_GIVEREF(__pyx_tuple__137);

  /* "pymola/generated/ModelicaParser.pyx":86
 *         buf.write("\n\65\f\65\16\65\u030b\13\65\7\65\u030d\n\65\f\65\16\65")
 *         buf.write("\u0310\13\65\3\65\3\65\3\65\3\66\3\66\3\66\3\66\3\66\3")
 *         buf.write("\66\7\66\u031b\n\66\f\66\16\66\u031e\13\66\3\66\3\66\3")             # <<<<<<<<<<<<<<
 *         buf.write("\66\3\66\3\66\3\66\7\66\u0326\n\66\f\66\16\66\u0329\13")
 *         buf.write("\66\7\66\u032b\n\66\f\66\16\66\u032e\13\66\3\66\3\66\3")
 */
  __pyx_tuple__138 = PyTuple_Pack(1, __pyx_kp_u_6_6_6_6_6_6_6_6); if (unlikely(!__pyx_tuple__138)) __PYX_ERR(0, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__138);
  __Pyx_GIVEREF(__pyx_tuple__138);

  /* "pymola/generated/ModelicaParser.pyx":87
 *         buf.write("\u0310\13\65\3\65\3\65\3\65\3\66\3\66\3\66\3\66\3\66\3")
 *         buf.write("\66\7\66\u031b\n\66\f\66\16\66\u031e\13\66\3\66\3\66\3")
 *         buf.write("\66\3\66\3\66\3\66\7\66\u0326\n\66\f\66\16\66\u0329\13")             # <<<<<<<<<<<<<<
 *         buf.write("\66\7\66\u032b\n\66\f\66\16\66\u032e\13\66\3\66\3\66\3")
 *         buf.write("\66\3\67\3\67\3\67\3\67\3\67\3\67\3\67\38\38\38\38\38")
 */
  __pyx_tuple__139 = PyTuple_Pack(1, __pyx_kp_u_6_6_6_6_6_6_6_6_2); if (unlikely(!__pyx_tuple__139)) __PYX_ERR(0, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__139);
  __Pyx_GIVEREF(__pyx_tuple__139);

  /* "pymola/generated/ModelicaParser.pyx":88
 *         buf.write("\66\7\66\u031b\n\66\f\66\16\66\u031e\13\66\3\66\3\66\3")
 *         buf.write("\66\3\66\3\66\3\66\7\66\u0326\n\66\f\66\16\66\u0329\13")
 *         buf.write("\66\7\66\u032b\n\66\f\66\16\66\u032e\13\66\3\66\3\66\3")             # <<<<<<<<<<<<<<
 *         buf.write("\66\3\67\3\67\3\67\3\67\3\67\3\67\3\67\38\38\38\38\38")
 *         buf.write("\38\38\38\38\38\78\u0344\n8\f8\168\u0347\138\38\38\38")
 */
  __pyx_tuple__140 = PyTuple_Pack(1, __pyx_kp_u_6_6_6_6_6_6_6_6_3); if (unlikely(!__pyx_tuple__140)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__140);
  __Pyx_GIVEREF(__pyx_tuple__140);

  /* "pymola/generated/ModelicaParser.pyx":89
 *         buf.write("\66\3\66\3\66\3\66\7\66\u0326\n\66\f\66\16\66\u0329\13")
 *         buf.write("\66\7\66\u032b\n\66\f\66\16\66\u032e\13\66\3\66\3\66\3")
 *         buf.write("\66\3\67\3\67\3\67\3\67\3\67\3\67\3\67\38\38\38\38\38")             # <<<<<<<<<<<<<<
 *         buf.write("\38\38\38\38\38\78\u0344\n8\f8\168\u0347\138\38\38\38")
 *         buf.write("\58\u034c\n8\39\39\39\39\39\59\u0353\n9\59\u0355\n9\3")
 */
  __pyx_tuple__141 = PyTuple_Pack(1, __pyx_kp_u_6_7_7_7_7_7_7_7_8_8_8_8_8); if (unlikely(!__pyx_tuple__141)) __PYX_ERR(0, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__141);
  __Pyx_GIVEREF(__pyx_tuple__141);

  /* "pymola/generated/ModelicaParser.pyx":90
 *         buf.write("\66\7\66\u032b\n\66\f\66\16\66\u032e\13\66\3\66\3\66\3")
 *         buf.write("\66\3\67\3\67\3\67\3\67\3\67\3\67\3\67\38\38\38\38\38")
 *         buf.write("\38\38\38\38\38\78\u0344\n8\f8\168\u0347\138\38\38\38")             # <<<<<<<<<<<<<<
 *         buf.write("\58\u034c\n8\39\39\39\39\39\59\u0353\n9\59\u0355\n9\3")
 *         buf.write(":\3:\3:\3:\3:\3:\3:\3:\3:\3:\5:\u0361\n:\3:\3:\3:\3:\3")
 */
  __pyx_tuple__142 = PyTuple_Pack(1, __pyx_kp_u_8_8_8_8_8_8_8_8_8_8_8_8_8); if (unlikely(!__pyx_tuple__142)) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__142);
  __Pyx_GIVEREF(__pyx_tuple__142);

  /* "pymola/generated/ModelicaParser.pyx":91
 *         buf.write("\66\3\67\3\67\3\67\3\67\3\67\3\67\3\67\38\38\38\38\38")
 *         buf.write("\38\38\38\38\38\78\u0344\n8\f8\168\u0347\138\38\38\38")
 *         buf.write("\58\u034c\n8\39\39\39\39\39\59\u0353\n9\59\u0355\n9\3")             # <<<<<<<<<<<<<<
 *         buf.write(":\3:\3:\3:\3:\3:\3:\3:\3:\3:\5:\u0361\n:\3:\3:\3:\3:\3")
 *         buf.write(":\3:\3:\3:\3:\3:\3:\3:\3:\3:\3:\7:\u0372\n:\f:\16:\u0375")
 */
  __pyx_tuple__143 = PyTuple_Pack(1, __pyx_kp_u_8_8_9_9_9_9_9_9_9_9_9); if (unlikely(!__pyx_tuple__143)) __PYX_ERR(0, 91, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__143);
  __Pyx_GIVEREF(__pyx_tuple__143);

  /* "pymola/generated/ModelicaParser.pyx":92
 *         buf.write("\38\38\38\38\38\78\u0344\n8\f8\168\u0347\138\38\38\38")
 *         buf.write("\58\u034c\n8\39\39\39\39\39\59\u0353\n9\59\u0355\n9\3")
 *         buf.write(":\3:\3:\3:\3:\3:\3:\3:\3:\3:\5:\u0361\n:\3:\3:\3:\3:\3")             # <<<<<<<<<<<<<<
 *         buf.write(":\3:\3:\3:\3:\3:\3:\3:\3:\3:\3:\7:\u0372\n:\f:\16:\u0375")
 *         buf.write("\13:\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3")
 */
  __pyx_tuple__145 = PyTuple_Pack(1, __pyx_kp_u__144); if (unlikely(!__pyx_tuple__145)) __PYX_ERR(0, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__145);
  __Pyx_GIVEREF(__pyx_tuple__145);

  /* "pymola/generated/ModelicaParser.pyx":93
 *         buf.write("\58\u034c\n8\39\39\39\39\39\59\u0353\n9\59\u0355\n9\3")
 *         buf.write(":\3:\3:\3:\3:\3:\3:\3:\3:\3:\5:\u0361\n:\3:\3:\3:\3:\3")
 *         buf.write(":\3:\3:\3:\3:\3:\3:\3:\3:\3:\3:\7:\u0372\n:\f:\16:\u0375")             # <<<<<<<<<<<<<<
 *         buf.write("\13:\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3")
 *         buf.write(";\3;\3;\3;\7;\u038b\n;\f;\16;\u038e\13;\3;\3;\3;\3;\3")
 */
  __pyx_tuple__147 = PyTuple_Pack(1, __pyx_kp_u__146); if (unlikely(!__pyx_tuple__147)) __PYX_ERR(0, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__147);
  __Pyx_GIVEREF(__pyx_tuple__147);

  /* "pymola/generated/ModelicaParser.pyx":94
 *         buf.write(":\3:\3:\3:\3:\3:\3:\3:\3:\3:\5:\u0361\n:\3:\3:\3:\3:\3")
 *         buf.write(":\3:\3:\3:\3:\3:\3:\3:\3:\3:\3:\7:\u0372\n:\f:\16:\u0375")
 *         buf.write("\13:\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3")             # <<<<<<<<<<<<<<
 *         buf.write(";\3;\3;\3;\7;\u038b\n;\f;\16;\u038e\13;\3;\3;\3;\3;\3")
 *         buf.write(";\3;\3;\5;\u0397\n;\3<\5<\u039a\n<\3<\3<\3<\7<\u039f\n")
 */
  __pyx_tuple__149 = PyTuple_Pack(1, __pyx_kp_u__148); if (unlikely(!__pyx_tuple__149)) __PYX_ERR(0, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__149);
  __Pyx_GIVEREF(__pyx_tuple__149);

  /* "pymola/generated/ModelicaParser.pyx":95
 *         buf.write(":\3:\3:\3:\3:\3:\3:\3:\3:\3:\3:\7:\u0372\n:\f:\16:\u0375")
 *         buf.write("\13:\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3")
 *         buf.write(";\3;\3;\3;\7;\u038b\n;\f;\16;\u038e\13;\3;\3;\3;\3;\3")             # <<<<<<<<<<<<<<
 *         buf.write(";\3;\3;\5;\u0397\n;\3<\5<\u039a\n<\3<\3<\3<\7<\u039f\n")
 *         buf.write("<\f<\16<\u03a2\13<\3=\3=\5=\u03a6\n=\3>\3>\3>\7>\u03ab")
 */
  __pyx_tuple__151 = PyTuple_Pack(1, __pyx_kp_u__150); if (unlikely(!__pyx_tuple__151)) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__151);
  __Pyx_GIVEREF(__pyx_tuple__151);

  /* "pymola/generated/ModelicaParser.pyx":96
 *         buf.write("\13:\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3;\3")
 *         buf.write(";\3;\3;\3;\7;\u038b\n;\f;\16;\u038e\13;\3;\3;\3;\3;\3")
 *         buf.write(";\3;\3;\5;\u0397\n;\3<\5<\u039a\n<\3<\3<\3<\7<\u039f\n")             # <<<<<<<<<<<<<<
 *         buf.write("<\f<\16<\u03a2\13<\3=\3=\5=\u03a6\n=\3>\3>\3>\7>\u03ab")
 *         buf.write("\n>\f>\16>\u03ae\13>\3?\3?\5?\u03b2\n?\3?\3?\3@\3@\3@")
 */
  __pyx_tuple__153 = PyTuple_Pack(1, __pyx_kp_u__152); if (unlikely(!__pyx_tuple__153)) __PYX_ERR(0, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__153);
  __Pyx_GIVEREF(__pyx_tuple__153);

  /* "pymola/generated/ModelicaParser.pyx":97
 *         buf.write(";\3;\3;\3;\7;\u038b\n;\f;\16;\u038e\13;\3;\3;\3;\3;\3")
 *         buf.write(";\3;\3;\5;\u0397\n;\3<\5<\u039a\n<\3<\3<\3<\7<\u039f\n")
 *         buf.write("<\f<\16<\u03a2\13<\3=\3=\5=\u03a6\n=\3>\3>\3>\7>\u03ab")             # <<<<<<<<<<<<<<
 *         buf.write("\n>\f>\16>\u03ae\13>\3?\3?\5?\u03b2\n?\3?\3?\3@\3@\3@")
 *         buf.write("\3@\3@\7@\u03bb\n@\f@\16@\u03be\13@\3@\5@\u03c1\n@\3A")
 */
  __pyx_tuple__155 = PyTuple_Pack(1, __pyx_kp_u__154); if (unlikely(!__pyx_tuple__155)) __PYX_ERR(0, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__155);
  __Pyx_GIVEREF(__pyx_tuple__155);

  /* "pymola/generated/ModelicaParser.pyx":98
 *         buf.write(";\3;\3;\5;\u0397\n;\3<\5<\u039a\n<\3<\3<\3<\7<\u039f\n")
 *         buf.write("<\f<\16<\u03a2\13<\3=\3=\5=\u03a6\n=\3>\3>\3>\7>\u03ab")
 *         buf.write("\n>\f>\16>\u03ae\13>\3?\3?\5?\u03b2\n?\3?\3?\3@\3@\3@")             # <<<<<<<<<<<<<<
 *         buf.write("\3@\3@\7@\u03bb\n@\f@\16@\u03be\13@\3@\5@\u03c1\n@\3A")
 *         buf.write("\3A\3A\7A\u03c6\nA\fA\16A\u03c9\13A\3B\3B\3B\3B\3C\3C")
 */
  __pyx_tuple__157 = PyTuple_Pack(1, __pyx_kp_u__156); if (unlikely(!__pyx_tuple__157)) __PYX_ERR(0, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__157);
  __Pyx_GIVEREF(__pyx_tuple__157);

  /* "pymola/generated/ModelicaParser.pyx":99
 *         buf.write("<\f<\16<\u03a2\13<\3=\3=\5=\u03a6\n=\3>\3>\3>\7>\u03ab")
 *         buf.write("\n>\f>\16>\u03ae\13>\3?\3?\5?\u03b2\n?\3?\3?\3@\3@\3@")
 *         buf.write("\3@\3@\7@\u03bb\n@\f@\16@\u03be\13@\3@\5@\u03c1\n@\3A")             # <<<<<<<<<<<<<<
 *         buf.write("\3A\3A\7A\u03c6\nA\fA\16A\u03c9\13A\3B\3B\3B\3B\3C\3C")
 *         buf.write("\3C\3C\5C\u03d3\nC\3C\3C\3C\5C\u03d8\nC\3D\5D\u03db\n")
 */
  __pyx_tuple__158 = PyTuple_Pack(1, __pyx_kp_u_A); if (unlikely(!__pyx_tuple__158)) __PYX_ERR(0, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__158);
  __Pyx_GIVEREF(__pyx_tuple__158);

  /* "pymola/generated/ModelicaParser.pyx":100
 *         buf.write("\n>\f>\16>\u03ae\13>\3?\3?\5?\u03b2\n?\3?\3?\3@\3@\3@")
 *         buf.write("\3@\3@\7@\u03bb\n@\f@\16@\u03be\13@\3@\5@\u03c1\n@\3A")
 *         buf.write("\3A\3A\7A\u03c6\nA\fA\16A\u03c9\13A\3B\3B\3B\3B\3C\3C")             # <<<<<<<<<<<<<<
 *         buf.write("\3C\3C\5C\u03d3\nC\3C\3C\3C\5C\u03d8\nC\3D\5D\u03db\n")
 *         buf.write("D\3D\3D\7D\u03df\nD\fD\16D\u03e2\13D\3E\3E\3E\7E\u03e7")
 */
  __pyx_tuple__159 = PyTuple_Pack(1, __pyx_kp_u_A_A_A_A_A_A_A_B_B_B_B_C_C); if (unlikely(!__pyx_tuple__159)) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__159);
  __Pyx_GIVEREF(__pyx_tuple__159);

  /* "pymola/generated/ModelicaParser.pyx":101
 *         buf.write("\3@\3@\7@\u03bb\n@\f@\16@\u03be\13@\3@\5@\u03c1\n@\3A")
 *         buf.write("\3A\3A\7A\u03c6\nA\fA\16A\u03c9\13A\3B\3B\3B\3B\3C\3C")
 *         buf.write("\3C\3C\5C\u03d3\nC\3C\3C\3C\5C\u03d8\nC\3D\5D\u03db\n")             # <<<<<<<<<<<<<<
 *         buf.write("D\3D\3D\7D\u03df\nD\fD\16D\u03e2\13D\3E\3E\3E\7E\u03e7")
 *         buf.write("\nE\fE\16E\u03ea\13E\3F\3F\3F\3F\7F\u03f0\nF\fF\16F\u03f3")
 */
  __pyx_tuple__160 = PyTuple_Pack(1, __pyx_kp_u_C_C_C_C_C_C_C_C_C_D_D); if (unlikely(!__pyx_tuple__160)) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__160);
  __Pyx_GIVEREF(__pyx_tuple__160);

  /* "pymola/generated/ModelicaParser.pyx":102
 *         buf.write("\3A\3A\7A\u03c6\nA\fA\16A\u03c9\13A\3B\3B\3B\3B\3C\3C")
 *         buf.write("\3C\3C\5C\u03d3\nC\3C\3C\3C\5C\u03d8\nC\3D\5D\u03db\n")
 *         buf.write("D\3D\3D\7D\u03df\nD\fD\16D\u03e2\13D\3E\3E\3E\7E\u03e7")             # <<<<<<<<<<<<<<
 *         buf.write("\nE\fE\16E\u03ea\13E\3F\3F\3F\3F\7F\u03f0\nF\fF\16F\u03f3")
 *         buf.write("\13F\3F\3F\3G\3G\5G\u03f9\nG\3H\3H\5H\u03fd\nH\3I\3I\3")
 */
  __pyx_tuple__161 = PyTuple_Pack(1, __pyx_kp_u_D_D_D_D_D_D_D_D_E_E_E_E); if (unlikely(!__pyx_tuple__161)) __PYX_ERR(0, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__161);
  __Pyx_GIVEREF(__pyx_tuple__161);

  /* "pymola/generated/ModelicaParser.pyx":103
 *         buf.write("\3C\3C\5C\u03d3\nC\3C\3C\3C\5C\u03d8\nC\3D\5D\u03db\n")
 *         buf.write("D\3D\3D\7D\u03df\nD\fD\16D\u03e2\13D\3E\3E\3E\7E\u03e7")
 *         buf.write("\nE\fE\16E\u03ea\13E\3F\3F\3F\3F\7F\u03f0\nF\fF\16F\u03f3")             # <<<<<<<<<<<<<<
 *         buf.write("\13F\3F\3F\3G\3G\5G\u03f9\nG\3H\3H\5H\u03fd\nH\3I\3I\3")
 *         buf.write("I\7I\u0402\nI\fI\16I\u0405\13I\5I\u0407\nI\3J\3J\3J\3")
 */
  __pyx_tuple__162 = PyTuple_Pack(1, __pyx_kp_u_E_E_E_E_F_F_F_F_F_F_F_F); if (unlikely(!__pyx_tuple__162)) __PYX_ERR(0, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__162);
  __Pyx_GIVEREF(__pyx_tuple__162);

  /* "pymola/generated/ModelicaParser.pyx":104
 *         buf.write("D\3D\3D\7D\u03df\nD\fD\16D\u03e2\13D\3E\3E\3E\7E\u03e7")
 *         buf.write("\nE\fE\16E\u03ea\13E\3F\3F\3F\3F\7F\u03f0\nF\fF\16F\u03f3")
 *         buf.write("\13F\3F\3F\3G\3G\5G\u03f9\nG\3H\3H\5H\u03fd\nH\3I\3I\3")             # <<<<<<<<<<<<<<
 *         buf.write("I\7I\u0402\nI\fI\16I\u0405\13I\5I\u0407\nI\3J\3J\3J\3")
 *         buf.write("J\2\3rK\2\4\6\b\n\f\16\20\22\24\26\30\32\34\36 \"$&(*")
 */
  __pyx_tuple__163 = PyTuple_Pack(1, __pyx_kp_u_F_F_F_G_G_G_G_H_H_H_H_I_I); if (unlikely(!__pyx_tuple__163)) __PYX_ERR(0, 104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__163);
  __Pyx_GIVEREF(__pyx_tuple__163);

  /* "pymola/generated/ModelicaParser.pyx":105
 *         buf.write("\nE\fE\16E\u03ea\13E\3F\3F\3F\3F\7F\u03f0\nF\fF\16F\u03f3")
 *         buf.write("\13F\3F\3F\3G\3G\5G\u03f9\nG\3H\3H\5H\u03fd\nH\3I\3I\3")
 *         buf.write("I\7I\u0402\nI\fI\16I\u0405\13I\5I\u0407\nI\3J\3J\3J\3")             # <<<<<<<<<<<<<<
 *         buf.write("J\2\3rK\2\4\6\b\n\f\16\20\22\24\26\30\32\34\36 \"$&(*")
 *         buf.write(",.\60\62\64\668:<>@BDFHJLNPRTVXZ\\^`bdfhjlnprtvxz|~\u0080")
 */
  __pyx_tuple__164 = PyTuple_Pack(1, __pyx_kp_u_I_I_I_I_I_I_I_I_J_J_J); if (unlikely(!__pyx_tuple__164)) __PYX_ERR(0, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__164);
  __Pyx_GIVEREF(__pyx_tuple__164);

  /* "pymola/generated/ModelicaParser.pyx":106
 *         buf.write("\13F\3F\3F\3G\3G\5G\u03f9\nG\3H\3H\5H\u03fd\nH\3I\3I\3")
 *         buf.write("I\7I\u0402\nI\fI\16I\u0405\13I\5I\u0407\nI\3J\3J\3J\3")
 *         buf.write("J\2\3rK\2\4\6\b\n\f\16\20\22\24\26\30\32\34\36 \"$&(*")             # <<<<<<<<<<<<<<
 *         buf.write(",.\60\62\64\668:<>@BDFHJLNPRTVXZ\\^`bdfhjlnprtvxz|~\u0080")
 *         buf.write("\u0082\u0084\u0086\u0088\u008a\u008c\u008e\u0090\u0092")
 */
  __pyx_tuple__165 = PyTuple_Pack(1, __pyx_kp_u_J_rK); if (unlikely(!__pyx_tuple__165)) __PYX_ERR(0, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__165);
  __Pyx_GIVEREF(__pyx_tuple__165);

  /* "pymola/generated/ModelicaParser.pyx":107
 *         buf.write("I\7I\u0402\nI\fI\16I\u0405\13I\5I\u0407\nI\3J\3J\3J\3")
 *         buf.write("J\2\3rK\2\4\6\b\n\f\16\20\22\24\26\30\32\34\36 \"$&(*")
 *         buf.write(",.\60\62\64\668:<>@BDFHJLNPRTVXZ\\^`bdfhjlnprtvxz|~\u0080")             # <<<<<<<<<<<<<<
 *         buf.write("\u0082\u0084\u0086\u0088\u008a\u008c\u008e\u0090\u0092")
 *         buf.write("\2\n\3\2\r\16\3\2#$\3\2%\'\3\2()\3\2;<\4\2\37\37=?\4\2")
 */
  __pyx_tuple__166 = PyTuple_Pack(1, __pyx_kp_u_02468_BDFHJLNPRTVXZ_bdfhjlnprtv); if (unlikely(!__pyx_tuple__166)) __PYX_ERR(0, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__166);
  __Pyx_GIVEREF(__pyx_tuple__166);

  /* "pymola/generated/ModelicaParser.pyx":108
 *         buf.write("J\2\3rK\2\4\6\b\n\f\16\20\22\24\26\30\32\34\36 \"$&(*")
 *         buf.write(",.\60\62\64\668:<>@BDFHJLNPRTVXZ\\^`bdfhjlnprtvxz|~\u0080")
 *         buf.write("\u0082\u0084\u0086\u0088\u008a\u008c\u008e\u0090\u0092")             # <<<<<<<<<<<<<<
 *         buf.write("\2\n\3\2\r\16\3\2#$\3\2%\'\3\2()\3\2;<\4\2\37\37=?\4\2")
 *         buf.write("::@B\3\2CH\2\u046f\2\u0099\3\2\2\2\4\u00a2\3\2\2\2\6\u00a8")
 */
  __pyx_tuple__168 = PyTuple_Pack(1, __pyx_kp_u__167); if (unlikely(!__pyx_tuple__168)) __PYX_ERR(0, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__168);
  __Pyx_GIVEREF(__pyx_tuple__168);

  /* "pymola/generated/ModelicaParser.pyx":109
 *         buf.write(",.\60\62\64\668:<>@BDFHJLNPRTVXZ\\^`bdfhjlnprtvxz|~\u0080")
 *         buf.write("\u0082\u0084\u0086\u0088\u008a\u008c\u008e\u0090\u0092")
 *         buf.write("\2\n\3\2\r\16\3\2#$\3\2%\'\3\2()\3\2;<\4\2\37\37=?\4\2")             # <<<<<<<<<<<<<<
 *         buf.write("::@B\3\2CH\2\u046f\2\u0099\3\2\2\2\4\u00a2\3\2\2\2\6\u00a8")
 *         buf.write("\3\2\2\2\b\u00ae\3\2\2\2\n\u00c7\3\2\2\2\f\u0102\3\2\2")
 */
  __pyx_tuple__170 = PyTuple_Pack(1, __pyx_kp_u__169); if (unlikely(!__pyx_tuple__170)) __PYX_ERR(0, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__170);
  __Pyx_GIVEREF(__pyx_tuple__170);

  /* "pymola/generated/ModelicaParser.pyx":110
 *         buf.write("\u0082\u0084\u0086\u0088\u008a\u008c\u008e\u0090\u0092")
 *         buf.write("\2\n\3\2\r\16\3\2#$\3\2%\'\3\2()\3\2;<\4\2\37\37=?\4\2")
 *         buf.write("::@B\3\2CH\2\u046f\2\u0099\3\2\2\2\4\u00a2\3\2\2\2\6\u00a8")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\b\u00ae\3\2\2\2\n\u00c7\3\2\2\2\f\u0102\3\2\2")
 *         buf.write("\2\16\u0104\3\2\2\2\20\u0106\3\2\2\2\22\u010e\3\2\2\2")
 */
  __pyx_tuple__171 = PyTuple_Pack(1, __pyx_kp_u_B_CH); if (unlikely(!__pyx_tuple__171)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__171);
  __Pyx_GIVEREF(__pyx_tuple__171);

  /* "pymola/generated/ModelicaParser.pyx":111
 *         buf.write("\2\n\3\2\r\16\3\2#$\3\2%\'\3\2()\3\2;<\4\2\37\37=?\4\2")
 *         buf.write("::@B\3\2CH\2\u046f\2\u0099\3\2\2\2\4\u00a2\3\2\2\2\6\u00a8")
 *         buf.write("\3\2\2\2\b\u00ae\3\2\2\2\n\u00c7\3\2\2\2\f\u0102\3\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\16\u0104\3\2\2\2\20\u0106\3\2\2\2\22\u010e\3\2\2\2")
 *         buf.write("\24\u0111\3\2\2\2\26\u012f\3\2\2\2\30\u0134\3\2\2\2\32")
 */
  __pyx_tuple__173 = PyTuple_Pack(1, __pyx_kp_u__172); if (unlikely(!__pyx_tuple__173)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__173);
  __Pyx_GIVEREF(__pyx_tuple__173);

  /* "pymola/generated/ModelicaParser.pyx":112
 *         buf.write("::@B\3\2CH\2\u046f\2\u0099\3\2\2\2\4\u00a2\3\2\2\2\6\u00a8")
 *         buf.write("\3\2\2\2\b\u00ae\3\2\2\2\n\u00c7\3\2\2\2\f\u0102\3\2\2")
 *         buf.write("\2\16\u0104\3\2\2\2\20\u0106\3\2\2\2\22\u010e\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\24\u0111\3\2\2\2\26\u012f\3\2\2\2\30\u0134\3\2\2\2\32")
 *         buf.write("\u0142\3\2\2\2\34\u0149\3\2\2\2\36\u014c\3\2\2\2 \u015c")
 */
  __pyx_tuple__175 = PyTuple_Pack(1, __pyx_kp_u__174); if (unlikely(!__pyx_tuple__175)) __PYX_ERR(0, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__175);
  __Pyx_GIVEREF(__pyx_tuple__175);

  /* "pymola/generated/ModelicaParser.pyx":113
 *         buf.write("\3\2\2\2\b\u00ae\3\2\2\2\n\u00c7\3\2\2\2\f\u0102\3\2\2")
 *         buf.write("\2\16\u0104\3\2\2\2\20\u0106\3\2\2\2\22\u010e\3\2\2\2")
 *         buf.write("\24\u0111\3\2\2\2\26\u012f\3\2\2\2\30\u0134\3\2\2\2\32")             # <<<<<<<<<<<<<<
 *         buf.write("\u0142\3\2\2\2\34\u0149\3\2\2\2\36\u014c\3\2\2\2 \u015c")
 *         buf.write("\3\2\2\2\"\u0171\3\2\2\2$\u0184\3\2\2\2&\u018c\3\2\2\2")
 */
  __pyx_tuple__177 = PyTuple_Pack(1, __pyx_kp_u__176); if (unlikely(!__pyx_tuple__177)) __PYX_ERR(0, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__177);
  __Pyx_GIVEREF(__pyx_tuple__177);

  /* "pymola/generated/ModelicaParser.pyx":114
 *         buf.write("\2\16\u0104\3\2\2\2\20\u0106\3\2\2\2\22\u010e\3\2\2\2")
 *         buf.write("\24\u0111\3\2\2\2\26\u012f\3\2\2\2\30\u0134\3\2\2\2\32")
 *         buf.write("\u0142\3\2\2\2\34\u0149\3\2\2\2\36\u014c\3\2\2\2 \u015c")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\"\u0171\3\2\2\2$\u0184\3\2\2\2&\u018c\3\2\2\2")
 *         buf.write("(\u0194\3\2\2\2*\u0199\3\2\2\2,\u01a1\3\2\2\2.\u01a9\3")
 */
  __pyx_tuple__179 = PyTuple_Pack(1, __pyx_kp_u__178); if (unlikely(!__pyx_tuple__179)) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__179);
  __Pyx_GIVEREF(__pyx_tuple__179);

  /* "pymola/generated/ModelicaParser.pyx":115
 *         buf.write("\24\u0111\3\2\2\2\26\u012f\3\2\2\2\30\u0134\3\2\2\2\32")
 *         buf.write("\u0142\3\2\2\2\34\u0149\3\2\2\2\36\u014c\3\2\2\2 \u015c")
 *         buf.write("\3\2\2\2\"\u0171\3\2\2\2$\u0184\3\2\2\2&\u018c\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("(\u0194\3\2\2\2*\u0199\3\2\2\2,\u01a1\3\2\2\2.\u01a9\3")
 *         buf.write("\2\2\2\60\u01ab\3\2\2\2\62\u01b3\3\2\2\2\64\u01b9\3\2")
 */
  __pyx_tuple__181 = PyTuple_Pack(1, __pyx_kp_u__180); if (unlikely(!__pyx_tuple__181)) __PYX_ERR(0, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__181);
  __Pyx_GIVEREF(__pyx_tuple__181);

  /* "pymola/generated/ModelicaParser.pyx":116
 *         buf.write("\u0142\3\2\2\2\34\u0149\3\2\2\2\36\u014c\3\2\2\2 \u015c")
 *         buf.write("\3\2\2\2\"\u0171\3\2\2\2$\u0184\3\2\2\2&\u018c\3\2\2\2")
 *         buf.write("(\u0194\3\2\2\2*\u0199\3\2\2\2,\u01a1\3\2\2\2.\u01a9\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\60\u01ab\3\2\2\2\62\u01b3\3\2\2\2\64\u01b9\3\2")
 *         buf.write("\2\2\66\u01bc\3\2\2\28\u01cc\3\2\2\2:\u01ce\3\2\2\2<\u01d4")
 */
  __pyx_tuple__183 = PyTuple_Pack(1, __pyx_kp_u__182); if (unlikely(!__pyx_tuple__183)) __PYX_ERR(0, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__183);
  __Pyx_GIVEREF(__pyx_tuple__183);

  /* "pymola/generated/ModelicaParser.pyx":117
 *         buf.write("\3\2\2\2\"\u0171\3\2\2\2$\u0184\3\2\2\2&\u018c\3\2\2\2")
 *         buf.write("(\u0194\3\2\2\2*\u0199\3\2\2\2,\u01a1\3\2\2\2.\u01a9\3")
 *         buf.write("\2\2\2\60\u01ab\3\2\2\2\62\u01b3\3\2\2\2\64\u01b9\3\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\66\u01bc\3\2\2\28\u01cc\3\2\2\2:\u01ce\3\2\2\2<\u01d4")
 *         buf.write("\3\2\2\2>\u01de\3\2\2\2@\u01e1\3\2\2\2B\u01ea\3\2\2\2")
 */
  __pyx_tuple__184 = PyTuple_Pack(1, __pyx_kp_u_0_2_4); if (unlikely(!__pyx_tuple__184)) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__184);
  __Pyx_GIVEREF(__pyx_tuple__184);

  /* "pymola/generated/ModelicaParser.pyx":118
 *         buf.write("(\u0194\3\2\2\2*\u0199\3\2\2\2,\u01a1\3\2\2\2.\u01a9\3")
 *         buf.write("\2\2\2\60\u01ab\3\2\2\2\62\u01b3\3\2\2\2\64\u01b9\3\2")
 *         buf.write("\2\2\66\u01bc\3\2\2\28\u01cc\3\2\2\2:\u01ce\3\2\2\2<\u01d4")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2>\u01de\3\2\2\2@\u01e1\3\2\2\2B\u01ea\3\2\2\2")
 *         buf.write("D\u01f0\3\2\2\2F\u01fe\3\2\2\2H\u0206\3\2\2\2J\u020a\3")
 */
  __pyx_tuple__185 = PyTuple_Pack(1, __pyx_kp_u_6_8); if (unlikely(!__pyx_tuple__185)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__185);
  __Pyx_GIVEREF(__pyx_tuple__185);

  /* "pymola/generated/ModelicaParser.pyx":119
 *         buf.write("\2\2\2\60\u01ab\3\2\2\2\62\u01b3\3\2\2\2\64\u01b9\3\2")
 *         buf.write("\2\2\66\u01bc\3\2\2\28\u01cc\3\2\2\2:\u01ce\3\2\2\2<\u01d4")
 *         buf.write("\3\2\2\2>\u01de\3\2\2\2@\u01e1\3\2\2\2B\u01ea\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("D\u01f0\3\2\2\2F\u01fe\3\2\2\2H\u0206\3\2\2\2J\u020a\3")
 *         buf.write("\2\2\2L\u020d\3\2\2\2N\u0227\3\2\2\2P\u0233\3\2\2\2R\u0249")
 */
  __pyx_tuple__186 = PyTuple_Pack(1, __pyx_kp_u_B); if (unlikely(!__pyx_tuple__186)) __PYX_ERR(0, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__186);
  __Pyx_GIVEREF(__pyx_tuple__186);

  /* "pymola/generated/ModelicaParser.pyx":120
 *         buf.write("\2\2\66\u01bc\3\2\2\28\u01cc\3\2\2\2:\u01ce\3\2\2\2<\u01d4")
 *         buf.write("\3\2\2\2>\u01de\3\2\2\2@\u01e1\3\2\2\2B\u01ea\3\2\2\2")
 *         buf.write("D\u01f0\3\2\2\2F\u01fe\3\2\2\2H\u0206\3\2\2\2J\u020a\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2L\u020d\3\2\2\2N\u0227\3\2\2\2P\u0233\3\2\2\2R\u0249")
 *         buf.write("\3\2\2\2T\u024b\3\2\2\2V\u0268\3\2\2\2X\u026a\3\2\2\2")
 */
  __pyx_tuple__187 = PyTuple_Pack(1, __pyx_kp_u_D_F_H_J); if (unlikely(!__pyx_tuple__187)) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__187);
  __Pyx_GIVEREF(__pyx_tuple__187);

  /* "pymola/generated/ModelicaParser.pyx":121
 *         buf.write("\3\2\2\2>\u01de\3\2\2\2@\u01e1\3\2\2\2B\u01ea\3\2\2\2")
 *         buf.write("D\u01f0\3\2\2\2F\u01fe\3\2\2\2H\u0206\3\2\2\2J\u020a\3")
 *         buf.write("\2\2\2L\u020d\3\2\2\2N\u0227\3\2\2\2P\u0233\3\2\2\2R\u0249")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2T\u024b\3\2\2\2V\u0268\3\2\2\2X\u026a\3\2\2\2")
 *         buf.write("Z\u026d\3\2\2\2\\\u0296\3\2\2\2^\u02bf\3\2\2\2`\u02cd")
 */
  __pyx_tuple__188 = PyTuple_Pack(1, __pyx_kp_u_L_N_P_R); if (unlikely(!__pyx_tuple__188)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__188);
  __Pyx_GIVEREF(__pyx_tuple__188);

  /* "pymola/generated/ModelicaParser.pyx":122
 *         buf.write("D\u01f0\3\2\2\2F\u01fe\3\2\2\2H\u0206\3\2\2\2J\u020a\3")
 *         buf.write("\2\2\2L\u020d\3\2\2\2N\u0227\3\2\2\2P\u0233\3\2\2\2R\u0249")
 *         buf.write("\3\2\2\2T\u024b\3\2\2\2V\u0268\3\2\2\2X\u026a\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("Z\u026d\3\2\2\2\\\u0296\3\2\2\2^\u02bf\3\2\2\2`\u02cd")
 *         buf.write("\3\2\2\2b\u02db\3\2\2\2d\u02e3\3\2\2\2f\u02e8\3\2\2\2")
 */
  __pyx_tuple__189 = PyTuple_Pack(1, __pyx_kp_u_T_V_X); if (unlikely(!__pyx_tuple__189)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__189);
  __Pyx_GIVEREF(__pyx_tuple__189);

  /* "pymola/generated/ModelicaParser.pyx":123
 *         buf.write("\2\2\2L\u020d\3\2\2\2N\u0227\3\2\2\2P\u0233\3\2\2\2R\u0249")
 *         buf.write("\3\2\2\2T\u024b\3\2\2\2V\u0268\3\2\2\2X\u026a\3\2\2\2")
 *         buf.write("Z\u026d\3\2\2\2\\\u0296\3\2\2\2^\u02bf\3\2\2\2`\u02cd")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2b\u02db\3\2\2\2d\u02e3\3\2\2\2f\u02e8\3\2\2\2")
 *         buf.write("h\u02f6\3\2\2\2j\u0314\3\2\2\2l\u0332\3\2\2\2n\u034b\3")
 */
  __pyx_tuple__190 = PyTuple_Pack(1, __pyx_kp_u_Z); if (unlikely(!__pyx_tuple__190)) __PYX_ERR(0, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__190);
  __Pyx_GIVEREF(__pyx_tuple__190);

  /* "pymola/generated/ModelicaParser.pyx":124
 *         buf.write("\3\2\2\2T\u024b\3\2\2\2V\u0268\3\2\2\2X\u026a\3\2\2\2")
 *         buf.write("Z\u026d\3\2\2\2\\\u0296\3\2\2\2^\u02bf\3\2\2\2`\u02cd")
 *         buf.write("\3\2\2\2b\u02db\3\2\2\2d\u02e3\3\2\2\2f\u02e8\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("h\u02f6\3\2\2\2j\u0314\3\2\2\2l\u0332\3\2\2\2n\u034b\3")
 *         buf.write("\2\2\2p\u034d\3\2\2\2r\u0360\3\2\2\2t\u0396\3\2\2\2v\u0399")
 */
  __pyx_tuple__191 = PyTuple_Pack(1, __pyx_kp_u_b_d_f); if (unlikely(!__pyx_tuple__191)) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__191);
  __Pyx_GIVEREF(__pyx_tuple__191);

  /* "pymola/generated/ModelicaParser.pyx":125
 *         buf.write("Z\u026d\3\2\2\2\\\u0296\3\2\2\2^\u02bf\3\2\2\2`\u02cd")
 *         buf.write("\3\2\2\2b\u02db\3\2\2\2d\u02e3\3\2\2\2f\u02e8\3\2\2\2")
 *         buf.write("h\u02f6\3\2\2\2j\u0314\3\2\2\2l\u0332\3\2\2\2n\u034b\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2p\u034d\3\2\2\2r\u0360\3\2\2\2t\u0396\3\2\2\2v\u0399")
 *         buf.write("\3\2\2\2x\u03a3\3\2\2\2z\u03a7\3\2\2\2|\u03af\3\2\2\2")
 */
  __pyx_tuple__192 = PyTuple_Pack(1, __pyx_kp_u_h_j_l_n); if (unlikely(!__pyx_tuple__192)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__192);
  __Pyx_GIVEREF(__pyx_tuple__192);

  /* "pymola/generated/ModelicaParser.pyx":126
 *         buf.write("\3\2\2\2b\u02db\3\2\2\2d\u02e3\3\2\2\2f\u02e8\3\2\2\2")
 *         buf.write("h\u02f6\3\2\2\2j\u0314\3\2\2\2l\u0332\3\2\2\2n\u034b\3")
 *         buf.write("\2\2\2p\u034d\3\2\2\2r\u0360\3\2\2\2t\u0396\3\2\2\2v\u0399")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2x\u03a3\3\2\2\2z\u03a7\3\2\2\2|\u03af\3\2\2\2")
 *         buf.write("~\u03c0\3\2\2\2\u0080\u03c2\3\2\2\2\u0082\u03ca\3\2\2")
 */
  __pyx_tuple__193 = PyTuple_Pack(1, __pyx_kp_u_p_r_t_v); if (unlikely(!__pyx_tuple__193)) __PYX_ERR(0, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__193);
  __Pyx_GIVEREF(__pyx_tuple__193);

  /* "pymola/generated/ModelicaParser.pyx":127
 *         buf.write("h\u02f6\3\2\2\2j\u0314\3\2\2\2l\u0332\3\2\2\2n\u034b\3")
 *         buf.write("\2\2\2p\u034d\3\2\2\2r\u0360\3\2\2\2t\u0396\3\2\2\2v\u0399")
 *         buf.write("\3\2\2\2x\u03a3\3\2\2\2z\u03a7\3\2\2\2|\u03af\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("~\u03c0\3\2\2\2\u0080\u03c2\3\2\2\2\u0082\u03ca\3\2\2")
 *         buf.write("\2\u0084\u03d7\3\2\2\2\u0086\u03da\3\2\2\2\u0088\u03e3")
 */
  __pyx_tuple__194 = PyTuple_Pack(1, __pyx_kp_u_x_z); if (unlikely(!__pyx_tuple__194)) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__194);
  __Pyx_GIVEREF(__pyx_tuple__194);

  /* "pymola/generated/ModelicaParser.pyx":128
 *         buf.write("\2\2\2p\u034d\3\2\2\2r\u0360\3\2\2\2t\u0396\3\2\2\2v\u0399")
 *         buf.write("\3\2\2\2x\u03a3\3\2\2\2z\u03a7\3\2\2\2|\u03af\3\2\2\2")
 *         buf.write("~\u03c0\3\2\2\2\u0080\u03c2\3\2\2\2\u0082\u03ca\3\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u0084\u03d7\3\2\2\2\u0086\u03da\3\2\2\2\u0088\u03e3")
 *         buf.write("\3\2\2\2\u008a\u03eb\3\2\2\2\u008c\u03f8\3\2\2\2\u008e")
 */
  __pyx_tuple__196 = PyTuple_Pack(1, __pyx_kp_u__195); if (unlikely(!__pyx_tuple__196)) __PYX_ERR(0, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__196);
  __Pyx_GIVEREF(__pyx_tuple__196);

  /* "pymola/generated/ModelicaParser.pyx":129
 *         buf.write("\3\2\2\2x\u03a3\3\2\2\2z\u03a7\3\2\2\2|\u03af\3\2\2\2")
 *         buf.write("~\u03c0\3\2\2\2\u0080\u03c2\3\2\2\2\u0082\u03ca\3\2\2")
 *         buf.write("\2\u0084\u03d7\3\2\2\2\u0086\u03da\3\2\2\2\u0088\u03e3")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u008a\u03eb\3\2\2\2\u008c\u03f8\3\2\2\2\u008e")
 *         buf.write("\u03fa\3\2\2\2\u0090\u0406\3\2\2\2\u0092\u0408\3\2\2\2")
 */
  __pyx_tuple__198 = PyTuple_Pack(1, __pyx_kp_u__197); if (unlikely(!__pyx_tuple__198)) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__198);
  __Pyx_GIVEREF(__pyx_tuple__198);

  /* "pymola/generated/ModelicaParser.pyx":130
 *         buf.write("~\u03c0\3\2\2\2\u0080\u03c2\3\2\2\2\u0082\u03ca\3\2\2")
 *         buf.write("\2\u0084\u03d7\3\2\2\2\u0086\u03da\3\2\2\2\u0088\u03e3")
 *         buf.write("\3\2\2\2\u008a\u03eb\3\2\2\2\u008c\u03f8\3\2\2\2\u008e")             # <<<<<<<<<<<<<<
 *         buf.write("\u03fa\3\2\2\2\u0090\u0406\3\2\2\2\u0092\u0408\3\2\2\2")
 *         buf.write("\u0094\u0096\7T\2\2\u0095\u0097\5z>\2\u0096\u0095\3\2")
 */
  __pyx_tuple__200 = PyTuple_Pack(1, __pyx_kp_u__199); if (unlikely(!__pyx_tuple__200)) __PYX_ERR(0, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__200);
  __Pyx_GIVEREF(__pyx_tuple__200);

  /* "pymola/generated/ModelicaParser.pyx":131
 *         buf.write("\2\u0084\u03d7\3\2\2\2\u0086\u03da\3\2\2\2\u0088\u03e3")
 *         buf.write("\3\2\2\2\u008a\u03eb\3\2\2\2\u008c\u03f8\3\2\2\2\u008e")
 *         buf.write("\u03fa\3\2\2\2\u0090\u0406\3\2\2\2\u0092\u0408\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0094\u0096\7T\2\2\u0095\u0097\5z>\2\u0096\u0095\3\2")
 *         buf.write("\2\2\u0096\u0097\3\2\2\2\u0097\u0098\3\2\2\2\u0098\u009a")
 */
  __pyx_tuple__202 = PyTuple_Pack(1, __pyx_kp_u__201); if (unlikely(!__pyx_tuple__202)) __PYX_ERR(0, 131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__202);
  __Pyx_GIVEREF(__pyx_tuple__202);

  /* "pymola/generated/ModelicaParser.pyx":132
 *         buf.write("\3\2\2\2\u008a\u03eb\3\2\2\2\u008c\u03f8\3\2\2\2\u008e")
 *         buf.write("\u03fa\3\2\2\2\u0090\u0406\3\2\2\2\u0092\u0408\3\2\2\2")
 *         buf.write("\u0094\u0096\7T\2\2\u0095\u0097\5z>\2\u0096\u0095\3\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\u0096\u0097\3\2\2\2\u0097\u0098\3\2\2\2\u0098\u009a")
 *         buf.write("\7\3\2\2\u0099\u0094\3\2\2\2\u0099\u009a\3\2\2\2\u009a")
 */
  __pyx_tuple__203 = PyTuple_Pack(1, __pyx_kp_u_T_z); if (unlikely(!__pyx_tuple__203)) __PYX_ERR(0, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__203);
  __Pyx_GIVEREF(__pyx_tuple__203);

  /* "pymola/generated/ModelicaParser.pyx":133
 *         buf.write("\u03fa\3\2\2\2\u0090\u0406\3\2\2\2\u0092\u0408\3\2\2\2")
 *         buf.write("\u0094\u0096\7T\2\2\u0095\u0097\5z>\2\u0096\u0095\3\2")
 *         buf.write("\2\2\u0096\u0097\3\2\2\2\u0097\u0098\3\2\2\2\u0098\u009a")             # <<<<<<<<<<<<<<
 *         buf.write("\7\3\2\2\u0099\u0094\3\2\2\2\u0099\u009a\3\2\2\2\u009a")
 *         buf.write("\u009e\3\2\2\2\u009b\u009d\5\4\3\2\u009c\u009b\3\2\2\2")
 */
  __pyx_tuple__205 = PyTuple_Pack(1, __pyx_kp_u__204); if (unlikely(!__pyx_tuple__205)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__205);
  __Pyx_GIVEREF(__pyx_tuple__205);

  /* "pymola/generated/ModelicaParser.pyx":134
 *         buf.write("\u0094\u0096\7T\2\2\u0095\u0097\5z>\2\u0096\u0095\3\2")
 *         buf.write("\2\2\u0096\u0097\3\2\2\2\u0097\u0098\3\2\2\2\u0098\u009a")
 *         buf.write("\7\3\2\2\u0099\u0094\3\2\2\2\u0099\u009a\3\2\2\2\u009a")             # <<<<<<<<<<<<<<
 *         buf.write("\u009e\3\2\2\2\u009b\u009d\5\4\3\2\u009c\u009b\3\2\2\2")
 *         buf.write("\u009d\u00a0\3\2\2\2\u009e\u009c\3\2\2\2\u009e\u009f\3")
 */
  __pyx_tuple__207 = PyTuple_Pack(1, __pyx_kp_u__206); if (unlikely(!__pyx_tuple__207)) __PYX_ERR(0, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__207);
  __Pyx_GIVEREF(__pyx_tuple__207);

  /* "pymola/generated/ModelicaParser.pyx":135
 *         buf.write("\2\2\u0096\u0097\3\2\2\2\u0097\u0098\3\2\2\2\u0098\u009a")
 *         buf.write("\7\3\2\2\u0099\u0094\3\2\2\2\u0099\u009a\3\2\2\2\u009a")
 *         buf.write("\u009e\3\2\2\2\u009b\u009d\5\4\3\2\u009c\u009b\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u009d\u00a0\3\2\2\2\u009e\u009c\3\2\2\2\u009e\u009f\3")
 *         buf.write("\2\2\2\u009f\3\3\2\2\2\u00a0\u009e\3\2\2\2\u00a1\u00a3")
 */
  __pyx_tuple__209 = PyTuple_Pack(1, __pyx_kp_u__208); if (unlikely(!__pyx_tuple__209)) __PYX_ERR(0, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__209);
  __Pyx_GIVEREF(__pyx_tuple__209);

  /* "pymola/generated/ModelicaParser.pyx":136
 *         buf.write("\7\3\2\2\u0099\u0094\3\2\2\2\u0099\u009a\3\2\2\2\u009a")
 *         buf.write("\u009e\3\2\2\2\u009b\u009d\5\4\3\2\u009c\u009b\3\2\2\2")
 *         buf.write("\u009d\u00a0\3\2\2\2\u009e\u009c\3\2\2\2\u009e\u009f\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u009f\3\3\2\2\2\u00a0\u009e\3\2\2\2\u00a1\u00a3")
 *         buf.write("\7S\2\2\u00a2\u00a1\3\2\2\2\u00a2\u00a3\3\2\2\2\u00a3")
 */
  __pyx_tuple__211 = PyTuple_Pack(1, __pyx_kp_u__210); if (unlikely(!__pyx_tuple__211)) __PYX_ERR(0, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__211);
  __Pyx_GIVEREF(__pyx_tuple__211);

  /* "pymola/generated/ModelicaParser.pyx":137
 *         buf.write("\u009e\3\2\2\2\u009b\u009d\5\4\3\2\u009c\u009b\3\2\2\2")
 *         buf.write("\u009d\u00a0\3\2\2\2\u009e\u009c\3\2\2\2\u009e\u009f\3")
 *         buf.write("\2\2\2\u009f\3\3\2\2\2\u00a0\u009e\3\2\2\2\u00a1\u00a3")             # <<<<<<<<<<<<<<
 *         buf.write("\7S\2\2\u00a2\u00a1\3\2\2\2\u00a2\u00a3\3\2\2\2\u00a3")
 *         buf.write("\u00a4\3\2\2\2\u00a4\u00a5\5\6\4\2\u00a5\u00a6\7\3\2\2")
 */
  __pyx_tuple__213 = PyTuple_Pack(1, __pyx_kp_u__212); if (unlikely(!__pyx_tuple__213)) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__213);
  __Pyx_GIVEREF(__pyx_tuple__213);

  /* "pymola/generated/ModelicaParser.pyx":138
 *         buf.write("\u009d\u00a0\3\2\2\2\u009e\u009c\3\2\2\2\u009e\u009f\3")
 *         buf.write("\2\2\2\u009f\3\3\2\2\2\u00a0\u009e\3\2\2\2\u00a1\u00a3")
 *         buf.write("\7S\2\2\u00a2\u00a1\3\2\2\2\u00a2\u00a3\3\2\2\2\u00a3")             # <<<<<<<<<<<<<<
 *         buf.write("\u00a4\3\2\2\2\u00a4\u00a5\5\6\4\2\u00a5\u00a6\7\3\2\2")
 *         buf.write("\u00a6\5\3\2\2\2\u00a7\u00a9\7U\2\2\u00a8\u00a7\3\2\2")
 */
  __pyx_tuple__214 = PyTuple_Pack(1, __pyx_kp_u_S); if (unlikely(!__pyx_tuple__214)) __PYX_ERR(0, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__214);
  __Pyx_GIVEREF(__pyx_tuple__214);

  /* "pymola/generated/ModelicaParser.pyx":139
 *         buf.write("\2\2\2\u009f\3\3\2\2\2\u00a0\u009e\3\2\2\2\u00a1\u00a3")
 *         buf.write("\7S\2\2\u00a2\u00a1\3\2\2\2\u00a2\u00a3\3\2\2\2\u00a3")
 *         buf.write("\u00a4\3\2\2\2\u00a4\u00a5\5\6\4\2\u00a5\u00a6\7\3\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u00a6\5\3\2\2\2\u00a7\u00a9\7U\2\2\u00a8\u00a7\3\2\2")
 *         buf.write("\2\u00a8\u00a9\3\2\2\2\u00a9\u00aa\3\2\2\2\u00aa\u00ab")
 */
  __pyx_tuple__216 = PyTuple_Pack(1, __pyx_kp_u__215); if (unlikely(!__pyx_tuple__216)) __PYX_ERR(0, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__216);
  __Pyx_GIVEREF(__pyx_tuple__216);

  /* "pymola/generated/ModelicaParser.pyx":140
 *         buf.write("\7S\2\2\u00a2\u00a1\3\2\2\2\u00a2\u00a3\3\2\2\2\u00a3")
 *         buf.write("\u00a4\3\2\2\2\u00a4\u00a5\5\6\4\2\u00a5\u00a6\7\3\2\2")
 *         buf.write("\u00a6\5\3\2\2\2\u00a7\u00a9\7U\2\2\u00a8\u00a7\3\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u00a8\u00a9\3\2\2\2\u00a9\u00aa\3\2\2\2\u00aa\u00ab")
 *         buf.write("\5\b\5\2\u00ab\u00ac\5\f\7\2\u00ac\7\3\2\2\2\u00ad\u00af")
 */
  __pyx_tuple__217 = PyTuple_Pack(1, __pyx_kp_u_U); if (unlikely(!__pyx_tuple__217)) __PYX_ERR(0, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__217);
  __Pyx_GIVEREF(__pyx_tuple__217);

  /* "pymola/generated/ModelicaParser.pyx":141
 *         buf.write("\u00a4\3\2\2\2\u00a4\u00a5\5\6\4\2\u00a5\u00a6\7\3\2\2")
 *         buf.write("\u00a6\5\3\2\2\2\u00a7\u00a9\7U\2\2\u00a8\u00a7\3\2\2")
 *         buf.write("\2\u00a8\u00a9\3\2\2\2\u00a9\u00aa\3\2\2\2\u00aa\u00ab")             # <<<<<<<<<<<<<<
 *         buf.write("\5\b\5\2\u00ab\u00ac\5\f\7\2\u00ac\7\3\2\2\2\u00ad\u00af")
 *         buf.write("\7R\2\2\u00ae\u00ad\3\2\2\2\u00ae\u00af\3\2\2\2\u00af")
 */
  __pyx_tuple__219 = PyTuple_Pack(1, __pyx_kp_u__218); if (unlikely(!__pyx_tuple__219)) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__219);
  __Pyx_GIVEREF(__pyx_tuple__219);

  /* "pymola/generated/ModelicaParser.pyx":142
 *         buf.write("\u00a6\5\3\2\2\2\u00a7\u00a9\7U\2\2\u00a8\u00a7\3\2\2")
 *         buf.write("\2\u00a8\u00a9\3\2\2\2\u00a9\u00aa\3\2\2\2\u00aa\u00ab")
 *         buf.write("\5\b\5\2\u00ab\u00ac\5\f\7\2\u00ac\7\3\2\2\2\u00ad\u00af")             # <<<<<<<<<<<<<<
 *         buf.write("\7R\2\2\u00ae\u00ad\3\2\2\2\u00ae\u00af\3\2\2\2\u00af")
 *         buf.write("\u00b0\3\2\2\2\u00b0\u00b1\5\n\6\2\u00b1\t\3\2\2\2\u00b2")
 */
  __pyx_tuple__221 = PyTuple_Pack(1, __pyx_kp_u__220); if (unlikely(!__pyx_tuple__221)) __PYX_ERR(0, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__221);
  __Pyx_GIVEREF(__pyx_tuple__221);

  /* "pymola/generated/ModelicaParser.pyx":143
 *         buf.write("\2\u00a8\u00a9\3\2\2\2\u00a9\u00aa\3\2\2\2\u00aa\u00ab")
 *         buf.write("\5\b\5\2\u00ab\u00ac\5\f\7\2\u00ac\7\3\2\2\2\u00ad\u00af")
 *         buf.write("\7R\2\2\u00ae\u00ad\3\2\2\2\u00ae\u00af\3\2\2\2\u00af")             # <<<<<<<<<<<<<<
 *         buf.write("\u00b0\3\2\2\2\u00b0\u00b1\5\n\6\2\u00b1\t\3\2\2\2\u00b2")
 *         buf.write("\u00c8\7\4\2\2\u00b3\u00c8\7\5\2\2\u00b4\u00b6\7\6\2\2")
 */
  __pyx_tuple__222 = PyTuple_Pack(1, __pyx_kp_u_R); if (unlikely(!__pyx_tuple__222)) __PYX_ERR(0, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__222);
  __Pyx_GIVEREF(__pyx_tuple__222);

  /* "pymola/generated/ModelicaParser.pyx":144
 *         buf.write("\5\b\5\2\u00ab\u00ac\5\f\7\2\u00ac\7\3\2\2\2\u00ad\u00af")
 *         buf.write("\7R\2\2\u00ae\u00ad\3\2\2\2\u00ae\u00af\3\2\2\2\u00af")
 *         buf.write("\u00b0\3\2\2\2\u00b0\u00b1\5\n\6\2\u00b1\t\3\2\2\2\u00b2")             # <<<<<<<<<<<<<<
 *         buf.write("\u00c8\7\4\2\2\u00b3\u00c8\7\5\2\2\u00b4\u00b6\7\6\2\2")
 *         buf.write("\u00b5\u00b4\3\2\2\2\u00b5\u00b6\3\2\2\2\u00b6\u00b7\3")
 */
  __pyx_tuple__224 = PyTuple_Pack(1, __pyx_kp_u__223); if (unlikely(!__pyx_tuple__224)) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__224);
  __Pyx_GIVEREF(__pyx_tuple__224);

  /* "pymola/generated/ModelicaParser.pyx":145
 *         buf.write("\7R\2\2\u00ae\u00ad\3\2\2\2\u00ae\u00af\3\2\2\2\u00af")
 *         buf.write("\u00b0\3\2\2\2\u00b0\u00b1\5\n\6\2\u00b1\t\3\2\2\2\u00b2")
 *         buf.write("\u00c8\7\4\2\2\u00b3\u00c8\7\5\2\2\u00b4\u00b6\7\6\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u00b5\u00b4\3\2\2\2\u00b5\u00b6\3\2\2\2\u00b6\u00b7\3")
 *         buf.write("\2\2\2\u00b7\u00c8\7\7\2\2\u00b8\u00c8\7\b\2\2\u00b9\u00bb")
 */
  __pyx_tuple__226 = PyTuple_Pack(1, __pyx_kp_u__225); if (unlikely(!__pyx_tuple__226)) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__226);
  __Pyx_GIVEREF(__pyx_tuple__226);

  /* "pymola/generated/ModelicaParser.pyx":146
 *         buf.write("\u00b0\3\2\2\2\u00b0\u00b1\5\n\6\2\u00b1\t\3\2\2\2\u00b2")
 *         buf.write("\u00c8\7\4\2\2\u00b3\u00c8\7\5\2\2\u00b4\u00b6\7\6\2\2")
 *         buf.write("\u00b5\u00b4\3\2\2\2\u00b5\u00b6\3\2\2\2\u00b6\u00b7\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u00b7\u00c8\7\7\2\2\u00b8\u00c8\7\b\2\2\u00b9\u00bb")
 *         buf.write("\7\t\2\2\u00ba\u00b9\3\2\2\2\u00ba\u00bb\3\2\2\2\u00bb")
 */
  __pyx_tuple__228 = PyTuple_Pack(1, __pyx_kp_u__227); if (unlikely(!__pyx_tuple__228)) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__228);
  __Pyx_GIVEREF(__pyx_tuple__228);

  /* "pymola/generated/ModelicaParser.pyx":147
 *         buf.write("\u00c8\7\4\2\2\u00b3\u00c8\7\5\2\2\u00b4\u00b6\7\6\2\2")
 *         buf.write("\u00b5\u00b4\3\2\2\2\u00b5\u00b6\3\2\2\2\u00b6\u00b7\3")
 *         buf.write("\2\2\2\u00b7\u00c8\7\7\2\2\u00b8\u00c8\7\b\2\2\u00b9\u00bb")             # <<<<<<<<<<<<<<
 *         buf.write("\7\t\2\2\u00ba\u00b9\3\2\2\2\u00ba\u00bb\3\2\2\2\u00bb")
 *         buf.write("\u00bc\3\2\2\2\u00bc\u00c8\7\n\2\2\u00bd\u00c8\7\13\2")
 */
  __pyx_tuple__230 = PyTuple_Pack(1, __pyx_kp_u__229); if (unlikely(!__pyx_tuple__230)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__230);
  __Pyx_GIVEREF(__pyx_tuple__230);

  /* "pymola/generated/ModelicaParser.pyx":148
 *         buf.write("\u00b5\u00b4\3\2\2\2\u00b5\u00b6\3\2\2\2\u00b6\u00b7\3")
 *         buf.write("\2\2\2\u00b7\u00c8\7\7\2\2\u00b8\u00c8\7\b\2\2\u00b9\u00bb")
 *         buf.write("\7\t\2\2\u00ba\u00b9\3\2\2\2\u00ba\u00bb\3\2\2\2\u00bb")             # <<<<<<<<<<<<<<
 *         buf.write("\u00bc\3\2\2\2\u00bc\u00c8\7\n\2\2\u00bd\u00c8\7\13\2")
 *         buf.write("\2\u00be\u00c8\7\f\2\2\u00bf\u00c1\t\2\2\2\u00c0\u00bf")
 */
  __pyx_tuple__232 = PyTuple_Pack(1, __pyx_kp_u__231); if (unlikely(!__pyx_tuple__232)) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__232);
  __Pyx_GIVEREF(__pyx_tuple__232);

  /* "pymola/generated/ModelicaParser.pyx":149
 *         buf.write("\2\2\2\u00b7\u00c8\7\7\2\2\u00b8\u00c8\7\b\2\2\u00b9\u00bb")
 *         buf.write("\7\t\2\2\u00ba\u00b9\3\2\2\2\u00ba\u00bb\3\2\2\2\u00bb")
 *         buf.write("\u00bc\3\2\2\2\u00bc\u00c8\7\n\2\2\u00bd\u00c8\7\13\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u00be\u00c8\7\f\2\2\u00bf\u00c1\t\2\2\2\u00c0\u00bf")
 *         buf.write("\3\2\2\2\u00c0\u00c1\3\2\2\2\u00c1\u00c3\3\2\2\2\u00c2")
 */
  __pyx_tuple__234 = PyTuple_Pack(1, __pyx_kp_u__233); if (unlikely(!__pyx_tuple__234)) __PYX_ERR(0, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__234);
  __Pyx_GIVEREF(__pyx_tuple__234);

  /* "pymola/generated/ModelicaParser.pyx":150
 *         buf.write("\7\t\2\2\u00ba\u00b9\3\2\2\2\u00ba\u00bb\3\2\2\2\u00bb")
 *         buf.write("\u00bc\3\2\2\2\u00bc\u00c8\7\n\2\2\u00bd\u00c8\7\13\2")
 *         buf.write("\2\u00be\u00c8\7\f\2\2\u00bf\u00c1\t\2\2\2\u00c0\u00bf")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u00c0\u00c1\3\2\2\2\u00c1\u00c3\3\2\2\2\u00c2")
 *         buf.write("\u00c4\7\6\2\2\u00c3\u00c2\3\2\2\2\u00c3\u00c4\3\2\2\2")
 */
  __pyx_tuple__236 = PyTuple_Pack(1, __pyx_kp_u__235); if (unlikely(!__pyx_tuple__236)) __PYX_ERR(0, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__236);
  __Pyx_GIVEREF(__pyx_tuple__236);

  /* "pymola/generated/ModelicaParser.pyx":151
 *         buf.write("\u00bc\3\2\2\2\u00bc\u00c8\7\n\2\2\u00bd\u00c8\7\13\2")
 *         buf.write("\2\u00be\u00c8\7\f\2\2\u00bf\u00c1\t\2\2\2\u00c0\u00bf")
 *         buf.write("\3\2\2\2\u00c0\u00c1\3\2\2\2\u00c1\u00c3\3\2\2\2\u00c2")             # <<<<<<<<<<<<<<
 *         buf.write("\u00c4\7\6\2\2\u00c3\u00c2\3\2\2\2\u00c3\u00c4\3\2\2\2")
 *         buf.write("\u00c4\u00c5\3\2\2\2\u00c5\u00c8\7\17\2\2\u00c6\u00c8")
 */
  __pyx_tuple__238 = PyTuple_Pack(1, __pyx_kp_u__237); if (unlikely(!__pyx_tuple__238)) __PYX_ERR(0, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__238);
  __Pyx_GIVEREF(__pyx_tuple__238);

  /* "pymola/generated/ModelicaParser.pyx":152
 *         buf.write("\2\u00be\u00c8\7\f\2\2\u00bf\u00c1\t\2\2\2\u00c0\u00bf")
 *         buf.write("\3\2\2\2\u00c0\u00c1\3\2\2\2\u00c1\u00c3\3\2\2\2\u00c2")
 *         buf.write("\u00c4\7\6\2\2\u00c3\u00c2\3\2\2\2\u00c3\u00c4\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u00c4\u00c5\3\2\2\2\u00c5\u00c8\7\17\2\2\u00c6\u00c8")
 *         buf.write("\7\6\2\2\u00c7\u00b2\3\2\2\2\u00c7\u00b3\3\2\2\2\u00c7")
 */
  __pyx_tuple__240 = PyTuple_Pack(1, __pyx_kp_u__239); if (unlikely(!__pyx_tuple__240)) __PYX_ERR(0, 152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__240);
  __Pyx_GIVEREF(__pyx_tuple__240);

  /* "pymola/generated/ModelicaParser.pyx":153
 *         buf.write("\3\2\2\2\u00c0\u00c1\3\2\2\2\u00c1\u00c3\3\2\2\2\u00c2")
 *         buf.write("\u00c4\7\6\2\2\u00c3\u00c2\3\2\2\2\u00c3\u00c4\3\2\2\2")
 *         buf.write("\u00c4\u00c5\3\2\2\2\u00c5\u00c8\7\17\2\2\u00c6\u00c8")             # <<<<<<<<<<<<<<
 *         buf.write("\7\6\2\2\u00c7\u00b2\3\2\2\2\u00c7\u00b3\3\2\2\2\u00c7")
 *         buf.write("\u00b5\3\2\2\2\u00c7\u00b8\3\2\2\2\u00c7\u00ba\3\2\2\2")
 */
  __pyx_tuple__242 = PyTuple_Pack(1, __pyx_kp_u__241); if (unlikely(!__pyx_tuple__242)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__242);
  __Pyx_GIVEREF(__pyx_tuple__242);

  /* "pymola/generated/ModelicaParser.pyx":154
 *         buf.write("\u00c4\7\6\2\2\u00c3\u00c2\3\2\2\2\u00c3\u00c4\3\2\2\2")
 *         buf.write("\u00c4\u00c5\3\2\2\2\u00c5\u00c8\7\17\2\2\u00c6\u00c8")
 *         buf.write("\7\6\2\2\u00c7\u00b2\3\2\2\2\u00c7\u00b3\3\2\2\2\u00c7")             # <<<<<<<<<<<<<<
 *         buf.write("\u00b5\3\2\2\2\u00c7\u00b8\3\2\2\2\u00c7\u00ba\3\2\2\2")
 *         buf.write("\u00c7\u00bd\3\2\2\2\u00c7\u00be\3\2\2\2\u00c7\u00c0\3")
 */
  __pyx_tuple__244 = PyTuple_Pack(1, __pyx_kp_u__243); if (unlikely(!__pyx_tuple__244)) __PYX_ERR(0, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__244);
  __Pyx_GIVEREF(__pyx_tuple__244);

  /* "pymola/generated/ModelicaParser.pyx":155
 *         buf.write("\u00c4\u00c5\3\2\2\2\u00c5\u00c8\7\17\2\2\u00c6\u00c8")
 *         buf.write("\7\6\2\2\u00c7\u00b2\3\2\2\2\u00c7\u00b3\3\2\2\2\u00c7")
 *         buf.write("\u00b5\3\2\2\2\u00c7\u00b8\3\2\2\2\u00c7\u00ba\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u00c7\u00bd\3\2\2\2\u00c7\u00be\3\2\2\2\u00c7\u00c0\3")
 *         buf.write("\2\2\2\u00c7\u00c6\3\2\2\2\u00c8\13\3\2\2\2\u00c9\u00ca")
 */
  __pyx_tuple__246 = PyTuple_Pack(1, __pyx_kp_u__245); if (unlikely(!__pyx_tuple__246)) __PYX_ERR(0, 155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__246);
  __Pyx_GIVEREF(__pyx_tuple__246);

  /* "pymola/generated/ModelicaParser.pyx":156
 *         buf.write("\7\6\2\2\u00c7\u00b2\3\2\2\2\u00c7\u00b3\3\2\2\2\u00c7")
 *         buf.write("\u00b5\3\2\2\2\u00c7\u00b8\3\2\2\2\u00c7\u00ba\3\2\2\2")
 *         buf.write("\u00c7\u00bd\3\2\2\2\u00c7\u00be\3\2\2\2\u00c7\u00c0\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u00c7\u00c6\3\2\2\2\u00c8\13\3\2\2\2\u00c9\u00ca")
 *         buf.write("\7Z\2\2\u00ca\u00cb\5\u0090I\2\u00cb\u00cc\5\24\13\2\u00cc")
 */
  __pyx_tuple__248 = PyTuple_Pack(1, __pyx_kp_u__247); if (unlikely(!__pyx_tuple__248)) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__248);
  __Pyx_GIVEREF(__pyx_tuple__248);

  /* "pymola/generated/ModelicaParser.pyx":157
 *         buf.write("\u00b5\3\2\2\2\u00c7\u00b8\3\2\2\2\u00c7\u00ba\3\2\2\2")
 *         buf.write("\u00c7\u00bd\3\2\2\2\u00c7\u00be\3\2\2\2\u00c7\u00c0\3")
 *         buf.write("\2\2\2\u00c7\u00c6\3\2\2\2\u00c8\13\3\2\2\2\u00c9\u00ca")             # <<<<<<<<<<<<<<
 *         buf.write("\7Z\2\2\u00ca\u00cb\5\u0090I\2\u00cb\u00cc\5\24\13\2\u00cc")
 *         buf.write("\u00cd\7\20\2\2\u00cd\u00ce\7Z\2\2\u00ce\u0103\3\2\2\2")
 */
  __pyx_tuple__250 = PyTuple_Pack(1, __pyx_kp_u__249); if (unlikely(!__pyx_tuple__250)) __PYX_ERR(0, 157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__250);
  __Pyx_GIVEREF(__pyx_tuple__250);

  /* "pymola/generated/ModelicaParser.pyx":158
 *         buf.write("\u00c7\u00bd\3\2\2\2\u00c7\u00be\3\2\2\2\u00c7\u00c0\3")
 *         buf.write("\2\2\2\u00c7\u00c6\3\2\2\2\u00c8\13\3\2\2\2\u00c9\u00ca")
 *         buf.write("\7Z\2\2\u00ca\u00cb\5\u0090I\2\u00cb\u00cc\5\24\13\2\u00cc")             # <<<<<<<<<<<<<<
 *         buf.write("\u00cd\7\20\2\2\u00cd\u00ce\7Z\2\2\u00ce\u0103\3\2\2\2")
 *         buf.write("\u00cf\u00d0\7Z\2\2\u00d0\u00d1\7\21\2\2\u00d1\u00d2\5")
 */
  __pyx_tuple__251 = PyTuple_Pack(1, __pyx_kp_u_Z_I); if (unlikely(!__pyx_tuple__251)) __PYX_ERR(0, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__251);
  __Pyx_GIVEREF(__pyx_tuple__251);

  /* "pymola/generated/ModelicaParser.pyx":159
 *         buf.write("\2\2\2\u00c7\u00c6\3\2\2\2\u00c8\13\3\2\2\2\u00c9\u00ca")
 *         buf.write("\7Z\2\2\u00ca\u00cb\5\u0090I\2\u00cb\u00cc\5\24\13\2\u00cc")
 *         buf.write("\u00cd\7\20\2\2\u00cd\u00ce\7Z\2\2\u00ce\u0103\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u00cf\u00d0\7Z\2\2\u00d0\u00d1\7\21\2\2\u00d1\u00d2\5")
 *         buf.write("\16\b\2\u00d2\u00d4\5v<\2\u00d3\u00d5\5\u008aF\2\u00d4")
 */
  __pyx_tuple__252 = PyTuple_Pack(1, __pyx_kp_u_Z_2); if (unlikely(!__pyx_tuple__252)) __PYX_ERR(0, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__252);
  __Pyx_GIVEREF(__pyx_tuple__252);

  /* "pymola/generated/ModelicaParser.pyx":160
 *         buf.write("\7Z\2\2\u00ca\u00cb\5\u0090I\2\u00cb\u00cc\5\24\13\2\u00cc")
 *         buf.write("\u00cd\7\20\2\2\u00cd\u00ce\7Z\2\2\u00ce\u0103\3\2\2\2")
 *         buf.write("\u00cf\u00d0\7Z\2\2\u00d0\u00d1\7\21\2\2\u00d1\u00d2\5")             # <<<<<<<<<<<<<<
 *         buf.write("\16\b\2\u00d2\u00d4\5v<\2\u00d3\u00d5\5\u008aF\2\u00d4")
 *         buf.write("\u00d3\3\2\2\2\u00d4\u00d5\3\2\2\2\u00d5\u00d7\3\2\2\2")
 */
  __pyx_tuple__253 = PyTuple_Pack(1, __pyx_kp_u_Z_3); if (unlikely(!__pyx_tuple__253)) __PYX_ERR(0, 160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__253);
  __Pyx_GIVEREF(__pyx_tuple__253);

  /* "pymola/generated/ModelicaParser.pyx":161
 *         buf.write("\u00cd\7\20\2\2\u00cd\u00ce\7Z\2\2\u00ce\u0103\3\2\2\2")
 *         buf.write("\u00cf\u00d0\7Z\2\2\u00d0\u00d1\7\21\2\2\u00d1\u00d2\5")
 *         buf.write("\16\b\2\u00d2\u00d4\5v<\2\u00d3\u00d5\5\u008aF\2\u00d4")             # <<<<<<<<<<<<<<
 *         buf.write("\u00d3\3\2\2\2\u00d4\u00d5\3\2\2\2\u00d5\u00d7\3\2\2\2")
 *         buf.write("\u00d6\u00d8\5:\36\2\u00d7\u00d6\3\2\2\2\u00d7\u00d8\3")
 */
  __pyx_tuple__254 = PyTuple_Pack(1, __pyx_kp_u_v_F); if (unlikely(!__pyx_tuple__254)) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__254);
  __Pyx_GIVEREF(__pyx_tuple__254);

  /* "pymola/generated/ModelicaParser.pyx":162
 *         buf.write("\u00cf\u00d0\7Z\2\2\u00d0\u00d1\7\21\2\2\u00d1\u00d2\5")
 *         buf.write("\16\b\2\u00d2\u00d4\5v<\2\u00d3\u00d5\5\u008aF\2\u00d4")
 *         buf.write("\u00d3\3\2\2\2\u00d4\u00d5\3\2\2\2\u00d5\u00d7\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u00d6\u00d8\5:\36\2\u00d7\u00d6\3\2\2\2\u00d7\u00d8\3")
 *         buf.write("\2\2\2\u00d8\u00d9\3\2\2\2\u00d9\u00da\5\u008eH\2\u00da")
 */
  __pyx_tuple__256 = PyTuple_Pack(1, __pyx_kp_u__255); if (unlikely(!__pyx_tuple__256)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__256);
  __Pyx_GIVEREF(__pyx_tuple__256);

  /* "pymola/generated/ModelicaParser.pyx":163
 *         buf.write("\16\b\2\u00d2\u00d4\5v<\2\u00d3\u00d5\5\u008aF\2\u00d4")
 *         buf.write("\u00d3\3\2\2\2\u00d4\u00d5\3\2\2\2\u00d5\u00d7\3\2\2\2")
 *         buf.write("\u00d6\u00d8\5:\36\2\u00d7\u00d6\3\2\2\2\u00d7\u00d8\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u00d8\u00d9\3\2\2\2\u00d9\u00da\5\u008eH\2\u00da")
 *         buf.write("\u0103\3\2\2\2\u00db\u00dc\7Z\2\2\u00dc\u00dd\7\21\2\2")
 */
  __pyx_tuple__258 = PyTuple_Pack(1, __pyx_kp_u__257); if (unlikely(!__pyx_tuple__258)) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__258);
  __Pyx_GIVEREF(__pyx_tuple__258);

  /* "pymola/generated/ModelicaParser.pyx":164
 *         buf.write("\u00d3\3\2\2\2\u00d4\u00d5\3\2\2\2\u00d5\u00d7\3\2\2\2")
 *         buf.write("\u00d6\u00d8\5:\36\2\u00d7\u00d6\3\2\2\2\u00d7\u00d8\3")
 *         buf.write("\2\2\2\u00d8\u00d9\3\2\2\2\u00d9\u00da\5\u008eH\2\u00da")             # <<<<<<<<<<<<<<
 *         buf.write("\u0103\3\2\2\2\u00db\u00dc\7Z\2\2\u00dc\u00dd\7\21\2\2")
 *         buf.write("\u00dd\u00de\7\22\2\2\u00de\u00e3\7\23\2\2\u00df\u00e1")
 */
  __pyx_tuple__259 = PyTuple_Pack(1, __pyx_kp_u_H); if (unlikely(!__pyx_tuple__259)) __PYX_ERR(0, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__259);
  __Pyx_GIVEREF(__pyx_tuple__259);

  /* "pymola/generated/ModelicaParser.pyx":165
 *         buf.write("\u00d6\u00d8\5:\36\2\u00d7\u00d6\3\2\2\2\u00d7\u00d8\3")
 *         buf.write("\2\2\2\u00d8\u00d9\3\2\2\2\u00d9\u00da\5\u008eH\2\u00da")
 *         buf.write("\u0103\3\2\2\2\u00db\u00dc\7Z\2\2\u00dc\u00dd\7\21\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u00dd\u00de\7\22\2\2\u00de\u00e3\7\23\2\2\u00df\u00e1")
 *         buf.write("\5\20\t\2\u00e0\u00df\3\2\2\2\u00e0\u00e1\3\2\2\2\u00e1")
 */
  __pyx_tuple__260 = PyTuple_Pack(1, __pyx_kp_u_Z_4); if (unlikely(!__pyx_tuple__260)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__260);
  __Pyx_GIVEREF(__pyx_tuple__260);

  /* "pymola/generated/ModelicaParser.pyx":166
 *         buf.write("\2\2\2\u00d8\u00d9\3\2\2\2\u00d9\u00da\5\u008eH\2\u00da")
 *         buf.write("\u0103\3\2\2\2\u00db\u00dc\7Z\2\2\u00dc\u00dd\7\21\2\2")
 *         buf.write("\u00dd\u00de\7\22\2\2\u00de\u00e3\7\23\2\2\u00df\u00e1")             # <<<<<<<<<<<<<<
 *         buf.write("\5\20\t\2\u00e0\u00df\3\2\2\2\u00e0\u00e1\3\2\2\2\u00e1")
 *         buf.write("\u00e4\3\2\2\2\u00e2\u00e4\7\24\2\2\u00e3\u00e0\3\2\2")
 */
  __pyx_tuple__262 = PyTuple_Pack(1, __pyx_kp_u__261); if (unlikely(!__pyx_tuple__262)) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__262);
  __Pyx_GIVEREF(__pyx_tuple__262);

  /* "pymola/generated/ModelicaParser.pyx":167
 *         buf.write("\u0103\3\2\2\2\u00db\u00dc\7Z\2\2\u00dc\u00dd\7\21\2\2")
 *         buf.write("\u00dd\u00de\7\22\2\2\u00de\u00e3\7\23\2\2\u00df\u00e1")
 *         buf.write("\5\20\t\2\u00e0\u00df\3\2\2\2\u00e0\u00e1\3\2\2\2\u00e1")             # <<<<<<<<<<<<<<
 *         buf.write("\u00e4\3\2\2\2\u00e2\u00e4\7\24\2\2\u00e3\u00e0\3\2\2")
 *         buf.write("\2\u00e3\u00e2\3\2\2\2\u00e4\u00e5\3\2\2\2\u00e5\u00e6")
 */
  __pyx_tuple__264 = PyTuple_Pack(1, __pyx_kp_u__263); if (unlikely(!__pyx_tuple__264)) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__264);
  __Pyx_GIVEREF(__pyx_tuple__264);

  /* "pymola/generated/ModelicaParser.pyx":168
 *         buf.write("\u00dd\u00de\7\22\2\2\u00de\u00e3\7\23\2\2\u00df\u00e1")
 *         buf.write("\5\20\t\2\u00e0\u00df\3\2\2\2\u00e0\u00e1\3\2\2\2\u00e1")
 *         buf.write("\u00e4\3\2\2\2\u00e2\u00e4\7\24\2\2\u00e3\u00e0\3\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u00e3\u00e2\3\2\2\2\u00e4\u00e5\3\2\2\2\u00e5\u00e6")
 *         buf.write("\7\25\2\2\u00e6\u0103\5\u008eH\2\u00e7\u00e8\7Z\2\2\u00e8")
 */
  __pyx_tuple__266 = PyTuple_Pack(1, __pyx_kp_u__265); if (unlikely(!__pyx_tuple__266)) __PYX_ERR(0, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__266);
  __Pyx_GIVEREF(__pyx_tuple__266);

  /* "pymola/generated/ModelicaParser.pyx":169
 *         buf.write("\5\20\t\2\u00e0\u00df\3\2\2\2\u00e0\u00e1\3\2\2\2\u00e1")
 *         buf.write("\u00e4\3\2\2\2\u00e2\u00e4\7\24\2\2\u00e3\u00e0\3\2\2")
 *         buf.write("\2\u00e3\u00e2\3\2\2\2\u00e4\u00e5\3\2\2\2\u00e5\u00e6")             # <<<<<<<<<<<<<<
 *         buf.write("\7\25\2\2\u00e6\u0103\5\u008eH\2\u00e7\u00e8\7Z\2\2\u00e8")
 *         buf.write("\u00e9\7\21\2\2\u00e9\u00ea\7\26\2\2\u00ea\u00eb\7\23")
 */
  __pyx_tuple__268 = PyTuple_Pack(1, __pyx_kp_u__267); if (unlikely(!__pyx_tuple__268)) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__268);
  __Pyx_GIVEREF(__pyx_tuple__268);

  /* "pymola/generated/ModelicaParser.pyx":170
 *         buf.write("\u00e4\3\2\2\2\u00e2\u00e4\7\24\2\2\u00e3\u00e0\3\2\2")
 *         buf.write("\2\u00e3\u00e2\3\2\2\2\u00e4\u00e5\3\2\2\2\u00e5\u00e6")
 *         buf.write("\7\25\2\2\u00e6\u0103\5\u008eH\2\u00e7\u00e8\7Z\2\2\u00e8")             # <<<<<<<<<<<<<<
 *         buf.write("\u00e9\7\21\2\2\u00e9\u00ea\7\26\2\2\u00ea\u00eb\7\23")
 *         buf.write("\2\2\u00eb\u00ec\5v<\2\u00ec\u00ed\7\27\2\2\u00ed\u00f2")
 */
  __pyx_tuple__269 = PyTuple_Pack(1, __pyx_kp_u_H_Z); if (unlikely(!__pyx_tuple__269)) __PYX_ERR(0, 170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__269);
  __Pyx_GIVEREF(__pyx_tuple__269);

  /* "pymola/generated/ModelicaParser.pyx":171
 *         buf.write("\2\u00e3\u00e2\3\2\2\2\u00e4\u00e5\3\2\2\2\u00e5\u00e6")
 *         buf.write("\7\25\2\2\u00e6\u0103\5\u008eH\2\u00e7\u00e8\7Z\2\2\u00e8")
 *         buf.write("\u00e9\7\21\2\2\u00e9\u00ea\7\26\2\2\u00ea\u00eb\7\23")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\u00eb\u00ec\5v<\2\u00ec\u00ed\7\27\2\2\u00ed\u00f2")
 *         buf.write("\7Z\2\2\u00ee\u00ef\7\27\2\2\u00ef\u00f1\7Z\2\2\u00f0")
 */
  __pyx_tuple__271 = PyTuple_Pack(1, __pyx_kp_u__270); if (unlikely(!__pyx_tuple__271)) __PYX_ERR(0, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__271);
  __Pyx_GIVEREF(__pyx_tuple__271);

  /* "pymola/generated/ModelicaParser.pyx":172
 *         buf.write("\7\25\2\2\u00e6\u0103\5\u008eH\2\u00e7\u00e8\7Z\2\2\u00e8")
 *         buf.write("\u00e9\7\21\2\2\u00e9\u00ea\7\26\2\2\u00ea\u00eb\7\23")
 *         buf.write("\2\2\u00eb\u00ec\5v<\2\u00ec\u00ed\7\27\2\2\u00ed\u00f2")             # <<<<<<<<<<<<<<
 *         buf.write("\7Z\2\2\u00ee\u00ef\7\27\2\2\u00ef\u00f1\7Z\2\2\u00f0")
 *         buf.write("\u00ee\3\2\2\2\u00f1\u00f4\3\2\2\2\u00f2\u00f0\3\2\2\2")
 */
  __pyx_tuple__272 = PyTuple_Pack(1, __pyx_kp_u_v); if (unlikely(!__pyx_tuple__272)) __PYX_ERR(0, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__272);
  __Pyx_GIVEREF(__pyx_tuple__272);

  /* "pymola/generated/ModelicaParser.pyx":173
 *         buf.write("\u00e9\7\21\2\2\u00e9\u00ea\7\26\2\2\u00ea\u00eb\7\23")
 *         buf.write("\2\2\u00eb\u00ec\5v<\2\u00ec\u00ed\7\27\2\2\u00ed\u00f2")
 *         buf.write("\7Z\2\2\u00ee\u00ef\7\27\2\2\u00ef\u00f1\7Z\2\2\u00f0")             # <<<<<<<<<<<<<<
 *         buf.write("\u00ee\3\2\2\2\u00f1\u00f4\3\2\2\2\u00f2\u00f0\3\2\2\2")
 *         buf.write("\u00f2\u00f3\3\2\2\2\u00f3\u00f5\3\2\2\2\u00f4\u00f2\3")
 */
  __pyx_tuple__273 = PyTuple_Pack(1, __pyx_kp_u_Z_Z); if (unlikely(!__pyx_tuple__273)) __PYX_ERR(0, 173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__273);
  __Pyx_GIVEREF(__pyx_tuple__273);

  /* "pymola/generated/ModelicaParser.pyx":174
 *         buf.write("\2\2\u00eb\u00ec\5v<\2\u00ec\u00ed\7\27\2\2\u00ed\u00f2")
 *         buf.write("\7Z\2\2\u00ee\u00ef\7\27\2\2\u00ef\u00f1\7Z\2\2\u00f0")
 *         buf.write("\u00ee\3\2\2\2\u00f1\u00f4\3\2\2\2\u00f2\u00f0\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u00f2\u00f3\3\2\2\2\u00f3\u00f5\3\2\2\2\u00f4\u00f2\3")
 *         buf.write("\2\2\2\u00f5\u00f6\7\25\2\2\u00f6\u00f7\5\u008eH\2\u00f7")
 */
  __pyx_tuple__275 = PyTuple_Pack(1, __pyx_kp_u__274); if (unlikely(!__pyx_tuple__275)) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__275);
  __Pyx_GIVEREF(__pyx_tuple__275);

  /* "pymola/generated/ModelicaParser.pyx":175
 *         buf.write("\7Z\2\2\u00ee\u00ef\7\27\2\2\u00ef\u00f1\7Z\2\2\u00f0")
 *         buf.write("\u00ee\3\2\2\2\u00f1\u00f4\3\2\2\2\u00f2\u00f0\3\2\2\2")
 *         buf.write("\u00f2\u00f3\3\2\2\2\u00f3\u00f5\3\2\2\2\u00f4\u00f2\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u00f5\u00f6\7\25\2\2\u00f6\u00f7\5\u008eH\2\u00f7")
 *         buf.write("\u0103\3\2\2\2\u00f8\u00f9\7\30\2\2\u00f9\u00fb\7Z\2\2")
 */
  __pyx_tuple__277 = PyTuple_Pack(1, __pyx_kp_u__276); if (unlikely(!__pyx_tuple__277)) __PYX_ERR(0, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__277);
  __Pyx_GIVEREF(__pyx_tuple__277);

  /* "pymola/generated/ModelicaParser.pyx":176
 *         buf.write("\u00ee\3\2\2\2\u00f1\u00f4\3\2\2\2\u00f2\u00f0\3\2\2\2")
 *         buf.write("\u00f2\u00f3\3\2\2\2\u00f3\u00f5\3\2\2\2\u00f4\u00f2\3")
 *         buf.write("\2\2\2\u00f5\u00f6\7\25\2\2\u00f6\u00f7\5\u008eH\2\u00f7")             # <<<<<<<<<<<<<<
 *         buf.write("\u0103\3\2\2\2\u00f8\u00f9\7\30\2\2\u00f9\u00fb\7Z\2\2")
 *         buf.write("\u00fa\u00fc\5:\36\2\u00fb\u00fa\3\2\2\2\u00fb\u00fc\3")
 */
  __pyx_tuple__278 = PyTuple_Pack(1, __pyx_kp_u_H_2); if (unlikely(!__pyx_tuple__278)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__278);
  __Pyx_GIVEREF(__pyx_tuple__278);

  /* "pymola/generated/ModelicaParser.pyx":177
 *         buf.write("\u00f2\u00f3\3\2\2\2\u00f3\u00f5\3\2\2\2\u00f4\u00f2\3")
 *         buf.write("\2\2\2\u00f5\u00f6\7\25\2\2\u00f6\u00f7\5\u008eH\2\u00f7")
 *         buf.write("\u0103\3\2\2\2\u00f8\u00f9\7\30\2\2\u00f9\u00fb\7Z\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u00fa\u00fc\5:\36\2\u00fb\u00fa\3\2\2\2\u00fb\u00fc\3")
 *         buf.write("\2\2\2\u00fc\u00fd\3\2\2\2\u00fd\u00fe\5\u0090I\2\u00fe")
 */
  __pyx_tuple__279 = PyTuple_Pack(1, __pyx_kp_u_Z_5); if (unlikely(!__pyx_tuple__279)) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__279);
  __Pyx_GIVEREF(__pyx_tuple__279);

  /* "pymola/generated/ModelicaParser.pyx":178
 *         buf.write("\2\2\2\u00f5\u00f6\7\25\2\2\u00f6\u00f7\5\u008eH\2\u00f7")
 *         buf.write("\u0103\3\2\2\2\u00f8\u00f9\7\30\2\2\u00f9\u00fb\7Z\2\2")
 *         buf.write("\u00fa\u00fc\5:\36\2\u00fb\u00fa\3\2\2\2\u00fb\u00fc\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u00fc\u00fd\3\2\2\2\u00fd\u00fe\5\u0090I\2\u00fe")
 *         buf.write("\u00ff\5\24\13\2\u00ff\u0100\7\20\2\2\u0100\u0101\7Z\2")
 */
  __pyx_tuple__281 = PyTuple_Pack(1, __pyx_kp_u__280); if (unlikely(!__pyx_tuple__281)) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__281);
  __Pyx_GIVEREF(__pyx_tuple__281);

  /* "pymola/generated/ModelicaParser.pyx":179
 *         buf.write("\u0103\3\2\2\2\u00f8\u00f9\7\30\2\2\u00f9\u00fb\7Z\2\2")
 *         buf.write("\u00fa\u00fc\5:\36\2\u00fb\u00fa\3\2\2\2\u00fb\u00fc\3")
 *         buf.write("\2\2\2\u00fc\u00fd\3\2\2\2\u00fd\u00fe\5\u0090I\2\u00fe")             # <<<<<<<<<<<<<<
 *         buf.write("\u00ff\5\24\13\2\u00ff\u0100\7\20\2\2\u0100\u0101\7Z\2")
 *         buf.write("\2\u0101\u0103\3\2\2\2\u0102\u00c9\3\2\2\2\u0102\u00cf")
 */
  __pyx_tuple__282 = PyTuple_Pack(1, __pyx_kp_u_I); if (unlikely(!__pyx_tuple__282)) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__282);
  __Pyx_GIVEREF(__pyx_tuple__282);

  /* "pymola/generated/ModelicaParser.pyx":180
 *         buf.write("\u00fa\u00fc\5:\36\2\u00fb\u00fa\3\2\2\2\u00fb\u00fc\3")
 *         buf.write("\2\2\2\u00fc\u00fd\3\2\2\2\u00fd\u00fe\5\u0090I\2\u00fe")
 *         buf.write("\u00ff\5\24\13\2\u00ff\u0100\7\20\2\2\u0100\u0101\7Z\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u0101\u0103\3\2\2\2\u0102\u00c9\3\2\2\2\u0102\u00cf")
 *         buf.write("\3\2\2\2\u0102\u00db\3\2\2\2\u0102\u00e7\3\2\2\2\u0102")
 */
  __pyx_tuple__283 = PyTuple_Pack(1, __pyx_kp_u_Z_6); if (unlikely(!__pyx_tuple__283)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__283);
  __Pyx_GIVEREF(__pyx_tuple__283);

  /* "pymola/generated/ModelicaParser.pyx":181
 *         buf.write("\2\2\2\u00fc\u00fd\3\2\2\2\u00fd\u00fe\5\u0090I\2\u00fe")
 *         buf.write("\u00ff\5\24\13\2\u00ff\u0100\7\20\2\2\u0100\u0101\7Z\2")
 *         buf.write("\2\u0101\u0103\3\2\2\2\u0102\u00c9\3\2\2\2\u0102\u00cf")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0102\u00db\3\2\2\2\u0102\u00e7\3\2\2\2\u0102")
 *         buf.write("\u00f8\3\2\2\2\u0103\r\3\2\2\2\u0104\u0105\5,\27\2\u0105")
 */
  __pyx_tuple__285 = PyTuple_Pack(1, __pyx_kp_u__284); if (unlikely(!__pyx_tuple__285)) __PYX_ERR(0, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__285);
  __Pyx_GIVEREF(__pyx_tuple__285);

  /* "pymola/generated/ModelicaParser.pyx":182
 *         buf.write("\u00ff\5\24\13\2\u00ff\u0100\7\20\2\2\u0100\u0101\7Z\2")
 *         buf.write("\2\u0101\u0103\3\2\2\2\u0102\u00c9\3\2\2\2\u0102\u00cf")
 *         buf.write("\3\2\2\2\u0102\u00db\3\2\2\2\u0102\u00e7\3\2\2\2\u0102")             # <<<<<<<<<<<<<<
 *         buf.write("\u00f8\3\2\2\2\u0103\r\3\2\2\2\u0104\u0105\5,\27\2\u0105")
 *         buf.write("\17\3\2\2\2\u0106\u010b\5\22\n\2\u0107\u0108\7\27\2\2")
 */
  __pyx_tuple__287 = PyTuple_Pack(1, __pyx_kp_u__286); if (unlikely(!__pyx_tuple__287)) __PYX_ERR(0, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__287);
  __Pyx_GIVEREF(__pyx_tuple__287);

  /* "pymola/generated/ModelicaParser.pyx":183
 *         buf.write("\2\u0101\u0103\3\2\2\2\u0102\u00c9\3\2\2\2\u0102\u00cf")
 *         buf.write("\3\2\2\2\u0102\u00db\3\2\2\2\u0102\u00e7\3\2\2\2\u0102")
 *         buf.write("\u00f8\3\2\2\2\u0103\r\3\2\2\2\u0104\u0105\5,\27\2\u0105")             # <<<<<<<<<<<<<<
 *         buf.write("\17\3\2\2\2\u0106\u010b\5\22\n\2\u0107\u0108\7\27\2\2")
 *         buf.write("\u0108\u010a\5\22\n\2\u0109\u0107\3\2\2\2\u010a\u010d")
 */
  __pyx_tuple__289 = PyTuple_Pack(1, __pyx_kp_u__288); if (unlikely(!__pyx_tuple__289)) __PYX_ERR(0, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__289);
  __Pyx_GIVEREF(__pyx_tuple__289);

  /* "pymola/generated/ModelicaParser.pyx":184
 *         buf.write("\3\2\2\2\u0102\u00db\3\2\2\2\u0102\u00e7\3\2\2\2\u0102")
 *         buf.write("\u00f8\3\2\2\2\u0103\r\3\2\2\2\u0104\u0105\5,\27\2\u0105")
 *         buf.write("\17\3\2\2\2\u0106\u010b\5\22\n\2\u0107\u0108\7\27\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0108\u010a\5\22\n\2\u0109\u0107\3\2\2\2\u010a\u010d")
 *         buf.write("\3\2\2\2\u010b\u0109\3\2\2\2\u010b\u010c\3\2\2\2\u010c")
 */
  __pyx_tuple__291 = PyTuple_Pack(1, __pyx_kp_u__290); if (unlikely(!__pyx_tuple__291)) __PYX_ERR(0, 184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__291);
  __Pyx_GIVEREF(__pyx_tuple__291);

  /* "pymola/generated/ModelicaParser.pyx":185
 *         buf.write("\u00f8\3\2\2\2\u0103\r\3\2\2\2\u0104\u0105\5,\27\2\u0105")
 *         buf.write("\17\3\2\2\2\u0106\u010b\5\22\n\2\u0107\u0108\7\27\2\2")
 *         buf.write("\u0108\u010a\5\22\n\2\u0109\u0107\3\2\2\2\u010a\u010d")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u010b\u0109\3\2\2\2\u010b\u010c\3\2\2\2\u010c")
 *         buf.write("\21\3\2\2\2\u010d\u010b\3\2\2\2\u010e\u010f\7Z\2\2\u010f")
 */
  __pyx_tuple__293 = PyTuple_Pack(1, __pyx_kp_u__292); if (unlikely(!__pyx_tuple__293)) __PYX_ERR(0, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__293);
  __Pyx_GIVEREF(__pyx_tuple__293);

  /* "pymola/generated/ModelicaParser.pyx":186
 *         buf.write("\17\3\2\2\2\u0106\u010b\5\22\n\2\u0107\u0108\7\27\2\2")
 *         buf.write("\u0108\u010a\5\22\n\2\u0109\u0107\3\2\2\2\u010a\u010d")
 *         buf.write("\3\2\2\2\u010b\u0109\3\2\2\2\u010b\u010c\3\2\2\2\u010c")             # <<<<<<<<<<<<<<
 *         buf.write("\21\3\2\2\2\u010d\u010b\3\2\2\2\u010e\u010f\7Z\2\2\u010f")
 *         buf.write("\u0110\5\u008eH\2\u0110\23\3\2\2\2\u0111\u011a\5\32\16")
 */
  __pyx_tuple__295 = PyTuple_Pack(1, __pyx_kp_u__294); if (unlikely(!__pyx_tuple__295)) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__295);
  __Pyx_GIVEREF(__pyx_tuple__295);

  /* "pymola/generated/ModelicaParser.pyx":187
 *         buf.write("\u0108\u010a\5\22\n\2\u0109\u0107\3\2\2\2\u010a\u010d")
 *         buf.write("\3\2\2\2\u010b\u0109\3\2\2\2\u010b\u010c\3\2\2\2\u010c")
 *         buf.write("\21\3\2\2\2\u010d\u010b\3\2\2\2\u010e\u010f\7Z\2\2\u010f")             # <<<<<<<<<<<<<<
 *         buf.write("\u0110\5\u008eH\2\u0110\23\3\2\2\2\u0111\u011a\5\32\16")
 *         buf.write("\2\u0112\u0113\7\31\2\2\u0113\u0119\5\32\16\2\u0114\u0115")
 */
  __pyx_tuple__296 = PyTuple_Pack(1, __pyx_kp_u_Z_7); if (unlikely(!__pyx_tuple__296)) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__296);
  __Pyx_GIVEREF(__pyx_tuple__296);

  /* "pymola/generated/ModelicaParser.pyx":188
 *         buf.write("\3\2\2\2\u010b\u0109\3\2\2\2\u010b\u010c\3\2\2\2\u010c")
 *         buf.write("\21\3\2\2\2\u010d\u010b\3\2\2\2\u010e\u010f\7Z\2\2\u010f")
 *         buf.write("\u0110\5\u008eH\2\u0110\23\3\2\2\2\u0111\u011a\5\32\16")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u0112\u0113\7\31\2\2\u0113\u0119\5\32\16\2\u0114\u0115")
 *         buf.write("\7\32\2\2\u0115\u0119\5\32\16\2\u0116\u0119\5N(\2\u0117")
 */
  __pyx_tuple__297 = PyTuple_Pack(1, __pyx_kp_u_H_3); if (unlikely(!__pyx_tuple__297)) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__297);
  __Pyx_GIVEREF(__pyx_tuple__297);

  /* "pymola/generated/ModelicaParser.pyx":189
 *         buf.write("\21\3\2\2\2\u010d\u010b\3\2\2\2\u010e\u010f\7Z\2\2\u010f")
 *         buf.write("\u0110\5\u008eH\2\u0110\23\3\2\2\2\u0111\u011a\5\32\16")
 *         buf.write("\2\u0112\u0113\7\31\2\2\u0113\u0119\5\32\16\2\u0114\u0115")             # <<<<<<<<<<<<<<
 *         buf.write("\7\32\2\2\u0115\u0119\5\32\16\2\u0116\u0119\5N(\2\u0117")
 *         buf.write("\u0119\5P)\2\u0118\u0112\3\2\2\2\u0118\u0114\3\2\2\2\u0118")
 */
  __pyx_tuple__299 = PyTuple_Pack(1, __pyx_kp_u__298); if (unlikely(!__pyx_tuple__299)) __PYX_ERR(0, 189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__299);
  __Pyx_GIVEREF(__pyx_tuple__299);

  /* "pymola/generated/ModelicaParser.pyx":190
 *         buf.write("\u0110\5\u008eH\2\u0110\23\3\2\2\2\u0111\u011a\5\32\16")
 *         buf.write("\2\u0112\u0113\7\31\2\2\u0113\u0119\5\32\16\2\u0114\u0115")
 *         buf.write("\7\32\2\2\u0115\u0119\5\32\16\2\u0116\u0119\5N(\2\u0117")             # <<<<<<<<<<<<<<
 *         buf.write("\u0119\5P)\2\u0118\u0112\3\2\2\2\u0118\u0114\3\2\2\2\u0118")
 *         buf.write("\u0116\3\2\2\2\u0118\u0117\3\2\2\2\u0119\u011c\3\2\2\2")
 */
  __pyx_tuple__300 = PyTuple_Pack(1, __pyx_kp_u_N); if (unlikely(!__pyx_tuple__300)) __PYX_ERR(0, 190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__300);
  __Pyx_GIVEREF(__pyx_tuple__300);

  /* "pymola/generated/ModelicaParser.pyx":191
 *         buf.write("\2\u0112\u0113\7\31\2\2\u0113\u0119\5\32\16\2\u0114\u0115")
 *         buf.write("\7\32\2\2\u0115\u0119\5\32\16\2\u0116\u0119\5N(\2\u0117")
 *         buf.write("\u0119\5P)\2\u0118\u0112\3\2\2\2\u0118\u0114\3\2\2\2\u0118")             # <<<<<<<<<<<<<<
 *         buf.write("\u0116\3\2\2\2\u0118\u0117\3\2\2\2\u0119\u011c\3\2\2\2")
 *         buf.write("\u011a\u0118\3\2\2\2\u011a\u011b\3\2\2\2\u011b\u0128\3")
 */
  __pyx_tuple__301 = PyTuple_Pack(1, __pyx_kp_u_P); if (unlikely(!__pyx_tuple__301)) __PYX_ERR(0, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__301);
  __Pyx_GIVEREF(__pyx_tuple__301);

  /* "pymola/generated/ModelicaParser.pyx":192
 *         buf.write("\7\32\2\2\u0115\u0119\5\32\16\2\u0116\u0119\5N(\2\u0117")
 *         buf.write("\u0119\5P)\2\u0118\u0112\3\2\2\2\u0118\u0114\3\2\2\2\u0118")
 *         buf.write("\u0116\3\2\2\2\u0118\u0117\3\2\2\2\u0119\u011c\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u011a\u0118\3\2\2\2\u011a\u011b\3\2\2\2\u011b\u0128\3")
 *         buf.write("\2\2\2\u011c\u011a\3\2\2\2\u011d\u011f\7\33\2\2\u011e")
 */
  __pyx_tuple__303 = PyTuple_Pack(1, __pyx_kp_u__302); if (unlikely(!__pyx_tuple__303)) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__303);
  __Pyx_GIVEREF(__pyx_tuple__303);

  /* "pymola/generated/ModelicaParser.pyx":193
 *         buf.write("\u0119\5P)\2\u0118\u0112\3\2\2\2\u0118\u0114\3\2\2\2\u0118")
 *         buf.write("\u0116\3\2\2\2\u0118\u0117\3\2\2\2\u0119\u011c\3\2\2\2")
 *         buf.write("\u011a\u0118\3\2\2\2\u011a\u011b\3\2\2\2\u011b\u0128\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u011c\u011a\3\2\2\2\u011d\u011f\7\33\2\2\u011e")
 *         buf.write("\u0120\5\26\f\2\u011f\u011e\3\2\2\2\u011f\u0120\3\2\2")
 */
  __pyx_tuple__305 = PyTuple_Pack(1, __pyx_kp_u__304); if (unlikely(!__pyx_tuple__305)) __PYX_ERR(0, 193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__305);
  __Pyx_GIVEREF(__pyx_tuple__305);

  /* "pymola/generated/ModelicaParser.pyx":194
 *         buf.write("\u0116\3\2\2\2\u0118\u0117\3\2\2\2\u0119\u011c\3\2\2\2")
 *         buf.write("\u011a\u0118\3\2\2\2\u011a\u011b\3\2\2\2\u011b\u0128\3")
 *         buf.write("\2\2\2\u011c\u011a\3\2\2\2\u011d\u011f\7\33\2\2\u011e")             # <<<<<<<<<<<<<<
 *         buf.write("\u0120\5\26\f\2\u011f\u011e\3\2\2\2\u011f\u0120\3\2\2")
 *         buf.write("\2\u0120\u0122\3\2\2\2\u0121\u0123\5\30\r\2\u0122\u0121")
 */
  __pyx_tuple__307 = PyTuple_Pack(1, __pyx_kp_u__306); if (unlikely(!__pyx_tuple__307)) __PYX_ERR(0, 194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__307);
  __Pyx_GIVEREF(__pyx_tuple__307);

  /* "pymola/generated/ModelicaParser.pyx":195
 *         buf.write("\u011a\u0118\3\2\2\2\u011a\u011b\3\2\2\2\u011b\u0128\3")
 *         buf.write("\2\2\2\u011c\u011a\3\2\2\2\u011d\u011f\7\33\2\2\u011e")
 *         buf.write("\u0120\5\26\f\2\u011f\u011e\3\2\2\2\u011f\u0120\3\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u0120\u0122\3\2\2\2\u0121\u0123\5\30\r\2\u0122\u0121")
 *         buf.write("\3\2\2\2\u0122\u0123\3\2\2\2\u0123\u0125\3\2\2\2\u0124")
 */
  __pyx_tuple__309 = PyTuple_Pack(1, __pyx_kp_u__308); if (unlikely(!__pyx_tuple__309)) __PYX_ERR(0, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__309);
  __Pyx_GIVEREF(__pyx_tuple__309);

  /* "pymola/generated/ModelicaParser.pyx":196
 *         buf.write("\2\2\2\u011c\u011a\3\2\2\2\u011d\u011f\7\33\2\2\u011e")
 *         buf.write("\u0120\5\26\f\2\u011f\u011e\3\2\2\2\u011f\u0120\3\2\2")
 *         buf.write("\2\u0120\u0122\3\2\2\2\u0121\u0123\5\30\r\2\u0122\u0121")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0122\u0123\3\2\2\2\u0123\u0125\3\2\2\2\u0124")
 *         buf.write("\u0126\5\u0092J\2\u0125\u0124\3\2\2\2\u0125\u0126\3\2")
 */
  __pyx_tuple__311 = PyTuple_Pack(1, __pyx_kp_u__310); if (unlikely(!__pyx_tuple__311)) __PYX_ERR(0, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__311);
  __Pyx_GIVEREF(__pyx_tuple__311);

  /* "pymola/generated/ModelicaParser.pyx":197
 *         buf.write("\u0120\5\26\f\2\u011f\u011e\3\2\2\2\u011f\u0120\3\2\2")
 *         buf.write("\2\u0120\u0122\3\2\2\2\u0121\u0123\5\30\r\2\u0122\u0121")
 *         buf.write("\3\2\2\2\u0122\u0123\3\2\2\2\u0123\u0125\3\2\2\2\u0124")             # <<<<<<<<<<<<<<
 *         buf.write("\u0126\5\u0092J\2\u0125\u0124\3\2\2\2\u0125\u0126\3\2")
 *         buf.write("\2\2\u0126\u0127\3\2\2\2\u0127\u0129\7\24\2\2\u0128\u011d")
 */
  __pyx_tuple__313 = PyTuple_Pack(1, __pyx_kp_u__312); if (unlikely(!__pyx_tuple__313)) __PYX_ERR(0, 197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__313);
  __Pyx_GIVEREF(__pyx_tuple__313);

  /* "pymola/generated/ModelicaParser.pyx":198
 *         buf.write("\2\u0120\u0122\3\2\2\2\u0121\u0123\5\30\r\2\u0122\u0121")
 *         buf.write("\3\2\2\2\u0122\u0123\3\2\2\2\u0123\u0125\3\2\2\2\u0124")
 *         buf.write("\u0126\5\u0092J\2\u0125\u0124\3\2\2\2\u0125\u0126\3\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\u0126\u0127\3\2\2\2\u0127\u0129\7\24\2\2\u0128\u011d")
 *         buf.write("\3\2\2\2\u0128\u0129\3\2\2\2\u0129\u012d\3\2\2\2\u012a")
 */
  __pyx_tuple__314 = PyTuple_Pack(1, __pyx_kp_u_J); if (unlikely(!__pyx_tuple__314)) __PYX_ERR(0, 198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__314);
  __Pyx_GIVEREF(__pyx_tuple__314);

  /* "pymola/generated/ModelicaParser.pyx":199
 *         buf.write("\3\2\2\2\u0122\u0123\3\2\2\2\u0123\u0125\3\2\2\2\u0124")
 *         buf.write("\u0126\5\u0092J\2\u0125\u0124\3\2\2\2\u0125\u0126\3\2")
 *         buf.write("\2\2\u0126\u0127\3\2\2\2\u0127\u0129\7\24\2\2\u0128\u011d")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0128\u0129\3\2\2\2\u0129\u012d\3\2\2\2\u012a")
 *         buf.write("\u012b\5\u0092J\2\u012b\u012c\7\3\2\2\u012c\u012e\3\2")
 */
  __pyx_tuple__316 = PyTuple_Pack(1, __pyx_kp_u__315); if (unlikely(!__pyx_tuple__316)) __PYX_ERR(0, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__316);
  __Pyx_GIVEREF(__pyx_tuple__316);

  /* "pymola/generated/ModelicaParser.pyx":200
 *         buf.write("\u0126\5\u0092J\2\u0125\u0124\3\2\2\2\u0125\u0126\3\2")
 *         buf.write("\2\2\u0126\u0127\3\2\2\2\u0127\u0129\7\24\2\2\u0128\u011d")
 *         buf.write("\3\2\2\2\u0128\u0129\3\2\2\2\u0129\u012d\3\2\2\2\u012a")             # <<<<<<<<<<<<<<
 *         buf.write("\u012b\5\u0092J\2\u012b\u012c\7\3\2\2\u012c\u012e\3\2")
 *         buf.write("\2\2\u012d\u012a\3\2\2\2\u012d\u012e\3\2\2\2\u012e\25")
 */
  __pyx_tuple__318 = PyTuple_Pack(1, __pyx_kp_u__317); if (unlikely(!__pyx_tuple__318)) __PYX_ERR(0, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__318);
  __Pyx_GIVEREF(__pyx_tuple__318);

  /* "pymola/generated/ModelicaParser.pyx":201
 *         buf.write("\2\2\u0126\u0127\3\2\2\2\u0127\u0129\7\24\2\2\u0128\u011d")
 *         buf.write("\3\2\2\2\u0128\u0129\3\2\2\2\u0129\u012d\3\2\2\2\u012a")
 *         buf.write("\u012b\5\u0092J\2\u012b\u012c\7\3\2\2\u012c\u012e\3\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\u012d\u012a\3\2\2\2\u012d\u012e\3\2\2\2\u012e\25")
 *         buf.write("\3\2\2\2\u012f\u0130\7[\2\2\u0130\27\3\2\2\2\u0131\u0132")
 */
  __pyx_tuple__319 = PyTuple_Pack(1, __pyx_kp_u_J_2); if (unlikely(!__pyx_tuple__319)) __PYX_ERR(0, 201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__319);
  __Pyx_GIVEREF(__pyx_tuple__319);

  /* "pymola/generated/ModelicaParser.pyx":202
 *         buf.write("\3\2\2\2\u0128\u0129\3\2\2\2\u0129\u012d\3\2\2\2\u012a")
 *         buf.write("\u012b\5\u0092J\2\u012b\u012c\7\3\2\2\u012c\u012e\3\2")
 *         buf.write("\2\2\u012d\u012a\3\2\2\2\u012d\u012e\3\2\2\2\u012e\25")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u012f\u0130\7[\2\2\u0130\27\3\2\2\2\u0131\u0132")
 *         buf.write("\5z>\2\u0132\u0133\7\21\2\2\u0133\u0135\3\2\2\2\u0134")
 */
  __pyx_tuple__321 = PyTuple_Pack(1, __pyx_kp_u__320); if (unlikely(!__pyx_tuple__321)) __PYX_ERR(0, 202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__321);
  __Pyx_GIVEREF(__pyx_tuple__321);

  /* "pymola/generated/ModelicaParser.pyx":203
 *         buf.write("\u012b\5\u0092J\2\u012b\u012c\7\3\2\2\u012c\u012e\3\2")
 *         buf.write("\2\2\u012d\u012a\3\2\2\2\u012d\u012e\3\2\2\2\u012e\25")
 *         buf.write("\3\2\2\2\u012f\u0130\7[\2\2\u0130\27\3\2\2\2\u0131\u0132")             # <<<<<<<<<<<<<<
 *         buf.write("\5z>\2\u0132\u0133\7\21\2\2\u0133\u0135\3\2\2\2\u0134")
 *         buf.write("\u0131\3\2\2\2\u0134\u0135\3\2\2\2\u0135\u0136\3\2\2\2")
 */
  __pyx_tuple__323 = PyTuple_Pack(1, __pyx_kp_u__322); if (unlikely(!__pyx_tuple__323)) __PYX_ERR(0, 203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__323);
  __Pyx_GIVEREF(__pyx_tuple__323);

  /* "pymola/generated/ModelicaParser.pyx":204
 *         buf.write("\2\2\u012d\u012a\3\2\2\2\u012d\u012e\3\2\2\2\u012e\25")
 *         buf.write("\3\2\2\2\u012f\u0130\7[\2\2\u0130\27\3\2\2\2\u0131\u0132")
 *         buf.write("\5z>\2\u0132\u0133\7\21\2\2\u0133\u0135\3\2\2\2\u0134")             # <<<<<<<<<<<<<<
 *         buf.write("\u0131\3\2\2\2\u0134\u0135\3\2\2\2\u0135\u0136\3\2\2\2")
 *         buf.write("\u0136\u0137\7Z\2\2\u0137\u0139\7\23\2\2\u0138\u013a\5")
 */
  __pyx_tuple__324 = PyTuple_Pack(1, __pyx_kp_u_z); if (unlikely(!__pyx_tuple__324)) __PYX_ERR(0, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__324);
  __Pyx_GIVEREF(__pyx_tuple__324);

  /* "pymola/generated/ModelicaParser.pyx":205
 *         buf.write("\3\2\2\2\u012f\u0130\7[\2\2\u0130\27\3\2\2\2\u0131\u0132")
 *         buf.write("\5z>\2\u0132\u0133\7\21\2\2\u0133\u0135\3\2\2\2\u0134")
 *         buf.write("\u0131\3\2\2\2\u0134\u0135\3\2\2\2\u0135\u0136\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0136\u0137\7Z\2\2\u0137\u0139\7\23\2\2\u0138\u013a\5")
 *         buf.write("\u0088E\2\u0139\u0138\3\2\2\2\u0139\u013a\3\2\2\2\u013a")
 */
  __pyx_tuple__326 = PyTuple_Pack(1, __pyx_kp_u__325); if (unlikely(!__pyx_tuple__326)) __PYX_ERR(0, 205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__326);
  __Pyx_GIVEREF(__pyx_tuple__326);

  /* "pymola/generated/ModelicaParser.pyx":206
 *         buf.write("\5z>\2\u0132\u0133\7\21\2\2\u0133\u0135\3\2\2\2\u0134")
 *         buf.write("\u0131\3\2\2\2\u0134\u0135\3\2\2\2\u0135\u0136\3\2\2\2")
 *         buf.write("\u0136\u0137\7Z\2\2\u0137\u0139\7\23\2\2\u0138\u013a\5")             # <<<<<<<<<<<<<<
 *         buf.write("\u0088E\2\u0139\u0138\3\2\2\2\u0139\u013a\3\2\2\2\u013a")
 *         buf.write("\u013b\3\2\2\2\u013b\u013c\7\25\2\2\u013c\31\3\2\2\2\u013d")
 */
  __pyx_tuple__327 = PyTuple_Pack(1, __pyx_kp_u_Z_8); if (unlikely(!__pyx_tuple__327)) __PYX_ERR(0, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__327);
  __Pyx_GIVEREF(__pyx_tuple__327);

  /* "pymola/generated/ModelicaParser.pyx":207
 *         buf.write("\u0131\3\2\2\2\u0134\u0135\3\2\2\2\u0135\u0136\3\2\2\2")
 *         buf.write("\u0136\u0137\7Z\2\2\u0137\u0139\7\23\2\2\u0138\u013a\5")
 *         buf.write("\u0088E\2\u0139\u0138\3\2\2\2\u0139\u013a\3\2\2\2\u013a")             # <<<<<<<<<<<<<<
 *         buf.write("\u013b\3\2\2\2\u013b\u013c\7\25\2\2\u013c\31\3\2\2\2\u013d")
 *         buf.write("\u013e\5\34\17\2\u013e\u013f\7\3\2\2\u013f\u0141\3\2\2")
 */
  __pyx_tuple__328 = PyTuple_Pack(1, __pyx_kp_u_E); if (unlikely(!__pyx_tuple__328)) __PYX_ERR(0, 207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__328);
  __Pyx_GIVEREF(__pyx_tuple__328);

  /* "pymola/generated/ModelicaParser.pyx":208
 *         buf.write("\u0136\u0137\7Z\2\2\u0137\u0139\7\23\2\2\u0138\u013a\5")
 *         buf.write("\u0088E\2\u0139\u0138\3\2\2\2\u0139\u013a\3\2\2\2\u013a")
 *         buf.write("\u013b\3\2\2\2\u013b\u013c\7\25\2\2\u013c\31\3\2\2\2\u013d")             # <<<<<<<<<<<<<<
 *         buf.write("\u013e\5\34\17\2\u013e\u013f\7\3\2\2\u013f\u0141\3\2\2")
 *         buf.write("\2\u0140\u013d\3\2\2\2\u0141\u0144\3\2\2\2\u0142\u0140")
 */
  __pyx_tuple__330 = PyTuple_Pack(1, __pyx_kp_u__329); if (unlikely(!__pyx_tuple__330)) __PYX_ERR(0, 208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__330);
  __Pyx_GIVEREF(__pyx_tuple__330);

  /* "pymola/generated/ModelicaParser.pyx":209
 *         buf.write("\u0088E\2\u0139\u0138\3\2\2\2\u0139\u013a\3\2\2\2\u013a")
 *         buf.write("\u013b\3\2\2\2\u013b\u013c\7\25\2\2\u013c\31\3\2\2\2\u013d")
 *         buf.write("\u013e\5\34\17\2\u013e\u013f\7\3\2\2\u013f\u0141\3\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u0140\u013d\3\2\2\2\u0141\u0144\3\2\2\2\u0142\u0140")
 *         buf.write("\3\2\2\2\u0142\u0143\3\2\2\2\u0143\33\3\2\2\2\u0144\u0142")
 */
  __pyx_tuple__332 = PyTuple_Pack(1, __pyx_kp_u__331); if (unlikely(!__pyx_tuple__332)) __PYX_ERR(0, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__332);
  __Pyx_GIVEREF(__pyx_tuple__332);

  /* "pymola/generated/ModelicaParser.pyx":210
 *         buf.write("\u013b\3\2\2\2\u013b\u013c\7\25\2\2\u013c\31\3\2\2\2\u013d")
 *         buf.write("\u013e\5\34\17\2\u013e\u013f\7\3\2\2\u013f\u0141\3\2\2")
 *         buf.write("\2\u0140\u013d\3\2\2\2\u0141\u0144\3\2\2\2\u0142\u0140")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0142\u0143\3\2\2\2\u0143\33\3\2\2\2\u0144\u0142")
 *         buf.write("\3\2\2\2\u0145\u014a\5\"\22\2\u0146\u014a\5&\24\2\u0147")
 */
  __pyx_tuple__334 = PyTuple_Pack(1, __pyx_kp_u__333); if (unlikely(!__pyx_tuple__334)) __PYX_ERR(0, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__334);
  __Pyx_GIVEREF(__pyx_tuple__334);

  /* "pymola/generated/ModelicaParser.pyx":211
 *         buf.write("\u013e\5\34\17\2\u013e\u013f\7\3\2\2\u013f\u0141\3\2\2")
 *         buf.write("\2\u0140\u013d\3\2\2\2\u0141\u0144\3\2\2\2\u0142\u0140")
 *         buf.write("\3\2\2\2\u0142\u0143\3\2\2\2\u0143\33\3\2\2\2\u0144\u0142")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0145\u014a\5\"\22\2\u0146\u014a\5&\24\2\u0147")
 *         buf.write("\u014a\5\36\20\2\u0148\u014a\5 \21\2\u0149\u0145\3\2\2")
 */
  __pyx_tuple__336 = PyTuple_Pack(1, __pyx_kp_u__335); if (unlikely(!__pyx_tuple__336)) __PYX_ERR(0, 211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__336);
  __Pyx_GIVEREF(__pyx_tuple__336);

  /* "pymola/generated/ModelicaParser.pyx":212
 *         buf.write("\2\u0140\u013d\3\2\2\2\u0141\u0144\3\2\2\2\u0142\u0140")
 *         buf.write("\3\2\2\2\u0142\u0143\3\2\2\2\u0143\33\3\2\2\2\u0144\u0142")
 *         buf.write("\3\2\2\2\u0145\u014a\5\"\22\2\u0146\u014a\5&\24\2\u0147")             # <<<<<<<<<<<<<<
 *         buf.write("\u014a\5\36\20\2\u0148\u014a\5 \21\2\u0149\u0145\3\2\2")
 *         buf.write("\2\u0149\u0146\3\2\2\2\u0149\u0147\3\2\2\2\u0149\u0148")
 */
  __pyx_tuple__338 = PyTuple_Pack(1, __pyx_kp_u__337); if (unlikely(!__pyx_tuple__338)) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__338);
  __Pyx_GIVEREF(__pyx_tuple__338);

  /* "pymola/generated/ModelicaParser.pyx":213
 *         buf.write("\3\2\2\2\u0142\u0143\3\2\2\2\u0143\33\3\2\2\2\u0144\u0142")
 *         buf.write("\3\2\2\2\u0145\u014a\5\"\22\2\u0146\u014a\5&\24\2\u0147")
 *         buf.write("\u014a\5\36\20\2\u0148\u014a\5 \21\2\u0149\u0145\3\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u0149\u0146\3\2\2\2\u0149\u0147\3\2\2\2\u0149\u0148")
 *         buf.write("\3\2\2\2\u014a\35\3\2\2\2\u014b\u014d\7V\2\2\u014c\u014b")
 */
  __pyx_tuple__340 = PyTuple_Pack(1, __pyx_kp_u__339); if (unlikely(!__pyx_tuple__340)) __PYX_ERR(0, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__340);
  __Pyx_GIVEREF(__pyx_tuple__340);

  /* "pymola/generated/ModelicaParser.pyx":214
 *         buf.write("\3\2\2\2\u0145\u014a\5\"\22\2\u0146\u014a\5&\24\2\u0147")
 *         buf.write("\u014a\5\36\20\2\u0148\u014a\5 \21\2\u0149\u0145\3\2\2")
 *         buf.write("\2\u0149\u0146\3\2\2\2\u0149\u0147\3\2\2\2\u0149\u0148")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u014a\35\3\2\2\2\u014b\u014d\7V\2\2\u014c\u014b")
 *         buf.write("\3\2\2\2\u014c\u014d\3\2\2\2\u014d\u014f\3\2\2\2\u014e")
 */
  __pyx_tuple__342 = PyTuple_Pack(1, __pyx_kp_u__341); if (unlikely(!__pyx_tuple__342)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__342);
  __Pyx_GIVEREF(__pyx_tuple__342);

  /* "pymola/generated/ModelicaParser.pyx":215
 *         buf.write("\u014a\5\36\20\2\u0148\u014a\5 \21\2\u0149\u0145\3\2\2")
 *         buf.write("\2\u0149\u0146\3\2\2\2\u0149\u0147\3\2\2\2\u0149\u0148")
 *         buf.write("\3\2\2\2\u014a\35\3\2\2\2\u014b\u014d\7V\2\2\u014c\u014b")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u014c\u014d\3\2\2\2\u014d\u014f\3\2\2\2\u014e")
 *         buf.write("\u0150\7S\2\2\u014f\u014e\3\2\2\2\u014f\u0150\3\2\2\2")
 */
  __pyx_tuple__343 = PyTuple_Pack(1, __pyx_kp_u_V); if (unlikely(!__pyx_tuple__343)) __PYX_ERR(0, 215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__343);
  __Pyx_GIVEREF(__pyx_tuple__343);

  /* "pymola/generated/ModelicaParser.pyx":216
 *         buf.write("\2\u0149\u0146\3\2\2\2\u0149\u0147\3\2\2\2\u0149\u0148")
 *         buf.write("\3\2\2\2\u014a\35\3\2\2\2\u014b\u014d\7V\2\2\u014c\u014b")
 *         buf.write("\3\2\2\2\u014c\u014d\3\2\2\2\u014d\u014f\3\2\2\2\u014e")             # <<<<<<<<<<<<<<
 *         buf.write("\u0150\7S\2\2\u014f\u014e\3\2\2\2\u014f\u0150\3\2\2\2")
 *         buf.write("\u0150\u0152\3\2\2\2\u0151\u0153\7W\2\2\u0152\u0151\3")
 */
  __pyx_tuple__345 = PyTuple_Pack(1, __pyx_kp_u__344); if (unlikely(!__pyx_tuple__345)) __PYX_ERR(0, 216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__345);
  __Pyx_GIVEREF(__pyx_tuple__345);

  /* "pymola/generated/ModelicaParser.pyx":217
 *         buf.write("\3\2\2\2\u014a\35\3\2\2\2\u014b\u014d\7V\2\2\u014c\u014b")
 *         buf.write("\3\2\2\2\u014c\u014d\3\2\2\2\u014d\u014f\3\2\2\2\u014e")
 *         buf.write("\u0150\7S\2\2\u014f\u014e\3\2\2\2\u014f\u0150\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0150\u0152\3\2\2\2\u0151\u0153\7W\2\2\u0152\u0151\3")
 *         buf.write("\2\2\2\u0152\u0153\3\2\2\2\u0153\u0155\3\2\2\2\u0154\u0156")
 */
  __pyx_tuple__346 = PyTuple_Pack(1, __pyx_kp_u_S_2); if (unlikely(!__pyx_tuple__346)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__346);
  __Pyx_GIVEREF(__pyx_tuple__346);

  /* "pymola/generated/ModelicaParser.pyx":218
 *         buf.write("\3\2\2\2\u014c\u014d\3\2\2\2\u014d\u014f\3\2\2\2\u014e")
 *         buf.write("\u0150\7S\2\2\u014f\u014e\3\2\2\2\u014f\u0150\3\2\2\2")
 *         buf.write("\u0150\u0152\3\2\2\2\u0151\u0153\7W\2\2\u0152\u0151\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u0152\u0153\3\2\2\2\u0153\u0155\3\2\2\2\u0154\u0156")
 *         buf.write("\7X\2\2\u0155\u0154\3\2\2\2\u0155\u0156\3\2\2\2\u0156")
 */
  __pyx_tuple__347 = PyTuple_Pack(1, __pyx_kp_u_W); if (unlikely(!__pyx_tuple__347)) __PYX_ERR(0, 218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__347);
  __Pyx_GIVEREF(__pyx_tuple__347);

  /* "pymola/generated/ModelicaParser.pyx":219
 *         buf.write("\u0150\7S\2\2\u014f\u014e\3\2\2\2\u014f\u0150\3\2\2\2")
 *         buf.write("\u0150\u0152\3\2\2\2\u0151\u0153\7W\2\2\u0152\u0151\3")
 *         buf.write("\2\2\2\u0152\u0153\3\2\2\2\u0153\u0155\3\2\2\2\u0154\u0156")             # <<<<<<<<<<<<<<
 *         buf.write("\7X\2\2\u0155\u0154\3\2\2\2\u0155\u0156\3\2\2\2\u0156")
 *         buf.write("\u0159\3\2\2\2\u0157\u015a\5\6\4\2\u0158\u015a\5*\26\2")
 */
  __pyx_tuple__349 = PyTuple_Pack(1, __pyx_kp_u__348); if (unlikely(!__pyx_tuple__349)) __PYX_ERR(0, 219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__349);
  __Pyx_GIVEREF(__pyx_tuple__349);

  /* "pymola/generated/ModelicaParser.pyx":220
 *         buf.write("\u0150\u0152\3\2\2\2\u0151\u0153\7W\2\2\u0152\u0151\3")
 *         buf.write("\2\2\2\u0152\u0153\3\2\2\2\u0153\u0155\3\2\2\2\u0154\u0156")
 *         buf.write("\7X\2\2\u0155\u0154\3\2\2\2\u0155\u0156\3\2\2\2\u0156")             # <<<<<<<<<<<<<<
 *         buf.write("\u0159\3\2\2\2\u0157\u015a\5\6\4\2\u0158\u015a\5*\26\2")
 *         buf.write("\u0159\u0157\3\2\2\2\u0159\u0158\3\2\2\2\u015a\37\3\2")
 */
  __pyx_tuple__350 = PyTuple_Pack(1, __pyx_kp_u_X); if (unlikely(!__pyx_tuple__350)) __PYX_ERR(0, 220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__350);
  __Pyx_GIVEREF(__pyx_tuple__350);

  /* "pymola/generated/ModelicaParser.pyx":221
 *         buf.write("\2\2\2\u0152\u0153\3\2\2\2\u0153\u0155\3\2\2\2\u0154\u0156")
 *         buf.write("\7X\2\2\u0155\u0154\3\2\2\2\u0155\u0156\3\2\2\2\u0156")
 *         buf.write("\u0159\3\2\2\2\u0157\u015a\5\6\4\2\u0158\u015a\5*\26\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0159\u0157\3\2\2\2\u0159\u0158\3\2\2\2\u015a\37\3\2")
 *         buf.write("\2\2\u015b\u015d\7V\2\2\u015c\u015b\3\2\2\2\u015c\u015d")
 */
  __pyx_tuple__352 = PyTuple_Pack(1, __pyx_kp_u__351); if (unlikely(!__pyx_tuple__352)) __PYX_ERR(0, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__352);
  __Pyx_GIVEREF(__pyx_tuple__352);

  /* "pymola/generated/ModelicaParser.pyx":222
 *         buf.write("\7X\2\2\u0155\u0154\3\2\2\2\u0155\u0156\3\2\2\2\u0156")
 *         buf.write("\u0159\3\2\2\2\u0157\u015a\5\6\4\2\u0158\u015a\5*\26\2")
 *         buf.write("\u0159\u0157\3\2\2\2\u0159\u0158\3\2\2\2\u015a\37\3\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\u015b\u015d\7V\2\2\u015c\u015b\3\2\2\2\u015c\u015d")
 *         buf.write("\3\2\2\2\u015d\u015f\3\2\2\2\u015e\u0160\7S\2\2\u015f")
 */
  __pyx_tuple__354 = PyTuple_Pack(1, __pyx_kp_u__353); if (unlikely(!__pyx_tuple__354)) __PYX_ERR(0, 222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__354);
  __Pyx_GIVEREF(__pyx_tuple__354);

  /* "pymola/generated/ModelicaParser.pyx":223
 *         buf.write("\u0159\3\2\2\2\u0157\u015a\5\6\4\2\u0158\u015a\5*\26\2")
 *         buf.write("\u0159\u0157\3\2\2\2\u0159\u0158\3\2\2\2\u015a\37\3\2")
 *         buf.write("\2\2\u015b\u015d\7V\2\2\u015c\u015b\3\2\2\2\u015c\u015d")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u015d\u015f\3\2\2\2\u015e\u0160\7S\2\2\u015f")
 *         buf.write("\u015e\3\2\2\2\u015f\u0160\3\2\2\2\u0160\u0162\3\2\2\2")
 */
  __pyx_tuple__355 = PyTuple_Pack(1, __pyx_kp_u_V_2); if (unlikely(!__pyx_tuple__355)) __PYX_ERR(0, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__355);
  __Pyx_GIVEREF(__pyx_tuple__355);

  /* "pymola/generated/ModelicaParser.pyx":224
 *         buf.write("\u0159\u0157\3\2\2\2\u0159\u0158\3\2\2\2\u015a\37\3\2")
 *         buf.write("\2\2\u015b\u015d\7V\2\2\u015c\u015b\3\2\2\2\u015c\u015d")
 *         buf.write("\3\2\2\2\u015d\u015f\3\2\2\2\u015e\u0160\7S\2\2\u015f")             # <<<<<<<<<<<<<<
 *         buf.write("\u015e\3\2\2\2\u015f\u0160\3\2\2\2\u0160\u0162\3\2\2\2")
 *         buf.write("\u0161\u0163\7W\2\2\u0162\u0161\3\2\2\2\u0162\u0163\3")
 */
  __pyx_tuple__356 = PyTuple_Pack(1, __pyx_kp_u_S_3); if (unlikely(!__pyx_tuple__356)) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__356);
  __Pyx_GIVEREF(__pyx_tuple__356);

  /* "pymola/generated/ModelicaParser.pyx":225
 *         buf.write("\2\2\u015b\u015d\7V\2\2\u015c\u015b\3\2\2\2\u015c\u015d")
 *         buf.write("\3\2\2\2\u015d\u015f\3\2\2\2\u015e\u0160\7S\2\2\u015f")
 *         buf.write("\u015e\3\2\2\2\u015f\u0160\3\2\2\2\u0160\u0162\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0161\u0163\7W\2\2\u0162\u0161\3\2\2\2\u0162\u0163\3")
 *         buf.write("\2\2\2\u0163\u0165\3\2\2\2\u0164\u0166\7X\2\2\u0165\u0164")
 */
  __pyx_tuple__358 = PyTuple_Pack(1, __pyx_kp_u__357); if (unlikely(!__pyx_tuple__358)) __PYX_ERR(0, 225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__358);
  __Pyx_GIVEREF(__pyx_tuple__358);

  /* "pymola/generated/ModelicaParser.pyx":226
 *         buf.write("\3\2\2\2\u015d\u015f\3\2\2\2\u015e\u0160\7S\2\2\u015f")
 *         buf.write("\u015e\3\2\2\2\u015f\u0160\3\2\2\2\u0160\u0162\3\2\2\2")
 *         buf.write("\u0161\u0163\7W\2\2\u0162\u0161\3\2\2\2\u0162\u0163\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u0163\u0165\3\2\2\2\u0164\u0166\7X\2\2\u0165\u0164")
 *         buf.write("\3\2\2\2\u0165\u0166\3\2\2\2\u0166\u0167\3\2\2\2\u0167")
 */
  __pyx_tuple__359 = PyTuple_Pack(1, __pyx_kp_u_W_2); if (unlikely(!__pyx_tuple__359)) __PYX_ERR(0, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__359);
  __Pyx_GIVEREF(__pyx_tuple__359);

  /* "pymola/generated/ModelicaParser.pyx":227
 *         buf.write("\u015e\3\2\2\2\u015f\u0160\3\2\2\2\u0160\u0162\3\2\2\2")
 *         buf.write("\u0161\u0163\7W\2\2\u0162\u0161\3\2\2\2\u0162\u0163\3")
 *         buf.write("\2\2\2\u0163\u0165\3\2\2\2\u0164\u0166\7X\2\2\u0165\u0164")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0165\u0166\3\2\2\2\u0166\u0167\3\2\2\2\u0167")
 *         buf.write("\u016a\7\34\2\2\u0168\u016b\5\6\4\2\u0169\u016b\5*\26")
 */
  __pyx_tuple__360 = PyTuple_Pack(1, __pyx_kp_u_X_2); if (unlikely(!__pyx_tuple__360)) __PYX_ERR(0, 227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__360);
  __Pyx_GIVEREF(__pyx_tuple__360);

  /* "pymola/generated/ModelicaParser.pyx":228
 *         buf.write("\u0161\u0163\7W\2\2\u0162\u0161\3\2\2\2\u0162\u0163\3")
 *         buf.write("\2\2\2\u0163\u0165\3\2\2\2\u0164\u0166\7X\2\2\u0165\u0164")
 *         buf.write("\3\2\2\2\u0165\u0166\3\2\2\2\u0166\u0167\3\2\2\2\u0167")             # <<<<<<<<<<<<<<
 *         buf.write("\u016a\7\34\2\2\u0168\u016b\5\6\4\2\u0169\u016b\5*\26")
 *         buf.write("\2\u016a\u0168\3\2\2\2\u016a\u0169\3\2\2\2\u016b\u016f")
 */
  __pyx_tuple__362 = PyTuple_Pack(1, __pyx_kp_u__361); if (unlikely(!__pyx_tuple__362)) __PYX_ERR(0, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__362);
  __Pyx_GIVEREF(__pyx_tuple__362);

  /* "pymola/generated/ModelicaParser.pyx":229
 *         buf.write("\2\2\2\u0163\u0165\3\2\2\2\u0164\u0166\7X\2\2\u0165\u0164")
 *         buf.write("\3\2\2\2\u0165\u0166\3\2\2\2\u0166\u0167\3\2\2\2\u0167")
 *         buf.write("\u016a\7\34\2\2\u0168\u016b\5\6\4\2\u0169\u016b\5*\26")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u016a\u0168\3\2\2\2\u016a\u0169\3\2\2\2\u016b\u016f")
 *         buf.write("\3\2\2\2\u016c\u016d\5(\25\2\u016d\u016e\5\u008eH\2\u016e")
 */
  __pyx_tuple__364 = PyTuple_Pack(1, __pyx_kp_u__363); if (unlikely(!__pyx_tuple__364)) __PYX_ERR(0, 229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__364);
  __Pyx_GIVEREF(__pyx_tuple__364);

  /* "pymola/generated/ModelicaParser.pyx":230
 *         buf.write("\3\2\2\2\u0165\u0166\3\2\2\2\u0166\u0167\3\2\2\2\u0167")
 *         buf.write("\u016a\7\34\2\2\u0168\u016b\5\6\4\2\u0169\u016b\5*\26")
 *         buf.write("\2\u016a\u0168\3\2\2\2\u016a\u0169\3\2\2\2\u016b\u016f")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u016c\u016d\5(\25\2\u016d\u016e\5\u008eH\2\u016e")
 *         buf.write("\u0170\3\2\2\2\u016f\u016c\3\2\2\2\u016f\u0170\3\2\2\2")
 */
  __pyx_tuple__366 = PyTuple_Pack(1, __pyx_kp_u__365); if (unlikely(!__pyx_tuple__366)) __PYX_ERR(0, 230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__366);
  __Pyx_GIVEREF(__pyx_tuple__366);

  /* "pymola/generated/ModelicaParser.pyx":231
 *         buf.write("\u016a\7\34\2\2\u0168\u016b\5\6\4\2\u0169\u016b\5*\26")
 *         buf.write("\2\u016a\u0168\3\2\2\2\u016a\u0169\3\2\2\2\u016b\u016f")
 *         buf.write("\3\2\2\2\u016c\u016d\5(\25\2\u016d\u016e\5\u008eH\2\u016e")             # <<<<<<<<<<<<<<
 *         buf.write("\u0170\3\2\2\2\u016f\u016c\3\2\2\2\u016f\u0170\3\2\2\2")
 *         buf.write("\u0170!\3\2\2\2\u0171\u0180\7\35\2\2\u0172\u0173\7Z\2")
 */
  __pyx_tuple__367 = PyTuple_Pack(1, __pyx_kp_u_H_4); if (unlikely(!__pyx_tuple__367)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__367);
  __Pyx_GIVEREF(__pyx_tuple__367);

  /* "pymola/generated/ModelicaParser.pyx":232
 *         buf.write("\2\u016a\u0168\3\2\2\2\u016a\u0169\3\2\2\2\u016b\u016f")
 *         buf.write("\3\2\2\2\u016c\u016d\5(\25\2\u016d\u016e\5\u008eH\2\u016e")
 *         buf.write("\u0170\3\2\2\2\u016f\u016c\3\2\2\2\u016f\u0170\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0170!\3\2\2\2\u0171\u0180\7\35\2\2\u0172\u0173\7Z\2")
 *         buf.write("\2\u0173\u0174\7\21\2\2\u0174\u0181\5z>\2\u0175\u017e")
 */
  __pyx_tuple__369 = PyTuple_Pack(1, __pyx_kp_u__368); if (unlikely(!__pyx_tuple__369)) __PYX_ERR(0, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__369);
  __Pyx_GIVEREF(__pyx_tuple__369);

  /* "pymola/generated/ModelicaParser.pyx":233
 *         buf.write("\3\2\2\2\u016c\u016d\5(\25\2\u016d\u016e\5\u008eH\2\u016e")
 *         buf.write("\u0170\3\2\2\2\u016f\u016c\3\2\2\2\u016f\u0170\3\2\2\2")
 *         buf.write("\u0170!\3\2\2\2\u0171\u0180\7\35\2\2\u0172\u0173\7Z\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u0173\u0174\7\21\2\2\u0174\u0181\5z>\2\u0175\u017e")
 *         buf.write("\5z>\2\u0176\u017c\7\36\2\2\u0177\u017d\7\37\2\2\u0178")
 */
  __pyx_tuple__370 = PyTuple_Pack(1, __pyx_kp_u_Z_9); if (unlikely(!__pyx_tuple__370)) __PYX_ERR(0, 233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__370);
  __Pyx_GIVEREF(__pyx_tuple__370);

  /* "pymola/generated/ModelicaParser.pyx":234
 *         buf.write("\u0170\3\2\2\2\u016f\u016c\3\2\2\2\u016f\u0170\3\2\2\2")
 *         buf.write("\u0170!\3\2\2\2\u0171\u0180\7\35\2\2\u0172\u0173\7Z\2")
 *         buf.write("\2\u0173\u0174\7\21\2\2\u0174\u0181\5z>\2\u0175\u017e")             # <<<<<<<<<<<<<<
 *         buf.write("\5z>\2\u0176\u017c\7\36\2\2\u0177\u017d\7\37\2\2\u0178")
 *         buf.write("\u0179\7 \2\2\u0179\u017a\5$\23\2\u017a\u017b\7!\2\2\u017b")
 */
  __pyx_tuple__371 = PyTuple_Pack(1, __pyx_kp_u_z_2); if (unlikely(!__pyx_tuple__371)) __PYX_ERR(0, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__371);
  __Pyx_GIVEREF(__pyx_tuple__371);

  /* "pymola/generated/ModelicaParser.pyx":235
 *         buf.write("\u0170!\3\2\2\2\u0171\u0180\7\35\2\2\u0172\u0173\7Z\2")
 *         buf.write("\2\u0173\u0174\7\21\2\2\u0174\u0181\5z>\2\u0175\u017e")
 *         buf.write("\5z>\2\u0176\u017c\7\36\2\2\u0177\u017d\7\37\2\2\u0178")             # <<<<<<<<<<<<<<
 *         buf.write("\u0179\7 \2\2\u0179\u017a\5$\23\2\u017a\u017b\7!\2\2\u017b")
 *         buf.write("\u017d\3\2\2\2\u017c\u0177\3\2\2\2\u017c\u0178\3\2\2\2")
 */
  __pyx_tuple__372 = PyTuple_Pack(1, __pyx_kp_u_z_3); if (unlikely(!__pyx_tuple__372)) __PYX_ERR(0, 235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__372);
  __Pyx_GIVEREF(__pyx_tuple__372);

  /* "pymola/generated/ModelicaParser.pyx":236
 *         buf.write("\2\u0173\u0174\7\21\2\2\u0174\u0181\5z>\2\u0175\u017e")
 *         buf.write("\5z>\2\u0176\u017c\7\36\2\2\u0177\u017d\7\37\2\2\u0178")
 *         buf.write("\u0179\7 \2\2\u0179\u017a\5$\23\2\u017a\u017b\7!\2\2\u017b")             # <<<<<<<<<<<<<<
 *         buf.write("\u017d\3\2\2\2\u017c\u0177\3\2\2\2\u017c\u0178\3\2\2\2")
 *         buf.write("\u017d\u017f\3\2\2\2\u017e\u0176\3\2\2\2\u017e\u017f\3")
 */
  __pyx_tuple__374 = PyTuple_Pack(1, __pyx_kp_u__373); if (unlikely(!__pyx_tuple__374)) __PYX_ERR(0, 236, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__374);
  __Pyx_GIVEREF(__pyx_tuple__374);

  /* "pymola/generated/ModelicaParser.pyx":237
 *         buf.write("\5z>\2\u0176\u017c\7\36\2\2\u0177\u017d\7\37\2\2\u0178")
 *         buf.write("\u0179\7 \2\2\u0179\u017a\5$\23\2\u017a\u017b\7!\2\2\u017b")
 *         buf.write("\u017d\3\2\2\2\u017c\u0177\3\2\2\2\u017c\u0178\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u017d\u017f\3\2\2\2\u017e\u0176\3\2\2\2\u017e\u017f\3")
 *         buf.write("\2\2\2\u017f\u0181\3\2\2\2\u0180\u0172\3\2\2\2\u0180\u0175")
 */
  __pyx_tuple__376 = PyTuple_Pack(1, __pyx_kp_u__375); if (unlikely(!__pyx_tuple__376)) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__376);
  __Pyx_GIVEREF(__pyx_tuple__376);

  /* "pymola/generated/ModelicaParser.pyx":238
 *         buf.write("\u0179\7 \2\2\u0179\u017a\5$\23\2\u017a\u017b\7!\2\2\u017b")
 *         buf.write("\u017d\3\2\2\2\u017c\u0177\3\2\2\2\u017c\u0178\3\2\2\2")
 *         buf.write("\u017d\u017f\3\2\2\2\u017e\u0176\3\2\2\2\u017e\u017f\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u017f\u0181\3\2\2\2\u0180\u0172\3\2\2\2\u0180\u0175")
 *         buf.write("\3\2\2\2\u0181\u0182\3\2\2\2\u0182\u0183\5\u008eH\2\u0183")
 */
  __pyx_tuple__378 = PyTuple_Pack(1, __pyx_kp_u__377); if (unlikely(!__pyx_tuple__378)) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__378);
  __Pyx_GIVEREF(__pyx_tuple__378);

  /* "pymola/generated/ModelicaParser.pyx":239
 *         buf.write("\u017d\3\2\2\2\u017c\u0177\3\2\2\2\u017c\u0178\3\2\2\2")
 *         buf.write("\u017d\u017f\3\2\2\2\u017e\u0176\3\2\2\2\u017e\u017f\3")
 *         buf.write("\2\2\2\u017f\u0181\3\2\2\2\u0180\u0172\3\2\2\2\u0180\u0175")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0181\u0182\3\2\2\2\u0182\u0183\5\u008eH\2\u0183")
 *         buf.write("#\3\2\2\2\u0184\u0189\7Z\2\2\u0185\u0186\7\27\2\2\u0186")
 */
  __pyx_tuple__380 = PyTuple_Pack(1, __pyx_kp_u__379); if (unlikely(!__pyx_tuple__380)) __PYX_ERR(0, 239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__380);
  __Pyx_GIVEREF(__pyx_tuple__380);

  /* "pymola/generated/ModelicaParser.pyx":240
 *         buf.write("\u017d\u017f\3\2\2\2\u017e\u0176\3\2\2\2\u017e\u017f\3")
 *         buf.write("\2\2\2\u017f\u0181\3\2\2\2\u0180\u0172\3\2\2\2\u0180\u0175")
 *         buf.write("\3\2\2\2\u0181\u0182\3\2\2\2\u0182\u0183\5\u008eH\2\u0183")             # <<<<<<<<<<<<<<
 *         buf.write("#\3\2\2\2\u0184\u0189\7Z\2\2\u0185\u0186\7\27\2\2\u0186")
 *         buf.write("\u0188\5$\23\2\u0187\u0185\3\2\2\2\u0188\u018b\3\2\2\2")
 */
  __pyx_tuple__381 = PyTuple_Pack(1, __pyx_kp_u_H_5); if (unlikely(!__pyx_tuple__381)) __PYX_ERR(0, 240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__381);
  __Pyx_GIVEREF(__pyx_tuple__381);

  /* "pymola/generated/ModelicaParser.pyx":241
 *         buf.write("\2\2\2\u017f\u0181\3\2\2\2\u0180\u0172\3\2\2\2\u0180\u0175")
 *         buf.write("\3\2\2\2\u0181\u0182\3\2\2\2\u0182\u0183\5\u008eH\2\u0183")
 *         buf.write("#\3\2\2\2\u0184\u0189\7Z\2\2\u0185\u0186\7\27\2\2\u0186")             # <<<<<<<<<<<<<<
 *         buf.write("\u0188\5$\23\2\u0187\u0185\3\2\2\2\u0188\u018b\3\2\2\2")
 *         buf.write("\u0189\u0187\3\2\2\2\u0189\u018a\3\2\2\2\u018a%\3\2\2")
 */
  __pyx_tuple__382 = PyTuple_Pack(1, __pyx_kp_u_Z_10); if (unlikely(!__pyx_tuple__382)) __PYX_ERR(0, 241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__382);
  __Pyx_GIVEREF(__pyx_tuple__382);

  /* "pymola/generated/ModelicaParser.pyx":242
 *         buf.write("\3\2\2\2\u0181\u0182\3\2\2\2\u0182\u0183\5\u008eH\2\u0183")
 *         buf.write("#\3\2\2\2\u0184\u0189\7Z\2\2\u0185\u0186\7\27\2\2\u0186")
 *         buf.write("\u0188\5$\23\2\u0187\u0185\3\2\2\2\u0188\u018b\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0189\u0187\3\2\2\2\u0189\u018a\3\2\2\2\u018a%\3\2\2")
 *         buf.write("\2\u018b\u0189\3\2\2\2\u018c\u018d\7\30\2\2\u018d\u018f")
 */
  __pyx_tuple__384 = PyTuple_Pack(1, __pyx_kp_u__383); if (unlikely(!__pyx_tuple__384)) __PYX_ERR(0, 242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__384);
  __Pyx_GIVEREF(__pyx_tuple__384);

  /* "pymola/generated/ModelicaParser.pyx":243
 *         buf.write("#\3\2\2\2\u0184\u0189\7Z\2\2\u0185\u0186\7\27\2\2\u0186")
 *         buf.write("\u0188\5$\23\2\u0187\u0185\3\2\2\2\u0188\u018b\3\2\2\2")
 *         buf.write("\u0189\u0187\3\2\2\2\u0189\u018a\3\2\2\2\u018a%\3\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u018b\u0189\3\2\2\2\u018c\u018d\7\30\2\2\u018d\u018f")
 *         buf.write("\5z>\2\u018e\u0190\5:\36\2\u018f\u018e\3\2\2\2\u018f\u0190")
 */
  __pyx_tuple__386 = PyTuple_Pack(1, __pyx_kp_u__385); if (unlikely(!__pyx_tuple__386)) __PYX_ERR(0, 243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__386);
  __Pyx_GIVEREF(__pyx_tuple__386);

  /* "pymola/generated/ModelicaParser.pyx":244
 *         buf.write("\u0188\5$\23\2\u0187\u0185\3\2\2\2\u0188\u018b\3\2\2\2")
 *         buf.write("\u0189\u0187\3\2\2\2\u0189\u018a\3\2\2\2\u018a%\3\2\2")
 *         buf.write("\2\u018b\u0189\3\2\2\2\u018c\u018d\7\30\2\2\u018d\u018f")             # <<<<<<<<<<<<<<
 *         buf.write("\5z>\2\u018e\u0190\5:\36\2\u018f\u018e\3\2\2\2\u018f\u0190")
 *         buf.write("\3\2\2\2\u0190\u0192\3\2\2\2\u0191\u0193\5\u0092J\2\u0192")
 */
  __pyx_tuple__388 = PyTuple_Pack(1, __pyx_kp_u__387); if (unlikely(!__pyx_tuple__388)) __PYX_ERR(0, 244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__388);
  __Pyx_GIVEREF(__pyx_tuple__388);

  /* "pymola/generated/ModelicaParser.pyx":245
 *         buf.write("\u0189\u0187\3\2\2\2\u0189\u018a\3\2\2\2\u018a%\3\2\2")
 *         buf.write("\2\u018b\u0189\3\2\2\2\u018c\u018d\7\30\2\2\u018d\u018f")
 *         buf.write("\5z>\2\u018e\u0190\5:\36\2\u018f\u018e\3\2\2\2\u018f\u0190")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0190\u0192\3\2\2\2\u0191\u0193\5\u0092J\2\u0192")
 *         buf.write("\u0191\3\2\2\2\u0192\u0193\3\2\2\2\u0193\'\3\2\2\2\u0194")
 */
  __pyx_tuple__389 = PyTuple_Pack(1, __pyx_kp_u_z_4); if (unlikely(!__pyx_tuple__389)) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__389);
  __Pyx_GIVEREF(__pyx_tuple__389);

  /* "pymola/generated/ModelicaParser.pyx":246
 *         buf.write("\2\u018b\u0189\3\2\2\2\u018c\u018d\7\30\2\2\u018d\u018f")
 *         buf.write("\5z>\2\u018e\u0190\5:\36\2\u018f\u018e\3\2\2\2\u018f\u0190")
 *         buf.write("\3\2\2\2\u0190\u0192\3\2\2\2\u0191\u0193\5\u0092J\2\u0192")             # <<<<<<<<<<<<<<
 *         buf.write("\u0191\3\2\2\2\u0192\u0193\3\2\2\2\u0193\'\3\2\2\2\u0194")
 *         buf.write("\u0195\7\"\2\2\u0195\u0197\5v<\2\u0196\u0198\5:\36\2\u0197")
 */
  __pyx_tuple__390 = PyTuple_Pack(1, __pyx_kp_u_J_3); if (unlikely(!__pyx_tuple__390)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__390);
  __Pyx_GIVEREF(__pyx_tuple__390);

  /* "pymola/generated/ModelicaParser.pyx":247
 *         buf.write("\5z>\2\u018e\u0190\5:\36\2\u018f\u018e\3\2\2\2\u018f\u0190")
 *         buf.write("\3\2\2\2\u0190\u0192\3\2\2\2\u0191\u0193\5\u0092J\2\u0192")
 *         buf.write("\u0191\3\2\2\2\u0192\u0193\3\2\2\2\u0193\'\3\2\2\2\u0194")             # <<<<<<<<<<<<<<
 *         buf.write("\u0195\7\"\2\2\u0195\u0197\5v<\2\u0196\u0198\5:\36\2\u0197")
 *         buf.write("\u0196\3\2\2\2\u0197\u0198\3\2\2\2\u0198)\3\2\2\2\u0199")
 */
  __pyx_tuple__392 = PyTuple_Pack(1, __pyx_kp_u__391); if (unlikely(!__pyx_tuple__392)) __PYX_ERR(0, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__392);
  __Pyx_GIVEREF(__pyx_tuple__392);

  /* "pymola/generated/ModelicaParser.pyx":248
 *         buf.write("\3\2\2\2\u0190\u0192\3\2\2\2\u0191\u0193\5\u0092J\2\u0192")
 *         buf.write("\u0191\3\2\2\2\u0192\u0193\3\2\2\2\u0193\'\3\2\2\2\u0194")
 *         buf.write("\u0195\7\"\2\2\u0195\u0197\5v<\2\u0196\u0198\5:\36\2\u0197")             # <<<<<<<<<<<<<<
 *         buf.write("\u0196\3\2\2\2\u0197\u0198\3\2\2\2\u0198)\3\2\2\2\u0199")
 *         buf.write("\u019a\5,\27\2\u019a\u019c\5.\30\2\u019b\u019d\5\u008a")
 */
  __pyx_tuple__393 = PyTuple_Pack(1, __pyx_kp_u_v_2); if (unlikely(!__pyx_tuple__393)) __PYX_ERR(0, 248, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__393);
  __Pyx_GIVEREF(__pyx_tuple__393);

  /* "pymola/generated/ModelicaParser.pyx":249
 *         buf.write("\u0191\3\2\2\2\u0192\u0193\3\2\2\2\u0193\'\3\2\2\2\u0194")
 *         buf.write("\u0195\7\"\2\2\u0195\u0197\5v<\2\u0196\u0198\5:\36\2\u0197")
 *         buf.write("\u0196\3\2\2\2\u0197\u0198\3\2\2\2\u0198)\3\2\2\2\u0199")             # <<<<<<<<<<<<<<
 *         buf.write("\u019a\5,\27\2\u019a\u019c\5.\30\2\u019b\u019d\5\u008a")
 *         buf.write("F\2\u019c\u019b\3\2\2\2\u019c\u019d\3\2\2\2\u019d\u019e")
 */
  __pyx_tuple__395 = PyTuple_Pack(1, __pyx_kp_u__394); if (unlikely(!__pyx_tuple__395)) __PYX_ERR(0, 249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__395);
  __Pyx_GIVEREF(__pyx_tuple__395);

  /* "pymola/generated/ModelicaParser.pyx":250
 *         buf.write("\u0195\7\"\2\2\u0195\u0197\5v<\2\u0196\u0198\5:\36\2\u0197")
 *         buf.write("\u0196\3\2\2\2\u0197\u0198\3\2\2\2\u0198)\3\2\2\2\u0199")
 *         buf.write("\u019a\5,\27\2\u019a\u019c\5.\30\2\u019b\u019d\5\u008a")             # <<<<<<<<<<<<<<
 *         buf.write("F\2\u019c\u019b\3\2\2\2\u019c\u019d\3\2\2\2\u019d\u019e")
 *         buf.write("\3\2\2\2\u019e\u019f\5\60\31\2\u019f+\3\2\2\2\u01a0\u01a2")
 */
  __pyx_tuple__397 = PyTuple_Pack(1, __pyx_kp_u__396); if (unlikely(!__pyx_tuple__397)) __PYX_ERR(0, 250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__397);
  __Pyx_GIVEREF(__pyx_tuple__397);

  /* "pymola/generated/ModelicaParser.pyx":251
 *         buf.write("\u0196\3\2\2\2\u0197\u0198\3\2\2\2\u0198)\3\2\2\2\u0199")
 *         buf.write("\u019a\5,\27\2\u019a\u019c\5.\30\2\u019b\u019d\5\u008a")
 *         buf.write("F\2\u019c\u019b\3\2\2\2\u019c\u019d\3\2\2\2\u019d\u019e")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u019e\u019f\5\60\31\2\u019f+\3\2\2\2\u01a0\u01a2")
 *         buf.write("\t\3\2\2\u01a1\u01a0\3\2\2\2\u01a1\u01a2\3\2\2\2\u01a2")
 */
  __pyx_tuple__398 = PyTuple_Pack(1, __pyx_kp_u_F); if (unlikely(!__pyx_tuple__398)) __PYX_ERR(0, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__398);
  __Pyx_GIVEREF(__pyx_tuple__398);

  /* "pymola/generated/ModelicaParser.pyx":252
 *         buf.write("\u019a\5,\27\2\u019a\u019c\5.\30\2\u019b\u019d\5\u008a")
 *         buf.write("F\2\u019c\u019b\3\2\2\2\u019c\u019d\3\2\2\2\u019d\u019e")
 *         buf.write("\3\2\2\2\u019e\u019f\5\60\31\2\u019f+\3\2\2\2\u01a0\u01a2")             # <<<<<<<<<<<<<<
 *         buf.write("\t\3\2\2\u01a1\u01a0\3\2\2\2\u01a1\u01a2\3\2\2\2\u01a2")
 *         buf.write("\u01a4\3\2\2\2\u01a3\u01a5\t\4\2\2\u01a4\u01a3\3\2\2\2")
 */
  __pyx_tuple__399 = PyTuple_Pack(1, __pyx_kp_u_0); if (unlikely(!__pyx_tuple__399)) __PYX_ERR(0, 252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__399);
  __Pyx_GIVEREF(__pyx_tuple__399);

  /* "pymola/generated/ModelicaParser.pyx":253
 *         buf.write("F\2\u019c\u019b\3\2\2\2\u019c\u019d\3\2\2\2\u019d\u019e")
 *         buf.write("\3\2\2\2\u019e\u019f\5\60\31\2\u019f+\3\2\2\2\u01a0\u01a2")
 *         buf.write("\t\3\2\2\u01a1\u01a0\3\2\2\2\u01a1\u01a2\3\2\2\2\u01a2")             # <<<<<<<<<<<<<<
 *         buf.write("\u01a4\3\2\2\2\u01a3\u01a5\t\4\2\2\u01a4\u01a3\3\2\2\2")
 *         buf.write("\u01a4\u01a5\3\2\2\2\u01a5\u01a7\3\2\2\2\u01a6\u01a8\t")
 */
  __pyx_tuple__401 = PyTuple_Pack(1, __pyx_kp_u__400); if (unlikely(!__pyx_tuple__401)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__401);
  __Pyx_GIVEREF(__pyx_tuple__401);

  /* "pymola/generated/ModelicaParser.pyx":254
 *         buf.write("\3\2\2\2\u019e\u019f\5\60\31\2\u019f+\3\2\2\2\u01a0\u01a2")
 *         buf.write("\t\3\2\2\u01a1\u01a0\3\2\2\2\u01a1\u01a2\3\2\2\2\u01a2")
 *         buf.write("\u01a4\3\2\2\2\u01a3\u01a5\t\4\2\2\u01a4\u01a3\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u01a4\u01a5\3\2\2\2\u01a5\u01a7\3\2\2\2\u01a6\u01a8\t")
 *         buf.write("\5\2\2\u01a7\u01a6\3\2\2\2\u01a7\u01a8\3\2\2\2\u01a8-")
 */
  __pyx_tuple__403 = PyTuple_Pack(1, __pyx_kp_u__402); if (unlikely(!__pyx_tuple__403)) __PYX_ERR(0, 254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__403);
  __Pyx_GIVEREF(__pyx_tuple__403);

  /* "pymola/generated/ModelicaParser.pyx":255
 *         buf.write("\t\3\2\2\u01a1\u01a0\3\2\2\2\u01a1\u01a2\3\2\2\2\u01a2")
 *         buf.write("\u01a4\3\2\2\2\u01a3\u01a5\t\4\2\2\u01a4\u01a3\3\2\2\2")
 *         buf.write("\u01a4\u01a5\3\2\2\2\u01a5\u01a7\3\2\2\2\u01a6\u01a8\t")             # <<<<<<<<<<<<<<
 *         buf.write("\5\2\2\u01a7\u01a6\3\2\2\2\u01a7\u01a8\3\2\2\2\u01a8-")
 *         buf.write("\3\2\2\2\u01a9\u01aa\5z>\2\u01aa/\3\2\2\2\u01ab\u01b0")
 */
  __pyx_tuple__405 = PyTuple_Pack(1, __pyx_kp_u__404); if (unlikely(!__pyx_tuple__405)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__405);
  __Pyx_GIVEREF(__pyx_tuple__405);

  /* "pymola/generated/ModelicaParser.pyx":256
 *         buf.write("\u01a4\3\2\2\2\u01a3\u01a5\t\4\2\2\u01a4\u01a3\3\2\2\2")
 *         buf.write("\u01a4\u01a5\3\2\2\2\u01a5\u01a7\3\2\2\2\u01a6\u01a8\t")
 *         buf.write("\5\2\2\u01a7\u01a6\3\2\2\2\u01a7\u01a8\3\2\2\2\u01a8-")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u01a9\u01aa\5z>\2\u01aa/\3\2\2\2\u01ab\u01b0")
 *         buf.write("\5\62\32\2\u01ac\u01ad\7\27\2\2\u01ad\u01af\5\62\32\2")
 */
  __pyx_tuple__407 = PyTuple_Pack(1, __pyx_kp_u__406); if (unlikely(!__pyx_tuple__407)) __PYX_ERR(0, 256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__407);
  __Pyx_GIVEREF(__pyx_tuple__407);

  /* "pymola/generated/ModelicaParser.pyx":257
 *         buf.write("\u01a4\u01a5\3\2\2\2\u01a5\u01a7\3\2\2\2\u01a6\u01a8\t")
 *         buf.write("\5\2\2\u01a7\u01a6\3\2\2\2\u01a7\u01a8\3\2\2\2\u01a8-")
 *         buf.write("\3\2\2\2\u01a9\u01aa\5z>\2\u01aa/\3\2\2\2\u01ab\u01b0")             # <<<<<<<<<<<<<<
 *         buf.write("\5\62\32\2\u01ac\u01ad\7\27\2\2\u01ad\u01af\5\62\32\2")
 *         buf.write("\u01ae\u01ac\3\2\2\2\u01af\u01b2\3\2\2\2\u01b0\u01ae\3")
 */
  __pyx_tuple__408 = PyTuple_Pack(1, __pyx_kp_u_z_5); if (unlikely(!__pyx_tuple__408)) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__408);
  __Pyx_GIVEREF(__pyx_tuple__408);

  /* "pymola/generated/ModelicaParser.pyx":258
 *         buf.write("\5\2\2\u01a7\u01a6\3\2\2\2\u01a7\u01a8\3\2\2\2\u01a8-")
 *         buf.write("\3\2\2\2\u01a9\u01aa\5z>\2\u01aa/\3\2\2\2\u01ab\u01b0")
 *         buf.write("\5\62\32\2\u01ac\u01ad\7\27\2\2\u01ad\u01af\5\62\32\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u01ae\u01ac\3\2\2\2\u01af\u01b2\3\2\2\2\u01b0\u01ae\3")
 *         buf.write("\2\2\2\u01b0\u01b1\3\2\2\2\u01b1\61\3\2\2\2\u01b2\u01b0")
 */
  __pyx_tuple__409 = PyTuple_Pack(1, __pyx_kp_u_2_2); if (unlikely(!__pyx_tuple__409)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__409);
  __Pyx_GIVEREF(__pyx_tuple__409);

  /* "pymola/generated/ModelicaParser.pyx":259
 *         buf.write("\3\2\2\2\u01a9\u01aa\5z>\2\u01aa/\3\2\2\2\u01ab\u01b0")
 *         buf.write("\5\62\32\2\u01ac\u01ad\7\27\2\2\u01ad\u01af\5\62\32\2")
 *         buf.write("\u01ae\u01ac\3\2\2\2\u01af\u01b2\3\2\2\2\u01b0\u01ae\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u01b0\u01b1\3\2\2\2\u01b1\61\3\2\2\2\u01b2\u01b0")
 *         buf.write("\3\2\2\2\u01b3\u01b5\5\66\34\2\u01b4\u01b6\5\64\33\2\u01b5")
 */
  __pyx_tuple__411 = PyTuple_Pack(1, __pyx_kp_u__410); if (unlikely(!__pyx_tuple__411)) __PYX_ERR(0, 259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__411);
  __Pyx_GIVEREF(__pyx_tuple__411);

  /* "pymola/generated/ModelicaParser.pyx":260
 *         buf.write("\5\62\32\2\u01ac\u01ad\7\27\2\2\u01ad\u01af\5\62\32\2")
 *         buf.write("\u01ae\u01ac\3\2\2\2\u01af\u01b2\3\2\2\2\u01b0\u01ae\3")
 *         buf.write("\2\2\2\u01b0\u01b1\3\2\2\2\u01b1\61\3\2\2\2\u01b2\u01b0")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u01b3\u01b5\5\66\34\2\u01b4\u01b6\5\64\33\2\u01b5")
 *         buf.write("\u01b4\3\2\2\2\u01b5\u01b6\3\2\2\2\u01b6\u01b7\3\2\2\2")
 */
  __pyx_tuple__412 = PyTuple_Pack(1, __pyx_kp_u_1); if (unlikely(!__pyx_tuple__412)) __PYX_ERR(0, 260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__412);
  __Pyx_GIVEREF(__pyx_tuple__412);

  /* "pymola/generated/ModelicaParser.pyx":261
 *         buf.write("\u01ae\u01ac\3\2\2\2\u01af\u01b2\3\2\2\2\u01b0\u01ae\3")
 *         buf.write("\2\2\2\u01b0\u01b1\3\2\2\2\u01b1\61\3\2\2\2\u01b2\u01b0")
 *         buf.write("\3\2\2\2\u01b3\u01b5\5\66\34\2\u01b4\u01b6\5\64\33\2\u01b5")             # <<<<<<<<<<<<<<
 *         buf.write("\u01b4\3\2\2\2\u01b5\u01b6\3\2\2\2\u01b6\u01b7\3\2\2\2")
 *         buf.write("\u01b7\u01b8\5\u008eH\2\u01b8\63\3\2\2\2\u01b9\u01ba\7")
 */
  __pyx_tuple__413 = PyTuple_Pack(1, __pyx_kp_u_6_4); if (unlikely(!__pyx_tuple__413)) __PYX_ERR(0, 261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__413);
  __Pyx_GIVEREF(__pyx_tuple__413);

  /* "pymola/generated/ModelicaParser.pyx":262
 *         buf.write("\2\2\2\u01b0\u01b1\3\2\2\2\u01b1\61\3\2\2\2\u01b2\u01b0")
 *         buf.write("\3\2\2\2\u01b3\u01b5\5\66\34\2\u01b4\u01b6\5\64\33\2\u01b5")
 *         buf.write("\u01b4\3\2\2\2\u01b5\u01b6\3\2\2\2\u01b6\u01b7\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u01b7\u01b8\5\u008eH\2\u01b8\63\3\2\2\2\u01b9\u01ba\7")
 *         buf.write("*\2\2\u01ba\u01bb\5n8\2\u01bb\65\3\2\2\2\u01bc\u01be\7")
 */
  __pyx_tuple__415 = PyTuple_Pack(1, __pyx_kp_u__414); if (unlikely(!__pyx_tuple__415)) __PYX_ERR(0, 262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__415);
  __Pyx_GIVEREF(__pyx_tuple__415);

  /* "pymola/generated/ModelicaParser.pyx":263
 *         buf.write("\3\2\2\2\u01b3\u01b5\5\66\34\2\u01b4\u01b6\5\64\33\2\u01b5")
 *         buf.write("\u01b4\3\2\2\2\u01b5\u01b6\3\2\2\2\u01b6\u01b7\3\2\2\2")
 *         buf.write("\u01b7\u01b8\5\u008eH\2\u01b8\63\3\2\2\2\u01b9\u01ba\7")             # <<<<<<<<<<<<<<
 *         buf.write("*\2\2\u01ba\u01bb\5n8\2\u01bb\65\3\2\2\2\u01bc\u01be\7")
 *         buf.write("Z\2\2\u01bd\u01bf\5\u008aF\2\u01be\u01bd\3\2\2\2\u01be")
 */
  __pyx_tuple__416 = PyTuple_Pack(1, __pyx_kp_u_H_3_2); if (unlikely(!__pyx_tuple__416)) __PYX_ERR(0, 263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__416);
  __Pyx_GIVEREF(__pyx_tuple__416);

  /* "pymola/generated/ModelicaParser.pyx":264
 *         buf.write("\u01b4\3\2\2\2\u01b5\u01b6\3\2\2\2\u01b6\u01b7\3\2\2\2")
 *         buf.write("\u01b7\u01b8\5\u008eH\2\u01b8\63\3\2\2\2\u01b9\u01ba\7")
 *         buf.write("*\2\2\u01ba\u01bb\5n8\2\u01bb\65\3\2\2\2\u01bc\u01be\7")             # <<<<<<<<<<<<<<
 *         buf.write("Z\2\2\u01bd\u01bf\5\u008aF\2\u01be\u01bd\3\2\2\2\u01be")
 *         buf.write("\u01bf\3\2\2\2\u01bf\u01c1\3\2\2\2\u01c0\u01c2\58\35\2")
 */
  __pyx_tuple__417 = PyTuple_Pack(1, __pyx_kp_u_n8_5); if (unlikely(!__pyx_tuple__417)) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__417);
  __Pyx_GIVEREF(__pyx_tuple__417);

  /* "pymola/generated/ModelicaParser.pyx":265
 *         buf.write("\u01b7\u01b8\5\u008eH\2\u01b8\63\3\2\2\2\u01b9\u01ba\7")
 *         buf.write("*\2\2\u01ba\u01bb\5n8\2\u01bb\65\3\2\2\2\u01bc\u01be\7")
 *         buf.write("Z\2\2\u01bd\u01bf\5\u008aF\2\u01be\u01bd\3\2\2\2\u01be")             # <<<<<<<<<<<<<<
 *         buf.write("\u01bf\3\2\2\2\u01bf\u01c1\3\2\2\2\u01c0\u01c2\58\35\2")
 *         buf.write("\u01c1\u01c0\3\2\2\2\u01c1\u01c2\3\2\2\2\u01c2\67\3\2")
 */
  __pyx_tuple__418 = PyTuple_Pack(1, __pyx_kp_u_Z_F); if (unlikely(!__pyx_tuple__418)) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__418);
  __Pyx_GIVEREF(__pyx_tuple__418);

  /* "pymola/generated/ModelicaParser.pyx":266
 *         buf.write("*\2\2\u01ba\u01bb\5n8\2\u01bb\65\3\2\2\2\u01bc\u01be\7")
 *         buf.write("Z\2\2\u01bd\u01bf\5\u008aF\2\u01be\u01bd\3\2\2\2\u01be")
 *         buf.write("\u01bf\3\2\2\2\u01bf\u01c1\3\2\2\2\u01c0\u01c2\58\35\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u01c1\u01c0\3\2\2\2\u01c1\u01c2\3\2\2\2\u01c2\67\3\2")
 *         buf.write("\2\2\u01c3\u01c6\5:\36\2\u01c4\u01c5\7\21\2\2\u01c5\u01c7")
 */
  __pyx_tuple__419 = PyTuple_Pack(1, __pyx_kp_u_8); if (unlikely(!__pyx_tuple__419)) __PYX_ERR(0, 266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__419);
  __Pyx_GIVEREF(__pyx_tuple__419);

  /* "pymola/generated/ModelicaParser.pyx":267
 *         buf.write("Z\2\2\u01bd\u01bf\5\u008aF\2\u01be\u01bd\3\2\2\2\u01be")
 *         buf.write("\u01bf\3\2\2\2\u01bf\u01c1\3\2\2\2\u01c0\u01c2\58\35\2")
 *         buf.write("\u01c1\u01c0\3\2\2\2\u01c1\u01c2\3\2\2\2\u01c2\67\3\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\u01c3\u01c6\5:\36\2\u01c4\u01c5\7\21\2\2\u01c5\u01c7")
 *         buf.write("\5n8\2\u01c6\u01c4\3\2\2\2\u01c6\u01c7\3\2\2\2\u01c7\u01cd")
 */
  __pyx_tuple__420 = PyTuple_Pack(1, __pyx_kp_u_7); if (unlikely(!__pyx_tuple__420)) __PYX_ERR(0, 267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__420);
  __Pyx_GIVEREF(__pyx_tuple__420);

  /* "pymola/generated/ModelicaParser.pyx":268
 *         buf.write("\u01bf\3\2\2\2\u01bf\u01c1\3\2\2\2\u01c0\u01c2\58\35\2")
 *         buf.write("\u01c1\u01c0\3\2\2\2\u01c1\u01c2\3\2\2\2\u01c2\67\3\2")
 *         buf.write("\2\2\u01c3\u01c6\5:\36\2\u01c4\u01c5\7\21\2\2\u01c5\u01c7")             # <<<<<<<<<<<<<<
 *         buf.write("\5n8\2\u01c6\u01c4\3\2\2\2\u01c6\u01c7\3\2\2\2\u01c7\u01cd")
 *         buf.write("\3\2\2\2\u01c8\u01c9\7\21\2\2\u01c9\u01cd\5n8\2\u01ca")
 */
  __pyx_tuple__422 = PyTuple_Pack(1, __pyx_kp_u__421); if (unlikely(!__pyx_tuple__422)) __PYX_ERR(0, 268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__422);
  __Pyx_GIVEREF(__pyx_tuple__422);

  /* "pymola/generated/ModelicaParser.pyx":269
 *         buf.write("\u01c1\u01c0\3\2\2\2\u01c1\u01c2\3\2\2\2\u01c2\67\3\2")
 *         buf.write("\2\2\u01c3\u01c6\5:\36\2\u01c4\u01c5\7\21\2\2\u01c5\u01c7")
 *         buf.write("\5n8\2\u01c6\u01c4\3\2\2\2\u01c6\u01c7\3\2\2\2\u01c7\u01cd")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u01c8\u01c9\7\21\2\2\u01c9\u01cd\5n8\2\u01ca")
 *         buf.write("\u01cb\7+\2\2\u01cb\u01cd\5n8\2\u01cc\u01c3\3\2\2\2\u01cc")
 */
  __pyx_tuple__423 = PyTuple_Pack(1, __pyx_kp_u_n8); if (unlikely(!__pyx_tuple__423)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__423);
  __Pyx_GIVEREF(__pyx_tuple__423);

  /* "pymola/generated/ModelicaParser.pyx":270
 *         buf.write("\2\2\u01c3\u01c6\5:\36\2\u01c4\u01c5\7\21\2\2\u01c5\u01c7")
 *         buf.write("\5n8\2\u01c6\u01c4\3\2\2\2\u01c6\u01c7\3\2\2\2\u01c7\u01cd")
 *         buf.write("\3\2\2\2\u01c8\u01c9\7\21\2\2\u01c9\u01cd\5n8\2\u01ca")             # <<<<<<<<<<<<<<
 *         buf.write("\u01cb\7+\2\2\u01cb\u01cd\5n8\2\u01cc\u01c3\3\2\2\2\u01cc")
 *         buf.write("\u01c8\3\2\2\2\u01cc\u01ca\3\2\2\2\u01cd9\3\2\2\2\u01ce")
 */
  __pyx_tuple__424 = PyTuple_Pack(1, __pyx_kp_u_n8_2); if (unlikely(!__pyx_tuple__424)) __PYX_ERR(0, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__424);
  __Pyx_GIVEREF(__pyx_tuple__424);

  /* "pymola/generated/ModelicaParser.pyx":271
 *         buf.write("\5n8\2\u01c6\u01c4\3\2\2\2\u01c6\u01c7\3\2\2\2\u01c7\u01cd")
 *         buf.write("\3\2\2\2\u01c8\u01c9\7\21\2\2\u01c9\u01cd\5n8\2\u01ca")
 *         buf.write("\u01cb\7+\2\2\u01cb\u01cd\5n8\2\u01cc\u01c3\3\2\2\2\u01cc")             # <<<<<<<<<<<<<<
 *         buf.write("\u01c8\3\2\2\2\u01cc\u01ca\3\2\2\2\u01cd9\3\2\2\2\u01ce")
 *         buf.write("\u01d0\7\23\2\2\u01cf\u01d1\5<\37\2\u01d0\u01cf\3\2\2")
 */
  __pyx_tuple__425 = PyTuple_Pack(1, __pyx_kp_u_n8_3); if (unlikely(!__pyx_tuple__425)) __PYX_ERR(0, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__425);
  __Pyx_GIVEREF(__pyx_tuple__425);

  /* "pymola/generated/ModelicaParser.pyx":272
 *         buf.write("\3\2\2\2\u01c8\u01c9\7\21\2\2\u01c9\u01cd\5n8\2\u01ca")
 *         buf.write("\u01cb\7+\2\2\u01cb\u01cd\5n8\2\u01cc\u01c3\3\2\2\2\u01cc")
 *         buf.write("\u01c8\3\2\2\2\u01cc\u01ca\3\2\2\2\u01cd9\3\2\2\2\u01ce")             # <<<<<<<<<<<<<<
 *         buf.write("\u01d0\7\23\2\2\u01cf\u01d1\5<\37\2\u01d0\u01cf\3\2\2")
 *         buf.write("\2\u01d0\u01d1\3\2\2\2\u01d1\u01d2\3\2\2\2\u01d2\u01d3")
 */
  __pyx_tuple__426 = PyTuple_Pack(1, __pyx_kp_u_9); if (unlikely(!__pyx_tuple__426)) __PYX_ERR(0, 272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__426);
  __Pyx_GIVEREF(__pyx_tuple__426);

  /* "pymola/generated/ModelicaParser.pyx":273
 *         buf.write("\u01cb\7+\2\2\u01cb\u01cd\5n8\2\u01cc\u01c3\3\2\2\2\u01cc")
 *         buf.write("\u01c8\3\2\2\2\u01cc\u01ca\3\2\2\2\u01cd9\3\2\2\2\u01ce")
 *         buf.write("\u01d0\7\23\2\2\u01cf\u01d1\5<\37\2\u01d0\u01cf\3\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u01d0\u01d1\3\2\2\2\u01d1\u01d2\3\2\2\2\u01d2\u01d3")
 *         buf.write("\7\25\2\2\u01d3;\3\2\2\2\u01d4\u01d9\5> \2\u01d5\u01d6")
 */
  __pyx_tuple__428 = PyTuple_Pack(1, __pyx_kp_u__427); if (unlikely(!__pyx_tuple__428)) __PYX_ERR(0, 273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__428);
  __Pyx_GIVEREF(__pyx_tuple__428);

  /* "pymola/generated/ModelicaParser.pyx":274
 *         buf.write("\u01c8\3\2\2\2\u01cc\u01ca\3\2\2\2\u01cd9\3\2\2\2\u01ce")
 *         buf.write("\u01d0\7\23\2\2\u01cf\u01d1\5<\37\2\u01d0\u01cf\3\2\2")
 *         buf.write("\2\u01d0\u01d1\3\2\2\2\u01d1\u01d2\3\2\2\2\u01d2\u01d3")             # <<<<<<<<<<<<<<
 *         buf.write("\7\25\2\2\u01d3;\3\2\2\2\u01d4\u01d9\5> \2\u01d5\u01d6")
 *         buf.write("\7\27\2\2\u01d6\u01d8\5> \2\u01d7\u01d5\3\2\2\2\u01d8")
 */
  __pyx_tuple__430 = PyTuple_Pack(1, __pyx_kp_u__429); if (unlikely(!__pyx_tuple__430)) __PYX_ERR(0, 274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__430);
  __Pyx_GIVEREF(__pyx_tuple__430);

  /* "pymola/generated/ModelicaParser.pyx":275
 *         buf.write("\u01d0\7\23\2\2\u01cf\u01d1\5<\37\2\u01d0\u01cf\3\2\2")
 *         buf.write("\2\u01d0\u01d1\3\2\2\2\u01d1\u01d2\3\2\2\2\u01d2\u01d3")
 *         buf.write("\7\25\2\2\u01d3;\3\2\2\2\u01d4\u01d9\5> \2\u01d5\u01d6")             # <<<<<<<<<<<<<<
 *         buf.write("\7\27\2\2\u01d6\u01d8\5> \2\u01d7\u01d5\3\2\2\2\u01d8")
 *         buf.write("\u01db\3\2\2\2\u01d9\u01d7\3\2\2\2\u01d9\u01da\3\2\2\2")
 */
  __pyx_tuple__432 = PyTuple_Pack(1, __pyx_kp_u__431); if (unlikely(!__pyx_tuple__432)) __PYX_ERR(0, 275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__432);
  __Pyx_GIVEREF(__pyx_tuple__432);

  /* "pymola/generated/ModelicaParser.pyx":276
 *         buf.write("\2\u01d0\u01d1\3\2\2\2\u01d1\u01d2\3\2\2\2\u01d2\u01d3")
 *         buf.write("\7\25\2\2\u01d3;\3\2\2\2\u01d4\u01d9\5> \2\u01d5\u01d6")
 *         buf.write("\7\27\2\2\u01d6\u01d8\5> \2\u01d7\u01d5\3\2\2\2\u01d8")             # <<<<<<<<<<<<<<
 *         buf.write("\u01db\3\2\2\2\u01d9\u01d7\3\2\2\2\u01d9\u01da\3\2\2\2")
 *         buf.write("\u01da=\3\2\2\2\u01db\u01d9\3\2\2\2\u01dc\u01df\5@!\2")
 */
  __pyx_tuple__434 = PyTuple_Pack(1, __pyx_kp_u__433); if (unlikely(!__pyx_tuple__434)) __PYX_ERR(0, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__434);
  __Pyx_GIVEREF(__pyx_tuple__434);

  /* "pymola/generated/ModelicaParser.pyx":277
 *         buf.write("\7\25\2\2\u01d3;\3\2\2\2\u01d4\u01d9\5> \2\u01d5\u01d6")
 *         buf.write("\7\27\2\2\u01d6\u01d8\5> \2\u01d7\u01d5\3\2\2\2\u01d8")
 *         buf.write("\u01db\3\2\2\2\u01d9\u01d7\3\2\2\2\u01d9\u01da\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u01da=\3\2\2\2\u01db\u01d9\3\2\2\2\u01dc\u01df\5@!\2")
 *         buf.write("\u01dd\u01df\5D#\2\u01de\u01dc\3\2\2\2\u01de\u01dd\3\2")
 */
  __pyx_tuple__436 = PyTuple_Pack(1, __pyx_kp_u__435); if (unlikely(!__pyx_tuple__436)) __PYX_ERR(0, 277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__436);
  __Pyx_GIVEREF(__pyx_tuple__436);

  /* "pymola/generated/ModelicaParser.pyx":278
 *         buf.write("\7\27\2\2\u01d6\u01d8\5> \2\u01d7\u01d5\3\2\2\2\u01d8")
 *         buf.write("\u01db\3\2\2\2\u01d9\u01d7\3\2\2\2\u01d9\u01da\3\2\2\2")
 *         buf.write("\u01da=\3\2\2\2\u01db\u01d9\3\2\2\2\u01dc\u01df\5@!\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u01dd\u01df\5D#\2\u01de\u01dc\3\2\2\2\u01de\u01dd\3\2")
 *         buf.write("\2\2\u01df?\3\2\2\2\u01e0\u01e2\7Q\2\2\u01e1\u01e0\3\2")
 */
  __pyx_tuple__438 = PyTuple_Pack(1, __pyx_kp_u__437); if (unlikely(!__pyx_tuple__438)) __PYX_ERR(0, 278, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__438);
  __Pyx_GIVEREF(__pyx_tuple__438);

  /* "pymola/generated/ModelicaParser.pyx":279
 *         buf.write("\u01db\3\2\2\2\u01d9\u01d7\3\2\2\2\u01d9\u01da\3\2\2\2")
 *         buf.write("\u01da=\3\2\2\2\u01db\u01d9\3\2\2\2\u01dc\u01df\5@!\2")
 *         buf.write("\u01dd\u01df\5D#\2\u01de\u01dc\3\2\2\2\u01de\u01dd\3\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\u01df?\3\2\2\2\u01e0\u01e2\7Q\2\2\u01e1\u01e0\3\2")
 *         buf.write("\2\2\u01e1\u01e2\3\2\2\2\u01e2\u01e4\3\2\2\2\u01e3\u01e5")
 */
  __pyx_tuple__439 = PyTuple_Pack(1, __pyx_kp_u_D); if (unlikely(!__pyx_tuple__439)) __PYX_ERR(0, 279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__439);
  __Pyx_GIVEREF(__pyx_tuple__439);

  /* "pymola/generated/ModelicaParser.pyx":280
 *         buf.write("\u01da=\3\2\2\2\u01db\u01d9\3\2\2\2\u01dc\u01df\5@!\2")
 *         buf.write("\u01dd\u01df\5D#\2\u01de\u01dc\3\2\2\2\u01de\u01dd\3\2")
 *         buf.write("\2\2\u01df?\3\2\2\2\u01e0\u01e2\7Q\2\2\u01e1\u01e0\3\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\u01e1\u01e2\3\2\2\2\u01e2\u01e4\3\2\2\2\u01e3\u01e5")
 *         buf.write("\7S\2\2\u01e4\u01e3\3\2\2\2\u01e4\u01e5\3\2\2\2\u01e5")
 */
  __pyx_tuple__440 = PyTuple_Pack(1, __pyx_kp_u_Q); if (unlikely(!__pyx_tuple__440)) __PYX_ERR(0, 280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__440);
  __Pyx_GIVEREF(__pyx_tuple__440);

  /* "pymola/generated/ModelicaParser.pyx":281
 *         buf.write("\u01dd\u01df\5D#\2\u01de\u01dc\3\2\2\2\u01de\u01dd\3\2")
 *         buf.write("\2\2\u01df?\3\2\2\2\u01e0\u01e2\7Q\2\2\u01e1\u01e0\3\2")
 *         buf.write("\2\2\u01e1\u01e2\3\2\2\2\u01e2\u01e4\3\2\2\2\u01e3\u01e5")             # <<<<<<<<<<<<<<
 *         buf.write("\7S\2\2\u01e4\u01e3\3\2\2\2\u01e4\u01e5\3\2\2\2\u01e5")
 *         buf.write("\u01e8\3\2\2\2\u01e6\u01e9\5B\"\2\u01e7\u01e9\5F$\2\u01e8")
 */
  __pyx_tuple__442 = PyTuple_Pack(1, __pyx_kp_u__441); if (unlikely(!__pyx_tuple__442)) __PYX_ERR(0, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__442);
  __Pyx_GIVEREF(__pyx_tuple__442);

  /* "pymola/generated/ModelicaParser.pyx":282
 *         buf.write("\2\2\u01df?\3\2\2\2\u01e0\u01e2\7Q\2\2\u01e1\u01e0\3\2")
 *         buf.write("\2\2\u01e1\u01e2\3\2\2\2\u01e2\u01e4\3\2\2\2\u01e3\u01e5")
 *         buf.write("\7S\2\2\u01e4\u01e3\3\2\2\2\u01e4\u01e5\3\2\2\2\u01e5")             # <<<<<<<<<<<<<<
 *         buf.write("\u01e8\3\2\2\2\u01e6\u01e9\5B\"\2\u01e7\u01e9\5F$\2\u01e8")
 *         buf.write("\u01e6\3\2\2\2\u01e8\u01e7\3\2\2\2\u01e9A\3\2\2\2\u01ea")
 */
  __pyx_tuple__443 = PyTuple_Pack(1, __pyx_kp_u_S_4); if (unlikely(!__pyx_tuple__443)) __PYX_ERR(0, 282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__443);
  __Pyx_GIVEREF(__pyx_tuple__443);

  /* "pymola/generated/ModelicaParser.pyx":283
 *         buf.write("\2\2\u01e1\u01e2\3\2\2\2\u01e2\u01e4\3\2\2\2\u01e3\u01e5")
 *         buf.write("\7S\2\2\u01e4\u01e3\3\2\2\2\u01e4\u01e5\3\2\2\2\u01e5")
 *         buf.write("\u01e8\3\2\2\2\u01e6\u01e9\5B\"\2\u01e7\u01e9\5F$\2\u01e8")             # <<<<<<<<<<<<<<
 *         buf.write("\u01e6\3\2\2\2\u01e8\u01e7\3\2\2\2\u01e9A\3\2\2\2\u01ea")
 *         buf.write("\u01ec\5z>\2\u01eb\u01ed\58\35\2\u01ec\u01eb\3\2\2\2\u01ec")
 */
  __pyx_tuple__444 = PyTuple_Pack(1, __pyx_kp_u_B_F); if (unlikely(!__pyx_tuple__444)) __PYX_ERR(0, 283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__444);
  __Pyx_GIVEREF(__pyx_tuple__444);

  /* "pymola/generated/ModelicaParser.pyx":284
 *         buf.write("\7S\2\2\u01e4\u01e3\3\2\2\2\u01e4\u01e5\3\2\2\2\u01e5")
 *         buf.write("\u01e8\3\2\2\2\u01e6\u01e9\5B\"\2\u01e7\u01e9\5F$\2\u01e8")
 *         buf.write("\u01e6\3\2\2\2\u01e8\u01e7\3\2\2\2\u01e9A\3\2\2\2\u01ea")             # <<<<<<<<<<<<<<
 *         buf.write("\u01ec\5z>\2\u01eb\u01ed\58\35\2\u01ec\u01eb\3\2\2\2\u01ec")
 *         buf.write("\u01ed\3\2\2\2\u01ed\u01ee\3\2\2\2\u01ee\u01ef\5\u0090")
 */
  __pyx_tuple__445 = PyTuple_Pack(1, __pyx_kp_u_A_2); if (unlikely(!__pyx_tuple__445)) __PYX_ERR(0, 284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__445);
  __Pyx_GIVEREF(__pyx_tuple__445);

  /* "pymola/generated/ModelicaParser.pyx":285
 *         buf.write("\u01e8\3\2\2\2\u01e6\u01e9\5B\"\2\u01e7\u01e9\5F$\2\u01e8")
 *         buf.write("\u01e6\3\2\2\2\u01e8\u01e7\3\2\2\2\u01e9A\3\2\2\2\u01ea")
 *         buf.write("\u01ec\5z>\2\u01eb\u01ed\58\35\2\u01ec\u01eb\3\2\2\2\u01ec")             # <<<<<<<<<<<<<<
 *         buf.write("\u01ed\3\2\2\2\u01ed\u01ee\3\2\2\2\u01ee\u01ef\5\u0090")
 *         buf.write("I\2\u01efC\3\2\2\2\u01f0\u01f2\7V\2\2\u01f1\u01f3\7Q\2")
 */
  __pyx_tuple__446 = PyTuple_Pack(1, __pyx_kp_u_z_8); if (unlikely(!__pyx_tuple__446)) __PYX_ERR(0, 285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__446);
  __Pyx_GIVEREF(__pyx_tuple__446);

  /* "pymola/generated/ModelicaParser.pyx":286
 *         buf.write("\u01e6\3\2\2\2\u01e8\u01e7\3\2\2\2\u01e9A\3\2\2\2\u01ea")
 *         buf.write("\u01ec\5z>\2\u01eb\u01ed\58\35\2\u01ec\u01eb\3\2\2\2\u01ec")
 *         buf.write("\u01ed\3\2\2\2\u01ed\u01ee\3\2\2\2\u01ee\u01ef\5\u0090")             # <<<<<<<<<<<<<<
 *         buf.write("I\2\u01efC\3\2\2\2\u01f0\u01f2\7V\2\2\u01f1\u01f3\7Q\2")
 *         buf.write("\2\u01f2\u01f1\3\2\2\2\u01f2\u01f3\3\2\2\2\u01f3\u01f5")
 */
  __pyx_tuple__448 = PyTuple_Pack(1, __pyx_kp_u__447); if (unlikely(!__pyx_tuple__448)) __PYX_ERR(0, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__448);
  __Pyx_GIVEREF(__pyx_tuple__448);

  /* "pymola/generated/ModelicaParser.pyx":287
 *         buf.write("\u01ec\5z>\2\u01eb\u01ed\58\35\2\u01ec\u01eb\3\2\2\2\u01ec")
 *         buf.write("\u01ed\3\2\2\2\u01ed\u01ee\3\2\2\2\u01ee\u01ef\5\u0090")
 *         buf.write("I\2\u01efC\3\2\2\2\u01f0\u01f2\7V\2\2\u01f1\u01f3\7Q\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u01f2\u01f1\3\2\2\2\u01f2\u01f3\3\2\2\2\u01f3\u01f5")
 *         buf.write("\3\2\2\2\u01f4\u01f6\7S\2\2\u01f5\u01f4\3\2\2\2\u01f5")
 */
  __pyx_tuple__449 = PyTuple_Pack(1, __pyx_kp_u_I_C_V_Q); if (unlikely(!__pyx_tuple__449)) __PYX_ERR(0, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__449);
  __Pyx_GIVEREF(__pyx_tuple__449);

  /* "pymola/generated/ModelicaParser.pyx":288
 *         buf.write("\u01ed\3\2\2\2\u01ed\u01ee\3\2\2\2\u01ee\u01ef\5\u0090")
 *         buf.write("I\2\u01efC\3\2\2\2\u01f0\u01f2\7V\2\2\u01f1\u01f3\7Q\2")
 *         buf.write("\2\u01f2\u01f1\3\2\2\2\u01f2\u01f3\3\2\2\2\u01f3\u01f5")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u01f4\u01f6\7S\2\2\u01f5\u01f4\3\2\2\2\u01f5")
 *         buf.write("\u01f6\3\2\2\2\u01f6\u01fc\3\2\2\2\u01f7\u01fa\5L\'\2")
 */
  __pyx_tuple__451 = PyTuple_Pack(1, __pyx_kp_u__450); if (unlikely(!__pyx_tuple__451)) __PYX_ERR(0, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__451);
  __Pyx_GIVEREF(__pyx_tuple__451);

  /* "pymola/generated/ModelicaParser.pyx":289
 *         buf.write("I\2\u01efC\3\2\2\2\u01f0\u01f2\7V\2\2\u01f1\u01f3\7Q\2")
 *         buf.write("\2\u01f2\u01f1\3\2\2\2\u01f2\u01f3\3\2\2\2\u01f3\u01f5")
 *         buf.write("\3\2\2\2\u01f4\u01f6\7S\2\2\u01f5\u01f4\3\2\2\2\u01f5")             # <<<<<<<<<<<<<<
 *         buf.write("\u01f6\3\2\2\2\u01f6\u01fc\3\2\2\2\u01f7\u01fa\5L\'\2")
 *         buf.write("\u01f8\u01fa\5H%\2\u01f9\u01f7\3\2\2\2\u01f9\u01f8\3\2")
 */
  __pyx_tuple__452 = PyTuple_Pack(1, __pyx_kp_u_S_5); if (unlikely(!__pyx_tuple__452)) __PYX_ERR(0, 289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__452);
  __Pyx_GIVEREF(__pyx_tuple__452);

  /* "pymola/generated/ModelicaParser.pyx":290
 *         buf.write("\2\u01f2\u01f1\3\2\2\2\u01f2\u01f3\3\2\2\2\u01f3\u01f5")
 *         buf.write("\3\2\2\2\u01f4\u01f6\7S\2\2\u01f5\u01f4\3\2\2\2\u01f5")
 *         buf.write("\u01f6\3\2\2\2\u01f6\u01fc\3\2\2\2\u01f7\u01fa\5L\'\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u01f8\u01fa\5H%\2\u01f9\u01f7\3\2\2\2\u01f9\u01f8\3\2")
 *         buf.write("\2\2\u01fa\u01fd\3\2\2\2\u01fb\u01fd\5F$\2\u01fc\u01f9")
 */
  __pyx_tuple__453 = PyTuple_Pack(1, __pyx_kp_u_L); if (unlikely(!__pyx_tuple__453)) __PYX_ERR(0, 290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__453);
  __Pyx_GIVEREF(__pyx_tuple__453);

  /* "pymola/generated/ModelicaParser.pyx":291
 *         buf.write("\3\2\2\2\u01f4\u01f6\7S\2\2\u01f5\u01f4\3\2\2\2\u01f5")
 *         buf.write("\u01f6\3\2\2\2\u01f6\u01fc\3\2\2\2\u01f7\u01fa\5L\'\2")
 *         buf.write("\u01f8\u01fa\5H%\2\u01f9\u01f7\3\2\2\2\u01f9\u01f8\3\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\u01fa\u01fd\3\2\2\2\u01fb\u01fd\5F$\2\u01fc\u01f9")
 *         buf.write("\3\2\2\2\u01fc\u01fb\3\2\2\2\u01fdE\3\2\2\2\u01fe\u0201")
 */
  __pyx_tuple__454 = PyTuple_Pack(1, __pyx_kp_u_H_6); if (unlikely(!__pyx_tuple__454)) __PYX_ERR(0, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__454);
  __Pyx_GIVEREF(__pyx_tuple__454);

  /* "pymola/generated/ModelicaParser.pyx":292
 *         buf.write("\u01f6\3\2\2\2\u01f6\u01fc\3\2\2\2\u01f7\u01fa\5L\'\2")
 *         buf.write("\u01f8\u01fa\5H%\2\u01f9\u01f7\3\2\2\2\u01f9\u01f8\3\2")
 *         buf.write("\2\2\u01fa\u01fd\3\2\2\2\u01fb\u01fd\5F$\2\u01fc\u01f9")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u01fc\u01fb\3\2\2\2\u01fdE\3\2\2\2\u01fe\u0201")
 *         buf.write("\7\34\2\2\u01ff\u0202\5L\'\2\u0200\u0202\5H%\2\u0201\u01ff")
 */
  __pyx_tuple__455 = PyTuple_Pack(1, __pyx_kp_u_F_2); if (unlikely(!__pyx_tuple__455)) __PYX_ERR(0, 292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__455);
  __Pyx_GIVEREF(__pyx_tuple__455);

  /* "pymola/generated/ModelicaParser.pyx":293
 *         buf.write("\u01f8\u01fa\5H%\2\u01f9\u01f7\3\2\2\2\u01f9\u01f8\3\2")
 *         buf.write("\2\2\u01fa\u01fd\3\2\2\2\u01fb\u01fd\5F$\2\u01fc\u01f9")
 *         buf.write("\3\2\2\2\u01fc\u01fb\3\2\2\2\u01fdE\3\2\2\2\u01fe\u0201")             # <<<<<<<<<<<<<<
 *         buf.write("\7\34\2\2\u01ff\u0202\5L\'\2\u0200\u0202\5H%\2\u0201\u01ff")
 *         buf.write("\3\2\2\2\u0201\u0200\3\2\2\2\u0202\u0204\3\2\2\2\u0203")
 */
  __pyx_tuple__456 = PyTuple_Pack(1, __pyx_kp_u_E_2); if (unlikely(!__pyx_tuple__456)) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__456);
  __Pyx_GIVEREF(__pyx_tuple__456);

  /* "pymola/generated/ModelicaParser.pyx":294
 *         buf.write("\2\2\u01fa\u01fd\3\2\2\2\u01fb\u01fd\5F$\2\u01fc\u01f9")
 *         buf.write("\3\2\2\2\u01fc\u01fb\3\2\2\2\u01fdE\3\2\2\2\u01fe\u0201")
 *         buf.write("\7\34\2\2\u01ff\u0202\5L\'\2\u0200\u0202\5H%\2\u0201\u01ff")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0201\u0200\3\2\2\2\u0202\u0204\3\2\2\2\u0203")
 *         buf.write("\u0205\5(\25\2\u0204\u0203\3\2\2\2\u0204\u0205\3\2\2\2")
 */
  __pyx_tuple__457 = PyTuple_Pack(1, __pyx_kp_u_L_H); if (unlikely(!__pyx_tuple__457)) __PYX_ERR(0, 294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__457);
  __Pyx_GIVEREF(__pyx_tuple__457);

  /* "pymola/generated/ModelicaParser.pyx":295
 *         buf.write("\3\2\2\2\u01fc\u01fb\3\2\2\2\u01fdE\3\2\2\2\u01fe\u0201")
 *         buf.write("\7\34\2\2\u01ff\u0202\5L\'\2\u0200\u0202\5H%\2\u0201\u01ff")
 *         buf.write("\3\2\2\2\u0201\u0200\3\2\2\2\u0202\u0204\3\2\2\2\u0203")             # <<<<<<<<<<<<<<
 *         buf.write("\u0205\5(\25\2\u0204\u0203\3\2\2\2\u0204\u0205\3\2\2\2")
 *         buf.write("\u0205G\3\2\2\2\u0206\u0207\5,\27\2\u0207\u0208\5.\30")
 */
  __pyx_tuple__459 = PyTuple_Pack(1, __pyx_kp_u__458); if (unlikely(!__pyx_tuple__459)) __PYX_ERR(0, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__459);
  __Pyx_GIVEREF(__pyx_tuple__459);

  /* "pymola/generated/ModelicaParser.pyx":296
 *         buf.write("\7\34\2\2\u01ff\u0202\5L\'\2\u0200\u0202\5H%\2\u0201\u01ff")
 *         buf.write("\3\2\2\2\u0201\u0200\3\2\2\2\u0202\u0204\3\2\2\2\u0203")
 *         buf.write("\u0205\5(\25\2\u0204\u0203\3\2\2\2\u0204\u0205\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0205G\3\2\2\2\u0206\u0207\5,\27\2\u0207\u0208\5.\30")
 *         buf.write("\2\u0208\u0209\5J&\2\u0209I\3\2\2\2\u020a\u020b\5\66\34")
 */
  __pyx_tuple__461 = PyTuple_Pack(1, __pyx_kp_u__460); if (unlikely(!__pyx_tuple__461)) __PYX_ERR(0, 296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__461);
  __Pyx_GIVEREF(__pyx_tuple__461);

  /* "pymola/generated/ModelicaParser.pyx":297
 *         buf.write("\3\2\2\2\u0201\u0200\3\2\2\2\u0202\u0204\3\2\2\2\u0203")
 *         buf.write("\u0205\5(\25\2\u0204\u0203\3\2\2\2\u0204\u0205\3\2\2\2")
 *         buf.write("\u0205G\3\2\2\2\u0206\u0207\5,\27\2\u0207\u0208\5.\30")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u0208\u0209\5J&\2\u0209I\3\2\2\2\u020a\u020b\5\66\34")
 *         buf.write("\2\u020b\u020c\5\u008eH\2\u020cK\3\2\2\2\u020d\u020e\5")
 */
  __pyx_tuple__462 = PyTuple_Pack(1, __pyx_kp_u_G); if (unlikely(!__pyx_tuple__462)) __PYX_ERR(0, 297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__462);
  __Pyx_GIVEREF(__pyx_tuple__462);

  /* "pymola/generated/ModelicaParser.pyx":298
 *         buf.write("\u0205\5(\25\2\u0204\u0203\3\2\2\2\u0204\u0205\3\2\2\2")
 *         buf.write("\u0205G\3\2\2\2\u0206\u0207\5,\27\2\u0207\u0208\5.\30")
 *         buf.write("\2\u0208\u0209\5J&\2\u0209I\3\2\2\2\u020a\u020b\5\66\34")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u020b\u020c\5\u008eH\2\u020cK\3\2\2\2\u020d\u020e\5")
 *         buf.write("\b\5\2\u020e\u020f\7Z\2\2\u020f\u0224\7\21\2\2\u0210\u0211")
 */
  __pyx_tuple__463 = PyTuple_Pack(1, __pyx_kp_u_J_I_6); if (unlikely(!__pyx_tuple__463)) __PYX_ERR(0, 298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__463);
  __Pyx_GIVEREF(__pyx_tuple__463);

  /* "pymola/generated/ModelicaParser.pyx":299
 *         buf.write("\u0205G\3\2\2\2\u0206\u0207\5,\27\2\u0207\u0208\5.\30")
 *         buf.write("\2\u0208\u0209\5J&\2\u0209I\3\2\2\2\u020a\u020b\5\66\34")
 *         buf.write("\2\u020b\u020c\5\u008eH\2\u020cK\3\2\2\2\u020d\u020e\5")             # <<<<<<<<<<<<<<
 *         buf.write("\b\5\2\u020e\u020f\7Z\2\2\u020f\u0224\7\21\2\2\u0210\u0211")
 *         buf.write("\5\16\b\2\u0211\u0213\5z>\2\u0212\u0214\5\u008aF\2\u0213")
 */
  __pyx_tuple__464 = PyTuple_Pack(1, __pyx_kp_u_H_K); if (unlikely(!__pyx_tuple__464)) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__464);
  __Pyx_GIVEREF(__pyx_tuple__464);

  /* "pymola/generated/ModelicaParser.pyx":300
 *         buf.write("\2\u0208\u0209\5J&\2\u0209I\3\2\2\2\u020a\u020b\5\66\34")
 *         buf.write("\2\u020b\u020c\5\u008eH\2\u020cK\3\2\2\2\u020d\u020e\5")
 *         buf.write("\b\5\2\u020e\u020f\7Z\2\2\u020f\u0224\7\21\2\2\u0210\u0211")             # <<<<<<<<<<<<<<
 *         buf.write("\5\16\b\2\u0211\u0213\5z>\2\u0212\u0214\5\u008aF\2\u0213")
 *         buf.write("\u0212\3\2\2\2\u0213\u0214\3\2\2\2\u0214\u0216\3\2\2\2")
 */
  __pyx_tuple__465 = PyTuple_Pack(1, __pyx_kp_u_Z_11); if (unlikely(!__pyx_tuple__465)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__465);
  __Pyx_GIVEREF(__pyx_tuple__465);

  /* "pymola/generated/ModelicaParser.pyx":301
 *         buf.write("\2\u020b\u020c\5\u008eH\2\u020cK\3\2\2\2\u020d\u020e\5")
 *         buf.write("\b\5\2\u020e\u020f\7Z\2\2\u020f\u0224\7\21\2\2\u0210\u0211")
 *         buf.write("\5\16\b\2\u0211\u0213\5z>\2\u0212\u0214\5\u008aF\2\u0213")             # <<<<<<<<<<<<<<
 *         buf.write("\u0212\3\2\2\2\u0213\u0214\3\2\2\2\u0214\u0216\3\2\2\2")
 *         buf.write("\u0215\u0217\5:\36\2\u0216\u0215\3\2\2\2\u0216\u0217\3")
 */
  __pyx_tuple__466 = PyTuple_Pack(1, __pyx_kp_u_z_F); if (unlikely(!__pyx_tuple__466)) __PYX_ERR(0, 301, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__466);
  __Pyx_GIVEREF(__pyx_tuple__466);

  /* "pymola/generated/ModelicaParser.pyx":302
 *         buf.write("\b\5\2\u020e\u020f\7Z\2\2\u020f\u0224\7\21\2\2\u0210\u0211")
 *         buf.write("\5\16\b\2\u0211\u0213\5z>\2\u0212\u0214\5\u008aF\2\u0213")
 *         buf.write("\u0212\3\2\2\2\u0213\u0214\3\2\2\2\u0214\u0216\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0215\u0217\5:\36\2\u0216\u0215\3\2\2\2\u0216\u0217\3")
 *         buf.write("\2\2\2\u0217\u0218\3\2\2\2\u0218\u0219\5\u008eH\2\u0219")
 */
  __pyx_tuple__468 = PyTuple_Pack(1, __pyx_kp_u__467); if (unlikely(!__pyx_tuple__468)) __PYX_ERR(0, 302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__468);
  __Pyx_GIVEREF(__pyx_tuple__468);

  /* "pymola/generated/ModelicaParser.pyx":303
 *         buf.write("\5\16\b\2\u0211\u0213\5z>\2\u0212\u0214\5\u008aF\2\u0213")
 *         buf.write("\u0212\3\2\2\2\u0213\u0214\3\2\2\2\u0214\u0216\3\2\2\2")
 *         buf.write("\u0215\u0217\5:\36\2\u0216\u0215\3\2\2\2\u0216\u0217\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u0217\u0218\3\2\2\2\u0218\u0219\5\u008eH\2\u0219")
 *         buf.write("\u0225\3\2\2\2\u021a\u021b\7\22\2\2\u021b\u0220\7\23\2")
 */
  __pyx_tuple__470 = PyTuple_Pack(1, __pyx_kp_u__469); if (unlikely(!__pyx_tuple__470)) __PYX_ERR(0, 303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__470);
  __Pyx_GIVEREF(__pyx_tuple__470);

  /* "pymola/generated/ModelicaParser.pyx":304
 *         buf.write("\u0212\3\2\2\2\u0213\u0214\3\2\2\2\u0214\u0216\3\2\2\2")
 *         buf.write("\u0215\u0217\5:\36\2\u0216\u0215\3\2\2\2\u0216\u0217\3")
 *         buf.write("\2\2\2\u0217\u0218\3\2\2\2\u0218\u0219\5\u008eH\2\u0219")             # <<<<<<<<<<<<<<
 *         buf.write("\u0225\3\2\2\2\u021a\u021b\7\22\2\2\u021b\u0220\7\23\2")
 *         buf.write("\2\u021c\u021e\5\20\t\2\u021d\u021c\3\2\2\2\u021d\u021e")
 */
  __pyx_tuple__471 = PyTuple_Pack(1, __pyx_kp_u_H_7); if (unlikely(!__pyx_tuple__471)) __PYX_ERR(0, 304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__471);
  __Pyx_GIVEREF(__pyx_tuple__471);

  /* "pymola/generated/ModelicaParser.pyx":305
 *         buf.write("\u0215\u0217\5:\36\2\u0216\u0215\3\2\2\2\u0216\u0217\3")
 *         buf.write("\2\2\2\u0217\u0218\3\2\2\2\u0218\u0219\5\u008eH\2\u0219")
 *         buf.write("\u0225\3\2\2\2\u021a\u021b\7\22\2\2\u021b\u0220\7\23\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u021c\u021e\5\20\t\2\u021d\u021c\3\2\2\2\u021d\u021e")
 *         buf.write("\3\2\2\2\u021e\u0221\3\2\2\2\u021f\u0221\7\24\2\2\u0220")
 */
  __pyx_tuple__473 = PyTuple_Pack(1, __pyx_kp_u__472); if (unlikely(!__pyx_tuple__473)) __PYX_ERR(0, 305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__473);
  __Pyx_GIVEREF(__pyx_tuple__473);

  /* "pymola/generated/ModelicaParser.pyx":306
 *         buf.write("\2\2\2\u0217\u0218\3\2\2\2\u0218\u0219\5\u008eH\2\u0219")
 *         buf.write("\u0225\3\2\2\2\u021a\u021b\7\22\2\2\u021b\u0220\7\23\2")
 *         buf.write("\2\u021c\u021e\5\20\t\2\u021d\u021c\3\2\2\2\u021d\u021e")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u021e\u0221\3\2\2\2\u021f\u0221\7\24\2\2\u0220")
 *         buf.write("\u021d\3\2\2\2\u0220\u021f\3\2\2\2\u0221\u0222\3\2\2\2")
 */
  __pyx_tuple__475 = PyTuple_Pack(1, __pyx_kp_u__474); if (unlikely(!__pyx_tuple__475)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__475);
  __Pyx_GIVEREF(__pyx_tuple__475);

  /* "pymola/generated/ModelicaParser.pyx":307
 *         buf.write("\u0225\3\2\2\2\u021a\u021b\7\22\2\2\u021b\u0220\7\23\2")
 *         buf.write("\2\u021c\u021e\5\20\t\2\u021d\u021c\3\2\2\2\u021d\u021e")
 *         buf.write("\3\2\2\2\u021e\u0221\3\2\2\2\u021f\u0221\7\24\2\2\u0220")             # <<<<<<<<<<<<<<
 *         buf.write("\u021d\3\2\2\2\u0220\u021f\3\2\2\2\u0221\u0222\3\2\2\2")
 *         buf.write("\u0222\u0223\7\25\2\2\u0223\u0225\5\u008eH\2\u0224\u0210")
 */
  __pyx_tuple__477 = PyTuple_Pack(1, __pyx_kp_u__476); if (unlikely(!__pyx_tuple__477)) __PYX_ERR(0, 307, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__477);
  __Pyx_GIVEREF(__pyx_tuple__477);

  /* "pymola/generated/ModelicaParser.pyx":308
 *         buf.write("\2\u021c\u021e\5\20\t\2\u021d\u021c\3\2\2\2\u021d\u021e")
 *         buf.write("\3\2\2\2\u021e\u0221\3\2\2\2\u021f\u0221\7\24\2\2\u0220")
 *         buf.write("\u021d\3\2\2\2\u0220\u021f\3\2\2\2\u0221\u0222\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0222\u0223\7\25\2\2\u0223\u0225\5\u008eH\2\u0224\u0210")
 *         buf.write("\3\2\2\2\u0224\u021a\3\2\2\2\u0225M\3\2\2\2\u0226\u0228")
 */
  __pyx_tuple__479 = PyTuple_Pack(1, __pyx_kp_u__478); if (unlikely(!__pyx_tuple__479)) __PYX_ERR(0, 308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__479);
  __Pyx_GIVEREF(__pyx_tuple__479);

  /* "pymola/generated/ModelicaParser.pyx":309
 *         buf.write("\3\2\2\2\u021e\u0221\3\2\2\2\u021f\u0221\7\24\2\2\u0220")
 *         buf.write("\u021d\3\2\2\2\u0220\u021f\3\2\2\2\u0221\u0222\3\2\2\2")
 *         buf.write("\u0222\u0223\7\25\2\2\u0223\u0225\5\u008eH\2\u0224\u0210")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0224\u021a\3\2\2\2\u0225M\3\2\2\2\u0226\u0228")
 *         buf.write("\7Y\2\2\u0227\u0226\3\2\2\2\u0227\u0228\3\2\2\2\u0228")
 */
  __pyx_tuple__480 = PyTuple_Pack(1, __pyx_kp_u_H_8); if (unlikely(!__pyx_tuple__480)) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__480);
  __Pyx_GIVEREF(__pyx_tuple__480);

  /* "pymola/generated/ModelicaParser.pyx":310
 *         buf.write("\u021d\3\2\2\2\u0220\u021f\3\2\2\2\u0221\u0222\3\2\2\2")
 *         buf.write("\u0222\u0223\7\25\2\2\u0223\u0225\5\u008eH\2\u0224\u0210")
 *         buf.write("\3\2\2\2\u0224\u021a\3\2\2\2\u0225M\3\2\2\2\u0226\u0228")             # <<<<<<<<<<<<<<
 *         buf.write("\7Y\2\2\u0227\u0226\3\2\2\2\u0227\u0228\3\2\2\2\u0228")
 *         buf.write("\u0229\3\2\2\2\u0229\u022f\7,\2\2\u022a\u022b\5T+\2\u022b")
 */
  __pyx_tuple__481 = PyTuple_Pack(1, __pyx_kp_u_M); if (unlikely(!__pyx_tuple__481)) __PYX_ERR(0, 310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__481);
  __Pyx_GIVEREF(__pyx_tuple__481);

  /* "pymola/generated/ModelicaParser.pyx":311
 *         buf.write("\u0222\u0223\7\25\2\2\u0223\u0225\5\u008eH\2\u0224\u0210")
 *         buf.write("\3\2\2\2\u0224\u021a\3\2\2\2\u0225M\3\2\2\2\u0226\u0228")
 *         buf.write("\7Y\2\2\u0227\u0226\3\2\2\2\u0227\u0228\3\2\2\2\u0228")             # <<<<<<<<<<<<<<
 *         buf.write("\u0229\3\2\2\2\u0229\u022f\7,\2\2\u022a\u022b\5T+\2\u022b")
 *         buf.write("\u022c\7\3\2\2\u022c\u022e\3\2\2\2\u022d\u022a\3\2\2\2")
 */
  __pyx_tuple__482 = PyTuple_Pack(1, __pyx_kp_u_Y); if (unlikely(!__pyx_tuple__482)) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__482);
  __Pyx_GIVEREF(__pyx_tuple__482);

  /* "pymola/generated/ModelicaParser.pyx":312
 *         buf.write("\3\2\2\2\u0224\u021a\3\2\2\2\u0225M\3\2\2\2\u0226\u0228")
 *         buf.write("\7Y\2\2\u0227\u0226\3\2\2\2\u0227\u0228\3\2\2\2\u0228")
 *         buf.write("\u0229\3\2\2\2\u0229\u022f\7,\2\2\u022a\u022b\5T+\2\u022b")             # <<<<<<<<<<<<<<
 *         buf.write("\u022c\7\3\2\2\u022c\u022e\3\2\2\2\u022d\u022a\3\2\2\2")
 *         buf.write("\u022e\u0231\3\2\2\2\u022f\u022d\3\2\2\2\u022f\u0230\3")
 */
  __pyx_tuple__483 = PyTuple_Pack(1, __pyx_kp_u_T); if (unlikely(!__pyx_tuple__483)) __PYX_ERR(0, 312, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__483);
  __Pyx_GIVEREF(__pyx_tuple__483);

  /* "pymola/generated/ModelicaParser.pyx":313
 *         buf.write("\7Y\2\2\u0227\u0226\3\2\2\2\u0227\u0228\3\2\2\2\u0228")
 *         buf.write("\u0229\3\2\2\2\u0229\u022f\7,\2\2\u022a\u022b\5T+\2\u022b")
 *         buf.write("\u022c\7\3\2\2\u022c\u022e\3\2\2\2\u022d\u022a\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u022e\u0231\3\2\2\2\u022f\u022d\3\2\2\2\u022f\u0230\3")
 *         buf.write("\2\2\2\u0230O\3\2\2\2\u0231\u022f\3\2\2\2\u0232\u0234")
 */
  __pyx_tuple__485 = PyTuple_Pack(1, __pyx_kp_u__484); if (unlikely(!__pyx_tuple__485)) __PYX_ERR(0, 313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__485);
  __Pyx_GIVEREF(__pyx_tuple__485);

  /* "pymola/generated/ModelicaParser.pyx":314
 *         buf.write("\u0229\3\2\2\2\u0229\u022f\7,\2\2\u022a\u022b\5T+\2\u022b")
 *         buf.write("\u022c\7\3\2\2\u022c\u022e\3\2\2\2\u022d\u022a\3\2\2\2")
 *         buf.write("\u022e\u0231\3\2\2\2\u022f\u022d\3\2\2\2\u022f\u0230\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u0230O\3\2\2\2\u0231\u022f\3\2\2\2\u0232\u0234")
 *         buf.write("\7Y\2\2\u0233\u0232\3\2\2\2\u0233\u0234\3\2\2\2\u0234")
 */
  __pyx_tuple__487 = PyTuple_Pack(1, __pyx_kp_u__486); if (unlikely(!__pyx_tuple__487)) __PYX_ERR(0, 314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__487);
  __Pyx_GIVEREF(__pyx_tuple__487);

  /* "pymola/generated/ModelicaParser.pyx":315
 *         buf.write("\u022c\7\3\2\2\u022c\u022e\3\2\2\2\u022d\u022a\3\2\2\2")
 *         buf.write("\u022e\u0231\3\2\2\2\u022f\u022d\3\2\2\2\u022f\u0230\3")
 *         buf.write("\2\2\2\u0230O\3\2\2\2\u0231\u022f\3\2\2\2\u0232\u0234")             # <<<<<<<<<<<<<<
 *         buf.write("\7Y\2\2\u0233\u0232\3\2\2\2\u0233\u0234\3\2\2\2\u0234")
 *         buf.write("\u0235\3\2\2\2\u0235\u023b\7-\2\2\u0236\u0237\5X-\2\u0237")
 */
  __pyx_tuple__488 = PyTuple_Pack(1, __pyx_kp_u_O); if (unlikely(!__pyx_tuple__488)) __PYX_ERR(0, 315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__488);
  __Pyx_GIVEREF(__pyx_tuple__488);

  /* "pymola/generated/ModelicaParser.pyx":316
 *         buf.write("\u022e\u0231\3\2\2\2\u022f\u022d\3\2\2\2\u022f\u0230\3")
 *         buf.write("\2\2\2\u0230O\3\2\2\2\u0231\u022f\3\2\2\2\u0232\u0234")
 *         buf.write("\7Y\2\2\u0233\u0232\3\2\2\2\u0233\u0234\3\2\2\2\u0234")             # <<<<<<<<<<<<<<
 *         buf.write("\u0235\3\2\2\2\u0235\u023b\7-\2\2\u0236\u0237\5X-\2\u0237")
 *         buf.write("\u0238\7\3\2\2\u0238\u023a\3\2\2\2\u0239\u0236\3\2\2\2")
 */
  __pyx_tuple__489 = PyTuple_Pack(1, __pyx_kp_u_Y_2); if (unlikely(!__pyx_tuple__489)) __PYX_ERR(0, 316, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__489);
  __Pyx_GIVEREF(__pyx_tuple__489);

  /* "pymola/generated/ModelicaParser.pyx":317
 *         buf.write("\2\2\2\u0230O\3\2\2\2\u0231\u022f\3\2\2\2\u0232\u0234")
 *         buf.write("\7Y\2\2\u0233\u0232\3\2\2\2\u0233\u0234\3\2\2\2\u0234")
 *         buf.write("\u0235\3\2\2\2\u0235\u023b\7-\2\2\u0236\u0237\5X-\2\u0237")             # <<<<<<<<<<<<<<
 *         buf.write("\u0238\7\3\2\2\u0238\u023a\3\2\2\2\u0239\u0236\3\2\2\2")
 *         buf.write("\u023a\u023d\3\2\2\2\u023b\u0239\3\2\2\2\u023b\u023c\3")
 */
  __pyx_tuple__490 = PyTuple_Pack(1, __pyx_kp_u_X_3); if (unlikely(!__pyx_tuple__490)) __PYX_ERR(0, 317, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__490);
  __Pyx_GIVEREF(__pyx_tuple__490);

  /* "pymola/generated/ModelicaParser.pyx":318
 *         buf.write("\7Y\2\2\u0233\u0232\3\2\2\2\u0233\u0234\3\2\2\2\u0234")
 *         buf.write("\u0235\3\2\2\2\u0235\u023b\7-\2\2\u0236\u0237\5X-\2\u0237")
 *         buf.write("\u0238\7\3\2\2\u0238\u023a\3\2\2\2\u0239\u0236\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u023a\u023d\3\2\2\2\u023b\u0239\3\2\2\2\u023b\u023c\3")
 *         buf.write("\2\2\2\u023cQ\3\2\2\2\u023d\u023b\3\2\2\2\u023e\u023f")
 */
  __pyx_tuple__492 = PyTuple_Pack(1, __pyx_kp_u__491); if (unlikely(!__pyx_tuple__492)) __PYX_ERR(0, 318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__492);
  __Pyx_GIVEREF(__pyx_tuple__492);

  /* "pymola/generated/ModelicaParser.pyx":319
 *         buf.write("\u0235\3\2\2\2\u0235\u023b\7-\2\2\u0236\u0237\5X-\2\u0237")
 *         buf.write("\u0238\7\3\2\2\u0238\u023a\3\2\2\2\u0239\u0236\3\2\2\2")
 *         buf.write("\u023a\u023d\3\2\2\2\u023b\u0239\3\2\2\2\u023b\u023c\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u023cQ\3\2\2\2\u023d\u023b\3\2\2\2\u023e\u023f")
 *         buf.write("\5p9\2\u023f\u0240\7\21\2\2\u0240\u0241\5n8\2\u0241\u024a")
 */
  __pyx_tuple__494 = PyTuple_Pack(1, __pyx_kp_u__493); if (unlikely(!__pyx_tuple__494)) __PYX_ERR(0, 319, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__494);
  __Pyx_GIVEREF(__pyx_tuple__494);

  /* "pymola/generated/ModelicaParser.pyx":320
 *         buf.write("\u0238\7\3\2\2\u0238\u023a\3\2\2\2\u0239\u0236\3\2\2\2")
 *         buf.write("\u023a\u023d\3\2\2\2\u023b\u0239\3\2\2\2\u023b\u023c\3")
 *         buf.write("\2\2\2\u023cQ\3\2\2\2\u023d\u023b\3\2\2\2\u023e\u023f")             # <<<<<<<<<<<<<<
 *         buf.write("\5p9\2\u023f\u0240\7\21\2\2\u0240\u0241\5n8\2\u0241\u024a")
 *         buf.write("\3\2\2\2\u0242\u024a\5Z.\2\u0243\u024a\5^\60\2\u0244\u024a")
 */
  __pyx_tuple__495 = PyTuple_Pack(1, __pyx_kp_u_Q_2); if (unlikely(!__pyx_tuple__495)) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__495);
  __Pyx_GIVEREF(__pyx_tuple__495);

  /* "pymola/generated/ModelicaParser.pyx":321
 *         buf.write("\u023a\u023d\3\2\2\2\u023b\u0239\3\2\2\2\u023b\u023c\3")
 *         buf.write("\2\2\2\u023cQ\3\2\2\2\u023d\u023b\3\2\2\2\u023e\u023f")
 *         buf.write("\5p9\2\u023f\u0240\7\21\2\2\u0240\u0241\5n8\2\u0241\u024a")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0242\u024a\5Z.\2\u0243\u024a\5^\60\2\u0244\u024a")
 *         buf.write("\5l\67\2\u0245\u024a\5h\65\2\u0246\u0247\5v<\2\u0247\u0248")
 */
  __pyx_tuple__496 = PyTuple_Pack(1, __pyx_kp_u_p9_n8); if (unlikely(!__pyx_tuple__496)) __PYX_ERR(0, 321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__496);
  __Pyx_GIVEREF(__pyx_tuple__496);

  /* "pymola/generated/ModelicaParser.pyx":322
 *         buf.write("\2\2\2\u023cQ\3\2\2\2\u023d\u023b\3\2\2\2\u023e\u023f")
 *         buf.write("\5p9\2\u023f\u0240\7\21\2\2\u0240\u0241\5n8\2\u0241\u024a")
 *         buf.write("\3\2\2\2\u0242\u024a\5Z.\2\u0243\u024a\5^\60\2\u0244\u024a")             # <<<<<<<<<<<<<<
 *         buf.write("\5l\67\2\u0245\u024a\5h\65\2\u0246\u0247\5v<\2\u0247\u0248")
 *         buf.write("\5|?\2\u0248\u024a\3\2\2\2\u0249\u023e\3\2\2\2\u0249\u0242")
 */
  __pyx_tuple__497 = PyTuple_Pack(1, __pyx_kp_u_Z_0); if (unlikely(!__pyx_tuple__497)) __PYX_ERR(0, 322, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__497);
  __Pyx_GIVEREF(__pyx_tuple__497);

  /* "pymola/generated/ModelicaParser.pyx":323
 *         buf.write("\5p9\2\u023f\u0240\7\21\2\2\u0240\u0241\5n8\2\u0241\u024a")
 *         buf.write("\3\2\2\2\u0242\u024a\5Z.\2\u0243\u024a\5^\60\2\u0244\u024a")
 *         buf.write("\5l\67\2\u0245\u024a\5h\65\2\u0246\u0247\5v<\2\u0247\u0248")             # <<<<<<<<<<<<<<
 *         buf.write("\5|?\2\u0248\u024a\3\2\2\2\u0249\u023e\3\2\2\2\u0249\u0242")
 *         buf.write("\3\2\2\2\u0249\u0243\3\2\2\2\u0249\u0244\3\2\2\2\u0249")
 */
  __pyx_tuple__498 = PyTuple_Pack(1, __pyx_kp_u_l7_h5_v); if (unlikely(!__pyx_tuple__498)) __PYX_ERR(0, 323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__498);
  __Pyx_GIVEREF(__pyx_tuple__498);

  /* "pymola/generated/ModelicaParser.pyx":324
 *         buf.write("\3\2\2\2\u0242\u024a\5Z.\2\u0243\u024a\5^\60\2\u0244\u024a")
 *         buf.write("\5l\67\2\u0245\u024a\5h\65\2\u0246\u0247\5v<\2\u0247\u0248")
 *         buf.write("\5|?\2\u0248\u024a\3\2\2\2\u0249\u023e\3\2\2\2\u0249\u0242")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0249\u0243\3\2\2\2\u0249\u0244\3\2\2\2\u0249")
 *         buf.write("\u0245\3\2\2\2\u0249\u0246\3\2\2\2\u024aS\3\2\2\2\u024b")
 */
  __pyx_tuple__500 = PyTuple_Pack(1, __pyx_kp_u__499); if (unlikely(!__pyx_tuple__500)) __PYX_ERR(0, 324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__500);
  __Pyx_GIVEREF(__pyx_tuple__500);

  /* "pymola/generated/ModelicaParser.pyx":325
 *         buf.write("\5l\67\2\u0245\u024a\5h\65\2\u0246\u0247\5v<\2\u0247\u0248")
 *         buf.write("\5|?\2\u0248\u024a\3\2\2\2\u0249\u023e\3\2\2\2\u0249\u0242")
 *         buf.write("\3\2\2\2\u0249\u0243\3\2\2\2\u0249\u0244\3\2\2\2\u0249")             # <<<<<<<<<<<<<<
 *         buf.write("\u0245\3\2\2\2\u0249\u0246\3\2\2\2\u024aS\3\2\2\2\u024b")
 *         buf.write("\u024c\5R*\2\u024c\u024d\5\u008eH\2\u024dU\3\2\2\2\u024e")
 */
  __pyx_tuple__502 = PyTuple_Pack(1, __pyx_kp_u__501); if (unlikely(!__pyx_tuple__502)) __PYX_ERR(0, 325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__502);
  __Pyx_GIVEREF(__pyx_tuple__502);

  /* "pymola/generated/ModelicaParser.pyx":326
 *         buf.write("\5|?\2\u0248\u024a\3\2\2\2\u0249\u023e\3\2\2\2\u0249\u0242")
 *         buf.write("\3\2\2\2\u0249\u0243\3\2\2\2\u0249\u0244\3\2\2\2\u0249")
 *         buf.write("\u0245\3\2\2\2\u0249\u0246\3\2\2\2\u024aS\3\2\2\2\u024b")             # <<<<<<<<<<<<<<
 *         buf.write("\u024c\5R*\2\u024c\u024d\5\u008eH\2\u024dU\3\2\2\2\u024e")
 *         buf.write("\u0252\5z>\2\u024f\u0250\7+\2\2\u0250\u0253\5n8\2\u0251")
 */
  __pyx_tuple__503 = PyTuple_Pack(1, __pyx_kp_u_S_6); if (unlikely(!__pyx_tuple__503)) __PYX_ERR(0, 326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__503);
  __Pyx_GIVEREF(__pyx_tuple__503);

  /* "pymola/generated/ModelicaParser.pyx":327
 *         buf.write("\3\2\2\2\u0249\u0243\3\2\2\2\u0249\u0244\3\2\2\2\u0249")
 *         buf.write("\u0245\3\2\2\2\u0249\u0246\3\2\2\2\u024aS\3\2\2\2\u024b")
 *         buf.write("\u024c\5R*\2\u024c\u024d\5\u008eH\2\u024dU\3\2\2\2\u024e")             # <<<<<<<<<<<<<<
 *         buf.write("\u0252\5z>\2\u024f\u0250\7+\2\2\u0250\u0253\5n8\2\u0251")
 *         buf.write("\u0253\5|?\2\u0252\u024f\3\2\2\2\u0252\u0251\3\2\2\2\u0253")
 */
  __pyx_tuple__504 = PyTuple_Pack(1, __pyx_kp_u_R_H_U); if (unlikely(!__pyx_tuple__504)) __PYX_ERR(0, 327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__504);
  __Pyx_GIVEREF(__pyx_tuple__504);

  /* "pymola/generated/ModelicaParser.pyx":328
 *         buf.write("\u0245\3\2\2\2\u0249\u0246\3\2\2\2\u024aS\3\2\2\2\u024b")
 *         buf.write("\u024c\5R*\2\u024c\u024d\5\u008eH\2\u024dU\3\2\2\2\u024e")
 *         buf.write("\u0252\5z>\2\u024f\u0250\7+\2\2\u0250\u0253\5n8\2\u0251")             # <<<<<<<<<<<<<<
 *         buf.write("\u0253\5|?\2\u0252\u024f\3\2\2\2\u0252\u0251\3\2\2\2\u0253")
 *         buf.write("\u0269\3\2\2\2\u0254\u0255\7\23\2\2\u0255\u025a\5z>\2")
 */
  __pyx_tuple__505 = PyTuple_Pack(1, __pyx_kp_u_z_n8); if (unlikely(!__pyx_tuple__505)) __PYX_ERR(0, 328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__505);
  __Pyx_GIVEREF(__pyx_tuple__505);

  /* "pymola/generated/ModelicaParser.pyx":329
 *         buf.write("\u024c\5R*\2\u024c\u024d\5\u008eH\2\u024dU\3\2\2\2\u024e")
 *         buf.write("\u0252\5z>\2\u024f\u0250\7+\2\2\u0250\u0253\5n8\2\u0251")
 *         buf.write("\u0253\5|?\2\u0252\u024f\3\2\2\2\u0252\u0251\3\2\2\2\u0253")             # <<<<<<<<<<<<<<
 *         buf.write("\u0269\3\2\2\2\u0254\u0255\7\23\2\2\u0255\u025a\5z>\2")
 *         buf.write("\u0256\u0257\7\27\2\2\u0257\u0259\5z>\2\u0258\u0256\3")
 */
  __pyx_tuple__507 = PyTuple_Pack(1, __pyx_kp_u__506); if (unlikely(!__pyx_tuple__507)) __PYX_ERR(0, 329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__507);
  __Pyx_GIVEREF(__pyx_tuple__507);

  /* "pymola/generated/ModelicaParser.pyx":330
 *         buf.write("\u0252\5z>\2\u024f\u0250\7+\2\2\u0250\u0253\5n8\2\u0251")
 *         buf.write("\u0253\5|?\2\u0252\u024f\3\2\2\2\u0252\u0251\3\2\2\2\u0253")
 *         buf.write("\u0269\3\2\2\2\u0254\u0255\7\23\2\2\u0255\u025a\5z>\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0256\u0257\7\27\2\2\u0257\u0259\5z>\2\u0258\u0256\3")
 *         buf.write("\2\2\2\u0259\u025c\3\2\2\2\u025a\u0258\3\2\2\2\u025a\u025b")
 */
  __pyx_tuple__508 = PyTuple_Pack(1, __pyx_kp_u_z_6); if (unlikely(!__pyx_tuple__508)) __PYX_ERR(0, 330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__508);
  __Pyx_GIVEREF(__pyx_tuple__508);

  /* "pymola/generated/ModelicaParser.pyx":331
 *         buf.write("\u0253\5|?\2\u0252\u024f\3\2\2\2\u0252\u0251\3\2\2\2\u0253")
 *         buf.write("\u0269\3\2\2\2\u0254\u0255\7\23\2\2\u0255\u025a\5z>\2")
 *         buf.write("\u0256\u0257\7\27\2\2\u0257\u0259\5z>\2\u0258\u0256\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u0259\u025c\3\2\2\2\u025a\u0258\3\2\2\2\u025a\u025b")
 *         buf.write("\3\2\2\2\u025b\u025d\3\2\2\2\u025c\u025a\3\2\2\2\u025d")
 */
  __pyx_tuple__509 = PyTuple_Pack(1, __pyx_kp_u_z_7); if (unlikely(!__pyx_tuple__509)) __PYX_ERR(0, 331, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__509);
  __Pyx_GIVEREF(__pyx_tuple__509);

  /* "pymola/generated/ModelicaParser.pyx":332
 *         buf.write("\u0269\3\2\2\2\u0254\u0255\7\23\2\2\u0255\u025a\5z>\2")
 *         buf.write("\u0256\u0257\7\27\2\2\u0257\u0259\5z>\2\u0258\u0256\3")
 *         buf.write("\2\2\2\u0259\u025c\3\2\2\2\u025a\u0258\3\2\2\2\u025a\u025b")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u025b\u025d\3\2\2\2\u025c\u025a\3\2\2\2\u025d")
 *         buf.write("\u025e\7\25\2\2\u025e\u025f\7+\2\2\u025f\u0260\5z>\2\u0260")
 */
  __pyx_tuple__511 = PyTuple_Pack(1, __pyx_kp_u__510); if (unlikely(!__pyx_tuple__511)) __PYX_ERR(0, 332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__511);
  __Pyx_GIVEREF(__pyx_tuple__511);

  /* "pymola/generated/ModelicaParser.pyx":333
 *         buf.write("\u0256\u0257\7\27\2\2\u0257\u0259\5z>\2\u0258\u0256\3")
 *         buf.write("\2\2\2\u0259\u025c\3\2\2\2\u025a\u0258\3\2\2\2\u025a\u025b")
 *         buf.write("\3\2\2\2\u025b\u025d\3\2\2\2\u025c\u025a\3\2\2\2\u025d")             # <<<<<<<<<<<<<<
 *         buf.write("\u025e\7\25\2\2\u025e\u025f\7+\2\2\u025f\u0260\5z>\2\u0260")
 *         buf.write("\u0261\5|?\2\u0261\u0269\3\2\2\2\u0262\u0269\7.\2\2\u0263")
 */
  __pyx_tuple__513 = PyTuple_Pack(1, __pyx_kp_u__512); if (unlikely(!__pyx_tuple__513)) __PYX_ERR(0, 333, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__513);
  __Pyx_GIVEREF(__pyx_tuple__513);

  /* "pymola/generated/ModelicaParser.pyx":334
 *         buf.write("\2\2\2\u0259\u025c\3\2\2\2\u025a\u0258\3\2\2\2\u025a\u025b")
 *         buf.write("\3\2\2\2\u025b\u025d\3\2\2\2\u025c\u025a\3\2\2\2\u025d")
 *         buf.write("\u025e\7\25\2\2\u025e\u025f\7+\2\2\u025f\u0260\5z>\2\u0260")             # <<<<<<<<<<<<<<
 *         buf.write("\u0261\5|?\2\u0261\u0269\3\2\2\2\u0262\u0269\7.\2\2\u0263")
 *         buf.write("\u0269\7/\2\2\u0264\u0269\5\\/\2\u0265\u0269\5`\61\2\u0266")
 */
  __pyx_tuple__514 = PyTuple_Pack(1, __pyx_kp_u_z_9); if (unlikely(!__pyx_tuple__514)) __PYX_ERR(0, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__514);
  __Pyx_GIVEREF(__pyx_tuple__514);

  /* "pymola/generated/ModelicaParser.pyx":335
 *         buf.write("\3\2\2\2\u025b\u025d\3\2\2\2\u025c\u025a\3\2\2\2\u025d")
 *         buf.write("\u025e\7\25\2\2\u025e\u025f\7+\2\2\u025f\u0260\5z>\2\u0260")
 *         buf.write("\u0261\5|?\2\u0261\u0269\3\2\2\2\u0262\u0269\7.\2\2\u0263")             # <<<<<<<<<<<<<<
 *         buf.write("\u0269\7/\2\2\u0264\u0269\5\\/\2\u0265\u0269\5`\61\2\u0266")
 *         buf.write("\u0269\5f\64\2\u0267\u0269\5j\66\2\u0268\u024e\3\2\2\2")
 */
  __pyx_tuple__516 = PyTuple_Pack(1, __pyx_kp_u__515); if (unlikely(!__pyx_tuple__516)) __PYX_ERR(0, 335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__516);
  __Pyx_GIVEREF(__pyx_tuple__516);

  /* "pymola/generated/ModelicaParser.pyx":336
 *         buf.write("\u025e\7\25\2\2\u025e\u025f\7+\2\2\u025f\u0260\5z>\2\u0260")
 *         buf.write("\u0261\5|?\2\u0261\u0269\3\2\2\2\u0262\u0269\7.\2\2\u0263")
 *         buf.write("\u0269\7/\2\2\u0264\u0269\5\\/\2\u0265\u0269\5`\61\2\u0266")             # <<<<<<<<<<<<<<
 *         buf.write("\u0269\5f\64\2\u0267\u0269\5j\66\2\u0268\u024e\3\2\2\2")
 *         buf.write("\u0268\u0254\3\2\2\2\u0268\u0262\3\2\2\2\u0268\u0263\3")
 */
  __pyx_tuple__517 = PyTuple_Pack(1, __pyx_kp_u_1_2); if (unlikely(!__pyx_tuple__517)) __PYX_ERR(0, 336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__517);
  __Pyx_GIVEREF(__pyx_tuple__517);

  /* "pymola/generated/ModelicaParser.pyx":337
 *         buf.write("\u0261\5|?\2\u0261\u0269\3\2\2\2\u0262\u0269\7.\2\2\u0263")
 *         buf.write("\u0269\7/\2\2\u0264\u0269\5\\/\2\u0265\u0269\5`\61\2\u0266")
 *         buf.write("\u0269\5f\64\2\u0267\u0269\5j\66\2\u0268\u024e\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0268\u0254\3\2\2\2\u0268\u0262\3\2\2\2\u0268\u0263\3")
 *         buf.write("\2\2\2\u0268\u0264\3\2\2\2\u0268\u0265\3\2\2\2\u0268\u0266")
 */
  __pyx_tuple__518 = PyTuple_Pack(1, __pyx_kp_u_f4_j6); if (unlikely(!__pyx_tuple__518)) __PYX_ERR(0, 337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__518);
  __Pyx_GIVEREF(__pyx_tuple__518);

  /* "pymola/generated/ModelicaParser.pyx":338
 *         buf.write("\u0269\7/\2\2\u0264\u0269\5\\/\2\u0265\u0269\5`\61\2\u0266")
 *         buf.write("\u0269\5f\64\2\u0267\u0269\5j\66\2\u0268\u024e\3\2\2\2")
 *         buf.write("\u0268\u0254\3\2\2\2\u0268\u0262\3\2\2\2\u0268\u0263\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u0268\u0264\3\2\2\2\u0268\u0265\3\2\2\2\u0268\u0266")
 *         buf.write("\3\2\2\2\u0268\u0267\3\2\2\2\u0269W\3\2\2\2\u026a\u026b")
 */
  __pyx_tuple__520 = PyTuple_Pack(1, __pyx_kp_u__519); if (unlikely(!__pyx_tuple__520)) __PYX_ERR(0, 338, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__520);
  __Pyx_GIVEREF(__pyx_tuple__520);

  /* "pymola/generated/ModelicaParser.pyx":339
 *         buf.write("\u0269\5f\64\2\u0267\u0269\5j\66\2\u0268\u024e\3\2\2\2")
 *         buf.write("\u0268\u0254\3\2\2\2\u0268\u0262\3\2\2\2\u0268\u0263\3")
 *         buf.write("\2\2\2\u0268\u0264\3\2\2\2\u0268\u0265\3\2\2\2\u0268\u0266")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0268\u0267\3\2\2\2\u0269W\3\2\2\2\u026a\u026b")
 *         buf.write("\5V,\2\u026b\u026c\5\u008eH\2\u026cY\3\2\2\2\u026d\u026e")
 */
  __pyx_tuple__522 = PyTuple_Pack(1, __pyx_kp_u__521); if (unlikely(!__pyx_tuple__522)) __PYX_ERR(0, 339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__522);
  __Pyx_GIVEREF(__pyx_tuple__522);

  /* "pymola/generated/ModelicaParser.pyx":340
 *         buf.write("\u0268\u0254\3\2\2\2\u0268\u0262\3\2\2\2\u0268\u0263\3")
 *         buf.write("\2\2\2\u0268\u0264\3\2\2\2\u0268\u0265\3\2\2\2\u0268\u0266")
 *         buf.write("\3\2\2\2\u0268\u0267\3\2\2\2\u0269W\3\2\2\2\u026a\u026b")             # <<<<<<<<<<<<<<
 *         buf.write("\5V,\2\u026b\u026c\5\u008eH\2\u026cY\3\2\2\2\u026d\u026e")
 *         buf.write("\7*\2\2\u026e\u026f\5n8\2\u026f\u0275\7\60\2\2\u0270\u0271")
 */
  __pyx_tuple__523 = PyTuple_Pack(1, __pyx_kp_u_W_3); if (unlikely(!__pyx_tuple__523)) __PYX_ERR(0, 340, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__523);
  __Pyx_GIVEREF(__pyx_tuple__523);

  /* "pymola/generated/ModelicaParser.pyx":341
 *         buf.write("\2\2\2\u0268\u0264\3\2\2\2\u0268\u0265\3\2\2\2\u0268\u0266")
 *         buf.write("\3\2\2\2\u0268\u0267\3\2\2\2\u0269W\3\2\2\2\u026a\u026b")
 *         buf.write("\5V,\2\u026b\u026c\5\u008eH\2\u026cY\3\2\2\2\u026d\u026e")             # <<<<<<<<<<<<<<
 *         buf.write("\7*\2\2\u026e\u026f\5n8\2\u026f\u0275\7\60\2\2\u0270\u0271")
 *         buf.write("\5T+\2\u0271\u0272\7\3\2\2\u0272\u0274\3\2\2\2\u0273\u0270")
 */
  __pyx_tuple__524 = PyTuple_Pack(1, __pyx_kp_u_V_H_Y); if (unlikely(!__pyx_tuple__524)) __PYX_ERR(0, 341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__524);
  __Pyx_GIVEREF(__pyx_tuple__524);

  /* "pymola/generated/ModelicaParser.pyx":342
 *         buf.write("\3\2\2\2\u0268\u0267\3\2\2\2\u0269W\3\2\2\2\u026a\u026b")
 *         buf.write("\5V,\2\u026b\u026c\5\u008eH\2\u026cY\3\2\2\2\u026d\u026e")
 *         buf.write("\7*\2\2\u026e\u026f\5n8\2\u026f\u0275\7\60\2\2\u0270\u0271")             # <<<<<<<<<<<<<<
 *         buf.write("\5T+\2\u0271\u0272\7\3\2\2\u0272\u0274\3\2\2\2\u0273\u0270")
 *         buf.write("\3\2\2\2\u0274\u0277\3\2\2\2\u0275\u0273\3\2\2\2\u0275")
 */
  __pyx_tuple__525 = PyTuple_Pack(1, __pyx_kp_u_n8_0); if (unlikely(!__pyx_tuple__525)) __PYX_ERR(0, 342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__525);
  __Pyx_GIVEREF(__pyx_tuple__525);

  /* "pymola/generated/ModelicaParser.pyx":343
 *         buf.write("\5V,\2\u026b\u026c\5\u008eH\2\u026cY\3\2\2\2\u026d\u026e")
 *         buf.write("\7*\2\2\u026e\u026f\5n8\2\u026f\u0275\7\60\2\2\u0270\u0271")
 *         buf.write("\5T+\2\u0271\u0272\7\3\2\2\u0272\u0274\3\2\2\2\u0273\u0270")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0274\u0277\3\2\2\2\u0275\u0273\3\2\2\2\u0275")
 *         buf.write("\u0276\3\2\2\2\u0276\u0285\3\2\2\2\u0277\u0275\3\2\2\2")
 */
  __pyx_tuple__526 = PyTuple_Pack(1, __pyx_kp_u_T_2); if (unlikely(!__pyx_tuple__526)) __PYX_ERR(0, 343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__526);
  __Pyx_GIVEREF(__pyx_tuple__526);

  /* "pymola/generated/ModelicaParser.pyx":344
 *         buf.write("\7*\2\2\u026e\u026f\5n8\2\u026f\u0275\7\60\2\2\u0270\u0271")
 *         buf.write("\5T+\2\u0271\u0272\7\3\2\2\u0272\u0274\3\2\2\2\u0273\u0270")
 *         buf.write("\3\2\2\2\u0274\u0277\3\2\2\2\u0275\u0273\3\2\2\2\u0275")             # <<<<<<<<<<<<<<
 *         buf.write("\u0276\3\2\2\2\u0276\u0285\3\2\2\2\u0277\u0275\3\2\2\2")
 *         buf.write("\u0278\u0279\7\61\2\2\u0279\u027a\5n8\2\u027a\u0280\7")
 */
  __pyx_tuple__528 = PyTuple_Pack(1, __pyx_kp_u__527); if (unlikely(!__pyx_tuple__528)) __PYX_ERR(0, 344, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__528);
  __Pyx_GIVEREF(__pyx_tuple__528);

  /* "pymola/generated/ModelicaParser.pyx":345
 *         buf.write("\5T+\2\u0271\u0272\7\3\2\2\u0272\u0274\3\2\2\2\u0273\u0270")
 *         buf.write("\3\2\2\2\u0274\u0277\3\2\2\2\u0275\u0273\3\2\2\2\u0275")
 *         buf.write("\u0276\3\2\2\2\u0276\u0285\3\2\2\2\u0277\u0275\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0278\u0279\7\61\2\2\u0279\u027a\5n8\2\u027a\u0280\7")
 *         buf.write("\60\2\2\u027b\u027c\5T+\2\u027c\u027d\7\3\2\2\u027d\u027f")
 */
  __pyx_tuple__530 = PyTuple_Pack(1, __pyx_kp_u__529); if (unlikely(!__pyx_tuple__530)) __PYX_ERR(0, 345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__530);
  __Pyx_GIVEREF(__pyx_tuple__530);

  /* "pymola/generated/ModelicaParser.pyx":346
 *         buf.write("\3\2\2\2\u0274\u0277\3\2\2\2\u0275\u0273\3\2\2\2\u0275")
 *         buf.write("\u0276\3\2\2\2\u0276\u0285\3\2\2\2\u0277\u0275\3\2\2\2")
 *         buf.write("\u0278\u0279\7\61\2\2\u0279\u027a\5n8\2\u027a\u0280\7")             # <<<<<<<<<<<<<<
 *         buf.write("\60\2\2\u027b\u027c\5T+\2\u027c\u027d\7\3\2\2\u027d\u027f")
 *         buf.write("\3\2\2\2\u027e\u027b\3\2\2\2\u027f\u0282\3\2\2\2\u0280")
 */
  __pyx_tuple__531 = PyTuple_Pack(1, __pyx_kp_u_1_n8); if (unlikely(!__pyx_tuple__531)) __PYX_ERR(0, 346, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__531);
  __Pyx_GIVEREF(__pyx_tuple__531);

  /* "pymola/generated/ModelicaParser.pyx":347
 *         buf.write("\u0276\3\2\2\2\u0276\u0285\3\2\2\2\u0277\u0275\3\2\2\2")
 *         buf.write("\u0278\u0279\7\61\2\2\u0279\u027a\5n8\2\u027a\u0280\7")
 *         buf.write("\60\2\2\u027b\u027c\5T+\2\u027c\u027d\7\3\2\2\u027d\u027f")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u027e\u027b\3\2\2\2\u027f\u0282\3\2\2\2\u0280")
 *         buf.write("\u027e\3\2\2\2\u0280\u0281\3\2\2\2\u0281\u0284\3\2\2\2")
 */
  __pyx_tuple__532 = PyTuple_Pack(1, __pyx_kp_u_0_T); if (unlikely(!__pyx_tuple__532)) __PYX_ERR(0, 347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__532);
  __Pyx_GIVEREF(__pyx_tuple__532);

  /* "pymola/generated/ModelicaParser.pyx":348
 *         buf.write("\u0278\u0279\7\61\2\2\u0279\u027a\5n8\2\u027a\u0280\7")
 *         buf.write("\60\2\2\u027b\u027c\5T+\2\u027c\u027d\7\3\2\2\u027d\u027f")
 *         buf.write("\3\2\2\2\u027e\u027b\3\2\2\2\u027f\u0282\3\2\2\2\u0280")             # <<<<<<<<<<<<<<
 *         buf.write("\u027e\3\2\2\2\u0280\u0281\3\2\2\2\u0281\u0284\3\2\2\2")
 *         buf.write("\u0282\u0280\3\2\2\2\u0283\u0278\3\2\2\2\u0284\u0287\3")
 */
  __pyx_tuple__534 = PyTuple_Pack(1, __pyx_kp_u__533); if (unlikely(!__pyx_tuple__534)) __PYX_ERR(0, 348, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__534);
  __Pyx_GIVEREF(__pyx_tuple__534);

  /* "pymola/generated/ModelicaParser.pyx":349
 *         buf.write("\60\2\2\u027b\u027c\5T+\2\u027c\u027d\7\3\2\2\u027d\u027f")
 *         buf.write("\3\2\2\2\u027e\u027b\3\2\2\2\u027f\u0282\3\2\2\2\u0280")
 *         buf.write("\u027e\3\2\2\2\u0280\u0281\3\2\2\2\u0281\u0284\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0282\u0280\3\2\2\2\u0283\u0278\3\2\2\2\u0284\u0287\3")
 *         buf.write("\2\2\2\u0285\u0283\3\2\2\2\u0285\u0286\3\2\2\2\u0286\u0291")
 */
  __pyx_tuple__536 = PyTuple_Pack(1, __pyx_kp_u__535); if (unlikely(!__pyx_tuple__536)) __PYX_ERR(0, 349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__536);
  __Pyx_GIVEREF(__pyx_tuple__536);

  /* "pymola/generated/ModelicaParser.pyx":350
 *         buf.write("\3\2\2\2\u027e\u027b\3\2\2\2\u027f\u0282\3\2\2\2\u0280")
 *         buf.write("\u027e\3\2\2\2\u0280\u0281\3\2\2\2\u0281\u0284\3\2\2\2")
 *         buf.write("\u0282\u0280\3\2\2\2\u0283\u0278\3\2\2\2\u0284\u0287\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u0285\u0283\3\2\2\2\u0285\u0286\3\2\2\2\u0286\u0291")
 *         buf.write("\3\2\2\2\u0287\u0285\3\2\2\2\u0288\u028e\7\62\2\2\u0289")
 */
  __pyx_tuple__538 = PyTuple_Pack(1, __pyx_kp_u__537); if (unlikely(!__pyx_tuple__538)) __PYX_ERR(0, 350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__538);
  __Pyx_GIVEREF(__pyx_tuple__538);

  /* "pymola/generated/ModelicaParser.pyx":351
 *         buf.write("\u027e\3\2\2\2\u0280\u0281\3\2\2\2\u0281\u0284\3\2\2\2")
 *         buf.write("\u0282\u0280\3\2\2\2\u0283\u0278\3\2\2\2\u0284\u0287\3")
 *         buf.write("\2\2\2\u0285\u0283\3\2\2\2\u0285\u0286\3\2\2\2\u0286\u0291")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0287\u0285\3\2\2\2\u0288\u028e\7\62\2\2\u0289")
 *         buf.write("\u028a\5T+\2\u028a\u028b\7\3\2\2\u028b\u028d\3\2\2\2\u028c")
 */
  __pyx_tuple__540 = PyTuple_Pack(1, __pyx_kp_u__539); if (unlikely(!__pyx_tuple__540)) __PYX_ERR(0, 351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__540);
  __Pyx_GIVEREF(__pyx_tuple__540);

  /* "pymola/generated/ModelicaParser.pyx":352
 *         buf.write("\u0282\u0280\3\2\2\2\u0283\u0278\3\2\2\2\u0284\u0287\3")
 *         buf.write("\2\2\2\u0285\u0283\3\2\2\2\u0285\u0286\3\2\2\2\u0286\u0291")
 *         buf.write("\3\2\2\2\u0287\u0285\3\2\2\2\u0288\u028e\7\62\2\2\u0289")             # <<<<<<<<<<<<<<
 *         buf.write("\u028a\5T+\2\u028a\u028b\7\3\2\2\u028b\u028d\3\2\2\2\u028c")
 *         buf.write("\u0289\3\2\2\2\u028d\u0290\3\2\2\2\u028e\u028c\3\2\2\2")
 */
  __pyx_tuple__541 = PyTuple_Pack(1, __pyx_kp_u_2); if (unlikely(!__pyx_tuple__541)) __PYX_ERR(0, 352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__541);
  __Pyx_GIVEREF(__pyx_tuple__541);

  /* "pymola/generated/ModelicaParser.pyx":353
 *         buf.write("\2\2\2\u0285\u0283\3\2\2\2\u0285\u0286\3\2\2\2\u0286\u0291")
 *         buf.write("\3\2\2\2\u0287\u0285\3\2\2\2\u0288\u028e\7\62\2\2\u0289")
 *         buf.write("\u028a\5T+\2\u028a\u028b\7\3\2\2\u028b\u028d\3\2\2\2\u028c")             # <<<<<<<<<<<<<<
 *         buf.write("\u0289\3\2\2\2\u028d\u0290\3\2\2\2\u028e\u028c\3\2\2\2")
 *         buf.write("\u028e\u028f\3\2\2\2\u028f\u0292\3\2\2\2\u0290\u028e\3")
 */
  __pyx_tuple__542 = PyTuple_Pack(1, __pyx_kp_u_T_3); if (unlikely(!__pyx_tuple__542)) __PYX_ERR(0, 353, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__542);
  __Pyx_GIVEREF(__pyx_tuple__542);

  /* "pymola/generated/ModelicaParser.pyx":354
 *         buf.write("\3\2\2\2\u0287\u0285\3\2\2\2\u0288\u028e\7\62\2\2\u0289")
 *         buf.write("\u028a\5T+\2\u028a\u028b\7\3\2\2\u028b\u028d\3\2\2\2\u028c")
 *         buf.write("\u0289\3\2\2\2\u028d\u0290\3\2\2\2\u028e\u028c\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u028e\u028f\3\2\2\2\u028f\u0292\3\2\2\2\u0290\u028e\3")
 *         buf.write("\2\2\2\u0291\u0288\3\2\2\2\u0291\u0292\3\2\2\2\u0292\u0293")
 */
  __pyx_tuple__544 = PyTuple_Pack(1, __pyx_kp_u__543); if (unlikely(!__pyx_tuple__544)) __PYX_ERR(0, 354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__544);
  __Pyx_GIVEREF(__pyx_tuple__544);

  /* "pymola/generated/ModelicaParser.pyx":355
 *         buf.write("\u028a\5T+\2\u028a\u028b\7\3\2\2\u028b\u028d\3\2\2\2\u028c")
 *         buf.write("\u0289\3\2\2\2\u028d\u0290\3\2\2\2\u028e\u028c\3\2\2\2")
 *         buf.write("\u028e\u028f\3\2\2\2\u028f\u0292\3\2\2\2\u0290\u028e\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u0291\u0288\3\2\2\2\u0291\u0292\3\2\2\2\u0292\u0293")
 *         buf.write("\3\2\2\2\u0293\u0294\7\20\2\2\u0294\u0295\7*\2\2\u0295")
 */
  __pyx_tuple__546 = PyTuple_Pack(1, __pyx_kp_u__545); if (unlikely(!__pyx_tuple__546)) __PYX_ERR(0, 355, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__546);
  __Pyx_GIVEREF(__pyx_tuple__546);

  /* "pymola/generated/ModelicaParser.pyx":356
 *         buf.write("\u0289\3\2\2\2\u028d\u0290\3\2\2\2\u028e\u028c\3\2\2\2")
 *         buf.write("\u028e\u028f\3\2\2\2\u028f\u0292\3\2\2\2\u0290\u028e\3")
 *         buf.write("\2\2\2\u0291\u0288\3\2\2\2\u0291\u0292\3\2\2\2\u0292\u0293")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0293\u0294\7\20\2\2\u0294\u0295\7*\2\2\u0295")
 *         buf.write("[\3\2\2\2\u0296\u0297\7*\2\2\u0297\u0298\5n8\2\u0298\u029e")
 */
  __pyx_tuple__548 = PyTuple_Pack(1, __pyx_kp_u__547); if (unlikely(!__pyx_tuple__548)) __PYX_ERR(0, 356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__548);
  __Pyx_GIVEREF(__pyx_tuple__548);

  /* "pymola/generated/ModelicaParser.pyx":357
 *         buf.write("\u028e\u028f\3\2\2\2\u028f\u0292\3\2\2\2\u0290\u028e\3")
 *         buf.write("\2\2\2\u0291\u0288\3\2\2\2\u0291\u0292\3\2\2\2\u0292\u0293")
 *         buf.write("\3\2\2\2\u0293\u0294\7\20\2\2\u0294\u0295\7*\2\2\u0295")             # <<<<<<<<<<<<<<
 *         buf.write("[\3\2\2\2\u0296\u0297\7*\2\2\u0297\u0298\5n8\2\u0298\u029e")
 *         buf.write("\7\60\2\2\u0299\u029a\5X-\2\u029a\u029b\7\3\2\2\u029b")
 */
  __pyx_tuple__550 = PyTuple_Pack(1, __pyx_kp_u__549); if (unlikely(!__pyx_tuple__550)) __PYX_ERR(0, 357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__550);
  __Pyx_GIVEREF(__pyx_tuple__550);

  /* "pymola/generated/ModelicaParser.pyx":358
 *         buf.write("\2\2\2\u0291\u0288\3\2\2\2\u0291\u0292\3\2\2\2\u0292\u0293")
 *         buf.write("\3\2\2\2\u0293\u0294\7\20\2\2\u0294\u0295\7*\2\2\u0295")
 *         buf.write("[\3\2\2\2\u0296\u0297\7*\2\2\u0297\u0298\5n8\2\u0298\u029e")             # <<<<<<<<<<<<<<
 *         buf.write("\7\60\2\2\u0299\u029a\5X-\2\u029a\u029b\7\3\2\2\u029b")
 *         buf.write("\u029d\3\2\2\2\u029c\u0299\3\2\2\2\u029d\u02a0\3\2\2\2")
 */
  __pyx_tuple__551 = PyTuple_Pack(1, __pyx_kp_u_n8_4); if (unlikely(!__pyx_tuple__551)) __PYX_ERR(0, 358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__551);
  __Pyx_GIVEREF(__pyx_tuple__551);

  /* "pymola/generated/ModelicaParser.pyx":359
 *         buf.write("\3\2\2\2\u0293\u0294\7\20\2\2\u0294\u0295\7*\2\2\u0295")
 *         buf.write("[\3\2\2\2\u0296\u0297\7*\2\2\u0297\u0298\5n8\2\u0298\u029e")
 *         buf.write("\7\60\2\2\u0299\u029a\5X-\2\u029a\u029b\7\3\2\2\u029b")             # <<<<<<<<<<<<<<
 *         buf.write("\u029d\3\2\2\2\u029c\u0299\3\2\2\2\u029d\u02a0\3\2\2\2")
 *         buf.write("\u029e\u029c\3\2\2\2\u029e\u029f\3\2\2\2\u029f\u02ae\3")
 */
  __pyx_tuple__552 = PyTuple_Pack(1, __pyx_kp_u_0_X); if (unlikely(!__pyx_tuple__552)) __PYX_ERR(0, 359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__552);
  __Pyx_GIVEREF(__pyx_tuple__552);

  /* "pymola/generated/ModelicaParser.pyx":360
 *         buf.write("[\3\2\2\2\u0296\u0297\7*\2\2\u0297\u0298\5n8\2\u0298\u029e")
 *         buf.write("\7\60\2\2\u0299\u029a\5X-\2\u029a\u029b\7\3\2\2\u029b")
 *         buf.write("\u029d\3\2\2\2\u029c\u0299\3\2\2\2\u029d\u02a0\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u029e\u029c\3\2\2\2\u029e\u029f\3\2\2\2\u029f\u02ae\3")
 *         buf.write("\2\2\2\u02a0\u029e\3\2\2\2\u02a1\u02a2\7\61\2\2\u02a2")
 */
  __pyx_tuple__554 = PyTuple_Pack(1, __pyx_kp_u__553); if (unlikely(!__pyx_tuple__554)) __PYX_ERR(0, 360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__554);
  __Pyx_GIVEREF(__pyx_tuple__554);

  /* "pymola/generated/ModelicaParser.pyx":361
 *         buf.write("\7\60\2\2\u0299\u029a\5X-\2\u029a\u029b\7\3\2\2\u029b")
 *         buf.write("\u029d\3\2\2\2\u029c\u0299\3\2\2\2\u029d\u02a0\3\2\2\2")
 *         buf.write("\u029e\u029c\3\2\2\2\u029e\u029f\3\2\2\2\u029f\u02ae\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u02a0\u029e\3\2\2\2\u02a1\u02a2\7\61\2\2\u02a2")
 *         buf.write("\u02a3\5n8\2\u02a3\u02a9\7\60\2\2\u02a4\u02a5\5X-\2\u02a5")
 */
  __pyx_tuple__556 = PyTuple_Pack(1, __pyx_kp_u__555); if (unlikely(!__pyx_tuple__556)) __PYX_ERR(0, 361, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__556);
  __Pyx_GIVEREF(__pyx_tuple__556);

  /* "pymola/generated/ModelicaParser.pyx":362
 *         buf.write("\u029d\3\2\2\2\u029c\u0299\3\2\2\2\u029d\u02a0\3\2\2\2")
 *         buf.write("\u029e\u029c\3\2\2\2\u029e\u029f\3\2\2\2\u029f\u02ae\3")
 *         buf.write("\2\2\2\u02a0\u029e\3\2\2\2\u02a1\u02a2\7\61\2\2\u02a2")             # <<<<<<<<<<<<<<
 *         buf.write("\u02a3\5n8\2\u02a3\u02a9\7\60\2\2\u02a4\u02a5\5X-\2\u02a5")
 *         buf.write("\u02a6\7\3\2\2\u02a6\u02a8\3\2\2\2\u02a7\u02a4\3\2\2\2")
 */
  __pyx_tuple__557 = PyTuple_Pack(1, __pyx_kp_u_1_3); if (unlikely(!__pyx_tuple__557)) __PYX_ERR(0, 362, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__557);
  __Pyx_GIVEREF(__pyx_tuple__557);

  /* "pymola/generated/ModelicaParser.pyx":363
 *         buf.write("\u029e\u029c\3\2\2\2\u029e\u029f\3\2\2\2\u029f\u02ae\3")
 *         buf.write("\2\2\2\u02a0\u029e\3\2\2\2\u02a1\u02a2\7\61\2\2\u02a2")
 *         buf.write("\u02a3\5n8\2\u02a3\u02a9\7\60\2\2\u02a4\u02a5\5X-\2\u02a5")             # <<<<<<<<<<<<<<
 *         buf.write("\u02a6\7\3\2\2\u02a6\u02a8\3\2\2\2\u02a7\u02a4\3\2\2\2")
 *         buf.write("\u02a8\u02ab\3\2\2\2\u02a9\u02a7\3\2\2\2\u02a9\u02aa\3")
 */
  __pyx_tuple__558 = PyTuple_Pack(1, __pyx_kp_u_n8_0_X); if (unlikely(!__pyx_tuple__558)) __PYX_ERR(0, 363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__558);
  __Pyx_GIVEREF(__pyx_tuple__558);

  /* "pymola/generated/ModelicaParser.pyx":364
 *         buf.write("\2\2\2\u02a0\u029e\3\2\2\2\u02a1\u02a2\7\61\2\2\u02a2")
 *         buf.write("\u02a3\5n8\2\u02a3\u02a9\7\60\2\2\u02a4\u02a5\5X-\2\u02a5")
 *         buf.write("\u02a6\7\3\2\2\u02a6\u02a8\3\2\2\2\u02a7\u02a4\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u02a8\u02ab\3\2\2\2\u02a9\u02a7\3\2\2\2\u02a9\u02aa\3")
 *         buf.write("\2\2\2\u02aa\u02ad\3\2\2\2\u02ab\u02a9\3\2\2\2\u02ac\u02a1")
 */
  __pyx_tuple__560 = PyTuple_Pack(1, __pyx_kp_u__559); if (unlikely(!__pyx_tuple__560)) __PYX_ERR(0, 364, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__560);
  __Pyx_GIVEREF(__pyx_tuple__560);

  /* "pymola/generated/ModelicaParser.pyx":365
 *         buf.write("\u02a3\5n8\2\u02a3\u02a9\7\60\2\2\u02a4\u02a5\5X-\2\u02a5")
 *         buf.write("\u02a6\7\3\2\2\u02a6\u02a8\3\2\2\2\u02a7\u02a4\3\2\2\2")
 *         buf.write("\u02a8\u02ab\3\2\2\2\u02a9\u02a7\3\2\2\2\u02a9\u02aa\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u02aa\u02ad\3\2\2\2\u02ab\u02a9\3\2\2\2\u02ac\u02a1")
 *         buf.write("\3\2\2\2\u02ad\u02b0\3\2\2\2\u02ae\u02ac\3\2\2\2\u02ae")
 */
  __pyx_tuple__562 = PyTuple_Pack(1, __pyx_kp_u__561); if (unlikely(!__pyx_tuple__562)) __PYX_ERR(0, 365, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__562);
  __Pyx_GIVEREF(__pyx_tuple__562);

  /* "pymola/generated/ModelicaParser.pyx":366
 *         buf.write("\u02a6\7\3\2\2\u02a6\u02a8\3\2\2\2\u02a7\u02a4\3\2\2\2")
 *         buf.write("\u02a8\u02ab\3\2\2\2\u02a9\u02a7\3\2\2\2\u02a9\u02aa\3")
 *         buf.write("\2\2\2\u02aa\u02ad\3\2\2\2\u02ab\u02a9\3\2\2\2\u02ac\u02a1")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u02ad\u02b0\3\2\2\2\u02ae\u02ac\3\2\2\2\u02ae")
 *         buf.write("\u02af\3\2\2\2\u02af\u02ba\3\2\2\2\u02b0\u02ae\3\2\2\2")
 */
  __pyx_tuple__564 = PyTuple_Pack(1, __pyx_kp_u__563); if (unlikely(!__pyx_tuple__564)) __PYX_ERR(0, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__564);
  __Pyx_GIVEREF(__pyx_tuple__564);

  /* "pymola/generated/ModelicaParser.pyx":367
 *         buf.write("\u02a8\u02ab\3\2\2\2\u02a9\u02a7\3\2\2\2\u02a9\u02aa\3")
 *         buf.write("\2\2\2\u02aa\u02ad\3\2\2\2\u02ab\u02a9\3\2\2\2\u02ac\u02a1")
 *         buf.write("\3\2\2\2\u02ad\u02b0\3\2\2\2\u02ae\u02ac\3\2\2\2\u02ae")             # <<<<<<<<<<<<<<
 *         buf.write("\u02af\3\2\2\2\u02af\u02ba\3\2\2\2\u02b0\u02ae\3\2\2\2")
 *         buf.write("\u02b1\u02b7\7\62\2\2\u02b2\u02b3\5X-\2\u02b3\u02b4\7")
 */
  __pyx_tuple__566 = PyTuple_Pack(1, __pyx_kp_u__565); if (unlikely(!__pyx_tuple__566)) __PYX_ERR(0, 367, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__566);
  __Pyx_GIVEREF(__pyx_tuple__566);

  /* "pymola/generated/ModelicaParser.pyx":368
 *         buf.write("\2\2\2\u02aa\u02ad\3\2\2\2\u02ab\u02a9\3\2\2\2\u02ac\u02a1")
 *         buf.write("\3\2\2\2\u02ad\u02b0\3\2\2\2\u02ae\u02ac\3\2\2\2\u02ae")
 *         buf.write("\u02af\3\2\2\2\u02af\u02ba\3\2\2\2\u02b0\u02ae\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u02b1\u02b7\7\62\2\2\u02b2\u02b3\5X-\2\u02b3\u02b4\7")
 *         buf.write("\3\2\2\u02b4\u02b6\3\2\2\2\u02b5\u02b2\3\2\2\2\u02b6\u02b9")
 */
  __pyx_tuple__568 = PyTuple_Pack(1, __pyx_kp_u__567); if (unlikely(!__pyx_tuple__568)) __PYX_ERR(0, 368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__568);
  __Pyx_GIVEREF(__pyx_tuple__568);

  /* "pymola/generated/ModelicaParser.pyx":369
 *         buf.write("\3\2\2\2\u02ad\u02b0\3\2\2\2\u02ae\u02ac\3\2\2\2\u02ae")
 *         buf.write("\u02af\3\2\2\2\u02af\u02ba\3\2\2\2\u02b0\u02ae\3\2\2\2")
 *         buf.write("\u02b1\u02b7\7\62\2\2\u02b2\u02b3\5X-\2\u02b3\u02b4\7")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\u02b4\u02b6\3\2\2\2\u02b5\u02b2\3\2\2\2\u02b6\u02b9")
 *         buf.write("\3\2\2\2\u02b7\u02b5\3\2\2\2\u02b7\u02b8\3\2\2\2\u02b8")
 */
  __pyx_tuple__569 = PyTuple_Pack(1, __pyx_kp_u_2_X); if (unlikely(!__pyx_tuple__569)) __PYX_ERR(0, 369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__569);
  __Pyx_GIVEREF(__pyx_tuple__569);

  /* "pymola/generated/ModelicaParser.pyx":370
 *         buf.write("\u02af\3\2\2\2\u02af\u02ba\3\2\2\2\u02b0\u02ae\3\2\2\2")
 *         buf.write("\u02b1\u02b7\7\62\2\2\u02b2\u02b3\5X-\2\u02b3\u02b4\7")
 *         buf.write("\3\2\2\u02b4\u02b6\3\2\2\2\u02b5\u02b2\3\2\2\2\u02b6\u02b9")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u02b7\u02b5\3\2\2\2\u02b7\u02b8\3\2\2\2\u02b8")
 *         buf.write("\u02bb\3\2\2\2\u02b9\u02b7\3\2\2\2\u02ba\u02b1\3\2\2\2")
 */
  __pyx_tuple__571 = PyTuple_Pack(1, __pyx_kp_u__570); if (unlikely(!__pyx_tuple__571)) __PYX_ERR(0, 370, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__571);
  __Pyx_GIVEREF(__pyx_tuple__571);

  /* "pymola/generated/ModelicaParser.pyx":371
 *         buf.write("\u02b1\u02b7\7\62\2\2\u02b2\u02b3\5X-\2\u02b3\u02b4\7")
 *         buf.write("\3\2\2\u02b4\u02b6\3\2\2\2\u02b5\u02b2\3\2\2\2\u02b6\u02b9")
 *         buf.write("\3\2\2\2\u02b7\u02b5\3\2\2\2\u02b7\u02b8\3\2\2\2\u02b8")             # <<<<<<<<<<<<<<
 *         buf.write("\u02bb\3\2\2\2\u02b9\u02b7\3\2\2\2\u02ba\u02b1\3\2\2\2")
 *         buf.write("\u02ba\u02bb\3\2\2\2\u02bb\u02bc\3\2\2\2\u02bc\u02bd\7")
 */
  __pyx_tuple__573 = PyTuple_Pack(1, __pyx_kp_u__572); if (unlikely(!__pyx_tuple__573)) __PYX_ERR(0, 371, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__573);
  __Pyx_GIVEREF(__pyx_tuple__573);

  /* "pymola/generated/ModelicaParser.pyx":372
 *         buf.write("\3\2\2\u02b4\u02b6\3\2\2\2\u02b5\u02b2\3\2\2\2\u02b6\u02b9")
 *         buf.write("\3\2\2\2\u02b7\u02b5\3\2\2\2\u02b7\u02b8\3\2\2\2\u02b8")
 *         buf.write("\u02bb\3\2\2\2\u02b9\u02b7\3\2\2\2\u02ba\u02b1\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u02ba\u02bb\3\2\2\2\u02bb\u02bc\3\2\2\2\u02bc\u02bd\7")
 *         buf.write("\20\2\2\u02bd\u02be\7*\2\2\u02be]\3\2\2\2\u02bf\u02c0")
 */
  __pyx_tuple__575 = PyTuple_Pack(1, __pyx_kp_u__574); if (unlikely(!__pyx_tuple__575)) __PYX_ERR(0, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__575);
  __Pyx_GIVEREF(__pyx_tuple__575);

  /* "pymola/generated/ModelicaParser.pyx":373
 *         buf.write("\3\2\2\2\u02b7\u02b5\3\2\2\2\u02b7\u02b8\3\2\2\2\u02b8")
 *         buf.write("\u02bb\3\2\2\2\u02b9\u02b7\3\2\2\2\u02ba\u02b1\3\2\2\2")
 *         buf.write("\u02ba\u02bb\3\2\2\2\u02bb\u02bc\3\2\2\2\u02bc\u02bd\7")             # <<<<<<<<<<<<<<
 *         buf.write("\20\2\2\u02bd\u02be\7*\2\2\u02be]\3\2\2\2\u02bf\u02c0")
 *         buf.write("\7\63\2\2\u02c0\u02c1\5b\62\2\u02c1\u02c7\7\64\2\2\u02c2")
 */
  __pyx_tuple__577 = PyTuple_Pack(1, __pyx_kp_u__576); if (unlikely(!__pyx_tuple__577)) __PYX_ERR(0, 373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__577);
  __Pyx_GIVEREF(__pyx_tuple__577);

  /* "pymola/generated/ModelicaParser.pyx":374
 *         buf.write("\u02bb\3\2\2\2\u02b9\u02b7\3\2\2\2\u02ba\u02b1\3\2\2\2")
 *         buf.write("\u02ba\u02bb\3\2\2\2\u02bb\u02bc\3\2\2\2\u02bc\u02bd\7")
 *         buf.write("\20\2\2\u02bd\u02be\7*\2\2\u02be]\3\2\2\2\u02bf\u02c0")             # <<<<<<<<<<<<<<
 *         buf.write("\7\63\2\2\u02c0\u02c1\5b\62\2\u02c1\u02c7\7\64\2\2\u02c2")
 *         buf.write("\u02c3\5T+\2\u02c3\u02c4\7\3\2\2\u02c4\u02c6\3\2\2\2\u02c5")
 */
  __pyx_tuple__579 = PyTuple_Pack(1, __pyx_kp_u__578); if (unlikely(!__pyx_tuple__579)) __PYX_ERR(0, 374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__579);
  __Pyx_GIVEREF(__pyx_tuple__579);

  /* "pymola/generated/ModelicaParser.pyx":375
 *         buf.write("\u02ba\u02bb\3\2\2\2\u02bb\u02bc\3\2\2\2\u02bc\u02bd\7")
 *         buf.write("\20\2\2\u02bd\u02be\7*\2\2\u02be]\3\2\2\2\u02bf\u02c0")
 *         buf.write("\7\63\2\2\u02c0\u02c1\5b\62\2\u02c1\u02c7\7\64\2\2\u02c2")             # <<<<<<<<<<<<<<
 *         buf.write("\u02c3\5T+\2\u02c3\u02c4\7\3\2\2\u02c4\u02c6\3\2\2\2\u02c5")
 *         buf.write("\u02c2\3\2\2\2\u02c6\u02c9\3\2\2\2\u02c7\u02c5\3\2\2\2")
 */
  __pyx_tuple__580 = PyTuple_Pack(1, __pyx_kp_u_3_b2_4); if (unlikely(!__pyx_tuple__580)) __PYX_ERR(0, 375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__580);
  __Pyx_GIVEREF(__pyx_tuple__580);

  /* "pymola/generated/ModelicaParser.pyx":376
 *         buf.write("\20\2\2\u02bd\u02be\7*\2\2\u02be]\3\2\2\2\u02bf\u02c0")
 *         buf.write("\7\63\2\2\u02c0\u02c1\5b\62\2\u02c1\u02c7\7\64\2\2\u02c2")
 *         buf.write("\u02c3\5T+\2\u02c3\u02c4\7\3\2\2\u02c4\u02c6\3\2\2\2\u02c5")             # <<<<<<<<<<<<<<
 *         buf.write("\u02c2\3\2\2\2\u02c6\u02c9\3\2\2\2\u02c7\u02c5\3\2\2\2")
 *         buf.write("\u02c7\u02c8\3\2\2\2\u02c8\u02ca\3\2\2\2\u02c9\u02c7\3")
 */
  __pyx_tuple__581 = PyTuple_Pack(1, __pyx_kp_u_T_4); if (unlikely(!__pyx_tuple__581)) __PYX_ERR(0, 376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__581);
  __Pyx_GIVEREF(__pyx_tuple__581);

  /* "pymola/generated/ModelicaParser.pyx":377
 *         buf.write("\7\63\2\2\u02c0\u02c1\5b\62\2\u02c1\u02c7\7\64\2\2\u02c2")
 *         buf.write("\u02c3\5T+\2\u02c3\u02c4\7\3\2\2\u02c4\u02c6\3\2\2\2\u02c5")
 *         buf.write("\u02c2\3\2\2\2\u02c6\u02c9\3\2\2\2\u02c7\u02c5\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u02c7\u02c8\3\2\2\2\u02c8\u02ca\3\2\2\2\u02c9\u02c7\3")
 *         buf.write("\2\2\2\u02ca\u02cb\7\20\2\2\u02cb\u02cc\7\63\2\2\u02cc")
 */
  __pyx_tuple__583 = PyTuple_Pack(1, __pyx_kp_u__582); if (unlikely(!__pyx_tuple__583)) __PYX_ERR(0, 377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__583);
  __Pyx_GIVEREF(__pyx_tuple__583);

  /* "pymola/generated/ModelicaParser.pyx":378
 *         buf.write("\u02c3\5T+\2\u02c3\u02c4\7\3\2\2\u02c4\u02c6\3\2\2\2\u02c5")
 *         buf.write("\u02c2\3\2\2\2\u02c6\u02c9\3\2\2\2\u02c7\u02c5\3\2\2\2")
 *         buf.write("\u02c7\u02c8\3\2\2\2\u02c8\u02ca\3\2\2\2\u02c9\u02c7\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u02ca\u02cb\7\20\2\2\u02cb\u02cc\7\63\2\2\u02cc")
 *         buf.write("_\3\2\2\2\u02cd\u02ce\7\63\2\2\u02ce\u02cf\5b\62\2\u02cf")
 */
  __pyx_tuple__585 = PyTuple_Pack(1, __pyx_kp_u__584); if (unlikely(!__pyx_tuple__585)) __PYX_ERR(0, 378, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__585);
  __Pyx_GIVEREF(__pyx_tuple__585);

  /* "pymola/generated/ModelicaParser.pyx":379
 *         buf.write("\u02c2\3\2\2\2\u02c6\u02c9\3\2\2\2\u02c7\u02c5\3\2\2\2")
 *         buf.write("\u02c7\u02c8\3\2\2\2\u02c8\u02ca\3\2\2\2\u02c9\u02c7\3")
 *         buf.write("\2\2\2\u02ca\u02cb\7\20\2\2\u02cb\u02cc\7\63\2\2\u02cc")             # <<<<<<<<<<<<<<
 *         buf.write("_\3\2\2\2\u02cd\u02ce\7\63\2\2\u02ce\u02cf\5b\62\2\u02cf")
 *         buf.write("\u02d5\7\64\2\2\u02d0\u02d1\5X-\2\u02d1\u02d2\7\3\2\2")
 */
  __pyx_tuple__586 = PyTuple_Pack(1, __pyx_kp_u_3); if (unlikely(!__pyx_tuple__586)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__586);
  __Pyx_GIVEREF(__pyx_tuple__586);

  /* "pymola/generated/ModelicaParser.pyx":380
 *         buf.write("\u02c7\u02c8\3\2\2\2\u02c8\u02ca\3\2\2\2\u02c9\u02c7\3")
 *         buf.write("\2\2\2\u02ca\u02cb\7\20\2\2\u02cb\u02cc\7\63\2\2\u02cc")
 *         buf.write("_\3\2\2\2\u02cd\u02ce\7\63\2\2\u02ce\u02cf\5b\62\2\u02cf")             # <<<<<<<<<<<<<<
 *         buf.write("\u02d5\7\64\2\2\u02d0\u02d1\5X-\2\u02d1\u02d2\7\3\2\2")
 *         buf.write("\u02d2\u02d4\3\2\2\2\u02d3\u02d0\3\2\2\2\u02d4\u02d7\3")
 */
  __pyx_tuple__587 = PyTuple_Pack(1, __pyx_kp_u_3_b2); if (unlikely(!__pyx_tuple__587)) __PYX_ERR(0, 380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__587);
  __Pyx_GIVEREF(__pyx_tuple__587);

  /* "pymola/generated/ModelicaParser.pyx":381
 *         buf.write("\2\2\2\u02ca\u02cb\7\20\2\2\u02cb\u02cc\7\63\2\2\u02cc")
 *         buf.write("_\3\2\2\2\u02cd\u02ce\7\63\2\2\u02ce\u02cf\5b\62\2\u02cf")
 *         buf.write("\u02d5\7\64\2\2\u02d0\u02d1\5X-\2\u02d1\u02d2\7\3\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u02d2\u02d4\3\2\2\2\u02d3\u02d0\3\2\2\2\u02d4\u02d7\3")
 *         buf.write("\2\2\2\u02d5\u02d3\3\2\2\2\u02d5\u02d6\3\2\2\2\u02d6\u02d8")
 */
  __pyx_tuple__588 = PyTuple_Pack(1, __pyx_kp_u_4_X); if (unlikely(!__pyx_tuple__588)) __PYX_ERR(0, 381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__588);
  __Pyx_GIVEREF(__pyx_tuple__588);

  /* "pymola/generated/ModelicaParser.pyx":382
 *         buf.write("_\3\2\2\2\u02cd\u02ce\7\63\2\2\u02ce\u02cf\5b\62\2\u02cf")
 *         buf.write("\u02d5\7\64\2\2\u02d0\u02d1\5X-\2\u02d1\u02d2\7\3\2\2")
 *         buf.write("\u02d2\u02d4\3\2\2\2\u02d3\u02d0\3\2\2\2\u02d4\u02d7\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u02d5\u02d3\3\2\2\2\u02d5\u02d6\3\2\2\2\u02d6\u02d8")
 *         buf.write("\3\2\2\2\u02d7\u02d5\3\2\2\2\u02d8\u02d9\7\20\2\2\u02d9")
 */
  __pyx_tuple__590 = PyTuple_Pack(1, __pyx_kp_u__589); if (unlikely(!__pyx_tuple__590)) __PYX_ERR(0, 382, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__590);
  __Pyx_GIVEREF(__pyx_tuple__590);

  /* "pymola/generated/ModelicaParser.pyx":383
 *         buf.write("\u02d5\7\64\2\2\u02d0\u02d1\5X-\2\u02d1\u02d2\7\3\2\2")
 *         buf.write("\u02d2\u02d4\3\2\2\2\u02d3\u02d0\3\2\2\2\u02d4\u02d7\3")
 *         buf.write("\2\2\2\u02d5\u02d3\3\2\2\2\u02d5\u02d6\3\2\2\2\u02d6\u02d8")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u02d7\u02d5\3\2\2\2\u02d8\u02d9\7\20\2\2\u02d9")
 *         buf.write("\u02da\7\63\2\2\u02daa\3\2\2\2\u02db\u02e0\5d\63\2\u02dc")
 */
  __pyx_tuple__592 = PyTuple_Pack(1, __pyx_kp_u__591); if (unlikely(!__pyx_tuple__592)) __PYX_ERR(0, 383, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__592);
  __Pyx_GIVEREF(__pyx_tuple__592);

  /* "pymola/generated/ModelicaParser.pyx":384
 *         buf.write("\u02d2\u02d4\3\2\2\2\u02d3\u02d0\3\2\2\2\u02d4\u02d7\3")
 *         buf.write("\2\2\2\u02d5\u02d3\3\2\2\2\u02d5\u02d6\3\2\2\2\u02d6\u02d8")
 *         buf.write("\3\2\2\2\u02d7\u02d5\3\2\2\2\u02d8\u02d9\7\20\2\2\u02d9")             # <<<<<<<<<<<<<<
 *         buf.write("\u02da\7\63\2\2\u02daa\3\2\2\2\u02db\u02e0\5d\63\2\u02dc")
 *         buf.write("\u02dd\7\27\2\2\u02dd\u02df\5d\63\2\u02de\u02dc\3\2\2")
 */
  __pyx_tuple__594 = PyTuple_Pack(1, __pyx_kp_u__593); if (unlikely(!__pyx_tuple__594)) __PYX_ERR(0, 384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__594);
  __Pyx_GIVEREF(__pyx_tuple__594);

  /* "pymola/generated/ModelicaParser.pyx":385
 *         buf.write("\2\2\2\u02d5\u02d3\3\2\2\2\u02d5\u02d6\3\2\2\2\u02d6\u02d8")
 *         buf.write("\3\2\2\2\u02d7\u02d5\3\2\2\2\u02d8\u02d9\7\20\2\2\u02d9")
 *         buf.write("\u02da\7\63\2\2\u02daa\3\2\2\2\u02db\u02e0\5d\63\2\u02dc")             # <<<<<<<<<<<<<<
 *         buf.write("\u02dd\7\27\2\2\u02dd\u02df\5d\63\2\u02de\u02dc\3\2\2")
 *         buf.write("\2\u02df\u02e2\3\2\2\2\u02e0\u02de\3\2\2\2\u02e0\u02e1")
 */
  __pyx_tuple__595 = PyTuple_Pack(1, __pyx_kp_u_3_a_d3); if (unlikely(!__pyx_tuple__595)) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__595);
  __Pyx_GIVEREF(__pyx_tuple__595);

  /* "pymola/generated/ModelicaParser.pyx":386
 *         buf.write("\3\2\2\2\u02d7\u02d5\3\2\2\2\u02d8\u02d9\7\20\2\2\u02d9")
 *         buf.write("\u02da\7\63\2\2\u02daa\3\2\2\2\u02db\u02e0\5d\63\2\u02dc")
 *         buf.write("\u02dd\7\27\2\2\u02dd\u02df\5d\63\2\u02de\u02dc\3\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u02df\u02e2\3\2\2\2\u02e0\u02de\3\2\2\2\u02e0\u02e1")
 *         buf.write("\3\2\2\2\u02e1c\3\2\2\2\u02e2\u02e0\3\2\2\2\u02e3\u02e6")
 */
  __pyx_tuple__596 = PyTuple_Pack(1, __pyx_kp_u_d3); if (unlikely(!__pyx_tuple__596)) __PYX_ERR(0, 386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__596);
  __Pyx_GIVEREF(__pyx_tuple__596);

  /* "pymola/generated/ModelicaParser.pyx":387
 *         buf.write("\u02da\7\63\2\2\u02daa\3\2\2\2\u02db\u02e0\5d\63\2\u02dc")
 *         buf.write("\u02dd\7\27\2\2\u02dd\u02df\5d\63\2\u02de\u02dc\3\2\2")
 *         buf.write("\2\u02df\u02e2\3\2\2\2\u02e0\u02de\3\2\2\2\u02e0\u02e1")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u02e1c\3\2\2\2\u02e2\u02e0\3\2\2\2\u02e3\u02e6")
 *         buf.write("\7Z\2\2\u02e4\u02e5\7\65\2\2\u02e5\u02e7\5n8\2\u02e6\u02e4")
 */
  __pyx_tuple__598 = PyTuple_Pack(1, __pyx_kp_u__597); if (unlikely(!__pyx_tuple__598)) __PYX_ERR(0, 387, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__598);
  __Pyx_GIVEREF(__pyx_tuple__598);

  /* "pymola/generated/ModelicaParser.pyx":388
 *         buf.write("\u02dd\7\27\2\2\u02dd\u02df\5d\63\2\u02de\u02dc\3\2\2")
 *         buf.write("\2\u02df\u02e2\3\2\2\2\u02e0\u02de\3\2\2\2\u02e0\u02e1")
 *         buf.write("\3\2\2\2\u02e1c\3\2\2\2\u02e2\u02e0\3\2\2\2\u02e3\u02e6")             # <<<<<<<<<<<<<<
 *         buf.write("\7Z\2\2\u02e4\u02e5\7\65\2\2\u02e5\u02e7\5n8\2\u02e6\u02e4")
 *         buf.write("\3\2\2\2\u02e6\u02e7\3\2\2\2\u02e7e\3\2\2\2\u02e8\u02e9")
 */
  __pyx_tuple__599 = PyTuple_Pack(1, __pyx_kp_u_c); if (unlikely(!__pyx_tuple__599)) __PYX_ERR(0, 388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__599);
  __Pyx_GIVEREF(__pyx_tuple__599);

  /* "pymola/generated/ModelicaParser.pyx":389
 *         buf.write("\2\u02df\u02e2\3\2\2\2\u02e0\u02de\3\2\2\2\u02e0\u02e1")
 *         buf.write("\3\2\2\2\u02e1c\3\2\2\2\u02e2\u02e0\3\2\2\2\u02e3\u02e6")
 *         buf.write("\7Z\2\2\u02e4\u02e5\7\65\2\2\u02e5\u02e7\5n8\2\u02e6\u02e4")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u02e6\u02e7\3\2\2\2\u02e7e\3\2\2\2\u02e8\u02e9")
 *         buf.write("\7\66\2\2\u02e9\u02ea\5n8\2\u02ea\u02f0\7\64\2\2\u02eb")
 */
  __pyx_tuple__600 = PyTuple_Pack(1, __pyx_kp_u_Z_5_n8); if (unlikely(!__pyx_tuple__600)) __PYX_ERR(0, 389, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__600);
  __Pyx_GIVEREF(__pyx_tuple__600);

  /* "pymola/generated/ModelicaParser.pyx":390
 *         buf.write("\3\2\2\2\u02e1c\3\2\2\2\u02e2\u02e0\3\2\2\2\u02e3\u02e6")
 *         buf.write("\7Z\2\2\u02e4\u02e5\7\65\2\2\u02e5\u02e7\5n8\2\u02e6\u02e4")
 *         buf.write("\3\2\2\2\u02e6\u02e7\3\2\2\2\u02e7e\3\2\2\2\u02e8\u02e9")             # <<<<<<<<<<<<<<
 *         buf.write("\7\66\2\2\u02e9\u02ea\5n8\2\u02ea\u02f0\7\64\2\2\u02eb")
 *         buf.write("\u02ec\5X-\2\u02ec\u02ed\7\3\2\2\u02ed\u02ef\3\2\2\2\u02ee")
 */
  __pyx_tuple__601 = PyTuple_Pack(1, __pyx_kp_u_e); if (unlikely(!__pyx_tuple__601)) __PYX_ERR(0, 390, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__601);
  __Pyx_GIVEREF(__pyx_tuple__601);

  /* "pymola/generated/ModelicaParser.pyx":391
 *         buf.write("\7Z\2\2\u02e4\u02e5\7\65\2\2\u02e5\u02e7\5n8\2\u02e6\u02e4")
 *         buf.write("\3\2\2\2\u02e6\u02e7\3\2\2\2\u02e7e\3\2\2\2\u02e8\u02e9")
 *         buf.write("\7\66\2\2\u02e9\u02ea\5n8\2\u02ea\u02f0\7\64\2\2\u02eb")             # <<<<<<<<<<<<<<
 *         buf.write("\u02ec\5X-\2\u02ec\u02ed\7\3\2\2\u02ed\u02ef\3\2\2\2\u02ee")
 *         buf.write("\u02eb\3\2\2\2\u02ef\u02f2\3\2\2\2\u02f0\u02ee\3\2\2\2")
 */
  __pyx_tuple__602 = PyTuple_Pack(1, __pyx_kp_u_6_n8_4); if (unlikely(!__pyx_tuple__602)) __PYX_ERR(0, 391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__602);
  __Pyx_GIVEREF(__pyx_tuple__602);

  /* "pymola/generated/ModelicaParser.pyx":392
 *         buf.write("\3\2\2\2\u02e6\u02e7\3\2\2\2\u02e7e\3\2\2\2\u02e8\u02e9")
 *         buf.write("\7\66\2\2\u02e9\u02ea\5n8\2\u02ea\u02f0\7\64\2\2\u02eb")
 *         buf.write("\u02ec\5X-\2\u02ec\u02ed\7\3\2\2\u02ed\u02ef\3\2\2\2\u02ee")             # <<<<<<<<<<<<<<
 *         buf.write("\u02eb\3\2\2\2\u02ef\u02f2\3\2\2\2\u02f0\u02ee\3\2\2\2")
 *         buf.write("\u02f0\u02f1\3\2\2\2\u02f1\u02f3\3\2\2\2\u02f2\u02f0\3")
 */
  __pyx_tuple__603 = PyTuple_Pack(1, __pyx_kp_u_X_4); if (unlikely(!__pyx_tuple__603)) __PYX_ERR(0, 392, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__603);
  __Pyx_GIVEREF(__pyx_tuple__603);

  /* "pymola/generated/ModelicaParser.pyx":393
 *         buf.write("\7\66\2\2\u02e9\u02ea\5n8\2\u02ea\u02f0\7\64\2\2\u02eb")
 *         buf.write("\u02ec\5X-\2\u02ec\u02ed\7\3\2\2\u02ed\u02ef\3\2\2\2\u02ee")
 *         buf.write("\u02eb\3\2\2\2\u02ef\u02f2\3\2\2\2\u02f0\u02ee\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u02f0\u02f1\3\2\2\2\u02f1\u02f3\3\2\2\2\u02f2\u02f0\3")
 *         buf.write("\2\2\2\u02f3\u02f4\7\20\2\2\u02f4\u02f5\7\66\2\2\u02f5")
 */
  __pyx_tuple__605 = PyTuple_Pack(1, __pyx_kp_u__604); if (unlikely(!__pyx_tuple__605)) __PYX_ERR(0, 393, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__605);
  __Pyx_GIVEREF(__pyx_tuple__605);

  /* "pymola/generated/ModelicaParser.pyx":394
 *         buf.write("\u02ec\5X-\2\u02ec\u02ed\7\3\2\2\u02ed\u02ef\3\2\2\2\u02ee")
 *         buf.write("\u02eb\3\2\2\2\u02ef\u02f2\3\2\2\2\u02f0\u02ee\3\2\2\2")
 *         buf.write("\u02f0\u02f1\3\2\2\2\u02f1\u02f3\3\2\2\2\u02f2\u02f0\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u02f3\u02f4\7\20\2\2\u02f4\u02f5\7\66\2\2\u02f5")
 *         buf.write("g\3\2\2\2\u02f6\u02f7\7\67\2\2\u02f7\u02f8\5n8\2\u02f8")
 */
  __pyx_tuple__607 = PyTuple_Pack(1, __pyx_kp_u__606); if (unlikely(!__pyx_tuple__607)) __PYX_ERR(0, 394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__607);
  __Pyx_GIVEREF(__pyx_tuple__607);

  /* "pymola/generated/ModelicaParser.pyx":395
 *         buf.write("\u02eb\3\2\2\2\u02ef\u02f2\3\2\2\2\u02f0\u02ee\3\2\2\2")
 *         buf.write("\u02f0\u02f1\3\2\2\2\u02f1\u02f3\3\2\2\2\u02f2\u02f0\3")
 *         buf.write("\2\2\2\u02f3\u02f4\7\20\2\2\u02f4\u02f5\7\66\2\2\u02f5")             # <<<<<<<<<<<<<<
 *         buf.write("g\3\2\2\2\u02f6\u02f7\7\67\2\2\u02f7\u02f8\5n8\2\u02f8")
 *         buf.write("\u02fe\7\60\2\2\u02f9\u02fa\5T+\2\u02fa\u02fb\7\3\2\2")
 */
  __pyx_tuple__608 = PyTuple_Pack(1, __pyx_kp_u_6); if (unlikely(!__pyx_tuple__608)) __PYX_ERR(0, 395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__608);
  __Pyx_GIVEREF(__pyx_tuple__608);

  /* "pymola/generated/ModelicaParser.pyx":396
 *         buf.write("\u02f0\u02f1\3\2\2\2\u02f1\u02f3\3\2\2\2\u02f2\u02f0\3")
 *         buf.write("\2\2\2\u02f3\u02f4\7\20\2\2\u02f4\u02f5\7\66\2\2\u02f5")
 *         buf.write("g\3\2\2\2\u02f6\u02f7\7\67\2\2\u02f7\u02f8\5n8\2\u02f8")             # <<<<<<<<<<<<<<
 *         buf.write("\u02fe\7\60\2\2\u02f9\u02fa\5T+\2\u02fa\u02fb\7\3\2\2")
 *         buf.write("\u02fb\u02fd\3\2\2\2\u02fc\u02f9\3\2\2\2\u02fd\u0300\3")
 */
  __pyx_tuple__609 = PyTuple_Pack(1, __pyx_kp_u_g_7_n8); if (unlikely(!__pyx_tuple__609)) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__609);
  __Pyx_GIVEREF(__pyx_tuple__609);

  /* "pymola/generated/ModelicaParser.pyx":397
 *         buf.write("\2\2\2\u02f3\u02f4\7\20\2\2\u02f4\u02f5\7\66\2\2\u02f5")
 *         buf.write("g\3\2\2\2\u02f6\u02f7\7\67\2\2\u02f7\u02f8\5n8\2\u02f8")
 *         buf.write("\u02fe\7\60\2\2\u02f9\u02fa\5T+\2\u02fa\u02fb\7\3\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u02fb\u02fd\3\2\2\2\u02fc\u02f9\3\2\2\2\u02fd\u0300\3")
 *         buf.write("\2\2\2\u02fe\u02fc\3\2\2\2\u02fe\u02ff\3\2\2\2\u02ff\u030e")
 */
  __pyx_tuple__610 = PyTuple_Pack(1, __pyx_kp_u_0_T_2); if (unlikely(!__pyx_tuple__610)) __PYX_ERR(0, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__610);
  __Pyx_GIVEREF(__pyx_tuple__610);

  /* "pymola/generated/ModelicaParser.pyx":398
 *         buf.write("g\3\2\2\2\u02f6\u02f7\7\67\2\2\u02f7\u02f8\5n8\2\u02f8")
 *         buf.write("\u02fe\7\60\2\2\u02f9\u02fa\5T+\2\u02fa\u02fb\7\3\2\2")
 *         buf.write("\u02fb\u02fd\3\2\2\2\u02fc\u02f9\3\2\2\2\u02fd\u0300\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u02fe\u02fc\3\2\2\2\u02fe\u02ff\3\2\2\2\u02ff\u030e")
 *         buf.write("\3\2\2\2\u0300\u02fe\3\2\2\2\u0301\u0302\78\2\2\u0302")
 */
  __pyx_tuple__612 = PyTuple_Pack(1, __pyx_kp_u__611); if (unlikely(!__pyx_tuple__612)) __PYX_ERR(0, 398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__612);
  __Pyx_GIVEREF(__pyx_tuple__612);

  /* "pymola/generated/ModelicaParser.pyx":399
 *         buf.write("\u02fe\7\60\2\2\u02f9\u02fa\5T+\2\u02fa\u02fb\7\3\2\2")
 *         buf.write("\u02fb\u02fd\3\2\2\2\u02fc\u02f9\3\2\2\2\u02fd\u0300\3")
 *         buf.write("\2\2\2\u02fe\u02fc\3\2\2\2\u02fe\u02ff\3\2\2\2\u02ff\u030e")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0300\u02fe\3\2\2\2\u0301\u0302\78\2\2\u0302")
 *         buf.write("\u0303\5n8\2\u0303\u0309\7\60\2\2\u0304\u0305\5T+\2\u0305")
 */
  __pyx_tuple__614 = PyTuple_Pack(1, __pyx_kp_u__613); if (unlikely(!__pyx_tuple__614)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__614);
  __Pyx_GIVEREF(__pyx_tuple__614);

  /* "pymola/generated/ModelicaParser.pyx":400
 *         buf.write("\u02fb\u02fd\3\2\2\2\u02fc\u02f9\3\2\2\2\u02fd\u0300\3")
 *         buf.write("\2\2\2\u02fe\u02fc\3\2\2\2\u02fe\u02ff\3\2\2\2\u02ff\u030e")
 *         buf.write("\3\2\2\2\u0300\u02fe\3\2\2\2\u0301\u0302\78\2\2\u0302")             # <<<<<<<<<<<<<<
 *         buf.write("\u0303\5n8\2\u0303\u0309\7\60\2\2\u0304\u0305\5T+\2\u0305")
 *         buf.write("\u0306\7\3\2\2\u0306\u0308\3\2\2\2\u0307\u0304\3\2\2\2")
 */
  __pyx_tuple__615 = PyTuple_Pack(1, __pyx_kp_u_8_2); if (unlikely(!__pyx_tuple__615)) __PYX_ERR(0, 400, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__615);
  __Pyx_GIVEREF(__pyx_tuple__615);

  /* "pymola/generated/ModelicaParser.pyx":401
 *         buf.write("\2\2\2\u02fe\u02fc\3\2\2\2\u02fe\u02ff\3\2\2\2\u02ff\u030e")
 *         buf.write("\3\2\2\2\u0300\u02fe\3\2\2\2\u0301\u0302\78\2\2\u0302")
 *         buf.write("\u0303\5n8\2\u0303\u0309\7\60\2\2\u0304\u0305\5T+\2\u0305")             # <<<<<<<<<<<<<<
 *         buf.write("\u0306\7\3\2\2\u0306\u0308\3\2\2\2\u0307\u0304\3\2\2\2")
 *         buf.write("\u0308\u030b\3\2\2\2\u0309\u0307\3\2\2\2\u0309\u030a\3")
 */
  __pyx_tuple__616 = PyTuple_Pack(1, __pyx_kp_u_n8_0_T); if (unlikely(!__pyx_tuple__616)) __PYX_ERR(0, 401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__616);
  __Pyx_GIVEREF(__pyx_tuple__616);

  /* "pymola/generated/ModelicaParser.pyx":402
 *         buf.write("\3\2\2\2\u0300\u02fe\3\2\2\2\u0301\u0302\78\2\2\u0302")
 *         buf.write("\u0303\5n8\2\u0303\u0309\7\60\2\2\u0304\u0305\5T+\2\u0305")
 *         buf.write("\u0306\7\3\2\2\u0306\u0308\3\2\2\2\u0307\u0304\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0308\u030b\3\2\2\2\u0309\u0307\3\2\2\2\u0309\u030a\3")
 *         buf.write("\2\2\2\u030a\u030d\3\2\2\2\u030b\u0309\3\2\2\2\u030c\u0301")
 */
  __pyx_tuple__618 = PyTuple_Pack(1, __pyx_kp_u__617); if (unlikely(!__pyx_tuple__618)) __PYX_ERR(0, 402, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__618);
  __Pyx_GIVEREF(__pyx_tuple__618);

  /* "pymola/generated/ModelicaParser.pyx":403
 *         buf.write("\u0303\5n8\2\u0303\u0309\7\60\2\2\u0304\u0305\5T+\2\u0305")
 *         buf.write("\u0306\7\3\2\2\u0306\u0308\3\2\2\2\u0307\u0304\3\2\2\2")
 *         buf.write("\u0308\u030b\3\2\2\2\u0309\u0307\3\2\2\2\u0309\u030a\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u030a\u030d\3\2\2\2\u030b\u0309\3\2\2\2\u030c\u0301")
 *         buf.write("\3\2\2\2\u030d\u0310\3\2\2\2\u030e\u030c\3\2\2\2\u030e")
 */
  __pyx_tuple__620 = PyTuple_Pack(1, __pyx_kp_u__619); if (unlikely(!__pyx_tuple__620)) __PYX_ERR(0, 403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__620);
  __Pyx_GIVEREF(__pyx_tuple__620);

  /* "pymola/generated/ModelicaParser.pyx":404
 *         buf.write("\u0306\7\3\2\2\u0306\u0308\3\2\2\2\u0307\u0304\3\2\2\2")
 *         buf.write("\u0308\u030b\3\2\2\2\u0309\u0307\3\2\2\2\u0309\u030a\3")
 *         buf.write("\2\2\2\u030a\u030d\3\2\2\2\u030b\u0309\3\2\2\2\u030c\u0301")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u030d\u0310\3\2\2\2\u030e\u030c\3\2\2\2\u030e")
 *         buf.write("\u030f\3\2\2\2\u030f\u0311\3\2\2\2\u0310\u030e\3\2\2\2")
 */
  __pyx_tuple__622 = PyTuple_Pack(1, __pyx_kp_u__621); if (unlikely(!__pyx_tuple__622)) __PYX_ERR(0, 404, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__622);
  __Pyx_GIVEREF(__pyx_tuple__622);

  /* "pymola/generated/ModelicaParser.pyx":405
 *         buf.write("\u0308\u030b\3\2\2\2\u0309\u0307\3\2\2\2\u0309\u030a\3")
 *         buf.write("\2\2\2\u030a\u030d\3\2\2\2\u030b\u0309\3\2\2\2\u030c\u0301")
 *         buf.write("\3\2\2\2\u030d\u0310\3\2\2\2\u030e\u030c\3\2\2\2\u030e")             # <<<<<<<<<<<<<<
 *         buf.write("\u030f\3\2\2\2\u030f\u0311\3\2\2\2\u0310\u030e\3\2\2\2")
 *         buf.write("\u0311\u0312\7\20\2\2\u0312\u0313\7\67\2\2\u0313i\3\2")
 */
  __pyx_tuple__624 = PyTuple_Pack(1, __pyx_kp_u__623); if (unlikely(!__pyx_tuple__624)) __PYX_ERR(0, 405, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__624);
  __Pyx_GIVEREF(__pyx_tuple__624);

  /* "pymola/generated/ModelicaParser.pyx":406
 *         buf.write("\2\2\2\u030a\u030d\3\2\2\2\u030b\u0309\3\2\2\2\u030c\u0301")
 *         buf.write("\3\2\2\2\u030d\u0310\3\2\2\2\u030e\u030c\3\2\2\2\u030e")
 *         buf.write("\u030f\3\2\2\2\u030f\u0311\3\2\2\2\u0310\u030e\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0311\u0312\7\20\2\2\u0312\u0313\7\67\2\2\u0313i\3\2")
 *         buf.write("\2\2\u0314\u0315\7\67\2\2\u0315\u0316\5n8\2\u0316\u031c")
 */
  __pyx_tuple__626 = PyTuple_Pack(1, __pyx_kp_u__625); if (unlikely(!__pyx_tuple__626)) __PYX_ERR(0, 406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__626);
  __Pyx_GIVEREF(__pyx_tuple__626);

  /* "pymola/generated/ModelicaParser.pyx":407
 *         buf.write("\3\2\2\2\u030d\u0310\3\2\2\2\u030e\u030c\3\2\2\2\u030e")
 *         buf.write("\u030f\3\2\2\2\u030f\u0311\3\2\2\2\u0310\u030e\3\2\2\2")
 *         buf.write("\u0311\u0312\7\20\2\2\u0312\u0313\7\67\2\2\u0313i\3\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\u0314\u0315\7\67\2\2\u0315\u0316\5n8\2\u0316\u031c")
 *         buf.write("\7\60\2\2\u0317\u0318\5X-\2\u0318\u0319\7\3\2\2\u0319")
 */
  __pyx_tuple__627 = PyTuple_Pack(1, __pyx_kp_u_7_i); if (unlikely(!__pyx_tuple__627)) __PYX_ERR(0, 407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__627);
  __Pyx_GIVEREF(__pyx_tuple__627);

  /* "pymola/generated/ModelicaParser.pyx":408
 *         buf.write("\u030f\3\2\2\2\u030f\u0311\3\2\2\2\u0310\u030e\3\2\2\2")
 *         buf.write("\u0311\u0312\7\20\2\2\u0312\u0313\7\67\2\2\u0313i\3\2")
 *         buf.write("\2\2\u0314\u0315\7\67\2\2\u0315\u0316\5n8\2\u0316\u031c")             # <<<<<<<<<<<<<<
 *         buf.write("\7\60\2\2\u0317\u0318\5X-\2\u0318\u0319\7\3\2\2\u0319")
 *         buf.write("\u031b\3\2\2\2\u031a\u0317\3\2\2\2\u031b\u031e\3\2\2\2")
 */
  __pyx_tuple__628 = PyTuple_Pack(1, __pyx_kp_u_7_n8); if (unlikely(!__pyx_tuple__628)) __PYX_ERR(0, 408, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__628);
  __Pyx_GIVEREF(__pyx_tuple__628);

  /* "pymola/generated/ModelicaParser.pyx":409
 *         buf.write("\u0311\u0312\7\20\2\2\u0312\u0313\7\67\2\2\u0313i\3\2")
 *         buf.write("\2\2\u0314\u0315\7\67\2\2\u0315\u0316\5n8\2\u0316\u031c")
 *         buf.write("\7\60\2\2\u0317\u0318\5X-\2\u0318\u0319\7\3\2\2\u0319")             # <<<<<<<<<<<<<<
 *         buf.write("\u031b\3\2\2\2\u031a\u0317\3\2\2\2\u031b\u031e\3\2\2\2")
 *         buf.write("\u031c\u031a\3\2\2\2\u031c\u031d\3\2\2\2\u031d\u032c\3")
 */
  __pyx_tuple__629 = PyTuple_Pack(1, __pyx_kp_u_0_X_2); if (unlikely(!__pyx_tuple__629)) __PYX_ERR(0, 409, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__629);
  __Pyx_GIVEREF(__pyx_tuple__629);

  /* "pymola/generated/ModelicaParser.pyx":410
 *         buf.write("\2\2\u0314\u0315\7\67\2\2\u0315\u0316\5n8\2\u0316\u031c")
 *         buf.write("\7\60\2\2\u0317\u0318\5X-\2\u0318\u0319\7\3\2\2\u0319")
 *         buf.write("\u031b\3\2\2\2\u031a\u0317\3\2\2\2\u031b\u031e\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u031c\u031a\3\2\2\2\u031c\u031d\3\2\2\2\u031d\u032c\3")
 *         buf.write("\2\2\2\u031e\u031c\3\2\2\2\u031f\u0320\78\2\2\u0320\u0321")
 */
  __pyx_tuple__631 = PyTuple_Pack(1, __pyx_kp_u__630); if (unlikely(!__pyx_tuple__631)) __PYX_ERR(0, 410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__631);
  __Pyx_GIVEREF(__pyx_tuple__631);

  /* "pymola/generated/ModelicaParser.pyx":411
 *         buf.write("\7\60\2\2\u0317\u0318\5X-\2\u0318\u0319\7\3\2\2\u0319")
 *         buf.write("\u031b\3\2\2\2\u031a\u0317\3\2\2\2\u031b\u031e\3\2\2\2")
 *         buf.write("\u031c\u031a\3\2\2\2\u031c\u031d\3\2\2\2\u031d\u032c\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u031e\u031c\3\2\2\2\u031f\u0320\78\2\2\u0320\u0321")
 *         buf.write("\5n8\2\u0321\u0327\7\60\2\2\u0322\u0323\5X-\2\u0323\u0324")
 */
  __pyx_tuple__633 = PyTuple_Pack(1, __pyx_kp_u__632); if (unlikely(!__pyx_tuple__633)) __PYX_ERR(0, 411, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__633);
  __Pyx_GIVEREF(__pyx_tuple__633);

  /* "pymola/generated/ModelicaParser.pyx":412
 *         buf.write("\u031b\3\2\2\2\u031a\u0317\3\2\2\2\u031b\u031e\3\2\2\2")
 *         buf.write("\u031c\u031a\3\2\2\2\u031c\u031d\3\2\2\2\u031d\u032c\3")
 *         buf.write("\2\2\2\u031e\u031c\3\2\2\2\u031f\u0320\78\2\2\u0320\u0321")             # <<<<<<<<<<<<<<
 *         buf.write("\5n8\2\u0321\u0327\7\60\2\2\u0322\u0323\5X-\2\u0323\u0324")
 *         buf.write("\7\3\2\2\u0324\u0326\3\2\2\2\u0325\u0322\3\2\2\2\u0326")
 */
  __pyx_tuple__634 = PyTuple_Pack(1, __pyx_kp_u_8_3); if (unlikely(!__pyx_tuple__634)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__634);
  __Pyx_GIVEREF(__pyx_tuple__634);

  /* "pymola/generated/ModelicaParser.pyx":413
 *         buf.write("\u031c\u031a\3\2\2\2\u031c\u031d\3\2\2\2\u031d\u032c\3")
 *         buf.write("\2\2\2\u031e\u031c\3\2\2\2\u031f\u0320\78\2\2\u0320\u0321")
 *         buf.write("\5n8\2\u0321\u0327\7\60\2\2\u0322\u0323\5X-\2\u0323\u0324")             # <<<<<<<<<<<<<<
 *         buf.write("\7\3\2\2\u0324\u0326\3\2\2\2\u0325\u0322\3\2\2\2\u0326")
 *         buf.write("\u0329\3\2\2\2\u0327\u0325\3\2\2\2\u0327\u0328\3\2\2\2")
 */
  __pyx_tuple__635 = PyTuple_Pack(1, __pyx_kp_u_n8_0_X_2); if (unlikely(!__pyx_tuple__635)) __PYX_ERR(0, 413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__635);
  __Pyx_GIVEREF(__pyx_tuple__635);

  /* "pymola/generated/ModelicaParser.pyx":414
 *         buf.write("\2\2\2\u031e\u031c\3\2\2\2\u031f\u0320\78\2\2\u0320\u0321")
 *         buf.write("\5n8\2\u0321\u0327\7\60\2\2\u0322\u0323\5X-\2\u0323\u0324")
 *         buf.write("\7\3\2\2\u0324\u0326\3\2\2\2\u0325\u0322\3\2\2\2\u0326")             # <<<<<<<<<<<<<<
 *         buf.write("\u0329\3\2\2\2\u0327\u0325\3\2\2\2\u0327\u0328\3\2\2\2")
 *         buf.write("\u0328\u032b\3\2\2\2\u0329\u0327\3\2\2\2\u032a\u031f\3")
 */
  __pyx_tuple__637 = PyTuple_Pack(1, __pyx_kp_u__636); if (unlikely(!__pyx_tuple__637)) __PYX_ERR(0, 414, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__637);
  __Pyx_GIVEREF(__pyx_tuple__637);

  /* "pymola/generated/ModelicaParser.pyx":415
 *         buf.write("\5n8\2\u0321\u0327\7\60\2\2\u0322\u0323\5X-\2\u0323\u0324")
 *         buf.write("\7\3\2\2\u0324\u0326\3\2\2\2\u0325\u0322\3\2\2\2\u0326")
 *         buf.write("\u0329\3\2\2\2\u0327\u0325\3\2\2\2\u0327\u0328\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0328\u032b\3\2\2\2\u0329\u0327\3\2\2\2\u032a\u031f\3")
 *         buf.write("\2\2\2\u032b\u032e\3\2\2\2\u032c\u032a\3\2\2\2\u032c\u032d")
 */
  __pyx_tuple__639 = PyTuple_Pack(1, __pyx_kp_u__638); if (unlikely(!__pyx_tuple__639)) __PYX_ERR(0, 415, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__639);
  __Pyx_GIVEREF(__pyx_tuple__639);

  /* "pymola/generated/ModelicaParser.pyx":416
 *         buf.write("\7\3\2\2\u0324\u0326\3\2\2\2\u0325\u0322\3\2\2\2\u0326")
 *         buf.write("\u0329\3\2\2\2\u0327\u0325\3\2\2\2\u0327\u0328\3\2\2\2")
 *         buf.write("\u0328\u032b\3\2\2\2\u0329\u0327\3\2\2\2\u032a\u031f\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u032b\u032e\3\2\2\2\u032c\u032a\3\2\2\2\u032c\u032d")
 *         buf.write("\3\2\2\2\u032d\u032f\3\2\2\2\u032e\u032c\3\2\2\2\u032f")
 */
  __pyx_tuple__641 = PyTuple_Pack(1, __pyx_kp_u__640); if (unlikely(!__pyx_tuple__641)) __PYX_ERR(0, 416, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__641);
  __Pyx_GIVEREF(__pyx_tuple__641);

  /* "pymola/generated/ModelicaParser.pyx":417
 *         buf.write("\u0329\3\2\2\2\u0327\u0325\3\2\2\2\u0327\u0328\3\2\2\2")
 *         buf.write("\u0328\u032b\3\2\2\2\u0329\u0327\3\2\2\2\u032a\u031f\3")
 *         buf.write("\2\2\2\u032b\u032e\3\2\2\2\u032c\u032a\3\2\2\2\u032c\u032d")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u032d\u032f\3\2\2\2\u032e\u032c\3\2\2\2\u032f")
 *         buf.write("\u0330\7\20\2\2\u0330\u0331\7\67\2\2\u0331k\3\2\2\2\u0332")
 */
  __pyx_tuple__643 = PyTuple_Pack(1, __pyx_kp_u__642); if (unlikely(!__pyx_tuple__643)) __PYX_ERR(0, 417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__643);
  __Pyx_GIVEREF(__pyx_tuple__643);

  /* "pymola/generated/ModelicaParser.pyx":418
 *         buf.write("\u0328\u032b\3\2\2\2\u0329\u0327\3\2\2\2\u032a\u031f\3")
 *         buf.write("\2\2\2\u032b\u032e\3\2\2\2\u032c\u032a\3\2\2\2\u032c\u032d")
 *         buf.write("\3\2\2\2\u032d\u032f\3\2\2\2\u032e\u032c\3\2\2\2\u032f")             # <<<<<<<<<<<<<<
 *         buf.write("\u0330\7\20\2\2\u0330\u0331\7\67\2\2\u0331k\3\2\2\2\u0332")
 *         buf.write("\u0333\79\2\2\u0333\u0334\7\23\2\2\u0334\u0335\5z>\2\u0335")
 */
  __pyx_tuple__645 = PyTuple_Pack(1, __pyx_kp_u__644); if (unlikely(!__pyx_tuple__645)) __PYX_ERR(0, 418, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__645);
  __Pyx_GIVEREF(__pyx_tuple__645);

  /* "pymola/generated/ModelicaParser.pyx":419
 *         buf.write("\2\2\2\u032b\u032e\3\2\2\2\u032c\u032a\3\2\2\2\u032c\u032d")
 *         buf.write("\3\2\2\2\u032d\u032f\3\2\2\2\u032e\u032c\3\2\2\2\u032f")
 *         buf.write("\u0330\7\20\2\2\u0330\u0331\7\67\2\2\u0331k\3\2\2\2\u0332")             # <<<<<<<<<<<<<<
 *         buf.write("\u0333\79\2\2\u0333\u0334\7\23\2\2\u0334\u0335\5z>\2\u0335")
 *         buf.write("\u0336\7\27\2\2\u0336\u0337\5z>\2\u0337\u0338\7\25\2\2")
 */
  __pyx_tuple__646 = PyTuple_Pack(1, __pyx_kp_u_7_k); if (unlikely(!__pyx_tuple__646)) __PYX_ERR(0, 419, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__646);
  __Pyx_GIVEREF(__pyx_tuple__646);

  /* "pymola/generated/ModelicaParser.pyx":420
 *         buf.write("\3\2\2\2\u032d\u032f\3\2\2\2\u032e\u032c\3\2\2\2\u032f")
 *         buf.write("\u0330\7\20\2\2\u0330\u0331\7\67\2\2\u0331k\3\2\2\2\u0332")
 *         buf.write("\u0333\79\2\2\u0333\u0334\7\23\2\2\u0334\u0335\5z>\2\u0335")             # <<<<<<<<<<<<<<
 *         buf.write("\u0336\7\27\2\2\u0336\u0337\5z>\2\u0337\u0338\7\25\2\2")
 *         buf.write("\u0338m\3\2\2\2\u0339\u034c\5p9\2\u033a\u033b\7*\2\2\u033b")
 */
  __pyx_tuple__647 = PyTuple_Pack(1, __pyx_kp_u_9_z); if (unlikely(!__pyx_tuple__647)) __PYX_ERR(0, 420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__647);
  __Pyx_GIVEREF(__pyx_tuple__647);

  /* "pymola/generated/ModelicaParser.pyx":421
 *         buf.write("\u0330\7\20\2\2\u0330\u0331\7\67\2\2\u0331k\3\2\2\2\u0332")
 *         buf.write("\u0333\79\2\2\u0333\u0334\7\23\2\2\u0334\u0335\5z>\2\u0335")
 *         buf.write("\u0336\7\27\2\2\u0336\u0337\5z>\2\u0337\u0338\7\25\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0338m\3\2\2\2\u0339\u034c\5p9\2\u033a\u033b\7*\2\2\u033b")
 *         buf.write("\u033c\5n8\2\u033c\u033d\7\60\2\2\u033d\u0345\5n8\2\u033e")
 */
  __pyx_tuple__648 = PyTuple_Pack(1, __pyx_kp_u_z_10); if (unlikely(!__pyx_tuple__648)) __PYX_ERR(0, 421, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__648);
  __Pyx_GIVEREF(__pyx_tuple__648);

  /* "pymola/generated/ModelicaParser.pyx":422
 *         buf.write("\u0333\79\2\2\u0333\u0334\7\23\2\2\u0334\u0335\5z>\2\u0335")
 *         buf.write("\u0336\7\27\2\2\u0336\u0337\5z>\2\u0337\u0338\7\25\2\2")
 *         buf.write("\u0338m\3\2\2\2\u0339\u034c\5p9\2\u033a\u033b\7*\2\2\u033b")             # <<<<<<<<<<<<<<
 *         buf.write("\u033c\5n8\2\u033c\u033d\7\60\2\2\u033d\u0345\5n8\2\u033e")
 *         buf.write("\u033f\7\61\2\2\u033f\u0340\5n8\2\u0340\u0341\7\60\2\2")
 */
  __pyx_tuple__649 = PyTuple_Pack(1, __pyx_kp_u_m_p9); if (unlikely(!__pyx_tuple__649)) __PYX_ERR(0, 422, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__649);
  __Pyx_GIVEREF(__pyx_tuple__649);

  /* "pymola/generated/ModelicaParser.pyx":423
 *         buf.write("\u0336\7\27\2\2\u0336\u0337\5z>\2\u0337\u0338\7\25\2\2")
 *         buf.write("\u0338m\3\2\2\2\u0339\u034c\5p9\2\u033a\u033b\7*\2\2\u033b")
 *         buf.write("\u033c\5n8\2\u033c\u033d\7\60\2\2\u033d\u0345\5n8\2\u033e")             # <<<<<<<<<<<<<<
 *         buf.write("\u033f\7\61\2\2\u033f\u0340\5n8\2\u0340\u0341\7\60\2\2")
 *         buf.write("\u0341\u0342\5n8\2\u0342\u0344\3\2\2\2\u0343\u033e\3\2")
 */
  __pyx_tuple__650 = PyTuple_Pack(1, __pyx_kp_u_n8_0_n8); if (unlikely(!__pyx_tuple__650)) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__650);
  __Pyx_GIVEREF(__pyx_tuple__650);

  /* "pymola/generated/ModelicaParser.pyx":424
 *         buf.write("\u0338m\3\2\2\2\u0339\u034c\5p9\2\u033a\u033b\7*\2\2\u033b")
 *         buf.write("\u033c\5n8\2\u033c\u033d\7\60\2\2\u033d\u0345\5n8\2\u033e")
 *         buf.write("\u033f\7\61\2\2\u033f\u0340\5n8\2\u0340\u0341\7\60\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0341\u0342\5n8\2\u0342\u0344\3\2\2\2\u0343\u033e\3\2")
 *         buf.write("\2\2\u0344\u0347\3\2\2\2\u0345\u0343\3\2\2\2\u0345\u0346")
 */
  __pyx_tuple__651 = PyTuple_Pack(1, __pyx_kp_u_1_n8_0); if (unlikely(!__pyx_tuple__651)) __PYX_ERR(0, 424, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__651);
  __Pyx_GIVEREF(__pyx_tuple__651);

  /* "pymola/generated/ModelicaParser.pyx":425
 *         buf.write("\u033c\5n8\2\u033c\u033d\7\60\2\2\u033d\u0345\5n8\2\u033e")
 *         buf.write("\u033f\7\61\2\2\u033f\u0340\5n8\2\u0340\u0341\7\60\2\2")
 *         buf.write("\u0341\u0342\5n8\2\u0342\u0344\3\2\2\2\u0343\u033e\3\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\u0344\u0347\3\2\2\2\u0345\u0343\3\2\2\2\u0345\u0346")
 *         buf.write("\3\2\2\2\u0346\u0348\3\2\2\2\u0347\u0345\3\2\2\2\u0348")
 */
  __pyx_tuple__652 = PyTuple_Pack(1, __pyx_kp_u_n8_6); if (unlikely(!__pyx_tuple__652)) __PYX_ERR(0, 425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__652);
  __Pyx_GIVEREF(__pyx_tuple__652);

  /* "pymola/generated/ModelicaParser.pyx":426
 *         buf.write("\u033f\7\61\2\2\u033f\u0340\5n8\2\u0340\u0341\7\60\2\2")
 *         buf.write("\u0341\u0342\5n8\2\u0342\u0344\3\2\2\2\u0343\u033e\3\2")
 *         buf.write("\2\2\u0344\u0347\3\2\2\2\u0345\u0343\3\2\2\2\u0345\u0346")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0346\u0348\3\2\2\2\u0347\u0345\3\2\2\2\u0348")
 *         buf.write("\u0349\7\62\2\2\u0349\u034a\5n8\2\u034a\u034c\3\2\2\2")
 */
  __pyx_tuple__654 = PyTuple_Pack(1, __pyx_kp_u__653); if (unlikely(!__pyx_tuple__654)) __PYX_ERR(0, 426, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__654);
  __Pyx_GIVEREF(__pyx_tuple__654);

  /* "pymola/generated/ModelicaParser.pyx":427
 *         buf.write("\u0341\u0342\5n8\2\u0342\u0344\3\2\2\2\u0343\u033e\3\2")
 *         buf.write("\2\2\u0344\u0347\3\2\2\2\u0345\u0343\3\2\2\2\u0345\u0346")
 *         buf.write("\3\2\2\2\u0346\u0348\3\2\2\2\u0347\u0345\3\2\2\2\u0348")             # <<<<<<<<<<<<<<
 *         buf.write("\u0349\7\62\2\2\u0349\u034a\5n8\2\u034a\u034c\3\2\2\2")
 *         buf.write("\u034b\u0339\3\2\2\2\u034b\u033a\3\2\2\2\u034co\3\2\2")
 */
  __pyx_tuple__656 = PyTuple_Pack(1, __pyx_kp_u__655); if (unlikely(!__pyx_tuple__656)) __PYX_ERR(0, 427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__656);
  __Pyx_GIVEREF(__pyx_tuple__656);

  /* "pymola/generated/ModelicaParser.pyx":428
 *         buf.write("\2\2\u0344\u0347\3\2\2\2\u0345\u0343\3\2\2\2\u0345\u0346")
 *         buf.write("\3\2\2\2\u0346\u0348\3\2\2\2\u0347\u0345\3\2\2\2\u0348")
 *         buf.write("\u0349\7\62\2\2\u0349\u034a\5n8\2\u034a\u034c\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u034b\u0339\3\2\2\2\u034b\u033a\3\2\2\2\u034co\3\2\2")
 *         buf.write("\2\u034d\u0354\5r:\2\u034e\u034f\7\24\2\2\u034f\u0352")
 */
  __pyx_tuple__657 = PyTuple_Pack(1, __pyx_kp_u_2_n8); if (unlikely(!__pyx_tuple__657)) __PYX_ERR(0, 428, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__657);
  __Pyx_GIVEREF(__pyx_tuple__657);

  /* "pymola/generated/ModelicaParser.pyx":429
 *         buf.write("\3\2\2\2\u0346\u0348\3\2\2\2\u0347\u0345\3\2\2\2\u0348")
 *         buf.write("\u0349\7\62\2\2\u0349\u034a\5n8\2\u034a\u034c\3\2\2\2")
 *         buf.write("\u034b\u0339\3\2\2\2\u034b\u033a\3\2\2\2\u034co\3\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u034d\u0354\5r:\2\u034e\u034f\7\24\2\2\u034f\u0352")
 *         buf.write("\5r:\2\u0350\u0351\7\24\2\2\u0351\u0353\5r:\2\u0352\u0350")
 */
  __pyx_tuple__658 = PyTuple_Pack(1, __pyx_kp_u_o); if (unlikely(!__pyx_tuple__658)) __PYX_ERR(0, 429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__658);
  __Pyx_GIVEREF(__pyx_tuple__658);

  /* "pymola/generated/ModelicaParser.pyx":430
 *         buf.write("\u0349\7\62\2\2\u0349\u034a\5n8\2\u034a\u034c\3\2\2\2")
 *         buf.write("\u034b\u0339\3\2\2\2\u034b\u033a\3\2\2\2\u034co\3\2\2")
 *         buf.write("\2\u034d\u0354\5r:\2\u034e\u034f\7\24\2\2\u034f\u0352")             # <<<<<<<<<<<<<<
 *         buf.write("\5r:\2\u0350\u0351\7\24\2\2\u0351\u0353\5r:\2\u0352\u0350")
 *         buf.write("\3\2\2\2\u0352\u0353\3\2\2\2\u0353\u0355\3\2\2\2\u0354")
 */
  __pyx_tuple__659 = PyTuple_Pack(1, __pyx_kp_u_r); if (unlikely(!__pyx_tuple__659)) __PYX_ERR(0, 430, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__659);
  __Pyx_GIVEREF(__pyx_tuple__659);

  /* "pymola/generated/ModelicaParser.pyx":431
 *         buf.write("\u034b\u0339\3\2\2\2\u034b\u033a\3\2\2\2\u034co\3\2\2")
 *         buf.write("\2\u034d\u0354\5r:\2\u034e\u034f\7\24\2\2\u034f\u0352")
 *         buf.write("\5r:\2\u0350\u0351\7\24\2\2\u0351\u0353\5r:\2\u0352\u0350")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0352\u0353\3\2\2\2\u0353\u0355\3\2\2\2\u0354")
 *         buf.write("\u034e\3\2\2\2\u0354\u0355\3\2\2\2\u0355q\3\2\2\2\u0356")
 */
  __pyx_tuple__660 = PyTuple_Pack(1, __pyx_kp_u_r_r); if (unlikely(!__pyx_tuple__660)) __PYX_ERR(0, 431, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__660);
  __Pyx_GIVEREF(__pyx_tuple__660);

  /* "pymola/generated/ModelicaParser.pyx":432
 *         buf.write("\2\u034d\u0354\5r:\2\u034e\u034f\7\24\2\2\u034f\u0352")
 *         buf.write("\5r:\2\u0350\u0351\7\24\2\2\u0351\u0353\5r:\2\u0352\u0350")
 *         buf.write("\3\2\2\2\u0352\u0353\3\2\2\2\u0353\u0355\3\2\2\2\u0354")             # <<<<<<<<<<<<<<
 *         buf.write("\u034e\3\2\2\2\u0354\u0355\3\2\2\2\u0355q\3\2\2\2\u0356")
 *         buf.write("\u0357\b:\1\2\u0357\u0358\7:\2\2\u0358\u0361\5r:\13\u0359")
 */
  __pyx_tuple__662 = PyTuple_Pack(1, __pyx_kp_u__661); if (unlikely(!__pyx_tuple__662)) __PYX_ERR(0, 432, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__662);
  __Pyx_GIVEREF(__pyx_tuple__662);

  /* "pymola/generated/ModelicaParser.pyx":433
 *         buf.write("\5r:\2\u0350\u0351\7\24\2\2\u0351\u0353\5r:\2\u0352\u0350")
 *         buf.write("\3\2\2\2\u0352\u0353\3\2\2\2\u0353\u0355\3\2\2\2\u0354")
 *         buf.write("\u034e\3\2\2\2\u0354\u0355\3\2\2\2\u0355q\3\2\2\2\u0356")             # <<<<<<<<<<<<<<
 *         buf.write("\u0357\b:\1\2\u0357\u0358\7:\2\2\u0358\u0361\5r:\13\u0359")
 *         buf.write("\u035a\5t;\2\u035a\u035b\t\6\2\2\u035b\u035c\5t;\2\u035c")
 */
  __pyx_tuple__663 = PyTuple_Pack(1, __pyx_kp_u_q); if (unlikely(!__pyx_tuple__663)) __PYX_ERR(0, 433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__663);
  __Pyx_GIVEREF(__pyx_tuple__663);

  /* "pymola/generated/ModelicaParser.pyx":434
 *         buf.write("\3\2\2\2\u0352\u0353\3\2\2\2\u0353\u0355\3\2\2\2\u0354")
 *         buf.write("\u034e\3\2\2\2\u0354\u0355\3\2\2\2\u0355q\3\2\2\2\u0356")
 *         buf.write("\u0357\b:\1\2\u0357\u0358\7:\2\2\u0358\u0361\5r:\13\u0359")             # <<<<<<<<<<<<<<
 *         buf.write("\u035a\5t;\2\u035a\u035b\t\6\2\2\u035b\u035c\5t;\2\u035c")
 *         buf.write("\u0361\3\2\2\2\u035d\u035e\7I\2\2\u035e\u0361\5r:\6\u035f")
 */
  __pyx_tuple__664 = PyTuple_Pack(1, __pyx_kp_u_r_2); if (unlikely(!__pyx_tuple__664)) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__664);
  __Pyx_GIVEREF(__pyx_tuple__664);

  /* "pymola/generated/ModelicaParser.pyx":435
 *         buf.write("\u034e\3\2\2\2\u0354\u0355\3\2\2\2\u0355q\3\2\2\2\u0356")
 *         buf.write("\u0357\b:\1\2\u0357\u0358\7:\2\2\u0358\u0361\5r:\13\u0359")
 *         buf.write("\u035a\5t;\2\u035a\u035b\t\6\2\2\u035b\u035c\5t;\2\u035c")             # <<<<<<<<<<<<<<
 *         buf.write("\u0361\3\2\2\2\u035d\u035e\7I\2\2\u035e\u0361\5r:\6\u035f")
 *         buf.write("\u0361\5t;\2\u0360\u0356\3\2\2\2\u0360\u0359\3\2\2\2\u0360")
 */
  __pyx_tuple__665 = PyTuple_Pack(1, __pyx_kp_u_t_t); if (unlikely(!__pyx_tuple__665)) __PYX_ERR(0, 435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__665);
  __Pyx_GIVEREF(__pyx_tuple__665);

  /* "pymola/generated/ModelicaParser.pyx":436
 *         buf.write("\u0357\b:\1\2\u0357\u0358\7:\2\2\u0358\u0361\5r:\13\u0359")
 *         buf.write("\u035a\5t;\2\u035a\u035b\t\6\2\2\u035b\u035c\5t;\2\u035c")
 *         buf.write("\u0361\3\2\2\2\u035d\u035e\7I\2\2\u035e\u0361\5r:\6\u035f")             # <<<<<<<<<<<<<<
 *         buf.write("\u0361\5t;\2\u0360\u0356\3\2\2\2\u0360\u0359\3\2\2\2\u0360")
 *         buf.write("\u035d\3\2\2\2\u0360\u035f\3\2\2\2\u0361\u0373\3\2\2\2")
 */
  __pyx_tuple__666 = PyTuple_Pack(1, __pyx_kp_u_I_r); if (unlikely(!__pyx_tuple__666)) __PYX_ERR(0, 436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__666);
  __Pyx_GIVEREF(__pyx_tuple__666);

  /* "pymola/generated/ModelicaParser.pyx":437
 *         buf.write("\u035a\5t;\2\u035a\u035b\t\6\2\2\u035b\u035c\5t;\2\u035c")
 *         buf.write("\u0361\3\2\2\2\u035d\u035e\7I\2\2\u035e\u0361\5r:\6\u035f")
 *         buf.write("\u0361\5t;\2\u0360\u0356\3\2\2\2\u0360\u0359\3\2\2\2\u0360")             # <<<<<<<<<<<<<<
 *         buf.write("\u035d\3\2\2\2\u0360\u035f\3\2\2\2\u0361\u0373\3\2\2\2")
 *         buf.write("\u0362\u0363\f\t\2\2\u0363\u0364\t\7\2\2\u0364\u0372\5")
 */
  __pyx_tuple__667 = PyTuple_Pack(1, __pyx_kp_u_t); if (unlikely(!__pyx_tuple__667)) __PYX_ERR(0, 437, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__667);
  __Pyx_GIVEREF(__pyx_tuple__667);

  /* "pymola/generated/ModelicaParser.pyx":438
 *         buf.write("\u0361\3\2\2\2\u035d\u035e\7I\2\2\u035e\u0361\5r:\6\u035f")
 *         buf.write("\u0361\5t;\2\u0360\u0356\3\2\2\2\u0360\u0359\3\2\2\2\u0360")
 *         buf.write("\u035d\3\2\2\2\u0360\u035f\3\2\2\2\u0361\u0373\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0362\u0363\f\t\2\2\u0363\u0364\t\7\2\2\u0364\u0372\5")
 *         buf.write("r:\n\u0365\u0366\f\b\2\2\u0366\u0367\t\b\2\2\u0367\u0372")
 */
  __pyx_tuple__669 = PyTuple_Pack(1, __pyx_kp_u__668); if (unlikely(!__pyx_tuple__669)) __PYX_ERR(0, 438, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__669);
  __Pyx_GIVEREF(__pyx_tuple__669);

  /* "pymola/generated/ModelicaParser.pyx":439
 *         buf.write("\u0361\5t;\2\u0360\u0356\3\2\2\2\u0360\u0359\3\2\2\2\u0360")
 *         buf.write("\u035d\3\2\2\2\u0360\u035f\3\2\2\2\u0361\u0373\3\2\2\2")
 *         buf.write("\u0362\u0363\f\t\2\2\u0363\u0364\t\7\2\2\u0364\u0372\5")             # <<<<<<<<<<<<<<
 *         buf.write("r:\n\u0365\u0366\f\b\2\2\u0366\u0367\t\b\2\2\u0367\u0372")
 *         buf.write("\5r:\t\u0368\u0369\f\7\2\2\u0369\u036a\t\t\2\2\u036a\u0372")
 */
  __pyx_tuple__671 = PyTuple_Pack(1, __pyx_kp_u__670); if (unlikely(!__pyx_tuple__671)) __PYX_ERR(0, 439, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__671);
  __Pyx_GIVEREF(__pyx_tuple__671);

  /* "pymola/generated/ModelicaParser.pyx":440
 *         buf.write("\u035d\3\2\2\2\u0360\u035f\3\2\2\2\u0361\u0373\3\2\2\2")
 *         buf.write("\u0362\u0363\f\t\2\2\u0363\u0364\t\7\2\2\u0364\u0372\5")
 *         buf.write("r:\n\u0365\u0366\f\b\2\2\u0366\u0367\t\b\2\2\u0367\u0372")             # <<<<<<<<<<<<<<
 *         buf.write("\5r:\t\u0368\u0369\f\7\2\2\u0369\u036a\t\t\2\2\u036a\u0372")
 *         buf.write("\5r:\b\u036b\u036c\f\5\2\2\u036c\u036d\7J\2\2\u036d\u0372")
 */
  __pyx_tuple__672 = PyTuple_Pack(1, __pyx_kp_u_r_3); if (unlikely(!__pyx_tuple__672)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__672);
  __Pyx_GIVEREF(__pyx_tuple__672);

  /* "pymola/generated/ModelicaParser.pyx":441
 *         buf.write("\u0362\u0363\f\t\2\2\u0363\u0364\t\7\2\2\u0364\u0372\5")
 *         buf.write("r:\n\u0365\u0366\f\b\2\2\u0366\u0367\t\b\2\2\u0367\u0372")
 *         buf.write("\5r:\t\u0368\u0369\f\7\2\2\u0369\u036a\t\t\2\2\u036a\u0372")             # <<<<<<<<<<<<<<
 *         buf.write("\5r:\b\u036b\u036c\f\5\2\2\u036c\u036d\7J\2\2\u036d\u0372")
 *         buf.write("\5r:\6\u036e\u036f\f\4\2\2\u036f\u0370\7K\2\2\u0370\u0372")
 */
  __pyx_tuple__673 = PyTuple_Pack(1, __pyx_kp_u_r_4); if (unlikely(!__pyx_tuple__673)) __PYX_ERR(0, 441, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__673);
  __Pyx_GIVEREF(__pyx_tuple__673);

  /* "pymola/generated/ModelicaParser.pyx":442
 *         buf.write("r:\n\u0365\u0366\f\b\2\2\u0366\u0367\t\b\2\2\u0367\u0372")
 *         buf.write("\5r:\t\u0368\u0369\f\7\2\2\u0369\u036a\t\t\2\2\u036a\u0372")
 *         buf.write("\5r:\b\u036b\u036c\f\5\2\2\u036c\u036d\7J\2\2\u036d\u0372")             # <<<<<<<<<<<<<<
 *         buf.write("\5r:\6\u036e\u036f\f\4\2\2\u036f\u0370\7K\2\2\u0370\u0372")
 *         buf.write("\5r:\5\u0371\u0362\3\2\2\2\u0371\u0365\3\2\2\2\u0371\u0368")
 */
  __pyx_tuple__674 = PyTuple_Pack(1, __pyx_kp_u_r_J); if (unlikely(!__pyx_tuple__674)) __PYX_ERR(0, 442, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__674);
  __Pyx_GIVEREF(__pyx_tuple__674);

  /* "pymola/generated/ModelicaParser.pyx":443
 *         buf.write("\5r:\t\u0368\u0369\f\7\2\2\u0369\u036a\t\t\2\2\u036a\u0372")
 *         buf.write("\5r:\b\u036b\u036c\f\5\2\2\u036c\u036d\7J\2\2\u036d\u0372")
 *         buf.write("\5r:\6\u036e\u036f\f\4\2\2\u036f\u0370\7K\2\2\u0370\u0372")             # <<<<<<<<<<<<<<
 *         buf.write("\5r:\5\u0371\u0362\3\2\2\2\u0371\u0365\3\2\2\2\u0371\u0368")
 *         buf.write("\3\2\2\2\u0371\u036b\3\2\2\2\u0371\u036e\3\2\2\2\u0372")
 */
  __pyx_tuple__675 = PyTuple_Pack(1, __pyx_kp_u_r_K); if (unlikely(!__pyx_tuple__675)) __PYX_ERR(0, 443, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__675);
  __Pyx_GIVEREF(__pyx_tuple__675);

  /* "pymola/generated/ModelicaParser.pyx":444
 *         buf.write("\5r:\b\u036b\u036c\f\5\2\2\u036c\u036d\7J\2\2\u036d\u0372")
 *         buf.write("\5r:\6\u036e\u036f\f\4\2\2\u036f\u0370\7K\2\2\u0370\u0372")
 *         buf.write("\5r:\5\u0371\u0362\3\2\2\2\u0371\u0365\3\2\2\2\u0371\u0368")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0371\u036b\3\2\2\2\u0371\u036e\3\2\2\2\u0372")
 *         buf.write("\u0375\3\2\2\2\u0373\u0371\3\2\2\2\u0373\u0374\3\2\2\2")
 */
  __pyx_tuple__676 = PyTuple_Pack(1, __pyx_kp_u_r_5); if (unlikely(!__pyx_tuple__676)) __PYX_ERR(0, 444, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__676);
  __Pyx_GIVEREF(__pyx_tuple__676);

  /* "pymola/generated/ModelicaParser.pyx":445
 *         buf.write("\5r:\6\u036e\u036f\f\4\2\2\u036f\u0370\7K\2\2\u0370\u0372")
 *         buf.write("\5r:\5\u0371\u0362\3\2\2\2\u0371\u0365\3\2\2\2\u0371\u0368")
 *         buf.write("\3\2\2\2\u0371\u036b\3\2\2\2\u0371\u036e\3\2\2\2\u0372")             # <<<<<<<<<<<<<<
 *         buf.write("\u0375\3\2\2\2\u0373\u0371\3\2\2\2\u0373\u0374\3\2\2\2")
 *         buf.write("\u0374s\3\2\2\2\u0375\u0373\3\2\2\2\u0376\u0397\7\\\2")
 */
  __pyx_tuple__678 = PyTuple_Pack(1, __pyx_kp_u__677); if (unlikely(!__pyx_tuple__678)) __PYX_ERR(0, 445, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__678);
  __Pyx_GIVEREF(__pyx_tuple__678);

  /* "pymola/generated/ModelicaParser.pyx":446
 *         buf.write("\5r:\5\u0371\u0362\3\2\2\2\u0371\u0365\3\2\2\2\u0371\u0368")
 *         buf.write("\3\2\2\2\u0371\u036b\3\2\2\2\u0371\u036e\3\2\2\2\u0372")
 *         buf.write("\u0375\3\2\2\2\u0373\u0371\3\2\2\2\u0373\u0374\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0374s\3\2\2\2\u0375\u0373\3\2\2\2\u0376\u0397\7\\\2")
 *         buf.write("\2\u0377\u0397\7[\2\2\u0378\u0397\7L\2\2\u0379\u0397\7")
 */
  __pyx_tuple__680 = PyTuple_Pack(1, __pyx_kp_u__679); if (unlikely(!__pyx_tuple__680)) __PYX_ERR(0, 446, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__680);
  __Pyx_GIVEREF(__pyx_tuple__680);

  /* "pymola/generated/ModelicaParser.pyx":447
 *         buf.write("\3\2\2\2\u0371\u036b\3\2\2\2\u0371\u036e\3\2\2\2\u0372")
 *         buf.write("\u0375\3\2\2\2\u0373\u0371\3\2\2\2\u0373\u0374\3\2\2\2")
 *         buf.write("\u0374s\3\2\2\2\u0375\u0373\3\2\2\2\u0376\u0397\7\\\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u0377\u0397\7[\2\2\u0378\u0397\7L\2\2\u0379\u0397\7")
 *         buf.write("M\2\2\u037a\u037b\5z>\2\u037b\u037c\5|?\2\u037c\u0397")
 */
  __pyx_tuple__681 = PyTuple_Pack(1, __pyx_kp_u_s); if (unlikely(!__pyx_tuple__681)) __PYX_ERR(0, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__681);
  __Pyx_GIVEREF(__pyx_tuple__681);

  /* "pymola/generated/ModelicaParser.pyx":448
 *         buf.write("\u0375\3\2\2\2\u0373\u0371\3\2\2\2\u0373\u0374\3\2\2\2")
 *         buf.write("\u0374s\3\2\2\2\u0375\u0373\3\2\2\2\u0376\u0397\7\\\2")
 *         buf.write("\2\u0377\u0397\7[\2\2\u0378\u0397\7L\2\2\u0379\u0397\7")             # <<<<<<<<<<<<<<
 *         buf.write("M\2\2\u037a\u037b\5z>\2\u037b\u037c\5|?\2\u037c\u0397")
 *         buf.write("\3\2\2\2\u037d\u037e\7\26\2\2\u037e\u0397\5|?\2\u037f")
 */
  __pyx_tuple__682 = PyTuple_Pack(1, __pyx_kp_u_L_2); if (unlikely(!__pyx_tuple__682)) __PYX_ERR(0, 448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__682);
  __Pyx_GIVEREF(__pyx_tuple__682);

  /* "pymola/generated/ModelicaParser.pyx":449
 *         buf.write("\u0374s\3\2\2\2\u0375\u0373\3\2\2\2\u0376\u0397\7\\\2")
 *         buf.write("\2\u0377\u0397\7[\2\2\u0378\u0397\7L\2\2\u0379\u0397\7")
 *         buf.write("M\2\2\u037a\u037b\5z>\2\u037b\u037c\5|?\2\u037c\u0397")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u037d\u037e\7\26\2\2\u037e\u0397\5|?\2\u037f")
 *         buf.write("\u0380\7Y\2\2\u0380\u0397\5|?\2\u0381\u0397\5z>\2\u0382")
 */
  __pyx_tuple__683 = PyTuple_Pack(1, __pyx_kp_u_M_z); if (unlikely(!__pyx_tuple__683)) __PYX_ERR(0, 449, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__683);
  __Pyx_GIVEREF(__pyx_tuple__683);

  /* "pymola/generated/ModelicaParser.pyx":450
 *         buf.write("\2\u0377\u0397\7[\2\2\u0378\u0397\7L\2\2\u0379\u0397\7")
 *         buf.write("M\2\2\u037a\u037b\5z>\2\u037b\u037c\5|?\2\u037c\u0397")
 *         buf.write("\3\2\2\2\u037d\u037e\7\26\2\2\u037e\u0397\5|?\2\u037f")             # <<<<<<<<<<<<<<
 *         buf.write("\u0380\7Y\2\2\u0380\u0397\5|?\2\u0381\u0397\5z>\2\u0382")
 *         buf.write("\u0383\7\23\2\2\u0383\u0384\5\u0086D\2\u0384\u0385\7\25")
 */
  __pyx_tuple__685 = PyTuple_Pack(1, __pyx_kp_u__684); if (unlikely(!__pyx_tuple__685)) __PYX_ERR(0, 450, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__685);
  __Pyx_GIVEREF(__pyx_tuple__685);

  /* "pymola/generated/ModelicaParser.pyx":451
 *         buf.write("M\2\2\u037a\u037b\5z>\2\u037b\u037c\5|?\2\u037c\u0397")
 *         buf.write("\3\2\2\2\u037d\u037e\7\26\2\2\u037e\u0397\5|?\2\u037f")
 *         buf.write("\u0380\7Y\2\2\u0380\u0397\5|?\2\u0381\u0397\5z>\2\u0382")             # <<<<<<<<<<<<<<
 *         buf.write("\u0383\7\23\2\2\u0383\u0384\5\u0086D\2\u0384\u0385\7\25")
 *         buf.write("\2\2\u0385\u0397\3\2\2\2\u0386\u0387\7N\2\2\u0387\u038c")
 */
  __pyx_tuple__686 = PyTuple_Pack(1, __pyx_kp_u_Y_z); if (unlikely(!__pyx_tuple__686)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__686);
  __Pyx_GIVEREF(__pyx_tuple__686);

  /* "pymola/generated/ModelicaParser.pyx":452
 *         buf.write("\3\2\2\2\u037d\u037e\7\26\2\2\u037e\u0397\5|?\2\u037f")
 *         buf.write("\u0380\7Y\2\2\u0380\u0397\5|?\2\u0381\u0397\5z>\2\u0382")
 *         buf.write("\u0383\7\23\2\2\u0383\u0384\5\u0086D\2\u0384\u0385\7\25")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\u0385\u0397\3\2\2\2\u0386\u0387\7N\2\2\u0387\u038c")
 *         buf.write("\5\u0088E\2\u0388\u0389\7\3\2\2\u0389\u038b\5\u0088E\2")
 */
  __pyx_tuple__687 = PyTuple_Pack(1, __pyx_kp_u_D_2); if (unlikely(!__pyx_tuple__687)) __PYX_ERR(0, 452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__687);
  __Pyx_GIVEREF(__pyx_tuple__687);

  /* "pymola/generated/ModelicaParser.pyx":453
 *         buf.write("\u0380\7Y\2\2\u0380\u0397\5|?\2\u0381\u0397\5z>\2\u0382")
 *         buf.write("\u0383\7\23\2\2\u0383\u0384\5\u0086D\2\u0384\u0385\7\25")
 *         buf.write("\2\2\u0385\u0397\3\2\2\2\u0386\u0387\7N\2\2\u0387\u038c")             # <<<<<<<<<<<<<<
 *         buf.write("\5\u0088E\2\u0388\u0389\7\3\2\2\u0389\u038b\5\u0088E\2")
 *         buf.write("\u038a\u0388\3\2\2\2\u038b\u038e\3\2\2\2\u038c\u038a\3")
 */
  __pyx_tuple__688 = PyTuple_Pack(1, __pyx_kp_u_N_2); if (unlikely(!__pyx_tuple__688)) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__688);
  __Pyx_GIVEREF(__pyx_tuple__688);

  /* "pymola/generated/ModelicaParser.pyx":454
 *         buf.write("\u0383\7\23\2\2\u0383\u0384\5\u0086D\2\u0384\u0385\7\25")
 *         buf.write("\2\2\u0385\u0397\3\2\2\2\u0386\u0387\7N\2\2\u0387\u038c")
 *         buf.write("\5\u0088E\2\u0388\u0389\7\3\2\2\u0389\u038b\5\u0088E\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u038a\u0388\3\2\2\2\u038b\u038e\3\2\2\2\u038c\u038a\3")
 *         buf.write("\2\2\2\u038c\u038d\3\2\2\2\u038d\u038f\3\2\2\2\u038e\u038c")
 */
  __pyx_tuple__689 = PyTuple_Pack(1, __pyx_kp_u_E_E); if (unlikely(!__pyx_tuple__689)) __PYX_ERR(0, 454, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__689);
  __Pyx_GIVEREF(__pyx_tuple__689);

  /* "pymola/generated/ModelicaParser.pyx":455
 *         buf.write("\2\2\u0385\u0397\3\2\2\2\u0386\u0387\7N\2\2\u0387\u038c")
 *         buf.write("\5\u0088E\2\u0388\u0389\7\3\2\2\u0389\u038b\5\u0088E\2")
 *         buf.write("\u038a\u0388\3\2\2\2\u038b\u038e\3\2\2\2\u038c\u038a\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u038c\u038d\3\2\2\2\u038d\u038f\3\2\2\2\u038e\u038c")
 *         buf.write("\3\2\2\2\u038f\u0390\7O\2\2\u0390\u0397\3\2\2\2\u0391")
 */
  __pyx_tuple__691 = PyTuple_Pack(1, __pyx_kp_u__690); if (unlikely(!__pyx_tuple__691)) __PYX_ERR(0, 455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__691);
  __Pyx_GIVEREF(__pyx_tuple__691);

  /* "pymola/generated/ModelicaParser.pyx":456
 *         buf.write("\5\u0088E\2\u0388\u0389\7\3\2\2\u0389\u038b\5\u0088E\2")
 *         buf.write("\u038a\u0388\3\2\2\2\u038b\u038e\3\2\2\2\u038c\u038a\3")
 *         buf.write("\2\2\2\u038c\u038d\3\2\2\2\u038d\u038f\3\2\2\2\u038e\u038c")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u038f\u0390\7O\2\2\u0390\u0397\3\2\2\2\u0391")
 *         buf.write("\u0392\7 \2\2\u0392\u0393\5~@\2\u0393\u0394\7!\2\2\u0394")
 */
  __pyx_tuple__693 = PyTuple_Pack(1, __pyx_kp_u__692); if (unlikely(!__pyx_tuple__693)) __PYX_ERR(0, 456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__693);
  __Pyx_GIVEREF(__pyx_tuple__693);

  /* "pymola/generated/ModelicaParser.pyx":457
 *         buf.write("\u038a\u0388\3\2\2\2\u038b\u038e\3\2\2\2\u038c\u038a\3")
 *         buf.write("\2\2\2\u038c\u038d\3\2\2\2\u038d\u038f\3\2\2\2\u038e\u038c")
 *         buf.write("\3\2\2\2\u038f\u0390\7O\2\2\u0390\u0397\3\2\2\2\u0391")             # <<<<<<<<<<<<<<
 *         buf.write("\u0392\7 \2\2\u0392\u0393\5~@\2\u0393\u0394\7!\2\2\u0394")
 *         buf.write("\u0397\3\2\2\2\u0395\u0397\7\20\2\2\u0396\u0376\3\2\2")
 */
  __pyx_tuple__694 = PyTuple_Pack(1, __pyx_kp_u_O_2); if (unlikely(!__pyx_tuple__694)) __PYX_ERR(0, 457, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__694);
  __Pyx_GIVEREF(__pyx_tuple__694);

  /* "pymola/generated/ModelicaParser.pyx":458
 *         buf.write("\2\2\2\u038c\u038d\3\2\2\2\u038d\u038f\3\2\2\2\u038e\u038c")
 *         buf.write("\3\2\2\2\u038f\u0390\7O\2\2\u0390\u0397\3\2\2\2\u0391")
 *         buf.write("\u0392\7 \2\2\u0392\u0393\5~@\2\u0393\u0394\7!\2\2\u0394")             # <<<<<<<<<<<<<<
 *         buf.write("\u0397\3\2\2\2\u0395\u0397\7\20\2\2\u0396\u0376\3\2\2")
 *         buf.write("\2\u0396\u0377\3\2\2\2\u0396\u0378\3\2\2\2\u0396\u0379")
 */
  __pyx_tuple__696 = PyTuple_Pack(1, __pyx_kp_u__695); if (unlikely(!__pyx_tuple__696)) __PYX_ERR(0, 458, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__696);
  __Pyx_GIVEREF(__pyx_tuple__696);

  /* "pymola/generated/ModelicaParser.pyx":459
 *         buf.write("\3\2\2\2\u038f\u0390\7O\2\2\u0390\u0397\3\2\2\2\u0391")
 *         buf.write("\u0392\7 \2\2\u0392\u0393\5~@\2\u0393\u0394\7!\2\2\u0394")
 *         buf.write("\u0397\3\2\2\2\u0395\u0397\7\20\2\2\u0396\u0376\3\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u0396\u0377\3\2\2\2\u0396\u0378\3\2\2\2\u0396\u0379")
 *         buf.write("\3\2\2\2\u0396\u037a\3\2\2\2\u0396\u037d\3\2\2\2\u0396")
 */
  __pyx_tuple__698 = PyTuple_Pack(1, __pyx_kp_u__697); if (unlikely(!__pyx_tuple__698)) __PYX_ERR(0, 459, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__698);
  __Pyx_GIVEREF(__pyx_tuple__698);

  /* "pymola/generated/ModelicaParser.pyx":460
 *         buf.write("\u0392\7 \2\2\u0392\u0393\5~@\2\u0393\u0394\7!\2\2\u0394")
 *         buf.write("\u0397\3\2\2\2\u0395\u0397\7\20\2\2\u0396\u0376\3\2\2")
 *         buf.write("\2\u0396\u0377\3\2\2\2\u0396\u0378\3\2\2\2\u0396\u0379")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0396\u037a\3\2\2\2\u0396\u037d\3\2\2\2\u0396")
 *         buf.write("\u037f\3\2\2\2\u0396\u0381\3\2\2\2\u0396\u0382\3\2\2\2")
 */
  __pyx_tuple__700 = PyTuple_Pack(1, __pyx_kp_u__699); if (unlikely(!__pyx_tuple__700)) __PYX_ERR(0, 460, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__700);
  __Pyx_GIVEREF(__pyx_tuple__700);

  /* "pymola/generated/ModelicaParser.pyx":461
 *         buf.write("\u0397\3\2\2\2\u0395\u0397\7\20\2\2\u0396\u0376\3\2\2")
 *         buf.write("\2\u0396\u0377\3\2\2\2\u0396\u0378\3\2\2\2\u0396\u0379")
 *         buf.write("\3\2\2\2\u0396\u037a\3\2\2\2\u0396\u037d\3\2\2\2\u0396")             # <<<<<<<<<<<<<<
 *         buf.write("\u037f\3\2\2\2\u0396\u0381\3\2\2\2\u0396\u0382\3\2\2\2")
 *         buf.write("\u0396\u0386\3\2\2\2\u0396\u0391\3\2\2\2\u0396\u0395\3")
 */
  __pyx_tuple__702 = PyTuple_Pack(1, __pyx_kp_u__701); if (unlikely(!__pyx_tuple__702)) __PYX_ERR(0, 461, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__702);
  __Pyx_GIVEREF(__pyx_tuple__702);

  /* "pymola/generated/ModelicaParser.pyx":462
 *         buf.write("\2\u0396\u0377\3\2\2\2\u0396\u0378\3\2\2\2\u0396\u0379")
 *         buf.write("\3\2\2\2\u0396\u037a\3\2\2\2\u0396\u037d\3\2\2\2\u0396")
 *         buf.write("\u037f\3\2\2\2\u0396\u0381\3\2\2\2\u0396\u0382\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0396\u0386\3\2\2\2\u0396\u0391\3\2\2\2\u0396\u0395\3")
 *         buf.write("\2\2\2\u0397u\3\2\2\2\u0398\u039a\7\36\2\2\u0399\u0398")
 */
  __pyx_tuple__704 = PyTuple_Pack(1, __pyx_kp_u__703); if (unlikely(!__pyx_tuple__704)) __PYX_ERR(0, 462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__704);
  __Pyx_GIVEREF(__pyx_tuple__704);

  /* "pymola/generated/ModelicaParser.pyx":463
 *         buf.write("\3\2\2\2\u0396\u037a\3\2\2\2\u0396\u037d\3\2\2\2\u0396")
 *         buf.write("\u037f\3\2\2\2\u0396\u0381\3\2\2\2\u0396\u0382\3\2\2\2")
 *         buf.write("\u0396\u0386\3\2\2\2\u0396\u0391\3\2\2\2\u0396\u0395\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u0397u\3\2\2\2\u0398\u039a\7\36\2\2\u0399\u0398")
 *         buf.write("\3\2\2\2\u0399\u039a\3\2\2\2\u039a\u039b\3\2\2\2\u039b")
 */
  __pyx_tuple__706 = PyTuple_Pack(1, __pyx_kp_u__705); if (unlikely(!__pyx_tuple__706)) __PYX_ERR(0, 463, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__706);
  __Pyx_GIVEREF(__pyx_tuple__706);

  /* "pymola/generated/ModelicaParser.pyx":464
 *         buf.write("\u037f\3\2\2\2\u0396\u0381\3\2\2\2\u0396\u0382\3\2\2\2")
 *         buf.write("\u0396\u0386\3\2\2\2\u0396\u0391\3\2\2\2\u0396\u0395\3")
 *         buf.write("\2\2\2\u0397u\3\2\2\2\u0398\u039a\7\36\2\2\u0399\u0398")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0399\u039a\3\2\2\2\u039a\u039b\3\2\2\2\u039b")
 *         buf.write("\u03a0\7Z\2\2\u039c\u039d\7\36\2\2\u039d\u039f\7Z\2\2")
 */
  __pyx_tuple__707 = PyTuple_Pack(1, __pyx_kp_u_u); if (unlikely(!__pyx_tuple__707)) __PYX_ERR(0, 464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__707);
  __Pyx_GIVEREF(__pyx_tuple__707);

  /* "pymola/generated/ModelicaParser.pyx":465
 *         buf.write("\u0396\u0386\3\2\2\2\u0396\u0391\3\2\2\2\u0396\u0395\3")
 *         buf.write("\2\2\2\u0397u\3\2\2\2\u0398\u039a\7\36\2\2\u0399\u0398")
 *         buf.write("\3\2\2\2\u0399\u039a\3\2\2\2\u039a\u039b\3\2\2\2\u039b")             # <<<<<<<<<<<<<<
 *         buf.write("\u03a0\7Z\2\2\u039c\u039d\7\36\2\2\u039d\u039f\7Z\2\2")
 *         buf.write("\u039e\u039c\3\2\2\2\u039f\u03a2\3\2\2\2\u03a0\u039e\3")
 */
  __pyx_tuple__709 = PyTuple_Pack(1, __pyx_kp_u__708); if (unlikely(!__pyx_tuple__709)) __PYX_ERR(0, 465, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__709);
  __Pyx_GIVEREF(__pyx_tuple__709);

  /* "pymola/generated/ModelicaParser.pyx":466
 *         buf.write("\2\2\2\u0397u\3\2\2\2\u0398\u039a\7\36\2\2\u0399\u0398")
 *         buf.write("\3\2\2\2\u0399\u039a\3\2\2\2\u039a\u039b\3\2\2\2\u039b")
 *         buf.write("\u03a0\7Z\2\2\u039c\u039d\7\36\2\2\u039d\u039f\7Z\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u039e\u039c\3\2\2\2\u039f\u03a2\3\2\2\2\u03a0\u039e\3")
 *         buf.write("\2\2\2\u03a0\u03a1\3\2\2\2\u03a1w\3\2\2\2\u03a2\u03a0")
 */
  __pyx_tuple__710 = PyTuple_Pack(1, __pyx_kp_u_Z_Z_2); if (unlikely(!__pyx_tuple__710)) __PYX_ERR(0, 466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__710);
  __Pyx_GIVEREF(__pyx_tuple__710);

  /* "pymola/generated/ModelicaParser.pyx":467
 *         buf.write("\3\2\2\2\u0399\u039a\3\2\2\2\u039a\u039b\3\2\2\2\u039b")
 *         buf.write("\u03a0\7Z\2\2\u039c\u039d\7\36\2\2\u039d\u039f\7Z\2\2")
 *         buf.write("\u039e\u039c\3\2\2\2\u039f\u03a2\3\2\2\2\u03a0\u039e\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u03a0\u03a1\3\2\2\2\u03a1w\3\2\2\2\u03a2\u03a0")
 *         buf.write("\3\2\2\2\u03a3\u03a5\7Z\2\2\u03a4\u03a6\5\u008aF\2\u03a5")
 */
  __pyx_tuple__712 = PyTuple_Pack(1, __pyx_kp_u__711); if (unlikely(!__pyx_tuple__712)) __PYX_ERR(0, 467, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__712);
  __Pyx_GIVEREF(__pyx_tuple__712);

  /* "pymola/generated/ModelicaParser.pyx":468
 *         buf.write("\u03a0\7Z\2\2\u039c\u039d\7\36\2\2\u039d\u039f\7Z\2\2")
 *         buf.write("\u039e\u039c\3\2\2\2\u039f\u03a2\3\2\2\2\u03a0\u039e\3")
 *         buf.write("\2\2\2\u03a0\u03a1\3\2\2\2\u03a1w\3\2\2\2\u03a2\u03a0")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u03a3\u03a5\7Z\2\2\u03a4\u03a6\5\u008aF\2\u03a5")
 *         buf.write("\u03a4\3\2\2\2\u03a5\u03a6\3\2\2\2\u03a6y\3\2\2\2\u03a7")
 */
  __pyx_tuple__713 = PyTuple_Pack(1, __pyx_kp_u_w); if (unlikely(!__pyx_tuple__713)) __PYX_ERR(0, 468, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__713);
  __Pyx_GIVEREF(__pyx_tuple__713);

  /* "pymola/generated/ModelicaParser.pyx":469
 *         buf.write("\u039e\u039c\3\2\2\2\u039f\u03a2\3\2\2\2\u03a0\u039e\3")
 *         buf.write("\2\2\2\u03a0\u03a1\3\2\2\2\u03a1w\3\2\2\2\u03a2\u03a0")
 *         buf.write("\3\2\2\2\u03a3\u03a5\7Z\2\2\u03a4\u03a6\5\u008aF\2\u03a5")             # <<<<<<<<<<<<<<
 *         buf.write("\u03a4\3\2\2\2\u03a5\u03a6\3\2\2\2\u03a6y\3\2\2\2\u03a7")
 *         buf.write("\u03ac\5x=\2\u03a8\u03a9\7\36\2\2\u03a9\u03ab\5x=\2\u03aa")
 */
  __pyx_tuple__714 = PyTuple_Pack(1, __pyx_kp_u_Z_F_2); if (unlikely(!__pyx_tuple__714)) __PYX_ERR(0, 469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__714);
  __Pyx_GIVEREF(__pyx_tuple__714);

  /* "pymola/generated/ModelicaParser.pyx":470
 *         buf.write("\2\2\2\u03a0\u03a1\3\2\2\2\u03a1w\3\2\2\2\u03a2\u03a0")
 *         buf.write("\3\2\2\2\u03a3\u03a5\7Z\2\2\u03a4\u03a6\5\u008aF\2\u03a5")
 *         buf.write("\u03a4\3\2\2\2\u03a5\u03a6\3\2\2\2\u03a6y\3\2\2\2\u03a7")             # <<<<<<<<<<<<<<
 *         buf.write("\u03ac\5x=\2\u03a8\u03a9\7\36\2\2\u03a9\u03ab\5x=\2\u03aa")
 *         buf.write("\u03a8\3\2\2\2\u03ab\u03ae\3\2\2\2\u03ac\u03aa\3\2\2\2")
 */
  __pyx_tuple__715 = PyTuple_Pack(1, __pyx_kp_u_y); if (unlikely(!__pyx_tuple__715)) __PYX_ERR(0, 470, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__715);
  __Pyx_GIVEREF(__pyx_tuple__715);

  /* "pymola/generated/ModelicaParser.pyx":471
 *         buf.write("\3\2\2\2\u03a3\u03a5\7Z\2\2\u03a4\u03a6\5\u008aF\2\u03a5")
 *         buf.write("\u03a4\3\2\2\2\u03a5\u03a6\3\2\2\2\u03a6y\3\2\2\2\u03a7")
 *         buf.write("\u03ac\5x=\2\u03a8\u03a9\7\36\2\2\u03a9\u03ab\5x=\2\u03aa")             # <<<<<<<<<<<<<<
 *         buf.write("\u03a8\3\2\2\2\u03ab\u03ae\3\2\2\2\u03ac\u03aa\3\2\2\2")
 *         buf.write("\u03ac\u03ad\3\2\2\2\u03ad{\3\2\2\2\u03ae\u03ac\3\2\2")
 */
  __pyx_tuple__716 = PyTuple_Pack(1, __pyx_kp_u_x_x); if (unlikely(!__pyx_tuple__716)) __PYX_ERR(0, 471, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__716);
  __Pyx_GIVEREF(__pyx_tuple__716);

  /* "pymola/generated/ModelicaParser.pyx":472
 *         buf.write("\u03a4\3\2\2\2\u03a5\u03a6\3\2\2\2\u03a6y\3\2\2\2\u03a7")
 *         buf.write("\u03ac\5x=\2\u03a8\u03a9\7\36\2\2\u03a9\u03ab\5x=\2\u03aa")
 *         buf.write("\u03a8\3\2\2\2\u03ab\u03ae\3\2\2\2\u03ac\u03aa\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u03ac\u03ad\3\2\2\2\u03ad{\3\2\2\2\u03ae\u03ac\3\2\2")
 *         buf.write("\2\u03af\u03b1\7\23\2\2\u03b0\u03b2\5~@\2\u03b1\u03b0")
 */
  __pyx_tuple__718 = PyTuple_Pack(1, __pyx_kp_u__717); if (unlikely(!__pyx_tuple__718)) __PYX_ERR(0, 472, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__718);
  __Pyx_GIVEREF(__pyx_tuple__718);

  /* "pymola/generated/ModelicaParser.pyx":473
 *         buf.write("\u03ac\5x=\2\u03a8\u03a9\7\36\2\2\u03a9\u03ab\5x=\2\u03aa")
 *         buf.write("\u03a8\3\2\2\2\u03ab\u03ae\3\2\2\2\u03ac\u03aa\3\2\2\2")
 *         buf.write("\u03ac\u03ad\3\2\2\2\u03ad{\3\2\2\2\u03ae\u03ac\3\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u03af\u03b1\7\23\2\2\u03b0\u03b2\5~@\2\u03b1\u03b0")
 *         buf.write("\3\2\2\2\u03b1\u03b2\3\2\2\2\u03b2\u03b3\3\2\2\2\u03b3")
 */
  __pyx_tuple__720 = PyTuple_Pack(1, __pyx_kp_u__719); if (unlikely(!__pyx_tuple__720)) __PYX_ERR(0, 473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__720);
  __Pyx_GIVEREF(__pyx_tuple__720);

  /* "pymola/generated/ModelicaParser.pyx":474
 *         buf.write("\u03a8\3\2\2\2\u03ab\u03ae\3\2\2\2\u03ac\u03aa\3\2\2\2")
 *         buf.write("\u03ac\u03ad\3\2\2\2\u03ad{\3\2\2\2\u03ae\u03ac\3\2\2")
 *         buf.write("\2\u03af\u03b1\7\23\2\2\u03b0\u03b2\5~@\2\u03b1\u03b0")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u03b1\u03b2\3\2\2\2\u03b2\u03b3\3\2\2\2\u03b3")
 *         buf.write("\u03b4\7\25\2\2\u03b4}\3\2\2\2\u03b5\u03bc\5\u0084C\2")
 */
  __pyx_tuple__722 = PyTuple_Pack(1, __pyx_kp_u__721); if (unlikely(!__pyx_tuple__722)) __PYX_ERR(0, 474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__722);
  __Pyx_GIVEREF(__pyx_tuple__722);

  /* "pymola/generated/ModelicaParser.pyx":475
 *         buf.write("\u03ac\u03ad\3\2\2\2\u03ad{\3\2\2\2\u03ae\u03ac\3\2\2")
 *         buf.write("\2\u03af\u03b1\7\23\2\2\u03b0\u03b2\5~@\2\u03b1\u03b0")
 *         buf.write("\3\2\2\2\u03b1\u03b2\3\2\2\2\u03b2\u03b3\3\2\2\2\u03b3")             # <<<<<<<<<<<<<<
 *         buf.write("\u03b4\7\25\2\2\u03b4}\3\2\2\2\u03b5\u03bc\5\u0084C\2")
 *         buf.write("\u03b6\u03b7\7\27\2\2\u03b7\u03bb\5\u0084C\2\u03b8\u03b9")
 */
  __pyx_tuple__724 = PyTuple_Pack(1, __pyx_kp_u__723); if (unlikely(!__pyx_tuple__724)) __PYX_ERR(0, 475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__724);
  __Pyx_GIVEREF(__pyx_tuple__724);

  /* "pymola/generated/ModelicaParser.pyx":476
 *         buf.write("\2\u03af\u03b1\7\23\2\2\u03b0\u03b2\5~@\2\u03b1\u03b0")
 *         buf.write("\3\2\2\2\u03b1\u03b2\3\2\2\2\u03b2\u03b3\3\2\2\2\u03b3")
 *         buf.write("\u03b4\7\25\2\2\u03b4}\3\2\2\2\u03b5\u03bc\5\u0084C\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u03b6\u03b7\7\27\2\2\u03b7\u03bb\5\u0084C\2\u03b8\u03b9")
 *         buf.write("\7\63\2\2\u03b9\u03bb\5b\62\2\u03ba\u03b6\3\2\2\2\u03ba")
 */
  __pyx_tuple__725 = PyTuple_Pack(1, __pyx_kp_u_C); if (unlikely(!__pyx_tuple__725)) __PYX_ERR(0, 476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__725);
  __Pyx_GIVEREF(__pyx_tuple__725);

  /* "pymola/generated/ModelicaParser.pyx":477
 *         buf.write("\3\2\2\2\u03b1\u03b2\3\2\2\2\u03b2\u03b3\3\2\2\2\u03b3")
 *         buf.write("\u03b4\7\25\2\2\u03b4}\3\2\2\2\u03b5\u03bc\5\u0084C\2")
 *         buf.write("\u03b6\u03b7\7\27\2\2\u03b7\u03bb\5\u0084C\2\u03b8\u03b9")             # <<<<<<<<<<<<<<
 *         buf.write("\7\63\2\2\u03b9\u03bb\5b\62\2\u03ba\u03b6\3\2\2\2\u03ba")
 *         buf.write("\u03b8\3\2\2\2\u03bb\u03be\3\2\2\2\u03bc\u03ba\3\2\2\2")
 */
  __pyx_tuple__726 = PyTuple_Pack(1, __pyx_kp_u_C_2); if (unlikely(!__pyx_tuple__726)) __PYX_ERR(0, 477, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__726);
  __Pyx_GIVEREF(__pyx_tuple__726);

  /* "pymola/generated/ModelicaParser.pyx":478
 *         buf.write("\u03b4\7\25\2\2\u03b4}\3\2\2\2\u03b5\u03bc\5\u0084C\2")
 *         buf.write("\u03b6\u03b7\7\27\2\2\u03b7\u03bb\5\u0084C\2\u03b8\u03b9")
 *         buf.write("\7\63\2\2\u03b9\u03bb\5b\62\2\u03ba\u03b6\3\2\2\2\u03ba")             # <<<<<<<<<<<<<<
 *         buf.write("\u03b8\3\2\2\2\u03bb\u03be\3\2\2\2\u03bc\u03ba\3\2\2\2")
 *         buf.write("\u03bc\u03bd\3\2\2\2\u03bd\u03c1\3\2\2\2\u03be\u03bc\3")
 */
  __pyx_tuple__727 = PyTuple_Pack(1, __pyx_kp_u_3_b2_2); if (unlikely(!__pyx_tuple__727)) __PYX_ERR(0, 478, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__727);
  __Pyx_GIVEREF(__pyx_tuple__727);

  /* "pymola/generated/ModelicaParser.pyx":479
 *         buf.write("\u03b6\u03b7\7\27\2\2\u03b7\u03bb\5\u0084C\2\u03b8\u03b9")
 *         buf.write("\7\63\2\2\u03b9\u03bb\5b\62\2\u03ba\u03b6\3\2\2\2\u03ba")
 *         buf.write("\u03b8\3\2\2\2\u03bb\u03be\3\2\2\2\u03bc\u03ba\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u03bc\u03bd\3\2\2\2\u03bd\u03c1\3\2\2\2\u03be\u03bc\3")
 *         buf.write("\2\2\2\u03bf\u03c1\5\u0080A\2\u03c0\u03b5\3\2\2\2\u03c0")
 */
  __pyx_tuple__729 = PyTuple_Pack(1, __pyx_kp_u__728); if (unlikely(!__pyx_tuple__729)) __PYX_ERR(0, 479, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__729);
  __Pyx_GIVEREF(__pyx_tuple__729);

  /* "pymola/generated/ModelicaParser.pyx":480
 *         buf.write("\7\63\2\2\u03b9\u03bb\5b\62\2\u03ba\u03b6\3\2\2\2\u03ba")
 *         buf.write("\u03b8\3\2\2\2\u03bb\u03be\3\2\2\2\u03bc\u03ba\3\2\2\2")
 *         buf.write("\u03bc\u03bd\3\2\2\2\u03bd\u03c1\3\2\2\2\u03be\u03bc\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u03bf\u03c1\5\u0080A\2\u03c0\u03b5\3\2\2\2\u03c0")
 *         buf.write("\u03bf\3\2\2\2\u03c1\177\3\2\2\2\u03c2\u03c7\5\u0082B")
 */
  __pyx_tuple__731 = PyTuple_Pack(1, __pyx_kp_u__730); if (unlikely(!__pyx_tuple__731)) __PYX_ERR(0, 480, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__731);
  __Pyx_GIVEREF(__pyx_tuple__731);

  /* "pymola/generated/ModelicaParser.pyx":481
 *         buf.write("\u03b8\3\2\2\2\u03bb\u03be\3\2\2\2\u03bc\u03ba\3\2\2\2")
 *         buf.write("\u03bc\u03bd\3\2\2\2\u03bd\u03c1\3\2\2\2\u03be\u03bc\3")
 *         buf.write("\2\2\2\u03bf\u03c1\5\u0080A\2\u03c0\u03b5\3\2\2\2\u03c0")             # <<<<<<<<<<<<<<
 *         buf.write("\u03bf\3\2\2\2\u03c1\177\3\2\2\2\u03c2\u03c7\5\u0082B")
 *         buf.write("\2\u03c3\u03c4\7\27\2\2\u03c4\u03c6\5\u0082B\2\u03c5\u03c3")
 */
  __pyx_tuple__732 = PyTuple_Pack(1, __pyx_kp_u_A_3); if (unlikely(!__pyx_tuple__732)) __PYX_ERR(0, 481, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__732);
  __Pyx_GIVEREF(__pyx_tuple__732);

  /* "pymola/generated/ModelicaParser.pyx":482
 *         buf.write("\u03bc\u03bd\3\2\2\2\u03bd\u03c1\3\2\2\2\u03be\u03bc\3")
 *         buf.write("\2\2\2\u03bf\u03c1\5\u0080A\2\u03c0\u03b5\3\2\2\2\u03c0")
 *         buf.write("\u03bf\3\2\2\2\u03c1\177\3\2\2\2\u03c2\u03c7\5\u0082B")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u03c3\u03c4\7\27\2\2\u03c4\u03c6\5\u0082B\2\u03c5\u03c3")
 *         buf.write("\3\2\2\2\u03c6\u03c9\3\2\2\2\u03c7\u03c5\3\2\2\2\u03c7")
 */
  __pyx_tuple__733 = PyTuple_Pack(1, __pyx_kp_u_B_2); if (unlikely(!__pyx_tuple__733)) __PYX_ERR(0, 482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__733);
  __Pyx_GIVEREF(__pyx_tuple__733);

  /* "pymola/generated/ModelicaParser.pyx":483
 *         buf.write("\2\2\2\u03bf\u03c1\5\u0080A\2\u03c0\u03b5\3\2\2\2\u03c0")
 *         buf.write("\u03bf\3\2\2\2\u03c1\177\3\2\2\2\u03c2\u03c7\5\u0082B")
 *         buf.write("\2\u03c3\u03c4\7\27\2\2\u03c4\u03c6\5\u0082B\2\u03c5\u03c3")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u03c6\u03c9\3\2\2\2\u03c7\u03c5\3\2\2\2\u03c7")
 *         buf.write("\u03c8\3\2\2\2\u03c8\u0081\3\2\2\2\u03c9\u03c7\3\2\2\2")
 */
  __pyx_tuple__734 = PyTuple_Pack(1, __pyx_kp_u_B_3); if (unlikely(!__pyx_tuple__734)) __PYX_ERR(0, 483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__734);
  __Pyx_GIVEREF(__pyx_tuple__734);

  /* "pymola/generated/ModelicaParser.pyx":484
 *         buf.write("\u03bf\3\2\2\2\u03c1\177\3\2\2\2\u03c2\u03c7\5\u0082B")
 *         buf.write("\2\u03c3\u03c4\7\27\2\2\u03c4\u03c6\5\u0082B\2\u03c5\u03c3")
 *         buf.write("\3\2\2\2\u03c6\u03c9\3\2\2\2\u03c7\u03c5\3\2\2\2\u03c7")             # <<<<<<<<<<<<<<
 *         buf.write("\u03c8\3\2\2\2\u03c8\u0081\3\2\2\2\u03c9\u03c7\3\2\2\2")
 *         buf.write("\u03ca\u03cb\7Z\2\2\u03cb\u03cc\7\21\2\2\u03cc\u03cd\5")
 */
  __pyx_tuple__736 = PyTuple_Pack(1, __pyx_kp_u__735); if (unlikely(!__pyx_tuple__736)) __PYX_ERR(0, 484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__736);
  __Pyx_GIVEREF(__pyx_tuple__736);

  /* "pymola/generated/ModelicaParser.pyx":485
 *         buf.write("\2\u03c3\u03c4\7\27\2\2\u03c4\u03c6\5\u0082B\2\u03c5\u03c3")
 *         buf.write("\3\2\2\2\u03c6\u03c9\3\2\2\2\u03c7\u03c5\3\2\2\2\u03c7")
 *         buf.write("\u03c8\3\2\2\2\u03c8\u0081\3\2\2\2\u03c9\u03c7\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u03ca\u03cb\7Z\2\2\u03cb\u03cc\7\21\2\2\u03cc\u03cd\5")
 *         buf.write("\u0084C\2\u03cd\u0083\3\2\2\2\u03ce\u03cf\7\17\2\2\u03cf")
 */
  __pyx_tuple__738 = PyTuple_Pack(1, __pyx_kp_u__737); if (unlikely(!__pyx_tuple__738)) __PYX_ERR(0, 485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__738);
  __Pyx_GIVEREF(__pyx_tuple__738);

  /* "pymola/generated/ModelicaParser.pyx":486
 *         buf.write("\3\2\2\2\u03c6\u03c9\3\2\2\2\u03c7\u03c5\3\2\2\2\u03c7")
 *         buf.write("\u03c8\3\2\2\2\u03c8\u0081\3\2\2\2\u03c9\u03c7\3\2\2\2")
 *         buf.write("\u03ca\u03cb\7Z\2\2\u03cb\u03cc\7\21\2\2\u03cc\u03cd\5")             # <<<<<<<<<<<<<<
 *         buf.write("\u0084C\2\u03cd\u0083\3\2\2\2\u03ce\u03cf\7\17\2\2\u03cf")
 *         buf.write("\u03d0\5v<\2\u03d0\u03d2\7\23\2\2\u03d1\u03d3\5\u0080")
 */
  __pyx_tuple__739 = PyTuple_Pack(1, __pyx_kp_u_Z_12); if (unlikely(!__pyx_tuple__739)) __PYX_ERR(0, 486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__739);
  __Pyx_GIVEREF(__pyx_tuple__739);

  /* "pymola/generated/ModelicaParser.pyx":487
 *         buf.write("\u03c8\3\2\2\2\u03c8\u0081\3\2\2\2\u03c9\u03c7\3\2\2\2")
 *         buf.write("\u03ca\u03cb\7Z\2\2\u03cb\u03cc\7\21\2\2\u03cc\u03cd\5")
 *         buf.write("\u0084C\2\u03cd\u0083\3\2\2\2\u03ce\u03cf\7\17\2\2\u03cf")             # <<<<<<<<<<<<<<
 *         buf.write("\u03d0\5v<\2\u03d0\u03d2\7\23\2\2\u03d1\u03d3\5\u0080")
 *         buf.write("A\2\u03d2\u03d1\3\2\2\2\u03d2\u03d3\3\2\2\2\u03d3\u03d4")
 */
  __pyx_tuple__740 = PyTuple_Pack(1, __pyx_kp_u_C_3); if (unlikely(!__pyx_tuple__740)) __PYX_ERR(0, 487, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__740);
  __Pyx_GIVEREF(__pyx_tuple__740);

  /* "pymola/generated/ModelicaParser.pyx":488
 *         buf.write("\u03ca\u03cb\7Z\2\2\u03cb\u03cc\7\21\2\2\u03cc\u03cd\5")
 *         buf.write("\u0084C\2\u03cd\u0083\3\2\2\2\u03ce\u03cf\7\17\2\2\u03cf")
 *         buf.write("\u03d0\5v<\2\u03d0\u03d2\7\23\2\2\u03d1\u03d3\5\u0080")             # <<<<<<<<<<<<<<
 *         buf.write("A\2\u03d2\u03d1\3\2\2\2\u03d2\u03d3\3\2\2\2\u03d3\u03d4")
 *         buf.write("\3\2\2\2\u03d4\u03d5\7\25\2\2\u03d5\u03d8\3\2\2\2\u03d6")
 */
  __pyx_tuple__741 = PyTuple_Pack(1, __pyx_kp_u_v_3); if (unlikely(!__pyx_tuple__741)) __PYX_ERR(0, 488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__741);
  __Pyx_GIVEREF(__pyx_tuple__741);

  /* "pymola/generated/ModelicaParser.pyx":489
 *         buf.write("\u0084C\2\u03cd\u0083\3\2\2\2\u03ce\u03cf\7\17\2\2\u03cf")
 *         buf.write("\u03d0\5v<\2\u03d0\u03d2\7\23\2\2\u03d1\u03d3\5\u0080")
 *         buf.write("A\2\u03d2\u03d1\3\2\2\2\u03d2\u03d3\3\2\2\2\u03d3\u03d4")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u03d4\u03d5\7\25\2\2\u03d5\u03d8\3\2\2\2\u03d6")
 *         buf.write("\u03d8\5n8\2\u03d7\u03ce\3\2\2\2\u03d7\u03d6\3\2\2\2\u03d8")
 */
  __pyx_tuple__742 = PyTuple_Pack(1, __pyx_kp_u_A_4); if (unlikely(!__pyx_tuple__742)) __PYX_ERR(0, 489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__742);
  __Pyx_GIVEREF(__pyx_tuple__742);

  /* "pymola/generated/ModelicaParser.pyx":490
 *         buf.write("\u03d0\5v<\2\u03d0\u03d2\7\23\2\2\u03d1\u03d3\5\u0080")
 *         buf.write("A\2\u03d2\u03d1\3\2\2\2\u03d2\u03d3\3\2\2\2\u03d3\u03d4")
 *         buf.write("\3\2\2\2\u03d4\u03d5\7\25\2\2\u03d5\u03d8\3\2\2\2\u03d6")             # <<<<<<<<<<<<<<
 *         buf.write("\u03d8\5n8\2\u03d7\u03ce\3\2\2\2\u03d7\u03d6\3\2\2\2\u03d8")
 *         buf.write("\u0085\3\2\2\2\u03d9\u03db\5n8\2\u03da\u03d9\3\2\2\2\u03da")
 */
  __pyx_tuple__744 = PyTuple_Pack(1, __pyx_kp_u__743); if (unlikely(!__pyx_tuple__744)) __PYX_ERR(0, 490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__744);
  __Pyx_GIVEREF(__pyx_tuple__744);

  /* "pymola/generated/ModelicaParser.pyx":491
 *         buf.write("A\2\u03d2\u03d1\3\2\2\2\u03d2\u03d3\3\2\2\2\u03d3\u03d4")
 *         buf.write("\3\2\2\2\u03d4\u03d5\7\25\2\2\u03d5\u03d8\3\2\2\2\u03d6")
 *         buf.write("\u03d8\5n8\2\u03d7\u03ce\3\2\2\2\u03d7\u03d6\3\2\2\2\u03d8")             # <<<<<<<<<<<<<<
 *         buf.write("\u0085\3\2\2\2\u03d9\u03db\5n8\2\u03da\u03d9\3\2\2\2\u03da")
 *         buf.write("\u03db\3\2\2\2\u03db\u03e0\3\2\2\2\u03dc\u03dd\7\27\2")
 */
  __pyx_tuple__745 = PyTuple_Pack(1, __pyx_kp_u_n8_7); if (unlikely(!__pyx_tuple__745)) __PYX_ERR(0, 491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__745);
  __Pyx_GIVEREF(__pyx_tuple__745);

  /* "pymola/generated/ModelicaParser.pyx":492
 *         buf.write("\3\2\2\2\u03d4\u03d5\7\25\2\2\u03d5\u03d8\3\2\2\2\u03d6")
 *         buf.write("\u03d8\5n8\2\u03d7\u03ce\3\2\2\2\u03d7\u03d6\3\2\2\2\u03d8")
 *         buf.write("\u0085\3\2\2\2\u03d9\u03db\5n8\2\u03da\u03d9\3\2\2\2\u03da")             # <<<<<<<<<<<<<<
 *         buf.write("\u03db\3\2\2\2\u03db\u03e0\3\2\2\2\u03dc\u03dd\7\27\2")
 *         buf.write("\2\u03dd\u03df\5n8\2\u03de\u03dc\3\2\2\2\u03df\u03e2\3")
 */
  __pyx_tuple__746 = PyTuple_Pack(1, __pyx_kp_u_n8_8); if (unlikely(!__pyx_tuple__746)) __PYX_ERR(0, 492, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__746);
  __Pyx_GIVEREF(__pyx_tuple__746);

  /* "pymola/generated/ModelicaParser.pyx":493
 *         buf.write("\u03d8\5n8\2\u03d7\u03ce\3\2\2\2\u03d7\u03d6\3\2\2\2\u03d8")
 *         buf.write("\u0085\3\2\2\2\u03d9\u03db\5n8\2\u03da\u03d9\3\2\2\2\u03da")
 *         buf.write("\u03db\3\2\2\2\u03db\u03e0\3\2\2\2\u03dc\u03dd\7\27\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\u03dd\u03df\5n8\2\u03de\u03dc\3\2\2\2\u03df\u03e2\3")
 *         buf.write("\2\2\2\u03e0\u03de\3\2\2\2\u03e0\u03e1\3\2\2\2\u03e1\u0087")
 */
  __pyx_tuple__748 = PyTuple_Pack(1, __pyx_kp_u__747); if (unlikely(!__pyx_tuple__748)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__748);
  __Pyx_GIVEREF(__pyx_tuple__748);

  /* "pymola/generated/ModelicaParser.pyx":494
 *         buf.write("\u0085\3\2\2\2\u03d9\u03db\5n8\2\u03da\u03d9\3\2\2\2\u03da")
 *         buf.write("\u03db\3\2\2\2\u03db\u03e0\3\2\2\2\u03dc\u03dd\7\27\2")
 *         buf.write("\2\u03dd\u03df\5n8\2\u03de\u03dc\3\2\2\2\u03df\u03e2\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u03e0\u03de\3\2\2\2\u03e0\u03e1\3\2\2\2\u03e1\u0087")
 *         buf.write("\3\2\2\2\u03e2\u03e0\3\2\2\2\u03e3\u03e8\5n8\2\u03e4\u03e5")
 */
  __pyx_tuple__749 = PyTuple_Pack(1, __pyx_kp_u_n8_9); if (unlikely(!__pyx_tuple__749)) __PYX_ERR(0, 494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__749);
  __Pyx_GIVEREF(__pyx_tuple__749);

  /* "pymola/generated/ModelicaParser.pyx":495
 *         buf.write("\u03db\3\2\2\2\u03db\u03e0\3\2\2\2\u03dc\u03dd\7\27\2")
 *         buf.write("\2\u03dd\u03df\5n8\2\u03de\u03dc\3\2\2\2\u03df\u03e2\3")
 *         buf.write("\2\2\2\u03e0\u03de\3\2\2\2\u03e0\u03e1\3\2\2\2\u03e1\u0087")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u03e2\u03e0\3\2\2\2\u03e3\u03e8\5n8\2\u03e4\u03e5")
 *         buf.write("\7\27\2\2\u03e5\u03e7\5n8\2\u03e6\u03e4\3\2\2\2\u03e7")
 */
  __pyx_tuple__751 = PyTuple_Pack(1, __pyx_kp_u__750); if (unlikely(!__pyx_tuple__751)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__751);
  __Pyx_GIVEREF(__pyx_tuple__751);

  /* "pymola/generated/ModelicaParser.pyx":496
 *         buf.write("\2\u03dd\u03df\5n8\2\u03de\u03dc\3\2\2\2\u03df\u03e2\3")
 *         buf.write("\2\2\2\u03e0\u03de\3\2\2\2\u03e0\u03e1\3\2\2\2\u03e1\u0087")
 *         buf.write("\3\2\2\2\u03e2\u03e0\3\2\2\2\u03e3\u03e8\5n8\2\u03e4\u03e5")             # <<<<<<<<<<<<<<
 *         buf.write("\7\27\2\2\u03e5\u03e7\5n8\2\u03e6\u03e4\3\2\2\2\u03e7")
 *         buf.write("\u03ea\3\2\2\2\u03e8\u03e6\3\2\2\2\u03e8\u03e9\3\2\2\2")
 */
  __pyx_tuple__752 = PyTuple_Pack(1, __pyx_kp_u_n8_10); if (unlikely(!__pyx_tuple__752)) __PYX_ERR(0, 496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__752);
  __Pyx_GIVEREF(__pyx_tuple__752);

  /* "pymola/generated/ModelicaParser.pyx":497
 *         buf.write("\2\2\2\u03e0\u03de\3\2\2\2\u03e0\u03e1\3\2\2\2\u03e1\u0087")
 *         buf.write("\3\2\2\2\u03e2\u03e0\3\2\2\2\u03e3\u03e8\5n8\2\u03e4\u03e5")
 *         buf.write("\7\27\2\2\u03e5\u03e7\5n8\2\u03e6\u03e4\3\2\2\2\u03e7")             # <<<<<<<<<<<<<<
 *         buf.write("\u03ea\3\2\2\2\u03e8\u03e6\3\2\2\2\u03e8\u03e9\3\2\2\2")
 *         buf.write("\u03e9\u0089\3\2\2\2\u03ea\u03e8\3\2\2\2\u03eb\u03ec\7")
 */
  __pyx_tuple__753 = PyTuple_Pack(1, __pyx_kp_u_n8_11); if (unlikely(!__pyx_tuple__753)) __PYX_ERR(0, 497, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__753);
  __Pyx_GIVEREF(__pyx_tuple__753);

  /* "pymola/generated/ModelicaParser.pyx":498
 *         buf.write("\3\2\2\2\u03e2\u03e0\3\2\2\2\u03e3\u03e8\5n8\2\u03e4\u03e5")
 *         buf.write("\7\27\2\2\u03e5\u03e7\5n8\2\u03e6\u03e4\3\2\2\2\u03e7")
 *         buf.write("\u03ea\3\2\2\2\u03e8\u03e6\3\2\2\2\u03e8\u03e9\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u03e9\u0089\3\2\2\2\u03ea\u03e8\3\2\2\2\u03eb\u03ec\7")
 *         buf.write("N\2\2\u03ec\u03f1\5\u008cG\2\u03ed\u03ee\7\27\2\2\u03ee")
 */
  __pyx_tuple__755 = PyTuple_Pack(1, __pyx_kp_u__754); if (unlikely(!__pyx_tuple__755)) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__755);
  __Pyx_GIVEREF(__pyx_tuple__755);

  /* "pymola/generated/ModelicaParser.pyx":499
 *         buf.write("\7\27\2\2\u03e5\u03e7\5n8\2\u03e6\u03e4\3\2\2\2\u03e7")
 *         buf.write("\u03ea\3\2\2\2\u03e8\u03e6\3\2\2\2\u03e8\u03e9\3\2\2\2")
 *         buf.write("\u03e9\u0089\3\2\2\2\u03ea\u03e8\3\2\2\2\u03eb\u03ec\7")             # <<<<<<<<<<<<<<
 *         buf.write("N\2\2\u03ec\u03f1\5\u008cG\2\u03ed\u03ee\7\27\2\2\u03ee")
 *         buf.write("\u03f0\5\u008cG\2\u03ef\u03ed\3\2\2\2\u03f0\u03f3\3\2")
 */
  __pyx_tuple__757 = PyTuple_Pack(1, __pyx_kp_u__756); if (unlikely(!__pyx_tuple__757)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__757);
  __Pyx_GIVEREF(__pyx_tuple__757);

  /* "pymola/generated/ModelicaParser.pyx":500
 *         buf.write("\u03ea\3\2\2\2\u03e8\u03e6\3\2\2\2\u03e8\u03e9\3\2\2\2")
 *         buf.write("\u03e9\u0089\3\2\2\2\u03ea\u03e8\3\2\2\2\u03eb\u03ec\7")
 *         buf.write("N\2\2\u03ec\u03f1\5\u008cG\2\u03ed\u03ee\7\27\2\2\u03ee")             # <<<<<<<<<<<<<<
 *         buf.write("\u03f0\5\u008cG\2\u03ef\u03ed\3\2\2\2\u03f0\u03f3\3\2")
 *         buf.write("\2\2\u03f1\u03ef\3\2\2\2\u03f1\u03f2\3\2\2\2\u03f2\u03f4")
 */
  __pyx_tuple__758 = PyTuple_Pack(1, __pyx_kp_u_N_G); if (unlikely(!__pyx_tuple__758)) __PYX_ERR(0, 500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__758);
  __Pyx_GIVEREF(__pyx_tuple__758);

  /* "pymola/generated/ModelicaParser.pyx":501
 *         buf.write("\u03e9\u0089\3\2\2\2\u03ea\u03e8\3\2\2\2\u03eb\u03ec\7")
 *         buf.write("N\2\2\u03ec\u03f1\5\u008cG\2\u03ed\u03ee\7\27\2\2\u03ee")
 *         buf.write("\u03f0\5\u008cG\2\u03ef\u03ed\3\2\2\2\u03f0\u03f3\3\2")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\u03f1\u03ef\3\2\2\2\u03f1\u03f2\3\2\2\2\u03f2\u03f4")
 *         buf.write("\3\2\2\2\u03f3\u03f1\3\2\2\2\u03f4\u03f5\7O\2\2\u03f5")
 */
  __pyx_tuple__759 = PyTuple_Pack(1, __pyx_kp_u_G_2); if (unlikely(!__pyx_tuple__759)) __PYX_ERR(0, 501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__759);
  __Pyx_GIVEREF(__pyx_tuple__759);

  /* "pymola/generated/ModelicaParser.pyx":502
 *         buf.write("N\2\2\u03ec\u03f1\5\u008cG\2\u03ed\u03ee\7\27\2\2\u03ee")
 *         buf.write("\u03f0\5\u008cG\2\u03ef\u03ed\3\2\2\2\u03f0\u03f3\3\2")
 *         buf.write("\2\2\u03f1\u03ef\3\2\2\2\u03f1\u03f2\3\2\2\2\u03f2\u03f4")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u03f3\u03f1\3\2\2\2\u03f4\u03f5\7O\2\2\u03f5")
 *         buf.write("\u008b\3\2\2\2\u03f6\u03f9\7\24\2\2\u03f7\u03f9\5n8\2")
 */
  __pyx_tuple__761 = PyTuple_Pack(1, __pyx_kp_u__760); if (unlikely(!__pyx_tuple__761)) __PYX_ERR(0, 502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__761);
  __Pyx_GIVEREF(__pyx_tuple__761);

  /* "pymola/generated/ModelicaParser.pyx":503
 *         buf.write("\u03f0\5\u008cG\2\u03ef\u03ed\3\2\2\2\u03f0\u03f3\3\2")
 *         buf.write("\2\2\u03f1\u03ef\3\2\2\2\u03f1\u03f2\3\2\2\2\u03f2\u03f4")
 *         buf.write("\3\2\2\2\u03f3\u03f1\3\2\2\2\u03f4\u03f5\7O\2\2\u03f5")             # <<<<<<<<<<<<<<
 *         buf.write("\u008b\3\2\2\2\u03f6\u03f9\7\24\2\2\u03f7\u03f9\5n8\2")
 *         buf.write("\u03f8\u03f6\3\2\2\2\u03f8\u03f7\3\2\2\2\u03f9\u008d\3")
 */
  __pyx_tuple__762 = PyTuple_Pack(1, __pyx_kp_u_O_3); if (unlikely(!__pyx_tuple__762)) __PYX_ERR(0, 503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__762);
  __Pyx_GIVEREF(__pyx_tuple__762);

  /* "pymola/generated/ModelicaParser.pyx":504
 *         buf.write("\2\2\u03f1\u03ef\3\2\2\2\u03f1\u03f2\3\2\2\2\u03f2\u03f4")
 *         buf.write("\3\2\2\2\u03f3\u03f1\3\2\2\2\u03f4\u03f5\7O\2\2\u03f5")
 *         buf.write("\u008b\3\2\2\2\u03f6\u03f9\7\24\2\2\u03f7\u03f9\5n8\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u03f8\u03f6\3\2\2\2\u03f8\u03f7\3\2\2\2\u03f9\u008d\3")
 *         buf.write("\2\2\2\u03fa\u03fc\5\u0090I\2\u03fb\u03fd\5\u0092J\2\u03fc")
 */
  __pyx_tuple__763 = PyTuple_Pack(1, __pyx_kp_u_n8_12); if (unlikely(!__pyx_tuple__763)) __PYX_ERR(0, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__763);
  __Pyx_GIVEREF(__pyx_tuple__763);

  /* "pymola/generated/ModelicaParser.pyx":505
 *         buf.write("\3\2\2\2\u03f3\u03f1\3\2\2\2\u03f4\u03f5\7O\2\2\u03f5")
 *         buf.write("\u008b\3\2\2\2\u03f6\u03f9\7\24\2\2\u03f7\u03f9\5n8\2")
 *         buf.write("\u03f8\u03f6\3\2\2\2\u03f8\u03f7\3\2\2\2\u03f9\u008d\3")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\2\u03fa\u03fc\5\u0090I\2\u03fb\u03fd\5\u0092J\2\u03fc")
 *         buf.write("\u03fb\3\2\2\2\u03fc\u03fd\3\2\2\2\u03fd\u008f\3\2\2\2")
 */
  __pyx_tuple__765 = PyTuple_Pack(1, __pyx_kp_u__764); if (unlikely(!__pyx_tuple__765)) __PYX_ERR(0, 505, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__765);
  __Pyx_GIVEREF(__pyx_tuple__765);

  /* "pymola/generated/ModelicaParser.pyx":506
 *         buf.write("\u008b\3\2\2\2\u03f6\u03f9\7\24\2\2\u03f7\u03f9\5n8\2")
 *         buf.write("\u03f8\u03f6\3\2\2\2\u03f8\u03f7\3\2\2\2\u03f9\u008d\3")
 *         buf.write("\2\2\2\u03fa\u03fc\5\u0090I\2\u03fb\u03fd\5\u0092J\2\u03fc")             # <<<<<<<<<<<<<<
 *         buf.write("\u03fb\3\2\2\2\u03fc\u03fd\3\2\2\2\u03fd\u008f\3\2\2\2")
 *         buf.write("\u03fe\u0403\7[\2\2\u03ff\u0400\7@\2\2\u0400\u0402\7[")
 */
  __pyx_tuple__766 = PyTuple_Pack(1, __pyx_kp_u_I_J); if (unlikely(!__pyx_tuple__766)) __PYX_ERR(0, 506, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__766);
  __Pyx_GIVEREF(__pyx_tuple__766);

  /* "pymola/generated/ModelicaParser.pyx":507
 *         buf.write("\u03f8\u03f6\3\2\2\2\u03f8\u03f7\3\2\2\2\u03f9\u008d\3")
 *         buf.write("\2\2\2\u03fa\u03fc\5\u0090I\2\u03fb\u03fd\5\u0092J\2\u03fc")
 *         buf.write("\u03fb\3\2\2\2\u03fc\u03fd\3\2\2\2\u03fd\u008f\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u03fe\u0403\7[\2\2\u03ff\u0400\7@\2\2\u0400\u0402\7[")
 *         buf.write("\2\2\u0401\u03ff\3\2\2\2\u0402\u0405\3\2\2\2\u0403\u0401")
 */
  __pyx_tuple__768 = PyTuple_Pack(1, __pyx_kp_u__767); if (unlikely(!__pyx_tuple__768)) __PYX_ERR(0, 507, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__768);
  __Pyx_GIVEREF(__pyx_tuple__768);

  /* "pymola/generated/ModelicaParser.pyx":508
 *         buf.write("\2\2\2\u03fa\u03fc\5\u0090I\2\u03fb\u03fd\5\u0092J\2\u03fc")
 *         buf.write("\u03fb\3\2\2\2\u03fc\u03fd\3\2\2\2\u03fd\u008f\3\2\2\2")
 *         buf.write("\u03fe\u0403\7[\2\2\u03ff\u0400\7@\2\2\u0400\u0402\7[")             # <<<<<<<<<<<<<<
 *         buf.write("\2\2\u0401\u03ff\3\2\2\2\u0402\u0405\3\2\2\2\u0403\u0401")
 *         buf.write("\3\2\2\2\u0403\u0404\3\2\2\2\u0404\u0407\3\2\2\2\u0405")
 */
  __pyx_tuple__770 = PyTuple_Pack(1, __pyx_kp_u__769); if (unlikely(!__pyx_tuple__770)) __PYX_ERR(0, 508, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__770);
  __Pyx_GIVEREF(__pyx_tuple__770);

  /* "pymola/generated/ModelicaParser.pyx":509
 *         buf.write("\u03fb\3\2\2\2\u03fc\u03fd\3\2\2\2\u03fd\u008f\3\2\2\2")
 *         buf.write("\u03fe\u0403\7[\2\2\u03ff\u0400\7@\2\2\u0400\u0402\7[")
 *         buf.write("\2\2\u0401\u03ff\3\2\2\2\u0402\u0405\3\2\2\2\u0403\u0401")             # <<<<<<<<<<<<<<
 *         buf.write("\3\2\2\2\u0403\u0404\3\2\2\2\u0404\u0407\3\2\2\2\u0405")
 *         buf.write("\u0403\3\2\2\2\u0406\u03fe\3\2\2\2\u0406\u0407\3\2\2\2")
 */
  __pyx_tuple__772 = PyTuple_Pack(1, __pyx_kp_u__771); if (unlikely(!__pyx_tuple__772)) __PYX_ERR(0, 509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__772);
  __Pyx_GIVEREF(__pyx_tuple__772);

  /* "pymola/generated/ModelicaParser.pyx":510
 *         buf.write("\u03fe\u0403\7[\2\2\u03ff\u0400\7@\2\2\u0400\u0402\7[")
 *         buf.write("\2\2\u0401\u03ff\3\2\2\2\u0402\u0405\3\2\2\2\u0403\u0401")
 *         buf.write("\3\2\2\2\u0403\u0404\3\2\2\2\u0404\u0407\3\2\2\2\u0405")             # <<<<<<<<<<<<<<
 *         buf.write("\u0403\3\2\2\2\u0406\u03fe\3\2\2\2\u0406\u0407\3\2\2\2")
 *         buf.write("\u0407\u0091\3\2\2\2\u0408\u0409\7P\2\2\u0409\u040a\5")
 */
  __pyx_tuple__774 = PyTuple_Pack(1, __pyx_kp_u__773); if (unlikely(!__pyx_tuple__774)) __PYX_ERR(0, 510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__774);
  __Pyx_GIVEREF(__pyx_tuple__774);

  /* "pymola/generated/ModelicaParser.pyx":511
 *         buf.write("\2\2\u0401\u03ff\3\2\2\2\u0402\u0405\3\2\2\2\u0403\u0401")
 *         buf.write("\3\2\2\2\u0403\u0404\3\2\2\2\u0404\u0407\3\2\2\2\u0405")
 *         buf.write("\u0403\3\2\2\2\u0406\u03fe\3\2\2\2\u0406\u0407\3\2\2\2")             # <<<<<<<<<<<<<<
 *         buf.write("\u0407\u0091\3\2\2\2\u0408\u0409\7P\2\2\u0409\u040a\5")
 *         buf.write(":\36\2\u040a\u0093\3\2\2\2\u0087\u0096\u0099\u009e\u00a2")
 */
  __pyx_tuple__776 = PyTuple_Pack(1, __pyx_kp_u__775); if (unlikely(!__pyx_tuple__776)) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__776);
  __Pyx_GIVEREF(__pyx_tuple__776);

  /* "pymola/generated/ModelicaParser.pyx":512
 *         buf.write("\3\2\2\2\u0403\u0404\3\2\2\2\u0404\u0407\3\2\2\2\u0405")
 *         buf.write("\u0403\3\2\2\2\u0406\u03fe\3\2\2\2\u0406\u0407\3\2\2\2")
 *         buf.write("\u0407\u0091\3\2\2\2\u0408\u0409\7P\2\2\u0409\u040a\5")             # <<<<<<<<<<<<<<
 *         buf.write(":\36\2\u040a\u0093\3\2\2\2\u0087\u0096\u0099\u009e\u00a2")
 *         buf.write("\u00a8\u00ae\u00b5\u00ba\u00c0\u00c3\u00c7\u00d4\u00d7")
 */
  __pyx_tuple__777 = PyTuple_Pack(1, __pyx_kp_u_P_2); if (unlikely(!__pyx_tuple__777)) __PYX_ERR(0, 512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__777);
  __Pyx_GIVEREF(__pyx_tuple__777);

  /* "pymola/generated/ModelicaParser.pyx":513
 *         buf.write("\u0403\3\2\2\2\u0406\u03fe\3\2\2\2\u0406\u0407\3\2\2\2")
 *         buf.write("\u0407\u0091\3\2\2\2\u0408\u0409\7P\2\2\u0409\u040a\5")
 *         buf.write(":\36\2\u040a\u0093\3\2\2\2\u0087\u0096\u0099\u009e\u00a2")             # <<<<<<<<<<<<<<
 *         buf.write("\u00a8\u00ae\u00b5\u00ba\u00c0\u00c3\u00c7\u00d4\u00d7")
 *         buf.write("\u00e0\u00e3\u00f2\u00fb\u0102\u010b\u0118\u011a\u011f")
 */
  __pyx_tuple__779 = PyTuple_Pack(1, __pyx_kp_u__778); if (unlikely(!__pyx_tuple__779)) __PYX_ERR(0, 513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__779);
  __Pyx_GIVEREF(__pyx_tuple__779);

  /* "pymola/generated/ModelicaParser.pyx":514
 *         buf.write("\u0407\u0091\3\2\2\2\u0408\u0409\7P\2\2\u0409\u040a\5")
 *         buf.write(":\36\2\u040a\u0093\3\2\2\2\u0087\u0096\u0099\u009e\u00a2")
 *         buf.write("\u00a8\u00ae\u00b5\u00ba\u00c0\u00c3\u00c7\u00d4\u00d7")             # <<<<<<<<<<<<<<
 *         buf.write("\u00e0\u00e3\u00f2\u00fb\u0102\u010b\u0118\u011a\u011f")
 *         buf.write("\u0122\u0125\u0128\u012d\u0134\u0139\u0142\u0149\u014c")
 */
  __pyx_tuple__781 = PyTuple_Pack(1, __pyx_kp_u__780); if (unlikely(!__pyx_tuple__781)) __PYX_ERR(0, 514, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__781);
  __Pyx_GIVEREF(__pyx_tuple__781);

  /* "pymola/generated/ModelicaParser.pyx":515
 *         buf.write(":\36\2\u040a\u0093\3\2\2\2\u0087\u0096\u0099\u009e\u00a2")
 *         buf.write("\u00a8\u00ae\u00b5\u00ba\u00c0\u00c3\u00c7\u00d4\u00d7")
 *         buf.write("\u00e0\u00e3\u00f2\u00fb\u0102\u010b\u0118\u011a\u011f")             # <<<<<<<<<<<<<<
 *         buf.write("\u0122\u0125\u0128\u012d\u0134\u0139\u0142\u0149\u014c")
 *         buf.write("\u014f\u0152\u0155\u0159\u015c\u015f\u0162\u0165\u016a")
 */
  __pyx_tuple__783 = PyTuple_Pack(1, __pyx_n_u__782); if (unlikely(!__pyx_tuple__783)) __PYX_ERR(0, 515, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__783);
  __Pyx_GIVEREF(__pyx_tuple__783);

  /* "pymola/generated/ModelicaParser.pyx":516
 *         buf.write("\u00a8\u00ae\u00b5\u00ba\u00c0\u00c3\u00c7\u00d4\u00d7")
 *         buf.write("\u00e0\u00e3\u00f2\u00fb\u0102\u010b\u0118\u011a\u011f")
 *         buf.write("\u0122\u0125\u0128\u012d\u0134\u0139\u0142\u0149\u014c")             # <<<<<<<<<<<<<<
 *         buf.write("\u014f\u0152\u0155\u0159\u015c\u015f\u0162\u0165\u016a")
 *         buf.write("\u016f\u017c\u017e\u0180\u0189\u018f\u0192\u0197\u019c")
 */
  __pyx_tuple__785 = PyTuple_Pack(1, __pyx_n_u__784); if (unlikely(!__pyx_tuple__785)) __PYX_ERR(0, 516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__785);
  __Pyx_GIVEREF(__pyx_tuple__785);

  /* "pymola/generated/ModelicaParser.pyx":517
 *         buf.write("\u00e0\u00e3\u00f2\u00fb\u0102\u010b\u0118\u011a\u011f")
 *         buf.write("\u0122\u0125\u0128\u012d\u0134\u0139\u0142\u0149\u014c")
 *         buf.write("\u014f\u0152\u0155\u0159\u015c\u015f\u0162\u0165\u016a")             # <<<<<<<<<<<<<<
 *         buf.write("\u016f\u017c\u017e\u0180\u0189\u018f\u0192\u0197\u019c")
 *         buf.write("\u01a1\u01a4\u01a7\u01b0\u01b5\u01be\u01c1\u01c6\u01cc")
 */
  __pyx_tuple__787 = PyTuple_Pack(1, __pyx_n_u__786); if (unlikely(!__pyx_tuple__787)) __PYX_ERR(0, 517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__787);
  __Pyx_GIVEREF(__pyx_tuple__787);

  /* "pymola/generated/ModelicaParser.pyx":518
 *         buf.write("\u0122\u0125\u0128\u012d\u0134\u0139\u0142\u0149\u014c")
 *         buf.write("\u014f\u0152\u0155\u0159\u015c\u015f\u0162\u0165\u016a")
 *         buf.write("\u016f\u017c\u017e\u0180\u0189\u018f\u0192\u0197\u019c")             # <<<<<<<<<<<<<<
 *         buf.write("\u01a1\u01a4\u01a7\u01b0\u01b5\u01be\u01c1\u01c6\u01cc")
 *         buf.write("\u01d0\u01d9\u01de\u01e1\u01e4\u01e8\u01ec\u01f2\u01f5")
 */
  __pyx_tuple__789 = PyTuple_Pack(1, __pyx_n_u__788); if (unlikely(!__pyx_tuple__789)) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__789);
  __Pyx_GIVEREF(__pyx_tuple__789);

  /* "pymola/generated/ModelicaParser.pyx":519
 *         buf.write("\u014f\u0152\u0155\u0159\u015c\u015f\u0162\u0165\u016a")
 *         buf.write("\u016f\u017c\u017e\u0180\u0189\u018f\u0192\u0197\u019c")
 *         buf.write("\u01a1\u01a4\u01a7\u01b0\u01b5\u01be\u01c1\u01c6\u01cc")             # <<<<<<<<<<<<<<
 *         buf.write("\u01d0\u01d9\u01de\u01e1\u01e4\u01e8\u01ec\u01f2\u01f5")
 *         buf.write("\u01f9\u01fc\u0201\u0204\u0213\u0216\u021d\u0220\u0224")
 */
  __pyx_tuple__791 = PyTuple_Pack(1, __pyx_n_u__790); if (unlikely(!__pyx_tuple__791)) __PYX_ERR(0, 519, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__791);
  __Pyx_GIVEREF(__pyx_tuple__791);

  /* "pymola/generated/ModelicaParser.pyx":520
 *         buf.write("\u016f\u017c\u017e\u0180\u0189\u018f\u0192\u0197\u019c")
 *         buf.write("\u01a1\u01a4\u01a7\u01b0\u01b5\u01be\u01c1\u01c6\u01cc")
 *         buf.write("\u01d0\u01d9\u01de\u01e1\u01e4\u01e8\u01ec\u01f2\u01f5")             # <<<<<<<<<<<<<<
 *         buf.write("\u01f9\u01fc\u0201\u0204\u0213\u0216\u021d\u0220\u0224")
 *         buf.write("\u0227\u022f\u0233\u023b\u0249\u0252\u025a\u0268\u0275")
 */
  __pyx_tuple__793 = PyTuple_Pack(1, __pyx_n_u__792); if (unlikely(!__pyx_tuple__793)) __PYX_ERR(0, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__793);
  __Pyx_GIVEREF(__pyx_tuple__793);

  /* "pymola/generated/ModelicaParser.pyx":521
 *         buf.write("\u01a1\u01a4\u01a7\u01b0\u01b5\u01be\u01c1\u01c6\u01cc")
 *         buf.write("\u01d0\u01d9\u01de\u01e1\u01e4\u01e8\u01ec\u01f2\u01f5")
 *         buf.write("\u01f9\u01fc\u0201\u0204\u0213\u0216\u021d\u0220\u0224")             # <<<<<<<<<<<<<<
 *         buf.write("\u0227\u022f\u0233\u023b\u0249\u0252\u025a\u0268\u0275")
 *         buf.write("\u0280\u0285\u028e\u0291\u029e\u02a9\u02ae\u02b7\u02ba")
 */
  __pyx_tuple__795 = PyTuple_Pack(1, __pyx_n_u__794); if (unlikely(!__pyx_tuple__795)) __PYX_ERR(0, 521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__795);
  __Pyx_GIVEREF(__pyx_tuple__795);

  /* "pymola/generated/ModelicaParser.pyx":522
 *         buf.write("\u01d0\u01d9\u01de\u01e1\u01e4\u01e8\u01ec\u01f2\u01f5")
 *         buf.write("\u01f9\u01fc\u0201\u0204\u0213\u0216\u021d\u0220\u0224")
 *         buf.write("\u0227\u022f\u0233\u023b\u0249\u0252\u025a\u0268\u0275")             # <<<<<<<<<<<<<<
 *         buf.write("\u0280\u0285\u028e\u0291\u029e\u02a9\u02ae\u02b7\u02ba")
 *         buf.write("\u02c7\u02d5\u02e0\u02e6\u02f0\u02fe\u0309\u030e\u031c")
 */
  __pyx_tuple__797 = PyTuple_Pack(1, __pyx_n_u__796); if (unlikely(!__pyx_tuple__797)) __PYX_ERR(0, 522, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__797);
  __Pyx_GIVEREF(__pyx_tuple__797);

  /* "pymola/generated/ModelicaParser.pyx":523
 *         buf.write("\u01f9\u01fc\u0201\u0204\u0213\u0216\u021d\u0220\u0224")
 *         buf.write("\u0227\u022f\u0233\u023b\u0249\u0252\u025a\u0268\u0275")
 *         buf.write("\u0280\u0285\u028e\u0291\u029e\u02a9\u02ae\u02b7\u02ba")             # <<<<<<<<<<<<<<
 *         buf.write("\u02c7\u02d5\u02e0\u02e6\u02f0\u02fe\u0309\u030e\u031c")
 *         buf.write("\u0327\u032c\u0345\u034b\u0352\u0354\u0360\u0371\u0373")
 */
  __pyx_tuple__799 = PyTuple_Pack(1, __pyx_n_u__798); if (unlikely(!__pyx_tuple__799)) __PYX_ERR(0, 523, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__799);
  __Pyx_GIVEREF(__pyx_tuple__799);

  /* "pymola/generated/ModelicaParser.pyx":524
 *         buf.write("\u0227\u022f\u0233\u023b\u0249\u0252\u025a\u0268\u0275")
 *         buf.write("\u0280\u0285\u028e\u0291\u029e\u02a9\u02ae\u02b7\u02ba")
 *         buf.write("\u02c7\u02d5\u02e0\u02e6\u02f0\u02fe\u0309\u030e\u031c")             # <<<<<<<<<<<<<<
 *         buf.write("\u0327\u032c\u0345\u034b\u0352\u0354\u0360\u0371\u0373")
 *         buf.write("\u038c\u0396\u0399\u03a0\u03a5\u03ac\u03b1\u03ba\u03bc")
 */
  __pyx_tuple__801 = PyTuple_Pack(1, __pyx_kp_u__800); if (unlikely(!__pyx_tuple__801)) __PYX_ERR(0, 524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__801);
  __Pyx_GIVEREF(__pyx_tuple__801);

  /* "pymola/generated/ModelicaParser.pyx":525
 *         buf.write("\u0280\u0285\u028e\u0291\u029e\u02a9\u02ae\u02b7\u02ba")
 *         buf.write("\u02c7\u02d5\u02e0\u02e6\u02f0\u02fe\u0309\u030e\u031c")
 *         buf.write("\u0327\u032c\u0345\u034b\u0352\u0354\u0360\u0371\u0373")             # <<<<<<<<<<<<<<
 *         buf.write("\u038c\u0396\u0399\u03a0\u03a5\u03ac\u03b1\u03ba\u03bc")
 *         buf.write("\u03c0\u03c7\u03d2\u03d7\u03da\u03e0\u03e8\u03f1\u03f8")
 */
  __pyx_tuple__803 = PyTuple_Pack(1, __pyx_kp_u__802); if (unlikely(!__pyx_tuple__803)) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__803);
  __Pyx_GIVEREF(__pyx_tuple__803);

  /* "pymola/generated/ModelicaParser.pyx":526
 *         buf.write("\u02c7\u02d5\u02e0\u02e6\u02f0\u02fe\u0309\u030e\u031c")
 *         buf.write("\u0327\u032c\u0345\u034b\u0352\u0354\u0360\u0371\u0373")
 *         buf.write("\u038c\u0396\u0399\u03a0\u03a5\u03ac\u03b1\u03ba\u03bc")             # <<<<<<<<<<<<<<
 *         buf.write("\u03c0\u03c7\u03d2\u03d7\u03da\u03e0\u03e8\u03f1\u03f8")
 *         buf.write("\u03fc\u0403\u0406")
 */
  __pyx_tuple__805 = PyTuple_Pack(1, __pyx_n_u__804); if (unlikely(!__pyx_tuple__805)) __PYX_ERR(0, 526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__805);
  __Pyx_GIVEREF(__pyx_tuple__805);

  /* "pymola/generated/ModelicaParser.pyx":527
 *         buf.write("\u0327\u032c\u0345\u034b\u0352\u0354\u0360\u0371\u0373")
 *         buf.write("\u038c\u0396\u0399\u03a0\u03a5\u03ac\u03b1\u03ba\u03bc")
 *         buf.write("\u03c0\u03c7\u03d2\u03d7\u03da\u03e0\u03e8\u03f1\u03f8")             # <<<<<<<<<<<<<<
 *         buf.write("\u03fc\u0403\u0406")
 *         return buf.getvalue()
 */
  __pyx_tuple__807 = PyTuple_Pack(1, __pyx_n_u__806); if (unlikely(!__pyx_tuple__807)) __PYX_ERR(0, 527, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__807);
  __Pyx_GIVEREF(__pyx_tuple__807);

  /* "pymola/generated/ModelicaParser.pyx":528
 *         buf.write("\u038c\u0396\u0399\u03a0\u03a5\u03ac\u03b1\u03ba\u03bc")
 *         buf.write("\u03c0\u03c7\u03d2\u03d7\u03da\u03e0\u03e8\u03f1\u03f8")
 *         buf.write("\u03fc\u0403\u0406")             # <<<<<<<<<<<<<<
 *         return buf.getvalue()
 * 
 */
  __pyx_tuple__809 = PyTuple_Pack(1, __pyx_n_u__808); if (unlikely(!__pyx_tuple__809)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__809);
  __Pyx_GIVEREF(__pyx_tuple__809);

  /* "pymola/generated/ModelicaParser.pyx":9
 * 
 * def serializedATN():
 *     with StringIO() as buf:             # <<<<<<<<<<<<<<
 *         buf.write("\3\u608b\ua72a\u8133\ub9ed\u417c\u3be7\u7786\u5964\3^")
 *         buf.write("\u040c\4\2\t\2\4\3\t\3\4\4\t\4\4\5\t\5\4\6\t\6\4\7\t\7")
 */
  __pyx_tuple__810 = PyTuple_Pack(3, Py_None, Py_None, Py_None); if (unlikely(!__pyx_tuple__810)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__810);
  __Pyx_GIVEREF(__pyx_tuple__810);
  __pyx_tuple__811 = PyTuple_Pack(3, Py_None, Py_None, Py_None); if (unlikely(!__pyx_tuple__811)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__811);
  __Pyx_GIVEREF(__pyx_tuple__811);

  /* "pymola/generated/ModelicaParser.pyx":777
 *     def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
 *         super().__init__(input, output)
 *         self.checkVersion("4.7")             # <<<<<<<<<<<<<<
 *         self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
 *         self._predicates = None
 */
  __pyx_tuple__812 = PyTuple_Pack(1, __pyx_kp_u_4_7); if (unlikely(!__pyx_tuple__812)) __PYX_ERR(0, 777, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__812);
  __Pyx_GIVEREF(__pyx_tuple__812);

  /* "pymola/generated/ModelicaParser.pyx":832
 *             self.state = 151
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.WITHIN:
 *                 self.state = 146
 */
  __pyx_tuple__813 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__813)) __PYX_ERR(0, 832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__813);
  __Pyx_GIVEREF(__pyx_tuple__813);

  /* "pymola/generated/ModelicaParser.pyx":838
 *                 self.state = 148
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if _la==ModelicaParser.IDENT:
 *                     self.state = 147
 */
  __pyx_tuple__814 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__814)) __PYX_ERR(0, 838, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__814);
  __Pyx_GIVEREF(__pyx_tuple__814);

  /* "pymola/generated/ModelicaParser.pyx":850
 *             self.state = 156
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__1) | (1 << ModelicaParser.T__2) | (1 << ModelicaParser.T__3) | (1 << ModelicaParser.T__4) | (1 << ModelicaParser.T__5) | (1 << ModelicaParser.T__6) | (1 << ModelicaParser.T__7) | (1 << ModelicaParser.T__8) | (1 << ModelicaParser.T__9) | (1 << ModelicaParser.T__10) | (1 << ModelicaParser.T__11) | (1 << ModelicaParser.T__12))) != 0) or ((((_la - 80)) & ~0x3f) == 0 and ((1 << (_la - 80)) & ((1 << (ModelicaParser.PARTIAL - 80)) | (1 << (ModelicaParser.FINAL - 80)) | (1 << (ModelicaParser.ENCAPSULATED - 80)))) != 0):
 *                 self.state = 153
 */
  __pyx_tuple__815 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__815)) __PYX_ERR(0, 850, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__815);
  __Pyx_GIVEREF(__pyx_tuple__815);

  /* "pymola/generated/ModelicaParser.pyx":856
 *                 self.state = 158
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *         except RecognitionException as re:
 */
  __pyx_tuple__816 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__816)) __PYX_ERR(0, 856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__816);
  __Pyx_GIVEREF(__pyx_tuple__816);

  /* "pymola/generated/ModelicaParser.pyx":908
 *             self.state = 160
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.FINAL:
 *                 self.state = 159
 */
  __pyx_tuple__817 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__817)) __PYX_ERR(0, 908, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__817);
  __Pyx_GIVEREF(__pyx_tuple__817);

  /* "pymola/generated/ModelicaParser.pyx":972
 *             self.state = 166
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.ENCAPSULATED:
 *                 self.state = 165
 */
  __pyx_tuple__818 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__818)) __PYX_ERR(0, 972, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__818);
  __Pyx_GIVEREF(__pyx_tuple__818);

  /* "pymola/generated/ModelicaParser.pyx":1032
 *             self.state = 172
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.PARTIAL:
 *                 self.state = 171
 */
  __pyx_tuple__819 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__819)) __PYX_ERR(0, 1032, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__819);
  __Pyx_GIVEREF(__pyx_tuple__819);

  /* "pymola/generated/ModelicaParser.pyx":1100
 *                 self.state = 179
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if _la==ModelicaParser.T__3:
 *                     self.state = 178
 */
  __pyx_tuple__820 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__820)) __PYX_ERR(0, 1100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__820);
  __Pyx_GIVEREF(__pyx_tuple__820);

  /* "pymola/generated/ModelicaParser.pyx":1120
 *                 self.state = 184
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if _la==ModelicaParser.T__6:
 *                     self.state = 183
 */
  __pyx_tuple__821 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__821)) __PYX_ERR(0, 1120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__821);
  __Pyx_GIVEREF(__pyx_tuple__821);

  /* "pymola/generated/ModelicaParser.pyx":1146
 *                 self.state = 190
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if _la==ModelicaParser.T__10 or _la==ModelicaParser.T__11:
 *                     self.state = 189
 */
  __pyx_tuple__822 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__822)) __PYX_ERR(0, 1146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__822);
  __Pyx_GIVEREF(__pyx_tuple__822);

  /* "pymola/generated/ModelicaParser.pyx":1149
 *                 if _la==ModelicaParser.T__10 or _la==ModelicaParser.T__11:
 *                     self.state = 189
 *                     _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                     if not(_la==ModelicaParser.T__10 or _la==ModelicaParser.T__11):
 *                         self._errHandler.recoverInline(self)
 */
  __pyx_tuple__823 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__823)) __PYX_ERR(0, 1149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__823);
  __Pyx_GIVEREF(__pyx_tuple__823);

  /* "pymola/generated/ModelicaParser.pyx":1159
 *                 self.state = 193
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if _la==ModelicaParser.T__3:
 *                     self.state = 192
 */
  __pyx_tuple__824 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__824)) __PYX_ERR(0, 1159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__824);
  __Pyx_GIVEREF(__pyx_tuple__824);

  /* "pymola/generated/ModelicaParser.pyx":1409
 *                 self.state = 210
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if _la==ModelicaParser.T__75:
 *                     self.state = 209
 */
  __pyx_tuple__825 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__825)) __PYX_ERR(0, 1409, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__825);
  __Pyx_GIVEREF(__pyx_tuple__825);

  /* "pymola/generated/ModelicaParser.pyx":1417
 *                 self.state = 213
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if _la==ModelicaParser.T__16:
 *                     self.state = 212
 */
  __pyx_tuple__826 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__826)) __PYX_ERR(0, 1417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__826);
  __Pyx_GIVEREF(__pyx_tuple__826);

  /* "pymola/generated/ModelicaParser.pyx":1440
 *                 self.state = 225
 *                 self._errHandler.sync(self)
 *                 token = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if token in [ModelicaParser.T__18, ModelicaParser.IDENT]:
 *                     self.state = 222
 */
  __pyx_tuple__827 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__827)) __PYX_ERR(0, 1440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__827);
  __Pyx_GIVEREF(__pyx_tuple__827);

  /* "pymola/generated/ModelicaParser.pyx":1444
 *                     self.state = 222
 *                     self._errHandler.sync(self)
 *                     _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                     if _la==ModelicaParser.IDENT:
 *                         self.state = 221
 */
  __pyx_tuple__828 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__828)) __PYX_ERR(0, 1444, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__828);
  __Pyx_GIVEREF(__pyx_tuple__828);

  /* "pymola/generated/ModelicaParser.pyx":1483
 *                 self.state = 240
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 while _la==ModelicaParser.T__20:
 *                     self.state = 236
 */
  __pyx_tuple__829 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__829)) __PYX_ERR(0, 1483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__829);
  __Pyx_GIVEREF(__pyx_tuple__829);

  /* "pymola/generated/ModelicaParser.pyx":1491
 *                     self.state = 242
 *                     self._errHandler.sync(self)
 *                     _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *                 self.state = 243
 */
  __pyx_tuple__830 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__830)) __PYX_ERR(0, 1491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__830);
  __Pyx_GIVEREF(__pyx_tuple__830);

  /* "pymola/generated/ModelicaParser.pyx":1508
 *                 self.state = 249
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if _la==ModelicaParser.T__16:
 *                     self.state = 248
 */
  __pyx_tuple__831 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__831)) __PYX_ERR(0, 1508, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__831);
  __Pyx_GIVEREF(__pyx_tuple__831);

  /* "pymola/generated/ModelicaParser.pyx":1623
 *             self.state = 265
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             while _la==ModelicaParser.T__20:
 *                 self.state = 261
 */
  __pyx_tuple__832 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__832)) __PYX_ERR(0, 1623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__832);
  __Pyx_GIVEREF(__pyx_tuple__832);

  /* "pymola/generated/ModelicaParser.pyx":1631
 *                 self.state = 267
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *         except RecognitionException as re:
 */
  __pyx_tuple__833 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__833)) __PYX_ERR(0, 1631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__833);
  __Pyx_GIVEREF(__pyx_tuple__833);

  /* "pymola/generated/ModelicaParser.pyx":1770
 *             self.state = 280
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__22) | (1 << ModelicaParser.T__23) | (1 << ModelicaParser.T__41) | (1 << ModelicaParser.T__42))) != 0) or _la==ModelicaParser.INITIAL:
 *                 self.state = 278
 */
  __pyx_tuple__834 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__834)) __PYX_ERR(0, 1770, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__834);
  __Pyx_GIVEREF(__pyx_tuple__834);

  /* "pymola/generated/ModelicaParser.pyx":1802
 *                 self.state = 282
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *             self.state = 294
 */
  __pyx_tuple__835 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__835)) __PYX_ERR(0, 1802, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__835);
  __Pyx_GIVEREF(__pyx_tuple__835);

  /* "pymola/generated/ModelicaParser.pyx":1806
 *             self.state = 294
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.T__24:
 *                 self.state = 283
 */
  __pyx_tuple__836 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__836)) __PYX_ERR(0, 1806, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__836);
  __Pyx_GIVEREF(__pyx_tuple__836);

  /* "pymola/generated/ModelicaParser.pyx":1812
 *                 self.state = 285
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if _la==ModelicaParser.STRING:
 *                     self.state = 284
 */
  __pyx_tuple__837 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__837)) __PYX_ERR(0, 1812, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__837);
  __Pyx_GIVEREF(__pyx_tuple__837);

  /* "pymola/generated/ModelicaParser.pyx":1820
 *                 self.state = 288
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if _la==ModelicaParser.IDENT:
 *                     self.state = 287
 */
  __pyx_tuple__838 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__838)) __PYX_ERR(0, 1820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__838);
  __Pyx_GIVEREF(__pyx_tuple__838);

  /* "pymola/generated/ModelicaParser.pyx":1828
 *                 self.state = 291
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if _la==ModelicaParser.T__77:
 *                     self.state = 290
 */
  __pyx_tuple__839 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__839)) __PYX_ERR(0, 1828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__839);
  __Pyx_GIVEREF(__pyx_tuple__839);

  /* "pymola/generated/ModelicaParser.pyx":1840
 *             self.state = 299
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.T__77:
 *                 self.state = 296
 */
  __pyx_tuple__840 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__840)) __PYX_ERR(0, 1840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__840);
  __Pyx_GIVEREF(__pyx_tuple__840);

  /* "pymola/generated/ModelicaParser.pyx":1961
 *             self.state = 311
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__13) | (1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__19) | (1 << ModelicaParser.T__29) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__55))) != 0) or ((((_la - 71)) & ~0x3f) == 0 and ((1 << (_la - 71)) & ((1 << (ModelicaParser.T__70 - 71)) | (1 << (ModelicaParser.T__73 - 71)) | (1 << (ModelicaParser.T__74 - 71)) | (1 << (ModelicaParser.T__75 - 71)) | (1 << (ModelicaParser.INITIAL - 71)) | (1 << (ModelicaParser.IDENT - 71)) | (1 << (ModelicaParser.STRING - 71)) | (1 << (ModelicaParser.UNSIGNED_NUMBER - 71)))) != 0):
 *                 self.state = 310
 */
  __pyx_tuple__841 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__841)) __PYX_ERR(0, 1961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__841);
  __Pyx_GIVEREF(__pyx_tuple__841);

  /* "pymola/generated/ModelicaParser.pyx":2019
 *             self.state = 320
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__1) | (1 << ModelicaParser.T__2) | (1 << ModelicaParser.T__3) | (1 << ModelicaParser.T__4) | (1 << ModelicaParser.T__5) | (1 << ModelicaParser.T__6) | (1 << ModelicaParser.T__7) | (1 << ModelicaParser.T__8) | (1 << ModelicaParser.T__9) | (1 << ModelicaParser.T__10) | (1 << ModelicaParser.T__11) | (1 << ModelicaParser.T__12) | (1 << ModelicaParser.T__21) | (1 << ModelicaParser.T__25) | (1 << ModelicaParser.T__26) | (1 << ModelicaParser.T__32) | (1 << ModelicaParser.T__33) | (1 << ModelicaParser.T__34) | (1 << ModelicaParser.T__35) | (1 << ModelicaParser.T__36) | (1 << ModelicaParser.T__37) | (1 << ModelicaParser.T__38))) != 0) or ((((_la - 80)) & ~0x3f) == 0 and ((1 << (_la - 80)) & ((1 << (ModelicaParser.PARTIAL - 80)) | (1 << (ModelicaParser.FINAL - 80)) | (1 << (ModelicaParser.ENCAPSULATED - 80)) | (1 << (ModelicaParser.REDECLARE - 80)) | (1 << (ModelicaParser.INNER - 80)) | (1 << (ModelicaParser.OUTER - 80)) | (1 << (ModelicaParser.IDENT - 80)))) != 0):
 *                 self.state = 315
 */
  __pyx_tuple__842 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__842)) __PYX_ERR(0, 2019, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__842);
  __Pyx_GIVEREF(__pyx_tuple__842);

  /* "pymola/generated/ModelicaParser.pyx":2027
 *                 self.state = 322
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *         except RecognitionException as re:
 */
  __pyx_tuple__843 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__843)) __PYX_ERR(0, 2027, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__843);
  __Pyx_GIVEREF(__pyx_tuple__843);

  /* "pymola/generated/ModelicaParser.pyx":2177
 *             self.state = 330
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.REDECLARE:
 *                 self.state = 329
 */
  __pyx_tuple__844 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__844)) __PYX_ERR(0, 2177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__844);
  __Pyx_GIVEREF(__pyx_tuple__844);

  /* "pymola/generated/ModelicaParser.pyx":2185
 *             self.state = 333
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.FINAL:
 *                 self.state = 332
 */
  __pyx_tuple__845 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__845)) __PYX_ERR(0, 2185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__845);
  __Pyx_GIVEREF(__pyx_tuple__845);

  /* "pymola/generated/ModelicaParser.pyx":2193
 *             self.state = 336
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.INNER:
 *                 self.state = 335
 */
  __pyx_tuple__846 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__846)) __PYX_ERR(0, 2193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__846);
  __Pyx_GIVEREF(__pyx_tuple__846);

  /* "pymola/generated/ModelicaParser.pyx":2201
 *             self.state = 339
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.OUTER:
 *                 self.state = 338
 */
  __pyx_tuple__847 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__847)) __PYX_ERR(0, 2201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__847);
  __Pyx_GIVEREF(__pyx_tuple__847);

  /* "pymola/generated/ModelicaParser.pyx":2209
 *             self.state = 343
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if token in [ModelicaParser.T__1, ModelicaParser.T__2, ModelicaParser.T__3, ModelicaParser.T__4, ModelicaParser.T__5, ModelicaParser.T__6, ModelicaParser.T__7, ModelicaParser.T__8, ModelicaParser.T__9, ModelicaParser.T__10, ModelicaParser.T__11, ModelicaParser.T__12, ModelicaParser.PARTIAL, ModelicaParser.ENCAPSULATED]:
 *                 self.state = 341
 */
  __pyx_tuple__848 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__848)) __PYX_ERR(0, 2209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__848);
  __Pyx_GIVEREF(__pyx_tuple__848);

  /* "pymola/generated/ModelicaParser.pyx":2294
 *             self.state = 346
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.REDECLARE:
 *                 self.state = 345
 */
  __pyx_tuple__849 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__849)) __PYX_ERR(0, 2294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__849);
  __Pyx_GIVEREF(__pyx_tuple__849);

  /* "pymola/generated/ModelicaParser.pyx":2302
 *             self.state = 349
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.FINAL:
 *                 self.state = 348
 */
  __pyx_tuple__850 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__850)) __PYX_ERR(0, 2302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__850);
  __Pyx_GIVEREF(__pyx_tuple__850);

  /* "pymola/generated/ModelicaParser.pyx":2310
 *             self.state = 352
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.INNER:
 *                 self.state = 351
 */
  __pyx_tuple__851 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__851)) __PYX_ERR(0, 2310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__851);
  __Pyx_GIVEREF(__pyx_tuple__851);

  /* "pymola/generated/ModelicaParser.pyx":2318
 *             self.state = 355
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.OUTER:
 *                 self.state = 354
 */
  __pyx_tuple__852 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__852)) __PYX_ERR(0, 2318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__852);
  __Pyx_GIVEREF(__pyx_tuple__852);

  /* "pymola/generated/ModelicaParser.pyx":2328
 *             self.state = 360
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if token in [ModelicaParser.T__1, ModelicaParser.T__2, ModelicaParser.T__3, ModelicaParser.T__4, ModelicaParser.T__5, ModelicaParser.T__6, ModelicaParser.T__7, ModelicaParser.T__8, ModelicaParser.T__9, ModelicaParser.T__10, ModelicaParser.T__11, ModelicaParser.T__12, ModelicaParser.PARTIAL, ModelicaParser.ENCAPSULATED]:
 *                 self.state = 358
 */
  __pyx_tuple__853 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__853)) __PYX_ERR(0, 2328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__853);
  __Pyx_GIVEREF(__pyx_tuple__853);

  /* "pymola/generated/ModelicaParser.pyx":2342
 *             self.state = 365
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.T__31:
 *                 self.state = 362
 */
  __pyx_tuple__854 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__854)) __PYX_ERR(0, 2342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__854);
  __Pyx_GIVEREF(__pyx_tuple__854);

  /* "pymola/generated/ModelicaParser.pyx":2425
 *                 self.state = 380
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if _la==ModelicaParser.T__27:
 *                     self.state = 372
 */
  __pyx_tuple__855 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__855)) __PYX_ERR(0, 2425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__855);
  __Pyx_GIVEREF(__pyx_tuple__855);

  /* "pymola/generated/ModelicaParser.pyx":2431
 *                     self.state = 378
 *                     self._errHandler.sync(self)
 *                     token = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                     if token in [ModelicaParser.T__28]:
 *                         self.state = 373
 */
  __pyx_tuple__856 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__856)) __PYX_ERR(0, 2431, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__856);
  __Pyx_GIVEREF(__pyx_tuple__856);

  /* "pymola/generated/ModelicaParser.pyx":2578
 *             self.state = 397
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.T__16:
 *                 self.state = 396
 */
  __pyx_tuple__857 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__857)) __PYX_ERR(0, 2578, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__857);
  __Pyx_GIVEREF(__pyx_tuple__857);

  /* "pymola/generated/ModelicaParser.pyx":2586
 *             self.state = 400
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.T__77:
 *                 self.state = 399
 */
  __pyx_tuple__858 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__858)) __PYX_ERR(0, 2586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__858);
  __Pyx_GIVEREF(__pyx_tuple__858);

  /* "pymola/generated/ModelicaParser.pyx":2647
 *             self.state = 405
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.T__16:
 *                 self.state = 404
 */
  __pyx_tuple__859 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__859)) __PYX_ERR(0, 2647, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__859);
  __Pyx_GIVEREF(__pyx_tuple__859);

  /* "pymola/generated/ModelicaParser.pyx":2716
 *             self.state = 410
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.T__75:
 *                 self.state = 409
 */
  __pyx_tuple__860 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__860)) __PYX_ERR(0, 2716, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__860);
  __Pyx_GIVEREF(__pyx_tuple__860);

  /* "pymola/generated/ModelicaParser.pyx":2768
 *             self.state = 415
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.T__32 or _la==ModelicaParser.T__33:
 *                 self.state = 414
 */
  __pyx_tuple__861 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__861)) __PYX_ERR(0, 2768, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__861);
  __Pyx_GIVEREF(__pyx_tuple__861);

  /* "pymola/generated/ModelicaParser.pyx":2771
 *             if _la==ModelicaParser.T__32 or _la==ModelicaParser.T__33:
 *                 self.state = 414
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if not(_la==ModelicaParser.T__32 or _la==ModelicaParser.T__33):
 *                     self._errHandler.recoverInline(self)
 */
  __pyx_tuple__862 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__862)) __PYX_ERR(0, 2771, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__862);
  __Pyx_GIVEREF(__pyx_tuple__862);

  /* "pymola/generated/ModelicaParser.pyx":2781
 *             self.state = 418
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__34) | (1 << ModelicaParser.T__35) | (1 << ModelicaParser.T__36))) != 0):
 *                 self.state = 417
 */
  __pyx_tuple__863 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__863)) __PYX_ERR(0, 2781, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__863);
  __Pyx_GIVEREF(__pyx_tuple__863);

  /* "pymola/generated/ModelicaParser.pyx":2784
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__34) | (1 << ModelicaParser.T__35) | (1 << ModelicaParser.T__36))) != 0):
 *                 self.state = 417
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__34) | (1 << ModelicaParser.T__35) | (1 << ModelicaParser.T__36))) != 0)):
 *                     self._errHandler.recoverInline(self)
 */
  __pyx_tuple__864 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__864)) __PYX_ERR(0, 2784, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__864);
  __Pyx_GIVEREF(__pyx_tuple__864);

  /* "pymola/generated/ModelicaParser.pyx":2794
 *             self.state = 421
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.T__37 or _la==ModelicaParser.T__38:
 *                 self.state = 420
 */
  __pyx_tuple__865 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__865)) __PYX_ERR(0, 2794, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__865);
  __Pyx_GIVEREF(__pyx_tuple__865);

  /* "pymola/generated/ModelicaParser.pyx":2797
 *             if _la==ModelicaParser.T__37 or _la==ModelicaParser.T__38:
 *                 self.state = 420
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if not(_la==ModelicaParser.T__37 or _la==ModelicaParser.T__38):
 *                     self._errHandler.recoverInline(self)
 */
  __pyx_tuple__866 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__866)) __PYX_ERR(0, 2797, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__866);
  __Pyx_GIVEREF(__pyx_tuple__866);

  /* "pymola/generated/ModelicaParser.pyx":2903
 *             self.state = 430
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             while _la==ModelicaParser.T__20:
 *                 self.state = 426
 */
  __pyx_tuple__867 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__867)) __PYX_ERR(0, 2903, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__867);
  __Pyx_GIVEREF(__pyx_tuple__867);

  /* "pymola/generated/ModelicaParser.pyx":2911
 *                 self.state = 432
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *         except RecognitionException as re:
 */
  __pyx_tuple__868 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__868)) __PYX_ERR(0, 2911, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__868);
  __Pyx_GIVEREF(__pyx_tuple__868);

  /* "pymola/generated/ModelicaParser.pyx":2970
 *             self.state = 435
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.T__39:
 *                 self.state = 434
 */
  __pyx_tuple__869 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__869)) __PYX_ERR(0, 2970, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__869);
  __Pyx_GIVEREF(__pyx_tuple__869);

  /* "pymola/generated/ModelicaParser.pyx":3082
 *             self.state = 444
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.T__75:
 *                 self.state = 443
 */
  __pyx_tuple__870 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__870)) __PYX_ERR(0, 3082, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__870);
  __Pyx_GIVEREF(__pyx_tuple__870);

  /* "pymola/generated/ModelicaParser.pyx":3090
 *             self.state = 447
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__14) | (1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__40))) != 0):
 *                 self.state = 446
 */
  __pyx_tuple__871 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__871)) __PYX_ERR(0, 3090, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__871);
  __Pyx_GIVEREF(__pyx_tuple__871);

  /* "pymola/generated/ModelicaParser.pyx":3207
 *             self.state = 458
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if token in [ModelicaParser.T__16]:
 *                 localctx = ModelicaParser.Modification_classContext(self, localctx)
 */
  __pyx_tuple__872 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__872)) __PYX_ERR(0, 3207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__872);
  __Pyx_GIVEREF(__pyx_tuple__872);

  /* "pymola/generated/ModelicaParser.pyx":3215
 *                 self.state = 452
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if _la==ModelicaParser.T__14:
 *                     self.state = 450
 */
  __pyx_tuple__873 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__873)) __PYX_ERR(0, 3215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__873);
  __Pyx_GIVEREF(__pyx_tuple__873);

  /* "pymola/generated/ModelicaParser.pyx":3292
 *             self.state = 462
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if ((((_la - 26)) & ~0x3f) == 0 and ((1 << (_la - 26)) & ((1 << (ModelicaParser.T__25 - 26)) | (1 << (ModelicaParser.EACH - 26)) | (1 << (ModelicaParser.FINAL - 26)) | (1 << (ModelicaParser.REDECLARE - 26)) | (1 << (ModelicaParser.IDENT - 26)))) != 0):
 *                 self.state = 461
 */
  __pyx_tuple__874 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__874)) __PYX_ERR(0, 3292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__874);
  __Pyx_GIVEREF(__pyx_tuple__874);

  /* "pymola/generated/ModelicaParser.pyx":3352
 *             self.state = 471
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             while _la==ModelicaParser.T__20:
 *                 self.state = 467
 */
  __pyx_tuple__875 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__875)) __PYX_ERR(0, 3352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__875);
  __Pyx_GIVEREF(__pyx_tuple__875);

  /* "pymola/generated/ModelicaParser.pyx":3360
 *                 self.state = 473
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *         except RecognitionException as re:
 */
  __pyx_tuple__876 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__876)) __PYX_ERR(0, 3360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__876);
  __Pyx_GIVEREF(__pyx_tuple__876);

  /* "pymola/generated/ModelicaParser.pyx":3411
 *             self.state = 476
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if token in [ModelicaParser.T__25, ModelicaParser.EACH, ModelicaParser.FINAL, ModelicaParser.IDENT]:
 *                 self.enterOuterAlt(localctx, 1)
 */
  __pyx_tuple__877 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__877)) __PYX_ERR(0, 3411, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__877);
  __Pyx_GIVEREF(__pyx_tuple__877);

  /* "pymola/generated/ModelicaParser.pyx":3482
 *             self.state = 479
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.EACH:
 *                 self.state = 478
 */
  __pyx_tuple__878 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__878)) __PYX_ERR(0, 3482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__878);
  __Pyx_GIVEREF(__pyx_tuple__878);

  /* "pymola/generated/ModelicaParser.pyx":3490
 *             self.state = 482
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.FINAL:
 *                 self.state = 481
 */
  __pyx_tuple__879 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__879)) __PYX_ERR(0, 3490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__879);
  __Pyx_GIVEREF(__pyx_tuple__879);

  /* "pymola/generated/ModelicaParser.pyx":3498
 *             self.state = 486
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if token in [ModelicaParser.IDENT]:
 *                 self.state = 484
 */
  __pyx_tuple__880 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__880)) __PYX_ERR(0, 3498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__880);
  __Pyx_GIVEREF(__pyx_tuple__880);

  /* "pymola/generated/ModelicaParser.pyx":3567
 *             self.state = 490
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__14) | (1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__40))) != 0):
 *                 self.state = 489
 */
  __pyx_tuple__881 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__881)) __PYX_ERR(0, 3567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__881);
  __Pyx_GIVEREF(__pyx_tuple__881);

  /* "pymola/generated/ModelicaParser.pyx":3641
 *             self.state = 496
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.EACH:
 *                 self.state = 495
 */
  __pyx_tuple__882 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__882)) __PYX_ERR(0, 3641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__882);
  __Pyx_GIVEREF(__pyx_tuple__882);

  /* "pymola/generated/ModelicaParser.pyx":3649
 *             self.state = 499
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.FINAL:
 *                 self.state = 498
 */
  __pyx_tuple__883 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__883)) __PYX_ERR(0, 3649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__883);
  __Pyx_GIVEREF(__pyx_tuple__883);

  /* "pymola/generated/ModelicaParser.pyx":3657
 *             self.state = 506
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if token in [ModelicaParser.T__1, ModelicaParser.T__2, ModelicaParser.T__3, ModelicaParser.T__4, ModelicaParser.T__5, ModelicaParser.T__6, ModelicaParser.T__7, ModelicaParser.T__8, ModelicaParser.T__9, ModelicaParser.T__10, ModelicaParser.T__11, ModelicaParser.T__12, ModelicaParser.T__32, ModelicaParser.T__33, ModelicaParser.T__34, ModelicaParser.T__35, ModelicaParser.T__36, ModelicaParser.T__37, ModelicaParser.T__38, ModelicaParser.PARTIAL, ModelicaParser.IDENT]:
 *                 self.state = 503
 */
  __pyx_tuple__884 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__884)) __PYX_ERR(0, 3657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__884);
  __Pyx_GIVEREF(__pyx_tuple__884);

  /* "pymola/generated/ModelicaParser.pyx":3661
 *                 self.state = 503
 *                 self._errHandler.sync(self)
 *                 token = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if token in [ModelicaParser.T__1, ModelicaParser.T__2, ModelicaParser.T__3, ModelicaParser.T__4, ModelicaParser.T__5, ModelicaParser.T__6, ModelicaParser.T__7, ModelicaParser.T__8, ModelicaParser.T__9, ModelicaParser.T__10, ModelicaParser.T__11, ModelicaParser.T__12, ModelicaParser.PARTIAL]:
 *                     self.state = 501
 */
  __pyx_tuple__885 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__885)) __PYX_ERR(0, 3661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__885);
  __Pyx_GIVEREF(__pyx_tuple__885);

  /* "pymola/generated/ModelicaParser.pyx":3738
 *             self.state = 511
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if token in [ModelicaParser.T__1, ModelicaParser.T__2, ModelicaParser.T__3, ModelicaParser.T__4, ModelicaParser.T__5, ModelicaParser.T__6, ModelicaParser.T__7, ModelicaParser.T__8, ModelicaParser.T__9, ModelicaParser.T__10, ModelicaParser.T__11, ModelicaParser.T__12, ModelicaParser.PARTIAL]:
 *                 self.state = 509
 */
  __pyx_tuple__886 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__886)) __PYX_ERR(0, 3738, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__886);
  __Pyx_GIVEREF(__pyx_tuple__886);

  /* "pymola/generated/ModelicaParser.pyx":3752
 *             self.state = 514
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.T__31:
 *                 self.state = 513
 */
  __pyx_tuple__887 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__887)) __PYX_ERR(0, 3752, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__887);
  __Pyx_GIVEREF(__pyx_tuple__887);

  /* "pymola/generated/ModelicaParser.pyx":3948
 *             self.state = 546
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if token in [ModelicaParser.T__32, ModelicaParser.T__33, ModelicaParser.T__34, ModelicaParser.T__35, ModelicaParser.T__36, ModelicaParser.T__37, ModelicaParser.T__38, ModelicaParser.IDENT]:
 *                 self.state = 526
 */
  __pyx_tuple__888 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__888)) __PYX_ERR(0, 3948, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__888);
  __Pyx_GIVEREF(__pyx_tuple__888);

  /* "pymola/generated/ModelicaParser.pyx":3956
 *                 self.state = 529
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if _la==ModelicaParser.T__75:
 *                     self.state = 528
 */
  __pyx_tuple__889 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__889)) __PYX_ERR(0, 3956, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__889);
  __Pyx_GIVEREF(__pyx_tuple__889);

  /* "pymola/generated/ModelicaParser.pyx":3964
 *                 self.state = 532
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if _la==ModelicaParser.T__16:
 *                     self.state = 531
 */
  __pyx_tuple__890 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__890)) __PYX_ERR(0, 3964, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__890);
  __Pyx_GIVEREF(__pyx_tuple__890);

  /* "pymola/generated/ModelicaParser.pyx":3980
 *                 self.state = 542
 *                 self._errHandler.sync(self)
 *                 token = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if token in [ModelicaParser.T__18, ModelicaParser.IDENT]:
 *                     self.state = 539
 */
  __pyx_tuple__891 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__891)) __PYX_ERR(0, 3980, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__891);
  __Pyx_GIVEREF(__pyx_tuple__891);

  /* "pymola/generated/ModelicaParser.pyx":3984
 *                     self.state = 539
 *                     self._errHandler.sync(self)
 *                     _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                     if _la==ModelicaParser.IDENT:
 *                         self.state = 538
 */
  __pyx_tuple__892 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__892)) __PYX_ERR(0, 3984, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__892);
  __Pyx_GIVEREF(__pyx_tuple__892);

  /* "pymola/generated/ModelicaParser.pyx":4059
 *             self.state = 549
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.INITIAL:
 *                 self.state = 548
 */
  __pyx_tuple__893 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__893)) __PYX_ERR(0, 4059, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__893);
  __Pyx_GIVEREF(__pyx_tuple__893);

  /* "pymola/generated/ModelicaParser.pyx":4133
 *             self.state = 561
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.INITIAL:
 *                 self.state = 560
 */
  __pyx_tuple__894 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__894)) __PYX_ERR(0, 4133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__894);
  __Pyx_GIVEREF(__pyx_tuple__894);

  /* "pymola/generated/ModelicaParser.pyx":4143
 *             self.state = 569
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:
 *                 self.state = 564
 */
  __pyx_tuple__895 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__895)) __PYX_ERR(0, 4143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__895);
  __Pyx_GIVEREF(__pyx_tuple__895);

  /* "pymola/generated/ModelicaParser.pyx":4151
 *                 self.state = 571
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *         except RecognitionException as re:
 */
  __pyx_tuple__896 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__896)) __PYX_ERR(0, 4151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__896);
  __Pyx_GIVEREF(__pyx_tuple__896);

  /* "pymola/generated/ModelicaParser.pyx":4682
 *             self.state = 614
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if token in [ModelicaParser.IDENT]:
 *                 localctx = ModelicaParser.Statement_component_referenceContext(self, localctx)
 */
  __pyx_tuple__897 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__897)) __PYX_ERR(0, 4682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__897);
  __Pyx_GIVEREF(__pyx_tuple__897);

  /* "pymola/generated/ModelicaParser.pyx":4690
 *                 self.state = 592
 *                 self._errHandler.sync(self)
 *                 token = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if token in [ModelicaParser.T__40]:
 *                     self.state = 589
 */
  __pyx_tuple__898 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__898)) __PYX_ERR(0, 4690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__898);
  __Pyx_GIVEREF(__pyx_tuple__898);

  /* "pymola/generated/ModelicaParser.pyx":4714
 *                 self.state = 600
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 while _la==ModelicaParser.T__20:
 *                     self.state = 596
 */
  __pyx_tuple__899 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__899)) __PYX_ERR(0, 4714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__899);
  __Pyx_GIVEREF(__pyx_tuple__899);

  /* "pymola/generated/ModelicaParser.pyx":4722
 *                     self.state = 602
 *                     self._errHandler.sync(self)
 *                     _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *                 self.state = 603
 */
  __pyx_tuple__900 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__900)) __PYX_ERR(0, 4722, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__900);
  __Pyx_GIVEREF(__pyx_tuple__900);

  /* "pymola/generated/ModelicaParser.pyx":4900
 *             self.state = 643
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             while _la==ModelicaParser.T__46:
 *                 self.state = 630
 */
  __pyx_tuple__901 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__901)) __PYX_ERR(0, 4900, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__901);
  __Pyx_GIVEREF(__pyx_tuple__901);

  /* "pymola/generated/ModelicaParser.pyx":4923
 *                 self.state = 645
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *             self.state = 655
 */
  __pyx_tuple__902 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__902)) __PYX_ERR(0, 4923, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__902);
  __Pyx_GIVEREF(__pyx_tuple__902);

  /* "pymola/generated/ModelicaParser.pyx":4927
 *             self.state = 655
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.T__47:
 *                 self.state = 646
 */
  __pyx_tuple__903 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__903)) __PYX_ERR(0, 4927, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__903);
  __Pyx_GIVEREF(__pyx_tuple__903);

  /* "pymola/generated/ModelicaParser.pyx":5013
 *             self.state = 668
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:
 *                 self.state = 663
 */
  __pyx_tuple__904 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__904)) __PYX_ERR(0, 5013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__904);
  __Pyx_GIVEREF(__pyx_tuple__904);

  /* "pymola/generated/ModelicaParser.pyx":5021
 *                 self.state = 670
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *             self.state = 684
 */
  __pyx_tuple__905 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__905)) __PYX_ERR(0, 5021, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__905);
  __Pyx_GIVEREF(__pyx_tuple__905);

  /* "pymola/generated/ModelicaParser.pyx":5025
 *             self.state = 684
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             while _la==ModelicaParser.T__46:
 *                 self.state = 671
 */
  __pyx_tuple__906 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__906)) __PYX_ERR(0, 5025, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__906);
  __Pyx_GIVEREF(__pyx_tuple__906);

  /* "pymola/generated/ModelicaParser.pyx":5035
 *                 self.state = 679
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:
 *                     self.state = 674
 */
  __pyx_tuple__907 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__907)) __PYX_ERR(0, 5035, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__907);
  __Pyx_GIVEREF(__pyx_tuple__907);

  /* "pymola/generated/ModelicaParser.pyx":5043
 *                     self.state = 681
 *                     self._errHandler.sync(self)
 *                     _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *                 self.state = 686
 */
  __pyx_tuple__908 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__908)) __PYX_ERR(0, 5043, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__908);
  __Pyx_GIVEREF(__pyx_tuple__908);

  /* "pymola/generated/ModelicaParser.pyx":5047
 *                 self.state = 686
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *             self.state = 696
 */
  __pyx_tuple__909 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__909)) __PYX_ERR(0, 5047, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__909);
  __Pyx_GIVEREF(__pyx_tuple__909);

  /* "pymola/generated/ModelicaParser.pyx":5051
 *             self.state = 696
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.T__47:
 *                 self.state = 687
 */
  __pyx_tuple__910 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__910)) __PYX_ERR(0, 5051, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__910);
  __Pyx_GIVEREF(__pyx_tuple__910);

  /* "pymola/generated/ModelicaParser.pyx":5057
 *                 self.state = 693
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:
 *                     self.state = 688
 */
  __pyx_tuple__911 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__911)) __PYX_ERR(0, 5057, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__911);
  __Pyx_GIVEREF(__pyx_tuple__911);

  /* "pymola/generated/ModelicaParser.pyx":5065
 *                     self.state = 695
 *                     self._errHandler.sync(self)
 *                     _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__912 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__912)) __PYX_ERR(0, 5065, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__912);
  __Pyx_GIVEREF(__pyx_tuple__912);

  /* "pymola/generated/ModelicaParser.pyx":5207
 *             self.state = 723
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:
 *                 self.state = 718
 */
  __pyx_tuple__913 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__913)) __PYX_ERR(0, 5207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__913);
  __Pyx_GIVEREF(__pyx_tuple__913);

  /* "pymola/generated/ModelicaParser.pyx":5215
 *                 self.state = 725
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *             self.state = 726
 */
  __pyx_tuple__914 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__914)) __PYX_ERR(0, 5215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__914);
  __Pyx_GIVEREF(__pyx_tuple__914);

  /* "pymola/generated/ModelicaParser.pyx":5335
 *             self.state = 740
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.T__50:
 *                 self.state = 738
 */
  __pyx_tuple__915 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__915)) __PYX_ERR(0, 5335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__915);
  __Pyx_GIVEREF(__pyx_tuple__915);

  /* "pymola/generated/ModelicaParser.pyx":5403
 *             self.state = 750
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:
 *                 self.state = 745
 */
  __pyx_tuple__916 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__916)) __PYX_ERR(0, 5403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__916);
  __Pyx_GIVEREF(__pyx_tuple__916);

  /* "pymola/generated/ModelicaParser.pyx":5411
 *                 self.state = 752
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *             self.state = 753
 */
  __pyx_tuple__917 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__917)) __PYX_ERR(0, 5411, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__917);
  __Pyx_GIVEREF(__pyx_tuple__917);

  /* "pymola/generated/ModelicaParser.pyx":5493
 *             self.state = 780
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             while _la==ModelicaParser.T__53:
 *                 self.state = 767
 */
  __pyx_tuple__918 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__918)) __PYX_ERR(0, 5493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__918);
  __Pyx_GIVEREF(__pyx_tuple__918);

  /* "pymola/generated/ModelicaParser.pyx":5516
 *                 self.state = 782
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *             self.state = 783
 */
  __pyx_tuple__919 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__919)) __PYX_ERR(0, 5516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__919);
  __Pyx_GIVEREF(__pyx_tuple__919);

  /* "pymola/generated/ModelicaParser.pyx":5585
 *             self.state = 794
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:
 *                 self.state = 789
 */
  __pyx_tuple__920 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__920)) __PYX_ERR(0, 5585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__920);
  __Pyx_GIVEREF(__pyx_tuple__920);

  /* "pymola/generated/ModelicaParser.pyx":5593
 *                 self.state = 796
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *             self.state = 810
 */
  __pyx_tuple__921 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__921)) __PYX_ERR(0, 5593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__921);
  __Pyx_GIVEREF(__pyx_tuple__921);

  /* "pymola/generated/ModelicaParser.pyx":5597
 *             self.state = 810
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             while _la==ModelicaParser.T__53:
 *                 self.state = 797
 */
  __pyx_tuple__922 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__922)) __PYX_ERR(0, 5597, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__922);
  __Pyx_GIVEREF(__pyx_tuple__922);

  /* "pymola/generated/ModelicaParser.pyx":5607
 *                 self.state = 805
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__43) | (1 << ModelicaParser.T__44) | (1 << ModelicaParser.T__48) | (1 << ModelicaParser.T__51) | (1 << ModelicaParser.T__52))) != 0) or _la==ModelicaParser.IDENT:
 *                     self.state = 800
 */
  __pyx_tuple__923 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__923)) __PYX_ERR(0, 5607, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__923);
  __Pyx_GIVEREF(__pyx_tuple__923);

  /* "pymola/generated/ModelicaParser.pyx":5615
 *                     self.state = 807
 *                     self._errHandler.sync(self)
 *                     _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *                 self.state = 812
 */
  __pyx_tuple__924 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__924)) __PYX_ERR(0, 5615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__924);
  __Pyx_GIVEREF(__pyx_tuple__924);

  /* "pymola/generated/ModelicaParser.pyx":5619
 *                 self.state = 812
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *             self.state = 813
 */
  __pyx_tuple__925 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__925)) __PYX_ERR(0, 5619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__925);
  __Pyx_GIVEREF(__pyx_tuple__925);

  /* "pymola/generated/ModelicaParser.pyx":5770
 *             self.state = 841
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if token in [ModelicaParser.T__13, ModelicaParser.T__16, ModelicaParser.T__19, ModelicaParser.T__29, ModelicaParser.T__55, ModelicaParser.T__70, ModelicaParser.T__73, ModelicaParser.T__74, ModelicaParser.T__75, ModelicaParser.INITIAL, ModelicaParser.IDENT, ModelicaParser.STRING, ModelicaParser.UNSIGNED_NUMBER]:
 *                 localctx = ModelicaParser.Expression_simpleContext(self, localctx)
 */
  __pyx_tuple__926 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__926)) __PYX_ERR(0, 5770, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__926);
  __Pyx_GIVEREF(__pyx_tuple__926);

  /* "pymola/generated/ModelicaParser.pyx":5790
 *                 self.state = 835
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 while _la==ModelicaParser.T__46:
 *                     self.state = 828
 */
  __pyx_tuple__927 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__927)) __PYX_ERR(0, 5790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__927);
  __Pyx_GIVEREF(__pyx_tuple__927);

  /* "pymola/generated/ModelicaParser.pyx":5802
 *                     self.state = 837
 *                     self._errHandler.sync(self)
 *                     _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *                 self.state = 838
 */
  __pyx_tuple__928 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__928)) __PYX_ERR(0, 5802, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__928);
  __Pyx_GIVEREF(__pyx_tuple__928);

  /* "pymola/generated/ModelicaParser.pyx":5861
 *             self.enterOuterAlt(localctx, 1)
 *             self.state = 843
 *             self.expr(0)             # <<<<<<<<<<<<<<
 *             self.state = 850
 *             self._errHandler.sync(self)
 */
  __pyx_tuple__929 = PyTuple_Pack(1, __pyx_int_0); if (unlikely(!__pyx_tuple__929)) __PYX_ERR(0, 5861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__929);
  __Pyx_GIVEREF(__pyx_tuple__929);

  /* "pymola/generated/ModelicaParser.pyx":5864
 *             self.state = 850
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.T__17:
 *                 self.state = 844
 */
  __pyx_tuple__930 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__930)) __PYX_ERR(0, 5864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__930);
  __Pyx_GIVEREF(__pyx_tuple__930);

  /* "pymola/generated/ModelicaParser.pyx":5869
 *                 self.match(ModelicaParser.T__17)
 *                 self.state = 845
 *                 self.expr(0)             # <<<<<<<<<<<<<<
 *                 self.state = 848
 *                 self._errHandler.sync(self)
 */
  __pyx_tuple__931 = PyTuple_Pack(1, __pyx_int_0); if (unlikely(!__pyx_tuple__931)) __PYX_ERR(0, 5869, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__931);
  __Pyx_GIVEREF(__pyx_tuple__931);

  /* "pymola/generated/ModelicaParser.pyx":5872
 *                 self.state = 848
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if _la==ModelicaParser.T__17:
 *                     self.state = 846
 */
  __pyx_tuple__932 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__932)) __PYX_ERR(0, 5872, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__932);
  __Pyx_GIVEREF(__pyx_tuple__932);

  /* "pymola/generated/ModelicaParser.pyx":5877
 *                     self.match(ModelicaParser.T__17)
 *                     self.state = 847
 *                     self.expr(0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__933 = PyTuple_Pack(1, __pyx_int_0); if (unlikely(!__pyx_tuple__933)) __PYX_ERR(0, 5877, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__933);
  __Pyx_GIVEREF(__pyx_tuple__933);

  /* "pymola/generated/ModelicaParser.pyx":6175
 *                 localctx.op = self.match(ModelicaParser.T__55)
 *                 self.state = 854
 *                 self.expr(9)             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
  __pyx_tuple__934 = PyTuple_Pack(1, __pyx_int_9); if (unlikely(!__pyx_tuple__934)) __PYX_ERR(0, 6175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__934);
  __Pyx_GIVEREF(__pyx_tuple__934);

  /* "pymola/generated/ModelicaParser.pyx":6185
 *                 self.primary()
 *                 self.state = 856
 *                 localctx.op = self._input.LT(1)             # <<<<<<<<<<<<<<
 *                 _la = self._input.LA(1)
 *                 if not(_la==ModelicaParser.T__56 or _la==ModelicaParser.T__57):
 */
  __pyx_tuple__935 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__935)) __PYX_ERR(0, 6185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__935);
  __Pyx_GIVEREF(__pyx_tuple__935);

  /* "pymola/generated/ModelicaParser.pyx":6186
 *                 self.state = 856
 *                 localctx.op = self._input.LT(1)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if not(_la==ModelicaParser.T__56 or _la==ModelicaParser.T__57):
 *                     localctx.op = self._errHandler.recoverInline(self)
 */
  __pyx_tuple__936 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__936)) __PYX_ERR(0, 6186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__936);
  __Pyx_GIVEREF(__pyx_tuple__936);

  /* "pymola/generated/ModelicaParser.pyx":6203
 *                 self.match(ModelicaParser.T__70)
 *                 self.state = 860
 *                 self.expr(4)             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
  __pyx_tuple__937 = PyTuple_Pack(1, __pyx_int_4); if (unlikely(!__pyx_tuple__937)) __PYX_ERR(0, 6203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__937);
  __Pyx_GIVEREF(__pyx_tuple__937);

  /* "pymola/generated/ModelicaParser.pyx":6215
 * 
 * 
 *             self._ctx.stop = self._input.LT(-1)             # <<<<<<<<<<<<<<
 *             self.state = 881
 *             self._errHandler.sync(self)
 */
  __pyx_tuple__938 = PyTuple_Pack(1, __pyx_int_neg_1); if (unlikely(!__pyx_tuple__938)) __PYX_ERR(0, 6215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__938);
  __Pyx_GIVEREF(__pyx_tuple__938);

  /* "pymola/generated/ModelicaParser.pyx":6235
 *                             raise FailedPredicateException(self, "self.precpred(self._ctx, 7)")
 *                         self.state = 865
 *                         localctx.op = self._input.LT(1)             # <<<<<<<<<<<<<<
 *                         _la = self._input.LA(1)
 *                         if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__28) | (1 << ModelicaParser.T__58) | (1 << ModelicaParser.T__59) | (1 << ModelicaParser.T__60))) != 0)):
 */
  __pyx_tuple__939 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__939)) __PYX_ERR(0, 6235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__939);
  __Pyx_GIVEREF(__pyx_tuple__939);

  /* "pymola/generated/ModelicaParser.pyx":6236
 *                         self.state = 865
 *                         localctx.op = self._input.LT(1)
 *                         _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                         if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__28) | (1 << ModelicaParser.T__58) | (1 << ModelicaParser.T__59) | (1 << ModelicaParser.T__60))) != 0)):
 *                             localctx.op = self._errHandler.recoverInline(self)
 */
  __pyx_tuple__940 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__940)) __PYX_ERR(0, 6236, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__940);
  __Pyx_GIVEREF(__pyx_tuple__940);

  /* "pymola/generated/ModelicaParser.pyx":6243
 *                             self.consume()
 *                         self.state = 866
 *                         self.expr(8)             # <<<<<<<<<<<<<<
 *                         pass
 * 
 */
  __pyx_tuple__941 = PyTuple_Pack(1, __pyx_int_8); if (unlikely(!__pyx_tuple__941)) __PYX_ERR(0, 6243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__941);
  __Pyx_GIVEREF(__pyx_tuple__941);

  /* "pymola/generated/ModelicaParser.pyx":6254
 *                             raise FailedPredicateException(self, "self.precpred(self._ctx, 6)")
 *                         self.state = 868
 *                         localctx.op = self._input.LT(1)             # <<<<<<<<<<<<<<
 *                         _la = self._input.LA(1)
 *                         if not(((((_la - 56)) & ~0x3f) == 0 and ((1 << (_la - 56)) & ((1 << (ModelicaParser.T__55 - 56)) | (1 << (ModelicaParser.T__61 - 56)) | (1 << (ModelicaParser.T__62 - 56)) | (1 << (ModelicaParser.T__63 - 56)))) != 0)):
 */
  __pyx_tuple__942 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__942)) __PYX_ERR(0, 6254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__942);
  __Pyx_GIVEREF(__pyx_tuple__942);

  /* "pymola/generated/ModelicaParser.pyx":6255
 *                         self.state = 868
 *                         localctx.op = self._input.LT(1)
 *                         _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                         if not(((((_la - 56)) & ~0x3f) == 0 and ((1 << (_la - 56)) & ((1 << (ModelicaParser.T__55 - 56)) | (1 << (ModelicaParser.T__61 - 56)) | (1 << (ModelicaParser.T__62 - 56)) | (1 << (ModelicaParser.T__63 - 56)))) != 0)):
 *                             localctx.op = self._errHandler.recoverInline(self)
 */
  __pyx_tuple__943 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__943)) __PYX_ERR(0, 6255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__943);
  __Pyx_GIVEREF(__pyx_tuple__943);

  /* "pymola/generated/ModelicaParser.pyx":6262
 *                             self.consume()
 *                         self.state = 869
 *                         self.expr(7)             # <<<<<<<<<<<<<<
 *                         pass
 * 
 */
  __pyx_tuple__944 = PyTuple_Pack(1, __pyx_int_7); if (unlikely(!__pyx_tuple__944)) __PYX_ERR(0, 6262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__944);
  __Pyx_GIVEREF(__pyx_tuple__944);

  /* "pymola/generated/ModelicaParser.pyx":6273
 *                             raise FailedPredicateException(self, "self.precpred(self._ctx, 5)")
 *                         self.state = 871
 *                         localctx.op = self._input.LT(1)             # <<<<<<<<<<<<<<
 *                         _la = self._input.LA(1)
 *                         if not(((((_la - 65)) & ~0x3f) == 0 and ((1 << (_la - 65)) & ((1 << (ModelicaParser.T__64 - 65)) | (1 << (ModelicaParser.T__65 - 65)) | (1 << (ModelicaParser.T__66 - 65)) | (1 << (ModelicaParser.T__67 - 65)) | (1 << (ModelicaParser.T__68 - 65)) | (1 << (ModelicaParser.T__69 - 65)))) != 0)):
 */
  __pyx_tuple__945 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__945)) __PYX_ERR(0, 6273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__945);
  __Pyx_GIVEREF(__pyx_tuple__945);

  /* "pymola/generated/ModelicaParser.pyx":6274
 *                         self.state = 871
 *                         localctx.op = self._input.LT(1)
 *                         _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                         if not(((((_la - 65)) & ~0x3f) == 0 and ((1 << (_la - 65)) & ((1 << (ModelicaParser.T__64 - 65)) | (1 << (ModelicaParser.T__65 - 65)) | (1 << (ModelicaParser.T__66 - 65)) | (1 << (ModelicaParser.T__67 - 65)) | (1 << (ModelicaParser.T__68 - 65)) | (1 << (ModelicaParser.T__69 - 65)))) != 0)):
 *                             localctx.op = self._errHandler.recoverInline(self)
 */
  __pyx_tuple__946 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__946)) __PYX_ERR(0, 6274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__946);
  __Pyx_GIVEREF(__pyx_tuple__946);

  /* "pymola/generated/ModelicaParser.pyx":6281
 *                             self.consume()
 *                         self.state = 872
 *                         self.expr(6)             # <<<<<<<<<<<<<<
 *                         pass
 * 
 */
  __pyx_tuple__947 = PyTuple_Pack(1, __pyx_int_6); if (unlikely(!__pyx_tuple__947)) __PYX_ERR(0, 6281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__947);
  __Pyx_GIVEREF(__pyx_tuple__947);

  /* "pymola/generated/ModelicaParser.pyx":6294
 *                         self.match(ModelicaParser.T__71)
 *                         self.state = 875
 *                         self.expr(4)             # <<<<<<<<<<<<<<
 *                         pass
 * 
 */
  __pyx_tuple__948 = PyTuple_Pack(1, __pyx_int_4); if (unlikely(!__pyx_tuple__948)) __PYX_ERR(0, 6294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__948);
  __Pyx_GIVEREF(__pyx_tuple__948);

  /* "pymola/generated/ModelicaParser.pyx":6307
 *                         self.match(ModelicaParser.T__72)
 *                         self.state = 878
 *                         self.expr(3)             # <<<<<<<<<<<<<<
 *                         pass
 * 
 */
  __pyx_tuple__949 = PyTuple_Pack(1, __pyx_int_3); if (unlikely(!__pyx_tuple__949)) __PYX_ERR(0, 6307, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__949);
  __Pyx_GIVEREF(__pyx_tuple__949);

  /* "pymola/generated/ModelicaParser.pyx":6726
 *                 self.state = 906
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 while _la==ModelicaParser.T__0:
 *                     self.state = 902
 */
  __pyx_tuple__950 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__950)) __PYX_ERR(0, 6726, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__950);
  __Pyx_GIVEREF(__pyx_tuple__950);

  /* "pymola/generated/ModelicaParser.pyx":6734
 *                     self.state = 908
 *                     self._errHandler.sync(self)
 *                     _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *                 self.state = 909
 */
  __pyx_tuple__951 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__951)) __PYX_ERR(0, 6734, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__951);
  __Pyx_GIVEREF(__pyx_tuple__951);

  /* "pymola/generated/ModelicaParser.pyx":6808
 *             self.state = 919
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.T__27:
 *                 self.state = 918
 */
  __pyx_tuple__952 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__952)) __PYX_ERR(0, 6808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__952);
  __Pyx_GIVEREF(__pyx_tuple__952);

  /* "pymola/generated/ModelicaParser.pyx":6818
 *             self.state = 926
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             while _la==ModelicaParser.T__27:
 *                 self.state = 922
 */
  __pyx_tuple__953 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__953)) __PYX_ERR(0, 6818, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__953);
  __Pyx_GIVEREF(__pyx_tuple__953);

  /* "pymola/generated/ModelicaParser.pyx":6826
 *                 self.state = 928
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *         except RecognitionException as re:
 */
  __pyx_tuple__954 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__954)) __PYX_ERR(0, 6826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__954);
  __Pyx_GIVEREF(__pyx_tuple__954);

  /* "pymola/generated/ModelicaParser.pyx":6996
 *             self.state = 943
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__12) | (1 << ModelicaParser.T__13) | (1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__19) | (1 << ModelicaParser.T__29) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__55))) != 0) or ((((_la - 71)) & ~0x3f) == 0 and ((1 << (_la - 71)) & ((1 << (ModelicaParser.T__70 - 71)) | (1 << (ModelicaParser.T__73 - 71)) | (1 << (ModelicaParser.T__74 - 71)) | (1 << (ModelicaParser.T__75 - 71)) | (1 << (ModelicaParser.INITIAL - 71)) | (1 << (ModelicaParser.IDENT - 71)) | (1 << (ModelicaParser.STRING - 71)) | (1 << (ModelicaParser.UNSIGNED_NUMBER - 71)))) != 0):
 *                 self.state = 942
 */
  __pyx_tuple__955 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__955)) __PYX_ERR(0, 6996, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__955);
  __Pyx_GIVEREF(__pyx_tuple__955);

  /* "pymola/generated/ModelicaParser.pyx":7071
 *                 self.state = 954
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 while _la==ModelicaParser.T__20 or _la==ModelicaParser.T__48:
 *                     self.state = 952
 */
  __pyx_tuple__956 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__956)) __PYX_ERR(0, 7071, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__956);
  __Pyx_GIVEREF(__pyx_tuple__956);

  /* "pymola/generated/ModelicaParser.pyx":7075
 *                     self.state = 952
 *                     self._errHandler.sync(self)
 *                     token = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                     if token in [ModelicaParser.T__20]:
 *                         self.state = 948
 */
  __pyx_tuple__957 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__957)) __PYX_ERR(0, 7075, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__957);
  __Pyx_GIVEREF(__pyx_tuple__957);

  /* "pymola/generated/ModelicaParser.pyx":7093
 *                     self.state = 956
 *                     self._errHandler.sync(self)
 *                     _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *                 pass
 */
  __pyx_tuple__958 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__958)) __PYX_ERR(0, 7093, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__958);
  __Pyx_GIVEREF(__pyx_tuple__958);

  /* "pymola/generated/ModelicaParser.pyx":7156
 *             self.state = 965
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             while _la==ModelicaParser.T__20:
 *                 self.state = 961
 */
  __pyx_tuple__959 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__959)) __PYX_ERR(0, 7156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__959);
  __Pyx_GIVEREF(__pyx_tuple__959);

  /* "pymola/generated/ModelicaParser.pyx":7164
 *                 self.state = 967
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *         except RecognitionException as re:
 */
  __pyx_tuple__960 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__960)) __PYX_ERR(0, 7164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__960);
  __Pyx_GIVEREF(__pyx_tuple__960);

  /* "pymola/generated/ModelicaParser.pyx":7305
 *             self.state = 981
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if token in [ModelicaParser.T__12]:
 *                 localctx = ModelicaParser.Argument_functionContext(self, localctx)
 */
  __pyx_tuple__961 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__961)) __PYX_ERR(0, 7305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__961);
  __Pyx_GIVEREF(__pyx_tuple__961);

  /* "pymola/generated/ModelicaParser.pyx":7317
 *                 self.state = 976
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 if _la==ModelicaParser.IDENT:
 *                     self.state = 975
 */
  __pyx_tuple__962 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__962)) __PYX_ERR(0, 7317, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__962);
  __Pyx_GIVEREF(__pyx_tuple__962);

  /* "pymola/generated/ModelicaParser.pyx":7385
 *             self.state = 984
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ModelicaParser.T__13) | (1 << ModelicaParser.T__16) | (1 << ModelicaParser.T__19) | (1 << ModelicaParser.T__29) | (1 << ModelicaParser.T__39) | (1 << ModelicaParser.T__55))) != 0) or ((((_la - 71)) & ~0x3f) == 0 and ((1 << (_la - 71)) & ((1 << (ModelicaParser.T__70 - 71)) | (1 << (ModelicaParser.T__73 - 71)) | (1 << (ModelicaParser.T__74 - 71)) | (1 << (ModelicaParser.T__75 - 71)) | (1 << (ModelicaParser.INITIAL - 71)) | (1 << (ModelicaParser.IDENT - 71)) | (1 << (ModelicaParser.STRING - 71)) | (1 << (ModelicaParser.UNSIGNED_NUMBER - 71)))) != 0):
 *                 self.state = 983
 */
  __pyx_tuple__963 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__963)) __PYX_ERR(0, 7385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__963);
  __Pyx_GIVEREF(__pyx_tuple__963);

  /* "pymola/generated/ModelicaParser.pyx":7393
 *             self.state = 990
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             while _la==ModelicaParser.T__20:
 *                 self.state = 986
 */
  __pyx_tuple__964 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__964)) __PYX_ERR(0, 7393, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__964);
  __Pyx_GIVEREF(__pyx_tuple__964);

  /* "pymola/generated/ModelicaParser.pyx":7401
 *                 self.state = 992
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *         except RecognitionException as re:
 */
  __pyx_tuple__965 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__965)) __PYX_ERR(0, 7401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__965);
  __Pyx_GIVEREF(__pyx_tuple__965);

  /* "pymola/generated/ModelicaParser.pyx":7455
 *             self.state = 998
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             while _la==ModelicaParser.T__20:
 *                 self.state = 994
 */
  __pyx_tuple__966 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__966)) __PYX_ERR(0, 7455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__966);
  __Pyx_GIVEREF(__pyx_tuple__966);

  /* "pymola/generated/ModelicaParser.pyx":7463
 *                 self.state = 1000
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *         except RecognitionException as re:
 */
  __pyx_tuple__967 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__967)) __PYX_ERR(0, 7463, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__967);
  __Pyx_GIVEREF(__pyx_tuple__967);

  /* "pymola/generated/ModelicaParser.pyx":7519
 *             self.state = 1007
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             while _la==ModelicaParser.T__20:
 *                 self.state = 1003
 */
  __pyx_tuple__968 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__968)) __PYX_ERR(0, 7519, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__968);
  __Pyx_GIVEREF(__pyx_tuple__968);

  /* "pymola/generated/ModelicaParser.pyx":7527
 *                 self.state = 1009
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 *             self.state = 1010
 */
  __pyx_tuple__969 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__969)) __PYX_ERR(0, 7527, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__969);
  __Pyx_GIVEREF(__pyx_tuple__969);

  /* "pymola/generated/ModelicaParser.pyx":7576
 *             self.state = 1014
 *             self._errHandler.sync(self)
 *             token = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if token in [ModelicaParser.T__17]:
 *                 self.enterOuterAlt(localctx, 1)
 */
  __pyx_tuple__970 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__970)) __PYX_ERR(0, 7576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__970);
  __Pyx_GIVEREF(__pyx_tuple__970);

  /* "pymola/generated/ModelicaParser.pyx":7643
 *             self.state = 1018
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.T__77:
 *                 self.state = 1017
 */
  __pyx_tuple__971 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__971)) __PYX_ERR(0, 7643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__971);
  __Pyx_GIVEREF(__pyx_tuple__971);

  /* "pymola/generated/ModelicaParser.pyx":7698
 *             self.state = 1028
 *             self._errHandler.sync(self)
 *             _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *             if _la==ModelicaParser.STRING:
 *                 self.state = 1020
 */
  __pyx_tuple__972 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__972)) __PYX_ERR(0, 7698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__972);
  __Pyx_GIVEREF(__pyx_tuple__972);

  /* "pymola/generated/ModelicaParser.pyx":7704
 *                 self.state = 1025
 *                 self._errHandler.sync(self)
 *                 _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 *                 while _la==ModelicaParser.T__61:
 *                     self.state = 1021
 */
  __pyx_tuple__973 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__973)) __PYX_ERR(0, 7704, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__973);
  __Pyx_GIVEREF(__pyx_tuple__973);

  /* "pymola/generated/ModelicaParser.pyx":7712
 *                     self.state = 1027
 *                     self._errHandler.sync(self)
 *                     _la = self._input.LA(1)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__974 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__974)) __PYX_ERR(0, 7712, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__974);
  __Pyx_GIVEREF(__pyx_tuple__974);

  /* "pymola/generated/ModelicaParser.pyx":8
 * import sys
 * 
 * def serializedATN():             # <<<<<<<<<<<<<<
 *     with StringIO() as buf:
 *         buf.write("\3\u608b\ua72a\u8133\ub9ed\u417c\u3be7\u7786\u5964\3^")
 */
  __pyx_tuple__976 = PyTuple_Pack(1, __pyx_n_s_buf); if (unlikely(!__pyx_tuple__976)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__976);
  __Pyx_GIVEREF(__pyx_tuple__976);
  __pyx_codeobj__977 = (PyObject*)__Pyx_PyCode_New(0, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__976, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_serializedATN, 8, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__977)) __PYX_ERR(0, 8, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":775
 *     WS=92
 * 
 *     def __init__(self, input:TokenStream, output:TextIO = sys.stdout):             # <<<<<<<<<<<<<<
 *         super().__init__(input, output)
 *         self.checkVersion("4.7")
 */
  __pyx_tuple__1006 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_input, __pyx_n_s_output); if (unlikely(!__pyx_tuple__1006)) __PYX_ERR(0, 775, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1006);
  __Pyx_GIVEREF(__pyx_tuple__1006);
  __pyx_codeobj__1007 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1006, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 775, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1007)) __PYX_ERR(0, 775, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":785
 *     class Stored_definitionContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1008 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1008)) __PYX_ERR(0, 785, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1008);
  __Pyx_GIVEREF(__pyx_tuple__1008);
  __pyx_codeobj__1009 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1008, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 785, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1009)) __PYX_ERR(0, 785, __pyx_L1_error)
  __pyx_tuple__1010 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1010)) __PYX_ERR(0, 785, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1010);
  __Pyx_GIVEREF(__pyx_tuple__1010);

  /* "pymola/generated/ModelicaParser.pyx":789
 *             self.parser = parser
 * 
 *         def WITHIN(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.WITHIN, 0)
 * 
 */
  __pyx_tuple__1011 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1011)) __PYX_ERR(0, 789, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1011);
  __Pyx_GIVEREF(__pyx_tuple__1011);
  __pyx_codeobj__1012 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1011, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_WITHIN, 789, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1012)) __PYX_ERR(0, 789, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":792
 *             return self.getToken(ModelicaParser.WITHIN, 0)
 * 
 *         def stored_definition_class(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Stored_definition_classContext)
 */
  __pyx_tuple__1013 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__1013)) __PYX_ERR(0, 792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1013);
  __Pyx_GIVEREF(__pyx_tuple__1013);
  __pyx_codeobj__1014 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1013, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_stored_definition_class, 792, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1014)) __PYX_ERR(0, 792, __pyx_L1_error)
  __pyx_tuple__1015 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__1015)) __PYX_ERR(0, 792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1015);
  __Pyx_GIVEREF(__pyx_tuple__1015);

  /* "pymola/generated/ModelicaParser.pyx":799
 * 
 * 
 *         def component_reference(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 */
  __pyx_tuple__1016 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1016)) __PYX_ERR(0, 799, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1016);
  __Pyx_GIVEREF(__pyx_tuple__1016);
  __pyx_codeobj__1017 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1016, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_component_reference, 799, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1017)) __PYX_ERR(0, 799, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":803
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_stored_definition
 * 
 */
  __pyx_tuple__1018 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1018)) __PYX_ERR(0, 803, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1018);
  __Pyx_GIVEREF(__pyx_tuple__1018);
  __pyx_codeobj__1019 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1018, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 803, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1019)) __PYX_ERR(0, 803, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":806
 *             return ModelicaParser.RULE_stored_definition
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStored_definition" ):
 *                 listener.enterStored_definition(self)
 */
  __pyx_tuple__1020 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1020)) __PYX_ERR(0, 806, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1020);
  __Pyx_GIVEREF(__pyx_tuple__1020);
  __pyx_codeobj__1021 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1020, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 806, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1021)) __PYX_ERR(0, 806, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":810
 *                 listener.enterStored_definition(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStored_definition" ):
 *                 listener.exitStored_definition(self)
 */
  __pyx_tuple__1022 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1022)) __PYX_ERR(0, 810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1022);
  __Pyx_GIVEREF(__pyx_tuple__1022);
  __pyx_codeobj__1023 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1022, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 810, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1023)) __PYX_ERR(0, 810, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":814
 *                 listener.exitStored_definition(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStored_definition" ):
 *                 return visitor.visitStored_definition(self)
 */
  __pyx_tuple__1024 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1024)) __PYX_ERR(0, 814, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1024);
  __Pyx_GIVEREF(__pyx_tuple__1024);
  __pyx_codeobj__1025 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1024, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 814, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1025)) __PYX_ERR(0, 814, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":823
 * 
 * 
 *     def stored_definition(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Stored_definitionContext(self, self._ctx, self.state)
 */
  __pyx_tuple__1026 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1026)) __PYX_ERR(0, 823, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1026);
  __Pyx_GIVEREF(__pyx_tuple__1026);
  __pyx_codeobj__1027 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1026, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_stored_definition, 823, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1027)) __PYX_ERR(0, 823, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":868
 *     class Stored_definition_classContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1028 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1028)) __PYX_ERR(0, 868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1028);
  __Pyx_GIVEREF(__pyx_tuple__1028);
  __pyx_codeobj__1029 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1028, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 868, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1029)) __PYX_ERR(0, 868, __pyx_L1_error)
  __pyx_tuple__1030 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1030)) __PYX_ERR(0, 868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1030);
  __Pyx_GIVEREF(__pyx_tuple__1030);

  /* "pymola/generated/ModelicaParser.pyx":872
 *             self.parser = parser
 * 
 *         def class_definition(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_definitionContext,0)
 * 
 */
  __pyx_tuple__1031 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1031)) __PYX_ERR(0, 872, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1031);
  __Pyx_GIVEREF(__pyx_tuple__1031);
  __pyx_codeobj__1032 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1031, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_class_definition, 872, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1032)) __PYX_ERR(0, 872, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":876
 * 
 * 
 *         def FINAL(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.FINAL, 0)
 * 
 */
  __pyx_tuple__1033 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1033)) __PYX_ERR(0, 876, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1033);
  __Pyx_GIVEREF(__pyx_tuple__1033);
  __pyx_codeobj__1034 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1033, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_FINAL, 876, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1034)) __PYX_ERR(0, 876, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":879
 *             return self.getToken(ModelicaParser.FINAL, 0)
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_stored_definition_class
 * 
 */
  __pyx_tuple__1035 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1035)) __PYX_ERR(0, 879, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1035);
  __Pyx_GIVEREF(__pyx_tuple__1035);
  __pyx_codeobj__1036 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1035, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 879, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1036)) __PYX_ERR(0, 879, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":882
 *             return ModelicaParser.RULE_stored_definition_class
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStored_definition_class" ):
 *                 listener.enterStored_definition_class(self)
 */
  __pyx_tuple__1037 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1037)) __PYX_ERR(0, 882, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1037);
  __Pyx_GIVEREF(__pyx_tuple__1037);
  __pyx_codeobj__1038 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1037, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 882, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1038)) __PYX_ERR(0, 882, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":886
 *                 listener.enterStored_definition_class(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStored_definition_class" ):
 *                 listener.exitStored_definition_class(self)
 */
  __pyx_tuple__1039 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1039)) __PYX_ERR(0, 886, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1039);
  __Pyx_GIVEREF(__pyx_tuple__1039);
  __pyx_codeobj__1040 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1039, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 886, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1040)) __PYX_ERR(0, 886, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":890
 *                 listener.exitStored_definition_class(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStored_definition_class" ):
 *                 return visitor.visitStored_definition_class(self)
 */
  __pyx_tuple__1041 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1041)) __PYX_ERR(0, 890, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1041);
  __Pyx_GIVEREF(__pyx_tuple__1041);
  __pyx_codeobj__1042 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1041, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 890, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1042)) __PYX_ERR(0, 890, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":899
 * 
 * 
 *     def stored_definition_class(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Stored_definition_classContext(self, self._ctx, self.state)
 */
  __pyx_tuple__1043 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1043)) __PYX_ERR(0, 899, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1043);
  __Pyx_GIVEREF(__pyx_tuple__1043);
  __pyx_codeobj__1044 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1043, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_stored_definition_class, 899, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1044)) __PYX_ERR(0, 899, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":928
 *     class Class_definitionContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1045 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1045)) __PYX_ERR(0, 928, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1045);
  __Pyx_GIVEREF(__pyx_tuple__1045);
  __pyx_codeobj__1046 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1045, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 928, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1046)) __PYX_ERR(0, 928, __pyx_L1_error)
  __pyx_tuple__1047 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1047)) __PYX_ERR(0, 928, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1047);
  __Pyx_GIVEREF(__pyx_tuple__1047);

  /* "pymola/generated/ModelicaParser.pyx":932
 *             self.parser = parser
 * 
 *         def class_prefixes(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_prefixesContext,0)
 * 
 */
  __pyx_tuple__1048 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1048)) __PYX_ERR(0, 932, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1048);
  __Pyx_GIVEREF(__pyx_tuple__1048);
  __pyx_codeobj__1049 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1048, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_class_prefixes, 932, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1049)) __PYX_ERR(0, 932, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":936
 * 
 * 
 *         def class_specifier(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_specifierContext,0)
 * 
 */
  __pyx_tuple__1050 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1050)) __PYX_ERR(0, 936, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1050);
  __Pyx_GIVEREF(__pyx_tuple__1050);
  __pyx_codeobj__1051 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1050, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_class_specifier, 936, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1051)) __PYX_ERR(0, 936, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":940
 * 
 * 
 *         def ENCAPSULATED(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.ENCAPSULATED, 0)
 * 
 */
  __pyx_tuple__1052 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1052)) __PYX_ERR(0, 940, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1052);
  __Pyx_GIVEREF(__pyx_tuple__1052);
  __pyx_codeobj__1053 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1052, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_ENCAPSULATED, 940, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1053)) __PYX_ERR(0, 940, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":943
 *             return self.getToken(ModelicaParser.ENCAPSULATED, 0)
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_class_definition
 * 
 */
  __pyx_tuple__1054 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1054)) __PYX_ERR(0, 943, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1054);
  __Pyx_GIVEREF(__pyx_tuple__1054);
  __pyx_codeobj__1055 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1054, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 943, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1055)) __PYX_ERR(0, 943, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":946
 *             return ModelicaParser.RULE_class_definition
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterClass_definition" ):
 *                 listener.enterClass_definition(self)
 */
  __pyx_tuple__1056 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1056)) __PYX_ERR(0, 946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1056);
  __Pyx_GIVEREF(__pyx_tuple__1056);
  __pyx_codeobj__1057 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1056, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 946, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1057)) __PYX_ERR(0, 946, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":950
 *                 listener.enterClass_definition(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitClass_definition" ):
 *                 listener.exitClass_definition(self)
 */
  __pyx_tuple__1058 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1058)) __PYX_ERR(0, 950, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1058);
  __Pyx_GIVEREF(__pyx_tuple__1058);
  __pyx_codeobj__1059 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1058, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 950, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1059)) __PYX_ERR(0, 950, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":954
 *                 listener.exitClass_definition(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitClass_definition" ):
 *                 return visitor.visitClass_definition(self)
 */
  __pyx_tuple__1060 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1060)) __PYX_ERR(0, 954, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1060);
  __Pyx_GIVEREF(__pyx_tuple__1060);
  __pyx_codeobj__1061 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1060, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 954, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1061)) __PYX_ERR(0, 954, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":963
 * 
 * 
 *     def class_definition(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Class_definitionContext(self, self._ctx, self.state)
 */
  __pyx_tuple__1062 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1062)) __PYX_ERR(0, 963, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1062);
  __Pyx_GIVEREF(__pyx_tuple__1062);
  __pyx_codeobj__1063 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1062, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_class_definition, 963, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1063)) __PYX_ERR(0, 963, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":992
 *     class Class_prefixesContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1064 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1064)) __PYX_ERR(0, 992, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1064);
  __Pyx_GIVEREF(__pyx_tuple__1064);
  __pyx_codeobj__1065 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1064, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 992, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1065)) __PYX_ERR(0, 992, __pyx_L1_error)
  __pyx_tuple__1066 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1066)) __PYX_ERR(0, 992, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1066);
  __Pyx_GIVEREF(__pyx_tuple__1066);

  /* "pymola/generated/ModelicaParser.pyx":996
 *             self.parser = parser
 * 
 *         def class_type(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_typeContext,0)
 * 
 */
  __pyx_tuple__1067 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1067)) __PYX_ERR(0, 996, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1067);
  __Pyx_GIVEREF(__pyx_tuple__1067);
  __pyx_codeobj__1068 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1067, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_class_type, 996, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1068)) __PYX_ERR(0, 996, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1000
 * 
 * 
 *         def PARTIAL(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.PARTIAL, 0)
 * 
 */
  __pyx_tuple__1069 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1069)) __PYX_ERR(0, 1000, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1069);
  __Pyx_GIVEREF(__pyx_tuple__1069);
  __pyx_codeobj__1070 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1069, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_PARTIAL, 1000, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1070)) __PYX_ERR(0, 1000, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1003
 *             return self.getToken(ModelicaParser.PARTIAL, 0)
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_class_prefixes
 * 
 */
  __pyx_tuple__1071 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1071)) __PYX_ERR(0, 1003, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1071);
  __Pyx_GIVEREF(__pyx_tuple__1071);
  __pyx_codeobj__1072 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1071, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 1003, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1072)) __PYX_ERR(0, 1003, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1006
 *             return ModelicaParser.RULE_class_prefixes
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterClass_prefixes" ):
 *                 listener.enterClass_prefixes(self)
 */
  __pyx_tuple__1073 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1073)) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1073);
  __Pyx_GIVEREF(__pyx_tuple__1073);
  __pyx_codeobj__1074 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1073, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 1006, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1074)) __PYX_ERR(0, 1006, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1010
 *                 listener.enterClass_prefixes(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitClass_prefixes" ):
 *                 listener.exitClass_prefixes(self)
 */
  __pyx_tuple__1075 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1075)) __PYX_ERR(0, 1010, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1075);
  __Pyx_GIVEREF(__pyx_tuple__1075);
  __pyx_codeobj__1076 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1075, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 1010, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1076)) __PYX_ERR(0, 1010, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1014
 *                 listener.exitClass_prefixes(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitClass_prefixes" ):
 *                 return visitor.visitClass_prefixes(self)
 */
  __pyx_tuple__1077 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1077)) __PYX_ERR(0, 1014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1077);
  __Pyx_GIVEREF(__pyx_tuple__1077);
  __pyx_codeobj__1078 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1077, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 1014, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1078)) __PYX_ERR(0, 1014, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1023
 * 
 * 
 *     def class_prefixes(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Class_prefixesContext(self, self._ctx, self.state)
 */
  __pyx_tuple__1079 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1079)) __PYX_ERR(0, 1023, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1079);
  __Pyx_GIVEREF(__pyx_tuple__1079);
  __pyx_codeobj__1080 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1079, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_class_prefixes, 1023, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1080)) __PYX_ERR(0, 1023, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1050
 *     class Class_typeContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1081 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1081)) __PYX_ERR(0, 1050, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1081);
  __Pyx_GIVEREF(__pyx_tuple__1081);
  __pyx_codeobj__1082 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1081, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 1050, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1082)) __PYX_ERR(0, 1050, __pyx_L1_error)
  __pyx_tuple__1083 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1083)) __PYX_ERR(0, 1050, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1083);
  __Pyx_GIVEREF(__pyx_tuple__1083);

  /* "pymola/generated/ModelicaParser.pyx":1055
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_class_type
 * 
 */
  __pyx_tuple__1084 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1084)) __PYX_ERR(0, 1055, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1084);
  __Pyx_GIVEREF(__pyx_tuple__1084);
  __pyx_codeobj__1085 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1084, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 1055, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1085)) __PYX_ERR(0, 1055, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1058
 *             return ModelicaParser.RULE_class_type
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterClass_type" ):
 *                 listener.enterClass_type(self)
 */
  __pyx_tuple__1086 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1086)) __PYX_ERR(0, 1058, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1086);
  __Pyx_GIVEREF(__pyx_tuple__1086);
  __pyx_codeobj__1087 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1086, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 1058, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1087)) __PYX_ERR(0, 1058, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1062
 *                 listener.enterClass_type(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitClass_type" ):
 *                 listener.exitClass_type(self)
 */
  __pyx_tuple__1088 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1088)) __PYX_ERR(0, 1062, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1088);
  __Pyx_GIVEREF(__pyx_tuple__1088);
  __pyx_codeobj__1089 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1088, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 1062, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1089)) __PYX_ERR(0, 1062, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1066
 *                 listener.exitClass_type(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitClass_type" ):
 *                 return visitor.visitClass_type(self)
 */
  __pyx_tuple__1090 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1090)) __PYX_ERR(0, 1066, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1090);
  __Pyx_GIVEREF(__pyx_tuple__1090);
  __pyx_codeobj__1091 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1090, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 1066, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1091)) __PYX_ERR(0, 1066, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1075
 * 
 * 
 *     def class_type(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Class_typeContext(self, self._ctx, self.state)
 */
  __pyx_tuple__1092 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la_2, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1092)) __PYX_ERR(0, 1075, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1092);
  __Pyx_GIVEREF(__pyx_tuple__1092);
  __pyx_codeobj__1093 = (PyObject*)__Pyx_PyCode_New(1, 0, 5, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1092, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_class_type, 1075, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1093)) __PYX_ERR(0, 1075, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1186
 *     class Class_specifierContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1094 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1094)) __PYX_ERR(0, 1186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1094);
  __Pyx_GIVEREF(__pyx_tuple__1094);
  __pyx_codeobj__1095 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1094, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 1186, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1095)) __PYX_ERR(0, 1186, __pyx_L1_error)
  __pyx_tuple__1096 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1096)) __PYX_ERR(0, 1186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1096);
  __Pyx_GIVEREF(__pyx_tuple__1096);

  /* "pymola/generated/ModelicaParser.pyx":1191
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_class_specifier
 * 
 */
  __pyx_tuple__1097 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1097)) __PYX_ERR(0, 1191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1097);
  __Pyx_GIVEREF(__pyx_tuple__1097);
  __pyx_codeobj__1098 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1097, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 1191, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1098)) __PYX_ERR(0, 1191, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1195
 * 
 * 
 *         def copyFrom(self, ctx:ParserRuleContext):             # <<<<<<<<<<<<<<
 *             super().copyFrom(ctx)
 * 
 */
  __pyx_tuple__1099 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__1099)) __PYX_ERR(0, 1195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1099);
  __Pyx_GIVEREF(__pyx_tuple__1099);
  __pyx_codeobj__1100 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1099, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_copyFrom, 1195, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1100)) __PYX_ERR(0, 1195, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1202
 *     class Class_spec_derContext(Class_specifierContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Class_specifierContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__1101 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__1101)) __PYX_ERR(0, 1202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1101);
  __Pyx_GIVEREF(__pyx_tuple__1101);
  __pyx_codeobj__1102 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1101, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 1202, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1102)) __PYX_ERR(0, 1202, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1206
 *             self.copyFrom(ctx)
 * 
 *         def IDENT(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTokens(ModelicaParser.IDENT)
 */
  __pyx_tuple__1103 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__1103)) __PYX_ERR(0, 1206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1103);
  __Pyx_GIVEREF(__pyx_tuple__1103);
  __pyx_codeobj__1104 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1103, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_IDENT, 1206, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1104)) __PYX_ERR(0, 1206, __pyx_L1_error)
  __pyx_tuple__1105 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__1105)) __PYX_ERR(0, 1206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1105);
  __Pyx_GIVEREF(__pyx_tuple__1105);

  /* "pymola/generated/ModelicaParser.pyx":1211
 *             else:
 *                 return self.getToken(ModelicaParser.IDENT, i)
 *         def name(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)
 * 
 */
  __pyx_tuple__1106 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1106)) __PYX_ERR(0, 1211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1106);
  __Pyx_GIVEREF(__pyx_tuple__1106);
  __pyx_codeobj__1107 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1106, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_name, 1211, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1107)) __PYX_ERR(0, 1211, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1214
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)
 * 
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */
  __pyx_tuple__1108 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1108)) __PYX_ERR(0, 1214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1108);
  __Pyx_GIVEREF(__pyx_tuple__1108);
  __pyx_codeobj__1109 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1108, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_comment, 1214, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1109)) __PYX_ERR(0, 1214, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1218
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterClass_spec_der" ):
 *                 listener.enterClass_spec_der(self)
 */
  __pyx_tuple__1110 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1110)) __PYX_ERR(0, 1218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1110);
  __Pyx_GIVEREF(__pyx_tuple__1110);
  __pyx_codeobj__1111 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1110, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 1218, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1111)) __PYX_ERR(0, 1218, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1222
 *                 listener.enterClass_spec_der(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitClass_spec_der" ):
 *                 listener.exitClass_spec_der(self)
 */
  __pyx_tuple__1112 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1112)) __PYX_ERR(0, 1222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1112);
  __Pyx_GIVEREF(__pyx_tuple__1112);
  __pyx_codeobj__1113 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1112, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 1222, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1113)) __PYX_ERR(0, 1222, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1226
 *                 listener.exitClass_spec_der(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitClass_spec_der" ):
 *                 return visitor.visitClass_spec_der(self)
 */
  __pyx_tuple__1114 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1114)) __PYX_ERR(0, 1226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1114);
  __Pyx_GIVEREF(__pyx_tuple__1114);
  __pyx_codeobj__1115 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1114, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 1226, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1115)) __PYX_ERR(0, 1226, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1235
 *     class Class_spec_enumContext(Class_specifierContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Class_specifierContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__1116 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__1116)) __PYX_ERR(0, 1235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1116);
  __Pyx_GIVEREF(__pyx_tuple__1116);
  __pyx_codeobj__1117 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1116, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 1235, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1117)) __PYX_ERR(0, 1235, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1239
 *             self.copyFrom(ctx)
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 *         def comment(self):
 */
  __pyx_tuple__1118 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1118)) __PYX_ERR(0, 1239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1118);
  __Pyx_GIVEREF(__pyx_tuple__1118);
  __pyx_codeobj__1119 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1118, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_IDENT, 1239, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1119)) __PYX_ERR(0, 1239, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1241
 *         def IDENT(self):
 *             return self.getToken(ModelicaParser.IDENT, 0)
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */
  __pyx_tuple__1120 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1120)) __PYX_ERR(0, 1241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1120);
  __Pyx_GIVEREF(__pyx_tuple__1120);
  __pyx_codeobj__1121 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1120, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_comment, 1241, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1121)) __PYX_ERR(0, 1241, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1244
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 *         def enum_list(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Enum_listContext,0)
 * 
 */
  __pyx_tuple__1122 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1122)) __PYX_ERR(0, 1244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1122);
  __Pyx_GIVEREF(__pyx_tuple__1122);
  __pyx_codeobj__1123 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1122, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enum_list, 1244, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1123)) __PYX_ERR(0, 1244, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1248
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterClass_spec_enum" ):
 *                 listener.enterClass_spec_enum(self)
 */
  __pyx_tuple__1124 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1124)) __PYX_ERR(0, 1248, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1124);
  __Pyx_GIVEREF(__pyx_tuple__1124);
  __pyx_codeobj__1125 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1124, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 1248, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1125)) __PYX_ERR(0, 1248, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1252
 *                 listener.enterClass_spec_enum(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitClass_spec_enum" ):
 *                 listener.exitClass_spec_enum(self)
 */
  __pyx_tuple__1126 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1126)) __PYX_ERR(0, 1252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1126);
  __Pyx_GIVEREF(__pyx_tuple__1126);
  __pyx_codeobj__1127 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1126, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 1252, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1127)) __PYX_ERR(0, 1252, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1256
 *                 listener.exitClass_spec_enum(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitClass_spec_enum" ):
 *                 return visitor.visitClass_spec_enum(self)
 */
  __pyx_tuple__1128 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1128)) __PYX_ERR(0, 1256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1128);
  __Pyx_GIVEREF(__pyx_tuple__1128);
  __pyx_codeobj__1129 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1128, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 1256, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1129)) __PYX_ERR(0, 1256, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1265
 *     class Class_spec_baseContext(Class_specifierContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Class_specifierContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__1130 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__1130)) __PYX_ERR(0, 1265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1130);
  __Pyx_GIVEREF(__pyx_tuple__1130);
  __pyx_codeobj__1131 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1130, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 1265, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1131)) __PYX_ERR(0, 1265, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1269
 *             self.copyFrom(ctx)
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 *         def base_prefix(self):
 */
  __pyx_tuple__1132 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1132)) __PYX_ERR(0, 1269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1132);
  __Pyx_GIVEREF(__pyx_tuple__1132);
  __pyx_codeobj__1133 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_IDENT, 1269, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1133)) __PYX_ERR(0, 1269, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1271
 *         def IDENT(self):
 *             return self.getToken(ModelicaParser.IDENT, 0)
 *         def base_prefix(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Base_prefixContext,0)
 * 
 */
  __pyx_tuple__1134 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1134)) __PYX_ERR(0, 1271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1134);
  __Pyx_GIVEREF(__pyx_tuple__1134);
  __pyx_codeobj__1135 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1134, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_base_prefix, 1271, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1135)) __PYX_ERR(0, 1271, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1274
 *             return self.getTypedRuleContext(ModelicaParser.Base_prefixContext,0)
 * 
 *         def name(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)
 * 
 */
  __pyx_tuple__1136 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1136)) __PYX_ERR(0, 1274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1136);
  __Pyx_GIVEREF(__pyx_tuple__1136);
  __pyx_codeobj__1137 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1136, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_name, 1274, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1137)) __PYX_ERR(0, 1274, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1277
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)
 * 
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */
  __pyx_tuple__1138 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1138)) __PYX_ERR(0, 1277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1138);
  __Pyx_GIVEREF(__pyx_tuple__1138);
  __pyx_codeobj__1139 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1138, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_comment, 1277, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1139)) __PYX_ERR(0, 1277, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1280
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 *         def array_subscripts(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Array_subscriptsContext,0)
 * 
 */
  __pyx_tuple__1140 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1140)) __PYX_ERR(0, 1280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1140);
  __Pyx_GIVEREF(__pyx_tuple__1140);
  __pyx_codeobj__1141 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1140, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_array_subscripts, 1280, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1141)) __PYX_ERR(0, 1280, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1283
 *             return self.getTypedRuleContext(ModelicaParser.Array_subscriptsContext,0)
 * 
 *         def class_modification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_modificationContext,0)
 * 
 */
  __pyx_tuple__1142 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1142)) __PYX_ERR(0, 1283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1142);
  __Pyx_GIVEREF(__pyx_tuple__1142);
  __pyx_codeobj__1143 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1142, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_class_modification, 1283, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1143)) __PYX_ERR(0, 1283, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1287
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterClass_spec_base" ):
 *                 listener.enterClass_spec_base(self)
 */
  __pyx_tuple__1144 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1144)) __PYX_ERR(0, 1287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1144);
  __Pyx_GIVEREF(__pyx_tuple__1144);
  __pyx_codeobj__1145 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1144, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 1287, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1145)) __PYX_ERR(0, 1287, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1291
 *                 listener.enterClass_spec_base(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitClass_spec_base" ):
 *                 listener.exitClass_spec_base(self)
 */
  __pyx_tuple__1146 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1146)) __PYX_ERR(0, 1291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1146);
  __Pyx_GIVEREF(__pyx_tuple__1146);
  __pyx_codeobj__1147 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1146, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 1291, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1147)) __PYX_ERR(0, 1291, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1295
 *                 listener.exitClass_spec_base(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitClass_spec_base" ):
 *                 return visitor.visitClass_spec_base(self)
 */
  __pyx_tuple__1148 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1148)) __PYX_ERR(0, 1295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1148);
  __Pyx_GIVEREF(__pyx_tuple__1148);
  __pyx_codeobj__1149 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1148, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 1295, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1149)) __PYX_ERR(0, 1295, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1304
 *     class Class_spec_compContext(Class_specifierContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Class_specifierContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__1150 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__1150)) __PYX_ERR(0, 1304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1150);
  __Pyx_GIVEREF(__pyx_tuple__1150);
  __pyx_codeobj__1151 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1150, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 1304, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1151)) __PYX_ERR(0, 1304, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1308
 *             self.copyFrom(ctx)
 * 
 *         def IDENT(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTokens(ModelicaParser.IDENT)
 */
  __pyx_tuple__1152 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__1152)) __PYX_ERR(0, 1308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1152);
  __Pyx_GIVEREF(__pyx_tuple__1152);
  __pyx_codeobj__1153 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1152, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_IDENT, 1308, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1153)) __PYX_ERR(0, 1308, __pyx_L1_error)
  __pyx_tuple__1154 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__1154)) __PYX_ERR(0, 1308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1154);
  __Pyx_GIVEREF(__pyx_tuple__1154);

  /* "pymola/generated/ModelicaParser.pyx":1313
 *             else:
 *                 return self.getToken(ModelicaParser.IDENT, i)
 *         def string_comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.String_commentContext,0)
 * 
 */
  __pyx_tuple__1155 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1155)) __PYX_ERR(0, 1313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1155);
  __Pyx_GIVEREF(__pyx_tuple__1155);
  __pyx_codeobj__1156 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1155, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_string_comment, 1313, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1156)) __PYX_ERR(0, 1313, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1316
 *             return self.getTypedRuleContext(ModelicaParser.String_commentContext,0)
 * 
 *         def composition(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CompositionContext,0)
 * 
 */
  __pyx_tuple__1157 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1157)) __PYX_ERR(0, 1316, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1157);
  __Pyx_GIVEREF(__pyx_tuple__1157);
  __pyx_codeobj__1158 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1157, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_composition, 1316, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1158)) __PYX_ERR(0, 1316, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1320
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterClass_spec_comp" ):
 *                 listener.enterClass_spec_comp(self)
 */
  __pyx_tuple__1159 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1159)) __PYX_ERR(0, 1320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1159);
  __Pyx_GIVEREF(__pyx_tuple__1159);
  __pyx_codeobj__1160 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1159, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 1320, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1160)) __PYX_ERR(0, 1320, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1324
 *                 listener.enterClass_spec_comp(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitClass_spec_comp" ):
 *                 listener.exitClass_spec_comp(self)
 */
  __pyx_tuple__1161 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1161)) __PYX_ERR(0, 1324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1161);
  __Pyx_GIVEREF(__pyx_tuple__1161);
  __pyx_codeobj__1162 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1161, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 1324, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1162)) __PYX_ERR(0, 1324, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1328
 *                 listener.exitClass_spec_comp(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitClass_spec_comp" ):
 *                 return visitor.visitClass_spec_comp(self)
 */
  __pyx_tuple__1163 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1163)) __PYX_ERR(0, 1328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1163);
  __Pyx_GIVEREF(__pyx_tuple__1163);
  __pyx_codeobj__1164 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1163, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 1328, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1164)) __PYX_ERR(0, 1328, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1337
 *     class Class_spec_extendsContext(Class_specifierContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Class_specifierContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__1165 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__1165)) __PYX_ERR(0, 1337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1165);
  __Pyx_GIVEREF(__pyx_tuple__1165);
  __pyx_codeobj__1166 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1165, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 1337, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1166)) __PYX_ERR(0, 1337, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1341
 *             self.copyFrom(ctx)
 * 
 *         def IDENT(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTokens(ModelicaParser.IDENT)
 */
  __pyx_tuple__1167 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__1167)) __PYX_ERR(0, 1341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1167);
  __Pyx_GIVEREF(__pyx_tuple__1167);
  __pyx_codeobj__1168 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_IDENT, 1341, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1168)) __PYX_ERR(0, 1341, __pyx_L1_error)
  __pyx_tuple__1169 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__1169)) __PYX_ERR(0, 1341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1169);
  __Pyx_GIVEREF(__pyx_tuple__1169);

  /* "pymola/generated/ModelicaParser.pyx":1346
 *             else:
 *                 return self.getToken(ModelicaParser.IDENT, i)
 *         def string_comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.String_commentContext,0)
 * 
 */
  __pyx_tuple__1170 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1170)) __PYX_ERR(0, 1346, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1170);
  __Pyx_GIVEREF(__pyx_tuple__1170);
  __pyx_codeobj__1171 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1170, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_string_comment, 1346, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1171)) __PYX_ERR(0, 1346, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1349
 *             return self.getTypedRuleContext(ModelicaParser.String_commentContext,0)
 * 
 *         def composition(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CompositionContext,0)
 * 
 */
  __pyx_tuple__1172 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1172)) __PYX_ERR(0, 1349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1172);
  __Pyx_GIVEREF(__pyx_tuple__1172);
  __pyx_codeobj__1173 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1172, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_composition, 1349, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1173)) __PYX_ERR(0, 1349, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1352
 *             return self.getTypedRuleContext(ModelicaParser.CompositionContext,0)
 * 
 *         def class_modification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_modificationContext,0)
 * 
 */
  __pyx_tuple__1174 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1174)) __PYX_ERR(0, 1352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1174);
  __Pyx_GIVEREF(__pyx_tuple__1174);
  __pyx_codeobj__1175 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1174, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_class_modification, 1352, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1175)) __PYX_ERR(0, 1352, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1356
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterClass_spec_extends" ):
 *                 listener.enterClass_spec_extends(self)
 */
  __pyx_tuple__1176 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1176)) __PYX_ERR(0, 1356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1176);
  __Pyx_GIVEREF(__pyx_tuple__1176);
  __pyx_codeobj__1177 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1176, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 1356, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1177)) __PYX_ERR(0, 1356, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1360
 *                 listener.enterClass_spec_extends(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitClass_spec_extends" ):
 *                 listener.exitClass_spec_extends(self)
 */
  __pyx_tuple__1178 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1178)) __PYX_ERR(0, 1360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1178);
  __Pyx_GIVEREF(__pyx_tuple__1178);
  __pyx_codeobj__1179 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1178, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 1360, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1179)) __PYX_ERR(0, 1360, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1364
 *                 listener.exitClass_spec_extends(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitClass_spec_extends" ):
 *                 return visitor.visitClass_spec_extends(self)
 */
  __pyx_tuple__1180 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1180)) __PYX_ERR(0, 1364, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1180);
  __Pyx_GIVEREF(__pyx_tuple__1180);
  __pyx_codeobj__1181 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1180, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 1364, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1181)) __PYX_ERR(0, 1364, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1372
 * 
 * 
 *     def class_specifier(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Class_specifierContext(self, self._ctx, self.state)
 */
  __pyx_tuple__1182 = PyTuple_Pack(6, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la_2, __pyx_n_s_la, __pyx_n_s_token, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1182)) __PYX_ERR(0, 1372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1182);
  __Pyx_GIVEREF(__pyx_tuple__1182);
  __pyx_codeobj__1183 = (PyObject*)__Pyx_PyCode_New(1, 0, 6, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1182, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_class_specifier, 1372, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1183)) __PYX_ERR(0, 1372, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1535
 *     class Base_prefixContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1184 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1184)) __PYX_ERR(0, 1535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1184);
  __Pyx_GIVEREF(__pyx_tuple__1184);
  __pyx_codeobj__1185 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1184, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 1535, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1185)) __PYX_ERR(0, 1535, __pyx_L1_error)
  __pyx_tuple__1186 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1186)) __PYX_ERR(0, 1535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1186);
  __Pyx_GIVEREF(__pyx_tuple__1186);

  /* "pymola/generated/ModelicaParser.pyx":1539
 *             self.parser = parser
 * 
 *         def type_prefix(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Type_prefixContext,0)
 * 
 */
  __pyx_tuple__1187 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1187)) __PYX_ERR(0, 1539, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1187);
  __Pyx_GIVEREF(__pyx_tuple__1187);
  __pyx_codeobj__1188 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1187, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_type_prefix, 1539, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1188)) __PYX_ERR(0, 1539, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1543
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_base_prefix
 * 
 */
  __pyx_tuple__1189 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1189)) __PYX_ERR(0, 1543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1189);
  __Pyx_GIVEREF(__pyx_tuple__1189);
  __pyx_codeobj__1190 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1189, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 1543, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1190)) __PYX_ERR(0, 1543, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1546
 *             return ModelicaParser.RULE_base_prefix
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterBase_prefix" ):
 *                 listener.enterBase_prefix(self)
 */
  __pyx_tuple__1191 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1191)) __PYX_ERR(0, 1546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1191);
  __Pyx_GIVEREF(__pyx_tuple__1191);
  __pyx_codeobj__1192 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1191, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 1546, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1192)) __PYX_ERR(0, 1546, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1550
 *                 listener.enterBase_prefix(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitBase_prefix" ):
 *                 listener.exitBase_prefix(self)
 */
  __pyx_tuple__1193 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1193)) __PYX_ERR(0, 1550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1193);
  __Pyx_GIVEREF(__pyx_tuple__1193);
  __pyx_codeobj__1194 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1193, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 1550, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1194)) __PYX_ERR(0, 1550, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1554
 *                 listener.exitBase_prefix(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitBase_prefix" ):
 *                 return visitor.visitBase_prefix(self)
 */
  __pyx_tuple__1195 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1195)) __PYX_ERR(0, 1554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1195);
  __Pyx_GIVEREF(__pyx_tuple__1195);
  __pyx_codeobj__1196 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1195, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 1554, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1196)) __PYX_ERR(0, 1554, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1563
 * 
 * 
 *     def base_prefix(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Base_prefixContext(self, self._ctx, self.state)
 */
  __pyx_tuple__1197 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1197)) __PYX_ERR(0, 1563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1197);
  __Pyx_GIVEREF(__pyx_tuple__1197);
  __pyx_codeobj__1198 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1197, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_base_prefix, 1563, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1198)) __PYX_ERR(0, 1563, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1581
 *     class Enum_listContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1199 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1199)) __PYX_ERR(0, 1581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1199);
  __Pyx_GIVEREF(__pyx_tuple__1199);
  __pyx_codeobj__1200 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1199, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 1581, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1200)) __PYX_ERR(0, 1581, __pyx_L1_error)
  __pyx_tuple__1201 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1201)) __PYX_ERR(0, 1581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1201);
  __Pyx_GIVEREF(__pyx_tuple__1201);

  /* "pymola/generated/ModelicaParser.pyx":1585
 *             self.parser = parser
 * 
 *         def enumeration_literal(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Enumeration_literalContext)
 */
  __pyx_tuple__1202 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__1202)) __PYX_ERR(0, 1585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1202);
  __Pyx_GIVEREF(__pyx_tuple__1202);
  __pyx_codeobj__1203 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1202, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enumeration_literal, 1585, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1203)) __PYX_ERR(0, 1585, __pyx_L1_error)
  __pyx_tuple__1204 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__1204)) __PYX_ERR(0, 1585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1204);
  __Pyx_GIVEREF(__pyx_tuple__1204);

  /* "pymola/generated/ModelicaParser.pyx":1592
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_enum_list
 * 
 */
  __pyx_tuple__1205 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1205)) __PYX_ERR(0, 1592, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1205);
  __Pyx_GIVEREF(__pyx_tuple__1205);
  __pyx_codeobj__1206 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1205, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 1592, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1206)) __PYX_ERR(0, 1592, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1595
 *             return ModelicaParser.RULE_enum_list
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterEnum_list" ):
 *                 listener.enterEnum_list(self)
 */
  __pyx_tuple__1207 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1207)) __PYX_ERR(0, 1595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1207);
  __Pyx_GIVEREF(__pyx_tuple__1207);
  __pyx_codeobj__1208 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1207, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 1595, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1208)) __PYX_ERR(0, 1595, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1599
 *                 listener.enterEnum_list(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitEnum_list" ):
 *                 listener.exitEnum_list(self)
 */
  __pyx_tuple__1209 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1209)) __PYX_ERR(0, 1599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1209);
  __Pyx_GIVEREF(__pyx_tuple__1209);
  __pyx_codeobj__1210 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 1599, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1210)) __PYX_ERR(0, 1599, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1603
 *                 listener.exitEnum_list(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitEnum_list" ):
 *                 return visitor.visitEnum_list(self)
 */
  __pyx_tuple__1211 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1211)) __PYX_ERR(0, 1603, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1211);
  __Pyx_GIVEREF(__pyx_tuple__1211);
  __pyx_codeobj__1212 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1211, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 1603, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1212)) __PYX_ERR(0, 1603, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1612
 * 
 * 
 *     def enum_list(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Enum_listContext(self, self._ctx, self.state)
 */
  __pyx_tuple__1213 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1213)) __PYX_ERR(0, 1612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1213);
  __Pyx_GIVEREF(__pyx_tuple__1213);
  __pyx_codeobj__1214 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1213, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enum_list, 1612, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1214)) __PYX_ERR(0, 1612, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1643
 *     class Enumeration_literalContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1215 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1215)) __PYX_ERR(0, 1643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1215);
  __Pyx_GIVEREF(__pyx_tuple__1215);
  __pyx_codeobj__1216 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1215, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 1643, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1216)) __PYX_ERR(0, 1643, __pyx_L1_error)
  __pyx_tuple__1217 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1217)) __PYX_ERR(0, 1643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1217);
  __Pyx_GIVEREF(__pyx_tuple__1217);

  /* "pymola/generated/ModelicaParser.pyx":1647
 *             self.parser = parser
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 */
  __pyx_tuple__1218 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1218)) __PYX_ERR(0, 1647, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1218);
  __Pyx_GIVEREF(__pyx_tuple__1218);
  __pyx_codeobj__1219 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1218, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_IDENT, 1647, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1219)) __PYX_ERR(0, 1647, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1650
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */
  __pyx_tuple__1220 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1220)) __PYX_ERR(0, 1650, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1220);
  __Pyx_GIVEREF(__pyx_tuple__1220);
  __pyx_codeobj__1221 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1220, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_comment, 1650, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1221)) __PYX_ERR(0, 1650, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1654
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_enumeration_literal
 * 
 */
  __pyx_tuple__1222 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1222)) __PYX_ERR(0, 1654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1222);
  __Pyx_GIVEREF(__pyx_tuple__1222);
  __pyx_codeobj__1223 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1222, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 1654, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1223)) __PYX_ERR(0, 1654, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1657
 *             return ModelicaParser.RULE_enumeration_literal
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterEnumeration_literal" ):
 *                 listener.enterEnumeration_literal(self)
 */
  __pyx_tuple__1224 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1224)) __PYX_ERR(0, 1657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1224);
  __Pyx_GIVEREF(__pyx_tuple__1224);
  __pyx_codeobj__1225 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1224, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 1657, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1225)) __PYX_ERR(0, 1657, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1661
 *                 listener.enterEnumeration_literal(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitEnumeration_literal" ):
 *                 listener.exitEnumeration_literal(self)
 */
  __pyx_tuple__1226 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1226)) __PYX_ERR(0, 1661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1226);
  __Pyx_GIVEREF(__pyx_tuple__1226);
  __pyx_codeobj__1227 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1226, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 1661, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1227)) __PYX_ERR(0, 1661, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1665
 *                 listener.exitEnumeration_literal(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitEnumeration_literal" ):
 *                 return visitor.visitEnumeration_literal(self)
 */
  __pyx_tuple__1228 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1228)) __PYX_ERR(0, 1665, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1228);
  __Pyx_GIVEREF(__pyx_tuple__1228);
  __pyx_codeobj__1229 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1228, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 1665, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1229)) __PYX_ERR(0, 1665, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1674
 * 
 * 
 *     def enumeration_literal(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Enumeration_literalContext(self, self._ctx, self.state)
 */
  __pyx_tuple__1230 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1230)) __PYX_ERR(0, 1674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1230);
  __Pyx_GIVEREF(__pyx_tuple__1230);
  __pyx_codeobj__1231 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1230, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enumeration_literal, 1674, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1231)) __PYX_ERR(0, 1674, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1694
 *     class CompositionContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1232 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1232)) __PYX_ERR(0, 1694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1232);
  __Pyx_GIVEREF(__pyx_tuple__1232);
  __pyx_codeobj__1233 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1232, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 1694, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1233)) __PYX_ERR(0, 1694, __pyx_L1_error)
  __pyx_tuple__1234 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1234)) __PYX_ERR(0, 1694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1234);
  __Pyx_GIVEREF(__pyx_tuple__1234);

  /* "pymola/generated/ModelicaParser.pyx":1703
 *             self.comp_annotation = None # AnnotationContext
 * 
 *         def element_list(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Element_listContext)
 */
  __pyx_tuple__1235 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__1235)) __PYX_ERR(0, 1703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1235);
  __Pyx_GIVEREF(__pyx_tuple__1235);
  __pyx_codeobj__1236 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1235, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_element_list, 1703, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1236)) __PYX_ERR(0, 1703, __pyx_L1_error)
  __pyx_tuple__1237 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__1237)) __PYX_ERR(0, 1703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1237);
  __Pyx_GIVEREF(__pyx_tuple__1237);

  /* "pymola/generated/ModelicaParser.pyx":1710
 * 
 * 
 *         def equation_section(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Equation_sectionContext)
 */
  __pyx_tuple__1238 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__1238)) __PYX_ERR(0, 1710, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1238);
  __Pyx_GIVEREF(__pyx_tuple__1238);
  __pyx_codeobj__1239 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1238, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_equation_section, 1710, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1239)) __PYX_ERR(0, 1710, __pyx_L1_error)
  __pyx_tuple__1240 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__1240)) __PYX_ERR(0, 1710, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1240);
  __Pyx_GIVEREF(__pyx_tuple__1240);

  /* "pymola/generated/ModelicaParser.pyx":1717
 * 
 * 
 *         def algorithm_section(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Algorithm_sectionContext)
 */
  __pyx_tuple__1241 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__1241)) __PYX_ERR(0, 1717, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1241);
  __Pyx_GIVEREF(__pyx_tuple__1241);
  __pyx_codeobj__1242 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1241, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_algorithm_section, 1717, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1242)) __PYX_ERR(0, 1717, __pyx_L1_error)
  __pyx_tuple__1243 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__1243)) __PYX_ERR(0, 1717, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1243);
  __Pyx_GIVEREF(__pyx_tuple__1243);

  /* "pymola/generated/ModelicaParser.pyx":1724
 * 
 * 
 *         def annotation(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.AnnotationContext)
 */
  __pyx_tuple__1244 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__1244)) __PYX_ERR(0, 1724, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1244);
  __Pyx_GIVEREF(__pyx_tuple__1244);
  __pyx_codeobj__1245 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1244, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_annotation, 1724, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1245)) __PYX_ERR(0, 1724, __pyx_L1_error)
  __pyx_tuple__1246 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__1246)) __PYX_ERR(0, 1724, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1246);
  __Pyx_GIVEREF(__pyx_tuple__1246);

  /* "pymola/generated/ModelicaParser.pyx":1731
 * 
 * 
 *         def language_specification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Language_specificationContext,0)
 * 
 */
  __pyx_tuple__1247 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1247)) __PYX_ERR(0, 1731, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1247);
  __Pyx_GIVEREF(__pyx_tuple__1247);
  __pyx_codeobj__1248 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1247, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_language_specification, 1731, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1248)) __PYX_ERR(0, 1731, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1735
 * 
 * 
 *         def external_function_call(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.External_function_callContext,0)
 * 
 */
  __pyx_tuple__1249 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1249)) __PYX_ERR(0, 1735, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1249);
  __Pyx_GIVEREF(__pyx_tuple__1249);
  __pyx_codeobj__1250 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1249, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_external_function_call, 1735, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1250)) __PYX_ERR(0, 1735, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1739
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_composition
 * 
 */
  __pyx_tuple__1251 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1251)) __PYX_ERR(0, 1739, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1251);
  __Pyx_GIVEREF(__pyx_tuple__1251);
  __pyx_codeobj__1252 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1251, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 1739, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1252)) __PYX_ERR(0, 1739, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1742
 *             return ModelicaParser.RULE_composition
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterComposition" ):
 *                 listener.enterComposition(self)
 */
  __pyx_tuple__1253 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1253)) __PYX_ERR(0, 1742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1253);
  __Pyx_GIVEREF(__pyx_tuple__1253);
  __pyx_codeobj__1254 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1253, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 1742, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1254)) __PYX_ERR(0, 1742, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1746
 *                 listener.enterComposition(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitComposition" ):
 *                 listener.exitComposition(self)
 */
  __pyx_tuple__1255 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1255)) __PYX_ERR(0, 1746, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1255);
  __Pyx_GIVEREF(__pyx_tuple__1255);
  __pyx_codeobj__1256 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1255, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 1746, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1256)) __PYX_ERR(0, 1746, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1750
 *                 listener.exitComposition(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitComposition" ):
 *                 return visitor.visitComposition(self)
 */
  __pyx_tuple__1257 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1257)) __PYX_ERR(0, 1750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1257);
  __Pyx_GIVEREF(__pyx_tuple__1257);
  __pyx_codeobj__1258 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1257, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 1750, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1258)) __PYX_ERR(0, 1750, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1759
 * 
 * 
 *     def composition(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.CompositionContext(self, self._ctx, self.state)
 */
  __pyx_tuple__1259 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la, __pyx_n_s_la_2, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1259)) __PYX_ERR(0, 1759, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1259);
  __Pyx_GIVEREF(__pyx_tuple__1259);
  __pyx_codeobj__1260 = (PyObject*)__Pyx_PyCode_New(1, 0, 5, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1259, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_composition, 1759, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1260)) __PYX_ERR(0, 1759, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1858
 *     class Language_specificationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1261 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1261)) __PYX_ERR(0, 1858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1261);
  __Pyx_GIVEREF(__pyx_tuple__1261);
  __pyx_codeobj__1262 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1261, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 1858, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1262)) __PYX_ERR(0, 1858, __pyx_L1_error)
  __pyx_tuple__1263 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1263)) __PYX_ERR(0, 1858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1263);
  __Pyx_GIVEREF(__pyx_tuple__1263);

  /* "pymola/generated/ModelicaParser.pyx":1862
 *             self.parser = parser
 * 
 *         def STRING(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.STRING, 0)
 * 
 */
  __pyx_tuple__1264 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1264)) __PYX_ERR(0, 1862, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1264);
  __Pyx_GIVEREF(__pyx_tuple__1264);
  __pyx_codeobj__1265 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1264, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_STRING, 1862, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1265)) __PYX_ERR(0, 1862, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1865
 *             return self.getToken(ModelicaParser.STRING, 0)
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_language_specification
 * 
 */
  __pyx_tuple__1266 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1266)) __PYX_ERR(0, 1865, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1266);
  __Pyx_GIVEREF(__pyx_tuple__1266);
  __pyx_codeobj__1267 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1266, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 1865, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1267)) __PYX_ERR(0, 1865, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1868
 *             return ModelicaParser.RULE_language_specification
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterLanguage_specification" ):
 *                 listener.enterLanguage_specification(self)
 */
  __pyx_tuple__1268 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1268)) __PYX_ERR(0, 1868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1268);
  __Pyx_GIVEREF(__pyx_tuple__1268);
  __pyx_codeobj__1269 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1268, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 1868, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1269)) __PYX_ERR(0, 1868, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1872
 *                 listener.enterLanguage_specification(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitLanguage_specification" ):
 *                 listener.exitLanguage_specification(self)
 */
  __pyx_tuple__1270 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1270)) __PYX_ERR(0, 1872, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1270);
  __Pyx_GIVEREF(__pyx_tuple__1270);
  __pyx_codeobj__1271 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1270, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 1872, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1271)) __PYX_ERR(0, 1872, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1876
 *                 listener.exitLanguage_specification(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitLanguage_specification" ):
 *                 return visitor.visitLanguage_specification(self)
 */
  __pyx_tuple__1272 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1272)) __PYX_ERR(0, 1876, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1272);
  __Pyx_GIVEREF(__pyx_tuple__1272);
  __pyx_codeobj__1273 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1272, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 1876, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1273)) __PYX_ERR(0, 1876, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1885
 * 
 * 
 *     def language_specification(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Language_specificationContext(self, self._ctx, self.state)
 */
  __pyx_tuple__1274 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1274)) __PYX_ERR(0, 1885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1274);
  __Pyx_GIVEREF(__pyx_tuple__1274);
  __pyx_codeobj__1275 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1274, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_language_specification, 1885, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1275)) __PYX_ERR(0, 1885, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1903
 *     class External_function_callContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1276 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1276)) __PYX_ERR(0, 1903, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1276);
  __Pyx_GIVEREF(__pyx_tuple__1276);
  __pyx_codeobj__1277 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1276, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 1903, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1277)) __PYX_ERR(0, 1903, __pyx_L1_error)
  __pyx_tuple__1278 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1278)) __PYX_ERR(0, 1903, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1278);
  __Pyx_GIVEREF(__pyx_tuple__1278);

  /* "pymola/generated/ModelicaParser.pyx":1907
 *             self.parser = parser
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 */
  __pyx_tuple__1279 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1279)) __PYX_ERR(0, 1907, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1279);
  __Pyx_GIVEREF(__pyx_tuple__1279);
  __pyx_codeobj__1280 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1279, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_IDENT, 1907, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1280)) __PYX_ERR(0, 1907, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1910
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 *         def component_reference(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 */
  __pyx_tuple__1281 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1281)) __PYX_ERR(0, 1910, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1281);
  __Pyx_GIVEREF(__pyx_tuple__1281);
  __pyx_codeobj__1282 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1281, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_component_reference, 1910, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1282)) __PYX_ERR(0, 1910, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1914
 * 
 * 
 *         def expression_list(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Expression_listContext,0)
 * 
 */
  __pyx_tuple__1283 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1283)) __PYX_ERR(0, 1914, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1283);
  __Pyx_GIVEREF(__pyx_tuple__1283);
  __pyx_codeobj__1284 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1283, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_expression_list, 1914, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1284)) __PYX_ERR(0, 1914, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1918
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_external_function_call
 * 
 */
  __pyx_tuple__1285 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1285)) __PYX_ERR(0, 1918, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1285);
  __Pyx_GIVEREF(__pyx_tuple__1285);
  __pyx_codeobj__1286 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1285, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 1918, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1286)) __PYX_ERR(0, 1918, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1921
 *             return ModelicaParser.RULE_external_function_call
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExternal_function_call" ):
 *                 listener.enterExternal_function_call(self)
 */
  __pyx_tuple__1287 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1287)) __PYX_ERR(0, 1921, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1287);
  __Pyx_GIVEREF(__pyx_tuple__1287);
  __pyx_codeobj__1288 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1287, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 1921, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1288)) __PYX_ERR(0, 1921, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1925
 *                 listener.enterExternal_function_call(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExternal_function_call" ):
 *                 listener.exitExternal_function_call(self)
 */
  __pyx_tuple__1289 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1289)) __PYX_ERR(0, 1925, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1289);
  __Pyx_GIVEREF(__pyx_tuple__1289);
  __pyx_codeobj__1290 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1289, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 1925, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1290)) __PYX_ERR(0, 1925, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1929
 *                 listener.exitExternal_function_call(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExternal_function_call" ):
 *                 return visitor.visitExternal_function_call(self)
 */
  __pyx_tuple__1291 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1291)) __PYX_ERR(0, 1929, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1291);
  __Pyx_GIVEREF(__pyx_tuple__1291);
  __pyx_codeobj__1292 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1291, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 1929, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1292)) __PYX_ERR(0, 1929, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1938
 * 
 * 
 *     def external_function_call(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.External_function_callContext(self, self._ctx, self.state)
 */
  __pyx_tuple__1293 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la_2, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1293)) __PYX_ERR(0, 1938, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1293);
  __Pyx_GIVEREF(__pyx_tuple__1293);
  __pyx_codeobj__1294 = (PyObject*)__Pyx_PyCode_New(1, 0, 5, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1293, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_external_function_call, 1938, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1294)) __PYX_ERR(0, 1938, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1979
 *     class Element_listContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1295 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1295)) __PYX_ERR(0, 1979, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1295);
  __Pyx_GIVEREF(__pyx_tuple__1295);
  __pyx_codeobj__1296 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1295, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 1979, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1296)) __PYX_ERR(0, 1979, __pyx_L1_error)
  __pyx_tuple__1297 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1297)) __PYX_ERR(0, 1979, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1297);
  __Pyx_GIVEREF(__pyx_tuple__1297);

  /* "pymola/generated/ModelicaParser.pyx":1983
 *             self.parser = parser
 * 
 *         def element(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ElementContext)
 */
  __pyx_tuple__1298 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__1298)) __PYX_ERR(0, 1983, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1298);
  __Pyx_GIVEREF(__pyx_tuple__1298);
  __pyx_codeobj__1299 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1298, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_element, 1983, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1299)) __PYX_ERR(0, 1983, __pyx_L1_error)
  __pyx_tuple__1300 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__1300)) __PYX_ERR(0, 1983, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1300);
  __Pyx_GIVEREF(__pyx_tuple__1300);

  /* "pymola/generated/ModelicaParser.pyx":1990
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_element_list
 * 
 */
  __pyx_tuple__1301 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1301)) __PYX_ERR(0, 1990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1301);
  __Pyx_GIVEREF(__pyx_tuple__1301);
  __pyx_codeobj__1302 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1301, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 1990, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1302)) __PYX_ERR(0, 1990, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1993
 *             return ModelicaParser.RULE_element_list
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterElement_list" ):
 *                 listener.enterElement_list(self)
 */
  __pyx_tuple__1303 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1303)) __PYX_ERR(0, 1993, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1303);
  __Pyx_GIVEREF(__pyx_tuple__1303);
  __pyx_codeobj__1304 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1303, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 1993, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1304)) __PYX_ERR(0, 1993, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":1997
 *                 listener.enterElement_list(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitElement_list" ):
 *                 listener.exitElement_list(self)
 */
  __pyx_tuple__1305 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1305)) __PYX_ERR(0, 1997, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1305);
  __Pyx_GIVEREF(__pyx_tuple__1305);
  __pyx_codeobj__1306 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1305, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 1997, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1306)) __PYX_ERR(0, 1997, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2001
 *                 listener.exitElement_list(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitElement_list" ):
 *                 return visitor.visitElement_list(self)
 */
  __pyx_tuple__1307 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1307)) __PYX_ERR(0, 2001, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1307);
  __Pyx_GIVEREF(__pyx_tuple__1307);
  __pyx_codeobj__1308 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1307, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 2001, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1308)) __PYX_ERR(0, 2001, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2010
 * 
 * 
 *     def element_list(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Element_listContext(self, self._ctx, self.state)
 */
  __pyx_tuple__1309 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1309)) __PYX_ERR(0, 2010, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1309);
  __Pyx_GIVEREF(__pyx_tuple__1309);
  __pyx_codeobj__1310 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1309, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_element_list, 2010, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1310)) __PYX_ERR(0, 2010, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2039
 *     class ElementContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1311 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1311)) __PYX_ERR(0, 2039, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1311);
  __Pyx_GIVEREF(__pyx_tuple__1311);
  __pyx_codeobj__1312 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1311, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 2039, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1312)) __PYX_ERR(0, 2039, __pyx_L1_error)
  __pyx_tuple__1313 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1313)) __PYX_ERR(0, 2039, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1313);
  __Pyx_GIVEREF(__pyx_tuple__1313);

  /* "pymola/generated/ModelicaParser.pyx":2043
 *             self.parser = parser
 * 
 *         def import_clause(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Import_clauseContext,0)
 * 
 */
  __pyx_tuple__1314 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1314)) __PYX_ERR(0, 2043, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1314);
  __Pyx_GIVEREF(__pyx_tuple__1314);
  __pyx_codeobj__1315 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1314, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_import_clause, 2043, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1315)) __PYX_ERR(0, 2043, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2047
 * 
 * 
 *         def extends_clause(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Extends_clauseContext,0)
 * 
 */
  __pyx_tuple__1316 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1316)) __PYX_ERR(0, 2047, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1316);
  __Pyx_GIVEREF(__pyx_tuple__1316);
  __pyx_codeobj__1317 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1316, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_extends_clause, 2047, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1317)) __PYX_ERR(0, 2047, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2051
 * 
 * 
 *         def regular_element(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Regular_elementContext,0)
 * 
 */
  __pyx_tuple__1318 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1318)) __PYX_ERR(0, 2051, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1318);
  __Pyx_GIVEREF(__pyx_tuple__1318);
  __pyx_codeobj__1319 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1318, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_regular_element, 2051, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1319)) __PYX_ERR(0, 2051, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2055
 * 
 * 
 *         def replaceable_element(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Replaceable_elementContext,0)
 * 
 */
  __pyx_tuple__1320 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1320)) __PYX_ERR(0, 2055, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1320);
  __Pyx_GIVEREF(__pyx_tuple__1320);
  __pyx_codeobj__1321 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1320, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_replaceable_element, 2055, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1321)) __PYX_ERR(0, 2055, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2059
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_element
 * 
 */
  __pyx_tuple__1322 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1322)) __PYX_ERR(0, 2059, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1322);
  __Pyx_GIVEREF(__pyx_tuple__1322);
  __pyx_codeobj__1323 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1322, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 2059, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1323)) __PYX_ERR(0, 2059, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2062
 *             return ModelicaParser.RULE_element
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterElement" ):
 *                 listener.enterElement(self)
 */
  __pyx_tuple__1324 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1324)) __PYX_ERR(0, 2062, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1324);
  __Pyx_GIVEREF(__pyx_tuple__1324);
  __pyx_codeobj__1325 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1324, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 2062, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1325)) __PYX_ERR(0, 2062, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2066
 *                 listener.enterElement(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitElement" ):
 *                 listener.exitElement(self)
 */
  __pyx_tuple__1326 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1326)) __PYX_ERR(0, 2066, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1326);
  __Pyx_GIVEREF(__pyx_tuple__1326);
  __pyx_codeobj__1327 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1326, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 2066, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1327)) __PYX_ERR(0, 2066, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2070
 *                 listener.exitElement(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitElement" ):
 *                 return visitor.visitElement(self)
 */
  __pyx_tuple__1328 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1328)) __PYX_ERR(0, 2070, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1328);
  __Pyx_GIVEREF(__pyx_tuple__1328);
  __pyx_codeobj__1329 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1328, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 2070, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1329)) __PYX_ERR(0, 2070, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2079
 * 
 * 
 *     def element(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.ElementContext(self, self._ctx, self.state)
 */
  __pyx_tuple__1330 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la_2, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1330)) __PYX_ERR(0, 2079, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1330);
  __Pyx_GIVEREF(__pyx_tuple__1330);
  __pyx_codeobj__1331 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1330, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_element, 2079, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1331)) __PYX_ERR(0, 2079, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2122
 *     class Regular_elementContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1332 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1332)) __PYX_ERR(0, 2122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1332);
  __Pyx_GIVEREF(__pyx_tuple__1332);
  __pyx_codeobj__1333 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1332, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 2122, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1333)) __PYX_ERR(0, 2122, __pyx_L1_error)
  __pyx_tuple__1334 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1334)) __PYX_ERR(0, 2122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1334);
  __Pyx_GIVEREF(__pyx_tuple__1334);

  /* "pymola/generated/ModelicaParser.pyx":2128
 *             self.comp_elem = None # Component_clauseContext
 * 
 *         def REDECLARE(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.REDECLARE, 0)
 * 
 */
  __pyx_tuple__1335 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1335)) __PYX_ERR(0, 2128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1335);
  __Pyx_GIVEREF(__pyx_tuple__1335);
  __pyx_codeobj__1336 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1335, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_REDECLARE, 2128, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1336)) __PYX_ERR(0, 2128, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2131
 *             return self.getToken(ModelicaParser.REDECLARE, 0)
 * 
 *         def FINAL(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.FINAL, 0)
 * 
 */
  __pyx_tuple__1337 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1337)) __PYX_ERR(0, 2131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1337);
  __Pyx_GIVEREF(__pyx_tuple__1337);
  __pyx_codeobj__1338 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1337, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_FINAL, 2131, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1338)) __PYX_ERR(0, 2131, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2134
 *             return self.getToken(ModelicaParser.FINAL, 0)
 * 
 *         def INNER(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.INNER, 0)
 * 
 */
  __pyx_tuple__1339 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1339)) __PYX_ERR(0, 2134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1339);
  __Pyx_GIVEREF(__pyx_tuple__1339);
  __pyx_codeobj__1340 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1339, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_INNER, 2134, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1340)) __PYX_ERR(0, 2134, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2137
 *             return self.getToken(ModelicaParser.INNER, 0)
 * 
 *         def OUTER(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.OUTER, 0)
 * 
 */
  __pyx_tuple__1341 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1341)) __PYX_ERR(0, 2137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1341);
  __Pyx_GIVEREF(__pyx_tuple__1341);
  __pyx_codeobj__1342 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1341, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_OUTER, 2137, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1342)) __PYX_ERR(0, 2137, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2140
 *             return self.getToken(ModelicaParser.OUTER, 0)
 * 
 *         def class_definition(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_definitionContext,0)
 * 
 */
  __pyx_tuple__1343 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1343)) __PYX_ERR(0, 2140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1343);
  __Pyx_GIVEREF(__pyx_tuple__1343);
  __pyx_codeobj__1344 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1343, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_class_definition, 2140, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1344)) __PYX_ERR(0, 2140, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2144
 * 
 * 
 *         def component_clause(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_clauseContext,0)
 * 
 */
  __pyx_tuple__1345 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1345)) __PYX_ERR(0, 2144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1345);
  __Pyx_GIVEREF(__pyx_tuple__1345);
  __pyx_codeobj__1346 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1345, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_component_clause, 2144, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1346)) __PYX_ERR(0, 2144, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2148
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_regular_element
 * 
 */
  __pyx_tuple__1347 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1347)) __PYX_ERR(0, 2148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1347);
  __Pyx_GIVEREF(__pyx_tuple__1347);
  __pyx_codeobj__1348 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1347, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 2148, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1348)) __PYX_ERR(0, 2148, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2151
 *             return ModelicaParser.RULE_regular_element
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterRegular_element" ):
 *                 listener.enterRegular_element(self)
 */
  __pyx_tuple__1349 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1349)) __PYX_ERR(0, 2151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1349);
  __Pyx_GIVEREF(__pyx_tuple__1349);
  __pyx_codeobj__1350 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1349, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 2151, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1350)) __PYX_ERR(0, 2151, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2155
 *                 listener.enterRegular_element(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitRegular_element" ):
 *                 listener.exitRegular_element(self)
 */
  __pyx_tuple__1351 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1351)) __PYX_ERR(0, 2155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1351);
  __Pyx_GIVEREF(__pyx_tuple__1351);
  __pyx_codeobj__1352 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1351, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 2155, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1352)) __PYX_ERR(0, 2155, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2159
 *                 listener.exitRegular_element(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitRegular_element" ):
 *                 return visitor.visitRegular_element(self)
 */
  __pyx_tuple__1353 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1353)) __PYX_ERR(0, 2159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1353);
  __Pyx_GIVEREF(__pyx_tuple__1353);
  __pyx_codeobj__1354 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1353, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 2159, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1354)) __PYX_ERR(0, 2159, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2168
 * 
 * 
 *     def regular_element(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Regular_elementContext(self, self._ctx, self.state)
 */
  __pyx_tuple__1355 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la, __pyx_n_s_token, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1355)) __PYX_ERR(0, 2168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1355);
  __Pyx_GIVEREF(__pyx_tuple__1355);
  __pyx_codeobj__1356 = (PyObject*)__Pyx_PyCode_New(1, 0, 5, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1355, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_regular_element, 2168, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1356)) __PYX_ERR(0, 2168, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2231
 *     class Replaceable_elementContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1357 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1357)) __PYX_ERR(0, 2231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1357);
  __Pyx_GIVEREF(__pyx_tuple__1357);
  __pyx_codeobj__1358 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1357, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 2231, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1358)) __PYX_ERR(0, 2231, __pyx_L1_error)
  __pyx_tuple__1359 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1359)) __PYX_ERR(0, 2231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1359);
  __Pyx_GIVEREF(__pyx_tuple__1359);

  /* "pymola/generated/ModelicaParser.pyx":2237
 *             self.comp_elem = None # Component_clauseContext
 * 
 *         def REDECLARE(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.REDECLARE, 0)
 * 
 */
  __pyx_tuple__1360 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1360)) __PYX_ERR(0, 2237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1360);
  __Pyx_GIVEREF(__pyx_tuple__1360);
  __pyx_codeobj__1361 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1360, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_REDECLARE, 2237, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1361)) __PYX_ERR(0, 2237, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2240
 *             return self.getToken(ModelicaParser.REDECLARE, 0)
 * 
 *         def FINAL(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.FINAL, 0)
 * 
 */
  __pyx_tuple__1362 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1362)) __PYX_ERR(0, 2240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1362);
  __Pyx_GIVEREF(__pyx_tuple__1362);
  __pyx_codeobj__1363 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1362, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_FINAL, 2240, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1363)) __PYX_ERR(0, 2240, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2243
 *             return self.getToken(ModelicaParser.FINAL, 0)
 * 
 *         def INNER(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.INNER, 0)
 * 
 */
  __pyx_tuple__1364 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1364)) __PYX_ERR(0, 2243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1364);
  __Pyx_GIVEREF(__pyx_tuple__1364);
  __pyx_codeobj__1365 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1364, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_INNER, 2243, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1365)) __PYX_ERR(0, 2243, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2246
 *             return self.getToken(ModelicaParser.INNER, 0)
 * 
 *         def OUTER(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.OUTER, 0)
 * 
 */
  __pyx_tuple__1366 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1366)) __PYX_ERR(0, 2246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1366);
  __Pyx_GIVEREF(__pyx_tuple__1366);
  __pyx_codeobj__1367 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1366, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_OUTER, 2246, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1367)) __PYX_ERR(0, 2246, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2249
 *             return self.getToken(ModelicaParser.OUTER, 0)
 * 
 *         def class_definition(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_definitionContext,0)
 * 
 */
  __pyx_tuple__1368 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1368)) __PYX_ERR(0, 2249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1368);
  __Pyx_GIVEREF(__pyx_tuple__1368);
  __pyx_codeobj__1369 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1368, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_class_definition, 2249, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1369)) __PYX_ERR(0, 2249, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2253
 * 
 * 
 *         def component_clause(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_clauseContext,0)
 * 
 */
  __pyx_tuple__1370 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1370)) __PYX_ERR(0, 2253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1370);
  __Pyx_GIVEREF(__pyx_tuple__1370);
  __pyx_codeobj__1371 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1370, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_component_clause, 2253, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1371)) __PYX_ERR(0, 2253, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2257
 * 
 * 
 *         def constraining_clause(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Constraining_clauseContext,0)
 * 
 */
  __pyx_tuple__1372 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1372)) __PYX_ERR(0, 2257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1372);
  __Pyx_GIVEREF(__pyx_tuple__1372);
  __pyx_codeobj__1373 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1372, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_constraining_clause, 2257, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1373)) __PYX_ERR(0, 2257, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2261
 * 
 * 
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */
  __pyx_tuple__1374 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1374)) __PYX_ERR(0, 2261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1374);
  __Pyx_GIVEREF(__pyx_tuple__1374);
  __pyx_codeobj__1375 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1374, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_comment, 2261, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1375)) __PYX_ERR(0, 2261, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2265
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_replaceable_element
 * 
 */
  __pyx_tuple__1376 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1376)) __PYX_ERR(0, 2265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1376);
  __Pyx_GIVEREF(__pyx_tuple__1376);
  __pyx_codeobj__1377 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1376, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 2265, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1377)) __PYX_ERR(0, 2265, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2268
 *             return ModelicaParser.RULE_replaceable_element
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterReplaceable_element" ):
 *                 listener.enterReplaceable_element(self)
 */
  __pyx_tuple__1378 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1378)) __PYX_ERR(0, 2268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1378);
  __Pyx_GIVEREF(__pyx_tuple__1378);
  __pyx_codeobj__1379 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1378, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 2268, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1379)) __PYX_ERR(0, 2268, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2272
 *                 listener.enterReplaceable_element(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitReplaceable_element" ):
 *                 listener.exitReplaceable_element(self)
 */
  __pyx_tuple__1380 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1380)) __PYX_ERR(0, 2272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1380);
  __Pyx_GIVEREF(__pyx_tuple__1380);
  __pyx_codeobj__1381 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1380, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 2272, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1381)) __PYX_ERR(0, 2272, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2276
 *                 listener.exitReplaceable_element(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitReplaceable_element" ):
 *                 return visitor.visitReplaceable_element(self)
 */
  __pyx_tuple__1382 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1382)) __PYX_ERR(0, 2276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1382);
  __Pyx_GIVEREF(__pyx_tuple__1382);
  __pyx_codeobj__1383 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1382, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 2276, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1383)) __PYX_ERR(0, 2276, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2285
 * 
 * 
 *     def replaceable_element(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Replaceable_elementContext(self, self._ctx, self.state)
 */
  __pyx_tuple__1384 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la, __pyx_n_s_token, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1384)) __PYX_ERR(0, 2285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1384);
  __Pyx_GIVEREF(__pyx_tuple__1384);
  __pyx_codeobj__1385 = (PyObject*)__Pyx_PyCode_New(1, 0, 5, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1384, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_replaceable_element, 2285, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1385)) __PYX_ERR(0, 2285, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2360
 *     class Import_clauseContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1386 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1386)) __PYX_ERR(0, 2360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1386);
  __Pyx_GIVEREF(__pyx_tuple__1386);
  __pyx_codeobj__1387 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1386, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 2360, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1387)) __PYX_ERR(0, 2360, __pyx_L1_error)
  __pyx_tuple__1388 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1388)) __PYX_ERR(0, 2360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1388);
  __Pyx_GIVEREF(__pyx_tuple__1388);

  /* "pymola/generated/ModelicaParser.pyx":2364
 *             self.parser = parser
 * 
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */
  __pyx_tuple__1389 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1389)) __PYX_ERR(0, 2364, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1389);
  __Pyx_GIVEREF(__pyx_tuple__1389);
  __pyx_codeobj__1390 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1389, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_comment, 2364, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1390)) __PYX_ERR(0, 2364, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2368
 * 
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 */
  __pyx_tuple__1391 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1391)) __PYX_ERR(0, 2368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1391);
  __Pyx_GIVEREF(__pyx_tuple__1391);
  __pyx_codeobj__1392 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1391, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_IDENT, 2368, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1392)) __PYX_ERR(0, 2368, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2371
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 *         def component_reference(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 */
  __pyx_tuple__1393 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1393)) __PYX_ERR(0, 2371, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1393);
  __Pyx_GIVEREF(__pyx_tuple__1393);
  __pyx_codeobj__1394 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1393, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_component_reference, 2371, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1394)) __PYX_ERR(0, 2371, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2375
 * 
 * 
 *         def import_list(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Import_listContext,0)
 * 
 */
  __pyx_tuple__1395 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1395)) __PYX_ERR(0, 2375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1395);
  __Pyx_GIVEREF(__pyx_tuple__1395);
  __pyx_codeobj__1396 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1395, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_import_list, 2375, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1396)) __PYX_ERR(0, 2375, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2379
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_import_clause
 * 
 */
  __pyx_tuple__1397 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1397)) __PYX_ERR(0, 2379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1397);
  __Pyx_GIVEREF(__pyx_tuple__1397);
  __pyx_codeobj__1398 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1397, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 2379, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1398)) __PYX_ERR(0, 2379, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2382
 *             return ModelicaParser.RULE_import_clause
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterImport_clause" ):
 *                 listener.enterImport_clause(self)
 */
  __pyx_tuple__1399 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1399)) __PYX_ERR(0, 2382, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1399);
  __Pyx_GIVEREF(__pyx_tuple__1399);
  __pyx_codeobj__1400 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1399, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 2382, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1400)) __PYX_ERR(0, 2382, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2386
 *                 listener.enterImport_clause(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitImport_clause" ):
 *                 listener.exitImport_clause(self)
 */
  __pyx_tuple__1401 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1401)) __PYX_ERR(0, 2386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1401);
  __Pyx_GIVEREF(__pyx_tuple__1401);
  __pyx_codeobj__1402 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1401, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 2386, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1402)) __PYX_ERR(0, 2386, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2390
 *                 listener.exitImport_clause(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitImport_clause" ):
 *                 return visitor.visitImport_clause(self)
 */
  __pyx_tuple__1403 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1403)) __PYX_ERR(0, 2390, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1403);
  __Pyx_GIVEREF(__pyx_tuple__1403);
  __pyx_codeobj__1404 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1403, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 2390, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1404)) __PYX_ERR(0, 2390, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2399
 * 
 * 
 *     def import_clause(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Import_clauseContext(self, self._ctx, self.state)
 */
  __pyx_tuple__1405 = PyTuple_Pack(6, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la_2, __pyx_n_s_la, __pyx_n_s_token, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1405)) __PYX_ERR(0, 2399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1405);
  __Pyx_GIVEREF(__pyx_tuple__1405);
  __pyx_codeobj__1406 = (PyObject*)__Pyx_PyCode_New(1, 0, 6, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1405, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_import_clause, 2399, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1406)) __PYX_ERR(0, 2399, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2464
 *     class Import_listContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1407 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1407)) __PYX_ERR(0, 2464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1407);
  __Pyx_GIVEREF(__pyx_tuple__1407);
  __pyx_codeobj__1408 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1407, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 2464, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1408)) __PYX_ERR(0, 2464, __pyx_L1_error)
  __pyx_tuple__1409 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1409)) __PYX_ERR(0, 2464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1409);
  __Pyx_GIVEREF(__pyx_tuple__1409);

  /* "pymola/generated/ModelicaParser.pyx":2468
 *             self.parser = parser
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 */
  __pyx_tuple__1410 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1410)) __PYX_ERR(0, 2468, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1410);
  __Pyx_GIVEREF(__pyx_tuple__1410);
  __pyx_codeobj__1411 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1410, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_IDENT, 2468, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1411)) __PYX_ERR(0, 2468, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2471
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 *         def import_list(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Import_listContext)
 */
  __pyx_tuple__1412 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__1412)) __PYX_ERR(0, 2471, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1412);
  __Pyx_GIVEREF(__pyx_tuple__1412);
  __pyx_codeobj__1413 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1412, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_import_list, 2471, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1413)) __PYX_ERR(0, 2471, __pyx_L1_error)
  __pyx_tuple__1414 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__1414)) __PYX_ERR(0, 2471, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1414);
  __Pyx_GIVEREF(__pyx_tuple__1414);

  /* "pymola/generated/ModelicaParser.pyx":2478
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_import_list
 * 
 */
  __pyx_tuple__1415 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1415)) __PYX_ERR(0, 2478, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1415);
  __Pyx_GIVEREF(__pyx_tuple__1415);
  __pyx_codeobj__1416 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1415, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 2478, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1416)) __PYX_ERR(0, 2478, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2481
 *             return ModelicaParser.RULE_import_list
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterImport_list" ):
 *                 listener.enterImport_list(self)
 */
  __pyx_tuple__1417 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1417)) __PYX_ERR(0, 2481, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1417);
  __Pyx_GIVEREF(__pyx_tuple__1417);
  __pyx_codeobj__1418 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1417, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 2481, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1418)) __PYX_ERR(0, 2481, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2485
 *                 listener.enterImport_list(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitImport_list" ):
 *                 listener.exitImport_list(self)
 */
  __pyx_tuple__1419 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1419)) __PYX_ERR(0, 2485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1419);
  __Pyx_GIVEREF(__pyx_tuple__1419);
  __pyx_codeobj__1420 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1419, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 2485, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1420)) __PYX_ERR(0, 2485, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2489
 *                 listener.exitImport_list(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitImport_list" ):
 *                 return visitor.visitImport_list(self)
 */
  __pyx_tuple__1421 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1421)) __PYX_ERR(0, 2489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1421);
  __Pyx_GIVEREF(__pyx_tuple__1421);
  __pyx_codeobj__1422 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1421, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 2489, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1422)) __PYX_ERR(0, 2489, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2498
 * 
 * 
 *     def import_list(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Import_listContext(self, self._ctx, self.state)
 */
  __pyx_tuple__1423 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_alt, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1423)) __PYX_ERR(0, 2498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1423);
  __Pyx_GIVEREF(__pyx_tuple__1423);
  __pyx_codeobj__1424 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1423, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_import_list, 2498, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1424)) __PYX_ERR(0, 2498, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2529
 *     class Extends_clauseContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1425 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1425)) __PYX_ERR(0, 2529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1425);
  __Pyx_GIVEREF(__pyx_tuple__1425);
  __pyx_codeobj__1426 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1425, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 2529, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1426)) __PYX_ERR(0, 2529, __pyx_L1_error)
  __pyx_tuple__1427 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1427)) __PYX_ERR(0, 2529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1427);
  __Pyx_GIVEREF(__pyx_tuple__1427);

  /* "pymola/generated/ModelicaParser.pyx":2533
 *             self.parser = parser
 * 
 *         def component_reference(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 */
  __pyx_tuple__1428 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1428)) __PYX_ERR(0, 2533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1428);
  __Pyx_GIVEREF(__pyx_tuple__1428);
  __pyx_codeobj__1429 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1428, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_component_reference, 2533, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1429)) __PYX_ERR(0, 2533, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2537
 * 
 * 
 *         def class_modification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_modificationContext,0)
 * 
 */
  __pyx_tuple__1430 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1430)) __PYX_ERR(0, 2537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1430);
  __Pyx_GIVEREF(__pyx_tuple__1430);
  __pyx_codeobj__1431 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1430, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_class_modification, 2537, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1431)) __PYX_ERR(0, 2537, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2541
 * 
 * 
 *         def annotation(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.AnnotationContext,0)
 * 
 */
  __pyx_tuple__1432 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1432)) __PYX_ERR(0, 2541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1432);
  __Pyx_GIVEREF(__pyx_tuple__1432);
  __pyx_codeobj__1433 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1432, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_annotation, 2541, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1433)) __PYX_ERR(0, 2541, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2545
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_extends_clause
 * 
 */
  __pyx_tuple__1434 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1434)) __PYX_ERR(0, 2545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1434);
  __Pyx_GIVEREF(__pyx_tuple__1434);
  __pyx_codeobj__1435 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1434, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 2545, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1435)) __PYX_ERR(0, 2545, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2548
 *             return ModelicaParser.RULE_extends_clause
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExtends_clause" ):
 *                 listener.enterExtends_clause(self)
 */
  __pyx_tuple__1436 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1436)) __PYX_ERR(0, 2548, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1436);
  __Pyx_GIVEREF(__pyx_tuple__1436);
  __pyx_codeobj__1437 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1436, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 2548, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1437)) __PYX_ERR(0, 2548, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2552
 *                 listener.enterExtends_clause(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExtends_clause" ):
 *                 listener.exitExtends_clause(self)
 */
  __pyx_tuple__1438 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1438)) __PYX_ERR(0, 2552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1438);
  __Pyx_GIVEREF(__pyx_tuple__1438);
  __pyx_codeobj__1439 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1438, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 2552, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1439)) __PYX_ERR(0, 2552, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2556
 *                 listener.exitExtends_clause(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExtends_clause" ):
 *                 return visitor.visitExtends_clause(self)
 */
  __pyx_tuple__1440 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1440)) __PYX_ERR(0, 2556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1440);
  __Pyx_GIVEREF(__pyx_tuple__1440);
  __pyx_codeobj__1441 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1440, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 2556, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1441)) __PYX_ERR(0, 2556, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2565
 * 
 * 
 *     def extends_clause(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Extends_clauseContext(self, self._ctx, self.state)
 */
  __pyx_tuple__1442 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1442)) __PYX_ERR(0, 2565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1442);
  __Pyx_GIVEREF(__pyx_tuple__1442);
  __pyx_codeobj__1443 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1442, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_extends_clause, 2565, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1443)) __PYX_ERR(0, 2565, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2602
 *     class Constraining_clauseContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1444 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1444)) __PYX_ERR(0, 2602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1444);
  __Pyx_GIVEREF(__pyx_tuple__1444);
  __pyx_codeobj__1445 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1444, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 2602, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1445)) __PYX_ERR(0, 2602, __pyx_L1_error)
  __pyx_tuple__1446 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1446)) __PYX_ERR(0, 2602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1446);
  __Pyx_GIVEREF(__pyx_tuple__1446);

  /* "pymola/generated/ModelicaParser.pyx":2606
 *             self.parser = parser
 * 
 *         def name(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)
 * 
 */
  __pyx_tuple__1447 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1447)) __PYX_ERR(0, 2606, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1447);
  __Pyx_GIVEREF(__pyx_tuple__1447);
  __pyx_codeobj__1448 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1447, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_name, 2606, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1448)) __PYX_ERR(0, 2606, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2610
 * 
 * 
 *         def class_modification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_modificationContext,0)
 * 
 */
  __pyx_tuple__1449 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1449)) __PYX_ERR(0, 2610, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1449);
  __Pyx_GIVEREF(__pyx_tuple__1449);
  __pyx_codeobj__1450 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1449, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_class_modification, 2610, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1450)) __PYX_ERR(0, 2610, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2614
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_constraining_clause
 * 
 */
  __pyx_tuple__1451 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1451)) __PYX_ERR(0, 2614, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1451);
  __Pyx_GIVEREF(__pyx_tuple__1451);
  __pyx_codeobj__1452 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1451, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 2614, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1452)) __PYX_ERR(0, 2614, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2617
 *             return ModelicaParser.RULE_constraining_clause
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterConstraining_clause" ):
 *                 listener.enterConstraining_clause(self)
 */
  __pyx_tuple__1453 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1453)) __PYX_ERR(0, 2617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1453);
  __Pyx_GIVEREF(__pyx_tuple__1453);
  __pyx_codeobj__1454 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1453, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 2617, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1454)) __PYX_ERR(0, 2617, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2621
 *                 listener.enterConstraining_clause(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitConstraining_clause" ):
 *                 listener.exitConstraining_clause(self)
 */
  __pyx_tuple__1455 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1455)) __PYX_ERR(0, 2621, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1455);
  __Pyx_GIVEREF(__pyx_tuple__1455);
  __pyx_codeobj__1456 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1455, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 2621, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1456)) __PYX_ERR(0, 2621, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2625
 *                 listener.exitConstraining_clause(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitConstraining_clause" ):
 *                 return visitor.visitConstraining_clause(self)
 */
  __pyx_tuple__1457 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1457)) __PYX_ERR(0, 2625, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1457);
  __Pyx_GIVEREF(__pyx_tuple__1457);
  __pyx_codeobj__1458 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1457, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 2625, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1458)) __PYX_ERR(0, 2625, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2634
 * 
 * 
 *     def constraining_clause(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Constraining_clauseContext(self, self._ctx, self.state)
 */
  __pyx_tuple__1459 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1459)) __PYX_ERR(0, 2634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1459);
  __Pyx_GIVEREF(__pyx_tuple__1459);
  __pyx_codeobj__1460 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1459, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_constraining_clause, 2634, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1460)) __PYX_ERR(0, 2634, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2663
 *     class Component_clauseContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1461 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1461)) __PYX_ERR(0, 2663, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1461);
  __Pyx_GIVEREF(__pyx_tuple__1461);
  __pyx_codeobj__1462 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1461, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 2663, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1462)) __PYX_ERR(0, 2663, __pyx_L1_error)
  __pyx_tuple__1463 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1463)) __PYX_ERR(0, 2663, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1463);
  __Pyx_GIVEREF(__pyx_tuple__1463);

  /* "pymola/generated/ModelicaParser.pyx":2667
 *             self.parser = parser
 * 
 *         def type_prefix(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Type_prefixContext,0)
 * 
 */
  __pyx_tuple__1464 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1464)) __PYX_ERR(0, 2667, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1464);
  __Pyx_GIVEREF(__pyx_tuple__1464);
  __pyx_codeobj__1465 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1464, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_type_prefix, 2667, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1465)) __PYX_ERR(0, 2667, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2671
 * 
 * 
 *         def type_specifier(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Type_specifierContext,0)
 * 
 */
  __pyx_tuple__1466 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1466)) __PYX_ERR(0, 2671, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1466);
  __Pyx_GIVEREF(__pyx_tuple__1466);
  __pyx_codeobj__1467 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1466, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_type_specifier, 2671, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1467)) __PYX_ERR(0, 2671, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2675
 * 
 * 
 *         def component_list(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_listContext,0)
 * 
 */
  __pyx_tuple__1468 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1468)) __PYX_ERR(0, 2675, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1468);
  __Pyx_GIVEREF(__pyx_tuple__1468);
  __pyx_codeobj__1469 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1468, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_component_list, 2675, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1469)) __PYX_ERR(0, 2675, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2679
 * 
 * 
 *         def array_subscripts(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Array_subscriptsContext,0)
 * 
 */
  __pyx_tuple__1470 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1470)) __PYX_ERR(0, 2679, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1470);
  __Pyx_GIVEREF(__pyx_tuple__1470);
  __pyx_codeobj__1471 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1470, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_array_subscripts, 2679, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1471)) __PYX_ERR(0, 2679, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2683
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_component_clause
 * 
 */
  __pyx_tuple__1472 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1472)) __PYX_ERR(0, 2683, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1472);
  __Pyx_GIVEREF(__pyx_tuple__1472);
  __pyx_codeobj__1473 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1472, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 2683, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1473)) __PYX_ERR(0, 2683, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2686
 *             return ModelicaParser.RULE_component_clause
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterComponent_clause" ):
 *                 listener.enterComponent_clause(self)
 */
  __pyx_tuple__1474 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1474)) __PYX_ERR(0, 2686, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1474);
  __Pyx_GIVEREF(__pyx_tuple__1474);
  __pyx_codeobj__1475 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1474, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 2686, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1475)) __PYX_ERR(0, 2686, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2690
 *                 listener.enterComponent_clause(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitComponent_clause" ):
 *                 listener.exitComponent_clause(self)
 */
  __pyx_tuple__1476 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1476)) __PYX_ERR(0, 2690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1476);
  __Pyx_GIVEREF(__pyx_tuple__1476);
  __pyx_codeobj__1477 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1476, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 2690, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1477)) __PYX_ERR(0, 2690, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2694
 *                 listener.exitComponent_clause(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitComponent_clause" ):
 *                 return visitor.visitComponent_clause(self)
 */
  __pyx_tuple__1478 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1478)) __PYX_ERR(0, 2694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1478);
  __Pyx_GIVEREF(__pyx_tuple__1478);
  __pyx_codeobj__1479 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1478, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 2694, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1479)) __PYX_ERR(0, 2694, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2703
 * 
 * 
 *     def component_clause(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Component_clauseContext(self, self._ctx, self.state)
 */
  __pyx_tuple__1480 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1480)) __PYX_ERR(0, 2703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1480);
  __Pyx_GIVEREF(__pyx_tuple__1480);
  __pyx_codeobj__1481 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1480, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_component_clause, 2703, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1481)) __PYX_ERR(0, 2703, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2734
 *     class Type_prefixContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1482 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1482)) __PYX_ERR(0, 2734, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1482);
  __Pyx_GIVEREF(__pyx_tuple__1482);
  __pyx_codeobj__1483 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1482, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 2734, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1483)) __PYX_ERR(0, 2734, __pyx_L1_error)
  __pyx_tuple__1484 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1484)) __PYX_ERR(0, 2734, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1484);
  __Pyx_GIVEREF(__pyx_tuple__1484);

  /* "pymola/generated/ModelicaParser.pyx":2739
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_type_prefix
 * 
 */
  __pyx_tuple__1485 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1485)) __PYX_ERR(0, 2739, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1485);
  __Pyx_GIVEREF(__pyx_tuple__1485);
  __pyx_codeobj__1486 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1485, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 2739, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1486)) __PYX_ERR(0, 2739, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2742
 *             return ModelicaParser.RULE_type_prefix
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterType_prefix" ):
 *                 listener.enterType_prefix(self)
 */
  __pyx_tuple__1487 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1487)) __PYX_ERR(0, 2742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1487);
  __Pyx_GIVEREF(__pyx_tuple__1487);
  __pyx_codeobj__1488 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1487, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 2742, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1488)) __PYX_ERR(0, 2742, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2746
 *                 listener.enterType_prefix(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitType_prefix" ):
 *                 listener.exitType_prefix(self)
 */
  __pyx_tuple__1489 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1489)) __PYX_ERR(0, 2746, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1489);
  __Pyx_GIVEREF(__pyx_tuple__1489);
  __pyx_codeobj__1490 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1489, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 2746, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1490)) __PYX_ERR(0, 2746, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2750
 *                 listener.exitType_prefix(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitType_prefix" ):
 *                 return visitor.visitType_prefix(self)
 */
  __pyx_tuple__1491 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1491)) __PYX_ERR(0, 2750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1491);
  __Pyx_GIVEREF(__pyx_tuple__1491);
  __pyx_codeobj__1492 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1491, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 2750, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1492)) __PYX_ERR(0, 2750, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2759
 * 
 * 
 *     def type_prefix(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Type_prefixContext(self, self._ctx, self.state)
 */
  __pyx_tuple__1493 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1493)) __PYX_ERR(0, 2759, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1493);
  __Pyx_GIVEREF(__pyx_tuple__1493);
  __pyx_codeobj__1494 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1493, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_type_prefix, 2759, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1494)) __PYX_ERR(0, 2759, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2815
 *     class Type_specifierContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1495 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1495)) __PYX_ERR(0, 2815, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1495);
  __Pyx_GIVEREF(__pyx_tuple__1495);
  __pyx_codeobj__1496 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1495, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 2815, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1496)) __PYX_ERR(0, 2815, __pyx_L1_error)
  __pyx_tuple__1497 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1497)) __PYX_ERR(0, 2815, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1497);
  __Pyx_GIVEREF(__pyx_tuple__1497);

  /* "pymola/generated/ModelicaParser.pyx":2819
 *             self.parser = parser
 * 
 *         def component_reference(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 */
  __pyx_tuple__1498 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1498)) __PYX_ERR(0, 2819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1498);
  __Pyx_GIVEREF(__pyx_tuple__1498);
  __pyx_codeobj__1499 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1498, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_component_reference, 2819, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1499)) __PYX_ERR(0, 2819, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2823
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_type_specifier
 * 
 */
  __pyx_tuple__1500 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1500)) __PYX_ERR(0, 2823, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1500);
  __Pyx_GIVEREF(__pyx_tuple__1500);
  __pyx_codeobj__1501 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1500, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 2823, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1501)) __PYX_ERR(0, 2823, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2826
 *             return ModelicaParser.RULE_type_specifier
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterType_specifier" ):
 *                 listener.enterType_specifier(self)
 */
  __pyx_tuple__1502 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1502)) __PYX_ERR(0, 2826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1502);
  __Pyx_GIVEREF(__pyx_tuple__1502);
  __pyx_codeobj__1503 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1502, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 2826, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1503)) __PYX_ERR(0, 2826, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2830
 *                 listener.enterType_specifier(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitType_specifier" ):
 *                 listener.exitType_specifier(self)
 */
  __pyx_tuple__1504 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1504)) __PYX_ERR(0, 2830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1504);
  __Pyx_GIVEREF(__pyx_tuple__1504);
  __pyx_codeobj__1505 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1504, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 2830, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1505)) __PYX_ERR(0, 2830, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2834
 *                 listener.exitType_specifier(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitType_specifier" ):
 *                 return visitor.visitType_specifier(self)
 */
  __pyx_tuple__1506 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1506)) __PYX_ERR(0, 2834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1506);
  __Pyx_GIVEREF(__pyx_tuple__1506);
  __pyx_codeobj__1507 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1506, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 2834, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1507)) __PYX_ERR(0, 2834, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2843
 * 
 * 
 *     def type_specifier(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Type_specifierContext(self, self._ctx, self.state)
 */
  __pyx_tuple__1508 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1508)) __PYX_ERR(0, 2843, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1508);
  __Pyx_GIVEREF(__pyx_tuple__1508);
  __pyx_codeobj__1509 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1508, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_type_specifier, 2843, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1509)) __PYX_ERR(0, 2843, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2861
 *     class Component_listContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1510 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1510)) __PYX_ERR(0, 2861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1510);
  __Pyx_GIVEREF(__pyx_tuple__1510);
  __pyx_codeobj__1511 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1510, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 2861, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1511)) __PYX_ERR(0, 2861, __pyx_L1_error)
  __pyx_tuple__1512 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1512)) __PYX_ERR(0, 2861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1512);
  __Pyx_GIVEREF(__pyx_tuple__1512);

  /* "pymola/generated/ModelicaParser.pyx":2865
 *             self.parser = parser
 * 
 *         def component_declaration(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Component_declarationContext)
 */
  __pyx_tuple__1513 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__1513)) __PYX_ERR(0, 2865, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1513);
  __Pyx_GIVEREF(__pyx_tuple__1513);
  __pyx_codeobj__1514 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1513, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_component_declaration, 2865, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1514)) __PYX_ERR(0, 2865, __pyx_L1_error)
  __pyx_tuple__1515 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__1515)) __PYX_ERR(0, 2865, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1515);
  __Pyx_GIVEREF(__pyx_tuple__1515);

  /* "pymola/generated/ModelicaParser.pyx":2872
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_component_list
 * 
 */
  __pyx_tuple__1516 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1516)) __PYX_ERR(0, 2872, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1516);
  __Pyx_GIVEREF(__pyx_tuple__1516);
  __pyx_codeobj__1517 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1516, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 2872, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1517)) __PYX_ERR(0, 2872, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2875
 *             return ModelicaParser.RULE_component_list
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterComponent_list" ):
 *                 listener.enterComponent_list(self)
 */
  __pyx_tuple__1518 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1518)) __PYX_ERR(0, 2875, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1518);
  __Pyx_GIVEREF(__pyx_tuple__1518);
  __pyx_codeobj__1519 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1518, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 2875, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1519)) __PYX_ERR(0, 2875, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2879
 *                 listener.enterComponent_list(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitComponent_list" ):
 *                 listener.exitComponent_list(self)
 */
  __pyx_tuple__1520 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1520)) __PYX_ERR(0, 2879, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1520);
  __Pyx_GIVEREF(__pyx_tuple__1520);
  __pyx_codeobj__1521 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1520, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 2879, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1521)) __PYX_ERR(0, 2879, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2883
 *                 listener.exitComponent_list(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitComponent_list" ):
 *                 return visitor.visitComponent_list(self)
 */
  __pyx_tuple__1522 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1522)) __PYX_ERR(0, 2883, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1522);
  __Pyx_GIVEREF(__pyx_tuple__1522);
  __pyx_codeobj__1523 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1522, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 2883, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1523)) __PYX_ERR(0, 2883, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2892
 * 
 * 
 *     def component_list(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Component_listContext(self, self._ctx, self.state)
 */
  __pyx_tuple__1524 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1524)) __PYX_ERR(0, 2892, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1524);
  __Pyx_GIVEREF(__pyx_tuple__1524);
  __pyx_codeobj__1525 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1524, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_component_list, 2892, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1525)) __PYX_ERR(0, 2892, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2923
 *     class Component_declarationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1526 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1526)) __PYX_ERR(0, 2923, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1526);
  __Pyx_GIVEREF(__pyx_tuple__1526);
  __pyx_codeobj__1527 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1526, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 2923, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1527)) __PYX_ERR(0, 2923, __pyx_L1_error)
  __pyx_tuple__1528 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1528)) __PYX_ERR(0, 2923, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1528);
  __Pyx_GIVEREF(__pyx_tuple__1528);

  /* "pymola/generated/ModelicaParser.pyx":2927
 *             self.parser = parser
 * 
 *         def declaration(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.DeclarationContext,0)
 * 
 */
  __pyx_tuple__1529 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1529)) __PYX_ERR(0, 2927, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1529);
  __Pyx_GIVEREF(__pyx_tuple__1529);
  __pyx_codeobj__1530 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1529, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_declaration, 2927, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1530)) __PYX_ERR(0, 2927, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2931
 * 
 * 
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */
  __pyx_tuple__1531 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1531)) __PYX_ERR(0, 2931, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1531);
  __Pyx_GIVEREF(__pyx_tuple__1531);
  __pyx_codeobj__1532 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1531, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_comment, 2931, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1532)) __PYX_ERR(0, 2931, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2935
 * 
 * 
 *         def condition_attribute(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Condition_attributeContext,0)
 * 
 */
  __pyx_tuple__1533 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1533)) __PYX_ERR(0, 2935, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1533);
  __Pyx_GIVEREF(__pyx_tuple__1533);
  __pyx_codeobj__1534 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1533, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_condition_attribute, 2935, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1534)) __PYX_ERR(0, 2935, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2939
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_component_declaration
 * 
 */
  __pyx_tuple__1535 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1535)) __PYX_ERR(0, 2939, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1535);
  __Pyx_GIVEREF(__pyx_tuple__1535);
  __pyx_codeobj__1536 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1535, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 2939, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1536)) __PYX_ERR(0, 2939, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2942
 *             return ModelicaParser.RULE_component_declaration
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterComponent_declaration" ):
 *                 listener.enterComponent_declaration(self)
 */
  __pyx_tuple__1537 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1537)) __PYX_ERR(0, 2942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1537);
  __Pyx_GIVEREF(__pyx_tuple__1537);
  __pyx_codeobj__1538 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1537, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 2942, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1538)) __PYX_ERR(0, 2942, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2946
 *                 listener.enterComponent_declaration(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitComponent_declaration" ):
 *                 listener.exitComponent_declaration(self)
 */
  __pyx_tuple__1539 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1539)) __PYX_ERR(0, 2946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1539);
  __Pyx_GIVEREF(__pyx_tuple__1539);
  __pyx_codeobj__1540 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1539, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 2946, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1540)) __PYX_ERR(0, 2946, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2950
 *                 listener.exitComponent_declaration(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitComponent_declaration" ):
 *                 return visitor.visitComponent_declaration(self)
 */
  __pyx_tuple__1541 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1541)) __PYX_ERR(0, 2950, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1541);
  __Pyx_GIVEREF(__pyx_tuple__1541);
  __pyx_codeobj__1542 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1541, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 2950, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1542)) __PYX_ERR(0, 2950, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2959
 * 
 * 
 *     def component_declaration(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Component_declarationContext(self, self._ctx, self.state)
 */
  __pyx_tuple__1543 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1543)) __PYX_ERR(0, 2959, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1543);
  __Pyx_GIVEREF(__pyx_tuple__1543);
  __pyx_codeobj__1544 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1543, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_component_declaration, 2959, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1544)) __PYX_ERR(0, 2959, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2988
 *     class Condition_attributeContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1545 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1545)) __PYX_ERR(0, 2988, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1545);
  __Pyx_GIVEREF(__pyx_tuple__1545);
  __pyx_codeobj__1546 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1545, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 2988, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1546)) __PYX_ERR(0, 2988, __pyx_L1_error)
  __pyx_tuple__1547 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1547)) __PYX_ERR(0, 2988, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1547);
  __Pyx_GIVEREF(__pyx_tuple__1547);

  /* "pymola/generated/ModelicaParser.pyx":2992
 *             self.parser = parser
 * 
 *         def expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 */
  __pyx_tuple__1548 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1548)) __PYX_ERR(0, 2992, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1548);
  __Pyx_GIVEREF(__pyx_tuple__1548);
  __pyx_codeobj__1549 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1548, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_expression, 2992, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1549)) __PYX_ERR(0, 2992, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2996
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_condition_attribute
 * 
 */
  __pyx_tuple__1550 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1550)) __PYX_ERR(0, 2996, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1550);
  __Pyx_GIVEREF(__pyx_tuple__1550);
  __pyx_codeobj__1551 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1550, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 2996, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1551)) __PYX_ERR(0, 2996, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":2999
 *             return ModelicaParser.RULE_condition_attribute
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterCondition_attribute" ):
 *                 listener.enterCondition_attribute(self)
 */
  __pyx_tuple__1552 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1552)) __PYX_ERR(0, 2999, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1552);
  __Pyx_GIVEREF(__pyx_tuple__1552);
  __pyx_codeobj__1553 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1552, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 2999, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1553)) __PYX_ERR(0, 2999, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3003
 *                 listener.enterCondition_attribute(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitCondition_attribute" ):
 *                 listener.exitCondition_attribute(self)
 */
  __pyx_tuple__1554 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1554)) __PYX_ERR(0, 3003, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1554);
  __Pyx_GIVEREF(__pyx_tuple__1554);
  __pyx_codeobj__1555 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1554, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 3003, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1555)) __PYX_ERR(0, 3003, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3007
 *                 listener.exitCondition_attribute(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitCondition_attribute" ):
 *                 return visitor.visitCondition_attribute(self)
 */
  __pyx_tuple__1556 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1556)) __PYX_ERR(0, 3007, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1556);
  __Pyx_GIVEREF(__pyx_tuple__1556);
  __pyx_codeobj__1557 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1556, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 3007, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1557)) __PYX_ERR(0, 3007, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3016
 * 
 * 
 *     def condition_attribute(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Condition_attributeContext(self, self._ctx, self.state)
 */
  __pyx_tuple__1558 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1558)) __PYX_ERR(0, 3016, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1558);
  __Pyx_GIVEREF(__pyx_tuple__1558);
  __pyx_codeobj__1559 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1558, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_condition_attribute, 3016, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1559)) __PYX_ERR(0, 3016, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3036
 *     class DeclarationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1560 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1560)) __PYX_ERR(0, 3036, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1560);
  __Pyx_GIVEREF(__pyx_tuple__1560);
  __pyx_codeobj__1561 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1560, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 3036, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1561)) __PYX_ERR(0, 3036, __pyx_L1_error)
  __pyx_tuple__1562 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1562)) __PYX_ERR(0, 3036, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1562);
  __Pyx_GIVEREF(__pyx_tuple__1562);

  /* "pymola/generated/ModelicaParser.pyx":3040
 *             self.parser = parser
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 */
  __pyx_tuple__1563 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1563)) __PYX_ERR(0, 3040, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1563);
  __Pyx_GIVEREF(__pyx_tuple__1563);
  __pyx_codeobj__1564 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1563, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_IDENT, 3040, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1564)) __PYX_ERR(0, 3040, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3043
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 *         def array_subscripts(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Array_subscriptsContext,0)
 * 
 */
  __pyx_tuple__1565 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1565)) __PYX_ERR(0, 3043, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1565);
  __Pyx_GIVEREF(__pyx_tuple__1565);
  __pyx_codeobj__1566 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1565, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_array_subscripts, 3043, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1566)) __PYX_ERR(0, 3043, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3047
 * 
 * 
 *         def modification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ModificationContext,0)
 * 
 */
  __pyx_tuple__1567 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1567)) __PYX_ERR(0, 3047, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1567);
  __Pyx_GIVEREF(__pyx_tuple__1567);
  __pyx_codeobj__1568 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1567, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_modification, 3047, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1568)) __PYX_ERR(0, 3047, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3051
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_declaration
 * 
 */
  __pyx_tuple__1569 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1569)) __PYX_ERR(0, 3051, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1569);
  __Pyx_GIVEREF(__pyx_tuple__1569);
  __pyx_codeobj__1570 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1569, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 3051, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1570)) __PYX_ERR(0, 3051, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3054
 *             return ModelicaParser.RULE_declaration
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterDeclaration" ):
 *                 listener.enterDeclaration(self)
 */
  __pyx_tuple__1571 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1571)) __PYX_ERR(0, 3054, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1571);
  __Pyx_GIVEREF(__pyx_tuple__1571);
  __pyx_codeobj__1572 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1571, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 3054, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1572)) __PYX_ERR(0, 3054, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3058
 *                 listener.enterDeclaration(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitDeclaration" ):
 *                 listener.exitDeclaration(self)
 */
  __pyx_tuple__1573 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1573)) __PYX_ERR(0, 3058, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1573);
  __Pyx_GIVEREF(__pyx_tuple__1573);
  __pyx_codeobj__1574 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1573, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 3058, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1574)) __PYX_ERR(0, 3058, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3062
 *                 listener.exitDeclaration(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitDeclaration" ):
 *                 return visitor.visitDeclaration(self)
 */
  __pyx_tuple__1575 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1575)) __PYX_ERR(0, 3062, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1575);
  __Pyx_GIVEREF(__pyx_tuple__1575);
  __pyx_codeobj__1576 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1575, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 3062, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1576)) __PYX_ERR(0, 3062, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3071
 * 
 * 
 *     def declaration(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.DeclarationContext(self, self._ctx, self.state)
 */
  __pyx_tuple__1577 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1577)) __PYX_ERR(0, 3071, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1577);
  __Pyx_GIVEREF(__pyx_tuple__1577);
  __pyx_codeobj__1578 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1577, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_declaration, 3071, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1578)) __PYX_ERR(0, 3071, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3106
 *     class ModificationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1579 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1579)) __PYX_ERR(0, 3106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1579);
  __Pyx_GIVEREF(__pyx_tuple__1579);
  __pyx_codeobj__1580 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1579, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 3106, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1580)) __PYX_ERR(0, 3106, __pyx_L1_error)
  __pyx_tuple__1581 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1581)) __PYX_ERR(0, 3106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1581);
  __Pyx_GIVEREF(__pyx_tuple__1581);

  /* "pymola/generated/ModelicaParser.pyx":3111
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_modification
 * 
 */
  __pyx_tuple__1582 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1582)) __PYX_ERR(0, 3111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1582);
  __Pyx_GIVEREF(__pyx_tuple__1582);
  __pyx_codeobj__1583 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1582, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 3111, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1583)) __PYX_ERR(0, 3111, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3115
 * 
 * 
 *         def copyFrom(self, ctx:ParserRuleContext):             # <<<<<<<<<<<<<<
 *             super().copyFrom(ctx)
 * 
 */
  __pyx_tuple__1584 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__1584)) __PYX_ERR(0, 3115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1584);
  __Pyx_GIVEREF(__pyx_tuple__1584);
  __pyx_codeobj__1585 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1584, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_copyFrom, 3115, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1585)) __PYX_ERR(0, 3115, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3122
 *     class Modification_classContext(ModificationContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ModificationContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__1586 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__1586)) __PYX_ERR(0, 3122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1586);
  __Pyx_GIVEREF(__pyx_tuple__1586);
  __pyx_codeobj__1587 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1586, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 3122, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1587)) __PYX_ERR(0, 3122, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3126
 *             self.copyFrom(ctx)
 * 
 *         def class_modification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_modificationContext,0)
 * 
 */
  __pyx_tuple__1588 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1588)) __PYX_ERR(0, 3126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1588);
  __Pyx_GIVEREF(__pyx_tuple__1588);
  __pyx_codeobj__1589 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1588, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_class_modification, 3126, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1589)) __PYX_ERR(0, 3126, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3129
 *             return self.getTypedRuleContext(ModelicaParser.Class_modificationContext,0)
 * 
 *         def expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 */
  __pyx_tuple__1590 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1590)) __PYX_ERR(0, 3129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1590);
  __Pyx_GIVEREF(__pyx_tuple__1590);
  __pyx_codeobj__1591 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1590, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_expression, 3129, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1591)) __PYX_ERR(0, 3129, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3133
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterModification_class" ):
 *                 listener.enterModification_class(self)
 */
  __pyx_tuple__1592 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1592)) __PYX_ERR(0, 3133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1592);
  __Pyx_GIVEREF(__pyx_tuple__1592);
  __pyx_codeobj__1593 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1592, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 3133, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1593)) __PYX_ERR(0, 3133, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3137
 *                 listener.enterModification_class(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitModification_class" ):
 *                 listener.exitModification_class(self)
 */
  __pyx_tuple__1594 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1594)) __PYX_ERR(0, 3137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1594);
  __Pyx_GIVEREF(__pyx_tuple__1594);
  __pyx_codeobj__1595 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1594, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 3137, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1595)) __PYX_ERR(0, 3137, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3141
 *                 listener.exitModification_class(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitModification_class" ):
 *                 return visitor.visitModification_class(self)
 */
  __pyx_tuple__1596 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1596)) __PYX_ERR(0, 3141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1596);
  __Pyx_GIVEREF(__pyx_tuple__1596);
  __pyx_codeobj__1597 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1596, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 3141, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1597)) __PYX_ERR(0, 3141, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3150
 *     class Modification_assignment2Context(ModificationContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ModificationContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__1598 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__1598)) __PYX_ERR(0, 3150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1598);
  __Pyx_GIVEREF(__pyx_tuple__1598);
  __pyx_codeobj__1599 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1598, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 3150, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1599)) __PYX_ERR(0, 3150, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3154
 *             self.copyFrom(ctx)
 * 
 *         def expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 */
  __pyx_tuple__1600 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1600)) __PYX_ERR(0, 3154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1600);
  __Pyx_GIVEREF(__pyx_tuple__1600);
  __pyx_codeobj__1601 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1600, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_expression, 3154, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1601)) __PYX_ERR(0, 3154, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3158
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterModification_assignment2" ):
 *                 listener.enterModification_assignment2(self)
 */
  __pyx_tuple__1602 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1602)) __PYX_ERR(0, 3158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1602);
  __Pyx_GIVEREF(__pyx_tuple__1602);
  __pyx_codeobj__1603 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1602, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 3158, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1603)) __PYX_ERR(0, 3158, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3162
 *                 listener.enterModification_assignment2(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitModification_assignment2" ):
 *                 listener.exitModification_assignment2(self)
 */
  __pyx_tuple__1604 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1604)) __PYX_ERR(0, 3162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1604);
  __Pyx_GIVEREF(__pyx_tuple__1604);
  __pyx_codeobj__1605 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1604, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 3162, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1605)) __PYX_ERR(0, 3162, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3166
 *                 listener.exitModification_assignment2(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitModification_assignment2" ):
 *                 return visitor.visitModification_assignment2(self)
 */
  __pyx_tuple__1606 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1606)) __PYX_ERR(0, 3166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1606);
  __Pyx_GIVEREF(__pyx_tuple__1606);
  __pyx_codeobj__1607 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1606, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 3166, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1607)) __PYX_ERR(0, 3166, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3175
 *     class Modification_assignmentContext(ModificationContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ModificationContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__1608 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__1608)) __PYX_ERR(0, 3175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1608);
  __Pyx_GIVEREF(__pyx_tuple__1608);
  __pyx_codeobj__1609 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1608, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 3175, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1609)) __PYX_ERR(0, 3175, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3179
 *             self.copyFrom(ctx)
 * 
 *         def expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 */
  __pyx_tuple__1610 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1610)) __PYX_ERR(0, 3179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1610);
  __Pyx_GIVEREF(__pyx_tuple__1610);
  __pyx_codeobj__1611 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1610, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_expression, 3179, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1611)) __PYX_ERR(0, 3179, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3183
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterModification_assignment" ):
 *                 listener.enterModification_assignment(self)
 */
  __pyx_tuple__1612 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1612)) __PYX_ERR(0, 3183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1612);
  __Pyx_GIVEREF(__pyx_tuple__1612);
  __pyx_codeobj__1613 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1612, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 3183, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1613)) __PYX_ERR(0, 3183, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3187
 *                 listener.enterModification_assignment(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitModification_assignment" ):
 *                 listener.exitModification_assignment(self)
 */
  __pyx_tuple__1614 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1614)) __PYX_ERR(0, 3187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1614);
  __Pyx_GIVEREF(__pyx_tuple__1614);
  __pyx_codeobj__1615 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1614, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 3187, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1615)) __PYX_ERR(0, 3187, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3191
 *                 listener.exitModification_assignment(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitModification_assignment" ):
 *                 return visitor.visitModification_assignment(self)
 */
  __pyx_tuple__1616 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1616)) __PYX_ERR(0, 3191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1616);
  __Pyx_GIVEREF(__pyx_tuple__1616);
  __pyx_codeobj__1617 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1616, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 3191, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1617)) __PYX_ERR(0, 3191, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3199
 * 
 * 
 *     def modification(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.ModificationContext(self, self._ctx, self.state)
 */
  __pyx_tuple__1618 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_token, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1618)) __PYX_ERR(0, 3199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1618);
  __Pyx_GIVEREF(__pyx_tuple__1618);
  __pyx_codeobj__1619 = (PyObject*)__Pyx_PyCode_New(1, 0, 5, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1618, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_modification, 3199, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1619)) __PYX_ERR(0, 3199, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3253
 *     class Class_modificationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1620 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1620)) __PYX_ERR(0, 3253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1620);
  __Pyx_GIVEREF(__pyx_tuple__1620);
  __pyx_codeobj__1621 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1620, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 3253, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1621)) __PYX_ERR(0, 3253, __pyx_L1_error)
  __pyx_tuple__1622 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1622)) __PYX_ERR(0, 3253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1622);
  __Pyx_GIVEREF(__pyx_tuple__1622);

  /* "pymola/generated/ModelicaParser.pyx":3257
 *             self.parser = parser
 * 
 *         def argument_list(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Argument_listContext,0)
 * 
 */
  __pyx_tuple__1623 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1623)) __PYX_ERR(0, 3257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1623);
  __Pyx_GIVEREF(__pyx_tuple__1623);
  __pyx_codeobj__1624 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1623, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_argument_list, 3257, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1624)) __PYX_ERR(0, 3257, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3261
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_class_modification
 * 
 */
  __pyx_tuple__1625 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1625)) __PYX_ERR(0, 3261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1625);
  __Pyx_GIVEREF(__pyx_tuple__1625);
  __pyx_codeobj__1626 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1625, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 3261, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1626)) __PYX_ERR(0, 3261, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3264
 *             return ModelicaParser.RULE_class_modification
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterClass_modification" ):
 *                 listener.enterClass_modification(self)
 */
  __pyx_tuple__1627 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1627)) __PYX_ERR(0, 3264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1627);
  __Pyx_GIVEREF(__pyx_tuple__1627);
  __pyx_codeobj__1628 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1627, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 3264, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1628)) __PYX_ERR(0, 3264, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3268
 *                 listener.enterClass_modification(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitClass_modification" ):
 *                 listener.exitClass_modification(self)
 */
  __pyx_tuple__1629 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1629)) __PYX_ERR(0, 3268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1629);
  __Pyx_GIVEREF(__pyx_tuple__1629);
  __pyx_codeobj__1630 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1629, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 3268, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1630)) __PYX_ERR(0, 3268, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3272
 *                 listener.exitClass_modification(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitClass_modification" ):
 *                 return visitor.visitClass_modification(self)
 */
  __pyx_tuple__1631 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1631)) __PYX_ERR(0, 3272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1631);
  __Pyx_GIVEREF(__pyx_tuple__1631);
  __pyx_codeobj__1632 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1631, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 3272, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1632)) __PYX_ERR(0, 3272, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3281
 * 
 * 
 *     def class_modification(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Class_modificationContext(self, self._ctx, self.state)
 */
  __pyx_tuple__1633 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1633)) __PYX_ERR(0, 3281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1633);
  __Pyx_GIVEREF(__pyx_tuple__1633);
  __pyx_codeobj__1634 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1633, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_class_modification, 3281, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1634)) __PYX_ERR(0, 3281, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3310
 *     class Argument_listContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1635 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1635)) __PYX_ERR(0, 3310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1635);
  __Pyx_GIVEREF(__pyx_tuple__1635);
  __pyx_codeobj__1636 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1635, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 3310, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1636)) __PYX_ERR(0, 3310, __pyx_L1_error)
  __pyx_tuple__1637 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1637)) __PYX_ERR(0, 3310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1637);
  __Pyx_GIVEREF(__pyx_tuple__1637);

  /* "pymola/generated/ModelicaParser.pyx":3314
 *             self.parser = parser
 * 
 *         def argument(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ArgumentContext)
 */
  __pyx_tuple__1638 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__1638)) __PYX_ERR(0, 3314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1638);
  __Pyx_GIVEREF(__pyx_tuple__1638);
  __pyx_codeobj__1639 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1638, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_argument, 3314, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1639)) __PYX_ERR(0, 3314, __pyx_L1_error)
  __pyx_tuple__1640 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__1640)) __PYX_ERR(0, 3314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1640);
  __Pyx_GIVEREF(__pyx_tuple__1640);

  /* "pymola/generated/ModelicaParser.pyx":3321
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_argument_list
 * 
 */
  __pyx_tuple__1641 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1641)) __PYX_ERR(0, 3321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1641);
  __Pyx_GIVEREF(__pyx_tuple__1641);
  __pyx_codeobj__1642 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1641, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 3321, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1642)) __PYX_ERR(0, 3321, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3324
 *             return ModelicaParser.RULE_argument_list
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterArgument_list" ):
 *                 listener.enterArgument_list(self)
 */
  __pyx_tuple__1643 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1643)) __PYX_ERR(0, 3324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1643);
  __Pyx_GIVEREF(__pyx_tuple__1643);
  __pyx_codeobj__1644 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1643, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 3324, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1644)) __PYX_ERR(0, 3324, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3328
 *                 listener.enterArgument_list(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitArgument_list" ):
 *                 listener.exitArgument_list(self)
 */
  __pyx_tuple__1645 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1645)) __PYX_ERR(0, 3328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1645);
  __Pyx_GIVEREF(__pyx_tuple__1645);
  __pyx_codeobj__1646 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1645, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 3328, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1646)) __PYX_ERR(0, 3328, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3332
 *                 listener.exitArgument_list(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitArgument_list" ):
 *                 return visitor.visitArgument_list(self)
 */
  __pyx_tuple__1647 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1647)) __PYX_ERR(0, 3332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1647);
  __Pyx_GIVEREF(__pyx_tuple__1647);
  __pyx_codeobj__1648 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1647, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 3332, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1648)) __PYX_ERR(0, 3332, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3341
 * 
 * 
 *     def argument_list(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Argument_listContext(self, self._ctx, self.state)
 */
  __pyx_tuple__1649 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1649)) __PYX_ERR(0, 3341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1649);
  __Pyx_GIVEREF(__pyx_tuple__1649);
  __pyx_codeobj__1650 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1649, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_argument_list, 3341, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1650)) __PYX_ERR(0, 3341, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3372
 *     class ArgumentContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1651 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1651)) __PYX_ERR(0, 3372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1651);
  __Pyx_GIVEREF(__pyx_tuple__1651);
  __pyx_codeobj__1652 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1651, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 3372, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1652)) __PYX_ERR(0, 3372, __pyx_L1_error)
  __pyx_tuple__1653 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1653)) __PYX_ERR(0, 3372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1653);
  __Pyx_GIVEREF(__pyx_tuple__1653);

  /* "pymola/generated/ModelicaParser.pyx":3376
 *             self.parser = parser
 * 
 *         def element_modification_or_replaceable(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Element_modification_or_replaceableContext,0)
 * 
 */
  __pyx_tuple__1654 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1654)) __PYX_ERR(0, 3376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1654);
  __Pyx_GIVEREF(__pyx_tuple__1654);
  __pyx_codeobj__1655 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1654, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_element_modification_or_replacea, 3376, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1655)) __PYX_ERR(0, 3376, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3380
 * 
 * 
 *         def element_redeclaration(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Element_redeclarationContext,0)
 * 
 */
  __pyx_tuple__1656 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1656)) __PYX_ERR(0, 3380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1656);
  __Pyx_GIVEREF(__pyx_tuple__1656);
  __pyx_codeobj__1657 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1656, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_element_redeclaration, 3380, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1657)) __PYX_ERR(0, 3380, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3384
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_argument
 * 
 */
  __pyx_tuple__1658 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1658)) __PYX_ERR(0, 3384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1658);
  __Pyx_GIVEREF(__pyx_tuple__1658);
  __pyx_codeobj__1659 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1658, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 3384, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1659)) __PYX_ERR(0, 3384, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3387
 *             return ModelicaParser.RULE_argument
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterArgument" ):
 *                 listener.enterArgument(self)
 */
  __pyx_tuple__1660 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1660)) __PYX_ERR(0, 3387, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1660);
  __Pyx_GIVEREF(__pyx_tuple__1660);
  __pyx_codeobj__1661 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1660, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 3387, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1661)) __PYX_ERR(0, 3387, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3391
 *                 listener.enterArgument(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitArgument" ):
 *                 listener.exitArgument(self)
 */
  __pyx_tuple__1662 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1662)) __PYX_ERR(0, 3391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1662);
  __Pyx_GIVEREF(__pyx_tuple__1662);
  __pyx_codeobj__1663 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1662, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 3391, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1663)) __PYX_ERR(0, 3391, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3395
 *                 listener.exitArgument(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitArgument" ):
 *                 return visitor.visitArgument(self)
 */
  __pyx_tuple__1664 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1664)) __PYX_ERR(0, 3395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1664);
  __Pyx_GIVEREF(__pyx_tuple__1664);
  __pyx_codeobj__1665 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1664, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 3395, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1665)) __PYX_ERR(0, 3395, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3404
 * 
 * 
 *     def argument(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.ArgumentContext(self, self._ctx, self.state)
 */
  __pyx_tuple__1666 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_token, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1666)) __PYX_ERR(0, 3404, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1666);
  __Pyx_GIVEREF(__pyx_tuple__1666);
  __pyx_codeobj__1667 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1666, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_argument, 3404, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1667)) __PYX_ERR(0, 3404, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3435
 *     class Element_modification_or_replaceableContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1668 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1668)) __PYX_ERR(0, 3435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1668);
  __Pyx_GIVEREF(__pyx_tuple__1668);
  __pyx_codeobj__1669 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1668, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 3435, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1669)) __PYX_ERR(0, 3435, __pyx_L1_error)
  __pyx_tuple__1670 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1670)) __PYX_ERR(0, 3435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1670);
  __Pyx_GIVEREF(__pyx_tuple__1670);

  /* "pymola/generated/ModelicaParser.pyx":3439
 *             self.parser = parser
 * 
 *         def element_modification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Element_modificationContext,0)
 * 
 */
  __pyx_tuple__1671 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1671)) __PYX_ERR(0, 3439, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1671);
  __Pyx_GIVEREF(__pyx_tuple__1671);
  __pyx_codeobj__1672 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1671, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_element_modification, 3439, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1672)) __PYX_ERR(0, 3439, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3443
 * 
 * 
 *         def element_replaceable(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Element_replaceableContext,0)
 * 
 */
  __pyx_tuple__1673 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1673)) __PYX_ERR(0, 3443, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1673);
  __Pyx_GIVEREF(__pyx_tuple__1673);
  __pyx_codeobj__1674 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1673, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_element_replaceable, 3443, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1674)) __PYX_ERR(0, 3443, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3447
 * 
 * 
 *         def EACH(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.EACH, 0)
 * 
 */
  __pyx_tuple__1675 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1675)) __PYX_ERR(0, 3447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1675);
  __Pyx_GIVEREF(__pyx_tuple__1675);
  __pyx_codeobj__1676 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1675, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_EACH, 3447, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1676)) __PYX_ERR(0, 3447, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3450
 *             return self.getToken(ModelicaParser.EACH, 0)
 * 
 *         def FINAL(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.FINAL, 0)
 * 
 */
  __pyx_tuple__1677 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1677)) __PYX_ERR(0, 3450, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1677);
  __Pyx_GIVEREF(__pyx_tuple__1677);
  __pyx_codeobj__1678 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1677, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_FINAL, 3450, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1678)) __PYX_ERR(0, 3450, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3453
 *             return self.getToken(ModelicaParser.FINAL, 0)
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_element_modification_or_replaceable
 * 
 */
  __pyx_tuple__1679 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1679)) __PYX_ERR(0, 3453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1679);
  __Pyx_GIVEREF(__pyx_tuple__1679);
  __pyx_codeobj__1680 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1679, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 3453, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1680)) __PYX_ERR(0, 3453, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3456
 *             return ModelicaParser.RULE_element_modification_or_replaceable
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterElement_modification_or_replaceable" ):
 *                 listener.enterElement_modification_or_replaceable(self)
 */
  __pyx_tuple__1681 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1681)) __PYX_ERR(0, 3456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1681);
  __Pyx_GIVEREF(__pyx_tuple__1681);
  __pyx_codeobj__1682 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1681, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 3456, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1682)) __PYX_ERR(0, 3456, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3460
 *                 listener.enterElement_modification_or_replaceable(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitElement_modification_or_replaceable" ):
 *                 listener.exitElement_modification_or_replaceable(self)
 */
  __pyx_tuple__1683 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1683)) __PYX_ERR(0, 3460, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1683);
  __Pyx_GIVEREF(__pyx_tuple__1683);
  __pyx_codeobj__1684 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1683, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 3460, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1684)) __PYX_ERR(0, 3460, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3464
 *                 listener.exitElement_modification_or_replaceable(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitElement_modification_or_replaceable" ):
 *                 return visitor.visitElement_modification_or_replaceable(self)
 */
  __pyx_tuple__1685 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1685)) __PYX_ERR(0, 3464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1685);
  __Pyx_GIVEREF(__pyx_tuple__1685);
  __pyx_codeobj__1686 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1685, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 3464, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1686)) __PYX_ERR(0, 3464, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3473
 * 
 * 
 *     def element_modification_or_replaceable(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Element_modification_or_replaceableContext(self, self._ctx, self.state)
 */
  __pyx_tuple__1687 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la, __pyx_n_s_token, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1687)) __PYX_ERR(0, 3473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1687);
  __Pyx_GIVEREF(__pyx_tuple__1687);
  __pyx_codeobj__1688 = (PyObject*)__Pyx_PyCode_New(1, 0, 5, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1687, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_element_modification_or_replacea, 3473, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1688)) __PYX_ERR(0, 3473, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3520
 *     class Element_modificationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1689 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1689)) __PYX_ERR(0, 3520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1689);
  __Pyx_GIVEREF(__pyx_tuple__1689);
  __pyx_codeobj__1690 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1689, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 3520, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1690)) __PYX_ERR(0, 3520, __pyx_L1_error)
  __pyx_tuple__1691 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1691)) __PYX_ERR(0, 3520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1691);
  __Pyx_GIVEREF(__pyx_tuple__1691);

  /* "pymola/generated/ModelicaParser.pyx":3524
 *             self.parser = parser
 * 
 *         def component_reference(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 */
  __pyx_tuple__1692 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1692)) __PYX_ERR(0, 3524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1692);
  __Pyx_GIVEREF(__pyx_tuple__1692);
  __pyx_codeobj__1693 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1692, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_component_reference, 3524, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1693)) __PYX_ERR(0, 3524, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3528
 * 
 * 
 *         def string_comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.String_commentContext,0)
 * 
 */
  __pyx_tuple__1694 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1694)) __PYX_ERR(0, 3528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1694);
  __Pyx_GIVEREF(__pyx_tuple__1694);
  __pyx_codeobj__1695 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1694, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_string_comment, 3528, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1695)) __PYX_ERR(0, 3528, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3532
 * 
 * 
 *         def modification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ModificationContext,0)
 * 
 */
  __pyx_tuple__1696 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1696)) __PYX_ERR(0, 3532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1696);
  __Pyx_GIVEREF(__pyx_tuple__1696);
  __pyx_codeobj__1697 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1696, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_modification, 3532, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1697)) __PYX_ERR(0, 3532, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3536
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_element_modification
 * 
 */
  __pyx_tuple__1698 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1698)) __PYX_ERR(0, 3536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1698);
  __Pyx_GIVEREF(__pyx_tuple__1698);
  __pyx_codeobj__1699 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1698, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 3536, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1699)) __PYX_ERR(0, 3536, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3539
 *             return ModelicaParser.RULE_element_modification
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterElement_modification" ):
 *                 listener.enterElement_modification(self)
 */
  __pyx_tuple__1700 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1700)) __PYX_ERR(0, 3539, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1700);
  __Pyx_GIVEREF(__pyx_tuple__1700);
  __pyx_codeobj__1701 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1700, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 3539, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1701)) __PYX_ERR(0, 3539, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3543
 *                 listener.enterElement_modification(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitElement_modification" ):
 *                 listener.exitElement_modification(self)
 */
  __pyx_tuple__1702 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1702)) __PYX_ERR(0, 3543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1702);
  __Pyx_GIVEREF(__pyx_tuple__1702);
  __pyx_codeobj__1703 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1702, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 3543, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1703)) __PYX_ERR(0, 3543, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3547
 *                 listener.exitElement_modification(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitElement_modification" ):
 *                 return visitor.visitElement_modification(self)
 */
  __pyx_tuple__1704 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1704)) __PYX_ERR(0, 3547, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1704);
  __Pyx_GIVEREF(__pyx_tuple__1704);
  __pyx_codeobj__1705 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1704, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 3547, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1705)) __PYX_ERR(0, 3547, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3556
 * 
 * 
 *     def element_modification(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Element_modificationContext(self, self._ctx, self.state)
 */
  __pyx_tuple__1706 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1706)) __PYX_ERR(0, 3556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1706);
  __Pyx_GIVEREF(__pyx_tuple__1706);
  __pyx_codeobj__1707 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1706, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_element_modification, 3556, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1707)) __PYX_ERR(0, 3556, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3585
 *     class Element_redeclarationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1708 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1708)) __PYX_ERR(0, 3585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1708);
  __Pyx_GIVEREF(__pyx_tuple__1708);
  __pyx_codeobj__1709 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1708, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 3585, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1709)) __PYX_ERR(0, 3585, __pyx_L1_error)
  __pyx_tuple__1710 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1710)) __PYX_ERR(0, 3585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1710);
  __Pyx_GIVEREF(__pyx_tuple__1710);

  /* "pymola/generated/ModelicaParser.pyx":3589
 *             self.parser = parser
 * 
 *         def REDECLARE(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.REDECLARE, 0)
 * 
 */
  __pyx_tuple__1711 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1711)) __PYX_ERR(0, 3589, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1711);
  __Pyx_GIVEREF(__pyx_tuple__1711);
  __pyx_codeobj__1712 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1711, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_REDECLARE, 3589, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1712)) __PYX_ERR(0, 3589, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3592
 *             return self.getToken(ModelicaParser.REDECLARE, 0)
 * 
 *         def element_replaceable(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Element_replaceableContext,0)
 * 
 */
  __pyx_tuple__1713 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1713)) __PYX_ERR(0, 3592, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1713);
  __Pyx_GIVEREF(__pyx_tuple__1713);
  __pyx_codeobj__1714 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1713, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_element_replaceable, 3592, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1714)) __PYX_ERR(0, 3592, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3596
 * 
 * 
 *         def EACH(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.EACH, 0)
 * 
 */
  __pyx_tuple__1715 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1715)) __PYX_ERR(0, 3596, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1715);
  __Pyx_GIVEREF(__pyx_tuple__1715);
  __pyx_codeobj__1716 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1715, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_EACH, 3596, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1716)) __PYX_ERR(0, 3596, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3599
 *             return self.getToken(ModelicaParser.EACH, 0)
 * 
 *         def FINAL(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.FINAL, 0)
 * 
 */
  __pyx_tuple__1717 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1717)) __PYX_ERR(0, 3599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1717);
  __Pyx_GIVEREF(__pyx_tuple__1717);
  __pyx_codeobj__1718 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1717, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_FINAL, 3599, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1718)) __PYX_ERR(0, 3599, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3602
 *             return self.getToken(ModelicaParser.FINAL, 0)
 * 
 *         def short_class_definition(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Short_class_definitionContext,0)
 * 
 */
  __pyx_tuple__1719 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1719)) __PYX_ERR(0, 3602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1719);
  __Pyx_GIVEREF(__pyx_tuple__1719);
  __pyx_codeobj__1720 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1719, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_short_class_definition, 3602, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1720)) __PYX_ERR(0, 3602, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3606
 * 
 * 
 *         def component_clause1(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_clause1Context,0)
 * 
 */
  __pyx_tuple__1721 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1721)) __PYX_ERR(0, 3606, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1721);
  __Pyx_GIVEREF(__pyx_tuple__1721);
  __pyx_codeobj__1722 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1721, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_component_clause1, 3606, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1722)) __PYX_ERR(0, 3606, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3610
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_element_redeclaration
 * 
 */
  __pyx_tuple__1723 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1723)) __PYX_ERR(0, 3610, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1723);
  __Pyx_GIVEREF(__pyx_tuple__1723);
  __pyx_codeobj__1724 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1723, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 3610, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1724)) __PYX_ERR(0, 3610, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3613
 *             return ModelicaParser.RULE_element_redeclaration
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterElement_redeclaration" ):
 *                 listener.enterElement_redeclaration(self)
 */
  __pyx_tuple__1725 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1725)) __PYX_ERR(0, 3613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1725);
  __Pyx_GIVEREF(__pyx_tuple__1725);
  __pyx_codeobj__1726 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1725, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 3613, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1726)) __PYX_ERR(0, 3613, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3617
 *                 listener.enterElement_redeclaration(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitElement_redeclaration" ):
 *                 listener.exitElement_redeclaration(self)
 */
  __pyx_tuple__1727 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1727)) __PYX_ERR(0, 3617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1727);
  __Pyx_GIVEREF(__pyx_tuple__1727);
  __pyx_codeobj__1728 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1727, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 3617, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1728)) __PYX_ERR(0, 3617, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3621
 *                 listener.exitElement_redeclaration(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitElement_redeclaration" ):
 *                 return visitor.visitElement_redeclaration(self)
 */
  __pyx_tuple__1729 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1729)) __PYX_ERR(0, 3621, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1729);
  __Pyx_GIVEREF(__pyx_tuple__1729);
  __pyx_codeobj__1730 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1729, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 3621, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1730)) __PYX_ERR(0, 3621, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3630
 * 
 * 
 *     def element_redeclaration(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Element_redeclarationContext(self, self._ctx, self.state)
 */
  __pyx_tuple__1731 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la, __pyx_n_s_token, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1731)) __PYX_ERR(0, 3630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1731);
  __Pyx_GIVEREF(__pyx_tuple__1731);
  __pyx_codeobj__1732 = (PyObject*)__Pyx_PyCode_New(1, 0, 5, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1731, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_element_redeclaration, 3630, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1732)) __PYX_ERR(0, 3630, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3691
 *     class Element_replaceableContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1733 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1733)) __PYX_ERR(0, 3691, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1733);
  __Pyx_GIVEREF(__pyx_tuple__1733);
  __pyx_codeobj__1734 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1733, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 3691, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1734)) __PYX_ERR(0, 3691, __pyx_L1_error)
  __pyx_tuple__1735 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1735)) __PYX_ERR(0, 3691, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1735);
  __Pyx_GIVEREF(__pyx_tuple__1735);

  /* "pymola/generated/ModelicaParser.pyx":3695
 *             self.parser = parser
 * 
 *         def short_class_definition(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Short_class_definitionContext,0)
 * 
 */
  __pyx_tuple__1736 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1736)) __PYX_ERR(0, 3695, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1736);
  __Pyx_GIVEREF(__pyx_tuple__1736);
  __pyx_codeobj__1737 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1736, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_short_class_definition, 3695, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1737)) __PYX_ERR(0, 3695, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3699
 * 
 * 
 *         def component_clause1(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_clause1Context,0)
 * 
 */
  __pyx_tuple__1738 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1738)) __PYX_ERR(0, 3699, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1738);
  __Pyx_GIVEREF(__pyx_tuple__1738);
  __pyx_codeobj__1739 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1738, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_component_clause1, 3699, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1739)) __PYX_ERR(0, 3699, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3703
 * 
 * 
 *         def constraining_clause(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Constraining_clauseContext,0)
 * 
 */
  __pyx_tuple__1740 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1740)) __PYX_ERR(0, 3703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1740);
  __Pyx_GIVEREF(__pyx_tuple__1740);
  __pyx_codeobj__1741 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1740, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_constraining_clause, 3703, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1741)) __PYX_ERR(0, 3703, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3707
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_element_replaceable
 * 
 */
  __pyx_tuple__1742 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1742)) __PYX_ERR(0, 3707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1742);
  __Pyx_GIVEREF(__pyx_tuple__1742);
  __pyx_codeobj__1743 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1742, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 3707, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1743)) __PYX_ERR(0, 3707, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3710
 *             return ModelicaParser.RULE_element_replaceable
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterElement_replaceable" ):
 *                 listener.enterElement_replaceable(self)
 */
  __pyx_tuple__1744 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1744)) __PYX_ERR(0, 3710, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1744);
  __Pyx_GIVEREF(__pyx_tuple__1744);
  __pyx_codeobj__1745 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1744, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 3710, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1745)) __PYX_ERR(0, 3710, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3714
 *                 listener.enterElement_replaceable(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitElement_replaceable" ):
 *                 listener.exitElement_replaceable(self)
 */
  __pyx_tuple__1746 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1746)) __PYX_ERR(0, 3714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1746);
  __Pyx_GIVEREF(__pyx_tuple__1746);
  __pyx_codeobj__1747 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1746, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 3714, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1747)) __PYX_ERR(0, 3714, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3718
 *                 listener.exitElement_replaceable(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitElement_replaceable" ):
 *                 return visitor.visitElement_replaceable(self)
 */
  __pyx_tuple__1748 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1748)) __PYX_ERR(0, 3718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1748);
  __Pyx_GIVEREF(__pyx_tuple__1748);
  __pyx_codeobj__1749 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1748, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 3718, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1749)) __PYX_ERR(0, 3718, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3727
 * 
 * 
 *     def element_replaceable(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Element_replaceableContext(self, self._ctx, self.state)
 */
  __pyx_tuple__1750 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_token, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1750)) __PYX_ERR(0, 3727, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1750);
  __Pyx_GIVEREF(__pyx_tuple__1750);
  __pyx_codeobj__1751 = (PyObject*)__Pyx_PyCode_New(1, 0, 5, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1750, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_element_replaceable, 3727, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1751)) __PYX_ERR(0, 3727, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3768
 *     class Component_clause1Context(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1752 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1752)) __PYX_ERR(0, 3768, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1752);
  __Pyx_GIVEREF(__pyx_tuple__1752);
  __pyx_codeobj__1753 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1752, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 3768, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1753)) __PYX_ERR(0, 3768, __pyx_L1_error)
  __pyx_tuple__1754 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1754)) __PYX_ERR(0, 3768, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1754);
  __Pyx_GIVEREF(__pyx_tuple__1754);

  /* "pymola/generated/ModelicaParser.pyx":3772
 *             self.parser = parser
 * 
 *         def type_prefix(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Type_prefixContext,0)
 * 
 */
  __pyx_tuple__1755 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1755)) __PYX_ERR(0, 3772, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1755);
  __Pyx_GIVEREF(__pyx_tuple__1755);
  __pyx_codeobj__1756 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1755, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_type_prefix, 3772, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1756)) __PYX_ERR(0, 3772, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3776
 * 
 * 
 *         def type_specifier(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Type_specifierContext,0)
 * 
 */
  __pyx_tuple__1757 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1757)) __PYX_ERR(0, 3776, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1757);
  __Pyx_GIVEREF(__pyx_tuple__1757);
  __pyx_codeobj__1758 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1757, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_type_specifier, 3776, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1758)) __PYX_ERR(0, 3776, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3780
 * 
 * 
 *         def component_declaration1(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_declaration1Context,0)
 * 
 */
  __pyx_tuple__1759 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1759)) __PYX_ERR(0, 3780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1759);
  __Pyx_GIVEREF(__pyx_tuple__1759);
  __pyx_codeobj__1760 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1759, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_component_declaration1, 3780, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1760)) __PYX_ERR(0, 3780, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3784
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_component_clause1
 * 
 */
  __pyx_tuple__1761 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1761)) __PYX_ERR(0, 3784, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1761);
  __Pyx_GIVEREF(__pyx_tuple__1761);
  __pyx_codeobj__1762 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1761, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 3784, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1762)) __PYX_ERR(0, 3784, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3787
 *             return ModelicaParser.RULE_component_clause1
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterComponent_clause1" ):
 *                 listener.enterComponent_clause1(self)
 */
  __pyx_tuple__1763 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1763)) __PYX_ERR(0, 3787, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1763);
  __Pyx_GIVEREF(__pyx_tuple__1763);
  __pyx_codeobj__1764 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1763, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 3787, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1764)) __PYX_ERR(0, 3787, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3791
 *                 listener.enterComponent_clause1(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitComponent_clause1" ):
 *                 listener.exitComponent_clause1(self)
 */
  __pyx_tuple__1765 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1765)) __PYX_ERR(0, 3791, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1765);
  __Pyx_GIVEREF(__pyx_tuple__1765);
  __pyx_codeobj__1766 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1765, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 3791, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1766)) __PYX_ERR(0, 3791, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3795
 *                 listener.exitComponent_clause1(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitComponent_clause1" ):
 *                 return visitor.visitComponent_clause1(self)
 */
  __pyx_tuple__1767 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1767)) __PYX_ERR(0, 3795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1767);
  __Pyx_GIVEREF(__pyx_tuple__1767);
  __pyx_codeobj__1768 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1767, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 3795, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1768)) __PYX_ERR(0, 3795, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3804
 * 
 * 
 *     def component_clause1(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Component_clause1Context(self, self._ctx, self.state)
 */
  __pyx_tuple__1769 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1769)) __PYX_ERR(0, 3804, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1769);
  __Pyx_GIVEREF(__pyx_tuple__1769);
  __pyx_codeobj__1770 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1769, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_component_clause1, 3804, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1770)) __PYX_ERR(0, 3804, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3826
 *     class Component_declaration1Context(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1771 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1771)) __PYX_ERR(0, 3826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1771);
  __Pyx_GIVEREF(__pyx_tuple__1771);
  __pyx_codeobj__1772 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1771, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 3826, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1772)) __PYX_ERR(0, 3826, __pyx_L1_error)
  __pyx_tuple__1773 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1773)) __PYX_ERR(0, 3826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1773);
  __Pyx_GIVEREF(__pyx_tuple__1773);

  /* "pymola/generated/ModelicaParser.pyx":3830
 *             self.parser = parser
 * 
 *         def declaration(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.DeclarationContext,0)
 * 
 */
  __pyx_tuple__1774 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1774)) __PYX_ERR(0, 3830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1774);
  __Pyx_GIVEREF(__pyx_tuple__1774);
  __pyx_codeobj__1775 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1774, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_declaration, 3830, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1775)) __PYX_ERR(0, 3830, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3834
 * 
 * 
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */
  __pyx_tuple__1776 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1776)) __PYX_ERR(0, 3834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1776);
  __Pyx_GIVEREF(__pyx_tuple__1776);
  __pyx_codeobj__1777 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1776, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_comment, 3834, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1777)) __PYX_ERR(0, 3834, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3838
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_component_declaration1
 * 
 */
  __pyx_tuple__1778 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1778)) __PYX_ERR(0, 3838, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1778);
  __Pyx_GIVEREF(__pyx_tuple__1778);
  __pyx_codeobj__1779 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1778, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 3838, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1779)) __PYX_ERR(0, 3838, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3841
 *             return ModelicaParser.RULE_component_declaration1
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterComponent_declaration1" ):
 *                 listener.enterComponent_declaration1(self)
 */
  __pyx_tuple__1780 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1780)) __PYX_ERR(0, 3841, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1780);
  __Pyx_GIVEREF(__pyx_tuple__1780);
  __pyx_codeobj__1781 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1780, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 3841, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1781)) __PYX_ERR(0, 3841, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3845
 *                 listener.enterComponent_declaration1(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitComponent_declaration1" ):
 *                 listener.exitComponent_declaration1(self)
 */
  __pyx_tuple__1782 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1782)) __PYX_ERR(0, 3845, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1782);
  __Pyx_GIVEREF(__pyx_tuple__1782);
  __pyx_codeobj__1783 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1782, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 3845, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1783)) __PYX_ERR(0, 3845, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3849
 *                 listener.exitComponent_declaration1(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitComponent_declaration1" ):
 *                 return visitor.visitComponent_declaration1(self)
 */
  __pyx_tuple__1784 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1784)) __PYX_ERR(0, 3849, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1784);
  __Pyx_GIVEREF(__pyx_tuple__1784);
  __pyx_codeobj__1785 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1784, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 3849, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1785)) __PYX_ERR(0, 3849, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3858
 * 
 * 
 *     def component_declaration1(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Component_declaration1Context(self, self._ctx, self.state)
 */
  __pyx_tuple__1786 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1786)) __PYX_ERR(0, 3858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1786);
  __Pyx_GIVEREF(__pyx_tuple__1786);
  __pyx_codeobj__1787 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1786, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_component_declaration1, 3858, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1787)) __PYX_ERR(0, 3858, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3878
 *     class Short_class_definitionContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1788 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1788)) __PYX_ERR(0, 3878, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1788);
  __Pyx_GIVEREF(__pyx_tuple__1788);
  __pyx_codeobj__1789 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1788, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 3878, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1789)) __PYX_ERR(0, 3878, __pyx_L1_error)
  __pyx_tuple__1790 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1790)) __PYX_ERR(0, 3878, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1790);
  __Pyx_GIVEREF(__pyx_tuple__1790);

  /* "pymola/generated/ModelicaParser.pyx":3882
 *             self.parser = parser
 * 
 *         def class_prefixes(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_prefixesContext,0)
 * 
 */
  __pyx_tuple__1791 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1791)) __PYX_ERR(0, 3882, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1791);
  __Pyx_GIVEREF(__pyx_tuple__1791);
  __pyx_codeobj__1792 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1791, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_class_prefixes, 3882, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1792)) __PYX_ERR(0, 3882, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3886
 * 
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 */
  __pyx_tuple__1793 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1793)) __PYX_ERR(0, 3886, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1793);
  __Pyx_GIVEREF(__pyx_tuple__1793);
  __pyx_codeobj__1794 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1793, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_IDENT, 3886, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1794)) __PYX_ERR(0, 3886, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3889
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 *         def base_prefix(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Base_prefixContext,0)
 * 
 */
  __pyx_tuple__1795 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1795)) __PYX_ERR(0, 3889, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1795);
  __Pyx_GIVEREF(__pyx_tuple__1795);
  __pyx_codeobj__1796 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1795, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_base_prefix, 3889, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1796)) __PYX_ERR(0, 3889, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3893
 * 
 * 
 *         def component_reference(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 */
  __pyx_tuple__1797 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1797)) __PYX_ERR(0, 3893, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1797);
  __Pyx_GIVEREF(__pyx_tuple__1797);
  __pyx_codeobj__1798 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1797, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_component_reference, 3893, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1798)) __PYX_ERR(0, 3893, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3897
 * 
 * 
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */
  __pyx_tuple__1799 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1799)) __PYX_ERR(0, 3897, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1799);
  __Pyx_GIVEREF(__pyx_tuple__1799);
  __pyx_codeobj__1800 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1799, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_comment, 3897, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1800)) __PYX_ERR(0, 3897, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3901
 * 
 * 
 *         def array_subscripts(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Array_subscriptsContext,0)
 * 
 */
  __pyx_tuple__1801 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1801)) __PYX_ERR(0, 3901, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1801);
  __Pyx_GIVEREF(__pyx_tuple__1801);
  __pyx_codeobj__1802 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1801, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_array_subscripts, 3901, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1802)) __PYX_ERR(0, 3901, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3905
 * 
 * 
 *         def class_modification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_modificationContext,0)
 * 
 */
  __pyx_tuple__1803 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1803)) __PYX_ERR(0, 3905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1803);
  __Pyx_GIVEREF(__pyx_tuple__1803);
  __pyx_codeobj__1804 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1803, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_class_modification, 3905, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1804)) __PYX_ERR(0, 3905, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3909
 * 
 * 
 *         def enum_list(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Enum_listContext,0)
 * 
 */
  __pyx_tuple__1805 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1805)) __PYX_ERR(0, 3909, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1805);
  __Pyx_GIVEREF(__pyx_tuple__1805);
  __pyx_codeobj__1806 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1805, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enum_list, 3909, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1806)) __PYX_ERR(0, 3909, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3913
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_short_class_definition
 * 
 */
  __pyx_tuple__1807 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1807)) __PYX_ERR(0, 3913, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1807);
  __Pyx_GIVEREF(__pyx_tuple__1807);
  __pyx_codeobj__1808 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1807, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 3913, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1808)) __PYX_ERR(0, 3913, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3916
 *             return ModelicaParser.RULE_short_class_definition
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterShort_class_definition" ):
 *                 listener.enterShort_class_definition(self)
 */
  __pyx_tuple__1809 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1809)) __PYX_ERR(0, 3916, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1809);
  __Pyx_GIVEREF(__pyx_tuple__1809);
  __pyx_codeobj__1810 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1809, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 3916, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1810)) __PYX_ERR(0, 3916, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3920
 *                 listener.enterShort_class_definition(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitShort_class_definition" ):
 *                 listener.exitShort_class_definition(self)
 */
  __pyx_tuple__1811 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1811)) __PYX_ERR(0, 3920, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1811);
  __Pyx_GIVEREF(__pyx_tuple__1811);
  __pyx_codeobj__1812 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1811, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 3920, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1812)) __PYX_ERR(0, 3920, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3924
 *                 listener.exitShort_class_definition(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitShort_class_definition" ):
 *                 return visitor.visitShort_class_definition(self)
 */
  __pyx_tuple__1813 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1813)) __PYX_ERR(0, 3924, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1813);
  __Pyx_GIVEREF(__pyx_tuple__1813);
  __pyx_codeobj__1814 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1813, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 3924, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1814)) __PYX_ERR(0, 3924, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":3933
 * 
 * 
 *     def short_class_definition(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Short_class_definitionContext(self, self._ctx, self.state)
 */
  __pyx_tuple__1815 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_token, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1815)) __PYX_ERR(0, 3933, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1815);
  __Pyx_GIVEREF(__pyx_tuple__1815);
  __pyx_codeobj__1816 = (PyObject*)__Pyx_PyCode_New(1, 0, 5, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1815, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_short_class_definition, 3933, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1816)) __PYX_ERR(0, 3933, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4016
 *     class Equation_sectionContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1817 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1817)) __PYX_ERR(0, 4016, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1817);
  __Pyx_GIVEREF(__pyx_tuple__1817);
  __pyx_codeobj__1818 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1817, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 4016, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1818)) __PYX_ERR(0, 4016, __pyx_L1_error)
  __pyx_tuple__1819 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1819)) __PYX_ERR(0, 4016, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1819);
  __Pyx_GIVEREF(__pyx_tuple__1819);

  /* "pymola/generated/ModelicaParser.pyx":4020
 *             self.parser = parser
 * 
 *         def INITIAL(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.INITIAL, 0)
 * 
 */
  __pyx_tuple__1820 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1820)) __PYX_ERR(0, 4020, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1820);
  __Pyx_GIVEREF(__pyx_tuple__1820);
  __pyx_codeobj__1821 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1820, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_INITIAL, 4020, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1821)) __PYX_ERR(0, 4020, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4023
 *             return self.getToken(ModelicaParser.INITIAL, 0)
 * 
 *         def equation(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.EquationContext)
 */
  __pyx_tuple__1822 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__1822)) __PYX_ERR(0, 4023, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1822);
  __Pyx_GIVEREF(__pyx_tuple__1822);
  __pyx_codeobj__1823 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1822, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_equation, 4023, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1823)) __PYX_ERR(0, 4023, __pyx_L1_error)
  __pyx_tuple__1824 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__1824)) __PYX_ERR(0, 4023, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1824);
  __Pyx_GIVEREF(__pyx_tuple__1824);

  /* "pymola/generated/ModelicaParser.pyx":4030
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_equation_section
 * 
 */
  __pyx_tuple__1825 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1825)) __PYX_ERR(0, 4030, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1825);
  __Pyx_GIVEREF(__pyx_tuple__1825);
  __pyx_codeobj__1826 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1825, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 4030, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1826)) __PYX_ERR(0, 4030, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4033
 *             return ModelicaParser.RULE_equation_section
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterEquation_section" ):
 *                 listener.enterEquation_section(self)
 */
  __pyx_tuple__1827 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1827)) __PYX_ERR(0, 4033, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1827);
  __Pyx_GIVEREF(__pyx_tuple__1827);
  __pyx_codeobj__1828 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1827, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 4033, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1828)) __PYX_ERR(0, 4033, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4037
 *                 listener.enterEquation_section(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitEquation_section" ):
 *                 listener.exitEquation_section(self)
 */
  __pyx_tuple__1829 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1829)) __PYX_ERR(0, 4037, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1829);
  __Pyx_GIVEREF(__pyx_tuple__1829);
  __pyx_codeobj__1830 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1829, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 4037, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1830)) __PYX_ERR(0, 4037, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4041
 *                 listener.exitEquation_section(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitEquation_section" ):
 *                 return visitor.visitEquation_section(self)
 */
  __pyx_tuple__1831 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1831)) __PYX_ERR(0, 4041, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1831);
  __Pyx_GIVEREF(__pyx_tuple__1831);
  __pyx_codeobj__1832 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1831, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 4041, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1832)) __PYX_ERR(0, 4041, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4050
 * 
 * 
 *     def equation_section(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Equation_sectionContext(self, self._ctx, self.state)
 */
  __pyx_tuple__1833 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la, __pyx_n_s_alt, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1833)) __PYX_ERR(0, 4050, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1833);
  __Pyx_GIVEREF(__pyx_tuple__1833);
  __pyx_codeobj__1834 = (PyObject*)__Pyx_PyCode_New(1, 0, 5, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1833, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_equation_section, 4050, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1834)) __PYX_ERR(0, 4050, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4090
 *     class Algorithm_sectionContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1835 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1835)) __PYX_ERR(0, 4090, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1835);
  __Pyx_GIVEREF(__pyx_tuple__1835);
  __pyx_codeobj__1836 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1835, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 4090, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1836)) __PYX_ERR(0, 4090, __pyx_L1_error)
  __pyx_tuple__1837 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1837)) __PYX_ERR(0, 4090, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1837);
  __Pyx_GIVEREF(__pyx_tuple__1837);

  /* "pymola/generated/ModelicaParser.pyx":4094
 *             self.parser = parser
 * 
 *         def INITIAL(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.INITIAL, 0)
 * 
 */
  __pyx_tuple__1838 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1838)) __PYX_ERR(0, 4094, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1838);
  __Pyx_GIVEREF(__pyx_tuple__1838);
  __pyx_codeobj__1839 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1838, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_INITIAL, 4094, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1839)) __PYX_ERR(0, 4094, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4097
 *             return self.getToken(ModelicaParser.INITIAL, 0)
 * 
 *         def statement(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.StatementContext)
 */
  __pyx_tuple__1840 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__1840)) __PYX_ERR(0, 4097, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1840);
  __Pyx_GIVEREF(__pyx_tuple__1840);
  __pyx_codeobj__1841 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1840, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_statement, 4097, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1841)) __PYX_ERR(0, 4097, __pyx_L1_error)
  __pyx_tuple__1842 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__1842)) __PYX_ERR(0, 4097, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1842);
  __Pyx_GIVEREF(__pyx_tuple__1842);

  /* "pymola/generated/ModelicaParser.pyx":4104
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_algorithm_section
 * 
 */
  __pyx_tuple__1843 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1843)) __PYX_ERR(0, 4104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1843);
  __Pyx_GIVEREF(__pyx_tuple__1843);
  __pyx_codeobj__1844 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1843, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 4104, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1844)) __PYX_ERR(0, 4104, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4107
 *             return ModelicaParser.RULE_algorithm_section
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterAlgorithm_section" ):
 *                 listener.enterAlgorithm_section(self)
 */
  __pyx_tuple__1845 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1845)) __PYX_ERR(0, 4107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1845);
  __Pyx_GIVEREF(__pyx_tuple__1845);
  __pyx_codeobj__1846 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1845, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 4107, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1846)) __PYX_ERR(0, 4107, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4111
 *                 listener.enterAlgorithm_section(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitAlgorithm_section" ):
 *                 listener.exitAlgorithm_section(self)
 */
  __pyx_tuple__1847 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1847)) __PYX_ERR(0, 4111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1847);
  __Pyx_GIVEREF(__pyx_tuple__1847);
  __pyx_codeobj__1848 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1847, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 4111, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1848)) __PYX_ERR(0, 4111, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4115
 *                 listener.exitAlgorithm_section(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitAlgorithm_section" ):
 *                 return visitor.visitAlgorithm_section(self)
 */
  __pyx_tuple__1849 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1849)) __PYX_ERR(0, 4115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1849);
  __Pyx_GIVEREF(__pyx_tuple__1849);
  __pyx_codeobj__1850 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1849, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 4115, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1850)) __PYX_ERR(0, 4115, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4124
 * 
 * 
 *     def algorithm_section(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Algorithm_sectionContext(self, self._ctx, self.state)
 */
  __pyx_tuple__1851 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1851)) __PYX_ERR(0, 4124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1851);
  __Pyx_GIVEREF(__pyx_tuple__1851);
  __pyx_codeobj__1852 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1851, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_algorithm_section, 4124, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1852)) __PYX_ERR(0, 4124, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4163
 *     class Equation_optionsContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1853 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1853)) __PYX_ERR(0, 4163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1853);
  __Pyx_GIVEREF(__pyx_tuple__1853);
  __pyx_codeobj__1854 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1853, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 4163, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1854)) __PYX_ERR(0, 4163, __pyx_L1_error)
  __pyx_tuple__1855 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1855)) __PYX_ERR(0, 4163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1855);
  __Pyx_GIVEREF(__pyx_tuple__1855);

  /* "pymola/generated/ModelicaParser.pyx":4168
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_equation_options
 * 
 */
  __pyx_tuple__1856 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1856)) __PYX_ERR(0, 4168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1856);
  __Pyx_GIVEREF(__pyx_tuple__1856);
  __pyx_codeobj__1857 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1856, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 4168, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1857)) __PYX_ERR(0, 4168, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4172
 * 
 * 
 *         def copyFrom(self, ctx:ParserRuleContext):             # <<<<<<<<<<<<<<
 *             super().copyFrom(ctx)
 * 
 */
  __pyx_tuple__1858 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__1858)) __PYX_ERR(0, 4172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1858);
  __Pyx_GIVEREF(__pyx_tuple__1858);
  __pyx_codeobj__1859 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1858, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_copyFrom, 4172, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1859)) __PYX_ERR(0, 4172, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4179
 *     class Equation_whenContext(Equation_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__1860 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__1860)) __PYX_ERR(0, 4179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1860);
  __Pyx_GIVEREF(__pyx_tuple__1860);
  __pyx_codeobj__1861 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1860, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 4179, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1861)) __PYX_ERR(0, 4179, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4183
 *             self.copyFrom(ctx)
 * 
 *         def when_equation(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.When_equationContext,0)
 * 
 */
  __pyx_tuple__1862 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1862)) __PYX_ERR(0, 4183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1862);
  __Pyx_GIVEREF(__pyx_tuple__1862);
  __pyx_codeobj__1863 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1862, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_when_equation, 4183, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1863)) __PYX_ERR(0, 4183, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4187
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterEquation_when" ):
 *                 listener.enterEquation_when(self)
 */
  __pyx_tuple__1864 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1864)) __PYX_ERR(0, 4187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1864);
  __Pyx_GIVEREF(__pyx_tuple__1864);
  __pyx_codeobj__1865 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1864, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 4187, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1865)) __PYX_ERR(0, 4187, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4191
 *                 listener.enterEquation_when(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitEquation_when" ):
 *                 listener.exitEquation_when(self)
 */
  __pyx_tuple__1866 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1866)) __PYX_ERR(0, 4191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1866);
  __Pyx_GIVEREF(__pyx_tuple__1866);
  __pyx_codeobj__1867 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1866, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 4191, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1867)) __PYX_ERR(0, 4191, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4195
 *                 listener.exitEquation_when(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitEquation_when" ):
 *                 return visitor.visitEquation_when(self)
 */
  __pyx_tuple__1868 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1868)) __PYX_ERR(0, 4195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1868);
  __Pyx_GIVEREF(__pyx_tuple__1868);
  __pyx_codeobj__1869 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1868, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 4195, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1869)) __PYX_ERR(0, 4195, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4204
 *     class Equation_connect_clauseContext(Equation_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__1870 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__1870)) __PYX_ERR(0, 4204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1870);
  __Pyx_GIVEREF(__pyx_tuple__1870);
  __pyx_codeobj__1871 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1870, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 4204, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1871)) __PYX_ERR(0, 4204, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4208
 *             self.copyFrom(ctx)
 * 
 *         def connect_clause(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Connect_clauseContext,0)
 * 
 */
  __pyx_tuple__1872 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1872)) __PYX_ERR(0, 4208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1872);
  __Pyx_GIVEREF(__pyx_tuple__1872);
  __pyx_codeobj__1873 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1872, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_connect_clause, 4208, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1873)) __PYX_ERR(0, 4208, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4212
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterEquation_connect_clause" ):
 *                 listener.enterEquation_connect_clause(self)
 */
  __pyx_tuple__1874 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1874)) __PYX_ERR(0, 4212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1874);
  __Pyx_GIVEREF(__pyx_tuple__1874);
  __pyx_codeobj__1875 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1874, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 4212, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1875)) __PYX_ERR(0, 4212, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4216
 *                 listener.enterEquation_connect_clause(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitEquation_connect_clause" ):
 *                 listener.exitEquation_connect_clause(self)
 */
  __pyx_tuple__1876 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1876)) __PYX_ERR(0, 4216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1876);
  __Pyx_GIVEREF(__pyx_tuple__1876);
  __pyx_codeobj__1877 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1876, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 4216, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1877)) __PYX_ERR(0, 4216, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4220
 *                 listener.exitEquation_connect_clause(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitEquation_connect_clause" ):
 *                 return visitor.visitEquation_connect_clause(self)
 */
  __pyx_tuple__1878 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1878)) __PYX_ERR(0, 4220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1878);
  __Pyx_GIVEREF(__pyx_tuple__1878);
  __pyx_codeobj__1879 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1878, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 4220, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1879)) __PYX_ERR(0, 4220, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4229
 *     class Equation_forContext(Equation_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__1880 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__1880)) __PYX_ERR(0, 4229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1880);
  __Pyx_GIVEREF(__pyx_tuple__1880);
  __pyx_codeobj__1881 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1880, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 4229, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1881)) __PYX_ERR(0, 4229, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4233
 *             self.copyFrom(ctx)
 * 
 *         def for_equation(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.For_equationContext,0)
 * 
 */
  __pyx_tuple__1882 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1882)) __PYX_ERR(0, 4233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1882);
  __Pyx_GIVEREF(__pyx_tuple__1882);
  __pyx_codeobj__1883 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1882, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_for_equation, 4233, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1883)) __PYX_ERR(0, 4233, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4237
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterEquation_for" ):
 *                 listener.enterEquation_for(self)
 */
  __pyx_tuple__1884 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1884)) __PYX_ERR(0, 4237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1884);
  __Pyx_GIVEREF(__pyx_tuple__1884);
  __pyx_codeobj__1885 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1884, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 4237, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1885)) __PYX_ERR(0, 4237, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4241
 *                 listener.enterEquation_for(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitEquation_for" ):
 *                 listener.exitEquation_for(self)
 */
  __pyx_tuple__1886 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1886)) __PYX_ERR(0, 4241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1886);
  __Pyx_GIVEREF(__pyx_tuple__1886);
  __pyx_codeobj__1887 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1886, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 4241, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1887)) __PYX_ERR(0, 4241, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4245
 *                 listener.exitEquation_for(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitEquation_for" ):
 *                 return visitor.visitEquation_for(self)
 */
  __pyx_tuple__1888 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1888)) __PYX_ERR(0, 4245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1888);
  __Pyx_GIVEREF(__pyx_tuple__1888);
  __pyx_codeobj__1889 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1888, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 4245, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1889)) __PYX_ERR(0, 4245, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4254
 *     class Equation_functionContext(Equation_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__1890 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__1890)) __PYX_ERR(0, 4254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1890);
  __Pyx_GIVEREF(__pyx_tuple__1890);
  __pyx_codeobj__1891 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1890, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 4254, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1891)) __PYX_ERR(0, 4254, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4258
 *             self.copyFrom(ctx)
 * 
 *         def name(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)
 * 
 */
  __pyx_tuple__1892 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1892)) __PYX_ERR(0, 4258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1892);
  __Pyx_GIVEREF(__pyx_tuple__1892);
  __pyx_codeobj__1893 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1892, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_name, 4258, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1893)) __PYX_ERR(0, 4258, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4261
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)
 * 
 *         def function_call_args(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Function_call_argsContext,0)
 * 
 */
  __pyx_tuple__1894 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1894)) __PYX_ERR(0, 4261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1894);
  __Pyx_GIVEREF(__pyx_tuple__1894);
  __pyx_codeobj__1895 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1894, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_function_call_args, 4261, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1895)) __PYX_ERR(0, 4261, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4265
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterEquation_function" ):
 *                 listener.enterEquation_function(self)
 */
  __pyx_tuple__1896 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1896)) __PYX_ERR(0, 4265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1896);
  __Pyx_GIVEREF(__pyx_tuple__1896);
  __pyx_codeobj__1897 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1896, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 4265, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1897)) __PYX_ERR(0, 4265, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4269
 *                 listener.enterEquation_function(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitEquation_function" ):
 *                 listener.exitEquation_function(self)
 */
  __pyx_tuple__1898 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1898)) __PYX_ERR(0, 4269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1898);
  __Pyx_GIVEREF(__pyx_tuple__1898);
  __pyx_codeobj__1899 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1898, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 4269, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1899)) __PYX_ERR(0, 4269, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4273
 *                 listener.exitEquation_function(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitEquation_function" ):
 *                 return visitor.visitEquation_function(self)
 */
  __pyx_tuple__1900 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1900)) __PYX_ERR(0, 4273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1900);
  __Pyx_GIVEREF(__pyx_tuple__1900);
  __pyx_codeobj__1901 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1900, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 4273, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1901)) __PYX_ERR(0, 4273, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4282
 *     class Equation_simpleContext(Equation_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__1902 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__1902)) __PYX_ERR(0, 4282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1902);
  __Pyx_GIVEREF(__pyx_tuple__1902);
  __pyx_codeobj__1903 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1902, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 4282, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1903)) __PYX_ERR(0, 4282, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4286
 *             self.copyFrom(ctx)
 * 
 *         def simple_expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Simple_expressionContext,0)
 * 
 */
  __pyx_tuple__1904 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1904)) __PYX_ERR(0, 4286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1904);
  __Pyx_GIVEREF(__pyx_tuple__1904);
  __pyx_codeobj__1905 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1904, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_simple_expression, 4286, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1905)) __PYX_ERR(0, 4286, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4289
 *             return self.getTypedRuleContext(ModelicaParser.Simple_expressionContext,0)
 * 
 *         def expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 */
  __pyx_tuple__1906 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1906)) __PYX_ERR(0, 4289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1906);
  __Pyx_GIVEREF(__pyx_tuple__1906);
  __pyx_codeobj__1907 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1906, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_expression, 4289, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1907)) __PYX_ERR(0, 4289, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4293
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterEquation_simple" ):
 *                 listener.enterEquation_simple(self)
 */
  __pyx_tuple__1908 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1908)) __PYX_ERR(0, 4293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1908);
  __Pyx_GIVEREF(__pyx_tuple__1908);
  __pyx_codeobj__1909 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1908, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 4293, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1909)) __PYX_ERR(0, 4293, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4297
 *                 listener.enterEquation_simple(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitEquation_simple" ):
 *                 listener.exitEquation_simple(self)
 */
  __pyx_tuple__1910 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1910)) __PYX_ERR(0, 4297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1910);
  __Pyx_GIVEREF(__pyx_tuple__1910);
  __pyx_codeobj__1911 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1910, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 4297, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1911)) __PYX_ERR(0, 4297, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4301
 *                 listener.exitEquation_simple(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitEquation_simple" ):
 *                 return visitor.visitEquation_simple(self)
 */
  __pyx_tuple__1912 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1912)) __PYX_ERR(0, 4301, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1912);
  __Pyx_GIVEREF(__pyx_tuple__1912);
  __pyx_codeobj__1913 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1912, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 4301, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1913)) __PYX_ERR(0, 4301, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4310
 *     class Equation_ifContext(Equation_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__1914 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__1914)) __PYX_ERR(0, 4310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1914);
  __Pyx_GIVEREF(__pyx_tuple__1914);
  __pyx_codeobj__1915 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1914, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 4310, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1915)) __PYX_ERR(0, 4310, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4314
 *             self.copyFrom(ctx)
 * 
 *         def if_equation(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.If_equationContext,0)
 * 
 */
  __pyx_tuple__1916 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1916)) __PYX_ERR(0, 4314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1916);
  __Pyx_GIVEREF(__pyx_tuple__1916);
  __pyx_codeobj__1917 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1916, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_if_equation, 4314, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1917)) __PYX_ERR(0, 4314, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4318
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterEquation_if" ):
 *                 listener.enterEquation_if(self)
 */
  __pyx_tuple__1918 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1918)) __PYX_ERR(0, 4318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1918);
  __Pyx_GIVEREF(__pyx_tuple__1918);
  __pyx_codeobj__1919 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1918, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 4318, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1919)) __PYX_ERR(0, 4318, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4322
 *                 listener.enterEquation_if(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitEquation_if" ):
 *                 listener.exitEquation_if(self)
 */
  __pyx_tuple__1920 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1920)) __PYX_ERR(0, 4322, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1920);
  __Pyx_GIVEREF(__pyx_tuple__1920);
  __pyx_codeobj__1921 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1920, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 4322, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1921)) __PYX_ERR(0, 4322, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4326
 *                 listener.exitEquation_if(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitEquation_if" ):
 *                 return visitor.visitEquation_if(self)
 */
  __pyx_tuple__1922 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1922)) __PYX_ERR(0, 4326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1922);
  __Pyx_GIVEREF(__pyx_tuple__1922);
  __pyx_codeobj__1923 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1922, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 4326, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1923)) __PYX_ERR(0, 4326, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4334
 * 
 * 
 *     def equation_options(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Equation_optionsContext(self, self._ctx, self.state)
 */
  __pyx_tuple__1924 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la_2, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1924)) __PYX_ERR(0, 4334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1924);
  __Pyx_GIVEREF(__pyx_tuple__1924);
  __pyx_codeobj__1925 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1924, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_equation_options, 4334, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1925)) __PYX_ERR(0, 4334, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4401
 *     class EquationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1926 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1926)) __PYX_ERR(0, 4401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1926);
  __Pyx_GIVEREF(__pyx_tuple__1926);
  __pyx_codeobj__1927 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1926, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 4401, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1927)) __PYX_ERR(0, 4401, __pyx_L1_error)
  __pyx_tuple__1928 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1928)) __PYX_ERR(0, 4401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1928);
  __Pyx_GIVEREF(__pyx_tuple__1928);

  /* "pymola/generated/ModelicaParser.pyx":4405
 *             self.parser = parser
 * 
 *         def equation_options(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Equation_optionsContext,0)
 * 
 */
  __pyx_tuple__1929 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1929)) __PYX_ERR(0, 4405, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1929);
  __Pyx_GIVEREF(__pyx_tuple__1929);
  __pyx_codeobj__1930 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1929, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_equation_options, 4405, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1930)) __PYX_ERR(0, 4405, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4409
 * 
 * 
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */
  __pyx_tuple__1931 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1931)) __PYX_ERR(0, 4409, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1931);
  __Pyx_GIVEREF(__pyx_tuple__1931);
  __pyx_codeobj__1932 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1931, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_comment, 4409, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1932)) __PYX_ERR(0, 4409, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4413
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_equation
 * 
 */
  __pyx_tuple__1933 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1933)) __PYX_ERR(0, 4413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1933);
  __Pyx_GIVEREF(__pyx_tuple__1933);
  __pyx_codeobj__1934 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1933, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 4413, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1934)) __PYX_ERR(0, 4413, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4416
 *             return ModelicaParser.RULE_equation
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterEquation" ):
 *                 listener.enterEquation(self)
 */
  __pyx_tuple__1935 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1935)) __PYX_ERR(0, 4416, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1935);
  __Pyx_GIVEREF(__pyx_tuple__1935);
  __pyx_codeobj__1936 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1935, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 4416, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1936)) __PYX_ERR(0, 4416, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4420
 *                 listener.enterEquation(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitEquation" ):
 *                 listener.exitEquation(self)
 */
  __pyx_tuple__1937 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1937)) __PYX_ERR(0, 4420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1937);
  __Pyx_GIVEREF(__pyx_tuple__1937);
  __pyx_codeobj__1938 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1937, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 4420, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1938)) __PYX_ERR(0, 4420, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4424
 *                 listener.exitEquation(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitEquation" ):
 *                 return visitor.visitEquation(self)
 */
  __pyx_tuple__1939 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1939)) __PYX_ERR(0, 4424, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1939);
  __Pyx_GIVEREF(__pyx_tuple__1939);
  __pyx_codeobj__1940 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1939, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 4424, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1940)) __PYX_ERR(0, 4424, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4433
 * 
 * 
 *     def equation(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.EquationContext(self, self._ctx, self.state)
 */
  __pyx_tuple__1941 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_re); if (unlikely(!__pyx_tuple__1941)) __PYX_ERR(0, 4433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1941);
  __Pyx_GIVEREF(__pyx_tuple__1941);
  __pyx_codeobj__1942 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1941, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_equation, 4433, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1942)) __PYX_ERR(0, 4433, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4453
 *     class Statement_optionsContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__1943 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__1943)) __PYX_ERR(0, 4453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1943);
  __Pyx_GIVEREF(__pyx_tuple__1943);
  __pyx_codeobj__1944 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1943, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 4453, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1944)) __PYX_ERR(0, 4453, __pyx_L1_error)
  __pyx_tuple__1945 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__1945)) __PYX_ERR(0, 4453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1945);
  __Pyx_GIVEREF(__pyx_tuple__1945);

  /* "pymola/generated/ModelicaParser.pyx":4458
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_statement_options
 * 
 */
  __pyx_tuple__1946 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1946)) __PYX_ERR(0, 4458, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1946);
  __Pyx_GIVEREF(__pyx_tuple__1946);
  __pyx_codeobj__1947 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1946, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 4458, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1947)) __PYX_ERR(0, 4458, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4462
 * 
 * 
 *         def copyFrom(self, ctx:ParserRuleContext):             # <<<<<<<<<<<<<<
 *             super().copyFrom(ctx)
 * 
 */
  __pyx_tuple__1948 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__1948)) __PYX_ERR(0, 4462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1948);
  __Pyx_GIVEREF(__pyx_tuple__1948);
  __pyx_codeobj__1949 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1948, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_copyFrom, 4462, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1949)) __PYX_ERR(0, 4462, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4469
 *     class Statement_breakContext(Statement_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__1950 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__1950)) __PYX_ERR(0, 4469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1950);
  __Pyx_GIVEREF(__pyx_tuple__1950);
  __pyx_codeobj__1951 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1950, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 4469, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1951)) __PYX_ERR(0, 4469, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4474
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStatement_break" ):
 *                 listener.enterStatement_break(self)
 */
  __pyx_tuple__1952 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1952)) __PYX_ERR(0, 4474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1952);
  __Pyx_GIVEREF(__pyx_tuple__1952);
  __pyx_codeobj__1953 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1952, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 4474, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1953)) __PYX_ERR(0, 4474, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4478
 *                 listener.enterStatement_break(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStatement_break" ):
 *                 listener.exitStatement_break(self)
 */
  __pyx_tuple__1954 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1954)) __PYX_ERR(0, 4478, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1954);
  __Pyx_GIVEREF(__pyx_tuple__1954);
  __pyx_codeobj__1955 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1954, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 4478, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1955)) __PYX_ERR(0, 4478, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4482
 *                 listener.exitStatement_break(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStatement_break" ):
 *                 return visitor.visitStatement_break(self)
 */
  __pyx_tuple__1956 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1956)) __PYX_ERR(0, 4482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1956);
  __Pyx_GIVEREF(__pyx_tuple__1956);
  __pyx_codeobj__1957 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1956, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 4482, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1957)) __PYX_ERR(0, 4482, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4491
 *     class Statement_whileContext(Statement_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__1958 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__1958)) __PYX_ERR(0, 4491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1958);
  __Pyx_GIVEREF(__pyx_tuple__1958);
  __pyx_codeobj__1959 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1958, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 4491, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1959)) __PYX_ERR(0, 4491, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4495
 *             self.copyFrom(ctx)
 * 
 *         def while_statement(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.While_statementContext,0)
 * 
 */
  __pyx_tuple__1960 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1960)) __PYX_ERR(0, 4495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1960);
  __Pyx_GIVEREF(__pyx_tuple__1960);
  __pyx_codeobj__1961 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1960, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_while_statement, 4495, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1961)) __PYX_ERR(0, 4495, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4499
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStatement_while" ):
 *                 listener.enterStatement_while(self)
 */
  __pyx_tuple__1962 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1962)) __PYX_ERR(0, 4499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1962);
  __Pyx_GIVEREF(__pyx_tuple__1962);
  __pyx_codeobj__1963 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1962, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 4499, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1963)) __PYX_ERR(0, 4499, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4503
 *                 listener.enterStatement_while(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStatement_while" ):
 *                 listener.exitStatement_while(self)
 */
  __pyx_tuple__1964 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1964)) __PYX_ERR(0, 4503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1964);
  __Pyx_GIVEREF(__pyx_tuple__1964);
  __pyx_codeobj__1965 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1964, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 4503, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1965)) __PYX_ERR(0, 4503, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4507
 *                 listener.exitStatement_while(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStatement_while" ):
 *                 return visitor.visitStatement_while(self)
 */
  __pyx_tuple__1966 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1966)) __PYX_ERR(0, 4507, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1966);
  __Pyx_GIVEREF(__pyx_tuple__1966);
  __pyx_codeobj__1967 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1966, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 4507, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1967)) __PYX_ERR(0, 4507, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4516
 *     class Statement_component_functionContext(Statement_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__1968 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__1968)) __PYX_ERR(0, 4516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1968);
  __Pyx_GIVEREF(__pyx_tuple__1968);
  __pyx_codeobj__1969 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1968, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 4516, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1969)) __PYX_ERR(0, 4516, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4520
 *             self.copyFrom(ctx)
 * 
 *         def component_reference(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Component_referenceContext)
 */
  __pyx_tuple__1970 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__1970)) __PYX_ERR(0, 4520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1970);
  __Pyx_GIVEREF(__pyx_tuple__1970);
  __pyx_codeobj__1971 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1970, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_component_reference, 4520, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1971)) __PYX_ERR(0, 4520, __pyx_L1_error)
  __pyx_tuple__1972 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__1972)) __PYX_ERR(0, 4520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1972);
  __Pyx_GIVEREF(__pyx_tuple__1972);

  /* "pymola/generated/ModelicaParser.pyx":4526
 *                 return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,i)
 * 
 *         def function_call_args(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Function_call_argsContext,0)
 * 
 */
  __pyx_tuple__1973 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1973)) __PYX_ERR(0, 4526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1973);
  __Pyx_GIVEREF(__pyx_tuple__1973);
  __pyx_codeobj__1974 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1973, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_function_call_args, 4526, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1974)) __PYX_ERR(0, 4526, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4530
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStatement_component_function" ):
 *                 listener.enterStatement_component_function(self)
 */
  __pyx_tuple__1975 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1975)) __PYX_ERR(0, 4530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1975);
  __Pyx_GIVEREF(__pyx_tuple__1975);
  __pyx_codeobj__1976 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1975, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 4530, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1976)) __PYX_ERR(0, 4530, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4534
 *                 listener.enterStatement_component_function(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStatement_component_function" ):
 *                 listener.exitStatement_component_function(self)
 */
  __pyx_tuple__1977 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1977)) __PYX_ERR(0, 4534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1977);
  __Pyx_GIVEREF(__pyx_tuple__1977);
  __pyx_codeobj__1978 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1977, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 4534, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1978)) __PYX_ERR(0, 4534, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4538
 *                 listener.exitStatement_component_function(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStatement_component_function" ):
 *                 return visitor.visitStatement_component_function(self)
 */
  __pyx_tuple__1979 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1979)) __PYX_ERR(0, 4538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1979);
  __Pyx_GIVEREF(__pyx_tuple__1979);
  __pyx_codeobj__1980 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1979, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 4538, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1980)) __PYX_ERR(0, 4538, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4547
 *     class Statement_returnContext(Statement_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__1981 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__1981)) __PYX_ERR(0, 4547, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1981);
  __Pyx_GIVEREF(__pyx_tuple__1981);
  __pyx_codeobj__1982 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1981, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 4547, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1982)) __PYX_ERR(0, 4547, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4552
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStatement_return" ):
 *                 listener.enterStatement_return(self)
 */
  __pyx_tuple__1983 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1983)) __PYX_ERR(0, 4552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1983);
  __Pyx_GIVEREF(__pyx_tuple__1983);
  __pyx_codeobj__1984 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1983, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 4552, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1984)) __PYX_ERR(0, 4552, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4556
 *                 listener.enterStatement_return(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStatement_return" ):
 *                 listener.exitStatement_return(self)
 */
  __pyx_tuple__1985 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1985)) __PYX_ERR(0, 4556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1985);
  __Pyx_GIVEREF(__pyx_tuple__1985);
  __pyx_codeobj__1986 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1985, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 4556, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1986)) __PYX_ERR(0, 4556, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4560
 *                 listener.exitStatement_return(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStatement_return" ):
 *                 return visitor.visitStatement_return(self)
 */
  __pyx_tuple__1987 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1987)) __PYX_ERR(0, 4560, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1987);
  __Pyx_GIVEREF(__pyx_tuple__1987);
  __pyx_codeobj__1988 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1987, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 4560, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1988)) __PYX_ERR(0, 4560, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4569
 *     class Statement_forContext(Statement_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__1989 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__1989)) __PYX_ERR(0, 4569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1989);
  __Pyx_GIVEREF(__pyx_tuple__1989);
  __pyx_codeobj__1990 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1989, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 4569, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1990)) __PYX_ERR(0, 4569, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4573
 *             self.copyFrom(ctx)
 * 
 *         def for_statement(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.For_statementContext,0)
 * 
 */
  __pyx_tuple__1991 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__1991)) __PYX_ERR(0, 4573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1991);
  __Pyx_GIVEREF(__pyx_tuple__1991);
  __pyx_codeobj__1992 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1991, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_for_statement, 4573, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1992)) __PYX_ERR(0, 4573, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4577
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStatement_for" ):
 *                 listener.enterStatement_for(self)
 */
  __pyx_tuple__1993 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1993)) __PYX_ERR(0, 4577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1993);
  __Pyx_GIVEREF(__pyx_tuple__1993);
  __pyx_codeobj__1994 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1993, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 4577, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1994)) __PYX_ERR(0, 4577, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4581
 *                 listener.enterStatement_for(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStatement_for" ):
 *                 listener.exitStatement_for(self)
 */
  __pyx_tuple__1995 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__1995)) __PYX_ERR(0, 4581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1995);
  __Pyx_GIVEREF(__pyx_tuple__1995);
  __pyx_codeobj__1996 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1995, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 4581, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1996)) __PYX_ERR(0, 4581, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4585
 *                 listener.exitStatement_for(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStatement_for" ):
 *                 return visitor.visitStatement_for(self)
 */
  __pyx_tuple__1997 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__1997)) __PYX_ERR(0, 4585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1997);
  __Pyx_GIVEREF(__pyx_tuple__1997);
  __pyx_codeobj__1998 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1997, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 4585, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__1998)) __PYX_ERR(0, 4585, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4594
 *     class Statement_whenContext(Statement_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__1999 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__1999)) __PYX_ERR(0, 4594, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__1999);
  __Pyx_GIVEREF(__pyx_tuple__1999);
  __pyx_codeobj__2000 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__1999, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 4594, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2000)) __PYX_ERR(0, 4594, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4598
 *             self.copyFrom(ctx)
 * 
 *         def when_statement(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.When_statementContext,0)
 * 
 */
  __pyx_tuple__2001 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2001)) __PYX_ERR(0, 4598, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2001);
  __Pyx_GIVEREF(__pyx_tuple__2001);
  __pyx_codeobj__2002 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2001, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_when_statement, 4598, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2002)) __PYX_ERR(0, 4598, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4602
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStatement_when" ):
 *                 listener.enterStatement_when(self)
 */
  __pyx_tuple__2003 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2003)) __PYX_ERR(0, 4602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2003);
  __Pyx_GIVEREF(__pyx_tuple__2003);
  __pyx_codeobj__2004 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2003, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 4602, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2004)) __PYX_ERR(0, 4602, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4606
 *                 listener.enterStatement_when(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStatement_when" ):
 *                 listener.exitStatement_when(self)
 */
  __pyx_tuple__2005 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2005)) __PYX_ERR(0, 4606, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2005);
  __Pyx_GIVEREF(__pyx_tuple__2005);
  __pyx_codeobj__2006 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2005, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 4606, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2006)) __PYX_ERR(0, 4606, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4610
 *                 listener.exitStatement_when(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStatement_when" ):
 *                 return visitor.visitStatement_when(self)
 */
  __pyx_tuple__2007 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2007)) __PYX_ERR(0, 4610, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2007);
  __Pyx_GIVEREF(__pyx_tuple__2007);
  __pyx_codeobj__2008 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2007, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 4610, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2008)) __PYX_ERR(0, 4610, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4619
 *     class Statement_ifContext(Statement_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__2009 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__2009)) __PYX_ERR(0, 4619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2009);
  __Pyx_GIVEREF(__pyx_tuple__2009);
  __pyx_codeobj__2010 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2009, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 4619, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2010)) __PYX_ERR(0, 4619, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4623
 *             self.copyFrom(ctx)
 * 
 *         def if_statement(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.If_statementContext,0)
 * 
 */
  __pyx_tuple__2011 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2011)) __PYX_ERR(0, 4623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2011);
  __Pyx_GIVEREF(__pyx_tuple__2011);
  __pyx_codeobj__2012 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2011, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_if_statement, 4623, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2012)) __PYX_ERR(0, 4623, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4627
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStatement_if" ):
 *                 listener.enterStatement_if(self)
 */
  __pyx_tuple__2013 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2013)) __PYX_ERR(0, 4627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2013);
  __Pyx_GIVEREF(__pyx_tuple__2013);
  __pyx_codeobj__2014 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2013, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 4627, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2014)) __PYX_ERR(0, 4627, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4631
 *                 listener.enterStatement_if(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStatement_if" ):
 *                 listener.exitStatement_if(self)
 */
  __pyx_tuple__2015 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2015)) __PYX_ERR(0, 4631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2015);
  __Pyx_GIVEREF(__pyx_tuple__2015);
  __pyx_codeobj__2016 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2015, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 4631, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2016)) __PYX_ERR(0, 4631, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4635
 *                 listener.exitStatement_if(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStatement_if" ):
 *                 return visitor.visitStatement_if(self)
 */
  __pyx_tuple__2017 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2017)) __PYX_ERR(0, 4635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2017);
  __Pyx_GIVEREF(__pyx_tuple__2017);
  __pyx_codeobj__2018 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2017, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 4635, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2018)) __PYX_ERR(0, 4635, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4644
 *     class Statement_component_referenceContext(Statement_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__2019 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__2019)) __PYX_ERR(0, 4644, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2019);
  __Pyx_GIVEREF(__pyx_tuple__2019);
  __pyx_codeobj__2020 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2019, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 4644, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2020)) __PYX_ERR(0, 4644, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4648
 *             self.copyFrom(ctx)
 * 
 *         def component_reference(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 */
  __pyx_tuple__2021 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2021)) __PYX_ERR(0, 4648, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2021);
  __Pyx_GIVEREF(__pyx_tuple__2021);
  __pyx_codeobj__2022 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2021, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_component_reference, 4648, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2022)) __PYX_ERR(0, 4648, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4651
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 *         def expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 */
  __pyx_tuple__2023 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2023)) __PYX_ERR(0, 4651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2023);
  __Pyx_GIVEREF(__pyx_tuple__2023);
  __pyx_codeobj__2024 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2023, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_expression, 4651, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2024)) __PYX_ERR(0, 4651, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4654
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 *         def function_call_args(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Function_call_argsContext,0)
 * 
 */
  __pyx_tuple__2025 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2025)) __PYX_ERR(0, 4654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2025);
  __Pyx_GIVEREF(__pyx_tuple__2025);
  __pyx_codeobj__2026 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2025, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_function_call_args, 4654, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2026)) __PYX_ERR(0, 4654, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4658
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStatement_component_reference" ):
 *                 listener.enterStatement_component_reference(self)
 */
  __pyx_tuple__2027 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2027)) __PYX_ERR(0, 4658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2027);
  __Pyx_GIVEREF(__pyx_tuple__2027);
  __pyx_codeobj__2028 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2027, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 4658, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2028)) __PYX_ERR(0, 4658, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4662
 *                 listener.enterStatement_component_reference(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStatement_component_reference" ):
 *                 listener.exitStatement_component_reference(self)
 */
  __pyx_tuple__2029 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2029)) __PYX_ERR(0, 4662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2029);
  __Pyx_GIVEREF(__pyx_tuple__2029);
  __pyx_codeobj__2030 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2029, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 4662, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2030)) __PYX_ERR(0, 4662, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4666
 *                 listener.exitStatement_component_reference(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStatement_component_reference" ):
 *                 return visitor.visitStatement_component_reference(self)
 */
  __pyx_tuple__2031 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2031)) __PYX_ERR(0, 4666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2031);
  __Pyx_GIVEREF(__pyx_tuple__2031);
  __pyx_codeobj__2032 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2031, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 4666, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2032)) __PYX_ERR(0, 4666, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4674
 * 
 * 
 *     def statement_options(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Statement_optionsContext(self, self._ctx, self.state)
 */
  __pyx_tuple__2033 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_token, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__2033)) __PYX_ERR(0, 4674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2033);
  __Pyx_GIVEREF(__pyx_tuple__2033);
  __pyx_codeobj__2034 = (PyObject*)__Pyx_PyCode_New(1, 0, 5, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2033, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_statement_options, 4674, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2034)) __PYX_ERR(0, 4674, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4782
 *     class StatementContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__2035 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__2035)) __PYX_ERR(0, 4782, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2035);
  __Pyx_GIVEREF(__pyx_tuple__2035);
  __pyx_codeobj__2036 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2035, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 4782, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2036)) __PYX_ERR(0, 4782, __pyx_L1_error)
  __pyx_tuple__2037 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__2037)) __PYX_ERR(0, 4782, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2037);
  __Pyx_GIVEREF(__pyx_tuple__2037);

  /* "pymola/generated/ModelicaParser.pyx":4786
 *             self.parser = parser
 * 
 *         def statement_options(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Statement_optionsContext,0)
 * 
 */
  __pyx_tuple__2038 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2038)) __PYX_ERR(0, 4786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2038);
  __Pyx_GIVEREF(__pyx_tuple__2038);
  __pyx_codeobj__2039 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2038, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_statement_options, 4786, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2039)) __PYX_ERR(0, 4786, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4790
 * 
 * 
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */
  __pyx_tuple__2040 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2040)) __PYX_ERR(0, 4790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2040);
  __Pyx_GIVEREF(__pyx_tuple__2040);
  __pyx_codeobj__2041 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2040, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_comment, 4790, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2041)) __PYX_ERR(0, 4790, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4794
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_statement
 * 
 */
  __pyx_tuple__2042 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2042)) __PYX_ERR(0, 4794, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2042);
  __Pyx_GIVEREF(__pyx_tuple__2042);
  __pyx_codeobj__2043 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2042, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 4794, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2043)) __PYX_ERR(0, 4794, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4797
 *             return ModelicaParser.RULE_statement
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStatement" ):
 *                 listener.enterStatement(self)
 */
  __pyx_tuple__2044 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2044)) __PYX_ERR(0, 4797, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2044);
  __Pyx_GIVEREF(__pyx_tuple__2044);
  __pyx_codeobj__2045 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2044, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 4797, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2045)) __PYX_ERR(0, 4797, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4801
 *                 listener.enterStatement(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStatement" ):
 *                 listener.exitStatement(self)
 */
  __pyx_tuple__2046 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2046)) __PYX_ERR(0, 4801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2046);
  __Pyx_GIVEREF(__pyx_tuple__2046);
  __pyx_codeobj__2047 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2046, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 4801, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2047)) __PYX_ERR(0, 4801, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4805
 *                 listener.exitStatement(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStatement" ):
 *                 return visitor.visitStatement(self)
 */
  __pyx_tuple__2048 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2048)) __PYX_ERR(0, 4805, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2048);
  __Pyx_GIVEREF(__pyx_tuple__2048);
  __pyx_codeobj__2049 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2048, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 4805, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2049)) __PYX_ERR(0, 4805, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4814
 * 
 * 
 *     def statement(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.StatementContext(self, self._ctx, self.state)
 */
  __pyx_tuple__2050 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_re); if (unlikely(!__pyx_tuple__2050)) __PYX_ERR(0, 4814, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2050);
  __Pyx_GIVEREF(__pyx_tuple__2050);
  __pyx_codeobj__2051 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2050, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_statement, 4814, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2051)) __PYX_ERR(0, 4814, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4834
 *     class If_equationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__2052 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__2052)) __PYX_ERR(0, 4834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2052);
  __Pyx_GIVEREF(__pyx_tuple__2052);
  __pyx_codeobj__2053 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2052, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 4834, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2053)) __PYX_ERR(0, 4834, __pyx_L1_error)
  __pyx_tuple__2054 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__2054)) __PYX_ERR(0, 4834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2054);
  __Pyx_GIVEREF(__pyx_tuple__2054);

  /* "pymola/generated/ModelicaParser.pyx":4838
 *             self.parser = parser
 * 
 *         def expression(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 */
  __pyx_tuple__2055 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__2055)) __PYX_ERR(0, 4838, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2055);
  __Pyx_GIVEREF(__pyx_tuple__2055);
  __pyx_codeobj__2056 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2055, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_expression, 4838, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2056)) __PYX_ERR(0, 4838, __pyx_L1_error)
  __pyx_tuple__2057 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__2057)) __PYX_ERR(0, 4838, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2057);
  __Pyx_GIVEREF(__pyx_tuple__2057);

  /* "pymola/generated/ModelicaParser.pyx":4845
 * 
 * 
 *         def equation(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.EquationContext)
 */
  __pyx_tuple__2058 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__2058)) __PYX_ERR(0, 4845, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2058);
  __Pyx_GIVEREF(__pyx_tuple__2058);
  __pyx_codeobj__2059 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2058, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_equation, 4845, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2059)) __PYX_ERR(0, 4845, __pyx_L1_error)
  __pyx_tuple__2060 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__2060)) __PYX_ERR(0, 4845, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2060);
  __Pyx_GIVEREF(__pyx_tuple__2060);

  /* "pymola/generated/ModelicaParser.pyx":4852
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_if_equation
 * 
 */
  __pyx_tuple__2061 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2061)) __PYX_ERR(0, 4852, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2061);
  __Pyx_GIVEREF(__pyx_tuple__2061);
  __pyx_codeobj__2062 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2061, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 4852, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2062)) __PYX_ERR(0, 4852, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4855
 *             return ModelicaParser.RULE_if_equation
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterIf_equation" ):
 *                 listener.enterIf_equation(self)
 */
  __pyx_tuple__2063 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2063)) __PYX_ERR(0, 4855, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2063);
  __Pyx_GIVEREF(__pyx_tuple__2063);
  __pyx_codeobj__2064 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2063, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 4855, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2064)) __PYX_ERR(0, 4855, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4859
 *                 listener.enterIf_equation(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitIf_equation" ):
 *                 listener.exitIf_equation(self)
 */
  __pyx_tuple__2065 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2065)) __PYX_ERR(0, 4859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2065);
  __Pyx_GIVEREF(__pyx_tuple__2065);
  __pyx_codeobj__2066 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2065, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 4859, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2066)) __PYX_ERR(0, 4859, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4863
 *                 listener.exitIf_equation(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitIf_equation" ):
 *                 return visitor.visitIf_equation(self)
 */
  __pyx_tuple__2067 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2067)) __PYX_ERR(0, 4863, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2067);
  __Pyx_GIVEREF(__pyx_tuple__2067);
  __pyx_codeobj__2068 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2067, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 4863, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2068)) __PYX_ERR(0, 4863, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4872
 * 
 * 
 *     def if_equation(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.If_equationContext(self, self._ctx, self.state)
 */
  __pyx_tuple__2069 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_alt, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__2069)) __PYX_ERR(0, 4872, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2069);
  __Pyx_GIVEREF(__pyx_tuple__2069);
  __pyx_codeobj__2070 = (PyObject*)__Pyx_PyCode_New(1, 0, 5, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2069, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_if_equation, 4872, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2070)) __PYX_ERR(0, 4872, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4960
 *     class If_statementContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__2071 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__2071)) __PYX_ERR(0, 4960, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2071);
  __Pyx_GIVEREF(__pyx_tuple__2071);
  __pyx_codeobj__2072 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2071, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 4960, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2072)) __PYX_ERR(0, 4960, __pyx_L1_error)
  __pyx_tuple__2073 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__2073)) __PYX_ERR(0, 4960, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2073);
  __Pyx_GIVEREF(__pyx_tuple__2073);

  /* "pymola/generated/ModelicaParser.pyx":4964
 *             self.parser = parser
 * 
 *         def expression(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 */
  __pyx_tuple__2074 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__2074)) __PYX_ERR(0, 4964, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2074);
  __Pyx_GIVEREF(__pyx_tuple__2074);
  __pyx_codeobj__2075 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2074, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_expression, 4964, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2075)) __PYX_ERR(0, 4964, __pyx_L1_error)
  __pyx_tuple__2076 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__2076)) __PYX_ERR(0, 4964, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2076);
  __Pyx_GIVEREF(__pyx_tuple__2076);

  /* "pymola/generated/ModelicaParser.pyx":4971
 * 
 * 
 *         def statement(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.StatementContext)
 */
  __pyx_tuple__2077 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__2077)) __PYX_ERR(0, 4971, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2077);
  __Pyx_GIVEREF(__pyx_tuple__2077);
  __pyx_codeobj__2078 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2077, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_statement, 4971, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2078)) __PYX_ERR(0, 4971, __pyx_L1_error)
  __pyx_tuple__2079 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__2079)) __PYX_ERR(0, 4971, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2079);
  __Pyx_GIVEREF(__pyx_tuple__2079);

  /* "pymola/generated/ModelicaParser.pyx":4978
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_if_statement
 * 
 */
  __pyx_tuple__2080 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2080)) __PYX_ERR(0, 4978, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2080);
  __Pyx_GIVEREF(__pyx_tuple__2080);
  __pyx_codeobj__2081 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2080, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 4978, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2081)) __PYX_ERR(0, 4978, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4981
 *             return ModelicaParser.RULE_if_statement
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterIf_statement" ):
 *                 listener.enterIf_statement(self)
 */
  __pyx_tuple__2082 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2082)) __PYX_ERR(0, 4981, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2082);
  __Pyx_GIVEREF(__pyx_tuple__2082);
  __pyx_codeobj__2083 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2082, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 4981, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2083)) __PYX_ERR(0, 4981, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4985
 *                 listener.enterIf_statement(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitIf_statement" ):
 *                 listener.exitIf_statement(self)
 */
  __pyx_tuple__2084 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2084)) __PYX_ERR(0, 4985, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2084);
  __Pyx_GIVEREF(__pyx_tuple__2084);
  __pyx_codeobj__2085 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2084, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 4985, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2085)) __PYX_ERR(0, 4985, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4989
 *                 listener.exitIf_statement(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitIf_statement" ):
 *                 return visitor.visitIf_statement(self)
 */
  __pyx_tuple__2086 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2086)) __PYX_ERR(0, 4989, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2086);
  __Pyx_GIVEREF(__pyx_tuple__2086);
  __pyx_codeobj__2087 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2086, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 4989, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2087)) __PYX_ERR(0, 4989, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":4998
 * 
 * 
 *     def if_statement(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.If_statementContext(self, self._ctx, self.state)
 */
  __pyx_tuple__2088 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__2088)) __PYX_ERR(0, 4998, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2088);
  __Pyx_GIVEREF(__pyx_tuple__2088);
  __pyx_codeobj__2089 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2088, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_if_statement, 4998, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2089)) __PYX_ERR(0, 4998, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5083
 *     class For_equationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__2090 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__2090)) __PYX_ERR(0, 5083, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2090);
  __Pyx_GIVEREF(__pyx_tuple__2090);
  __pyx_codeobj__2091 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2090, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 5083, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2091)) __PYX_ERR(0, 5083, __pyx_L1_error)
  __pyx_tuple__2092 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__2092)) __PYX_ERR(0, 5083, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2092);
  __Pyx_GIVEREF(__pyx_tuple__2092);

  /* "pymola/generated/ModelicaParser.pyx":5087
 *             self.parser = parser
 * 
 *         def for_indices(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.For_indicesContext,0)
 * 
 */
  __pyx_tuple__2093 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2093)) __PYX_ERR(0, 5087, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2093);
  __Pyx_GIVEREF(__pyx_tuple__2093);
  __pyx_codeobj__2094 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2093, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_for_indices, 5087, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2094)) __PYX_ERR(0, 5087, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5091
 * 
 * 
 *         def equation(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.EquationContext)
 */
  __pyx_tuple__2095 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__2095)) __PYX_ERR(0, 5091, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2095);
  __Pyx_GIVEREF(__pyx_tuple__2095);
  __pyx_codeobj__2096 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2095, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_equation, 5091, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2096)) __PYX_ERR(0, 5091, __pyx_L1_error)
  __pyx_tuple__2097 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__2097)) __PYX_ERR(0, 5091, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2097);
  __Pyx_GIVEREF(__pyx_tuple__2097);

  /* "pymola/generated/ModelicaParser.pyx":5098
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_for_equation
 * 
 */
  __pyx_tuple__2098 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2098)) __PYX_ERR(0, 5098, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2098);
  __Pyx_GIVEREF(__pyx_tuple__2098);
  __pyx_codeobj__2099 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2098, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 5098, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2099)) __PYX_ERR(0, 5098, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5101
 *             return ModelicaParser.RULE_for_equation
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterFor_equation" ):
 *                 listener.enterFor_equation(self)
 */
  __pyx_tuple__2100 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2100)) __PYX_ERR(0, 5101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2100);
  __Pyx_GIVEREF(__pyx_tuple__2100);
  __pyx_codeobj__2101 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2100, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 5101, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2101)) __PYX_ERR(0, 5101, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5105
 *                 listener.enterFor_equation(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitFor_equation" ):
 *                 listener.exitFor_equation(self)
 */
  __pyx_tuple__2102 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2102)) __PYX_ERR(0, 5105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2102);
  __Pyx_GIVEREF(__pyx_tuple__2102);
  __pyx_codeobj__2103 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2102, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 5105, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2103)) __PYX_ERR(0, 5105, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5109
 *                 listener.exitFor_equation(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitFor_equation" ):
 *                 return visitor.visitFor_equation(self)
 */
  __pyx_tuple__2104 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2104)) __PYX_ERR(0, 5109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2104);
  __Pyx_GIVEREF(__pyx_tuple__2104);
  __pyx_codeobj__2105 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2104, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 5109, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2105)) __PYX_ERR(0, 5109, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5118
 * 
 * 
 *     def for_equation(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.For_equationContext(self, self._ctx, self.state)
 */
  __pyx_tuple__2106 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_alt, __pyx_n_s_re); if (unlikely(!__pyx_tuple__2106)) __PYX_ERR(0, 5118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2106);
  __Pyx_GIVEREF(__pyx_tuple__2106);
  __pyx_codeobj__2107 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2106, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_for_equation, 5118, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2107)) __PYX_ERR(0, 5118, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5157
 *     class For_statementContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__2108 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__2108)) __PYX_ERR(0, 5157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2108);
  __Pyx_GIVEREF(__pyx_tuple__2108);
  __pyx_codeobj__2109 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2108, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 5157, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2109)) __PYX_ERR(0, 5157, __pyx_L1_error)
  __pyx_tuple__2110 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__2110)) __PYX_ERR(0, 5157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2110);
  __Pyx_GIVEREF(__pyx_tuple__2110);

  /* "pymola/generated/ModelicaParser.pyx":5161
 *             self.parser = parser
 * 
 *         def for_indices(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.For_indicesContext,0)
 * 
 */
  __pyx_tuple__2111 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2111)) __PYX_ERR(0, 5161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2111);
  __Pyx_GIVEREF(__pyx_tuple__2111);
  __pyx_codeobj__2112 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2111, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_for_indices, 5161, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2112)) __PYX_ERR(0, 5161, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5165
 * 
 * 
 *         def statement(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.StatementContext)
 */
  __pyx_tuple__2113 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__2113)) __PYX_ERR(0, 5165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2113);
  __Pyx_GIVEREF(__pyx_tuple__2113);
  __pyx_codeobj__2114 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2113, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_statement, 5165, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2114)) __PYX_ERR(0, 5165, __pyx_L1_error)
  __pyx_tuple__2115 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__2115)) __PYX_ERR(0, 5165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2115);
  __Pyx_GIVEREF(__pyx_tuple__2115);

  /* "pymola/generated/ModelicaParser.pyx":5172
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_for_statement
 * 
 */
  __pyx_tuple__2116 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2116)) __PYX_ERR(0, 5172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2116);
  __Pyx_GIVEREF(__pyx_tuple__2116);
  __pyx_codeobj__2117 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2116, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 5172, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2117)) __PYX_ERR(0, 5172, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5175
 *             return ModelicaParser.RULE_for_statement
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterFor_statement" ):
 *                 listener.enterFor_statement(self)
 */
  __pyx_tuple__2118 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2118)) __PYX_ERR(0, 5175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2118);
  __Pyx_GIVEREF(__pyx_tuple__2118);
  __pyx_codeobj__2119 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2118, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 5175, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2119)) __PYX_ERR(0, 5175, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5179
 *                 listener.enterFor_statement(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitFor_statement" ):
 *                 listener.exitFor_statement(self)
 */
  __pyx_tuple__2120 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2120)) __PYX_ERR(0, 5179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2120);
  __Pyx_GIVEREF(__pyx_tuple__2120);
  __pyx_codeobj__2121 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2120, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 5179, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2121)) __PYX_ERR(0, 5179, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5183
 *                 listener.exitFor_statement(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitFor_statement" ):
 *                 return visitor.visitFor_statement(self)
 */
  __pyx_tuple__2122 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2122)) __PYX_ERR(0, 5183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2122);
  __Pyx_GIVEREF(__pyx_tuple__2122);
  __pyx_codeobj__2123 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2122, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 5183, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2123)) __PYX_ERR(0, 5183, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5192
 * 
 * 
 *     def for_statement(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.For_statementContext(self, self._ctx, self.state)
 */
  __pyx_tuple__2124 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__2124)) __PYX_ERR(0, 5192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2124);
  __Pyx_GIVEREF(__pyx_tuple__2124);
  __pyx_codeobj__2125 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2124, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_for_statement, 5192, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2125)) __PYX_ERR(0, 5192, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5231
 *     class For_indicesContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__2126 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__2126)) __PYX_ERR(0, 5231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2126);
  __Pyx_GIVEREF(__pyx_tuple__2126);
  __pyx_codeobj__2127 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2126, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 5231, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2127)) __PYX_ERR(0, 5231, __pyx_L1_error)
  __pyx_tuple__2128 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__2128)) __PYX_ERR(0, 5231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2128);
  __Pyx_GIVEREF(__pyx_tuple__2128);

  /* "pymola/generated/ModelicaParser.pyx":5235
 *             self.parser = parser
 * 
 *         def for_index(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.For_indexContext)
 */
  __pyx_tuple__2129 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__2129)) __PYX_ERR(0, 5235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2129);
  __Pyx_GIVEREF(__pyx_tuple__2129);
  __pyx_codeobj__2130 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2129, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_for_index, 5235, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2130)) __PYX_ERR(0, 5235, __pyx_L1_error)
  __pyx_tuple__2131 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__2131)) __PYX_ERR(0, 5235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2131);
  __Pyx_GIVEREF(__pyx_tuple__2131);

  /* "pymola/generated/ModelicaParser.pyx":5242
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_for_indices
 * 
 */
  __pyx_tuple__2132 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2132)) __PYX_ERR(0, 5242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2132);
  __Pyx_GIVEREF(__pyx_tuple__2132);
  __pyx_codeobj__2133 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 5242, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2133)) __PYX_ERR(0, 5242, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5245
 *             return ModelicaParser.RULE_for_indices
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterFor_indices" ):
 *                 listener.enterFor_indices(self)
 */
  __pyx_tuple__2134 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2134)) __PYX_ERR(0, 5245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2134);
  __Pyx_GIVEREF(__pyx_tuple__2134);
  __pyx_codeobj__2135 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2134, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 5245, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2135)) __PYX_ERR(0, 5245, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5249
 *                 listener.enterFor_indices(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitFor_indices" ):
 *                 listener.exitFor_indices(self)
 */
  __pyx_tuple__2136 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2136)) __PYX_ERR(0, 5249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2136);
  __Pyx_GIVEREF(__pyx_tuple__2136);
  __pyx_codeobj__2137 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2136, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 5249, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2137)) __PYX_ERR(0, 5249, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5253
 *                 listener.exitFor_indices(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitFor_indices" ):
 *                 return visitor.visitFor_indices(self)
 */
  __pyx_tuple__2138 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2138)) __PYX_ERR(0, 5253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2138);
  __Pyx_GIVEREF(__pyx_tuple__2138);
  __pyx_codeobj__2139 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2138, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 5253, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2139)) __PYX_ERR(0, 5253, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5262
 * 
 * 
 *     def for_indices(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.For_indicesContext(self, self._ctx, self.state)
 */
  __pyx_tuple__2140 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_alt, __pyx_n_s_re); if (unlikely(!__pyx_tuple__2140)) __PYX_ERR(0, 5262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2140);
  __Pyx_GIVEREF(__pyx_tuple__2140);
  __pyx_codeobj__2141 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2140, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_for_indices, 5262, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2141)) __PYX_ERR(0, 5262, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5293
 *     class For_indexContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__2142 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__2142)) __PYX_ERR(0, 5293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2142);
  __Pyx_GIVEREF(__pyx_tuple__2142);
  __pyx_codeobj__2143 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2142, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 5293, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2143)) __PYX_ERR(0, 5293, __pyx_L1_error)
  __pyx_tuple__2144 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__2144)) __PYX_ERR(0, 5293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2144);
  __Pyx_GIVEREF(__pyx_tuple__2144);

  /* "pymola/generated/ModelicaParser.pyx":5297
 *             self.parser = parser
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 */
  __pyx_tuple__2145 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2145)) __PYX_ERR(0, 5297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2145);
  __Pyx_GIVEREF(__pyx_tuple__2145);
  __pyx_codeobj__2146 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2145, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_IDENT, 5297, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2146)) __PYX_ERR(0, 5297, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5300
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 *         def expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 */
  __pyx_tuple__2147 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2147)) __PYX_ERR(0, 5300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2147);
  __Pyx_GIVEREF(__pyx_tuple__2147);
  __pyx_codeobj__2148 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2147, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_expression, 5300, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2148)) __PYX_ERR(0, 5300, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5304
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_for_index
 * 
 */
  __pyx_tuple__2149 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2149)) __PYX_ERR(0, 5304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2149);
  __Pyx_GIVEREF(__pyx_tuple__2149);
  __pyx_codeobj__2150 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2149, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 5304, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2150)) __PYX_ERR(0, 5304, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5307
 *             return ModelicaParser.RULE_for_index
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterFor_index" ):
 *                 listener.enterFor_index(self)
 */
  __pyx_tuple__2151 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2151)) __PYX_ERR(0, 5307, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2151);
  __Pyx_GIVEREF(__pyx_tuple__2151);
  __pyx_codeobj__2152 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2151, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 5307, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2152)) __PYX_ERR(0, 5307, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5311
 *                 listener.enterFor_index(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitFor_index" ):
 *                 listener.exitFor_index(self)
 */
  __pyx_tuple__2153 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2153)) __PYX_ERR(0, 5311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2153);
  __Pyx_GIVEREF(__pyx_tuple__2153);
  __pyx_codeobj__2154 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2153, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 5311, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2154)) __PYX_ERR(0, 5311, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5315
 *                 listener.exitFor_index(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitFor_index" ):
 *                 return visitor.visitFor_index(self)
 */
  __pyx_tuple__2155 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2155)) __PYX_ERR(0, 5315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2155);
  __Pyx_GIVEREF(__pyx_tuple__2155);
  __pyx_codeobj__2156 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2155, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 5315, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2156)) __PYX_ERR(0, 5315, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5324
 * 
 * 
 *     def for_index(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.For_indexContext(self, self._ctx, self.state)
 */
  __pyx_tuple__2157 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__2157)) __PYX_ERR(0, 5324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2157);
  __Pyx_GIVEREF(__pyx_tuple__2157);
  __pyx_codeobj__2158 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2157, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_for_index, 5324, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2158)) __PYX_ERR(0, 5324, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5353
 *     class While_statementContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__2159 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__2159)) __PYX_ERR(0, 5353, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2159);
  __Pyx_GIVEREF(__pyx_tuple__2159);
  __pyx_codeobj__2160 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2159, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 5353, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2160)) __PYX_ERR(0, 5353, __pyx_L1_error)
  __pyx_tuple__2161 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__2161)) __PYX_ERR(0, 5353, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2161);
  __Pyx_GIVEREF(__pyx_tuple__2161);

  /* "pymola/generated/ModelicaParser.pyx":5357
 *             self.parser = parser
 * 
 *         def expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 */
  __pyx_tuple__2162 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2162)) __PYX_ERR(0, 5357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2162);
  __Pyx_GIVEREF(__pyx_tuple__2162);
  __pyx_codeobj__2163 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2162, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_expression, 5357, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2163)) __PYX_ERR(0, 5357, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5361
 * 
 * 
 *         def statement(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.StatementContext)
 */
  __pyx_tuple__2164 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__2164)) __PYX_ERR(0, 5361, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2164);
  __Pyx_GIVEREF(__pyx_tuple__2164);
  __pyx_codeobj__2165 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2164, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_statement, 5361, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2165)) __PYX_ERR(0, 5361, __pyx_L1_error)
  __pyx_tuple__2166 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__2166)) __PYX_ERR(0, 5361, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2166);
  __Pyx_GIVEREF(__pyx_tuple__2166);

  /* "pymola/generated/ModelicaParser.pyx":5368
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_while_statement
 * 
 */
  __pyx_tuple__2167 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2167)) __PYX_ERR(0, 5368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2167);
  __Pyx_GIVEREF(__pyx_tuple__2167);
  __pyx_codeobj__2168 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 5368, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2168)) __PYX_ERR(0, 5368, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5371
 *             return ModelicaParser.RULE_while_statement
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterWhile_statement" ):
 *                 listener.enterWhile_statement(self)
 */
  __pyx_tuple__2169 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2169)) __PYX_ERR(0, 5371, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2169);
  __Pyx_GIVEREF(__pyx_tuple__2169);
  __pyx_codeobj__2170 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 5371, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2170)) __PYX_ERR(0, 5371, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5375
 *                 listener.enterWhile_statement(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitWhile_statement" ):
 *                 listener.exitWhile_statement(self)
 */
  __pyx_tuple__2171 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2171)) __PYX_ERR(0, 5375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2171);
  __Pyx_GIVEREF(__pyx_tuple__2171);
  __pyx_codeobj__2172 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2171, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 5375, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2172)) __PYX_ERR(0, 5375, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5379
 *                 listener.exitWhile_statement(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitWhile_statement" ):
 *                 return visitor.visitWhile_statement(self)
 */
  __pyx_tuple__2173 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2173)) __PYX_ERR(0, 5379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2173);
  __Pyx_GIVEREF(__pyx_tuple__2173);
  __pyx_codeobj__2174 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2173, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 5379, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2174)) __PYX_ERR(0, 5379, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5388
 * 
 * 
 *     def while_statement(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.While_statementContext(self, self._ctx, self.state)
 */
  __pyx_tuple__2175 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__2175)) __PYX_ERR(0, 5388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2175);
  __Pyx_GIVEREF(__pyx_tuple__2175);
  __pyx_codeobj__2176 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2175, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_while_statement, 5388, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2176)) __PYX_ERR(0, 5388, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5427
 *     class When_equationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__2177 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__2177)) __PYX_ERR(0, 5427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2177);
  __Pyx_GIVEREF(__pyx_tuple__2177);
  __pyx_codeobj__2178 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2177, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 5427, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2178)) __PYX_ERR(0, 5427, __pyx_L1_error)
  __pyx_tuple__2179 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__2179)) __PYX_ERR(0, 5427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2179);
  __Pyx_GIVEREF(__pyx_tuple__2179);

  /* "pymola/generated/ModelicaParser.pyx":5431
 *             self.parser = parser
 * 
 *         def expression(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 */
  __pyx_tuple__2180 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__2180)) __PYX_ERR(0, 5431, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2180);
  __Pyx_GIVEREF(__pyx_tuple__2180);
  __pyx_codeobj__2181 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2180, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_expression, 5431, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2181)) __PYX_ERR(0, 5431, __pyx_L1_error)
  __pyx_tuple__2182 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__2182)) __PYX_ERR(0, 5431, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2182);
  __Pyx_GIVEREF(__pyx_tuple__2182);

  /* "pymola/generated/ModelicaParser.pyx":5438
 * 
 * 
 *         def equation(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.EquationContext)
 */
  __pyx_tuple__2183 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__2183)) __PYX_ERR(0, 5438, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2183);
  __Pyx_GIVEREF(__pyx_tuple__2183);
  __pyx_codeobj__2184 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2183, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_equation, 5438, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2184)) __PYX_ERR(0, 5438, __pyx_L1_error)
  __pyx_tuple__2185 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__2185)) __PYX_ERR(0, 5438, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2185);
  __Pyx_GIVEREF(__pyx_tuple__2185);

  /* "pymola/generated/ModelicaParser.pyx":5445
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_when_equation
 * 
 */
  __pyx_tuple__2186 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2186)) __PYX_ERR(0, 5445, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2186);
  __Pyx_GIVEREF(__pyx_tuple__2186);
  __pyx_codeobj__2187 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2186, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 5445, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2187)) __PYX_ERR(0, 5445, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5448
 *             return ModelicaParser.RULE_when_equation
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterWhen_equation" ):
 *                 listener.enterWhen_equation(self)
 */
  __pyx_tuple__2188 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2188)) __PYX_ERR(0, 5448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2188);
  __Pyx_GIVEREF(__pyx_tuple__2188);
  __pyx_codeobj__2189 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2188, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 5448, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2189)) __PYX_ERR(0, 5448, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5452
 *                 listener.enterWhen_equation(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitWhen_equation" ):
 *                 listener.exitWhen_equation(self)
 */
  __pyx_tuple__2190 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2190)) __PYX_ERR(0, 5452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2190);
  __Pyx_GIVEREF(__pyx_tuple__2190);
  __pyx_codeobj__2191 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2190, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 5452, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2191)) __PYX_ERR(0, 5452, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5456
 *                 listener.exitWhen_equation(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitWhen_equation" ):
 *                 return visitor.visitWhen_equation(self)
 */
  __pyx_tuple__2192 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2192)) __PYX_ERR(0, 5456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2192);
  __Pyx_GIVEREF(__pyx_tuple__2192);
  __pyx_codeobj__2193 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2192, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 5456, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2193)) __PYX_ERR(0, 5456, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5465
 * 
 * 
 *     def when_equation(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.When_equationContext(self, self._ctx, self.state)
 */
  __pyx_tuple__2194 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_alt, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__2194)) __PYX_ERR(0, 5465, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2194);
  __Pyx_GIVEREF(__pyx_tuple__2194);
  __pyx_codeobj__2195 = (PyObject*)__Pyx_PyCode_New(1, 0, 5, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2194, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_when_equation, 5465, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2195)) __PYX_ERR(0, 5465, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5532
 *     class When_statementContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__2196 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__2196)) __PYX_ERR(0, 5532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2196);
  __Pyx_GIVEREF(__pyx_tuple__2196);
  __pyx_codeobj__2197 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2196, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 5532, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2197)) __PYX_ERR(0, 5532, __pyx_L1_error)
  __pyx_tuple__2198 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__2198)) __PYX_ERR(0, 5532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2198);
  __Pyx_GIVEREF(__pyx_tuple__2198);

  /* "pymola/generated/ModelicaParser.pyx":5536
 *             self.parser = parser
 * 
 *         def expression(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 */
  __pyx_tuple__2199 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__2199)) __PYX_ERR(0, 5536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2199);
  __Pyx_GIVEREF(__pyx_tuple__2199);
  __pyx_codeobj__2200 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2199, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_expression, 5536, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2200)) __PYX_ERR(0, 5536, __pyx_L1_error)
  __pyx_tuple__2201 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__2201)) __PYX_ERR(0, 5536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2201);
  __Pyx_GIVEREF(__pyx_tuple__2201);

  /* "pymola/generated/ModelicaParser.pyx":5543
 * 
 * 
 *         def statement(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.StatementContext)
 */
  __pyx_tuple__2202 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__2202)) __PYX_ERR(0, 5543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2202);
  __Pyx_GIVEREF(__pyx_tuple__2202);
  __pyx_codeobj__2203 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2202, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_statement, 5543, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2203)) __PYX_ERR(0, 5543, __pyx_L1_error)
  __pyx_tuple__2204 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__2204)) __PYX_ERR(0, 5543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2204);
  __Pyx_GIVEREF(__pyx_tuple__2204);

  /* "pymola/generated/ModelicaParser.pyx":5550
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_when_statement
 * 
 */
  __pyx_tuple__2205 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2205)) __PYX_ERR(0, 5550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2205);
  __Pyx_GIVEREF(__pyx_tuple__2205);
  __pyx_codeobj__2206 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2205, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 5550, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2206)) __PYX_ERR(0, 5550, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5553
 *             return ModelicaParser.RULE_when_statement
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterWhen_statement" ):
 *                 listener.enterWhen_statement(self)
 */
  __pyx_tuple__2207 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2207)) __PYX_ERR(0, 5553, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2207);
  __Pyx_GIVEREF(__pyx_tuple__2207);
  __pyx_codeobj__2208 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2207, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 5553, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2208)) __PYX_ERR(0, 5553, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5557
 *                 listener.enterWhen_statement(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitWhen_statement" ):
 *                 listener.exitWhen_statement(self)
 */
  __pyx_tuple__2209 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2209)) __PYX_ERR(0, 5557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2209);
  __Pyx_GIVEREF(__pyx_tuple__2209);
  __pyx_codeobj__2210 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 5557, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2210)) __PYX_ERR(0, 5557, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5561
 *                 listener.exitWhen_statement(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitWhen_statement" ):
 *                 return visitor.visitWhen_statement(self)
 */
  __pyx_tuple__2211 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2211)) __PYX_ERR(0, 5561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2211);
  __Pyx_GIVEREF(__pyx_tuple__2211);
  __pyx_codeobj__2212 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2211, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 5561, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2212)) __PYX_ERR(0, 5561, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5570
 * 
 * 
 *     def when_statement(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.When_statementContext(self, self._ctx, self.state)
 */
  __pyx_tuple__2213 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__2213)) __PYX_ERR(0, 5570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2213);
  __Pyx_GIVEREF(__pyx_tuple__2213);
  __pyx_codeobj__2214 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2213, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_when_statement, 5570, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2214)) __PYX_ERR(0, 5570, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5635
 *     class Connect_clauseContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__2215 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__2215)) __PYX_ERR(0, 5635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2215);
  __Pyx_GIVEREF(__pyx_tuple__2215);
  __pyx_codeobj__2216 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2215, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 5635, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2216)) __PYX_ERR(0, 5635, __pyx_L1_error)
  __pyx_tuple__2217 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__2217)) __PYX_ERR(0, 5635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2217);
  __Pyx_GIVEREF(__pyx_tuple__2217);

  /* "pymola/generated/ModelicaParser.pyx":5639
 *             self.parser = parser
 * 
 *         def component_reference(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Component_referenceContext)
 */
  __pyx_tuple__2218 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__2218)) __PYX_ERR(0, 5639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2218);
  __Pyx_GIVEREF(__pyx_tuple__2218);
  __pyx_codeobj__2219 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2218, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_component_reference, 5639, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2219)) __PYX_ERR(0, 5639, __pyx_L1_error)
  __pyx_tuple__2220 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__2220)) __PYX_ERR(0, 5639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2220);
  __Pyx_GIVEREF(__pyx_tuple__2220);

  /* "pymola/generated/ModelicaParser.pyx":5646
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_connect_clause
 * 
 */
  __pyx_tuple__2221 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2221)) __PYX_ERR(0, 5646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2221);
  __Pyx_GIVEREF(__pyx_tuple__2221);
  __pyx_codeobj__2222 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2221, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 5646, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2222)) __PYX_ERR(0, 5646, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5649
 *             return ModelicaParser.RULE_connect_clause
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterConnect_clause" ):
 *                 listener.enterConnect_clause(self)
 */
  __pyx_tuple__2223 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2223)) __PYX_ERR(0, 5649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2223);
  __Pyx_GIVEREF(__pyx_tuple__2223);
  __pyx_codeobj__2224 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2223, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 5649, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2224)) __PYX_ERR(0, 5649, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5653
 *                 listener.enterConnect_clause(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitConnect_clause" ):
 *                 listener.exitConnect_clause(self)
 */
  __pyx_tuple__2225 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2225)) __PYX_ERR(0, 5653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2225);
  __Pyx_GIVEREF(__pyx_tuple__2225);
  __pyx_codeobj__2226 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2225, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 5653, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2226)) __PYX_ERR(0, 5653, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5657
 *                 listener.exitConnect_clause(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitConnect_clause" ):
 *                 return visitor.visitConnect_clause(self)
 */
  __pyx_tuple__2227 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2227)) __PYX_ERR(0, 5657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2227);
  __Pyx_GIVEREF(__pyx_tuple__2227);
  __pyx_codeobj__2228 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2227, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 5657, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2228)) __PYX_ERR(0, 5657, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5666
 * 
 * 
 *     def connect_clause(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Connect_clauseContext(self, self._ctx, self.state)
 */
  __pyx_tuple__2229 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_re); if (unlikely(!__pyx_tuple__2229)) __PYX_ERR(0, 5666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2229);
  __Pyx_GIVEREF(__pyx_tuple__2229);
  __pyx_codeobj__2230 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2229, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_connect_clause, 5666, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2230)) __PYX_ERR(0, 5666, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5694
 *     class ExpressionContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__2231 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__2231)) __PYX_ERR(0, 5694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2231);
  __Pyx_GIVEREF(__pyx_tuple__2231);
  __pyx_codeobj__2232 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2231, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 5694, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2232)) __PYX_ERR(0, 5694, __pyx_L1_error)
  __pyx_tuple__2233 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__2233)) __PYX_ERR(0, 5694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2233);
  __Pyx_GIVEREF(__pyx_tuple__2233);

  /* "pymola/generated/ModelicaParser.pyx":5699
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_expression
 * 
 */
  __pyx_tuple__2234 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2234)) __PYX_ERR(0, 5699, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2234);
  __Pyx_GIVEREF(__pyx_tuple__2234);
  __pyx_codeobj__2235 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2234, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 5699, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2235)) __PYX_ERR(0, 5699, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5703
 * 
 * 
 *         def copyFrom(self, ctx:ParserRuleContext):             # <<<<<<<<<<<<<<
 *             super().copyFrom(ctx)
 * 
 */
  __pyx_tuple__2236 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__2236)) __PYX_ERR(0, 5703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2236);
  __Pyx_GIVEREF(__pyx_tuple__2236);
  __pyx_codeobj__2237 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2236, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_copyFrom, 5703, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2237)) __PYX_ERR(0, 5703, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5710
 *     class Expression_ifContext(ExpressionContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExpressionContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__2238 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__2238)) __PYX_ERR(0, 5710, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2238);
  __Pyx_GIVEREF(__pyx_tuple__2238);
  __pyx_codeobj__2239 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2238, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 5710, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2239)) __PYX_ERR(0, 5710, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5714
 *             self.copyFrom(ctx)
 * 
 *         def expression(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 */
  __pyx_tuple__2240 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__2240)) __PYX_ERR(0, 5714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2240);
  __Pyx_GIVEREF(__pyx_tuple__2240);
  __pyx_codeobj__2241 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2240, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_expression, 5714, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2241)) __PYX_ERR(0, 5714, __pyx_L1_error)
  __pyx_tuple__2242 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__2242)) __PYX_ERR(0, 5714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2242);
  __Pyx_GIVEREF(__pyx_tuple__2242);

  /* "pymola/generated/ModelicaParser.pyx":5721
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpression_if" ):
 *                 listener.enterExpression_if(self)
 */
  __pyx_tuple__2243 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2243)) __PYX_ERR(0, 5721, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2243);
  __Pyx_GIVEREF(__pyx_tuple__2243);
  __pyx_codeobj__2244 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2243, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 5721, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2244)) __PYX_ERR(0, 5721, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5725
 *                 listener.enterExpression_if(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpression_if" ):
 *                 listener.exitExpression_if(self)
 */
  __pyx_tuple__2245 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2245)) __PYX_ERR(0, 5725, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2245);
  __Pyx_GIVEREF(__pyx_tuple__2245);
  __pyx_codeobj__2246 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2245, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 5725, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2246)) __PYX_ERR(0, 5725, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5729
 *                 listener.exitExpression_if(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpression_if" ):
 *                 return visitor.visitExpression_if(self)
 */
  __pyx_tuple__2247 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2247)) __PYX_ERR(0, 5729, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2247);
  __Pyx_GIVEREF(__pyx_tuple__2247);
  __pyx_codeobj__2248 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2247, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 5729, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2248)) __PYX_ERR(0, 5729, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5738
 *     class Expression_simpleContext(ExpressionContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExpressionContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__2249 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__2249)) __PYX_ERR(0, 5738, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2249);
  __Pyx_GIVEREF(__pyx_tuple__2249);
  __pyx_codeobj__2250 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2249, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 5738, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2250)) __PYX_ERR(0, 5738, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5742
 *             self.copyFrom(ctx)
 * 
 *         def simple_expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Simple_expressionContext,0)
 * 
 */
  __pyx_tuple__2251 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2251)) __PYX_ERR(0, 5742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2251);
  __Pyx_GIVEREF(__pyx_tuple__2251);
  __pyx_codeobj__2252 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2251, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_simple_expression, 5742, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2252)) __PYX_ERR(0, 5742, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5746
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpression_simple" ):
 *                 listener.enterExpression_simple(self)
 */
  __pyx_tuple__2253 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2253)) __PYX_ERR(0, 5746, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2253);
  __Pyx_GIVEREF(__pyx_tuple__2253);
  __pyx_codeobj__2254 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2253, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 5746, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2254)) __PYX_ERR(0, 5746, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5750
 *                 listener.enterExpression_simple(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpression_simple" ):
 *                 listener.exitExpression_simple(self)
 */
  __pyx_tuple__2255 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2255)) __PYX_ERR(0, 5750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2255);
  __Pyx_GIVEREF(__pyx_tuple__2255);
  __pyx_codeobj__2256 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2255, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 5750, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2256)) __PYX_ERR(0, 5750, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5754
 *                 listener.exitExpression_simple(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpression_simple" ):
 *                 return visitor.visitExpression_simple(self)
 */
  __pyx_tuple__2257 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2257)) __PYX_ERR(0, 5754, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2257);
  __Pyx_GIVEREF(__pyx_tuple__2257);
  __pyx_codeobj__2258 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2257, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 5754, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2258)) __PYX_ERR(0, 5754, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5762
 * 
 * 
 *     def expression(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.ExpressionContext(self, self._ctx, self.state)
 */
  __pyx_tuple__2259 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_token, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__2259)) __PYX_ERR(0, 5762, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2259);
  __Pyx_GIVEREF(__pyx_tuple__2259);
  __pyx_codeobj__2260 = (PyObject*)__Pyx_PyCode_New(1, 0, 5, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2259, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_expression, 5762, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2260)) __PYX_ERR(0, 5762, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5822
 *     class Simple_expressionContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__2261 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__2261)) __PYX_ERR(0, 5822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2261);
  __Pyx_GIVEREF(__pyx_tuple__2261);
  __pyx_codeobj__2262 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2261, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 5822, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2262)) __PYX_ERR(0, 5822, __pyx_L1_error)
  __pyx_tuple__2263 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__2263)) __PYX_ERR(0, 5822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2263);
  __Pyx_GIVEREF(__pyx_tuple__2263);

  /* "pymola/generated/ModelicaParser.pyx":5826
 *             self.parser = parser
 * 
 *         def expr(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 */
  __pyx_tuple__2264 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__2264)) __PYX_ERR(0, 5826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2264);
  __Pyx_GIVEREF(__pyx_tuple__2264);
  __pyx_codeobj__2265 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2264, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_expr, 5826, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2265)) __PYX_ERR(0, 5826, __pyx_L1_error)
  __pyx_tuple__2266 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__2266)) __PYX_ERR(0, 5826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2266);
  __Pyx_GIVEREF(__pyx_tuple__2266);

  /* "pymola/generated/ModelicaParser.pyx":5833
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_simple_expression
 * 
 */
  __pyx_tuple__2267 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2267)) __PYX_ERR(0, 5833, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2267);
  __Pyx_GIVEREF(__pyx_tuple__2267);
  __pyx_codeobj__2268 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2267, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 5833, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2268)) __PYX_ERR(0, 5833, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5836
 *             return ModelicaParser.RULE_simple_expression
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterSimple_expression" ):
 *                 listener.enterSimple_expression(self)
 */
  __pyx_tuple__2269 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2269)) __PYX_ERR(0, 5836, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2269);
  __Pyx_GIVEREF(__pyx_tuple__2269);
  __pyx_codeobj__2270 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2269, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 5836, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2270)) __PYX_ERR(0, 5836, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5840
 *                 listener.enterSimple_expression(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitSimple_expression" ):
 *                 listener.exitSimple_expression(self)
 */
  __pyx_tuple__2271 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2271)) __PYX_ERR(0, 5840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2271);
  __Pyx_GIVEREF(__pyx_tuple__2271);
  __pyx_codeobj__2272 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2271, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 5840, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2272)) __PYX_ERR(0, 5840, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5844
 *                 listener.exitSimple_expression(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitSimple_expression" ):
 *                 return visitor.visitSimple_expression(self)
 */
  __pyx_tuple__2273 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2273)) __PYX_ERR(0, 5844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2273);
  __Pyx_GIVEREF(__pyx_tuple__2273);
  __pyx_codeobj__2274 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2273, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 5844, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2274)) __PYX_ERR(0, 5844, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5853
 * 
 * 
 *     def simple_expression(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Simple_expressionContext(self, self._ctx, self.state)
 */
  __pyx_tuple__2275 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__2275)) __PYX_ERR(0, 5853, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2275);
  __Pyx_GIVEREF(__pyx_tuple__2275);
  __pyx_codeobj__2276 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2275, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_simple_expression, 5853, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2276)) __PYX_ERR(0, 5853, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5892
 *     class ExprContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__2277 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__2277)) __PYX_ERR(0, 5892, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2277);
  __Pyx_GIVEREF(__pyx_tuple__2277);
  __pyx_codeobj__2278 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2277, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 5892, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2278)) __PYX_ERR(0, 5892, __pyx_L1_error)
  __pyx_tuple__2279 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__2279)) __PYX_ERR(0, 5892, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2279);
  __Pyx_GIVEREF(__pyx_tuple__2279);

  /* "pymola/generated/ModelicaParser.pyx":5897
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_expr
 * 
 */
  __pyx_tuple__2280 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2280)) __PYX_ERR(0, 5897, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2280);
  __Pyx_GIVEREF(__pyx_tuple__2280);
  __pyx_codeobj__2281 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2280, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 5897, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2281)) __PYX_ERR(0, 5897, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5901
 * 
 * 
 *         def copyFrom(self, ctx:ParserRuleContext):             # <<<<<<<<<<<<<<
 *             super().copyFrom(ctx)
 * 
 */
  __pyx_tuple__2282 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__2282)) __PYX_ERR(0, 5901, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2282);
  __Pyx_GIVEREF(__pyx_tuple__2282);
  __pyx_codeobj__2283 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2282, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_copyFrom, 5901, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2283)) __PYX_ERR(0, 5901, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5907
 *     class Expr_negContext(ExprContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.op = None # Token
 */
  __pyx_tuple__2284 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__2284)) __PYX_ERR(0, 5907, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2284);
  __Pyx_GIVEREF(__pyx_tuple__2284);
  __pyx_codeobj__2285 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2284, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 5907, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2285)) __PYX_ERR(0, 5907, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5912
 *             self.copyFrom(ctx)
 * 
 *         def expr(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExprContext,0)
 * 
 */
  __pyx_tuple__2286 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2286)) __PYX_ERR(0, 5912, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2286);
  __Pyx_GIVEREF(__pyx_tuple__2286);
  __pyx_codeobj__2287 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2286, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_expr, 5912, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2287)) __PYX_ERR(0, 5912, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5916
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpr_neg" ):
 *                 listener.enterExpr_neg(self)
 */
  __pyx_tuple__2288 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2288)) __PYX_ERR(0, 5916, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2288);
  __Pyx_GIVEREF(__pyx_tuple__2288);
  __pyx_codeobj__2289 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2288, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 5916, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2289)) __PYX_ERR(0, 5916, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5920
 *                 listener.enterExpr_neg(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpr_neg" ):
 *                 listener.exitExpr_neg(self)
 */
  __pyx_tuple__2290 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2290)) __PYX_ERR(0, 5920, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2290);
  __Pyx_GIVEREF(__pyx_tuple__2290);
  __pyx_codeobj__2291 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2290, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 5920, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2291)) __PYX_ERR(0, 5920, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5924
 *                 listener.exitExpr_neg(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpr_neg" ):
 *                 return visitor.visitExpr_neg(self)
 */
  __pyx_tuple__2292 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2292)) __PYX_ERR(0, 5924, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2292);
  __Pyx_GIVEREF(__pyx_tuple__2292);
  __pyx_codeobj__2293 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2292, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 5924, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2293)) __PYX_ERR(0, 5924, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5933
 *     class Expr_addContext(ExprContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.op = None # Token
 */
  __pyx_tuple__2294 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__2294)) __PYX_ERR(0, 5933, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2294);
  __Pyx_GIVEREF(__pyx_tuple__2294);
  __pyx_codeobj__2295 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2294, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 5933, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2295)) __PYX_ERR(0, 5933, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5938
 *             self.copyFrom(ctx)
 * 
 *         def expr(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 */
  __pyx_tuple__2296 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__2296)) __PYX_ERR(0, 5938, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2296);
  __Pyx_GIVEREF(__pyx_tuple__2296);
  __pyx_codeobj__2297 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2296, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_expr, 5938, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2297)) __PYX_ERR(0, 5938, __pyx_L1_error)
  __pyx_tuple__2298 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__2298)) __PYX_ERR(0, 5938, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2298);
  __Pyx_GIVEREF(__pyx_tuple__2298);

  /* "pymola/generated/ModelicaParser.pyx":5945
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpr_add" ):
 *                 listener.enterExpr_add(self)
 */
  __pyx_tuple__2299 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2299)) __PYX_ERR(0, 5945, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2299);
  __Pyx_GIVEREF(__pyx_tuple__2299);
  __pyx_codeobj__2300 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2299, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 5945, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2300)) __PYX_ERR(0, 5945, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5949
 *                 listener.enterExpr_add(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpr_add" ):
 *                 listener.exitExpr_add(self)
 */
  __pyx_tuple__2301 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2301)) __PYX_ERR(0, 5949, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2301);
  __Pyx_GIVEREF(__pyx_tuple__2301);
  __pyx_codeobj__2302 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2301, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 5949, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2302)) __PYX_ERR(0, 5949, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5953
 *                 listener.exitExpr_add(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpr_add" ):
 *                 return visitor.visitExpr_add(self)
 */
  __pyx_tuple__2303 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2303)) __PYX_ERR(0, 5953, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2303);
  __Pyx_GIVEREF(__pyx_tuple__2303);
  __pyx_codeobj__2304 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2303, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 5953, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2304)) __PYX_ERR(0, 5953, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5962
 *     class Expr_expContext(ExprContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.op = None # Token
 */
  __pyx_tuple__2305 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__2305)) __PYX_ERR(0, 5962, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2305);
  __Pyx_GIVEREF(__pyx_tuple__2305);
  __pyx_codeobj__2306 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2305, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 5962, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2306)) __PYX_ERR(0, 5962, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5967
 *             self.copyFrom(ctx)
 * 
 *         def primary(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.PrimaryContext)
 */
  __pyx_tuple__2307 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__2307)) __PYX_ERR(0, 5967, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2307);
  __Pyx_GIVEREF(__pyx_tuple__2307);
  __pyx_codeobj__2308 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2307, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_primary, 5967, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2308)) __PYX_ERR(0, 5967, __pyx_L1_error)
  __pyx_tuple__2309 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__2309)) __PYX_ERR(0, 5967, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2309);
  __Pyx_GIVEREF(__pyx_tuple__2309);

  /* "pymola/generated/ModelicaParser.pyx":5974
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpr_exp" ):
 *                 listener.enterExpr_exp(self)
 */
  __pyx_tuple__2310 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2310)) __PYX_ERR(0, 5974, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2310);
  __Pyx_GIVEREF(__pyx_tuple__2310);
  __pyx_codeobj__2311 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2310, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 5974, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2311)) __PYX_ERR(0, 5974, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5978
 *                 listener.enterExpr_exp(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpr_exp" ):
 *                 listener.exitExpr_exp(self)
 */
  __pyx_tuple__2312 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2312)) __PYX_ERR(0, 5978, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2312);
  __Pyx_GIVEREF(__pyx_tuple__2312);
  __pyx_codeobj__2313 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2312, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 5978, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2313)) __PYX_ERR(0, 5978, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5982
 *                 listener.exitExpr_exp(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpr_exp" ):
 *                 return visitor.visitExpr_exp(self)
 */
  __pyx_tuple__2314 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2314)) __PYX_ERR(0, 5982, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2314);
  __Pyx_GIVEREF(__pyx_tuple__2314);
  __pyx_codeobj__2315 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2314, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 5982, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2315)) __PYX_ERR(0, 5982, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5991
 *     class Expr_orContext(ExprContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__2316 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__2316)) __PYX_ERR(0, 5991, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2316);
  __Pyx_GIVEREF(__pyx_tuple__2316);
  __pyx_codeobj__2317 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2316, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 5991, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2317)) __PYX_ERR(0, 5991, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":5995
 *             self.copyFrom(ctx)
 * 
 *         def expr(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 */
  __pyx_tuple__2318 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__2318)) __PYX_ERR(0, 5995, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2318);
  __Pyx_GIVEREF(__pyx_tuple__2318);
  __pyx_codeobj__2319 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2318, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_expr, 5995, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2319)) __PYX_ERR(0, 5995, __pyx_L1_error)
  __pyx_tuple__2320 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__2320)) __PYX_ERR(0, 5995, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2320);
  __Pyx_GIVEREF(__pyx_tuple__2320);

  /* "pymola/generated/ModelicaParser.pyx":6002
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpr_or" ):
 *                 listener.enterExpr_or(self)
 */
  __pyx_tuple__2321 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2321)) __PYX_ERR(0, 6002, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2321);
  __Pyx_GIVEREF(__pyx_tuple__2321);
  __pyx_codeobj__2322 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2321, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 6002, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2322)) __PYX_ERR(0, 6002, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6006
 *                 listener.enterExpr_or(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpr_or" ):
 *                 listener.exitExpr_or(self)
 */
  __pyx_tuple__2323 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2323)) __PYX_ERR(0, 6006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2323);
  __Pyx_GIVEREF(__pyx_tuple__2323);
  __pyx_codeobj__2324 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2323, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 6006, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2324)) __PYX_ERR(0, 6006, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6010
 *                 listener.exitExpr_or(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpr_or" ):
 *                 return visitor.visitExpr_or(self)
 */
  __pyx_tuple__2325 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2325)) __PYX_ERR(0, 6010, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2325);
  __Pyx_GIVEREF(__pyx_tuple__2325);
  __pyx_codeobj__2326 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2325, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 6010, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2326)) __PYX_ERR(0, 6010, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6019
 *     class Expr_primaryContext(ExprContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__2327 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__2327)) __PYX_ERR(0, 6019, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2327);
  __Pyx_GIVEREF(__pyx_tuple__2327);
  __pyx_codeobj__2328 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2327, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 6019, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2328)) __PYX_ERR(0, 6019, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6023
 *             self.copyFrom(ctx)
 * 
 *         def primary(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.PrimaryContext,0)
 * 
 */
  __pyx_tuple__2329 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2329)) __PYX_ERR(0, 6023, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2329);
  __Pyx_GIVEREF(__pyx_tuple__2329);
  __pyx_codeobj__2330 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2329, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_primary, 6023, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2330)) __PYX_ERR(0, 6023, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6027
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpr_primary" ):
 *                 listener.enterExpr_primary(self)
 */
  __pyx_tuple__2331 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2331)) __PYX_ERR(0, 6027, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2331);
  __Pyx_GIVEREF(__pyx_tuple__2331);
  __pyx_codeobj__2332 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2331, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 6027, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2332)) __PYX_ERR(0, 6027, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6031
 *                 listener.enterExpr_primary(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpr_primary" ):
 *                 listener.exitExpr_primary(self)
 */
  __pyx_tuple__2333 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2333)) __PYX_ERR(0, 6031, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2333);
  __Pyx_GIVEREF(__pyx_tuple__2333);
  __pyx_codeobj__2334 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2333, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 6031, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2334)) __PYX_ERR(0, 6031, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6035
 *                 listener.exitExpr_primary(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpr_primary" ):
 *                 return visitor.visitExpr_primary(self)
 */
  __pyx_tuple__2335 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2335)) __PYX_ERR(0, 6035, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2335);
  __Pyx_GIVEREF(__pyx_tuple__2335);
  __pyx_codeobj__2336 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2335, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 6035, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2336)) __PYX_ERR(0, 6035, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6044
 *     class Expr_andContext(ExprContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__2337 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__2337)) __PYX_ERR(0, 6044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2337);
  __Pyx_GIVEREF(__pyx_tuple__2337);
  __pyx_codeobj__2338 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2337, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 6044, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2338)) __PYX_ERR(0, 6044, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6048
 *             self.copyFrom(ctx)
 * 
 *         def expr(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 */
  __pyx_tuple__2339 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__2339)) __PYX_ERR(0, 6048, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2339);
  __Pyx_GIVEREF(__pyx_tuple__2339);
  __pyx_codeobj__2340 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2339, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_expr, 6048, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2340)) __PYX_ERR(0, 6048, __pyx_L1_error)
  __pyx_tuple__2341 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__2341)) __PYX_ERR(0, 6048, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2341);
  __Pyx_GIVEREF(__pyx_tuple__2341);

  /* "pymola/generated/ModelicaParser.pyx":6055
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpr_and" ):
 *                 listener.enterExpr_and(self)
 */
  __pyx_tuple__2342 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2342)) __PYX_ERR(0, 6055, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2342);
  __Pyx_GIVEREF(__pyx_tuple__2342);
  __pyx_codeobj__2343 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2342, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 6055, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2343)) __PYX_ERR(0, 6055, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6059
 *                 listener.enterExpr_and(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpr_and" ):
 *                 listener.exitExpr_and(self)
 */
  __pyx_tuple__2344 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2344)) __PYX_ERR(0, 6059, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2344);
  __Pyx_GIVEREF(__pyx_tuple__2344);
  __pyx_codeobj__2345 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2344, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 6059, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2345)) __PYX_ERR(0, 6059, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6063
 *                 listener.exitExpr_and(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpr_and" ):
 *                 return visitor.visitExpr_and(self)
 */
  __pyx_tuple__2346 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2346)) __PYX_ERR(0, 6063, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2346);
  __Pyx_GIVEREF(__pyx_tuple__2346);
  __pyx_codeobj__2347 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2346, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 6063, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2347)) __PYX_ERR(0, 6063, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6072
 *     class Expr_relContext(ExprContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.op = None # Token
 */
  __pyx_tuple__2348 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__2348)) __PYX_ERR(0, 6072, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2348);
  __Pyx_GIVEREF(__pyx_tuple__2348);
  __pyx_codeobj__2349 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2348, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 6072, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2349)) __PYX_ERR(0, 6072, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6077
 *             self.copyFrom(ctx)
 * 
 *         def expr(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 */
  __pyx_tuple__2350 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__2350)) __PYX_ERR(0, 6077, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2350);
  __Pyx_GIVEREF(__pyx_tuple__2350);
  __pyx_codeobj__2351 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2350, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_expr, 6077, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2351)) __PYX_ERR(0, 6077, __pyx_L1_error)
  __pyx_tuple__2352 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__2352)) __PYX_ERR(0, 6077, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2352);
  __Pyx_GIVEREF(__pyx_tuple__2352);

  /* "pymola/generated/ModelicaParser.pyx":6084
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpr_rel" ):
 *                 listener.enterExpr_rel(self)
 */
  __pyx_tuple__2353 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2353)) __PYX_ERR(0, 6084, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2353);
  __Pyx_GIVEREF(__pyx_tuple__2353);
  __pyx_codeobj__2354 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2353, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 6084, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2354)) __PYX_ERR(0, 6084, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6088
 *                 listener.enterExpr_rel(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpr_rel" ):
 *                 listener.exitExpr_rel(self)
 */
  __pyx_tuple__2355 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2355)) __PYX_ERR(0, 6088, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2355);
  __Pyx_GIVEREF(__pyx_tuple__2355);
  __pyx_codeobj__2356 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2355, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 6088, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2356)) __PYX_ERR(0, 6088, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6092
 *                 listener.exitExpr_rel(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpr_rel" ):
 *                 return visitor.visitExpr_rel(self)
 */
  __pyx_tuple__2357 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2357)) __PYX_ERR(0, 6092, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2357);
  __Pyx_GIVEREF(__pyx_tuple__2357);
  __pyx_codeobj__2358 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2357, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 6092, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2358)) __PYX_ERR(0, 6092, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6101
 *     class Expr_notContext(ExprContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__2359 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__2359)) __PYX_ERR(0, 6101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2359);
  __Pyx_GIVEREF(__pyx_tuple__2359);
  __pyx_codeobj__2360 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2359, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 6101, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2360)) __PYX_ERR(0, 6101, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6105
 *             self.copyFrom(ctx)
 * 
 *         def expr(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExprContext,0)
 * 
 */
  __pyx_tuple__2361 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2361)) __PYX_ERR(0, 6105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2361);
  __Pyx_GIVEREF(__pyx_tuple__2361);
  __pyx_codeobj__2362 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2361, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_expr, 6105, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2362)) __PYX_ERR(0, 6105, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6109
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpr_not" ):
 *                 listener.enterExpr_not(self)
 */
  __pyx_tuple__2363 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2363)) __PYX_ERR(0, 6109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2363);
  __Pyx_GIVEREF(__pyx_tuple__2363);
  __pyx_codeobj__2364 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2363, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 6109, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2364)) __PYX_ERR(0, 6109, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6113
 *                 listener.enterExpr_not(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpr_not" ):
 *                 listener.exitExpr_not(self)
 */
  __pyx_tuple__2365 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2365)) __PYX_ERR(0, 6113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2365);
  __Pyx_GIVEREF(__pyx_tuple__2365);
  __pyx_codeobj__2366 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2365, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 6113, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2366)) __PYX_ERR(0, 6113, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6117
 *                 listener.exitExpr_not(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpr_not" ):
 *                 return visitor.visitExpr_not(self)
 */
  __pyx_tuple__2367 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2367)) __PYX_ERR(0, 6117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2367);
  __Pyx_GIVEREF(__pyx_tuple__2367);
  __pyx_codeobj__2368 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2367, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 6117, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2368)) __PYX_ERR(0, 6117, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6126
 *     class Expr_mulContext(ExprContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.op = None # Token
 */
  __pyx_tuple__2369 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__2369)) __PYX_ERR(0, 6126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2369);
  __Pyx_GIVEREF(__pyx_tuple__2369);
  __pyx_codeobj__2370 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2369, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 6126, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2370)) __PYX_ERR(0, 6126, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6131
 *             self.copyFrom(ctx)
 * 
 *         def expr(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 */
  __pyx_tuple__2371 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__2371)) __PYX_ERR(0, 6131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2371);
  __Pyx_GIVEREF(__pyx_tuple__2371);
  __pyx_codeobj__2372 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2371, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_expr, 6131, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2372)) __PYX_ERR(0, 6131, __pyx_L1_error)
  __pyx_tuple__2373 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__2373)) __PYX_ERR(0, 6131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2373);
  __Pyx_GIVEREF(__pyx_tuple__2373);

  /* "pymola/generated/ModelicaParser.pyx":6138
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpr_mul" ):
 *                 listener.enterExpr_mul(self)
 */
  __pyx_tuple__2374 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2374)) __PYX_ERR(0, 6138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2374);
  __Pyx_GIVEREF(__pyx_tuple__2374);
  __pyx_codeobj__2375 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2374, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 6138, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2375)) __PYX_ERR(0, 6138, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6142
 *                 listener.enterExpr_mul(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpr_mul" ):
 *                 listener.exitExpr_mul(self)
 */
  __pyx_tuple__2376 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2376)) __PYX_ERR(0, 6142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2376);
  __Pyx_GIVEREF(__pyx_tuple__2376);
  __pyx_codeobj__2377 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2376, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 6142, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2377)) __PYX_ERR(0, 6142, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6146
 *                 listener.exitExpr_mul(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpr_mul" ):
 *                 return visitor.visitExpr_mul(self)
 */
  __pyx_tuple__2378 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2378)) __PYX_ERR(0, 6146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2378);
  __Pyx_GIVEREF(__pyx_tuple__2378);
  __pyx_codeobj__2379 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2378, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 6146, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2379)) __PYX_ERR(0, 6146, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6154
 * 
 * 
 *     def expr(self, _p:int=0):             # <<<<<<<<<<<<<<
 *         _parentctx = self._ctx
 *         _parentState = self.state
 */
  __pyx_tuple__2380 = PyTuple_Pack(12, __pyx_n_s_self, __pyx_n_s_p, __pyx_n_s_parentctx, __pyx_n_s_parentState, __pyx_n_s_localctx, __pyx_n_s_prevctx, __pyx_n_s_startState, __pyx_n_s_la_2, __pyx_n_s_la, __pyx_n_s_alt, __pyx_n_s_FailedPredicateException, __pyx_n_s_re); if (unlikely(!__pyx_tuple__2380)) __PYX_ERR(0, 6154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2380);
  __Pyx_GIVEREF(__pyx_tuple__2380);
  __pyx_codeobj__2381 = (PyObject*)__Pyx_PyCode_New(2, 0, 12, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2380, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_expr, 6154, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2381)) __PYX_ERR(0, 6154, __pyx_L1_error)
  __pyx_tuple__2382 = PyTuple_Pack(1, ((PyObject *)__pyx_int_0)); if (unlikely(!__pyx_tuple__2382)) __PYX_ERR(0, 6154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2382);
  __Pyx_GIVEREF(__pyx_tuple__2382);

  /* "pymola/generated/ModelicaParser.pyx":6325
 *     class PrimaryContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__2383 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__2383)) __PYX_ERR(0, 6325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2383);
  __Pyx_GIVEREF(__pyx_tuple__2383);
  __pyx_codeobj__2384 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2383, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 6325, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2384)) __PYX_ERR(0, 6325, __pyx_L1_error)
  __pyx_tuple__2385 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__2385)) __PYX_ERR(0, 6325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2385);
  __Pyx_GIVEREF(__pyx_tuple__2385);

  /* "pymola/generated/ModelicaParser.pyx":6330
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_primary
 * 
 */
  __pyx_tuple__2386 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2386)) __PYX_ERR(0, 6330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2386);
  __Pyx_GIVEREF(__pyx_tuple__2386);
  __pyx_codeobj__2387 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2386, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 6330, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2387)) __PYX_ERR(0, 6330, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6334
 * 
 * 
 *         def copyFrom(self, ctx:ParserRuleContext):             # <<<<<<<<<<<<<<
 *             super().copyFrom(ctx)
 * 
 */
  __pyx_tuple__2388 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__2388)) __PYX_ERR(0, 6334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2388);
  __Pyx_GIVEREF(__pyx_tuple__2388);
  __pyx_codeobj__2389 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2388, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_copyFrom, 6334, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2389)) __PYX_ERR(0, 6334, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6341
 *     class Primary_stringContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__2390 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__2390)) __PYX_ERR(0, 6341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2390);
  __Pyx_GIVEREF(__pyx_tuple__2390);
  __pyx_codeobj__2391 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2390, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 6341, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2391)) __PYX_ERR(0, 6341, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6345
 *             self.copyFrom(ctx)
 * 
 *         def STRING(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.STRING, 0)
 * 
 */
  __pyx_tuple__2392 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2392)) __PYX_ERR(0, 6345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2392);
  __Pyx_GIVEREF(__pyx_tuple__2392);
  __pyx_codeobj__2393 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2392, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_STRING, 6345, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2393)) __PYX_ERR(0, 6345, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6348
 *             return self.getToken(ModelicaParser.STRING, 0)
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_string" ):
 *                 listener.enterPrimary_string(self)
 */
  __pyx_tuple__2394 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2394)) __PYX_ERR(0, 6348, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2394);
  __Pyx_GIVEREF(__pyx_tuple__2394);
  __pyx_codeobj__2395 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2394, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 6348, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2395)) __PYX_ERR(0, 6348, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6352
 *                 listener.enterPrimary_string(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_string" ):
 *                 listener.exitPrimary_string(self)
 */
  __pyx_tuple__2396 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2396)) __PYX_ERR(0, 6352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2396);
  __Pyx_GIVEREF(__pyx_tuple__2396);
  __pyx_codeobj__2397 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2396, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 6352, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2397)) __PYX_ERR(0, 6352, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6356
 *                 listener.exitPrimary_string(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_string" ):
 *                 return visitor.visitPrimary_string(self)
 */
  __pyx_tuple__2398 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2398)) __PYX_ERR(0, 6356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2398);
  __Pyx_GIVEREF(__pyx_tuple__2398);
  __pyx_codeobj__2399 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2398, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 6356, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2399)) __PYX_ERR(0, 6356, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6365
 *     class Primary_endContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__2400 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__2400)) __PYX_ERR(0, 6365, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2400);
  __Pyx_GIVEREF(__pyx_tuple__2400);
  __pyx_codeobj__2401 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2400, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 6365, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2401)) __PYX_ERR(0, 6365, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6370
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_end" ):
 *                 listener.enterPrimary_end(self)
 */
  __pyx_tuple__2402 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2402)) __PYX_ERR(0, 6370, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2402);
  __Pyx_GIVEREF(__pyx_tuple__2402);
  __pyx_codeobj__2403 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2402, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 6370, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2403)) __PYX_ERR(0, 6370, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6374
 *                 listener.enterPrimary_end(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_end" ):
 *                 listener.exitPrimary_end(self)
 */
  __pyx_tuple__2404 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2404)) __PYX_ERR(0, 6374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2404);
  __Pyx_GIVEREF(__pyx_tuple__2404);
  __pyx_codeobj__2405 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2404, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 6374, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2405)) __PYX_ERR(0, 6374, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6378
 *                 listener.exitPrimary_end(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_end" ):
 *                 return visitor.visitPrimary_end(self)
 */
  __pyx_tuple__2406 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2406)) __PYX_ERR(0, 6378, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2406);
  __Pyx_GIVEREF(__pyx_tuple__2406);
  __pyx_codeobj__2407 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2406, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 6378, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2407)) __PYX_ERR(0, 6378, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6387
 *     class Primary_output_expression_listContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__2408 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__2408)) __PYX_ERR(0, 6387, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2408);
  __Pyx_GIVEREF(__pyx_tuple__2408);
  __pyx_codeobj__2409 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2408, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 6387, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2409)) __PYX_ERR(0, 6387, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6391
 *             self.copyFrom(ctx)
 * 
 *         def output_expression_list(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Output_expression_listContext,0)
 * 
 */
  __pyx_tuple__2410 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2410)) __PYX_ERR(0, 6391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2410);
  __Pyx_GIVEREF(__pyx_tuple__2410);
  __pyx_codeobj__2411 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2410, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_output_expression_list, 6391, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2411)) __PYX_ERR(0, 6391, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6395
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_output_expression_list" ):
 *                 listener.enterPrimary_output_expression_list(self)
 */
  __pyx_tuple__2412 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2412)) __PYX_ERR(0, 6395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2412);
  __Pyx_GIVEREF(__pyx_tuple__2412);
  __pyx_codeobj__2413 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2412, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 6395, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2413)) __PYX_ERR(0, 6395, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6399
 *                 listener.enterPrimary_output_expression_list(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_output_expression_list" ):
 *                 listener.exitPrimary_output_expression_list(self)
 */
  __pyx_tuple__2414 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2414)) __PYX_ERR(0, 6399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2414);
  __Pyx_GIVEREF(__pyx_tuple__2414);
  __pyx_codeobj__2415 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2414, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 6399, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2415)) __PYX_ERR(0, 6399, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6403
 *                 listener.exitPrimary_output_expression_list(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_output_expression_list" ):
 *                 return visitor.visitPrimary_output_expression_list(self)
 */
  __pyx_tuple__2416 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2416)) __PYX_ERR(0, 6403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2416);
  __Pyx_GIVEREF(__pyx_tuple__2416);
  __pyx_codeobj__2417 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2416, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 6403, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2417)) __PYX_ERR(0, 6403, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6412
 *     class Primary_unsigned_numberContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__2418 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__2418)) __PYX_ERR(0, 6412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2418);
  __Pyx_GIVEREF(__pyx_tuple__2418);
  __pyx_codeobj__2419 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2418, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 6412, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2419)) __PYX_ERR(0, 6412, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6416
 *             self.copyFrom(ctx)
 * 
 *         def UNSIGNED_NUMBER(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.UNSIGNED_NUMBER, 0)
 * 
 */
  __pyx_tuple__2420 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2420)) __PYX_ERR(0, 6416, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2420);
  __Pyx_GIVEREF(__pyx_tuple__2420);
  __pyx_codeobj__2421 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2420, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_UNSIGNED_NUMBER, 6416, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2421)) __PYX_ERR(0, 6416, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6419
 *             return self.getToken(ModelicaParser.UNSIGNED_NUMBER, 0)
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_unsigned_number" ):
 *                 listener.enterPrimary_unsigned_number(self)
 */
  __pyx_tuple__2422 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2422)) __PYX_ERR(0, 6419, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2422);
  __Pyx_GIVEREF(__pyx_tuple__2422);
  __pyx_codeobj__2423 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2422, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 6419, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2423)) __PYX_ERR(0, 6419, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6423
 *                 listener.enterPrimary_unsigned_number(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_unsigned_number" ):
 *                 listener.exitPrimary_unsigned_number(self)
 */
  __pyx_tuple__2424 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2424)) __PYX_ERR(0, 6423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2424);
  __Pyx_GIVEREF(__pyx_tuple__2424);
  __pyx_codeobj__2425 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2424, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 6423, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2425)) __PYX_ERR(0, 6423, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6427
 *                 listener.exitPrimary_unsigned_number(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_unsigned_number" ):
 *                 return visitor.visitPrimary_unsigned_number(self)
 */
  __pyx_tuple__2426 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2426)) __PYX_ERR(0, 6427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2426);
  __Pyx_GIVEREF(__pyx_tuple__2426);
  __pyx_codeobj__2427 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2426, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 6427, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2427)) __PYX_ERR(0, 6427, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6436
 *     class Primary_function_argumentsContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__2428 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__2428)) __PYX_ERR(0, 6436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2428);
  __Pyx_GIVEREF(__pyx_tuple__2428);
  __pyx_codeobj__2429 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2428, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 6436, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2429)) __PYX_ERR(0, 6436, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6440
 *             self.copyFrom(ctx)
 * 
 *         def function_arguments(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Function_argumentsContext,0)
 * 
 */
  __pyx_tuple__2430 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2430)) __PYX_ERR(0, 6440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2430);
  __Pyx_GIVEREF(__pyx_tuple__2430);
  __pyx_codeobj__2431 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2430, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_function_arguments, 6440, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2431)) __PYX_ERR(0, 6440, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6444
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_function_arguments" ):
 *                 listener.enterPrimary_function_arguments(self)
 */
  __pyx_tuple__2432 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2432)) __PYX_ERR(0, 6444, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2432);
  __Pyx_GIVEREF(__pyx_tuple__2432);
  __pyx_codeobj__2433 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2432, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 6444, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2433)) __PYX_ERR(0, 6444, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6448
 *                 listener.enterPrimary_function_arguments(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_function_arguments" ):
 *                 listener.exitPrimary_function_arguments(self)
 */
  __pyx_tuple__2434 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2434)) __PYX_ERR(0, 6448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2434);
  __Pyx_GIVEREF(__pyx_tuple__2434);
  __pyx_codeobj__2435 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2434, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 6448, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2435)) __PYX_ERR(0, 6448, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6452
 *                 listener.exitPrimary_function_arguments(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_function_arguments" ):
 *                 return visitor.visitPrimary_function_arguments(self)
 */
  __pyx_tuple__2436 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2436)) __PYX_ERR(0, 6452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2436);
  __Pyx_GIVEREF(__pyx_tuple__2436);
  __pyx_codeobj__2437 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2436, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 6452, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2437)) __PYX_ERR(0, 6452, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6461
 *     class Primary_falseContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__2438 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__2438)) __PYX_ERR(0, 6461, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2438);
  __Pyx_GIVEREF(__pyx_tuple__2438);
  __pyx_codeobj__2439 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2438, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 6461, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2439)) __PYX_ERR(0, 6461, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6466
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_false" ):
 *                 listener.enterPrimary_false(self)
 */
  __pyx_tuple__2440 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2440)) __PYX_ERR(0, 6466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2440);
  __Pyx_GIVEREF(__pyx_tuple__2440);
  __pyx_codeobj__2441 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2440, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 6466, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2441)) __PYX_ERR(0, 6466, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6470
 *                 listener.enterPrimary_false(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_false" ):
 *                 listener.exitPrimary_false(self)
 */
  __pyx_tuple__2442 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2442)) __PYX_ERR(0, 6470, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2442);
  __Pyx_GIVEREF(__pyx_tuple__2442);
  __pyx_codeobj__2443 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2442, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 6470, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2443)) __PYX_ERR(0, 6470, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6474
 *                 listener.exitPrimary_false(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_false" ):
 *                 return visitor.visitPrimary_false(self)
 */
  __pyx_tuple__2444 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2444)) __PYX_ERR(0, 6474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2444);
  __Pyx_GIVEREF(__pyx_tuple__2444);
  __pyx_codeobj__2445 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2444, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 6474, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2445)) __PYX_ERR(0, 6474, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6483
 *     class Primary_derivativeContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__2446 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__2446)) __PYX_ERR(0, 6483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2446);
  __Pyx_GIVEREF(__pyx_tuple__2446);
  __pyx_codeobj__2447 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2446, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 6483, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2447)) __PYX_ERR(0, 6483, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6487
 *             self.copyFrom(ctx)
 * 
 *         def function_call_args(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Function_call_argsContext,0)
 * 
 */
  __pyx_tuple__2448 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2448)) __PYX_ERR(0, 6487, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2448);
  __Pyx_GIVEREF(__pyx_tuple__2448);
  __pyx_codeobj__2449 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2448, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_function_call_args, 6487, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2449)) __PYX_ERR(0, 6487, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6491
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_derivative" ):
 *                 listener.enterPrimary_derivative(self)
 */
  __pyx_tuple__2450 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2450)) __PYX_ERR(0, 6491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2450);
  __Pyx_GIVEREF(__pyx_tuple__2450);
  __pyx_codeobj__2451 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2450, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 6491, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2451)) __PYX_ERR(0, 6491, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6495
 *                 listener.enterPrimary_derivative(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_derivative" ):
 *                 listener.exitPrimary_derivative(self)
 */
  __pyx_tuple__2452 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2452)) __PYX_ERR(0, 6495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2452);
  __Pyx_GIVEREF(__pyx_tuple__2452);
  __pyx_codeobj__2453 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2452, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 6495, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2453)) __PYX_ERR(0, 6495, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6499
 *                 listener.exitPrimary_derivative(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_derivative" ):
 *                 return visitor.visitPrimary_derivative(self)
 */
  __pyx_tuple__2454 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2454)) __PYX_ERR(0, 6499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2454);
  __Pyx_GIVEREF(__pyx_tuple__2454);
  __pyx_codeobj__2455 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2454, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 6499, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2455)) __PYX_ERR(0, 6499, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6508
 *     class Primary_component_referenceContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__2456 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__2456)) __PYX_ERR(0, 6508, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2456);
  __Pyx_GIVEREF(__pyx_tuple__2456);
  __pyx_codeobj__2457 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2456, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 6508, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2457)) __PYX_ERR(0, 6508, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6512
 *             self.copyFrom(ctx)
 * 
 *         def component_reference(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 */
  __pyx_tuple__2458 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2458)) __PYX_ERR(0, 6512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2458);
  __Pyx_GIVEREF(__pyx_tuple__2458);
  __pyx_codeobj__2459 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2458, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_component_reference, 6512, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2459)) __PYX_ERR(0, 6512, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6516
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_component_reference" ):
 *                 listener.enterPrimary_component_reference(self)
 */
  __pyx_tuple__2460 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2460)) __PYX_ERR(0, 6516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2460);
  __Pyx_GIVEREF(__pyx_tuple__2460);
  __pyx_codeobj__2461 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2460, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 6516, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2461)) __PYX_ERR(0, 6516, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6520
 *                 listener.enterPrimary_component_reference(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_component_reference" ):
 *                 listener.exitPrimary_component_reference(self)
 */
  __pyx_tuple__2462 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2462)) __PYX_ERR(0, 6520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2462);
  __Pyx_GIVEREF(__pyx_tuple__2462);
  __pyx_codeobj__2463 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2462, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 6520, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2463)) __PYX_ERR(0, 6520, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6524
 *                 listener.exitPrimary_component_reference(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_component_reference" ):
 *                 return visitor.visitPrimary_component_reference(self)
 */
  __pyx_tuple__2464 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2464)) __PYX_ERR(0, 6524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2464);
  __Pyx_GIVEREF(__pyx_tuple__2464);
  __pyx_codeobj__2465 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2464, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 6524, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2465)) __PYX_ERR(0, 6524, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6533
 *     class Primary_expression_listContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__2466 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__2466)) __PYX_ERR(0, 6533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2466);
  __Pyx_GIVEREF(__pyx_tuple__2466);
  __pyx_codeobj__2467 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2466, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 6533, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2467)) __PYX_ERR(0, 6533, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6537
 *             self.copyFrom(ctx)
 * 
 *         def expression_list(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Expression_listContext)
 */
  __pyx_tuple__2468 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__2468)) __PYX_ERR(0, 6537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2468);
  __Pyx_GIVEREF(__pyx_tuple__2468);
  __pyx_codeobj__2469 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2468, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_expression_list, 6537, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2469)) __PYX_ERR(0, 6537, __pyx_L1_error)
  __pyx_tuple__2470 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__2470)) __PYX_ERR(0, 6537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2470);
  __Pyx_GIVEREF(__pyx_tuple__2470);

  /* "pymola/generated/ModelicaParser.pyx":6544
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_expression_list" ):
 *                 listener.enterPrimary_expression_list(self)
 */
  __pyx_tuple__2471 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2471)) __PYX_ERR(0, 6544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2471);
  __Pyx_GIVEREF(__pyx_tuple__2471);
  __pyx_codeobj__2472 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2471, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 6544, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2472)) __PYX_ERR(0, 6544, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6548
 *                 listener.enterPrimary_expression_list(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_expression_list" ):
 *                 listener.exitPrimary_expression_list(self)
 */
  __pyx_tuple__2473 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2473)) __PYX_ERR(0, 6548, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2473);
  __Pyx_GIVEREF(__pyx_tuple__2473);
  __pyx_codeobj__2474 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2473, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 6548, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2474)) __PYX_ERR(0, 6548, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6552
 *                 listener.exitPrimary_expression_list(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_expression_list" ):
 *                 return visitor.visitPrimary_expression_list(self)
 */
  __pyx_tuple__2475 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2475)) __PYX_ERR(0, 6552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2475);
  __Pyx_GIVEREF(__pyx_tuple__2475);
  __pyx_codeobj__2476 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2475, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 6552, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2476)) __PYX_ERR(0, 6552, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6561
 *     class Primary_trueContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__2477 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__2477)) __PYX_ERR(0, 6561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2477);
  __Pyx_GIVEREF(__pyx_tuple__2477);
  __pyx_codeobj__2478 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2477, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 6561, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2478)) __PYX_ERR(0, 6561, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6566
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_true" ):
 *                 listener.enterPrimary_true(self)
 */
  __pyx_tuple__2479 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2479)) __PYX_ERR(0, 6566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2479);
  __Pyx_GIVEREF(__pyx_tuple__2479);
  __pyx_codeobj__2480 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2479, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 6566, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2480)) __PYX_ERR(0, 6566, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6570
 *                 listener.enterPrimary_true(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_true" ):
 *                 listener.exitPrimary_true(self)
 */
  __pyx_tuple__2481 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2481)) __PYX_ERR(0, 6570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2481);
  __Pyx_GIVEREF(__pyx_tuple__2481);
  __pyx_codeobj__2482 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2481, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 6570, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2482)) __PYX_ERR(0, 6570, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6574
 *                 listener.exitPrimary_true(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_true" ):
 *                 return visitor.visitPrimary_true(self)
 */
  __pyx_tuple__2483 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2483)) __PYX_ERR(0, 6574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2483);
  __Pyx_GIVEREF(__pyx_tuple__2483);
  __pyx_codeobj__2484 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2483, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 6574, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2484)) __PYX_ERR(0, 6574, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6583
 *     class Primary_functionContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__2485 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__2485)) __PYX_ERR(0, 6583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2485);
  __Pyx_GIVEREF(__pyx_tuple__2485);
  __pyx_codeobj__2486 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2485, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 6583, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2486)) __PYX_ERR(0, 6583, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6587
 *             self.copyFrom(ctx)
 * 
 *         def component_reference(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 */
  __pyx_tuple__2487 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2487)) __PYX_ERR(0, 6587, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2487);
  __Pyx_GIVEREF(__pyx_tuple__2487);
  __pyx_codeobj__2488 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2487, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_component_reference, 6587, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2488)) __PYX_ERR(0, 6587, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6590
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 *         def function_call_args(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Function_call_argsContext,0)
 * 
 */
  __pyx_tuple__2489 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2489)) __PYX_ERR(0, 6590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2489);
  __Pyx_GIVEREF(__pyx_tuple__2489);
  __pyx_codeobj__2490 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2489, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_function_call_args, 6590, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2490)) __PYX_ERR(0, 6590, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6594
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_function" ):
 *                 listener.enterPrimary_function(self)
 */
  __pyx_tuple__2491 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2491)) __PYX_ERR(0, 6594, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2491);
  __Pyx_GIVEREF(__pyx_tuple__2491);
  __pyx_codeobj__2492 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2491, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 6594, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2492)) __PYX_ERR(0, 6594, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6598
 *                 listener.enterPrimary_function(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_function" ):
 *                 listener.exitPrimary_function(self)
 */
  __pyx_tuple__2493 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2493)) __PYX_ERR(0, 6598, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2493);
  __Pyx_GIVEREF(__pyx_tuple__2493);
  __pyx_codeobj__2494 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2493, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 6598, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2494)) __PYX_ERR(0, 6598, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6602
 *                 listener.exitPrimary_function(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_function" ):
 *                 return visitor.visitPrimary_function(self)
 */
  __pyx_tuple__2495 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2495)) __PYX_ERR(0, 6602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2495);
  __Pyx_GIVEREF(__pyx_tuple__2495);
  __pyx_codeobj__2496 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2495, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 6602, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2496)) __PYX_ERR(0, 6602, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6611
 *     class Primary_initialContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__2497 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__2497)) __PYX_ERR(0, 6611, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2497);
  __Pyx_GIVEREF(__pyx_tuple__2497);
  __pyx_codeobj__2498 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2497, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 6611, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2498)) __PYX_ERR(0, 6611, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6615
 *             self.copyFrom(ctx)
 * 
 *         def function_call_args(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Function_call_argsContext,0)
 * 
 */
  __pyx_tuple__2499 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2499)) __PYX_ERR(0, 6615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2499);
  __Pyx_GIVEREF(__pyx_tuple__2499);
  __pyx_codeobj__2500 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2499, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_function_call_args, 6615, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2500)) __PYX_ERR(0, 6615, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6619
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_initial" ):
 *                 listener.enterPrimary_initial(self)
 */
  __pyx_tuple__2501 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2501)) __PYX_ERR(0, 6619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2501);
  __Pyx_GIVEREF(__pyx_tuple__2501);
  __pyx_codeobj__2502 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2501, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 6619, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2502)) __PYX_ERR(0, 6619, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6623
 *                 listener.enterPrimary_initial(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_initial" ):
 *                 listener.exitPrimary_initial(self)
 */
  __pyx_tuple__2503 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2503)) __PYX_ERR(0, 6623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2503);
  __Pyx_GIVEREF(__pyx_tuple__2503);
  __pyx_codeobj__2504 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2503, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 6623, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2504)) __PYX_ERR(0, 6623, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6627
 *                 listener.exitPrimary_initial(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_initial" ):
 *                 return visitor.visitPrimary_initial(self)
 */
  __pyx_tuple__2505 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2505)) __PYX_ERR(0, 6627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2505);
  __Pyx_GIVEREF(__pyx_tuple__2505);
  __pyx_codeobj__2506 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2505, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 6627, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2506)) __PYX_ERR(0, 6627, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6635
 * 
 * 
 *     def primary(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.PrimaryContext(self, self._ctx, self.state)
 */
  __pyx_tuple__2507 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la_2, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__2507)) __PYX_ERR(0, 6635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2507);
  __Pyx_GIVEREF(__pyx_tuple__2507);
  __pyx_codeobj__2508 = (PyObject*)__Pyx_PyCode_New(1, 0, 5, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2507, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_primary, 6635, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2508)) __PYX_ERR(0, 6635, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6769
 *     class NameContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__2509 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__2509)) __PYX_ERR(0, 6769, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2509);
  __Pyx_GIVEREF(__pyx_tuple__2509);
  __pyx_codeobj__2510 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2509, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 6769, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2510)) __PYX_ERR(0, 6769, __pyx_L1_error)
  __pyx_tuple__2511 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__2511)) __PYX_ERR(0, 6769, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2511);
  __Pyx_GIVEREF(__pyx_tuple__2511);

  /* "pymola/generated/ModelicaParser.pyx":6773
 *             self.parser = parser
 * 
 *         def IDENT(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTokens(ModelicaParser.IDENT)
 */
  __pyx_tuple__2512 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__2512)) __PYX_ERR(0, 6773, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2512);
  __Pyx_GIVEREF(__pyx_tuple__2512);
  __pyx_codeobj__2513 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2512, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_IDENT, 6773, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2513)) __PYX_ERR(0, 6773, __pyx_L1_error)
  __pyx_tuple__2514 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__2514)) __PYX_ERR(0, 6773, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2514);
  __Pyx_GIVEREF(__pyx_tuple__2514);

  /* "pymola/generated/ModelicaParser.pyx":6779
 *                 return self.getToken(ModelicaParser.IDENT, i)
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_name
 * 
 */
  __pyx_tuple__2515 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2515)) __PYX_ERR(0, 6779, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2515);
  __Pyx_GIVEREF(__pyx_tuple__2515);
  __pyx_codeobj__2516 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2515, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 6779, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2516)) __PYX_ERR(0, 6779, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6782
 *             return ModelicaParser.RULE_name
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterName" ):
 *                 listener.enterName(self)
 */
  __pyx_tuple__2517 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2517)) __PYX_ERR(0, 6782, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2517);
  __Pyx_GIVEREF(__pyx_tuple__2517);
  __pyx_codeobj__2518 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2517, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 6782, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2518)) __PYX_ERR(0, 6782, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6786
 *                 listener.enterName(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitName" ):
 *                 listener.exitName(self)
 */
  __pyx_tuple__2519 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2519)) __PYX_ERR(0, 6786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2519);
  __Pyx_GIVEREF(__pyx_tuple__2519);
  __pyx_codeobj__2520 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2519, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 6786, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2520)) __PYX_ERR(0, 6786, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6790
 *                 listener.exitName(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitName" ):
 *                 return visitor.visitName(self)
 */
  __pyx_tuple__2521 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2521)) __PYX_ERR(0, 6790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2521);
  __Pyx_GIVEREF(__pyx_tuple__2521);
  __pyx_codeobj__2522 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2521, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 6790, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2522)) __PYX_ERR(0, 6790, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6799
 * 
 * 
 *     def name(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.NameContext(self, self._ctx, self.state)
 */
  __pyx_tuple__2523 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__2523)) __PYX_ERR(0, 6799, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2523);
  __Pyx_GIVEREF(__pyx_tuple__2523);
  __pyx_codeobj__2524 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2523, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_name, 6799, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2524)) __PYX_ERR(0, 6799, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6838
 *     class Component_reference_elementContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__2525 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__2525)) __PYX_ERR(0, 6838, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2525);
  __Pyx_GIVEREF(__pyx_tuple__2525);
  __pyx_codeobj__2526 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2525, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 6838, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2526)) __PYX_ERR(0, 6838, __pyx_L1_error)
  __pyx_tuple__2527 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__2527)) __PYX_ERR(0, 6838, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2527);
  __Pyx_GIVEREF(__pyx_tuple__2527);

  /* "pymola/generated/ModelicaParser.pyx":6842
 *             self.parser = parser
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 */
  __pyx_tuple__2528 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2528)) __PYX_ERR(0, 6842, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2528);
  __Pyx_GIVEREF(__pyx_tuple__2528);
  __pyx_codeobj__2529 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2528, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_IDENT, 6842, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2529)) __PYX_ERR(0, 6842, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6845
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 *         def array_subscripts(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Array_subscriptsContext,0)
 * 
 */
  __pyx_tuple__2530 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2530)) __PYX_ERR(0, 6845, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2530);
  __Pyx_GIVEREF(__pyx_tuple__2530);
  __pyx_codeobj__2531 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2530, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_array_subscripts, 6845, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2531)) __PYX_ERR(0, 6845, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6849
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_component_reference_element
 * 
 */
  __pyx_tuple__2532 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2532)) __PYX_ERR(0, 6849, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2532);
  __Pyx_GIVEREF(__pyx_tuple__2532);
  __pyx_codeobj__2533 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2532, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 6849, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2533)) __PYX_ERR(0, 6849, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6852
 *             return ModelicaParser.RULE_component_reference_element
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterComponent_reference_element" ):
 *                 listener.enterComponent_reference_element(self)
 */
  __pyx_tuple__2534 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2534)) __PYX_ERR(0, 6852, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2534);
  __Pyx_GIVEREF(__pyx_tuple__2534);
  __pyx_codeobj__2535 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2534, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 6852, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2535)) __PYX_ERR(0, 6852, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6856
 *                 listener.enterComponent_reference_element(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitComponent_reference_element" ):
 *                 listener.exitComponent_reference_element(self)
 */
  __pyx_tuple__2536 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2536)) __PYX_ERR(0, 6856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2536);
  __Pyx_GIVEREF(__pyx_tuple__2536);
  __pyx_codeobj__2537 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2536, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 6856, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2537)) __PYX_ERR(0, 6856, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6860
 *                 listener.exitComponent_reference_element(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitComponent_reference_element" ):
 *                 return visitor.visitComponent_reference_element(self)
 */
  __pyx_tuple__2538 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2538)) __PYX_ERR(0, 6860, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2538);
  __Pyx_GIVEREF(__pyx_tuple__2538);
  __pyx_codeobj__2539 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2538, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 6860, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2539)) __PYX_ERR(0, 6860, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6869
 * 
 * 
 *     def component_reference_element(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Component_reference_elementContext(self, self._ctx, self.state)
 */
  __pyx_tuple__2540 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la_2, __pyx_n_s_re); if (unlikely(!__pyx_tuple__2540)) __PYX_ERR(0, 6869, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2540);
  __Pyx_GIVEREF(__pyx_tuple__2540);
  __pyx_codeobj__2541 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2540, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_component_reference_element, 6869, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2541)) __PYX_ERR(0, 6869, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6895
 *     class Component_referenceContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__2542 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__2542)) __PYX_ERR(0, 6895, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2542);
  __Pyx_GIVEREF(__pyx_tuple__2542);
  __pyx_codeobj__2543 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2542, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 6895, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2543)) __PYX_ERR(0, 6895, __pyx_L1_error)
  __pyx_tuple__2544 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__2544)) __PYX_ERR(0, 6895, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2544);
  __Pyx_GIVEREF(__pyx_tuple__2544);

  /* "pymola/generated/ModelicaParser.pyx":6899
 *             self.parser = parser
 * 
 *         def component_reference_element(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Component_reference_elementContext)
 */
  __pyx_tuple__2545 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__2545)) __PYX_ERR(0, 6899, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2545);
  __Pyx_GIVEREF(__pyx_tuple__2545);
  __pyx_codeobj__2546 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2545, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_component_reference_element, 6899, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2546)) __PYX_ERR(0, 6899, __pyx_L1_error)
  __pyx_tuple__2547 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__2547)) __PYX_ERR(0, 6899, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2547);
  __Pyx_GIVEREF(__pyx_tuple__2547);

  /* "pymola/generated/ModelicaParser.pyx":6906
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_component_reference
 * 
 */
  __pyx_tuple__2548 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2548)) __PYX_ERR(0, 6906, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2548);
  __Pyx_GIVEREF(__pyx_tuple__2548);
  __pyx_codeobj__2549 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2548, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 6906, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2549)) __PYX_ERR(0, 6906, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6909
 *             return ModelicaParser.RULE_component_reference
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterComponent_reference" ):
 *                 listener.enterComponent_reference(self)
 */
  __pyx_tuple__2550 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2550)) __PYX_ERR(0, 6909, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2550);
  __Pyx_GIVEREF(__pyx_tuple__2550);
  __pyx_codeobj__2551 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2550, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 6909, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2551)) __PYX_ERR(0, 6909, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6913
 *                 listener.enterComponent_reference(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitComponent_reference" ):
 *                 listener.exitComponent_reference(self)
 */
  __pyx_tuple__2552 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2552)) __PYX_ERR(0, 6913, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2552);
  __Pyx_GIVEREF(__pyx_tuple__2552);
  __pyx_codeobj__2553 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2552, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 6913, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2553)) __PYX_ERR(0, 6913, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6917
 *                 listener.exitComponent_reference(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitComponent_reference" ):
 *                 return visitor.visitComponent_reference(self)
 */
  __pyx_tuple__2554 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2554)) __PYX_ERR(0, 6917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2554);
  __Pyx_GIVEREF(__pyx_tuple__2554);
  __pyx_codeobj__2555 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2554, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 6917, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2555)) __PYX_ERR(0, 6917, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6926
 * 
 * 
 *     def component_reference(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Component_referenceContext(self, self._ctx, self.state)
 */
  __pyx_tuple__2556 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_alt, __pyx_n_s_re); if (unlikely(!__pyx_tuple__2556)) __PYX_ERR(0, 6926, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2556);
  __Pyx_GIVEREF(__pyx_tuple__2556);
  __pyx_codeobj__2557 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2556, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_component_reference, 6926, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2557)) __PYX_ERR(0, 6926, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6957
 *     class Function_call_argsContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__2558 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__2558)) __PYX_ERR(0, 6957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2558);
  __Pyx_GIVEREF(__pyx_tuple__2558);
  __pyx_codeobj__2559 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2558, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 6957, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2559)) __PYX_ERR(0, 6957, __pyx_L1_error)
  __pyx_tuple__2560 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__2560)) __PYX_ERR(0, 6957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2560);
  __Pyx_GIVEREF(__pyx_tuple__2560);

  /* "pymola/generated/ModelicaParser.pyx":6961
 *             self.parser = parser
 * 
 *         def function_arguments(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Function_argumentsContext,0)
 * 
 */
  __pyx_tuple__2561 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2561)) __PYX_ERR(0, 6961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2561);
  __Pyx_GIVEREF(__pyx_tuple__2561);
  __pyx_codeobj__2562 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2561, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_function_arguments, 6961, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2562)) __PYX_ERR(0, 6961, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6965
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_function_call_args
 * 
 */
  __pyx_tuple__2563 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2563)) __PYX_ERR(0, 6965, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2563);
  __Pyx_GIVEREF(__pyx_tuple__2563);
  __pyx_codeobj__2564 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2563, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 6965, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2564)) __PYX_ERR(0, 6965, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6968
 *             return ModelicaParser.RULE_function_call_args
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterFunction_call_args" ):
 *                 listener.enterFunction_call_args(self)
 */
  __pyx_tuple__2565 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2565)) __PYX_ERR(0, 6968, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2565);
  __Pyx_GIVEREF(__pyx_tuple__2565);
  __pyx_codeobj__2566 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2565, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 6968, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2566)) __PYX_ERR(0, 6968, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6972
 *                 listener.enterFunction_call_args(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitFunction_call_args" ):
 *                 listener.exitFunction_call_args(self)
 */
  __pyx_tuple__2567 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2567)) __PYX_ERR(0, 6972, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2567);
  __Pyx_GIVEREF(__pyx_tuple__2567);
  __pyx_codeobj__2568 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2567, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 6972, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2568)) __PYX_ERR(0, 6972, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6976
 *                 listener.exitFunction_call_args(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitFunction_call_args" ):
 *                 return visitor.visitFunction_call_args(self)
 */
  __pyx_tuple__2569 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2569)) __PYX_ERR(0, 6976, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2569);
  __Pyx_GIVEREF(__pyx_tuple__2569);
  __pyx_codeobj__2570 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2569, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 6976, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2570)) __PYX_ERR(0, 6976, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":6985
 * 
 * 
 *     def function_call_args(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Function_call_argsContext(self, self._ctx, self.state)
 */
  __pyx_tuple__2571 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__2571)) __PYX_ERR(0, 6985, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2571);
  __Pyx_GIVEREF(__pyx_tuple__2571);
  __pyx_codeobj__2572 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2571, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_function_call_args, 6985, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2572)) __PYX_ERR(0, 6985, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7014
 *     class Function_argumentsContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__2573 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__2573)) __PYX_ERR(0, 7014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2573);
  __Pyx_GIVEREF(__pyx_tuple__2573);
  __pyx_codeobj__2574 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2573, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 7014, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2574)) __PYX_ERR(0, 7014, __pyx_L1_error)
  __pyx_tuple__2575 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__2575)) __PYX_ERR(0, 7014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2575);
  __Pyx_GIVEREF(__pyx_tuple__2575);

  /* "pymola/generated/ModelicaParser.pyx":7018
 *             self.parser = parser
 * 
 *         def function_argument(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Function_argumentContext)
 */
  __pyx_tuple__2576 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__2576)) __PYX_ERR(0, 7018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2576);
  __Pyx_GIVEREF(__pyx_tuple__2576);
  __pyx_codeobj__2577 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2576, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_function_argument, 7018, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2577)) __PYX_ERR(0, 7018, __pyx_L1_error)
  __pyx_tuple__2578 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__2578)) __PYX_ERR(0, 7018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2578);
  __Pyx_GIVEREF(__pyx_tuple__2578);

  /* "pymola/generated/ModelicaParser.pyx":7025
 * 
 * 
 *         def for_indices(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.For_indicesContext)
 */
  __pyx_tuple__2579 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__2579)) __PYX_ERR(0, 7025, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2579);
  __Pyx_GIVEREF(__pyx_tuple__2579);
  __pyx_codeobj__2580 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2579, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_for_indices, 7025, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2580)) __PYX_ERR(0, 7025, __pyx_L1_error)
  __pyx_tuple__2581 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__2581)) __PYX_ERR(0, 7025, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2581);
  __Pyx_GIVEREF(__pyx_tuple__2581);

  /* "pymola/generated/ModelicaParser.pyx":7032
 * 
 * 
 *         def named_arguments(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Named_argumentsContext,0)
 * 
 */
  __pyx_tuple__2582 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2582)) __PYX_ERR(0, 7032, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2582);
  __Pyx_GIVEREF(__pyx_tuple__2582);
  __pyx_codeobj__2583 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2582, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_named_arguments, 7032, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2583)) __PYX_ERR(0, 7032, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7036
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_function_arguments
 * 
 */
  __pyx_tuple__2584 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2584)) __PYX_ERR(0, 7036, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2584);
  __Pyx_GIVEREF(__pyx_tuple__2584);
  __pyx_codeobj__2585 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2584, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 7036, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2585)) __PYX_ERR(0, 7036, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7039
 *             return ModelicaParser.RULE_function_arguments
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterFunction_arguments" ):
 *                 listener.enterFunction_arguments(self)
 */
  __pyx_tuple__2586 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2586)) __PYX_ERR(0, 7039, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2586);
  __Pyx_GIVEREF(__pyx_tuple__2586);
  __pyx_codeobj__2587 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2586, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 7039, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2587)) __PYX_ERR(0, 7039, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7043
 *                 listener.enterFunction_arguments(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitFunction_arguments" ):
 *                 listener.exitFunction_arguments(self)
 */
  __pyx_tuple__2588 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2588)) __PYX_ERR(0, 7043, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2588);
  __Pyx_GIVEREF(__pyx_tuple__2588);
  __pyx_codeobj__2589 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2588, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 7043, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2589)) __PYX_ERR(0, 7043, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7047
 *                 listener.exitFunction_arguments(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitFunction_arguments" ):
 *                 return visitor.visitFunction_arguments(self)
 */
  __pyx_tuple__2590 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2590)) __PYX_ERR(0, 7047, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2590);
  __Pyx_GIVEREF(__pyx_tuple__2590);
  __pyx_codeobj__2591 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2590, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 7047, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2591)) __PYX_ERR(0, 7047, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7056
 * 
 * 
 *     def function_arguments(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Function_argumentsContext(self, self._ctx, self.state)
 */
  __pyx_tuple__2592 = PyTuple_Pack(6, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la_2, __pyx_n_s_la, __pyx_n_s_token, __pyx_n_s_re); if (unlikely(!__pyx_tuple__2592)) __PYX_ERR(0, 7056, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2592);
  __Pyx_GIVEREF(__pyx_tuple__2592);
  __pyx_codeobj__2593 = (PyObject*)__Pyx_PyCode_New(1, 0, 6, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2592, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_function_arguments, 7056, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2593)) __PYX_ERR(0, 7056, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7114
 *     class Named_argumentsContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__2594 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__2594)) __PYX_ERR(0, 7114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2594);
  __Pyx_GIVEREF(__pyx_tuple__2594);
  __pyx_codeobj__2595 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2594, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 7114, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2595)) __PYX_ERR(0, 7114, __pyx_L1_error)
  __pyx_tuple__2596 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__2596)) __PYX_ERR(0, 7114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2596);
  __Pyx_GIVEREF(__pyx_tuple__2596);

  /* "pymola/generated/ModelicaParser.pyx":7118
 *             self.parser = parser
 * 
 *         def named_argument(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Named_argumentContext)
 */
  __pyx_tuple__2597 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__2597)) __PYX_ERR(0, 7118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2597);
  __Pyx_GIVEREF(__pyx_tuple__2597);
  __pyx_codeobj__2598 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2597, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_named_argument, 7118, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2598)) __PYX_ERR(0, 7118, __pyx_L1_error)
  __pyx_tuple__2599 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__2599)) __PYX_ERR(0, 7118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2599);
  __Pyx_GIVEREF(__pyx_tuple__2599);

  /* "pymola/generated/ModelicaParser.pyx":7125
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_named_arguments
 * 
 */
  __pyx_tuple__2600 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2600)) __PYX_ERR(0, 7125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2600);
  __Pyx_GIVEREF(__pyx_tuple__2600);
  __pyx_codeobj__2601 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2600, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 7125, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2601)) __PYX_ERR(0, 7125, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7128
 *             return ModelicaParser.RULE_named_arguments
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterNamed_arguments" ):
 *                 listener.enterNamed_arguments(self)
 */
  __pyx_tuple__2602 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2602)) __PYX_ERR(0, 7128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2602);
  __Pyx_GIVEREF(__pyx_tuple__2602);
  __pyx_codeobj__2603 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2602, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 7128, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2603)) __PYX_ERR(0, 7128, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7132
 *                 listener.enterNamed_arguments(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitNamed_arguments" ):
 *                 listener.exitNamed_arguments(self)
 */
  __pyx_tuple__2604 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2604)) __PYX_ERR(0, 7132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2604);
  __Pyx_GIVEREF(__pyx_tuple__2604);
  __pyx_codeobj__2605 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2604, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 7132, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2605)) __PYX_ERR(0, 7132, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7136
 *                 listener.exitNamed_arguments(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitNamed_arguments" ):
 *                 return visitor.visitNamed_arguments(self)
 */
  __pyx_tuple__2606 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2606)) __PYX_ERR(0, 7136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2606);
  __Pyx_GIVEREF(__pyx_tuple__2606);
  __pyx_codeobj__2607 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2606, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 7136, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2607)) __PYX_ERR(0, 7136, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7145
 * 
 * 
 *     def named_arguments(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Named_argumentsContext(self, self._ctx, self.state)
 */
  __pyx_tuple__2608 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__2608)) __PYX_ERR(0, 7145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2608);
  __Pyx_GIVEREF(__pyx_tuple__2608);
  __pyx_codeobj__2609 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2608, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_named_arguments, 7145, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2609)) __PYX_ERR(0, 7145, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7176
 *     class Named_argumentContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__2610 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__2610)) __PYX_ERR(0, 7176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2610);
  __Pyx_GIVEREF(__pyx_tuple__2610);
  __pyx_codeobj__2611 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2610, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 7176, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2611)) __PYX_ERR(0, 7176, __pyx_L1_error)
  __pyx_tuple__2612 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__2612)) __PYX_ERR(0, 7176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2612);
  __Pyx_GIVEREF(__pyx_tuple__2612);

  /* "pymola/generated/ModelicaParser.pyx":7180
 *             self.parser = parser
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 */
  __pyx_tuple__2613 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2613)) __PYX_ERR(0, 7180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2613);
  __Pyx_GIVEREF(__pyx_tuple__2613);
  __pyx_codeobj__2614 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2613, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_IDENT, 7180, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2614)) __PYX_ERR(0, 7180, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7183
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 *         def function_argument(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Function_argumentContext,0)
 * 
 */
  __pyx_tuple__2615 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2615)) __PYX_ERR(0, 7183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2615);
  __Pyx_GIVEREF(__pyx_tuple__2615);
  __pyx_codeobj__2616 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2615, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_function_argument, 7183, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2616)) __PYX_ERR(0, 7183, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7187
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_named_argument
 * 
 */
  __pyx_tuple__2617 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2617)) __PYX_ERR(0, 7187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2617);
  __Pyx_GIVEREF(__pyx_tuple__2617);
  __pyx_codeobj__2618 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2617, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 7187, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2618)) __PYX_ERR(0, 7187, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7190
 *             return ModelicaParser.RULE_named_argument
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterNamed_argument" ):
 *                 listener.enterNamed_argument(self)
 */
  __pyx_tuple__2619 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2619)) __PYX_ERR(0, 7190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2619);
  __Pyx_GIVEREF(__pyx_tuple__2619);
  __pyx_codeobj__2620 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2619, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 7190, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2620)) __PYX_ERR(0, 7190, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7194
 *                 listener.enterNamed_argument(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitNamed_argument" ):
 *                 listener.exitNamed_argument(self)
 */
  __pyx_tuple__2621 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2621)) __PYX_ERR(0, 7194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2621);
  __Pyx_GIVEREF(__pyx_tuple__2621);
  __pyx_codeobj__2622 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2621, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 7194, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2622)) __PYX_ERR(0, 7194, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7198
 *                 listener.exitNamed_argument(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitNamed_argument" ):
 *                 return visitor.visitNamed_argument(self)
 */
  __pyx_tuple__2623 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2623)) __PYX_ERR(0, 7198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2623);
  __Pyx_GIVEREF(__pyx_tuple__2623);
  __pyx_codeobj__2624 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2623, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 7198, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2624)) __PYX_ERR(0, 7198, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7207
 * 
 * 
 *     def named_argument(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Named_argumentContext(self, self._ctx, self.state)
 */
  __pyx_tuple__2625 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_re); if (unlikely(!__pyx_tuple__2625)) __PYX_ERR(0, 7207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2625);
  __Pyx_GIVEREF(__pyx_tuple__2625);
  __pyx_codeobj__2626 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2625, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_named_argument, 7207, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2626)) __PYX_ERR(0, 7207, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7229
 *     class Function_argumentContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__2627 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__2627)) __PYX_ERR(0, 7229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2627);
  __Pyx_GIVEREF(__pyx_tuple__2627);
  __pyx_codeobj__2628 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2627, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 7229, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2628)) __PYX_ERR(0, 7229, __pyx_L1_error)
  __pyx_tuple__2629 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__2629)) __PYX_ERR(0, 7229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2629);
  __Pyx_GIVEREF(__pyx_tuple__2629);

  /* "pymola/generated/ModelicaParser.pyx":7234
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_function_argument
 * 
 */
  __pyx_tuple__2630 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2630)) __PYX_ERR(0, 7234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2630);
  __Pyx_GIVEREF(__pyx_tuple__2630);
  __pyx_codeobj__2631 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2630, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 7234, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2631)) __PYX_ERR(0, 7234, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7238
 * 
 * 
 *         def copyFrom(self, ctx:ParserRuleContext):             # <<<<<<<<<<<<<<
 *             super().copyFrom(ctx)
 * 
 */
  __pyx_tuple__2632 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__2632)) __PYX_ERR(0, 7238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2632);
  __Pyx_GIVEREF(__pyx_tuple__2632);
  __pyx_codeobj__2633 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2632, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_copyFrom, 7238, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2633)) __PYX_ERR(0, 7238, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7245
 *     class Argument_expressionContext(Function_argumentContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Function_argumentContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__2634 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__2634)) __PYX_ERR(0, 7245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2634);
  __Pyx_GIVEREF(__pyx_tuple__2634);
  __pyx_codeobj__2635 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2634, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 7245, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2635)) __PYX_ERR(0, 7245, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7249
 *             self.copyFrom(ctx)
 * 
 *         def expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 */
  __pyx_tuple__2636 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2636)) __PYX_ERR(0, 7249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2636);
  __Pyx_GIVEREF(__pyx_tuple__2636);
  __pyx_codeobj__2637 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2636, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_expression, 7249, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2637)) __PYX_ERR(0, 7249, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7253
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterArgument_expression" ):
 *                 listener.enterArgument_expression(self)
 */
  __pyx_tuple__2638 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2638)) __PYX_ERR(0, 7253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2638);
  __Pyx_GIVEREF(__pyx_tuple__2638);
  __pyx_codeobj__2639 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2638, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 7253, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2639)) __PYX_ERR(0, 7253, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7257
 *                 listener.enterArgument_expression(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitArgument_expression" ):
 *                 listener.exitArgument_expression(self)
 */
  __pyx_tuple__2640 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2640)) __PYX_ERR(0, 7257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2640);
  __Pyx_GIVEREF(__pyx_tuple__2640);
  __pyx_codeobj__2641 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2640, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 7257, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2641)) __PYX_ERR(0, 7257, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7261
 *                 listener.exitArgument_expression(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitArgument_expression" ):
 *                 return visitor.visitArgument_expression(self)
 */
  __pyx_tuple__2642 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2642)) __PYX_ERR(0, 7261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2642);
  __Pyx_GIVEREF(__pyx_tuple__2642);
  __pyx_codeobj__2643 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2642, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 7261, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2643)) __PYX_ERR(0, 7261, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7270
 *     class Argument_functionContext(Function_argumentContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Function_argumentContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_tuple__2644 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_ctx_2); if (unlikely(!__pyx_tuple__2644)) __PYX_ERR(0, 7270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2644);
  __Pyx_GIVEREF(__pyx_tuple__2644);
  __pyx_codeobj__2645 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2644, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 7270, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2645)) __PYX_ERR(0, 7270, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7274
 *             self.copyFrom(ctx)
 * 
 *         def name(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)
 * 
 */
  __pyx_tuple__2646 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2646)) __PYX_ERR(0, 7274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2646);
  __Pyx_GIVEREF(__pyx_tuple__2646);
  __pyx_codeobj__2647 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2646, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_name, 7274, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2647)) __PYX_ERR(0, 7274, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7277
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)
 * 
 *         def named_arguments(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Named_argumentsContext,0)
 * 
 */
  __pyx_tuple__2648 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2648)) __PYX_ERR(0, 7277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2648);
  __Pyx_GIVEREF(__pyx_tuple__2648);
  __pyx_codeobj__2649 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2648, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_named_arguments, 7277, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2649)) __PYX_ERR(0, 7277, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7281
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterArgument_function" ):
 *                 listener.enterArgument_function(self)
 */
  __pyx_tuple__2650 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2650)) __PYX_ERR(0, 7281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2650);
  __Pyx_GIVEREF(__pyx_tuple__2650);
  __pyx_codeobj__2651 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2650, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 7281, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2651)) __PYX_ERR(0, 7281, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7285
 *                 listener.enterArgument_function(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitArgument_function" ):
 *                 listener.exitArgument_function(self)
 */
  __pyx_tuple__2652 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2652)) __PYX_ERR(0, 7285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2652);
  __Pyx_GIVEREF(__pyx_tuple__2652);
  __pyx_codeobj__2653 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2652, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 7285, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2653)) __PYX_ERR(0, 7285, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7289
 *                 listener.exitArgument_function(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitArgument_function" ):
 *                 return visitor.visitArgument_function(self)
 */
  __pyx_tuple__2654 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2654)) __PYX_ERR(0, 7289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2654);
  __Pyx_GIVEREF(__pyx_tuple__2654);
  __pyx_codeobj__2655 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2654, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 7289, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2655)) __PYX_ERR(0, 7289, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7297
 * 
 * 
 *     def function_argument(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Function_argumentContext(self, self._ctx, self.state)
 */
  __pyx_tuple__2656 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_token, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__2656)) __PYX_ERR(0, 7297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2656);
  __Pyx_GIVEREF(__pyx_tuple__2656);
  __pyx_codeobj__2657 = (PyObject*)__Pyx_PyCode_New(1, 0, 5, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2656, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_function_argument, 7297, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2657)) __PYX_ERR(0, 7297, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7345
 *     class Output_expression_listContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__2658 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__2658)) __PYX_ERR(0, 7345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2658);
  __Pyx_GIVEREF(__pyx_tuple__2658);
  __pyx_codeobj__2659 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2658, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 7345, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2659)) __PYX_ERR(0, 7345, __pyx_L1_error)
  __pyx_tuple__2660 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__2660)) __PYX_ERR(0, 7345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2660);
  __Pyx_GIVEREF(__pyx_tuple__2660);

  /* "pymola/generated/ModelicaParser.pyx":7349
 *             self.parser = parser
 * 
 *         def expression(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 */
  __pyx_tuple__2661 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__2661)) __PYX_ERR(0, 7349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2661);
  __Pyx_GIVEREF(__pyx_tuple__2661);
  __pyx_codeobj__2662 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2661, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_expression, 7349, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2662)) __PYX_ERR(0, 7349, __pyx_L1_error)
  __pyx_tuple__2663 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__2663)) __PYX_ERR(0, 7349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2663);
  __Pyx_GIVEREF(__pyx_tuple__2663);

  /* "pymola/generated/ModelicaParser.pyx":7356
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_output_expression_list
 * 
 */
  __pyx_tuple__2664 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2664)) __PYX_ERR(0, 7356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2664);
  __Pyx_GIVEREF(__pyx_tuple__2664);
  __pyx_codeobj__2665 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2664, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 7356, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2665)) __PYX_ERR(0, 7356, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7359
 *             return ModelicaParser.RULE_output_expression_list
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterOutput_expression_list" ):
 *                 listener.enterOutput_expression_list(self)
 */
  __pyx_tuple__2666 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2666)) __PYX_ERR(0, 7359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2666);
  __Pyx_GIVEREF(__pyx_tuple__2666);
  __pyx_codeobj__2667 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2666, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 7359, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2667)) __PYX_ERR(0, 7359, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7363
 *                 listener.enterOutput_expression_list(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitOutput_expression_list" ):
 *                 listener.exitOutput_expression_list(self)
 */
  __pyx_tuple__2668 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2668)) __PYX_ERR(0, 7363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2668);
  __Pyx_GIVEREF(__pyx_tuple__2668);
  __pyx_codeobj__2669 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2668, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 7363, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2669)) __PYX_ERR(0, 7363, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7367
 *                 listener.exitOutput_expression_list(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitOutput_expression_list" ):
 *                 return visitor.visitOutput_expression_list(self)
 */
  __pyx_tuple__2670 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2670)) __PYX_ERR(0, 7367, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2670);
  __Pyx_GIVEREF(__pyx_tuple__2670);
  __pyx_codeobj__2671 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2670, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 7367, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2671)) __PYX_ERR(0, 7367, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7376
 * 
 * 
 *     def output_expression_list(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Output_expression_listContext(self, self._ctx, self.state)
 */
  __pyx_tuple__2672 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__2672)) __PYX_ERR(0, 7376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2672);
  __Pyx_GIVEREF(__pyx_tuple__2672);
  __pyx_codeobj__2673 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2672, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_output_expression_list, 7376, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2673)) __PYX_ERR(0, 7376, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7413
 *     class Expression_listContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__2674 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__2674)) __PYX_ERR(0, 7413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2674);
  __Pyx_GIVEREF(__pyx_tuple__2674);
  __pyx_codeobj__2675 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2674, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 7413, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2675)) __PYX_ERR(0, 7413, __pyx_L1_error)
  __pyx_tuple__2676 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__2676)) __PYX_ERR(0, 7413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2676);
  __Pyx_GIVEREF(__pyx_tuple__2676);

  /* "pymola/generated/ModelicaParser.pyx":7417
 *             self.parser = parser
 * 
 *         def expression(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 */
  __pyx_tuple__2677 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__2677)) __PYX_ERR(0, 7417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2677);
  __Pyx_GIVEREF(__pyx_tuple__2677);
  __pyx_codeobj__2678 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2677, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_expression, 7417, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2678)) __PYX_ERR(0, 7417, __pyx_L1_error)
  __pyx_tuple__2679 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__2679)) __PYX_ERR(0, 7417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2679);
  __Pyx_GIVEREF(__pyx_tuple__2679);

  /* "pymola/generated/ModelicaParser.pyx":7424
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_expression_list
 * 
 */
  __pyx_tuple__2680 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2680)) __PYX_ERR(0, 7424, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2680);
  __Pyx_GIVEREF(__pyx_tuple__2680);
  __pyx_codeobj__2681 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2680, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 7424, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2681)) __PYX_ERR(0, 7424, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7427
 *             return ModelicaParser.RULE_expression_list
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpression_list" ):
 *                 listener.enterExpression_list(self)
 */
  __pyx_tuple__2682 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2682)) __PYX_ERR(0, 7427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2682);
  __Pyx_GIVEREF(__pyx_tuple__2682);
  __pyx_codeobj__2683 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2682, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 7427, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2683)) __PYX_ERR(0, 7427, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7431
 *                 listener.enterExpression_list(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpression_list" ):
 *                 listener.exitExpression_list(self)
 */
  __pyx_tuple__2684 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2684)) __PYX_ERR(0, 7431, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2684);
  __Pyx_GIVEREF(__pyx_tuple__2684);
  __pyx_codeobj__2685 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2684, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 7431, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2685)) __PYX_ERR(0, 7431, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7435
 *                 listener.exitExpression_list(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpression_list" ):
 *                 return visitor.visitExpression_list(self)
 */
  __pyx_tuple__2686 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2686)) __PYX_ERR(0, 7435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2686);
  __Pyx_GIVEREF(__pyx_tuple__2686);
  __pyx_codeobj__2687 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2686, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 7435, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2687)) __PYX_ERR(0, 7435, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7444
 * 
 * 
 *     def expression_list(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Expression_listContext(self, self._ctx, self.state)
 */
  __pyx_tuple__2688 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__2688)) __PYX_ERR(0, 7444, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2688);
  __Pyx_GIVEREF(__pyx_tuple__2688);
  __pyx_codeobj__2689 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2688, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_expression_list, 7444, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2689)) __PYX_ERR(0, 7444, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7475
 *     class Array_subscriptsContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__2690 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__2690)) __PYX_ERR(0, 7475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2690);
  __Pyx_GIVEREF(__pyx_tuple__2690);
  __pyx_codeobj__2691 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2690, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 7475, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2691)) __PYX_ERR(0, 7475, __pyx_L1_error)
  __pyx_tuple__2692 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__2692)) __PYX_ERR(0, 7475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2692);
  __Pyx_GIVEREF(__pyx_tuple__2692);

  /* "pymola/generated/ModelicaParser.pyx":7479
 *             self.parser = parser
 * 
 *         def subscript(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.SubscriptContext)
 */
  __pyx_tuple__2693 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__2693)) __PYX_ERR(0, 7479, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2693);
  __Pyx_GIVEREF(__pyx_tuple__2693);
  __pyx_codeobj__2694 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2693, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_subscript, 7479, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2694)) __PYX_ERR(0, 7479, __pyx_L1_error)
  __pyx_tuple__2695 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__2695)) __PYX_ERR(0, 7479, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2695);
  __Pyx_GIVEREF(__pyx_tuple__2695);

  /* "pymola/generated/ModelicaParser.pyx":7486
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_array_subscripts
 * 
 */
  __pyx_tuple__2696 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2696)) __PYX_ERR(0, 7486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2696);
  __Pyx_GIVEREF(__pyx_tuple__2696);
  __pyx_codeobj__2697 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2696, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 7486, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2697)) __PYX_ERR(0, 7486, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7489
 *             return ModelicaParser.RULE_array_subscripts
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterArray_subscripts" ):
 *                 listener.enterArray_subscripts(self)
 */
  __pyx_tuple__2698 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2698)) __PYX_ERR(0, 7489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2698);
  __Pyx_GIVEREF(__pyx_tuple__2698);
  __pyx_codeobj__2699 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2698, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 7489, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2699)) __PYX_ERR(0, 7489, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7493
 *                 listener.enterArray_subscripts(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitArray_subscripts" ):
 *                 listener.exitArray_subscripts(self)
 */
  __pyx_tuple__2700 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2700)) __PYX_ERR(0, 7493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2700);
  __Pyx_GIVEREF(__pyx_tuple__2700);
  __pyx_codeobj__2701 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2700, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 7493, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2701)) __PYX_ERR(0, 7493, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7497
 *                 listener.exitArray_subscripts(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitArray_subscripts" ):
 *                 return visitor.visitArray_subscripts(self)
 */
  __pyx_tuple__2702 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2702)) __PYX_ERR(0, 7497, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2702);
  __Pyx_GIVEREF(__pyx_tuple__2702);
  __pyx_codeobj__2703 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2702, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 7497, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2703)) __PYX_ERR(0, 7497, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7506
 * 
 * 
 *     def array_subscripts(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Array_subscriptsContext(self, self._ctx, self.state)
 */
  __pyx_tuple__2704 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__2704)) __PYX_ERR(0, 7506, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2704);
  __Pyx_GIVEREF(__pyx_tuple__2704);
  __pyx_codeobj__2705 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2704, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_array_subscripts, 7506, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2705)) __PYX_ERR(0, 7506, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7541
 *     class SubscriptContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__2706 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__2706)) __PYX_ERR(0, 7541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2706);
  __Pyx_GIVEREF(__pyx_tuple__2706);
  __pyx_codeobj__2707 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2706, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 7541, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2707)) __PYX_ERR(0, 7541, __pyx_L1_error)
  __pyx_tuple__2708 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__2708)) __PYX_ERR(0, 7541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2708);
  __Pyx_GIVEREF(__pyx_tuple__2708);

  /* "pymola/generated/ModelicaParser.pyx":7545
 *             self.parser = parser
 * 
 *         def expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 */
  __pyx_tuple__2709 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2709)) __PYX_ERR(0, 7545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2709);
  __Pyx_GIVEREF(__pyx_tuple__2709);
  __pyx_codeobj__2710 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2709, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_expression, 7545, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2710)) __PYX_ERR(0, 7545, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7549
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_subscript
 * 
 */
  __pyx_tuple__2711 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2711)) __PYX_ERR(0, 7549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2711);
  __Pyx_GIVEREF(__pyx_tuple__2711);
  __pyx_codeobj__2712 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2711, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 7549, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2712)) __PYX_ERR(0, 7549, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7552
 *             return ModelicaParser.RULE_subscript
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterSubscript" ):
 *                 listener.enterSubscript(self)
 */
  __pyx_tuple__2713 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2713)) __PYX_ERR(0, 7552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2713);
  __Pyx_GIVEREF(__pyx_tuple__2713);
  __pyx_codeobj__2714 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2713, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 7552, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2714)) __PYX_ERR(0, 7552, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7556
 *                 listener.enterSubscript(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitSubscript" ):
 *                 listener.exitSubscript(self)
 */
  __pyx_tuple__2715 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2715)) __PYX_ERR(0, 7556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2715);
  __Pyx_GIVEREF(__pyx_tuple__2715);
  __pyx_codeobj__2716 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2715, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 7556, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2716)) __PYX_ERR(0, 7556, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7560
 *                 listener.exitSubscript(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitSubscript" ):
 *                 return visitor.visitSubscript(self)
 */
  __pyx_tuple__2717 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2717)) __PYX_ERR(0, 7560, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2717);
  __Pyx_GIVEREF(__pyx_tuple__2717);
  __pyx_codeobj__2718 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2717, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 7560, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2718)) __PYX_ERR(0, 7560, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7569
 * 
 * 
 *     def subscript(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.SubscriptContext(self, self._ctx, self.state)
 */
  __pyx_tuple__2719 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_token, __pyx_n_s_re); if (unlikely(!__pyx_tuple__2719)) __PYX_ERR(0, 7569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2719);
  __Pyx_GIVEREF(__pyx_tuple__2719);
  __pyx_codeobj__2720 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2719, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_subscript, 7569, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2720)) __PYX_ERR(0, 7569, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7600
 *     class CommentContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__2721 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__2721)) __PYX_ERR(0, 7600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2721);
  __Pyx_GIVEREF(__pyx_tuple__2721);
  __pyx_codeobj__2722 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2721, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 7600, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2722)) __PYX_ERR(0, 7600, __pyx_L1_error)
  __pyx_tuple__2723 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__2723)) __PYX_ERR(0, 7600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2723);
  __Pyx_GIVEREF(__pyx_tuple__2723);

  /* "pymola/generated/ModelicaParser.pyx":7604
 *             self.parser = parser
 * 
 *         def string_comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.String_commentContext,0)
 * 
 */
  __pyx_tuple__2724 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2724)) __PYX_ERR(0, 7604, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2724);
  __Pyx_GIVEREF(__pyx_tuple__2724);
  __pyx_codeobj__2725 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2724, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_string_comment, 7604, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2725)) __PYX_ERR(0, 7604, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7608
 * 
 * 
 *         def annotation(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.AnnotationContext,0)
 * 
 */
  __pyx_tuple__2726 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2726)) __PYX_ERR(0, 7608, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2726);
  __Pyx_GIVEREF(__pyx_tuple__2726);
  __pyx_codeobj__2727 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2726, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_annotation, 7608, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2727)) __PYX_ERR(0, 7608, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7612
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_comment
 * 
 */
  __pyx_tuple__2728 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2728)) __PYX_ERR(0, 7612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2728);
  __Pyx_GIVEREF(__pyx_tuple__2728);
  __pyx_codeobj__2729 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2728, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 7612, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2729)) __PYX_ERR(0, 7612, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7615
 *             return ModelicaParser.RULE_comment
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterComment" ):
 *                 listener.enterComment(self)
 */
  __pyx_tuple__2730 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2730)) __PYX_ERR(0, 7615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2730);
  __Pyx_GIVEREF(__pyx_tuple__2730);
  __pyx_codeobj__2731 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2730, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 7615, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2731)) __PYX_ERR(0, 7615, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7619
 *                 listener.enterComment(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitComment" ):
 *                 listener.exitComment(self)
 */
  __pyx_tuple__2732 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2732)) __PYX_ERR(0, 7619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2732);
  __Pyx_GIVEREF(__pyx_tuple__2732);
  __pyx_codeobj__2733 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2732, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 7619, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2733)) __PYX_ERR(0, 7619, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7623
 *                 listener.exitComment(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitComment" ):
 *                 return visitor.visitComment(self)
 */
  __pyx_tuple__2734 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2734)) __PYX_ERR(0, 7623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2734);
  __Pyx_GIVEREF(__pyx_tuple__2734);
  __pyx_codeobj__2735 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2734, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 7623, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2735)) __PYX_ERR(0, 7623, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7632
 * 
 * 
 *     def comment(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.CommentContext(self, self._ctx, self.state)
 */
  __pyx_tuple__2736 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__2736)) __PYX_ERR(0, 7632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2736);
  __Pyx_GIVEREF(__pyx_tuple__2736);
  __pyx_codeobj__2737 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2736, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_comment, 7632, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2737)) __PYX_ERR(0, 7632, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7659
 *     class String_commentContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__2738 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__2738)) __PYX_ERR(0, 7659, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2738);
  __Pyx_GIVEREF(__pyx_tuple__2738);
  __pyx_codeobj__2739 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2738, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 7659, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2739)) __PYX_ERR(0, 7659, __pyx_L1_error)
  __pyx_tuple__2740 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__2740)) __PYX_ERR(0, 7659, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2740);
  __Pyx_GIVEREF(__pyx_tuple__2740);

  /* "pymola/generated/ModelicaParser.pyx":7663
 *             self.parser = parser
 * 
 *         def STRING(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTokens(ModelicaParser.STRING)
 */
  __pyx_tuple__2741 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__2741)) __PYX_ERR(0, 7663, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2741);
  __Pyx_GIVEREF(__pyx_tuple__2741);
  __pyx_codeobj__2742 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2741, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_STRING, 7663, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2742)) __PYX_ERR(0, 7663, __pyx_L1_error)
  __pyx_tuple__2743 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__2743)) __PYX_ERR(0, 7663, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2743);
  __Pyx_GIVEREF(__pyx_tuple__2743);

  /* "pymola/generated/ModelicaParser.pyx":7669
 *                 return self.getToken(ModelicaParser.STRING, i)
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_string_comment
 * 
 */
  __pyx_tuple__2744 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2744)) __PYX_ERR(0, 7669, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2744);
  __Pyx_GIVEREF(__pyx_tuple__2744);
  __pyx_codeobj__2745 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2744, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 7669, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2745)) __PYX_ERR(0, 7669, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7672
 *             return ModelicaParser.RULE_string_comment
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterString_comment" ):
 *                 listener.enterString_comment(self)
 */
  __pyx_tuple__2746 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2746)) __PYX_ERR(0, 7672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2746);
  __Pyx_GIVEREF(__pyx_tuple__2746);
  __pyx_codeobj__2747 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2746, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 7672, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2747)) __PYX_ERR(0, 7672, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7676
 *                 listener.enterString_comment(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitString_comment" ):
 *                 listener.exitString_comment(self)
 */
  __pyx_tuple__2748 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2748)) __PYX_ERR(0, 7676, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2748);
  __Pyx_GIVEREF(__pyx_tuple__2748);
  __pyx_codeobj__2749 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2748, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 7676, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2749)) __PYX_ERR(0, 7676, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7680
 *                 listener.exitString_comment(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitString_comment" ):
 *                 return visitor.visitString_comment(self)
 */
  __pyx_tuple__2750 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2750)) __PYX_ERR(0, 7680, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2750);
  __Pyx_GIVEREF(__pyx_tuple__2750);
  __pyx_codeobj__2751 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2750, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 7680, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2751)) __PYX_ERR(0, 7680, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7689
 * 
 * 
 *     def string_comment(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.String_commentContext(self, self._ctx, self.state)
 */
  __pyx_tuple__2752 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_la, __pyx_n_s_re); if (unlikely(!__pyx_tuple__2752)) __PYX_ERR(0, 7689, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2752);
  __Pyx_GIVEREF(__pyx_tuple__2752);
  __pyx_codeobj__2753 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2752, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_string_comment, 7689, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2753)) __PYX_ERR(0, 7689, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7726
 *     class AnnotationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_tuple__2754 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parser, __pyx_n_s_parent, __pyx_n_s_invokingState); if (unlikely(!__pyx_tuple__2754)) __PYX_ERR(0, 7726, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2754);
  __Pyx_GIVEREF(__pyx_tuple__2754);
  __pyx_codeobj__2755 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2754, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_init, 7726, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2755)) __PYX_ERR(0, 7726, __pyx_L1_error)
  __pyx_tuple__2756 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)__pyx_int_neg_1)); if (unlikely(!__pyx_tuple__2756)) __PYX_ERR(0, 7726, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2756);
  __Pyx_GIVEREF(__pyx_tuple__2756);

  /* "pymola/generated/ModelicaParser.pyx":7730
 *             self.parser = parser
 * 
 *         def class_modification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_modificationContext,0)
 * 
 */
  __pyx_tuple__2757 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2757)) __PYX_ERR(0, 7730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2757);
  __Pyx_GIVEREF(__pyx_tuple__2757);
  __pyx_codeobj__2758 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2757, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_class_modification, 7730, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2758)) __PYX_ERR(0, 7730, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7734
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_annotation
 * 
 */
  __pyx_tuple__2759 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__2759)) __PYX_ERR(0, 7734, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2759);
  __Pyx_GIVEREF(__pyx_tuple__2759);
  __pyx_codeobj__2760 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2759, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_getRuleIndex, 7734, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2760)) __PYX_ERR(0, 7734, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7737
 *             return ModelicaParser.RULE_annotation
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterAnnotation" ):
 *                 listener.enterAnnotation(self)
 */
  __pyx_tuple__2761 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2761)) __PYX_ERR(0, 7737, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2761);
  __Pyx_GIVEREF(__pyx_tuple__2761);
  __pyx_codeobj__2762 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2761, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRule, 7737, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2762)) __PYX_ERR(0, 7737, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7741
 *                 listener.enterAnnotation(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitAnnotation" ):
 *                 listener.exitAnnotation(self)
 */
  __pyx_tuple__2763 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_listener); if (unlikely(!__pyx_tuple__2763)) __PYX_ERR(0, 7741, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2763);
  __Pyx_GIVEREF(__pyx_tuple__2763);
  __pyx_codeobj__2764 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2763, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRule, 7741, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2764)) __PYX_ERR(0, 7741, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7745
 *                 listener.exitAnnotation(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitAnnotation" ):
 *                 return visitor.visitAnnotation(self)
 */
  __pyx_tuple__2765 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visitor); if (unlikely(!__pyx_tuple__2765)) __PYX_ERR(0, 7745, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2765);
  __Pyx_GIVEREF(__pyx_tuple__2765);
  __pyx_codeobj__2766 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2765, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_accept, 7745, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2766)) __PYX_ERR(0, 7745, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7754
 * 
 * 
 *     def annotation(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.AnnotationContext(self, self._ctx, self.state)
 */
  __pyx_tuple__2767 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_re); if (unlikely(!__pyx_tuple__2767)) __PYX_ERR(0, 7754, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2767);
  __Pyx_GIVEREF(__pyx_tuple__2767);
  __pyx_codeobj__2768 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2767, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_annotation, 7754, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2768)) __PYX_ERR(0, 7754, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7774
 * 
 * 
 *     def sempred(self, localctx:RuleContext, ruleIndex:int, predIndex:int):             # <<<<<<<<<<<<<<
 *         if self._predicates == None:
 *             self._predicates = dict()
 */
  __pyx_tuple__2769 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_ruleIndex, __pyx_n_s_predIndex, __pyx_n_s_pred); if (unlikely(!__pyx_tuple__2769)) __PYX_ERR(0, 7774, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2769);
  __Pyx_GIVEREF(__pyx_tuple__2769);
  __pyx_codeobj__2770 = (PyObject*)__Pyx_PyCode_New(4, 0, 5, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2769, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_sempred, 7774, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2770)) __PYX_ERR(0, 7774, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":7784
 *             return pred(localctx, predIndex)
 * 
 *     def expr_sempred(self, localctx:ExprContext, predIndex:int):             # <<<<<<<<<<<<<<
 *             if predIndex == 0:
 *                 return self.precpred(self._ctx, 7)
 */
  __pyx_tuple__2771 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_localctx, __pyx_n_s_predIndex); if (unlikely(!__pyx_tuple__2771)) __PYX_ERR(0, 7784, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2771);
  __Pyx_GIVEREF(__pyx_tuple__2771);
  __pyx_codeobj__2772 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2771, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_expr_sempred, 7784, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2772)) __PYX_ERR(0, 7784, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4 = PyInt_FromLong(4); if (unlikely(!__pyx_int_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_5 = PyInt_FromLong(5); if (unlikely(!__pyx_int_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_6 = PyInt_FromLong(6); if (unlikely(!__pyx_int_6)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_7 = PyInt_FromLong(7); if (unlikely(!__pyx_int_7)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_8 = PyInt_FromLong(8); if (unlikely(!__pyx_int_8)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_9 = PyInt_FromLong(9); if (unlikely(!__pyx_int_9)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_10 = PyInt_FromLong(10); if (unlikely(!__pyx_int_10)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_11 = PyInt_FromLong(11); if (unlikely(!__pyx_int_11)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_12 = PyInt_FromLong(12); if (unlikely(!__pyx_int_12)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_13 = PyInt_FromLong(13); if (unlikely(!__pyx_int_13)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_14 = PyInt_FromLong(14); if (unlikely(!__pyx_int_14)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_15 = PyInt_FromLong(15); if (unlikely(!__pyx_int_15)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_16 = PyInt_FromLong(16); if (unlikely(!__pyx_int_16)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_17 = PyInt_FromLong(17); if (unlikely(!__pyx_int_17)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_18 = PyInt_FromLong(18); if (unlikely(!__pyx_int_18)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_19 = PyInt_FromLong(19); if (unlikely(!__pyx_int_19)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_20 = PyInt_FromLong(20); if (unlikely(!__pyx_int_20)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_21 = PyInt_FromLong(21); if (unlikely(!__pyx_int_21)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_22 = PyInt_FromLong(22); if (unlikely(!__pyx_int_22)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_23 = PyInt_FromLong(23); if (unlikely(!__pyx_int_23)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_24 = PyInt_FromLong(24); if (unlikely(!__pyx_int_24)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_25 = PyInt_FromLong(25); if (unlikely(!__pyx_int_25)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_26 = PyInt_FromLong(26); if (unlikely(!__pyx_int_26)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_27 = PyInt_FromLong(27); if (unlikely(!__pyx_int_27)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_28 = PyInt_FromLong(28); if (unlikely(!__pyx_int_28)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_29 = PyInt_FromLong(29); if (unlikely(!__pyx_int_29)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_30 = PyInt_FromLong(30); if (unlikely(!__pyx_int_30)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_31 = PyInt_FromLong(31); if (unlikely(!__pyx_int_31)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_32 = PyInt_FromLong(32); if (unlikely(!__pyx_int_32)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_33 = PyInt_FromLong(33); if (unlikely(!__pyx_int_33)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_34 = PyInt_FromLong(34); if (unlikely(!__pyx_int_34)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_35 = PyInt_FromLong(35); if (unlikely(!__pyx_int_35)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_36 = PyInt_FromLong(36); if (unlikely(!__pyx_int_36)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_37 = PyInt_FromLong(37); if (unlikely(!__pyx_int_37)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_38 = PyInt_FromLong(38); if (unlikely(!__pyx_int_38)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_39 = PyInt_FromLong(39); if (unlikely(!__pyx_int_39)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_40 = PyInt_FromLong(40); if (unlikely(!__pyx_int_40)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_41 = PyInt_FromLong(41); if (unlikely(!__pyx_int_41)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_42 = PyInt_FromLong(42); if (unlikely(!__pyx_int_42)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_43 = PyInt_FromLong(43); if (unlikely(!__pyx_int_43)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_44 = PyInt_FromLong(44); if (unlikely(!__pyx_int_44)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_45 = PyInt_FromLong(45); if (unlikely(!__pyx_int_45)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_46 = PyInt_FromLong(46); if (unlikely(!__pyx_int_46)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_47 = PyInt_FromLong(47); if (unlikely(!__pyx_int_47)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_48 = PyInt_FromLong(48); if (unlikely(!__pyx_int_48)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_49 = PyInt_FromLong(49); if (unlikely(!__pyx_int_49)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_50 = PyInt_FromLong(50); if (unlikely(!__pyx_int_50)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_51 = PyInt_FromLong(51); if (unlikely(!__pyx_int_51)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_52 = PyInt_FromLong(52); if (unlikely(!__pyx_int_52)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_53 = PyInt_FromLong(53); if (unlikely(!__pyx_int_53)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_54 = PyInt_FromLong(54); if (unlikely(!__pyx_int_54)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_55 = PyInt_FromLong(55); if (unlikely(!__pyx_int_55)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_56 = PyInt_FromLong(56); if (unlikely(!__pyx_int_56)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_57 = PyInt_FromLong(57); if (unlikely(!__pyx_int_57)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_58 = PyInt_FromLong(58); if (unlikely(!__pyx_int_58)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_59 = PyInt_FromLong(59); if (unlikely(!__pyx_int_59)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_60 = PyInt_FromLong(60); if (unlikely(!__pyx_int_60)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_61 = PyInt_FromLong(61); if (unlikely(!__pyx_int_61)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_62 = PyInt_FromLong(62); if (unlikely(!__pyx_int_62)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_63 = PyInt_FromLong(63); if (unlikely(!__pyx_int_63)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_64 = PyInt_FromLong(64); if (unlikely(!__pyx_int_64)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_65 = PyInt_FromLong(65); if (unlikely(!__pyx_int_65)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_66 = PyInt_FromLong(66); if (unlikely(!__pyx_int_66)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_67 = PyInt_FromLong(67); if (unlikely(!__pyx_int_67)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_68 = PyInt_FromLong(68); if (unlikely(!__pyx_int_68)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_69 = PyInt_FromLong(69); if (unlikely(!__pyx_int_69)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_70 = PyInt_FromLong(70); if (unlikely(!__pyx_int_70)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_71 = PyInt_FromLong(71); if (unlikely(!__pyx_int_71)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_72 = PyInt_FromLong(72); if (unlikely(!__pyx_int_72)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_73 = PyInt_FromLong(73); if (unlikely(!__pyx_int_73)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_74 = PyInt_FromLong(74); if (unlikely(!__pyx_int_74)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_75 = PyInt_FromLong(75); if (unlikely(!__pyx_int_75)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_76 = PyInt_FromLong(76); if (unlikely(!__pyx_int_76)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_77 = PyInt_FromLong(77); if (unlikely(!__pyx_int_77)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_78 = PyInt_FromLong(78); if (unlikely(!__pyx_int_78)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_79 = PyInt_FromLong(79); if (unlikely(!__pyx_int_79)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_80 = PyInt_FromLong(80); if (unlikely(!__pyx_int_80)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_81 = PyInt_FromLong(81); if (unlikely(!__pyx_int_81)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_82 = PyInt_FromLong(82); if (unlikely(!__pyx_int_82)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_83 = PyInt_FromLong(83); if (unlikely(!__pyx_int_83)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_84 = PyInt_FromLong(84); if (unlikely(!__pyx_int_84)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_85 = PyInt_FromLong(85); if (unlikely(!__pyx_int_85)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_86 = PyInt_FromLong(86); if (unlikely(!__pyx_int_86)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_87 = PyInt_FromLong(87); if (unlikely(!__pyx_int_87)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_88 = PyInt_FromLong(88); if (unlikely(!__pyx_int_88)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_89 = PyInt_FromLong(89); if (unlikely(!__pyx_int_89)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_90 = PyInt_FromLong(90); if (unlikely(!__pyx_int_90)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_91 = PyInt_FromLong(91); if (unlikely(!__pyx_int_91)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_92 = PyInt_FromLong(92); if (unlikely(!__pyx_int_92)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_94 = PyInt_FromLong(94); if (unlikely(!__pyx_int_94)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_96 = PyInt_FromLong(96); if (unlikely(!__pyx_int_96)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_98 = PyInt_FromLong(98); if (unlikely(!__pyx_int_98)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_99 = PyInt_FromLong(99); if (unlikely(!__pyx_int_99)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_100 = PyInt_FromLong(100); if (unlikely(!__pyx_int_100)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_102 = PyInt_FromLong(102); if (unlikely(!__pyx_int_102)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_104 = PyInt_FromLong(104); if (unlikely(!__pyx_int_104)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_106 = PyInt_FromLong(106); if (unlikely(!__pyx_int_106)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_108 = PyInt_FromLong(108); if (unlikely(!__pyx_int_108)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_109 = PyInt_FromLong(109); if (unlikely(!__pyx_int_109)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_110 = PyInt_FromLong(110); if (unlikely(!__pyx_int_110)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_111 = PyInt_FromLong(111); if (unlikely(!__pyx_int_111)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_112 = PyInt_FromLong(112); if (unlikely(!__pyx_int_112)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_113 = PyInt_FromLong(113); if (unlikely(!__pyx_int_113)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_114 = PyInt_FromLong(114); if (unlikely(!__pyx_int_114)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_116 = PyInt_FromLong(116); if (unlikely(!__pyx_int_116)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_117 = PyInt_FromLong(117); if (unlikely(!__pyx_int_117)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_118 = PyInt_FromLong(118); if (unlikely(!__pyx_int_118)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_120 = PyInt_FromLong(120); if (unlikely(!__pyx_int_120)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_121 = PyInt_FromLong(121); if (unlikely(!__pyx_int_121)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_122 = PyInt_FromLong(122); if (unlikely(!__pyx_int_122)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_124 = PyInt_FromLong(124); if (unlikely(!__pyx_int_124)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_126 = PyInt_FromLong(126); if (unlikely(!__pyx_int_126)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_128 = PyInt_FromLong(128); if (unlikely(!__pyx_int_128)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_130 = PyInt_FromLong(130); if (unlikely(!__pyx_int_130)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_132 = PyInt_FromLong(132); if (unlikely(!__pyx_int_132)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_134 = PyInt_FromLong(134); if (unlikely(!__pyx_int_134)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_136 = PyInt_FromLong(136); if (unlikely(!__pyx_int_136)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_138 = PyInt_FromLong(138); if (unlikely(!__pyx_int_138)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_140 = PyInt_FromLong(140); if (unlikely(!__pyx_int_140)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_142 = PyInt_FromLong(142); if (unlikely(!__pyx_int_142)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_144 = PyInt_FromLong(144); if (unlikely(!__pyx_int_144)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_146 = PyInt_FromLong(146); if (unlikely(!__pyx_int_146)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_147 = PyInt_FromLong(147); if (unlikely(!__pyx_int_147)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_148 = PyInt_FromLong(148); if (unlikely(!__pyx_int_148)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_150 = PyInt_FromLong(150); if (unlikely(!__pyx_int_150)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_151 = PyInt_FromLong(151); if (unlikely(!__pyx_int_151)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_153 = PyInt_FromLong(153); if (unlikely(!__pyx_int_153)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_156 = PyInt_FromLong(156); if (unlikely(!__pyx_int_156)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_158 = PyInt_FromLong(158); if (unlikely(!__pyx_int_158)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_159 = PyInt_FromLong(159); if (unlikely(!__pyx_int_159)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_160 = PyInt_FromLong(160); if (unlikely(!__pyx_int_160)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_162 = PyInt_FromLong(162); if (unlikely(!__pyx_int_162)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_163 = PyInt_FromLong(163); if (unlikely(!__pyx_int_163)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_165 = PyInt_FromLong(165); if (unlikely(!__pyx_int_165)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_166 = PyInt_FromLong(166); if (unlikely(!__pyx_int_166)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_168 = PyInt_FromLong(168); if (unlikely(!__pyx_int_168)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_169 = PyInt_FromLong(169); if (unlikely(!__pyx_int_169)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_171 = PyInt_FromLong(171); if (unlikely(!__pyx_int_171)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_172 = PyInt_FromLong(172); if (unlikely(!__pyx_int_172)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_174 = PyInt_FromLong(174); if (unlikely(!__pyx_int_174)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_176 = PyInt_FromLong(176); if (unlikely(!__pyx_int_176)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_177 = PyInt_FromLong(177); if (unlikely(!__pyx_int_177)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_178 = PyInt_FromLong(178); if (unlikely(!__pyx_int_178)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_179 = PyInt_FromLong(179); if (unlikely(!__pyx_int_179)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_181 = PyInt_FromLong(181); if (unlikely(!__pyx_int_181)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_182 = PyInt_FromLong(182); if (unlikely(!__pyx_int_182)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_183 = PyInt_FromLong(183); if (unlikely(!__pyx_int_183)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_184 = PyInt_FromLong(184); if (unlikely(!__pyx_int_184)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_186 = PyInt_FromLong(186); if (unlikely(!__pyx_int_186)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_187 = PyInt_FromLong(187); if (unlikely(!__pyx_int_187)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_188 = PyInt_FromLong(188); if (unlikely(!__pyx_int_188)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_189 = PyInt_FromLong(189); if (unlikely(!__pyx_int_189)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_190 = PyInt_FromLong(190); if (unlikely(!__pyx_int_190)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_192 = PyInt_FromLong(192); if (unlikely(!__pyx_int_192)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_193 = PyInt_FromLong(193); if (unlikely(!__pyx_int_193)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_195 = PyInt_FromLong(195); if (unlikely(!__pyx_int_195)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_196 = PyInt_FromLong(196); if (unlikely(!__pyx_int_196)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_197 = PyInt_FromLong(197); if (unlikely(!__pyx_int_197)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_199 = PyInt_FromLong(199); if (unlikely(!__pyx_int_199)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_200 = PyInt_FromLong(200); if (unlikely(!__pyx_int_200)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_201 = PyInt_FromLong(201); if (unlikely(!__pyx_int_201)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_202 = PyInt_FromLong(202); if (unlikely(!__pyx_int_202)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_203 = PyInt_FromLong(203); if (unlikely(!__pyx_int_203)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_205 = PyInt_FromLong(205); if (unlikely(!__pyx_int_205)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_206 = PyInt_FromLong(206); if (unlikely(!__pyx_int_206)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_207 = PyInt_FromLong(207); if (unlikely(!__pyx_int_207)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_208 = PyInt_FromLong(208); if (unlikely(!__pyx_int_208)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_209 = PyInt_FromLong(209); if (unlikely(!__pyx_int_209)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_210 = PyInt_FromLong(210); if (unlikely(!__pyx_int_210)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_212 = PyInt_FromLong(212); if (unlikely(!__pyx_int_212)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_213 = PyInt_FromLong(213); if (unlikely(!__pyx_int_213)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_215 = PyInt_FromLong(215); if (unlikely(!__pyx_int_215)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_217 = PyInt_FromLong(217); if (unlikely(!__pyx_int_217)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_218 = PyInt_FromLong(218); if (unlikely(!__pyx_int_218)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_219 = PyInt_FromLong(219); if (unlikely(!__pyx_int_219)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_220 = PyInt_FromLong(220); if (unlikely(!__pyx_int_220)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_221 = PyInt_FromLong(221); if (unlikely(!__pyx_int_221)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_222 = PyInt_FromLong(222); if (unlikely(!__pyx_int_222)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_224 = PyInt_FromLong(224); if (unlikely(!__pyx_int_224)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_225 = PyInt_FromLong(225); if (unlikely(!__pyx_int_225)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_227 = PyInt_FromLong(227); if (unlikely(!__pyx_int_227)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_228 = PyInt_FromLong(228); if (unlikely(!__pyx_int_228)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_229 = PyInt_FromLong(229); if (unlikely(!__pyx_int_229)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_230 = PyInt_FromLong(230); if (unlikely(!__pyx_int_230)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_231 = PyInt_FromLong(231); if (unlikely(!__pyx_int_231)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_232 = PyInt_FromLong(232); if (unlikely(!__pyx_int_232)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_233 = PyInt_FromLong(233); if (unlikely(!__pyx_int_233)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_234 = PyInt_FromLong(234); if (unlikely(!__pyx_int_234)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_235 = PyInt_FromLong(235); if (unlikely(!__pyx_int_235)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_236 = PyInt_FromLong(236); if (unlikely(!__pyx_int_236)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_237 = PyInt_FromLong(237); if (unlikely(!__pyx_int_237)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_240 = PyInt_FromLong(240); if (unlikely(!__pyx_int_240)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_242 = PyInt_FromLong(242); if (unlikely(!__pyx_int_242)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_243 = PyInt_FromLong(243); if (unlikely(!__pyx_int_243)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_244 = PyInt_FromLong(244); if (unlikely(!__pyx_int_244)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_246 = PyInt_FromLong(246); if (unlikely(!__pyx_int_246)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_247 = PyInt_FromLong(247); if (unlikely(!__pyx_int_247)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_248 = PyInt_FromLong(248); if (unlikely(!__pyx_int_248)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_249 = PyInt_FromLong(249); if (unlikely(!__pyx_int_249)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_251 = PyInt_FromLong(251); if (unlikely(!__pyx_int_251)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_252 = PyInt_FromLong(252); if (unlikely(!__pyx_int_252)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_253 = PyInt_FromLong(253); if (unlikely(!__pyx_int_253)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_254 = PyInt_FromLong(254); if (unlikely(!__pyx_int_254)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_256 = PyInt_FromLong(256); if (unlikely(!__pyx_int_256)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_258 = PyInt_FromLong(258); if (unlikely(!__pyx_int_258)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_260 = PyInt_FromLong(260); if (unlikely(!__pyx_int_260)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_261 = PyInt_FromLong(261); if (unlikely(!__pyx_int_261)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_262 = PyInt_FromLong(262); if (unlikely(!__pyx_int_262)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_265 = PyInt_FromLong(265); if (unlikely(!__pyx_int_265)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_267 = PyInt_FromLong(267); if (unlikely(!__pyx_int_267)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_268 = PyInt_FromLong(268); if (unlikely(!__pyx_int_268)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_269 = PyInt_FromLong(269); if (unlikely(!__pyx_int_269)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_271 = PyInt_FromLong(271); if (unlikely(!__pyx_int_271)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_272 = PyInt_FromLong(272); if (unlikely(!__pyx_int_272)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_273 = PyInt_FromLong(273); if (unlikely(!__pyx_int_273)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_274 = PyInt_FromLong(274); if (unlikely(!__pyx_int_274)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_275 = PyInt_FromLong(275); if (unlikely(!__pyx_int_275)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_276 = PyInt_FromLong(276); if (unlikely(!__pyx_int_276)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_277 = PyInt_FromLong(277); if (unlikely(!__pyx_int_277)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_278 = PyInt_FromLong(278); if (unlikely(!__pyx_int_278)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_280 = PyInt_FromLong(280); if (unlikely(!__pyx_int_280)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_282 = PyInt_FromLong(282); if (unlikely(!__pyx_int_282)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_283 = PyInt_FromLong(283); if (unlikely(!__pyx_int_283)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_284 = PyInt_FromLong(284); if (unlikely(!__pyx_int_284)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_285 = PyInt_FromLong(285); if (unlikely(!__pyx_int_285)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_287 = PyInt_FromLong(287); if (unlikely(!__pyx_int_287)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_288 = PyInt_FromLong(288); if (unlikely(!__pyx_int_288)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_290 = PyInt_FromLong(290); if (unlikely(!__pyx_int_290)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_291 = PyInt_FromLong(291); if (unlikely(!__pyx_int_291)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_293 = PyInt_FromLong(293); if (unlikely(!__pyx_int_293)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_294 = PyInt_FromLong(294); if (unlikely(!__pyx_int_294)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_296 = PyInt_FromLong(296); if (unlikely(!__pyx_int_296)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_297 = PyInt_FromLong(297); if (unlikely(!__pyx_int_297)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_299 = PyInt_FromLong(299); if (unlikely(!__pyx_int_299)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_301 = PyInt_FromLong(301); if (unlikely(!__pyx_int_301)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_303 = PyInt_FromLong(303); if (unlikely(!__pyx_int_303)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_304 = PyInt_FromLong(304); if (unlikely(!__pyx_int_304)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_306 = PyInt_FromLong(306); if (unlikely(!__pyx_int_306)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_308 = PyInt_FromLong(308); if (unlikely(!__pyx_int_308)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_309 = PyInt_FromLong(309); if (unlikely(!__pyx_int_309)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_310 = PyInt_FromLong(310); if (unlikely(!__pyx_int_310)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_311 = PyInt_FromLong(311); if (unlikely(!__pyx_int_311)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_313 = PyInt_FromLong(313); if (unlikely(!__pyx_int_313)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_315 = PyInt_FromLong(315); if (unlikely(!__pyx_int_315)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_316 = PyInt_FromLong(316); if (unlikely(!__pyx_int_316)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_320 = PyInt_FromLong(320); if (unlikely(!__pyx_int_320)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_322 = PyInt_FromLong(322); if (unlikely(!__pyx_int_322)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_323 = PyInt_FromLong(323); if (unlikely(!__pyx_int_323)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_324 = PyInt_FromLong(324); if (unlikely(!__pyx_int_324)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_325 = PyInt_FromLong(325); if (unlikely(!__pyx_int_325)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_326 = PyInt_FromLong(326); if (unlikely(!__pyx_int_326)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_327 = PyInt_FromLong(327); if (unlikely(!__pyx_int_327)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_329 = PyInt_FromLong(329); if (unlikely(!__pyx_int_329)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_330 = PyInt_FromLong(330); if (unlikely(!__pyx_int_330)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_332 = PyInt_FromLong(332); if (unlikely(!__pyx_int_332)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_333 = PyInt_FromLong(333); if (unlikely(!__pyx_int_333)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_335 = PyInt_FromLong(335); if (unlikely(!__pyx_int_335)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_336 = PyInt_FromLong(336); if (unlikely(!__pyx_int_336)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_338 = PyInt_FromLong(338); if (unlikely(!__pyx_int_338)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_339 = PyInt_FromLong(339); if (unlikely(!__pyx_int_339)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_341 = PyInt_FromLong(341); if (unlikely(!__pyx_int_341)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_342 = PyInt_FromLong(342); if (unlikely(!__pyx_int_342)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_343 = PyInt_FromLong(343); if (unlikely(!__pyx_int_343)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_345 = PyInt_FromLong(345); if (unlikely(!__pyx_int_345)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_346 = PyInt_FromLong(346); if (unlikely(!__pyx_int_346)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_348 = PyInt_FromLong(348); if (unlikely(!__pyx_int_348)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_349 = PyInt_FromLong(349); if (unlikely(!__pyx_int_349)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_351 = PyInt_FromLong(351); if (unlikely(!__pyx_int_351)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_352 = PyInt_FromLong(352); if (unlikely(!__pyx_int_352)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_354 = PyInt_FromLong(354); if (unlikely(!__pyx_int_354)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_355 = PyInt_FromLong(355); if (unlikely(!__pyx_int_355)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_357 = PyInt_FromLong(357); if (unlikely(!__pyx_int_357)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_358 = PyInt_FromLong(358); if (unlikely(!__pyx_int_358)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_359 = PyInt_FromLong(359); if (unlikely(!__pyx_int_359)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_360 = PyInt_FromLong(360); if (unlikely(!__pyx_int_360)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_362 = PyInt_FromLong(362); if (unlikely(!__pyx_int_362)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_363 = PyInt_FromLong(363); if (unlikely(!__pyx_int_363)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_365 = PyInt_FromLong(365); if (unlikely(!__pyx_int_365)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_367 = PyInt_FromLong(367); if (unlikely(!__pyx_int_367)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_368 = PyInt_FromLong(368); if (unlikely(!__pyx_int_368)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_369 = PyInt_FromLong(369); if (unlikely(!__pyx_int_369)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_370 = PyInt_FromLong(370); if (unlikely(!__pyx_int_370)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_371 = PyInt_FromLong(371); if (unlikely(!__pyx_int_371)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_372 = PyInt_FromLong(372); if (unlikely(!__pyx_int_372)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_373 = PyInt_FromLong(373); if (unlikely(!__pyx_int_373)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_374 = PyInt_FromLong(374); if (unlikely(!__pyx_int_374)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_375 = PyInt_FromLong(375); if (unlikely(!__pyx_int_375)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_376 = PyInt_FromLong(376); if (unlikely(!__pyx_int_376)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_378 = PyInt_FromLong(378); if (unlikely(!__pyx_int_378)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_380 = PyInt_FromLong(380); if (unlikely(!__pyx_int_380)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_382 = PyInt_FromLong(382); if (unlikely(!__pyx_int_382)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_384 = PyInt_FromLong(384); if (unlikely(!__pyx_int_384)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_386 = PyInt_FromLong(386); if (unlikely(!__pyx_int_386)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_387 = PyInt_FromLong(387); if (unlikely(!__pyx_int_387)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_388 = PyInt_FromLong(388); if (unlikely(!__pyx_int_388)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_391 = PyInt_FromLong(391); if (unlikely(!__pyx_int_391)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_393 = PyInt_FromLong(393); if (unlikely(!__pyx_int_393)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_394 = PyInt_FromLong(394); if (unlikely(!__pyx_int_394)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_395 = PyInt_FromLong(395); if (unlikely(!__pyx_int_395)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_396 = PyInt_FromLong(396); if (unlikely(!__pyx_int_396)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_397 = PyInt_FromLong(397); if (unlikely(!__pyx_int_397)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_399 = PyInt_FromLong(399); if (unlikely(!__pyx_int_399)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_400 = PyInt_FromLong(400); if (unlikely(!__pyx_int_400)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_402 = PyInt_FromLong(402); if (unlikely(!__pyx_int_402)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_403 = PyInt_FromLong(403); if (unlikely(!__pyx_int_403)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_404 = PyInt_FromLong(404); if (unlikely(!__pyx_int_404)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_405 = PyInt_FromLong(405); if (unlikely(!__pyx_int_405)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_407 = PyInt_FromLong(407); if (unlikely(!__pyx_int_407)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_408 = PyInt_FromLong(408); if (unlikely(!__pyx_int_408)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_409 = PyInt_FromLong(409); if (unlikely(!__pyx_int_409)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_410 = PyInt_FromLong(410); if (unlikely(!__pyx_int_410)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_412 = PyInt_FromLong(412); if (unlikely(!__pyx_int_412)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_414 = PyInt_FromLong(414); if (unlikely(!__pyx_int_414)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_415 = PyInt_FromLong(415); if (unlikely(!__pyx_int_415)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_417 = PyInt_FromLong(417); if (unlikely(!__pyx_int_417)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_418 = PyInt_FromLong(418); if (unlikely(!__pyx_int_418)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_420 = PyInt_FromLong(420); if (unlikely(!__pyx_int_420)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_421 = PyInt_FromLong(421); if (unlikely(!__pyx_int_421)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_423 = PyInt_FromLong(423); if (unlikely(!__pyx_int_423)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_425 = PyInt_FromLong(425); if (unlikely(!__pyx_int_425)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_426 = PyInt_FromLong(426); if (unlikely(!__pyx_int_426)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_427 = PyInt_FromLong(427); if (unlikely(!__pyx_int_427)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_430 = PyInt_FromLong(430); if (unlikely(!__pyx_int_430)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_432 = PyInt_FromLong(432); if (unlikely(!__pyx_int_432)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_433 = PyInt_FromLong(433); if (unlikely(!__pyx_int_433)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_434 = PyInt_FromLong(434); if (unlikely(!__pyx_int_434)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_435 = PyInt_FromLong(435); if (unlikely(!__pyx_int_435)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_437 = PyInt_FromLong(437); if (unlikely(!__pyx_int_437)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_439 = PyInt_FromLong(439); if (unlikely(!__pyx_int_439)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_440 = PyInt_FromLong(440); if (unlikely(!__pyx_int_440)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_442 = PyInt_FromLong(442); if (unlikely(!__pyx_int_442)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_443 = PyInt_FromLong(443); if (unlikely(!__pyx_int_443)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_444 = PyInt_FromLong(444); if (unlikely(!__pyx_int_444)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_446 = PyInt_FromLong(446); if (unlikely(!__pyx_int_446)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_447 = PyInt_FromLong(447); if (unlikely(!__pyx_int_447)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_449 = PyInt_FromLong(449); if (unlikely(!__pyx_int_449)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_450 = PyInt_FromLong(450); if (unlikely(!__pyx_int_450)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_451 = PyInt_FromLong(451); if (unlikely(!__pyx_int_451)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_452 = PyInt_FromLong(452); if (unlikely(!__pyx_int_452)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_454 = PyInt_FromLong(454); if (unlikely(!__pyx_int_454)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_455 = PyInt_FromLong(455); if (unlikely(!__pyx_int_455)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_456 = PyInt_FromLong(456); if (unlikely(!__pyx_int_456)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_457 = PyInt_FromLong(457); if (unlikely(!__pyx_int_457)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_458 = PyInt_FromLong(458); if (unlikely(!__pyx_int_458)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_460 = PyInt_FromLong(460); if (unlikely(!__pyx_int_460)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_461 = PyInt_FromLong(461); if (unlikely(!__pyx_int_461)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_462 = PyInt_FromLong(462); if (unlikely(!__pyx_int_462)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_464 = PyInt_FromLong(464); if (unlikely(!__pyx_int_464)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_466 = PyInt_FromLong(466); if (unlikely(!__pyx_int_466)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_467 = PyInt_FromLong(467); if (unlikely(!__pyx_int_467)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_468 = PyInt_FromLong(468); if (unlikely(!__pyx_int_468)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_471 = PyInt_FromLong(471); if (unlikely(!__pyx_int_471)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_473 = PyInt_FromLong(473); if (unlikely(!__pyx_int_473)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_474 = PyInt_FromLong(474); if (unlikely(!__pyx_int_474)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_475 = PyInt_FromLong(475); if (unlikely(!__pyx_int_475)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_476 = PyInt_FromLong(476); if (unlikely(!__pyx_int_476)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_478 = PyInt_FromLong(478); if (unlikely(!__pyx_int_478)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_479 = PyInt_FromLong(479); if (unlikely(!__pyx_int_479)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_481 = PyInt_FromLong(481); if (unlikely(!__pyx_int_481)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_482 = PyInt_FromLong(482); if (unlikely(!__pyx_int_482)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_484 = PyInt_FromLong(484); if (unlikely(!__pyx_int_484)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_485 = PyInt_FromLong(485); if (unlikely(!__pyx_int_485)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_486 = PyInt_FromLong(486); if (unlikely(!__pyx_int_486)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_488 = PyInt_FromLong(488); if (unlikely(!__pyx_int_488)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_489 = PyInt_FromLong(489); if (unlikely(!__pyx_int_489)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_490 = PyInt_FromLong(490); if (unlikely(!__pyx_int_490)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_492 = PyInt_FromLong(492); if (unlikely(!__pyx_int_492)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_494 = PyInt_FromLong(494); if (unlikely(!__pyx_int_494)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_495 = PyInt_FromLong(495); if (unlikely(!__pyx_int_495)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_496 = PyInt_FromLong(496); if (unlikely(!__pyx_int_496)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_498 = PyInt_FromLong(498); if (unlikely(!__pyx_int_498)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_499 = PyInt_FromLong(499); if (unlikely(!__pyx_int_499)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_501 = PyInt_FromLong(501); if (unlikely(!__pyx_int_501)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_502 = PyInt_FromLong(502); if (unlikely(!__pyx_int_502)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_503 = PyInt_FromLong(503); if (unlikely(!__pyx_int_503)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_505 = PyInt_FromLong(505); if (unlikely(!__pyx_int_505)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_506 = PyInt_FromLong(506); if (unlikely(!__pyx_int_506)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_508 = PyInt_FromLong(508); if (unlikely(!__pyx_int_508)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_509 = PyInt_FromLong(509); if (unlikely(!__pyx_int_509)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_510 = PyInt_FromLong(510); if (unlikely(!__pyx_int_510)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_511 = PyInt_FromLong(511); if (unlikely(!__pyx_int_511)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_513 = PyInt_FromLong(513); if (unlikely(!__pyx_int_513)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_514 = PyInt_FromLong(514); if (unlikely(!__pyx_int_514)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_516 = PyInt_FromLong(516); if (unlikely(!__pyx_int_516)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_517 = PyInt_FromLong(517); if (unlikely(!__pyx_int_517)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_518 = PyInt_FromLong(518); if (unlikely(!__pyx_int_518)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_520 = PyInt_FromLong(520); if (unlikely(!__pyx_int_520)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_521 = PyInt_FromLong(521); if (unlikely(!__pyx_int_521)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_523 = PyInt_FromLong(523); if (unlikely(!__pyx_int_523)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_524 = PyInt_FromLong(524); if (unlikely(!__pyx_int_524)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_525 = PyInt_FromLong(525); if (unlikely(!__pyx_int_525)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_526 = PyInt_FromLong(526); if (unlikely(!__pyx_int_526)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_527 = PyInt_FromLong(527); if (unlikely(!__pyx_int_527)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_528 = PyInt_FromLong(528); if (unlikely(!__pyx_int_528)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_529 = PyInt_FromLong(529); if (unlikely(!__pyx_int_529)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_531 = PyInt_FromLong(531); if (unlikely(!__pyx_int_531)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_532 = PyInt_FromLong(532); if (unlikely(!__pyx_int_532)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_534 = PyInt_FromLong(534); if (unlikely(!__pyx_int_534)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_536 = PyInt_FromLong(536); if (unlikely(!__pyx_int_536)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_537 = PyInt_FromLong(537); if (unlikely(!__pyx_int_537)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_538 = PyInt_FromLong(538); if (unlikely(!__pyx_int_538)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_539 = PyInt_FromLong(539); if (unlikely(!__pyx_int_539)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_541 = PyInt_FromLong(541); if (unlikely(!__pyx_int_541)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_542 = PyInt_FromLong(542); if (unlikely(!__pyx_int_542)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_544 = PyInt_FromLong(544); if (unlikely(!__pyx_int_544)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_545 = PyInt_FromLong(545); if (unlikely(!__pyx_int_545)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_546 = PyInt_FromLong(546); if (unlikely(!__pyx_int_546)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_548 = PyInt_FromLong(548); if (unlikely(!__pyx_int_548)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_549 = PyInt_FromLong(549); if (unlikely(!__pyx_int_549)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_551 = PyInt_FromLong(551); if (unlikely(!__pyx_int_551)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_552 = PyInt_FromLong(552); if (unlikely(!__pyx_int_552)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_553 = PyInt_FromLong(553); if (unlikely(!__pyx_int_553)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_557 = PyInt_FromLong(557); if (unlikely(!__pyx_int_557)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_559 = PyInt_FromLong(559); if (unlikely(!__pyx_int_559)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_560 = PyInt_FromLong(560); if (unlikely(!__pyx_int_560)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_561 = PyInt_FromLong(561); if (unlikely(!__pyx_int_561)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_563 = PyInt_FromLong(563); if (unlikely(!__pyx_int_563)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_564 = PyInt_FromLong(564); if (unlikely(!__pyx_int_564)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_565 = PyInt_FromLong(565); if (unlikely(!__pyx_int_565)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_569 = PyInt_FromLong(569); if (unlikely(!__pyx_int_569)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_571 = PyInt_FromLong(571); if (unlikely(!__pyx_int_571)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_572 = PyInt_FromLong(572); if (unlikely(!__pyx_int_572)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_573 = PyInt_FromLong(573); if (unlikely(!__pyx_int_573)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_574 = PyInt_FromLong(574); if (unlikely(!__pyx_int_574)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_576 = PyInt_FromLong(576); if (unlikely(!__pyx_int_576)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_577 = PyInt_FromLong(577); if (unlikely(!__pyx_int_577)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_578 = PyInt_FromLong(578); if (unlikely(!__pyx_int_578)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_579 = PyInt_FromLong(579); if (unlikely(!__pyx_int_579)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_580 = PyInt_FromLong(580); if (unlikely(!__pyx_int_580)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_581 = PyInt_FromLong(581); if (unlikely(!__pyx_int_581)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_583 = PyInt_FromLong(583); if (unlikely(!__pyx_int_583)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_585 = PyInt_FromLong(585); if (unlikely(!__pyx_int_585)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_586 = PyInt_FromLong(586); if (unlikely(!__pyx_int_586)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_588 = PyInt_FromLong(588); if (unlikely(!__pyx_int_588)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_589 = PyInt_FromLong(589); if (unlikely(!__pyx_int_589)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_590 = PyInt_FromLong(590); if (unlikely(!__pyx_int_590)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_591 = PyInt_FromLong(591); if (unlikely(!__pyx_int_591)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_592 = PyInt_FromLong(592); if (unlikely(!__pyx_int_592)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_594 = PyInt_FromLong(594); if (unlikely(!__pyx_int_594)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_595 = PyInt_FromLong(595); if (unlikely(!__pyx_int_595)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_596 = PyInt_FromLong(596); if (unlikely(!__pyx_int_596)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_597 = PyInt_FromLong(597); if (unlikely(!__pyx_int_597)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_600 = PyInt_FromLong(600); if (unlikely(!__pyx_int_600)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_602 = PyInt_FromLong(602); if (unlikely(!__pyx_int_602)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_603 = PyInt_FromLong(603); if (unlikely(!__pyx_int_603)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_604 = PyInt_FromLong(604); if (unlikely(!__pyx_int_604)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_605 = PyInt_FromLong(605); if (unlikely(!__pyx_int_605)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_606 = PyInt_FromLong(606); if (unlikely(!__pyx_int_606)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_608 = PyInt_FromLong(608); if (unlikely(!__pyx_int_608)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_609 = PyInt_FromLong(609); if (unlikely(!__pyx_int_609)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_610 = PyInt_FromLong(610); if (unlikely(!__pyx_int_610)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_611 = PyInt_FromLong(611); if (unlikely(!__pyx_int_611)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_612 = PyInt_FromLong(612); if (unlikely(!__pyx_int_612)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_613 = PyInt_FromLong(613); if (unlikely(!__pyx_int_613)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_614 = PyInt_FromLong(614); if (unlikely(!__pyx_int_614)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_616 = PyInt_FromLong(616); if (unlikely(!__pyx_int_616)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_617 = PyInt_FromLong(617); if (unlikely(!__pyx_int_617)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_619 = PyInt_FromLong(619); if (unlikely(!__pyx_int_619)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_620 = PyInt_FromLong(620); if (unlikely(!__pyx_int_620)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_621 = PyInt_FromLong(621); if (unlikely(!__pyx_int_621)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_622 = PyInt_FromLong(622); if (unlikely(!__pyx_int_622)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_623 = PyInt_FromLong(623); if (unlikely(!__pyx_int_623)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_627 = PyInt_FromLong(627); if (unlikely(!__pyx_int_627)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_629 = PyInt_FromLong(629); if (unlikely(!__pyx_int_629)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_630 = PyInt_FromLong(630); if (unlikely(!__pyx_int_630)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_631 = PyInt_FromLong(631); if (unlikely(!__pyx_int_631)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_632 = PyInt_FromLong(632); if (unlikely(!__pyx_int_632)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_633 = PyInt_FromLong(633); if (unlikely(!__pyx_int_633)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_634 = PyInt_FromLong(634); if (unlikely(!__pyx_int_634)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_638 = PyInt_FromLong(638); if (unlikely(!__pyx_int_638)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_640 = PyInt_FromLong(640); if (unlikely(!__pyx_int_640)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_643 = PyInt_FromLong(643); if (unlikely(!__pyx_int_643)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_645 = PyInt_FromLong(645); if (unlikely(!__pyx_int_645)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_646 = PyInt_FromLong(646); if (unlikely(!__pyx_int_646)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_647 = PyInt_FromLong(647); if (unlikely(!__pyx_int_647)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_648 = PyInt_FromLong(648); if (unlikely(!__pyx_int_648)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_652 = PyInt_FromLong(652); if (unlikely(!__pyx_int_652)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_654 = PyInt_FromLong(654); if (unlikely(!__pyx_int_654)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_655 = PyInt_FromLong(655); if (unlikely(!__pyx_int_655)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_657 = PyInt_FromLong(657); if (unlikely(!__pyx_int_657)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_658 = PyInt_FromLong(658); if (unlikely(!__pyx_int_658)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_660 = PyInt_FromLong(660); if (unlikely(!__pyx_int_660)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_661 = PyInt_FromLong(661); if (unlikely(!__pyx_int_661)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_662 = PyInt_FromLong(662); if (unlikely(!__pyx_int_662)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_663 = PyInt_FromLong(663); if (unlikely(!__pyx_int_663)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_664 = PyInt_FromLong(664); if (unlikely(!__pyx_int_664)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_668 = PyInt_FromLong(668); if (unlikely(!__pyx_int_668)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_670 = PyInt_FromLong(670); if (unlikely(!__pyx_int_670)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_671 = PyInt_FromLong(671); if (unlikely(!__pyx_int_671)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_672 = PyInt_FromLong(672); if (unlikely(!__pyx_int_672)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_673 = PyInt_FromLong(673); if (unlikely(!__pyx_int_673)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_674 = PyInt_FromLong(674); if (unlikely(!__pyx_int_674)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_675 = PyInt_FromLong(675); if (unlikely(!__pyx_int_675)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_679 = PyInt_FromLong(679); if (unlikely(!__pyx_int_679)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_681 = PyInt_FromLong(681); if (unlikely(!__pyx_int_681)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_684 = PyInt_FromLong(684); if (unlikely(!__pyx_int_684)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_686 = PyInt_FromLong(686); if (unlikely(!__pyx_int_686)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_687 = PyInt_FromLong(687); if (unlikely(!__pyx_int_687)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_688 = PyInt_FromLong(688); if (unlikely(!__pyx_int_688)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_689 = PyInt_FromLong(689); if (unlikely(!__pyx_int_689)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_693 = PyInt_FromLong(693); if (unlikely(!__pyx_int_693)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_695 = PyInt_FromLong(695); if (unlikely(!__pyx_int_695)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_696 = PyInt_FromLong(696); if (unlikely(!__pyx_int_696)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_698 = PyInt_FromLong(698); if (unlikely(!__pyx_int_698)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_699 = PyInt_FromLong(699); if (unlikely(!__pyx_int_699)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_701 = PyInt_FromLong(701); if (unlikely(!__pyx_int_701)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_702 = PyInt_FromLong(702); if (unlikely(!__pyx_int_702)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_703 = PyInt_FromLong(703); if (unlikely(!__pyx_int_703)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_704 = PyInt_FromLong(704); if (unlikely(!__pyx_int_704)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_705 = PyInt_FromLong(705); if (unlikely(!__pyx_int_705)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_709 = PyInt_FromLong(709); if (unlikely(!__pyx_int_709)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_711 = PyInt_FromLong(711); if (unlikely(!__pyx_int_711)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_712 = PyInt_FromLong(712); if (unlikely(!__pyx_int_712)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_713 = PyInt_FromLong(713); if (unlikely(!__pyx_int_713)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_715 = PyInt_FromLong(715); if (unlikely(!__pyx_int_715)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_716 = PyInt_FromLong(716); if (unlikely(!__pyx_int_716)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_717 = PyInt_FromLong(717); if (unlikely(!__pyx_int_717)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_718 = PyInt_FromLong(718); if (unlikely(!__pyx_int_718)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_719 = PyInt_FromLong(719); if (unlikely(!__pyx_int_719)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_723 = PyInt_FromLong(723); if (unlikely(!__pyx_int_723)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_725 = PyInt_FromLong(725); if (unlikely(!__pyx_int_725)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_726 = PyInt_FromLong(726); if (unlikely(!__pyx_int_726)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_727 = PyInt_FromLong(727); if (unlikely(!__pyx_int_727)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_729 = PyInt_FromLong(729); if (unlikely(!__pyx_int_729)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_730 = PyInt_FromLong(730); if (unlikely(!__pyx_int_730)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_731 = PyInt_FromLong(731); if (unlikely(!__pyx_int_731)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_734 = PyInt_FromLong(734); if (unlikely(!__pyx_int_734)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_736 = PyInt_FromLong(736); if (unlikely(!__pyx_int_736)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_737 = PyInt_FromLong(737); if (unlikely(!__pyx_int_737)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_738 = PyInt_FromLong(738); if (unlikely(!__pyx_int_738)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_739 = PyInt_FromLong(739); if (unlikely(!__pyx_int_739)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_740 = PyInt_FromLong(740); if (unlikely(!__pyx_int_740)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_742 = PyInt_FromLong(742); if (unlikely(!__pyx_int_742)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_743 = PyInt_FromLong(743); if (unlikely(!__pyx_int_743)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_744 = PyInt_FromLong(744); if (unlikely(!__pyx_int_744)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_745 = PyInt_FromLong(745); if (unlikely(!__pyx_int_745)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_746 = PyInt_FromLong(746); if (unlikely(!__pyx_int_746)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_750 = PyInt_FromLong(750); if (unlikely(!__pyx_int_750)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_752 = PyInt_FromLong(752); if (unlikely(!__pyx_int_752)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_753 = PyInt_FromLong(753); if (unlikely(!__pyx_int_753)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_754 = PyInt_FromLong(754); if (unlikely(!__pyx_int_754)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_756 = PyInt_FromLong(756); if (unlikely(!__pyx_int_756)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_757 = PyInt_FromLong(757); if (unlikely(!__pyx_int_757)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_758 = PyInt_FromLong(758); if (unlikely(!__pyx_int_758)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_759 = PyInt_FromLong(759); if (unlikely(!__pyx_int_759)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_760 = PyInt_FromLong(760); if (unlikely(!__pyx_int_760)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_764 = PyInt_FromLong(764); if (unlikely(!__pyx_int_764)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_766 = PyInt_FromLong(766); if (unlikely(!__pyx_int_766)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_767 = PyInt_FromLong(767); if (unlikely(!__pyx_int_767)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_768 = PyInt_FromLong(768); if (unlikely(!__pyx_int_768)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_769 = PyInt_FromLong(769); if (unlikely(!__pyx_int_769)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_770 = PyInt_FromLong(770); if (unlikely(!__pyx_int_770)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_771 = PyInt_FromLong(771); if (unlikely(!__pyx_int_771)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_775 = PyInt_FromLong(775); if (unlikely(!__pyx_int_775)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_777 = PyInt_FromLong(777); if (unlikely(!__pyx_int_777)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_780 = PyInt_FromLong(780); if (unlikely(!__pyx_int_780)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_782 = PyInt_FromLong(782); if (unlikely(!__pyx_int_782)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_783 = PyInt_FromLong(783); if (unlikely(!__pyx_int_783)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_784 = PyInt_FromLong(784); if (unlikely(!__pyx_int_784)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_786 = PyInt_FromLong(786); if (unlikely(!__pyx_int_786)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_787 = PyInt_FromLong(787); if (unlikely(!__pyx_int_787)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_788 = PyInt_FromLong(788); if (unlikely(!__pyx_int_788)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_789 = PyInt_FromLong(789); if (unlikely(!__pyx_int_789)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_790 = PyInt_FromLong(790); if (unlikely(!__pyx_int_790)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_794 = PyInt_FromLong(794); if (unlikely(!__pyx_int_794)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_796 = PyInt_FromLong(796); if (unlikely(!__pyx_int_796)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_797 = PyInt_FromLong(797); if (unlikely(!__pyx_int_797)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_798 = PyInt_FromLong(798); if (unlikely(!__pyx_int_798)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_799 = PyInt_FromLong(799); if (unlikely(!__pyx_int_799)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_800 = PyInt_FromLong(800); if (unlikely(!__pyx_int_800)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_801 = PyInt_FromLong(801); if (unlikely(!__pyx_int_801)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_805 = PyInt_FromLong(805); if (unlikely(!__pyx_int_805)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_807 = PyInt_FromLong(807); if (unlikely(!__pyx_int_807)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_810 = PyInt_FromLong(810); if (unlikely(!__pyx_int_810)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_812 = PyInt_FromLong(812); if (unlikely(!__pyx_int_812)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_813 = PyInt_FromLong(813); if (unlikely(!__pyx_int_813)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_814 = PyInt_FromLong(814); if (unlikely(!__pyx_int_814)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_816 = PyInt_FromLong(816); if (unlikely(!__pyx_int_816)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_817 = PyInt_FromLong(817); if (unlikely(!__pyx_int_817)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_818 = PyInt_FromLong(818); if (unlikely(!__pyx_int_818)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_819 = PyInt_FromLong(819); if (unlikely(!__pyx_int_819)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_820 = PyInt_FromLong(820); if (unlikely(!__pyx_int_820)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_821 = PyInt_FromLong(821); if (unlikely(!__pyx_int_821)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_823 = PyInt_FromLong(823); if (unlikely(!__pyx_int_823)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_824 = PyInt_FromLong(824); if (unlikely(!__pyx_int_824)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_825 = PyInt_FromLong(825); if (unlikely(!__pyx_int_825)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_826 = PyInt_FromLong(826); if (unlikely(!__pyx_int_826)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_827 = PyInt_FromLong(827); if (unlikely(!__pyx_int_827)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_828 = PyInt_FromLong(828); if (unlikely(!__pyx_int_828)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_829 = PyInt_FromLong(829); if (unlikely(!__pyx_int_829)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_830 = PyInt_FromLong(830); if (unlikely(!__pyx_int_830)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_831 = PyInt_FromLong(831); if (unlikely(!__pyx_int_831)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_835 = PyInt_FromLong(835); if (unlikely(!__pyx_int_835)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_837 = PyInt_FromLong(837); if (unlikely(!__pyx_int_837)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_838 = PyInt_FromLong(838); if (unlikely(!__pyx_int_838)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_839 = PyInt_FromLong(839); if (unlikely(!__pyx_int_839)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_841 = PyInt_FromLong(841); if (unlikely(!__pyx_int_841)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_843 = PyInt_FromLong(843); if (unlikely(!__pyx_int_843)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_844 = PyInt_FromLong(844); if (unlikely(!__pyx_int_844)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_845 = PyInt_FromLong(845); if (unlikely(!__pyx_int_845)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_846 = PyInt_FromLong(846); if (unlikely(!__pyx_int_846)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_847 = PyInt_FromLong(847); if (unlikely(!__pyx_int_847)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_848 = PyInt_FromLong(848); if (unlikely(!__pyx_int_848)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_850 = PyInt_FromLong(850); if (unlikely(!__pyx_int_850)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_853 = PyInt_FromLong(853); if (unlikely(!__pyx_int_853)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_854 = PyInt_FromLong(854); if (unlikely(!__pyx_int_854)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_855 = PyInt_FromLong(855); if (unlikely(!__pyx_int_855)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_856 = PyInt_FromLong(856); if (unlikely(!__pyx_int_856)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_857 = PyInt_FromLong(857); if (unlikely(!__pyx_int_857)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_859 = PyInt_FromLong(859); if (unlikely(!__pyx_int_859)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_860 = PyInt_FromLong(860); if (unlikely(!__pyx_int_860)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_861 = PyInt_FromLong(861); if (unlikely(!__pyx_int_861)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_862 = PyInt_FromLong(862); if (unlikely(!__pyx_int_862)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_864 = PyInt_FromLong(864); if (unlikely(!__pyx_int_864)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_865 = PyInt_FromLong(865); if (unlikely(!__pyx_int_865)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_866 = PyInt_FromLong(866); if (unlikely(!__pyx_int_866)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_867 = PyInt_FromLong(867); if (unlikely(!__pyx_int_867)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_868 = PyInt_FromLong(868); if (unlikely(!__pyx_int_868)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_869 = PyInt_FromLong(869); if (unlikely(!__pyx_int_869)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_870 = PyInt_FromLong(870); if (unlikely(!__pyx_int_870)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_871 = PyInt_FromLong(871); if (unlikely(!__pyx_int_871)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_872 = PyInt_FromLong(872); if (unlikely(!__pyx_int_872)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_873 = PyInt_FromLong(873); if (unlikely(!__pyx_int_873)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_874 = PyInt_FromLong(874); if (unlikely(!__pyx_int_874)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_875 = PyInt_FromLong(875); if (unlikely(!__pyx_int_875)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_876 = PyInt_FromLong(876); if (unlikely(!__pyx_int_876)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_877 = PyInt_FromLong(877); if (unlikely(!__pyx_int_877)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_878 = PyInt_FromLong(878); if (unlikely(!__pyx_int_878)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_879 = PyInt_FromLong(879); if (unlikely(!__pyx_int_879)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_881 = PyInt_FromLong(881); if (unlikely(!__pyx_int_881)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_883 = PyInt_FromLong(883); if (unlikely(!__pyx_int_883)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_884 = PyInt_FromLong(884); if (unlikely(!__pyx_int_884)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_885 = PyInt_FromLong(885); if (unlikely(!__pyx_int_885)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_886 = PyInt_FromLong(886); if (unlikely(!__pyx_int_886)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_887 = PyInt_FromLong(887); if (unlikely(!__pyx_int_887)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_888 = PyInt_FromLong(888); if (unlikely(!__pyx_int_888)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_889 = PyInt_FromLong(889); if (unlikely(!__pyx_int_889)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_891 = PyInt_FromLong(891); if (unlikely(!__pyx_int_891)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_892 = PyInt_FromLong(892); if (unlikely(!__pyx_int_892)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_893 = PyInt_FromLong(893); if (unlikely(!__pyx_int_893)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_894 = PyInt_FromLong(894); if (unlikely(!__pyx_int_894)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_895 = PyInt_FromLong(895); if (unlikely(!__pyx_int_895)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_896 = PyInt_FromLong(896); if (unlikely(!__pyx_int_896)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_897 = PyInt_FromLong(897); if (unlikely(!__pyx_int_897)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_898 = PyInt_FromLong(898); if (unlikely(!__pyx_int_898)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_900 = PyInt_FromLong(900); if (unlikely(!__pyx_int_900)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_901 = PyInt_FromLong(901); if (unlikely(!__pyx_int_901)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_902 = PyInt_FromLong(902); if (unlikely(!__pyx_int_902)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_903 = PyInt_FromLong(903); if (unlikely(!__pyx_int_903)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_906 = PyInt_FromLong(906); if (unlikely(!__pyx_int_906)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_908 = PyInt_FromLong(908); if (unlikely(!__pyx_int_908)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_909 = PyInt_FromLong(909); if (unlikely(!__pyx_int_909)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_911 = PyInt_FromLong(911); if (unlikely(!__pyx_int_911)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_912 = PyInt_FromLong(912); if (unlikely(!__pyx_int_912)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_913 = PyInt_FromLong(913); if (unlikely(!__pyx_int_913)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_915 = PyInt_FromLong(915); if (unlikely(!__pyx_int_915)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_916 = PyInt_FromLong(916); if (unlikely(!__pyx_int_916)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_918 = PyInt_FromLong(918); if (unlikely(!__pyx_int_918)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_919 = PyInt_FromLong(919); if (unlikely(!__pyx_int_919)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_921 = PyInt_FromLong(921); if (unlikely(!__pyx_int_921)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_922 = PyInt_FromLong(922); if (unlikely(!__pyx_int_922)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_923 = PyInt_FromLong(923); if (unlikely(!__pyx_int_923)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_926 = PyInt_FromLong(926); if (unlikely(!__pyx_int_926)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_928 = PyInt_FromLong(928); if (unlikely(!__pyx_int_928)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_929 = PyInt_FromLong(929); if (unlikely(!__pyx_int_929)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_930 = PyInt_FromLong(930); if (unlikely(!__pyx_int_930)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_931 = PyInt_FromLong(931); if (unlikely(!__pyx_int_931)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_933 = PyInt_FromLong(933); if (unlikely(!__pyx_int_933)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_934 = PyInt_FromLong(934); if (unlikely(!__pyx_int_934)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_935 = PyInt_FromLong(935); if (unlikely(!__pyx_int_935)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_938 = PyInt_FromLong(938); if (unlikely(!__pyx_int_938)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_940 = PyInt_FromLong(940); if (unlikely(!__pyx_int_940)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_941 = PyInt_FromLong(941); if (unlikely(!__pyx_int_941)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_942 = PyInt_FromLong(942); if (unlikely(!__pyx_int_942)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_943 = PyInt_FromLong(943); if (unlikely(!__pyx_int_943)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_945 = PyInt_FromLong(945); if (unlikely(!__pyx_int_945)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_947 = PyInt_FromLong(947); if (unlikely(!__pyx_int_947)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_948 = PyInt_FromLong(948); if (unlikely(!__pyx_int_948)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_949 = PyInt_FromLong(949); if (unlikely(!__pyx_int_949)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_950 = PyInt_FromLong(950); if (unlikely(!__pyx_int_950)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_951 = PyInt_FromLong(951); if (unlikely(!__pyx_int_951)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_952 = PyInt_FromLong(952); if (unlikely(!__pyx_int_952)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_954 = PyInt_FromLong(954); if (unlikely(!__pyx_int_954)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_956 = PyInt_FromLong(956); if (unlikely(!__pyx_int_956)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_957 = PyInt_FromLong(957); if (unlikely(!__pyx_int_957)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_958 = PyInt_FromLong(958); if (unlikely(!__pyx_int_958)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_960 = PyInt_FromLong(960); if (unlikely(!__pyx_int_960)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_961 = PyInt_FromLong(961); if (unlikely(!__pyx_int_961)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_962 = PyInt_FromLong(962); if (unlikely(!__pyx_int_962)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_965 = PyInt_FromLong(965); if (unlikely(!__pyx_int_965)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_967 = PyInt_FromLong(967); if (unlikely(!__pyx_int_967)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_968 = PyInt_FromLong(968); if (unlikely(!__pyx_int_968)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_969 = PyInt_FromLong(969); if (unlikely(!__pyx_int_969)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_970 = PyInt_FromLong(970); if (unlikely(!__pyx_int_970)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_972 = PyInt_FromLong(972); if (unlikely(!__pyx_int_972)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_973 = PyInt_FromLong(973); if (unlikely(!__pyx_int_973)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_974 = PyInt_FromLong(974); if (unlikely(!__pyx_int_974)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_975 = PyInt_FromLong(975); if (unlikely(!__pyx_int_975)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_976 = PyInt_FromLong(976); if (unlikely(!__pyx_int_976)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_978 = PyInt_FromLong(978); if (unlikely(!__pyx_int_978)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_980 = PyInt_FromLong(980); if (unlikely(!__pyx_int_980)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_981 = PyInt_FromLong(981); if (unlikely(!__pyx_int_981)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_983 = PyInt_FromLong(983); if (unlikely(!__pyx_int_983)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_984 = PyInt_FromLong(984); if (unlikely(!__pyx_int_984)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_986 = PyInt_FromLong(986); if (unlikely(!__pyx_int_986)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_987 = PyInt_FromLong(987); if (unlikely(!__pyx_int_987)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_990 = PyInt_FromLong(990); if (unlikely(!__pyx_int_990)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_992 = PyInt_FromLong(992); if (unlikely(!__pyx_int_992)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_993 = PyInt_FromLong(993); if (unlikely(!__pyx_int_993)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_994 = PyInt_FromLong(994); if (unlikely(!__pyx_int_994)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_995 = PyInt_FromLong(995); if (unlikely(!__pyx_int_995)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_998 = PyInt_FromLong(998); if (unlikely(!__pyx_int_998)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1000 = PyInt_FromLong(1000); if (unlikely(!__pyx_int_1000)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1001 = PyInt_FromLong(1001); if (unlikely(!__pyx_int_1001)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1002 = PyInt_FromLong(1002); if (unlikely(!__pyx_int_1002)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1003 = PyInt_FromLong(1003); if (unlikely(!__pyx_int_1003)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1004 = PyInt_FromLong(1004); if (unlikely(!__pyx_int_1004)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1007 = PyInt_FromLong(1007); if (unlikely(!__pyx_int_1007)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1009 = PyInt_FromLong(1009); if (unlikely(!__pyx_int_1009)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1010 = PyInt_FromLong(1010); if (unlikely(!__pyx_int_1010)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1012 = PyInt_FromLong(1012); if (unlikely(!__pyx_int_1012)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1013 = PyInt_FromLong(1013); if (unlikely(!__pyx_int_1013)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1014 = PyInt_FromLong(1014); if (unlikely(!__pyx_int_1014)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1016 = PyInt_FromLong(1016); if (unlikely(!__pyx_int_1016)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1017 = PyInt_FromLong(1017); if (unlikely(!__pyx_int_1017)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1018 = PyInt_FromLong(1018); if (unlikely(!__pyx_int_1018)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1020 = PyInt_FromLong(1020); if (unlikely(!__pyx_int_1020)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1021 = PyInt_FromLong(1021); if (unlikely(!__pyx_int_1021)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1022 = PyInt_FromLong(1022); if (unlikely(!__pyx_int_1022)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1025 = PyInt_FromLong(1025); if (unlikely(!__pyx_int_1025)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1027 = PyInt_FromLong(1027); if (unlikely(!__pyx_int_1027)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1028 = PyInt_FromLong(1028); if (unlikely(!__pyx_int_1028)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1030 = PyInt_FromLong(1030); if (unlikely(!__pyx_int_1030)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1031 = PyInt_FromLong(1031); if (unlikely(!__pyx_int_1031)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC initModelicaParser(void); /*proto*/
PyMODINIT_FUNC initModelicaParser(void)
#else
PyMODINIT_FUNC PyInit_ModelicaParser(void); /*proto*/
PyMODINIT_FUNC PyInit_ModelicaParser(void)
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  Py_ssize_t __pyx_t_10;
  PyObject *(*__pyx_t_11)(PyObject *);
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  PyObject *__pyx_t_25 = NULL;
  PyObject *__pyx_t_26 = NULL;
  PyObject *__pyx_t_27 = NULL;
  PyObject *__pyx_t_28 = NULL;
  PyObject *__pyx_t_29 = NULL;
  PyObject *__pyx_t_30 = NULL;
  PyObject *__pyx_t_31 = NULL;
  PyObject *__pyx_t_32 = NULL;
  PyObject *__pyx_t_33 = NULL;
  PyObject *__pyx_t_34 = NULL;
  PyObject *__pyx_t_35 = NULL;
  PyObject *__pyx_t_36 = NULL;
  PyObject *__pyx_t_37 = NULL;
  PyObject *__pyx_t_38 = NULL;
  PyObject *__pyx_t_39 = NULL;
  PyObject *__pyx_t_40 = NULL;
  PyObject *__pyx_t_41 = NULL;
  PyObject *__pyx_t_42 = NULL;
  PyObject *__pyx_t_43 = NULL;
  PyObject *__pyx_t_44 = NULL;
  PyObject *__pyx_t_45 = NULL;
  PyObject *__pyx_t_46 = NULL;
  PyObject *__pyx_t_47 = NULL;
  PyObject *__pyx_t_48 = NULL;
  PyObject *__pyx_t_49 = NULL;
  PyObject *__pyx_t_50 = NULL;
  PyObject *__pyx_t_51 = NULL;
  PyObject *__pyx_t_52 = NULL;
  PyObject *__pyx_t_53 = NULL;
  PyObject *__pyx_t_54 = NULL;
  PyObject *__pyx_t_55 = NULL;
  PyObject *__pyx_t_56 = NULL;
  PyObject *__pyx_t_57 = NULL;
  PyObject *__pyx_t_58 = NULL;
  PyObject *__pyx_t_59 = NULL;
  PyObject *__pyx_t_60 = NULL;
  PyObject *__pyx_t_61 = NULL;
  PyObject *__pyx_t_62 = NULL;
  PyObject *__pyx_t_63 = NULL;
  PyObject *__pyx_t_64 = NULL;
  PyObject *__pyx_t_65 = NULL;
  PyObject *__pyx_t_66 = NULL;
  PyObject *__pyx_t_67 = NULL;
  PyObject *__pyx_t_68 = NULL;
  PyObject *__pyx_t_69 = NULL;
  PyObject *__pyx_t_70 = NULL;
  PyObject *__pyx_t_71 = NULL;
  PyObject *__pyx_t_72 = NULL;
  PyObject *__pyx_t_73 = NULL;
  PyObject *__pyx_t_74 = NULL;
  PyObject *__pyx_t_75 = NULL;
  PyObject *__pyx_t_76 = NULL;
  PyObject *__pyx_t_77 = NULL;
  PyObject *__pyx_t_78 = NULL;
  PyObject *__pyx_t_79 = NULL;
  PyObject *__pyx_t_80 = NULL;
  PyObject *__pyx_t_81 = NULL;
  PyObject *__pyx_t_82 = NULL;
  PyObject *__pyx_t_83 = NULL;
  PyObject *__pyx_t_84 = NULL;
  PyObject *__pyx_t_85 = NULL;
  PyObject *__pyx_t_86 = NULL;
  PyObject *__pyx_t_87 = NULL;
  PyObject *__pyx_t_88 = NULL;
  PyObject *__pyx_t_89 = NULL;
  PyObject *__pyx_t_90 = NULL;
  PyObject *__pyx_t_91 = NULL;
  PyObject *__pyx_t_92 = NULL;
  PyObject *__pyx_t_93 = NULL;
  PyObject *__pyx_t_94 = NULL;
  PyObject *__pyx_t_95 = NULL;
  PyObject *__pyx_t_96 = NULL;
  PyObject *__pyx_t_97 = NULL;
  PyObject *__pyx_t_98 = NULL;
  PyObject *__pyx_t_99 = NULL;
  PyObject *__pyx_t_100 = NULL;
  PyObject *__pyx_t_101 = NULL;
  PyObject *__pyx_t_102 = NULL;
  PyObject *__pyx_t_103 = NULL;
  PyObject *__pyx_t_104 = NULL;
  PyObject *__pyx_t_105 = NULL;
  PyObject *__pyx_t_106 = NULL;
  PyObject *__pyx_t_107 = NULL;
  PyObject *__pyx_t_108 = NULL;
  PyObject *__pyx_t_109 = NULL;
  PyObject *__pyx_t_110 = NULL;
  PyObject *__pyx_t_111 = NULL;
  PyObject *__pyx_t_112 = NULL;
  PyObject *__pyx_t_113 = NULL;
  PyObject *__pyx_t_114 = NULL;
  PyObject *__pyx_t_115 = NULL;
  PyObject *__pyx_t_116 = NULL;
  PyObject *__pyx_t_117 = NULL;
  PyObject *__pyx_t_118 = NULL;
  PyObject *__pyx_t_119 = NULL;
  PyObject *__pyx_t_120 = NULL;
  PyObject *__pyx_t_121 = NULL;
  PyObject *__pyx_t_122 = NULL;
  PyObject *__pyx_t_123 = NULL;
  PyObject *__pyx_t_124 = NULL;
  PyObject *__pyx_t_125 = NULL;
  PyObject *__pyx_t_126 = NULL;
  PyObject *__pyx_t_127 = NULL;
  PyObject *__pyx_t_128 = NULL;
  PyObject *__pyx_t_129 = NULL;
  PyObject *__pyx_t_130 = NULL;
  PyObject *__pyx_t_131 = NULL;
  PyObject *__pyx_t_132 = NULL;
  __Pyx_RefNannyDeclarations
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
      PyErr_Clear();
      __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
      if (!__Pyx_RefNanny)
          Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  __Pyx_RefNannySetupContext("PyMODINIT_FUNC PyInit_ModelicaParser(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("ModelicaParser", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_pymola__generated__ModelicaParser) {
    if (PyObject_SetAttrString(__pyx_m, "__name__", __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "pymola.generated.ModelicaParser")) {
      if (unlikely(PyDict_SetItemString(modules, "pymola.generated.ModelicaParser", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global init code ---*/
  /*--- Variable export code ---*/
  /*--- Function export code ---*/
  /*--- Type init code ---*/
  /*--- Type import code ---*/
  /*--- Variable import code ---*/
  /*--- Function import code ---*/
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "pymola/generated/ModelicaParser.pyx":3
 * # Generated from Modelica.g4 by ANTLR 4.7
 * # encoding: utf-8
 * from antlr4 import *             # <<<<<<<<<<<<<<
 * from io import StringIO
 * from typing.io import TextIO
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s__975);
  __Pyx_GIVEREF(__pyx_n_s__975);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s__975);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_antlr4, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_import_star(__pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4
 * # encoding: utf-8
 * from antlr4 import *
 * from io import StringIO             # <<<<<<<<<<<<<<
 * from typing.io import TextIO
 * import sys
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_StringIO);
  __Pyx_GIVEREF(__pyx_n_s_StringIO);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_StringIO);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_io, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_StringIO); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StringIO, __pyx_t_2) < 0) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5
 * from antlr4 import *
 * from io import StringIO
 * from typing.io import TextIO             # <<<<<<<<<<<<<<
 * import sys
 * 
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_TextIO);
  __Pyx_GIVEREF(__pyx_n_s_TextIO);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_TextIO);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_typing_io, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_TextIO); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TextIO, __pyx_t_1) < 0) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6
 * from io import StringIO
 * from typing.io import TextIO
 * import sys             # <<<<<<<<<<<<<<
 * 
 * def serializedATN():
 */
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_sys, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_sys, __pyx_t_2) < 0) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pymola/generated/ModelicaParser.pyx":8
 * import sys
 * 
 * def serializedATN():             # <<<<<<<<<<<<<<
 *     with StringIO() as buf:
 *         buf.write("\3\u608b\ua72a\u8133\ub9ed\u417c\u3be7\u7786\u5964\3^")
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_1serializedATN, NULL, __pyx_n_s_pymola_generated_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_serializedATN, __pyx_t_2) < 0) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pymola/generated/ModelicaParser.pyx":532
 * 
 * 
 * class ModelicaParser ( Parser ):             # <<<<<<<<<<<<<<
 * 
 *     grammarFileName = "Modelica.g4"
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_Parser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_1, __pyx_n_s_ModelicaParser, __pyx_n_s_ModelicaParser, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "pymola/generated/ModelicaParser.pyx":534
 * class ModelicaParser ( Parser ):
 * 
 *     grammarFileName = "Modelica.g4"             # <<<<<<<<<<<<<<
 * 
 *     atn = ATNDeserializer().deserialize(serializedATN())
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_grammarFileName, __pyx_kp_u_Modelica_g4) < 0) __PYX_ERR(0, 534, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":536
 *     grammarFileName = "Modelica.g4"
 * 
 *     atn = ATNDeserializer().deserialize(serializedATN())             # <<<<<<<<<<<<<<
 * 
 *     decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]
 */
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ATNDeserializer); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  if (__pyx_t_8) {
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_8); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 536, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  } else {
    __pyx_t_6 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 536, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_deserialize); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_serializedATN); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
    }
  }
  if (__pyx_t_9) {
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 536, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  } else {
    __pyx_t_6 = __Pyx_PyObject_CallNoArg(__pyx_t_8); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 536, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  if (!__pyx_t_8) {
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 536, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GOTREF(__pyx_t_5);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_7)) {
      PyObject *__pyx_temp[2] = {__pyx_t_8, __pyx_t_6};
      __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 536, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
      PyObject *__pyx_temp[2] = {__pyx_t_8, __pyx_t_6};
      __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 536, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_9 = PyTuple_New(1+1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 536, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_8); __pyx_t_8 = NULL;
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_9, 0+1, __pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_9, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 536, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_atn, __pyx_t_5) < 0) __PYX_ERR(0, 536, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":538
 *     atn = ATNDeserializer().deserialize(serializedATN())
 * 
 *     decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]             # <<<<<<<<<<<<<<
 * 
 *     sharedContextCache = PredictionContextCache()
 */
  { /* enter inner scope */
    PyObject *__pyx_7genexpr__pyx_v_6pymola_9generated_14ModelicaParser_14ModelicaParser_i = NULL;
    PyObject *__pyx_7genexpr__pyx_v_6pymola_9generated_14ModelicaParser_14ModelicaParser_ds = NULL;
    __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 538, __pyx_L4_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_int_0);
    __pyx_t_7 = __pyx_int_0;
    __pyx_t_9 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_atn);
    if (unlikely(!__pyx_t_9)) {
      PyErr_Clear();
      __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_atn);
    }
    if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 538, __pyx_L4_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_decisionToState); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 538, __pyx_L4_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (likely(PyList_CheckExact(__pyx_t_6)) || PyTuple_CheckExact(__pyx_t_6)) {
      __pyx_t_9 = __pyx_t_6; __Pyx_INCREF(__pyx_t_9); __pyx_t_10 = 0;
      __pyx_t_11 = NULL;
    } else {
      __pyx_t_10 = -1; __pyx_t_9 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 538, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_11 = Py_TYPE(__pyx_t_9)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 538, __pyx_L4_error)
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    for (;;) {
      if (likely(!__pyx_t_11)) {
        if (likely(PyList_CheckExact(__pyx_t_9))) {
          if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_9)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyList_GET_ITEM(__pyx_t_9, __pyx_t_10); __Pyx_INCREF(__pyx_t_6); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 538, __pyx_L4_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_9, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 538, __pyx_L4_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        } else {
          if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_9)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_9, __pyx_t_10); __Pyx_INCREF(__pyx_t_6); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 538, __pyx_L4_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_9, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 538, __pyx_L4_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        }
      } else {
        __pyx_t_6 = __pyx_t_11(__pyx_t_9);
        if (unlikely(!__pyx_t_6)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 538, __pyx_L4_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_6);
      }
      __Pyx_XDECREF_SET(__pyx_7genexpr__pyx_v_6pymola_9generated_14ModelicaParser_14ModelicaParser_ds, __pyx_t_6);
      __pyx_t_6 = 0;
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_XDECREF_SET(__pyx_7genexpr__pyx_v_6pymola_9generated_14ModelicaParser_14ModelicaParser_i, __pyx_t_7);
      __pyx_t_6 = __Pyx_PyInt_AddObjC(__pyx_t_7, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 538, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_7);
      __pyx_t_7 = __pyx_t_6;
      __pyx_t_6 = 0;
      __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_DFA); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 538, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_12 = NULL;
      __pyx_t_13 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_12)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_12);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
          __pyx_t_13 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_8)) {
        PyObject *__pyx_temp[3] = {__pyx_t_12, __pyx_7genexpr__pyx_v_6pymola_9generated_14ModelicaParser_14ModelicaParser_ds, __pyx_7genexpr__pyx_v_6pymola_9generated_14ModelicaParser_14ModelicaParser_i};
        __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_13, 2+__pyx_t_13); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 538, __pyx_L4_error)
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_GOTREF(__pyx_t_6);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
        PyObject *__pyx_temp[3] = {__pyx_t_12, __pyx_7genexpr__pyx_v_6pymola_9generated_14ModelicaParser_14ModelicaParser_ds, __pyx_7genexpr__pyx_v_6pymola_9generated_14ModelicaParser_14ModelicaParser_i};
        __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_13, 2+__pyx_t_13); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 538, __pyx_L4_error)
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_GOTREF(__pyx_t_6);
      } else
      #endif
      {
        __pyx_t_14 = PyTuple_New(2+__pyx_t_13); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 538, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_14);
        if (__pyx_t_12) {
          __Pyx_GIVEREF(__pyx_t_12); PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_12); __pyx_t_12 = NULL;
        }
        __Pyx_INCREF(__pyx_7genexpr__pyx_v_6pymola_9generated_14ModelicaParser_14ModelicaParser_ds);
        __Pyx_GIVEREF(__pyx_7genexpr__pyx_v_6pymola_9generated_14ModelicaParser_14ModelicaParser_ds);
        PyTuple_SET_ITEM(__pyx_t_14, 0+__pyx_t_13, __pyx_7genexpr__pyx_v_6pymola_9generated_14ModelicaParser_14ModelicaParser_ds);
        __Pyx_INCREF(__pyx_7genexpr__pyx_v_6pymola_9generated_14ModelicaParser_14ModelicaParser_i);
        __Pyx_GIVEREF(__pyx_7genexpr__pyx_v_6pymola_9generated_14ModelicaParser_14ModelicaParser_i);
        PyTuple_SET_ITEM(__pyx_t_14, 1+__pyx_t_13, __pyx_7genexpr__pyx_v_6pymola_9generated_14ModelicaParser_14ModelicaParser_i);
        __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_14, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 538, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      }
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_5, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 538, __pyx_L4_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_7genexpr__pyx_v_6pymola_9generated_14ModelicaParser_14ModelicaParser_i);
    __Pyx_XDECREF(__pyx_7genexpr__pyx_v_6pymola_9generated_14ModelicaParser_14ModelicaParser_ds);
    goto __pyx_L7_exit_scope;
    __pyx_L4_error:;
    __Pyx_XDECREF(__pyx_7genexpr__pyx_v_6pymola_9generated_14ModelicaParser_14ModelicaParser_i);
    __Pyx_XDECREF(__pyx_7genexpr__pyx_v_6pymola_9generated_14ModelicaParser_14ModelicaParser_ds);
    goto __pyx_L1_error;
    __pyx_L7_exit_scope:;
  } /* exit inner scope */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_decisionsToDFA, __pyx_t_5) < 0) __PYX_ERR(0, 538, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":540
 *     decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]
 * 
 *     sharedContextCache = PredictionContextCache()             # <<<<<<<<<<<<<<
 * 
 *     literalNames = [ "<INVALID>", "';'", "'class'", "'model'", "'operator'",
 */
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_PredictionContextCache); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  if (__pyx_t_9) {
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 540, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  } else {
    __pyx_t_5 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 540, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_sharedContextCache, __pyx_t_5) < 0) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":542
 *     sharedContextCache = PredictionContextCache()
 * 
 *     literalNames = [ "<INVALID>", "';'", "'class'", "'model'", "'operator'",             # <<<<<<<<<<<<<<
 *                      "'record'", "'block'", "'expandable'", "'connector'",
 *                      "'type'", "'package'", "'pure'", "'impure'", "'function'",
 */
  __pyx_t_5 = PyList_New(88); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 0, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u__978);
  __Pyx_GIVEREF(__pyx_kp_u__978);
  PyList_SET_ITEM(__pyx_t_5, 1, __pyx_kp_u__978);
  __Pyx_INCREF(__pyx_kp_u_class);
  __Pyx_GIVEREF(__pyx_kp_u_class);
  PyList_SET_ITEM(__pyx_t_5, 2, __pyx_kp_u_class);
  __Pyx_INCREF(__pyx_kp_u_model);
  __Pyx_GIVEREF(__pyx_kp_u_model);
  PyList_SET_ITEM(__pyx_t_5, 3, __pyx_kp_u_model);
  __Pyx_INCREF(__pyx_kp_u_operator);
  __Pyx_GIVEREF(__pyx_kp_u_operator);
  PyList_SET_ITEM(__pyx_t_5, 4, __pyx_kp_u_operator);
  __Pyx_INCREF(__pyx_kp_u_record);
  __Pyx_GIVEREF(__pyx_kp_u_record);
  PyList_SET_ITEM(__pyx_t_5, 5, __pyx_kp_u_record);
  __Pyx_INCREF(__pyx_kp_u_block);
  __Pyx_GIVEREF(__pyx_kp_u_block);
  PyList_SET_ITEM(__pyx_t_5, 6, __pyx_kp_u_block);
  __Pyx_INCREF(__pyx_kp_u_expandable);
  __Pyx_GIVEREF(__pyx_kp_u_expandable);
  PyList_SET_ITEM(__pyx_t_5, 7, __pyx_kp_u_expandable);
  __Pyx_INCREF(__pyx_kp_u_connector);
  __Pyx_GIVEREF(__pyx_kp_u_connector);
  PyList_SET_ITEM(__pyx_t_5, 8, __pyx_kp_u_connector);
  __Pyx_INCREF(__pyx_kp_u_type);
  __Pyx_GIVEREF(__pyx_kp_u_type);
  PyList_SET_ITEM(__pyx_t_5, 9, __pyx_kp_u_type);
  __Pyx_INCREF(__pyx_kp_u_package);
  __Pyx_GIVEREF(__pyx_kp_u_package);
  PyList_SET_ITEM(__pyx_t_5, 10, __pyx_kp_u_package);
  __Pyx_INCREF(__pyx_kp_u_pure);
  __Pyx_GIVEREF(__pyx_kp_u_pure);
  PyList_SET_ITEM(__pyx_t_5, 11, __pyx_kp_u_pure);
  __Pyx_INCREF(__pyx_kp_u_impure);
  __Pyx_GIVEREF(__pyx_kp_u_impure);
  PyList_SET_ITEM(__pyx_t_5, 12, __pyx_kp_u_impure);
  __Pyx_INCREF(__pyx_kp_u_function);
  __Pyx_GIVEREF(__pyx_kp_u_function);
  PyList_SET_ITEM(__pyx_t_5, 13, __pyx_kp_u_function);
  __Pyx_INCREF(__pyx_kp_u_end);
  __Pyx_GIVEREF(__pyx_kp_u_end);
  PyList_SET_ITEM(__pyx_t_5, 14, __pyx_kp_u_end);
  __Pyx_INCREF(__pyx_kp_u__979);
  __Pyx_GIVEREF(__pyx_kp_u__979);
  PyList_SET_ITEM(__pyx_t_5, 15, __pyx_kp_u__979);
  __Pyx_INCREF(__pyx_kp_u_enumeration);
  __Pyx_GIVEREF(__pyx_kp_u_enumeration);
  PyList_SET_ITEM(__pyx_t_5, 16, __pyx_kp_u_enumeration);
  __Pyx_INCREF(__pyx_kp_u__980);
  __Pyx_GIVEREF(__pyx_kp_u__980);
  PyList_SET_ITEM(__pyx_t_5, 17, __pyx_kp_u__980);
  __Pyx_INCREF(__pyx_kp_u__981);
  __Pyx_GIVEREF(__pyx_kp_u__981);
  PyList_SET_ITEM(__pyx_t_5, 18, __pyx_kp_u__981);
  __Pyx_INCREF(__pyx_kp_u__982);
  __Pyx_GIVEREF(__pyx_kp_u__982);
  PyList_SET_ITEM(__pyx_t_5, 19, __pyx_kp_u__982);
  __Pyx_INCREF(__pyx_kp_u_der);
  __Pyx_GIVEREF(__pyx_kp_u_der);
  PyList_SET_ITEM(__pyx_t_5, 20, __pyx_kp_u_der);
  __Pyx_INCREF(__pyx_kp_u__983);
  __Pyx_GIVEREF(__pyx_kp_u__983);
  PyList_SET_ITEM(__pyx_t_5, 21, __pyx_kp_u__983);
  __Pyx_INCREF(__pyx_kp_u_extends);
  __Pyx_GIVEREF(__pyx_kp_u_extends);
  PyList_SET_ITEM(__pyx_t_5, 22, __pyx_kp_u_extends);
  __Pyx_INCREF(__pyx_kp_u_public);
  __Pyx_GIVEREF(__pyx_kp_u_public);
  PyList_SET_ITEM(__pyx_t_5, 23, __pyx_kp_u_public);
  __Pyx_INCREF(__pyx_kp_u_protected);
  __Pyx_GIVEREF(__pyx_kp_u_protected);
  PyList_SET_ITEM(__pyx_t_5, 24, __pyx_kp_u_protected);
  __Pyx_INCREF(__pyx_kp_u_external);
  __Pyx_GIVEREF(__pyx_kp_u_external);
  PyList_SET_ITEM(__pyx_t_5, 25, __pyx_kp_u_external);
  __Pyx_INCREF(__pyx_kp_u_replaceable);
  __Pyx_GIVEREF(__pyx_kp_u_replaceable);
  PyList_SET_ITEM(__pyx_t_5, 26, __pyx_kp_u_replaceable);
  __Pyx_INCREF(__pyx_kp_u_import_2);
  __Pyx_GIVEREF(__pyx_kp_u_import_2);
  PyList_SET_ITEM(__pyx_t_5, 27, __pyx_kp_u_import_2);
  __Pyx_INCREF(__pyx_kp_u__984);
  __Pyx_GIVEREF(__pyx_kp_u__984);
  PyList_SET_ITEM(__pyx_t_5, 28, __pyx_kp_u__984);
  __Pyx_INCREF(__pyx_kp_u__985);
  __Pyx_GIVEREF(__pyx_kp_u__985);
  PyList_SET_ITEM(__pyx_t_5, 29, __pyx_kp_u__985);
  __Pyx_INCREF(__pyx_kp_u__986);
  __Pyx_GIVEREF(__pyx_kp_u__986);
  PyList_SET_ITEM(__pyx_t_5, 30, __pyx_kp_u__986);
  __Pyx_INCREF(__pyx_kp_u__987);
  __Pyx_GIVEREF(__pyx_kp_u__987);
  PyList_SET_ITEM(__pyx_t_5, 31, __pyx_kp_u__987);
  __Pyx_INCREF(__pyx_kp_u_constrainedby);
  __Pyx_GIVEREF(__pyx_kp_u_constrainedby);
  PyList_SET_ITEM(__pyx_t_5, 32, __pyx_kp_u_constrainedby);
  __Pyx_INCREF(__pyx_kp_u_flow);
  __Pyx_GIVEREF(__pyx_kp_u_flow);
  PyList_SET_ITEM(__pyx_t_5, 33, __pyx_kp_u_flow);
  __Pyx_INCREF(__pyx_kp_u_stream);
  __Pyx_GIVEREF(__pyx_kp_u_stream);
  PyList_SET_ITEM(__pyx_t_5, 34, __pyx_kp_u_stream);
  __Pyx_INCREF(__pyx_kp_u_discrete);
  __Pyx_GIVEREF(__pyx_kp_u_discrete);
  PyList_SET_ITEM(__pyx_t_5, 35, __pyx_kp_u_discrete);
  __Pyx_INCREF(__pyx_kp_u_parameter);
  __Pyx_GIVEREF(__pyx_kp_u_parameter);
  PyList_SET_ITEM(__pyx_t_5, 36, __pyx_kp_u_parameter);
  __Pyx_INCREF(__pyx_kp_u_constant);
  __Pyx_GIVEREF(__pyx_kp_u_constant);
  PyList_SET_ITEM(__pyx_t_5, 37, __pyx_kp_u_constant);
  __Pyx_INCREF(__pyx_kp_u_input_3);
  __Pyx_GIVEREF(__pyx_kp_u_input_3);
  PyList_SET_ITEM(__pyx_t_5, 38, __pyx_kp_u_input_3);
  __Pyx_INCREF(__pyx_kp_u_output_2);
  __Pyx_GIVEREF(__pyx_kp_u_output_2);
  PyList_SET_ITEM(__pyx_t_5, 39, __pyx_kp_u_output_2);
  __Pyx_INCREF(__pyx_kp_u_if);
  __Pyx_GIVEREF(__pyx_kp_u_if);
  PyList_SET_ITEM(__pyx_t_5, 40, __pyx_kp_u_if);
  __Pyx_INCREF(__pyx_kp_u__988);
  __Pyx_GIVEREF(__pyx_kp_u__988);
  PyList_SET_ITEM(__pyx_t_5, 41, __pyx_kp_u__988);
  __Pyx_INCREF(__pyx_kp_u_equation_2);
  __Pyx_GIVEREF(__pyx_kp_u_equation_2);
  PyList_SET_ITEM(__pyx_t_5, 42, __pyx_kp_u_equation_2);
  __Pyx_INCREF(__pyx_kp_u_algorithm);
  __Pyx_GIVEREF(__pyx_kp_u_algorithm);
  PyList_SET_ITEM(__pyx_t_5, 43, __pyx_kp_u_algorithm);
  __Pyx_INCREF(__pyx_kp_u_break);
  __Pyx_GIVEREF(__pyx_kp_u_break);
  PyList_SET_ITEM(__pyx_t_5, 44, __pyx_kp_u_break);
  __Pyx_INCREF(__pyx_kp_u_return);
  __Pyx_GIVEREF(__pyx_kp_u_return);
  PyList_SET_ITEM(__pyx_t_5, 45, __pyx_kp_u_return);
  __Pyx_INCREF(__pyx_kp_u_then);
  __Pyx_GIVEREF(__pyx_kp_u_then);
  PyList_SET_ITEM(__pyx_t_5, 46, __pyx_kp_u_then);
  __Pyx_INCREF(__pyx_kp_u_elseif);
  __Pyx_GIVEREF(__pyx_kp_u_elseif);
  PyList_SET_ITEM(__pyx_t_5, 47, __pyx_kp_u_elseif);
  __Pyx_INCREF(__pyx_kp_u_else);
  __Pyx_GIVEREF(__pyx_kp_u_else);
  PyList_SET_ITEM(__pyx_t_5, 48, __pyx_kp_u_else);
  __Pyx_INCREF(__pyx_kp_u_for);
  __Pyx_GIVEREF(__pyx_kp_u_for);
  PyList_SET_ITEM(__pyx_t_5, 49, __pyx_kp_u_for);
  __Pyx_INCREF(__pyx_kp_u_loop);
  __Pyx_GIVEREF(__pyx_kp_u_loop);
  PyList_SET_ITEM(__pyx_t_5, 50, __pyx_kp_u_loop);
  __Pyx_INCREF(__pyx_kp_u_in);
  __Pyx_GIVEREF(__pyx_kp_u_in);
  PyList_SET_ITEM(__pyx_t_5, 51, __pyx_kp_u_in);
  __Pyx_INCREF(__pyx_kp_u_while);
  __Pyx_GIVEREF(__pyx_kp_u_while);
  PyList_SET_ITEM(__pyx_t_5, 52, __pyx_kp_u_while);
  __Pyx_INCREF(__pyx_kp_u_when);
  __Pyx_GIVEREF(__pyx_kp_u_when);
  PyList_SET_ITEM(__pyx_t_5, 53, __pyx_kp_u_when);
  __Pyx_INCREF(__pyx_kp_u_elsewhen);
  __Pyx_GIVEREF(__pyx_kp_u_elsewhen);
  PyList_SET_ITEM(__pyx_t_5, 54, __pyx_kp_u_elsewhen);
  __Pyx_INCREF(__pyx_kp_u_connect);
  __Pyx_GIVEREF(__pyx_kp_u_connect);
  PyList_SET_ITEM(__pyx_t_5, 55, __pyx_kp_u_connect);
  __Pyx_INCREF(__pyx_kp_u__989);
  __Pyx_GIVEREF(__pyx_kp_u__989);
  PyList_SET_ITEM(__pyx_t_5, 56, __pyx_kp_u__989);
  __Pyx_INCREF(__pyx_kp_u__990);
  __Pyx_GIVEREF(__pyx_kp_u__990);
  PyList_SET_ITEM(__pyx_t_5, 57, __pyx_kp_u__990);
  __Pyx_INCREF(__pyx_kp_u__991);
  __Pyx_GIVEREF(__pyx_kp_u__991);
  PyList_SET_ITEM(__pyx_t_5, 58, __pyx_kp_u__991);
  __Pyx_INCREF(__pyx_kp_u__992);
  __Pyx_GIVEREF(__pyx_kp_u__992);
  PyList_SET_ITEM(__pyx_t_5, 59, __pyx_kp_u__992);
  __Pyx_INCREF(__pyx_kp_u__993);
  __Pyx_GIVEREF(__pyx_kp_u__993);
  PyList_SET_ITEM(__pyx_t_5, 60, __pyx_kp_u__993);
  __Pyx_INCREF(__pyx_kp_u__994);
  __Pyx_GIVEREF(__pyx_kp_u__994);
  PyList_SET_ITEM(__pyx_t_5, 61, __pyx_kp_u__994);
  __Pyx_INCREF(__pyx_kp_u__995);
  __Pyx_GIVEREF(__pyx_kp_u__995);
  PyList_SET_ITEM(__pyx_t_5, 62, __pyx_kp_u__995);
  __Pyx_INCREF(__pyx_kp_u__996);
  __Pyx_GIVEREF(__pyx_kp_u__996);
  PyList_SET_ITEM(__pyx_t_5, 63, __pyx_kp_u__996);
  __Pyx_INCREF(__pyx_kp_u__997);
  __Pyx_GIVEREF(__pyx_kp_u__997);
  PyList_SET_ITEM(__pyx_t_5, 64, __pyx_kp_u__997);
  __Pyx_INCREF(__pyx_kp_u__998);
  __Pyx_GIVEREF(__pyx_kp_u__998);
  PyList_SET_ITEM(__pyx_t_5, 65, __pyx_kp_u__998);
  __Pyx_INCREF(__pyx_kp_u__999);
  __Pyx_GIVEREF(__pyx_kp_u__999);
  PyList_SET_ITEM(__pyx_t_5, 66, __pyx_kp_u__999);
  __Pyx_INCREF(__pyx_kp_u__1000);
  __Pyx_GIVEREF(__pyx_kp_u__1000);
  PyList_SET_ITEM(__pyx_t_5, 67, __pyx_kp_u__1000);
  __Pyx_INCREF(__pyx_kp_u__1001);
  __Pyx_GIVEREF(__pyx_kp_u__1001);
  PyList_SET_ITEM(__pyx_t_5, 68, __pyx_kp_u__1001);
  __Pyx_INCREF(__pyx_kp_u__1002);
  __Pyx_GIVEREF(__pyx_kp_u__1002);
  PyList_SET_ITEM(__pyx_t_5, 69, __pyx_kp_u__1002);
  __Pyx_INCREF(__pyx_kp_u__1003);
  __Pyx_GIVEREF(__pyx_kp_u__1003);
  PyList_SET_ITEM(__pyx_t_5, 70, __pyx_kp_u__1003);
  __Pyx_INCREF(__pyx_kp_u_not);
  __Pyx_GIVEREF(__pyx_kp_u_not);
  PyList_SET_ITEM(__pyx_t_5, 71, __pyx_kp_u_not);
  __Pyx_INCREF(__pyx_kp_u_and);
  __Pyx_GIVEREF(__pyx_kp_u_and);
  PyList_SET_ITEM(__pyx_t_5, 72, __pyx_kp_u_and);
  __Pyx_INCREF(__pyx_kp_u_or);
  __Pyx_GIVEREF(__pyx_kp_u_or);
  PyList_SET_ITEM(__pyx_t_5, 73, __pyx_kp_u_or);
  __Pyx_INCREF(__pyx_kp_u_false);
  __Pyx_GIVEREF(__pyx_kp_u_false);
  PyList_SET_ITEM(__pyx_t_5, 74, __pyx_kp_u_false);
  __Pyx_INCREF(__pyx_kp_u_true);
  __Pyx_GIVEREF(__pyx_kp_u_true);
  PyList_SET_ITEM(__pyx_t_5, 75, __pyx_kp_u_true);
  __Pyx_INCREF(__pyx_kp_u__1004);
  __Pyx_GIVEREF(__pyx_kp_u__1004);
  PyList_SET_ITEM(__pyx_t_5, 76, __pyx_kp_u__1004);
  __Pyx_INCREF(__pyx_kp_u__1005);
  __Pyx_GIVEREF(__pyx_kp_u__1005);
  PyList_SET_ITEM(__pyx_t_5, 77, __pyx_kp_u__1005);
  __Pyx_INCREF(__pyx_kp_u_annotation_2);
  __Pyx_GIVEREF(__pyx_kp_u_annotation_2);
  PyList_SET_ITEM(__pyx_t_5, 78, __pyx_kp_u_annotation_2);
  __Pyx_INCREF(__pyx_kp_u_each);
  __Pyx_GIVEREF(__pyx_kp_u_each);
  PyList_SET_ITEM(__pyx_t_5, 79, __pyx_kp_u_each);
  __Pyx_INCREF(__pyx_kp_u_partial);
  __Pyx_GIVEREF(__pyx_kp_u_partial);
  PyList_SET_ITEM(__pyx_t_5, 80, __pyx_kp_u_partial);
  __Pyx_INCREF(__pyx_kp_u_final);
  __Pyx_GIVEREF(__pyx_kp_u_final);
  PyList_SET_ITEM(__pyx_t_5, 81, __pyx_kp_u_final);
  __Pyx_INCREF(__pyx_kp_u_within);
  __Pyx_GIVEREF(__pyx_kp_u_within);
  PyList_SET_ITEM(__pyx_t_5, 82, __pyx_kp_u_within);
  __Pyx_INCREF(__pyx_kp_u_encapsulated);
  __Pyx_GIVEREF(__pyx_kp_u_encapsulated);
  PyList_SET_ITEM(__pyx_t_5, 83, __pyx_kp_u_encapsulated);
  __Pyx_INCREF(__pyx_kp_u_redeclare);
  __Pyx_GIVEREF(__pyx_kp_u_redeclare);
  PyList_SET_ITEM(__pyx_t_5, 84, __pyx_kp_u_redeclare);
  __Pyx_INCREF(__pyx_kp_u_inner);
  __Pyx_GIVEREF(__pyx_kp_u_inner);
  PyList_SET_ITEM(__pyx_t_5, 85, __pyx_kp_u_inner);
  __Pyx_INCREF(__pyx_kp_u_outer);
  __Pyx_GIVEREF(__pyx_kp_u_outer);
  PyList_SET_ITEM(__pyx_t_5, 86, __pyx_kp_u_outer);
  __Pyx_INCREF(__pyx_kp_u_initial);
  __Pyx_GIVEREF(__pyx_kp_u_initial);
  PyList_SET_ITEM(__pyx_t_5, 87, __pyx_kp_u_initial);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_literalNames, __pyx_t_5) < 0) __PYX_ERR(0, 542, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":560
 *                      "'inner'", "'outer'", "'initial'" ]
 * 
 *     symbolicNames = [ "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>",             # <<<<<<<<<<<<<<
 *                       "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>",
 *                       "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>",
 */
  __pyx_t_5 = PyList_New(93); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 560, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 0, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 1, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 2, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 3, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 4, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 5, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 6, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 7, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 8, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 9, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 10, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 11, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 12, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 13, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 14, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 15, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 16, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 17, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 18, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 19, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 20, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 21, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 22, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 23, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 24, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 25, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 26, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 27, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 28, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 29, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 30, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 31, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 32, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 33, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 34, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 35, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 36, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 37, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 38, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 39, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 40, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 41, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 42, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 43, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 44, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 45, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 46, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 47, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 48, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 49, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 50, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 51, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 52, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 53, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 54, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 55, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 56, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 57, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 58, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 59, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 60, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 61, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 62, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 63, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 64, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 65, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 66, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 67, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 68, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 69, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 70, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 71, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 72, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 73, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 74, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 75, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 76, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 77, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_kp_u_INVALID);
  __Pyx_GIVEREF(__pyx_kp_u_INVALID);
  PyList_SET_ITEM(__pyx_t_5, 78, __pyx_kp_u_INVALID);
  __Pyx_INCREF(__pyx_n_u_EACH);
  __Pyx_GIVEREF(__pyx_n_u_EACH);
  PyList_SET_ITEM(__pyx_t_5, 79, __pyx_n_u_EACH);
  __Pyx_INCREF(__pyx_n_u_PARTIAL);
  __Pyx_GIVEREF(__pyx_n_u_PARTIAL);
  PyList_SET_ITEM(__pyx_t_5, 80, __pyx_n_u_PARTIAL);
  __Pyx_INCREF(__pyx_n_u_FINAL);
  __Pyx_GIVEREF(__pyx_n_u_FINAL);
  PyList_SET_ITEM(__pyx_t_5, 81, __pyx_n_u_FINAL);
  __Pyx_INCREF(__pyx_n_u_WITHIN);
  __Pyx_GIVEREF(__pyx_n_u_WITHIN);
  PyList_SET_ITEM(__pyx_t_5, 82, __pyx_n_u_WITHIN);
  __Pyx_INCREF(__pyx_n_u_ENCAPSULATED);
  __Pyx_GIVEREF(__pyx_n_u_ENCAPSULATED);
  PyList_SET_ITEM(__pyx_t_5, 83, __pyx_n_u_ENCAPSULATED);
  __Pyx_INCREF(__pyx_n_u_REDECLARE);
  __Pyx_GIVEREF(__pyx_n_u_REDECLARE);
  PyList_SET_ITEM(__pyx_t_5, 84, __pyx_n_u_REDECLARE);
  __Pyx_INCREF(__pyx_n_u_INNER);
  __Pyx_GIVEREF(__pyx_n_u_INNER);
  PyList_SET_ITEM(__pyx_t_5, 85, __pyx_n_u_INNER);
  __Pyx_INCREF(__pyx_n_u_OUTER);
  __Pyx_GIVEREF(__pyx_n_u_OUTER);
  PyList_SET_ITEM(__pyx_t_5, 86, __pyx_n_u_OUTER);
  __Pyx_INCREF(__pyx_n_u_INITIAL);
  __Pyx_GIVEREF(__pyx_n_u_INITIAL);
  PyList_SET_ITEM(__pyx_t_5, 87, __pyx_n_u_INITIAL);
  __Pyx_INCREF(__pyx_n_u_IDENT);
  __Pyx_GIVEREF(__pyx_n_u_IDENT);
  PyList_SET_ITEM(__pyx_t_5, 88, __pyx_n_u_IDENT);
  __Pyx_INCREF(__pyx_n_u_STRING);
  __Pyx_GIVEREF(__pyx_n_u_STRING);
  PyList_SET_ITEM(__pyx_t_5, 89, __pyx_n_u_STRING);
  __Pyx_INCREF(__pyx_n_u_UNSIGNED_NUMBER);
  __Pyx_GIVEREF(__pyx_n_u_UNSIGNED_NUMBER);
  PyList_SET_ITEM(__pyx_t_5, 90, __pyx_n_u_UNSIGNED_NUMBER);
  __Pyx_INCREF(__pyx_n_u_COMMENT);
  __Pyx_GIVEREF(__pyx_n_u_COMMENT);
  PyList_SET_ITEM(__pyx_t_5, 91, __pyx_n_u_COMMENT);
  __Pyx_INCREF(__pyx_n_u_WS);
  __Pyx_GIVEREF(__pyx_n_u_WS);
  PyList_SET_ITEM(__pyx_t_5, 92, __pyx_n_u_WS);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_symbolicNames, __pyx_t_5) < 0) __PYX_ERR(0, 560, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":584
 *                       "COMMENT", "WS" ]
 * 
 *     RULE_stored_definition = 0             # <<<<<<<<<<<<<<
 *     RULE_stored_definition_class = 1
 *     RULE_class_definition = 2
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_stored_definition, __pyx_int_0) < 0) __PYX_ERR(0, 584, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":585
 * 
 *     RULE_stored_definition = 0
 *     RULE_stored_definition_class = 1             # <<<<<<<<<<<<<<
 *     RULE_class_definition = 2
 *     RULE_class_prefixes = 3
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_stored_definition_class, __pyx_int_1) < 0) __PYX_ERR(0, 585, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":586
 *     RULE_stored_definition = 0
 *     RULE_stored_definition_class = 1
 *     RULE_class_definition = 2             # <<<<<<<<<<<<<<
 *     RULE_class_prefixes = 3
 *     RULE_class_type = 4
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_class_definition, __pyx_int_2) < 0) __PYX_ERR(0, 586, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":587
 *     RULE_stored_definition_class = 1
 *     RULE_class_definition = 2
 *     RULE_class_prefixes = 3             # <<<<<<<<<<<<<<
 *     RULE_class_type = 4
 *     RULE_class_specifier = 5
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_class_prefixes, __pyx_int_3) < 0) __PYX_ERR(0, 587, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":588
 *     RULE_class_definition = 2
 *     RULE_class_prefixes = 3
 *     RULE_class_type = 4             # <<<<<<<<<<<<<<
 *     RULE_class_specifier = 5
 *     RULE_base_prefix = 6
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_class_type, __pyx_int_4) < 0) __PYX_ERR(0, 588, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":589
 *     RULE_class_prefixes = 3
 *     RULE_class_type = 4
 *     RULE_class_specifier = 5             # <<<<<<<<<<<<<<
 *     RULE_base_prefix = 6
 *     RULE_enum_list = 7
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_class_specifier, __pyx_int_5) < 0) __PYX_ERR(0, 589, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":590
 *     RULE_class_type = 4
 *     RULE_class_specifier = 5
 *     RULE_base_prefix = 6             # <<<<<<<<<<<<<<
 *     RULE_enum_list = 7
 *     RULE_enumeration_literal = 8
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_base_prefix, __pyx_int_6) < 0) __PYX_ERR(0, 590, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":591
 *     RULE_class_specifier = 5
 *     RULE_base_prefix = 6
 *     RULE_enum_list = 7             # <<<<<<<<<<<<<<
 *     RULE_enumeration_literal = 8
 *     RULE_composition = 9
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_enum_list, __pyx_int_7) < 0) __PYX_ERR(0, 591, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":592
 *     RULE_base_prefix = 6
 *     RULE_enum_list = 7
 *     RULE_enumeration_literal = 8             # <<<<<<<<<<<<<<
 *     RULE_composition = 9
 *     RULE_language_specification = 10
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_enumeration_literal, __pyx_int_8) < 0) __PYX_ERR(0, 592, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":593
 *     RULE_enum_list = 7
 *     RULE_enumeration_literal = 8
 *     RULE_composition = 9             # <<<<<<<<<<<<<<
 *     RULE_language_specification = 10
 *     RULE_external_function_call = 11
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_composition, __pyx_int_9) < 0) __PYX_ERR(0, 593, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":594
 *     RULE_enumeration_literal = 8
 *     RULE_composition = 9
 *     RULE_language_specification = 10             # <<<<<<<<<<<<<<
 *     RULE_external_function_call = 11
 *     RULE_element_list = 12
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_language_specification, __pyx_int_10) < 0) __PYX_ERR(0, 594, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":595
 *     RULE_composition = 9
 *     RULE_language_specification = 10
 *     RULE_external_function_call = 11             # <<<<<<<<<<<<<<
 *     RULE_element_list = 12
 *     RULE_element = 13
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_external_function_call, __pyx_int_11) < 0) __PYX_ERR(0, 595, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":596
 *     RULE_language_specification = 10
 *     RULE_external_function_call = 11
 *     RULE_element_list = 12             # <<<<<<<<<<<<<<
 *     RULE_element = 13
 *     RULE_regular_element = 14
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_element_list, __pyx_int_12) < 0) __PYX_ERR(0, 596, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":597
 *     RULE_external_function_call = 11
 *     RULE_element_list = 12
 *     RULE_element = 13             # <<<<<<<<<<<<<<
 *     RULE_regular_element = 14
 *     RULE_replaceable_element = 15
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_element, __pyx_int_13) < 0) __PYX_ERR(0, 597, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":598
 *     RULE_element_list = 12
 *     RULE_element = 13
 *     RULE_regular_element = 14             # <<<<<<<<<<<<<<
 *     RULE_replaceable_element = 15
 *     RULE_import_clause = 16
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_regular_element, __pyx_int_14) < 0) __PYX_ERR(0, 598, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":599
 *     RULE_element = 13
 *     RULE_regular_element = 14
 *     RULE_replaceable_element = 15             # <<<<<<<<<<<<<<
 *     RULE_import_clause = 16
 *     RULE_import_list = 17
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_replaceable_element, __pyx_int_15) < 0) __PYX_ERR(0, 599, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":600
 *     RULE_regular_element = 14
 *     RULE_replaceable_element = 15
 *     RULE_import_clause = 16             # <<<<<<<<<<<<<<
 *     RULE_import_list = 17
 *     RULE_extends_clause = 18
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_import_clause, __pyx_int_16) < 0) __PYX_ERR(0, 600, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":601
 *     RULE_replaceable_element = 15
 *     RULE_import_clause = 16
 *     RULE_import_list = 17             # <<<<<<<<<<<<<<
 *     RULE_extends_clause = 18
 *     RULE_constraining_clause = 19
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_import_list, __pyx_int_17) < 0) __PYX_ERR(0, 601, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":602
 *     RULE_import_clause = 16
 *     RULE_import_list = 17
 *     RULE_extends_clause = 18             # <<<<<<<<<<<<<<
 *     RULE_constraining_clause = 19
 *     RULE_component_clause = 20
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_extends_clause, __pyx_int_18) < 0) __PYX_ERR(0, 602, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":603
 *     RULE_import_list = 17
 *     RULE_extends_clause = 18
 *     RULE_constraining_clause = 19             # <<<<<<<<<<<<<<
 *     RULE_component_clause = 20
 *     RULE_type_prefix = 21
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_constraining_clause, __pyx_int_19) < 0) __PYX_ERR(0, 603, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":604
 *     RULE_extends_clause = 18
 *     RULE_constraining_clause = 19
 *     RULE_component_clause = 20             # <<<<<<<<<<<<<<
 *     RULE_type_prefix = 21
 *     RULE_type_specifier = 22
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_component_clause, __pyx_int_20) < 0) __PYX_ERR(0, 604, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":605
 *     RULE_constraining_clause = 19
 *     RULE_component_clause = 20
 *     RULE_type_prefix = 21             # <<<<<<<<<<<<<<
 *     RULE_type_specifier = 22
 *     RULE_component_list = 23
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_type_prefix, __pyx_int_21) < 0) __PYX_ERR(0, 605, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":606
 *     RULE_component_clause = 20
 *     RULE_type_prefix = 21
 *     RULE_type_specifier = 22             # <<<<<<<<<<<<<<
 *     RULE_component_list = 23
 *     RULE_component_declaration = 24
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_type_specifier, __pyx_int_22) < 0) __PYX_ERR(0, 606, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":607
 *     RULE_type_prefix = 21
 *     RULE_type_specifier = 22
 *     RULE_component_list = 23             # <<<<<<<<<<<<<<
 *     RULE_component_declaration = 24
 *     RULE_condition_attribute = 25
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_component_list, __pyx_int_23) < 0) __PYX_ERR(0, 607, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":608
 *     RULE_type_specifier = 22
 *     RULE_component_list = 23
 *     RULE_component_declaration = 24             # <<<<<<<<<<<<<<
 *     RULE_condition_attribute = 25
 *     RULE_declaration = 26
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_component_declaration, __pyx_int_24) < 0) __PYX_ERR(0, 608, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":609
 *     RULE_component_list = 23
 *     RULE_component_declaration = 24
 *     RULE_condition_attribute = 25             # <<<<<<<<<<<<<<
 *     RULE_declaration = 26
 *     RULE_modification = 27
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_condition_attribute, __pyx_int_25) < 0) __PYX_ERR(0, 609, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":610
 *     RULE_component_declaration = 24
 *     RULE_condition_attribute = 25
 *     RULE_declaration = 26             # <<<<<<<<<<<<<<
 *     RULE_modification = 27
 *     RULE_class_modification = 28
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_declaration, __pyx_int_26) < 0) __PYX_ERR(0, 610, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":611
 *     RULE_condition_attribute = 25
 *     RULE_declaration = 26
 *     RULE_modification = 27             # <<<<<<<<<<<<<<
 *     RULE_class_modification = 28
 *     RULE_argument_list = 29
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_modification, __pyx_int_27) < 0) __PYX_ERR(0, 611, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":612
 *     RULE_declaration = 26
 *     RULE_modification = 27
 *     RULE_class_modification = 28             # <<<<<<<<<<<<<<
 *     RULE_argument_list = 29
 *     RULE_argument = 30
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_class_modification, __pyx_int_28) < 0) __PYX_ERR(0, 612, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":613
 *     RULE_modification = 27
 *     RULE_class_modification = 28
 *     RULE_argument_list = 29             # <<<<<<<<<<<<<<
 *     RULE_argument = 30
 *     RULE_element_modification_or_replaceable = 31
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_argument_list, __pyx_int_29) < 0) __PYX_ERR(0, 613, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":614
 *     RULE_class_modification = 28
 *     RULE_argument_list = 29
 *     RULE_argument = 30             # <<<<<<<<<<<<<<
 *     RULE_element_modification_or_replaceable = 31
 *     RULE_element_modification = 32
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_argument, __pyx_int_30) < 0) __PYX_ERR(0, 614, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":615
 *     RULE_argument_list = 29
 *     RULE_argument = 30
 *     RULE_element_modification_or_replaceable = 31             # <<<<<<<<<<<<<<
 *     RULE_element_modification = 32
 *     RULE_element_redeclaration = 33
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_element_modification_or_rep, __pyx_int_31) < 0) __PYX_ERR(0, 615, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":616
 *     RULE_argument = 30
 *     RULE_element_modification_or_replaceable = 31
 *     RULE_element_modification = 32             # <<<<<<<<<<<<<<
 *     RULE_element_redeclaration = 33
 *     RULE_element_replaceable = 34
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_element_modification, __pyx_int_32) < 0) __PYX_ERR(0, 616, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":617
 *     RULE_element_modification_or_replaceable = 31
 *     RULE_element_modification = 32
 *     RULE_element_redeclaration = 33             # <<<<<<<<<<<<<<
 *     RULE_element_replaceable = 34
 *     RULE_component_clause1 = 35
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_element_redeclaration, __pyx_int_33) < 0) __PYX_ERR(0, 617, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":618
 *     RULE_element_modification = 32
 *     RULE_element_redeclaration = 33
 *     RULE_element_replaceable = 34             # <<<<<<<<<<<<<<
 *     RULE_component_clause1 = 35
 *     RULE_component_declaration1 = 36
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_element_replaceable, __pyx_int_34) < 0) __PYX_ERR(0, 618, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":619
 *     RULE_element_redeclaration = 33
 *     RULE_element_replaceable = 34
 *     RULE_component_clause1 = 35             # <<<<<<<<<<<<<<
 *     RULE_component_declaration1 = 36
 *     RULE_short_class_definition = 37
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_component_clause1, __pyx_int_35) < 0) __PYX_ERR(0, 619, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":620
 *     RULE_element_replaceable = 34
 *     RULE_component_clause1 = 35
 *     RULE_component_declaration1 = 36             # <<<<<<<<<<<<<<
 *     RULE_short_class_definition = 37
 *     RULE_equation_section = 38
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_component_declaration1, __pyx_int_36) < 0) __PYX_ERR(0, 620, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":621
 *     RULE_component_clause1 = 35
 *     RULE_component_declaration1 = 36
 *     RULE_short_class_definition = 37             # <<<<<<<<<<<<<<
 *     RULE_equation_section = 38
 *     RULE_algorithm_section = 39
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_short_class_definition, __pyx_int_37) < 0) __PYX_ERR(0, 621, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":622
 *     RULE_component_declaration1 = 36
 *     RULE_short_class_definition = 37
 *     RULE_equation_section = 38             # <<<<<<<<<<<<<<
 *     RULE_algorithm_section = 39
 *     RULE_equation_options = 40
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_equation_section, __pyx_int_38) < 0) __PYX_ERR(0, 622, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":623
 *     RULE_short_class_definition = 37
 *     RULE_equation_section = 38
 *     RULE_algorithm_section = 39             # <<<<<<<<<<<<<<
 *     RULE_equation_options = 40
 *     RULE_equation = 41
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_algorithm_section, __pyx_int_39) < 0) __PYX_ERR(0, 623, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":624
 *     RULE_equation_section = 38
 *     RULE_algorithm_section = 39
 *     RULE_equation_options = 40             # <<<<<<<<<<<<<<
 *     RULE_equation = 41
 *     RULE_statement_options = 42
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_equation_options, __pyx_int_40) < 0) __PYX_ERR(0, 624, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":625
 *     RULE_algorithm_section = 39
 *     RULE_equation_options = 40
 *     RULE_equation = 41             # <<<<<<<<<<<<<<
 *     RULE_statement_options = 42
 *     RULE_statement = 43
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_equation, __pyx_int_41) < 0) __PYX_ERR(0, 625, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":626
 *     RULE_equation_options = 40
 *     RULE_equation = 41
 *     RULE_statement_options = 42             # <<<<<<<<<<<<<<
 *     RULE_statement = 43
 *     RULE_if_equation = 44
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_statement_options, __pyx_int_42) < 0) __PYX_ERR(0, 626, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":627
 *     RULE_equation = 41
 *     RULE_statement_options = 42
 *     RULE_statement = 43             # <<<<<<<<<<<<<<
 *     RULE_if_equation = 44
 *     RULE_if_statement = 45
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_statement, __pyx_int_43) < 0) __PYX_ERR(0, 627, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":628
 *     RULE_statement_options = 42
 *     RULE_statement = 43
 *     RULE_if_equation = 44             # <<<<<<<<<<<<<<
 *     RULE_if_statement = 45
 *     RULE_for_equation = 46
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_if_equation, __pyx_int_44) < 0) __PYX_ERR(0, 628, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":629
 *     RULE_statement = 43
 *     RULE_if_equation = 44
 *     RULE_if_statement = 45             # <<<<<<<<<<<<<<
 *     RULE_for_equation = 46
 *     RULE_for_statement = 47
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_if_statement, __pyx_int_45) < 0) __PYX_ERR(0, 629, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":630
 *     RULE_if_equation = 44
 *     RULE_if_statement = 45
 *     RULE_for_equation = 46             # <<<<<<<<<<<<<<
 *     RULE_for_statement = 47
 *     RULE_for_indices = 48
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_for_equation, __pyx_int_46) < 0) __PYX_ERR(0, 630, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":631
 *     RULE_if_statement = 45
 *     RULE_for_equation = 46
 *     RULE_for_statement = 47             # <<<<<<<<<<<<<<
 *     RULE_for_indices = 48
 *     RULE_for_index = 49
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_for_statement, __pyx_int_47) < 0) __PYX_ERR(0, 631, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":632
 *     RULE_for_equation = 46
 *     RULE_for_statement = 47
 *     RULE_for_indices = 48             # <<<<<<<<<<<<<<
 *     RULE_for_index = 49
 *     RULE_while_statement = 50
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_for_indices, __pyx_int_48) < 0) __PYX_ERR(0, 632, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":633
 *     RULE_for_statement = 47
 *     RULE_for_indices = 48
 *     RULE_for_index = 49             # <<<<<<<<<<<<<<
 *     RULE_while_statement = 50
 *     RULE_when_equation = 51
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_for_index, __pyx_int_49) < 0) __PYX_ERR(0, 633, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":634
 *     RULE_for_indices = 48
 *     RULE_for_index = 49
 *     RULE_while_statement = 50             # <<<<<<<<<<<<<<
 *     RULE_when_equation = 51
 *     RULE_when_statement = 52
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_while_statement, __pyx_int_50) < 0) __PYX_ERR(0, 634, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":635
 *     RULE_for_index = 49
 *     RULE_while_statement = 50
 *     RULE_when_equation = 51             # <<<<<<<<<<<<<<
 *     RULE_when_statement = 52
 *     RULE_connect_clause = 53
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_when_equation, __pyx_int_51) < 0) __PYX_ERR(0, 635, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":636
 *     RULE_while_statement = 50
 *     RULE_when_equation = 51
 *     RULE_when_statement = 52             # <<<<<<<<<<<<<<
 *     RULE_connect_clause = 53
 *     RULE_expression = 54
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_when_statement, __pyx_int_52) < 0) __PYX_ERR(0, 636, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":637
 *     RULE_when_equation = 51
 *     RULE_when_statement = 52
 *     RULE_connect_clause = 53             # <<<<<<<<<<<<<<
 *     RULE_expression = 54
 *     RULE_simple_expression = 55
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_connect_clause, __pyx_int_53) < 0) __PYX_ERR(0, 637, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":638
 *     RULE_when_statement = 52
 *     RULE_connect_clause = 53
 *     RULE_expression = 54             # <<<<<<<<<<<<<<
 *     RULE_simple_expression = 55
 *     RULE_expr = 56
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_expression, __pyx_int_54) < 0) __PYX_ERR(0, 638, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":639
 *     RULE_connect_clause = 53
 *     RULE_expression = 54
 *     RULE_simple_expression = 55             # <<<<<<<<<<<<<<
 *     RULE_expr = 56
 *     RULE_primary = 57
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_simple_expression, __pyx_int_55) < 0) __PYX_ERR(0, 639, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":640
 *     RULE_expression = 54
 *     RULE_simple_expression = 55
 *     RULE_expr = 56             # <<<<<<<<<<<<<<
 *     RULE_primary = 57
 *     RULE_name = 58
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_expr, __pyx_int_56) < 0) __PYX_ERR(0, 640, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":641
 *     RULE_simple_expression = 55
 *     RULE_expr = 56
 *     RULE_primary = 57             # <<<<<<<<<<<<<<
 *     RULE_name = 58
 *     RULE_component_reference_element = 59
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_primary, __pyx_int_57) < 0) __PYX_ERR(0, 641, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":642
 *     RULE_expr = 56
 *     RULE_primary = 57
 *     RULE_name = 58             # <<<<<<<<<<<<<<
 *     RULE_component_reference_element = 59
 *     RULE_component_reference = 60
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_name, __pyx_int_58) < 0) __PYX_ERR(0, 642, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":643
 *     RULE_primary = 57
 *     RULE_name = 58
 *     RULE_component_reference_element = 59             # <<<<<<<<<<<<<<
 *     RULE_component_reference = 60
 *     RULE_function_call_args = 61
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_component_reference_element, __pyx_int_59) < 0) __PYX_ERR(0, 643, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":644
 *     RULE_name = 58
 *     RULE_component_reference_element = 59
 *     RULE_component_reference = 60             # <<<<<<<<<<<<<<
 *     RULE_function_call_args = 61
 *     RULE_function_arguments = 62
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_component_reference, __pyx_int_60) < 0) __PYX_ERR(0, 644, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":645
 *     RULE_component_reference_element = 59
 *     RULE_component_reference = 60
 *     RULE_function_call_args = 61             # <<<<<<<<<<<<<<
 *     RULE_function_arguments = 62
 *     RULE_named_arguments = 63
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_function_call_args, __pyx_int_61) < 0) __PYX_ERR(0, 645, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":646
 *     RULE_component_reference = 60
 *     RULE_function_call_args = 61
 *     RULE_function_arguments = 62             # <<<<<<<<<<<<<<
 *     RULE_named_arguments = 63
 *     RULE_named_argument = 64
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_function_arguments, __pyx_int_62) < 0) __PYX_ERR(0, 646, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":647
 *     RULE_function_call_args = 61
 *     RULE_function_arguments = 62
 *     RULE_named_arguments = 63             # <<<<<<<<<<<<<<
 *     RULE_named_argument = 64
 *     RULE_function_argument = 65
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_named_arguments, __pyx_int_63) < 0) __PYX_ERR(0, 647, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":648
 *     RULE_function_arguments = 62
 *     RULE_named_arguments = 63
 *     RULE_named_argument = 64             # <<<<<<<<<<<<<<
 *     RULE_function_argument = 65
 *     RULE_output_expression_list = 66
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_named_argument, __pyx_int_64) < 0) __PYX_ERR(0, 648, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":649
 *     RULE_named_arguments = 63
 *     RULE_named_argument = 64
 *     RULE_function_argument = 65             # <<<<<<<<<<<<<<
 *     RULE_output_expression_list = 66
 *     RULE_expression_list = 67
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_function_argument, __pyx_int_65) < 0) __PYX_ERR(0, 649, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":650
 *     RULE_named_argument = 64
 *     RULE_function_argument = 65
 *     RULE_output_expression_list = 66             # <<<<<<<<<<<<<<
 *     RULE_expression_list = 67
 *     RULE_array_subscripts = 68
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_output_expression_list, __pyx_int_66) < 0) __PYX_ERR(0, 650, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":651
 *     RULE_function_argument = 65
 *     RULE_output_expression_list = 66
 *     RULE_expression_list = 67             # <<<<<<<<<<<<<<
 *     RULE_array_subscripts = 68
 *     RULE_subscript = 69
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_expression_list, __pyx_int_67) < 0) __PYX_ERR(0, 651, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":652
 *     RULE_output_expression_list = 66
 *     RULE_expression_list = 67
 *     RULE_array_subscripts = 68             # <<<<<<<<<<<<<<
 *     RULE_subscript = 69
 *     RULE_comment = 70
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_array_subscripts, __pyx_int_68) < 0) __PYX_ERR(0, 652, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":653
 *     RULE_expression_list = 67
 *     RULE_array_subscripts = 68
 *     RULE_subscript = 69             # <<<<<<<<<<<<<<
 *     RULE_comment = 70
 *     RULE_string_comment = 71
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_subscript, __pyx_int_69) < 0) __PYX_ERR(0, 653, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":654
 *     RULE_array_subscripts = 68
 *     RULE_subscript = 69
 *     RULE_comment = 70             # <<<<<<<<<<<<<<
 *     RULE_string_comment = 71
 *     RULE_annotation = 72
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_comment, __pyx_int_70) < 0) __PYX_ERR(0, 654, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":655
 *     RULE_subscript = 69
 *     RULE_comment = 70
 *     RULE_string_comment = 71             # <<<<<<<<<<<<<<
 *     RULE_annotation = 72
 * 
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_string_comment, __pyx_int_71) < 0) __PYX_ERR(0, 655, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":656
 *     RULE_comment = 70
 *     RULE_string_comment = 71
 *     RULE_annotation = 72             # <<<<<<<<<<<<<<
 * 
 *     ruleNames =  [ "stored_definition", "stored_definition_class", "class_definition",
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_RULE_annotation, __pyx_int_72) < 0) __PYX_ERR(0, 656, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":658
 *     RULE_annotation = 72
 * 
 *     ruleNames =  [ "stored_definition", "stored_definition_class", "class_definition",             # <<<<<<<<<<<<<<
 *                    "class_prefixes", "class_type", "class_specifier", "base_prefix",
 *                    "enum_list", "enumeration_literal", "composition", "language_specification",
 */
  __pyx_t_5 = PyList_New(73); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_n_u_stored_definition);
  __Pyx_GIVEREF(__pyx_n_u_stored_definition);
  PyList_SET_ITEM(__pyx_t_5, 0, __pyx_n_u_stored_definition);
  __Pyx_INCREF(__pyx_n_u_stored_definition_class);
  __Pyx_GIVEREF(__pyx_n_u_stored_definition_class);
  PyList_SET_ITEM(__pyx_t_5, 1, __pyx_n_u_stored_definition_class);
  __Pyx_INCREF(__pyx_n_u_class_definition);
  __Pyx_GIVEREF(__pyx_n_u_class_definition);
  PyList_SET_ITEM(__pyx_t_5, 2, __pyx_n_u_class_definition);
  __Pyx_INCREF(__pyx_n_u_class_prefixes);
  __Pyx_GIVEREF(__pyx_n_u_class_prefixes);
  PyList_SET_ITEM(__pyx_t_5, 3, __pyx_n_u_class_prefixes);
  __Pyx_INCREF(__pyx_n_u_class_type);
  __Pyx_GIVEREF(__pyx_n_u_class_type);
  PyList_SET_ITEM(__pyx_t_5, 4, __pyx_n_u_class_type);
  __Pyx_INCREF(__pyx_n_u_class_specifier);
  __Pyx_GIVEREF(__pyx_n_u_class_specifier);
  PyList_SET_ITEM(__pyx_t_5, 5, __pyx_n_u_class_specifier);
  __Pyx_INCREF(__pyx_n_u_base_prefix);
  __Pyx_GIVEREF(__pyx_n_u_base_prefix);
  PyList_SET_ITEM(__pyx_t_5, 6, __pyx_n_u_base_prefix);
  __Pyx_INCREF(__pyx_n_u_enum_list);
  __Pyx_GIVEREF(__pyx_n_u_enum_list);
  PyList_SET_ITEM(__pyx_t_5, 7, __pyx_n_u_enum_list);
  __Pyx_INCREF(__pyx_n_u_enumeration_literal);
  __Pyx_GIVEREF(__pyx_n_u_enumeration_literal);
  PyList_SET_ITEM(__pyx_t_5, 8, __pyx_n_u_enumeration_literal);
  __Pyx_INCREF(__pyx_n_u_composition);
  __Pyx_GIVEREF(__pyx_n_u_composition);
  PyList_SET_ITEM(__pyx_t_5, 9, __pyx_n_u_composition);
  __Pyx_INCREF(__pyx_n_u_language_specification);
  __Pyx_GIVEREF(__pyx_n_u_language_specification);
  PyList_SET_ITEM(__pyx_t_5, 10, __pyx_n_u_language_specification);
  __Pyx_INCREF(__pyx_n_u_external_function_call);
  __Pyx_GIVEREF(__pyx_n_u_external_function_call);
  PyList_SET_ITEM(__pyx_t_5, 11, __pyx_n_u_external_function_call);
  __Pyx_INCREF(__pyx_n_u_element_list);
  __Pyx_GIVEREF(__pyx_n_u_element_list);
  PyList_SET_ITEM(__pyx_t_5, 12, __pyx_n_u_element_list);
  __Pyx_INCREF(__pyx_n_u_element);
  __Pyx_GIVEREF(__pyx_n_u_element);
  PyList_SET_ITEM(__pyx_t_5, 13, __pyx_n_u_element);
  __Pyx_INCREF(__pyx_n_u_regular_element);
  __Pyx_GIVEREF(__pyx_n_u_regular_element);
  PyList_SET_ITEM(__pyx_t_5, 14, __pyx_n_u_regular_element);
  __Pyx_INCREF(__pyx_n_u_replaceable_element);
  __Pyx_GIVEREF(__pyx_n_u_replaceable_element);
  PyList_SET_ITEM(__pyx_t_5, 15, __pyx_n_u_replaceable_element);
  __Pyx_INCREF(__pyx_n_u_import_clause);
  __Pyx_GIVEREF(__pyx_n_u_import_clause);
  PyList_SET_ITEM(__pyx_t_5, 16, __pyx_n_u_import_clause);
  __Pyx_INCREF(__pyx_n_u_import_list);
  __Pyx_GIVEREF(__pyx_n_u_import_list);
  PyList_SET_ITEM(__pyx_t_5, 17, __pyx_n_u_import_list);
  __Pyx_INCREF(__pyx_n_u_extends_clause);
  __Pyx_GIVEREF(__pyx_n_u_extends_clause);
  PyList_SET_ITEM(__pyx_t_5, 18, __pyx_n_u_extends_clause);
  __Pyx_INCREF(__pyx_n_u_constraining_clause);
  __Pyx_GIVEREF(__pyx_n_u_constraining_clause);
  PyList_SET_ITEM(__pyx_t_5, 19, __pyx_n_u_constraining_clause);
  __Pyx_INCREF(__pyx_n_u_component_clause);
  __Pyx_GIVEREF(__pyx_n_u_component_clause);
  PyList_SET_ITEM(__pyx_t_5, 20, __pyx_n_u_component_clause);
  __Pyx_INCREF(__pyx_n_u_type_prefix);
  __Pyx_GIVEREF(__pyx_n_u_type_prefix);
  PyList_SET_ITEM(__pyx_t_5, 21, __pyx_n_u_type_prefix);
  __Pyx_INCREF(__pyx_n_u_type_specifier);
  __Pyx_GIVEREF(__pyx_n_u_type_specifier);
  PyList_SET_ITEM(__pyx_t_5, 22, __pyx_n_u_type_specifier);
  __Pyx_INCREF(__pyx_n_u_component_list);
  __Pyx_GIVEREF(__pyx_n_u_component_list);
  PyList_SET_ITEM(__pyx_t_5, 23, __pyx_n_u_component_list);
  __Pyx_INCREF(__pyx_n_u_component_declaration);
  __Pyx_GIVEREF(__pyx_n_u_component_declaration);
  PyList_SET_ITEM(__pyx_t_5, 24, __pyx_n_u_component_declaration);
  __Pyx_INCREF(__pyx_n_u_condition_attribute);
  __Pyx_GIVEREF(__pyx_n_u_condition_attribute);
  PyList_SET_ITEM(__pyx_t_5, 25, __pyx_n_u_condition_attribute);
  __Pyx_INCREF(__pyx_n_u_declaration);
  __Pyx_GIVEREF(__pyx_n_u_declaration);
  PyList_SET_ITEM(__pyx_t_5, 26, __pyx_n_u_declaration);
  __Pyx_INCREF(__pyx_n_u_modification);
  __Pyx_GIVEREF(__pyx_n_u_modification);
  PyList_SET_ITEM(__pyx_t_5, 27, __pyx_n_u_modification);
  __Pyx_INCREF(__pyx_n_u_class_modification);
  __Pyx_GIVEREF(__pyx_n_u_class_modification);
  PyList_SET_ITEM(__pyx_t_5, 28, __pyx_n_u_class_modification);
  __Pyx_INCREF(__pyx_n_u_argument_list);
  __Pyx_GIVEREF(__pyx_n_u_argument_list);
  PyList_SET_ITEM(__pyx_t_5, 29, __pyx_n_u_argument_list);
  __Pyx_INCREF(__pyx_n_u_argument);
  __Pyx_GIVEREF(__pyx_n_u_argument);
  PyList_SET_ITEM(__pyx_t_5, 30, __pyx_n_u_argument);
  __Pyx_INCREF(__pyx_n_u_element_modification_or_replacea);
  __Pyx_GIVEREF(__pyx_n_u_element_modification_or_replacea);
  PyList_SET_ITEM(__pyx_t_5, 31, __pyx_n_u_element_modification_or_replacea);
  __Pyx_INCREF(__pyx_n_u_element_modification);
  __Pyx_GIVEREF(__pyx_n_u_element_modification);
  PyList_SET_ITEM(__pyx_t_5, 32, __pyx_n_u_element_modification);
  __Pyx_INCREF(__pyx_n_u_element_redeclaration);
  __Pyx_GIVEREF(__pyx_n_u_element_redeclaration);
  PyList_SET_ITEM(__pyx_t_5, 33, __pyx_n_u_element_redeclaration);
  __Pyx_INCREF(__pyx_n_u_element_replaceable);
  __Pyx_GIVEREF(__pyx_n_u_element_replaceable);
  PyList_SET_ITEM(__pyx_t_5, 34, __pyx_n_u_element_replaceable);
  __Pyx_INCREF(__pyx_n_u_component_clause1);
  __Pyx_GIVEREF(__pyx_n_u_component_clause1);
  PyList_SET_ITEM(__pyx_t_5, 35, __pyx_n_u_component_clause1);
  __Pyx_INCREF(__pyx_n_u_component_declaration1);
  __Pyx_GIVEREF(__pyx_n_u_component_declaration1);
  PyList_SET_ITEM(__pyx_t_5, 36, __pyx_n_u_component_declaration1);
  __Pyx_INCREF(__pyx_n_u_short_class_definition);
  __Pyx_GIVEREF(__pyx_n_u_short_class_definition);
  PyList_SET_ITEM(__pyx_t_5, 37, __pyx_n_u_short_class_definition);
  __Pyx_INCREF(__pyx_n_u_equation_section);
  __Pyx_GIVEREF(__pyx_n_u_equation_section);
  PyList_SET_ITEM(__pyx_t_5, 38, __pyx_n_u_equation_section);
  __Pyx_INCREF(__pyx_n_u_algorithm_section);
  __Pyx_GIVEREF(__pyx_n_u_algorithm_section);
  PyList_SET_ITEM(__pyx_t_5, 39, __pyx_n_u_algorithm_section);
  __Pyx_INCREF(__pyx_n_u_equation_options);
  __Pyx_GIVEREF(__pyx_n_u_equation_options);
  PyList_SET_ITEM(__pyx_t_5, 40, __pyx_n_u_equation_options);
  __Pyx_INCREF(__pyx_n_u_equation);
  __Pyx_GIVEREF(__pyx_n_u_equation);
  PyList_SET_ITEM(__pyx_t_5, 41, __pyx_n_u_equation);
  __Pyx_INCREF(__pyx_n_u_statement_options);
  __Pyx_GIVEREF(__pyx_n_u_statement_options);
  PyList_SET_ITEM(__pyx_t_5, 42, __pyx_n_u_statement_options);
  __Pyx_INCREF(__pyx_n_u_statement);
  __Pyx_GIVEREF(__pyx_n_u_statement);
  PyList_SET_ITEM(__pyx_t_5, 43, __pyx_n_u_statement);
  __Pyx_INCREF(__pyx_n_u_if_equation);
  __Pyx_GIVEREF(__pyx_n_u_if_equation);
  PyList_SET_ITEM(__pyx_t_5, 44, __pyx_n_u_if_equation);
  __Pyx_INCREF(__pyx_n_u_if_statement);
  __Pyx_GIVEREF(__pyx_n_u_if_statement);
  PyList_SET_ITEM(__pyx_t_5, 45, __pyx_n_u_if_statement);
  __Pyx_INCREF(__pyx_n_u_for_equation);
  __Pyx_GIVEREF(__pyx_n_u_for_equation);
  PyList_SET_ITEM(__pyx_t_5, 46, __pyx_n_u_for_equation);
  __Pyx_INCREF(__pyx_n_u_for_statement);
  __Pyx_GIVEREF(__pyx_n_u_for_statement);
  PyList_SET_ITEM(__pyx_t_5, 47, __pyx_n_u_for_statement);
  __Pyx_INCREF(__pyx_n_u_for_indices);
  __Pyx_GIVEREF(__pyx_n_u_for_indices);
  PyList_SET_ITEM(__pyx_t_5, 48, __pyx_n_u_for_indices);
  __Pyx_INCREF(__pyx_n_u_for_index);
  __Pyx_GIVEREF(__pyx_n_u_for_index);
  PyList_SET_ITEM(__pyx_t_5, 49, __pyx_n_u_for_index);
  __Pyx_INCREF(__pyx_n_u_while_statement);
  __Pyx_GIVEREF(__pyx_n_u_while_statement);
  PyList_SET_ITEM(__pyx_t_5, 50, __pyx_n_u_while_statement);
  __Pyx_INCREF(__pyx_n_u_when_equation);
  __Pyx_GIVEREF(__pyx_n_u_when_equation);
  PyList_SET_ITEM(__pyx_t_5, 51, __pyx_n_u_when_equation);
  __Pyx_INCREF(__pyx_n_u_when_statement);
  __Pyx_GIVEREF(__pyx_n_u_when_statement);
  PyList_SET_ITEM(__pyx_t_5, 52, __pyx_n_u_when_statement);
  __Pyx_INCREF(__pyx_n_u_connect_clause);
  __Pyx_GIVEREF(__pyx_n_u_connect_clause);
  PyList_SET_ITEM(__pyx_t_5, 53, __pyx_n_u_connect_clause);
  __Pyx_INCREF(__pyx_n_u_expression);
  __Pyx_GIVEREF(__pyx_n_u_expression);
  PyList_SET_ITEM(__pyx_t_5, 54, __pyx_n_u_expression);
  __Pyx_INCREF(__pyx_n_u_simple_expression);
  __Pyx_GIVEREF(__pyx_n_u_simple_expression);
  PyList_SET_ITEM(__pyx_t_5, 55, __pyx_n_u_simple_expression);
  __Pyx_INCREF(__pyx_n_u_expr);
  __Pyx_GIVEREF(__pyx_n_u_expr);
  PyList_SET_ITEM(__pyx_t_5, 56, __pyx_n_u_expr);
  __Pyx_INCREF(__pyx_n_u_primary);
  __Pyx_GIVEREF(__pyx_n_u_primary);
  PyList_SET_ITEM(__pyx_t_5, 57, __pyx_n_u_primary);
  __Pyx_INCREF(__pyx_n_u_name);
  __Pyx_GIVEREF(__pyx_n_u_name);
  PyList_SET_ITEM(__pyx_t_5, 58, __pyx_n_u_name);
  __Pyx_INCREF(__pyx_n_u_component_reference_element);
  __Pyx_GIVEREF(__pyx_n_u_component_reference_element);
  PyList_SET_ITEM(__pyx_t_5, 59, __pyx_n_u_component_reference_element);
  __Pyx_INCREF(__pyx_n_u_component_reference);
  __Pyx_GIVEREF(__pyx_n_u_component_reference);
  PyList_SET_ITEM(__pyx_t_5, 60, __pyx_n_u_component_reference);
  __Pyx_INCREF(__pyx_n_u_function_call_args);
  __Pyx_GIVEREF(__pyx_n_u_function_call_args);
  PyList_SET_ITEM(__pyx_t_5, 61, __pyx_n_u_function_call_args);
  __Pyx_INCREF(__pyx_n_u_function_arguments);
  __Pyx_GIVEREF(__pyx_n_u_function_arguments);
  PyList_SET_ITEM(__pyx_t_5, 62, __pyx_n_u_function_arguments);
  __Pyx_INCREF(__pyx_n_u_named_arguments);
  __Pyx_GIVEREF(__pyx_n_u_named_arguments);
  PyList_SET_ITEM(__pyx_t_5, 63, __pyx_n_u_named_arguments);
  __Pyx_INCREF(__pyx_n_u_named_argument);
  __Pyx_GIVEREF(__pyx_n_u_named_argument);
  PyList_SET_ITEM(__pyx_t_5, 64, __pyx_n_u_named_argument);
  __Pyx_INCREF(__pyx_n_u_function_argument);
  __Pyx_GIVEREF(__pyx_n_u_function_argument);
  PyList_SET_ITEM(__pyx_t_5, 65, __pyx_n_u_function_argument);
  __Pyx_INCREF(__pyx_n_u_output_expression_list);
  __Pyx_GIVEREF(__pyx_n_u_output_expression_list);
  PyList_SET_ITEM(__pyx_t_5, 66, __pyx_n_u_output_expression_list);
  __Pyx_INCREF(__pyx_n_u_expression_list);
  __Pyx_GIVEREF(__pyx_n_u_expression_list);
  PyList_SET_ITEM(__pyx_t_5, 67, __pyx_n_u_expression_list);
  __Pyx_INCREF(__pyx_n_u_array_subscripts);
  __Pyx_GIVEREF(__pyx_n_u_array_subscripts);
  PyList_SET_ITEM(__pyx_t_5, 68, __pyx_n_u_array_subscripts);
  __Pyx_INCREF(__pyx_n_u_subscript);
  __Pyx_GIVEREF(__pyx_n_u_subscript);
  PyList_SET_ITEM(__pyx_t_5, 69, __pyx_n_u_subscript);
  __Pyx_INCREF(__pyx_n_u_comment);
  __Pyx_GIVEREF(__pyx_n_u_comment);
  PyList_SET_ITEM(__pyx_t_5, 70, __pyx_n_u_comment);
  __Pyx_INCREF(__pyx_n_u_string_comment);
  __Pyx_GIVEREF(__pyx_n_u_string_comment);
  PyList_SET_ITEM(__pyx_t_5, 71, __pyx_n_u_string_comment);
  __Pyx_INCREF(__pyx_n_u_annotation);
  __Pyx_GIVEREF(__pyx_n_u_annotation);
  PyList_SET_ITEM(__pyx_t_5, 72, __pyx_n_u_annotation);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_ruleNames, __pyx_t_5) < 0) __PYX_ERR(0, 658, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":681
 *                    "subscript", "comment", "string_comment", "annotation" ]
 * 
 *     EOF = Token.EOF             # <<<<<<<<<<<<<<
 *     T__0=1
 *     T__1=2
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_Token); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_EOF); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_EOF, __pyx_t_7) < 0) __PYX_ERR(0, 681, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":682
 * 
 *     EOF = Token.EOF
 *     T__0=1             # <<<<<<<<<<<<<<
 *     T__1=2
 *     T__2=3
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__0, __pyx_int_1) < 0) __PYX_ERR(0, 682, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":683
 *     EOF = Token.EOF
 *     T__0=1
 *     T__1=2             # <<<<<<<<<<<<<<
 *     T__2=3
 *     T__3=4
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__1, __pyx_int_2) < 0) __PYX_ERR(0, 683, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":684
 *     T__0=1
 *     T__1=2
 *     T__2=3             # <<<<<<<<<<<<<<
 *     T__3=4
 *     T__4=5
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__2, __pyx_int_3) < 0) __PYX_ERR(0, 684, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":685
 *     T__1=2
 *     T__2=3
 *     T__3=4             # <<<<<<<<<<<<<<
 *     T__4=5
 *     T__5=6
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__3, __pyx_int_4) < 0) __PYX_ERR(0, 685, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":686
 *     T__2=3
 *     T__3=4
 *     T__4=5             # <<<<<<<<<<<<<<
 *     T__5=6
 *     T__6=7
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__4, __pyx_int_5) < 0) __PYX_ERR(0, 686, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":687
 *     T__3=4
 *     T__4=5
 *     T__5=6             # <<<<<<<<<<<<<<
 *     T__6=7
 *     T__7=8
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__5, __pyx_int_6) < 0) __PYX_ERR(0, 687, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":688
 *     T__4=5
 *     T__5=6
 *     T__6=7             # <<<<<<<<<<<<<<
 *     T__7=8
 *     T__8=9
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__6, __pyx_int_7) < 0) __PYX_ERR(0, 688, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":689
 *     T__5=6
 *     T__6=7
 *     T__7=8             # <<<<<<<<<<<<<<
 *     T__8=9
 *     T__9=10
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__7, __pyx_int_8) < 0) __PYX_ERR(0, 689, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":690
 *     T__6=7
 *     T__7=8
 *     T__8=9             # <<<<<<<<<<<<<<
 *     T__9=10
 *     T__10=11
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__8, __pyx_int_9) < 0) __PYX_ERR(0, 690, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":691
 *     T__7=8
 *     T__8=9
 *     T__9=10             # <<<<<<<<<<<<<<
 *     T__10=11
 *     T__11=12
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__9, __pyx_int_10) < 0) __PYX_ERR(0, 691, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":692
 *     T__8=9
 *     T__9=10
 *     T__10=11             # <<<<<<<<<<<<<<
 *     T__11=12
 *     T__12=13
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__10, __pyx_int_11) < 0) __PYX_ERR(0, 692, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":693
 *     T__9=10
 *     T__10=11
 *     T__11=12             # <<<<<<<<<<<<<<
 *     T__12=13
 *     T__13=14
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__11, __pyx_int_12) < 0) __PYX_ERR(0, 693, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":694
 *     T__10=11
 *     T__11=12
 *     T__12=13             # <<<<<<<<<<<<<<
 *     T__13=14
 *     T__14=15
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__12, __pyx_int_13) < 0) __PYX_ERR(0, 694, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":695
 *     T__11=12
 *     T__12=13
 *     T__13=14             # <<<<<<<<<<<<<<
 *     T__14=15
 *     T__15=16
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__13, __pyx_int_14) < 0) __PYX_ERR(0, 695, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":696
 *     T__12=13
 *     T__13=14
 *     T__14=15             # <<<<<<<<<<<<<<
 *     T__15=16
 *     T__16=17
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__14, __pyx_int_15) < 0) __PYX_ERR(0, 696, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":697
 *     T__13=14
 *     T__14=15
 *     T__15=16             # <<<<<<<<<<<<<<
 *     T__16=17
 *     T__17=18
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__15, __pyx_int_16) < 0) __PYX_ERR(0, 697, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":698
 *     T__14=15
 *     T__15=16
 *     T__16=17             # <<<<<<<<<<<<<<
 *     T__17=18
 *     T__18=19
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__16, __pyx_int_17) < 0) __PYX_ERR(0, 698, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":699
 *     T__15=16
 *     T__16=17
 *     T__17=18             # <<<<<<<<<<<<<<
 *     T__18=19
 *     T__19=20
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__17, __pyx_int_18) < 0) __PYX_ERR(0, 699, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":700
 *     T__16=17
 *     T__17=18
 *     T__18=19             # <<<<<<<<<<<<<<
 *     T__19=20
 *     T__20=21
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__18, __pyx_int_19) < 0) __PYX_ERR(0, 700, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":701
 *     T__17=18
 *     T__18=19
 *     T__19=20             # <<<<<<<<<<<<<<
 *     T__20=21
 *     T__21=22
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__19, __pyx_int_20) < 0) __PYX_ERR(0, 701, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":702
 *     T__18=19
 *     T__19=20
 *     T__20=21             # <<<<<<<<<<<<<<
 *     T__21=22
 *     T__22=23
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__20, __pyx_int_21) < 0) __PYX_ERR(0, 702, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":703
 *     T__19=20
 *     T__20=21
 *     T__21=22             # <<<<<<<<<<<<<<
 *     T__22=23
 *     T__23=24
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__21, __pyx_int_22) < 0) __PYX_ERR(0, 703, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":704
 *     T__20=21
 *     T__21=22
 *     T__22=23             # <<<<<<<<<<<<<<
 *     T__23=24
 *     T__24=25
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__22, __pyx_int_23) < 0) __PYX_ERR(0, 704, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":705
 *     T__21=22
 *     T__22=23
 *     T__23=24             # <<<<<<<<<<<<<<
 *     T__24=25
 *     T__25=26
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__23, __pyx_int_24) < 0) __PYX_ERR(0, 705, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":706
 *     T__22=23
 *     T__23=24
 *     T__24=25             # <<<<<<<<<<<<<<
 *     T__25=26
 *     T__26=27
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__24, __pyx_int_25) < 0) __PYX_ERR(0, 706, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":707
 *     T__23=24
 *     T__24=25
 *     T__25=26             # <<<<<<<<<<<<<<
 *     T__26=27
 *     T__27=28
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__25, __pyx_int_26) < 0) __PYX_ERR(0, 707, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":708
 *     T__24=25
 *     T__25=26
 *     T__26=27             # <<<<<<<<<<<<<<
 *     T__27=28
 *     T__28=29
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__26, __pyx_int_27) < 0) __PYX_ERR(0, 708, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":709
 *     T__25=26
 *     T__26=27
 *     T__27=28             # <<<<<<<<<<<<<<
 *     T__28=29
 *     T__29=30
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__27, __pyx_int_28) < 0) __PYX_ERR(0, 709, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":710
 *     T__26=27
 *     T__27=28
 *     T__28=29             # <<<<<<<<<<<<<<
 *     T__29=30
 *     T__30=31
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__28, __pyx_int_29) < 0) __PYX_ERR(0, 710, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":711
 *     T__27=28
 *     T__28=29
 *     T__29=30             # <<<<<<<<<<<<<<
 *     T__30=31
 *     T__31=32
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__29, __pyx_int_30) < 0) __PYX_ERR(0, 711, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":712
 *     T__28=29
 *     T__29=30
 *     T__30=31             # <<<<<<<<<<<<<<
 *     T__31=32
 *     T__32=33
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__30, __pyx_int_31) < 0) __PYX_ERR(0, 712, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":713
 *     T__29=30
 *     T__30=31
 *     T__31=32             # <<<<<<<<<<<<<<
 *     T__32=33
 *     T__33=34
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__31, __pyx_int_32) < 0) __PYX_ERR(0, 713, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":714
 *     T__30=31
 *     T__31=32
 *     T__32=33             # <<<<<<<<<<<<<<
 *     T__33=34
 *     T__34=35
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__32, __pyx_int_33) < 0) __PYX_ERR(0, 714, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":715
 *     T__31=32
 *     T__32=33
 *     T__33=34             # <<<<<<<<<<<<<<
 *     T__34=35
 *     T__35=36
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__33, __pyx_int_34) < 0) __PYX_ERR(0, 715, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":716
 *     T__32=33
 *     T__33=34
 *     T__34=35             # <<<<<<<<<<<<<<
 *     T__35=36
 *     T__36=37
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__34, __pyx_int_35) < 0) __PYX_ERR(0, 716, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":717
 *     T__33=34
 *     T__34=35
 *     T__35=36             # <<<<<<<<<<<<<<
 *     T__36=37
 *     T__37=38
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__35, __pyx_int_36) < 0) __PYX_ERR(0, 717, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":718
 *     T__34=35
 *     T__35=36
 *     T__36=37             # <<<<<<<<<<<<<<
 *     T__37=38
 *     T__38=39
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__36, __pyx_int_37) < 0) __PYX_ERR(0, 718, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":719
 *     T__35=36
 *     T__36=37
 *     T__37=38             # <<<<<<<<<<<<<<
 *     T__38=39
 *     T__39=40
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__37, __pyx_int_38) < 0) __PYX_ERR(0, 719, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":720
 *     T__36=37
 *     T__37=38
 *     T__38=39             # <<<<<<<<<<<<<<
 *     T__39=40
 *     T__40=41
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__38, __pyx_int_39) < 0) __PYX_ERR(0, 720, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":721
 *     T__37=38
 *     T__38=39
 *     T__39=40             # <<<<<<<<<<<<<<
 *     T__40=41
 *     T__41=42
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__39, __pyx_int_40) < 0) __PYX_ERR(0, 721, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":722
 *     T__38=39
 *     T__39=40
 *     T__40=41             # <<<<<<<<<<<<<<
 *     T__41=42
 *     T__42=43
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__40, __pyx_int_41) < 0) __PYX_ERR(0, 722, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":723
 *     T__39=40
 *     T__40=41
 *     T__41=42             # <<<<<<<<<<<<<<
 *     T__42=43
 *     T__43=44
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__41, __pyx_int_42) < 0) __PYX_ERR(0, 723, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":724
 *     T__40=41
 *     T__41=42
 *     T__42=43             # <<<<<<<<<<<<<<
 *     T__43=44
 *     T__44=45
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__42, __pyx_int_43) < 0) __PYX_ERR(0, 724, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":725
 *     T__41=42
 *     T__42=43
 *     T__43=44             # <<<<<<<<<<<<<<
 *     T__44=45
 *     T__45=46
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__43, __pyx_int_44) < 0) __PYX_ERR(0, 725, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":726
 *     T__42=43
 *     T__43=44
 *     T__44=45             # <<<<<<<<<<<<<<
 *     T__45=46
 *     T__46=47
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__44, __pyx_int_45) < 0) __PYX_ERR(0, 726, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":727
 *     T__43=44
 *     T__44=45
 *     T__45=46             # <<<<<<<<<<<<<<
 *     T__46=47
 *     T__47=48
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__45, __pyx_int_46) < 0) __PYX_ERR(0, 727, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":728
 *     T__44=45
 *     T__45=46
 *     T__46=47             # <<<<<<<<<<<<<<
 *     T__47=48
 *     T__48=49
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__46, __pyx_int_47) < 0) __PYX_ERR(0, 728, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":729
 *     T__45=46
 *     T__46=47
 *     T__47=48             # <<<<<<<<<<<<<<
 *     T__48=49
 *     T__49=50
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__47, __pyx_int_48) < 0) __PYX_ERR(0, 729, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":730
 *     T__46=47
 *     T__47=48
 *     T__48=49             # <<<<<<<<<<<<<<
 *     T__49=50
 *     T__50=51
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__48, __pyx_int_49) < 0) __PYX_ERR(0, 730, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":731
 *     T__47=48
 *     T__48=49
 *     T__49=50             # <<<<<<<<<<<<<<
 *     T__50=51
 *     T__51=52
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__49, __pyx_int_50) < 0) __PYX_ERR(0, 731, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":732
 *     T__48=49
 *     T__49=50
 *     T__50=51             # <<<<<<<<<<<<<<
 *     T__51=52
 *     T__52=53
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__50, __pyx_int_51) < 0) __PYX_ERR(0, 732, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":733
 *     T__49=50
 *     T__50=51
 *     T__51=52             # <<<<<<<<<<<<<<
 *     T__52=53
 *     T__53=54
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__51, __pyx_int_52) < 0) __PYX_ERR(0, 733, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":734
 *     T__50=51
 *     T__51=52
 *     T__52=53             # <<<<<<<<<<<<<<
 *     T__53=54
 *     T__54=55
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__52, __pyx_int_53) < 0) __PYX_ERR(0, 734, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":735
 *     T__51=52
 *     T__52=53
 *     T__53=54             # <<<<<<<<<<<<<<
 *     T__54=55
 *     T__55=56
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__53, __pyx_int_54) < 0) __PYX_ERR(0, 735, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":736
 *     T__52=53
 *     T__53=54
 *     T__54=55             # <<<<<<<<<<<<<<
 *     T__55=56
 *     T__56=57
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__54, __pyx_int_55) < 0) __PYX_ERR(0, 736, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":737
 *     T__53=54
 *     T__54=55
 *     T__55=56             # <<<<<<<<<<<<<<
 *     T__56=57
 *     T__57=58
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__55, __pyx_int_56) < 0) __PYX_ERR(0, 737, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":738
 *     T__54=55
 *     T__55=56
 *     T__56=57             # <<<<<<<<<<<<<<
 *     T__57=58
 *     T__58=59
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__56, __pyx_int_57) < 0) __PYX_ERR(0, 738, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":739
 *     T__55=56
 *     T__56=57
 *     T__57=58             # <<<<<<<<<<<<<<
 *     T__58=59
 *     T__59=60
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__57, __pyx_int_58) < 0) __PYX_ERR(0, 739, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":740
 *     T__56=57
 *     T__57=58
 *     T__58=59             # <<<<<<<<<<<<<<
 *     T__59=60
 *     T__60=61
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__58, __pyx_int_59) < 0) __PYX_ERR(0, 740, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":741
 *     T__57=58
 *     T__58=59
 *     T__59=60             # <<<<<<<<<<<<<<
 *     T__60=61
 *     T__61=62
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__59, __pyx_int_60) < 0) __PYX_ERR(0, 741, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":742
 *     T__58=59
 *     T__59=60
 *     T__60=61             # <<<<<<<<<<<<<<
 *     T__61=62
 *     T__62=63
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__60, __pyx_int_61) < 0) __PYX_ERR(0, 742, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":743
 *     T__59=60
 *     T__60=61
 *     T__61=62             # <<<<<<<<<<<<<<
 *     T__62=63
 *     T__63=64
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__61, __pyx_int_62) < 0) __PYX_ERR(0, 743, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":744
 *     T__60=61
 *     T__61=62
 *     T__62=63             # <<<<<<<<<<<<<<
 *     T__63=64
 *     T__64=65
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__62, __pyx_int_63) < 0) __PYX_ERR(0, 744, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":745
 *     T__61=62
 *     T__62=63
 *     T__63=64             # <<<<<<<<<<<<<<
 *     T__64=65
 *     T__65=66
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__63, __pyx_int_64) < 0) __PYX_ERR(0, 745, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":746
 *     T__62=63
 *     T__63=64
 *     T__64=65             # <<<<<<<<<<<<<<
 *     T__65=66
 *     T__66=67
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__64, __pyx_int_65) < 0) __PYX_ERR(0, 746, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":747
 *     T__63=64
 *     T__64=65
 *     T__65=66             # <<<<<<<<<<<<<<
 *     T__66=67
 *     T__67=68
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__65, __pyx_int_66) < 0) __PYX_ERR(0, 747, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":748
 *     T__64=65
 *     T__65=66
 *     T__66=67             # <<<<<<<<<<<<<<
 *     T__67=68
 *     T__68=69
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__66, __pyx_int_67) < 0) __PYX_ERR(0, 748, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":749
 *     T__65=66
 *     T__66=67
 *     T__67=68             # <<<<<<<<<<<<<<
 *     T__68=69
 *     T__69=70
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__67, __pyx_int_68) < 0) __PYX_ERR(0, 749, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":750
 *     T__66=67
 *     T__67=68
 *     T__68=69             # <<<<<<<<<<<<<<
 *     T__69=70
 *     T__70=71
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__68, __pyx_int_69) < 0) __PYX_ERR(0, 750, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":751
 *     T__67=68
 *     T__68=69
 *     T__69=70             # <<<<<<<<<<<<<<
 *     T__70=71
 *     T__71=72
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__69, __pyx_int_70) < 0) __PYX_ERR(0, 751, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":752
 *     T__68=69
 *     T__69=70
 *     T__70=71             # <<<<<<<<<<<<<<
 *     T__71=72
 *     T__72=73
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__70, __pyx_int_71) < 0) __PYX_ERR(0, 752, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":753
 *     T__69=70
 *     T__70=71
 *     T__71=72             # <<<<<<<<<<<<<<
 *     T__72=73
 *     T__73=74
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__71, __pyx_int_72) < 0) __PYX_ERR(0, 753, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":754
 *     T__70=71
 *     T__71=72
 *     T__72=73             # <<<<<<<<<<<<<<
 *     T__73=74
 *     T__74=75
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__72, __pyx_int_73) < 0) __PYX_ERR(0, 754, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":755
 *     T__71=72
 *     T__72=73
 *     T__73=74             # <<<<<<<<<<<<<<
 *     T__74=75
 *     T__75=76
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__73, __pyx_int_74) < 0) __PYX_ERR(0, 755, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":756
 *     T__72=73
 *     T__73=74
 *     T__74=75             # <<<<<<<<<<<<<<
 *     T__75=76
 *     T__76=77
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__74, __pyx_int_75) < 0) __PYX_ERR(0, 756, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":757
 *     T__73=74
 *     T__74=75
 *     T__75=76             # <<<<<<<<<<<<<<
 *     T__76=77
 *     T__77=78
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__75, __pyx_int_76) < 0) __PYX_ERR(0, 757, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":758
 *     T__74=75
 *     T__75=76
 *     T__76=77             # <<<<<<<<<<<<<<
 *     T__77=78
 *     EACH=79
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__76, __pyx_int_77) < 0) __PYX_ERR(0, 758, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":759
 *     T__75=76
 *     T__76=77
 *     T__77=78             # <<<<<<<<<<<<<<
 *     EACH=79
 *     PARTIAL=80
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_T__77, __pyx_int_78) < 0) __PYX_ERR(0, 759, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":760
 *     T__76=77
 *     T__77=78
 *     EACH=79             # <<<<<<<<<<<<<<
 *     PARTIAL=80
 *     FINAL=81
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_EACH, __pyx_int_79) < 0) __PYX_ERR(0, 760, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":761
 *     T__77=78
 *     EACH=79
 *     PARTIAL=80             # <<<<<<<<<<<<<<
 *     FINAL=81
 *     WITHIN=82
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_PARTIAL, __pyx_int_80) < 0) __PYX_ERR(0, 761, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":762
 *     EACH=79
 *     PARTIAL=80
 *     FINAL=81             # <<<<<<<<<<<<<<
 *     WITHIN=82
 *     ENCAPSULATED=83
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_FINAL, __pyx_int_81) < 0) __PYX_ERR(0, 762, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":763
 *     PARTIAL=80
 *     FINAL=81
 *     WITHIN=82             # <<<<<<<<<<<<<<
 *     ENCAPSULATED=83
 *     REDECLARE=84
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_WITHIN, __pyx_int_82) < 0) __PYX_ERR(0, 763, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":764
 *     FINAL=81
 *     WITHIN=82
 *     ENCAPSULATED=83             # <<<<<<<<<<<<<<
 *     REDECLARE=84
 *     INNER=85
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_ENCAPSULATED, __pyx_int_83) < 0) __PYX_ERR(0, 764, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":765
 *     WITHIN=82
 *     ENCAPSULATED=83
 *     REDECLARE=84             # <<<<<<<<<<<<<<
 *     INNER=85
 *     OUTER=86
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_REDECLARE, __pyx_int_84) < 0) __PYX_ERR(0, 765, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":766
 *     ENCAPSULATED=83
 *     REDECLARE=84
 *     INNER=85             # <<<<<<<<<<<<<<
 *     OUTER=86
 *     INITIAL=87
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_INNER, __pyx_int_85) < 0) __PYX_ERR(0, 766, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":767
 *     REDECLARE=84
 *     INNER=85
 *     OUTER=86             # <<<<<<<<<<<<<<
 *     INITIAL=87
 *     IDENT=88
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_OUTER, __pyx_int_86) < 0) __PYX_ERR(0, 767, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":768
 *     INNER=85
 *     OUTER=86
 *     INITIAL=87             # <<<<<<<<<<<<<<
 *     IDENT=88
 *     STRING=89
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_INITIAL, __pyx_int_87) < 0) __PYX_ERR(0, 768, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":769
 *     OUTER=86
 *     INITIAL=87
 *     IDENT=88             # <<<<<<<<<<<<<<
 *     STRING=89
 *     UNSIGNED_NUMBER=90
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_IDENT, __pyx_int_88) < 0) __PYX_ERR(0, 769, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":770
 *     INITIAL=87
 *     IDENT=88
 *     STRING=89             # <<<<<<<<<<<<<<
 *     UNSIGNED_NUMBER=90
 *     COMMENT=91
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_STRING, __pyx_int_89) < 0) __PYX_ERR(0, 770, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":771
 *     IDENT=88
 *     STRING=89
 *     UNSIGNED_NUMBER=90             # <<<<<<<<<<<<<<
 *     COMMENT=91
 *     WS=92
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_UNSIGNED_NUMBER, __pyx_int_90) < 0) __PYX_ERR(0, 771, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":772
 *     STRING=89
 *     UNSIGNED_NUMBER=90
 *     COMMENT=91             # <<<<<<<<<<<<<<
 *     WS=92
 * 
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_COMMENT, __pyx_int_91) < 0) __PYX_ERR(0, 772, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":773
 *     UNSIGNED_NUMBER=90
 *     COMMENT=91
 *     WS=92             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
 */
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_WS, __pyx_int_92) < 0) __PYX_ERR(0, 773, __pyx_L1_error)

  /* "pymola/generated/ModelicaParser.pyx":775
 *     WS=92
 * 
 *     def __init__(self, input:TokenStream, output:TextIO = sys.stdout):             # <<<<<<<<<<<<<<
 *         super().__init__(input, output)
 *         self.checkVersion("4.7")
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 775, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_TokenStream); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 775, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_input, __pyx_t_5) < 0) __PYX_ERR(0, 775, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_TextIO); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 775, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_output, __pyx_t_5) < 0) __PYX_ERR(0, 775, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_1__init__, 0, __pyx_n_s_ModelicaParser___init, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1007)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 775, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_t_5);
  PyList_Append(__pyx_t_4, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_5, sizeof(__pyx_defaults), 1)) __PYX_ERR(0, 775, __pyx_L1_error)
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_sys); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 775, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_stdout); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 775, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_t_5)->__pyx_arg_output = __pyx_t_6;
  __Pyx_GIVEREF(__pyx_t_6);
  __pyx_t_6 = 0;
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_5, __pyx_pf_6pymola_9generated_14ModelicaParser_2__defaults__);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_init, __pyx_t_5) < 0) __PYX_ERR(0, 775, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":783
 * 
 * 
 *     class Stored_definitionContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Stored_definitionContext, __pyx_n_s_ModelicaParser_Stored_definition, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_9 = PyList_New(0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "pymola/generated/ModelicaParser.pyx":785
 *     class Stored_definitionContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 785, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_14 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 785, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_parent, __pyx_t_14) < 0) __PYX_ERR(0, 785, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 785, __pyx_L1_error)
  __pyx_t_14 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_1__init__, 0, __pyx_n_s_ModelicaParser_Stored_definition_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1009)); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 785, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_INCREF(__pyx_t_14);
  PyList_Append(__pyx_t_9, __pyx_t_14);
  __Pyx_GIVEREF(__pyx_t_14);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_14, __pyx_tuple__1010);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_14, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_14) < 0) __PYX_ERR(0, 785, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

  /* "pymola/generated/ModelicaParser.pyx":789
 *             self.parser = parser
 * 
 *         def WITHIN(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.WITHIN, 0)
 * 
 */
  __pyx_t_14 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_3WITHIN, 0, __pyx_n_s_ModelicaParser_Stored_definition_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1012)); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 789, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_WITHIN, __pyx_t_14) < 0) __PYX_ERR(0, 789, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

  /* "pymola/generated/ModelicaParser.pyx":792
 *             return self.getToken(ModelicaParser.WITHIN, 0)
 * 
 *         def stored_definition_class(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Stored_definition_classContext)
 */
  __pyx_t_14 = PyDict_New(); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  if (PyDict_SetItem(__pyx_t_14, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 792, __pyx_L1_error)
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_5stored_definition_class, 0, __pyx_n_s_ModelicaParser_Stored_definition_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1014)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_8, __pyx_tuple__1015);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_14);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_stored_definition_class, __pyx_t_8) < 0) __PYX_ERR(0, 792, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaParser.pyx":799
 * 
 * 
 *         def component_reference(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 */
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_7component_reference, 0, __pyx_n_s_ModelicaParser_Stored_definition_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1017)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 799, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_component_reference, __pyx_t_8) < 0) __PYX_ERR(0, 799, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaParser.pyx":803
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_stored_definition
 * 
 */
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_9getRuleIndex, 0, __pyx_n_s_ModelicaParser_Stored_definition_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1019)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 803, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_8) < 0) __PYX_ERR(0, 803, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaParser.pyx":806
 *             return ModelicaParser.RULE_stored_definition
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStored_definition" ):
 *                 listener.enterStored_definition(self)
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 806, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_14 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 806, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_listener, __pyx_t_14) < 0) __PYX_ERR(0, 806, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __pyx_t_14 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_11enterRule, 0, __pyx_n_s_ModelicaParser_Stored_definition_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1021)); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 806, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_14, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_14) < 0) __PYX_ERR(0, 806, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

  /* "pymola/generated/ModelicaParser.pyx":810
 *                 listener.enterStored_definition(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStored_definition" ):
 *                 listener.exitStored_definition(self)
 */
  __pyx_t_14 = PyDict_New(); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_t_14, __pyx_n_s_listener, __pyx_t_8) < 0) __PYX_ERR(0, 810, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_13exitRule, 0, __pyx_n_s_ModelicaParser_Stored_definition_8, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1023)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_14);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_8) < 0) __PYX_ERR(0, 810, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaParser.pyx":814
 *                 listener.exitStored_definition(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStored_definition" ):
 *                 return visitor.visitStored_definition(self)
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 814, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_14 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 814, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_visitor, __pyx_t_14) < 0) __PYX_ERR(0, 814, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __pyx_t_14 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Stored_definitionContext_15accept, 0, __pyx_n_s_ModelicaParser_Stored_definition_9, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1025)); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 814, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_14, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_14) < 0) __PYX_ERR(0, 814, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

  /* "pymola/generated/ModelicaParser.pyx":783
 * 
 * 
 *     class Stored_definitionContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_14 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Stored_definitionContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_9, __pyx_t_14) < 0) __PYX_ERR(0, 783, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Stored_definitionContext, __pyx_t_14) < 0) __PYX_ERR(0, 783, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":823
 * 
 * 
 *     def stored_definition(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Stored_definitionContext(self, self._ctx, self.state)
 */
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_3stored_definition, 0, __pyx_n_s_ModelicaParser_stored_definition, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1027)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 823, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_stored_definition, __pyx_t_7) < 0) __PYX_ERR(0, 823, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":866
 *         return localctx
 * 
 *     class Stored_definition_classContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Stored_definition_classContext, __pyx_n_s_ModelicaParser_Stored_definition_10, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_14 = PyList_New(0); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);

  /* "pymola/generated/ModelicaParser.pyx":868
 *     class Stored_definition_classContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_12 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_parent, __pyx_t_12) < 0) __PYX_ERR(0, 868, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 868, __pyx_L1_error)
  __pyx_t_12 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_1__init__, 0, __pyx_n_s_ModelicaParser_Stored_definition_11, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1029)); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_INCREF(__pyx_t_12);
  PyList_Append(__pyx_t_14, __pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_12);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_12, __pyx_tuple__1030);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_12, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_12) < 0) __PYX_ERR(0, 868, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

  /* "pymola/generated/ModelicaParser.pyx":872
 *             self.parser = parser
 * 
 *         def class_definition(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_definitionContext,0)
 * 
 */
  __pyx_t_12 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_3class_definition, 0, __pyx_n_s_ModelicaParser_Stored_definition_12, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1032)); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 872, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_class_definition, __pyx_t_12) < 0) __PYX_ERR(0, 872, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

  /* "pymola/generated/ModelicaParser.pyx":876
 * 
 * 
 *         def FINAL(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.FINAL, 0)
 * 
 */
  __pyx_t_12 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_5FINAL, 0, __pyx_n_s_ModelicaParser_Stored_definition_13, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1034)); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 876, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_FINAL, __pyx_t_12) < 0) __PYX_ERR(0, 876, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

  /* "pymola/generated/ModelicaParser.pyx":879
 *             return self.getToken(ModelicaParser.FINAL, 0)
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_stored_definition_class
 * 
 */
  __pyx_t_12 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_7getRuleIndex, 0, __pyx_n_s_ModelicaParser_Stored_definition_14, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1036)); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 879, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_12) < 0) __PYX_ERR(0, 879, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

  /* "pymola/generated/ModelicaParser.pyx":882
 *             return ModelicaParser.RULE_stored_definition_class
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStored_definition_class" ):
 *                 listener.enterStored_definition_class(self)
 */
  __pyx_t_12 = PyDict_New(); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 882, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 882, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_t_12, __pyx_n_s_listener, __pyx_t_8) < 0) __PYX_ERR(0, 882, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_9enterRule, 0, __pyx_n_s_ModelicaParser_Stored_definition_15, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1038)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 882, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_12);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_8) < 0) __PYX_ERR(0, 882, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaParser.pyx":886
 *                 listener.enterStored_definition_class(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStored_definition_class" ):
 *                 listener.exitStored_definition_class(self)
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 886, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_12 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 886, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_listener, __pyx_t_12) < 0) __PYX_ERR(0, 886, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_t_12 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_11exitRule, 0, __pyx_n_s_ModelicaParser_Stored_definition_16, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1040)); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 886, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_12, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_12) < 0) __PYX_ERR(0, 886, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

  /* "pymola/generated/ModelicaParser.pyx":890
 *                 listener.exitStored_definition_class(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStored_definition_class" ):
 *                 return visitor.visitStored_definition_class(self)
 */
  __pyx_t_12 = PyDict_New(); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 890, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 890, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_t_12, __pyx_n_s_visitor, __pyx_t_8) < 0) __PYX_ERR(0, 890, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_30Stored_definition_classContext_13accept, 0, __pyx_n_s_ModelicaParser_Stored_definition_17, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1042)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 890, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_12);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_8) < 0) __PYX_ERR(0, 890, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaParser.pyx":866
 *         return localctx
 * 
 *     class Stored_definition_classContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_8 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Stored_definition_classContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_14, __pyx_t_8) < 0) __PYX_ERR(0, 866, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Stored_definition_classContext, __pyx_t_8) < 0) __PYX_ERR(0, 866, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":899
 * 
 * 
 *     def stored_definition_class(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Stored_definition_classContext(self, self._ctx, self.state)
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_5stored_definition_class, 0, __pyx_n_s_ModelicaParser_stored_definition_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1044)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 899, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_stored_definition_class, __pyx_t_5) < 0) __PYX_ERR(0, 899, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":926
 *         return localctx
 * 
 *     class Class_definitionContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 926, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 926, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 926, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Class_definitionContext, __pyx_n_s_ModelicaParser_Class_definitionC, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 926, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_8 = PyList_New(0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 926, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);

  /* "pymola/generated/ModelicaParser.pyx":928
 *     class Class_definitionContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_12 = PyDict_New(); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 928, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_15 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 928, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  if (PyDict_SetItem(__pyx_t_12, __pyx_n_s_parent, __pyx_t_15) < 0) __PYX_ERR(0, 928, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (PyDict_SetItem(__pyx_t_12, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 928, __pyx_L1_error)
  __pyx_t_15 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Class_definitionContext_1__init__, 0, __pyx_n_s_ModelicaParser_Class_definitionC_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1046)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 928, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_INCREF(__pyx_t_15);
  PyList_Append(__pyx_t_8, __pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_15);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_15, __pyx_tuple__1047);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_15, __pyx_t_12);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_15) < 0) __PYX_ERR(0, 928, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "pymola/generated/ModelicaParser.pyx":932
 *             self.parser = parser
 * 
 *         def class_prefixes(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_prefixesContext,0)
 * 
 */
  __pyx_t_15 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Class_definitionContext_3class_prefixes, 0, __pyx_n_s_ModelicaParser_Class_definitionC_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1049)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 932, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_class_prefixes, __pyx_t_15) < 0) __PYX_ERR(0, 932, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "pymola/generated/ModelicaParser.pyx":936
 * 
 * 
 *         def class_specifier(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_specifierContext,0)
 * 
 */
  __pyx_t_15 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Class_definitionContext_5class_specifier, 0, __pyx_n_s_ModelicaParser_Class_definitionC_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1051)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 936, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_class_specifier, __pyx_t_15) < 0) __PYX_ERR(0, 936, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "pymola/generated/ModelicaParser.pyx":940
 * 
 * 
 *         def ENCAPSULATED(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.ENCAPSULATED, 0)
 * 
 */
  __pyx_t_15 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Class_definitionContext_7ENCAPSULATED, 0, __pyx_n_s_ModelicaParser_Class_definitionC_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1053)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 940, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_ENCAPSULATED, __pyx_t_15) < 0) __PYX_ERR(0, 940, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "pymola/generated/ModelicaParser.pyx":943
 *             return self.getToken(ModelicaParser.ENCAPSULATED, 0)
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_class_definition
 * 
 */
  __pyx_t_15 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Class_definitionContext_9getRuleIndex, 0, __pyx_n_s_ModelicaParser_Class_definitionC_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1055)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 943, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_15) < 0) __PYX_ERR(0, 943, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "pymola/generated/ModelicaParser.pyx":946
 *             return ModelicaParser.RULE_class_definition
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterClass_definition" ):
 *                 listener.enterClass_definition(self)
 */
  __pyx_t_15 = PyDict_New(); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_12 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  if (PyDict_SetItem(__pyx_t_15, __pyx_n_s_listener, __pyx_t_12) < 0) __PYX_ERR(0, 946, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_t_12 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Class_definitionContext_11enterRule, 0, __pyx_n_s_ModelicaParser_Class_definitionC_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1057)); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_12, __pyx_t_15);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_12) < 0) __PYX_ERR(0, 946, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

  /* "pymola/generated/ModelicaParser.pyx":950
 *                 listener.enterClass_definition(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitClass_definition" ):
 *                 listener.exitClass_definition(self)
 */
  __pyx_t_12 = PyDict_New(); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 950, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_15 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 950, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  if (PyDict_SetItem(__pyx_t_12, __pyx_n_s_listener, __pyx_t_15) < 0) __PYX_ERR(0, 950, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Class_definitionContext_13exitRule, 0, __pyx_n_s_ModelicaParser_Class_definitionC_8, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1059)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 950, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_15, __pyx_t_12);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_15) < 0) __PYX_ERR(0, 950, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "pymola/generated/ModelicaParser.pyx":954
 *                 listener.exitClass_definition(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitClass_definition" ):
 *                 return visitor.visitClass_definition(self)
 */
  __pyx_t_15 = PyDict_New(); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 954, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_12 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 954, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  if (PyDict_SetItem(__pyx_t_15, __pyx_n_s_visitor, __pyx_t_12) < 0) __PYX_ERR(0, 954, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_t_12 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Class_definitionContext_15accept, 0, __pyx_n_s_ModelicaParser_Class_definitionC_9, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1061)); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 954, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_12, __pyx_t_15);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_12) < 0) __PYX_ERR(0, 954, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

  /* "pymola/generated/ModelicaParser.pyx":926
 *         return localctx
 * 
 *     class Class_definitionContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_12 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Class_definitionContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 926, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_8, __pyx_t_12) < 0) __PYX_ERR(0, 926, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Class_definitionContext, __pyx_t_12) < 0) __PYX_ERR(0, 926, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":963
 * 
 * 
 *     def class_definition(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Class_definitionContext(self, self._ctx, self.state)
 */
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_7class_definition, 0, __pyx_n_s_ModelicaParser_class_definition, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1063)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 963, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_class_definition, __pyx_t_7) < 0) __PYX_ERR(0, 963, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":990
 *         return localctx
 * 
 *     class Class_prefixesContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Class_prefixesContext, __pyx_n_s_ModelicaParser_Class_prefixesCon, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_12 = PyList_New(0); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);

  /* "pymola/generated/ModelicaParser.pyx":992
 *     class Class_prefixesContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_15 = PyDict_New(); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 992, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_16 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 992, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  if (PyDict_SetItem(__pyx_t_15, __pyx_n_s_parent, __pyx_t_16) < 0) __PYX_ERR(0, 992, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  if (PyDict_SetItem(__pyx_t_15, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 992, __pyx_L1_error)
  __pyx_t_16 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_1__init__, 0, __pyx_n_s_ModelicaParser_Class_prefixesCon_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1065)); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 992, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_INCREF(__pyx_t_16);
  PyList_Append(__pyx_t_12, __pyx_t_16);
  __Pyx_GIVEREF(__pyx_t_16);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_16, __pyx_tuple__1066);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_16, __pyx_t_15);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_16) < 0) __PYX_ERR(0, 992, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "pymola/generated/ModelicaParser.pyx":996
 *             self.parser = parser
 * 
 *         def class_type(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_typeContext,0)
 * 
 */
  __pyx_t_16 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_3class_type, 0, __pyx_n_s_ModelicaParser_Class_prefixesCon_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1068)); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 996, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_class_type, __pyx_t_16) < 0) __PYX_ERR(0, 996, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1000
 * 
 * 
 *         def PARTIAL(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.PARTIAL, 0)
 * 
 */
  __pyx_t_16 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_5PARTIAL, 0, __pyx_n_s_ModelicaParser_Class_prefixesCon_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1070)); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1000, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_PARTIAL, __pyx_t_16) < 0) __PYX_ERR(0, 1000, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1003
 *             return self.getToken(ModelicaParser.PARTIAL, 0)
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_class_prefixes
 * 
 */
  __pyx_t_16 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_7getRuleIndex, 0, __pyx_n_s_ModelicaParser_Class_prefixesCon_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1072)); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1003, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_16) < 0) __PYX_ERR(0, 1003, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1006
 *             return ModelicaParser.RULE_class_prefixes
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterClass_prefixes" ):
 *                 listener.enterClass_prefixes(self)
 */
  __pyx_t_16 = PyDict_New(); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_15 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  if (PyDict_SetItem(__pyx_t_16, __pyx_n_s_listener, __pyx_t_15) < 0) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_9enterRule, 0, __pyx_n_s_ModelicaParser_Class_prefixesCon_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1074)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_15, __pyx_t_16);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_15) < 0) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1010
 *                 listener.enterClass_prefixes(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitClass_prefixes" ):
 *                 listener.exitClass_prefixes(self)
 */
  __pyx_t_15 = PyDict_New(); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1010, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_16 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1010, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  if (PyDict_SetItem(__pyx_t_15, __pyx_n_s_listener, __pyx_t_16) < 0) __PYX_ERR(0, 1010, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __pyx_t_16 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_11exitRule, 0, __pyx_n_s_ModelicaParser_Class_prefixesCon_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1076)); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1010, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_16, __pyx_t_15);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_16) < 0) __PYX_ERR(0, 1010, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1014
 *                 listener.exitClass_prefixes(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitClass_prefixes" ):
 *                 return visitor.visitClass_prefixes(self)
 */
  __pyx_t_16 = PyDict_New(); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_15 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  if (PyDict_SetItem(__pyx_t_16, __pyx_n_s_visitor, __pyx_t_15) < 0) __PYX_ERR(0, 1014, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Class_prefixesContext_13accept, 0, __pyx_n_s_ModelicaParser_Class_prefixesCon_8, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1078)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_15, __pyx_t_16);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_15) < 0) __PYX_ERR(0, 1014, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "pymola/generated/ModelicaParser.pyx":990
 *         return localctx
 * 
 *     class Class_prefixesContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_15 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Class_prefixesContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_12, __pyx_t_15) < 0) __PYX_ERR(0, 990, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Class_prefixesContext, __pyx_t_15) < 0) __PYX_ERR(0, 990, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1023
 * 
 * 
 *     def class_prefixes(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Class_prefixesContext(self, self._ctx, self.state)
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_9class_prefixes, 0, __pyx_n_s_ModelicaParser_class_prefixes, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1080)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1023, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_class_prefixes, __pyx_t_5) < 0) __PYX_ERR(0, 1023, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1048
 *         return localctx
 * 
 *     class Class_typeContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1048, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1048, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1048, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Class_typeContext, __pyx_n_s_ModelicaParser_Class_typeContext, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1048, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_15 = PyList_New(0); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1048, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);

  /* "pymola/generated/ModelicaParser.pyx":1050
 *     class Class_typeContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_16 = PyDict_New(); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1050, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_17 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1050, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  if (PyDict_SetItem(__pyx_t_16, __pyx_n_s_parent, __pyx_t_17) < 0) __PYX_ERR(0, 1050, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  if (PyDict_SetItem(__pyx_t_16, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 1050, __pyx_L1_error)
  __pyx_t_17 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_17Class_typeContext_1__init__, 0, __pyx_n_s_ModelicaParser_Class_typeContext_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1082)); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1050, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_INCREF(__pyx_t_17);
  PyList_Append(__pyx_t_15, __pyx_t_17);
  __Pyx_GIVEREF(__pyx_t_17);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_17, __pyx_tuple__1083);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_17, __pyx_t_16);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_17) < 0) __PYX_ERR(0, 1050, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1055
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_class_type
 * 
 */
  __pyx_t_17 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_17Class_typeContext_3getRuleIndex, 0, __pyx_n_s_ModelicaParser_Class_typeContext_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1085)); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1055, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_17) < 0) __PYX_ERR(0, 1055, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1058
 *             return ModelicaParser.RULE_class_type
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterClass_type" ):
 *                 listener.enterClass_type(self)
 */
  __pyx_t_17 = PyDict_New(); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1058, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __pyx_t_16 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1058, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  if (PyDict_SetItem(__pyx_t_17, __pyx_n_s_listener, __pyx_t_16) < 0) __PYX_ERR(0, 1058, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __pyx_t_16 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_17Class_typeContext_5enterRule, 0, __pyx_n_s_ModelicaParser_Class_typeContext_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1087)); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1058, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_16, __pyx_t_17);
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_16) < 0) __PYX_ERR(0, 1058, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1062
 *                 listener.enterClass_type(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitClass_type" ):
 *                 listener.exitClass_type(self)
 */
  __pyx_t_16 = PyDict_New(); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1062, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_17 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1062, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  if (PyDict_SetItem(__pyx_t_16, __pyx_n_s_listener, __pyx_t_17) < 0) __PYX_ERR(0, 1062, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __pyx_t_17 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_17Class_typeContext_7exitRule, 0, __pyx_n_s_ModelicaParser_Class_typeContext_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1089)); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1062, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_17, __pyx_t_16);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_17) < 0) __PYX_ERR(0, 1062, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1066
 *                 listener.exitClass_type(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitClass_type" ):
 *                 return visitor.visitClass_type(self)
 */
  __pyx_t_17 = PyDict_New(); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1066, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __pyx_t_16 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1066, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  if (PyDict_SetItem(__pyx_t_17, __pyx_n_s_visitor, __pyx_t_16) < 0) __PYX_ERR(0, 1066, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __pyx_t_16 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_17Class_typeContext_9accept, 0, __pyx_n_s_ModelicaParser_Class_typeContext_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1091)); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1066, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_16, __pyx_t_17);
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_16) < 0) __PYX_ERR(0, 1066, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1048
 *         return localctx
 * 
 *     class Class_typeContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_16 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Class_typeContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1048, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_15, __pyx_t_16) < 0) __PYX_ERR(0, 1048, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Class_typeContext, __pyx_t_16) < 0) __PYX_ERR(0, 1048, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1075
 * 
 * 
 *     def class_type(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Class_typeContext(self, self._ctx, self.state)
 */
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_11class_type, 0, __pyx_n_s_ModelicaParser_class_type, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1093)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1075, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_class_type, __pyx_t_7) < 0) __PYX_ERR(0, 1075, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1184
 *         return localctx
 * 
 *     class Class_specifierContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Class_specifierContext, __pyx_n_s_ModelicaParser_Class_specifierCo, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_16 = PyList_New(0); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);

  /* "pymola/generated/ModelicaParser.pyx":1186
 *     class Class_specifierContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_17 = PyDict_New(); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __pyx_t_18 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  if (PyDict_SetItem(__pyx_t_17, __pyx_n_s_parent, __pyx_t_18) < 0) __PYX_ERR(0, 1186, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  if (PyDict_SetItem(__pyx_t_17, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 1186, __pyx_L1_error)
  __pyx_t_18 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_specifierContext_1__init__, 0, __pyx_n_s_ModelicaParser_Class_specifierCo_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1095)); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_INCREF(__pyx_t_18);
  PyList_Append(__pyx_t_16, __pyx_t_18);
  __Pyx_GIVEREF(__pyx_t_18);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_18, __pyx_tuple__1096);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_18, __pyx_t_17);
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_18) < 0) __PYX_ERR(0, 1186, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1191
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_class_specifier
 * 
 */
  __pyx_t_18 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_specifierContext_3getRuleIndex, 0, __pyx_n_s_ModelicaParser_Class_specifierCo_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1098)); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_18) < 0) __PYX_ERR(0, 1191, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1195
 * 
 * 
 *         def copyFrom(self, ctx:ParserRuleContext):             # <<<<<<<<<<<<<<
 *             super().copyFrom(ctx)
 * 
 */
  __pyx_t_18 = PyDict_New(); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_17 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  if (PyDict_SetItem(__pyx_t_18, __pyx_n_s_ctx_2, __pyx_t_17) < 0) __PYX_ERR(0, 1195, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __pyx_t_17 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_specifierContext_5copyFrom, 0, __pyx_n_s_ModelicaParser_Class_specifierCo_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1100)); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_INCREF(__pyx_t_17);
  PyList_Append(__pyx_t_16, __pyx_t_17);
  __Pyx_GIVEREF(__pyx_t_17);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_17, __pyx_t_18);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_copyFrom, __pyx_t_17) < 0) __PYX_ERR(0, 1195, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1184
 *         return localctx
 * 
 *     class Class_specifierContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_17 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Class_specifierContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_16, __pyx_t_17) < 0) __PYX_ERR(0, 1184, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Class_specifierContext, __pyx_t_17) < 0) __PYX_ERR(0, 1184, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1200
 * 
 * 
 *     class Class_spec_derContext(Class_specifierContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Class_specifierContext
 */
  __pyx_t_5 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_Class_specifierContext);
  if (unlikely(!__pyx_t_5)) {
    PyErr_Clear();
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_Class_specifierContext);
  }
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Class_spec_derContext, __pyx_n_s_ModelicaParser_Class_spec_derCon, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_17 = PyList_New(0); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);

  /* "pymola/generated/ModelicaParser.pyx":1202
 *     class Class_spec_derContext(Class_specifierContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Class_specifierContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_18 = PyDict_New(); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_19 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  if (PyDict_SetItem(__pyx_t_18, __pyx_n_s_ctx_2, __pyx_t_19) < 0) __PYX_ERR(0, 1202, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __pyx_t_19 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_1__init__, 0, __pyx_n_s_ModelicaParser_Class_spec_derCon_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1102)); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_INCREF(__pyx_t_19);
  PyList_Append(__pyx_t_17, __pyx_t_19);
  __Pyx_GIVEREF(__pyx_t_19);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_19, __pyx_t_18);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_19) < 0) __PYX_ERR(0, 1202, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1206
 *             self.copyFrom(ctx)
 * 
 *         def IDENT(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTokens(ModelicaParser.IDENT)
 */
  __pyx_t_19 = PyDict_New(); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  if (PyDict_SetItem(__pyx_t_19, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 1206, __pyx_L1_error)
  __pyx_t_18 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_3IDENT, 0, __pyx_n_s_ModelicaParser_Class_spec_derCon_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1104)); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_18, __pyx_tuple__1105);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_18, __pyx_t_19);
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_IDENT, __pyx_t_18) < 0) __PYX_ERR(0, 1206, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1211
 *             else:
 *                 return self.getToken(ModelicaParser.IDENT, i)
 *         def name(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)
 * 
 */
  __pyx_t_18 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_5name, 0, __pyx_n_s_ModelicaParser_Class_spec_derCon_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1107)); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_name, __pyx_t_18) < 0) __PYX_ERR(0, 1211, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1214
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)
 * 
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */
  __pyx_t_18 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_7comment, 0, __pyx_n_s_ModelicaParser_Class_spec_derCon_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1109)); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_comment, __pyx_t_18) < 0) __PYX_ERR(0, 1214, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1218
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterClass_spec_der" ):
 *                 listener.enterClass_spec_der(self)
 */
  __pyx_t_18 = PyDict_New(); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_19 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  if (PyDict_SetItem(__pyx_t_18, __pyx_n_s_listener, __pyx_t_19) < 0) __PYX_ERR(0, 1218, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __pyx_t_19 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_9enterRule, 0, __pyx_n_s_ModelicaParser_Class_spec_derCon_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1111)); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_19, __pyx_t_18);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_19) < 0) __PYX_ERR(0, 1218, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1222
 *                 listener.enterClass_spec_der(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitClass_spec_der" ):
 *                 listener.exitClass_spec_der(self)
 */
  __pyx_t_19 = PyDict_New(); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __pyx_t_18 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  if (PyDict_SetItem(__pyx_t_19, __pyx_n_s_listener, __pyx_t_18) < 0) __PYX_ERR(0, 1222, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  __pyx_t_18 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_11exitRule, 0, __pyx_n_s_ModelicaParser_Class_spec_derCon_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1113)); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_18, __pyx_t_19);
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_18) < 0) __PYX_ERR(0, 1222, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1226
 *                 listener.exitClass_spec_der(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitClass_spec_der" ):
 *                 return visitor.visitClass_spec_der(self)
 */
  __pyx_t_18 = PyDict_New(); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_19 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  if (PyDict_SetItem(__pyx_t_18, __pyx_n_s_visitor, __pyx_t_19) < 0) __PYX_ERR(0, 1226, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __pyx_t_19 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Class_spec_derContext_13accept, 0, __pyx_n_s_ModelicaParser_Class_spec_derCon_8, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1115)); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_19, __pyx_t_18);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_19) < 0) __PYX_ERR(0, 1226, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1200
 * 
 * 
 *     class Class_spec_derContext(Class_specifierContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Class_specifierContext
 */
  __pyx_t_19 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Class_spec_derContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_17, __pyx_t_19) < 0) __PYX_ERR(0, 1200, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Class_spec_derContext, __pyx_t_19) < 0) __PYX_ERR(0, 1200, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1233
 * 
 * 
 *     class Class_spec_enumContext(Class_specifierContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Class_specifierContext
 */
  __pyx_t_7 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_Class_specifierContext);
  if (unlikely(!__pyx_t_7)) {
    PyErr_Clear();
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_Class_specifierContext);
  }
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Class_spec_enumContext, __pyx_n_s_ModelicaParser_Class_spec_enumCo, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_19 = PyList_New(0); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);

  /* "pymola/generated/ModelicaParser.pyx":1235
 *     class Class_spec_enumContext(Class_specifierContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Class_specifierContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_18 = PyDict_New(); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_20 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  if (PyDict_SetItem(__pyx_t_18, __pyx_n_s_ctx_2, __pyx_t_20) < 0) __PYX_ERR(0, 1235, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __pyx_t_20 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_1__init__, 0, __pyx_n_s_ModelicaParser_Class_spec_enumCo_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1117)); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_INCREF(__pyx_t_20);
  PyList_Append(__pyx_t_19, __pyx_t_20);
  __Pyx_GIVEREF(__pyx_t_20);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_20, __pyx_t_18);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_20) < 0) __PYX_ERR(0, 1235, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1239
 *             self.copyFrom(ctx)
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 *         def comment(self):
 */
  __pyx_t_20 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_3IDENT, 0, __pyx_n_s_ModelicaParser_Class_spec_enumCo_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1119)); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_IDENT, __pyx_t_20) < 0) __PYX_ERR(0, 1239, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1241
 *         def IDENT(self):
 *             return self.getToken(ModelicaParser.IDENT, 0)
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */
  __pyx_t_20 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_5comment, 0, __pyx_n_s_ModelicaParser_Class_spec_enumCo_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1121)); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_comment, __pyx_t_20) < 0) __PYX_ERR(0, 1241, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1244
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 *         def enum_list(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Enum_listContext,0)
 * 
 */
  __pyx_t_20 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_7enum_list, 0, __pyx_n_s_ModelicaParser_Class_spec_enumCo_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1123)); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enum_list, __pyx_t_20) < 0) __PYX_ERR(0, 1244, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1248
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterClass_spec_enum" ):
 *                 listener.enterClass_spec_enum(self)
 */
  __pyx_t_20 = PyDict_New(); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1248, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_18 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1248, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  if (PyDict_SetItem(__pyx_t_20, __pyx_n_s_listener, __pyx_t_18) < 0) __PYX_ERR(0, 1248, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  __pyx_t_18 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_9enterRule, 0, __pyx_n_s_ModelicaParser_Class_spec_enumCo_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1125)); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1248, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_18, __pyx_t_20);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_18) < 0) __PYX_ERR(0, 1248, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1252
 *                 listener.enterClass_spec_enum(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitClass_spec_enum" ):
 *                 listener.exitClass_spec_enum(self)
 */
  __pyx_t_18 = PyDict_New(); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_20 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  if (PyDict_SetItem(__pyx_t_18, __pyx_n_s_listener, __pyx_t_20) < 0) __PYX_ERR(0, 1252, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __pyx_t_20 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_11exitRule, 0, __pyx_n_s_ModelicaParser_Class_spec_enumCo_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1127)); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_20, __pyx_t_18);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_20) < 0) __PYX_ERR(0, 1252, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1256
 *                 listener.exitClass_spec_enum(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitClass_spec_enum" ):
 *                 return visitor.visitClass_spec_enum(self)
 */
  __pyx_t_20 = PyDict_New(); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_18 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  if (PyDict_SetItem(__pyx_t_20, __pyx_n_s_visitor, __pyx_t_18) < 0) __PYX_ERR(0, 1256, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  __pyx_t_18 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_enumContext_13accept, 0, __pyx_n_s_ModelicaParser_Class_spec_enumCo_8, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1129)); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_18, __pyx_t_20);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_18) < 0) __PYX_ERR(0, 1256, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1233
 * 
 * 
 *     class Class_spec_enumContext(Class_specifierContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Class_specifierContext
 */
  __pyx_t_18 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Class_spec_enumContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_19, __pyx_t_18) < 0) __PYX_ERR(0, 1233, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Class_spec_enumContext, __pyx_t_18) < 0) __PYX_ERR(0, 1233, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1263
 * 
 * 
 *     class Class_spec_baseContext(Class_specifierContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Class_specifierContext
 */
  __pyx_t_5 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_Class_specifierContext);
  if (unlikely(!__pyx_t_5)) {
    PyErr_Clear();
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_Class_specifierContext);
  }
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Class_spec_baseContext, __pyx_n_s_ModelicaParser_Class_spec_baseCo, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_18 = PyList_New(0); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);

  /* "pymola/generated/ModelicaParser.pyx":1265
 *     class Class_spec_baseContext(Class_specifierContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Class_specifierContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_20 = PyDict_New(); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_21 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  if (PyDict_SetItem(__pyx_t_20, __pyx_n_s_ctx_2, __pyx_t_21) < 0) __PYX_ERR(0, 1265, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __pyx_t_21 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_1__init__, 0, __pyx_n_s_ModelicaParser_Class_spec_baseCo_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1131)); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_INCREF(__pyx_t_21);
  PyList_Append(__pyx_t_18, __pyx_t_21);
  __Pyx_GIVEREF(__pyx_t_21);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_21, __pyx_t_20);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_21) < 0) __PYX_ERR(0, 1265, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1269
 *             self.copyFrom(ctx)
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 *         def base_prefix(self):
 */
  __pyx_t_21 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_3IDENT, 0, __pyx_n_s_ModelicaParser_Class_spec_baseCo_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1133)); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_IDENT, __pyx_t_21) < 0) __PYX_ERR(0, 1269, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1271
 *         def IDENT(self):
 *             return self.getToken(ModelicaParser.IDENT, 0)
 *         def base_prefix(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Base_prefixContext,0)
 * 
 */
  __pyx_t_21 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_5base_prefix, 0, __pyx_n_s_ModelicaParser_Class_spec_baseCo_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1135)); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_base_prefix, __pyx_t_21) < 0) __PYX_ERR(0, 1271, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1274
 *             return self.getTypedRuleContext(ModelicaParser.Base_prefixContext,0)
 * 
 *         def name(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)
 * 
 */
  __pyx_t_21 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_7name, 0, __pyx_n_s_ModelicaParser_Class_spec_baseCo_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1137)); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_name, __pyx_t_21) < 0) __PYX_ERR(0, 1274, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1277
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)
 * 
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */
  __pyx_t_21 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_9comment, 0, __pyx_n_s_ModelicaParser_Class_spec_baseCo_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1139)); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_comment, __pyx_t_21) < 0) __PYX_ERR(0, 1277, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1280
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 *         def array_subscripts(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Array_subscriptsContext,0)
 * 
 */
  __pyx_t_21 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_11array_subscripts, 0, __pyx_n_s_ModelicaParser_Class_spec_baseCo_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1141)); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_array_subscripts, __pyx_t_21) < 0) __PYX_ERR(0, 1280, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1283
 *             return self.getTypedRuleContext(ModelicaParser.Array_subscriptsContext,0)
 * 
 *         def class_modification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_modificationContext,0)
 * 
 */
  __pyx_t_21 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_13class_modification, 0, __pyx_n_s_ModelicaParser_Class_spec_baseCo_8, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1143)); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_class_modification, __pyx_t_21) < 0) __PYX_ERR(0, 1283, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1287
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterClass_spec_base" ):
 *                 listener.enterClass_spec_base(self)
 */
  __pyx_t_21 = PyDict_New(); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_20 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  if (PyDict_SetItem(__pyx_t_21, __pyx_n_s_listener, __pyx_t_20) < 0) __PYX_ERR(0, 1287, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __pyx_t_20 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_15enterRule, 0, __pyx_n_s_ModelicaParser_Class_spec_baseCo_9, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1145)); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_20, __pyx_t_21);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_20) < 0) __PYX_ERR(0, 1287, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1291
 *                 listener.enterClass_spec_base(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitClass_spec_base" ):
 *                 listener.exitClass_spec_base(self)
 */
  __pyx_t_20 = PyDict_New(); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_21 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  if (PyDict_SetItem(__pyx_t_20, __pyx_n_s_listener, __pyx_t_21) < 0) __PYX_ERR(0, 1291, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __pyx_t_21 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_17exitRule, 0, __pyx_n_s_ModelicaParser_Class_spec_baseCo_10, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1147)); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_21, __pyx_t_20);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_21) < 0) __PYX_ERR(0, 1291, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1295
 *                 listener.exitClass_spec_base(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitClass_spec_base" ):
 *                 return visitor.visitClass_spec_base(self)
 */
  __pyx_t_21 = PyDict_New(); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_20 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  if (PyDict_SetItem(__pyx_t_21, __pyx_n_s_visitor, __pyx_t_20) < 0) __PYX_ERR(0, 1295, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __pyx_t_20 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_baseContext_19accept, 0, __pyx_n_s_ModelicaParser_Class_spec_baseCo_11, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1149)); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_20, __pyx_t_21);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_20) < 0) __PYX_ERR(0, 1295, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1263
 * 
 * 
 *     class Class_spec_baseContext(Class_specifierContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Class_specifierContext
 */
  __pyx_t_20 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Class_spec_baseContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_18, __pyx_t_20) < 0) __PYX_ERR(0, 1263, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Class_spec_baseContext, __pyx_t_20) < 0) __PYX_ERR(0, 1263, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1302
 * 
 * 
 *     class Class_spec_compContext(Class_specifierContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Class_specifierContext
 */
  __pyx_t_7 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_Class_specifierContext);
  if (unlikely(!__pyx_t_7)) {
    PyErr_Clear();
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_Class_specifierContext);
  }
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Class_spec_compContext, __pyx_n_s_ModelicaParser_Class_spec_compCo, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_20 = PyList_New(0); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);

  /* "pymola/generated/ModelicaParser.pyx":1304
 *     class Class_spec_compContext(Class_specifierContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Class_specifierContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_21 = PyDict_New(); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_22 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  if (PyDict_SetItem(__pyx_t_21, __pyx_n_s_ctx_2, __pyx_t_22) < 0) __PYX_ERR(0, 1304, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __pyx_t_22 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_1__init__, 0, __pyx_n_s_ModelicaParser_Class_spec_compCo_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1151)); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_INCREF(__pyx_t_22);
  PyList_Append(__pyx_t_20, __pyx_t_22);
  __Pyx_GIVEREF(__pyx_t_22);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_22, __pyx_t_21);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_22) < 0) __PYX_ERR(0, 1304, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1308
 *             self.copyFrom(ctx)
 * 
 *         def IDENT(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTokens(ModelicaParser.IDENT)
 */
  __pyx_t_22 = PyDict_New(); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  if (PyDict_SetItem(__pyx_t_22, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 1308, __pyx_L1_error)
  __pyx_t_21 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_3IDENT, 0, __pyx_n_s_ModelicaParser_Class_spec_compCo_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1153)); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_21, __pyx_tuple__1154);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_21, __pyx_t_22);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_IDENT, __pyx_t_21) < 0) __PYX_ERR(0, 1308, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1313
 *             else:
 *                 return self.getToken(ModelicaParser.IDENT, i)
 *         def string_comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.String_commentContext,0)
 * 
 */
  __pyx_t_21 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_5string_comment, 0, __pyx_n_s_ModelicaParser_Class_spec_compCo_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1156)); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_string_comment, __pyx_t_21) < 0) __PYX_ERR(0, 1313, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1316
 *             return self.getTypedRuleContext(ModelicaParser.String_commentContext,0)
 * 
 *         def composition(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CompositionContext,0)
 * 
 */
  __pyx_t_21 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_7composition, 0, __pyx_n_s_ModelicaParser_Class_spec_compCo_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1158)); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1316, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_composition, __pyx_t_21) < 0) __PYX_ERR(0, 1316, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1320
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterClass_spec_comp" ):
 *                 listener.enterClass_spec_comp(self)
 */
  __pyx_t_21 = PyDict_New(); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_22 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  if (PyDict_SetItem(__pyx_t_21, __pyx_n_s_listener, __pyx_t_22) < 0) __PYX_ERR(0, 1320, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __pyx_t_22 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_9enterRule, 0, __pyx_n_s_ModelicaParser_Class_spec_compCo_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1160)); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_22, __pyx_t_21);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_22) < 0) __PYX_ERR(0, 1320, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1324
 *                 listener.enterClass_spec_comp(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitClass_spec_comp" ):
 *                 listener.exitClass_spec_comp(self)
 */
  __pyx_t_22 = PyDict_New(); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_t_21 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  if (PyDict_SetItem(__pyx_t_22, __pyx_n_s_listener, __pyx_t_21) < 0) __PYX_ERR(0, 1324, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __pyx_t_21 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_11exitRule, 0, __pyx_n_s_ModelicaParser_Class_spec_compCo_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1162)); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_21, __pyx_t_22);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_21) < 0) __PYX_ERR(0, 1324, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1328
 *                 listener.exitClass_spec_comp(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitClass_spec_comp" ):
 *                 return visitor.visitClass_spec_comp(self)
 */
  __pyx_t_21 = PyDict_New(); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_22 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  if (PyDict_SetItem(__pyx_t_21, __pyx_n_s_visitor, __pyx_t_22) < 0) __PYX_ERR(0, 1328, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __pyx_t_22 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Class_spec_compContext_13accept, 0, __pyx_n_s_ModelicaParser_Class_spec_compCo_8, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1164)); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_22, __pyx_t_21);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_22) < 0) __PYX_ERR(0, 1328, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1302
 * 
 * 
 *     class Class_spec_compContext(Class_specifierContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Class_specifierContext
 */
  __pyx_t_22 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Class_spec_compContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_20, __pyx_t_22) < 0) __PYX_ERR(0, 1302, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Class_spec_compContext, __pyx_t_22) < 0) __PYX_ERR(0, 1302, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1335
 * 
 * 
 *     class Class_spec_extendsContext(Class_specifierContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Class_specifierContext
 */
  __pyx_t_5 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_Class_specifierContext);
  if (unlikely(!__pyx_t_5)) {
    PyErr_Clear();
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_Class_specifierContext);
  }
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Class_spec_extendsContext, __pyx_n_s_ModelicaParser_Class_spec_extend, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_22 = PyList_New(0); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);

  /* "pymola/generated/ModelicaParser.pyx":1337
 *     class Class_spec_extendsContext(Class_specifierContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Class_specifierContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_21 = PyDict_New(); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_23 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  if (PyDict_SetItem(__pyx_t_21, __pyx_n_s_ctx_2, __pyx_t_23) < 0) __PYX_ERR(0, 1337, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __pyx_t_23 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_1__init__, 0, __pyx_n_s_ModelicaParser_Class_spec_extend_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1166)); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_INCREF(__pyx_t_23);
  PyList_Append(__pyx_t_22, __pyx_t_23);
  __Pyx_GIVEREF(__pyx_t_23);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_23, __pyx_t_21);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_23) < 0) __PYX_ERR(0, 1337, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1341
 *             self.copyFrom(ctx)
 * 
 *         def IDENT(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTokens(ModelicaParser.IDENT)
 */
  __pyx_t_23 = PyDict_New(); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  if (PyDict_SetItem(__pyx_t_23, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 1341, __pyx_L1_error)
  __pyx_t_21 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_3IDENT, 0, __pyx_n_s_ModelicaParser_Class_spec_extend_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1168)); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_21, __pyx_tuple__1169);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_21, __pyx_t_23);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_IDENT, __pyx_t_21) < 0) __PYX_ERR(0, 1341, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1346
 *             else:
 *                 return self.getToken(ModelicaParser.IDENT, i)
 *         def string_comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.String_commentContext,0)
 * 
 */
  __pyx_t_21 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_5string_comment, 0, __pyx_n_s_ModelicaParser_Class_spec_extend_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1171)); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1346, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_string_comment, __pyx_t_21) < 0) __PYX_ERR(0, 1346, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1349
 *             return self.getTypedRuleContext(ModelicaParser.String_commentContext,0)
 * 
 *         def composition(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CompositionContext,0)
 * 
 */
  __pyx_t_21 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_7composition, 0, __pyx_n_s_ModelicaParser_Class_spec_extend_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1173)); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_composition, __pyx_t_21) < 0) __PYX_ERR(0, 1349, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1352
 *             return self.getTypedRuleContext(ModelicaParser.CompositionContext,0)
 * 
 *         def class_modification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_modificationContext,0)
 * 
 */
  __pyx_t_21 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_9class_modification, 0, __pyx_n_s_ModelicaParser_Class_spec_extend_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1175)); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_class_modification, __pyx_t_21) < 0) __PYX_ERR(0, 1352, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1356
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterClass_spec_extends" ):
 *                 listener.enterClass_spec_extends(self)
 */
  __pyx_t_21 = PyDict_New(); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_23 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  if (PyDict_SetItem(__pyx_t_21, __pyx_n_s_listener, __pyx_t_23) < 0) __PYX_ERR(0, 1356, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __pyx_t_23 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_11enterRule, 0, __pyx_n_s_ModelicaParser_Class_spec_extend_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1177)); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_23, __pyx_t_21);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_23) < 0) __PYX_ERR(0, 1356, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1360
 *                 listener.enterClass_spec_extends(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitClass_spec_extends" ):
 *                 listener.exitClass_spec_extends(self)
 */
  __pyx_t_23 = PyDict_New(); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_21 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  if (PyDict_SetItem(__pyx_t_23, __pyx_n_s_listener, __pyx_t_21) < 0) __PYX_ERR(0, 1360, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __pyx_t_21 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_13exitRule, 0, __pyx_n_s_ModelicaParser_Class_spec_extend_8, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1179)); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_21, __pyx_t_23);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_21) < 0) __PYX_ERR(0, 1360, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1364
 *                 listener.exitClass_spec_extends(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitClass_spec_extends" ):
 *                 return visitor.visitClass_spec_extends(self)
 */
  __pyx_t_21 = PyDict_New(); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1364, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_23 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1364, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  if (PyDict_SetItem(__pyx_t_21, __pyx_n_s_visitor, __pyx_t_23) < 0) __PYX_ERR(0, 1364, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __pyx_t_23 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_25Class_spec_extendsContext_15accept, 0, __pyx_n_s_ModelicaParser_Class_spec_extend_9, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1181)); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1364, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_23, __pyx_t_21);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_23) < 0) __PYX_ERR(0, 1364, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1335
 * 
 * 
 *     class Class_spec_extendsContext(Class_specifierContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Class_specifierContext
 */
  __pyx_t_23 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Class_spec_extendsContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_22, __pyx_t_23) < 0) __PYX_ERR(0, 1335, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Class_spec_extendsContext, __pyx_t_23) < 0) __PYX_ERR(0, 1335, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1372
 * 
 * 
 *     def class_specifier(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Class_specifierContext(self, self._ctx, self.state)
 */
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_13class_specifier, 0, __pyx_n_s_ModelicaParser_class_specifier, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1183)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_class_specifier, __pyx_t_7) < 0) __PYX_ERR(0, 1372, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1533
 *         return localctx
 * 
 *     class Base_prefixContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Base_prefixContext, __pyx_n_s_ModelicaParser_Base_prefixContex, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_23 = PyList_New(0); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);

  /* "pymola/generated/ModelicaParser.pyx":1535
 *     class Base_prefixContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_21 = PyDict_New(); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_24 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  if (PyDict_SetItem(__pyx_t_21, __pyx_n_s_parent, __pyx_t_24) < 0) __PYX_ERR(0, 1535, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  if (PyDict_SetItem(__pyx_t_21, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 1535, __pyx_L1_error)
  __pyx_t_24 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18Base_prefixContext_1__init__, 0, __pyx_n_s_ModelicaParser_Base_prefixContex_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1185)); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_INCREF(__pyx_t_24);
  PyList_Append(__pyx_t_23, __pyx_t_24);
  __Pyx_GIVEREF(__pyx_t_24);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_24, __pyx_tuple__1186);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_24, __pyx_t_21);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_24) < 0) __PYX_ERR(0, 1535, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1539
 *             self.parser = parser
 * 
 *         def type_prefix(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Type_prefixContext,0)
 * 
 */
  __pyx_t_24 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18Base_prefixContext_3type_prefix, 0, __pyx_n_s_ModelicaParser_Base_prefixContex_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1188)); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1539, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_type_prefix, __pyx_t_24) < 0) __PYX_ERR(0, 1539, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1543
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_base_prefix
 * 
 */
  __pyx_t_24 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18Base_prefixContext_5getRuleIndex, 0, __pyx_n_s_ModelicaParser_Base_prefixContex_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1190)); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_24) < 0) __PYX_ERR(0, 1543, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1546
 *             return ModelicaParser.RULE_base_prefix
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterBase_prefix" ):
 *                 listener.enterBase_prefix(self)
 */
  __pyx_t_24 = PyDict_New(); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __pyx_t_21 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  if (PyDict_SetItem(__pyx_t_24, __pyx_n_s_listener, __pyx_t_21) < 0) __PYX_ERR(0, 1546, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __pyx_t_21 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18Base_prefixContext_7enterRule, 0, __pyx_n_s_ModelicaParser_Base_prefixContex_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1192)); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_21, __pyx_t_24);
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_21) < 0) __PYX_ERR(0, 1546, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1550
 *                 listener.enterBase_prefix(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitBase_prefix" ):
 *                 listener.exitBase_prefix(self)
 */
  __pyx_t_21 = PyDict_New(); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_24 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  if (PyDict_SetItem(__pyx_t_21, __pyx_n_s_listener, __pyx_t_24) < 0) __PYX_ERR(0, 1550, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __pyx_t_24 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18Base_prefixContext_9exitRule, 0, __pyx_n_s_ModelicaParser_Base_prefixContex_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1194)); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_24, __pyx_t_21);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_24) < 0) __PYX_ERR(0, 1550, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1554
 *                 listener.exitBase_prefix(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitBase_prefix" ):
 *                 return visitor.visitBase_prefix(self)
 */
  __pyx_t_24 = PyDict_New(); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __pyx_t_21 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  if (PyDict_SetItem(__pyx_t_24, __pyx_n_s_visitor, __pyx_t_21) < 0) __PYX_ERR(0, 1554, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __pyx_t_21 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18Base_prefixContext_11accept, 0, __pyx_n_s_ModelicaParser_Base_prefixContex_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1196)); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_21, __pyx_t_24);
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_21) < 0) __PYX_ERR(0, 1554, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1533
 *         return localctx
 * 
 *     class Base_prefixContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_21 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Base_prefixContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_23, __pyx_t_21) < 0) __PYX_ERR(0, 1533, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Base_prefixContext, __pyx_t_21) < 0) __PYX_ERR(0, 1533, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1563
 * 
 * 
 *     def base_prefix(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Base_prefixContext(self, self._ctx, self.state)
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_15base_prefix, 0, __pyx_n_s_ModelicaParser_base_prefix, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1198)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_base_prefix, __pyx_t_5) < 0) __PYX_ERR(0, 1563, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1579
 *         return localctx
 * 
 *     class Enum_listContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Enum_listContext, __pyx_n_s_ModelicaParser_Enum_listContext, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_21 = PyList_New(0); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);

  /* "pymola/generated/ModelicaParser.pyx":1581
 *     class Enum_listContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_24 = PyDict_New(); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __pyx_t_25 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 1581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  if (PyDict_SetItem(__pyx_t_24, __pyx_n_s_parent, __pyx_t_25) < 0) __PYX_ERR(0, 1581, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
  if (PyDict_SetItem(__pyx_t_24, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 1581, __pyx_L1_error)
  __pyx_t_25 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_16Enum_listContext_1__init__, 0, __pyx_n_s_ModelicaParser_Enum_listContext_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1200)); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 1581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_INCREF(__pyx_t_25);
  PyList_Append(__pyx_t_21, __pyx_t_25);
  __Pyx_GIVEREF(__pyx_t_25);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_25, __pyx_tuple__1201);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_25, __pyx_t_24);
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_25) < 0) __PYX_ERR(0, 1581, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1585
 *             self.parser = parser
 * 
 *         def enumeration_literal(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Enumeration_literalContext)
 */
  __pyx_t_25 = PyDict_New(); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 1585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  if (PyDict_SetItem(__pyx_t_25, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 1585, __pyx_L1_error)
  __pyx_t_24 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_16Enum_listContext_3enumeration_literal, 0, __pyx_n_s_ModelicaParser_Enum_listContext_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1203)); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_24, __pyx_tuple__1204);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_24, __pyx_t_25);
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enumeration_literal, __pyx_t_24) < 0) __PYX_ERR(0, 1585, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1592
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_enum_list
 * 
 */
  __pyx_t_24 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_16Enum_listContext_5getRuleIndex, 0, __pyx_n_s_ModelicaParser_Enum_listContext_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1206)); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1592, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_24) < 0) __PYX_ERR(0, 1592, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1595
 *             return ModelicaParser.RULE_enum_list
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterEnum_list" ):
 *                 listener.enterEnum_list(self)
 */
  __pyx_t_24 = PyDict_New(); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __pyx_t_25 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 1595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  if (PyDict_SetItem(__pyx_t_24, __pyx_n_s_listener, __pyx_t_25) < 0) __PYX_ERR(0, 1595, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
  __pyx_t_25 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_16Enum_listContext_7enterRule, 0, __pyx_n_s_ModelicaParser_Enum_listContext_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1208)); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 1595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_25, __pyx_t_24);
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_25) < 0) __PYX_ERR(0, 1595, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1599
 *                 listener.enterEnum_list(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitEnum_list" ):
 *                 listener.exitEnum_list(self)
 */
  __pyx_t_25 = PyDict_New(); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 1599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __pyx_t_24 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  if (PyDict_SetItem(__pyx_t_25, __pyx_n_s_listener, __pyx_t_24) < 0) __PYX_ERR(0, 1599, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __pyx_t_24 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_16Enum_listContext_9exitRule, 0, __pyx_n_s_ModelicaParser_Enum_listContext_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1210)); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_24, __pyx_t_25);
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_24) < 0) __PYX_ERR(0, 1599, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1603
 *                 listener.exitEnum_list(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitEnum_list" ):
 *                 return visitor.visitEnum_list(self)
 */
  __pyx_t_24 = PyDict_New(); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1603, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __pyx_t_25 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 1603, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  if (PyDict_SetItem(__pyx_t_24, __pyx_n_s_visitor, __pyx_t_25) < 0) __PYX_ERR(0, 1603, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
  __pyx_t_25 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_16Enum_listContext_11accept, 0, __pyx_n_s_ModelicaParser_Enum_listContext_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1212)); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 1603, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_25, __pyx_t_24);
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_25) < 0) __PYX_ERR(0, 1603, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1579
 *         return localctx
 * 
 *     class Enum_listContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_25 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Enum_listContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 1579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_21, __pyx_t_25) < 0) __PYX_ERR(0, 1579, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Enum_listContext, __pyx_t_25) < 0) __PYX_ERR(0, 1579, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1612
 * 
 * 
 *     def enum_list(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Enum_listContext(self, self._ctx, self.state)
 */
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_17enum_list, 0, __pyx_n_s_ModelicaParser_enum_list, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1214)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_enum_list, __pyx_t_7) < 0) __PYX_ERR(0, 1612, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1641
 *         return localctx
 * 
 *     class Enumeration_literalContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Enumeration_literalContext, __pyx_n_s_ModelicaParser_Enumeration_liter, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_25 = PyList_New(0); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 1641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);

  /* "pymola/generated/ModelicaParser.pyx":1643
 *     class Enumeration_literalContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_24 = PyDict_New(); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __pyx_t_26 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 1643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  if (PyDict_SetItem(__pyx_t_24, __pyx_n_s_parent, __pyx_t_26) < 0) __PYX_ERR(0, 1643, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
  if (PyDict_SetItem(__pyx_t_24, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 1643, __pyx_L1_error)
  __pyx_t_26 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_1__init__, 0, __pyx_n_s_ModelicaParser_Enumeration_liter_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1216)); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 1643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __Pyx_INCREF(__pyx_t_26);
  PyList_Append(__pyx_t_25, __pyx_t_26);
  __Pyx_GIVEREF(__pyx_t_26);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_26, __pyx_tuple__1217);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_26, __pyx_t_24);
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_26) < 0) __PYX_ERR(0, 1643, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1647
 *             self.parser = parser
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 */
  __pyx_t_26 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_3IDENT, 0, __pyx_n_s_ModelicaParser_Enumeration_liter_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1219)); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 1647, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_IDENT, __pyx_t_26) < 0) __PYX_ERR(0, 1647, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1650
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */
  __pyx_t_26 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_5comment, 0, __pyx_n_s_ModelicaParser_Enumeration_liter_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1221)); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 1650, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_comment, __pyx_t_26) < 0) __PYX_ERR(0, 1650, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1654
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_enumeration_literal
 * 
 */
  __pyx_t_26 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_7getRuleIndex, 0, __pyx_n_s_ModelicaParser_Enumeration_liter_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1223)); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 1654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_26) < 0) __PYX_ERR(0, 1654, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1657
 *             return ModelicaParser.RULE_enumeration_literal
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterEnumeration_literal" ):
 *                 listener.enterEnumeration_literal(self)
 */
  __pyx_t_26 = PyDict_New(); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 1657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __pyx_t_24 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  if (PyDict_SetItem(__pyx_t_26, __pyx_n_s_listener, __pyx_t_24) < 0) __PYX_ERR(0, 1657, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __pyx_t_24 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_9enterRule, 0, __pyx_n_s_ModelicaParser_Enumeration_liter_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1225)); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_24, __pyx_t_26);
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_24) < 0) __PYX_ERR(0, 1657, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1661
 *                 listener.enterEnumeration_literal(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitEnumeration_literal" ):
 *                 listener.exitEnumeration_literal(self)
 */
  __pyx_t_24 = PyDict_New(); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __pyx_t_26 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 1661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  if (PyDict_SetItem(__pyx_t_24, __pyx_n_s_listener, __pyx_t_26) < 0) __PYX_ERR(0, 1661, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
  __pyx_t_26 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_11exitRule, 0, __pyx_n_s_ModelicaParser_Enumeration_liter_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1227)); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 1661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_26, __pyx_t_24);
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_26) < 0) __PYX_ERR(0, 1661, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1665
 *                 listener.exitEnumeration_literal(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitEnumeration_literal" ):
 *                 return visitor.visitEnumeration_literal(self)
 */
  __pyx_t_26 = PyDict_New(); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 1665, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __pyx_t_24 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1665, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  if (PyDict_SetItem(__pyx_t_26, __pyx_n_s_visitor, __pyx_t_24) < 0) __PYX_ERR(0, 1665, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __pyx_t_24 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Enumeration_literalContext_13accept, 0, __pyx_n_s_ModelicaParser_Enumeration_liter_8, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1229)); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1665, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_24, __pyx_t_26);
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_24) < 0) __PYX_ERR(0, 1665, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1641
 *         return localctx
 * 
 *     class Enumeration_literalContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_24 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Enumeration_literalContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_25, __pyx_t_24) < 0) __PYX_ERR(0, 1641, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Enumeration_literalContext, __pyx_t_24) < 0) __PYX_ERR(0, 1641, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1674
 * 
 * 
 *     def enumeration_literal(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Enumeration_literalContext(self, self._ctx, self.state)
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_19enumeration_literal, 0, __pyx_n_s_ModelicaParser_enumeration_liter, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1231)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_enumeration_literal, __pyx_t_5) < 0) __PYX_ERR(0, 1674, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1692
 *         return localctx
 * 
 *     class CompositionContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_CompositionContext, __pyx_n_s_ModelicaParser_CompositionContex, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_24 = PyList_New(0); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);

  /* "pymola/generated/ModelicaParser.pyx":1694
 *     class CompositionContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_26 = PyDict_New(); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 1694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __pyx_t_27 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 1694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  if (PyDict_SetItem(__pyx_t_26, __pyx_n_s_parent, __pyx_t_27) < 0) __PYX_ERR(0, 1694, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
  if (PyDict_SetItem(__pyx_t_26, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 1694, __pyx_L1_error)
  __pyx_t_27 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18CompositionContext_1__init__, 0, __pyx_n_s_ModelicaParser_CompositionContex_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1233)); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 1694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_INCREF(__pyx_t_27);
  PyList_Append(__pyx_t_24, __pyx_t_27);
  __Pyx_GIVEREF(__pyx_t_27);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_27, __pyx_tuple__1234);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_27, __pyx_t_26);
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_27) < 0) __PYX_ERR(0, 1694, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1703
 *             self.comp_annotation = None # AnnotationContext
 * 
 *         def element_list(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Element_listContext)
 */
  __pyx_t_27 = PyDict_New(); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 1703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  if (PyDict_SetItem(__pyx_t_27, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 1703, __pyx_L1_error)
  __pyx_t_26 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18CompositionContext_3element_list, 0, __pyx_n_s_ModelicaParser_CompositionContex_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1236)); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 1703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_26, __pyx_tuple__1237);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_26, __pyx_t_27);
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_element_list, __pyx_t_26) < 0) __PYX_ERR(0, 1703, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1710
 * 
 * 
 *         def equation_section(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Equation_sectionContext)
 */
  __pyx_t_26 = PyDict_New(); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 1710, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  if (PyDict_SetItem(__pyx_t_26, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 1710, __pyx_L1_error)
  __pyx_t_27 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18CompositionContext_5equation_section, 0, __pyx_n_s_ModelicaParser_CompositionContex_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1239)); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 1710, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_27, __pyx_tuple__1240);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_27, __pyx_t_26);
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_equation_section, __pyx_t_27) < 0) __PYX_ERR(0, 1710, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1717
 * 
 * 
 *         def algorithm_section(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Algorithm_sectionContext)
 */
  __pyx_t_27 = PyDict_New(); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 1717, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  if (PyDict_SetItem(__pyx_t_27, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 1717, __pyx_L1_error)
  __pyx_t_26 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18CompositionContext_7algorithm_section, 0, __pyx_n_s_ModelicaParser_CompositionContex_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1242)); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 1717, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_26, __pyx_tuple__1243);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_26, __pyx_t_27);
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_algorithm_section, __pyx_t_26) < 0) __PYX_ERR(0, 1717, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1724
 * 
 * 
 *         def annotation(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.AnnotationContext)
 */
  __pyx_t_26 = PyDict_New(); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 1724, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  if (PyDict_SetItem(__pyx_t_26, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 1724, __pyx_L1_error)
  __pyx_t_27 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18CompositionContext_9annotation, 0, __pyx_n_s_ModelicaParser_CompositionContex_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1245)); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 1724, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_27, __pyx_tuple__1246);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_27, __pyx_t_26);
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_annotation, __pyx_t_27) < 0) __PYX_ERR(0, 1724, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1731
 * 
 * 
 *         def language_specification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Language_specificationContext,0)
 * 
 */
  __pyx_t_27 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18CompositionContext_11language_specification, 0, __pyx_n_s_ModelicaParser_CompositionContex_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1248)); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 1731, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_language_specification, __pyx_t_27) < 0) __PYX_ERR(0, 1731, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1735
 * 
 * 
 *         def external_function_call(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.External_function_callContext,0)
 * 
 */
  __pyx_t_27 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18CompositionContext_13external_function_call, 0, __pyx_n_s_ModelicaParser_CompositionContex_8, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1250)); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 1735, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_external_function_call, __pyx_t_27) < 0) __PYX_ERR(0, 1735, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1739
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_composition
 * 
 */
  __pyx_t_27 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18CompositionContext_15getRuleIndex, 0, __pyx_n_s_ModelicaParser_CompositionContex_9, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1252)); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 1739, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_27) < 0) __PYX_ERR(0, 1739, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1742
 *             return ModelicaParser.RULE_composition
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterComposition" ):
 *                 listener.enterComposition(self)
 */
  __pyx_t_27 = PyDict_New(); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 1742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __pyx_t_26 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 1742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  if (PyDict_SetItem(__pyx_t_27, __pyx_n_s_listener, __pyx_t_26) < 0) __PYX_ERR(0, 1742, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
  __pyx_t_26 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18CompositionContext_17enterRule, 0, __pyx_n_s_ModelicaParser_CompositionContex_10, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1254)); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 1742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_26, __pyx_t_27);
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_26) < 0) __PYX_ERR(0, 1742, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1746
 *                 listener.enterComposition(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitComposition" ):
 *                 listener.exitComposition(self)
 */
  __pyx_t_26 = PyDict_New(); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 1746, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __pyx_t_27 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 1746, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  if (PyDict_SetItem(__pyx_t_26, __pyx_n_s_listener, __pyx_t_27) < 0) __PYX_ERR(0, 1746, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
  __pyx_t_27 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18CompositionContext_19exitRule, 0, __pyx_n_s_ModelicaParser_CompositionContex_11, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1256)); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 1746, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_27, __pyx_t_26);
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_27) < 0) __PYX_ERR(0, 1746, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1750
 *                 listener.exitComposition(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitComposition" ):
 *                 return visitor.visitComposition(self)
 */
  __pyx_t_27 = PyDict_New(); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 1750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __pyx_t_26 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 1750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  if (PyDict_SetItem(__pyx_t_27, __pyx_n_s_visitor, __pyx_t_26) < 0) __PYX_ERR(0, 1750, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
  __pyx_t_26 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18CompositionContext_21accept, 0, __pyx_n_s_ModelicaParser_CompositionContex_12, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1258)); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 1750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_26, __pyx_t_27);
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_26) < 0) __PYX_ERR(0, 1750, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1692
 *         return localctx
 * 
 *     class CompositionContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_26 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_CompositionContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 1692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_24, __pyx_t_26) < 0) __PYX_ERR(0, 1692, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_CompositionContext, __pyx_t_26) < 0) __PYX_ERR(0, 1692, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1759
 * 
 * 
 *     def composition(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.CompositionContext(self, self._ctx, self.state)
 */
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_21composition, 0, __pyx_n_s_ModelicaParser_composition, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1260)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1759, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_composition, __pyx_t_7) < 0) __PYX_ERR(0, 1759, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1856
 *         return localctx
 * 
 *     class Language_specificationContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Language_specificationContext, __pyx_n_s_ModelicaParser_Language_specific, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_26 = PyList_New(0); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 1856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);

  /* "pymola/generated/ModelicaParser.pyx":1858
 *     class Language_specificationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_27 = PyDict_New(); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 1858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __pyx_t_28 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 1858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_28);
  if (PyDict_SetItem(__pyx_t_27, __pyx_n_s_parent, __pyx_t_28) < 0) __PYX_ERR(0, 1858, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
  if (PyDict_SetItem(__pyx_t_27, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 1858, __pyx_L1_error)
  __pyx_t_28 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_29Language_specificationContext_1__init__, 0, __pyx_n_s_ModelicaParser_Language_specific_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1262)); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 1858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_28);
  __Pyx_INCREF(__pyx_t_28);
  PyList_Append(__pyx_t_26, __pyx_t_28);
  __Pyx_GIVEREF(__pyx_t_28);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_28, __pyx_tuple__1263);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_28, __pyx_t_27);
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_28) < 0) __PYX_ERR(0, 1858, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1862
 *             self.parser = parser
 * 
 *         def STRING(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.STRING, 0)
 * 
 */
  __pyx_t_28 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_29Language_specificationContext_3STRING, 0, __pyx_n_s_ModelicaParser_Language_specific_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1265)); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 1862, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_28);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_STRING, __pyx_t_28) < 0) __PYX_ERR(0, 1862, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1865
 *             return self.getToken(ModelicaParser.STRING, 0)
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_language_specification
 * 
 */
  __pyx_t_28 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_29Language_specificationContext_5getRuleIndex, 0, __pyx_n_s_ModelicaParser_Language_specific_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1267)); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 1865, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_28);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_28) < 0) __PYX_ERR(0, 1865, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1868
 *             return ModelicaParser.RULE_language_specification
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterLanguage_specification" ):
 *                 listener.enterLanguage_specification(self)
 */
  __pyx_t_28 = PyDict_New(); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 1868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_28);
  __pyx_t_27 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 1868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  if (PyDict_SetItem(__pyx_t_28, __pyx_n_s_listener, __pyx_t_27) < 0) __PYX_ERR(0, 1868, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
  __pyx_t_27 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_29Language_specificationContext_7enterRule, 0, __pyx_n_s_ModelicaParser_Language_specific_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1269)); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 1868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_27, __pyx_t_28);
  __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_27) < 0) __PYX_ERR(0, 1868, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1872
 *                 listener.enterLanguage_specification(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitLanguage_specification" ):
 *                 listener.exitLanguage_specification(self)
 */
  __pyx_t_27 = PyDict_New(); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 1872, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __pyx_t_28 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 1872, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_28);
  if (PyDict_SetItem(__pyx_t_27, __pyx_n_s_listener, __pyx_t_28) < 0) __PYX_ERR(0, 1872, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
  __pyx_t_28 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_29Language_specificationContext_9exitRule, 0, __pyx_n_s_ModelicaParser_Language_specific_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1271)); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 1872, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_28);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_28, __pyx_t_27);
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_28) < 0) __PYX_ERR(0, 1872, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1876
 *                 listener.exitLanguage_specification(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitLanguage_specification" ):
 *                 return visitor.visitLanguage_specification(self)
 */
  __pyx_t_28 = PyDict_New(); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 1876, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_28);
  __pyx_t_27 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 1876, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  if (PyDict_SetItem(__pyx_t_28, __pyx_n_s_visitor, __pyx_t_27) < 0) __PYX_ERR(0, 1876, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
  __pyx_t_27 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_29Language_specificationContext_11accept, 0, __pyx_n_s_ModelicaParser_Language_specific_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1273)); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 1876, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_27, __pyx_t_28);
  __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_27) < 0) __PYX_ERR(0, 1876, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1856
 *         return localctx
 * 
 *     class Language_specificationContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_27 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Language_specificationContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 1856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_26, __pyx_t_27) < 0) __PYX_ERR(0, 1856, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Language_specificationContext, __pyx_t_27) < 0) __PYX_ERR(0, 1856, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1885
 * 
 * 
 *     def language_specification(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Language_specificationContext(self, self._ctx, self.state)
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_23language_specification, 0, __pyx_n_s_ModelicaParser_language_specific, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1275)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_language_specification, __pyx_t_5) < 0) __PYX_ERR(0, 1885, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1901
 *         return localctx
 * 
 *     class External_function_callContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1901, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1901, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1901, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_External_function_callContext, __pyx_n_s_ModelicaParser_External_function, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1901, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_27 = PyList_New(0); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 1901, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);

  /* "pymola/generated/ModelicaParser.pyx":1903
 *     class External_function_callContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_28 = PyDict_New(); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 1903, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_28);
  __pyx_t_29 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 1903, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  if (PyDict_SetItem(__pyx_t_28, __pyx_n_s_parent, __pyx_t_29) < 0) __PYX_ERR(0, 1903, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
  if (PyDict_SetItem(__pyx_t_28, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 1903, __pyx_L1_error)
  __pyx_t_29 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_29External_function_callContext_1__init__, 0, __pyx_n_s_ModelicaParser_External_function_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1277)); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 1903, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __Pyx_INCREF(__pyx_t_29);
  PyList_Append(__pyx_t_27, __pyx_t_29);
  __Pyx_GIVEREF(__pyx_t_29);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_29, __pyx_tuple__1278);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_29, __pyx_t_28);
  __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_29) < 0) __PYX_ERR(0, 1903, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1907
 *             self.parser = parser
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 */
  __pyx_t_29 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_29External_function_callContext_3IDENT, 0, __pyx_n_s_ModelicaParser_External_function_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1280)); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 1907, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_IDENT, __pyx_t_29) < 0) __PYX_ERR(0, 1907, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1910
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 *         def component_reference(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 */
  __pyx_t_29 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_29External_function_callContext_5component_reference, 0, __pyx_n_s_ModelicaParser_External_function_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1282)); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 1910, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_component_reference, __pyx_t_29) < 0) __PYX_ERR(0, 1910, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1914
 * 
 * 
 *         def expression_list(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Expression_listContext,0)
 * 
 */
  __pyx_t_29 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_29External_function_callContext_7expression_list, 0, __pyx_n_s_ModelicaParser_External_function_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1284)); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 1914, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_expression_list, __pyx_t_29) < 0) __PYX_ERR(0, 1914, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1918
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_external_function_call
 * 
 */
  __pyx_t_29 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_29External_function_callContext_9getRuleIndex, 0, __pyx_n_s_ModelicaParser_External_function_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1286)); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 1918, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_29) < 0) __PYX_ERR(0, 1918, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1921
 *             return ModelicaParser.RULE_external_function_call
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExternal_function_call" ):
 *                 listener.enterExternal_function_call(self)
 */
  __pyx_t_29 = PyDict_New(); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 1921, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __pyx_t_28 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 1921, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_28);
  if (PyDict_SetItem(__pyx_t_29, __pyx_n_s_listener, __pyx_t_28) < 0) __PYX_ERR(0, 1921, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
  __pyx_t_28 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_29External_function_callContext_11enterRule, 0, __pyx_n_s_ModelicaParser_External_function_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1288)); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 1921, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_28);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_28, __pyx_t_29);
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_28) < 0) __PYX_ERR(0, 1921, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1925
 *                 listener.enterExternal_function_call(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExternal_function_call" ):
 *                 listener.exitExternal_function_call(self)
 */
  __pyx_t_28 = PyDict_New(); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 1925, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_28);
  __pyx_t_29 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 1925, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  if (PyDict_SetItem(__pyx_t_28, __pyx_n_s_listener, __pyx_t_29) < 0) __PYX_ERR(0, 1925, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
  __pyx_t_29 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_29External_function_callContext_13exitRule, 0, __pyx_n_s_ModelicaParser_External_function_8, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1290)); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 1925, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_29, __pyx_t_28);
  __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_29) < 0) __PYX_ERR(0, 1925, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1929
 *                 listener.exitExternal_function_call(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExternal_function_call" ):
 *                 return visitor.visitExternal_function_call(self)
 */
  __pyx_t_29 = PyDict_New(); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 1929, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __pyx_t_28 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 1929, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_28);
  if (PyDict_SetItem(__pyx_t_29, __pyx_n_s_visitor, __pyx_t_28) < 0) __PYX_ERR(0, 1929, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
  __pyx_t_28 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_29External_function_callContext_15accept, 0, __pyx_n_s_ModelicaParser_External_function_9, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1292)); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 1929, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_28);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_28, __pyx_t_29);
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_28) < 0) __PYX_ERR(0, 1929, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1901
 *         return localctx
 * 
 *     class External_function_callContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_28 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_External_function_callContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 1901, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_28);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_27, __pyx_t_28) < 0) __PYX_ERR(0, 1901, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_External_function_callContext, __pyx_t_28) < 0) __PYX_ERR(0, 1901, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1938
 * 
 * 
 *     def external_function_call(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.External_function_callContext(self, self._ctx, self.state)
 */
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_25external_function_call, 0, __pyx_n_s_ModelicaParser_external_function, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1294)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1938, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_external_function_call, __pyx_t_7) < 0) __PYX_ERR(0, 1938, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1977
 *         return localctx
 * 
 *     class Element_listContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1977, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1977, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1977, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Element_listContext, __pyx_n_s_ModelicaParser_Element_listConte, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1977, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_28 = PyList_New(0); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 1977, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_28);

  /* "pymola/generated/ModelicaParser.pyx":1979
 *     class Element_listContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_29 = PyDict_New(); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 1979, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __pyx_t_30 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 1979, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  if (PyDict_SetItem(__pyx_t_29, __pyx_n_s_parent, __pyx_t_30) < 0) __PYX_ERR(0, 1979, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
  if (PyDict_SetItem(__pyx_t_29, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 1979, __pyx_L1_error)
  __pyx_t_30 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19Element_listContext_1__init__, 0, __pyx_n_s_ModelicaParser_Element_listConte_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1296)); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 1979, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __Pyx_INCREF(__pyx_t_30);
  PyList_Append(__pyx_t_28, __pyx_t_30);
  __Pyx_GIVEREF(__pyx_t_30);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_30, __pyx_tuple__1297);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_30, __pyx_t_29);
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_30) < 0) __PYX_ERR(0, 1979, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1983
 *             self.parser = parser
 * 
 *         def element(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ElementContext)
 */
  __pyx_t_30 = PyDict_New(); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 1983, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  if (PyDict_SetItem(__pyx_t_30, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 1983, __pyx_L1_error)
  __pyx_t_29 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19Element_listContext_3element, 0, __pyx_n_s_ModelicaParser_Element_listConte_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1299)); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 1983, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_29, __pyx_tuple__1300);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_29, __pyx_t_30);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_element, __pyx_t_29) < 0) __PYX_ERR(0, 1983, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1990
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_element_list
 * 
 */
  __pyx_t_29 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19Element_listContext_5getRuleIndex, 0, __pyx_n_s_ModelicaParser_Element_listConte_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1302)); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 1990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_29) < 0) __PYX_ERR(0, 1990, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1993
 *             return ModelicaParser.RULE_element_list
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterElement_list" ):
 *                 listener.enterElement_list(self)
 */
  __pyx_t_29 = PyDict_New(); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 1993, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __pyx_t_30 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 1993, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  if (PyDict_SetItem(__pyx_t_29, __pyx_n_s_listener, __pyx_t_30) < 0) __PYX_ERR(0, 1993, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
  __pyx_t_30 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19Element_listContext_7enterRule, 0, __pyx_n_s_ModelicaParser_Element_listConte_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1304)); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 1993, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_30, __pyx_t_29);
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_30) < 0) __PYX_ERR(0, 1993, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1997
 *                 listener.enterElement_list(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitElement_list" ):
 *                 listener.exitElement_list(self)
 */
  __pyx_t_30 = PyDict_New(); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 1997, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_29 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 1997, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  if (PyDict_SetItem(__pyx_t_30, __pyx_n_s_listener, __pyx_t_29) < 0) __PYX_ERR(0, 1997, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
  __pyx_t_29 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19Element_listContext_9exitRule, 0, __pyx_n_s_ModelicaParser_Element_listConte_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1306)); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 1997, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_29, __pyx_t_30);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_29) < 0) __PYX_ERR(0, 1997, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2001
 *                 listener.exitElement_list(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitElement_list" ):
 *                 return visitor.visitElement_list(self)
 */
  __pyx_t_29 = PyDict_New(); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 2001, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __pyx_t_30 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 2001, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  if (PyDict_SetItem(__pyx_t_29, __pyx_n_s_visitor, __pyx_t_30) < 0) __PYX_ERR(0, 2001, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
  __pyx_t_30 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19Element_listContext_11accept, 0, __pyx_n_s_ModelicaParser_Element_listConte_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1308)); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 2001, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_30, __pyx_t_29);
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_30) < 0) __PYX_ERR(0, 2001, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1977
 *         return localctx
 * 
 *     class Element_listContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_30 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Element_listContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 1977, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_28, __pyx_t_30) < 0) __PYX_ERR(0, 1977, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Element_listContext, __pyx_t_30) < 0) __PYX_ERR(0, 1977, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2010
 * 
 * 
 *     def element_list(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Element_listContext(self, self._ctx, self.state)
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_27element_list, 0, __pyx_n_s_ModelicaParser_element_list, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1310)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2010, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_element_list, __pyx_t_5) < 0) __PYX_ERR(0, 2010, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2037
 *         return localctx
 * 
 *     class ElementContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2037, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2037, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2037, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_ElementContext, __pyx_n_s_ModelicaParser_ElementContext, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2037, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_30 = PyList_New(0); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 2037, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);

  /* "pymola/generated/ModelicaParser.pyx":2039
 *     class ElementContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_29 = PyDict_New(); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 2039, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __pyx_t_31 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 2039, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  if (PyDict_SetItem(__pyx_t_29, __pyx_n_s_parent, __pyx_t_31) < 0) __PYX_ERR(0, 2039, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
  if (PyDict_SetItem(__pyx_t_29, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 2039, __pyx_L1_error)
  __pyx_t_31 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ElementContext_1__init__, 0, __pyx_n_s_ModelicaParser_ElementContext_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1312)); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 2039, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_INCREF(__pyx_t_31);
  PyList_Append(__pyx_t_30, __pyx_t_31);
  __Pyx_GIVEREF(__pyx_t_31);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_31, __pyx_tuple__1313);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_31, __pyx_t_29);
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_31) < 0) __PYX_ERR(0, 2039, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2043
 *             self.parser = parser
 * 
 *         def import_clause(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Import_clauseContext,0)
 * 
 */
  __pyx_t_31 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ElementContext_3import_clause, 0, __pyx_n_s_ModelicaParser_ElementContext_im, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1315)); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 2043, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_import_clause, __pyx_t_31) < 0) __PYX_ERR(0, 2043, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2047
 * 
 * 
 *         def extends_clause(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Extends_clauseContext,0)
 * 
 */
  __pyx_t_31 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ElementContext_5extends_clause, 0, __pyx_n_s_ModelicaParser_ElementContext_ex, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1317)); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 2047, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_extends_clause, __pyx_t_31) < 0) __PYX_ERR(0, 2047, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2051
 * 
 * 
 *         def regular_element(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Regular_elementContext,0)
 * 
 */
  __pyx_t_31 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ElementContext_7regular_element, 0, __pyx_n_s_ModelicaParser_ElementContext_re, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1319)); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 2051, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_regular_element, __pyx_t_31) < 0) __PYX_ERR(0, 2051, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2055
 * 
 * 
 *         def replaceable_element(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Replaceable_elementContext,0)
 * 
 */
  __pyx_t_31 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ElementContext_9replaceable_element, 0, __pyx_n_s_ModelicaParser_ElementContext_re_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1321)); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 2055, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_replaceable_element, __pyx_t_31) < 0) __PYX_ERR(0, 2055, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2059
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_element
 * 
 */
  __pyx_t_31 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ElementContext_11getRuleIndex, 0, __pyx_n_s_ModelicaParser_ElementContext_ge, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1323)); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 2059, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_31) < 0) __PYX_ERR(0, 2059, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2062
 *             return ModelicaParser.RULE_element
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterElement" ):
 *                 listener.enterElement(self)
 */
  __pyx_t_31 = PyDict_New(); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 2062, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __pyx_t_29 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 2062, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  if (PyDict_SetItem(__pyx_t_31, __pyx_n_s_listener, __pyx_t_29) < 0) __PYX_ERR(0, 2062, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
  __pyx_t_29 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ElementContext_13enterRule, 0, __pyx_n_s_ModelicaParser_ElementContext_en, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1325)); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 2062, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_29, __pyx_t_31);
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_29) < 0) __PYX_ERR(0, 2062, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2066
 *                 listener.enterElement(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitElement" ):
 *                 listener.exitElement(self)
 */
  __pyx_t_29 = PyDict_New(); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 2066, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __pyx_t_31 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 2066, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  if (PyDict_SetItem(__pyx_t_29, __pyx_n_s_listener, __pyx_t_31) < 0) __PYX_ERR(0, 2066, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
  __pyx_t_31 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ElementContext_15exitRule, 0, __pyx_n_s_ModelicaParser_ElementContext_ex_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1327)); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 2066, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_31, __pyx_t_29);
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_31) < 0) __PYX_ERR(0, 2066, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2070
 *                 listener.exitElement(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitElement" ):
 *                 return visitor.visitElement(self)
 */
  __pyx_t_31 = PyDict_New(); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 2070, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __pyx_t_29 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 2070, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  if (PyDict_SetItem(__pyx_t_31, __pyx_n_s_visitor, __pyx_t_29) < 0) __PYX_ERR(0, 2070, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
  __pyx_t_29 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ElementContext_17accept, 0, __pyx_n_s_ModelicaParser_ElementContext_ac, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1329)); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 2070, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_29, __pyx_t_31);
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_29) < 0) __PYX_ERR(0, 2070, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2037
 *         return localctx
 * 
 *     class ElementContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_29 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_ElementContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 2037, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_30, __pyx_t_29) < 0) __PYX_ERR(0, 2037, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_ElementContext, __pyx_t_29) < 0) __PYX_ERR(0, 2037, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2079
 * 
 * 
 *     def element(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.ElementContext(self, self._ctx, self.state)
 */
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_29element, 0, __pyx_n_s_ModelicaParser_element, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1331)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2079, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_element, __pyx_t_7) < 0) __PYX_ERR(0, 2079, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2120
 *         return localctx
 * 
 *     class Regular_elementContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Regular_elementContext, __pyx_n_s_ModelicaParser_Regular_elementCo, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_29 = PyList_New(0); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 2120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);

  /* "pymola/generated/ModelicaParser.pyx":2122
 *     class Regular_elementContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_31 = PyDict_New(); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 2122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __pyx_t_32 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 2122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  if (PyDict_SetItem(__pyx_t_31, __pyx_n_s_parent, __pyx_t_32) < 0) __PYX_ERR(0, 2122, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  if (PyDict_SetItem(__pyx_t_31, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 2122, __pyx_L1_error)
  __pyx_t_32 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Regular_elementContext_1__init__, 0, __pyx_n_s_ModelicaParser_Regular_elementCo_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1333)); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 2122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __Pyx_INCREF(__pyx_t_32);
  PyList_Append(__pyx_t_29, __pyx_t_32);
  __Pyx_GIVEREF(__pyx_t_32);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_32, __pyx_tuple__1334);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_32, __pyx_t_31);
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_32) < 0) __PYX_ERR(0, 2122, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2128
 *             self.comp_elem = None # Component_clauseContext
 * 
 *         def REDECLARE(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.REDECLARE, 0)
 * 
 */
  __pyx_t_32 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Regular_elementContext_3REDECLARE, 0, __pyx_n_s_ModelicaParser_Regular_elementCo_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1336)); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 2128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_REDECLARE, __pyx_t_32) < 0) __PYX_ERR(0, 2128, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2131
 *             return self.getToken(ModelicaParser.REDECLARE, 0)
 * 
 *         def FINAL(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.FINAL, 0)
 * 
 */
  __pyx_t_32 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Regular_elementContext_5FINAL, 0, __pyx_n_s_ModelicaParser_Regular_elementCo_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1338)); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 2131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_FINAL, __pyx_t_32) < 0) __PYX_ERR(0, 2131, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2134
 *             return self.getToken(ModelicaParser.FINAL, 0)
 * 
 *         def INNER(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.INNER, 0)
 * 
 */
  __pyx_t_32 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Regular_elementContext_7INNER, 0, __pyx_n_s_ModelicaParser_Regular_elementCo_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1340)); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 2134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_INNER, __pyx_t_32) < 0) __PYX_ERR(0, 2134, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2137
 *             return self.getToken(ModelicaParser.INNER, 0)
 * 
 *         def OUTER(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.OUTER, 0)
 * 
 */
  __pyx_t_32 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Regular_elementContext_9OUTER, 0, __pyx_n_s_ModelicaParser_Regular_elementCo_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1342)); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 2137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_OUTER, __pyx_t_32) < 0) __PYX_ERR(0, 2137, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2140
 *             return self.getToken(ModelicaParser.OUTER, 0)
 * 
 *         def class_definition(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_definitionContext,0)
 * 
 */
  __pyx_t_32 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Regular_elementContext_11class_definition, 0, __pyx_n_s_ModelicaParser_Regular_elementCo_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1344)); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 2140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_class_definition, __pyx_t_32) < 0) __PYX_ERR(0, 2140, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2144
 * 
 * 
 *         def component_clause(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_clauseContext,0)
 * 
 */
  __pyx_t_32 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Regular_elementContext_13component_clause, 0, __pyx_n_s_ModelicaParser_Regular_elementCo_8, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1346)); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 2144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_component_clause, __pyx_t_32) < 0) __PYX_ERR(0, 2144, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2148
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_regular_element
 * 
 */
  __pyx_t_32 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Regular_elementContext_15getRuleIndex, 0, __pyx_n_s_ModelicaParser_Regular_elementCo_9, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1348)); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 2148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_32) < 0) __PYX_ERR(0, 2148, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2151
 *             return ModelicaParser.RULE_regular_element
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterRegular_element" ):
 *                 listener.enterRegular_element(self)
 */
  __pyx_t_32 = PyDict_New(); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 2151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_31 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 2151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  if (PyDict_SetItem(__pyx_t_32, __pyx_n_s_listener, __pyx_t_31) < 0) __PYX_ERR(0, 2151, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
  __pyx_t_31 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Regular_elementContext_17enterRule, 0, __pyx_n_s_ModelicaParser_Regular_elementCo_10, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1350)); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 2151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_31, __pyx_t_32);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_31) < 0) __PYX_ERR(0, 2151, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2155
 *                 listener.enterRegular_element(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitRegular_element" ):
 *                 listener.exitRegular_element(self)
 */
  __pyx_t_31 = PyDict_New(); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 2155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __pyx_t_32 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 2155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  if (PyDict_SetItem(__pyx_t_31, __pyx_n_s_listener, __pyx_t_32) < 0) __PYX_ERR(0, 2155, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  __pyx_t_32 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Regular_elementContext_19exitRule, 0, __pyx_n_s_ModelicaParser_Regular_elementCo_11, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1352)); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 2155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_32, __pyx_t_31);
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_32) < 0) __PYX_ERR(0, 2155, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2159
 *                 listener.exitRegular_element(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitRegular_element" ):
 *                 return visitor.visitRegular_element(self)
 */
  __pyx_t_32 = PyDict_New(); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 2159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_31 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 2159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  if (PyDict_SetItem(__pyx_t_32, __pyx_n_s_visitor, __pyx_t_31) < 0) __PYX_ERR(0, 2159, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
  __pyx_t_31 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Regular_elementContext_21accept, 0, __pyx_n_s_ModelicaParser_Regular_elementCo_12, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1354)); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 2159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_31, __pyx_t_32);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_31) < 0) __PYX_ERR(0, 2159, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2120
 *         return localctx
 * 
 *     class Regular_elementContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_31 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Regular_elementContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 2120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_29, __pyx_t_31) < 0) __PYX_ERR(0, 2120, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Regular_elementContext, __pyx_t_31) < 0) __PYX_ERR(0, 2120, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2168
 * 
 * 
 *     def regular_element(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Regular_elementContext(self, self._ctx, self.state)
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_31regular_element, 0, __pyx_n_s_ModelicaParser_regular_element, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1356)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_regular_element, __pyx_t_5) < 0) __PYX_ERR(0, 2168, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2229
 *         return localctx
 * 
 *     class Replaceable_elementContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Replaceable_elementContext, __pyx_n_s_ModelicaParser_Replaceable_eleme, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_31 = PyList_New(0); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 2229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);

  /* "pymola/generated/ModelicaParser.pyx":2231
 *     class Replaceable_elementContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_32 = PyDict_New(); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 2231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_33 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 2231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  if (PyDict_SetItem(__pyx_t_32, __pyx_n_s_parent, __pyx_t_33) < 0) __PYX_ERR(0, 2231, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
  if (PyDict_SetItem(__pyx_t_32, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 2231, __pyx_L1_error)
  __pyx_t_33 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_1__init__, 0, __pyx_n_s_ModelicaParser_Replaceable_eleme_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1358)); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 2231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __Pyx_INCREF(__pyx_t_33);
  PyList_Append(__pyx_t_31, __pyx_t_33);
  __Pyx_GIVEREF(__pyx_t_33);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_33, __pyx_tuple__1359);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_33, __pyx_t_32);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_33) < 0) __PYX_ERR(0, 2231, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2237
 *             self.comp_elem = None # Component_clauseContext
 * 
 *         def REDECLARE(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.REDECLARE, 0)
 * 
 */
  __pyx_t_33 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_3REDECLARE, 0, __pyx_n_s_ModelicaParser_Replaceable_eleme_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1361)); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 2237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_REDECLARE, __pyx_t_33) < 0) __PYX_ERR(0, 2237, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2240
 *             return self.getToken(ModelicaParser.REDECLARE, 0)
 * 
 *         def FINAL(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.FINAL, 0)
 * 
 */
  __pyx_t_33 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_5FINAL, 0, __pyx_n_s_ModelicaParser_Replaceable_eleme_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1363)); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 2240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_FINAL, __pyx_t_33) < 0) __PYX_ERR(0, 2240, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2243
 *             return self.getToken(ModelicaParser.FINAL, 0)
 * 
 *         def INNER(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.INNER, 0)
 * 
 */
  __pyx_t_33 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_7INNER, 0, __pyx_n_s_ModelicaParser_Replaceable_eleme_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1365)); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 2243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_INNER, __pyx_t_33) < 0) __PYX_ERR(0, 2243, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2246
 *             return self.getToken(ModelicaParser.INNER, 0)
 * 
 *         def OUTER(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.OUTER, 0)
 * 
 */
  __pyx_t_33 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_9OUTER, 0, __pyx_n_s_ModelicaParser_Replaceable_eleme_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1367)); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 2246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_OUTER, __pyx_t_33) < 0) __PYX_ERR(0, 2246, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2249
 *             return self.getToken(ModelicaParser.OUTER, 0)
 * 
 *         def class_definition(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_definitionContext,0)
 * 
 */
  __pyx_t_33 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_11class_definition, 0, __pyx_n_s_ModelicaParser_Replaceable_eleme_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1369)); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 2249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_class_definition, __pyx_t_33) < 0) __PYX_ERR(0, 2249, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2253
 * 
 * 
 *         def component_clause(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_clauseContext,0)
 * 
 */
  __pyx_t_33 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_13component_clause, 0, __pyx_n_s_ModelicaParser_Replaceable_eleme_8, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1371)); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 2253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_component_clause, __pyx_t_33) < 0) __PYX_ERR(0, 2253, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2257
 * 
 * 
 *         def constraining_clause(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Constraining_clauseContext,0)
 * 
 */
  __pyx_t_33 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_15constraining_clause, 0, __pyx_n_s_ModelicaParser_Replaceable_eleme_9, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1373)); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 2257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_constraining_clause, __pyx_t_33) < 0) __PYX_ERR(0, 2257, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2261
 * 
 * 
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */
  __pyx_t_33 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_17comment, 0, __pyx_n_s_ModelicaParser_Replaceable_eleme_10, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1375)); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 2261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_comment, __pyx_t_33) < 0) __PYX_ERR(0, 2261, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2265
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_replaceable_element
 * 
 */
  __pyx_t_33 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_19getRuleIndex, 0, __pyx_n_s_ModelicaParser_Replaceable_eleme_11, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1377)); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 2265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_33) < 0) __PYX_ERR(0, 2265, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2268
 *             return ModelicaParser.RULE_replaceable_element
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterReplaceable_element" ):
 *                 listener.enterReplaceable_element(self)
 */
  __pyx_t_33 = PyDict_New(); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 2268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __pyx_t_32 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 2268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  if (PyDict_SetItem(__pyx_t_33, __pyx_n_s_listener, __pyx_t_32) < 0) __PYX_ERR(0, 2268, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  __pyx_t_32 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_21enterRule, 0, __pyx_n_s_ModelicaParser_Replaceable_eleme_12, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1379)); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 2268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_32, __pyx_t_33);
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_32) < 0) __PYX_ERR(0, 2268, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2272
 *                 listener.enterReplaceable_element(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitReplaceable_element" ):
 *                 listener.exitReplaceable_element(self)
 */
  __pyx_t_32 = PyDict_New(); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 2272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_33 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 2272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  if (PyDict_SetItem(__pyx_t_32, __pyx_n_s_listener, __pyx_t_33) < 0) __PYX_ERR(0, 2272, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
  __pyx_t_33 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_23exitRule, 0, __pyx_n_s_ModelicaParser_Replaceable_eleme_13, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1381)); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 2272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_33, __pyx_t_32);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_33) < 0) __PYX_ERR(0, 2272, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2276
 *                 listener.exitReplaceable_element(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitReplaceable_element" ):
 *                 return visitor.visitReplaceable_element(self)
 */
  __pyx_t_33 = PyDict_New(); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 2276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __pyx_t_32 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 2276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  if (PyDict_SetItem(__pyx_t_33, __pyx_n_s_visitor, __pyx_t_32) < 0) __PYX_ERR(0, 2276, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  __pyx_t_32 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Replaceable_elementContext_25accept, 0, __pyx_n_s_ModelicaParser_Replaceable_eleme_14, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1383)); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 2276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_32, __pyx_t_33);
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_32) < 0) __PYX_ERR(0, 2276, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2229
 *         return localctx
 * 
 *     class Replaceable_elementContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_32 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Replaceable_elementContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 2229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_31, __pyx_t_32) < 0) __PYX_ERR(0, 2229, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Replaceable_elementContext, __pyx_t_32) < 0) __PYX_ERR(0, 2229, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2285
 * 
 * 
 *     def replaceable_element(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Replaceable_elementContext(self, self._ctx, self.state)
 */
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_33replaceable_element, 0, __pyx_n_s_ModelicaParser_replaceable_eleme, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1385)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_replaceable_element, __pyx_t_7) < 0) __PYX_ERR(0, 2285, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2358
 *         return localctx
 * 
 *     class Import_clauseContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Import_clauseContext, __pyx_n_s_ModelicaParser_Import_clauseCont, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_32 = PyList_New(0); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 2358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);

  /* "pymola/generated/ModelicaParser.pyx":2360
 *     class Import_clauseContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_33 = PyDict_New(); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 2360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __pyx_t_34 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 2360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  if (PyDict_SetItem(__pyx_t_33, __pyx_n_s_parent, __pyx_t_34) < 0) __PYX_ERR(0, 2360, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;
  if (PyDict_SetItem(__pyx_t_33, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 2360, __pyx_L1_error)
  __pyx_t_34 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20Import_clauseContext_1__init__, 0, __pyx_n_s_ModelicaParser_Import_clauseCont_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1387)); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 2360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  __Pyx_INCREF(__pyx_t_34);
  PyList_Append(__pyx_t_32, __pyx_t_34);
  __Pyx_GIVEREF(__pyx_t_34);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_34, __pyx_tuple__1388);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_34, __pyx_t_33);
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_34) < 0) __PYX_ERR(0, 2360, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2364
 *             self.parser = parser
 * 
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */
  __pyx_t_34 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20Import_clauseContext_3comment, 0, __pyx_n_s_ModelicaParser_Import_clauseCont_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1390)); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 2364, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_comment, __pyx_t_34) < 0) __PYX_ERR(0, 2364, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2368
 * 
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 */
  __pyx_t_34 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20Import_clauseContext_5IDENT, 0, __pyx_n_s_ModelicaParser_Import_clauseCont_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1392)); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 2368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_IDENT, __pyx_t_34) < 0) __PYX_ERR(0, 2368, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2371
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 *         def component_reference(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 */
  __pyx_t_34 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20Import_clauseContext_7component_reference, 0, __pyx_n_s_ModelicaParser_Import_clauseCont_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1394)); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 2371, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_component_reference, __pyx_t_34) < 0) __PYX_ERR(0, 2371, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2375
 * 
 * 
 *         def import_list(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Import_listContext,0)
 * 
 */
  __pyx_t_34 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20Import_clauseContext_9import_list, 0, __pyx_n_s_ModelicaParser_Import_clauseCont_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1396)); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 2375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_import_list, __pyx_t_34) < 0) __PYX_ERR(0, 2375, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2379
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_import_clause
 * 
 */
  __pyx_t_34 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20Import_clauseContext_11getRuleIndex, 0, __pyx_n_s_ModelicaParser_Import_clauseCont_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1398)); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 2379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_34) < 0) __PYX_ERR(0, 2379, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2382
 *             return ModelicaParser.RULE_import_clause
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterImport_clause" ):
 *                 listener.enterImport_clause(self)
 */
  __pyx_t_34 = PyDict_New(); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 2382, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  __pyx_t_33 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 2382, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  if (PyDict_SetItem(__pyx_t_34, __pyx_n_s_listener, __pyx_t_33) < 0) __PYX_ERR(0, 2382, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
  __pyx_t_33 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20Import_clauseContext_13enterRule, 0, __pyx_n_s_ModelicaParser_Import_clauseCont_8, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1400)); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 2382, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_33, __pyx_t_34);
  __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_33) < 0) __PYX_ERR(0, 2382, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2386
 *                 listener.enterImport_clause(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitImport_clause" ):
 *                 listener.exitImport_clause(self)
 */
  __pyx_t_33 = PyDict_New(); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 2386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __pyx_t_34 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 2386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  if (PyDict_SetItem(__pyx_t_33, __pyx_n_s_listener, __pyx_t_34) < 0) __PYX_ERR(0, 2386, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;
  __pyx_t_34 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20Import_clauseContext_15exitRule, 0, __pyx_n_s_ModelicaParser_Import_clauseCont_9, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1402)); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 2386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_34, __pyx_t_33);
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_34) < 0) __PYX_ERR(0, 2386, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2390
 *                 listener.exitImport_clause(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitImport_clause" ):
 *                 return visitor.visitImport_clause(self)
 */
  __pyx_t_34 = PyDict_New(); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 2390, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  __pyx_t_33 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 2390, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  if (PyDict_SetItem(__pyx_t_34, __pyx_n_s_visitor, __pyx_t_33) < 0) __PYX_ERR(0, 2390, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
  __pyx_t_33 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20Import_clauseContext_17accept, 0, __pyx_n_s_ModelicaParser_Import_clauseCont_10, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1404)); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 2390, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_33, __pyx_t_34);
  __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_33) < 0) __PYX_ERR(0, 2390, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2358
 *         return localctx
 * 
 *     class Import_clauseContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_33 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Import_clauseContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 2358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_32, __pyx_t_33) < 0) __PYX_ERR(0, 2358, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Import_clauseContext, __pyx_t_33) < 0) __PYX_ERR(0, 2358, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2399
 * 
 * 
 *     def import_clause(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Import_clauseContext(self, self._ctx, self.state)
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_35import_clause, 0, __pyx_n_s_ModelicaParser_import_clause, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1406)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_import_clause, __pyx_t_5) < 0) __PYX_ERR(0, 2399, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2462
 *         return localctx
 * 
 *     class Import_listContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Import_listContext, __pyx_n_s_ModelicaParser_Import_listContex, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_33 = PyList_New(0); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 2462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);

  /* "pymola/generated/ModelicaParser.pyx":2464
 *     class Import_listContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_34 = PyDict_New(); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 2464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  __pyx_t_35 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 2464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_35);
  if (PyDict_SetItem(__pyx_t_34, __pyx_n_s_parent, __pyx_t_35) < 0) __PYX_ERR(0, 2464, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_35); __pyx_t_35 = 0;
  if (PyDict_SetItem(__pyx_t_34, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 2464, __pyx_L1_error)
  __pyx_t_35 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18Import_listContext_1__init__, 0, __pyx_n_s_ModelicaParser_Import_listContex_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1408)); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 2464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_35);
  __Pyx_INCREF(__pyx_t_35);
  PyList_Append(__pyx_t_33, __pyx_t_35);
  __Pyx_GIVEREF(__pyx_t_35);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_35, __pyx_tuple__1409);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_35, __pyx_t_34);
  __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_35) < 0) __PYX_ERR(0, 2464, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_35); __pyx_t_35 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2468
 *             self.parser = parser
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 */
  __pyx_t_35 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18Import_listContext_3IDENT, 0, __pyx_n_s_ModelicaParser_Import_listContex_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1411)); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 2468, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_35);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_IDENT, __pyx_t_35) < 0) __PYX_ERR(0, 2468, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_35); __pyx_t_35 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2471
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 *         def import_list(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Import_listContext)
 */
  __pyx_t_35 = PyDict_New(); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 2471, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_35);
  if (PyDict_SetItem(__pyx_t_35, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 2471, __pyx_L1_error)
  __pyx_t_34 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18Import_listContext_5import_list, 0, __pyx_n_s_ModelicaParser_Import_listContex_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1413)); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 2471, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_34, __pyx_tuple__1414);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_34, __pyx_t_35);
  __Pyx_DECREF(__pyx_t_35); __pyx_t_35 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_import_list, __pyx_t_34) < 0) __PYX_ERR(0, 2471, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2478
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_import_list
 * 
 */
  __pyx_t_34 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18Import_listContext_7getRuleIndex, 0, __pyx_n_s_ModelicaParser_Import_listContex_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1416)); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 2478, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_34) < 0) __PYX_ERR(0, 2478, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2481
 *             return ModelicaParser.RULE_import_list
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterImport_list" ):
 *                 listener.enterImport_list(self)
 */
  __pyx_t_34 = PyDict_New(); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 2481, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  __pyx_t_35 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 2481, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_35);
  if (PyDict_SetItem(__pyx_t_34, __pyx_n_s_listener, __pyx_t_35) < 0) __PYX_ERR(0, 2481, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_35); __pyx_t_35 = 0;
  __pyx_t_35 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18Import_listContext_9enterRule, 0, __pyx_n_s_ModelicaParser_Import_listContex_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1418)); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 2481, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_35);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_35, __pyx_t_34);
  __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_35) < 0) __PYX_ERR(0, 2481, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_35); __pyx_t_35 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2485
 *                 listener.enterImport_list(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitImport_list" ):
 *                 listener.exitImport_list(self)
 */
  __pyx_t_35 = PyDict_New(); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 2485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_35);
  __pyx_t_34 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 2485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  if (PyDict_SetItem(__pyx_t_35, __pyx_n_s_listener, __pyx_t_34) < 0) __PYX_ERR(0, 2485, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;
  __pyx_t_34 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18Import_listContext_11exitRule, 0, __pyx_n_s_ModelicaParser_Import_listContex_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1420)); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 2485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_34, __pyx_t_35);
  __Pyx_DECREF(__pyx_t_35); __pyx_t_35 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_34) < 0) __PYX_ERR(0, 2485, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2489
 *                 listener.exitImport_list(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitImport_list" ):
 *                 return visitor.visitImport_list(self)
 */
  __pyx_t_34 = PyDict_New(); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 2489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  __pyx_t_35 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 2489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_35);
  if (PyDict_SetItem(__pyx_t_34, __pyx_n_s_visitor, __pyx_t_35) < 0) __PYX_ERR(0, 2489, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_35); __pyx_t_35 = 0;
  __pyx_t_35 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18Import_listContext_13accept, 0, __pyx_n_s_ModelicaParser_Import_listContex_8, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1422)); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 2489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_35);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_35, __pyx_t_34);
  __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_35) < 0) __PYX_ERR(0, 2489, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_35); __pyx_t_35 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2462
 *         return localctx
 * 
 *     class Import_listContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_35 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Import_listContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 2462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_35);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_33, __pyx_t_35) < 0) __PYX_ERR(0, 2462, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Import_listContext, __pyx_t_35) < 0) __PYX_ERR(0, 2462, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_35); __pyx_t_35 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2498
 * 
 * 
 *     def import_list(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Import_listContext(self, self._ctx, self.state)
 */
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_37import_list, 0, __pyx_n_s_ModelicaParser_import_list, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1424)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_import_list, __pyx_t_7) < 0) __PYX_ERR(0, 2498, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2527
 *         return localctx
 * 
 *     class Extends_clauseContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2527, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2527, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2527, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Extends_clauseContext, __pyx_n_s_ModelicaParser_Extends_clauseCon, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2527, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_35 = PyList_New(0); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 2527, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_35);

  /* "pymola/generated/ModelicaParser.pyx":2529
 *     class Extends_clauseContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_34 = PyDict_New(); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 2529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  __pyx_t_36 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 2529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_36);
  if (PyDict_SetItem(__pyx_t_34, __pyx_n_s_parent, __pyx_t_36) < 0) __PYX_ERR(0, 2529, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;
  if (PyDict_SetItem(__pyx_t_34, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 2529, __pyx_L1_error)
  __pyx_t_36 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_1__init__, 0, __pyx_n_s_ModelicaParser_Extends_clauseCon_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1426)); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 2529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_36);
  __Pyx_INCREF(__pyx_t_36);
  PyList_Append(__pyx_t_35, __pyx_t_36);
  __Pyx_GIVEREF(__pyx_t_36);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_36, __pyx_tuple__1427);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_36, __pyx_t_34);
  __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_36) < 0) __PYX_ERR(0, 2529, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2533
 *             self.parser = parser
 * 
 *         def component_reference(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 */
  __pyx_t_36 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_3component_reference, 0, __pyx_n_s_ModelicaParser_Extends_clauseCon_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1429)); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 2533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_36);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_component_reference, __pyx_t_36) < 0) __PYX_ERR(0, 2533, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2537
 * 
 * 
 *         def class_modification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_modificationContext,0)
 * 
 */
  __pyx_t_36 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_5class_modification, 0, __pyx_n_s_ModelicaParser_Extends_clauseCon_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1431)); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 2537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_36);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_class_modification, __pyx_t_36) < 0) __PYX_ERR(0, 2537, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2541
 * 
 * 
 *         def annotation(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.AnnotationContext,0)
 * 
 */
  __pyx_t_36 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_7annotation, 0, __pyx_n_s_ModelicaParser_Extends_clauseCon_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1433)); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 2541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_36);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_annotation, __pyx_t_36) < 0) __PYX_ERR(0, 2541, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2545
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_extends_clause
 * 
 */
  __pyx_t_36 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_9getRuleIndex, 0, __pyx_n_s_ModelicaParser_Extends_clauseCon_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1435)); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 2545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_36);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_36) < 0) __PYX_ERR(0, 2545, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2548
 *             return ModelicaParser.RULE_extends_clause
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExtends_clause" ):
 *                 listener.enterExtends_clause(self)
 */
  __pyx_t_36 = PyDict_New(); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 2548, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_36);
  __pyx_t_34 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 2548, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  if (PyDict_SetItem(__pyx_t_36, __pyx_n_s_listener, __pyx_t_34) < 0) __PYX_ERR(0, 2548, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;
  __pyx_t_34 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_11enterRule, 0, __pyx_n_s_ModelicaParser_Extends_clauseCon_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1437)); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 2548, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_34, __pyx_t_36);
  __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_34) < 0) __PYX_ERR(0, 2548, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2552
 *                 listener.enterExtends_clause(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExtends_clause" ):
 *                 listener.exitExtends_clause(self)
 */
  __pyx_t_34 = PyDict_New(); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 2552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  __pyx_t_36 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 2552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_36);
  if (PyDict_SetItem(__pyx_t_34, __pyx_n_s_listener, __pyx_t_36) < 0) __PYX_ERR(0, 2552, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;
  __pyx_t_36 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_13exitRule, 0, __pyx_n_s_ModelicaParser_Extends_clauseCon_8, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1439)); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 2552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_36);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_36, __pyx_t_34);
  __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_36) < 0) __PYX_ERR(0, 2552, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2556
 *                 listener.exitExtends_clause(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExtends_clause" ):
 *                 return visitor.visitExtends_clause(self)
 */
  __pyx_t_36 = PyDict_New(); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 2556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_36);
  __pyx_t_34 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 2556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  if (PyDict_SetItem(__pyx_t_36, __pyx_n_s_visitor, __pyx_t_34) < 0) __PYX_ERR(0, 2556, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;
  __pyx_t_34 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Extends_clauseContext_15accept, 0, __pyx_n_s_ModelicaParser_Extends_clauseCon_9, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1441)); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 2556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_34, __pyx_t_36);
  __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_34) < 0) __PYX_ERR(0, 2556, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2527
 *         return localctx
 * 
 *     class Extends_clauseContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_34 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Extends_clauseContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 2527, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_35, __pyx_t_34) < 0) __PYX_ERR(0, 2527, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_35); __pyx_t_35 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Extends_clauseContext, __pyx_t_34) < 0) __PYX_ERR(0, 2527, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2565
 * 
 * 
 *     def extends_clause(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Extends_clauseContext(self, self._ctx, self.state)
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_39extends_clause, 0, __pyx_n_s_ModelicaParser_extends_clause, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1443)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_extends_clause, __pyx_t_5) < 0) __PYX_ERR(0, 2565, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2600
 *         return localctx
 * 
 *     class Constraining_clauseContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Constraining_clauseContext, __pyx_n_s_ModelicaParser_Constraining_clau, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_34 = PyList_New(0); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 2600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);

  /* "pymola/generated/ModelicaParser.pyx":2602
 *     class Constraining_clauseContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_36 = PyDict_New(); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 2602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_36);
  __pyx_t_37 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 2602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  if (PyDict_SetItem(__pyx_t_36, __pyx_n_s_parent, __pyx_t_37) < 0) __PYX_ERR(0, 2602, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
  if (PyDict_SetItem(__pyx_t_36, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 2602, __pyx_L1_error)
  __pyx_t_37 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_1__init__, 0, __pyx_n_s_ModelicaParser_Constraining_clau_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1445)); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 2602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  __Pyx_INCREF(__pyx_t_37);
  PyList_Append(__pyx_t_34, __pyx_t_37);
  __Pyx_GIVEREF(__pyx_t_37);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_37, __pyx_tuple__1446);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_37, __pyx_t_36);
  __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_37) < 0) __PYX_ERR(0, 2602, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2606
 *             self.parser = parser
 * 
 *         def name(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)
 * 
 */
  __pyx_t_37 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_3name, 0, __pyx_n_s_ModelicaParser_Constraining_clau_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1448)); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 2606, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_name, __pyx_t_37) < 0) __PYX_ERR(0, 2606, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2610
 * 
 * 
 *         def class_modification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_modificationContext,0)
 * 
 */
  __pyx_t_37 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_5class_modification, 0, __pyx_n_s_ModelicaParser_Constraining_clau_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1450)); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 2610, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_class_modification, __pyx_t_37) < 0) __PYX_ERR(0, 2610, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2614
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_constraining_clause
 * 
 */
  __pyx_t_37 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_7getRuleIndex, 0, __pyx_n_s_ModelicaParser_Constraining_clau_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1452)); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 2614, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_37) < 0) __PYX_ERR(0, 2614, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2617
 *             return ModelicaParser.RULE_constraining_clause
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterConstraining_clause" ):
 *                 listener.enterConstraining_clause(self)
 */
  __pyx_t_37 = PyDict_New(); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 2617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  __pyx_t_36 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 2617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_36);
  if (PyDict_SetItem(__pyx_t_37, __pyx_n_s_listener, __pyx_t_36) < 0) __PYX_ERR(0, 2617, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;
  __pyx_t_36 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_9enterRule, 0, __pyx_n_s_ModelicaParser_Constraining_clau_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1454)); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 2617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_36);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_36, __pyx_t_37);
  __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_36) < 0) __PYX_ERR(0, 2617, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2621
 *                 listener.enterConstraining_clause(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitConstraining_clause" ):
 *                 listener.exitConstraining_clause(self)
 */
  __pyx_t_36 = PyDict_New(); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 2621, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_36);
  __pyx_t_37 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 2621, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  if (PyDict_SetItem(__pyx_t_36, __pyx_n_s_listener, __pyx_t_37) < 0) __PYX_ERR(0, 2621, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
  __pyx_t_37 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_11exitRule, 0, __pyx_n_s_ModelicaParser_Constraining_clau_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1456)); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 2621, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_37, __pyx_t_36);
  __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_37) < 0) __PYX_ERR(0, 2621, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2625
 *                 listener.exitConstraining_clause(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitConstraining_clause" ):
 *                 return visitor.visitConstraining_clause(self)
 */
  __pyx_t_37 = PyDict_New(); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 2625, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  __pyx_t_36 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 2625, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_36);
  if (PyDict_SetItem(__pyx_t_37, __pyx_n_s_visitor, __pyx_t_36) < 0) __PYX_ERR(0, 2625, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;
  __pyx_t_36 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Constraining_clauseContext_13accept, 0, __pyx_n_s_ModelicaParser_Constraining_clau_8, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1458)); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 2625, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_36);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_36, __pyx_t_37);
  __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_36) < 0) __PYX_ERR(0, 2625, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2600
 *         return localctx
 * 
 *     class Constraining_clauseContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_36 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Constraining_clauseContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 2600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_36);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_34, __pyx_t_36) < 0) __PYX_ERR(0, 2600, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Constraining_clauseContext, __pyx_t_36) < 0) __PYX_ERR(0, 2600, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2634
 * 
 * 
 *     def constraining_clause(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Constraining_clauseContext(self, self._ctx, self.state)
 */
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_41constraining_clause, 0, __pyx_n_s_ModelicaParser_constraining_clau, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1460)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_constraining_clause, __pyx_t_7) < 0) __PYX_ERR(0, 2634, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2661
 *         return localctx
 * 
 *     class Component_clauseContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Component_clauseContext, __pyx_n_s_ModelicaParser_Component_clauseC, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_36 = PyList_New(0); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 2661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_36);

  /* "pymola/generated/ModelicaParser.pyx":2663
 *     class Component_clauseContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_37 = PyDict_New(); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 2663, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  __pyx_t_38 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_38)) __PYX_ERR(0, 2663, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_38);
  if (PyDict_SetItem(__pyx_t_37, __pyx_n_s_parent, __pyx_t_38) < 0) __PYX_ERR(0, 2663, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_38); __pyx_t_38 = 0;
  if (PyDict_SetItem(__pyx_t_37, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 2663, __pyx_L1_error)
  __pyx_t_38 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Component_clauseContext_1__init__, 0, __pyx_n_s_ModelicaParser_Component_clauseC_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1462)); if (unlikely(!__pyx_t_38)) __PYX_ERR(0, 2663, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_38);
  __Pyx_INCREF(__pyx_t_38);
  PyList_Append(__pyx_t_36, __pyx_t_38);
  __Pyx_GIVEREF(__pyx_t_38);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_38, __pyx_tuple__1463);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_38, __pyx_t_37);
  __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_38) < 0) __PYX_ERR(0, 2663, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_38); __pyx_t_38 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2667
 *             self.parser = parser
 * 
 *         def type_prefix(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Type_prefixContext,0)
 * 
 */
  __pyx_t_38 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Component_clauseContext_3type_prefix, 0, __pyx_n_s_ModelicaParser_Component_clauseC_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1465)); if (unlikely(!__pyx_t_38)) __PYX_ERR(0, 2667, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_38);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_type_prefix, __pyx_t_38) < 0) __PYX_ERR(0, 2667, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_38); __pyx_t_38 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2671
 * 
 * 
 *         def type_specifier(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Type_specifierContext,0)
 * 
 */
  __pyx_t_38 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Component_clauseContext_5type_specifier, 0, __pyx_n_s_ModelicaParser_Component_clauseC_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1467)); if (unlikely(!__pyx_t_38)) __PYX_ERR(0, 2671, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_38);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_type_specifier, __pyx_t_38) < 0) __PYX_ERR(0, 2671, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_38); __pyx_t_38 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2675
 * 
 * 
 *         def component_list(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_listContext,0)
 * 
 */
  __pyx_t_38 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Component_clauseContext_7component_list, 0, __pyx_n_s_ModelicaParser_Component_clauseC_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1469)); if (unlikely(!__pyx_t_38)) __PYX_ERR(0, 2675, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_38);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_component_list, __pyx_t_38) < 0) __PYX_ERR(0, 2675, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_38); __pyx_t_38 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2679
 * 
 * 
 *         def array_subscripts(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Array_subscriptsContext,0)
 * 
 */
  __pyx_t_38 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Component_clauseContext_9array_subscripts, 0, __pyx_n_s_ModelicaParser_Component_clauseC_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1471)); if (unlikely(!__pyx_t_38)) __PYX_ERR(0, 2679, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_38);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_array_subscripts, __pyx_t_38) < 0) __PYX_ERR(0, 2679, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_38); __pyx_t_38 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2683
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_component_clause
 * 
 */
  __pyx_t_38 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Component_clauseContext_11getRuleIndex, 0, __pyx_n_s_ModelicaParser_Component_clauseC_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1473)); if (unlikely(!__pyx_t_38)) __PYX_ERR(0, 2683, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_38);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_38) < 0) __PYX_ERR(0, 2683, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_38); __pyx_t_38 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2686
 *             return ModelicaParser.RULE_component_clause
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterComponent_clause" ):
 *                 listener.enterComponent_clause(self)
 */
  __pyx_t_38 = PyDict_New(); if (unlikely(!__pyx_t_38)) __PYX_ERR(0, 2686, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_38);
  __pyx_t_37 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 2686, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  if (PyDict_SetItem(__pyx_t_38, __pyx_n_s_listener, __pyx_t_37) < 0) __PYX_ERR(0, 2686, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
  __pyx_t_37 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Component_clauseContext_13enterRule, 0, __pyx_n_s_ModelicaParser_Component_clauseC_8, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1475)); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 2686, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_37, __pyx_t_38);
  __Pyx_DECREF(__pyx_t_38); __pyx_t_38 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_37) < 0) __PYX_ERR(0, 2686, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2690
 *                 listener.enterComponent_clause(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitComponent_clause" ):
 *                 listener.exitComponent_clause(self)
 */
  __pyx_t_37 = PyDict_New(); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 2690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  __pyx_t_38 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_38)) __PYX_ERR(0, 2690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_38);
  if (PyDict_SetItem(__pyx_t_37, __pyx_n_s_listener, __pyx_t_38) < 0) __PYX_ERR(0, 2690, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_38); __pyx_t_38 = 0;
  __pyx_t_38 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Component_clauseContext_15exitRule, 0, __pyx_n_s_ModelicaParser_Component_clauseC_9, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1477)); if (unlikely(!__pyx_t_38)) __PYX_ERR(0, 2690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_38);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_38, __pyx_t_37);
  __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_38) < 0) __PYX_ERR(0, 2690, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_38); __pyx_t_38 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2694
 *                 listener.exitComponent_clause(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitComponent_clause" ):
 *                 return visitor.visitComponent_clause(self)
 */
  __pyx_t_38 = PyDict_New(); if (unlikely(!__pyx_t_38)) __PYX_ERR(0, 2694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_38);
  __pyx_t_37 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 2694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  if (PyDict_SetItem(__pyx_t_38, __pyx_n_s_visitor, __pyx_t_37) < 0) __PYX_ERR(0, 2694, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
  __pyx_t_37 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Component_clauseContext_17accept, 0, __pyx_n_s_ModelicaParser_Component_clauseC_10, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1479)); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 2694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_37, __pyx_t_38);
  __Pyx_DECREF(__pyx_t_38); __pyx_t_38 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_37) < 0) __PYX_ERR(0, 2694, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2661
 *         return localctx
 * 
 *     class Component_clauseContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_37 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Component_clauseContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 2661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_36, __pyx_t_37) < 0) __PYX_ERR(0, 2661, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Component_clauseContext, __pyx_t_37) < 0) __PYX_ERR(0, 2661, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2703
 * 
 * 
 *     def component_clause(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Component_clauseContext(self, self._ctx, self.state)
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_43component_clause, 0, __pyx_n_s_ModelicaParser_component_clause, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1481)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_component_clause, __pyx_t_5) < 0) __PYX_ERR(0, 2703, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2732
 *         return localctx
 * 
 *     class Type_prefixContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2732, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2732, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2732, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Type_prefixContext, __pyx_n_s_ModelicaParser_Type_prefixContex, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2732, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_37 = PyList_New(0); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 2732, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);

  /* "pymola/generated/ModelicaParser.pyx":2734
 *     class Type_prefixContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_38 = PyDict_New(); if (unlikely(!__pyx_t_38)) __PYX_ERR(0, 2734, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_38);
  __pyx_t_39 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_39)) __PYX_ERR(0, 2734, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_39);
  if (PyDict_SetItem(__pyx_t_38, __pyx_n_s_parent, __pyx_t_39) < 0) __PYX_ERR(0, 2734, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_39); __pyx_t_39 = 0;
  if (PyDict_SetItem(__pyx_t_38, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 2734, __pyx_L1_error)
  __pyx_t_39 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18Type_prefixContext_1__init__, 0, __pyx_n_s_ModelicaParser_Type_prefixContex_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1483)); if (unlikely(!__pyx_t_39)) __PYX_ERR(0, 2734, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_39);
  __Pyx_INCREF(__pyx_t_39);
  PyList_Append(__pyx_t_37, __pyx_t_39);
  __Pyx_GIVEREF(__pyx_t_39);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_39, __pyx_tuple__1484);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_39, __pyx_t_38);
  __Pyx_DECREF(__pyx_t_38); __pyx_t_38 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_39) < 0) __PYX_ERR(0, 2734, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_39); __pyx_t_39 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2739
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_type_prefix
 * 
 */
  __pyx_t_39 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18Type_prefixContext_3getRuleIndex, 0, __pyx_n_s_ModelicaParser_Type_prefixContex_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1486)); if (unlikely(!__pyx_t_39)) __PYX_ERR(0, 2739, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_39);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_39) < 0) __PYX_ERR(0, 2739, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_39); __pyx_t_39 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2742
 *             return ModelicaParser.RULE_type_prefix
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterType_prefix" ):
 *                 listener.enterType_prefix(self)
 */
  __pyx_t_39 = PyDict_New(); if (unlikely(!__pyx_t_39)) __PYX_ERR(0, 2742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_39);
  __pyx_t_38 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_38)) __PYX_ERR(0, 2742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_38);
  if (PyDict_SetItem(__pyx_t_39, __pyx_n_s_listener, __pyx_t_38) < 0) __PYX_ERR(0, 2742, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_38); __pyx_t_38 = 0;
  __pyx_t_38 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18Type_prefixContext_5enterRule, 0, __pyx_n_s_ModelicaParser_Type_prefixContex_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1488)); if (unlikely(!__pyx_t_38)) __PYX_ERR(0, 2742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_38);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_38, __pyx_t_39);
  __Pyx_DECREF(__pyx_t_39); __pyx_t_39 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_38) < 0) __PYX_ERR(0, 2742, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_38); __pyx_t_38 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2746
 *                 listener.enterType_prefix(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitType_prefix" ):
 *                 listener.exitType_prefix(self)
 */
  __pyx_t_38 = PyDict_New(); if (unlikely(!__pyx_t_38)) __PYX_ERR(0, 2746, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_38);
  __pyx_t_39 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_39)) __PYX_ERR(0, 2746, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_39);
  if (PyDict_SetItem(__pyx_t_38, __pyx_n_s_listener, __pyx_t_39) < 0) __PYX_ERR(0, 2746, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_39); __pyx_t_39 = 0;
  __pyx_t_39 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18Type_prefixContext_7exitRule, 0, __pyx_n_s_ModelicaParser_Type_prefixContex_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1490)); if (unlikely(!__pyx_t_39)) __PYX_ERR(0, 2746, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_39);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_39, __pyx_t_38);
  __Pyx_DECREF(__pyx_t_38); __pyx_t_38 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_39) < 0) __PYX_ERR(0, 2746, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_39); __pyx_t_39 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2750
 *                 listener.exitType_prefix(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitType_prefix" ):
 *                 return visitor.visitType_prefix(self)
 */
  __pyx_t_39 = PyDict_New(); if (unlikely(!__pyx_t_39)) __PYX_ERR(0, 2750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_39);
  __pyx_t_38 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_38)) __PYX_ERR(0, 2750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_38);
  if (PyDict_SetItem(__pyx_t_39, __pyx_n_s_visitor, __pyx_t_38) < 0) __PYX_ERR(0, 2750, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_38); __pyx_t_38 = 0;
  __pyx_t_38 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18Type_prefixContext_9accept, 0, __pyx_n_s_ModelicaParser_Type_prefixContex_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1492)); if (unlikely(!__pyx_t_38)) __PYX_ERR(0, 2750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_38);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_38, __pyx_t_39);
  __Pyx_DECREF(__pyx_t_39); __pyx_t_39 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_38) < 0) __PYX_ERR(0, 2750, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_38); __pyx_t_38 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2732
 *         return localctx
 * 
 *     class Type_prefixContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_38 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Type_prefixContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_38)) __PYX_ERR(0, 2732, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_38);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_37, __pyx_t_38) < 0) __PYX_ERR(0, 2732, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Type_prefixContext, __pyx_t_38) < 0) __PYX_ERR(0, 2732, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_38); __pyx_t_38 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2759
 * 
 * 
 *     def type_prefix(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Type_prefixContext(self, self._ctx, self.state)
 */
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_45type_prefix, 0, __pyx_n_s_ModelicaParser_type_prefix, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1494)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2759, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_type_prefix, __pyx_t_7) < 0) __PYX_ERR(0, 2759, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2813
 *         return localctx
 * 
 *     class Type_specifierContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Type_specifierContext, __pyx_n_s_ModelicaParser_Type_specifierCon, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_38 = PyList_New(0); if (unlikely(!__pyx_t_38)) __PYX_ERR(0, 2813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_38);

  /* "pymola/generated/ModelicaParser.pyx":2815
 *     class Type_specifierContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_39 = PyDict_New(); if (unlikely(!__pyx_t_39)) __PYX_ERR(0, 2815, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_39);
  __pyx_t_40 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 2815, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_40);
  if (PyDict_SetItem(__pyx_t_39, __pyx_n_s_parent, __pyx_t_40) < 0) __PYX_ERR(0, 2815, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_40); __pyx_t_40 = 0;
  if (PyDict_SetItem(__pyx_t_39, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 2815, __pyx_L1_error)
  __pyx_t_40 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Type_specifierContext_1__init__, 0, __pyx_n_s_ModelicaParser_Type_specifierCon_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1496)); if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 2815, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_40);
  __Pyx_INCREF(__pyx_t_40);
  PyList_Append(__pyx_t_38, __pyx_t_40);
  __Pyx_GIVEREF(__pyx_t_40);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_40, __pyx_tuple__1497);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_40, __pyx_t_39);
  __Pyx_DECREF(__pyx_t_39); __pyx_t_39 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_40) < 0) __PYX_ERR(0, 2815, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_40); __pyx_t_40 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2819
 *             self.parser = parser
 * 
 *         def component_reference(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 */
  __pyx_t_40 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Type_specifierContext_3component_reference, 0, __pyx_n_s_ModelicaParser_Type_specifierCon_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1499)); if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 2819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_40);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_component_reference, __pyx_t_40) < 0) __PYX_ERR(0, 2819, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_40); __pyx_t_40 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2823
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_type_specifier
 * 
 */
  __pyx_t_40 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Type_specifierContext_5getRuleIndex, 0, __pyx_n_s_ModelicaParser_Type_specifierCon_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1501)); if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 2823, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_40);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_40) < 0) __PYX_ERR(0, 2823, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_40); __pyx_t_40 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2826
 *             return ModelicaParser.RULE_type_specifier
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterType_specifier" ):
 *                 listener.enterType_specifier(self)
 */
  __pyx_t_40 = PyDict_New(); if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 2826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_40);
  __pyx_t_39 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_39)) __PYX_ERR(0, 2826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_39);
  if (PyDict_SetItem(__pyx_t_40, __pyx_n_s_listener, __pyx_t_39) < 0) __PYX_ERR(0, 2826, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_39); __pyx_t_39 = 0;
  __pyx_t_39 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Type_specifierContext_7enterRule, 0, __pyx_n_s_ModelicaParser_Type_specifierCon_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1503)); if (unlikely(!__pyx_t_39)) __PYX_ERR(0, 2826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_39);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_39, __pyx_t_40);
  __Pyx_DECREF(__pyx_t_40); __pyx_t_40 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_39) < 0) __PYX_ERR(0, 2826, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_39); __pyx_t_39 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2830
 *                 listener.enterType_specifier(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitType_specifier" ):
 *                 listener.exitType_specifier(self)
 */
  __pyx_t_39 = PyDict_New(); if (unlikely(!__pyx_t_39)) __PYX_ERR(0, 2830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_39);
  __pyx_t_40 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 2830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_40);
  if (PyDict_SetItem(__pyx_t_39, __pyx_n_s_listener, __pyx_t_40) < 0) __PYX_ERR(0, 2830, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_40); __pyx_t_40 = 0;
  __pyx_t_40 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Type_specifierContext_9exitRule, 0, __pyx_n_s_ModelicaParser_Type_specifierCon_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1505)); if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 2830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_40);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_40, __pyx_t_39);
  __Pyx_DECREF(__pyx_t_39); __pyx_t_39 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_40) < 0) __PYX_ERR(0, 2830, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_40); __pyx_t_40 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2834
 *                 listener.exitType_specifier(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitType_specifier" ):
 *                 return visitor.visitType_specifier(self)
 */
  __pyx_t_40 = PyDict_New(); if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 2834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_40);
  __pyx_t_39 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_39)) __PYX_ERR(0, 2834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_39);
  if (PyDict_SetItem(__pyx_t_40, __pyx_n_s_visitor, __pyx_t_39) < 0) __PYX_ERR(0, 2834, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_39); __pyx_t_39 = 0;
  __pyx_t_39 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Type_specifierContext_11accept, 0, __pyx_n_s_ModelicaParser_Type_specifierCon_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1507)); if (unlikely(!__pyx_t_39)) __PYX_ERR(0, 2834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_39);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_39, __pyx_t_40);
  __Pyx_DECREF(__pyx_t_40); __pyx_t_40 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_39) < 0) __PYX_ERR(0, 2834, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_39); __pyx_t_39 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2813
 *         return localctx
 * 
 *     class Type_specifierContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_39 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Type_specifierContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_39)) __PYX_ERR(0, 2813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_39);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_38, __pyx_t_39) < 0) __PYX_ERR(0, 2813, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_38); __pyx_t_38 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Type_specifierContext, __pyx_t_39) < 0) __PYX_ERR(0, 2813, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_39); __pyx_t_39 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2843
 * 
 * 
 *     def type_specifier(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Type_specifierContext(self, self._ctx, self.state)
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_47type_specifier, 0, __pyx_n_s_ModelicaParser_type_specifier, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1509)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2843, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_type_specifier, __pyx_t_5) < 0) __PYX_ERR(0, 2843, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2859
 *         return localctx
 * 
 *     class Component_listContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Component_listContext, __pyx_n_s_ModelicaParser_Component_listCon, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_39 = PyList_New(0); if (unlikely(!__pyx_t_39)) __PYX_ERR(0, 2859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_39);

  /* "pymola/generated/ModelicaParser.pyx":2861
 *     class Component_listContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_40 = PyDict_New(); if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 2861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_40);
  __pyx_t_41 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_41)) __PYX_ERR(0, 2861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_41);
  if (PyDict_SetItem(__pyx_t_40, __pyx_n_s_parent, __pyx_t_41) < 0) __PYX_ERR(0, 2861, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_41); __pyx_t_41 = 0;
  if (PyDict_SetItem(__pyx_t_40, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 2861, __pyx_L1_error)
  __pyx_t_41 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Component_listContext_1__init__, 0, __pyx_n_s_ModelicaParser_Component_listCon_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1511)); if (unlikely(!__pyx_t_41)) __PYX_ERR(0, 2861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_41);
  __Pyx_INCREF(__pyx_t_41);
  PyList_Append(__pyx_t_39, __pyx_t_41);
  __Pyx_GIVEREF(__pyx_t_41);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_41, __pyx_tuple__1512);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_41, __pyx_t_40);
  __Pyx_DECREF(__pyx_t_40); __pyx_t_40 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_41) < 0) __PYX_ERR(0, 2861, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_41); __pyx_t_41 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2865
 *             self.parser = parser
 * 
 *         def component_declaration(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Component_declarationContext)
 */
  __pyx_t_41 = PyDict_New(); if (unlikely(!__pyx_t_41)) __PYX_ERR(0, 2865, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_41);
  if (PyDict_SetItem(__pyx_t_41, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 2865, __pyx_L1_error)
  __pyx_t_40 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Component_listContext_3component_declaration, 0, __pyx_n_s_ModelicaParser_Component_listCon_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1514)); if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 2865, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_40);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_40, __pyx_tuple__1515);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_40, __pyx_t_41);
  __Pyx_DECREF(__pyx_t_41); __pyx_t_41 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_component_declaration, __pyx_t_40) < 0) __PYX_ERR(0, 2865, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_40); __pyx_t_40 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2872
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_component_list
 * 
 */
  __pyx_t_40 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Component_listContext_5getRuleIndex, 0, __pyx_n_s_ModelicaParser_Component_listCon_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1517)); if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 2872, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_40);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_40) < 0) __PYX_ERR(0, 2872, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_40); __pyx_t_40 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2875
 *             return ModelicaParser.RULE_component_list
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterComponent_list" ):
 *                 listener.enterComponent_list(self)
 */
  __pyx_t_40 = PyDict_New(); if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 2875, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_40);
  __pyx_t_41 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_41)) __PYX_ERR(0, 2875, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_41);
  if (PyDict_SetItem(__pyx_t_40, __pyx_n_s_listener, __pyx_t_41) < 0) __PYX_ERR(0, 2875, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_41); __pyx_t_41 = 0;
  __pyx_t_41 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Component_listContext_7enterRule, 0, __pyx_n_s_ModelicaParser_Component_listCon_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1519)); if (unlikely(!__pyx_t_41)) __PYX_ERR(0, 2875, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_41);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_41, __pyx_t_40);
  __Pyx_DECREF(__pyx_t_40); __pyx_t_40 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_41) < 0) __PYX_ERR(0, 2875, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_41); __pyx_t_41 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2879
 *                 listener.enterComponent_list(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitComponent_list" ):
 *                 listener.exitComponent_list(self)
 */
  __pyx_t_41 = PyDict_New(); if (unlikely(!__pyx_t_41)) __PYX_ERR(0, 2879, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_41);
  __pyx_t_40 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 2879, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_40);
  if (PyDict_SetItem(__pyx_t_41, __pyx_n_s_listener, __pyx_t_40) < 0) __PYX_ERR(0, 2879, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_40); __pyx_t_40 = 0;
  __pyx_t_40 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Component_listContext_9exitRule, 0, __pyx_n_s_ModelicaParser_Component_listCon_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1521)); if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 2879, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_40);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_40, __pyx_t_41);
  __Pyx_DECREF(__pyx_t_41); __pyx_t_41 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_40) < 0) __PYX_ERR(0, 2879, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_40); __pyx_t_40 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2883
 *                 listener.exitComponent_list(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitComponent_list" ):
 *                 return visitor.visitComponent_list(self)
 */
  __pyx_t_40 = PyDict_New(); if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 2883, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_40);
  __pyx_t_41 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_41)) __PYX_ERR(0, 2883, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_41);
  if (PyDict_SetItem(__pyx_t_40, __pyx_n_s_visitor, __pyx_t_41) < 0) __PYX_ERR(0, 2883, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_41); __pyx_t_41 = 0;
  __pyx_t_41 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Component_listContext_11accept, 0, __pyx_n_s_ModelicaParser_Component_listCon_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1523)); if (unlikely(!__pyx_t_41)) __PYX_ERR(0, 2883, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_41);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_41, __pyx_t_40);
  __Pyx_DECREF(__pyx_t_40); __pyx_t_40 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_41) < 0) __PYX_ERR(0, 2883, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_41); __pyx_t_41 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2859
 *         return localctx
 * 
 *     class Component_listContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_41 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Component_listContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_41)) __PYX_ERR(0, 2859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_41);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_39, __pyx_t_41) < 0) __PYX_ERR(0, 2859, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_39); __pyx_t_39 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Component_listContext, __pyx_t_41) < 0) __PYX_ERR(0, 2859, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_41); __pyx_t_41 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2892
 * 
 * 
 *     def component_list(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Component_listContext(self, self._ctx, self.state)
 */
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_49component_list, 0, __pyx_n_s_ModelicaParser_component_list, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1525)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2892, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_component_list, __pyx_t_7) < 0) __PYX_ERR(0, 2892, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2921
 *         return localctx
 * 
 *     class Component_declarationContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2921, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2921, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2921, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Component_declarationContext, __pyx_n_s_ModelicaParser_Component_declara, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2921, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_41 = PyList_New(0); if (unlikely(!__pyx_t_41)) __PYX_ERR(0, 2921, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_41);

  /* "pymola/generated/ModelicaParser.pyx":2923
 *     class Component_declarationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_40 = PyDict_New(); if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 2923, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_40);
  __pyx_t_42 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_42)) __PYX_ERR(0, 2923, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_42);
  if (PyDict_SetItem(__pyx_t_40, __pyx_n_s_parent, __pyx_t_42) < 0) __PYX_ERR(0, 2923, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_42); __pyx_t_42 = 0;
  if (PyDict_SetItem(__pyx_t_40, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 2923, __pyx_L1_error)
  __pyx_t_42 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_28Component_declarationContext_1__init__, 0, __pyx_n_s_ModelicaParser_Component_declara_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1527)); if (unlikely(!__pyx_t_42)) __PYX_ERR(0, 2923, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_42);
  __Pyx_INCREF(__pyx_t_42);
  PyList_Append(__pyx_t_41, __pyx_t_42);
  __Pyx_GIVEREF(__pyx_t_42);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_42, __pyx_tuple__1528);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_42, __pyx_t_40);
  __Pyx_DECREF(__pyx_t_40); __pyx_t_40 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_42) < 0) __PYX_ERR(0, 2923, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_42); __pyx_t_42 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2927
 *             self.parser = parser
 * 
 *         def declaration(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.DeclarationContext,0)
 * 
 */
  __pyx_t_42 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_28Component_declarationContext_3declaration, 0, __pyx_n_s_ModelicaParser_Component_declara_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1530)); if (unlikely(!__pyx_t_42)) __PYX_ERR(0, 2927, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_42);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_declaration, __pyx_t_42) < 0) __PYX_ERR(0, 2927, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_42); __pyx_t_42 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2931
 * 
 * 
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */
  __pyx_t_42 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_28Component_declarationContext_5comment, 0, __pyx_n_s_ModelicaParser_Component_declara_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1532)); if (unlikely(!__pyx_t_42)) __PYX_ERR(0, 2931, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_42);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_comment, __pyx_t_42) < 0) __PYX_ERR(0, 2931, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_42); __pyx_t_42 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2935
 * 
 * 
 *         def condition_attribute(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Condition_attributeContext,0)
 * 
 */
  __pyx_t_42 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_28Component_declarationContext_7condition_attribute, 0, __pyx_n_s_ModelicaParser_Component_declara_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1534)); if (unlikely(!__pyx_t_42)) __PYX_ERR(0, 2935, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_42);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_condition_attribute, __pyx_t_42) < 0) __PYX_ERR(0, 2935, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_42); __pyx_t_42 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2939
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_component_declaration
 * 
 */
  __pyx_t_42 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_28Component_declarationContext_9getRuleIndex, 0, __pyx_n_s_ModelicaParser_Component_declara_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1536)); if (unlikely(!__pyx_t_42)) __PYX_ERR(0, 2939, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_42);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_42) < 0) __PYX_ERR(0, 2939, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_42); __pyx_t_42 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2942
 *             return ModelicaParser.RULE_component_declaration
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterComponent_declaration" ):
 *                 listener.enterComponent_declaration(self)
 */
  __pyx_t_42 = PyDict_New(); if (unlikely(!__pyx_t_42)) __PYX_ERR(0, 2942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_42);
  __pyx_t_40 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 2942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_40);
  if (PyDict_SetItem(__pyx_t_42, __pyx_n_s_listener, __pyx_t_40) < 0) __PYX_ERR(0, 2942, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_40); __pyx_t_40 = 0;
  __pyx_t_40 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_28Component_declarationContext_11enterRule, 0, __pyx_n_s_ModelicaParser_Component_declara_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1538)); if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 2942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_40);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_40, __pyx_t_42);
  __Pyx_DECREF(__pyx_t_42); __pyx_t_42 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_40) < 0) __PYX_ERR(0, 2942, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_40); __pyx_t_40 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2946
 *                 listener.enterComponent_declaration(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitComponent_declaration" ):
 *                 listener.exitComponent_declaration(self)
 */
  __pyx_t_40 = PyDict_New(); if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 2946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_40);
  __pyx_t_42 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_42)) __PYX_ERR(0, 2946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_42);
  if (PyDict_SetItem(__pyx_t_40, __pyx_n_s_listener, __pyx_t_42) < 0) __PYX_ERR(0, 2946, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_42); __pyx_t_42 = 0;
  __pyx_t_42 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_28Component_declarationContext_13exitRule, 0, __pyx_n_s_ModelicaParser_Component_declara_8, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1540)); if (unlikely(!__pyx_t_42)) __PYX_ERR(0, 2946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_42);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_42, __pyx_t_40);
  __Pyx_DECREF(__pyx_t_40); __pyx_t_40 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_42) < 0) __PYX_ERR(0, 2946, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_42); __pyx_t_42 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2950
 *                 listener.exitComponent_declaration(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitComponent_declaration" ):
 *                 return visitor.visitComponent_declaration(self)
 */
  __pyx_t_42 = PyDict_New(); if (unlikely(!__pyx_t_42)) __PYX_ERR(0, 2950, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_42);
  __pyx_t_40 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 2950, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_40);
  if (PyDict_SetItem(__pyx_t_42, __pyx_n_s_visitor, __pyx_t_40) < 0) __PYX_ERR(0, 2950, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_40); __pyx_t_40 = 0;
  __pyx_t_40 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_28Component_declarationContext_15accept, 0, __pyx_n_s_ModelicaParser_Component_declara_9, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1542)); if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 2950, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_40);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_40, __pyx_t_42);
  __Pyx_DECREF(__pyx_t_42); __pyx_t_42 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_40) < 0) __PYX_ERR(0, 2950, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_40); __pyx_t_40 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2921
 *         return localctx
 * 
 *     class Component_declarationContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_40 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Component_declarationContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 2921, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_40);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_41, __pyx_t_40) < 0) __PYX_ERR(0, 2921, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_41); __pyx_t_41 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Component_declarationContext, __pyx_t_40) < 0) __PYX_ERR(0, 2921, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_40); __pyx_t_40 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2959
 * 
 * 
 *     def component_declaration(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Component_declarationContext(self, self._ctx, self.state)
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_51component_declaration, 0, __pyx_n_s_ModelicaParser_component_declara, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1544)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2959, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_component_declaration, __pyx_t_5) < 0) __PYX_ERR(0, 2959, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2986
 *         return localctx
 * 
 *     class Condition_attributeContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2986, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2986, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2986, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Condition_attributeContext, __pyx_n_s_ModelicaParser_Condition_attribu, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2986, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_40 = PyList_New(0); if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 2986, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_40);

  /* "pymola/generated/ModelicaParser.pyx":2988
 *     class Condition_attributeContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_42 = PyDict_New(); if (unlikely(!__pyx_t_42)) __PYX_ERR(0, 2988, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_42);
  __pyx_t_43 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 2988, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_43);
  if (PyDict_SetItem(__pyx_t_42, __pyx_n_s_parent, __pyx_t_43) < 0) __PYX_ERR(0, 2988, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_43); __pyx_t_43 = 0;
  if (PyDict_SetItem(__pyx_t_42, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 2988, __pyx_L1_error)
  __pyx_t_43 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_1__init__, 0, __pyx_n_s_ModelicaParser_Condition_attribu_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1546)); if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 2988, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_43);
  __Pyx_INCREF(__pyx_t_43);
  PyList_Append(__pyx_t_40, __pyx_t_43);
  __Pyx_GIVEREF(__pyx_t_43);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_43, __pyx_tuple__1547);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_43, __pyx_t_42);
  __Pyx_DECREF(__pyx_t_42); __pyx_t_42 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_43) < 0) __PYX_ERR(0, 2988, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_43); __pyx_t_43 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2992
 *             self.parser = parser
 * 
 *         def expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 */
  __pyx_t_43 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_3expression, 0, __pyx_n_s_ModelicaParser_Condition_attribu_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1549)); if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 2992, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_43);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_expression, __pyx_t_43) < 0) __PYX_ERR(0, 2992, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_43); __pyx_t_43 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2996
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_condition_attribute
 * 
 */
  __pyx_t_43 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_5getRuleIndex, 0, __pyx_n_s_ModelicaParser_Condition_attribu_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1551)); if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 2996, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_43);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_43) < 0) __PYX_ERR(0, 2996, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_43); __pyx_t_43 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2999
 *             return ModelicaParser.RULE_condition_attribute
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterCondition_attribute" ):
 *                 listener.enterCondition_attribute(self)
 */
  __pyx_t_43 = PyDict_New(); if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 2999, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_43);
  __pyx_t_42 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_42)) __PYX_ERR(0, 2999, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_42);
  if (PyDict_SetItem(__pyx_t_43, __pyx_n_s_listener, __pyx_t_42) < 0) __PYX_ERR(0, 2999, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_42); __pyx_t_42 = 0;
  __pyx_t_42 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_7enterRule, 0, __pyx_n_s_ModelicaParser_Condition_attribu_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1553)); if (unlikely(!__pyx_t_42)) __PYX_ERR(0, 2999, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_42);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_42, __pyx_t_43);
  __Pyx_DECREF(__pyx_t_43); __pyx_t_43 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_42) < 0) __PYX_ERR(0, 2999, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_42); __pyx_t_42 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3003
 *                 listener.enterCondition_attribute(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitCondition_attribute" ):
 *                 listener.exitCondition_attribute(self)
 */
  __pyx_t_42 = PyDict_New(); if (unlikely(!__pyx_t_42)) __PYX_ERR(0, 3003, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_42);
  __pyx_t_43 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 3003, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_43);
  if (PyDict_SetItem(__pyx_t_42, __pyx_n_s_listener, __pyx_t_43) < 0) __PYX_ERR(0, 3003, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_43); __pyx_t_43 = 0;
  __pyx_t_43 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_9exitRule, 0, __pyx_n_s_ModelicaParser_Condition_attribu_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1555)); if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 3003, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_43);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_43, __pyx_t_42);
  __Pyx_DECREF(__pyx_t_42); __pyx_t_42 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_43) < 0) __PYX_ERR(0, 3003, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_43); __pyx_t_43 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3007
 *                 listener.exitCondition_attribute(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitCondition_attribute" ):
 *                 return visitor.visitCondition_attribute(self)
 */
  __pyx_t_43 = PyDict_New(); if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 3007, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_43);
  __pyx_t_42 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_42)) __PYX_ERR(0, 3007, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_42);
  if (PyDict_SetItem(__pyx_t_43, __pyx_n_s_visitor, __pyx_t_42) < 0) __PYX_ERR(0, 3007, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_42); __pyx_t_42 = 0;
  __pyx_t_42 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Condition_attributeContext_11accept, 0, __pyx_n_s_ModelicaParser_Condition_attribu_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1557)); if (unlikely(!__pyx_t_42)) __PYX_ERR(0, 3007, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_42);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_42, __pyx_t_43);
  __Pyx_DECREF(__pyx_t_43); __pyx_t_43 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_42) < 0) __PYX_ERR(0, 3007, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_42); __pyx_t_42 = 0;

  /* "pymola/generated/ModelicaParser.pyx":2986
 *         return localctx
 * 
 *     class Condition_attributeContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_42 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Condition_attributeContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_42)) __PYX_ERR(0, 2986, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_42);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_40, __pyx_t_42) < 0) __PYX_ERR(0, 2986, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_40); __pyx_t_40 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Condition_attributeContext, __pyx_t_42) < 0) __PYX_ERR(0, 2986, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_42); __pyx_t_42 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3016
 * 
 * 
 *     def condition_attribute(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Condition_attributeContext(self, self._ctx, self.state)
 */
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_53condition_attribute, 0, __pyx_n_s_ModelicaParser_condition_attribu, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1559)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3016, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_condition_attribute, __pyx_t_7) < 0) __PYX_ERR(0, 3016, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3034
 *         return localctx
 * 
 *     class DeclarationContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_DeclarationContext, __pyx_n_s_ModelicaParser_DeclarationContex, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_42 = PyList_New(0); if (unlikely(!__pyx_t_42)) __PYX_ERR(0, 3034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_42);

  /* "pymola/generated/ModelicaParser.pyx":3036
 *     class DeclarationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_43 = PyDict_New(); if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 3036, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_43);
  __pyx_t_44 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_44)) __PYX_ERR(0, 3036, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_44);
  if (PyDict_SetItem(__pyx_t_43, __pyx_n_s_parent, __pyx_t_44) < 0) __PYX_ERR(0, 3036, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_44); __pyx_t_44 = 0;
  if (PyDict_SetItem(__pyx_t_43, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 3036, __pyx_L1_error)
  __pyx_t_44 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18DeclarationContext_1__init__, 0, __pyx_n_s_ModelicaParser_DeclarationContex_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1561)); if (unlikely(!__pyx_t_44)) __PYX_ERR(0, 3036, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_44);
  __Pyx_INCREF(__pyx_t_44);
  PyList_Append(__pyx_t_42, __pyx_t_44);
  __Pyx_GIVEREF(__pyx_t_44);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_44, __pyx_tuple__1562);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_44, __pyx_t_43);
  __Pyx_DECREF(__pyx_t_43); __pyx_t_43 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_44) < 0) __PYX_ERR(0, 3036, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_44); __pyx_t_44 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3040
 *             self.parser = parser
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 */
  __pyx_t_44 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18DeclarationContext_3IDENT, 0, __pyx_n_s_ModelicaParser_DeclarationContex_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1564)); if (unlikely(!__pyx_t_44)) __PYX_ERR(0, 3040, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_44);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_IDENT, __pyx_t_44) < 0) __PYX_ERR(0, 3040, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_44); __pyx_t_44 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3043
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 *         def array_subscripts(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Array_subscriptsContext,0)
 * 
 */
  __pyx_t_44 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18DeclarationContext_5array_subscripts, 0, __pyx_n_s_ModelicaParser_DeclarationContex_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1566)); if (unlikely(!__pyx_t_44)) __PYX_ERR(0, 3043, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_44);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_array_subscripts, __pyx_t_44) < 0) __PYX_ERR(0, 3043, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_44); __pyx_t_44 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3047
 * 
 * 
 *         def modification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ModificationContext,0)
 * 
 */
  __pyx_t_44 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18DeclarationContext_7modification, 0, __pyx_n_s_ModelicaParser_DeclarationContex_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1568)); if (unlikely(!__pyx_t_44)) __PYX_ERR(0, 3047, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_44);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_modification, __pyx_t_44) < 0) __PYX_ERR(0, 3047, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_44); __pyx_t_44 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3051
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_declaration
 * 
 */
  __pyx_t_44 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18DeclarationContext_9getRuleIndex, 0, __pyx_n_s_ModelicaParser_DeclarationContex_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1570)); if (unlikely(!__pyx_t_44)) __PYX_ERR(0, 3051, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_44);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_44) < 0) __PYX_ERR(0, 3051, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_44); __pyx_t_44 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3054
 *             return ModelicaParser.RULE_declaration
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterDeclaration" ):
 *                 listener.enterDeclaration(self)
 */
  __pyx_t_44 = PyDict_New(); if (unlikely(!__pyx_t_44)) __PYX_ERR(0, 3054, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_44);
  __pyx_t_43 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 3054, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_43);
  if (PyDict_SetItem(__pyx_t_44, __pyx_n_s_listener, __pyx_t_43) < 0) __PYX_ERR(0, 3054, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_43); __pyx_t_43 = 0;
  __pyx_t_43 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18DeclarationContext_11enterRule, 0, __pyx_n_s_ModelicaParser_DeclarationContex_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1572)); if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 3054, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_43);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_43, __pyx_t_44);
  __Pyx_DECREF(__pyx_t_44); __pyx_t_44 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_43) < 0) __PYX_ERR(0, 3054, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_43); __pyx_t_43 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3058
 *                 listener.enterDeclaration(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitDeclaration" ):
 *                 listener.exitDeclaration(self)
 */
  __pyx_t_43 = PyDict_New(); if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 3058, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_43);
  __pyx_t_44 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_44)) __PYX_ERR(0, 3058, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_44);
  if (PyDict_SetItem(__pyx_t_43, __pyx_n_s_listener, __pyx_t_44) < 0) __PYX_ERR(0, 3058, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_44); __pyx_t_44 = 0;
  __pyx_t_44 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18DeclarationContext_13exitRule, 0, __pyx_n_s_ModelicaParser_DeclarationContex_8, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1574)); if (unlikely(!__pyx_t_44)) __PYX_ERR(0, 3058, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_44);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_44, __pyx_t_43);
  __Pyx_DECREF(__pyx_t_43); __pyx_t_43 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_44) < 0) __PYX_ERR(0, 3058, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_44); __pyx_t_44 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3062
 *                 listener.exitDeclaration(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitDeclaration" ):
 *                 return visitor.visitDeclaration(self)
 */
  __pyx_t_44 = PyDict_New(); if (unlikely(!__pyx_t_44)) __PYX_ERR(0, 3062, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_44);
  __pyx_t_43 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 3062, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_43);
  if (PyDict_SetItem(__pyx_t_44, __pyx_n_s_visitor, __pyx_t_43) < 0) __PYX_ERR(0, 3062, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_43); __pyx_t_43 = 0;
  __pyx_t_43 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18DeclarationContext_15accept, 0, __pyx_n_s_ModelicaParser_DeclarationContex_9, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1576)); if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 3062, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_43);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_43, __pyx_t_44);
  __Pyx_DECREF(__pyx_t_44); __pyx_t_44 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_43) < 0) __PYX_ERR(0, 3062, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_43); __pyx_t_43 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3034
 *         return localctx
 * 
 *     class DeclarationContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_43 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_DeclarationContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 3034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_43);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_42, __pyx_t_43) < 0) __PYX_ERR(0, 3034, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_42); __pyx_t_42 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_DeclarationContext, __pyx_t_43) < 0) __PYX_ERR(0, 3034, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_43); __pyx_t_43 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3071
 * 
 * 
 *     def declaration(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.DeclarationContext(self, self._ctx, self.state)
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_55declaration, 0, __pyx_n_s_ModelicaParser_declaration, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1578)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3071, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_declaration, __pyx_t_5) < 0) __PYX_ERR(0, 3071, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3104
 *         return localctx
 * 
 *     class ModificationContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_ModificationContext, __pyx_n_s_ModelicaParser_ModificationConte, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_43 = PyList_New(0); if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 3104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_43);

  /* "pymola/generated/ModelicaParser.pyx":3106
 *     class ModificationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_44 = PyDict_New(); if (unlikely(!__pyx_t_44)) __PYX_ERR(0, 3106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_44);
  __pyx_t_45 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 3106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_45);
  if (PyDict_SetItem(__pyx_t_44, __pyx_n_s_parent, __pyx_t_45) < 0) __PYX_ERR(0, 3106, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;
  if (PyDict_SetItem(__pyx_t_44, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 3106, __pyx_L1_error)
  __pyx_t_45 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19ModificationContext_1__init__, 0, __pyx_n_s_ModelicaParser_ModificationConte_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1580)); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 3106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_45);
  __Pyx_INCREF(__pyx_t_45);
  PyList_Append(__pyx_t_43, __pyx_t_45);
  __Pyx_GIVEREF(__pyx_t_45);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_45, __pyx_tuple__1581);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_45, __pyx_t_44);
  __Pyx_DECREF(__pyx_t_44); __pyx_t_44 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_45) < 0) __PYX_ERR(0, 3106, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3111
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_modification
 * 
 */
  __pyx_t_45 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19ModificationContext_3getRuleIndex, 0, __pyx_n_s_ModelicaParser_ModificationConte_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1583)); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 3111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_45);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_45) < 0) __PYX_ERR(0, 3111, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3115
 * 
 * 
 *         def copyFrom(self, ctx:ParserRuleContext):             # <<<<<<<<<<<<<<
 *             super().copyFrom(ctx)
 * 
 */
  __pyx_t_45 = PyDict_New(); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 3115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_45);
  __pyx_t_44 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_44)) __PYX_ERR(0, 3115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_44);
  if (PyDict_SetItem(__pyx_t_45, __pyx_n_s_ctx_2, __pyx_t_44) < 0) __PYX_ERR(0, 3115, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_44); __pyx_t_44 = 0;
  __pyx_t_44 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19ModificationContext_5copyFrom, 0, __pyx_n_s_ModelicaParser_ModificationConte_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1585)); if (unlikely(!__pyx_t_44)) __PYX_ERR(0, 3115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_44);
  __Pyx_INCREF(__pyx_t_44);
  PyList_Append(__pyx_t_43, __pyx_t_44);
  __Pyx_GIVEREF(__pyx_t_44);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_44, __pyx_t_45);
  __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_copyFrom, __pyx_t_44) < 0) __PYX_ERR(0, 3115, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_44); __pyx_t_44 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3104
 *         return localctx
 * 
 *     class ModificationContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_44 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_ModificationContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_44)) __PYX_ERR(0, 3104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_44);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_43, __pyx_t_44) < 0) __PYX_ERR(0, 3104, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_43); __pyx_t_43 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_ModificationContext, __pyx_t_44) < 0) __PYX_ERR(0, 3104, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_44); __pyx_t_44 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3120
 * 
 * 
 *     class Modification_classContext(ModificationContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ModificationContext
 */
  __pyx_t_7 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_ModificationContext);
  if (unlikely(!__pyx_t_7)) {
    PyErr_Clear();
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModificationContext);
  }
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Modification_classContext, __pyx_n_s_ModelicaParser_Modification_clas, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_44 = PyList_New(0); if (unlikely(!__pyx_t_44)) __PYX_ERR(0, 3120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_44);

  /* "pymola/generated/ModelicaParser.pyx":3122
 *     class Modification_classContext(ModificationContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ModificationContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_45 = PyDict_New(); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 3122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_45);
  __pyx_t_46 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 3122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_46);
  if (PyDict_SetItem(__pyx_t_45, __pyx_n_s_ctx_2, __pyx_t_46) < 0) __PYX_ERR(0, 3122, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
  __pyx_t_46 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_25Modification_classContext_1__init__, 0, __pyx_n_s_ModelicaParser_Modification_clas_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1587)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 3122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_46);
  __Pyx_INCREF(__pyx_t_46);
  PyList_Append(__pyx_t_44, __pyx_t_46);
  __Pyx_GIVEREF(__pyx_t_46);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_46, __pyx_t_45);
  __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_46) < 0) __PYX_ERR(0, 3122, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3126
 *             self.copyFrom(ctx)
 * 
 *         def class_modification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_modificationContext,0)
 * 
 */
  __pyx_t_46 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_25Modification_classContext_3class_modification, 0, __pyx_n_s_ModelicaParser_Modification_clas_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1589)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 3126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_46);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_class_modification, __pyx_t_46) < 0) __PYX_ERR(0, 3126, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3129
 *             return self.getTypedRuleContext(ModelicaParser.Class_modificationContext,0)
 * 
 *         def expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 */
  __pyx_t_46 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_25Modification_classContext_5expression, 0, __pyx_n_s_ModelicaParser_Modification_clas_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1591)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 3129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_46);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_expression, __pyx_t_46) < 0) __PYX_ERR(0, 3129, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3133
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterModification_class" ):
 *                 listener.enterModification_class(self)
 */
  __pyx_t_46 = PyDict_New(); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 3133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_46);
  __pyx_t_45 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 3133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_45);
  if (PyDict_SetItem(__pyx_t_46, __pyx_n_s_listener, __pyx_t_45) < 0) __PYX_ERR(0, 3133, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;
  __pyx_t_45 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_25Modification_classContext_7enterRule, 0, __pyx_n_s_ModelicaParser_Modification_clas_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1593)); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 3133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_45);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_45, __pyx_t_46);
  __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_45) < 0) __PYX_ERR(0, 3133, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3137
 *                 listener.enterModification_class(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitModification_class" ):
 *                 listener.exitModification_class(self)
 */
  __pyx_t_45 = PyDict_New(); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 3137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_45);
  __pyx_t_46 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 3137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_46);
  if (PyDict_SetItem(__pyx_t_45, __pyx_n_s_listener, __pyx_t_46) < 0) __PYX_ERR(0, 3137, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
  __pyx_t_46 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_25Modification_classContext_9exitRule, 0, __pyx_n_s_ModelicaParser_Modification_clas_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1595)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 3137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_46);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_46, __pyx_t_45);
  __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_46) < 0) __PYX_ERR(0, 3137, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3141
 *                 listener.exitModification_class(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitModification_class" ):
 *                 return visitor.visitModification_class(self)
 */
  __pyx_t_46 = PyDict_New(); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 3141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_46);
  __pyx_t_45 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 3141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_45);
  if (PyDict_SetItem(__pyx_t_46, __pyx_n_s_visitor, __pyx_t_45) < 0) __PYX_ERR(0, 3141, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;
  __pyx_t_45 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_25Modification_classContext_11accept, 0, __pyx_n_s_ModelicaParser_Modification_clas_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1597)); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 3141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_45);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_45, __pyx_t_46);
  __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_45) < 0) __PYX_ERR(0, 3141, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3120
 * 
 * 
 *     class Modification_classContext(ModificationContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ModificationContext
 */
  __pyx_t_45 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Modification_classContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 3120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_45);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_44, __pyx_t_45) < 0) __PYX_ERR(0, 3120, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_44); __pyx_t_44 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Modification_classContext, __pyx_t_45) < 0) __PYX_ERR(0, 3120, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3148
 * 
 * 
 *     class Modification_assignment2Context(ModificationContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ModificationContext
 */
  __pyx_t_5 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_ModificationContext);
  if (unlikely(!__pyx_t_5)) {
    PyErr_Clear();
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModificationContext);
  }
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Modification_assignment2Context, __pyx_n_s_ModelicaParser_Modification_assi, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_45 = PyList_New(0); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 3148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_45);

  /* "pymola/generated/ModelicaParser.pyx":3150
 *     class Modification_assignment2Context(ModificationContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ModificationContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_46 = PyDict_New(); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 3150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_46);
  __pyx_t_47 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 3150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_47);
  if (PyDict_SetItem(__pyx_t_46, __pyx_n_s_ctx_2, __pyx_t_47) < 0) __PYX_ERR(0, 3150, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
  __pyx_t_47 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_31Modification_assignment2Context_1__init__, 0, __pyx_n_s_ModelicaParser_Modification_assi_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1599)); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 3150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_47);
  __Pyx_INCREF(__pyx_t_47);
  PyList_Append(__pyx_t_45, __pyx_t_47);
  __Pyx_GIVEREF(__pyx_t_47);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_47, __pyx_t_46);
  __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_47) < 0) __PYX_ERR(0, 3150, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3154
 *             self.copyFrom(ctx)
 * 
 *         def expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 */
  __pyx_t_47 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_31Modification_assignment2Context_3expression, 0, __pyx_n_s_ModelicaParser_Modification_assi_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1601)); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 3154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_47);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_expression, __pyx_t_47) < 0) __PYX_ERR(0, 3154, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3158
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterModification_assignment2" ):
 *                 listener.enterModification_assignment2(self)
 */
  __pyx_t_47 = PyDict_New(); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 3158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_47);
  __pyx_t_46 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 3158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_46);
  if (PyDict_SetItem(__pyx_t_47, __pyx_n_s_listener, __pyx_t_46) < 0) __PYX_ERR(0, 3158, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
  __pyx_t_46 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_31Modification_assignment2Context_5enterRule, 0, __pyx_n_s_ModelicaParser_Modification_assi_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1603)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 3158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_46);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_46, __pyx_t_47);
  __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_46) < 0) __PYX_ERR(0, 3158, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3162
 *                 listener.enterModification_assignment2(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitModification_assignment2" ):
 *                 listener.exitModification_assignment2(self)
 */
  __pyx_t_46 = PyDict_New(); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 3162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_46);
  __pyx_t_47 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 3162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_47);
  if (PyDict_SetItem(__pyx_t_46, __pyx_n_s_listener, __pyx_t_47) < 0) __PYX_ERR(0, 3162, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
  __pyx_t_47 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_31Modification_assignment2Context_7exitRule, 0, __pyx_n_s_ModelicaParser_Modification_assi_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1605)); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 3162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_47);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_47, __pyx_t_46);
  __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_47) < 0) __PYX_ERR(0, 3162, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3166
 *                 listener.exitModification_assignment2(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitModification_assignment2" ):
 *                 return visitor.visitModification_assignment2(self)
 */
  __pyx_t_47 = PyDict_New(); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 3166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_47);
  __pyx_t_46 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 3166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_46);
  if (PyDict_SetItem(__pyx_t_47, __pyx_n_s_visitor, __pyx_t_46) < 0) __PYX_ERR(0, 3166, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
  __pyx_t_46 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_31Modification_assignment2Context_9accept, 0, __pyx_n_s_ModelicaParser_Modification_assi_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1607)); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 3166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_46);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_46, __pyx_t_47);
  __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_46) < 0) __PYX_ERR(0, 3166, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3148
 * 
 * 
 *     class Modification_assignment2Context(ModificationContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ModificationContext
 */
  __pyx_t_46 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Modification_assignment2Context, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 3148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_46);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_45, __pyx_t_46) < 0) __PYX_ERR(0, 3148, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Modification_assignment2Context, __pyx_t_46) < 0) __PYX_ERR(0, 3148, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3173
 * 
 * 
 *     class Modification_assignmentContext(ModificationContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ModificationContext
 */
  __pyx_t_7 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_ModificationContext);
  if (unlikely(!__pyx_t_7)) {
    PyErr_Clear();
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModificationContext);
  }
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Modification_assignmentContext, __pyx_n_s_ModelicaParser_Modification_assi_7, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_46 = PyList_New(0); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 3173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_46);

  /* "pymola/generated/ModelicaParser.pyx":3175
 *     class Modification_assignmentContext(ModificationContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ModificationContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_47 = PyDict_New(); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 3175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_47);
  __pyx_t_48 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_48)) __PYX_ERR(0, 3175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  if (PyDict_SetItem(__pyx_t_47, __pyx_n_s_ctx_2, __pyx_t_48) < 0) __PYX_ERR(0, 3175, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;
  __pyx_t_48 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_30Modification_assignmentContext_1__init__, 0, __pyx_n_s_ModelicaParser_Modification_assi_8, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1609)); if (unlikely(!__pyx_t_48)) __PYX_ERR(0, 3175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  __Pyx_INCREF(__pyx_t_48);
  PyList_Append(__pyx_t_46, __pyx_t_48);
  __Pyx_GIVEREF(__pyx_t_48);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_48, __pyx_t_47);
  __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_48) < 0) __PYX_ERR(0, 3175, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3179
 *             self.copyFrom(ctx)
 * 
 *         def expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 */
  __pyx_t_48 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_30Modification_assignmentContext_3expression, 0, __pyx_n_s_ModelicaParser_Modification_assi_9, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1611)); if (unlikely(!__pyx_t_48)) __PYX_ERR(0, 3179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_expression, __pyx_t_48) < 0) __PYX_ERR(0, 3179, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3183
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterModification_assignment" ):
 *                 listener.enterModification_assignment(self)
 */
  __pyx_t_48 = PyDict_New(); if (unlikely(!__pyx_t_48)) __PYX_ERR(0, 3183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  __pyx_t_47 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 3183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_47);
  if (PyDict_SetItem(__pyx_t_48, __pyx_n_s_listener, __pyx_t_47) < 0) __PYX_ERR(0, 3183, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
  __pyx_t_47 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_30Modification_assignmentContext_5enterRule, 0, __pyx_n_s_ModelicaParser_Modification_assi_10, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1613)); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 3183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_47);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_47, __pyx_t_48);
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_47) < 0) __PYX_ERR(0, 3183, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3187
 *                 listener.enterModification_assignment(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitModification_assignment" ):
 *                 listener.exitModification_assignment(self)
 */
  __pyx_t_47 = PyDict_New(); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 3187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_47);
  __pyx_t_48 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_48)) __PYX_ERR(0, 3187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  if (PyDict_SetItem(__pyx_t_47, __pyx_n_s_listener, __pyx_t_48) < 0) __PYX_ERR(0, 3187, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;
  __pyx_t_48 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_30Modification_assignmentContext_7exitRule, 0, __pyx_n_s_ModelicaParser_Modification_assi_11, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1615)); if (unlikely(!__pyx_t_48)) __PYX_ERR(0, 3187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_48, __pyx_t_47);
  __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_48) < 0) __PYX_ERR(0, 3187, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3191
 *                 listener.exitModification_assignment(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitModification_assignment" ):
 *                 return visitor.visitModification_assignment(self)
 */
  __pyx_t_48 = PyDict_New(); if (unlikely(!__pyx_t_48)) __PYX_ERR(0, 3191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  __pyx_t_47 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 3191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_47);
  if (PyDict_SetItem(__pyx_t_48, __pyx_n_s_visitor, __pyx_t_47) < 0) __PYX_ERR(0, 3191, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
  __pyx_t_47 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_30Modification_assignmentContext_9accept, 0, __pyx_n_s_ModelicaParser_Modification_assi_12, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1617)); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 3191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_47);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_47, __pyx_t_48);
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_47) < 0) __PYX_ERR(0, 3191, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3173
 * 
 * 
 *     class Modification_assignmentContext(ModificationContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ModificationContext
 */
  __pyx_t_47 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Modification_assignmentContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 3173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_47);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_46, __pyx_t_47) < 0) __PYX_ERR(0, 3173, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Modification_assignmentContext, __pyx_t_47) < 0) __PYX_ERR(0, 3173, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3199
 * 
 * 
 *     def modification(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.ModificationContext(self, self._ctx, self.state)
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_57modification, 0, __pyx_n_s_ModelicaParser_modification, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1619)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_modification, __pyx_t_5) < 0) __PYX_ERR(0, 3199, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3251
 *         return localctx
 * 
 *     class Class_modificationContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Class_modificationContext, __pyx_n_s_ModelicaParser_Class_modificatio, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_47 = PyList_New(0); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 3251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_47);

  /* "pymola/generated/ModelicaParser.pyx":3253
 *     class Class_modificationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_48 = PyDict_New(); if (unlikely(!__pyx_t_48)) __PYX_ERR(0, 3253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  __pyx_t_49 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_49)) __PYX_ERR(0, 3253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_49);
  if (PyDict_SetItem(__pyx_t_48, __pyx_n_s_parent, __pyx_t_49) < 0) __PYX_ERR(0, 3253, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_49); __pyx_t_49 = 0;
  if (PyDict_SetItem(__pyx_t_48, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 3253, __pyx_L1_error)
  __pyx_t_49 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_25Class_modificationContext_1__init__, 0, __pyx_n_s_ModelicaParser_Class_modificatio_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1621)); if (unlikely(!__pyx_t_49)) __PYX_ERR(0, 3253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_49);
  __Pyx_INCREF(__pyx_t_49);
  PyList_Append(__pyx_t_47, __pyx_t_49);
  __Pyx_GIVEREF(__pyx_t_49);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_49, __pyx_tuple__1622);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_49, __pyx_t_48);
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_49) < 0) __PYX_ERR(0, 3253, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_49); __pyx_t_49 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3257
 *             self.parser = parser
 * 
 *         def argument_list(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Argument_listContext,0)
 * 
 */
  __pyx_t_49 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_25Class_modificationContext_3argument_list, 0, __pyx_n_s_ModelicaParser_Class_modificatio_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1624)); if (unlikely(!__pyx_t_49)) __PYX_ERR(0, 3257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_49);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_argument_list, __pyx_t_49) < 0) __PYX_ERR(0, 3257, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_49); __pyx_t_49 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3261
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_class_modification
 * 
 */
  __pyx_t_49 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_25Class_modificationContext_5getRuleIndex, 0, __pyx_n_s_ModelicaParser_Class_modificatio_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1626)); if (unlikely(!__pyx_t_49)) __PYX_ERR(0, 3261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_49);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_49) < 0) __PYX_ERR(0, 3261, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_49); __pyx_t_49 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3264
 *             return ModelicaParser.RULE_class_modification
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterClass_modification" ):
 *                 listener.enterClass_modification(self)
 */
  __pyx_t_49 = PyDict_New(); if (unlikely(!__pyx_t_49)) __PYX_ERR(0, 3264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_49);
  __pyx_t_48 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_48)) __PYX_ERR(0, 3264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  if (PyDict_SetItem(__pyx_t_49, __pyx_n_s_listener, __pyx_t_48) < 0) __PYX_ERR(0, 3264, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;
  __pyx_t_48 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_25Class_modificationContext_7enterRule, 0, __pyx_n_s_ModelicaParser_Class_modificatio_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1628)); if (unlikely(!__pyx_t_48)) __PYX_ERR(0, 3264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_48, __pyx_t_49);
  __Pyx_DECREF(__pyx_t_49); __pyx_t_49 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_48) < 0) __PYX_ERR(0, 3264, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3268
 *                 listener.enterClass_modification(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitClass_modification" ):
 *                 listener.exitClass_modification(self)
 */
  __pyx_t_48 = PyDict_New(); if (unlikely(!__pyx_t_48)) __PYX_ERR(0, 3268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  __pyx_t_49 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_49)) __PYX_ERR(0, 3268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_49);
  if (PyDict_SetItem(__pyx_t_48, __pyx_n_s_listener, __pyx_t_49) < 0) __PYX_ERR(0, 3268, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_49); __pyx_t_49 = 0;
  __pyx_t_49 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_25Class_modificationContext_9exitRule, 0, __pyx_n_s_ModelicaParser_Class_modificatio_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1630)); if (unlikely(!__pyx_t_49)) __PYX_ERR(0, 3268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_49);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_49, __pyx_t_48);
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_49) < 0) __PYX_ERR(0, 3268, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_49); __pyx_t_49 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3272
 *                 listener.exitClass_modification(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitClass_modification" ):
 *                 return visitor.visitClass_modification(self)
 */
  __pyx_t_49 = PyDict_New(); if (unlikely(!__pyx_t_49)) __PYX_ERR(0, 3272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_49);
  __pyx_t_48 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_48)) __PYX_ERR(0, 3272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  if (PyDict_SetItem(__pyx_t_49, __pyx_n_s_visitor, __pyx_t_48) < 0) __PYX_ERR(0, 3272, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;
  __pyx_t_48 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_25Class_modificationContext_11accept, 0, __pyx_n_s_ModelicaParser_Class_modificatio_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1632)); if (unlikely(!__pyx_t_48)) __PYX_ERR(0, 3272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_48, __pyx_t_49);
  __Pyx_DECREF(__pyx_t_49); __pyx_t_49 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_48) < 0) __PYX_ERR(0, 3272, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3251
 *         return localctx
 * 
 *     class Class_modificationContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_48 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Class_modificationContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_48)) __PYX_ERR(0, 3251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_47, __pyx_t_48) < 0) __PYX_ERR(0, 3251, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Class_modificationContext, __pyx_t_48) < 0) __PYX_ERR(0, 3251, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3281
 * 
 * 
 *     def class_modification(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Class_modificationContext(self, self._ctx, self.state)
 */
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_59class_modification, 0, __pyx_n_s_ModelicaParser_class_modificatio, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1634)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_class_modification, __pyx_t_7) < 0) __PYX_ERR(0, 3281, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3308
 *         return localctx
 * 
 *     class Argument_listContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Argument_listContext, __pyx_n_s_ModelicaParser_Argument_listCont, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_48 = PyList_New(0); if (unlikely(!__pyx_t_48)) __PYX_ERR(0, 3308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);

  /* "pymola/generated/ModelicaParser.pyx":3310
 *     class Argument_listContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_49 = PyDict_New(); if (unlikely(!__pyx_t_49)) __PYX_ERR(0, 3310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_49);
  __pyx_t_50 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_50)) __PYX_ERR(0, 3310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_50);
  if (PyDict_SetItem(__pyx_t_49, __pyx_n_s_parent, __pyx_t_50) < 0) __PYX_ERR(0, 3310, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_50); __pyx_t_50 = 0;
  if (PyDict_SetItem(__pyx_t_49, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 3310, __pyx_L1_error)
  __pyx_t_50 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20Argument_listContext_1__init__, 0, __pyx_n_s_ModelicaParser_Argument_listCont_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1636)); if (unlikely(!__pyx_t_50)) __PYX_ERR(0, 3310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_50);
  __Pyx_INCREF(__pyx_t_50);
  PyList_Append(__pyx_t_48, __pyx_t_50);
  __Pyx_GIVEREF(__pyx_t_50);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_50, __pyx_tuple__1637);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_50, __pyx_t_49);
  __Pyx_DECREF(__pyx_t_49); __pyx_t_49 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_50) < 0) __PYX_ERR(0, 3310, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_50); __pyx_t_50 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3314
 *             self.parser = parser
 * 
 *         def argument(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ArgumentContext)
 */
  __pyx_t_50 = PyDict_New(); if (unlikely(!__pyx_t_50)) __PYX_ERR(0, 3314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_50);
  if (PyDict_SetItem(__pyx_t_50, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 3314, __pyx_L1_error)
  __pyx_t_49 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20Argument_listContext_3argument, 0, __pyx_n_s_ModelicaParser_Argument_listCont_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1639)); if (unlikely(!__pyx_t_49)) __PYX_ERR(0, 3314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_49);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_49, __pyx_tuple__1640);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_49, __pyx_t_50);
  __Pyx_DECREF(__pyx_t_50); __pyx_t_50 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_argument, __pyx_t_49) < 0) __PYX_ERR(0, 3314, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_49); __pyx_t_49 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3321
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_argument_list
 * 
 */
  __pyx_t_49 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20Argument_listContext_5getRuleIndex, 0, __pyx_n_s_ModelicaParser_Argument_listCont_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1642)); if (unlikely(!__pyx_t_49)) __PYX_ERR(0, 3321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_49);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_49) < 0) __PYX_ERR(0, 3321, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_49); __pyx_t_49 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3324
 *             return ModelicaParser.RULE_argument_list
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterArgument_list" ):
 *                 listener.enterArgument_list(self)
 */
  __pyx_t_49 = PyDict_New(); if (unlikely(!__pyx_t_49)) __PYX_ERR(0, 3324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_49);
  __pyx_t_50 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_50)) __PYX_ERR(0, 3324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_50);
  if (PyDict_SetItem(__pyx_t_49, __pyx_n_s_listener, __pyx_t_50) < 0) __PYX_ERR(0, 3324, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_50); __pyx_t_50 = 0;
  __pyx_t_50 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20Argument_listContext_7enterRule, 0, __pyx_n_s_ModelicaParser_Argument_listCont_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1644)); if (unlikely(!__pyx_t_50)) __PYX_ERR(0, 3324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_50);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_50, __pyx_t_49);
  __Pyx_DECREF(__pyx_t_49); __pyx_t_49 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_50) < 0) __PYX_ERR(0, 3324, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_50); __pyx_t_50 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3328
 *                 listener.enterArgument_list(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitArgument_list" ):
 *                 listener.exitArgument_list(self)
 */
  __pyx_t_50 = PyDict_New(); if (unlikely(!__pyx_t_50)) __PYX_ERR(0, 3328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_50);
  __pyx_t_49 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_49)) __PYX_ERR(0, 3328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_49);
  if (PyDict_SetItem(__pyx_t_50, __pyx_n_s_listener, __pyx_t_49) < 0) __PYX_ERR(0, 3328, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_49); __pyx_t_49 = 0;
  __pyx_t_49 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20Argument_listContext_9exitRule, 0, __pyx_n_s_ModelicaParser_Argument_listCont_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1646)); if (unlikely(!__pyx_t_49)) __PYX_ERR(0, 3328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_49);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_49, __pyx_t_50);
  __Pyx_DECREF(__pyx_t_50); __pyx_t_50 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_49) < 0) __PYX_ERR(0, 3328, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_49); __pyx_t_49 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3332
 *                 listener.exitArgument_list(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitArgument_list" ):
 *                 return visitor.visitArgument_list(self)
 */
  __pyx_t_49 = PyDict_New(); if (unlikely(!__pyx_t_49)) __PYX_ERR(0, 3332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_49);
  __pyx_t_50 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_50)) __PYX_ERR(0, 3332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_50);
  if (PyDict_SetItem(__pyx_t_49, __pyx_n_s_visitor, __pyx_t_50) < 0) __PYX_ERR(0, 3332, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_50); __pyx_t_50 = 0;
  __pyx_t_50 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20Argument_listContext_11accept, 0, __pyx_n_s_ModelicaParser_Argument_listCont_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1648)); if (unlikely(!__pyx_t_50)) __PYX_ERR(0, 3332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_50);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_50, __pyx_t_49);
  __Pyx_DECREF(__pyx_t_49); __pyx_t_49 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_50) < 0) __PYX_ERR(0, 3332, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_50); __pyx_t_50 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3308
 *         return localctx
 * 
 *     class Argument_listContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_50 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Argument_listContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_50)) __PYX_ERR(0, 3308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_50);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_48, __pyx_t_50) < 0) __PYX_ERR(0, 3308, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Argument_listContext, __pyx_t_50) < 0) __PYX_ERR(0, 3308, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_50); __pyx_t_50 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3341
 * 
 * 
 *     def argument_list(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Argument_listContext(self, self._ctx, self.state)
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_61argument_list, 0, __pyx_n_s_ModelicaParser_argument_list, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1650)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_argument_list, __pyx_t_5) < 0) __PYX_ERR(0, 3341, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3370
 *         return localctx
 * 
 *     class ArgumentContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3370, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3370, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3370, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_ArgumentContext, __pyx_n_s_ModelicaParser_ArgumentContext, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3370, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_50 = PyList_New(0); if (unlikely(!__pyx_t_50)) __PYX_ERR(0, 3370, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_50);

  /* "pymola/generated/ModelicaParser.pyx":3372
 *     class ArgumentContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_49 = PyDict_New(); if (unlikely(!__pyx_t_49)) __PYX_ERR(0, 3372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_49);
  __pyx_t_51 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_51)) __PYX_ERR(0, 3372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_51);
  if (PyDict_SetItem(__pyx_t_49, __pyx_n_s_parent, __pyx_t_51) < 0) __PYX_ERR(0, 3372, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_51); __pyx_t_51 = 0;
  if (PyDict_SetItem(__pyx_t_49, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 3372, __pyx_L1_error)
  __pyx_t_51 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15ArgumentContext_1__init__, 0, __pyx_n_s_ModelicaParser_ArgumentContext_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1652)); if (unlikely(!__pyx_t_51)) __PYX_ERR(0, 3372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_51);
  __Pyx_INCREF(__pyx_t_51);
  PyList_Append(__pyx_t_50, __pyx_t_51);
  __Pyx_GIVEREF(__pyx_t_51);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_51, __pyx_tuple__1653);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_51, __pyx_t_49);
  __Pyx_DECREF(__pyx_t_49); __pyx_t_49 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_51) < 0) __PYX_ERR(0, 3372, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_51); __pyx_t_51 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3376
 *             self.parser = parser
 * 
 *         def element_modification_or_replaceable(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Element_modification_or_replaceableContext,0)
 * 
 */
  __pyx_t_51 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15ArgumentContext_3element_modification_or_replaceable, 0, __pyx_n_s_ModelicaParser_ArgumentContext_e, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1655)); if (unlikely(!__pyx_t_51)) __PYX_ERR(0, 3376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_51);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_element_modification_or_replacea, __pyx_t_51) < 0) __PYX_ERR(0, 3376, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_51); __pyx_t_51 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3380
 * 
 * 
 *         def element_redeclaration(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Element_redeclarationContext,0)
 * 
 */
  __pyx_t_51 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15ArgumentContext_5element_redeclaration, 0, __pyx_n_s_ModelicaParser_ArgumentContext_e_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1657)); if (unlikely(!__pyx_t_51)) __PYX_ERR(0, 3380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_51);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_element_redeclaration, __pyx_t_51) < 0) __PYX_ERR(0, 3380, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_51); __pyx_t_51 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3384
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_argument
 * 
 */
  __pyx_t_51 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15ArgumentContext_7getRuleIndex, 0, __pyx_n_s_ModelicaParser_ArgumentContext_g, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1659)); if (unlikely(!__pyx_t_51)) __PYX_ERR(0, 3384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_51);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_51) < 0) __PYX_ERR(0, 3384, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_51); __pyx_t_51 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3387
 *             return ModelicaParser.RULE_argument
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterArgument" ):
 *                 listener.enterArgument(self)
 */
  __pyx_t_51 = PyDict_New(); if (unlikely(!__pyx_t_51)) __PYX_ERR(0, 3387, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_51);
  __pyx_t_49 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_49)) __PYX_ERR(0, 3387, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_49);
  if (PyDict_SetItem(__pyx_t_51, __pyx_n_s_listener, __pyx_t_49) < 0) __PYX_ERR(0, 3387, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_49); __pyx_t_49 = 0;
  __pyx_t_49 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15ArgumentContext_9enterRule, 0, __pyx_n_s_ModelicaParser_ArgumentContext_e_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1661)); if (unlikely(!__pyx_t_49)) __PYX_ERR(0, 3387, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_49);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_49, __pyx_t_51);
  __Pyx_DECREF(__pyx_t_51); __pyx_t_51 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_49) < 0) __PYX_ERR(0, 3387, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_49); __pyx_t_49 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3391
 *                 listener.enterArgument(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitArgument" ):
 *                 listener.exitArgument(self)
 */
  __pyx_t_49 = PyDict_New(); if (unlikely(!__pyx_t_49)) __PYX_ERR(0, 3391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_49);
  __pyx_t_51 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_51)) __PYX_ERR(0, 3391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_51);
  if (PyDict_SetItem(__pyx_t_49, __pyx_n_s_listener, __pyx_t_51) < 0) __PYX_ERR(0, 3391, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_51); __pyx_t_51 = 0;
  __pyx_t_51 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15ArgumentContext_11exitRule, 0, __pyx_n_s_ModelicaParser_ArgumentContext_e_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1663)); if (unlikely(!__pyx_t_51)) __PYX_ERR(0, 3391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_51);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_51, __pyx_t_49);
  __Pyx_DECREF(__pyx_t_49); __pyx_t_49 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_51) < 0) __PYX_ERR(0, 3391, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_51); __pyx_t_51 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3395
 *                 listener.exitArgument(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitArgument" ):
 *                 return visitor.visitArgument(self)
 */
  __pyx_t_51 = PyDict_New(); if (unlikely(!__pyx_t_51)) __PYX_ERR(0, 3395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_51);
  __pyx_t_49 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_49)) __PYX_ERR(0, 3395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_49);
  if (PyDict_SetItem(__pyx_t_51, __pyx_n_s_visitor, __pyx_t_49) < 0) __PYX_ERR(0, 3395, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_49); __pyx_t_49 = 0;
  __pyx_t_49 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15ArgumentContext_13accept, 0, __pyx_n_s_ModelicaParser_ArgumentContext_a, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1665)); if (unlikely(!__pyx_t_49)) __PYX_ERR(0, 3395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_49);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_49, __pyx_t_51);
  __Pyx_DECREF(__pyx_t_51); __pyx_t_51 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_49) < 0) __PYX_ERR(0, 3395, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_49); __pyx_t_49 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3370
 *         return localctx
 * 
 *     class ArgumentContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_49 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_ArgumentContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_49)) __PYX_ERR(0, 3370, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_49);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_50, __pyx_t_49) < 0) __PYX_ERR(0, 3370, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_50); __pyx_t_50 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_ArgumentContext, __pyx_t_49) < 0) __PYX_ERR(0, 3370, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_49); __pyx_t_49 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3404
 * 
 * 
 *     def argument(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.ArgumentContext(self, self._ctx, self.state)
 */
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_63argument, 0, __pyx_n_s_ModelicaParser_argument, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1667)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3404, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_argument, __pyx_t_7) < 0) __PYX_ERR(0, 3404, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3433
 *         return localctx
 * 
 *     class Element_modification_or_replaceableContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Element_modification_or_replacea, __pyx_n_s_ModelicaParser_Element_modificat, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_49 = PyList_New(0); if (unlikely(!__pyx_t_49)) __PYX_ERR(0, 3433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_49);

  /* "pymola/generated/ModelicaParser.pyx":3435
 *     class Element_modification_or_replaceableContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_51 = PyDict_New(); if (unlikely(!__pyx_t_51)) __PYX_ERR(0, 3435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_51);
  __pyx_t_52 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_52)) __PYX_ERR(0, 3435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_52);
  if (PyDict_SetItem(__pyx_t_51, __pyx_n_s_parent, __pyx_t_52) < 0) __PYX_ERR(0, 3435, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_52); __pyx_t_52 = 0;
  if (PyDict_SetItem(__pyx_t_51, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 3435, __pyx_L1_error)
  __pyx_t_52 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_1__init__, 0, __pyx_n_s_ModelicaParser_Element_modificat_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1669)); if (unlikely(!__pyx_t_52)) __PYX_ERR(0, 3435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_52);
  __Pyx_INCREF(__pyx_t_52);
  PyList_Append(__pyx_t_49, __pyx_t_52);
  __Pyx_GIVEREF(__pyx_t_52);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_52, __pyx_tuple__1670);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_52, __pyx_t_51);
  __Pyx_DECREF(__pyx_t_51); __pyx_t_51 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_52) < 0) __PYX_ERR(0, 3435, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_52); __pyx_t_52 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3439
 *             self.parser = parser
 * 
 *         def element_modification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Element_modificationContext,0)
 * 
 */
  __pyx_t_52 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_3element_modification, 0, __pyx_n_s_ModelicaParser_Element_modificat_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1672)); if (unlikely(!__pyx_t_52)) __PYX_ERR(0, 3439, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_52);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_element_modification, __pyx_t_52) < 0) __PYX_ERR(0, 3439, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_52); __pyx_t_52 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3443
 * 
 * 
 *         def element_replaceable(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Element_replaceableContext,0)
 * 
 */
  __pyx_t_52 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_5element_replaceable, 0, __pyx_n_s_ModelicaParser_Element_modificat_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1674)); if (unlikely(!__pyx_t_52)) __PYX_ERR(0, 3443, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_52);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_element_replaceable, __pyx_t_52) < 0) __PYX_ERR(0, 3443, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_52); __pyx_t_52 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3447
 * 
 * 
 *         def EACH(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.EACH, 0)
 * 
 */
  __pyx_t_52 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_7EACH, 0, __pyx_n_s_ModelicaParser_Element_modificat_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1676)); if (unlikely(!__pyx_t_52)) __PYX_ERR(0, 3447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_52);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_EACH, __pyx_t_52) < 0) __PYX_ERR(0, 3447, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_52); __pyx_t_52 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3450
 *             return self.getToken(ModelicaParser.EACH, 0)
 * 
 *         def FINAL(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.FINAL, 0)
 * 
 */
  __pyx_t_52 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_9FINAL, 0, __pyx_n_s_ModelicaParser_Element_modificat_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1678)); if (unlikely(!__pyx_t_52)) __PYX_ERR(0, 3450, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_52);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_FINAL, __pyx_t_52) < 0) __PYX_ERR(0, 3450, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_52); __pyx_t_52 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3453
 *             return self.getToken(ModelicaParser.FINAL, 0)
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_element_modification_or_replaceable
 * 
 */
  __pyx_t_52 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_11getRuleIndex, 0, __pyx_n_s_ModelicaParser_Element_modificat_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1680)); if (unlikely(!__pyx_t_52)) __PYX_ERR(0, 3453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_52);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_52) < 0) __PYX_ERR(0, 3453, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_52); __pyx_t_52 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3456
 *             return ModelicaParser.RULE_element_modification_or_replaceable
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterElement_modification_or_replaceable" ):
 *                 listener.enterElement_modification_or_replaceable(self)
 */
  __pyx_t_52 = PyDict_New(); if (unlikely(!__pyx_t_52)) __PYX_ERR(0, 3456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_52);
  __pyx_t_51 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_51)) __PYX_ERR(0, 3456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_51);
  if (PyDict_SetItem(__pyx_t_52, __pyx_n_s_listener, __pyx_t_51) < 0) __PYX_ERR(0, 3456, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_51); __pyx_t_51 = 0;
  __pyx_t_51 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_13enterRule, 0, __pyx_n_s_ModelicaParser_Element_modificat_8, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1682)); if (unlikely(!__pyx_t_51)) __PYX_ERR(0, 3456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_51);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_51, __pyx_t_52);
  __Pyx_DECREF(__pyx_t_52); __pyx_t_52 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_51) < 0) __PYX_ERR(0, 3456, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_51); __pyx_t_51 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3460
 *                 listener.enterElement_modification_or_replaceable(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitElement_modification_or_replaceable" ):
 *                 listener.exitElement_modification_or_replaceable(self)
 */
  __pyx_t_51 = PyDict_New(); if (unlikely(!__pyx_t_51)) __PYX_ERR(0, 3460, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_51);
  __pyx_t_52 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_52)) __PYX_ERR(0, 3460, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_52);
  if (PyDict_SetItem(__pyx_t_51, __pyx_n_s_listener, __pyx_t_52) < 0) __PYX_ERR(0, 3460, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_52); __pyx_t_52 = 0;
  __pyx_t_52 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_15exitRule, 0, __pyx_n_s_ModelicaParser_Element_modificat_9, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1684)); if (unlikely(!__pyx_t_52)) __PYX_ERR(0, 3460, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_52);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_52, __pyx_t_51);
  __Pyx_DECREF(__pyx_t_51); __pyx_t_51 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_52) < 0) __PYX_ERR(0, 3460, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_52); __pyx_t_52 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3464
 *                 listener.exitElement_modification_or_replaceable(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitElement_modification_or_replaceable" ):
 *                 return visitor.visitElement_modification_or_replaceable(self)
 */
  __pyx_t_52 = PyDict_New(); if (unlikely(!__pyx_t_52)) __PYX_ERR(0, 3464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_52);
  __pyx_t_51 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_51)) __PYX_ERR(0, 3464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_51);
  if (PyDict_SetItem(__pyx_t_52, __pyx_n_s_visitor, __pyx_t_51) < 0) __PYX_ERR(0, 3464, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_51); __pyx_t_51 = 0;
  __pyx_t_51 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_42Element_modification_or_replaceableContext_17accept, 0, __pyx_n_s_ModelicaParser_Element_modificat_10, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1686)); if (unlikely(!__pyx_t_51)) __PYX_ERR(0, 3464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_51);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_51, __pyx_t_52);
  __Pyx_DECREF(__pyx_t_52); __pyx_t_52 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_51) < 0) __PYX_ERR(0, 3464, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_51); __pyx_t_51 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3433
 *         return localctx
 * 
 *     class Element_modification_or_replaceableContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_51 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Element_modification_or_replacea, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_51)) __PYX_ERR(0, 3433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_51);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_49, __pyx_t_51) < 0) __PYX_ERR(0, 3433, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_49); __pyx_t_49 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Element_modification_or_replacea, __pyx_t_51) < 0) __PYX_ERR(0, 3433, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_51); __pyx_t_51 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3473
 * 
 * 
 *     def element_modification_or_replaceable(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Element_modification_or_replaceableContext(self, self._ctx, self.state)
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_65element_modification_or_replaceable, 0, __pyx_n_s_ModelicaParser_element_modificat, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1688)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_element_modification_or_replacea, __pyx_t_5) < 0) __PYX_ERR(0, 3473, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3518
 *         return localctx
 * 
 *     class Element_modificationContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Element_modificationContext, __pyx_n_s_ModelicaParser_Element_modificat_11, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_51 = PyList_New(0); if (unlikely(!__pyx_t_51)) __PYX_ERR(0, 3518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_51);

  /* "pymola/generated/ModelicaParser.pyx":3520
 *     class Element_modificationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_52 = PyDict_New(); if (unlikely(!__pyx_t_52)) __PYX_ERR(0, 3520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_52);
  __pyx_t_53 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_53)) __PYX_ERR(0, 3520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_53);
  if (PyDict_SetItem(__pyx_t_52, __pyx_n_s_parent, __pyx_t_53) < 0) __PYX_ERR(0, 3520, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_53); __pyx_t_53 = 0;
  if (PyDict_SetItem(__pyx_t_52, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 3520, __pyx_L1_error)
  __pyx_t_53 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_27Element_modificationContext_1__init__, 0, __pyx_n_s_ModelicaParser_Element_modificat_12, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1690)); if (unlikely(!__pyx_t_53)) __PYX_ERR(0, 3520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_53);
  __Pyx_INCREF(__pyx_t_53);
  PyList_Append(__pyx_t_51, __pyx_t_53);
  __Pyx_GIVEREF(__pyx_t_53);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_53, __pyx_tuple__1691);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_53, __pyx_t_52);
  __Pyx_DECREF(__pyx_t_52); __pyx_t_52 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_53) < 0) __PYX_ERR(0, 3520, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_53); __pyx_t_53 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3524
 *             self.parser = parser
 * 
 *         def component_reference(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 */
  __pyx_t_53 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_27Element_modificationContext_3component_reference, 0, __pyx_n_s_ModelicaParser_Element_modificat_13, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1693)); if (unlikely(!__pyx_t_53)) __PYX_ERR(0, 3524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_53);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_component_reference, __pyx_t_53) < 0) __PYX_ERR(0, 3524, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_53); __pyx_t_53 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3528
 * 
 * 
 *         def string_comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.String_commentContext,0)
 * 
 */
  __pyx_t_53 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_27Element_modificationContext_5string_comment, 0, __pyx_n_s_ModelicaParser_Element_modificat_14, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1695)); if (unlikely(!__pyx_t_53)) __PYX_ERR(0, 3528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_53);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_string_comment, __pyx_t_53) < 0) __PYX_ERR(0, 3528, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_53); __pyx_t_53 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3532
 * 
 * 
 *         def modification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ModificationContext,0)
 * 
 */
  __pyx_t_53 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_27Element_modificationContext_7modification, 0, __pyx_n_s_ModelicaParser_Element_modificat_15, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1697)); if (unlikely(!__pyx_t_53)) __PYX_ERR(0, 3532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_53);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_modification, __pyx_t_53) < 0) __PYX_ERR(0, 3532, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_53); __pyx_t_53 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3536
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_element_modification
 * 
 */
  __pyx_t_53 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_27Element_modificationContext_9getRuleIndex, 0, __pyx_n_s_ModelicaParser_Element_modificat_16, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1699)); if (unlikely(!__pyx_t_53)) __PYX_ERR(0, 3536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_53);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_53) < 0) __PYX_ERR(0, 3536, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_53); __pyx_t_53 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3539
 *             return ModelicaParser.RULE_element_modification
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterElement_modification" ):
 *                 listener.enterElement_modification(self)
 */
  __pyx_t_53 = PyDict_New(); if (unlikely(!__pyx_t_53)) __PYX_ERR(0, 3539, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_53);
  __pyx_t_52 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_52)) __PYX_ERR(0, 3539, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_52);
  if (PyDict_SetItem(__pyx_t_53, __pyx_n_s_listener, __pyx_t_52) < 0) __PYX_ERR(0, 3539, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_52); __pyx_t_52 = 0;
  __pyx_t_52 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_27Element_modificationContext_11enterRule, 0, __pyx_n_s_ModelicaParser_Element_modificat_17, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1701)); if (unlikely(!__pyx_t_52)) __PYX_ERR(0, 3539, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_52);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_52, __pyx_t_53);
  __Pyx_DECREF(__pyx_t_53); __pyx_t_53 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_52) < 0) __PYX_ERR(0, 3539, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_52); __pyx_t_52 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3543
 *                 listener.enterElement_modification(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitElement_modification" ):
 *                 listener.exitElement_modification(self)
 */
  __pyx_t_52 = PyDict_New(); if (unlikely(!__pyx_t_52)) __PYX_ERR(0, 3543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_52);
  __pyx_t_53 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_53)) __PYX_ERR(0, 3543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_53);
  if (PyDict_SetItem(__pyx_t_52, __pyx_n_s_listener, __pyx_t_53) < 0) __PYX_ERR(0, 3543, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_53); __pyx_t_53 = 0;
  __pyx_t_53 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_27Element_modificationContext_13exitRule, 0, __pyx_n_s_ModelicaParser_Element_modificat_18, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1703)); if (unlikely(!__pyx_t_53)) __PYX_ERR(0, 3543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_53);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_53, __pyx_t_52);
  __Pyx_DECREF(__pyx_t_52); __pyx_t_52 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_53) < 0) __PYX_ERR(0, 3543, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_53); __pyx_t_53 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3547
 *                 listener.exitElement_modification(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitElement_modification" ):
 *                 return visitor.visitElement_modification(self)
 */
  __pyx_t_53 = PyDict_New(); if (unlikely(!__pyx_t_53)) __PYX_ERR(0, 3547, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_53);
  __pyx_t_52 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_52)) __PYX_ERR(0, 3547, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_52);
  if (PyDict_SetItem(__pyx_t_53, __pyx_n_s_visitor, __pyx_t_52) < 0) __PYX_ERR(0, 3547, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_52); __pyx_t_52 = 0;
  __pyx_t_52 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_27Element_modificationContext_15accept, 0, __pyx_n_s_ModelicaParser_Element_modificat_19, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1705)); if (unlikely(!__pyx_t_52)) __PYX_ERR(0, 3547, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_52);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_52, __pyx_t_53);
  __Pyx_DECREF(__pyx_t_53); __pyx_t_53 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_52) < 0) __PYX_ERR(0, 3547, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_52); __pyx_t_52 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3518
 *         return localctx
 * 
 *     class Element_modificationContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_52 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Element_modificationContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_52)) __PYX_ERR(0, 3518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_52);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_51, __pyx_t_52) < 0) __PYX_ERR(0, 3518, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_51); __pyx_t_51 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Element_modificationContext, __pyx_t_52) < 0) __PYX_ERR(0, 3518, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_52); __pyx_t_52 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3556
 * 
 * 
 *     def element_modification(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Element_modificationContext(self, self._ctx, self.state)
 */
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_67element_modification, 0, __pyx_n_s_ModelicaParser_element_modificat_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1707)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_element_modification, __pyx_t_7) < 0) __PYX_ERR(0, 3556, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3583
 *         return localctx
 * 
 *     class Element_redeclarationContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Element_redeclarationContext, __pyx_n_s_ModelicaParser_Element_redeclara, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_52 = PyList_New(0); if (unlikely(!__pyx_t_52)) __PYX_ERR(0, 3583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_52);

  /* "pymola/generated/ModelicaParser.pyx":3585
 *     class Element_redeclarationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_53 = PyDict_New(); if (unlikely(!__pyx_t_53)) __PYX_ERR(0, 3585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_53);
  __pyx_t_54 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_54)) __PYX_ERR(0, 3585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_54);
  if (PyDict_SetItem(__pyx_t_53, __pyx_n_s_parent, __pyx_t_54) < 0) __PYX_ERR(0, 3585, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_54); __pyx_t_54 = 0;
  if (PyDict_SetItem(__pyx_t_53, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 3585, __pyx_L1_error)
  __pyx_t_54 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_1__init__, 0, __pyx_n_s_ModelicaParser_Element_redeclara_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1709)); if (unlikely(!__pyx_t_54)) __PYX_ERR(0, 3585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_54);
  __Pyx_INCREF(__pyx_t_54);
  PyList_Append(__pyx_t_52, __pyx_t_54);
  __Pyx_GIVEREF(__pyx_t_54);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_54, __pyx_tuple__1710);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_54, __pyx_t_53);
  __Pyx_DECREF(__pyx_t_53); __pyx_t_53 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_54) < 0) __PYX_ERR(0, 3585, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_54); __pyx_t_54 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3589
 *             self.parser = parser
 * 
 *         def REDECLARE(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.REDECLARE, 0)
 * 
 */
  __pyx_t_54 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_3REDECLARE, 0, __pyx_n_s_ModelicaParser_Element_redeclara_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1712)); if (unlikely(!__pyx_t_54)) __PYX_ERR(0, 3589, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_54);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_REDECLARE, __pyx_t_54) < 0) __PYX_ERR(0, 3589, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_54); __pyx_t_54 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3592
 *             return self.getToken(ModelicaParser.REDECLARE, 0)
 * 
 *         def element_replaceable(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Element_replaceableContext,0)
 * 
 */
  __pyx_t_54 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_5element_replaceable, 0, __pyx_n_s_ModelicaParser_Element_redeclara_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1714)); if (unlikely(!__pyx_t_54)) __PYX_ERR(0, 3592, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_54);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_element_replaceable, __pyx_t_54) < 0) __PYX_ERR(0, 3592, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_54); __pyx_t_54 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3596
 * 
 * 
 *         def EACH(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.EACH, 0)
 * 
 */
  __pyx_t_54 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_7EACH, 0, __pyx_n_s_ModelicaParser_Element_redeclara_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1716)); if (unlikely(!__pyx_t_54)) __PYX_ERR(0, 3596, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_54);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_EACH, __pyx_t_54) < 0) __PYX_ERR(0, 3596, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_54); __pyx_t_54 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3599
 *             return self.getToken(ModelicaParser.EACH, 0)
 * 
 *         def FINAL(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.FINAL, 0)
 * 
 */
  __pyx_t_54 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_9FINAL, 0, __pyx_n_s_ModelicaParser_Element_redeclara_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1718)); if (unlikely(!__pyx_t_54)) __PYX_ERR(0, 3599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_54);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_FINAL, __pyx_t_54) < 0) __PYX_ERR(0, 3599, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_54); __pyx_t_54 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3602
 *             return self.getToken(ModelicaParser.FINAL, 0)
 * 
 *         def short_class_definition(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Short_class_definitionContext,0)
 * 
 */
  __pyx_t_54 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_11short_class_definition, 0, __pyx_n_s_ModelicaParser_Element_redeclara_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1720)); if (unlikely(!__pyx_t_54)) __PYX_ERR(0, 3602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_54);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_short_class_definition, __pyx_t_54) < 0) __PYX_ERR(0, 3602, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_54); __pyx_t_54 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3606
 * 
 * 
 *         def component_clause1(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_clause1Context,0)
 * 
 */
  __pyx_t_54 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_13component_clause1, 0, __pyx_n_s_ModelicaParser_Element_redeclara_8, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1722)); if (unlikely(!__pyx_t_54)) __PYX_ERR(0, 3606, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_54);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_component_clause1, __pyx_t_54) < 0) __PYX_ERR(0, 3606, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_54); __pyx_t_54 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3610
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_element_redeclaration
 * 
 */
  __pyx_t_54 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_15getRuleIndex, 0, __pyx_n_s_ModelicaParser_Element_redeclara_9, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1724)); if (unlikely(!__pyx_t_54)) __PYX_ERR(0, 3610, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_54);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_54) < 0) __PYX_ERR(0, 3610, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_54); __pyx_t_54 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3613
 *             return ModelicaParser.RULE_element_redeclaration
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterElement_redeclaration" ):
 *                 listener.enterElement_redeclaration(self)
 */
  __pyx_t_54 = PyDict_New(); if (unlikely(!__pyx_t_54)) __PYX_ERR(0, 3613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_54);
  __pyx_t_53 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_53)) __PYX_ERR(0, 3613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_53);
  if (PyDict_SetItem(__pyx_t_54, __pyx_n_s_listener, __pyx_t_53) < 0) __PYX_ERR(0, 3613, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_53); __pyx_t_53 = 0;
  __pyx_t_53 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_17enterRule, 0, __pyx_n_s_ModelicaParser_Element_redeclara_10, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1726)); if (unlikely(!__pyx_t_53)) __PYX_ERR(0, 3613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_53);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_53, __pyx_t_54);
  __Pyx_DECREF(__pyx_t_54); __pyx_t_54 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_53) < 0) __PYX_ERR(0, 3613, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_53); __pyx_t_53 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3617
 *                 listener.enterElement_redeclaration(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitElement_redeclaration" ):
 *                 listener.exitElement_redeclaration(self)
 */
  __pyx_t_53 = PyDict_New(); if (unlikely(!__pyx_t_53)) __PYX_ERR(0, 3617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_53);
  __pyx_t_54 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_54)) __PYX_ERR(0, 3617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_54);
  if (PyDict_SetItem(__pyx_t_53, __pyx_n_s_listener, __pyx_t_54) < 0) __PYX_ERR(0, 3617, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_54); __pyx_t_54 = 0;
  __pyx_t_54 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_19exitRule, 0, __pyx_n_s_ModelicaParser_Element_redeclara_11, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1728)); if (unlikely(!__pyx_t_54)) __PYX_ERR(0, 3617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_54);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_54, __pyx_t_53);
  __Pyx_DECREF(__pyx_t_53); __pyx_t_53 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_54) < 0) __PYX_ERR(0, 3617, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_54); __pyx_t_54 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3621
 *                 listener.exitElement_redeclaration(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitElement_redeclaration" ):
 *                 return visitor.visitElement_redeclaration(self)
 */
  __pyx_t_54 = PyDict_New(); if (unlikely(!__pyx_t_54)) __PYX_ERR(0, 3621, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_54);
  __pyx_t_53 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_53)) __PYX_ERR(0, 3621, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_53);
  if (PyDict_SetItem(__pyx_t_54, __pyx_n_s_visitor, __pyx_t_53) < 0) __PYX_ERR(0, 3621, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_53); __pyx_t_53 = 0;
  __pyx_t_53 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_28Element_redeclarationContext_21accept, 0, __pyx_n_s_ModelicaParser_Element_redeclara_12, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1730)); if (unlikely(!__pyx_t_53)) __PYX_ERR(0, 3621, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_53);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_53, __pyx_t_54);
  __Pyx_DECREF(__pyx_t_54); __pyx_t_54 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_53) < 0) __PYX_ERR(0, 3621, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_53); __pyx_t_53 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3583
 *         return localctx
 * 
 *     class Element_redeclarationContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_53 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Element_redeclarationContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_53)) __PYX_ERR(0, 3583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_53);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_52, __pyx_t_53) < 0) __PYX_ERR(0, 3583, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_52); __pyx_t_52 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Element_redeclarationContext, __pyx_t_53) < 0) __PYX_ERR(0, 3583, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_53); __pyx_t_53 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3630
 * 
 * 
 *     def element_redeclaration(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Element_redeclarationContext(self, self._ctx, self.state)
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_69element_redeclaration, 0, __pyx_n_s_ModelicaParser_element_redeclara, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1732)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_element_redeclaration, __pyx_t_5) < 0) __PYX_ERR(0, 3630, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3689
 *         return localctx
 * 
 *     class Element_replaceableContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3689, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3689, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3689, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Element_replaceableContext, __pyx_n_s_ModelicaParser_Element_replaceab, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3689, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_53 = PyList_New(0); if (unlikely(!__pyx_t_53)) __PYX_ERR(0, 3689, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_53);

  /* "pymola/generated/ModelicaParser.pyx":3691
 *     class Element_replaceableContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_54 = PyDict_New(); if (unlikely(!__pyx_t_54)) __PYX_ERR(0, 3691, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_54);
  __pyx_t_55 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_55)) __PYX_ERR(0, 3691, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_55);
  if (PyDict_SetItem(__pyx_t_54, __pyx_n_s_parent, __pyx_t_55) < 0) __PYX_ERR(0, 3691, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_55); __pyx_t_55 = 0;
  if (PyDict_SetItem(__pyx_t_54, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 3691, __pyx_L1_error)
  __pyx_t_55 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_1__init__, 0, __pyx_n_s_ModelicaParser_Element_replaceab_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1734)); if (unlikely(!__pyx_t_55)) __PYX_ERR(0, 3691, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_55);
  __Pyx_INCREF(__pyx_t_55);
  PyList_Append(__pyx_t_53, __pyx_t_55);
  __Pyx_GIVEREF(__pyx_t_55);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_55, __pyx_tuple__1735);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_55, __pyx_t_54);
  __Pyx_DECREF(__pyx_t_54); __pyx_t_54 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_55) < 0) __PYX_ERR(0, 3691, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_55); __pyx_t_55 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3695
 *             self.parser = parser
 * 
 *         def short_class_definition(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Short_class_definitionContext,0)
 * 
 */
  __pyx_t_55 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_3short_class_definition, 0, __pyx_n_s_ModelicaParser_Element_replaceab_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1737)); if (unlikely(!__pyx_t_55)) __PYX_ERR(0, 3695, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_55);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_short_class_definition, __pyx_t_55) < 0) __PYX_ERR(0, 3695, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_55); __pyx_t_55 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3699
 * 
 * 
 *         def component_clause1(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_clause1Context,0)
 * 
 */
  __pyx_t_55 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_5component_clause1, 0, __pyx_n_s_ModelicaParser_Element_replaceab_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1739)); if (unlikely(!__pyx_t_55)) __PYX_ERR(0, 3699, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_55);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_component_clause1, __pyx_t_55) < 0) __PYX_ERR(0, 3699, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_55); __pyx_t_55 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3703
 * 
 * 
 *         def constraining_clause(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Constraining_clauseContext,0)
 * 
 */
  __pyx_t_55 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_7constraining_clause, 0, __pyx_n_s_ModelicaParser_Element_replaceab_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1741)); if (unlikely(!__pyx_t_55)) __PYX_ERR(0, 3703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_55);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_constraining_clause, __pyx_t_55) < 0) __PYX_ERR(0, 3703, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_55); __pyx_t_55 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3707
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_element_replaceable
 * 
 */
  __pyx_t_55 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_9getRuleIndex, 0, __pyx_n_s_ModelicaParser_Element_replaceab_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1743)); if (unlikely(!__pyx_t_55)) __PYX_ERR(0, 3707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_55);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_55) < 0) __PYX_ERR(0, 3707, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_55); __pyx_t_55 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3710
 *             return ModelicaParser.RULE_element_replaceable
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterElement_replaceable" ):
 *                 listener.enterElement_replaceable(self)
 */
  __pyx_t_55 = PyDict_New(); if (unlikely(!__pyx_t_55)) __PYX_ERR(0, 3710, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_55);
  __pyx_t_54 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_54)) __PYX_ERR(0, 3710, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_54);
  if (PyDict_SetItem(__pyx_t_55, __pyx_n_s_listener, __pyx_t_54) < 0) __PYX_ERR(0, 3710, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_54); __pyx_t_54 = 0;
  __pyx_t_54 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_11enterRule, 0, __pyx_n_s_ModelicaParser_Element_replaceab_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1745)); if (unlikely(!__pyx_t_54)) __PYX_ERR(0, 3710, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_54);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_54, __pyx_t_55);
  __Pyx_DECREF(__pyx_t_55); __pyx_t_55 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_54) < 0) __PYX_ERR(0, 3710, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_54); __pyx_t_54 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3714
 *                 listener.enterElement_replaceable(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitElement_replaceable" ):
 *                 listener.exitElement_replaceable(self)
 */
  __pyx_t_54 = PyDict_New(); if (unlikely(!__pyx_t_54)) __PYX_ERR(0, 3714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_54);
  __pyx_t_55 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_55)) __PYX_ERR(0, 3714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_55);
  if (PyDict_SetItem(__pyx_t_54, __pyx_n_s_listener, __pyx_t_55) < 0) __PYX_ERR(0, 3714, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_55); __pyx_t_55 = 0;
  __pyx_t_55 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_13exitRule, 0, __pyx_n_s_ModelicaParser_Element_replaceab_8, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1747)); if (unlikely(!__pyx_t_55)) __PYX_ERR(0, 3714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_55);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_55, __pyx_t_54);
  __Pyx_DECREF(__pyx_t_54); __pyx_t_54 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_55) < 0) __PYX_ERR(0, 3714, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_55); __pyx_t_55 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3718
 *                 listener.exitElement_replaceable(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitElement_replaceable" ):
 *                 return visitor.visitElement_replaceable(self)
 */
  __pyx_t_55 = PyDict_New(); if (unlikely(!__pyx_t_55)) __PYX_ERR(0, 3718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_55);
  __pyx_t_54 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_54)) __PYX_ERR(0, 3718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_54);
  if (PyDict_SetItem(__pyx_t_55, __pyx_n_s_visitor, __pyx_t_54) < 0) __PYX_ERR(0, 3718, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_54); __pyx_t_54 = 0;
  __pyx_t_54 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Element_replaceableContext_15accept, 0, __pyx_n_s_ModelicaParser_Element_replaceab_9, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1749)); if (unlikely(!__pyx_t_54)) __PYX_ERR(0, 3718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_54);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_54, __pyx_t_55);
  __Pyx_DECREF(__pyx_t_55); __pyx_t_55 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_54) < 0) __PYX_ERR(0, 3718, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_54); __pyx_t_54 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3689
 *         return localctx
 * 
 *     class Element_replaceableContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_54 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Element_replaceableContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_54)) __PYX_ERR(0, 3689, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_54);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_53, __pyx_t_54) < 0) __PYX_ERR(0, 3689, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_53); __pyx_t_53 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Element_replaceableContext, __pyx_t_54) < 0) __PYX_ERR(0, 3689, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_54); __pyx_t_54 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3727
 * 
 * 
 *     def element_replaceable(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Element_replaceableContext(self, self._ctx, self.state)
 */
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_71element_replaceable, 0, __pyx_n_s_ModelicaParser_element_replaceab, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1751)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3727, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_element_replaceable, __pyx_t_7) < 0) __PYX_ERR(0, 3727, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3766
 *         return localctx
 * 
 *     class Component_clause1Context(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Component_clause1Context, __pyx_n_s_ModelicaParser_Component_clause1, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_54 = PyList_New(0); if (unlikely(!__pyx_t_54)) __PYX_ERR(0, 3766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_54);

  /* "pymola/generated/ModelicaParser.pyx":3768
 *     class Component_clause1Context(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_55 = PyDict_New(); if (unlikely(!__pyx_t_55)) __PYX_ERR(0, 3768, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_55);
  __pyx_t_56 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_56)) __PYX_ERR(0, 3768, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_56);
  if (PyDict_SetItem(__pyx_t_55, __pyx_n_s_parent, __pyx_t_56) < 0) __PYX_ERR(0, 3768, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_56); __pyx_t_56 = 0;
  if (PyDict_SetItem(__pyx_t_55, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 3768, __pyx_L1_error)
  __pyx_t_56 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Component_clause1Context_1__init__, 0, __pyx_n_s_ModelicaParser_Component_clause1_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1753)); if (unlikely(!__pyx_t_56)) __PYX_ERR(0, 3768, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_56);
  __Pyx_INCREF(__pyx_t_56);
  PyList_Append(__pyx_t_54, __pyx_t_56);
  __Pyx_GIVEREF(__pyx_t_56);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_56, __pyx_tuple__1754);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_56, __pyx_t_55);
  __Pyx_DECREF(__pyx_t_55); __pyx_t_55 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_56) < 0) __PYX_ERR(0, 3768, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_56); __pyx_t_56 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3772
 *             self.parser = parser
 * 
 *         def type_prefix(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Type_prefixContext,0)
 * 
 */
  __pyx_t_56 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Component_clause1Context_3type_prefix, 0, __pyx_n_s_ModelicaParser_Component_clause1_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1756)); if (unlikely(!__pyx_t_56)) __PYX_ERR(0, 3772, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_56);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_type_prefix, __pyx_t_56) < 0) __PYX_ERR(0, 3772, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_56); __pyx_t_56 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3776
 * 
 * 
 *         def type_specifier(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Type_specifierContext,0)
 * 
 */
  __pyx_t_56 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Component_clause1Context_5type_specifier, 0, __pyx_n_s_ModelicaParser_Component_clause1_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1758)); if (unlikely(!__pyx_t_56)) __PYX_ERR(0, 3776, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_56);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_type_specifier, __pyx_t_56) < 0) __PYX_ERR(0, 3776, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_56); __pyx_t_56 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3780
 * 
 * 
 *         def component_declaration1(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_declaration1Context,0)
 * 
 */
  __pyx_t_56 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Component_clause1Context_7component_declaration1, 0, __pyx_n_s_ModelicaParser_Component_clause1_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1760)); if (unlikely(!__pyx_t_56)) __PYX_ERR(0, 3780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_56);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_component_declaration1, __pyx_t_56) < 0) __PYX_ERR(0, 3780, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_56); __pyx_t_56 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3784
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_component_clause1
 * 
 */
  __pyx_t_56 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Component_clause1Context_9getRuleIndex, 0, __pyx_n_s_ModelicaParser_Component_clause1_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1762)); if (unlikely(!__pyx_t_56)) __PYX_ERR(0, 3784, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_56);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_56) < 0) __PYX_ERR(0, 3784, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_56); __pyx_t_56 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3787
 *             return ModelicaParser.RULE_component_clause1
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterComponent_clause1" ):
 *                 listener.enterComponent_clause1(self)
 */
  __pyx_t_56 = PyDict_New(); if (unlikely(!__pyx_t_56)) __PYX_ERR(0, 3787, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_56);
  __pyx_t_55 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_55)) __PYX_ERR(0, 3787, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_55);
  if (PyDict_SetItem(__pyx_t_56, __pyx_n_s_listener, __pyx_t_55) < 0) __PYX_ERR(0, 3787, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_55); __pyx_t_55 = 0;
  __pyx_t_55 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Component_clause1Context_11enterRule, 0, __pyx_n_s_ModelicaParser_Component_clause1_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1764)); if (unlikely(!__pyx_t_55)) __PYX_ERR(0, 3787, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_55);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_55, __pyx_t_56);
  __Pyx_DECREF(__pyx_t_56); __pyx_t_56 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_55) < 0) __PYX_ERR(0, 3787, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_55); __pyx_t_55 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3791
 *                 listener.enterComponent_clause1(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitComponent_clause1" ):
 *                 listener.exitComponent_clause1(self)
 */
  __pyx_t_55 = PyDict_New(); if (unlikely(!__pyx_t_55)) __PYX_ERR(0, 3791, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_55);
  __pyx_t_56 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_56)) __PYX_ERR(0, 3791, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_56);
  if (PyDict_SetItem(__pyx_t_55, __pyx_n_s_listener, __pyx_t_56) < 0) __PYX_ERR(0, 3791, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_56); __pyx_t_56 = 0;
  __pyx_t_56 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Component_clause1Context_13exitRule, 0, __pyx_n_s_ModelicaParser_Component_clause1_8, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1766)); if (unlikely(!__pyx_t_56)) __PYX_ERR(0, 3791, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_56);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_56, __pyx_t_55);
  __Pyx_DECREF(__pyx_t_55); __pyx_t_55 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_56) < 0) __PYX_ERR(0, 3791, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_56); __pyx_t_56 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3795
 *                 listener.exitComponent_clause1(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitComponent_clause1" ):
 *                 return visitor.visitComponent_clause1(self)
 */
  __pyx_t_56 = PyDict_New(); if (unlikely(!__pyx_t_56)) __PYX_ERR(0, 3795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_56);
  __pyx_t_55 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_55)) __PYX_ERR(0, 3795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_55);
  if (PyDict_SetItem(__pyx_t_56, __pyx_n_s_visitor, __pyx_t_55) < 0) __PYX_ERR(0, 3795, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_55); __pyx_t_55 = 0;
  __pyx_t_55 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Component_clause1Context_15accept, 0, __pyx_n_s_ModelicaParser_Component_clause1_9, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1768)); if (unlikely(!__pyx_t_55)) __PYX_ERR(0, 3795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_55);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_55, __pyx_t_56);
  __Pyx_DECREF(__pyx_t_56); __pyx_t_56 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_55) < 0) __PYX_ERR(0, 3795, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_55); __pyx_t_55 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3766
 *         return localctx
 * 
 *     class Component_clause1Context(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_55 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Component_clause1Context, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_55)) __PYX_ERR(0, 3766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_55);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_54, __pyx_t_55) < 0) __PYX_ERR(0, 3766, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_54); __pyx_t_54 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Component_clause1Context, __pyx_t_55) < 0) __PYX_ERR(0, 3766, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_55); __pyx_t_55 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3804
 * 
 * 
 *     def component_clause1(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Component_clause1Context(self, self._ctx, self.state)
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_73component_clause1, 0, __pyx_n_s_ModelicaParser_component_clause1, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1770)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3804, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_component_clause1, __pyx_t_5) < 0) __PYX_ERR(0, 3804, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3824
 *         return localctx
 * 
 *     class Component_declaration1Context(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3824, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3824, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3824, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Component_declaration1Context, __pyx_n_s_ModelicaParser_Component_declara_10, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3824, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_55 = PyList_New(0); if (unlikely(!__pyx_t_55)) __PYX_ERR(0, 3824, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_55);

  /* "pymola/generated/ModelicaParser.pyx":3826
 *     class Component_declaration1Context(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_56 = PyDict_New(); if (unlikely(!__pyx_t_56)) __PYX_ERR(0, 3826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_56);
  __pyx_t_57 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_57)) __PYX_ERR(0, 3826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_57);
  if (PyDict_SetItem(__pyx_t_56, __pyx_n_s_parent, __pyx_t_57) < 0) __PYX_ERR(0, 3826, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_57); __pyx_t_57 = 0;
  if (PyDict_SetItem(__pyx_t_56, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 3826, __pyx_L1_error)
  __pyx_t_57 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_1__init__, 0, __pyx_n_s_ModelicaParser_Component_declara_11, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1772)); if (unlikely(!__pyx_t_57)) __PYX_ERR(0, 3826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_57);
  __Pyx_INCREF(__pyx_t_57);
  PyList_Append(__pyx_t_55, __pyx_t_57);
  __Pyx_GIVEREF(__pyx_t_57);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_57, __pyx_tuple__1773);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_57, __pyx_t_56);
  __Pyx_DECREF(__pyx_t_56); __pyx_t_56 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_57) < 0) __PYX_ERR(0, 3826, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_57); __pyx_t_57 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3830
 *             self.parser = parser
 * 
 *         def declaration(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.DeclarationContext,0)
 * 
 */
  __pyx_t_57 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_3declaration, 0, __pyx_n_s_ModelicaParser_Component_declara_12, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1775)); if (unlikely(!__pyx_t_57)) __PYX_ERR(0, 3830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_57);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_declaration, __pyx_t_57) < 0) __PYX_ERR(0, 3830, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_57); __pyx_t_57 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3834
 * 
 * 
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */
  __pyx_t_57 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_5comment, 0, __pyx_n_s_ModelicaParser_Component_declara_13, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1777)); if (unlikely(!__pyx_t_57)) __PYX_ERR(0, 3834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_57);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_comment, __pyx_t_57) < 0) __PYX_ERR(0, 3834, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_57); __pyx_t_57 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3838
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_component_declaration1
 * 
 */
  __pyx_t_57 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_7getRuleIndex, 0, __pyx_n_s_ModelicaParser_Component_declara_14, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1779)); if (unlikely(!__pyx_t_57)) __PYX_ERR(0, 3838, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_57);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_57) < 0) __PYX_ERR(0, 3838, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_57); __pyx_t_57 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3841
 *             return ModelicaParser.RULE_component_declaration1
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterComponent_declaration1" ):
 *                 listener.enterComponent_declaration1(self)
 */
  __pyx_t_57 = PyDict_New(); if (unlikely(!__pyx_t_57)) __PYX_ERR(0, 3841, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_57);
  __pyx_t_56 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_56)) __PYX_ERR(0, 3841, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_56);
  if (PyDict_SetItem(__pyx_t_57, __pyx_n_s_listener, __pyx_t_56) < 0) __PYX_ERR(0, 3841, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_56); __pyx_t_56 = 0;
  __pyx_t_56 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_9enterRule, 0, __pyx_n_s_ModelicaParser_Component_declara_15, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1781)); if (unlikely(!__pyx_t_56)) __PYX_ERR(0, 3841, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_56);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_56, __pyx_t_57);
  __Pyx_DECREF(__pyx_t_57); __pyx_t_57 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_56) < 0) __PYX_ERR(0, 3841, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_56); __pyx_t_56 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3845
 *                 listener.enterComponent_declaration1(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitComponent_declaration1" ):
 *                 listener.exitComponent_declaration1(self)
 */
  __pyx_t_56 = PyDict_New(); if (unlikely(!__pyx_t_56)) __PYX_ERR(0, 3845, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_56);
  __pyx_t_57 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_57)) __PYX_ERR(0, 3845, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_57);
  if (PyDict_SetItem(__pyx_t_56, __pyx_n_s_listener, __pyx_t_57) < 0) __PYX_ERR(0, 3845, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_57); __pyx_t_57 = 0;
  __pyx_t_57 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_11exitRule, 0, __pyx_n_s_ModelicaParser_Component_declara_16, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1783)); if (unlikely(!__pyx_t_57)) __PYX_ERR(0, 3845, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_57);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_57, __pyx_t_56);
  __Pyx_DECREF(__pyx_t_56); __pyx_t_56 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_57) < 0) __PYX_ERR(0, 3845, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_57); __pyx_t_57 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3849
 *                 listener.exitComponent_declaration1(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitComponent_declaration1" ):
 *                 return visitor.visitComponent_declaration1(self)
 */
  __pyx_t_57 = PyDict_New(); if (unlikely(!__pyx_t_57)) __PYX_ERR(0, 3849, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_57);
  __pyx_t_56 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_56)) __PYX_ERR(0, 3849, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_56);
  if (PyDict_SetItem(__pyx_t_57, __pyx_n_s_visitor, __pyx_t_56) < 0) __PYX_ERR(0, 3849, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_56); __pyx_t_56 = 0;
  __pyx_t_56 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_29Component_declaration1Context_13accept, 0, __pyx_n_s_ModelicaParser_Component_declara_17, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1785)); if (unlikely(!__pyx_t_56)) __PYX_ERR(0, 3849, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_56);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_56, __pyx_t_57);
  __Pyx_DECREF(__pyx_t_57); __pyx_t_57 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_56) < 0) __PYX_ERR(0, 3849, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_56); __pyx_t_56 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3824
 *         return localctx
 * 
 *     class Component_declaration1Context(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_56 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Component_declaration1Context, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_56)) __PYX_ERR(0, 3824, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_56);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_55, __pyx_t_56) < 0) __PYX_ERR(0, 3824, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_55); __pyx_t_55 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Component_declaration1Context, __pyx_t_56) < 0) __PYX_ERR(0, 3824, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_56); __pyx_t_56 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3858
 * 
 * 
 *     def component_declaration1(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Component_declaration1Context(self, self._ctx, self.state)
 */
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_75component_declaration1, 0, __pyx_n_s_ModelicaParser_component_declara_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1787)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_component_declaration1, __pyx_t_7) < 0) __PYX_ERR(0, 3858, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3876
 *         return localctx
 * 
 *     class Short_class_definitionContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3876, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3876, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3876, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Short_class_definitionContext, __pyx_n_s_ModelicaParser_Short_class_defin, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3876, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_56 = PyList_New(0); if (unlikely(!__pyx_t_56)) __PYX_ERR(0, 3876, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_56);

  /* "pymola/generated/ModelicaParser.pyx":3878
 *     class Short_class_definitionContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_57 = PyDict_New(); if (unlikely(!__pyx_t_57)) __PYX_ERR(0, 3878, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_57);
  __pyx_t_58 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_58)) __PYX_ERR(0, 3878, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_58);
  if (PyDict_SetItem(__pyx_t_57, __pyx_n_s_parent, __pyx_t_58) < 0) __PYX_ERR(0, 3878, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_58); __pyx_t_58 = 0;
  if (PyDict_SetItem(__pyx_t_57, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 3878, __pyx_L1_error)
  __pyx_t_58 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_1__init__, 0, __pyx_n_s_ModelicaParser_Short_class_defin_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1789)); if (unlikely(!__pyx_t_58)) __PYX_ERR(0, 3878, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_58);
  __Pyx_INCREF(__pyx_t_58);
  PyList_Append(__pyx_t_56, __pyx_t_58);
  __Pyx_GIVEREF(__pyx_t_58);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_58, __pyx_tuple__1790);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_58, __pyx_t_57);
  __Pyx_DECREF(__pyx_t_57); __pyx_t_57 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_58) < 0) __PYX_ERR(0, 3878, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_58); __pyx_t_58 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3882
 *             self.parser = parser
 * 
 *         def class_prefixes(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_prefixesContext,0)
 * 
 */
  __pyx_t_58 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_3class_prefixes, 0, __pyx_n_s_ModelicaParser_Short_class_defin_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1792)); if (unlikely(!__pyx_t_58)) __PYX_ERR(0, 3882, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_58);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_class_prefixes, __pyx_t_58) < 0) __PYX_ERR(0, 3882, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_58); __pyx_t_58 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3886
 * 
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 */
  __pyx_t_58 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_5IDENT, 0, __pyx_n_s_ModelicaParser_Short_class_defin_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1794)); if (unlikely(!__pyx_t_58)) __PYX_ERR(0, 3886, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_58);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_IDENT, __pyx_t_58) < 0) __PYX_ERR(0, 3886, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_58); __pyx_t_58 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3889
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 *         def base_prefix(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Base_prefixContext,0)
 * 
 */
  __pyx_t_58 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_7base_prefix, 0, __pyx_n_s_ModelicaParser_Short_class_defin_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1796)); if (unlikely(!__pyx_t_58)) __PYX_ERR(0, 3889, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_58);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_base_prefix, __pyx_t_58) < 0) __PYX_ERR(0, 3889, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_58); __pyx_t_58 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3893
 * 
 * 
 *         def component_reference(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 */
  __pyx_t_58 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_9component_reference, 0, __pyx_n_s_ModelicaParser_Short_class_defin_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1798)); if (unlikely(!__pyx_t_58)) __PYX_ERR(0, 3893, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_58);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_component_reference, __pyx_t_58) < 0) __PYX_ERR(0, 3893, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_58); __pyx_t_58 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3897
 * 
 * 
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */
  __pyx_t_58 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_11comment, 0, __pyx_n_s_ModelicaParser_Short_class_defin_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1800)); if (unlikely(!__pyx_t_58)) __PYX_ERR(0, 3897, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_58);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_comment, __pyx_t_58) < 0) __PYX_ERR(0, 3897, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_58); __pyx_t_58 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3901
 * 
 * 
 *         def array_subscripts(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Array_subscriptsContext,0)
 * 
 */
  __pyx_t_58 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_13array_subscripts, 0, __pyx_n_s_ModelicaParser_Short_class_defin_8, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1802)); if (unlikely(!__pyx_t_58)) __PYX_ERR(0, 3901, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_58);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_array_subscripts, __pyx_t_58) < 0) __PYX_ERR(0, 3901, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_58); __pyx_t_58 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3905
 * 
 * 
 *         def class_modification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_modificationContext,0)
 * 
 */
  __pyx_t_58 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_15class_modification, 0, __pyx_n_s_ModelicaParser_Short_class_defin_9, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1804)); if (unlikely(!__pyx_t_58)) __PYX_ERR(0, 3905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_58);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_class_modification, __pyx_t_58) < 0) __PYX_ERR(0, 3905, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_58); __pyx_t_58 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3909
 * 
 * 
 *         def enum_list(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Enum_listContext,0)
 * 
 */
  __pyx_t_58 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_17enum_list, 0, __pyx_n_s_ModelicaParser_Short_class_defin_10, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1806)); if (unlikely(!__pyx_t_58)) __PYX_ERR(0, 3909, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_58);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enum_list, __pyx_t_58) < 0) __PYX_ERR(0, 3909, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_58); __pyx_t_58 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3913
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_short_class_definition
 * 
 */
  __pyx_t_58 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_19getRuleIndex, 0, __pyx_n_s_ModelicaParser_Short_class_defin_11, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1808)); if (unlikely(!__pyx_t_58)) __PYX_ERR(0, 3913, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_58);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_58) < 0) __PYX_ERR(0, 3913, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_58); __pyx_t_58 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3916
 *             return ModelicaParser.RULE_short_class_definition
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterShort_class_definition" ):
 *                 listener.enterShort_class_definition(self)
 */
  __pyx_t_58 = PyDict_New(); if (unlikely(!__pyx_t_58)) __PYX_ERR(0, 3916, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_58);
  __pyx_t_57 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_57)) __PYX_ERR(0, 3916, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_57);
  if (PyDict_SetItem(__pyx_t_58, __pyx_n_s_listener, __pyx_t_57) < 0) __PYX_ERR(0, 3916, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_57); __pyx_t_57 = 0;
  __pyx_t_57 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_21enterRule, 0, __pyx_n_s_ModelicaParser_Short_class_defin_12, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1810)); if (unlikely(!__pyx_t_57)) __PYX_ERR(0, 3916, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_57);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_57, __pyx_t_58);
  __Pyx_DECREF(__pyx_t_58); __pyx_t_58 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_57) < 0) __PYX_ERR(0, 3916, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_57); __pyx_t_57 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3920
 *                 listener.enterShort_class_definition(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitShort_class_definition" ):
 *                 listener.exitShort_class_definition(self)
 */
  __pyx_t_57 = PyDict_New(); if (unlikely(!__pyx_t_57)) __PYX_ERR(0, 3920, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_57);
  __pyx_t_58 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_58)) __PYX_ERR(0, 3920, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_58);
  if (PyDict_SetItem(__pyx_t_57, __pyx_n_s_listener, __pyx_t_58) < 0) __PYX_ERR(0, 3920, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_58); __pyx_t_58 = 0;
  __pyx_t_58 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_23exitRule, 0, __pyx_n_s_ModelicaParser_Short_class_defin_13, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1812)); if (unlikely(!__pyx_t_58)) __PYX_ERR(0, 3920, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_58);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_58, __pyx_t_57);
  __Pyx_DECREF(__pyx_t_57); __pyx_t_57 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_58) < 0) __PYX_ERR(0, 3920, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_58); __pyx_t_58 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3924
 *                 listener.exitShort_class_definition(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitShort_class_definition" ):
 *                 return visitor.visitShort_class_definition(self)
 */
  __pyx_t_58 = PyDict_New(); if (unlikely(!__pyx_t_58)) __PYX_ERR(0, 3924, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_58);
  __pyx_t_57 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_57)) __PYX_ERR(0, 3924, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_57);
  if (PyDict_SetItem(__pyx_t_58, __pyx_n_s_visitor, __pyx_t_57) < 0) __PYX_ERR(0, 3924, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_57); __pyx_t_57 = 0;
  __pyx_t_57 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_29Short_class_definitionContext_25accept, 0, __pyx_n_s_ModelicaParser_Short_class_defin_14, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1814)); if (unlikely(!__pyx_t_57)) __PYX_ERR(0, 3924, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_57);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_57, __pyx_t_58);
  __Pyx_DECREF(__pyx_t_58); __pyx_t_58 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_57) < 0) __PYX_ERR(0, 3924, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_57); __pyx_t_57 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3876
 *         return localctx
 * 
 *     class Short_class_definitionContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_57 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Short_class_definitionContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_57)) __PYX_ERR(0, 3876, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_57);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_56, __pyx_t_57) < 0) __PYX_ERR(0, 3876, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_56); __pyx_t_56 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Short_class_definitionContext, __pyx_t_57) < 0) __PYX_ERR(0, 3876, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_57); __pyx_t_57 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":3933
 * 
 * 
 *     def short_class_definition(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Short_class_definitionContext(self, self._ctx, self.state)
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_77short_class_definition, 0, __pyx_n_s_ModelicaParser_short_class_defin, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1816)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3933, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_short_class_definition, __pyx_t_5) < 0) __PYX_ERR(0, 3933, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4014
 *         return localctx
 * 
 *     class Equation_sectionContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Equation_sectionContext, __pyx_n_s_ModelicaParser_Equation_sectionC, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_57 = PyList_New(0); if (unlikely(!__pyx_t_57)) __PYX_ERR(0, 4014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_57);

  /* "pymola/generated/ModelicaParser.pyx":4016
 *     class Equation_sectionContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_58 = PyDict_New(); if (unlikely(!__pyx_t_58)) __PYX_ERR(0, 4016, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_58);
  __pyx_t_59 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_59)) __PYX_ERR(0, 4016, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_59);
  if (PyDict_SetItem(__pyx_t_58, __pyx_n_s_parent, __pyx_t_59) < 0) __PYX_ERR(0, 4016, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_59); __pyx_t_59 = 0;
  if (PyDict_SetItem(__pyx_t_58, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 4016, __pyx_L1_error)
  __pyx_t_59 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_1__init__, 0, __pyx_n_s_ModelicaParser_Equation_sectionC_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1818)); if (unlikely(!__pyx_t_59)) __PYX_ERR(0, 4016, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_59);
  __Pyx_INCREF(__pyx_t_59);
  PyList_Append(__pyx_t_57, __pyx_t_59);
  __Pyx_GIVEREF(__pyx_t_59);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_59, __pyx_tuple__1819);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_59, __pyx_t_58);
  __Pyx_DECREF(__pyx_t_58); __pyx_t_58 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_59) < 0) __PYX_ERR(0, 4016, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_59); __pyx_t_59 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4020
 *             self.parser = parser
 * 
 *         def INITIAL(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.INITIAL, 0)
 * 
 */
  __pyx_t_59 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_3INITIAL, 0, __pyx_n_s_ModelicaParser_Equation_sectionC_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1821)); if (unlikely(!__pyx_t_59)) __PYX_ERR(0, 4020, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_59);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_INITIAL, __pyx_t_59) < 0) __PYX_ERR(0, 4020, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_59); __pyx_t_59 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4023
 *             return self.getToken(ModelicaParser.INITIAL, 0)
 * 
 *         def equation(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.EquationContext)
 */
  __pyx_t_59 = PyDict_New(); if (unlikely(!__pyx_t_59)) __PYX_ERR(0, 4023, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_59);
  if (PyDict_SetItem(__pyx_t_59, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 4023, __pyx_L1_error)
  __pyx_t_58 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_5equation, 0, __pyx_n_s_ModelicaParser_Equation_sectionC_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1823)); if (unlikely(!__pyx_t_58)) __PYX_ERR(0, 4023, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_58);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_58, __pyx_tuple__1824);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_58, __pyx_t_59);
  __Pyx_DECREF(__pyx_t_59); __pyx_t_59 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_equation, __pyx_t_58) < 0) __PYX_ERR(0, 4023, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_58); __pyx_t_58 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4030
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_equation_section
 * 
 */
  __pyx_t_58 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_7getRuleIndex, 0, __pyx_n_s_ModelicaParser_Equation_sectionC_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1826)); if (unlikely(!__pyx_t_58)) __PYX_ERR(0, 4030, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_58);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_58) < 0) __PYX_ERR(0, 4030, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_58); __pyx_t_58 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4033
 *             return ModelicaParser.RULE_equation_section
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterEquation_section" ):
 *                 listener.enterEquation_section(self)
 */
  __pyx_t_58 = PyDict_New(); if (unlikely(!__pyx_t_58)) __PYX_ERR(0, 4033, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_58);
  __pyx_t_59 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_59)) __PYX_ERR(0, 4033, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_59);
  if (PyDict_SetItem(__pyx_t_58, __pyx_n_s_listener, __pyx_t_59) < 0) __PYX_ERR(0, 4033, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_59); __pyx_t_59 = 0;
  __pyx_t_59 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_9enterRule, 0, __pyx_n_s_ModelicaParser_Equation_sectionC_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1828)); if (unlikely(!__pyx_t_59)) __PYX_ERR(0, 4033, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_59);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_59, __pyx_t_58);
  __Pyx_DECREF(__pyx_t_58); __pyx_t_58 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_59) < 0) __PYX_ERR(0, 4033, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_59); __pyx_t_59 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4037
 *                 listener.enterEquation_section(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitEquation_section" ):
 *                 listener.exitEquation_section(self)
 */
  __pyx_t_59 = PyDict_New(); if (unlikely(!__pyx_t_59)) __PYX_ERR(0, 4037, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_59);
  __pyx_t_58 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_58)) __PYX_ERR(0, 4037, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_58);
  if (PyDict_SetItem(__pyx_t_59, __pyx_n_s_listener, __pyx_t_58) < 0) __PYX_ERR(0, 4037, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_58); __pyx_t_58 = 0;
  __pyx_t_58 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_11exitRule, 0, __pyx_n_s_ModelicaParser_Equation_sectionC_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1830)); if (unlikely(!__pyx_t_58)) __PYX_ERR(0, 4037, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_58);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_58, __pyx_t_59);
  __Pyx_DECREF(__pyx_t_59); __pyx_t_59 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_58) < 0) __PYX_ERR(0, 4037, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_58); __pyx_t_58 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4041
 *                 listener.exitEquation_section(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitEquation_section" ):
 *                 return visitor.visitEquation_section(self)
 */
  __pyx_t_58 = PyDict_New(); if (unlikely(!__pyx_t_58)) __PYX_ERR(0, 4041, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_58);
  __pyx_t_59 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_59)) __PYX_ERR(0, 4041, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_59);
  if (PyDict_SetItem(__pyx_t_58, __pyx_n_s_visitor, __pyx_t_59) < 0) __PYX_ERR(0, 4041, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_59); __pyx_t_59 = 0;
  __pyx_t_59 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Equation_sectionContext_13accept, 0, __pyx_n_s_ModelicaParser_Equation_sectionC_8, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1832)); if (unlikely(!__pyx_t_59)) __PYX_ERR(0, 4041, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_59);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_59, __pyx_t_58);
  __Pyx_DECREF(__pyx_t_58); __pyx_t_58 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_59) < 0) __PYX_ERR(0, 4041, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_59); __pyx_t_59 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4014
 *         return localctx
 * 
 *     class Equation_sectionContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_59 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Equation_sectionContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_59)) __PYX_ERR(0, 4014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_59);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_57, __pyx_t_59) < 0) __PYX_ERR(0, 4014, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_57); __pyx_t_57 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Equation_sectionContext, __pyx_t_59) < 0) __PYX_ERR(0, 4014, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_59); __pyx_t_59 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4050
 * 
 * 
 *     def equation_section(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Equation_sectionContext(self, self._ctx, self.state)
 */
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_79equation_section, 0, __pyx_n_s_ModelicaParser_equation_section, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1834)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4050, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_equation_section, __pyx_t_7) < 0) __PYX_ERR(0, 4050, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4088
 *         return localctx
 * 
 *     class Algorithm_sectionContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4088, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4088, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4088, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Algorithm_sectionContext, __pyx_n_s_ModelicaParser_Algorithm_section, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4088, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_59 = PyList_New(0); if (unlikely(!__pyx_t_59)) __PYX_ERR(0, 4088, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_59);

  /* "pymola/generated/ModelicaParser.pyx":4090
 *     class Algorithm_sectionContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_58 = PyDict_New(); if (unlikely(!__pyx_t_58)) __PYX_ERR(0, 4090, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_58);
  __pyx_t_60 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_60)) __PYX_ERR(0, 4090, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_60);
  if (PyDict_SetItem(__pyx_t_58, __pyx_n_s_parent, __pyx_t_60) < 0) __PYX_ERR(0, 4090, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_60); __pyx_t_60 = 0;
  if (PyDict_SetItem(__pyx_t_58, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 4090, __pyx_L1_error)
  __pyx_t_60 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_1__init__, 0, __pyx_n_s_ModelicaParser_Algorithm_section_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1836)); if (unlikely(!__pyx_t_60)) __PYX_ERR(0, 4090, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_60);
  __Pyx_INCREF(__pyx_t_60);
  PyList_Append(__pyx_t_59, __pyx_t_60);
  __Pyx_GIVEREF(__pyx_t_60);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_60, __pyx_tuple__1837);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_60, __pyx_t_58);
  __Pyx_DECREF(__pyx_t_58); __pyx_t_58 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_60) < 0) __PYX_ERR(0, 4090, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_60); __pyx_t_60 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4094
 *             self.parser = parser
 * 
 *         def INITIAL(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.INITIAL, 0)
 * 
 */
  __pyx_t_60 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_3INITIAL, 0, __pyx_n_s_ModelicaParser_Algorithm_section_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1839)); if (unlikely(!__pyx_t_60)) __PYX_ERR(0, 4094, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_60);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_INITIAL, __pyx_t_60) < 0) __PYX_ERR(0, 4094, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_60); __pyx_t_60 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4097
 *             return self.getToken(ModelicaParser.INITIAL, 0)
 * 
 *         def statement(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.StatementContext)
 */
  __pyx_t_60 = PyDict_New(); if (unlikely(!__pyx_t_60)) __PYX_ERR(0, 4097, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_60);
  if (PyDict_SetItem(__pyx_t_60, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 4097, __pyx_L1_error)
  __pyx_t_58 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_5statement, 0, __pyx_n_s_ModelicaParser_Algorithm_section_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1841)); if (unlikely(!__pyx_t_58)) __PYX_ERR(0, 4097, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_58);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_58, __pyx_tuple__1842);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_58, __pyx_t_60);
  __Pyx_DECREF(__pyx_t_60); __pyx_t_60 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_statement, __pyx_t_58) < 0) __PYX_ERR(0, 4097, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_58); __pyx_t_58 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4104
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_algorithm_section
 * 
 */
  __pyx_t_58 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_7getRuleIndex, 0, __pyx_n_s_ModelicaParser_Algorithm_section_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1844)); if (unlikely(!__pyx_t_58)) __PYX_ERR(0, 4104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_58);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_58) < 0) __PYX_ERR(0, 4104, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_58); __pyx_t_58 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4107
 *             return ModelicaParser.RULE_algorithm_section
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterAlgorithm_section" ):
 *                 listener.enterAlgorithm_section(self)
 */
  __pyx_t_58 = PyDict_New(); if (unlikely(!__pyx_t_58)) __PYX_ERR(0, 4107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_58);
  __pyx_t_60 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_60)) __PYX_ERR(0, 4107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_60);
  if (PyDict_SetItem(__pyx_t_58, __pyx_n_s_listener, __pyx_t_60) < 0) __PYX_ERR(0, 4107, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_60); __pyx_t_60 = 0;
  __pyx_t_60 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_9enterRule, 0, __pyx_n_s_ModelicaParser_Algorithm_section_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1846)); if (unlikely(!__pyx_t_60)) __PYX_ERR(0, 4107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_60);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_60, __pyx_t_58);
  __Pyx_DECREF(__pyx_t_58); __pyx_t_58 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_60) < 0) __PYX_ERR(0, 4107, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_60); __pyx_t_60 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4111
 *                 listener.enterAlgorithm_section(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitAlgorithm_section" ):
 *                 listener.exitAlgorithm_section(self)
 */
  __pyx_t_60 = PyDict_New(); if (unlikely(!__pyx_t_60)) __PYX_ERR(0, 4111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_60);
  __pyx_t_58 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_58)) __PYX_ERR(0, 4111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_58);
  if (PyDict_SetItem(__pyx_t_60, __pyx_n_s_listener, __pyx_t_58) < 0) __PYX_ERR(0, 4111, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_58); __pyx_t_58 = 0;
  __pyx_t_58 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_11exitRule, 0, __pyx_n_s_ModelicaParser_Algorithm_section_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1848)); if (unlikely(!__pyx_t_58)) __PYX_ERR(0, 4111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_58);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_58, __pyx_t_60);
  __Pyx_DECREF(__pyx_t_60); __pyx_t_60 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_58) < 0) __PYX_ERR(0, 4111, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_58); __pyx_t_58 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4115
 *                 listener.exitAlgorithm_section(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitAlgorithm_section" ):
 *                 return visitor.visitAlgorithm_section(self)
 */
  __pyx_t_58 = PyDict_New(); if (unlikely(!__pyx_t_58)) __PYX_ERR(0, 4115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_58);
  __pyx_t_60 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_60)) __PYX_ERR(0, 4115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_60);
  if (PyDict_SetItem(__pyx_t_58, __pyx_n_s_visitor, __pyx_t_60) < 0) __PYX_ERR(0, 4115, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_60); __pyx_t_60 = 0;
  __pyx_t_60 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Algorithm_sectionContext_13accept, 0, __pyx_n_s_ModelicaParser_Algorithm_section_8, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1850)); if (unlikely(!__pyx_t_60)) __PYX_ERR(0, 4115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_60);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_60, __pyx_t_58);
  __Pyx_DECREF(__pyx_t_58); __pyx_t_58 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_60) < 0) __PYX_ERR(0, 4115, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_60); __pyx_t_60 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4088
 *         return localctx
 * 
 *     class Algorithm_sectionContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_60 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Algorithm_sectionContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_60)) __PYX_ERR(0, 4088, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_60);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_59, __pyx_t_60) < 0) __PYX_ERR(0, 4088, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_59); __pyx_t_59 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Algorithm_sectionContext, __pyx_t_60) < 0) __PYX_ERR(0, 4088, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_60); __pyx_t_60 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4124
 * 
 * 
 *     def algorithm_section(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Algorithm_sectionContext(self, self._ctx, self.state)
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_81algorithm_section, 0, __pyx_n_s_ModelicaParser_algorithm_section, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1852)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_algorithm_section, __pyx_t_5) < 0) __PYX_ERR(0, 4124, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4161
 *         return localctx
 * 
 *     class Equation_optionsContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Equation_optionsContext, __pyx_n_s_ModelicaParser_Equation_optionsC, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_60 = PyList_New(0); if (unlikely(!__pyx_t_60)) __PYX_ERR(0, 4161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_60);

  /* "pymola/generated/ModelicaParser.pyx":4163
 *     class Equation_optionsContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_58 = PyDict_New(); if (unlikely(!__pyx_t_58)) __PYX_ERR(0, 4163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_58);
  __pyx_t_61 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_61)) __PYX_ERR(0, 4163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_61);
  if (PyDict_SetItem(__pyx_t_58, __pyx_n_s_parent, __pyx_t_61) < 0) __PYX_ERR(0, 4163, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_61); __pyx_t_61 = 0;
  if (PyDict_SetItem(__pyx_t_58, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 4163, __pyx_L1_error)
  __pyx_t_61 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Equation_optionsContext_1__init__, 0, __pyx_n_s_ModelicaParser_Equation_optionsC_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1854)); if (unlikely(!__pyx_t_61)) __PYX_ERR(0, 4163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_61);
  __Pyx_INCREF(__pyx_t_61);
  PyList_Append(__pyx_t_60, __pyx_t_61);
  __Pyx_GIVEREF(__pyx_t_61);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_61, __pyx_tuple__1855);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_61, __pyx_t_58);
  __Pyx_DECREF(__pyx_t_58); __pyx_t_58 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_61) < 0) __PYX_ERR(0, 4163, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_61); __pyx_t_61 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4168
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_equation_options
 * 
 */
  __pyx_t_61 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Equation_optionsContext_3getRuleIndex, 0, __pyx_n_s_ModelicaParser_Equation_optionsC_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1857)); if (unlikely(!__pyx_t_61)) __PYX_ERR(0, 4168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_61);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_61) < 0) __PYX_ERR(0, 4168, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_61); __pyx_t_61 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4172
 * 
 * 
 *         def copyFrom(self, ctx:ParserRuleContext):             # <<<<<<<<<<<<<<
 *             super().copyFrom(ctx)
 * 
 */
  __pyx_t_61 = PyDict_New(); if (unlikely(!__pyx_t_61)) __PYX_ERR(0, 4172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_61);
  __pyx_t_58 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_58)) __PYX_ERR(0, 4172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_58);
  if (PyDict_SetItem(__pyx_t_61, __pyx_n_s_ctx_2, __pyx_t_58) < 0) __PYX_ERR(0, 4172, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_58); __pyx_t_58 = 0;
  __pyx_t_58 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Equation_optionsContext_5copyFrom, 0, __pyx_n_s_ModelicaParser_Equation_optionsC_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1859)); if (unlikely(!__pyx_t_58)) __PYX_ERR(0, 4172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_58);
  __Pyx_INCREF(__pyx_t_58);
  PyList_Append(__pyx_t_60, __pyx_t_58);
  __Pyx_GIVEREF(__pyx_t_58);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_58, __pyx_t_61);
  __Pyx_DECREF(__pyx_t_61); __pyx_t_61 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_copyFrom, __pyx_t_58) < 0) __PYX_ERR(0, 4172, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_58); __pyx_t_58 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4161
 *         return localctx
 * 
 *     class Equation_optionsContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_58 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Equation_optionsContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_58)) __PYX_ERR(0, 4161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_58);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_60, __pyx_t_58) < 0) __PYX_ERR(0, 4161, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_60); __pyx_t_60 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Equation_optionsContext, __pyx_t_58) < 0) __PYX_ERR(0, 4161, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_58); __pyx_t_58 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4177
 * 
 * 
 *     class Equation_whenContext(Equation_optionsContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext
 */
  __pyx_t_7 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_Equation_optionsContext);
  if (unlikely(!__pyx_t_7)) {
    PyErr_Clear();
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_Equation_optionsContext);
  }
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Equation_whenContext, __pyx_n_s_ModelicaParser_Equation_whenCont, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_58 = PyList_New(0); if (unlikely(!__pyx_t_58)) __PYX_ERR(0, 4177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_58);

  /* "pymola/generated/ModelicaParser.pyx":4179
 *     class Equation_whenContext(Equation_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_61 = PyDict_New(); if (unlikely(!__pyx_t_61)) __PYX_ERR(0, 4179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_61);
  __pyx_t_62 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_62)) __PYX_ERR(0, 4179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_62);
  if (PyDict_SetItem(__pyx_t_61, __pyx_n_s_ctx_2, __pyx_t_62) < 0) __PYX_ERR(0, 4179, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_62); __pyx_t_62 = 0;
  __pyx_t_62 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20Equation_whenContext_1__init__, 0, __pyx_n_s_ModelicaParser_Equation_whenCont_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1861)); if (unlikely(!__pyx_t_62)) __PYX_ERR(0, 4179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_62);
  __Pyx_INCREF(__pyx_t_62);
  PyList_Append(__pyx_t_58, __pyx_t_62);
  __Pyx_GIVEREF(__pyx_t_62);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_62, __pyx_t_61);
  __Pyx_DECREF(__pyx_t_61); __pyx_t_61 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_62) < 0) __PYX_ERR(0, 4179, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_62); __pyx_t_62 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4183
 *             self.copyFrom(ctx)
 * 
 *         def when_equation(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.When_equationContext,0)
 * 
 */
  __pyx_t_62 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20Equation_whenContext_3when_equation, 0, __pyx_n_s_ModelicaParser_Equation_whenCont_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1863)); if (unlikely(!__pyx_t_62)) __PYX_ERR(0, 4183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_62);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_when_equation, __pyx_t_62) < 0) __PYX_ERR(0, 4183, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_62); __pyx_t_62 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4187
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterEquation_when" ):
 *                 listener.enterEquation_when(self)
 */
  __pyx_t_62 = PyDict_New(); if (unlikely(!__pyx_t_62)) __PYX_ERR(0, 4187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_62);
  __pyx_t_61 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_61)) __PYX_ERR(0, 4187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_61);
  if (PyDict_SetItem(__pyx_t_62, __pyx_n_s_listener, __pyx_t_61) < 0) __PYX_ERR(0, 4187, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_61); __pyx_t_61 = 0;
  __pyx_t_61 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20Equation_whenContext_5enterRule, 0, __pyx_n_s_ModelicaParser_Equation_whenCont_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1865)); if (unlikely(!__pyx_t_61)) __PYX_ERR(0, 4187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_61);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_61, __pyx_t_62);
  __Pyx_DECREF(__pyx_t_62); __pyx_t_62 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_61) < 0) __PYX_ERR(0, 4187, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_61); __pyx_t_61 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4191
 *                 listener.enterEquation_when(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitEquation_when" ):
 *                 listener.exitEquation_when(self)
 */
  __pyx_t_61 = PyDict_New(); if (unlikely(!__pyx_t_61)) __PYX_ERR(0, 4191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_61);
  __pyx_t_62 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_62)) __PYX_ERR(0, 4191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_62);
  if (PyDict_SetItem(__pyx_t_61, __pyx_n_s_listener, __pyx_t_62) < 0) __PYX_ERR(0, 4191, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_62); __pyx_t_62 = 0;
  __pyx_t_62 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20Equation_whenContext_7exitRule, 0, __pyx_n_s_ModelicaParser_Equation_whenCont_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1867)); if (unlikely(!__pyx_t_62)) __PYX_ERR(0, 4191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_62);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_62, __pyx_t_61);
  __Pyx_DECREF(__pyx_t_61); __pyx_t_61 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_62) < 0) __PYX_ERR(0, 4191, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_62); __pyx_t_62 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4195
 *                 listener.exitEquation_when(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitEquation_when" ):
 *                 return visitor.visitEquation_when(self)
 */
  __pyx_t_62 = PyDict_New(); if (unlikely(!__pyx_t_62)) __PYX_ERR(0, 4195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_62);
  __pyx_t_61 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_61)) __PYX_ERR(0, 4195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_61);
  if (PyDict_SetItem(__pyx_t_62, __pyx_n_s_visitor, __pyx_t_61) < 0) __PYX_ERR(0, 4195, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_61); __pyx_t_61 = 0;
  __pyx_t_61 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20Equation_whenContext_9accept, 0, __pyx_n_s_ModelicaParser_Equation_whenCont_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1869)); if (unlikely(!__pyx_t_61)) __PYX_ERR(0, 4195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_61);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_61, __pyx_t_62);
  __Pyx_DECREF(__pyx_t_62); __pyx_t_62 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_61) < 0) __PYX_ERR(0, 4195, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_61); __pyx_t_61 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4177
 * 
 * 
 *     class Equation_whenContext(Equation_optionsContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext
 */
  __pyx_t_61 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Equation_whenContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_61)) __PYX_ERR(0, 4177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_61);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_58, __pyx_t_61) < 0) __PYX_ERR(0, 4177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_58); __pyx_t_58 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Equation_whenContext, __pyx_t_61) < 0) __PYX_ERR(0, 4177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_61); __pyx_t_61 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4202
 * 
 * 
 *     class Equation_connect_clauseContext(Equation_optionsContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext
 */
  __pyx_t_5 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_Equation_optionsContext);
  if (unlikely(!__pyx_t_5)) {
    PyErr_Clear();
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_Equation_optionsContext);
  }
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Equation_connect_clauseContext, __pyx_n_s_ModelicaParser_Equation_connect, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_61 = PyList_New(0); if (unlikely(!__pyx_t_61)) __PYX_ERR(0, 4202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_61);

  /* "pymola/generated/ModelicaParser.pyx":4204
 *     class Equation_connect_clauseContext(Equation_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_62 = PyDict_New(); if (unlikely(!__pyx_t_62)) __PYX_ERR(0, 4204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_62);
  __pyx_t_63 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_63)) __PYX_ERR(0, 4204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_63);
  if (PyDict_SetItem(__pyx_t_62, __pyx_n_s_ctx_2, __pyx_t_63) < 0) __PYX_ERR(0, 4204, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_63); __pyx_t_63 = 0;
  __pyx_t_63 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_30Equation_connect_clauseContext_1__init__, 0, __pyx_n_s_ModelicaParser_Equation_connect_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1871)); if (unlikely(!__pyx_t_63)) __PYX_ERR(0, 4204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_63);
  __Pyx_INCREF(__pyx_t_63);
  PyList_Append(__pyx_t_61, __pyx_t_63);
  __Pyx_GIVEREF(__pyx_t_63);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_63, __pyx_t_62);
  __Pyx_DECREF(__pyx_t_62); __pyx_t_62 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_63) < 0) __PYX_ERR(0, 4204, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_63); __pyx_t_63 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4208
 *             self.copyFrom(ctx)
 * 
 *         def connect_clause(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Connect_clauseContext,0)
 * 
 */
  __pyx_t_63 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_30Equation_connect_clauseContext_3connect_clause, 0, __pyx_n_s_ModelicaParser_Equation_connect_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1873)); if (unlikely(!__pyx_t_63)) __PYX_ERR(0, 4208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_63);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_connect_clause, __pyx_t_63) < 0) __PYX_ERR(0, 4208, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_63); __pyx_t_63 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4212
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterEquation_connect_clause" ):
 *                 listener.enterEquation_connect_clause(self)
 */
  __pyx_t_63 = PyDict_New(); if (unlikely(!__pyx_t_63)) __PYX_ERR(0, 4212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_63);
  __pyx_t_62 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_62)) __PYX_ERR(0, 4212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_62);
  if (PyDict_SetItem(__pyx_t_63, __pyx_n_s_listener, __pyx_t_62) < 0) __PYX_ERR(0, 4212, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_62); __pyx_t_62 = 0;
  __pyx_t_62 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_30Equation_connect_clauseContext_5enterRule, 0, __pyx_n_s_ModelicaParser_Equation_connect_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1875)); if (unlikely(!__pyx_t_62)) __PYX_ERR(0, 4212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_62);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_62, __pyx_t_63);
  __Pyx_DECREF(__pyx_t_63); __pyx_t_63 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_62) < 0) __PYX_ERR(0, 4212, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_62); __pyx_t_62 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4216
 *                 listener.enterEquation_connect_clause(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitEquation_connect_clause" ):
 *                 listener.exitEquation_connect_clause(self)
 */
  __pyx_t_62 = PyDict_New(); if (unlikely(!__pyx_t_62)) __PYX_ERR(0, 4216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_62);
  __pyx_t_63 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_63)) __PYX_ERR(0, 4216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_63);
  if (PyDict_SetItem(__pyx_t_62, __pyx_n_s_listener, __pyx_t_63) < 0) __PYX_ERR(0, 4216, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_63); __pyx_t_63 = 0;
  __pyx_t_63 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_30Equation_connect_clauseContext_7exitRule, 0, __pyx_n_s_ModelicaParser_Equation_connect_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1877)); if (unlikely(!__pyx_t_63)) __PYX_ERR(0, 4216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_63);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_63, __pyx_t_62);
  __Pyx_DECREF(__pyx_t_62); __pyx_t_62 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_63) < 0) __PYX_ERR(0, 4216, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_63); __pyx_t_63 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4220
 *                 listener.exitEquation_connect_clause(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitEquation_connect_clause" ):
 *                 return visitor.visitEquation_connect_clause(self)
 */
  __pyx_t_63 = PyDict_New(); if (unlikely(!__pyx_t_63)) __PYX_ERR(0, 4220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_63);
  __pyx_t_62 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_62)) __PYX_ERR(0, 4220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_62);
  if (PyDict_SetItem(__pyx_t_63, __pyx_n_s_visitor, __pyx_t_62) < 0) __PYX_ERR(0, 4220, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_62); __pyx_t_62 = 0;
  __pyx_t_62 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_30Equation_connect_clauseContext_9accept, 0, __pyx_n_s_ModelicaParser_Equation_connect_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1879)); if (unlikely(!__pyx_t_62)) __PYX_ERR(0, 4220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_62);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_62, __pyx_t_63);
  __Pyx_DECREF(__pyx_t_63); __pyx_t_63 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_62) < 0) __PYX_ERR(0, 4220, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_62); __pyx_t_62 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4202
 * 
 * 
 *     class Equation_connect_clauseContext(Equation_optionsContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext
 */
  __pyx_t_62 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Equation_connect_clauseContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_62)) __PYX_ERR(0, 4202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_62);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_61, __pyx_t_62) < 0) __PYX_ERR(0, 4202, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_61); __pyx_t_61 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Equation_connect_clauseContext, __pyx_t_62) < 0) __PYX_ERR(0, 4202, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_62); __pyx_t_62 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4227
 * 
 * 
 *     class Equation_forContext(Equation_optionsContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext
 */
  __pyx_t_7 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_Equation_optionsContext);
  if (unlikely(!__pyx_t_7)) {
    PyErr_Clear();
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_Equation_optionsContext);
  }
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Equation_forContext, __pyx_n_s_ModelicaParser_Equation_forConte, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_62 = PyList_New(0); if (unlikely(!__pyx_t_62)) __PYX_ERR(0, 4227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_62);

  /* "pymola/generated/ModelicaParser.pyx":4229
 *     class Equation_forContext(Equation_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_63 = PyDict_New(); if (unlikely(!__pyx_t_63)) __PYX_ERR(0, 4229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_63);
  __pyx_t_64 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_64)) __PYX_ERR(0, 4229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_64);
  if (PyDict_SetItem(__pyx_t_63, __pyx_n_s_ctx_2, __pyx_t_64) < 0) __PYX_ERR(0, 4229, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_64); __pyx_t_64 = 0;
  __pyx_t_64 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19Equation_forContext_1__init__, 0, __pyx_n_s_ModelicaParser_Equation_forConte_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1881)); if (unlikely(!__pyx_t_64)) __PYX_ERR(0, 4229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_64);
  __Pyx_INCREF(__pyx_t_64);
  PyList_Append(__pyx_t_62, __pyx_t_64);
  __Pyx_GIVEREF(__pyx_t_64);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_64, __pyx_t_63);
  __Pyx_DECREF(__pyx_t_63); __pyx_t_63 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_64) < 0) __PYX_ERR(0, 4229, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_64); __pyx_t_64 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4233
 *             self.copyFrom(ctx)
 * 
 *         def for_equation(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.For_equationContext,0)
 * 
 */
  __pyx_t_64 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19Equation_forContext_3for_equation, 0, __pyx_n_s_ModelicaParser_Equation_forConte_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1883)); if (unlikely(!__pyx_t_64)) __PYX_ERR(0, 4233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_64);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_for_equation, __pyx_t_64) < 0) __PYX_ERR(0, 4233, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_64); __pyx_t_64 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4237
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterEquation_for" ):
 *                 listener.enterEquation_for(self)
 */
  __pyx_t_64 = PyDict_New(); if (unlikely(!__pyx_t_64)) __PYX_ERR(0, 4237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_64);
  __pyx_t_63 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_63)) __PYX_ERR(0, 4237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_63);
  if (PyDict_SetItem(__pyx_t_64, __pyx_n_s_listener, __pyx_t_63) < 0) __PYX_ERR(0, 4237, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_63); __pyx_t_63 = 0;
  __pyx_t_63 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19Equation_forContext_5enterRule, 0, __pyx_n_s_ModelicaParser_Equation_forConte_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1885)); if (unlikely(!__pyx_t_63)) __PYX_ERR(0, 4237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_63);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_63, __pyx_t_64);
  __Pyx_DECREF(__pyx_t_64); __pyx_t_64 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_63) < 0) __PYX_ERR(0, 4237, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_63); __pyx_t_63 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4241
 *                 listener.enterEquation_for(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitEquation_for" ):
 *                 listener.exitEquation_for(self)
 */
  __pyx_t_63 = PyDict_New(); if (unlikely(!__pyx_t_63)) __PYX_ERR(0, 4241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_63);
  __pyx_t_64 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_64)) __PYX_ERR(0, 4241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_64);
  if (PyDict_SetItem(__pyx_t_63, __pyx_n_s_listener, __pyx_t_64) < 0) __PYX_ERR(0, 4241, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_64); __pyx_t_64 = 0;
  __pyx_t_64 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19Equation_forContext_7exitRule, 0, __pyx_n_s_ModelicaParser_Equation_forConte_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1887)); if (unlikely(!__pyx_t_64)) __PYX_ERR(0, 4241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_64);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_64, __pyx_t_63);
  __Pyx_DECREF(__pyx_t_63); __pyx_t_63 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_64) < 0) __PYX_ERR(0, 4241, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_64); __pyx_t_64 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4245
 *                 listener.exitEquation_for(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitEquation_for" ):
 *                 return visitor.visitEquation_for(self)
 */
  __pyx_t_64 = PyDict_New(); if (unlikely(!__pyx_t_64)) __PYX_ERR(0, 4245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_64);
  __pyx_t_63 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_63)) __PYX_ERR(0, 4245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_63);
  if (PyDict_SetItem(__pyx_t_64, __pyx_n_s_visitor, __pyx_t_63) < 0) __PYX_ERR(0, 4245, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_63); __pyx_t_63 = 0;
  __pyx_t_63 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19Equation_forContext_9accept, 0, __pyx_n_s_ModelicaParser_Equation_forConte_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1889)); if (unlikely(!__pyx_t_63)) __PYX_ERR(0, 4245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_63);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_63, __pyx_t_64);
  __Pyx_DECREF(__pyx_t_64); __pyx_t_64 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_63) < 0) __PYX_ERR(0, 4245, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_63); __pyx_t_63 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4227
 * 
 * 
 *     class Equation_forContext(Equation_optionsContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext
 */
  __pyx_t_63 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Equation_forContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_63)) __PYX_ERR(0, 4227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_63);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_62, __pyx_t_63) < 0) __PYX_ERR(0, 4227, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_62); __pyx_t_62 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Equation_forContext, __pyx_t_63) < 0) __PYX_ERR(0, 4227, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_63); __pyx_t_63 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4252
 * 
 * 
 *     class Equation_functionContext(Equation_optionsContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext
 */
  __pyx_t_5 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_Equation_optionsContext);
  if (unlikely(!__pyx_t_5)) {
    PyErr_Clear();
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_Equation_optionsContext);
  }
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Equation_functionContext, __pyx_n_s_ModelicaParser_Equation_function, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_63 = PyList_New(0); if (unlikely(!__pyx_t_63)) __PYX_ERR(0, 4252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_63);

  /* "pymola/generated/ModelicaParser.pyx":4254
 *     class Equation_functionContext(Equation_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_64 = PyDict_New(); if (unlikely(!__pyx_t_64)) __PYX_ERR(0, 4254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_64);
  __pyx_t_65 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_65)) __PYX_ERR(0, 4254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_65);
  if (PyDict_SetItem(__pyx_t_64, __pyx_n_s_ctx_2, __pyx_t_65) < 0) __PYX_ERR(0, 4254, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_65); __pyx_t_65 = 0;
  __pyx_t_65 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Equation_functionContext_1__init__, 0, __pyx_n_s_ModelicaParser_Equation_function_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1891)); if (unlikely(!__pyx_t_65)) __PYX_ERR(0, 4254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_65);
  __Pyx_INCREF(__pyx_t_65);
  PyList_Append(__pyx_t_63, __pyx_t_65);
  __Pyx_GIVEREF(__pyx_t_65);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_65, __pyx_t_64);
  __Pyx_DECREF(__pyx_t_64); __pyx_t_64 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_65) < 0) __PYX_ERR(0, 4254, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_65); __pyx_t_65 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4258
 *             self.copyFrom(ctx)
 * 
 *         def name(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)
 * 
 */
  __pyx_t_65 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Equation_functionContext_3name, 0, __pyx_n_s_ModelicaParser_Equation_function_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1893)); if (unlikely(!__pyx_t_65)) __PYX_ERR(0, 4258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_65);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_name, __pyx_t_65) < 0) __PYX_ERR(0, 4258, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_65); __pyx_t_65 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4261
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)
 * 
 *         def function_call_args(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Function_call_argsContext,0)
 * 
 */
  __pyx_t_65 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Equation_functionContext_5function_call_args, 0, __pyx_n_s_ModelicaParser_Equation_function_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1895)); if (unlikely(!__pyx_t_65)) __PYX_ERR(0, 4261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_65);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_function_call_args, __pyx_t_65) < 0) __PYX_ERR(0, 4261, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_65); __pyx_t_65 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4265
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterEquation_function" ):
 *                 listener.enterEquation_function(self)
 */
  __pyx_t_65 = PyDict_New(); if (unlikely(!__pyx_t_65)) __PYX_ERR(0, 4265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_65);
  __pyx_t_64 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_64)) __PYX_ERR(0, 4265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_64);
  if (PyDict_SetItem(__pyx_t_65, __pyx_n_s_listener, __pyx_t_64) < 0) __PYX_ERR(0, 4265, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_64); __pyx_t_64 = 0;
  __pyx_t_64 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Equation_functionContext_7enterRule, 0, __pyx_n_s_ModelicaParser_Equation_function_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1897)); if (unlikely(!__pyx_t_64)) __PYX_ERR(0, 4265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_64);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_64, __pyx_t_65);
  __Pyx_DECREF(__pyx_t_65); __pyx_t_65 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_64) < 0) __PYX_ERR(0, 4265, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_64); __pyx_t_64 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4269
 *                 listener.enterEquation_function(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitEquation_function" ):
 *                 listener.exitEquation_function(self)
 */
  __pyx_t_64 = PyDict_New(); if (unlikely(!__pyx_t_64)) __PYX_ERR(0, 4269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_64);
  __pyx_t_65 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_65)) __PYX_ERR(0, 4269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_65);
  if (PyDict_SetItem(__pyx_t_64, __pyx_n_s_listener, __pyx_t_65) < 0) __PYX_ERR(0, 4269, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_65); __pyx_t_65 = 0;
  __pyx_t_65 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Equation_functionContext_9exitRule, 0, __pyx_n_s_ModelicaParser_Equation_function_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1899)); if (unlikely(!__pyx_t_65)) __PYX_ERR(0, 4269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_65);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_65, __pyx_t_64);
  __Pyx_DECREF(__pyx_t_64); __pyx_t_64 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_65) < 0) __PYX_ERR(0, 4269, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_65); __pyx_t_65 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4273
 *                 listener.exitEquation_function(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitEquation_function" ):
 *                 return visitor.visitEquation_function(self)
 */
  __pyx_t_65 = PyDict_New(); if (unlikely(!__pyx_t_65)) __PYX_ERR(0, 4273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_65);
  __pyx_t_64 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_64)) __PYX_ERR(0, 4273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_64);
  if (PyDict_SetItem(__pyx_t_65, __pyx_n_s_visitor, __pyx_t_64) < 0) __PYX_ERR(0, 4273, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_64); __pyx_t_64 = 0;
  __pyx_t_64 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Equation_functionContext_11accept, 0, __pyx_n_s_ModelicaParser_Equation_function_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1901)); if (unlikely(!__pyx_t_64)) __PYX_ERR(0, 4273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_64);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_64, __pyx_t_65);
  __Pyx_DECREF(__pyx_t_65); __pyx_t_65 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_64) < 0) __PYX_ERR(0, 4273, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_64); __pyx_t_64 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4252
 * 
 * 
 *     class Equation_functionContext(Equation_optionsContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext
 */
  __pyx_t_64 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Equation_functionContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_64)) __PYX_ERR(0, 4252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_64);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_63, __pyx_t_64) < 0) __PYX_ERR(0, 4252, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_63); __pyx_t_63 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Equation_functionContext, __pyx_t_64) < 0) __PYX_ERR(0, 4252, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_64); __pyx_t_64 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4280
 * 
 * 
 *     class Equation_simpleContext(Equation_optionsContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext
 */
  __pyx_t_7 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_Equation_optionsContext);
  if (unlikely(!__pyx_t_7)) {
    PyErr_Clear();
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_Equation_optionsContext);
  }
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Equation_simpleContext, __pyx_n_s_ModelicaParser_Equation_simpleCo, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_64 = PyList_New(0); if (unlikely(!__pyx_t_64)) __PYX_ERR(0, 4280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_64);

  /* "pymola/generated/ModelicaParser.pyx":4282
 *     class Equation_simpleContext(Equation_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_65 = PyDict_New(); if (unlikely(!__pyx_t_65)) __PYX_ERR(0, 4282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_65);
  __pyx_t_66 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_66)) __PYX_ERR(0, 4282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_66);
  if (PyDict_SetItem(__pyx_t_65, __pyx_n_s_ctx_2, __pyx_t_66) < 0) __PYX_ERR(0, 4282, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_66); __pyx_t_66 = 0;
  __pyx_t_66 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_1__init__, 0, __pyx_n_s_ModelicaParser_Equation_simpleCo_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1903)); if (unlikely(!__pyx_t_66)) __PYX_ERR(0, 4282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_66);
  __Pyx_INCREF(__pyx_t_66);
  PyList_Append(__pyx_t_64, __pyx_t_66);
  __Pyx_GIVEREF(__pyx_t_66);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_66, __pyx_t_65);
  __Pyx_DECREF(__pyx_t_65); __pyx_t_65 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_66) < 0) __PYX_ERR(0, 4282, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_66); __pyx_t_66 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4286
 *             self.copyFrom(ctx)
 * 
 *         def simple_expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Simple_expressionContext,0)
 * 
 */
  __pyx_t_66 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_3simple_expression, 0, __pyx_n_s_ModelicaParser_Equation_simpleCo_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1905)); if (unlikely(!__pyx_t_66)) __PYX_ERR(0, 4286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_66);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_simple_expression, __pyx_t_66) < 0) __PYX_ERR(0, 4286, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_66); __pyx_t_66 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4289
 *             return self.getTypedRuleContext(ModelicaParser.Simple_expressionContext,0)
 * 
 *         def expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 */
  __pyx_t_66 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_5expression, 0, __pyx_n_s_ModelicaParser_Equation_simpleCo_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1907)); if (unlikely(!__pyx_t_66)) __PYX_ERR(0, 4289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_66);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_expression, __pyx_t_66) < 0) __PYX_ERR(0, 4289, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_66); __pyx_t_66 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4293
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterEquation_simple" ):
 *                 listener.enterEquation_simple(self)
 */
  __pyx_t_66 = PyDict_New(); if (unlikely(!__pyx_t_66)) __PYX_ERR(0, 4293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_66);
  __pyx_t_65 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_65)) __PYX_ERR(0, 4293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_65);
  if (PyDict_SetItem(__pyx_t_66, __pyx_n_s_listener, __pyx_t_65) < 0) __PYX_ERR(0, 4293, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_65); __pyx_t_65 = 0;
  __pyx_t_65 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_7enterRule, 0, __pyx_n_s_ModelicaParser_Equation_simpleCo_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1909)); if (unlikely(!__pyx_t_65)) __PYX_ERR(0, 4293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_65);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_65, __pyx_t_66);
  __Pyx_DECREF(__pyx_t_66); __pyx_t_66 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_65) < 0) __PYX_ERR(0, 4293, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_65); __pyx_t_65 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4297
 *                 listener.enterEquation_simple(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitEquation_simple" ):
 *                 listener.exitEquation_simple(self)
 */
  __pyx_t_65 = PyDict_New(); if (unlikely(!__pyx_t_65)) __PYX_ERR(0, 4297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_65);
  __pyx_t_66 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_66)) __PYX_ERR(0, 4297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_66);
  if (PyDict_SetItem(__pyx_t_65, __pyx_n_s_listener, __pyx_t_66) < 0) __PYX_ERR(0, 4297, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_66); __pyx_t_66 = 0;
  __pyx_t_66 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_9exitRule, 0, __pyx_n_s_ModelicaParser_Equation_simpleCo_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1911)); if (unlikely(!__pyx_t_66)) __PYX_ERR(0, 4297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_66);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_66, __pyx_t_65);
  __Pyx_DECREF(__pyx_t_65); __pyx_t_65 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_66) < 0) __PYX_ERR(0, 4297, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_66); __pyx_t_66 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4301
 *                 listener.exitEquation_simple(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitEquation_simple" ):
 *                 return visitor.visitEquation_simple(self)
 */
  __pyx_t_66 = PyDict_New(); if (unlikely(!__pyx_t_66)) __PYX_ERR(0, 4301, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_66);
  __pyx_t_65 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_65)) __PYX_ERR(0, 4301, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_65);
  if (PyDict_SetItem(__pyx_t_66, __pyx_n_s_visitor, __pyx_t_65) < 0) __PYX_ERR(0, 4301, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_65); __pyx_t_65 = 0;
  __pyx_t_65 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Equation_simpleContext_11accept, 0, __pyx_n_s_ModelicaParser_Equation_simpleCo_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1913)); if (unlikely(!__pyx_t_65)) __PYX_ERR(0, 4301, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_65);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_65, __pyx_t_66);
  __Pyx_DECREF(__pyx_t_66); __pyx_t_66 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_65) < 0) __PYX_ERR(0, 4301, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_65); __pyx_t_65 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4280
 * 
 * 
 *     class Equation_simpleContext(Equation_optionsContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext
 */
  __pyx_t_65 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Equation_simpleContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_65)) __PYX_ERR(0, 4280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_65);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_64, __pyx_t_65) < 0) __PYX_ERR(0, 4280, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_64); __pyx_t_64 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Equation_simpleContext, __pyx_t_65) < 0) __PYX_ERR(0, 4280, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_65); __pyx_t_65 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4308
 * 
 * 
 *     class Equation_ifContext(Equation_optionsContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext
 */
  __pyx_t_5 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_Equation_optionsContext);
  if (unlikely(!__pyx_t_5)) {
    PyErr_Clear();
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_Equation_optionsContext);
  }
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Equation_ifContext, __pyx_n_s_ModelicaParser_Equation_ifContex, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_65 = PyList_New(0); if (unlikely(!__pyx_t_65)) __PYX_ERR(0, 4308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_65);

  /* "pymola/generated/ModelicaParser.pyx":4310
 *     class Equation_ifContext(Equation_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_66 = PyDict_New(); if (unlikely(!__pyx_t_66)) __PYX_ERR(0, 4310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_66);
  __pyx_t_67 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_67)) __PYX_ERR(0, 4310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_67);
  if (PyDict_SetItem(__pyx_t_66, __pyx_n_s_ctx_2, __pyx_t_67) < 0) __PYX_ERR(0, 4310, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_67); __pyx_t_67 = 0;
  __pyx_t_67 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18Equation_ifContext_1__init__, 0, __pyx_n_s_ModelicaParser_Equation_ifContex_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1915)); if (unlikely(!__pyx_t_67)) __PYX_ERR(0, 4310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_67);
  __Pyx_INCREF(__pyx_t_67);
  PyList_Append(__pyx_t_65, __pyx_t_67);
  __Pyx_GIVEREF(__pyx_t_67);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_67, __pyx_t_66);
  __Pyx_DECREF(__pyx_t_66); __pyx_t_66 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_67) < 0) __PYX_ERR(0, 4310, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_67); __pyx_t_67 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4314
 *             self.copyFrom(ctx)
 * 
 *         def if_equation(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.If_equationContext,0)
 * 
 */
  __pyx_t_67 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18Equation_ifContext_3if_equation, 0, __pyx_n_s_ModelicaParser_Equation_ifContex_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1917)); if (unlikely(!__pyx_t_67)) __PYX_ERR(0, 4314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_67);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_if_equation, __pyx_t_67) < 0) __PYX_ERR(0, 4314, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_67); __pyx_t_67 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4318
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterEquation_if" ):
 *                 listener.enterEquation_if(self)
 */
  __pyx_t_67 = PyDict_New(); if (unlikely(!__pyx_t_67)) __PYX_ERR(0, 4318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_67);
  __pyx_t_66 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_66)) __PYX_ERR(0, 4318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_66);
  if (PyDict_SetItem(__pyx_t_67, __pyx_n_s_listener, __pyx_t_66) < 0) __PYX_ERR(0, 4318, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_66); __pyx_t_66 = 0;
  __pyx_t_66 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18Equation_ifContext_5enterRule, 0, __pyx_n_s_ModelicaParser_Equation_ifContex_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1919)); if (unlikely(!__pyx_t_66)) __PYX_ERR(0, 4318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_66);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_66, __pyx_t_67);
  __Pyx_DECREF(__pyx_t_67); __pyx_t_67 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_66) < 0) __PYX_ERR(0, 4318, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_66); __pyx_t_66 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4322
 *                 listener.enterEquation_if(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitEquation_if" ):
 *                 listener.exitEquation_if(self)
 */
  __pyx_t_66 = PyDict_New(); if (unlikely(!__pyx_t_66)) __PYX_ERR(0, 4322, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_66);
  __pyx_t_67 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_67)) __PYX_ERR(0, 4322, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_67);
  if (PyDict_SetItem(__pyx_t_66, __pyx_n_s_listener, __pyx_t_67) < 0) __PYX_ERR(0, 4322, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_67); __pyx_t_67 = 0;
  __pyx_t_67 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18Equation_ifContext_7exitRule, 0, __pyx_n_s_ModelicaParser_Equation_ifContex_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1921)); if (unlikely(!__pyx_t_67)) __PYX_ERR(0, 4322, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_67);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_67, __pyx_t_66);
  __Pyx_DECREF(__pyx_t_66); __pyx_t_66 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_67) < 0) __PYX_ERR(0, 4322, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_67); __pyx_t_67 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4326
 *                 listener.exitEquation_if(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitEquation_if" ):
 *                 return visitor.visitEquation_if(self)
 */
  __pyx_t_67 = PyDict_New(); if (unlikely(!__pyx_t_67)) __PYX_ERR(0, 4326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_67);
  __pyx_t_66 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_66)) __PYX_ERR(0, 4326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_66);
  if (PyDict_SetItem(__pyx_t_67, __pyx_n_s_visitor, __pyx_t_66) < 0) __PYX_ERR(0, 4326, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_66); __pyx_t_66 = 0;
  __pyx_t_66 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18Equation_ifContext_9accept, 0, __pyx_n_s_ModelicaParser_Equation_ifContex_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1923)); if (unlikely(!__pyx_t_66)) __PYX_ERR(0, 4326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_66);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_66, __pyx_t_67);
  __Pyx_DECREF(__pyx_t_67); __pyx_t_67 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_66) < 0) __PYX_ERR(0, 4326, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_66); __pyx_t_66 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4308
 * 
 * 
 *     class Equation_ifContext(Equation_optionsContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Equation_optionsContext
 */
  __pyx_t_66 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Equation_ifContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_66)) __PYX_ERR(0, 4308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_66);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_65, __pyx_t_66) < 0) __PYX_ERR(0, 4308, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_65); __pyx_t_65 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Equation_ifContext, __pyx_t_66) < 0) __PYX_ERR(0, 4308, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_66); __pyx_t_66 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4334
 * 
 * 
 *     def equation_options(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Equation_optionsContext(self, self._ctx, self.state)
 */
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_83equation_options, 0, __pyx_n_s_ModelicaParser_equation_options, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1925)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_equation_options, __pyx_t_7) < 0) __PYX_ERR(0, 4334, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4399
 *         return localctx
 * 
 *     class EquationContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_EquationContext, __pyx_n_s_ModelicaParser_EquationContext, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_66 = PyList_New(0); if (unlikely(!__pyx_t_66)) __PYX_ERR(0, 4399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_66);

  /* "pymola/generated/ModelicaParser.pyx":4401
 *     class EquationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_67 = PyDict_New(); if (unlikely(!__pyx_t_67)) __PYX_ERR(0, 4401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_67);
  __pyx_t_68 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_68)) __PYX_ERR(0, 4401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_68);
  if (PyDict_SetItem(__pyx_t_67, __pyx_n_s_parent, __pyx_t_68) < 0) __PYX_ERR(0, 4401, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_68); __pyx_t_68 = 0;
  if (PyDict_SetItem(__pyx_t_67, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 4401, __pyx_L1_error)
  __pyx_t_68 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15EquationContext_1__init__, 0, __pyx_n_s_ModelicaParser_EquationContext_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1927)); if (unlikely(!__pyx_t_68)) __PYX_ERR(0, 4401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_68);
  __Pyx_INCREF(__pyx_t_68);
  PyList_Append(__pyx_t_66, __pyx_t_68);
  __Pyx_GIVEREF(__pyx_t_68);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_68, __pyx_tuple__1928);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_68, __pyx_t_67);
  __Pyx_DECREF(__pyx_t_67); __pyx_t_67 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_68) < 0) __PYX_ERR(0, 4401, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_68); __pyx_t_68 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4405
 *             self.parser = parser
 * 
 *         def equation_options(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Equation_optionsContext,0)
 * 
 */
  __pyx_t_68 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15EquationContext_3equation_options, 0, __pyx_n_s_ModelicaParser_EquationContext_e, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1930)); if (unlikely(!__pyx_t_68)) __PYX_ERR(0, 4405, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_68);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_equation_options, __pyx_t_68) < 0) __PYX_ERR(0, 4405, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_68); __pyx_t_68 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4409
 * 
 * 
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */
  __pyx_t_68 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15EquationContext_5comment, 0, __pyx_n_s_ModelicaParser_EquationContext_c, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1932)); if (unlikely(!__pyx_t_68)) __PYX_ERR(0, 4409, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_68);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_comment, __pyx_t_68) < 0) __PYX_ERR(0, 4409, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_68); __pyx_t_68 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4413
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_equation
 * 
 */
  __pyx_t_68 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15EquationContext_7getRuleIndex, 0, __pyx_n_s_ModelicaParser_EquationContext_g, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1934)); if (unlikely(!__pyx_t_68)) __PYX_ERR(0, 4413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_68);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_68) < 0) __PYX_ERR(0, 4413, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_68); __pyx_t_68 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4416
 *             return ModelicaParser.RULE_equation
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterEquation" ):
 *                 listener.enterEquation(self)
 */
  __pyx_t_68 = PyDict_New(); if (unlikely(!__pyx_t_68)) __PYX_ERR(0, 4416, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_68);
  __pyx_t_67 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_67)) __PYX_ERR(0, 4416, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_67);
  if (PyDict_SetItem(__pyx_t_68, __pyx_n_s_listener, __pyx_t_67) < 0) __PYX_ERR(0, 4416, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_67); __pyx_t_67 = 0;
  __pyx_t_67 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15EquationContext_9enterRule, 0, __pyx_n_s_ModelicaParser_EquationContext_e_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1936)); if (unlikely(!__pyx_t_67)) __PYX_ERR(0, 4416, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_67);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_67, __pyx_t_68);
  __Pyx_DECREF(__pyx_t_68); __pyx_t_68 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_67) < 0) __PYX_ERR(0, 4416, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_67); __pyx_t_67 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4420
 *                 listener.enterEquation(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitEquation" ):
 *                 listener.exitEquation(self)
 */
  __pyx_t_67 = PyDict_New(); if (unlikely(!__pyx_t_67)) __PYX_ERR(0, 4420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_67);
  __pyx_t_68 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_68)) __PYX_ERR(0, 4420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_68);
  if (PyDict_SetItem(__pyx_t_67, __pyx_n_s_listener, __pyx_t_68) < 0) __PYX_ERR(0, 4420, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_68); __pyx_t_68 = 0;
  __pyx_t_68 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15EquationContext_11exitRule, 0, __pyx_n_s_ModelicaParser_EquationContext_e_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1938)); if (unlikely(!__pyx_t_68)) __PYX_ERR(0, 4420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_68);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_68, __pyx_t_67);
  __Pyx_DECREF(__pyx_t_67); __pyx_t_67 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_68) < 0) __PYX_ERR(0, 4420, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_68); __pyx_t_68 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4424
 *                 listener.exitEquation(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitEquation" ):
 *                 return visitor.visitEquation(self)
 */
  __pyx_t_68 = PyDict_New(); if (unlikely(!__pyx_t_68)) __PYX_ERR(0, 4424, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_68);
  __pyx_t_67 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_67)) __PYX_ERR(0, 4424, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_67);
  if (PyDict_SetItem(__pyx_t_68, __pyx_n_s_visitor, __pyx_t_67) < 0) __PYX_ERR(0, 4424, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_67); __pyx_t_67 = 0;
  __pyx_t_67 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15EquationContext_13accept, 0, __pyx_n_s_ModelicaParser_EquationContext_a, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1940)); if (unlikely(!__pyx_t_67)) __PYX_ERR(0, 4424, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_67);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_67, __pyx_t_68);
  __Pyx_DECREF(__pyx_t_68); __pyx_t_68 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_67) < 0) __PYX_ERR(0, 4424, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_67); __pyx_t_67 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4399
 *         return localctx
 * 
 *     class EquationContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_67 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_EquationContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_67)) __PYX_ERR(0, 4399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_67);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_66, __pyx_t_67) < 0) __PYX_ERR(0, 4399, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_66); __pyx_t_66 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_EquationContext, __pyx_t_67) < 0) __PYX_ERR(0, 4399, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_67); __pyx_t_67 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4433
 * 
 * 
 *     def equation(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.EquationContext(self, self._ctx, self.state)
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_85equation, 0, __pyx_n_s_ModelicaParser_equation, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1942)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_equation, __pyx_t_5) < 0) __PYX_ERR(0, 4433, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4451
 *         return localctx
 * 
 *     class Statement_optionsContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Statement_optionsContext, __pyx_n_s_ModelicaParser_Statement_options, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_67 = PyList_New(0); if (unlikely(!__pyx_t_67)) __PYX_ERR(0, 4451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_67);

  /* "pymola/generated/ModelicaParser.pyx":4453
 *     class Statement_optionsContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_68 = PyDict_New(); if (unlikely(!__pyx_t_68)) __PYX_ERR(0, 4453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_68);
  __pyx_t_69 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_69)) __PYX_ERR(0, 4453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_69);
  if (PyDict_SetItem(__pyx_t_68, __pyx_n_s_parent, __pyx_t_69) < 0) __PYX_ERR(0, 4453, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_69); __pyx_t_69 = 0;
  if (PyDict_SetItem(__pyx_t_68, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 4453, __pyx_L1_error)
  __pyx_t_69 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Statement_optionsContext_1__init__, 0, __pyx_n_s_ModelicaParser_Statement_options_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1944)); if (unlikely(!__pyx_t_69)) __PYX_ERR(0, 4453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_69);
  __Pyx_INCREF(__pyx_t_69);
  PyList_Append(__pyx_t_67, __pyx_t_69);
  __Pyx_GIVEREF(__pyx_t_69);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_69, __pyx_tuple__1945);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_69, __pyx_t_68);
  __Pyx_DECREF(__pyx_t_68); __pyx_t_68 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_69) < 0) __PYX_ERR(0, 4453, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_69); __pyx_t_69 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4458
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_statement_options
 * 
 */
  __pyx_t_69 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Statement_optionsContext_3getRuleIndex, 0, __pyx_n_s_ModelicaParser_Statement_options_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1947)); if (unlikely(!__pyx_t_69)) __PYX_ERR(0, 4458, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_69);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_69) < 0) __PYX_ERR(0, 4458, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_69); __pyx_t_69 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4462
 * 
 * 
 *         def copyFrom(self, ctx:ParserRuleContext):             # <<<<<<<<<<<<<<
 *             super().copyFrom(ctx)
 * 
 */
  __pyx_t_69 = PyDict_New(); if (unlikely(!__pyx_t_69)) __PYX_ERR(0, 4462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_69);
  __pyx_t_68 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_68)) __PYX_ERR(0, 4462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_68);
  if (PyDict_SetItem(__pyx_t_69, __pyx_n_s_ctx_2, __pyx_t_68) < 0) __PYX_ERR(0, 4462, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_68); __pyx_t_68 = 0;
  __pyx_t_68 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Statement_optionsContext_5copyFrom, 0, __pyx_n_s_ModelicaParser_Statement_options_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1949)); if (unlikely(!__pyx_t_68)) __PYX_ERR(0, 4462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_68);
  __Pyx_INCREF(__pyx_t_68);
  PyList_Append(__pyx_t_67, __pyx_t_68);
  __Pyx_GIVEREF(__pyx_t_68);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_68, __pyx_t_69);
  __Pyx_DECREF(__pyx_t_69); __pyx_t_69 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_copyFrom, __pyx_t_68) < 0) __PYX_ERR(0, 4462, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_68); __pyx_t_68 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4451
 *         return localctx
 * 
 *     class Statement_optionsContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_68 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Statement_optionsContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_68)) __PYX_ERR(0, 4451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_68);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_67, __pyx_t_68) < 0) __PYX_ERR(0, 4451, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_67); __pyx_t_67 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Statement_optionsContext, __pyx_t_68) < 0) __PYX_ERR(0, 4451, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_68); __pyx_t_68 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4467
 * 
 * 
 *     class Statement_breakContext(Statement_optionsContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext
 */
  __pyx_t_7 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_Statement_optionsContext);
  if (unlikely(!__pyx_t_7)) {
    PyErr_Clear();
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_Statement_optionsContext);
  }
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4467, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4467, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4467, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Statement_breakContext, __pyx_n_s_ModelicaParser_Statement_breakCo, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4467, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_68 = PyList_New(0); if (unlikely(!__pyx_t_68)) __PYX_ERR(0, 4467, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_68);

  /* "pymola/generated/ModelicaParser.pyx":4469
 *     class Statement_breakContext(Statement_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_69 = PyDict_New(); if (unlikely(!__pyx_t_69)) __PYX_ERR(0, 4469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_69);
  __pyx_t_70 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_70)) __PYX_ERR(0, 4469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_70);
  if (PyDict_SetItem(__pyx_t_69, __pyx_n_s_ctx_2, __pyx_t_70) < 0) __PYX_ERR(0, 4469, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_70); __pyx_t_70 = 0;
  __pyx_t_70 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Statement_breakContext_1__init__, 0, __pyx_n_s_ModelicaParser_Statement_breakCo_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1951)); if (unlikely(!__pyx_t_70)) __PYX_ERR(0, 4469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_70);
  __Pyx_INCREF(__pyx_t_70);
  PyList_Append(__pyx_t_68, __pyx_t_70);
  __Pyx_GIVEREF(__pyx_t_70);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_70, __pyx_t_69);
  __Pyx_DECREF(__pyx_t_69); __pyx_t_69 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_70) < 0) __PYX_ERR(0, 4469, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_70); __pyx_t_70 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4474
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStatement_break" ):
 *                 listener.enterStatement_break(self)
 */
  __pyx_t_70 = PyDict_New(); if (unlikely(!__pyx_t_70)) __PYX_ERR(0, 4474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_70);
  __pyx_t_69 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_69)) __PYX_ERR(0, 4474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_69);
  if (PyDict_SetItem(__pyx_t_70, __pyx_n_s_listener, __pyx_t_69) < 0) __PYX_ERR(0, 4474, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_69); __pyx_t_69 = 0;
  __pyx_t_69 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Statement_breakContext_3enterRule, 0, __pyx_n_s_ModelicaParser_Statement_breakCo_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1953)); if (unlikely(!__pyx_t_69)) __PYX_ERR(0, 4474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_69);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_69, __pyx_t_70);
  __Pyx_DECREF(__pyx_t_70); __pyx_t_70 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_69) < 0) __PYX_ERR(0, 4474, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_69); __pyx_t_69 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4478
 *                 listener.enterStatement_break(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStatement_break" ):
 *                 listener.exitStatement_break(self)
 */
  __pyx_t_69 = PyDict_New(); if (unlikely(!__pyx_t_69)) __PYX_ERR(0, 4478, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_69);
  __pyx_t_70 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_70)) __PYX_ERR(0, 4478, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_70);
  if (PyDict_SetItem(__pyx_t_69, __pyx_n_s_listener, __pyx_t_70) < 0) __PYX_ERR(0, 4478, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_70); __pyx_t_70 = 0;
  __pyx_t_70 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Statement_breakContext_5exitRule, 0, __pyx_n_s_ModelicaParser_Statement_breakCo_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1955)); if (unlikely(!__pyx_t_70)) __PYX_ERR(0, 4478, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_70);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_70, __pyx_t_69);
  __Pyx_DECREF(__pyx_t_69); __pyx_t_69 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_70) < 0) __PYX_ERR(0, 4478, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_70); __pyx_t_70 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4482
 *                 listener.exitStatement_break(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStatement_break" ):
 *                 return visitor.visitStatement_break(self)
 */
  __pyx_t_70 = PyDict_New(); if (unlikely(!__pyx_t_70)) __PYX_ERR(0, 4482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_70);
  __pyx_t_69 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_69)) __PYX_ERR(0, 4482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_69);
  if (PyDict_SetItem(__pyx_t_70, __pyx_n_s_visitor, __pyx_t_69) < 0) __PYX_ERR(0, 4482, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_69); __pyx_t_69 = 0;
  __pyx_t_69 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Statement_breakContext_7accept, 0, __pyx_n_s_ModelicaParser_Statement_breakCo_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1957)); if (unlikely(!__pyx_t_69)) __PYX_ERR(0, 4482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_69);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_69, __pyx_t_70);
  __Pyx_DECREF(__pyx_t_70); __pyx_t_70 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_69) < 0) __PYX_ERR(0, 4482, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_69); __pyx_t_69 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4467
 * 
 * 
 *     class Statement_breakContext(Statement_optionsContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext
 */
  __pyx_t_69 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Statement_breakContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_69)) __PYX_ERR(0, 4467, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_69);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_68, __pyx_t_69) < 0) __PYX_ERR(0, 4467, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_68); __pyx_t_68 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Statement_breakContext, __pyx_t_69) < 0) __PYX_ERR(0, 4467, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_69); __pyx_t_69 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4489
 * 
 * 
 *     class Statement_whileContext(Statement_optionsContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext
 */
  __pyx_t_5 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_Statement_optionsContext);
  if (unlikely(!__pyx_t_5)) {
    PyErr_Clear();
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_Statement_optionsContext);
  }
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Statement_whileContext, __pyx_n_s_ModelicaParser_Statement_whileCo, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_69 = PyList_New(0); if (unlikely(!__pyx_t_69)) __PYX_ERR(0, 4489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_69);

  /* "pymola/generated/ModelicaParser.pyx":4491
 *     class Statement_whileContext(Statement_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_70 = PyDict_New(); if (unlikely(!__pyx_t_70)) __PYX_ERR(0, 4491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_70);
  __pyx_t_71 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_71)) __PYX_ERR(0, 4491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_71);
  if (PyDict_SetItem(__pyx_t_70, __pyx_n_s_ctx_2, __pyx_t_71) < 0) __PYX_ERR(0, 4491, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_71); __pyx_t_71 = 0;
  __pyx_t_71 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Statement_whileContext_1__init__, 0, __pyx_n_s_ModelicaParser_Statement_whileCo_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1959)); if (unlikely(!__pyx_t_71)) __PYX_ERR(0, 4491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_71);
  __Pyx_INCREF(__pyx_t_71);
  PyList_Append(__pyx_t_69, __pyx_t_71);
  __Pyx_GIVEREF(__pyx_t_71);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_71, __pyx_t_70);
  __Pyx_DECREF(__pyx_t_70); __pyx_t_70 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_71) < 0) __PYX_ERR(0, 4491, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_71); __pyx_t_71 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4495
 *             self.copyFrom(ctx)
 * 
 *         def while_statement(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.While_statementContext,0)
 * 
 */
  __pyx_t_71 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Statement_whileContext_3while_statement, 0, __pyx_n_s_ModelicaParser_Statement_whileCo_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1961)); if (unlikely(!__pyx_t_71)) __PYX_ERR(0, 4495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_71);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_while_statement, __pyx_t_71) < 0) __PYX_ERR(0, 4495, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_71); __pyx_t_71 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4499
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStatement_while" ):
 *                 listener.enterStatement_while(self)
 */
  __pyx_t_71 = PyDict_New(); if (unlikely(!__pyx_t_71)) __PYX_ERR(0, 4499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_71);
  __pyx_t_70 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_70)) __PYX_ERR(0, 4499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_70);
  if (PyDict_SetItem(__pyx_t_71, __pyx_n_s_listener, __pyx_t_70) < 0) __PYX_ERR(0, 4499, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_70); __pyx_t_70 = 0;
  __pyx_t_70 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Statement_whileContext_5enterRule, 0, __pyx_n_s_ModelicaParser_Statement_whileCo_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1963)); if (unlikely(!__pyx_t_70)) __PYX_ERR(0, 4499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_70);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_70, __pyx_t_71);
  __Pyx_DECREF(__pyx_t_71); __pyx_t_71 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_70) < 0) __PYX_ERR(0, 4499, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_70); __pyx_t_70 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4503
 *                 listener.enterStatement_while(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStatement_while" ):
 *                 listener.exitStatement_while(self)
 */
  __pyx_t_70 = PyDict_New(); if (unlikely(!__pyx_t_70)) __PYX_ERR(0, 4503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_70);
  __pyx_t_71 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_71)) __PYX_ERR(0, 4503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_71);
  if (PyDict_SetItem(__pyx_t_70, __pyx_n_s_listener, __pyx_t_71) < 0) __PYX_ERR(0, 4503, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_71); __pyx_t_71 = 0;
  __pyx_t_71 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Statement_whileContext_7exitRule, 0, __pyx_n_s_ModelicaParser_Statement_whileCo_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1965)); if (unlikely(!__pyx_t_71)) __PYX_ERR(0, 4503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_71);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_71, __pyx_t_70);
  __Pyx_DECREF(__pyx_t_70); __pyx_t_70 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_71) < 0) __PYX_ERR(0, 4503, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_71); __pyx_t_71 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4507
 *                 listener.exitStatement_while(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStatement_while" ):
 *                 return visitor.visitStatement_while(self)
 */
  __pyx_t_71 = PyDict_New(); if (unlikely(!__pyx_t_71)) __PYX_ERR(0, 4507, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_71);
  __pyx_t_70 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_70)) __PYX_ERR(0, 4507, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_70);
  if (PyDict_SetItem(__pyx_t_71, __pyx_n_s_visitor, __pyx_t_70) < 0) __PYX_ERR(0, 4507, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_70); __pyx_t_70 = 0;
  __pyx_t_70 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Statement_whileContext_9accept, 0, __pyx_n_s_ModelicaParser_Statement_whileCo_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1967)); if (unlikely(!__pyx_t_70)) __PYX_ERR(0, 4507, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_70);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_70, __pyx_t_71);
  __Pyx_DECREF(__pyx_t_71); __pyx_t_71 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_70) < 0) __PYX_ERR(0, 4507, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_70); __pyx_t_70 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4489
 * 
 * 
 *     class Statement_whileContext(Statement_optionsContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext
 */
  __pyx_t_70 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Statement_whileContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_70)) __PYX_ERR(0, 4489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_70);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_69, __pyx_t_70) < 0) __PYX_ERR(0, 4489, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_69); __pyx_t_69 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Statement_whileContext, __pyx_t_70) < 0) __PYX_ERR(0, 4489, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_70); __pyx_t_70 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4514
 * 
 * 
 *     class Statement_component_functionContext(Statement_optionsContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext
 */
  __pyx_t_7 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_Statement_optionsContext);
  if (unlikely(!__pyx_t_7)) {
    PyErr_Clear();
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_Statement_optionsContext);
  }
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4514, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4514, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4514, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Statement_component_functionCont, __pyx_n_s_ModelicaParser_Statement_compone, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4514, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_70 = PyList_New(0); if (unlikely(!__pyx_t_70)) __PYX_ERR(0, 4514, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_70);

  /* "pymola/generated/ModelicaParser.pyx":4516
 *     class Statement_component_functionContext(Statement_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_71 = PyDict_New(); if (unlikely(!__pyx_t_71)) __PYX_ERR(0, 4516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_71);
  __pyx_t_72 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_72)) __PYX_ERR(0, 4516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_72);
  if (PyDict_SetItem(__pyx_t_71, __pyx_n_s_ctx_2, __pyx_t_72) < 0) __PYX_ERR(0, 4516, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_72); __pyx_t_72 = 0;
  __pyx_t_72 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_1__init__, 0, __pyx_n_s_ModelicaParser_Statement_compone_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1969)); if (unlikely(!__pyx_t_72)) __PYX_ERR(0, 4516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_72);
  __Pyx_INCREF(__pyx_t_72);
  PyList_Append(__pyx_t_70, __pyx_t_72);
  __Pyx_GIVEREF(__pyx_t_72);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_72, __pyx_t_71);
  __Pyx_DECREF(__pyx_t_71); __pyx_t_71 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_72) < 0) __PYX_ERR(0, 4516, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_72); __pyx_t_72 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4520
 *             self.copyFrom(ctx)
 * 
 *         def component_reference(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Component_referenceContext)
 */
  __pyx_t_72 = PyDict_New(); if (unlikely(!__pyx_t_72)) __PYX_ERR(0, 4520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_72);
  if (PyDict_SetItem(__pyx_t_72, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 4520, __pyx_L1_error)
  __pyx_t_71 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_3component_reference, 0, __pyx_n_s_ModelicaParser_Statement_compone_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1971)); if (unlikely(!__pyx_t_71)) __PYX_ERR(0, 4520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_71);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_71, __pyx_tuple__1972);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_71, __pyx_t_72);
  __Pyx_DECREF(__pyx_t_72); __pyx_t_72 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_component_reference, __pyx_t_71) < 0) __PYX_ERR(0, 4520, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_71); __pyx_t_71 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4526
 *                 return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,i)
 * 
 *         def function_call_args(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Function_call_argsContext,0)
 * 
 */
  __pyx_t_71 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_5function_call_args, 0, __pyx_n_s_ModelicaParser_Statement_compone_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1974)); if (unlikely(!__pyx_t_71)) __PYX_ERR(0, 4526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_71);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_function_call_args, __pyx_t_71) < 0) __PYX_ERR(0, 4526, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_71); __pyx_t_71 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4530
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStatement_component_function" ):
 *                 listener.enterStatement_component_function(self)
 */
  __pyx_t_71 = PyDict_New(); if (unlikely(!__pyx_t_71)) __PYX_ERR(0, 4530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_71);
  __pyx_t_72 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_72)) __PYX_ERR(0, 4530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_72);
  if (PyDict_SetItem(__pyx_t_71, __pyx_n_s_listener, __pyx_t_72) < 0) __PYX_ERR(0, 4530, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_72); __pyx_t_72 = 0;
  __pyx_t_72 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_7enterRule, 0, __pyx_n_s_ModelicaParser_Statement_compone_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1976)); if (unlikely(!__pyx_t_72)) __PYX_ERR(0, 4530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_72);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_72, __pyx_t_71);
  __Pyx_DECREF(__pyx_t_71); __pyx_t_71 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_72) < 0) __PYX_ERR(0, 4530, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_72); __pyx_t_72 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4534
 *                 listener.enterStatement_component_function(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStatement_component_function" ):
 *                 listener.exitStatement_component_function(self)
 */
  __pyx_t_72 = PyDict_New(); if (unlikely(!__pyx_t_72)) __PYX_ERR(0, 4534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_72);
  __pyx_t_71 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_71)) __PYX_ERR(0, 4534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_71);
  if (PyDict_SetItem(__pyx_t_72, __pyx_n_s_listener, __pyx_t_71) < 0) __PYX_ERR(0, 4534, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_71); __pyx_t_71 = 0;
  __pyx_t_71 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_9exitRule, 0, __pyx_n_s_ModelicaParser_Statement_compone_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1978)); if (unlikely(!__pyx_t_71)) __PYX_ERR(0, 4534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_71);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_71, __pyx_t_72);
  __Pyx_DECREF(__pyx_t_72); __pyx_t_72 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_71) < 0) __PYX_ERR(0, 4534, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_71); __pyx_t_71 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4538
 *                 listener.exitStatement_component_function(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStatement_component_function" ):
 *                 return visitor.visitStatement_component_function(self)
 */
  __pyx_t_71 = PyDict_New(); if (unlikely(!__pyx_t_71)) __PYX_ERR(0, 4538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_71);
  __pyx_t_72 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_72)) __PYX_ERR(0, 4538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_72);
  if (PyDict_SetItem(__pyx_t_71, __pyx_n_s_visitor, __pyx_t_72) < 0) __PYX_ERR(0, 4538, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_72); __pyx_t_72 = 0;
  __pyx_t_72 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_35Statement_component_functionContext_11accept, 0, __pyx_n_s_ModelicaParser_Statement_compone_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1980)); if (unlikely(!__pyx_t_72)) __PYX_ERR(0, 4538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_72);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_72, __pyx_t_71);
  __Pyx_DECREF(__pyx_t_71); __pyx_t_71 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_72) < 0) __PYX_ERR(0, 4538, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_72); __pyx_t_72 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4514
 * 
 * 
 *     class Statement_component_functionContext(Statement_optionsContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext
 */
  __pyx_t_72 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Statement_component_functionCont, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_72)) __PYX_ERR(0, 4514, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_72);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_70, __pyx_t_72) < 0) __PYX_ERR(0, 4514, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_70); __pyx_t_70 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Statement_component_functionCont, __pyx_t_72) < 0) __PYX_ERR(0, 4514, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_72); __pyx_t_72 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4545
 * 
 * 
 *     class Statement_returnContext(Statement_optionsContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext
 */
  __pyx_t_5 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_Statement_optionsContext);
  if (unlikely(!__pyx_t_5)) {
    PyErr_Clear();
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_Statement_optionsContext);
  }
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Statement_returnContext, __pyx_n_s_ModelicaParser_Statement_returnC, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_72 = PyList_New(0); if (unlikely(!__pyx_t_72)) __PYX_ERR(0, 4545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_72);

  /* "pymola/generated/ModelicaParser.pyx":4547
 *     class Statement_returnContext(Statement_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_71 = PyDict_New(); if (unlikely(!__pyx_t_71)) __PYX_ERR(0, 4547, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_71);
  __pyx_t_73 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_73)) __PYX_ERR(0, 4547, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_73);
  if (PyDict_SetItem(__pyx_t_71, __pyx_n_s_ctx_2, __pyx_t_73) < 0) __PYX_ERR(0, 4547, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_73); __pyx_t_73 = 0;
  __pyx_t_73 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Statement_returnContext_1__init__, 0, __pyx_n_s_ModelicaParser_Statement_returnC_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1982)); if (unlikely(!__pyx_t_73)) __PYX_ERR(0, 4547, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_73);
  __Pyx_INCREF(__pyx_t_73);
  PyList_Append(__pyx_t_72, __pyx_t_73);
  __Pyx_GIVEREF(__pyx_t_73);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_73, __pyx_t_71);
  __Pyx_DECREF(__pyx_t_71); __pyx_t_71 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_73) < 0) __PYX_ERR(0, 4547, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_73); __pyx_t_73 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4552
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStatement_return" ):
 *                 listener.enterStatement_return(self)
 */
  __pyx_t_73 = PyDict_New(); if (unlikely(!__pyx_t_73)) __PYX_ERR(0, 4552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_73);
  __pyx_t_71 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_71)) __PYX_ERR(0, 4552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_71);
  if (PyDict_SetItem(__pyx_t_73, __pyx_n_s_listener, __pyx_t_71) < 0) __PYX_ERR(0, 4552, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_71); __pyx_t_71 = 0;
  __pyx_t_71 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Statement_returnContext_3enterRule, 0, __pyx_n_s_ModelicaParser_Statement_returnC_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1984)); if (unlikely(!__pyx_t_71)) __PYX_ERR(0, 4552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_71);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_71, __pyx_t_73);
  __Pyx_DECREF(__pyx_t_73); __pyx_t_73 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_71) < 0) __PYX_ERR(0, 4552, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_71); __pyx_t_71 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4556
 *                 listener.enterStatement_return(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStatement_return" ):
 *                 listener.exitStatement_return(self)
 */
  __pyx_t_71 = PyDict_New(); if (unlikely(!__pyx_t_71)) __PYX_ERR(0, 4556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_71);
  __pyx_t_73 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_73)) __PYX_ERR(0, 4556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_73);
  if (PyDict_SetItem(__pyx_t_71, __pyx_n_s_listener, __pyx_t_73) < 0) __PYX_ERR(0, 4556, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_73); __pyx_t_73 = 0;
  __pyx_t_73 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Statement_returnContext_5exitRule, 0, __pyx_n_s_ModelicaParser_Statement_returnC_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1986)); if (unlikely(!__pyx_t_73)) __PYX_ERR(0, 4556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_73);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_73, __pyx_t_71);
  __Pyx_DECREF(__pyx_t_71); __pyx_t_71 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_73) < 0) __PYX_ERR(0, 4556, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_73); __pyx_t_73 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4560
 *                 listener.exitStatement_return(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStatement_return" ):
 *                 return visitor.visitStatement_return(self)
 */
  __pyx_t_73 = PyDict_New(); if (unlikely(!__pyx_t_73)) __PYX_ERR(0, 4560, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_73);
  __pyx_t_71 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_71)) __PYX_ERR(0, 4560, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_71);
  if (PyDict_SetItem(__pyx_t_73, __pyx_n_s_visitor, __pyx_t_71) < 0) __PYX_ERR(0, 4560, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_71); __pyx_t_71 = 0;
  __pyx_t_71 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Statement_returnContext_7accept, 0, __pyx_n_s_ModelicaParser_Statement_returnC_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1988)); if (unlikely(!__pyx_t_71)) __PYX_ERR(0, 4560, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_71);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_71, __pyx_t_73);
  __Pyx_DECREF(__pyx_t_73); __pyx_t_73 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_71) < 0) __PYX_ERR(0, 4560, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_71); __pyx_t_71 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4545
 * 
 * 
 *     class Statement_returnContext(Statement_optionsContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext
 */
  __pyx_t_71 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Statement_returnContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_71)) __PYX_ERR(0, 4545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_71);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_72, __pyx_t_71) < 0) __PYX_ERR(0, 4545, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_72); __pyx_t_72 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Statement_returnContext, __pyx_t_71) < 0) __PYX_ERR(0, 4545, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_71); __pyx_t_71 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4567
 * 
 * 
 *     class Statement_forContext(Statement_optionsContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext
 */
  __pyx_t_7 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_Statement_optionsContext);
  if (unlikely(!__pyx_t_7)) {
    PyErr_Clear();
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_Statement_optionsContext);
  }
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Statement_forContext, __pyx_n_s_ModelicaParser_Statement_forCont, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_71 = PyList_New(0); if (unlikely(!__pyx_t_71)) __PYX_ERR(0, 4567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_71);

  /* "pymola/generated/ModelicaParser.pyx":4569
 *     class Statement_forContext(Statement_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_73 = PyDict_New(); if (unlikely(!__pyx_t_73)) __PYX_ERR(0, 4569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_73);
  __pyx_t_74 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_74)) __PYX_ERR(0, 4569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_74);
  if (PyDict_SetItem(__pyx_t_73, __pyx_n_s_ctx_2, __pyx_t_74) < 0) __PYX_ERR(0, 4569, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_74); __pyx_t_74 = 0;
  __pyx_t_74 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20Statement_forContext_1__init__, 0, __pyx_n_s_ModelicaParser_Statement_forCont_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1990)); if (unlikely(!__pyx_t_74)) __PYX_ERR(0, 4569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_74);
  __Pyx_INCREF(__pyx_t_74);
  PyList_Append(__pyx_t_71, __pyx_t_74);
  __Pyx_GIVEREF(__pyx_t_74);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_74, __pyx_t_73);
  __Pyx_DECREF(__pyx_t_73); __pyx_t_73 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_74) < 0) __PYX_ERR(0, 4569, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_74); __pyx_t_74 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4573
 *             self.copyFrom(ctx)
 * 
 *         def for_statement(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.For_statementContext,0)
 * 
 */
  __pyx_t_74 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20Statement_forContext_3for_statement, 0, __pyx_n_s_ModelicaParser_Statement_forCont_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1992)); if (unlikely(!__pyx_t_74)) __PYX_ERR(0, 4573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_74);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_for_statement, __pyx_t_74) < 0) __PYX_ERR(0, 4573, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_74); __pyx_t_74 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4577
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStatement_for" ):
 *                 listener.enterStatement_for(self)
 */
  __pyx_t_74 = PyDict_New(); if (unlikely(!__pyx_t_74)) __PYX_ERR(0, 4577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_74);
  __pyx_t_73 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_73)) __PYX_ERR(0, 4577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_73);
  if (PyDict_SetItem(__pyx_t_74, __pyx_n_s_listener, __pyx_t_73) < 0) __PYX_ERR(0, 4577, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_73); __pyx_t_73 = 0;
  __pyx_t_73 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20Statement_forContext_5enterRule, 0, __pyx_n_s_ModelicaParser_Statement_forCont_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1994)); if (unlikely(!__pyx_t_73)) __PYX_ERR(0, 4577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_73);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_73, __pyx_t_74);
  __Pyx_DECREF(__pyx_t_74); __pyx_t_74 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_73) < 0) __PYX_ERR(0, 4577, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_73); __pyx_t_73 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4581
 *                 listener.enterStatement_for(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStatement_for" ):
 *                 listener.exitStatement_for(self)
 */
  __pyx_t_73 = PyDict_New(); if (unlikely(!__pyx_t_73)) __PYX_ERR(0, 4581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_73);
  __pyx_t_74 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_74)) __PYX_ERR(0, 4581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_74);
  if (PyDict_SetItem(__pyx_t_73, __pyx_n_s_listener, __pyx_t_74) < 0) __PYX_ERR(0, 4581, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_74); __pyx_t_74 = 0;
  __pyx_t_74 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20Statement_forContext_7exitRule, 0, __pyx_n_s_ModelicaParser_Statement_forCont_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1996)); if (unlikely(!__pyx_t_74)) __PYX_ERR(0, 4581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_74);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_74, __pyx_t_73);
  __Pyx_DECREF(__pyx_t_73); __pyx_t_73 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_74) < 0) __PYX_ERR(0, 4581, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_74); __pyx_t_74 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4585
 *                 listener.exitStatement_for(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStatement_for" ):
 *                 return visitor.visitStatement_for(self)
 */
  __pyx_t_74 = PyDict_New(); if (unlikely(!__pyx_t_74)) __PYX_ERR(0, 4585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_74);
  __pyx_t_73 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_73)) __PYX_ERR(0, 4585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_73);
  if (PyDict_SetItem(__pyx_t_74, __pyx_n_s_visitor, __pyx_t_73) < 0) __PYX_ERR(0, 4585, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_73); __pyx_t_73 = 0;
  __pyx_t_73 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20Statement_forContext_9accept, 0, __pyx_n_s_ModelicaParser_Statement_forCont_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__1998)); if (unlikely(!__pyx_t_73)) __PYX_ERR(0, 4585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_73);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_73, __pyx_t_74);
  __Pyx_DECREF(__pyx_t_74); __pyx_t_74 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_73) < 0) __PYX_ERR(0, 4585, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_73); __pyx_t_73 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4567
 * 
 * 
 *     class Statement_forContext(Statement_optionsContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext
 */
  __pyx_t_73 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Statement_forContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_73)) __PYX_ERR(0, 4567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_73);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_71, __pyx_t_73) < 0) __PYX_ERR(0, 4567, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_71); __pyx_t_71 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Statement_forContext, __pyx_t_73) < 0) __PYX_ERR(0, 4567, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_73); __pyx_t_73 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4592
 * 
 * 
 *     class Statement_whenContext(Statement_optionsContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext
 */
  __pyx_t_5 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_Statement_optionsContext);
  if (unlikely(!__pyx_t_5)) {
    PyErr_Clear();
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_Statement_optionsContext);
  }
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4592, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4592, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4592, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Statement_whenContext, __pyx_n_s_ModelicaParser_Statement_whenCon, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4592, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_73 = PyList_New(0); if (unlikely(!__pyx_t_73)) __PYX_ERR(0, 4592, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_73);

  /* "pymola/generated/ModelicaParser.pyx":4594
 *     class Statement_whenContext(Statement_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_74 = PyDict_New(); if (unlikely(!__pyx_t_74)) __PYX_ERR(0, 4594, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_74);
  __pyx_t_75 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_75)) __PYX_ERR(0, 4594, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_75);
  if (PyDict_SetItem(__pyx_t_74, __pyx_n_s_ctx_2, __pyx_t_75) < 0) __PYX_ERR(0, 4594, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_75); __pyx_t_75 = 0;
  __pyx_t_75 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Statement_whenContext_1__init__, 0, __pyx_n_s_ModelicaParser_Statement_whenCon_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2000)); if (unlikely(!__pyx_t_75)) __PYX_ERR(0, 4594, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_75);
  __Pyx_INCREF(__pyx_t_75);
  PyList_Append(__pyx_t_73, __pyx_t_75);
  __Pyx_GIVEREF(__pyx_t_75);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_75, __pyx_t_74);
  __Pyx_DECREF(__pyx_t_74); __pyx_t_74 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_75) < 0) __PYX_ERR(0, 4594, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_75); __pyx_t_75 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4598
 *             self.copyFrom(ctx)
 * 
 *         def when_statement(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.When_statementContext,0)
 * 
 */
  __pyx_t_75 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Statement_whenContext_3when_statement, 0, __pyx_n_s_ModelicaParser_Statement_whenCon_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2002)); if (unlikely(!__pyx_t_75)) __PYX_ERR(0, 4598, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_75);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_when_statement, __pyx_t_75) < 0) __PYX_ERR(0, 4598, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_75); __pyx_t_75 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4602
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStatement_when" ):
 *                 listener.enterStatement_when(self)
 */
  __pyx_t_75 = PyDict_New(); if (unlikely(!__pyx_t_75)) __PYX_ERR(0, 4602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_75);
  __pyx_t_74 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_74)) __PYX_ERR(0, 4602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_74);
  if (PyDict_SetItem(__pyx_t_75, __pyx_n_s_listener, __pyx_t_74) < 0) __PYX_ERR(0, 4602, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_74); __pyx_t_74 = 0;
  __pyx_t_74 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Statement_whenContext_5enterRule, 0, __pyx_n_s_ModelicaParser_Statement_whenCon_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2004)); if (unlikely(!__pyx_t_74)) __PYX_ERR(0, 4602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_74);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_74, __pyx_t_75);
  __Pyx_DECREF(__pyx_t_75); __pyx_t_75 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_74) < 0) __PYX_ERR(0, 4602, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_74); __pyx_t_74 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4606
 *                 listener.enterStatement_when(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStatement_when" ):
 *                 listener.exitStatement_when(self)
 */
  __pyx_t_74 = PyDict_New(); if (unlikely(!__pyx_t_74)) __PYX_ERR(0, 4606, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_74);
  __pyx_t_75 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_75)) __PYX_ERR(0, 4606, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_75);
  if (PyDict_SetItem(__pyx_t_74, __pyx_n_s_listener, __pyx_t_75) < 0) __PYX_ERR(0, 4606, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_75); __pyx_t_75 = 0;
  __pyx_t_75 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Statement_whenContext_7exitRule, 0, __pyx_n_s_ModelicaParser_Statement_whenCon_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2006)); if (unlikely(!__pyx_t_75)) __PYX_ERR(0, 4606, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_75);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_75, __pyx_t_74);
  __Pyx_DECREF(__pyx_t_74); __pyx_t_74 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_75) < 0) __PYX_ERR(0, 4606, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_75); __pyx_t_75 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4610
 *                 listener.exitStatement_when(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStatement_when" ):
 *                 return visitor.visitStatement_when(self)
 */
  __pyx_t_75 = PyDict_New(); if (unlikely(!__pyx_t_75)) __PYX_ERR(0, 4610, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_75);
  __pyx_t_74 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_74)) __PYX_ERR(0, 4610, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_74);
  if (PyDict_SetItem(__pyx_t_75, __pyx_n_s_visitor, __pyx_t_74) < 0) __PYX_ERR(0, 4610, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_74); __pyx_t_74 = 0;
  __pyx_t_74 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Statement_whenContext_9accept, 0, __pyx_n_s_ModelicaParser_Statement_whenCon_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2008)); if (unlikely(!__pyx_t_74)) __PYX_ERR(0, 4610, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_74);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_74, __pyx_t_75);
  __Pyx_DECREF(__pyx_t_75); __pyx_t_75 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_74) < 0) __PYX_ERR(0, 4610, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_74); __pyx_t_74 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4592
 * 
 * 
 *     class Statement_whenContext(Statement_optionsContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext
 */
  __pyx_t_74 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Statement_whenContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_74)) __PYX_ERR(0, 4592, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_74);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_73, __pyx_t_74) < 0) __PYX_ERR(0, 4592, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_73); __pyx_t_73 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Statement_whenContext, __pyx_t_74) < 0) __PYX_ERR(0, 4592, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_74); __pyx_t_74 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4617
 * 
 * 
 *     class Statement_ifContext(Statement_optionsContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext
 */
  __pyx_t_7 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_Statement_optionsContext);
  if (unlikely(!__pyx_t_7)) {
    PyErr_Clear();
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_Statement_optionsContext);
  }
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Statement_ifContext, __pyx_n_s_ModelicaParser_Statement_ifConte, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_74 = PyList_New(0); if (unlikely(!__pyx_t_74)) __PYX_ERR(0, 4617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_74);

  /* "pymola/generated/ModelicaParser.pyx":4619
 *     class Statement_ifContext(Statement_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_75 = PyDict_New(); if (unlikely(!__pyx_t_75)) __PYX_ERR(0, 4619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_75);
  __pyx_t_76 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_76)) __PYX_ERR(0, 4619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_76);
  if (PyDict_SetItem(__pyx_t_75, __pyx_n_s_ctx_2, __pyx_t_76) < 0) __PYX_ERR(0, 4619, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_76); __pyx_t_76 = 0;
  __pyx_t_76 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19Statement_ifContext_1__init__, 0, __pyx_n_s_ModelicaParser_Statement_ifConte_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2010)); if (unlikely(!__pyx_t_76)) __PYX_ERR(0, 4619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_76);
  __Pyx_INCREF(__pyx_t_76);
  PyList_Append(__pyx_t_74, __pyx_t_76);
  __Pyx_GIVEREF(__pyx_t_76);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_76, __pyx_t_75);
  __Pyx_DECREF(__pyx_t_75); __pyx_t_75 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_76) < 0) __PYX_ERR(0, 4619, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_76); __pyx_t_76 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4623
 *             self.copyFrom(ctx)
 * 
 *         def if_statement(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.If_statementContext,0)
 * 
 */
  __pyx_t_76 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19Statement_ifContext_3if_statement, 0, __pyx_n_s_ModelicaParser_Statement_ifConte_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2012)); if (unlikely(!__pyx_t_76)) __PYX_ERR(0, 4623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_76);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_if_statement, __pyx_t_76) < 0) __PYX_ERR(0, 4623, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_76); __pyx_t_76 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4627
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStatement_if" ):
 *                 listener.enterStatement_if(self)
 */
  __pyx_t_76 = PyDict_New(); if (unlikely(!__pyx_t_76)) __PYX_ERR(0, 4627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_76);
  __pyx_t_75 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_75)) __PYX_ERR(0, 4627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_75);
  if (PyDict_SetItem(__pyx_t_76, __pyx_n_s_listener, __pyx_t_75) < 0) __PYX_ERR(0, 4627, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_75); __pyx_t_75 = 0;
  __pyx_t_75 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19Statement_ifContext_5enterRule, 0, __pyx_n_s_ModelicaParser_Statement_ifConte_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2014)); if (unlikely(!__pyx_t_75)) __PYX_ERR(0, 4627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_75);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_75, __pyx_t_76);
  __Pyx_DECREF(__pyx_t_76); __pyx_t_76 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_75) < 0) __PYX_ERR(0, 4627, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_75); __pyx_t_75 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4631
 *                 listener.enterStatement_if(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStatement_if" ):
 *                 listener.exitStatement_if(self)
 */
  __pyx_t_75 = PyDict_New(); if (unlikely(!__pyx_t_75)) __PYX_ERR(0, 4631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_75);
  __pyx_t_76 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_76)) __PYX_ERR(0, 4631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_76);
  if (PyDict_SetItem(__pyx_t_75, __pyx_n_s_listener, __pyx_t_76) < 0) __PYX_ERR(0, 4631, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_76); __pyx_t_76 = 0;
  __pyx_t_76 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19Statement_ifContext_7exitRule, 0, __pyx_n_s_ModelicaParser_Statement_ifConte_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2016)); if (unlikely(!__pyx_t_76)) __PYX_ERR(0, 4631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_76);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_76, __pyx_t_75);
  __Pyx_DECREF(__pyx_t_75); __pyx_t_75 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_76) < 0) __PYX_ERR(0, 4631, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_76); __pyx_t_76 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4635
 *                 listener.exitStatement_if(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStatement_if" ):
 *                 return visitor.visitStatement_if(self)
 */
  __pyx_t_76 = PyDict_New(); if (unlikely(!__pyx_t_76)) __PYX_ERR(0, 4635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_76);
  __pyx_t_75 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_75)) __PYX_ERR(0, 4635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_75);
  if (PyDict_SetItem(__pyx_t_76, __pyx_n_s_visitor, __pyx_t_75) < 0) __PYX_ERR(0, 4635, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_75); __pyx_t_75 = 0;
  __pyx_t_75 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19Statement_ifContext_9accept, 0, __pyx_n_s_ModelicaParser_Statement_ifConte_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2018)); if (unlikely(!__pyx_t_75)) __PYX_ERR(0, 4635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_75);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_75, __pyx_t_76);
  __Pyx_DECREF(__pyx_t_76); __pyx_t_76 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_75) < 0) __PYX_ERR(0, 4635, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_75); __pyx_t_75 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4617
 * 
 * 
 *     class Statement_ifContext(Statement_optionsContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext
 */
  __pyx_t_75 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Statement_ifContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_75)) __PYX_ERR(0, 4617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_75);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_74, __pyx_t_75) < 0) __PYX_ERR(0, 4617, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_74); __pyx_t_74 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Statement_ifContext, __pyx_t_75) < 0) __PYX_ERR(0, 4617, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_75); __pyx_t_75 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4642
 * 
 * 
 *     class Statement_component_referenceContext(Statement_optionsContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext
 */
  __pyx_t_5 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_Statement_optionsContext);
  if (unlikely(!__pyx_t_5)) {
    PyErr_Clear();
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_Statement_optionsContext);
  }
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Statement_component_referenceCon, __pyx_n_s_ModelicaParser_Statement_compone_8, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_75 = PyList_New(0); if (unlikely(!__pyx_t_75)) __PYX_ERR(0, 4642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_75);

  /* "pymola/generated/ModelicaParser.pyx":4644
 *     class Statement_component_referenceContext(Statement_optionsContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_76 = PyDict_New(); if (unlikely(!__pyx_t_76)) __PYX_ERR(0, 4644, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_76);
  __pyx_t_77 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_77)) __PYX_ERR(0, 4644, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_77);
  if (PyDict_SetItem(__pyx_t_76, __pyx_n_s_ctx_2, __pyx_t_77) < 0) __PYX_ERR(0, 4644, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_77); __pyx_t_77 = 0;
  __pyx_t_77 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_1__init__, 0, __pyx_n_s_ModelicaParser_Statement_compone_9, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2020)); if (unlikely(!__pyx_t_77)) __PYX_ERR(0, 4644, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_77);
  __Pyx_INCREF(__pyx_t_77);
  PyList_Append(__pyx_t_75, __pyx_t_77);
  __Pyx_GIVEREF(__pyx_t_77);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_77, __pyx_t_76);
  __Pyx_DECREF(__pyx_t_76); __pyx_t_76 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_77) < 0) __PYX_ERR(0, 4644, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_77); __pyx_t_77 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4648
 *             self.copyFrom(ctx)
 * 
 *         def component_reference(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 */
  __pyx_t_77 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_3component_reference, 0, __pyx_n_s_ModelicaParser_Statement_compone_10, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2022)); if (unlikely(!__pyx_t_77)) __PYX_ERR(0, 4648, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_77);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_component_reference, __pyx_t_77) < 0) __PYX_ERR(0, 4648, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_77); __pyx_t_77 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4651
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 *         def expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 */
  __pyx_t_77 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_5expression, 0, __pyx_n_s_ModelicaParser_Statement_compone_11, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2024)); if (unlikely(!__pyx_t_77)) __PYX_ERR(0, 4651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_77);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_expression, __pyx_t_77) < 0) __PYX_ERR(0, 4651, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_77); __pyx_t_77 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4654
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 *         def function_call_args(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Function_call_argsContext,0)
 * 
 */
  __pyx_t_77 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_7function_call_args, 0, __pyx_n_s_ModelicaParser_Statement_compone_12, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2026)); if (unlikely(!__pyx_t_77)) __PYX_ERR(0, 4654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_77);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_function_call_args, __pyx_t_77) < 0) __PYX_ERR(0, 4654, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_77); __pyx_t_77 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4658
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStatement_component_reference" ):
 *                 listener.enterStatement_component_reference(self)
 */
  __pyx_t_77 = PyDict_New(); if (unlikely(!__pyx_t_77)) __PYX_ERR(0, 4658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_77);
  __pyx_t_76 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_76)) __PYX_ERR(0, 4658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_76);
  if (PyDict_SetItem(__pyx_t_77, __pyx_n_s_listener, __pyx_t_76) < 0) __PYX_ERR(0, 4658, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_76); __pyx_t_76 = 0;
  __pyx_t_76 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_9enterRule, 0, __pyx_n_s_ModelicaParser_Statement_compone_13, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2028)); if (unlikely(!__pyx_t_76)) __PYX_ERR(0, 4658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_76);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_76, __pyx_t_77);
  __Pyx_DECREF(__pyx_t_77); __pyx_t_77 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_76) < 0) __PYX_ERR(0, 4658, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_76); __pyx_t_76 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4662
 *                 listener.enterStatement_component_reference(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStatement_component_reference" ):
 *                 listener.exitStatement_component_reference(self)
 */
  __pyx_t_76 = PyDict_New(); if (unlikely(!__pyx_t_76)) __PYX_ERR(0, 4662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_76);
  __pyx_t_77 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_77)) __PYX_ERR(0, 4662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_77);
  if (PyDict_SetItem(__pyx_t_76, __pyx_n_s_listener, __pyx_t_77) < 0) __PYX_ERR(0, 4662, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_77); __pyx_t_77 = 0;
  __pyx_t_77 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_11exitRule, 0, __pyx_n_s_ModelicaParser_Statement_compone_14, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2030)); if (unlikely(!__pyx_t_77)) __PYX_ERR(0, 4662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_77);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_77, __pyx_t_76);
  __Pyx_DECREF(__pyx_t_76); __pyx_t_76 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_77) < 0) __PYX_ERR(0, 4662, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_77); __pyx_t_77 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4666
 *                 listener.exitStatement_component_reference(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStatement_component_reference" ):
 *                 return visitor.visitStatement_component_reference(self)
 */
  __pyx_t_77 = PyDict_New(); if (unlikely(!__pyx_t_77)) __PYX_ERR(0, 4666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_77);
  __pyx_t_76 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_76)) __PYX_ERR(0, 4666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_76);
  if (PyDict_SetItem(__pyx_t_77, __pyx_n_s_visitor, __pyx_t_76) < 0) __PYX_ERR(0, 4666, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_76); __pyx_t_76 = 0;
  __pyx_t_76 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_36Statement_component_referenceContext_13accept, 0, __pyx_n_s_ModelicaParser_Statement_compone_15, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2032)); if (unlikely(!__pyx_t_76)) __PYX_ERR(0, 4666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_76);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_76, __pyx_t_77);
  __Pyx_DECREF(__pyx_t_77); __pyx_t_77 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_76) < 0) __PYX_ERR(0, 4666, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_76); __pyx_t_76 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4642
 * 
 * 
 *     class Statement_component_referenceContext(Statement_optionsContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Statement_optionsContext
 */
  __pyx_t_76 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Statement_component_referenceCon, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_76)) __PYX_ERR(0, 4642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_76);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_75, __pyx_t_76) < 0) __PYX_ERR(0, 4642, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_75); __pyx_t_75 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Statement_component_referenceCon, __pyx_t_76) < 0) __PYX_ERR(0, 4642, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_76); __pyx_t_76 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4674
 * 
 * 
 *     def statement_options(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Statement_optionsContext(self, self._ctx, self.state)
 */
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_87statement_options, 0, __pyx_n_s_ModelicaParser_statement_options, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2034)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_statement_options, __pyx_t_7) < 0) __PYX_ERR(0, 4674, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4780
 *         return localctx
 * 
 *     class StatementContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_StatementContext, __pyx_n_s_ModelicaParser_StatementContext, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_76 = PyList_New(0); if (unlikely(!__pyx_t_76)) __PYX_ERR(0, 4780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_76);

  /* "pymola/generated/ModelicaParser.pyx":4782
 *     class StatementContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_77 = PyDict_New(); if (unlikely(!__pyx_t_77)) __PYX_ERR(0, 4782, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_77);
  __pyx_t_78 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_78)) __PYX_ERR(0, 4782, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_78);
  if (PyDict_SetItem(__pyx_t_77, __pyx_n_s_parent, __pyx_t_78) < 0) __PYX_ERR(0, 4782, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_78); __pyx_t_78 = 0;
  if (PyDict_SetItem(__pyx_t_77, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 4782, __pyx_L1_error)
  __pyx_t_78 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_16StatementContext_1__init__, 0, __pyx_n_s_ModelicaParser_StatementContext_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2036)); if (unlikely(!__pyx_t_78)) __PYX_ERR(0, 4782, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_78);
  __Pyx_INCREF(__pyx_t_78);
  PyList_Append(__pyx_t_76, __pyx_t_78);
  __Pyx_GIVEREF(__pyx_t_78);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_78, __pyx_tuple__2037);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_78, __pyx_t_77);
  __Pyx_DECREF(__pyx_t_77); __pyx_t_77 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_78) < 0) __PYX_ERR(0, 4782, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_78); __pyx_t_78 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4786
 *             self.parser = parser
 * 
 *         def statement_options(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Statement_optionsContext,0)
 * 
 */
  __pyx_t_78 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_16StatementContext_3statement_options, 0, __pyx_n_s_ModelicaParser_StatementContext_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2039)); if (unlikely(!__pyx_t_78)) __PYX_ERR(0, 4786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_78);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_statement_options, __pyx_t_78) < 0) __PYX_ERR(0, 4786, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_78); __pyx_t_78 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4790
 * 
 * 
 *         def comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.CommentContext,0)
 * 
 */
  __pyx_t_78 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_16StatementContext_5comment, 0, __pyx_n_s_ModelicaParser_StatementContext_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2041)); if (unlikely(!__pyx_t_78)) __PYX_ERR(0, 4790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_78);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_comment, __pyx_t_78) < 0) __PYX_ERR(0, 4790, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_78); __pyx_t_78 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4794
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_statement
 * 
 */
  __pyx_t_78 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_16StatementContext_7getRuleIndex, 0, __pyx_n_s_ModelicaParser_StatementContext_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2043)); if (unlikely(!__pyx_t_78)) __PYX_ERR(0, 4794, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_78);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_78) < 0) __PYX_ERR(0, 4794, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_78); __pyx_t_78 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4797
 *             return ModelicaParser.RULE_statement
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterStatement" ):
 *                 listener.enterStatement(self)
 */
  __pyx_t_78 = PyDict_New(); if (unlikely(!__pyx_t_78)) __PYX_ERR(0, 4797, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_78);
  __pyx_t_77 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_77)) __PYX_ERR(0, 4797, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_77);
  if (PyDict_SetItem(__pyx_t_78, __pyx_n_s_listener, __pyx_t_77) < 0) __PYX_ERR(0, 4797, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_77); __pyx_t_77 = 0;
  __pyx_t_77 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_16StatementContext_9enterRule, 0, __pyx_n_s_ModelicaParser_StatementContext_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2045)); if (unlikely(!__pyx_t_77)) __PYX_ERR(0, 4797, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_77);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_77, __pyx_t_78);
  __Pyx_DECREF(__pyx_t_78); __pyx_t_78 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_77) < 0) __PYX_ERR(0, 4797, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_77); __pyx_t_77 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4801
 *                 listener.enterStatement(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitStatement" ):
 *                 listener.exitStatement(self)
 */
  __pyx_t_77 = PyDict_New(); if (unlikely(!__pyx_t_77)) __PYX_ERR(0, 4801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_77);
  __pyx_t_78 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_78)) __PYX_ERR(0, 4801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_78);
  if (PyDict_SetItem(__pyx_t_77, __pyx_n_s_listener, __pyx_t_78) < 0) __PYX_ERR(0, 4801, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_78); __pyx_t_78 = 0;
  __pyx_t_78 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_16StatementContext_11exitRule, 0, __pyx_n_s_ModelicaParser_StatementContext_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2047)); if (unlikely(!__pyx_t_78)) __PYX_ERR(0, 4801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_78);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_78, __pyx_t_77);
  __Pyx_DECREF(__pyx_t_77); __pyx_t_77 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_78) < 0) __PYX_ERR(0, 4801, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_78); __pyx_t_78 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4805
 *                 listener.exitStatement(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitStatement" ):
 *                 return visitor.visitStatement(self)
 */
  __pyx_t_78 = PyDict_New(); if (unlikely(!__pyx_t_78)) __PYX_ERR(0, 4805, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_78);
  __pyx_t_77 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_77)) __PYX_ERR(0, 4805, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_77);
  if (PyDict_SetItem(__pyx_t_78, __pyx_n_s_visitor, __pyx_t_77) < 0) __PYX_ERR(0, 4805, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_77); __pyx_t_77 = 0;
  __pyx_t_77 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_16StatementContext_13accept, 0, __pyx_n_s_ModelicaParser_StatementContext_8, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2049)); if (unlikely(!__pyx_t_77)) __PYX_ERR(0, 4805, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_77);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_77, __pyx_t_78);
  __Pyx_DECREF(__pyx_t_78); __pyx_t_78 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_77) < 0) __PYX_ERR(0, 4805, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_77); __pyx_t_77 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4780
 *         return localctx
 * 
 *     class StatementContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_77 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_StatementContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_77)) __PYX_ERR(0, 4780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_77);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_76, __pyx_t_77) < 0) __PYX_ERR(0, 4780, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_76); __pyx_t_76 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_StatementContext, __pyx_t_77) < 0) __PYX_ERR(0, 4780, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_77); __pyx_t_77 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4814
 * 
 * 
 *     def statement(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.StatementContext(self, self._ctx, self.state)
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_89statement, 0, __pyx_n_s_ModelicaParser_statement, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2051)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4814, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_statement, __pyx_t_5) < 0) __PYX_ERR(0, 4814, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4832
 *         return localctx
 * 
 *     class If_equationContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_If_equationContext, __pyx_n_s_ModelicaParser_If_equationContex, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_77 = PyList_New(0); if (unlikely(!__pyx_t_77)) __PYX_ERR(0, 4832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_77);

  /* "pymola/generated/ModelicaParser.pyx":4834
 *     class If_equationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_78 = PyDict_New(); if (unlikely(!__pyx_t_78)) __PYX_ERR(0, 4834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_78);
  __pyx_t_79 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_79)) __PYX_ERR(0, 4834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_79);
  if (PyDict_SetItem(__pyx_t_78, __pyx_n_s_parent, __pyx_t_79) < 0) __PYX_ERR(0, 4834, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_79); __pyx_t_79 = 0;
  if (PyDict_SetItem(__pyx_t_78, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 4834, __pyx_L1_error)
  __pyx_t_79 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18If_equationContext_1__init__, 0, __pyx_n_s_ModelicaParser_If_equationContex_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2053)); if (unlikely(!__pyx_t_79)) __PYX_ERR(0, 4834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_79);
  __Pyx_INCREF(__pyx_t_79);
  PyList_Append(__pyx_t_77, __pyx_t_79);
  __Pyx_GIVEREF(__pyx_t_79);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_79, __pyx_tuple__2054);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_79, __pyx_t_78);
  __Pyx_DECREF(__pyx_t_78); __pyx_t_78 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_79) < 0) __PYX_ERR(0, 4834, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_79); __pyx_t_79 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4838
 *             self.parser = parser
 * 
 *         def expression(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 */
  __pyx_t_79 = PyDict_New(); if (unlikely(!__pyx_t_79)) __PYX_ERR(0, 4838, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_79);
  if (PyDict_SetItem(__pyx_t_79, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 4838, __pyx_L1_error)
  __pyx_t_78 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18If_equationContext_3expression, 0, __pyx_n_s_ModelicaParser_If_equationContex_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2056)); if (unlikely(!__pyx_t_78)) __PYX_ERR(0, 4838, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_78);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_78, __pyx_tuple__2057);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_78, __pyx_t_79);
  __Pyx_DECREF(__pyx_t_79); __pyx_t_79 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_expression, __pyx_t_78) < 0) __PYX_ERR(0, 4838, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_78); __pyx_t_78 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4845
 * 
 * 
 *         def equation(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.EquationContext)
 */
  __pyx_t_78 = PyDict_New(); if (unlikely(!__pyx_t_78)) __PYX_ERR(0, 4845, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_78);
  if (PyDict_SetItem(__pyx_t_78, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 4845, __pyx_L1_error)
  __pyx_t_79 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18If_equationContext_5equation, 0, __pyx_n_s_ModelicaParser_If_equationContex_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2059)); if (unlikely(!__pyx_t_79)) __PYX_ERR(0, 4845, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_79);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_79, __pyx_tuple__2060);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_79, __pyx_t_78);
  __Pyx_DECREF(__pyx_t_78); __pyx_t_78 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_equation, __pyx_t_79) < 0) __PYX_ERR(0, 4845, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_79); __pyx_t_79 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4852
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_if_equation
 * 
 */
  __pyx_t_79 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18If_equationContext_7getRuleIndex, 0, __pyx_n_s_ModelicaParser_If_equationContex_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2062)); if (unlikely(!__pyx_t_79)) __PYX_ERR(0, 4852, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_79);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_79) < 0) __PYX_ERR(0, 4852, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_79); __pyx_t_79 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4855
 *             return ModelicaParser.RULE_if_equation
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterIf_equation" ):
 *                 listener.enterIf_equation(self)
 */
  __pyx_t_79 = PyDict_New(); if (unlikely(!__pyx_t_79)) __PYX_ERR(0, 4855, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_79);
  __pyx_t_78 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_78)) __PYX_ERR(0, 4855, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_78);
  if (PyDict_SetItem(__pyx_t_79, __pyx_n_s_listener, __pyx_t_78) < 0) __PYX_ERR(0, 4855, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_78); __pyx_t_78 = 0;
  __pyx_t_78 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18If_equationContext_9enterRule, 0, __pyx_n_s_ModelicaParser_If_equationContex_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2064)); if (unlikely(!__pyx_t_78)) __PYX_ERR(0, 4855, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_78);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_78, __pyx_t_79);
  __Pyx_DECREF(__pyx_t_79); __pyx_t_79 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_78) < 0) __PYX_ERR(0, 4855, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_78); __pyx_t_78 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4859
 *                 listener.enterIf_equation(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitIf_equation" ):
 *                 listener.exitIf_equation(self)
 */
  __pyx_t_78 = PyDict_New(); if (unlikely(!__pyx_t_78)) __PYX_ERR(0, 4859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_78);
  __pyx_t_79 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_79)) __PYX_ERR(0, 4859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_79);
  if (PyDict_SetItem(__pyx_t_78, __pyx_n_s_listener, __pyx_t_79) < 0) __PYX_ERR(0, 4859, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_79); __pyx_t_79 = 0;
  __pyx_t_79 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18If_equationContext_11exitRule, 0, __pyx_n_s_ModelicaParser_If_equationContex_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2066)); if (unlikely(!__pyx_t_79)) __PYX_ERR(0, 4859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_79);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_79, __pyx_t_78);
  __Pyx_DECREF(__pyx_t_78); __pyx_t_78 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_79) < 0) __PYX_ERR(0, 4859, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_79); __pyx_t_79 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4863
 *                 listener.exitIf_equation(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitIf_equation" ):
 *                 return visitor.visitIf_equation(self)
 */
  __pyx_t_79 = PyDict_New(); if (unlikely(!__pyx_t_79)) __PYX_ERR(0, 4863, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_79);
  __pyx_t_78 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_78)) __PYX_ERR(0, 4863, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_78);
  if (PyDict_SetItem(__pyx_t_79, __pyx_n_s_visitor, __pyx_t_78) < 0) __PYX_ERR(0, 4863, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_78); __pyx_t_78 = 0;
  __pyx_t_78 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18If_equationContext_13accept, 0, __pyx_n_s_ModelicaParser_If_equationContex_8, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2068)); if (unlikely(!__pyx_t_78)) __PYX_ERR(0, 4863, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_78);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_78, __pyx_t_79);
  __Pyx_DECREF(__pyx_t_79); __pyx_t_79 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_78) < 0) __PYX_ERR(0, 4863, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_78); __pyx_t_78 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4832
 *         return localctx
 * 
 *     class If_equationContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_78 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_If_equationContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_78)) __PYX_ERR(0, 4832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_78);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_77, __pyx_t_78) < 0) __PYX_ERR(0, 4832, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_77); __pyx_t_77 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_If_equationContext, __pyx_t_78) < 0) __PYX_ERR(0, 4832, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_78); __pyx_t_78 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4872
 * 
 * 
 *     def if_equation(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.If_equationContext(self, self._ctx, self.state)
 */
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_91if_equation, 0, __pyx_n_s_ModelicaParser_if_equation, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2070)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4872, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_if_equation, __pyx_t_7) < 0) __PYX_ERR(0, 4872, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4958
 *         return localctx
 * 
 *     class If_statementContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4958, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4958, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4958, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_If_statementContext, __pyx_n_s_ModelicaParser_If_statementConte, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4958, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_78 = PyList_New(0); if (unlikely(!__pyx_t_78)) __PYX_ERR(0, 4958, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_78);

  /* "pymola/generated/ModelicaParser.pyx":4960
 *     class If_statementContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_79 = PyDict_New(); if (unlikely(!__pyx_t_79)) __PYX_ERR(0, 4960, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_79);
  __pyx_t_80 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_80)) __PYX_ERR(0, 4960, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_80);
  if (PyDict_SetItem(__pyx_t_79, __pyx_n_s_parent, __pyx_t_80) < 0) __PYX_ERR(0, 4960, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_80); __pyx_t_80 = 0;
  if (PyDict_SetItem(__pyx_t_79, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 4960, __pyx_L1_error)
  __pyx_t_80 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19If_statementContext_1__init__, 0, __pyx_n_s_ModelicaParser_If_statementConte_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2072)); if (unlikely(!__pyx_t_80)) __PYX_ERR(0, 4960, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_80);
  __Pyx_INCREF(__pyx_t_80);
  PyList_Append(__pyx_t_78, __pyx_t_80);
  __Pyx_GIVEREF(__pyx_t_80);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_80, __pyx_tuple__2073);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_80, __pyx_t_79);
  __Pyx_DECREF(__pyx_t_79); __pyx_t_79 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_80) < 0) __PYX_ERR(0, 4960, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_80); __pyx_t_80 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4964
 *             self.parser = parser
 * 
 *         def expression(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 */
  __pyx_t_80 = PyDict_New(); if (unlikely(!__pyx_t_80)) __PYX_ERR(0, 4964, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_80);
  if (PyDict_SetItem(__pyx_t_80, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 4964, __pyx_L1_error)
  __pyx_t_79 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19If_statementContext_3expression, 0, __pyx_n_s_ModelicaParser_If_statementConte_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2075)); if (unlikely(!__pyx_t_79)) __PYX_ERR(0, 4964, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_79);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_79, __pyx_tuple__2076);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_79, __pyx_t_80);
  __Pyx_DECREF(__pyx_t_80); __pyx_t_80 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_expression, __pyx_t_79) < 0) __PYX_ERR(0, 4964, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_79); __pyx_t_79 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4971
 * 
 * 
 *         def statement(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.StatementContext)
 */
  __pyx_t_79 = PyDict_New(); if (unlikely(!__pyx_t_79)) __PYX_ERR(0, 4971, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_79);
  if (PyDict_SetItem(__pyx_t_79, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 4971, __pyx_L1_error)
  __pyx_t_80 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19If_statementContext_5statement, 0, __pyx_n_s_ModelicaParser_If_statementConte_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2078)); if (unlikely(!__pyx_t_80)) __PYX_ERR(0, 4971, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_80);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_80, __pyx_tuple__2079);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_80, __pyx_t_79);
  __Pyx_DECREF(__pyx_t_79); __pyx_t_79 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_statement, __pyx_t_80) < 0) __PYX_ERR(0, 4971, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_80); __pyx_t_80 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4978
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_if_statement
 * 
 */
  __pyx_t_80 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19If_statementContext_7getRuleIndex, 0, __pyx_n_s_ModelicaParser_If_statementConte_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2081)); if (unlikely(!__pyx_t_80)) __PYX_ERR(0, 4978, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_80);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_80) < 0) __PYX_ERR(0, 4978, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_80); __pyx_t_80 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4981
 *             return ModelicaParser.RULE_if_statement
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterIf_statement" ):
 *                 listener.enterIf_statement(self)
 */
  __pyx_t_80 = PyDict_New(); if (unlikely(!__pyx_t_80)) __PYX_ERR(0, 4981, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_80);
  __pyx_t_79 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_79)) __PYX_ERR(0, 4981, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_79);
  if (PyDict_SetItem(__pyx_t_80, __pyx_n_s_listener, __pyx_t_79) < 0) __PYX_ERR(0, 4981, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_79); __pyx_t_79 = 0;
  __pyx_t_79 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19If_statementContext_9enterRule, 0, __pyx_n_s_ModelicaParser_If_statementConte_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2083)); if (unlikely(!__pyx_t_79)) __PYX_ERR(0, 4981, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_79);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_79, __pyx_t_80);
  __Pyx_DECREF(__pyx_t_80); __pyx_t_80 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_79) < 0) __PYX_ERR(0, 4981, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_79); __pyx_t_79 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4985
 *                 listener.enterIf_statement(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitIf_statement" ):
 *                 listener.exitIf_statement(self)
 */
  __pyx_t_79 = PyDict_New(); if (unlikely(!__pyx_t_79)) __PYX_ERR(0, 4985, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_79);
  __pyx_t_80 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_80)) __PYX_ERR(0, 4985, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_80);
  if (PyDict_SetItem(__pyx_t_79, __pyx_n_s_listener, __pyx_t_80) < 0) __PYX_ERR(0, 4985, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_80); __pyx_t_80 = 0;
  __pyx_t_80 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19If_statementContext_11exitRule, 0, __pyx_n_s_ModelicaParser_If_statementConte_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2085)); if (unlikely(!__pyx_t_80)) __PYX_ERR(0, 4985, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_80);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_80, __pyx_t_79);
  __Pyx_DECREF(__pyx_t_79); __pyx_t_79 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_80) < 0) __PYX_ERR(0, 4985, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_80); __pyx_t_80 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4989
 *                 listener.exitIf_statement(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitIf_statement" ):
 *                 return visitor.visitIf_statement(self)
 */
  __pyx_t_80 = PyDict_New(); if (unlikely(!__pyx_t_80)) __PYX_ERR(0, 4989, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_80);
  __pyx_t_79 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_79)) __PYX_ERR(0, 4989, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_79);
  if (PyDict_SetItem(__pyx_t_80, __pyx_n_s_visitor, __pyx_t_79) < 0) __PYX_ERR(0, 4989, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_79); __pyx_t_79 = 0;
  __pyx_t_79 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19If_statementContext_13accept, 0, __pyx_n_s_ModelicaParser_If_statementConte_8, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2087)); if (unlikely(!__pyx_t_79)) __PYX_ERR(0, 4989, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_79);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_79, __pyx_t_80);
  __Pyx_DECREF(__pyx_t_80); __pyx_t_80 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_79) < 0) __PYX_ERR(0, 4989, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_79); __pyx_t_79 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4958
 *         return localctx
 * 
 *     class If_statementContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_79 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_If_statementContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_79)) __PYX_ERR(0, 4958, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_79);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_78, __pyx_t_79) < 0) __PYX_ERR(0, 4958, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_78); __pyx_t_78 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_If_statementContext, __pyx_t_79) < 0) __PYX_ERR(0, 4958, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_79); __pyx_t_79 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":4998
 * 
 * 
 *     def if_statement(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.If_statementContext(self, self._ctx, self.state)
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_93if_statement, 0, __pyx_n_s_ModelicaParser_if_statement, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2089)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4998, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_if_statement, __pyx_t_5) < 0) __PYX_ERR(0, 4998, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5081
 *         return localctx
 * 
 *     class For_equationContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5081, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5081, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5081, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_For_equationContext, __pyx_n_s_ModelicaParser_For_equationConte, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5081, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_79 = PyList_New(0); if (unlikely(!__pyx_t_79)) __PYX_ERR(0, 5081, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_79);

  /* "pymola/generated/ModelicaParser.pyx":5083
 *     class For_equationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_80 = PyDict_New(); if (unlikely(!__pyx_t_80)) __PYX_ERR(0, 5083, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_80);
  __pyx_t_81 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_81)) __PYX_ERR(0, 5083, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_81);
  if (PyDict_SetItem(__pyx_t_80, __pyx_n_s_parent, __pyx_t_81) < 0) __PYX_ERR(0, 5083, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_81); __pyx_t_81 = 0;
  if (PyDict_SetItem(__pyx_t_80, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 5083, __pyx_L1_error)
  __pyx_t_81 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19For_equationContext_1__init__, 0, __pyx_n_s_ModelicaParser_For_equationConte_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2091)); if (unlikely(!__pyx_t_81)) __PYX_ERR(0, 5083, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_81);
  __Pyx_INCREF(__pyx_t_81);
  PyList_Append(__pyx_t_79, __pyx_t_81);
  __Pyx_GIVEREF(__pyx_t_81);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_81, __pyx_tuple__2092);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_81, __pyx_t_80);
  __Pyx_DECREF(__pyx_t_80); __pyx_t_80 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_81) < 0) __PYX_ERR(0, 5083, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_81); __pyx_t_81 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5087
 *             self.parser = parser
 * 
 *         def for_indices(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.For_indicesContext,0)
 * 
 */
  __pyx_t_81 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19For_equationContext_3for_indices, 0, __pyx_n_s_ModelicaParser_For_equationConte_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2094)); if (unlikely(!__pyx_t_81)) __PYX_ERR(0, 5087, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_81);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_for_indices, __pyx_t_81) < 0) __PYX_ERR(0, 5087, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_81); __pyx_t_81 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5091
 * 
 * 
 *         def equation(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.EquationContext)
 */
  __pyx_t_81 = PyDict_New(); if (unlikely(!__pyx_t_81)) __PYX_ERR(0, 5091, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_81);
  if (PyDict_SetItem(__pyx_t_81, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 5091, __pyx_L1_error)
  __pyx_t_80 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19For_equationContext_5equation, 0, __pyx_n_s_ModelicaParser_For_equationConte_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2096)); if (unlikely(!__pyx_t_80)) __PYX_ERR(0, 5091, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_80);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_80, __pyx_tuple__2097);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_80, __pyx_t_81);
  __Pyx_DECREF(__pyx_t_81); __pyx_t_81 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_equation, __pyx_t_80) < 0) __PYX_ERR(0, 5091, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_80); __pyx_t_80 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5098
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_for_equation
 * 
 */
  __pyx_t_80 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19For_equationContext_7getRuleIndex, 0, __pyx_n_s_ModelicaParser_For_equationConte_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2099)); if (unlikely(!__pyx_t_80)) __PYX_ERR(0, 5098, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_80);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_80) < 0) __PYX_ERR(0, 5098, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_80); __pyx_t_80 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5101
 *             return ModelicaParser.RULE_for_equation
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterFor_equation" ):
 *                 listener.enterFor_equation(self)
 */
  __pyx_t_80 = PyDict_New(); if (unlikely(!__pyx_t_80)) __PYX_ERR(0, 5101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_80);
  __pyx_t_81 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_81)) __PYX_ERR(0, 5101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_81);
  if (PyDict_SetItem(__pyx_t_80, __pyx_n_s_listener, __pyx_t_81) < 0) __PYX_ERR(0, 5101, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_81); __pyx_t_81 = 0;
  __pyx_t_81 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19For_equationContext_9enterRule, 0, __pyx_n_s_ModelicaParser_For_equationConte_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2101)); if (unlikely(!__pyx_t_81)) __PYX_ERR(0, 5101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_81);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_81, __pyx_t_80);
  __Pyx_DECREF(__pyx_t_80); __pyx_t_80 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_81) < 0) __PYX_ERR(0, 5101, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_81); __pyx_t_81 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5105
 *                 listener.enterFor_equation(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitFor_equation" ):
 *                 listener.exitFor_equation(self)
 */
  __pyx_t_81 = PyDict_New(); if (unlikely(!__pyx_t_81)) __PYX_ERR(0, 5105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_81);
  __pyx_t_80 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_80)) __PYX_ERR(0, 5105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_80);
  if (PyDict_SetItem(__pyx_t_81, __pyx_n_s_listener, __pyx_t_80) < 0) __PYX_ERR(0, 5105, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_80); __pyx_t_80 = 0;
  __pyx_t_80 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19For_equationContext_11exitRule, 0, __pyx_n_s_ModelicaParser_For_equationConte_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2103)); if (unlikely(!__pyx_t_80)) __PYX_ERR(0, 5105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_80);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_80, __pyx_t_81);
  __Pyx_DECREF(__pyx_t_81); __pyx_t_81 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_80) < 0) __PYX_ERR(0, 5105, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_80); __pyx_t_80 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5109
 *                 listener.exitFor_equation(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitFor_equation" ):
 *                 return visitor.visitFor_equation(self)
 */
  __pyx_t_80 = PyDict_New(); if (unlikely(!__pyx_t_80)) __PYX_ERR(0, 5109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_80);
  __pyx_t_81 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_81)) __PYX_ERR(0, 5109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_81);
  if (PyDict_SetItem(__pyx_t_80, __pyx_n_s_visitor, __pyx_t_81) < 0) __PYX_ERR(0, 5109, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_81); __pyx_t_81 = 0;
  __pyx_t_81 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19For_equationContext_13accept, 0, __pyx_n_s_ModelicaParser_For_equationConte_8, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2105)); if (unlikely(!__pyx_t_81)) __PYX_ERR(0, 5109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_81);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_81, __pyx_t_80);
  __Pyx_DECREF(__pyx_t_80); __pyx_t_80 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_81) < 0) __PYX_ERR(0, 5109, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_81); __pyx_t_81 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5081
 *         return localctx
 * 
 *     class For_equationContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_81 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_For_equationContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_81)) __PYX_ERR(0, 5081, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_81);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_79, __pyx_t_81) < 0) __PYX_ERR(0, 5081, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_79); __pyx_t_79 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_For_equationContext, __pyx_t_81) < 0) __PYX_ERR(0, 5081, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_81); __pyx_t_81 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5118
 * 
 * 
 *     def for_equation(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.For_equationContext(self, self._ctx, self.state)
 */
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_95for_equation, 0, __pyx_n_s_ModelicaParser_for_equation, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2107)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_for_equation, __pyx_t_7) < 0) __PYX_ERR(0, 5118, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5155
 *         return localctx
 * 
 *     class For_statementContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_For_statementContext, __pyx_n_s_ModelicaParser_For_statementCont, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_81 = PyList_New(0); if (unlikely(!__pyx_t_81)) __PYX_ERR(0, 5155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_81);

  /* "pymola/generated/ModelicaParser.pyx":5157
 *     class For_statementContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_80 = PyDict_New(); if (unlikely(!__pyx_t_80)) __PYX_ERR(0, 5157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_80);
  __pyx_t_82 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_82)) __PYX_ERR(0, 5157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_82);
  if (PyDict_SetItem(__pyx_t_80, __pyx_n_s_parent, __pyx_t_82) < 0) __PYX_ERR(0, 5157, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_82); __pyx_t_82 = 0;
  if (PyDict_SetItem(__pyx_t_80, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 5157, __pyx_L1_error)
  __pyx_t_82 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20For_statementContext_1__init__, 0, __pyx_n_s_ModelicaParser_For_statementCont_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2109)); if (unlikely(!__pyx_t_82)) __PYX_ERR(0, 5157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_82);
  __Pyx_INCREF(__pyx_t_82);
  PyList_Append(__pyx_t_81, __pyx_t_82);
  __Pyx_GIVEREF(__pyx_t_82);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_82, __pyx_tuple__2110);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_82, __pyx_t_80);
  __Pyx_DECREF(__pyx_t_80); __pyx_t_80 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_82) < 0) __PYX_ERR(0, 5157, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_82); __pyx_t_82 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5161
 *             self.parser = parser
 * 
 *         def for_indices(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.For_indicesContext,0)
 * 
 */
  __pyx_t_82 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20For_statementContext_3for_indices, 0, __pyx_n_s_ModelicaParser_For_statementCont_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2112)); if (unlikely(!__pyx_t_82)) __PYX_ERR(0, 5161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_82);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_for_indices, __pyx_t_82) < 0) __PYX_ERR(0, 5161, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_82); __pyx_t_82 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5165
 * 
 * 
 *         def statement(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.StatementContext)
 */
  __pyx_t_82 = PyDict_New(); if (unlikely(!__pyx_t_82)) __PYX_ERR(0, 5165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_82);
  if (PyDict_SetItem(__pyx_t_82, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 5165, __pyx_L1_error)
  __pyx_t_80 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20For_statementContext_5statement, 0, __pyx_n_s_ModelicaParser_For_statementCont_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2114)); if (unlikely(!__pyx_t_80)) __PYX_ERR(0, 5165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_80);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_80, __pyx_tuple__2115);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_80, __pyx_t_82);
  __Pyx_DECREF(__pyx_t_82); __pyx_t_82 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_statement, __pyx_t_80) < 0) __PYX_ERR(0, 5165, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_80); __pyx_t_80 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5172
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_for_statement
 * 
 */
  __pyx_t_80 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20For_statementContext_7getRuleIndex, 0, __pyx_n_s_ModelicaParser_For_statementCont_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2117)); if (unlikely(!__pyx_t_80)) __PYX_ERR(0, 5172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_80);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_80) < 0) __PYX_ERR(0, 5172, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_80); __pyx_t_80 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5175
 *             return ModelicaParser.RULE_for_statement
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterFor_statement" ):
 *                 listener.enterFor_statement(self)
 */
  __pyx_t_80 = PyDict_New(); if (unlikely(!__pyx_t_80)) __PYX_ERR(0, 5175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_80);
  __pyx_t_82 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_82)) __PYX_ERR(0, 5175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_82);
  if (PyDict_SetItem(__pyx_t_80, __pyx_n_s_listener, __pyx_t_82) < 0) __PYX_ERR(0, 5175, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_82); __pyx_t_82 = 0;
  __pyx_t_82 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20For_statementContext_9enterRule, 0, __pyx_n_s_ModelicaParser_For_statementCont_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2119)); if (unlikely(!__pyx_t_82)) __PYX_ERR(0, 5175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_82);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_82, __pyx_t_80);
  __Pyx_DECREF(__pyx_t_80); __pyx_t_80 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_82) < 0) __PYX_ERR(0, 5175, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_82); __pyx_t_82 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5179
 *                 listener.enterFor_statement(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitFor_statement" ):
 *                 listener.exitFor_statement(self)
 */
  __pyx_t_82 = PyDict_New(); if (unlikely(!__pyx_t_82)) __PYX_ERR(0, 5179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_82);
  __pyx_t_80 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_80)) __PYX_ERR(0, 5179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_80);
  if (PyDict_SetItem(__pyx_t_82, __pyx_n_s_listener, __pyx_t_80) < 0) __PYX_ERR(0, 5179, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_80); __pyx_t_80 = 0;
  __pyx_t_80 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20For_statementContext_11exitRule, 0, __pyx_n_s_ModelicaParser_For_statementCont_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2121)); if (unlikely(!__pyx_t_80)) __PYX_ERR(0, 5179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_80);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_80, __pyx_t_82);
  __Pyx_DECREF(__pyx_t_82); __pyx_t_82 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_80) < 0) __PYX_ERR(0, 5179, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_80); __pyx_t_80 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5183
 *                 listener.exitFor_statement(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitFor_statement" ):
 *                 return visitor.visitFor_statement(self)
 */
  __pyx_t_80 = PyDict_New(); if (unlikely(!__pyx_t_80)) __PYX_ERR(0, 5183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_80);
  __pyx_t_82 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_82)) __PYX_ERR(0, 5183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_82);
  if (PyDict_SetItem(__pyx_t_80, __pyx_n_s_visitor, __pyx_t_82) < 0) __PYX_ERR(0, 5183, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_82); __pyx_t_82 = 0;
  __pyx_t_82 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20For_statementContext_13accept, 0, __pyx_n_s_ModelicaParser_For_statementCont_8, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2123)); if (unlikely(!__pyx_t_82)) __PYX_ERR(0, 5183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_82);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_82, __pyx_t_80);
  __Pyx_DECREF(__pyx_t_80); __pyx_t_80 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_82) < 0) __PYX_ERR(0, 5183, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_82); __pyx_t_82 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5155
 *         return localctx
 * 
 *     class For_statementContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_82 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_For_statementContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_82)) __PYX_ERR(0, 5155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_82);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_81, __pyx_t_82) < 0) __PYX_ERR(0, 5155, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_81); __pyx_t_81 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_For_statementContext, __pyx_t_82) < 0) __PYX_ERR(0, 5155, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_82); __pyx_t_82 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5192
 * 
 * 
 *     def for_statement(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.For_statementContext(self, self._ctx, self.state)
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_97for_statement, 0, __pyx_n_s_ModelicaParser_for_statement, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2125)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_for_statement, __pyx_t_5) < 0) __PYX_ERR(0, 5192, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5229
 *         return localctx
 * 
 *     class For_indicesContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_For_indicesContext, __pyx_n_s_ModelicaParser_For_indicesContex, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_82 = PyList_New(0); if (unlikely(!__pyx_t_82)) __PYX_ERR(0, 5229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_82);

  /* "pymola/generated/ModelicaParser.pyx":5231
 *     class For_indicesContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_80 = PyDict_New(); if (unlikely(!__pyx_t_80)) __PYX_ERR(0, 5231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_80);
  __pyx_t_83 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_83)) __PYX_ERR(0, 5231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_83);
  if (PyDict_SetItem(__pyx_t_80, __pyx_n_s_parent, __pyx_t_83) < 0) __PYX_ERR(0, 5231, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_83); __pyx_t_83 = 0;
  if (PyDict_SetItem(__pyx_t_80, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 5231, __pyx_L1_error)
  __pyx_t_83 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18For_indicesContext_1__init__, 0, __pyx_n_s_ModelicaParser_For_indicesContex_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2127)); if (unlikely(!__pyx_t_83)) __PYX_ERR(0, 5231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_83);
  __Pyx_INCREF(__pyx_t_83);
  PyList_Append(__pyx_t_82, __pyx_t_83);
  __Pyx_GIVEREF(__pyx_t_83);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_83, __pyx_tuple__2128);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_83, __pyx_t_80);
  __Pyx_DECREF(__pyx_t_80); __pyx_t_80 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_83) < 0) __PYX_ERR(0, 5231, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_83); __pyx_t_83 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5235
 *             self.parser = parser
 * 
 *         def for_index(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.For_indexContext)
 */
  __pyx_t_83 = PyDict_New(); if (unlikely(!__pyx_t_83)) __PYX_ERR(0, 5235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_83);
  if (PyDict_SetItem(__pyx_t_83, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 5235, __pyx_L1_error)
  __pyx_t_80 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18For_indicesContext_3for_index, 0, __pyx_n_s_ModelicaParser_For_indicesContex_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2130)); if (unlikely(!__pyx_t_80)) __PYX_ERR(0, 5235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_80);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_80, __pyx_tuple__2131);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_80, __pyx_t_83);
  __Pyx_DECREF(__pyx_t_83); __pyx_t_83 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_for_index, __pyx_t_80) < 0) __PYX_ERR(0, 5235, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_80); __pyx_t_80 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5242
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_for_indices
 * 
 */
  __pyx_t_80 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18For_indicesContext_5getRuleIndex, 0, __pyx_n_s_ModelicaParser_For_indicesContex_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2133)); if (unlikely(!__pyx_t_80)) __PYX_ERR(0, 5242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_80);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_80) < 0) __PYX_ERR(0, 5242, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_80); __pyx_t_80 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5245
 *             return ModelicaParser.RULE_for_indices
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterFor_indices" ):
 *                 listener.enterFor_indices(self)
 */
  __pyx_t_80 = PyDict_New(); if (unlikely(!__pyx_t_80)) __PYX_ERR(0, 5245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_80);
  __pyx_t_83 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_83)) __PYX_ERR(0, 5245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_83);
  if (PyDict_SetItem(__pyx_t_80, __pyx_n_s_listener, __pyx_t_83) < 0) __PYX_ERR(0, 5245, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_83); __pyx_t_83 = 0;
  __pyx_t_83 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18For_indicesContext_7enterRule, 0, __pyx_n_s_ModelicaParser_For_indicesContex_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2135)); if (unlikely(!__pyx_t_83)) __PYX_ERR(0, 5245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_83);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_83, __pyx_t_80);
  __Pyx_DECREF(__pyx_t_80); __pyx_t_80 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_83) < 0) __PYX_ERR(0, 5245, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_83); __pyx_t_83 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5249
 *                 listener.enterFor_indices(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitFor_indices" ):
 *                 listener.exitFor_indices(self)
 */
  __pyx_t_83 = PyDict_New(); if (unlikely(!__pyx_t_83)) __PYX_ERR(0, 5249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_83);
  __pyx_t_80 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_80)) __PYX_ERR(0, 5249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_80);
  if (PyDict_SetItem(__pyx_t_83, __pyx_n_s_listener, __pyx_t_80) < 0) __PYX_ERR(0, 5249, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_80); __pyx_t_80 = 0;
  __pyx_t_80 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18For_indicesContext_9exitRule, 0, __pyx_n_s_ModelicaParser_For_indicesContex_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2137)); if (unlikely(!__pyx_t_80)) __PYX_ERR(0, 5249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_80);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_80, __pyx_t_83);
  __Pyx_DECREF(__pyx_t_83); __pyx_t_83 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_80) < 0) __PYX_ERR(0, 5249, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_80); __pyx_t_80 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5253
 *                 listener.exitFor_indices(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitFor_indices" ):
 *                 return visitor.visitFor_indices(self)
 */
  __pyx_t_80 = PyDict_New(); if (unlikely(!__pyx_t_80)) __PYX_ERR(0, 5253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_80);
  __pyx_t_83 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_83)) __PYX_ERR(0, 5253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_83);
  if (PyDict_SetItem(__pyx_t_80, __pyx_n_s_visitor, __pyx_t_83) < 0) __PYX_ERR(0, 5253, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_83); __pyx_t_83 = 0;
  __pyx_t_83 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18For_indicesContext_11accept, 0, __pyx_n_s_ModelicaParser_For_indicesContex_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2139)); if (unlikely(!__pyx_t_83)) __PYX_ERR(0, 5253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_83);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_83, __pyx_t_80);
  __Pyx_DECREF(__pyx_t_80); __pyx_t_80 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_83) < 0) __PYX_ERR(0, 5253, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_83); __pyx_t_83 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5229
 *         return localctx
 * 
 *     class For_indicesContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_83 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_For_indicesContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_83)) __PYX_ERR(0, 5229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_83);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_82, __pyx_t_83) < 0) __PYX_ERR(0, 5229, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_82); __pyx_t_82 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_For_indicesContext, __pyx_t_83) < 0) __PYX_ERR(0, 5229, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_83); __pyx_t_83 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5262
 * 
 * 
 *     def for_indices(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.For_indicesContext(self, self._ctx, self.state)
 */
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_99for_indices, 0, __pyx_n_s_ModelicaParser_for_indices, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2141)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_for_indices, __pyx_t_7) < 0) __PYX_ERR(0, 5262, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5291
 *         return localctx
 * 
 *     class For_indexContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_For_indexContext, __pyx_n_s_ModelicaParser_For_indexContext, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_83 = PyList_New(0); if (unlikely(!__pyx_t_83)) __PYX_ERR(0, 5291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_83);

  /* "pymola/generated/ModelicaParser.pyx":5293
 *     class For_indexContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_80 = PyDict_New(); if (unlikely(!__pyx_t_80)) __PYX_ERR(0, 5293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_80);
  __pyx_t_84 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_84)) __PYX_ERR(0, 5293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_84);
  if (PyDict_SetItem(__pyx_t_80, __pyx_n_s_parent, __pyx_t_84) < 0) __PYX_ERR(0, 5293, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_84); __pyx_t_84 = 0;
  if (PyDict_SetItem(__pyx_t_80, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 5293, __pyx_L1_error)
  __pyx_t_84 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_16For_indexContext_1__init__, 0, __pyx_n_s_ModelicaParser_For_indexContext_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2143)); if (unlikely(!__pyx_t_84)) __PYX_ERR(0, 5293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_84);
  __Pyx_INCREF(__pyx_t_84);
  PyList_Append(__pyx_t_83, __pyx_t_84);
  __Pyx_GIVEREF(__pyx_t_84);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_84, __pyx_tuple__2144);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_84, __pyx_t_80);
  __Pyx_DECREF(__pyx_t_80); __pyx_t_80 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_84) < 0) __PYX_ERR(0, 5293, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_84); __pyx_t_84 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5297
 *             self.parser = parser
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 */
  __pyx_t_84 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_16For_indexContext_3IDENT, 0, __pyx_n_s_ModelicaParser_For_indexContext_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2146)); if (unlikely(!__pyx_t_84)) __PYX_ERR(0, 5297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_84);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_IDENT, __pyx_t_84) < 0) __PYX_ERR(0, 5297, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_84); __pyx_t_84 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5300
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 *         def expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 */
  __pyx_t_84 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_16For_indexContext_5expression, 0, __pyx_n_s_ModelicaParser_For_indexContext_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2148)); if (unlikely(!__pyx_t_84)) __PYX_ERR(0, 5300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_84);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_expression, __pyx_t_84) < 0) __PYX_ERR(0, 5300, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_84); __pyx_t_84 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5304
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_for_index
 * 
 */
  __pyx_t_84 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_16For_indexContext_7getRuleIndex, 0, __pyx_n_s_ModelicaParser_For_indexContext_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2150)); if (unlikely(!__pyx_t_84)) __PYX_ERR(0, 5304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_84);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_84) < 0) __PYX_ERR(0, 5304, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_84); __pyx_t_84 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5307
 *             return ModelicaParser.RULE_for_index
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterFor_index" ):
 *                 listener.enterFor_index(self)
 */
  __pyx_t_84 = PyDict_New(); if (unlikely(!__pyx_t_84)) __PYX_ERR(0, 5307, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_84);
  __pyx_t_80 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_80)) __PYX_ERR(0, 5307, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_80);
  if (PyDict_SetItem(__pyx_t_84, __pyx_n_s_listener, __pyx_t_80) < 0) __PYX_ERR(0, 5307, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_80); __pyx_t_80 = 0;
  __pyx_t_80 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_16For_indexContext_9enterRule, 0, __pyx_n_s_ModelicaParser_For_indexContext_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2152)); if (unlikely(!__pyx_t_80)) __PYX_ERR(0, 5307, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_80);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_80, __pyx_t_84);
  __Pyx_DECREF(__pyx_t_84); __pyx_t_84 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_80) < 0) __PYX_ERR(0, 5307, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_80); __pyx_t_80 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5311
 *                 listener.enterFor_index(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitFor_index" ):
 *                 listener.exitFor_index(self)
 */
  __pyx_t_80 = PyDict_New(); if (unlikely(!__pyx_t_80)) __PYX_ERR(0, 5311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_80);
  __pyx_t_84 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_84)) __PYX_ERR(0, 5311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_84);
  if (PyDict_SetItem(__pyx_t_80, __pyx_n_s_listener, __pyx_t_84) < 0) __PYX_ERR(0, 5311, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_84); __pyx_t_84 = 0;
  __pyx_t_84 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_16For_indexContext_11exitRule, 0, __pyx_n_s_ModelicaParser_For_indexContext_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2154)); if (unlikely(!__pyx_t_84)) __PYX_ERR(0, 5311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_84);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_84, __pyx_t_80);
  __Pyx_DECREF(__pyx_t_80); __pyx_t_80 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_84) < 0) __PYX_ERR(0, 5311, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_84); __pyx_t_84 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5315
 *                 listener.exitFor_index(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitFor_index" ):
 *                 return visitor.visitFor_index(self)
 */
  __pyx_t_84 = PyDict_New(); if (unlikely(!__pyx_t_84)) __PYX_ERR(0, 5315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_84);
  __pyx_t_80 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_80)) __PYX_ERR(0, 5315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_80);
  if (PyDict_SetItem(__pyx_t_84, __pyx_n_s_visitor, __pyx_t_80) < 0) __PYX_ERR(0, 5315, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_80); __pyx_t_80 = 0;
  __pyx_t_80 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_16For_indexContext_13accept, 0, __pyx_n_s_ModelicaParser_For_indexContext_8, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2156)); if (unlikely(!__pyx_t_80)) __PYX_ERR(0, 5315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_80);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_80, __pyx_t_84);
  __Pyx_DECREF(__pyx_t_84); __pyx_t_84 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_80) < 0) __PYX_ERR(0, 5315, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_80); __pyx_t_80 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5291
 *         return localctx
 * 
 *     class For_indexContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_80 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_For_indexContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_80)) __PYX_ERR(0, 5291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_80);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_83, __pyx_t_80) < 0) __PYX_ERR(0, 5291, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_83); __pyx_t_83 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_For_indexContext, __pyx_t_80) < 0) __PYX_ERR(0, 5291, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_80); __pyx_t_80 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5324
 * 
 * 
 *     def for_index(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.For_indexContext(self, self._ctx, self.state)
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_101for_index, 0, __pyx_n_s_ModelicaParser_for_index, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2158)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_for_index, __pyx_t_5) < 0) __PYX_ERR(0, 5324, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5351
 *         return localctx
 * 
 *     class While_statementContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_While_statementContext, __pyx_n_s_ModelicaParser_While_statementCo, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_80 = PyList_New(0); if (unlikely(!__pyx_t_80)) __PYX_ERR(0, 5351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_80);

  /* "pymola/generated/ModelicaParser.pyx":5353
 *     class While_statementContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_84 = PyDict_New(); if (unlikely(!__pyx_t_84)) __PYX_ERR(0, 5353, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_84);
  __pyx_t_85 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_85)) __PYX_ERR(0, 5353, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_85);
  if (PyDict_SetItem(__pyx_t_84, __pyx_n_s_parent, __pyx_t_85) < 0) __PYX_ERR(0, 5353, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_85); __pyx_t_85 = 0;
  if (PyDict_SetItem(__pyx_t_84, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 5353, __pyx_L1_error)
  __pyx_t_85 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22While_statementContext_1__init__, 0, __pyx_n_s_ModelicaParser_While_statementCo_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2160)); if (unlikely(!__pyx_t_85)) __PYX_ERR(0, 5353, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_85);
  __Pyx_INCREF(__pyx_t_85);
  PyList_Append(__pyx_t_80, __pyx_t_85);
  __Pyx_GIVEREF(__pyx_t_85);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_85, __pyx_tuple__2161);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_85, __pyx_t_84);
  __Pyx_DECREF(__pyx_t_84); __pyx_t_84 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_85) < 0) __PYX_ERR(0, 5353, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_85); __pyx_t_85 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5357
 *             self.parser = parser
 * 
 *         def expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 */
  __pyx_t_85 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22While_statementContext_3expression, 0, __pyx_n_s_ModelicaParser_While_statementCo_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2163)); if (unlikely(!__pyx_t_85)) __PYX_ERR(0, 5357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_85);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_expression, __pyx_t_85) < 0) __PYX_ERR(0, 5357, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_85); __pyx_t_85 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5361
 * 
 * 
 *         def statement(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.StatementContext)
 */
  __pyx_t_85 = PyDict_New(); if (unlikely(!__pyx_t_85)) __PYX_ERR(0, 5361, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_85);
  if (PyDict_SetItem(__pyx_t_85, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 5361, __pyx_L1_error)
  __pyx_t_84 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22While_statementContext_5statement, 0, __pyx_n_s_ModelicaParser_While_statementCo_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2165)); if (unlikely(!__pyx_t_84)) __PYX_ERR(0, 5361, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_84);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_84, __pyx_tuple__2166);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_84, __pyx_t_85);
  __Pyx_DECREF(__pyx_t_85); __pyx_t_85 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_statement, __pyx_t_84) < 0) __PYX_ERR(0, 5361, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_84); __pyx_t_84 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5368
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_while_statement
 * 
 */
  __pyx_t_84 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22While_statementContext_7getRuleIndex, 0, __pyx_n_s_ModelicaParser_While_statementCo_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2168)); if (unlikely(!__pyx_t_84)) __PYX_ERR(0, 5368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_84);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_84) < 0) __PYX_ERR(0, 5368, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_84); __pyx_t_84 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5371
 *             return ModelicaParser.RULE_while_statement
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterWhile_statement" ):
 *                 listener.enterWhile_statement(self)
 */
  __pyx_t_84 = PyDict_New(); if (unlikely(!__pyx_t_84)) __PYX_ERR(0, 5371, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_84);
  __pyx_t_85 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_85)) __PYX_ERR(0, 5371, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_85);
  if (PyDict_SetItem(__pyx_t_84, __pyx_n_s_listener, __pyx_t_85) < 0) __PYX_ERR(0, 5371, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_85); __pyx_t_85 = 0;
  __pyx_t_85 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22While_statementContext_9enterRule, 0, __pyx_n_s_ModelicaParser_While_statementCo_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2170)); if (unlikely(!__pyx_t_85)) __PYX_ERR(0, 5371, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_85);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_85, __pyx_t_84);
  __Pyx_DECREF(__pyx_t_84); __pyx_t_84 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_85) < 0) __PYX_ERR(0, 5371, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_85); __pyx_t_85 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5375
 *                 listener.enterWhile_statement(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitWhile_statement" ):
 *                 listener.exitWhile_statement(self)
 */
  __pyx_t_85 = PyDict_New(); if (unlikely(!__pyx_t_85)) __PYX_ERR(0, 5375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_85);
  __pyx_t_84 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_84)) __PYX_ERR(0, 5375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_84);
  if (PyDict_SetItem(__pyx_t_85, __pyx_n_s_listener, __pyx_t_84) < 0) __PYX_ERR(0, 5375, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_84); __pyx_t_84 = 0;
  __pyx_t_84 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22While_statementContext_11exitRule, 0, __pyx_n_s_ModelicaParser_While_statementCo_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2172)); if (unlikely(!__pyx_t_84)) __PYX_ERR(0, 5375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_84);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_84, __pyx_t_85);
  __Pyx_DECREF(__pyx_t_85); __pyx_t_85 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_84) < 0) __PYX_ERR(0, 5375, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_84); __pyx_t_84 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5379
 *                 listener.exitWhile_statement(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitWhile_statement" ):
 *                 return visitor.visitWhile_statement(self)
 */
  __pyx_t_84 = PyDict_New(); if (unlikely(!__pyx_t_84)) __PYX_ERR(0, 5379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_84);
  __pyx_t_85 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_85)) __PYX_ERR(0, 5379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_85);
  if (PyDict_SetItem(__pyx_t_84, __pyx_n_s_visitor, __pyx_t_85) < 0) __PYX_ERR(0, 5379, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_85); __pyx_t_85 = 0;
  __pyx_t_85 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22While_statementContext_13accept, 0, __pyx_n_s_ModelicaParser_While_statementCo_8, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2174)); if (unlikely(!__pyx_t_85)) __PYX_ERR(0, 5379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_85);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_85, __pyx_t_84);
  __Pyx_DECREF(__pyx_t_84); __pyx_t_84 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_85) < 0) __PYX_ERR(0, 5379, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_85); __pyx_t_85 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5351
 *         return localctx
 * 
 *     class While_statementContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_85 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_While_statementContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_85)) __PYX_ERR(0, 5351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_85);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_80, __pyx_t_85) < 0) __PYX_ERR(0, 5351, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_80); __pyx_t_80 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_While_statementContext, __pyx_t_85) < 0) __PYX_ERR(0, 5351, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_85); __pyx_t_85 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5388
 * 
 * 
 *     def while_statement(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.While_statementContext(self, self._ctx, self.state)
 */
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_103while_statement, 0, __pyx_n_s_ModelicaParser_while_statement, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2176)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_while_statement, __pyx_t_7) < 0) __PYX_ERR(0, 5388, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5425
 *         return localctx
 * 
 *     class When_equationContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_When_equationContext, __pyx_n_s_ModelicaParser_When_equationCont, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_85 = PyList_New(0); if (unlikely(!__pyx_t_85)) __PYX_ERR(0, 5425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_85);

  /* "pymola/generated/ModelicaParser.pyx":5427
 *     class When_equationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_84 = PyDict_New(); if (unlikely(!__pyx_t_84)) __PYX_ERR(0, 5427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_84);
  __pyx_t_86 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_86)) __PYX_ERR(0, 5427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_86);
  if (PyDict_SetItem(__pyx_t_84, __pyx_n_s_parent, __pyx_t_86) < 0) __PYX_ERR(0, 5427, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_86); __pyx_t_86 = 0;
  if (PyDict_SetItem(__pyx_t_84, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 5427, __pyx_L1_error)
  __pyx_t_86 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20When_equationContext_1__init__, 0, __pyx_n_s_ModelicaParser_When_equationCont_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2178)); if (unlikely(!__pyx_t_86)) __PYX_ERR(0, 5427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_86);
  __Pyx_INCREF(__pyx_t_86);
  PyList_Append(__pyx_t_85, __pyx_t_86);
  __Pyx_GIVEREF(__pyx_t_86);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_86, __pyx_tuple__2179);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_86, __pyx_t_84);
  __Pyx_DECREF(__pyx_t_84); __pyx_t_84 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_86) < 0) __PYX_ERR(0, 5427, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_86); __pyx_t_86 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5431
 *             self.parser = parser
 * 
 *         def expression(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 */
  __pyx_t_86 = PyDict_New(); if (unlikely(!__pyx_t_86)) __PYX_ERR(0, 5431, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_86);
  if (PyDict_SetItem(__pyx_t_86, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 5431, __pyx_L1_error)
  __pyx_t_84 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20When_equationContext_3expression, 0, __pyx_n_s_ModelicaParser_When_equationCont_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2181)); if (unlikely(!__pyx_t_84)) __PYX_ERR(0, 5431, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_84);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_84, __pyx_tuple__2182);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_84, __pyx_t_86);
  __Pyx_DECREF(__pyx_t_86); __pyx_t_86 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_expression, __pyx_t_84) < 0) __PYX_ERR(0, 5431, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_84); __pyx_t_84 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5438
 * 
 * 
 *         def equation(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.EquationContext)
 */
  __pyx_t_84 = PyDict_New(); if (unlikely(!__pyx_t_84)) __PYX_ERR(0, 5438, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_84);
  if (PyDict_SetItem(__pyx_t_84, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 5438, __pyx_L1_error)
  __pyx_t_86 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20When_equationContext_5equation, 0, __pyx_n_s_ModelicaParser_When_equationCont_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2184)); if (unlikely(!__pyx_t_86)) __PYX_ERR(0, 5438, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_86);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_86, __pyx_tuple__2185);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_86, __pyx_t_84);
  __Pyx_DECREF(__pyx_t_84); __pyx_t_84 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_equation, __pyx_t_86) < 0) __PYX_ERR(0, 5438, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_86); __pyx_t_86 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5445
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_when_equation
 * 
 */
  __pyx_t_86 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20When_equationContext_7getRuleIndex, 0, __pyx_n_s_ModelicaParser_When_equationCont_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2187)); if (unlikely(!__pyx_t_86)) __PYX_ERR(0, 5445, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_86);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_86) < 0) __PYX_ERR(0, 5445, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_86); __pyx_t_86 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5448
 *             return ModelicaParser.RULE_when_equation
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterWhen_equation" ):
 *                 listener.enterWhen_equation(self)
 */
  __pyx_t_86 = PyDict_New(); if (unlikely(!__pyx_t_86)) __PYX_ERR(0, 5448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_86);
  __pyx_t_84 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_84)) __PYX_ERR(0, 5448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_84);
  if (PyDict_SetItem(__pyx_t_86, __pyx_n_s_listener, __pyx_t_84) < 0) __PYX_ERR(0, 5448, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_84); __pyx_t_84 = 0;
  __pyx_t_84 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20When_equationContext_9enterRule, 0, __pyx_n_s_ModelicaParser_When_equationCont_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2189)); if (unlikely(!__pyx_t_84)) __PYX_ERR(0, 5448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_84);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_84, __pyx_t_86);
  __Pyx_DECREF(__pyx_t_86); __pyx_t_86 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_84) < 0) __PYX_ERR(0, 5448, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_84); __pyx_t_84 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5452
 *                 listener.enterWhen_equation(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitWhen_equation" ):
 *                 listener.exitWhen_equation(self)
 */
  __pyx_t_84 = PyDict_New(); if (unlikely(!__pyx_t_84)) __PYX_ERR(0, 5452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_84);
  __pyx_t_86 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_86)) __PYX_ERR(0, 5452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_86);
  if (PyDict_SetItem(__pyx_t_84, __pyx_n_s_listener, __pyx_t_86) < 0) __PYX_ERR(0, 5452, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_86); __pyx_t_86 = 0;
  __pyx_t_86 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20When_equationContext_11exitRule, 0, __pyx_n_s_ModelicaParser_When_equationCont_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2191)); if (unlikely(!__pyx_t_86)) __PYX_ERR(0, 5452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_86);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_86, __pyx_t_84);
  __Pyx_DECREF(__pyx_t_84); __pyx_t_84 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_86) < 0) __PYX_ERR(0, 5452, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_86); __pyx_t_86 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5456
 *                 listener.exitWhen_equation(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitWhen_equation" ):
 *                 return visitor.visitWhen_equation(self)
 */
  __pyx_t_86 = PyDict_New(); if (unlikely(!__pyx_t_86)) __PYX_ERR(0, 5456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_86);
  __pyx_t_84 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_84)) __PYX_ERR(0, 5456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_84);
  if (PyDict_SetItem(__pyx_t_86, __pyx_n_s_visitor, __pyx_t_84) < 0) __PYX_ERR(0, 5456, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_84); __pyx_t_84 = 0;
  __pyx_t_84 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20When_equationContext_13accept, 0, __pyx_n_s_ModelicaParser_When_equationCont_8, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2193)); if (unlikely(!__pyx_t_84)) __PYX_ERR(0, 5456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_84);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_84, __pyx_t_86);
  __Pyx_DECREF(__pyx_t_86); __pyx_t_86 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_84) < 0) __PYX_ERR(0, 5456, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_84); __pyx_t_84 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5425
 *         return localctx
 * 
 *     class When_equationContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_84 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_When_equationContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_84)) __PYX_ERR(0, 5425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_84);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_85, __pyx_t_84) < 0) __PYX_ERR(0, 5425, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_85); __pyx_t_85 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_When_equationContext, __pyx_t_84) < 0) __PYX_ERR(0, 5425, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_84); __pyx_t_84 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5465
 * 
 * 
 *     def when_equation(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.When_equationContext(self, self._ctx, self.state)
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_105when_equation, 0, __pyx_n_s_ModelicaParser_when_equation, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2195)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5465, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_when_equation, __pyx_t_5) < 0) __PYX_ERR(0, 5465, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5530
 *         return localctx
 * 
 *     class When_statementContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_When_statementContext, __pyx_n_s_ModelicaParser_When_statementCon, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_84 = PyList_New(0); if (unlikely(!__pyx_t_84)) __PYX_ERR(0, 5530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_84);

  /* "pymola/generated/ModelicaParser.pyx":5532
 *     class When_statementContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_86 = PyDict_New(); if (unlikely(!__pyx_t_86)) __PYX_ERR(0, 5532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_86);
  __pyx_t_87 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_87)) __PYX_ERR(0, 5532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_87);
  if (PyDict_SetItem(__pyx_t_86, __pyx_n_s_parent, __pyx_t_87) < 0) __PYX_ERR(0, 5532, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_87); __pyx_t_87 = 0;
  if (PyDict_SetItem(__pyx_t_86, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 5532, __pyx_L1_error)
  __pyx_t_87 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21When_statementContext_1__init__, 0, __pyx_n_s_ModelicaParser_When_statementCon_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2197)); if (unlikely(!__pyx_t_87)) __PYX_ERR(0, 5532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_87);
  __Pyx_INCREF(__pyx_t_87);
  PyList_Append(__pyx_t_84, __pyx_t_87);
  __Pyx_GIVEREF(__pyx_t_87);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_87, __pyx_tuple__2198);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_87, __pyx_t_86);
  __Pyx_DECREF(__pyx_t_86); __pyx_t_86 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_87) < 0) __PYX_ERR(0, 5532, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_87); __pyx_t_87 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5536
 *             self.parser = parser
 * 
 *         def expression(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 */
  __pyx_t_87 = PyDict_New(); if (unlikely(!__pyx_t_87)) __PYX_ERR(0, 5536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_87);
  if (PyDict_SetItem(__pyx_t_87, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 5536, __pyx_L1_error)
  __pyx_t_86 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21When_statementContext_3expression, 0, __pyx_n_s_ModelicaParser_When_statementCon_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2200)); if (unlikely(!__pyx_t_86)) __PYX_ERR(0, 5536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_86);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_86, __pyx_tuple__2201);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_86, __pyx_t_87);
  __Pyx_DECREF(__pyx_t_87); __pyx_t_87 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_expression, __pyx_t_86) < 0) __PYX_ERR(0, 5536, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_86); __pyx_t_86 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5543
 * 
 * 
 *         def statement(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.StatementContext)
 */
  __pyx_t_86 = PyDict_New(); if (unlikely(!__pyx_t_86)) __PYX_ERR(0, 5543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_86);
  if (PyDict_SetItem(__pyx_t_86, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 5543, __pyx_L1_error)
  __pyx_t_87 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21When_statementContext_5statement, 0, __pyx_n_s_ModelicaParser_When_statementCon_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2203)); if (unlikely(!__pyx_t_87)) __PYX_ERR(0, 5543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_87);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_87, __pyx_tuple__2204);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_87, __pyx_t_86);
  __Pyx_DECREF(__pyx_t_86); __pyx_t_86 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_statement, __pyx_t_87) < 0) __PYX_ERR(0, 5543, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_87); __pyx_t_87 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5550
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_when_statement
 * 
 */
  __pyx_t_87 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21When_statementContext_7getRuleIndex, 0, __pyx_n_s_ModelicaParser_When_statementCon_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2206)); if (unlikely(!__pyx_t_87)) __PYX_ERR(0, 5550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_87);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_87) < 0) __PYX_ERR(0, 5550, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_87); __pyx_t_87 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5553
 *             return ModelicaParser.RULE_when_statement
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterWhen_statement" ):
 *                 listener.enterWhen_statement(self)
 */
  __pyx_t_87 = PyDict_New(); if (unlikely(!__pyx_t_87)) __PYX_ERR(0, 5553, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_87);
  __pyx_t_86 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_86)) __PYX_ERR(0, 5553, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_86);
  if (PyDict_SetItem(__pyx_t_87, __pyx_n_s_listener, __pyx_t_86) < 0) __PYX_ERR(0, 5553, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_86); __pyx_t_86 = 0;
  __pyx_t_86 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21When_statementContext_9enterRule, 0, __pyx_n_s_ModelicaParser_When_statementCon_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2208)); if (unlikely(!__pyx_t_86)) __PYX_ERR(0, 5553, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_86);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_86, __pyx_t_87);
  __Pyx_DECREF(__pyx_t_87); __pyx_t_87 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_86) < 0) __PYX_ERR(0, 5553, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_86); __pyx_t_86 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5557
 *                 listener.enterWhen_statement(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitWhen_statement" ):
 *                 listener.exitWhen_statement(self)
 */
  __pyx_t_86 = PyDict_New(); if (unlikely(!__pyx_t_86)) __PYX_ERR(0, 5557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_86);
  __pyx_t_87 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_87)) __PYX_ERR(0, 5557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_87);
  if (PyDict_SetItem(__pyx_t_86, __pyx_n_s_listener, __pyx_t_87) < 0) __PYX_ERR(0, 5557, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_87); __pyx_t_87 = 0;
  __pyx_t_87 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21When_statementContext_11exitRule, 0, __pyx_n_s_ModelicaParser_When_statementCon_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2210)); if (unlikely(!__pyx_t_87)) __PYX_ERR(0, 5557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_87);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_87, __pyx_t_86);
  __Pyx_DECREF(__pyx_t_86); __pyx_t_86 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_87) < 0) __PYX_ERR(0, 5557, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_87); __pyx_t_87 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5561
 *                 listener.exitWhen_statement(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitWhen_statement" ):
 *                 return visitor.visitWhen_statement(self)
 */
  __pyx_t_87 = PyDict_New(); if (unlikely(!__pyx_t_87)) __PYX_ERR(0, 5561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_87);
  __pyx_t_86 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_86)) __PYX_ERR(0, 5561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_86);
  if (PyDict_SetItem(__pyx_t_87, __pyx_n_s_visitor, __pyx_t_86) < 0) __PYX_ERR(0, 5561, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_86); __pyx_t_86 = 0;
  __pyx_t_86 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21When_statementContext_13accept, 0, __pyx_n_s_ModelicaParser_When_statementCon_8, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2212)); if (unlikely(!__pyx_t_86)) __PYX_ERR(0, 5561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_86);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_86, __pyx_t_87);
  __Pyx_DECREF(__pyx_t_87); __pyx_t_87 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_86) < 0) __PYX_ERR(0, 5561, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_86); __pyx_t_86 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5530
 *         return localctx
 * 
 *     class When_statementContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_86 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_When_statementContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_86)) __PYX_ERR(0, 5530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_86);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_84, __pyx_t_86) < 0) __PYX_ERR(0, 5530, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_84); __pyx_t_84 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_When_statementContext, __pyx_t_86) < 0) __PYX_ERR(0, 5530, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_86); __pyx_t_86 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5570
 * 
 * 
 *     def when_statement(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.When_statementContext(self, self._ctx, self.state)
 */
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_107when_statement, 0, __pyx_n_s_ModelicaParser_when_statement, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2214)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_when_statement, __pyx_t_7) < 0) __PYX_ERR(0, 5570, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5633
 *         return localctx
 * 
 *     class Connect_clauseContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Connect_clauseContext, __pyx_n_s_ModelicaParser_Connect_clauseCon, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_86 = PyList_New(0); if (unlikely(!__pyx_t_86)) __PYX_ERR(0, 5633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_86);

  /* "pymola/generated/ModelicaParser.pyx":5635
 *     class Connect_clauseContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_87 = PyDict_New(); if (unlikely(!__pyx_t_87)) __PYX_ERR(0, 5635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_87);
  __pyx_t_88 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_88)) __PYX_ERR(0, 5635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_88);
  if (PyDict_SetItem(__pyx_t_87, __pyx_n_s_parent, __pyx_t_88) < 0) __PYX_ERR(0, 5635, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_88); __pyx_t_88 = 0;
  if (PyDict_SetItem(__pyx_t_87, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 5635, __pyx_L1_error)
  __pyx_t_88 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_1__init__, 0, __pyx_n_s_ModelicaParser_Connect_clauseCon_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2216)); if (unlikely(!__pyx_t_88)) __PYX_ERR(0, 5635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_88);
  __Pyx_INCREF(__pyx_t_88);
  PyList_Append(__pyx_t_86, __pyx_t_88);
  __Pyx_GIVEREF(__pyx_t_88);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_88, __pyx_tuple__2217);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_88, __pyx_t_87);
  __Pyx_DECREF(__pyx_t_87); __pyx_t_87 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_88) < 0) __PYX_ERR(0, 5635, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_88); __pyx_t_88 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5639
 *             self.parser = parser
 * 
 *         def component_reference(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Component_referenceContext)
 */
  __pyx_t_88 = PyDict_New(); if (unlikely(!__pyx_t_88)) __PYX_ERR(0, 5639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_88);
  if (PyDict_SetItem(__pyx_t_88, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 5639, __pyx_L1_error)
  __pyx_t_87 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_3component_reference, 0, __pyx_n_s_ModelicaParser_Connect_clauseCon_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2219)); if (unlikely(!__pyx_t_87)) __PYX_ERR(0, 5639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_87);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_87, __pyx_tuple__2220);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_87, __pyx_t_88);
  __Pyx_DECREF(__pyx_t_88); __pyx_t_88 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_component_reference, __pyx_t_87) < 0) __PYX_ERR(0, 5639, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_87); __pyx_t_87 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5646
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_connect_clause
 * 
 */
  __pyx_t_87 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_5getRuleIndex, 0, __pyx_n_s_ModelicaParser_Connect_clauseCon_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2222)); if (unlikely(!__pyx_t_87)) __PYX_ERR(0, 5646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_87);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_87) < 0) __PYX_ERR(0, 5646, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_87); __pyx_t_87 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5649
 *             return ModelicaParser.RULE_connect_clause
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterConnect_clause" ):
 *                 listener.enterConnect_clause(self)
 */
  __pyx_t_87 = PyDict_New(); if (unlikely(!__pyx_t_87)) __PYX_ERR(0, 5649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_87);
  __pyx_t_88 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_88)) __PYX_ERR(0, 5649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_88);
  if (PyDict_SetItem(__pyx_t_87, __pyx_n_s_listener, __pyx_t_88) < 0) __PYX_ERR(0, 5649, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_88); __pyx_t_88 = 0;
  __pyx_t_88 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_7enterRule, 0, __pyx_n_s_ModelicaParser_Connect_clauseCon_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2224)); if (unlikely(!__pyx_t_88)) __PYX_ERR(0, 5649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_88);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_88, __pyx_t_87);
  __Pyx_DECREF(__pyx_t_87); __pyx_t_87 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_88) < 0) __PYX_ERR(0, 5649, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_88); __pyx_t_88 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5653
 *                 listener.enterConnect_clause(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitConnect_clause" ):
 *                 listener.exitConnect_clause(self)
 */
  __pyx_t_88 = PyDict_New(); if (unlikely(!__pyx_t_88)) __PYX_ERR(0, 5653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_88);
  __pyx_t_87 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_87)) __PYX_ERR(0, 5653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_87);
  if (PyDict_SetItem(__pyx_t_88, __pyx_n_s_listener, __pyx_t_87) < 0) __PYX_ERR(0, 5653, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_87); __pyx_t_87 = 0;
  __pyx_t_87 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_9exitRule, 0, __pyx_n_s_ModelicaParser_Connect_clauseCon_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2226)); if (unlikely(!__pyx_t_87)) __PYX_ERR(0, 5653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_87);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_87, __pyx_t_88);
  __Pyx_DECREF(__pyx_t_88); __pyx_t_88 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_87) < 0) __PYX_ERR(0, 5653, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_87); __pyx_t_87 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5657
 *                 listener.exitConnect_clause(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitConnect_clause" ):
 *                 return visitor.visitConnect_clause(self)
 */
  __pyx_t_87 = PyDict_New(); if (unlikely(!__pyx_t_87)) __PYX_ERR(0, 5657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_87);
  __pyx_t_88 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_88)) __PYX_ERR(0, 5657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_88);
  if (PyDict_SetItem(__pyx_t_87, __pyx_n_s_visitor, __pyx_t_88) < 0) __PYX_ERR(0, 5657, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_88); __pyx_t_88 = 0;
  __pyx_t_88 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Connect_clauseContext_11accept, 0, __pyx_n_s_ModelicaParser_Connect_clauseCon_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2228)); if (unlikely(!__pyx_t_88)) __PYX_ERR(0, 5657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_88);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_88, __pyx_t_87);
  __Pyx_DECREF(__pyx_t_87); __pyx_t_87 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_88) < 0) __PYX_ERR(0, 5657, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_88); __pyx_t_88 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5633
 *         return localctx
 * 
 *     class Connect_clauseContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_88 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Connect_clauseContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_88)) __PYX_ERR(0, 5633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_88);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_86, __pyx_t_88) < 0) __PYX_ERR(0, 5633, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_86); __pyx_t_86 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Connect_clauseContext, __pyx_t_88) < 0) __PYX_ERR(0, 5633, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_88); __pyx_t_88 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5666
 * 
 * 
 *     def connect_clause(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Connect_clauseContext(self, self._ctx, self.state)
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_109connect_clause, 0, __pyx_n_s_ModelicaParser_connect_clause, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2230)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_connect_clause, __pyx_t_5) < 0) __PYX_ERR(0, 5666, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5692
 *         return localctx
 * 
 *     class ExpressionContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_ExpressionContext, __pyx_n_s_ModelicaParser_ExpressionContext, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_88 = PyList_New(0); if (unlikely(!__pyx_t_88)) __PYX_ERR(0, 5692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_88);

  /* "pymola/generated/ModelicaParser.pyx":5694
 *     class ExpressionContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_87 = PyDict_New(); if (unlikely(!__pyx_t_87)) __PYX_ERR(0, 5694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_87);
  __pyx_t_89 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_89)) __PYX_ERR(0, 5694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_89);
  if (PyDict_SetItem(__pyx_t_87, __pyx_n_s_parent, __pyx_t_89) < 0) __PYX_ERR(0, 5694, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_89); __pyx_t_89 = 0;
  if (PyDict_SetItem(__pyx_t_87, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 5694, __pyx_L1_error)
  __pyx_t_89 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_17ExpressionContext_1__init__, 0, __pyx_n_s_ModelicaParser_ExpressionContext_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2232)); if (unlikely(!__pyx_t_89)) __PYX_ERR(0, 5694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_89);
  __Pyx_INCREF(__pyx_t_89);
  PyList_Append(__pyx_t_88, __pyx_t_89);
  __Pyx_GIVEREF(__pyx_t_89);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_89, __pyx_tuple__2233);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_89, __pyx_t_87);
  __Pyx_DECREF(__pyx_t_87); __pyx_t_87 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_89) < 0) __PYX_ERR(0, 5694, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_89); __pyx_t_89 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5699
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_expression
 * 
 */
  __pyx_t_89 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_17ExpressionContext_3getRuleIndex, 0, __pyx_n_s_ModelicaParser_ExpressionContext_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2235)); if (unlikely(!__pyx_t_89)) __PYX_ERR(0, 5699, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_89);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_89) < 0) __PYX_ERR(0, 5699, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_89); __pyx_t_89 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5703
 * 
 * 
 *         def copyFrom(self, ctx:ParserRuleContext):             # <<<<<<<<<<<<<<
 *             super().copyFrom(ctx)
 * 
 */
  __pyx_t_89 = PyDict_New(); if (unlikely(!__pyx_t_89)) __PYX_ERR(0, 5703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_89);
  __pyx_t_87 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_87)) __PYX_ERR(0, 5703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_87);
  if (PyDict_SetItem(__pyx_t_89, __pyx_n_s_ctx_2, __pyx_t_87) < 0) __PYX_ERR(0, 5703, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_87); __pyx_t_87 = 0;
  __pyx_t_87 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_17ExpressionContext_5copyFrom, 0, __pyx_n_s_ModelicaParser_ExpressionContext_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2237)); if (unlikely(!__pyx_t_87)) __PYX_ERR(0, 5703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_87);
  __Pyx_INCREF(__pyx_t_87);
  PyList_Append(__pyx_t_88, __pyx_t_87);
  __Pyx_GIVEREF(__pyx_t_87);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_87, __pyx_t_89);
  __Pyx_DECREF(__pyx_t_89); __pyx_t_89 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_copyFrom, __pyx_t_87) < 0) __PYX_ERR(0, 5703, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_87); __pyx_t_87 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5692
 *         return localctx
 * 
 *     class ExpressionContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_87 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_ExpressionContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_87)) __PYX_ERR(0, 5692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_87);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_88, __pyx_t_87) < 0) __PYX_ERR(0, 5692, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_88); __pyx_t_88 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_ExpressionContext, __pyx_t_87) < 0) __PYX_ERR(0, 5692, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_87); __pyx_t_87 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5708
 * 
 * 
 *     class Expression_ifContext(ExpressionContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExpressionContext
 */
  __pyx_t_7 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_ExpressionContext);
  if (unlikely(!__pyx_t_7)) {
    PyErr_Clear();
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ExpressionContext);
  }
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Expression_ifContext, __pyx_n_s_ModelicaParser_Expression_ifCont, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_87 = PyList_New(0); if (unlikely(!__pyx_t_87)) __PYX_ERR(0, 5708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_87);

  /* "pymola/generated/ModelicaParser.pyx":5710
 *     class Expression_ifContext(ExpressionContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExpressionContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_89 = PyDict_New(); if (unlikely(!__pyx_t_89)) __PYX_ERR(0, 5710, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_89);
  __pyx_t_90 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_90)) __PYX_ERR(0, 5710, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_90);
  if (PyDict_SetItem(__pyx_t_89, __pyx_n_s_ctx_2, __pyx_t_90) < 0) __PYX_ERR(0, 5710, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_90); __pyx_t_90 = 0;
  __pyx_t_90 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20Expression_ifContext_1__init__, 0, __pyx_n_s_ModelicaParser_Expression_ifCont_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2239)); if (unlikely(!__pyx_t_90)) __PYX_ERR(0, 5710, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_90);
  __Pyx_INCREF(__pyx_t_90);
  PyList_Append(__pyx_t_87, __pyx_t_90);
  __Pyx_GIVEREF(__pyx_t_90);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_90, __pyx_t_89);
  __Pyx_DECREF(__pyx_t_89); __pyx_t_89 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_90) < 0) __PYX_ERR(0, 5710, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_90); __pyx_t_90 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5714
 *             self.copyFrom(ctx)
 * 
 *         def expression(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 */
  __pyx_t_90 = PyDict_New(); if (unlikely(!__pyx_t_90)) __PYX_ERR(0, 5714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_90);
  if (PyDict_SetItem(__pyx_t_90, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 5714, __pyx_L1_error)
  __pyx_t_89 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20Expression_ifContext_3expression, 0, __pyx_n_s_ModelicaParser_Expression_ifCont_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2241)); if (unlikely(!__pyx_t_89)) __PYX_ERR(0, 5714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_89);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_89, __pyx_tuple__2242);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_89, __pyx_t_90);
  __Pyx_DECREF(__pyx_t_90); __pyx_t_90 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_expression, __pyx_t_89) < 0) __PYX_ERR(0, 5714, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_89); __pyx_t_89 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5721
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpression_if" ):
 *                 listener.enterExpression_if(self)
 */
  __pyx_t_89 = PyDict_New(); if (unlikely(!__pyx_t_89)) __PYX_ERR(0, 5721, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_89);
  __pyx_t_90 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_90)) __PYX_ERR(0, 5721, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_90);
  if (PyDict_SetItem(__pyx_t_89, __pyx_n_s_listener, __pyx_t_90) < 0) __PYX_ERR(0, 5721, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_90); __pyx_t_90 = 0;
  __pyx_t_90 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20Expression_ifContext_5enterRule, 0, __pyx_n_s_ModelicaParser_Expression_ifCont_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2244)); if (unlikely(!__pyx_t_90)) __PYX_ERR(0, 5721, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_90);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_90, __pyx_t_89);
  __Pyx_DECREF(__pyx_t_89); __pyx_t_89 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_90) < 0) __PYX_ERR(0, 5721, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_90); __pyx_t_90 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5725
 *                 listener.enterExpression_if(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpression_if" ):
 *                 listener.exitExpression_if(self)
 */
  __pyx_t_90 = PyDict_New(); if (unlikely(!__pyx_t_90)) __PYX_ERR(0, 5725, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_90);
  __pyx_t_89 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_89)) __PYX_ERR(0, 5725, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_89);
  if (PyDict_SetItem(__pyx_t_90, __pyx_n_s_listener, __pyx_t_89) < 0) __PYX_ERR(0, 5725, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_89); __pyx_t_89 = 0;
  __pyx_t_89 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20Expression_ifContext_7exitRule, 0, __pyx_n_s_ModelicaParser_Expression_ifCont_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2246)); if (unlikely(!__pyx_t_89)) __PYX_ERR(0, 5725, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_89);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_89, __pyx_t_90);
  __Pyx_DECREF(__pyx_t_90); __pyx_t_90 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_89) < 0) __PYX_ERR(0, 5725, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_89); __pyx_t_89 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5729
 *                 listener.exitExpression_if(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpression_if" ):
 *                 return visitor.visitExpression_if(self)
 */
  __pyx_t_89 = PyDict_New(); if (unlikely(!__pyx_t_89)) __PYX_ERR(0, 5729, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_89);
  __pyx_t_90 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_90)) __PYX_ERR(0, 5729, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_90);
  if (PyDict_SetItem(__pyx_t_89, __pyx_n_s_visitor, __pyx_t_90) < 0) __PYX_ERR(0, 5729, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_90); __pyx_t_90 = 0;
  __pyx_t_90 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20Expression_ifContext_9accept, 0, __pyx_n_s_ModelicaParser_Expression_ifCont_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2248)); if (unlikely(!__pyx_t_90)) __PYX_ERR(0, 5729, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_90);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_90, __pyx_t_89);
  __Pyx_DECREF(__pyx_t_89); __pyx_t_89 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_90) < 0) __PYX_ERR(0, 5729, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_90); __pyx_t_90 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5708
 * 
 * 
 *     class Expression_ifContext(ExpressionContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExpressionContext
 */
  __pyx_t_90 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Expression_ifContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_90)) __PYX_ERR(0, 5708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_90);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_87, __pyx_t_90) < 0) __PYX_ERR(0, 5708, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_87); __pyx_t_87 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Expression_ifContext, __pyx_t_90) < 0) __PYX_ERR(0, 5708, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_90); __pyx_t_90 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5736
 * 
 * 
 *     class Expression_simpleContext(ExpressionContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExpressionContext
 */
  __pyx_t_5 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_ExpressionContext);
  if (unlikely(!__pyx_t_5)) {
    PyErr_Clear();
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ExpressionContext);
  }
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Expression_simpleContext, __pyx_n_s_ModelicaParser_Expression_simple, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_90 = PyList_New(0); if (unlikely(!__pyx_t_90)) __PYX_ERR(0, 5736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_90);

  /* "pymola/generated/ModelicaParser.pyx":5738
 *     class Expression_simpleContext(ExpressionContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExpressionContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_89 = PyDict_New(); if (unlikely(!__pyx_t_89)) __PYX_ERR(0, 5738, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_89);
  __pyx_t_91 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_91)) __PYX_ERR(0, 5738, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_91);
  if (PyDict_SetItem(__pyx_t_89, __pyx_n_s_ctx_2, __pyx_t_91) < 0) __PYX_ERR(0, 5738, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_91); __pyx_t_91 = 0;
  __pyx_t_91 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Expression_simpleContext_1__init__, 0, __pyx_n_s_ModelicaParser_Expression_simple_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2250)); if (unlikely(!__pyx_t_91)) __PYX_ERR(0, 5738, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_91);
  __Pyx_INCREF(__pyx_t_91);
  PyList_Append(__pyx_t_90, __pyx_t_91);
  __Pyx_GIVEREF(__pyx_t_91);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_91, __pyx_t_89);
  __Pyx_DECREF(__pyx_t_89); __pyx_t_89 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_91) < 0) __PYX_ERR(0, 5738, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_91); __pyx_t_91 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5742
 *             self.copyFrom(ctx)
 * 
 *         def simple_expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Simple_expressionContext,0)
 * 
 */
  __pyx_t_91 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Expression_simpleContext_3simple_expression, 0, __pyx_n_s_ModelicaParser_Expression_simple_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2252)); if (unlikely(!__pyx_t_91)) __PYX_ERR(0, 5742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_91);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_simple_expression, __pyx_t_91) < 0) __PYX_ERR(0, 5742, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_91); __pyx_t_91 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5746
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpression_simple" ):
 *                 listener.enterExpression_simple(self)
 */
  __pyx_t_91 = PyDict_New(); if (unlikely(!__pyx_t_91)) __PYX_ERR(0, 5746, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_91);
  __pyx_t_89 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_89)) __PYX_ERR(0, 5746, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_89);
  if (PyDict_SetItem(__pyx_t_91, __pyx_n_s_listener, __pyx_t_89) < 0) __PYX_ERR(0, 5746, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_89); __pyx_t_89 = 0;
  __pyx_t_89 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Expression_simpleContext_5enterRule, 0, __pyx_n_s_ModelicaParser_Expression_simple_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2254)); if (unlikely(!__pyx_t_89)) __PYX_ERR(0, 5746, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_89);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_89, __pyx_t_91);
  __Pyx_DECREF(__pyx_t_91); __pyx_t_91 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_89) < 0) __PYX_ERR(0, 5746, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_89); __pyx_t_89 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5750
 *                 listener.enterExpression_simple(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpression_simple" ):
 *                 listener.exitExpression_simple(self)
 */
  __pyx_t_89 = PyDict_New(); if (unlikely(!__pyx_t_89)) __PYX_ERR(0, 5750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_89);
  __pyx_t_91 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_91)) __PYX_ERR(0, 5750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_91);
  if (PyDict_SetItem(__pyx_t_89, __pyx_n_s_listener, __pyx_t_91) < 0) __PYX_ERR(0, 5750, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_91); __pyx_t_91 = 0;
  __pyx_t_91 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Expression_simpleContext_7exitRule, 0, __pyx_n_s_ModelicaParser_Expression_simple_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2256)); if (unlikely(!__pyx_t_91)) __PYX_ERR(0, 5750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_91);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_91, __pyx_t_89);
  __Pyx_DECREF(__pyx_t_89); __pyx_t_89 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_91) < 0) __PYX_ERR(0, 5750, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_91); __pyx_t_91 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5754
 *                 listener.exitExpression_simple(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpression_simple" ):
 *                 return visitor.visitExpression_simple(self)
 */
  __pyx_t_91 = PyDict_New(); if (unlikely(!__pyx_t_91)) __PYX_ERR(0, 5754, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_91);
  __pyx_t_89 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_89)) __PYX_ERR(0, 5754, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_89);
  if (PyDict_SetItem(__pyx_t_91, __pyx_n_s_visitor, __pyx_t_89) < 0) __PYX_ERR(0, 5754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_89); __pyx_t_89 = 0;
  __pyx_t_89 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Expression_simpleContext_9accept, 0, __pyx_n_s_ModelicaParser_Expression_simple_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2258)); if (unlikely(!__pyx_t_89)) __PYX_ERR(0, 5754, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_89);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_89, __pyx_t_91);
  __Pyx_DECREF(__pyx_t_91); __pyx_t_91 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_89) < 0) __PYX_ERR(0, 5754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_89); __pyx_t_89 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5736
 * 
 * 
 *     class Expression_simpleContext(ExpressionContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExpressionContext
 */
  __pyx_t_89 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Expression_simpleContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_89)) __PYX_ERR(0, 5736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_89);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_90, __pyx_t_89) < 0) __PYX_ERR(0, 5736, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_90); __pyx_t_90 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Expression_simpleContext, __pyx_t_89) < 0) __PYX_ERR(0, 5736, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_89); __pyx_t_89 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5762
 * 
 * 
 *     def expression(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.ExpressionContext(self, self._ctx, self.state)
 */
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_111expression, 0, __pyx_n_s_ModelicaParser_expression, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2260)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5762, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_expression, __pyx_t_7) < 0) __PYX_ERR(0, 5762, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5820
 *         return localctx
 * 
 *     class Simple_expressionContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Simple_expressionContext, __pyx_n_s_ModelicaParser_Simple_expression, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_89 = PyList_New(0); if (unlikely(!__pyx_t_89)) __PYX_ERR(0, 5820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_89);

  /* "pymola/generated/ModelicaParser.pyx":5822
 *     class Simple_expressionContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_91 = PyDict_New(); if (unlikely(!__pyx_t_91)) __PYX_ERR(0, 5822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_91);
  __pyx_t_92 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_92)) __PYX_ERR(0, 5822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_92);
  if (PyDict_SetItem(__pyx_t_91, __pyx_n_s_parent, __pyx_t_92) < 0) __PYX_ERR(0, 5822, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_92); __pyx_t_92 = 0;
  if (PyDict_SetItem(__pyx_t_91, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 5822, __pyx_L1_error)
  __pyx_t_92 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_1__init__, 0, __pyx_n_s_ModelicaParser_Simple_expression_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2262)); if (unlikely(!__pyx_t_92)) __PYX_ERR(0, 5822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_92);
  __Pyx_INCREF(__pyx_t_92);
  PyList_Append(__pyx_t_89, __pyx_t_92);
  __Pyx_GIVEREF(__pyx_t_92);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_92, __pyx_tuple__2263);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_92, __pyx_t_91);
  __Pyx_DECREF(__pyx_t_91); __pyx_t_91 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_92) < 0) __PYX_ERR(0, 5822, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_92); __pyx_t_92 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5826
 *             self.parser = parser
 * 
 *         def expr(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 */
  __pyx_t_92 = PyDict_New(); if (unlikely(!__pyx_t_92)) __PYX_ERR(0, 5826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_92);
  if (PyDict_SetItem(__pyx_t_92, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 5826, __pyx_L1_error)
  __pyx_t_91 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_3expr, 0, __pyx_n_s_ModelicaParser_Simple_expression_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2265)); if (unlikely(!__pyx_t_91)) __PYX_ERR(0, 5826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_91);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_91, __pyx_tuple__2266);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_91, __pyx_t_92);
  __Pyx_DECREF(__pyx_t_92); __pyx_t_92 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_expr, __pyx_t_91) < 0) __PYX_ERR(0, 5826, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_91); __pyx_t_91 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5833
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_simple_expression
 * 
 */
  __pyx_t_91 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_5getRuleIndex, 0, __pyx_n_s_ModelicaParser_Simple_expression_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2268)); if (unlikely(!__pyx_t_91)) __PYX_ERR(0, 5833, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_91);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_91) < 0) __PYX_ERR(0, 5833, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_91); __pyx_t_91 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5836
 *             return ModelicaParser.RULE_simple_expression
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterSimple_expression" ):
 *                 listener.enterSimple_expression(self)
 */
  __pyx_t_91 = PyDict_New(); if (unlikely(!__pyx_t_91)) __PYX_ERR(0, 5836, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_91);
  __pyx_t_92 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_92)) __PYX_ERR(0, 5836, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_92);
  if (PyDict_SetItem(__pyx_t_91, __pyx_n_s_listener, __pyx_t_92) < 0) __PYX_ERR(0, 5836, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_92); __pyx_t_92 = 0;
  __pyx_t_92 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_7enterRule, 0, __pyx_n_s_ModelicaParser_Simple_expression_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2270)); if (unlikely(!__pyx_t_92)) __PYX_ERR(0, 5836, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_92);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_92, __pyx_t_91);
  __Pyx_DECREF(__pyx_t_91); __pyx_t_91 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_92) < 0) __PYX_ERR(0, 5836, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_92); __pyx_t_92 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5840
 *                 listener.enterSimple_expression(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitSimple_expression" ):
 *                 listener.exitSimple_expression(self)
 */
  __pyx_t_92 = PyDict_New(); if (unlikely(!__pyx_t_92)) __PYX_ERR(0, 5840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_92);
  __pyx_t_91 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_91)) __PYX_ERR(0, 5840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_91);
  if (PyDict_SetItem(__pyx_t_92, __pyx_n_s_listener, __pyx_t_91) < 0) __PYX_ERR(0, 5840, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_91); __pyx_t_91 = 0;
  __pyx_t_91 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_9exitRule, 0, __pyx_n_s_ModelicaParser_Simple_expression_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2272)); if (unlikely(!__pyx_t_91)) __PYX_ERR(0, 5840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_91);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_91, __pyx_t_92);
  __Pyx_DECREF(__pyx_t_92); __pyx_t_92 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_91) < 0) __PYX_ERR(0, 5840, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_91); __pyx_t_91 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5844
 *                 listener.exitSimple_expression(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitSimple_expression" ):
 *                 return visitor.visitSimple_expression(self)
 */
  __pyx_t_91 = PyDict_New(); if (unlikely(!__pyx_t_91)) __PYX_ERR(0, 5844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_91);
  __pyx_t_92 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_92)) __PYX_ERR(0, 5844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_92);
  if (PyDict_SetItem(__pyx_t_91, __pyx_n_s_visitor, __pyx_t_92) < 0) __PYX_ERR(0, 5844, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_92); __pyx_t_92 = 0;
  __pyx_t_92 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Simple_expressionContext_11accept, 0, __pyx_n_s_ModelicaParser_Simple_expression_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2274)); if (unlikely(!__pyx_t_92)) __PYX_ERR(0, 5844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_92);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_92, __pyx_t_91);
  __Pyx_DECREF(__pyx_t_91); __pyx_t_91 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_92) < 0) __PYX_ERR(0, 5844, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_92); __pyx_t_92 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5820
 *         return localctx
 * 
 *     class Simple_expressionContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_92 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Simple_expressionContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_92)) __PYX_ERR(0, 5820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_92);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_89, __pyx_t_92) < 0) __PYX_ERR(0, 5820, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_89); __pyx_t_89 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Simple_expressionContext, __pyx_t_92) < 0) __PYX_ERR(0, 5820, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_92); __pyx_t_92 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5853
 * 
 * 
 *     def simple_expression(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Simple_expressionContext(self, self._ctx, self.state)
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_113simple_expression, 0, __pyx_n_s_ModelicaParser_simple_expression, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2276)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5853, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_simple_expression, __pyx_t_5) < 0) __PYX_ERR(0, 5853, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5890
 *         return localctx
 * 
 *     class ExprContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5890, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5890, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5890, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_ExprContext, __pyx_n_s_ModelicaParser_ExprContext, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5890, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_92 = PyList_New(0); if (unlikely(!__pyx_t_92)) __PYX_ERR(0, 5890, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_92);

  /* "pymola/generated/ModelicaParser.pyx":5892
 *     class ExprContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_91 = PyDict_New(); if (unlikely(!__pyx_t_91)) __PYX_ERR(0, 5892, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_91);
  __pyx_t_93 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_93)) __PYX_ERR(0, 5892, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_93);
  if (PyDict_SetItem(__pyx_t_91, __pyx_n_s_parent, __pyx_t_93) < 0) __PYX_ERR(0, 5892, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_93); __pyx_t_93 = 0;
  if (PyDict_SetItem(__pyx_t_91, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 5892, __pyx_L1_error)
  __pyx_t_93 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_11ExprContext_1__init__, 0, __pyx_n_s_ModelicaParser_ExprContext___ini, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2278)); if (unlikely(!__pyx_t_93)) __PYX_ERR(0, 5892, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_93);
  __Pyx_INCREF(__pyx_t_93);
  PyList_Append(__pyx_t_92, __pyx_t_93);
  __Pyx_GIVEREF(__pyx_t_93);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_93, __pyx_tuple__2279);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_93, __pyx_t_91);
  __Pyx_DECREF(__pyx_t_91); __pyx_t_91 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_93) < 0) __PYX_ERR(0, 5892, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_93); __pyx_t_93 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5897
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_expr
 * 
 */
  __pyx_t_93 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_11ExprContext_3getRuleIndex, 0, __pyx_n_s_ModelicaParser_ExprContext_getRu, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2281)); if (unlikely(!__pyx_t_93)) __PYX_ERR(0, 5897, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_93);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_93) < 0) __PYX_ERR(0, 5897, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_93); __pyx_t_93 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5901
 * 
 * 
 *         def copyFrom(self, ctx:ParserRuleContext):             # <<<<<<<<<<<<<<
 *             super().copyFrom(ctx)
 * 
 */
  __pyx_t_93 = PyDict_New(); if (unlikely(!__pyx_t_93)) __PYX_ERR(0, 5901, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_93);
  __pyx_t_91 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_91)) __PYX_ERR(0, 5901, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_91);
  if (PyDict_SetItem(__pyx_t_93, __pyx_n_s_ctx_2, __pyx_t_91) < 0) __PYX_ERR(0, 5901, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_91); __pyx_t_91 = 0;
  __pyx_t_91 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_11ExprContext_5copyFrom, 0, __pyx_n_s_ModelicaParser_ExprContext_copyF, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2283)); if (unlikely(!__pyx_t_91)) __PYX_ERR(0, 5901, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_91);
  __Pyx_INCREF(__pyx_t_91);
  PyList_Append(__pyx_t_92, __pyx_t_91);
  __Pyx_GIVEREF(__pyx_t_91);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_91, __pyx_t_93);
  __Pyx_DECREF(__pyx_t_93); __pyx_t_93 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_copyFrom, __pyx_t_91) < 0) __PYX_ERR(0, 5901, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_91); __pyx_t_91 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5890
 *         return localctx
 * 
 *     class ExprContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_91 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_ExprContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_91)) __PYX_ERR(0, 5890, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_91);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_92, __pyx_t_91) < 0) __PYX_ERR(0, 5890, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_92); __pyx_t_92 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_ExprContext, __pyx_t_91) < 0) __PYX_ERR(0, 5890, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_91); __pyx_t_91 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5905
 * 
 * 
 *     class Expr_negContext(ExprContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext
 */
  __pyx_t_7 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_ExprContext);
  if (unlikely(!__pyx_t_7)) {
    PyErr_Clear();
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ExprContext);
  }
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Expr_negContext, __pyx_n_s_ModelicaParser_Expr_negContext, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_91 = PyList_New(0); if (unlikely(!__pyx_t_91)) __PYX_ERR(0, 5905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_91);

  /* "pymola/generated/ModelicaParser.pyx":5907
 *     class Expr_negContext(ExprContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.op = None # Token
 */
  __pyx_t_93 = PyDict_New(); if (unlikely(!__pyx_t_93)) __PYX_ERR(0, 5907, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_93);
  __pyx_t_94 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_94)) __PYX_ERR(0, 5907, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_94);
  if (PyDict_SetItem(__pyx_t_93, __pyx_n_s_ctx_2, __pyx_t_94) < 0) __PYX_ERR(0, 5907, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_94); __pyx_t_94 = 0;
  __pyx_t_94 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_negContext_1__init__, 0, __pyx_n_s_ModelicaParser_Expr_negContext_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2285)); if (unlikely(!__pyx_t_94)) __PYX_ERR(0, 5907, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_94);
  __Pyx_INCREF(__pyx_t_94);
  PyList_Append(__pyx_t_91, __pyx_t_94);
  __Pyx_GIVEREF(__pyx_t_94);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_94, __pyx_t_93);
  __Pyx_DECREF(__pyx_t_93); __pyx_t_93 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_94) < 0) __PYX_ERR(0, 5907, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_94); __pyx_t_94 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5912
 *             self.copyFrom(ctx)
 * 
 *         def expr(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExprContext,0)
 * 
 */
  __pyx_t_94 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_negContext_3expr, 0, __pyx_n_s_ModelicaParser_Expr_negContext_e, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2287)); if (unlikely(!__pyx_t_94)) __PYX_ERR(0, 5912, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_94);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_expr, __pyx_t_94) < 0) __PYX_ERR(0, 5912, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_94); __pyx_t_94 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5916
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpr_neg" ):
 *                 listener.enterExpr_neg(self)
 */
  __pyx_t_94 = PyDict_New(); if (unlikely(!__pyx_t_94)) __PYX_ERR(0, 5916, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_94);
  __pyx_t_93 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_93)) __PYX_ERR(0, 5916, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_93);
  if (PyDict_SetItem(__pyx_t_94, __pyx_n_s_listener, __pyx_t_93) < 0) __PYX_ERR(0, 5916, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_93); __pyx_t_93 = 0;
  __pyx_t_93 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_negContext_5enterRule, 0, __pyx_n_s_ModelicaParser_Expr_negContext_e_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2289)); if (unlikely(!__pyx_t_93)) __PYX_ERR(0, 5916, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_93);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_93, __pyx_t_94);
  __Pyx_DECREF(__pyx_t_94); __pyx_t_94 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_93) < 0) __PYX_ERR(0, 5916, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_93); __pyx_t_93 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5920
 *                 listener.enterExpr_neg(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpr_neg" ):
 *                 listener.exitExpr_neg(self)
 */
  __pyx_t_93 = PyDict_New(); if (unlikely(!__pyx_t_93)) __PYX_ERR(0, 5920, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_93);
  __pyx_t_94 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_94)) __PYX_ERR(0, 5920, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_94);
  if (PyDict_SetItem(__pyx_t_93, __pyx_n_s_listener, __pyx_t_94) < 0) __PYX_ERR(0, 5920, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_94); __pyx_t_94 = 0;
  __pyx_t_94 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_negContext_7exitRule, 0, __pyx_n_s_ModelicaParser_Expr_negContext_e_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2291)); if (unlikely(!__pyx_t_94)) __PYX_ERR(0, 5920, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_94);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_94, __pyx_t_93);
  __Pyx_DECREF(__pyx_t_93); __pyx_t_93 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_94) < 0) __PYX_ERR(0, 5920, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_94); __pyx_t_94 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5924
 *                 listener.exitExpr_neg(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpr_neg" ):
 *                 return visitor.visitExpr_neg(self)
 */
  __pyx_t_94 = PyDict_New(); if (unlikely(!__pyx_t_94)) __PYX_ERR(0, 5924, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_94);
  __pyx_t_93 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_93)) __PYX_ERR(0, 5924, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_93);
  if (PyDict_SetItem(__pyx_t_94, __pyx_n_s_visitor, __pyx_t_93) < 0) __PYX_ERR(0, 5924, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_93); __pyx_t_93 = 0;
  __pyx_t_93 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_negContext_9accept, 0, __pyx_n_s_ModelicaParser_Expr_negContext_a, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2293)); if (unlikely(!__pyx_t_93)) __PYX_ERR(0, 5924, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_93);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_93, __pyx_t_94);
  __Pyx_DECREF(__pyx_t_94); __pyx_t_94 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_93) < 0) __PYX_ERR(0, 5924, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_93); __pyx_t_93 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5905
 * 
 * 
 *     class Expr_negContext(ExprContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext
 */
  __pyx_t_93 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Expr_negContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_93)) __PYX_ERR(0, 5905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_93);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_91, __pyx_t_93) < 0) __PYX_ERR(0, 5905, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_91); __pyx_t_91 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Expr_negContext, __pyx_t_93) < 0) __PYX_ERR(0, 5905, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_93); __pyx_t_93 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5931
 * 
 * 
 *     class Expr_addContext(ExprContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext
 */
  __pyx_t_5 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_ExprContext);
  if (unlikely(!__pyx_t_5)) {
    PyErr_Clear();
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ExprContext);
  }
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5931, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5931, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5931, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Expr_addContext, __pyx_n_s_ModelicaParser_Expr_addContext, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5931, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_93 = PyList_New(0); if (unlikely(!__pyx_t_93)) __PYX_ERR(0, 5931, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_93);

  /* "pymola/generated/ModelicaParser.pyx":5933
 *     class Expr_addContext(ExprContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.op = None # Token
 */
  __pyx_t_94 = PyDict_New(); if (unlikely(!__pyx_t_94)) __PYX_ERR(0, 5933, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_94);
  __pyx_t_95 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_95)) __PYX_ERR(0, 5933, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_95);
  if (PyDict_SetItem(__pyx_t_94, __pyx_n_s_ctx_2, __pyx_t_95) < 0) __PYX_ERR(0, 5933, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_95); __pyx_t_95 = 0;
  __pyx_t_95 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_addContext_1__init__, 0, __pyx_n_s_ModelicaParser_Expr_addContext_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2295)); if (unlikely(!__pyx_t_95)) __PYX_ERR(0, 5933, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_95);
  __Pyx_INCREF(__pyx_t_95);
  PyList_Append(__pyx_t_93, __pyx_t_95);
  __Pyx_GIVEREF(__pyx_t_95);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_95, __pyx_t_94);
  __Pyx_DECREF(__pyx_t_94); __pyx_t_94 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_95) < 0) __PYX_ERR(0, 5933, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_95); __pyx_t_95 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5938
 *             self.copyFrom(ctx)
 * 
 *         def expr(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 */
  __pyx_t_95 = PyDict_New(); if (unlikely(!__pyx_t_95)) __PYX_ERR(0, 5938, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_95);
  if (PyDict_SetItem(__pyx_t_95, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 5938, __pyx_L1_error)
  __pyx_t_94 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_addContext_3expr, 0, __pyx_n_s_ModelicaParser_Expr_addContext_e, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2297)); if (unlikely(!__pyx_t_94)) __PYX_ERR(0, 5938, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_94);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_94, __pyx_tuple__2298);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_94, __pyx_t_95);
  __Pyx_DECREF(__pyx_t_95); __pyx_t_95 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_expr, __pyx_t_94) < 0) __PYX_ERR(0, 5938, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_94); __pyx_t_94 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5945
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpr_add" ):
 *                 listener.enterExpr_add(self)
 */
  __pyx_t_94 = PyDict_New(); if (unlikely(!__pyx_t_94)) __PYX_ERR(0, 5945, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_94);
  __pyx_t_95 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_95)) __PYX_ERR(0, 5945, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_95);
  if (PyDict_SetItem(__pyx_t_94, __pyx_n_s_listener, __pyx_t_95) < 0) __PYX_ERR(0, 5945, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_95); __pyx_t_95 = 0;
  __pyx_t_95 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_addContext_5enterRule, 0, __pyx_n_s_ModelicaParser_Expr_addContext_e_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2300)); if (unlikely(!__pyx_t_95)) __PYX_ERR(0, 5945, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_95);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_95, __pyx_t_94);
  __Pyx_DECREF(__pyx_t_94); __pyx_t_94 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_95) < 0) __PYX_ERR(0, 5945, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_95); __pyx_t_95 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5949
 *                 listener.enterExpr_add(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpr_add" ):
 *                 listener.exitExpr_add(self)
 */
  __pyx_t_95 = PyDict_New(); if (unlikely(!__pyx_t_95)) __PYX_ERR(0, 5949, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_95);
  __pyx_t_94 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_94)) __PYX_ERR(0, 5949, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_94);
  if (PyDict_SetItem(__pyx_t_95, __pyx_n_s_listener, __pyx_t_94) < 0) __PYX_ERR(0, 5949, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_94); __pyx_t_94 = 0;
  __pyx_t_94 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_addContext_7exitRule, 0, __pyx_n_s_ModelicaParser_Expr_addContext_e_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2302)); if (unlikely(!__pyx_t_94)) __PYX_ERR(0, 5949, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_94);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_94, __pyx_t_95);
  __Pyx_DECREF(__pyx_t_95); __pyx_t_95 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_94) < 0) __PYX_ERR(0, 5949, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_94); __pyx_t_94 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5953
 *                 listener.exitExpr_add(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpr_add" ):
 *                 return visitor.visitExpr_add(self)
 */
  __pyx_t_94 = PyDict_New(); if (unlikely(!__pyx_t_94)) __PYX_ERR(0, 5953, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_94);
  __pyx_t_95 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_95)) __PYX_ERR(0, 5953, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_95);
  if (PyDict_SetItem(__pyx_t_94, __pyx_n_s_visitor, __pyx_t_95) < 0) __PYX_ERR(0, 5953, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_95); __pyx_t_95 = 0;
  __pyx_t_95 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_addContext_9accept, 0, __pyx_n_s_ModelicaParser_Expr_addContext_a, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2304)); if (unlikely(!__pyx_t_95)) __PYX_ERR(0, 5953, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_95);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_95, __pyx_t_94);
  __Pyx_DECREF(__pyx_t_94); __pyx_t_94 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_95) < 0) __PYX_ERR(0, 5953, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_95); __pyx_t_95 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5931
 * 
 * 
 *     class Expr_addContext(ExprContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext
 */
  __pyx_t_95 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Expr_addContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_95)) __PYX_ERR(0, 5931, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_95);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_93, __pyx_t_95) < 0) __PYX_ERR(0, 5931, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_93); __pyx_t_93 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Expr_addContext, __pyx_t_95) < 0) __PYX_ERR(0, 5931, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_95); __pyx_t_95 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5960
 * 
 * 
 *     class Expr_expContext(ExprContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext
 */
  __pyx_t_7 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_ExprContext);
  if (unlikely(!__pyx_t_7)) {
    PyErr_Clear();
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ExprContext);
  }
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5960, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5960, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5960, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Expr_expContext, __pyx_n_s_ModelicaParser_Expr_expContext, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5960, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_95 = PyList_New(0); if (unlikely(!__pyx_t_95)) __PYX_ERR(0, 5960, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_95);

  /* "pymola/generated/ModelicaParser.pyx":5962
 *     class Expr_expContext(ExprContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.op = None # Token
 */
  __pyx_t_94 = PyDict_New(); if (unlikely(!__pyx_t_94)) __PYX_ERR(0, 5962, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_94);
  __pyx_t_96 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_96)) __PYX_ERR(0, 5962, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_96);
  if (PyDict_SetItem(__pyx_t_94, __pyx_n_s_ctx_2, __pyx_t_96) < 0) __PYX_ERR(0, 5962, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_96); __pyx_t_96 = 0;
  __pyx_t_96 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_expContext_1__init__, 0, __pyx_n_s_ModelicaParser_Expr_expContext_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2306)); if (unlikely(!__pyx_t_96)) __PYX_ERR(0, 5962, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_96);
  __Pyx_INCREF(__pyx_t_96);
  PyList_Append(__pyx_t_95, __pyx_t_96);
  __Pyx_GIVEREF(__pyx_t_96);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_96, __pyx_t_94);
  __Pyx_DECREF(__pyx_t_94); __pyx_t_94 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_96) < 0) __PYX_ERR(0, 5962, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_96); __pyx_t_96 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5967
 *             self.copyFrom(ctx)
 * 
 *         def primary(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.PrimaryContext)
 */
  __pyx_t_96 = PyDict_New(); if (unlikely(!__pyx_t_96)) __PYX_ERR(0, 5967, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_96);
  if (PyDict_SetItem(__pyx_t_96, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 5967, __pyx_L1_error)
  __pyx_t_94 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_expContext_3primary, 0, __pyx_n_s_ModelicaParser_Expr_expContext_p, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2308)); if (unlikely(!__pyx_t_94)) __PYX_ERR(0, 5967, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_94);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_94, __pyx_tuple__2309);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_94, __pyx_t_96);
  __Pyx_DECREF(__pyx_t_96); __pyx_t_96 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_primary, __pyx_t_94) < 0) __PYX_ERR(0, 5967, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_94); __pyx_t_94 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5974
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpr_exp" ):
 *                 listener.enterExpr_exp(self)
 */
  __pyx_t_94 = PyDict_New(); if (unlikely(!__pyx_t_94)) __PYX_ERR(0, 5974, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_94);
  __pyx_t_96 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_96)) __PYX_ERR(0, 5974, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_96);
  if (PyDict_SetItem(__pyx_t_94, __pyx_n_s_listener, __pyx_t_96) < 0) __PYX_ERR(0, 5974, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_96); __pyx_t_96 = 0;
  __pyx_t_96 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_expContext_5enterRule, 0, __pyx_n_s_ModelicaParser_Expr_expContext_e, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2311)); if (unlikely(!__pyx_t_96)) __PYX_ERR(0, 5974, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_96);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_96, __pyx_t_94);
  __Pyx_DECREF(__pyx_t_94); __pyx_t_94 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_96) < 0) __PYX_ERR(0, 5974, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_96); __pyx_t_96 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5978
 *                 listener.enterExpr_exp(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpr_exp" ):
 *                 listener.exitExpr_exp(self)
 */
  __pyx_t_96 = PyDict_New(); if (unlikely(!__pyx_t_96)) __PYX_ERR(0, 5978, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_96);
  __pyx_t_94 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_94)) __PYX_ERR(0, 5978, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_94);
  if (PyDict_SetItem(__pyx_t_96, __pyx_n_s_listener, __pyx_t_94) < 0) __PYX_ERR(0, 5978, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_94); __pyx_t_94 = 0;
  __pyx_t_94 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_expContext_7exitRule, 0, __pyx_n_s_ModelicaParser_Expr_expContext_e_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2313)); if (unlikely(!__pyx_t_94)) __PYX_ERR(0, 5978, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_94);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_94, __pyx_t_96);
  __Pyx_DECREF(__pyx_t_96); __pyx_t_96 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_94) < 0) __PYX_ERR(0, 5978, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_94); __pyx_t_94 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5982
 *                 listener.exitExpr_exp(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpr_exp" ):
 *                 return visitor.visitExpr_exp(self)
 */
  __pyx_t_94 = PyDict_New(); if (unlikely(!__pyx_t_94)) __PYX_ERR(0, 5982, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_94);
  __pyx_t_96 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_96)) __PYX_ERR(0, 5982, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_96);
  if (PyDict_SetItem(__pyx_t_94, __pyx_n_s_visitor, __pyx_t_96) < 0) __PYX_ERR(0, 5982, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_96); __pyx_t_96 = 0;
  __pyx_t_96 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_expContext_9accept, 0, __pyx_n_s_ModelicaParser_Expr_expContext_a, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2315)); if (unlikely(!__pyx_t_96)) __PYX_ERR(0, 5982, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_96);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_96, __pyx_t_94);
  __Pyx_DECREF(__pyx_t_94); __pyx_t_94 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_96) < 0) __PYX_ERR(0, 5982, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_96); __pyx_t_96 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5960
 * 
 * 
 *     class Expr_expContext(ExprContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext
 */
  __pyx_t_96 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Expr_expContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_96)) __PYX_ERR(0, 5960, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_96);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_95, __pyx_t_96) < 0) __PYX_ERR(0, 5960, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_95); __pyx_t_95 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Expr_expContext, __pyx_t_96) < 0) __PYX_ERR(0, 5960, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_96); __pyx_t_96 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5989
 * 
 * 
 *     class Expr_orContext(ExprContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext
 */
  __pyx_t_5 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_ExprContext);
  if (unlikely(!__pyx_t_5)) {
    PyErr_Clear();
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ExprContext);
  }
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5989, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5989, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5989, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Expr_orContext, __pyx_n_s_ModelicaParser_Expr_orContext, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5989, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_96 = PyList_New(0); if (unlikely(!__pyx_t_96)) __PYX_ERR(0, 5989, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_96);

  /* "pymola/generated/ModelicaParser.pyx":5991
 *     class Expr_orContext(ExprContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_94 = PyDict_New(); if (unlikely(!__pyx_t_94)) __PYX_ERR(0, 5991, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_94);
  __pyx_t_97 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_97)) __PYX_ERR(0, 5991, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_97);
  if (PyDict_SetItem(__pyx_t_94, __pyx_n_s_ctx_2, __pyx_t_97) < 0) __PYX_ERR(0, 5991, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_97); __pyx_t_97 = 0;
  __pyx_t_97 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14Expr_orContext_1__init__, 0, __pyx_n_s_ModelicaParser_Expr_orContext_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2317)); if (unlikely(!__pyx_t_97)) __PYX_ERR(0, 5991, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_97);
  __Pyx_INCREF(__pyx_t_97);
  PyList_Append(__pyx_t_96, __pyx_t_97);
  __Pyx_GIVEREF(__pyx_t_97);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_97, __pyx_t_94);
  __Pyx_DECREF(__pyx_t_94); __pyx_t_94 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_97) < 0) __PYX_ERR(0, 5991, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_97); __pyx_t_97 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5995
 *             self.copyFrom(ctx)
 * 
 *         def expr(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 */
  __pyx_t_97 = PyDict_New(); if (unlikely(!__pyx_t_97)) __PYX_ERR(0, 5995, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_97);
  if (PyDict_SetItem(__pyx_t_97, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 5995, __pyx_L1_error)
  __pyx_t_94 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14Expr_orContext_3expr, 0, __pyx_n_s_ModelicaParser_Expr_orContext_ex, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2319)); if (unlikely(!__pyx_t_94)) __PYX_ERR(0, 5995, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_94);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_94, __pyx_tuple__2320);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_94, __pyx_t_97);
  __Pyx_DECREF(__pyx_t_97); __pyx_t_97 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_expr, __pyx_t_94) < 0) __PYX_ERR(0, 5995, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_94); __pyx_t_94 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6002
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpr_or" ):
 *                 listener.enterExpr_or(self)
 */
  __pyx_t_94 = PyDict_New(); if (unlikely(!__pyx_t_94)) __PYX_ERR(0, 6002, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_94);
  __pyx_t_97 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_97)) __PYX_ERR(0, 6002, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_97);
  if (PyDict_SetItem(__pyx_t_94, __pyx_n_s_listener, __pyx_t_97) < 0) __PYX_ERR(0, 6002, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_97); __pyx_t_97 = 0;
  __pyx_t_97 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14Expr_orContext_5enterRule, 0, __pyx_n_s_ModelicaParser_Expr_orContext_en, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2322)); if (unlikely(!__pyx_t_97)) __PYX_ERR(0, 6002, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_97);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_97, __pyx_t_94);
  __Pyx_DECREF(__pyx_t_94); __pyx_t_94 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_97) < 0) __PYX_ERR(0, 6002, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_97); __pyx_t_97 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6006
 *                 listener.enterExpr_or(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpr_or" ):
 *                 listener.exitExpr_or(self)
 */
  __pyx_t_97 = PyDict_New(); if (unlikely(!__pyx_t_97)) __PYX_ERR(0, 6006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_97);
  __pyx_t_94 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_94)) __PYX_ERR(0, 6006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_94);
  if (PyDict_SetItem(__pyx_t_97, __pyx_n_s_listener, __pyx_t_94) < 0) __PYX_ERR(0, 6006, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_94); __pyx_t_94 = 0;
  __pyx_t_94 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14Expr_orContext_7exitRule, 0, __pyx_n_s_ModelicaParser_Expr_orContext_ex_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2324)); if (unlikely(!__pyx_t_94)) __PYX_ERR(0, 6006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_94);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_94, __pyx_t_97);
  __Pyx_DECREF(__pyx_t_97); __pyx_t_97 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_94) < 0) __PYX_ERR(0, 6006, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_94); __pyx_t_94 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6010
 *                 listener.exitExpr_or(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpr_or" ):
 *                 return visitor.visitExpr_or(self)
 */
  __pyx_t_94 = PyDict_New(); if (unlikely(!__pyx_t_94)) __PYX_ERR(0, 6010, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_94);
  __pyx_t_97 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_97)) __PYX_ERR(0, 6010, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_97);
  if (PyDict_SetItem(__pyx_t_94, __pyx_n_s_visitor, __pyx_t_97) < 0) __PYX_ERR(0, 6010, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_97); __pyx_t_97 = 0;
  __pyx_t_97 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14Expr_orContext_9accept, 0, __pyx_n_s_ModelicaParser_Expr_orContext_ac, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2326)); if (unlikely(!__pyx_t_97)) __PYX_ERR(0, 6010, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_97);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_97, __pyx_t_94);
  __Pyx_DECREF(__pyx_t_94); __pyx_t_94 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_97) < 0) __PYX_ERR(0, 6010, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_97); __pyx_t_97 = 0;

  /* "pymola/generated/ModelicaParser.pyx":5989
 * 
 * 
 *     class Expr_orContext(ExprContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext
 */
  __pyx_t_97 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Expr_orContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_97)) __PYX_ERR(0, 5989, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_97);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_96, __pyx_t_97) < 0) __PYX_ERR(0, 5989, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_96); __pyx_t_96 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Expr_orContext, __pyx_t_97) < 0) __PYX_ERR(0, 5989, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_97); __pyx_t_97 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6017
 * 
 * 
 *     class Expr_primaryContext(ExprContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext
 */
  __pyx_t_7 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_ExprContext);
  if (unlikely(!__pyx_t_7)) {
    PyErr_Clear();
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ExprContext);
  }
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6017, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6017, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6017, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Expr_primaryContext, __pyx_n_s_ModelicaParser_Expr_primaryConte, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6017, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_97 = PyList_New(0); if (unlikely(!__pyx_t_97)) __PYX_ERR(0, 6017, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_97);

  /* "pymola/generated/ModelicaParser.pyx":6019
 *     class Expr_primaryContext(ExprContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_94 = PyDict_New(); if (unlikely(!__pyx_t_94)) __PYX_ERR(0, 6019, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_94);
  __pyx_t_98 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_98)) __PYX_ERR(0, 6019, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_98);
  if (PyDict_SetItem(__pyx_t_94, __pyx_n_s_ctx_2, __pyx_t_98) < 0) __PYX_ERR(0, 6019, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_98); __pyx_t_98 = 0;
  __pyx_t_98 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19Expr_primaryContext_1__init__, 0, __pyx_n_s_ModelicaParser_Expr_primaryConte_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2328)); if (unlikely(!__pyx_t_98)) __PYX_ERR(0, 6019, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_98);
  __Pyx_INCREF(__pyx_t_98);
  PyList_Append(__pyx_t_97, __pyx_t_98);
  __Pyx_GIVEREF(__pyx_t_98);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_98, __pyx_t_94);
  __Pyx_DECREF(__pyx_t_94); __pyx_t_94 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_98) < 0) __PYX_ERR(0, 6019, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_98); __pyx_t_98 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6023
 *             self.copyFrom(ctx)
 * 
 *         def primary(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.PrimaryContext,0)
 * 
 */
  __pyx_t_98 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19Expr_primaryContext_3primary, 0, __pyx_n_s_ModelicaParser_Expr_primaryConte_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2330)); if (unlikely(!__pyx_t_98)) __PYX_ERR(0, 6023, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_98);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_primary, __pyx_t_98) < 0) __PYX_ERR(0, 6023, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_98); __pyx_t_98 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6027
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpr_primary" ):
 *                 listener.enterExpr_primary(self)
 */
  __pyx_t_98 = PyDict_New(); if (unlikely(!__pyx_t_98)) __PYX_ERR(0, 6027, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_98);
  __pyx_t_94 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_94)) __PYX_ERR(0, 6027, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_94);
  if (PyDict_SetItem(__pyx_t_98, __pyx_n_s_listener, __pyx_t_94) < 0) __PYX_ERR(0, 6027, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_94); __pyx_t_94 = 0;
  __pyx_t_94 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19Expr_primaryContext_5enterRule, 0, __pyx_n_s_ModelicaParser_Expr_primaryConte_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2332)); if (unlikely(!__pyx_t_94)) __PYX_ERR(0, 6027, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_94);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_94, __pyx_t_98);
  __Pyx_DECREF(__pyx_t_98); __pyx_t_98 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_94) < 0) __PYX_ERR(0, 6027, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_94); __pyx_t_94 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6031
 *                 listener.enterExpr_primary(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpr_primary" ):
 *                 listener.exitExpr_primary(self)
 */
  __pyx_t_94 = PyDict_New(); if (unlikely(!__pyx_t_94)) __PYX_ERR(0, 6031, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_94);
  __pyx_t_98 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_98)) __PYX_ERR(0, 6031, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_98);
  if (PyDict_SetItem(__pyx_t_94, __pyx_n_s_listener, __pyx_t_98) < 0) __PYX_ERR(0, 6031, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_98); __pyx_t_98 = 0;
  __pyx_t_98 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19Expr_primaryContext_7exitRule, 0, __pyx_n_s_ModelicaParser_Expr_primaryConte_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2334)); if (unlikely(!__pyx_t_98)) __PYX_ERR(0, 6031, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_98);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_98, __pyx_t_94);
  __Pyx_DECREF(__pyx_t_94); __pyx_t_94 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_98) < 0) __PYX_ERR(0, 6031, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_98); __pyx_t_98 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6035
 *                 listener.exitExpr_primary(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpr_primary" ):
 *                 return visitor.visitExpr_primary(self)
 */
  __pyx_t_98 = PyDict_New(); if (unlikely(!__pyx_t_98)) __PYX_ERR(0, 6035, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_98);
  __pyx_t_94 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_94)) __PYX_ERR(0, 6035, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_94);
  if (PyDict_SetItem(__pyx_t_98, __pyx_n_s_visitor, __pyx_t_94) < 0) __PYX_ERR(0, 6035, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_94); __pyx_t_94 = 0;
  __pyx_t_94 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19Expr_primaryContext_9accept, 0, __pyx_n_s_ModelicaParser_Expr_primaryConte_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2336)); if (unlikely(!__pyx_t_94)) __PYX_ERR(0, 6035, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_94);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_94, __pyx_t_98);
  __Pyx_DECREF(__pyx_t_98); __pyx_t_98 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_94) < 0) __PYX_ERR(0, 6035, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_94); __pyx_t_94 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6017
 * 
 * 
 *     class Expr_primaryContext(ExprContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext
 */
  __pyx_t_94 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Expr_primaryContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_94)) __PYX_ERR(0, 6017, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_94);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_97, __pyx_t_94) < 0) __PYX_ERR(0, 6017, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_97); __pyx_t_97 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Expr_primaryContext, __pyx_t_94) < 0) __PYX_ERR(0, 6017, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_94); __pyx_t_94 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6042
 * 
 * 
 *     class Expr_andContext(ExprContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext
 */
  __pyx_t_5 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_ExprContext);
  if (unlikely(!__pyx_t_5)) {
    PyErr_Clear();
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ExprContext);
  }
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6042, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6042, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6042, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Expr_andContext, __pyx_n_s_ModelicaParser_Expr_andContext, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6042, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_94 = PyList_New(0); if (unlikely(!__pyx_t_94)) __PYX_ERR(0, 6042, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_94);

  /* "pymola/generated/ModelicaParser.pyx":6044
 *     class Expr_andContext(ExprContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_98 = PyDict_New(); if (unlikely(!__pyx_t_98)) __PYX_ERR(0, 6044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_98);
  __pyx_t_99 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_99)) __PYX_ERR(0, 6044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_99);
  if (PyDict_SetItem(__pyx_t_98, __pyx_n_s_ctx_2, __pyx_t_99) < 0) __PYX_ERR(0, 6044, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_99); __pyx_t_99 = 0;
  __pyx_t_99 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_andContext_1__init__, 0, __pyx_n_s_ModelicaParser_Expr_andContext_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2338)); if (unlikely(!__pyx_t_99)) __PYX_ERR(0, 6044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_99);
  __Pyx_INCREF(__pyx_t_99);
  PyList_Append(__pyx_t_94, __pyx_t_99);
  __Pyx_GIVEREF(__pyx_t_99);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_99, __pyx_t_98);
  __Pyx_DECREF(__pyx_t_98); __pyx_t_98 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_99) < 0) __PYX_ERR(0, 6044, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_99); __pyx_t_99 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6048
 *             self.copyFrom(ctx)
 * 
 *         def expr(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 */
  __pyx_t_99 = PyDict_New(); if (unlikely(!__pyx_t_99)) __PYX_ERR(0, 6048, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_99);
  if (PyDict_SetItem(__pyx_t_99, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 6048, __pyx_L1_error)
  __pyx_t_98 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_andContext_3expr, 0, __pyx_n_s_ModelicaParser_Expr_andContext_e, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2340)); if (unlikely(!__pyx_t_98)) __PYX_ERR(0, 6048, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_98);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_98, __pyx_tuple__2341);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_98, __pyx_t_99);
  __Pyx_DECREF(__pyx_t_99); __pyx_t_99 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_expr, __pyx_t_98) < 0) __PYX_ERR(0, 6048, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_98); __pyx_t_98 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6055
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpr_and" ):
 *                 listener.enterExpr_and(self)
 */
  __pyx_t_98 = PyDict_New(); if (unlikely(!__pyx_t_98)) __PYX_ERR(0, 6055, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_98);
  __pyx_t_99 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_99)) __PYX_ERR(0, 6055, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_99);
  if (PyDict_SetItem(__pyx_t_98, __pyx_n_s_listener, __pyx_t_99) < 0) __PYX_ERR(0, 6055, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_99); __pyx_t_99 = 0;
  __pyx_t_99 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_andContext_5enterRule, 0, __pyx_n_s_ModelicaParser_Expr_andContext_e_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2343)); if (unlikely(!__pyx_t_99)) __PYX_ERR(0, 6055, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_99);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_99, __pyx_t_98);
  __Pyx_DECREF(__pyx_t_98); __pyx_t_98 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_99) < 0) __PYX_ERR(0, 6055, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_99); __pyx_t_99 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6059
 *                 listener.enterExpr_and(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpr_and" ):
 *                 listener.exitExpr_and(self)
 */
  __pyx_t_99 = PyDict_New(); if (unlikely(!__pyx_t_99)) __PYX_ERR(0, 6059, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_99);
  __pyx_t_98 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_98)) __PYX_ERR(0, 6059, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_98);
  if (PyDict_SetItem(__pyx_t_99, __pyx_n_s_listener, __pyx_t_98) < 0) __PYX_ERR(0, 6059, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_98); __pyx_t_98 = 0;
  __pyx_t_98 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_andContext_7exitRule, 0, __pyx_n_s_ModelicaParser_Expr_andContext_e_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2345)); if (unlikely(!__pyx_t_98)) __PYX_ERR(0, 6059, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_98);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_98, __pyx_t_99);
  __Pyx_DECREF(__pyx_t_99); __pyx_t_99 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_98) < 0) __PYX_ERR(0, 6059, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_98); __pyx_t_98 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6063
 *                 listener.exitExpr_and(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpr_and" ):
 *                 return visitor.visitExpr_and(self)
 */
  __pyx_t_98 = PyDict_New(); if (unlikely(!__pyx_t_98)) __PYX_ERR(0, 6063, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_98);
  __pyx_t_99 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_99)) __PYX_ERR(0, 6063, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_99);
  if (PyDict_SetItem(__pyx_t_98, __pyx_n_s_visitor, __pyx_t_99) < 0) __PYX_ERR(0, 6063, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_99); __pyx_t_99 = 0;
  __pyx_t_99 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_andContext_9accept, 0, __pyx_n_s_ModelicaParser_Expr_andContext_a, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2347)); if (unlikely(!__pyx_t_99)) __PYX_ERR(0, 6063, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_99);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_99, __pyx_t_98);
  __Pyx_DECREF(__pyx_t_98); __pyx_t_98 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_99) < 0) __PYX_ERR(0, 6063, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_99); __pyx_t_99 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6042
 * 
 * 
 *     class Expr_andContext(ExprContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext
 */
  __pyx_t_99 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Expr_andContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_99)) __PYX_ERR(0, 6042, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_99);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_94, __pyx_t_99) < 0) __PYX_ERR(0, 6042, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_94); __pyx_t_94 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Expr_andContext, __pyx_t_99) < 0) __PYX_ERR(0, 6042, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_99); __pyx_t_99 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6070
 * 
 * 
 *     class Expr_relContext(ExprContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext
 */
  __pyx_t_7 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_ExprContext);
  if (unlikely(!__pyx_t_7)) {
    PyErr_Clear();
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ExprContext);
  }
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6070, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6070, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6070, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Expr_relContext, __pyx_n_s_ModelicaParser_Expr_relContext, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6070, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_99 = PyList_New(0); if (unlikely(!__pyx_t_99)) __PYX_ERR(0, 6070, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_99);

  /* "pymola/generated/ModelicaParser.pyx":6072
 *     class Expr_relContext(ExprContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.op = None # Token
 */
  __pyx_t_98 = PyDict_New(); if (unlikely(!__pyx_t_98)) __PYX_ERR(0, 6072, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_98);
  __pyx_t_100 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_100)) __PYX_ERR(0, 6072, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_100);
  if (PyDict_SetItem(__pyx_t_98, __pyx_n_s_ctx_2, __pyx_t_100) < 0) __PYX_ERR(0, 6072, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_100); __pyx_t_100 = 0;
  __pyx_t_100 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_relContext_1__init__, 0, __pyx_n_s_ModelicaParser_Expr_relContext_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2349)); if (unlikely(!__pyx_t_100)) __PYX_ERR(0, 6072, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_100);
  __Pyx_INCREF(__pyx_t_100);
  PyList_Append(__pyx_t_99, __pyx_t_100);
  __Pyx_GIVEREF(__pyx_t_100);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_100, __pyx_t_98);
  __Pyx_DECREF(__pyx_t_98); __pyx_t_98 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_100) < 0) __PYX_ERR(0, 6072, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_100); __pyx_t_100 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6077
 *             self.copyFrom(ctx)
 * 
 *         def expr(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 */
  __pyx_t_100 = PyDict_New(); if (unlikely(!__pyx_t_100)) __PYX_ERR(0, 6077, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_100);
  if (PyDict_SetItem(__pyx_t_100, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 6077, __pyx_L1_error)
  __pyx_t_98 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_relContext_3expr, 0, __pyx_n_s_ModelicaParser_Expr_relContext_e, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2351)); if (unlikely(!__pyx_t_98)) __PYX_ERR(0, 6077, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_98);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_98, __pyx_tuple__2352);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_98, __pyx_t_100);
  __Pyx_DECREF(__pyx_t_100); __pyx_t_100 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_expr, __pyx_t_98) < 0) __PYX_ERR(0, 6077, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_98); __pyx_t_98 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6084
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpr_rel" ):
 *                 listener.enterExpr_rel(self)
 */
  __pyx_t_98 = PyDict_New(); if (unlikely(!__pyx_t_98)) __PYX_ERR(0, 6084, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_98);
  __pyx_t_100 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_100)) __PYX_ERR(0, 6084, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_100);
  if (PyDict_SetItem(__pyx_t_98, __pyx_n_s_listener, __pyx_t_100) < 0) __PYX_ERR(0, 6084, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_100); __pyx_t_100 = 0;
  __pyx_t_100 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_relContext_5enterRule, 0, __pyx_n_s_ModelicaParser_Expr_relContext_e_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2354)); if (unlikely(!__pyx_t_100)) __PYX_ERR(0, 6084, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_100);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_100, __pyx_t_98);
  __Pyx_DECREF(__pyx_t_98); __pyx_t_98 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_100) < 0) __PYX_ERR(0, 6084, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_100); __pyx_t_100 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6088
 *                 listener.enterExpr_rel(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpr_rel" ):
 *                 listener.exitExpr_rel(self)
 */
  __pyx_t_100 = PyDict_New(); if (unlikely(!__pyx_t_100)) __PYX_ERR(0, 6088, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_100);
  __pyx_t_98 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_98)) __PYX_ERR(0, 6088, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_98);
  if (PyDict_SetItem(__pyx_t_100, __pyx_n_s_listener, __pyx_t_98) < 0) __PYX_ERR(0, 6088, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_98); __pyx_t_98 = 0;
  __pyx_t_98 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_relContext_7exitRule, 0, __pyx_n_s_ModelicaParser_Expr_relContext_e_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2356)); if (unlikely(!__pyx_t_98)) __PYX_ERR(0, 6088, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_98);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_98, __pyx_t_100);
  __Pyx_DECREF(__pyx_t_100); __pyx_t_100 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_98) < 0) __PYX_ERR(0, 6088, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_98); __pyx_t_98 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6092
 *                 listener.exitExpr_rel(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpr_rel" ):
 *                 return visitor.visitExpr_rel(self)
 */
  __pyx_t_98 = PyDict_New(); if (unlikely(!__pyx_t_98)) __PYX_ERR(0, 6092, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_98);
  __pyx_t_100 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_100)) __PYX_ERR(0, 6092, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_100);
  if (PyDict_SetItem(__pyx_t_98, __pyx_n_s_visitor, __pyx_t_100) < 0) __PYX_ERR(0, 6092, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_100); __pyx_t_100 = 0;
  __pyx_t_100 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_relContext_9accept, 0, __pyx_n_s_ModelicaParser_Expr_relContext_a, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2358)); if (unlikely(!__pyx_t_100)) __PYX_ERR(0, 6092, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_100);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_100, __pyx_t_98);
  __Pyx_DECREF(__pyx_t_98); __pyx_t_98 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_100) < 0) __PYX_ERR(0, 6092, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_100); __pyx_t_100 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6070
 * 
 * 
 *     class Expr_relContext(ExprContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext
 */
  __pyx_t_100 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Expr_relContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_100)) __PYX_ERR(0, 6070, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_100);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_99, __pyx_t_100) < 0) __PYX_ERR(0, 6070, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_99); __pyx_t_99 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Expr_relContext, __pyx_t_100) < 0) __PYX_ERR(0, 6070, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_100); __pyx_t_100 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6099
 * 
 * 
 *     class Expr_notContext(ExprContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext
 */
  __pyx_t_5 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_ExprContext);
  if (unlikely(!__pyx_t_5)) {
    PyErr_Clear();
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ExprContext);
  }
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6099, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6099, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6099, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Expr_notContext, __pyx_n_s_ModelicaParser_Expr_notContext, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6099, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_100 = PyList_New(0); if (unlikely(!__pyx_t_100)) __PYX_ERR(0, 6099, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_100);

  /* "pymola/generated/ModelicaParser.pyx":6101
 *     class Expr_notContext(ExprContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_98 = PyDict_New(); if (unlikely(!__pyx_t_98)) __PYX_ERR(0, 6101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_98);
  __pyx_t_101 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_101)) __PYX_ERR(0, 6101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_101);
  if (PyDict_SetItem(__pyx_t_98, __pyx_n_s_ctx_2, __pyx_t_101) < 0) __PYX_ERR(0, 6101, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_101); __pyx_t_101 = 0;
  __pyx_t_101 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_notContext_1__init__, 0, __pyx_n_s_ModelicaParser_Expr_notContext_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2360)); if (unlikely(!__pyx_t_101)) __PYX_ERR(0, 6101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_101);
  __Pyx_INCREF(__pyx_t_101);
  PyList_Append(__pyx_t_100, __pyx_t_101);
  __Pyx_GIVEREF(__pyx_t_101);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_101, __pyx_t_98);
  __Pyx_DECREF(__pyx_t_98); __pyx_t_98 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_101) < 0) __PYX_ERR(0, 6101, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_101); __pyx_t_101 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6105
 *             self.copyFrom(ctx)
 * 
 *         def expr(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExprContext,0)
 * 
 */
  __pyx_t_101 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_notContext_3expr, 0, __pyx_n_s_ModelicaParser_Expr_notContext_e, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2362)); if (unlikely(!__pyx_t_101)) __PYX_ERR(0, 6105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_101);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_expr, __pyx_t_101) < 0) __PYX_ERR(0, 6105, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_101); __pyx_t_101 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6109
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpr_not" ):
 *                 listener.enterExpr_not(self)
 */
  __pyx_t_101 = PyDict_New(); if (unlikely(!__pyx_t_101)) __PYX_ERR(0, 6109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_101);
  __pyx_t_98 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_98)) __PYX_ERR(0, 6109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_98);
  if (PyDict_SetItem(__pyx_t_101, __pyx_n_s_listener, __pyx_t_98) < 0) __PYX_ERR(0, 6109, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_98); __pyx_t_98 = 0;
  __pyx_t_98 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_notContext_5enterRule, 0, __pyx_n_s_ModelicaParser_Expr_notContext_e_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2364)); if (unlikely(!__pyx_t_98)) __PYX_ERR(0, 6109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_98);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_98, __pyx_t_101);
  __Pyx_DECREF(__pyx_t_101); __pyx_t_101 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_98) < 0) __PYX_ERR(0, 6109, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_98); __pyx_t_98 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6113
 *                 listener.enterExpr_not(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpr_not" ):
 *                 listener.exitExpr_not(self)
 */
  __pyx_t_98 = PyDict_New(); if (unlikely(!__pyx_t_98)) __PYX_ERR(0, 6113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_98);
  __pyx_t_101 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_101)) __PYX_ERR(0, 6113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_101);
  if (PyDict_SetItem(__pyx_t_98, __pyx_n_s_listener, __pyx_t_101) < 0) __PYX_ERR(0, 6113, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_101); __pyx_t_101 = 0;
  __pyx_t_101 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_notContext_7exitRule, 0, __pyx_n_s_ModelicaParser_Expr_notContext_e_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2366)); if (unlikely(!__pyx_t_101)) __PYX_ERR(0, 6113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_101);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_101, __pyx_t_98);
  __Pyx_DECREF(__pyx_t_98); __pyx_t_98 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_101) < 0) __PYX_ERR(0, 6113, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_101); __pyx_t_101 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6117
 *                 listener.exitExpr_not(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpr_not" ):
 *                 return visitor.visitExpr_not(self)
 */
  __pyx_t_101 = PyDict_New(); if (unlikely(!__pyx_t_101)) __PYX_ERR(0, 6117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_101);
  __pyx_t_98 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_98)) __PYX_ERR(0, 6117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_98);
  if (PyDict_SetItem(__pyx_t_101, __pyx_n_s_visitor, __pyx_t_98) < 0) __PYX_ERR(0, 6117, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_98); __pyx_t_98 = 0;
  __pyx_t_98 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_notContext_9accept, 0, __pyx_n_s_ModelicaParser_Expr_notContext_a, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2368)); if (unlikely(!__pyx_t_98)) __PYX_ERR(0, 6117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_98);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_98, __pyx_t_101);
  __Pyx_DECREF(__pyx_t_101); __pyx_t_101 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_98) < 0) __PYX_ERR(0, 6117, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_98); __pyx_t_98 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6099
 * 
 * 
 *     class Expr_notContext(ExprContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext
 */
  __pyx_t_98 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Expr_notContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_98)) __PYX_ERR(0, 6099, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_98);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_100, __pyx_t_98) < 0) __PYX_ERR(0, 6099, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_100); __pyx_t_100 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Expr_notContext, __pyx_t_98) < 0) __PYX_ERR(0, 6099, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_98); __pyx_t_98 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6124
 * 
 * 
 *     class Expr_mulContext(ExprContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext
 */
  __pyx_t_7 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_ExprContext);
  if (unlikely(!__pyx_t_7)) {
    PyErr_Clear();
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ExprContext);
  }
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Expr_mulContext, __pyx_n_s_ModelicaParser_Expr_mulContext, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_98 = PyList_New(0); if (unlikely(!__pyx_t_98)) __PYX_ERR(0, 6124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_98);

  /* "pymola/generated/ModelicaParser.pyx":6126
 *     class Expr_mulContext(ExprContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.op = None # Token
 */
  __pyx_t_101 = PyDict_New(); if (unlikely(!__pyx_t_101)) __PYX_ERR(0, 6126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_101);
  __pyx_t_102 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_102)) __PYX_ERR(0, 6126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_102);
  if (PyDict_SetItem(__pyx_t_101, __pyx_n_s_ctx_2, __pyx_t_102) < 0) __PYX_ERR(0, 6126, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_102); __pyx_t_102 = 0;
  __pyx_t_102 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_mulContext_1__init__, 0, __pyx_n_s_ModelicaParser_Expr_mulContext_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2370)); if (unlikely(!__pyx_t_102)) __PYX_ERR(0, 6126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_102);
  __Pyx_INCREF(__pyx_t_102);
  PyList_Append(__pyx_t_98, __pyx_t_102);
  __Pyx_GIVEREF(__pyx_t_102);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_102, __pyx_t_101);
  __Pyx_DECREF(__pyx_t_101); __pyx_t_101 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_102) < 0) __PYX_ERR(0, 6126, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_102); __pyx_t_102 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6131
 *             self.copyFrom(ctx)
 * 
 *         def expr(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExprContext)
 */
  __pyx_t_102 = PyDict_New(); if (unlikely(!__pyx_t_102)) __PYX_ERR(0, 6131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_102);
  if (PyDict_SetItem(__pyx_t_102, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 6131, __pyx_L1_error)
  __pyx_t_101 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_mulContext_3expr, 0, __pyx_n_s_ModelicaParser_Expr_mulContext_e, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2372)); if (unlikely(!__pyx_t_101)) __PYX_ERR(0, 6131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_101);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_101, __pyx_tuple__2373);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_101, __pyx_t_102);
  __Pyx_DECREF(__pyx_t_102); __pyx_t_102 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_expr, __pyx_t_101) < 0) __PYX_ERR(0, 6131, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_101); __pyx_t_101 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6138
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpr_mul" ):
 *                 listener.enterExpr_mul(self)
 */
  __pyx_t_101 = PyDict_New(); if (unlikely(!__pyx_t_101)) __PYX_ERR(0, 6138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_101);
  __pyx_t_102 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_102)) __PYX_ERR(0, 6138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_102);
  if (PyDict_SetItem(__pyx_t_101, __pyx_n_s_listener, __pyx_t_102) < 0) __PYX_ERR(0, 6138, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_102); __pyx_t_102 = 0;
  __pyx_t_102 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_mulContext_5enterRule, 0, __pyx_n_s_ModelicaParser_Expr_mulContext_e_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2375)); if (unlikely(!__pyx_t_102)) __PYX_ERR(0, 6138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_102);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_102, __pyx_t_101);
  __Pyx_DECREF(__pyx_t_101); __pyx_t_101 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_102) < 0) __PYX_ERR(0, 6138, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_102); __pyx_t_102 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6142
 *                 listener.enterExpr_mul(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpr_mul" ):
 *                 listener.exitExpr_mul(self)
 */
  __pyx_t_102 = PyDict_New(); if (unlikely(!__pyx_t_102)) __PYX_ERR(0, 6142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_102);
  __pyx_t_101 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_101)) __PYX_ERR(0, 6142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_101);
  if (PyDict_SetItem(__pyx_t_102, __pyx_n_s_listener, __pyx_t_101) < 0) __PYX_ERR(0, 6142, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_101); __pyx_t_101 = 0;
  __pyx_t_101 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_mulContext_7exitRule, 0, __pyx_n_s_ModelicaParser_Expr_mulContext_e_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2377)); if (unlikely(!__pyx_t_101)) __PYX_ERR(0, 6142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_101);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_101, __pyx_t_102);
  __Pyx_DECREF(__pyx_t_102); __pyx_t_102 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_101) < 0) __PYX_ERR(0, 6142, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_101); __pyx_t_101 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6146
 *                 listener.exitExpr_mul(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpr_mul" ):
 *                 return visitor.visitExpr_mul(self)
 */
  __pyx_t_101 = PyDict_New(); if (unlikely(!__pyx_t_101)) __PYX_ERR(0, 6146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_101);
  __pyx_t_102 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_102)) __PYX_ERR(0, 6146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_102);
  if (PyDict_SetItem(__pyx_t_101, __pyx_n_s_visitor, __pyx_t_102) < 0) __PYX_ERR(0, 6146, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_102); __pyx_t_102 = 0;
  __pyx_t_102 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_15Expr_mulContext_9accept, 0, __pyx_n_s_ModelicaParser_Expr_mulContext_a, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2379)); if (unlikely(!__pyx_t_102)) __PYX_ERR(0, 6146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_102);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_102, __pyx_t_101);
  __Pyx_DECREF(__pyx_t_101); __pyx_t_101 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_102) < 0) __PYX_ERR(0, 6146, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_102); __pyx_t_102 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6124
 * 
 * 
 *     class Expr_mulContext(ExprContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.ExprContext
 */
  __pyx_t_102 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Expr_mulContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_102)) __PYX_ERR(0, 6124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_102);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_98, __pyx_t_102) < 0) __PYX_ERR(0, 6124, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_98); __pyx_t_98 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Expr_mulContext, __pyx_t_102) < 0) __PYX_ERR(0, 6124, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_102); __pyx_t_102 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6154
 * 
 * 
 *     def expr(self, _p:int=0):             # <<<<<<<<<<<<<<
 *         _parentctx = self._ctx
 *         _parentState = self.state
 */
  __pyx_t_5 = PyDict_New(); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_p, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 6154, __pyx_L1_error)
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_115expr, 0, __pyx_n_s_ModelicaParser_expr, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2381)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_tuple__2382);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_expr, __pyx_t_7) < 0) __PYX_ERR(0, 6154, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6323
 *         return localctx
 * 
 *     class PrimaryContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_PrimaryContext, __pyx_n_s_ModelicaParser_PrimaryContext, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_102 = PyList_New(0); if (unlikely(!__pyx_t_102)) __PYX_ERR(0, 6323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_102);

  /* "pymola/generated/ModelicaParser.pyx":6325
 *     class PrimaryContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_101 = PyDict_New(); if (unlikely(!__pyx_t_101)) __PYX_ERR(0, 6325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_101);
  __pyx_t_103 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_103)) __PYX_ERR(0, 6325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_103);
  if (PyDict_SetItem(__pyx_t_101, __pyx_n_s_parent, __pyx_t_103) < 0) __PYX_ERR(0, 6325, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_103); __pyx_t_103 = 0;
  if (PyDict_SetItem(__pyx_t_101, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 6325, __pyx_L1_error)
  __pyx_t_103 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14PrimaryContext_1__init__, 0, __pyx_n_s_ModelicaParser_PrimaryContext_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2384)); if (unlikely(!__pyx_t_103)) __PYX_ERR(0, 6325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_103);
  __Pyx_INCREF(__pyx_t_103);
  PyList_Append(__pyx_t_102, __pyx_t_103);
  __Pyx_GIVEREF(__pyx_t_103);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_103, __pyx_tuple__2385);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_103, __pyx_t_101);
  __Pyx_DECREF(__pyx_t_101); __pyx_t_101 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_103) < 0) __PYX_ERR(0, 6325, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_103); __pyx_t_103 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6330
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_primary
 * 
 */
  __pyx_t_103 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14PrimaryContext_3getRuleIndex, 0, __pyx_n_s_ModelicaParser_PrimaryContext_ge, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2387)); if (unlikely(!__pyx_t_103)) __PYX_ERR(0, 6330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_103);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_103) < 0) __PYX_ERR(0, 6330, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_103); __pyx_t_103 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6334
 * 
 * 
 *         def copyFrom(self, ctx:ParserRuleContext):             # <<<<<<<<<<<<<<
 *             super().copyFrom(ctx)
 * 
 */
  __pyx_t_103 = PyDict_New(); if (unlikely(!__pyx_t_103)) __PYX_ERR(0, 6334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_103);
  __pyx_t_101 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_101)) __PYX_ERR(0, 6334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_101);
  if (PyDict_SetItem(__pyx_t_103, __pyx_n_s_ctx_2, __pyx_t_101) < 0) __PYX_ERR(0, 6334, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_101); __pyx_t_101 = 0;
  __pyx_t_101 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14PrimaryContext_5copyFrom, 0, __pyx_n_s_ModelicaParser_PrimaryContext_co, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2389)); if (unlikely(!__pyx_t_101)) __PYX_ERR(0, 6334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_101);
  __Pyx_INCREF(__pyx_t_101);
  PyList_Append(__pyx_t_102, __pyx_t_101);
  __Pyx_GIVEREF(__pyx_t_101);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_101, __pyx_t_103);
  __Pyx_DECREF(__pyx_t_103); __pyx_t_103 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_copyFrom, __pyx_t_101) < 0) __PYX_ERR(0, 6334, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_101); __pyx_t_101 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6323
 *         return localctx
 * 
 *     class PrimaryContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_101 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_PrimaryContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_101)) __PYX_ERR(0, 6323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_101);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_102, __pyx_t_101) < 0) __PYX_ERR(0, 6323, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_102); __pyx_t_102 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_PrimaryContext, __pyx_t_101) < 0) __PYX_ERR(0, 6323, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_101); __pyx_t_101 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6339
 * 
 * 
 *     class Primary_stringContext(PrimaryContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 */
  __pyx_t_5 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_PrimaryContext);
  if (unlikely(!__pyx_t_5)) {
    PyErr_Clear();
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_PrimaryContext);
  }
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Primary_stringContext, __pyx_n_s_ModelicaParser_Primary_stringCon, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_101 = PyList_New(0); if (unlikely(!__pyx_t_101)) __PYX_ERR(0, 6339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_101);

  /* "pymola/generated/ModelicaParser.pyx":6341
 *     class Primary_stringContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_103 = PyDict_New(); if (unlikely(!__pyx_t_103)) __PYX_ERR(0, 6341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_103);
  __pyx_t_104 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_104)) __PYX_ERR(0, 6341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_104);
  if (PyDict_SetItem(__pyx_t_103, __pyx_n_s_ctx_2, __pyx_t_104) < 0) __PYX_ERR(0, 6341, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_104); __pyx_t_104 = 0;
  __pyx_t_104 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Primary_stringContext_1__init__, 0, __pyx_n_s_ModelicaParser_Primary_stringCon_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2391)); if (unlikely(!__pyx_t_104)) __PYX_ERR(0, 6341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_104);
  __Pyx_INCREF(__pyx_t_104);
  PyList_Append(__pyx_t_101, __pyx_t_104);
  __Pyx_GIVEREF(__pyx_t_104);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_104, __pyx_t_103);
  __Pyx_DECREF(__pyx_t_103); __pyx_t_103 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_104) < 0) __PYX_ERR(0, 6341, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_104); __pyx_t_104 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6345
 *             self.copyFrom(ctx)
 * 
 *         def STRING(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.STRING, 0)
 * 
 */
  __pyx_t_104 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Primary_stringContext_3STRING, 0, __pyx_n_s_ModelicaParser_Primary_stringCon_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2393)); if (unlikely(!__pyx_t_104)) __PYX_ERR(0, 6345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_104);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_STRING, __pyx_t_104) < 0) __PYX_ERR(0, 6345, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_104); __pyx_t_104 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6348
 *             return self.getToken(ModelicaParser.STRING, 0)
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_string" ):
 *                 listener.enterPrimary_string(self)
 */
  __pyx_t_104 = PyDict_New(); if (unlikely(!__pyx_t_104)) __PYX_ERR(0, 6348, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_104);
  __pyx_t_103 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_103)) __PYX_ERR(0, 6348, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_103);
  if (PyDict_SetItem(__pyx_t_104, __pyx_n_s_listener, __pyx_t_103) < 0) __PYX_ERR(0, 6348, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_103); __pyx_t_103 = 0;
  __pyx_t_103 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Primary_stringContext_5enterRule, 0, __pyx_n_s_ModelicaParser_Primary_stringCon_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2395)); if (unlikely(!__pyx_t_103)) __PYX_ERR(0, 6348, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_103);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_103, __pyx_t_104);
  __Pyx_DECREF(__pyx_t_104); __pyx_t_104 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_103) < 0) __PYX_ERR(0, 6348, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_103); __pyx_t_103 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6352
 *                 listener.enterPrimary_string(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_string" ):
 *                 listener.exitPrimary_string(self)
 */
  __pyx_t_103 = PyDict_New(); if (unlikely(!__pyx_t_103)) __PYX_ERR(0, 6352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_103);
  __pyx_t_104 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_104)) __PYX_ERR(0, 6352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_104);
  if (PyDict_SetItem(__pyx_t_103, __pyx_n_s_listener, __pyx_t_104) < 0) __PYX_ERR(0, 6352, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_104); __pyx_t_104 = 0;
  __pyx_t_104 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Primary_stringContext_7exitRule, 0, __pyx_n_s_ModelicaParser_Primary_stringCon_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2397)); if (unlikely(!__pyx_t_104)) __PYX_ERR(0, 6352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_104);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_104, __pyx_t_103);
  __Pyx_DECREF(__pyx_t_103); __pyx_t_103 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_104) < 0) __PYX_ERR(0, 6352, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_104); __pyx_t_104 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6356
 *                 listener.exitPrimary_string(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_string" ):
 *                 return visitor.visitPrimary_string(self)
 */
  __pyx_t_104 = PyDict_New(); if (unlikely(!__pyx_t_104)) __PYX_ERR(0, 6356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_104);
  __pyx_t_103 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_103)) __PYX_ERR(0, 6356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_103);
  if (PyDict_SetItem(__pyx_t_104, __pyx_n_s_visitor, __pyx_t_103) < 0) __PYX_ERR(0, 6356, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_103); __pyx_t_103 = 0;
  __pyx_t_103 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Primary_stringContext_9accept, 0, __pyx_n_s_ModelicaParser_Primary_stringCon_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2399)); if (unlikely(!__pyx_t_103)) __PYX_ERR(0, 6356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_103);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_103, __pyx_t_104);
  __Pyx_DECREF(__pyx_t_104); __pyx_t_104 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_103) < 0) __PYX_ERR(0, 6356, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_103); __pyx_t_103 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6339
 * 
 * 
 *     class Primary_stringContext(PrimaryContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 */
  __pyx_t_103 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Primary_stringContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_103)) __PYX_ERR(0, 6339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_103);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_101, __pyx_t_103) < 0) __PYX_ERR(0, 6339, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_101); __pyx_t_101 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Primary_stringContext, __pyx_t_103) < 0) __PYX_ERR(0, 6339, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_103); __pyx_t_103 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6363
 * 
 * 
 *     class Primary_endContext(PrimaryContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 */
  __pyx_t_7 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_PrimaryContext);
  if (unlikely(!__pyx_t_7)) {
    PyErr_Clear();
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_PrimaryContext);
  }
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Primary_endContext, __pyx_n_s_ModelicaParser_Primary_endContex, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_103 = PyList_New(0); if (unlikely(!__pyx_t_103)) __PYX_ERR(0, 6363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_103);

  /* "pymola/generated/ModelicaParser.pyx":6365
 *     class Primary_endContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_104 = PyDict_New(); if (unlikely(!__pyx_t_104)) __PYX_ERR(0, 6365, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_104);
  __pyx_t_105 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_105)) __PYX_ERR(0, 6365, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_105);
  if (PyDict_SetItem(__pyx_t_104, __pyx_n_s_ctx_2, __pyx_t_105) < 0) __PYX_ERR(0, 6365, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_105); __pyx_t_105 = 0;
  __pyx_t_105 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18Primary_endContext_1__init__, 0, __pyx_n_s_ModelicaParser_Primary_endContex_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2401)); if (unlikely(!__pyx_t_105)) __PYX_ERR(0, 6365, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_105);
  __Pyx_INCREF(__pyx_t_105);
  PyList_Append(__pyx_t_103, __pyx_t_105);
  __Pyx_GIVEREF(__pyx_t_105);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_105, __pyx_t_104);
  __Pyx_DECREF(__pyx_t_104); __pyx_t_104 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_105) < 0) __PYX_ERR(0, 6365, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_105); __pyx_t_105 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6370
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_end" ):
 *                 listener.enterPrimary_end(self)
 */
  __pyx_t_105 = PyDict_New(); if (unlikely(!__pyx_t_105)) __PYX_ERR(0, 6370, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_105);
  __pyx_t_104 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_104)) __PYX_ERR(0, 6370, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_104);
  if (PyDict_SetItem(__pyx_t_105, __pyx_n_s_listener, __pyx_t_104) < 0) __PYX_ERR(0, 6370, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_104); __pyx_t_104 = 0;
  __pyx_t_104 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18Primary_endContext_3enterRule, 0, __pyx_n_s_ModelicaParser_Primary_endContex_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2403)); if (unlikely(!__pyx_t_104)) __PYX_ERR(0, 6370, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_104);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_104, __pyx_t_105);
  __Pyx_DECREF(__pyx_t_105); __pyx_t_105 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_104) < 0) __PYX_ERR(0, 6370, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_104); __pyx_t_104 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6374
 *                 listener.enterPrimary_end(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_end" ):
 *                 listener.exitPrimary_end(self)
 */
  __pyx_t_104 = PyDict_New(); if (unlikely(!__pyx_t_104)) __PYX_ERR(0, 6374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_104);
  __pyx_t_105 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_105)) __PYX_ERR(0, 6374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_105);
  if (PyDict_SetItem(__pyx_t_104, __pyx_n_s_listener, __pyx_t_105) < 0) __PYX_ERR(0, 6374, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_105); __pyx_t_105 = 0;
  __pyx_t_105 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18Primary_endContext_5exitRule, 0, __pyx_n_s_ModelicaParser_Primary_endContex_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2405)); if (unlikely(!__pyx_t_105)) __PYX_ERR(0, 6374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_105);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_105, __pyx_t_104);
  __Pyx_DECREF(__pyx_t_104); __pyx_t_104 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_105) < 0) __PYX_ERR(0, 6374, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_105); __pyx_t_105 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6378
 *                 listener.exitPrimary_end(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_end" ):
 *                 return visitor.visitPrimary_end(self)
 */
  __pyx_t_105 = PyDict_New(); if (unlikely(!__pyx_t_105)) __PYX_ERR(0, 6378, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_105);
  __pyx_t_104 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_104)) __PYX_ERR(0, 6378, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_104);
  if (PyDict_SetItem(__pyx_t_105, __pyx_n_s_visitor, __pyx_t_104) < 0) __PYX_ERR(0, 6378, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_104); __pyx_t_104 = 0;
  __pyx_t_104 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_18Primary_endContext_7accept, 0, __pyx_n_s_ModelicaParser_Primary_endContex_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2407)); if (unlikely(!__pyx_t_104)) __PYX_ERR(0, 6378, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_104);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_104, __pyx_t_105);
  __Pyx_DECREF(__pyx_t_105); __pyx_t_105 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_104) < 0) __PYX_ERR(0, 6378, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_104); __pyx_t_104 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6363
 * 
 * 
 *     class Primary_endContext(PrimaryContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 */
  __pyx_t_104 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Primary_endContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_104)) __PYX_ERR(0, 6363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_104);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_103, __pyx_t_104) < 0) __PYX_ERR(0, 6363, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_103); __pyx_t_103 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Primary_endContext, __pyx_t_104) < 0) __PYX_ERR(0, 6363, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_104); __pyx_t_104 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6385
 * 
 * 
 *     class Primary_output_expression_listContext(PrimaryContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 */
  __pyx_t_5 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_PrimaryContext);
  if (unlikely(!__pyx_t_5)) {
    PyErr_Clear();
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_PrimaryContext);
  }
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Primary_output_expression_listCo, __pyx_n_s_ModelicaParser_Primary_output_ex, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_104 = PyList_New(0); if (unlikely(!__pyx_t_104)) __PYX_ERR(0, 6385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_104);

  /* "pymola/generated/ModelicaParser.pyx":6387
 *     class Primary_output_expression_listContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_105 = PyDict_New(); if (unlikely(!__pyx_t_105)) __PYX_ERR(0, 6387, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_105);
  __pyx_t_106 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_106)) __PYX_ERR(0, 6387, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_106);
  if (PyDict_SetItem(__pyx_t_105, __pyx_n_s_ctx_2, __pyx_t_106) < 0) __PYX_ERR(0, 6387, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_106); __pyx_t_106 = 0;
  __pyx_t_106 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_37Primary_output_expression_listContext_1__init__, 0, __pyx_n_s_ModelicaParser_Primary_output_ex_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2409)); if (unlikely(!__pyx_t_106)) __PYX_ERR(0, 6387, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_106);
  __Pyx_INCREF(__pyx_t_106);
  PyList_Append(__pyx_t_104, __pyx_t_106);
  __Pyx_GIVEREF(__pyx_t_106);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_106, __pyx_t_105);
  __Pyx_DECREF(__pyx_t_105); __pyx_t_105 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_106) < 0) __PYX_ERR(0, 6387, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_106); __pyx_t_106 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6391
 *             self.copyFrom(ctx)
 * 
 *         def output_expression_list(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Output_expression_listContext,0)
 * 
 */
  __pyx_t_106 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_37Primary_output_expression_listContext_3output_expression_list, 0, __pyx_n_s_ModelicaParser_Primary_output_ex_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2411)); if (unlikely(!__pyx_t_106)) __PYX_ERR(0, 6391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_106);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_output_expression_list, __pyx_t_106) < 0) __PYX_ERR(0, 6391, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_106); __pyx_t_106 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6395
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_output_expression_list" ):
 *                 listener.enterPrimary_output_expression_list(self)
 */
  __pyx_t_106 = PyDict_New(); if (unlikely(!__pyx_t_106)) __PYX_ERR(0, 6395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_106);
  __pyx_t_105 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_105)) __PYX_ERR(0, 6395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_105);
  if (PyDict_SetItem(__pyx_t_106, __pyx_n_s_listener, __pyx_t_105) < 0) __PYX_ERR(0, 6395, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_105); __pyx_t_105 = 0;
  __pyx_t_105 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_37Primary_output_expression_listContext_5enterRule, 0, __pyx_n_s_ModelicaParser_Primary_output_ex_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2413)); if (unlikely(!__pyx_t_105)) __PYX_ERR(0, 6395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_105);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_105, __pyx_t_106);
  __Pyx_DECREF(__pyx_t_106); __pyx_t_106 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_105) < 0) __PYX_ERR(0, 6395, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_105); __pyx_t_105 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6399
 *                 listener.enterPrimary_output_expression_list(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_output_expression_list" ):
 *                 listener.exitPrimary_output_expression_list(self)
 */
  __pyx_t_105 = PyDict_New(); if (unlikely(!__pyx_t_105)) __PYX_ERR(0, 6399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_105);
  __pyx_t_106 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_106)) __PYX_ERR(0, 6399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_106);
  if (PyDict_SetItem(__pyx_t_105, __pyx_n_s_listener, __pyx_t_106) < 0) __PYX_ERR(0, 6399, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_106); __pyx_t_106 = 0;
  __pyx_t_106 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_37Primary_output_expression_listContext_7exitRule, 0, __pyx_n_s_ModelicaParser_Primary_output_ex_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2415)); if (unlikely(!__pyx_t_106)) __PYX_ERR(0, 6399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_106);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_106, __pyx_t_105);
  __Pyx_DECREF(__pyx_t_105); __pyx_t_105 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_106) < 0) __PYX_ERR(0, 6399, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_106); __pyx_t_106 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6403
 *                 listener.exitPrimary_output_expression_list(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_output_expression_list" ):
 *                 return visitor.visitPrimary_output_expression_list(self)
 */
  __pyx_t_106 = PyDict_New(); if (unlikely(!__pyx_t_106)) __PYX_ERR(0, 6403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_106);
  __pyx_t_105 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_105)) __PYX_ERR(0, 6403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_105);
  if (PyDict_SetItem(__pyx_t_106, __pyx_n_s_visitor, __pyx_t_105) < 0) __PYX_ERR(0, 6403, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_105); __pyx_t_105 = 0;
  __pyx_t_105 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_37Primary_output_expression_listContext_9accept, 0, __pyx_n_s_ModelicaParser_Primary_output_ex_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2417)); if (unlikely(!__pyx_t_105)) __PYX_ERR(0, 6403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_105);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_105, __pyx_t_106);
  __Pyx_DECREF(__pyx_t_106); __pyx_t_106 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_105) < 0) __PYX_ERR(0, 6403, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_105); __pyx_t_105 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6385
 * 
 * 
 *     class Primary_output_expression_listContext(PrimaryContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 */
  __pyx_t_105 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Primary_output_expression_listCo, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_105)) __PYX_ERR(0, 6385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_105);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_104, __pyx_t_105) < 0) __PYX_ERR(0, 6385, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_104); __pyx_t_104 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Primary_output_expression_listCo, __pyx_t_105) < 0) __PYX_ERR(0, 6385, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_105); __pyx_t_105 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6410
 * 
 * 
 *     class Primary_unsigned_numberContext(PrimaryContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 */
  __pyx_t_7 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_PrimaryContext);
  if (unlikely(!__pyx_t_7)) {
    PyErr_Clear();
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_PrimaryContext);
  }
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Primary_unsigned_numberContext, __pyx_n_s_ModelicaParser_Primary_unsigned, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_105 = PyList_New(0); if (unlikely(!__pyx_t_105)) __PYX_ERR(0, 6410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_105);

  /* "pymola/generated/ModelicaParser.pyx":6412
 *     class Primary_unsigned_numberContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_106 = PyDict_New(); if (unlikely(!__pyx_t_106)) __PYX_ERR(0, 6412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_106);
  __pyx_t_107 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_107)) __PYX_ERR(0, 6412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_107);
  if (PyDict_SetItem(__pyx_t_106, __pyx_n_s_ctx_2, __pyx_t_107) < 0) __PYX_ERR(0, 6412, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_107); __pyx_t_107 = 0;
  __pyx_t_107 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_30Primary_unsigned_numberContext_1__init__, 0, __pyx_n_s_ModelicaParser_Primary_unsigned_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2419)); if (unlikely(!__pyx_t_107)) __PYX_ERR(0, 6412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_107);
  __Pyx_INCREF(__pyx_t_107);
  PyList_Append(__pyx_t_105, __pyx_t_107);
  __Pyx_GIVEREF(__pyx_t_107);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_107, __pyx_t_106);
  __Pyx_DECREF(__pyx_t_106); __pyx_t_106 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_107) < 0) __PYX_ERR(0, 6412, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_107); __pyx_t_107 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6416
 *             self.copyFrom(ctx)
 * 
 *         def UNSIGNED_NUMBER(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.UNSIGNED_NUMBER, 0)
 * 
 */
  __pyx_t_107 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_30Primary_unsigned_numberContext_3UNSIGNED_NUMBER, 0, __pyx_n_s_ModelicaParser_Primary_unsigned_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2421)); if (unlikely(!__pyx_t_107)) __PYX_ERR(0, 6416, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_107);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_UNSIGNED_NUMBER, __pyx_t_107) < 0) __PYX_ERR(0, 6416, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_107); __pyx_t_107 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6419
 *             return self.getToken(ModelicaParser.UNSIGNED_NUMBER, 0)
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_unsigned_number" ):
 *                 listener.enterPrimary_unsigned_number(self)
 */
  __pyx_t_107 = PyDict_New(); if (unlikely(!__pyx_t_107)) __PYX_ERR(0, 6419, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_107);
  __pyx_t_106 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_106)) __PYX_ERR(0, 6419, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_106);
  if (PyDict_SetItem(__pyx_t_107, __pyx_n_s_listener, __pyx_t_106) < 0) __PYX_ERR(0, 6419, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_106); __pyx_t_106 = 0;
  __pyx_t_106 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_30Primary_unsigned_numberContext_5enterRule, 0, __pyx_n_s_ModelicaParser_Primary_unsigned_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2423)); if (unlikely(!__pyx_t_106)) __PYX_ERR(0, 6419, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_106);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_106, __pyx_t_107);
  __Pyx_DECREF(__pyx_t_107); __pyx_t_107 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_106) < 0) __PYX_ERR(0, 6419, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_106); __pyx_t_106 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6423
 *                 listener.enterPrimary_unsigned_number(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_unsigned_number" ):
 *                 listener.exitPrimary_unsigned_number(self)
 */
  __pyx_t_106 = PyDict_New(); if (unlikely(!__pyx_t_106)) __PYX_ERR(0, 6423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_106);
  __pyx_t_107 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_107)) __PYX_ERR(0, 6423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_107);
  if (PyDict_SetItem(__pyx_t_106, __pyx_n_s_listener, __pyx_t_107) < 0) __PYX_ERR(0, 6423, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_107); __pyx_t_107 = 0;
  __pyx_t_107 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_30Primary_unsigned_numberContext_7exitRule, 0, __pyx_n_s_ModelicaParser_Primary_unsigned_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2425)); if (unlikely(!__pyx_t_107)) __PYX_ERR(0, 6423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_107);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_107, __pyx_t_106);
  __Pyx_DECREF(__pyx_t_106); __pyx_t_106 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_107) < 0) __PYX_ERR(0, 6423, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_107); __pyx_t_107 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6427
 *                 listener.exitPrimary_unsigned_number(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_unsigned_number" ):
 *                 return visitor.visitPrimary_unsigned_number(self)
 */
  __pyx_t_107 = PyDict_New(); if (unlikely(!__pyx_t_107)) __PYX_ERR(0, 6427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_107);
  __pyx_t_106 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_106)) __PYX_ERR(0, 6427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_106);
  if (PyDict_SetItem(__pyx_t_107, __pyx_n_s_visitor, __pyx_t_106) < 0) __PYX_ERR(0, 6427, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_106); __pyx_t_106 = 0;
  __pyx_t_106 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_30Primary_unsigned_numberContext_9accept, 0, __pyx_n_s_ModelicaParser_Primary_unsigned_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2427)); if (unlikely(!__pyx_t_106)) __PYX_ERR(0, 6427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_106);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_106, __pyx_t_107);
  __Pyx_DECREF(__pyx_t_107); __pyx_t_107 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_106) < 0) __PYX_ERR(0, 6427, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_106); __pyx_t_106 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6410
 * 
 * 
 *     class Primary_unsigned_numberContext(PrimaryContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 */
  __pyx_t_106 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Primary_unsigned_numberContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_106)) __PYX_ERR(0, 6410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_106);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_105, __pyx_t_106) < 0) __PYX_ERR(0, 6410, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_105); __pyx_t_105 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Primary_unsigned_numberContext, __pyx_t_106) < 0) __PYX_ERR(0, 6410, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_106); __pyx_t_106 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6434
 * 
 * 
 *     class Primary_function_argumentsContext(PrimaryContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 */
  __pyx_t_5 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_PrimaryContext);
  if (unlikely(!__pyx_t_5)) {
    PyErr_Clear();
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_PrimaryContext);
  }
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Primary_function_argumentsContex, __pyx_n_s_ModelicaParser_Primary_function, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_106 = PyList_New(0); if (unlikely(!__pyx_t_106)) __PYX_ERR(0, 6434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_106);

  /* "pymola/generated/ModelicaParser.pyx":6436
 *     class Primary_function_argumentsContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_107 = PyDict_New(); if (unlikely(!__pyx_t_107)) __PYX_ERR(0, 6436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_107);
  __pyx_t_108 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_108)) __PYX_ERR(0, 6436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_108);
  if (PyDict_SetItem(__pyx_t_107, __pyx_n_s_ctx_2, __pyx_t_108) < 0) __PYX_ERR(0, 6436, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_108); __pyx_t_108 = 0;
  __pyx_t_108 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_33Primary_function_argumentsContext_1__init__, 0, __pyx_n_s_ModelicaParser_Primary_function_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2429)); if (unlikely(!__pyx_t_108)) __PYX_ERR(0, 6436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_108);
  __Pyx_INCREF(__pyx_t_108);
  PyList_Append(__pyx_t_106, __pyx_t_108);
  __Pyx_GIVEREF(__pyx_t_108);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_108, __pyx_t_107);
  __Pyx_DECREF(__pyx_t_107); __pyx_t_107 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_108) < 0) __PYX_ERR(0, 6436, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_108); __pyx_t_108 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6440
 *             self.copyFrom(ctx)
 * 
 *         def function_arguments(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Function_argumentsContext,0)
 * 
 */
  __pyx_t_108 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_33Primary_function_argumentsContext_3function_arguments, 0, __pyx_n_s_ModelicaParser_Primary_function_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2431)); if (unlikely(!__pyx_t_108)) __PYX_ERR(0, 6440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_108);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_function_arguments, __pyx_t_108) < 0) __PYX_ERR(0, 6440, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_108); __pyx_t_108 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6444
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_function_arguments" ):
 *                 listener.enterPrimary_function_arguments(self)
 */
  __pyx_t_108 = PyDict_New(); if (unlikely(!__pyx_t_108)) __PYX_ERR(0, 6444, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_108);
  __pyx_t_107 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_107)) __PYX_ERR(0, 6444, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_107);
  if (PyDict_SetItem(__pyx_t_108, __pyx_n_s_listener, __pyx_t_107) < 0) __PYX_ERR(0, 6444, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_107); __pyx_t_107 = 0;
  __pyx_t_107 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_33Primary_function_argumentsContext_5enterRule, 0, __pyx_n_s_ModelicaParser_Primary_function_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2433)); if (unlikely(!__pyx_t_107)) __PYX_ERR(0, 6444, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_107);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_107, __pyx_t_108);
  __Pyx_DECREF(__pyx_t_108); __pyx_t_108 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_107) < 0) __PYX_ERR(0, 6444, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_107); __pyx_t_107 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6448
 *                 listener.enterPrimary_function_arguments(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_function_arguments" ):
 *                 listener.exitPrimary_function_arguments(self)
 */
  __pyx_t_107 = PyDict_New(); if (unlikely(!__pyx_t_107)) __PYX_ERR(0, 6448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_107);
  __pyx_t_108 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_108)) __PYX_ERR(0, 6448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_108);
  if (PyDict_SetItem(__pyx_t_107, __pyx_n_s_listener, __pyx_t_108) < 0) __PYX_ERR(0, 6448, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_108); __pyx_t_108 = 0;
  __pyx_t_108 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_33Primary_function_argumentsContext_7exitRule, 0, __pyx_n_s_ModelicaParser_Primary_function_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2435)); if (unlikely(!__pyx_t_108)) __PYX_ERR(0, 6448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_108);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_108, __pyx_t_107);
  __Pyx_DECREF(__pyx_t_107); __pyx_t_107 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_108) < 0) __PYX_ERR(0, 6448, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_108); __pyx_t_108 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6452
 *                 listener.exitPrimary_function_arguments(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_function_arguments" ):
 *                 return visitor.visitPrimary_function_arguments(self)
 */
  __pyx_t_108 = PyDict_New(); if (unlikely(!__pyx_t_108)) __PYX_ERR(0, 6452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_108);
  __pyx_t_107 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_107)) __PYX_ERR(0, 6452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_107);
  if (PyDict_SetItem(__pyx_t_108, __pyx_n_s_visitor, __pyx_t_107) < 0) __PYX_ERR(0, 6452, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_107); __pyx_t_107 = 0;
  __pyx_t_107 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_33Primary_function_argumentsContext_9accept, 0, __pyx_n_s_ModelicaParser_Primary_function_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2437)); if (unlikely(!__pyx_t_107)) __PYX_ERR(0, 6452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_107);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_107, __pyx_t_108);
  __Pyx_DECREF(__pyx_t_108); __pyx_t_108 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_107) < 0) __PYX_ERR(0, 6452, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_107); __pyx_t_107 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6434
 * 
 * 
 *     class Primary_function_argumentsContext(PrimaryContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 */
  __pyx_t_107 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Primary_function_argumentsContex, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_107)) __PYX_ERR(0, 6434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_107);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_106, __pyx_t_107) < 0) __PYX_ERR(0, 6434, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_106); __pyx_t_106 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Primary_function_argumentsContex, __pyx_t_107) < 0) __PYX_ERR(0, 6434, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_107); __pyx_t_107 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6459
 * 
 * 
 *     class Primary_falseContext(PrimaryContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 */
  __pyx_t_7 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_PrimaryContext);
  if (unlikely(!__pyx_t_7)) {
    PyErr_Clear();
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_PrimaryContext);
  }
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6459, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6459, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6459, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Primary_falseContext, __pyx_n_s_ModelicaParser_Primary_falseCont, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6459, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_107 = PyList_New(0); if (unlikely(!__pyx_t_107)) __PYX_ERR(0, 6459, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_107);

  /* "pymola/generated/ModelicaParser.pyx":6461
 *     class Primary_falseContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_108 = PyDict_New(); if (unlikely(!__pyx_t_108)) __PYX_ERR(0, 6461, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_108);
  __pyx_t_109 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_109)) __PYX_ERR(0, 6461, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_109);
  if (PyDict_SetItem(__pyx_t_108, __pyx_n_s_ctx_2, __pyx_t_109) < 0) __PYX_ERR(0, 6461, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_109); __pyx_t_109 = 0;
  __pyx_t_109 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20Primary_falseContext_1__init__, 0, __pyx_n_s_ModelicaParser_Primary_falseCont_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2439)); if (unlikely(!__pyx_t_109)) __PYX_ERR(0, 6461, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_109);
  __Pyx_INCREF(__pyx_t_109);
  PyList_Append(__pyx_t_107, __pyx_t_109);
  __Pyx_GIVEREF(__pyx_t_109);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_109, __pyx_t_108);
  __Pyx_DECREF(__pyx_t_108); __pyx_t_108 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_109) < 0) __PYX_ERR(0, 6461, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_109); __pyx_t_109 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6466
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_false" ):
 *                 listener.enterPrimary_false(self)
 */
  __pyx_t_109 = PyDict_New(); if (unlikely(!__pyx_t_109)) __PYX_ERR(0, 6466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_109);
  __pyx_t_108 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_108)) __PYX_ERR(0, 6466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_108);
  if (PyDict_SetItem(__pyx_t_109, __pyx_n_s_listener, __pyx_t_108) < 0) __PYX_ERR(0, 6466, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_108); __pyx_t_108 = 0;
  __pyx_t_108 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20Primary_falseContext_3enterRule, 0, __pyx_n_s_ModelicaParser_Primary_falseCont_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2441)); if (unlikely(!__pyx_t_108)) __PYX_ERR(0, 6466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_108);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_108, __pyx_t_109);
  __Pyx_DECREF(__pyx_t_109); __pyx_t_109 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_108) < 0) __PYX_ERR(0, 6466, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_108); __pyx_t_108 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6470
 *                 listener.enterPrimary_false(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_false" ):
 *                 listener.exitPrimary_false(self)
 */
  __pyx_t_108 = PyDict_New(); if (unlikely(!__pyx_t_108)) __PYX_ERR(0, 6470, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_108);
  __pyx_t_109 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_109)) __PYX_ERR(0, 6470, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_109);
  if (PyDict_SetItem(__pyx_t_108, __pyx_n_s_listener, __pyx_t_109) < 0) __PYX_ERR(0, 6470, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_109); __pyx_t_109 = 0;
  __pyx_t_109 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20Primary_falseContext_5exitRule, 0, __pyx_n_s_ModelicaParser_Primary_falseCont_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2443)); if (unlikely(!__pyx_t_109)) __PYX_ERR(0, 6470, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_109);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_109, __pyx_t_108);
  __Pyx_DECREF(__pyx_t_108); __pyx_t_108 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_109) < 0) __PYX_ERR(0, 6470, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_109); __pyx_t_109 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6474
 *                 listener.exitPrimary_false(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_false" ):
 *                 return visitor.visitPrimary_false(self)
 */
  __pyx_t_109 = PyDict_New(); if (unlikely(!__pyx_t_109)) __PYX_ERR(0, 6474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_109);
  __pyx_t_108 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_108)) __PYX_ERR(0, 6474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_108);
  if (PyDict_SetItem(__pyx_t_109, __pyx_n_s_visitor, __pyx_t_108) < 0) __PYX_ERR(0, 6474, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_108); __pyx_t_108 = 0;
  __pyx_t_108 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_20Primary_falseContext_7accept, 0, __pyx_n_s_ModelicaParser_Primary_falseCont_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2445)); if (unlikely(!__pyx_t_108)) __PYX_ERR(0, 6474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_108);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_108, __pyx_t_109);
  __Pyx_DECREF(__pyx_t_109); __pyx_t_109 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_108) < 0) __PYX_ERR(0, 6474, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_108); __pyx_t_108 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6459
 * 
 * 
 *     class Primary_falseContext(PrimaryContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 */
  __pyx_t_108 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Primary_falseContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_108)) __PYX_ERR(0, 6459, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_108);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_107, __pyx_t_108) < 0) __PYX_ERR(0, 6459, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_107); __pyx_t_107 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Primary_falseContext, __pyx_t_108) < 0) __PYX_ERR(0, 6459, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_108); __pyx_t_108 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6481
 * 
 * 
 *     class Primary_derivativeContext(PrimaryContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 */
  __pyx_t_5 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_PrimaryContext);
  if (unlikely(!__pyx_t_5)) {
    PyErr_Clear();
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_PrimaryContext);
  }
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6481, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6481, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6481, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Primary_derivativeContext, __pyx_n_s_ModelicaParser_Primary_derivativ, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6481, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_108 = PyList_New(0); if (unlikely(!__pyx_t_108)) __PYX_ERR(0, 6481, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_108);

  /* "pymola/generated/ModelicaParser.pyx":6483
 *     class Primary_derivativeContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_109 = PyDict_New(); if (unlikely(!__pyx_t_109)) __PYX_ERR(0, 6483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_109);
  __pyx_t_110 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_110)) __PYX_ERR(0, 6483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_110);
  if (PyDict_SetItem(__pyx_t_109, __pyx_n_s_ctx_2, __pyx_t_110) < 0) __PYX_ERR(0, 6483, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_110); __pyx_t_110 = 0;
  __pyx_t_110 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_25Primary_derivativeContext_1__init__, 0, __pyx_n_s_ModelicaParser_Primary_derivativ_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2447)); if (unlikely(!__pyx_t_110)) __PYX_ERR(0, 6483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_110);
  __Pyx_INCREF(__pyx_t_110);
  PyList_Append(__pyx_t_108, __pyx_t_110);
  __Pyx_GIVEREF(__pyx_t_110);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_110, __pyx_t_109);
  __Pyx_DECREF(__pyx_t_109); __pyx_t_109 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_110) < 0) __PYX_ERR(0, 6483, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_110); __pyx_t_110 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6487
 *             self.copyFrom(ctx)
 * 
 *         def function_call_args(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Function_call_argsContext,0)
 * 
 */
  __pyx_t_110 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_25Primary_derivativeContext_3function_call_args, 0, __pyx_n_s_ModelicaParser_Primary_derivativ_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2449)); if (unlikely(!__pyx_t_110)) __PYX_ERR(0, 6487, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_110);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_function_call_args, __pyx_t_110) < 0) __PYX_ERR(0, 6487, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_110); __pyx_t_110 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6491
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_derivative" ):
 *                 listener.enterPrimary_derivative(self)
 */
  __pyx_t_110 = PyDict_New(); if (unlikely(!__pyx_t_110)) __PYX_ERR(0, 6491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_110);
  __pyx_t_109 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_109)) __PYX_ERR(0, 6491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_109);
  if (PyDict_SetItem(__pyx_t_110, __pyx_n_s_listener, __pyx_t_109) < 0) __PYX_ERR(0, 6491, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_109); __pyx_t_109 = 0;
  __pyx_t_109 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_25Primary_derivativeContext_5enterRule, 0, __pyx_n_s_ModelicaParser_Primary_derivativ_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2451)); if (unlikely(!__pyx_t_109)) __PYX_ERR(0, 6491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_109);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_109, __pyx_t_110);
  __Pyx_DECREF(__pyx_t_110); __pyx_t_110 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_109) < 0) __PYX_ERR(0, 6491, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_109); __pyx_t_109 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6495
 *                 listener.enterPrimary_derivative(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_derivative" ):
 *                 listener.exitPrimary_derivative(self)
 */
  __pyx_t_109 = PyDict_New(); if (unlikely(!__pyx_t_109)) __PYX_ERR(0, 6495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_109);
  __pyx_t_110 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_110)) __PYX_ERR(0, 6495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_110);
  if (PyDict_SetItem(__pyx_t_109, __pyx_n_s_listener, __pyx_t_110) < 0) __PYX_ERR(0, 6495, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_110); __pyx_t_110 = 0;
  __pyx_t_110 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_25Primary_derivativeContext_7exitRule, 0, __pyx_n_s_ModelicaParser_Primary_derivativ_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2453)); if (unlikely(!__pyx_t_110)) __PYX_ERR(0, 6495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_110);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_110, __pyx_t_109);
  __Pyx_DECREF(__pyx_t_109); __pyx_t_109 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_110) < 0) __PYX_ERR(0, 6495, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_110); __pyx_t_110 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6499
 *                 listener.exitPrimary_derivative(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_derivative" ):
 *                 return visitor.visitPrimary_derivative(self)
 */
  __pyx_t_110 = PyDict_New(); if (unlikely(!__pyx_t_110)) __PYX_ERR(0, 6499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_110);
  __pyx_t_109 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_109)) __PYX_ERR(0, 6499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_109);
  if (PyDict_SetItem(__pyx_t_110, __pyx_n_s_visitor, __pyx_t_109) < 0) __PYX_ERR(0, 6499, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_109); __pyx_t_109 = 0;
  __pyx_t_109 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_25Primary_derivativeContext_9accept, 0, __pyx_n_s_ModelicaParser_Primary_derivativ_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2455)); if (unlikely(!__pyx_t_109)) __PYX_ERR(0, 6499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_109);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_109, __pyx_t_110);
  __Pyx_DECREF(__pyx_t_110); __pyx_t_110 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_109) < 0) __PYX_ERR(0, 6499, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_109); __pyx_t_109 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6481
 * 
 * 
 *     class Primary_derivativeContext(PrimaryContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 */
  __pyx_t_109 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Primary_derivativeContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_109)) __PYX_ERR(0, 6481, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_109);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_108, __pyx_t_109) < 0) __PYX_ERR(0, 6481, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_108); __pyx_t_108 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Primary_derivativeContext, __pyx_t_109) < 0) __PYX_ERR(0, 6481, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_109); __pyx_t_109 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6506
 * 
 * 
 *     class Primary_component_referenceContext(PrimaryContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 */
  __pyx_t_7 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_PrimaryContext);
  if (unlikely(!__pyx_t_7)) {
    PyErr_Clear();
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_PrimaryContext);
  }
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6506, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6506, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6506, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Primary_component_referenceConte, __pyx_n_s_ModelicaParser_Primary_component, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6506, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_109 = PyList_New(0); if (unlikely(!__pyx_t_109)) __PYX_ERR(0, 6506, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_109);

  /* "pymola/generated/ModelicaParser.pyx":6508
 *     class Primary_component_referenceContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_110 = PyDict_New(); if (unlikely(!__pyx_t_110)) __PYX_ERR(0, 6508, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_110);
  __pyx_t_111 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_111)) __PYX_ERR(0, 6508, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_111);
  if (PyDict_SetItem(__pyx_t_110, __pyx_n_s_ctx_2, __pyx_t_111) < 0) __PYX_ERR(0, 6508, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_111); __pyx_t_111 = 0;
  __pyx_t_111 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_34Primary_component_referenceContext_1__init__, 0, __pyx_n_s_ModelicaParser_Primary_component_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2457)); if (unlikely(!__pyx_t_111)) __PYX_ERR(0, 6508, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_111);
  __Pyx_INCREF(__pyx_t_111);
  PyList_Append(__pyx_t_109, __pyx_t_111);
  __Pyx_GIVEREF(__pyx_t_111);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_111, __pyx_t_110);
  __Pyx_DECREF(__pyx_t_110); __pyx_t_110 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_111) < 0) __PYX_ERR(0, 6508, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_111); __pyx_t_111 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6512
 *             self.copyFrom(ctx)
 * 
 *         def component_reference(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 */
  __pyx_t_111 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_34Primary_component_referenceContext_3component_reference, 0, __pyx_n_s_ModelicaParser_Primary_component_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2459)); if (unlikely(!__pyx_t_111)) __PYX_ERR(0, 6512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_111);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_component_reference, __pyx_t_111) < 0) __PYX_ERR(0, 6512, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_111); __pyx_t_111 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6516
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_component_reference" ):
 *                 listener.enterPrimary_component_reference(self)
 */
  __pyx_t_111 = PyDict_New(); if (unlikely(!__pyx_t_111)) __PYX_ERR(0, 6516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_111);
  __pyx_t_110 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_110)) __PYX_ERR(0, 6516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_110);
  if (PyDict_SetItem(__pyx_t_111, __pyx_n_s_listener, __pyx_t_110) < 0) __PYX_ERR(0, 6516, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_110); __pyx_t_110 = 0;
  __pyx_t_110 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_34Primary_component_referenceContext_5enterRule, 0, __pyx_n_s_ModelicaParser_Primary_component_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2461)); if (unlikely(!__pyx_t_110)) __PYX_ERR(0, 6516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_110);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_110, __pyx_t_111);
  __Pyx_DECREF(__pyx_t_111); __pyx_t_111 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_110) < 0) __PYX_ERR(0, 6516, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_110); __pyx_t_110 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6520
 *                 listener.enterPrimary_component_reference(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_component_reference" ):
 *                 listener.exitPrimary_component_reference(self)
 */
  __pyx_t_110 = PyDict_New(); if (unlikely(!__pyx_t_110)) __PYX_ERR(0, 6520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_110);
  __pyx_t_111 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_111)) __PYX_ERR(0, 6520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_111);
  if (PyDict_SetItem(__pyx_t_110, __pyx_n_s_listener, __pyx_t_111) < 0) __PYX_ERR(0, 6520, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_111); __pyx_t_111 = 0;
  __pyx_t_111 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_34Primary_component_referenceContext_7exitRule, 0, __pyx_n_s_ModelicaParser_Primary_component_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2463)); if (unlikely(!__pyx_t_111)) __PYX_ERR(0, 6520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_111);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_111, __pyx_t_110);
  __Pyx_DECREF(__pyx_t_110); __pyx_t_110 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_111) < 0) __PYX_ERR(0, 6520, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_111); __pyx_t_111 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6524
 *                 listener.exitPrimary_component_reference(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_component_reference" ):
 *                 return visitor.visitPrimary_component_reference(self)
 */
  __pyx_t_111 = PyDict_New(); if (unlikely(!__pyx_t_111)) __PYX_ERR(0, 6524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_111);
  __pyx_t_110 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_110)) __PYX_ERR(0, 6524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_110);
  if (PyDict_SetItem(__pyx_t_111, __pyx_n_s_visitor, __pyx_t_110) < 0) __PYX_ERR(0, 6524, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_110); __pyx_t_110 = 0;
  __pyx_t_110 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_34Primary_component_referenceContext_9accept, 0, __pyx_n_s_ModelicaParser_Primary_component_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2465)); if (unlikely(!__pyx_t_110)) __PYX_ERR(0, 6524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_110);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_110, __pyx_t_111);
  __Pyx_DECREF(__pyx_t_111); __pyx_t_111 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_110) < 0) __PYX_ERR(0, 6524, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_110); __pyx_t_110 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6506
 * 
 * 
 *     class Primary_component_referenceContext(PrimaryContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 */
  __pyx_t_110 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Primary_component_referenceConte, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_110)) __PYX_ERR(0, 6506, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_110);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_109, __pyx_t_110) < 0) __PYX_ERR(0, 6506, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_109); __pyx_t_109 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Primary_component_referenceConte, __pyx_t_110) < 0) __PYX_ERR(0, 6506, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_110); __pyx_t_110 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6531
 * 
 * 
 *     class Primary_expression_listContext(PrimaryContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 */
  __pyx_t_5 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_PrimaryContext);
  if (unlikely(!__pyx_t_5)) {
    PyErr_Clear();
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_PrimaryContext);
  }
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Primary_expression_listContext, __pyx_n_s_ModelicaParser_Primary_expressio, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_110 = PyList_New(0); if (unlikely(!__pyx_t_110)) __PYX_ERR(0, 6531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_110);

  /* "pymola/generated/ModelicaParser.pyx":6533
 *     class Primary_expression_listContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_111 = PyDict_New(); if (unlikely(!__pyx_t_111)) __PYX_ERR(0, 6533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_111);
  __pyx_t_112 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_112)) __PYX_ERR(0, 6533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_112);
  if (PyDict_SetItem(__pyx_t_111, __pyx_n_s_ctx_2, __pyx_t_112) < 0) __PYX_ERR(0, 6533, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_112); __pyx_t_112 = 0;
  __pyx_t_112 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_30Primary_expression_listContext_1__init__, 0, __pyx_n_s_ModelicaParser_Primary_expressio_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2467)); if (unlikely(!__pyx_t_112)) __PYX_ERR(0, 6533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_112);
  __Pyx_INCREF(__pyx_t_112);
  PyList_Append(__pyx_t_110, __pyx_t_112);
  __Pyx_GIVEREF(__pyx_t_112);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_112, __pyx_t_111);
  __Pyx_DECREF(__pyx_t_111); __pyx_t_111 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_112) < 0) __PYX_ERR(0, 6533, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_112); __pyx_t_112 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6537
 *             self.copyFrom(ctx)
 * 
 *         def expression_list(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Expression_listContext)
 */
  __pyx_t_112 = PyDict_New(); if (unlikely(!__pyx_t_112)) __PYX_ERR(0, 6537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_112);
  if (PyDict_SetItem(__pyx_t_112, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 6537, __pyx_L1_error)
  __pyx_t_111 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_30Primary_expression_listContext_3expression_list, 0, __pyx_n_s_ModelicaParser_Primary_expressio_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2469)); if (unlikely(!__pyx_t_111)) __PYX_ERR(0, 6537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_111);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_111, __pyx_tuple__2470);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_111, __pyx_t_112);
  __Pyx_DECREF(__pyx_t_112); __pyx_t_112 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_expression_list, __pyx_t_111) < 0) __PYX_ERR(0, 6537, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_111); __pyx_t_111 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6544
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_expression_list" ):
 *                 listener.enterPrimary_expression_list(self)
 */
  __pyx_t_111 = PyDict_New(); if (unlikely(!__pyx_t_111)) __PYX_ERR(0, 6544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_111);
  __pyx_t_112 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_112)) __PYX_ERR(0, 6544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_112);
  if (PyDict_SetItem(__pyx_t_111, __pyx_n_s_listener, __pyx_t_112) < 0) __PYX_ERR(0, 6544, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_112); __pyx_t_112 = 0;
  __pyx_t_112 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_30Primary_expression_listContext_5enterRule, 0, __pyx_n_s_ModelicaParser_Primary_expressio_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2472)); if (unlikely(!__pyx_t_112)) __PYX_ERR(0, 6544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_112);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_112, __pyx_t_111);
  __Pyx_DECREF(__pyx_t_111); __pyx_t_111 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_112) < 0) __PYX_ERR(0, 6544, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_112); __pyx_t_112 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6548
 *                 listener.enterPrimary_expression_list(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_expression_list" ):
 *                 listener.exitPrimary_expression_list(self)
 */
  __pyx_t_112 = PyDict_New(); if (unlikely(!__pyx_t_112)) __PYX_ERR(0, 6548, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_112);
  __pyx_t_111 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_111)) __PYX_ERR(0, 6548, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_111);
  if (PyDict_SetItem(__pyx_t_112, __pyx_n_s_listener, __pyx_t_111) < 0) __PYX_ERR(0, 6548, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_111); __pyx_t_111 = 0;
  __pyx_t_111 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_30Primary_expression_listContext_7exitRule, 0, __pyx_n_s_ModelicaParser_Primary_expressio_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2474)); if (unlikely(!__pyx_t_111)) __PYX_ERR(0, 6548, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_111);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_111, __pyx_t_112);
  __Pyx_DECREF(__pyx_t_112); __pyx_t_112 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_111) < 0) __PYX_ERR(0, 6548, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_111); __pyx_t_111 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6552
 *                 listener.exitPrimary_expression_list(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_expression_list" ):
 *                 return visitor.visitPrimary_expression_list(self)
 */
  __pyx_t_111 = PyDict_New(); if (unlikely(!__pyx_t_111)) __PYX_ERR(0, 6552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_111);
  __pyx_t_112 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_112)) __PYX_ERR(0, 6552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_112);
  if (PyDict_SetItem(__pyx_t_111, __pyx_n_s_visitor, __pyx_t_112) < 0) __PYX_ERR(0, 6552, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_112); __pyx_t_112 = 0;
  __pyx_t_112 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_30Primary_expression_listContext_9accept, 0, __pyx_n_s_ModelicaParser_Primary_expressio_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2476)); if (unlikely(!__pyx_t_112)) __PYX_ERR(0, 6552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_112);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_112, __pyx_t_111);
  __Pyx_DECREF(__pyx_t_111); __pyx_t_111 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_112) < 0) __PYX_ERR(0, 6552, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_112); __pyx_t_112 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6531
 * 
 * 
 *     class Primary_expression_listContext(PrimaryContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 */
  __pyx_t_112 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Primary_expression_listContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_112)) __PYX_ERR(0, 6531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_112);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_110, __pyx_t_112) < 0) __PYX_ERR(0, 6531, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_110); __pyx_t_110 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Primary_expression_listContext, __pyx_t_112) < 0) __PYX_ERR(0, 6531, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_112); __pyx_t_112 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6559
 * 
 * 
 *     class Primary_trueContext(PrimaryContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 */
  __pyx_t_7 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_PrimaryContext);
  if (unlikely(!__pyx_t_7)) {
    PyErr_Clear();
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_PrimaryContext);
  }
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6559, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6559, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6559, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Primary_trueContext, __pyx_n_s_ModelicaParser_Primary_trueConte, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6559, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_112 = PyList_New(0); if (unlikely(!__pyx_t_112)) __PYX_ERR(0, 6559, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_112);

  /* "pymola/generated/ModelicaParser.pyx":6561
 *     class Primary_trueContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_111 = PyDict_New(); if (unlikely(!__pyx_t_111)) __PYX_ERR(0, 6561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_111);
  __pyx_t_113 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_113)) __PYX_ERR(0, 6561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_113);
  if (PyDict_SetItem(__pyx_t_111, __pyx_n_s_ctx_2, __pyx_t_113) < 0) __PYX_ERR(0, 6561, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_113); __pyx_t_113 = 0;
  __pyx_t_113 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19Primary_trueContext_1__init__, 0, __pyx_n_s_ModelicaParser_Primary_trueConte_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2478)); if (unlikely(!__pyx_t_113)) __PYX_ERR(0, 6561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_113);
  __Pyx_INCREF(__pyx_t_113);
  PyList_Append(__pyx_t_112, __pyx_t_113);
  __Pyx_GIVEREF(__pyx_t_113);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_113, __pyx_t_111);
  __Pyx_DECREF(__pyx_t_111); __pyx_t_111 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_113) < 0) __PYX_ERR(0, 6561, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_113); __pyx_t_113 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6566
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_true" ):
 *                 listener.enterPrimary_true(self)
 */
  __pyx_t_113 = PyDict_New(); if (unlikely(!__pyx_t_113)) __PYX_ERR(0, 6566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_113);
  __pyx_t_111 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_111)) __PYX_ERR(0, 6566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_111);
  if (PyDict_SetItem(__pyx_t_113, __pyx_n_s_listener, __pyx_t_111) < 0) __PYX_ERR(0, 6566, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_111); __pyx_t_111 = 0;
  __pyx_t_111 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19Primary_trueContext_3enterRule, 0, __pyx_n_s_ModelicaParser_Primary_trueConte_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2480)); if (unlikely(!__pyx_t_111)) __PYX_ERR(0, 6566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_111);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_111, __pyx_t_113);
  __Pyx_DECREF(__pyx_t_113); __pyx_t_113 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_111) < 0) __PYX_ERR(0, 6566, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_111); __pyx_t_111 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6570
 *                 listener.enterPrimary_true(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_true" ):
 *                 listener.exitPrimary_true(self)
 */
  __pyx_t_111 = PyDict_New(); if (unlikely(!__pyx_t_111)) __PYX_ERR(0, 6570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_111);
  __pyx_t_113 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_113)) __PYX_ERR(0, 6570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_113);
  if (PyDict_SetItem(__pyx_t_111, __pyx_n_s_listener, __pyx_t_113) < 0) __PYX_ERR(0, 6570, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_113); __pyx_t_113 = 0;
  __pyx_t_113 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19Primary_trueContext_5exitRule, 0, __pyx_n_s_ModelicaParser_Primary_trueConte_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2482)); if (unlikely(!__pyx_t_113)) __PYX_ERR(0, 6570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_113);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_113, __pyx_t_111);
  __Pyx_DECREF(__pyx_t_111); __pyx_t_111 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_113) < 0) __PYX_ERR(0, 6570, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_113); __pyx_t_113 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6574
 *                 listener.exitPrimary_true(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_true" ):
 *                 return visitor.visitPrimary_true(self)
 */
  __pyx_t_113 = PyDict_New(); if (unlikely(!__pyx_t_113)) __PYX_ERR(0, 6574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_113);
  __pyx_t_111 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_111)) __PYX_ERR(0, 6574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_111);
  if (PyDict_SetItem(__pyx_t_113, __pyx_n_s_visitor, __pyx_t_111) < 0) __PYX_ERR(0, 6574, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_111); __pyx_t_111 = 0;
  __pyx_t_111 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_19Primary_trueContext_7accept, 0, __pyx_n_s_ModelicaParser_Primary_trueConte_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2484)); if (unlikely(!__pyx_t_111)) __PYX_ERR(0, 6574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_111);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_111, __pyx_t_113);
  __Pyx_DECREF(__pyx_t_113); __pyx_t_113 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_111) < 0) __PYX_ERR(0, 6574, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_111); __pyx_t_111 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6559
 * 
 * 
 *     class Primary_trueContext(PrimaryContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 */
  __pyx_t_111 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Primary_trueContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_111)) __PYX_ERR(0, 6559, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_111);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_112, __pyx_t_111) < 0) __PYX_ERR(0, 6559, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_112); __pyx_t_112 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Primary_trueContext, __pyx_t_111) < 0) __PYX_ERR(0, 6559, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_111); __pyx_t_111 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6581
 * 
 * 
 *     class Primary_functionContext(PrimaryContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 */
  __pyx_t_5 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_PrimaryContext);
  if (unlikely(!__pyx_t_5)) {
    PyErr_Clear();
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_PrimaryContext);
  }
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Primary_functionContext, __pyx_n_s_ModelicaParser_Primary_functionC, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_111 = PyList_New(0); if (unlikely(!__pyx_t_111)) __PYX_ERR(0, 6581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_111);

  /* "pymola/generated/ModelicaParser.pyx":6583
 *     class Primary_functionContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_113 = PyDict_New(); if (unlikely(!__pyx_t_113)) __PYX_ERR(0, 6583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_113);
  __pyx_t_114 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_114)) __PYX_ERR(0, 6583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_114);
  if (PyDict_SetItem(__pyx_t_113, __pyx_n_s_ctx_2, __pyx_t_114) < 0) __PYX_ERR(0, 6583, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_114); __pyx_t_114 = 0;
  __pyx_t_114 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Primary_functionContext_1__init__, 0, __pyx_n_s_ModelicaParser_Primary_functionC_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2486)); if (unlikely(!__pyx_t_114)) __PYX_ERR(0, 6583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_114);
  __Pyx_INCREF(__pyx_t_114);
  PyList_Append(__pyx_t_111, __pyx_t_114);
  __Pyx_GIVEREF(__pyx_t_114);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_114, __pyx_t_113);
  __Pyx_DECREF(__pyx_t_113); __pyx_t_113 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_114) < 0) __PYX_ERR(0, 6583, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_114); __pyx_t_114 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6587
 *             self.copyFrom(ctx)
 * 
 *         def component_reference(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 */
  __pyx_t_114 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Primary_functionContext_3component_reference, 0, __pyx_n_s_ModelicaParser_Primary_functionC_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2488)); if (unlikely(!__pyx_t_114)) __PYX_ERR(0, 6587, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_114);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_component_reference, __pyx_t_114) < 0) __PYX_ERR(0, 6587, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_114); __pyx_t_114 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6590
 *             return self.getTypedRuleContext(ModelicaParser.Component_referenceContext,0)
 * 
 *         def function_call_args(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Function_call_argsContext,0)
 * 
 */
  __pyx_t_114 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Primary_functionContext_5function_call_args, 0, __pyx_n_s_ModelicaParser_Primary_functionC_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2490)); if (unlikely(!__pyx_t_114)) __PYX_ERR(0, 6590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_114);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_function_call_args, __pyx_t_114) < 0) __PYX_ERR(0, 6590, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_114); __pyx_t_114 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6594
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_function" ):
 *                 listener.enterPrimary_function(self)
 */
  __pyx_t_114 = PyDict_New(); if (unlikely(!__pyx_t_114)) __PYX_ERR(0, 6594, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_114);
  __pyx_t_113 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_113)) __PYX_ERR(0, 6594, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_113);
  if (PyDict_SetItem(__pyx_t_114, __pyx_n_s_listener, __pyx_t_113) < 0) __PYX_ERR(0, 6594, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_113); __pyx_t_113 = 0;
  __pyx_t_113 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Primary_functionContext_7enterRule, 0, __pyx_n_s_ModelicaParser_Primary_functionC_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2492)); if (unlikely(!__pyx_t_113)) __PYX_ERR(0, 6594, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_113);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_113, __pyx_t_114);
  __Pyx_DECREF(__pyx_t_114); __pyx_t_114 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_113) < 0) __PYX_ERR(0, 6594, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_113); __pyx_t_113 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6598
 *                 listener.enterPrimary_function(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_function" ):
 *                 listener.exitPrimary_function(self)
 */
  __pyx_t_113 = PyDict_New(); if (unlikely(!__pyx_t_113)) __PYX_ERR(0, 6598, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_113);
  __pyx_t_114 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_114)) __PYX_ERR(0, 6598, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_114);
  if (PyDict_SetItem(__pyx_t_113, __pyx_n_s_listener, __pyx_t_114) < 0) __PYX_ERR(0, 6598, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_114); __pyx_t_114 = 0;
  __pyx_t_114 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Primary_functionContext_9exitRule, 0, __pyx_n_s_ModelicaParser_Primary_functionC_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2494)); if (unlikely(!__pyx_t_114)) __PYX_ERR(0, 6598, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_114);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_114, __pyx_t_113);
  __Pyx_DECREF(__pyx_t_113); __pyx_t_113 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_114) < 0) __PYX_ERR(0, 6598, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_114); __pyx_t_114 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6602
 *                 listener.exitPrimary_function(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_function" ):
 *                 return visitor.visitPrimary_function(self)
 */
  __pyx_t_114 = PyDict_New(); if (unlikely(!__pyx_t_114)) __PYX_ERR(0, 6602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_114);
  __pyx_t_113 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_113)) __PYX_ERR(0, 6602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_113);
  if (PyDict_SetItem(__pyx_t_114, __pyx_n_s_visitor, __pyx_t_113) < 0) __PYX_ERR(0, 6602, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_113); __pyx_t_113 = 0;
  __pyx_t_113 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Primary_functionContext_11accept, 0, __pyx_n_s_ModelicaParser_Primary_functionC_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2496)); if (unlikely(!__pyx_t_113)) __PYX_ERR(0, 6602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_113);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_113, __pyx_t_114);
  __Pyx_DECREF(__pyx_t_114); __pyx_t_114 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_113) < 0) __PYX_ERR(0, 6602, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_113); __pyx_t_113 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6581
 * 
 * 
 *     class Primary_functionContext(PrimaryContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 */
  __pyx_t_113 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Primary_functionContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_113)) __PYX_ERR(0, 6581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_113);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_111, __pyx_t_113) < 0) __PYX_ERR(0, 6581, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_111); __pyx_t_111 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Primary_functionContext, __pyx_t_113) < 0) __PYX_ERR(0, 6581, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_113); __pyx_t_113 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6609
 * 
 * 
 *     class Primary_initialContext(PrimaryContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 */
  __pyx_t_7 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_PrimaryContext);
  if (unlikely(!__pyx_t_7)) {
    PyErr_Clear();
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_PrimaryContext);
  }
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6609, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6609, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6609, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Primary_initialContext, __pyx_n_s_ModelicaParser_Primary_initialCo, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6609, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_113 = PyList_New(0); if (unlikely(!__pyx_t_113)) __PYX_ERR(0, 6609, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_113);

  /* "pymola/generated/ModelicaParser.pyx":6611
 *     class Primary_initialContext(PrimaryContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_114 = PyDict_New(); if (unlikely(!__pyx_t_114)) __PYX_ERR(0, 6611, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_114);
  __pyx_t_115 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_115)) __PYX_ERR(0, 6611, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_115);
  if (PyDict_SetItem(__pyx_t_114, __pyx_n_s_ctx_2, __pyx_t_115) < 0) __PYX_ERR(0, 6611, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_115); __pyx_t_115 = 0;
  __pyx_t_115 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Primary_initialContext_1__init__, 0, __pyx_n_s_ModelicaParser_Primary_initialCo_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2498)); if (unlikely(!__pyx_t_115)) __PYX_ERR(0, 6611, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_115);
  __Pyx_INCREF(__pyx_t_115);
  PyList_Append(__pyx_t_113, __pyx_t_115);
  __Pyx_GIVEREF(__pyx_t_115);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_115, __pyx_t_114);
  __Pyx_DECREF(__pyx_t_114); __pyx_t_114 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_115) < 0) __PYX_ERR(0, 6611, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_115); __pyx_t_115 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6615
 *             self.copyFrom(ctx)
 * 
 *         def function_call_args(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Function_call_argsContext,0)
 * 
 */
  __pyx_t_115 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Primary_initialContext_3function_call_args, 0, __pyx_n_s_ModelicaParser_Primary_initialCo_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2500)); if (unlikely(!__pyx_t_115)) __PYX_ERR(0, 6615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_115);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_function_call_args, __pyx_t_115) < 0) __PYX_ERR(0, 6615, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_115); __pyx_t_115 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6619
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterPrimary_initial" ):
 *                 listener.enterPrimary_initial(self)
 */
  __pyx_t_115 = PyDict_New(); if (unlikely(!__pyx_t_115)) __PYX_ERR(0, 6619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_115);
  __pyx_t_114 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_114)) __PYX_ERR(0, 6619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_114);
  if (PyDict_SetItem(__pyx_t_115, __pyx_n_s_listener, __pyx_t_114) < 0) __PYX_ERR(0, 6619, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_114); __pyx_t_114 = 0;
  __pyx_t_114 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Primary_initialContext_5enterRule, 0, __pyx_n_s_ModelicaParser_Primary_initialCo_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2502)); if (unlikely(!__pyx_t_114)) __PYX_ERR(0, 6619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_114);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_114, __pyx_t_115);
  __Pyx_DECREF(__pyx_t_115); __pyx_t_115 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_114) < 0) __PYX_ERR(0, 6619, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_114); __pyx_t_114 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6623
 *                 listener.enterPrimary_initial(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitPrimary_initial" ):
 *                 listener.exitPrimary_initial(self)
 */
  __pyx_t_114 = PyDict_New(); if (unlikely(!__pyx_t_114)) __PYX_ERR(0, 6623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_114);
  __pyx_t_115 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_115)) __PYX_ERR(0, 6623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_115);
  if (PyDict_SetItem(__pyx_t_114, __pyx_n_s_listener, __pyx_t_115) < 0) __PYX_ERR(0, 6623, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_115); __pyx_t_115 = 0;
  __pyx_t_115 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Primary_initialContext_7exitRule, 0, __pyx_n_s_ModelicaParser_Primary_initialCo_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2504)); if (unlikely(!__pyx_t_115)) __PYX_ERR(0, 6623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_115);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_115, __pyx_t_114);
  __Pyx_DECREF(__pyx_t_114); __pyx_t_114 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_115) < 0) __PYX_ERR(0, 6623, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_115); __pyx_t_115 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6627
 *                 listener.exitPrimary_initial(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitPrimary_initial" ):
 *                 return visitor.visitPrimary_initial(self)
 */
  __pyx_t_115 = PyDict_New(); if (unlikely(!__pyx_t_115)) __PYX_ERR(0, 6627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_115);
  __pyx_t_114 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_114)) __PYX_ERR(0, 6627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_114);
  if (PyDict_SetItem(__pyx_t_115, __pyx_n_s_visitor, __pyx_t_114) < 0) __PYX_ERR(0, 6627, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_114); __pyx_t_114 = 0;
  __pyx_t_114 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Primary_initialContext_9accept, 0, __pyx_n_s_ModelicaParser_Primary_initialCo_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2506)); if (unlikely(!__pyx_t_114)) __PYX_ERR(0, 6627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_114);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_114, __pyx_t_115);
  __Pyx_DECREF(__pyx_t_115); __pyx_t_115 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_114) < 0) __PYX_ERR(0, 6627, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_114); __pyx_t_114 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6609
 * 
 * 
 *     class Primary_initialContext(PrimaryContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.PrimaryContext
 */
  __pyx_t_114 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Primary_initialContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_114)) __PYX_ERR(0, 6609, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_114);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_113, __pyx_t_114) < 0) __PYX_ERR(0, 6609, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_113); __pyx_t_113 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Primary_initialContext, __pyx_t_114) < 0) __PYX_ERR(0, 6609, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_114); __pyx_t_114 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6635
 * 
 * 
 *     def primary(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.PrimaryContext(self, self._ctx, self.state)
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_117primary, 0, __pyx_n_s_ModelicaParser_primary, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2508)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_primary, __pyx_t_5) < 0) __PYX_ERR(0, 6635, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6767
 *         return localctx
 * 
 *     class NameContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6767, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6767, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6767, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_NameContext, __pyx_n_s_ModelicaParser_NameContext, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6767, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_114 = PyList_New(0); if (unlikely(!__pyx_t_114)) __PYX_ERR(0, 6767, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_114);

  /* "pymola/generated/ModelicaParser.pyx":6769
 *     class NameContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_115 = PyDict_New(); if (unlikely(!__pyx_t_115)) __PYX_ERR(0, 6769, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_115);
  __pyx_t_116 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_116)) __PYX_ERR(0, 6769, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_116);
  if (PyDict_SetItem(__pyx_t_115, __pyx_n_s_parent, __pyx_t_116) < 0) __PYX_ERR(0, 6769, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_116); __pyx_t_116 = 0;
  if (PyDict_SetItem(__pyx_t_115, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 6769, __pyx_L1_error)
  __pyx_t_116 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_11NameContext_1__init__, 0, __pyx_n_s_ModelicaParser_NameContext___ini, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2510)); if (unlikely(!__pyx_t_116)) __PYX_ERR(0, 6769, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_116);
  __Pyx_INCREF(__pyx_t_116);
  PyList_Append(__pyx_t_114, __pyx_t_116);
  __Pyx_GIVEREF(__pyx_t_116);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_116, __pyx_tuple__2511);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_116, __pyx_t_115);
  __Pyx_DECREF(__pyx_t_115); __pyx_t_115 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_116) < 0) __PYX_ERR(0, 6769, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_116); __pyx_t_116 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6773
 *             self.parser = parser
 * 
 *         def IDENT(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTokens(ModelicaParser.IDENT)
 */
  __pyx_t_116 = PyDict_New(); if (unlikely(!__pyx_t_116)) __PYX_ERR(0, 6773, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_116);
  if (PyDict_SetItem(__pyx_t_116, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 6773, __pyx_L1_error)
  __pyx_t_115 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_11NameContext_3IDENT, 0, __pyx_n_s_ModelicaParser_NameContext_IDENT, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2513)); if (unlikely(!__pyx_t_115)) __PYX_ERR(0, 6773, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_115);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_115, __pyx_tuple__2514);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_115, __pyx_t_116);
  __Pyx_DECREF(__pyx_t_116); __pyx_t_116 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_IDENT, __pyx_t_115) < 0) __PYX_ERR(0, 6773, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_115); __pyx_t_115 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6779
 *                 return self.getToken(ModelicaParser.IDENT, i)
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_name
 * 
 */
  __pyx_t_115 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_11NameContext_5getRuleIndex, 0, __pyx_n_s_ModelicaParser_NameContext_getRu, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2516)); if (unlikely(!__pyx_t_115)) __PYX_ERR(0, 6779, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_115);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_115) < 0) __PYX_ERR(0, 6779, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_115); __pyx_t_115 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6782
 *             return ModelicaParser.RULE_name
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterName" ):
 *                 listener.enterName(self)
 */
  __pyx_t_115 = PyDict_New(); if (unlikely(!__pyx_t_115)) __PYX_ERR(0, 6782, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_115);
  __pyx_t_116 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_116)) __PYX_ERR(0, 6782, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_116);
  if (PyDict_SetItem(__pyx_t_115, __pyx_n_s_listener, __pyx_t_116) < 0) __PYX_ERR(0, 6782, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_116); __pyx_t_116 = 0;
  __pyx_t_116 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_11NameContext_7enterRule, 0, __pyx_n_s_ModelicaParser_NameContext_enter, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2518)); if (unlikely(!__pyx_t_116)) __PYX_ERR(0, 6782, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_116);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_116, __pyx_t_115);
  __Pyx_DECREF(__pyx_t_115); __pyx_t_115 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_116) < 0) __PYX_ERR(0, 6782, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_116); __pyx_t_116 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6786
 *                 listener.enterName(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitName" ):
 *                 listener.exitName(self)
 */
  __pyx_t_116 = PyDict_New(); if (unlikely(!__pyx_t_116)) __PYX_ERR(0, 6786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_116);
  __pyx_t_115 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_115)) __PYX_ERR(0, 6786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_115);
  if (PyDict_SetItem(__pyx_t_116, __pyx_n_s_listener, __pyx_t_115) < 0) __PYX_ERR(0, 6786, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_115); __pyx_t_115 = 0;
  __pyx_t_115 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_11NameContext_9exitRule, 0, __pyx_n_s_ModelicaParser_NameContext_exitR, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2520)); if (unlikely(!__pyx_t_115)) __PYX_ERR(0, 6786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_115);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_115, __pyx_t_116);
  __Pyx_DECREF(__pyx_t_116); __pyx_t_116 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_115) < 0) __PYX_ERR(0, 6786, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_115); __pyx_t_115 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6790
 *                 listener.exitName(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitName" ):
 *                 return visitor.visitName(self)
 */
  __pyx_t_115 = PyDict_New(); if (unlikely(!__pyx_t_115)) __PYX_ERR(0, 6790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_115);
  __pyx_t_116 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_116)) __PYX_ERR(0, 6790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_116);
  if (PyDict_SetItem(__pyx_t_115, __pyx_n_s_visitor, __pyx_t_116) < 0) __PYX_ERR(0, 6790, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_116); __pyx_t_116 = 0;
  __pyx_t_116 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_11NameContext_11accept, 0, __pyx_n_s_ModelicaParser_NameContext_accep, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2522)); if (unlikely(!__pyx_t_116)) __PYX_ERR(0, 6790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_116);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_116, __pyx_t_115);
  __Pyx_DECREF(__pyx_t_115); __pyx_t_115 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_116) < 0) __PYX_ERR(0, 6790, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_116); __pyx_t_116 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6767
 *         return localctx
 * 
 *     class NameContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_116 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_NameContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_116)) __PYX_ERR(0, 6767, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_116);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_114, __pyx_t_116) < 0) __PYX_ERR(0, 6767, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_114); __pyx_t_114 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_NameContext, __pyx_t_116) < 0) __PYX_ERR(0, 6767, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_116); __pyx_t_116 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6799
 * 
 * 
 *     def name(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.NameContext(self, self._ctx, self.state)
 */
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_119name, 0, __pyx_n_s_ModelicaParser_name, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2524)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6799, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_name, __pyx_t_7) < 0) __PYX_ERR(0, 6799, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6836
 *         return localctx
 * 
 *     class Component_reference_elementContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6836, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6836, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6836, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Component_reference_elementConte, __pyx_n_s_ModelicaParser_Component_referen, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6836, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_116 = PyList_New(0); if (unlikely(!__pyx_t_116)) __PYX_ERR(0, 6836, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_116);

  /* "pymola/generated/ModelicaParser.pyx":6838
 *     class Component_reference_elementContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_115 = PyDict_New(); if (unlikely(!__pyx_t_115)) __PYX_ERR(0, 6838, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_115);
  __pyx_t_117 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_117)) __PYX_ERR(0, 6838, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_117);
  if (PyDict_SetItem(__pyx_t_115, __pyx_n_s_parent, __pyx_t_117) < 0) __PYX_ERR(0, 6838, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_117); __pyx_t_117 = 0;
  if (PyDict_SetItem(__pyx_t_115, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 6838, __pyx_L1_error)
  __pyx_t_117 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_1__init__, 0, __pyx_n_s_ModelicaParser_Component_referen_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2526)); if (unlikely(!__pyx_t_117)) __PYX_ERR(0, 6838, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_117);
  __Pyx_INCREF(__pyx_t_117);
  PyList_Append(__pyx_t_116, __pyx_t_117);
  __Pyx_GIVEREF(__pyx_t_117);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_117, __pyx_tuple__2527);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_117, __pyx_t_115);
  __Pyx_DECREF(__pyx_t_115); __pyx_t_115 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_117) < 0) __PYX_ERR(0, 6838, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_117); __pyx_t_117 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6842
 *             self.parser = parser
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 */
  __pyx_t_117 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_3IDENT, 0, __pyx_n_s_ModelicaParser_Component_referen_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2529)); if (unlikely(!__pyx_t_117)) __PYX_ERR(0, 6842, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_117);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_IDENT, __pyx_t_117) < 0) __PYX_ERR(0, 6842, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_117); __pyx_t_117 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6845
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 *         def array_subscripts(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Array_subscriptsContext,0)
 * 
 */
  __pyx_t_117 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_5array_subscripts, 0, __pyx_n_s_ModelicaParser_Component_referen_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2531)); if (unlikely(!__pyx_t_117)) __PYX_ERR(0, 6845, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_117);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_array_subscripts, __pyx_t_117) < 0) __PYX_ERR(0, 6845, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_117); __pyx_t_117 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6849
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_component_reference_element
 * 
 */
  __pyx_t_117 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_7getRuleIndex, 0, __pyx_n_s_ModelicaParser_Component_referen_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2533)); if (unlikely(!__pyx_t_117)) __PYX_ERR(0, 6849, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_117);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_117) < 0) __PYX_ERR(0, 6849, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_117); __pyx_t_117 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6852
 *             return ModelicaParser.RULE_component_reference_element
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterComponent_reference_element" ):
 *                 listener.enterComponent_reference_element(self)
 */
  __pyx_t_117 = PyDict_New(); if (unlikely(!__pyx_t_117)) __PYX_ERR(0, 6852, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_117);
  __pyx_t_115 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_115)) __PYX_ERR(0, 6852, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_115);
  if (PyDict_SetItem(__pyx_t_117, __pyx_n_s_listener, __pyx_t_115) < 0) __PYX_ERR(0, 6852, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_115); __pyx_t_115 = 0;
  __pyx_t_115 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_9enterRule, 0, __pyx_n_s_ModelicaParser_Component_referen_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2535)); if (unlikely(!__pyx_t_115)) __PYX_ERR(0, 6852, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_115);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_115, __pyx_t_117);
  __Pyx_DECREF(__pyx_t_117); __pyx_t_117 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_115) < 0) __PYX_ERR(0, 6852, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_115); __pyx_t_115 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6856
 *                 listener.enterComponent_reference_element(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitComponent_reference_element" ):
 *                 listener.exitComponent_reference_element(self)
 */
  __pyx_t_115 = PyDict_New(); if (unlikely(!__pyx_t_115)) __PYX_ERR(0, 6856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_115);
  __pyx_t_117 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_117)) __PYX_ERR(0, 6856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_117);
  if (PyDict_SetItem(__pyx_t_115, __pyx_n_s_listener, __pyx_t_117) < 0) __PYX_ERR(0, 6856, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_117); __pyx_t_117 = 0;
  __pyx_t_117 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_11exitRule, 0, __pyx_n_s_ModelicaParser_Component_referen_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2537)); if (unlikely(!__pyx_t_117)) __PYX_ERR(0, 6856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_117);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_117, __pyx_t_115);
  __Pyx_DECREF(__pyx_t_115); __pyx_t_115 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_117) < 0) __PYX_ERR(0, 6856, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_117); __pyx_t_117 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6860
 *                 listener.exitComponent_reference_element(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitComponent_reference_element" ):
 *                 return visitor.visitComponent_reference_element(self)
 */
  __pyx_t_117 = PyDict_New(); if (unlikely(!__pyx_t_117)) __PYX_ERR(0, 6860, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_117);
  __pyx_t_115 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_115)) __PYX_ERR(0, 6860, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_115);
  if (PyDict_SetItem(__pyx_t_117, __pyx_n_s_visitor, __pyx_t_115) < 0) __PYX_ERR(0, 6860, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_115); __pyx_t_115 = 0;
  __pyx_t_115 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_34Component_reference_elementContext_13accept, 0, __pyx_n_s_ModelicaParser_Component_referen_8, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2539)); if (unlikely(!__pyx_t_115)) __PYX_ERR(0, 6860, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_115);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_115, __pyx_t_117);
  __Pyx_DECREF(__pyx_t_117); __pyx_t_117 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_115) < 0) __PYX_ERR(0, 6860, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_115); __pyx_t_115 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6836
 *         return localctx
 * 
 *     class Component_reference_elementContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_115 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Component_reference_elementConte, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_115)) __PYX_ERR(0, 6836, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_115);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_116, __pyx_t_115) < 0) __PYX_ERR(0, 6836, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_116); __pyx_t_116 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Component_reference_elementConte, __pyx_t_115) < 0) __PYX_ERR(0, 6836, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_115); __pyx_t_115 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6869
 * 
 * 
 *     def component_reference_element(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Component_reference_elementContext(self, self._ctx, self.state)
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_121component_reference_element, 0, __pyx_n_s_ModelicaParser_component_referen, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2541)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6869, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_component_reference_element, __pyx_t_5) < 0) __PYX_ERR(0, 6869, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6893
 *         return localctx
 * 
 *     class Component_referenceContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6893, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6893, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6893, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Component_referenceContext, __pyx_n_s_ModelicaParser_Component_referen_9, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6893, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_115 = PyList_New(0); if (unlikely(!__pyx_t_115)) __PYX_ERR(0, 6893, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_115);

  /* "pymola/generated/ModelicaParser.pyx":6895
 *     class Component_referenceContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_117 = PyDict_New(); if (unlikely(!__pyx_t_117)) __PYX_ERR(0, 6895, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_117);
  __pyx_t_118 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_118)) __PYX_ERR(0, 6895, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_118);
  if (PyDict_SetItem(__pyx_t_117, __pyx_n_s_parent, __pyx_t_118) < 0) __PYX_ERR(0, 6895, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_118); __pyx_t_118 = 0;
  if (PyDict_SetItem(__pyx_t_117, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 6895, __pyx_L1_error)
  __pyx_t_118 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Component_referenceContext_1__init__, 0, __pyx_n_s_ModelicaParser_Component_referen_10, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2543)); if (unlikely(!__pyx_t_118)) __PYX_ERR(0, 6895, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_118);
  __Pyx_INCREF(__pyx_t_118);
  PyList_Append(__pyx_t_115, __pyx_t_118);
  __Pyx_GIVEREF(__pyx_t_118);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_118, __pyx_tuple__2544);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_118, __pyx_t_117);
  __Pyx_DECREF(__pyx_t_117); __pyx_t_117 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_118) < 0) __PYX_ERR(0, 6895, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_118); __pyx_t_118 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6899
 *             self.parser = parser
 * 
 *         def component_reference_element(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Component_reference_elementContext)
 */
  __pyx_t_118 = PyDict_New(); if (unlikely(!__pyx_t_118)) __PYX_ERR(0, 6899, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_118);
  if (PyDict_SetItem(__pyx_t_118, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 6899, __pyx_L1_error)
  __pyx_t_117 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Component_referenceContext_3component_reference_element, 0, __pyx_n_s_ModelicaParser_Component_referen_11, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2546)); if (unlikely(!__pyx_t_117)) __PYX_ERR(0, 6899, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_117);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_117, __pyx_tuple__2547);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_117, __pyx_t_118);
  __Pyx_DECREF(__pyx_t_118); __pyx_t_118 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_component_reference_element, __pyx_t_117) < 0) __PYX_ERR(0, 6899, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_117); __pyx_t_117 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6906
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_component_reference
 * 
 */
  __pyx_t_117 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Component_referenceContext_5getRuleIndex, 0, __pyx_n_s_ModelicaParser_Component_referen_12, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2549)); if (unlikely(!__pyx_t_117)) __PYX_ERR(0, 6906, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_117);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_117) < 0) __PYX_ERR(0, 6906, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_117); __pyx_t_117 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6909
 *             return ModelicaParser.RULE_component_reference
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterComponent_reference" ):
 *                 listener.enterComponent_reference(self)
 */
  __pyx_t_117 = PyDict_New(); if (unlikely(!__pyx_t_117)) __PYX_ERR(0, 6909, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_117);
  __pyx_t_118 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_118)) __PYX_ERR(0, 6909, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_118);
  if (PyDict_SetItem(__pyx_t_117, __pyx_n_s_listener, __pyx_t_118) < 0) __PYX_ERR(0, 6909, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_118); __pyx_t_118 = 0;
  __pyx_t_118 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Component_referenceContext_7enterRule, 0, __pyx_n_s_ModelicaParser_Component_referen_13, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2551)); if (unlikely(!__pyx_t_118)) __PYX_ERR(0, 6909, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_118);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_118, __pyx_t_117);
  __Pyx_DECREF(__pyx_t_117); __pyx_t_117 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_118) < 0) __PYX_ERR(0, 6909, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_118); __pyx_t_118 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6913
 *                 listener.enterComponent_reference(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitComponent_reference" ):
 *                 listener.exitComponent_reference(self)
 */
  __pyx_t_118 = PyDict_New(); if (unlikely(!__pyx_t_118)) __PYX_ERR(0, 6913, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_118);
  __pyx_t_117 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_117)) __PYX_ERR(0, 6913, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_117);
  if (PyDict_SetItem(__pyx_t_118, __pyx_n_s_listener, __pyx_t_117) < 0) __PYX_ERR(0, 6913, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_117); __pyx_t_117 = 0;
  __pyx_t_117 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Component_referenceContext_9exitRule, 0, __pyx_n_s_ModelicaParser_Component_referen_14, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2553)); if (unlikely(!__pyx_t_117)) __PYX_ERR(0, 6913, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_117);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_117, __pyx_t_118);
  __Pyx_DECREF(__pyx_t_118); __pyx_t_118 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_117) < 0) __PYX_ERR(0, 6913, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_117); __pyx_t_117 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6917
 *                 listener.exitComponent_reference(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitComponent_reference" ):
 *                 return visitor.visitComponent_reference(self)
 */
  __pyx_t_117 = PyDict_New(); if (unlikely(!__pyx_t_117)) __PYX_ERR(0, 6917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_117);
  __pyx_t_118 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_118)) __PYX_ERR(0, 6917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_118);
  if (PyDict_SetItem(__pyx_t_117, __pyx_n_s_visitor, __pyx_t_118) < 0) __PYX_ERR(0, 6917, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_118); __pyx_t_118 = 0;
  __pyx_t_118 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Component_referenceContext_11accept, 0, __pyx_n_s_ModelicaParser_Component_referen_15, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2555)); if (unlikely(!__pyx_t_118)) __PYX_ERR(0, 6917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_118);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_118, __pyx_t_117);
  __Pyx_DECREF(__pyx_t_117); __pyx_t_117 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_118) < 0) __PYX_ERR(0, 6917, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_118); __pyx_t_118 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6893
 *         return localctx
 * 
 *     class Component_referenceContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_118 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Component_referenceContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_118)) __PYX_ERR(0, 6893, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_118);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_115, __pyx_t_118) < 0) __PYX_ERR(0, 6893, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_115); __pyx_t_115 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Component_referenceContext, __pyx_t_118) < 0) __PYX_ERR(0, 6893, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_118); __pyx_t_118 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6926
 * 
 * 
 *     def component_reference(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Component_referenceContext(self, self._ctx, self.state)
 */
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_123component_reference, 0, __pyx_n_s_ModelicaParser_component_referen_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2557)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6926, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_component_reference, __pyx_t_7) < 0) __PYX_ERR(0, 6926, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6955
 *         return localctx
 * 
 *     class Function_call_argsContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6955, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6955, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6955, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Function_call_argsContext, __pyx_n_s_ModelicaParser_Function_call_arg, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6955, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_118 = PyList_New(0); if (unlikely(!__pyx_t_118)) __PYX_ERR(0, 6955, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_118);

  /* "pymola/generated/ModelicaParser.pyx":6957
 *     class Function_call_argsContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_117 = PyDict_New(); if (unlikely(!__pyx_t_117)) __PYX_ERR(0, 6957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_117);
  __pyx_t_119 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_119)) __PYX_ERR(0, 6957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_119);
  if (PyDict_SetItem(__pyx_t_117, __pyx_n_s_parent, __pyx_t_119) < 0) __PYX_ERR(0, 6957, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_119); __pyx_t_119 = 0;
  if (PyDict_SetItem(__pyx_t_117, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 6957, __pyx_L1_error)
  __pyx_t_119 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_1__init__, 0, __pyx_n_s_ModelicaParser_Function_call_arg_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2559)); if (unlikely(!__pyx_t_119)) __PYX_ERR(0, 6957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_119);
  __Pyx_INCREF(__pyx_t_119);
  PyList_Append(__pyx_t_118, __pyx_t_119);
  __Pyx_GIVEREF(__pyx_t_119);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_119, __pyx_tuple__2560);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_119, __pyx_t_117);
  __Pyx_DECREF(__pyx_t_117); __pyx_t_117 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_119) < 0) __PYX_ERR(0, 6957, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_119); __pyx_t_119 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6961
 *             self.parser = parser
 * 
 *         def function_arguments(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Function_argumentsContext,0)
 * 
 */
  __pyx_t_119 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_3function_arguments, 0, __pyx_n_s_ModelicaParser_Function_call_arg_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2562)); if (unlikely(!__pyx_t_119)) __PYX_ERR(0, 6961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_119);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_function_arguments, __pyx_t_119) < 0) __PYX_ERR(0, 6961, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_119); __pyx_t_119 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6965
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_function_call_args
 * 
 */
  __pyx_t_119 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_5getRuleIndex, 0, __pyx_n_s_ModelicaParser_Function_call_arg_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2564)); if (unlikely(!__pyx_t_119)) __PYX_ERR(0, 6965, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_119);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_119) < 0) __PYX_ERR(0, 6965, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_119); __pyx_t_119 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6968
 *             return ModelicaParser.RULE_function_call_args
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterFunction_call_args" ):
 *                 listener.enterFunction_call_args(self)
 */
  __pyx_t_119 = PyDict_New(); if (unlikely(!__pyx_t_119)) __PYX_ERR(0, 6968, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_119);
  __pyx_t_117 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_117)) __PYX_ERR(0, 6968, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_117);
  if (PyDict_SetItem(__pyx_t_119, __pyx_n_s_listener, __pyx_t_117) < 0) __PYX_ERR(0, 6968, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_117); __pyx_t_117 = 0;
  __pyx_t_117 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_7enterRule, 0, __pyx_n_s_ModelicaParser_Function_call_arg_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2566)); if (unlikely(!__pyx_t_117)) __PYX_ERR(0, 6968, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_117);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_117, __pyx_t_119);
  __Pyx_DECREF(__pyx_t_119); __pyx_t_119 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_117) < 0) __PYX_ERR(0, 6968, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_117); __pyx_t_117 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6972
 *                 listener.enterFunction_call_args(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitFunction_call_args" ):
 *                 listener.exitFunction_call_args(self)
 */
  __pyx_t_117 = PyDict_New(); if (unlikely(!__pyx_t_117)) __PYX_ERR(0, 6972, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_117);
  __pyx_t_119 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_119)) __PYX_ERR(0, 6972, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_119);
  if (PyDict_SetItem(__pyx_t_117, __pyx_n_s_listener, __pyx_t_119) < 0) __PYX_ERR(0, 6972, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_119); __pyx_t_119 = 0;
  __pyx_t_119 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_9exitRule, 0, __pyx_n_s_ModelicaParser_Function_call_arg_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2568)); if (unlikely(!__pyx_t_119)) __PYX_ERR(0, 6972, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_119);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_119, __pyx_t_117);
  __Pyx_DECREF(__pyx_t_117); __pyx_t_117 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_119) < 0) __PYX_ERR(0, 6972, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_119); __pyx_t_119 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6976
 *                 listener.exitFunction_call_args(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitFunction_call_args" ):
 *                 return visitor.visitFunction_call_args(self)
 */
  __pyx_t_119 = PyDict_New(); if (unlikely(!__pyx_t_119)) __PYX_ERR(0, 6976, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_119);
  __pyx_t_117 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_117)) __PYX_ERR(0, 6976, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_117);
  if (PyDict_SetItem(__pyx_t_119, __pyx_n_s_visitor, __pyx_t_117) < 0) __PYX_ERR(0, 6976, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_117); __pyx_t_117 = 0;
  __pyx_t_117 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_25Function_call_argsContext_11accept, 0, __pyx_n_s_ModelicaParser_Function_call_arg_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2570)); if (unlikely(!__pyx_t_117)) __PYX_ERR(0, 6976, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_117);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_117, __pyx_t_119);
  __Pyx_DECREF(__pyx_t_119); __pyx_t_119 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_117) < 0) __PYX_ERR(0, 6976, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_117); __pyx_t_117 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6955
 *         return localctx
 * 
 *     class Function_call_argsContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_117 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Function_call_argsContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_117)) __PYX_ERR(0, 6955, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_117);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_118, __pyx_t_117) < 0) __PYX_ERR(0, 6955, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_118); __pyx_t_118 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Function_call_argsContext, __pyx_t_117) < 0) __PYX_ERR(0, 6955, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_117); __pyx_t_117 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":6985
 * 
 * 
 *     def function_call_args(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Function_call_argsContext(self, self._ctx, self.state)
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_125function_call_args, 0, __pyx_n_s_ModelicaParser_function_call_arg, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2572)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6985, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_function_call_args, __pyx_t_5) < 0) __PYX_ERR(0, 6985, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7012
 *         return localctx
 * 
 *     class Function_argumentsContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7012, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7012, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7012, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Function_argumentsContext, __pyx_n_s_ModelicaParser_Function_argument, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7012, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_117 = PyList_New(0); if (unlikely(!__pyx_t_117)) __PYX_ERR(0, 7012, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_117);

  /* "pymola/generated/ModelicaParser.pyx":7014
 *     class Function_argumentsContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_119 = PyDict_New(); if (unlikely(!__pyx_t_119)) __PYX_ERR(0, 7014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_119);
  __pyx_t_120 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_120)) __PYX_ERR(0, 7014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_120);
  if (PyDict_SetItem(__pyx_t_119, __pyx_n_s_parent, __pyx_t_120) < 0) __PYX_ERR(0, 7014, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_120); __pyx_t_120 = 0;
  if (PyDict_SetItem(__pyx_t_119, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 7014, __pyx_L1_error)
  __pyx_t_120 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_1__init__, 0, __pyx_n_s_ModelicaParser_Function_argument_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2574)); if (unlikely(!__pyx_t_120)) __PYX_ERR(0, 7014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_120);
  __Pyx_INCREF(__pyx_t_120);
  PyList_Append(__pyx_t_117, __pyx_t_120);
  __Pyx_GIVEREF(__pyx_t_120);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_120, __pyx_tuple__2575);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_120, __pyx_t_119);
  __Pyx_DECREF(__pyx_t_119); __pyx_t_119 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_120) < 0) __PYX_ERR(0, 7014, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_120); __pyx_t_120 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7018
 *             self.parser = parser
 * 
 *         def function_argument(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Function_argumentContext)
 */
  __pyx_t_120 = PyDict_New(); if (unlikely(!__pyx_t_120)) __PYX_ERR(0, 7018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_120);
  if (PyDict_SetItem(__pyx_t_120, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 7018, __pyx_L1_error)
  __pyx_t_119 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_3function_argument, 0, __pyx_n_s_ModelicaParser_Function_argument_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2577)); if (unlikely(!__pyx_t_119)) __PYX_ERR(0, 7018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_119);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_119, __pyx_tuple__2578);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_119, __pyx_t_120);
  __Pyx_DECREF(__pyx_t_120); __pyx_t_120 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_function_argument, __pyx_t_119) < 0) __PYX_ERR(0, 7018, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_119); __pyx_t_119 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7025
 * 
 * 
 *         def for_indices(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.For_indicesContext)
 */
  __pyx_t_119 = PyDict_New(); if (unlikely(!__pyx_t_119)) __PYX_ERR(0, 7025, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_119);
  if (PyDict_SetItem(__pyx_t_119, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 7025, __pyx_L1_error)
  __pyx_t_120 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_5for_indices, 0, __pyx_n_s_ModelicaParser_Function_argument_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2580)); if (unlikely(!__pyx_t_120)) __PYX_ERR(0, 7025, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_120);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_120, __pyx_tuple__2581);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_120, __pyx_t_119);
  __Pyx_DECREF(__pyx_t_119); __pyx_t_119 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_for_indices, __pyx_t_120) < 0) __PYX_ERR(0, 7025, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_120); __pyx_t_120 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7032
 * 
 * 
 *         def named_arguments(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Named_argumentsContext,0)
 * 
 */
  __pyx_t_120 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_7named_arguments, 0, __pyx_n_s_ModelicaParser_Function_argument_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2583)); if (unlikely(!__pyx_t_120)) __PYX_ERR(0, 7032, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_120);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_named_arguments, __pyx_t_120) < 0) __PYX_ERR(0, 7032, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_120); __pyx_t_120 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7036
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_function_arguments
 * 
 */
  __pyx_t_120 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_9getRuleIndex, 0, __pyx_n_s_ModelicaParser_Function_argument_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2585)); if (unlikely(!__pyx_t_120)) __PYX_ERR(0, 7036, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_120);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_120) < 0) __PYX_ERR(0, 7036, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_120); __pyx_t_120 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7039
 *             return ModelicaParser.RULE_function_arguments
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterFunction_arguments" ):
 *                 listener.enterFunction_arguments(self)
 */
  __pyx_t_120 = PyDict_New(); if (unlikely(!__pyx_t_120)) __PYX_ERR(0, 7039, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_120);
  __pyx_t_119 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_119)) __PYX_ERR(0, 7039, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_119);
  if (PyDict_SetItem(__pyx_t_120, __pyx_n_s_listener, __pyx_t_119) < 0) __PYX_ERR(0, 7039, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_119); __pyx_t_119 = 0;
  __pyx_t_119 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_11enterRule, 0, __pyx_n_s_ModelicaParser_Function_argument_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2587)); if (unlikely(!__pyx_t_119)) __PYX_ERR(0, 7039, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_119);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_119, __pyx_t_120);
  __Pyx_DECREF(__pyx_t_120); __pyx_t_120 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_119) < 0) __PYX_ERR(0, 7039, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_119); __pyx_t_119 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7043
 *                 listener.enterFunction_arguments(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitFunction_arguments" ):
 *                 listener.exitFunction_arguments(self)
 */
  __pyx_t_119 = PyDict_New(); if (unlikely(!__pyx_t_119)) __PYX_ERR(0, 7043, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_119);
  __pyx_t_120 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_120)) __PYX_ERR(0, 7043, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_120);
  if (PyDict_SetItem(__pyx_t_119, __pyx_n_s_listener, __pyx_t_120) < 0) __PYX_ERR(0, 7043, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_120); __pyx_t_120 = 0;
  __pyx_t_120 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_13exitRule, 0, __pyx_n_s_ModelicaParser_Function_argument_8, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2589)); if (unlikely(!__pyx_t_120)) __PYX_ERR(0, 7043, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_120);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_120, __pyx_t_119);
  __Pyx_DECREF(__pyx_t_119); __pyx_t_119 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_120) < 0) __PYX_ERR(0, 7043, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_120); __pyx_t_120 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7047
 *                 listener.exitFunction_arguments(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitFunction_arguments" ):
 *                 return visitor.visitFunction_arguments(self)
 */
  __pyx_t_120 = PyDict_New(); if (unlikely(!__pyx_t_120)) __PYX_ERR(0, 7047, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_120);
  __pyx_t_119 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_119)) __PYX_ERR(0, 7047, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_119);
  if (PyDict_SetItem(__pyx_t_120, __pyx_n_s_visitor, __pyx_t_119) < 0) __PYX_ERR(0, 7047, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_119); __pyx_t_119 = 0;
  __pyx_t_119 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_25Function_argumentsContext_15accept, 0, __pyx_n_s_ModelicaParser_Function_argument_9, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2591)); if (unlikely(!__pyx_t_119)) __PYX_ERR(0, 7047, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_119);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_119, __pyx_t_120);
  __Pyx_DECREF(__pyx_t_120); __pyx_t_120 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_119) < 0) __PYX_ERR(0, 7047, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_119); __pyx_t_119 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7012
 *         return localctx
 * 
 *     class Function_argumentsContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_119 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Function_argumentsContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_119)) __PYX_ERR(0, 7012, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_119);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_117, __pyx_t_119) < 0) __PYX_ERR(0, 7012, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_117); __pyx_t_117 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Function_argumentsContext, __pyx_t_119) < 0) __PYX_ERR(0, 7012, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_119); __pyx_t_119 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7056
 * 
 * 
 *     def function_arguments(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Function_argumentsContext(self, self._ctx, self.state)
 */
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_127function_arguments, 0, __pyx_n_s_ModelicaParser_function_argument, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2593)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7056, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_function_arguments, __pyx_t_7) < 0) __PYX_ERR(0, 7056, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7112
 *         return localctx
 * 
 *     class Named_argumentsContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Named_argumentsContext, __pyx_n_s_ModelicaParser_Named_argumentsCo, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_119 = PyList_New(0); if (unlikely(!__pyx_t_119)) __PYX_ERR(0, 7112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_119);

  /* "pymola/generated/ModelicaParser.pyx":7114
 *     class Named_argumentsContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_120 = PyDict_New(); if (unlikely(!__pyx_t_120)) __PYX_ERR(0, 7114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_120);
  __pyx_t_121 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_121)) __PYX_ERR(0, 7114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_121);
  if (PyDict_SetItem(__pyx_t_120, __pyx_n_s_parent, __pyx_t_121) < 0) __PYX_ERR(0, 7114, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_121); __pyx_t_121 = 0;
  if (PyDict_SetItem(__pyx_t_120, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 7114, __pyx_L1_error)
  __pyx_t_121 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_1__init__, 0, __pyx_n_s_ModelicaParser_Named_argumentsCo_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2595)); if (unlikely(!__pyx_t_121)) __PYX_ERR(0, 7114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_121);
  __Pyx_INCREF(__pyx_t_121);
  PyList_Append(__pyx_t_119, __pyx_t_121);
  __Pyx_GIVEREF(__pyx_t_121);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_121, __pyx_tuple__2596);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_121, __pyx_t_120);
  __Pyx_DECREF(__pyx_t_120); __pyx_t_120 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_121) < 0) __PYX_ERR(0, 7114, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_121); __pyx_t_121 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7118
 *             self.parser = parser
 * 
 *         def named_argument(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.Named_argumentContext)
 */
  __pyx_t_121 = PyDict_New(); if (unlikely(!__pyx_t_121)) __PYX_ERR(0, 7118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_121);
  if (PyDict_SetItem(__pyx_t_121, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 7118, __pyx_L1_error)
  __pyx_t_120 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_3named_argument, 0, __pyx_n_s_ModelicaParser_Named_argumentsCo_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2598)); if (unlikely(!__pyx_t_120)) __PYX_ERR(0, 7118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_120);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_120, __pyx_tuple__2599);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_120, __pyx_t_121);
  __Pyx_DECREF(__pyx_t_121); __pyx_t_121 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_named_argument, __pyx_t_120) < 0) __PYX_ERR(0, 7118, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_120); __pyx_t_120 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7125
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_named_arguments
 * 
 */
  __pyx_t_120 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_5getRuleIndex, 0, __pyx_n_s_ModelicaParser_Named_argumentsCo_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2601)); if (unlikely(!__pyx_t_120)) __PYX_ERR(0, 7125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_120);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_120) < 0) __PYX_ERR(0, 7125, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_120); __pyx_t_120 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7128
 *             return ModelicaParser.RULE_named_arguments
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterNamed_arguments" ):
 *                 listener.enterNamed_arguments(self)
 */
  __pyx_t_120 = PyDict_New(); if (unlikely(!__pyx_t_120)) __PYX_ERR(0, 7128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_120);
  __pyx_t_121 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_121)) __PYX_ERR(0, 7128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_121);
  if (PyDict_SetItem(__pyx_t_120, __pyx_n_s_listener, __pyx_t_121) < 0) __PYX_ERR(0, 7128, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_121); __pyx_t_121 = 0;
  __pyx_t_121 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_7enterRule, 0, __pyx_n_s_ModelicaParser_Named_argumentsCo_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2603)); if (unlikely(!__pyx_t_121)) __PYX_ERR(0, 7128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_121);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_121, __pyx_t_120);
  __Pyx_DECREF(__pyx_t_120); __pyx_t_120 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_121) < 0) __PYX_ERR(0, 7128, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_121); __pyx_t_121 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7132
 *                 listener.enterNamed_arguments(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitNamed_arguments" ):
 *                 listener.exitNamed_arguments(self)
 */
  __pyx_t_121 = PyDict_New(); if (unlikely(!__pyx_t_121)) __PYX_ERR(0, 7132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_121);
  __pyx_t_120 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_120)) __PYX_ERR(0, 7132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_120);
  if (PyDict_SetItem(__pyx_t_121, __pyx_n_s_listener, __pyx_t_120) < 0) __PYX_ERR(0, 7132, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_120); __pyx_t_120 = 0;
  __pyx_t_120 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_9exitRule, 0, __pyx_n_s_ModelicaParser_Named_argumentsCo_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2605)); if (unlikely(!__pyx_t_120)) __PYX_ERR(0, 7132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_120);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_120, __pyx_t_121);
  __Pyx_DECREF(__pyx_t_121); __pyx_t_121 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_120) < 0) __PYX_ERR(0, 7132, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_120); __pyx_t_120 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7136
 *                 listener.exitNamed_arguments(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitNamed_arguments" ):
 *                 return visitor.visitNamed_arguments(self)
 */
  __pyx_t_120 = PyDict_New(); if (unlikely(!__pyx_t_120)) __PYX_ERR(0, 7136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_120);
  __pyx_t_121 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_121)) __PYX_ERR(0, 7136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_121);
  if (PyDict_SetItem(__pyx_t_120, __pyx_n_s_visitor, __pyx_t_121) < 0) __PYX_ERR(0, 7136, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_121); __pyx_t_121 = 0;
  __pyx_t_121 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Named_argumentsContext_11accept, 0, __pyx_n_s_ModelicaParser_Named_argumentsCo_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2607)); if (unlikely(!__pyx_t_121)) __PYX_ERR(0, 7136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_121);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_121, __pyx_t_120);
  __Pyx_DECREF(__pyx_t_120); __pyx_t_120 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_121) < 0) __PYX_ERR(0, 7136, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_121); __pyx_t_121 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7112
 *         return localctx
 * 
 *     class Named_argumentsContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_121 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Named_argumentsContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_121)) __PYX_ERR(0, 7112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_121);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_119, __pyx_t_121) < 0) __PYX_ERR(0, 7112, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_119); __pyx_t_119 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Named_argumentsContext, __pyx_t_121) < 0) __PYX_ERR(0, 7112, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_121); __pyx_t_121 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7145
 * 
 * 
 *     def named_arguments(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Named_argumentsContext(self, self._ctx, self.state)
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_129named_arguments, 0, __pyx_n_s_ModelicaParser_named_arguments, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2609)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_named_arguments, __pyx_t_5) < 0) __PYX_ERR(0, 7145, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7174
 *         return localctx
 * 
 *     class Named_argumentContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Named_argumentContext, __pyx_n_s_ModelicaParser_Named_argumentCon, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_121 = PyList_New(0); if (unlikely(!__pyx_t_121)) __PYX_ERR(0, 7174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_121);

  /* "pymola/generated/ModelicaParser.pyx":7176
 *     class Named_argumentContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_120 = PyDict_New(); if (unlikely(!__pyx_t_120)) __PYX_ERR(0, 7176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_120);
  __pyx_t_122 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_122)) __PYX_ERR(0, 7176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_122);
  if (PyDict_SetItem(__pyx_t_120, __pyx_n_s_parent, __pyx_t_122) < 0) __PYX_ERR(0, 7176, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_122); __pyx_t_122 = 0;
  if (PyDict_SetItem(__pyx_t_120, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 7176, __pyx_L1_error)
  __pyx_t_122 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Named_argumentContext_1__init__, 0, __pyx_n_s_ModelicaParser_Named_argumentCon_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2611)); if (unlikely(!__pyx_t_122)) __PYX_ERR(0, 7176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_122);
  __Pyx_INCREF(__pyx_t_122);
  PyList_Append(__pyx_t_121, __pyx_t_122);
  __Pyx_GIVEREF(__pyx_t_122);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_122, __pyx_tuple__2612);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_122, __pyx_t_120);
  __Pyx_DECREF(__pyx_t_120); __pyx_t_120 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_122) < 0) __PYX_ERR(0, 7176, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_122); __pyx_t_122 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7180
 *             self.parser = parser
 * 
 *         def IDENT(self):             # <<<<<<<<<<<<<<
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 */
  __pyx_t_122 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Named_argumentContext_3IDENT, 0, __pyx_n_s_ModelicaParser_Named_argumentCon_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2614)); if (unlikely(!__pyx_t_122)) __PYX_ERR(0, 7180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_122);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_IDENT, __pyx_t_122) < 0) __PYX_ERR(0, 7180, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_122); __pyx_t_122 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7183
 *             return self.getToken(ModelicaParser.IDENT, 0)
 * 
 *         def function_argument(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Function_argumentContext,0)
 * 
 */
  __pyx_t_122 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Named_argumentContext_5function_argument, 0, __pyx_n_s_ModelicaParser_Named_argumentCon_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2616)); if (unlikely(!__pyx_t_122)) __PYX_ERR(0, 7183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_122);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_function_argument, __pyx_t_122) < 0) __PYX_ERR(0, 7183, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_122); __pyx_t_122 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7187
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_named_argument
 * 
 */
  __pyx_t_122 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Named_argumentContext_7getRuleIndex, 0, __pyx_n_s_ModelicaParser_Named_argumentCon_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2618)); if (unlikely(!__pyx_t_122)) __PYX_ERR(0, 7187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_122);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_122) < 0) __PYX_ERR(0, 7187, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_122); __pyx_t_122 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7190
 *             return ModelicaParser.RULE_named_argument
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterNamed_argument" ):
 *                 listener.enterNamed_argument(self)
 */
  __pyx_t_122 = PyDict_New(); if (unlikely(!__pyx_t_122)) __PYX_ERR(0, 7190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_122);
  __pyx_t_120 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_120)) __PYX_ERR(0, 7190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_120);
  if (PyDict_SetItem(__pyx_t_122, __pyx_n_s_listener, __pyx_t_120) < 0) __PYX_ERR(0, 7190, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_120); __pyx_t_120 = 0;
  __pyx_t_120 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Named_argumentContext_9enterRule, 0, __pyx_n_s_ModelicaParser_Named_argumentCon_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2620)); if (unlikely(!__pyx_t_120)) __PYX_ERR(0, 7190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_120);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_120, __pyx_t_122);
  __Pyx_DECREF(__pyx_t_122); __pyx_t_122 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_120) < 0) __PYX_ERR(0, 7190, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_120); __pyx_t_120 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7194
 *                 listener.enterNamed_argument(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitNamed_argument" ):
 *                 listener.exitNamed_argument(self)
 */
  __pyx_t_120 = PyDict_New(); if (unlikely(!__pyx_t_120)) __PYX_ERR(0, 7194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_120);
  __pyx_t_122 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_122)) __PYX_ERR(0, 7194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_122);
  if (PyDict_SetItem(__pyx_t_120, __pyx_n_s_listener, __pyx_t_122) < 0) __PYX_ERR(0, 7194, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_122); __pyx_t_122 = 0;
  __pyx_t_122 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Named_argumentContext_11exitRule, 0, __pyx_n_s_ModelicaParser_Named_argumentCon_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2622)); if (unlikely(!__pyx_t_122)) __PYX_ERR(0, 7194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_122);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_122, __pyx_t_120);
  __Pyx_DECREF(__pyx_t_120); __pyx_t_120 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_122) < 0) __PYX_ERR(0, 7194, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_122); __pyx_t_122 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7198
 *                 listener.exitNamed_argument(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitNamed_argument" ):
 *                 return visitor.visitNamed_argument(self)
 */
  __pyx_t_122 = PyDict_New(); if (unlikely(!__pyx_t_122)) __PYX_ERR(0, 7198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_122);
  __pyx_t_120 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_120)) __PYX_ERR(0, 7198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_120);
  if (PyDict_SetItem(__pyx_t_122, __pyx_n_s_visitor, __pyx_t_120) < 0) __PYX_ERR(0, 7198, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_120); __pyx_t_120 = 0;
  __pyx_t_120 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21Named_argumentContext_13accept, 0, __pyx_n_s_ModelicaParser_Named_argumentCon_8, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2624)); if (unlikely(!__pyx_t_120)) __PYX_ERR(0, 7198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_120);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_120, __pyx_t_122);
  __Pyx_DECREF(__pyx_t_122); __pyx_t_122 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_120) < 0) __PYX_ERR(0, 7198, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_120); __pyx_t_120 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7174
 *         return localctx
 * 
 *     class Named_argumentContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_120 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Named_argumentContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_120)) __PYX_ERR(0, 7174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_120);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_121, __pyx_t_120) < 0) __PYX_ERR(0, 7174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_121); __pyx_t_121 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Named_argumentContext, __pyx_t_120) < 0) __PYX_ERR(0, 7174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_120); __pyx_t_120 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7207
 * 
 * 
 *     def named_argument(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Named_argumentContext(self, self._ctx, self.state)
 */
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_131named_argument, 0, __pyx_n_s_ModelicaParser_named_argument, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2626)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_named_argument, __pyx_t_7) < 0) __PYX_ERR(0, 7207, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7227
 *         return localctx
 * 
 *     class Function_argumentContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Function_argumentContext, __pyx_n_s_ModelicaParser_Function_argument_10, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_120 = PyList_New(0); if (unlikely(!__pyx_t_120)) __PYX_ERR(0, 7227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_120);

  /* "pymola/generated/ModelicaParser.pyx":7229
 *     class Function_argumentContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_122 = PyDict_New(); if (unlikely(!__pyx_t_122)) __PYX_ERR(0, 7229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_122);
  __pyx_t_123 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_123)) __PYX_ERR(0, 7229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_123);
  if (PyDict_SetItem(__pyx_t_122, __pyx_n_s_parent, __pyx_t_123) < 0) __PYX_ERR(0, 7229, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_123); __pyx_t_123 = 0;
  if (PyDict_SetItem(__pyx_t_122, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 7229, __pyx_L1_error)
  __pyx_t_123 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Function_argumentContext_1__init__, 0, __pyx_n_s_ModelicaParser_Function_argument_11, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2628)); if (unlikely(!__pyx_t_123)) __PYX_ERR(0, 7229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_123);
  __Pyx_INCREF(__pyx_t_123);
  PyList_Append(__pyx_t_120, __pyx_t_123);
  __Pyx_GIVEREF(__pyx_t_123);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_123, __pyx_tuple__2629);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_123, __pyx_t_122);
  __Pyx_DECREF(__pyx_t_122); __pyx_t_122 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_123) < 0) __PYX_ERR(0, 7229, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_123); __pyx_t_123 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7234
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_function_argument
 * 
 */
  __pyx_t_123 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Function_argumentContext_3getRuleIndex, 0, __pyx_n_s_ModelicaParser_Function_argument_12, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2631)); if (unlikely(!__pyx_t_123)) __PYX_ERR(0, 7234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_123);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_123) < 0) __PYX_ERR(0, 7234, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_123); __pyx_t_123 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7238
 * 
 * 
 *         def copyFrom(self, ctx:ParserRuleContext):             # <<<<<<<<<<<<<<
 *             super().copyFrom(ctx)
 * 
 */
  __pyx_t_123 = PyDict_New(); if (unlikely(!__pyx_t_123)) __PYX_ERR(0, 7238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_123);
  __pyx_t_122 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_122)) __PYX_ERR(0, 7238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_122);
  if (PyDict_SetItem(__pyx_t_123, __pyx_n_s_ctx_2, __pyx_t_122) < 0) __PYX_ERR(0, 7238, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_122); __pyx_t_122 = 0;
  __pyx_t_122 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Function_argumentContext_5copyFrom, 0, __pyx_n_s_ModelicaParser_Function_argument_13, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2633)); if (unlikely(!__pyx_t_122)) __PYX_ERR(0, 7238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_122);
  __Pyx_INCREF(__pyx_t_122);
  PyList_Append(__pyx_t_120, __pyx_t_122);
  __Pyx_GIVEREF(__pyx_t_122);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_122, __pyx_t_123);
  __Pyx_DECREF(__pyx_t_123); __pyx_t_123 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_copyFrom, __pyx_t_122) < 0) __PYX_ERR(0, 7238, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_122); __pyx_t_122 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7227
 *         return localctx
 * 
 *     class Function_argumentContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_122 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Function_argumentContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_122)) __PYX_ERR(0, 7227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_122);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_120, __pyx_t_122) < 0) __PYX_ERR(0, 7227, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_120); __pyx_t_120 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Function_argumentContext, __pyx_t_122) < 0) __PYX_ERR(0, 7227, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_122); __pyx_t_122 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7243
 * 
 * 
 *     class Argument_expressionContext(Function_argumentContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Function_argumentContext
 */
  __pyx_t_5 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_Function_argumentContext);
  if (unlikely(!__pyx_t_5)) {
    PyErr_Clear();
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_Function_argumentContext);
  }
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Argument_expressionContext, __pyx_n_s_ModelicaParser_Argument_expressi, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_122 = PyList_New(0); if (unlikely(!__pyx_t_122)) __PYX_ERR(0, 7243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_122);

  /* "pymola/generated/ModelicaParser.pyx":7245
 *     class Argument_expressionContext(Function_argumentContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Function_argumentContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_123 = PyDict_New(); if (unlikely(!__pyx_t_123)) __PYX_ERR(0, 7245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_123);
  __pyx_t_124 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_124)) __PYX_ERR(0, 7245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_124);
  if (PyDict_SetItem(__pyx_t_123, __pyx_n_s_ctx_2, __pyx_t_124) < 0) __PYX_ERR(0, 7245, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_124); __pyx_t_124 = 0;
  __pyx_t_124 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Argument_expressionContext_1__init__, 0, __pyx_n_s_ModelicaParser_Argument_expressi_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2635)); if (unlikely(!__pyx_t_124)) __PYX_ERR(0, 7245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_124);
  __Pyx_INCREF(__pyx_t_124);
  PyList_Append(__pyx_t_122, __pyx_t_124);
  __Pyx_GIVEREF(__pyx_t_124);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_124, __pyx_t_123);
  __Pyx_DECREF(__pyx_t_123); __pyx_t_123 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_124) < 0) __PYX_ERR(0, 7245, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_124); __pyx_t_124 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7249
 *             self.copyFrom(ctx)
 * 
 *         def expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 */
  __pyx_t_124 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Argument_expressionContext_3expression, 0, __pyx_n_s_ModelicaParser_Argument_expressi_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2637)); if (unlikely(!__pyx_t_124)) __PYX_ERR(0, 7249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_124);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_expression, __pyx_t_124) < 0) __PYX_ERR(0, 7249, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_124); __pyx_t_124 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7253
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterArgument_expression" ):
 *                 listener.enterArgument_expression(self)
 */
  __pyx_t_124 = PyDict_New(); if (unlikely(!__pyx_t_124)) __PYX_ERR(0, 7253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_124);
  __pyx_t_123 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_123)) __PYX_ERR(0, 7253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_123);
  if (PyDict_SetItem(__pyx_t_124, __pyx_n_s_listener, __pyx_t_123) < 0) __PYX_ERR(0, 7253, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_123); __pyx_t_123 = 0;
  __pyx_t_123 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Argument_expressionContext_5enterRule, 0, __pyx_n_s_ModelicaParser_Argument_expressi_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2639)); if (unlikely(!__pyx_t_123)) __PYX_ERR(0, 7253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_123);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_123, __pyx_t_124);
  __Pyx_DECREF(__pyx_t_124); __pyx_t_124 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_123) < 0) __PYX_ERR(0, 7253, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_123); __pyx_t_123 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7257
 *                 listener.enterArgument_expression(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitArgument_expression" ):
 *                 listener.exitArgument_expression(self)
 */
  __pyx_t_123 = PyDict_New(); if (unlikely(!__pyx_t_123)) __PYX_ERR(0, 7257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_123);
  __pyx_t_124 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_124)) __PYX_ERR(0, 7257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_124);
  if (PyDict_SetItem(__pyx_t_123, __pyx_n_s_listener, __pyx_t_124) < 0) __PYX_ERR(0, 7257, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_124); __pyx_t_124 = 0;
  __pyx_t_124 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Argument_expressionContext_7exitRule, 0, __pyx_n_s_ModelicaParser_Argument_expressi_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2641)); if (unlikely(!__pyx_t_124)) __PYX_ERR(0, 7257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_124);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_124, __pyx_t_123);
  __Pyx_DECREF(__pyx_t_123); __pyx_t_123 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_124) < 0) __PYX_ERR(0, 7257, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_124); __pyx_t_124 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7261
 *                 listener.exitArgument_expression(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitArgument_expression" ):
 *                 return visitor.visitArgument_expression(self)
 */
  __pyx_t_124 = PyDict_New(); if (unlikely(!__pyx_t_124)) __PYX_ERR(0, 7261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_124);
  __pyx_t_123 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_123)) __PYX_ERR(0, 7261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_123);
  if (PyDict_SetItem(__pyx_t_124, __pyx_n_s_visitor, __pyx_t_123) < 0) __PYX_ERR(0, 7261, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_123); __pyx_t_123 = 0;
  __pyx_t_123 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_26Argument_expressionContext_9accept, 0, __pyx_n_s_ModelicaParser_Argument_expressi_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2643)); if (unlikely(!__pyx_t_123)) __PYX_ERR(0, 7261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_123);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_123, __pyx_t_124);
  __Pyx_DECREF(__pyx_t_124); __pyx_t_124 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_123) < 0) __PYX_ERR(0, 7261, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_123); __pyx_t_123 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7243
 * 
 * 
 *     class Argument_expressionContext(Function_argumentContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Function_argumentContext
 */
  __pyx_t_123 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Argument_expressionContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_123)) __PYX_ERR(0, 7243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_123);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_122, __pyx_t_123) < 0) __PYX_ERR(0, 7243, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_122); __pyx_t_122 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Argument_expressionContext, __pyx_t_123) < 0) __PYX_ERR(0, 7243, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_123); __pyx_t_123 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7268
 * 
 * 
 *     class Argument_functionContext(Function_argumentContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Function_argumentContext
 */
  __pyx_t_7 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_Function_argumentContext);
  if (unlikely(!__pyx_t_7)) {
    PyErr_Clear();
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_Function_argumentContext);
  }
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Argument_functionContext, __pyx_n_s_ModelicaParser_Argument_function, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_123 = PyList_New(0); if (unlikely(!__pyx_t_123)) __PYX_ERR(0, 7268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_123);

  /* "pymola/generated/ModelicaParser.pyx":7270
 *     class Argument_functionContext(Function_argumentContext):
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Function_argumentContext             # <<<<<<<<<<<<<<
 *             super().__init__(parser)
 *             self.copyFrom(ctx)
 */
  __pyx_t_124 = PyDict_New(); if (unlikely(!__pyx_t_124)) __PYX_ERR(0, 7270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_124);
  __pyx_t_125 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_125)) __PYX_ERR(0, 7270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_125);
  if (PyDict_SetItem(__pyx_t_124, __pyx_n_s_ctx_2, __pyx_t_125) < 0) __PYX_ERR(0, 7270, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_125); __pyx_t_125 = 0;
  __pyx_t_125 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Argument_functionContext_1__init__, 0, __pyx_n_s_ModelicaParser_Argument_function_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2645)); if (unlikely(!__pyx_t_125)) __PYX_ERR(0, 7270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_125);
  __Pyx_INCREF(__pyx_t_125);
  PyList_Append(__pyx_t_123, __pyx_t_125);
  __Pyx_GIVEREF(__pyx_t_125);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_125, __pyx_t_124);
  __Pyx_DECREF(__pyx_t_124); __pyx_t_124 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_125) < 0) __PYX_ERR(0, 7270, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_125); __pyx_t_125 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7274
 *             self.copyFrom(ctx)
 * 
 *         def name(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)
 * 
 */
  __pyx_t_125 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Argument_functionContext_3name, 0, __pyx_n_s_ModelicaParser_Argument_function_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2647)); if (unlikely(!__pyx_t_125)) __PYX_ERR(0, 7274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_125);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_name, __pyx_t_125) < 0) __PYX_ERR(0, 7274, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_125); __pyx_t_125 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7277
 *             return self.getTypedRuleContext(ModelicaParser.NameContext,0)
 * 
 *         def named_arguments(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Named_argumentsContext,0)
 * 
 */
  __pyx_t_125 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Argument_functionContext_5named_arguments, 0, __pyx_n_s_ModelicaParser_Argument_function_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2649)); if (unlikely(!__pyx_t_125)) __PYX_ERR(0, 7277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_125);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_named_arguments, __pyx_t_125) < 0) __PYX_ERR(0, 7277, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_125); __pyx_t_125 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7281
 * 
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterArgument_function" ):
 *                 listener.enterArgument_function(self)
 */
  __pyx_t_125 = PyDict_New(); if (unlikely(!__pyx_t_125)) __PYX_ERR(0, 7281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_125);
  __pyx_t_124 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_124)) __PYX_ERR(0, 7281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_124);
  if (PyDict_SetItem(__pyx_t_125, __pyx_n_s_listener, __pyx_t_124) < 0) __PYX_ERR(0, 7281, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_124); __pyx_t_124 = 0;
  __pyx_t_124 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Argument_functionContext_7enterRule, 0, __pyx_n_s_ModelicaParser_Argument_function_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2651)); if (unlikely(!__pyx_t_124)) __PYX_ERR(0, 7281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_124);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_124, __pyx_t_125);
  __Pyx_DECREF(__pyx_t_125); __pyx_t_125 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_124) < 0) __PYX_ERR(0, 7281, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_124); __pyx_t_124 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7285
 *                 listener.enterArgument_function(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitArgument_function" ):
 *                 listener.exitArgument_function(self)
 */
  __pyx_t_124 = PyDict_New(); if (unlikely(!__pyx_t_124)) __PYX_ERR(0, 7285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_124);
  __pyx_t_125 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_125)) __PYX_ERR(0, 7285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_125);
  if (PyDict_SetItem(__pyx_t_124, __pyx_n_s_listener, __pyx_t_125) < 0) __PYX_ERR(0, 7285, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_125); __pyx_t_125 = 0;
  __pyx_t_125 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Argument_functionContext_9exitRule, 0, __pyx_n_s_ModelicaParser_Argument_function_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2653)); if (unlikely(!__pyx_t_125)) __PYX_ERR(0, 7285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_125);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_125, __pyx_t_124);
  __Pyx_DECREF(__pyx_t_124); __pyx_t_124 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_125) < 0) __PYX_ERR(0, 7285, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_125); __pyx_t_125 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7289
 *                 listener.exitArgument_function(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitArgument_function" ):
 *                 return visitor.visitArgument_function(self)
 */
  __pyx_t_125 = PyDict_New(); if (unlikely(!__pyx_t_125)) __PYX_ERR(0, 7289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_125);
  __pyx_t_124 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_124)) __PYX_ERR(0, 7289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_124);
  if (PyDict_SetItem(__pyx_t_125, __pyx_n_s_visitor, __pyx_t_124) < 0) __PYX_ERR(0, 7289, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_124); __pyx_t_124 = 0;
  __pyx_t_124 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_24Argument_functionContext_11accept, 0, __pyx_n_s_ModelicaParser_Argument_function_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2655)); if (unlikely(!__pyx_t_124)) __PYX_ERR(0, 7289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_124);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_124, __pyx_t_125);
  __Pyx_DECREF(__pyx_t_125); __pyx_t_125 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_124) < 0) __PYX_ERR(0, 7289, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_124); __pyx_t_124 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7268
 * 
 * 
 *     class Argument_functionContext(Function_argumentContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, ctx:ParserRuleContext): # actually a ModelicaParser.Function_argumentContext
 */
  __pyx_t_124 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Argument_functionContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_124)) __PYX_ERR(0, 7268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_124);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_123, __pyx_t_124) < 0) __PYX_ERR(0, 7268, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_123); __pyx_t_123 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Argument_functionContext, __pyx_t_124) < 0) __PYX_ERR(0, 7268, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_124); __pyx_t_124 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7297
 * 
 * 
 *     def function_argument(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Function_argumentContext(self, self._ctx, self.state)
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_133function_argument, 0, __pyx_n_s_ModelicaParser_function_argument_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2657)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_function_argument, __pyx_t_5) < 0) __PYX_ERR(0, 7297, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7343
 *         return localctx
 * 
 *     class Output_expression_listContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Output_expression_listContext, __pyx_n_s_ModelicaParser_Output_expression, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_124 = PyList_New(0); if (unlikely(!__pyx_t_124)) __PYX_ERR(0, 7343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_124);

  /* "pymola/generated/ModelicaParser.pyx":7345
 *     class Output_expression_listContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_125 = PyDict_New(); if (unlikely(!__pyx_t_125)) __PYX_ERR(0, 7345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_125);
  __pyx_t_126 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_126)) __PYX_ERR(0, 7345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_126);
  if (PyDict_SetItem(__pyx_t_125, __pyx_n_s_parent, __pyx_t_126) < 0) __PYX_ERR(0, 7345, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_126); __pyx_t_126 = 0;
  if (PyDict_SetItem(__pyx_t_125, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 7345, __pyx_L1_error)
  __pyx_t_126 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_1__init__, 0, __pyx_n_s_ModelicaParser_Output_expression_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2659)); if (unlikely(!__pyx_t_126)) __PYX_ERR(0, 7345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_126);
  __Pyx_INCREF(__pyx_t_126);
  PyList_Append(__pyx_t_124, __pyx_t_126);
  __Pyx_GIVEREF(__pyx_t_126);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_126, __pyx_tuple__2660);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_126, __pyx_t_125);
  __Pyx_DECREF(__pyx_t_125); __pyx_t_125 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_126) < 0) __PYX_ERR(0, 7345, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_126); __pyx_t_126 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7349
 *             self.parser = parser
 * 
 *         def expression(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 */
  __pyx_t_126 = PyDict_New(); if (unlikely(!__pyx_t_126)) __PYX_ERR(0, 7349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_126);
  if (PyDict_SetItem(__pyx_t_126, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 7349, __pyx_L1_error)
  __pyx_t_125 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_3expression, 0, __pyx_n_s_ModelicaParser_Output_expression_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2662)); if (unlikely(!__pyx_t_125)) __PYX_ERR(0, 7349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_125);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_125, __pyx_tuple__2663);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_125, __pyx_t_126);
  __Pyx_DECREF(__pyx_t_126); __pyx_t_126 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_expression, __pyx_t_125) < 0) __PYX_ERR(0, 7349, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_125); __pyx_t_125 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7356
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_output_expression_list
 * 
 */
  __pyx_t_125 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_5getRuleIndex, 0, __pyx_n_s_ModelicaParser_Output_expression_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2665)); if (unlikely(!__pyx_t_125)) __PYX_ERR(0, 7356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_125);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_125) < 0) __PYX_ERR(0, 7356, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_125); __pyx_t_125 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7359
 *             return ModelicaParser.RULE_output_expression_list
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterOutput_expression_list" ):
 *                 listener.enterOutput_expression_list(self)
 */
  __pyx_t_125 = PyDict_New(); if (unlikely(!__pyx_t_125)) __PYX_ERR(0, 7359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_125);
  __pyx_t_126 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_126)) __PYX_ERR(0, 7359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_126);
  if (PyDict_SetItem(__pyx_t_125, __pyx_n_s_listener, __pyx_t_126) < 0) __PYX_ERR(0, 7359, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_126); __pyx_t_126 = 0;
  __pyx_t_126 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_7enterRule, 0, __pyx_n_s_ModelicaParser_Output_expression_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2667)); if (unlikely(!__pyx_t_126)) __PYX_ERR(0, 7359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_126);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_126, __pyx_t_125);
  __Pyx_DECREF(__pyx_t_125); __pyx_t_125 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_126) < 0) __PYX_ERR(0, 7359, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_126); __pyx_t_126 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7363
 *                 listener.enterOutput_expression_list(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitOutput_expression_list" ):
 *                 listener.exitOutput_expression_list(self)
 */
  __pyx_t_126 = PyDict_New(); if (unlikely(!__pyx_t_126)) __PYX_ERR(0, 7363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_126);
  __pyx_t_125 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_125)) __PYX_ERR(0, 7363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_125);
  if (PyDict_SetItem(__pyx_t_126, __pyx_n_s_listener, __pyx_t_125) < 0) __PYX_ERR(0, 7363, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_125); __pyx_t_125 = 0;
  __pyx_t_125 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_9exitRule, 0, __pyx_n_s_ModelicaParser_Output_expression_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2669)); if (unlikely(!__pyx_t_125)) __PYX_ERR(0, 7363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_125);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_125, __pyx_t_126);
  __Pyx_DECREF(__pyx_t_126); __pyx_t_126 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_125) < 0) __PYX_ERR(0, 7363, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_125); __pyx_t_125 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7367
 *                 listener.exitOutput_expression_list(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitOutput_expression_list" ):
 *                 return visitor.visitOutput_expression_list(self)
 */
  __pyx_t_125 = PyDict_New(); if (unlikely(!__pyx_t_125)) __PYX_ERR(0, 7367, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_125);
  __pyx_t_126 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_126)) __PYX_ERR(0, 7367, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_126);
  if (PyDict_SetItem(__pyx_t_125, __pyx_n_s_visitor, __pyx_t_126) < 0) __PYX_ERR(0, 7367, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_126); __pyx_t_126 = 0;
  __pyx_t_126 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_29Output_expression_listContext_11accept, 0, __pyx_n_s_ModelicaParser_Output_expression_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2671)); if (unlikely(!__pyx_t_126)) __PYX_ERR(0, 7367, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_126);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_126, __pyx_t_125);
  __Pyx_DECREF(__pyx_t_125); __pyx_t_125 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_126) < 0) __PYX_ERR(0, 7367, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_126); __pyx_t_126 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7343
 *         return localctx
 * 
 *     class Output_expression_listContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_126 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Output_expression_listContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_126)) __PYX_ERR(0, 7343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_126);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_124, __pyx_t_126) < 0) __PYX_ERR(0, 7343, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_124); __pyx_t_124 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Output_expression_listContext, __pyx_t_126) < 0) __PYX_ERR(0, 7343, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_126); __pyx_t_126 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7376
 * 
 * 
 *     def output_expression_list(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Output_expression_listContext(self, self._ctx, self.state)
 */
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_135output_expression_list, 0, __pyx_n_s_ModelicaParser_output_expression, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2673)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_output_expression_list, __pyx_t_7) < 0) __PYX_ERR(0, 7376, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7411
 *         return localctx
 * 
 *     class Expression_listContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7411, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7411, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7411, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_Expression_listContext, __pyx_n_s_ModelicaParser_Expression_listCo, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7411, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_126 = PyList_New(0); if (unlikely(!__pyx_t_126)) __PYX_ERR(0, 7411, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_126);

  /* "pymola/generated/ModelicaParser.pyx":7413
 *     class Expression_listContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_125 = PyDict_New(); if (unlikely(!__pyx_t_125)) __PYX_ERR(0, 7413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_125);
  __pyx_t_127 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_127)) __PYX_ERR(0, 7413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_127);
  if (PyDict_SetItem(__pyx_t_125, __pyx_n_s_parent, __pyx_t_127) < 0) __PYX_ERR(0, 7413, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_127); __pyx_t_127 = 0;
  if (PyDict_SetItem(__pyx_t_125, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 7413, __pyx_L1_error)
  __pyx_t_127 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Expression_listContext_1__init__, 0, __pyx_n_s_ModelicaParser_Expression_listCo_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2675)); if (unlikely(!__pyx_t_127)) __PYX_ERR(0, 7413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_127);
  __Pyx_INCREF(__pyx_t_127);
  PyList_Append(__pyx_t_126, __pyx_t_127);
  __Pyx_GIVEREF(__pyx_t_127);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_127, __pyx_tuple__2676);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_127, __pyx_t_125);
  __Pyx_DECREF(__pyx_t_125); __pyx_t_125 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_127) < 0) __PYX_ERR(0, 7413, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_127); __pyx_t_127 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7417
 *             self.parser = parser
 * 
 *         def expression(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.ExpressionContext)
 */
  __pyx_t_127 = PyDict_New(); if (unlikely(!__pyx_t_127)) __PYX_ERR(0, 7417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_127);
  if (PyDict_SetItem(__pyx_t_127, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 7417, __pyx_L1_error)
  __pyx_t_125 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Expression_listContext_3expression, 0, __pyx_n_s_ModelicaParser_Expression_listCo_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2678)); if (unlikely(!__pyx_t_125)) __PYX_ERR(0, 7417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_125);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_125, __pyx_tuple__2679);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_125, __pyx_t_127);
  __Pyx_DECREF(__pyx_t_127); __pyx_t_127 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_expression, __pyx_t_125) < 0) __PYX_ERR(0, 7417, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_125); __pyx_t_125 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7424
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_expression_list
 * 
 */
  __pyx_t_125 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Expression_listContext_5getRuleIndex, 0, __pyx_n_s_ModelicaParser_Expression_listCo_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2681)); if (unlikely(!__pyx_t_125)) __PYX_ERR(0, 7424, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_125);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_125) < 0) __PYX_ERR(0, 7424, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_125); __pyx_t_125 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7427
 *             return ModelicaParser.RULE_expression_list
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterExpression_list" ):
 *                 listener.enterExpression_list(self)
 */
  __pyx_t_125 = PyDict_New(); if (unlikely(!__pyx_t_125)) __PYX_ERR(0, 7427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_125);
  __pyx_t_127 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_127)) __PYX_ERR(0, 7427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_127);
  if (PyDict_SetItem(__pyx_t_125, __pyx_n_s_listener, __pyx_t_127) < 0) __PYX_ERR(0, 7427, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_127); __pyx_t_127 = 0;
  __pyx_t_127 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Expression_listContext_7enterRule, 0, __pyx_n_s_ModelicaParser_Expression_listCo_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2683)); if (unlikely(!__pyx_t_127)) __PYX_ERR(0, 7427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_127);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_127, __pyx_t_125);
  __Pyx_DECREF(__pyx_t_125); __pyx_t_125 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_127) < 0) __PYX_ERR(0, 7427, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_127); __pyx_t_127 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7431
 *                 listener.enterExpression_list(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitExpression_list" ):
 *                 listener.exitExpression_list(self)
 */
  __pyx_t_127 = PyDict_New(); if (unlikely(!__pyx_t_127)) __PYX_ERR(0, 7431, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_127);
  __pyx_t_125 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_125)) __PYX_ERR(0, 7431, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_125);
  if (PyDict_SetItem(__pyx_t_127, __pyx_n_s_listener, __pyx_t_125) < 0) __PYX_ERR(0, 7431, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_125); __pyx_t_125 = 0;
  __pyx_t_125 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Expression_listContext_9exitRule, 0, __pyx_n_s_ModelicaParser_Expression_listCo_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2685)); if (unlikely(!__pyx_t_125)) __PYX_ERR(0, 7431, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_125);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_125, __pyx_t_127);
  __Pyx_DECREF(__pyx_t_127); __pyx_t_127 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_125) < 0) __PYX_ERR(0, 7431, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_125); __pyx_t_125 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7435
 *                 listener.exitExpression_list(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitExpression_list" ):
 *                 return visitor.visitExpression_list(self)
 */
  __pyx_t_125 = PyDict_New(); if (unlikely(!__pyx_t_125)) __PYX_ERR(0, 7435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_125);
  __pyx_t_127 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_127)) __PYX_ERR(0, 7435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_127);
  if (PyDict_SetItem(__pyx_t_125, __pyx_n_s_visitor, __pyx_t_127) < 0) __PYX_ERR(0, 7435, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_127); __pyx_t_127 = 0;
  __pyx_t_127 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_22Expression_listContext_11accept, 0, __pyx_n_s_ModelicaParser_Expression_listCo_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2687)); if (unlikely(!__pyx_t_127)) __PYX_ERR(0, 7435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_127);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_127, __pyx_t_125);
  __Pyx_DECREF(__pyx_t_125); __pyx_t_125 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_127) < 0) __PYX_ERR(0, 7435, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_127); __pyx_t_127 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7411
 *         return localctx
 * 
 *     class Expression_listContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_127 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_Expression_listContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_127)) __PYX_ERR(0, 7411, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_127);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_126, __pyx_t_127) < 0) __PYX_ERR(0, 7411, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_126); __pyx_t_126 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Expression_listContext, __pyx_t_127) < 0) __PYX_ERR(0, 7411, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_127); __pyx_t_127 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7444
 * 
 * 
 *     def expression_list(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Expression_listContext(self, self._ctx, self.state)
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_137expression_list, 0, __pyx_n_s_ModelicaParser_expression_list, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2689)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7444, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_expression_list, __pyx_t_5) < 0) __PYX_ERR(0, 7444, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7473
 *         return localctx
 * 
 *     class Array_subscriptsContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_Array_subscriptsContext, __pyx_n_s_ModelicaParser_Array_subscriptsC, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_127 = PyList_New(0); if (unlikely(!__pyx_t_127)) __PYX_ERR(0, 7473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_127);

  /* "pymola/generated/ModelicaParser.pyx":7475
 *     class Array_subscriptsContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_125 = PyDict_New(); if (unlikely(!__pyx_t_125)) __PYX_ERR(0, 7475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_125);
  __pyx_t_128 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_128)) __PYX_ERR(0, 7475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_128);
  if (PyDict_SetItem(__pyx_t_125, __pyx_n_s_parent, __pyx_t_128) < 0) __PYX_ERR(0, 7475, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_128); __pyx_t_128 = 0;
  if (PyDict_SetItem(__pyx_t_125, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 7475, __pyx_L1_error)
  __pyx_t_128 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_1__init__, 0, __pyx_n_s_ModelicaParser_Array_subscriptsC_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2691)); if (unlikely(!__pyx_t_128)) __PYX_ERR(0, 7475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_128);
  __Pyx_INCREF(__pyx_t_128);
  PyList_Append(__pyx_t_127, __pyx_t_128);
  __Pyx_GIVEREF(__pyx_t_128);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_128, __pyx_tuple__2692);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_128, __pyx_t_125);
  __Pyx_DECREF(__pyx_t_125); __pyx_t_125 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_128) < 0) __PYX_ERR(0, 7475, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_128); __pyx_t_128 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7479
 *             self.parser = parser
 * 
 *         def subscript(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTypedRuleContexts(ModelicaParser.SubscriptContext)
 */
  __pyx_t_128 = PyDict_New(); if (unlikely(!__pyx_t_128)) __PYX_ERR(0, 7479, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_128);
  if (PyDict_SetItem(__pyx_t_128, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 7479, __pyx_L1_error)
  __pyx_t_125 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_3subscript, 0, __pyx_n_s_ModelicaParser_Array_subscriptsC_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2694)); if (unlikely(!__pyx_t_125)) __PYX_ERR(0, 7479, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_125);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_125, __pyx_tuple__2695);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_125, __pyx_t_128);
  __Pyx_DECREF(__pyx_t_128); __pyx_t_128 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_subscript, __pyx_t_125) < 0) __PYX_ERR(0, 7479, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_125); __pyx_t_125 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7486
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_array_subscripts
 * 
 */
  __pyx_t_125 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_5getRuleIndex, 0, __pyx_n_s_ModelicaParser_Array_subscriptsC_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2697)); if (unlikely(!__pyx_t_125)) __PYX_ERR(0, 7486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_125);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_125) < 0) __PYX_ERR(0, 7486, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_125); __pyx_t_125 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7489
 *             return ModelicaParser.RULE_array_subscripts
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterArray_subscripts" ):
 *                 listener.enterArray_subscripts(self)
 */
  __pyx_t_125 = PyDict_New(); if (unlikely(!__pyx_t_125)) __PYX_ERR(0, 7489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_125);
  __pyx_t_128 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_128)) __PYX_ERR(0, 7489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_128);
  if (PyDict_SetItem(__pyx_t_125, __pyx_n_s_listener, __pyx_t_128) < 0) __PYX_ERR(0, 7489, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_128); __pyx_t_128 = 0;
  __pyx_t_128 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_7enterRule, 0, __pyx_n_s_ModelicaParser_Array_subscriptsC_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2699)); if (unlikely(!__pyx_t_128)) __PYX_ERR(0, 7489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_128);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_128, __pyx_t_125);
  __Pyx_DECREF(__pyx_t_125); __pyx_t_125 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_128) < 0) __PYX_ERR(0, 7489, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_128); __pyx_t_128 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7493
 *                 listener.enterArray_subscripts(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitArray_subscripts" ):
 *                 listener.exitArray_subscripts(self)
 */
  __pyx_t_128 = PyDict_New(); if (unlikely(!__pyx_t_128)) __PYX_ERR(0, 7493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_128);
  __pyx_t_125 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_125)) __PYX_ERR(0, 7493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_125);
  if (PyDict_SetItem(__pyx_t_128, __pyx_n_s_listener, __pyx_t_125) < 0) __PYX_ERR(0, 7493, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_125); __pyx_t_125 = 0;
  __pyx_t_125 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_9exitRule, 0, __pyx_n_s_ModelicaParser_Array_subscriptsC_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2701)); if (unlikely(!__pyx_t_125)) __PYX_ERR(0, 7493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_125);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_125, __pyx_t_128);
  __Pyx_DECREF(__pyx_t_128); __pyx_t_128 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_125) < 0) __PYX_ERR(0, 7493, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_125); __pyx_t_125 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7497
 *                 listener.exitArray_subscripts(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitArray_subscripts" ):
 *                 return visitor.visitArray_subscripts(self)
 */
  __pyx_t_125 = PyDict_New(); if (unlikely(!__pyx_t_125)) __PYX_ERR(0, 7497, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_125);
  __pyx_t_128 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_128)) __PYX_ERR(0, 7497, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_128);
  if (PyDict_SetItem(__pyx_t_125, __pyx_n_s_visitor, __pyx_t_128) < 0) __PYX_ERR(0, 7497, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_128); __pyx_t_128 = 0;
  __pyx_t_128 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_23Array_subscriptsContext_11accept, 0, __pyx_n_s_ModelicaParser_Array_subscriptsC_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2703)); if (unlikely(!__pyx_t_128)) __PYX_ERR(0, 7497, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_128);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_128, __pyx_t_125);
  __Pyx_DECREF(__pyx_t_125); __pyx_t_125 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_128) < 0) __PYX_ERR(0, 7497, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_128); __pyx_t_128 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7473
 *         return localctx
 * 
 *     class Array_subscriptsContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_128 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Array_subscriptsContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_128)) __PYX_ERR(0, 7473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_128);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_127, __pyx_t_128) < 0) __PYX_ERR(0, 7473, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_127); __pyx_t_127 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_Array_subscriptsContext, __pyx_t_128) < 0) __PYX_ERR(0, 7473, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_128); __pyx_t_128 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7506
 * 
 * 
 *     def array_subscripts(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.Array_subscriptsContext(self, self._ctx, self.state)
 */
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_139array_subscripts, 0, __pyx_n_s_ModelicaParser_array_subscripts, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2705)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7506, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_array_subscripts, __pyx_t_7) < 0) __PYX_ERR(0, 7506, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7539
 *         return localctx
 * 
 *     class SubscriptContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7539, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7539, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7539, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_SubscriptContext, __pyx_n_s_ModelicaParser_SubscriptContext, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7539, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_128 = PyList_New(0); if (unlikely(!__pyx_t_128)) __PYX_ERR(0, 7539, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_128);

  /* "pymola/generated/ModelicaParser.pyx":7541
 *     class SubscriptContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_125 = PyDict_New(); if (unlikely(!__pyx_t_125)) __PYX_ERR(0, 7541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_125);
  __pyx_t_129 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_129)) __PYX_ERR(0, 7541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_129);
  if (PyDict_SetItem(__pyx_t_125, __pyx_n_s_parent, __pyx_t_129) < 0) __PYX_ERR(0, 7541, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_129); __pyx_t_129 = 0;
  if (PyDict_SetItem(__pyx_t_125, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 7541, __pyx_L1_error)
  __pyx_t_129 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_16SubscriptContext_1__init__, 0, __pyx_n_s_ModelicaParser_SubscriptContext_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2707)); if (unlikely(!__pyx_t_129)) __PYX_ERR(0, 7541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_129);
  __Pyx_INCREF(__pyx_t_129);
  PyList_Append(__pyx_t_128, __pyx_t_129);
  __Pyx_GIVEREF(__pyx_t_129);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_129, __pyx_tuple__2708);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_129, __pyx_t_125);
  __Pyx_DECREF(__pyx_t_125); __pyx_t_125 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_129) < 0) __PYX_ERR(0, 7541, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_129); __pyx_t_129 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7545
 *             self.parser = parser
 * 
 *         def expression(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.ExpressionContext,0)
 * 
 */
  __pyx_t_129 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_16SubscriptContext_3expression, 0, __pyx_n_s_ModelicaParser_SubscriptContext_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2710)); if (unlikely(!__pyx_t_129)) __PYX_ERR(0, 7545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_129);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_expression, __pyx_t_129) < 0) __PYX_ERR(0, 7545, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_129); __pyx_t_129 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7549
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_subscript
 * 
 */
  __pyx_t_129 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_16SubscriptContext_5getRuleIndex, 0, __pyx_n_s_ModelicaParser_SubscriptContext_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2712)); if (unlikely(!__pyx_t_129)) __PYX_ERR(0, 7549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_129);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_129) < 0) __PYX_ERR(0, 7549, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_129); __pyx_t_129 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7552
 *             return ModelicaParser.RULE_subscript
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterSubscript" ):
 *                 listener.enterSubscript(self)
 */
  __pyx_t_129 = PyDict_New(); if (unlikely(!__pyx_t_129)) __PYX_ERR(0, 7552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_129);
  __pyx_t_125 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_125)) __PYX_ERR(0, 7552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_125);
  if (PyDict_SetItem(__pyx_t_129, __pyx_n_s_listener, __pyx_t_125) < 0) __PYX_ERR(0, 7552, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_125); __pyx_t_125 = 0;
  __pyx_t_125 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_16SubscriptContext_7enterRule, 0, __pyx_n_s_ModelicaParser_SubscriptContext_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2714)); if (unlikely(!__pyx_t_125)) __PYX_ERR(0, 7552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_125);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_125, __pyx_t_129);
  __Pyx_DECREF(__pyx_t_129); __pyx_t_129 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_125) < 0) __PYX_ERR(0, 7552, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_125); __pyx_t_125 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7556
 *                 listener.enterSubscript(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitSubscript" ):
 *                 listener.exitSubscript(self)
 */
  __pyx_t_125 = PyDict_New(); if (unlikely(!__pyx_t_125)) __PYX_ERR(0, 7556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_125);
  __pyx_t_129 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_129)) __PYX_ERR(0, 7556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_129);
  if (PyDict_SetItem(__pyx_t_125, __pyx_n_s_listener, __pyx_t_129) < 0) __PYX_ERR(0, 7556, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_129); __pyx_t_129 = 0;
  __pyx_t_129 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_16SubscriptContext_9exitRule, 0, __pyx_n_s_ModelicaParser_SubscriptContext_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2716)); if (unlikely(!__pyx_t_129)) __PYX_ERR(0, 7556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_129);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_129, __pyx_t_125);
  __Pyx_DECREF(__pyx_t_125); __pyx_t_125 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_129) < 0) __PYX_ERR(0, 7556, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_129); __pyx_t_129 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7560
 *                 listener.exitSubscript(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitSubscript" ):
 *                 return visitor.visitSubscript(self)
 */
  __pyx_t_129 = PyDict_New(); if (unlikely(!__pyx_t_129)) __PYX_ERR(0, 7560, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_129);
  __pyx_t_125 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_125)) __PYX_ERR(0, 7560, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_125);
  if (PyDict_SetItem(__pyx_t_129, __pyx_n_s_visitor, __pyx_t_125) < 0) __PYX_ERR(0, 7560, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_125); __pyx_t_125 = 0;
  __pyx_t_125 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_16SubscriptContext_11accept, 0, __pyx_n_s_ModelicaParser_SubscriptContext_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2718)); if (unlikely(!__pyx_t_125)) __PYX_ERR(0, 7560, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_125);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_125, __pyx_t_129);
  __Pyx_DECREF(__pyx_t_129); __pyx_t_129 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_125) < 0) __PYX_ERR(0, 7560, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_125); __pyx_t_125 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7539
 *         return localctx
 * 
 *     class SubscriptContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_125 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_SubscriptContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_125)) __PYX_ERR(0, 7539, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_125);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_128, __pyx_t_125) < 0) __PYX_ERR(0, 7539, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_128); __pyx_t_128 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_SubscriptContext, __pyx_t_125) < 0) __PYX_ERR(0, 7539, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_125); __pyx_t_125 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7569
 * 
 * 
 *     def subscript(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.SubscriptContext(self, self._ctx, self.state)
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_141subscript, 0, __pyx_n_s_ModelicaParser_subscript, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2720)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_subscript, __pyx_t_5) < 0) __PYX_ERR(0, 7569, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7598
 *         return localctx
 * 
 *     class CommentContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7598, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7598, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7598, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_CommentContext, __pyx_n_s_ModelicaParser_CommentContext, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7598, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_125 = PyList_New(0); if (unlikely(!__pyx_t_125)) __PYX_ERR(0, 7598, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_125);

  /* "pymola/generated/ModelicaParser.pyx":7600
 *     class CommentContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_129 = PyDict_New(); if (unlikely(!__pyx_t_129)) __PYX_ERR(0, 7600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_129);
  __pyx_t_130 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_130)) __PYX_ERR(0, 7600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_130);
  if (PyDict_SetItem(__pyx_t_129, __pyx_n_s_parent, __pyx_t_130) < 0) __PYX_ERR(0, 7600, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_130); __pyx_t_130 = 0;
  if (PyDict_SetItem(__pyx_t_129, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 7600, __pyx_L1_error)
  __pyx_t_130 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14CommentContext_1__init__, 0, __pyx_n_s_ModelicaParser_CommentContext_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2722)); if (unlikely(!__pyx_t_130)) __PYX_ERR(0, 7600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_130);
  __Pyx_INCREF(__pyx_t_130);
  PyList_Append(__pyx_t_125, __pyx_t_130);
  __Pyx_GIVEREF(__pyx_t_130);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_130, __pyx_tuple__2723);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_130, __pyx_t_129);
  __Pyx_DECREF(__pyx_t_129); __pyx_t_129 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_130) < 0) __PYX_ERR(0, 7600, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_130); __pyx_t_130 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7604
 *             self.parser = parser
 * 
 *         def string_comment(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.String_commentContext,0)
 * 
 */
  __pyx_t_130 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14CommentContext_3string_comment, 0, __pyx_n_s_ModelicaParser_CommentContext_st, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2725)); if (unlikely(!__pyx_t_130)) __PYX_ERR(0, 7604, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_130);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_string_comment, __pyx_t_130) < 0) __PYX_ERR(0, 7604, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_130); __pyx_t_130 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7608
 * 
 * 
 *         def annotation(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.AnnotationContext,0)
 * 
 */
  __pyx_t_130 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14CommentContext_5annotation, 0, __pyx_n_s_ModelicaParser_CommentContext_an, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2727)); if (unlikely(!__pyx_t_130)) __PYX_ERR(0, 7608, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_130);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_annotation, __pyx_t_130) < 0) __PYX_ERR(0, 7608, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_130); __pyx_t_130 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7612
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_comment
 * 
 */
  __pyx_t_130 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14CommentContext_7getRuleIndex, 0, __pyx_n_s_ModelicaParser_CommentContext_ge, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2729)); if (unlikely(!__pyx_t_130)) __PYX_ERR(0, 7612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_130);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_130) < 0) __PYX_ERR(0, 7612, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_130); __pyx_t_130 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7615
 *             return ModelicaParser.RULE_comment
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterComment" ):
 *                 listener.enterComment(self)
 */
  __pyx_t_130 = PyDict_New(); if (unlikely(!__pyx_t_130)) __PYX_ERR(0, 7615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_130);
  __pyx_t_129 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_129)) __PYX_ERR(0, 7615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_129);
  if (PyDict_SetItem(__pyx_t_130, __pyx_n_s_listener, __pyx_t_129) < 0) __PYX_ERR(0, 7615, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_129); __pyx_t_129 = 0;
  __pyx_t_129 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14CommentContext_9enterRule, 0, __pyx_n_s_ModelicaParser_CommentContext_en, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2731)); if (unlikely(!__pyx_t_129)) __PYX_ERR(0, 7615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_129);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_129, __pyx_t_130);
  __Pyx_DECREF(__pyx_t_130); __pyx_t_130 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_129) < 0) __PYX_ERR(0, 7615, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_129); __pyx_t_129 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7619
 *                 listener.enterComment(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitComment" ):
 *                 listener.exitComment(self)
 */
  __pyx_t_129 = PyDict_New(); if (unlikely(!__pyx_t_129)) __PYX_ERR(0, 7619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_129);
  __pyx_t_130 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_130)) __PYX_ERR(0, 7619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_130);
  if (PyDict_SetItem(__pyx_t_129, __pyx_n_s_listener, __pyx_t_130) < 0) __PYX_ERR(0, 7619, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_130); __pyx_t_130 = 0;
  __pyx_t_130 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14CommentContext_11exitRule, 0, __pyx_n_s_ModelicaParser_CommentContext_ex, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2733)); if (unlikely(!__pyx_t_130)) __PYX_ERR(0, 7619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_130);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_130, __pyx_t_129);
  __Pyx_DECREF(__pyx_t_129); __pyx_t_129 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_130) < 0) __PYX_ERR(0, 7619, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_130); __pyx_t_130 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7623
 *                 listener.exitComment(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitComment" ):
 *                 return visitor.visitComment(self)
 */
  __pyx_t_130 = PyDict_New(); if (unlikely(!__pyx_t_130)) __PYX_ERR(0, 7623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_130);
  __pyx_t_129 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_129)) __PYX_ERR(0, 7623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_129);
  if (PyDict_SetItem(__pyx_t_130, __pyx_n_s_visitor, __pyx_t_129) < 0) __PYX_ERR(0, 7623, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_129); __pyx_t_129 = 0;
  __pyx_t_129 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14CommentContext_13accept, 0, __pyx_n_s_ModelicaParser_CommentContext_ac, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2735)); if (unlikely(!__pyx_t_129)) __PYX_ERR(0, 7623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_129);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_129, __pyx_t_130);
  __Pyx_DECREF(__pyx_t_130); __pyx_t_130 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_129) < 0) __PYX_ERR(0, 7623, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_129); __pyx_t_129 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7598
 *         return localctx
 * 
 *     class CommentContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_129 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_CommentContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_129)) __PYX_ERR(0, 7598, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_129);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_125, __pyx_t_129) < 0) __PYX_ERR(0, 7598, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_125); __pyx_t_125 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_CommentContext, __pyx_t_129) < 0) __PYX_ERR(0, 7598, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_129); __pyx_t_129 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7632
 * 
 * 
 *     def comment(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.CommentContext(self, self._ctx, self.state)
 */
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_143comment, 0, __pyx_n_s_ModelicaParser_comment, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2737)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_comment, __pyx_t_7) < 0) __PYX_ERR(0, 7632, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7657
 *         return localctx
 * 
 *     class String_commentContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_5, __pyx_n_s_String_commentContext, __pyx_n_s_ModelicaParser_String_commentCon, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_129 = PyList_New(0); if (unlikely(!__pyx_t_129)) __PYX_ERR(0, 7657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_129);

  /* "pymola/generated/ModelicaParser.pyx":7659
 *     class String_commentContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_130 = PyDict_New(); if (unlikely(!__pyx_t_130)) __PYX_ERR(0, 7659, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_130);
  __pyx_t_131 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_131)) __PYX_ERR(0, 7659, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_131);
  if (PyDict_SetItem(__pyx_t_130, __pyx_n_s_parent, __pyx_t_131) < 0) __PYX_ERR(0, 7659, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_131); __pyx_t_131 = 0;
  if (PyDict_SetItem(__pyx_t_130, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 7659, __pyx_L1_error)
  __pyx_t_131 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21String_commentContext_1__init__, 0, __pyx_n_s_ModelicaParser_String_commentCon_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2739)); if (unlikely(!__pyx_t_131)) __PYX_ERR(0, 7659, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_131);
  __Pyx_INCREF(__pyx_t_131);
  PyList_Append(__pyx_t_129, __pyx_t_131);
  __Pyx_GIVEREF(__pyx_t_131);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_131, __pyx_tuple__2740);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_131, __pyx_t_130);
  __Pyx_DECREF(__pyx_t_130); __pyx_t_130 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_131) < 0) __PYX_ERR(0, 7659, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_131); __pyx_t_131 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7663
 *             self.parser = parser
 * 
 *         def STRING(self, i:int=None):             # <<<<<<<<<<<<<<
 *             if i is None:
 *                 return self.getTokens(ModelicaParser.STRING)
 */
  __pyx_t_131 = PyDict_New(); if (unlikely(!__pyx_t_131)) __PYX_ERR(0, 7663, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_131);
  if (PyDict_SetItem(__pyx_t_131, __pyx_n_s_i, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 7663, __pyx_L1_error)
  __pyx_t_130 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21String_commentContext_3STRING, 0, __pyx_n_s_ModelicaParser_String_commentCon_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2742)); if (unlikely(!__pyx_t_130)) __PYX_ERR(0, 7663, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_130);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_130, __pyx_tuple__2743);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_130, __pyx_t_131);
  __Pyx_DECREF(__pyx_t_131); __pyx_t_131 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_STRING, __pyx_t_130) < 0) __PYX_ERR(0, 7663, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_130); __pyx_t_130 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7669
 *                 return self.getToken(ModelicaParser.STRING, i)
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_string_comment
 * 
 */
  __pyx_t_130 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21String_commentContext_5getRuleIndex, 0, __pyx_n_s_ModelicaParser_String_commentCon_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2745)); if (unlikely(!__pyx_t_130)) __PYX_ERR(0, 7669, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_130);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_130) < 0) __PYX_ERR(0, 7669, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_130); __pyx_t_130 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7672
 *             return ModelicaParser.RULE_string_comment
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterString_comment" ):
 *                 listener.enterString_comment(self)
 */
  __pyx_t_130 = PyDict_New(); if (unlikely(!__pyx_t_130)) __PYX_ERR(0, 7672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_130);
  __pyx_t_131 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_131)) __PYX_ERR(0, 7672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_131);
  if (PyDict_SetItem(__pyx_t_130, __pyx_n_s_listener, __pyx_t_131) < 0) __PYX_ERR(0, 7672, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_131); __pyx_t_131 = 0;
  __pyx_t_131 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21String_commentContext_7enterRule, 0, __pyx_n_s_ModelicaParser_String_commentCon_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2747)); if (unlikely(!__pyx_t_131)) __PYX_ERR(0, 7672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_131);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_131, __pyx_t_130);
  __Pyx_DECREF(__pyx_t_130); __pyx_t_130 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_131) < 0) __PYX_ERR(0, 7672, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_131); __pyx_t_131 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7676
 *                 listener.enterString_comment(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitString_comment" ):
 *                 listener.exitString_comment(self)
 */
  __pyx_t_131 = PyDict_New(); if (unlikely(!__pyx_t_131)) __PYX_ERR(0, 7676, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_131);
  __pyx_t_130 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_130)) __PYX_ERR(0, 7676, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_130);
  if (PyDict_SetItem(__pyx_t_131, __pyx_n_s_listener, __pyx_t_130) < 0) __PYX_ERR(0, 7676, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_130); __pyx_t_130 = 0;
  __pyx_t_130 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21String_commentContext_9exitRule, 0, __pyx_n_s_ModelicaParser_String_commentCon_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2749)); if (unlikely(!__pyx_t_130)) __PYX_ERR(0, 7676, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_130);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_130, __pyx_t_131);
  __Pyx_DECREF(__pyx_t_131); __pyx_t_131 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_130) < 0) __PYX_ERR(0, 7676, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_130); __pyx_t_130 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7680
 *                 listener.exitString_comment(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitString_comment" ):
 *                 return visitor.visitString_comment(self)
 */
  __pyx_t_130 = PyDict_New(); if (unlikely(!__pyx_t_130)) __PYX_ERR(0, 7680, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_130);
  __pyx_t_131 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_131)) __PYX_ERR(0, 7680, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_131);
  if (PyDict_SetItem(__pyx_t_130, __pyx_n_s_visitor, __pyx_t_131) < 0) __PYX_ERR(0, 7680, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_131); __pyx_t_131 = 0;
  __pyx_t_131 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_21String_commentContext_11accept, 0, __pyx_n_s_ModelicaParser_String_commentCon_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2751)); if (unlikely(!__pyx_t_131)) __PYX_ERR(0, 7680, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_131);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_131, __pyx_t_130);
  __Pyx_DECREF(__pyx_t_130); __pyx_t_130 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_131) < 0) __PYX_ERR(0, 7680, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_131); __pyx_t_131 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7657
 *         return localctx
 * 
 *     class String_commentContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_131 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_String_commentContext, __pyx_t_5, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_131)) __PYX_ERR(0, 7657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_131);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_129, __pyx_t_131) < 0) __PYX_ERR(0, 7657, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_129); __pyx_t_129 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_String_commentContext, __pyx_t_131) < 0) __PYX_ERR(0, 7657, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_131); __pyx_t_131 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7689
 * 
 * 
 *     def string_comment(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.String_commentContext(self, self._ctx, self.state)
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_145string_comment, 0, __pyx_n_s_ModelicaParser_string_comment, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2753)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7689, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_string_comment, __pyx_t_5) < 0) __PYX_ERR(0, 7689, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7724
 *         return localctx
 * 
 *     class AnnotationContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7724, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7724, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7724, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_7, __pyx_n_s_AnnotationContext, __pyx_n_s_ModelicaParser_AnnotationContext, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaParser, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 7724, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_131 = PyList_New(0); if (unlikely(!__pyx_t_131)) __PYX_ERR(0, 7724, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_131);

  /* "pymola/generated/ModelicaParser.pyx":7726
 *     class AnnotationContext(ParserRuleContext):
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):             # <<<<<<<<<<<<<<
 *             super().__init__(parent, invokingState)
 *             self.parser = parser
 */
  __pyx_t_130 = PyDict_New(); if (unlikely(!__pyx_t_130)) __PYX_ERR(0, 7726, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_130);
  __pyx_t_132 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParserRuleContext); if (unlikely(!__pyx_t_132)) __PYX_ERR(0, 7726, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_132);
  if (PyDict_SetItem(__pyx_t_130, __pyx_n_s_parent, __pyx_t_132) < 0) __PYX_ERR(0, 7726, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_132); __pyx_t_132 = 0;
  if (PyDict_SetItem(__pyx_t_130, __pyx_n_s_invokingState, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 7726, __pyx_L1_error)
  __pyx_t_132 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_17AnnotationContext_1__init__, 0, __pyx_n_s_ModelicaParser_AnnotationContext_2, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2755)); if (unlikely(!__pyx_t_132)) __PYX_ERR(0, 7726, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_132);
  __Pyx_INCREF(__pyx_t_132);
  PyList_Append(__pyx_t_131, __pyx_t_132);
  __Pyx_GIVEREF(__pyx_t_132);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_132, __pyx_tuple__2756);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_132, __pyx_t_130);
  __Pyx_DECREF(__pyx_t_130); __pyx_t_130 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_init, __pyx_t_132) < 0) __PYX_ERR(0, 7726, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_132); __pyx_t_132 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7730
 *             self.parser = parser
 * 
 *         def class_modification(self):             # <<<<<<<<<<<<<<
 *             return self.getTypedRuleContext(ModelicaParser.Class_modificationContext,0)
 * 
 */
  __pyx_t_132 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_17AnnotationContext_3class_modification, 0, __pyx_n_s_ModelicaParser_AnnotationContext_3, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2758)); if (unlikely(!__pyx_t_132)) __PYX_ERR(0, 7730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_132);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_class_modification, __pyx_t_132) < 0) __PYX_ERR(0, 7730, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_132); __pyx_t_132 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7734
 * 
 * 
 *         def getRuleIndex(self):             # <<<<<<<<<<<<<<
 *             return ModelicaParser.RULE_annotation
 * 
 */
  __pyx_t_132 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_17AnnotationContext_5getRuleIndex, 0, __pyx_n_s_ModelicaParser_AnnotationContext_4, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2760)); if (unlikely(!__pyx_t_132)) __PYX_ERR(0, 7734, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_132);
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_getRuleIndex, __pyx_t_132) < 0) __PYX_ERR(0, 7734, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_132); __pyx_t_132 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7737
 *             return ModelicaParser.RULE_annotation
 * 
 *         def enterRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "enterAnnotation" ):
 *                 listener.enterAnnotation(self)
 */
  __pyx_t_132 = PyDict_New(); if (unlikely(!__pyx_t_132)) __PYX_ERR(0, 7737, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_132);
  __pyx_t_130 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_130)) __PYX_ERR(0, 7737, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_130);
  if (PyDict_SetItem(__pyx_t_132, __pyx_n_s_listener, __pyx_t_130) < 0) __PYX_ERR(0, 7737, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_130); __pyx_t_130 = 0;
  __pyx_t_130 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_17AnnotationContext_7enterRule, 0, __pyx_n_s_ModelicaParser_AnnotationContext_5, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2762)); if (unlikely(!__pyx_t_130)) __PYX_ERR(0, 7737, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_130);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_130, __pyx_t_132);
  __Pyx_DECREF(__pyx_t_132); __pyx_t_132 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRule, __pyx_t_130) < 0) __PYX_ERR(0, 7737, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_130); __pyx_t_130 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7741
 *                 listener.enterAnnotation(self)
 * 
 *         def exitRule(self, listener:ParseTreeListener):             # <<<<<<<<<<<<<<
 *             if hasattr( listener, "exitAnnotation" ):
 *                 listener.exitAnnotation(self)
 */
  __pyx_t_130 = PyDict_New(); if (unlikely(!__pyx_t_130)) __PYX_ERR(0, 7741, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_130);
  __pyx_t_132 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_132)) __PYX_ERR(0, 7741, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_132);
  if (PyDict_SetItem(__pyx_t_130, __pyx_n_s_listener, __pyx_t_132) < 0) __PYX_ERR(0, 7741, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_132); __pyx_t_132 = 0;
  __pyx_t_132 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_17AnnotationContext_9exitRule, 0, __pyx_n_s_ModelicaParser_AnnotationContext_6, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2764)); if (unlikely(!__pyx_t_132)) __PYX_ERR(0, 7741, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_132);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_132, __pyx_t_130);
  __Pyx_DECREF(__pyx_t_130); __pyx_t_130 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRule, __pyx_t_132) < 0) __PYX_ERR(0, 7741, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_132); __pyx_t_132 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7745
 *                 listener.exitAnnotation(self)
 * 
 *         def accept(self, visitor:ParseTreeVisitor):             # <<<<<<<<<<<<<<
 *             if hasattr( visitor, "visitAnnotation" ):
 *                 return visitor.visitAnnotation(self)
 */
  __pyx_t_132 = PyDict_New(); if (unlikely(!__pyx_t_132)) __PYX_ERR(0, 7745, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_132);
  __pyx_t_130 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeVisitor); if (unlikely(!__pyx_t_130)) __PYX_ERR(0, 7745, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_130);
  if (PyDict_SetItem(__pyx_t_132, __pyx_n_s_visitor, __pyx_t_130) < 0) __PYX_ERR(0, 7745, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_130); __pyx_t_130 = 0;
  __pyx_t_130 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_17AnnotationContext_11accept, 0, __pyx_n_s_ModelicaParser_AnnotationContext_7, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2766)); if (unlikely(!__pyx_t_130)) __PYX_ERR(0, 7745, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_130);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_130, __pyx_t_132);
  __Pyx_DECREF(__pyx_t_132); __pyx_t_132 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_accept, __pyx_t_130) < 0) __PYX_ERR(0, 7745, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_130); __pyx_t_130 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7724
 *         return localctx
 * 
 *     class AnnotationContext(ParserRuleContext):             # <<<<<<<<<<<<<<
 * 
 *         def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
 */
  __pyx_t_130 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_AnnotationContext, __pyx_t_7, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_130)) __PYX_ERR(0, 7724, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_130);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_131, __pyx_t_130) < 0) __PYX_ERR(0, 7724, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_131); __pyx_t_131 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_AnnotationContext, __pyx_t_130) < 0) __PYX_ERR(0, 7724, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_130); __pyx_t_130 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7754
 * 
 * 
 *     def annotation(self):             # <<<<<<<<<<<<<<
 * 
 *         localctx = ModelicaParser.AnnotationContext(self, self._ctx, self.state)
 */
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_147annotation, 0, __pyx_n_s_ModelicaParser_annotation, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2768)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7754, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_annotation, __pyx_t_7) < 0) __PYX_ERR(0, 7754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7774
 * 
 * 
 *     def sempred(self, localctx:RuleContext, ruleIndex:int, predIndex:int):             # <<<<<<<<<<<<<<
 *         if self._predicates == None:
 *             self._predicates = dict()
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7774, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_RuleContext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7774, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_localctx, __pyx_t_5) < 0) __PYX_ERR(0, 7774, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ruleIndex, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 7774, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_predIndex, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 7774, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_149sempred, 0, __pyx_n_s_ModelicaParser_sempred, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2770)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7774, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_sempred, __pyx_t_5) < 0) __PYX_ERR(0, 7774, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pymola/generated/ModelicaParser.pyx":7784
 *             return pred(localctx, predIndex)
 * 
 *     def expr_sempred(self, localctx:ExprContext, predIndex:int):             # <<<<<<<<<<<<<<
 *             if predIndex == 0:
 *                 return self.precpred(self._ctx, 7)
 */
  __pyx_t_5 = PyDict_New(); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 7784, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyObject_GetItem(__pyx_t_3, __pyx_n_s_ExprContext);
  if (unlikely(!__pyx_t_7)) {
    PyErr_Clear();
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ExprContext);
  }
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7784, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_localctx, __pyx_t_7) < 0) __PYX_ERR(0, 7784, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_predIndex, ((PyObject *)(&PyInt_Type))) < 0) __PYX_ERR(0, 7784, __pyx_L1_error)
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_14ModelicaParser_14ModelicaParser_151expr_sempred, 0, __pyx_n_s_ModelicaParser_expr_sempred, NULL, __pyx_n_s_pymola_generated_ModelicaParser, __pyx_d, ((PyObject *)__pyx_codeobj__2772)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 7784, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyObject_SetItem(__pyx_t_3, __pyx_n_s_expr_sempred, __pyx_t_7) < 0) __PYX_ERR(0, 7784, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaParser.pyx":532
 * 
 * 
 * class ModelicaParser ( Parser ):             # <<<<<<<<<<<<<<
 * 
 *     grammarFileName = "Modelica.g4"
 */
  __pyx_t_7 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_ModelicaParser, __pyx_t_1, __pyx_t_3, NULL, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_4, __pyx_t_7) < 0) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ModelicaParser, __pyx_t_7) < 0) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaParser.pyx":1
 * # Generated from Modelica.g4 by ANTLR 4.7             # <<<<<<<<<<<<<<
 * # encoding: utf-8
 * from antlr4 import *
 */
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_XDECREF(__pyx_t_21);
  __Pyx_XDECREF(__pyx_t_22);
  __Pyx_XDECREF(__pyx_t_23);
  __Pyx_XDECREF(__pyx_t_24);
  __Pyx_XDECREF(__pyx_t_25);
  __Pyx_XDECREF(__pyx_t_26);
  __Pyx_XDECREF(__pyx_t_27);
  __Pyx_XDECREF(__pyx_t_28);
  __Pyx_XDECREF(__pyx_t_29);
  __Pyx_XDECREF(__pyx_t_30);
  __Pyx_XDECREF(__pyx_t_31);
  __Pyx_XDECREF(__pyx_t_32);
  __Pyx_XDECREF(__pyx_t_33);
  __Pyx_XDECREF(__pyx_t_34);
  __Pyx_XDECREF(__pyx_t_35);
  __Pyx_XDECREF(__pyx_t_36);
  __Pyx_XDECREF(__pyx_t_37);
  __Pyx_XDECREF(__pyx_t_38);
  __Pyx_XDECREF(__pyx_t_39);
  __Pyx_XDECREF(__pyx_t_40);
  __Pyx_XDECREF(__pyx_t_41);
  __Pyx_XDECREF(__pyx_t_42);
  __Pyx_XDECREF(__pyx_t_43);
  __Pyx_XDECREF(__pyx_t_44);
  __Pyx_XDECREF(__pyx_t_45);
  __Pyx_XDECREF(__pyx_t_46);
  __Pyx_XDECREF(__pyx_t_47);
  __Pyx_XDECREF(__pyx_t_48);
  __Pyx_XDECREF(__pyx_t_49);
  __Pyx_XDECREF(__pyx_t_50);
  __Pyx_XDECREF(__pyx_t_51);
  __Pyx_XDECREF(__pyx_t_52);
  __Pyx_XDECREF(__pyx_t_53);
  __Pyx_XDECREF(__pyx_t_54);
  __Pyx_XDECREF(__pyx_t_55);
  __Pyx_XDECREF(__pyx_t_56);
  __Pyx_XDECREF(__pyx_t_57);
  __Pyx_XDECREF(__pyx_t_58);
  __Pyx_XDECREF(__pyx_t_59);
  __Pyx_XDECREF(__pyx_t_60);
  __Pyx_XDECREF(__pyx_t_61);
  __Pyx_XDECREF(__pyx_t_62);
  __Pyx_XDECREF(__pyx_t_63);
  __Pyx_XDECREF(__pyx_t_64);
  __Pyx_XDECREF(__pyx_t_65);
  __Pyx_XDECREF(__pyx_t_66);
  __Pyx_XDECREF(__pyx_t_67);
  __Pyx_XDECREF(__pyx_t_68);
  __Pyx_XDECREF(__pyx_t_69);
  __Pyx_XDECREF(__pyx_t_70);
  __Pyx_XDECREF(__pyx_t_71);
  __Pyx_XDECREF(__pyx_t_72);
  __Pyx_XDECREF(__pyx_t_73);
  __Pyx_XDECREF(__pyx_t_74);
  __Pyx_XDECREF(__pyx_t_75);
  __Pyx_XDECREF(__pyx_t_76);
  __Pyx_XDECREF(__pyx_t_77);
  __Pyx_XDECREF(__pyx_t_78);
  __Pyx_XDECREF(__pyx_t_79);
  __Pyx_XDECREF(__pyx_t_80);
  __Pyx_XDECREF(__pyx_t_81);
  __Pyx_XDECREF(__pyx_t_82);
  __Pyx_XDECREF(__pyx_t_83);
  __Pyx_XDECREF(__pyx_t_84);
  __Pyx_XDECREF(__pyx_t_85);
  __Pyx_XDECREF(__pyx_t_86);
  __Pyx_XDECREF(__pyx_t_87);
  __Pyx_XDECREF(__pyx_t_88);
  __Pyx_XDECREF(__pyx_t_89);
  __Pyx_XDECREF(__pyx_t_90);
  __Pyx_XDECREF(__pyx_t_91);
  __Pyx_XDECREF(__pyx_t_92);
  __Pyx_XDECREF(__pyx_t_93);
  __Pyx_XDECREF(__pyx_t_94);
  __Pyx_XDECREF(__pyx_t_95);
  __Pyx_XDECREF(__pyx_t_96);
  __Pyx_XDECREF(__pyx_t_97);
  __Pyx_XDECREF(__pyx_t_98);
  __Pyx_XDECREF(__pyx_t_99);
  __Pyx_XDECREF(__pyx_t_100);
  __Pyx_XDECREF(__pyx_t_101);
  __Pyx_XDECREF(__pyx_t_102);
  __Pyx_XDECREF(__pyx_t_103);
  __Pyx_XDECREF(__pyx_t_104);
  __Pyx_XDECREF(__pyx_t_105);
  __Pyx_XDECREF(__pyx_t_106);
  __Pyx_XDECREF(__pyx_t_107);
  __Pyx_XDECREF(__pyx_t_108);
  __Pyx_XDECREF(__pyx_t_109);
  __Pyx_XDECREF(__pyx_t_110);
  __Pyx_XDECREF(__pyx_t_111);
  __Pyx_XDECREF(__pyx_t_112);
  __Pyx_XDECREF(__pyx_t_113);
  __Pyx_XDECREF(__pyx_t_114);
  __Pyx_XDECREF(__pyx_t_115);
  __Pyx_XDECREF(__pyx_t_116);
  __Pyx_XDECREF(__pyx_t_117);
  __Pyx_XDECREF(__pyx_t_118);
  __Pyx_XDECREF(__pyx_t_119);
  __Pyx_XDECREF(__pyx_t_120);
  __Pyx_XDECREF(__pyx_t_121);
  __Pyx_XDECREF(__pyx_t_122);
  __Pyx_XDECREF(__pyx_t_123);
  __Pyx_XDECREF(__pyx_t_124);
  __Pyx_XDECREF(__pyx_t_125);
  __Pyx_XDECREF(__pyx_t_126);
  __Pyx_XDECREF(__pyx_t_127);
  __Pyx_XDECREF(__pyx_t_128);
  __Pyx_XDECREF(__pyx_t_129);
  __Pyx_XDECREF(__pyx_t_130);
  __Pyx_XDECREF(__pyx_t_131);
  __Pyx_XDECREF(__pyx_t_132);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init pymola.generated.ModelicaParser", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_DECREF(__pyx_m); __pyx_m = 0;
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init pymola.generated.ModelicaParser");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if PY_MAJOR_VERSION < 3
  return;
  #else
  return __pyx_m;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule((char *)modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, (char *)"RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* GetModuleGlobalName */
static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name) {
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
    result = PyDict_GetItem(__pyx_d, name);
    if (likely(result)) {
        Py_INCREF(result);
    } else {
#else
    result = PyObject_GetItem(__pyx_d, name);
    if (!result) {
        PyErr_Clear();
#endif
        result = __Pyx_GetBuiltinName(name);
    }
    return result;
}

/* PyCFunctionFastCall */
  #if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    return (*((__Pyx_PyCFunctionFast)meth)) (self, args, nargs, NULL);
}
#endif  // CYTHON_FAST_PYCCALL

/* PyFunctionFastCall */
  #if CYTHON_FAST_PYCALL
#include "frameobject.h"
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = PyThreadState_GET();
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = f->f_localsplus;
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif  // CPython < 3.6
#endif  // CYTHON_FAST_PYCALL

/* PyObjectCall */
  #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
  #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
  #if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || PyObject_TypeCheck(func, __pyx_CyFunctionType))) {
#else
    if (likely(PyCFunction_Check(func))) {
#endif
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyObjectCallNoArg */
    #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || PyObject_TypeCheck(func, __pyx_CyFunctionType))) {
#else
    if (likely(PyCFunction_Check(func))) {
#endif
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* SaveResetException */
      #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* GetException */
      #if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb) {
#endif
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* PyErrFetchRestore */
        #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseArgTupleInvalid */
        static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
        static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
        static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* PyIntBinop */
        #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED int inplace) {
    if (op1 == op2) {
        Py_RETURN_TRUE;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        if (a == b) {
            Py_RETURN_TRUE;
        } else {
            Py_RETURN_FALSE;
        }
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a;
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                #if PyLong_SHIFT < 30 && PyLong_SHIFT != 15
                default: return PyLong_Type.tp_richcompare(op1, op2, Py_EQ);
                #else
                default: Py_RETURN_FALSE;
                #endif
            }
        }
            if (a == b) {
                Py_RETURN_TRUE;
            } else {
                Py_RETURN_FALSE;
            }
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            if ((double)a == (double)b) {
                Py_RETURN_TRUE;
            } else {
                Py_RETURN_FALSE;
            }
    }
    return PyObject_RichCompare(op1, op2, Py_EQ);
}
#endif

/* PyIntBinop */
        #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED int inplace) {
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a - b);
            if (likely((x^a) >= 0 || (x^~b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                default: return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
            }
        }
                x = a - b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla - llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("subtract", return NULL)
            result = ((double)a) - (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceSubtract : PyNumber_Subtract)(op1, op2);
}
#endif

/* PyErrExceptionMatches */
        #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    return PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* SwapException */
        #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* RaiseException */
        #if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
#if PY_VERSION_HEX >= 0x03030000
    if (cause) {
#else
    if (cause && cause != Py_None) {
#endif
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = PyThreadState_GET();
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* Import */
          static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_VERSION_HEX < 0x03030000
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                #if PY_VERSION_HEX < 0x03030000
                PyObject *py_level = PyInt_FromLong(1);
                if (!py_level)
                    goto bad;
                module = PyObject_CallFunctionObjArgs(py_import,
                    name, global_dict, empty_dict, list, py_level, NULL);
                Py_DECREF(py_level);
                #else
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                #endif
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_VERSION_HEX < 0x03030000
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_VERSION_HEX < 0x03030000
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
          static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* SetItemInt */
          static CYTHON_INLINE int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v) {
    int r;
    if (!j) return -1;
    r = PyObject_SetItem(o, j, v);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v, int is_list,
                                               CYTHON_NCP_UNUSED int wraparound, CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = (!wraparound) ? i : ((likely(i >= 0)) ? i : i + PyList_GET_SIZE(o));
        if ((!boundscheck) || likely((n >= 0) & (n < PyList_GET_SIZE(o)))) {
            PyObject* old = PyList_GET_ITEM(o, n);
            Py_INCREF(v);
            PyList_SET_ITEM(o, n, v);
            Py_DECREF(old);
            return 1;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_ass_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return -1;
                    PyErr_Clear();
                }
            }
            return m->sq_ass_item(o, i, v);
        }
    }
#else
#if CYTHON_COMPILING_IN_PYPY
    if (is_list || (PySequence_Check(o) && !PyDict_Check(o))) {
#else
    if (is_list || PySequence_Check(o)) {
#endif
        return PySequence_SetItem(o, i, v);
    }
#endif
    return __Pyx_SetItemInt_Generic(o, PyInt_FromSsize_t(i), v);
}

/* CalculateMetaclass */
            static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases) {
    Py_ssize_t i, nbases = PyTuple_GET_SIZE(bases);
    for (i=0; i < nbases; i++) {
        PyTypeObject *tmptype;
        PyObject *tmp = PyTuple_GET_ITEM(bases, i);
        tmptype = Py_TYPE(tmp);
#if PY_MAJOR_VERSION < 3
        if (tmptype == &PyClass_Type)
            continue;
#endif
        if (!metaclass) {
            metaclass = tmptype;
            continue;
        }
        if (PyType_IsSubtype(metaclass, tmptype))
            continue;
        if (PyType_IsSubtype(tmptype, metaclass)) {
            metaclass = tmptype;
            continue;
        }
        PyErr_SetString(PyExc_TypeError,
                        "metaclass conflict: "
                        "the metaclass of a derived class "
                        "must be a (non-strict) subclass "
                        "of the metaclasses of all its bases");
        return NULL;
    }
    if (!metaclass) {
#if PY_MAJOR_VERSION < 3
        metaclass = &PyClass_Type;
#else
        metaclass = &PyType_Type;
#endif
    }
    Py_INCREF((PyObject*) metaclass);
    return (PyObject*) metaclass;
}

/* PyIntBinop */
            #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED int inplace) {
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* FetchCommonType */
            static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CythonFunction */
            static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(__pyx_CyFunctionObject, func.m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromString(m->func.m_ml->ml_name);
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_New(PyTypeObject *type, PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    __pyx_CyFunctionObject *op = PyObject_GC_New(__pyx_CyFunctionObject, type);
    if (op == NULL)
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    PyObject_GC_Track(op);
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0 = PySequence_ITEM(arg, 0);
                if (unlikely(!arg0)) return NULL;
                result = (*meth)(self, arg0);
                Py_DECREF(arg0);
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    return __Pyx_CyFunction_CallMethod(func, ((PyCFunctionObject*)func)->m_self, arg, kw);
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
        argc = PyTuple_GET_SIZE(args);
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (__pyx_CyFunctionType == NULL) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (!m->defaults)
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* Py3ClassCreate */
                static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name,
                                           PyObject *qualname, PyObject *mkw, PyObject *modname, PyObject *doc) {
    PyObject *ns;
    if (metaclass) {
        PyObject *prep = __Pyx_PyObject_GetAttrStr(metaclass, __pyx_n_s_prepare);
        if (prep) {
            PyObject *pargs = PyTuple_Pack(2, name, bases);
            if (unlikely(!pargs)) {
                Py_DECREF(prep);
                return NULL;
            }
            ns = PyObject_Call(prep, pargs, mkw);
            Py_DECREF(prep);
            Py_DECREF(pargs);
        } else {
            if (unlikely(!PyErr_ExceptionMatches(PyExc_AttributeError)))
                return NULL;
            PyErr_Clear();
            ns = PyDict_New();
        }
    } else {
        ns = PyDict_New();
    }
    if (unlikely(!ns))
        return NULL;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_module, modname) < 0)) goto bad;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_qualname, qualname) < 0)) goto bad;
    if (unlikely(doc && PyObject_SetItem(ns, __pyx_n_s_doc, doc) < 0)) goto bad;
    return ns;
bad:
    Py_DECREF(ns);
    return NULL;
}
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases,
                                      PyObject *dict, PyObject *mkw,
                                      int calculate_metaclass, int allow_py2_metaclass) {
    PyObject *result, *margs;
    PyObject *owned_metaclass = NULL;
    if (allow_py2_metaclass) {
        owned_metaclass = PyObject_GetItem(dict, __pyx_n_s_metaclass);
        if (owned_metaclass) {
            metaclass = owned_metaclass;
        } else if (likely(PyErr_ExceptionMatches(PyExc_KeyError))) {
            PyErr_Clear();
        } else {
            return NULL;
        }
    }
    if (calculate_metaclass && (!metaclass || PyType_Check(metaclass))) {
        metaclass = __Pyx_CalculateMetaclass((PyTypeObject*) metaclass, bases);
        Py_XDECREF(owned_metaclass);
        if (unlikely(!metaclass))
            return NULL;
        owned_metaclass = metaclass;
    }
    margs = PyTuple_Pack(3, name, bases, dict);
    if (unlikely(!margs)) {
        result = NULL;
    } else {
        result = PyObject_Call(metaclass, margs, mkw);
        Py_DECREF(margs);
    }
    Py_XDECREF(owned_metaclass);
    return result;
}

/* CodeObjectCache */
                static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
                #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    py_code = __pyx_find_code_object(c_line ? c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        PyThreadState_GET(), /*PyThreadState *tstate,*/
        py_code,             /*PyCodeObject *code,*/
        __pyx_d,      /*PyObject *globals,*/
        0                    /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntToPy */
                static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CyFunctionClassCell */
                static CYTHON_INLINE int __Pyx_CyFunction_InitClassCell(PyObject *cyfunctions, PyObject *classobj) {
    Py_ssize_t i, count = PyList_GET_SIZE(cyfunctions);
    for (i = 0; i < count; i++) {
        __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *)
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            PyList_GET_ITEM(cyfunctions, i);
#else
            PySequence_ITEM(cyfunctions, i);
        if (unlikely(!m))
            return -1;
#endif
        Py_INCREF(classobj);
        m->func_classobj = classobj;
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
        Py_DECREF((PyObject*)m);
#endif
    }
    return 0;
}

/* CIntFromPyVerify */
                #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntFromPy */
                static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
                static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CStringEquals */
                static CYTHON_INLINE int __Pyx_StrEq(const char *s1, const char *s2) {
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    return *s1 == *s2;
}

/* CheckBinaryVersion */
                static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* InitStrings */
                static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if CYTHON_COMPILING_IN_CPYTHON && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
#if PY_VERSION_HEX < 0x03030000
        char* defenc_c;
        PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
        if (!defenc) return NULL;
        defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        {
            char* end = defenc_c + PyBytes_GET_SIZE(defenc);
            char* c;
            for (c = defenc_c; c < end; c++) {
                if ((unsigned char) (*c) >= 128) {
                    PyUnicode_AsASCIIString(o);
                    return NULL;
                }
            }
        }
#endif
        *length = PyBytes_GET_SIZE(defenc);
        return defenc_c;
#else
        if (__Pyx_PyUnicode_READY(o) == -1) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        if (PyUnicode_IS_ASCII(o)) {
            *length = PyUnicode_GET_LENGTH(o);
            return PyUnicode_AsUTF8(o);
        } else {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
#else
        return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (PyInt_Check(x) || PyLong_Check(x))
#else
  if (PyLong_Check(x))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = PyNumber_Long(x);
  }
  #else
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Long(x);
  }
  #endif
#else
  res = PyNumber_Int(x);
#endif
  if (res) {
#if PY_MAJOR_VERSION < 3
    if (!PyInt_Check(res) && !PyLong_Check(res)) {
#else
    if (!PyLong_Check(res)) {
#endif
      PyErr_Format(PyExc_TypeError,
                   "__%.4s__ returned non-%.4s (type %.200s)",
                   name, name, Py_TYPE(res)->tp_name);
      Py_DECREF(res);
      return NULL;
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(x);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
